//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.gBase.*;

import java.util.Collections;

import ARoad0.ErrB.BaseError;

/**
 * This class is responsible for creating a MySQL as an access control system (ACS).
 * The requirements for the MySQL access controls are described
 * in http://dev.mysql.com/doc/refman/5.6/en/request-access.html and the near pages.
 * Almost all the work is done by the super class, and this class has just to use
 * the right methods of the superclass.
 * The limitations on the MySQL server simulation are described in the MySQL gBase
 * documentation. <P>
 * Called by gDMak.ActionNewACSyst and MySQL.gWork.AcsFactory.MySQL.
 * This class does not use for its GUI any BeanInfo.
 * It is a javabean without specific bound properties.
 * @see ARoad0.AcsAddon.Accbee.MySQL.gBaseBeanInfo.ACSMySQLImplBeanInfo
 */
public class ACSFactoryMySQLImpl extends ACSFactoryImpl {
  private static final int    PRIME = 1000003;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  static {
      EMPTY_INSTANCE = new ACSFactoryMySQLImpl();
  }


 /**
  * Default constructor.
  * Calls ACSFactoryImpl.addAcsVocabulary() to define the AcsAddon classes for MySQL, including
  * the classes BasesSetMySQLImpl and ProxyMySQLImpl.
  * Sets the current ACS addon version.
  */
  public ACSFactoryMySQLImpl() {

       super();
       try {
            // for an AcsAddon ACS,
            // the ACS vocabulary must be set before the call to
            // AssociateUtilityImpl.controlAssociations()
            // that factory.addAssociationsToNewObject(), addXXXroots() perform.

            // no changes on ACSFactoryUtilityImpl.ACSADDON_EDITOR, ACSADDON_PACKAGE_EDITOR

           addAcsVocabulary(
                ACSFactoryUtilityImpl.ACSADDON_NAME, new String[] {
                    "MySQL"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.ACSFACTORY_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.ACSFactoryMySQLImpl"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.ACS_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.ACSMySQLImpl"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.RESOURCE_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.ResourceMySQLImpl"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.ACTOR_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.ActorMySQLImpl"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.DIRECTORY_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.USER_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.UserIDMySQLImpl"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.GROUP_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.GroupIDMySQLImpl"});
           addAcsVocabulary(
                ACSFactoryUtilityImpl.VIRTUALFOLDER_CLASS, new String[] {
                    "ARoad0.AcsAddon.Accbee.MySQL.gBase.VirtualFolderMySQLImpl"});

           // the extension classes
           addAcsVocabulary(
                ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES, new String[] {
                    "BasesSet/is/Directory",
                    "Proxy/is/UserID"
                    //"Role/is/GroupID"
           });

           addAcsVocabulary(
                ACSFactoryUtilityImpl.ACSADDON_VERSION, new String[] {"0.7.2"});

       }
       catch (BaseError b) {
           throw new InternalError(b.toString());
       }
  } // constructor

  /**
   * Called by MySQL.gWork.AcsFactoryMySQL.
   * Calls addAcsVocabulary() to define the AcsAddon classes for MySQL, including
   * the classes BasesSetMySQLImpl and ProxyMySQLImpl.
   * @param _iSname the name of the IS containing this new ACS. May be null, and
   * ISFactoryUtilityImpl.DEFAULT_IS_NAME is then set as IS name and it is created
   * BaseManagerImpl.
   * Sets the current ACS addon version.
   * @param _name the name of the new ACS. Never null.
   * @param _ACStype the ACS type. Never null.
   * @param _parentName the parent of this new ACS. No use if _iSname is null.
   * Otherwise, it has to contain the IS name.
   * @exception BaseError if _type is null, _parentName does not start with _iSname.
   */
  public ACSFactoryMySQLImpl(ImmutableName _iSname, ImmutableName _name,
          String _ACStype, ImmutableName _parentName) throws BaseError {

       super(_iSname, _name, _ACStype, _parentName);

        // for an AcsAddon ACS,
        // the ACS vocabulary must be set before the call to
        // AssociateUtilityImpl.controlAssociations()
        // that factory.addAssociationsToNewObject(), addXXXroots() perform.
       
       // no changes on ACSFactoryUtilityImpl.ACSADDON_EDITOR, ACSADDON_PACKAGE_EDITOR

       addAcsVocabulary(
            ACSFactoryUtilityImpl.ACSADDON_NAME, new String[] {
                "MySQL"});
       addAcsVocabulary(
            ACSFactoryUtilityImpl.ACSFACTORY_CLASS, new String[] {
                "ARoad0.AcsAddon.Accbee.MySQL.gBase.ACSFactoryMySQLImpl"});
       //addAcsVocabulary(
         //   ACSFactoryUtilityImpl.ACS_CLASS, new String[] {
           //     "ARoad0.gBase.ACSImpl"});
       addAcsVocabulary(
            ACSFactoryUtilityImpl.RESOURCE_CLASS, new String[] {
                "ARoad0.AcsAddon.Accbee.MySQL.gBase.ResourceMySQLImpl"});
       addAcsVocabulary(
            ACSFactoryUtilityImpl.ACTOR_CLASS, new String[] {
                "ARoad0.AcsAddon.Accbee.MySQL.gBase.ActorMySQLImpl"});
       addAcsVocabulary(
            ACSFactoryUtilityImpl.DIRECTORY_CLASS, new String[] {
                "ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl"});
       addAcsVocabulary(
            ACSFactoryUtilityImpl.USER_CLASS, new String[] {
                "ARoad0.AcsAddon.Accbee.MySQL.gBase.UserIDMySQLImpl"});
       addAcsVocabulary(
            ACSFactoryUtilityImpl.GROUP_CLASS, new String[] {
                "ARoad0.AcsAddon.Accbee.MySQL.gBase.GroupIDMySQLImpl"});
       addAcsVocabulary(
            ACSFactoryUtilityImpl.VIRTUALFOLDER_CLASS, new String[] {
                "ARoad0.AcsAddon.Accbee.MySQL.gBase.VirtualFolderMySQLImpl"});

       // the extension classes
       addAcsVocabulary(
            ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES, new String[] {
                "BasesSet/is/Directory",
                "Proxy/is/UserID"
                //"Role/is/GroupID"
       });

       addAcsVocabulary(
            ACSFactoryUtilityImpl.ACSADDON_VERSION, new String[] {"0.7.2"});
  } // constructor

  /**
   * Initializes the ACS constraints maps before the call to ACSImpl.valueOf().
   * Adds to ACSFactoryImpl.l_ImmutableRules_ some specific rules: <P>
   * - ACL right: ACL for actors only as targets <P>
   * This AcsAddon method adds also an informal description of the right selection algorithm.
   * An ACSFactoryMySQLImpl for a new ACS derived from a MySQL ACS
   * has already all these values.
   * Called by createACS(). Overrides and calls first the super method.
   */
  public synchronized void initializeConstraintsBeforeCreation() {

      super.initializeConstraintsBeforeCreation();

      if (l_ImmutableRules_.contains("ACL right: ACL for actors only as targets")) {
          l_ImmutableRules_.remove(
                  "ACL right: ACL for actors only as targets");
      }

      Collections.sort(l_ImmutableRules_);

      // informal description
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    END OF THE FORMAL ACCESS ROAD RULES.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    ------- THE SPECIFIC RULES FOR THE MySQL SERVER -------------------");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    HOW THE STRUCTURE IS MANAGED -----------------");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    'MySQL server' is the root of the structure of Resources.");
      l_ImmutableRules_.add("    Its children are databases or bases sets. Only databases have children.");
      l_ImmutableRules_.add("    They are tables, view delegates, stored procedures, stored functions");
      l_ImmutableRules_.add("    and triggers. Only tables have children, and they are columns");
      l_ImmutableRules_.add("    and indexes. Only a view delegate has an alias, and it is a view.");
      l_ImmutableRules_.add("    The databases and the bases sets are called 'database-related targets',");
      l_ImmutableRules_.add("    because all the AclEntries of the  bases set are copied");
      l_ImmutableRules_.add("    into the direct ACL list of every database that matches");
      l_ImmutableRules_.add("    the bases set name. For instance, the bases set named 'on%'");
      l_ImmutableRules_.add("    matches the database named 'one', and all the 'on%' direct ACL");
      l_ImmutableRules_.add("    are then copied into the direct ACL list of the database 'one'.");
      l_ImmutableRules_.add("    These terms define the main types of the MySQL ACS access targets.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    An user in MySQL is simulated by an UserID having the type 'account'.");
      l_ImmutableRules_.add("    A simple host in MySQL is simulated by a 'host' GroupID.");
      l_ImmutableRules_.add("    In MySQL, it is possible to set with one statement like 'tom@%'");
      l_ImmutableRules_.add("    some rights for all the accounts 'tom' whatever their current hosts.");
      l_ImmutableRules_.add("    From the security point of view, this feature is not acceptable,");
      l_ImmutableRules_.add("    because there is no way to state it is always the same 'tom'.");
      l_ImmutableRules_.add("    This is why Access Road allows to create the accounts 'tom', 'tom(2)'");
      l_ImmutableRules_.add("    and tom(A) from 3 differents hosts, for setting their common rights");
      l_ImmutableRules_.add("    through a group named 'tom§@%', in this example of MySQL statement.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    The group 'localhost' models the MySQL server host.");
      l_ImmutableRules_.add("    Otherwise, a new host has to model an open ACS.");
      l_ImmutableRules_.add("    The name last component of a new host must have the form");
      l_ImmutableRules_.add("    'xxx:yy:zzz:tttt', and it models then the ACS");
      l_ImmutableRules_.add("    ' xxx:: yy:: zzz:: tttt::  '.");
      l_ImmutableRules_.add("    There is exactly one simple host per account, at the creation");
      l_ImmutableRules_.add("    of the account, and this first host cannot be changed.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    A pattern-name group is created by the user to contains");
      l_ImmutableRules_.add("    all the accounts or groups having a given pattern in their names.");
      l_ImmutableRules_.add("    Its name contains one of the two '%' and '_' wildcard characters:");
      l_ImmutableRules_.add("    - '%' matches any number of characters, even zero characters,");
      l_ImmutableRules_.add("    - '_' matches exactly one character.");
      l_ImmutableRules_.add("    While MySQL Server allows to have two wildcards, or more,");
      l_ImmutableRules_.add("    in the same name of a pattern-name group, Access Road");
      l_ImmutableRules_.add("    does not permit it, because the algorithm sorting for these names");
      l_ImmutableRules_.add("    is not specified and quite confusing.");
      l_ImmutableRules_.add("    An IP wildcard value can match only IP addresses,");
      l_ImmutableRules_.add("    and not host names. '1.2.example.com' never matches '1.2.%'.");
      l_ImmutableRules_.add("    The MySQL ACS addon ensures that accounts or groups are members");
      l_ImmutableRules_.add("    of the pattern-name groups that match to them.");
      l_ImmutableRules_.add("    Hosts groups and accounts-for-hosts groups are the two types of ");
      l_ImmutableRules_.add("    pattern-name groups. They are defined hereinafter:");
      l_ImmutableRules_.add("    (1) A pattern-name hosts group may be created by the Access Road");
      l_ImmutableRules_.add("    user to contain all the hosts having a given pattern in their names.");
      l_ImmutableRules_.add("    '1.2.%' is an example of hosts group nick name.");
      l_ImmutableRules_.add("    A hosts group may have ACL only for the database level.");
      l_ImmutableRules_.add("    (2) A pattern-name accounts-for-hosts group may be created");
      l_ImmutableRules_.add("    by the Access Road user to contain all the accounts");
      l_ImmutableRules_.add("    having a given pattern, like 'tom§@1.2.%' for 'tom' and 'tom(2)'.");
      l_ImmutableRules_.add("    The pattern-name accounts-for-hosts group nick name");
      l_ImmutableRules_.add("    has the form 'nnnn§@YYYY', where 'YYYY' contain a wildcard character,");
      l_ImmutableRules_.add("    and 'nnnn' is the nick name of an account, like 'tom'.");
      l_ImmutableRules_.add("    In the example, 'tom' and 'tom(2)' are set by Access Road");
      l_ImmutableRules_.add("    as members of the group 'tom§@1.2.%'. This is proper to Access Road,");
      l_ImmutableRules_.add("    since MySQL would use 'tom@%' in the privilege statement.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    A proxy is an UserID that uses an proxied account to access.");
      l_ImmutableRules_.add("    A proxy has no own rights. Any account may have several proxies.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    A simple host group in MySQL has only accounts and proxies as members.");
      l_ImmutableRules_.add("    Accounts and simple host memberships are created by the user.");
      l_ImmutableRules_.add("    A hostS group has only simple hosts as members.");
      l_ImmutableRules_.add("    An accounts-for-hosts group has only accounts, not proxies, as members.");
      l_ImmutableRules_.add("    From their names, Access Road sets the members of these two groups.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    HOW THE RIGHTS ARE MANAGED -------------------");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    MySQL server combines the information in the various grant tables");
      l_ImmutableRules_.add("    to form a complete description of a user's MySQL privileges.");
      l_ImmutableRules_.add("    Into Access Road, the effective access rights on a MySQL resource");
      l_ImmutableRules_.add("    are selected by a two-steps process based on AclEntries (or ACLs).");
      l_ImmutableRules_.add("    The Access Road AclEntries simulate the MySQL privileges.");
      l_ImmutableRules_.add("    The Access Road AclEntry inheritance simulate the inheritance");
      l_ImmutableRules_.add("    of the privileges among the MySQL objects, from the server to");
      l_ImmutableRules_.add("    a table column, for instance.");
      l_ImmutableRules_.add("    For a MySQL user/host, the user is always member of one simple host.");
      l_ImmutableRules_.add("    Typically, the user may be the source of a conditional ACL,");
      l_ImmutableRules_.add("    and the simple host is the single condition group of the ACL.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    The effective rights of an user on a resource");
      l_ImmutableRules_.add("    are selected by a two-steps algorithm, with the global rights first.");
      l_ImmutableRules_.add("    At the FIRST step of access control, only the AclEntries");
      l_ImmutableRules_.add("    of the root 'MySQL server' are implied, to select the first pair");
      l_ImmutableRules_.add("    (user2, simple host or accounts-for-hosts group)");
      l_ImmutableRules_.add("    that (user, host) matches.");
      l_ImmutableRules_.add("    This selected pair is called by MySQL the CURRENT_USER.");
      l_ImmutableRules_.add("    'user2' is 'user' or the 'anonymous' account simulating '*' or ''.");
      l_ImmutableRules_.add("    Matching an accounts-for-hosts group means 'user' is a member.");
      l_ImmutableRules_.add("    For instance, all the effective rights of (jerry, 'localhost')");
      l_ImmutableRules_.add("    may be provided by the current user ('jerry', '%').");
      l_ImmutableRules_.add("    No selected pair means no access to the MySQL server.");
      l_ImmutableRules_.add("    At each step, only the rights of the current user matter,");
      l_ImmutableRules_.add("    except for views or routines in the DEFINER mode we will see later.");
      l_ImmutableRules_.add("    The direct AclEntries of the root 'MySQL server' may be:");
      l_ImmutableRules_.add("    - conditional user/simple host ACL on the root");
      l_ImmutableRules_.add("    EXCLUSIVE OR (the first matched ACL is selected)");
      l_ImmutableRules_.add("    - non-conditional 'accounts-for-hosts group' ACL on this root.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    For a given user, there is at most one user/host ACL and several");
      l_ImmutableRules_.add("    'accounts-for-hosts group' ACL on the root, and this is also true");
      l_ImmutableRules_.add("    for the other ACL targets under the root.");
      l_ImmutableRules_.add("    If a right is granted at this first step (at least the 'USAGE' right)");
      l_ImmutableRules_.add("    complementary rights may be added, for the SAME current user,");
      l_ImmutableRules_.add("    at each parent level from the database to the direct parent,");
      l_ImmutableRules_.add("    then other rights may be added from the direct ACL of the resource.");
      l_ImmutableRules_.add("    To model this behavior, all the direct AclEntries of each parent");
      l_ImmutableRules_.add("    are simply put into the inherited ACL list of the resource.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    At the root, at every parent level, and for the resource itself,");
      l_ImmutableRules_.add("    and from a parsing of the SORTED list of ACL,");
      l_ImmutableRules_.add("    the FIRST ACL that matches the current user, if any,");
      l_ImmutableRules_.add("    (or sometimes the first couple of ACL, but only for a database)");
      l_ImmutableRules_.add("    defines the rights to add to the result from this parent level.");
      l_ImmutableRules_.add("    This is why the sorting of the AclEntries is so important.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    At the SECOND step of access control, remember that the databases");
      l_ImmutableRules_.add("    and the bases sets are called 'database-related targets'.");
      l_ImmutableRules_.add("    They are at the second parent level, after the root 'MySQL server'.");
      l_ImmutableRules_.add("    Some rights from a database-related target may then be added");
      l_ImmutableRules_.add("    at the SECOND step, following this algorithm:");
      l_ImmutableRules_.add("    - conditional user/simple host ACL");
      l_ImmutableRules_.add("    EXCLUSIVE OR");
      l_ImmutableRules_.add("    - non-conditional 'accounts-for-hosts group' ACL");
      l_ImmutableRules_.add("    EXCLUSIVE OR");
      l_ImmutableRules_.add("    (non-conditional user ACL -------- AND ----------");
      l_ImmutableRules_.add("    simple host or hosts group non-conditional ACL).");
      l_ImmutableRules_.add("    These two previous ACL are called to-be-mixed ACL,");
      l_ImmutableRules_.add("    because they have to be combined to produce some rights, if any.");
      l_ImmutableRules_.add("    Due to the bases set and the to-be-mixed ACL, the level");
      l_ImmutableRules_.add("    of the databases is the most complex one.");
      l_ImmutableRules_.add("    Finally, the SECOND step is ended by a parsing on the children");
      l_ImmutableRules_.add("    of the database, if any, for detecting the matched ACL");
      l_ImmutableRules_.add("    at each parent level, and for the current user.");
      l_ImmutableRules_.add("    These rules are the same rules the root 'MySQL server' applies:");
      l_ImmutableRules_.add("    - conditional user/simple host ACL on the table/view delegate/");
      l_ImmutableRules_.add("    stored procedure/stored function,");
      l_ImmutableRules_.add("    EXCLUSIVE OR (the first matched ACL is selected)");
      l_ImmutableRules_.add("    - non-conditional 'accounts-for-hosts group' ACL on the table/");
      l_ImmutableRules_.add("    view delegate/stored procedure/stored function,");
      l_ImmutableRules_.add("    then, at the most inferior parent level, and if the resource");
      l_ImmutableRules_.add("    is a table column or a trigger:");
      l_ImmutableRules_.add("    - conditional user/simple host ACL on the table column/trigger");
      l_ImmutableRules_.add("    EXCLUSIVE OR (the first matched ACL is selected)");
      l_ImmutableRules_.add("    - non-conditional 'accounts-for-hosts group' ACL on");
      l_ImmutableRules_.add("    the table column/trigger");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    The table index and the view has no ACL at all.");
      l_ImmutableRules_.add("    The view is accessed only through the ACL of its view delegate.");
      l_ImmutableRules_.add("    The view delegate, the trigger, the stored function and");
      l_ImmutableRules_.add("    the stored procedure is modelled as an Actor having");
      l_ImmutableRules_.add("    ACL as target, but never ACL as source.");
      l_ImmutableRules_.add("    The 4 types of Actor operate on the tables and columns");
      l_ImmutableRules_.add("    of the same database. Two modes of operation are possible:");
      l_ImmutableRules_.add("    the DEFINER mode and the INVOKER mode.");
      l_ImmutableRules_.add("    The definer attribute names a MySQL account. The mode");
      l_ImmutableRules_.add("    specifies whether the object executes in the definer or");
      l_ImmutableRules_.add("    invoker context, that is indeed the MySQL current user.");
      l_ImmutableRules_.add("    The definer is modelled by the starting account of the Actor.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    HOW THE MATCHED RIGHTS ARE SELECTED -----------");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    The MySQL server analyzes the rights in a process where");
      l_ImmutableRules_.add("    it takes the first applicable row in the rights tables.");
      l_ImmutableRules_.add("    In most cases, such a row is simulated through an ACL in Access Road.");
      l_ImmutableRules_.add("    A matched ACL is one of the selected ones for a given current user");
      l_ImmutableRules_.add("    and a given target. Except for the hosts group ACL,");
      l_ImmutableRules_.add("    a matched ACL has always non-null rights.");
      l_ImmutableRules_.add("    Each target has a sorted list of direct ACLs, in its 'ACL' tab, ");
      l_ImmutableRules_.add("    and a sorted list of inherited ACLs, in its 'ACL Inheritance' tab. ");
      l_ImmutableRules_.add("    In most of the cases, an inherited ACL is an ACL for which");
      l_ImmutableRules_.add("    its target is a direct or indirect parent of the access target, ");
      l_ImmutableRules_.add("    except for the bases set ACL copied into the matched databases ACL lists. ");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    We have seen that at every parent level, through a parsing");
      l_ImmutableRules_.add("    of the sorted list of ACL, the first ACL that matches");
      l_ImmutableRules_.add("    the current user, if any, is the only applicable ACL.");
      l_ImmutableRules_.add("    The single exception is for the selection of two to-be-mixed ACL");
      l_ImmutableRules_.add("    at the database level (see above).");
      l_ImmutableRules_.add("    (R1) The first ACL SORTING RULE is to apply the order used");
      l_ImmutableRules_.add("    in this presentation to describe the varied types of AclEntries,");
      l_ImmutableRules_.add("    into the previous section HOW THE RIGHTS ARE MANAGED. Then,");
      l_ImmutableRules_.add("    a matched conditional user/host ACL is always selected before");
      l_ImmutableRules_.add("    any matched non-conditional 'accounts-for-hosts group' ACL.");
      l_ImmutableRules_.add("    For the database level, a matched non-conditional 'accounts-for-hosts");
      l_ImmutableRules_.add("    group' ACL is always selected before any matched to-be-mixed ACL couple.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    For the root 'MySQL server' and any database children,");
      l_ImmutableRules_.add("    there are two other sorting rules for their AclEntries:");
      l_ImmutableRules_.add("    - (R2) the conditional ACL of the 'anonymous' account at the end");
      l_ImmutableRules_.add("    of all the conditional ACL for the other accounts.");
      l_ImmutableRules_.add("    - (R3) among the non-conditional ACL, which are all for");
      l_ImmutableRules_.add("    accounts-for-hosts groups, the hosts group sorting is applied");
      l_ImmutableRules_.add("    (there is no ordering among the accounts).");
      l_ImmutableRules_.add("    The ACL list is not totally ordered, since the accounts ");
      l_ImmutableRules_.add("    conditional ACL are not sorted if the accounts are not 'anomymous'.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    For the databases, the ACL sorting is more complex, since");
      l_ImmutableRules_.add("    the (R1) and (R3) rules remains but they are completed by:");
      l_ImmutableRules_.add("    - (R4) replaces (R2): the ACL of the 'anonymous' account at the end");
      l_ImmutableRules_.add("    of all the ACL having the same type for the other accounts.");
      l_ImmutableRules_.add("    - (R5) the ACL for the 'database' target are put before");
      l_ImmutableRules_.add("     the ACL for a 'bases set' target.");
      l_ImmutableRules_.add("    - (R6) among the ACL for a 'bases set' target, and having");
      l_ImmutableRules_.add("     the same type, the bases set sorting is applied.");
      l_ImmutableRules_.add("    - (R7) among the to-be-mixed ACL, the ACL of host(s) ");
      l_ImmutableRules_.add("    is put after the ACL of account.");
      l_ImmutableRules_.add("    - (R8) among the to-be-mixed ACL of account, ");
      l_ImmutableRules_.add("    the set R4/R5/R6 is applied.");
      l_ImmutableRules_.add("    - (R9) among the to-be-mixed ACL of host(s),");
      l_ImmutableRules_.add("    the hosts group sorting is applied first, then the set R4/R5/R6.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    A last sorting rule is applied to the inherited ACL list, to put first");
      l_ImmutableRules_.add("    the ACL for the root, then the ACL for the database level, and so on.");
      l_ImmutableRules_.add("    This sorting has no impact on the selection of AclEntries, because");
      l_ImmutableRules_.add("    the selection at each parent level is independent.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    HOW THE BASES SETS AND THE HOSTS GROUPS ARE SORTED");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    The bases set sorting is complex and beyond the specification.");
      l_ImmutableRules_.add("    The simpliest rule is to put first the databases, before the bases sets,");
      l_ImmutableRules_.add("    and there is no sorting among the databases.");
      l_ImmutableRules_.add("    A typical example of sorted list for database and bases sets,");
      l_ImmutableRules_.add("    having a match to the database 'one' for each item, is the following list:");
      l_ImmutableRules_.add("    'one', on%', 'on_', 'o%', 'o%e', 'o_e', '%', '_ne', '%e', '%ne'.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    The hosts group sorting is different and also beyond the specification.");
      l_ImmutableRules_.add("    The simpliest rule is to put first the simple hosts, before the hosts groups,");
      l_ImmutableRules_.add("    and there is no sorting among the simple hosts.");
      l_ImmutableRules_.add("    A typical example of sorted list for host(s) group, having a match");
      l_ImmutableRules_.add("    to the simple host 'localhost' for each item, is the following list:");
      l_ImmutableRules_.add("    'localhost', localhos_', '_ocalhost', 'localhost%', 'localhos%',");
      l_ImmutableRules_.add("    'lo%', 'lo%t', 'lo%ost', 'l%t', '%', '%ocalhost', '%t'.");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    The simulation of these sortings is a great feature");
      l_ImmutableRules_.add("    for the Access Road users!");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    THESE ACCESS PATHS ARE DELETED BY THE MySQL ADDON");
      l_ImmutableRules_.add("    AFTER THEIR CREATION BY THE GENERIC FUNCTIONS:");
      l_ImmutableRules_.add("    ");
      l_ImmutableRules_.add("    - path containing an accounts-for-host group as intermediate node.");
      l_ImmutableRules_.add("    - path containing, as initial sequence,");
      l_ImmutableRules_.add("       accounts-for-hosts group/acl/actor/run_under...");
      l_ImmutableRules_.add("    - path containing contains a sequence ...account/member/group/, then an ACL.");
      l_ImmutableRules_.add("    - path ending with the sequence ");
      l_ImmutableRules_.add("       (Actor view_delegate)/aclink/GroupIDMember/acLink/Node.");
      l_ImmutableRules_.add("    - path from/through a trigger to an object or");
      l_ImmutableRules_.add("       from an actor not being in the same table.");
      l_ImmutableRules_.add("    - path from/through a MySQL view delegate to an object");
      l_ImmutableRules_.add("       the MySQL view does not have as element.");
      l_ImmutableRules_.add("       ");


  } // initializeConstraintsBeforeCreation()

  /**
   * Creates the ACSMySQLImpl with its structural arguments.
   * Overrides and copies the ACSFactoryImpl method, with simply, as difference,
   * the call to the ACSMySQLImpl constructor.
   * @see ARoad0.gBase.ACSFactoryImpl#createACS
   * @exception BaseError if the integrity rules are not followed (with the list of
   *   all the user errors, including up to 2500 words when all the rules
   *   are unmatched), if the ACS is already known by BaseManagerImpl.
   * @return the new ACSMySQLImpl
   */
  public synchronized ACSImpl createACS() throws BaseError {

      String message = checkACSbeforeCreation();
      if (message != null) throw new BaseError(message);

                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACS(): call 1");

      initializeConstraintsBeforeCreation();

                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACS(): call 2");

      // computes the policy checksum
      int checksum = m_Structure_.values().hashCode();
      int inter = 1;
      for (String[] l_ri : m_TypesAndExplorer_.values()) {
          for (int i = 0; i < l_ri.length; i++) {
              inter = PRIME*inter/10000 + l_ri[i].hashCode();
          }
      } // for l_ri
      checksum = PRIME*checksum + inter;
      inter = 1;
      for (String[] l_ass : m_AssociationsForNewObjects_.values()) {
          for (int i = 0; i < l_ass.length; i++) {
              inter = PRIME*inter/10000 + l_ass[i].hashCode();
          }
      } // for l_ass
      checksum = PRIME*checksum + inter;
      inter = 1;
      // does not use StringRight.hashcode() because it integrates the ACS name
      for (StringRight[] l_ri : m_StandardRights_.values()) {
          for (int i = 0; i < l_ri.length; i++) {
              inter = PRIME*inter/10000 + l_ri[i].getRight().hashCode();
          }
      } // for l_ri
      checksum = PRIME*checksum + inter;
      inter = 1;
      // does not use StringRight.hashcode() because it integrates the ACS name
      for (StringRight[] l_ri : m_SpecializedRights_.values()) {
          for (int i = 0; i < l_ri.length; i++) {
              inter = PRIME*inter/10000 + l_ri[i].getRight().hashCode();
          }
      } // for l_ri
      checksum = PRIME*checksum + inter;
      inter = 1;
      if (m_SubAndGroupPolicy_ != null) {
          for (String[] l_na : m_SubAndGroupPolicy_.values()) {
              for (int i = 0; i < l_na.length; i++) {
                  inter = PRIME*inter/10000 + l_na[i].hashCode();
              } // for i
          } // for l_na
      } // if
      checksum = PRIME*checksum + inter;

      if (type_.indexOf(" | ") != -1)
          type_ = type_.substring(0, type_.indexOf(" | ")) + " | " + Integer.toString(checksum);
      else
          type_ = type_ + " | " + Integer.toString(checksum);


      // creates the ACS without registering in BaseManagerImpl
      ACSMySQLImpl acs = (ACSMySQLImpl) ACSMySQLImpl.valueOf(
                      iSname_, parentName_, name_,
                      type_, m_Structure_, m_TypesAndExplorer_,
                      AssociateUtilityImpl.getM_FilteredAssociations(
                        m_AssociationsForNewObjects_, true),
                      m_StandardRights_, m_SpecializedRights_, l_Rights_,
                      l_ImmutableRules_, m_CommentsOnProperties_,
                      m_AcsVocabulary_);

      message = null;

      // registering in BaseManagerImpl
      return createACSforGenericBase(acs);
  } // createACS()

} // ACSFactoryMySQLImpl
