//Title:        Access Road
//Version:      0.7.1
//Copyright:    Copyright 2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.gBaseBeanInfo.ACSImplBeanInfo;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;
import ARoad0.Gui1.Desktop;

import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.TreeMap;
import java.util.SortedMap;
import java.util.HashMap;
import java.util.Map;
import java.util.HashSet;
import java.util.Set;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Collections;

import javax.swing.ImageIcon;
import javax.swing.Icon;

import java.net.MalformedURLException;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyDescriptor;

/**
 * This class is responsible for creating a new ACS, defining its structure,
 * and creating the restrictions on the objects, rights, privileges and types values
 * when an ACS is created. With ACSImpl, this is the most complex class
 * of the gBase package.
 * The main methods which ensure the ACS integrity rules are checkACSbeforeCreation()
 * and ACSFactoryUtilityImpl.controlAcsStructure().
 * Called by gDMak.ActionNewACSyst and gWork.AcsFactory.
 * This class does not implements the interface ACSCatalog, but most of the methods
 * of this interface are implemented.
 * Caution: This class uses for its GUI two BeanInfo; ACSFactoryImplBeanInfo and 
 * ACSPreFactoryImplBeanInfo.
 * ACSPreFactoryImplBeanInfo ensures the creation of the ACSFactory instance and
 * the acsrights list.
 * <P> 
 * This class methods have to be called in the right order: <P>
 * - through ACSPreFactoryImplBeanInfo, creates the instance constructor, 
 *   including the acs name and the type, then creates the optional acsrights, <P>
 * - change the dialog frame to create most of the acs properties, through ACSFactoryImplBeanInfo: <P>
 *   - the ACS structure, <P>
 *   - the IS nametree properties for the ACS: IS name, hub property, visible space, links, <P>
 *   - the EP and Resource types, if any, and the specialized values, <P>
 *   - the 4 set/addL_XXXRights(), if any, <P>
 * - checkACSbeforeCreation() may be used at any step by the user, <P>
 * - createACS(), which sets the new ACS and finalizes this class instance, <P>
 * This class updates the nametrees, implying NameISTrees which have some specific uses:<P>
 * - an ACS may belong to one information system (IS), or not, <P>
 * - there are IS nametrees where ACS belonging to the IS are registered, <P>
 * - a nametree has empty nodes and ACS nodes, <P>
 * - boolean 'is_empty' property for each node, false for the ACS node,
 *      which means that a non-empty node is an active entity (say, an ACS). <P>
 * - boolean 'is_hub' property for each ACS, which means, if true, that 
 *      any direct child of this node may have exchanges with any other direct child,
 *      if the rights allow them. <P>
 * - 'VisibleSpace' is a mandatory ACS property, and the possible String values 
 *      are NameISTree constants <P>
 * - String 'is_authorization_server' as a Links property, which indicates
 *      an ACS which is an authorization server,
 *      which has methods to generate dynamic rigths on requests of a right source;
 *      then, the property 'is_running_on' may be true or not. <P>
 * Note: couples of Strings to indicate couples (property, target_node_name)
 * are possible in Links. IS nametrees have also components which are managed directly
 * by BaseManagerImpl.
 * <P>
 * It is a javabean with the following bound properties:
 * 'ISname', 'ParentName', 'HubProperty', 'VisibleSpace',
 * 'ResourceTypes', 'EPTypes', 'TypesAndExplorerSet',
 * 'SpecializedRightsSet', 'HavePrivilegeSpecializedRights', 'AssociationForNewObjectsMap',
 * 'ARightsSet', 'GORightsSet', 'AclRightsSet',
 * 'PrivilegeRightsSet', 'BridgeRightsSet',
 * 'ManageResources', 'ManageResourceTrees', 'ManageAccounts', 
 * 'ManageGroupsOfAccounts', 'ManageGroupsOfGroups', 'ManageInternalAcls',
 * 'ManageExternalAcls', 'ManageFlexibleConditionalAcls', 'ManageRightDefinedConditionalAcls',
 * 'ManageActors', 'ManageBridges',
 * 'ManageSecondaryGroupsForActor', 'ManageAGORight', 'ManageNxExecutables',
 * 'ManageDirectoryEPs', 'ManagePrivilegeRights',
 * 'ManageAcsRights', 'ManageMetaRights',
 * 'VirtualFolders', 'ManageGrantRights', 'ManageDenyRights',
 * 'TiedToParentACS', 'RunningOnParentACS', 'isRightServer',
 * 'ComponentNodes', 'CompositeNodes', 'BelongsToComposite', 'SubAndGroupPolicy',
 * 'Alternates', 'OpeningAndClosingRoots', 'OpeningNoClosingRoots', 'IntegrityRoots',
 * 'Icon', 'CommentsOnProperties',
 * 'ManageGrantingRightPrivileges', 'ManageDenyingRightPrivileges',
 * 'ManageSimpleTypedPrivileges', 'ManageSimpleLinkedPrivileges',
 * 'ManageResourceVirtualFolders', 'ManageEPVirtualFolders',
 * 'ManageActorInEPVirtualFolders', 'ManageUserIDInEPVirtualFolders',
 * 'ManageGroupIDInEPVirtualFolders', 'ManageSimpleResourceInResourceVirtualFolders',
 * 'ManageDirectoryInResourceVirtualFolders', 'ManageDirectoryEPinEPVirtualFolders',
 * 'ManageVirtualFolderTrees', 'OneTimeInVirtualFolderTreeForMember',
 * 'IncompleteStructureModeling', 'IncompleteBehaviorModeling',
 * 'PrivilegeForTypeTypes', 'PrivilegeForLinksTypes',
 * 'ResVirtualFolderTypes', 'EPVirtualFolderTypes',
 * 'AcsVocabulary', 'ManageACLRightInheritance', 'ManageAGORightInheritance',
 * "ManageLPRIRightInheritance', 'ManagePrivilegesInVirtualFolders',
 * 'ManagePrivilegesInResources', 'ManageOneTimeInVirtualFolderTreeForMember',
 * 'ManageInternalHardAlias', 'ManageSoftAlias',
 * 'ManageIfAllSourceLinkedPrivileges', 'ManageInternalForOneToOneLinkedPrivileges',
 * 'ManageExternalForOneToOneLinkedPrivileges', 'ManageClassedTypedPrivileges',
 * 'ManageClassedSpecificTypedPrivileges', 'ManageAuthorizationServerTypedPrivileges'.
 * <P>
 * This class has several subclasses in the AcsAddon packages.
 * Most of the variables are protected.
 * @see ARoad0.gBaseBeanInfo.ACSImplBeanInfo
 * @version 0.7.1
 */
public class ACSFactoryImpl implements BaseObject, BoundBean, ACSRun {
  // public attributes
      // no right set for 'root', which has 'full_control' as mandatory right
    /**
     * for Account rights only, in m_StandardRights_
     */
  public static final String  A_RIGHTS_SET = "Resource.AccountRightsSet";
    /**
     * for rights in m_StandardRights_
     */
  public static final String  GO_RIGHTS_SET = "Resource.GORightsSet";
  public static final String  ACL_RIGHTS_SET = "AclEntry.AclRightsSet";
  public static final String  PRI_RIGHTS_SET = "Privilege.PrivilegeRightsSet";
  public static final String  BDG_RIGHTS_SET = "Actor.BridgeRightsSet"; // no use
  public static final String  SPE_RIGHTS_SET = "SpecializedRightsSet"; // no use
      
  // private attributes
  private static final int    PRIME = 1000003;
  private static final int    INITIAL_CAPACITY_25 = 25; // default value for map: 101
  private static final int    INITIAL_CAPACITY_5 = 5;
  private static final        StringRight[] L_EMPTY_ARRAY_2 = new StringRight[0];
  private static final        String[] L_EMPTY_ARRAY = new String[0];

  // protected attributes
  // size of the ACS icon
  protected static final      int ICON_HEIGHT_MAX = 24;
  protected static final      int ICON_WIDTH_MAX = 24;
  /**
   * For links in the IS nametree
   */
  //protected static ArrayList    l_authorizedLinks__ = new ArrayList();
  
  /**
   * For the specialized rights. This is less complete than
   * ACSFactoryUtilityImpl.L_ACS_GENERIC_CLASSES.
   */
  protected static String[]   l_AllowedACSInterfacesForSpecializedRights_ = null;

  static {
      l_AllowedACSInterfacesForSpecializedRights_ = new String[] {
              "Actor",
              "Directory",
              "GroupID",
              "Node",
              "Resource",
              "UserID",
              "Basic",
              "Privilege",
              "VirtualFolder",
              "DirectoryEP"
              };
  } // static
  
  protected ImmutableName       name_;
  protected String              type_;
  protected String              comment_ = "";
  protected ImmutableName       iSname_;
  protected Name                parentName_;
  protected boolean             is_hub_ = false;
  protected String              visibleSpace_ = NameISTreeImpl.ALL_THE_IS_SPACE;
  
  protected String[]            l_resourceTypes_ = new String[]{"<undefined>"};
  protected String[]            l_epTypes_ = new String[]{"<undefined>"};
  protected String[]            l_resvfTypes_ = new String[]{"<undefined>"};
  protected String[]            l_epvfTypes_ = new String[]{"<undefined>"};
  protected String[]            l_privilegeForTypeTypes_ = new String[]{"<undefined>"};
  protected String[]            l_privilegeForLinksTypes_ = new String[]{"<undefined>"};
  protected StringRight[]       l_AccountRightsSet_ = L_EMPTY_ARRAY_2;
  protected StringRight[]       l_GORightsSet_ = L_EMPTY_ARRAY_2;
  protected StringRight[]       l_AclRightsSet_ = L_EMPTY_ARRAY_2;
  protected StringRight[]       l_bridgeRightsSet_ = L_EMPTY_ARRAY_2;
  protected StringRight[]       l_Rights_;
  /**
  * To define the privilege rights following their main type (standard rights)
  * or second type (specialized rights)
  * @since 0.7.0
  */
  protected HashMap<String,StringRight[]>  m_PrivilegeRights_ =
          new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25);
  
  /**
   * true if some specialized rights have been set in the privilege rights;
   * Not used in the GUI nor outside this class.
   * @since 0.7.0
   */
  protected boolean             havePrivilegeSpecializedRights_ = false;
   
  /**
  * The active booleans which define the generic ACS structure and features.
   * The name of the parameter is the key, and the value indicates if
   * the parameter is managed by this ACS. By default, the value
   * is false.
   * Some complementary, non-generic values are allowed for the AcsAddons.
   * This map may be directly updated at the creation of an AcsAddon ACS,
   * for adding new keys of non-generic immutable properties which are ridden
   * by their proper AcsAddon classes.
   */
  protected SortedMap<String,Boolean>  m_Structure_ = new TreeMap<String,Boolean>();
  
     /**
   * Map of mandatory associations on the ACS new objects. Updated by
   * add/removeAssociationsToNewObject().
   */
  protected HashMap<String,String[]>  m_AssociationsForNewObjects_ =
          new HashMap<String,String[]>(INITIAL_CAPACITY_25);
  /**
   * Map of restrictions on the ACS general String properties. Updated from l_resourceTypes_,
   * l_epTypes_, l_resvfTypes_, l_epvfTypes_, l_privilegeForTypeTypes_, l_privilegeForLinksTypes_
   * and add/removeTypesAndExplorer().
   */
  protected HashMap<String,String[]>  m_TypesAndExplorer_ =
          new HashMap<String,String[]>(INITIAL_CAPACITY_25);
     /**
   * Map of restrictions on the ACS StringRight properties
   */
  protected HashMap<String,StringRight[]>  m_StandardRights_ =
          new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25);
  /**
  * To extend or to reduce the standard rights following the 
  * target/source interface or type
  */
  protected HashMap<String,StringRight[]>  m_SpecializedRights_ =
          new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25);
  /**
  * The map key has the format
  * ACS.Name.<Value>.XXX which is a parent ACS name having the value XXX.
  * The map values represents a script of instructions.
  * @since 0.7.0
  */
  protected HashMap<String, String[]> m_RootsForAcsOpeningAndClosing_;
  /**
  * The map key has the format
  * ACS.Name.<Value>.XXX which is a parent ACS name having the value XXX.
  * The map values represents a script of instructions.
  * @since 0.7.0
  */
  protected HashMap<String, String[]> m_RootsForAcsOpeningNoClosing_;
  /**
  * The map key has the format
  * ACS.Name.<Value>.XXX which is a parent ACS name having the value XXX.
  * The map values represents a script of instructions.
  * @since 0.7.0
  */
  protected HashMap<String, String[]> m_RootsForAcsIntegrity_;
  /**
   * List of the ACS rules which are derivated from the structure, like
   * 'AGO right: A rights overlay G rights'.
   * @since 0.7.0
   */
  protected List<String>              l_ImmutableRules_ = null;

  /**
   * Get the immutable subACSs and their dedicated nodes to display in the GUI.
   * In the returned value, the key is the subACS name, in one word, and 
   * the value is an array which may contain the following values: <P>
   * 'Resources', 'UserIDs' and 'GroupIDs'. <P>
   * For instance, 'Resources' means that a node ACSTree.RESOURCES has to be displayed
   * directly under the subACS node, in the explorer.
   */
  protected HashMap<String, String[]>   m_SubAndGroupPolicy_ = null;

  /**
   * Immutable map which says if the ACS or a subACS belongs to a Composite ACS 
   * in the IS Name Tree. <P>
   * In the returned value, the key is the subACS name, in one word, 
   * or the ACS name last component, and the value is the Composite name.
   * The returned map through the dedicated method is not exactly this property,
   * since it is a Map<String, ImmutableName[]>.
   * @since 0.7.0
   */
  protected HashMap<String, ImmutableName>  m_BelongsToComposite_ = null;

    /**
   * Map of physical and logical components under the ACS or its parent. Updated
   * by add/removeComponentNodes().
   * The key starts with 'NewAcsParent.' or 'NewAcs.' to say that the component
   * node is under the parent or the new ACS, and the key ends with the subname
   * of the component, in other words, its name as last component, and its parent,
   * if any, between it and 'NewAcsParent.' or 'NewAcs.'. These optional parents
   * are always components defined in the same map.
   * The value is an array which may contain the following values: <P>
   * 'physical' and 'logical'.
   * @since 0.7.0
   */
  protected HashMap<String,String[]>  m_ComponentNodes_ = null;

    /**
   * Map of the new composite nodes for the IS nametree, outside the current
   * ACS creation in progress. Updated by add/removeCompositeNodes().
   * The key is the unique composite name.
   * The value is an array which may contain only one value among: <P>
   * 'acs', 'acs.authorization_server' and 'subacs'.
   * @since 0.7.0
   */
  protected HashMap<String,String[]>  m_CompositeNodes_ = null;

    /**
   * Map of the alternates of the ACS or the subACS, if any, with all the
   * associated properties. Updated by add/removeAlternates().
   * The key starts with the ACS name last component or the subACS name (no point
   * at the end, on the contrary of getL_MapKeysStartsForAlternates()), and then,
   * the key ends with the name of the alternate in the same IS. <P>
   * The value is an array which may contain the following values: <P>
   * - exactly one value among the types in ISFactoryUtilityImpl.getAlternateTypes(), <P>
   * - one or several right types, among the 3 in ISFactoryUtilityImpl, <P>
   * - as options, the values OPTIONAL_ALTERNATE and OPTIONAL_GROUND in ISFactoryUtilityImpl. <P>
   * Then, the values array lenght is from 2 to 6.
   * The returned map through the dedicated method is not exactly this property,
   * since it is a Map&lt;String, String[]&gt;. &lt;GeneralCriterion&gt;
   * @since 0.7.0
   */
  protected HashMap<ImmutableName, String[]> m_alternateProperties_ = null;
  /**
   * The file name of the icon of the aCS in the beamer - selected at the creation
   *  or the beamer uses the default value. Exemple of value: 'linux.gif'
   */
  protected String                icon_;
  /**
   * The image the user cannot set directly, and it is used by ActionNewACS to copy
   * the image icon of an open ACS. Used to create the ACS only if icon_ is null.
   */
  protected ImageIcon             imageIcon_;

  /**
   * Map of the comments on each ACS property, to explain the simulation.
   * Used with add/removeCommentsOnProperties(). Default value is null.
   * @since 0.7.0
   */
  protected HashMap<String,String[]>  m_CommentsOnProperties_ = null;

  /**
   * The ACS vocabulary defines several independant domains: <P>
   * - how to name the ACSFactoryUtilityImpl.L_ACS_TREE_GENERIC_UPPER_NODES constants for this ACS,
   * as names of the ACS nodes in the proper ACS internal frame, <P>
   * - what are the gBase classes to use, since some generic classes may be
   * subclassed in an AcsAddon gBase package (see ACSFactoryUtilityImpl.L_CLASSIC_CLASS_KEYS), <P>
   * - if the ACS is an AcsAddon, what are the specific new classes in
   * the AcsAddon gBase package (see ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES).
   * The default value is ACSFactoryUtilityImpl.M_GENERIC_ACS_VOCABULARY,
   * but some of its keys are removed if the ACS does not use the associated classes.
   * Caution: UtilityImpl.getACSVocabularyFromArrayOrInterface() reads
   * this property from an instance created through an empty constructor
   * of this class or one of its subclass.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#M_GENERIC_ACS_VOCABULARY
   */
  protected HashMap<String,String[]> m_AcsVocabulary_ = (HashMap<String,String[]>)
          ACSFactoryUtilityImpl.M_GENERIC_ACS_VOCABULARY.clone();

  /**
   * manages all the property change listeners
   */
  protected transient PropertyChangeSupport changeSupport_ = null;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  protected static ACSFactoryImpl EMPTY_INSTANCE = new ACSFactoryImpl();

  
 /**
  * Default constructor
  */
    public ACSFactoryImpl() {
      name_ = NameImpl.EMPTY_INSTANCE;
      type_ = ""; 
      changeSupport_ = new PropertyChangeSupport(this);
      l_Rights_ = new StringRight[0];
    } // constructor
    
  /**
   * Called by gDMak.ActionNewACSyst and gWork.AcsFactory.
   * @param _iSname the name of the IS containing this new ACS. May be null, and
   * ISFactoryUtilityImpl.DEFAULT_IS_NAME is then set as IS name and it is created
   * BaseManagerImpl.
   * @param _name the name of the new ACS. Never null.
   * @param _ACStype the ACS type. Never null.
   * @param _parentName the parent of this new ACS. No use if _iSname is null.
   * Otherwise, it has to contain the IS name.
   * @exception BaseError if _type is null, _parentName does not start with _iSname.
   */
  public ACSFactoryImpl(ImmutableName _iSname, ImmutableName _name, 
          String _ACStype, ImmutableName _parentName) 
          throws BaseError {

      if (_ACStype == null) throw new BaseError(
        " The ACS type must be defined.");
      if (_name == null || _name.equals(new NameImpl("no_acs"))) throw new BaseError(
              " A null ACS name or the ACS name 'no_acs' is not allowed.");
      
      if (_iSname != null) 
          iSname_ = (ImmutableName) _iSname.clone();
      else {
          iSname_ = new NameImpl(ISFactoryUtilityImpl.DEFAULT_IS_NAME);
      }

      if (!_name.subname(0, iSname_.noOfComponents()).equals(iSname_) ) {
          iSname_ = null;
          throw new BaseError(" The ACS name does not"+
                  " start with the IS name.");
      } // if

      type_ = _ACStype;

      if (_parentName != null) {
          if (!_parentName.subname(0, iSname_.noOfComponents()).equals(iSname_) ) {
              iSname_ = null;
              throw new BaseError(" The parent name does not"+
                      " start with the IS name.");
          }
          parentName_ = (Name)_parentName.clone();
      }
      else {
          parentName_ = (Name)iSname_.clone();
      }
      
      if (!BaseManagerImpl.getInstance().knowsIS(iSname_)) {
          BaseManagerImpl.getInstance().addIS(iSname_, true);
      } // if

      name_ = (Name)_name.clone();
      changeSupport_ = new PropertyChangeSupport(this);
      l_Rights_ = new StringRight[0];
      
            // for testing and tracking
            //System.out.println("in ACSFactoryImpl constructor, "+
              //"l_structure_ is: ");
            //for (int i = 0 ; i < l_structure_.length ; i++) {
              // System.out.println(i+": "+l_structure_[i]); 
            //}
      
      // initializes the structure
      m_Structure_ = ACSFactoryUtilityImpl.getDefaultAcsStructure();
      
  } // constructor

  /**
   * Creates the ACS with its structural arguments.
   * To construct the ACS, calls ACSImpl.valueOf() then createACSforGenericBase(). <P>
   * Before the ACS creation, calls initializeConstraintsBeforeCreation(), and computes
   * the ACS policy checksum to add to type_. The checksum is computed from the hashcodes
   * of all the values into the properties m_Structure_, m_TypesAndExplorer_,
   * m_AssociationsForNewObjects_, m_StandardRights_, m_SpecializedRights_ and
   * m_SubAndGroupPolicy_. The ACS type ends with ' | checksum'.
   * Calls createACSforGenericBase() to set the ACS structure and relations in its IS.
   * The ACSFactory instance is finalized at the end of createACSforGenericBase().
   * Overridden by the AcsAddon ACSFactory subclasses.
   * @exception BaseError if the integrity rules are not followed (with the list of 
   * all the user errors, including up to 2500 words when all the rules
   * are unmatched), if the ACS is already known by BaseManagerImpl.
   * @return the new configured ACSImpl
   */
  public synchronized ACSImpl createACS() throws BaseError {
      
      String message = checkACSbeforeCreation();
      if (message != null) {
          
                                    // for testing and tracking
                                    System.out.println("ACSFactoryImpl createACS(): "+
                                                "error from checkACSbeforeCreation()");

          BaseError b = new BaseError(message);
          b.printStackTrace();
          throw b;
      } // if

                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACS(): call 1");

      initializeConstraintsBeforeCreation();   
      
                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACS(): call 2");

      // computes the policy checksum
      int checksum = m_Structure_.values().hashCode();
      int inter = 1;
      for (String[] l_ri : m_TypesAndExplorer_.values()) {
          for (int i = 0; i < l_ri.length; i++) {
              inter = PRIME*inter/10000 + l_ri[i].hashCode();
          }
      } // for l_ri
      checksum = PRIME*checksum + inter;
      inter = 1;
      for (String[] l_ass : m_AssociationsForNewObjects_.values()) {
          for (int i = 0; i < l_ass.length; i++) {
              inter = PRIME*inter/10000 + l_ass[i].hashCode();
          }
      } // for l_ass
      checksum = PRIME*checksum + inter;
      inter = 1;
      // does not use StringRight.hashcode() because it integrates the ACS name
      for (StringRight[] l_ri : m_StandardRights_.values()) {
          for (int i = 0; i < l_ri.length; i++) {
              inter = PRIME*inter/10000 + l_ri[i].getRight().hashCode();
          }
      } // for l_ri
      checksum = PRIME*checksum + inter;
      inter = 1;
      // does not use StringRight.hashcode() because it integrates the ACS name
      for (StringRight[] l_ri : m_SpecializedRights_.values()) {
          for (int i = 0; i < l_ri.length; i++) {
              inter = PRIME*inter/10000 + l_ri[i].getRight().hashCode();
          }
      } // for l_ri
      checksum = PRIME*checksum + inter;
      inter = 1;
      if (m_SubAndGroupPolicy_ != null) {
          for (String[] l_na : m_SubAndGroupPolicy_.values()) {
              for (int i = 0; i < l_na.length; i++) {
                  inter = PRIME*inter/10000 + l_na[i].hashCode();
              } // for i
          } // for l_na
      } // if
      checksum = PRIME*checksum + inter;
      
      if (type_.indexOf(" | ") != -1)
          type_ = type_.substring(0, type_.indexOf(" | ")) + " | " + Integer.toString(checksum);
      else
          type_ = type_ + " | " + Integer.toString(checksum);

      // creates the ACS without registering in BaseManagerImpl
      ACSImpl acs = ACSImpl.valueOf(iSname_, parentName_, name_,
                      type_, is_hub_, visibleSpace_, m_Structure_, m_TypesAndExplorer_,
                      AssociateUtilityImpl.getM_FilteredAssociations(
                        m_AssociationsForNewObjects_, true),
                      m_StandardRights_, m_SpecializedRights_, l_Rights_, 
                      l_ImmutableRules_, m_CommentsOnProperties_,
                      m_AcsVocabulary_);

      message = null;

      return createACSforGenericBase(acs);
  } // createACS()

  /**
   * Sets the ACS structure and relations in its IS.
   * If the IS is new, creates a new IS in BaseManagerImpl with a hub root node.
   * Registers the ACS in BaseManagerImpl. The parent ACS, if any, has to be open.
   * After the ACS creation, calls initializeConstraintsAfterCreation().
   * Called by createACS() and by its versions in the AcsAddon subclasses.
   * The ACSFactory instance is finalized at the end of this method.
   * @since 0.7.0
   * @param _acs new ACS to configure
   * @exception BaseError if the integrity rules are not followed (with the list
   *   of all the user errors, including up to 2500 words when all the rules
   *   are unmatched), if the ACS is already known by BaseManagerImpl.
   * @return the new configured ACS
   */
  protected synchronized ACSImpl createACSforGenericBase(ACSImpl _acs) throws BaseError {
      
      boolean               ISroot_is_hub = true;
      boolean               secondaryProperties = false;
      ImmutableName         name = null;
      Name                  name2 = null;
      ImmutableName         name3 = null;
      List<ImmutableName>   l_alternates = new ArrayList<ImmutableName>(INITIAL_CAPACITY_25);
      Map<ImmutableName, String[]> m_alternateProperties = 
              new HashMap<ImmutableName, String[]>(INITIAL_CAPACITY_5);
      ACS                   parentACS = null;

      _acs.initializeAcsRelationsInIS(
          m_BelongsToComposite_,
          m_RootsForAcsOpeningAndClosing_,
          m_RootsForAcsOpeningNoClosing_,
          m_RootsForAcsIntegrity_,
          m_SubAndGroupPolicy_);

      if (icon_ != null) 
          _acs.setEorIcon(icon_);
      else if (imageIcon_ != null)
          _acs.icon_ = imageIcon_;

      _acs.setComment(comment_);

      // registers the new composite nodes, on one level under the IS root
      if (m_CompositeNodes_ != null) {
          for (String key : m_CompositeNodes_.keySet()) {
              if (((String[])m_CompositeNodes_.get(key))[0].equals("acs"))
                  BaseManagerImpl.getInstance().addCompositeNodeToIS(
                          iSname_, key, ISFactoryUtilityImpl.COMPOSITE_ACS_NODE, false);
              if (((String[])m_CompositeNodes_.get(key))[0].equals("acs.authorization_server"))
                  BaseManagerImpl.getInstance().addCompositeNodeToIS(
                          iSname_, key, ISFactoryUtilityImpl.COMPOSITE_ACS_NODE, true);
              if (((String[])m_CompositeNodes_.get(key))[0].equals("subacs"))
                  BaseManagerImpl.getInstance().addCompositeNodeToIS(
                          iSname_, key, ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE, false);
          } // for key
      } // if (m_CompositeNodes_ != null)

      // adds this to the map of ACSs in BaseManagerImpl.
      // throws BaseError if already exists.
      // TAKE OUT THE FOLLOWING LIGNES for testing the base without BaseManager,
      // and without AcsControllers (which always use BaseManager).
      BaseManagerImpl.getInstance().newEorACS(_acs, ISroot_is_hub);
      
                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACSforGenericBase(): call 3");

      if (getTiedToParentACS()) {
          parentACS = (ACS) BaseManagerImpl.m_OpenAcs__.get(parentName_);
          if (parentACS != null) {
              List<ImmutableName> l_tieds = parentACS.getL_TiedChildACS();
              l_tieds.add(name_);
              parentACS.setL_TiedChildACS(l_tieds);
              l_tieds = null;
          } // if (parentACS != null
      } // if

      // sets l_alternates and m_alternateProperties from m_alternateProperties_
      // with the keys which start with the ACS name last component
      if (m_alternateProperties_ != null) {
          for (Map.Entry entry : m_alternateProperties_.entrySet()) {
              if (!((ImmutableName)entry.getKey()).component(0).
                   equals(name_.lastComponent()))
                      continue;
              // adds the key without the first component
              name = ((ImmutableName)entry.getKey()).subname(1,
                      ((ImmutableName)entry.getKey()).noOfComponents());
              l_alternates.add(name);
              m_alternateProperties.put(
                      name,
                      (String[])entry.getValue());
          } // for entry
      } // if

      if (getTiedToParentACS() ||  
              (m_BelongsToComposite_ != null && 
                m_BelongsToComposite_.get(name_.lastComponent()) != null) ||
              !l_alternates.isEmpty())
                    secondaryProperties = true;
      
                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACSforGenericBase(): call 4");

      // registers the ACS in the IS nametree
      BaseManagerImpl.getInstance().addAcsOrSubAcsNodeToIS(
            iSname_,  parentName_, name_,
            ISFactoryUtilityImpl.ACS_NODE, is_hub_,
            visibleSpace_, getIsAuthorizationServer(), !secondaryProperties);
      
                        // for testing and tracking
                        //BaseManagerImpl.getInstance().getL_IS();

      // registers the new component nodes
      if (m_ComponentNodes_ != null) {

        try {



                                        // for testing and tracking
                                        //System.out.println("ACSFactoryImpl.createACSforGenericBase(), "+
                                          //      "the component name is "+name);

              for (String key : m_ComponentNodes_.keySet()) {

                    // cuts the leading and trailing whitespaces, for which the number may vary,
                    // then adds the right whitespaces to build up the String format of a name
                  name = NameImpl.getNameFromString( " "+key.substring(
                        key.indexOf(".") + 1 ).trim() + "  ");
                  name2 = new NameImpl();

                  // if 'name' contains 3 components, create 3 nodes in a sequence
                  // from the upper node
                  for (Object compo : name.components()) {
                      name2 = name2.extended((String)compo);

                      if (key.startsWith("IsChildOfNewAcsParent.")) {
                        name3 = ((Name)((NameImpl)parentName_).clone()).extended(name2);
                      }
                      else if (key.startsWith("IsChildOfNewAcs.")) {
                        name3 = ((Name)((NameImpl)name_).clone()).extended(name2);
                      }

                                // for testing and tracking
                                //System.out.println("ACSFactoryImpl createACSforGenericBase() calls"+
                                  //      " BaseManagerImpl.getInstance().addComponentNodeToIS() for "+
                                    //    name2+" in "+name3);


                      // with the filter on 'level', the parent node is created first
                      if (((String[])m_ComponentNodes_.get(key))[0].equals("physical"))
                          BaseManagerImpl.getInstance().addComponentNodeToIS(
                                  iSname_, name3, ISFactoryUtilityImpl.PHYSICAL_COMPONENT_NODE);
                      if (((String[])m_ComponentNodes_.get(key))[0].equals("logical"))
                          BaseManagerImpl.getInstance().addComponentNodeToIS(
                                  iSname_, name3, ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE);
                  } // for name3
              } // for key

        } catch (BaseError b) {
            throw new BaseError(" The component node key does not end with a true name. "+
                    b.toString());
        }


      } // if (m_ComponentNodes_ != null)
      
                                // for testing and tracking
                                //System.out.println("ACSFactoryImpl createACSforGenericBase(): call 5");

      if (secondaryProperties) {
        if (m_BelongsToComposite_ != null)
          BaseManagerImpl.getInstance().addSecondaryPropertiesToAcsOrSubNode(
            iSname_, name_, getTiedToParentACS(), 
            m_BelongsToComposite_.get(name_.lastComponent()),
            l_alternates.toArray(new ImmutableName[0]),
            m_alternateProperties, true); // (m_SubAndGroupPolicy_ == null)
        else
          BaseManagerImpl.getInstance().addSecondaryPropertiesToAcsOrSubNode(
            iSname_, name_, getTiedToParentACS(),
            null,
            l_alternates.toArray(new ImmutableName[0]),
            m_alternateProperties, true); // (m_SubAndGroupPolicy_ == null)
      } // if
      
                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACSforGenericBase(): call 6");

      // registers the subACSs in the IS nametree
      if (m_SubAndGroupPolicy_ != null) {

          for (String subname : m_SubAndGroupPolicy_.keySet()) {

              // does not process the group tree roots
              if (m_SubAndGroupPolicy_.get(subname)[0].equals("<GroupTree>") )
                  continue;
              // processes '<ResourceSubACS>' and '<GroupSubACS>'

              // sets l_alternates and m_alternateProperties from m_alternateProperties_
              // with the keys which start with the subACS name
              m_alternateProperties.clear();
              l_alternates.clear();
              secondaryProperties = false;
              if (m_alternateProperties_ != null) {
                  for (Map.Entry entry : m_alternateProperties_.entrySet()) {
                      if (!((ImmutableName)entry.getKey()).component(0).equals(subname))
                              continue;
                      // adds the key without the first component
                      name = ((ImmutableName)entry.getKey()).subname(1,
                              ((ImmutableName)entry.getKey()).noOfComponents());
                      l_alternates.add(name);
                      m_alternateProperties.put(
                              name,
                              (String[])entry.getValue());
                  } // for entry
              } // if

              if (!l_alternates.isEmpty() || 
                      (m_BelongsToComposite_ != null && 
                      m_BelongsToComposite_.get(subname) != null))
                            secondaryProperties = true;

              // is_hub_ and visibleSpace_ are from the ACS
              // a subACS is never an authorization server
              BaseManagerImpl.getInstance().addAcsOrSubAcsNodeToIS(
                    iSname_,  (Name)name_,
                    ((Name)((NameImpl)name_).clone()).extended(subname),
                    ISFactoryUtilityImpl.SUB_ACS_NODE, is_hub_,
                    visibleSpace_, false, !secondaryProperties);

              // a subACS has never a tied child relation with its ACS
              if (secondaryProperties ) {
                  if (m_BelongsToComposite_ != null)
                      BaseManagerImpl.getInstance().
                        addSecondaryPropertiesToAcsOrSubNode(iSname_,
                        ((Name)((NameImpl)name_).clone()).extended(subname),
                        false,  m_BelongsToComposite_.get(subname),
                        l_alternates.toArray(new ImmutableName[0]),
                        m_alternateProperties,
                        true);
                  else
                      BaseManagerImpl.getInstance().
                        addSecondaryPropertiesToAcsOrSubNode(iSname_,
                        ((Name)((NameImpl)name_).clone()).extended(subname),
                        false,  null,
                        l_alternates.toArray(new ImmutableName[0]),
                        m_alternateProperties,
                        true);
              } // if (secondaryProperties )
          } // for subname

      } // if
      
                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACSforGenericBase(): call 7");

      initializeConstraintsAfterCreation(_acs);

                        // for testing and tracking
                        //System.out.println("ACSFactoryImpl createACSforGenericBase(): call 8");
      
      // This factory instance is finalized
      try {
          finalizeForUser();
      } catch (UpDateError u) {
          throw new BaseError(u.toString());
      }
      
      name = null;
      l_alternates = null;
      m_alternateProperties = null;
      parentACS = null;
      name2 = null;
      name3 = null;

      return _acs;
  } // createACSforGenericBase()


  /**
   * Controls the integrity of the ACS with all its arguments,
   * before the call to ACSImpl.valueOf().
   * If the integrity rules are not followed on the ACS structure, return an user message
   * (with the list of all the user errors, including up to 2500 words 
   * when all the rules are unmatched). 
   * The ACS should have a type.
   * When a type of rights is managed, it is mandatory to have a set of
   * authorized rights for this right type (account rights, for instance), with
   * two cases: through the standard rights array, or through the
   * specialized rights map.
   * It is not mandatory to have some associated types when a kind of ACSObject is managed.
   * It is mandatory to manage Privileges when the VirtualFolders are managed.
   * Controls the constraints on subACSs and groups tree roots, since they must
   * have dedicated types set in the type constraints both for the keys
   * 'Explorer.SubAcsNode' and 'Explorer.AcsGroupTree' which controls the explorer.
   * For the key 'CreationByBeamer.NoType' which forbids the subACS types to be chosen
   * for the ACSObjects the user may create through the beamer, the
   * initializeConstraintsBeforeCreation method sets as forbidden
   * all the types that contains '&lt;byAcsCreator&gt;'.
   * Controls the new object associations and the ACS roots handling.
   * Called by createACS(). Calls ACSFactoruUtilityImpl.controlAcsStructure() and
   * adds complementary controls.
   * @return null if the ACS arguments are correct, or an error message.
   */
  public synchronized String checkACSbeforeCreation() {
      
      //if (m_Structure_ == null) return "The structure has to be defined " +
        //      "before the ACS creation."; 
      // variables
      Map.Entry<String,String[]>     entry = null;
      String                        message = ACSFactoryUtilityImpl.
                                        controlAcsStructure(m_Structure_);
      Set                           spTypes = null;
      List<String>                  l_NodesToDisplay = null;
      boolean                       withResourceSubAcs = false;
      //boolean                       withGroupSubAcs = false;
      boolean                       withGroupTree = false;
      boolean                       found = false;
      Name                          name = null;
      String[]                      l_values = null;
      String[]                      l_current = null;
      String                        key = null;
      
      // first control
      if (message != null) {
                return message;
      }
      
      if (type_ == null) {
                return "The ACS type has to be defined before the ACS creation.";  
      }

      // analysis of getL_LimitedValuesForTypesAndExplorer()
      // for the subACSs and the groups tree roots
      if (m_SubAndGroupPolicy_ != null) {
          for (Map.Entry ent : m_SubAndGroupPolicy_.entrySet()) {
            l_values = (String[]) ent.getValue();

            if (l_values == null) {
                message = "The subACS or groups tree root "+ent.getKey()+" has no value "+
                        "in its map";
                break;
            }

            if (Arrays.asList(l_values).contains("<ResourceSubACS>") ) {
                found = false;
                for (int i = 0; i < l_resourceTypes_.length; i++) {
                    if (l_resourceTypes_[i].indexOf("<byAcsCreator> "+ent.getKey()) != -1) {
                        found = true;
                        break;
                    }
                } // for i
                if (!found) {
                    message = "The Resources subACS "+ent.getKey()+" has no type "+
                        "in the resource types";
                    break;
                }
                //
                l_current = m_TypesAndExplorer_.get("Explorer.SubAcsNode");
                for (int i = 0; i < l_current.length; i++) {
                    if (l_current[i].indexOf("<byAcsCreator> "+ent.getKey()) != -1) {
                        found = true;
                        break;
                    }
                } // for i
                if (!found) {
                    message = "The Resources subACS "+ent.getKey()+" has no type "+
                        "in 'Types and their constraints for the explorer'";
                    break;
                }
            } // if

            if (Arrays.asList(l_values).contains("<GroupSubACS>") ) {
                found = false;
                for (int i = 0; i < l_epTypes_.length; i++) {
                    if (l_epTypes_[i].indexOf("<byAcsCreator> "+ent.getKey()) != -1) {
                        found = true;
                        break;
                    }
                } // for i
                if (!found) {
                    message = "The Groups subACS "+ent.getKey()+" has no type "+
                        "in the EligibleParty types";
                    break;
                }

                l_current = m_TypesAndExplorer_.get("Explorer.SubAcsNode");
                for (int i = 0; i < l_current.length; i++) {
                    if (l_current[i].indexOf("<byAcsCreator> "+ent.getKey()) != -1) {
                        found = true;
                        break;
                    }
                } // for i
                if (!found) {
                    message = "The Groups subACS "+ent.getKey()+" has no type "+
                        "in 'Types and their constraints for the explorer'";
                    break;
                }
            } // if

            if (Arrays.asList(l_values).contains("<GroupTree>") ) {
                found = false;
                for (int i = 0; i < l_epTypes_.length; i++) {
                    if (l_epTypes_[i].indexOf("<byAcsCreator> "+ent.getKey()) != -1) {
                        found = true;
                        break;
                    }
                } // for i
                if (!found) {
                    message = "The Groups tree "+ent.getKey()+" has no type "+
                        "in the EligibleParty types";
                    break;
                }

                l_current = m_TypesAndExplorer_.get("Explorer.AcsGroupTree");
                for (int i = 0; i < l_current.length; i++) {
                    if (l_current[i].indexOf("<byAcsCreator> "+ent.getKey()) != -1) {
                        found = true;
                        break;
                    }
                } // for i
                if (!found) {
                    message = "The Groups tree "+ent.getKey()+" has no type "+
                        "in 'Types and their constraints for the explorer'";
                    break;
                }
            } // if
          } // for entry
      } // if

      if (message != null) {
          l_values = null;
          l_current = null;
          return message;
      }

      l_values = null;
      l_current = null;

      l_NodesToDisplay = getL_LimitedValuesForTypesAndExplorer("Explorer.SubAcsNode");
      if (l_NodesToDisplay != null) {
          for (Iterator ite = l_NodesToDisplay.iterator(); ite.hasNext(); ) {
              key = (String)ite.next();
              if ( key.startsWith("ResourceType.<byAcsCreator> ")  ) {
                  withResourceSubAcs = true;
              }
              else if ( key.startsWith("GroupType.<byAcsCreator> ")  ) {
                  //withGroupSubAcs = true;
              }
          } // for ite
      } // if

      l_NodesToDisplay = getL_LimitedValuesForTypesAndExplorer("Explorer.AcsGroupTree");
      if (l_NodesToDisplay != null) {
          for (Iterator ite = l_NodesToDisplay.iterator(); ite.hasNext(); ) {
              if ( ((String)ite.next()).startsWith("GroupType.<byAcsCreator> ")  ) {
                  withGroupTree = true;
                  break;
              }
          } // for ite
      } // if

      l_NodesToDisplay = null;
      key = null;
      
      spTypes = getTypesInSpecializedRights();
      
      if (getManageAccounts() &&
           l_AccountRightsSet_.length == 0 && l_AclRightsSet_.length == 0 &&
           m_PrivilegeRights_.isEmpty() &&
           !spTypes.contains("AccRightsSet") && !spTypes.contains("AclRightsSet") ) {
                spTypes = null;
                return "The account, ACL or privilege rights, as specialized ones or not,"+
                      " have to be set since accounts are managed.";
      }
      if (!getManageAccounts() && (l_AccountRightsSet_.length != 0 || 
           spTypes.contains("AccRightsSet"))) {
                spTypes = null;
                return "The account rights, as specialized ones or not,"+
                      " are set but accounts are not managed.";
      }
      
      if (getManageGroups() && 
           l_GORightsSet_.length == 0 && l_AclRightsSet_.length == 0 &&
           m_PrivilegeRights_.isEmpty() &&
           !spTypes.contains("AclRightsSet") ) {
                spTypes = null;
                return "The group, ACL or privilege rights, as specialized ones or not,"+
                      " have to be set since groups are managed.";
      }
      if (!getManageGroups() && (l_GORightsSet_.length != 0 || 
           spTypes.contains("GpORightsSet"))) {
                spTypes = null;
                return "The group rights, as specialized ones or not,"+
                      " are set but groups are not managed.";
      }
      
      if (getManageInternalAcls() || getManageExternalAcls()) {
          if (l_AclRightsSet_.length == 0 && !spTypes.contains("AclRightsSet")) {
                spTypes = null;
                return "The set of ACL rights, as specialized ones or not,"+
                      " have to be defined since they are managed.";
          } // if
      } // if
      if (!getManageInternalAcls() && !getManageExternalAcls()) {
          if (l_AclRightsSet_.length != 0 || spTypes.contains("AclRightsSet")) 
                  return "The set of ACL rights, as specialized ones or not,"+
                      " is set but ACL are not managed.";
      }
      
      if (getManageBridges() && l_bridgeRightsSet_.length == 0 && 
           !spTypes.contains("BdgRightsSet")) {
                spTypes = null;
                return "The set of bridge rights, as specialized ones or not,"+
                      " have to be defined since they are managed.";
      }
      if (!getManageBridges() && (l_bridgeRightsSet_.length != 0 || 
           spTypes.contains("BdgRightsSet"))) {
                spTypes = null;
                return "The bridge rights, as specialized ones or not,"+
                      " are set but bridges are not managed.";
      }

      spTypes = null;

      if (getManagePrivilegeRights() && m_PrivilegeRights_.isEmpty()) {
                return "The set of privilege rights, as specialized ones or not,"+
                      " have to be defined since the privileges are managed.";
      }
      if (!getManagePrivilegeRights() && !m_PrivilegeRights_.isEmpty()) {
                return "The privilege rights, as specialized ones or not,"+
                      " are set while they are not managed.";
      }
      if (getManageSimpleTypedPrivileges() ) {
          // m_PrivilegeRights_ not empty, since getManagePrivileges() true
          found = false;
          for (String st : m_PrivilegeRights_.keySet() ) {
              if (st.equals("MainType."+PrivilegeAbst.TYPED) ||
                      st.startsWith("PFType")) {
                  found = true;
                  break;
              } // if
          } // for st
          if (!found)
              return "The simple typed privileges are set,"+
                      " while there is no typed privilege rights.";
      }
      if (getManageSimpleLinkedPrivileges() ) {
          // m_PrivilegeRights_ not empty, since getManagePrivileges() true
          found = false;
          for (String st : m_PrivilegeRights_.keySet() ) {
              if (st.equals("MainType."+PrivilegeAbst.LINKED) ||
                      st.startsWith("PFLink")) {
                  found = true;
                  break;
              } // if
          } // for st
          if (!found)
              return "The simple linked privileges are set,"+
                      " while there is no typed privilege rights.";
      }
      
      //if (getManagePrivilegeRights() && l_PrivRightsSet_.length == 0 &&
        //   !spTypes.contains("PrvRightsSet")) 
          //    return "The set of privileges, as specialized ones or not,"+
            //          " have to be defined since they are managed.";      
      //if (!getManagePrivilegeRights() && (l_PrivRightsSet_.length != 0 ||
        //   spTypes.contains("PrvRightsSet"))) 
          //    return "The privilege rights, as specialized ones or not,"+
            //          "are set but privileges are not managed.";
      
      if (withResourceSubAcs && !getManageResourceTrees()) {
                return "The resource trees have to be managed since"+
                      " there are subACS resources for the key 'Explorer.SubAcsNode'.";
      }
      if (withResourceSubAcs && !getManageResourceTrees()) {
                return "The resource trees have to be managed since"+
                      " there are subACS resources for the key 'Explorer.SubAcsNode'.";
      }
      // withGroupSubAcs may be true without group trees, like in Linux Ubuntu
      if (withGroupTree && !getManageGroupTrees()) {
                return "The group trees have to be managed since"+
                      " there are group trees for the key 'Explorer.AcsGroupTree'.";
      }

      if (getManageResources() && m_TypesAndExplorer_ != null &&
              m_TypesAndExplorer_.get("Directory.TypesOfChildFor."+"NULL") != null) {
          if ( ((String[])m_TypesAndExplorer_.get(
              "Directory.TypesOfChildFor."+"NULL")).length == 0) {
                    return "The resources are managed while"+
                          " all the types are forbidden by the key 'Directory.TypesOfChildFor.NULL'.";
          } // if
      } // if (getManageResourceTrees() && m_TypesAndExplorer_ != null
      if (getManageVirtualFolders() && m_TypesAndExplorer_ != null &&
              m_TypesAndExplorer_.get("VirtualFolder.TypesOfChildFor."+"NULL") != null) {
          if ( ((String[])m_TypesAndExplorer_.get(
              "VirtualFolder.TypesOfChildFor."+"NULL")).length == 0) {
                    return "The virtual folders are managed while"+
                          " all the types are forbidden by the key 'VirtualFolder.TypesOfChildFor.NULL'.";
          } // if
      } // if (getManageVirtualFolderTrees() && m_TypesAndExplorer_ != null

              // for testing and tracking
              //System.out.println("in ACSFactoryImpl.checkACSbeforeCreation(): ");
      
      if (m_AssociationsForNewObjects_ != null) {
          for (Iterator ite = m_AssociationsForNewObjects_.entrySet().iterator(); ite.hasNext(); ) {
             entry = (Map.Entry<String,String[]>)ite.next();
             if (entry.getValue() == null) {
                 message = "null actions array for "+entry.getKey();
                 break;
             }

             l_values = (String[])entry.getValue();
             if (l_values.length > 1) {
                 for (int i = 1; i < l_values.length; i++) {
                     if (!l_values[i].substring(1, 2).equals(".") &&
                         !l_values[i].substring(2, 3).equals(".") &&
                         !l_values[i].substring(3, 4).equals(".")) {
                            message = "No-numbered action for the key "+entry.getKey()+
                                    "and the value "+l_values[i];
                            break;
                     } // if
                 } // for i
             } // for i

             if (message == null) {
                 for (Iterator ita = AssociateUtilityImpl.splitActionSetsForAssociations(
                         (String[])entry.getValue() ).iterator(); ita.hasNext(); ) {
                     message = AssociateUtilityImpl.controlAssociations(
                                    this,
                                    m_AssociationsForNewObjects_,
                                    entry.getKey(), (String[])ita.next(), true);
                     if (message != null) break;
                 } // for ita
             } // if

             if (message != null) break;
          } // for ite
          if (message != null) {
              entry = null;
              name = null;
              l_values = null;
              return "The scripts for new objects are not validated; "+message;
          }
      } // if (m_AssociationsForNewObjects_

      if (m_ComponentNodes_ != null) {
          for (String[] l_val : m_ComponentNodes_.values()) {
              if (l_val == null || l_val.length != 1) {
                  message = "A component node has an unknown type value.";
                  break;
              }
          }
          if (message != null) {
              entry = null;
              name = null;
              l_values = null;
              return message;
          }
      } // if m_ComponentNodes_
      
      if (m_CompositeNodes_ != null) {
          for (String[] l_val : m_CompositeNodes_.values()) {
              if (l_val == null || l_val.length != 1) {
                  message = "A composite node has an unknown type value.";
                  break;
              }
          }
          if (message != null) return message;
      } // if m_ComponentNodes_
      
      if (m_BelongsToComposite_ != null) {
          for (ImmutableName comp : m_BelongsToComposite_.values()) {
              if (comp == null) {
                  message = "The ACS or a subACS has no defined composite that it belongs to.";
                  break;
              }
          }
          if (message != null) return message;
      } // if m_ComponentNodes_

      if (m_alternateProperties_ != null) {
          for (Map.Entry entry2 : m_alternateProperties_.entrySet()) {
              name = (Name)entry2.getKey();
              if (((String[])entry2.getValue()).length < 2 ||
                  !ISFactoryUtilityImpl.getAlternateTypes().contains(
                    ((String[])entry2.getValue())[0]) ||
                  !ISFactoryUtilityImpl.getRightTypesForAlternate().contains(
                    ((String[])entry2.getValue())[1]) ) {
                        message = name.component(0)+"."+
                                    name.subname(1, name.noOfComponents()).toString();
                        break;
              } // if
          } // for entry

          if (message != null) {
              entry = null;
              name = null;
              l_values = null;
              return "Wrong first or second alternate property for the key "+message;
          }

      } // if (m_alternateProperties_ != null)

      if (m_RootsForAcsOpeningAndClosing_ != null) {
          for (Iterator ite = m_RootsForAcsOpeningAndClosing_.entrySet().iterator(); ite.hasNext(); ) {
             entry = (Map.Entry<String,String[]>)ite.next();
             if (entry.getValue() == null) {
                 message = "null actions array for "+entry.getKey();
                 break;
             }

             l_values = (String[])entry.getValue();
             if (l_values.length > 1) {
                 for (int i = 1; i < l_values.length; i++) {
                     if (!l_values[i].substring(1, 2).equals(".") &&
                         !l_values[i].substring(2, 3).equals(".") &&
                         !l_values[i].substring(3, 4).equals(".")) {
                            message = "No-numbered action for the key "+entry.getKey()+
                                    "and the value "+l_values[i];
                            break;
                     } // if
                 } // for i
             } // for i

             if (message == null) {
                 for (Iterator ita = AssociateUtilityImpl.splitActionSetsForAssociations(
                         (String[])entry.getValue() ).iterator(); ita.hasNext(); ) {
                     message = AssociateUtilityImpl.controlAssociations(
                                    this,
                                    m_RootsForAcsOpeningAndClosing_,
                                    entry.getKey(), (String[])ita.next(), true);
                     if (message != null) break;
                 } // for ita
             } // if

             if (message != null) break;
          } // for ite

          if (message != null) {
              entry = null;
              name = null;
              l_values = null;


              System.out.println("in ACSFactoryImpl.checkACSbeforeCreation(), error message: "+
                      "Opening and closing roots are not validated - "+message);
              return "Opening and closing roots are not validated - "+message;
          }
      } // if (m_RootsForAcsOpeningAndClosing_

      if (m_RootsForAcsOpeningNoClosing_ != null) {
          for (Iterator ite = m_RootsForAcsOpeningNoClosing_.entrySet().iterator(); ite.hasNext(); ) {
             entry = (Map.Entry<String,String[]>)ite.next();
             if (entry.getValue() == null) {
                 message = "null actions array for "+entry.getKey();
                 break;
             }

             l_values = (String[])entry.getValue();
             if (l_values.length > 1) {
                 for (int i = 1; i < l_values.length; i++) {
                     if (!l_values[i].substring(1, 2).equals(".") &&
                         !l_values[i].substring(2, 3).equals(".") &&
                         !l_values[i].substring(3, 4).equals(".")) {
                            message = "No-numbered action for the key "+entry.getKey()+
                                    "and the value "+l_values[i]+" at "+i;
                            break;
                     } // if
                 } // for i
             } // for i

             if (message == null) {
                 for (Iterator ita = AssociateUtilityImpl.splitActionSetsForAssociations(
                         (String[])entry.getValue() ).iterator(); ita.hasNext(); ) {
                     message = AssociateUtilityImpl.controlAssociations(
                                    this,
                                    m_RootsForAcsOpeningNoClosing_,
                                    entry.getKey(), (String[])ita.next(), true);
                     if (message != null) break;
                 } // for ita
             } // if

             if (message != null) break;
          } // for ite

          if (message != null) {
              entry = null;
              name = null;
              l_values = null;

              System.out.println("in ACSFactoryImpl.checkACSbeforeCreation(), error message: "+
                      "Opening no closing roots are not validated - "+message);
              return "Opening no closing roots are not validated - "+message;
          }
      } // if (m_RootsForAcsOpeningNoClosing_

      if (m_RootsForAcsIntegrity_ != null) {
          for (Iterator ite = m_RootsForAcsIntegrity_.entrySet().iterator(); ite.hasNext(); ) {
             entry = (Map.Entry<String,String[]>)ite.next();
             if (entry.getValue() == null) {
                 message = "null actions array for "+entry.getKey();
                 break;
             }

             l_values = (String[])entry.getValue();
             if (l_values.length > 1) {
                 for (int i = 1; i < l_values.length; i++) {
                     if (!l_values[i].substring(1, 2).equals(".") &&
                         !l_values[i].substring(2, 3).equals(".") &&
                         !l_values[i].substring(3, 4).equals(".")) {
                            message = "No-numbered action for the key "+entry.getKey()+
                                    "and the value "+l_values[i];
                            break;
                     } // if
                 } // for i
             } // for i

             if (message == null) {
                 for (Iterator ita = AssociateUtilityImpl.splitActionSetsForAssociations(
                         (String[])entry.getValue() ).iterator(); ita.hasNext(); ) {
                     message = AssociateUtilityImpl.controlAssociations(
                                    this,
                                    m_RootsForAcsIntegrity_,
                                    entry.getKey(), (String[])ita.next(), true);
                     if (message != null) break;
                 } // for ita
             } // if

             if (message != null) break;
          } // for ite
          if (message != null) {
              entry = null;
              name = null;
              l_values = null;


              System.out.println("in ACSFactoryImpl.checkACSbeforeCreation(), error message: "+
                      "Integrity roots are not validated - "+message);
              return "Integrity roots are not validated - "+message;
          }
      } // if (m_RootsForAcsIntegrity_
      
      entry = null;     
      name = null;
      l_values = null;
      message = null;
      return null;
  } // checkACSbeforeCreation()
  
  /**
   * @return String comment which is associated to the acs. May be null.
   */
  public String getComment() {
      return comment_;
  }
  
  /**
  * Set any comment related to this acs. It is a bound property.
  * @param _s comment which is associated to the acs.
  */
  public synchronized void setComment(String _s) {
      String oldComment = comment_;
      comment_ = _s;
      try {
              firePropertyChange("Comment",oldComment,comment_);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
      }
      oldComment = null;
 }

  /**
   * The possible specialized right types are 'AccRightsSet',
   * 'GpORightsSet', 'AclRightsSet', 'BdgRightsSet'. The privilege rights
   * use another method.
   * @return the set of the used right types in the specialized rights.
   * May be empty but never null.
   */
  public Set<String> getTypesInSpecializedRights() {
      Set<String> result = new HashSet();
      String key = null;
      
      // collect in m_SpecializedRights_ the right types
      if (!m_SpecializedRights_.isEmpty()) {
          for (Iterator ite = m_SpecializedRights_.keySet().iterator(); ite.hasNext(); ) {
              key = (String)ite.next();
              if (key.startsWith("AccRightsSet", 7))
                  result.add("AccRightsSet");
              else if (key.startsWith("GpORightsSet", 7))
                  result.add("GpORightsSet");
              else if (key.startsWith("AclRightsSet", 7))
                  result.add("AclRightsSet");
              //else if (key.startsWith("PrvRightsSet", 7))
                //  result.add("PrvRightsSet");
              else if (key.startsWith("BdgRightsSet", 7))
                  result.add("BdgRightsSet");
          } // for ite
      } // if      
      return result;
  } // getTypesInSpecializedRights()

  /**
   * The possible specialized rights.
   * @return the set of the used rights in the specialized rights.
   */
  public Set<StringRight> getRightsInSpecializedRights() {
      Set<StringRight> result = new HashSet();
      Map.Entry entry = null;
      StringRight[] l_arr = null;
      
      // collect in m_SpecializedRights_ the rights
      if (!m_SpecializedRights_.isEmpty()) {
          for (Iterator ite = m_SpecializedRights_.entrySet().iterator(); ite.hasNext(); ) {
              entry = (Map.Entry)ite.next();
              l_arr = (StringRight[])entry.getValue();
              for (int i = 0; i < l_arr.length; i++) {
                  result.add(l_arr[i]);
              } // for i
          } // for ite
      } // if      
      return result;
  } // getRightsInSpecializedRights()
  
  /**
   * Initializes the ACS constraints maps before the call to ACSImpl.valueOf().
   * This method registers each used right in the rights lists,
   * sorts the global rights list, sets the Grant/Deny rights boolean.
   * To forbid these types for any creation through the beamer, this method
   * associates all the '&lt;byAcsCreator&gt;' Resource and EligibleParty types
   * to the key 'CreationByBeamer.NoType' in the specialized types map. <P>
   * Each map key is a String of ACSObject class interface (like Resource) ended by
   * a bound property name (like ResourceType), with '.' as separator.
   * For m_StandardRights_, each map key includes 'Rights' to
   * facilitate BaseUtilityImpl.getL_AcsRestrictedValues(). <P>
   * The m_TypesAndExplorer_ key 'Actor.ResourceType' is defined,
   * if Actors are managed, as the types being both into the Resource and
   * the EligibleParty types. If one of these two type sets is empty, the other set
   * is simply copied for having the Actor set. <P>
   * Caution: The design of these ACS constraints has to take account that
   * BaseUtilityImpl.getL_AcsRestrictedValues() detects the correct key to apply
   * after a search among all the interfaces of the ACSObject, and this search
   * is done without any order. The only exception is for Actor that is selected
   * before Resource. Since the first matching key is used to get
   * the constraints, it is not possible to guaranty the continuity of results if
   * two keys are applicable in the map for a given object.
   * Called by createACS(). Calls ACSFactoryUtilityImpl.getAcsImmutableRules().
   */
  public synchronized void initializeConstraintsBeforeCreation() {
      Set<StringRight> l_AllUsedRights = new HashSet(INITIAL_CAPACITY_25);
      
      if (getManageResources()) {
          m_TypesAndExplorer_.put(
            "Resource.ResourceType",
            l_resourceTypes_);

          try {
              // to forbid the '<byAcsCreator>' types for any Resource creation
              // through the beamer
              for (int i = 0; i < l_resourceTypes_.length; i++) {
                  if (l_resourceTypes_[i].indexOf("<byAcsCreator>") != -1)
                      addTypesAndExplorer(
                              "CreationByBeamer.NoType",
                              new String[] {"ResourceType."+l_resourceTypes_[i]});
              } // for i
          } catch (UpDateError u) {
              throw new InternalError("ACSFactoryImpl.initializeConstraintsBeforeCreation() "+
                      "error :"+u);
          }
      } else
          m_TypesAndExplorer_.put(
            "Resource.ResourceType",
            L_EMPTY_ARRAY);
      
      if (getManageAccounts() || getManageGroups() || getManageActors()) {
          m_TypesAndExplorer_.put(
            "EligibleParty.EPType",
            l_epTypes_);

          try {
              // to forbid the '<byAcsCreator>' types for any EligibleParty creation
              // through the beamer
              for (int i = 0; i < l_epTypes_.length; i++) {
                  if (l_epTypes_[i].startsWith("<byAcsCreator>"))
                      addTypesAndExplorer(
                              "CreationByBeamer.NoType",
                              new String[] {"EPType."+l_epTypes_[i]});
              } // for i
          } catch (UpDateError u) {
              throw new InternalError("ACSFactoryImpl.initializeConstraintsBeforeCreation() "+
                      "error :"+u);
          }
      } else
          m_TypesAndExplorer_.put(
            "EligibleParty.EPType",
            L_EMPTY_ARRAY);

      // select the types that are both Resource type and EligibleParty type
      // to put them as Actor types
      if (getManageActors() &&
           (((Object[])m_TypesAndExplorer_.get("Resource.ResourceType")).length != 0 ||
           ((Object[])m_TypesAndExplorer_.get("EligibleParty.EPType")).length != 0)) {

                String[] l_res = m_TypesAndExplorer_.get("Resource.ResourceType");
                String[] l_ep = m_TypesAndExplorer_.get("EligibleParty.EPType");
                String[] l_copy = null;

                if (l_res.length != 0 && l_ep.length != 0) {
                    Set l_sum = new HashSet(20);
                    for (int i = 0; i < l_res.length; i++) {
                        if (Arrays.asList(l_ep).contains(l_res[i]))
                            l_sum.add(l_res[i]);
                    } // for i
                    m_TypesAndExplorer_.put(
                            "Actor.ResourceType",
                            (String[]) l_sum.toArray(new String[0]));
                    l_sum = null;
                }
                else if (l_res.length != 0) {
                    l_copy = new String[l_res.length];
                    System.arraycopy(l_res, 0, l_copy, 0, l_res.length);
                    m_TypesAndExplorer_.put("Actor.ResourceType", l_copy);
                }
                else {
                    l_copy = new String[l_ep.length];
                    System.arraycopy(l_ep, 0, l_copy, 0, l_ep.length); 
                    m_TypesAndExplorer_.put("Actor.ResourceType", l_copy);
                } // else

                l_res = null;
                l_ep = null;
                l_copy = null;
      } // if (getManageActors()
      else if (getManageActors()) {
          m_TypesAndExplorer_.put(
            "Actor.ResourceType",
            L_EMPTY_ARRAY);
      } // else
      
      if (getManageVirtualFolders()) {
          m_TypesAndExplorer_.put(
            "VirtualFolder.ResourceVFType",
            l_resvfTypes_);
          m_TypesAndExplorer_.put(
            "VirtualFolder.EPVFType",
            l_epvfTypes_);

          try {
              // to forbid the '<byAcsCreator>' types for any VirtualFolder creation
              // through the beamer
              for (int i = 0; i < l_resvfTypes_.length; i++) {
                  if (l_resvfTypes_[i].indexOf("<byAcsCreator>") != -1)
                      addTypesAndExplorer(
                              "CreationByBeamer.NoType",
                              new String[] {"VirtualFolderType."+l_resvfTypes_[i]});
              } // for i
              for (int i = 0; i < l_epvfTypes_.length; i++) {
                  if (l_epvfTypes_[i].indexOf("<byAcsCreator>") != -1)
                      addTypesAndExplorer(
                              "CreationByBeamer.NoType",
                              new String[] {"VirtualFolderType."+l_epvfTypes_[i]});
              } // for i
          } catch (UpDateError u) {
              throw new InternalError("ACSFactoryImpl.initializeConstraintsBeforeCreation() "+
                      "error : "+u);
          }
      } else {
          m_TypesAndExplorer_.put(
            "VirtualFolder.ResourceVFType",
            L_EMPTY_ARRAY);
          m_TypesAndExplorer_.put(
            "VirtualFolder.EPVFType",
            L_EMPTY_ARRAY);
      } // else

      if (getManageSimpleTypedPrivileges()) {
          m_TypesAndExplorer_.put(
            "PrivilegeForType.SecondType",
            l_privilegeForTypeTypes_);
          try {
              // to forbid the '<byAcsCreator>' types for any EligibleParty creation
              // through the beamer
              for (int i = 0; i < l_privilegeForTypeTypes_.length; i++) {
                  if (l_privilegeForTypeTypes_[i].startsWith("<byAcsCreator>"))
                      addTypesAndExplorer(
                              "CreationByBeamer.NoType",
                              new String[] {"PrivilegeForTypeSecondType."+
                                      l_privilegeForTypeTypes_[i]});
              } // for i
          } catch (UpDateError u) {
              throw new InternalError("ACSFactoryImpl.initializeConstraintsBeforeCreation() "+
                      "error: "+u);
          }
      } else {
          m_TypesAndExplorer_.put(
            "PrivilegeForType.SecondType",
            L_EMPTY_ARRAY);
      }

      if (getManageSimpleLinkedPrivileges()) {
          m_TypesAndExplorer_.put(
            "PrivilegeForLinks.SecondType",
            l_privilegeForLinksTypes_);
          try {
              // to forbid the '<byAcsCreator>' types for any EligibleParty creation
              // through the beamer
              for (int i = 0; i < l_privilegeForLinksTypes_.length; i++) {
                  if (l_privilegeForLinksTypes_[i].startsWith("<byAcsCreator>"))
                      addTypesAndExplorer(
                              "CreationByBeamer.NoType",
                              new String[] {"PrivilegeForLinksSecondType."+
                                      l_privilegeForLinksTypes_[i]});
              } // for i
          } catch (UpDateError u) {
              throw new InternalError("ACSFactoryImpl.initializeConstraintsBeforeCreation() "+
                      "error: "+u);
          }
      } else {
          m_TypesAndExplorer_.put(
            "PrivilegeForLinks.SecondType",
            L_EMPTY_ARRAY);
      }
      
      if (getManageAGORights()) {
          if (getManageAccounts()) {
              m_StandardRights_.put(
                "Resource.OwnerRights",
                l_AccountRightsSet_);
              l_AllUsedRights.addAll( Arrays.asList(l_AccountRightsSet_));
          } else m_StandardRights_.put(
                "Resource.OwnerRights",
                L_EMPTY_ARRAY_2);
          
          if (getManageGroups()) {
              m_StandardRights_.put(
                "Resource.GroupRights",
                l_GORightsSet_);
              l_AllUsedRights.addAll( Arrays.asList(l_GORightsSet_));
          } else m_StandardRights_.put(
                "Resource.GroupRights",
                L_EMPTY_ARRAY_2);                      
      } else {
          m_StandardRights_.put(
              "Resource.GroupRights",
              L_EMPTY_ARRAY_2);
          m_StandardRights_.put(
              "Resource.OwnerRights",
              L_EMPTY_ARRAY_2);
      } // else

      if (getManageInternalAcls() || getManageExternalAcls()) {
          m_StandardRights_.put(
              "Resource.AclRights",
              l_AclRightsSet_);
          l_AllUsedRights.addAll( Arrays.asList(l_AclRightsSet_));
      } else {
          m_StandardRights_.put(
              "Resource.AclRights",
              L_EMPTY_ARRAY_2);   
      } 
      
      if (getManageBridges()) {
          m_StandardRights_.put(
              "Actor.BridgeRights",
              l_bridgeRightsSet_); 
          l_AllUsedRights.addAll( Arrays.asList(l_bridgeRightsSet_));
      } else {
          m_StandardRights_.put(
              "Actor.BridgeRights",
              L_EMPTY_ARRAY_2);   
      }       

      for (String st : PrivilegeAbst.getL_PrivilegeTypes()) {
          if (getManagePrivilegeRights() &&
               m_PrivilegeRights_.get("MainType."+st) != null) {
                  m_StandardRights_.put(
                      "PrivilegeRights.MainType."+st,
                      m_PrivilegeRights_.get("MainType."+st));

                  l_AllUsedRights.addAll(
                          Arrays.asList(m_PrivilegeRights_.get("MainType."+st)));
          } else {
              m_StandardRights_.put(
                  "PrivilegeRights.MainType."+st,
                  L_EMPTY_ARRAY_2);
          }
      } // for st
 
      for (String st : l_privilegeForTypeTypes_) {
          // if m_PrivilegeRights_.get("PFType.SecondType."+st) is null,
            // there is no applicable rights for the second type, and
            // the standard rights are applicable, if any
          if (m_PrivilegeRights_.get("PFType.SecondType."+st) != null)
              m_SpecializedRights_.put(
                "PFType.PrvRightsSet.Type."+st,
                m_PrivilegeRights_.get("PFType.SecondType."+st));
          else
              m_SpecializedRights_.put(
                "PFType.PrvRightsSet.Type."+st,
                L_EMPTY_ARRAY_2);
      } // for st 
      for (String st : l_privilegeForLinksTypes_) {
          if (m_PrivilegeRights_.get("PFLink.SecondType."+st) != null)
              m_SpecializedRights_.put(
                "PFLink.PrvRightsSet.Type."+st,
                m_PrivilegeRights_.get("PFLink.SecondType."+st));
          else
              m_SpecializedRights_.put(
                "PFLink.PrvRightsSet.Type."+st,
                L_EMPTY_ARRAY_2);
      } // for st

      l_AllUsedRights.addAll(getRightsInSpecializedRights());

      // updates l_Rights_ with all the used acsrights and metarights,
      // and do not delete the unused ones
      StringRight[] l_currentRights = getL_Rights();
      boolean deny = false;
      boolean grant = false;
      StringRight right = null;
      //try {
          // remove from l_Rights_ the unused rights

          //for (int i = 0; i < l_currentRights.length; i++) {
            // if (!l_AllUsedRights.contains(l_currentRights[i])) {
              //   if (!(l_currentRights[i] instanceof StringMetaRight))
                //     l_currentRights[i].finalizeForUser(); // call removeRight()
                 //else moveRight(l_currentRights[i]);
             //} // if
          //} // for i

          // some used metarights may not be registered in l_Rights_
          for (Iterator ite = l_AllUsedRights.iterator(); ite.hasNext(); ) {
              right = (StringRight)ite.next();
              if (right instanceof StringMetaRight && 
                      !Arrays.asList(l_currentRights).contains(right)) {
                            newRight(right);
              } // if
              if (right.isGrantingRight()) grant = true;
              else deny = true;
          } // for ite
      //} catch (UpDateError u) {
        //  System.out.print("ACSFactoryImpl.initializeConstraints(): "+u);
      //}
      
      // adds 'full_control' metaright, which is used by resource owner,
      // if there is no equivallent in the rights list
      boolean full_ctrl = false;
      l_currentRights = getL_Rights();
      for (int i = 0; i < l_currentRights.length; i++) {
            if (l_currentRights[i].getEorMetaRight().getRight().equals("full_control")) {
                full_ctrl = true;
                break;
            }          
      } // for i
      if (!full_ctrl)
          newRight( BaseManagerImpl.getInstance().getMetaRight("full_control") );

        // sort the rights list: NOK
      //TreeSet l_tree = new TreeSet( Arrays.asList(l_Rights_) );
      //l_Rights_ = (StringRight[]) l_tree.toArray(new StringRight[0]);
      //l_tree = null;      
      
                        // for testing and tracking
                      //System.out.println("From ACSFactoryImpl.initializeConstraints(), rights array: ");
                      //l_currentRights = getL_Rights();
                      //for (int i = 0; i < l_currentRights.length; i++) {
                        //    System.out.println("------------- at "+i+": "+l_currentRights[i]);
                      //} // for i
      
      // in l_Structure, 'acs: acsrights' has been updated by new/removeRight()
      // but not 'acs: metarights' which has been set to true in the constructor
      m_Structure_.put("access metarule: denying rights", new Boolean(deny));
      m_Structure_.put("access metarule: no granting rights", new Boolean(!grant));

      // the gWork.AcsFactory class or the user cannot add directly
      // new values in l_ImmutableRules_
      l_ImmutableRules_ = ACSFactoryUtilityImpl.getAcsImmutableRules(m_Structure_, type_);
      Collections.sort(l_ImmutableRules_);

      // in m_AcsVocabulary_, remove the keys for the classes that are not used
      if (!getManageActors())
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.ACTOR_CLASS);

      if (!getManageAccounts())
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.USER_CLASS);

      if (!getManageGroups())
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.GROUP_CLASS);

      if (!getManageResources())
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.RESOURCE_CLASS);

      if (!getManageInternalAcls() && !getManageExternalAcls())
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.ACLENTRY_CLASS);

      if (!getManageResourceTrees())
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.DIRECTORY_CLASS);

      if (!getManageAccounts() && !getManageGroups()) {
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.ELIGIBLEPARTY_CLASS);
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.GROUPMEMBER_CLASS);
      }

      if (!getManageResources() && !getManageAccounts() && !getManageGroups())
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.BASIC_CLASS);

      if (!getManageVirtualFolders() )
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.VIRTUALFOLDER_CLASS);

      if (!getManagePrivilegeRights() ) {
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.PRIVILEGEABST_CLASS);
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.PRIVILEGEFORTYPE_CLASS);
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.PRIVILEGEFORLINKS_CLASS);
      }
      else if (!getManageSimpleLinkedPrivileges() ) {
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.PRIVILEGEFORLINKS_CLASS);
      }
      else if (!getManageSimpleTypedPrivileges() ) {
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.PRIVILEGEFORTYPE_CLASS);
      }

      if (!getManageDirectoryEPs() )
          m_AcsVocabulary_.remove(ACSFactoryUtilityImpl.DIRECTORYEP_CLASS);


      l_AllUsedRights = null;
      l_currentRights = null;
      right = null;
  } // initializeConstraintsBeforeCreation()
  
  /**
   * Initializes the ACS constraints maps, after the call to ACSImpl.valueOf().
   * Creates the subACS nodes as Directories without parent,
   * with the names from the key 'Explorer.SubAcsNode'.
   * The directory may be an AcsAddon Directory, if necessary.
   * Called by createACSforGenericBase().
   * @param _acs to initialize just after its creation
   * @exception BaseError if thrown by the Directory constructor.
   */
  public synchronized void initializeConstraintsAfterCreation(ACS _acs) throws BaseError {
    String                  val = null;
    ImmutableName           name = null;
    List                    l_secondNodes = null;
    String                  className = null;
    Class                   dirClass = null;
    Class<?>[]              l_paramClass = new Class<?>[] {
                                ImmutableName.class, ACS.class, Directory.class,
                                UserID.class, String.class};
    
    l_secondNodes = ACSFactoryUtilityImpl.getL_SubAcsToDisplay(_acs);
    
    // create the to-display typed node with the user object Directory 
    // named <ffff_subacs>, placed just under the RESOURCES node
    if (l_secondNodes != null && _acs.getManageResources()) {
        try {
          for (Iterator ite = l_secondNodes.iterator(); ite.hasNext(); ) {
              val = (String)ite.next();
              if (val.startsWith("ResourceType.<byAcsCreator> ") ) {
                  name = ((Name)_acs.getName()).extended("<"+val.substring(28)+"_subacs>");
                  if (_acs instanceof ACSImpl &&
                          !_acs.containsOwnResourceName(name)) {

                                // works also for the Directory of an AcsAddon ACS
                                className = UtilityImpl.getFirstInstance().
                                                getEffectiveClassName(m_AcsVocabulary_,
                                                ACSFactoryUtilityImpl.DIRECTORY_CLASS);
                                dirClass = Class.forName(className);
                                dirClass.getConstructor(l_paramClass).newInstance(
                                    new Object[] {name, _acs, null, null, "<byAcsCreator>"});

                                        // for testing and tracking
                                        //System.out.println("ACSFactoryImpl.initializeConstraintsAfterCreation() "+
                                          //      "acs.getEorM_Resources().get(name): "+
                                            //    ((ACSImpl)_acs).getEorM_Resources().get(name)+
                                              //  " with the type <byAcsCreator> ");
                                
                  } // if (_acs
              } // if (val
          }  // for ite 
        } catch (InstantiationException i) {
            System.out.println("ACSFactoryImpl.initializeConstraintsAfterCreation()"+
                    "on a subACS Directory: "+i.toString());
            throw new BaseError(i.toString()+" on "+i.getMessage());
        } catch (IllegalArgumentException l) {
            System.out.println("ACSFactoryImpl.initializeConstraintsAfterCreation()"+
                    "on a subACS Directory: "+l.toString());
            throw new BaseError(l.toString()+" on "+l.getMessage());
        } catch (Exception e) {
            System.out.println("ACSFactoryImpl.initializeConstraintsAfterCreation()"+
                    "on a subACS Directory: "+e.toString());
            throw new BaseError(e.toString()+" on "+e.getMessage());
        }
    } // if
    
    val = null;
    name = null;
    l_secondNodes = null;
    className = null;
    dirClass = null;
    l_paramClass = null;
  } // initializeConstraintsAfterCreation()

  /**
   * Gets the allowed types for the resources.
   * The '&lt;undefined&gt;' value is always in the array.
   * @return String[] of allowed types. Never null nor empty.
   */
  public final String[] getL_ResourceTypes() {
    String[] result = new String[l_resourceTypes_.length];
    System.arraycopy(l_resourceTypes_, 0, result, 0, l_resourceTypes_.length);
    return result;
  }
  
  /**
   * Sets the types for the resources, but it is not mandatory.
   * The rule is that the 'immutable' word, if it is used, must be at the start
   * or at the end of the type. Otherwise, it is not added and no exception is thrown.
   * No action if limited set of values is not allowed.
   * The 'undefined' value is always added in the array if it is not there,
   * because the Resource BeanInfo says that the resource type is mandatory
   * in the constructor, and it is used in Privileges.
   * If the lenght of a type is superior to 100, it is not added and no exception is thrown.
   * Fires a PropertyChange for 'ResourceTypes'.
   * @param _s is the array of allowed types.
   *   May be null or contain one null element.
   */
  public synchronized void setL_ResourceTypes(String[] _s)	{
    if (l_GORightsSet_ == null) return;
    
    String[] l_oldTypes = new String[l_resourceTypes_.length];
    System.arraycopy(l_resourceTypes_, 0, l_oldTypes, 0, l_resourceTypes_.length);

    //String[] l_defaultTypes = {"full control"};
    Set l_newTypes = new HashSet(INITIAL_CAPACITY_25);
    boolean withUndefined = false;
    boolean withToDisplayType = false;
    
    //for (int i = 0; i < l_defaultTypes.length; i++) {
      //l_newTypes.add(l_defaultTypes[i]);
    //} // for i

    if (_s != null) {

      for (int i = 0; i < _s.length; i++) {
        if (_s[i] == null || _s[i].length() >= 100)
            continue;
        if (_s[i].contains("<immutable> ") &&
                !_s[i].startsWith("<immutable> ") && !_s[i].endsWith("<immutable> "))
            continue;

        l_newTypes.add(_s[i]);
        if (_s[i].equals("<undefined>")) withUndefined = true;
        if (_s[i].startsWith("<byAcsCreator> ")) withToDisplayType = true;
      } // for i

      if (!withUndefined) l_newTypes.add("<undefined>");
      if (withToDisplayType) l_newTypes.add("<byAcsCreator>");
      
    } else {
        l_newTypes.add("<undefined>");
    }
    
    l_resourceTypes_ = (String[])l_newTypes.toArray(new String[l_newTypes.size()]);

    try {
        firePropertyChange("ResourceTypes", l_oldTypes,
           l_newTypes.toArray(new String[l_newTypes.size()]));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_ResourceTypes(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldTypes = null;
    l_newTypes = null;

  } // setL_ResourceTypes 
  
  /**
   * Sets the optional hub property for the ACS in the IS nametree.
   * Fires a PropertyChange for 'HubProperty'.
   * @since 0.7.0
   * @param _is_hub is the hub property
   */
  public synchronized void setHubProperty(boolean _is_hub)	{
    if (iSname_ == null ) return;
    
    boolean old = is_hub_;
    is_hub_ = _is_hub;

    try {
        firePropertyChange("HubProperty", new Boolean(old), new Boolean(_is_hub));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_EPTypes(), firePropertChange() throws:");
        e.printStackTrace();
    }

  } // setHubProperty()
   
  /**
   * Set the visible space for the ACS in the IS nametree.
   * The default value is NameISTreeImpl.ALL_THE_IS_SPACE.
   * Fires a PropertyChange for 'VisibleSpace'.
   * @since 0.7.0
   * @param _visibleSpace is the visible space from the ACS in the IS nametree.
   *   Has to be a NameISTreeImpl constant.
   */
  public synchronized void setVisibleSpace(String _visibleSpace) {
    if (iSname_ == null ) return;
    if ( !Arrays.asList( NameISTreeImpl.EMPTY_INSTANCE.getExtendedVisibleSpaceConstants()).
       contains(_visibleSpace) ) return;
    
    String old = visibleSpace_;
    visibleSpace_ = _visibleSpace;

    try {
        firePropertyChange("VisibleSpace", old, _visibleSpace);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_EPTypes(), firePropertChange() throws:");
        e.printStackTrace();
    }

    old = null;

  } // setVisibleSpace()

  /**
   * Gets the allowed types for the EPs.
   * The '&lt;undefined>' value is always in the array.
   * @return String[] of allowed types. Never null nor empty.
   */
  public final String[] getL_EPTypes() {
    String[] result = new String[l_epTypes_.length];
    System.arraycopy(l_epTypes_, 0, result, 0, l_epTypes_.length);
    return result;
  }
  
  /**
   * Sets the types for the EPs, but it is not mandatory.
   * The rule is that the 'immutable' word, if it is used, must be at the start
   * or at the end of the type. Otherwise, it is not added and no exception is thrown.
   * The 'undefined' value is always added in the array if it is not there,
   * because the EP BeanInfo says that the type is mandatory in the constructor.
   * The 'byAcsCreator' value is added if the argument contains
   * at most one value with the format 'byAcsCreator ffff'.
   * If the lenght of a type is superior to 100, it is not added and no exception is thrown.
   * Fires a PropertyChange for 'EPTypes'.
   * @param _s allowed types. May be null or contain one null element.
   */
  public synchronized void setL_EPTypes(String[] _s)	{
    if (l_GORightsSet_ == null) return;
    
    String[] l_oldTypes = new String[l_epTypes_.length];
    System.arraycopy(l_epTypes_, 0, l_oldTypes, 0, l_epTypes_.length);

    Set l_newTypes = new HashSet(INITIAL_CAPACITY_25);
    boolean withUndefined = false;
    boolean withToDisplayType = false;
    
    //for (int i = 0; i < l_defaultRights.length; i++) {
      //l_newRights.add(l_defaultRights[i]);
    //} // for i

    if (_s != null) {

      for (int i = 0; i < _s.length; i++) {
        if (_s[i] == null || _s[i].length() >= 100)
            continue;
        if (_s[i].contains("<immutable> ") &&
                !_s[i].startsWith("<immutable> ") && !_s[i].endsWith("<immutable> "))
            continue;

        l_newTypes.add(_s[i]);
        if (_s[i].equals("<undefined>")) withUndefined = true;
        if (_s[i].startsWith("<byAcsCreator> ")) withToDisplayType = true;
      } // for i

      if (!withUndefined) l_newTypes.add("<undefined>");
      if (withToDisplayType) l_newTypes.add("<byAcsCreator>");
    } else {
        l_newTypes.add("<undefined>");
    }
    
    l_epTypes_ = (String[])l_newTypes.toArray(new String[l_newTypes.size()]);

    try {
        firePropertyChange("EPTypes", l_oldTypes,
           l_newTypes.toArray(new String[l_newTypes.size()]));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_EPTypes(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldTypes = null;
    l_newTypes = null;

  } // setL_EPTypes()

  /**
   * Gets the allowed types for the Resource Virtual Folders.
   * The '&lt;undefined&gt;' value is always in the array.
   * @since 0.7.0
   * @return String[] of types. Never null nor empty.
   */
  public final String[] getL_ResourceVirtualFolderTypes() {
    String[] result = new String[l_resvfTypes_.length];
    System.arraycopy(l_resvfTypes_, 0, result, 0, l_resvfTypes_.length);
    return result;
  }

  /**
   * Gets the allowed types for the Eligible Party Virtual Folders.
   * The '&lt;undefined&gt;' value is always in the array.
   * @since 0.7.0
   * @return String[] of types. Never null nor empty.
   */
  public final String[] getL_EPVirtualFolderTypes() {
    String[] result = new String[l_epvfTypes_.length];
    System.arraycopy(l_epvfTypes_, 0, result, 0, l_epvfTypes_.length);
    return result;
  }
  
  /**
   * Sets the types for the Resource VirtualFolders, but it is not mandatory.
   * The rule is that the 'immutable' word, if it is used, must be at the start
   * or at the end of the type. Otherwise, it is not added and no exception is thrown.
   * The 'undefined' value is always added in the array if it is not there,
   * because the VirtualFolder BeanInfo says that the type argument is mandatory
   * in the VirtualFolder constructor.
   * The 'byAcsCreator' value is added if the argument contains
   * at most one value with the format 'byAcsCreator ffff'.
   * If the lenght of a type is superior to 100, it is not added and no exception is thrown.
   * Fires a PropertyChange for 'ResVirtualFolderTypes'.
   * @since 0.7.0
   * @param _s String[] of types.
   *   May be null or contain one null element.
   */
  public synchronized void setL_ResourceVirtualFolderTypes(String[] _s)	{
    if (l_GORightsSet_ == null) return;
    
    String[] l_oldTypes = new String[l_resvfTypes_.length];
    System.arraycopy(l_resvfTypes_, 0, l_oldTypes, 0, l_resvfTypes_.length);

    Set l_newTypes = new HashSet(INITIAL_CAPACITY_25);
    boolean withUndefined = false;
    boolean withToDisplayType = false;
    
    //for (int i = 0; i < l_defaultRights.length; i++) {
      //l_newRights.add(l_defaultRights[i]);
    //} // for i

    if (_s != null) {

      for (int i = 0; i < _s.length; i++) {
        if (_s[i] == null || _s[i].length() >= 100)
            continue;
        if (_s[i].contains("<immutable> ") &&
                !_s[i].startsWith("<immutable> ") && !_s[i].endsWith("<immutable> "))
            continue;
        
        l_newTypes.add(_s[i]);
        if (_s[i].equals("<undefined>")) withUndefined = true;
        if (_s[i].startsWith("<byAcsCreator> ")) withToDisplayType = true;
      } // for i

      if (!withUndefined) l_newTypes.add("<undefined>"); // done if _s is empty
      if (withToDisplayType) l_newTypes.add("<byAcsCreator>");
    } else {
        l_newTypes.add("<undefined>");
    }
    
    l_resvfTypes_ = (String[])l_newTypes.toArray(new String[l_newTypes.size()]);

    try {
        firePropertyChange("ResVirtualFolderTypes", l_oldTypes,
           l_newTypes.toArray(new String[l_newTypes.size()]));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_VirtualFolderTypes(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldTypes = null;
    l_newTypes = null;

  } // setL_ResourceVirtualFolderTypes()
  
  /**
   * Sets the types for the Eligible Party VirtualFolders, but it is not mandatory.
   * The rule is that the 'immutable' word, if it is used, must be at the start
   * or at the end of the type. Otherwise, it is not added and no exception is thrown.
   * The 'undefined' value is always added in the array if it is not there,
   * because the VirtualFolder BeanInfo says that the type argument is mandatory
   * in the VirtualFolder constructor.
   * The 'byAcsCreator' value is added if the argument contains
   * at most one value with the format 'byAcsCreator ffff'.
   * If the lenght of a type is superior to 100, it is not added and no exception is thrown.
   * Fires a PropertyChange for 'EPVirtualFolderTypes'.
   * @since 0.7.0
   * @param _s String[] of types.
   *   May be null or contain one null element.
   */
  public synchronized void setL_EPVirtualFolderTypes(String[] _s)	{
    if (l_GORightsSet_ == null) return;
    
    String[] l_oldTypes = new String[l_epvfTypes_.length];
    System.arraycopy(l_epvfTypes_, 0, l_oldTypes, 0, l_epvfTypes_.length);

    Set l_newTypes = new HashSet(INITIAL_CAPACITY_25);
    boolean withUndefined = false;
    boolean withToDisplayType = false;
    
    //for (int i = 0; i < l_defaultRights.length; i++) {
      //l_newRights.add(l_defaultRights[i]);
    //} // for i

    if (_s != null) {
      for (int i = 0; i < _s.length; i++) {
        if (_s[i] == null || _s[i].length() >= 100)
            continue;
        if (_s[i].contains("<immutable> ") &&
                !_s[i].startsWith("<immutable> ") && !_s[i].endsWith("<immutable> "))
            continue;

        l_newTypes.add(_s[i]);
        if (_s[i].equals("<undefined>")) withUndefined = true;
        if (_s[i].startsWith("<byAcsCreator> ")) withToDisplayType = true;
      } // for i

      if (!withUndefined) l_newTypes.add("<undefined>"); // done if _s is empty
      if (withToDisplayType) l_newTypes.add("<byAcsCreator>");
    } else {
        l_newTypes.add("<undefined>");
    }
    
    l_epvfTypes_ = (String[])l_newTypes.toArray(new String[l_newTypes.size()]);

    try {
        firePropertyChange("EPVirtualFolderTypes", l_oldTypes,
           l_newTypes.toArray(new String[l_newTypes.size()]));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_VirtualFolderTypes(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldTypes = null;
    l_newTypes = null;

  } // setL_EPVirtualFolderTypes()

  /**
   * Gets the allowed second types for the PrivilegeForTypes.
   * The '&lt;undefined&gt;' value is always in the array.
   * @since 0.7.0
   * @return String[] of types. Never null nor empty.
   */
  public final String[] getL_PrivilegeForTypeTypes() {
    String[] result = new String[l_privilegeForTypeTypes_.length];
    System.arraycopy(l_privilegeForTypeTypes_, 0, result, 0, l_privilegeForTypeTypes_.length);
    return result;
  }
  
  /**
   * Sets the second types for the PrivilegeForTypes, but it is not mandatory.
   * The '&lt;undefined>' value is always added in the array if it is not there,
   * because the Privilege BeanInfos state that the type is mandatory in the constructor.
   * The '&lt;byAcsCreator>' value is added if the argument contains
   * at most one value with the format '<byAcsCreator> ffff'.
   * If the lenght of a type is superior to 100, it is not added and no exception is thrown.
   * Fires a PropertyChange for 'PrivilegeForTypeTypes'.
   * @since 0.7.0
   * @param _s String[] of types.
   *   May be null or contain one null element.
   */
  public synchronized void setL_PrivilegeForTypeTypes(String[] _s)	{
    if (l_GORightsSet_ == null) return;
    
    String[] l_oldTypes = new String[l_privilegeForTypeTypes_.length];
    System.arraycopy(l_privilegeForTypeTypes_, 0, l_oldTypes, 0, l_privilegeForTypeTypes_.length);

    Set l_newTypes = new HashSet(INITIAL_CAPACITY_25);
    boolean withUndefined = false;
    boolean withToDisplayType = false;

    if (_s != null) {
      for (int i = 0; i < _s.length; i++) {
        if (_s[i] != null && _s[i].length() < 100) {
            l_newTypes.add(_s[i]);
            if (_s[i].equals("<undefined>")) withUndefined = true;
            if (_s[i].startsWith("<byAcsCreator> ")) withToDisplayType = true;
        } // if
      } // for i
      if (!withUndefined) l_newTypes.add("<undefined>");
      if (withToDisplayType) l_newTypes.add("<byAcsCreator>");
    } else {
        l_newTypes.add("<undefined>");
    }
    
    l_privilegeForTypeTypes_ = (String[])l_newTypes.toArray(new String[l_newTypes.size()]);

    try {
        firePropertyChange("PrivilegeForTypeTypes", l_oldTypes,
           l_newTypes.toArray(new String[l_newTypes.size()]));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_PrivilegeForTypeTypes(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldTypes = null;
    l_newTypes = null;
  } // setL_PrivilegeForTypeTypes()
 
  /**
   * Gets the allowed second types for the PrivilegeForLinks.
   * The '&lt;undefined>' value is always in the array.
   * @return String[] of types. Never null nor empty.
   */
  public final String[] getL_PrivilegeForLinksTypes() {
    String[] result = new String[l_privilegeForLinksTypes_.length];
    System.arraycopy(l_privilegeForLinksTypes_, 0, result, 0, l_privilegeForLinksTypes_.length);
    return result;
  }
  
  /**
   * Sets the second types for the PrivilegeForLinks, but it is not mandatory.
   * The '&lt;undefined&gt;' value is always added in the array if it is not there,
   * because the Privilege BeanInfos state that the type is mandatory in the constructor.
   * The '&lt;byAcsCreator&gt;' value is added if the argument contains
   * at most one value with the format '&lt;byAcsCreator&gt; ffff'.
   * If the lenght of a type is superior to 100, it is not added and no exception is thrown.
   * Fires a PropertyChange for 'PrivilegeForLinksTypes'.
   * @since 0.7.0
   * @param _s String[] of types.
   *   May be null or contain one null element.
   */
  public synchronized void setL_PrivilegeForLinksTypes(String[] _s)	{
    if (l_GORightsSet_ == null) return;
    
    String[] l_oldTypes = new String[l_privilegeForLinksTypes_.length];
    System.arraycopy(l_privilegeForLinksTypes_, 0, l_oldTypes, 0, l_privilegeForLinksTypes_.length);

    Set l_newTypes = new HashSet(INITIAL_CAPACITY_25);
    boolean withUndefined = false;
    boolean withToDisplayType = false;

    if (_s != null) {
      for (int i = 0; i < _s.length; i++) {
        if (_s[i] != null && _s[i].length() < 100) {
            l_newTypes.add(_s[i]);
            if (_s[i].equals("<undefined>")) withUndefined = true;
            if (_s[i].startsWith("<byAcsCreator> ")) withToDisplayType = true;
        } // if
      } // for i
      if (!withUndefined) l_newTypes.add("<undefined>");
      if (withToDisplayType) l_newTypes.add("<byAcsCreator>");
    } else {
        l_newTypes.add("<undefined>");
    }
    
    l_privilegeForLinksTypes_ = (String[])l_newTypes.toArray(new String[l_newTypes.size()]);

    try {
        firePropertyChange("PrivilegeForLinksTypes", l_oldTypes,
           l_newTypes.toArray(new String[l_newTypes.size()]));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_PrivilegeForLinksTypes(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldTypes = null;
    l_newTypes = null;
  } // setL_PrivilegeForLinksTypes()

  /**
   * The full name is unique for the Access Road program.
   * Its is equal to the ACS name.
   * @return a copy of the name. May be empty, but not null.
   */
    public ImmutableName getFullName() {
       return getName();
    }
  
  /**
   * Gets the futur ACS name, including the IS name if any.
   * @return ImmutableName of the acs. May be empty, but not null.
   */
  public final ImmutableName getName() {
      if (name_ == null) return NameImpl.EMPTY_INSTANCE;
      else return (ImmutableName)name_.clone();       
  } // getName()
    
   /**
   * This short name is NOT always unique for the Access Road program.
   * @return the ACS name last component. Never null.
   */
   public String getNickName() {
        if (name_ == null) return "";
        else return name_.lastComponent();
   }
  
  /**
   * Returns the ACS type, as ACSImpl.GENERIC, LINUX,...
   * or a user-defined type.
   * @return the type
   */
  public String getACSType() {
        return type_;
  }

  /**
   * The modeling of the structure is incomplete when some ACSObjects miss
   * while they participate to the decision for some important access controls,
   * or simply when these missed ACSObjects are important for the overall ACS security,
   * or when their important relations with some other ACSObjects are not set.
   * On the other hand, the modeling remains complete if some secondary ACSObjects
   * are not modelled, like for instance image files in an operating system file system.
   * When the modeling of an ACS structure is incomplete, all the views which
   * use this ACS are set incomplete in their results.
   * @since 0.7.0
   * @return true if the modeling of the ACS structure is incomplete.
   */
  public boolean getIncompleteStructureModeling() {
          Boolean result = m_Structure_.get(
                  "access metarule: incomplete modeling of the ACS structure");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * Note: if the structure map does not contain the key
   * 'access metarule: incomplete modeling of the ACS structure', this method
   * throws a NullPointerException.
   * @since 0.7.0
   * @param _b true if the modeling of the ACS structure is incomplete.
   */
  public void setIncompleteStructureModeling(boolean _b) {

                          // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.setIncompleteStructureModeling(), "+
                      //      "m_Structure_ at the start: "+m_Structure_);

      boolean old = (boolean)m_Structure_.get(
              "access metarule: incomplete modeling of the ACS structure");
      if (old != _b) {
          m_Structure_.put(
                  "access metarule: incomplete modeling of the ACS structure",
                  new Boolean(_b));
          firePropertyChange("IncompleteStructureModeling", old, _b);
      }
  }

  /**
   * The modeling of the behavior is incomplete when important constraints on
   * ACSObjects are not modelled, for instance to set their allowed rights, or
   * their effective default rights.
   * On the other hand, the modeling remains complete if some secondary relations
   * or constraints are not modelled by the program, like for instance the presence
   * of image files in an operating system file system for a given application software.
   * The user may complements manually the properties of an ACSObject for which
   * the behavior has not been totally defined in the ACS model.
   * When the modeling of an ACS behavior is incomplete, the views which
   * use this ACS remain complete in their results, but the creation or deletion
   * of a new ACSObject should be reviewed by the user to ensure that all the properties
   * are still well defined in the ACS.
   * @since 0.7.0
   * @return true if the modeling of the ACS behavior is incomplete.
   */
  public boolean getIncompleteBehaviorModeling() {
          Boolean result = m_Structure_.get(
                  "access metarule: incomplete modeling of the ACS behavior");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the modeling of the ACS behavior is incomplete.
   */
  public void setIncompleteBehaviorModeling(boolean _b) {
      boolean old = (boolean)m_Structure_.get(
              "access metarule: incomplete modeling of the ACS behavior");
      if (old != _b) {
          m_Structure_.put(
                  "access metarule: incomplete modeling of the ACS behavior",
                  new Boolean(_b));
          firePropertyChange("IncompleteBehaviorModeling", old, _b);
      }
  }

  /**
   * @return true if the resources are managed. 
   */
  public boolean getManageResources() {
          Boolean result = m_Structure_.get("resource: no resources");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * @param _b true if the resources are managed.
   */
  public void setManageResources(boolean _b) {
      boolean old = (boolean)m_Structure_.get("resource: no resources");
      if (old != !_b) {
          m_Structure_.put("resource: no resources", new Boolean(!_b));
          firePropertyChange("ManageResources", !old, _b);    
      }
  }
  
  /**
   * @return true if the resource trees are managed.
   */
  public boolean getManageResourceTrees() {
          Boolean result = m_Structure_.get("resource: directories of resources trees");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @param _b true if the resource directories are managed.
   */
  public void setManageResourceTrees(boolean _b) {
      boolean old = (boolean)m_Structure_.get("resource: directories of resources trees");
      if (old != _b) {
          m_Structure_.put("resource: directories of resources trees", new Boolean(_b));
          firePropertyChange("ManageResourceTrees", old, _b);    
      }
  }
  
  /**
   * @return true if the accounts (or users) are managed.
   */
  public boolean getManageAccounts() {
          Boolean result = m_Structure_.get("EP: accounts");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @param _b true if the accounts are managed.
   */
  public void setManageAccounts(boolean _b) {
      boolean old = (boolean)m_Structure_.get("EP: accounts");
      if (old != _b) {
          m_Structure_.put("EP: accounts", new Boolean(_b));
          firePropertyChange("ManageAccounts", old, _b);  
      }
  }
  
  /**
   * @return true if the groups of accounts are managed. 
   */
  public boolean getManageGroups() {
          Boolean result = m_Structure_.get("EP: groups of accounts");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @param _b true if the groups of accounts are managed.
   */
  public void setManageGroups(boolean _b) {
      boolean old = (boolean)m_Structure_.get("EP: groups of accounts");
      if (old != _b) {
          m_Structure_.put("EP: groups of accounts", new Boolean(_b));
          firePropertyChange("ManageGroupsOfAccounts", old, _b);  
      }
  }
  
  /**
   * Caution: the name of this method is confusing, because a collection of
   * nested groups may be more than one single group tree.
   * @return true if the nested groups are managed.
   */
  public boolean getManageGroupTrees() {
          Boolean result = m_Structure_.get("EP: group of groups");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * Caution: the name of this method is confusing, because a collection of
   * nested groups may be more than one single group tree.
   * @param _b true if the nested groups are managed.
   */
  public void setManageGroupTrees(boolean _b) {
      boolean old = (boolean)m_Structure_.get("EP: group of groups");
          
      if (old != _b) {
            m_Structure_.put("EP: group of groups", new Boolean(_b));
            firePropertyChange("ManageGroupsOfGroups", old, _b);  
      }
  }
  
  /**
   * @return true if the internal acl entries are managed.
   */
  public boolean getManageInternalAcls() {
          Boolean result = m_Structure_.get("ACL right: internal ACL");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @param _b true if the internal acl entries are managed.
   */
  public void setManageInternalAcls(boolean _b) {
      boolean old = (boolean)m_Structure_.get("ACL right: internal ACL");
 
            // for testing and tracking
            //System.out.println("From ACSFactoryImpl.setManageInternalAcls(), "+
              //      "old value: "+old+" and new value: "+_b);
      
      if (old != _b) {
            m_Structure_.put("ACL right: internal ACL", new Boolean(_b));
            firePropertyChange("ManageInternalAcls", old, _b);
      }
  }
  
  /** 
   * @return true if the external acl entries are managed, as
   *   controler of other-acs objects, and/or as acs which is controled by other acs.
   */
  public boolean getManageExternalAcls() {
      Boolean one = m_Structure_.get("ACL right: external ACL with entity in this acs");
      Boolean two = m_Structure_.get("ACL right: pure external ACL without entity in this acs");
      if ((one != null && one.booleanValue()) || (two != null && two.booleanValue())) {
          one = null;
          two = null;
          return true;
      }
      one = null;
      two = null;
      return false;
  }

  /**
   * @param _b true if the external acl entries are managed.
   */
  public void setManageExternalAcls(boolean _b) {
      boolean old = false;
      if ((boolean)m_Structure_.get("ACL right: external ACL with entity in this acs")
           || (boolean)m_Structure_.get("ACL right: pure external ACL without entity in this acs"))
                old = true;
      
      if (old != _b) {
            m_Structure_.put("ACL right: external ACL with entity in this acs", new Boolean(_b));
            m_Structure_.put("ACL right: pure external ACL without entity in this acs", new Boolean(_b));
            firePropertyChange("ManageExternalAcls", old, _b);
      }
  }
  
  /** 
   * @since 0.7.0
   * @return true if the flexible conditional AclEntries are managed
   */
  public boolean getManageFlexibleConditionalAcls() {
      Boolean result = m_Structure_.get("ACL right: flexible conditional ACL with "+
          "condition groups the groupIDmember has to be member");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the flexible conditional AclEntries are managed.
   */
  public void setManageFlexibleConditionalAcls(boolean _b) {
      boolean old = (boolean)m_Structure_.get("ACL right: flexible conditional ACL with "+
          "condition groups the groupIDmember has to be member");
          
      if (old != _b) {
            m_Structure_.put("ACL right: flexible conditional ACL with "+
          "condition groups the groupIDmember has to be member", new Boolean(_b));
            firePropertyChange("ManageFlexibleConditionalAcls", old, _b);
      }
  }
  
  /** 
   * @since 0.7.0
   * @return true if the right-defined conditional AclEntries are managed.
   */
  public boolean getManageRightDefinedConditionalAcls() {
      Boolean result = m_Structure_.get("ACL right: right-defined conditional ACL with "+
            "predefined condition groups the groupIDmember has to be member");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the right-defined conditional AclEntries are managed.
   */
  public void setManageRightDefinedConditionalAcls(boolean _b) {
      boolean old = (boolean)m_Structure_.get("ACL right: right-defined conditional ACL with "+
            "predefined condition groups the groupIDmember has to be member");
          
      if (old != _b) {
            m_Structure_.put("ACL right: right-defined conditional ACL with "+
            "predefined condition groups the groupIDmember has to be member", new Boolean(_b));
            firePropertyChange("ManageRightDefinedConditionalAcls", old, _b);
      }
  }
  
  /**
   * @since 0.7.0
   * @return true if the aclentry rights inheritance from the parent are managed
   * (not necessary all these rights).
   */
  public boolean getManageACLRightsInheritance() { 
      Boolean result = m_Structure_.get("ACL inheritance 1: ACL inheritance");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the aclentry rights inheritance from the parent are managed.
   */
  public void setManageACLRightsInheritance(boolean _b) {
      boolean old = (boolean)m_Structure_.get("ACL inheritance 1: ACL inheritance");
          
      if (old != _b) {
            m_Structure_.put("ACL inheritance 1: ACL inheritance", new Boolean(_b));
            firePropertyChange("ManageACLRightInheritance", old, _b);
      }
  }
  
  /**
   * @return true if the actors are managed. 
   */
  public boolean getManageActors() {
          Boolean result = m_Structure_.get("resource: no actors");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * @param _b true if the actors are managed.
   */
  public void setManageActors(boolean _b) {
      boolean old = (boolean)m_Structure_.get("resource: no actors");
          
      if (old != !_b) {
            m_Structure_.put("resource: no actors", new Boolean(!_b));
            firePropertyChange("ManageActors", !old, _b);
      }
  }
  
  /**
   * @return true if the ACS bridges are managed.
   */
  public boolean getManageBridges() {
      Boolean result = m_Structure_.get("bridge right: no BriDGes");

                // for testing and tracking
                //System.out.println("Call of ACSFactoryImpl.getManageBridges() ");


      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @param _b true if the ACS bridges are managed.
   */
  public void setManageBridges(boolean _b) {
      boolean old = (boolean)m_Structure_.get("bridge right: no BriDGes");

                    // for testing and tracking
                    //System.out.println("Call of ACSFactoryImpl.setManageBridges("+_b+") from "+!old);  
      
      if (old != !_b) {
            m_Structure_.put("bridge right: no BriDGes", new Boolean(!_b));
            firePropertyChange("ManageBridges", !old, _b);
      }
  }
  
  /**
   * @return true if the secondary groups are managed for actors (no
   *   answer on directoryEPs). 
   */
  public boolean getManageSecGroupsForActor() {
      Boolean one = m_Structure_.get("EP 1: secondary groups for actors as right users");
      Boolean two = m_Structure_.get("EP: secondary groups for actors, excluding area groups");
      if ((one != null && one.booleanValue()) || (two != null && two.booleanValue())) {
          one = null;
          two = null;
          return true;
      }
      one = null;
      two = null;
      return false;     
  }

  /**
   * @param _b true if the secondary groups for actors are managed.
   */
  public void setManageSecGroupsForActor(boolean _b) {
      boolean old = false;
      if ((boolean)m_Structure_.get("EP 1: secondary groups for actors as right users")
           || (boolean)m_Structure_.get("EP: secondary groups for actors, excluding area groups"))
                old = true;
      if (old != _b) {
          m_Structure_.put("EP 1: secondary groups for actors as right users", new Boolean(_b));
          m_Structure_.put("EP: secondary groups for actors, excluding area groups", new Boolean(_b));
          firePropertyChange("ManageSecondaryGroupsForActor", old, _b);    
      }
  }
  
  /**
   * @return true if Account-Group rights are managed (but not
   * necessary all these rights).
   */
  public boolean getManageAGORights() { 
      Boolean result = m_Structure_.get("AGO right: Account and Group rights");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @param _b true if the Account/Group rights are managed.
   */
  public void setManageAGORights(boolean _b) {
      boolean old = (boolean)m_Structure_.get("AGO right: Account and Group rights");
          
      if (old != _b) {
            m_Structure_.put("AGO right: Account and Group rights", new Boolean(_b));
            firePropertyChange("ManageAGORight", old, _b);
      }
  }
  
  /**
   * @since 0.7.0
   * @return true if Account/Group rights inheritance from the parent are managed (not
   * necessary all these rights).
   */
  public boolean getManageAGORightsInheritance() { 
      Boolean result = m_Structure_.get("AGO inheritance 1: AG inheritance");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the Account/Group rights inheritance from the parent are managed.
   */
  public void setManageAGORightsInheritance(boolean _b) {
      boolean old = (boolean)m_Structure_.get("AGO inheritance 1: AG inheritance");
          
      if (old != _b) {
            m_Structure_.put("AGO inheritance 1: AG inheritance", new Boolean(_b));
            firePropertyChange("ManageAGORightInheritance", old, _b);
      }
  }

  /**
   * @return true if the directoryEPs are managed.
   */
  public boolean getManageDirectoryEPs() {
      Boolean result = m_Structure_.get("resource: eligible party directories");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @param _b true if the directoryEP are managed.
   */
  public void setManageDirectoryEPs(boolean _b) {
      boolean old = (boolean)m_Structure_.get("resource: eligible party directories");
         
      if (old != _b) {
            m_Structure_.put("resource: eligible party directories", new Boolean(_b));
            firePropertyChange("ManageDirectoryEPs", old, _b); 
      }
   return;
  }
  
  /**
   * @since 0.7.0
   * @return true if the privilege rights are managed.
   */
  public boolean getManagePrivilegeRights() {
      Boolean result = m_Structure_.get("PRI right: no PRI rights");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the privilege rights are managed.
   */
  public void setManagePrivilegeRights(boolean _b) {
      boolean old = (boolean)m_Structure_.get("PRI right: no PRI rights");
          
      if (old != !_b) {
            m_Structure_.put("PRI right: no PRI rights", new Boolean(!_b));
            firePropertyChange("ManagePrivilegeRights", !old, _b); 
      }
  }
  
  /**
   * @since 0.7.0
   * @return true if the simple linked privilege rights are managed.
   */
  public boolean getManageSimpleLinkedPrivileges() {
      Boolean result = m_Structure_.get("PRI right: simple LPRI");
      if (result != null && result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the simple linked privilege rights are managed.
   */
  public void setManageSimpleLinkedPrivileges(boolean _b) {
      boolean old = (boolean)m_Structure_.get("PRI right: simple LPRI");
          
      if (old != _b) {
            m_Structure_.put("PRI right: simple LPRI", new Boolean(_b));
            firePropertyChange("ManageSimpleLinkedPrivileges", old, _b);
      }
  }
  
  /**
   * @since 0.7.0
   * @return true if the simple typed privilege rights are managed. 
   */
  public boolean getManageSimpleTypedPrivileges() {
      Boolean result = m_Structure_.get("PRI right: simple TPRI");
      if (result != null && result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the simple typed privilege rights are managed.
   */
  public void setManageSimpleTypedPrivileges(boolean _b) {
      boolean old = (boolean)m_Structure_.get("PRI right: simple TPRI");
          
      if (old != _b) {
            m_Structure_.put("PRI right: simple TPRI", new Boolean(_b));
            firePropertyChange("ManageSimpleTypedPrivileges", old, _b); 
      }
  }

  /**
   * @since 0.7.0
   * @return true if the granting rights in privileges are managed.
   */
  public boolean getManageGrantingRightPrivileges() {
      Boolean result = m_Structure_.get("PRI right: no granting rights");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the granting rights in privileges are managed.
   */
  public void setManageGrantingRightPrivileges(boolean _b) {
      boolean old = (boolean)m_Structure_.get("PRI right: no granting rights");

      if (old != !_b) {
            m_Structure_.put("PRI right: no granting rights", new Boolean(!_b));
            firePropertyChange("ManageGrantingRightPrivileges", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the denying rights in privileges are managed.
   */
  public boolean getManageDenyingRightPrivileges() {
      Boolean result = m_Structure_.get("PRI right: no denying rights");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the denying rights in privileges are managed.
   */
  public void setManageDenyingRightPrivileges(boolean _b) {
      boolean old = (boolean)m_Structure_.get("PRI right: no denying rights");

      if (old != !_b) {
            m_Structure_.put("PRI right: no denying rights", new Boolean(!_b));
            firePropertyChange("ManageDenyingRightPrivileges", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the privilege rights in Resources are managed.
   */
  public boolean getManagePrivilegesInResources() {
          Boolean result = m_Structure_.get("PRI right: for Resource");
          if (result != null && result.booleanValue()) return true;
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the privilege rights in Resources are managed.
   */
  public void setManagePrivilegesInResources(boolean _b) {
      boolean old = (boolean)m_Structure_.get("PRI right: for Resource");

      if (old != _b) {
            m_Structure_.put("PRI right: for Resource", new Boolean(_b));
            firePropertyChange("ManagePrivilegesInResources", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the privilege rights in VirtualFolders are managed.
   */
  public boolean getManagePrivilegesInVirtualFolders() {
          Boolean result = m_Structure_.get("PRI right: for VirtualFolder");
          if (result != null && result.booleanValue()) return true;
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the privilege rights in VirtualFolders are managed.
   */
  public void setManagePrivilegesInVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get("PRI right: for VirtualFolder");
          
      if (old != _b) {
            m_Structure_.put("PRI right: for VirtualFolder", new Boolean(_b));
            firePropertyChange("ManagePrivilegesInVirtualFolders", old, _b);
      }
  }
  
  /**
   * @since 0.7.0
   * @return true if the linked privilege rights inheritance from the parent are managed
   * (not necessary all these rights).
   */
  public boolean getManageLPRIRightsInheritance() { 
      Boolean result = m_Structure_.get("LPRI inheritance 1: Linked PRI inheritance");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the linked privilege rights inheritance from the parent are managed.
   */
  public void setManageLPRIRightsInheritance(boolean _b) {
      Boolean olD = (Boolean)m_Structure_.get("LPRI inheritance 1: Linked PRI inheritance");
      boolean old = !_b;
      if (olD != null) old = olD;
          
      if (old != _b) {
            m_Structure_.put("LPRI inheritance 1: Linked PRI inheritance", new Boolean(_b));
            firePropertyChange("ManageLPRIRightInheritance", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the if-all-sources linked privilege rights are managed.
   */
  public boolean getManageIfAllSourceLinkedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: If-All-Sources LPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the if-all-sources linked privilege rights are managed.
   */
  public void setManageIfAllSourceLinkedPrivileges(boolean _b) {
      Boolean olD = (Boolean)m_Structure_.get("PRI right: If-All-Sources LPRI");
      boolean old = !_b;
      if (olD != null) old = olD;

      if (old != _b) {
            m_Structure_.put("PRI right: If-All-Sources LPRI", new Boolean(_b));
            firePropertyChange("ManageIfAllSourceLinkedPrivileges", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the internal for-one-to-one linked privilege rights are managed.
   */
  public boolean getManageInternalForOneToOneLinkedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: internal One-To-One LPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the internal for-one-to-one linked privilege rights are managed.
   */
  public void setManageInternalForOneToOneLinkedPrivileges(boolean _b) {
      Boolean olD = (Boolean)m_Structure_.get("PRI right: internal One-To-One LPRI");
      boolean old = !_b;
      if (olD != null) old = olD;

      if (old != _b) {
            m_Structure_.put("PRI right: internal One-To-One LPRI", new Boolean(_b));
            firePropertyChange("ManageInternalForOneToOneLinkedPrivileges", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the external for-one-to-one linked privilege rights are managed.
   */
  public boolean getManageExternalForOneToOneLinkedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: external One-To-One LPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be true if the privileges are managed but all the other privilege features
   * are not managed, to allow an own component in the privilege of another ACS.
   * Managing inner external privileges always implies that all the own sources of this ACS
   * may have an inner external privilege. To allow an own resource or an own virtual folder
   * of this ACS to get an inner external privilege, the privileges in resources or
   * in virtual folders have to be managed. This allows to limit the
   * inner external privileges to the own sources of the ACS, excluding the own targets,
   * if there are no internal privileges to manage. On the other hand, having
   * internal privileges means that the inner external privileges always accept 
   * as target an own resource or own virtual folder.
   * @since 0.7.0
   * @param _b true if the external one-to-one linked privilege rights are managed
   * as inner privileges or as privileges of another ACS.
   */
  public void setManageExternalForOneToOneLinkedPrivileges(boolean _b) {
      Boolean olD = (Boolean)m_Structure_.get("PRI right: external One-To-One LPRI");
      boolean old = !_b;
      if (olD != null) old = olD;

      if (old != _b) {
            m_Structure_.put("PRI right: external One-To-One LPRI", new Boolean(_b));
            firePropertyChange("ManageExternalForOneToOneLinkedPrivileges", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the classed typed privilege rights are managed.
   */
  public boolean getManageClassedTypedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: Typed PRI for classed"+
                  "with contraints on source and target classes");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the classed typed privilege rights are managed.
   */
  public void setManageClassedTypedPrivileges(boolean _b) {
      Boolean olD = (Boolean)m_Structure_.get("PRI right: Typed PRI for classed"+
                  "with contraints on source and target classes");
      boolean old = !_b;
      if (olD != null) old = olD;

      if (old != _b) {
            m_Structure_.put("PRI right: Typed PRI for classed"+
                  "with contraints on source and target classes", new Boolean(_b));
            firePropertyChange("ManageClassedTypedPrivileges", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the classed specific typed privilege rights are managed,
   * for designated source and target.
   */
  public boolean getManageClassedSpecificTypedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: Typed PRI for classed specific"+
                  "for designated source and target");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the classed specific typed privilege rights are managed,
   * for designated source and target.
   */
  public void setManageClassedSpecificTypedPrivileges(boolean _b) {
      Boolean olD = (Boolean)m_Structure_.get("PRI right: Typed PRI for classed specific"+
                  "for designated source and target");
      boolean old = !_b;
      if (olD != null) old = olD;

      if (old != _b) {
            m_Structure_.put("PRI right: Typed PRI for classed specific"+
                  "for designated source and target", new Boolean(_b));
            firePropertyChange("ManageClassedSpecificTypedPrivileges", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the typed privilege rights for authorization servers are managed.
   */
  public boolean getManageAuthorizationServerTypedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: Typed PRI for authorization server");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the typed privilege rights for authorization servers are managed.
   */
  public void setManageAuthorizationServerTypedPrivileges(boolean _b) {
      Boolean olD = (Boolean)m_Structure_.get("PRI right: Typed PRI for authorization server");
      boolean old = !_b;
      if (olD != null) old = olD;

      if (old != _b) {
            m_Structure_.put("PRI right: Typed PRI for authorization server", new Boolean(_b));
            firePropertyChange("ManageAuthorizationServerTypedPrivileges", old, _b);
      }
  }
  
  /**
   * @since 0.7.0
   * @return true if the virtual folders are managed for the ACS.
   */
  public boolean getManageVirtualFolders() {
      Boolean result = m_Structure_.get("virtual folder: no virtual folders");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the virtual folders are managed.
   */
  public void setManageVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get("virtual folder: no virtual folders");
        
      if (old != !_b) {
            m_Structure_.put("virtual folder: no virtual folders", new Boolean(!_b));
            firePropertyChange("ManageVirtualFolders", !old, _b);   
      }
  }
  
  /**
   * @since 0.7.0
   * @return true if the virtual folders of resources are managed for the ACS. 
   */
  public boolean getManageResourceVirtualFolders() {
      Boolean result = m_Structure_.get("virtual folder: no virtual folders of Resources");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the virtual folders of resources are managed.
   */
  public void setManageResourceVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get("virtual folder: no virtual folders of Resources");
        
      if (old != !_b) {
            m_Structure_.put("virtual folder: no virtual folders of Resources", new Boolean(!_b));
            firePropertyChange("ManageResourceVirtualFolders", !old, _b);   
      }
  }

  /**
   * @since 0.7.0
   * @return true if the virtual folders of eligible parties are managed for the ACS.
   */
  public boolean getManageEPVirtualFolders() {
      Boolean result = m_Structure_.get("virtual folder: no virtual folders of EP");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the virtual folders of eligible parties are managed.
   */
  public void setManageEPVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get("virtual folder: no virtual folders of EP");

      if (old != !_b) {
            m_Structure_.put("virtual folder: no virtual folders of EP", new Boolean(!_b));
            firePropertyChange("ManageEPVirtualFolders", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the virtual folders of simple resources are managed for the ACS.
   */
  public boolean getManageSimpleResourceVirtualFolders() {
      Boolean result = m_Structure_.get(
              "virtual folder: no simple resources in virtual folders of Resources");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the virtual folders of simple resources are managed.
   */
  public void setManageSimpleResourceVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get(
              "virtual folder: no simple resources in virtual folders of Resources");

      if (old != !_b) {
            m_Structure_.put(
                    "virtual folder: no simple resources in virtual folders of Resources",
                    new Boolean(!_b));
            firePropertyChange("ManageSimpleResourceVirtualFolders", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the resource virtual folders manage directories for the ACS.
   */
  public boolean getManageDirectoryInResourceVirtualFolders() {
      Boolean result = m_Structure_.get(
              "virtual folder: no directories in virtual folders of Resources");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if resource virtual folders manage directories are managed.
   */
  public void setManageDirectoryInResourceVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get(
              "virtual folder: no directories in virtual folders of Resources");

      if (old != !_b) {
          m_Structure_.put(
                    "virtual folder: no directories in virtual folders of Resources",
                    new Boolean(!_b));
          firePropertyChange("ManageDirectoryInResourceVirtualFolders", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the virtual folders of resources manage actors for the ACS.
   */
  public boolean getManageDirectoryEPinEPVirtualFolders() {
      Boolean result = m_Structure_.get(
              "virtual folder: no directoryEPs in virtual folders of EP");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the virtual folders of resources manage actors.
   */
  public void setManageDirectoryEPinEPVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get(
              "virtual folder: no directoryEPs in virtual folders of EP");

      if (old != !_b) {
            m_Structure_.put(
                    "virtual folder: no directoryEPs in virtual folders of EP",
                    new Boolean(!_b));
            firePropertyChange("ManageDirectoryEPinEPVirtualFolders", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the EP virtual folders manage actors for the ACS.
   */
  public boolean getManageActorInEPVirtualFolders() {
      Boolean result = m_Structure_.get(
              "virtual folder: no actors in virtual folders of EP");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the virtual folders of eligible parties manage actors.
   */
  public void setManageActorInEPVirtualFolders(boolean _b) {
      boolean old = (boolean)m_Structure_.get(
              "virtual folder: no actors in virtual folders of EP");

      if (old != !_b) {
            m_Structure_.put(
                    "virtual folder: no actors in virtual folders of EP",
                    new Boolean(!_b));
            firePropertyChange("ManageActorInEPVirtualFolders", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the virtual folder trees are managed for the ACS.
   */
  public boolean getManageVirtualFolderTrees() {
      Boolean result = m_Structure_.get(
              "virtual folder: no virtual folder trees");
      if (result != null && !result.booleanValue()) return true;
      else return false;
  }

  /**
   * @since 0.7.0 "virtual folder: a member is no more than one time"+
              "in each virtual folder tree"
   * @param _b true if the virtual folder trees are managed for the ACS.
   */
  public void setManageVirtualFolderTrees(boolean _b) {
      boolean old = (boolean)m_Structure_.get(
              "virtual folder: no virtual folder trees");

      if (old != !_b) {
            m_Structure_.put(
                    "virtual folder: no virtual folder trees",
                    new Boolean(!_b));
            firePropertyChange("ManageVirtualFolderTrees", !old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if a member is no more than one time in each virtual folder tree.
   */
  public boolean getManageOneTimeInVirtualFolderTreeForMember() {
      Boolean result = m_Structure_.get(
              "virtual folder: a member is no more than one time"+
                "in each virtual folder tree");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if a member is no more than one time in each virtual folder tree.
   */
  public void setManageOneTimeInVirtualFolderTreeForMember(boolean _b) {
      boolean old = false;

      if (m_Structure_.get(
              "virtual folder: a member is no more than one time"+
                        "in each virtual folder tree") != null)
          old = (boolean)m_Structure_.get(
              "virtual folder: a member is no more than one time"+
                        "in each virtual folder tree");

      if (old != _b) {
            m_Structure_.put(
                    "virtual folder: a member is no more than one time"+
                        "in each virtual folder tree",
                    new Boolean(_b));
            firePropertyChange("ManageOneTimeInVirtualFolderTreeForMember", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if a virtual folder propagates its target rights to all its members.
   */
  public boolean getManageVirtualFolderRightsPropagationToMembers() {
      Boolean result = m_Structure_.get(
              "virtual folder: propagation of target rights to all the members");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0 
   * @param _b true if a virtual folder propagates its target rights to all its members.
   */
  public void setManageVirtualFolderRightsPropagationToMembers(boolean _b) {
      boolean old = false;

      if (m_Structure_.get(
              "virtual folder: propagation of target rights to all the members") != null)
          old = (boolean)m_Structure_.get(
              "virtual folder: propagation of target rights to all the members");

      if (old != _b) {
            m_Structure_.put(
                    "virtual folder: propagation of target rights to all the members",
                    new Boolean(_b));
            firePropertyChange("ManageOneTimeInVirtualFolderTreeForMember", old, _b);
      }
  }

  /**
   * Only an AcsAddon can manage the soft aliasing.
   * @since 0.7.0
   * @return true if the acs manages the external or internal soft aliases
   */
  public boolean getManageSoftAlias() {
      Boolean result = m_Structure_.get(
              "Aliasing: acs object aliasing with external or internal soft alias");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * Only an AcsAddon can manage the soft aliasing.
   * @since 0.7.0 
   * @param _b true if the acs manages the external soft aliases
   */
  public void setManageSoftAlias(boolean _b) {
      boolean old = false;

      if (m_Structure_.get(
              "Aliasing: acs object aliasing with external or internal soft alias") != null)
          old = (boolean)m_Structure_.get(
              "Aliasing: acs object aliasing with external or internal soft alias");

      if (old != _b) {
            m_Structure_.put(
                    "Aliasing: acs object aliasing with external or internal soft alias",
                    new Boolean(_b));
            firePropertyChange("ManageSoftAlias", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the acs manages the internal hard aliases
   */
  public boolean getManageInternalHardAlias() {
      Boolean result = m_Structure_.get(
              "Aliasing: acs object aliasing with internal hard alias");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the acs manages the internal hard aliases
   */
  public void setManageInternalHardAlias(boolean _b) {
      boolean old = false;

      if (m_Structure_.get(
              "Aliasing: acs object aliasing with internal hard alias") != null)
          old = (boolean)m_Structure_.get(
              "Aliasing: acs object aliasing with internal hard alias");

      if (old != _b) {
            m_Structure_.put(
                    "Aliasing: acs object aliasing with internal hard alias",
                    new Boolean(_b));
            firePropertyChange("ManageInternalHardAlias", old, _b);
      }
  }

  /**
   * @since 0.7.0
   * @return true if the acs is tied to its parent acs.
   */
  public boolean getTiedToParentACS() {
      Boolean result = m_Structure_.get("embedded: this ACS is running on "+
              "and is tied to the parent ACS");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * Sets if the new ACS is tied to its ACS parent or not.
   * @since 0.7.0
   * @param _b true if the acs is tied to its parent acs.
   * @exception UpDateError if:<P>
   *    - _b true when the parent is not an ACS, in the IS.
   */
  public void setTiedToParentACS(boolean _b) throws UpDateError {
      try {
          if (_b &&
            !BaseManagerImpl.getInstance().getISTree(iSname_).
                getType(parentName_).equals(ISFactoryUtilityImpl.ACS_NODE) )
                    throw new UpDateError("The parent is not an ACS."+
                            " This new ACS cannot be tied to its parent.");
      }
      catch (BaseError b) {
          b.printStackTrace();
          throw new InternalError(" This new ACS cannot be tied to its parent: "+b);
      }

      boolean old = false;

      if (m_Structure_.get(
              "embedded: this ACS is running on and is tied to the parent ACS") != null)
          old = (boolean)m_Structure_.get("embedded: this ACS is running on "+
              "and is tied to the parent ACS");
         
      if (old != _b) {
            m_Structure_.put("embedded: this ACS is running on and is tied to the parent ACS",
                    new Boolean(_b));
            firePropertyChange("TiedToParentACS", old, _b);
      }
  } // setTiedToParentACS()
  
  /**
   * @return true if the acs is an authorization store as a LDAP directory for instance.
   */
  public boolean getIsAuthorizationServer() {
      Boolean result = m_Structure_.get("acs: this ACS is an authorization store");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @param _b true if the acs is an authorization server as a LDAP directory.
   */
  public void setIsAuthorizationServer(boolean _b) {
      boolean old = (boolean)m_Structure_.get("acs: this ACS is an authorization store");
          
      if (old != _b) {
            m_Structure_.put("acs: this ACS is an authorization store", new Boolean(_b));
            firePropertyChange("IsAuthorizationServer", old, _b);
      }
  } 
   
  /**
   * @since 0.7.0
   * @return true if the acs manages ACS rights (it may manage metarights too).
   */
  public boolean getManageAcsRights() {
      Boolean result = m_Structure_.get("acs: acsrights");
      if (result != null) return result.booleanValue();
      else return false;
  } 
 
  /**
   * @param _b true if the acs manages ACS rights (it may manage metarights too).
   */
  public void setManageAcsRights(boolean _b) {
      boolean old = (boolean)m_Structure_.get("acs: acsrights");
          
      if (old != _b) {
            m_Structure_.put("acs: acsrights", new Boolean(_b));
            firePropertyChange("ManageAcsRights", old, _b);
      }
  }
  
  /**
   * @return true if the acs manages metarights (it may manage ACS rights too).
   */
  public boolean getManageMetaRights() {
      Boolean result = m_Structure_.get("acs: metarights");
      if (result != null) return result.booleanValue();
      else return false;
  } 
 
  /**
   * @param _b true if the acs manages metarights (it may manage ACS rights too).
   */
  public void setManageMetaRights(boolean _b) {
      boolean old = (boolean)m_Structure_.get("acs: metarights");
          
      if (old != _b) {
            m_Structure_.put("acs: metarights", new Boolean(_b));
            firePropertyChange("ManageMetaRights", old, _b);
      }
  }
  
  /**
   * @return true if some specialized rights have been defined.
   */
  //public boolean getHavePrivilegeSpecializedRights() {
    //      return havePrivilegeSpecializedRights_;
  //}
   
  /**
   * Gets the name of the information system which owns this acs.
   * @since 0.7.0
   * @return a copy of the IS name. May be empty, but not null.
   */
  public final ImmutableName getISName() {
      if (iSname_ == null) return null;
      else return (ImmutableName)iSname_.clone();
  }

  /**
   * Sets the name of the information system which owns this acs.
   * @since 0.7.0
   * @param _name the IS name. May be empty or null.
   */
  public final void setISName(ImmutableName _name) {
    ImmutableName old = iSname_;
    if (_name == null) iSname_ = null;
    else iSname_ = (ImmutableName)_name.clone();

    if ((old == null && _name != null) ||
              (old != null && _name == null) ||
              (old != null && !old.equals(_name))) {
        firePropertyChange("ISname", old, _name);
    }
    old = null;
  }
  
  /**
   * Gets the optional parent name for the ACS in the IS nametree.
   * @return the parent name. May be null or empty.
   */
  public ImmutableName getParentName()	{
    if (parentName_ == null) return null;
    else return (ImmutableName)parentName_.clone();
  } // getParentName()

  /**
   * Sets the parent name for the ACS in the IS nametree.
   * @param _name the parent name. May be null or empty.
   */
  public void setParentName(ImmutableName _name)	{
    ImmutableName old = parentName_;
    if (_name == null) parentName_ = null;
    else parentName_ = (Name)_name.clone();
    
    if ((old == null && _name != null) ||
              (old != null && _name == null) ||
              (old != null && !old.equals(_name))) {
        firePropertyChange("ParentName", old, _name);
    }
    old = null;
  } // setParentName()
  
  /**
   * Gets the hub property for the ACS node in the IS nametree.
   * @since 0.7.0
   * @return the is_hub property
   */
  public boolean getHubProperty()	{
    return is_hub_;
  } 
  
  /**
   * Gets the visible space for the ACS in the IS nametree.
   * @since 0.7.0
   * @return the visible space property
   */
  public String getVisibleSpace() {
    return visibleSpace_;
  } 

  
  /**
   * @return the authorized values for the ACS node links in the IS nametree.
   */
  //public final List getL_authorizedLinks()	{
    //  return (List)l_authorizedLinks__.clone();
  //}  // getL_authorizedLinks()

  /**
   * Gets all the acsrights which are allowed in this ACS.
   * Some of these acsrights may not be used in the final ACS.
   * @return all the rights used in this acs.
   */
  public final StringRight[] getL_Rights() {
    if (l_Rights_ == null) return null;
    StringRight[] result = new StringRight[l_Rights_.length];
    System.arraycopy(l_Rights_, 0, result, 0, l_Rights_.length);
    return result;
  }
  
  /**
   * Gets the first acsright which has 'full_control' as metaright, or this metaright
   * if no acsright complies this condition.
   * 'full_control' metaright has 'deny_all' metaright for opposite,
   * and its single user upper right is 'modify_security_descriptor'.
   * @return the right. Never null. 
   */
  public StringRight getFullControlRight() {
    if (l_Rights_ == null) return null;
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        
                // for testing and tracking
                //System.out.println("---------------------- "+l_Rights_[i].getName());
        
        if (!(l_Rights_[i] instanceof StringMetaRight) &&
                l_Rights_[i].getEorMetaRight().getRight().equals("full_control")) {
            result = l_Rights_[i];
            break;
        }
    } // for
    if (result == null) 
        result = BaseManagerImpl.getInstance().getMetaRight("full_control");
    return result;      
  } // getFullControlRight()
  
  /**
   * Gets a right which is registered by this ACS. It may be a StringMetaRight.
   * @param _name is the name of the right, provided by StringRight.getName().
   * @return the right; null if it does not exist.
   */
  public final StringRight getOneRight(ImmutableName _name) {
    if (l_Rights_ == null || _name == null) return null;
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        if (l_Rights_[i].getName().equals(_name)) {
            result = l_Rights_[i];
            break;
        }
    } // for
    

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.getOneRight(), "+
                      //      "result is "+result+" for the name "+_name);     
    
    return result;
  } //getOneRight()
  
  /**
   * Gets a right which is registered by this ACS. It may be a StringMetaRight.
   * @param _right is the right, provided by StringRight.getRight().
   * @return the right; null if it does not exist.
   */
  public final StringRight getOneRight(String _right) {
    if (l_Rights_ == null || _right == null) return null;
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        if (l_Rights_[i].getRight().equals(_right)) {
            result = l_Rights_[i];
            break;
        }
    } // for
    return result;
  } //getOneRight() 
  
  /**
   * Gets a right which is registered by this ACS. It may be a StringMetaRight.
   * @param _metaRight is the right of the metaright of the searched right, 
   *             provided by StringRight.getRight().
   * @return the right; null if it does not exist.
   */
  public final StringRight getFirstRightFromMeta(String _metaRight) {
                
                // for testing and tracking
                //System.out.println("Call to ACSImpl.getFirstRightFromMeta() for the string "+_metaRight);
                
    if (l_Rights_ == null || _metaRight == null) return null;
    
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        if (l_Rights_[i].getEorMetaRight().getRight().equals(_metaRight)) {
            result = l_Rights_[i];
            break;
        }
    } // for
    return result;
  } //getFirstRightFromMeta() 
  
  /**
   * Registers a new ACS right through the ACSpreFactoryBeanInfo or from gWork.
   * Updates the structure to indicate the type of rights: ACS right. <P>
   * Note: The NoThanViewInBases use StringRight.compareTo() between the rights. The
   * right internal key is the first compared property if the rights have the same sens.
   * Thus, a NoThanViewInBase may detect its criterion right is not fulfilled while
   * the effective right has no direct path to the criterion right, in the network of rights.
   * This behavior of NoThanViewInBase shows the need to
   * set a consistent network of rights, including well-chosen internal keys. <P>
   * The user don't register the metarights through this method.
   * Metarights are added by initializeConstraints() which calls this method
   * before the ACS creation.
   * Not used through ACSFactoryBeanInfo. 'OneAcsRight' is a bound property.
   * Calls newRight().
   * @param _right is the new acsright to register. Not null.
   * @exception CreateError if the right is null, a metaright or if
   *   the right is already used.
   */
   public void addRight(StringRight _right ) throws CreateError {
      if (_right == null || _right.getEorMetaRight() == null)
          throw new CreateError("in ACSFactoryImpl.addRight(),"+
            " the right is null or has no metaright.");
      
      // exceptions
      if (_right instanceof StringMetaRight) {
         throw new CreateError("in ACSFactoryImpl.addRight(),"+
            " the right is a metaright.");
      }

      // TO REMOVE for StringRight testing only
      if (getOneRight(_right.getRight()) != null) {
          
                System.out.println("From ACSFactoryImpl.addRight(), in exception on right already known: ");
                for (int i = 0; i < l_Rights_.length; i++) {
                                        // for testing and tracking
                    System.out.println("----------- "+l_Rights_[i]);
                }
          
         throw new CreateError(
            "In ACSFactoryImpl.addRight(), the new right "+_right+" is already known.");
      }
 
      newRight(_right );

            // for testing and tracking
            //System.out.println("From ACSFactoryImpl.addRight(), firePropertChange() on 'OneAcsRight'");
      
      try {
            firePropertyChange("OneAcsRight", null, _right);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeRight(), firePropertChange() throws:");
            e.printStackTrace();
      }
   } // addRight()

  /**
   * Unregisters a right through the ACSpreFactoryBeanInfo.
   * Updates the structure to indicate the type of rights: acsright.
   * Used by the GUI and StringRight.finalizeForUser() method. 
   * Not used through ACSFactoryBeanInfo. Call moveRight().
   * 'OneAcsRight' is a bound property.
   * @param _right is the right to unregister.
   * @exception UpDateError if the folder is null or not known.
   */
   public void removeRight(StringRight _right ) throws UpDateError {
        // exception
        if (_right == null || getOneRight(_right.getRight()) == null)
          throw new UpDateError("in ACSFactoryImpl.removeRight(),"+
            " the right is null or unknown."); 
        
        moveRight(_right );
        
        try {
            firePropertyChange("OneAcsRight", _right, null);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeRight(), firePropertChange() throws:");
            e.printStackTrace();
        }
   } // removeRight()    
  
  /**
   * Sets the allowed granting right values for the AGO Group rights.
   * "full_control" is the default right, but it is not mandatory.
   * No action if the 'group' rights are not managed,
   * or if the instance has been finalized. A right may be not registered in getL_Rights().
   * Fires a PropertyChange for 'GORightsSet'.
   * @param _s is an array of rights.
   *   Not null, no null elements, but may be a 0-length array nor negative rights.
   * @exception UpDateError if _s is  null, if the GO rights are not allowed, or if
   *    there is no fixed set of rights in this acs
   */
  public synchronized void setL_GroupRightsSet(StringRight[] _s) throws UpDateError {
    if (!m_Structure_.get("AGO right: Account and Group rights") && _s.length != 0)
        throw new UpDateError(" The AG rights are not allowed.");
    if (l_GORightsSet_ == null)
        throw new UpDateError(" The old G/O rights set is null.");
    
    controlRights(_s, false); // throws UpDateError
    
    StringRight[] l_oldRights = new StringRight[l_GORightsSet_.length];
    System.arraycopy(l_GORightsSet_, 0, l_oldRights, 0, l_GORightsSet_.length);
    
    l_GORightsSet_ = new StringRight[_s.length];
    System.arraycopy(_s, 0, l_GORightsSet_, 0, _s.length);

    try {
        firePropertyChange("GORightsSet", l_oldRights, _s);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_GroupRightsSet(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldRights = null;;
  } // setL_GroupRightsSet()

  /**
   * Gets the allowed AGO right values for the groups.
   * @return array of rights. May be empty, but not null.
   */
  public final StringRight[] getL_GroupRightsSet() {
    StringRight[] result = new StringRight[l_GORightsSet_.length];
    System.arraycopy(l_GORightsSet_, 0, result, 0, l_GORightsSet_.length);
    return result;
  }  
  
  /**
   * Sets the allowed granting right values for the Account (an UserID).
   * "full_control" is the default right, but it is not mandatory.
   * No action if account rights or limited set of values are not managed, or if
   * the instance has been finalized.
   * A right may be not registered in getL_Rights().
   * Fires a PropertyChange for 'ARightsSet'.
   * @param _s is an array of rights for the UserID which owns this resource,
   *   in the main ACS.
   *   Not null, no null or negative right, but may be a 0-length array.
   * @exception UpDateError if _s is  null, if the AGO rights are not allowed, or if
   *    there is no fixed set of rights in this acs
   */
  public synchronized void setL_AccountRightsSet(StringRight[] _s) throws UpDateError {
    if (!m_Structure_.get("EP: accounts") && _s.length != 0)
        throw new UpDateError(" The accounts are not allowed.");
    if (!m_Structure_.get("AGO right: Account and Group rights") && _s.length != 0)
        throw new UpDateError(" The account rights are not allowed.");
    if (l_AccountRightsSet_ == null )
          throw new UpDateError(" The old Account rights set is null.");
    
    controlRights(_s, false); // throws UpDateError
    
    StringRight[] l_oldRights = new StringRight[l_AccountRightsSet_.length];
    System.arraycopy(l_AccountRightsSet_, 0, l_oldRights, 0, l_AccountRightsSet_.length);
    
    l_AccountRightsSet_ = new StringRight[_s.length];
    System.arraycopy(_s, 0, l_AccountRightsSet_, 0, _s.length);

    try {
        firePropertyChange("ARightsSet", l_oldRights, _s);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.setL_AGORightsSet(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldRights = null;
  } // setL_AccountRightsSet()

  /**
   * Gets the allowed granting right values for the Account (an UserID).
   * @return array of rights. May be empty, but not null.
   */
  public final StringRight[] getL_AccountRightsSet() {
    StringRight[] result = new StringRight[l_AccountRightsSet_.length];
    System.arraycopy(l_AccountRightsSet_, 0, result, 0, l_AccountRightsSet_.length);
    return result;
  }  

  /**
   * Gets the allowed interfaces for the objects which are handled by an acs.
   * For instance, 'Directory' if the acs has the ACS manages the resources tree.
   * @return array of some interfaces in the Access Road base.
   *   The immutable types are not included.
   *   Never empty nor null.
   */
  public static final String[] getL_AllowedACSInterfacesForSpecializedRights() {
    return l_AllowedACSInterfacesForSpecializedRights_;
  }   
  
  /**
   * Adds some specialized rights to extend or to constrain
   * the allowed values which are defined for each right type, as in setL_xxxRightsSet(). 
   * Specialized rights are always possible, whatever the ACS structure.
   * If internal and external AclRights are not allowed and the criteria contains them,
   * these rights are allowed for internal ACL only, not external ones. In all cases,
   * the 'Acl' specialized rights are not added to the AclRights set. The other right
   * categories are managed following the same way.
   * If it is about types, this method have to be called after the creation of
   * the allowed types. If an immutable type is used as key for AclEntry rights,
   * the single effect is to avoid the user to select rights which will be rejected
   * by the AclEntry class, but then, the user has no information about the cause
   * (the Resource or the EligibleParty is immutable). This is why immutable types
   * as not recommanded as keys for AclEntry rights. <P>
   * The criteria is a quartet of Strings. 
   * Examples of quartets (object.domain.key.subkey) are: <P>
   *     - 'Target.AclRightsSet.Interface.Directory', for ACL rights when the target is a Directory instance <P>
   *     - 'Target.AccRightsSet.Type.script', for Account rights when the target has the type 'script' <P>
   *     - 'Source.AclRightsSet.Type.group', for ACL rights when the source has the type 'group' <P>
   * The object is Target or Source to indicate that these rights are for the target
   * (or source) which has the correct interface or type. <P>
   * The domain may be one of the right categories:
   * 'AccRightsSet' for Target only, 'GpORightsSet' for Target only, 
   * 'AclRightsSet', 'BdgRightsSet', 'RooRightsSet' for Source and Target. <P>
   * The subdomain is 'Interface' or 'Type' to indicate that the subkey belongs to one
   * of these categories. <P>
   * The subkey is an interface name or an EligibleParty type (or Resource type if Target),
   * for which the specialized rights have to be used. <P>
   * Caution: addPrivilegeRights() define the specialized rights for privileges.
   * These privilege specialized rights never change this current property.
   * A right may be not registered in getL_Rights().
   * Fires PropertyChanges for 'SpecializedRightsSet'.
   * @param _criteria is the map key to apply these specialized rights.
   * @param _l_rights is a StringRights array, from the correct ACS, which contain 
   *   the specialized rights for the criteria. They are added to the previous rights, if any.
   *   Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - the criteria is null, unknown or does not end with '&lt;undefined&gt;', <P>
   *    - a StringRight is not from the ACS.
   */
  public synchronized void addSpecializedRights(String _criteria, 
    StringRight[] _l_rights) throws UpDateError {

                    // for testing and tracking
                    //System.out.println("Call to ACSFactoryImpl.addSpecializedRights() for "+
                      //      _criteria);
      
    if (_criteria == null || _l_rights == null ) {
        throw new UpDateError(" The criteria or the rights array is not well formed.");
    }
    
    List<String>[] list = ACSFactoryUtilityImpl.getL_MapKeysStartsForSpecializedRights();
    boolean error = true;
    for(Iterator ite = list[0].iterator(); ite.hasNext(); ) {
        if (_criteria.startsWith((String)ite.next()) ) {
            error = false;
            break;
        }
    } // for ite
    if (error && !_criteria.endsWith("<undefined>"))
        throw new UpDateError(_criteria+" is not the start of an allowed key.");
    else if (error)
        return;
    error = true;
    for(Iterator ite = list[1].iterator(); ite.hasNext(); ) {
        if (_criteria.startsWith((String)ite.next(), 7) ) {
            error = false;
            break;
        }
    } // for ite
    if (error) throw new UpDateError(" In "+_criteria+", the domain is not well formed.");
    error = true;
    for(Iterator ite = list[2].iterator(); ite.hasNext(); ) {
        if (_criteria.startsWith((String)ite.next(), 20) ) {
            error = false;
            break;
        }
    } // for ite
    if (error) throw new UpDateError(" In "+_criteria+", the subdomain is not Interface or Type.");    
 
    // other controls on criteria
    if (_criteria.startsWith("Source") && 
            (_criteria.startsWith("AccRightsSet", 7) || _criteria.startsWith("GpORightsSet", 7) )) {
        throw new UpDateError(" These rights are not allowed to a Source, only to a Target.");
    }
    
    // controls the interface
    if (_criteria.startsWith("Interface", 20) && getL_AllowedACSInterfacesForSpecializedRights() != null) {
        if (!Arrays.asList( getL_AllowedACSInterfacesForSpecializedRights() ).contains( _criteria.substring(30) ) ) {
            

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addSpecializedRights(), "+
                      //      "exception on the key for interface with the entered key: "+
                      //      _criteria.substring(30));            
            
            throw new UpDateError(" In "+_criteria+
                    ", the subkey is an interface which is not allowed.");
        }
    } // if
    
    controlRights(_l_rights, true); //throws UpDateError    
    
    // no negative root/A/bridge rights (but possible for GO and ACL)
    if (_criteria.startsWith("AccRightsSet") ||
            _criteria.startsWith("BdgRightsSet") || _criteria.startsWith("RooRightsSet")) {
        for (int i = 0; i < _l_rights.length; i++) {
          if (!_l_rights[i].isGrantingRight() )  {
              error = true;
              break;
          } 
        } // for i
    } // if
    if (error) throw new UpDateError(" The negative rights are forbidden for Root, Account,"+
            " and Bridge rights");     

    // controls the management of rights
    if (_criteria.startsWith("AclRightsSet", 7) && !getManageInternalAcls() ) {
            throw new UpDateError(" The internal ACL are not managed");
    } else if (_criteria.startsWith("AccRightsSet", 7) && 
         (!getManageAccounts() || !getManageAGORights() ) ) {
            throw new UpDateError(" The accounts are not managed");
    } else if (_criteria.startsWith("GpORightsSet", 7) && 
         (!getManageGroups() ||  !getManageAGORights() )) {
            throw new UpDateError(" The group rights are not managed");
    } else if (_criteria.startsWith("BdgRightsSet", 7) && !getManageBridges() ) {
            throw new UpDateError(" The bridges are not managed");
    } // else

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addSpecializedRights(), "+
                      //      "the rights are controlled OK ");
                    
    // adds _criteria as a Resource, EP type if necessary.
    // not for a Privilege second type, which is set in addPrivilegeRights().
    // not for the VirtualFolders.
    if (_criteria.startsWith("Type", 20)) {
        if (_criteria.startsWith("Target") && getManageResources() &&
                !Arrays.asList( l_resourceTypes_ ).contains( _criteria.substring(25) )) {
                     String[] l_array = new String[l_resourceTypes_.length + 1];
                     System.arraycopy(l_resourceTypes_, 0, l_array, 0, l_resourceTypes_.length);
                     l_array[l_array.length - 1] = _criteria.substring(25);
                     setL_ResourceTypes(l_array);
                     l_array = null;
        } // if (_criteria.startsWith("Target") 
        if (_criteria.startsWith("Source") &&
                (getManageAccounts() || getManageGroups() || getManageActors())
                &&
                !Arrays.asList( l_epTypes_ ).contains( _criteria.substring(25) ) ) {
                     String[] l_array = new String[l_epTypes_.length + 1];
                     System.arraycopy(l_epTypes_, 0, l_array, 0, l_epTypes_.length);
                     l_array[l_array.length - 1] = _criteria.substring(25);
                     setL_EPTypes(l_array);
                     l_array = null;
        } // if (_criteria.startsWith("Source")
        
        // no adding into the VirtualFolder types as source, since they are not 
        // AG, ACL, PRI Source, and not as targets that is handled in addPrivilegeRights().
        //if (_criteria.startsWith("Source") && getManageResourceVirtualFolders() &&
          //      !Arrays.asList( l_resvfTypes_ ).contains( _criteria.substring(25) ) ) {
            //         String[] l_array = new String[l_resvfTypes_.length + 1];
              //       System.arraycopy(l_resvfTypes_, 0, l_array, 0, l_resvfTypes_.length);
                //     l_array[l_array.length - 1] = _criteria.substring(25);
                  //   setL_ResourceVirtualFolderTypes(l_array);
                    // l_array = null;
        //} // if (_criteria.startsWith("Source")
        //if (_criteria.startsWith("Source") && getManageEPVirtualFolders() &&
          //      !Arrays.asList( l_epvfTypes_ ).contains( _criteria.substring(25) ) ) {
            //         String[] l_array = new String[l_epvfTypes_.length + 1];
              //       System.arraycopy(l_epvfTypes_, 0, l_array, 0, l_epvfTypes_.length);
                //     l_array[l_array.length - 1] = _criteria.substring(25);
                  //   setL_EPVirtualFolderTypes(l_array);
                    // l_array = null;
        //} // if (_criteria.startsWith("Source")
    } // if
    
    HashMap m_oldRights =  null;
    if (m_SpecializedRights_ != null) 
        m_oldRights = (HashMap) m_SpecializedRights_.clone();
    else {
        m_SpecializedRights_ = new HashMap(INITIAL_CAPACITY_25);
    }

    if (m_SpecializedRights_.get(_criteria) != null) {
        StringRight[] l_old = m_SpecializedRights_.get(_criteria);
        if (l_old.length != 0) {
            Set<StringRight> l_new = new HashSet(INITIAL_CAPACITY_25);
            for (int i = 0; i < l_old.length; i++) {
                l_new.add(l_old[i]);
            } // for i
            l_new.addAll(Arrays.asList(_l_rights) );
            m_SpecializedRights_.put(_criteria, (StringRight[])l_new.toArray(new StringRight[0]));
            l_new = null;
        } 
        else {
            m_SpecializedRights_.put(_criteria, _l_rights);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addSpecializedRights(), "+
                      //      "for the key "+_criteria+", add the rights:");
                    //for (int i = 0; i < _l_rights.length; i++) {
                      //  System.out.println("------- for i = "+i+"; right = "+_l_rights[i]);
                    //} // for i
        
        } // else
        l_old = null;       
    }
    else m_SpecializedRights_.put(_criteria, _l_rights); 

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addSpecializedRights(), "+
                      //      "fire the property change 'SpecializedRightsSet' ");
    
    try {
        firePropertyChange("SpecializedRightsSet", m_oldRights, 
                (HashMap) m_SpecializedRights_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addSpecializedRights(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    m_oldRights = null;
    list = null;
  } // addSpecializedRights()

  /**
   * Removes some specialized rights.
   * The criteria is a quartet of Strings. 
   * Examples of quartets (object.domain.key.subkey) are: <P>
   *     - 'Target.AclRightsSet.Interface.Directory', for ACL rights when the target is a Directory instance <P>
   *     - 'Target.AccRightsSet.Type.script', for Account rights when the target has the type 'script' <P>
   *     - 'Source.AclRightsSet.Type.group', for ACL rights when the source has the type 'group' <P>
   * Fires a PropertyChange for 'SpecializedRightsSet'.
   * The criteria is not removed from the map if the rights argument contains
   * all the current rights which is associated to the criteria.
   * A removed type is not removed from the relevant types list, even if
   * it has been added in the list by addSpecializedRights().
   * @param _criteria for the relevant specialized rights.
   * @param _l_rights StringRights array which contain only
   *   the specialized rights to remove for the criteria. Null to remove the criteria. 
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - _l_rights is null or one right is null <P>
   *    - a right in _l_rights is not registered.
   */
  public synchronized void removeSpecializedRights(String _criteria, 
    StringRight[] _l_rights) throws UpDateError {
    
    if (_criteria == null ) {
        throw new UpDateError(" The criteria is not well formed.");
    }
    if (m_SpecializedRights_ == null) {
        throw new UpDateError(" The specialized rights are not well initialized.");
    }    
    if (!m_SpecializedRights_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeSpecializedRights(), "+
                      //      "rights to remove for "+_criteria);
    
    HashMap m_oldRights = (HashMap) m_SpecializedRights_.clone();   
    if ( _l_rights == null) {
        m_SpecializedRights_.remove(_criteria);
        if (m_SpecializedRights_.isEmpty()) havePrivilegeSpecializedRights_ = false;
        try {
            firePropertyChange("SpecializedRightsSet", m_oldRights,
                    (HashMap) m_SpecializedRights_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeSpecializedRightst(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch 
        m_oldRights = null;
        return;
    } // if ( _l_rights == null) 
    
    boolean error = false;
    
    for (int i = 0; i < _l_rights.length; i++) {
      if (_l_rights[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_rights[i]);          
      }
    } // for i
    if (error) {
        throw new UpDateError(" A specialized right in the argument is null.");
    }     
    
    StringRight[] l_current = m_SpecializedRights_.get(_criteria); 
    List list = new ArrayList( 20 );
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);
    }
    
    short count = 0;
    boolean found = false;

    for (int  i = 0 ; i < _l_rights.length ; i++) {
        found = list.remove(_l_rights[i]);
        
        if (!found) {
            error = true;
            break;
        } // if
        else count++;
            
    } // for i

    if (error) {
        m_oldRights = null;
        l_current = null;
        list = null;
        throw new UpDateError(" A specialized right to remove is not found.");
    }     
    
    if (l_current.length != count) {
        m_SpecializedRights_.put(
                _criteria, 
                (StringRight[]) list.toArray(new StringRight[list.size()]));
    } else { // if _l_rights contains all the current rights.
        m_SpecializedRights_.put(
                _criteria, new StringRight[0]);
    }

    try {
        firePropertyChange("SpecializedRightsSet", m_oldRights,
                (HashMap) m_SpecializedRights_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeSpecializedRights(), firePropertChange() throws:");
        e.printStackTrace();
    } // catch
    
    m_oldRights = null;
    l_current = null;
    list = null;
  } // removeSpecializedRights()  
  
  /**
   * Gets the specialized rights extending or constraining
   * the allowed values which are defined for each right type, as in setL_xxxRightsSet(). 
   * Specialized rights are always possible, whatever the ACS structure.
   * As map keys, examples of quartets (object.domain.subdomain.subkey) are: <P>
   *     - 'Target.allRights.Interface.Directory', for all rights when the target is a Directory instance <P>
   *     - 'Target.AccRights.Type.script', for Account rights when the target has the type 'script' <P>
   *     - 'Source.AclRights.Type.group', for ACL rights when the source has the type 'group' <P>
   * The object is Target or Source to indicate that these rights are for the target
   * (or source) which has the correct interface or type. <P>
   * The domain may be one of the right categories:
   * 'AccRightsSet' for Target only, 'GpORightsSet' for Target only, 
   * 'AclRightsSet' for Target only, 'PrvRightsSet', 'BdgRightsSet', 'RooRightsSet' for root. <P>
   * The subdomain is 'Interface' or 'Type' to demand that the subkey belongs to one
   * of these categories. <P>
   * The subkey is an interface name or an EligibleParty type (or Resource type if Target),
   * for which the specialized rights have to be used. <P>
   * Caution: getM_PrivilegeRights() may also return some ACS specialized rights, which
   * will be set in the new ACS. These privilege specialized rights are independant
   * to this current property.
   * A right may be not registered in getL_Rights().
   * @return a copied Map where the key is a quartet of Strings, and the value is an array
   * of StringRights. May be empty, but not null.
   */
  public final Map getM_SpecializedRights() {
      if (m_SpecializedRights_ == null) return new HashMap(INITIAL_CAPACITY_5);

      HashMap<String,StringRight[]>  m_result = 
         new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25); 
      StringRight[] l_rights = null;
      Map.Entry<String,StringRight[]> entry = null;
      StringRight[] result = null;

              // for testing and tracking
              //System.out.println("in ACSFactoryImpl.getM_SpecializedRights(): ");
      
      for (Iterator ite = m_SpecializedRights_.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,StringRight[]>)ite.next();

         l_rights = entry.getValue();
         result = new StringRight[l_rights.length];
         System.arraycopy(l_rights, 0, result, 0, l_rights.length);  
         m_result.put(entry.getKey(), result);
         
              // for testing and tracking
              //System.out.println("--- for the key "+entry.getKey()); 
              //for (int i = 0; i < result.length; i++) {
                //    System.out.println("--------- at "+i+": "+result[i]);                  
              //} // for i
         
      } // for ite

      
      l_rights = null;
      entry = null;
      result = null;     
      return m_result;
  } // getM_SpecializedRights()
  
  /**
   * Adds the values to manage the display in the explorer, or
   * to set the type-oriented policy following the ACSObject type.
   * The criteria is a double (key.subkey) or a triplet (key.subkey.type). <P>
   * This includes all the keys for the ACSImpl property m_TypesAndGUIPolicy_,
   * excepted for the five keys 'ResourceType', 'EPType',
   * 'VirtualFolderType', 'PrivilegeForTypeType', 'PrivilegeForLinksType'.
   * This also includes all the items to put in the ACS property m_TypeOrientedPolicy_. <P>
   * The key word 'immutable' is forbidden in the type of a triplet, since
   * all the policy is applied without this key word.
   * See the addSubAndGroupPolicy method about the specific handling of the keys
   * 'Explorer.SubAcsNode' and 'Explorer.AcsGroupTree'.
   * See the initializeConstraintsBeforeCreation method about the handling of the
   * key 'CreationByBeamer.NoType' to manage all the current 'byAcsCreator' types. <P>
   * Fires property changes for 'TypesAndExplorerSet'.
   * @see ACSFactoryUtilityImpl#getL_LimitedValuesForTypesAndExplorer
   * @param _criteria the map key to apply these specialized values.
   * @param _l_values String array which contain 
   *   the specialized values for the criteria. May be already known.
   *   They are added to the previous values, if any.
   *   Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not allowed, <P>
   *    - a value is not allowed. 
   */
  public synchronized void addTypesAndExplorer(String _criteria, String[] _l_values)
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("Call of ACSFactoryImpl.addTypesAndExplorer() ");;
      
    if (_criteria == null || _l_values == null ) {
        throw new UpDateError(" The key or the values is null.");
    }
    if (_criteria.indexOf("<immutable>") != -1) {
        throw new UpDateError(" The key cannot be for an immutable type.");
    }
    
    Set<String> list = getL_MapKeysForTypePolicies();
    if (!list.contains(_criteria))
        throw new UpDateError(" "+_criteria+" is not allowed as key.");
    
    boolean error = false;
    int index = -1;
    for (int i = 0; i < _l_values.length; i++) {
        if (_criteria.equals("Explorer.NoAcsNode") && !_l_values[i].startsWith("ACSTree.")) {
            error = true;
            index = i;
            break;
        } 
        else if (_criteria.equals("Explorer.SubAcsNode")
                && !_l_values[i].startsWith("ResourceType.")
                && !_l_values[i].startsWith("GroupType.") ) {
            error = true;
            index = i;
            break;
        } 
        else if (_criteria.equals("Explorer.AcsGroupTree") 
                && !_l_values[i].startsWith("GroupType.") ) {
            error = true;
            index = i;
            break;
        }
    } // for i
    if(error)
        throw new UpDateError(
            " The key "+_criteria+" does not comply with the value "+_l_values[index]);
    
    if (_criteria.startsWith("GroupIDMember.RecommandedGroupsAsMember.") &&
            _l_values.length > 4)
        throw new UpDateError(" "+_criteria+" cannot have more than 4 associated values.");

    if (!getManageResourceTrees() && 
      _criteria.equals("Directory.TypesOfChildFor."+"NULL")) {
        throw new UpDateError(" "+_criteria+" cannot be used when the resource trees "+
                "are not managed.");
    } // if (!getManageResourceTrees() &&
    if (!getManageVirtualFolderTrees() &&
      _criteria.equals("VirtualFolder.TypesOfChildFor."+"NULL")) {
        throw new UpDateError(" "+_criteria+" cannot be used when the virtual folder trees "+
                "are not managed.");
    } // if (!getManageVirtualFolderTrees() &&

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addTypesAndExplorer(), "+
                      //      "the rights are controlled OK "); 
    
    HashMap m_oldRights =  null;
    if (m_TypesAndExplorer_ != null)
        m_oldRights = (HashMap) m_TypesAndExplorer_.clone();
    else {
        m_TypesAndExplorer_ = new HashMap(INITIAL_CAPACITY_25);
    }

    if (m_TypesAndExplorer_.get(_criteria) != null) {
        String[] l_old = m_TypesAndExplorer_.get(_criteria);
        if (l_old.length != 0) {
            Set<String> l_new = new HashSet(INITIAL_CAPACITY_25);
            for (int i = 0; i < l_old.length; i++) {
                l_new.add(l_old[i]);
            } // for i
            l_new.addAll(Arrays.asList(_l_values) );
            m_TypesAndExplorer_.put(
                    _criteria, (String[])l_new.toArray(new String[0]));
            l_new = null;
        } 
        else m_TypesAndExplorer_.put(_criteria, _l_values);
        l_old = null;       
    }
    else m_TypesAndExplorer_.put(_criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addTypesAndExplorer(), "+
                      //      "fire the property change 'TypesAndExplorerSet' ");
    
    try {
        firePropertyChange("TypesAndExplorerSet", m_oldRights,
                (HashMap) m_TypesAndExplorer_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addTypesAndExplorer(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    m_oldRights = null;
    list = null;
  } // addTypesAndExplorer()

  /**
   * Removes some specialized values to manage the display in the explorer,
   * to set the allowed types and the group policy following the type.
   * The criteria is a double (key.subkey). <P>
   * Fires a PropertyChange for 'TypesAndExplorerSet'.
   * The criteria is not removed from the map if _l_values contains all the current
   * values which is associated to the criteria.
   * @param _criteria for the relevant specialized values.
   * @param _l_values Strings array which contain only
   *   the specialized values to remove for the criteria. Null to remove the criteria.
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is unknown.
   */
  public synchronized void removeTypesAndExplorer(String _criteria,
    String[] _l_values) throws UpDateError {
    
    if (_criteria == null ) {
        throw new UpDateError(" The criteria is not well formed.");
    }
    if (m_TypesAndExplorer_ == null) {
        throw new UpDateError(" The values are not well initialized.");
    }
    if (!m_TypesAndExplorer_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeTypesAndExplorer(), "+
                      //      "values to remove for "+_criteria);
    
    HashMap m_oldRights = (HashMap) m_TypesAndExplorer_.clone();
    if ( _l_values == null) {
        m_TypesAndExplorer_.remove(_criteria);

        try {
            firePropertyChange("TypesAndExplorerSet", m_oldRights,
                    (HashMap) m_TypesAndExplorer_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeTypesAndExplorer(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch 
        m_oldRights = null;
        return;
    } // if ( _l_values == null) 
    
    boolean error = false;
    
    for (int i = 0; i < _l_values.length; i++) {
      if (_l_values[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_values[i]);          
      }
    } // for i
    if (error) {
        throw new UpDateError(" A value in the argument is null.");
    }     
    
    String[] l_current = m_TypesAndExplorer_.get(_criteria);
    List list = new ArrayList(20);
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);
    }
    
    short count = 0;
    boolean found = false;

    for (int  i = 0 ; i < _l_values.length ; i++) {
        found = list.remove(_l_values[i]);
        
        if (!found) {
            error = true;
            break;
        } // if
        else count++;
            
    } // for i

    if (error) {
        m_oldRights = null;
        l_current = null;
        list = null;
        throw new UpDateError(" A value to remove is not found.");
    }     
    
    if (l_current.length != count) {
        m_TypesAndExplorer_.put(
                _criteria, 
                (String[]) list.toArray(new String[list.size()]));

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeTypesAndExplorer(), "+
                      //      "the new values for "+_criteria+" are: ");
                    //for (Object val : list) {
                      //  System.out.println("      "+val);
                    //}

    } else { // if _l_values contains all the current values
        m_TypesAndExplorer_.put(
                _criteria, new String[0]);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeTypesAndExplorer(), "+
                      //      "the new values for "+_criteria+" are an empty array. ");

    }

    try {
        firePropertyChange("TypesAndExplorerSet", m_oldRights,
                (HashMap) m_TypesAndExplorer_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeTypesAndExplorer(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch
    
    m_oldRights = null;
    l_current = null;
    list = null;
  } // removeTypesAndExplorer()
  
  /**
   * Gets the specialized values to manage the display in the explorer,
   * to set the allowed types and the group policy following the type.
   * The keys list is defined by getL_MapKeysForTypePolicies().
   * @return a copied Map where the key is a double of Strings, and the value is an array
   *   of Strings. May be empty, but not null. A value may be empty but is never null.
   */
  public final Map getM_TypesAndExplorer() {
      if (m_TypesAndExplorer_ == null)
          return new HashMap(INITIAL_CAPACITY_5);

      HashMap<String,String[]>  m_result = 
         new HashMap<String,String[]>(INITIAL_CAPACITY_25); 
      String[] l_values = null;
      Map.Entry<String,String[]> entry = null;
      String[] result = null;

              // for testing and tracking
              //System.out.println("in ACSFactoryImpl.getM_TypesAndExplorer(): ");
      
      for (Iterator ite = m_TypesAndExplorer_.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,String[]>)ite.next();

         l_values = entry.getValue();
         result = new String[l_values.length];
         System.arraycopy(l_values, 0, result, 0, l_values.length);  
         m_result.put(entry.getKey(), result);
         
              // for testing and tracking
              //System.out.println("--- for the key "+entry.getKey()); 
              //for (int i = 0; i < result.length; i++) {
                //    System.out.println("--------- at "+i+": "+result[i]);                  
              //} // for i
         
      } // for ite

      
      l_values = null;
      entry = null;
      result = null;     
      return m_result;
  } // getM_TypesAndExplorer()

  /**
   * This method returns the allowed keys which are used as map keys
   * for specialized types and explorer values. They are used in the two ACS maps
   * m_TypesAndGUIPolicy_ and m_TypeOrientedPolicy_. This does not include
   * the 5 following values which are set in another ACSFactoryImpl methods:
   * 'Resource.ResourceType', 'EligibleParty.EPType', 'VirtualFolderType',
   * 'PrivilegeForTypeType' and 'PrivilegeForLinksType'. <P>
   * Called by SimplePropertyEditor.getRestrictedValues() from the
   * ACSFactoryImplBeanInfo specification for StringMapPropertyEditor, and
   * by addTypesAndExplorer(). The key word 'immutable' may be
   * in an ACSObject type, but it is never used in the keys for type policies. <P>
   * The 35 key forms (key.subkey) or (key.subkey.type) are listed hereafter: <P>
   *     - 'Explorer.NoAcsNode',  <P>
   *     - 'Explorer.SubAcsNode', <P>
   *     - 'Explorer.AcsGroupTree', <P>
   *     - 'CreationByBeamer.NoType',  <P>
   *     - 'GroupIDMember.NoMemberOf',  <P>
   *     - 'GroupIDMember.IsNotConditionalAclSource', <P>
   *     - 'GroupID.NoPrimaryGroup',  <P>
   *     - 'GroupID.NoMainGroup', <P>
   *     - 'GroupID.NoSecondaryGroup', <P>
   *     - 'GroupID.MemberFromAdmin', <P>
   *     - 'GroupID.NoUserIDAsMember', <P>
   *     - 'GroupID.NoGroupIDAsMember', <P>
   *     - 'GroupID.NoMoreThanOneMember', <P>
   *     - 'GroupID.ConditionalAclGroup', <P>
   *     - 'Resource.OneRightACL', <P>
   *     - 'Resource.NoAccount', <P>
   *     - 'Resource.NoConditionalACL', <P>
   *     - 'Resource.NoNonConditionalACL', <P>
   *     - 'Resource.OneConditionGroupInACL', <P>
   *     - 'Actor.NoCurrentAccount', <P>
   *     - 'Actor.NoCurrentGroup', <P>
   *     - 'Actor.NoNullCurrentAccount', <P> 
   *     - 'Actor.IsNotBridgeTarget', <P>
   *     - 'Actor.IsNotBridgeSource', <P>
   *     - 'Actor.IsNotConditionalAclSource', <P>
   * <P>
   * The following keys are always ended by an allowed type: <P>
   *     - 'GroupID.TypesOfMemberFor.'type for EligibleParty', <P>
   *     - 'GroupIDMember.NoMoreThanOneGroup.'type of groupIDmember', <P>
   *     - 'GroupIDMember.RecommandedGroupsAsMember.'type of groupIDmember', <P>
   *     - 'Directory.TypesOfChildFor.'type of Resource or NULL', <P0>
   *     - 'VirtualFolder.TypesOfMemberFor.allowed_VirtualFolder_type', <P>
   *     - 'VirtualFolder.TypesOfChildFor.allowed_VirtualFolder_type or NULL', <P>
   *     - 'Resource.SecondTypesOfPrivilegeFor.allowed_Resource_type', <P>
   *     - 'EligibleParty.SecondTypesOfPrivilegeFor.allowed_EligibleParty_type', <P>
   *     - 'VirtualFolder.SecondTypesOfPrivilegeFor.allowed_VirtualFolder_type', <P>
   *     - 'Resource.TypesOfSourceForConditionalACL.allowed_Resource_type'. <P>
   * Calls BaseUtilityImpl.getObjectTypeWithoutImmutable().
   * @see ACSImpl#getM_TypesAndGUIPolicy
   * @see ACSFactoryImpl#getL_LimitedValuesForTypesAndExplorer
   * @return set of the allowed keys
   */
  public Set<String> getL_MapKeysForTypePolicies() {
    String[]    l_types = null;
    String      type = null;

    Set<String> l_result = new HashSet<String>(50);
    l_result.add("Explorer.NoAcsNode");
    l_result.add("Explorer.SubAcsNode");
    l_result.add("Explorer.AcsGroupTree");
    l_result.add("CreationByBeamer.NoType");
    l_result.add("GroupID.NoPrimaryGroup");
    l_result.add("GroupIDMember.NoMemberOf");
    l_result.add("GroupIDMember.IsNotConditionalAclSource");
    l_result.add("GroupID.NoMainGroup");
    l_result.add("GroupID.NoSecondaryGroup");
    l_result.add("GroupID.MemberFromAdmin");
    l_result.add("GroupID.NoUserIDAsMember");
    l_result.add("GroupID.NoGroupIDAsMember");
    l_result.add("GroupID.NoMoreThanOneMember");
    l_result.add("GroupID.ConditionalAclGroup");
    l_result.add("Resource.OneRightACL");
    l_result.add("Resource.NoAccount");
    l_result.add("Resource.NoConditionalACL");
    l_result.add("Resource.NoNonConditionalACL");
    l_result.add("Resource.OneConditionGroupInACL");
    l_result.add("Actor.NoCurrentAccount");
    l_result.add("Actor.NoCurrentGroup");
    l_result.add("Actor.NoNullCurrentAccount");
    l_result.add("Actor.IsNotBridgeTarget");
    l_result.add("Actor.IsNotBridgeSource");
    l_result.add("Actor.IsNotConditionalAclSource");

    l_types = getL_EPTypes();
    for (int i = 0; i < l_types.length; i++) {
        type = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(l_types[i]);
        l_result.add("GroupID.TypesOfMemberFor."+l_types[i]);
        l_result.add("GroupIDMember.RecommandedGroupsAsMember."+type);
        l_result.add("GroupIDMember.NoMoreThanOneGroup."+type);
        l_result.add("EligibleParty.SecondTypesOfPrivilegeFor."+type);
    } // for i

    l_result.add("Directory.TypesOfChildFor."+"NULL");
    l_types = getL_ResourceTypes();
    for (int i = 0; i < l_types.length; i++) {
        type = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(l_types[i]);
        l_result.add("Directory.TypesOfChildFor."+type);
        l_result.add("Resource.SecondTypesOfPrivilegeFor."+type);
        l_result.add("Resource.TypesOfSourceForConditionalACL."+type);
    } // for i

    l_result.add("VirtualFolder.TypesOfChildFor."+"NULL");
    l_types = getL_ResourceVirtualFolderTypes();
    for (int i = 0; i < l_types.length; i++) {
        type = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(l_types[i]);
        l_result.add("VirtualFolder.TypesOfMemberFor."+type);
        l_result.add("VirtualFolder.TypesOfChildFor."+type);
        l_result.add("VirtualFolder.SecondTypesOfPrivilegeFor."+type);
    } // for i

    l_types = getL_EPVirtualFolderTypes();
    for (int i = 0; i < l_types.length; i++) {
        type = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(l_types[i]);
        l_result.add("VirtualFolder.TypesOfMemberFor."+type);
        l_result.add("VirtualFolder.TypesOfChildFor."+type);
        l_result.add("VirtualFolder.SecondTypesOfPrivilegeFor."+type);
    } // for i

    l_types = null;
    type = null;
    return l_result;
  } // getL_MapKeysForTypePolicies()
  
  /**
   * Gets the specialized types which are allowed by the program for a key in the
   * types-and-explorer map and the group/node policy map.
   * This method is set in the BeanInfo, called
   * by StringMapPropertyEditor from a specification by ACSFactoryImplBeanInfo,
   * and is just a wraper for the ACSFactoryUtilityImpl method.
   * Called by checkACSbeforeCreation() only for the keys 'Explorer.SubAcsNode'
   * and 'Explorer.AcsGroupTree'.
   * @see ARoad0.gBase.ACSFactoryUtilityImpl#getL_LimitedValuesForTypesAndExplorer
   * @param _key for the constrained values
   * @return list of the allowed values. Never empty nor null. 
   */
  public final List<String> getL_LimitedValuesForTypesAndExplorer(String _key) {
      return ACSFactoryUtilityImpl.getL_LimitedValuesForTypesAndExplorer(this, _key);
  } // getL_LimitedValuesForTypesAndExplorer()
 
  /**
   * Adds one association to handle at the creation or the deletion of ACSObjects
   * in this ACS. The criteria is a map key among six allowed formats.
   * The allowed String map values depends lightly on the map key.
   * The non-empty values define fully or partially one action to do if the key matches.
   * One action is always a constructor or a method to call. <P>
   * This method provides an assistance to the user to build up a valid set
   * of actions with the proper commands and arguments.
   * See addAssociations() for understanding the complex logic.
   * The control of types in the keys implies that this editor must be called by the user
   * after the definition of the Resource and EligibleParty types in the acs.
   * Most of the keys are totally checked, except the keys 'Resource.ParentNode.&lt;Value&gt;.'.
   * The format is fully described at AssociateUtilityImpl.controlAssociations().
   * Calls the addAssociations method for the main work.
   * Fires PropertyChanges for 'AssociationForNewObjectsMap'.
   * @see AssociateUtilityImpl#getL_MapKeyStartsForAssociationsOnNewObject
   * @see ARoad0.Gui1.AssociationMapPropertyEditor
   * @param _criteria is the map key to apply these associations
   * @param _l_values a Strings array which contain the values for the criteria,
   *    but without the action number 'N.'. Values are added to the previous values, if any. 
   *    Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - a value is not allowed.
   */
  public synchronized void addAssociationsToNewObject(String _criteria, String[] _l_values) 
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("ACSFactoryImpl.addAssociationsToNewObject() called "+
                      //      "with criteria "+_criteria);
      
    if (_criteria == null || _l_values == null ) {
        throw new UpDateError(" The criteria or the values is not well formed.");
    }
    
    List<String>    list = AssociateUtilityImpl.getL_MapKeyStartsForAssociationsOnNewObject();
    boolean         error = true;
    HashMap         m_oldRights = null;

    for (Iterator ite = list.iterator(); ite.hasNext(); ) {
        if (_criteria.startsWith((String)ite.next())) {
            error = false;
            break;
        }
    }
    if (error)
        throw new UpDateError(" "+_criteria+" is not allowed as criteria.");  
    
    if (m_AssociationsForNewObjects_ != null) 
        m_oldRights = (HashMap) m_AssociationsForNewObjects_.clone();
    else {
        m_AssociationsForNewObjects_ = new HashMap(INITIAL_CAPACITY_25);
    }

    m_AssociationsForNewObjects_ = addAssociations(
          m_AssociationsForNewObjects_, _criteria, _l_values);
    
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addAssociationsToNewObject(), "+
                      //      "fire the property change 'AssociationForNewObjectsMap' ");
    
    try {
        firePropertyChange("AssociationForNewObjectsMap", m_oldRights, 
                (HashMap) m_AssociationsForNewObjects_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addAssociationsToNewObject(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }
    
    m_oldRights = null; 
    list = null;
  } // addAssociationsToNewObject()

  /**
   * Removes some associations to handle at the creation or the deletion of ACSObjects.
   * The format is fully described at AssociateUtilityImpl.controlAssociations().
   * Fires a PropertyChange for 'AssociationForNewObjectsMap'.
   * The criteria is not removed from the map if _l_values contains all the current values
   * which is associated to the criteria.
   * After the removing, the set of values is controlled and an error message is
   * added if the set is not compliant. A previous error message may be removed
   * ('Error:' or 'Error to delete:'), and also the action numbers.
   * Call removeAssociations() for the main work.
   * @see ARoad0.gBase.AssociateUtilityImpl#controlAssociations
   * @param _criteria is the map key to apply these associations
   * @param _l_values a Strings array which contain the values for the criteria,
   *    with an action number 'N.' or not. Null to remove the criteria. May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - _l_values is null or one value is null <P>
   *    - a value is unknown <P>
   *    - items in _l_lvalues are not in a sequence in the current values array.
   */
  public synchronized void removeAssociationsToNewObject(String _criteria, 
    String[] _l_values) throws UpDateError {
    
    if (_criteria == null) {
        throw new UpDateError(" The key is not well formed.");
    }
    if (m_AssociationsForNewObjects_ == null) {
        throw new UpDateError(" The property is not well initialized.");
    }    
    if (m_AssociationsForNewObjects_.get(_criteria) == null) {
        throw new UpDateError(" The key is unknown.");
    }     
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeAssociationsToNewObject(), "+
                      //      "values to remove for "+_criteria);
    
    HashMap             m_old = (HashMap) m_AssociationsForNewObjects_.clone();
    
    if ( _l_values == null) {
        m_AssociationsForNewObjects_.remove(_criteria);

        try {
            firePropertyChange("AssociationForNewObjectsMap", m_old,
                    (HashMap) m_AssociationsForNewObjects_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeAssociationsToNewObject(), firePropertChange() throws:");
            e.printStackTrace();
        } // catch 
        m_old = null;
        return;
    } // if ( _l_values == null) 

    m_AssociationsForNewObjects_ = removeAssociations(
          m_AssociationsForNewObjects_, _criteria, _l_values);

    try {
        firePropertyChange("AssociationForNewObjectsMap", m_old,
                (HashMap) m_AssociationsForNewObjects_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeAssociationsToNewObject(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch
    
    m_old = null;
  } // removeAssociationsToNewObject()  
  
  /**
   * Gets the associations to handle at the creation or the deletion of some ACSObjects,
   * by the mean of actions to do by the program. Each action has a number from 1 to 99.
   * The format is fully described at AssociateUtilityImpl.controlAssociations().
   * The internal map may contain values where the first 2 characters are not 'N.', but
   * the first no-number value and its followings are returned by this method which is used by
   * ActionMapPropertyEditor to edit the internal map.
   * Calls AssociateUtilityImpl.getM_FilteredAssociations() directly.
   * Called by gWork.AcsFactory.
   * @see ARoad0.gBase.AssociateUtilityImpl#controlAssociations
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings started by a number between 1 to 9 followed by a point.
   *   May be empty, but not null.
   */
  public final Map<String, String[]> getM_AssociationsOnNewObject() {
      return AssociateUtilityImpl.getM_FilteredAssociations(
              m_AssociationsForNewObjects_, false);
  } // getM_AssociationsOnNewObject()
 
  /**
   * Gets the set of the authorized values for the first value of an association.
   * Called by the AssociateMapPropertyEditor.
   * @see AssociateUtilityImpl#getL_LimitedFirstValuesForAssociation
   * @param _key for the constrained values. The single workable value is 'ForChildACS'.
   * @return list of the allowed values. Never empty nor null. 
   */
  public final List<String> getL_LimitedFirstValuesForAssociation(String _key) {
      return AssociateUtilityImpl.getL_LimitedFirstValuesForAssociation(_key);
  } // getL_LimitedFirstValuesForAssociation()

  /** 
   * Sets the fixed set of rights for the AclEntries.
   * A right may be not registered in getL_Rights().
   * Fires a PropertyChange for 'AclRightsSet'.
   * @param _s is an array of access rights for the AclEntries,
   *  in the main ACS. Not null, no null element, but may be a 0-length array.
   * @exception UpDateError if _s or one element is null, if the ACL rights are not allowed,
   *    or if there is no fixed set of rights for this acs
   */
  public synchronized void setL_AclRightsSet(StringRight[] _s) throws UpDateError {
    if (!m_Structure_.get("ACL right: internal ACL") && 
            !getManageExternalAcls() &&
            _s.length != 0)
        throw new UpDateError(" The internal or external ACL are not allowed.");
    if (l_AclRightsSet_ == null )
        throw new UpDateError(" The old ACL rights set is null.");

    controlRights(_s, true);
    
    StringRight[] l_oldRights = new StringRight[l_AclRightsSet_.length];
    System.arraycopy(l_AclRightsSet_, 0, l_oldRights, 0, l_AclRightsSet_.length);

    l_AclRightsSet_ = new StringRight[_s.length];
    System.arraycopy(_s, 0, l_AclRightsSet_, 0, _s.length);

    try {
        firePropertyChange("AclRightsSet", l_oldRights, _s);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFActoryImpl.setL_AclRightsSet(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldRights = null;

  } // setL_AclRightsSet()
  
  /**
   * @return set of allowed access rights for the AclEntries,
   *  in the main ACS. May be empty, but not null.
   */
  public final StringRight[] getL_AclRightsSet()	{
      StringRight[] result = new StringRight[l_AclRightsSet_.length];
      System.arraycopy(l_AclRightsSet_, 0, result, 0, l_AclRightsSet_.length);
      return result;
  }

  /**
   * Adds some privilege rights constraints, to apply as standard or specialized rights
   * to the privileges which have a given main type or a given second type.
   * The key defines the domain of the constraint. It may be for the standard or
   * for the specialized privilege rights. The possible keys are: <P>
   * - 'MainType.main type' for each of the 7 PrivilegeAbst main types, which sets
   * the 7 keys 'PrivilegeRights.MainType.main type' for the standard rights in ACSImpl, <P>
   * - 'PFType/PFLink.SecondType.second type' for the specialized rights, and
   * there is a returned value for each typed privilege second type
   * and each linked privilege second type. <P>
   * Note: this version manages only simple privileges (TYPED and LINKED types), and
   * the 5 other main types produce an exception.
   * A right may be not registered in getL_Rights().
   * Calls ACSFactoryUtilityImpl.getL_MapKeysForPrivilegeRights().
   * Fires PropertyChanges for 'PrivilegeRightsSet' and 'HavePrivilegeSpecializedRights'.
   * @since 0.7.0
   * @param _criteria is the map key to apply these rights.
   * @param _l_rights is a StringRights array, from the correct ACS, which contain 
   *   the new rights for the criteria. They are added to the previous rights, if any.
   *   Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a StringRight is not from the ACS.
   */
  public synchronized void addPrivilegeRights(String _criteria, 
    StringRight[] _l_rights) throws UpDateError {

                    // for testing and tracking
                    //System.out.println("Call to" +
                      //      " ACSFactoryImpl.addPrivilegeRights() ");;

    if (m_Structure_.get("PRI right: no PRI rights"))
        throw new UpDateError(" The privileges are not allowed.");

    if (_criteria == null || _l_rights == null ) {
        throw new UpDateError(" The criteria or the rights array is not well formed.");
    }
    if (_criteria.equals("MainType."+PrivilegeAbst.TYPED_CLASSED)
            || _criteria.equals("MainType."+PrivilegeAbst.TYPED_CLASSED_SPECIFIC)
            || _criteria.equals("MainType."+PrivilegeAbst.TYPED_FOR_SEED)
            || _criteria.equals("MainType."+PrivilegeAbst.LINKED_FOR_ONE_TO_ONE)
            || _criteria.equals("MainType."+PrivilegeAbst.LINKED_IF_ALL_SOURCES) )
        throw new UpDateError(" The criteria is applied to a privilege main type "+
                "which is not fully managed in this version of the program.");

    // controls the management of rights
    if (!getManageSimpleTypedPrivileges() && 
            (_criteria.startsWith("PFType") ||
            _criteria.equals("MainType."+PrivilegeAbst.TYPED)) )
                throw new UpDateError(" The simple typed privileges are not managed");
    if (!getManageSimpleLinkedPrivileges() &&
            (_criteria.startsWith("PFLink") ||
            _criteria.equals("MainType."+PrivilegeAbst.LINKED)) )
                throw new UpDateError(" The simple linked privileges are not managed");

    List<String> list = ACSFactoryUtilityImpl.getL_MapKeysForPrivilegeRights(this);
    boolean error = true;
    for(Iterator ite = list.iterator(); ite.hasNext(); ) {
        if (_criteria.equals((String)ite.next()) ) {
            error = false;
            break;
        }
    } // for ite
    if (error) throw new UpDateError(" "+_criteria+" is not allowed.");
    
    controlRights(_l_rights, true); //throws UpDateError        
    
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addPrivilegeRights(), "+
                      //      "the rights are controlled OK ");
                    
    // adds the type as a Privilege second type, if necessary
    if (!_criteria.startsWith("MainType")) {
        if (_criteria.startsWith("PFType") &&
                !Arrays.asList( l_privilegeForTypeTypes_ ).contains( _criteria.substring(18) )) {
                     String[] l_array = new String[l_privilegeForTypeTypes_.length + 1];
                     System.arraycopy(l_privilegeForTypeTypes_, 0, l_array, 0, l_privilegeForTypeTypes_.length);
                     l_array[l_array.length - 1] = _criteria.substring(18);
                     setL_PrivilegeForTypeTypes(l_array);
                     l_array = null;
        } // if (_criteria.startsWith("PFType")
        else if (_criteria.startsWith("PFLink") &&
                !Arrays.asList( l_privilegeForLinksTypes_ ).contains( _criteria.substring(18) )) {
                     String[] l_array = new String[l_privilegeForLinksTypes_.length + 1];
                     System.arraycopy(l_privilegeForLinksTypes_, 0, l_array, 0, l_privilegeForLinksTypes_.length);
                     l_array[l_array.length - 1] = _criteria.substring(18);
                     setL_PrivilegeForLinksTypes(l_array);
                     l_array = null;
        } // if (_criteria.startsWith("PFLink")
    } // if
    
    HashMap m_oldRights =  null;

    if (m_PrivilegeRights_ != null) 
        m_oldRights = (HashMap) m_PrivilegeRights_.clone();
    else {
        m_PrivilegeRights_ = new HashMap(INITIAL_CAPACITY_25);
    }

    if (m_PrivilegeRights_.get(_criteria) != null) {
        StringRight[] l_old = m_PrivilegeRights_.get(_criteria);
        if (l_old.length != 0) {
            Set<StringRight> l_new = new HashSet(INITIAL_CAPACITY_25);
            for (int i = 0; i < l_old.length; i++) {
                l_new.add(l_old[i]);
            } // for i
            l_new.addAll(Arrays.asList(_l_rights) );
            m_PrivilegeRights_.put(_criteria, (StringRight[])l_new.toArray(new StringRight[0]));
            l_new = null;
        } 
        else m_PrivilegeRights_.put(_criteria, _l_rights);
        l_old = null;       
    }
    else m_PrivilegeRights_.put(_criteria, _l_rights); 

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addPrivilegeRights(), "+
                            //"fires the property change 'PrivilegeRightsSet' ");
    
    try {
        firePropertyChange("PrivilegeRightsSet", m_oldRights, 
                (HashMap) m_PrivilegeRights_.clone());

        if (!havePrivilegeSpecializedRights_ && !_criteria.startsWith("MainType")) {
            havePrivilegeSpecializedRights_ = true;
            firePropertyChange("HavePrivilegeSpecializedRights", false, true);
        }
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addPrivilegeRights(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    m_oldRights = null;
    list = null;
  } // addPrivilegeRights()

  /**
   * Removes some privilege rights constraints, to apply as standard or specialized rights
   * to the privileges which have a given main type or a given second type.
   * The key defines the domain of the constraint. It may be for the standard or
   * for the specialized privilege rights. The possible keys are: <P>
   * - 'MainType.main type' for each of the 6 PrivilegeAbst main types, which sets
   * the 7 keys 'PrivilegeRights.MainType.main type' for the standard rights in ACSImpl, <P>
   * - 'PFType/PFLink.SecondType.second type' for the specialized rights, and
   * there is a returned value for each typed privilege second type
   * and each linked privilege second type. <P>
   * Fires a PropertyChange for 'PrivilegedRightsSet' and 'HavePrivilegeSpecializedRights'.
   * The criteria is not removed from the map if _l_rights contains all the current rights
   * which is associated to the criteria.
   * A removed type is not removed from the relevant types list, even if it has
   * been added in the list by addPrivilegeRights().
   * @since 0.7.0
   * @param _criteria for the relevant privilege rights.
   * @param _l_rights StringRights array which contain only
   *   the privilege rights to remove for the criteria. Null to remove the criteria.
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - _l_rights is null or one right is null <P>
   *    - a right in _l_rights is not registered.
   */
  public synchronized void removePrivilegeRights(String _criteria, 
    StringRight[] _l_rights) throws UpDateError {
    
    if (_criteria == null ) {
        throw new UpDateError(" The criteria is not well formed.");
    }
    if (m_PrivilegeRights_ == null) {
        throw new UpDateError(" The privilege rights are not well initialized.");
    }    
    if (!m_PrivilegeRights_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removePrivilegeRights(), "+
                      //      "rights to remove for "+_criteria);
    
    HashMap m_oldRights = (HashMap) m_PrivilegeRights_.clone();   
    if ( _l_rights == null) {
        m_PrivilegeRights_.remove(_criteria);
        try {
            firePropertyChange("PrivilegeRightsSet", m_oldRights,
                    (HashMap) m_PrivilegeRights_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removePrivilegeRightst(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch 
        m_oldRights = null;
        return;
    } // if ( _l_rights == null) 
    
    boolean error = false;
    
    for (int i = 0; i < _l_rights.length; i++) {
      if (_l_rights[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_rights[i]);          
      }
    } // for i
    if (error) {
        throw new UpDateError(" A privilege right in the argument is null.");
    }     
    
    StringRight[] l_current = m_PrivilegeRights_.get(_criteria); 
    List list = new ArrayList( 20 );
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);
    }
    
    short count = 0;
    boolean found = false;

    for (int  i = 0 ; i < _l_rights.length ; i++) {
        found = list.remove(_l_rights[i]);
        
        if (!found) {
            error = true;
            break;
        } // if
        else count++;
            
    } // for i

    if (error) {
        m_oldRights = null;
        l_current = null;
        list = null;
        throw new UpDateError(" A privilege right to remove is not found.");
    }     
    
    if (l_current.length != count) {
        m_PrivilegeRights_.put(
                _criteria, 
                (StringRight[]) list.toArray(new StringRight[list.size()]));
    } else { // if _l_rights contains all the current rights.
        m_PrivilegeRights_.put(
                _criteria, new StringRight[0]);
    }

    try {
        firePropertyChange("PrivilegeRightsSet", m_oldRights,
                (HashMap) m_PrivilegeRights_.clone());

        if (havePrivilegeSpecializedRights_) {
            havePrivilegeSpecializedRights_ = false;
            for (String st : m_PrivilegeRights_.keySet()) {
                if (!st.startsWith("MainType")) {
                    havePrivilegeSpecializedRights_ = true;
                    break;
                } // if
            } // for
            if (!havePrivilegeSpecializedRights_)
                firePropertyChange("HavePrivilegeSpecializedRights", true, false);
        } // if
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removePrivilegeRights(), firePropertChange() throws:");
        e.printStackTrace();
    } // catch
    
    m_oldRights = null;
    l_current = null;
    list = null;
  } // removePrivilegeRights()  
  
  /**
   * Gets the privilege rights constraints, to apply as standard or specialized rights
   * to the privileges which have a given main type or a given second type.
   * The key defines the domain of the constraint.
   * The key may have one of the following format: <P>
   * - 'MainType.main type' for each of the 7 PrivilegeAbst main types, which sets
   * the 7 keys 'PrivilegeRights.MainType.main type' for the standard rights in ACSImpl, <P>
   * - 'PFType/PFLink.PrvRightsSet.Type.second type' for the specialized rights, and
   * there is a returned value for each typed privilege second type
   * and each linked privilege second type. <P>
   * Note: this version does manage only simple TYPED and simple LINKED privileges.
   * A right may be not registered in getL_Rights().
   * @since 0.7.0
   * @return a copied Map where the key is a structured String, and the value is an array
   * of StringRights. May be empty, but not null.
   */
  public final Map getM_PrivilegeRights() {
      if (m_PrivilegeRights_ == null) return new HashMap(INITIAL_CAPACITY_5);

      HashMap<String,StringRight[]>  m_result = 
         new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25); 
      StringRight[] l_rights = null;
      Map.Entry<String,StringRight[]> entry = null;
      StringRight[] result = null;

              // for testing and tracking
              //System.out.println("in ACSFactoryImpl.getM_PrivilegeRights(): ");
      
      for (Iterator ite = m_PrivilegeRights_.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,StringRight[]>)ite.next();

         l_rights = entry.getValue();
         result = new StringRight[l_rights.length];
         System.arraycopy(l_rights, 0, result, 0, l_rights.length);  
         m_result.put(entry.getKey(), result);
         
              // for testing and tracking
              //System.out.println("--- for the key "+entry.getKey()); 
              //for (int i = 0; i < result.length; i++) {
                //    System.out.println("--------- at "+i+": "+result[i]);                  
              //} // for i
         
      } // for ite

      
      l_rights = null;
      entry = null;
      result = null;     
      return m_result;
  } // getM_PrivilegeRights()

  /**
   * This method returns the allowed keys for the standard and for the specialized
   * privilege rights: <P>
   * - 'MainType.main type' for each of the 6 PrivilegeAbst main types, which sets
   * the 6 keys 'PrivilegeRights.MainType.main type' for the standard rights in ACSImpl, <P>
   * - 'PFType/PFLink.PrvRightsSet.Type.second type' for the specialized rights, and
   * there is a returned value for each typed privilege second type
   * and each linked privilege second type. <P>
   * Note: the typed privilege types and the linked privilege types have to be all known.<P>
   * Calls ACSFactoryUtilityImpl.getL_MapKeysForPrivilegeRights(). Called for
   * BaseObjectMapPropertyEditor.actionPerformed() (through StringMapPropertyEditor.addKeyElement()
   * and SimplePropertyEditor.getRestrictedValues()).
   * @since 0.7.0
   * @return list of the keys for the privilege rights
   */
  public List<String> getL_MapKeysForPrivilegeRights() {
      return ACSFactoryUtilityImpl.getL_MapKeysForPrivilegeRights(this);
  }

  /**
   * Sets the allowed right values for the bridges from actors to other ACS.
   * A right may be not registered in getL_Rights().
   * Fires a PropertyChange for 'BridgeRightsSet'.
   * @param _s is an array of access rights for the ACS.
   *     Not null, no null or negative right, but may be a 0-length array.
   * @exception UpDateError if _s or one element is null, if the bridge rights are not allowed,
   *    or if there is no fixed set of rights for this acs
   */
  public synchronized void setL_bridgeRightsSet(StringRight[] _s) throws UpDateError {
    if (m_Structure_.get("bridge right: no BriDGes") && _s.length != 0)
        throw new UpDateError(" The bridges are not allowed.");
    if (l_bridgeRightsSet_ == null)
        throw new UpDateError(" The old bridge rights set is null.");

    controlRights(_s, false); // throws UpDateError
    
    StringRight[] l_oldRights = new StringRight[l_bridgeRightsSet_.length];
    System.arraycopy(l_bridgeRightsSet_, 0, l_oldRights, 0, l_bridgeRightsSet_.length);

    l_bridgeRightsSet_ = new StringRight[_s.length];
    System.arraycopy(_s, 0, l_bridgeRightsSet_, 0, _s.length);

    try {
        firePropertyChange("BridgeRightsSet", l_oldRights, _s);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFActoryImpl.setL_otherRights(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldRights = null;
  } // setL_bridgeRightsSet()
  
  /**
   * @return array of access rights for the privileges,
   *  in the main ACS. May be empty, but not null.
   */
  public final StringRight[] getL_bridgeRightsSet()	{
      StringRight[] result = new StringRight[l_bridgeRightsSet_.length];
      System.arraycopy(l_bridgeRightsSet_, 0, result, 0, l_bridgeRightsSet_.length);
      return result;
  }    

  /**
   * Adds a subACS name and its mode (for Resources or for GroupIDs) or a
   * root name for a groups tree, as a node to display in the explorer and the IS structure.
   * The key is the subACS name or a group root name, in one word,
   * and the value is an array which may contain the following values: <P>
   * &lt;ResourceSubACS&gt;, &lt;GroupSubACS&gt; and &lt;GroupTree&gt;. <P>
   * &lt;GroupTree&gt; is reserved to the group trees, and the other values are then forbidden.
   * A subACS may have one or two of the two values which are associated to.
   * This method does not define the Resource or EligibleParty type(s) which
   * have to be associated to a subACS or a group tree root. This has to be done by
   * the keys 'Explorer.SubAcsNode' and 'Explorer.AcsGroupTree'
   * in m_TypesAndExplorer_. To prepare the updating of this property
   * by the user, it is necessary to define previously the Resource or 
   * EligibleParty type(s) which are dedicated to these subACS or group tree root.
   * The values in m_TypesAndExplorer_ fulfill a policy that the
   * getL_LimitedValuesForTypesAndExplorer method sets. It states that: <P>
   *     - 'Explorer.SubAcsNode' sets some subACS nodes in the explorer from any
   *            current type starting with '&lt;byAcsCreator&gt; XXX' for Resources or EligibleParties,
   *            where the values have the form 'ResourceType.§lt;byAcsCreator&gt; XXX'
   *            or 'GroupType.&lt;byAcsCreator&gt; XXX', to specify the display
   *            under a dedicated node '&lt;XXX_subacs&gt;', of all the resources or groups
   *            having this relevant type. This type starting
   *            is reserved to these children in the explorer.  <P>
   *     - 'Explorer.AcsGroupTree' sets some group root nodes in the explorer from any
   *            current type starting with '&lt;byAcsCreator&gt; YYY' for EligibleParties,
   *            where the values have the form 'GroupType.&lt;byAcsCreator&gt; YYY',
   *            to specify the displaying directly under a dedicated node '&lt;YYY_tree&gt;'
   *            of all the groupIDs having the relevant type. This type starting
   *            is reserved to these children in the explorer.  <P>
   * This is why this method creates the type '&lt;byAcsCreator&gt; NNN' for Resources or
   * EligibleParties, where 'NNN' is the name of the new subACS or groups tree root,
   * and sets it for the new subACS or group tree root through a call
   * to the method addTypesAndExplorer.
   * That creation is mandatory to allow the next updating of m_TypesAndExplorer_.
   * The user may enter manually some other types like
   * '&lt;byAcsCreator&gt; NNN KKKK', if they are necessary in the subACS hierarchy or the group tree.
   * After such an addition, the user may also remove the new '&lt;byAcsCreator&gt; NNN' type.
   * All the types that contains '&lt;byAcsCreator&gt;' will be set
   * by initializeConstraintsBeforeCreation as forbidden to create a new object through
   * the beamer. <P>
   * May update l_resourceTypes_, l_epTypes_ and calls addTypesAndExplorer().
   * Fires PropertyChanges for 'SubAndGroupPolicy'.
   * @see ACSFactoryUtilityImpl#getL_LimitedValuesForSubAndGroupPolicy
   * @param _criteria the map key.
   * @param _l_values contains the allowed values for the criteria.
   * Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not allowed, <P>
   *    - a value is not allowed.
   */
  public synchronized void addSubAndGroupPolicy(String _criteria, String[] _l_values)
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("Call of ACSFactoryImpl.addSubAndGroupPolicy() ");;

    if (_criteria == null || _l_values == null ) {
        throw new UpDateError(" The key or the values are not well formed.");
    }

    boolean error = false;
    int     index = -1;
    List    l_allowed = ACSFactoryUtilityImpl.getL_LimitedValuesForSubAndGroupPolicy();
    List    l_val = null;

    for (int i = 0; i < _l_values.length; i++) {
        if (!l_allowed.contains(_l_values[i])) {
            error = true;
            index = i;
            break;
        }
    } // for i
    l_allowed = null;
    if (error)
        throw new InternalError("In ACSFactoryImpl.addSubAndGroupPolicy(),"+
            " this property does not allow the value "+_l_values[index]);
    
    if (_l_values.length != 0) {
        l_val = Arrays.asList(_l_values);
    } // if

    HashMap m_old =  null;
    if (m_SubAndGroupPolicy_ != null)
        m_old = (HashMap) m_SubAndGroupPolicy_.clone();
    else {
        m_SubAndGroupPolicy_ = new HashMap(INITIAL_CAPACITY_25);
    }

    if (m_SubAndGroupPolicy_.get(_criteria) != null) {
        String[] l_old = m_SubAndGroupPolicy_.get(_criteria);
        if (l_old.length != 0) {
            Set<String> l_new = new HashSet(INITIAL_CAPACITY_25);
            for (int i = 0; i < l_old.length; i++) {
                l_new.add(l_old[i]);
            } // for i
            l_new.addAll(Arrays.asList(_l_values) );

            if (l_new.contains("<GroupTree>") && l_new.size() != 1) {
                l_old = null;
                l_val = null;
                l_new = null;
                throw new UpDateError(" The value '<GroupTree>' has to be alone.");
            }

            m_SubAndGroupPolicy_.put(
                    _criteria, (String[])l_new.toArray(new String[0]));
            l_new = null;
        }
        else {
            if (l_val != null && l_val.contains("<GroupTree>") && l_val.size() != 1) {
                l_old = null;
                l_val = null;
                throw new UpDateError(" The value '<GroupTree>' has to be alone.");
            }

            m_SubAndGroupPolicy_.put(_criteria, _l_values);
        }
        l_old = null;
    }
    else m_SubAndGroupPolicy_.put(_criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addSubAndGroupPolicy(), "+
                      //      "fire the property change 'SubAndGroupPolicy' ");

    try {
        firePropertyChange("SubAndGroupPolicy", m_old, (HashMap) m_SubAndGroupPolicy_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addSubAndGroupPolicy(), firePropertChange() throws:");
        e.printStackTrace();
        return;
    }

    if (l_val != null) {
        // creates the '<byAcsCreator> NNN' type
        if (l_val.contains("<ResourceSubACS>") &&
                !Arrays.asList(l_resourceTypes_).contains("<byAcsCreator> "+_criteria) ) {
            String[] l_Types = new String[l_resourceTypes_.length + 1];
            System.arraycopy(l_resourceTypes_, 0, l_Types, 0, l_resourceTypes_.length);
            l_Types[l_Types.length - 1] = "<byAcsCreator> "+_criteria;
            setL_ResourceTypes(l_Types);
            l_Types = null;
        } // if
        if ((l_val.contains("<GroupSubACS>") || l_val.contains("<GroupTree>")) &&
                !Arrays.asList(l_epTypes_).contains("<byAcsCreator> "+_criteria) ) {
            String[] l_Types = new String[l_epTypes_.length + 1];
            System.arraycopy(l_epTypes_, 0, l_Types, 0, l_epTypes_.length);
            l_Types[l_Types.length - 1] = "<byAcsCreator> "+_criteria;
            setL_EPTypes(l_Types);
            l_Types = null;
        } // if

        // associates the '<byAcsCreator> NNN' type to the subACS or the group tree root.
        // The user may add other types directly.
        try {
            if (l_val.contains("<ResourceSubACS>"))
                  addTypesAndExplorer(
                          "Explorer.SubAcsNode",
                          new String[] {"ResourceType."+"<byAcsCreator> "+_criteria});
            if (l_val.contains("<GroupSubACS>"))
                  addTypesAndExplorer(
                          "Explorer.SubAcsNode",
                          new String[] {"GroupType."+"<byAcsCreator> "+_criteria});
            if (l_val.contains("<GroupTree>"))
                  addTypesAndExplorer(
                          "Explorer.AcsGroupTree",
                          new String[] {"GroupType."+"<byAcsCreator> "+_criteria});
        } catch (UpDateError u) {
          throw new InternalError("ACSFactoryImpl.addSubAndGroupPolicy() "+
                  "error :"+u);
        }
    } // l_val

    m_old = null;
    l_val = null;
  } // addSubAndGroupPolicy()

  /**
   * Removes a subACS or a groups tree node, and its dedicated node in the explorer.
   * The key is the subACS name or the groups tree name, in one word, and
   * the value is an array which may contain the following values: <P>
   * &lt;ResourceSubACS&gt;, &lt;GroupSubACS&gt; and &lt;GroupTree&gt;. <P>
   * Fires a PropertyChange for 'SubAndGroupPolicy'.
   * @param _criteria for the relevant value.
   * @param _l_values contains only the value to remove for the criteria.
   * Null to remove the criteria. May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is not registered.
   */
  public synchronized void removeSubAndGroupPolicy(String _criteria,
    String[] _l_values) throws UpDateError {

    if (_criteria == null ) {
        throw new UpDateError(" The criteria is not well formed.");
    }
    if (m_SubAndGroupPolicy_ == null) {
        throw new UpDateError(" The values are not well initialized.");
    }
    if (!m_SubAndGroupPolicy_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeSubAndGroupPolicy(), "+
                      //      "values to remove for "+_criteria+" are: ");
                    //if (_l_values != null) {
                      //  for (int i = 0; i < _l_values.length; i++) {
                        //    System.out.println("--------- "+_l_values[i]);
                       // }
                    //} // if

    HashMap     m_old = (HashMap) m_SubAndGroupPolicy_.clone();
    String[]    l_current = m_SubAndGroupPolicy_.get(_criteria);
    List        l_Types = null;

    if ( _l_values == null) {
        m_SubAndGroupPolicy_.remove(_criteria);

        try {
            firePropertyChange("SubAndGroupPolicy", m_old,
                    (HashMap) m_SubAndGroupPolicy_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeSubAndGroupPolicy(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;


        // deletes the '<byAcsCreator> NNN' type
        if (Arrays.asList(l_resourceTypes_).contains("<byAcsCreator> "+_criteria) ) {
            l_Types = new ArrayList(20);
            for (int i = 0; i < l_resourceTypes_.length; i++) {
                if (!l_resourceTypes_[i].equals("<byAcsCreator> "+_criteria))
                    l_Types.add(l_resourceTypes_[i]);
            }
            setL_ResourceTypes((String[])l_Types.toArray(new String[0]));
        } // if
        if (Arrays.asList(l_epTypes_).contains("<byAcsCreator> "+_criteria) ) {
            l_Types = new ArrayList(20);
            for (int i = 0; i < l_epTypes_.length; i++) {
                if (!l_epTypes_[i].equals("<byAcsCreator> "+_criteria))
                    l_Types.add(l_epTypes_[i]);
            }
            setL_EPTypes((String[])l_Types.toArray(new String[0]));
        } // if

        // deletes the '<byAcsCreator> NNN' type for the subACS or the group tree root.
        // The user may have added other similar types directly.
        try {
            l_current = m_TypesAndExplorer_.get("Explorer.SubAcsNode");
            if (l_current != null && l_current.length != 0 &&
                  Arrays.asList(l_current).contains("ResourceType."+"<byAcsCreator> "+_criteria))
                      removeTypesAndExplorer(
                              "Explorer.SubAcsNode",
                              new String[] {"ResourceType."+"<byAcsCreator> "+_criteria});
            if (l_current != null && l_current.length != 0 &&
                  Arrays.asList(l_current).contains("GroupType."+"<byAcsCreator> "+_criteria))
                      removeTypesAndExplorer(
                              "Explorer.SubAcsNode",
                              new String[] {"GroupType."+"<byAcsCreator> "+_criteria});

            l_current = m_TypesAndExplorer_.get("Explorer.AcsGroupTree");
            if (l_current != null && l_current.length != 0 &&
                  Arrays.asList(l_current).contains("GroupType."+"<byAcsCreator> "+_criteria))
                      removeTypesAndExplorer(
                              "Explorer.AcsGroupTree",
                              new String[] {"GroupType."+"<byAcsCreator> "+_criteria});
        } catch (UpDateError u) {
          throw new InternalError("ACSFactoryImpl.removeSubAndGroupPolicy() "+
                  "error: "+u);
        }

        m_old = null;
        l_current = null;
        l_Types = null;
        return;
    } // if ( _l_values == null)

    boolean error = false;

    for (int i = 0; i < _l_values.length; i++) {
      if (_l_values[i] == null)  {
          error = true;
          break;
      }
      //else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_values[i]);
      //}
    } // for i
    if (error) {
        m_old = null;
        l_current = null;
        l_Types = null;
        throw new UpDateError(" A value in the argument is null.");
    }

    List    list = new ArrayList( 20 );
    short   count = 0;
    boolean found = false;
    boolean noValue = false;
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);

                // for testing and tracking
                //System.out.println("From ACSFactoryImpl.removeSubAndGroupPolicy():"+
                  //      "current value for "+_criteria+" is "+l_current[i]);
    } // for i

    for (int  i = 0 ; i < _l_values.length ; i++) {
        found = list.remove(_l_values[i]);

        if (!found) {
            error = true;
            break;
        } // if
        else count++;

    } // for i

    if (error) {
        m_old = null;
        l_current = null;
        list = null;
        l_Types = null;
        throw new UpDateError(" A value to remove is not found.");
    }

    if (l_current.length != count) {
        m_SubAndGroupPolicy_.put(
                _criteria,
                (String[]) list.toArray(new String[list.size()]));
    } else { // if _l_values contains all the current values
        noValue = true;
        m_SubAndGroupPolicy_.put(
                _criteria, new String[0]);
    }

    try {
        firePropertyChange("SubAndGroupPolicy", m_old,
                (HashMap) m_SubAndGroupPolicy_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeSubAndGroupPolicy(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
        return;
    } // catch

    if (noValue) {
        m_old = null;
        l_current = null;
        list = null;
        l_Types = null;

                // for testing and tracking
                //System.out.println("From ACSFactoryImpl.removeSubAndGroupPolicy():"+
                  //      "no value for "+_criteria);

        return;
    } // if

    // deletes the '<byAcsCreator> NNN' correct type if there is no value
    if (Arrays.asList(l_resourceTypes_).contains("<byAcsCreator> "+_criteria) &&
              Arrays.asList(_l_values).contains("<ResourceSubACS>")) {
        l_Types = new ArrayList(20);
        for (int i = 0; i < l_resourceTypes_.length; i++) {
            if (!l_resourceTypes_[i].equals("<byAcsCreator> "+_criteria))
                l_Types.add(l_resourceTypes_[i]);
        }
        setL_ResourceTypes((String[])l_Types.toArray(new String[0]));
    } // if

                    //if (l_epTypes_ != null) {
                      //  System.out.println("From ACSFactoryImpl.removeSubAndGroupPolicy(), l_epTypes_:");
                      //  for (int i = 0; i < l_epTypes_.length; i++) {
                        //    System.out.println("--------- "+l_epTypes_[i]);
                      //  }
                    //} // if

    if (Arrays.asList(l_epTypes_).contains("<byAcsCreator> "+_criteria) &&
              (!list.contains("<GroupSubACS>") &&
              !list.contains("<<GroupTree>"))) {
        l_Types = new ArrayList(20);
        for (int i = 0; i < l_epTypes_.length; i++) {
            if (!l_epTypes_[i].equals("<byAcsCreator> "+_criteria))
                l_Types.add(l_epTypes_[i]);
        }
        setL_EPTypes((String[])l_Types.toArray(new String[0]));


                    //if (l_epTypes_ != null) {
                      //  System.out.println("From ACSFactoryImpl.removeSubAndGroupPolicy(), "+
                        //        "after the removing of "+_criteria+" in l_epTypes_:");
                        //for (int i = 0; i < l_epTypes_.length; i++) {
                          //  System.out.println("--------- "+l_epTypes_[i]);
                        //}
                    //} // if

    } // if

    // deletes the '<byAcsCreator> NNN' type for the subACS or the group tree root.
    // The user may have added other types manually.
    try {
        l_current = m_TypesAndExplorer_.get("Explorer.SubAcsNode");
        if (l_current != null && l_current.length != 0 &&
              Arrays.asList(l_current).contains("ResourceType."+"<byAcsCreator> "+_criteria) &&
              Arrays.asList(_l_values).contains("<ResourceSubACS>"))
                  removeTypesAndExplorer(
                          "Explorer.SubAcsNode",
                          new String[] {"ResourceType."+"<byAcsCreator> "+_criteria});
        if (l_current != null && l_current.length != 0 &&
              Arrays.asList(l_current).contains("GroupType."+"<byAcsCreator> "+_criteria) &&
              Arrays.asList(_l_values).contains("<GroupSubACS>"))
                  removeTypesAndExplorer(
                          "Explorer.SubAcsNode",
                          new String[] {"GroupType."+"<byAcsCreator> "+_criteria});

        l_current = m_TypesAndExplorer_.get("Explorer.AcsGroupTree");
        if (l_current != null && l_current.length != 0 &&
              Arrays.asList(l_current).contains("GroupType."+"<byAcsCreator> "+_criteria) &&
              Arrays.asList(_l_values).contains("<GroupTree>"))
                  removeTypesAndExplorer(
                          "Explorer.AcsGroupTree",
                          new String[] {"GroupType."+"<byAcsCreator> "+_criteria});
    } catch (UpDateError u) {
      throw new InternalError("ACSFactoryImpl.removeSubAndGroupPolicy() "+
              "error: "+u);
    }

    m_old = null;
    l_current = null;
    list = null;
    l_Types = null;
  } // removeSubAndGroupPolicy()

  /**
   * Gets the subACS names and their mode (for Resources or for GroupIDs) and the
   * root names for groups trees, as nodes to display in the explorer.
   * A subACS is always open with its ACS.
   * In the returned value, the key is the subACS name or a group root name, in one word,
   * and the value is an array which may contain one or two items among the following values: <P>
   * &lt;ResourceSubACS&gt;, &lt;GroupSubACS&gt; and &lt;GroupTree&gt;. <P>
   * For instance, the value 'yyy' for the key &lt;GroupTree&gt;,
   * means that a group root node '&lt;yyy_tree&gt;' has to be displayed
   * directly under the GroupIDs node, in the explorer. For the two other keys,
   * a value 'xxxx' means that a node '&lt;xxxx_subacs&gt;' has to be displayed directly
   * under the Resources and the AclEntries nodes, or under the GroupIDs node.
   * @return the map of the subACSs and the groups tree roots. May be empty but never null.
   */
  public Map<String, String[]> getM_SubAndGroupPolicy() {
      if (m_SubAndGroupPolicy_ != null)
          return (HashMap<String, String[]>)m_SubAndGroupPolicy_.clone();
      else return new HashMap(INITIAL_CAPACITY_5);      
  }  

  /**
   * This method defines the subacs and the groups tree nodes in the GUI explorer.
   * Called by ACSFactoryImpl for StringMapPropertyEditor.
   * '&lt;ResourceSubACS&gt;' means that a node for this subACS has to be displayed
   * directly under the Resources node, in the explorer.
   * '&lt;GroupSubACS&gt;' for a subACS means that a node has to be displayed
   * directly under the Groups node, in the explorer.
   * '&lt;GroupTree&gt;' for a groups tree means that a node has to be displayed
   * directly under the Groups node, in the explorer, as root of a specific tree.
   * @return list of the allowed values. Never empty nor null.
   */
  public final List<String> getL_LimitedValuesForSubAndGroupPolicy(String _key) {
      return ACSFactoryUtilityImpl.getL_LimitedValuesForSubAndGroupPolicy();
  }

  /**
   * Gets the map which tells the ACS or a subACS belongs to a Composite ACS
   * in the IS Name Tree. <P>
   * In the returned value, the key is the subACS name, in one word, 
   * or the ACS name last component, and the value is one Composite name in an array.
   * @since 0.7.0
   * @return the map of the composite relations. Never null and may be empty.
   */
  public Map<String, ImmutableName[]> getM_BelongsToComposite() {
      if (m_BelongsToComposite_ != null) {
          HashMap<String, ImmutableName[]> l_result = new HashMap(INITIAL_CAPACITY_5);
          ImmutableName[] l_values = null;

          for (Map.Entry entry : m_BelongsToComposite_.entrySet()) {
              if (entry.getValue() != null)
                  l_values = new ImmutableName[] {(ImmutableName)entry.getValue()};
              else
                  l_values = new ImmutableName[0];
              l_result.put((String)entry.getKey(), l_values);
          } // for entry
          l_values = null;
          return l_result;
      }
      else return new HashMap(INITIAL_CAPACITY_5);
  } // getM_BelongsToComposite()
  
  /**
   * Links a composite node in the IS nametree to the ACS or to a subACS.
   * The key is the subACS name, in one word,
   * or the ACS name last component, and the value is one Composite name
   * in an array.
   * Fires PropertyChanges for 'BelongsToComposite'.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#getL_LimitedValuesForComponentNodes
   * @param _criteria the map key.
   * @param _l_values array which contains the allowed values
   * for the criteria. Never null. The lenght may be 0 or 1.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not allowed, <P>
   *    - a value is not allowed, or there is already a value.
   */
  public synchronized void addBelongsToComposite(String _criteria,
          ImmutableName[] _l_values) throws UpDateError {

                    // for testing and tracking
                    //System.out.println("Call to ACSFactoryImpl.addBelongsToComposite() ");

    if (_criteria == null || _l_values == null) {
        throw new UpDateError(" Sorry, the key or the value is null.");
    }

    if (_l_values.length > 1) {
        throw new UpDateError(" Sorry, no more than one value.");
    }

    if (m_SubAndGroupPolicy_ == null &&
            !name_.lastComponent().equals(_criteria))
        throw new UpDateError(" The key is not a known subACS nor the ACS name last component.");

    if (m_SubAndGroupPolicy_ != null &&
            !m_SubAndGroupPolicy_.keySet().contains(_criteria) &&
                !name_.lastComponent().equals(_criteria))
        throw new UpDateError(" The key is not a known subACS nor the ACS name last component.");

    if (_l_values.length != 0 &&
            !getL_LimitedValuesForBelongsToComposite(_criteria).contains(_l_values[0]) )
        throw new UpDateError(" The value is not a current or new composite.");

    Map m_old =  null;
    if (m_BelongsToComposite_ != null)
        m_old = getM_BelongsToComposite();
    else {
        m_BelongsToComposite_ = new HashMap(INITIAL_CAPACITY_25);
    }

    if (m_BelongsToComposite_.get(_criteria) != null) {
        throw new UpDateError(
            " This property needs exactly one value per key.");
    }
    else if (_l_values.length == 0) {
        m_BelongsToComposite_.put(_criteria, null);
    } else m_BelongsToComposite_.put(_criteria, _l_values[0]);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addBelongsToComposite(), "+
                      //      "fire the property change 'BelongsToComposite' ");

    try {
        firePropertyChange("BelongsToComposite", m_old, getM_BelongsToComposite());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addBelongsToComposite(), firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null;
  } // addBelongsToComposite()

  /**
   * Removes the link to a composite node in the IS nametree to the ACS or to a subACS.
   * Fire a PropertyChange for 'BelongsToComposite'.
   * The criteria is not removed from the map if _l_values contains all the current
   * values which is associated to the criteria.
   * @since 0.7.0
   * @param _criteria for the relevant value.
   * @param _l_values array which contains only
   *   the value to remove for the criteria. Null to remove the criteria.
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is not registered.
   */
  public synchronized void removeBelongsToComposite(String _criteria, 
    ImmutableName[] _l_values) throws UpDateError {

    if (_criteria == null || (_l_values != null && _l_values.length != 1)) {
        throw new UpDateError(" The key or the value is not well formed.");
    }
    if (m_BelongsToComposite_ == null) {
        throw new UpDateError(" The property is not well initialized.");
    }
    if (!m_BelongsToComposite_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeBelongsToComposite(), "+
                      //      "values to remove for "+_criteria);

    Map m_old = getM_BelongsToComposite();
    
    if ( _l_values != null &&
            !((ImmutableName)m_BelongsToComposite_.get(_criteria)).equals(_l_values[0]))
        throw new UpDateError(" Unknown value to remove.");

    m_BelongsToComposite_.remove(_criteria);

    try {
        firePropertyChange("BelongsToComposite", m_old, getM_BelongsToComposite());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeBelongsToComposite(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
  } // removeBelongsToComposite()

  /**
   * Gets the current ACS and subACS name last compoents.
   * This method is set in the BeanInfo, called
   * by BaseObjectMapPropertyEditor from a specification by ACSFactoryImplBeanInfo.
   * @since 0.7.0
   * @return the current ACS and subACS name last compoents. Never null nor empty.
   */
  public final List<String> getL_currentAcsAndSubAcsNickNames() {
      List<String>  l_result = new ArrayList(INITIAL_CAPACITY_5);
      l_result.add(getNickName());
      for (String subacs : getM_SubAndGroupPolicy().keySet()) {
            l_result.add(subacs);
      }
      return l_result;
  } // getL_currentAcsAndSubAcsNickNames()

  /**
   * Gets the current composites in the IS, and adds the new composites that
   * this instance will create. If the argument _key is the new ACS last component,
   * then only ACS composite names are returned. Otherwise, only subACS composite
   * names are returned. The composites map is used for the map keys which have
   * a defined type as value in the composite map. The returned current composites have not
   * to be open.
   * This method is set in the BeanInfo, called
   * by BaseObjectMapPropertyEditor from a specification by ACSFactoryImplBeanInfo.
   * Calls BaseManagerImpl.getL_CompositeNodesOfIS() and getISTree(), and
   * ImmutableNameISTree.getType().
   * @since 0.7.0
   * @param _key the ACS name last component or a subACS name
   * @return the allowed composite names. Never null and may be empty.
   */
  public final List<ImmutableName> getL_LimitedValuesForBelongsToComposite(String _key) {
      List<ImmutableName>   l_result = new ArrayList(INITIAL_CAPACITY_5);
      boolean               keyIsForTheNewACS = false; // if false, the key is for a subACS
      ImmutableNameISTree   nametree = BaseManagerImpl.getInstance().getISTree(iSname_);
      Map.Entry<String,String[]> entry = null;
      ImmutableName         compositeName = null;

      if (name_.lastComponent().equals(_key)) keyIsForTheNewACS = true;

      // select the right current composites
      try {
          for (ImmutableName name :
              BaseManagerImpl.getInstance().getL_CompositeNodesOfIS( iSname_)) {
                if (keyIsForTheNewACS &&
                        !nametree.getType(name).equals(
                            ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE))
                                l_result.add(name);
                else if (!keyIsForTheNewACS &&
                        nametree.getType(name).equals(
                            ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE))
                                l_result.add(name);
          } // for name
      } catch (BaseError b) {
          throw new InternalError("ACSFactoryImpl.getL_LimitedValuesForBelongsToComposite(), "+
                  " for the key "+_key+": "+b.toString());
      }

      if (m_CompositeNodes_ != null) {
          for (Iterator ite = m_CompositeNodes_.entrySet().iterator(); ite.hasNext(); ) {
             entry = (Map.Entry<String,String[]>)ite.next();

             // the current map key which have a defined type as value
             if ( ((String[])entry.getValue()).length == 0 ) continue;

             compositeName = (ImmutableName) ((NameImpl)iSname_.clone()).
                                                extended((String)entry.getKey());

             if (keyIsForTheNewACS &&
                    !((String[])entry.getValue())[0].equals("subacs"))
                            l_result.add(compositeName);
             else if (!keyIsForTheNewACS &&
                    ((String[])entry.getValue())[0].equals("subacs"))
                            l_result.add(compositeName);

          } // for ite
      } // if (m_CompositeNodes_

      nametree = null;
      entry = null;
      compositeName = null;
      return l_result;
  } // getL_LimitedValuesForBelongsToComposite()

  /**
   * Adds some component nodes in the IS, from the new ACS node or from its parent.
   * The criteria has the format key.subname, where subname is the String form
   * of a Name which contains one component or more. The property editor allows only
   * one value per key. The allowed values are 'physical' and 'logical',
   * and the property editor controls them. <P>
   * Calls getL_MapKeysStartsForComponents(). No control on the current component nodes,
   * to set there is no double.
   * Fires PropertyChanges for 'ComponentNodes'.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#getL_LimitedValuesForComponentNodes
   * @param _criteria the map key.
   * @param _l_values String array which contains the allowed values
   * for the criteria. Never null. The lenght may be 0 or 1.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not allowed, <P>
   *    - a value is not allowed, or there is still one value.
   */
  public synchronized void addComponentNode(String _criteria, String[] _l_values)
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("Call of ACSFactoryImpl.addComponentNode() ");;

    if (_criteria == null || _l_values == null) {
        throw new UpDateError(" The key or the value is null.");
    }

    if (_l_values.length > 1) {
        throw new UpDateError(" More than one value is not allowed.");
    }

    List<String> list = getL_MapKeysStartsForComponents();
    if (!list.contains(_criteria.substring(0, _criteria.indexOf(".") + 1)))
        throw new UpDateError(" The key has to start with 'IsChildOfNewAcs.'"+
                " or 'IsChildOfNewAcsParent.'.");

    HashMap m_old =  null;
    boolean error = false;
    Name    name = null;

    try {

        // cuts the leading and trailing whitespaces, for which the number may vary,
        // then adds the right whitespaces to build up the String format of a name
        name = NameImpl.getNameFromString( " "+_criteria.substring(
                _criteria.indexOf(".") + 1 ).trim() + "  ");

                                // for testing and tracking
                                System.out.println("ACSFactoryImpl.addComponentNode(), "+
                                        "the component name is "+name);

    } catch (BaseError b) {
        throw new UpDateError(" The key does not end with a true name. "+b.toString());
    }

    int index = -1;
    for (int i = 0; i < _l_values.length; i++) {
        if (!_l_values[i].equals("physical") && !_l_values[i].equals("logical")) {
            error = true;
            index = i;
            break;
        }
    } // for i
    if (error)
        throw new UpDateError(
            " The value "+_l_values[index]+" is forbidden.");

    if (m_ComponentNodes_ != null)
        m_old = (HashMap) m_ComponentNodes_.clone();
    else {
        m_ComponentNodes_ = new HashMap(INITIAL_CAPACITY_25);
    }

    if (m_ComponentNodes_.get(_criteria) != null &&
            ((String[])m_ComponentNodes_.get(_criteria)).length != 0) {
        throw new UpDateError(
            " This property enforces exactly one value per key.");
    }
    else m_ComponentNodes_.put(_criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addTypesAndExplorer(), "+
                      //      "fire the property change 'TypesAndExplorerSet' ");

    try {
        firePropertyChange("ComponentNodes", m_old, (HashMap) m_ComponentNodes_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addComponentNode(), firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null;
    list = null;
    name = null;
  } // addComponentNode()

  /**
   * Removes a new component node in the IS nametree.
   * Fires a PropertyChange for 'ComponentNodes'.
   * The criteria is not removed from the map if _l_values contains all the current
   * values which is associated to the criteria.
   * @since 0.7.0
   * @param _criteria for the relevant specialized values.
   * @param _l_values array which contains only
   *   the value to remove for the criteria. Null to remove the criteria.
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is not registered.
   */
  public synchronized void removeComponentNode(String _criteria,
    String[] _l_values) throws UpDateError {

    if (_criteria == null ) {
        throw new UpDateError(" The criteria is not well formed.");
    }
    if (m_ComponentNodes_ == null) {
        throw new UpDateError(" The values are not well initialized.");
    }
    if (!m_ComponentNodes_.keySet().contains(_criteria)) {
        throw new UpDateError(" Unknown key.");
    }
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeComponentNode(), "+
                      //      "values to remove for "+_criteria);

    HashMap m_old = (HashMap) m_ComponentNodes_.clone();
    if ( _l_values == null) {
        m_ComponentNodes_.remove(_criteria);

        try {
            firePropertyChange("ComponentNodes", m_old,
                    (HashMap) m_ComponentNodes_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeComponentNode(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)

    boolean error = false;

    for (int i = 0; i < _l_values.length; i++) {
      if (_l_values[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_values[i]);
      }
    } // for i
    if (error) {
        throw new UpDateError(" A value in the argument is null.");
    }

    String[] l_current = m_ComponentNodes_.get(_criteria);
    List list = new ArrayList( 20 );
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);
    }

    short count = 0;
    boolean found = false;

    for (int  i = 0 ; i < _l_values.length ; i++) {
        found = list.remove(_l_values[i]);

        if (!found) {
            error = true;
            break;
        } // if
        else count++;

    } // for i

    if (error) {
        m_old = null;
        l_current = null;
        list = null;
        throw new UpDateError(" A value to remove is not found.");
    }

    if (l_current.length != count) {
        m_ComponentNodes_.put(
                _criteria,
                (String[]) list.toArray(new String[list.size()]));
    } else { // if _l_values contains all the current values
        m_ComponentNodes_.put(
                _criteria, new String[0]);
    }

    try {
        firePropertyChange("ComponentNodes", m_old,
                (HashMap) m_ComponentNodes_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeComponentNode(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
    l_current = null;
    list = null;
  } // removeComponentNode()

  /**
   * Gets the new component nodes in the IS nametree.
   * @since 0.7.0
   * @return a copied Map where the key is a double of Strings, and the value is
   * an array of one String. May be empty, but not null.
   * A value may be empty but is never null.
   */
  public final Map getM_ComponentNodes() {
      if (m_ComponentNodes_ == null) return new HashMap(INITIAL_CAPACITY_5);

      HashMap<String,String[]>  m_result =
         new HashMap<String,String[]>(INITIAL_CAPACITY_25);
      String[] l_values = null;
      Map.Entry<String,String[]> entry = null;
      String[] result = null;

              // for testing and tracking
              //System.out.println("in ACSFactoryImpl.getM_ComponentNodes(): ");

      for (Iterator ite = m_ComponentNodes_.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,String[]>)ite.next();

         l_values = entry.getValue();
         result = new String[l_values.length];
         System.arraycopy(l_values, 0, result, 0, l_values.length);
         m_result.put(entry.getKey(), result);

              // for testing and tracking
              //System.out.println("--- for the key "+entry.getKey());
              //for (int i = 0; i < result.length; i++) {
                //    System.out.println("--------- at "+i+": "+result[i]);
              //} // for i

      } // for ite

      l_values = null;
      entry = null;
      result = null;
      return m_result;
  } // getM_ComponentNodes()

  /**
   * This method returns the allowed values which is used as map key startings
   * for component node values.
   * Called by SimplePropertyEditor.getRestrictedValues() from the
   * ACSFactoryImplBeanInfo specification for the StringMapPropertyEditor,
   * by addComponentNode().
   * The starting formats (key) are a range of values listed hereafter: <P>
   * - 'IsChildOfNewAcs.', to set a new component under the new ACS, <P>
   * - 'IsChildOfNewAcsParent.', to set a new component under the parent of the new ACS. <P>
   * The keys are always ended by the components of the new component name,
   * and they may be more than one since it is possible to define a tree of
   * components from the new ACS or from its parents.
   * @since 0.7.0
   * @return List containing the allowed key startings
   */
  public List<String> getL_MapKeysStartsForComponents() {
    List<String> l_result = new ArrayList<String>(10);
    l_result.add("IsChildOfNewAcs.");
    l_result.add("IsChildOfNewAcsParent.");

    return l_result;
  } // getL_MapKeysStartsForComponents()
  
  /**
   * Gets the values which are allowed by the program for a component node.
   * This method is set in the BeanInfo, called
   * by StringTwoKeysMapPropertyEditor from a specification by ACSFactoryImplBeanInfo,
   * and is just a wraper for the ACSFactoryUtilityImpl method.
   * @since 0.7.0
   * @see ARoad0.gBase.ACSFactoryUtilityImpl#getL_LimitedValuesForTypesAndExplorer
   * @param _key for the constrained values
   * @return list of the allowed values. Never empty nor null.
   */
  public final List<String> getL_LimitedValuesForComponentNodes(String _key) {

      return ACSFactoryUtilityImpl.getL_LimitedValuesForComponentNodes(this, _key);
  } // getL_LimitedValuesForComponentNodes()
  
  /**
   * Adds a new composite node in the IS nametree.
   * The key is an unique composite name, and the property editor allows only
   * one value per key.
   * The value which may be 'acs', 'acs.authorization_server' and 'subacs' as
   * a composite type, and the property editor contols them. <P>
   * Fires PropertyChanges for 'CompositeNodes'.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#getL_LimitedValuesForComponentNodes
   * @param _criteria the map key.
   * @param _l_values String array which contains the allowed values
   * for the criteria. Never null. The lenght may be 0 or 1.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not allowed, <P>
   *    - a value is not allowed, or there is already a value.
   */
  public synchronized void addCompositeNode(String _criteria, String[] _l_values) 
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("Call of ACSFactoryImpl.addCompositeNode() ");;

    if (_criteria == null || _l_values == null) {
        throw new UpDateError(" Sorry, the key or the value is null.");
    }

    if (_l_values.length > 1) {
        throw new UpDateError(" Sorry, no more than one value.");
    }

    if (BaseManagerImpl.getInstance().getL_CompositeNodesOfIS(iSname_).contains(
            new NameImpl(_criteria)))
                throw new UpDateError(" The key is already known.");

    boolean error = false;
    int index = -1;
    for (int i = 0; i < _l_values.length; i++) {
        if (!_l_values[i].equals("acs") && !_l_values[i].equals("acs.authorization_server") &&
                !_l_values[i].equals("subacs")) {
            error = true;
            index = i;
            break;
        }
    } // for i
    if (error)
        throw new UpDateError(
            " The key "+_criteria+" does not comply with the value "+_l_values[index]);

    HashMap m_old =  null;
    if (m_CompositeNodes_ != null)
        m_old = (HashMap) m_CompositeNodes_.clone();
    else {
        m_CompositeNodes_ = new HashMap(INITIAL_CAPACITY_25);
    }

    if (m_CompositeNodes_.get(_criteria) != null &&
            ((String[])m_CompositeNodes_.get(_criteria)).length != 0) {
        throw new UpDateError(
            " This property needs exactly one value per key.");
    }
    else m_CompositeNodes_.put(_criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addCompositeNode(), "+
                      //      "fire the property change 'CompositeNodes' ");

    try {
        firePropertyChange("CompositeNodes", m_old, (HashMap) m_CompositeNodes_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addCompositeNode(), firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null;
  } // addCompositeNode()

  /**
   * Removes a new component node in the IS nametree.
   * Fires a PropertyChange for 'CompositeNodes'.
   * The criteria is not removed from the map if _l_values contains all the current
   * values which is associated to the criteria.
   * @since 0.7.0
   * @param _criteria for the relevant value.
   * @param _l_values array which contains only
   *   the value to remove for the criteria. Null to remove the criteria.
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is not registered.
   */
  public synchronized void removeCompositeNode(String _criteria, 
    String[] _l_values) throws UpDateError {

    if (_criteria == null ) {
        throw new UpDateError(" The key is not well formed.");
    }
    if (m_CompositeNodes_ == null) {
        throw new UpDateError(" The property is not well initialized.");
    }
    if (!m_CompositeNodes_.keySet().contains(_criteria)) {
        throw new UpDateError(" Unknown key.");
    }
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeCompositeNode(), "+
                      //      "values to remove for "+_criteria);

    HashMap m_old = (HashMap) m_CompositeNodes_.clone();
    if ( _l_values == null) {
        m_CompositeNodes_.remove(_criteria);

        try {
            firePropertyChange("CompositeNodes", m_old,
                    (HashMap) m_CompositeNodes_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeCompositeNode(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)

    boolean error = false;

    for (int i = 0; i < _l_values.length; i++) {
      if (_l_values[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_values[i]);
      }
    } // for i
    if (error) {
        throw new UpDateError(" A value in the argument is null.");
    }

    String[] l_current = m_CompositeNodes_.get(_criteria);
    List list = new ArrayList( 20);
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);
    }

    short count = 0;
    boolean found = false;

    for (int  i = 0 ; i < _l_values.length ; i++) {
        found = list.remove(_l_values[i]);

        if (!found) {
            error = true;
            break;
        } // if
        else count++;

    } // for i

    if (error) {
        m_old = null;
        l_current = null;
        list = null;
        throw new UpDateError(" A value to remove is not found.");
    }

    if (l_current.length != count) {
        m_CompositeNodes_.put(
                _criteria,
                (String[]) list.toArray(new String[list.size()]));
    } else { // if _l_values contains all the current values
        m_CompositeNodes_.put(
                _criteria, new String[0]);
    }

    try {
        firePropertyChange("CompositeNodes", m_old,
                (HashMap) m_CompositeNodes_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeCompositeNode(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
    l_current = null;
    list = null;
  } // removeCompositeNode()  
  
  /**
   * Gets the new composite nodes in the IS nametree.
   * The single value for each key is the type of the composite;
   * it may be 'acs', 'acs.authorization_server' and 'subacs'.
   * @since 0.7.0
   * @return a copied Map where the key is the composite name, and the value is
   * an array which contains none or one String. May be empty, but not null.
   * A value may be empty but is never null.
   */
  public final Map getM_CompositeNodes() {
      if (m_CompositeNodes_ == null) return new HashMap(INITIAL_CAPACITY_5);

      HashMap<String,String[]>  m_result = 
         new HashMap<String,String[]>(INITIAL_CAPACITY_25); 
      String[] l_values = null;
      Map.Entry<String,String[]> entry = null;
      String[] result = null;

              // for testing and tracking
              //System.out.println("in ACSFactoryImpl.getM_CompositeNodes(): ");
      
      for (Iterator ite = m_CompositeNodes_.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,String[]>)ite.next();

         l_values = entry.getValue();
         result = new String[l_values.length];
         System.arraycopy(l_values, 0, result, 0, l_values.length);  
         m_result.put(entry.getKey(), result);
         
              // for testing and tracking
              //System.out.println("--- for the key "+entry.getKey()); 
              //for (int i = 0; i < result.length; i++) {
                //    System.out.println("--------- at "+i+": "+result[i]);                  
              //} // for i
         
      } // for ite

      
      l_values = null;
      entry = null;
      result = null;     
      return m_result;
  } // getM_CompositeNodes()  
  
  /**
   * Gets the values which are allowed by the program for a composite node.
   * This method is set in the BeanInfo, called
   * by StringMapPropertyEditor from a specification by ACSFactoryImplBeanInfo,
   * and is just a wraper for the ACSFactoryUtilityImpl method.
   * @since 0.7.0
   * @param _key for the constrained values
   * @return list of the allowed values. Never empty nor null. 
   */
  public final List<String> getL_LimitedValuesForCompositeNodes(String _key) {

      return ACSFactoryUtilityImpl.getL_LimitedValuesForCompositeNodes();
  } // getL_LimitedValuesForCompositeNodes()
  
  /**
   * Adds an alternate of the ACS or the subACS, if any, with all the
   * associated properties.
   * The key starts with the ACS name last component or the subACS name, ended with a point,
   * and the key ends with the name of a current or new alternate in the same IS. <P>
   * The value is an array which may contain the following values, in this order: <P>
   * - exactly one value among the types in ISFactoryUtilityImpl.getAlternateTypes(), <P>
   * - one or several right types, among the 3 in ISFactoryUtilityImpl, <P>
   * - as options, the values OPTIONAL_ALTERNATE and OPTIONAL_GROUND in ISFactoryUtilityImpl. <P>
   * Then, the values array lenght is from 2 to 6.
   * Calls ACSFactoryUtilityImpl.controlAlternateValues().
   * Fires a Property Change for 'Alternates'.
   * @since 0.7.0
   * @param _criteria the map key.
   * @param _l_values String array which contains the allowed values
   * for the criteria. Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not allowed, <P>
   *    - a value is not allowed.
   */
  public synchronized void addAlternate(String _criteria, String[] _l_values)
          throws UpDateError {

    if (_criteria == null || _l_values == null) {
        throw new UpDateError(" The key or the values are null.");
    }
    if (_l_values.length > 6) {
        throw new UpDateError(" The values are limited to 6 items.");
    }

    boolean         error = true;
    boolean         found = false;
    List<String>    list = null;
    Map             m_old =  null;
    Name            alter = null;
    String          message = null;
    String          type = null;
    String          keyRoot = null;
    Name            trueKey = null;
    List<String>    l_new = null;

    
                                // for testing and tracking
                                //System.out.println("ACSFactoryImpl.addAlternate() - key: "+
                                  //      _criteria+", and _l_values: ");
                                //for (int i = 0; i < _l_values.length; i++) {
                                  //  System.out.println("--- "+_l_values[i]);
                                //}


    list = getL_MapKeysStartsForAlternates();

    for (String name : list) {
        if (_criteria.startsWith(name)) {
            error = false;
            break;
        }
    } // for name
    if (error)
        throw new UpDateError(" "+_criteria+" starting is not allowed.");

    // keyRoot is the ACS nick name or a subACS name
    keyRoot = _criteria.substring(0, _criteria.indexOf("."));

    try {

        // cuts the leading and trailing whitespaces, for which the number may vary,
        // then adds the right whitespaces to build up the String format of a name
        alter = NameImpl.getNameFromString( " "+_criteria.substring(
                _criteria.indexOf(".") + 1 ).trim() + "  ");
    
                                // for testing and tracking
                                //System.out.println("ACSFactoryImpl.addAlternate() - alter: "+
                                  //      alter);

        // the image of _criteria in m_alternateProperties_
        trueKey = (new NameImpl( keyRoot )).extended(alter);

        if (m_CompositeNodes_ != null) {
            for (String compo : m_CompositeNodes_.keySet()) {
                if (alter.equals( ((NameImpl)iSname_.clone()).extended(compo) )) {
    
                                // for testing and tracking
                                //System.out.println("ACSFactoryImpl.addAlternate() - "+
                                 //       "matched composite node: "+compo);

                    found = true;
                    break;
                }
            } // for compo
        } // if (m_CompositeNodes_

        if (!found && !BaseManagerImpl.getInstance().getISTree(iSname_).isOpen(alter))
            throw new UpDateError(" The key does not point to an open node in the IS.");

        if (!found) {
            type = BaseManagerImpl.getInstance().getISTree(iSname_).getType(alter);

            if (!type.equals(ISFactoryUtilityImpl.ACS_NODE) &&
                        !type.equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) &&
                            _criteria.startsWith(getNickName()) )
                throw new UpDateError(" The new ACS cannot have as alternate this subkey, that "+
                        "it is not an ACS or a composite ACS node, in the information system.");

            if (!type.equals(ISFactoryUtilityImpl.SUB_ACS_NODE) &&
                        !type.equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE) &&
                            !_criteria.startsWith(getNickName()) )
                throw new UpDateError(" The new subACS cannot have as alternate this subkey, that "+
                        "it is not a subACS node in the information system.");
            
            type = null;
        } // if (!found
        else { // the alternate is a new composite
            if (((String[])m_CompositeNodes_.get(alter.lastComponent()))[0].equals("subacs") &&
                            _criteria.startsWith(getNickName()) )
                throw new UpDateError(" The new ACS cannot have as alternate this subkey, that "+
                        "it is a composite subACS node in the information system.");

            if (!((String[])m_CompositeNodes_.get(alter.lastComponent()))[0].equals("subacs") &&
                            !_criteria.startsWith(getNickName()) )
                throw new UpDateError(" The new subACS cannot have as alternate this subkey, that "+
                        "it is not a composite subACS node in the information system.");
        } // else
        
    } catch (BaseError b) {
        throw new UpDateError(" The key does not end with a true name. "+b.toString());
    }

    if (m_alternateProperties_ != null)
        m_old = getM_Alternates();
    else {
        m_alternateProperties_ = new HashMap(INITIAL_CAPACITY_5);
    }

    if (m_alternateProperties_.get(trueKey) != null) {
        String[] l_old = m_alternateProperties_.get(trueKey);
        if (l_old.length != 0) {
            // l_new remains null if there is no value for trueKey
            l_new = new ArrayList(INITIAL_CAPACITY_5);

            for (int i = 0; i < l_old.length; i++) {
                l_new.add(l_old[i]);
            } // for i

            if (_l_values.length > 0) {
                for (int i = 0; i < _l_values.length; i++) {
                   l_new.add(_l_values[i]);
                }
            } // if

            if (l_new.size() > 6)
                throw new UpDateError(
                    " The length for the new values array is more than 6.");
        } // if
        l_old = null;
    } // if

    if (l_new == null ) {
        message = ACSFactoryUtilityImpl.controlAlternateValues(_l_values);
        if (message != null) {
            alter = null;
            keyRoot = null;
            trueKey = null;

                                // for testing and tracking
                                //System.out.println("From ACSFactoryImpl.addAlternate(), "+
                                  //      "ACSFactoryUtilityImpl.controlAlternateValues() fires "+
                                    //    "the message "+message);

            throw new UpDateError(message);
        } // if
    } // if m_alternateProperties_.get(trueKey)
    else {
        message = ACSFactoryUtilityImpl.controlAlternateValues(
                (String[])l_new.toArray(new String[0]) );
        if (message != null) {
            alter = null;
            keyRoot = null;
            trueKey = null;
            l_new = null;

                                // for testing and tracking
                                //System.out.println("From ACSFactoryImpl.addAlternate(), "+
                                  //      "ACSFactoryUtilityImpl.controlAlternateValues() fires "+
                                    //    "the message "+message);
                                
            throw new UpDateError(message);
        } // if
    } // else if l_new == null

                // test SWITCHING_ALTERNATE unicity for each ground and right type
    if (_l_values.length > 0 && _l_values[0].equals(ISFactoryUtilityImpl.SWITCHING_ALTERNATE)) {

                                // for testing and tracking
                                //System.out.println("From ACSFactoryImpl.addAlternate(), "+
                                  //      "start of SWITCHING_ALTERNATE unicity control. ");

        for (Map.Entry entry : m_alternateProperties_.entrySet()) {
            if (!((ImmutableName)entry.getKey()).component(0).equals(keyRoot))
                        continue;
            if (!Arrays.asList((String[]) entry.getValue()).contains(
                    ISFactoryUtilityImpl.SWITCHING_ALTERNATE))
                        continue;

                                // for testing and tracking
                                //System.out.println("From ACSFactoryImpl.addAlternate(), "+
                                  //      "SWITCHING_ALTERNATE unicity control, step 1. ");

            for (int i = 1; i < _l_values.length; i++) {
                if (!ISFactoryUtilityImpl.getRightTypesForAlternate().contains(
                        _l_values[i]) ) continue;


                            // for testing and tracking
                            //System.out.println("From ACSFactoryImpl.addAlternate(), "+
                              //      "SWITCHING_ALTERNATE unicity control, step 2. ");

                if (Arrays.asList((String[]) entry.getValue()).
                        contains(_l_values[i])) {
                            message = ((ImmutableName)entry.getKey()).lastComponent()+
                                        " is already a switching alternate for "+
                                        _l_values[i];
                            break;
                } // if
            } // for i
            if (message != null) break;
        } // for entry
    } // if
    if (message != null) {
        alter = null;
        keyRoot = null;
        trueKey = null;
        l_new = null;
        throw new UpDateError(message);
    } // if

    if (m_alternateProperties_.get(trueKey) != null) {
        if (l_new != null) {
            m_alternateProperties_.put(
                    trueKey,
                    (String[])l_new.toArray(new String[0]));
        }
        else m_alternateProperties_.put(
                trueKey,
                _l_values);
    }
    else m_alternateProperties_.put(
            trueKey,
            _l_values);

                                    // for testing and tracking
                                    //System.out.println("From ACSFactoryImpl.addAlternate(), "+
                                      //      "fire the property change 'Alternates' ");

    try {
        firePropertyChange("Alternates", m_old, getM_Alternates());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addAlternate(), firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null;
    list = null;
    alter = null;
    keyRoot = null;
    trueKey = null;
    l_new = null;
  } // addAlternate()

  /**
   * Removes an alternate of the ACS or the subACS, if any, with all the
   * associated properties.
   * The key starts with the ACS name last component or the subACS name,
   * and the key ends with the name of the alternate in the same IS. <P>
   * The value is an array which may contain the following values, in this order: <P>
   * - exactly one value among the types in ISFactoryUtilityImpl.getAlternateTypes(), <P>
   * - one or several right types, among the 3 in ISFactoryUtilityImpl, <P>
   * - as options, the values OPTIONAL_ALTERNATE and OPTIONAL_GROUND in ISFactoryUtilityImpl. <P>
   * Then, the values array lenght is from 2 to 6.
   * Fires a PropertyChange for 'Alternates'.
   * The criteria is not removed from the map if _l_values contains all the current
   * values which is associated to the criteria.
   * All the values are removed if _l_values contains the alternate type, in other
   * words, the mandatory first value of the current array.
   * @since 0.7.0
   * @param _criteria for the relevant value.
   * @param _l_values array which contains only
   *   the value to remove for the criteria. Null to remove the criteria.
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is not registered.
   */
  public synchronized void removeAlternate(String _criteria,
    String[] _l_values) throws UpDateError {

    if (_criteria == null ) {
        throw new UpDateError(" The key or the value is not well formed.");
    }
    if (m_alternateProperties_ == null) {
        throw new UpDateError(" The property is not well initialized.");
    }
    if (!getM_Alternates().keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeAlternate(), "+
                      //      "values to remove for "+_criteria);

    Map             m_old = getM_Alternates();
    Name            alter = null;
    Name            trueKey = null;
    String          keyRoot = null;
    boolean         error = false;
    List<String>    list = null;
    String[]        l_current = null;
    String          message = null;
    short           count = 0;
    boolean         found = false;


    // keyRoot is the ACS nick name or a subACS name
    keyRoot = _criteria.substring(0, _criteria.indexOf("."));

    try {

        // cuts the leading and trailing whitespaces, for which the number may vary,
        // then adds the right whitespaces to build up the String format of a name
        alter = NameImpl.getNameFromString( " "+_criteria.substring(
                _criteria.indexOf(".") + 1 ).trim() + "  ");

        // the image of _criteria in m_alternateProperties_
        trueKey = (new NameImpl( keyRoot )).extended(alter);

    } catch (BaseError b) {
        throw new UpDateError(" The key does not end with a true name.");
    }

    // remove the key if the alternate type is to remove
    if ( _l_values == null) {

        m_alternateProperties_.remove(trueKey);

        try {
            firePropertyChange("Alternates", m_old, getM_Alternates());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeAlternate(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        alter = null;
        list = null;
        keyRoot = null;
        trueKey = null;
        return;
    } // if ( _l_values == null)

    l_current = m_alternateProperties_.get(trueKey);
    list = new ArrayList( 20 );
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);
    }

    for (int  i = 0 ; i < _l_values.length ; i++) {
        found = list.remove(_l_values[i]);

        if (!found) {
            error = true;
            break;
        } // if
        else count++;

    } // for i

    if (error) {
        m_old = null;
        alter = null;
        l_current = null;
        list = null;
        keyRoot = null;
        trueKey = null;
        throw new UpDateError(" A value to remove is not found.");
    }

    if (!list.isEmpty())
        message = ACSFactoryUtilityImpl.controlAlternateValues(
            (String[])list.toArray(new String[0]) );
    if (message != null) {
        m_old = null;
        alter = null;
        keyRoot = null;
        trueKey = null;
        list = null;
        throw new UpDateError(message);
    } // if

    if (l_current.length != count) {
        m_alternateProperties_.put(
                trueKey,
                (String[]) list.toArray(new String[list.size()]));
    } else { // if _l_values contains all the current values
        m_alternateProperties_.put(
                trueKey,
                new String[0]);
    } // else

    try {
        firePropertyChange("Alternates", m_old, getM_Alternates());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeAlternates(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
    alter = null;
    l_current = null;
    list = null;
    keyRoot = null;
    trueKey = null;
  } // removeAlternate()

  /**
   * Gets the alternates of the ACS or the subACS, if any, with all the
   * associated properties.
   * The key starts with the ACS name last component or the subACS name,
   * and the key, after a '.' separator,
   * ends with the toString() form of the alternate name. <P>
   * The value is an array which may contain the following values: <P>
   * - exactly one value among the types in ISFactoryUtilityImpl.getAlternateTypes(), <P>
   * - one or several right types, among the 3 in ISFactoryUtilityImpl, <P>
   * - as options, the values OPTIONAL_ALTERNATE and OPTIONAL_GROUND in ISFactoryUtilityImpl. <P>
   * The values array lenght is from 2 to 6.
   * @since 0.7.0
   * @return the map of the alternate relations. Never null and may be empty.
   */
  public Map<String, String[]> getM_Alternates() {
      Name name = null;
      ImmutableName alter = null;

      if (m_alternateProperties_ != null) {
          HashMap<String, String[]> l_result = new HashMap(INITIAL_CAPACITY_5);

          for (Map.Entry entry : m_alternateProperties_.entrySet()) {
              name = (Name)entry.getKey();
              alter = name.subname(1, name.noOfComponents());

              l_result.put(
                      name.component(0)+"."+alter.toString(),
                      (String[])entry.getValue());
          } // for entry
          return l_result;
      }
      else return new HashMap(INITIAL_CAPACITY_5);
  } // getM_Alternates()

  /**
   * Gets the Alternate key startings with the ACS name last component or the subACS names,
   * all ended with a point.
   * @since 0.7.0
   * @return List of the allowed starting values
   */
  public List<String> getL_MapKeysStartsForAlternates() {

    List<String> l_result = new ArrayList<String>(10);
    l_result.add(getNickName()+".");
    for (String subacs : getM_SubAndGroupPolicy().keySet()) {
        l_result.add(subacs + ".");
    }
    return l_result;
  } // getL_MapKeysStartsForAlternates()

  /**
   * This method defines the alternate types, the right types for alternates and 
   * the options. Call ISFactoryUtilityImpl.getAlternateTypes() and
   * getRightTypesAndOptionsForAlternate(). Only one alternate type
   * should be chosen by the user, and there is no other constraint.
   * Called by StringTwoKeysMapPropertyEditor,
   * from the value returned by ACSFactoryImplBeanInfo.
   * @since 0.7.0
   * @param _key is unused, and it is here simply to be compliant with the StringMapPropertyEditor
   *    handling
   * @return list of the allowed values. Never empty nor null.
   */
  public List<String> getL_LimitedValuesForAlternates(String _key) {
      return ACSFactoryUtilityImpl.getL_LimitedValuesForAlternates();
  } // getL_LimitedValuesForAlternates()

  /**
   * Adds associations to create opening and closing roots in the parent ACSs. <P>
   * This relation sets that the parent ACS ACSObject, named root,
   * or one of its property value,
   * has to be there when the ACS is opened, to ensure the ACS ability to run,
   * and has to be closed or updated (for a property value) at the closing of the ACS. <P>
   * The criteria is a map key with one allowed format ('ACS.Name.&lt;Value&gt;.xxxx'),
   * due to the control by AssociationMapPropertyEditor and ACSFactoryImplBeanInfo.
   * The allowed String map values does not depend on the map key.
   * The non-empty values define fully or partially one action to do if the key matches.
   * One action is always a constructor or a method to call. <P>
   * Calls addAssociations() for the main work.
   * The format of the keys and values is described at AssociateUtilityImpl.controlAssociations().
   * Fires PropertyChanges for 'OpeningAndClosingRoots'.
   * @since 0.7.0
   * @see #addAssociations
   * @see ARoad0.Gui1.AssociationMapPropertyEditor
   * @param _criteria is the map key to apply these associations.
   *    It must be the parent ACS name.
   * @param _l_values a Strings array which contain the values for the criteria,
   *    but without the action number 'N.'. Values are added to the previous values, if any.
   *    Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - a value is not allowed.
   */
  public synchronized void addOpeningAndClosingRoots(String _criteria, String[] _l_values)
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("ACSFactoryImpl.addOpeningAndClosingRoots() called "+
                      //      "with criteria "+_criteria);

    if (_criteria == null || _l_values == null ) {
      throw new UpDateError(" The key or the value is null.");
    }
    try {
        ImmutableName name = NameImpl.getNameFromString(_criteria.substring(17));
        if (!BaseManagerImpl.getInstance().isOpenACS(name))
            throw new UpDateError(" The key does not end with the name of an open ACS.");
        name = null;
    }
    catch (BaseError b) {
        throw new UpDateError(" The key does not end with a name.");
    }

    HashMap         m_old = null;

    if (m_RootsForAcsOpeningAndClosing_ != null)
        m_old = (HashMap) m_RootsForAcsOpeningAndClosing_.clone();
    else {
        m_RootsForAcsOpeningAndClosing_ = new HashMap(INITIAL_CAPACITY_25);
    }

    m_RootsForAcsOpeningAndClosing_ = addAssociations(
          m_RootsForAcsOpeningAndClosing_, _criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addOpeningAndClosingRoots(), "+
                      //      "fire the property change 'OpeningAndClosingRoots' ");

    try {
        firePropertyChange("OpeningAndClosingRoots", m_old,
                (HashMap) m_RootsForAcsOpeningAndClosing_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addOpeningAndClosingRoots(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null; 
  } // addOpeningAndClosingRoots()

  /**
   * Removes associations to create opening and closing roots in the parent ACSs. <P>
   * The format is described at AssociateUtilityImpl.controlAssociations().
   * Fires a PropertyChange for 'OpeningAndClosingRoots'.
   * The criteria is not removed from the map if _l_values contains all the current values
   * which is associated to the criteria.
   * After the removing, the set of values is controlled and an error message is
   * added if the set is not compliant. A previous error message may be removed
   * ('Error:' or 'Error to delete:'), and also the action numbers.
   * Call removeAssociations() for the main work.
   * @since 0.7.0
   * @param _criteria is the map key to apply these associations
   * @param _l_values a Strings array which contain the values for the criteria,
   *    with an action number 'N.' or not. Null to remove the criteria. May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - _l_values is null or one value is null <P>
   *    - a value is unknown <P>
   *    - items in _l_lvalues are not in a sequence in the current values array.
   */
  public synchronized void removeOpeningAndClosingRoots(String _criteria,
    String[] _l_values) throws UpDateError {

    if (_criteria == null) {
        throw new UpDateError(" The key is not well formed.");
    }
    if (m_RootsForAcsOpeningAndClosing_ == null) {
        throw new UpDateError(" The property is not well initialized.");
    }
    if (!m_RootsForAcsOpeningAndClosing_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeOpeningAndClosingRoots(), "+
                      //      "values to remove for "+_criteria);

    HashMap m_old = (HashMap) m_RootsForAcsOpeningAndClosing_.clone();

    if ( _l_values == null) {
        m_RootsForAcsOpeningAndClosing_.remove(_criteria);

        try {
            firePropertyChange("OpeningAndClosingRoots", m_old,
                    (HashMap) m_RootsForAcsOpeningAndClosing_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");;
            System.out.println("From ACSFactoryImpl.removeOpeningAndClosingRoots(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)

    m_RootsForAcsOpeningAndClosing_ = removeAssociations(
          m_RootsForAcsOpeningAndClosing_, _criteria, _l_values);

    try {
        firePropertyChange("OpeningAndClosingRoots", m_old,
                (HashMap) m_RootsForAcsOpeningAndClosing_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeOpeningAndClosingRoots(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
  } // removeOpeningAndClosingRoots()

  /**
   * Gets associations to create opening and closing roots in the parent ACSs,
   * by the mean of actions to do by the program. Each action has a number from 1 to 99.
   * The criteria is a map key with one allowed format ('ACS.Name.&lt;Value&gt;.xxxx').
   * The values format is fully described at AssociateUtilityImpl.controlAssociations().
   * The internal map may contain values where the first 2 characters are not 'N.', but
   * the first no-number value and its followings are returned by this method which is used by
   * ActionMapPropertyEditor to edit the internal map.
   * Call AssociateUtilityImpl.getM_FilteredAssociations() directly.
   * @since 0.7.0
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings started by a number between 1 to 9 followed by a point.
   *   May be empty, but not null.
   */
  public final Map getM_OpeningAndClosingRoots() {
      return AssociateUtilityImpl.getM_FilteredAssociations(
              m_RootsForAcsOpeningAndClosing_, false);
  } // getM_OpeningAndClosingRoots()

  /**
   * Adds associations to create opening without closing roots in the parent ACSs. <P>
   * This relation sets that the parent ACS ACSObject, named root,
   * or one of its property value,
   * has to be there when the ACS is opened, to ensure the ACS ability to run,
   * but has not to be closed or updated (for a property value) at the closing 
   * of the ACS nor at the deleting of the ACS. <P>
   * The criteria is a map key with one allowed format ('ACS.Name.&lt;Value&gt;.xxxx'),
   * due to the control by AssociationMapPropertyEditor and ACSFactoryImplBeanInfo.
   * The allowed String map values does not depend on the map key.
   * The non-empty values define fully or partially one action to do if the key matches.
   * One action is always a constructor or a method to call. <P>
   * Calls addAssociations(). If this method detects an error, a message starting with
   * 'Error ' is added at the end of the m_RootsForAcsOpeningNoClosing_ map for
   * the key _criteria.
   * The format of the keys and values is described at
   * AssociateUtilityImpl.controlAssociations().
   * Fires PropertyChanges for 'OpeningNoClosingRoots'.
   * @since 0.7.0
   * @see #addAssociations
   * @see ARoad0.Gui1.AssociationMapPropertyEditor
   * @param _criteria is the map key to apply these associations. 
   *    It must be the parent ACS name.
   * @param _l_values a Strings array which contain the values for the criteria,
   *    but without the action number 'N.'. Values are added to the previous values, if any.
   *    Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - a value is not allowed.
   */
  public synchronized void addOpeningNoClosingRoots(String _criteria, String[] _l_values)
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("ACSFactoryImpl.addOpeningNoClosingRoots() called "+
                      //      "with criteria "+_criteria);

    if (_criteria == null || _l_values == null ) {
        throw new UpDateError(" The key or the value is null.");
    }
    try {
        ImmutableName name = NameImpl.getNameFromString(_criteria.substring(17));
        if (!BaseManagerImpl.getInstance().isOpenACS(name))
            throw new UpDateError(" The key does not end with the name of an open ACS.");
        name = null;
    }
    catch (BaseError b) {
        throw new UpDateError(" The key does not end with a name.");
    }

    HashMap         m_old = null;

    if (m_RootsForAcsOpeningNoClosing_ != null)
        m_old = (HashMap) m_RootsForAcsOpeningNoClosing_.clone();
    else {
        m_RootsForAcsOpeningNoClosing_ = new HashMap(INITIAL_CAPACITY_25);
    }

    m_RootsForAcsOpeningNoClosing_ = addAssociations(
          m_RootsForAcsOpeningNoClosing_, _criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addOpeningNoClosingRoots(), "+
                      //      "fire the property change 'OpeningNoClosingRoots' ");

    try {
        firePropertyChange("OpeningNoClosingRoots", m_old,
                (HashMap) m_RootsForAcsOpeningNoClosing_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addOpeningNoClosingRoots(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null; 
  } // addOpeningNoClosingRoots()

  /**
   * Removes associations to create opening without closing roots in the parent ACSs. <P>
   * The format is described at AssociateUtilityImpl.controlAssociations().
   * Fires a PropertyChange for 'OpeningNoClosingRoots'.
   * The criteria is not removed from the map if _l_values contains all the current values
   * which is associated to the criteria.
   * After the removing, the set of values is controlled and an error message is
   * added if the set is not compliant. A previous error message may be removed
   * ('Error:' or 'Error to delete:'), and also the action numbers.
   * Calls removeAssociations() for the main work.
   * @since 0.7.0
   * @param _criteria is the map key to apply these associations
   * @param _l_values a Strings array which contain the values for the criteria,
   *    with an action number 'N.' or not. Null to remove the criteria. May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - _l_values is null or one value is null <P>
   *    - a value is unknown <P>
   *    - items in _l_lvalues are not in a sequence in the current values array.
   */
  public synchronized void removeOpeningNoClosingRoots(String _criteria,
    String[] _l_values) throws UpDateError {

    if (_criteria == null) {
        throw new UpDateError(" The key is not well formed.");
    }
    if (m_RootsForAcsOpeningNoClosing_ == null) {
        throw new UpDateError(" The property is not well initialized.");
    }
    if (!m_RootsForAcsOpeningNoClosing_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeOpeningNoClosingRoots(), "+
                      //      "values to remove for "+_criteria);

    HashMap             m_old = (HashMap) m_RootsForAcsOpeningNoClosing_.clone();

    if ( _l_values == null) {
        m_RootsForAcsOpeningNoClosing_.remove(_criteria);

        try {
            firePropertyChange("OpeningNoClosingRoots", m_old,
                    (HashMap) m_RootsForAcsOpeningNoClosing_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeOpeningNoClosingRoots(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)

    m_RootsForAcsOpeningNoClosing_ = removeAssociations(
          m_RootsForAcsOpeningNoClosing_, _criteria, _l_values);

    try {
        firePropertyChange("OpeningNoClosingRoots", m_old,
                (HashMap) m_RootsForAcsOpeningNoClosing_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeOpeningNoClosingRoots(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
  } // removeOpeningNoClosingRoots()

  /**
   * Gets associations to create opening without closing roots in the parent ACSs,
   * by the mean of actions to do by the program. Each action has a number from 1 to 99.
   * The criteria is a map key with one allowed format ('ACS.Name.&lt;Value&gt;.xxxx').
   * The values format is fully described at AssociateUtilityImpl.controlAssociations().
   * The internal map may contain values where the first 2 characters are not 'N.', but
   * the first no-number value and its followings are returned by this method which is used by
   * ActionMapPropertyEditor to edit the internal map.
   * Calls AssociateUtilityImpl.getM_FilteredAssociations() directly.
   * @since 0.7.0
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings started by a number between 1 to 9 followed by a point.
   *   May be empty, but not null.
   */
  public final Map getM_OpeningNoClosingRoots() {
      return AssociateUtilityImpl.getM_FilteredAssociations(
              m_RootsForAcsOpeningNoClosing_, false);
  } // getM_OpeningNoClosingRoots()

  /**
   * Adds associations to set integrity root contraints in the parent ACSs. <P>
   * This relation sets that the parent ACS ACSObject, named root, 
   * or one of its property value,
   * has to be properly stated, to ensure the ability of the ACS to run. <P>
   * The criteria is a map key with one allowed format ('ACS.Name.&lt;Value&gt;.xxxx'),
   * due to the control by AssociationMapPropertyEditor and ACSFactoryImplBeanInfo.
   * The allowed String map values does not depend on the map key.
   * The non-empty values define fully or partially one action to do if the key matches.
   * One action is always a constructor or a method to call. <P>
   * This method provides an assistance to the user to build up a valid set
   * of integrity controls with the proper commands and arguments.
   * See addAssociations() for understanding the complex logic. <P>
   * Calls addAssociations().
   * The format of the keys and values is described at AssociateUtilityImpl.controlAssociations().
   * Fire PropertyChanges for 'IntegrityRoots'.
   * @since 0.7.0
   * @see #addAssociations
   * @see ARoad0.Gui1.AssociationMapPropertyEditor
   * @param _criteria is the map key to apply these associations. It must be an ACS name.
   * @param _l_values a Strings array which contain the values for the criteria,
   *    but without the action number 'N.'. Values are added to the previous values, if any.
   *    Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - a value is not allowed.
   */
  public synchronized void addIntegrityRoots(String _criteria, String[] _l_values)
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("ACSFactoryImpl.addIntegrityRoots() called "+
                      //      "with criteria "+_criteria);

    if (_criteria == null || _l_values == null ) {
      throw new UpDateError(" The key or the value is null.");
    }
    try {
        ImmutableName name = NameImpl.getNameFromString(_criteria.substring(17));
        if (!BaseManagerImpl.getInstance().isOpenACS(name))
            throw new UpDateError(" The key does not end with the name of an open ACS.");
        name = null;
    }
    catch (BaseError b) {
        throw new UpDateError(" The key does not end with a name.");
    }

    HashMap         m_old = null;

    if (m_RootsForAcsIntegrity_ != null)
        m_old = (HashMap) m_RootsForAcsIntegrity_.clone();
    else {
        m_RootsForAcsIntegrity_ = new HashMap(INITIAL_CAPACITY_25);
    }

    m_RootsForAcsIntegrity_ = addAssociations(
          m_RootsForAcsIntegrity_, _criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addIntegrityRoots(), "+
                      //      "fire the property change 'IntegrityRoots' ");

    try {
        firePropertyChange("IntegrityRoots", m_old,
                (HashMap) m_RootsForAcsIntegrity_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addIntegrityRoots(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null; 
  } // addIntegrityRoots()

  /**
   * Removes associations to set integrity root contraints in the parent ACSs. <P>
   * The format is described at AssociateUtilityImpl.controlAssociations().
   * Fires a PropertyChange for 'IntegrityRoots'.
   * The criteria is not removed from the map if _l_values contains all the current values
   * which is associated to the criteria.
   * After the removing, the set of values is controlled and an error message is
   * added if the set is not compliant. A previous error message may be removed
   * ('Error:' or 'Error to delete:'), and also the action numbers.
   * Calls removeAssociations() for the main work.
   * @since 0.7.0
   * @param _criteria is the map key to apply these associations
   * @param _l_values a Strings array which contain the values for the criteria,
   *    with an action number 'N.' or not. Null to remove the criteria. May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - _l_values is null or one value is null <P>
   *    - a value is unknown <P>
   *    - items in _l_lvalues are not in a sequence in the current values array.
   */
  public synchronized void removeIntegrityRoots(String _criteria,
    String[] _l_values) throws UpDateError {

    if (_criteria == null) {
        throw new UpDateError(" The key is not well formed.");
    }
    if (m_RootsForAcsIntegrity_ == null) {
        throw new UpDateError(" The property is not well initialized.");
    }
    if (!m_RootsForAcsIntegrity_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeIntegrityRoots(), "+
                      //      "values to remove for "+_criteria);

    HashMap             m_old = (HashMap) m_RootsForAcsIntegrity_.clone();

    if ( _l_values == null) {
        m_RootsForAcsIntegrity_.remove(_criteria);

        try {
            firePropertyChange("IntegrityRoots", m_old,
                    (HashMap) m_RootsForAcsIntegrity_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeIntegrityRoots(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)

    m_RootsForAcsIntegrity_ = removeAssociations(
          m_RootsForAcsIntegrity_, _criteria, _l_values);

    try {
        firePropertyChange("IntegrityRoots", m_old,
                (HashMap) m_RootsForAcsIntegrity_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeIntegrityRoots(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
  } // removeIntegrityRoots()

  /**
   * Gets associations to set integrity root contraints in the parent ACSs.
   * Each action has a number from 1 to 99.
   * The criteria is a map key with one allowed format ('ACS.Name.&lt;Value&gt;.xxxx').
   * The values format is fully described at AssociateUtilityImpl.controlAssociations().
   * The internal map may contain values where the first 2 characters are not 'N.', but
   * the first no-number value and its followings are returned by this method 
   * which is used by ActionMapPropertyEditor to edit the internal map.
   * Calls AssociateUtilityImpl.getM_FilteredAssociations() directly.
   * @since 0.7.0
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings started by a number between 1 to 9 followed by a point.
   *   May be empty, but not null.
   */
  public final Map getM_IntegrityRoots() {
      return AssociateUtilityImpl.getM_FilteredAssociations(
              m_RootsForAcsIntegrity_, false);
  } // getM_IntegrityRoots()

  /**
   * @since 0.7.0
   * @return the file name of the icon associated to the ACS. May be null.
   */
  public String getIcon() {
    return icon_;
  }

  /**
   * @since 0.7.0
   * @param _icon file name of the icon associated to the ACS. May be null.
   */
  public void setEorIcon(ImageIcon _icon) {
    imageIcon_ = _icon;
  }


  /**
   * Sets the ACS icon. The icon size is limited by ICON_HEIGHT_MAX and ICON_WIDTH_MAX.
   * The image has to be in the folder 'ARoad0.Gui1.images' or, for an AcsAddon,
   * in 'ARoad0.AcsAddon.Editor_name.AcsAddon_name.gBase'.
   * Fires 'icon'.
   * @since 0.7.0
   * @param _icon file name (linux.png for instance) of the icon associated to the ACS.
   * May be null, and with leading and trailing whitespaces.
   * @exception UpDateError if _icon is not the name of a gif, png or jpeg image file
   * in the folder
   */
  public void setIcon(String _icon) throws UpDateError {
    String old = icon_;
    Icon icon = null;
    try {

        if (getClass().toString().indexOf("ARoad0.AcsAddon.") == -1) {
            icon = new ImageIcon(
                    Desktop.AR_URI.resolve("Gui1"+Desktop.URL_SEPAR+"images"+
                        Desktop.URL_SEPAR+_icon).toURL() );
        }
        else {

            String editor = ((String[])m_AcsVocabulary_.get(
                    ACSFactoryUtilityImpl.ACSADDON_PACKAGE_EDITOR))[0];
            String addon = ((String[])m_AcsVocabulary_.get(
                    ACSFactoryUtilityImpl.ACSADDON_NAME))[0];

            icon = new ImageIcon(
                    Desktop.AR_URI.resolve("AcsAddon"+Desktop.URL_SEPAR+editor+
                        Desktop.URL_SEPAR+addon+Desktop.URL_SEPAR+"gBase"+
                        Desktop.URL_SEPAR+_icon).toURL() );
        } // else

    }
    catch (MalformedURLException m) {
        // if the AcsAddon has no proper image
        if (getClass().toString().indexOf("ARoad0.AcsAddon.") != -1) {
            try {
                icon = new ImageIcon(
                    Desktop.AR_URI.resolve("Gui1"+Desktop.URL_SEPAR+"images"+
                        Desktop.URL_SEPAR+_icon).toURL() );
            } catch (MalformedURLException p) {
                throw new UpDateError(" The image"+
                        " in 'Gui1/images' is not found: "+p.toString());
            }
        } // if
        else
            throw new UpDateError(" The image is not found: "+
              m.toString());
    } // catch
    catch (IllegalArgumentException i) {
        throw new UpDateError(" Wrong argument. The correct form is 'xxxx.jpg',"+
                " 'yyyyyyyy.png' or 'zz.gif'. ");
    }

    if (icon.getIconHeight() <= 0)
        throw new UpDateError(" The image has a null height ");

    if (icon.getIconHeight() > ICON_HEIGHT_MAX) {
        throw new UpDateError(" The image height "+
                "is larger than "+ICON_HEIGHT_MAX);
    }
    
    if (icon.getIconWidth() > ICON_WIDTH_MAX) {
        throw new UpDateError(" The image width "+
                "is larger than "+ICON_WIDTH_MAX);
    }
    
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.setIcon(), "+
                      //      "new icon height: "+icon.getIconHeight());

    icon_ = _icon;
    try {
        firePropertyChange("Icon", old, icon_);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeIntegrityRoots(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch
  } // setIcon()

  /**
   * Adds some comments to explain how this ACS simulates a real system/software.
   * The criteria is an ACS property displaying name. The value is a String array
   * of free user comments on the settings of the related ACS property. <P>
   * Fires Property Changes for 'CommentsOnProperties'.
   * @since 0.7.0
   * @param _criteria should be is an ACS property displaying name.
   * @param _l_values a Strings array of free user comments.
   *   Never null but the lenght may be 0.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not allowed
   */
  public synchronized void addCommentsOnProperties(String _criteria, String[] _l_values)
          throws UpDateError {
      if (m_CommentsOnProperties_ == null)
          m_CommentsOnProperties_ = new HashMap(INITIAL_CAPACITY_25);

                    // for testing and tracking
                    //System.out.println("Call of ACSFactoryImpl.addCommentsOnProperties() ");;

    if (_criteria == null || _l_values == null ) {
        throw new UpDateError(" The key or the values is not well formed.");
    }

    List<String> list = getL_MapKeysForCommentsOnProperties();
    if (!list.contains(_criteria))
        throw new UpDateError(" "+_criteria+" is not allowed as key.");

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addCommentsOnProperties(), "+
                      //      "the rights are controlled OK ");

    HashMap m_old =  null;
    m_old = (HashMap) m_CommentsOnProperties_.clone();


    if (m_CommentsOnProperties_.get(_criteria) != null) {
        String[] l_old = m_CommentsOnProperties_.get(_criteria);
        if (l_old.length != 0) {
            Set<String> l_new = new HashSet(INITIAL_CAPACITY_25);
            for (int i = 0; i < l_old.length; i++) {
                l_new.add(l_old[i]);
            } // for i
            l_new.addAll(Arrays.asList(_l_values) );
            m_CommentsOnProperties_.put(
                    _criteria, (String[])l_new.toArray(new String[0]));
            l_new = null;
        }
        else m_CommentsOnProperties_.put(_criteria, _l_values);
        l_old = null;
    }
    else m_CommentsOnProperties_.put(_criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addCommentsOnProperties(), "+
                      //      "fire the property change 'CommentsOnProperties' ");

    try {
        firePropertyChange("CommentsOnProperties", m_old,
                (HashMap) m_CommentsOnProperties_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addCommentsOnProperties(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null;
    list = null;
  } // addCommentsOnProperties()

  /**
   * Removes some comments to explain how this ACS simulates a real system/software.
   * The criteria is an ACS property name. The value is a String array
   * of free user comments on the settings of the related ACS property. <P>
   * Fires a PropertyChange for 'CommentsOnProperties'.
   * The criteria is not removed from the map if _l_values contains all the current
   * values which is associated to the criteria.
   * @since 0.7.0
   * @param _criteria for the relevant comments.
   * @param _l_values Strings array which contains only
   *   the specialized values to remove for the criteria. Null to remove the criteria.
   *   May be empty.
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is unknown.
   */
  public synchronized void removeCommentsOnProperties(String _criteria,
    String[] _l_values) throws UpDateError {

    if (_criteria == null ) {
        throw new UpDateError(" The criteria is not well formed.");
    }
    if (m_CommentsOnProperties_ == null) {
        throw new UpDateError(" The values are not well initialized.");
    }
    if (!m_CommentsOnProperties_.keySet().contains(_criteria) ) {
        throw new UpDateError(" Unknown key.");
    }

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeCommentsOnProperties(), "+
                      //      "values to remove for "+_criteria);

    HashMap m_old = (HashMap) m_CommentsOnProperties_.clone();
    if ( _l_values == null) {
        m_CommentsOnProperties_.remove(_criteria);

        try {
            firePropertyChange("CommentsOnProperties", m_old,
                    (HashMap) m_CommentsOnProperties_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeCommentsOnProperties(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)

    boolean error = false;

    for (int i = 0; i < _l_values.length; i++) {
      if (_l_values[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_values[i]);
      }
    } // for i
    if (error) {
        throw new UpDateError(" A value in the argument is null.");
    }

    String[] l_current = m_CommentsOnProperties_.get(_criteria);
    List list = new ArrayList(20);
    for (int i = 0; i < l_current.length; i++) {
        list.add(l_current[i]);
    }

    short count = 0;
    boolean found = false;

    for (int  i = 0 ; i < _l_values.length ; i++) {
        found = list.remove(_l_values[i]);

        if (!found) {
            error = true;
            break;
        } // if
        else count++;

    } // for i

    if (error) {
        m_old = null;
        l_current = null;
        list = null;
        throw new UpDateError(" A value to remove is not found.");
    }

    if (l_current.length != count) {
        m_CommentsOnProperties_.put(
                _criteria,
                (String[]) list.toArray(new String[list.size()]));

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeCommentsOnProperties(), "+
                      //      "the new values for "+_criteria+" are: ");
                    //for (Object val : list) {
                      //  System.out.println("      "+val);
                    //}

    } else { // if _l_values contains all the current values
        m_CommentsOnProperties_.put(
                _criteria, new String[0]);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeCommentsOnProperties(), "+
                      //      "the new values for "+_criteria+" are an empty array. ");

    }

    try {
        firePropertyChange("CommentsOnProperties", m_old,
                (HashMap) m_CommentsOnProperties_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeCommentsOnProperties(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
    l_current = null;
    list = null;
  } // removeCommentsOnProperties()

  /**
   * Gets comments to explain how this ACS simulates a real system/software.
   * The criteria is an ACS property displaying name. The value is a String array
   * of free user comments on the settings of the related ACS property.
   * @since 0.7.0
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings. May be empty, but not null. A value may be empty but is never null.
   */
  public final Map getM_CommentsOnProperties() {
      if (m_CommentsOnProperties_ == null) return new HashMap(INITIAL_CAPACITY_5);

      HashMap<String,String[]>  m_result =
         new HashMap<String,String[]>(INITIAL_CAPACITY_25);
      String[] l_values = null;
      Map.Entry<String,String[]> entry = null;
      String[] result = null;

              // for testing and tracking
              //System.out.println("in ACSFactoryImpl.getM_CommentsOnProperties(): ");

      for (Iterator ite = m_CommentsOnProperties_.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,String[]>)ite.next();

         l_values = entry.getValue();
         result = new String[l_values.length];
         System.arraycopy(l_values, 0, result, 0, l_values.length);
         m_result.put(entry.getKey(), result);

              // for testing and tracking
              //System.out.println("--- for the key "+entry.getKey());
              //for (int i = 0; i < result.length; i++) {
                //    System.out.println("--------- at "+i+": "+result[i]);
              //} // for i

      } // for ite


      l_values = null;
      entry = null;
      result = null;
      return m_result;
  } // getM_CommentsOnProperties()

  /**
   * This method returns the allowed keys which are used as map keys for
   * comments on properties.
   * The keys are all the ACS property displaying names, from the ACSImplBeanInfo.
   * The property 'ImmutableRules' has an empty display name, thus it is not returned
   * and it is not possible to comment it.
   * @since 0.7.0
   * @return List of the allowed keys
   */
  public List<String> getL_MapKeysForCommentsOnProperties() {
    PropertyDescriptor[] l_descriptors = (new ACSImplBeanInfo()).getPropertyDescriptors();

    List<String> l_result = new ArrayList<String>(10);

    for (int i = 0; i < l_descriptors.length; i++) {
        if (l_descriptors[i].getDisplayName() != null &&
                l_descriptors[i].getDisplayName().length() != 0)
            l_result.add(l_descriptors[i].getDisplayName());
    } // for i

    l_descriptors = null;
    return l_result;
  } // getL_MapKeysForCommentsOnProperties()

  /**
   * Sets an ACS vocabulary, without adding to the previous values which are deleted.
   * This may cover several independant domains: <P>
   * - the AcsAddon name, AcsAddon editor and AcsAddon package editor, in three different keys, <P>
   * - How to name the L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS constants for this ACS,
   * as names of the ACS tree nodes in the proper ACS internal frame. An example
   * is to replace the tree node 'Actors (right user target)' by 'Scripts'. <P>
   * - If the ACS is an AcsAddon, what are the generic gBase classes that are subclassed
   * in the AcsAddon gBase package by classic classes (see L_CLASSIC_CLASS_KEYS),
   * like when ResourceImpl is subclassed by ResourceUbuntuImpl in the AcsAddon Ubuntu.
   * The full Java name of the classic class has to be put in the value array. <P>
   * - If the ACS is an AcsAddon, what are the extension classes in
   * the AcsAddon gBase package (see ACSADDON_EXTENSION_CLASSES). In this case,
   * the class descriptions are separated by one escape, in any order, like in
   * 'Executable/is/Actor XXX/is/Resource'. <P>
   * Called by the ACSFactory constructor of any AcsAddon.
   * Fires a PropertyChange for 'AcsVocabulary'.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS
   * @see ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ACSFactoryUtilityImpl#ACSADDON_EXTENSION_CLASSES
   * @see ACSFactoryUtilityImpl#ACSADDON_NAME
   * @see ACSFactoryUtilityImpl#ACSADDON_EDITOR
   * @see ACSFactoryUtilityImpl#ACSADDON_PACKAGE_EDITOR
   * @see ACSFactoryUtilityImpl#ACSADDON_SPECIALIZED_RIGHT_DOMAINS
   * @param _criteria one of the ACSFactoryUtilityImpl constants
   * @param _l_values one item or more than one sometimes for the keys
   * ACSADDON_EXTENSION_CLASSES and ACSADDON_SPECIALIZED_RIGHT_DOMAINS.
   * Never null but the lenght may be 0.
   * @exception BaseError if:<P>
   *    - the criteria is null or not allowed<P>
   *    - the value is null or it contains more than one item<P>
   *    - the value is not allowed for the keys ACSADDON_EXTENSION_CLASSES and ACSADDON_SPECIALIZED_RIGHT_DOMAINS
   */
  public synchronized void addAcsVocabulary(String _criteria, String[] _l_values)
          throws BaseError {

                    // for testing and tracking
                    //System.out.println("Call of ACSFactoryImpl.addAcsVocabulary() ");

    if (_criteria == null || _l_values == null ||
            (_l_values.length > 1 &&
                !_criteria.equals(ACSFactoryUtilityImpl.ACSADDON_SPECIALIZED_RIGHT_DOMAINS) &&
                !_criteria.equals(ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES))) {
        throw new BaseError(" The key or the value is null, or the value contains more than"+
                " one item while this key does not allow it.");
    }

    List<String>    list = getL_MapKeysForAcsVocabulary();
    String[]        l_segments = null;
    boolean         error = false;
    int             index = -1;

    if (!list.contains(_criteria))
        throw new BaseError(" "+_criteria+" is not allowed as a key.");

    // control the complex String value for ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES
    if (_criteria.equals(ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES) && 
            _l_values.length != 0) {

        for (int i = 0; i < _l_values.length; i++) {
            if (_l_values[i].indexOf("/is/") == -1 ) {
                error = true;
                index = i;
                break;
            }
        } // for i

        if (error)
                throw new BaseError(
                        " A value does not use the separator '/is/'.");

        for (int i = 0; i < _l_values.length; i++) {
            l_segments = _l_values[i].split("/is/");
            if (l_segments[0].isEmpty()) {
                error = true;
                index = i;
                break;
            } // if
            if (!ACSFactoryUtilityImpl.L_CLASSIC_CLASS_KEYS.contains(l_segments[1])) {
                error = true;
                index = i;
                break;
            }
        } // for i

        if (error)
                throw new BaseError(" The first item in the value at "+index+
                        " does not name a new class,"+
                        " or the last item does not name a generic class.");

    } // if (_criteria.equals(ACSFactoryUtilityImpl.ACSADDON_NEW_CLASSES)

    // control the value for ACSFactoryUtilityImpl.ACSADDON_SPECIALIZED_RIGHT_DOMAINS
    if (_criteria.equals(ACSFactoryUtilityImpl.ACSADDON_SPECIALIZED_RIGHT_DOMAINS) && 
            _l_values.length != 0) {

        for (int i = 0; i < _l_values.length; i++) {
            if (_l_values[i].indexOf(".") == -1 ||
                    _l_values[i].indexOf(".") != _l_values[i].lastIndexOf(".")) {
                

                                    // for testing and tracking
                                    //System.out.println("ACSFactoryImpl.addAcsVocabulary() "+
                                      //      "with error on "+_l_values[i]);

                error = true;
                break;
            } // if
        } // for i

        if (error)
                throw new BaseError(" A value does not use the separator '.' one time.");

        for (int i = 0; i < _l_values.length; i++) {
            if (_l_values[i].substring(0, _l_values[i].indexOf(".")).indexOf("Rights") == -1) {
                error = true;
                break;
            }
        } // for i

        if (error)
                throw new BaseError(" A value does not contain 'Rights' in its first part.");

        for (int i = 0; i < _l_values.length; i++) {
            if (!ACSFactoryUtilityImpl.L_SPECIALIZED_RIGHT_DOMAINS.contains(
                    _l_values[i].substring(_l_values[i].indexOf(".") + 1))) {
                error = true;
                break;
            }
        } // for i

        if (error)
            throw new BaseError(" A value does not contain a specialized right "+
                    "domain name in its second part.");

    } // _criteria.equals(ACSFactoryUtilityImpl.ACSADDON_SPECIALIZED_RIGHT_DOMAINS)

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addAcsVocabulary(), "+
                      //      "the rights are controlled OK ");

    HashMap m_old = (HashMap) m_AcsVocabulary_.clone();

    m_AcsVocabulary_.put(_criteria, _l_values);

                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.addAcsVocabulary(), "+
                      //      "fire the property change 'AcsVocabulary' ");

    try {
        firePropertyChange("AcsVocabulary", m_old, (HashMap) m_AcsVocabulary_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.addAcsVocabulary(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    m_old = null;
    list = null;
    l_segments = null;
  } // addAcsVocabulary()

  /**
   * Removes an ACS vocabulary. This may cover several independant domains: <P>
   * - the AcsAddon editor and name, in two different keys, <P>
   * - How to name the L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS constants for this ACS,
   * as names of the ACS tree nodes in the proper ACS internal frame. An example
   * is to replace the tree node 'Actors (right user target)' by 'Scripts'. <P>
   * - If the ACS is an AcsAddon, what are the generic gBase classes that are subclassed
   * in the AcsAddon gBase package by classic classes (see L_CLASSIC_CLASS_KEYS),
   * like when ResourceImpl is subclassed by ResourceUbuntuImpl in the AcsAddon Ubuntu.
   * The full Java name of the classic class has to be put in the value array. <P>
   * - If the ACS is an AcsAddon, what are the extension classes in
   * the AcsAddon gBase package (see ACSADDON_EXTENSION_CLASSES). In this case,
   * the class descriptions are separated by one escape, like in
   * 'Executable/is/Actor XXX/is/Resource'. <P>
   * Fires a PropertyChange for 'AcsVocabulary'.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS
   * @see ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ACSFactoryUtilityImpl#ACSADDON_EXTENSION_CLASSES
   * @see ACSFactoryUtilityImpl#ACSADDON_NAME
   * @see ACSFactoryUtilityImpl#ACSADDON_EDITOR
   * @see ACSFactoryUtilityImpl#ACSADDON_PACKAGE_EDITOR
   * @param _criteria one of the ACSFactoryUtilityImpl constants
   * @param _l_values values to remove. This array is limited to one item by the BeanInfo.
   * @exception BaseError if:<P>
   *    - the criteria is null or not known, <P>
   *    - a value in _l_values is null, <P>
   *    - a value is unknown.
   */
  public synchronized void removeAcsVocabulary(String _criteria,
    String[] _l_values) throws BaseError {

    if (_criteria == null ) {
        throw new BaseError(" The criteria is not well formed.");
    }
    if (m_AcsVocabulary_ == null) {
        throw new BaseError(" The map is not well initialized.");
    }
    if (!m_AcsVocabulary_.keySet().contains(_criteria) ) {
        throw new BaseError(" Unknown key: "+_criteria);
    }

                                // for testing and tracking
                                System.out.println("From ACSFactoryImpl.removeAcsVocabulary(), "+
                                        ", for the key: "+_criteria+", the values to remove are: ");
                                if (_l_values != null) {
                                    for (int i = 0; i < _l_values.length; i++) {
                                        System.out.println("--- "+_l_values[i]);
                                    }
                                }
                                else {
                                    System.out.println("--- null");
                                }

    HashMap m_old = (HashMap) m_AcsVocabulary_.clone();
    if ( _l_values == null) {
        m_AcsVocabulary_.remove(_criteria);

        try {
            firePropertyChange("AcsVocabulary", m_old, (HashMap) m_AcsVocabulary_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
            System.out.println("From ACSFactoryImpl.removeAcsVocabulary(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)
    else {
        if (_l_values.length > 1 ) {
          m_old = null;
          throw new BaseError(" The value to remove is more than one item.");
        }
        if (_l_values.length == 1 &&
                !_l_values.equals(m_AcsVocabulary_.get(_criteria)) ) {
          m_old = null;
          String[] l_old = m_AcsVocabulary_.get(_criteria);
          
                                System.out.println("Unknown value to remove in:");
                                for (int i = 0; i < l_old.length; i++) {
                                    System.out.println("--- "+l_old[i]);
                                }

          throw new BaseError(" Unknown value to remove");
        }

                    // for testing and tracking
                    System.out.println("From ACSFactoryImpl.removeAcsVocabulary(), "+
                            "removed value: "+m_AcsVocabulary_.get(_criteria));

        m_AcsVocabulary_.put(_criteria, new String[0]);
    } // else

    try {
        firePropertyChange("AcsVocabulary", m_old, (HashMap) m_AcsVocabulary_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data updating, from outside the database."+
              " You can continue the ACS creation.");
        System.out.println("From ACSFactoryImpl.removeAcsVocabulary(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
  } // removeAcsVocabulary()

  /**
   * Gets an ACS vocabulary. This may cover several independant domains: <P>
   * - the AcsAddon name, AcsAddon editor and AcsAddon package editor, in three different keys, <P>
   * - How to name the L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS constants for this ACS,
   * as names of the ACS tree nodes in the proper ACS internal frame. An example
   * is to replace the tree node 'Actors (right user target)' by 'Scripts'. <P>
   * - If the ACS is an AcsAddon, what are the generic gBase classes that are subclassed
   * in the AcsAddon gBase package by classic classes (see L_CLASSIC_CLASS_KEYS),
   * like when ResourceImpl is subclassed by ResourceUbuntuImpl in the AcsAddon Ubuntu.
   * The full Java name of the classic class has to be put in the value array. <P>
   * - If the ACS is an AcsAddon, what are the extension classes in
   * the AcsAddon gBase package (see ACSADDON_EXTENSION_CLASSES). In this case,
   * the class descriptions are separated by one escape, like in
   * 'Executable/is/Actor XXX/is/Resource'. <P>
   * - If the ACS is an AcsAddon, what are the non-generic rights in the
   * ACSObjects on which the ACS constraints of the specialized rights
   * are applicable, and in that case, what domain is applicable. An example
   * is the domain 'GpORightsSet' for the property 'OtherRights', in Ubuntu,
   * that is declared by the value 'OtherRights.GpORightsSet'. <P>
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS
   * @see ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ACSFactoryUtilityImpl#ACSADDON_EXTENSION_CLASSES
   * @see ACSFactoryUtilityImpl#ACSADDON_NAME
   * @see ACSFactoryUtilityImpl#ACSADDON_EDITOR
   * @see ACSFactoryUtilityImpl#ACSADDON_PACKAGE_EDITOR
   * @see ACSFactoryUtilityImpl#ACSADDON_SPECIALIZED_RIGHT_DOMAINS
   * @return an unmodifiable Map where the key is a String, and the value is a String
   * array with at most one item. May be empty but not null.
   * A map value is never null but may be empty.
   */
  public final Map<String,String[]> getM_AcsVocabulary() {
      return Collections.unmodifiableMap(m_AcsVocabulary_);
  } // getM_AcsVocabulary()

  /**
   * This method returns the allowed keys which are used as map keys for
   * the ACS vocabulary. The values are constants, provided by ACSFactoryUtilityImpl.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS
   * @see ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ACSFactoryUtilityImpl#ACSADDON_EXTENSION_CLASSES
   * @see ACSFactoryUtilityImpl#ACSADDON_NAME
   * @see ACSFactoryUtilityImpl#ACSADDON_EDITOR
   * @see ACSFactoryUtilityImpl#ACSADDON_PACKAGE_EDITOR
   * @see ACSFactoryUtilityImpl#ACSADDON_SPECIALIZED_RIGHT_DOMAINS
   * @see ACSFactoryUtilityImpl#ACSADDON_VERSION
   * @return List containing the allowed keys
   */
  public List<String> getL_MapKeysForAcsVocabulary() {
    List<String> l_result = new ArrayList<String>(INITIAL_CAPACITY_25);

    l_result.addAll(ACSFactoryUtilityImpl.L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS);
    l_result.addAll(ACSFactoryUtilityImpl.L_CLASSIC_CLASS_KEYS);
    l_result.add(ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES);
    l_result.add(ACSFactoryUtilityImpl.ACSADDON_NAME);
    l_result.add(ACSFactoryUtilityImpl.ACSADDON_EDITOR);
    l_result.add(ACSFactoryUtilityImpl.ACSADDON_PACKAGE_EDITOR);
    l_result.add(ACSFactoryUtilityImpl.ACSADDON_SPECIALIZED_RIGHT_DOMAINS);
    l_result.add(ACSFactoryUtilityImpl.ACSADDON_VERSION);

    return l_result;
  } // getL_MapKeysForAcsVocabulary()


  /**
  * Adds a listener to the bean, for all the bound properties.
  * Used by ActionNewACSyst.
  * @param _l PropertyChangeListener to add
  */
  public void addPropertyChangeListener(PropertyChangeListener _l){

      synchronized (changeSupport_) 
        {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'ParentName', 'HubProperty', 'VisibleSpace',
  * 'ResourceTypes', 'EPTypes', 'TypesAndExplorerSet', 'OneAcsRight',
  * 'SpecializedRightsSet', 'HavePrivilegeSpecializedRights', 'AssociationForNewObjectsMap',
  * 'ARightsSet', 'GORightsSet', 'AclRightsSet',
  * 'PrivilegeRightsSet', 'BridgeRightsSet',
  * 'ManageResources', 'ManageResourceTrees', 'ManageAccounts', 
  * 'ManageGroupsOfAccounts', 'ManageGroupsOfGroups', 'ManageInternalAcls',
  * 'ManageExternalAcls', 'ManageFlexibleConditionalAcls', 'ManageRightDefinedConditionalAcls',
  * 'ManageActors', 'ManageBridges',
  * 'ManageSecondaryGroupsForActor', 'ManageAGORight', 'ManageNxExecutables',
  * 'ManageDirectoryEPs', 'ManagePrivilegeRights',
  * 'VirtualFolders', 'ManageGrantRights', 'ManageDenyRights',
  * 'TiedToParentACS', 'RunningOnParentACS', 'isRightServer',
  * 'ComponentNodes', 'CompositeNodes', 'BelongsToComposite', 'SubAndGroupPolicy',
  * 'Alternates', 'OpeningAndClosingRoots', 'OpeningNoClosingRoots', 'IntegrityRoots',
  * 'CommentsOnProperties', 'Icon',
  * 'ManageGrantingRightPrivileges', 'ManageDenyingRightPrivileges',
  * 'ManageSimpleTypedPrivileges', 'ManageSimpleLinkedPrivileges',
  * 'ManageResourceVirtualFolders', 'ManageEPVirtualFolders',
  * 'ManageActorInEPVirtualFolders', 'ManageUserIDInEPVirtualFolders',
  * 'ManageGroupIDInEPVirtualFolders', 'ManageSimpleResourceInResourceVirtualFolders',
  * 'ManageDirectoryInResourceVirtualFolders', 'ManageDirectoryEPinEPVirtualFolders',
  * 'ManageVirtualFolderTrees', 'OneTimeInVirtualFolderTreeForMember',
  * 'ManageACLRightInheritance', 'ManageAGORightInheritance',
  * "ManageLPRIRightInheritance', 'ManagePrivilegesInVirtualFolders',
  * 'ManagePrivilegesInResources',
  * 'ManageIfAllSourceLinkedPrivileges', 'ManageInternalForOneToOneLinkedPrivileges',
  * 'ManageExternalForOneToOneLinkedPrivileges', 'ManageClassedTypedPrivileges',
  * 'ManageClassedSpecificTypedPrivileges', 'ManageAuthorizationServerTypedPrivileges',
  * 'IncompleteStructureModeling', 'IncompleteBehaviorModeling',
  * 'ResVirtualFolderTypes', 'EPVirtualFolderTypes',
  * 'PrivilegeForTypeTypes', 'PrivilegeForLinksTypes',
  * 'AcsVocabulary'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      
                // for testing and tracking 
                //if (_propertyName.equals("AssociationForNewObjectsMap"))
                  //  System.out.println("in ACSFactoryImpl.addPropertyChangeListener(), call for "+
                    //    _l+" for the property "+_propertyName);
      
      synchronized (changeSupport_) 
      {changeSupport_.addPropertyChangeListener(_propertyName, _l);}   
                
                // for testing
                //PropertyChangeListener[] l_prop = changeSupport_.
                  //      getPropertyChangeListeners("AssociationForNewObjectsMap");
                //if (_propertyName.equals("AssociationForNewObjectsMap")) {
                  //  for (int i = 0; i < l_prop.length; i++) {
                    //    System.out.println("--------- "+l_prop[i]);
                    //} // for i
                //} // if
                
  } // addPropertyChangeListener(String _propertyName,

  /**
  * Removes a listener to the bean, for all the bound properties.
  * Used by ActionNewACSyst.
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(PropertyChangeListener _l){
    
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;
    
      synchronized (changeSupport_) 
      {changeSupport_.removePropertyChangeListener(_l);} 
  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'ParentName', 'HubProperty', 'VisibleSpace',
  * 'ResourceTypes', 'EPTypes', 'TypesAndExplorerSet', 'OneAcsRight',
  * 'SpecializedRightsSet', 'HavePrivilegeSpecializedRights', 'AssociationForNewObjectsMap',
  * 'ARightsSet', 'GORightsSet', 'AclRightsSet',
  * 'PrivilegeRightsSet', 'BridgeRightsSet',
  * 'ManageResources', 'ManageResourceTrees', 'ManageAccounts', 
  * 'ManageGroupsOfAccounts', 'ManageGroupsOfGroups', 'ManageInternalAcls',
  * 'ManageExternalAcls', 'ManageFlexibleConditionalAcls', 'ManageRightDefinedConditionalAcls',
  * 'ManageActors', 'ManageBridges',
  * 'ManageSecondaryGroupsForActor', 'ManageAGORight', 'ManageNxExecutables',
  * 'ManageDirectoryEPs', 'ManagePrivilegeRights',
  * 'VirtualFolders', 'ManageGrantRights', 'ManageDenyRights',
  * 'TiedToParentACS', 'RunningOnParentACS', 'isRightServer',
  * 'ComponentNodes', 'CompositeNodes', 'BelongsToComposite', 'SubAndGroupPolicy',
  * 'Alternates', 'OpeningAndClosingRoots', 'OpeningNoClosingRoots', 'IntegrityRoots',
  * 'CommentsOnProperties', 'Icon',
  * 'ManageGrantingRightPrivileges', 'ManageDenyingRightPrivileges',
  * 'ManageSimpleTypedPrivileges', 'ManageSimpleLinkedPrivileges',
  * 'ManageResourceVirtualFolders', 'ManageEPVirtualFolders',
  * 'ManageActorInEPVirtualFolders', 'ManageUserIDInEPVirtualFolders',
  * 'ManageGroupIDInEPVirtualFolders', 'ManageSimpleResourceInResourceVirtualFolders',
  * 'ManageDirectoryInResourceVirtualFolders', 'ManageDirectoryEPinEPVirtualFolders',
  * 'ManageVirtualFolderTrees', 'OneTimeInVirtualFolderTreeForMember',
  * 'ManageACLRightInheritance', 'ManageAGORightInheritance',
  * "ManageLPRIRightInheritance',
  * 'ManageIfAllSourceLinkedPrivileges', 'ManageInternalForOneToOneLinkedPrivileges',
  * 'ManageExternalForOneToOneLinkedPrivileges', 'ManageClassedTypedPrivileges',
  * 'ManageClassedSpecificTypedPrivileges', 'ManageAuthorizationServerTypedPrivileges',
  * 'IncompleteStructureModeling', 'IncompleteBehaviorModeling',
  * 'ResVirtualFolderTypes', 'EPVirtualFolderTypes',
  * 'PrivilegeForTypeTypes', 'PrivilegeForLinksTypes',
  * 'AcsVocabulary'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
    
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;
    
      synchronized (changeSupport_) 
        {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()
 
    /**
   * @return the hashcode value for the ACS, based on the super method.
   */
  public int hashCode() {
    return super.hashCode();
  } // hashcode()

  /**
  * Compare only the type and the name.
  * @param _obj object to compare
  * @return true if the two objects are equal instances of ACSFactoryImpl
  */
  public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      ACSFactoryImpl obj = (ACSFactoryImpl) _obj;

      if (!type_.equals(obj.getACSType())) return false;

      if (!name_.equals(obj.getName())) return false;

      return true;
  } //equals()
  
  /**
   * Gets the name last component of the instance.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    //sb.append("ACSFactory" );
    sb.append(name_.lastComponent() );

    /*
        sb.append("; size of the resources map = ");
        sb.append(m_Resources_.size());

        sb.append("; size of the eligible parties map = ");
        sb.append(m_EligibleParties_.size());

        sb.append("; size of the acl entries array = ");
        sb.append(l_AclEntry_.length);
    */

    return " " + sb;
  } // toString()
  
  //
  // ---------------- Protected and pivate methods -----------------------------------
  //
 
  /**
  * Fires an event to every registered listener, in any order.
  * For the bound properties 
  * 'ISname', 'ParentName', 'HubProperty', 'VisibleSpace',
  * 'ResourceTypes', 'EPTypes', 'TypesAndExplorerSet', 'OneAcsRight',
  * 'SpecializedRightsSet', 'HavePrivilegeSpecializedRights',
  * 'ARightsSet', 'GORightsSet', 'AclRightsSet',
  * 'PrivilegeRightsSet', 'BridgeRightsSet',
  * 'ManageResources', 'ManageResourceTrees', 'ManageAccounts', 
  * 'ManageGroupsOfAccounts', 'ManageGroupsOfGroups', 'ManageInternalAcls',
  * 'ManageExternalAcls', 'ManageFlexibleConditionalAcls', 'ManageRightDefinedConditionalAcls',
  * 'ManageActors', 'ManageBridges',
  * 'ManageSecondaryGroupsForActor', 'ManageAGORight', 'ManageNxExecutables',
  * 'ManageDirectoryEPs', 'ManagePrivilegeRights',
  * 'VirtualFolders', 'ManageGrantRights', 'ManageDenyRights',
  * 'TiedToParentACS', 'RunningOnParentACS', 'isRightServer',
  * 'ComponentNodes', 'CompositeNodes', 'BelongsToComposite', 'SubAndGroupPolicy',
  * 'Alternates', 'OpeningAndClosingRoots', 'OpeningNoClosingRoots', 'IntegrityRoots',
  * 'CommentsOnProperties', 'Icon',
  * 'IncompleteStructureModeling', 'IncompleteBehaviorModeling',
  * 'ManageGrantingRightPrivileges', 'ManageDenyingRightPrivileges',
  * 'ManageSimpleTypedPrivileges', 'ManageSimpleLinkedPrivileges',
  * 'ManageResourceVirtualFolders', 'ManageEPVirtualFolders',
  * 'ManageActorInEPVirtualFolders', 'ManageUserIDInEPVirtualFolders',
  * 'ManageGroupIDInEPVirtualFolders', 'ManageSimpleResourceInResourceVirtualFolders',
  * 'ManageDirectoryInResourceVirtualFolders', 'ManageDirectoryEPinEPVirtualFolders',
  * 'ManageVirtualFolderTrees', 'ManageOneTimeInVirtualFolderTreeForMember',
  * 'ManageACLRightInheritance', 'ManageAGORightInheritance',
  * 'ManageLPRIRightInheritance', 'ManagePrivilegesInVirtualFolders',
  * 'ManagePrivilegesInResources',
  * 'ManageIfAllSourceLinkedPrivileges', 'ManageInternalForOneToOneLinkedPrivileges',
  * 'ManageExternalForOneToOneLinkedPrivileges', 'ManageClassedTypedPrivileges',
  * 'ManageClassedSpecificTypedPrivileges', 'ManageAuthorizationServerTypedPrivileges',
  * 'ManageInternalHardAlias', 'ManageSoftAlias',
  * 'IncompleteStructureModeling', 'IncompleteBehaviorModeling',
  * 'ResVirtualFolderTypes', 'EPVirtualFolderTypes',
  * 'PrivilegeForTypeTypes', 'PrivilegeForLinksTypes',
  * 'AcsVocabulary'.
  * Called by all the updating methods.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {
      
                // for testing and tracking
                //if (_propertyName.equals("AssociationForNewObjectsMap")) {
                  //  System.out.println("in ACSFactoryImpl.firePropertyChange(), call"+
                  //" for the property "+_propertyName);
                    //if (_oldValue.equals(_newValue))
                      //  System.out.println("in ACSFactoryImpl.firePropertyChange(), the old "+
                        //        "and the new values are equals for the property "+_propertyName);
                //} // if
      
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;
    
      synchronized (changeSupport_) 
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);} 

  } // firePropertyChange()
  
  /**
   * This method removes the right from the ACS list, but there is no
   * guaranee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForUser();
    } catch (UpDateError u) {u.printStackTrace();}

    super.finalize();
  } // finalize()  
  
  /**
  * Closes the instance. The list and map variables are not cleared since they
  * set the ACSImpl variables without cloning.
  * Called by ActionNewACSyst and createACS().
  * @exception UpDateError if the super class throws it
  */
  public synchronized void finalizeForUser() throws UpDateError {

              // for testing purpose
              //System.out.println("call of ACSFactoryImpl.finalizeForUser() for "+getName());

      name_ = null;
      type_ = null; 

      iSname_ = null;
      parentName_ = null;
      visibleSpace_  = null;
      changeSupport_ = null;

      l_resourceTypes_ = null; 
      l_epTypes_ = null; 
      l_resvfTypes_ = null;
      l_epvfTypes_ = null;
      l_privilegeForTypeTypes_ = null;
      l_privilegeForLinksTypes_ = null;
      l_AccountRightsSet_ = null;
      l_GORightsSet_ = null; 
      l_AclRightsSet_ = null;
      m_PrivilegeRights_ = null;
      l_bridgeRightsSet_ = null; 
      l_Rights_ = null;

      // no clear() on maps, since some of them are directly used in the ACS constructor
      m_SpecializedRights_ = null;
      m_Structure_ = null; 
      m_TypesAndExplorer_ = null;
      m_AssociationsForNewObjects_ = null;
      m_StandardRights_ = null;
      l_ImmutableRules_ = null;

      // new in 0.7.0
      m_BelongsToComposite_ = null;
      m_SubAndGroupPolicy_ = null;
      m_ComponentNodes_ = null;
      m_CompositeNodes_ = null;
      m_RootsForAcsOpeningAndClosing_ = null;
      m_RootsForAcsOpeningNoClosing_ = null;
      m_RootsForAcsIntegrity_ = null;
      m_alternateProperties_ = null;
      icon_ = null;
      m_CommentsOnProperties_ = null;
      
      try {
          super.finalize();
      } catch (Throwable t) {throw new UpDateError(t.toString());}
  } // finalizeForUser()  
  
  /**
   * Controls the right values for the acs, which have to be registered only
   * if there are acsrights.
   * The rights set has not to be dense, since a right may have no neighbourhood
   * (lower or upper rights) in the array.
   * The right sens is controlled for denying rights, not for granting rights.
   * @param _s is an array of acsrights and metarights
   * @param _allowDeny true if _s may contain denying rights
   * @exception UpDateError if _s is  null, or desactivated, 
   * or it is an ACS right not registered, without a metaright or the correct acs.
   */
  protected void controlRights(StringRight[] _s, boolean _allowDeny) throws UpDateError {
    if (_s == null)
        throw new UpDateError(" The new rights set is null.");
    if (_s.length == 0) return;
    
    int error = -1;
    int error2 = -1;    
    int error3 = -1;
    int error4 = -1;
    int error5 = -1;
    int error6 = -1;
    
    for (int i = 0; i < _s.length; i++) {
      if (_s[i] == null)  {
          error = i;
          break;
      }
      if (!_s[i].isGrantingRight() && !_allowDeny)  {
          error5 = i;
          break;
      }
      
      // controls the acs name and the metaright if it is not a metaright
      // control it is not desactivated
      else if (!(_s[i] instanceof StringMetaRight) ) {
          if (!_s[i].getAcsName().equals(name_)) { 
              error2 = i;
              break;          
          } else if (_s[i].getEorMetaRight() == null ) { 
              error6 = i;
              break;
          } else if (_s[i].isDesactivated() ) { 
              error3 = i;
              break;
          } else if (getOneRight(_s[i].getName()) == null) { 
              error4 = i;
              break;
          }
      } // else
      else if (_s[i].isDesactivated() ) { 
              error3 = i;
              break;
      } // else
    } // for i
    
    if (error != -1)
        throw new UpDateError(" A new right is null, at "+error);
    else if (error2 != -1)
        throw new UpDateError(" A new right has not the good ACS name, at "+error2);
    else if (error3 != -1)
        throw new UpDateError(" A new right "+_s[error3]+" is desactivated, at "+error3);
    else if (error4 != -1)
        throw new UpDateError(" A new right "+_s[error4]+" is not known by the ACSFactory, at "+error4);
    else if (error6 != -1)
        throw new UpDateError(" A new right "+_s[error6]+" has a null metaright, at "+error6);
    else if (error5 != -1)
        throw new UpDateError(" A new right "+_s[error5]+" at "+error5+", is negative" +
          " and this is not allowed for this right category.");
    
  } // controlRights()
  
  /**
   * Registers a new right through the ACSpreFactoryBeanInfo.
   * Used to register the new acsrights and, in initializeConstraints(),
   * to register the used metarights just before the ACS creation.
   * Updates the structure to indicate the type of rights if it is an acsright.
   * Sorts the rights array.
   * Called by addRight() and initializeConstraints().
   * @param _right is the new right to register. No operation if null or empty.
   */
   protected void newRight(StringRight _right ) {
      
      // update the l_structure_ about acsrights - metarights are always true
      // for the PreACSFactory
      boolean acs = getManageAcsRights();
      if (!(_right instanceof StringMetaRight) && !acs)
                    m_Structure_.put("acs: acsrights", Boolean.TRUE);      
                        
      // adds _right in l_Rights_ 
      l_Rights_ = BaseUtilityImpl.addRightInDecreasingSort(l_Rights_, _right);
   } // newRight()

  /**
   * Unregisters a right through the ACSpreFactoryBeanInfo.
   * Updates the structure to indicate the type of rights: acsright or metaright.
   * Called by removeRight().
   * @param _right is the folder to unregister.
   * @exception UpDateError if the folder is null or not known.
   */
   protected void moveRight(StringRight _right ) throws UpDateError {
        
        // removes _right in l_Rights_
        int               shiftingBack = 0;
        StringRight[]     newArray = new StringRight[l_Rights_.length];
        StringRight[]     result = new StringRight[0];
        boolean           containACSright = false;
        
                // this code works even if shiftingBack > 1
        for (int  i = 0 ; i < l_Rights_.length ; i++) {
             // delete the empty l_Rights_[i]
            if (l_Rights_[i] == null || l_Rights_[i].isEmpty() ) shiftingBack++; 
            if (l_Rights_[i] == _right ) {                        // == rather than equals
                // incrementing shiftingBack deletes l_Rights_[i]
              shiftingBack++;
            }
            else {
                newArray[i - shiftingBack] = (StringRight)l_Rights_[i];
                if (!(l_Rights_[i] instanceof StringMetaRight))  containACSright = true;
            } // else
        } // for i

        if (l_Rights_.length != shiftingBack) {
            result = new StringRight[l_Rights_.length - shiftingBack];
            System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
        }
        l_Rights_ = result;

        // exceptions
        if (shiftingBack > 1) {
            throw new UpDateError("In ACSImpl.removeRight(_right),"+
              " _right is recorded more than once.");
        }
      
        // updates the l_structure_
        boolean acs = getManageAcsRights();
        if (!containACSright && acs)
                    m_Structure_.put("acs: acsrights", Boolean.FALSE);      
   
        newArray = null;
        result = null;

   } // moveRight()  
   
  /**
   * Adds one association to handle at the creation or the deletion of ACSObjects.
   * The criteria is a map key among six allowed formats.
   * The allowed String map values depends lightly on the map key.
   * The non-empty values define fully or partially one action to do if the key matches.
   * One action is always a constructor or a method to call. <P>
   * There is a fully-defined association when the values defines it with all the required
   * arguments. When _l_values is defined partially, this means that the constructor
   * or method is not well defined, or that its arguments are not all defined, or
   * are poorly defined in the values. Every partially-defined array for _l_values
   * is detected by this method, which adds to the array, at its end, a message to
   * the user to tell him the nature of the current error. The property is then
   * updated and displayed to the user, so he should update the values correctly.
   * When there is a fully-defined action, this method adds an action number 
   * to each value. So, the validated associations appear to the user as a
   * list of numbered values, which is ended, if any, by the last partially-defined
   * association without action number and with an error message. <P>
   * The number of actions is limited to 999.
   * See AssociateUtilityImpl.controlAssociations() for a full description of the forms.
   * Fires no Property Change.
   * Called by addAssociationsToNewObject(), addOpeningAndClosingRoots(),
   * addOpeningNoClosingRoots() and addIntegrityRoots().
   * @see ARoad0.gBase.AssociateUtilityImpl#getL_MapKeyStartsForAssociationsOnNewObject
   * @see ARoad0.gBase.AssociateUtilityImpl#controlAssociations
   * @see ARoad0.Gui1.AssociationMapPropertyEditor
   * @param _m_upd_Associations actions to do. May be null or empty.
   * @param _criteria is the map key to apply these associations. Never null.
   * @param _l_values a Strings array which contain the values for the criteria,
   *    but without the action number 'NNN.'. Values are added to the previous values, if any.
   *    Never null but the lenght may be 0.
   * @return the updated _m_upd_Associations
   * @exception UpDateError if:<P>
   *    - a value is not allowed.
   */
  protected synchronized HashMap<String, String[]> addAssociations(
          HashMap<String, String[]> _m_upd_Associations, String _criteria, String[] _l_values) 
          throws UpDateError {

                    // for testing and tracking
                    //System.out.println("ACSFactoryImpl.addAssociations() called "+
                      //      "with criteria "+_criteria);
    
    Integer             lastNumber = -1;
    Integer             newNumber = -1;
    boolean             toAdd = false;
    String[]            l_old = _m_upd_Associations.get(_criteria);
    String[]            l_filtered = null;
    LinkedList<String>  l_toControl = null;
    LinkedList<String>  l_new = new LinkedList();
    String              control = null;   

    if (l_old != null && l_old.length != 0) {
        l_toControl = new LinkedList();
        // controls the set of values for the last action
        l_filtered = (String[]) AssociateUtilityImpl.getM_FilteredAssociations(
                _m_upd_Associations, true).get(_criteria);
        
        if (l_filtered != null && l_filtered.length != 0) {
            if ( l_filtered[l_filtered.length - 1].substring(3, 4).equals(".") )
                lastNumber = new Integer(l_filtered[l_filtered.length - 1].substring(0, 3));
            else if ( l_filtered[l_filtered.length - 1].substring(2, 3).equals(".") )
                lastNumber = new Integer(l_filtered[l_filtered.length - 1].substring(0, 2));
            else lastNumber = new Integer(l_filtered[l_filtered.length - 1].substring(0, 1));
        } // else lastNumber == -1
        
        if (lastNumber == 999) {
            System.out.println("in ACSFactoryImpl.addAssociations() "+
                    _criteria+" has already 999 actions.");
            throw new UpDateError(" "+_criteria+" cannot have more than 999 actions.");
        }
        
                        // for testing and tracking
                        //if (lastNumber == 99) {
                           //System.out.println("ACSFactoryImpl.addAssociations()");
                          //  if (l_filtered[l_filtered.length - 1] != null)
                            //    System.out.println("lastNumber: "+lastNumber+" from "+
                              //          l_filtered[l_filtered.length - 1]);
                            //else System.out.println("lastNumber: "+lastNumber);
                        //} // if
                                     

        if (lastNumber != -1) {
            // puts in l_toControl the last no-number values in l_old, without the errors
            for (int i = 0; i < l_old.length; i++) {
                if (l_old[i].startsWith(lastNumber.toString()+".")) {
                    toAdd = true;
                    continue;
                } 
                if (toAdd && !l_old[i].startsWith(lastNumber.toString()+".") && 
                        !l_old[i].startsWith("Error")) 
                    l_toControl.add(l_old[i]);
            } // for i 
        } // if
        else {
            // puts in l_toControl all the (no-number) values in l_old, without the errors
            for (int i = 0; i < l_old.length; i++) {
                if (!l_old[i].startsWith("Error")) l_toControl.add(l_old[i]);
            } // for i             
        } // else

                    // for testing and tracking
                    //System.out.println("toControl before adding of _l_values: ");
                    //for (Iterator ite = l_toControl.iterator(); ite.hasNext(); ) {
                      //  System.out.println("------"+ite.next());
                    //} // for ite
        
        l_toControl.addAll(Arrays.asList(_l_values) );

                            // for testing and tracking
                            //if (lastNumber == 99 || lastNumber == 100) {
                                //System.out.println("ACSFactoryImpl.addAssociations()");
                                //System.out.println("toControl after the adding of _l_values, with "+
                                  //      "lastNumber = "+lastNumber);
                                //for (Iterator ite = l_toControl.iterator(); ite.hasNext(); ) {
                                  //  System.out.println("------"+ite.next());
                                //} // for ite
                                //String[] lcontrol = (String[])l_toControl.toArray(new String[0]);
                                        // for testing and tracking
                                //System.out.println("toControl as array: ");
                                //for (int i = 0; i < lcontrol.length; i++ ) {
                                  //  System.out.println("------"+lcontrol[i]);
                                //} // for i
                                //lcontrol = null;
                            //} // if
        
        // the returned value starts with 'Error ' if not null
        control = AssociateUtilityImpl.controlAssociations(
                this,
                _m_upd_Associations,
                _criteria, (String[])l_toControl.toArray(new String[0]), false);
    } // if
    else if (_l_values.length != 0)
        // the returned value starts with 'Error ' if not null
        control = AssociateUtilityImpl.controlAssociations(
                this,
                _m_upd_Associations,
                _criteria, _l_values, false);
    
                            // for testing and tracking
                            //System.out.println("From ACSFactoryImpl.addAssociations(), "+
                              //      "the rights are controlled OK ");
    

    if (l_old != null && l_old.length != 0) {
        // the values are not well formed; no number is added at their starts
        // and the control message is added at the end of the array to inform
        // the user; removes the previous error message
        if (control != null) {
            for (int i = 0; i < l_old.length; i++) {
                if (!l_old[i].startsWith("Error")) l_new.add(l_old[i]);
            } // for i
            l_new.addAll(Arrays.asList(_l_values) );
            l_new.add(control);

                            // for testing and tracking
                            System.out.println("From ACSFactoryImpl.addAssociations(), "+
                                    "the first values set for _criteria "+_criteria+
                                    " has a non-null control: "+control);

        } else {
            // the values are well formed, the right number is added to the
            // start of the last current no-number values and to the new values;
            // removes the last current error message
            if (lastNumber != -1) {
                newNumber = new Integer(lastNumber.intValue()+1);
                l_new.addAll(Arrays.asList(l_filtered) );
            }
            else newNumber = 1;  

                            // for testing and tracking
                            //if (lastNumber == 99 || lastNumber == 100)
                              //  System.out.println("l_new after the adding of the number, with "+
                                //        "lastNumber = "+lastNumber+" and newNumber "+newNumber);

            for (Iterator ite = l_toControl.iterator(); ite.hasNext(); ) {
                l_new.add(newNumber.toString()+"."+(String)ite.next());

                            // for testing and tracking
                            //if (lastNumber == 99 || lastNumber == 100)
                              //  System.out.println("---------- "+l_new.getLast());
            } // for ite
        } // else
    }
    else { // l_old is null  
        // adds the new values and the control message
        if (control != null) {
            l_new.addAll(Arrays.asList(_l_values) );
            l_new.add(control);

                            // for testing and tracking
                            System.out.println("From ACSFactoryImpl.addAssociations(), "+
                                    "the first values set for _criteria "+_criteria+
                                    " has a non-null control: "+control);

        } else {
            // adds the new values starting with the number '1'.
            // _l_values may be null.
            newNumber = 1;

                            // for testing and tracking
                            //System.out.println("From ACSFactoryImpl.addAssociations(), "+
                              //      "the first values set for _criteria "+_criteria+
                                //    " has a null control and the numbered values are: ");

            for (Iterator ite = Arrays.asList(_l_values).iterator(); ite.hasNext(); ) {
                l_new.add(newNumber.toString()+"."+(String)ite.next());

                            // for testing and tracking
                            //System.out.println("---------- "+l_new.getLast());
            } // for ite            
        } // else
    } // else

    if (!l_new.isEmpty())
        _m_upd_Associations.put(
            _criteria, (String[])l_new.toArray(new String[0])); 
    else
        _m_upd_Associations.put( _criteria, new String[0]); 
    
    l_old = null;
    l_filtered = null;
    l_toControl = null;
    l_new = null;
    control = null;
    return _m_upd_Associations;
  } // addAssociations()

  /**
   * Removes some associations to handle at the creation or the deletion of ACSObjects.
   * See getM_AssociationsOnNewObject() for a full description of the formats.
   * Does not fire a property change.
   * The criteria is not removed from the map if _l_values contains all the current values
   * which is associated to the criteria.
   * After the removing, the set of values is controlled and an error message is
   * added if the set is not compliant. A previous error message may be removed
   * ('Error:' or 'Error to delete:'), and also the action numbers.
   * Called by removeAssociationsToNewObject(), removeOpeningAndClosingRoots(),
   * removeOpeningNoClosingRoots() and removeIntegrityRoots().
   * @param _m_upd_Associations actions to update. Never null nor empty,
   * and the map value for _criteria is never null.
   * @param _criteria is the map key to apply these associations. Never null.
   * @param _l_values a Strings array which contain the values for the criteria,
   *    with an action number 'NNN.' or not. Never null. May be empty.
   * @return updated _m_upd_Associations
   * @exception UpDateError if:<P>
   *    - the criteria is null or not known, <P>
   *    - _l_values is null or one value is null <P>
   *    - a value is unknown <P>
   *    - items in _l_lvalues are not in a sequence in the current values array.
   */
  protected synchronized HashMap<String, String[]> removeAssociations(
          HashMap<String, String[]> _m_upd_Associations,
          String _criteria, String[] _l_values) throws UpDateError {
        
                    // for testing and tracking
                    //System.out.println("From ACSFactoryImpl.removeAssociations(), "+
                      //      "values to remove for "+_criteria);
    
    boolean error = false;
    
    for (int i = 0; i < _l_values.length; i++) {
      if (_l_values[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_values[i]);          
      }
    } // for i
    if (error) {
      throw new UpDateError(" A value is null.");
    }     
    
    // variables
    String[]            l_current = _m_upd_Associations.get(_criteria);
    LinkedList<String>  list = new LinkedList( Arrays.asList(l_current) );
    Integer             lastNumber = -1;
    boolean             toAdd = false;
    String[]            l_filtered = null;
    LinkedList<String>  l_toControl = null;
    Set<String>         l_numbersToRemove = new HashSet(5);
    Set<String>         l_numbersNotToControl = new HashSet(5);
    String              control = null; 
    String              value = null;    
    short               count = 0;
    boolean             found = false;
    int                 firstValueIndex = -1;
    int                 index = -1;
    
    firstValueIndex = list.indexOf(_l_values[0]);
    if (firstValueIndex == -1)
        throw new UpDateError(" The first value to remove is not found.");

    // removes the argument items in list
    for (int  i = 0 ; i < _l_values.length ; i++) {
        if (firstValueIndex + count == list.indexOf(_l_values[i])) {
            found = true;
            // get the number N if _l_values[i] starts with 'N.'
            if (_l_values[i].substring(1, 2).equals("."))
                l_numbersToRemove.add(_l_values[i].substring(0, 1));
            else if (_l_values[i].substring(2, 3).equals("."))
                l_numbersToRemove.add(_l_values[i].substring(0, 2));
            else if (_l_values[i].substring(3, 4).equals("."))
                l_numbersToRemove.add(_l_values[i].substring(0, 3));
        } // if
        
        if (!found) {
            error = true;
            break;
        } // if        
        
        list.remove(_l_values[i]);
        //count++; // no, since list.remove(_l_values[i])
        found = false;
    } // for i

    if (error) {
        l_current = null;
        list = null;
        l_numbersToRemove = null;
        
                // for testing
                //System.out.println("ACSFactoryImpl.removeAssociations(), "+
                  //                "UpDateError with _l_values: ");
                //for (int  i = 0 ; i < _l_values.length ; i++) {
                  //  System.out.println("------ at "+i+": "+_l_values[i]);
                //}
        
        throw new UpDateError(" A value to remove is not found,"+
                "or it is not just after the previous item to remove. "+
                "First index of the first value to remove: "+firstValueIndex);
    } // if
    
    // removes the number 'N.' in the argument when 'N' is in l_numbersToRemove.
    // the order of items is not changed.
    found = true;
    while (!l_numbersToRemove.isEmpty() && found) {
        found = false;
        for (Iterator ite = list.iterator(); ite.hasNext(); ) {
            value = (String)ite.next();
            if (l_numbersToRemove.contains(value.substring(0, 1)) ||
                    l_numbersToRemove.contains(value.substring(0, 2))) {
                index = list.indexOf(value);
                ite.remove();
                found = true;
                break;
            }
        } // for ite
        if (found) {
            if (value.substring(1, 2).equals("."))
                list.add(index, value.substring(2));
            else list.add(index, value.substring(3));
        } 
    } // while
    
    // the removing is validated, but the new set of values has to be controlled,
    // commented if there is an error. All the actions having the action number
    // of an argument, have their number removed.
    if (!list.isEmpty()) {
        l_toControl = new LinkedList();
        // controls the set of values for the last action - updates of list has not
        // updated this array
        l_filtered = (String[]) AssociateUtilityImpl.getM_FilteredAssociations(
                _m_upd_Associations, true).get(_criteria);

                              // for testing and tracking
                        //System.out.println("ACSFactoryImpl.removeAssociations(), "+
                          //"list is not null: ");
                        //for (Iterator ite = list.iterator(); ite.hasNext(); ) {
                          //  System.out.println("------ "+ite.next());
                        //} // for ite
                        //if (l_filtered != null) {
                          //  System.out.println("and l_filtered not null: ");
                            //for (int i = 0; i < l_filtered.length; i++) {
                              //  System.out.println("------ at "+i+": "+l_filtered[i]);
                            //} // for i
                        //}
                        //for (Iterator ite = l_numbersToRemove.iterator(); ite.hasNext(); ) {
                          //  System.out.println("------ "+ite.next());
                        //} // for ite
                        
        if (l_filtered != null && l_filtered.length != 0) {
            if ( l_filtered[l_filtered.length - 1].substring(2, 3).equals(".") )
                lastNumber = new Integer(l_filtered[l_filtered.length - 1].substring(0, 2));
            else lastNumber = new Integer(l_filtered[l_filtered.length - 1].substring(0, 1));            
        } // else lastNumber == -1
                        
        if (lastNumber != -1) {
            if ( l_numbersToRemove.contains(lastNumber.toString()) )
                index = lastNumber.intValue();
            else
                index = lastNumber.intValue() + 1;
            
            for (int i = 0; i < index; i++) {
                l_numbersNotToControl.add((new Integer(i)).toString());
            } // for i
            
                            // for testing and tracking
                        //System.out.println("lastNumber: "+lastNumber);
        } // if

        if (lastNumber != -1 && !l_numbersToRemove.contains(lastNumber.toString())) {
            // puts in l_toControl the last no-number values in list,
            // without the errors which are also removed from list
            for (Iterator ite = list.iterator(); ite.hasNext(); ) {
                value = (String)ite.next();
                if (value.startsWith(lastNumber.toString()+".")) {
                    toAdd = true;
                    continue;
                } 
                if (toAdd && !value.startsWith(lastNumber.toString()+".") && 
                        !value.startsWith("Error")) 
                    l_toControl.add(value);
                if (value.startsWith("Error")) ite.remove();
            } // for ite
        } // if
        else {
            // puts in l_toControl all the (no-number) values in list,
            // without the errors which are also removed from list,
            // and without the validated numbered values which stay in list
            for (Iterator ite = list.iterator(); ite.hasNext(); ) {
                value = (String)ite.next();
                if (!value.startsWith("Error")) {
                    if (!l_numbersNotToControl.contains(value.substring(0, 1)) &&
                            !l_numbersNotToControl.contains(value.substring(0, 2))) {
                        l_toControl.add(value);
                    } // if          
                } // if
                // starts with 'Error'
                else ite.remove();
            } // for ite            
        } // else

        // the returned value starts with 'Error ' if not null
        if (!l_toControl.isEmpty()) {
            control = AssociateUtilityImpl.controlAssociations(
                this,
                _m_upd_Associations,
                _criteria, (String[])l_toControl.toArray(new String[0]), false);
        
            // adds the new control message
            if (control != null) list.add(control);
        } // if (l_toControl
    } // if (!list.isEmpty())
    
    if (!list.isEmpty()) {
        _m_upd_Associations.put(
                _criteria, 
                (String[]) list.toArray(new String[list.size()]));
    } else { // if _l_values contains all the current values.
        _m_upd_Associations.put(
                _criteria, new String[0]);
    }
    
    l_current = null;
    list = null;
    l_filtered = null;
    l_toControl = null;
    control = null; 
    value = null;
    l_numbersToRemove = null;
    l_numbersNotToControl = null;
    return _m_upd_Associations;
  } // removeAssociations()

} // ACSFactoryImpl