//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.HashMap;

/**
 * This empty class is responsible for modeling a MySQL 5 server as an access control system (ACS).
 * It ensures the modelization of a MySQL 5 server, simply to change the ACS BeanInfo class that
 * has to provide a property descriptor for the VirtualFolders.
 * All the work is done by the super class.
 * It is a javabean without proper properties.
 * @see ARoad0.gBase.BaseManagerImpl
 * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.ACSFactoryMySQLImpl
 * @version 0.7.0
 */
public class ACSMySQLImpl extends ACSImpl implements AcsAddon {

    /**
    * Default constructor used by valueOf() and the subclasses
    */
    public ACSMySQLImpl() {
    }

    /**
    * Factory method to use for creating an instance.
    * Called by ACSFactoryMySQLImpl.created().
    * Does not register the new instance in BaseManagerImpl.
    * Calls ACSImpl.valueOfForGenericBase() that makes the main work.
    * @param _iSname is the IS name. May be null.
    * @param _parentName is the name of the parent node in the IS nametree.
    *        No use if _iSname is null. May be null.
    * @param _name is the acs name
    * @param _type is 'MySQL 5 server on Linux' or a derived value.
    * @param _m_structure is the map of booleans which defines the ACS structure. Not null.
    * @param _m_constraintsForStringValues is the map where
    *   the key is a String of an ACSObject interface (like Resource) ended by
    *   a bound property name (like ResourceType), with '.' as separator,
    *   and the value is a String array of the authorized values for this bound property
    *   (like {file, directory, executable, script...} for ResourceType in Linux). May be null.
    * @param _m_constraintsForNewObjects is the map where
    *   the key is a String following one of the 7 possible formats,
    *   and the value is a String array of the authorized values among 17. May be null.
    * @param _m_constraintsForRightValues is the map where
    *   the key is a String of an ACSObject interface (like Resource) ended by
    *   a bound property name (like OthersRights), with '.' as separator,
    *   and the value is a StringRight array of the authorized values for this bound property
    *   (like {d, r, w, x} for OthersRights in Unix). Never null.
    * @param _m_SpecializedRights is the map where
    *   the key is a quartet (object.domain.key.subkey)
    *   and the value is is a StringRight array of the authorized values. Never null.
    * @param _l_Rights array of acsrights and/or metarights which are workable. Never null.
    * @param _l_ImmutableRules list of the immutable rules derivated from the structure.
    *    Never null.
    * @param _m_CommentsOnProperties comments on each ACS property, to explain the simulation.
    *    May be null.
    * @param _m_AcsVocabulary defines several independant domains, like
    * how to name the constants in Gui2.ACSTree.M_ACS_TREE_UPPER_NODES for this ACS,
    * as names of the ACS nodes in the proper ACS internal frame, and,
    * if this ACS is an AcsAddon, what are the subclassed classes in the generic gBase.
    * May be null.
    * @return the new ACSMySQLImpl
    * @exception BaseError if _l_structure is not null with a lenght is not 20,
    * if there is a rule about _l_structure which is not applied, or if the ACS
    * is already known by BaseManagerImpl.
    */
    protected static ACSImpl valueOf(ImmutableName _iSname, ImmutableName _parentName,
    ImmutableName _name, String _type,
    SortedMap<String,Boolean> _m_structure,
    HashMap<String,String[]>  _m_constraintsForStringValues,
    HashMap<String,String[]>  _m_constraintsForNewObjects,
    HashMap<String,StringRight[]>  _m_constraintsForRightValues,
    HashMap<String,StringRight[]>  _m_SpecializedRights,
    StringRight[] _l_Rights, List<String> _l_ImmutableRules,
    HashMap<String,String[]>  _m_CommentsOnProperties,
    HashMap<String,String[]>  _m_AcsVocabulary) throws BaseError {

      if (_name == null) {
           throw new BaseError(" in the ACS constructor, "+
             "the ACS name is null.");
      } // if

      String message = null;

      if (_m_structure != null)
          message = ACSFactoryUtilityImpl.controlAcsStructure(_m_structure);

      if (_type == null) throw new BaseError(" in the ACS constructor, "+
        " the ACS type must be defined.");
      if (message != null) throw new BaseError(message);

      ACSMySQLImpl result = new ACSMySQLImpl();

      return valueOfForGenericBase(
        result,
        _iSname, _parentName,
        _name, _type,
        _m_structure,
        _m_constraintsForStringValues,
        _m_constraintsForNewObjects,
        _m_constraintsForRightValues,
        _m_SpecializedRights,
        _l_Rights, _l_ImmutableRules,
        _m_CommentsOnProperties,
        _m_AcsVocabulary);
    } // valueOf() constructor

    /**
    * Registers a new own resource with a type control,
    * and fires the bound properties 'M_OwnResources' or
    * 'M_ExternalResources', 'OneResource' and 'OneBaseObject' only if it is not
    * an Actor.
    * Overrides the super method and calls it without other processing.
    * @param _res is just created in a Resource constructor, or
    *   external resource which is associated to an aclEntry owned by this ACS
    * @exception CreateError (and baseError) if the Name is already known
    *  in ACSCatalog, if the Name is null, or if the resource type is not allowed
    *  by this ACS
    */
    protected void newEorResource(ImmutableResource _res) throws CreateError {
      super.newEorResource(_res);
    } // newEorResource()

    /**
    * Unregisters the own resources.
    * Fires property change events on "M_OwnResources" or "M_ExternalResources",
    * and fire "OneResource", "OneBaseObject".
    * Overrides the super method and calls it without other processing.
    * @param       _res internal or external open ImmutableResource
    *                to be taken off in the ACS.
    * @exception   UpDateError (and baseError) if the Name is unknown
    *                in the ACS.
    */
    protected void removeEorResource(ImmutableResource _res) throws UpDateError {
      super.removeEorResource(_res);
    } // removeEorResource()

    /**
    * Registers a new own eligible party in the ACS and
    * controls that the eligible party type is allowed. <P>
    * Fires a property change on "M_OwnEligibleParties" or "M_ExternalEligibleParties",
    *  "OneEligibleParty" and "OneBaseObject".
    * Overrides the super method and calls it without other processing.
    * @param       _ep is just created in EligibleParty()
    *                or in ActorImpl(), or associated to a newEorAclEntry() call
    * @exception   CreateError (and baseError) if the Name is already known
    *                in ACSCatalog, if the Name is null or if the ep type is not allowed.
    */
    protected void newEorEligibleParty(ImmutableEligibleParty _ep) throws CreateError {
      super.newEorEligibleParty(_ep);
    }

    /**
    * Removes the EligibleParty in m_EligibleParties_.
    * May updates the closed bridge targets map.
    * Fires property change on "M_OwnEligibleParties" or "M_ExternalEligibleParties",
    *  "OneEligibleParty", and "OneBaseObject" only if it is not an Actor.
    * Overrides the super method and calls it without other processing.
    * @param _ep eligible party to remove.
    * @exception UpDateError (and baseError) if the Name is unknown in the ACS.
    */
    protected void removeEorEligibleParty(ImmutableEligibleParty _ep) throws UpDateError {
      super.removeEorEligibleParty(_ep);
    }

  /**
   * Gets the specific events to listen on a Resource or a VirtualFolder,
   * its direct and indirect parents,
   * because these events should change the rights of the Resource or VirtualFolder in an ACS Addon.
   * Then, these events are listened to update the views containing these ACSObjects.
   * This method covers only the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBaseInterface.ACSObjectAddon.getAddonInnerPropertiesToListen()
   * for the other cases).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single basic object is the Resource or VirtualFolder, <P>
   * - 'ACS.method_name' to say that the list of Resources (or VirtualFolders) is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: PARENT (mandatory), <P>
   * - method_name: 'get' method name of each basic objet parent
   * that is associated, in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is 'TRUE', 'FALSE', 'NULL' or 'NOT-NULL', to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map key starts with 'ACS', the ACS method to call has no argument.
   * When the map value starts with 'PARENT', the parent method to call has no argument.
   * Called by ARoad0.gBase.EPRViewInBaseImpl, NoThanViewInBaseImpl
   * and ARoad0.gWork.RightsMediatorImpl.
   * For instance, 'PARENT.getAGOAccessFromTheParentOther IF FALSE' is the value
   * the ACS Addon Ubuntu ACS returns for the key ACS.getEorL_FromOtherInCurrentContextActors,
   * to request the listening of the 'FromTheParentOther' event on all the indirect parents
   * of a ResourceUbuntu in a view.
   * @return map to define the specific ACS Addon events to listen. May be null.
   */
  public Map<String, String> getAddonParentPropertiesToListen() {
    return null;
  } // getAddonInheritedPropertiesToListen()

} // ACSMySQLImpl


