//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.Gui1.Desktop;
import ARoad0.ErrW.DialogError;

import java.util.List;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.SortedMap;
import java.util.HashMap;
import java.util.Map;
import java.util.HashSet;
import java.util.Set;
import java.util.Collections;
import java.util.Iterator;
import java.util.Arrays;
import java.io.Serializable;
import java.io.ObjectStreamField;

//import java.beans.PropertyChangeSupport;
//import javax.swing.event.SwingPropertyChangeSupport;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import javax.swing.ImageIcon;
import javax.swing.Icon;
import java.net.MalformedURLException;

/**
 * This class is responsible for modeling an Access Control System (ACS) which
 * is the main unit of modelization for software.
 * It registers its internal and external resources,
 * its internal and external eligible parties (accounts, groups, actors),
 * its internal and external virtual folders,
 * its AclEntries, its Privileges and its ACS bridges.
 * It also registers the other ACS which control this ACS.
 * It cuts or restores in its objects, the external references to other ACS
 * when it is closed or open.
 * After ACSFactoryImpl, this is the most complex class of the gBase package.
 * <P>
 * The ACS is created by ACSFactoryImpl.createACS(), which set the booleans and string
 * parameters for the new ACS. They cannot be changed after the ACS creation.
 * They define the ACS general structure and behavior. At the ACS creation,
 * the compliance to varied integrity rules is verified.
 * <P>
 * It is a javabean with the following bound properties:
 * 'ExternalOrientedPrivilege', 'Controlled', 'ExternalOriented', 'OneController',
 * 'M_OwnResources', 'OneExternalResource', 'OneResource' (when one resource is created or deleted),
 * 'OneEligibleParty', 'OneBaseObject' for a Resource, VirtualFolder or EligibleParty,
 * 'AclEntry', 'M_OwnEligibleParties', 'OneExternalEligibleParty',
 * 'OwnResourcesRootName', 'OwnGroupsRootName', 'OwnResourceFoldersRootName', 
 * 'OwnEPFoldersRootName', 'BridgeSources', 'BridgeTargets', 'Comment',
 * 'OneAcsRight', 'OneNoRightProxyActor', 'Privileges',
 * 'TiedChildACSs', 'RootsForChildACS', 'OneTypedPrivilege', 'OneLinkedPrivilege',
 * 'OneResourceVirtualFolder', 'OneEPVirtualFolder', 'IncompleteStructureModeling'.
 * <P>
 * The ACSImpl class is a tricky case for the thread management.
 * There is very few direct accesses to ACS instances in the access paths search algorithms.
 * This is recommanded to do so in the AcsAddons.
 * For the rare ACS methods that are used or should be used by a view,
 * the method documentation includes the comment 'may be called from a view worker thread'.
 * These methods are fast, as far as possible. Furthermore, all the ACS new/remove methods
 * on the classes implementing Resource, EligibleParty, VirtualFolder, AclEntry
 * or Privilege are synchronized.
 * <P>
 * The listeners are proper to one ACSImpl instance, outside gBase,
 * transient (not serialized), and they are called in any order, excepted for:<P>
 * 1/ The sequence for a Resource change is : ExternalOrientedAclEntry,
 * ExternalOriented, M_OwnResources (or External), OneResource, OneBaseObject.
 * It is very similar for an EligibleParty change and a VirtualFolder change.<P>
 * 2/ For Actors, 'OneResource' is always fired when 'OneEligibleParty' is fired,
 *     and it is fired before this event for a resource creation,
 *     and after this first event for a resource deletion.
 *  'OneBaseObject' is fired when the operation is completed,
 *     and it is fired always just after 'OneEligibleParty' if it is an actor creation,
 *     and always just after 'OneResource' if it is an actor deletion.<P>
 * Excepting for 'OneResource', 'OneEligibleParty', 'OneBaseObject', 'AclEntry',
 * 'OneTypedPrivilege', 'OneLinkedPrivilege', 'OneVirtualFolder',
 * the listeners receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user. <P>
 * This class has several subclasses in the AcsAddon packages.
 * Most of the variables are protected.
 * @see ARoad0.gBase.BaseManagerImpl
 * @version 0.7.1
 */
public class ACSImpl implements ACS, BoundBean, Serializable {
  // public attributes
  public final static String  LINUX = "Linux";
  public final static String  GENERIC = "Generic";
    
  // protected attributes
  protected static final long serialVersionUID__ = 0;
  protected static final int  PRIME = 1000003;
  protected static final int  INITIAL_CAPACITY_25 = 25; // default value for map: 101
  protected static final int  INITIAL_CAPACITY_5 = 5;
  protected static final int  INCREMENT_CAPACITY = 10;
  /*
   * To limit the number of typed privileges. Applicable only to ACS which are
   * not authorization servers.
   */
  protected static final int  MAX_FOR_TYPED_PRIVILEGES_NUMBER = 50;

  /**
   * Name of the IS which owns this ACS
   */
  protected ImmutableName       iSname_;
  /**
   * Name of the acs, which starts with the IS name
   */
  protected ImmutableName       name_;
  /**
   * Name of the parent acs in the IS nametree. May be null.
   */
  protected ImmutableName       parentName_;
  /**
   * BaseError for this acs
   */
  protected BaseError           baseError_; 
  /**
   * Map of all the resources of this ACS
   */
  protected final SortedMap<ImmutableName, ImmutableResource>  m_Resources_ = new TreeMap(); // only open resources
  /**
   * Map of all the eligible parties of this ACS
   */
  protected final SortedMap<ImmutableName, ImmutableEligibleParty>  m_EligibleParties_ = new TreeMap();
     /**
   * Map of mandatory associations on the ACS new objects.
   */
  protected HashMap<String, String[]>  m_ConstraintsForNewObjects_ =
          new HashMap<String,String[]>(INITIAL_CAPACITY_25);
  /**
   * Map of restrictions on the ACS types and the GUI creation/display policy
   */
  protected HashMap<String, String[]>  m_TypesAndGUIPolicy_ =
          new HashMap<String,String[]>(INITIAL_CAPACITY_25);
  /**
   * Map of restrictions on the ACS GroupID/Node policy
   */
  protected HashMap<String, String[]>  m_TypeOrientedPolicy_ =
          new HashMap<String,String[]>(INITIAL_CAPACITY_25);  
  /**
   * Map of restrictions on the ACS Right properties
   */
  protected HashMap<String, StringRight[]>  m_StandardRights_ =
          new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25);
  /**
  * To extend or to reduce the allowed AGO, ACL, PRI rights following the 
  * target/source interface or type
  */
  protected HashMap<String, StringRight[]>  m_SpecializedRights_ =
          new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25);
  /**
   * List of the ACS rules which are derivated from the structure, like
   * 'AGO right: A rights overlay G rights'.
   */
  protected List<String>        l_ImmutableRules_ = null;
  /**
   * true if some specialized rights have been set
   */
  protected boolean             haveSpecializedRights_ = false;
  /**
   * All the open aclentries of this ACS
   */
  protected ImmutableAclEntry[] l_AclEntry_;
  protected ImmutableName       ownResourcesRootName_;
  protected ImmutableName       ownGroupsRootName_;
  protected ImmutableName       ownResourceFoldersRootName_;
  protected ImmutableName       ownEPFoldersRootName_;
  /**
   * The icon of the aCS in the beamer - selected at the creation
   *  or the GUI uses the default value.
   */
  protected ImageIcon           icon_;
  protected String              iconName_;
  protected String              editor_;
  protected boolean             is_hub_ = false;
  protected String              visibleSpace_ = NameISTreeImpl.ALL_THE_IS_SPACE;
  
  /**
   * the open controllers of this acs
   */
  protected ImmutableName[]     l_OpenAcsControllerNames_;
  protected String              comment_;

  /**
   * all the (ACS or meta) rights managed in this acs
   */
  protected StringRight[]       l_Rights_;
  /**
   * Array of all the ACS privileges
   */
  protected Privilege[]         l_Privilege_;
  /**
   * Array of all the ACS virtual folders
   */
  protected VirtualFolder[]     l_VirtualFolders_;
  /**
   * The possible specialized right types are 'AccRightsSet',
   * 'GpORightsSet', 'AclRightsSet', 'PrvRightsSet', 'BdgRightsSet'.
   */
  protected Set<String>        categoriesInSpecializedRights_;
  
    /**
     * Defines the ACS type, which may be GENERIC, LINUX, WINDOWS, LDAP...
     * or a user-defined type.
     * ACSImpl.GENERIC is the default value. 
     */
  protected String              aCSType_ = ACSImpl.GENERIC;
  /**
     * True if this ACS has open controllers. They are others ACS which have
     * external AclEntries or Privileges handling targets (not sources) from this ACS.
     */
  protected boolean             controlled_ = false;
    /**
     * if true, is a controller for other open ACS - that is the ACS
     * has some open external AclEntries or Privileges.
     */
  protected boolean             externalOriented_ = false;
    /**
     * Set of the ACS names of both controlled ACS and referenced ACS
     * that control this ACS, which are in the closed acl entries
     */
  protected Set<ImmutableName>  l_ClosedAclEntriesACSNames_ = null;
    /**
     * Set of only closed external aclEntries, that is those which have their
     * ACS that is currently closed.
     * BaseManagerImpl.importACS() handles directly this property.
     */
  protected Set<ImmutableAclEntry>       l_ClosedAclEntries_ = null;
    /**
     * Set of the ACS names of both controlled ACS and referenced ACS
     * that control this ACS, which are in the closed privileges
     */
  protected Set<ImmutableName>  l_ClosedPrivilegesACSNames_ = null;
    /**
     * Set of only closed external privileges, that is those which have their
     * ACS that is currently closed.
     * BaseManagerImpl.importACS() handles directly this property.
     */
  protected Set<PrivilegeForLinks>       l_ClosedPrivileges_ = null;
     /**
     * Map of the closed external Actor names
     * which are targets of own actor bridge sources.
     * The key is the own actor bridge source name. May be null.
     * BaseManagerImpl.importACS() handles directly this property.
     */
  protected SortedMap<ImmutableName, ImmutableName>
          m_closedTargetsOfActorBridgeSources_  = null;
    /**
     * Map of the closed external Actor names
     * which are sources of own actor bridge targets.
     * The key is the own actor bridge target name. May be null.
     * BaseManagerImpl.importACS() handles directly this property.
     */
  protected SortedMap<ImmutableName, ArrayList<ImmutableName>>
          m_closedSourcesOfActorBridgeTargets_  = null;
  /**
  * The active booleans which define the ACS structure and features.
   * The name of the parameter is the key, and the Boolean value indicates if
   * the parameter is managed by this ACS. By default, the parameter
   * is not in this map as key, and the value to consider is then false.
   * Some complementary, non-generic values are allowed for the AcsAddons.
   * This map may be directly updated at the creation of an AcsAddon ACS,
   * for adding new keys of non-generic immutable properties that are ridden
   * by their proper AcsAddon classes.
   */
  protected SortedMap<String, Boolean>  m_Structure_ = new TreeMap<String,Boolean>();
  
  /**
   * A child tied ACS has to be opened when this instance is opened.
   * Like any child ACS, it has to be closed when this instance is closed, and 
   * it may be closed alone.
   */
  protected ArrayList<ImmutableName> l_TiedChildACSs_ = null;

  /**
   * Gets the subACS names and their mode (for Resources or for GroupIDs) and the
   * root names for groups trees, as nodes to display in the explorer.
   * A subACS is always open with its ACS.
   * In the returned value, the key is the subACS name or a group root name, in one word,
   * and the value is an array which may contain one or two items among the following values:
   * <ResourceSubACS>, <GroupSubACS> and <GroupTree>.
   */
  protected HashMap<String, String[]>   m_SubAndGroupPolicy_ = null;

  /**
   * Immutable map which says if the ACS or a subACS belongs to a Composite ACS 
   * in the IS Name Tree. <P>
   * In the returned value, the key is the subACS name, in one word, 
   * or the ACS name last component, and the value is the Composite name.
   */
  protected HashMap<String, ImmutableName>  m_BelongsToComposite_ = null;

  /**
   * For a child ACS, the list of the ACSObject roots in the parent ACS which 
   * have a root relation with this child ACS. <P>
   * This relation set that the root parent ACSObject, or one of its property value, 
   * has to be there when the child ACS is opened, to ensure its ability to exist,
   * and has to be closed or updated (for a property value) at its closing.
   */
  protected HashMap<String, String[]> m_RootCommandsForAcsOpeningAndClosing_ = null;

  /**
   * For a child ACS, the list of the ACSObject roots in the parent ACS which 
   * have a root relation with this child ACS. <P>
   * This relation set that the root parent ACSObject, or one of its property value, 
   * has to be there when the child ACS is opened, to ensure its ability to exist,
   * but there is NO constraint at its closing.
   */
  protected HashMap<String, String[]> m_RootCommandsForAcsOpeningNoClosing_ = null;

  /**
   * Get for a child ACS, the list of the ACSObject roots in the parent ACS which 
   * have a root relation with this child ACS. <P>
   * This relation set that the root parent ACSObject, or one of its property value, 
   * has to be there when the child ACS is opened, to ensure its ability to work well, 
   * in other words, following its specifications.
   */
  protected HashMap<String, String[]> m_RootCommandsForAcsIntegrity_ = null;

  /**
   * The map of the ACSObject roots which have with an opened or closed child ACS,
   * a relation among those set by ACSFactoryUtilityImpl.getKeysForConstrainedChildACSMap().
   * These relations set that the deleting or the updating of the ACSObject will
   * stop the child ACS or will modify its ability to work well.
   * The key is the child ACS name in its String form, and the value is
   * an array of the root names in this ACS.
   */ 
  protected Map<String, ImmutableName[]> m_RootsForChildACS_ = null;

  /**
   * Map of the comments on each ACS property, to explain the simulation.
   * Used with add/removeCommentsOnProperties().
   */
  protected HashMap<String, String[]>  m_CommentsOnProperties_ = null;

  /**
   * The list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root remains in the parent ACS at the closing of this ACS.
   */
  protected List <ImmutableName> l_RootNamesForAcsOpeningNoClosing_ = null;

  /**
   * The ACS vocabulary defines several independant domains: <P>
   * - How to name the L_ACS_TREE_GENERIC_UPPER_NODES constants for this ACS,
   * as names of the ACS tree nodes in the proper ACS internal frame. An example
   * is to replace the tree node 'Actors (right user target)' by 'Scripts'. <P>
   * - If the ACS is an AcsAddon, what are the generic gBase classes that are subclassed
   * in the AcsAddon gBase package (see L_CLASSIC_CLASS_KEYS), like when
   * the Addon Ubuntu subclasses ResourceImpl with ResourceUbuntuImpl. <P>
   * - If the ACS is an AcsAddon, what are the specific new classes in
   * the AcsAddon gBase package (see ACSADDON_NEW_CLASSES). In this case,
   * the class names are separated by one space, like in
   * 'ExecutableUbuntu XXXXUbuntu YYYYUbuntu'.
   * The default value is ACSFactoryUtilityImpl.M_GENERIC_ACS_VOCABULARY.
   * @see ACSFactoryUtilityImpl#M_GENERIC_ACS_VOCABULARY
   * @see ACSFactoryUtilityImpl#L_ACS_TREE_GENERIC_UPPER_NODES
   * @see ACSFactoryUtilityImpl#L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS
   * @see ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ACSFactoryUtilityImpl#ACSADDON_EXTENSION_CLASSES
   */
  protected HashMap<String,String[]> m_AcsVocabulary_ = null;

  // ------------------- TRANSIENT VARIABLES ------------------------------------
    /**
     * if true, contains at most one actor which is a bridge target for another-ACS actor.
     **/
  protected transient boolean             containsBridgeTarget_ = false;
    /**
     *if true, contains at most one actor which is a bridge source for another-ACS actor.
     **/
  protected transient boolean             containsBridgeSource_ = false;
    /**
     * if true, contains some root actions and then, some ACS roots in other ACS.
     **/
  protected transient boolean             containsRootActions_ = false;
  /**
     * Map (ImmutableName, Actor) of own Actors which are bridge targets
     * for external actor sources. Used only in the two methods removeOwnActorBridgeTarget.
     * The key is the external actor bridge source name.
     */
  protected transient Map<ImmutableName, ImmutableActor>  m_ActorBridgeTargets_ = null;
    /**
     * List of own Actors which are bridge targets for external actors. One actor
     * may be the target of several sources. Synchronized with m_ActorBridgeTargets_.
     */
  protected transient List<ImmutableActor>         l_ActorBridgeTargets_ = null;
  /**
     * List of own Actors which are bridge sources for external actors.
     */
  protected transient List<ImmutableActor>         l_ActorBridgeSources_ = null;
  /**
   * List of own Actors which are not right proxies, then which are
   * secure and without any trojan able to deliver hidden rights to the right users
   * executing these actors.
   * The default state of right proxy is based on no confidence in the security
   * of actor code, because no code failure is an exception. Thus, setting an actor
   * as been not a right proxy is an important user choice to do with caution.
   */
  protected transient List<ImmutableActor>         l_NoRightProxyActors_ = null;
   /**
   * manage all the property change listeners
   */
  protected transient PropertyChangeSupport changeSupport_ = null;

    /**
   * The list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root has to be in the parent ACS at the opening of this ACS, and it is
   * deleted at the closing of the ACS.
   */
  protected transient List<ImmutableName> l_RootNamesForAcsOpeningWithClosing_ = null;

    /**
   * The list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root in the parent ACS has some property updatings which may weaken
   * the integrity of this ACS.
   */
  protected transient List<ImmutableName> l_RootNamesForAcsIntegrity_ = null;

  /**
   * Gets the ACS rights and the metarights this ACS handles and for a given internal key value
   * of the associated metaright. The map key is the metaright internal key.
   * This map is always filled up by fillRightsOfInternalKey(), called by
   * BaseManagerImpl.restoreACSInternalLinks() at each opening and ACS import,
   * and by valueOfForGenericBase() at the ACS creation.
   */
  protected transient SortedMap<String, StringRight[]> m_RightsOfInternalKey_ = null;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ACSImpl EMPTY_INSTANCE = new ACSImpl();

/**
 * variable for the JDK 2 serialization
 * @serial ("serialVersionUID",Long.TYPE), ("iSname_",NameImpl.class),<P>
 * ("name_",NameImpl.class), ("parentName_",NameImpl.class), <P>
 * ("l_OpenAcsControllerNames_",NameImpl[].class),<P>
 * ("controlled_",Boolean.TYPE),<P>
 * ("m_Resources_",TreeMap.class), ("m_EligibleParties_",TreeMap.class),<P>
 * ("ownResourcesRootName_",NameImpl.class), ("ownGroupsRootName_",NameImpl.class),<P>
 * ("ownResourceFoldersRootName_",NameImpl.class),<P>
 * ("ownEPFoldersRootName_",NameImpl.class), ("baseError_",BaseError.class),<P>
 * ("l_ClosedAclEntriesACSNames_",HashSet.class), ("l_ClosedAclEntries_",HashSet.class),<P>
 * ("l_AclEntry_",AclEntry[].class), ("externalOriented_",Boolean.TYPE),<P>
 * ("m_closedTargetsOfActorBridgeSources_",TreeMap.class),<P>
 * ("m_closedSourcesOfActorBridgeTargets_",TreeMap.class), <P>
 * ("m_Structure_",TreeMap<String,Boolean>.class),<P> 
 * ("m_TypesAndGUIPolicy_",HashMap<String,String[]>.class),<P>
 * ("m_TypeOrientedPolicy_",HashMap<String,String[]>.class),<P>
 * ("m_ConstraintsForNewObjects_",HashMap<String,String[]>.class),<P>
 * ("m_StandardRights_",HashMap<String,StringRight[]>.class),<P> 
 * ("m_SpecializedRights_",HashMap<String,StringRight[]>.class),<P>
 * ("categoriesInSpecializedRights_",HashSet.class),<P>
 * ("aCSType_",String.class),<P>
 * ("l_Rights_",StringRight[].class) <P>
 * ("icon_",Icon.class) <P>
 * ("l_ImmutableRules_",List.class) <P>
 * ("l_TiedChildACSs_",ArrayList<ImmutableName>.class) <P>
 * ("m_SubAndGroupPolicy_",HashMap<String, String[]>.class) <P>
 * ("m_BelongsToComposite_",HashMap<String, ImmutableName>.class) <P>
 * ("m_RootCommandsForAcsOpeningAndClosing_",HashMap<String, String[]>.class) <P>
 * ("m_RootCommandsForAcsOpeningNoClosing_",HashMap<String, String[]>.class) <P>
 * ("m_RootCommandsForAcsIntegrity_",HashMap<String, String[]>.class) <P>
 * ("m_RootsForChildACS_",HashMap<String, ImmutableName[]>.class) <P>
 * ("m_CommentsOnProperties_",HashMap<String, String[]>.class) <P>
 * ("l_RootNamesForAcsOpeningNoClosing_",ArrayList.class) <P>
 * ("l_Privilege_",Privilege[].class) <P>
 * ("l_VirtualFolders_",VirtualFolder[].class) <P>
 * ("m_AcsVocabulary_",HashMap.class) <P>
 * ("editor_",String.class) <P>
 */
 public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("serialVersionUID__",Long.TYPE),
      new ObjectStreamField("iSname_",NameImpl.class),
      new ObjectStreamField("name_",NameImpl.class),
      new ObjectStreamField("parentName_",NameImpl.class),              
      new ObjectStreamField("l_OpenAcsControllerNames_",NameImpl[].class),
      new ObjectStreamField("controlled_",Boolean.TYPE),
      new ObjectStreamField("aCSType_",String.class),         
      new ObjectStreamField("m_Resources_",TreeMap.class),
      new ObjectStreamField("m_EligibleParties_",TreeMap.class),
      new ObjectStreamField("ownResourcesRootName_",NameImpl.class),
      new ObjectStreamField("ownGroupsRootName_",NameImpl.class),
      new ObjectStreamField("ownResourceFoldersRootName_",NameImpl.class),
      new ObjectStreamField("ownEPFoldersRootName_",NameImpl.class),      
      new ObjectStreamField("baseError_",BaseError.class),
      new ObjectStreamField("l_ClosedAclEntriesACSNames_",HashSet.class),
      new ObjectStreamField("l_ClosedAclEntries_",HashSet.class),
      new ObjectStreamField("l_AclEntry_",AclEntry[].class),
      new ObjectStreamField("externalOriented_",Boolean.TYPE),
      new ObjectStreamField("m_closedTargetsOfActorBridgeSources_",TreeMap.class),
      new ObjectStreamField("m_closedSourcesOfActorBridgeTargets_",TreeMap.class),
      new ObjectStreamField("l_ACSRulesForBasics_",Boolean[].class), 
      new ObjectStreamField("l_ACSFirstRulesForRights_",Boolean[].class),
      new ObjectStreamField("l_ACSSecondRulesForRights_",Boolean[].class),
      new ObjectStreamField("m_Structure_",TreeMap.class),             
      new ObjectStreamField("m_TypesAndGUIPolicy_",HashMap.class), 
      new ObjectStreamField("m_TypeOrientedPolicy_",HashMap.class),
      new ObjectStreamField("m_ConstraintsForNewObjects_",HashMap.class),
      new ObjectStreamField("m_StandardRights_",HashMap.class),
      new ObjectStreamField("m_SpecializedRights_",HashMap.class),
      new ObjectStreamField("categoriesInSpecializedRights_",HashSet.class),
      new ObjectStreamField("comment_",String.class),
      new ObjectStreamField("l_Rights_",StringRight[].class), 
      new ObjectStreamField("icon_",Icon.class),
      new ObjectStreamField("l_ImmutableRules_",List.class),
      new ObjectStreamField("l_TiedChildACSs_",ArrayList.class),
      new ObjectStreamField("m_SubAndGroupPolicy_",HashMap.class),
      new ObjectStreamField("m_BelongsToComposite_",HashMap.class),
      new ObjectStreamField("m_RootCommandsForAcsOpeningAndClosing_",HashMap.class),
      new ObjectStreamField("m_RootCommandsForAcsOpeningNoClosing_",HashMap.class),
      new ObjectStreamField("m_RootCommandsForAcsIntegrity_",HashMap.class),
      new ObjectStreamField("m_RootsForChildACS_",HashMap.class),
      new ObjectStreamField("m_CommentsOnProperties_",HashMap.class),
      new ObjectStreamField("l_RootNamesForAcsOpeningNoClosing_",ArrayList.class),
      new ObjectStreamField("l_Privilege_",Privilege[].class),
      new ObjectStreamField("l_VirtualFolders_",VirtualFolder[].class),
      new ObjectStreamField("m_AcsVocabulary_",HashMap.class),
      new ObjectStreamField("editor_",String.class),
      new ObjectStreamField("is_hub_",Boolean.class),
      new ObjectStreamField("visibleSpace_",String.class)
 };
 
 // 
 // ---------- CONSTRUCTORS --------------------------------
 //
 
 /**
  * Default constructor used by valueOf() and the subclasses
  */
  public ACSImpl() {
      name_ = new NameImpl();
      iSname_ = null;
      // the 0.7.0 version initializes with 'NameImpl[1]'
      // and this is why getL_AcsControllers() removes the null values.
      l_OpenAcsControllerNames_ = (ImmutableName[]) new NameImpl[0];
      baseError_ = new BaseError();
      //ownResourcesRootName_ = NameImpl.EMPTY_INSTANCE;
      ownGroupsRootName_ = NameImpl.EMPTY_INSTANCE;
      ownResourceFoldersRootName_ = NameImpl.EMPTY_INSTANCE;
      ownEPFoldersRootName_ = NameImpl.EMPTY_INSTANCE;      
      l_AclEntry_ = (ImmutableAclEntry[]) new AclEntryImpl[0];
      comment_ = "";
      l_Privilege_ = new Privilege[0];
      l_VirtualFolders_ = new VirtualFolder[0];
      l_Rights_ = new StringRight[0];
      categoriesInSpecializedRights_ = new HashSet();
      changeSupport_ = new PropertyChangeSupport(this);
    } // constructor
    
  /**
   * Factory method to use for creating an instance.
   * Called by ACSFactoryImpl.createdACS(). Does not register the new instance
   * in BaseManagerImpl.
   * @param _iSname is the IS name. May be null.
   * @param _parentName is the name of the parent node in the IS nametree.
   *        No use if _iSname is null. May be null.
   * @param _name is the acs name
   * @param _type is ACSImpl.GENERIC (default value), LINUX, another ACSImpl constant
   *   or another user-defined value. Must be non-null. No use if _l_structure is null:
   *   the type is then set to LINUX.
   * @param _is_hub the IS nametree property for the ACS
   * @param _visibleSpace the IS nametree property for the ACS
   * @param _m_structure is the map of booleans which defines the ACS structure. Is null
   *   for LINUX-type ACS.
   * @param _m_constraintsForStringValues is the map where
   *   the key is a String of an ACSObject interface (like Resource) ended by
   *   a bound property name (like ResourceType), with '.' as separator,
   *   and the value is a String array of the authorized values for this bound property
   *   (like {file, directory, executable, script...} for ResourceType). May be null.
   * @param _m_constraintsForNewObjects is the map where
   *   the key is a String following one of the 7 possible formats,
   *   and the value is a String array of the authorized values among 17. May be null.
   * @param _m_constraintsForRightValues is the map where
   *   the key is a String of an ACSObject interface (like Resource) ended by
   *   a bound property name (like the AGO GroupRights), with '.' as separator,
   *   and the value is a StringRight array of the authorized values for this bound property
   *   (like {d, r, w, x} for GroupRights). Never null.
   * @param _m_SpecializedRights is the map where
   *   the key is a quartet (object.domain.key.subkey)
   *   and the value is is a StringRight array of the authorized values. Never null.
   * @param _l_Rights array of acsrights and/or metarights which are workable. Never null. 
   * @param _l_ImmutableRules list of the immutable rules derivated from the structure.
   *    Never null.
   * @param _m_CommentsOnProperties comments on each ACS property, to explain the simulation.
   *    May be null.
   * @param _m_AcsVocabulary defines several independant domains, like
   * how to name the constants in Gui2.ACSTree.M_ACS_TREE_UPPER_NODES for this ACS,
   * as names of the ACS nodes in the proper ACS internal frame, and,
   * if this ACS is an AcsAddon, what are the subclassed classes in the generic gBase.
   * May be null.
   * @return the new ACSImpl
   * @exception BaseError if _l_structure lenght is not 20, if there is a rule about
   *   _l_structure which is not applied, or if the ACS
   *   is already known by BaseManagerImpl. 
   */
  protected static ACSImpl valueOf(ImmutableName _iSname, ImmutableName _parentName, 
    ImmutableName _name, String _type,
    boolean _is_hub, String _visibleSpace,
    SortedMap<String,Boolean> _m_structure, 
    HashMap<String,String[]>  _m_constraintsForStringValues,
    HashMap<String,String[]>  _m_constraintsForNewObjects,
    HashMap<String,StringRight[]>  _m_constraintsForRightValues,
    HashMap<String,StringRight[]>  _m_SpecializedRights,
    StringRight[] _l_Rights, List<String> _l_ImmutableRules,
    HashMap<String,String[]>  _m_CommentsOnProperties,
    HashMap<String,String[]>  _m_AcsVocabulary) throws BaseError {

      if (_name == null) {
           throw new BaseError("in the ACS constructor, "+
             "the ACS name is null.");
      } // if

      String message = null;

      if (_m_structure != null) 
          message = ACSFactoryUtilityImpl.controlAcsStructure(_m_structure);

      if (_type == null) throw new BaseError("in the ACS constructor, "+
        "the ACS type must be defined.");
      if (message != null) throw new BaseError(message);
      
      ACSImpl result = new ACSImpl();

      result.is_hub_ = _is_hub;
      result.visibleSpace_ = _visibleSpace;

      return valueOfForGenericBase(
        result,
        _iSname, _parentName,
        _name, _type,
        _m_structure,
        _m_constraintsForStringValues,
        _m_constraintsForNewObjects,
        _m_constraintsForRightValues,
        _m_SpecializedRights,
        _l_Rights, _l_ImmutableRules,
        _m_CommentsOnProperties,
        _m_AcsVocabulary);
  } // valueOf() constructor
     
  /**
   * Factory method to use for creating an instance.
   * Called only by valueOf() from this class and from the subclasses.
   * Does not register the new instance
   * in BaseManagerImpl.
   * @param _result the ACS to configure
   * @param _iSname is the IS name. May be null.
   * @param _parentName is the name of the parent node in the IS nametree.
   *        No use if _iSname is null. May be null.
   * @param _name is the acs name
   * @param _type is ACSImpl.GENERIC (default value), LINUX, another ACSImpl constant
   *   or another user-defined value. Must be non-null. No use if _l_structure is null:
   *   the type is then set to LINUX.
   * @param _m_structure is the map of booleans which defines the ACS structure. Is null
   *   for LINUX-type ACS.
   * @param _m_constraintsForStringValues is the map where
   *   the key is a String of an ACSObject interface (like Resource) ended by
   *   a bound property name (like ResourceType), with '.' as separator,
   *   and the value is a String array of the authorized values for this bound property
   *   (like {file, directory, executable, script...} for ResourceType). May be null.
   * @param _m_constraintsForNewObjects is the map where
   *   the key is a String following one of the 7 possible formats,
   *   and the value is a String array of the authorized values among 17. May be null.
   * @param _m_constraintsForRightValues is the map where
   *   the key is a String of an ACSObject interface (like Resource) ended by
   *   a bound property name (like OthersRights), with '.' as separator,
   *   and the value is a StringRight array of the authorized values for this bound property
   *   (like {d, r, w, x} for OthersRights in Unix). Never null.
   * @param _m_SpecializedRights is the map where
   *   the key is a quartet (object.domain.key.subkey)
   *   and the value is is a StringRight array of the authorized values. Never null.
   * @param _l_Rights array of acsrights and/or metarights which are workable. Never null. 
   * @param _l_ImmutableRules list of the immutable rules derivated from the structure.
   *    Never null.
   * @param _m_CommentsOnProperties comments on each ACS property, to explain the simulation.
   *    May be null.
   * @param _m_AcsVocabulary defines several independant domains, like
   * how to name the constants in Gui2.ACSTree.M_ACS_TREE_UPPER_NODES for this ACS,
   * as names of the ACS nodes in the proper ACS internal frame, and,
   * if this ACS is an AcsAddon, what are the subclassed classes in the generic gBase.
   * May be null.
   * @return the new ACSImpl
   * @exception BaseError if _l_structure lenght is not 20, if there is a rule about
   *   _l_structure which is not applied, or if the ACS
   *   is already known by BaseManagerImpl. 
   */
  protected static ACSImpl valueOfForGenericBase(
    ACSImpl _result,
    ImmutableName _iSname, ImmutableName _parentName, 
    ImmutableName _name, String _type, 
    SortedMap<String,Boolean> _m_structure, 
    HashMap<String,String[]>  _m_constraintsForStringValues,
    HashMap<String,String[]>  _m_constraintsForNewObjects,
    HashMap<String,StringRight[]>  _m_constraintsForRightValues,
    HashMap<String,StringRight[]>  _m_SpecializedRights,
    StringRight[] _l_Rights, List<String> _l_ImmutableRules,
    HashMap<String,String[]>  _m_CommentsOnProperties,
    HashMap<String,String[]>  _m_AcsVocabulary) throws BaseError {

      _result.iSname_ = _iSname;
      _result.parentName_ = _parentName;
      _result.name_ = _name;
      _result.ownResourcesRootName_ = null;
      _result.l_Privilege_ = (Privilege[]) new Privilege[0];
      _result.l_VirtualFolders_ = new VirtualFolder[0];
      _result.l_Rights_ = _l_Rights;
      _result.fillRightsOfInternalKey();
      
      _result.aCSType_ = _type;
      _result.l_ImmutableRules_ = _l_ImmutableRules;
      
      if (_m_structure != null ) {

          _result.m_Structure_ = new TreeMap(_m_structure);
          
      } else { // creates a Linux structure
          
          _result.m_Structure_.put("resource: no resources", Boolean.FALSE);
          _result.m_Structure_.put("resource: directories of resources trees", Boolean.TRUE);
          _result.m_Structure_.put("EP: accounts", Boolean.TRUE);
          _result.m_Structure_.put("EP: groups of accounts", Boolean.TRUE);
          _result.m_Structure_.put("EP: group of groups", Boolean.FALSE);
          _result.m_Structure_.put("ACL right: internal ACL", Boolean.TRUE);         
          //_result.m_Structure_.put("ACL right: external ACL with entity in this acs", Boolean.FALSE);
          //_result.m_Structure_.put("ACL right: pure external ACL without entity in this acs", Boolean.FALSE);
          _result.m_Structure_.put("resource: no actors", Boolean.FALSE);
          _result.m_Structure_.put("bridge right: no BriDGes", Boolean.FALSE);
          _result.m_Structure_.put("EP 1: secondary groups for actors as right users", Boolean.TRUE);         
          _result.m_Structure_.put("AGO right: Account and Group rights", Boolean.TRUE);
          _result.m_Structure_.put("resource: no NX executables", Boolean.FALSE); 
          
          _result.aCSType_ = LINUX;
      } // else 
      
      // fills up m_TypesAndGUIPolicy_ and m_TypeOrientedPolicy_
      Map.Entry entry = null;
      if (_m_constraintsForStringValues != null && !_m_constraintsForStringValues.isEmpty()) {
          for (Iterator ite = _m_constraintsForStringValues.entrySet().iterator(); ite.hasNext(); ) {
              entry = (Map.Entry)ite.next();
              if (entry.getKey().equals("Resource.ResourceType") ||
                      entry.getKey().equals("Directory.ResourceType") ||
                      entry.getKey().equals("Actor.ResourceType") ||
                      entry.getKey().equals("EligibleParty.EPType") ||
                      entry.getKey().equals("UserID.EPType") ||
                      entry.getKey().equals("GroupID.EPType") ||
                      entry.getKey().equals("VirtualFolder.ResourceVFType") ||
                      entry.getKey().equals("VirtualFolder.EPVFType") ||
                      entry.getKey().equals("PrivilegeForLinks.SecondType") ||
                      entry.getKey().equals("PrivilegeForType.SecondType") ||
                      entry.getKey().equals("Explorer.SubAcsNode") ||
                      entry.getKey().equals("Explorer.AcsGroupTree") ||
                      entry.getKey().equals("Explorer.NoAcsNode") ||
                      entry.getKey().equals("CreationByBeamer.NoType"))
                  _result.m_TypesAndGUIPolicy_.put(
                          (String)entry.getKey(), (String[])entry.getValue());
              else
                  _result.m_TypeOrientedPolicy_.put(
                          (String)entry.getKey(), (String[])entry.getValue());
          } // for ite
      } // if
      entry = null;
      
      if (_m_constraintsForNewObjects != null)
          _result.m_ConstraintsForNewObjects_ = _m_constraintsForNewObjects;
      
      if (_m_constraintsForRightValues != null)
          _result.m_StandardRights_ = _m_constraintsForRightValues;

       
      if (_m_SpecializedRights != null && !_m_SpecializedRights.isEmpty()) {
          _result.m_SpecializedRights_ = _m_SpecializedRights;
          _result.haveSpecializedRights_ = true;
      }
      
      // collects in m_SpecializedRights_ the current right types
      String key = null;
      if (!_result.m_SpecializedRights_.isEmpty()) { 
          for (Iterator ite = _result.m_SpecializedRights_.keySet().iterator(); ite.hasNext(); ) {
              key = (String)ite.next();
              if (key.startsWith("AccRightsSet", 7))
                  _result.categoriesInSpecializedRights_.add("AccRightsSet");
              else if (key.startsWith("GpORightsSet", 7))
                  _result.categoriesInSpecializedRights_.add("GpORightsSet");
              else if (key.startsWith("AclRightsSet", 7))
                  _result.categoriesInSpecializedRights_.add("AclRightsSet");
              else if (key.startsWith("PrvRightsSet", 7))
                  _result.categoriesInSpecializedRights_.add("PrvRightsSet");
              else if (key.startsWith("BdgRightsSet", 7))
                  _result.categoriesInSpecializedRights_.add("BdgRightsSet");
          } // for ite
      } // if 
      key = null;

      _result.m_CommentsOnProperties_ = _m_CommentsOnProperties;
      _result.m_AcsVocabulary_ = _m_AcsVocabulary;

      if (_result.m_AcsVocabulary_ != null)
          _result.editor_ = "Editor: "+
              ((String[])_result.m_AcsVocabulary_.get(ACSFactoryUtilityImpl.ACSADDON_EDITOR))[0]+
              " - AcsAddon: "+
              ((String[])_result.m_AcsVocabulary_.get(ACSFactoryUtilityImpl.ACSADDON_NAME))[0];

      return _result;
  } // valueOfForGenericBase()

  /**
   * NO USE. TO REVIEW.
  * This method creates an ACS copy with another name, without registering it.
  * This copy method breaks the copy() standard form.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy).
  * The order of copy is:<P>
  * - EligibleParties, excepted for Actors,<P>
  * - Resources (including Actors), with their aclEntries,<P>
  * - Privileges<P>
  * The GroupIDMembers are copied and registered in the copied GroupID.
  * The bridge source is copied with its link to
  * the same bridge target, and the bridge target is copied
  * with its link to a copied new source.
  * The copy does not cover the links with closed ACS,
  * both the aclEntries links and bridge links, and does not copy the associations.<P>
  * Calls BaseUtilityImpl.copyDirectoryChildrenForACS()
  * and EligibleParty.copyWithoutAcl().
  * @param _iSname is the IS name; may be null
  * @param _parentName is the name of the parent node in the IS nametree;
  *   may be null only if _iSname is null
  * @param _name of the ACS. It is copied.
  * @param _is_hub is true if the ACS is a node hub, which means that for this node,
  *    any direct child may have exchanges with any other direct child,
  *    if the rights allow them.
  * @param _visibleSpace is a NameISTreeImpl constant which defines the nodes
  *    in the IS nametree which are visible from this ACS node
  * @param _acs is the ACSImpl to copy.
  * @param _toRegisterInManager true to register the copy in BaseManagerImpl.
  * @exception BaseError if _iSname or _name is null or already known in BaseManagerImpl,
  * or if _acs is not known
  * @return ACSImpl as copy
  */
  public static ACSImpl copy(ImmutableName _iSname, ImmutableName _parentName,
    ImmutableName _name, boolean _is_hub, String _visibleSpace, ACSImpl _acs,
    boolean _toRegisterInManager) throws BaseError {

      if (_acs == null || _name == null ||
        !BaseManagerImpl.getInstance().knowsACS(_acs.getName() )) {
          throw new BaseError("in ACSImpl.copy(),"+
          " _iSname, _name or _acs is null, or _acs is not known by the base manager.");
      }
      HashMap m_StringConstraints = new HashMap(_acs.m_TypesAndGUIPolicy_);
      m_StringConstraints.putAll(_acs.m_TypeOrientedPolicy_);
      
      ACSImpl result = ACSImpl.valueOf(_iSname, _parentName, _name,  
              _acs.getACSType(), _acs.is_hub_, _acs.visibleSpace_,
              new TreeMap<String,Boolean>(_acs.m_Structure_), 
              m_StringConstraints,
              new HashMap(_acs.m_ConstraintsForNewObjects_),              
              new HashMap(_acs.m_StandardRights_),
              new HashMap(_acs.m_SpecializedRights_),
              _acs.l_Rights_, _acs.l_ImmutableRules_, null, null); // TODO: l_Rights_ copy

      m_StringConstraints = null;

                            // for testing and tracking
                            //System.out.println("in ACSImpl.copy(), "+
                              //"result just after the constructor: "+result);
      
      // variables
      Resource res = null;
      Resource resCopy = null;
      EligibleParty ep = null;
      EligibleParty epCopy = null;
      GroupIDMemberImpl grpMember = null;
      
        // adds result to the map of ACS in BaseManagerImpl.
        // throws BaseError if already exists.
        // TAKE OUT THIS LIGNE for testing the base without ACSManager
      //if (_toRegisterInManager) {
        //BaseManagerImpl.getInstance().newEorACS(result);
      //}
      
      result.baseError_ =_acs.getEorBaseError();
      
         // copies the EP before the resources, to be matched
         // to Resource.copy()
      try {
          // first loop on ep - excepted for actors
          for (Iterator ite = _acs.getEorM_EligibleParties().values().iterator(); ite.hasNext();) {
              ep = (EligibleParty)ite.next();
              // the external ep are registered through the aclEntry copy
              // so, only own ep are processed
              if (ep.getEorACS() == _acs && !(ep instanceof Actor)) {

                            // for testing and tracking
                            //System.out.println("in ACSImpl.copy(), "+
                              //"copy of the no-actor EP: "+ep);
                  
                  // copies the groupIDMember and registers the copy in result,
                  // but without the internal acl which are copied in Resource.copy()
                  epCopy = (EligibleParty)ep.copyWithoutInternalAcl(
                    ((Name)result.getName()).extended(ep.getName().lastComponent()),
                    result, null, Boolean.TRUE);

              } // if
          } // for ite          

          // second loop on groupIDs to copy the group members
          // which has been created in the first loop
          for (Iterator ite = _acs.getEorM_EligibleParties().values().iterator(); ite.hasNext();) {
              ep = (EligibleParty)ite.next();
              // the external ep are registered through the aclEntry copy
              // so, only own ep are processed
              if (ep.getEorACS() == _acs) {
                  // copies all the group members in result
                  if (ep instanceof GroupID && ((GroupID)ep).getEorL_Members().length != 0) {

                            // for testing and tracking
                            //System.out.println("in ACSImpl.copy(), "+
                              //"copy of the members for the group: "+ep);
                  
                      // gets the copy
                      epCopy = (EligibleParty)result.getEorM_EligibleParties().get(
                        ((Name)result.getName()).extended(ep.getName().lastComponent()) );                      
                      
                      ImmutableGroupIDMember[] l_toCopyMembers = ((GroupID)ep).getEorL_Members();
                      Name newName;
                      for (int i = 0; i < l_toCopyMembers.length; i++ ) {
                            if (l_toCopyMembers[i].getName().lastComponent() == null) continue;
                            newName = ((Name)epCopy.getName()).context().extended(
                                   l_toCopyMembers[i].getName().lastComponent());
                            // registers the right member
                            if (epCopy.getEorACS().containsOwnEligiblePartyName(newName)) {
                               grpMember = (GroupIDMemberImpl)
                                 epCopy.getEorACS().getEorM_EligibleParties().get(newName);
                               if (!((GroupID)epCopy).containsAsMemberWithoutAliasing(grpMember))
                                ((GroupID)epCopy).addEorMember(grpMember);
                            } // if
                      } // for
                      l_toCopyMembers = null;
                      newName = null;
                      grpMember = null;
                  } // if
              } // if
          } // for ite 
          
          // copies each _acs resource with result as ACS
             // through the copy of all the first-level directories
          BaseUtilityImpl.copyDirectoryChildrenForACS(_acs, result);

      } catch (CreateError c) {
          throw new BaseError("in ACSImpl.copy(): "+c.toString());
      }

                            // for testing and tracking
                            //System.out.println("End of ACSImpl.copy()");
      
      return result;
  } // copy
  
  //
  // --------------------- methods of the interfaces -------------
  //
  
  /**
   * Called by BeamerManager.getNotDisplayThisProperty().
   * @return true if some specialized rights have been defined in this ACS.
   */
  public boolean getHaveSpecializedRights() {
          return haveSpecializedRights_;
  }
  
  /**
   * Get the name of the information system which owns this acs.
   * @return a copy of the IS name. May be empty, but not null.
   */
  public final ImmutableName getISName() {
      if (iSname_ == null) return null;
      else return (ImmutableName)iSname_.clone();
  }
  
  /**
   * Get the BaseObject name.
   * The information system name is the first part of the name,
   * if it is non-null.
   * @return a copy of the ACS name. May be empty, but not null.
   */
  public final ImmutableName getName() {
      if (name_ == null) return NameImpl.EMPTY_INSTANCE;
      else return (ImmutableName)name_.clone();
  }
  
  /**
   * The full name is unique for the Access Road program.
   * Its is equal to the ACS name.
   * @return a copy of the name. May be empty, but not null.
   */
    public ImmutableName getFullName() {
       if (name_ == null) return NameImpl.EMPTY_INSTANCE;
       return getName();
    }
    
  /**
   * This short name is NOT always unique for the Access Road program.
   * @return the ACS name last component. Never null.
   */
    public String getNickName() {
        if (name_ == null) return "";
        else return name_.lastComponent();
    } 
   
  /**
   * Gets the mandatory parent name for the ACS in the IS nametree.
   * @return the parent name. Never null nor empty.
   */
  public ImmutableName getParentName()	{
    if (parentName_ == null) return null;
    else return (ImmutableName)parentName_.clone();
  } // getParentName() 
  
  /**
  * Returns the ACS type, as ACSImpl.GENERIC, LINUX, or defined by the user,
  * and it is completed by the ACS policy checksum.
  */
  public String getACSType() {
     return aCSType_;
  }
  
  /**
  * True if an other open ACS controls this ACS,
  * that is has an external AclEntry or Linked Privilege
  * for a resource or a virtual folder from this ACS.
  * If an another ACS has an external AclEntry or Privilege
  * which is connected to an eligible party from this ACS,
  * this ACS is NOT controlled. The concept of control is for targets only.
  * @return true if this ACS is controlled by one or several ACS
  */
  public boolean getIsControlled() {
      return controlled_;
  }
  
  /**
   * @return the name of the icon associated to the ACS. May be null.
   */
  public String getIconName() {
    return iconName_;
  }
  
  /**
   * @return the icon associated to the ACS. May be null.
   */
  public ImageIcon getEorIcon() {
    return icon_;
  }
  
  /**
   * The form is 'Editor: full_editor_name - AcsAddon: addon_name'.
   * It is derived from the values returned by BaseManagerImpl.getL_AcsAddonDescriptors().
   * The AcsAddon name is 'NONE' if the ACS is not from an AcsAddon.
   * This is a simple copy of the values in the vocabulary map.
   * Caution: the full editor name is at the key ACSFactoryUtilityImpl.ACSADDON_EDITOR,
   * not at the key ACSFactoryUtilityImpl.ACSADDON_PACKAGE_EDITOR.
   * @return the full editor name and the AcsAddon name for the ACS. Never null.
   */
  public String getEditorAndAddonNames() {
    return editor_;
  }

  /**
   * To detect an AcsAddon, this method tests
   * (ACS.getClass().toString().indexOf("ARoad0.AcsAddon.") != -1).
   */
  public boolean getIsFromAcsAddon() {
      return (getClass().toString().indexOf("ARoad0.AcsAddon.") != -1);
  }
  
  /**
   * @return long is the serialVersionUID used by the serialization process.
   */
  public synchronized long getSerialVersionUID() {
    return serialVersionUID__;
  }

  /**
   * @return the BaseError instance
   */
  public BaseError getEorBaseError() {
    return baseError_;
  }

  /**
   * The resources are own or external ones. They may be directories, actors
   * or nx executables.
   * May be called from a view worker thread.
   * @return SortedMap of registered open resources.
   *   Unmodifiable map where the name is the key.
   */
  public final SortedMap<ImmutableName, ImmutableResource> getEorM_Resources()	{

                // for testing and tracking
                //System.out.println("in ACSImpl.getEorM_Resources(): ");
                //for (Iterator ite = m_Resources_.values().iterator() ; ite.hasNext() ; ) {
                //  System.out.println(ite.next());
                //}

    return Collections.unmodifiableSortedMap(m_Resources_);
  }

  /**
   * The eligible parties are own or external ones. They may be accounts, groups or
   * actors.
   * Called in the generic access paths search, then may be called
   * from a view worker thread.
   * @return SortedMap of registered open eligible parties.
   *   Unmodifiable map where the name is the key.
   */
  public final SortedMap<ImmutableName, ImmutableEligibleParty> getEorM_EligibleParties()	{
    return Collections.unmodifiableSortedMap(m_EligibleParties_);
  }

  /**
   * Copy of the own eligible parties in a new map.
   * @return SortedMap of registered own eligible parties
   *  in the ACS base. Unmodifiable map where the name is the key.
   *  Returns a subset of the getEorM_EligibleParties method,
   *  thus it is slower.
   */
  public final SortedMap<ImmutableName, ImmutableEligibleParty> getEorM_OwnEligibleParties() {
    TreeMap result = new TreeMap(m_EligibleParties_);

    for (Iterator ite = result.values().iterator() ; ite.hasNext() ; ) {
      if ( ((EligibleParty) ite.next()).getEorACS() != this) ite.remove();
    }

    return Collections.unmodifiableSortedMap(result);
  }

  /**
   * Copy of the own resources in a new map.
   * @return SortedMap of registred own resources copies,
   *  for which it is the main ACS. Unmodifiable map where the name is the key.
   * Returns a subset of the getEorM_Resources method, thus it is slower.
   */
  public final SortedMap<ImmutableName, ImmutableResource> getEorM_OwnResources() {
    TreeMap result = new TreeMap(m_Resources_);

    for (Iterator ite = result.values().iterator() ; ite.hasNext() ; ) {
      if ( ((Resource) ite.next()).getEorACS() != this) ite.remove();
    }

    return Collections.unmodifiableSortedMap(result);
  }

  /**
   * Copy of the own no-parent resources in a new map.
   * @return SortedMap of registred own resources copies,
   *  for which there is no parent. Unmodifiable map where the name is the key.
   * Returns a subset of the getEorM_Resources method, thus it is much slower.
   */
  public final SortedMap<ImmutableName, ImmutableResource> getEorM_OwnNoParentResources() {
    TreeMap result = new TreeMap(m_Resources_);
    Resource resource;

    for (Iterator ite = result.values().iterator() ; ite.hasNext() ; ) {
      resource = (Resource)ite.next();
      if ( resource.getEorACS() != this || resource.getEorParent() != null) {

                                // for testing
                                //System.out.println("ACSImpl.getEorM_OwnNoParentResources(), "+
                                  //      "removed resource in the result: " +
                                    //    resource.getName()+"- with the parent: "+
                                      //  resource.getEorParent());


        ite.remove();
      } // if
      else {

                                // for testing
                                //System.out.println("ACSImpl.getEorM_OwnNoParentResources(), "+
                                  //      "returned resource in the result: " +
                                    //    resource.getName()+"- with the parent: "+
                                      //  resource.getEorParent());
      }


    } // for ite

    return Collections.unmodifiableSortedMap(result);
  }

  /**
   * Copy of the external resources in a new map. They are defined as owned by another ACS
   * but controlled by this ACS through an external AclEntry or Linked Privilege.
   * @return SortedMap of registred external resources copies,
   * for which it is NOT the main ACS. Unmodifiable map where the name is the key.
   * Returns a subset of getEorM_Resources(), thus it is slower.
   * May be empty but never null.
   */
  public final SortedMap<ImmutableName, ImmutableResource> getEorM_ExternalResources() {
    TreeMap result = new TreeMap();
    Resource resource = ResourceImpl.EMPTY_INSTANCE;

    for (Iterator ite = m_Resources_.values().iterator() ; ite.hasNext() ; ) {
      resource = (Resource) ite.next();
      if ( resource.getEorACS() != this )
          result.put(resource.getName(), resource);
    } // for
    resource = null;
    return Collections.unmodifiableSortedMap(result);
  } // getEorM_ExternalResources()

  /**
   * Copy of the external eligible parties in a new map.
   * They are defined as owned by another ACS
   * but having rights through an external AclEntry or Linked Privilege
   * that is managed by this ACS.
   * @return SortedMap of registred external eligible party copies,
   * for which it is NOT the main ACS. Unmodifiable map where the name is the key.
   * Returns a subset of getEorM_EligibleParty(), thus it is slower.
   * May be empty but never null.
   */
  public final SortedMap<ImmutableName, ImmutableEligibleParty> getEorM_ExternalEligibleParties() {
    TreeMap result = new TreeMap();
    EligibleParty ep = EligiblePartyImpl.EMPTY_INSTANCE;

    for (Iterator ite = m_EligibleParties_.values().iterator() ; ite.hasNext() ; ) {
      ep = (EligibleParty) ite.next();
      if ( ep.getEorACS() != this )
          result.put(ep.getName(), ep);
    } // for
    ep = null;
    return Collections.unmodifiableSortedMap(result);
  } // getEorM_ExternalEligibleParties()

  /**
   * Copy of the external virtual folders in a new array. 
   * They are defined as owned by another ACS but having rights through 
   * an external Privilege that is managed by this ACS.
   * @return copy of the virtual folders list. May be null.
  */
  public final VirtualFolder[] getEorL_ExternalVirtualFolders() {
    if (l_VirtualFolders_ == null ||
            l_VirtualFolders_.length == 0) return null;
    
    int             count = 0;
    VirtualFolder[] l_temp = new VirtualFolder[l_VirtualFolders_.length];
    VirtualFolder[] result = null;
    
    for (int i = 0; i < l_VirtualFolders_.length; i++) {
        if (l_VirtualFolders_[i].getEorACS() != this) {
            l_temp[count] = l_VirtualFolders_[i];
            count++;
        }
    } // for i
    
    if (count != 0) {
        result = new VirtualFolder[count];
        System.arraycopy(l_temp, 0, result, 0, count);
    }
    
    l_temp = null;
    return result;
  } // getEorM_ExternalVirtualFolders()

  /**
   * Copy of all the actors in a new map.
  * The actors are controlled resources; they are internal or external.
  * They can't be external eligible actors, referenced in some AclEntries.
  * @return new SortedMap of registred actors.
  * Unmodifiable map where the name is the key. 
  * Return a subset of getM_Resources(), thus it is slower.
  */
  public SortedMap<ImmutableName, ImmutableActor> getEorM_Actors() {
    TreeMap result = new TreeMap(m_Resources_);

    for (Iterator ite = result.values().iterator() ; ite.hasNext() ; ) {
      if ( !((Resource) ite.next()).getActorActivated() ) ite.remove();
    }
    //  with result.entrySet(), result.remove( result.entrySet().iterator().next().getKey() );
    //  is faulty in some cases, depending on the name list

    return Collections.unmodifiableSortedMap(result);
  }

  /**
   * Copy of the own actors in a new map.
  * Get new unmodifiable SortedMap of registred own actors,
  *  for which it is the main ACS.
  * @return unmodifiable SortedMap of registered own actors,
  *  for which it is the main ACS. Unmodifiable map where the name is the key.
  *  Return a subset of getM_OwnResources(), thus it
  *  is slower than getM_OwnResources().
  */
  public SortedMap<ImmutableName, ImmutableActor> getEorM_OwnActors() {
    TreeMap result = new TreeMap(m_Resources_);
    Resource res = null;

    for (Iterator ite = result.values().iterator() ; ite.hasNext() ; ) {
        res = (Resource)ite.next();
        if ( !res.getActorActivated() || res.getEorACS() != this)
            ite.remove();
    } // for ite

    res = null;
    return Collections.unmodifiableSortedMap(result);
  } // getEorM_OwnActors()

/**
 * List of own Actors which are not right proxies, then which are supposed to be
 * secure and without any virus able to deliver rights to the rights users
 * which execute these actors.
 * The default state is to be a right proxy, that is to suppose no
 * confidence in the security of actor code. For the user, it is an important choice
 * to set an actor as not been a right proxy.
 * @return unmodifiable List of actors. May be empty. 
 */
public List<ImmutableActor> getEorL_NoRightProxyActors() {
    if (l_NoRightProxyActors_ == null)
        l_NoRightProxyActors_ = new ArrayList(INITIAL_CAPACITY_25);    
    return (List)((ArrayList)l_NoRightProxyActors_).clone();
} // getEorL_NoRightProxyActors()

  /**
 * Gets the number of own resources in the ACS,
 * without the external objects. The returned value is not updated through
 * a property change.
 * @return number of resources owned by the ACS
 */
  public synchronized int getNumberOfResources() {
    int result = 0;
    if (m_Resources_ != null) result += m_Resources_.size(); 
    return result;
  }  // getNumberOfResources()

  /**
 * Gets the number of own eligible parties in the ACS,
 * without the external objects. The returned value is not updated through
 * a property change.
 * @return number of eligible parties owned by the ACS
 */
  public synchronized int getNumberOfEPs() {
    int result = 0; 
    if (m_EligibleParties_ != null) result += m_EligibleParties_.size();
    return result;
  }  // getNumberOfEPs()

  /**
   * The current root is the single directory
   * without parent, if any. If there are two or more
   * directories without parent, returns an empty name.
   * @return Never null. May be equal to NameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnResourcesRootName() {
      if (ownResourcesRootName_ == null) return NameImpl.EMPTY_INSTANCE;
      else return (ImmutableName)ownResourcesRootName_.clone();
  }

  
  /**
   * The current root is the first non-finalized group
   * without parent which has been created.
   * @return name of the root for the recorded own groups in the ACS base.
   *  May be equal to NameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnGroupsRootName() {
    return ownGroupsRootName_;
  }
  
  /**
   * The current root is the first non-finalized virtual folder for resources
   * without parent which has been created.
   * @return name of the root for the recorded virtual folders in the ACS base.
   *   May be equal to NameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnResourceFoldersRootName() {
    return ownResourceFoldersRootName_;
  }

  /**
   * The current root is the first non-finalized virtual folder for
   * eligible parties, without parent which has been created.
   * @return name of the root for the recorded virtual folders in the ACS base.
   *  May be equal to NameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnEPFoldersRootName() {
    return ownEPFoldersRootName_;
  }
  
  /**
  * Get the internal and external AclEntries managed by this ACS. They connect a resource
  * and an eligible party from this ACS or from other ones.
  * They can't be deleted directly. Use their resources for doing that.
  * @return ImmutableAclEntry[]: copy of the AclEntries of this ACS. May be null,
  * with empty aclEntries, or equal to 'new AclEntryImpl[] {new AclEntryImpl()}'.
  */
  public ImmutableAclEntry[] getEorL_AclEntries() {
    if (l_AclEntry_ == null) return null;
    ImmutableAclEntry[] result = new AclEntryImpl[l_AclEntry_.length];
    System.arraycopy(l_AclEntry_, 0, result, 0, l_AclEntry_.length);
    return result;
  }
  
  /**
  * Get the external AclEntries managed by this ACS. They connect a resource
  * and an eligible party, and at least one of them is an external object for this ACS.
  * @return copy of the external AclEntries of this ACS. May be null.
  */
  public ImmutableAclEntry[] getEorL_ExternalAclEntries() {
    if (l_AclEntry_ == null || l_AclEntry_.length == 0) return null;
    List list = new ArrayList(INITIAL_CAPACITY_25*10);
    
    for (int i = 0; i < l_AclEntry_.length; i++) {
        if (l_AclEntry_[i] != null && 
                (l_AclEntry_[i].getEorResACS() != this || 
                l_AclEntry_[i].getEorEP().getEorACS() != this)  )
            list.add(l_AclEntry_[i]);
    } // for i
    if (!list.isEmpty()) {
        ImmutableAclEntry[] result = (ImmutableAclEntry[]) list.toArray(new ImmutableAclEntry[0]);
        list = null;
        return result;
    }
    else {
        list = null;
        return null;
    }
  } // getEorL_ExternalAclEntries()
  
  /**
  * Gets an AclEntry managed by this ACS from its detailled name.
  * @param _detailledName AclEntry detailled name
  * @return the AclEntry may be null.
  */
  public AclEntry getOneAclEntryFromDetailledName(String _detailledName) {

                // for testing and tracking
                //System.out.println("ACSImpl.getOneAclEntryFromDetailledName(),"+
                  //      " on the ACS "+name_.lastComponent()+
                    //    " for the detailled name: "+_detailledName);

    if (l_AclEntry_ == null || _detailledName == null)
        return null;
    AclEntry result = null;

    for (int i = 0; i < l_AclEntry_.length; i++) {

                // for testing and tracking
                //System.out.println("--- compare "+l_AclEntry_[i].getDetailledName());

        if (l_AclEntry_[i].getDetailledName().equals(_detailledName)) {
            result = (AclEntry)l_AclEntry_[i];
            break;
        }
    } // for
    return result;
  } // getOneAclEntryFromDetailledName()
  
  /**
   * Gets all the rights which are allowed in this ACS.
   * May be metarights, which then do not belong to this ACS, or may be acsrights.
   * Called in the generic access paths search, then may be called
   * from a view worker thread.
   * @return all the rights used in this acs.
   */
  public final synchronized StringRight[] getL_Rights() {
    if (l_Rights_ == null) return null;
    StringRight[] result = new StringRight[l_Rights_.length];
    System.arraycopy(l_Rights_, 0, result, 0, l_Rights_.length);
    return result;
  }
  
  /**
   * Gets the first acsright which has 'full_control' as metaright, or this metaright
   * if no acsright complies this condition.
   * 'full_control' metaright has 'deny_all' metaright as opposite.
   * @return the right. Never null. 
   */
  public StringRight getFullControlRight() {
    if (l_Rights_ == null) return null;
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        
                // for testing and tracking
                //System.out.println("---------------------- "+l_Rights_[i].getName());
        
        if (!(l_Rights_[i] instanceof StringMetaRight) &&
                l_Rights_[i].getEorMetaRight().getRight().equals("full_control")) {
            result = l_Rights_[i];
            break;
        }
    } // for
    if (result == null) 
        result = BaseManagerImpl.getInstance().getMetaRight("full_control");
    return result;      
  } // getFullControlRight()
  
  /**
   * Gets a right which is registered by this ACS. It may be a StringMetaRight.
   * @param _name is the name of the right, provided by StringRight.getName().
   * @return the right. Null if it does not exist.
   */
  public final StringRight getOneRight(ImmutableName _name) {
      
                // for testing and tracking
                //System.out.println("in ACSImpl.getOneRight() for the name "+_name); 
                //System.out.println("in ACSImpl.getOneRight(), l_Rights_ names are: ");
      
    if (l_Rights_ == null || _name == null) return null;
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        
                // for testing and tracking
                //System.out.println("---------------------- "+l_Rights_[i].getName());
        
        if (l_Rights_[i].getName().equals(_name)) {
            result = l_Rights_[i];
            break;
        }
    } // for
    return result;
  } //getOneRight()
  
  /**
   * Gets a right which is registered by this ACS. It may be a StringMetaRight.
   * Called by StringRightImpl.getRightFromNameAsString() and
   * getL_SelectedAcsRights().
   * @param _right is the right, provided by StringRight.getRight(). May be null.
   * @return the right. Null if it does not exist.
   */
  public final StringRight getOneRight(String _right) {
                
                // for testing and tracking
                //System.out.println("Call to ACSImpl.getOneRight() for the string "+_nickName);
                
    if (l_Rights_ == null || _right == null) return null;
    
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        if (l_Rights_[i].getRight().equals(_right)) {
            result = l_Rights_[i];
            break;
        }
    } // for
    return result;
  } //getOneRight() 
  
  /**
   * Gets the first right which is registered by this ACS with the correct metaright.
   * Several acsrights in the ACS may be associated to the same metaright.
   * It may be a StringMetaRight. It may be desactivated.
   * @param _metaRight is the right of the metaright of the searched right, 
   *             provided by StringRight.getRight().
   * @return the first-found right. Null if it does not exist.
   */
  public final StringRight getFirstRightFromMeta(String _metaRight) {
                
                // for testing and tracking
                //System.out.println("Call to ACSImpl.getFirstRightFromMeta() for the string "+_metaRight);
                
    if (l_Rights_ == null || _metaRight == null) return null;
    
    StringRight result = null;
    for (int i = 0; i < l_Rights_.length; i++) {
        if (l_Rights_[i].getEorMetaRight().getRight().equals(_metaRight)) {
            result = l_Rights_[i];
            break;
        }
    } // for
    return result;
  } //getFirstRightFromMeta() 
  
  /**
   * UNUSED METHOD, to implement ACSRun.
   * Register a new acsright or a used metaright in this ACS. 
   * Not described in the BeanInfo.
   * @param _right is the new right to register. Not null.
   * @exception CreateError if the method is called.
   */
   public void addRight(StringRight _right ) throws CreateError {
      return;
   } // addRight()

  /**
   * UNUSED METHOD, to implement ACSRun.
   * Unregisters a right in this ACS.
   * Used by the StringRight.finalizeForUser() method.
   * Not described in the BeanInfo. 'OneRight' is a bound property.
   * @param _right the right to unregister.
   * @exception UpDateError if the right is null or not known.
   */
   public void removeRight(StringRight _right ) throws UpDateError {
      return;
   } // removeRight()    

  /**
  * Gets the privileges managed by this ACS. They may connect several resources
  * and several eligible parties from this ACS.
  * @return a copy of the privileges list. May be null or empty.
  */
  public Privilege[] getEorL_Privileges() {
    if (l_Privilege_ == null) return null;
    Privilege[] result = new Privilege[l_Privilege_.length];
    System.arraycopy(l_Privilege_, 0, result, 0, l_Privilege_.length);
    return result;
  }

  /**
  * Gets a privilege managed by this ACS from its detailled name.
  * @param _detailledName the privilege detailled name
  * @return the Privilege. May be null.
  */
  public Privilege getOnePrivilegeFromDetailledName(String _detailledName) {
    if (l_Privilege_ == null || _detailledName == null)
        return null;
    Privilege result = null;
                
                        // for testing and tracking
                        //System.out.println("ACSImpl.getOnePrivilegeFromDetailledName()"+
                          //      " for the detailled name "+_detailledName);
    
    for (int i = 0; i < l_Privilege_.length; i++) {

                        // for testing and tracking
                        //System.out.println(" --- compare to: "+
                          //      l_Privilege_[i].getDetailledName());

        if (l_Privilege_[i].getDetailledName().equals(_detailledName)) {
            result = l_Privilege_[i];
            break;
        }
    } // for
    return result;
  } // getOnePrivilegeFromDetailledName()

   /**
   * Gets all the PrivilegeForLinks of the ACS, not the
   * PrivilegeForTypes.
   * @return List of PrivilegeForLinks. May be empty and then immutable, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForLinks() {
      if (l_Privilege_ == null) return Collections.emptyList();
      List<Privilege> result = new ArrayList(INITIAL_CAPACITY_25);
      for(int i = 0; i < l_Privilege_.length; i++) {
          if (l_Privilege_[i] instanceof PrivilegeForLinks)
              result.add(l_Privilege_[i]);
      } // for
      return result;
  } // getEorPrivilegeForLinks()

   /**
   * Gets all the PrivilegeForTypes of the ACS.
   * @return List of PrivilegeForTypes. May be empty and then immutable, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForTypes() {
      if (l_Privilege_ == null) return Collections.emptyList();
      List<Privilege> result = new ArrayList(INITIAL_CAPACITY_25);
      for(int i = 0; i < l_Privilege_.length; i++) {
          if (l_Privilege_[i] instanceof PrivilegeForType)
              result.add(l_Privilege_[i]);
      } // for
      return result;
  } // getEorPrivilegeForTypes()

   /**
   * Gets all the PrivilegeForTypes which have the right source type.
   * The key word 'immutable' is removed before the comparison of types.
   * There is an inverse method gWorkInterface.LinkRights.getEorL_SourcesForPrivilegeForType().
   * Caution: works for TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges, but
   * not for TYPED_FOR_SEED privileges.
   * Called by BasicImpl.setType() only for for TYPED and TYPED_CLASSED privileges.
   * Calls UtilityImpl.getObjectTypeWithoutImmutable().
   * @param _type is the source type or, for TYPED_CLASSED_SPECIFIC, it is the source name.
   * No action if null.
   * @return List of PrivilegeForTypes. May be empty and then immutable, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForTypesForSource(String _type) {
      if (l_Privilege_ == null || _type == null) return Collections.emptyList();

      List<Privilege> result = new ArrayList(INITIAL_CAPACITY_25);
      String sType = null;
      String type = null;
      type = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(_type);

      for (int i = 0; i < l_Privilege_.length; i++) {
            if (!(l_Privilege_[i] instanceof PrivilegeForType))
              continue;

            sType = ((PrivilegeForType)l_Privilege_[i]).getSourceType();
            sType = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(sType);

            if (type.equals( sType ) ) result.add(l_Privilege_[i]);
      } // for
      return result;
  } // getEorL_PrivilegeForTypesForSource()
  
   /**
   * Gets all the PrivilegeForTypes which have the right target type.
   * The key word 'immutable' is not taken in account in the comparison of types.
   * There is an inverse method gWorkInterface.LinkRights.getEorL_TargetsForPrivilegeForType().
   * Caution: works for TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges, but
   * not for TYPED_FOR_SEED privileges.
   * Called by BasicImpl.setType() and VirtualFolderImpl.setType()
   * only for for TYPED and TYPED_CLASSED privileges.
   * Calls UtilityImpl.getObjectTypeWithoutImmutable().
   * @param _type is the target type or, for TYPED_CLASSED_SPECIFIC, it is the target name.
   * No action if null.
   * @return List of PrivilegeForTypes. May be empty and then immutable, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForTypesForTarget(String _type) {
      if (l_Privilege_ == null || _type == null) return Collections.emptyList();

      List<Privilege> result = new ArrayList(INITIAL_CAPACITY_25);
      String sType = null;
      String type = null;
      type = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(_type);


      for (int i = 0; i < l_Privilege_.length; i++) {
            if (!(l_Privilege_[i] instanceof PrivilegeForType))
              continue;
            sType = ((PrivilegeForType)l_Privilege_[i]).getTargetType();
            sType = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(sType);

            if (type.equals( sType ) ) result.add(l_Privilege_[i]);
      } // for
      return result;
  } // getEorL_PrivilegeForTypesForTarget()
  
  /**
  * Gets the virtual folders managed by this ACS.
  * Called in the generic access paths search, then may be called
  * from a view worker thread.
  * @return VirtualFolder[] is a copy of the virtual folders list. May be empty or null.
  */
  public VirtualFolder[] getEorL_VirtualFolders() {
    if (l_VirtualFolders_ == null) return null;
    VirtualFolder[] result = new VirtualFolder[l_VirtualFolders_.length];
    System.arraycopy(l_VirtualFolders_, 0, result, 0, l_VirtualFolders_.length);
    return result;
  }

  /**
  * Gets the virtual folders managed by this ACS and which contain only
  * resources.
  * @return the relevant virtual folders. May be empty but not null.
  */
  public List<VirtualFolder> getEorL_ResourceVirtualFolders() {
      if (l_VirtualFolders_ == null || l_VirtualFolders_.length == 0)
          return Collections.emptyList();


            // for testing and tracking
            //System.out.println("in ACSImpl.getEorL_ResourceVirtualFolders() when not null: ");
            //for (int i = 0; i < l_VirtualFolders_.length; i++ ) {
              //  System.out.println(" ----- "+l_VirtualFolders_[i]);
            //}

      List<VirtualFolder> result = new ArrayList(INITIAL_CAPACITY_25);
      for(int i = 0; i < l_VirtualFolders_.length; i++) {
          if (l_VirtualFolders_[i].getMainVFType().equals(VirtualFolderImpl.RES_VFOLDER) )
              result.add(l_VirtualFolders_[i]);
      } // for

            // for testing and tracking
            //System.out.println("in ACSImpl.getEorL_ResourceVirtualFolders(), returned value: ");
            //for (VirtualFolder vf : result) {
              //  System.out.println(" ----- "+vf);
            //}

      return result;
  } // getEorL_ResourceVirtualFolders()
  
  /**
  * Gets the virtual folders managed by this ACS and which contain only
  * eligible parties.
  * @return List of the virtual folders list. May be empty but not null.
  */
  public List<VirtualFolder> getEorL_EPVirtualFolders() {
      if (l_VirtualFolders_ == null || l_VirtualFolders_.length == 0) {

                          // for testing purpose 
                          //System.out.println("ACSImpl.getEorL_EPVirtualFolders() "+
                            //                    "with a null VirtualFolders list" );

          return Collections.emptyList();
      }
      List result = new ArrayList(INITIAL_CAPACITY_25);
      for (int i = 0; i < l_VirtualFolders_.length; i++) {
          if (l_VirtualFolders_[i].getMainVFType().equals(VirtualFolderImpl.EP_VFOLDER) ) {
              result.add(l_VirtualFolders_[i]);
          } // if
      } // for
      return result;
  } // getEorL_EPVirtualFolders()
  
  /**
  * Gets a virtual folder managed by this ACS.
  * @param _name is the folder name
  * @return the VirtualFolder - May be null.
  */
  public VirtualFolder getOneVirtualFolder(ImmutableName _name) {
    if (l_VirtualFolders_ == null || _name == null || _name.isEmpty())
        return null;
    VirtualFolder result = null;
    for (int i = 0; i < l_VirtualFolders_.length; i++) {
        if (l_VirtualFolders_[i].getName().equals(_name)) {
            result = l_VirtualFolders_[i];
            break;
        }
    } // for
    return result;
  } // getOneVirtualFolder()

  /**
   * Gets all the open ACS that have an external AclEntry or an external Linked Privilege
   * for a resource or a virtual folder from this ACS.
   * For compatibility purpose about the version 0.7.0, this method removes
   * the null values in the array.
   * @return the open ACS controller names. May be empty but not null.
   * No null value in the array.
   */
  public final ImmutableName[] getL_AcsControllers() {
        // compatibility: l_OpenAcsControllerNames_[i] from 0.7.0 may be null
        // and it is removed there.
        int                 shiftingBack = 0;
        ImmutableName[]     newArray = new ImmutableName[l_OpenAcsControllerNames_.length];
        ImmutableName[]     result = new ImmutableName[0];
                // this code works even if shiftingBack > 1
        for (int  i = 0 ; i < l_OpenAcsControllerNames_.length ; i++) {
             // deletes the empty l_OpenAcsControllerNames_[i]
            if (l_OpenAcsControllerNames_[i] == null)
                shiftingBack++;
            else newArray[i - shiftingBack] = (ImmutableName)l_OpenAcsControllerNames_[i];
        } // for i

        if (l_OpenAcsControllerNames_.length != shiftingBack) {
            result = new ImmutableName[l_OpenAcsControllerNames_.length - shiftingBack];
            System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
        }

        l_OpenAcsControllerNames_ = new NameImpl[result.length];
        System.arraycopy( result, 0, l_OpenAcsControllerNames_, 0, result.length);
        newArray = null;
        return result;
  } // getL_AcsControllers()

  /**
   * Gets all the own actors which are bridge sources for external actors.
   * @return List of ImmutableActors. May be empty but not null.
   */
  public final List<ImmutableActor> getEorL_OwnActorBridgeSources(){
      if (l_ActorBridgeSources_ == null ) return new ArrayList(INITIAL_CAPACITY_5);
             
            // for testing and tracking
            //System.out.println("in ACSImpl.getEorL_OwnActorBridgeSources() when non-null: ");      
            //for (Iterator itetest = l_ActorBridgeSources_.iterator(); itetest.hasNext(); ) {
                //System.out.println(" ----- "+(BaseObject)itetest.next());
            //}
      
      return Collections.unmodifiableList(l_ActorBridgeSources_);
  }

  /**
   * Gets all the own actors which are bridge targets for external actors.
   * A returned value may be a bridge target for several sources in other ACS.
   * @return List of ImmutableActor. May be empty but not null.
   */
  public final List<ImmutableActor> getEorL_OwnActorBridgeTargets(){
      if (l_ActorBridgeTargets_ == null ) return new ArrayList(INITIAL_CAPACITY_5);
      
            // for testing and tracking
            //System.out.println("in ACSImpl.getEorL_OwnActorBridgeTargets() when non-null: ");      
            //for (Iterator itetest = l_ActorBridgeTargets_.iterator(); itetest.hasNext(); ) {
                //System.out.println(" ----- "+(BaseObject)itetest.next());
            //}
      
      return Collections.unmodifiableList(l_ActorBridgeTargets_);
  }

  /**
   * The modeling of the structure is incomplete when some ACSObjects miss
   * while they participate to the decision for some important access controls,
   * or simply when these missed ACSObjects are important for the overall ACS security,
   * or when their important relations with some other ACSObjects are not set.
   * On the other hand, the modeling remains complete if some secondary ACSObjects
   * are not modeled, like for instance image files in an operating system file system.
   * When the modeling of an ACS structure is incomplete, all the views which
   * use this ACS are set incomplete in their results.
   * @return true if the modeling of the ACS structure is incomplete.
   */
  public boolean getIncompleteStructureModeling() {
          Boolean result = m_Structure_.get(
                  "access metarule: incomplete modeling of the ACS structure");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * Fires the event 'IncompleteStructureModeling'.
   * Note: if the structure map does not contain the key
   * 'access metarule: incomplete modeling of the ACS structure', this method
   * throws a NullPointerException.
   * @since 0.7.0
   * @param _b true if the modeling of the ACS structure is incomplete.
   */
  public void setIncompleteStructureModeling(boolean _b) {

                          // for testing and tracking
                    //System.out.println("From ACSImpl.setIncompleteStructureModeling(), "+
                      //      "m_Structure_ at the start: "+m_Structure_);

      boolean old = (boolean)m_Structure_.get(
              "access metarule: incomplete modeling of the ACS structure");
      if (old != _b) {
          m_Structure_.put(
                  "access metarule: incomplete modeling of the ACS structure",
                  new Boolean(_b));
          firePropertyChange("IncompleteStructureModeling", old, _b);
      }
  }

  /**
   * The modeling of the behavior is incomplete when important constraints on
   * ACSObjects are not modelled, for instance to set their allowed rights, or
   * their effective default rights.
   * On the other hand, the modeling remains complete if some secondary relations
   * or constraints are not modeled by the program, like for instance the presence
   * of image files in an operating system file system for a given application software.
   * The user may complements manually the properties of an ACSObject for which
   * the behavior has not been totally defined in the ACS model.
   * When the modeling of an ACS behavior is incomplete, the views which
   * use this ACS remain complete in their results, but the creation or deletion
   * of a new ACSObject should be reviewed by the user to ensure that all the properties
   * are still well defined in the ACS.
   * @return true if the modeling of the ACS behavior is incomplete.
   */
  public boolean getIncompleteBehaviorModeling() {
          Boolean result = m_Structure_.get(
                  "access metarule: incomplete modeling of the ACS behavior");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * Gets the hub property for the ACS node in the IS nametree.
   * This property is not described in the BeanInfo. It is a simple copy
   * of the IS nametree property.
   * @since 0.7.0
   * @return the 'is_hub' property
   */
  public boolean getHubProperty() {
    return is_hub_;
  } 
  
  /**
   * Gets the visible space for the ACS in the IS nametree.
   * This property is not described in the BeanInfo. It is a simple copy
   * of the IS nametree property.
   * @since 0.7.0
   * @return the visible space property
   */
  public String getVisibleSpace() {
    return visibleSpace_;
  }

  /**
   * @since 0.7.0
   * @return true if the acs is tied to its parent acs.
   */
  public boolean getTiedToParentACS() {
      Boolean result = m_Structure_.get("embedded: this ACS is running on "+
              "and is tied to the parent ACS");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * May be called from a view worker thread.
   * @return true if the resources are managed.
   */
  public boolean getManageResources() {
          Boolean result = m_Structure_.get("resource: no resources");
          if (result != null) return !result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @return true if the resource trees are managed.
   */
  public boolean getManageResourceTrees() {
          Boolean result = m_Structure_.get("resource: directories of resources trees");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @return true if the accounts (or users) are managed. 
   */
  public boolean getManageAccounts() {
          Boolean result = m_Structure_.get("EP: accounts");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @return true if the groups of accounts are managed.
   */
  public boolean getManageGroups() {
          Boolean result = m_Structure_.get("EP: groups of accounts");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * Caution: the name of this method is confusing, because a collection of
   * nested groups may be more than one single group tree.
   * @return true if the nested groups are managed
   */
  public boolean getManageGroupTrees() {
          Boolean result = m_Structure_.get("EP: group of groups");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @return true if the internal acl entries are managed.
   */
  public boolean getManageInternalAcls() {
          Boolean result = m_Structure_.get("ACL right: internal ACL");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @return true if the external AclEntries are managed, as
   *   controler of other-acs objects, and/or as acs which is controled by other acs.
   */
  public boolean getManageExternalAcls() {
      Boolean one = m_Structure_.get("ACL right: external ACL with entity in this acs");
      Boolean two = m_Structure_.get("ACL right: pure external ACL without entity in this acs");
      if ((one != null && one.booleanValue()) || (two != null && two.booleanValue())) {
          one = null;
          two = null;
          return true;
      }
      one = null;
      two = null;
      return false;
  }
  
  /** 
   * The ACS cannot manage this property and the right-defined conditional ACLs.
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the flexible conditional AclEntries are managed
   */
  public boolean getManageFlexibleConditionalAcls() {
      Boolean result = m_Structure_.get("ACL right: flexible conditional ACL with "+
          "condition groups the groupIDmember has to be member");
      if (result != null) return result.booleanValue();
      else return false;
  }
  
  /** 
   * The ACS cannot manage this property and the flexible conditional ACLs.
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the right-defined conditional AclEntries are managed.
   */
  public boolean getManageRightDefinedConditionalAcls() {
      Boolean result = m_Structure_.get("ACL right: right-defined conditional ACL with "+
            "predefined condition groups the groupIDmember has to be member");
      if (result != null) return result.booleanValue();
      else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @return true if the aclentry rights inheritance from the parent are managed
   * (not necessary all these rights).
   */
  public boolean getManageACLRightsInheritance() { 
      Boolean result = m_Structure_.get("ACL inheritance 1: ACL inheritance");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * May be called from a view worker thread.
   * @return true if the actors are managed.
   */
  public boolean getManageActors() {
          Boolean result = m_Structure_.get("resource: no actors");
          if (result != null) return !result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @return true if the directoryEP are managed.
   */
  public boolean getManageDirectoryEPs() {
          Boolean result = m_Structure_.get("resource: eligible party directories");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @return true if the ACS bridges are managed.
   */
  public boolean getManageBridges() {
          Boolean result = m_Structure_.get("bridge right: no BriDGes");
          if (result != null) return !result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * Called in the access paths search, then may be called from a view worker thread.
   * @return true if the secondary groups are managed for actors;
   * not applicable to DirectoryEPs
   */
  public boolean getManageSecGroupsForActor() {    
      Boolean one = m_Structure_.get("EP 1: secondary groups for actors as right users");
      Boolean two = m_Structure_.get("EP: secondary groups for actors, excluding area groups");
      if ((one != null && one.booleanValue()) || (two != null && two.booleanValue())) {
          one = null;
          two = null;
          return true;
      }
      one = null;
      two = null;
      return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @return true if some account-group-other rights are managed (not
   * necessary all these rights).
   */
  public boolean getManageAGORights() { 
          Boolean result = m_Structure_.get("AGO right: Account and Group rights");
          if (result != null) return result.booleanValue();
          else return false;      
  }
  
  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if Account-Group rights inheritance from the parent are managed (not 
   * necessary all these rights).
   */
  public boolean getManageAGORightsInheritance() { 
      Boolean result = m_Structure_.get("AGO inheritance 1: AG inheritance");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the privileges are managed. 
   */
  public boolean getManagePrivilegeRights() {
          Boolean result = m_Structure_.get("PRI right: no PRI rights");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the virtual folders are managed for the ACS. 
   */
  public boolean getManageVirtualFolders() {
          Boolean result = m_Structure_.get("virtual folder: no virtual folders");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the simple linked privilege rights are managed.
   */
  public boolean getManageSimpleLinkedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: simple LPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the simple typed privilege rights are managed. 
   */
  public boolean getManageSimpleTypedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: simple TPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the if-all-sources linked privilege rights are managed.
   */
  public boolean getManageIfAllSourceLinkedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: If-All-Sources LPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the internal one-to-one linked privilege rights are managed. 
   */
  public boolean getManageInternalForOneToOneLinkedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: internal One-To-One LPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be true if the privileges are managed but all the other privilege features
   * are not managed, to allow an own component (source or target) in the privilege of another ACS.
   * Managing external privileges always implies that all the own sources of this ACS
   * may have an inner external privilege. To allow an own resource or an own virtual folder
   * of this ACS to get an inner external privilege, the privileges in resources or
   * in virtual folders have to be managed. This allows to limit the
   * inner external privileges to the own sources of the ACS, excluding the own targets,
   * if there are no internal privileges to manage. On the other hand, having
   * internal privileges means that the inner external privileges always accept 
   * as target an own resource or own virtual folder.
   * @since 0.7.0
   * @return true if the external one-to-one linked privilege rights are managed
   * as inner privileges or as privileges of another ACS.
   */
  public boolean getManageExternalForOneToOneLinkedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: external One-To-One LPRI");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the classed typed privilege rights are managed.
   */
  public boolean getManageClassedTypedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: Typed PRI for classed"+
                  "with contraints on source and target classes");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the classed specific typed privilege rights are managed,
   * for designated source and target.
   */
  public boolean getManageClassedSpecificTypedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: Typed PRI for classed specific"+
                  "for designated source and target");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the typed privilege rights for authorization servers are managed.
   */
  public boolean getManageAuthorizationServerTypedPrivileges() {
          Boolean result = m_Structure_.get("PRI right: Typed PRI for authorization server");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the granting rights in privileges are managed.
   */
  public boolean getManageGrantingRightPrivileges() {
          Boolean result = m_Structure_.get("PRI right: no granting rights");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the denying rights in privileges are managed.
   */
  public boolean getManageDenyingRightPrivileges() {
          Boolean result = m_Structure_.get("PRI right: no denying rights");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * Caution: if getManageExternalForOneToOneLinkedPrivileges() returns true,
   * the privilege of another ACS may always have as target a Resource of this ACS.
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the inner privileges in Resources are managed.
   */
  public boolean getManagePrivilegesInResources() {
          Boolean result = m_Structure_.get("PRI right: for Resource");
          if (result != null) return result.booleanValue();
          else return false;
  }

  /**
   * Caution: if getManageExternalForOneToOneLinkedPrivileges() returns true,
   * the privilege of another ACS may always have as target a VirtualFloder of this ACS.
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the inner privileges in VirtualFolders are managed.
   */
  public boolean getManagePrivilegesInVirtualFolders() {
          Boolean result = m_Structure_.get("PRI right: for VirtualFolder");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the linked privilege rights inheritance from the parent are managed
   * (not necessary all these rights).
   */
  public boolean getManageLPRIRightsInheritance() { 
      Boolean result = m_Structure_.get("LPRI inheritance 1: Linked PRI inheritance");
      if (result != null) return result.booleanValue();
      else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the virtual folders of resources are managed for the ACS,
   * and where actors are forbidden.
   */
  public boolean getManageResourceVirtualFolders() {
          Boolean result = m_Structure_.get("virtual folder: no virtual folders of Resources");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the virtual folders of actors are managed for the ACS,
   * and where other resources are forbidden.
   */
  public boolean getManageEPVirtualFolders() {
          Boolean result = m_Structure_.get("virtual folder: no virtual folders of EP");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @return true if the virtual folders of simple resources are managed for the ACS.
   */
  public boolean getManageSimpleResourceVirtualFolders() {
          Boolean result = m_Structure_.get(
                  "virtual folder: no simple resources in virtual folders of Resources");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the resource virtual folders manage directories for the ACS.
   */
  public boolean getManageDirectoryInResourceVirtualFolders() {
          Boolean result = m_Structure_.get(
                  "virtual folder: no directories in virtual folders of Resources");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the virtual folders of eligible parties manage actors for the ACS.
   */
  public boolean getManageActorInEPVirtualFolders() {
          Boolean result = m_Structure_.get(
                  "virtual folder: no actors in virtual folders of EP");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }


  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the EP virtual folders manage the directoryEPs for the ACS.
   */
  public boolean getManageDirectoryEPinEPVirtualFolders() {
          Boolean result = m_Structure_.get(
                  "virtual folder: no directoryEPs in virtual folders of EP");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * May be called from a view worker thread.
   * @since 0.7.0
   * @return true if the virtual folder trees are managed for the ACS.
   */
  public boolean getManageVirtualFolderTrees() {
          Boolean result = m_Structure_.get(
                  "virtual folder: no virtual folder trees");
          if (result != null && !result.booleanValue()) return true;
          else return false;
  }

  /**
   * @since 0.7.0
   * @return true if a member is no more than one time in each virtual folder tree.
   */
  public boolean getManageOneTimeInVirtualFolderTreeForMember() {
      Boolean result = m_Structure_.get(
              "virtual folder: a member is no more than one time"+
              "in each virtual folder tree");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @return true if a virtual folder propagates its target rights to all its members.
   */
  public boolean getManageVirtualFolderRightsPropagationToMembers() {
      Boolean result = m_Structure_.get(
              "virtual folder: propagation of target rights to all the members");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * Only an AcsAddon can manage the soft aliasing.
   * @since 0.7.0
   * @return true if the acs manages the external soft aliases
   */
  public boolean getManageSoftAlias() {
      Boolean result = m_Structure_.get(
              "Aliasing: acs object aliasing with external or internal soft alias");
      if (result != null) return result.booleanValue();
      else return false;
  }

  /**
   * @since 0.7.0
   * @return true if the acs manages the internal hard aliases
   */
  public boolean getManageInternalHardAlias() {
      Boolean result = m_Structure_.get(
              "Aliasing: acs object aliasing with internal hard alias");
      if (result != null) return result.booleanValue();
      else return false;
  }  

  /**
   * This method is semantically included in getL_TiedChildACS().
   * @deprecated
   * @return true if the acs is embedded in the parent acs.
   */
  public boolean getEmbeddedInParentACS() {
          Boolean result = m_Structure_.get("embedded: this ACS is running "+
                  "on and is embedded in the parent ACS");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * This method is semantically included in getM_RootCommandsForOpeningAndClosing(),
   * getM_RootsForOpeningNoClosing() and getM_RootsForIntegrity().
   * @deprecated
   * @return true if the acs is running on the parent acs.
   */
  public boolean getRunningOnParentACS() {
          Boolean result = m_Structure_.get("embedded: this ACS is running on the parent ACS");
          if (result != null) return result.booleanValue();
          else return false;
  } 
  
  /**
   * May be called from a view worker thread.
   * @return true if the acs is an authorization server, as an LDAP directory for instance.
   */
  public boolean getIsAuthorizationServer() {
          Boolean result = m_Structure_.get("acs: this ACS is an authorization store");
          if (result != null) return result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @return true if the Grant rights are managed. 
   */
  public boolean getManageGrantRights() {
          Boolean result = m_Structure_.get("access metarule: no granting rights");
          if (result != null) return !result.booleanValue();
          else return false;
  }
  
  /**
   * May be called from a view worker thread.
   * @return true if the Deny rights are managed. 
   */
  public boolean getManageDenyRights() {
          Boolean result = m_Structure_.get("access metarule: denying rights");
          if (result != null) return result.booleanValue();
          else return false;
  }       
  
  /**
   * May be called from a view worker thread.
   * @return true if the acs manages ACS rights (it main manages
   *    metarights too).
   */
  public boolean getManageAcsRights() {
          Boolean result = m_Structure_.get("acs: acsrights");
          if (result != null) return result.booleanValue();
          else return false;
  } 
  
  /**
   * May be called from a view worker thread.
   * @return true if the acs manages metarights (it may manage
   *    ACS rights too).
   */
  public boolean getManageMetaRights() {
          Boolean result = m_Structure_.get("acs: metarights");
          if (result != null) return result.booleanValue();
          else return false;
  } 
  
  /**
  * This map is initialized at the ACS creation to set the standard values
  * for the ACSObject Right properties (for example, the Linux rights), but only  
  * when there is no applicable specialized rights. The possible keys for
  * the generic ACS are: <P>
  * - 'Resource.OwnerRights',<P>
  * - 'Resource.GroupRights',<P>
  * - 'Resource.AclRights',<P>
  * - 'Actor.BridgeRights', <P>
  * - 'PrivilegeRights.MainType.'+PrivilegeAbst.TYPED, <P>
  * - 'PrivilegeRights.MainType.'+PrivilegeAbst.LINKED, <P>
  * An AcsAddon ACS may have complementary keys, like for instance 'Resource.OtherRights'
  * in the ACSUbuntuImpl class. In all cases, these AcsAddon keys have to start
  * with the name of an interface, to be detected by
  * ACSFactoryUtilityImpl.getL_MapKeysForSpecializedRights(),
  * and they have to end with a property change name containing 'Rights',
  * to be detected by BaseUtilityImpl.getRestrictedValues().
  * @see #getM_SpecializedRights()
  * @return unmodifiable HashMap where
  *   the key is a String of ACSObject class interface (like Resource) ended by
  *   a bound property name including 'Rights'(like GroupRights), with '.' as separator,
  *   and the value is a StringRight array of the authorized standard values for 
  *   this bound property. May be null but never empty.
  */ 
  public Map<String, StringRight[]> getM_StandardRights() {
      if (m_StandardRights_.isEmpty())
          return null;
      return Collections.unmodifiableMap(m_StandardRights_);
  }
  
  /**
   * Get the String values which define the allowed ACSObject types.
   * The default value is always all the possible values, when there is no key
   * to match. The types are used to manage the display of nodes in the explorer,
   * and to forbid the creation of instances having some specified types for a given class.
   * There are 10 possible keys, and the last one is the most complex. They are
   * described hereinafter: <P>
   *     - 'Resource.ResourceType' defines the allowed types for the Resources <P>
   *     - 'Actor.ResourceType' defines the allowed types for the Actors <P>
   *     - 'EligibleParty.EPType' defines the allowed types for the Eligible Parties <P>
   *     - 'VirtualFolder.resourceVFType' defines the allowed String types for the Resource VirtualFolders of the ACS <P>
   *     - 'VirtualFolder.EPVFType' defines the allowed String types for the Right User VirtualFolders of the ACS <P>
   *     - 'PrivilegeForType.SecondType' defines the allowed String types for the PrivilegeForTypes of the ACS <P>
   *     - 'PrivilegeForLinks.SecondType' defines the allowed String types for the PrivilegeForLinks of the ACS <P>
   *     - 'Explorer.NoAcsNode', where the values have the format 'ACSTree.XXX',
   *            to forbid the display in the explorer, of some standard nodes like
   *            'AclEntries...' or 'GroupIDs', <P>
   *     - 'Explorer.SubAcsNode' sets some subacs nodes in the explorer,
   *            where the values have the format 'ResourceType.&lt;byAcsCreator&gt; XXX'
   *            or 'GroupType.&lt;byAcsCreator&gt; XXX', to specify the display
   *            under a dedicated node '&lt;XXX_subacs&gt;', of all the resources or groups
   *            having a type starting with  '&lt;byAcsCreator&gt; XXX'. This type starting
   *            is reserved to these children.  <P>
   *     - 'Explorer.AcsGroupTree' sets some group trees in the explorer,
   *            where the values have the format 'GroupType.&lt;byAcsCreator&gt; XXX',
   *            to specify the display under a dedicated node '&lt;XXX_tree&gt;'
   *            as its direct children, all the groupIDs having a type which starts
   *            with '&lt;byAcsCreator&gt; YYY'. This type starting
   *            is reserved to these children.  <P>
   *     - 'CreationByBeamer.NoType' where the values are some forbidden types,
   *            to forbid the creation by the user and through the GUI,
   *            of ACSObjects having these types for a given class. <P>
   * With the key 'CreationByBeamer.NoType', the user can create some ACSObjects
   * only with the types the ACS authorizes. Typically, the types
   * 'ResourceType.&lt;byAcsCreator&gt;' and 'EPType.&lt;byAcsCreator&gt;' are not allowed.
   * After the first selection of allowed types, it may have a removing of some values
   * from the getM_TypesAndGUIPolicy() key 'CreationByBeamer.NoType'.
   * The forbidden types are the map values, with one of the following forms:
   * 'ResourceType.forbidden type', 'EPType.forbidden type', 'DirectoryType.forbidden type',
   * 'ActorType.forbidden type', 'UserIDType.forbidden type', 'GroupIDType.forbidden type',
   * 'VirtualFolderType.forbidden type', 'PrivilegeSecondType.forbidden type'. <P>
   * A forbidden type for EligibleParty is not forbidden to all the EligibleParty subclasses.
   * This is also true for a forbidden type for Resource. Like the order of interfaces,
   * the 'ActorType.forbidden type' list is the single list to applied to an Actor,
   * just like 'DirectoryType.forbidden type' for the Directories,
   * and 'UserIDType.forbidden type' for the UserID.
  * @return an unmodifiable HashMap where
  *   the key is a String of ACSObject class interface (like Resource) ended by
  *   a bound property name (like ResourceType), with '.' as separator,
  *   and the value is a String array of the values for this key.
  *   May be null but never empty.
  */ 
  public Map<String, String[]> getM_TypesAndGUIPolicy() {
      if (m_TypesAndGUIPolicy_.isEmpty())
          return null;
      return Collections.unmodifiableMap(m_TypesAndGUIPolicy_);
  }
  
  /**
   * Gets the ACS constraints on the allowed or forbidden ACSObject types
   * to set the GroupID, Node, Resource, AclEntry and Privilege policies.
   * The values are types of ACSObject, that is
   * the type of a Basic or a VirtualFolder, or the second type of a Privilege.
   * The default value is always to allow all the possible values, when there is no key
   * to match.
   * The keys are composed of a structured word which defines the application domain,
   * like 'GroupID.NoPrimaryGroup', and there is sometimes a type at the end.
   * The key word 'immutable' may be in an ACSObject type,
   * but it is never used in the keys for type policies. This means
   * that an immutable type has to apply the policy of its simplified form. <P>
   * There are 31 possible forms for the key, presented hereinafter: <P>
   *     - 'GroupIDMember.NoMemberOf', where the values are the types of the groupmembers for
   *            which the role of member of another group is forbidden <P>
   *     - 'GroupIDMember.IsNotConditionalAclSource', where the values are the types of
   *            the GroupIDmembers for which the conditional ACL is forbidden as source <P>
   *     - 'GroupID.NoPrimaryGroup', where the values are the types of the groups for
   *            which the role of primary group for an account is forbidden <P>
   *     - 'GroupID.NoMainGroup' where the values are the types of the groups for
   *            which the role of main group of a resource is forbidden <P>
   *     - 'GroupID.NoSecondaryGroup' where the values are the types of the groups for
   *            which the role of secondary group of an actor is forbidden <P>
   *     - 'GroupID.MemberFromAdmin' where the values are the types of the groups
   *            where every member has to be an administrator <P>
   *     - 'GroupID.NoUserIDMember' where the values are the types of the groups
   *            where every member has to not be an UserID <P>
   *     - 'GroupID.NoGroupIDMember' where the values are the types of the groups
   *            where every member has to not be a GroupID <P>
   *     - 'GroupID.NoMoreThanOneMember' where the values are the types of the groups
   *            where only one member is allowed <P>
   *     - 'GroupID.ConditionalAclGroup' where the values are the types of the groups
   *            which may be condition groups in some flexible or right-defined conditional ACLs <P>
   *     - 'Resource.OneRightACL' where the values are the types of the resources
   *            which handled only AclEntries with no or one right <P>
   *     - 'Resource.NoAccount' where the values are the types of the resources
   *            which do not handle an account as owner (a group is possible from parent) <P>
   *     - 'Resource.NoConditionalACL' where the values are the types of the resources
   *            which never handle flexible or right-defined conditional AclEntries <P>
   *     - 'Resource.NoNonConditionalACL' where the values are the types of the resources
   *            for which an AclEntry without condition group(s) is never operational <P>
   *     - 'Resource.OneConditionGroupInACL' where the values are the types of the resources
   *            which handled only conditional AclEntries with no or one condition group <P>
   *     - 'Actor.NoCurrentAccount' where the values are the types of the actors
   *            for which a current account is forbidden <P>
   *     - 'Actor.NoCurrentGroup' where the values are the types of the actors
   *            for which a current group is forbidden <P> 
   *     - 'Actor.NoNullCurrentAccount' where the values are the types of the actors
   *            for which a current account is mandatory <P>
   *     - 'Actor.IsNotBridgeTarget' where the values are the types of the actors
   *            which cannot be bridge targets <P>
   *     - 'Actor.IsNotBridgeSource' where the values are the types of the actors
   *            which cannot be bridge sources <P>
   *     - 'Actor.IsNotConditionalAclSource', where the values are the types of
   *            the Actors for which the conditional ACL is forbidden as source <P>
   * <P>
   * The following keys include an ACSObject type, after the last point:
   *     - 'GroupID.TypesOfMemberFor.allowed_GroupID_type' where the values are
   *            the allowed types of the members for a group having this type <P>
   *     - 'GroupIDMember.NoMoreThanOneGroup.allowed_GroupIDMember_type' where the values
   *            are the GroupID types for which the groupIDmember may be member of
   *            one group of this type at most <P>
   *     - 'GroupIDMember.RecommandedGroupsAsMember.allowed_GroupIDMember_type'
   *            where the values are the GroupID types for which a message
   *            recommands to the user to put the groupIDMember as member;
   *            there, it is possible to describe some alternatives
   *            like 'group_type_one<OR>group_type_two' <P>
   *     - 'Directory.TypesOfChildFor.allowed_Directory_type' where the values are
   *            the allowed types of the children for a Directory having this type,
   *            or if the type is 'NULL', for a Resource without parent <P>
   *     - 'VirtualFolder.TypesOfMemberFor.allowed_VirtualFolder_type' where the values are
   *            the allowed types of the members for a VirtualFolder having this type <P>
   *     - 'VirtualFolder.TypesOfChildFor.allowed_VirtualFolder_type' where the values are
   *            the allowed types of the children for a VirtualFolder having this type,
   *            or if the type is 'NULL', for a VirtualFolder without parent <P>
   *     - 'Resource.SecondTypesOfPrivilegeFor.allowed_Resource_type' where the values are
   *            the allowed second types of the privileges for a Resource having this type <P>
   *     - 'EligibleParty.SecondTypesOfPrivilegeFor.allowed_EligibleParty_type' where the values are
   *            the allowed second types of the privileges for an EligibleParty having this type <P>
   *     - 'VirtualFolder.SecondTypesOfPrivilegeFor.allowed_VirtualFolder_type' where the values are
   *            the allowed second types of the privileges for a VirtualFolder having this type <P>
   *     - 'Resource.TypesOfSourceForConditionalACL.allowed_Resource_type' where the values are
   *            the allowed EligibleParty types as source of a conditional ACL
   *            in a Resource having this type <P>
   * @return an unmodifiable HashMap<String, String[]> where
  *   the key is an ACSObject class interface (like Resource) ended by
  *   a bound property name (like ResourceType), with '.' as separator,
  *   and the value is a String array of the authorized types.
  *   May be null but never empty.
  */ 
  public Map<String, String[]> getM_TypeOrientedPolicy() {
      if (m_TypeOrientedPolicy_.isEmpty())
          return new HashMap(INITIAL_CAPACITY_5);
      return Collections.unmodifiableMap(m_TypeOrientedPolicy_);
  }
  
  /**
   * Gets the specialized rights extending or constraining the allowed values
   * which are defined for each right type set in getM_StandardRights().
   * The specialized rights are then superior to the standard ones, which
   * are unworkable if some specialized rights are applicable.
   * If several sets of specialized rights are applicable, it is their intersection
   * which is workable. <P>
   * As map keys, examples of quartets (object.domain.key.subkey) are: <P>
   *     - 'Target.AclRightsSet.Interface.Directory', for ACL rights 
   *           when the target is a Directory instance <P>
   *     - 'Target.AccRightsSet.Type.script', for Account rights 
   *           when the target has the type 'script' <P>
   *     - 'Source.AclRightsSet.Type.group', for ACL rights 
   *           when the source has the type 'group' <P>
   *     - 'PFType.PrvRightsSet.Type.role one', for PrivilegeForTypes
   *            having the second type 'role one' <P>
   * The object is Target or Source to indicate that these rights are for the target
   * (or source) which has the correct interface or type. <P>
   * For Privileges only, 'PFType' indicates rights for PrivilegeForTypes, and
   * 'PFLink' indicates that it is for PrivilegeForLinks. <P>
   * The domain may be one of the right categories:
   * 'AccRightsSet' for Target only, 'GpORightsSet' for Target only, 
   * 'AclRightsSet' for Target only, 'BdgRightsSet', 'RooRightsSet' for root.
   * 'PrvRightsSet' for PrivilegeForTypes and PrivilegeForLinks only. <P>
   * The key is 'Interface' or 'Type' to indicate that the subkey belongs to one
   * of these categories. <P>
   * The subkey is an interface name or a Basic type (or second type for a Privilege),
   * for which the specialized rights have to be used. <P>
   * The value in the returned map is an array of StringRight which may have 
   * a lenght of 0.
   * @see BaseUtilityImpl#getL_AcsSpecializedRights
   * @see #getM_StandardRights()
   * @return a copied Map where the key is a quartet of Strings, and the value 
   * is a copy of an array of StringRights. May be null but never empty.
   */
  public final Map<String, StringRight[]> getM_SpecializedRights() {
      if (m_SpecializedRights_ == null) return null;

      HashMap<String,StringRight[]>  m_result = 
         new HashMap<String,StringRight[]>(INITIAL_CAPACITY_25); 
      StringRight[] l_rights = null;
      Map.Entry<String,StringRight[]> entry = null;
      StringRight[] result = null;
      
      for (Iterator ite = m_SpecializedRights_.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,StringRight[]>)ite.next();
         l_rights = entry.getValue();
         if (l_rights.length != 0) {
             result = new StringRight[l_rights.length];
             System.arraycopy(l_rights, 0, result, 0, l_rights.length); 
         } else
             result = new StringRight[0];
         m_result.put(entry.getKey(), result);
      } // for ite
      
      l_rights = null;
      entry = null;
      result = null;     
      return m_result;
  } // getM_SpecializedRights()

  /**
   * The possible specialized right categories are 'AccRightsSet',
   * 'GpORightsSet', 'AclRightsSet', 'PrvRightsSet', 'BdgRightsSet', 'RooRightsSet'.
   * This property is not displayed in the BeanInfo.
   * @return the set of the used right types in the specialized rights.
   */
  public Set<String> getCategoriesInSpecializedRights() {
      return (Set) ((HashSet)categoriesInSpecializedRights_).clone();
  } // getCategoriesInSpecializedRights()
  
  /**
   * Get the associations to handle at the creation or the deletion of some ACSObjects,
   * by the mean of actions to do by the program. Actions have numbers from 1 to 999.
   * The criteria is a map key among the followings: <P>
   * - 'Class'.'Property'.&lt;Value&gt;.XXX: a new object property has the value XXX; <P>
   * - 'Class'.'Property'.&lt;Valuecontains&gt;.XXX: a new object property value contains the value XXX; <P>
   * - 'Class'.'Property'.&lt;Class&gt;.XXX: a new object property has a class with
   *        exactly the simple name XXX; <P>
   * - 'Class'.'Property'.&lt;NameStartsWith&gt;.XXX: the new object property name
   *        starts with the ACS name plus the subname XXX; <P>
   * - 'Class'.&lt;NameStartsWith&gt;.XXX: the new object name of the class
   *        starts with the ACS name plus the subname XXX; <P>
   * - &lt;Class&gt;.&lt;Value&gt;.XXX: the new object class is XXX in a gBase package; <P>
   * <P>
   * The allowed String map values depends lightly on the map key.
   * The values define one or several actions to do if the key matches.
   * There are 5 possible formats for the first value of the action N: <P>
   * - N.&lt;AssociateClass&gt;.XXX which specifies the class name XXX in gBase, without the Impl String,<P>
   *    of an associate object to build up or to finalize when the master is build up <P>
   *    or finalized <P>
   * - N.&lt;CalledMethod&gt;.&lt;Master&gt;.YYY which specifies the method YYY to call on the master, <P>
   * - N.&lt;CalledMethod&gt;.&lt;ACS&gt;.YYY which specifies the method YYY to call on the master ACS <P>
   * - N.&lt;CalledMethod&gt;.&lt;Action&gt;.C.ZZZ which specifies the method ZZZ <P>
   *    to call on the associate previously created by the action number C. <P>
   * - &lt;CalledMethod&gt;.&lt;Object&gt;.XXX+&lt;Class&gt;+YYY+&lt;Value&gt;+ZZZ which specifies the method ZZZ <P>
   *    to call on the future ACSObject having the name XXX and the class YYY,
   *     where the parameters must be the master or an associate 
   *    to set the reverse operation when it is finalized. <P>
   * After each first value, there is the full list of the constraints on the arguments
   * of the constructor (case 1) or the method (cases 2 to 5),
   * and there is one constraint specification per argument. Note that the method
   * does not need to know the property name, but it is usefull to the user.
   * Each argument constraint specification belongs to a set of 9 allowed formats,
   * presented herein for the action with the N number: <P>
   * - N.'PropertyName'.&lt;Null&gt;: null value for the argument having the name 'PropertyName' <P>
   * - N.'PropertyName'.&lt;Master&gt;: the master is the argument <P>
   * - N.'PropertyName'.&lt;MasterCreation&gt;: boolean true if the master is creating,
   *        false if it is finalizing <P>
   * - N.'PropertyName'.&lt;ArgumentValue&gt;.C: the associate or argument number C object <P>
   * - N.'PropertyName'.&lt;Value&gt;.XXX: the argument is XXX as String, ImmutableName
   *        if it is the property type, true/false if the type is boolean,
   *        or any BaseObject get from its name <P>
   * - N.'PropertyName'.&lt;Value&gt;.XXX+&lt;MasterValue&gt;+YYY: the master value <P>
   *        for the same property in its BeanInfo (the read method has to not return an array),
   *        with some optional pre String value 'XXX' and post value 'YYY'
   *        if the property type is String <P>
   * - N.'PropertyName'.&lt;Value&gt;.XXX+&lt;MasterNameLastComponent&gt;+YYY: <P>
   *        the last component in the master name, with some optional  <P>
   *        pre and post Name values as Strings <P>
   * - N.'PropertyName'.&lt;Value&gt;.&lt;MasterNameContext&gt;+YYY <P>
   *        the context in the master name, with a post Name value as String <P>
   * - N.'PropertyName'.&lt;Value&gt;.&lt;ACSName&gt;+YYY: the master ACS name, <P>
   *        with a post Name value as String if it is 
   *        the property type, or any BaseObject get from its name <P>
   * The internal map may contain values where the first characters are not 'N.', but
   * the first no-number value and its followings are returned by this method which is used by
   * StringMapPropertyEditor to edit the internal map.
   * Called by ActionNewACSyst.fillDerivedACSproperties() to populate the new acs associations.
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings started by a number between 1 to 9 followed by a point.
   *   May be empty, but not null.
   */
  public final Map<String, String[]> getM_AssociationsOnNewObject() {
      return Collections.unmodifiableMap(m_ConstraintsForNewObjects_);
  } // getM_AssociationsOnNewObject()
  
  /**
  * Add a listener to the bean, for all the bound properties.
  * Used by Gui1.SimplePropertyEditor and Gui2.ACSTreeBaseListenerImpl.
   * <P>
   * The listeners are proper to one ACSImpl instance, outside gBase,
   * transient (not serialized), and they are called in any order, excepted for:<P>
 * 1/ The sequence for a Resource change is : ExternalOrientedAclEntry,
 * ExternalOriented, M_OwnResources (or External), OneResource, OneBaseObject.
 * It is very similar for an EligibleParty change and a VirtualFolder change.<P>
   * 2/ For Actors, 'OneResource' is always fired when 'OneEligibleParty' is fired,
   *     and it is fired before this event for a resource creation,
   *     and after this first event for a resource deletion.
   *  'OneBaseObject' is fired when the operation is completed, so,
   *     and it is fired always just after 'OneEligibleParty' if it is an actor creation,
   *     and always just after 'OneResource' if it is an actor deletion.
   * <P>
  * Caution: when this method is used to detect a new ACSObject instance
  * send by ACSImpl in a PropertyChangeEvent, the new instance
  * may be not still fully initialized.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  * @since 0.5.1
  */
  public void addPropertyChangeListener(PropertyChangeListener _l){
      if (changeSupport_ == null) 
          changeSupport_ = new PropertyChangeSupport(this);
      
      synchronized (changeSupport_) 
      {changeSupport_.addPropertyChangeListener(_l);}  
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Called by gDMak.ACSTreeBaseListenerImpl,
  * Gui1.CollectionPropertyEditor, Gui1.SimplePropertyEditor,
  * ViewInBaseImpl.
  * <P>
  * The listeners are proper to one ACSImpl instance, outside gBase,
  * transient (not serialized), and they are called in any order, excepted for:<P>
 * 1/ The sequence for a Resource change is : ExternalOrientedAclEntry,
 * ExternalOriented, M_OwnResources (or External), OneResource, OneBaseObject.
 * It is very similar for an EligibleParty change and a VirtualFolder change.<P>
  * 2/ For Actors, 'OneResource' is always fired when 'OneEligibleParty' is fired,
  *     and it is fired before this event for a resource creation,
  *     and after this first event for a resource deletion.
  *  'OneBaseObject' is fired when the operation is completed, so,
  *     and it is fired always just after 'OneEligibleParty' if it is an actor creation,
  *     and always just after 'OneResource' if it is an actor deletion.<P>
  * Caution: when this method is used to detect a new ACSObject instance
  * send by ACSImpl in a PropertyChangeEvent, the new instance
  * may be not still fully initialized.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      if (changeSupport_ == null) 
          changeSupport_ = new PropertyChangeSupport(this);
      
      synchronized (changeSupport_) 
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}

            // for testing and tracking
            //if (_propertyName.equals("OneBaseObject")) {
              //  System.out.println( "ACSImpl.addPropertyChangeListener()"+
                //        ", for a 'OneBaseObject' event "); 
                //System.out.println("PropertyChangeListeners for 'OneBaseObject' : ");
                //PropertyChangeListener[] l_prop = changeSupport_.getPropertyChangeListeners("OneBaseObject");
                //for (int i = 0; i < l_prop.length; i++) {
                  //  System.out.println(" -------------- "+l_prop[i].toString());
                //}
            //} // if
      
  } // addPropertyChangeListener()

  /**
  * Remove a listener to the bean.
  * To be compliant to the BaseBeanManager interface, but no using so far.
  * @param _l PropertyChangeListener to remove
  * @since 0.5.1
  */
  public void removePropertyChangeListener(PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return; 
    
      synchronized (changeSupport_) 
      {changeSupport_.removePropertyChangeListener(_l);} 

  }  // removePropertyChangeListener()

  /**
  * Remove a listener to the bean.
  * Used by gDMak.ACSTreeBaseListenerImpl and Gui1.XXXPropertyEditor.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;
    
      synchronized (changeSupport_) 
      {changeSupport_.removePropertyChangeListener(_propertyName, _l);} 

  } // removePropertyChangeListener()

  /**
   * Test if the resource is opened, and owned or controlled by the ACS.
   * @param _name ImmutableName of the resource to test in the ACS
   * @return true if the ACS knows the resource
   */
  public final boolean containsResourceName(ImmutableName _name) {
  if (m_Resources_.containsKey(_name)) return true;
  else return false;
  }

  /**
   * Synchronized on the resources map.
   * @param _name of the resource to test
   * @return true if the ACS owns this resource, false if this resource is
   *  unknown or if it is an external one, known through an aclEntry.
   */
  public final boolean containsOwnResourceName(ImmutableName _name) {
    synchronized (m_Resources_) {
      if (m_Resources_.containsKey(_name)) {
        if ( ((Resource)m_Resources_.get(_name)).getEorACS() == this) {
          return true;
        }
      }
    } // synchro
    return false;
  }

  /**
   * Test if the eligible party is opened, and owned by the ACS or external
   * to the ACS and connected to it through an ACS aclEntry.
   * @param _name ImmutableName of the eligible party to test
   * @return true if the ACS knows this EP as its own or as an external one
   */
  public final boolean containsEligiblePartyName(ImmutableName _name) {
    if (m_EligibleParties_.containsKey(_name)) return true;
    else return false;
  }

  /**
   * True if the ACS owns this EP.
   * @param _name ImmutableName of the eligible party to test
   * @return true if the ACS owns this EP
   */
  public final boolean containsOwnEligiblePartyName(ImmutableName _name) {

                      // for testing and tracking
                      //System.out.println("In ACSImpl.containsOwnEligiblePartyName(), "+
                        //"the set of m_EligibleParties_ is: ");
                      //for (Iterator itetest = m_EligibleParties_.keySet().iterator() ; itetest.hasNext() ; ) {
                        //System.out.println((Name)itetest.next());
                      //} // for
      
    synchronized (m_EligibleParties_) {
      if (m_EligibleParties_.containsKey(_name)) {
        if ( ((EligibleParty)m_EligibleParties_.get(_name)).getEorACS() == this) {
          return true;
        }
                      // for testing and tracking
                      //System.out.println("In ACSImpl.containsOwnEligiblePartyName(), for "+
                        //name_+", detects a registered external EP: "+_name);
        
      }
    } // synchro
    return false;
  }


  /**
   * Test if an AclEntry is known. Use the equals() method.
   * @param _acl  AclEntry to test
   * @return boolean true if the ACS has recorded this aclEntry instance
   */
  public final boolean containsAclEntry(ImmutableAclEntry _acl) {
  boolean result = false;
  if (l_AclEntry_ == null || _acl == null) return result;
  for (int i = 0 ; i < l_AclEntry_.length ; i++) {
      if (_acl.equals(l_AclEntry_[i])) {
          result = true;
          break;
      }
  }
  return result;
  }

  /**
   * Tests if a privilege is known.
   * @param  _name of the privilege to test
   * @return boolean true if the ACS has recorded this privilege instance
   */
  public final boolean containsPrivilegeName(ImmutableName _name) {
      boolean result = false;
      if (l_Privilege_ == null || _name == null) return result;
      for (int i = 0 ; i < l_Privilege_.length ; i++) {
          if (_name.equals(l_Privilege_[i].getName())) {
              result = true;
              break;
          }
      }
      return result;
  } // containsPrivilege()

  /**
   * Tests if a virtual folder is known.
   * @param  _name of the virtual folder to test
   * @return boolean true if the ACS has recorded this virtual folder instance
   */
  public final boolean containsVirtualFolderName(ImmutableName _name) {
      boolean result = false;
      if (l_VirtualFolders_ == null || _name == null) return result;
      for (int i = 0 ; i < l_VirtualFolders_.length ; i++) {
          if (_name.equals(l_VirtualFolders_[i].getName())) {
              result = true;
              break;
          }
      }
      return result;
  } // containsVirtualFolder()

  /**
   * Tests if the ACS owns the virtual folder.
   * @param  _name of the virtual folder to test
   * @return boolean true if this virtual folder is from this ACS
   */
  public final boolean containsOwnVirtualFolderName(ImmutableName _name) {
      boolean result = false;
      if (l_VirtualFolders_ == null || _name == null) return result;
      for (int i = 0 ; i < l_VirtualFolders_.length ; i++) {
          if (_name.equals(l_VirtualFolders_[i].getName()) &&
                  l_VirtualFolders_[i].getEorACS() == this) {
              result = true;
              break;
          }
      }
      return result;
  } // containsOwnVirtualFolder()

/**
 * @return true if, through its open external AclEntries or Linked Privileges,
 * the ACS is a controller of the target of another ACS, or if
 * it delivers some rights to an eligible party in another ACS.
 * Closed AclEntries and Privileges are not taken in account.
 */
 public boolean getIsExternalOriented()	{
    return externalOriented_;
 }

/**
 * Gets the presence of actors which are bridge targets.
 * @return boolean true if, for the bridge relations,
 * the ACS contains own bridge targets.
 */
 public boolean getContainsBridgeTarget() {
    return containsBridgeTarget_;
 }

/**
 * Gets the presence of actors which are bridge sources.
 * @return boolean true if, for the bridge relations,
 * the ACS contains own bridge sources.
 */
 public boolean getContainsBridgeSource() {
    return containsBridgeSource_;
 }

  /**
   * Get the standard ACS types, but not those created by the user.
   * Standard types are 'LINUX' and 'GENERIC'.
  * @return the registered ACS types which are used
  *   by the ACS constructors.
  */
  public static final List<String> getL_AcsTypes() {
      List result = new ArrayList(INITIAL_CAPACITY_5);
      result.add(LINUX);
      result.add(GENERIC);
      return result;
  }

  /**
   * Get the structural rules which are derived from the ACS structure and which
   * define the processing of objects and rights in this ACS.
  * @return the immutable rules which are derived from the ACS structure, without
  *           cloning of the list nor Immutable wrapping, to be fast. 
   *            Never empty nor null.
  */
  public final List<String> getL_ImmutableRules() {
      return l_ImmutableRules_;
  }
  
  /**
   * @return String comment which is associated to the acs. May be null.
   */
  public String getComment()	{
      return comment_;
  }
  
  /**
  * Set any comment related to this acs. It is a bound property.
  * @param _s comment which is associated to the acs.
  */
  public synchronized void setComment(String _s) {
      String oldComment = comment_;
      comment_ = _s;
      try {
              firePropertyChange("Comment",oldComment,comment_);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
      }
      oldComment = null;
      return;
 }

    /**
     * @param _obj object to compare 
     * @return true if the two objects are instances of classes
     *  that are ACSImpl or extend ACSImpl (and not ACSRun, ACSCatalog or ImmutableACS)
     */
    public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      ACSImpl obj = (ACSImpl) _obj;

      if (!aCSType_.equals(obj.getACSType())) return false;
      if (!m_Structure_.equals(obj.m_Structure_)) return false;
      if (!m_StandardRights_.equals(obj.m_StandardRights_)) return false;      
      if (!m_TypesAndGUIPolicy_.equals(obj.m_TypesAndGUIPolicy_)) return false; 
      if (!m_TypeOrientedPolicy_.equals(obj.m_TypeOrientedPolicy_)) return false;
      if (!m_ConstraintsForNewObjects_.equals(obj.m_ConstraintsForNewObjects_)) return false;
      
      // controls on AcsControllers names only, and in the same order, and tests of null values
      if (getL_AcsControllers().length != obj.getL_AcsControllers().length )
          return false;
      ImmutableName[] l_myCont = getL_AcsControllers();
      ImmutableName[] l_yourCont = obj.getL_AcsControllers();
      for (int i = 0; i < l_myCont.length; i++) {
            if (!l_myCont[i].equals( l_yourCont[i] ))
              return false;
      } // for on i
      l_myCont = null;
      l_yourCont = null;

      if (name_ == null || obj.getName() == null) {
          if (name_ != null || obj.getName() != null)
            return false;  // if
      } // if
      else {
        if (!name_.equals( obj.getName() )) return false; }

      if (ownResourcesRootName_ == null || obj.ownResourcesRootName_ == null) {
          if (ownResourcesRootName_ != null || obj.ownResourcesRootName_ != null)
            return false;  // if
      } // if
      else if (!ownResourcesRootName_.equals(obj.getOwnResourcesRootName()) ) return false;
      
      if (!ownGroupsRootName_.equals(obj.getOwnGroupsRootName()) ) return false;
      if (!ownResourceFoldersRootName_.equals(obj.getOwnResourceFoldersRootName()) ) return false;
      if (!ownEPFoldersRootName_.equals(obj.getOwnEPFoldersRootName()) ) return false;      
      
      if (!baseError_.equals(obj.getEorBaseError()) ) return false;

      if (m_Resources_ == null || obj.getEorM_Resources() == null) {
          if (m_Resources_ != null || obj.getEorM_Resources() != null)
            return false;  // if
      } // if
      else {
        if (!m_Resources_.keySet().equals( obj.getEorM_Resources().keySet() )) return false; }


      if (m_EligibleParties_ == null || obj.getEorM_OwnEligibleParties() == null) {
          if (m_EligibleParties_ != null || obj.getEorM_OwnEligibleParties() != null)
            return false;  // if
      } // if
      else {
        if (!m_EligibleParties_.keySet().equals( obj.getEorM_OwnEligibleParties().keySet() ))
          return false; }

      if (controlled_ != obj.getIsControlled() ) return false;
      if (externalOriented_ != obj.getIsExternalOriented() ) return false;

      if (l_ClosedAclEntriesACSNames_ != null ) {
          if (!l_ClosedAclEntriesACSNames_.equals(
            ((ACSImpl)obj).getL_ClosedAclEntriesACSNames()) ) return false;
      }

      if (l_AclEntry_ == null && obj.getEorL_AclEntries() != null) return false;
      if (l_AclEntry_.length != obj.getEorL_AclEntries().length )
          return false;
      for (int i = 0; i < l_AclEntry_.length; i++) {
        if (l_AclEntry_[i] == null || obj.getEorL_AclEntries()[i] == null) {
          if (l_AclEntry_[i] != null || obj.getEorL_AclEntries()[i] != null)
            return false;
          continue; // i++
        } // if
        else {
            if (!l_AclEntry_[i].equals( obj.getEorL_AclEntries()[i] ))
              return false;
        } // else
      } // for on i

      if (l_ClosedAclEntries_ != null) {
        if (!l_ClosedAclEntries_.equals( ((ACSImpl)obj).getL_ClosedAclEntries() )) return false;
      }
      
      if (l_Privilege_ == null && obj.getEorL_Privileges() != null) return false;
      if (l_Privilege_.length != obj.getEorL_Privileges().length )
          return false;
      for (int i = 0; i < l_Privilege_.length; i++) {
        if (l_Privilege_[i] == null || obj.getEorL_Privileges()[i] == null) {
          if (l_Privilege_[i] != null || obj.getEorL_Privileges()[i] != null)
            return false;
          continue; // i++
        } // if
        else {
            if (!l_Privilege_[i].equals( obj.getEorL_Privileges()[i] ))
              return false;
        } // else
      } // for on i
      
      if (l_VirtualFolders_ == null && obj.getEorL_VirtualFolders() != null) return false;
      if (l_VirtualFolders_.length != obj.getEorL_VirtualFolders().length )
          return false;
      for (int i = 0; i < l_VirtualFolders_.length; i++) {
        if (l_VirtualFolders_[i] == null || obj.getEorL_VirtualFolders()[i] == null) {
          if (l_VirtualFolders_[i] != null || obj.getEorL_VirtualFolders()[i] != null)
            return false;
          continue; // i++
        } // if
        else {
            if (!l_VirtualFolders_[i].equals( obj.getEorL_VirtualFolders()[i] ))
              return false;
        } // else
      } // for on i
      
      return true;
    } //equals()

    /**
   * @return the hashcode value for the ACS, based only on its type and its name.
   */

  public int hashCode() {
    int result = 0;
    result = aCSType_.hashCode();
    
    if (name_ != null) {
      result = PRIME*result + name_.hashCode();
    }

    return result;
  } // hashcode()

  /**
   * Get the name last component of the instance.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append(getNickName() );

    /*
        sb.append("; size of the resources map = ");
        sb.append(m_Resources_.size());

        sb.append("; size of the eligible parties map = ");
        sb.append(m_EligibleParties_.size());

        sb.append("; size of the acl entries array = ");
        sb.append(l_AclEntry_.length);
     
        sb.append("; comment_ = ");
        sb.append(comment_);
    */

    return " " + sb;
  } // toString()
  
  /**
   * A tied child ACS has to be opened when this instance is opened.
   * Like any child ACS, it has to be closed when this instance is closed, and 
   * it may be closed alone. The subACSs are not tied children.
   * @return the names of the child ACS which are tied up to this instance.
   *    May be empty, but never null.
   */
  public List<ImmutableName> getL_TiedChildACS() {
      if (l_TiedChildACSs_ != null)
          return (List<ImmutableName>)l_TiedChildACSs_.clone();
      else return new ArrayList(INITIAL_CAPACITY_5);
  }
  
  /**
   * A tied child ACS has to be opened when this instance is opened.
   * Like any child ACS, it has to be closed when this instance is closed, and 
   * it may be closed alone. Called by ACSFactoryImpl.createACS().
   * Fires the property change 'TiedChildACSs'.
   * @param _l_names the names of the child ACS which are tied up to this instance.
   *        May be null or empty.
   */
  public void setL_TiedChildACS(List<ImmutableName> _l_names) {
      List<ImmutableName> l_old = null;
      
      if (l_TiedChildACSs_ != null)
          l_old = (List<ImmutableName>)l_TiedChildACSs_.clone();
      
      if (l_TiedChildACSs_ == null)
          l_TiedChildACSs_ = new ArrayList(INITIAL_CAPACITY_5);

      if (_l_names == null)
          l_TiedChildACSs_.clear();
      else {
          // does not check that each name is the name of a child acs
          //if (!BaseManagerImpl.getInstance().getISTree(iSname_).
            //      getL_DirectChildren(name_).containsAll(_l_names))
              //          throw new InternalError("ACSImpl.setL_TiedChildACS() "+
                //                " sets with at most a name which is not a direct child.");

          l_TiedChildACSs_.clear();
          if (!_l_names.isEmpty())
              l_TiedChildACSs_.addAll(_l_names);
      } // else

      try {
            firePropertyChange("TiedChildACSs", l_old, 
                    (List<ImmutableName>)l_TiedChildACSs_.clone());
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setL_TiedChildACS(), firePropertChange() throws:");
            e.printStackTrace();
      } 
  } // setL_TiedChildACS()

  /**
   * Gets the subACS names and their mode (for Resources or for GroupIDs) and the
   * root names for groups trees, as nodes to display in the explorer.
   * A subACS is always open with its ACS.
   * In the returned value, the key is the subACS name or a group root name, in one word,
   * and the value is an array which may contain one or two items among the following values: <P>
   * &lt;ResourceSubACS&gt;, &lt;GroupSubACS&gt; and &lt;GroupTree&gt;. <P>
   * For instance, the value 'yyy' for the key &lt;GroupTree&gt;,
   * means that a group root node '&lt;yyy_tree&gt;' has to be displayed
   * directly under the GroupIDs node, in the explorer. For the two other keys,
   * a value 'xxxx' means that a node '&lt;xxxx_subacs&gt;' has to be displayed directly
   * under the Resources and the AclEntries nodes, or under the GroupIDs node.
   * @return the cloned map of the subACSs and the groups tree roots.
   * May be empty but never null.
   */
  public Map<String, String[]> getM_SubAndGroupPolicy() {
      if (m_SubAndGroupPolicy_ != null)
          return (HashMap<String, String[]>)m_SubAndGroupPolicy_.clone();
      else return new HashMap(INITIAL_CAPACITY_5);      
  }  

  /**
   * Gets the map which says if the ACS or a subACS belongs to a Composite ACS
   * in the IS Name Tree. <P>
   * In the returned value, the key is the subACS name, in one word, 
   * or the ACS name last component, and the value is the Composite name.
   * @return the map of the composite relations. Never null and may be empty.
   */
  public Map<String, ImmutableName> getM_BelongsToComposite() {
      if (m_BelongsToComposite_ != null)
          return (HashMap<String, ImmutableName>)m_BelongsToComposite_.clone();
      else return new HashMap(INITIAL_CAPACITY_5);      
  } 

  /**
   * For a parent ACS, gets the list of the inner ACSObjects having with an opened child ACS,
   * a relation among &lt;ForChildAcsOpeningAndClosing&gt;, &lt;ForChildAcsOpeningNoClosing&gt;
   * and &lt;ForChildAcsIntegrity&gt;.
   * These relations set that the deleting or the updating of the ACSObject will
   * stop the child ACS or will modify its ability to run correctly.
   * @see #getM_RootCommandsForOpeningAndClosing
   * @see #getM_RootCommandsForOpeningNoClosing
   * @see #getM_RootCommandsForIntegrity
   * @return the map of the inner ACSObjects which support at least one child ACS.
   * The child ACS names are the keys, and the values are arrays of root names
   * from this ACS. Never null and may be empty.
   */
  public Map<String, ImmutableName[]> getM_InnerRootsForChildACS() {
      if (m_RootsForChildACS_ != null)
          return (Map<String, ImmutableName[]>)((HashMap)m_RootsForChildACS_).clone();
      else return new HashMap(INITIAL_CAPACITY_5);
  } // getM_InnerRootsForChildACS()

  /**
   * Gets for a child ACS, the list of associations to roots in its parent ACSs
   * for running. <P>
   * This relation set that the root parent ACSObject, or one of its property value,
   * has to be there when the child ACS is opened, to ensure its ability to exist,
   * and has to be closed or updated (for a property value) at its closing.
   * @see AssociateUtilityImpl
   * @return The map key has the format: <P>
   * - ACS.Name.&lt;Value&gt;.XXX: a parent ACS name having the value XXX. <P>
   * The map value is a sequence of commands which are executed by AssociateUtilityImpl.
   * Never null.
   */
  public Map<String, String[]> getM_RootCommandsForOpeningAndClosing() {
      if (m_RootCommandsForAcsOpeningAndClosing_ != null)
          return (HashMap<String, String[]>)m_RootCommandsForAcsOpeningAndClosing_.clone();
      else return new HashMap(INITIAL_CAPACITY_5);       
  }

  /**
   * Gets for a child ACS, the list of associations to roots in its parent ACSs
   * for running. <P>
   * This relation set that the root parent ACSObject, or one of its property value,
   * has to be there when the child ACS is opened, to ensure its ability to exist,
   * but there is NO constraint to free the associations at its closing.
   * @see AssociateUtilityImpl
   * @return The map key has the format: <P>
   * - ACS.Name.&lt;Value&gt;.XXX: a parent ACS name having the value XXX. <P>
   * The map value is a sequence of commands which are executed by AssociateUtilityImpl.
   */
  public Map<String, String[]> getM_RootCommandsForOpeningNoClosing() {
      if (m_RootCommandsForAcsOpeningNoClosing_ != null)
          return (HashMap<String, String[]>)m_RootCommandsForAcsOpeningNoClosing_.clone();
      else return new HashMap(INITIAL_CAPACITY_5);       
  }
  
  /**
   * Gets for a child ACS, the list of associations to roots in its parent ACSs
   * for integrity. <P>
   * This relation set that the root parent ACSObject, or one of its property value,
   * has to be there when the child ACS is opened, to ensure its ability to work well,
   * in other words, following its specifications.
   * @see AssociateUtilityImpl
   * @return The map key has the format: <P>
   * - ACS.Name.&lt;Value&gt;.XXX: a parent ACS name having the value XXX. <P>
   * The map value is a sequence of commands which are executed by AssociateUtilityImpl.
   */
  public Map<String, String[]> getM_RootCommandsForIntegrity() {
      if (m_RootCommandsForAcsIntegrity_ != null)
          return (HashMap<String, String[]>)m_RootCommandsForAcsIntegrity_.clone();
      else return new HashMap(INITIAL_CAPACITY_5);       
  }

  /**
   * Gets the list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root has to be in the parent ACS at the opening of this ACS to ensure its
   * working.
   * There is no common roots with the values in getL_RootsForIntegrity(), but
   * some common values may exist with getL_RootsForClosing().
   * This list is generated by AssociateUtilityImpl from the commands in
   * getM_RootCommandsForOpeningAndClosing() and getM_RootCommandsForOpeningNoClosing().
   * @return the root names. May be empty but never null.
   */
  public List<ImmutableName> getL_RootsForOpening() {
      List<ImmutableName> l_result = null;
      if (l_RootNamesForAcsOpeningWithClosing_ != null)
          l_result = (List)((ArrayList)l_RootNamesForAcsOpeningWithClosing_).clone();
      if (l_RootNamesForAcsOpeningNoClosing_ != null) {
          if (l_result == null)
              l_result = (List)((ArrayList)l_RootNamesForAcsOpeningNoClosing_).clone();
          else
              l_result.addAll(l_RootNamesForAcsOpeningNoClosing_);
      }

      if (l_result == null)
          l_result = new ArrayList(INITIAL_CAPACITY_5);
      return l_result;
  } // getL_RootsForOpening()

  /**
   * Gets the list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root has to be deleted in the parent ACS at the closing of this ACS.
   * There is no common roots with the values in getL_RootsForIntegrity(), but
   * all the values should exist in getL_RootsForOpening().
   * This list is generated by AssociateUtilityImpl from the commands in
   * getM_RootCommandsForOpeningAndClosing().
   * @return the root names. May be empty but never null.
   */
  public List<ImmutableName> getL_RootsForClosing() {
      if (l_RootNamesForAcsOpeningWithClosing_ != null)
          return (List)((ArrayList)l_RootNamesForAcsOpeningWithClosing_).clone();
      else return new ArrayList(INITIAL_CAPACITY_5);
  }

  /**
   * Get the list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root in the parent ACS has some property updatings which may weaken
   * the integrity of this ACS. There is no common roots with the values in
   * getL_RootsForOpening() and getL_RootsForClosing().
   * This list is generated by AssociateUtilityImpl from the commands in 
   * getM_RootCommandsForIntegrity().
   * @return the root names. May be empty but never null.
   */
  public List<ImmutableName> getL_RootsForIntegrity() {
      if (l_RootNamesForAcsIntegrity_ != null)
          return (List)((ArrayList)l_RootNamesForAcsIntegrity_).clone();
      else return new ArrayList(INITIAL_CAPACITY_5);
  }

  /**
   * Called by BeamerManager.getNotDisplayThisProperty().
   * @return true if the ACS has some roots in a parent ACS
   */
  public boolean getContainsRootActions() {
      return containsRootActions_;
  }

  /**
   * Gets comments to explain how this ACS simulates a real system/software.
   * The criteria is an ACS property displaying name. The value is a String array
   * of free user comments on the settings of the related ACS property. <P>
   * @return an unmodifiable Map where the key is a String, and the value is an array
   *   of Strings. May be empty, but not null. A map value may be empty but is never null.
   */
  public final Map<String, String[]> getM_CommentsOnProperties() {
      if (m_CommentsOnProperties_ == null) return new HashMap(INITIAL_CAPACITY_5);

      return Collections.unmodifiableMap(m_CommentsOnProperties_);
  } // getM_CommentsOnProperties()

  /**
   * Gets the ACS vocabulary. This may cover several independant domains: <P>
   * - the AcsAddon editor, the AcsAddon package editor and the AcsAddon name, in three keys, <P>
   * - how to name the L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS constants for this ACS,
   * as names of the ACS tree nodes in the proper ACS internal frame. An example
   * is to replace the tree node 'UserIDs (right user)' by 'Accounts'. <P>
   * - what are the gBase classes to use, from the generic package or from
   * the AcsAddon gBase package (see L_CLASSIC_CLASS_KEYS), like when
   * the Addon Ubuntu subclasses ResourceImpl with ResourceUbuntuImpl. The
   * full Java name of the subclass has to be put in the value array. <P>
   * - If the ACS is an AcsAddon, what are the specific new classes in
   * the AcsAddon gBase package (see ACSADDON_NEW_CLASSES). In this case,
   * the class names are separated by one space, like in
   * 'Executable/is/Actor XXX/is/Resource'. <P>
   * - If the ACS is an AcsAddon, what are the non-generic rights in the
   * ACSObjects on which the ACS constraints of the specialized rights
   * are applicable, and in that case, what domain is applicable. An example
   * is the domain 'GpORightsSet' for the property 'OtherRights', in Ubuntu,
   * that is declared by the value 'OtherRights.GpORightsSet'. <P>
   * The method getEditorAndAddonNames() also returns the editor and AcsAddon names.
   * @since 0.7.0
   * @see ACSFactoryUtilityImpl#L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS
   * @see ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ACSFactoryUtilityImpl#ACSADDON_EXTENSION_CLASSES
   * @return an unmodifiable Map where the key is a String, and the value 
   * is a String array. May be empty, but not null.
   * A map value is never null but may be empty.
   */
  public final Map<String, String[]> getM_AcsVocabulary() {
      if (m_AcsVocabulary_ == null) return new HashMap(INITIAL_CAPACITY_5);

      return Collections.unmodifiableMap(m_AcsVocabulary_);
  } // getM_AcsVocabulary()

  /**
   * Gets the structure of this ACS. For the generic properties, that are
   * defined in gBase.ACSImpl), there is a dedicated method getManageXXX to read
   * the value of each property.
   * Some complementary, non-generic values are allowed for the AcsAddons.
   * This map may be directly updated at the creation of an AcsAddon ACS,
   * for adding new keys of non-generic properties which are ridden
   * by their proper AcsAddon classes.
   * @return the unmodifiable ACS structure. Never empty nor null.
   */
  public SortedMap<String, Boolean> getM_Structure() {
      return Collections.unmodifiableSortedMap(m_Structure_);
  }

  /**
   * Gets the ACS rights and the metarights this ACS handles and for a given internal key value
   * of the associated metaright. The map key is the metaright internal key.
   * @return the unmodifiable rights map. Never empty nor null.
   */
  public SortedMap<String, StringRight[]> getM_RightsOfInternalKey() {
      return Collections.unmodifiableSortedMap(m_RightsOfInternalKey_);
  } // getM_RightsOfInternalKey()

  /**
  * Gets the ACS or meta rights having one of the arguments as nick name, and
  * used by the ACS.
  * Called by AssociateUtilityImpl.addAssociateActions() to get some rights array
  * for the roots to create. This is why the five arguments are not put in a single array or list.
  * @param _r1 nick name of an ACS right. May be null.
  * @param _r2 nick name of an ACS right. May be null.
  * @param _r3 nick name of an ACS right. May be null.
  * @param _r4 nick name of an ACS right. May be null.
  * @param _r5 nick name of an ACS right. May be null.
  * @return from zero to five required acs rights or meta rights
  */
  public StringRight[] getL_SelectedRights(String _r1, String _r2, String _r3,
      String _r4, String _r5) {
        List<StringRight>   l_result = new ArrayList(5);
        StringRight         right = null;

        right = getOneRight(_r1);
        if (right != null) l_result.add(right);
        right = getOneRight(_r2);
        if (right != null) l_result.add(right);
        right = getOneRight(_r3);
        if (right != null) l_result.add(right);
        right = getOneRight(_r4);
        if (right != null) l_result.add(right);
        right = getOneRight(_r5);
        if (right != null) l_result.add(right);

        right = null;
        if (!l_result.isEmpty())
            return l_result.toArray(new StringRight[0]);
        else {
            l_result = null;
            return new StringRight[0];
        }
  } // getL_SelectedRights()

  /**
   * Gets the reference in this ACS for an external or internal soft alias,
   * by the way of a specific algorithm. The internal reference may have several aliases.
   * This method returns null in this ACSImpl class.
   * It has to be overridden in the AcsAddons
   * where an external or internal soft alias is managed, and it is then
   * responsible to update the soft alias and the soft reference properties.
   * This means that a call to this method is the only mean to update these bound
   * properties. If the soft aliasing links may vary, it is the responsivility
   * of the AcsAddon to call this method at the right moment. This is true
   * typically when an object is deleted, or when there is a change into
   * the parameters defining the soft aliasing rules.
   * This method is not declared in the ACS BeanInfo.
   * @param _alias ACSObject from another ACS
   * @return the internal ACSObject which is the reference of the argument.
   * May be null.
   */
  public ACSObject getReferenceOfSoftAlias(ACSObject _alias) {
      return null;
  }

  //  
  // --------------  methods out of the interfaces ----------------
  //

  
  /**
   * Used only by ACSFactoryImpl or a subclass. No property change firing.
   * For an AcsAddon, the icon is searching in the AcsAddon gBase package, not
   * in the Gui1/images package. For the copy of an open ACS, ActionNewACS
   * sets imageIcon_ in the ACSFactoryImpl, and it is used to set directly
   * icon_ in this instance, without calling to this method.
   * @param _iconName file name of the icon associated to the ACS. Never null.
   */
  protected void setEorIcon(String _iconName) {
    try {
        
        if (getClass().toString().indexOf("ARoad0.AcsAddon.") == -1) {
            icon_ = new ImageIcon(
                    Desktop.AR_URI.resolve("Gui1"+Desktop.URL_SEPAR+"images"+
                        Desktop.URL_SEPAR+_iconName).toURL() );
        }
        else {

            String editor = ((String[])m_AcsVocabulary_.get(
                    ACSFactoryUtilityImpl.ACSADDON_PACKAGE_EDITOR))[0];
            String addon = ((String[])m_AcsVocabulary_.get(
                    ACSFactoryUtilityImpl.ACSADDON_NAME))[0];

            icon_ = new ImageIcon(
                    Desktop.AR_URI.resolve("AcsAddon"+Desktop.URL_SEPAR+editor+
                        Desktop.URL_SEPAR+addon+Desktop.URL_SEPAR+"gBase"+
                        Desktop.URL_SEPAR+_iconName).toURL() );
        } // else

    } catch (MalformedURLException m) {
      throw new InternalError("ACSImpl.setEorIcon(): "+m.toString());
    }
    iconName_ = _iconName;
  } // setEorIcon()
  
  /**
  * Registers a new own resource with a type control.
  * This method restricted to calls from resource constructors, to
  * control if the resource type is allowed by theis acs, the call addResource()
  * which registers the resource and fires the events.
  * Overridden by the AcsAddon subclasses.
  * Calls addResource().
  * @param _res is just created in a Resource constructor, or
  *   external resource which is associated to an aclEntry owned by this ACS
  * @exception CreateError (and baseError) if the Name is already known
  *  in ACSCatalog, if the Name is null, or if the resource type is not allowed
  *  by this ACS
  */
  protected synchronized void newEorResource(ImmutableResource _res) throws CreateError {

    if (_res == null) {
      baseError_.addException("In ACSImpl.newEorResource(Name), the resource is null");

                // for testing and tracking
                //System.out.println("in ACSImpl.newEorResource(), "+
                //  ", the resource is null");

      throw new CreateError("The resource is null.");
    }
    if (_res.getName() == null) {
      baseError_.addException("In ACSImpl.newEorResource(Name), the Name is null");

                // for testing and tracking
                //System.out.println("in ACSImpl.newEorResource(), "+
                //  ", the Name is null");

      throw new CreateError("The Name is null.");
    }

    String[] l_restrictedTypes = null;
    if (!(_res instanceof Actor))
        l_restrictedTypes = (String[]) m_TypesAndGUIPolicy_.get("Resource.ResourceType");
    else {
        l_restrictedTypes = (String[]) m_TypesAndGUIPolicy_.get("Actor.ResourceType");
        if (l_restrictedTypes == null)
           l_restrictedTypes = (String[]) m_TypesAndGUIPolicy_.get("Resource.ResourceType");
    } // else

    if (l_restrictedTypes != null) {
        Set l_limitedTypes = new HashSet(INITIAL_CAPACITY_25);
        for (int i = 0; i < l_restrictedTypes.length; i++) {
          l_limitedTypes.add(l_restrictedTypes[i]);
          
                          // for testing and tracking
                //System.out.println("in ACSImpl.newEorResource(), "+l_restrictedTypes[i]+
                  //" is a resource type.");
          
        } // for i
        String type = _res.getResourceType();
        
        if (type == null && 
          !l_limitedTypes.contains(null)) {
            l_limitedTypes = null;
            l_restrictedTypes = null;
            type = null;
            throw new CreateError(" The resource type of "+_res.getName()+
                    " is null, and it is forbidden by the ACS.");
        } else if (type != null && 
          !l_limitedTypes.contains(type)) {
            l_limitedTypes = null;
            l_restrictedTypes = null;
            type = null;
                      
                          // for testing and tracking
                //System.out.println("in ACSImpl.newEorResource(), "+type+
                  //" is not a resource type.");
            
            
            throw new CreateError(" The resource type of "+_res.getName()+
                    " is not allowed by the ACS.");
        }
        l_limitedTypes = null;
        l_restrictedTypes = null;
    } // if
    
    addResource(_res);
  } // newEorResource()

  /**
  * Registers a new own or a controlled resource without type control,
  * and fires the bound properties 'M_OwnResources' or
  * 'OneExternalResource', 'OneResource' and 'OneBaseObject' only if it is not
  * an Actor.
  * External Object Reference: _res.getName() and _res are directly put
  * in the resource map.
  * Called from newEorResource(), newEorAclEntry() for external AclEntries,
  * PrivilegeForLinksImpl.addEorPrivilegeTargetLink() for external Linked Privileges.
  * @param _res is just created in a Resource constructor, or
  *   external resource which is associated to an aclEntry owned by this ACS
  * @exception CreateError (and baseError) if the Name is already known,
  *  if the Name is null.
  */
  protected void addResource(ImmutableResource _res) throws CreateError {
    if (m_Resources_.containsKey(_res.getName()) || 
            BaseManagerImpl.getInstance().knowsACS(_res.getName()) || 
            BaseManagerImpl.getInstance().knowsView(_res.getName() )) {
      baseError_.addException("In ACSImpl.addResource(Name): "+
        "this name is already known in the ACS, or as an ACS or view name.");

                // for testing and tracking
                //System.out.println("in ACSImpl.addResource(), "+
                //  ", the Name is already known in ACSCatalog");

      throw new CreateError(" The name is already known in the ACS"+
              " as Resource, or as ACS or view.");
    }
    
    SortedMap m_OldResources;
    if (_res.getEorACS() == this)        
          m_OldResources = getEorM_OwnResources();
    else m_OldResources = getEorM_ExternalResources();
    
    synchronized(m_Resources_) {
      m_Resources_.put( (ImmutableName)_res.getName(), _res);
    }

    try {
        if (_res.getEorACS() == this ) {
            firePropertyChange("M_OwnResources", m_OldResources, getEorM_OwnResources());
        }
        else {
            // adds this ACS in the resource ACS as controller
            if (!Arrays.asList(_res.getEorACS().getL_AcsControllers()).contains(name_)) {
                ((ACSImpl)_res.getEorACS()).addAcsController(name_); // throw UpDateError
            } // if 

            if (!externalOriented_) {
                externalOriented_ = true;
                firePropertyChange("ExternalOriented", false, externalOriented_);
            }

            firePropertyChange("OneExternalResource", null,_res);
        } // else

        firePropertyChange("OneResource", null, _res);
        // if _res is an ImmutableActor, 'OneBaseObject' is fired later,
        // through newEorEligibleParty()
        if (!(_res instanceof ImmutableActor))
            firePropertyChange("OneBaseObject",null,_res);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updatinge, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.addResource(), firePropertChange() throws:");
        e.printStackTrace();
    }
        
                // for testing and tracking
                //System.out.println("in ACSImpl.addEorResource(), "+
                //  "the new resource is: "+_res.getName());

    m_OldResources = null;
  } // addResource()  
  
  /**
   * Unregisters the own and the external resources.
   * An own resource is already unregistered in its parent, GroupID and UserID;
   * its AclEntries are already deleted. Its ACS and ACS name are
   * still known by the resource. <P>
   * Updates the resources root name if _res is this root.
   * Called by removeEorAclEntry() and PrivilegeForLinks.removeEorPrivilegeTargetLink()
   * for an external resource. Not called by removeEorPrivilege().
   * For an external resource, calls removeAcsController() on its ACS.
   * Fires property change events on "M_OwnResources" or "OneExternalResource",
   * and fire "OneResource", "OneBaseObject".
   * Overridden by the AcsAddon subclasses.
   * @param _res internal or external open ImmutableResource to be taken off in the ACS.
   * @exception UpDateError (and baseError) if the Name is unknown in the ACS.
   */
  protected synchronized void removeEorResource(ImmutableResource _res) throws UpDateError {
      
                        // for testing and tracking
                        //System.out.println("in ACSImpl.removeEorResource(), on "+
                          //      name_.lastComponent()+
                            //    "the resource is: "+_res.getName());

    if (!m_Resources_.containsKey(_res.getName() )) {
      baseError_.addException("In ACSImpl.removeEorResource(Resource), "+
        "the Resource "+_res.getName()+" is unknown in the ACS");
      throw new UpDateError("In ACSImpl.removeEorResource(Resource), "+
        "the Resource "+_res.getName()+" is unknown in the ACS");
    } // if
    
    SortedMap m_OldResources;
    if (_res.getEorACS() == this)        
          m_OldResources = getEorM_OwnResources();
    else m_OldResources = getEorM_ExternalResources();
    
    synchronized (m_Resources_) {
        m_Resources_.remove( (ImmutableName)_res.getName());
    }

    if (ownResourcesRootName_ != null &&
            ownResourcesRootName_.equals(_res.getName())) 
                 setOwnResourcesRootName(null);

    try {
        if (_res.getEorACS() == this) {
          firePropertyChange("M_OwnResources", m_OldResources, getEorM_OwnResources());
        }
        else {
            boolean controllerOfResAcs = false;
            // removes this ACS in the EP ACS if necessary
            for (ImmutableResource res : getEorM_ExternalResources().values() ) {
                // m_Resources_ is already updated
                if ( res.getAcsName().equals(_res.getAcsName() ) ) {

                        // for testing and tracking
                        //System.out.println("in ACSImpl.removeEorResource(), on "+
                          //      name_.lastComponent()+
                            //    ", the external resource to keep is: "+res.getName());

                  controllerOfResAcs = true;
                  break;
                } // if
            } // for res
            if (!controllerOfResAcs) {
                // unregistered this as a controller in resource ACS
                ((ACSImpl)_res.getEorACS() ).removeAcsController(name_); // throw UpDateError
            } // if (!controllerOfEpAcs)

            if (!controllerOfResAcs)
                setExternalOriented();

            firePropertyChange("OneExternalResource", _res, null);
        } // else

        firePropertyChange("OneResource", _res, null);
        // for an actor, which is also handled as an eligible party in the TreeBaseListeners,
        // removeEorEligibleParty() is called before this method but do not fire 'OneBaseObject'
        firePropertyChange("OneBaseObject", _res, null);
     } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.removeEorResource(), firePropertChange() throws:");
        e.printStackTrace();
    }

    m_OldResources = null;
  } // removeEorResource()

  /**
   * Registers a new own eligible party in the ACS and
   * controls that the eligible party type is allowed. <P>
   * Calls addEligibleParty() which does the work and fires a property change on
   * "M_OwnEligibleParties", "OneEligibleParty" and "OneBaseObject". <P>
   * Overridden by the AcsAddon subclasses.
   * Called from the subclasses and EligibleParty implementations.
   * @param       _ep is just created in EligibleParty()
   *                or in ActorImpl(), or associated to a newEorAclEntry() call
   * @exception   CreateError (and baseError) if the Name is already known
   *                in ACSCatalog, if the Name is null or if the ep type is not allowed.
   */
  protected synchronized void newEorEligibleParty(ImmutableEligibleParty _ep) throws CreateError {

    if (_ep == null) {
      baseError_.addException("In ACSImpl.newEorEligibleParty(Name), "+
                "the eligible party is null");

                // for testing and tracking
                //System.out.println("in ACSImpl.newEorEligibleParty(), "+
                  //"baseError_: "+baseError_.toString());

      throw new CreateError("From ACSImpl.newEorEligibleParty(), "+
              "the eligible party is null.");
    }
    if (_ep.getName() == null) {
      baseError_.addException("In ACSImpl.newEorEligibleParty(Name), "+
                "the Name of the eligibleparty is null");

                // for testing and tracking
                //System.out.println("in ACSImpl.newEorEligibleParty(), "+
                  //"baseError_: "+baseError_.toString());

      throw new CreateError("From ACSImpl.newEorEligibleParty(), "+
              "the name of the EligibleParty is null in "+getNickName());
    }

    String[] l_restrictedTypes = null;

    if (!(_ep instanceof Actor))
        l_restrictedTypes = (String[]) m_TypesAndGUIPolicy_.get("EligibleParty.EPType");
    else {
        l_restrictedTypes = (String[]) m_TypesAndGUIPolicy_.get("Actor.ResourceType");
        if (l_restrictedTypes == null)
           l_restrictedTypes = (String[]) m_TypesAndGUIPolicy_.get("EligibleParty.EPType");
    } // else

    if (l_restrictedTypes != null) {
        Set l_limitedTypes = new HashSet(INITIAL_CAPACITY_25);
        for (int i = 0; i < l_restrictedTypes.length; i++) {
          l_limitedTypes.add(l_restrictedTypes[i]);
        } // for i
        String type = _ep.getEPType();
        
        if (type == null ) {
            l_limitedTypes = null;
            l_restrictedTypes = null;
            throw new CreateError("From ACSImpl.newEorEligibleParty(), "+
                    "the type is null and "+
                      "it is forbidden by the ACS.");
        } else if (type != null && 
          !l_limitedTypes.contains(type)) {
            l_limitedTypes = null;
            l_restrictedTypes = null;
            //type = null;
            throw new CreateError("From ACSImpl.newEorEligibleParty(), "+
                    "the type "+type+" is not allowed by the ACS "+
                    getNickName());
        }
        l_limitedTypes = null;
        l_restrictedTypes = null;
    } // if
    
                // for testing and tracking
                //System.out.println("Call of the ACS "+name_+".addEligibleParty() for "+
                  //      _ep.getName());
    
    addEligibleParty(_ep);
  } // newEorEligibleParty()


  /**
   * Registers a new own or an external eligible party in the ACS, without
   * type control.
   * Fires property changes on 'M_OwnEligibleParties' or 'OneExternalEligibleParty',
   * 'OneEligibleParty' and 'OneBaseObject'.
   * Called by newEorAclEntry() for external AclEntries, newEorEligibleParty(),
   * PrivilegeForLinksImpl.addEorPrivilegeSourceLink() for external Linked Privileges.
   * @param  _ep is just created in EligibleParty()
   * or in ActorImpl(), or associated to a newEorAclEntry() call
   * @exception CreateError (and baseError) if the Name is already known
   * in the base, if the Name is null.
   */
  protected void addEligibleParty(ImmutableEligibleParty _ep) throws CreateError {
    if (m_EligibleParties_.containsKey(_ep.getName()) || 
            BaseManagerImpl.getInstance().knowsACS(_ep.getName()) || 
            BaseManagerImpl.getInstance().knowsView(_ep.getName())
            ) {
      baseError_.addException("In ACSImpl.addEligibleParty(Name), "+
        "the name "+_ep.getName()+" is already known in the ACS or as an ACS or view name.");
      
                // for testing and tracking
                //System.out.println("in ACSImpl.newEorEligibleParty(), baseError_: "+baseError_.toString());

      throw new CreateError(" The name is already known in the ACS "+
              "as EligibleParty, or as ACS or view.");
    } // if
    
                                            // for testing and tracking
                                            //System.out.println("Call to the ACS "+name_+
                                              //      ".addEligibleParty() for"+_ep);

    SortedMap m_OldOwnEligibleParties = getEorM_OwnEligibleParties();
    
    synchronized (m_EligibleParties_) {
      m_EligibleParties_.put( (ImmutableName)_ep.getName(), _ep);
    }

    try {
        if (_ep.getEorACS() == this) {
            firePropertyChange("M_OwnEligibleParties", m_OldOwnEligibleParties,
                getEorM_OwnEligibleParties());
        }
        else {
            // also for ACSTreeBaseListenerImpl
            firePropertyChange("OneExternalEligibleParty", null, _ep);
        }
        firePropertyChange("OneEligibleParty", null, _ep);
        firePropertyChange("OneBaseObject", null, _ep);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.addEligibleParty(), firePropertChange() throws:");
        e.printStackTrace();
    }

    m_OldOwnEligibleParties = null;
  } // addEligibleParty()  
  
  /**
   * Removes the own and the external EligibleParty.
   * May update the closed bridge targets map.
   * Called by PrivilegeForLinks.removeEorPrivilegeTargetLink().
   * Fire property changes on 'M_OwnEligibleParties' or 'OneExternalEligibleParty',
   * and 'OneEligibleParty'.
   * @param _ep to remove
   * @exception UpDateError (and BaseError) if the name is unknown
   */
  protected synchronized void removeEorEligibleParty(ImmutableEligibleParty _ep)
          throws UpDateError {
    if(!m_EligibleParties_.containsKey(_ep.getName() )) {
      baseError_.addException("In ACSImpl.removeEorEligibleParty(),"+
        " the EligibleParty is unknown in the ACS");
      throw new UpDateError("In ACSImpl.removeEorEligibleParty(),"+
        " the EligibleParty "+_ep+" is unknown in the ACS "+name_);
    }

    SortedMap m_OldOwnEligibleParties = getEorM_OwnEligibleParties();

    synchronized (m_EligibleParties_) {
      m_EligibleParties_.remove(_ep.getName() );
    }

    if (ownGroupsRootName_.equals(_ep.getName())) setOwnGroupsRootName(null);

                                    // for testing and tracking
                                    //System.out.println("call to ACSImpl.removeEorEligibleParty() on "+
                                      //      _ep.getNickName()+", and fires 'OneEligibleParty'");

    try {
        if (_ep.getEorACS() == this) {
          firePropertyChange("M_OwnEligibleParties", m_OldOwnEligibleParties,
            getEorM_OwnEligibleParties());
        }
        else {
            // also for ACSTreeBaseListenerImpl
            firePropertyChange("OneExternalEligibleParty", _ep, null);
        }

        firePropertyChange("OneEligibleParty", _ep, null);
        // if _ep is an ImmutableActor, 'OneBaseObject' is fired later,
        // through removeEorResource()
        if (!(_ep instanceof ImmutableActor))
            firePropertyChange("OneBaseObject", _ep, null);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.removeEorEligibleParty(), firePropertChange() throws:");
        e.printStackTrace();
    }

    // removes _ep if it is in a bridge source in a closed bridge link
    if (_ep instanceof ActorImpl && m_closedTargetsOfActorBridgeSources_ != null &&
      m_closedTargetsOfActorBridgeSources_.containsKey(_ep.getName())) {
          
          m_closedTargetsOfActorBridgeSources_.remove(_ep.getName());
          
    } // if (_ep instanceof ImmutableActor
    
    m_OldOwnEligibleParties = null;
  } // removeEorEligibleParty()

  /**
   * Method called by Resource.addAclEntry() before the updating of
   * Resource.getL_aclEntry() and, indirectly, also called by checkClosedAclEntries(). <P>
   * If the resource ACS is different to this ACS and if it is unknown
   * by this ACS, the resource is recorded as an external one. If the resource ACS
   * unknowns this ACS as a controller, this ACS is recorded as a controller
   * in the resource ACS, but only if the acl ACS is not closed.
   * Calls setExternalOriented().
   * The ACL rights are not controlled there, like in the new methods for resource
   * and for eligible party, because the ACS restrictions are controlled
   * in the AclEntry constructor. <P>
   * The change events are fired after the updating of the AclEntry internal list.
   * Fires an event 'AclEntry' without copy, and for an external object,
   * 'OneExternalResource', 'OneResource', or 'OneExternalEligibleParty' and 'OneEligibleParty'.
   * May invoke addResource() and addEligibleParty()
   * if _acl is the first aclEntry to associate these external objects to this ACS.
   * The sequence of the property change firings is: events from the
   * external addResource/EligibleParty methods, 'Controlled' and 'OneController'
   * from the controlled ACS, 'ExternalOriented', then 'AclEntry'. <P>
   * Caution: These property changes are not to be used for updating the views,
   * since they are fired before the associated firings from
   * the AclEntry Resource and EligibleParty on which the changes are followed by the views.
   * @param _acl new AclEntry managed by this ACS. It is associated to
   * one resource and one eligible party, and they may belong to this ACS or not
   * @exception   CreateError (and baseError) if the resource name is null,
   * if the _acl ACS is closed or null.
   * No test of unicity for _acl, done in Resource.addAclEntry()
   */
  protected synchronized final void newEorAclEntry(ImmutableAclEntry _acl)
    throws CreateError {
      // exception
      if (_acl.isEmpty() ) throw new CreateError(" The new AclEntry is empty.");
      if (!name_.equals( ((AclEntryImpl)_acl).getAcsName() ) )
        throw new CreateError(" The ACS name of the new AclEntry is not correct.");
      
      // variables
      boolean             recorded = false;
      ImmutableName       resAcsName = ((AclEntryImpl)_acl).getNameOfResACS();
      ImmutableResource   aclResource = _acl.getEorResource();
      ImmutableEligibleParty  aclEP = _acl.getEorEP();
      Set<ImmutableName>  l_openACS = new HashSet(BaseManagerImpl.getInstance().
                                        getEorM_OpenACS().keySet());
      
      //exception
      if (resAcsName == null || aclResource == null )
        throw new CreateError(" The resource ACS name or the reference to resource "+
            "is not correct.");

                          // for testing and tracking
                          //System.out.println(
                            //      "ACSImpl.newEorAclEntry() on "+name_.lastComponent()+
                              //    " starting, for the ACL: "+ _acl);
      
      // adds _acl in l_AclEntry_
          // in a new array if l_AclEntry_ == null
      if (l_AclEntry_ == null || l_AclEntry_.length == 0)
        l_AclEntry_ = new AclEntryImpl[]{ ((AclEntryImpl)_acl) };
      else {
          // in an empty place if possible
        for (int i = 0 ; i < l_AclEntry_.length ; i++) {
            if (l_AclEntry_[i] == null || l_AclEntry_[i].isEmpty() ) {
                l_AclEntry_[i] = _acl;
                recorded = true;
                break;
            }
        }
            // or at the end of l_AclEntry_
        if (!recorded) {
            ImmutableAclEntry[] newArray = new AclEntryImpl[
              l_AclEntry_.length + 1];
            System.arraycopy(l_AclEntry_, 0, newArray, 0, l_AclEntry_.length);
            newArray[l_AclEntry_.length] = _acl;
            //l_AclEntry_ = new AclEntryImpl[newArray.length];
            l_AclEntry_ = newArray;
        }
      } // else

                          // for testing and tracking
                          //System.out.println(
                            //      "BEFORE Arrays.sort(l_AclEntry_), sorting: ");
                          //for (int i = 0; i < l_AclEntry_.length; i++) {
                            //  System.out.println(
                              //    "--- i = "+i+"; ACL: "+l_AclEntry_[i]);
                          //} // for i

      // use AclEntry.compareTo()
      Arrays.sort(l_AclEntry_);
                          
                          // for testing and tracking
                          //System.out.println(
                            //      "AFTER Arrays.sort(l_AclEntry_), sorting: ");
                          //for (int i = 0; i < l_AclEntry_.length; i++) {
                            //  System.out.println(
                              //    "--- i = "+i+"; ACL: "+l_AclEntry_[i]);
                          //} // for i
      
      // records the _acl resource and ep in this ACS instance if they are unknown.
      // fires from zero to several events.
      if (_acl.getEorResACS() != this) {
          if (!m_Resources_.containsKey(aclResource.getName()) ) {
            addResource(aclResource);
          } // if
      } // if
      if (!_acl.getNameOfEpACS().equals(name_)) {
          // test if it is a new external eligible party
          if ( !m_EligibleParties_.containsKey( aclEP.getName() ) ) {
            addEligibleParty(aclEP);
          } // if
      } // if

      // fires an event to the AclEntry listeners, without sending an acl copy
      try {
          firePropertyChange("AclEntry", null, _acl);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.newEorAclEntry(), firePropertChange() throws:");
        e.printStackTrace();
      }

      aclResource = null;
      aclEP = null;
      resAcsName = null;
      l_openACS = null;
  } // newEorAclEntry(ImmutableAclEntry _acl)

  /**
   * Removes the AclEntry in the internal properties.
   * Called by AclEntryImpl.finalizeForUser() after the updating of
   * the resource list and before the updating of the acs controller list
   * in the resource. This method is indirectly called by closeExternalAclEntries().
   * It uses a direct equal operator ('==') to find the AclEntry to remove. <P>
   * The change events are fired after the updating of the AclEntry internal list.
   * Fires an event 'OneExternalEligibleParty' or 'OneExternalResource',
   * when the removed AclEntry has a Resource or an EligibleParty
   * that belongs to another ACS. Calls setExternalOriented().
   * Fires an event for the 'AclEntry' property using a clone of _acl.
   * May invoke removeEorResource(), removeEorEligibleParty() for external objects.
   * May call removeAcsController(), but only if the acl ACS is not closed.
   * if _acl is the last AclEntry to associate these external objects to this ACS.
   * The sequence of the optional property change firings is: event
   * 'OneExternalEligibleParty' or 'OneExternalResource',
   * 'ExternalOriented', 'Controlled' from the controlled ACS, then 'AclEntry'.
   * @param _acl aclEntry to remove.
   * @exception UpDateError if _acl is recorded more than once,
   *  if the resource name is unknown in the ACS,
   *  if _acl ACS is not a recorded controller of this ACS.
   */
  protected synchronized final void removeEorAclEntry(ImmutableAclEntry _acl) throws UpDateError {
        // variables
        int                 shiftingBack = 0;
        boolean             toKeep = false;
        boolean             controllerOfResAcs = false;
        ImmutableName       resName = _acl.getResName();
        ImmutableName       epName = _acl.getEpName();
        ImmutableName       resAcsName = _acl.getNameOfResACS();
        Name                nullName = NameImpl.EMPTY_INSTANCE;
        ImmutableResource   aclResource = _acl.getEorResource();
        ImmutableEligibleParty  aclEP = _acl.getEorEP();
        AclEntry[]          newArray = new AclEntryImpl[l_AclEntry_.length];
        AclEntry[]          result = new AclEntryImpl[0];
        
          // clones the value
        //ImmutableAclEntry[] l_oldAclEntry = new ImmutableAclEntry[l_AclEntry_.length];
        //System.arraycopy(l_AclEntry_, 0, l_oldAclEntry, 0, l_AclEntry_.length);

        // exception
        if (resAcsName == null || nullName.equals(resName) ) throw
          new UpDateError("in ACSImpl.removeEorAclEntry,"+
            " the _acl resource or its ACS is unknown");

                    // for testing and tracking
                    //System.out.println(name_+".removeEorAclEntry() for: "+ _acl);

        // removes _acl from l_AclEntry_:
                // this code works even if count > 1
        for (int  i = 0 ; i < l_AclEntry_.length ; i++) {
            if (l_AclEntry_[i] == null || l_AclEntry_[i].isEmpty() ) shiftingBack++;  // delete the empty l_AclEntry_[i]
            if (l_AclEntry_[i] == _acl ) {                        // == rather than equals
                // incrementing shiftingBack deletes l_AclEntry_[i]
              shiftingBack++;
            }
            else newArray[i - shiftingBack] = (AclEntry)l_AclEntry_[i];
        } // for i

        if (l_AclEntry_.length != shiftingBack) {
            result = new AclEntryImpl[l_AclEntry_.length - shiftingBack];
            System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
        }
                // else, l_AclEntry_ = (AclEntry[]) new AclEntryImpl[] {new AclEntryImpl()};
        //l_AclEntry_  = new AclEntryImpl[result.length];
        l_AclEntry_ = result;

        // exceptions
        if (shiftingBack > 1) {
            getEorBaseError().addException("In ACSImpl.removeEorAclEntry(_acl),"+
              " _acl is recorded more than once ");
            throw new UpDateError("In ACSImpl.removeEorAclEntry(_acl),"+
              " _acl is recorded more than once ");
        }
        if (shiftingBack == 0) {
            getEorBaseError().addException("In ACSImpl.removeEorAclEntry(_acl),"+
              " _acl is not recorded.");
            throw new UpDateError("In ACSImpl.removeEorAclEntry(_acl),"+
              " _acl is not recorded.");
        }
        
        // removes resource of _acl from m_Resources_, if this resource
        // is not owned to this ACS, and if it is not the resource
        // of an another AclEntry. May call removeAcsController() on the external
        // resource ACS.
        if (!containsOwnResourceName(resName) ) {
            toKeep = BaseUtilityImpl.isExternalACSobject(this, resName, false);
            if (!toKeep) {

                  // for testing
                  //System.out.println("Remove the external resource: "+aclResource+
                    //"in the ACS: "+name_);

              removeEorResource(aclResource);
            } // if
        }

        // removes the eligible party of _acl from m_EligibleParties_,
        // if this eligible party is not owned to this ACS, and if it is not
        // the eligible party of an another AclEntry
        toKeep = false;
        if (!containsOwnEligiblePartyName(epName) ) {
            toKeep = BaseUtilityImpl.isExternalACSobject(this, epName, true);
            if (!toKeep) {

                  // for testing
                  //System.out.println("in ACSImpl.removeEorAclEntry(), "+
                    //"for the resource: "+aclResource.getName()+
                    //"remove the external EP: "+epName+
                    //" in the ACS: "+name_);

                // used by the processing of firePropertyChange("ExternalOrientedAclEntry")
              removeEorEligibleParty(aclEP);
            } // if
        } // if (!containsOwnEligiblePartyName      
        
        try {

            // fire an event to the AclEntry listeners
            firePropertyChange("AclEntry", _acl, null);
            
         } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.removeEorAclEntry(), firePropertChange() throws:");
            e.printStackTrace();
        } // catch

        newArray = null;
        resName = null;
        epName = null;
        resAcsName = null;
        nullName = null;
        aclResource = null;
        aclEP = null;
  } // removeEorAclEntry()

  /**
   * The ACS is external oriented when it has at least one external AclEntry or
   * one external Privilege, and thus at least one external resource, virtual folder
   * or eligible party.
   * Note: The methods addResource, addEligibleParty and addVirtualFolder
   * directly update this property and fire the event.
   * Fires an event 'ExternalOriented' if the value has changed.
   * Called by removeEorResource(), removeEorVirtualFolder() and removeEorEligibleParty().
   */
  protected void setExternalOriented() {

      boolean old = externalOriented_;

      externalOriented_ = false;
      // a loop on l_AclEntry_ is faster than
      // getEorM_ExternalResources().isEmpty()
      for (int i = 0 ; i < l_AclEntry_.length ; i++) {
        if ( !l_AclEntry_[i].isEmpty()) {
            // test the resource, then the eligible party
            if ( l_AclEntry_[i].isExternalAclEntry() ) {
                  externalOriented_ = true;
                  break;
            } // if
        } // if
      } // for i

      if (!externalOriented_) {
          for (int i = 0 ; i < l_Privilege_.length ; i++) {
            if ( !l_Privilege_[i].isEmpty()) {
                // test the resource, then the eligible party
                if ( l_Privilege_[i].isExternalPrivilege() ) {
                      externalOriented_ = true;
                      break;
                } // if
            } // if
          } // for i
      } // if (!externalOriented_)

      try {
            // fire an event if necessary
          if (old != externalOriented_)
            firePropertyChange("ExternalOriented", old, externalOriented_);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.setExternalOriented(), firePropertChange() throws:");
        e.printStackTrace();
      } // catch

  } // setExternalOriented()
  
  /**
   * Registers a new privilege in this ACS.
   * Used by the Privilege constructors and by the methods
   * PrivilegeForLinksImpl.addEorPrivilegeSourceLink/TargetLink(),
   * which controls the unicity of the privilege detailled name, and indirectly,
   * also called by checkClosedPrivileges().
   * For a typed privilege, checks MAX_FOR_TYPED_PRIVILEGES_NUMBER if this ACS
   * is not an authorization server.
   * If the target ACS is different to this ACS and if it is unknown
   * by this ACS, the target is recorded as an external one. If the target ACS
   * unknowns this ACS as a controller, this ACS is recorded as a controller
   * in the target ACS, but only if the privilege ACS is not closed.
   * Calls setExternalOriented().
   * The setting of the privilege management in the ACS is controlled
   * in the PrivilegeAbst constructor. <P>
   * The sequence of the property change firings is: 'Privileges' then
   * 'OneTypedPrivilege' or 'OneLinkedPrivilege', and 'OneBaseObject'.
   * @exception CreateError if typed privilege, and if the number of typed privileges is
   *    at its maximum number, if another privilege as the same detailled name.
   * @param _priv is the new privilege to register. Not null.
   * @exception InternalError if this ACS does not manage the privileges.
   */
   protected synchronized void newEorPrivilege(Privilege _priv ) throws CreateError {
        Map<ImmutableName, ImmutableACS> m_openACS = BaseManagerImpl.getInstance().
                                                        getEorM_OpenACS();
        Privilege[]     l_old = l_Privilege_;

      if (_priv instanceof PrivilegeForType && !getIsAuthorizationServer()) {
          int count = 0;
          for(int i = 0; i < l_Privilege_.length; i++) {
              if (l_Privilege_[i] instanceof PrivilegeForType)
                  count++;
          } // for
          if (count >= MAX_FOR_TYPED_PRIVILEGES_NUMBER ) {
            m_openACS = null;
            l_old = null;
            throw new CreateError("in ACSImpl.newEorPrivilege(),"+
                " the typed privileges number is at its upper limit: "+
                MAX_FOR_TYPED_PRIVILEGES_NUMBER);
          }
      } // if
        
      // adds _priv in l_Privilege_
          // in a new array if l_Privilege_ == null
      if (l_Privilege_ == null || l_Privilege_.length == 0)
        l_Privilege_ = (Privilege[]) new Privilege[]{ (_priv) };
      else {
        boolean recorded = false;
          // in an empty place if possible
        for (int i = 0 ; i < l_Privilege_.length ; i++) {
            if (l_Privilege_[i] == null || l_Privilege_[i].isEmpty() ) {
                l_Privilege_[i] = _priv;
                recorded = true;
                break;
            }
        }
            // or at the end of l_Privilege_
        if (!recorded) {
            Privilege[] newArray = new Privilege[
              l_Privilege_.length + 1];
            System.arraycopy(l_Privilege_, 0, newArray, 0, l_Privilege_.length);
            newArray[l_Privilege_.length] = _priv;
            l_Privilege_ = newArray;
        }
      } // else

      Arrays.sort(l_Privilege_);

      // fires an event to the listeners, without sending a _priv copy
      try {
          firePropertyChange("Privileges", l_old, getEorL_Privileges());
          if (_priv instanceof PrivilegeForType)
              firePropertyChange("OneTypedPrivilege", null, _priv);
          else
              firePropertyChange("OneLinkedPrivilege", null, _priv);
          firePropertyChange("OneBaseObject", null, _priv);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.newEorPrivilege(), firePropertChange() throws:");
        e.printStackTrace();
      }

      m_openACS = null;
      l_old = null;
   } // newEorPrivilege()

  /**
   * Removes the Privilege in the internal properties.
   * Called by Privilege.finalizeForBase(). <P>
   * The change events are fired after the updating of the Privilege internal lists.
   * Fires the events 'Privileges',
   * 'OneTypedPrivilege' or 'OneLinkedPrivilege', and 'OneBaseObject'.
   * @param _priv to unregister.
   * @exception UpDateError if the privilege is null or not known.
   */
   protected synchronized void removeEorPrivilege(Privilege _priv ) throws UpDateError {
        // exception
        if (_priv == null)
          throw new UpDateError("in ACSImpl.removeEorPrivilege(),"+
            " the privilege is null.");
        
        // removes _priv in l_Privilege_
        int             shiftingBack = 0;
        Privilege[]     newArray = new Privilege[l_Privilege_.length];
        Privilege[]     result = new Privilege[0];
        //boolean         toKeep = false;
        //boolean         controllerOfResAcs = false;
        //ImmutableName   tgName = null;
        //ImmutableName   epName = null;
        Map<ImmutableName, ImmutableACS> m_openACS = BaseManagerImpl.getInstance().
                                                        getEorM_OpenACS();
        Privilege[]     l_old = l_Privilege_;

                // this code works even if shiftingBack > 1
        for (int  i = 0 ; i < l_Privilege_.length ; i++) {
             // delete the empty l_Privilege_[i]
            if (l_Privilege_[i] == null || l_Privilege_[i].isEmpty() ) shiftingBack++;
            if (l_Privilege_[i] == _priv ) {                        // == rather than equals
                // incrementing shiftingBack deletes l_Privilege_[i]
              shiftingBack++;
            }
            else newArray[i - shiftingBack] = (Privilege)l_Privilege_[i];
        } // for i

        if (l_Privilege_.length != shiftingBack) {
            result = new Privilege[l_Privilege_.length - shiftingBack];
            System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
        }
        l_Privilege_ = result;

        // exceptions
        if (shiftingBack > 1) {
            getEorBaseError().addException("In ACSImpl.removeEorPrivilege(_priv),"+
              " _priv is recorded more than once.");
            throw new UpDateError("In ACSImpl.removeEorPrivilege(_priv),"+
              " _priv is recorded more than once.");
        }

      // fires an event to the listeners, without sending a _priv copy
      try {
          // for the beamer when the displayed privilege is deleted
          firePropertyChange("Privileges", l_old, getEorL_Privileges());
          if (_priv instanceof PrivilegeForType)
              firePropertyChange("OneTypedPrivilege", _priv, null);
          else
              firePropertyChange("OneLinkedPrivilege", _priv, null);

          firePropertyChange("OneBaseObject", _priv, null);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.removeEorPrivilege(), firePropertChange() throws:");
        e.printStackTrace();
      }

      newArray = null;
      result = null;
      //tgName = null;
      //epName = null;
      result = null;
      m_openACS = null;
      l_old = null;
   } // removeEorPrivilege()
   
  /**
   * Currently unused, this method allows to control the EP and/or resource types.
   * It is used by PrivilegeForTypeImpl.setSourceClass() and setTargetClass().
   * Caution: does not manage 'Actor.ResourceType' in m_TypesAndGUIPolicy_.
   * @param _epType is the type of the object which have the access rights.
   *   Cannot be null, but '&lt;undefined&gt;' is a neutral always-allowed value.
   * @param _resourceType is the type of the object on which access rights are applied.
   *   Cannot be null, but '&lt;undefined&gt;' is a neutral always-allowed value.
   * @exception InternalError if an argument type is null,
   *    if this ACS does not manage Resources or EligibleParties
   * @return short '0' if the two types are allowed by the ACS,
   *    '-1' if only the EP type is not allowed,
   *    '-2' if only the resource type is not allowed,
   *    '-3' if the two types are not allowed.
   */
    protected final short controlEPandResourceTypes(
      String _epType, String _resourceType) {
        
        if (_epType == null || _resourceType == null)
            throw new InternalError("in ACSImpl.controlEPandResourceTypes(), "+
              "an type argument is null.");
       
        // init
        short result = 0;
        String[] l_resourceTypes = (String[])m_TypesAndGUIPolicy_.
              get("Resource.ResourceType");
        String[] l_epTypes = (String[])m_TypesAndGUIPolicy_.
              get("EligibleParty.EPType");
        boolean detected = false;
     
        if (l_resourceTypes != null ) {
            for (int i = 0; i < l_resourceTypes.length; i++) {
              if (l_resourceTypes[i].equals(_resourceType)) {
                  detected = true;
                  break;
              }
            } // for i
            if (!detected) result--;
        } // if (l_resourceTypes
        if (l_epTypes != null ) {
            detected = false;
            for (int i = 0; i < l_epTypes.length; i++) {
              if (l_epTypes[i].equals(_epType)) {
                  detected = true;
                  break;
              }
            } // for i
            if (!detected) { // result sets to -2 or -3
                result--;
                result--;
            }
        } // if (l_resourceTypes
        l_resourceTypes = null;
        l_epTypes = null;
          
        return result;
    } // controlEPandResourceTypes()

  /**
   * Registers a new virtual folder in this ACS.
   * Used only by the VirtualFolder constructors. No control on the type.
   * Fires the 'OneResourceVirtualFolder' or 'OneEPVirtualFolder',
   * 'ExternalOriented' and 'OneBaseObject' events. No 'M_ExternalVirtualFolders' event.
   * Called by newEorVirtualFolder() and
   * PrivilegeForLinksImpl.addEorPrivilegeTargetLink() for external Linked Privileges.
   * @param _fold is the new folder to register. Not null.
   * @exception CreateError if the VirtualFolder is null, if
   *   the name is already used, or if this ACS does not manage this object type.
   */
   protected synchronized void addVirtualFolder(VirtualFolder _fold ) throws CreateError {

      // adds _fold in l_VirtualFolders_
          // in a new array if l_VirtualFolders_ == null
      if (l_VirtualFolders_ == null || l_VirtualFolders_.length == 0)
        l_VirtualFolders_ = (VirtualFolder[]) new VirtualFolder[]{ _fold };
      else {
        boolean recorded = false;
          // in an empty place if possible
        for (int i = 0 ; i < l_VirtualFolders_.length ; i++) {
            if (l_VirtualFolders_[i] == null ) {
                l_VirtualFolders_[i] = _fold;
                recorded = true;
                break;
            }
        }
            // or at the end of l_VirtualFolders_
        if (!recorded) {
            VirtualFolder[] newArray = new VirtualFolder[l_VirtualFolders_.length + 1];
            System.arraycopy(l_VirtualFolders_, 0, newArray, 0, l_VirtualFolders_.length);
            newArray[l_VirtualFolders_.length] = _fold;
            l_VirtualFolders_ = newArray;
        }
      } // else


                            // for testing and tracking
                            //System.out.println("ACSImpl.addEorVirtualFolder(), "+
                              //      "l_VirtualFolders_ at the end :");
                            //System.out.println(" VF to add: "+_fold);
                            //System.out.println(" resulting list of VF: ");
                            //for (int i = 0; i < l_VirtualFolders_.length; i++) {
                              //  System.out.println(" -------- "+l_VirtualFolders_[i]);
                            //}



      // fires an event to the listeners, without _fold copy
      try {
          if (_fold.getEorACS() != this ) {
                // adds this ACS in the VF ACS as controller
                if (!Arrays.asList(_fold.getEorACS().getL_AcsControllers()).contains(name_)) {
                    ((ACSImpl)_fold.getEorACS()).addAcsController(name_); // throw UpDateError
                } // if

                if (!externalOriented_) {
                    externalOriented_ = true;
                    firePropertyChange("ExternalOriented", false, externalOriented_);
                }
          } // else

          if (VirtualFolderImpl.EP_VFOLDER.equals( _fold.getMainVFType()))
              firePropertyChange("OneEPVirtualFolder", null, _fold);
          else
              firePropertyChange("OneResourceVirtualFolder", null, _fold);
          
          firePropertyChange("OneBaseObject", null, _fold);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.newEorVirtualFolder(), firePropertChange() throws:");
        e.printStackTrace();
      }

   } // addVirtualFolder()


  /**
   * Registers a new virtual folder in this ACS.
   * Used only by the VirtualFolder constructors. No control on the type,
   * since the control is done in the VirtualFolder constructor.
   * Calls addVirtualFolder().
   * @param _fold is the new folder to register. Not null.
   * @exception CreateError if the VirtualFolder is null, if
   *   the name is already used, or if this ACS does not manage this object type.
   */
   protected synchronized void newEorVirtualFolder(VirtualFolder _fold ) throws CreateError {
      if (
        !getManageVirtualFolders()
      ) {
        throw new InternalError("This ACS does not manage the virtual folders.");
      }
      if (_fold == null)
        throw new CreateError("in ACSImpl.newEorVirtualFolder(),"+
            " the folder is null.");
      if (
        !getManageEPVirtualFolders() &&
            VirtualFolderImpl.EP_VFOLDER.equals( _fold.getType() )
      ) {
        throw new InternalError("This ACS does not manage the EP virtual folders.");
      }
      if (
        !getManageResourceVirtualFolders() &&
            VirtualFolderImpl.RES_VFOLDER.equals( _fold.getType() )
      ) {
        throw new InternalError("This ACS does not manage the resource virtual folders.");
      }
      
      // exception
      if (getOneVirtualFolder(_fold.getName()) != null) {
         throw new CreateError("in ACSImpl.newEorVirtualFolder(),"+
            " the folder name already exists.");
      }
      
      addVirtualFolder(_fold );

   } // newEorVirtualFolder()

  /**
   * Unregisters a virtual folder in this ACS.
   * Called by the VirtualFolder.finalizeForUser() method.
   * Called by PrivilegeForLinks.removeEorPrivilegeTargetLink()
   * for an external virtual folder. Not called by removeEorPrivilege().
   * For an external virtual folder, calls removeAcsController() on its ACS.
   * Fires the 'OneResourceVirtualFolder' or 'OneEPVirtualFolder' event,
   * and fires a 'OneBaseObject' event. No 'M_ExternalVirtualFolders' event.
   * @param _fold to unregister.
   * @exception UpDateError if the folder is null or not known.
   */
   protected synchronized void removeEorVirtualFolder(VirtualFolder _fold )
     throws UpDateError {
        // exception
        if (_fold == null)
          throw new UpDateError("in ACSImpl.removeEorVirtualFolder(),"+
            " the folder is null.");
        if (getOneVirtualFolder(_fold.getName()) == null)
         throw new UpDateError("in ACSImpl.removeEorVirtualFolder(),"+
            " the folder has a null name.");
        
        if (ownResourceFoldersRootName_.equals(_fold.getName()))
            setOwnResourceFoldersRootName(null);
        if (ownEPFoldersRootName_.equals(_fold.getName()))
            setOwnEPFoldersRootName(null);


                            // for testing and tracking
                            //System.out.println("ACSImpl.removeEorVirtualFolder(), "+
                              //      "l_VirtualFolders_ at the start :");
                            //for (int i = 0; i < l_VirtualFolders_.length; i++) {
                              //  System.out.println(" -------- "+l_VirtualFolders_[i]);
                            //}
                            //System.out.println(" VF to remove: "+_fold);
                
        // removes _fold in l_VirtualFolder_
        int                 shiftingBack = 0;
        VirtualFolder[]     newArray = new VirtualFolder[l_VirtualFolders_.length];
        VirtualFolder[]     result = new VirtualFolder[0];
                // this code works even if shiftingBack > 1
        for (int  i = 0 ; i < l_VirtualFolders_.length ; i++) {
             // deletes the empty l_VirtualFolders_[i]
            if (l_VirtualFolders_[i] == null)
                shiftingBack++;
            if (l_VirtualFolders_[i] == _fold ) {                        // == rather than equals
                // incrementing shiftingBack deletes l_VirtualFolders_[i]
                shiftingBack++;
            }
            else newArray[i - shiftingBack] = (VirtualFolder)l_VirtualFolders_[i];
        } // for i

        if (l_VirtualFolders_.length != shiftingBack) {


                // for testing and tracking
                //System.out.println("ACSImpl.removeEorVirtualFolder(), l_VirtualFolders_ lenght:"+
                  //  l_VirtualFolders_.length);
                //System.out.println("ACSImpl.removeEorVirtualFolder(), l_VirtualFolders_ shiftingBack:"+
                  //  shiftingBack);

            result = new VirtualFolder[l_VirtualFolders_.length - shiftingBack];
            System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
        }
        l_VirtualFolders_ = result;

        // exceptions
        if (shiftingBack > 1) {
            getEorBaseError().addException("In ACSImpl.removeEorVirtualFolder(_fold),"+
              " _fold is recorded more than once.");
            throw new UpDateError("In ACSImpl.removeEorVirtualFolder(_fold),"+
              " _fold is recorded more than once.");
        }

      // fires an event to the listeners, without a _fold copy
      try {
        if (_fold.getEorACS() != this) {
            boolean controllerOfVfAcs = false;
            VirtualFolder[] l_extVF = getEorL_ExternalVirtualFolders();
            
            if (l_extVF != null) {
                // removes this ACS in the EP ACS if necessary
                for (int i = 0; i < l_extVF.length; i++ ) {
                    // l_VirtualFolders_ is already updated
                    if ( l_extVF[i].getAcsName().equals(_fold.getAcsName() ) ) {
                      controllerOfVfAcs = true;
                      break;
                    } // if
                } // for i
            } // if (l_extVF != null
            l_extVF = null;

            if (!controllerOfVfAcs) {
                // unregistered this as a controller in the VF ACS
                ((ACSImpl)_fold.getEorACS() ).removeAcsController(name_); // throw UpDateError
            } // if (!controllerOfVfAcs)

            if (!controllerOfVfAcs)
                setExternalOriented();
        } // if (_fold.getEorACS() != this)

        if (VirtualFolderImpl.EP_VFOLDER.equals( _fold.getMainVFType()))
              firePropertyChange("OneEPVirtualFolder", _fold, null);
        else
              firePropertyChange("OneResourceVirtualFolder", _fold, null);
        
        firePropertyChange("OneBaseObject", _fold, null);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ACSImpl.removeEorVirtualFolder(), firePropertChange() throws:");
        e.printStackTrace();
      }

      newArray = null;
      result = null;
   } // removeEorVirtualFolder()
  
  /**
   * Gets all the closed ACS names if this ACS is a controller of external resource(s)
   * or eligible, through its aclEntries, or if this ACS is controlled by other ACS.
   * The current ACS instance manages only references to open ACS.
   * @return immutable set of closed or non closed ACS names,
   * for which there are aclEntries in getL_ClosedAclEntries(). May be null.
     */
   protected final Set<ImmutableName> getL_ClosedAclEntriesACSNames()	{
      if (l_ClosedAclEntriesACSNames_ == null) return null;
      else return Collections.unmodifiableSet(l_ClosedAclEntriesACSNames_);
   }

    /**
     * Called by equals(), finalizeForBase().
     * @return Set of the closed own aclEntries, without copy. May be null.
     */
    protected final Set<ImmutableAclEntry> getL_ClosedAclEntries() {
      return l_ClosedAclEntries_;
    }

    /**
     * Gets the own bridge sources which are not activated
     * because their targets belong to a closed ACS.
     * Called by BaseManagerImpl.
     * @return immutable Set of actor names. May be null.
     */
    protected final Set<ImmutableName> getL_ClosedACSBridgeSources() {
      if (m_closedTargetsOfActorBridgeSources_ == null ||
           m_closedTargetsOfActorBridgeSources_.isEmpty()) {

                        // for testing and tracking
                        //System.out.println("ACSImpl.getL_ClosedACSBridgeSources(),"+
                          //      " return null in the ACS "+name_);

          return null;
      }
      else
          return Collections.unmodifiableSet(m_closedTargetsOfActorBridgeSources_.keySet());
    }    
  
  /**
   * Gets all the closed ACS names if this ACS is a controller of external resource(s)
   * or eligible, through its Privileges, or if this ACS is controlled by other ACS.
   * The current ACS instance manages only references to open ACS.
   * @return immutable set of closed or non closed ACS names,
   * for which there are Privileges in getL_ClosedPrivileges(). May be null.
     */
   protected final Set<ImmutableName> getL_ClosedPrivilegesACSNames()	{
      if (l_ClosedPrivilegesACSNames_ == null) return null;
      else return Collections.unmodifiableSet(l_ClosedPrivilegesACSNames_);
   }

    /**
     * Called by BaseManagerImpl and another opening ACS for identifying
     * its resources that are controlled in Privileges by this ACS.
     * @return Set of the closed own Privileges, without copy. May be null.
     */
    protected final Set<PrivilegeForLinks> getL_ClosedPrivileges() {
      return l_ClosedPrivileges_;
    }

  /**
  * Closes the ACS. It is not removed from the ACS list in the BaseManager.
  * The ACS is not saved.
  * Removes the roots if necessary, and updates the roots map in the parent ACS.
  * The other associations are internal to the ACS. They are not explicitly
  * handled by a call to the finalizeForUser method on each ACSObject, since
  * all the ACSObjects are finalized.
  * Called only by BaseManagerImpl.closeACS().
  *  In this method, THE OBJECT IS SET TO NULL.
  *   - If there are listener(s) for the BaseManager bound property 'openACS',
  *   an event has been fired to them without any order by BaseManagerImpl before
  *   the call of this method.
  *   - the listeners, in the GUI, for the bound properties of the ACSImpl objects
  *   are not managed here, since the graphical objects (XXXTreeBaseListener)
  *   which are dedicated to this ACS and which are listeners, are finalized
  *   through the BaseManager bound property 'openACS'.
  */
  protected synchronized void finalizeForBase() {
    // variables
    AclEntry        nullAcl = new AclEntryImpl();
    EligibleParty   ep;
    Resource        res;
    AclEntryImpl[]  l_acl;
    SortedMap       m_ep;
    SortedMap       m_res;
    Set             l_closedAcl;
    Set             l_closedPri;
    Privilege[]     l_priv;
    VirtualFolder[] l_fold;

                        // for testing and tracking
                        //System.out.println("call to ACSImpl.finalizeForBase() on"+name_);
    
    //finalizes the main ACS objects to free memory
          // initialization
                // some variables must be independant to the internal _acs variables,
                // to allow iteration with finalizeXXX(), so use copies:
    l_acl = (AclEntryImpl[])getEorL_AclEntries();
    m_ep = getEorM_OwnEligibleParties();
    m_res = getEorM_OwnResources();
                // some variables must be directly returned, so:
    l_closedAcl = getL_ClosedAclEntries();
    l_closedPri = getL_ClosedPrivileges();
    l_priv = getEorL_Privileges();
    l_fold = getEorL_VirtualFolders();
    
          // deletes all open aclEntries of the ACS
    try {
      for (int i = 0 ; i < l_acl.length ; i++) {

            // for testing and tracking
            //if (l_acl[i] != null) System.out.println(
            //  "in ACSImpl.finalizeForBase(), open acl:"+l_acl[i]+
            //  ", and direct link to EP:"+l_acl[i].getEorEP());
            //else System.out.println("in ACSImpl.finalizeForBase(), there is a null acl");

        if (l_acl[i] == null || nullAcl.equals(l_acl[i]) ) {
            continue;
        }
        // call back removeEorAclEntry() to remove the ACL
        if (!l_acl[i].getL_ConditionSourceGroups().isEmpty()) {

           ((ResourceImpl)l_acl[i].getEorResource()).removeACL(
                   l_acl[i].getEorACS(),
                   l_acl[i].getEorEP(),
                   l_acl[i].getPositiveRight(),
                   l_acl[i].getL_ConditionSourceGroups().iterator().next());

        }
        else ( (ResourceImpl)l_acl[i].getEorResource() ).removeACL(
          this, l_acl[i].getEorEP(), l_acl[i].getPositiveRight(), null);
      } // for i
    } catch (UpDateError u) {
      throw new InternalError("in ACSImpl.finalizeForBase(), 1 - "+u.toString() );
    }
    
    try {
          // deletes the closed AclEntries and Privileges
      if (l_closedAcl != null) {
        for (Iterator ita = l_closedAcl.iterator();ita.hasNext();){
          ((AclEntryImpl)ita.next()).finalizeClosedAcl();
          ita.remove();
        }
      }
      if (l_closedPri != null) {
        for (Iterator ita = l_closedPri.iterator();ita.hasNext();){
          ((PrivilegeForLinksImpl)ita.next()).finalizeClosedPrivilege();
          ita.remove();
        }
      }

      // deletes all the groupIDMembers of _acs
      if (m_ep != null) {
        for (Iterator ita = m_ep.values().iterator();ita.hasNext();){
          ep = (EligibleParty) ita.next();
          if (ep == null || ep.equals(EligiblePartyImpl.EMPTY_INSTANCE ) ) continue;
          if (!(ep instanceof GroupIDMember) ) continue;
          if (ep.equals(GroupIDImpl.EMPTY_INSTANCE ) ||
            ep.equals(UserIDImpl.EMPTY_INSTANCE ) ) continue;
          
                        // for testing and tracking
                        //System.out.println("in ACSImpl.finalizeForBase(), ep has to be finalized: "+
                            //ep.getName());
          
          ep.finalizeForBase();
        }
      }
          // deletes all the actors of _acs
          // deletes the bridge sources and targets
      if (m_ep != null) {
        for (Iterator ita = m_ep.values().iterator();ita.hasNext();){
          ep = (EligibleParty) ita.next();
          if (ep == null || ep.equals(EligiblePartyImpl.EMPTY_INSTANCE ) ) continue;
          if (ep instanceof GroupIDMember ) continue;
          if (ep.equals(ActorImpl.EMPTY_INSTANCE) ) continue;
          
                        // for testing and tracking
                        //System.out.println("in ACSImpl.finalizeForBase(), "+
                            // "ep has to be finalized: "+ep.getName());
          
          ep.finalizeForBase();
        }
      }
      // deletes all resources of _acs, except Actor instances already deleted,
          // finalized instances (where res.getL_aclEntry() == null) and empty Directories
      if (m_res != null) {
        for (Iterator ita = m_res.values().iterator();ita.hasNext();){
          res = (Resource) ita.next();
          if (res instanceof Actor) continue;
          if (res == null || res.getL_aclEntry() == null ) continue;
          if (res.equals(ResourceImpl.EMPTY_INSTANCE) ||
            res.equals(DirectoryImpl.EMPTY_INSTANCE ) ) continue;

          res.finalizeForBase();
        } // for ita
      } // if

        // deletes all the privileges of _acs, from a l_Privilege_ copy
        if (l_priv != null) {
          for (int i = 0 ; i < l_priv.length ; i++) {

                // for testing and tracking
                //if (l_priv[i] != null) System.out.println(
                //  "in ACSImpl.finalizeForBase(), privilege:"+l_priv[i]);
                //else System.out.println("in ACSImpl.finalizeForBase(), "+
                //  "there is a null privilege");

            if (l_priv[i] == null || l_priv[i].isEmpty() ) continue;
            else l_priv[i].finalizeForBase();
          } // for i
        } // if (l_priv
      
        // deletes all the virtual folders of _acs, from a l_VirtualFolders_ copy
        if (l_fold != null) {
          for (int i = 0 ; i < l_fold.length ; i++) {

                // for testing and tracking
                //if (l_fold[i] != null) System.out.println(
                //  "in ACSImpl.finalizeForBase(), virtual folder:"+l_fold[i]);
                //else System.out.println("in ACSImpl.finalizeForBase(), "+
                //  "there is a null folder");

            if (l_fold[i] == null || l_fold[i].isEmpty() ) continue;
            else l_fold[i].finalizeForBase();
          } // for i
        } // if (l_fold
    
    } catch (UpDateError u) {
        u.printStackTrace();
        throw new InternalError("in ACSImpl.finalizeForBase(), "+u.toString() );
    }

    // --------- handles parent ACS relations ---------------------
    // roots with closing are removed from the parent, but not roots without closing
    List<ImmutableName> l_roots = null;
    if (l_RootNamesForAcsOpeningWithClosing_ != null)
        l_roots = new ArrayList(l_RootNamesForAcsOpeningWithClosing_);

    if (l_roots != null) {
        ImmutableName acsname = null;
        ACSImpl       acs = null;
        ACSObject     obj = null;

        try {
          // removes the with-closing roots in the parent ACSs
          for (ImmutableName name : l_roots) {
              obj = (ACSObject)BaseUtilityImpl.getEorBaseObject(name);
              acsname = obj.getAcsName();
              acs = (ACSImpl)BaseManagerImpl.m_OpenAcs__.get(acsname);

                        // for testing and tracking
                        //System.out.println(
                          //"in ACSImpl.finalizeForBase(), call to removeRootForChildACS() on:"+
                          //acs.getNickName()+" for "+obj.getName());

              acs.removeRootForChildACS(name_, obj);

              //if (obj instanceof Resource) {
                //((ResourceImpl)obj).removeConstrainedChildACS(
                  //      "<ForChildAcsOpeningAndClosing>", new ImmutableACS[] {this});
              //} // if
              //else if (obj instanceof EligibleParty) {
                //((EligiblePartyImpl)obj).removeConstrainedChildACS(
                  //      "<ForChildAcsOpeningAndClosing>", new ImmutableACS[] {this});
              //} // else
          } // for name
        } catch (CreateError c) {
          throw new InternalError("ACSImpl.finalizeForBase(): "+c.toString());
        } catch (BaseError b) {
          throw new InternalError("ACSImpl.finalizeForBase(): "+b.toString());
        //} catch (UpDateError u) {
          //throw new InternalError("ACSImpl.finalizeForBase(): "+u.toString());
        }

        acsname = null;
        acs = null;
        obj = null;
    } // if l_roots != null
    l_roots = null;

    try {
        if (m_RootCommandsForAcsOpeningAndClosing_ != null) {
            AssociateUtilityImpl.removeRootsInParentACSs(
                    m_RootCommandsForAcsOpeningAndClosing_, this);
        }
    } catch (BaseError b) {
        throw new InternalError("ACSImpl.finalizeForBase(): "+b.toString());
    }

    l_acl = null;
    m_ep = null;
    m_res = null;
    l_closedAcl = null;
    ep = null;
    res = null;
    nullAcl = null;
    l_priv = null;
    l_fold = null;

    iSname_ = null;    
    //name_ = (ImmutableName) NameImpl.EMPTY_INSTANCE;
    parentName_ = null;
    l_OpenAcsControllerNames_ = (ImmutableName[]) new NameImpl[0];
    baseError_ = new BaseError();
    ownResourcesRootName_ = null;
    ownGroupsRootName_ = NameImpl.EMPTY_INSTANCE;
    ownResourceFoldersRootName_ = NameImpl.EMPTY_INSTANCE;
    ownEPFoldersRootName_ = NameImpl.EMPTY_INSTANCE;
    
    l_AclEntry_ = (ImmutableAclEntry[]) new AclEntryImpl[0];
    m_Resources_.clear();
    m_EligibleParties_.clear();
    m_closedTargetsOfActorBridgeSources_  = null;
    l_ActorBridgeSources_ = null;
    l_ActorBridgeTargets_ = null;
    comment_ = null;
    l_ImmutableRules_ = null;
    changeSupport_ = null;
    m_Structure_ = null;
    l_ClosedAclEntriesACSNames_ = null;
    l_ClosedAclEntries_ = null;
    m_closedTargetsOfActorBridgeSources_  = null;
    
    // new in v0.7
    l_TiedChildACSs_ = null;
    m_SubAndGroupPolicy_ = null;
    m_BelongsToComposite_ = null;
    m_RootCommandsForAcsOpeningAndClosing_ = null;
    m_RootCommandsForAcsOpeningNoClosing_ = null;
    m_RootCommandsForAcsIntegrity_ = null;
    l_RootNamesForAcsOpeningWithClosing_ = null;
    l_RootNamesForAcsOpeningNoClosing_ = null;
    l_RootNamesForAcsIntegrity_ = null;
    m_RootsForChildACS_ = null;
    icon_ = null;
    iconName_ = null;
    l_Privilege_ = null;
    l_VirtualFolders_ = null;
    editor_ = null;

    l_ClosedPrivilegesACSNames_ = null;
    l_ClosedPrivileges_ = null;
    m_closedSourcesOfActorBridgeTargets_  = null;

  } // finalizeForBase()

  /**
   * This method removes the object from the application space,
   * but there is no guaranee that the JVM calls this method before
   * the end of the application.
   * @exception Throwable
   */
  protected void finalize() throws Throwable {
    finalizeForBase();
    super.finalize();
  }

  /**
  * Fires an event to every registered listener, in any order.
  * For the properties  * 'ExternalOrientedPrivilege',
  * 'Controlled', 'OneController', 'ExternalOriented', 'M_OwnResources', 'OneExternalResource',
  * 'OneResource' (when one resource is created or deleted),
  * 'OneEligibleParty', 'OneBaseObject' (Resource or EligibleParty) , 'AclEntry',
  * 'M_OwnEligibleParties', 'OneExternalEligibleParty',
  * 'OwnResourcesRootName', 'OwnGroupsRootName', 'OwnResourceFoldersRootName',
  * 'OwnEPFoldersRootName', 'BridgeSources', 'BridgeTargets', 'Comment',
  * 'OneAcsRight', 'OneNoRightProxyActor', 'TiedChildACSs', 'RootsForChildACS', 
  * 'Privileges', 'OneTypedPrivilege', 'OneLinkedPrivilege', 'OneResourceVirtualFolder',
  * 'OneEPVirtualFolder', 'IncompleteStructureModeling'.<P>
  * The listeners are proper to one ACSImpl instance, outside gBase,
  * transient (not serialized), and they are called in any order, excepted for:<P>
  * 1/ The order for a Resource change is : M_OwnResources (or External),
  *  OneResource, OneBaseObject.
  *  It is the same for the EligibleParty changes.<P>
  * 2/ For Actors, 'OneResource' is always fired when 'OneEligibleParty' is fired,
  *     and it is fired before this event for a resource creation,
  *     and after this event for a resource deletion.
  *  'OneBaseObject' is fired when the operation is completed, so,
  *     and it is fired always just after 'OneEligibleParty' if it is an actor creation,
  *     and always just after 'OneResource' if it is an actor deletion.
  * @param _propertyName name of the bound property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and if the two values are not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;


            // for testing and tracking
            //if (//_oldValue != null &&
              //  _propertyName.equals("OneBaseObject")) {
                //System.out.println( "ACSImpl.firePropertyChange() with an old value: "+
                  //      _oldValue+", for a 'OneBaseObject' event ");
                //System.out.println("PropertyChangeListeners for 'OneBaseObject' : ");
                //PropertyChangeListener[] l_prop = changeSupport_.getPropertyChangeListeners("OneBaseObject");
                //for (int i = 0; i < l_prop.length; i++) {
                  //  System.out.println(" -------------- "+l_prop[i].getClass()+
                    //        " : "+l_prop[i].toString());
                //}
            //} // if
      
      synchronized (changeSupport_) 
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);} 
  } // firePropertyChange()

  /**
   * Tests a listener and returns true if it is registered for the listening of
   * a given bound property.
   * Called by RightsMediatorImpl to avoid multiple registrations.
   * @param _listener to test
   * @param _propertyName name of the bound property to listen
   * @return true if _listener is registered
   */
  public boolean isPropertyChangeListener(
          String _propertyName, PropertyChangeListener _listener) {
      if (changeSupport_ == null) return false;
      boolean result = false;

      PropertyChangeListener[] l_listeners = changeSupport_.
                                    getPropertyChangeListeners(_propertyName);

      for (int i = 0; i < l_listeners.length; i++) {
          if (l_listeners[i] == _listener) {
              result = true;
              break;
          }
      } // for i

      l_listeners = null;
      return result;
  } // isPropertyChangeListener(

    /**
    * Closes all the inner AclEntries relevant to an another ACS,
    * to avoid any failed external references at the closing of one of theses ACS.
    * If an _acs resource or an _acs eligible party is controlled by an AclEntry
    * of this ACS, the AclEntry is deleted after a copyForExternal has been put
    * in l_ClosedAclEntries_.
    * l_ClosedAclEntriesACSNames_ is updated.
    * No action if _acs is this instance.
    * Called by BaseManagerImpl.cutACSExternalLinks().
    * @param _acs to perform. Not null.
    */
    protected final void closeExternalAclEntries(ACS _acs) {

        if (l_AclEntry_ == null || l_AclEntry_.length == 0) return;
        if (_acs == this) return;

        // attributes
        AclEntryImpl[]        l_aclCopy           = new AclEntryImpl[l_AclEntry_.length];
        Name                  acsName             = (Name)_acs.getName();
        Set                   l_aclEntryToClose   = new HashSet(INITIAL_CAPACITY_25);
        AclEntry              acl                 = new AclEntryImpl();
        //AclEntry              nullAcl             = new AclEntryImpl();

        // initialization
        if (l_ClosedAclEntries_ == null) l_ClosedAclEntries_ =
          new HashSet(INITIAL_CAPACITY_25);
        if (l_ClosedAclEntriesACSNames_ == null)
          l_ClosedAclEntriesACSNames_ = new HashSet(INITIAL_CAPACITY_5);
        System.arraycopy(l_AclEntry_, 0, l_aclCopy, 0, l_AclEntry_.length);

        // iterates on all aclEntries of this ACS through l_aclCopy
        // because resource.removeAclEntry() updates l_AclEntry_
        for (int i = 0 ; i < l_aclCopy.length ; i++) {
            if (l_aclCopy[i] == null) continue;
            if ( l_aclCopy[i].isEmpty() ) continue;
            if ( l_aclCopy[i].getEorResource() == null) throw new InternalError(
              "in ACSImpl.closeExternalAclEntries(), an AclEntry has a null resource");
            
            // selects the AclEntries with a resource from _acs,
            // and puts them in the saving lists
            if ( l_aclCopy[i].getNameOfResACS().equals(acsName ) ) {

                                        // for testing and tracking
                                     //System.out.println("ACSImpl.closeExternalAclEntries() on"+
                                       //     " for "+name_+", close the ACL at "+i+
                                         //   ": "+l_aclCopy[i].toString());

                l_ClosedAclEntries_.add((AclEntry)l_aclCopy[i].copyForExternal() );
                l_ClosedAclEntriesACSNames_.add((ImmutableName)acsName.clone() );
                l_aclEntryToClose.add(l_aclCopy[i]);
            }
            // selects the AclEntries with an ep from _acs,
            // and put them in the saving lists
            else if ( l_aclCopy[i].getNameOfEpACS().equals(acsName ) ) {

                                        // for testing and tracking
                                     //System.out.println("ACSImpl.closeExternalAclEntries() on"+
                                       //     " for "+name_+", close the ACL at "+i+
                                         //   ": "+l_aclCopy[i].toString());

                l_ClosedAclEntries_.add((AclEntry)l_aclCopy[i].copyForExternal() );
                l_ClosedAclEntriesACSNames_.add((ImmutableName)acsName.clone() );
                l_aclEntryToClose.add(l_aclCopy[i]);
            }
        } // for i

        // removes the AclEntries collected in l_aclEntryToClose
        // updates l_AclEntry_
        for (Iterator ite = l_aclEntryToClose.iterator() ; ite.hasNext() ; ) {
            try {
                acl = (AclEntry)ite.next();

                        // for testing and tracking
                        //System.out.println("ACSImpl.closeExternalAclEntries() on"+
                          //name_+" with removeAclEntry() for: "+ acl);

                // calls removeEorAclEntry()
                // works because the resource is still registered in its ACS
                if (!acl.getL_ConditionSourceGroups().isEmpty())
                    ( (Resource)acl.getEorResource() ).removeAclEntry(
                      this, acl.getEorEP(), acl.getPositiveRight(),
                      acl.getL_ConditionSourceGroups().iterator().next());
                else
                    ( (Resource)acl.getEorResource() ).removeAclEntry(
                      this, acl.getEorEP(), acl.getPositiveRight(),
                      null);
                // associated StringRights are not finalized, but the user can do it
                
            } catch (UpDateError u) {
                u.printStackTrace();
                throw new InternalError("From ACSImpl.closeExternalAclEntries() on the ACL "+acl);
            }
        } // for ite

                                        // for testing and tracking
                                        //if (l_ClosedAclEntries_ != null) {
                                          //       // for testing and tracking
                                            //    System.out.println("ACSImpl.closeExternalAclEntries() on"+
                                              //    name_+" ends with the l_ClosedAclEntries_ list:");

                                            //for (ImmutableAclEntry acl2 : l_ClosedAclEntries_) {
                                                 // for testing and tracking
                                              //  System.out.println("----"+acl2);
                                            //}
                                        //}
                                        //else {
                                                // for testing and tracking
                                          //      System.out.println("ACSImpl.closeExternalAclEntries() on"+
                                            //      name_+" ends with a null l_ClosedAclEntries_");
                                        //} // else

        l_aclCopy = null;
        acsName = null;
        l_aclEntryToClose = null;
        acl = null;
    } // closeExternalAclEntries()
  
    /**
    * Opens any closed external AclEntry that has both its resource,
    * its eligible party, the first condition group and the associated ACS(s)
    * which are currently open.
    * If not, the resource and/or the eligible party are not set as external objects.
    * Creates these AclEntries in their resources.
    * Selects the right first condition group from the resource ACS or the EP ACS.
    * If an external closed AclEntry has a Resource or an EligibleParty
    * which has been deleted in its open ACS, the closed AclEntry is deleted.
    * The metarights in the AclEntries are removed if they have been finalized,
    * and else, are switched to the reference metarights in BaseManagerImpl.
    * Called by BaseManagerImpl.restoreACSExternalLinks().
    * Updates l_ClosedAclEntries_ and l_ClosedAclEntriesACSNames_.
    * @param _acs an open ACS
    */
    protected final void checkClosedExternalAclEntries(ACS _acs) {

                                      // for testing and tracking
                                      //System.out.println("ACSImpl.checkClosedExternalAclEntries() "+
                                        //      "called on "+name_+"with the argument "+
                                          //    _acs.getNickName());

        if (l_ClosedAclEntries_ == null || l_ClosedAclEntries_.isEmpty()) {

                                      // for testing and tracking
                                      //System.out.println("ACSImpl.checkClosedExternalAclEntries() "+
                                        //      "called on "+name_+", l_ClosedAclEntries_ is empty. ");
            
            return;
        }
        if (_acs == this) return;

                                                 // for testing and tracking
                                                //System.out.println("ACSImpl.checkClosedExternalAclEntries() on"+
                                                  //name_+" starts with the l_ClosedAclEntries_ list:");
                                                //for (ImmutableAclEntry acl2 : l_ClosedAclEntries_) {
                                                     // for testing and tracking
                                                  //  System.out.println("----"+acl2);
                                                //}


        // initialization
        AclEntryImpl            acl;
        ImmutableACS            acs;
        ImmutableACS            acs_2;
        ImmutableEligibleParty  ep;
        ImmutableName           name;
        Resource                resource;
        GroupID                 condGrp = null;
        Map                     m_openACS = BaseManagerImpl.getInstance().
                                  getEorM_OpenACS();
                                // names of potentially closed ACS
        Set                     l_ACSToTest = new HashSet(INITIAL_CAPACITY_5); // set of names
        boolean                 known = false;
        StringRight[]           l_rights = null;
        StringRight             right = null;
        ArrayList               l_newRights = new ArrayList(INITIAL_CAPACITY_5);

        //if (l_ClosedAclEntriesACSNames_ == null)
        //  l_ClosedAclEntriesACSNames_ = new HashSet(INITIAL_CAPACITY_5);
            // for testing and tracking
            //System.out.println("start of "+name_+".checkClosedAclEntries()");

        // selects the right AclEntries to open, then opens them
        // and removes them in l_ClosedAclEntries_
        try {
          for (Iterator ite = l_ClosedAclEntries_.iterator() ; ite.hasNext() ; ) {
              acl = (AclEntryImpl)ite.next();

                                              // for testing and tracking
                                              //System.out.println("ACSImpl.checkClosedAclEntries(), process"+
                                                //      " the closed ACL: "+acl);

              if (acl == null || acl.isEmpty()) {
                ite.remove();

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                            //"remove the ACL "+acl);

                continue;
              } // if

              // for the ACL rights, removes a metaright if
              // it is unknown by the base manager, or restore it.
              // deletes the empty metarights.
              int count = 0;
              l_newRights.clear();
              l_rights = (StringRight[]) acl.getNoConditionalRight();
              for (int j = 0; j < l_rights.length; j++) {
                if (!(l_rights[j] instanceof StringMetaRightImpl)) {
                    l_newRights.add(l_rights[j]);
                    continue;
                }

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                            //"process the metaright "+l_rights[j]);

                if (l_rights[j].getRight() == null) {
                    count++;

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                            //"there is a finalized metaright");

                }
                else if ( BaseManagerImpl.getInstance().getMetaRight(
                                  l_rights[j].getRight()) == null ) {
                     // so l_rights[j] becames empty in all the standard and specialized rights of acs
                     ((StringMetaRightImpl)l_rights[j]).finalizeClosedRight();
                     count++;

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                            //"a metaright is not in the base manager ");

                } // if
                else {
                    right = BaseManagerImpl.getInstance().getMetaRight( l_rights[j].getRight());
                    // no ((StringMetaRightImpl)l_rights[j]).finalizeClosedRight();
                    // since it may be still used in the acs
                    l_newRights.add(right);
                    count++;

                          // for testing and tracking
                          //System.out.println("---- add right "+right);
                          //System.out.println("---- right equals to itself: "+
                            //right.equals(right));

                } // else
              } // for i

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                            //"the rights are set ");

              // no use of setRight() since the acs reference is null, and
              // no risk of error since there is just the removing of some metarights
              if (!l_newRights.isEmpty())
                  acl.l_right_ = (StringRight[])l_newRights.toArray(new StringRight[0]);
              else
                  acl.l_right_ = new StringRight[0];
              // clean the condition groups set if one does not exist,
              // and the returned value does not matter
              if (getManageRightDefinedConditionalAcls() &&
                        acl.l_right_.length != 0 && 
                        acl.l_right_[0].getRight().indexOf("<IF>") != -1) {
                  acl.setConditionSourceGroupsFromRights();
              }
              else {
                  // there is at most one condition group
                  List<ImmutableGroupID> l_grps = acl.getL_ConditionSourceGroups();
                  try {
                      for (ImmutableGroupID grp : l_grps) {
                          if (UtilityImpl.getFirstInstance().getBaseObject(grp.getName()) == null) {
                              acl.removeConditionSourceGroup(grp);
                          }
                      } // for grp
                  }
                  catch (CreateError c) {
                      c.printStackTrace();
                      System.out.println("The first condition group is unknown, then the AclEntry "+
                              acl+
                              "is finalized in ACSImpl.checkClosedExternalAclEntries().");
                      acl.finalizeClosedAcl();
                      ite.remove();
                      continue;
                  } // catch
              } // else

              // tests the 2 ACS in the acl, and do nothing if an ACS is closed
              // or if _acs is not implied.
              // acs and acs_2 may be this acs.
              acs = null;
              acs_2 = null;
              resource = null;
              ep = null;
              acs = (ImmutableACS)m_openACS.get( acl.getNameOfResACS() );
              if (acs == null) {

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                            //" the Resource ACS "+acl.getNameOfResACS()+" is unknown. ");

                  continue;
              }
              acs_2 = (ImmutableACS)m_openACS.get( acl.getNameOfEpACS() );
              if (acs_2 == null) {

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                            //" the EP ACS "+acl.getNameOfEpACS()+" is unknown. ");
                          
                  continue;
              }
              if (acs != _acs && acs_2 != _acs) continue;

              // tests the resource state from the acl clone, and deletes the AclEntry
              // only if the resource no more exists in its ACS.
              resource = (Resource)acs.getEorM_OwnResources().get( acl.resName_ );
              if (resource == null ) {
                  if (acs != this)
                      l_ACSToTest.add(acs.getName() ); // to update l_ClosedAclEntriesACSNames_

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedExternalAclEntries(), "+
                           //" the ACL resource "+acl.resName_+" is unknown in the ACS "+
                           //acs.getNickName());

                  ((AclEntryImpl)acl).finalizeClosedAcl();
                  ite.remove();
                  continue;
              } // if resource

              // tests the EligibleParty and deletes the AclEntry
              // only if the EligibleParty doesn't exist.
              ep = (ImmutableEligibleParty)acs_2.getEorM_OwnEligibleParties().
                get( acl.epName_ );
              if (ep == null ) {
                  if (acs_2 != this)
                      l_ACSToTest.add(acs_2.getName() );

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedAclEntries(), "+
                           //" the ACL EP "+acl.epName_+" is unknown in the ACS "+
                           //acs_2.getNickName());

                  ((AclEntryImpl)acl).finalizeClosedAcl();
                  ite.remove();
                  continue;
              } // if ep

              // no test here if it is an internal oriented AclEntry
              // since closeExternalAclEntries() does not allow it.

              // all is ok, so creates the AclEntry, removes it in l_ClosedAclEntries_,
              // and adds the two (possible) ACS in l_ACSToTest

                                  // for testing and tracking
                                  //System.out.println("in ACSImpl.checkClosedAclEntries(), "+
                                    //"in the acs "+name_+
                                    //", and for the ACS "+_acs.getName());
                                  //System.out.println("--- the ACS resource name is "+
                                    //acl.getNameOfResACS()+
                                    //", the ACS EP name is "+acl.getNameOfEpACS());
                                  //System.out.println("--- resource.addAclEntry() with resource: "+
                                    //resource.getName()
                                    //+" and EP: "+ep.getName());
                                  //System.out.println("--- ACL no-conditional rights: ");
                                  //l_rights = (StringRight[]) acl.getNoConditionalRight();
                                  //for (int j = 0; j < l_rights.length; j++) {
                                    //System.out.println("-------- "+l_rights[j]);
                                  //} // for j

              // there is at most one condition group, from 'acs' or 'acs_2'
              if (!acl.getL_ConditionSourceGroups().isEmpty()) {
                  if (acs.getName().equals(
                          ((GroupID)acl.getL_ConditionSourceGroups().iterator().next()).
                          getAcsName() )) {
                        condGrp = (GroupID) acs.getEorM_OwnEligibleParties().get(
                            ((GroupID)acl.getL_ConditionSourceGroups().iterator().next()).getName());
                  }
                  else {
                        condGrp = (GroupID) acs_2.getEorM_OwnEligibleParties().get(
                            ((GroupID)acl.getL_ConditionSourceGroups().iterator().next()).getName());
                  }

                  if (condGrp == null)
                      throw new InternalError("unknown condition group into the ACL "+acl );

                  resource.addAclEntry(this, ep, acl.getNoConditionalRight(),
                      acl.getPositiveRight(),
                      condGrp);
              }
              else {
                  resource.addAclEntry(this, ep, acl.getNoConditionalRight(),
                      acl.getPositiveRight(),
                      null);
              }

              if (acs != this)
                  l_ACSToTest.add(acs.getName() );
              if (acs_2 != this)
                  l_ACSToTest.add(acs_2.getName() );
              ite.remove();
          } // for ite = l_ClosedAclEntries_.iterator()

        } catch (CreateError c) {
            c.printStackTrace();
            throw new InternalError("the last error is through "+
                    "ACSImpl.checkClosedExternalAclEntries()" );
        } // catch

        // if an ACS name in l_ACSToTest is unknown in updated l_ClosedAclEntries_,
        // and is known in l_ClosedAclEntriesACSNames_, removes it
        // in l_ClosedAclEntriesACSNames_
        for (Iterator ite = l_ACSToTest.iterator() ; ite.hasNext() ; ) {
            name = (ImmutableName)ite.next();
            // it may be an open ACS
            if(!l_ClosedAclEntriesACSNames_.contains(name)) {
                   // for testing and tracking
                  //System.out.println("in ACSImpl.checkClosedAclEntries(), "+
                    //"test of HashSet.contains. "+name+" must be an open ACS.");

                continue;
            }
            // search if name is still known in l_ClosedAclEntries_
            known = false;
            for (Iterator ite2 = l_ClosedAclEntries_.iterator() ; ite2.hasNext() ; ) {
                acl = (AclEntryImpl)ite2.next();
                if (name.equals(acl.getNameOfEpACS() ) ||
                  name.equals(acl.getNameOfResACS() ) ) {
                    known = true;
                    break;
                }
            } // for ite2
            if (!known) l_ClosedAclEntriesACSNames_.remove(name);
        } // for ite

        acl = null;
        acs = null;
        acs_2 = null;
        ep = null;
        name = null;
        m_openACS = null;
        l_ACSToTest = null;
        l_rights = null;
        right = null;
        l_newRights = null;
    } // checkClosedAclEntries()

/**
 * Checks if there are own actor bridge sources or bridge targets for a given ACS.
 * They are linked through a bridge to an ACS. This method closes these links
 * and fills up the two persistant maps of these actors. The maps are used
 * by the method checkBridges() at the opening of the ACS, to restore the bridge links
 * with the actors from other open ACS.
 * Called by BaseManagerImpl.cutACSExternalLinks().
 * @param _acs process the actors from this ACS
 */
  protected final void closeBridges(ACS _acs) {
      if (!getContainsBridgeSource() && !getContainsBridgeTarget()) {

                                // for testing and tracking
                                //System.out.println("ACSImpl.closeBridges(), no source "+
                                  //      "nor target in "+name_);

          return;
      } // if
      if (_acs == this) return;
      
      ImmutableActor[]      l_sources2 = null;
      List<ImmutableActor>  l_copy = null;
      ArrayList<ImmutableName>   l_sources = null;

      if (l_ActorBridgeSources_ != null && !l_ActorBridgeSources_.isEmpty()) {
          // l_ActorBridgeSources_ is updated in the loop
          l_copy = (List<ImmutableActor>) ((ArrayList)l_ActorBridgeSources_).clone();

                    // for testing and tracking
                    //System.out.println("ACSImpl.closeBridges(), at the start, the sources are:");
                    //for (Iterator itetest = l_copy.iterator(); itetest.hasNext(); ) {
                      //  System.out.println(" -------- "+((ACSObject)itetest.next()).getName());
                    //} // for

          // process the sources
          for (ImmutableActor source : l_copy) {
                // actor.getEorACSBridgeTarget() is never null
                if (source.getEorACSBridgeTarget().getEorACS() != _acs)
                  continue;

                  // done in removeOwnActorBridgeSource()
                //if (m_closedTargetsOfActorBridgeSources_ == null)
                  //  m_closedTargetsOfActorBridgeSources_ = new TreeMap();
                //m_closedTargetsOfActorBridgeSources_.put(
                  //      source.getName(),
                    //    source.getEorACSBridgeTarget().getName() );

                                                // for testing and tracking
                                                //System.out.println("ACSImpl.closeBridges(), "+
                                                  //      "in m_closedTargetsOfActorBridgeSources_"+
                                                    //        " puts the key "+source.getName());

                try {
                    // calls removeOwnActorBridgeSource()
                    // and calls removeOwnActorBridgeTarget() on the target ACS
                    ((ActorImpl)source).setEorACSBridgeTarget(null, this);
                } catch (CreateError c) {
                    throw new InternalError("in ACSImpl.closeBridges() for "+
                            name_+", the bridge to the ACS "+_acs.getName()+
                            " is not deleted for the bridge source "+
                            source.getName()+", with the error "+c.toString());
                } // catch


          } // for actor
      } // if (l_ActorBridgeSources_ != null
      else {

                    // for testing and tracking
                    //System.out.println("ACSImpl.closeBridges(), the sources are empty"+
                      //      " in the ACS "+name_);
                    
      } // else

      if (l_ActorBridgeTargets_ == null || l_ActorBridgeTargets_.isEmpty()) {

                    // for testing and tracking
                    //System.out.println("ACSImpl.closeBridges(), the targets are empty"+
                      //      " in the ACS "+name_);

          l_copy = null;
          return;
      } // if (l_ActorBridgeTargets_ == null

      // l_ActorBridgeTargets_ is updated in the loop
      l_copy = (List<ImmutableActor>) ((ArrayList)l_ActorBridgeTargets_).clone();

      l_sources = new ArrayList(INITIAL_CAPACITY_25);
      // process the targets
      for (ImmutableActor target : l_copy) {

          l_sources2 = target.getEorL_bridgeSources();
          l_sources.clear();

          for (int i = 0; i < l_sources2.length; i++) {
                if (l_sources2[i].getEorACS() != _acs)
                    continue;
                l_sources.add(l_sources2[i].getName());

                try {
                    // calls removeOwnActorBridgeTarget()
                    ((ActorImpl)l_sources2[i]).setEorACSBridgeTarget(null, this);
                } catch (CreateError c) {
                    throw new InternalError("in ACSImpl.closeBridges() for "+
                            name_+", the bridge to the ACS "+_acs.getName()+
                            " is not deleted for the bridge source "+
                            l_sources2[i].getName()+", with the error "+c.toString());
                } // catch
          } // for i

      } // for target

      l_copy = null;
      l_sources = null;
      l_sources2 = null;
  } // closeBridges()
    
/**
 * Checks if there are bridge targets to create when the instance or
 * the argument is an opening ACS. If not, the closed target remains still
 * the right ACS is opening.
 * Note: the maps m_closedTargetsOfActorBridgeSources_ and m_closedSourcesOfActorBridgeTargets_
 * keep the data when an ACS or an external actor is deleted. The main difference
 * is that m_closedSourcesOfActorBridgeTargets_ is not ridden by this method, and
 * it simply updated when a closed source resets the bridge through
 * addOwnActorBridgeSource().
 * Called by BaseManagerImpl.restoreACSExternalLinks().
 * @param _acs an ACS
 */
  protected final void checkClosedBridges(ACS _acs) {
      
      if ((m_closedTargetsOfActorBridgeSources_ == null ||
        m_closedTargetsOfActorBridgeSources_.isEmpty()) &&
            (m_closedSourcesOfActorBridgeTargets_ == null ||
            m_closedSourcesOfActorBridgeTargets_.isEmpty())) {

                            // for testing and tracking
                            //System.out.println("ACSImpl.checkClosedBridges(), at the start, "+
                              //      "m_closedTargetsOfActorBridgeSources_ and"+
                                //    " m_closedSourcesOfActorBridgeTargets_ are empty.");

          return;
      } // if
      
      Map.Entry<ImmutableName, ImmutableName>   entry;
      ActorImpl                                 source = null;
      Actor                                     target;
      boolean                                   errorOne = false;
      boolean                                   errorTwo = false;
      SortedMap<ImmutableName, ImmutableName>
          m_closedTargetsOfActorBridgeSources_copy = new TreeMap();

                            // for testing and tracking
                            //System.out.println("ACSImpl.checkClosedBridges(), in "+name_+
                              //      " at the start, "+
                                //    "the sources in m_closedTargetsOfActorBridgeSources_ are:");
                            //for (ImmutableName name : m_closedTargetsOfActorBridgeSources_.keySet() ) {
                              //  System.out.println(" -------- "+name);
                            //} // for
      

      // provides a full copy to avoid a ConcurrentModificationException on entry
      for (Map.Entry<ImmutableName, ImmutableName> entry2 :
        m_closedTargetsOfActorBridgeSources_.entrySet()) {

          m_closedTargetsOfActorBridgeSources_copy.put(
                  (ImmutableName)entry2.getKey().clone(),
                  (ImmutableName)entry2.getValue().clone());

      } // for entry2

      for (Iterator ite = m_closedTargetsOfActorBridgeSources_copy.entrySet().iterator(); ite.hasNext();) {
          entry = (Map.Entry)ite.next();
          // if target is null, the bridge target ACS is not open or target has been deleted
          // and m_closedTargetsOfActorBridgeSources_ keeps the data
          target = (Actor) UtilityImpl.getFirstInstance().getBaseObject(
                                (ImmutableName)entry.getValue());

          // cannot delete a target its ACS has deleted when this ACS was closed
          if (target == null) {
              continue;
          }
          if (target.getEorACS() != _acs) continue;

          source = (ActorImpl) m_Resources_.get(entry.getKey() );
                // calls back ACSImpl.addOwnActorBridgeSource()
          if (source != null) {
              try {

                            // for testing and tracking
                            //System.out.println("ACSImpl.checkClosedBridges(), in "+name_+
                              //      "call to setEorACSBridgeTarget(target, this) on the source "+
                                //    source.getName());

                  // updates m_closedTargetsOfActorBridgeSources_
                  source.setEorACSBridgeTarget(target, this);
              } catch (CreateError c) {

                            // for testing and tracking
                            //System.out.println("ACSImpl.checkClosedBridges(), call to "+
                              //      "setEorACSBridgeTarget(target, this) on the source "+
                                //    source.getName()+" with error "+c.toString());

                  errorOne = true;
                  break;
              } // catch
          } // if (source != null)
          else {
              errorTwo = true;
              break;
          }
      } // for ite

      if (errorOne) throw new InternalError("in ACSImpl.checkClosedBridges() for "+
                            name_+", the bridge for the ACS "+_acs.getName()+
                            " is not created for the source "+source.getName());
      if (errorTwo) throw new InternalError("in ACSImpl.checkClosedBridges(),"+
            "a bridge target is not found for the source "+source.getName());

      if (m_closedSourcesOfActorBridgeTargets_ == null ||
            m_closedSourcesOfActorBridgeTargets_.isEmpty()) {
          source = null;
          target = null;
          entry = null;
          m_closedTargetsOfActorBridgeSources_copy = null;
          return;
      } // if

                            // for testing and tracking
                            //System.out.println("ACSImpl.checkClosedBridges(), in "+name_+
                              //      " at the start, "+
                                //    "the targets in m_closedSourcesOfActorBridgeTargets_ are:");
                            //for (ImmutableName name : m_closedSourcesOfActorBridgeTargets_.keySet() ) {
                              //  System.out.println(" -------- "+name);
                            //} // for

      // variables
      SortedMap<ImmutableName, ArrayList<ImmutableName>>
          m_closedSourcesOfActorBridgeTargets_copy = new TreeMap();
      ArrayList<ImmutableName> l_names = new ArrayList(INITIAL_CAPACITY_25);

      // provides a full copy to avoid a ConcurrentModificationException on entry2
      for (Map.Entry<ImmutableName, ArrayList<ImmutableName>> entry2 : 
        m_closedSourcesOfActorBridgeTargets_.entrySet()) {
          
          for (ImmutableName name : entry2.getValue()) {
              l_names.add(name);
          } // for name

          m_closedSourcesOfActorBridgeTargets_copy.put(
                  (ImmutableName)entry2.getKey().clone(),
                  (ArrayList<ImmutableName>)l_names.clone());

          l_names.clear();
      } // for entry2

        // reset the relevant bridges
      for (Map.Entry<ImmutableName, ArrayList<ImmutableName>> entry2 :
        m_closedSourcesOfActorBridgeTargets_copy.entrySet()) {

          for (Iterator ite = ((List)entry2.getValue()).iterator(); ite.hasNext(); ) {
              // if the source is null, the bridge target ACS is not open
              // and m_closedSourcesOfActorBridgeTargets_ keeps the data
              source = (ActorImpl) UtilityImpl.getFirstInstance().
                            getBaseObject((ImmutableName)ite.next());

              // cannot delete a source its ACS has deleted when this ACS was closed.
              // this may produce 'ghost' bridges.
              if (source == null) {
                  //ite.remove();
                  continue;
              }
              if (source.getEorACS() != _acs) continue;

              target = (Actor) m_Resources_.get(entry2.getKey() );
                    // calls back ACSImpl.addOwnActorBridgeSource()
              if (target != null) {
                  try {

                                // for testing and tracking
                                //System.out.println("ACSImpl.checkClosedBridges(), in "+name_+
                                  //      "call to setEorACSBridgeTarget(target, this) on the source "+
                                    //    source.getName());

                      // updates m_closedSourcesOfActorBridgeTargets_
                      source.setEorACSBridgeTarget(target, this);
                  } catch (CreateError c) {

                                // for testing and tracking
                                //System.out.println("ACSImpl.checkClosedBridges(), call to "+
                                  //      "setEorACSBridgeTarget(target, this) on the source "+
                                    //    source.getName()+" with error "+c.toString());

                      errorOne = true;
                      break;
                  } // catch
              } // if (source != null)
              else {
                  errorTwo = true;
                  break;
              }
          } // for ite
      } // for entry2
 
      if (errorOne) throw new InternalError("in ACSImpl.checkClosedBridges() for "+
                            name_+", the bridge for the ACS "+_acs.getName()+
                            " is not created for the source "+source.getName());
      if (errorTwo) throw new InternalError("in ACSImpl.checkClosedBridges(),"+
            "a bridge target is not found for the source "+source.getName());

      source = null;
      target = null;
      entry = null;
      l_names = null;
      m_closedTargetsOfActorBridgeSources_copy = null;
      m_closedSourcesOfActorBridgeTargets_copy = null;
  } // checkClosedBridges()
 
    /**
    * Closes all the inner linked privileges relevant to an another ACS,
    * to avoid any failed external references at the closing of one of these ACS.
    * If an _acs resource or an _acs virtual folder is controlled by a Privilege
    * of this ACS, the Privilege is deleted and its copyForExternal is put
    * in l_ClosedExternalLinkedPrivileges_.
    * The TYPED_FOR_SEED PrivilegeForTypes are external privileges but they are
    * never closed.
    * l_ClosedPrivilegesACSNames_ is updated.
    * No action if _acs is this instance.
    * Called by BaseManagerImpl.cutACSExternalLinks().
    * @param _acs to perform. Not null.
    */
    protected final void closeExternalLinkedPrivileges(ACS _acs) {

        if (l_Privilege_ == null || l_Privilege_.length == 0) return;
        if (_acs == this) return;

        // attributes
        Privilege[]     l_privCopy = new Privilege[l_Privilege_.length];
        Name            acsName = (Name)_acs.getName();
        Set<Privilege>  l_privToClose = new HashSet(INITIAL_CAPACITY_25);
        PrivilegeForLinksImpl       priv = new PrivilegeForLinksImpl();

        // initialization
        if (l_ClosedPrivileges_ == null) 
            l_ClosedPrivileges_ = new HashSet(INITIAL_CAPACITY_25);
        if (l_ClosedPrivilegesACSNames_ == null)
          l_ClosedPrivilegesACSNames_ = new HashSet(INITIAL_CAPACITY_5);
        System.arraycopy(l_Privilege_, 0, l_privCopy, 0, l_Privilege_.length);

        // iterates on all Privileges of this ACS through l_privCopy
        // because removeEorPrivilege() updates l_Privilege_
        for (int i = 0 ; i < l_privCopy.length ; i++) {
            if (l_privCopy[i] == null)
                continue;
            if ( l_privCopy[i].isEmpty() || !l_privCopy[i].isExternalPrivilege() ||
                    !l_privCopy[i].getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE))
                continue;

            priv = (PrivilegeForLinksImpl)l_privCopy[i];

                            // for testing
                         //System.out.println("ACS.closeExternalLinkedPrivileges() on"+
                           //     name_+" close the PRI at "+i+
                             //   ": "+priv.toString());
            
            // selects the Privileges with a target from _acs,
            // and puts them in the saving lists
            if (priv.getNameOfSingleTargetACS() != null &&
                    priv.getNameOfSingleTargetACS().equals(acsName ) ) {
                l_ClosedPrivileges_.add(priv.copyForExternal() );
                l_ClosedPrivilegesACSNames_.add((ImmutableName)acsName.clone() );
                l_privToClose.add(priv);
            }
            // selects the Privileges with an ep from _acs,
            // and put them in the saving lists
            else if (priv.getNameOfSingleSourceACS() != null &&
                    priv.getNameOfSingleSourceACS().equals(acsName ) ) {
                l_ClosedPrivileges_.add(priv.copyForExternal() );
                l_ClosedPrivilegesACSNames_.add((ImmutableName)acsName.clone() );
                l_privToClose.add(priv);
            }
        } // for i

        // removes the Privileges collected in l_privToClose
        // updates l_Privilege_
        try {
            for (Iterator ite = l_privToClose.iterator() ; ite.hasNext() ; ) {
            
                priv = (PrivilegeForLinksImpl)ite.next();

                        // for testing and tracking
                        //System.out.println(name_+"ACSImpl.closeExternalLinkedPrivileges()"+
                          //" finalizes "+ priv);

                priv.finalizeForUser();
                // associated StringRights are not finalized, but the user can do it
            } // for ite
        } catch (UpDateError u) {
            throw new InternalError("In ACSImpl.closeExternalLinkedPrivileges(): "+u.toString());
        }

        l_privCopy = null;
        acsName = null;
        l_privToClose = null;
        priv = null;
    } // closeExternalLinkedPrivileges()
  
    /**
    * Opens any closed external linked one-to-one Privileges that has both its target,
    * its eligible party and the associated ACS(s) which are currently open.
    * If not, the resource and/or the eligible party are not set as external objects.
    * Creates these Privileges in this ACS.
    * The TYPED_FOR_SEED PrivilegeForTypes are external privileges but they are
    * never closed and restored.
    * If an external closed Privilege has a Target or an EligibleParty
    * which has been deleted in its open ACS, the closed Privilege is deleted.
    * The metarights in the Privileges are removed if they have been finalized,
    * and else, they are switched to the reference metarights in BaseManagerImpl.
    * Called by BaseManagerImpl.restoreACSExternalLinks().
    * Updates l_ClosedPrivileges_, l_ClosedPrivilegesACSNames_, externalOriented_.
    * @param _acs an open ACS
    */
    protected final void checkClosedExternalLinkedPrivileges(ACS _acs) {

                          // for testing and tracking
                          //System.out.println("ACSImpl.checkClosedExternalLinkedPrivileges() called for "+
                            //      _acs.getNickName()+" on the ACS "+name_.lastComponent());

        if (l_ClosedPrivileges_ == null || l_ClosedPrivileges_.isEmpty())
          return;
        if (_acs == this) return;
        if (!l_ClosedPrivilegesACSNames_.contains(_acs.getName())) return;

        // initialization
        PrivilegeForLinksImpl   priv;
        ImmutableACS            acs;
        ImmutableACS            acs_2;
        ImmutableEligibleParty  ep;
        ImmutableName           name;
        PrivilegeTarget         target;
        Map                     m_openACS = BaseManagerImpl.getInstance().getEorM_OpenACS();
                                // names of potentially closed ACS
        Set                     l_ACSToTest = new HashSet(INITIAL_CAPACITY_5); // set of names
        boolean                 known = false;
        StringRight[]           l_rights = null;
        StringRight             right = null;
        ArrayList               l_newRights = new ArrayList(INITIAL_CAPACITY_5);

        //if (l_ClosedPrivilegesACSNames_ == null)
        //  l_ClosedPrivilegesACSNames_ = new HashSet(INITIAL_CAPACITY_5);
            // for testing and tracking
            //System.out.println("start of "+name_+".checkClosedPrivileges()");

        // selects the right Privileges to open, then opens them
        // and removes them in l_ClosedPrivileges_
        try {
          for (Iterator ite = l_ClosedPrivileges_.iterator() ; ite.hasNext() ; ) {
              priv = (PrivilegeForLinksImpl)ite.next();

                          // for testing and tracking
                          //System.out.println("ACSImpl.checkClosedPrivileges(), for "+priv);
                            //"remove a privilege ");

              if (priv == null || priv.isEmpty()) {
                ite.remove();

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                            //"remove a privilege ");

                continue;
              } // if

              // tests the 2 ACS in the privilege, and do nothing if an ACS is closed
              // or if _acs is not implied.
              // 'acs' and 'acs_2' may be this ACS.
              acs = null;
              acs_2 = null;
              target = null;
              ep = null;
              if (priv.getNameOfSingleTargetACS() != null) {
                  acs = (ImmutableACS)m_openACS.get( priv.getNameOfSingleTargetACS() );
                  if (acs == null) {

                              // for testing and tracking
                              //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                                //" the target ACS "+priv.getNameOfSingleTargetACS()+" is unknown. ");

                      continue;
                  }
              } // if (priv.getNameOfSingleTargetACS()
              if (priv.getNameOfSingleSourceACS() != null) {
                  acs_2 = (ImmutableACS)m_openACS.get( priv.getNameOfSingleSourceACS() );
                  if (acs_2 == null) {

                              // for testing and tracking
                              //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                                //" the EP ACS "+priv.getNameOfSingleSourceACS()+" is unknown. ");

                      continue;
                  }
              } // if (priv.getNameOfSingleSourceACS()

              // no target and no source implies a null 'acs' and a null 'acs_2'
              if (acs != _acs && acs_2 != _acs) continue;
              
              // for the PRI rights, removes a metaright if
              // it is unknown by the base manager, or restores it.
              // deletes the empty metarights.
              int count = 0;
              l_newRights.clear();
              l_rights = (StringRight[]) priv.getL_Rights();
              for (int j = 0; j < l_rights.length; j++) {
                if (!(l_rights[j] instanceof StringMetaRightImpl)) {
                    l_newRights.add(l_rights[j]);
                    continue;
                }

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                            //"process the metaright "+l_rights[j]);

                if (l_rights[j].getRight() == null) {
                    count++;

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                            //"there is a finalized metaright");

                }
                else if ( BaseManagerImpl.getInstance().getMetaRight(
                                  l_rights[j].getRight()) == null ) {
                     // l_rights[j] becames empty in all the standard and specialized rights of the acs
                     ((StringMetaRightImpl)l_rights[j]).finalizeClosedRight();
                     count++;

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                            //"a metaright is not in the base manager ");

                } // if
                else {
                    right = BaseManagerImpl.getInstance().getMetaRight( l_rights[j].getRight());
                    // no ((StringMetaRightImpl)l_rights[j]).finalizeClosedRight();
                    // since it may be still used in the acs
                    l_newRights.add(right);
                    count++;

                          // for testing and tracking
                          //System.out.println("---- add right "+right);
                          //System.out.println("---- right equals to itself: "+
                            //right.equals(right));

                } // else
              } // for i

                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                            //"the rights are set ");

              // no use of setRight() since the acs reference is null, and
              // no risk of error since there is just the removing of some metarights
              if (!l_newRights.isEmpty())
                  priv.l_rights_ = (StringRight[])l_newRights.toArray(new StringRight[0]);
              else
                  priv.l_rights_ = new StringRight[0];

              // tests the target state from the priv clone, and deletes the Privilege
              // only if the target no more exists in its ACS.
              if (priv.getSingleTargetName() != null) {
                  if (acs.containsOwnResourceName(priv.getSingleTargetName()) )
                      target = (PrivilegeTarget)acs.getEorM_OwnResources().
                            get( priv.getSingleTargetName() );
                  else
                      // the target may be a VirtualFolder
                      target = acs.getOneVirtualFolder( priv.getSingleTargetName() );

                  if (target == null) {
                      if (acs != this)
                          l_ACSToTest.add(acs.getName() ); // to update l_ClosedPrivilegesACSNames_

                              // for testing and tracking
                              //System.out.println("in ACSImpl.checkClosedExternalLinkedPrivileges(), "+
                               //" the PRI target "+priv.getSingleTargetName()+" is unknown in the ACS "+
                               //acs.getNickName());

                      priv.finalizeClosedPrivilege();
                      ite.remove();
                      continue;
                  } // if (target != null
              } // if (priv.getSingleTargetName() != null)

              // tests the EligibleParty and deletes the privilege
              // only if the EligibleParty doesn't exist.
              if (priv.getSingleSourceName() != null) {
                  ep = (ImmutableEligibleParty)acs_2.getEorM_OwnEligibleParties().
                            get( priv.getSingleSourceName() );

                  if (ep == null ) {
                      if (acs_2 != this)
                          l_ACSToTest.add(acs_2.getName() );

                              // for testing and tracking
                              //System.out.println("in ACSImpl.checkClosedExternalLinkedPrivileges(), "+
                               //" the PRI EP "+priv.getSingleSourceName()+" is unknown in the ACS "+
                               //acs_2.getNickName());

                      priv.finalizeClosedPrivilege();
                      ite.remove();
                      continue;
                  } // if ep
              } // if (priv.getSingleSourceName()

              // no test here if it is an internal privilege
              // since closeExternalPrivileges() does not allow it.

              // all is ok, so creates the Privilege with its source and target, if any,
              // removes it in l_ClosedExternalLinkedPrivileges_,
              // and adds the other ACS in l_ACSToTest

                              // for testing and tracking
                              //System.out.println("in ACSImpl.checkClosedExternalLinkedPrivileges(), "+
                                //"in the acs "+name_+
                                //", and for the opening ACS "+_acs.getName());
                              //System.out.println("---- the ACS target name is "+priv.getNameOfSingleTargetACS()+
                                //", the ACS EP name is "+priv.getNameOfSingleSourceACS());
                              //System.out.println("---- creates the privilege with the target: "+priv.getSingleTargetName()
                                //+" and EP: "+priv.getSingleSourceName());

              priv = new PrivilegeForLinksImpl(this, priv.getNickName(), priv.getMainType(),
                        priv.l_rights_, priv.getSecondType(), true);

              // one of these 2 'add' methods change the detailled name
              if (ep != null)
                  priv.addEorPrivilegeSourceLink((PrivilegeSource)ep);
              if (target != null)
                  priv.addEorPrivilegeTargetLink(target);

              if (acs != null && acs != this)
                  l_ACSToTest.add(acs.getName() );
              if (acs_2 != null && acs_2 != this)
                  l_ACSToTest.add(acs_2.getName() );
              ite.remove();
          } // for ite = l_ClosedPrivileges_.iterator()

        } catch (CreateError c) {
            c.printStackTrace();
            throw new InternalError("the last error is from ACSImpl."+
                    "checkClosedExternalLinkedPrivileges()" );
        } catch (BaseError b) {
            b.printStackTrace();
            throw new InternalError("the last error is from ACSImpl."+
                    "checkClosedExternalLinkedPrivileges()" );
        } // catch

        // if an ACS name in l_ACSToTest is unknown in updated l_ClosedPrivileges_,
        // and is known in l_ClosedPrivilegesACSNames_, removes it
        // in l_ClosedPrivilegesACSNames_
        for (Iterator ite = l_ACSToTest.iterator() ; ite.hasNext() ; ) {
            name = (ImmutableName)ite.next();

                               // for testing and tracking
                              //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                                //"test in l_ACSToTest: "+name.lastComponent());

            // it may be an open ACS
            if (!l_ClosedPrivilegesACSNames_.contains(name)) {

                   // for testing and tracking
                  //System.out.println("in ACSImpl.checkClosedPrivileges(), "+
                    //"test of HashSet.contains. "+name+" must be an open ACS.");

                continue;
            }
            // search if name is still known in l_ClosedPrivileges_
            known = false;
            for (Iterator ite2 = l_ClosedPrivileges_.iterator() ; ite2.hasNext() ; ) {
                priv = (PrivilegeForLinksImpl)ite2.next();
                if ((priv.getNameOfSingleSourceACS() != null &&
                        name.equals( priv.getNameOfSingleSourceACS() )) ||
                  (priv.getNameOfSingleTargetACS() != null &&
                  name.equals( priv.getNameOfSingleTargetACS()) ) ) {
                    known = true;
                    break;
                }
            } // for ite2
            if (!known) l_ClosedPrivilegesACSNames_.remove(name);
        } // for ite
                    
                              // for testing and tracking
                              //System.out.println(
                                //      "End of ACSImpl.checkClosedExternalLinkedPrivileges()");

        priv = null;
        acs = null;
        acs_2 = null;
        ep = null;
        name = null;
        m_openACS = null;
        l_ACSToTest = null;
        l_rights = null;
        right = null;
        l_newRights = null;
    } // checkClosedExternalLinkedPrivileges()

/**
 * For the ACS rights, unregisters them in their metarights.
 * The metarights are serialized without their map of acsrights,
 * but with all the links to the upper and lower metarights.
 * Called by BaseManagerImpl.cutACSExternalLinks().
 */
  protected final void closeAcsRights() {

                      // for testing and tracking
                      //System.out.println("ACSImpl.closeAcsRights(), rights for the acs: "+
                        //      name_);
      
      for (int i = 0; i < l_Rights_.length; i++) {
          
                      // for testing and tracking
                      //System.out.println("------------ at "+i+" metaright of right is "+
                        //      l_Rights_[i].getEorMetaRight());
          
            if (!(l_Rights_[i] instanceof StringMetaRight)) {
                ((StringMetaRightImpl)l_Rights_[i].getEorMetaRight()).removeACSRight(
                    (StringRightImpl)l_Rights_[i]); // no call back to the acsright
                // metaRight_ allows to restore meta_ at the next opening
                ((StringRightImpl)l_Rights_[i]).meta_ = null;
            }
            // else, it is a metaright which is serialized without its map of acsrights,
            // but its network of upper and lower metarights is serialized
        
      } // for i
  } // closeAcsRights()
    
/**
 * Checks if there are ACS rights to register in their respective metarights.
 * During the import of an ACS, a deserialized metaright which is unknown
 * is added in the BaseManager.
 * The deserialized metarights in the rights array are not switched to their equivallent
 * in the BaseManager, nor the deserialized metarights 
 * in the ACS constraints maps like 'AccRights'. See checkClosedExternalAclEntries()
 * where only the metarights in AclEntries are switched. This is also true
 * for the closed Privileges. Called by BaseManagerImpl.restoreACSExternalLinks().
 */
  protected final void checkClosedAcsRights() { 
      /*
      // first loop on metarights only
      for (int i = 0; i < l_Rights_.length; i++) {
          
          if (l_Rights_[i] instanceof StringMetaRight) { 
              // the deserialized-metaright is replaced by the current metaright
              // to cut the network of serialized upper and lower metarights, but
              // the deserialized metarights in the ACS constraints maps like 'AclRights'
              // remain, and have no updated m_ACSRights_ as in the metarights BaseManager, 
              //  and this is why StringMetaRightImpl.equals() does not control m_ACSRights_
              l_Rights_[i] = BaseManagerImpl.
                      getInstance().getMetaRight(l_Rights_[i].getRight());              

                      // for testing and tracking
                      //System.out.println("in ACSImpl.checkClosedAcsRights(), rights:");
                      // for testing and tracking
                      //System.out.println("------------ at "+i+" metaright is "+
                             // l_Rights_[i]);
          
          
          } // if
      } // for i
       * */
 
                          // for testing and tracking
                          //System.out.println("in ACSImpl.checkClosedAcsRights(), start for "+
                                  //name_);
      
      if (!getManageAcsRights()) return;      

      try {
          // loop on ACS rights for restoring the metaright links
          for (int i = 0; i < l_Rights_.length; i++) {

              if (l_Rights_[i] instanceof StringMetaRight) {
                  if (BaseManagerImpl.getInstance().getMetaRight(
                            ((StringRightImpl)l_Rights_[i]).metaRight_) == null) {

                      BaseManagerImpl.newMetaRight(
                              (StringMetaRight) l_Rights_[i], true);
                  } // if

                  continue;
              } // if

              // the deserialized-acsright metaright is replaced by the current metaright,
              // if it exists
              if (BaseManagerImpl.getInstance().getMetaRight(
                            ((StringRightImpl)l_Rights_[i]).metaRight_) != null) {

                  ((StringRightImpl)l_Rights_[i]).meta_ = BaseManagerImpl.
                          getInstance().getMetaRight(
                                ((StringRightImpl)l_Rights_[i]).metaRight_);
                  // registers in the metaright
                  ((StringMetaRightImpl)l_Rights_[i].getEorMetaRight()).addACSRight(
                        (StringRightImpl)l_Rights_[i]);

              } // if

                          // for testing and tracking
                          //System.out.println("---- at "+i+" the ACS right is "+
                                  //((StringRightImpl)l_Rights_[i])); //.meta_);

          } // for i
      }
      catch (CreateError c) {
            throw new InternalError(" in ACSImpl.checkClosedAcsRights(), a closed metaright "+
                    "is rejected by the BaseManager.");
      } // catch
  } // checkClosedAcsRights()

/**
 * Fills the ACS internal property list, after an ACS deserialization.
 * On the contrary of getEorM_Actors(), an Actor is tested as instance,
 * and not through Resource.getActorActivated().
 * Called by BaseManagerImpl.restoreACSInternalLinks().
 * Overridden by the AcsAddon Ubuntu.
 */
protected synchronized void fillNoRightProxyActors() {
    
    if (l_NoRightProxyActors_ == null)
        l_NoRightProxyActors_ = new ArrayList(INITIAL_CAPACITY_25);

    Resource res = null;
    Actor act = null;
                    
                                // for testing and tracking
                                //System.out.println("ACSImpl.fillNoRightProxyActors()"); 
    
    for (Iterator ite = m_Resources_.values().iterator() ; ite.hasNext() ; ) {
        res = (Resource) ite.next();
        if ( !(res instanceof Actor)) continue;
        act = (Actor)res;
        if ( act.getIsRightProxy() ) continue;
        
        l_NoRightProxyActors_.add(act);
                    
                                // for testing and tracking
                                //System.out.println("--------- ");        
        
    } // for ite 
    
    res = null;
    act = null;
} // fillNoRightProxyActors()

/**
 * Adds an own Actor which is not right proxy, then which is supposed to be
 * secure and without any virus able to deliver rights to the rights users
 * which execute these actors.
 * Fires the event 'OneNoRightProxyActor'.
 * Called by Actor.setIsRightProxy().
 * @param _act complies the criteria
 */
protected synchronized void addNoRightProxyActor(ImmutableActor _act) {
    
    if (l_NoRightProxyActors_ == null)
        l_NoRightProxyActors_ = new ArrayList(INITIAL_CAPACITY_25);

    List old = (List)((ArrayList)l_NoRightProxyActors_).clone();    
    
    if (!l_NoRightProxyActors_.contains(_act)) {
        l_NoRightProxyActors_.add(_act);
    
        try {
              firePropertyChange("OneNoRightProxyActor", 
                      old, ((ArrayList)l_NoRightProxyActors_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.addNoRightProxyActor(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
    }
    old = null;
} // addNoRightProxyActor()

/**
 * Removes an own Actor which is not right proxy, then which is supposed to be
 * secure and without any virus able to deliver rights to the rights users
 * which execute these actors.
 * Fires the event 'OneNoRightProxyActor'.
 * Called by Actor.setIsRightProxy().
 * @param _act does not comply the criteria
 */
protected synchronized void removeNoRightProxyActor(ImmutableActor _act) {
    
    if (l_NoRightProxyActors_ == null)
        l_NoRightProxyActors_ = new ArrayList(INITIAL_CAPACITY_25);

    List old = (List)((ArrayList)l_NoRightProxyActors_).clone();    
    
    if (l_NoRightProxyActors_.contains(_act)) {
        l_NoRightProxyActors_.remove(_act);
    
        try {
              firePropertyChange("OneNoRightProxyActor", 
                      old, ((ArrayList)l_NoRightProxyActors_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.addNoRightProxyActor(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
    }
    old = null;
} // removeNoRightProxyActor()

/**
 * Sets the VirtualFolder internal properties, before any ACS deserialization.
 * Called by BaseManagerImpl.cutACSInternalLinks().
 */
protected synchronized void closeVirtualFolders() {

    if (l_VirtualFolders_ == null || l_VirtualFolders_.length == 0) {

                                // for testing and tracking
                                //System.out.println("ACSImpl.closeVirtualFolders() in "+name_+
                                  //      ", null VF list.");

        return;
    }

                                // for testing and tracking
                                //System.out.println("ACSImpl.closeVirtualFolders() in "+name_+
                                  //      " for the VF:");
                                
    for (int i = 0; i < l_VirtualFolders_.length; i++) {
        if (l_VirtualFolders_[i] != null)
            ((VirtualFolderImpl)l_VirtualFolders_[i]).prepareBeforeSaving();
                    
                                // for testing and tracking
                                //System.out.println("--------- "+l_VirtualFolders_[i]);
        
    } // for i
} // closeVirtualFolders()

/**
 * Sets the VirtualFolder internal properties, after any ACS deserialization.
 * It is not necessary to call this method when closeVirtualFolders() has been called
 * to save the ACS, since the workable properties of the VirtualFolders remains
 * usable after closeVirtualFolders(). If the virtual folder is already restored,
 * there is no action. Thus, BaseManagerImpl may call this method at the request
 * of a method AcsFactoryAcsAddon.configure().
 * Called by BaseManagerImpl.restoreACSInternalLinks().
 */
protected synchronized void fillVirtualFolders() {

    if (l_VirtualFolders_ == null || l_VirtualFolders_.length == 0)
        return;

    try {
        for (int i = 0; i < l_VirtualFolders_.length; i++) {
            if (l_VirtualFolders_[i] != null)
                ((VirtualFolderImpl)l_VirtualFolders_[i]).prepareAfterRestoring();

                                    // for testing and tracking
                                    //System.out.println("--------- ");

        } // for i
    } catch (DialogError d) {
        throw new InternalError("from ACSImpl.fillVirtualFolders() for "+
                name_+", "+d.toString());
    }
} // fillVirtualFolders() 


/**
 * Sets the ACS rights and the metarights this ACS handles and for a given internal key value
 * of the associated metaright. The map key is the metaright internal key.
 * If the map is already set, there is no action.
 * Thus, BaseManagerImpl may call this method at the request
 * of a method AcsFactoryAcsAddon.configure(), after the ACS creation.
 * Called by BaseManagerImpl.restoreACSInternalLinks() and valueOfForGenericBase().
 */
protected synchronized void fillRightsOfInternalKey() {

    if (m_RightsOfInternalKey_ != null && !m_RightsOfInternalKey_.isEmpty())
        return;

    m_RightsOfInternalKey_ = new TreeMap<String, StringRight[]>();
    String          key = null;
    StringRight[]   l_rights = null;
    StringRight[]   l_result = null;
   
    for (int i = 0; i < l_Rights_.length; i++) {
        if (l_Rights_[i] == null)
            continue;

        key = (new Integer(l_Rights_[i].getEorMetaRight().getInternalKey())).toString();

        if (m_RightsOfInternalKey_.get(key) == null ) {
            m_RightsOfInternalKey_.put(key, new StringRight[] {l_Rights_[i]});
        }
        else {
            l_rights = m_RightsOfInternalKey_.get(key);
            l_result = new StringRight[l_rights.length + 1];
            System.arraycopy(l_rights, 0, l_result, 0, l_rights.length);
            l_result[l_result.length - 1] = l_Rights_[i];
            m_RightsOfInternalKey_.put(key, l_result);
        } // else

                                // for testing and tracking
                                //System.out.println("ACSImpl.fillRightsOfInternalKey() adds "+
                                  //       "one right on the key "+key);

    } // for i

    key = null;
    l_rights = null;
} // fillRightsOfInternalKey()

 /**
 * Noop. This hook is there to be overriden. <P>
 * May be called by AssociateUtilityImpl.add/removeAssociateActions() to update
 * the ACS after the opening or the closing of one of its child ACS, following an
 * association script. For instance, it should be used to make some complex works
 * the scripts cannot do.<P>
 * On the other hand, the scripts have the property
 * to be ridden in the GUI by the advanced users of Access Road,
 * on the contrary of this method.
 * @param _root is the ACSObject of this parent ACS which is a root of a child ACS
 * @param _opening true if this method is called for the opening of the child ACS,
 * and false for its closing
 */
 protected void handleAssociationsForRootsOfChildACS(ACSObject _root, boolean _opening) {
    
 } // handleAssociationsForRootsOfChildACS()


 /**
 * Noop. This hook is there to be overriden. <P>
 * Called by AssociateUtilityImpl.add/removeAssociateActions() to update 
 * the ACS after the creating or the deleting of one of its ACSObject, following an
 * association script. For instance, it should be used to make some complex works
 * the scripts cannot do.<P>
 * On the other hand, the scripts have the property
 * to be ridden in the GUI by the advanced users of Access Road,
 * on the contrary of this method.
 * @param _obj is the ACSObject of this ACS which is creating or deleting
 * @param _creating false if _obj is finalizing
 */
 protected void handleAssociationsForInnerObject(ACSObject _obj, boolean _creating) {
    
 } // handleAssociationsForInnerObject()

  //
  // ----------------- PACKAGE METHODS -------------------------------------
  //

  /**
   * Set name of the root for the ACS own resources.
   * The current root is the first non-finalized directory
   * without parent which has been created.
   * So, if there is already a resource root, the only allowed
   * operation is to nullify the root name.
   * Called by the Directory constructor and finalizer.
   * Fire changes on the 'OwnResourcesRootName' property.
   * @param _name ImmutableName of the root for the recorded own resources
   * in the ACS base. May be null or empty.
   * The directory has to be already recorded in the ACS base.
   * @exception UpDateError if the Node has a parent,
   *  or if it doesn't owned to this ACS.
   */
  protected void setOwnResourcesRootName(ImmutableName _name) throws UpDateError {
    if (_name == null ) {
      if (ownResourcesRootName_!= null ) {
        Name clone = (Name) ownResourcesRootName_.clone();
        ownResourcesRootName_ = null;
        try {
            firePropertyChange("OwnResourcesRootName", clone, NameImpl.EMPTY_INSTANCE);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnResourcesRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
        clone = null;
      } else return; // no update
    } else {
          Name clone;
          if (ownResourcesRootName_ != null) clone = (Name) ownResourcesRootName_.clone();
          else clone = NameImpl.EMPTY_INSTANCE;
          ownResourcesRootName_ = (ImmutableName)_name.clone();
          try {
             firePropertyChange("OwnResourcesRootName", clone, ownResourcesRootName_.clone());
          } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnResourcesRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
          }
          clone = null;
    } // else
  } // setOwnResourcesRootName()

  /**
   * Set the name of the root for the ACS own groups.
   * The current root is the first non-finalized group
   * without parent which has been created.
   * So, if there is already a group root, the only allowed
   * operation is to nullify the root name.
   * Called by the GroupImpl constructor and finalizer.
   * Fire changes on the 'OwnGroupsRootName' property.
   * @param _name of the root for the recorded own groups
   * in the ACS base. May be null or empty.
   * The group has to be already recorded in the ACS base.
   * @exception UpDateError if the group has a parent,
   *  or if it doesn't owned to this ACS.
   */
  protected void setOwnGroupsRootName(ImmutableName _name) throws UpDateError {
    if (_name == null || _name.equals(NameImpl.EMPTY_INSTANCE) ) {
      if (!ownGroupsRootName_.equals(NameImpl.EMPTY_INSTANCE)) {
        Name clone = (Name) ownGroupsRootName_.clone();
        ownGroupsRootName_ = NameImpl.EMPTY_INSTANCE;
        try {
            firePropertyChange("OwnGroupsRootName", clone, NameImpl.EMPTY_INSTANCE);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updatinge, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnGroupsRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
        clone = null;
      } else return; // no update
    } else if (!ownGroupsRootName_.equals(NameImpl.EMPTY_INSTANCE)) {
        return; // no direct update from a name to another name
    } else if (containsOwnEligiblePartyName( _name) &&
      m_EligibleParties_.get(_name) instanceof GroupID &&
      ((GroupID)m_EligibleParties_.get(_name)).getEorL_GroupID().length == 0) {
          Name clone = (Name) ownGroupsRootName_.clone();
          ownGroupsRootName_ = (ImmutableName)_name.clone();
          try {
             firePropertyChange("OwnGroupsRootName", clone, ownGroupsRootName_.clone());
          } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnGroupsRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
          }
          clone = null;
    } else {
        baseError_.addException("In ACSImpl.setOwnGroupsRoot("+
          "ImmutableName _name), the root is not a GroupID without parent,"+
          "or it doesn't owned to this ACS");
        throw new UpDateError("In ACSImpl.setOwnGroupsRoot("+
          "ImmutableName _name), the root is not a GroupID without parent,"+
          "or it doesn't owned to this ACS");
    }
  } // setOwnGroupsRootName()
  
  /**
   * Sets name of the root for the ACS own resource virtual folders.
   * The current root is the first non-finalized folder
   * without parent which has been created.
   * So, if there is already a resource virtual folder, the only allowed
   * operation is to nullify the root name.
   * Called by the VirtualFolderImpl constructor and finalizer,
   * and by removeEorVirtualFolder().
   * Fire change on 'OwnResourceFoldersRootName' property.
   * @param _name of the root for the recorded own resource folders
   * in the ACS base. May be null or empty.
   * The folder has to be already recorded in the ACS base.
   * @exception UpDateError if the folder has a parent,
   *  or if it doesn't owned to this ACS.
   */
  protected void setOwnResourceFoldersRootName(ImmutableName _name) throws UpDateError {
    if (_name == null || _name.equals(NameImpl.EMPTY_INSTANCE) ) {
      if (!ownResourceFoldersRootName_.equals(NameImpl.EMPTY_INSTANCE)) {
        Name clone = (Name) ownResourceFoldersRootName_.clone();
        ownResourceFoldersRootName_ = NameImpl.EMPTY_INSTANCE;
        try {
            firePropertyChange("OwnResourceFoldersRootName", clone, NameImpl.EMPTY_INSTANCE);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnResourceFoldersRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
        clone = null;
      } else return; // no update
    } else if (!ownResourceFoldersRootName_.equals(NameImpl.EMPTY_INSTANCE)) {
        return; // no direct update from a name to another name
        
    } else if (containsVirtualFolderName( _name) &&
      (getOneVirtualFolder(_name).getEorParent() == null)) {
          Name old = (Name) ownResourceFoldersRootName_.clone();
          ownResourceFoldersRootName_ = (ImmutableName)_name.clone();
          try {
             firePropertyChange("OwnResourceFoldersRootName", old, ownResourceFoldersRootName_.clone());
          } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnResourceFoldersRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
          }
          old = null;
    } else {
        baseError_.addException("In ACSImpl.setOwnResourceFoldersRootName("+
          "ImmutableName _name), the root is not a folder without parent,"+
          "or it doesn't owned to this ACS");
        throw new UpDateError("In ACSImpl.setOwnResourceFoldersRootName("+
          "ImmutableName _name), the root is not a folder without parent,"+
          "or it doesn't owned to this ACS");
    }
  } // setOwnResourceFoldersRootName()

  /**
   * Sets name of the root for the ACS own eligible party virtual folders.
   * The current root is the first non-finalized folder
   * without parent which has been created.
   * So, if there is already a EP virtual folder, the only allowed
   * operation is to nullify the root name.
   * Called by the VirtualFolderImpl constructor and finalizer,
   * and by removeEorVirtualFolder().
   * Fire change on 'OwnEPFoldersRootName' property.
   * @param _name of the root for the recorded own EP folders
   * in the ACS base. May be null or empty.
   * The folder has to be already recorded in the ACS base.
   * @exception UpDateError if the folder has a parent,
   *  or if it doesn't owned to this ACS.
   */
  protected void setOwnEPFoldersRootName(ImmutableName _name) throws UpDateError {
    if (_name == null || _name.equals(NameImpl.EMPTY_INSTANCE) ) {
      if (!ownEPFoldersRootName_.equals(NameImpl.EMPTY_INSTANCE)) {
        Name clone = (Name) ownEPFoldersRootName_.clone();
        ownEPFoldersRootName_ = NameImpl.EMPTY_INSTANCE;
        try {
            firePropertyChange("OwnEPFoldersRootName", clone, NameImpl.EMPTY_INSTANCE);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnEPFoldersRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
        clone = null;
      } else return; // no update
    } else if (!ownEPFoldersRootName_.equals(NameImpl.EMPTY_INSTANCE)) {
        return; // no direct update from a name to another name
        
    } else if (containsVirtualFolderName( _name) &&
      (getOneVirtualFolder(_name).getEorParent() == null)) {
          Name old = (Name) ownEPFoldersRootName_.clone();
          ownEPFoldersRootName_ = (ImmutableName)_name.clone();
          try {
             firePropertyChange("OwnEPFoldersRootName", old, ownEPFoldersRootName_.clone());
          } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.setOwnEPFoldersRootName(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
          }
          old = null;
    } else {
        baseError_.addException("In ACSImpl.setOwnEPFoldersRootName("+
          "ImmutableName _name), the root is not a folder without parent,"+
          "or it doesn't owned to this ACS");
        throw new UpDateError("In ACSImpl.setOwnEPFoldersRootName("+
          "ImmutableName _name), the root is not a folder without parent,"+
          "or it doesn't owned to this ACS");
    }
  } // setOwnEPFoldersRootName()
  
/**
 * Sets an own actor as a bridge target for an external actor.
 * Fires the property event 'BridgeTargets'.
 * Called by _source.
 * @param _source is a non-null external actor which is the source
 * @param _target is a non-null own actor which is the target
 * @exception CreateError if a parameter is null or does not belong to
 *  the right ACS.
 */
  protected final void addOwnActorBridgeTarget(ImmutableActor _source, ImmutableActor _target)
      throws CreateError {

        ArrayList<ImmutableName>  l_sources = null;

        if (!getManageBridges())
          throw new CreateError("The bridge target ACS "+
               "does not manage bridges to other ACS.");
        if (_source == null || _source.getEorACS() == this) throw new CreateError(
          "in ACSImpl.addOwnActorBridgeTarget(), _source is null or "+
          "belongs to the ACS.");
        if (_target == null || _target.getEorACS() != this) throw new CreateError(
          "in ACSImpl.addOwnActorBridgeTarget(), _target is null or "+
          "does not belong to the ACS.");
        
        if (m_ActorBridgeTargets_ == null)
            m_ActorBridgeTargets_ = new TreeMap();
        m_ActorBridgeTargets_.put(_source.getName(), _target );

        if (m_closedSourcesOfActorBridgeTargets_ != null &&
                !m_closedSourcesOfActorBridgeTargets_.isEmpty()) {

              l_sources = m_closedSourcesOfActorBridgeTargets_.get(_target.getName());

              if (l_sources != null && !l_sources.isEmpty()) {
                  // may be not in l_sources
                  l_sources.remove(_source.getName());
                  if (l_sources.isEmpty())
                      m_closedSourcesOfActorBridgeTargets_.remove(_target.getName());
              } 
        } // if (m_closedSourcesOfActorBridgeTargets_ != null
        
        if (l_ActorBridgeTargets_ == null)
            l_ActorBridgeTargets_ = new ArrayList(INITIAL_CAPACITY_5);

        if (!l_ActorBridgeTargets_.contains(_target)) { 
            List l_OldTargets = (List) ((ArrayList)l_ActorBridgeTargets_).clone();
            l_ActorBridgeTargets_.add(_target);
            containsBridgeTarget_ = true;

                                    // for testing and tracking
                                    //System.out.println("ACSImpl.addOwnActorBridgeTarget() l_ActorBridgeTargets_"+
                                      //      " at the end:");
                                    //for (Iterator itetest = l_ActorBridgeTargets_.iterator(); itetest.hasNext(); ) {
                                      //  System.out.println(" -------- "+((ACSObject)itetest.next()).getName());
                                    //}

            try {
                firePropertyChange("BridgeTargets",l_OldTargets,
                        ((ArrayList)l_ActorBridgeTargets_).clone());
            } catch (Exception e) {
                ARoad0.Gui1.DialogBox.showAlarmMessage(
                  "There is an exception after the base updating, from outside the database."+
                  "You should save the base and stop the program, then run it once again.");
                System.out.println("From ACSImpl.addOwnActorBridgeTarge(), "+
                        "firePropertChange() throws:");
                e.printStackTrace();
            }
            l_OldTargets = null;        
        } // if

  } // addOwnActorBridgeTarget()

/**
 * Removes an own actor as bridge targets for one or several external actor sources.
 * Searches in m_ActorBridgeTargets_ all the sources which have the argument
 * as bridge target.
 * Does not fire any property event.
 * Called by the own BridgeTarget.finalizeForUser() and closeBridges().
 * Uses removeOwnActorBridgeTargetFromSource().
 * @param _target non-null own actor which is a bridge target
 * @exception CreateError if the source is null or not registered, or
 *   InternalError if the associated source is not found
 */
  protected final void removeOwnActorBridgeTarget(ImmutableActor _target)
      throws CreateError {
        if (_target == null || !l_ActorBridgeTargets_.contains(_target))
          throw new CreateError(
            "in ACSImpl.removeOwnActorBridgeTarget(), _target is null or "+
              "is not registered as an own target.");
        
        Map.Entry entry = null;;
        Map m_copy = (Map) ((TreeMap)m_ActorBridgeTargets_).clone();
        try {
            for (Iterator ite = m_copy.entrySet().iterator(); ite.hasNext(); ) {
                entry = (Map.Entry) ite.next();

                if ( entry.getValue() == _target) {
                        // updates m_ActorBridgeTargets_
                     removeOwnActorBridgeTargetFromSource( (Actor)
                          BaseUtilityImpl.getEorBaseObject((ImmutableName)entry.getKey()),
                          this);
                } // if
            } // for
        } catch (BaseError b) {
            throw new CreateError("in ACSImpl.removeOwnActorBridgeTarget(): "+b);
        }
        entry = null;
        m_copy = null;        
  } // removeOwnActorBridgeTarget()
  
/**
 * Removes an own actor as a bridge target for an external actor source.
 * Called by the source instance from setEorACSBridgeTarget() and finalizeForBase(),
 * and by removeOwnActorBridgeTarget(). Calls _source.setEorACSBridgeTarget()
 * if the caller is not the source.
 * Does not update the _source ACS.
 * Fires a property change 'BridgeTargets' only if the relevant own target
 * has _source as the unique bridge source.
 * @param _source non-null actor which is an axternal source (not the target!)
 * @param _caller is used to call to the source if it is not the caller
 * @exception CreateError if the source is null or not registered
 */
  protected final void removeOwnActorBridgeTargetFromSource(
          ImmutableActor _source, BaseObject _caller) throws CreateError {
        if (_source == null || !m_ActorBridgeTargets_.containsKey(_source.getName()))
          throw new CreateError(
            "in ACSImpl.removeOwnActorBridgeTargetFromSource(), _source is null or "+
              "is not registered as an external source.");
        
        Actor target = (Actor) m_ActorBridgeTargets_.get(_source.getName());
        ArrayList<ImmutableName> l_sources = null;

        if (_caller != _source)
            ((ActorImpl)_source).setEorACSBridgeTarget(null, this);
        
        m_ActorBridgeTargets_.remove(_source.getName());


        if (m_closedSourcesOfActorBridgeTargets_ == null) {
                m_closedSourcesOfActorBridgeTargets_ = new TreeMap();
        } // if
        if (m_closedSourcesOfActorBridgeTargets_.get(target.getName()) == null) {
            l_sources = new ArrayList(INITIAL_CAPACITY_25);
            l_sources.add(_source.getName());

            m_closedSourcesOfActorBridgeTargets_.put(
              target.getName(),
              l_sources);
        } // if
        else {
            l_sources = m_closedSourcesOfActorBridgeTargets_.get(target.getName());
            l_sources.add(_source.getName());
        }

                                    // for testing and tracking
                                    //System.out.println("ACSImpl.removeOwnActorBridgeTargetFromSource(), for "+
                                      //      name_+", in m_closedSourcesOfActorBridgeTargets_"+
                                        //        " updates the values for the key "+target.getName());

        // 'target' may be the target for several sources in m_ActorBridgeTargets_
        if (!m_ActorBridgeTargets_.containsValue(target)) {
            List l_OldTargets = (List) ((ArrayList)l_ActorBridgeTargets_).clone();
            l_ActorBridgeTargets_.remove(target);
            firePropertyChange(
                    "BridgeTargets",l_OldTargets,((ArrayList)l_ActorBridgeTargets_).clone());
            l_OldTargets = null;


                                // for testing and tracking
                                //System.out.println("ACSImpl.removeOwnActorBridgeTargetFromSource(), "+
                                  //      "l_ActorBridgeTargets_ after the 'BridgeTargets' event firing:");
                                //for (Iterator itetest = l_ActorBridgeTargets_.iterator(); itetest.hasNext(); ) {
                                  //  System.out.println(" -------- "+((ACSObject)itetest.next()).getName());
                                //}

        } // if

        if (m_ActorBridgeTargets_.isEmpty())
            containsBridgeTarget_ = false;
        
        target = null;
        l_sources = null;
  } // removeOwnActorBridgeTargetFromSource()

/**
 * Sets an own actor as a bridge source for an external actor.
 * Fires the property event 'BridgeSources'.
 * Called by _actor.setEorACSBridgeTarget().
 * @param _actor non-null actor which is a source
 * @exception CreateError if the actor does not belong to the ACS
 */
  protected final void addOwnActorBridgeSource(ImmutableActor _actor)
      throws CreateError {
        if (!getManageBridges())
          throw new CreateError("The bridge source ACS "+
               "does not manage bridges to other ACS.");
        if (_actor == null || _actor.getEorACS() != this) throw new CreateError(
          "in ACSImpl.addOwnActorBridgeSource(), _actor is null or "+
          "does not belong to the ACS.");
        
        if (l_ActorBridgeSources_ == null)
            l_ActorBridgeSources_ = new ArrayList(INITIAL_CAPACITY_5);
        if (l_ActorBridgeSources_.contains(_actor)) throw new CreateError(
          "in ACSImpl.addOwnActorBridgeSource(), _actor is already known.");
        
        List l_OldSources = (List) ((ArrayList)l_ActorBridgeSources_).clone();
        l_ActorBridgeSources_.add(_actor);
        containsBridgeSource_ = true;

                                    // for testing and tracking
                                    //System.out.println("in ACSImpl.addOwnActorBridgeSource() l_ActorBridgeSources_"+
                                      //      " at the end:");
                                    //for (Iterator itetest = l_ActorBridgeSources_.iterator(); itetest.hasNext(); ) {
                                      //  System.out.println(" -------- "+((ACSObject)itetest.next()).getName());
                                    //}

        if (m_closedTargetsOfActorBridgeSources_ != null &&
             m_closedTargetsOfActorBridgeSources_.get(_actor.getName()) != null) {
                m_closedTargetsOfActorBridgeSources_.remove(_actor.getName());
        } // if

        try {
            firePropertyChange("BridgeSources",l_OldSources,
              ((ArrayList)l_ActorBridgeSources_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.addOwnActorBridgeSource(), firePropertChange() throws:");
            e.printStackTrace();
        }

        l_OldSources = null;          
  } // addOwnActorBridgeSource()
  
/**
 * Removes an own actor which is a bridge source for an external actor.
 * Fires the property event 'BridgeSources'.
 * Called by _actor.setEorACSBridgeTarget().
 * Updates m_closedTargetsOfActorBridgeSources_ and calls back
 * getEorACSBridgeTarget() on the argument.
 * @param _actor non-null actor which is a source
 * @exception CreateError if not registered
 */
  protected final void removeOwnActorBridgeSource(ImmutableActor _actor)
      throws CreateError {
        
                                // for testing and tracking
                                //System.out.println("ACSImpl.removeOwnActorBridgeSource(), starting"+
                                 //" for the actor: "+_actor.getName());

          
        if (_actor == null || !l_ActorBridgeSources_.contains(_actor))
          throw new CreateError(
            "in ACSImpl.removeOwnActorBridgeSource(), _actor is null or "+
            "is not registered as source. _actor: "+_actor);
        
        List l_OldSources = (List) ((ArrayList)l_ActorBridgeSources_).clone();
        l_ActorBridgeSources_.remove(_actor);
        if (l_ActorBridgeSources_.isEmpty()) {

                                // for testing and tracking
                                //System.out.println("ACSImpl.removeOwnActorBridgeSource(), "+
                                 //" containsBridgeSource_ set to 'false' for the actor: "+
                                 //_actor.getName());

            containsBridgeSource_ = false;
        }

        if (m_closedTargetsOfActorBridgeSources_ == null)
            m_closedTargetsOfActorBridgeSources_ = new TreeMap();
        m_closedTargetsOfActorBridgeSources_.put(
                _actor.getName(),
                _actor.getEorACSBridgeTarget().getName() );

                                    // for testing and tracking
                                    //System.out.println("ACSImpl.removeOwnActorBridgeSource(), for "+
                                      //      name_+", in m_closedTargetsOfActorBridgeSources_"+
                                        //        " put the key "+_actor.getName());

                                    // for testing and tracking
                                    //System.out.println("in ACSImpl.removeOwnActorBridgeSource() after the update:");
                                    //for (Iterator itetest = l_ActorBridgeSources_.iterator(); itetest.hasNext(); ) {
                                        //System.out.println(" -------- "+((ACSObject)itetest.next()).getName());
                                    //}
        try {
            firePropertyChange("BridgeSources", l_OldSources,
                ((ArrayList)l_ActorBridgeSources_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.removeOwnActorBridgeSource(), firePropertChange() throws:");
            e.printStackTrace();
        }

        l_OldSources = null; 
        return;           
  } // removeOwnActorBridgeSource()

/**
 * Adds an ACSObject which has at least one relation with any opened child ACS,
 * among the possible relations &lt;ForChildAcsOpeningAndClosing&gt;,
 * &lt;ForChildAcsOpeningNoClosing&gt; and &lt;ForChildAcsIntegrity&gt;.
 * These relations set that the deleting or the updating of the ACSObject will
 * stop the child ACS or will modify its ability to run correctly.
 * Fires the property event 'RootsForChildACS'.
 * Called by the child ACS setAcsRelationsInIS() method.
 * @param _child child ACS name
 * @param _root from this ACS to add as root for _child. Noop if it is already known.
 * @exception CreateError if an argument is null, _child is this ACS or if
 * _root does not belong to this ACS
 */
  protected final void addRootForChildACS(ImmutableName _child, ACSObject _root)
     throws CreateError {

                        // for testing and tracking
                        //System.out.println("call to ACSImpl.addRootForChildACS() in "+name_+
                          //      " for the root "+_root.getName());

      if (_child == null || _child.isEmpty() || _child.equals(name_)) throw new CreateError(
          "in ACSImpl.addRootForChildACS(), _child is null or this ACS: "+_child);
      if (!BaseManagerImpl.getInstance().isOpenACS(name_) ) throw new CreateError(
          "in ACSImpl.addRootForChildACS(), the ACS is closed ");
      if (_root == null || !_root.getAcsName().equals(name_)) throw new CreateError(
          "in ACSImpl.addRootForChildACS(), _root is null or does not belong "+
          "to this ACS. _root: "+_root);
      
        if (m_RootsForChildACS_ == null)
            m_RootsForChildACS_ = new HashMap(INITIAL_CAPACITY_5);
        
        Map old = (Map) ((HashMap)m_RootsForChildACS_).clone();
        
        if (m_RootsForChildACS_.get(_child.toString()) == null)
            m_RootsForChildACS_.put(_child.toString(), new ImmutableName[] {_root.getName()});
        else {
            if ( Arrays.asList((ImmutableName[])m_RootsForChildACS_.get(
                    _child.toString())).contains(_root.getName())) {
                old = null;
                return;
            } // if

            ImmutableName[] array = new ImmutableName[1 +
                    ((ImmutableName[])m_RootsForChildACS_.get(_child.toString())).length];
            System.arraycopy(
                    (ImmutableName[])m_RootsForChildACS_.get(_child.toString()),
                            0, array, 0, array.length - 1);
            array[array.length - 1] = _root.getName();
            m_RootsForChildACS_.put(_child.toString(), array);
            array = null;
        } // else

        try {
            firePropertyChange("RootsForChildACS", old, ((HashMap)m_RootsForChildACS_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.addRootForChildACS(), firePropertChange() throws:");
            e.printStackTrace();
        }

        old = null;          
  } // addRootForChildACS()
  
/**
 * Removes an ACSObject which has no more relations with any opened child ACS,
 * among the possible relations &lt;ForChildAcsOpeningAndClosing&gt;,
 * &lt;ForChildAcsOpeningNoClosing&gt; and &lt;ForChildAcsIntegrity&gt;.
 * These relations set that the deleting or the updating of the ACSObject will
 * stop the child ACS or will modify its ability to run correctly.
 * Fires the property event 'RootsForChildACS'.
 * Called by the child ACS.finalizeForBase() method for with-closing roots.
 * At the ACS opening, setAcsRelationsInIS() checks the deleted child ACSs
 * for without-closing roots, and deleted them directly in the property whithout
 * any call to this method.
 * @param _child child ACS name
 * @param _root to remove
 * @exception CreateError if an argument is null, _child is this ACS, if
 * _root does not belong to this ACS, or if _obj is unknown.
 */
  protected final void removeRootForChildACS(ImmutableName _child, ACSObject _root)
      throws CreateError {
          
                // for testing and tracking
                //System.out.println("call to ACSImpl.removeRootForChildACS() in "+name_+
                 //", the object to remove is "+_root.getName());

        if (_child == null || _root == null || _child.equals(name_) ||
                !_root.getAcsName().equals(name_)) throw new CreateError(
          "in ACSImpl.removeRootForChildACS(), an argument is null, _child is this ACS or "+
          "_root does not belong to this ACS.");

        if (m_RootsForChildACS_ == null) throw new CreateError(
          "in ACSImpl.removeRootForChildACS(), the map is empty.");

        Map             old = (Map) ((HashMap)m_RootsForChildACS_).clone();
        ImmutableName[] array = m_RootsForChildACS_.get(_child.toString());

        if (array == null)
            throw new CreateError("in ACSImpl.removeRootForChildACS(), _child is unknown.");

        ImmutableName[] tempArray = new ImmutableName[array.length];
        int             shiftingBack = 0;


        if ( !Arrays.asList(array).contains(_root.getName())) {
            old = null;
            array = null;
            tempArray = null;
            throw new CreateError("in ACSImpl.removeRootForChildACS(), _root is unknown.");
        } // if
        else if (array.length == 1) {
            m_RootsForChildACS_.remove(_child.toString());
        }
        else {
            for (int  i = 0 ; i < array.length ; i++) {
                if (array[i] == null) shiftingBack++;  // deletes the empty item
                else if (array[i].equals(_root.getName()) ) {
                    // incrementing shiftingBack deletes the item
                  shiftingBack++;
                }
                else tempArray[i - shiftingBack] = array[i];
            } // for i

            array = new ImmutableName[array.length - shiftingBack];
            System.arraycopy(tempArray, 0, array, 0, tempArray.length - shiftingBack);

            m_RootsForChildACS_.put(_child.toString(), array);
        } // else

        try {
            firePropertyChange("RootsForChildACS", old, ((HashMap)m_RootsForChildACS_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.removeRootForChildACS(), firePropertChange() throws:");
            e.printStackTrace();
        }

        old = null;
        array = null;
        tempArray = null;
  } // removeRootForChildACS()

  /**
   * Sets the maps and lists describing the relations of this new ACS in the IS Name Tree.
   * Called by ACSFactoryImpl just after the creation of the instance.
   * Creates the roots if necessary, and updates the roots map in the parent ACS.
   * Does not fire any property change. Calls setAcsRelationsInIS().
   * @param _m_isComposite the map of the composite relations. 
   * The key is the subACS name, in one word,
   * or the ACS name last component, and the value is the Composite name.
   * May be null but not empty.
   * @param _m_RootsForAcsOpeningAndClosing where the map key has the format
   * ACS.Name.<Value>.XXX which is a parent ACS name having the value XXX.
   * The map values is a set of commands. May be null but not empty.
   * @param _m_RootsForAcsOpeningNoClosing where the map key has the format
   * ACS.Name.<Value>.XXX which is a parent ACS name having the value XXX.
   * The map values is a set of commands. May be null but not empty.
   * @param _m_RootsForAcsIntegrity where the map key has the format
   * ACS.Name.<Value>.XXX which is a parent ACS name having the value XXX.
   * The map values is a set of commands. May be null but not empty.
   * @param _m_SubAndGroupPolicy where the map key is the subACS name, in one word,
   * and the map value is an array which may contain the following values: <P>
   * 'Resources', 'UserIDs' and 'GroupIDs'. May be null but not empty. <P>
   * @exception BaseError from AssociateUtilityImpl or addRootForChildACS()
   */
  protected void initializeAcsRelationsInIS(
          HashMap<String, ImmutableName> _m_isComposite,
          HashMap<String, String[]> _m_RootsForAcsOpeningAndClosing,
          HashMap<String, String[]> _m_RootsForAcsOpeningNoClosing,
          HashMap<String, String[]> _m_RootsForAcsIntegrity,
          HashMap<String, String[]> _m_SubAndGroupPolicy) throws BaseError {

      m_BelongsToComposite_ = _m_isComposite;
      m_SubAndGroupPolicy_ = _m_SubAndGroupPolicy;

      m_RootCommandsForAcsOpeningAndClosing_ = _m_RootsForAcsOpeningAndClosing;
      m_RootCommandsForAcsOpeningNoClosing_ = _m_RootsForAcsOpeningNoClosing;
      m_RootCommandsForAcsIntegrity_ = _m_RootsForAcsIntegrity;

      setAcsRelationsInIS(true, false);
  } // initializeAcsRelationsInIS()

  /**
   * Sets the maps and lists describing the relations of this ACS in the IS Name Tree.
   * Called by initializeAcsRelationsInIS() and by BaseManagerImpl
   * after the opening or the importation of the ACS.
   * Creates the roots if necessary, and updates the roots map in the parent ACS
   * and the roots themself. Checks the child ACS roots, and deletes the without-closing roots
   * if the child ACS has been deleted from the base.
   * Caution: any returned ACSObject by AssociateUtilityImpl.addRootsInParentACSs()
   * is put in the root names lists if it does not belong to this ACS. When a root
   * command calls a method to get an argument for a next root command, it should
   * not be such an ACSObject, to avoid to be seen as a root.
   * The BaseErrors are not thrown during an importation, during the creation of the roots.
   * The importation operation is not stopped by such an error, and it may deliver
   * a partial set of roots in the parent ACS. <P>
   * Does not fire any property change. Calls addRootForChildACS(),
   * AssociateUtilityImpl.addRootsInParentACSs()
   * and getL_AssociatesAndCalledMethods().
   * @param _atCreation true at the creation or importation of the ACS, false at its opening
   * @param _atImport true only at the importation of the ACS
   * @exception BaseError from AssociateUtilityImpl, addRootForChildACS(),
   *        NameImpl.getNameFromString(), Resource/EligibleParty.finalizeForUser()
   */
  protected void setAcsRelationsInIS(boolean _atCreation, boolean _atImport) throws BaseError {

      List<ImmutableName>   l_roots = null;
      Map                   m_asso = null;
      //Set<ImmutableName>    l_noDeletingRoots = null;
      ImmutableName         name2 = null;
      Map.Entry             entry2 = null;
      boolean               importError = false;

      m_asso = getM_RootCommandsForOpeningAndClosing();

      if (!m_asso.isEmpty()) {
          try {
              // creates or updates the with-closing roots in the parent ACSs,
              // but does not update the parent ACS map for child ACSs nor the relevant root property.
              // throw BaseError
              l_RootNamesForAcsOpeningWithClosing_ = new ArrayList(INITIAL_CAPACITY_5);
              for (Object ob : AssociateUtilityImpl.addRootsInParentACSs(this, m_asso)) {
                  if (ob instanceof ACSObject &&
                          !((ACSObject)ob).getAcsName().equals(name_) )
                                l_RootNamesForAcsOpeningWithClosing_.add(((ACSObject)ob).getName());
              } // for ob
          }
          catch (BaseError u) {
                // no throwing in an importation
                if (!_atImport)
                    throw new BaseError(u.toString());
          }
      } // if

      m_asso = getM_RootCommandsForOpeningNoClosing();

      if (_atCreation && !m_asso.isEmpty()) {
          try {
              // creates the no-closing roots in the parent ACSs,
              // but does not update the parent ACS map for child ACSs nor the relevant root property.
              // throw BaseError
              if (l_RootNamesForAcsOpeningNoClosing_ == null)
                  l_RootNamesForAcsOpeningNoClosing_ = new ArrayList(INITIAL_CAPACITY_5);
              for (Object ob : AssociateUtilityImpl.addRootsInParentACSs(this, m_asso)) {
                  if (ob instanceof ACSObject &&
                          !((ACSObject)ob).getAcsName().equals(name_) )
                                l_RootNamesForAcsOpeningNoClosing_.add(((ACSObject)ob).getName());
              } // for ob
          }
          catch (BaseError u) {
                // no throwing in an importation
                if (!_atImport)
                    throw new BaseError(u.toString());
          }
      } // if

      /*
      if (_atCreation && l_RootNamesForAcsOpeningWithClosing_ != null &&
              l_RootNamesForAcsOpeningNoClosing_ != null) {
          l_RootNamesForAcsOpeningWithClosing_.addAll(l_RootNamesForAcsOpeningNoClosing_);
          //l_noDeletingRoots = AssociateUtilityImpl.getL_RootsWithoutDeletingInParentACSs(this);
          //if (l_noDeletingRoots != null)
            //  l_RootNamesForAcsOpeningNoClosing_.removeAll(l_noDeletingRoots);
          //l_noDeletingRoots = null;
      } // if
      else if (_atCreation && l_RootNamesForAcsOpeningWithClosing_ == null &&
              l_RootNamesForAcsOpeningNoClosing_ != null)
          l_RootNamesForAcsOpeningWithClosing_ =
                  (List) ((ArrayList)l_RootNamesForAcsOpeningNoClosing_).clone();
       * */

      if (m_RootCommandsForAcsIntegrity_ != null) {
          try {
              for (Map.Entry entry : m_RootCommandsForAcsIntegrity_.entrySet()) {
                  if (l_RootNamesForAcsIntegrity_ == null)
                      // does not update the parent ACSs
                      // throw BaseError
                      l_RootNamesForAcsIntegrity_ = new ArrayList(
                          AssociateUtilityImpl.getL_AssociatesAndCalledMethods(
                          BaseUtilityImpl.getEorBaseObject(NameImpl.getNameFromString(
                            (String)entry.getKey())),
                          (String[])entry.getValue(), this).keySet());
                  else
                      l_RootNamesForAcsIntegrity_.addAll(
                          AssociateUtilityImpl.getL_AssociatesAndCalledMethods(
                          BaseUtilityImpl.getEorBaseObject(NameImpl.getNameFromString(
                            (String)entry.getKey())),
                          (String[])entry.getValue(), this).keySet());
              } // for entry
          }
          catch (BaseError u) {
                // no throwing in an importation
                if (!_atImport)
                    throw new BaseError(u.toString());
          }
      } // if (m_RootCommandsForAcsIntegrity_

      // updates the parent ACS and root properties at each opening
      if (l_RootNamesForAcsOpeningWithClosing_ != null) {
          try {
              addInParentACS(l_RootNamesForAcsOpeningWithClosing_);
          } catch (CreateError c) {
                // no throwing in an importation
                if (!_atImport)
                    throw new InternalError("ACSImpl.setAcsRelationsInIS() for "+
                      "l_RootNamesForAcsOpeningWithClosing_: "+c.toString());
                else {
                    importError = true;
                }
          }
          if (!importError) {
              containsRootActions_ = true;
              importError = false;
          } // if
      } // if
      if (_atCreation && l_RootNamesForAcsOpeningNoClosing_ != null) {
          try {
              addInParentACS(l_RootNamesForAcsOpeningNoClosing_);
          } catch (CreateError c) {
              // no throwing in an importation
                if (!_atImport)
                    throw new InternalError("ACSImpl.setAcsRelationsInIS() for "+
                      "l_RootNamesForAcsOpeningNoClosing_: "+c.toString());
          }
      } // if
      // works after creation
      if (m_RootCommandsForAcsOpeningNoClosing_ != null) {
          containsRootActions_ = true;
      } // if
      if (l_RootNamesForAcsIntegrity_ != null) {
          try {
              addInParentACS(l_RootNamesForAcsIntegrity_);
          } catch (CreateError c) {
              // no throwing in an importation
                if (!_atImport) {
                    throw new InternalError("ACSImpl.setAcsRelationsInIS() for "+
                      "l_RootNamesForAcsIntegrity_: "+c.toString());
                } // if
                else {
                    importError = true;
                }
          } // catch
          if (!importError)
              containsRootActions_ = true;
      } // if

      // deletes in the parent ACS, the child ACS no-closing roots if their ACS is deleted
      // (not if the ACS is closed)
      try {
          if (m_RootsForChildACS_ != null) {
              for (Iterator ite = m_RootsForChildACS_.entrySet().iterator(); ite.hasNext(); ) {
                  entry2 = (Map.Entry)ite.next();
                  name2 = NameImpl.getNameFromString((String)entry2.getKey());
                  if (BaseManagerImpl.getInstance().knowsACS(name2))
                      continue;
                  else {
                      // deletes the roots
                      l_roots = Arrays.asList( (ImmutableName[])entry2.getValue() );
                      
                      for (ImmutableName name3 : l_roots) {

                                        // for testing and tracking
                                        //System.out.println("ACSImpl.setAcsRelationsInIS() in "+name_+
                                          //      " deletes the root: "+name3);

                          // finalizes, even with the '<byAcsCreator>' type.
                          // the object may be null if one of its directory has been finalized.
                          if (m_Resources_.get(name3) != null)
                              ((Basic) m_Resources_.get(name3)).finalizeForBase();
                      } // for
                      
                      // deletes the child ACS
                      ite.remove();
                  } // else
              } // for ite
          } // if
      }
      catch (UpDateError u) {
            throw new BaseError(u.toString());
      }

                        // for testing and tracking
                        //System.out.println("end of ACSImpl.setAcsRelationsInIS() in "+name_+
                          //      " containsRootActions_ is "+containsRootActions_);

      m_asso = null;
      l_roots = null;
      //l_noDeletingRoots = null;
      name2 = null;
      entry2 = null;
  } // setAcsRelationsInIS()
  
    //
    // ------------- PRIVATE METHODS ------------------------------------------
    //
    
  /**
  * This method is called by removeEorResource() and removeEorVirtualFolder().
  * It takes off _ACSname into the list of Acs Controllers of this ACS,
  * and it updates controlled_. 
  * Note: there is an ACS controllers list in Resource, but not on VirtualFolder.
  * Fires the PropertyChange events 'OneController' and 'Controlled'.
  * @param _ACSname open ACS that controls this ACS
  * @exception UpDateError if _ACSname is unknown.
  */
  protected final void removeAcsController(ImmutableName _ACSname)
      throws UpDateError {

                                    // for testing and tracking
                                    //System.out.println("ACSImpl.removeAcsController() starts on "+
                                      //      name_.lastComponent()+", for the controller: "+
                                        //    _ACSname.lastComponent());
      
        int                 shiftingBack = 0;
        ImmutableName[]     newArray = new ImmutableName[l_OpenAcsControllerNames_.length];
        ImmutableName[]     result = new ImmutableName[0];
                // this code works even if shiftingBack > 1
        for (int  i = 0 ; i < l_OpenAcsControllerNames_.length ; i++) {
             // deletes the empty l_OpenAcsControllerNames_[i]
            if (l_OpenAcsControllerNames_[i] == null)
                shiftingBack++;
            if (l_OpenAcsControllerNames_[i].equals(_ACSname) ) {
                // incrementing shiftingBack deletes l_OpenAcsControllerNames_[i]
                shiftingBack++;
            }
            else newArray[i - shiftingBack] = (ImmutableName)l_OpenAcsControllerNames_[i];
        } // for i

        if (l_OpenAcsControllerNames_.length != shiftingBack) {

                        // for testing and tracking
                        //System.out.println("ACSImpl.removeEorVirtualFolder(), l_OpenAcsControllerNames_ lenght:"+
                          //  l_OpenAcsControllerNames_.length);
                        //System.out.println("ACSImpl.removeEorVirtualFolder(), l_OpenAcsControllerNames_ shiftingBack:"+
                          //  shiftingBack);

            result = new ImmutableName[l_OpenAcsControllerNames_.length - shiftingBack];
            System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
        }
        l_OpenAcsControllerNames_ = result;

                        // for testing and tracking
                        //System.out.println("ACSImpl.removeEorVirtualFolder(), "+
                          //      "l_OpenAcsControllerNames_ lenght: "+
                            //    l_OpenAcsControllerNames_.length);

      if (shiftingBack == 0) {
          getEorBaseError().addException(
                  "ACSImpl.removeAcsController(Name): Name is unknown ");
          throw new UpDateError("ACS.removeAcsController() in "+name_.lastComponent()+
                  " where the ACS argument "+_ACSname.lastComponent()+" is unknown ");
      }

      // updates controlled_
      if (l_OpenAcsControllerNames_.length == 0) {
          controlled_ = false;
          try {
                firePropertyChange("Controlled", Boolean.TRUE, Boolean.FALSE);
          } catch (Exception e) {
                ARoad0.Gui1.DialogBox.showAlarmMessage(
                  "There is an exception after the base updating, from outside the database."+
                  "You should save the base and stop the program, then run it once again.");
                System.out.println("From ACSImpl.removeAcsController(), firePropertChange() throws:");
                e.printStackTrace();
          } // catch
      } // if

      try {
            firePropertyChange("OneController", _ACSname, null);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.removeAcsController(), firePropertChange() throws:");
            e.printStackTrace();
      } // catch


      newArray = null;
      result = null;
  } // removeAcsController()

/**
 * Called by addResource() and addVirtualFolder().
 * @param _ACSname of open ACS that controls a resource
 * of this ACS. This method is called by newEorAclEntry() from the AclEntry ACS
 * if it is an another ACS, and by PrivilegeforLinks.addEorPrivilegeSource/TargetLink()
 * if it is from another ACS.
 * If _ACSName is unknown, it adds _ACSname in the list of AcsControllers,
 * sets controlled_ to true.
 * Fires the PropertyChange events 'OneController' and 'Controlled'.
 * @exception CreateError if the ACS is closed
 * (no exception if it is already known).
 */
  protected final void addAcsController(ImmutableName _ACSname)
      throws CreateError {

                                    // for testing and tracking
                                    //System.out.println("ACSImpl.addAcsController() starts on "+
                                      //      name_.lastComponent()+", for the controller: "+
                                        //    _ACSname.lastComponent());

      // method variables:
    Map           m_openACS = BaseManagerImpl.getInstance().getEorM_OpenACS();
    boolean       recorded = false;
    ImmutableName[] newArray = null;

    // exception if _ACSname is closed, as in a SaveAs action
    if (!m_openACS.containsKey(_ACSname) ) {
        getEorBaseError().addException("Resource.addAcsController(Name): Name of a closed ACS");
        throw new CreateError(" The ACS controller is a closed ACS.");
    }

    // return if _ACSname is already known:
    for (int i = 0 ; i < l_OpenAcsControllerNames_.length ; i++) {
        // compatibility: l_OpenAcsControllerNames_[i] from 0.7.0 may be null
        // and it is removed by getL_AcsControllers()
        if (l_OpenAcsControllerNames_[i] != null &&
                l_OpenAcsControllerNames_[i].equals(_ACSname) ) {
            recorded = true;
            break;
        }
    }
    if (recorded) {
        m_openACS = null;
        return;
    }
    else {
        newArray = new NameImpl[l_OpenAcsControllerNames_.length + 1];
        System.arraycopy(l_OpenAcsControllerNames_, 0, newArray, 0,
                l_OpenAcsControllerNames_.length);
        newArray[l_OpenAcsControllerNames_.length] = (Name)_ACSname.clone();
        l_OpenAcsControllerNames_ = newArray;
    } // else

    // updates controlled_
    if (!controlled_) {
      controlled_ = true;
      try {
            firePropertyChange("Controlled", Boolean.FALSE, Boolean.TRUE);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ACSImpl.removeAcsController(), firePropertChange() throws:");
            e.printStackTrace();
      } // catch
    } // if

    try {
      firePropertyChange("OneController", null, _ACSname);
    } catch (Exception e) {
    ARoad0.Gui1.DialogBox.showAlarmMessage(
      "There is an exception after the dbase updating, from outside the database."+
      "You should save the base and stop the program, then run it once again.");
    System.out.println("From ACSImpl.addAcsController(), firePropertChange() throws:");
    e.printStackTrace();
    }

    m_openACS = null;
    newArray = null;
  } // addAcsController()

/**
 * Adds the root properties to the parent ACS, and to the roots themself in the parent ACS.
 * Called by setAcsRelationsInIS().
 * @param _l_roots list of the roots to add in the parent ACSs
 * @exception CreateError if the ACS is closed
 * (no exception if it is already known).
 */
  protected final void addInParentACS(List<ImmutableName> _l_roots)
    throws CreateError, BaseError {
      ACSObject obj = null;
      ImmutableName acsname = null;
      ACSImpl acs = null;

      // updates the parent ACSs
      for (ImmutableName name : _l_roots) {
          obj = (ACSObject)BaseUtilityImpl.getEorBaseObject(name);
          acsname = obj.getAcsName();
          acs = (ACSImpl)BaseManagerImpl.m_OpenAcs__.get(acsname);

          acs.addRootForChildACS(name_, obj);

          if (obj instanceof Basic) {
            // the key is <ForChildAcsOpeningAndClosing>,
              // <ForChildAcsOpeningNoClosing>
                // or <ForChildAcsIntegrity>
            if (l_RootNamesForAcsOpeningWithClosing_ != null &&
                    l_RootNamesForAcsOpeningWithClosing_.contains(name) )
                ((BasicImpl)obj).addConstrainedChildACS(
                    "<ForChildAcsOpeningAndClosing>", new ImmutableName[] {name_});
            else if (l_RootNamesForAcsOpeningNoClosing_ != null &&
                    l_RootNamesForAcsOpeningNoClosing_.contains(name))
                ((BasicImpl)obj).addConstrainedChildACS(
                    "<ForChildAcsOpeningNoClosing>", new ImmutableName[] {name_});
            else if (l_RootNamesForAcsIntegrity_ != null &&
                    l_RootNamesForAcsIntegrity_.contains(name) )
                ((BasicImpl)obj).addConstrainedChildACS(
                    "<ForChildAcsIntegrity>", new ImmutableName[] {name_});
          } // if
      } // for name

      obj = null;
      acsname = null;
      acs = null;
  } // addInParentACS()

} // ACSImpl

