//Title:        Access Road
//Version:      0.7.1
//Copyright:    2011-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.Map;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Arrays;
import java.util.List;
import java.util.Collections;

import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.BaseError;

//import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;

/**
 * This class is responsible for modeling a MySQL Actor to implement varied constraints
 * on the instance, the property isRightProxy_ and the inherited AclEntries.
 * The editing of the default true value on the property isRightProxy_ is forbidden.
 * This class handles the Resource types 'trigger', 'view_delegate', 'stored_function',
 * 'stored_procedure', 'user socket'.
 * Note: the constructor forbids to have a Host that is not the host group
 * of the Definer account. This is why there is an error message
 * 'The Host is not the host group of the Definer'. Nonetheless, this is not
 * the behavior of the MySQL Server, where it is possible to have a disabled Definer. <P>
 * This class manages a map of primary rights, as inherited AclEntry rights,
 * and a map of global rights that says what users have a global access on the resource database.
 * The final access rights does not depends directly on the primary rights map, but
 * no global rights for a couple user/host means no access at all to the target. <P>
 * The class VirtualFolderMySQLImpl is responsible to create an Actor having the type
 * 'view delegate' under the database node in the resources tree, for the database
 * the virtual folder is associate to.
 * The ACS types policy, with 'Resource.NoNonConditionalACL' and 'Resource.NoConditionalACL',
 * sets that there is an empty 'AclEntry' property for the 'view_delegate' and
 * the 'trigger' actors as Respources, while the 'stored_function' and 'stored_procedure' actors
 * allows direct ACL to simulate the MySQL privileges table 'procs_priv'.
 * <P>
 * This class implements the following rules, and informs the Access Road user
 * with the same sentence as error message when they are not fulfilled with: <P>
 * - The Host is mandatory when there is a Definer. <P>
 * - The Host ACS is not the ACS of this actor. <P>
 * - The Host have to be a simple host group. <P>
 * - The Host is not the host group of the Definer. <P>
 * - A non-null Definer User must have a non-null Definer Host. <P>
 * - A trigger must have a non-null Definer user and a non-null Definer host. <P>
 * - An host group can have an AclEntry only for a database. <P>
 * - The AclEntry of a MySQL Actor must have a condition group. <P>
 * - A MySQL conditional AclEntry cannot use a standard group. <P>
 * This class has no subclass in this package.
 * It is a javabean without specific bound properties. The 'InheritedTargetAclEntries'
 * events are bound when the primary AclEntries map is changing.
 * Excepting for EPRViewInBaseImpl, the listeners are outside gBase,
 * they are transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * @version 0.7.1
 */
public class ActorMySQLImpl extends ActorImpl implements ActorMySQL {
  // private variables
  private static final long   serialVersionUID__ = 0;
  // protected variables

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ActorMySQLImpl EMPTY_INSTANCE = new ActorMySQLImpl();

 /**
  * The basic constructor
  */
  public ActorMySQLImpl() {


                          // for testing purpose
                     //System.out.println("Call to ActorMySQLImpl constructor");
  }
  
  /**
  * Used for the creation of an instance through the GUI.
  * An association script in AcsFactoryMySQL.create() uses a constructor from the superclass
  * to create a 'connection' actor with a bridge target, but without the argument _host
  * nor inheritance of the parent ACL.
  * This constructor is for the view delegates, the stored procedures, the stored functions
  * and the triggers. The Actor bridge target is set then to null.
  * The user error messages are: <P>
  * - The Host is mandatory when there is a Definer. <P>
  * - The Host ACS is not the ACS of this actor. <P>
  * - The Host have to be a simple host group. <P>
  * - The Host is not the host group of the Definer. <P>
  * - A non-null Definer User must have a non-null Definer Host. <P>
  * - A trigger must have a non-null Definer user and a non-null Definer host.
  * @param _name of the actor as a resource.
  * @param _acs main ACS of the actor.
  * @param _parent of the actor.
  * @param _owner owns the resource and is its current UserID; may be null.
  * @param _resourceType is also the EP type, and may be null if the ACS allows it.
  * @param _host the current GroupID. Has to be the simple host of _owner. May be null.
  * Never null if _owner is not null.
  * @exception CreateError from the super constructor, or _host null with _owner not null.
  */
  public ActorMySQLImpl(ImmutableName _name, ACS _acs, Directory _parent, 
    UserID _owner, String _resourceType, GroupIDMySQL _host)
    throws CreateError {
      
      super(_name, _acs, _parent, _owner, _resourceType, null, true); // null bridge target

      if (_host != null) {
          if (_owner == null) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From ActorMySQLImpl constructor. "+
                      "Error on the null Definer. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Definer is mandatory when there is a Host.");
          }
          else if (_host.getEorACS() != aCS_) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From ActorMySQLImpl constructor. "+
                      "Error on the Host ACS. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Host ACS is not the ACS of this actor.");
          }
          else if (!_host.isSimpleHost()) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From ActorMySQLImpl constructor. "+
                      "Error on the no-simple Host. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Host have to be a simple host group.");
          }
          else if (((UserIDMySQL) _owner).getEorFirstHost() != _host) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From ActorMySQLImpl constructor. "+
                      "Error on the Host that is not the Definer host. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Host is not the host group of the Definer.");
          } // else if


          try {
              setEorCurrentGroupID(_host);
          }
          catch (UpDateError u) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError n) {
                  throw new CreateError(" From ActorMySQLImpl constructor. "+
                      "Error on the Host as current Group. "+
                      "Error on finalizing "+n.toString());
              }

              throw new CreateError(u.toString());
          } // catch u

      } // if (_host != null)
      else {
          // host is mandatory if there is an owner
          if (_owner != null) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From ActorMySQLImpl constructor. "+
                      "Error on the null Host. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" A non-null Definer User must have a non-null Definer Host. ");
          } // if
          
          // 'Actor.NoNullCurrentAccount' forbids a null Definer for triggers
          if (_resourceType.indexOf("trigger") != -1) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From ActorMySQLImpl constructor. "+
                      "Error on the Host or the User for a trigger. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" A trigger must have a non-null Definer user and"+
                      "a non-null Definer host. ");
          } // if
      } // else if (_host == null)

      if (_parent != null ) {

        // the AclEntries and the inherited AclEntries from the parent are added
        AclEntry    acl2 = null;
        AclEntry[]  l_acl = _parent.getL_aclEntry();
        ArrayList<AclEntry> l_inh_acl = ((ResourceImpl)_parent).getEorL_TargetInheritedAclEntries();
        boolean     noInheritance = false;

                                    // for testing and tracking
                                    //System.out.println("");
                                    //System.out.println("ActorMySQLImpl constructor for "+
                                      //   name_+
                                        // " adds the direct ACLs of the parent as inherited ACLs ");

        for (int i = 0; i < l_acl.length; i++) {
            acl2 = selectInheritedAclEntry(l_acl[i], type_);
            if (acl2 == null) {
                noInheritance = true;
                continue;
            }

            addInheritedAclEntry(acl2);
        } // for i

                                  // for testing and tracking
                                  //System.out.println("ActorMySQLImpl constructor for "+
                                    //     name_+
                                      //   " adds the inherited ACLs of the parent as inherited ACLs ");

        for (AclEntry ac : l_inh_acl) {
            acl2 = selectInheritedAclEntry(ac, type_);
            if (acl2 == null) {
                noInheritance = true;
                continue;
            }

            addInheritedAclEntry(acl2);
        } // for ac

        if (noInheritance)
            setComment(comment_ + " No ACL inheritance for this type.");

        l_acl = null;
        l_inh_acl = null;
        acl2 = null;
      } // if (_parent != null )

                                  // for testing and tracking
                                  //System.out.println("end of the ActorMySQLImpl constructor for "+
                                    //     name_+" with the current userID "+currentUserID_);

  } // constructor

  //
  // To implements the ActorImpl methods
  //

  /**
   * No operation, and fires an exception 'An Actor in the AcsAddon MySQL cannot
   * change the 'right-proxy' value'.
   * @param _b true if this Actor may be a proxy in the access paths, that is,
   *    as rights user, it gets some rights to the other
   *    rights users which have an 'execute' right on it
  * @exception UpDateError in all cases
   */
  public void setIsRightProxy(boolean _b) throws UpDateError {
      throw new UpDateError(" An Actor in the AcsAddon MySQL cannot change "+
              "the 'right-proxy' value.");
  }

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon.Returns always null,
   * because these events are listened only to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: is optional, where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL. <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject
   * as one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()

  // ------------- protected methods -----------------------------------------


  // ------------------ copied methods from ResourceMySQLImpl ----------------

    /**
    * Creates the AclEntry, and sorts the new AclEntry list.
    * On every child of the root 'MySQL server', a new conditional ACL for a pair
    * (user@host or accounts-for-hosts group) produces a new ACL
    * on the root for the same pair if it does not exist. This new root ACL has
    * the right USAGE, and it is not deleted by default when the initial ACL is deleted.
    * For a new non-conditional ACL on a database, there is no adding on the root. <P>
    * The ACS types policy, with 'Resource.NoNonConditionalACL' and 'Resource.NoConditionalACL',
    * sets that there is an empty 'AclEntry' property for the 'view_delegate' and
    * the 'trigger' actors, while the 'stored_function' and 'stored_procedure' actors
    * allows direct ACL to simulate the MySQL privileges table 'procs_priv'.
    * The generic firing order on 'AclRights' and 'EPsThroughAclEntry' is inversed
    * in this method.
    * The four types of host groups can be used as ACL sources only for
    * the databases, which are not ActorMySQL.
    * Calls the super.addAclEntry() method to do it, and BaseUtilityMySQLImpl.addInSortedAclList().
    * @param _aclAcs ImmutableACS which owns the new AclEntry, and to be recorded
    *  as an ACS controller of this resource if it is not its main ACS.
    * @param _ep eligible party connected to the resource through a new aclEntry.
    * @param _l_right is an array of positive or negative rights of _ep on this resource
    *  (that is, if READ is a right, _ep can read on the resource). May be null.
    * @param _positiveRight is true if the rights are set to authorize (grant),
    *  and false if they are set to forbid (deny).
    * @param _conditionGroup first condition group. May be null.
    * @exception InternalError if this resource has a null name or a null ACS,
    *  or if _ep is not an EligiblePartyImpl nor an ActorImpl.
    * @exception CreateError from the super method or not
    * @return the new AclEntry.
    */
  public synchronized AclEntry addAclEntry(ImmutableACS _aclAcs,
    ImmutableEligibleParty _ep, StringRight[] _l_right, Boolean _positiveRight,
    ImmutableGroupID _conditionGroup) throws CreateError {

        PropertyChangeListener[]    l_listeners = null;
        AclEntry[]                  l_oldAclEntry = null;
        AclEntry                    acl = null;
        ArrayList<AclEntry>         l_acl = null;

        if (type_.indexOf("user socket") == -1) {
            if (_conditionGroup != null && _ep.getEPType().indexOf("account") == -1)
                throw new CreateError(" A MySQL routine allows conditional ACL "+
                        "only to accounts.");
            else if (_conditionGroup == null &&
                    _ep.getEPType().indexOf("pattern-name_accounts-for-hosts group") == -1)
                throw new CreateError(" A MySQL routine allows non-conditional ACL"+
                        "only to accounts-for-hosts groups.");
        } // if

        // stop the property change firing on 'AclRights', and not for 'EPsThroughAclEntry'
        l_listeners = getPropertyChangeListeners("AclRights");

        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                removePropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if

        l_oldAclEntry = getL_aclEntry();

        // firing on 'EPsThroughAclEntry'
        try {
            acl = super.addAclEntry(_aclAcs, _ep, _l_right, _positiveRight, _conditionGroup);
        } catch (CreateError c) {
            // restore the property change firing on 'AclRights'
            if (l_listeners != null) {
                for (int i = 0; i < l_listeners.length; i++) {
                    addPropertyChangeListener("AclRights", l_listeners[i]);
                } // for i
            } // if
            l_listeners = null;
            l_oldAclEntry = null;
            throw c;
        } // catch

        l_acl = new ArrayList(Arrays.asList(l_aclEntry_));

        // The new AclEntry has been added at the end of the current list, but no use here
        l_acl.remove(acl);
        l_acl = BaseUtilityMySQLImpl.addInSortedAclList(l_acl, acl);
        l_aclEntry_ = l_acl.toArray(new AclEntry[0]);

        // restore the property change firing on 'AclRights'
        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                addPropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if
 
        // add the primary rights directly
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                null,
                m_inheritedT_ACLRights_,
                true);

        // add to 'MySQL server' an ACL with the USAGE right, if necessary
        boolean found = false;
        DirectoryMySQLImpl server = null;
        if (l_inheritedT_AclEntries_ != null) {
            // no direct reading of server.l_aclEntry_
            for (AclEntry acl2 : l_inheritedT_AclEntries_) {

                if (!acl2.getResType().equals("MySQL server <byAcsCreator>"))
                   break;
                
                if (acl2.getEorEP() != _ep)
                    continue;

                if (acl2.getL_ConditionSourceGroups().isEmpty() &&
                        _conditionGroup == null) {
                    found = true;
                    break;
                }
                else if (acl2.getL_ConditionSourceGroups().isEmpty() &&
                        _conditionGroup != null) {
                    continue;
                }
                else if (acl2.getL_ConditionSourceGroups().get(0) ==
                            _conditionGroup) {
                    found = true;
                    break;
                } // if
            } // for i
        } // if

        if (!found) {
            server = (DirectoryMySQLImpl) aCS_.getEorM_Resources().get(
                        ((Name)aCS_Name_.clone()).extended("MySQL server"));

            // calls back addInheritedAclEntry() on all the 'MySQL server' children,
            // including this instance.
            server.addAclEntry(_aclAcs, _ep,
                    new StringRight[] {aCS_.getOneRight("USAGE")},
                    _positiveRight, _conditionGroup);
        } // if

        // the ACL ACS and the ACL source are already informed
        try {
            firePropertyChange("AclRights", l_oldAclEntry, getL_aclEntry());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorMySQLImpl.addAclEntry(), firePropertChange() throws:");
            e.printStackTrace();
        }

        l_listeners = null;
        l_oldAclEntry = null;
        l_acl = null;
        server = null;
        return acl;
    } // addAclEntry()

    /**
    * Removes the AclEntry and the primary rights.
    * Calls the overridden method first.
    * Fires 'InheritedTargetAclEntries' events only to update the display
    * of the primary rights map.
    * The detailed sequence of calls for adding is described at the method addAclEntry()
    * of DirectoryMySQLImpl.
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public synchronized void removeAclEntry(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

                                          // for testing and tracking
                                          //System.out.println(
                                            //  "start of ActorMySQLImpl.removeAclEntry() in: "+
                                              //name_+" for _ep "+_ep.getNickName()+
                                              //", _positiveRight "+_positiveRight+
                                              //", _conditionGroup "+_conditionGroup.getNickName());

        // firing on 'AclRights' and 'EPsThroughAclEntry'
        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);
        
        // removes the primary rights
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                null,
                m_inheritedT_ACLRights_,
                true);
        
                                      // for testing and tracking
                                      //System.out.println("ActorMySQLImpl.removeAclEntry() in: "+
                                        //      name_+" fires the 'InheritedTargetAclEntries' event");

        // to update the display of the primary rights map
        try {
            firePropertyChange(
                    "InheritedTargetAclEntries", null, l_inheritedT_AclEntries_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From DirectoryMySQLImpl.removeAclEntry(), "+
                    "firePropertChange() throws");
            e.printStackTrace();
        } // catch


                          // for testing and tracking
                  //System.out.println("at the end of ActorMySQLImpl.removeAclEntry(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}
  } // removeAclEntry()

    /**
    * Adds the inherited AclEntry in the good sorting. Called by DirectoryMySQLImpl.addAclEntry()
    * and the same method, and by the constructor. An inherited AclEntry may have
    * a Resource which is not this instance.
    * Copied from ResourceMySQLImpl.
    * Calls getEorL_TargetInheritedAclEntries(), BaseUtilityMySQLImpl.addInSortedInheritedAclList()
    * and ResourceImpl.setEorL_TargetInheritedAclEntries() that fires 'InheritedTargetAclEntries'.
    * Not described in the BeanInfo since the user does not call it.
    * The detailed sequence of calls is described at DirectoryMySQLImpl.addAclEntry(). <P>
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _acl inherited AclEntry to put in the list at the right order
    */
  public synchronized void addInheritedAclEntry(AclEntry _acl) {
    ArrayList<AclEntry> l_acl = null;

    l_acl = getEorL_TargetInheritedAclEntries();

                      // for testing and tracking
                      //System.out.println("ActorMySQLImpl.addInheritedAclEntry() for "+
                        //     name_+", call to ResourceMySQLImpl.addInSortedAclList()"+
                          //   "for the new acl "+_acl+", with current AclEntries: ");
                      //for (Object obj : l_acl) {
                        //  System.out.println("----- "+obj);
                      //}

    l_acl = BaseUtilityMySQLImpl.addInSortedInheritedAclList(l_acl, _acl);

    setEorL_TargetInheritedAclEntries(l_acl);

    l_acl = null;
  } // addInheritedAclEntry()

    /**
    * Removes the inherited AclEntries in the children.
    * Copied in DirectoryMySQLImpl.
    * Calls getEorL_TargetInheritedAclEntries() and
    * ResourceImpl.setEorL_TargetInheritedAclEntries() that fires 'InheritedTargetAclEntries'.
    * Not described in the BeanInfo since the user does not call it.
    * The detailed sequence of calls is described at the method DirectoryMySQLImpl.addAclEntry(). <P>
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _acl inherited AclEntry
    * @exception UpDateError if _acl is null or unknown
    */
  public synchronized void removeInheritedAclEntry(AclEntry _acl) throws UpDateError {

                      // for testing and tracking
                      //System.out.println("Call to ActorMySQLImpl.removeInheritedAclEntry() in: "+
                        //      name_+" for the ACL "+_acl);

      ArrayList<AclEntry>   l_acl = getEorL_TargetInheritedAclEntries();
      AclEntry              acl = null;
      boolean               found = false;

      for (Iterator ite = l_acl.iterator(); ite.hasNext(); ) {
        acl = (AclEntry)ite.next();
        if (acl != null && !acl.isEmpty()) {
          if (acl.equals(_acl)) {
              ite.remove();
              found = true;
              break;
          }
        } // if
        //else
          //  ite.remove();
      } // for ite

      if (!found)
          throw new UpDateError(" The AclEntry is unknown "+_acl);
      else {
                      // for testing and tracking
                      //System.out.println("ActorMySQLImpl.removeInheritedAclEntry() in: "+
                        //      name_+" removes the ACL "+acl);
      } // else

      setEorL_TargetInheritedAclEntries(l_acl);

      l_acl = null;
      acl = null;
  } // removeInheritedAclEntry()

    /**
    * Returns the AclEntry to use as inherited AclEntry in a child.
    * Returns null if the type contains 'index', 'directory', or starts with 'byAcsCreator'.
    * Otherwise, returns the argument.
    * Called by the constructors. 
    * Overrides the super method and does not call it.
    * Calls the ResourceMySQL static method having the same name.
    * @param _acl the initial AclEntry which may be given to the child
    * @param _type type of a child, as Resource
    * @return the same AclEntry if the child inherits this initial AclEntry, or null
    */
  public AclEntry selectInheritedAclEntry(AclEntry _acl, String _type) {
    return ResourceMySQLImpl.selectInheritedAclEntry(this, _acl, _type);
  } // selectInheritedAclEntry()

    /**
    * Resets the inherited AclEntry rights and the primary rights after an user edition.
    * Called by the parent when there is a change in the effective rights of
    * an inherited AclEntry. Derived from setM_TargetInheritedACLRights().
    * Called by Directory.resetInheritedAclEntryRightsOfChildren().
    * Overrides the super method to update the two inherited ACL maps,
    * and does not call it.
    * Calls setM_TargetInheritedACLRights(), since it is not possible to reset
    * the two inherited rights maps without analyzing all the inherited AclEntries.
    * Fires the 'InheritedTargetAclEntries' event.
    * @param _acl inherited AclEntry with updated rights. Never null.
    */
  public synchronized void resetInheritedAclEntryRights(AclEntry _acl) {
      
                                // for testing and tracking
                                //if (l_inheritedT_AclEntries_ == null)
                                  //  System.out.println(
                                    //"in ActorMySQLImpl.resetInheritedAclEntryRights(), for "+
                                    //name_+", l_inheritedT_AclEntries_ is null.");

    setM_TargetInheritedACLRights(l_inheritedT_AclEntries_);

    // to update the display of the primary rights map
    try {
        firePropertyChange(
                "InheritedTargetAclEntries", null, l_inheritedT_AclEntries_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ActorMySQLImpl.resetInheritedAclEntryRights(), "+
                "firePropertChange() throws");
        e.printStackTrace();
    } // catch
  } // resetInheritedAclEntryRights()

  /**
   * Gets the internal map of the primary rights from each parent level of this resource.
   * The primary rights are displayed into the beamer. They help the user
   * to understand the current users, since it covers the root 'MySQL server' rights,
   * and to understand the complex database rights. Because this method overrides
   * the Resource method, its name is 'TargetInheritedACL' rather than 'TargetPrimaryACL'.
   * Overrides the Resource method to describe its use by the MySQL addon.
   * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl#getM_TargetInheritedACLRights()
   * @return unmodifiable map of primary rights from the direct and inherited AclEntries.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedACLRights() {
      if (m_inheritedT_ACLRights_ == null) return Collections.EMPTY_MAP;
      else {
          return Collections.unmodifiableMap(m_inheritedT_ACLRights_);
      } // else;
  }

    /**
    * Removes the direct AclEntry, but without event firing. The purpose of this method 
    * is to avoid an unuseful sequence of firings when a lot of AclEntries are removed. 
    * It is called from the root 'MySQL server' method removeAclEntry() when it deletes 
    * all the direct ACL of a pair on the root children when the root ACL for this pair is deleted.
    * Calls the overridden method, and disables before all the firings for the 'AclRights'
    * and 'EPsThroughAclEntry' events.
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public void removeAclEntryWithoutFiring(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {
      
        PropertyChangeListener[]    l_listeners = null;
        PropertyChangeListener[]    l_listeners2 = null;

                                      // for testing and tracking
                                      //System.out.println(
                                        //      "in ActorMySQLImpl.removeAclEntryWithoutFiring(),"+
                                          //    "call to the super method in "+
                                            //  name_);

        // 'EPsThroughAclEntry' and 'AclRights' are bound properties in ResourceImpl
        // do unregister before the call to super.removeAclEntry()
        l_listeners = getPropertyChangeListeners("AclRights");
        l_listeners2 = getPropertyChangeListeners("EPsThroughAclEntry");

        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                removePropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                removePropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);

        // restore the property change firings
        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                addPropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                addPropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

        // removes the primary rights
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                null,
                m_inheritedT_ACLRights_,
                true);

                          // for testing and tracking
                  //System.out.println("at the end of ActorMySQLImpl.removeAclEntryWithoutFiring(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}

        l_listeners = null;
        l_listeners2 = null;
  } // removeAclEntryWithoutFiring()

 /**
  * Clones without registering of the new object in the ACSs.
   * Caution: don't use any updating method on this clone,
   * because that may destroy the base integrity. A finalized instance returns
   * a clone with getEorACS() which returns null,
   * but getName() and getAcsName() operates.
   * @exception CloneNotSupportedException
   * @return Object; is not registred in the ACSs. The names, the aclEntries,
   * the AG rights, the ACS controller names are cloned.
   * The AclEntries are weakly cloned, so they have no direct link to this resource.
   * The external object references are not cloned
   */
  public synchronized Object clone() {

      ResourceMySQLImpl clon = (ResourceMySQLImpl) super.clone();

      return clon;
  } // clone()

   /**
   * This method sets to null the primary inherited rights map.
   * Overrides the super method and calls it at the end.
   * Note: when this method is called by the ResourceImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the main ACS do not known the resource (if
   * it is a clone without setName(), typically)
   */
  public synchronized void finalizeForBase() throws UpDateError {

                    // for testing and tracking
                //System.out.println("In ResourceMySQLImpl, call of " + name_ + ".finalizeForBase()");

    if (folderMember_ == null) return;

    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * This method provides a copy of the instance
  * with another name, following Copyable.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Each aclEntry is copied, but not the list of ViewInBase
  * that contains this instance. The bridge target is copied if
  * this instance is a bridge source.
  * Caution: do not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Calls copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually created as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
     May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ActorMySQLImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ||
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError(
          " The name or the acs is null,"+
          " or the parent ACS is not the instance ACS.");
      }

      ActorMySQLImpl result = new ActorMySQLImpl(
            _name, _acs, _parent, (UserIDMySQL)currentUserID_,
            getResourceType(), (GroupIDMySQL)currentGroupID_);

      result = copyOwnVariables(result, false);

      return result;
  } // copy(,,,)


  /**
  * This method provides a copy of the instance
  * as copy() but without copying the internal AclEntries.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Each aclEntry is copied, but not the list of ViewInBase
  * that contains this instance. The bridge target is copied if
  * this instance is a bridge source.
  * Caution: does not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Caution: does not copy the bridge sources, if they exists.
  * Calls copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually created as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
     May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ActorMySQLImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ||
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError(" in ActorMySQLImpl.copy(),"+
          " _name or _acs is null,"+
          " or _parent ACS is not _acs.");
      }

      ActorMySQLImpl result = new ActorMySQLImpl(
            _name, _acs, _parent, (UserIDMySQL)currentUserID_,
            getResourceType(), (GroupIDMySQL)currentGroupID_);

      // copies the external AclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclEntry copy

      //if (result.getEorL_AclConnectedResources().isEmpty()) return result;
      //else return EligiblePartyImpl.deleteAclEntries(result);
      return result;
  } // copyWithoutAcl(,,,)

  //
  // -------------- protected methods as ResourceMySQLImpl copies -----------------------
  //
  
  /**
  * This method provides a copy of the instance variables.
  * Copies the AclEntries.
  * Caution: copies the current user and group, and the secondary groups,
  * if it is the same ACS,
  * or else in every UserID and GroupID copy which is found.
  * In this second case, the user or group name must be made
  * of the acs name ended by the user or group string name.
  * Caution: does not copy the bridge sources, if they exist, nor the actor alias.
  * Uses the super overridden method with one argument.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result ActorMySQLImpl to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError
  **/
  protected ActorMySQLImpl copyOwnVariables(ActorMySQLImpl _result, boolean _withoutInternalAcl)
    throws CreateError {
      
      return (ActorMySQLImpl)super.copyOwnVariables(_result, _withoutInternalAcl);
  } // copyOwnVariables(,,,)

  /**
   * NO USE.
   * This method sets to null the primary inherited rights map, since it is used
   * after a deserialization.
   * Overrides the super method and calls it at the end.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In ResourceMySQLImpl, call of " + name_ +
                //".finalizeFromDeserialization()");

    if (folderMember_ == null) return;

    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

   /**
   * Sets all the inherited AclEntries for this target. The AclEntry may be a clone.
   * To use by the parent class in the gBase package.
   * No use in this generic gBase package, so it is reserved to the AcsAddon packages.
   * Calls setM_TargetInheritedACLRights() to set the map (source name + order, StringRight[]).
   * Fires 'InheritedTargetAclEntries' events.
   * The detailed sequence of calls for adding is described at the method
   * DirectoryMySQLImpl.addAclEntry(). <P>
   * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
   * @param _l_acl list of AclEntries. May be empty, but never null.
   */
  protected void setEorL_TargetInheritedAclEntries(ArrayList<AclEntry> _l_acl) {
      if (_l_acl== null) throw new InternalError("ActorMySQLImpl."+
              "setEorL_TargetInheritedAclEntries() with a null argument.");

      ArrayList l_oldList = null;

      if (l_inheritedT_AclEntries_ == null) {
          l_oldList = new ArrayList(INITIAL_CAPACITY_2);
      } else {
          l_oldList = (ArrayList) l_inheritedT_AclEntries_.clone();
      } // else

      l_inheritedT_AclEntries_ = _l_acl;

      setM_TargetInheritedACLRights(l_inheritedT_AclEntries_);
      
      try {
        firePropertyChange(
                "InheritedTargetAclEntries", l_oldList, l_inheritedT_AclEntries_.clone());
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ActorMySQLImpl.setEorL_TargetInheritedAclEntries(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
      } // catch

      l_oldList = null;
  } // setEorL_TargetInheritedAclEntries()

  /**
   * Sets the internal maps of the inherited AclEntry rights and the primary rights.
   * The map value is a non-null StringRight array of inherited rights, and it may be empty.
   * Called by setEorL_TargetInheritedAclEntries() only, and do not fire any event
   * while the caller method fires 'InheritedTargetAclEntries' events.
   * Calls the BaseUtilityMySQL static method having the same name.
   * The detailed sequence of calls for adding is described at the method
   * DirectoryMySQLImpl.addAclEntry(). <P>
   * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
   * @param _l_acl current list of inherited AclEntries.
   * May be empty, but never null.
   */
  protected void setM_TargetInheritedACLRights(List<AclEntry> _l_acl) {
      
                                      // for testing and tracking
                                      //System.out.println("Call to ActorMySQLImpl."+
                                        //      "setM_TargetInheritedACLRights() in: "+
                                          //    name_);

      if (m_inheritedT_ACLRights_ == null)
          m_inheritedT_ACLRights_ = new TreeMap();

      BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                _l_acl,
                null,
                m_inheritedT_ACLRights_,
                false);
  } // setM_TargetInheritedACLRights()

  //
  // -------- protected overridden methods from ActorImpl --------------------

  /**
   * This method is rejected because the bridge may be from any ACS to simulate
   * the MySQL client bugs and failures.
  * Adds an actor as source to this instance which is then a bridge target for this source.
  * Sets the bridge source for a 'connection', since this is the only case.
  * Checks if the bridge source is not from the good ACS, which matches to
  * the current Group of the target, and otherwise, fires an InternalError
  * which is converted by ActorImpl in a CreateError for the user.
  * Calls the super method to provide the main work.
  * Invoked by Actor.setEorACSBridgeTarget().
  * Fire an event on "l_bridgeSources".
  * @param _act is a bridge source for this actor as target.
  */
  //protected void addBridgeSource(Actor _act) {
    //  ImmutableName hostName = null;

      //try {
        //  hostName = GroupIDMySQLImpl.getNameFromHostComponent(
          //    currentGroupID_.getNickName());
      //}
      //catch (BaseError b) {
        //  throw new InternalError("ActorMySQLImpl.addBridgeSource() throws "+b);
      //}

      //if (!_act.getAcsName().equals(hostName))
        //  throw new InternalError("This bridge source is not from the good ACS, "+
          //        "which matches to the current Group of the target.");

      //super.addBridgeSource(_act);
  //} // addBridgeSource()

} // ActorMySQLImpl


