//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import ARoad0.gBaseInterface.*;
import ARoad0.gBase.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.Pattern.Singleton;

import ARoad0.AcsAddon.Accbee.Ubuntu.gBase.ActorUbuntuImpl;

/**
* This class is an utility class for the creation of the standard ACS
* which are configured by code, like the RBAC application.
* It uses the class ARoad0.gBase.ACSFactoryImpl for creating the ACS
* in the base. It offers only public static methods.
* This class has several subclasses in the AcsAddon packages, and they implements
* the interfaces Pattern.Singleton and Pattern.AcsFactoryForAcsAddon.
* this class has no variables.
* @version 0.7.1
*/
public class AcsFactory implements Singleton {

    /* Constructors
     */

  /**
   * with this public constructor for the subclasses.
  */
  public AcsFactory() {
  }

 
  /**
   * Creates an ACS with the type Role-Based Access Control application.
   * Calls configureRBAConLinux().
   * configureRBAC() is called after this method by ActionNewACSyst.fillWithAcsObjects().
   * @param _name ImmutableName for the ACS to create.
   *   No test of null value nor empty.
   * @param _type is 'RBAC application'.
   * @param _core is the core ACS (for instance, the operating system) 
   *   on which the new ACS is build up. May be null.
   * @exception BaseError if the ACS is not known, the ACS is open, the open ACS has not
   *   the same name, the core ACS is null when it is mandatory, or 
   *   ClassNotFoundException, IOException with
   *   invalidClassException if ACSImpl HAS CHANGED,
   *   or if the type or the specified structure are not correct.
   * @return the new ACS
   */
  public static synchronized ACSImpl createRBAC(ImmutableName _name,
          String _type, ACSImpl _core) throws BaseError {
        
        if (!_type.equals("RBAC application")) 
                         throw new BaseError("The ACS type is not correct.");
        if (_core == null) 
            throw new BaseError("The parent ACS is null for the new RBAC application.");
        else if ( _core.getACSType().indexOf("Linux") == -1)
            throw new BaseError("The core ACS is not a Linux for the new RBAC application.");
        
        ACSImpl acs = null;
        ACSFactoryImpl factory = null;
        StringRightImpl adm = null;
        StringRightImpl mdr = null;
        StringRightImpl ctrl = null;
        StringRightImpl dctrl = null;
        StringRightImpl aud = null;
        StringRightImpl daud = null;
        StringRightImpl dw = null;
        StringRightImpl dr = null;
        StringRightImpl w = null;
        StringRightImpl r = null; 
        StringRightImpl co = null;
        StringRightImpl dco = null;
        StringRightImpl lau = null;
        
        String[] l_asso = null;

        // creates the ACS
        try {
            factory = new ACSFactoryImpl(_core.getISName(), _name, _type, 
                    _core.getName());

            // creates the acsrights -------------------------------------------
            adm = new StringRightImpl("administer", 
                    BaseManagerImpl.getInstance().getMetaRight("|is_superadmin|"),
                    990, null, null, factory );
            mdr = new StringRightImpl("right update", 
                    BaseManagerImpl.getInstance().getMetaRight("|modify_security_descriptor|"),
                    900, new StringRightImpl[] {adm}, null, factory );
            ctrl = new StringRightImpl("control", 
                    BaseManagerImpl.getInstance().getMetaRight("full_control"),
                    970, null, null, factory );
            aud = new StringRightImpl("audit update", 
                    BaseManagerImpl.getInstance().getMetaRight("execute"),
                    600, null, null, factory ); 
            w = new StringRightImpl("enter data", 
                    BaseManagerImpl.getInstance().getMetaRight("execute_&_write"),
                    700, new StringRightImpl[] {ctrl}, null, factory );
            r = new StringRightImpl("browse data", 
                    BaseManagerImpl.getInstance().getMetaRight("read_&_execute"),
                    400, new StringRightImpl[] {w}, null, factory );           
            co = new StringRightImpl("connect", 
                    BaseManagerImpl.getInstance().getMetaRight("execute"),
                    100, new StringRightImpl[] {r, aud}, null, factory ); 

            dctrl = new StringRightImpl("deny_control", 
                    BaseManagerImpl.getInstance().getMetaRight("deny_full_control"),
                    970, null, null, factory );
            daud = new StringRightImpl("deny_audit", 
                    BaseManagerImpl.getInstance().getMetaRight("deny_execute"),
                    600, null, null, factory );
            dw = new StringRightImpl("deny_enter data", 
                    BaseManagerImpl.getInstance().getMetaRight("deny_execute_&_write"),
                    700, new StringRightImpl[] {dctrl}, null, factory );
            dr = new StringRightImpl("deny_browse data", 
                    BaseManagerImpl.getInstance().getMetaRight("deny_read_&_execute"),
                    400, new StringRightImpl[] {dw}, null, factory );            
            dco = new StringRightImpl("deny_connect", 
                    BaseManagerImpl.getInstance().getMetaRight("deny_all"),
                    100, new StringRightImpl[] {dr, daud}, null, factory );

            // new in 0.7.1, from 'transaction' to 'to database',
            // and for the external ACL to the Linux roots
            lau = new StringRightImpl("launch",
                    BaseManagerImpl.getInstance().getMetaRight("execute"),
                    800, new StringRightImpl[] {ctrl}, null, factory );

            factory.setManageAcsRights(true);

            // creates the types ---------------------------------------------
            factory.setL_EPTypes(new String[] {
                "personal user account", 
                "<immutable> personal user account",
                "<byAcsCreator>",
                "<byAcsCreator><immutable> ",
                // for GroupID
                "<byAcsCreator> physical",
                "<byAcsCreator> functional",
                "manager group",
                "physical group",
                "functional group",
                "operational role",
                "transient role",
                "operational audit role",
                "security audit role",
                "manager role",
                "administrator role",
                "<immutable> group",
                "<immutable> physical group",
                "<immutable> functional group",                
                "<immutable> operational role",
                "<immutable> transient role",
                "<immutable> operational audit role",
                "<immutable> security audit role",
                "<immutable> manager role",
                "<immutable> administrator role",
                // new in 0.7.1
                "transaction",
                "<immutable> transaction",
                "to database",
                "<immutable> to database"
            });
            
            factory.setL_ResourceTypes(new String[] {
                "<byAcsCreator>",
                "transaction",
                "transaction set",
                "<immutable> transaction",
                "<immutable> transaction set",
                // new in 0.7.1
                "to database",
                "<immutable> to database"});

            // creates the structure -------------------------------------------
            factory.setManageResources(true);
            factory.setManageResourceTrees(true);
            factory.setManageAccounts(true);
            factory.setManageGroups(true);
            factory.setManageAGORights(false);
            factory.setManageActors(true); // new on 0.7.1
            factory.setManageInternalAcls(true);
            factory.setManageExternalAcls(true); // new on 0.7.1
            factory.setManageBridges(false);
            factory.setManageSecGroupsForActor(false);
            factory.setManageGroupTrees(true);

            // creates the SubAcsNodes and creates the relevant types and policy --------
            // (nothing for the key Explorer.SubAcsNode)
            // for Explorer.AcsGroupTree
            factory.addSubAndGroupPolicy("physical",
                    new String[] {"<GroupTree>"});
            factory.addSubAndGroupPolicy("functional",
                    new String[] {"<GroupTree>"});

            // specialized types ------------------------------------------------
            // no 'Explorer.NoAcsNode' key to forbid some nodes in the explorer
            // CreationByBeamer.NoType to forbid creation of these types by the user
            factory.addTypesAndExplorer("CreationByBeamer.NoType",
                    new String[] {
                        // for Resource, only 'transaction' is allowed
                        "ResourceType.<undefined>",
                        "ResourceType.<byAcsCreator>",
                        "ResourceType.transaction set",
                        "ResourceType.<immutable> transaction set",
                            // new in 0.7.1
                        "ResourceType.to database",
                        "ResourceType.<immutable> to database",
                        // for Directory, only 'transaction set' is allowed
                        "DirectoryType.<undefined>",
                        "DirectoryType.<byAcsCreator>",
                        "DirectoryType.transaction",
                        "DirectoryType.<immutable> transaction",
                            // new in 0.7.1
                        "ResourceType.to database",
                        "ResourceType.<immutable> to database",
                        // for UserID, only 'personal user account' is allowed
                        "UserIDType.<byAcsCreator><immutable> ",
                        "UserIDType.<undefined>",
                        "UserIDType.<byAcsCreator>",
                        "UserIDType.<byAcsCreator> physical",
                        "UserIDType.<byAcsCreator> functional",
                        "UserIDType.manager group",
                        "UserIDType.physical group",
                        "UserIDType.functional group",
                        "UserIDType.operational role",
                        "UserIDType.transient role",
                        "UserIDType.operational audit role",
                        "UserIDType.security audit role",
                        "UserIDType.manager role",
                        "UserIDType.administrator role",
                        "UserIDType.<immutable> group",
                        "UserIDType.<immutable> manager group",
                        "UserIDType.<immutable> physical group",
                        "UserIDType.<immutable> functional group",
                        "UserIDType.<immutable> operational role",
                        "UserIDType.<immutable> transient role",
                        "UserIDType.<immutable> operational audit role",
                        "UserIDType.<immutable> security audit role",
                        "UserIDType.<immutable> manager role",
                        "UserIDType.<immutable> administrator role",
                            // new in 0.7.1
                        "UserIDType.to database",
                        "UserIDType.<immutable> to database",
                        // for GroupID, most of the values are authorized
                        "GroupIDType.<byAcsCreator><immutable> ",
                        "GroupIDType.<undefined>",
                        "GroupIDType.<byAcsCreator>",
                        "GroupIDType.<byAcsCreator> physical",
                        "GroupIDType.<byAcsCreator> functional",
                        "GroupIDType.personal user account",
                        "GroupIDType.<immutable> personal user account",
                            // new in 0.7.1
                        "GroupIDType.to database",
                        "GroupIDType.<immutable> to database",
            });

                // see the documentation of
                // ACSFactoryUtilityImpl.getL_LimitedValuesForTypesAndExplorer()
            // GroupID.NoPrimaryGroup,
            // GroupID.MemberFromAdmin, GroupID.NoMoreThanOneMember for the group policy
            factory.addTypesAndExplorer("GroupID.NoPrimaryGroup",
                    new String[] {
                        "<byAcsCreator>",
                        // for GroupID
                        "<byAcsCreator> physical",
                        "<byAcsCreator> functional",
                        "operational role",
                        "manager group",
                        "physical group",
                        "functional group", 
                        "transient role",
                        "operational audit role",
                        "security audit role",
                        "manager role",
                        "administrator role"
            });
            factory.addTypesAndExplorer("GroupID.MemberFromAdmin",
                    new String[] {
                        "administrator role",
            });
            factory.addTypesAndExplorer("GroupID.NoUserIDAsMember",
                    new String[] {
                        "<byAcsCreator> physical",
                        "<byAcsCreator> functional",
                        "functional group",
                        "manager group"
            });
            factory.addTypesAndExplorer("GroupID.NoGroupIDAsMember",
                    new String[] {
                        "operational role",
                        "transient role",
                        "operational audit role",
                        "security audit role",
                        "manager role",
                        "administrator role"
            });
            // no 'GroupID.NoMoreThanOneMember' key 
            // with 'GroupID.TypesOfMemberFor.', each possible group type has a set of members
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "<byAcsCreator> physical",
                    new String[] {
                        "physical group"
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "physical group", 
                    new String[] {
                        "physical group",
                        "personal user account"
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "<byAcsCreator> functional",
                    new String[] {
                        "functional group",
                        "manager group"
            }); 
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "functional group", 
                    new String[] {
                        "functional group",
                        "operational role",
                        "transient role",
                        "operational audit role",
                        "security audit role",
                        "manager role",
                        "administrator role"                       
            }); 
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "manager group", 
                    new String[] {
                        "manager role"                      
            }); 
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "operational role", 
                    new String[] {
                        "personal user account"                      
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "transient role", 
                    new String[] {
                        "personal user account"                      
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "operational audit role", 
                    new String[] {
                        "personal user account"                      
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "security audit role", 
                    new String[] {
                        "personal user account"                      
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "manager role", 
                    new String[] {
                        "personal user account"                      
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "administrator role", 
                    new String[] {
                        "personal user account"                      
            });
            
            factory.addTypesAndExplorer("GroupIDMember.NoMoreThanOneGroup."+
                    "personal user account", 
                    new String[] {
                        "physical group",
                        "operational role",
                        "transient role",
                        "manager role",
                        "administrator role" 
            });
            // in this version, there is no control on the alternative
            // 'operational or transient or ...'            
            factory.addTypesAndExplorer("GroupIDMember.RecommandedGroupsAsMember."+
                    "personal user account", 
                    new String[] {
                        "physical group",
                        "operational role<OR>transient role<OR>manager role<OR>" +
                                "operational audit role<OR>security audit role<OR>"+
                                "administrator role"
            });
            
            // on Directories
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "NULL",
                    new String[] {
                        "<byAcsCreator>",
                            // new in 0.7.1
                        "to database",
                        "<immutable> to database",
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "<byAcsCreator>",
                    new String[] {
                        "transaction set"
            });           
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "transaction set", 
                    new String[] {
                        "transaction"
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "transaction", 
                    new String[0]);

            // creates the rights sets for targets ---------------------------------------------
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.<byAcsCreator>",
                    new StringRight[0]); 
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.transaction", 
                    new StringRight[] {co, adm, aud, mdr, ctrl, w, r, dctrl, daud, dw, dr, dco});           
            // no rights for directories 'transaction set'
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.transaction set", 
                    new StringRight[0]);
                    // new in 0.7.1
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.to database", 
                    new StringRight[] {lau});
            
            // creates the rights sets for sources ------------------
            // put 'deny_control' to have a right not used in any target,
            // so there is on common right, and the result is no right for this type
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> physical",
                    new StringRight[] {dctrl} );
            // all users have 'co' right on 'application' through the node '<functions>'
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> functional",
                    new StringRight[] {co} ); 
            // groups like 'accounting' have 'r' rights
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.functional group", 
                    new StringRight[] {r} );
            // a manager has 'w' on other-group transactions,
            // and 'r' on audit transactions
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.manager group", 
                    new StringRight[] { w, r}); 
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.operational role", 
                    new StringRight[] {w, r, co}); 
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.manager role", 
                    new StringRight[] {ctrl, w, co});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.transient role", 
                    new StringRight[0] ); 
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.administrator role", 
                    new StringRight[] {adm, mdr, co}); 
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.operational audit role", 
                    new StringRight[] {aud, co});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.security audit role", 
                    new StringRight[] {aud, co});
            // to forbid some rights from some sites
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.physical group", 
                    new StringRight[] {dctrl, daud, dw, dr, dco});
                    // new in 0.7.1
            // from transaction to 'to database'
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.transaction",
                    new StringRight[] {lau});
            // external ACL to the Linux targets
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.to database",
                    new StringRight[] {lau});

            // 25 characters per line
            factory.addCommentsOnProperties("All the rights",
                    new String[] {"The RBAC application is ",
                                    "modeled mostly as a tree",
                                    "of transactions ",
                                    "(in the resources tree),",
                                    "a tree of functional",
                                    "groups which contains",
                                    "the application personal",
                                    "users (the userIDs), and",
                                    "a set of AclEntries (ACLs)",
                                    "from the groups to ",
                                    "the transactions.",
                                    "",
                                    "There are specialized rights,",
                                    "without standard rights:",
                                    "- no AG, PRI, BDG rights,",
                                    "- if the target is a transaction,",
                                    "  all the rights are allowed",
                                    "- if the target is a transaction",
                                    "  set, or if it has the type",
                                    "  <byAcsACreator>, there is",
                                    "  no right",
                                    "- most of the rights are",
                                    "  defined with the source as",
                                    "  criteria,",
                                    "  and this implies keys in",
                                    "  the specialized rights",
                                    "- these keys give to",
                                    "  a group or transaction, ",
                                    "  as source, the rights adapted",
                                    "  to its function",
                                    "- for instance, the group",
                                    "  'administrator role' is",
                                    "  the only one to have",
                                    "  the right 'administer'."
                                    });
            
            // associations ------------------------------------------------------
            // no associations to any new object

            // sets the creation of the accounts, groups, directories and resources
            // roots of the RBAC ACS in its parent ACS
            configureRBAConLinux(_name, _core, factory);

                    // for testing and tracking
                    //System.out.println("AcsFactory.createRBAC() "+
                      //"before the ACS creation of "+_name);
            
            // final check and creation
            factory.setTiedToParentACS(true);
            acs = factory.createACS();

        } catch (CreateError c) {

                // for testing and tracking
                throw new BaseError("From AcsFactory.createRBAC(), exception: "+c);                 
        } catch (UpDateError u) {

                // for testing and tracking
                throw new BaseError("From AcsFactory.createRBAC(), exception: "+u);                 
        }

        factory = null;
        mdr = null;
        ctrl = null;
        dctrl = null;
        w = null;
        co = null;
        r = null;
        dco = null;
        adm = null;
        daud = null;
        aud = null;
        dw = null;
        dr = null;

                    // for testing and tracking
                    //System.out.println("AcsFactory.createRBAC() "+
                      //"before the ACS saving of "+_name);
        
        // saves the Linux parent only, since the new ACS is saved in ActionNewACSyst
        BaseManagerImpl.getInstance().saveOpenACS(_core);
        return acs;
  } // createRBAC()

  /**
   * Creates a specific ACS to test the VirtualFolders and the Privileges.
   * This ACS has Resources, Directories, Actors, UserIDs,
   * VirtualFolders for Resources and EligibleParty Resources,
   * TypedPrivileges, internal and external LinkedPrivileges.
   * It has no Groups, AGO rights, AclEntries.
   * configureTestForVFandPRI() is called after this method by
   * ActionNewACSyst.fillWithAcsObjects().
   * @param _iSname the IS of the new ACS. May be null.
   * @param _name of the ACS to create.
   *   No test of null value nor empty.
   * @param _type must be 'An ACS for testing VirtualFolders and Privileges'
   * @param _parent is the parent node of the new ACS in the IS. Never null.
   * @exception BaseError if the ACS is known, the ACS is open,
   *   the core ACS is null when it is mandatory, or
   *   ClassNotFoundException, IOException with
   *   invalidClassException if ACSImpl HAS CHANGED,
   *   or if the type or the specified structure are not correct.
   * @return the new ACS
   */
  public static ACSImpl createTestForVFandPRI(ImmutableName _iSname,
          ImmutableName _name, String _type, ImmutableName _parent) throws BaseError {
        
        if (!_type.equals("An ACS for testing VirtualFolders and Privileges"))
                         throw new BaseError("The ACS type is false.");
      
        ACSImpl acs = null;
        ImmutableNameISTree tree = null;
        ACSFactoryImpl factory = null;
        StringRightImpl acc = null;
        StringRightImpl dacc = null;
        StringRightImpl go = null;
        StringRightImpl dgo = null;

        String[] l_asso = null;

            // creates the ACS
        try {
            factory = new ACSFactoryImpl(_iSname, _name, _type, _parent);

            factory.setIncompleteBehaviorModeling(true);
            factory.setIncompleteStructureModeling(true);

            if (_iSname != null) {
                 tree = (ImmutableNameISTree) BaseManagerImpl.getInstance().getISTree(_iSname);
                 if (tree != null) {
                     // unuseful, since this is the default value
                     factory.setVisibleSpace(NameISTreeImpl.ALL_THE_IS_SPACE);
                     factory.setHubProperty(true);
                     factory.setTiedToParentACS(false);
                     factory.setIsAuthorizationServer(false);
                 } // if
            } // if


            // creates the acsrights -------------------------------------------
            acc = new StringRightImpl("ACCESS",
                    BaseManagerImpl.getInstance().getMetaRight("full_control"),
                    990, null, null, factory );
            dacc = new StringRightImpl("NOACCESS",
                    BaseManagerImpl.getInstance().getMetaRight("deny_full_control"),
                    990, null, null, factory );
            go = new StringRightImpl("GO",
                    BaseManagerImpl.getInstance().getMetaRight("read"),
                    700, new StringRightImpl[] {acc}, null, factory );
            dgo = new StringRightImpl("NOGO",
                    BaseManagerImpl.getInstance().getMetaRight("deny_read"),
                    700, new StringRightImpl[] {dacc}, null, factory );

            factory.setManageAcsRights(true);

            // creates the structure --------------------------------------------
            factory.setManageResources(true);
            factory.setManageResourceTrees(true);
            factory.setManageAccounts(true);
            factory.setManageGroups(false);
            factory.setManageAGORights(false);
            factory.setManageActors(true);
            factory.setManageInternalAcls(false);
            factory.setManageExternalAcls(false);
            factory.setManageBridges(false);
            factory.setManageSecGroupsForActor(false);
            factory.setManageGroupTrees(false);
                // VirtualFolders for Resources
            factory.setManageVirtualFolders(true);
            factory.setManageVirtualFolderRightsPropagationToMembers(true);
            factory.setManageResourceVirtualFolders(true);
            factory.setManageSimpleResourceVirtualFolders(true);
            factory.setManageDirectoryInResourceVirtualFolders(false);
                // VirtualFolders for EPs
            factory.setManageEPVirtualFolders(true);
            factory.setManageActorInEPVirtualFolders(true);
            factory.setManageDirectoryEPinEPVirtualFolders(false);
            // Typed and Linked Privileges
            factory.setManagePrivilegeRights(true);
            factory.setManageGrantingRightPrivileges(true);
            factory.setManageDenyingRightPrivileges(true);
            factory.setManageSimpleTypedPrivileges(true);
            factory.setManageSimpleLinkedPrivileges(true);
            factory.setManagePrivilegesInResources(true);
            factory.setManagePrivilegesInVirtualFolders(true);
            factory.setManageInternalForOneToOneLinkedPrivileges(true);
            factory.setManageExternalForOneToOneLinkedPrivileges(true);


            factory.addCommentsOnProperties("All the rights",
                    new String[] {"There are 4 types of rights:",
                                    "- ACCESS has NOACCESS has",
                                    "  opposite,",
                                    "- GO has NOGO has",
                                    "  opposite, and its",
                                    "  upper right is ACCESS."
                                    });
            

            // creates the types ---------------------------------------------
            factory.setL_EPTypes(new String[] {
                "account_1", "account_2", "account_3",
                "actor_1", "actor_2", "actor_3",
                "<byAcsCreator>"});

            factory.setL_ResourceTypes(new String[] {
                "resource_1",  "resource_2", "resource_3",
                "actor_1",  "actor_2", "actor_3",
                "directory",
                "<byAcsCreator>"});
            
            factory.setL_ResourceVirtualFolderTypes(new String[] {
                "VF__for_simple_resources_1",  "VF__for_simple_resources_2",
                "<byAcsCreator>"});

            factory.setL_EPVirtualFolderTypes(new String[] {
                "VF__for_actors_1", "VF__for_actors_2",
                "<byAcsCreator>"});

            factory.setL_PrivilegeForTypeTypes(new String[] {
                "typed_PRI_for_account_1_resource_1",
                "typed_PRI_for_account_2_resource_2",
                "typed_PRI_for_account_3_resource_3",
                "typed_PRI_for_actor_1_resource_1",
                "typed_PRI_for_actor_1_directory",
                "<byAcsCreator>"});

            factory.setL_PrivilegeForLinksTypes(new String[] {
                "linked_PRI_for_account_1",  "linked_PRI_for_account_2", "linked_PRI_for_account_3",
                "linked_PRI_for_actor_1",  "linked_PRI_for_actor_2", "linked_PRI_for_actor_3",
                "linked_PRI_for_account_1_VF__for_actors_1",
                "linked_PRI_for_account_2_VF__for_actors_2",
                "linked_PRI_for_account_3_resource_3",
                "<byAcsCreator>"});


            // no SubAcsNodes ---------------------------------------------------
            // no AcsGroupTree

            // specialized types ------------------------------------------------
            // no 'Explorer.NoAcsNode' key to forbid some nodes in the explorer
            // no CreationByBeamer.NoType to forbid creation of these types by the user

            factory.addCommentsOnProperties("All the ACS object types",
                    new String[] {"The types are designed to",
                                    "help the construction of",
                                    "structured privileges",
                                    "from Account/Actor to ",
                                    "Resource/Actor/VirtualFolder"
                                    });

            // no GroupID.NoPrimaryGroup,
            // no GroupID.MemberFromAdmin, GroupID.NoMoreThanOneMember for the group policy

            // on Nodes and Resources
            // only 'directory' for effective Directories
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "directory",
                    new String[] {
                        "resource_1",  "resource_2", "resource_3",
                        "actor_1",  "actor_2", "actor_3"
            });             
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "resource_1",
                    new String[0]);
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "resource_2", 
                    new String[0]);
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "resource_3",
                    new String[0]);
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "actor_1",
                    new String[0]);
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "actor_2",
                    new String[0]);
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "actor_3",
                    new String[0]);

            // sets the Privilege restrictions for each Resource type
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "resource_1",
                    new String[] {
                        "typed_PRI_for_account_1_resource_1",
                        "typed_PRI_for_actor_1_resource_1",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "resource_2",
                    new String[] {
                        "typed_PRI_for_account_2_resource_2",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2", 
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "resource_3",
                    new String[] {
                        "typed_PRI_for_account_3_resource_3",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3",
                        "linked_PRI_for_account_3_resource_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "actor_1",
                    new String[] {
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_account_1_VF__for_actors_1",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "actor_2",
                    new String[] {
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_account_2_VF__for_actors_2",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "actor_3",
                    new String[] {
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "directory",
                    new String[] {
                        "typed_PRI_for_actor_1_directory"
            });

            // sets the Privilege restrictions for each VirtualFolder type
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_simple_resources_1",
                    new String[] {
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_simple_resources_2",
                    new String[] {
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_actors_1",
                    new String[] {
                        "linked_PRI_for_account_1_VF__for_actors_1"
            });
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_actors_2",
                    new String[] {
                        "linked_PRI_for_account_2_VF__for_actors_2"
            });

            // sets the Privilege restrictions for each EligibleParty type
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "account_1",
                    new String[] {
                        "typed_PRI_for_account_1_resource_1",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_1_VF__for_actors_1"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "account_2",
                    new String[] {
                        "typed_PRI_for_account_2_resource_2",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_2_VF__for_actors_2"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "account_3",
                    new String[] {
                        "typed_PRI_for_account_3_resource_3",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_account_3_resource_3"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "actor_1",
                    new String[] {
                        "typed_PRI_for_actor_1_resource_1",
                        "typed_PRI_for_actor_1_directory",
                        "linked_PRI_for_actor_1"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "actor_2",
                    new String[] {
                        "linked_PRI_for_actor_2"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "actor_3",
                    new String[] {
                        "linked_PRI_for_actor_3"
            });

            // no associations ------------------------------------------------------

            // creates the rights sets ---------------------------------------------

            factory.addPrivilegeRights(
                    "MainType.TYPED",
                    new StringRight[] {acc, dacc, go, dgo});
            factory.addPrivilegeRights(
                    "MainType.LINKED",
                    new StringRight[] {acc, dacc, go, dgo});

            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_account_1_resource_1",
                    new StringRight[] {acc, go});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_account_2_resource_2", 
                    new StringRight[] {acc, go});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_account_3_resource_3", 
                    new StringRight[] {go, dgo});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_actor_1_resource_1",
                    new StringRight[] {acc, go, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_actor_1_directory",
                    new StringRight[] {acc, go, dacc, dgo});

            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_1",
                    new StringRight[] {acc, go});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_2",
                    new StringRight[] {acc, go});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_3",
                    new StringRight[] {acc, go});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_actor_1",
                    new StringRight[] {acc, go, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_actor_2",
                    new StringRight[] {acc, go, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_actor_3",
                    new StringRight[] {acc, go, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_1_VF__for_actors_1",
                    new StringRight[] {acc, go, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_2_VF__for_actors_2",
                    new StringRight[] {acc, go, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_3_resource_3",
                    new StringRight[] {acc, go, dacc, dgo});


            factory.addCommentsOnProperties("Specialized rights",
                    new String[] {"Simple Typed privileges have ",
                                    "(acc, go) rights for ",
                                    "account/resource 1 and 2, ",
                                    "(go, dgo) rights for ",
                                    "account/resource 3, ",
                                    "(acc, go, dacc, dgo) rights ",
                                    "for actor 1 on resource_1 ",
                                    "and directory, ",
                                    "and simple Linked privileges",
                                    "have all rights but not ",
                                    "for account_1 which is limited ",
                                    "to (acc, go)."
                                    });

            // final check and creation
            acs = factory.createACS();
            
        } catch (CreateError c) {

                // for testing and tracking
                c.printStackTrace();
                throw new BaseError("From AcsFactory.createTestForVFandPRI(), "+
                        "exception: "+c);                 
        } catch (UpDateError u) {

                // for testing and tracking
                u.printStackTrace();
                throw new BaseError("From AcsFactory.createTestForVFandPRI(), "+
                        "exception: "+u);                 
        }

 
                    // for testing and tracking
                    //System.out.println("in AcsFactory.createTestForVFandPRI(), "+
                      //"end of the method");

        factory = null;
        acc = null;
        dacc = null;
        go = null;
        dgo = null;

        return acs;
  } // createTestForVFandPRI()


  /**
   * Creates a specific ACS to test all the options for the structure, the types,
   * the constraints in an ACS.
   * This ACS has Resources, Directories, Actors, UserIDs, GroupIDs, tree of GroupIDs,
   * VirtualFolders for Resources and EP Resources. It has TypedPrivileges and LinkedPrivileges,
   * AGO rights without 'other' rights, internal and external AclEntries, Bridges.
   * It has no Resource SubACS, and one Group SubACS.
   * configureTestOfAll() is called after this method by
   * ActionNewACSyst.fillWithAcsObjects().
   * @param _iSname the IS of the new ACS. May be null.
   * @param _name of the ACS to create.
   *   No test of null value nor empty.
   * @param _type must be 'An ACS for testing AG/ACL/PRI rights and AG inheritance'
   * @param _parent is the parent node of the new ACS in the IS. Never null.
   * @exception BaseError if the ACS is known, the ACS is open,
   *   the core ACS is null when it is mandatory, or
   *   ClassNotFoundException, IOException with
   *   invalidClassException if ACSImpl HAS CHANGED,
   *   or if the type or the specified structure are not correct.
   * @return the new ACS
   */
  public static ACSImpl createTestOfAll(ImmutableName _iSname,
          ImmutableName _name, String _type, ImmutableName _parent) throws BaseError {

        if (!_type.equals("An ACS for testing AG/ACL/PRI rights and AG inheritance"))
                         throw new BaseError("The ACS type is false.");

        ACSImpl acs = null;
        ACSFactoryImpl  factory = null;
        StringRightImpl acc = null;
        StringRightImpl dacc = null;
        StringRightImpl activ = null;
        StringRightImpl dactiv = null;
        StringRightImpl go = null;
        StringRightImpl dgo = null;
        StringRightImpl doifcas = null;
        StringRightImpl doifc = null;
        StringRightImpl doifas = null;
        StringRightImpl doi = null;
        StringRightImpl ddoi = null;

        String[]        l_asso = null;

            // creates the ACS
        try {
            factory = new ACSFactoryImpl(_iSname, _name, _type, _parent);

            // creates the ACS rights -------------------------------------------
                // see BaseUtilityImpl.withAcrossToDirectoryAGORights():
                // should contain at least one equivallent of the metarights
                // 'gotochild', 'execute_for_nxdirectory', 'read_for_directory' or 'full_control'

            acc = new StringRightImpl("ACCESS",
                    BaseManagerImpl.getInstance().getMetaRight("full_control"),
                    990, null, null, factory );
            dacc = new StringRightImpl("NOACCESS",
                    BaseManagerImpl.getInstance().getMetaRight("deny_full_control"),
                    990, null, null, factory );
            activ = new StringRightImpl("MODIFY",
                    BaseManagerImpl.getInstance().getMetaRight("write"),
                    700, new StringRightImpl[] {acc}, null, factory );
            dactiv = new StringRightImpl("NOMODIFY",
                    BaseManagerImpl.getInstance().getMetaRight("deny_write"),
                    700, new StringRightImpl[] {dacc}, null, factory );
            activ = new StringRightImpl("ACTIVATE",
                    BaseManagerImpl.getInstance().getMetaRight("execute"),
                    600, new StringRightImpl[] {acc}, null, factory );
            dactiv = new StringRightImpl("NOACTIVATE",
                    BaseManagerImpl.getInstance().getMetaRight("deny_execute"),
                    600, new StringRightImpl[] {dacc}, null, factory );
            go = new StringRightImpl("GO",
                    BaseManagerImpl.getInstance().getMetaRight("read"),
                    400, new StringRightImpl[] {acc}, null, factory );
            dgo = new StringRightImpl("NOGO",
                    BaseManagerImpl.getInstance().getMetaRight("deny_read"),
                    400, new StringRightImpl[] {dacc}, null, factory );

            doifcas = new StringRightImpl("authorize<IF><Console><IF><ActiveSession>",
                    BaseManagerImpl.getInstance().getMetaRight("do_action_one"),
                    100, null, null, factory );
            doifc = new StringRightImpl("authorize<IF><Console>",
                    BaseManagerImpl.getInstance().getMetaRight("do_action_one"),
                    100, null, null, factory );
            doifas = new StringRightImpl("authorize<IF><ActiveSession>",
                    BaseManagerImpl.getInstance().getMetaRight("do_action_one"),
                    100, null, null, factory );
            doi = new StringRightImpl("authorize",
                    BaseManagerImpl.getInstance().getMetaRight("do_action_one"),
                    100, null, null, factory );
            // 'deny_authorize' will delete all the 'authorize' rights
            ddoi = new StringRightImpl("deny_authorize",
                    BaseManagerImpl.getInstance().getMetaRight("deny_do_action_one"),
                    100, null, null, factory );

            factory.setManageAcsRights(true);

            factory.addCommentsOnProperties("All the rights",
                    new String[] {"There are 17 types of rights:",
                                    "- ACCESS is a 'full_control'",
                                    "  right; it has NOACCESS has",
                                    "  opposite,",
                                    "- ACTIVATE is an 'execute'",
                                    "  right; it has NOACTIVATE",
                                    "  has opposite,",
                                    "- GO has NOGO has",
                                    "  opposite, and its",
                                    "  upper right is ACCESS,",
                                    "- 'authorize', its opposite",
                                    "  and its 3 derived rights are",
                                    "  for the ACL only."
                                    });

            // creates the types ---------------------------------------------
            factory.setL_EPTypes(new String[] {
                "account_1", "account_2", "account_3",
                "<immutable> account_1", "<immutable> account_2", "<immutable> account_3",
                "actor_1", "actor_2", "actor_3",
                "<immutable> actor_1", "<immutable> actor_2", "<immutable> actor_3",
                "<byAcsCreator>", "<byAcsCreator><immutable> ",
                "account", "<immutable> account",
                "group", "<immutable> group",
                "personal account", "<immutable> personal account",
                "personal group", "<immutable> personal group",
                // for Actor
                "link", "<immutable> link",
                "executable",  "<immutable> executable",
                "script shell", "<immutable> script shell",
                "<byAcsCreator> authorization",
                "<byAcsCreator> authorization for active session",
                "<byAcsCreator> authorization for admin active session"});

            factory.setL_ResourceTypes(new String[] {
                "<immutable> resource_1",  "<immutable> resource_2", "<immutable> resource_3",
                "resource_1",  "resource_2", "resource_3",
                "<immutable> actor_1",  "<immutable> actor_2", "<immutable> actor_3",
                "actor_1",  "actor_2", "actor_3",
                "<byAcsCreator>",  "<byAcsCreator><immutable> ",
                "file", "<immutable> file",
                "directory", "<immutable> directory",
                "shared memory", "<immutable> shared memory",
                "fifo", "<immutable> fifo",
                // for Actor
                "link", "<immutable> link",
                "socket", "<immutable> socket",
                "executable",  "<immutable> executable",
                "script shell", "<immutable> script shell",
                "<byAcsCreator> authorization"});

            factory.setL_ResourceVirtualFolderTypes(new String[] {
                "VF__for_simple_resources_1",  "VF__for_simple_resources_2",
                "<byAcsCreator>"});

            factory.setL_EPVirtualFolderTypes(new String[] {
                "VF__for_actors_1", "VF__for_actors_2",
                "<byAcsCreator>"});

            factory.setL_PrivilegeForTypeTypes(new String[] {
                "typed_PRI_for_account_1_resource_1",
                "typed_PRI_for_account_2_resource_2",
                "typed_PRI_for_account_3_resource_3",
                "typed_PRI_for_actor_1_resource_1",
                "typed_PRI_for_actor_1_directory",
                "<byAcsCreator>"});

            factory.setL_PrivilegeForLinksTypes(new String[] {
                "linked_PRI_for_account_1",  "linked_PRI_for_account_2", "linked_PRI_for_account_3",
                "linked_PRI_for_actor_1",  "linked_PRI_for_actor_2", "linked_PRI_for_actor_3",
                "linked_PRI_for_account_1_VF__for_actors_1",
                "linked_PRI_for_account_2_VF__for_actors_2",
                "linked_PRI_for_account_3_resource_3",
                "<byAcsCreator>"});

            // creates the structure --------------------------------------------
            factory.setManageResources(true);
            factory.setManageResourceTrees(true);
            factory.setManageAccounts(true);
            factory.setManageGroups(true);
            factory.setManageAGORights(true);
            factory.setManageActors(true);
            factory.setManageInternalAcls(true);
            factory.setManageExternalAcls(true);
            factory.setManageBridges(true);
            factory.setManageSecGroupsForActor(true);
            factory.setManageGroupTrees(true);
            factory.setManageAGORightsInheritance(true);
                // VirtualFolders for Resources
            factory.setManageVirtualFolders(true);
            factory.setManageVirtualFolderRightsPropagationToMembers(true);
            factory.setManageResourceVirtualFolders(true);
            factory.setManageSimpleResourceVirtualFolders(true);
            factory.setManageDirectoryInResourceVirtualFolders(false);
                // VirtualFolders for EPs
            factory.setManageEPVirtualFolders(true);
            factory.setManageActorInEPVirtualFolders(true);
            factory.setManageDirectoryEPinEPVirtualFolders(false);
            // Typed and Linked Privileges
            factory.setManagePrivilegeRights(true);
            factory.setManageGrantingRightPrivileges(true);
            factory.setManageDenyingRightPrivileges(true);
            factory.setManageSimpleTypedPrivileges(true);
            factory.setManageSimpleLinkedPrivileges(true);
            factory.setManagePrivilegesInResources(true);


            // creates the SubAcsNodes and creates the relevant types and policy
            factory.addSubAndGroupPolicy("authorization",
                    new String[] {"<GroupSubACS>", "<ResourceSubACS>"});

            // specialized types ------------------------------------------------
            // no 'Explorer.NoAcsNode' key to forbid some nodes in the explorer
            // CreationByBeamer.NoType to forbid creation of these types by the user
            factory.addTypesAndExplorer("CreationByBeamer.NoType",
                    new String[] {
                        "ResourceType.<byAcsCreator>",
                        "ResourceType.<byAcsCreator><immutable> ",
                        "ResourceType.<byAcsCreator> authorization",
                        // for Directory, only 'directory' & all are allowed
                        "DirectoryType.resource_1",
                        "DirectoryType.resource_2",
                        "DirectoryType.resource_3",
                        "DirectoryType.actor_1",
                        "DirectoryType.actor_2",
                        "DirectoryType.actor_3",
                        "DirectoryType.file",
                        "DirectoryType.link",
                        "DirectoryType.shared memory",
                        "DirectoryType.socket",
                        "DirectoryType.fifo",
                        "DirectoryType.executable",
                        "DirectoryType.script shell",
                        // for Actor, then also for EP since the type is the same
                        "ActorType.file",
                        "ActorType.directory",
                        "ActorType.shared memory",
                        "ActorType.fifo",

                         // for EP
                        "EPType.<byAcsCreator>",
                        "EPType.<byAcsCreator><immutable> ",
                        //"EPType.<byAcsCreator> network",
                        "EPType.<byAcsCreator> capability",
                        "EPType.<byAcsCreator> authorization",
                        "EPType.<byAcsCreator> authorization for active session",
                        "EPType.<byAcsCreator> authorization for admin active session",
                        // for UserID, only 'account', 'personal account'
                        "UserIDType.group",
                        "UserIDType.personal group",
                        "UserIDType.link",
                        "UserIDType.executable",
                        "UserIDType.script shell",
                        // for GroupID, only 'group', 'personal group'
                        "GroupIDType.account",
                        "GroupIDType.personal account",
                        "GroupIDType.link",
                        "GroupIDType.executable",
                        "GroupIDType.script shell"
            });

            factory.addCommentsOnProperties("All the ACS object types",
                    new String[] {"To understand this complex property,",
                                    "read the documentation. For instance,",
                                    "'CreationByBeamer.NoType' forbids",
                                    "the creation of some objects:",
                                    "- some types which do not match with",
                                    "      the object class,",
                                    "- all the '<byAcsCreator>' types",
                                    "       are reserved to the program."
                                    });

            // GroupID.NoPrimaryGroup, 
            // GroupID.MemberFromAdmin, GroupID.NoMoreThanOneMember for the group policy
            factory.addTypesAndExplorer("GroupID.NoPrimaryGroup",
                    new String[] {
                        "<byAcsCreator> authorization",
                        "<byAcsCreator> authorization for active session",
                        "<byAcsCreator> authorization for admin active session"
            });
            factory.addTypesAndExplorer("GroupID.MemberFromAdmin",
                    new String[] {
                        "<byAcsCreator> authorization for admin active session"
            });
            factory.addTypesAndExplorer("GroupID.NoMoreThanOneMember",
                    new String[] {
                        "<byAcsCreator> authorization for active session",
                        "<byAcsCreator> authorization for admin active session"
            });

            // on Nodes and Resources
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "<byAcsCreator>",
                    new String[] {
                        "<byAcsCreator>",
                        "<byAcsCreator> authorization",
                        //"<byAcsCreator> network",
                        "directory"
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "directory",
                    new String[] {
                        "resource_1",  "resource_2", "resource_3",
                        "actor_1",  "actor_2", "actor_3",
                        "<byAcsCreator>",
                        "<byAcsCreator> authorization",
                        "directory",
                        "file", "executable",
                        "script shell", "link",
                        "shared memory", 
                        "socket", 
                        "fifo"
            });
            factory.addTypesAndExplorer("Resource.OneRightACL",
                    new String[] {
                        "<byAcsCreator> authorization"
            });

            // sets the Privilege restrictions for each Resource type
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "resource_1",
                    new String[] {
                        "typed_PRI_for_account_1_resource_1",
                        "typed_PRI_for_actor_1_resource_1",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "resource_2",
                    new String[] {
                        "typed_PRI_for_account_2_resource_2",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "resource_3",
                    new String[] {
                        "typed_PRI_for_account_3_resource_3",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3",
                        "linked_PRI_for_account_3_resource_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "actor_1",
                    new String[] {
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_account_1_VF__for_actors_1",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "actor_2",
                    new String[] {
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_account_2_VF__for_actors_2",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "actor_3",
                    new String[] {
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                    "directory",
                    new String[] {
                        "typed_PRI_for_actor_1_directory"
            });
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "file",
                new String[0]);
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "shared memory",
                new String[0]);
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "fifo",
                new String[0]);
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "link",
                new String[0]);
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "socket",
                new String[0]);
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "executable",
                new String[0]);
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "script shell",
                new String[0]);
            factory.addTypesAndExplorer("Resource.SecondTypesOfPrivilegeFor."+
                "<byAcsCreator> authorization",
                new String[0]);

            // sets the Privilege restrictions for each VirtualFolder type
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_simple_resources_1",
                    new String[] {
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_simple_resources_2",
                    new String[] {
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_actor_1",
                        "linked_PRI_for_actor_2",
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_actors_1",
                    new String[] {
                        "linked_PRI_for_account_1_VF__for_actors_1"
            });
            factory.addTypesAndExplorer("VirtualFolder.SecondTypesOfPrivilegeFor."+
                    "VF__for_actors_2",
                    new String[] {
                        "linked_PRI_for_account_2_VF__for_actors_2"
            });

            // sets the Privilege restrictions for each EligibleParty type
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "account_1",
                    new String[] {
                        "typed_PRI_for_account_1_resource_1",
                        "linked_PRI_for_account_1",
                        "linked_PRI_for_account_1_VF__for_actors_1"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "account_2",
                    new String[] {
                        "typed_PRI_for_account_2_resource_2",
                        "linked_PRI_for_account_2",
                        "linked_PRI_for_account_2_VF__for_actors_2"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "account_3",
                    new String[] {
                        "typed_PRI_for_account_3_resource_3",
                        "linked_PRI_for_account_3",
                        "linked_PRI_for_account_3_resource_3"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "actor_1",
                    new String[] {
                        "typed_PRI_for_actor_1_resource_1",
                        "typed_PRI_for_actor_1_directory",
                        "linked_PRI_for_actor_1"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "actor_2",
                    new String[] {
                        "linked_PRI_for_actor_2"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                    "actor_3",
                    new String[] {
                        "linked_PRI_for_actor_3"
            });
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "<byAcsCreator>",
                 new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "account",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "group",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "personal account",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "personal group",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "link",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "executable",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "script shell",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "<byAcsCreator> authorization",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "<byAcsCreator> authorization for active session",
                new String[0]);
            factory.addTypesAndExplorer("EligibleParty.SecondTypesOfPrivilegeFor."+
                "<byAcsCreator> authorization for admin active session",
                new String[0]);  

            // associations to any account
                // account in '<AllUserAccounts>'
                    // note: registration is deleted when the account is deleted 
            factory.addAssociationsToNewObject(
                    "<Class>.<Value>.UserID", 
                    new String[] {
                        "<CalledMethod>.<Object>."+
                        ((Name)factory.getName()).extended("<G>").extended(
                                "<AllUserAccounts>").toString()+
                                "+<Class>+GroupID+<Value>+addEorMember",
                        "GroupMember.<Master>"});  

            // associations to a new personal account: the master
                // personal group creation
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<AssociateClass>.GroupID",
                        "Name.<Value>.<MasterNameLastComponent>",
                        "ACS.<Value>.<MasterValue>",
                        "EP Type.<Value>.personal group",
                        "Is Admin.<Value>.false"});
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<AssociateClass>.GroupID "+l_asso[l_asso.length - 1]);             

                // personal group as group of the master
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Action>.1.addEorMember",
                        "GroupMember.<Master>"});
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<CalledMethod>.<Action>.1.addEorMember: "+l_asso[l_asso.length - 1]);

            // immutable personal group
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Action>.1.setEPType",
                        "EP Type.<Value>.<immutable> personal group"});  
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<CalledMethod>.<Action>.1.setEPType: "+l_asso[l_asso.length - 1]);

                // personal account in 'users'
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Object>."+
                        ((Name)factory.getName()).extended("<G>").extended("users").toString()+
                        "+<Class>+GroupID+<Value>+addEorMember",
                        "GroupMember.<Master>"}); 
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<CalledMethod>.<Object>."+
                        ((Name)factory.getName()).extended("<G>").extended("users").toString()+
                        "+<Class>+GroupID+<Value>+addEorMember: "+
                        l_asso[l_asso.length - 1]);

                // personal directory creation
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<AssociateClass>.Directory",
                        "Name.<Value>.<MasterNameLastComponent>",
                        "ACS.<Value>.<MasterValue>",
                        "Parent Node.<Value>.<ACSName>+ <files_tree>:: home::  ",
                        "UserID.<Master>",
                        "Resource Type.<Value>.<immutable> directory"});  
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<AssociateClass>.Directory: "+
                        l_asso[l_asso.length - 1]);

            // new directory comment
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Action>.5.setComment",
                        "Comment.<Value>.This is the personal directory of the same-name account"}); 
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<CalledMethod>.<Action>.6.setComment: "+
                        l_asso[l_asso.length - 1]);

            // master primary group
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Master>.setEorPrimGroup",
                        "Prim Group.<ArgumentValue>.1"});
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<CalledMethod>.<Master>.setEorPrimGroup: "+
                        l_asso[l_asso.length - 1]);

            // immutable master
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Master>.setEPType",
                        "EP Type.<Value>.<immutable> personal account"});
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<CalledMethod>.<Master>.setEPType: "+
                        l_asso[l_asso.length - 1]);

            factory.addCommentsOnProperties("Actions for association to a new object",
                    new String[] {"To understand this complex property,",
                                    "read the documentation.",
                                    "For instance,",
                                    "'UserID.EPType.<ValueContains>.",
                                    "personal account' is a key which sets",
                                    "a script of 8 actions when",
                                    "a new UserID is dedicated to a person.",
                                    "For each new personal account,",
                                    "the action 1 creates a personal group,",
                                    "the action 5 creates a personal directory.",
                                    "The other actions set some associated",
                                    "properties, like the personal account",
                                    "membership to its personal group."
                                    });

            // creates the rights sets ---------------------------------------------
            factory.setL_AccountRightsSet(new StringRight[] {
                acc, activ, go});
            factory.setL_GroupRightsSet(new StringRight[] {
                acc, activ, go});
            factory.setL_bridgeRightsSet(new StringRight[] {activ});
            factory.addSpecializedRights(
                    "Target.AccRightsSet.Interface.Directory",
                    new StringRight[] {acc, activ, go});
            factory.addSpecializedRights(
                    "Target.GpORightsSet.Interface.Directory",
                    new StringRight[] {acc, activ, go});

            // ACL for resources
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.actor_1",
                    new StringRight[] {acc, activ, go});
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.actor_2",
                    new StringRight[] {acc, activ, go});
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.actor_3",
                    new StringRight[] {acc, activ, go});
            
            // ACL for authorizations
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.<byAcsCreator> authorization",
                    new StringRight[] {doi, doifc, doifas, doifcas, ddoi});

            // ACL for sources
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> authorization",
                    new StringRight[] {doi});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> authorization for active session",
                    new StringRight[] {doi});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> authorization for admin active session",
                    new StringRight[] {doi});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.personal account",
                    new StringRight[] {doi, doifc, doifas, doifcas, ddoi});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> capability",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.group",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.personal group",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<immutable> group",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<immutable> personal group",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator>",
                    new StringRight[0]);

            // for Privileges
            factory.addPrivilegeRights(
                    "MainType.TYPED",
                    new StringRight[] {acc, dacc, activ, dactiv, go, dgo});
            factory.addPrivilegeRights(
                    "MainType.LINKED",
                    new StringRight[] {acc, dacc, activ, dactiv, go, dgo});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_account_1_resource_1",
                    new StringRight[] {acc, activ, go});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_account_2_resource_2",
                    new StringRight[] {acc, activ, go});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_account_3_resource_3",
                    new StringRight[] {go, dgo});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_actor_1_resource_1",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFType.SecondType.typed_PRI_for_actor_1_directory",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_1",
                    new StringRight[] {acc, activ, go});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_2",
                    new StringRight[] {acc, activ, go});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_3",
                    new StringRight[] {acc, activ, go});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_actor_1",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_actor_2",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_actor_3",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_1_VF__for_actors_1",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_2_VF__for_actors_2",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});
            factory.addPrivilegeRights(
                    "PFLink.SecondType.linked_PRI_for_account_3_resource_3",
                    new StringRight[] {acc, go, activ, dactiv, dacc, dgo});


            factory.addCommentsOnProperties("Specialized rights",
                    new String[] {"Simple Typed privileges have ",
                                    "(acc, activ, go) rights for ",
                                    "account/resource 1 and 2, ",
                                    "(go, dgo) rights for ",
                                    "account/resource 3, ",
                                    "(acc, go, activ, dactiv, dacc, dgo) ",
                                    "rights for actor 1 on resource_1 ",
                                    "and directory, ",
                                    "and simple Linked privileges",
                                    "have all rights but not ",
                                    "for account_1 which is limited ",
                                    "to (acc, activ, go)."
                                    });

            // final check and creation
            acs = factory.createACS();

        } catch (CreateError c) {

                // for testing and tracking
                c.printStackTrace();
                throw new BaseError("From AcsFactory.createTestOfAll(), "+
                        "exception: "+c);
        } catch (UpDateError u) {

                // for testing and tracking
                u.printStackTrace();
                throw new BaseError("From AcsFactory.createTestOfAll(), "+
                        "exception: "+u);
        }


                    // for testing and tracking
                    //System.out.println("in AcsFactory.createTestOfAll(), "+
                      //"end of the method");

        factory = null;
        acc = null;
        dacc = null;
        activ = null;
        dactiv = null;
        go = null;
        dgo = null;
        l_asso = null;
        doifcas = null;
        doifc = null;
        doifas = null;
        doi = null;
        ddoi = null;

        return acs;
  } // createTestOfAll()
  
  /**
  * Creates the elements of the RBAC application ACS.
  * Creates a resources tree with directories, leaves, and the group trees.
  * Called by ActionNewACSyst.fillWithAcsObjects().
  * @param _result to construct
  * @param _core the core ACS of _result
  * @return ACS with the initial and standard trees of resources and
  *   eligible parties.
  * @exception InternalError if the method fails
  */
  public static final ACS configureRBAC(ACSImpl _result, ACS _core) {

                            // for testing and tracking
                            //System.out.println("call to AcsFactory.configureRBAC()");

    if (!_result.getManageExternalAcls())
        throw new InternalError(" This RBAC application is from an Access Road"+
                " version before 0.7.1");

    //  creates GroupIDs
    ImmutableName   naACS = _result.getName();
    Name            naRBAC = ((Name)naACS.clone()).extended("<"+naACS.lastComponent()+
                                " operations>");
                                // new in 0.7.1
    Name            toBase = ((Name)naACS.clone()).extended("<database handler>");
    Name            naGrpOne = ((Name)naACS.clone()).extended("<sites>");
    GroupIDImpl     grpOne = null;
    Name            naGrpTwo = ((Name)naACS.clone()).extended("<functions>");
    // 'accounts & purchase'
    GroupIDImpl     grpTwo = null;
    GroupIDImpl     grpThree = null;
    GroupIDImpl     grpFour = null;
    GroupIDImpl     grpFive = null;    
    GroupIDImpl     grpSix = null;
    GroupIDImpl     grpSeven = null; 
    GroupIDImpl     grpEight = null;
    // 'accounts manager'
    GroupIDImpl     grpNine = null;
    // 'finance manager'
    GroupIDImpl     grpTen = null; 
    GroupIDImpl     grpE = null;    
    GroupIDImpl     grpN = null;
    GroupIDImpl     grpT = null; 
    
    GroupIDImpl     grp = null;
    GroupIDImpl     grp2 = null;
    // accounts
    UserIDImpl      useOne = null;
    UserIDImpl      useTwo = null;

    try {
      // <sites> in the site system
      grpOne = new GroupIDImpl(naGrpOne,
              _result, "<byAcsCreator> physical", false);
      grpOne.setComment("Contains all the physical sites for the application users");    
      grpTwo = new GroupIDImpl(((Name)naACS.clone()).extended("Paris site"),
              _result, "physical group", false);
      grpTwo.setComment("Contains the Paris site users");
      grpThree = new GroupIDImpl(((Name)naACS.clone()).extended("Cilaos site"),
              _result, "physical group", false);
      grpThree.setComment("Contains the Cilaos site users");
      grpOne.setEorL_Members(new GroupIDImpl[] {grpTwo, grpThree});
      grpOne = new GroupIDImpl(((Name)naACS.clone()).extended("Rahjan street"),
              _result, "physical group", false);
      grpTwo.setComment("Contains the Rahjan users"); 
      grpThree.setEorL_Members(new GroupIDImpl[] {grpOne});
      
      // user accounts
      useOne = new UserIDImpl(((Name)naACS.clone()).extended("John"),
                    _result, "personal user account", false);
      useTwo = new UserIDImpl(((Name)naACS.clone()).extended("Aicha"),
                    _result, "personal user account", false);
      // John in 'Rahjan'
      grpOne.setEorL_Members(new GroupIDMemberImpl[] {useOne});
      // Aicha in 'Paris'
      grpTwo.setEorL_Members(new GroupIDMemberImpl[] {useTwo});
      
      // <functions> in the functional system
      grpOne = new GroupIDImpl(naGrpTwo,
              _result, "<byAcsCreator> functional", false);
      grpOne.setComment("Contains all the functional roles of the users");            
      
      grpTwo = new GroupIDImpl(((Name)naACS.clone()).extended("accounts & purchase"),
              _result, "functional group", false);
      grpTwo.setComment("Contains the accounting groups");
      grpNine = new GroupIDImpl(((Name)naACS.clone()).extended("accounts manager"),
              _result, "manager role", false);
      grpNine.setComment("Contains the account managers");
      // John in 'accounts manager'
      grpNine.addEorMember(useOne);
      // Aicha in 'accounts manager'
      grpNine.addEorMember(useTwo);     
      
      grpN = new GroupIDImpl(((Name)naACS.clone()).extended("accounts operational"),
              _result, "operational role", false);
      grpN.setComment("Contains the accounts operational users");      
      grpT = new GroupIDImpl(((Name)naACS.clone()).extended("accounts transient"),
              _result, "transient role", false);
      grpT.setComment("Contains the accounts transient users"); 
      grpTwo.setEorL_Members(new GroupIDImpl[] {grpNine, grpN, grpT});      
      
      grpThree = new GroupIDImpl(((Name)naACS.clone()).extended("finance"),
              _result, "functional group", false);
      grpThree.setComment("Contains the finance groups");
      grpTen = new GroupIDImpl(((Name)naACS.clone()).extended("finance manager"),
              _result, "manager role", false);
      grpTen.setComment("Contains the finance managers");
      grpN = new GroupIDImpl(((Name)naACS.clone()).extended("finance operational"),
              _result, "operational role", false);
      grpN.setComment("Contains the finance operational users");      
      grpT = new GroupIDImpl(((Name)naACS.clone()).extended("finance transient"),
              _result, "transient role", false);
      grpT.setComment("Contains the finance transient users"); 
      grpThree.setEorL_Members(new GroupIDImpl[] {grpTen, grpN, grpT}); 
      
      grpFour = new GroupIDImpl(((Name)naACS.clone()).extended("sales"),
              _result, "functional group", false);
      grpFour.setComment("Contains the sales groups"); 
      grpE = new GroupIDImpl(((Name)naACS.clone()).extended("sales manager"),
              _result, "manager role", false);
      grpE.setComment("Contains the sales managers");
      grpN = new GroupIDImpl(((Name)naACS.clone()).extended("sales operational"),
              _result, "operational role", false);
      grpN.setComment("Contains the sales operational users");      
      grpT = new GroupIDImpl(((Name)naACS.clone()).extended("sales transient"),
              _result, "transient role", false);
      grpT.setComment("Contains the sales transient users"); 
      grpFour.setEorL_Members(new GroupIDImpl[] {grpE, grpN, grpT});       
      
      grpFive = new GroupIDImpl(((Name)naACS.clone()).extended("communication"),
              _result, "functional group", false);
      grpFive.setComment("Contains the communication groups");  
      grpE = new GroupIDImpl(((Name)naACS.clone()).extended("communication manager"),
              _result, "manager role", false);
      grpE.setComment("Contains the communication managers");
      grpN = new GroupIDImpl(((Name)naACS.clone()).extended("communication operational"),
              _result, "operational role", false);
      grpN.setComment("Contains the communication operational users");      
      grpT = new GroupIDImpl(((Name)naACS.clone()).extended("communication transient"),
              _result, "transient role", false);
      grpT.setComment("Contains the communication transient users"); 
      grpFive.setEorL_Members(new GroupIDImpl[] {grpE, grpN, grpT});             
      
      grpSix = new GroupIDImpl(((Name)naACS.clone()).extended("audit"),
              _result, "functional group", false);
      grpSix.setComment("Contains the application auditor groups");
      grpE = new GroupIDImpl(((Name)naACS.clone()).extended("finance audit"),
              _result, "operational audit role", false);
      grpE.setComment("Contains the finance auditors");
      grpN = new GroupIDImpl(((Name)naACS.clone()).extended("security audit"),
              _result, "security audit role", false);
      grpN.setComment("Contains the security auditors");       
      grpSix.setEorL_Members(new GroupIDImpl[] {grpE, grpN}); 
      
      grpSeven = new GroupIDImpl(((Name)naACS.clone()).extended("administration"),
              _result, "functional group", false);
      grpSeven.setComment("Contains the application administration groups");
      grpE = new GroupIDImpl(((Name)naACS.clone()).extended("general administration"),
              _result, "administrator role", false);
      grpE.setComment("Contains the application administrators");
      grpN = new GroupIDImpl(((Name)naACS.clone()).extended("rights administration"),
              _result, "administrator role", false);
      grpNine.setComment("Contains the rights administrators");       
      grpSeven.setEorL_Members(new GroupIDImpl[] {grpE, grpN});  
      
      grpEight = new GroupIDImpl(((Name)naACS.clone()).extended("extended management"),
              _result, "manager group", false);
      grpEight.setComment("Contains some manager groups having 'enter data' rights");  
      grpEight.setEorL_Members(new GroupIDImpl[] {grpNine, grpTen});      
      
      grpOne.setEorL_Members(new GroupIDImpl[] {grpTwo, grpThree,
        grpFour, grpFive, grpSix, grpSeven, grpEight});
      
      // creates a resources tree with directories and resources
      Actor         act = null;
      Actor         base = null;
      Directory     rbac = null;
      Directory     dir = null;
      
      StringRight[] l_right = new StringRight[] {
                                    _result.getOneRight("browse data")
                                };
      StringRight[] l_right2 = new StringRight[] {
                                    _result.getOneRight("enter data")
                                };
      StringRight[] l_right3 = new StringRight[] {
                                    _result.getOneRight("enter data"),
                                    _result.getOneRight("control")
                                };
      StringRight[] l_right4 = new StringRight[] {
                                    _result.getOneRight("audit update")
                                };
      StringRight[] l_right5 = new StringRight[] {
                                    _result.getOneRight("connect"),
                                    _result.getOneRight("administer")
                                };             
      StringRight[] l_right6 = new StringRight[] {
                                    _result.getOneRight("connect"),
                                    _result.getOneRight("right update")
                                }; 
      StringRight[] l_right7 = new StringRight[] {
                                    _result.getOneRight("launch")
                                };
      
                // for testing and tracking
                //System.out.println("AcsFactory.configureRBAC() after groups creation ");
      
      // creates the files tree, where only the root '<name operations>' is immutable   
      rbac = new DirectoryImpl(naRBAC, _result, null, null, "<byAcsCreator>");
      base = new ActorImpl(toBase, _result, null, null, "to database", null);

      dir = new DirectoryImpl(((Name)naRBAC.clone()).extended("accountancy"),
              _result, rbac, null, "transaction set");
      act = new ActorImpl(((Name)dir.getName()).extended("loads"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'accounts & purchase'
      act.addAclEntry(_result, grpTwo, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'accounts operational'
      grp = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("accounts operational") );      
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'accounts manager'
      grp2 = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("accounts manager") );      
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'     
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("sales"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'accounts & purchase'
      act.addAclEntry(_result, grpTwo, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'accounts operational'     
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'accounts manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'    
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("taxes"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'accounts & purchase'
      act.addAclEntry(_result, grpTwo, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'accounts operational'     
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'accounts manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'    
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("shares"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'accounts & purchase'
      act.addAclEntry(_result, grpTwo, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'accounts operational'     
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'accounts manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'    
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      
      dir = new DirectoryImpl(((Name)naRBAC.clone()).extended("purchase"),
              _result, rbac, null, "transaction set");
      act = new ActorImpl(((Name)dir.getName()).extended("projects"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'accounts & purchase'
      act.addAclEntry(_result, grpTwo, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'accounts operational'      
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'accounts manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'     
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("recurrent purchase"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'accounts & purchase'
      act.addAclEntry(_result, grpTwo, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'accounts operational'      
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'accounts manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'     
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("subcontracting"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'accounts & purchase'
      act.addAclEntry(_result, grpTwo, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'accounts operational'      
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'accounts manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'     
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      
      dir = new DirectoryImpl(((Name)naRBAC.clone()).extended("finance"),
              _result, rbac, null, "transaction set");
      act = new ActorImpl(((Name)dir.getName()).extended("treasury"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'finance'
      act.addAclEntry(_result, grpThree, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'finance operational'
      grp = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("finance operational") );      
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'finance manager'
      grp2 = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("finance manager") );      
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'     
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("stocks"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'finance'
      act.addAclEntry(_result, grpThree, l_right, true, null);
      // 'enter data' for 'finance operational'     
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'finance manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'     
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("shareholders"),
              _result, dir, null, "transaction", null);
      // 'browse' for 'finance'
      act.addAclEntry(_result, grpThree, l_right, true, null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'enter data' for 'finance operational'     
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'finance manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'     
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      
      dir = new DirectoryImpl(((Name)naRBAC.clone()).extended("sales"),
              _result, rbac, null, "transaction set");
      act = new ActorImpl(((Name)dir.getName()).extended("communication"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'browse' for 'sales'
      act.addAclEntry(_result, grpFour, l_right, true, null);
      // 'enter data' for 'sales operational'
      grp = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("sales operational") );      
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'sales manager'
      grp2 = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("sales manager") );      
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("sales plan"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'browse' for 'sales'
      act.addAclEntry(_result, grpFour, l_right, true, null);
      // 'enter data' for 'sales operational'     
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'sales manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("sale contacts"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'browse' for 'sales'
      act.addAclEntry(_result, grpFour, l_right, true, null);
      // 'enter data' for 'sales operational'     
      act.addAclEntry(_result, grp, l_right2, true, null);
      // 'control' for 'sales manager'     
      act.addAclEntry(_result, grp2, l_right3, true, null);
      // 'enter data' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right2, true, null);
      
      dir = new DirectoryImpl(((Name)naRBAC.clone()).extended(naACS.lastComponent()+" audit"),
              _result, rbac, null, "transaction set");
      act = new ActorImpl(((Name)dir.getName()).extended("finance audit"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'audit update' for 'finance audit'
      grp2 = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("finance audit") );      
      act.addAclEntry(_result, grp2, l_right4, true, null);
      // 'browse' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("norm audit"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'audit update' for 'finance audit'      
      act.addAclEntry(_result, grp2, l_right4, true, null);
      // 'browse' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("security audit"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'audit update' for 'security audit'      
      grp2 = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("security audit") );
      act.addAclEntry(_result, grp2, l_right4, true, null);
      // 'browse' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right, true, null);
      
      dir = new DirectoryImpl(((Name)naRBAC.clone()).extended(naACS.lastComponent()+" administration"),
              _result, rbac, null, "transaction set");
      act = new ActorImpl(((Name)dir.getName()).extended("application dashboard"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'administer' for 'general administration'
      grp2 = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("general administration") );
      act.addAclEntry(_result, grp2, l_right5, true, null);
      // 'browse' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("users"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'administer' for 'general administration'     
      act.addAclEntry(_result, grp2, l_right5, true, null);
      // 'browse' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right, true, null);
      act = new ActorImpl(((Name)dir.getName()).extended("rights"),
              _result, dir, null, "transaction", null);
      base.addAclEntry(_result, act, l_right7, true, null);
      // 'right update' for 'rights administration'
      grp2 = (GroupIDImpl)_result.getEorM_EligibleParties().get( 
              ((Name) naACS.clone()).extended("<G>").extended("rights administration") );      
      act.addAclEntry(_result, grp2, l_right6, true, null);
      // 'browse' for 'extended management'      
      act.addAclEntry(_result, grpEight, l_right, true, null);
      
                // for testing and tracking
                //System.out.println("AcsFactory.configureRBAC() end ");
      
      
      naACS = null;
      naRBAC = null;
      act = null;
      base = null;
      naGrpOne = null;
      grpOne = null;
      naGrpTwo = null;
      grpTwo = null;
      grpThree = null;
      grpFour = null;
      grpFive = null;    
      grpSix = null;
      grpSeven = null; 
      grpEight = null;
      grpNine = null;
      grpTen = null; 
      grpE = null;    
      grpN = null;
      grpT = null; 
      grp = null;
      grp2 = null;

      dir = null;
      rbac = null;
      l_right = null;
      l_right2 = null;
      l_right3 = null;
      l_right4 = null;
      l_right5 = null;
      l_right6 = null;
      
    } // try
    catch (CreateError e) {
      e.printStackTrace();
      throw new InternalError(e.toString()+ 
              "AcsFactory.configureRBAC() failed ");
    }

    /*
     * TEST OF THE PATHS BETWEEN
     * <U>::  Aicha, accountancy::  loads, and finance::  shareholders.
     * It covers the paths:
     * - Aicha -> loads, 3 paths:
     *      account/member/group/acl/resource,
     *      account/member/group/indirect mamber/group/acl/resource (x 2)
     *      Main result: the indirect member link is found; the resulting 'control'
     *      right is computed as the sum of the 3 paths
     *
     * The view should generate the following SeeWhy text:
     *
     *   FROM A RIGHT USER TO AN ACCESS TARGET
     *    FROM  HO:: stub:: tub:: trab:: <U>::  Aicha
     *    TO  HO:: stub:: tub:: trab:: <trab operations>:: accountancy::  loads.
     *      PATH number 1:
     *               Link number 1:
     *                     Member
     *               Intermediate node:
     *                      HO:: stub:: tub:: trab:: <G>::  accounts manager
     *               Link number 2:
     *                     ACL ( control )  -- All Rights: control
     *      PATH number 2:
     *               Link number 1:
     *                     Member
     *               Intermediate node:
     *                      HO:: stub:: tub:: trab:: <G>::  accounts manager
     *               Link number 2:
     *                     Indirect Member
     *               Intermediate node:
     *                      HO:: stub:: tub:: trab:: <G>::  extended management
     *               Link number 3:
     *                     ACL ( enter data)  -- All Rights: enter data
     *      PATH number 3:
     *               Link number 1:
     *                     Member
     *               Intermediate node:
     *                      HO:: stub:: tub:: trab:: <G>::  accounts manager
     *               Link number 2:
     *                     Indirect Member
     *               Intermediate node:
     *                      HO:: stub:: tub:: trab:: <G>::  accounts & purchase
     *               Link number 3:
     *                     ACL ( browse data)  -- All Rights: browse data
     */

    return _result;
  } // configureRBAC()
  
  /**
  * Creates the elements of the ACS test for VirtualFolders and Privileges.
  * @param _result to construct
  * @return ACS with the initial and standard trees of resources, eligible parties.
  */
  public static final ACS configureTestForVFandPRI(ACSImpl _result) {

            // for testing and tracking
            //System.out.println("call to AcsFactory.configureTestForVFandPRI()");

    try {
        // creates the Resource Virtual Folders
        new VirtualFolderImpl(new NameImpl("vf_res_1_one"), _result, null,
              "VF__for_simple_resources_1", true);
        new VirtualFolderImpl(new NameImpl("vf_res_2_one"), _result, null,
              "VF__for_simple_resources_2", true);
        // creates the EP Virtual Folders
        new VirtualFolderImpl(new NameImpl("vf_act_1_one"), _result, null,
              "VF__for_actors_1", false);
        VirtualFolder vf_act_2_one = new VirtualFolderImpl(
                new NameImpl("vf_act_2_one"), _result, null,
              "VF__for_actors_2", false);

        // creates the accounts
        UserID acc_one = new UserIDImpl(((Name)_result.getName()).extended("acc_one"),
                _result, "account_1", true, true);
        UserID acc_two = new UserIDImpl(((Name)_result.getName()).extended("acc_two"),
                _result, "account_2", false, true);
        UserID acc_three = new UserIDImpl(((Name)_result.getName()).extended("acc_three"),
                _result, "account_3", false, true);

        // creates the privileges
        new PrivilegeForTypeImpl(_result, "actor_1", "resource_1",
            "T_one", PrivilegeAbst.TYPED, new StringRight[] {_result.getOneRight("GO")} ,
            "typed_PRI_for_actor_1_resource_1", true);
        new PrivilegeForTypeImpl(_result, "account_1", "resource_1",
            "T_two", PrivilegeAbst.TYPED, new StringRight[] {_result.getOneRight("GO")} ,
            "typed_PRI_for_account_1_resource_1", true);
        new PrivilegeForTypeImpl(_result, "account_2", "resource_2",
            "T_three", PrivilegeAbst.TYPED, new StringRight[] {_result.getOneRight("NOACCESS")} ,
            "typed_PRI_for_account_2_resource_2", true);
        PrivilegeForLinksImpl L_one = new PrivilegeForLinksImpl(
                _result, "L_one", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("ACCESS")},
            "linked_PRI_for_account_1_VF__for_actors_1");
        PrivilegeForLinksImpl L_two = new PrivilegeForLinksImpl(
                _result, "L_two", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("GO")},
            "linked_PRI_for_account_2_VF__for_actors_2");
        new PrivilegeForLinksImpl(_result, "L_three", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("GO")},
            "linked_PRI_for_actor_2");
        PrivilegeForLinksImpl L_four = new PrivilegeForLinksImpl(
                _result, "L_four", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("ACCESS")},
            "linked_PRI_for_account_2");

        // creates the resource tree
        Name naC = ((Name)_result.getName()).extended("root_directory");
        DirectoryImpl c = new DirectoryImpl(naC, _result, null, null, "<byAcsCreator>");
        DirectoryImpl b = new DirectoryImpl(((Name)c.getName()).extended("actors"),
                _result, c, null, "<byAcsCreator>");
        
        ActorImpl act_one = new ActorImpl(((Name)b.getName()).extended("act_one"),
              _result, b, null, "actor_1", null);
        act_one.setEorCurrentUserIDForUser(acc_three);
        new ActorImpl(((Name)b.getName()).extended("act_one2"),
              _result, b, null, "actor_1", null);
        new ActorImpl(((Name)b.getName()).extended("act_two"),
              _result, b, null, "actor_2", null);
        Actor act_three = new ActorImpl(((Name)b.getName()).extended("act_three"),
              _result, b, null, "actor_3", null);
        Resource res_one = new ResourceImpl(((Name)naC.clone()).extended(
                "res_one"), _result, c, null,
              "resource_1", true);
        new ResourceImpl(((Name)naC.clone()).extended("res_two"), _result, c, null,
              "resource_2", true);
        new ResourceImpl(((Name)naC.clone()).extended("res_three"), _result, c, null,
              "resource_3", true);

        // L_one from acc_one to act_one
        // L_four from acc_two to res_one,
        // L_two from acc_two to vf_act_2_one
        L_one.addEorPrivilegeSourceLink(acc_one);
        L_one.addEorPrivilegeTargetLink(act_one);
        L_four.addEorPrivilegeSourceLink(acc_two);
        L_four.addEorPrivilegeTargetLink(res_one);
        L_two.addEorPrivilegeSourceLink(acc_two);
        L_two.addEorPrivilegeTargetLink(vf_act_2_one);
        // vf_act_2_one members
        vf_act_2_one.addEorBaseObject(act_three);

        // after the Actor creation, to fill the no-proxy actors list
        BaseManagerImpl.getInstance().restoreACSInternalLinks(_result);

        naC = null;
        b = null;
        c = null;
        L_four = null;
        L_two = null;
        acc_two = null;
        res_one = null;
        vf_act_2_one = null;
        act_three = null;
        acc_one = null;
        act_one = null;
        acc_three = null;
    } catch (CreateError c) {
        throw new InternalError("AcsFactory.configureTestForVFandPRI(): "+c.toString());
    } catch (BaseError b) {
        throw new InternalError("AcsFactory.configureTestForVFandPRI(): "+b.toString());
    } catch (UpDateError u) {
        throw new InternalError("AcsFactory.configureTestForVFandPRI(): "+u.toString());
    }

    /**
     * 1/3
     * TEST OF THE PATHS between
         IO:: one:: vfpr:: <U>:: acc_one::
         IO:: one:: vfpr:: <U>:: acc_two::
         IO:: one:: vfpr:: root_directory:: actors:: act_one::
         IO:: one:: vfpr:: root_directory:: actors:: act_one2::
         IO:: one:: vfpr:: root_directory:: actors:: act_three::
         IO:: one:: vfpr:: root_directory:: res_one::
         IO:: one:: vfpr:: root_directory:: res_two::
         IO:: one:: vfpr:: vf_act_2_one::
     *
     * 
FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: <U>::  acc_one
 TO  IO:: two:: vfi4:: root_directory:: actors::  act_one.
-- PATH NUMBER 1:
            Link number 1:
                  Linked PRI (ACCESS)  --  PRI: L_one  -- All Rights: ACCESS

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: <U>::  acc_one
 TO  IO:: two:: vfi4:: root_directory::  res_one.
-- PATH NUMBER 1:
            Link number 1:
                  Typed PRI (GO)  --  PRI: T_two  -- All Rights: GO

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: <U>::  acc_two
 TO  IO:: two:: vfi4:: root_directory::  res_one.
-- PATH NUMBER 1:
            Link number 1:
                  Linked PRI (ACCESS)  --  PRI: L_four  -- All Rights: ACCESS

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: <U>::  acc_two
 TO  IO:: two:: vfi4:: root_directory::  res_two.
-- PATH NUMBER 1:
            Link number 1:
                  Negative Typed PRI (NOACCESS)  --  PRI: T_three  -- All Rights: NOACCESS

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: <U>::  acc_two
 TO  IO:: two:: vfi4::  vf_act_2_one.
-- PATH NUMBER 1:
            Link number 1:
                  Linked PRI (GO)  --  PRI: L_two  -- All Rights: GO

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: root_directory:: actors::  act_one
 TO  IO:: two:: vfi4:: root_directory::  res_one.
-- PATH NUMBER 1:
            Link number 1:
                  Typed PRI (GO)  --  PRI: T_one  -- All Rights: GO

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: root_directory:: actors::  act_one2
 TO  IO:: two:: vfi4:: root_directory::  res_one.
-- PATH NUMBER 1:
            Link number 1:
                  Typed PRI (GO)  --  PRI: T_one  -- All Rights: GO

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4::  vf_act_2_one
 TO  IO:: two:: vfi4:: root_directory:: actors::  act_three.
-- PATH NUMBER 1:
            Link number 1:
                  Virtual Member  --  (no right)
     *
     * 2/3
     * TEST OF THE PATHS IN A NOTHAN VIEW
     * FROM the 4 EP in the previous EPR view having access to res_one
     * TO res_one
     * No-More-Than Right : GO red
     * No-Less-Than Right : GO green
     *
     * Action: removing of res_one in L_four
     * Result:
     * No-More-Than Right : GO green
     * No-Less-Than Right : GO red
     *
     *
     * 3/3
     * TEST OF THE EXTERNAL ONE-TO-ONE LPRI TO/FROM ANOTHER
     * 'An ACS for testing VirtualFolders and Privileges' ACS, named 'vfi5'
     *
     * - create in this ACS 'Linked PRI (ACCESS)  --  PRI: e_one'
     * from a vfi5 UserID to the resource 'res_one'
     *
     * FROM IO:: two:: vfi5:: <U>::  acc_two
     * TO  IO:: two:: vfi4:: root_directory::  res_one
     * -- PATH NUMBER 1:
     *  Link number 1:
     *          Linked PRI (ACCESS)  --  PRI: e_one  -- All Rights: ACCESS
     *
     * - create in this ACS 'Linked PRI (ACCESS)  --  PRI: e_two'
     * from 'acc_three' to a vfi5 resource
     *
FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: two:: vfi4:: <U>::  acc_one
 TO  IO:: two:: vfi5:: root_directory::  res_one.
-- PATH NUMBER 1:
            Link number 1:
                  Linked PRI (ACCESS)  --  PRI: L_one <Sets Owner>
            Intermediate node:
                   IO:: two:: vfi4:: root_directory:: actors::  act_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: two:: vfi4:: <U>::  acc_three
            Link number 3:
                  Linked PRI (ACCESS)  --  PRI: e_two  -- All Rights: ACCESS
     *
     *
     */

    return _result;
  } // configureTestForVFandPRI()

  /**
  * Creates the elements of the ACS test for VirtualFolders and Privileges.
  * Note: since the AGO 'other' rights are not activated in this strange  'Linux' ACS,
  * the inherited AGO rights are very restricted; to have an AGO right, the same account
  * or group must be applicable for each directory in the parent path.
  * @param _result to construct
  * @return ACS with the initial and standard trees of resources, eligible parties.
  */
  public static final ACS configureTestOfAll(ACSImpl _result) {

            // for testing and tracking
            //System.out.println("call to AcsFactory.configureTestOfAll()");

    ImmutableName   naACS = _result.getName();
    GroupIDImpl     grpWheel = null;
    Name            naRoot = ((Name)naACS.clone()).extended("root_account");
    UserIDImpl      root = null;
    GroupIDImpl     grpThree = null;
    GroupIDMemberImpl[] l_ep_one = null;
    UserIDImpl      tom = null;
    UserIDImpl      jerry = null;
    Resource        ajerry = null;
    Actor           bridgeSource = null;
    GroupIDImpl     grpJerry = null;
    Directory       djerry = null;

    try {
        //creates the Resource Virtual Folders
        VirtualFolder vfr = new VirtualFolderImpl(new NameImpl("vf_res_1_one"), _result, null,
              "VF__for_simple_resources_1", true);
        VirtualFolder vfr2 = new VirtualFolderImpl(new NameImpl("vf_res_2_one"), _result, null,
              "VF__for_simple_resources_2", true);
        //creates the EP Virtual Folders
        VirtualFolder vfe = new VirtualFolderImpl(new NameImpl("vf_act_1_one"), _result, null,
              "VF__for_actors_1", false);
        VirtualFolder vfe2 = new VirtualFolderImpl(new NameImpl("vf_act_2_one"), _result, null,
              "VF__for_actors_2", false);

        // <AllUserAccounts> in the authorization system,
        // contains all the new userids, by the mean of an associate command
        grpThree = new GroupIDImpl(((Name)naACS.clone()).extended("<AllUserAccounts>"),
              _result, "<byAcsCreator> authorization", false);
        grpThree.setComment("Contains all the personal and applicative accounts of Linux");

        // wrong EP type to allow the updating
        grpWheel = new GroupIDImpl(naRoot, _result, null, true);
        root = new UserIDImpl(naRoot, _result, null, true);
        grpWheel.setOrder(0);
        root.setOrder(0);
        root.setEPType("<byAcsCreator>");
        grpWheel.setEPType("<byAcsCreator>");

        // GroupIDs for AUTHORIZATION system
        // <Console> in the authorization system
        grpThree = new GroupIDImpl(((Name)naACS.clone()).extended("<Console>"),
              _result, "<byAcsCreator> authorization", false);
        grpThree.setComment("Put in this group, the accounts which are associated to a console");

        // <ActiveSession> in the authorization system
        // with a type in the 'GroupID.NoMoreThanOneMember' values
        grpThree = new GroupIDImpl(((Name)naACS.clone()).extended("<ActiveSession>"),
              _result, "<byAcsCreator> authorization for active session", false);
        grpThree.setComment("Put in this group, the account which is associated to an active session");

        // users group
        grpThree = new GroupIDImpl(((Name)naACS.clone()).extended("users"),
              _result, "<byAcsCreator>", false);
        grpThree.setOrder(100);
        
        grpThree = new GroupIDImpl(((Name)naACS.clone()).extended("sys"),
              _result, "<byAcsCreator>", false);
        grpThree.setOrder(3);
        root = new UserIDImpl(((Name)naACS.clone()).extended("sys"), _result, null, false);
        l_ep_one = new GroupIDMemberImpl[] {root};
        grpThree.setEorL_Members(l_ep_one);
        root.setOrder(3);
        root.setEorPrimGroup(grpThree);
        root.setEPType("<byAcsCreator>");

        // creates the other accounts
        new UserIDImpl(((Name)_result.getName()).extended("acc_one"),
                _result, "account_1", true, true);
        UserIDImpl acc2 = new UserIDImpl(((Name)_result.getName()).extended("acc_two"),
                _result, "account_2", false, true);
        new UserIDImpl(((Name)_result.getName()).extended("acc_three"),
                _result, "account_3", false, true);

        // creates the privileges
        new PrivilegeForTypeImpl(_result, "actor_1", "resource_1",
            "T_one", PrivilegeAbst.TYPED, new StringRight[] {_result.getOneRight("GO")} ,
            "typed_PRI_for_actor_1_resource_1", true);
        new PrivilegeForTypeImpl(_result, "account_1", "resource_1",
            "T_two", PrivilegeAbst.TYPED, new StringRight[] {_result.getOneRight("GO")} ,
            "typed_PRI_for_account_1_resource_1", true);
        new PrivilegeForTypeImpl(_result, "account_2", "resource_2",
            "T_three", PrivilegeAbst.TYPED, new StringRight[] {_result.getOneRight("NOACCESS")} ,
            "typed_PRI_for_account_2_resource_2", true);
        new PrivilegeForLinksImpl(_result, "L_one", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("ACCESS")},
            "linked_PRI_for_account_1_VF__for_actors_1");
        new PrivilegeForLinksImpl(_result, "L_two", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("GO")},
            "linked_PRI_for_account_2_VF__for_actors_2");
        new PrivilegeForLinksImpl(_result, "L_three", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("GO")},
            "linked_PRI_for_actor_2");
        new PrivilegeForLinksImpl(_result, "L_four", PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("ACCESS")},
            "linked_PRI_for_account_2");
        PrivilegeForLinksImpl pri5 = new PrivilegeForLinksImpl(_result, "L_five",
            PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("ACCESS")},
            "linked_PRI_for_actor_2");
        PrivilegeForLinksImpl pri6 = new PrivilegeForLinksImpl(_result, "L_six",
            PrivilegeAbst.LINKED,
            new StringRight[] {_result.getOneRight("ACCESS")},
            "linked_PRI_for_account_2_VF__for_actors_2");

        // creates the resource tree and fills the virtual folders
        Name naC = ((Name)_result.getName()).extended("root_directory");
        DirectoryImpl c = new DirectoryImpl(naC, _result, null, null, "directory");
        DirectoryImpl b = new DirectoryImpl(((Name)c.getName()).extended("actors"),
                _result, c, null, "directory");

        Actor ac = new ActorImpl(((Name)b.getName()).extended("act_one"),
              _result, b, null, "actor_1", null);
        Actor ac2 = new ActorImpl(((Name)b.getName()).extended("act_one2"),
              _result, b, null, "actor_1", null);
        Actor ac3 = new ActorImpl(((Name)b.getName()).extended("act_two"),
              _result, b, null, "actor_2", null);
        Actor ac4 = new ActorImpl(((Name)b.getName()).extended("act_three"),
              _result, b, null, "actor_2", null);
        Resource re = new ResourceImpl(((Name)naC.clone()).extended("res_one"), _result, c,
                null, "resource_1", true);
        Resource re2 = new ResourceImpl(((Name)naC.clone()).extended("res_two"), _result, c,
                null, "resource_2", true);
        new ResourceImpl(((Name)naC.clone()).extended("res_three"), _result, c,
                null, "resource_3", true);

        vfr.addEorBaseObject(re);
        vfr.addEorBaseObject(re2);
        vfe.addEorBaseObject(ac);
        vfe.addEorBaseObject(ac2);
        vfe2.addEorBaseObject(ac3);
        pri5.addEorPrivilegeSourceLink(ac3);
        ac3.setEorCurrentUserIDForUser(acc2);
        pri5.addEorPrivilegeTargetLink(re);
        pri5.addEorPrivilegeTargetLink(re2);
        pri6.addEorPrivilegeSourceLink(acc2);
        pri6.addEorPrivilegeTargetLink(vfe2);

        ac4.setEorCurrentUserIDForUser(acc2);

        // '<files_tree>' directory
        c = new DirectoryImpl(((Name)_result.getName()).extended("<files_tree>"),
                _result, null, acc2, "directory");
        b = new DirectoryImpl(((Name)c.getName()).extended("home"),
                _result, c, acc2, "directory");

        // creates the personal accounts
        jerry = new UserIDImpl(((Name)_result.getName()).extended("jerry"),
                _result, "personal account", false, true);
        tom = new UserIDImpl(((Name)_result.getName()).extended("tom"),
                _result, "personal account", false, true);

        // add the associates
        AssociateUtilityImpl.addAssociatesAfterCreation(jerry);
        AssociateUtilityImpl.addAssociatesAfterCreation(tom);

        grpJerry = (GroupIDImpl)_result.getEorM_EligibleParties().get(
                ((Name)_result.getName()).extended("<G>").extended("jerry") );
        grpJerry.setEPType("personal group");

        // 'acc_two' under the 'jerry' group created by addAssociatesAfterCreation(jerry)
        // and 'jerry' is its primary group
        grpJerry.addEorMember(acc2);
        acc2.setEorPrimGroup(grpJerry);

        // 'acc_two' and 'grpjerry' for the '<files_tree>' and 'home' directories
        c.setEorUserID(acc2);
        b.setEorUserID(acc2);
        c.setL_groupRights(new StringRight[] {_result.getOneRight("ACCESS")});
        b.setL_groupRights(new StringRight[] {_result.getOneRight("ACCESS")});

        djerry = ((Directory)_result.getEorM_Resources().get(((Name)b.getName()).extended("jerry")));
        djerry.setResourceType("directory");
        djerry.setL_groupRights(new StringRight[] {_result.getOneRight("ACCESS")});

        // a_jerry under the directory 'jerry' with the owner right 'ACCESS'
        // (the directory is created by addAssociatesAfterCreation(jerry))
        ajerry = new ResourceImpl(((Name)b.getName()).extended("jerry").extended("a_jerry"),
                _result,
                (Directory)_result.getEorM_Resources().get(((Name)b.getName()).extended("jerry")),
                acc2, "file", true);
        ajerry.setL_ownerRights(new StringRight[] {_result.getOneRight("ACCESS")});
        ajerry.setL_groupRights(new StringRight[] {_result.getOneRight("ACCESS")});

        // searching for a Linux Ubuntu ACS to add a bridge from  its 'to_"+_result.getNickName()' actor
        // to 'act_three' in this ACS
        ACS acs = null;
        for (Object obj : BaseManagerImpl.getInstance().getEorM_OpenACS().values()) {
            if (((ACS)obj).getACSType().indexOf("Linux Ubuntu") != -1)
                acs = (ACS)obj;
        } // for obj
        if (acs != null) {
            new ActorUbuntuImpl(
                    ((Name)acs.getName()).extended("<files_tree>").extended("home").
                        extended("jerry").extended("d_jerry").extended("to_"+_result.getNickName()),
                    acs,
                    (Directory)acs.getEorM_Resources().get(
                        ((Name)acs.getName()).extended("<files_tree>").extended("home").
                        extended("jerry").extended("d_jerry") ),
                    (UserID)acs.getEorM_EligibleParties().get(
                        ((Name)acs.getName()).extended("<U>").extended("jerry") ),
                    "executable",
                    ac4);
        } // if (acs

        // after the Actor creation, to fill the no-proxy actors list
        BaseManagerImpl.getInstance().restoreACSInternalLinks(_result);

        naC = null;
        b = null;
        c = null;
        vfr = null;
        vfr2 = null;
        vfe = null;
        vfe2 = null;
        ac = null;
        ac2 = null;
        ac3 = null;
        ac4 = null;
        re = null;
        re2 = null;
        pri5 = null;
        pri6 = null;
        l_ep_one = null;
        jerry = null;
        tom = null;
        ajerry = null;
        bridgeSource = null;
    } catch (CreateError c) {
        throw new InternalError("AcsFactory.configureTestOfAll(): "+c.toString());
    } catch (BaseError b) {
        throw new InternalError("AcsFactory.configureTestOfAll(): "+b.toString());
    } catch (UpDateError u) {
        throw new InternalError("AcsFactory.configureTestOfAll(): "+u.toString());
    }

    /*
     * 1/3 - TEST OF PATHS BETWEEN act_two and res_two
     *      Main result: the NOACCESS right and the ACCESS right are subtracted,
     *      and the result is MODIF GO ACTIV
     * FROM A RIGHT USER TO AN ACCESS TARGET
     * FROM  IO:: one:: full:: root_directory:: actors::  act_two
     * TO  IO:: one:: full:: root_directory::  res_two.
     * PATH NUMBER 1:
     *       Link number 1:
     *             Negative Typed PRI  -- Linked PRI (ACCESS)  --  PRI: L_five <includes negative rights from other paths>  -- All Rights: NOACCESS/ACCESS
     * PATH NUMBER 2:
     *       Link number 1:
     *             Runs Under
     *       Intermediate node:
     *              IO:: one:: full:: <U>::  acc_two
     *       Link number 2:
     *             Negative Typed PRI (NOACCESS)  --  PRI: T_three  -- All Rights: NOACCESS
     *
     *
     * 2/3 - TEST OF THE PATHS BETWEEN acc_two, res_two and a_jerry.
     * It covers the paths:
     * - acc_two -> a_jerry:
     *      account/own/resource with AGO inheritance,
     *      Main result: the primary group of the account on the a_jerry direct parent
     *      delivers rename/view/create /goto/copy/delete rights
     * - acc_two -> res_two, 2 paths:
     *      account/negative linked privilege/resource,
     *      account/linked privilege/virtual folder/virtual member/actor/linked privilege/resource
     *      Main result: the NOACCESS right and the ACCESS right are subtracted,
     *      and the result is MODIF GO ACTIV
     * 
     * The view should generate the following SeeWhy text:
     * 
     *    FROM A RIGHT USER TO AN ACCESS TARGET
     *    FROM  LE:: sful:: fulo:: <U>::  acc_two
     *    TO  LE:: sful:: fulo:: <files_tree>:: home:: jerry::  a_jerry.
     *      PATH number 1:
     *               Link number 1:
     *                     Owns  --  <inherited AG rights>  -- All Rights: rename/view/create /goto/copy/delete/ACCESS
     *
     *   FROM A RIGHT USER TO AN ACCESS TARGET
     *    FROM  LE:: sful:: fulo:: <U>::  acc_two
     *    TO  LE:: sful:: fulo:: root_directory::  res_two.
     *      PATH number 1:
     *               Link number 1:
     *                     Negative Typed PRI (NOACCESS)  --  PRI: T_three  -- All Rights: NOACCESS
     *      PATH number 2:
     *               Link number 1:
     *                     Linked PRI ( ACCESS)
     *               Intermediate node:
     *                      LE:: sful:: fulo::  vf_act_2_one
     *               Link number 2:
     *                     Virtual Member
     *               Intermediate node:
     *                      LE:: sful:: fulo:: root_directory:: actors::  act_two
     *               Link number 3:
     *                     Linked PRI ( ACCESS)  -- All Rights: ACCESS
     */

    /*
     * 3/3 - TEST OF THE PATHS BETWEEN Linux-ACS e_jerry and this-ACS res_one.
     * It covers the path of 8 nodes:
     * - e_jerry -> res_one:
     *      Main result: find a complex combination of
     *      AGO/Bridge/Linked Privilege/Virtual Member links in 2 ACS
     * 
     *     FROM A RIGHT USER TO AN ACCESS TARGET
     *      FROM  HO:: stub:: tub:: <files_tree>:: home:: jerry:: d_jerry::  e_jerry
     *      TO  LE:: sful:: fulo:: root_directory::  res_one.
     *      PATH number 1:
     *       Link number 1:
     *             Run Under 
     *       Intermediate node:
     *              HO:: stub:: tub:: <U>::  jerry
     *       Link number 2:
     *             Owner 
     *       Intermediate node:
     *              HO:: stub:: tub:: <files_tree>:: home:: jerry:: d_jerry::  to_fulo
     *       Link number 3:
     *             Bridge 
     *       Intermediate node:
     *              LE:: sful:: fulo:: root_directory:: actors::  act_three
     *       Link number 4:
     *             Run Under 
     *       Intermediate node:
     *              LE:: sful:: fulo:: <U>::  acc_two
     *       Link number 5:
     *             Linked PRI ( ACCESS) 
     *       Intermediate node:
     *              LE:: sful:: fulo::  vf_act_2_one
     *       Link number 6:
     *             Virtual Member 
     *       Intermediate node:
     *              LE:: sful:: fulo:: root_directory:: actors::  act_two
     *       Link number 7:
     *             Linked PRI ( ACCESS)  -- All Rights: ACCESS
     * 
     */


    return _result;
  } // configureTestOfAll()

  /**
  * Creates the elements of the Apache ACS.
  *
  * Under Linux:
  * GroupIDMembers are:
  *                              www-data group
   * 
  *
  * Creates a resources tree with directories, leaves
  * and executables (then actors):
  * @param _result to construct
  * @param _core is the foundation on which _result is build up. Noop if
  * its type is not "Linux 8.04 Desktop".
  * @return ACS with the initial and standard trees of resources,
  *   eligible parties.
  */
  public static final ACSImpl createApache(ACSImpl _result, ACS _core) {

                        // for testing and tracking
                        //System.out.println("call to AcsFactory.createApacheACS()");

    if (!_core.getACSType().equals("Linux 8.04 Desktop"))
        return null;
      
    // ---------------------------
    // create Apache under Impl
    // --------------------------
      /*
    //  creates 'www-data' group and account
    Name coreACS = (Name)_core.getName();
    Name resACS = (Name)_result.getName();
    GroupIDImpl grpWheel = null;
    Name naRoot = ((Name)coreACS.clone()).extended("<G>").extended("root");
    Name nbRoot = ((Name)coreACS.clone()).extended("<U>").extended("root");
    UserIDImpl root = null;
    Name naOne = ((Name)coreACS.clone()).extended("www-data");
    GroupIDImpl grpOne = null;

    UserIDImpl useOne = null;
    
    try {
      grpWheel =(GroupIDImpl) _core.getEorM_EligibleParties().get(naRoot);
      root =(UserIDImpl) _core.getEorM_EligibleParties().get(nbRoot);
      
      grpOne = new GroupIDImpl(naOne, _core, "group");
      grpOne.setOrder(33);
      useOne = new UserIDImpl((Name)naOne.clone(), _core, "account");
      useOne.setOrder(550);      // not in grpOne
    } catch (CreateError c) {
      throw new InternalError("In AcsFactory.createApacheACS(): "+c);
    }

    try {
                // for testing and tracking
                //System.out.println("in AcsFactory.createApacheACS(), "+
                  //"exeType ="+exeType+"\n"+"resType ="+resType+"\n"+"dirType ="+dirType);
      
      // creates 'apache2' tree under /etc
      Resource res;
      Directory program;
      Name name;
      Name aName;

      // creates /etc/apache2 directory and subdirectories
      Directory etc = (Directory)_core.getEorM_Resources().get(
              ((Name)coreACS.clone()).extended("<files_tree>").extended("etc"));
      
      Directory apache = new DirectoryImpl(
              ((Name)etc.getName().clone()).extended("apache2"),
              _core, etc, "directory");
      apache.setEorUserID(root);
      apache.setEorGroupID(grpWheel);
      apache.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      apache.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      apache.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")}); 
      
      program = new DirectoryImpl(
              ((Name)apache.getName().clone()).extended("conf.d"),
              _core, apache, "directory");
      program.setEorUserID(root);
      program.setEorGroupID(grpWheel);
      program.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")}); 
      res = new ResourceImpl(
              ((Name) program.getName().clone()).extended("<typical_conf_file.conf>"), 
              _core, program, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});       

      program = new DirectoryImpl(
              ((Name)apache.getName().clone()).extended("mods-available"),
              _core, apache, "directory");
      program.setEorUserID(root);
      program.setEorGroupID(grpWheel);
      program.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res = new ResourceImpl(
              ((Name) program.getName().clone()).extended("<typical_conf_file.conf>"), 
              _core, program, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      
      program = new DirectoryImpl(
              ((Name)apache.getName().clone()).extended("mods-enabled"),
              _core, apache, "directory");
      program.setEorUserID(root);
      program.setEorGroupID(grpWheel);
      program.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res = new ResourceImpl(
              ((Name) program.getName().clone()).extended("<typical_conf_link>"), 
              _core, program, "link");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      
      program = new DirectoryImpl(
              ((Name)apache.getName().clone()).extended("sites-available"),
              _core, apache, "directory");
      program.setEorUserID(root);
      program.setEorGroupID(grpWheel);
      program.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res = new ResourceImpl(
              ((Name) program.getName().clone()).extended("default"), 
              _core, program, "link");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      
      program = new DirectoryImpl(
              ((Name)apache.getName().clone()).extended("sites-enabled"),
              _core, apache, "directory");
      program.setEorUserID(root);
      program.setEorGroupID(grpWheel);
      program.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res = new ResourceImpl(
              ((Name) program.getName().clone()).extended("000-default"), 
              _core, program, "link");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      
      // creates apache2 resources    
      res = new ResourceImpl(
              ((Name) apache.getName().clone()).extended("apache2.conf"), 
              _core, apache, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
     
      res = new ResourceImpl(
              ((Name) apache.getName().clone()).extended("apache2.conf"), 
              _core, apache, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")}); 
      
      res = new ResourceImpl(
              ((Name) apache.getName().clone()).extended("httpd.conf"), 
              _core, apache, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")}); 
      
      res = new ResourceImpl(
              ((Name) apache.getName().clone()).extended("ports.conf"), 
              _core, apache, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});       
      
      res = new ResourceImpl(
              ((Name) apache.getName().clone()).extended("envvars"), 
              _core, apache, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel);
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      
      // usr/lib/apache2 tree
      Directory usrlib = (Directory)_core.getEorM_Resources().get(
              ((Name)coreACS.clone()).extended("<files_tree>").extended("usr").extended("lib"));      
      apache = new DirectoryImpl(
              ((Name)usrlib.getName().clone()).extended("apache2"),
              _core, usrlib, "directory");
      apache.setEorUserID(root);
      apache.setEorGroupID(grpWheel);
      apache.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      apache.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      apache.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res = new ExecutableImpl(
              ((Name)apache.getName().clone()).extended("suexec"),
              _core, apache, "program", "program",
              null, null);
      res.setEorUserID(root);
      res.setEorGroupID(grpOne); // www-data group
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      program = new DirectoryImpl(
              ((Name)apache.getName().clone()).extended("modules"),
              _core, apache, "directory");
      program.setEorUserID(root);
      program.setEorGroupID(grpWheel);
      program.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      program.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      res = new ResourceImpl(
              ((Name) program.getName().clone()).extended("<typical_shared_library.so>"), 
              _core, program, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel); 
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});

      // /var/run and /var/log directories
      Directory varrun = (Directory)_core.getEorM_Resources().get(
              ((Name)coreACS.clone()).extended("<files_tree>").extended("var").extended("run"));       
      Directory var;
      if (varrun == null) {
          var = (Directory)_core.getEorM_Resources().get(
              ((Name)coreACS.clone()).extended("<files_tree>").extended("var"));
          varrun = new DirectoryImpl(
              ((Name)var.getName().clone()).extended("run"),
              _core, var, "directory");
          varrun.setEorUserID(root);
          varrun.setEorGroupID(grpWheel);
          varrun.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
          varrun.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
          varrun.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});          
      } // if
      res = new ResourceImpl(
              ((Name) varrun.getName().clone()).extended("apache2.pid"), 
              _core, varrun, "file");
      res.setEorUserID(root);
      res.setEorGroupID(grpWheel); 
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});    

      Directory varlog = (Directory)_core.getEorM_Resources().get(
              ((Name)coreACS.clone()).extended("<files_tree>").extended("var").extended("log"));
      if (varlog == null) {
          var = (Directory)_core.getEorM_Resources().get(
              ((Name)coreACS.clone()).extended("<files_tree>").extended("var"));
          varlog = new DirectoryImpl(
              ((Name)var.getName().clone()).extended("log"),
              _core, var, "directory");
          varlog.setEorUserID(root);
          varlog.setEorGroupID(grpWheel);
          varlog.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
          varlog.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
          varlog.setL_otherRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});         
      } // if 
      
      apache = new DirectoryImpl(
              ((Name)varlog.getName().clone()).extended("apache2"),
              _core, varlog, "directory");
      apache.setEorUserID(root);
      apache.setEorGroupID(  (GroupIDImpl) _core.getEorM_EligibleParties().get(
              ((Name)coreACS.clone()).extended("<G>").extended("adm")  )  );
      apache.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      apache.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("execute")});
      apache.setL_otherRights(new StringRight[] {});      
      
      res = new ResourceImpl(
              ((Name)apache.getName().clone()).extended("access.log.N.gz"),
              _core, apache, "file");
      res.setEorUserID(root);
      res.setEorGroupID(  (GroupIDImpl) _core.getEorM_EligibleParties().get(
              ((Name)coreACS.clone()).extended("<G>").extended("adm")  )  );
      res.setL_ownerRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read"),
          BaseManagerImpl.getInstance().getMetaRight("write")});
      res.setL_groupRights(new StringRight[] {
          BaseManagerImpl.getInstance().getMetaRight("read")});
      res.setL_otherRights(new StringRight[] {});  
      
    } // try
    catch (CreateError e) {
      throw new InternalError(e.toString()+ 
              "BaseManagerImpl.createApacheACS() failed: "+e);
    }
    catch (UpDateError u) {
        throw new InternalError(u.toString()+ 
                "BaseManagerImpl.createApacheACS() failed: "+u);
    }
      
    // ---------------------------
    // create Apache as ACS
    // --------------------------    
    
    */
    
    return _result;
  } // createApache()   
    
  /**
  * Creates the roots of the RBAC application on the Impl ACS. They are only
  * opening and closing roots.
  * Creates the account '_nameLastComponent_appli' and the group '_nameLastComponent_group'.
  * Creates a resources tree with directories, leaves and executables (then actors).
  * Called by createRBAC().
  * @param _name ImmutableName for the ACS to create.
  *   No test of null value nor empty.
  * @param _core is the Linux on which the application is building up, and
  *     it is updated by the initial and standard trees of resources
  *     and eligible parties of the RBAC application
  * @param _factory the RBAC ACS factory
  */
  public static final void configureRBAConLinux(
          ImmutableName _name, ACS _core, ACSFactoryImpl _factory) {

                        // for testing and tracking
                        //System.out.println("call to configureRBAConLinux()");

    if (_core.getACSType().indexOf("Linux") == -1)
        return;

    StringRight[] l_right = new StringRight[] {
                                    _core.getOneRight("|mdown|"),
                                    _core.getOneRight("ctrl")
                                };
    StringRight[] l_right3 = new StringRight[] {
                                    _core.getOneRight("rdir"),
                                    _core.getOneRight("xdir")
                                };    
    StringRight[] l_right5 = new StringRight[] {
                                    _core.getOneRight("r"),
                                    _core.getOneRight("x")
                                };    
    StringRight[] l_right6 = new StringRight[] {
                                    _core.getOneRight("r")
                                }; 
    StringRight[] l_right7 = new StringRight[] {
                                    _core.getOneRight("r"),
                                    _core.getOneRight("w")
                                }; 
    StringRight[] l_right8 = new StringRight[] {
                                    _core.getOneRight("|mdown|"),
                                    _core.getOneRight("r"),
                                    _core.getOneRight("x")
                                }; 
    StringRight[] l_right9 = new StringRight[] {
                                    _core.getOneRight("|mdown|"),
                                    _core.getOneRight("r"),
                                    _core.getOneRight("w")
                                };
 
/*
            // associations to a new personal account: the master
                // personal group creation
            _factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<AssociateClass>.GroupID",
                        "Name.<Value>.<MasterNameLastComponent>",
                        "ACS.<Value>.<MasterValue>",
                        "EP Type.<Value>.personal group",
                        "Is Admin.<Value>.false"});
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));
            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                throw new BaseError("AcsFactory.createACS(): error for "+
                        "<AssociateClass>.GroupID "+l_asso[l_asso.length - 1]);


                        // for testing and tracking
                        //System.out.println("AcsFactory.createACS(), "+
                          //      "l_asso:");
                        //for (int i = 0; i < l_asso.length; i++) {
                          //  System.out.println("-------- "+l_asso[i]);
                        //} // for i                

                // personal group as group of the master
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Action>.1.addEorMember",
                        "GroupMember.<Master>"});
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("UserID.EPType.<ValueContains>.personal account"));


                        // for testing and tracking
                        //System.out.println("AcsFactory.createACS(), "+
                          //      "l_asso:");
                        //for (int i = 0; i < l_asso.length; i++) {
                          //  System.out.println("-------- "+l_asso[i]);
                        //} // for i 

            // immutable personal group
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Action>.1.setEPType",
                        "EP Type.<Value>.<immutable> personal group"});         
                // personal account in 'users'
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Object>."+
                        ((Name)factory.getName()).extended("<G>").extended("users").toString()+
                        "+<Class>+GroupID+<Value>+addEorMember",
                        "GroupMember.<Master>"});            
                // master UMask
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<CalledMethod>.<Master>.setUMask",
                        "UMask.<Value>.--- -w- rwx "});
                // personal directory creation
            factory.addAssociationsToNewObject(
                    "UserID.EPType.<ValueContains>.personal account", 
                    new String[] {
                        "<AssociateClass>.Directory",
                        "Name.<Value>.<MasterNameLastComponent>",
                        "ACS.<Value>.<MasterValue>",
                        "Parent Node.<Value>.<ACSName>+ <files_tree>:: home::  ",
                        "UserID.<Master>",
                        "Resource Type.<Value>.<immutable> directory"});

 */
    // ---------------------------
    // creates RBAC application under Linux
    // --------------------------

    //  creates 'application' group and account
    Name        coreACS = (Name)_core.getName();
    Name        nbRoot = ((Name)coreACS.clone()).extended("<U>").extended("root");
    UserIDImpl  root = null;
    //Name naOne = ((Name)coreACS.clone()).extended(_name.lastComponent()+"_appli");
    //Name naTwo = ((Name)coreACS.clone()).extended(_name.lastComponent()+"_group");
    GroupIDImpl grpOne = null;
    UserIDImpl  useOne = null;
    Resource    res = null;
    Directory   program = null;
    Directory   etc = null;
    Directory   rbac = null;
    Directory   varlog = null;
    Directory   varlib = null;
    Directory   varrun = null;
    Actor       container = null;
    String[]    l_asso = null;
    
    try {
      // 1 - sets grpOne = new GroupIDImpl(naTwo, _core, "<byAcsCreator>", false);
      // and sets ActionNewEligibleParty.handleNewGroup(grpOne);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.GroupID",
                "Name.<Value>."+"<ChildNameLastComponent>"+"_group",
                "ACS.<Value>.<Master>",
                "EP Type.<Value>.<byAcsCreator>",
                "Is Admin.<Value>.false"});

      // 2 - sets grpOne.setOrder(124);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.1.setOrder",
                "Order.<Value>.124"});

      // 3 - useOne = new UserIDImpl(naOne, _core, null, false);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.UserID",
                "Name.<Value>."+"<ChildNameLastComponent>"+"_appli",
                "ACS.<Value>.<Master>",
                "EP Type.<Null>",
                "Is Admin.<Value>.false"});

      // 4 - sets grpOne.setEorL_Members(new GroupIDMemberImpl[] {useOne});
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.1.addEorMember",
                "GroupMember.<ArgumentValue>.3"});

      // 5 - sets useOne.setOrder(245);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.3.setOrder",
                "Order.<Value>.245"});

      // 6 - useOne.setEorPrimGroup(grpOne);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.3.setEorPrimGroup",
                "GroupID.<ArgumentValue>.1"});

      // 7 - useOne.setEPType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.3.setEPType",
                "EP Type.<Value>.<byAcsCreator>"});

    } catch (UpDateError u) {
      throw new InternalError("In AcsFactory.configureRBAConLinux(): "+u);
    }
    
    try {
                // for testing and tracking
                //System.out.println("in AcsFactory.createMysqlACS(), mysql account: "+
                  //useOne);
      
      // creates 'rbac' tree under /etc
      // creates /etc/_name.lastComponent() directory and subdirectories
      
      // 8 - sets rbac = new DirectoryImpl(
              //((Name)etc.getName()).extended(_name.lastComponent()),
              //_core, etc, root, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>."+"<ChildNameLastComponent>",
                "ACS.<Value>.<Master>",
                "Parent Node.<Value>."+
                        "<ACSName>+"+" <files_tree>:: etc::  ",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>"});

      
                          // for testing and tracking
                          //l_asso = (String[]) _factory.getM_OpeningNoClosingRoots().
                            //get( "ACS.Name.<Value>."+_core.getName().toString() );
                          //System.out.println("AcsFactory.configureRBAConLinux(), "+
                            //    "factory.getM_OpeningNoClosingRoots() after "+
                              //  "the creation command 8 for the Directory "+_name.lastComponent());
                          //for (int i = 0; i < l_asso.length; i++) {
                            //  System.out.println("-------- "+l_asso[i]);
                          //} // for i


      // 9 - sets rbac.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.8.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 10 - sets l_right = new StringRight[] {
                                   // _core.getOneRight("|mdown|"),
                                   // _core.getOneRight("ctrl")
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<ACS>.getL_SelectedRights",
                "String Right.<Value>.|mdown|",
                "String Right.<Value>.ctrl",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>"});
              
      // 11 - sets l_right3 = new StringRight[] {
                                   // _core.getOneRight("rdir"),
                                    //_core.getOneRight("xdir"
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<ACS>.getL_SelectedRights",
                "String Right.<Value>.rdir",
                "String Right.<Value>.xdir",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>"});

      // 12 - sets l_right5 = new StringRight[] {
                                    //_core.getOneRight("r"),
                                    //_core.getOneRight("x")
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<ACS>.getL_SelectedRights",
                "String Right.<Value>.r",
                "String Right.<Value>.x",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>"});

      // 13 - sets l_right6 = new StringRight[] {
                                    //_core.getOneRight("r")
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<ACS>.getL_SelectedRights",
                "String Right.<Value>.r",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>"});

      // 14 - sets l_right7 = new StringRight[] {
                                    //_core.getOneRight("r"),
                                    //_core.getOneRight("w")
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<ACS>.getL_SelectedRights",
                "String Right.<Value>.r",
                "String Right.<Value>.w",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>"});

      // 15 - sets l_right9 = new StringRight[] {
                                   // _core.getOneRight("|mdown|"),
                                   // _core.getOneRight("r"),
                                   // _core.getOneRight("w")
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<ACS>.getL_SelectedRights",
                "String Right.<Value>.|mdown|",
                "String Right.<Value>.r",
                "String Right.<Value>.w",
                "String Right.<Null>",
                "String Right.<Null>"});
      
      // 16 - sets l_right10 = new StringRight[0];
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<ACS>.getL_SelectedRights",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>",
                "String Right.<Null>"});

      // 17 - sets rbac.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.8.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 18 - sets rbac.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.8.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 19 - sets rbac.setL_otherRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.8.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 20 - sets rbac.setComment("Directory of a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.8.setComment",
                "Comments.<Value>.Directory of a RBAC application"});

      // 21 - sets rbac.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.8.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});
      
      // 22 - sets res = new DirectoryImpl(
              //((Name) rbac.getName()).extended("conf.d"),
              //_core, rbac, root, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>.conf.d",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.8",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>"});


                          // for testing and tracking
                          //l_asso = (String[]) _factory.getM_OpeningNoClosingRoots().
                            //get( "ACS.Name.<Value>."+_core.getName().toString() );
                          //System.out.println("AcsFactory.configureRBAConLinux(), "+
                            //    "factory.getM_OpeningNoClosingRoots() after "+
                              //  "the creation command 22 for the Directory 'conf.d':");
                          //for (int i = 0; i < l_asso.length; i++) {
                            //  System.out.println("-------- "+l_asso[i]);
                          //} // for i

      // 23 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.22.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 24 - sets res.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.22.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 25 - sets res.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.22.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 26 - sets res.setL_otherRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.22.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 27 - sets res.setComment("Directory of the RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.22.setComment",
                "Comments.<Value>.Directory of a RBAC application"});

      // 28 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.22.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 29 - sets res = new ResourceImpl(
                //((Name) rbac.getName()).extended("debian.cnf"),
                //_core, rbac, root, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Resource",
                "Name.<Value>.debian.cnf",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.8",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 30 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.29.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 31 - sets res.setL_ownerRights(l_right5);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.29.setL_ownerRights",
                "Rights.<ArgumentValue>.12"});

      // 32 - sets res.setL_groupRights(l_right10);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.29.setL_groupRights",
                "Rights.<ArgumentValue>.16"});

      // 33 - sets res.setL_otherRights(l_right10);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.29.setL_otherRights",
                "Rights.<ArgumentValue>.16"});

      // 34 - sets res.setComment("Directory of the RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.29.setComment",
                "Comments.<Value>.for the RBAC application"});

      // 35 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.29.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 36 - sets res = new ResourceImpl(
                //((Name) rbac.getName()).extended("rbac.cnf"),
                //_core, rbac, root, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Resource",
                "Name.<Value>.rbac.cnf",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.8",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 37 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.36.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 38 - sets res.setL_ownerRights(l_right5);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.36.setL_ownerRights",
                "Rights.<ArgumentValue>.12"});

      // 39 - sets res.setL_groupRights(l_right6);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.36.setL_groupRights",
                "Rights.<ArgumentValue>.13"});

      // 40 - sets res.setL_otherRights(l_right6);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.36.setL_otherRights",
                "Rights.<ArgumentValue>.13"});

      // 41 - sets res.setComment("for the RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.36.setComment",
                "Comments.<Value>.for the RBAC application"});

      // 42 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.36.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});

      // creates /usr/share rbac files
      //Directory usrshare = (Directory)_core.getEorM_Resources().get(
        //      ((Name)coreACS.clone()).extended("<files_tree>").extended("usr").extended("share"));   

      // 43 - sets program = new DirectoryImpl(
              //((Name)usrshare.getName().clone()).extended(_name.lastComponent()),
              //_core, usrshare, root, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>."+"<ChildNameLastComponent>",
                "ACS.<Value>.<Master>",
                "Parent Node.<Value>."+
                        "<ACSName>+"+" <files_tree>:: usr:: share::  ",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 44 - sets program.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.43.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 45 - sets program.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.43.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 46 - sets program.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.43.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 47 - sets program.setL_otherRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.43.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 48 - sets program.setComment("for the RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.43.setComment",
                "Comments.<Value>.for the RBAC application"});

      // 49 - sets program.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.43.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});

                          // for testing and tracking
                          //l_asso = (String[]) _factory.getM_OpeningNoClosingRoots().
                            //get( "ACS.Name.<Value>."+_core.getName().toString() );
                          //System.out.println("AcsFactory.configureRBAConLinux(), "+
                            //    "factory.getM_OpeningNoClosingRoots() after "+
                              //  "the creation command 49 for the Directory under usr/share/:");
                          //for (int i = 0; i < l_asso.length; i++) {
                            //  System.out.println("-------- "+l_asso[i]);
                          //} // for i

      // 50 - sets res = new ExecutableImpl(
              // ((Name)program.getName()).extended("debian-start.inc.sh"),
              // _core, program, root, "<byAcsCreator>", null);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Executable",
                "Name.<Value>.debian-start.inc.sh",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.43",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>",
                "ACS Bridge Target.<Null>"});


                          // for testing and tracking
                          //l_asso = (String[]) _factory.getM_OpeningNoClosingRoots().
                            //get( "ACS.Name.<Value>."+_core.getName().toString() );
                          //System.out.println("AcsFactory.configureRBAConLinux(), "+
                            //    "factory.getM_OpeningNoClosingRoots() after "+
                              //  "the creation command 50 for the Executable debian-start.inc.sh:");
                          //for (int i = 0; i < l_asso.length; i++) {
                            //  System.out.println("-------- "+l_asso[i]);
                          //} // for i


      // 51 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.50.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 52 - sets res.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.50.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 53 - sets res.setL_groupRights(l_right5);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.50.setL_groupRights",
                "Rights.<ArgumentValue>.12"});

      // 54 - sets res.setL_otherRights(l_right5);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.50.setL_otherRights",
                "Rights.<ArgumentValue>.12"});

      // 55 - sets res.setComment("for the RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.50.setComment",
                "Comments.<Value>.for the RBAC application"});

      // 56 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.50.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 57 - sets res = new ExecutableImpl(
              //((Name)program.getName()).extended("rbac_multi.server"),
              //_core, program, root, "<byAcsCreator>", null);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Executable",
                "Name.<Value>.rbac_multi.server",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.43",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>",
                "ACS Bridge Target.<Null>"});

      // 58 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.57.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 59 - sets res.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.57.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 60 - sets res.setL_groupRights(l_right5);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.57.setL_groupRights",
                "Rights.<ArgumentValue>.12"});

      // 61 - sets res.setL_otherRights(l_right5);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.57.setL_otherRights",
                "Rights.<ArgumentValue>.12"});

      // 62 - sets res.setComment("for the RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.57.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 63 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.57.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 64 - sets rbac = new DirectoryImpl(
              //((Name)usrshare.getName()).extended(_name.lastComponent()+"-common"),
              //_core, usrshare, root, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>."+"<ChildNameLastComponent>"+"-common",
                "ACS.<Value>.<Master>",
                "Parent Node.<Value>."+
                        "<ACSName>+"+" <files_tree>:: usr:: share::  ",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 65 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.64.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 66 - sets res.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.64.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 67 - sets res.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.64.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 68 - sets res.setL_otherRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.64.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 69 - sets res.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.64.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 70 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.64.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 71 - sets program = new DirectoryImpl(
                    //((Name)rbac.getName()).extended("internal-use-only"),
                    //_core, rbac, root, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>.internal-use-only",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.64",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 72 - sets program.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.71.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 73 - sets program.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.71.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 74 - sets program.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.71.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 75 - sets program.setL_otherRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.71.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 76 - sets program.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.71.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 77 - sets program.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.71.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 78 - sets res = new ExecutableImpl(
                //((Name)program.getName()).extended("_etc_init.d_rbac"),
                //_core, program, root, "<byAcsCreator>", null);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Executable",
                "Name.<Value>._etc_init.d_rbac",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.71",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>",
                "ACS Bridge Target.<Null>"});

      // 79 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.78.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 80 - sets res.setL_ownerRights(l_right9);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.78.setL_ownerRights",
                "Rights.<ArgumentValue>.15"});

      // 81 - sets res.setL_groupRights(l_right6);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.78.setL_groupRights",
                "Rights.<ArgumentValue>.13"});

      // 82 - sets res.setL_otherRights(l_right6);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.78.setL_otherRights",
                "Rights.<ArgumentValue>.13"});

      // 83 - sets res.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.78.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 84 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.78.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 85 - sets res = new ExecutableImpl(
                //((Name)program.getName()).extended("_etc_rbac_debian-start"),
                //_core, program, root, "<byAcsCreator>", null);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Executable",
                "Name.<Value>._etc_rbac_debian-start",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.71",
                "UserID.<Value>."+
                        "<ACSName>+"+" <U>:: root::  ",
                "Resource Type.<Value>.<byAcsCreator>",
                "ACS Bridge Target.<Null>"});

      // 86 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.85.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 87 - sets res.setL_ownerRights(l_right9);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.85.setL_ownerRights",
                "Rights.<ArgumentValue>.15"});

      // 88 - sets res.setL_groupRights(l_right6);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.85.setL_groupRights",
                "Rights.<ArgumentValue>.13"});

      // 89 - sets res.setL_otherRights(l_right6);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.85.setL_otherRights",
                "Rights.<ArgumentValue>.13"});

      // 90 - sets res.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.85.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 91 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.85.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});

      // get /var/lib, /var/run and /var/log rbac directories
      //varlib = (Directory)_core.getEorM_Resources().get(
        //      ((Name)coreACS.clone()).extended("<files_tree>").extended("var").extended("lib"));

      // 92 - sets program = new DirectoryImpl(
                //((Name)varlib.getName()).extended(_name.lastComponent()),
                //_core, varlib, useOne, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>."+"<ChildNameLastComponent>",
                "ACS.<Value>.<Master>",
                "Parent Node.<Value>."+
                        "<ACSName>+"+" <files_tree>:: var:: lib::  ",
                "UserID.<ArgumentValue>.3",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 93 - sets program.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.92.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 94 - sets program.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.92.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 95 - sets program.setL_groupRights(l_right11);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.92.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 96 - sets program.setL_otherRights(l_right11);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.92.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 97 - sets program.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.92.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 98 - sets program.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.92.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 99 - sets res = new ResourceImpl(
                //((Name)program.getName()).extended("ibdata1"),
                //_core, program, useOne, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Resource",
                "Name.<Value>.ibdata1",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.92",
                "UserID.<ArgumentValue>.3",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 100 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.99.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 101 - sets res.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.99.setL_ownerRights",
                "Rights.<ArgumentValue>.15"});

      // 102 - sets res.setL_groupRights(l_right11);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.99.setL_groupRights",
                "Rights.<ArgumentValue>.14"});

      // 103 - sets res.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.99.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 104 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.99.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 105 - sets res = new ResourceImpl(
                //((Name)program.getName()).extended("ib_logfileX"),
                //_core, program, useOne, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Resource",
                "Name.<Value>.ib_logfileX",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.92",
                "UserID.<ArgumentValue>.3",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 106 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.105.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 107 - sets res.setL_ownerRights(l_right9);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.105.setL_ownerRights",
                "Rights.<ArgumentValue>.15"});

      // 108 - sets res.setL_groupRights(l_right7);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.105.setL_groupRights",
                "Rights.<ArgumentValue>.14"});

      // 109 - sets res.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.105.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 110 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.105.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 111 - sets rbac = new DirectoryImpl(
                //((Name)program.getName()).extended("rbac"),
                //_core, program, useOne, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>.rbac",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.92",
                "UserID.<ArgumentValue>.3",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 112 - sets rbac.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.111.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 113 - sets rbac.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.111.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 114 - sets rbac.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.111.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 115 - sets rbac.setL_otherRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.111.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 116 - sets rbac.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.111.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 117 - sets rbac.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.111.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});


      // 118 - sets res = new ResourceImpl(
                //((Name)rbac.getName()).extended("<SQL_bases_&_tables_descriptions>"),
                //_core, rbac, useOne, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Resource",
                "Name.<Value>.<SQL_bases_&_tables_descriptions>",
                "ACS.<Value>.<Master>",
                "Parent Node.<ArgumentValue>.111",
                "UserID.<ArgumentValue>.3",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 119 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.118.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 120 - sets res.setL_ownerRights(l_right9);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.118.setL_ownerRights",
                "Rights.<ArgumentValue>.15"});

      // 121 - sets res.setL_groupRights(l_right7);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.118.setL_groupRights",
                "Rights.<ArgumentValue>.14"});

      // 122 - sets res.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.118.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 123 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.118.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});

      //varrun = (Directory)_core.getEorM_Resources().get(
        //      ((Name)coreACS.clone()).extended("<files_tree>").extended("var").extended("run"));

      // 124 - sets res = new ExecutableImpl(
                //((Name)varrun.getName()).extended(_name.lastComponent()+"d"),
                //_core, varrun, useOne, "<byAcsCreator>", null);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Executable",
                "Name.<Value>."+"<ChildNameLastComponent>"+"d",
                "ACS.<Value>.<Master>",
                "Parent Node.<Value>."+
                        "<ACSName>+"+" <files_tree>:: var:: run::  ",
                "UserID.<ArgumentValue>.3",
                "Resource Type.<Value>.<byAcsCreator>",
                "ACS Bridge Target.<Null>"});

      // 125 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.124.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 126 - sets res.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.124.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 127 - sets res.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.124.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 128 - sets res.setL_otherRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.124.setL_otherRights",
                "Rights.<ArgumentValue>.11"});

      // 129 - sets res.setComment("main program of a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.124.setComment",
                "Comments.<Value>.main program of a RBAC application"});

      // 130 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.124.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});

      //varlog = (Directory)_core.getEorM_Resources().get(
        //      ((Name)coreACS.clone()).extended("<files_tree>").extended("var").extended("log"));

      // 131 - sets res = new DirectoryImpl(
                //((Name)varlog.getName()).extended(_name.lastComponent()),
                //_core, varlog, useOne, "<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<AssociateClass>.Directory",
                "Name.<Value>."+"<ChildNameLastComponent>",
                "ACS.<Value>.<Master>",
                "Parent Node.<Value>."+
                        "<ACSName>+"+" <files_tree>:: var:: log::  ",
                "UserID.<ArgumentValue>.3",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 132 - sets res.setResourceType("<12345678901234567890>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.131.setResourceType",
                "Resource Type.<Value>.<12345678901234567890>"});

      // 133 - sets res.setEorGroupID(  (GroupIDImpl) _core.getEorM_EligibleParties().get(
                //((Name)coreACS.clone()).extended("<G>").extended("adm")  )  );
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.131.setEorGroupID",
                "GroupID.<Value>."+
                    "<ACSName>+"+" <G>:: adm::  " });

      // 134 - sets res.setL_ownerRights(l_right);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.131.setL_ownerRights",
                "Rights.<ArgumentValue>.10"});

      // 135 - sets res.setL_groupRights(l_right3);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.131.setL_groupRights",
                "Rights.<ArgumentValue>.11"});

      // 136 - sets res.setStickyBit(true);
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.131.setStickyBit",
                "Sticky bit.<Value>.true"});

      // 137 - sets res.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.131.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 138 - sets res.setResourceType("<byAcsCreator>");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.131.setResourceType",
                "Resource Type.<Value>.<byAcsCreator>"});

      // 139 - sets grpOne.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.1.setComment",
                "Comments.<Value>.for a RBAC application"});

      // 140 - sets useOne.setComment("for a RBAC application");
      _factory.addOpeningNoClosingRoots(
            "ACS.Name.<Value>."+_core.getName().toString(),
            new String[] {
                "<CalledMethod>.<Action>.3.setComment",
                "Comments.<Value>.for a RBAC application"});

                          // for testing and tracking
                          //l_asso = (String[]) _factory.getM_OpeningNoClosingRoots().
                            //get( "ACS.Name.<Value>."+_core.getName().toString() );
                          //System.out.println("AcsFactory.configureRBAConLinux(), "+
                            //    "factory.getM_OpeningNoClosingRoots() after "+
                              //  "the last creation command 138:");
                          //for (int i = 0; i < l_asso.length; i++) {
                            //  System.out.println("-------- "+l_asso[i]);
                          //} // for i
      
    } // try
    catch (UpDateError u) {
        u.printStackTrace();
        throw new InternalError("AcsFactory.configureRBAConLinux() failed ");
    }

    l_right = null;
    l_right3 = null;     
    l_right5 = null;    
    l_right6 = null; 
    l_right7 = null; 
    l_right9 = null;    
    
    coreACS = null;
    nbRoot = null;
    root = null;
    //naOne = null;
    //naTwo = null;
    grpOne = null;
    useOne = null; 
    program = null;
    etc = null;
    rbac = null; 
    res = null;
    program = null;
    varlog = null;
    varrun = null;
    varlib = null;
  } // configureRBAConLinux()
  
} // AcsFactory
