//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Collections;
import java.util.Arrays;

import java.text.StringCharacterIterator;
import java.text.CharacterIterator;

/**
* This class is a utility class for the MySQL ACS,
* providing synchronized static methods to sort the AclEntry list and to set
* the internal maps of the inherited AclEntry rights and the primary rights. <P>
* Generally speaking, the sorting for ACLs having a different parent level is independant.
* For the same parent level, the ACL are sorted first from the more-specific sources
* to the least-specific ones, then they are sorted with a similar logic for the hosts
* (the ACL condition group) and finally, with a special sorting for the bases sets, if any.
* This sorting is both for direct AclEntries and inherited AclEntries.
* The method addInSortedAclList) inserts a new AclEntry into the ACL list,
* at the right MySQL sorting, and for a given access target.
* The method handles three independant cases: <P>
* (1) if the _acl target is not a database-specific target (database or base set), <P>
* (2) if the _acl target is a database-specific target and _acl is NOT a to-be-mixed ACL, <P>
* (3) if the _acl target is a database-specific target and _acl is a to-be-mixed ACL. <P>
* This method is so a commutator for the methods addAclNOTForDatabase(),
* addAclForDatabaseNOTForToBeMixedAcl() and addAclForDatabaseForToBeMixedAcl().
* This sorting drives the selection process of the first matched AclEntry(ies) at a parent level,
* for a couple user/host and a target. The ACL selection is the responsability
* of the methods NodeRightsMySQLImpl.detectL_aclEntryRights() and
* getL_accessRightsThroughNodesTree(), respectively for the direct and inherited ACL.<P>
* Important: Sorting the host (or the hosts group) first, then sorting the base
* (or the bases set) does NOT fulfill with the MySQL documentation at
* http://dev.mysql.com/doc/refman/5.6/en/request-access.html, because the true
* behavior of the MySQL Server is to sort the hosts first.
* To go further, see the documentation of the method addInSortedAclList(). <P>
* For having some rights on a MySQL Resource, it is mandatory for a pair (user, host)
* to match on the MySQL Server. This means to match on the Resources root 'MySQL server',
* a pair (user, host) or a pair (user, hosts group) which has at least an USAGE right.
* The first matching pair becomes the current user, and it is used to get
* the rights for the requesting pair (user, host).
* The current user is returned by the method NodeRightsMySQLImpl.getConnectionPair().
* If the current user has only an USAGE right and if, for a given Resource,
* the current user has no direct ACL and no inherited ACL, then no right is granted
* to the pair (user, host). Otherwise, the rights are the sum of the current user
* rights from the first matched direct ACL and from the inherited ACL. For these
* inherited ACL, the rights from each parent are added to the final result. A given
* parent delivers the rights of the first matched inherited ACL, for the current user.
* This ACL may be real or virtual, where the real ACL are the classical ACL displayed
* in the beamer lists. There are two types of virtual ACL: the virtual
* 'user@accounts-for-hosts group' ACL and the mixed ACL. <P>
* To take in account the ACLs from different parent level, the inherited ACLs of
* a Resource are simply the full copy of all the ACL of its direct and indirect parents, 
* including for instance all the direct ACL on the database. <P>
* For the inherited AclEntries, the MySQL Resource description contains
* a specific map for the primary rights. This is not a MySQL concept.
* The primary rights for each applicable user@host pair are specically created to help
* the user to better understand both the current user and the complex database rights.
* It is the responsibility of DirectoryMySQLImpl to create and edit the primary rights.
* To go further about the inherited ACL rights, see the documentation of the method
* setM_TargetInheritedACLRights(). The documentation of the method 
* DirectoryMySQLImpl.addAclEntry() describes in details the calling sequence
* of all these methods when an ACL is created, including DirectoryMySQLImpl,
* BasesSetMySQLImpl, ResourceMySQLImpl and AclEntryImpl. <P>
* The methods of the superclass are not called by the consumers of this class.
* The calls to this class are reserved to this package, even if
* its methods are public since they overread the super methods.
* Caution: this class is not defined in ACSFactoryUtilityImpl.M_GENERIC_ACS_VOCABULARY,
* then it is not possible to get it outside this package through
* ARoad0.gBaseInterface.ACSRun.getM_AcsVocabulary().
* This class has no subclass in this package.
* @version 0.7.0
*/
public class BaseUtilityMySQLImpl extends BaseUtilityImpl {

  static {
      instance__ = new BaseUtilityMySQLImpl();
  }

    // Constructors

  /**
   * Public constructor
  */
  public BaseUtilityMySQLImpl() {
  }

    /**
    * Only method to obtain the unique instance of BaseUtilityMySQLImpl.
    * @return a BaseUtilityMySQLImpl instance
    */
  public static BaseUtilityImpl getInstance() {
    return instance__;
  } // getInstance()

    /**
    * This method inserts an inherited AclEntry into the inherited ACL list,
    * at the right MySQL sorting, for a given resource.
    * It is a simple adaptor for addInSortedAclList() to set the good list of ACL
    * as argument.
    * Called by ResourceMySQL.addInheritedAclEntry() in all its implementations.
    * Calls addInSortedAclList() which does the main job.
    * @see #addInSortedAclList(java.util.ArrayList, ARoad0.gBaseInterface.AclEntry)
    * @param _l_acl inherited AclEntry sorted list. May be empty but not null.
    * @param _acl inherited AclEntry to put in the list at the right order. Never null.
    * @return the new inherited AclEntry sorted list
    */
  public static ArrayList<AclEntry> addInSortedInheritedAclList(
          ArrayList<AclEntry> _l_acl, AclEntry _acl) {

                                      // for testing and tracking
                                      //System.out.println(" ");
                                      //System.out.println("BaseUtilityMySQLImpl.addInSortedInheritedAclList() start"+
                                        //      " for ACL "+_acl);

    if (_l_acl.isEmpty()) {
        _l_acl.add(_acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addInSortedInheritedAclList() ends "+
                                    //      ", with one ACL "+_acl);

        return _l_acl;
    } // if

    ArrayList<AclEntry>     l_acl = new ArrayList(50);
    ArrayList<AclEntry>     l_result = new ArrayList(50);
    ArrayList<AclEntry>     l_lowerParentAcl = new ArrayList(50);
    ImmutableName           resName = _acl.getResName();

    if (_acl.getResType().indexOf("database") == -1 &&
            _acl.getResType().indexOf("pattern-name_bases set") == -1) {

        for (AclEntry acl : _l_acl) {
            if (resName.equals(acl.getResName()))
                l_acl.add(acl);
            else if (
                    // 'acl' resource is a parent of '_acl' resource
                    resName.contains(acl.getResName()) ||
                    // 'acl' is a bases set ACL to set before a non-root ACL
                    (_acl.getResType().indexOf("MySQL server") == -1 &&
                        acl.getResType().indexOf("bases set") != -1) )
                l_result.add(acl);
            else
                l_lowerParentAcl.add(acl);
        } // for acl

        // l_acl contains only ACL for the same resource
        l_acl = addInSortedAclList(l_acl, _acl);
        l_result.addAll(l_acl);
        l_result.addAll(l_lowerParentAcl);

    }
    else {
        // for the databases
        for (AclEntry acl : _l_acl) {
            if (acl.getResType().indexOf("database") != -1 ||
              acl.getResType().indexOf("pattern-name_bases set") != -1)
                l_acl.add(acl);
            else if (acl.getResType().indexOf("MySQL server") != -1)
                l_result.add(acl);
            else
                l_lowerParentAcl.add(acl);
        } // for

        // l_acl contains only the ACL for the same database
        l_acl = addInSortedAclList(l_acl, _acl);
        l_result.addAll(l_acl);
        l_result.addAll(l_lowerParentAcl);

    } // else

    l_acl = null;
    l_lowerParentAcl = null;
    return l_result;
  } // addInSortedInheritedAclList()

    /**
    * This important method inserts an AclEntry into the ACL list, at the right MySQL sorting,
    * for a given access target. This sorting is only for the list of the direct AclEntries,
    * and then, the argument _upd_l_acl contains only such AclEntries.
    * After the call to this method, the list of the inherited AclEntries is based on
    * the sorting of the ACL targets to insert at the right place each sorted list of direct ACL.<P>
    * This method handles three independant cases: <P>
    * (1) if the _acl target is not a database-specific target (database or base set), <P>
    * (2) if the _acl target is a database-specific target and _acl is NOT a to-be-mixed ACL, <P>
    * (3) if the _acl target is a database-specific target and _acl is a to-be-mixed ACL. <P>
    * This method is then a commutator for the methods addAclNOTForDatabase(),
    * addAclForDatabaseNOTForToBeMixedAcl() and addAclForDatabaseForToBeMixedAcl().
    * The detailed sequence of calls for adding an ACL is described at the method
    * DirectoryMySQLImpl.addAclEntry(). <P>
    * Called by addInSortedInheritedAclList() and ResourceMySQLImpl.addInheritedAclEntry()
    * and addAclEntry(), DirectoryMySQLImpl and ActorMySQLImpl.addAclEntry().
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see ARoad0.AcsAddon.Accbee.MySQL.gWork.NodeRightsMySQLImpl#getL_accessRightsThroughNodesTree(ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see #setM_TargetInheritedACLRights(ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.ResourceMySQL, java.util.List, java.util.Map, java.util.Map, boolean) 
    * @see #getPreviousSortedTypesOfResourceInSortedAclList(java.util.ArrayList, java.lang.String)
    * @see #addAclNOTForDatabase(java.util.ArrayList, ARoad0.gBaseInterface.AclEntry)
    * @see #addAclForDatabaseNOTForToBeMixedAcl(java.util.ArrayList, ARoad0.gBaseInterface.AclEntry)
    * @see #addAclForDatabaseNOTForToBeMixedAcl(java.util.ArrayList, ARoad0.gBaseInterface.AclEntry) 
    * @param _upd_l_acl direct AclEntry sorted list, with all ACL not for a database-specific target
    * or with all ACL for a database-specific target, following the nature of _acl.
    * May be empty but not null.
    * @param _acl direct AclEntry to put in the list at the right order. Never null.
    * @return the updated _upd_l_acl
    */
  public static ArrayList<AclEntry> addInSortedAclList(
          ArrayList<AclEntry> _upd_l_acl, AclEntry _acl) {

                                      // for testing and tracking
                                      //System.out.println(" ");
                                      //System.out.println("BaseUtilityMySQLImpl.addInSortedAclList() start"+
                                        //      " for ACL "+_acl);
      
    if (_upd_l_acl.isEmpty()) {
        _upd_l_acl.add(_acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addInSortedAclList() ends "+
                                    //      ", with one ACL "+_acl);

        return _upd_l_acl;
    }

    if (_acl.getResType().indexOf("database") == -1 &&
            _acl.getResType().indexOf("pattern-name_bases set") == -1) {

        for (AclEntry acl : _upd_l_acl) {

            if (acl.getResType().indexOf("database") != -1 ||
              acl.getResType().indexOf("pattern-name_bases set") != -1)
                throw new InternalError("BaseUtilityMySQLImpl.addInSortedAclList() "+
                        "where the list contains a wrong acl for a database-specific target: "+
                        acl);
        } // for

        return addAclNOTForDatabase(_upd_l_acl, _acl);

    }
    else if (_acl.isConditionalAclEntry() ||
            _acl.getEorEP().getEPType().indexOf("pattern-name_accounts-for-hosts group") != -1) {
        
        for (AclEntry acl : _upd_l_acl) {
            
            if (acl.getResType().indexOf("database") == -1 &&
              acl.getResType().indexOf("pattern-name_bases set") == -1)
                throw new InternalError("BaseUtilityMySQLImpl.addInSortedAclList() "+
                        "where the list contains a wrong acl, "+
                        "that is not for a database-specific target: "+
                        acl);
            //if (!acl.isConditionalAclEntry() &&
              //acl.getEorEP().getEPType().indexOf("pattern-name_accounts-for-hosts group") == -1)
                //throw new InternalError("BaseUtilityMySQLImpl.addInSortedAclList() "+
                  //      "where the list contains a wrong acl, "+
                    //    "that is a to-be-mixed ACL: "+
                      //  acl);
        } // for acl

        return addAclForDatabaseNOTForToBeMixedAcl(_upd_l_acl, _acl);

    }
    else {
        if (_acl.isConditionalAclEntry())
                throw new InternalError("BaseUtilityMySQLImpl.addInSortedAclList() "+
                        "where the list contains a wrong ACL argument, "+
                        "that is conditional and to-be-mixed ACL: "+
                        _acl);

        for (AclEntry acl : _upd_l_acl) {

            if (acl.getResType().indexOf("database") == -1 &&
              acl.getResType().indexOf("pattern-name_bases set") == -1)
                throw new InternalError("BaseUtilityMySQLImpl.addInSortedAclList() "+
                        "where the list contains a wrong acl, "+
                        "that is not for a database-specific target: "+
                        acl);
            //if (acl.isConditionalAclEntry() ||
              //acl.getEorEP().getEPType().indexOf("pattern-name_accounts-for-hosts group") != -1)
                //throw new InternalError("BaseUtilityMySQLImpl.addInSortedAclList() "+
                  //      "where the list contains a wrong acl, "+
                    //    "that is not a to-be-mixed ACL: "+
                      //  acl);
        } // for acl

        return addAclForDatabaseForToBeMixedAcl(_upd_l_acl, _acl);
    } // else
  } // addInSortedAclList()

    /**
    * This method inserts an AclEntry into the ACL list, following the MySQL sorting,
    * when the AclEntry has a given access target that is NOT a database or a bases set,
    * just like the ACL targets in the list. This sorting is only for the list
    * of the direct AclEntries, and then, the argument _upd_l_acl contains only such AclEntries.
    * This method is for the ACL target 'MySQL server' and all the children of a database. <P>
    * The 4 sorting rules are the followings, in this order: <P>
    * (1) among the ACL targets, the ACL for the parent before the ACL for the child, <P>
    * (2) the conditional ACL before the non-conditional ACL, so that the ACL
    * for accounts-for-hosts groups are at the end, <P>
    * (3) among the conditional ACL, the ACL of the 'anonymous' account after the ACL
    * for other accounts, <P>
    * (4) among the non-conditional ACL, which are all for accounts-for-hosts groups,
    * the hosts group sorting is applied (there is no order among the accounts). <P>
    * The first order is implemented by this method, but it is not used because the calling method
    * manages it to put only direct ACL in the argument _upd_l_acl.
    * The third and fourth order are indeed at the same priority.
    * An accounts-for-hosts group never handles a simple host, but always a hosts group.
    * The accounts-for-hosts group name has the format 'xxxx§@yyyyy', where 'yyyyy'
    * is a hosts group name last component which contains exactly one wildcard '_' or '%'.
    * The hosts group sorting is complex. A typical example of a sorted list for hosts groups,
    * all matching 'localhost', is the following hosts group last components: <P>
    * 'localhos_', '_ocalhost', 'localhost%', 'localhos%', 'lo%', 'lo%t', 'lo%ost',
    * 'l%t', '%', '%ocalhost', '%t'. <P>
    * For sorting the hosts groups of the accounts-for-hosts groups,
    * the method isAfterAsPatternNameHostsGroup() is called.
    * The detailed sequence of calls for adding an ACL is described at the method
    * DirectoryMySQLImpl.addAclEntry(). Called by addInSortedAclList().
    * Calls getPreviousSortedTypesOfResourceInSortedAclList(),
    * getTypeOfResourceLastOrderInSortedAclList(),
    * getTypeOfResourceFirstOrderInSortedAclList(),
    * isAfterAsPatternNameHostsGroup().
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see ARoad0.AcsAddon.Accbee.MySQL.gWork.NodeRightsMySQLImpl#getL_accessRightsThroughNodesTree(ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see #isAfterAsPatternNameHostsGroup(java.lang.String, java.lang.String)
    * @param _upd_l_acl direct AclEntry sorted list, where the target is not
    * a database-specific target. May be empty but not null.
    * @param _acl direct AclEntry to put in the list at the right order. Never null
    * and not into _upd_l_acl.
    * @return the updated _upd_l_acl
    */
  protected static ArrayList<AclEntry> addAclNOTForDatabase(
          ArrayList<AclEntry> _upd_l_acl, AclEntry _acl) {

                                      // for testing and tracking
                                      //System.out.println(" ");
                                      //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() start"+
                                        //      " for ACL "+_acl);

    // to implement: (1) among the ACL targets, the ACL for the parent before the ACL for the child
    int                 firstOrder = 0;
    // to implement: (2) the conditional ACL before the non-conditional ACL
    int                 secondOrder = 0;
    // to implement: (3) among the conditional ACL, the ACL of the 'anonymous' account after the ACL
    // for other accounts
    int                 thirdOrder = 0;
    // to implement: (4) among the non-conditional ACL (there are all for accounts-for-hosts groups),
    // the hosts group sorting is applied
    int                 fourthOrder = 0;
    int                 resTypeOrder = -1;
    String              aclResType = _acl.getResType();
    String              resType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(aclResType);
    String              rAclType = null;

    // sorted list of the resource types which are in the AclEntry list
    // and which are before resType in the target-oriented sorting.
    ArrayList<String>   l_previousTypes = getPreviousSortedTypesOfResourceInSortedAclList(
                                                _upd_l_acl, resType);
    int                 lastResTypeOrder = -1;
    AclEntry            acl = null;
    boolean             isConditional = _acl.isConditionalAclEntry();
    boolean             stopAtSecondOrder = false;
    String              hostGroupOne = null;
    String              hostGroupTwo = null;

    if (!l_previousTypes.isEmpty()) {
        // the order of the last AclEntry in _upd_l_acl having a Resource
        // with the last order before resType in its type, or -1 if not found
        lastResTypeOrder = getTypeOfResourceLastOrderInSortedAclList(
                            _upd_l_acl, l_previousTypes.get(l_previousTypes.size() - 1));

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclNOTForDatabase() with non-nulll_previousTypes "+
                                      //    l_previousTypes.get(l_previousTypes.size() - 1)+
                                        //  " and lastResTypeOrder = "+lastResTypeOrder);
                                  
    } // if

    // the order of the first AclEntry in _upd_l_acl having a Resource
    // with resType in its type, or -1 if not found
    resTypeOrder = getTypeOfResourceFirstOrderInSortedAclList(_upd_l_acl, resType);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclNOTForDatabase() with resTypeOrder = "+resTypeOrder+
                                      //    ", lastResTypeOrder = "+lastResTypeOrder+
                                        //  " and _upd_l_acl size = "+_upd_l_acl.size());

    if (resTypeOrder != -1) {
        firstOrder = resTypeOrder;
    } // if
    else if (lastResTypeOrder != -1 && _upd_l_acl.size() != lastResTypeOrder + 1) {
        firstOrder = lastResTypeOrder + 1;
    } // else
    else if (lastResTypeOrder != -1  && _upd_l_acl.size() == lastResTypeOrder + 1) {

        _upd_l_acl.add(lastResTypeOrder + 1, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() ends"+
                                    //      ", with only the first logic");
                                  //System.out.println("-- with l_previousTypes.get(l_previousTypes.size() - 1) "+
                                    //      l_previousTypes.get(l_previousTypes.size() - 1)+
                                      //    " with lastResTypeOrder "+lastResTypeOrder+
                                        //  ", with _upd_l_acl.size() "+_upd_l_acl.size());
                                  //System.out.println("-- with result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        aclResType = null;
        l_previousTypes = null;
        resType = null;;
        rAclType = null;
        acl = null;
        return _upd_l_acl;
    } // else
    else {
      // when (resTypeOrder == -1 && lastResTypeOrder == -1)
        aclResType = null;
        l_previousTypes = null;
        resType = null;
        rAclType = null;
        acl = null;

        _upd_l_acl.add(0, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() ends "+
                                    //      ", with only the first logic");
                                  //System.out.println("-- with resTypeOrder "+resTypeOrder+
                                    //      ", with lastResTypeOrder "+lastResTypeOrder+
                                      //    ", with _upd_l_acl.size() "+_upd_l_acl.size());
                                  //System.out.println("-- with result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // else

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclNOTForDatabase() with resTypeOrder = "+resTypeOrder+
                                      //    ", lastResTypeOrder = "+lastResTypeOrder+
                                        //  " and firstOrder = "+firstOrder);

    // if only the first logic to apply, with firstOrder = lastResTypeOrder + 1
    if (resTypeOrder == -1) {
        aclResType = null;
        l_previousTypes = null;
        resType = null;
        rAclType = null;
        acl = null;

        _upd_l_acl.add(firstOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() ends "+
                                    //      ", with only the first logic, resTypeOrder == -1, and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() for "+
                                    //      "the second logic. ");


    // 'firstOrder' is set, and there is an ACL at this position.
    // The second logic is from the condition group AclEntry order, which sets
    // the conditional ACL before the non-conditional ACL.
    acl = _upd_l_acl.get(firstOrder);
    if (!isConditional && acl.isConditionalAclEntry()) { // otherwise, secondOrder == 0

        for (Iterator ite = _upd_l_acl.listIterator(firstOrder); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclNOTForDatabase() for the secondOrder, "+
                                     //     "process the ACL "+acl);


            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

            // search on the same target type
            if (!rAclType.equals(resType)) {
                break;
            }

            if (!isConditional && acl.isConditionalAclEntry()) {
                secondOrder++;
                continue;
            }
            else if (isConditional && !acl.isConditionalAclEntry()) {
                // no third or fourth order to process
                stopAtSecondOrder = true;
                break;
            }
            else {
                // third or fourth order to process
                break;
            }

        } // for ite

    } // if (!isConditional

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclNOTForDatabase() with firstOrder = "+firstOrder+
                                      //    " and secondOrder = "+secondOrder);

    // if only the first and second logic to apply
    if (firstOrder + secondOrder == _upd_l_acl.size() || stopAtSecondOrder) {
        aclResType = null;
        l_previousTypes = null;
        resType = null;
        rAclType = null;
        acl = null;

        _upd_l_acl.add(firstOrder + secondOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() "+
                                    //      "ends, with only the first and second logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if
    
    // 'firstOrder' and 'secondOrder' are set, and there is an ACL at (firstOrder + secondOrder).
    // The third logic is among the conditional ACL, to put the ACL of the 'anonymous' account
    // after the ACL for other accounts
    acl = _upd_l_acl.get(firstOrder + secondOrder);
    if (isConditional && acl.isConditionalAclEntry() &&
            _acl.getEpName().lastComponent().equals("anonymous")) { // otherwise, thirdOrder == 0

        for (Iterator ite = _upd_l_acl.listIterator(firstOrder + secondOrder); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclNOTForDatabase() for the thirdOrder, "+
                                      //    "process the ACL "+acl);


            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

            // search on the same target type
            if (!rAclType.equals(resType)) {
                break;
            }

            if (acl.isConditionalAclEntry()) {
                thirdOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (!isConditional


                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() for "+
                                    //      "the third and fourth logics. ");

    // if third logic to apply, no fourth logic
    if (isConditional) {
        aclResType = null;
        l_previousTypes = null;
        resType = null;
        rAclType = null;
        acl = null;

        _upd_l_acl.add(firstOrder + secondOrder + thirdOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() "+
                                    //      "ends, with only the first, second and third logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // 'firstOrder' and 'secondOrder' are set, and there is an ACL at (firstOrder + secondOrder).
    // The fourth logic is among the non-conditional ACL (there are all for accounts-for-hosts groups),
    // to apply the hosts group sorting.
    if (!isConditional && !acl.isConditionalAclEntry()) { // otherwise, fourthOrder == 0

        hostGroupOne = _acl.getEpName().lastComponent().substring(
                _acl.getEpName().lastComponent().indexOf("§@") + 2);

        for (Iterator ite = _upd_l_acl.listIterator(firstOrder + secondOrder); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclNOTForDatabase() for the fourthOrder, "+
                                     //     "process the ACL "+acl);


            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

            // search on the same target type
            if (!rAclType.equals(resType)) {
                break;
            }

            if (!isConditional && !acl.isConditionalAclEntry()) {

                hostGroupTwo = acl.getEpName().lastComponent().substring(
                        acl.getEpName().lastComponent().indexOf("§@") + 2);

                if (isAfterAsPatternNameHostsGroup(hostGroupOne, hostGroupTwo)) {
                    fourthOrder++;
                    continue;
                }
                else {
                    break;
                }
            }
            else {
                break;
            } // else

        } // for ite

    } // if (!isConditional

    _upd_l_acl.add(firstOrder + secondOrder + fourthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclNOTForDatabase() ends "+
                                    //      "after the fourth sorting on hosts groups, with result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

    aclResType = null;
    l_previousTypes = null;
    resType = null;
    rAclType = null;
    hostGroupOne = null;
    hostGroupTwo = null;
    
    return _upd_l_acl;
  } // addAclNOTForDatabase()

    /**
    * This method inserts an AclEntry into the ACL list, following the MySQL sorting,
    * when the _acl target is a database-specific target and _acl is NOT a to-be-mixed ACL.
    * This sorting is only for the list of the direct AclEntries,
    * and then, the argument _upd_l_acl contains only such AclEntries for a database-specific target.
    * The _acl target has the type 'database' or 'bases set', like the ACL targets in the list. <P>
    * The 7 sorting rules include the 4 rules for addAclNOTForDatabase(),
    * where the fourth rule becomes the fifth rule.
    * They are the followings, in this order: <P>
    * (1) among the ACL targets, the ACL for the parent before the ACL for the child, <P>
    * (2) the conditional ACL before the non-conditional ACL, <P>
    * (3) among the conditional ACL, the ACL of the 'anonymous' account after the ACL
    * for other accounts, <P>
    * (4) among the conditional ACL, after the rule 3, if the account is 'anonymous' or not,
    * the ACL sorting follows the target sorting, that is all
    * the 'database' targets first, then the 'bases set' target sorting,  <P>
    * (5) among the non-conditional ACL (there are all for accounts-for-hosts groups),
    * the hosts group sorting is applied first (there is no order among the accounts), <P>
    * (6) among the non-conditional ACL, after the rule 5,
    * the ACL sorting follows the target sorting, that is all
    * the 'database' targets first, then the 'bases set' target sorting, <P>
    * (7) among the non-conditional ACL and for a database-specific target,
    * the to-be-mixed ACL are at the end of the list. <P>
    * The first order is not used because the calling method
    * manages it to put, in the argument _upd_l_acl, only database-specific target ACL.
    * The seventh order may be seen as an interface with the method
    * addAclForDatabaseForToBeMixedAcl().
    * An accounts-for-hosts group never handles a simple host, but always a hosts group.
    * The accounts-for-hosts group name has the format 'xxxx§@yyyyy', where 'yyyyy'
    * is a hosts group name last component which contains exactly one wildcard ('_' or '%').
    * The hosts group sorting is complex. A typical example of a sorted list for hosts groups,
    * all matching 'localhost', is the following hosts group last components: <P>
    * 'localhos_', '_ocalhost', 'localhost%', 'localhos%', 'lo%', 'lo%t', 'lo%ost',
    * 'l%t', '%', '%ocalhost', '%t'. <P>
    * For sorting the hosts groups of the accounts-for-hosts groups,
    * the method isAfterAsPatternNameHostsGroup() is called. <P>
    * A 'pattern-name_bases set' target has a name last component which contains
    * exactly one wildcard '_' or '%'. The 'bases set' sorting is specific and also complex.
    * A typical example of a sorted list for bases sets, all matching 'one',
    * is the following bases set last components: <P>
    * 'on%', 'on_', 'o%', 'o%e', 'o_e', '%', '_ne', '%e', '%ne'. <P>
    * For sorting the bases sets, the method isAfterAsPatternNameBasesSet() is called. <P>
    * The detailed sequence of calls for adding an ACL is described at the method
    * DirectoryMySQLImpl.addAclEntry(). Called by addInSortedAclList().
    * Calls getPreviousSortedTypesOfResourceInSortedAclList(),
    * getTypeOfResourceLastOrderInSortedAclList(),
    * getTypeOfResourceFirstOrderInSortedAclList(),
    * isAfterAsPatternNameHostsGroup().
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see ARoad0.AcsAddon.Accbee.MySQL.gWork.NodeRightsMySQLImpl#getL_accessRightsThroughNodesTree(ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see #isAfterAsPatternNameHostsGroup(java.lang.String, java.lang.String)
    * @see #isAfterAsPatternNameBasesSet(java.lang.String, java.lang.String)
    * @param _upd_l_acl direct AclEntry sorted list for a database-specific target.
    * May be empty but not null.
    * @param _acl direct AclEntry to put in the list at the right order. Never null
    * and not into _upd_l_acl.
    * @return the updated _upd_l_acl
    */
  protected static ArrayList<AclEntry> addAclForDatabaseNOTForToBeMixedAcl(
          ArrayList<AclEntry> _upd_l_acl, AclEntry _acl) {

                                      // for testing and tracking
                                      //System.out.println(" ");
                                      //System.out.println("BaseUtilityMySQLImpl."+
                                        //      "addAclForDatabaseNOTForToBeMixedAcl() start"+
                                          //    " for ACL "+_acl);

    // to not implement: (1) among the ACL targets, the ACL for the parent
                                      // before the ACL for the child
    //int                 firstOrder = 0;
    // to implement: (2) the conditional ACL before the non-conditional ACL
    int                 secondOrder = 0;
    // to implement: (3) among the conditional ACL, the ACL of the 'anonymous' account after the ACL
    // for other accounts
    int                 thirdOrder = 0;
    // to implement: (4) among the conditional ACL, after the rule 3,
    // if the account is 'anonymous' or not,
    // the ACL sorting follows the target sorting, that is all
    // the 'database' targets first, then the 'bases set' target sorting
    int                 fourthOrder = 0;
    // to implement: (5) among the non-conditional ACL (there are all for accounts-for-hosts groups),
    // the hosts group sorting is applied first
    int                 fifthOrder = 0;
    // to implement: (6) among the non-conditional ACL, after the rule 5,
    // the ACL sorting follows the target sorting, that is all
    // the 'database' targets first, then the 'bases set' target sorting
    int                 sixthOrder = 0;

    String              aclResType = _acl.getResType();
    String              resType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(aclResType);
    String              rAclType = null;

    AclEntry            acl = null;
    boolean             isConditional = _acl.isConditionalAclEntry();
    boolean             stopAtSecondOrder = false;
    String              hostGroupOne = null;
    String              hostGroupTwo = null;
    String              basesSetOne = null;
    String              basesSetTwo = null;

    // 'firstOrder' is set at 0, and there is an ACL at this position.
    // The second logic is from the condition group AclEntry order, which sets
    // the conditional ACL before the non-conditional ACL.
    acl = _upd_l_acl.get(0);
    if (!isConditional && acl.isConditionalAclEntry()) { // otherwise, secondOrder == 0

        for (Iterator ite = _upd_l_acl.listIterator(); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the secondOrder, "+
                                      //    "process the ACL "+acl);

            if (!isConditional && acl.isConditionalAclEntry()) {
                secondOrder++;
                continue;
            }
            else if (isConditional && !acl.isConditionalAclEntry()) {
                // no superior order to process
                stopAtSecondOrder = true;
                break;
            }
            else {
                // superior order to process
                break;
            }

        } // for ite

    } // if (!isConditional &&

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() with firstOrder = "+firstOrder+
                                      //    " and secondOrder = "+secondOrder);

    // if only the first and second logic to apply
    if (secondOrder == _upd_l_acl.size() || stopAtSecondOrder) {
        aclResType = null;
        resType = null;
        acl = null;

        _upd_l_acl.add(secondOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseNOTForToBeMixedAcl() "+
                                    //      "ends, with only the first and second logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // 'firstOrder' and 'secondOrder' are set, and there is an ACL at (secondOrder).
    // The third logic is among the conditional ACL, to put the ACL of the 'anonymous' account
    // after the ACL for other accounts
    acl = _upd_l_acl.get(secondOrder);
    if (isConditional && acl.isConditionalAclEntry() &&
            _acl.getEpName().lastComponent().equals("anonymous")) { // otherwise, thirdOrder == 0

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the thirdOrder, "+
                                      //    "process the ACL "+acl);

            if (acl.isConditionalAclEntry() &&
                    !acl.getEpName().lastComponent().equals("anonymous")) {
                thirdOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (!isConditional

    // if only the 3 first logic to apply
    if (isConditional &&
            (secondOrder + thirdOrder == _upd_l_acl.size() ||
            resType.indexOf("database") != -1)) { // that is, fourthOrder = 0
        aclResType = null;
        resType = null;
        acl = null;

        _upd_l_acl.add(secondOrder + thirdOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseNOTForToBeMixedAcl() "+
                                    //      "ends, with only the 3 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // The fourth logic among the conditional ACL, after the rule 3,
    // if the account is 'anonymous' or not,
    // the ACL sorting follows the target sorting, that is
    // the 'database' target first (fourthOrder = 0), then the 'bases set' target sorting
    acl = _upd_l_acl.get(secondOrder + thirdOrder);
    if (isConditional && acl.isConditionalAclEntry() &&
            resType.indexOf("database") == -1) { // otherwise, fourthOrder == 0

        basesSetOne = _acl.getResName().lastComponent();

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder + thirdOrder);
          ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the fourthOrder, "+
                                      //    "process the ACL "+acl);

            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());
            
            // recall of the third logic among the 'pattern-name_bases set' ACL
            if (!_acl.getEpName().lastComponent().equals("anonymous") &&
                    acl.getEpName().lastComponent().equals("anonymous")) {
                break;
            }

            if (acl.isConditionalAclEntry()) {

                if (rAclType.indexOf("database") != -1) {
                        fourthOrder++;

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the fourthOrder, "+
                                      //        "acl is a database. Incremented 'fourthOrder': "+
                                        //      fourthOrder);

                        continue;
                } // if

                basesSetTwo = acl.getResName().lastComponent();

                if (isAfterAsPatternNameBasesSet(basesSetOne, basesSetTwo)) {
                    fourthOrder++;

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the fourthOrder, "+
                                      //        "acl is before. Incremented 'fourthOrder': "+
                                        //      fourthOrder);

                    continue;
                }
                else {
                    break;
                }
            }
            else {
                break;
            } // else

        } // for ite

    } // if (isConditional

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseNOTForToBeMixedAcl(), "+
                                    //      "end of the third and fourth logics. ");

    // if conditional ACL, the fourth logic is the last one
    if (isConditional) {
        aclResType = null;
        resType = null;
        rAclType = null;
        acl = null;
        basesSetOne = null;
        basesSetTwo = null;

        _upd_l_acl.add(secondOrder + thirdOrder + fourthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseNOTForToBeMixedAcl() "+
                                    //      "ends, with only the 4 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // 'firstOrder' and 'secondOrder' are set, and there is an ACL at (firstOrder + secondOrder).
    // The fifth logic is among the non-conditional accounts-for-hosts groups ACL,
    // to apply the hosts group sorting.
    acl = _upd_l_acl.get(secondOrder); // redondant but clearer
    if (!acl.isConditionalAclEntry() &&
         _acl.getEorEP().getEPType().indexOf("pattern-name_accounts-for-hosts group") != -1) { // otherwise, fifthOrder == 0

        hostGroupOne = _acl.getEpName().lastComponent().substring(
                _acl.getEpName().lastComponent().indexOf("§@") + 2);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the fifthOrder, "+
                                      //    "process the _acl "+_acl);

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the fifthOrder, "+
                                      //    "process the acl "+acl);

            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

            // search on the same target type
            if (!rAclType.equals(resType)) {
                break;
            }

            if (acl.getEorEP().getEPType().indexOf("pattern-name_accounts-for-hosts group") != -1) {

                hostGroupTwo = acl.getEpName().lastComponent().substring(
                        acl.getEpName().lastComponent().indexOf("§@") + 2);

                if (isAfterAsPatternNameHostsGroup(hostGroupOne, hostGroupTwo)) {
                    fifthOrder++;
                    continue;
                }
                else {
                    break;
                }
            }
            else {
                break;
            } // else

        } // for ite

    } // if (!isConditional

    // If only the 5 first logic to apply.
    if (secondOrder + fifthOrder == _upd_l_acl.size() ||
            resType.indexOf("database") != -1) { // that is, sixthOrder == 0
        aclResType = null;
        resType = null;
        rAclType = null;
        acl = null;
        hostGroupOne = null;
        hostGroupTwo = null;
        basesSetOne = null;
        basesSetTwo = null;

        _upd_l_acl.add(secondOrder + fifthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseNOTForToBeMixedAcl() "+
                                    //      "ends, with only the 3 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // The sixth logic is among the non-conditional ACL, and
    // the ACL sorting follows the target sorting, that is
    // the 'database' target first (sixthOrder == 0), then the 'bases set' target sorting.
    // to apply the seventh logic, to find the first to-be-mixed ACL stops the iteration.
    acl = _upd_l_acl.get(secondOrder + fifthOrder);
    if (!acl.isConditionalAclEntry() && resType.indexOf("database") == -1) { // otherwise, sixthOrder == 0

        basesSetOne = _acl.getResName().lastComponent();

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder + fifthOrder);
          ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseNOTForToBeMixedAcl() for the sixthOrder, "+
                                      //    "process the ACL "+acl);

            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

            // to apply the seventh logic
            if (acl.getEorEP().getEPType().indexOf("pattern-name_accounts-for-hosts group") == -1) {
                break;
            }

            if (rAclType.indexOf("database") != -1) {
                sixthOrder++;
                continue;
            }

            // search on the same target type
            if (!rAclType.equals(resType)) {
                break;
            }

            basesSetTwo = acl.getResName().lastComponent();

            if (isAfterAsPatternNameBasesSet(basesSetOne, basesSetTwo)) {
                sixthOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (isConditional

    _upd_l_acl.add(secondOrder + fifthOrder + sixthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseNOTForToBeMixedAcl() ends "+
                                    //      "after the seven sorting, with result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

    aclResType = null;
    resType = null;
    rAclType = null;
    hostGroupOne = null;
    hostGroupTwo = null;
    basesSetOne = null;
    basesSetTwo = null;

    return _upd_l_acl;
  } // addAclForDatabaseNOTForToBeMixedAcl()

    /**
    * This method inserts an AclEntry into the ACL list, following the MySQL sorting,
    * when the _acl target is a database-specific target and _acl is a to-be-mixed ACL.
    * This sorting is only for the list of the direct AclEntries,
    * and then, the argument _upd_l_acl contains only such AclEntries for a database-specific target.
    * The _acl target has the type 'database' or 'bases set', like the targets in the ACL list. <P>
    * The 8 sorting rules are derived from the 7 rules for addAclForDatabaseNOTForToBeMixedAcl(),
    * where the 3-to-6 rules are inserted in the positions 5 to 8.
    * They are the followings, in this order: <P>
    * (1) among the ACL targets, the ACL for the parent before the ACL for the child, <P>
    * (2) the conditional ACL before the non-conditional ACL, <P>
    * (3) among the non-conditional ACL and for a database-specific target,
    * the to-be-mixed ACL are at the end of the list, after
    * the ACL for the type "pattern-name_accounts-for-hosts group". <P>
    * (4) among the to-be-mixed ACL, the ACL of host(s) are after the ACL
    * of account, <P>
    * (5) among the to-be-mixed ACL of accounts,
    * the ACL of the 'anonymous' account after the ACL of other accounts, <P>
    * (6) among the to-be-mixed ACL of accounts, after the rule 5,
    * if the account is 'anonymous' or not,
    * the ACL sorting follows the target sorting, that is
    * the 'database' target first, then the 'bases set' target sorting,  <P>
    * (7) among the to-be-mixed ACL of host(s),
    * the ACL sorting follows the hosts group sorting, that is
    * the 'host' source first, then the 'hosts group' source sorting,  <P>
    * (8) among the to-be-mixed ACL of host or hosts group, after the rule 7,
    * the ACL sorting follows the target sorting, that is
    * the 'database' target first, then the 'bases set' target sorting. <P>
    * The first order is not used because the calling method
    * manages it to put in the argument _upd_l_acl, only database-specific target ACL.
    * A to-be-mixed ACL is a non-conditional ACL.
    * The second and third orders may be seen as an interface with the method
    * addAclForDatabaseNOTForToBeMixedAcl().
    * The hosts group sorting is complex. A typical example of a sorted list for hosts groups,
    * all matching 'localhost', is the following hosts group last components: <P>
    * 'localhos_', '_ocalhost', 'localhost%', 'localhos%', 'lo%', 'lo%t', 'lo%ost',
    * 'l%t', '%', '%ocalhost', '%t'. <P>
    * For sorting the hosts groups of the accounts-for-hosts groups,
    * the method isAfterAsPatternNameHostsGroup() is called. <P>
    * A 'pattern-name_bases set' target has a name last component which contains
    * exactly one wildcard '_' or '%'. The 'bases set' sorting is specific and also complex.
    * A typical example of a sorted list for bases sets, all matching 'one',
    * is the following bases set last components: <P>
    * 'on%', 'on_', 'o%', 'o%e', 'o_e', '%', '_ne', '%e', '%ne'. <P>
    * For sorting the bases sets, the method isAfterAsPatternNameBasesSet() is called. <P>
    * The detailed sequence of calls for adding an ACL is described at the method
    * DirectoryMySQLImpl.addAclEntry(). Called by addInSortedAclList().
    * Calls getPreviousSortedTypesOfResourceInSortedAclList(),
    * getTypeOfResourceLastOrderInSortedAclList(),
    * getTypeOfResourceFirstOrderInSortedAclList(),
    * isAfterAsPatternNameHostsGroup().
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see ARoad0.AcsAddon.Accbee.MySQL.gWork.NodeRightsMySQLImpl#getL_accessRightsThroughNodesTree(ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.ImmutableGroupID)
    * @see #isAfterAsPatternNameHostsGroup(java.lang.String, java.lang.String)
    * @see #isAfterAsPatternNameBasesSet(java.lang.String, java.lang.String)
    * @param _upd_l_acl direct AclEntry sorted list for a database-specific target.
    * May be empty but not null.
    * @param _acl direct AclEntry to put in the list at the right order. Never null
    * and not into _upd_l_acl.
    * @return the updated _upd_l_acl
    */
  protected static ArrayList<AclEntry> addAclForDatabaseForToBeMixedAcl(
          ArrayList<AclEntry> _upd_l_acl, AclEntry _acl) {

                                      // for testing and tracking
                                      //System.out.println(" ");
                                      //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() start"+
                                        //      " for ACL "+_acl);

    // to not implement: (1) among the ACL targets, the ACL for the parent before the ACL for the child
    //int                 firstOrder = 0;
    // to implement: (2) the conditional ACL before the non-conditional ACL
    int                 secondOrder = 0;
    // to implement: (3) among the non-conditional ACL and for a database-specific target,
    // the to-be-mixed ACL are at the end of the list, after
    // the ACL for the type "pattern-name_accounts-for-hosts group"
    int                 thirdOrder = 0;
    // to implement: (4) among the to-be-mixed ACL, the ACL of host(s) are after the ACL
    // of account
    int                 fourthOrder = 0;
    // to implement: (5) among the to-be-mixed ACL of accounts,
    // the ACL of the 'anonymous' account after the ACL of other accounts
    int                 fifthOrder = 0;
    // to implement: (6) among the to-be-mixed ACL of accounts, after the rule 5,
    // if the account is 'anonymous' or not,
    // the ACL sorting follows the target sorting, that is
    // the 'database' target first, then the 'bases set' target sorting
    int                 sixthOrder = 0;
    // to implement: (7) among the to-be-mixed ACL of host(s),
    // the ACL sorting follows the hosts group sorting, that is
    // the 'host' source first, then the 'hosts group' source sorting
    int                 seventhOrder = 0;
    // to implement: (8) among the to-be-mixed ACL of host or hosts group, after the rule 7,
    // the ACL sorting follows the target sorting, that is
    // the 'database' target first, then the 'bases set' target sorting
    int                 eightOrder = 0;

    String              aclResType = _acl.getResType();
    String              resType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(aclResType);
    String              rAclType = null;

    AclEntry            acl = null;
    boolean             isConditional = _acl.isConditionalAclEntry(); // always false
    //boolean             stopAtSecondOrder = false;
    String              hostGroupOne = null;
    String              hostGroupTwo = null;
    String              basesSetOne = null;
    String              basesSetTwo = null;

    // 'firstOrder' is set at 0, and there is an ACL at this position.
    // The second logic is from the condition group AclEntry order, which sets
    // the conditional ACL before the non-conditional ACL.
    acl = _upd_l_acl.get(0);
    if (acl.isConditionalAclEntry()) { // otherwise, secondOrder == 0

        for (Iterator ite = _upd_l_acl.listIterator(); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseForToBeMixedAcl() for the secondOrder, "+
                                      //    "process the ACL "+acl);

            if (!isConditional && acl.isConditionalAclEntry()) {
                secondOrder++;
                continue;
            }
            //else if (isConditional && !acl.isConditionalAclEntry()) {
                // no superior order to process
              //  stopAtSecondOrder = true;
                //break;
            //}
            else {
                // superior order to process
                break;
            }

        } // for ite

    } // if (!isConditional

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseForToBeMixedAcl() with firstOrder = "+firstOrder+
                                      //    " and secondOrder = "+secondOrder);

    // if only the first and second logic to apply
    if (secondOrder == _upd_l_acl.size()) {
        aclResType = null;
        resType = null;
        acl = null;

        _upd_l_acl.add(secondOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() "+
                                    //      "ends, with only the first and second logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // 'firstOrder' and 'secondOrder' are set, and there is an ACL at (secondOrder).
    // The third logic is among the non-conditional ACL and for a database-specific target,
    // where the to-be-mixed ACL are at the end of the list, after
    // the ACL for the type "pattern-name_accounts-for-hosts group"
    acl = _upd_l_acl.get(secondOrder);
    if (acl.getEorEP().getType().indexOf("pattern-name_accounts-for-hosts group") != -1) {

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseForToBeMixedAcl() for the thirdOrder, "+
                                      //    "process the ACL "+acl);

            if (acl.getEorEP().getType().indexOf("pattern-name_accounts-for-hosts group") != -1) {
                thirdOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (acl.getEorEP().getType().indexOf("pattern-name_accounts-for-hosts group")

    // if only the 3 first logic to apply
    if (secondOrder + thirdOrder == _upd_l_acl.size() ) {
        aclResType = null;
        resType = null;
        acl = null;

        _upd_l_acl.add(secondOrder + thirdOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() "+
                                    //      "ends, with only the 3 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // The fourth logic is among the to-be-mixed ACL, where the ACL of host(s)
    // are after the ACL of account
    acl = _upd_l_acl.get(secondOrder + thirdOrder);
    if (_acl.getEorEP() instanceof GroupID && acl.getEorEP() instanceof UserID) { // otherwise, fourthOrder == 0

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder + thirdOrder); ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseForToBeMixedAcl() for the thirdOrder, "+
                                      //    "process the ACL "+acl);

            if (acl.getEorEP() instanceof UserID) {
                fourthOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (_acl.getEorEP() instanceof GroupID

    // if only the 4 first logic to apply
    if (secondOrder + thirdOrder + fourthOrder == _upd_l_acl.size() ) {
        aclResType = null;
        resType = null;
        acl = null;

        _upd_l_acl.add(secondOrder + thirdOrder + fourthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() "+
                                    //      "ends, with only the 4 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // The fifth logic is among the to-be-mixed ACL of accounts,
    // the ACL of the 'anonymous' account after the ACL of other accounts
    acl = _upd_l_acl.get(secondOrder + thirdOrder + fourthOrder);
    if (_acl.getEorEP() instanceof UserID &&
            acl.getEorEP() instanceof UserID &&
            _acl.getEpName().lastComponent().equals("anonymous") &&
            !acl.getEpName().lastComponent().equals("anonymous")) { // otherwise, fifthOrder == 0

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder + thirdOrder + fourthOrder);
          ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseForToBeMixedAcl() for the fifthOrder, "+
                                      //    "process the ACL "+acl);

            if (acl.getEorEP() instanceof UserID &&
                    !acl.getEpName().lastComponent().equals("anonymous")) {
                fifthOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (!isConditional

    // if only the 5 first logic to apply
    if (secondOrder + thirdOrder + fourthOrder + fifthOrder == _upd_l_acl.size()) {
        aclResType = null;
        resType = null;
        acl = null;

        _upd_l_acl.add(secondOrder + thirdOrder + fourthOrder + fifthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() "+
                                    //      "ends, with only the 5 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // The sixth logic among the to-be-mixed ACL of accounts, after the rule 5,
    // if the account is 'anonymous' or not,
    // the ACL sorting follows the target sorting, that is
    // the 'database' target first, then the 'bases set' target sorting
    acl = _upd_l_acl.get(secondOrder + thirdOrder + fourthOrder + fifthOrder);
    if (_acl.getEorEP() instanceof UserID &&
            acl.getEorEP() instanceof UserID &&
            resType.indexOf("database") == -1) { // otherwise, sixthOrder == 0

        basesSetOne = _acl.getResName().lastComponent();

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder + thirdOrder + fourthOrder + fifthOrder);
          ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseForToBeMixedAcl() for the sixthtOrder, "+
                                      //    "process the ACL "+acl);

            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

            // recall of the third logic among the 'pattern-name_bases set' ACL
            if (!_acl.getEpName().lastComponent().equals("anonymous") &&
                    acl.getEpName().lastComponent().equals("anonymous")) {
                break;
            }

            if (rAclType.indexOf("database") != -1) {
                    sixthOrder++;
                    continue;
            }

            // search on the same target type
            if (!rAclType.equals(resType)) {
                break;
            }

            basesSetTwo = acl.getResName().lastComponent();

            if (isAfterAsPatternNameBasesSet(basesSetOne, basesSetTwo)) {
                sixthOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (_acl.getEorEP() instanceof UserID

    // if there are only the 6 first logics to apply
    if (secondOrder + thirdOrder + fourthOrder + fifthOrder + sixthOrder == _upd_l_acl.size()) {
        aclResType = null;
        resType = null;
        acl = null;

        _upd_l_acl.add(secondOrder + thirdOrder + fourthOrder + fifthOrder + sixthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() "+
                                    //      "ends, with only the 6 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() for "+
                                    //      "the third and fourth logics. ");

    // if the sixth logic is the last one
    if (_acl.getEorEP() instanceof UserID) {
        aclResType = null;
        resType = null;
        rAclType = null;
        acl = null;
        basesSetOne = null;
        basesSetTwo = null;

        _upd_l_acl.add(secondOrder + thirdOrder + fourthOrder + fifthOrder + sixthOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() "+
                                    //      "ends, with only the 6 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // 'secondOrder', 'thirdOrder' and 'fourthOrder' are set, and
    // there is an ACL at (secondOrder + thirdOrder + fourthOrder).
    // The seventh logic is among the to-be-mixed ACL of host(s),
    // the ACL sorting follows the hosts group sorting, that is
    // the 'host' source first, then the 'hosts group' source sorting
    acl = _upd_l_acl.get(secondOrder + thirdOrder + fourthOrder); // redondant but clearer
    if (_acl.getEorEP().getEPType().indexOf("hosts group") != -1) { // otherwise, seventhOrder == 0

        hostGroupOne = _acl.getEpName().lastComponent();

        for (Iterator ite = _upd_l_acl.listIterator(secondOrder + thirdOrder + fourthOrder);
          ite.hasNext(); ) {
            acl = (AclEntry) ite.next();

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "addAclForDatabaseForToBeMixedAcl() for the seventhOrder, "+
                                      //    "process the ACL "+acl);

            rAclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

            // if 'acl' is for a simple host
            if (acl.getEorEP().getEPType().indexOf("hosts group") == -1) {
                seventhOrder++;
                continue;
            }

            // 'hosts group' ACL on 'database' and 'hosts group' ACL on 'pattern-name_bases set'
            // are not compared there
            if (!rAclType.equals(resType)) {
                if (acl.getResType().indexOf("bases set") == -1) {
                    // _acl is for 'hosts group' on a 'pattern-name_bases set',
                    // then all the 'hosts group' ACL on 'database' are jumped.
                    seventhOrder++;
                    continue;
                }
                else {
                    // _acl is for 'hosts group' on a 'database'
                    break;
                }
            } // if (!rAclType.equals(resType))

            hostGroupTwo = acl.getEpName().lastComponent();

            if (isAfterAsPatternNameHostsGroup(hostGroupOne, hostGroupTwo)) {
                seventhOrder++;
                continue;
            }
            else {
                break;
            }

        } // for ite

    } // if (_acl.getEorEP().getEPType().indexOf("hosts group") != -1)

    // If only the 7 first logic to apply.
    if (secondOrder + thirdOrder + fourthOrder + seventhOrder == _upd_l_acl.size() ||
            resType.indexOf("database") != -1) { // that is, eightOrder == 0
        aclResType = null;
        resType = null;
        rAclType = null;
        acl = null;
        hostGroupOne = null;
        hostGroupTwo = null;
        basesSetOne = null;
        basesSetTwo = null;

        _upd_l_acl.add(secondOrder + thirdOrder + fourthOrder + seventhOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() "+
                                    //      "ends, with only the 7 first logics to apply,"+
                                      //    " and result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

        return _upd_l_acl;
    } // if

    // The eigth logic is among the to-be-mixed ACL of host or hosts group,
    // where the ACL sorting follows the target sorting, that is
    // the 'database' target first, then the 'bases set' target sorting
    basesSetOne = _acl.getResName().lastComponent();

    for (Iterator ite = _upd_l_acl.listIterator(secondOrder + thirdOrder + fourthOrder + seventhOrder);
      ite.hasNext(); ) {
        acl = (AclEntry) ite.next();

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "addAclForDatabaseForToBeMixedAcl() for the eightOrder, "+
                                  //    "process the ACL "+acl);


        rAclType = UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    acl.getResType());

        if (rAclType.indexOf("database") != -1) {
                eightOrder++;
                continue;
        }

        // search on the same target type
        if (!rAclType.equals(resType)) {
            break;
        }

        basesSetTwo = acl.getResName().lastComponent();

        if (isAfterAsPatternNameBasesSet(basesSetOne, basesSetTwo)) {
            eightOrder++;
            continue;
        }
        else {
            break;
        }

    } // for ite

    _upd_l_acl.add(secondOrder + thirdOrder + fourthOrder + seventhOrder + eightOrder, _acl);

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl.addAclForDatabaseForToBeMixedAcl() ends "+
                                   //       "after the eight sorting, with result: ");
                                  //for (Object obj : _upd_l_acl) {
                                    //  System.out.println("----- "+obj);
                                  //}

    aclResType = null;
    resType = null;
    rAclType = null;
    hostGroupOne = null;
    hostGroupTwo = null;
    basesSetOne = null;
    basesSetTwo = null;

    return _upd_l_acl;
  } // addAclForDatabaseForToBeMixedAcl()

  /**
   * Gets the Resource types which are before _type in the sorted AclEntry list.
   * This method applies the first AclEntry sorting logic, based on the parent tree
   * for the ACL target.
   * This target-oriented sorting sets first 'MySQL server', then 'database' and
   * 'pattern-name_bases set', then as database children and at the same level, 'table',
   * 'view_delegate', 'stored_function' and 'stored_procedure',
   * then as table children, 'table_column'. <P>
   * For instance, if the argument _type is 'table', the returned list may contain
   * at most 'MySQL server', 'database' and 'pattern-name_bases set' as previous types.
   * The types 'byAcsCreator', 'directory', 'file', 'trigger' and 'index' are for Actors which
   * cannot be AclEntry targets. <P>
   * Note: this target-oriented logic is never altered by the other logics in the final ACL list,
   * except for the sorting between 'database' and 'pattern-name_bases set', where, for instance,
   * the conditional ACL 'user@host' for a 'pattern-name_bases set' is set before
   * the non-conditional ACL 'accounts-for-hosts group' for 'database'.
   * The general rule is that the host-oriented sorting is prior to the base-oriented sorting.
   * This is why 'database' is returned for the _type 'pattern-name_bases set',
   * and 'pattern-name_bases set' is returned for the _type 'database'.
   * When this target-oriented logic is altered, it is the responsability of the calling
   * method to handle properly the changes in the ACL sorting. <P>
   * Called by addAclNOTForDatabase().
   * @param _l_acl list of already-sorted AclEntries
   * @param _type a Resource type without 'immutable'. May be the type 
   * of an AclEntry Resource in the list.
   * @return sorted list of the Resource types which are in the AclEntry list
   * and which are before _type. The first type in the list is the first type in _l_acl.
   * May be empty but not null.
   */
  public static ArrayList<String> getPreviousSortedTypesOfResourceInSortedAclList(
          ArrayList<AclEntry> _l_acl, String _type) {

                                      // for testing and tracking
                                      //System.out.println("BaseUtilityMySQLImpl."+
                                        //      "getPreviousSortedTypesOfResourceInSortedAclList()"+
                                          //    " starts for the type "+_type);

      ArrayList<String> l_result = new ArrayList(10);
      //String            type = UtilityImpl.getFirstInstance().
        //                            getObjectTypeWithoutImmutable(_type);
      String            aclType = null;

      for (AclEntry acl : _l_acl) {

                                      // for testing and tracking
                                      //System.out.println("BaseUtilityMySQLImpl."+
                                        //      "getPreviousSortedTypesOfResourceInSortedAclList()"+
                                          //    " analyzes the ACL "+acl);

          aclType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        acl.getResType());

          if (!l_result.contains("MySQL server <byAcsCreator>") &&
                  aclType.equals("MySQL server <byAcsCreator>") ) {

              if (!_type.equals("MySQL server <byAcsCreator>")) {

                  l_result.add("MySQL server <byAcsCreator>");
                  continue;

              }
              else
                  break;
          } // if

          if (!l_result.contains("database") &&
                                      aclType.equals("database")) {

              if (!_type.equals("MySQL server <byAcsCreator>") &&
                      !_type.equals("database") ) {

                  l_result.add("database");
                  continue;

              }
              else
                  break;
          } // if

          if (!l_result.contains("pattern-name_bases set") &&
                                      aclType.equals("pattern-name_bases set")) {

              if (!_type.equals("MySQL server <byAcsCreator>") &&
                      !_type.equals("database") &&
                      !_type.equals("pattern-name_bases set")) {

                  l_result.add("pattern-name_bases set");
                  continue;

              }
              else
                  break;
          } // if
          
          // view_delegate <byAcsCreator>, stored_function and stored_procedure
          // are children of database without children: noop
          if (aclType.equals("view_delegate <byAcsCreator>") ||
                  aclType.equals("stored_function") ||
                  aclType.equals("stored_procedure") ) {
              break;
          } // if

          if (!l_result.contains("table") && aclType.equals("table")) {

              if (_type.equals("table_column") ) {

                  l_result.add("table");
                  break;
                  
              }
              else
                  break;
          } // if

      } // for acl

                                      // for testing and tracking
                                      //System.out.println("BaseUtilityMySQLImpl."+
                                        //      "getPreviousSortedTypesOfResourceInSortedAclList()"+
                                          //    " ends with the result for the type "+_type);
                                      //for (String r : l_result) {
                                        //  System.out.println("----- "+r);
                                      //}

      return l_result;
  } // getPreviousSortedTypesOfResourceInSortedAclList()

  /**
   * Gets the order of the last AclEntry in _upd_l_acl having a Resource
   * with the argument _type in its type.
   * The target-oriented first sorting sets first 'MySQL server', then 'database' and
   * 'pattern-name_bases set', then as database children and at the same level, 'table',
   * 'view_delegate', 'stored_function' and 'stored_procedure',
   * then as table children, 'table_column'. <P>
   * This order is both for direct AclEntries and inherited AclEntries.
   * Called by addAclNOTForDatabase().
   * @param _l_acl list of AclEntries in the correct sorting from the Resource type
   * @param _type a Resource type without immutable
   * @return the order of the last AclEntry in _upd_l_acl having a Resource
   * with _type in its type, or -1 if _type is not found
   */
  public static int getTypeOfResourceLastOrderInSortedAclList(
          ArrayList<AclEntry> _l_acl, String _type) {

      int       result = _l_acl.size() - 1;
      boolean   find = false;
      boolean   amongTheFirstTypes = false;
      AclEntry  acl = null;
      String    aclType = null;

      if (_type.equals("MySQL server <byAcsCreator>") || _type.equals("database") )
          amongTheFirstTypes = true;

      for (int i = _l_acl.size() - 1; i > -1; i--) {
          acl = _l_acl.get(i);
          aclType = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                        acl.getResType());

          if (aclType.equals(_type)) {
              find = true;
              break;
          }

          // stop at the middle of the type list, if it is possible,
          // to speed up the search. Works only if the sort of AclEntries
          // in _upd_l_acl is correct.
          if ((aclType.equals("MySQL server <byAcsCreator>") || aclType.equals("database")) &&
                  !amongTheFirstTypes) {
              break;
          }

          result--;
      } // for i

      if (!find) result = -1;

      acl = null;
      aclType = null;
      return result;
  } // getTypeOfResourceLastOrderInSortedAclList()
  
  /**
   * Gets the order of the first AclEntry in _upd_l_acl having a Resource
   * with the argument _type in its type.
   * The target-oriented first sorting sets first 'MySQL server', then 'database' and
   * 'pattern-name_bases set', then as database children and at the same level, 'table',
   * 'view_delegate', 'stored_function' and 'stored_procedure',
   * then as table children, 'table_column'. <P>
   * This order is both for direct AclEntries and inherited AclEntries.
   * Called by addAclNOTForDatabase().
   * @param _l_acl list of AclEntries in the correct sorting from the Resource type
   * @param _type a Resource type without 'immutable'
   * @return the order of the first AclEntry in _upd_l_acl having a Resource
   * with _type in its type, or -1 if _type is not found
   */
  public static int getTypeOfResourceFirstOrderInSortedAclList(
          ArrayList<AclEntry> _l_acl, String _type) {
      int       result = 0;
      boolean   find = false;
      boolean   amongTheFirstTypes = false;
      String    aclType = null;

      if (_type.equals("MySQL server <byAcsCreator>") || _type.equals("database"))
          amongTheFirstTypes = true;

      for (AclEntry acl: _l_acl) {
          aclType = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                        acl.getResType());

          if (aclType.equals(_type)) {
              find = true;
              break;
          }

          // stop at the middle of the type list, if it is possible,
          // to speed up the search. Works only if the sort of AclEntries
          // in _upd_l_acl is correct.
          if (!aclType.equals("MySQL server <byAcsCreator>") && !aclType.equals("database") &&
                  amongTheFirstTypes) {
              break;
          }

          result++;
      } // for acl

      if (!find) result = -1;

      aclType = null;
      return result;
  } // getTypeOfResourceFirstOrderInSortedAclList()

  /**
   * Detects from two hosts group name last components, if the first argument
   * has to be sorted after the second argument. Each argument contains exactly
   * one wildcard '_' or '%'. The hosts group sorting from the names is complex. 
   * A typical example of a sorted list for hosts groups, all matching the host 'localhost',
   * is the following hosts group last components: <P>
   * 'localhos_', '_ocalhost', 'localhost%', 'localhos%', 'lo%', 'lo%t', 'lo%ost',
   * 'l%t', '%', '%ocalhost', '%t'. <P>
   * This method implements the following algorithm: <P>
   *
   * (1) if a component contains '_' and the other component contains '%',
   * the '%' component is after ('_ocalhost' before '%ocalhost'), <P>
   *
   * (2) if both components contain '_': <P>
   * (2.1) if the 2 '_' have not the same location, the smaller location is after
   * ('localhos_' before '_ocalhost'), <P>
   * (2.2) otherwise the components are equal, <P>
   *
   * (3) if both components contain '%': <P>
   * (3.1) if the 2 '%' have not the same location, the smaller location is after
   * ('localhost%' before 'localhos%'), <P>
   * (3.2) if the '%' location is 0: <P>
   * (3.2.1) the component '%' is never after ('%' before '%ocalhost'), <P>
   * (3.2.2) otherwise, the smaller component is after ('%ocalhost' before '%t'), <P>
   * (3.3) otherwise, the greater component is after ('lo%t' before 'lo%ost'), <P>
   * 
   * This algorithm is based on MySQL server testing. It is not from the official documentation,
   * which declares simply that the most-specific group is sorted first. <P>
   * Called by addAclNOTForDatabase(), addAclForDatabaseNOTForToBeMixedAcl(),
   * addAclForDatabaseForToBeMixedAcl(), isAfterInTheGroupIDMemberOrder(), and
   * NodeRightsMySQLImpl.detectL_aclEntryRightsNOTForDatabase().
   * @param _compo_one last component of a hosts group name of an ACL to sort.
   * Must contain exactly one wildcard '_' or '%'.
   * @param _compo_two last component of a hosts group name of an ACL to sort.
   * Must contain exactly one wildcard '_' or '%'.
   * @return true if the ACL of _compo_one has to be after the ACL of _compo_two,
   * following the AclEntry pattern-name hosts group sorting
   */
  public static boolean isAfterAsPatternNameHostsGroup(
          String _compo_one, String _compo_two) {

      if (_compo_one.indexOf("%") == -1 && _compo_one.lastIndexOf("_") == -1)
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameHostsGroup() "+
                  "without one '%' or '_' in "+_compo_one);
      if (_compo_two.indexOf("%") == -1 && _compo_two.lastIndexOf("_") == -1)
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameHostsGroup() "+
                  "without one '%' or '_' in "+_compo_two);

      if (_compo_one.indexOf("%") != _compo_one.lastIndexOf("%"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameHostsGroup() "+
                  "with more than one '%' in "+_compo_one);
      if (_compo_two.indexOf("%") != _compo_two.lastIndexOf("%"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameHostsGroup() "+
                  "with more than one '%' in "+_compo_two);
      if (_compo_one.indexOf("_") != _compo_one.lastIndexOf("_"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameHostsGroup() "+
                  "with more than one '_' in "+_compo_one);
      if (_compo_two.indexOf("_") != _compo_two.lastIndexOf("_"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameHostsGroup() "+
                  "with more than one '_' in "+_compo_two);

      // the default value, so only true result is detected
      boolean   result = false;
      int       count_one = -1;
      int       count_two = -1;

      // (1) if a component contains '_' and the other component contains '%',
      // the '%' component is after
      if (_compo_one.indexOf("%") != -1 && _compo_two.indexOf("%") == -1)
              result = true;

      // (2) if both components contain '_'...
      else if (_compo_one.indexOf("_") != -1 && _compo_two.indexOf("_") != -1) {

          count_one = _compo_one.indexOf("_");
          count_two = _compo_two.indexOf("_");
          // (2.1) if the 2 '_' have not the same location, the smaller location is after
          if (count_one < count_two) {
              result = true;
          } // if
          // (2.2) otherwise the components are equal
      }

      // (3) if both components contain '%'...
      else if (_compo_one.indexOf("%") != -1 && _compo_two.indexOf("%") != -1) {

          count_one = _compo_one.indexOf("%");
          count_two = _compo_two.indexOf("%");

          // (3.1) if the 2 '%' have not the same location, the smaller location is after
          if (count_one < count_two) {
                  result = true;
          }
          else if (count_one == count_two) {

              // (3.2) if the '%' location is 0
              if (count_one == 0) {
                // (3.2.1) the component '%' is never after ('%' before '%ocalhost')
                if (_compo_two.equals("%"))
                    result = true;
                // (3.2.2) otherwise, the smaller component is after ('%ocalhost' before '%t')
                else if (_compo_one.length() < _compo_two.length())
                    result = true;
              }
              // (3.3) otherwise, the greater component is after
              else if (_compo_one.length() > _compo_two.length()) {
                  result = true;
              }

          } // else if (count_one == count_two)

      } // else if

      return result;
  } // isAfterAsPatternNameHostsGroup()

  /**
   * Detects from two bases set name last components, if the first argument
   * has to be sorted after the second argument. Each argument contains exactly
   * one wildcard '_' or '%'. The bases set sorting from the names is complex,
   * and it is not like the host group sorting.
   * A typical example of a sorted list for bases sets, all matching the base 'one',
   * is the following bases set last components: <P>
   * 'on%', 'on_', 'o%', 'o%e', 'o_e', '%', '_ne', '%e', '%ne'. <P>
   * This method implements the following algorithm: <P>
   * 
   * (1) if a component contains '_' and the other component contains '%': <P>
   * (1.1) with different locations, the smaller location is after ('on_' before 'o%'),
   * ('o%e' before '_ne'), <P>
   * (1.2) otherwise, if the '%' location is 0: <P>
   * (1.2.1) the component '%' is never after ('%' before '_ne'), <P>
   * (1.2.2) otherwise, the '%' component is after ('_ne' before '%e'), <P>
   * (1.3) otherwise, the '_' component is after ('on%' before 'on_'), <P>
   *
   * (2) if both components contain '_': <P>
   * (2.1) if the 2 '_' have not the same location, the smaller location is after
   * ('on_' before 'o_e'), <P>
   * (2.2) otherwise the components are equal, <P>
   *
   * (3) if both components contain '%': <P>
   * (3.1) if the 2 '%' have not the same location, the smaller location is after
   * ('on%' before 'o%'), <P>
   * (3.2) if the '%' location is 0: <P>
   * (3.2.1) the component '%' is never after ('%' before '%e'), <P>
   * (3.2.2) otherwise, the greater component is after ('%e' before '%ne'), <P>
   * (3.3) otherwise, the greater component is after ('o%' before 'o%e'), <P>
   *
   * This algorithm is based on MySQL server testing. It is not from the official documentation,
   * which declares simply the most-specific bases set is sorted first. <P>
   * Called by addAclNOTForDatabase(), addAclForDatabaseNOTForToBeMixedAcl()
   * and addAclForDatabaseForToBeMixedAcl().
   * @param _compo_one last component of a hosts group name of an ACL to sort.
   * Must contain exactly one wildcard '_' or '%'.
   * @param _compo_two last component of a hosts group name of an ACL to sort.
   * Must contain exactly one wildcard '_' or '%'.
   * @return true if the ACL of _compo_one has to be after the ACL of _compo_two,
   * following the AclEntry pattern-name hosts group sorting
   */
  public static boolean isAfterAsPatternNameBasesSet(
          String _compo_one, String _compo_two) {

                                      // for testing and tracking
                                      //System.out.println("BaseUtilityMySQLImpl."+
                                        //      "isAfterAsPatternNameBasesSet()"+
                                          //    " for _compo_one: "+_compo_one+
                                            //  " and _compo_two: "+_compo_two);

      if (_compo_one.indexOf("%") == -1 && _compo_one.lastIndexOf("_") == -1)
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameBasesSet() "+
                  "without one '%' or '_' in "+_compo_one);
      if (_compo_two.indexOf("%") == -1 && _compo_two.lastIndexOf("_") == -1)
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameBasesSet() "+
                  "without one '%' or '_' in "+_compo_two);

      if (_compo_one.indexOf("%") != _compo_one.lastIndexOf("%"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameBasesSet() "+
                  "with more than one '%' in "+_compo_one);
      if (_compo_two.indexOf("%") != _compo_two.lastIndexOf("%"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameBasesSet() "+
                  "with more than one '%' in "+_compo_two);
      if (_compo_one.indexOf("_") != _compo_one.lastIndexOf("_"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameBasesSet() "+
                  "with more than one '_' in "+_compo_one);
      if (_compo_two.indexOf("_") != _compo_two.lastIndexOf("_"))
          throw new InternalError("BaseUtilityMySQLImpl.isAfterAsPatternNameBasesSet() "+
                  "with more than one '_' in "+_compo_two);

      // the default value, so only true result is detected
      boolean   result = false;
      int       count_one = -1;
      int       count_two = -1;

      // (1) if a component contains '_' and the other component contains '%'...(case 1)
      if ( _compo_one.indexOf("%") != -1 && _compo_two.indexOf("%") == -1) {

          count_one = _compo_one.indexOf("%");
          count_two = _compo_two.indexOf("_");

                                      // for testing and tracking
                                      //System.out.println("BaseUtilityMySQLImpl."+
                                        //      "isAfterAsPatternNameBasesSet()"+
                                          //    " with count_one: "+count_one+
                                            //  " with count_two: "+count_two);


          // (1.1) with different locations, the smaller location is after
          if (count_one < count_two) {
              result = true;
          } // if
          // (1.2) otherwise, if the '%' location is 0
          else if (count_one == count_two && count_one == 0) {
              // (1.2.1) the component '%' is never after
              // (1.2.2) otherwise, the '%' component is after
              if (!_compo_one.equals("%")) {
                  result = true;
              }
          }
          // (1.3) otherwise, the '_' component is after
      }
      // (1) if a component contains '_' and the other component contains '%'...(case 2)
      else if ( _compo_one.indexOf("%") == -1 && _compo_two.indexOf("%") != -1 ) {

          count_one = _compo_one.indexOf("_");
          count_two = _compo_two.indexOf("%");
          // (1.1) with different locations, the smaller location is after
          if (count_one < count_two) {
              result = true;
          } // if
          // (1.2) otherwise, if the '%' location is 0
          else if (count_one == count_two && count_two == 0) {
              // (1.2.1) the component '%' is never after
              // (1.2.2) otherwise, the '%' component is after
              if (_compo_two.equals("%")) {
                  result = true;
              }
          }
          // (1.3) otherwise, the '_' component is after
          else if (count_one == count_two) {
              result = true;
          }
      }

      // (2) if both components contain '_'
      else if (_compo_one.indexOf("_") != -1 && _compo_two.indexOf("_") != -1) {

          count_one = _compo_one.indexOf("_");
          count_two = _compo_two.indexOf("_");
          // (2.1) if the 2 '_' have not the same location, the smaller location is after
          if (count_one < count_two) {
              result = true;
          } // if
          // (2.2) otherwise the components are equal
      }

      // (3) if both components contain '%'
      else if (_compo_one.indexOf("%") != -1 && _compo_two.indexOf("%") != -1) {

          count_one = _compo_one.indexOf("%");
          count_two = _compo_two.indexOf("%");

          // (3.1) if the 2 '%' have not the same location, the smaller location is after
          if (count_one < count_two) {
                  result = true;
          }
          else if (count_one == count_two) {

              // (3.2) if the '%' location is 0
              if (count_one == 0) {
                // (3.2.1) the component '%' is never after
                if (_compo_two.equals("%"))
                    result = true;
                // (3.2.2) otherwise, the greater component is after
                else if (_compo_one.length() > _compo_two.length())
                    result = true;
              }
              // (3.3) otherwise, the greater component is after
              else if (_compo_one.length() > _compo_two.length()) {
                  result = true;
              }

          } // else if (count_one == count_two)

      } // else if

                                      // for testing and tracking
                                      //System.out.println("BaseUtilityMySQLImpl."+
                                        //      "isAfterAsPatternNameBasesSet(),"+
                                         //     " result: "+result);

      return result;
  } // isAfterAsPatternNameBasesSet()

  /**
   * Detects for two types of a GroupIDMember without immutable, if the first argument
   * has to be after the second argument, following the MySQL sorting order.
   * For the sources, the order is 'account', 'host', 'pattern-name_accounts-for-hosts group',
   * the sorting among the accounts-for-hosts groups from their names,
   * then, 'pattern-name_hosts group',
   * the sorting among the hosts groups from their names,
   * and finally, 'standard pattern-name_accounts-for-hosts group',
   * 'standard pattern-name_hosts group' and 'standard for simple hosts'. <P>
   * Called by UserIDMySQLImpl and GroupIDMySQLImpl.addGroupToGM().
   * Calls isAfterAsPatternNameHostsGroup().
   * @param _typeOne type of a GroupIDMember without immutable. Nerver null.
   * @param _typeTwo type of a GroupIDMember without immutable. Nerver null.
   * @param _nameOne name last component of the GroupIDMember having _typeOne. Nerver null.
   * @param _nameTwo name last component of the GroupIDMember having _typeTwo. Nerver null.
   * @return true if _nameOne has to be after _nameTwo in the sorted list,
   * because it is least specific
   */
  public static boolean isAfterInTheGroupIDMemberOrder(
          String _typeOne, String _typeTwo, String _nameOne, String _nameTwo) {

      // default value
      boolean   result = true;
      String    nameOne = null;
      String    nameTwo = null;

      if (_typeOne.indexOf("view_delegate <byAcsCreator>") != -1 ||
          _typeOne.indexOf("trigger") != -1 ||
          _typeOne.indexOf("stored_function") != -1 ||
          _typeOne.indexOf("stored_procedure") != -1 )
            throw new InternalError("BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder()"+
                    " with a wrong argument "+_typeOne);
      if (_typeTwo.indexOf("view_delegate <byAcsCreator>") != -1 ||
          _typeTwo.indexOf("trigger") != -1 ||
          _typeTwo.indexOf("stored_function") != -1 ||
          _typeTwo.indexOf("stored_procedure") != -1 )
            throw new InternalError("BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder()"+
                    " with a wrong argument "+_typeTwo);

      // for having 'account' before, while 'pattern-name_accounts-for-hosts group' contain 'accounts'
      if (_typeOne.indexOf("account") != -1 && _typeOne.indexOf("accounts") == - 1)
          result = false;
      else if (_typeTwo.indexOf("account") != -1 && _typeTwo.indexOf("accounts") == -1)
          result = true;

      // for having 'host', 'host ', ' host' before, 'hosts group' and 'simple hosts' being after
      else if (_typeOne.indexOf("hosts group") == -1 && _typeOne.indexOf("simple hosts") == -1)
          result = false;
      else if (_typeTwo.indexOf("hosts group") == -1 && _typeTwo.indexOf("simple hosts") == -1)
          result = true;

      // for having 'accounts-for-hosts group' before
      else if (_typeOne.indexOf("accounts-for-hosts group") != -1 &&
              _typeOne.indexOf("standard") == -1 &&
              _typeTwo.indexOf("accounts-for-hosts group") == -1)
          result = false;
      else if (_typeOne.indexOf("accounts-for-hosts group") == -1 &&
              _typeTwo.indexOf("accounts-for-hosts group") != -1 &&
              _typeTwo.indexOf("standard") == -1)
          result = true;

      else if (_typeOne.indexOf("accounts-for-hosts group") != -1 &&
              _typeOne.indexOf("standard") == -1 &&
              _typeTwo.indexOf("accounts-for-hosts group") != -1 &&
              _typeTwo.indexOf("standard") == -1) {

          nameOne = _nameOne.substring(_nameOne.indexOf("§@") + 2);
          nameTwo = _nameTwo.substring(_nameTwo.indexOf("§@") + 2);

          result = isAfterAsPatternNameHostsGroup(nameOne, nameTwo);
          nameOne = null;
          nameTwo = null;
      } // if _typeOne.indexOf("accounts-for-hosts group") != -1

      // for having 'pattern-name_hosts group' before
      else if (_typeOne.indexOf("pattern-name_hosts group") != -1 &&
              _typeOne.indexOf("standard") == -1 &&
              _typeTwo.indexOf("pattern-name_hosts group") == -1)
          result = false;
      else if (_typeOne.indexOf("pattern-name_hosts group") == -1 &&
              _typeTwo.indexOf("pattern-name_hosts group") != -1 &&
              _typeTwo.indexOf("standard") == -1)
          result = true;

      else if (_typeOne.indexOf("pattern-name_hosts group") != -1 &&
              _typeOne.indexOf("standard") == -1 &&
              _typeTwo.indexOf("pattern-name_hosts group") != -1 &&
              _typeTwo.indexOf("standard") == -1) {

          result = isAfterAsPatternNameHostsGroup(_nameOne, _nameTwo);
      } // if _typeOne.indexOf("accounts-for-hosts group") != -1

      // for having 'standard accounts-for-hosts group' before
      else if (_typeOne.indexOf("standard accounts-for-hosts group") != -1 &&
              _typeTwo.indexOf("standard accounts-for-hosts group") == -1)
          result = false;
      // for having 'standard pattern-name_hosts group' before
      else if (_typeOne.indexOf("standard pattern-name_hosts group") == -1 &&
              _typeTwo.indexOf("standard pattern-name_hosts group") != -1)
          result = false;

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "isAfterInTheGroupIDMemberOrder() for "+
                                      //    "_nameOne"+_nameOne+" and _nameTwo "+_nameTwo+
                                       //   ", with result: "+result);

      return result;
  } // isAfterInTheGroupIDMemberOrder()



  /**
   * Currently unused. Gets the order of the last AclEntry in _upd_l_acl having an EligibleParty
   * with _type in its type. 
   * @param _l_acl list of AclEntries in the correct sorting from the EP type
   * @param _type an EligibleParty type without 'immutable'
   * @return the order of the last AclEntry in _upd_l_acl having an EligibleParty
   * with _type in its type, or -1 if _type is not found
   */
  public static int getTypeOfEPLastOrderInSortedAclList(
          ArrayList<AclEntry> _l_acl, String _type) {
      int       result = _l_acl.size() - 1;
      boolean   find = false;
      boolean   amongTheFirstTypes = false;
      AclEntry  acl = null;
      //String    type = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(_type);
      String    aclType = null;

      if (_type.equals("account"))
          amongTheFirstTypes = true;

      for (int i = _l_acl.size() - 1; i > -1; i--) {
          acl = _l_acl.get(i);
          aclType = UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                        acl.getEorEP().getEPType());

          if (aclType.equals(_type)) {
              find = true;
              break;
          }

          // stop at the middle of the type list, if it is possible,
          // to speed up the search. Works only if the sort of AclEntries
          // in _upd_l_acl is correct.
          if (aclType.equals("account") &&
                  !amongTheFirstTypes) {
              break;
          }

          result--;
      } // for i

      if (!find) result = -1;

      acl = null;
      aclType = null;
      //type = null;
      return result;
  } // getTypeOfEPLastOrderInSortedAclList()

    /**
    * This important method sets the internal maps of the inherited AclEntry rights
    * and the primary rights from the new inherited AclEntries list. These two internal maps
    * have lists of rights as values, and they are the two method arguments to update.
    * For the inherited ACL rights map and a given couple user/host and a target, 
    * the complete list of inherited rights is the addition of all the rights 
    * from the first keys at each parent level, if any, that matches the current user
    * of the couple user/host. The rights map keys of a parent level are not sorted
    * following the ACL sorting. This implies the rights maps are not self-suffisant
    * to define the effective rights of a couple user/host on a target.
    * They may be viewed as intermediate rights in the rights analysis process.
    * <P>
    * To take in account the ACLs from different parent level, the inherited ACLs of
    * a Resource are simply the full copy of all the ACL of its direct and indirect parents,
    * including for instance all the direct ACL on the database.
    * The sorting of the new inherited AclEntries list, for each parent level,
    * has been provided by addInSortedAclList(). This sorting drives the selection process
    * of the matched AclEntries at each parent level, for a given couple user/host and a target.
    * This selection is not handled by this current method. It is the responsibility of
    * the methods detectL_aclEntryRights() and getL_accessRightsThroughNodesTree()
    * on NodeRightsMySQLImpl, in the search of the access paths.
    * <P>
    * The rights, as items into the values of the two maps,
    * may be from a real ACL or from a virtual ACL. Only the real ACL
    * are displayed by the beamer, as conditional or non-conditional ACL.
    * A real ACL in the new inherited AclEntries list produces rights
    * directly into the two maps. The new inherited AclEntries list is also
    * used to produce some processed rights, as virtual ACL, into the two right maps.
    * <P>
    * The primary rights are not a MySQL concept. It is an help for the Access Road user.
    * They are not used by the gWork classes, to search for the access paths.
    * For the inherited AclEntries, the MySQL Resource description contains
    * a specific map for the primary rights from the real and virtual ACL.
    * A primary right for a Resource is delivered by the ACL list of the Resources root
    * 'MySQL server', or by the ACL list of the database which is the parent of the Resource.
    * The root 'MySQL server' has primary rights from its direct ACL, and a database
    * has primary rights from the root and its direct ACL. The primary rights are 
    * so designed to help the user to better understand both the current users
    * and the complex database rights.
    * <P>
    * To have some access rights on a MySQL Resource, a pair (user, host) has
    * to match on the root 'MySQL Server' a pair (user, host) or a pair (user, hosts group)
    * which has at least an USAGE right. The FIRST matching pair becomes the current user.
    * It is used to get the rights for the initial pair (user, host) at each parent level.
    * The current user is returned by the method NodeRightsMySQLImpl.getConnectionPair(),
    * but not by this current method where the rights selection is not considered.
    * If the current user has only an USAGE right and if, for the given Resource,
    * the current user has no direct ACL and no inherited ACL, then no right is granted
    * to the pair (user, host). Otherwise, the rights are the sum of the current user
    * rights from the FIRST matched direct ACL and from the FIRST matched inherited ACL
    * at EACH parent level. So, for these inherited ACL, the rights from each parent
    * are added to the final result. The first matched inherited ACL, for the current user,
    * may be real or virtual. The real ACL are the classical ACL displayed
    * in the beamer lists. 
    * <P>
    * This section is about the virtual ACL in the MySQL addon.
    * There are two types of virtual ACL: the virtual
    * 'user@accounts-for-hosts group' ACL on any Resource, which derives from
    * one 'accounts-for-hosts group' real ACL, and the mixed ACL only
    * on databases, which derives from two to-be-mixed real ACL.
    * The virtual user@accounts-for-hosts group ACL is so derived
    * from a non-conditional accounts-for-hosts group ACL, and the target may have any type.
    * There are processed rights for each account member of the accounts-for-hosts group.
    * If there are 3 members into the accounts-for-hosts group, there are 3 sets of
    * processed rights for three keys in the rights maps. All these rights are
    * simply the copy of the accounts-for-hosts group ACL rights. This is not a MySQL concept,
    * just as the accounts-for-hosts group is not.
    * <P>
    * The to-be-mixed real ACL are the only case where two real ACL are selected
    * at a level (this is indeed always the database level) to build up some rights
    * in a virtual mixed ACL. To-be-mixed and mixed ACL are not MySQL concepts,
    * but they implement a MySQL algorithm.
    * Two non-conditional to-be-mixed ACL on a database-specific target, respectively
    * for an user U and a host H, are combined to define the processed rights
    * for the pair (U, H). The processed rights are the union of the rights
    * from the two real ACL. This define a map key even if the union of rights is empty.
    * This is the only case where a key in the two maps, may have an empty set
    * of effective rights. The database level is the most complex case.
    * It is the only parent level to handle the virtual mixed ACL, and also to handle
    * the real bases set ACLs.
    * <P>
    * This section is about the keys into the primary rights map.
    * Each parent level may produce a matched key for a given couple user/host and a target.
    * We have seen that the ACL sorting at the parent level drives the selection process
    * of the matched key, in the class NodeRightsMySQLImpl.
    * For the inherited ACL rights and the primary ACL rights,
    * the map key may have varied forms, but the parent order is always at the end.
    * The parent order tells the distance from the target to the parent which delivers the rights.
    * To set the parent order as a number, the direct parent of a resource has the string '1'.
    * The direct parent of this parent has the string '2' as parent order
    * in the last name component, and so on. 
    * Into the two maps, the first difference between the map keys is that the key for
    * inherited ACL rights is a Name, and the primary ACL rights key is the String form
    * of this same Name. The second and last difference is to allow a parent order '0'
    * into a key for the primary rights of the root 'MySQL server' or a database, 
    * while this is forbidden for the inherited ACL rights.<P>
    * The ten key forms are listed hereinafter: <P>
    * - (user name @ host name last component) ::  parent order, 
    * for instance 'one :: mysql :: jerry@localhost :: 2',<P>
    * - &lt;Proc&gt;:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '&lt;Proc&gt;:: one :: mysql :: jerry(2)@192.168.1.% :: 0',<P>
    * '&lt;Proc&gt;' is a generic tag that means 'processed rights', to tell the key and
    * its rights are not get from the direct reading of a real ACL.
    * These two first forms are the only used ones when the parent level is not a database.
    * These forms are also used for a database, but the followings are additional forms in this case:
    * - &lt;Proc&gt;&lt;from 'bases set name last component'&gt;:: (user name @ host name last component):: parent order, <P>
    * - &lt;Proc&gt;&lt;from 'bases set name last component'&gt;:: (user name @ accounts-for-hosts group name
    * last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;:: (user name @ host name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'&gt;:: (user name @ host name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'>&lt;from 'bases set name last component'&gt;::
    * (user name @ host name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;:: (user name @ hosts group name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'&gt;:: (user name @ hosts group name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'>&lt;from 'bases set name last component'&gt;::
    * (user name @ hosts group name last component) :: parent order.
    * <P>
    * In this catalog of key forms for the database, '&lt;Mixed&gt;' means this is a virtual mixed ACL,
    * and '&lt;from 'bases set name last component'&gt;' means a bases set ACL is the real ACL
    * from which the virtual ACL is derived. The tag '&lt;Mixed&gt;' is proper to MySQL
    * but it may be reused in another ACS Addons. The tag '&lt;from 'xxxx'v' is a generic form
    * another ACS Addon may use.
    * A key contains two times '&lt;from 'bases set name last component'&gt;' when there are
    * both a bases set ACL for the user, and a bases set ACL for the host
    * or the hosts group. These key forms are designed to help the Access Road user
    * to find the origin of a key into a rights map.
    * A key has exactly one origin, which is one real ACL in most of the cases,
    * except the virtual mixed ACL where two real to-be-mixed ACL are processed.
    * We have seen that the rights map keys of a parent level are not sorted
    * following the ACL sorting, or following any other sorting.
    * The value in a rights map is always a non-null StringRight array.
    * This array may be empty only for a mixed ACL. This array may contain only
    * an 'USAGE' right when the key is provided by an ACL on the root 'MySQL server',
    * and this single right does not generate primary rights for the other targets.
    * <P>
    * The overall rights analysis process for the inherited ACL is described
    * into the documentation of the class DirectoryMySQLImpl.
    * The documentation of the method DirectoryMySQLImpl.addAclEntry() describes also
    * the calling sequence when an ACL is created, including the roles of the classes
    * DirectoryMySQLImpl, BaseSetMySQLImpl, ResourceMySQLImpl and AclEntryImpl.
    * Called by the non-static methods having the same name in ResourceMySQLImpl,
    * ActorMySQLImpl, DirectoryMySQLImpl, BaseSetMySQLImpl.
    * This method is mainly a commutator for the methods
    * setM_TargetInheritedACLRightsNOTForDatabaseAcl(),
    * setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() and
    * setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl(), to handle
    * the three following independant parts of the _l_acl list: <P>
    * (1) all the ACL where the target is not a database-specific target (database or base set), <P>
    * (2) all the ACL where the target is a database-specific target and the ACL is NOT a to-be-mixed ACL, <P>
    * (3) all the ACL where the target is a database-specific target and the ACL is a to-be-mixed ACL. <P>.
    * @see DirectoryMySQLImpl
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _res resource to update
    * @param _l_acl list of AclEntries, which is the current list
    * of inherited or direct AclEntries of _res. May be empty, but never null.
    * @param _upd_m_globalT_ACLRights _res map to update. May be null and no updating then.
    * Otherwise, this argument is cleared before the updating.
    * @param _upd_m_primaryT_ACLRights _res primary rights map to update.
    * May be empty, but never null.
    * @param _forDirectACL false if _l_acl is a list of inherited ACL, true if
    * _l_acl is a list of direct ACL
    */
  public static void setM_TargetInheritedACLRights(
         ResourceMySQL _res, List<AclEntry> _l_acl,
         Map<ImmutableName, StringRight[]> _upd_m_globalT_ACLRights,
         Map<String, StringRight[]> _upd_m_primaryT_ACLRights,
         boolean _forDirectACL) {

                                              // for testing and tracking
                                              //System.out.println("BaseUtilityMySQLImpl."+
                                                //      "setM_TargetInheritedACLRights() "+
                                                  //    "starts for the Resource "+
                                                    //  _res.getNickName());

     //if (_res.getType().indexOf("bases set") != -1 ||
       //      _res.getType().equals("virtual_objects <byAcsCreator>"))
         // throw new InternalError("BaseUtilityMySQLImpl."+
           //   "setM_TargetInheritedACLIRights(), the resource argument "+
             // "cannot be a bases set or its parent.");

     Map.Entry entry = null;

     if (_upd_m_globalT_ACLRights != null)
         _upd_m_globalT_ACLRights.clear();

     if (!_forDirectACL) { 
              // _upd_m_primaryT_ACLRights is updated for inherited ACL
              // with keys which does NOT end with '0  '.
              // For those keys, the primary rights map is cleared,
              // while all the other ones ending with '0  ' are not cleared.
          if (!_upd_m_primaryT_ACLRights.isEmpty()) {
              for (Iterator ite = _upd_m_primaryT_ACLRights.entrySet().iterator(); ite.hasNext();) {
                    entry = (Map.Entry) ite.next();

                    if ( ((String)entry.getKey()).endsWith("0::  ") ) {
                      continue;
                    }

                                              // for testing and tracking
                                              //System.out.println("BaseUtilityMySQLImpl."+
                                                //      "setM_TargetInheritedACLRights() "+
                                                 //     "removes first the key "+entry.getKey());

                    ite.remove();
              } // for ite
          } // if (!_upd_m_primaryT_ACLRights_.isEmpty())
     } // else
     else { // for updating the direct ACL
              // _upd_m_primaryT_ACLRights is updated for direct ACL
              // with keys which ends with '0  '.
              // For those keys, the primary rights map is cleared,
              // while all the other ones are not cleared.
          if (!_upd_m_primaryT_ACLRights.isEmpty()) {
              for (Iterator ite = _upd_m_primaryT_ACLRights.entrySet().iterator(); ite.hasNext();) {
                    entry = (Map.Entry) ite.next();

                    if ( !((String)entry.getKey()).endsWith("0::  ") ) {
                      continue;
                    }

                                              // for testing and tracking
                                              //System.out.println("BaseUtilityMySQLImpl."+
                                                //      "setM_TargetInheritedACLRights() "+
                                                  //    "removes first the key "+entry.getKey());

                    ite.remove();
              } // for ite
          } // if (!_upd_m_primaryT_ACLRights_.isEmpty())
     } // else

     if (_l_acl.isEmpty()) {
         return;
     }

     setM_TargetInheritedACLRightsNOTForDatabaseAcl(_res,
        _l_acl, _upd_m_globalT_ACLRights, _upd_m_primaryT_ACLRights);

     setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl(_res,
         _l_acl, _upd_m_primaryT_ACLRights);

     // there are also to-be-mixed ACL in the inherited ACL list of a database child
     setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl(_res,
         _l_acl, _upd_m_primaryT_ACLRights);

     entry = null;
  } // setM_TargetInheritedACLRights()

    /**
    * This method sets the internal maps of the inherited AclEntry rights
    * and the primary rights from the new inherited AclEntries list,
    * when the AclEntries are not for the database. These two internal maps
    * have lists of rights as values, and they are the two method arguments to update.
    * For the inherited ACL rights map and a given couple user/host and a target, 
    * the complete list of inherited rights is the addition of all the rights 
    * from the first keys at each parent level, if any, that matches the current user
    * of the couple user/host. The rights map keys of a parent level are not sorted
    * following the ACL sorting. This implies the rights maps are not self-suffisant
    * to define the effective rights of a couple user/host on a target.
    * They may be viewed as intermediate rights in the rights analysis process. <P>
    * The two key forms this method handles are listed hereinafter: <P>
    * - (user name @ host name last component) ::  parent order,
    * for instance 'one :: mysql :: jerry@localhost :: 2',<P>
    * - &lt;Proc&gt;:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '&lt;Proc&gt;:: one :: mysql :: jerry(2)@192.168.1.% :: 0',<P>
    * '&lt;Proc&gt;' is a generic tag that means 'processed rights', to tell the key and
    * its rights are not get from the direct reading of a real ACL.
    * Only the root 'MySQL server' may have as rights the single 'USAGE' right,
    * in the value arrays of the two rights maps.
    * Called by setM_TargetInheritedACLRights().
    * Calls addSimplestKeysInM_TargetInheritedACLRights().
    * @see #setM_TargetInheritedACLRights(ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.ResourceMySQL, java.util.List, java.util.Map, java.util.Map, boolean)
    * @param _res resource to update
    * @param _l_acl list of AclEntries, which is the current list
    * of inherited or direct (only for the root 'MySQL server') AclEntries of _res.
    * May be empty, but never null.
    * @param _upd_m_globalT_ACLRights inherited ACL map of _res to update.
    * May be null and no updating then. No key nor value removing on this argument.
    * @param _upd_m_primaryT_ACLRights primary ACL map of _res to update.
    * May be empty, but never null. No key nor value removing on this argument.
    * No adding for the database-specific ACL keys.
    *
    */
  public static void setM_TargetInheritedACLRightsNOTForDatabaseAcl(
         ResourceMySQL _res, List<AclEntry> _l_acl,
         Map<ImmutableName, StringRight[]> _upd_m_globalT_ACLRights,
         Map<String, StringRight[]> _upd_m_primaryT_ACLRights) {

      StringRight[]     l_rights = null;
      String            order = "-1";
      boolean           error = false;
      AclEntry          acl_error = null;

                                      // for testing and tracking
                                      //System.out.println("Call to BaseUtilityMySQLImpl."+
                                        //      "setM_TargetInheritedACLRightsNOTForDatabaseAcl() in "+
                                          //    _res.getNickName());
      
      try {
         for (AclEntry acl : _l_acl) {

            // no processing nor removing for databases ACL
            if (acl.getResType().indexOf("database") != -1 ||
                    acl.getResType().indexOf("bases set") != -1)
                continue;
            l_rights = acl.getL_Rights(); // may be empty

            // for a conditional ACL, empty rights if the source is not member
            // of the condition groups
            if (l_rights.length == 0 ||
                    (!acl.getResType().equals("MySQL server <byAcsCreator>") &&
                    l_rights.length == 1 && l_rights[0].getRight().equals("USAGE"))) {

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "setM_TargetInheritedACLRightsNOTForDatabaseAcl() "+
                                  //    "NO processing for the null-rights ACL "+acl);

                continue;
            } // if

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "setM_TargetInheritedACLRightsNOTForDatabaseAcl() "+
                                  //    "processing for the non-null-rights ACL "+acl);

            if (acl.getResName().equals(_res.getName()))
                order = "0";
            else
                order = UtilityImpl.getFirstInstance().getParentOrder(
                  _res,
                  (ImmutableNode) UtilityImpl.getFirstInstance().getBaseObject(
                    acl.getResName()));

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() "+
                                  //    ", parent order = "+order+
                                    //  ", Integer.valueOf(order) = "+Integer.valueOf(order));

            if (Integer.valueOf(order) < 0) {
              error = true;
              acl_error = acl;
              break;
            }

            addSimplestKeysInM_TargetInheritedACLRights(acl, order, l_rights,
                _upd_m_globalT_ACLRights,
                _upd_m_primaryT_ACLRights);

         } // for acl
      } catch (NumberFormatException n) {
          throw new InternalError("in BaseUtilityMySQLImpl."+
                  "setM_TargetInheritedACLRightsNOTForDatabaseAcl(), "+
                    "the exception for "+order+" is: "+n.toString());
      }

      if (error)
          throw new InternalError("BaseUtilityMySQLImpl."+
              "setM_TargetInheritedACLIRightsNOTForDatabase(), the target of an ACL "+
              "is not correct, with _res: "+_res.getName()+
              " with order: "+order+" and the acl resource: "+acl_error.getResName());

      l_rights = null;
      order = null;
  } // setM_TargetInheritedACLRightsNOTForDatabaseAcl()

    /**
    * This method sets the primary rights from the new inherited AclEntries list,
    * when the AclEntries are for the database and are not to-be-mixed. This internal map
    * has lists of rights as values, and it is a method argument to update.
    * For the rights map and a given couple user/host and a target,
    * the complete list of inherited rights is the addition of all the rights
    * from the first keys at each parent level, if any, that matches the current user
    * of the couple user/host. The rights map keys of a parent level are not sorted
    * following the ACL sorting. This implies the rights maps are not self-suffisant
    * to define the effective rights of a couple user/host on a target.
    * They may be viewed as intermediate rights in the rights analysis process. <P>
    * The two first key forms this method handles are listed hereinafter: <P>
    * - (user name @ host name last component) ::  parent order,
    * for instance 'one :: mysql :: jerry@localhost :: 2', <P>
    * - &lt;Proc&gt;:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '&lt;Proc&gt;:: one :: mysql :: jerry(2)@192.168.1.% :: 0', <P>
    * '&lt;Proc&gt;' is a generic tag that means 'processed rights', to tell the key and
    * its rights are not get from the direct reading of a real ACL.
    * These previous forms are completed by additional forms excluding to-be-mixed ACL:
    * - &lt;Proc&gt;&lt;from 'bases set name last component'&gt;:: (user name @ host name last component):: parent order, <P>
    * - &lt;Proc&gt;&lt;from 'bases set name last component'&gt;:: (user name @ accounts-for-hosts group name
    * last component) :: parent order, <P>
    * Called by setM_TargetInheritedACLRights(). Derived from setM_TargetInheritedACLRightsNOTForDatabaseAcl().
    * Calls addSimplestKeysInM_TargetInheritedACLRights() for the two first key forms.
    * @see #setM_TargetInheritedACLRights(ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.ResourceMySQL, java.util.List, java.util.Map, java.util.Map, boolean)
    * @param _res resource to update
    * @param _l_acl list of AclEntries, which is the current list of inherited AclEntries of _res,
    * or direct AclEntries only when _res is a database. May be empty, but never null.
    * @param _upd_m_primaryT_ACLRights primary ACL map of _res to update.
    * May be empty, but never null. No key nor value removing on this argument.
    */
  public static void setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl(
         ResourceMySQL _res, List<AclEntry> _l_acl,
         Map<String, StringRight[]> _upd_m_primaryT_ACLRights) {

      StringRight[]     l_rights = null;
      Set<StringRight>  l_newRights = new HashSet(); // = 16
      String            order = "-1";
      String            singleOrder = "-1";
      boolean           error = false;
      boolean           error2 = false;
      Name              keyName = null;
      ImmutableGroupID  grp = null;
      ImmutableGroupIDMember[] l_users = null;
      String            primaryKey = null;
      ImmutableName     parentBaseName = null;
      ImmutableResource parentName = null;

      //if (_res.getType().indexOf("database") != -1)
          //throw new InternalError("BaseUtilityMySQLImpl."+
            //  "setM_TargetInheritedACLIRightsForDatabaseNOTForToBeMixedAcl(), "+
              //"canot be called for a database as resource argument.");

                                      // for testing and tracking
                                      //System.out.println("Call to BaseUtilityMySQLImpl."+
                                        //      "setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() in "+
                                          //    _res.getNickName());

      try {
         for (AclEntry acl : _l_acl) {

            // only for database-specific targets
            if (acl.getResType().indexOf("database") == -1 &&
                    acl.getResType().indexOf("bases set") == -1)
                continue;
            // conditions to avoid the to-be-mixed ACL
            if (!acl.isConditionalAclEntry() &&
                    acl.getEorEP().getEPType().indexOf("accounts-for-hosts group") == -1)
                continue;

            l_rights = acl.getL_Rights(); // may be empty

            // for a conditional ACL, noop if the source is not member of the condition groups
            if (l_rights.length == 0 ||
                    (l_rights.length == 1 && l_rights[0].getRight().equals("USAGE"))) {

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() "+
                                  //    "NO processing for the null-rights ACL "+acl);

                continue;
            } // if

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() "+
                                  //    "processing for the non-null-rights ACL "+acl);

            if (acl.getResType().indexOf("database") != -1) {
                parentBaseName = acl.getResName();
            }
            else { // searching for the database parent of _res
                parentName = _res;
                parentBaseName = parentName.getName();
                while (parentName.getType().indexOf("database") == -1) {
                    parentName = (ImmutableResource) parentName.getEorParent();
                    parentBaseName = parentName.getName();
                } // while
            } // else

            if (parentBaseName.equals(_res.getName()))
                order = "0";
            else
                order = UtilityImpl.getFirstInstance().getParentOrder(
                  _res,
                  (ImmutableNode) UtilityImpl.getFirstInstance().getBaseObject(parentBaseName));

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() "+
                                  //    ", parent order = "+order+
                                    //  ", Integer.valueOf(order) = "+Integer.valueOf(order));

            if (Integer.valueOf(order) < 0) {
              error = true;
              break;
            }
            if (singleOrder.equals("-1")) {
                singleOrder = order;
            } else if (!singleOrder.equals(order)) {
              // for _res and for the database-specific targets, 'order' must be constant
              error2 = true;
              break;
            }

            if (acl.getResType().indexOf("database") != -1) {

                    addSimplestKeysInM_TargetInheritedACLRights(acl, order, l_rights,
                        null,
                        _upd_m_primaryT_ACLRights);
            }
            else {
                // for bases set only, derived from addSimplestKeysInM_TargetInheritedACLRights()

                // <Proc><from 'bases set name last component'>::
                    // (user name @ host name last component):: parent order
              if (acl.isConditionalAclEntry()) {
                    grp = (GroupID)acl.getL_ConditionSourceGroups().get(0);

                    keyName = new NameImpl("<Proc><from "+acl.getResName().lastComponent()+">");
                    keyName.extended(((Name)acl.getEpName().clone()).context()).
                          extended(acl.getEpName().lastComponent()+
                            "@"+
                            grp.getName().lastComponent()).extended(
                            order);
                    primaryKey = keyName.toString();

                    // always a database-specific there
                    if (_upd_m_primaryT_ACLRights.get(primaryKey) == null) {

                        _upd_m_primaryT_ACLRights.put(
                              primaryKey,
                              l_rights);

                    } // if
                    else {
                        l_newRights.clear();
                        l_newRights.addAll(
                                Arrays.asList(_upd_m_primaryT_ACLRights.get(primaryKey)));
                        l_newRights.addAll(Arrays.asList(l_rights));

                        _upd_m_primaryT_ACLRights.put(
                              primaryKey,
                              l_newRights.toArray(new StringRight[0]));

                    } // else
              }
                // <Proc><from 'bases set name last component'>::
                    // (user name @ accounts-for-hosts group name last component without '§') ::
                    // parent order
                // for every user which is member of the accounts-for-hosts group having an ACL
              else {

                    grp = (ImmutableGroupID) acl.getEorEP();
                    l_users = grp.getEorL_Members();

                    for (int i = 0; i < l_users.length; i++) {
                        keyName = new NameImpl("<Proc><from "+acl.getResName().lastComponent()+">");
                        keyName.extended(((Name)l_users[i].getName()).context()).
                                extended(l_users[i].getName().lastComponent()+
                                "@"+
                                grp.getName().lastComponent().substring(
                                    grp.getName().lastComponent().indexOf("@")+1)).
                                extended(order);
                        primaryKey = keyName.toString();

                        // always a database-specific there
                        if (_upd_m_primaryT_ACLRights.get(primaryKey) == null) {

                            _upd_m_primaryT_ACLRights.put(
                                  primaryKey,
                                  l_rights);

                        } // if
                        else {
                            l_newRights.clear();
                            l_newRights.addAll(
                                    Arrays.asList(_upd_m_primaryT_ACLRights.get(primaryKey)));
                            l_newRights.addAll(Arrays.asList(l_rights));

                            _upd_m_primaryT_ACLRights.put(
                                  primaryKey,
                                  l_newRights.toArray(new StringRight[0]));

                        } // else

                    } // for i on grp members

              } // else on 'if (_acl.isConditionalAclEntry())'
            } // else

         } // for acl
      } catch (NumberFormatException n) {
          throw new InternalError("in BaseUtilityMySQLImpl."+
                  "setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl(), "+
                    "the exception for "+order+" is: "+n.toString());
      }

      if (error)
          throw new InternalError("BaseUtilityMySQLImpl."+
              "setM_TargetInheritedACLIRightsForDatabaseNOTForToBeMixedAcl(), "+
              "the target of an ACL is not the resource or one of its parents, "+
              "or the non-conditional ACL "+
              "is not for an accounts-for-hosts group.");
      if (error2)
          throw new InternalError("BaseUtilityMySQLImpl."+
              "setM_TargetInheritedACLIRightsForDatabaseNOTForToBeMixedAcl(), "+
              "the order is not constant. Current order: "+order);

      l_rights = null;
      l_newRights = null;
      order = null;
      keyName = null;
      grp = null;
      l_users = null;
      primaryKey = null;
  } // setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl()

    /**
    * This method sets the primary rights from the new inherited AclEntries list,
    * when the AclEntries are for the database and are to-be-mixed. This internal map
    * has lists of rights as values, and it is a method arguments to update.
    * For the rights map and a given couple user/host and a target,
    * the complete list of inherited rights is the addition of all the rights
    * from the first keys at each parent level, if any, that matches the current user
    * of the couple user/host. The rights map keys of a parent level are not sorted
    * following the ACL sorting. This implies the rights maps are not self-suffisant
    * to define the effective rights of a couple user/host on a target.
    * They may be viewed as intermediate rights in the rights analysis process.<P>
    * The six key forms this method handles for the virtual mixed ACL are listed hereinafter: <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;:: (user name @ host name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'&gt;:: (user name @ host name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'&gt;&lt;from 'bases set name last component'&gt;::
    * (user name @ host name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;:: (user name @ hosts group name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'&gt;:: (user name @ hosts group name last component) :: parent order, <P>
    * - &lt;Proc&gt;&lt;Mixed&gt;&lt;from 'bases set name last component'&gt;&lt;from 'bases set name last component'&gt;::
    * (user name @ hosts group name last component) :: parent order.
    * <P>
    * In this catalog of key forms for the database, '&lt;Mixed&gt;' means
    * this is a virtual mixed ACL, and '&lt;from 'bases set name last component'&gt;'
    * means a bases set ACL is the real ACL from which the virtual ACL is derived.
    * A key contains two times '&lt;from 'bases set name last component'&gt;' when there are
    * both a real bases set ACL for the user, and a real bases set ACL for the host
    * or the hosts group. The value in a rights map is always a non-null StringRight array.
    * This rights array may be empty only for a mixed ACL. <P>
    * Called by setM_TargetInheritedACLRights().
    * Derived from setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl().
    * @see #setM_TargetInheritedACLRights(ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.ResourceMySQL, java.util.List, java.util.Map, java.util.Map, boolean) 
    * @param _res resource to update
    * @param _l_acl list of AclEntries, which is the current list of inherited AclEntries of _res,
    * or direct AclEntries only when _res is a database. May be empty, but never null.
    * @param _upd_m_primaryT_ACLRights primary ACL map of _res to update.
    * May be empty, but never null. No key nor value removing on this argument.
    */
  public static void setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl(
         ResourceMySQL _res, List<AclEntry> _l_acl,
         Map<String, StringRight[]> _upd_m_primaryT_ACLRights) {

      StringRight[]     l_rights = null;
      Set<StringRight>  l_newRights = new HashSet(); // = 16
      String            order = "-1";
      String            singleOrder = "-1";
      boolean           error = false;
      boolean           error2 = false;
      Name              keyName = null;
      ImmutableGroupID  grp = null;
      String            primaryKey = null;
      ImmutableName     parentBaseName = null;
      ImmutableResource parentName = null;
      Set<AclEntry>     l_userNonConditionalACL = new HashSet(INITIAL_CAPACITY*5); // = 50
      Set<AclEntry>     l_hostNonConditionalACL = new HashSet(INITIAL_CAPACITY*5); // = 50
      ImmutableUserID   usr = null;
      StringRight       right = null;
      
      //if (_res.getType().indexOf("database") != -1)
        //  throw new InternalError("BaseUtilityMySQLImpl."+
          //    "setM_TargetInheritedACLIRightsForDatabaseForToBeMixedAcl(), "+
            //  "cannot be called for a database as resource argument.");

                                      // for testing and tracking
                                      //System.out.println("Call to BaseUtilityMySQLImpl."+
                                        //      "setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl() in "+
                                          //    _res.getNickName());

      try {
        for (AclEntry acl : _l_acl) {

            // only for database-specific targets
            if (acl.getResType().indexOf("database") == -1 &&
                    acl.getResType().indexOf("bases set") == -1)
                continue;

            // initialize l_userNonConditionalACL and l_hostNonConditionalACL
            // to prepare the processing of the real to-be-mixed ACL
            if (!acl.isConditionalAclEntry() &&
                    acl.getEorEP().getEPType().indexOf("accounts-for-hosts group") == -1) {

                if (acl.getEorEP() instanceof UserID)
                    l_userNonConditionalACL.add(acl);
                else if (acl.getEorEP() instanceof GroupID)
                    l_hostNonConditionalACL.add(acl);

                              // for testing and tracking
                              //System.out.println("BaseUtilityMySQLImpl."+
                                //      "setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl() "+
                                  //    "to-be-mixed ACL "+acl.getEpName());

            } // if (!acl.isConditionalAclEntry()
        } // for acl

        // no mixed ACL to add
        if (l_userNonConditionalACL.isEmpty() || l_hostNonConditionalACL.isEmpty()) {
            l_newRights = null;
            l_userNonConditionalACL = null;
            l_hostNonConditionalACL = null;
            return;
        } // if

        // add the virtual mixed ACLs in the inherited and primary rights maps.
        for (AclEntry u : l_userNonConditionalACL) {
            for (AclEntry h : l_hostNonConditionalACL) {

                grp = (ImmutableGroupID) h.getEorEP();
                usr = (ImmutableUserID) u.getEorEP();

                if (u.getResType().indexOf("database") != -1) {
                    parentBaseName = u.getResName();
                }
                else { // searching for the database parent of _res
                    parentName = _res;
                    parentBaseName = parentName.getName();
                    while (parentName.getType().indexOf("database") == -1) {
                        parentName = (ImmutableResource) parentName.getEorParent();
                        parentBaseName = parentName.getName();
                    } // while
                } // else

                if (parentBaseName.equals(_res.getName()))
                    order = "0";
                else
                    order = UtilityImpl.getFirstInstance().getParentOrder(
                          _res,
                          (ImmutableNode) UtilityImpl.getFirstInstance().getBaseObject(
                            parentBaseName));

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() "+
                                      //    ", parent order = "+order+
                                        //  ", Integer.valueOf(order) = "+Integer.valueOf(order));

                if (Integer.valueOf(order) < 0) {
                  error = true;
                  break;
                }
                if (singleOrder.equals("-1")) {
                    singleOrder = order;
                } else if (!singleOrder.equals(order)) {
                  // for _res and for the database-specific targets, 'order' must be constant
                  error2 = true;
                  break;
                }

                // creates the 6 following map keys:
                // 1. <Proc><Mixed>:: user name @ host name last component) :: parent order, <P>
                // 2. <Proc><Mixed><from 'bases set name last component'>:: user name @ host name last component) :: parent order, <P>
                // 3. <Proc><Mixed><from 'bases set name last component'><from 'bases set name last component'>::
                    // user name @ host name last component) :: parent order, <P>
                // 4. <Proc><Mixed>:: user name @ hosts group name last component) :: parent order, <P>
                // 5. <Proc><Mixed><from 'bases set name last component'>::
                    // user name @ hosts group name last component) :: parent order, <P>
                // 6. <Proc><Mixed><from 'bases set name last component'><from 'bases set name last component'>::
                    // user name @ hosts group name last component) :: parent order,

                // The host has to be an (in)direct group of the user
                if (usr.getEorL_GroupID().length == 0 ||
                        (!grp.containsAsMember(usr) && !grp.containsAsIndirectMember(usr)))
                    continue;

                // The rights of this mixed ACL are the intersection
                // of the rights in the two non-conditional ACLs
                if (Collections.disjoint(
                        Arrays.asList(u.getL_Rights()),
                        Arrays.asList(h.getL_Rights())))
                            continue;

                l_newRights.clear();
                l_newRights.addAll(Arrays.asList(u.getL_Rights()));
                for (Iterator ite = l_newRights.iterator(); ite.hasNext(); ) {
                    right = (StringRight) ite.next();
                    if (!Arrays.asList(h.getL_Rights()).contains(right))
                        ite.remove();
                } // for ite

                l_rights = l_newRights.toArray(new StringRight[0]);
                if (l_rights.length == 1 && l_rights[0].getRight().equals("USAGE")) {

                                  // for testing and tracking
                                  //System.out.println("BaseUtilityMySQLImpl."+
                                    //      "setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl() "+
                                      //    "NO processing for the USAGE-right for the mixed ACL from "+
                                        //  u+" and "+h);

                    continue;
                } // if

                // 1 and 4. <Proc><Mixed>:: user name @ host (4: group) name last component) ::
                    // parent order
                if (u.getResType().indexOf("bases set") == -1 &&
                        h.getResType().indexOf("bases set") == -1) {
                    keyName = new NameImpl("<Proc><Mixed>");
                    keyName.extended(usr.getName().context()).
                              extended(usr.getName().lastComponent()+
                                "@"+
                                grp.getName().lastComponent()).
                                extended(order);
                    primaryKey = keyName.toString();
                } // if
                // 2 and 5. <Proc><Mixed><from 'bases set name last component'>::
                    // user name @ host (5: group) name last component) ::
                    // parent order
                else if (u.getResType().indexOf("bases set") != -1 &&
                        h.getResType().indexOf("bases set") == -1) {
                    keyName = new NameImpl("<Proc><Mixed><from "+
                                    // 'u' as bases set
                                    u.getResName().lastComponent()+">");
                    keyName.extended(usr.getName().context()).
                              extended(usr.getName().lastComponent()+
                                "@"+
                                grp.getName().lastComponent()).
                                extended(order);
                    primaryKey = keyName.toString();
                } // if
                // 2 and 5. <Proc><Mixed><from 'bases set name last component'>::
                    // user name @ host (5: group) name last component) ::
                    // parent order
                else if (u.getResType().indexOf("bases set") == -1 &&
                        h.getResType().indexOf("bases set") != -1) {
                    keyName = new NameImpl("<Proc><Mixed><from "+
                                    // 'h' as bases set
                                    h.getResName().lastComponent()+">");
                    keyName.extended(usr.getName().context()).
                              extended(usr.getName().lastComponent()+
                                "@"+
                                grp.getName().lastComponent()).
                                extended(order);
                    primaryKey = keyName.toString();
                } // if
                // 3 and 6. <Proc><Mixed><from 'bases set name last component'>
                    // <from 'bases set name last component'>::
                    // user name @ host (6: group) name last component) ::  
                    // parent order
                else if (u.getResType().indexOf("bases set") != -1 &&
                        h.getResType().indexOf("bases set") != -1) {
                    keyName = new NameImpl("<Proc><Mixed><from "+
                                    // 'u' as bases set
                                    u.getResName().lastComponent()+
                                    "><from "+
                                    // 'h' as bases set
                                    h.getResName().lastComponent()+
                                    ">");
                    keyName.extended(usr.getName().context()).
                              extended(usr.getName().lastComponent()+
                                "@"+
                                grp.getName().lastComponent()).
                                extended(order);
                    primaryKey = keyName.toString();
                } // if

                                             // for testing and tracking
                                        //System.out.println(
                                          //    "BaseUtilityMySQLImpl.setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl(), "+
                                            //  "for matched key : "+keyName);
                                        //for (StringRight ri : l_newRights) {
                                          //    System.out.println("----- add the inherited right "+ri);
                                        //} // for ri

                _upd_m_primaryT_ACLRights.put(
                        primaryKey,
                        l_rights);
            } // for h

            if (error)
                break;
            if (error2)
                break;
        } // for u
      } catch (NumberFormatException n) {
          throw new InternalError("in BaseUtilityMySQLImpl."+
                  "setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl(), "+
                    "the exception for "+order+" is: "+n.toString());
      }

      if (error)
          throw new InternalError("BaseUtilityMySQLImpl."+
              "setM_TargetInheritedACLIRightsForDatabaseForToBeMixedAcl(), the target of an ACL "+
              "is not the resource or one of its parents, or the non-conditional ACL "+
              "is not for an accounts-for-hosts group.");
      if (error2)
          throw new InternalError("BaseUtilityMySQLImpl."+
              "setM_TargetInheritedACLIRightsForDatabaseForToBeMixedAcl(), "+
              "the order is not constant. Current order: "+order);


      l_rights = null;
      l_newRights = null;
      order = null;
      keyName = null;
      grp = null;
      primaryKey = null;
      l_userNonConditionalACL = null;
      l_hostNonConditionalACL = null;
      right = null;
  } // setM_TargetInheritedACLRightsForDatabaseAclForToBeMixedAcl()

    /**
    * This method sets the internal maps of the global rights
    * and the primary rights, only for the simpliest AclEntries. These two internal maps
    * have lists of rights as values, and they are the two method arguments to update.
    * For the primary ACL rights map and a given couple user/host and a target,
    * the complete list of inherited rights is the addition of all the rights 
    * from the first keys at each parent level, if any, that matches the current user
    * of the couple user/host. The rights map keys of a parent level are not sorted
    * following the ACL sorting. This implies the rights maps are not self-suffisant
    * to define the effective rights of a couple user/host on a target.
    * The global rights may be viewed as intermediate rights in the rights analysis process.
    * The main role of the primary rights is to explain to the user the resulting rights. <P>
    * The two simple key forms this method handles are the most common ones.
    * They are listed hereinafter: <P>
    * - (user name @ host name last component) ::  parent order,
    * for instance 'one :: mysql :: jerry@localhost :: 2',<P>
    * - &lt;Proc&gt;:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '&lt;Proc&gt;:: one :: mysql :: jerry(2)@192.168.1.% :: 0',<P>
    * '&lt;Proc&gt;' is a generic tag that means 'processed rights', to tell the key
    * and its rights are not get from the direct reading of a real ACL.
    * Called by setM_TargetInheritedACLRightsNOTForDatabaseAcl(),
    * setM_TargetInheritedACLRightsForDatabaseAclNOTForToBeMixedAcl() and
    * setM_PrimaryTargetACLRightsFromDirectACL().
    * @see #setM_TargetInheritedACLRights(ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.ResourceMySQL, java.util.List, java.util.Map, java.util.Map, boolean) 
    * @param _acl simpliest ACL to process
    * @param _order parent order of _acl for the Resource. Never below '0'.
    * @param _l_rights effective rights to put as value in the two maps
    * @param _upd_m_globalT_ACLRights global rights map of _res to update.
    * May be null and no updating then. No key nor value removing on this argument.
    * @param _upd_m_primaryT_ACLRights primary ACL map of _res to update.
    * May be empty, but never null. No key nor value removing on this argument.
    */
  public static void addSimplestKeysInM_TargetInheritedACLRights(
         AclEntry _acl, String _order, StringRight[]  _l_rights,
         Map<ImmutableName, StringRight[]> _upd_m_globalT_ACLRights,
         Map<String, StringRight[]> _upd_m_primaryT_ACLRights) {

      Set<StringRight>  l_newRights = new HashSet(); // = 16
      boolean           error = false;
      Name              keyName = null;
      String            aclResourceType = _acl.getResType();
      ImmutableGroupID  grp = null;
      ImmutableGroupIDMember[] l_users = null;
      String            primaryKey = null;

                                      // for testing and tracking
                                      //System.out.println("Call to BaseUtilityMySQLImpl."+
                                        //      "addSimplestKeysInM_TargetInheritedACLRights() for "+
                                          //    _acl));

        // (user name @ host name last component) ::  parent order
      if (_acl.isConditionalAclEntry()) {
            keyName = ((Name)_acl.getEpName().clone()).context().
                  extended(_acl.getEpName().lastComponent()+
                    "@"+
                    ((GroupID)_acl.getL_ConditionSourceGroups().get(0)).
                        getName().lastComponent()).extended(_order);
            primaryKey = keyName.toString();

            // updates _upd_m_globalT_ACLRights only from a root ACL
            if (_upd_m_globalT_ACLRights != null &&
              aclResourceType.equals("MySQL server <byAcsCreator>")) {
                
                if (_upd_m_globalT_ACLRights.get(keyName) == null) {

                    _upd_m_globalT_ACLRights.put(keyName, _l_rights);
                } // if
                else {
                    l_newRights.clear();
                    l_newRights.addAll(
                            Arrays.asList(_upd_m_globalT_ACLRights.get(keyName)));
                    l_newRights.addAll(Arrays.asList(_l_rights));

                    _upd_m_globalT_ACLRights.put(
                            keyName,
                            l_newRights.toArray(new StringRight[0]));
                } // else
            } // if (_acl.getResType().equals("MySQL server <byAcsCreator>")

            // updates _upd_m_primaryT_ACLRights
            if (_upd_m_primaryT_ACLRights.get(primaryKey) == null) {

                _upd_m_primaryT_ACLRights.put(
                      primaryKey,
                      _l_rights);

            } // if
            else {
                l_newRights.clear();
                l_newRights.addAll(
                        Arrays.asList(_upd_m_primaryT_ACLRights.get(primaryKey)));
                l_newRights.addAll(Arrays.asList(_l_rights));

                _upd_m_primaryT_ACLRights.put(
                      primaryKey,
                      l_newRights.toArray(new StringRight[0]));

            } // else

      }
        // (<Proc>-user name @ accounts-for-hosts group name last component without '§') ::  -parent order
        // for every user which is member of the accounts-for-hosts group having an ACL
      else {
            if (_acl.getEorEP().getEPType().indexOf("accounts-for-hosts group") == -1) {
              error = true;
            }

            grp = (ImmutableGroupID) _acl.getEorEP();
            l_users = grp.getEorL_Members();

            for (int i = 0; i < l_users.length; i++) {
                keyName = new NameImpl("<Proc>");
                keyName.extended(l_users[i].getName().context()).
                      extended(l_users[i].getName().lastComponent()+
                        "@"+
                        grp.getName().lastComponent().substring(
                            grp.getName().lastComponent().indexOf("@")+1)).
                        extended(_order);
                primaryKey = keyName.toString();

                // updates _upd_m_globalT_ACLRights only from a root ACL
                if (_upd_m_globalT_ACLRights != null &&
                  aclResourceType.equals("MySQL server <byAcsCreator>")) {

                    if (_upd_m_globalT_ACLRights.get(keyName) == null) {

                        _upd_m_globalT_ACLRights.put(keyName, _l_rights);
                    } // if
                    else {
                        l_newRights.clear();
                        l_newRights.addAll(
                                Arrays.asList(_upd_m_globalT_ACLRights.get(keyName)));
                        l_newRights.addAll(Arrays.asList(_l_rights));

                        _upd_m_globalT_ACLRights.put(
                                keyName,
                                l_newRights.toArray(new StringRight[0]));
                    } // else
                } // if (aclResourceType.equals("MySQL server <byAcsCreator>"))

                // updates _upd_m_primaryT_ACLRights
                if (_upd_m_primaryT_ACLRights.get(primaryKey) == null) {

                    _upd_m_primaryT_ACLRights.put(
                          primaryKey,
                          _l_rights);

                } // if
                else {
                    l_newRights.clear();
                    l_newRights.addAll(
                            Arrays.asList(_upd_m_primaryT_ACLRights.get(primaryKey)));
                    l_newRights.addAll(Arrays.asList(_l_rights));

                    _upd_m_primaryT_ACLRights.put(
                          primaryKey,
                          l_newRights.toArray(new StringRight[0]));

                } // else

            } // for i on grp members

      } // else on 'if (_acl.isConditionalAclEntry())'

      if (error)
          throw new InternalError("BaseUtilityMySQLImpl."+
              "addSimplestKeysInM_TargetInheritedACLRights(), the non-conditional ACL "+
              "is not for an accounts-for-hosts group.");

      l_newRights = null;
      keyName = null;
      aclResourceType = null;
      grp = null;
      l_users = null;
      primaryKey = null;
  } // addSimplestKeysInM_TargetInheritedACLRights()

/**
 * Currently unused.
 * @param _text non-null text
 * @return the number of '_' in the text
 */
 public static int countUnderscoreCharacter(String _text) {
     StringCharacterIterator iter = new StringCharacterIterator(_text);
     int result = 0;

     for(char c = iter.first(); c != CharacterIterator.DONE; c = iter.next()) {
         if (c == '_') result++;
     }

     iter = null;
     return result;
 } // countUnderscoreCharacter()

/**
 * Currently unused.
 * @param _text non-null text
 * @return the number of '%' in the text
 */
 public static int countPercentageCharacter(String _text) {
     StringCharacterIterator iter = new StringCharacterIterator(_text);
     int result = 0;

     for(char c = iter.first(); c != CharacterIterator.DONE; c = iter.next()) {
         if (c == '%') result++;
     }

     iter = null;
     return result;
 } // countPercentageCharacter()

} // BaseUtilityMySQLImpl

