//Title:        Access Road
//Version:      0.7.0
//Copyright:    2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.Map;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Collections;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;

/**
 * This class is responsible for modeling a BasesSetMySQL which is a set of databases
 * having a wildcard in its name, like 'BASE O%', as child of the Directory 
 * '|pattern-name_bases sets|'. It has to be a Directory to be called
 * by the AclEntry when there is a rights changing
 * (through resetInheritedAclEntryRightsOfChildren()),
 * but it is not a ResourceMySQL with primary rights.
 * The documentation of DirectoryMySQLImpl describes in details the ACL management.
 * The direct AclEntries and the inherited AclEntries of the parent '|pattern-name_bases sets|'
 * are not copied into the inherited ACL list of a bases set.
 * The direct AclEntries of this class are copied into the inherited AclEntries list
 * of each database that the databases set owns. For instance, the databases set
 * 'BASE O%' copies its ACL into the lists of the two databases 'BASE One' and 'BASE OFF'. <P>
 * A bases set is a child of the '|pattern-name_bases sets|' directory.
 * A bases set name must start with 'BASES', and contains one
 * wildcard character '%' or '_'.
 * The type of a BasesSetMySQL is always 'pattern-name_bases set'.
 * A database does not know the bases sets where it is registered.
 * This class has no getAddonInnerPropertiesToListen() method
 * because the 'ConnectedDatabases' events do not change the views.
 * The error messages to the Access Road user are the followings: <P>
 * - A bases set must have '|pattern-name_bases sets|' as parent. <P>
 * - A bases set must have the type 'bases set'. <P>
 * - A bases set name must contain one wildcard '%' or '_'. <P>
 * - MySQL allows conditional ACL only to accounts. <P>
 * - MySQL allows a conditional ACL only with a simple host as conditional group.
 * To set 'tom@%' for instance, creates the accounts-for-groups group
 * 'tom§@%' and sets on it a non-conditional ACL. <P>
 * - MySQL allows conditional ACL only with the account first host as condition group. <P>
 * - A MySQL bases set allows non-conditional ACL only to accounts,
 * accounts-for-hosts groups, hosts and hosts groups. <P>
 * It is a javabean with the following bound properties: 'ConnectedDatabases'.
 * This class has no subclass in this package.
 * This class is declared in the ACS vocabulory by the key
 * ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES.
 * @version 0.7.0
 */
public class BasesSetMySQLImpl extends DirectoryImpl implements BasesSetMySQL {
  // private variables
  private static final long   serialVersionUID__ = 0;
  // protected variables
  List<DirectoryMySQL>        l_ConnectedDatabases_ = null;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final BasesSetMySQLImpl EMPTY_INSTANCE = new BasesSetMySQLImpl();


  /**
   * variable for the JDK 2 serialization
   * @serial ("l_ConnectedDatabases_", List.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("l_ConnectedDatabases_", List.class)
  };

 /**
  * The basic constructor
  */
  public BasesSetMySQLImpl() {


                          // for testing purpose
                     //System.out.println("Call to BasesSetMySQLImpl constructor");
  }

   /**
   * This constructor is set by the beamer editor for the ACS resources list,
   * from the 'New' button, or from the gWork.AcsFactoryUbuntu.
   * @param _name of the directory
   * @param _acs is the main ACS of the directory
   * @param _parent of the directory; may be null.
   * @param _resourceType is the type, and may be null if the ACS allows it.
   * @param _owner owns the resource; may be null. Never used.
   * @exception CreateError if ACSImpl.newEorResource() throws it -
   *  the instance must then be deleted !
   */
  public BasesSetMySQLImpl(ImmutableName _name, ACS _acs, Directory _parent,
          UserID _owner, String _resourceType) throws CreateError {

      this(_name, _acs, (Directory)_parent, null, _resourceType, true);

  } // constructor

   /**
   * This constructor is called by the other constructor.
   * A bases set name must start with 'BASES', and contains one
   * wildcard character '%' or '_'. The type must be a 'bases set'
   * and the parent must be the '|pattern-name_bases sets|' node.
   * The direct AclEntries and the inherited AclEntries of the parent are not added,
   * unlike the super constructor.
   * @param _name of the directory
   * @param _acs is the main ACS of the directory
   * @param _parent of the directory; may be null.
   * @param _resourceType is the type, and may be null if the ACS allows it.
   * @param _owner owns the resource. May be null. Never used.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in the ACS.
   * @exception CreateError if ACSImpl.newEorResource() throws it, or error on the name,
   * the type, the parent - the instance must then be deleted !
   */
  public BasesSetMySQLImpl(ImmutableName _name, ACS _acs, Directory _parent,
    UserID _owner, String _resourceType, boolean _toRegisterInManager)
    throws CreateError {

      super(_name, _acs, _parent, null,_resourceType, true);

      if (_parent == null || !_parent.getType().equals("virtual_objects <byAcsCreator>")) {
          try {
              finalizeForBase();
          }
          catch (UpDateError u) {
              throw new CreateError(" From BasesSetMySQLImpl constructor. "+
                  "Error on the parent"+
                  ". Error on finalizing "+u.toString());
          }
          
          throw new CreateError(" A bases set must have '|pattern-name_bases sets|' as parent.");
      } // if

      // a bases set must have the type 'bases set'
      if (_resourceType.indexOf("bases set") == -1 ) {
          try {
              finalizeForBase();
          }
          catch (UpDateError u) {
              throw new CreateError(" From BasesSetMySQLImpl constructor. "+
                  "Error on the bases set type. "+
                  "Error on finalizing "+u.toString());
          }

          throw new CreateError(" A bases set must have the type 'bases set'.");
      } // if

      // a bases set must have 'BASES ' at the beginning of its name
      if (!_name.lastComponent().startsWith("BASES ") ) {
          try {
              finalizeForBase();
          }
          catch (UpDateError u) {
              throw new CreateError("From BasesSetMySQLImpl constructor. "+
                  "Error on the bases set name. "+
                  "Error on finalizing: "+u.toString());
          }

          throw new CreateError(" A bases set name must start with 'BASES '.");
      } // if

      if (_name.lastComponent().indexOf("%") == -1 &&
              _name.lastComponent().indexOf("_") == -1) {
          try {
              finalizeForBase();
          }
          catch (UpDateError u) {
              throw new CreateError("From BasesSetMySQLImpl constructor. "+
                  "Error on the name. "+
                  "Error on finalizing: "+u.toString());
          }

          throw new CreateError(" A bases set name must contain one wildcard '%' or '_'.");
      } // if
      
      // a bases set cannot have a null Resource.l_inheritedT_AclEntries_
      // to avoid NodeRights.withAccessThroughNodesTreeFromEP()
      // to return false, but it does not inherit from the root 'MySQL server'
      // through '|pattern-name_bases sets|'
      l_inheritedT_AclEntries_ = new ArrayList(INITIAL_CAPACITY);

      // registers the matched databases
      Directory             server = (Directory) ((Resource)_parent).getEorParent();
      ImmutableResource[]   l_children = server.getEorL_Children();
      Pattern               pat = null;
      Matcher               matcher = null;
      String                name = "BASE"+_name.lastComponent().substring(5); // remove 'BASES'

      pat = Pattern.compile(
            name.replace("_", ".").replace("%", ".*"),
            Pattern.CASE_INSENSITIVE);

      for (int i = 0; i < l_children.length; i++) {
         matcher = pat.matcher(l_children[i].getName().lastComponent());
         if (l_children[i].getType().indexOf("database") != -1 && matcher.matches())
                 addConnectedDatabase((DirectoryMySQL) l_children[i]);
      } // for i

      server = null;
      l_children = null;
      pat = null;
      matcher = null;
      name = null;
  } // constructor

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon. Returns always null,
   * because these events are listened only to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject has
   * one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()

    /**
    * Adds the inherited AclEntry. 
    * This method is mandatory to be a ResourceMySQL, but there is no operation.
    * @param _acl inherited AclEntry to put in the list at the right order
    */
  public void addInheritedAclEntry(AclEntry _acl) {
      
  }
  
    /**
    * Removes the inherited AclEntries in the children.
    * This method is mandatory to be a ResourceMySQL, but there is no operation.
    * @param _acl inherited AclEntry
    * @exception UpDateError if _acl is null or unknown
    */
  public void removeInheritedAclEntry(AclEntry _acl) throws UpDateError {
      
  }

    /**
    * Resets the inherited AclEntry rights after an user edition.
    * This method is mandatory to be a ResourceMySQL, but there is no operation.
    * @param _acl inherited AclEntry with updated rights
    */
  public void resetInheritedAclEntryRights(AclEntry _acl) {
      
  }

  /**
   * Overridden and forbidden method.
   * @return an InternalException because the property ResourceImpl m_inheritedT_ACLRights_
   * is not used in this class.
   */
  public Map<String, StringRight[]> getM_TargetInheritedACLRights() {
      throw new InternalError("BasesSetMySQLImpl.getM_TargetInheritedACLRights() is forbidden.");
  }
  
    /**
    * Adds the AclEntry as direct AclEntry on this bases set and as
    * inherited AclEntry in the connected databases, and sorts the new direct AclEntry list.
    * This method stops then restores the property change firing on 'AclRights'.
    * The generic order on 'AclRights' and 'EPsThroughAclEntry' firings is inversed.
    * Copies the overridden method and calls it after a removing of all the property change
    * listeners for the event 'AclRights'.
    * Calls addInheritedAclEntry() on the connected databases.
    * The detailed sequence of calls into this method is described at the same DirectoryMySQLImpl
    * method, where the calls to the children are replaced by calls to the connected databases,
    * without primary rights handling.
    * @param _aclAcs ImmutableACS which owns the new AclEntry, and to be recorded
    *  as an ACS controller of this resource if it is not its main ACS.
    * @param _ep eligible party connected to the resource through a new aclEntry.
    * @param _l_right is an array of positive or negative rights of _ep on this resource
    *  (that is, if READ is a right, _ep can read on the resource). May be null.
    * @param _positiveRight is true if the rights are set to authorize (grant),
    *  and false if they are set to forbid (deny).
    * @param _conditionGroup first condition group. May be null.
    * @exception CreateError if _ep or _aclAcs is null, or if _ep is already recorded
    * in this resource with the same ACS, or thrown by the ACS.newEorAclEntry(),
    * or without a condition group not for an account, a host or a host group,
    * or with condition group while the source is not an account.
    * @return the new AclEntry.
    */
  public synchronized AclEntry addAclEntry(ImmutableACS _aclAcs,
    ImmutableEligibleParty _ep, StringRight[] _l_right, Boolean _positiveRight,
    ImmutableGroupID _conditionGroup) throws CreateError {

        PropertyChangeListener[]    l_listeners = null;
        AclEntry[]                  l_oldAclEntry = null;
        AclEntry                    acl = null;
        ArrayList<AclEntry>         l_acl = null;

        if (_conditionGroup != null) {
            if (_ep.getEPType().indexOf("account") == -1)
                throw new CreateError(" MySQL allows conditional ACL "+
                    "only to accounts.");
            if (_conditionGroup.getEPType().indexOf("host") == -1 ||
                _conditionGroup.getEPType().indexOf("hosts") != -1)
                    throw new CreateError(" MySQL allows a conditional ACL"+
                        " only with a simple host as conditional group. "+
                        "To set 'tom@%' for instance, creates the accounts-for-groups group"+
                        "'tom§@%' and sets on it a non-conditional ACL.");
            if (((UserIDMySQL)_ep).getEorFirstHost() != _conditionGroup)
                throw new CreateError(" MySQL allows conditional ACL "+
                    "only with the account first host as condition group.");
        }
        else {

                // The five types that can be used as ACL sources
                if (_ep.getEPType().indexOf("host") == -1 &&
                   // _ep.getEPType().indexOf("host <byAcsCreator>") == -1 &&
                   // _ep.getEPType().indexOf("pattern-name_hosts group") == -1 &&
                   // _ep.getEPType().indexOf("pattern-name_accounts-for-hosts group") == -1 &&
                   _ep.getEPType().indexOf("account") == -1)
                throw new CreateError(" A MySQL bases set allows non-conditional ACL"+
                        "only to accounts, accounts-for-hosts groups, hosts and hosts groups.");

        } // else if (_conditionGroup != null)

        // stop the property change firing on 'AclRights', and not for 'EPsThroughAclEntry'
        l_listeners = getPropertyChangeListeners("AclRights");

        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                removePropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if

        l_oldAclEntry = getL_aclEntry();

        // firing on 'EPsThroughAclEntry'
        acl = super.addAclEntry(_aclAcs, _ep, _l_right, _positiveRight, _conditionGroup);
        l_acl = new ArrayList(Arrays.asList(l_aclEntry_));

        // The new AclEntry has been added at the end of the current list, but no use here
        l_acl.remove(acl);
        l_acl = BaseUtilityMySQLImpl.addInSortedAclList(l_acl, acl);
        l_aclEntry_ = l_acl.toArray(new AclEntry[0]);

        // restore the property change firing on 'AclRights'
        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                addPropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if

        // the ACL ACS and the ACL source are already informed
        try {
            firePropertyChange("AclRights", l_oldAclEntry, getL_aclEntry());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BasesSetMySQLImpl.addAclEntry(), firePropertChange() throws");
            e.printStackTrace();
        }

        // add to 'MySQL server' an ACL with the USAGE right, if necessary
        boolean             found = false;
        DirectoryMySQLImpl  server = (DirectoryMySQLImpl) aCS_.getEorM_Resources().get(
                                ((Name)aCS_Name_.clone()).extended("MySQL server"));
        AclEntry[]          l_acl2 = server.getL_aclEntry();

        if (l_acl2.length != 0) {
            for (int i = 0; i < l_acl2.length; i++) {

                if (!l_acl2[i].getResType().equals("MySQL server <byAcsCreator>"))
                   break;

                if (l_acl2[i].getEorEP() != _ep)
                    continue;

                if (l_acl2[i].getL_ConditionSourceGroups().isEmpty() &&
                        _conditionGroup == null) {
                    found = true;
                    break;
                }
                else if (l_acl2[i].getL_ConditionSourceGroups().isEmpty() &&
                        _conditionGroup != null) {
                    continue;
                }
                else if (_conditionGroup == null) {
                    continue;
                }
                else if ( ((GroupID)l_acl2[i].getL_ConditionSourceGroups().get(0)).
                        getName().equals(_conditionGroup.getName())) {
                    found = true;
                    break;
                } // if
            } // for i
        } // if

        if (!found) {
            // calls back addInheritedAclEntry() on all the 'MySQL server' children,
            // but not this instance.
            server.addAclEntry(_aclAcs, _ep,
                    new StringRight[] {aCS_.getOneRight("USAGE")},
                    _positiveRight, _conditionGroup);
        } // if

        // add the new ACL to the matched databases
        if (l_ConnectedDatabases_ != null) {
            for (DirectoryMySQL base : l_ConnectedDatabases_) {
                base.addInheritedAclEntry(acl);
            }
        } // if

        l_listeners = null;
        l_oldAclEntry = null;
        l_acl = null;
        server = null;
        l_acl2 = null;
        return acl;
    } // addAclEntry()

    /**
    * Removes the AclEntry as direct AclEntry on this bases set and as
    * inherited AclEntry in the connected databases.
    * Calls the overridden method at last.
    * Calls setEorL_TargetInheritedAclEntries() on the connected databases.
    * Fires 'InheritedTargetAclEntries' events only to update the display of
    * the primary rights map.
    * The detailed sequence of calls for adding is described at the method addAclEntry(). <P>
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public synchronized void removeAclEntry(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

                                          // for testing and tracking
                                          //System.out.println(
                                            //  "start of BasesSetMySQLImpl.removeAclEntry() in: "+
                                              //name_+" for _ep "+_ep.getNickName()+
                                              //", _positiveRight "+_positiveRight+
                                              //", _conditionGroup "+_conditionGroup.getNickName());


        // remove the new ACL into the matched databases
        if (l_ConnectedDatabases_ != null) {
            String detailledName = null;
            try {
                if (_conditionGroup != null)
                    detailledName = AclEntryImpl.buildUpDetailledName(
                        ((Name)_ep.getName()).subname(
                                aCS_Name_.noOfComponents(), _ep.getName().noOfComponents()),
                        ((Name)name_.clone()).subname(
                                aCS_Name_.noOfComponents(), name_.noOfComponents()),
                        true, _conditionGroup.getNickName());
                else
                    detailledName = AclEntryImpl.buildUpDetailledName(
                        ((Name)_ep.getName()).subname(
                                aCS_Name_.noOfComponents(), _ep.getName().noOfComponents()),
                        ((Name)name_.clone()).subname(
                                aCS_Name_.noOfComponents(), name_.noOfComponents()),
                        true, null);
            }
            catch (CreateError c) {
                throw new UpDateError("DirectoryMySQLImpl.removeAclEntry() fires "+c);
            }
            ImmutableResource[] l_child = getEorL_Children();
            if (l_child != null) {
                AclEntry    acl = null;

                for (int i = 0; i < l_aclEntry_.length; i++) {
                    if (l_aclEntry_[i] != null &&
                            l_aclEntry_[i].getDetailledName().equals(detailledName)) {

                            acl = l_aclEntry_[i];
                            break;
                    } // if (l_aclEntry_[i] != null
                } // for i

                if (acl == null) {
                    throw new UpDateError("BasesSetMySQLImpl.removeAclEntry(), "+
                            "unknown AclEntry from "+
                            _ep.getName().lastComponent()+" to "+name_.lastComponent()+
                            " and positive "+_positiveRight);
                } // if

                for (DirectoryMySQL base : l_ConnectedDatabases_) {
                    base.removeInheritedAclEntry(acl);
                }

                acl = null;
            } // if (l_child != null)
        } // if (l_ConnectedDatabases_ != null)

                                      // for testing and tracking
                                      //System.out.println(
                                        //      "in BasesSetMySQLImpl.removeAclEntry(),"+
                                          //    "call to the super method in "+
                                            //  name_);

        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);

                          // for testing and tracking
                  //System.out.println("at the end of BasesSetMySQLImpl.removeAclEntry(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}

  } // removeAclEntry()

    /**
    * At each rights changing into a direct ACL, resets the inherited AclEntries
    * of the connected databases (which are not indeed children).
    * Does not call the super method.
    * Called by an own AclEntry when there is a change in its effective rights.
    * This generic Directory method is used to not subclass the generic AclEntry class,
    * but its name is not fully good because it does not process the inherited
    * children AclEntries there.
    * Note: the creation of the inherited AclEntry in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited AclEntry
    * is handled through this method.<P>
    * Not described in the BeanInfo since the user does not call it.
    * Does not fire property change for this instance, but fires the 'InheritedTargetAclEntries'
    * change from each database, to reset their inherited and primary rights.
    * Called by AclEntryImpl.resetRightsAsInheritedAclEntry().
    * Calls DirectoryMySQLImpl.setEorL_TargetInheritedAclEntries().
    * @param _acl own AclEntry with updated rights
    */
  public synchronized void resetInheritedAclEntryRightsOfChildren(AclEntry _acl) {

    // reset the inherited rights of each base
    for (DirectoryMySQL base : l_ConnectedDatabases_) {
            ((DirectoryMySQLImpl)base).setEorL_TargetInheritedAclEntries(
                    getEorL_TargetInheritedAclEntries(), true);
    } // for

  } // resetInheritedAclEntryRightsOfChildren()

  /**
   * Get the databases which are defined by the name of this BasesSetMySQL.
   * @return the databases; they have copies of all the AclEntries of this bases set,
   * or empty. Never null.
   */
  public List<DirectoryMySQL> getEorConnectedDatabases() {
      if (l_ConnectedDatabases_ == null)
          return Collections.EMPTY_LIST;
      else
          return Collections.unmodifiableList(l_ConnectedDatabases_);
  }

    /**
    * Removes the direct AclEntry but without event firing. The purpose of this method 
    * is to avoid an unuseful sequence of firings when a lot of AclEntries are removed. 
    * It is called from the root 'MySQL server' method removeAclEntry() when 
    * it deletes all the direct ACL of a pair on the root children when the root ACL 
    * for this pair is deleted.
    * Calls the overridden method and disables before all the firings for 
    * the 'AclRights' and 'EPsThroughAclEntry' events.
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public void removeAclEntryWithoutFiring(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

        if (type_.equals("MySQL server <byAcsCreator>"))
            throw new InternalError("call to BasesSetMySQLImpl.removeAclEntryWithoutFiring()"+
                    "on the root");

                                          // for testing and tracking
                                          //System.out.println(
                                            //  "start of BasesSetMySQLImpl.removeAclEntryWithoutFiring() in: "+
                                              //name_+" for _ep "+_ep.getNickName()+
                                              //", _positiveRight "+_positiveRight+
                                              //", _conditionGroup "+_conditionGroup.getNickName());
      
        PropertyChangeListener[]    l_listeners = null;
        PropertyChangeListener[]    l_listeners2 = null;

        // remove the new ACL into the matched databases
        if (l_ConnectedDatabases_ != null) {
            String detailledName = null;
            try {
                if (_conditionGroup != null)
                    detailledName = AclEntryImpl.buildUpDetailledName(
                        ((Name)_ep.getName()).subname(
                                aCS_Name_.noOfComponents(), _ep.getName().noOfComponents()),
                        ((Name)name_.clone()).subname(
                                aCS_Name_.noOfComponents(), name_.noOfComponents()),
                        true, _conditionGroup.getNickName());
                else
                    detailledName = AclEntryImpl.buildUpDetailledName(
                        ((Name)_ep.getName()).subname(
                                aCS_Name_.noOfComponents(), _ep.getName().noOfComponents()),
                        ((Name)name_.clone()).subname(
                                aCS_Name_.noOfComponents(), name_.noOfComponents()),
                        true, null);
            }
            catch (CreateError c) {
                throw new UpDateError("DirectoryMySQLImpl.removeAclEntry() fires "+c);
            }
            ImmutableResource[] l_child = getEorL_Children();
            if (l_child != null) {
                AclEntry    acl = null;

                for (int i = 0; i < l_aclEntry_.length; i++) {
                    if (l_aclEntry_[i] != null &&
                            l_aclEntry_[i].getDetailledName().equals(detailledName)) {

                            acl = l_aclEntry_[i];
                            break;
                    } // if (l_aclEntry_[i] != null
                } // for i

                if (acl == null) {
                    throw new UpDateError("BasesSetMySQLImpl.removeAclEntry(), "+
                            "unknown AclEntry from "+
                            _ep.getName().lastComponent()+" to "+name_.lastComponent()+
                            " and positive "+_positiveRight);
                } // if

                for (DirectoryMySQL base : l_ConnectedDatabases_) {

                    // stop the property change firing on 'InheritedTargetAclEntries'
                    // fired by setEorL_TargetInheritedAclEntries() from removeInheritedAclEntry()
                    l_listeners = base.getPropertyChangeListeners(
                            "InheritedTargetAclEntries");

                    if (l_listeners != null) {
                        for (int j = 0; j < l_listeners.length; j++) {
                            removePropertyChangeListener("InheritedTargetAclEntries", l_listeners[j]);
                        } // for i
                    } // if

                    base.removeInheritedAclEntry(acl);

                    // restore the property change firing on 'InheritedTargetAclEntries'
                    if (l_listeners != null) {
                        for (int j = 0; j < l_listeners.length; j++) {
                            addPropertyChangeListener("InheritedTargetAclEntries", l_listeners[j]);
                        } // for i
                    } // if

                } // for base

                acl = null;
            } // if (l_child != null)
        } // if (l_ConnectedDatabases_ != null)


        // 'EPsThroughAclEntry' and 'AclRights' are bound properties in ResourceImpl
        // do unregister before the call to super.removeAclEntry()
        l_listeners = getPropertyChangeListeners("AclRights");
        l_listeners2 = getPropertyChangeListeners("EPsThroughAclEntry");

        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                removePropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                removePropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);

        // restore the property change firings
        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                addPropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                addPropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

                          // for testing and tracking
                  //System.out.println("at the end of BasesSetMySQLImpl.removeAclEntryWithoutFiring(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}

        l_listeners = null;
        l_listeners2 = null;
  } // removeAclEntryWithoutFiring()

 /**
  * Clones without registering of the new object in the ACSs.
   * Caution: don't use any updating method on this clone,
   * because that may destroy the base integrity. A finalized instance returns
   * a clone with getEorACS() which returns null,
   * but getName() and getAcsName() still operate.
   * @exception CloneNotSupportedException
   * @return Object; is not registred in the ACSs. The names, the aclEntries,
   * the AG rights, the ACS controller names are cloned.
   * The AclEntries are weakly cloned, so they have no direct link to this resource.
   * The external object references are not cloned
   */
  public synchronized Object clone() {

      BasesSetMySQLImpl clon = (BasesSetMySQLImpl) super.clone();

      if (l_ConnectedDatabases_ != null) {
          clon.l_ConnectedDatabases_ = new ArrayList();
          for (DirectoryMySQL dat : l_ConnectedDatabases_) {
             clon.l_ConnectedDatabases_.add(dat);
          }
      } // if

      return clon;
  } // clone()

   /**
   * This method sets to null the primary inherited rights map.
   * Overrides the super method and calls it at the end.
   * Note: when this method is called by the ResourceImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the main ACS do not known the resource (if
   * it is a clone without setName(), typically)
   */
  public synchronized void finalizeForBase() throws UpDateError {

                    // for testing and tracking
                //System.out.println("In BasesSetMySQLImpl, call of " + name_ +
                    //".finalizeForBase()");

    if (folderMember_ == null) return;

    l_ConnectedDatabases_ = null;

    super.finalizeForBase();
  } // finalizeForBase()

  //
  // ---------------- protected methods ------------------------------
  //

  /**
   * Registers the database in this BasesSetMySQL. Called by the database constructor.
   * 'ConnectedDatabases' is a bound property. 
   * @param _base the database to add. Never null.
   */
  protected void addConnectedDatabase(DirectoryMySQL _base) {
    List<DirectoryMySQL> l_old = null;

    if (_base.getType().indexOf("database") == -1)
        throw new InternalError("BasesSetMySQLImpl.addConnectedDatabase() for an argument "+
                "which is not a database.");

    if (l_ConnectedDatabases_ == null)
        l_ConnectedDatabases_ = new ArrayList(INITIAL_CAPACITY);
    else
        l_old = (List<DirectoryMySQL>)((ArrayList)l_ConnectedDatabases_).clone();

    l_ConnectedDatabases_.add(_base);

    try {
        firePropertyChange("ConnectedDatabases", l_old, ((ArrayList)l_ConnectedDatabases_).clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BasesSetMySQLImpl.addConnectedDatabase(), "+
                "firePropertChange() throws");
        e.printStackTrace();
    }
  } // addConnectedDatabase()

  /**
   * Registers the database in this BasesSetMySQL. Called by the database constructor.
   * 'ConnectedDatabases' is a bound property.
   * @param _base the database to add. Never null.
   */
  protected void removeConnectedDatabase(DirectoryMySQL _base) {
    List<DirectoryMySQL> l_old = null;

    if (l_ConnectedDatabases_ == null)
      throw new InternalError("BasesSetMySQLImpl.removeConnectedDatabase() called for a null list.");
    else if (!l_ConnectedDatabases_.contains(_base))
      throw new InternalError("BasesSetMySQLImpl.removeConnectedDatabase() called for an unknown item.");
    else
        l_old = (List<DirectoryMySQL>)((ArrayList)l_ConnectedDatabases_).clone();

    l_ConnectedDatabases_.remove(_base);

    try {
        firePropertyChange("ConnectedDatabases", l_old, ((ArrayList)l_ConnectedDatabases_).clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BasesSetMySQLImpl.addConnectedDatabase(), firePropertChange() throws:");
        e.printStackTrace();
    }
  } // removeConnectedDatabase()

  //
  // ------------------ overridden protected methods -------------------------------------------
  //

  /**
  * This method provides a copy of the instance variables.
  * Does not copy the children.
  * Called by the copy() methods.
  * Calls super.copyOwnVariables().
  * @param _result to update with this instance variables copies.
  * @param _withoutInternalAcl true if the internal aclEntries
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception if super throws it
  **/
  protected BasesSetMySQLImpl copyOwnVariables(BasesSetMySQLImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      BasesSetMySQLImpl result = (BasesSetMySQLImpl) super.copyOwnVariables(
                _result, _withoutInternalAcl);

      if (l_ConnectedDatabases_ != null) {
          result.l_ConnectedDatabases_ = new ArrayList();
          for (DirectoryMySQL dat : l_ConnectedDatabases_) {
             result.l_ConnectedDatabases_.add(dat);
          }
      } // if

      return result;
  } // copyOwnVariables(,,,)

  /**
   * NO USE.
   * This method sets to null the primary inherited rights map, since it is used
   * after a deserialization.
   * Overrides the super method and calls it at the end.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In ResourceMySQLImpl, call of " + name_ +
                //".finalizeFromDeserialization()");

    if (folderMember_ == null) return;

    l_ConnectedDatabases_ = null;

    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()


} // BasesSetMySQLImpl

