//Title:        Access Road
//Version:      0.7.1
//Copyright:    2011-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.Map;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Arrays;
import java.util.List;
import java.util.Collections;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;

/**
 * This class is responsible to manage the inherited AclEntries of the children,
 * to manage the global rights, and to provide the main documentation for all
 * the ResourceMySQL classes. <P>
 * The children of a DirectoryMySQLImpl are ResourceMySQL, which is implemented
 * by this class, and by ResourceMySQLImpl, ActorMySQLImpl and BasesSetMySQLImpl.
 * This class manages a map of primary inherited AclEntry rights.
 * The final access rights include these primary rights, but only if there is
 * at least one right on the root 'MySQL server' for the user.
 * Otherwise, the user has no right on the resource.
 * On every child of the root 'MySQL server', a new ACL for a pair
 * (user@host or accounts-for-hosts group) produces a new ACL
 * on the root for the same pair if it does not exist. This new root ACL has
 * the right USAGE, and it is not deleted by default when the initial ACL is deleted.
 * For a new non-conditional ACL for an user, a host or a hosts group, there is
 * no adding on the root. The inverse operation is to delete all the direct ACL
 * of a pair on the root children when the root ACL for this pair is deleted.
 * <P> <P>
 * In the ACS MySQL, the effective rights of a pair (user, host) on a resource are selected
 * by a two-steps algorithm, derived from the MySQL specification, that is described hereinafter: <P>
 * At the first step of access control, only the ACL on the root 'MySQL server' are implied
 * to select the first pair (user, host or accounts-for-hosts group) that user matches.
 * An accounts-for-hosts group contains some accounts to which it transfers the rights
 * of all its ACL. Its behavior is like a Privilege having the type LINKED_IF_ALL_SOURCES.
 * This pair (user, host or accounts-for-hosts group) is called by MySQL the current user,
 * and it may be the initial pair (user, host) or not. For instance, all the effective rights
 * of a pair (jerry, localhost) may be provided by the current user ('jerry', '%').
 * The AclEntries on the root 'MySQL server', for this first step of access control, may be: <P>
 * - conditional user/host ACL on the root 'MySQL server' EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on the root 'MySQL server'.  <P>
 * The ACL list is not totally ordered, since the accounts conditional ACL 
 * are not sorted if the accounts are not 'anomymous'.
 * If a right is granted at this first step (at least the 'USAGE' right),
 * some complementary rights may be added, for the SAME current user,
 * at each parent level from the database to the resource direct parent, as inherited ACL,
 * then other complementary rights may be added from the direct ACLs of the resource.
 * At each of these parent levels, and for the resource itself, into the sorted list
 * of inherited or direct ACLs, the FIRST ACL or the FIRST couple of ACL for
 * the current user, if any, defines the effective rights,
 * at this level, for the pair (user, host) on the resource. This is the sens of
 * the 'EXCLUSIVE OR' into each description of the two-steps algorithm. This
 * first-selected-single-output rule is also applied into each category of ACL,
 * like into the sorted list of the non-conditional 'accounts-for-hosts group' ACL
 * for a resource.<P>
 * The database and the bases set are both called 'database-specific target',
 * because the bases set AclEntries are copied into the direct ACL list of every
 * database that matches the bases set name. This copy is mainly provided by the class
 * BasesSetMySQLImpl. For instance, the bases set named 'on%' matches
 * the database named 'one', and all the 'on%' direct ACL are then copied
 * by BasesSetMySQLImpl into the direct ACL list of the database 'one'.
 * As we have seen, if a right is granted at the first step on the root 'MySQL server',
 * some complementary rights may be added from the other levels: <P>
 * first from the database-specific targets: <P>
 * - AND ((conditional user/host ACL on a database-specific target EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on a database-specific target)
 * EXCLUSIVE OR (user non-conditional ACL on a database-specific target AND
 * host non-conditional ACL on the database-specific target)), <P>
 * and secondly, some complementary rights may be also added, for the SAME current user
 * (user, host or group of hosts), from the rights on all the database children
 * which are parents of the resource or which is the resource itself: <P>
 * - AND (conditional user/host ACL on a table EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this table),  <P>
 * - AND (conditional user/host ACL on a column EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this column),  <P>
 * - AND (conditional user/host ACL on a routine EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this routine),  <P>
 * - AND (conditional user/host ACL on a view delegate EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this view delegate),  <P> <P>
 * Routines are the stored functions and the stored procedures.
 * The rights on a resource may use any direct ACL rights on itself and
 * also may use any of the rights on any direct or indirect parent.
 * The rights on a database are the most complex. They use the ACL on the root,
 * on the related bases sets and on this database. <P> <P>
 * Only the first applicable ACL is selected at each parent level
 * (since there is the 'AND' logical operator between all of them),
 * except for the to-be-mixed ACL on the databases.
 * This selection rule explains why the ACL sorting into each ACL list is so important.
 * The only exception is at the database level, when two to-be-mixed ACL
 * are used into the sentence (user non-conditional ACL on a database-specific target AND
 * host non-conditional ACL on the database-specific target) to produce the effective rights
 * from the database level. This is true only if the previous ACL at this level do not match,
 * because the ACL sorting puts first the conditional user/host ACL, then
 * the non-conditional 'accounts-for-hosts group' ACL, and so, the to-be-mixed ACL
 * are always at the end of the sorted list for a database. <P>
 * Among the conditional user/host ACL, only one ACL matches a given pair (user, host).
 * There is no sorting to apply among the hosts into this category of ACL. On the other hand,
 * for several category of ACL into this description, like for instance
 * 'non-conditional 'accounts-for-hosts group' ACL on the root 'MySQL server'', 
 * two specific ACL sortings are required, so that the first matched ACL is get
 * for the effective rights. For the 'non-conditional 'accounts-for-hosts group' ACL
 * and the host to-be-mixed ACL, the first sorting is based on the hosts group names,
 * where for instance '192.168.1.%' is selected before '192.168.%' because it is sorted first.
 * For the bases set ACL at the database level, there is an another sorting
 * for the bases set names, where for instance 'on%' is selected before '%'.
 * All the ACL sorting for a given parent level is the responsibility of the method
 * BaseUtilityMySQLImpl.addInSortedAclList() and the related methods in the same class,
 * like addAclNOTForDatabase(). Detecting what ACL match the current user
 * is the responsability of the class gWork.NodeRightsMySQLImpl.
 * <P> <P>
 * For summarizing the ResourceMySQL AclEntries, the beamer uses the ResourceImpl property
 * m_inheritedT_ACLRights_ to show the primary rights to the user, which are defined as
 * the rights statements of a given parent level. The map keys mimic the 'user@host'
 * MySQL statements the administers use to declare the rights. This map does not contain
 * only the inherited ACL, as its generic name seems to declare it, but also the direct ACL.
 * The interface ResourceMySQL adds a new map for the global rights on the root 'MySQL server'.
 * Comparing the keys, the single difference between the map keys is that the key for
 * the primary rights is a String form of a Name, and the global rights key is
 * the same Name.
 * This is the responsibility of DirectoryMySQLImpl to create and edit these two rights maps.
 * The map of global rights is reserved to NodeRightsMySQLImpl to detect quickly if
 * an user has no access on the root 'MySQL server'. <P>
 * There are three flows of data to manage the updating of AclEntries:
 * (1) at the ACL creation/deletion, (2) when the inner rights of an ACL are editing,
 * (3) only for the conditional ACL, at any change of the membership relation
 * between the source and a condition group. <P>
 * The flow of method calls at the ACL creation is summarized hereinafter.
 * The process is fully under the responsability of the MySQL AcsAddon,
 * with the calling of varied objects and methods: <P>
 * - ResourceMySQL.addAclEntry() is the starting entry method, typically when it is called
 * by the beamer, <P>
 * - BaseUtilityMySQLImpl.setM_PrimaryTargetACLRightsFromDirectACL() to set both
 * the map of the primary AclEntry rights and the map of the global AclEntry rights. <P>
 * - if the ResourceMySQL is a BasesSetMySQLImpl, call to the matched databases
 * to copy the new ACL, <P>
 * - if the ResourceMySQL is a DirectoryMySQLImpl, ResourceMySQL.addInheritedAclEntry()
 * on all the children, so that the updating of each child is independant,
 * calling ResourceMySQL.setEorL_TargetInheritedAclEntries() that calls
 * BaseUtilityMySQLImpl.setM_TargetInheritedACLRights() to set both
 * the map of the primary AclEntry rights and the map of the global AclEntry rights
 * for each child. <P>
 * The detailed sequence of these calls is given at the method addAclEntry(). <P>
 * To manage the rights editing of an AclEntry among the primary AclEntries,
 * the process is partially managed by the MySQL AcsAddon. At each inner rights change,
 * AclEntryImpl calls Directory.resetInheritedAclEntryRightsOfChildren() if its target
 * is a Directory, and this method is overridden in DirectoryMySQL. <P>
 * The calling sequence is the following: <P>
 * - DirectoryMySQL calls the overridden resetInheritedAclEntryRightsOfChildren() method,
 * to call Resource.resetInheritedAclEntryRights() on each child, <P>
 * - if the DirectoryMySQL is a BasesSetMySQLImpl, it calls
 * Resource.resetInheritedAclEntryRights() on each matched database, <P>
 * - it calls BaseUtilityMySQLImpl.setM_PrimaryTargetACLRightsFromDirectACL().
 * <P> <P>
 * The 'InheritedTargetAclEntries' events are bound following specific rules: <P>
 * - if the inherited AclEntries list changes, then the primary AclEntries map may change,
 * but it is not mandatory for a to-be-mixed ACL, <P>
 * - the global AclEntries map change each time the ACL on the root 'MySQL server' change. <P>
 * The use of the property change event 'InheritedTargetAclEntries' does not strictly
 * fulfill with the BeanInfoPattern. In the second case, the event is fired to inform
 * the beamer about the primary rights, while the changing is about a direct ACL, not
 * an inherited one. This is due to the fact that the ACS addon has chosen
 * to not alterate the generic handling of a change of rights on a direct ACL,
 * since the addon does not subclass the AclEntryImpl class.
 * On the other hand, a graphic view that contains a ResourceMySQL
 * listens the direct ACL and the inherited ACL events. Then, the graphic view
 * receives several events about the same basic event about a change among the rights
 * of an ACL. The timer delay in Gui2.GraphiViewBaseListenerImpl allows
 * to avoid some unuseful processings for updating the graphic view. <P>
 * A database is registered into the matched bases sets which is implemented by BasesSetMySQLImpl.
 * This class implements the following rules, and when the user enter is not fulfilled with,
 * it delivers an error message to the Access Road user with the same sentence: <P>
 * - A database must have 'BASE ' at the beginning of its name. <P>
 * - A table must have 'table' at the end of its name. <P>
 * - MySQL allows conditional ACL only to accounts. <P>
 * - MySQL allows a conditional ACL only with a simple host as conditional group.
 * To set 'tom@%' for instance, creates the accounts-for-groups group
 * 'tom§@%', and creates then a new non-conditional ACL on this new group. <P>
 * - MySQL allows conditional ACL only with the account first host as condition group. <P>
 * - A MySQL bases set allows non-conditional ACL only to accounts,
 * accounts-for-hosts groups, hosts and hosts groups. <P>
 * Excepting for EPRViewInBaseImpl, the listeners are outside gBase,
 * they are transient (not serialized in this instance backup).
 * They are called in any order. They may receive a copy of the new value, a null
 * or an empty value. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user. <P>
 * This class has no subclass into this package.
 * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.BaseUtilityMySQLImpl
 * @version 0.7.1
 */
public class DirectoryMySQLImpl extends DirectoryImpl implements DirectoryMySQL {
  // private variables
  private static final long   serialVersionUID__ = 0;
  // protected variables

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final DirectoryMySQLImpl EMPTY_INSTANCE = new DirectoryMySQLImpl();

  /**
   * TreeMap of global AclEntry rights, that are the direct rights statements
   * for the root 'MySQL server'. The global access depends on this map, so each
   * directory has such a map to accelerate the rights search.
   * A ActorMySQL, ResourceMySQL or a BasesSetMySQL has no such property.
   * The types of key are described at the method 
   * BaseUtilityMySQLImpl.setM_TargetInheritedACLRights() which updates this map.
   * The map key into the primary rights map uses the same Name, but in its String form.
   * The types 'virtual_objects' and 'bases set' have no global rights.
   * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.BaseUtilityMySQLImpl#setM_TargetInheritedACLRights(ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.ResourceMySQL, java.util.List, java.util.Map, java.util.Map, boolean)
   */
  protected Map<ImmutableName, StringRight[]>   m_globalT_ACLRights_ = new TreeMap();

  /**
   * variable for the JDK 2 serialization
   * @serial ("m_globalT_ACLRights_", TreeMap.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("m_globalT_ACLRights_", TreeMap.class)
  };

 /**
  * The basic constructor
  */
  public DirectoryMySQLImpl() {


                          // for testing purpose
                     //System.out.println("Call to DirectoryMySQLImpl constructor");
  }

   /**
   * This constructor is set by the beamer editor for the ACS resources list,
   * from the 'New' button, or from the gWork.AcsFactoryUbuntu.
   * @param _name of the directory
   * @param _acs is the main ACS of the directory
   * @param _parent of the directory; may be null.
   * @param _resourceType is the type, and may be null if the ACS allows it.
   * @param _owner owns the resource; may be null. Never used.
   * @exception CreateError if ACSImpl.newEorResource() throws it -
   *  the instance must then be deleted !
   */
  public DirectoryMySQLImpl(ImmutableName _name, ACS _acs, Directory _parent, 
          UserID _owner, String _resourceType) throws CreateError {

      this(_name, _acs, (Directory)_parent, null, _resourceType, true);

  } // constructor

   /**
   * This constructor is called by the other constructor.
   * For a database, it registers this instance into the matched bases sets.
   * A bases set is implemented by the subclass BasesSetMySQLImpl.
   * A table must have 'table' at the end of its name.
   * The AclEntries and the inherited AclEntries from the parent are added.
   * Calls the super constructor.
   * @param _name of the directory
   * @param _acs is the main ACS of the directory
   * @param _parent of the directory; may be null.
   * @param _resourceType is the type, and may be null if the ACS allows it.
   * @param _owner owns the resource; may be null. Never used.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in the ACS.
   * @exception CreateError if ACSImpl.newEorResource() throws it, or error on the name -
   *  the instance must then be deleted !
   */
  public DirectoryMySQLImpl(ImmutableName _name, ACS _acs, Directory _parent,
    UserID _owner, String _resourceType, boolean _toRegisterInManager)
    throws CreateError {

      super(_name, _acs, _parent, null,_resourceType, true);

      if (_parent != null ) {
          // a database must have 'BASE' at the beginning of its name
          if (_resourceType.indexOf("database") != -1 &&
                    !_name.lastComponent().startsWith("BASE ") ) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From DirectoryMySQLImpl constructor. "+
                      "Error on the database name "+
                      ". Error on finalizing "+u.toString());
              }

              throw new CreateError(" A database name must start with 'BASE '.");
          } // if

          // a table must have 'table' at the end of its name
          if (_resourceType.indexOf("table") != -1 &&
                    !_name.lastComponent().endsWith("table") ) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From DirectoryMySQLImpl constructor. "+
                      "Error on the table name "+
                      ". Error on finalizing: "+u.toString());
              }

              throw new CreateError(" A table must have 'table' at the end of its name.");
          } // if

        // the AclEntries and the inherited AclEntries from the parent are added,
          // if it is not the 'pattern-name_bases sets' directory, or other types
          // ridden by selectInheritedAclEntry()
        AclEntry    acl2 = null;
        AclEntry[]  l_acl = _parent.getL_aclEntry();
        ArrayList<AclEntry> l_inh_acl = ((ResourceImpl)_parent).getEorL_TargetInheritedAclEntries();
        boolean     noInheritance = false;

        for (int i = 0; i < l_acl.length; i++) {
            acl2 = selectInheritedAclEntry(l_acl[i], type_);
            if (acl2 == null) {
                noInheritance = true;
                continue;
            }

            addInheritedAclEntry(acl2);
        } // for i

        for (AclEntry ac : l_inh_acl) {
            acl2 = selectInheritedAclEntry(ac, type_);
            if (acl2 == null) {
                noInheritance = true;
                continue;
            }

            addInheritedAclEntry(acl2);
        } // for ac

        if (noInheritance)
            setComment(comment_ + " No ACL inheritance for this type.");
        
        l_acl = null;
        l_inh_acl = null;
        acl2 = null;

      } // if (_parent != null)
      else if (name_.lastComponent().equals("MySQL server")) {
          // 'MySQL server' cannot have a null Resource.l_inheritedT_AclEntries_
          // to avoid NodeRights.withAccessThroughNodesTreeFromEP()
          // to return false while they are direct rights to apply
          l_inheritedT_AclEntries_ = new ArrayList(INITIAL_CAPACITY);
      } // else if (_parent != null)

      // registering into the matched bases sets
      if (_resourceType.indexOf("database") != -1) {
          ImmutableResource[]   l_children = ((Directory)_parent).getEorL_Children();
          Pattern               pat = null;
          Matcher               matcher = null;
          int                   basesIndex = -1;

          for (int i = 0; i < l_children.length; i++) {
              if (l_children[i].getName().lastComponent().equals("|pattern-name_bases sets|")) {
                  basesIndex = i;
                  break;
              }
          } // for

          if (basesIndex != -1)
              l_children = ((Directory)l_children[basesIndex]).getEorL_Children();
          else
              throw new InternalError("DirectoryMySQLImpl constructor on a database does not"+
                      " find the '|pattern-name_bases sets|' directory.");

          for (int i = 0; i < l_children.length; i++) {

              if (!(l_children[i] instanceof BasesSetMySQLImpl))
                    throw new InternalError("DirectoryMySQLImpl constructor on a database does not"+
                      " find a bases set under the '|pattern-name_bases sets|' directory: "+
                      l_children[i].getName());
              // remove 'BASES '
              pat = Pattern.compile(
                    l_children[i].getName().lastComponent().substring(6).
                        replace("_", ".").replace("%", ".*"),
                    Pattern.CASE_INSENSITIVE);
              // remove 'BASE '
              matcher = pat.matcher(_name.lastComponent().substring(5));
              if (matcher.matches())
                    ((BasesSetMySQLImpl)l_children[i]).addConnectedDatabase(this);

          } // for i

          l_children = null;
          pat = null;
          matcher = null;
      } // if (_resourceType.indexOf("database") != -1

  } // constructor

  //
  // To implement the ResourceImpl methods
  //

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon. Returns always null.
   * These events are listened to update the views containing these ACSObjects,
   * and this is not necessary for m_globalT_ACLRights_.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject has
   * one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()

  // ------------------ public methods -----------------------------------------

    /**
    * Updates the inherited AclEntry in the children, the primary and the global rights maps,
    * if it is applicable, and sorts the new direct AclEntry list.
    * This is true for the root 'MySQL server' and for any of its child, with few
    * exceptions since the types 'directory', 'virtual_objects' and 'bases set' have
    * no global rights.
    * On every child of the root 'MySQL server', a new ACL for a pair
    * (user@host or accounts-for-hosts group) produces a new ACL
    * on the root for the same pair if it does not exist. This new root ACL has
    * the right USAGE, and it is not deleted by default when the initial ACL is deleted.
    * For a new non-conditional ACL for an user, a host or a hosts group, there is
    * no adding on the root. The inverse operation is to delete all the ACL of a pair
    * on the root children when the root ACL for this pair is deleted.
    * It is not possible to give a CREATE (or DROP) right on a table
    * if the same grantee cannot CREATE (or DROP) the database itself. <P>
    * This method stops then restores the property change firing on 'AclRights'.
    * The generic order on 'AclRights' and 'EPsThroughAclEntry' firings is inversed.
    * It fires the 'InheritedTargetAclEntries' events only to update the display of
    * the primary rights map.
    * Copies the overridden method and calls it after a removing of all the property change
    * listeners for the event 'AclRights'. Calls addInheritedAclEntry() on the children,
    * and calls BaseUtilityMySQLImpl.setM_PrimaryTargetACLRightsFromDirectACL().
    * The detailed sequence of calls from this method is the following: <P>
    * - removePropertyChangeListener() for 'AclRights', to remove all the listeners, <P>
    * - super.addAclEntry(), which has the calls sequence: <P>
    * ----- AclEntryImpl constructor, <P>
    * ----- ACS.newAclEntry(), <P>
    * ----- EligibleParty.addEorAclEntryToEP(), <P>
    * ----- updating of ResourceImpl.l_aclEntry_, <P>
    * ----- ResourceImpl.addAcsController(), <P>
    * ----- ResourceImpl.firePropertyChange() for 'EPsThroughAclEntry', <P>
    * - BaseUtilityMySQLImpl.addInSortedAclList() with the original ResourceImpl.l_aclEntry_
    * as argument, <P>
    * - updating of ResourceImpl.l_aclEntry_ with the returned sorted list, <P>
    * - addPropertyChangeListener() for 'AclRights', to restore the initial listeners, <P>
    * - BaseUtilityMySQLImpl.setM_TargetInheritedACLRights() for updating
    * the primary and the global rights maps, <P>
    * - for a conditional ACL and if the root node has no ACL with the same pair,
    * addAclEntry() on the root node to add an equivalent ACL having only the USAGE right, <P>
    * - for each child, selectInheritedAclEntry(), <P>
    * - on each child, ResourceMySQL.addInheritedAclEntry(), which has the calls sequence: <P>
    * ----- BaseUtilityMySQLImpl.addInSortedInheritedAclList() to insert the new
    * ACL in the inherited ACL list, <P>
    * ---------- BaseUtilityMySQLImpl.addInSortedAclList() with the original inherited ACL list
    * as argument, <P>
    * ----- setEorL_TargetInheritedAclEntries() with the new inherited ACL list,
    * which itself uses the sequence of calls: <P>
    * ---------- updating of ResourceImpl.l_inheritedT_AclEntries_, <P>
    * ---------- setM_TargetInheritedACLRights(), which calls: <P>
    * --------------- BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(), which includes: <P>
    * -------------------- updating of ResourceImpl.m_inheritedT_ACLRights_, <P>
    * -------------------- updating of ResourceMySQL.m_globalT_ACLRights_, <P>
    * ---------- firePropertyChange() for 'InheritedTargetAclEntries', <P>
    * ----- if the child is a Directory, repeats the calls to selectInheritedAclEntry() and
    * addInheritedAclEntry() on its proper children, <P>
    * - firePropertyChange() for 'AclRights' and 'InheritedTargetAclEntries' from this method. <P>
    * @param _aclAcs ImmutableACS which owns the new AclEntry, and to be recorded
    *  as an ACS controller of this resource if it is not its main ACS.
    * @param _ep eligible party connected to the resource through a new aclEntry.
    * @param _l_right is an array of positive or negative rights of _ep on this resource
    *  (that is, if READ is a right, _ep can read on the resource). May be null.
    * @param _positiveRight is true if the rights are set to authorize (grant),
    *  and false if they are set to forbid (deny).
    * @param _conditionGroup first condition group. May be null.
    * @exception CreateError if _ep or _aclAcs is null, or if _ep is already recorded
    *  in this resource with the same ACS, or thrown by the ACS.newEorAclEntry(),
    * or with a condition group while the source is not an account,
    * or without a condition group, and if this directory is not a database while
    * the source is not an accounts-for-hosts group, or for a database
    * when it is not for an account, a host or a host group (including the accounts-for-hosts groups).
    * @return the new AclEntry.
    */
  public synchronized AclEntry addAclEntry(ImmutableACS _aclAcs,
    ImmutableEligibleParty _ep, StringRight[] _l_right, Boolean _positiveRight,
    ImmutableGroupID _conditionGroup) throws CreateError {

        PropertyChangeListener[]    l_listeners = null;
        AclEntry[]                  l_oldAclEntry = null;
        AclEntry                    acl = null;
        ArrayList<AclEntry>         l_acl = null;
        boolean                     found = false;
        boolean                     create = false;
        boolean                     drop = false;
        boolean                     tocreate = true;
        boolean                     todrop = true;

        if (_conditionGroup != null) {
            if (_ep.getEPType().indexOf("account") == -1)
                throw new CreateError(" MySQL allows conditional ACL "+
                    "only to accounts.");
            if (_conditionGroup.getEPType().indexOf("host") == -1 ||
                _conditionGroup.getEPType().indexOf("hosts") != -1)
                    throw new CreateError(" MySQL allows a "+
                        "conditional ACL only with a simple host as conditional group. "+
                        "To set 'tom@nn%' for instance, creates the accounts-for-groups group"+
                        "'tom§@nn%', and creates then a new non-conditional ACL on this new group.");
            if (((UserIDMySQL)_ep).getEorFirstHost() != _conditionGroup)
                throw new CreateError(" MySQL allows conditional ACL "+
                    "only with the account first host as condition group.");
        }
        else {
            if (type_.indexOf("database") == -1 &&
                _ep.getEPType().indexOf("pattern-name_accounts-for-hosts group") == -1)
                    throw new CreateError(" A MySQL resource being not a database allows "+
                        "non-conditional ACL only to accounts-for-hosts groups.");

            else if (type_.indexOf("database") != -1) {
                // The five types that can be used as ACL sources
                if (_ep.getEPType().indexOf("host") == -1 &&
                   // _ep.getEPType().indexOf("host <byAcsCreator>") == -1 &&
                   // _ep.getEPType().indexOf("pattern-name_hosts group") == -1 &&
                   // _ep.getEPType().indexOf("pattern-name_accounts-for-hosts group") == -1 &&
                   _ep.getEPType().indexOf("account") == -1)
                throw new CreateError(" A MySQL database allows non-conditional ACL"+
                        "only to accounts, accounts-for-hosts groups, hosts and hosts groups.");
            } // else if (type_.indexOf("database") != -1 &&
        } // else if (_conditionGroup != null)

        // It is not possible to give a CREATE or DROP right on a table
        // if the same grantee cannot CREATE or DROP the database itself.
        if (type_.indexOf("table") != -1 && type_.indexOf("table_column") == -1) {
            if (_l_right != null && 
                    (Arrays.asList(_l_right).contains(aCS_.getOneRight("CREATE")) ||
                    Arrays.asList(_l_right).contains(aCS_.getOneRight("DROP")))) {
                            // parent_ is always non null
                        l_acl = new ArrayList(Arrays.asList(parent_.getL_aclEntry()));
                        if ((Arrays.asList(_l_right).contains(aCS_.getOneRight("CREATE"))))
                                create = true;
                        if ((Arrays.asList(_l_right).contains(aCS_.getOneRight("DROP"))))
                                drop = true;

                        for (AclEntry ac : l_acl) {
                            if (ac.getEorEP() != _ep)
                                continue;

                            if (ac.getL_ConditionSourceGroups().isEmpty() &&
                                    _conditionGroup != null) {
                                continue;
                            }
                            else if (!ac.getL_ConditionSourceGroups().isEmpty() &&
                                    _conditionGroup == null) {
                                continue;
                            } // if

                            // it is the same grantee
                            if (create &&
                                    !Arrays.asList(ac.getL_Rights()).contains(
                                        aCS_.getOneRight("CREATE"))) {
                                tocreate = false;
                            }
                            if (drop &&
                                    !Arrays.asList(ac.getL_Rights()).contains(
                                        aCS_.getOneRight("DROP"))) {
                                todrop = false;
                            }
                            break;
                        } // for ac

                        l_acl = null;
                        if (!tocreate) {
                            throw new CreateError(" MySQL does not allow to give "+
                                "a CREATE right on a table if the same grantee"+
                                "cannot CREATE the database itself.");
                        } // if
                        if (!todrop) {
                            throw new CreateError(" MySQL does not allow to give "+
                                "a DROP right on a table if the same grantee"+
                                "cannot DROP the database itself.");
                        } // if

            } // if (_l_right != null
        } // if (type_.indexOf("table_column") == -1


        // stop the property change firing on 'AclRights', and not for 'EPsThroughAclEntry'
        l_listeners = getPropertyChangeListeners("AclRights");

        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                removePropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if

        l_oldAclEntry = getL_aclEntry();

        // firing on 'EPsThroughAclEntry'
        acl = super.addAclEntry(_aclAcs, _ep, _l_right, _positiveRight, _conditionGroup);
        l_acl = new ArrayList(Arrays.asList(l_aclEntry_));

        // The new AclEntry has been added at the end of the current list, but no use here
        l_acl.remove(acl);
        l_acl = BaseUtilityMySQLImpl.addInSortedAclList(l_acl, acl);
        l_aclEntry_ = l_acl.toArray(new AclEntry[0]);

        // restore the property change firing on 'AclRights'
        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                addPropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if

        // add the primary rights directly
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                m_globalT_ACLRights_,
                m_inheritedT_ACLRights_,
                true);
        
        found = false;
        DirectoryMySQLImpl server = null;
        // add to 'MySQL server' an ACL with the USAGE right
        if ((_conditionGroup != null || _ep.getType().indexOf("accounts-for-hosts group") != -1) &&
          !type_.equals("MySQL server <byAcsCreator>")) {

            if (l_inheritedT_AclEntries_ != null) {
                // no direct reading of server.l_aclEntry_
                for (AclEntry acl2 : l_inheritedT_AclEntries_) {

                    if (!acl2.getResType().equals("MySQL server <byAcsCreator>"))
                       break;

                    if (acl2.getEorEP() != _ep)
                        continue;

                    if (acl2.getL_ConditionSourceGroups().isEmpty() &&
                            _conditionGroup == null) {
                        found = true;
                        break;
                    }
                    else if (acl2.getL_ConditionSourceGroups().isEmpty() &&
                            _conditionGroup != null) {
                        continue;
                    }
                    else if (acl2.getL_ConditionSourceGroups().get(0) ==
                                _conditionGroup) {
                        found = true;
                        break;
                    } // if
                } // for i
            } // if

            if (!found) {
                server = (DirectoryMySQLImpl) aCS_.getEorM_Resources().get(
                            ((Name)aCS_Name_.clone()).extended("MySQL server"));

                // calls back addInheritedAclEntry() on all the 'MySQL server' children,
                // including this instance.
                server.addAclEntry(_aclAcs, _ep,
                        new StringRight[] {aCS_.getOneRight("USAGE")},
                        _positiveRight, _conditionGroup);
            } // if 
        } // if (_conditionGroup != null

        // add the new ACL as inherited ACL on the children
        AclEntry acl2 = null;
        ImmutableResource[] l_child = getEorL_Children();
        if (l_child == null) return acl;

        for (int i = 0; i < l_child.length; i++) {

            acl2 = selectInheritedAclEntry(acl, l_child[i].getResourceType());
            if (acl2 == null) continue;

            ((ResourceMySQL)l_child[i]).addInheritedAclEntry(acl2);

        } // for i

        // the ACL ACS and the ACL source are already informed.
        // to update the display of the primary rights map.
        try {
            firePropertyChange(
                    "InheritedTargetAclEntries", null, l_inheritedT_AclEntries_.clone());
            firePropertyChange("AclRights", l_oldAclEntry, getL_aclEntry());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From DirectoryMySQLImpl.addAclEntry(), firePropertChange() throws:");
            e.printStackTrace();
        }

        l_listeners = null;
        l_oldAclEntry = null;
        l_acl = null;
        l_child = null;
        acl2 = null;
        server = null;
        return acl;
    } // addAclEntry()

    /**
    * Removes the AclEntry and the inherited AclEntries in the children.
    * On the root 'MySQL server', updates the global rights and
    * deletes all the direct ACL of a pair on the root children when the root ACL
    * for this pair is deleted. Calls the overridden method at last.
    * Calls removeInheritedAclEntry() on the children.
    * Fires 'InheritedTargetAclEntries' events only to update the display
    * of the primary rights map.
    * The detailed sequence of calls for adding is described at the method addAclEntry().
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public synchronized void removeAclEntry(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

                                          // for testing and tracking
                                          //System.out.println(
                                            //  "start of DirectoryMySQLImpl.removeAclEntry() in "+
                                              //name_+" for _ep "+_ep.getNickName()+
                                              //", _positiveRight "+_positiveRight+
                                              //", _conditionGroup "+_conditionGroup.getNickName());


        // process the inherited ACLs on the children
        String detailledName = null;
        try {

            if (_conditionGroup != null)
                detailledName = AclEntryImpl.buildUpDetailledName(
                    ((Name)_ep.getName()).subname(
                        _ep.getAcsName().noOfComponents(), _ep.getName().noOfComponents()),
                    ((Name)name_.clone()).subname(
                        getAcsName().noOfComponents(), name_.noOfComponents()),
                    _positiveRight, _conditionGroup.getName().lastComponent());
            else
                detailledName = AclEntryImpl.buildUpDetailledName(
                    ((Name)_ep.getName()).subname(
                        _ep.getAcsName().noOfComponents(), _ep.getName().noOfComponents()),
                    ((Name)name_.clone()).subname(
                        getAcsName().noOfComponents(), name_.noOfComponents()),
                    _positiveRight, null);

        }
        catch (CreateError c) {
            throw new UpDateError("DirectoryMySQLImpl.removeAclEntry() fires "+c);
        }


                                          // for testing and tracking
                                          //System.out.println(
                                            //  "DirectoryMySQLImpl.removeAclEntry() "+
                                              //"remove the ACL detailled name"+
                                              //detailledName);

        ImmutableResource[] l_child = getEorL_Children();
        if (l_child != null) {
            AclEntry    acl = null;
            AclEntry    acl2 = null;

            for (int i = 0; i < l_aclEntry_.length; i++) {
                if (l_aclEntry_[i] != null && 
                      l_aclEntry_[i].getDetailledName().equals(detailledName)) {

                        acl = l_aclEntry_[i];
                        break;
                } // if (l_aclEntry_[i] != null
            } // for i

            if (acl == null) {
                throw new UpDateError(" DirectoryMySQLImpl.removeAclEntry(), "+
                        "unknown AclEntry from "+
                        _ep.getName().lastComponent()+" to "+name_.lastComponent()+
                        " and positive "+_positiveRight);
            } // if

            for (int i = 0; i < l_child.length; i++) {

                acl2 = selectInheritedAclEntry(acl, l_child[i].getResourceType());
                if (acl2 == null) continue;

                ((ResourceMySQL)l_child[i]).removeInheritedAclEntry(acl2);

            } // for i

            acl = null;
        } // if (l_child != null)


        ImmutableAclEntry[]     l_acl = aCS_.getEorL_AclEntries();
        List<ImmutableAclEntry> l_matchedACL = new ArrayList(INITIAL_CAPACITY);
        // removes the 'MySQL server' ACL pair on each target having this pair for an ACL
        if (type_.equals("MySQL server <byAcsCreator>")) {
            
            for (int i = 0; i < l_acl.length; i++) {
                if (l_acl[i].getResType().equals("MySQL server <byAcsCreator>") ||
                        l_acl[i].getEorEP() != _ep)
                    continue;

                if (l_acl[i].getL_ConditionSourceGroups().isEmpty() &&
                        _conditionGroup == null) {
                    l_matchedACL.add(l_acl[i]);
                }
                else if (l_acl[i].getL_ConditionSourceGroups().get(0) != null &&
                        _conditionGroup != null &&
                        ((GroupID)l_acl[i].getL_ConditionSourceGroups().get(0)).
                            getName().equals(_conditionGroup.getName())) {
                    l_matchedACL.add(l_acl[i]);
                }
            } // for i

            if (!l_matchedACL.isEmpty()) {
                for (ImmutableAclEntry acl2 : l_matchedACL) {
                    ((ResourceMySQL)acl2.getEorResource()).removeAclEntryWithoutFiring(
                            _aclAcs, _ep, _positiveRight, _conditionGroup);
                }
            } // if
        } // if (_conditionGroup != null

                                      // for testing and tracking
                                      //System.out.println(
                                        //      "in DirectoryMySQLImpl.removeAclEntry(),"+
                                          //    "call to the super method in "+
                                            //  name_);

        // firing on 'AclRights' and 'EPsThroughAclEntry'
        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);
        
        // removes the primary rights
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                m_globalT_ACLRights_,
                m_inheritedT_ACLRights_,
                true);
                                      // for testing and tracking
                                      //System.out.println("DirectoryMySQLImpl.removeAclEntry() in: "+
                                        //      name_+" fires the 'InheritedTargetAclEntries' event");

        // to update the display of the primary rights map
        try {
            firePropertyChange(
                    "InheritedTargetAclEntries", null, l_inheritedT_AclEntries_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From DirectoryMySQLImpl.removeAclEntry(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch


                          // for testing and tracking
                  //System.out.println("at the end of DirectoryMySQLImpl.removeAclEntry(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}

        l_child = null;
        l_matchedACL = null;
        l_acl = null;
        detailledName = null;
  } // removeAclEntry()

    /**
    * Adds the inherited AclEntry. Called by DirectoryMySQLImpl.addAclEntry()
    * and the same method, and by the constructor. An inherited AclEntry may have
    * a Resource which is not this instance.
    * The sorting rules in the ACL list are described in BaseUtilityMySQLImpl.
    * Copied from ResourceMySQLImpl.
    * Calls getEorL_TargetInheritedAclEntries(), BaseUtilityMySQLImpl.addInSortedInheritedAclList()
    * and ResourceImpl.setEorL_TargetInheritedAclEntries() that fires 'InheritedTargetAclEntries'.
    * Not described in the BeanInfo since the user does not call it.
    * Calls the children addInheritedAclEntry() method at last.
    * @see #addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _acl inherited AclEntry to put in the list at the right order.
    * Its target is not this instance.
    */
  public synchronized void addInheritedAclEntry(AclEntry _acl) {
    ArrayList<AclEntry> l_acl = null;
    AclEntry acl = null;

    l_acl = getEorL_TargetInheritedAclEntries();

    l_acl = BaseUtilityMySQLImpl.addInSortedInheritedAclList(l_acl, _acl);

    setEorL_TargetInheritedAclEntries(l_acl, true);

    // add the new inherited ACL as inherited ACL on the children
    ImmutableResource[] l_child = getEorL_Children();
    if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            acl = selectInheritedAclEntry(_acl, l_child[i].getResourceType());
            if (acl == null) continue;

            ((ResourceMySQL)l_child[i]).addInheritedAclEntry(acl);

        } // for i
    } // if (l_child != null)

    l_acl = null;
    acl = null;
  } // addInheritedAclEntry()

    /**
    * Removes the inherited AclEntries in the children.
    * Called by DirectoryMySQLImpl.
    * Copied from ResourceMySQLImpl.
    * Calls getEorL_TargetInheritedAclEntries() and
    * setEorL_TargetInheritedAclEntries() that fires 'InheritedTargetAclEntries'.
    * Not described in the BeanInfo since the user does not call it.
    * Calls the children removeInheritedAclEntry() method at last.
    * @param _acl inherited AclEntry. Its target is not this instance.
    * @exception UpDateError if _acl is null or unknown
    */
  public synchronized void removeInheritedAclEntry(AclEntry _acl) throws UpDateError {

              // for testing and tracking
              //System.out.println("Call to DirectoryMySQLImpl.removeInheritedAclEntry() in: "+name_);

      // this is a simple copy of the ResourceMySQKImpl method
      ArrayList<AclEntry>   l_acl = getEorL_TargetInheritedAclEntries();
      AclEntry              acl = null;
      AclEntry              acl2 = null;
      boolean               found = false;

      for (Iterator ite = l_acl.iterator(); ite.hasNext(); ) {
        acl = (AclEntry)ite.next();
        if (acl != null && !acl.isEmpty()) {
          if (acl.equals(_acl)) {
              ite.remove();
              found = true;
              break;
          }
        } // if
      } // for ite

      if (!found) {
          throw new UpDateError(" DirectoryMySQLImpl.removeInheritedAclEntry(), "+
                  "AclEntry is unknown "+_acl);
      }
      else {
                      // for testing and tracking
                      //System.out.println("DirectoryMySQLImpl.removeInheritedAclEntry() in: "+
                        //      name_+" removes the ACL "+acl);
      } // else

      setEorL_TargetInheritedAclEntries(l_acl, true);

      // process the inherited ACLs on the children
      ImmutableResource[] l_child = getEorL_Children();
      if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            acl2 = selectInheritedAclEntry(acl, l_child[i].getResourceType());
            if (acl2 == null) continue;

            ((ResourceMySQL)l_child[i]).removeInheritedAclEntry(acl2);

        } // for i
      } // if (l_child != null)

      l_acl = null;
      acl = null;
      acl2 = null;
  } // removeInheritedAclEntry()

    /**
    * Returns the AclEntry to use as inherited AclEntry in a child.
    * Returns null if the type contains 'index', 'directory', or starts with 'byAcsCreator'.
    * Otherwise, returns the argument.
    * Called by the constructors, resetInheritedAclEntryRights() and
    * Directory.resetInheritedAclEntryRightsOfChildren().
    * Overrides the super method and does not call it.
    * Calls the ResourceMySQL static method having the same name.
    * @param _acl the initial AclEntry which may be given to the child
    * @param _type type of a child, as Resource
    * @return the same AclEntry if the child inherits this initial AclEntry, or null
    */
  public AclEntry selectInheritedAclEntry(AclEntry _acl, String _type) {
    return ResourceMySQLImpl.selectInheritedAclEntry(this, _acl, _type);
  } // selectInheritedAclEntry()

    /**
    * Resets the inherited AclEntry rights and the primary rights after an user edition.
    * Called by the parent when there is a change in the effective rights of
    * an inherited AclEntry. Derived from setM_TargetInheritedACLRights().
    * Called by Directory.resetInheritedAclEntryRightsOfChildren().
    * Overrides the super method to update m_globalT_ACLRights_, and does not call it.
    * Calls setM_TargetInheritedACLRights(), since it is not possible to reset
    * the two inherited rights maps without analyzing all the inherited AclEntries.
    * Calls resetInheritedAclEntryRights() on the children.
    * Fires the 'InheritedTargetAclEntries' event.
    * @param _acl inherited AclEntry with updated rights. Never null.
    */
  public synchronized void resetInheritedAclEntryRights(AclEntry _acl) {

      // _acl is into l_inheritedT_AclEntries_
    setM_TargetInheritedACLRights(l_inheritedT_AclEntries_);

    // as Directory, reset the ACL as inherited ACL on the children
    AclEntry            acl = null;
    ImmutableResource[] l_child = getEorL_Children();

    if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            acl = selectInheritedAclEntry(_acl, l_child[i].getResourceType());
            if (acl == null) continue;

            ((Resource)l_child[i]).resetInheritedAclEntryRights(acl);

        } // for i
    } // if (l_child != null)

    // to update the display of the primary rights map
    try {
        firePropertyChange(
                "InheritedTargetAclEntries", null, l_inheritedT_AclEntries_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From DirectoryMySQLImpl.resetInheritedAclEntryRights(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

                                  // for testing and tracking
                                  //System.out.println("end of DirectoryMySQLImpl."+
                                    //      "resetInheritedAclEntryRights() ");
                                  
    acl = null;
    l_child = null;
  } // resetInheritedAclEntryRights()

    /**
    * At each rights changing into a direct ACL, if the ACL inheritance is applicable,
    * resets the primary rights, the global rights and the inherited AclEntry rights
    * in the children.
    * There are three flows of data to manage the inherited AclEntries: <P>
    * - one at the ACL creation/deletion, and this is not done in the generic gBase package, <P>
    * - one at their inner rights editing, <P>
    * - and the last flow, only for the condition ACL, at the editing of the membership
    * relation between the source and a condition group. <P>
    * This method is called for handling the second and the third cases,
    * to update the children of the Directory which is the target of the AclEntry.
    * This generic Directory method is used to not subclass the generic AclEntry class. <P>
    * Called by an own AclEntry when there is a change in its effective rights,
    * to update the children of the Directory which is its target.
    * Not described in the BeanInfo since the user does not call it.
    * Overrides the super method and calls it at the end.
    * Note: the creation of the inherited AclEntry in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited AclEntry
    * is handled through this method.
    * Fires 'InheritedTargetAclEntries' events only to update the display of
    * the primary rights map.
    * Calls BaseUtilityMySQLImpl.setM_PrimaryTargetACLRights().
    * Called by AclEntryImpl.resetRightsAsInheritedAclEntry(), but not by the
    * subclass BasesSetMySQLImpl.
    * @param _acl own AclEntry with updated rights
    */
  public synchronized void resetInheritedAclEntryRightsOfChildren(AclEntry _acl) {

    // reset the primary and global rights,
      // and this includes a reading of the _acl rights
    BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
            this,
            Arrays.asList(getL_aclEntry()),
            m_globalT_ACLRights_,
            m_inheritedT_ACLRights_,
            true);

    // to update the display of the primary rights map
    try {
        firePropertyChange(
                "InheritedTargetAclEntries", null, Collections.EMPTY_LIST);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From DirectoryMySQLImpl.resetInheritedAclEntryRightsOfChildren(), "+
                "firePropertChange() throws");
        e.printStackTrace();
    } // catch

    super.resetInheritedAclEntryRightsOfChildren(_acl);
  } // resetInheritedAclEntryRightsOfChildren()

  /**
   * Gets the internal map of the primary rights from each parent level of this resource.
    * The primary rights are displayed into the beamer. They help the user
    * to understand the current users, since it covers the root 'MySQL server' rights,
    * and to understand the complex database rights. Because this method overrides
    * the Resource method, its name is 'TargetInheritedACL' rather than 'TargetPrimaryACL'.
    * The map has lists of rights as values, and strongly contrained String keys.
    * The rights may be from a real ACL or from a virtual ACL.
    * The AclEntries list is used to produce some processed rights, as virtual ACL into the map.
    * Only the real ACL are displayed by the beamer, as conditional or non-conditional ACL.
    * A real ACL in the inherited AclEntries list produces rights directly into the map.
    * <P>
    * The primary rights are not an official MySQL concept. It is an help
    * for the Access Road user and it is necessary for the gWork classes.
    * It describes both the inherited AclEntries and the direct AclEntries.
    * The root 'MySQL server' has primary rights only from its direct ACL,
    * and a database has primary rights from the root and its direct ACL.
    * <P>
    * To have some access rights on a MySQL Resource, a pair (user, host) has
    * to match on the root 'MySQL Server' a pair (user, host) or a pair
    * (user, accounts-for-hosts group) which has at least an USAGE right.
    * The FIRST matching pair in the 'MySQL server' list becomes the current user.
    * It is used to get the rights for the initial pair (user, host) at each parent level.
    * The current user is returned by the method NodeRightsMySQLImpl.getConnectionPair().
    * If the current user has only an USAGE right and if, for a given Resource,
    * the curren user has no direct ACL and no inherited ACL, then no right is granted.
    * Otherwise, the rights are the sum of the current user rights
    * from the FIRST matched direct ACL and from the FIRST matched inherited ACL
    * at EACH parent level. For these inherited ACL, the rights from each parent
    * are added to the final result.
    * <P>
    * This section is about the virtual ACL in the MySQL addon.
    * There are two types of virtual ACL: the virtual
    * 'user@accounts-for-hosts group' ACL on any Resource, which derives from
    * one 'accounts-for-hosts group' real ACL, and the mixed ACL which is only
    * on databases. The mixed ACL derives from two to-be-mixed real ACL.
    * The virtual user@accounts-for-hosts group ACL is so derived
    * from a non-conditional accounts-for-hosts group ACL, and the target may have any type.
    * There are processed rights for each account member of the accounts-for-hosts group.
    * If there are 3 members into the accounts-for-hosts group, there are 3 sets of
    * processed rights for three keys in the rights maps. All these rights are
    * simply the copy of the accounts-for-hosts group ACL rights.
    * <P>
    * The database level is the most complex case. It is the only parent level
    * to handle the virtual mixed ACL, the to-be-mixed real ACL and the real bases set ACLs.
    * The to-be-mixed real ACL are the only case where two real ACL are selected
    * at a level (this is indeed always the database level) to build up some rights
    * in a virtual mixed ACL. To-be-mixed and mixed ACL are not official MySQL concepts,
    * but they implement a MySQL algorithm.
    * Two non-conditional to-be-mixed ACL on a database-related target, respectively
    * for an user U and a host H, are combined to define the processed rights
    * for the pair (U, H). The processed rights are the union of the rights
    * from the two real ACL. This define a map key even if the union of rights is empty.
    * This is the only case where a key in the two maps, may have an empty set
    * of effective rights.
    * <P>
    * This section is about the keys into the primary rights map.
    * Each parent level may produce a matched key for a given couple user/host and a target.
    * Into the primary ACL rights map,
    * the map key may have varied forms, where the parent order is always at the end.
    * The parent order tells the distance from the target to the parent which delivers
    * the referenced rights. For instance, the direct parent of a resource has the string '1'.
    * The direct parent of this parent has the string '2' as parent order, and so on.
    * Into the two rights maps, the single difference between the keys is that the key for
    * the primary rights is a String form of a Name, and the global rights key is
    * the same Name. <P>
    * The ten key forms are listed hereinafter: <P>
    * - (user name @ host name last component) ::  parent order,
    * for instance 'one :: mysql :: jerry@localhost :: 2',<P>
    * - <Proc>:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '<Proc>:: one :: mysql :: jerry@192.168.1.% :: 0',<P>
    * '<Proc>' is a generic tag that means 'processed rights', to tell the key and
    * its rights are not get from the direct reading of a real ACL.
    * These two first forms are the only used ones when the parent level is not a database.
    * These forms are also used for a database, but the followings are additional forms in this case:
    * - <Proc><from 'bases set name last component'>:: (user name @ host name last component):: parent order, <P>
    * - <Proc><from 'bases set name last component'>:: (user name @ accounts-for-hosts group name
    * last component) :: parent order, <P>
    * - <Proc><Mixed>:: (user name @ host name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'>:: (user name @ host name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'><from 'bases set name last component'>::
    * (user name @ host name last component) :: parent order, <P>
    * - <Proc><Mixed>:: (user name @ hosts group name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'>:: (user name @ hosts group name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'><from 'bases set name last component'>::
    * (user name @ hosts group name last component) :: parent order.
    * <P>
    * In this catalog of key forms for the database, '<Mixed>' means this is a virtual mixed ACL,
    * and '<from 'bases set name last component'>' means a bases set ACL is the real ACL
    * from which the virtual ACL is derived. The tag '<Mixed>' is proper to MySQL
    * but it may be reused in another ACS Addons. The tag '<from 'xxxx'>' is a generic form
    * another ACS Addon may use.
    * A key contains two times '<from 'bases set name last component'>' when there are
    * both a real bases set ACL for the user, and a real bases set ACL for the host
    * or the hosts group. These key forms are designed to help the Access Road user
    * to find the origin of a key into a rights map.
    * A key has exactly one origin, which is one real ACL in most of the cases,
    * except the virtual mixed ACL where two real to-be-mixed ACL are processed.
    * The rights map keys of a parent level are not sorted
    * following the ACL sorting, or following any other sorting.
    * The value in a rights map is always a non-null StringRight array.
    * This array may be empty only for a mixed virtual ACL.
   * Overrides the Resource method to describe its use by the MySQL addon.
   * @return unmodifiable map of primary rights from the direct and inherited AclEntries.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedACLRights() {
      if (m_inheritedT_ACLRights_ == null) return Collections.EMPTY_MAP;
      else {
          return Collections.unmodifiableMap(m_inheritedT_ACLRights_);
      } // else;
  }

  /**
    * Gets the map of the global rights from the root 'MySQL server'.
    * The map has lists of rights as values, and strongly contrained Name keys.
    * The rights may be from a real ACL or from a virtual ACL.
    * The AclEntries list is used to produce some processed rights, as virtual ACL into the map.
    * Only the real ACL are displayed by the beamer, as conditional or non-conditional ACL.
    * <P>
    * The global rights are necessary for the gWork classes, and the first user is
    * NodeRightsMySQLImpl.getConnectionPair().
    * To have some access rights on a MySQL Resource, a pair (user, host) has
    * to match on the root 'MySQL Server' a pair (user, host) or a pair
    * (user, accounts-for-hosts group) which has at least an USAGE right.
    * The FIRST matching pair in the 'MySQL server' list becomes the current user.
    * It is used to get the rights for the initial pair (user, host) at each parent level.
    * If the current user has only an USAGE right on the root and if, for a given Resource,
    * the current user has no other inherited ACL on the target, then no right is granted. <P>
    * Into the global ACL rights map,
    * the map key may have varied forms, where the parent order is always at the end.
    * The parent order tells the distance from the target to the parent which delivers
    * the referenced rights. For instance, the direct parent of a resource has the string '1'.
    * The direct parent of this parent has the string '2' as parent order, and so on.
    * The two key forms are listed hereinafter, where the parent order is always 0: <P>
    * - (user name @ host name last component) ::  parent order,
    * for instance 'one :: mysql :: jerry@localhost :: 0',<P>
    * - <Proc>:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '<Proc>:: one :: mysql :: jerry@192.168.1.% :: 0',<P>
    * '<Proc>' is a generic tag that means 'processed rights', to tell the key and
    * its rights are not get from the processing of a real ACL.
    * A key has exactly one origin, which is one real ACL.
    * The value in a rights map is always a non-null StringRight array.
   * @return the map of global rights from the root direct AclEntries.
   * This is a direct reference without protection, to be fast.
   * May be empty but never null.
   */
  public Map<ImmutableName, StringRight[]> getM_GlobalTargetACLRights() {
        // includes the root 'MySQL server'
      if (m_globalT_ACLRights_ == null)
          return Collections.EMPTY_MAP;
      else
          return m_globalT_ACLRights_;
  } // getM_PrimaryTargetACLRights()

 /**
  * Deep cloning without registering of the new object in the ACSs.
   * Caution: don't use any updating method on this clone,
   * because that may destroy the base integrity. A finalized instance returns
   * a clone with getEorACS() which returns null,
   * but getName() and getAcsName() operates.
   * @exception CloneNotSupportedException
   * @return Object; is not registred in the ACSs. The names, the aclEntries,
   * the AG rights, the ACS controller names are cloned.
   * The AclEntries are weakly cloned, so they have no direct link to this resource.
   * The external object references are not cloned
   */
  public synchronized Object clone() {

      DirectoryMySQLImpl clon = (DirectoryMySQLImpl) super.clone();

      if (m_globalT_ACLRights_ != null) {
          clon.m_globalT_ACLRights_ = new TreeMap();
          for (Map.Entry entry : m_globalT_ACLRights_.entrySet()) {
             clon.m_globalT_ACLRights_.put(
                     (ImmutableName)((Name)entry.getKey()).clone(),
                     Arrays.copyOf((StringRight[])entry.getValue(),
                        ((StringRight[])entry.getValue()).length));
          }
      } // if

      return clon;
  } // clone()

   /**
   * This method sets to null the primary inherited rights map,
   * and unregisters this instance into the matched bases sets.
   * Overrides the super method and calls it at the end.
   * Note: when this method is called by the ResourceImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the main ACS do not known the resource (if
   * it is a clone without setName(), typically)
   */
  public synchronized void finalizeForBase() throws UpDateError {

                    // for testing and tracking
                //System.out.println("In ResourceMySQLImpl, call of " + name_ + ".finalizeForBase()");

    if (folderMember_ == null) return;

    m_globalT_ACLRights_ = null;

      // unregistering into the matched bases sets
    if (type_.indexOf("database") != -1) {
      ImmutableResource[]   l_children = ((Directory)parent_).getEorL_Children();
      Pattern               pat = null;
      Matcher               matcher = null;

      if (l_children != null) {
          for (int i = 0; i < l_children.length; i++) {

             if (!(l_children[i] instanceof BasesSetMySQLImpl))
                    continue;

             pat = Pattern.compile(
                    l_children[i].getName().lastComponent().replace("_", ".").replace("%", ".*"),
                    Pattern.CASE_INSENSITIVE);
             matcher = pat.matcher(name_.lastComponent());
             if (matcher.matches())
                    ((BasesSetMySQLImpl)l_children[i]).removeConnectedDatabase(this);

          } // for i
      } // if

      l_children = null;
      pat = null;
      matcher = null;
    } // if (type_.indexOf("database") != -1

    super.finalizeForBase();
  } // finalizeForBase()
  
  /**
  * This method provides a copy of the instance
  * with another name, without children copying
  * because the copy order will not be always correct.
  * Caution:  the user and the group will not be copied if
  * the copy ACS is different and without similar user or group.
  * This method is overridden by the subclasses.
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
  *   May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return copy as DirectoryMySQLImpl instance
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null || 
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError(" in DirectoryMySQLImpl.copy(),"+
          " _name or _acs is null"+
          " or _parent ACS is not _acs.");
      }
      
      DirectoryMySQLImpl result = new DirectoryMySQLImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());
      
      result = copyOwnVariables(result, false);
      
      return result;
  } // copy()
  
  /**
  * This method provides a copy of the instance as Copyable.copy(),
  * but without copying the internal aclEntries.
  * It is used for ACSImpl.copy(), because all the EP must be copied
  * first before the copy of all the resources with their aclEntries.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy),
  * but without copy for the internal aclEntry nor the list of ViewInBase
  * that contains this instance.
  * Called from BaseUtilitiesImpl.copyDirectoriesAndActors().
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new DirectoryMySQLImpl instance
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs,
    Directory _parent, Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError(" in DirectoryMySQLImpl.copyWithoutInternalAcl(),"+
          " _name or _acs is null.");
      }

      DirectoryMySQLImpl result = new DirectoryMySQLImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclentries copy
      
      //if (result.m_AclEntriesResources_ == null) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copyWithoutInternalAcl(,,,)

    /**
    * Removes the direct AclEntry and the inherited AclEntries in the children, but
    * without event firing. The purpose of this method is to avoid an unuseful sequence
    * of firings when a lot of AclEntries are removed. It is called from the root
    * 'MySQL server' method removeAclEntry() when it deletes all the direct ACL of a pair
    * on the root children when the root ACL for this pair is deleted.
    * Calls removeInheritedAclEntry() on the children and also calls the overridden method,
    * and disables before all the firings for the 'InheritedTargetAclEntries', 'AclRights'
    * and 'EPsThroughAclEntry' events.
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public void removeAclEntryWithoutFiring(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

        if (type_.equals("MySQL server <byAcsCreator>"))
            throw new InternalError("call to DirectoryMySQLImpl.removeAclEntryWithoutFiring()"+
                    "on the root");

                                          // for testing and tracking
                                          //System.out.println(
                                            //  "start of DirectoryMySQLImpl.removeAclEntryWithoutFiring() in: "+
                                              //name_+" for _ep "+_ep.getNickName()+
                                              //", _positiveRight "+_positiveRight+
                                              //", _conditionGroup "+_conditionGroup.getNickName());
      
        PropertyChangeListener[]    l_listeners = null;
        PropertyChangeListener[]    l_listeners2 = null;

        // process the inherited ACLs on the children
        String detailledName = null;
        try {

            if (_conditionGroup != null)
                detailledName = AclEntryImpl.buildUpDetailledName(
                    ((Name)_ep.getName()).subname(
                        _ep.getAcsName().noOfComponents(), _ep.getName().noOfComponents()),
                    ((Name)name_.clone()).subname(
                        getAcsName().noOfComponents(), name_.noOfComponents()),
                    _positiveRight, _conditionGroup.getName().lastComponent());
            else
                detailledName = AclEntryImpl.buildUpDetailledName(
                    ((Name)_ep.getName()).subname(
                        _ep.getAcsName().noOfComponents(), _ep.getName().noOfComponents()),
                    ((Name)name_.clone()).subname(
                        getAcsName().noOfComponents(), name_.noOfComponents()),
                    _positiveRight, null);

        }
        catch (CreateError c) {
            throw new UpDateError("DirectoryMySQLImpl.removeAclEntry() fires "+c);
        }
        ImmutableResource[] l_child = getEorL_Children();
        if (l_child != null) {
            AclEntry    acl = null;
            AclEntry    acl2 = null;

            for (int i = 0; i < l_aclEntry_.length; i++) {
                if (l_aclEntry_[i] != null &&
                        l_aclEntry_[i].getDetailledName().equals(detailledName)) {

                        acl = l_aclEntry_[i];
                        break;
                } // if (l_aclEntry_[i] != null
            } // for i

            if (acl == null) {
                throw new UpDateError(" DirectoryMySQLImpl.removeAclEntryWithoutFiring(), "+
                        "unknown AclEntry from "+
                        _ep.getName().lastComponent()+" to "+name_.lastComponent()+
                        " and positive "+_positiveRight);
            } // if

            for (int i = 0; i < l_child.length; i++) {

                acl2 = selectInheritedAclEntry(acl, l_child[i].getResourceType());
                if (acl2 == null) continue;

                // stop the property change firing on 'InheritedTargetAclEntries'
                // fired by setEorL_TargetInheritedAclEntries() from removeInheritedAclEntry()
                l_listeners = ((ResourceMySQL)l_child[i]).getPropertyChangeListeners(
                        "InheritedTargetAclEntries");

                if (l_listeners != null) {
                    for (int j = 0; j < l_listeners.length; j++) {
                        removePropertyChangeListener("InheritedTargetAclEntries", l_listeners[j]);
                    } // for i
                } // if

                ((ResourceMySQL)l_child[i]).removeInheritedAclEntry(acl2);

                // restore the property change firing on 'InheritedTargetAclEntries'
                if (l_listeners != null) {
                    for (int j = 0; j < l_listeners.length; j++) {
                        addPropertyChangeListener("InheritedTargetAclEntries", l_listeners[j]);
                    } // for i
                } // if
            } // for i

            acl = null;
        } // if (l_child != null)
        l_listeners = null;

                                      // for testing and tracking
                                      //System.out.println(
                                        //      "in DirectoryMySQLImpl.removeAclEntryWithoutFiring(),"+
                                          //    "call to the super method in "+
                                            //  name_);

        // 'EPsThroughAclEntry' and 'AclRights' are bound properties in ResourceImpl
        // do unregister before the call to super.removeAclEntry()
        l_listeners = getPropertyChangeListeners("AclRights");
        l_listeners2 = getPropertyChangeListeners("EPsThroughAclEntry");

        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                removePropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                removePropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);

        // restore the property change firings
        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                addPropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                addPropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

        // removes the primary rights
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                m_globalT_ACLRights_,
                m_inheritedT_ACLRights_,
                true);

                          // for testing and tracking
                  //System.out.println("at the end of DirectoryMySQLImpl."+
                    //"removeAclEntryWithoutFiring(), the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}

        l_child = null;
        l_listeners = null;
        l_listeners2 = null;
        detailledName = null;
  } // removeAclEntryWithoutFiring()

  //
  // ------------------ overridden protected methods -------------------------------------------
  //
  
  /**
  * This method provides a copy of the instance variables.
  * Does not copy the children.
  * Called by copy(), and by the copyOwnVariables() methods
  * of the subclasses.
  * Calls super.copyOwnVariables().
  * Caution:  the user and the group will not be copied if
  * the copy ACS is different and without similar user or group.
  * @param _result to update with this instance variables copies.
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception if super throws it
  **/
  protected DirectoryMySQLImpl copyOwnVariables(DirectoryMySQLImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      super.copyOwnVariables(_result, _withoutInternalAcl);

      if (m_globalT_ACLRights_ != null) {
          _result.m_globalT_ACLRights_ = new TreeMap();
          for (Map.Entry entry : m_globalT_ACLRights_.entrySet()) {
             _result.m_globalT_ACLRights_.put(
                     (ImmutableName)((Name)entry.getKey()).clone(),
                     Arrays.copyOf((StringRight[])entry.getValue(), 
                        ((StringRight[])entry.getValue()).length));
          }
      } // if


      return _result;
  } // copyOwnVariables(,,,)

  /**
   * Currently unused.
   * This method sets to null the primary inherited rights map, since it is used
   * after a deserialization.
   * Overrides the super method and calls it at the end.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In ResourceMySQLImpl, call of " + name_ +
                //".finalizeFromDeserialization()");

    if (folderMember_ == null) return;

    m_globalT_ACLRights_ = null;

    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  //
  // -------------- protected methods as ResourceMySQLImpl copies -----------------------
  //

   /**
   * Sets all the inherited AclEntries, the global and the primary rights for this target.
   * The AclEntry may be a clone.
   * Called by add/removeInheritedAclEntry(), removeAclEntryWithoutFiring()
   * and by BasesSetMySQLImpl.resetInheritedAclEntryRightsOfChildren() for a database.
   * Calls setM_TargetInheritedACLRights() to set the primary rights map.
   * Fires 'InheritedTargetAclEntries' events.
   * The detailed sequence of calls for adding is described at the method
   * DirectoryMySQLImpl.addAclEntry(). <P>
   * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
   * @param _l_acl list of AclEntries. May be empty, but never null.
   * @param _withFiring true to fire 'InheritedTargetAclEntries'. False only
   * when called by removeAclEntryWithoutFiring().
   */
  protected void setEorL_TargetInheritedAclEntries(ArrayList<AclEntry> _l_acl,
          boolean _withFiring) {
      if (_l_acl== null) throw new InternalError("DirectoryMySQLImpl."+
              "setEorL_TargetInheritedAclEntries() with a null argument.");

      ArrayList l_oldList = null;

      if (l_inheritedT_AclEntries_ == null) {
          l_oldList = new ArrayList(INITIAL_CAPACITY_2);
      } else {
          l_oldList = (ArrayList) l_inheritedT_AclEntries_.clone();
      } // else

      l_inheritedT_AclEntries_ = _l_acl;

      // no firing
      setM_TargetInheritedACLRights(l_inheritedT_AclEntries_);

      try {
        if (_withFiring)
            firePropertyChange(
                "InheritedTargetAclEntries", l_oldList, l_inheritedT_AclEntries_.clone());
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From DirectoryMySQLImpl.setEorL_TargetInheritedAclEntries(), "+
                "firePropertChange() throws");
        e.printStackTrace();
      } // catch

      l_oldList = null;
  } // setEorL_TargetInheritedAclEntries()

  /**
   * Sets the internal maps of the primary and global rights.
   * Does not fire a property change event.
   * Called by setEorL_TargetInheritedAclEntries() only.
   * Does not fire any event while the caller method may fire
   * 'InheritedTargetAclEntries' events.
   * Calls the BaseUtilityMySQLImpl method having the same name.
   * The detailed sequence of calls for adding is described at the method
   * DirectoryMySQLImpl.addAclEntry(). <P>
   * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
   * @param _l_acl new list of inherited AclEntries. May be empty, but never null.
   */
  protected void setM_TargetInheritedACLRights(List<AclEntry> _l_acl) {
      
                                      // for testing and tracking
                                      //System.out.println("Call to DirectoryMySQLImpl."+
                                        //      "setM_TargetInheritedACLRights() in: "+
                                          //    name_);

      if (m_inheritedT_ACLRights_ == null)
          m_inheritedT_ACLRights_ = new TreeMap();

      if (m_globalT_ACLRights_ == null)
          m_globalT_ACLRights_ = new TreeMap();

        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                _l_acl,
                m_globalT_ACLRights_,
                m_inheritedT_ACLRights_,
                false);
  } // setM_TargetInheritedACLRights()

} // DirectoryMySQLImpl
