//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.UpDateError;
import ARoad0.gDMak.ActionNewEligibleParty;

import java.io.ObjectStreamField;

/**
 * This class is responsible for modeling a MySQL GroupID to manage the constraints
 * on the account membership and group names. The allowed basic types are 'host',
 * 'accounts-for-hosts group', 'hosts group' or one of the three standard groups.
 * The standard pattern-name accounts-for-hosts group has all the accounts-for-hosts groups
 * as members. The standard pattern-name hosts group has all the hosts groups as members.
 * The standard for simple hosts group has all the hosts as members.
 * <P>
 * A host in MySQL is simulated by a GroupID having a 'host' type, including
 * the group 'localhost' for the MySQL server host, which has the type 'host byAcsCreator'
 * to forbid its deleting by the Access Road user.
 * A new host, if it is not 'localhost', has to be an open ACS name as last component.
 * In this case, the name last component of a new host must have the form 'xxx:yy:zzz:tttt',
 * and it is converted to the ACS name ' xxx:: yy:: zzz:: tttt::  ' to checkup
 * it is an open ACS. After the group creation, if the open ACS is closed,
 * the MySQL host group remains unchanged.
 * There is a constraint to enforce exactly one simple host per account,
 * at the creation of the account, and this first host cannot be changed. <P>
 * A pattern-name group may be created by the user to contains all the accounts
 * or groups having a given pattern in their names. The convention for these special groups
 * is to use in their names the two '%' and '_' wildcard characters: <P>
 * - '%' matches any number of characters, even zero characters, <P>
 * - '_' matches exactly one character. <P>
 * While MySQL Server allows to have two wildcards, or more, in the same name
 * of a pattern-name group, Access Road does not permit it, because the algorithm
 * of sorting for these names is not specified and very confusing.
 * An IP wildcard value can match only IP addresses,
 * not host names. '1.2.example.com' never match '1.2.%'.
 * This class ensures that accounts or groups are members of the pattern-name groups
 * that match to them, while the proxy accounts are not implied there. <P>
 * There are two types of pattern-name groups, for hosts and for accounts-for-hosts,
 * listed hereinafter: <P>
 * (1) A pattern-name hosts group may be created by the Access Road user to contain
 * all the hosts having a given pattern in their names. <P>
 * (2) A pattern-name accounts-for-hosts group may be created by the Access Road user
 * to contain all the accounts having a given pattern, like 'tom' and 'tom(2)'.
 * The pattern-name accounts-for-hosts group name
 * has then the form 'name§@YYYY', where YYYY contain a wildcard character. <P>
 * To deliver some rights to a given source 'tom' whatever its current host,
 * a pattern-name accounts-for-hosts group 'tom§@%' is required.
 * This form is proper to Access Road, since MySQL would rather uses 'tom@%'.<P>
 * The host to use, if it is not 'localhost', has to be an open ACS having
 * an account with the new userID name as name (modulo '(N)' at the end),
 * and the user 'anonymous' is the single exception to this rule.
 * To summerize, a host in MySQL has only accounts or proxy accounts,
 * which are UserIDMySQLs, as members.
 * An accounts-for-hosts group has only accounts as members.
 * A hosts group has only hosts as members.
 * <P>
 * The rights of 'tom@host_one' and 'tom@host_two' can be defined in two conditional ACLs
 * but not as two accounts. In the database rights, MySQL handles the concept of user
 * independantly of the concept of host, so it is necessary to modelize accounts
 * as UserIDs. The account 'tom' may be set as member of only one host, and it may
 * be necessary to define the account 'tom(2)' on another host. <P>
 * This class implements the following rules, and informs the Access Road user
 * with the same sentence as error message when they are not fulfilled with: <P>
 * - The MySQL group names can be up to 60 characters long, but no more. <P>
 * - A group must not have the characters '%', '_', '§', '@' in its name
 * if it is not a pattern-name group. <P>
 * - This is the name of the MySQL ACS parent, and it is already modelled
 * by the group 'localhost'. <P>
 * - The host name is not a valid name with the form 'xxx:yy:zzz:tttt'. <P>
 * - The host name is not the name of an open ACS. <P>
 * - This is a wrong name, since an accounts-for-hosts group name must
 * have the format 'xxxx§@yyyyy' with a wildcard '%' or '_' in the sequence yyyy. <P>
 * - This is a wrong name , since a hosts group name must have exactly
 * one wildcard '%' or '_'. <P>
 * - A pattern-name group cannot be immutable at its creation. <P>
 * - It is not possible to change by hand the members of a pattern-name group. <P>
 * - An account cannot have a second simple host. <P>
 * - A proxy account cannot have a second simple host. <P>
 * - It is not possible to add a member to a non-standard host if
 * it is not an account nor a proxy. <P>
 * - The name 'ddd' is not associated to an open ACS. <P>
 * - The MySQL account is already associated to the same UserID, in the ACS of this group. <P>
 * - 'bbb' is not the name of an account in the open ACS 'ddd' this group represents. <P>
 * - It is not possible to change directly the members of a pattern-name group. <P>
 * - This account cannot be directly removed from this host, because this group
 * is its first host. <P>
 * This class adds a group for which this GroupIDMember is a member,
 * and it sorts the groups in an ordered array that follows the sorting order
 * for the MySQL AclEntries. The sorting of the source types is defined by
 * BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder(). <P>
 * This class has one subclass in this package, RoleMySQLImpl.
 * It is a javabean with the following bound properties: 'AssociatedGroups',
 * for the overridden method addGroupToGM().
 * @version 0.7.0
 */
public class GroupIDMySQLImpl extends GroupIDImpl implements GroupIDMySQL {
  // private variables
  private static final long   serialVersionUID__ = 0;
  // protected variables

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final GroupIDMySQLImpl EMPTY_INSTANCE = new GroupIDMySQLImpl();

  /**
   * empty variable for the JDK 2 serialization
   */
  public final static ObjectStreamField[] serialPersistentFields = new ObjectStreamField[0];

 /**
  * The basic constructor
  */
  public GroupIDMySQLImpl() {


                          // for testing purpose
                     //System.out.println("Call to GroupIDMySQLImpl constructor");
  }


   /**
    * Constructor used in the beamer by the editor for the ACS EP list, from 'New'.
    * Calls the other constructor.
    * @param _name ImmutableName of the groupID
    * @param _acs ACS of the groupID
    * @param _epType is the eligible party type, and may be null if the ACS allows it.
    * @param _isAdmin true only if the instance has an administrative role in the ACS,
    *    for the technical administration or the users.
    * @exception CreateError if _name ends with '_tree>', or if
    *    ACSImpl.newEorEligibleParty() throws it - the instance must then be deleted !
    */
  public GroupIDMySQLImpl(ImmutableName _name, ACS _acs, String _epType, Boolean _isAdmin)
   throws CreateError {

      this(_name, _acs, _epType, _isAdmin, true);

  } // constructor


   /**
   * Used for the creation of a GroupIDMySQL. Calls the super constructor.
   * This constructor fills up a new pattern-name group with the members
   * which match to it. A new host has to have an open ACS name as last component.
   * In this case, the name last component of a new host must have the form 'xxx:yy:zzz:tttt',
   * and it is converted to the ACS name ' xxx:: yy:: zzz:: tttt::  ' to check
   * if it is an open ACS. For an IPv4/6 address, the standard information system name format
   * is used.
   * The MySQL group names can be up to 60 characters long, but no more.
   * The new group is added to the relevant groups, if necessary.
   * A host in MySQL has only accounts, which are UserIDs, as members, and
   * they are created after the host.
   * An accounts-for-hosts group has also only accounts as members, and they are added
   * by this constructor.
   * A hosts group has only hosts as members, and they are added by this constructor.
   * The standard pattern-name accounts-for-hosts group has all the accounts-for-hosts groups
   * as members. The standard pattern-name hosts group has all the hosts groups as members.
   * The standard for simple hosts group has all the hosts as members. All these three
   * standard groups may be updated by this constructor.
   * Calls to isMemberOfMySQLPatternNameGroup(), addEorMember(),
   * ActionNewEligibleParty.handleNewGroup().
   * This method is used by copy().
   * @param _name ImmutableName of the eligible party to create
   * @param _acs Main ACS of the eligible party to create
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if ACSImpl.newEorEligibleParty() throws it, or if
   * the pattern-name policy is not fulfilled
   * - the instance must then be delete
   */
  public GroupIDMySQLImpl(ImmutableName _name, ACS _acs, String _epType, boolean _isAdmin,
    boolean _toRegisterInManager) throws CreateError {

      super(_name, _acs, _epType, _isAdmin, _toRegisterInManager);
      
      if (type_.indexOf("<byAcsCreator> standard") != -1 ) {
            return;
      } // if

      TreeMap                   m_EPs = new TreeMap(_acs.getEorM_OwnEligibleParties());
      TreeMap                   m_patternGrps = m_EPs;
      String                    grpType = null;
      ImmutableEligibleParty    ep = null;
      ImmutableName             acs_name = null;

      // The MySQL group name can be up to 60 characters long, but no more.
      if (name_.lastComponent().length() > 60) {
          try {
              finalizeForBase();
          }
          catch (UpDateError u) {
              throw new CreateError(" From GroupIDMySQLImpl constructor. "+
                  "Error on finalizing "+u.toString());
          }

          throw new CreateError(" The MySQL group name can be up to 60 characters long, "+
                  "but no more.");
      } // if (name_.lastComponent().length() > 60)

      // a group must not have the characters '%', '_', '§', '@' in its name
      // if it is not a pattern-name group
      if (type_.indexOf("pattern-name") == -1 &&
              (name_.lastComponent().contains("%") ||
              name_.lastComponent().contains("_") ||
              name_.lastComponent().contains("§") ||
              name_.lastComponent().contains("@"))) {
          try {
              finalizeForBase();
          }
          catch (UpDateError u) {
              throw new CreateError(" From GroupIDMySQLImpl constructor. "+
                  "Error on finalizing "+u.toString());
          }

          throw new CreateError(" A group must not have the characters '%', '_', '§', '@'"+
                  "in its name if it is not a pattern-name group.");
      } // if (type_.indexOf("pattern-name") == -1

      try {
          if (name_.lastComponent().indexOf(":") != -1 &&
                  getNameFromHostComponent(name_.lastComponent()).equals(
                    getEorACS().getParentName())) {
              finalizeForBase();
              throw new CreateError(" This is the name of the MySQL ACS parent, "+
                      "and it is already modeled by the group 'localhost'.");
          }
      }
      catch (BaseError b) { // from getNameFromHostComponent()
              throw new CreateError(" This is the name of the MySQL ACS parent; "+
                  "it is already modeled by the group 'localhost'."+
                  "Error on the host name "+b.toString());
      }
      catch (UpDateError u) { // from finalizeForBase()
              throw new CreateError(" This is a wrong name, since it is the ACS parent name;"+
                  " it is already modeled by the group 'localhost'."+
                  "Error on finalizing "+u.toString());
      }

      // the simple host, if not 'localhost', has to be an open ACS
      // the name last component of a new host has to have the form 'xxx:yy:zzz:tttt',
      // and it is converted to the ACS name ' xxx:: yy:: zzz:: tttt::  '.
      if (type_.indexOf("host") != -1 && type_.indexOf("hosts") == -1 &&
              !type_.equals("host <byAcsCreator>")) {
        try {
            acs_name = getNameFromHostComponent(name_.lastComponent());
        }
        catch (BaseError b) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" This is a wrong name, "+
                 "since the host name is not a valid name. "+
                          " Error on finalizing "+u.toString());
              }
              throw new CreateError(" The host name is not a valid name "+
                      "with the form 'xxx:yy:zzz:tttt' "+b.toString());
        }
        if (!BaseManagerImpl.getInstance().isOpenACS(acs_name)) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" This is a wrong name, "+
                 "since the host name is not the name of an open ACS."+
                          " Error on finalizing "+u.toString());
              }
              throw new CreateError(" The host name is not the name of an open ACS. ");
        } // if

      } // if (type_.equals("host"
      // an accounts-for-hosts group name must have the format 'xxxx§@yyyyy'
      // with one wildcard '%' or '_' in the sequence yyyy
      else if (type_.indexOf("pattern-name_accounts-for-hosts group") != -1 ) {

          /* 
             pat = Pattern.compile(
                    grpName.replace("_", ".").replace("%", ".*"),
                    Pattern.CASE_INSENSITIVE);
             matcher = pat.matcher(name);
             result = matcher.matches()
           */

        String nickName = name_.lastComponent();
        if (!Pattern.matches(".*§@.*%.*", nickName) &&
                !Pattern.matches(".*§@.*_.*", nickName)) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                 throw new CreateError(" This is a wrong name, "+
                 "since an accounts-for-hosts group name must have the format 'xxxx§@yyyyy'"+
                 " with a wildcard '%' or '_' in the sequence yyyy."+
                          " Error on finalizing: "+u.toString());
              }
              throw new CreateError(" This is a wrong name "+nickName+
                 ", since an accounts-for-hosts group name must have the format 'xxxx§@yyyyy'"+
                 " with a wildcard '%' or '_' in the sequence yyyy.");
        } // if
        if ((nickName.indexOf("_") != nickName.lastIndexOf("_")) ||
            (nickName.indexOf("%") != nickName.lastIndexOf("%")) ) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                 throw new CreateError(" This is a wrong name, "+
                 "since an accounts-for-hosts group name must have"+
                 " exactly one wildcard '%' or '_'"+
                          " Error on finalizing: "+u.toString());
              }
              throw new CreateError(" This is a wrong name "+nickName+
                 ", since an accounts-for-hosts group name must have"+
                 " exactly one wildcard '%' or '_'.");
        } // if
      } // else
      // a hosts group name must have exactly one wildcard '%' or '_'
      else if (type_.indexOf("pattern-name_hosts group") != -1 ) {

          /* 
             pat = Pattern.compile(
                    grpName.replace("_", ".").replace("%", ".*"),
                    Pattern.CASE_INSENSITIVE);
             matcher = pat.matcher(name);
             result = matcher.matches()
           */

        String nickName = name_.lastComponent();
        if ((nickName.indexOf("_") != nickName.lastIndexOf("_")) ||
            (nickName.indexOf("%") != nickName.lastIndexOf("%")) ) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                 throw new CreateError(" This is a wrong name, "+
                 "since a hosts group name must have"+
                 " exactly one wildcard '%' or '_'"+
                          " Error on finalizing "+u.toString());
              }
              throw new CreateError(" This is a wrong name "+nickName+
                 ", since a hosts group name must have"+
                 " exactly one wildcard '%' or '_'.");
        } // if
        if (nickName.contains("§") || nickName.contains("@") ) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                 throw new CreateError(" This is a wrong name, "+
                 "since a hosts group name must not have"+
                 " a character '§' or '@'"+
                          " Error on finalizing "+u.toString());
              }
              throw new CreateError(" This is a wrong name "+nickName+
                 ", since a hosts group name must not have"+
                 " a character '§' or '@'.");
        } // if
      } // else

      if (type_.indexOf("pattern-name") != -1 ) {

          // a pattern-name group must be mutable at its creation
          if (type_.indexOf("<immutable>") != -1) {
               try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError("From GroupIDMySQLImpl constructor. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError("  A pattern-name group cannot be immutable at its creation. ");
          } // if (type_.equals("pattern-name_hosts group")
      } // if

      // end of the exception checkings
      
      // if this group is a pattern-name hosts group, add to it all its current members,
      // if its name complies with the rules, and add it to the correct standard group
      if (type_.indexOf("pattern-name_hosts group") != -1 ) {

          ((GroupID)m_EPs.get(((Name)_acs.getName()).extended("<G>").
                      extended("|pattern-name_hosts groups|"))).addEorMember(this);

          // necessary for updating the members of this new group in the explorer
          ActionNewEligibleParty.handleNewGroup(this);

          try {
                for (Object me : m_EPs.values()) {
                  if (me instanceof ImmutableGroupID && me != this &&
                          isMemberOfMySQLPatternNameGroup((GroupIDMember)me, this))
                      addEorMemberToPatternNameGroup((GroupIDMember)me);
                } // for me

                setComment("In the name of a MySQL pattern-name group, '%' replaces "+
                        "any group of characters, and '_' replaces one character.");
          }
          catch (BaseError b) {
               try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError("From GroupIDMySQLImpl constructor. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" From the constructor "+b.toString());
          }          
      } // if (type_.equals("pattern-name_hosts group")
      // if this group is a pattern-name accounts-for-hosts group, add to it all its current members,
      // if its name complies with the rules, and add it to the correct standard group
      else if (type_.indexOf("pattern-name_accounts-for-hosts group") != -1 ) {

                                  // for testing purpose
                             //System.out.println("Call to GroupIDMySQLImpl constructor for the "+
                               //      "accounts-for-hosts group "+name_);

          ((GroupID)m_EPs.get(((Name)_acs.getName()).extended("<G>").
                      extended("|pattern-name_accounts-for-hosts|"))).addEorMember(this);

          // necessary for updating the members of this new group in the explorer
          ActionNewEligibleParty.handleNewGroup(this);

          try {
                for (Object me : m_EPs.values()) {
                    // do not add the proxies
                  if (me instanceof ImmutableUserID && 
                          ((ImmutableUserID)me).getType().indexOf("account") != -1 &&
                          isMemberOfMySQLPatternNameGroup((GroupIDMember)me, this)) {

                                  // for testing purpose
                             //System.out.println("GroupIDMySQLImpl constructor for the "+
                               //      "accounts-for-hosts group "+name_+
                                 //    ", calls addEorMemberToPatternNameGroup() for "+
                                   //  ((GroupIDMember)me).getNickName());

                      addEorMemberToPatternNameGroup((GroupIDMember)me);
                  } // if
                } // for me

                setComment("In the name of a pattern-name group, '%' replaces "+
                        "any group of characters, and '_' replaces one character.");
          }
          catch (BaseError b) {
               try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From GroupIDMySQLImpl constructor. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(" From the constructor "+b.toString());
          }
      } // if (type_.equals("pattern-name_accounts group")
      
      // if this group is a host, add it to the correct standard group and hosts groups.
      // the accounts, as members of the host, are created after the creation of the host.
      else if (type_.indexOf("host") != -1 && type_.indexOf("hosts group") == -1 ) {

          ((GroupID)m_EPs.get(((Name)_acs.getName()).extended("<G>").
                      extended("|simple hosts|"))).addEorMember(this);

          // necessary for updating the members of this new group in the explorer
          ActionNewEligibleParty.handleNewGroup(this);

          // add this new host to the matched non-standard pattern-name hosts groups
          for (Iterator ite = m_patternGrps.values().iterator() ; ite.hasNext() ; ) {
            ep = (ImmutableEligibleParty) ite.next();

            if (!(ep instanceof ImmutableGroupID) || ep == this) {
                ite.remove();
                continue;
            }

            grpType = ((ImmutableGroupID)ep).getEPType();
            if (grpType.indexOf("standard") != -1 || grpType.indexOf("hosts group") == -1) {
                ite.remove();
            }
          } // for ite

          try {
            for (Object grp : m_patternGrps.values()) {
              if (isMemberOfMySQLPatternNameGroup(this, (GroupID)grp))
                  ((GroupIDMySQLImpl)grp).addEorMemberToPatternNameGroup(this);
            } // for obj
          }
          catch (BaseError b) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From the constructor. "+
                      "Error on finalizing "+u.toString());
              }

              throw new CreateError(b.toString());
          } // catch

      } // if (type_.equals("host")
      
      m_EPs = null;
      m_patternGrps = null;
      grpType = null;
      ep = null;
      acs_name = null;
  } // constructor

  //
  // To implements the GroupIDImpl methods
  //

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon.Returns always null,
   * because these events are listened only to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject has
   * one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()

  /**
  * Adds an UserIDMySQL as a member of this GroupIDMySQL, which may have the type 'host'
  * or one of the three standard groups.
  * A host in MySQL has only accounts and proxy accounts, which are UserIDMySQLs, as members,
  * and each account has exactly one host.
  * The standard pattern-name accounts-for-hosts group has all the accounts-for-hosts groups
  * as members. The standard pattern-name hosts group has all the hosts groups as members.
  * The standard for simple hosts group has all the hosts as members.
  * An accounts-for-hosts group has only accounts as members, and not proxy accounts,
  * but it uses addEorMemberToPatternNameGroup() rather than this method.
  * A hosts group has only hosts as members, but it uses
  * addEorMemberToPatternNameGroup() rather than this method. <P>
  * An account or proxy account may be member of a host only if the account exists
  * in the ACS the host represents. The account 'anonymous' is the single exception
  * for which no match to a true ACS account is required. Nonetheless, it is possible
  * to have two 'jerry' account in two hosts by putting, as member of one host,
  * a MySQL account like 'jerry(2)' or 'jerry(Z)' to represent the account 'jerry'. <P>
  * The super method resets only the conditional ACLs. For the non-conditional ones,
  * there are several cases: <P>
  * - a host has only accounts as members, and the call
  * to this method at the creation of the host is always
  * before the creation of any account ACL, so there is no issue
  * on the current non-conditional ACL of the account or of the host. <P>
  * - a standard pattern-name group has no ACL, then there is no issue there. <p>
  * UserIDMySQLImpl.setValueOf() and ProxyMySQLImpl.setValueOf() call this method
  * at the creation of an account.
  * Overrides and calls the super method.
  * @param _gm new member of the GroupID. No action if null, empty.
  * @exception CreateError if _gm is a GroupID which has this instance has
  *   direct or indirect member, if _gm is already a member, or if the _gm type
  *   is not authorized by the ACS, or if the type does not allow the operation,
  *   if this group is immutable, or if it is mutable and not the new member.
  */
  public void addEorMember(GroupIDMember _gm) throws CreateError {

    if (type_.indexOf("standard") != -1 ) {
        super.addEorMember(_gm);
        return;
    } // if

    if (type_.indexOf("pattern-name_accounts-for-hosts group") != -1 ||
         type_.indexOf("pattern-name_hosts group") != -1) {
      throw new CreateError(" It is not possible to change by hand the members of "+
              "a pattern-name group.");
    } // if

    // only for adding to a host

    if (_gm.getType().equals("account")) {
        if (((UserIDMySQL)_gm).getEorFirstHost() != null)
            throw new CreateError(" An account cannot have a second simple host.");
    }
    else if (_gm.getType().equals("proxy")) {
        if (((UserIDMySQL)_gm).getEorFirstHost() != null)
            throw new CreateError(" A proxy account cannot have a second simple host.");
    }
    else {
      throw new CreateError(" It is not possible to add a member to a non-standard host if "+
              "it is not an account nor a proxy.");
    }

    // direct processing if it is 'anonymous'
    if (_gm.getName().lastComponent().equals("anonymous")) {
        super.addEorMember(_gm);
        return;
    } // if

    // the MySQL account or proxy account has to exist as UserID in the host as ACS
    boolean                   forLocalHost = false;
    ImmutableName             acs_name = null;
    ImmutableACS              acs = null;
    ImmutableName             account_name = null;
    UserID                    account = null;
    String                    lastCompo = name_.lastComponent();
    String                    gmNickName = null;
    UserIDMySQL               usr = null;
    String                    usrNickName = null;

    if (type_.equals("host <byAcsCreator>"))
        forLocalHost = true;

    if (!forLocalHost) {

        try {
            acs_name = GroupIDMySQLImpl.getNameFromHostComponent(lastCompo);
        }
        catch (BaseError b) {
            // this group has to be correct
            throw new InternalError(" To add an account to a host, "+
                    "the host name has to be a valid name "+b.toString());
        }
        // the host has to be an open ACS
        if (!BaseManagerImpl.getInstance().isOpenACS(acs_name)) {
            throw new CreateError(" The name "+acs_name.lastComponent()+
                    " is not associated to an open ACS. ");
        } // if

        acs = (ImmutableACS) BaseManagerImpl.getInstance().getEorM_OpenACS().get(acs_name);

    } // if (!lastCompo.equals("localhost"))
    else {
        // the ACS parent is the true local host
        acs_name = aCS_.getParentName();
        acs = (ImmutableACS) BaseManagerImpl.getInstance().getEorM_OpenACS().
                get(acs_name);
    } // else

    gmNickName = _gm.getName().lastComponent();
    if (Pattern.matches(".*\\(.\\)", gmNickName)) {

                            // for testing and tracking
                            //System.out.println("in GroupIDMySQLImpl.addEorMember(), for "+
                              //      getName()+", "+gmNickName+" matches '.*\\(.\\)'. ");

        gmNickName = gmNickName.substring(0, gmNickName.length() - 3);
    }

    // no adding if the associated account in this instance already exists as member
    for (Object ob : aCS_.getEorM_EligibleParties().values()) {
        if (!(ob instanceof UserIDMySQL)) continue;
        usr = (UserIDMySQL) ob;

        usrNickName = usr.getName().lastComponent();
        if (Pattern.matches(".*\\(.\\)", usrNickName))
            usrNickName = usrNickName.substring(0, usrNickName.length() - 3);

        if (gmNickName.equals(usrNickName) && usr.getEorFirstHost() == this) {
            gmNickName = null;
            break;
        }
    } // for ob

    if (gmNickName == null)
        throw new CreateError(" The MySQL account "+usr.getName().lastComponent()+
                " is already associated to the same UserID, in the ACS of this group.");

                            // for testing and tracking
                            //System.out.println("in GroupIDMySQLImpl.addEorMember(), for "+
                              //      getName()+", gmNickName: "+gmNickName);

    account_name = ((Name)acs_name.clone()).extended("<U>").extended(gmNickName );

    account = (UserID) acs.getEorM_EligibleParties().get(account_name);

    // the MySQL account has to be an existing account, modulo '(X)' at the end
    if (account == null && !account_name.lastComponent().equals("anonymous")) {
        throw new CreateError(" "+account_name.lastComponent()+
                " is not the name of an account in the open ACS '"+
                acs_name.lastComponent()+"' this group represents.");
    } // if (account == null)

    super.addEorMember(_gm);

    acs_name = null;
    acs = null;
    account_name = null;
    account = null;
    lastCompo = null;
    usr = null;
    usrNickName = null;
  } // addEorMember()

  /**
  * Called by the GUI, and throws an exception if the group is not standard.
  * Never called by the finalizeForBase() method of the members of a host,
  * which use removeEorMemberWithoutControl().
  * @param _gm member of the GroupID. 
  * @exception CreateError to explain that this operation is forbidden and why
  */
  public void removeEorMember(GroupIDMember _gm) throws CreateError {

      if (type_.indexOf("standard") != -1 ) {
        super.removeEorMember(_gm);
        return;
      } // if

      if (type_.indexOf("pattern-name_accounts-for-hosts group") != -1 ||
             type_.indexOf("pattern-name_hosts group") != -1 ) {
          throw new CreateError(" It is not possible to change directly the members of "+
                  "a pattern-name group.");
      } // if (type_.indexOf("pattern-name_accounts group") != -1

      if (type_.indexOf("hosts group") == -1 && type_.indexOf("host") != -1) {
          throw new CreateError(" This account cannot be directly removed "+
                      "from this host, because this group is its first host. ");
      } // if (type_

      //super.removeEorMember(_gm);

  } // removeEorMember()
  
   /**
   * Adds a group for which this GroupIDMember is a member,
   * and sorts the groups in an ordered array that follows the sorting order
   * for the MySQL AclEntries, from BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder().
   * Calls BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder().
   * Overrides the GroupIDMemberImpl method and does not call it.
   * Fires changes on 'AssociatedGroups' like the overridden method. Used only by GroupIDImpl.
   * @param _grp is the groupID to add. No action if it is null, or with null name.
   */
  public synchronized void addGroupToGM(GroupID _grp) {
    // variables
    boolean recorded = false;
    GroupIDImpl[] newArray;
    GroupIDImpl[] l_oldGroups;
    String typeOne = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(_grp.getEPType());

    if (_grp == null) return;
    // test if _grp is not known
    for (int i = 0 ; i < l_Groups_.length ; i++) {
        if (_grp == l_Groups_[i] ) {
          recorded = true; //_grp is known
          break;
        }
    } // for i
    if (recorded) return;

    // saves the old Group list, since l_Groups_ is a bound property
    l_oldGroups = new GroupIDImpl[l_Groups_.length];
    System.arraycopy(l_Groups_, 0, l_oldGroups, 0, l_Groups_.length);

                            // for testing and tracking
                            //System.out.println("in GroupIDMySQLImpl.addGroupToGM(), for "+getName()+
                              //      ", l_Groups_ before the adding:");
                              //for (int i = 0 ; i < l_Groups_.length ; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("-- at "+i+": "+l_Groups_[i].getName());
                                  //else
                                    //  System.out.println("-- at "+i+": NULL");
                              //} // for

        // adds at the right order
    newArray = new GroupIDImpl[l_Groups_.length + 1];
    if (l_Groups_.length != 0) {
        for (int i = 0 ; i < l_Groups_.length ; i++) {
            if (!recorded &&
                    BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder(
                        typeOne,
                        UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(l_Groups_[i].getEPType()),
                        _grp.getName().lastComponent(),
                        l_Groups_[i].getName().lastComponent())) {
                newArray[i] = l_Groups_[i];
                continue;
            } // if

            if (!recorded) {
                newArray[i] = (GroupIDImpl)_grp;
                recorded = true;
            } // if

            newArray[i + 1] = l_Groups_[i];
        } // for

        if (!recorded)
            newArray[l_Groups_.length] = (GroupIDImpl)_grp;
    }
    else {
        newArray[0] = (GroupIDImpl)_grp;
    } // else
    l_Groups_ = newArray;

                            // for testing and tracking
                            //System.out.println("in GroupIDMySQLImpl.addGroupToGM(), for "+getName()+
                              //      ", l_Groups_ after the adding:");
                              //for (int i = 0 ; i < l_Groups_.length ; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("-- at "+i+": "+l_Groups_[i].getName());
                                  //else
                                    //  System.out.println("-- at "+i+": NULL");
                              //} // for

    try {
        firePropertyChange("AssociatedGroups", l_oldGroups, getEorL_GroupID());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From GroupIDMySQLImpl.addGroupToGM(), firePropertChange() throws:");
        e.printStackTrace();
    }

    newArray = null;
    l_oldGroups = null;
  } // addGroupToGM()

  /**
  * Removes the groupID member of a host, without any control like in removeEorMember().
  * Called by UserIDMySQLImpl.finalizeForBase().
  * Simply calls the super method removeEorMember().
  * @param _gm member of the GroupID. No action if unknown, null, empty.
  * @exception CreateError if the type does not allow the operation
  */
  public void removeEorMemberWithoutControl(GroupIDMember _gm) throws CreateError {
      super.removeEorMember(_gm);
  } // removeEorMemberWithoutControl()
  
   /**
   * This method, if this instance is a host, finalizes the accounts that
   * are members having this group as first host.
   * This method removes this instance in its groups directly.
   * The superclass cannot be used, since it is done through a call to
   * GroupIDMySQLImpl.removeEorMember(), which may fire an exception.
   * A pattern-name group does not finalize its members when it is finalized.
   * Overrides the super method and calls it at the end.
   * An unused object has to be deleted to inform the other objects, and this method
   * HAS TO BE USED before the deleting of the GroupID instance.
   * Note: when this method is called by the EligiblePartyImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if GroupIDMySQLImpl.removeEorMemberWithoutControl() throws it
   */
   public synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call of GroupIDMySQLImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (folderMember_ == null ) {
          //super.finalizeForBase();
          return;
    }
    
    if (isSimpleHost() && l_Members_ != null) {
      ImmutableGroupIDMember[] l_oldMembers;

      l_oldMembers = new GroupIDMemberImpl[l_Members_.length];
      System.arraycopy(l_Members_, 0, l_oldMembers, 0, l_Members_.length);

      for (int  i = 0 ; i < l_oldMembers.length ; i++) {
          // calls back removeEorMemberWithoutControl()
          if (l_oldMembers[i] != null && ((UserIDMySQL)l_oldMembers[i]).getEorFirstHost() == this)
                l_oldMembers[i].finalizeForBase();
      } // for i

      l_oldMembers = null;
    } // if (isSimpleHost()

    // set l_Groups_ to an empty array
    if (l_Groups_ != null) {
        List<GroupIDMySQLImpl> l_grps = new ArrayList(Arrays.asList(l_Groups_));
        try {

            for (GroupIDMySQLImpl grp : l_grps) {
                // calls back removeGroupToGM() and updates l_Groups_.
                grp.removeEorMemberWithoutControl(this);
            } // for i

        }
        catch (CreateError c) {
            throw new UpDateError(" GroupIDMySQLImpl.finalizeForBase() "+c.toString());
        }
    } // if (l_Groups_ != null)
    
    super.finalizeForBase();
  } // finalizeForBase()
  
  /**
  * This method provides a copy of the instance
  * with another name.
  * The copy order is 3 times the source order.
  * Does not copy the GroupID members, nor the own resources,
  * the current actors, the own actors.
  * To copy all these links, you have to copy first the GroupID,
  * then to copy the GroupIDMembers, the Resources and the Actors
  * to copy the links. 
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of GroupIDMySQLImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager)
    throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in GroupIDMySQLImpl.copy(),"+
          " _name or _acs is null.");
      }
      
      GroupIDMySQLImpl result = new GroupIDMySQLImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);

             // for testing and tracking
             //System.out.println("in GroupIDMySQLImpl.copy(), the source: "+this+" and the result: "+result );
      
      
      return result;
  } // copy(,,,)

  
  /**
  * This method provides a copy of the instance
  * as copy(), but without copying the EligibleParty internal aclEntries.
  * The copy order is 3 times the source order.
  * Does not copy the GroupID members, nor the own resources,
  * the current actors, the own actors.
  * To copy all these links, you have to copy first the GroupID,
  * then to copy the GroupIDMembers, the Resources and the Actors
  * to copy the links. 
  * Calls copyOwnVariables().
  * Called by ACSImpl.copy().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of GroupIDMySQLImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in GroupIDMySQLImpl.copy(),"+
          " _name or _acs is null.");
      }
      
      GroupIDMySQLImpl result = new GroupIDMySQLImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclEntry copy
      
      // if (result.getEorL_AclConnectedResources().isEmpty()) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copy(,,,)

  /**
   * Returns true if the type is one of the 4 simple host types, and false
   * otherwise, like for a pattern-name_hosts group.
   * @return true if this GroupID models a MySQL simple host
   */
  public boolean isSimpleHost() {
      if (type_.equals("host") ||
          type_.equals("<immutable> host") ||
          type_.equals("host <byAcsCreator>") ||
          type_.equals("<immutable> host <byAcsCreator>"))
                return true;
      else
                return false;
  } // isSimpleHost()

  // ------------------ protected methods -----------------------------------------

  /**
  * Adds a GroupIDMember as a member of a non-standard pattern-name GroupID.
  * The two types of such group is 'pattern-name_accounts-for-hosts group' and
  * 'pattern-name_hosts group', and not the standard groups.
  * A host in MySQL is simulated by a group with a specific type.
  * The super method resets only the conditional ACLs. For the non-conditional ones,
  * there are several cases: <P>
  * - an accounts-for-hosts group has only accounts as members, and the call
  * to this method at the creation of the account or the group is always
  * before the creation of any ACL, so there is no issue
  * on the current non-conditional ACL of the account or of the group, <P>
  * - a hosts group has only hosts as members, and the call
  * to this method at the creation of the host or the hosts group is always
  * before the creation of any ACL, so there is no issue
  * on the current non-conditional ACL of the host or of the hosts group. <P>
  * Called by the constructor.
  * Overrides and calls the super method at the end.
  * @param _gm new member of the GroupID. No action if null, empty.
  * @exception CreateError if _gm is a GroupID which has this instance has
  *   direct or indirect member, if _gm is already a member, or if the _gm type
  *   is not authorized by the ACS, or if the type does not allow the operation,
  *   if this group is immutable, or if it is mutable and not the new member.
  */
  protected void addEorMemberToPatternNameGroup(GroupIDMember _gm) throws CreateError {
    if ((type_.indexOf("pattern-name_accounts-for-hosts group") == -1 &&
         type_.indexOf("pattern-name_hosts group") == -1) ||
         type_.indexOf("standard") != -1) {
      throw new InternalError(" This method is reserved to the non-standard pattern-name groups.");
    } // if

    super.addEorMember(_gm);
  } // addEorMemberToPatternNameGroup()

 /**
 * This static method detects if an account or a group is a member of a pattern-name group
 * that match to it, on the basis of the comparison of their name last components.
 * The name convention for the pattern-name groups
 * is to use in their names one of the two '%' and '_' wildcard characters: <P>
 * - '%' matches any number of characters, even zero characters, <P>
 * - '_' matches exactly one character. <P>
 * For instance, the host 'localhost' matches the pattern-name hosts group 'local%'.
 * A pattern-name group has not 'standard' in its type. There is
 * no operation if one argument is a standard group. <P>
 * There are two types of pattern-name groups, for hosts and for accounts-for-hosts,
 * listed hereinafter: <P>
 * (1) A pattern-name hosts group may be created by the Access Road user to contain
 * all the hosts having a given pattern in their names. <P>
 * (2) A pattern-name accounts-for-hosts group may be created by the Access Road user
 * to contain all the accounts having a given pattern, like 'tom' and 'tom(2)'.
 * The pattern-name accounts-for-hosts group name
 * has then the form 'name§@YYYY', where YYYY contain a wildcard character. <P>
 * An IP value can match only pure IPv4 addresses, not host names. 
 * '1.2.example.com' never match '1.2.%'. This is why the group '1.2.example.com'
 * is not a member of the pattern-name_hosts group '1.2.%'. This method detects
 * an IP v4 address has containing the character '.' and no 'a-z', 'A-Z' characters,
 * after a removing of the IS name.
 * Called by the constructor.
 * @param _memb MySQL account or group name
 * @param _patternNameGroup pattern-name group
 * @return true if _name matches the MySQL pattern for _patternNameGroup, and
 * so _memb has to be member of _patternNameGroup
 * @exception BaseError if _patternNameGroup is not a MySQL pattern-name group
 */
  protected static boolean isMemberOfMySQLPatternNameGroup(
          ImmutableGroupIDMember _memb, ImmutableGroupID _patternNameGroup)
          throws BaseError {
     
                                       // for testing purpose
                                 //System.out.println("GroupIDMySQLImpl.isMemberOfMySQLPatternNameGroup() start with:");
                                 //System.out.println("--- for _memb name: "+_memb.getName());
                                 //System.out.println("--- for _patternNameGroup: "+_patternNameGroup);

     if (_memb.getEPType().indexOf("standard") != -1)
         return false;

     boolean    result = false;
     String     grpType = _patternNameGroup.getEPType();
     String     grpName = _patternNameGroup.getName().lastComponent();

     if (grpType.indexOf("standard") != -1 ||
             (grpType.indexOf("pattern-name_accounts-for-hosts group") == -1 &&
             grpType.indexOf("pattern-name_hosts group") == -1) )
         throw new BaseError(grpName+
                 " has a standard type, or it is not a pattern-name group "+grpType);

     if (grpName.indexOf("%") == -1 && grpName.indexOf("_") == -1)
         throw new BaseError(grpName+" is not a pattern-name group name, "+
                 "which has to contain one of the two characters '_' and '%'.");

     if (grpType.indexOf("pattern-name_accounts-for-hosts group") != -1 &&
             !(_memb instanceof UserIDMySQL) ) {
         grpType = null;
         grpName = null;
         return result;
     }
     else if (grpType.indexOf("pattern-name_hosts group") != -1 &&
             !(_memb instanceof GroupIDMySQL)  ) {
         grpType = null;
         grpName = null;
         return result;
     } // else

     Pattern    pat = null;
     Matcher    matcher = null;
     String     name = _memb.getName().lastComponent();

     if (grpType.indexOf("pattern-name_hosts group") != -1) {

         pat = Pattern.compile(
                grpName.replace("_", ".").replace("%", ".*"),
                Pattern.CASE_INSENSITIVE);

     } // if
     else { // for pattern-name_accounts-for-hosts group named 'xxxx§@YYYY'
         // check 'xxxx'
         if (Pattern.matches(".*\\(.\\)", name))
             name = name.substring(0, name.length() - 3);
         if (!name.equals(grpName.substring(0, grpName.indexOf("§@")))) {
             grpType = null;
             grpName = null;
             name = null;
             return result;
         }
         
         // check 'YYYY'
         name = ((UserIDMySQL)_memb).getEorFirstHost().getName().lastComponent();
         pat = Pattern.compile(
                grpName.substring(grpName.indexOf("§@") + 2).
                    replace("_", ".").replace("%", ".*"),
                Pattern.CASE_INSENSITIVE);
     } // else
     
     matcher = pat.matcher(name);
     result = matcher.matches();

     //boolean    ipPourcentageHostGroupName = Pattern.matches(
       //                                          "\\d+%|"+
         //                                        "\\d+:%|"+
           //                                      "\\d+:\\d+%|"+
             //                                    "\\d+:\\d+:%|"+
               //                                  "\\d+:\\d+:\\d+%|"+
                 //                                "\\d+:\\d+:\\d+:%|"+
                   //                              "\\d+:\\d+:\\d+:\\d+%|",
                     //                            grpName);
    
    //boolean    isIpHostGroupName = Pattern.matches("[^a-zA-Z]+", grpName);
    //boolean    isIpMemberName = Pattern.matches("[^a-zA-Z]+", name);

                                       // for testing purpose
                                 //System.out.println(
                                   //      "GroupIDMySQLImpl.isMemberOfMySQLPatternNameGroup() "+
                                     //    "for grpName: "+grpName+" and member name: "+name);
                                 //System.out.println("--- grpName isIpHostGroupName: "+isIpHostGroupName);
                                 //System.out.println("--- name isIpMemberName: "+isIpMemberName);

     // An IP value can match only pure IPv4 addresses, not host names.
     // '1.2.example.com' never match '1.2.%'
     if (result) {
            // remove the IS name ('ISFactoryUtilityImpl.IPV4_ADDRESSES_SPACE:' or another name)
            grpName = grpName.substring(grpName.indexOf(":")+1);
            name = name.substring(name.indexOf(":")+1);

            if (grpName.contains(".") && grpType.indexOf("pattern-name_hosts group") != -1) {
                // an IP v4 address contains '.' (false into an IP v6 address)
                // 'a' to 'z', 'A' to 'Z' are forbidden if Pattern.matches() is true
                if (Pattern.matches("[^a-zA-Z]+", grpName) &&
                    !Pattern.matches("[^a-zA-Z]+", name)) {

                     result = false;
                } // if
            }
            else if (grpName.contains(".") && grpType.indexOf(
                    "pattern-name_accounts-for-hosts group") != -1) {
                // an IP v4 address contains '.' (false into an IP v6 address)
                // 'a' to 'z', 'A' to 'Z' are forbidden if Pattern.matches() is true
                if (Pattern.matches("[^a-zA-Z]+", grpName.substring(grpName.indexOf("§@") + 2)) &&
                    !Pattern.matches("[^a-zA-Z]+", name)) {

                     result = false;
                } // if
            } // else
     } // if

                                       // for testing purpose
                                 //System.out.println(
                                   //      "GroupIDMySQLImpl.isMemberOfMySQLPatternNameGroup() "+
                                     //    "for grpName: "+grpName+" and member name: "+name);
                                 //System.out.println("--- result: "+result);

     grpType = null;
     grpName = null;
     pat = null;
     matcher = null;
     name = null;
     return result;
  } // isMemberOfMySQLPatternNameGroup()

  /**
   * Gets the ACS name. The argument has to contain ':'.
   * @param _host_name host name last component, where the name components are
   * separated by ':'
   * @exception BaseError if the argument has a wrong format
   * @return ACS name
   */
  protected static ImmutableName getNameFromHostComponent(String _host_name) throws BaseError {
      if (_host_name == null)
          throw new BaseError("GroupIDMySQLImpl.getNameFromHostComponent() with a null argument");
      else if (_host_name.indexOf("::") != -1)
          throw new BaseError(" The host name cannot contain '::'");
      else if (_host_name.indexOf(":") == -1)
          throw new BaseError(" The host name does not contain ':'");


                           // for testing purpose
                     //System.out.println("GroupIDMySQLImpl.getNameFromHostComponent() argument :");
                     //System.out.println("--- '"+_host_name+"'");

      String[] l_compo = _host_name.split(":");

                        //System.out.println("--- l_compo:");
                        //for (int i = 0; i < l_compo.length; i++) {
                          //  System.out.println("------- "+l_compo[i]);
                        //}

      ArrayList     compo = new ArrayList(Arrays.asList(l_compo));
      ImmutableName result = new NameImpl(compo);

      l_compo = null;
      compo = null;

                           // for testing purpose
                     //System.out.println("GroupIDMySQLImpl.getNameFromHostComponent() result :");
                     //System.out.println("--- "+result);

      return result;
  } // getNameFromHostComponent()

  /**
  * This method provides a copy of the instance variables.
  * Copies the own actors and the result as groupIDMember
  * (done by GroupIDMemberImpl.copyOwnVariables()).
  * Does not copy the groupIDMembers (done by the copy() caller).
  * Does not copy the own resources, nor
  * the current actors,
  * but the resource and actor copies do the job.
  * The copy order is 3 times the source order.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError 
  **/
  protected GroupIDMySQLImpl copyOwnVariables(GroupIDMySQLImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      return (GroupIDMySQLImpl)super.copyOwnVariables(_result, _withoutInternalAcl);
  } // copyOwnVariables(,,,)

} // GroupIDMySQLImpl
