//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.Map;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import java.io.ObjectStreamField;

/**
 * This class is responsible for modeling a MySQL proxy user having a proxied account
 * as a non-generic alias reference. The generic soft aliasing feature is very partially used.
 * This class copies most of the UserIDMySQLImpl features,
 * but there are important differences: <P>
 * - a proxy user is never member of an account-for-hosts group. <P>
 * - for a new proxy user 'xxxx', there is no empty non-conditional ACL on 'MySQL Server'
 * for the accounts-for-hosts group 'xxxx§@%'. <P>
 * - the ACS policy allows no ACL to a proxy account, since it is simply
 * the alias of another account. <P>
 * The proxied account is never a proxy account.
 * The host of the proxied account is independant to the host of its proxy account.
 * This class works with UserIDMySQLImpl that implements the proxy account.
 * This AcsAddon does not use the ACS management of the soft aliasing between
 * ProxyMySQLImpl and UserIDMySQLImpl. This is why CompoundRightsFactoryImpl
 * does not use its method selectDirectReferencesInViewForAlias() for this proxy link.
 * The ALIAS link is implemented directly by the AcsAddon, but however,
 * the class UserIDMySQLImpl uses the Basic property 'SoftAlias' to keep
 * the names of the proxy accounts, for a given proxied account.
 * This class implements the following rules, and informs the Access Road user
 * with the same sentence as error message when they are not fulfilled with: <P>
 * - The host of an account cannot be null. <P>
 * - The name of an account cannot contain the characters '%', '_', '§' or '@'. <P>
 * - User names can be up to 16 characters long, but no more. <P>
 * - The host of an account has to be a simple host. <P>
 * - The proxied account cannot be a proxy account. <P>
 * - The host name is not the name of an open ACS. <P>
 * - The name xxx is not the name of an account in the open ACS. <P>
 * This class has no subclass in this package.
 * It is a javabean with the following bound properties: 'CurrentGroupID'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive the new value, without copy.
 * All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * @version 0.7.0
 */
public class ProxyMySQLImpl extends UserIDImpl implements ProxyMySQL {
  // private variables
  private static final long   serialVersionUID__ = 0;
  // protected variables
  /*
   * The current host of this MySQL user. This account is a host member.
   */
  protected GroupIDMySQLImpl  host_ = null;
  /*
   * The proxied account under which this proxy account gets MySQL rights.
   * It is initialized at the creation of the instance, and it is not a bound property.
   */
  protected UserIDMySQLImpl   proxiedAccount_ = null;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ProxyMySQLImpl EMPTY_INSTANCE = new ProxyMySQLImpl();

  /**
   * variable for the JDK 2 serialization
   * @serial ("host_", GroupIDMySQLImpl.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("host_", GroupIDMySQLImpl.class),
      new ObjectStreamField("proxiedAccount_", UserIDMySQLImpl.class)
  };

 /**
  * The basic constructor
  */
  public ProxyMySQLImpl() {


                          // for testing purpose
                     //System.out.println("Call to ProxyMySQLImpl constructor");
  }

   /**
 * Constructor used in the beamer, from the 'New' button, to create a proxy account.<P>
 * A host in MySQL is simulated by a GroupID having a 'host' type, with the
 * special case of the group 'localhost' that has the type 'host byAcsCreator'
 * to forbid its deleting by the Access Road user.
 * The host to use, if it is not 'localhost', has to be an open ACS having
 * an account with the new userID name as name,
 * and the user 'anonymous' is the single exception to this rule. <P>
 * This constructor ensures that the new account is member of the pattern-name
 * accounts-for-hosts groups that match to it, if it is not the user 'anonymous'.
 * For a new account 'user', an empty non-conditional ACL on 'MySQL Server' is set
 * without rights for the accounts-for-hosts group 'user§@%', if it does not exist.
 * Calls setValueOf(). Calls the super constructor.
 * @param _name ImmutableName of the userID as a kind of eligible party
 * @param _acs ACS owning the userID
 * @param _epType is the eligible party type, and may be null if the ACS allows it.
 * @param _isAdmin true only if the instance has an administrative role in the ACS,
 *    for the technical administration or the users
 * @param _host mandatory host for an account. Never null.
 * @param _proxiedAccount proxied account of this instance. Never null.
 * @exception CreateError if ACS.newEorEligibleParty() throws it - the instance
 * must then be deleted !
 */
  public ProxyMySQLImpl(ImmutableName _name, ACS _acs, String _epType, Boolean _isAdmin,
     GroupIDMySQLImpl _host, UserIDMySQLImpl _proxiedAccount) throws CreateError  {

        super(_name, _acs, _epType, _isAdmin);

        if (_host == null || _proxiedAccount == null)
            throw new CreateError(" The host of a proxy account cannot be null, "+
                    "nor its proxied account. ");

          // for testing purpose
          //System.out.println("call to ProxyMySQLImpl constructor with 2 arguments, for "+_name);

        try {
            setValueOf(this, _host, _proxiedAccount);
        }
        catch (CreateError c) {
            try {
                finalizeForBase();
            }
            catch (UpDateError u) {
                throw new CreateError("From ProxyMySQLImpl constructor "+c.toString()+
                        ". Error in finalizing "+u.toString());
            }

            throw c;
        } // catch c

  } // ProxyMySQLImpl()

   /**
   * Unused for the creation of a ProxyMySQL.
   * Calls the super constructor and setValueOf().
   * @param _name ImmutableName of the eligible party to create
   * @param _acs Main ACS of the eligible party to create
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users
   * @param _host mandatory host for an account. Never null.
 * @param _proxiedAccount proxied account of this instance. Never null.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if ACS.newEorEligibleParty() throws it -
   *  the instance must then be deleted !
   */
  public ProxyMySQLImpl(ImmutableName _name, ACS _acs, String _epType, boolean _isAdmin,
    GroupIDMySQLImpl _host, UserIDMySQLImpl _proxiedAccount, boolean _toRegisterInManager)
    throws CreateError {

        super(_name, _acs, _epType, _isAdmin, _toRegisterInManager);

        if (_host == null || _proxiedAccount == null)
            throw new CreateError(" The host of a proxy account cannot be null, "+
                    "nor its proxied account. ");

              // for testing purpose
              //System.out.println("call to ProxyMySQLImpl constructor with 3 arguments, for "+_name);

        try {
            setValueOf(this, _host, _proxiedAccount);
        }
        catch (CreateError c) {
            try {
                finalizeForBase();
            }
            catch (UpDateError u) {
                throw new CreateError("From ProxyMySQLImpl constructor "+c.toString()+
                        ". Error in finalizing "+u.toString());
            }
            // for an user message
            throw c;
        } // catch u

  } // ProxyMySQLImpl()

  // --------------- ProxyIDMySQL public methods ------------------------------

  /**
   * Gets the proxied account under which this proxy account gets MySQL rights.
   * @return the proxied account. Never null.
   */
  public UserIDMySQL getProxiedAccount() {
    return proxiedAccount_;
  } // getProxiedAccount()

  //
  // To implements the UserIDImpl methods
  //

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
     return serialVersionUID__;
  }

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon.Returns always null,
   * because these events are listened only to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject has
   * one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()

   /**
   * This method removes this instance in its groups directly.
   * The superclass cannot be used, since it is done through a call to
   * GroupIDMySQLImpl.removeEorMember(), which may fire an exception.
   * Calls GroupIDMySQLImpl.removeEorMemberWithoutControl() and
   * UserIdMySQLImpl.removeEorSoftAlias().
   * Overrides the super method and calls it at the end.
   * This method HAS TO BE USED before deleting the UserID.
   * Note: when this method is called by the EligiblePartyImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the instance cannot be removed from its first host group
   */
   public synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("UserIDImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (folderMember_ == null) {
      return;
    }

    // set l_Groups_ to an empty array
    List<GroupIDMySQLImpl> l_grps = null;
    try {
        if (l_Groups_ != null) { // false if called by EligiblePartyImpl constructor
            l_grps = new ArrayList(Arrays.asList(l_Groups_));

            for (GroupIDMySQLImpl grp : l_grps) {
                // calls back removeGroupToGM() and updates l_Groups_
                grp.removeEorMemberWithoutControl(this);
            } // for i
        } // if
    }
    catch (CreateError c) {

              // for testing purpose
              //System.out.println("ProxyMySQLImpl.finalizeForBase() with exception ");

        throw new UpDateError("ProxyMySQLImpl.finalizeForBase() "+c.toString());
    }

    if (proxiedAccount_ != null)
        proxiedAccount_.removeEorSoftAlias(name_);

    l_grps = null;
    host_ = null;
    proxiedAccount_ = null;
    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'CurrentGroupID'.
  * Overrides the super method and calls it.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      if (_propertyName.equals("CurrentGroupID")  )
          changeSupport_.addPropertyChangeListener(_propertyName, _l);
      else super.addPropertyChangeListener(_propertyName, _l);

  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties: 'CurrentGroupID'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("ProxyMySQLImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("CurrentGroupID") )
          changeSupport_.removePropertyChangeListener(_propertyName, _l);
      else super.removePropertyChangeListener(_propertyName, _l);

  } // removePropertyChangeListener()

  // --------------- UserIDMySQL public methods ------------------------------

  /**
   * @return the immutable first host of this MySQL user. May be null only at the creation
   * or if finalized.
   */
  public GroupIDMySQLImpl getEorFirstHost() {
       return host_;
  }

  /**
  * This method provides a copy of the instance with another name.
  * Do not copy the own resources or the current actors.
  * The copy order is 3 times the source order.
  * Caution: do not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of UserIDImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null  ) {
          throw new CreateError("in ProxyMySQLImpl.copy(),"+
          " _name or _acs is null.");
      }

      ProxyMySQLImpl result = new ProxyMySQLImpl(
          _name, _acs, getEPType(), false, host_, proxiedAccount_,
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);

      return result;
  } // copy(,,,)

  /**
  * This method provides a copy of the instance
  * as copy() but without copying the internal AclEntries.
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of UserIDImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, 
    Directory _parent, Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in ProxyMySQLImpl.copy(),"+
          " _name or _acs is null.");
      }

      ProxyMySQLImpl result = new ProxyMySQLImpl(
          _name, _acs, getEPType(), false, host_, proxiedAccount_,
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclEntry copy

      //if (result.getEorL_AclConnectedResources().isEmpty()) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copy(,,,)

  // ------------------ protected methods -----------------------------------------

  /**
  * This method provides a copy of the instance variables.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal AclEntries
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError
  **/
  protected ProxyMySQLImpl copyOwnVariables(ProxyMySQLImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in ProxyMySQLImpl.copyOwnVariables(),"+
          " _result is null.");
      }

      _result = (ProxyMySQLImpl)super.copyOwnVariables(_result, _withoutInternalAcl);

      _result.host_ = host_;
      _result.proxiedAccount_ = proxiedAccount_;

      return _result;
  } // copyOwnVariables(,,,)

/**
 * Sets, for the constructors, a new account as member of the matched pattern-name groups
 * and member of its mandatory host.
 * A host in MySQL is simulated by a GroupID having a 'host' type, with the
 * special case of the group 'localhost' that has the type 'host byAcsCreator'
 * to forbid its deleting by the Access Road user.
 * The host to use, if it is not 'localhost', has to be an open ACS having
 * an account with the new userID name as name,
 * and the user 'anonymous' is the single exception to this rule. <P>
 * This constructor ensures that the new account is member of the pattern-name
 * accounts-for-hosts groups that match to it, if it is not the user 'anonymous'.
 * For a new account 'user', an empty non-conditional ACL on 'MySQL Server' is set
 * without rights for the accounts-for-hosts group 'user§@%', if it does not exist.
 * Called by the constructor.
 * Calls GroupIDMySQLImpl.isMemberOfMySQLPatternNameGroup() and addEorMemberToPatternNameGroup().
 * @param _user new instance of this class
 * @param _host host group is mandatory. Never null.
 * @param _proxiedAccount proxied account of the proxy _user. Never null.
 * @exception CreateError if the constraints are not fulfilled with
 */
  protected static void setValueOf(ProxyMySQLImpl _user, GroupIDMySQLImpl _host, 
          UserIDMySQLImpl _proxiedAccount) throws CreateError {

    if (_user.getName().lastComponent().indexOf("%") != -1 ||
            _user.getName().lastComponent().indexOf("_") != -1 ||
            _user.getName().lastComponent().indexOf("§") != -1 ||
            _user.getName().lastComponent().indexOf("@") != -1) {
        throw new CreateError(" The name of a MySQL account cannot contain "+
                "one of the characters '%','_', '§' or '@'."+
                " In an ACL, the condition group is the account host.");
    } // if
    if (_user.getName().lastComponent().length() > 16) {
        throw new CreateError(" User names can be up to 16 characters long, but no more.");
    } // if
    if (_host.getType().indexOf("host") == -1 ||
            _host.getType().indexOf("hosts group") != -1 ||
            _host.getType().indexOf("standard") != -1)
        throw new CreateError(" The host of an account has to be a simple host. ");
    if (_proxiedAccount instanceof ProxyMySQL) {
        throw new CreateError(" The proxied account cannot be a proxy account.");
    }

    // set the proxied account
    try {
      // set as reference of this user
      _proxiedAccount.addEorSoftAlias(_user.getName());
      _user.proxiedAccount_ = _proxiedAccount;

    } catch (UpDateError u) {
      try {
          _user.finalizeForBase();
      } catch (UpDateError r) {
        throw new CreateError("ProxyMySQLImpl constructor: "+
              "wrong handling of the proxied account. "+
              "Error on finalizing "+
              r.toString());
      } // catch

      throw new CreateError(
              " Wrong handling of the proxied account "+
              u.toString());
    } // UpDateError

    // host is mandatory. May throws CreateError.
    _host.addEorMember(_user);
    _user.host_ = _host;

    _user.setComment("The proxied account of this proxy (see 'Alias') "+
            "is the only mean to deliver MySQL rights to this proxy.");
  } // setValueOf()

} // ProxyMySQLImpl
