//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.util.Map;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Arrays;
import java.util.List;
import java.util.Collections;

//import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
//import java.beans.PropertyChangeSupport;

/**
 * This class is responsible for modeling a MySQL resource, mainly to implement
 * the AclEntry sorting for the direct and the inherited ACL.
 * The documentation of DirectoryMySQLImpl describes in details the ACL management. <P> <P>
 * This class implements the following rules, and informs the Access Road user
 * with the same sentence as error message when they are not fulfilled with: <P>
 * - A table column name must end with 'column'. <P>
 * - An host group can have an AclEntry only for the target 'MySQL server'. <P>
 * - A MySQL simple resource allows only conditional AclEntries. <P>
 * - A MySQL conditional AclEntry cannot use a standard group.
 * It is a javabean without specific bound properties. The 'InheritedTargetAclEntries'
 * events are bound when the primary AclEntries map is changing.
 * Excepting for EPRViewInBaseImpl, the listeners are outside gBase,
 * they are transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * This class has no subclass in this package.
 * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.ResourceMySQLImpl
 * @version 0.7.0
 */
public class ResourceMySQLImpl extends ResourceImpl implements ResourceMySQL {
  // private variables
  private static final long   serialVersionUID__ = 0;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ResourceMySQLImpl EMPTY_INSTANCE = new ResourceMySQLImpl();

   /**
   * Default constructor. No registration in the ACS.
   */
  public ResourceMySQLImpl() {

                          // for testing purpose
                     //System.out.println("Call to ResourceMySQLImpl constructor");
  }

   /**
   * Used by the beamer editor for the ACS resources list, from the 'New' button,
   * and by the gWork.AcsFactory.
   * @param _name of the resource
   * @param _acs main ACS of the resource
   * @param _parent of the resource; may be null.
   * @param _owner owns the resource; may be null. Never used.
   * @param _resourceType the resource type which drives the rights and the immutability
   * of the instance. If null, the value is set to "<undefined>".
   * @exception CreateError if ACSImpl.newEorResource() throws it;
   * the instance has then to be deleted !
   */
  public ResourceMySQLImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType) throws CreateError {

      this(_name, _acs, (Directory)_parent, null, _resourceType, true);

                          // for testing purpose
                     //System.out.println("Call to ResourceMySQLImpl constructor("+
                       // "ImmutableName _name, ACS _acs, Node _parent, String _resourceType)"+
                       // " for "+_name);

  } // ResourceImpl constructor

   /**
   * Used by the other constructor.
   * The inherited AclEntries from the parent are added.
   * A table column must have 'column' at the end of its name.
   * Calls first the super constructor. Calls addDatabaseGroupsACL().
   * @param _name of the resource
   * @param _acs main ACS of the resource
   * @param _parent owns the resource; may be null.
   * @param _owner owns the resource; may be null. Never used.
   * @param _resourceType the resource type which drives the rights and the immutability
   * of the instance. If null, the value is set to "<undefined>".
   * @param _toRegisterInManager false if this method must not register
   *   the instance in its ACS.
   * @exception CreateError if ACSImpl.newEorResource() throws it.
   * The instance has then to be deleted !
   */
  public ResourceMySQLImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType, boolean _toRegisterInManager) throws CreateError {

      super(_name, _acs, _parent, null, _resourceType, true);

                          // for testing purpose
                     //System.out.println("ResourceMySQLImpl constructor for "+_name);

      if (_parent != null ) {
            // a table column must have 'column' at the end of its name
            if (_resourceType.indexOf("table_column") != -1 &&
                    !_name.lastComponent().endsWith("column") ) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError("From ResourceMySQLImpl constructor. "+
                      "error on the table_column name "+
                      ". Error on finalizing "+u.toString());
              }

              throw new CreateError(" A table column name must end with 'column'.");
            } // if

            if (_resourceType.indexOf("index") != -1) {
                setComment(comment_ + " An index is a passive object. No direct ACL.");
            }

            // the AclEntries and the inherited AclEntries from the parent are added
            AclEntry    acl2 = null;
            AclEntry[]  l_acl = _parent.getL_aclEntry();
            ArrayList<AclEntry> l_inh_acl = ((ResourceImpl)_parent).getEorL_TargetInheritedAclEntries();
            boolean     noInheritance = false;

            for (int i = 0; i < l_acl.length; i++) {
                acl2 = selectInheritedAclEntry(l_acl[i], type_);
                if (acl2 == null) {
                    noInheritance = true;
                    continue;
                }

                addInheritedAclEntry(acl2);
            } // for i

            for (AclEntry ac : l_inh_acl) {
                acl2 = selectInheritedAclEntry(ac, type_);
                if (acl2 == null) {
                    noInheritance = true;
                    continue;
                }

                addInheritedAclEntry(acl2);
            } // for ac

            if (noInheritance)
                setComment(comment_ + " No ACL inheritance for this type.");

            l_acl = null;
            l_inh_acl = null;
            acl2 = null;
      } // if (_parent != null)

  } // ResourceMySQLImpl constructor

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon.Returns always null,
   * because these events are listened only to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: is optional, where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL. <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject
   * as one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()

    /**
    * Creates the AclEntry, and sorts the new AclEntry list.
    * This method is derived from the same method in DirectoryMySQLImpl.
    * On every child of the root 'MySQL server', a new conditional ACL for a pair
    * (user@host or accounts-for-hosts group) produces a new ACL
    * on the root for the same pair if it does not exist. This new root ACL has
    * the right USAGE, and it is not deleted by default when the initial ACL is deleted.
    * For a new non-conditional ACL on a database, there is no adding on the root. <P>
    * The ACS types policy, with 'Resource.NoNonConditionalACL' and 'Resource.NoConditionalACL',
    * sets that there is an empty 'AclEntry' property for the 'index' resources.
    * Stops then restores the property change firing on 'AclRights'.
    * The generic firing order on 'AclRights' and 'EPsThroughAclEntry' is inversed
    * in this method.
    * The four types of host groups can be used as ACL sources only for the databases,
    * which are not ResourcemySQL.
    * Calls BaseUtilityMySQLImpl.addInSortedAclList().
    * Calls the overridden method before the sorting of the list.
    * The detailed sequence of calls is described at the method DirectoryMySQLImpl.addAclEntry(). <P>
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the new AclEntry, and to be recorded
    *  as an ACS controller of this resource if it is not its main ACS.
    * @param _ep eligible party connected to the resource through a new aclEntry.
    * @param _l_right is an array of positive or negative rights of _ep on this resource
    *  (that is, if READ is a right, _ep can read on the resource). May be null.
    * @param _positiveRight is true if the rights are set to authorize (grant),
    *  and false if they are set to forbid (deny).
    * @param _conditionGroup first condition group. May be null.
    * @exception InternalError if this resource has a null name or a null ACS,
    *  or if _ep is not an EligiblePartyImpl nor an ActorImpl.
    * @exception CreateError if _ep or _aclAcs is null, or if _ep is already recorded
    *  in this resource with the same ACS, or thrown by the ACS.newEorAclEntry(),
    * or without a condition group and not for an accounts-for-hosts group,
    * or with condition group while the source is not an account
    * @return the new AclEntry.
    */
  public synchronized AclEntry addAclEntry(ImmutableACS _aclAcs,
    ImmutableEligibleParty _ep, StringRight[] _l_right, Boolean _positiveRight,
    ImmutableGroupID _conditionGroup) throws CreateError {

        PropertyChangeListener[]    l_listeners = null;
        AclEntry[]                  l_oldAclEntry = null;
        AclEntry                    acl = null;
        ArrayList<AclEntry>         l_acl = null;

        // The five types that can be used as ACL sources
        //_ep.getEPType().indexOf("host") != -1 ||
        //_ep.getEPType().indexOf("host <byAcsCreator>") != -1 ||
        //_ep.getEPType().indexOf("pattern-name_hosts group") != -1 ||
        //_ep.getEPType().indexOf(pattern-name_accounts-for-hosts group") != -1 ||
        //_ep.getEPType().indexOf("account") != -1) {

        if (_conditionGroup != null && _ep.getEPType().indexOf("account") == -1)
            throw new CreateError(" A MySQL simple resource allows conditional ACL "+
                    "only to accounts.");
        else if (_conditionGroup == null &&
                _ep.getEPType().indexOf("pattern-name_accounts-for-hosts group") == -1)
            throw new CreateError(" A MySQL simple resource allows non-conditional ACL"+
                    "only to accounts-for-hosts groups.");

        // stop the property change firing on 'AclRights', and not for 'EPsThroughAclEntry'
        l_listeners = getPropertyChangeListeners("AclRights");
        
        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                removePropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if

        l_oldAclEntry = getL_aclEntry();

        // firing on 'EPsThroughAclEntry'
        try {
            acl = super.addAclEntry(_aclAcs, _ep, _l_right, _positiveRight, _conditionGroup);
        } catch (CreateError c) {
            // restore the property change firing on 'AclRights'
            if (l_listeners != null) {
                for (int i = 0; i < l_listeners.length; i++) {
                    addPropertyChangeListener("AclRights", l_listeners[i]);
                } // for i
            } // if
            l_listeners = null;
            l_oldAclEntry = null;
            throw c;
        } // catch

        l_acl = new ArrayList(Arrays.asList(l_aclEntry_));

        // The new AclEntry has been added at the end of the current list, but no use here
        l_acl.remove(acl);
        l_acl = BaseUtilityMySQLImpl.addInSortedAclList(l_acl, acl);
        l_aclEntry_ = l_acl.toArray(new AclEntry[0]);

        // restore the property change firing on 'AclRights'
        if (l_listeners != null) {
            for (int i = 0; i < l_listeners.length; i++) {
                addPropertyChangeListener("AclRights", l_listeners[i]);
            } // for i
        } // if
 
        // add the primary rights directly
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                null,
                m_inheritedT_ACLRights_,
                true);

        // add to 'MySQL server' an ACL with the USAGE right, if necessary
        boolean found = false;
        DirectoryMySQLImpl server = null;
        if (l_inheritedT_AclEntries_ != null) {
            // no direct reading of server.l_aclEntry_
            for (AclEntry acl2 : l_inheritedT_AclEntries_) {

                if (!acl2.getResType().equals("MySQL server <byAcsCreator>"))
                   break;
                
                if (acl2.getEorEP() != _ep)
                    continue;

                if (acl2.getL_ConditionSourceGroups().isEmpty() &&
                        _conditionGroup == null) {
                    found = true;
                    break;
                }
                else if (acl2.getL_ConditionSourceGroups().isEmpty() &&
                        _conditionGroup != null) {
                    continue;
                }
                else if (acl2.getL_ConditionSourceGroups().get(0) ==
                            _conditionGroup) {
                    found = true;
                    break;
                } // if
            } // for i
        } // if

        if (!found) {
            server = (DirectoryMySQLImpl) aCS_.getEorM_Resources().get(
                        ((Name)aCS_Name_.clone()).extended("MySQL server"));

            // calls back addInheritedAclEntry() on all the 'MySQL server' children,
            // including this instance.
            server.addAclEntry(_aclAcs, _ep,
                    new StringRight[] {aCS_.getOneRight("USAGE")},
                    _positiveRight, _conditionGroup);
        } // if

        // the ACL ACS and the ACL source are already informed
        try {
            firePropertyChange("AclRights", l_oldAclEntry, getL_aclEntry());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ResourceMySQLImpl.addAclEntry(), firePropertChange() throws");
            e.printStackTrace();
        }

        l_listeners = null;
        l_oldAclEntry = null;
        l_acl = null;
        server = null;
        return acl;
    } // addAclEntry()

    /**
    * Removes the AclEntry and the primary rights.
    * Calls the overridden method first.
    * Fires 'InheritedTargetAclEntries' events only to update the display
    * of the primary rights map.
    * The detailed sequence of calls for adding is described at the method addAclEntry()
    * of DirectoryMySQLImpl.
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public synchronized void removeAclEntry(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

                                          // for testing and tracking
                                          //System.out.println(
                                            //  "start of ResourceMySQLImpl.removeAclEntry() in: "+
                                              //name_+" for _ep "+_ep.getNickName()+
                                              //", _positiveRight "+_positiveRight+
                                              //", _conditionGroup "+_conditionGroup.getNickName());

        // firing on 'AclRights' and 'EPsThroughAclEntry'
        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);
        
        // removes the primary rights
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                null,
                m_inheritedT_ACLRights_,
                true);

                                      // for testing and tracking
                                      //System.out.println("ResourceMySQLImpl.removeAclEntry() in: "+
                                        //      name_+" fires the 'InheritedTargetAclEntries' event");

        // to update the display of the primary rights map
        try {
            firePropertyChange(
                    "InheritedTargetAclEntries", null, l_inheritedT_AclEntries_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From DirectoryMySQLImpl.removeAclEntry(), "+
                    "firePropertChange() throws");
            e.printStackTrace();
        } // catch


                          // for testing and tracking
                  //System.out.println("at the end of ResourceMySQLImpl.removeAclEntry(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}
  } // removeAclEntry()

    /**
    * Adds the inherited AclEntry. Called by DirectoryMySQLImpl.addAclEntry()
    * and the same method, and by the constructor. An inherited AclEntry may have
    * a Resource which is not this instance.
    * The sorting in the ACL list is described in BaseUtilityMySQLImpl.
    * From ResourceMySQL. Copied in DirectoryMySQLImpl and ActorMySQLImpl.
    * Calls getEorL_TargetInheritedAclEntries(), BaseUtilityMySQLImpl.addInSortedInheritedAclList()
    * and super.setEorL_TargetInheritedAclEntries() that fires 'InheritedTargetAclEntries'.
    * Not described in the BeanInfo since the user does not call it.
    * The detailed sequence of calls is described at the method DirectoryMySQLImpl.addAclEntry(). <P>
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _acl inherited AclEntry to put in the list at the right order
    */
  public synchronized void addInheritedAclEntry(AclEntry _acl) {
    ArrayList<AclEntry> l_acl = null;

    l_acl = getEorL_TargetInheritedAclEntries();

    l_acl = BaseUtilityMySQLImpl.addInSortedInheritedAclList(l_acl, _acl);

    setEorL_TargetInheritedAclEntries(l_acl);

    l_acl = null;
  } // addInheritedAclEntry()

    /**
    * Removes the inherited AclEntries in the children.
    * From ResourceMySQL. Copied in DirectoryMySQLImpl and ActorMySQLImpl.
    * Calls getEorL_TargetInheritedAclEntries() and
    * setEorL_TargetInheritedAclEntries() that fires 'InheritedTargetAclEntries'.
    * Not described in the BeanInfo since the user does not call it.
    * The detailed sequence of calls for adding is described at the method
    * DirectoryMySQLImpl.addAclEntry(). <P>
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _acl inherited AclEntry
    * @exception UpDateError if _acl is null or unknown
    */
  public synchronized void removeInheritedAclEntry(AclEntry _acl) throws UpDateError {

              // for testing and tracking
              //System.out.println("Call to ResourceMySQLImpl.removeInheritedAclEntry() in: "+name_);

      ArrayList<AclEntry>   l_acl = getEorL_TargetInheritedAclEntries();
      AclEntry              acl = null;
      boolean               found = false;

      for (Iterator ite = l_acl.iterator(); ite.hasNext(); ) {
        acl = (AclEntry)ite.next();
        if (acl != null && !acl.isEmpty()) {
          if (acl.equals(_acl)) {
              ite.remove();
              found = true;
              break;
          }
        } // if
      } // for ite

      if (!found)
          throw new UpDateError("ResourceMySQLImpl.removeInheritedAclEntry(), "+
                  "AclEntry is unknown "+_acl);

      setEorL_TargetInheritedAclEntries(l_acl);

      l_acl = null;
      acl = null;
  } // removeInheritedAclEntry()

    /**
    * Returns the AclEntry to use as inherited AclEntry in a child.
    * Returns null if the type contains 'index', 'directory', or starts with 'byAcsCreator'.
    * Otherwise, returns the argument.
    * Called by the constructors. 
    * Overrides the super method and does not call it.
    * Calls the ResourceMySQL static method having the same name.
    * @param _acl the initial AclEntry which may be given to the child
    * @param _type type of a child, as Resource
    * @return the same AclEntry if the child inherits this initial AclEntry, or null
    */
  public AclEntry selectInheritedAclEntry(AclEntry _acl, String _type) {
    return selectInheritedAclEntry(this, _acl, _type);
  } // selectInheritedAclEntry()

    /**
    * Returns the AclEntry to use as inherited AclEntry in a chil of the MySQL resources tree.
    * Returns null if the type contains 'index', 'connections', 'user socket' or 'virtual_objects',
    * of it it starts with 'byAcsCreator'.
    * Otherwise, returns the argument. Called by the constructors.
    * Overrides the super method and does not call it.
    * @param _res the current instance is not used in the method
    * @param _acl the initial AclEntry which may be given to the child
    * @param _type type of a child, as Resource
    * @return the same AclEntry if the child inherits this initial AclEntry, or null
    */
  public static AclEntry selectInheritedAclEntry(ImmutableResource _res,
          AclEntry _acl, String _type) {
    if (_type.indexOf("index") != -1)
        return null;
    else if (_type.startsWith("<byAcsCreator>"))
        return null;
    else if (_type.equals("connections <byAcsCreator>") ||
            _type.equals("virtual_objects <byAcsCreator>") ||
            _type.endsWith("user socket <byAcsCreator>"))
        return null;
    else
        return _acl;
  } // selectInheritedAclEntry()
  
    /**
    * Resets the inherited AclEntry rights and the primary rights after the user editing
    * of an AclEntry.
    * Called by the parent when there is a change in the effective rights of
    * an inherited AclEntry, including a change in the condition group.
    * Called by Directory.resetInheritedAclEntryRightsOfChildren().
    * Overreads the super method to update direct the two inherited ACL maps,
    * and does not call it. The generic gBase package provides
    * the handling of the listeners for the property changes
    * 'InheritedTargetAclEntries', but without any firing
    * of these changes.
    * Calls setM_TargetInheritedACLRights(), since it is not possible to reset
    * the two inherited rights maps without analyzing all the inherited AclEntries.
    * Fires the 'InheritedTargetAclEntries' event.
    * @param _acl inherited AclEntry with updated rights. Never null.
    */
  public synchronized void resetInheritedAclEntryRights(AclEntry _acl) {
      
        setM_TargetInheritedACLRights(l_inheritedT_AclEntries_);

        // to update the display of the primary rights map
        try {
            firePropertyChange(
                    "InheritedTargetAclEntries", null, l_inheritedT_AclEntries_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the data update, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ResourceMySQLImpl.resetInheritedAclEntryRights(), "+
                    "firePropertChange() throws");
            e.printStackTrace();
        } // catch
  } // resetInheritedAclEntryRights()

  /**
   * Gets the internal map of the primary rights from each parent level of this resource.
   * The primary rights are displayed into the beamer. They help the user
   * to understand the connection pairs, since it covers the root 'MySQL server' rights,
   * and to understand the complex database rights. Because this method overrides
   * the Resource method, its name is 'TargetInheritedACL' rather than 'TargetPrimaryACL'.
   * Overrides the Resource method to describe its use by the MySQL addon.
   * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl#getM_TargetInheritedACLRights()
   * @return unmodifiable map of primary rights from the direct and inherited AclEntries.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedACLRights() {
      if (m_inheritedT_ACLRights_ == null) return Collections.EMPTY_MAP;
      else {
          return Collections.unmodifiableMap(m_inheritedT_ACLRights_);
      } // else;
  }

    /**
    * Removes the direct AclEntry but without event firing. The purpose of this method 
    * is to avoid an unuseful sequence of firings when a lot of AclEntries are removed. 
    * It is called from the root 'MySQL server' method removeAclEntry() when 
    * it deletes all the direct ACL of a pair on the root children when the root ACL 
    * for this pair is deleted.
    * Calls the overridden method and disables before all the firings for 
    * the 'AclRights' and 'EPsThroughAclEntry' events.
    * @see DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID)
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public void removeAclEntryWithoutFiring(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

        if (type_.equals("MySQL server <byAcsCreator>"))
            throw new InternalError("call to ResourceMySQLImpl.removeAclEntryWithoutFiring()"+
                    "on the root");

                                          // for testing and tracking
                                          //System.out.println(
                                            //  "start of ResourceMySQLImpl.removeAclEntryWithoutFiring() in: "+
                                              //name_+" for _ep "+_ep.getNickName()+
                                              //", _positiveRight "+_positiveRight+
                                              //", _conditionGroup "+_conditionGroup.getNickName());
      
        PropertyChangeListener[]    l_listeners = null;
        PropertyChangeListener[]    l_listeners2 = null;

        // 'EPsThroughAclEntry' and 'AclRights' are bound properties in ResourceImpl
        // do unregister before the call to super.removeAclEntry()
        l_listeners = getPropertyChangeListeners("AclRights");
        l_listeners2 = getPropertyChangeListeners("EPsThroughAclEntry");

        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                removePropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                removePropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

        super.removeAclEntry(_aclAcs, _ep, _positiveRight, _conditionGroup);

        // restore the property change firings
        if (l_listeners != null) {
            for (int j = 0; j < l_listeners.length; j++) {
                addPropertyChangeListener("AclRights", l_listeners[j]);
            } // for i
        } // if
        if (l_listeners2 != null) {
            for (int j = 0; j < l_listeners2.length; j++) {
                addPropertyChangeListener("EPsThroughAclEntry", l_listeners2[j]);
            } // for i
        } // if

        // removes the primary rights
        BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                Arrays.asList(getL_aclEntry()),
                null,
                m_inheritedT_ACLRights_,
                true);

                          // for testing and tracking
                  //System.out.println("at the end of ResourceMySQLImpl.removeAclEntryWithoutFiring(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}

        l_listeners = null;
        l_listeners2 = null;
  } // removeAclEntryWithoutFiring()

 /**
  * Clones without registering of the new object in the ACSs.
   * Caution: don't use any updating method on this clone,
   * because that may destroy the base integrity. A finalized instance returns
   * a clone with getEorACS() which returns null,
   * but getName() and getAcsName() operates.
   * @exception CloneNotSupportedException
   * @return Object; is not registred in the ACSs. The names, the aclEntries,
   * the AG rights, the ACS controller names are cloned.
   * The AclEntries are weakly cloned, so they have no direct link to this resource.
   * The external object references are not cloned
   */
  public synchronized Object clone() {

      ResourceMySQLImpl clon = (ResourceMySQLImpl) super.clone();

      return clon;
  } // clone()

   /**
   * This method sets to null the primary inherited rights map.
   * Overrides the super method and calls it at the end.
   * @exception UpDateError if the main ACS do not known the resource (if
   * it is a clone without setName(), typically)
   */
  public synchronized void finalizeForBase() throws UpDateError {

                    // for testing and tracking
                //System.out.println("In ResourceMySQLImpl, call of " + name_ + ".finalizeForBase()");

    if (folderMember_ == null) return;

    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * Provides a copy of the instance
  * with another name. Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Most of the aclEntries are copied, but not the list of ViewInBase
  * that contains this instance.
  * Caution: does not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same last component in this instance user or group names.
  * <P>
  * For aclEntries, if _result ACS is not this resource ACS, and if
  * the aclEntry ACS is this resource ACS, sets _result ACS as
  * the aclEntry ACS of the copied acl.
  * Caution: does not copy an aclEntry if: <P>
  *  - _acs is not this resource ACS, <P>
  *  - the acl eligible party ACS is this resource ACS, <P>
  *  - and there is no matched EligibleParty (with the same last components)
  *      in _acs <P>
  * Calls copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
     May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ResourceImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public synchronized Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ||
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError(
          "The name or the acs is null, or the bridge target is not a Resource,"+
          " or the parent ACS is not the instance ACS.");
      }

      ResourceMySQLImpl result = new ResourceMySQLImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());

      // copies all the aclEntries
      result = copyOwnVariables(result, false);

      return result;
  } // copy(,,,)

  /**
  * Provides a copy of the instance as Copyable.copy(),
  * but without copying the internal aclEntries.
  * It is used for ACSImpl.copy(), because all the EP must be copied
  * first before the copy of all the resources with their aclEntries.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy),
  * but without copy for the internal aclEntry nor the list of ViewInBase
  * that contains this instance.
  * Called from BaseUtilitiesImpl.copyDirectoriesAndActors().
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of this class
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public synchronized Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs,
    Directory _parent, Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in ResourceImpl.copyWithoutInternalAcl(),"+
          " _name or _acs is null.");
      }

      ResourceMySQLImpl result = new ResourceMySQLImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclentries copy

      //if (result.m_AclEntriesResources_ == null) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copyWithoutInternalAcl(,,,)


  //
  // ------------------ overridden protected methods -------------------------------------------
  //
  
  /**
  * This method provides a copy of the instance variables.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy).
  * Most of the AclEntries are copied, but not the list of ViewInBase
  * that contains this instance, nor comments.
  * In most of the cases, the aclEntry arguments are copied,
  * but see the exceptions.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * Calls copyOwnAcl().
  * Caution: does not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same last component in this instance user or group names.<P>
  * For aclEntries, if _result ACS is not this resource ACS, and if
  * the aclEntry ACS is this resource ACS, sets _result ACS as
  * the aclEntry ACS of the copied acl.
  * Caution: does not copy an aclEntry if: 
  *  - _result ACS is not this resource ACS,
  *  - the eligible party ACS is this resource ACS,
  *  - and there is no matched EligibleParty (with the same last components)
  *      in the _result ACS
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return the updated _result
  * @exception CreateError of addAclEntry()
  **/
  protected ResourceMySQLImpl copyOwnVariables(ResourceMySQLImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      return (ResourceMySQLImpl)super.copyOwnVariables(
              _result, _withoutInternalAcl);
  } // copyOwnVariables(,,,) 
  
  /**
  * This method provides a copy of the AclEntries,
  * all the AclEntries (see restrictions) or only the external ones.
  * Most of the aclEntries are copied, but not the list of ViewInBase
  * that contains this instance, nor comments.
  * If _result ACS is not this resource ACS, and if
  * the aclEntry ACS is this resource ACS, sets _result ACS as
  * the aclEntry ACS of the copied acl.
  * Called by copyOwnVariables() and by
  * BaseUtilityMySQLImpl.copyDirectoryChildrenForACS().
  * @param _result to update with this instance AclEntries copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @param _withoutExternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError of addAclEntry()
  **/
  protected ResourceMySQLImpl copyOwnAcl(ResourceMySQLImpl _result,
    boolean _withoutInternalAcl, boolean _withoutExternalAcl) throws CreateError {

      return (ResourceMySQLImpl)super.copyOwnAcl(
              _result, _withoutInternalAcl, _withoutExternalAcl);
  } // copyOwnAcl(,,,)

  /**
   * Currently unused.
   * This method sets to null the primary inherited rights map, since it is used
   * after a deserialization.
   * Overrides the super method and calls it at the end.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In ResourceMySQLImpl, call of " + name_ +
                //".finalizeFromDeserialization()");

    if (folderMember_ == null) return;

    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  // ------------------- protected methods -------------------------------------

   /**
   * Sets all the inherited AclEntries for this target. The AclEntry may be a clone.
   * To use by the parent class in the AcsAddon gBase package.
   * Calls setM_TargetInheritedACLRights() to set the maps (source @ host order, StringRight[]).
   * Fires 'InheritedTargetAclEntries' events.
   * @param _l_acl list of inherited AclEntries.
   * May be empty, but never null.
   */
  protected void setEorL_TargetInheritedAclEntries(ArrayList<AclEntry> _l_acl) {
      if (_l_acl== null) throw new InternalError("ResourceMySQLImpl."+
              "setEorL_TargetInheritedAclEntries() with a null argument.");

      ArrayList l_oldList = null;

      if (l_inheritedT_AclEntries_ == null) {
          l_oldList = new ArrayList(INITIAL_CAPACITY_2);
      } else {
          l_oldList = (ArrayList) l_inheritedT_AclEntries_.clone();
      } // else

      l_inheritedT_AclEntries_ = _l_acl;

      // update l_inheritedT_AclEntries_
      setM_TargetInheritedACLRights(l_inheritedT_AclEntries_);

      try {
        firePropertyChange(
                "InheritedTargetAclEntries", l_oldList, l_inheritedT_AclEntries_.clone());
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceMySQLImpl.setEorL_TargetInheritedAclEntries(), "+
                "firePropertChange() throws");
        e.printStackTrace();
      } // catch

      l_oldList = null;
  } // setEorL_TargetInheritedAclEntries()

  /**
   * Sets the internal maps of the inherited AclEntry rights and the primary rights.
   * The map value is a non-null StringRight array of inherited rights, and it may be empty.
   * Called by setEorL_TargetInheritedAclEntries() and resetInheritedAclEntryRights().
   * Does not fire any property change event.
   * Calls the BaseUtilityMySQLImpl method having the same name.
   * @see BaseUtilityMySQLImpl#setM_TargetInheritedACLRights(ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.ResourceMySQL, java.util.List, java.util.Map, java.util.Map, boolean)
   * @param _l_acl current list of inherited AclEntries.
   * May be empty, but never null.
   */
  protected void setM_TargetInheritedACLRights(List<AclEntry> _l_acl) {

                                      // for testing and tracking
                                      //System.out.println("start of ResourceMySQLImpl."+
                                        //      "setM_TargetInheritedACLRights() in: "+
                                          //    name_);

      if (m_inheritedT_ACLRights_ == null)
          m_inheritedT_ACLRights_ = new TreeMap();

      BaseUtilityMySQLImpl.setM_TargetInheritedACLRights(
                this,
                _l_acl,
                null,
                m_inheritedT_ACLRights_,
                false);

                                      // for testing and tracking
                                      //if (name_.lastComponent().equals("product_price column")) {
                                        //  StringRight[] l_rights = null;

                                          //System.out.println("end of ResourceMySQLImpl."+
                                            //  "setM_TargetInheritedACLRights() for 'product_price column' "+
                                              //"with m_inheritedT_ACLRights_:");
                                          //for (Map.Entry entry : m_inheritedT_ACLRights_.entrySet()) {
                                            //  System.out.println("-- key: "+entry.getKey());
                                              //System.out.println("-- value: ");
                                              //l_rights = (StringRight[]) entry.getValue();
                                              //for (int i = 0; i < l_rights.length; i++) {
                                                //  System.out.println("------ right: "+l_rights[i]);
                                              //} // for i
                                          //} // for entry

                                          //l_rights = null;
                                      //} // if


  } // setM_TargetInheritedACLRights()

  //
  // ------------------ public static methods -----------------------------------------
  //
  
  /**
   * Currently unused.
   * Sets the new AclEntries for a database component, to put it in the relevant
   * database groups among '|ToFirstLevelItemsOf-xxx|', '|ToTablesOf-xxx|',
   * '|ToViewsOf-xxx|', '|ToStoredFunctionsOf-xxx|',
   * '|ToStoredProceduresOf-xxx|', where 'xxx' is the database name.
   * Noop if the argument is not the direct or indirect child of a database.
   * Calls Resource.addAclEntry(). Called by the constructor in ResourceMySQLImpl,
   * DirectoryMySQLImpl and ActorMysqLImpl.
   * @param _res to update with two AclEntries on the database groups
   * @exception CreateError from addAclEntry() or if the resource is not a child
   * of the MySQL root node
   */
  public static void addDatabaseGroupsACL(Resource _res) throws CreateError {

                      // for testing and tracking
                      //System.out.println("ResourceMySQLImpl.addDatabaseGroupsACL() start for "+
                        //      "the resource "+_res.getNickName()+
                          //    " and the type "+_res.getType());
                      
      int DatabaseIndex = _res.getName().components().indexOf("MySQL server") + 1;
      if (DatabaseIndex == 0)
          throw new CreateError("ResourceMySQLImpl.addDatabaseGroupsACL() "+
                  "with a wrong resource name");

      if (_res.getType().contains("table") && !_res.getType().contains("table_column")) {
          if (DatabaseIndex == _res.getName().noOfComponents() - 2)
              _res.addAclEntry(_res.getEorACS(),
                  (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                    get(((Name)_res.getEorACS().getName()).extended("<G>").
                    extended("|ToFirstLevelItemsOf-"+
                    _res.getName().component(DatabaseIndex)+"|")),
                  new StringRight[] {_res.getEorACS().getOneRight("CREATE"),
                            _res.getEorACS().getOneRight("TRIGGER"),
                            _res.getEorACS().getOneRight("DROP"),
                            _res.getEorACS().getOneRight("SELECT"),
                            _res.getEorACS().getOneRight("LOCK TABLES")},
                  true, null);

          _res.addAclEntry(_res.getEorACS(),
              (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                get(((Name)_res.getEorACS().getName()).extended("<G>").
                extended("|ToTablesOf-"+
                _res.getName().component(DatabaseIndex)+"|")),
              new StringRight[] {_res.getEorACS().getOneRight("CREATE"),
                    _res.getEorACS().getOneRight("TRIGGER"),
                    _res.getEorACS().getOneRight("DROP"),
                    _res.getEorACS().getOneRight("SELECT"),
                    _res.getEorACS().getOneRight("LOCK TABLES")},
              true, null);
      } // if (_res.getType().contains("table"))

      else if (_res.getType().contains("view_delegate")) {
          if (DatabaseIndex == _res.getName().noOfComponents() - 2)
              _res.addAclEntry(_res.getEorACS(),
                  (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                    get(((Name)_res.getEorACS().getName()).extended("<G>").
                    extended("|ToFirstLevelItemsOf-"+
                    _res.getName().component(DatabaseIndex)+"|")),
                  new StringRight[] {_res.getEorACS().getOneRight("DROP"),
                        _res.getEorACS().getOneRight("SHOW VIEW"),
                        _res.getEorACS().getOneRight("CREATE VIEW"),
                        _res.getEorACS().getOneRight("EXECUTE"),
                        _res.getEorACS().getOneRight("SELECT")},
                  true, null);

          _res.addAclEntry(_res.getEorACS(),
              (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                get(((Name)_res.getEorACS().getName()).extended("<G>").
                extended("|ToViewsOf-"+
                _res.getName().component(DatabaseIndex)+"|")),
              new StringRight[] {_res.getEorACS().getOneRight("DROP"),
                    _res.getEorACS().getOneRight("SHOW VIEW"),
                    _res.getEorACS().getOneRight("CREATE VIEW"),
                    _res.getEorACS().getOneRight("EXECUTE"),
                    _res.getEorACS().getOneRight("SELECT")},
              true, null);
      } // else if (_res.getType().contains("view delegate"))

      else if (_res.getType().contains("stored_function")) {
          if (DatabaseIndex == _res.getName().noOfComponents() - 2)
              _res.addAclEntry(_res.getEorACS(),
                  (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                    get(((Name)_res.getEorACS().getName()).extended("<G>").
                    extended("|ToFirstLevelItemsOf-"+
                    _res.getName().component(DatabaseIndex)+"|")),
                  new StringRight[] {_res.getEorACS().getOneRight("EXECUTE"),
                        _res.getEorACS().getOneRight("ALTER ROUTINE")},
                  true, null);

          _res.addAclEntry(_res.getEorACS(),
              (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                get(((Name)_res.getEorACS().getName()).extended("<G>").
                extended("|ToStoredFunctionsOf-"+
                _res.getName().component(DatabaseIndex)+"|")),
              new StringRight[] {_res.getEorACS().getOneRight("EXECUTE"),
                    _res.getEorACS().getOneRight("ALTER ROUTINE")},
              true, null);
      } // else if (_res.getType().contains("stored function"))

      else if (_res.getType().contains("stored_procedure")) {
          if (DatabaseIndex == _res.getName().noOfComponents() - 2)
              _res.addAclEntry(_res.getEorACS(),
                  (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                    get(((Name)_res.getEorACS().getName()).extended("<G>").
                    extended("|ToFirstLevelItemsOf-"+
                    _res.getName().component(DatabaseIndex)+"|")),
                  new StringRight[] {_res.getEorACS().getOneRight("EXECUTE"),
                       _res.getEorACS().getOneRight("ALTER ROUTINE")},
                  true, null);

          _res.addAclEntry(_res.getEorACS(),
              (ImmutableEligibleParty)_res.getEorACS().getEorM_EligibleParties().
                get(((Name)_res.getEorACS().getName()).extended("<G>").
                extended("|ToStoredProceduresOf-"+
                _res.getName().component(DatabaseIndex)+"|")),
              new StringRight[] {_res.getEorACS().getOneRight("EXECUTE"),
                   _res.getEorACS().getOneRight("ALTER ROUTINE")},
              true, null);
      } // else if (_res.getType().contains("stored procedure"))

                              // for testing and tracking
                              //System.out.println("ResourceMySQLImpl.addDatabaseGroupsACL() end "+
                                //      "for the resource "+_res.getNickName()+
                                  //    " and the type "+_res.getType());
                      
  } // addDatabaseGroupsACL()

} // ResourceMySQLImpl


