//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.UpDateError;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import java.io.ObjectStreamField;

/**
 * This class is responsible for modeling a MySQL account to
 * manage the constraints on the account membership, always with the type 'account'. <P>
 * A host in MySQL is simulated by a GroupID having a 'host' type, with the
 * special case of the group 'localhost' that has the type 'host byAcsCreator'
 * to forbid its deleting by the Access Road user.
 * The host to use, if it is not 'localhost', has to be an open ACS having
 * an account with the new userID name as name (modulo '(N)' at the end),
 * and the user 'anonymous' is the single exception to this rule.
 * MySQL does not allow any wildcard in the user names. <P>
 * The rights of 'tom@host_one' and 'tom@host_two' can be defined in two conditional ACLs
 * but not as two accounts. In the database rights, MySQL handles the concept of user
 * independantly of the concept of host, so it is necessary to modelize accounts
 * as UserIDs. The account 'tom' may be set as member of only one host, and it may
 * be necessary to define the account 'tom(2)' on another host. <P>
 * There are two types of pattern-name groups, for hosts and for accounts-for-hosts,
 * listed hereinafter: <P>
 * (1) A pattern-name hosts group may be created by the Access Road user to contain
 * all the hosts having a given pattern in their names. <P>
 * (2) A pattern-name accounts-for-hosts group may be created by the Access Road user
 * to contain all the accounts having a given pattern, like 'tom' and 'tom(2)'.
 * The pattern-name accounts-for-hosts group name
 * has then the form 'name§@YYYY', where YYYY contain a wildcard character. <P>
 * The convention for these two types of pattern-name groups
 * is to use in their names one '%' or '_' wildcard character.
 * This class ensures that the accounts or groups are members of the two types of
 * pattern-name groups that match to them.
 * To deliver some rights to a given source 'tom' whatever its current host,
 * a pattern-name accounts-for-hosts group 'tom§@%' is required.
 * This form is proper to Access Road, since MySQL would rather uses 'tom@%'.<P>
 * For a new account 'xxxx', an empty non-conditional ACL on 'MySQL Server'
 * for the accounts-for-hosts group 'xxxx§@%' is set without rights, if it does not exist.
 * This class implements the following rules, and informs the Access Road user
 * with the same sentence as error message when they are not fulfilled with: <P>
 * - The host of an account cannot be null. <P>
 * - The name of an account cannot contain the characters '%', '_', '§' or '@'.
 * - In an ACL, the condition group is the account host. <P>
 * - User names can be up to 16 characters long, but no more. <P>
 * - The host of an account has to be a simple host. <P>
 * - The condition group must have a simple host type, that is not be a pattern-name group. <P>
 * - The host name is not the name of an open ACS. <P>
 * - The name xxx is not the name of an account in the open ACS. <P>
 * For the account groups, the sorting is 'host',
 * 'pattern-name_accounts-for-hosts group', then
 * the sorting among the accounts-for-hosts groups from their names. <P>
 * This class has no subclass in this package. ProxyMySQLImpl handles
 * the proxy users, and it is derived from this class.
 * It is a javabean with the following bound properties: 'CurrentGroupID'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive the new value, without copy.
 * All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * @version 0.7.0
 */
public class UserIDMySQLImpl extends UserIDImpl implements UserIDMySQL {
  // private variables
  private static final long   serialVersionUID__ = 0;
  // protected variables
  /*
   * The current host of this MySQL user. This account is a host member.
   */
  protected GroupIDMySQLImpl  host_ = null;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final UserIDMySQLImpl EMPTY_INSTANCE = new UserIDMySQLImpl();

  /**
   * variable for the JDK 2 serialization
   * @serial ("host_", GroupIDMySQLImpl.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("host_", GroupIDMySQLImpl.class)
  };

 /**
  * The basic constructor
  */
  public UserIDMySQLImpl() {


                          // for testing purpose
                     //System.out.println("Call to UserIDMySQLImpl constructor");
  }

   /**
 * Constructor used in the beamer, from the 'New' button, to create an account.<P>
 * A host in MySQL is simulated by a GroupID having a 'host' type, with the
 * special case of the group 'localhost' that has the type 'host byAcsCreator'
 * to forbid its deleting by the Access Road user.
 * The host to use, if it is not 'localhost', has to be an open ACS having
 * an account with the new userID name as name,
 * and the user 'anonymous' is the single exception to this rule. <P>
 * This constructor ensures that the new account is member of the pattern-name
 * accounts-for-hosts groups that match to it, if it is not the user 'anonymous'.
 * For a new account 'user', an empty non-conditional ACL on 'MySQL Server' is set
 * without rights for the accounts-for-hosts group 'user§@%', if it does not exist.
 * Calls setValueOf(). Calls the super constructor.
 * @param _name ImmutableName of the userID as a kind of eligible party
 * @param _acs ACS owning the userID
 * @param _epType is the eligible party type, and may be null if the ACS allows it.
 * @param _isAdmin true only if the instance has an administrative role in the ACS,
 *    for the technical administration or the users
 * @param _host mandatory host for an account. Never null.
 * @exception CreateError if ACS.newEorEligibleParty() throws it - the instance
 * must then be deleted !
 */
  public UserIDMySQLImpl(ImmutableName _name, ACS _acs, String _epType, Boolean _isAdmin,
          GroupIDMySQLImpl _host) throws CreateError  {

        super(_name, _acs, _epType, _isAdmin);

        if (_host == null)
            throw new CreateError(" The host of an account cannot be null. ");

          // for testing purpose
          //System.out.println("call to UserIDMySQLImpl constructor with 2 arguments, for "+_name);

        try {
            setValueOf(this, _host);
        }
        catch (CreateError c) {
            try {
                finalizeForBase();
            }
            catch (UpDateError u) {
                throw new CreateError("From UserIDMySQLImpl constructor "+c.toString()+
                        ". Error in finalizing "+u.toString());
            }

            throw c;
        } // catch c

  } // UserIDMySQLImpl()

   /**
   * Currently unused.
   * Calls the super constructor and setValueOf().
   * @param _name ImmutableName of the eligible party to create
   * @param _acs Main ACS of the eligible party to create
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users
   * @param _host mandatory host for an account. Never null.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if ACS.newEorEligibleParty() throws it -
   *  the instance must then be deleted !
   */
  public UserIDMySQLImpl(ImmutableName _name, ACS _acs, String _epType, boolean _isAdmin,
    GroupIDMySQLImpl _host, boolean _toRegisterInManager) throws CreateError {

        super(_name, _acs, _epType, _toRegisterInManager);

        if (_host == null)
            throw new CreateError(" The first host of the account cannot be null. ");

              // for testing purpose
              //System.out.println("call to UserIDMySQLImpl constructor with 3 arguments, for "+_name);

        try {
            setValueOf(this, _host);
        }
        catch (CreateError c) {
            try {
                finalizeForBase();
            }
            catch (UpDateError u) {
                throw new CreateError("From UserIDMySQLImpl constructor "+c.toString()+
                        ". Error in finalizing "+u.toString());
            }
            // for an user message
            throw c;
        } // catch u

  } // UserIDMySQLImpl()

  //
  // To implements the UserIDImpl methods
  //

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon.Returns always null,
   * because these events are listened only to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject has
   * one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()
  
   /**
   * Adds a group for which this GroupIDMember is a member,
   * and sorts the groups in an sortered array that follows the sorting order
   * for the MySQL AclEntries, from the more-specific source to the least-specific ones.
   * There are, for the users, the groups 'host' then 'pattern-name_accounts group',
   * with an order among the 'pattern-name_accounts groups'.
   * Overrides the GroupIDMemberImpl method and does not call it.
   * Does not calls setEorPrimGroup(_grp) like in UserIDImpl, since there is no primary group.
   * Fires changes on 'AssociatedGroups'. Used only by GroupIDImpl.
   * Calls BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder().
   * @param _grp is the groupID to add. No action if it is null, or with null name.
   */
  public synchronized void addGroupToGM(GroupID _grp) {
    // variables
    boolean recorded = false;
    GroupIDImpl[] newArray;
    GroupIDImpl[] l_oldGroups;
    String typeOne = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(_grp.getEPType());

    if (_grp == null) return;

    // test if _grp is not known
    for (int i = 0 ; i < l_Groups_.length ; i++) {
        if (_grp == l_Groups_[i] ) {
          recorded = true; //_grp is known
          break;
        }
    } // for i
    if (recorded) return;

                            // for testing and tracking
                            //System.out.println("in UserIDMySQLImpl.addGroupToGM(), for "+getName()+
                              //      ", l_Groups_ before the adding:");
                              //for (int i = 0 ; i < l_Groups_.length ; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("-- at "+i+": "+l_Groups_[i].getName());
                                  //else
                                    //  System.out.println("-- at "+i+": NULL");
                              //} // for

    // saves the old Group list, since l_Groups_ is a bound property
    l_oldGroups = new GroupIDImpl[l_Groups_.length];
    System.arraycopy(l_Groups_, 0, l_oldGroups, 0, l_Groups_.length);

        // adds at the right order
    newArray = new GroupIDImpl[l_Groups_.length + 1];
    if (l_Groups_.length != 0) {
        for (int i = 0 ; i < l_Groups_.length ; i++) {
            if (!recorded &&
                    BaseUtilityMySQLImpl.isAfterInTheGroupIDMemberOrder(
                        typeOne,
                        UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(l_Groups_[i].getEPType()),
                        _grp.getName().lastComponent(),
                        l_Groups_[i].getName().lastComponent())) {
                newArray[i] = l_Groups_[i];
                continue;
            } // if

            if (!recorded) {
                newArray[i] = (GroupIDImpl)_grp;
                recorded = true;
            } // if

            newArray[i + 1] = l_Groups_[i];
        } // for i

        if (!recorded)
            newArray[l_Groups_.length] = (GroupIDImpl)_grp;
    }
    else {
        newArray[0] = (GroupIDImpl)_grp;
    } // else
    l_Groups_ = newArray;

                            // for testing and tracking
                            //System.out.println("in UserIDMySQLImpl.addGroupToGM(), for "+getName()+
                              //      ", l_Groups_ after the adding:");
                              //for (int i = 0 ; i < l_Groups_.length ; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("-- at "+i+": "+l_Groups_[i].getName());
                                  //else
                                    //  System.out.println("-- at "+i+": NULL");
                              //} // for

    try {
        firePropertyChange("AssociatedGroups", l_oldGroups, getEorL_GroupID());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDMySQLImpl.addGroupToGM(), firePropertChange() throws");
        e.printStackTrace();
    }

    newArray = null;
    l_oldGroups = null;
  } // addGroupToGM()

   /**
   * This method removes this instance in its groups directly.
   * The superclass cannot be used, since it is done through a call to
   * GroupIDMySQLImpl.removeEorMember(), which may fire an exception.
   * Calls GroupIDMySQLImpl.removeEorMemberWithoutControl().
   * Overrides the super method and calls it at the end.
   * This method HAS TO BE USED before deleting the UserID.
   * Note: when this method is called by the EligiblePartyImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the instance cannot be removed from its first host group
   */
   public synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("UserIDImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (folderMember_ == null) {
      return;
    }

    // set l_Groups_ to an empty array
    List<GroupIDMySQLImpl> l_grps = null;
    try {
        if (l_Groups_ != null) { // false if called by EligiblePartyImpl constructor
            l_grps = new ArrayList(Arrays.asList(l_Groups_));

            for (GroupIDMySQLImpl grp : l_grps) {
                // calls back removeGroupToGM() and updates l_Groups_
                grp.removeEorMemberWithoutControl(this);
            } // for i
        } // if
    }
    catch (CreateError c) {

              // for testing purpose
              //System.out.println("UserIDMySQLImpl.finalizeForBase() with exception ");

        throw new UpDateError("UserIDMySQLImpl.finalizeForBase() "+c.toString());
    }
    
    l_grps = null;
    host_ = null;
    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'CurrentGroupID'.
  * Overrides the super method and calls it.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      if (_propertyName.equals("CurrentGroupID")  )
          changeSupport_.addPropertyChangeListener(_propertyName, _l);
      else super.addPropertyChangeListener(_propertyName, _l);

  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties: 'CurrentGroupID'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("UserIDMySQLImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("CurrentGroupID") )
          changeSupport_.removePropertyChangeListener(_propertyName, _l);
      else super.removePropertyChangeListener(_propertyName, _l);

  } // removePropertyChangeListener()

  // --------------- UserIDMySQL public methods ------------------------------

  /**
   * @return the immutable first host of this MySQL user. May be null only at the creation
   * or if finalized.
   */
  public GroupIDMySQLImpl getEorFirstHost() {
          return host_;
  }

  /**
  * This method provides a copy of the instance with another name.
  * Do not copy the own resources or the current actors.
  * The copy order is 3 times the source order.
  * Caution: do not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of UserIDImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager)
    throws CreateError {

      if (_acs == null || _name == null  ) {
          throw new CreateError("in UserIDMySQLImpl.copy(),"+
          " _name or _acs is null.");
      }

      UserIDMySQLImpl result = new UserIDMySQLImpl(
          _name, _acs, getEPType(), false, host_,
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);

      return result;
  } // copy(,,,)

  /**
  * This method provides a copy of the instance
  * as copy() but without copying the internal AclEntries.
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of UserIDImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in UserIDMySQLImpl.copy(),"+
          " _name or _acs is null.");
      }

      UserIDMySQLImpl result = new UserIDMySQLImpl(
          _name, _acs, getEPType(), false, host_,
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclEntry copy

      //if (result.getEorL_AclConnectedResources().isEmpty()) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copy(,,,)

  // ------------------ protected methods -----------------------------------------
  
  /**
  * This method provides a copy of the instance variables.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal AclEntries
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError 
  **/
  protected UserIDMySQLImpl copyOwnVariables(UserIDMySQLImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in UserIDMySQLImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      _result = (UserIDMySQLImpl)super.copyOwnVariables(_result, _withoutInternalAcl);

      // todo: set variables
      
      return _result;
  } // copyOwnVariables(,,,)

/**
 * Sets, for the constructors, a new account as member of the matched pattern-name groups
 * and member of its mandatory host.
 * A host in MySQL is simulated by a GroupID having a 'host' type, with the
 * special case of the group 'localhost' that has the type 'host byAcsCreator'
 * to forbid its deleting by the Access Road user.
 * The host to use, if it is not 'localhost', has to be an open ACS having
 * an account with the new UserID name as name,
 * and the user 'anonymous' is the single exception to this rule. <P>
 * This constructor ensures that the new account is member of the pattern-name
 * accounts-for-hosts groups that match to it, if it is not the user 'anonymous'.
 * For a new account 'user', an empty non-conditional ACL on 'MySQL Server' is set
 * without rights for the accounts-for-hosts group 'user§@%', if it does not exist.
 * Called by the constructors.
 * Calls GroupIDMySQLImpl.isMemberOfMySQLPatternNameGroup() and addEorMemberToPatternNameGroup().
 * @param _user new instance of this class
 * @param _host host group is mandatory. Never null.
 * @exception CreateError if the constraints are not fulfilled with
 */
  protected static void setValueOf(UserIDMySQLImpl _user, GroupIDMySQLImpl _host)
          throws CreateError {

    if (_user.getName().lastComponent().indexOf("%") != -1 ||
            _user.getName().lastComponent().indexOf("_") != -1 ||
            _user.getName().lastComponent().indexOf("§") != -1 ||
            _user.getName().lastComponent().indexOf("@") != -1) {
        throw new CreateError(" The name of a MySQL account cannot contain "+
                "one of the characters '%','_', '§' or '@'."+
                " In an ACL, the condition group is the account host.");
    } // if
    if (_user.getName().lastComponent().length() > 16) {
        throw new CreateError(" User names can be up to 16 characters long, but no more.");
    } // if
    if (_host.getType().indexOf("host") == -1 ||
            _host.getType().indexOf("hosts group") != -1 ||
            _host.getType().indexOf("standard") != -1)
        throw new CreateError(" The host of an account has to be a simple host. ");

    List<ImmutableEligibleParty>  m_grps = new ArrayList(
                                        _user.getEorACS().getEorM_OwnEligibleParties().values());

    String                    grpType = null;
    ImmutableEligibleParty    ep = null;

    // host is mandatory. May throws CreateError.
    // resets the non-conditional and conditional ACLs in the ACS, if necessary.
    _host.addEorMember(_user);
    _user.host_ = _host;

    // For a new account 'user', an empty non-conditional ACL on 'MySQL Server' is set
    // without rights for the accounts-for-hosts group 'user§@%', if it does not exist.
    String name = null;
    DirectoryMySQL serv = null;
    if (_user.getNickName().matches(".*\\(.\\)"))
        name = _user.getNickName().substring(0, _user.getNickName().length() - 3);
    else
        name = _user.getNickName();

    GroupIDMySQLImpl afh = (GroupIDMySQLImpl) _user.aCS_.getEorM_EligibleParties().
            get(((Name)_user.aCS_.getName()).extended("<G>").extended(name+"§@%"));

    if (afh == null) {
        afh = new GroupIDMySQLImpl(
                ((Name)_user.aCS_.getName()).extended(name+"§@%"),
                (ACS)_user.aCS_, "pattern-name_accounts-for-hosts group", false, true);

        serv= (DirectoryMySQL) _user.aCS_.getEorM_Resources().
                get(((Name)_user.aCS_.getName()).extended("MySQL server"));
        serv.addAclEntry(_user.aCS_, afh, null, Boolean.TRUE, null);
    } // if

    // adds _user as member of the accounts-for-hosts groups, including 'user§@%'
    if (!_user.getName().lastComponent().equals("anonymous")) {
        for (Iterator<ImmutableEligibleParty> ite = m_grps.iterator() ; ite.hasNext() ; ) {
          ep = ite.next();

          if (!(ep instanceof ImmutableGroupID)) {
              ite.remove();
              continue;
          } // if

          grpType = ((ImmutableGroupID)ep).getEPType();
          if (grpType.indexOf("standard") != -1) {
                ite.remove();
                continue;
          }
          if (grpType.indexOf("pattern-name_accounts-for-hosts group") == -1 )
            ite.remove();
        } // for ite

        // add _user to the relevant accounts-for-hosts group(s)
        try {
          for (Object obj : m_grps) {
            if (GroupIDMySQLImpl.isMemberOfMySQLPatternNameGroup(_user, (GroupID)obj))
              ((GroupIDMySQLImpl)obj).addEorMemberToPatternNameGroup(_user);
          } // for obj
        }
        catch (BaseError b) {
          throw new CreateError(b.toString());
        }
    } // if (!_user.getName().lastComponent().equals("anonymous")

    m_grps = null;
    grpType = null;
    ep = null;
    name = null;
    afh = null;
    serv = null;
  } // setValueOf()

  /**
   * Adds a proxy account as source of a PROXY right on this instance.
   * Caution: the standard way of use for the soft aliasing is not followed there,
   * since the MySQL ACS does not manage this feature! The name of this method
   * is quite confusing. ImmutableSource.getL_AcsOfSoftReferences() is not workable
   * on this instance. The property 'SoftAlias' is only used to display in the beamer
   * the proxy accounts of an UserIDMySQL, using ImmutableSource.getL_SoftAlias().<P>
   * Fires the events 'SoftAlias' and 'hasSoftAlias'.
   * @param _proxy No action if null, or already known.
   * @exception UpDateError the type is immutable
   */
  protected void addEorSoftAlias(ImmutableName _proxy) throws UpDateError {
      super.addEorSoftAlias(_proxy);
  }

  /**
   * Removes a proxy account as source of a PROXY right on this instance.
   * Caution: the standard way of use for the soft aliasing is not followed there,
   * since the MySQL ACS does not manage this feature! The name of this method
   * is quite confusing. ImmutableSource.getL_AcsOfSoftReferences() is not workable
   * on this instance. The property 'SoftAlias' is only used to display in the beamer
   * the proxy accounts of an UserIDMySQL, using ImmutableSource.getL_SoftAlias().<P>
   * Fires the 'SoftAlias' and the 'hasSoftAlias' property change events.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the list is null.
   * @param _alias is a registered alias. No action if null.
   * @exception UpDateError if _proxy is not known.
   */
  protected void removeEorSoftAlias(ImmutableName _alias) throws UpDateError {
      super.removeEorSoftAlias(_alias);
  }

} // UserIDMySQLImpl

