//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

import java.util.Map;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
//import ARoad0.ErrB.BaseError;

//import java.io.ObjectStreamField;

/**
* This class is responsible to simulate a MySQL view,
* adding a view delegate under the database node in the resources tree.
* This class controls that the VirtuelFolder elements are only from the right database.
* The DEFINER mode is the default one for the MySQL views. It creates a VirtualFolder
* with a couple UserID/GroupID as parameter of the constructor.
* This UserID becomes the Account of the AG context for the view delegate.
* The GroupID is the host of the account.
* This feature implies the creation of a VirtualFolderMySQLImpl class in the AcsAddon
* gBase package, as subclass of VirtualFolderImpl in the generic gBase package.  <P>
* The view definition is 'frozen' at creation time in MySQL, but it is not useful
* to extend this constraint to the view simulation in Access Road.
* As actor, the view delegate in the resources tree runs under the couple UserID/GroupID
* passed on to the VirtualFolder constructor of the view, if any. This is also
* the responsibility of the VirtualFolderMySQLImpl constructor. <P>
* It is a javabean with the bound properties of the superclass. <P>
* This class has no subclass in this package,and no variables.
* Note: no equals() nor clone() methods.
* See the <a href="{@docRoot}/copyright.html">Copyright</a>.
*/
public class VirtualFolderMySQLImpl extends VirtualFolderImpl implements VirtualFolderMySQL {

  // ------------- public attributes -----------------------------------------
  /*
   * Used to create a new instance in CollectionPropertyEditor from the BeanInfo description
   * for ACSs and VirtualFolderMySQLs
   */
  public static final VirtualFolderMySQLImpl EMPTY_INSTANCE = new VirtualFolderMySQLImpl();

 /**
  * Default constructor
  */
  public VirtualFolderMySQLImpl() {
    super();
  }

  /**
   * This is the main constructor.
   * The arguments _definer and _host are not properties of this class, but
   * they are used to create the view delegate, as alias of this virtual folder.
   * A view must have 'view' at the end of its name.
   * The property aliasHandler_ is intialized in this constructor.
   * Calls the superclass constructor and setEorAliasAsInternalHardReference().<P>
   * With a definer and a host, the user message errors are: <P>
   * - The definer ACS is not the ACS of this virtual folder. <P>
   * - The host cannot be null when the definer is not null. <P>
   * - The host ACS is not the ACS of this virtual folder. <P>
   * - The host is not a simple host. <P>
   * - The Host is not the simple host of the definer.
   * @param _name name of the folder name. Its last component is added to the _acs name
   * to get the folder name. It is not added to the parent name, since this is
   * the job of valueOf(). Never null.
   * @param _acs is the main ACS of the folder to create. Never null.
   * @param _parent is the folder which contains this folder, and with the same type.
   * May be null or immutable. The last component has the form 'database_name views'
   * if the new VirtualFolder is a view.
   * @param _type is the folder type. Used like the Resource type. If null, the internal value
   *  is "<undefined>". Not the main type, get through getMainVFType.
   * @param _definer not null to set the view delegate as running under this account.
   * May be null.
  * @param _host the host of _definer, from the same ACS. May be null.
  * Never null if _definer is not null.
  * @exception CreateError if _m_BaseObjects contains a null,
  *     no ACSObject, empty-name or not _acs object, or with double link in
  *     the folders tree for a ACS, if _parent has not the same folder type,
  *     if the name is the name of a known VirtualFolder, Resource or EligibleParty,
  *     or if the object type is not managed by the ACS.
   */
  public VirtualFolderMySQLImpl(ImmutableName _name, ACS _acs, VirtualFolderMySQL _parent,
          String _type, UserID _definer, GroupIDMySQL _host)
          throws CreateError {

      // it is unuseful to register in the ACS at the end of this constructor
      super(_name, _acs,  _parent, _type, false, true);

                          // for testing purpose
                          //System.out.println("in VirtualFolderMySQLImpl.constructor, "+
                            //                    "current EP VirtualFolders:" );
                          //for (VirtualFolder vf : _acs.getEorL_EPVirtualFolders()) {
                            //  if (vf != null)
                              //    System.out.println("---- "+vf.getNickName());
                              //else
                                //  System.out.println("---- null VirtualFolder");
                          //}


      if (_parent == null) {
          if (_acs.getEorL_EPVirtualFolders().size() != 1) {

              try {
                  finalizeForBase();
              } catch (UpDateError r) {
                throw new CreateError(" VirtualFolderMySQLImpl constructor. "+
                      "The views tree must have only one root."+
                      " Error on finalizing "+r.toString());
              } // catch

              throw new CreateError(
                      " The views tree must have only one root.");
          } // if (!_acs.getEorL_EPVirtualFolders().isEmpty())

          // for the '<immutable> no-child node' type
          return;
      } // if (_parent == null)

      // for the 'root for a set' type
      if (_type.equals("<immutable> root for a set")) return;

      // A view must have 'view' at the end of its name
      if (_type.indexOf("view") != -1 &&
                !_name.lastComponent().endsWith("view") ) {
          try {
              finalizeForBase();
          }
          catch (UpDateError u) {
              throw new CreateError(" From VirtualFolderMySQLImpl constructor. "+
                  "error on the view name "+
                  ". Error on finalizing "+u.toString());
          }

          throw new CreateError(" A view must have 'view' at the end of its name.");
      } // if

      // creates the view delegate under the database Directory
                          // remove ' views' in the parent last component
      if (_definer != null) {
          if (_definer.getEorACS() != aCS_) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From VirtualFolderMySQLImpl constructor. "+
                      "error on the Definer ACS "+
                      ". Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Definer ACS is not the ACS of this virtual folder.");
          }
          if (_host == null) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From VirtualFolderMySQLImpl constructor. "+
                      "error on the null Host "+
                      ". Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Host cannot be null when the Definer is not null.");
          }
          if (_host.getEorACS() != aCS_) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From VirtualFolderMySQLImpl constructor. "+
                      "error on the Host ACS "+
                      ". Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Host ACS is not the ACS of this virtual folder.");
          }
          if (!_host.isSimpleHost()) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError(" From VirtualFolderMySQLImpl constructor. "+
                      "error on the no-simple Host "+
                      ". Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Host is not a simple host.");
          }
          if (((UserIDMySQL)_definer).getEorFirstHost() != _host) {
              try {
                  finalizeForBase();
              }
              catch (UpDateError u) {
                  throw new CreateError("From VirtualFolderMySQLImpl constructor. "+
                      "error on the simple Host "+
                      ". Error on finalizing "+u.toString());
              }

              throw new CreateError(" The Host is not the simple host of the definer.");
          }
      } // if (_definer != null)

      Name baseName = ((Name)_acs.getName()).extended("MySQL server").extended(
                    _parent.getName().lastComponent().substring(0,
                        _parent.getName().lastComponent().length() - 6));

      if (_acs.getEorM_Resources().get(baseName) == null) {

                            // for testing purpose
                            //System.out.println("");
                            //System.out.println("in VirtualFolderMySQLImpl.constructor, "+
                              //                  "unknown parent in the ACS - parent name "+
                                //                baseName);
                            //System.out.println("");
                            
              baseName = null;
              return;
      } // if

      Actor delegate = new ActorMySQLImpl(
              // adds ' alias'
              ((Name)baseName.clone()).extended(_name.lastComponent() + " alias"),
              _acs,
             (Directory) _acs.getEorM_Resources().get(baseName),
              _definer, "view_delegate <byAcsCreator>", _host);
      delegate.setComment("Alias of the view, and target of the view AclEntries");
   
      try {
          // set the delegate as alias of this view
          setEorAliasAsInternalHardReference(delegate);

      } catch (UpDateError u) {
          try {
              finalizeForBase();
          } catch (UpDateError r) {
            throw new CreateError(" VirtualFolderMySQLImpl constructor. "+
                  "Wrong handling of hard alias and current UserID. "+
                  "Error on finalizing "+
                  r.toString());
          } // catch

          throw new CreateError(
                  " Wrong handling of hard alias and current UserID "+
                  u.toString());
      } // UpDateError

                                // for testing and tracking
                              //System.out.println("end of the VirtualFolderMySQLImpl.constructor for "+
                                //            getName()+" with the BaseObjects: ");
                              //if (m_BaseObjects_ != null) {
                                //  for (Iterator ite = m_BaseObjects_.entrySet().iterator(); ite.hasNext();) {
                                  //          System.out.println("------- "+
                                    //                ((Map.Entry)ite.next()).getKey() );
                                  //} // for ite
                              //}
                              //else {
                                //  System.out.println("----- null map ");
                              //}


      baseName = null;
      delegate = null;
  } // constructor

  /**
   * No use, since CollectionPropertyEditor calls BaseObjectHandler.
   * Calls the main constructor.
   * Caution: it has to not be used as a Resource name in the same ACS,
   * because linked privileges have resources and folders as targets.
   * @param _name of the view. It is copied. 
   * @param _acs is the main ACS of the folder to create. Never null.
   * @param _parent is the folder which contains this folder, and with the same type;
   *   may be null.
   * @param _type is the folder type. If null, the internal value is "<undefined>".
   * Used for some typed privileges. The main type is get through getMainVFType.
   * @param _definer not null to set the view delegate as running under this account.
   * May be null.
   * @param _host the host of _definer. May be null. Never null if _definer is not null.
   * @exception CreateError if _acs is null, or from the constructor.
   */
  public static VirtualFolderMySQL valueOf(ImmutableName _name, ACS _acs,
          VirtualFolderMySQL _parent, String _type,
          UserID _definer, GroupIDMySQL _host) throws CreateError {

                          // for testing purpose
                          //System.out.println("call to VirtualFolderMySQLImpl.valueOf() ");

      if (_acs == null) throw new CreateError(" VirtualFolderImpl constructor "+
              "with a null ACS");

      return new VirtualFolderMySQLImpl(_name, _acs, _parent, _type, _definer, _host);

  } // valueOf()

  /**
   * In the MySQL DEFINER mode, the view delegate runs under an account.
   * @return the account under which the view delegate runs. Null when there is no delegate.
   */
  public ImmutableUserID getEorDefinerForDelegate() {
      if (getEorAliasAsInternalHardReference() == null)
          return null;
      else
        return ((Actor)getEorAliasAsInternalHardReference()).getEorCurrentUserID();
  }

  /**
   * In the MySQL DEFINER mode, the view delegate runs under an account and a current host.
   * @return the host under which the view delegate runs. Null when there is no delegate.
   */
  public GroupIDMySQL getEorDefinerHostForDelegate() {
      if (getEorAliasAsInternalHardReference() == null)
          return null;
      else
        return (GroupIDMySQL)((Actor)getEorAliasAsInternalHardReference()).getEorCurrentGroupID();
  }

  /**
   * Adds a member to this virtual folder.
   * Overrides the super method to control that the VirtualFolder elements are
   * only from the right database, then calls the super method.
   * @param _obj non-null ACSObject to add in the folder
   * @exception CreateError if the object is already known by the view,
   *   if it is known in the tree when it is restricted to one instance,
   *   or it is not a resource or an actor (following the main type),
   *   or if its ACS is not the folder ACS, or the object type is not authorized,
   *   or if it is not an object from the view database.
   */
  public synchronized void addEorBaseObject(BaseObject _obj) throws CreateError {

      // replace in the ACS the policy "VirtualFolder.TypesOfMemberFor.no-child node"
      // to fire the right exception before 'The new element does not belong to the view database.'
    if ("no-child node".equals(UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(type_)) ||
        "root for a set".equals(UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(type_)))
        throw new CreateError(" This is not a true view. It is just a passive node without elements.");

    Name baseName = ((Name)aCS_.getName()).extended("MySQL server").extended(
                    parent_.getName().lastComponent().substring(0,
                        parent_.getName().lastComponent().length() - 6));
      
    if (!_obj.getName().contains(baseName)) {
        getEorBaseError().addException(" in VirtualFolderMySQLImpl.addEorBaseObjects()"+
          "the argument is not from the view database.");
        throw new CreateError(" The new element does not belong to the view database.");
    }

    super.addEorBaseObject(_obj);

    baseName = null;
  } // addEorBaseObject()

  /**
   * Gets the specific events to listen from an ACSObject, because these events
   * should change the rights of an ACSObject in an ACS Addon.
   * These events are listened to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject has
   * one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.UtilityImpl.removeAddonParentPropertyChangeListener().
   * Overridden by the subclass DirectoryMySQLImpl.
   * @return map to define the specific ACS Addon events to listen. Is always null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    //Map<String, String> m_result = new HashMap();
    //m_result.put(
      //      "<THIS>",
        //    "<THIS>.getL_otherRights");
    return null;
  } // getAddonInnerPropertiesToListen()

  /**
  * Finalizes the view delegate if not null.
  * Called by finalizeForUser() and by the associations manager
  * when the instance is the root of the database views.
  * Calls superclass finalizeForBase().
  * @exception UpDateError from Basic.removeInVFolderList()
  */
  public synchronized void finalizeForBase() throws UpDateError {
    // noop if it is the second call to this method
    if (nodeHandler_ == null) {
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");

      return;
    }

    if (parent_ != null) {
        // finalizes the view delegate under the database Directory
        Name baseName = ((Name)aCS_.getName()).extended("MySQL server").extended(
                    parent_.getName().lastComponent());
        Actor delegate = (Actor) aCS_.getEorM_Resources().get(
              baseName.extended(name_.lastComponent()));
        // null delegate for a '<immutable> root for a set' or
        // '<immutable> no-child node' type VF
        if (delegate != null)
            delegate.finalizeForBase();
        baseName = null;
        delegate = null;
    }

    super.finalizeForBase();
  } // finalizeForBase()
  
  /**
  * This method provides a copy of the instance
  * with another name, without children copying
  * because the copy order will not be always correct.
  * @param _name is the copy name. Caution: it has to not be used
  *   as a Resource name in the same ACS, because linked privileges have
  *    resources and folders as targets.
  * @param _acs is the main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is the VirtualFolderMySQL (not Node!) which belongs to _acs
  *   and will have the copy. May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return Object is a copy as VirtualFolderMySQLImpl instance
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent is not a virtual folder, or _parent ACS is not _acs.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Node _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null || 
        (_parent != null && ((VirtualFolder)_parent).getEorACS() != _acs) ||
        !(_parent instanceof VirtualFolder)) {
          throw new CreateError(" in VirtualFolderImpl.copy(),"+
          " _name or _acs is null,"+
          " or _parent is not a virtual folder, or _parent ACS is not _acs.");
      }
      
      boolean forRes;
      if (VFType_.equals(RES_VFOLDER) ) forRes = true;
      else forRes = false;
      
      VirtualFolderImpl result = new VirtualFolderImpl(
          _name, _acs, (VirtualFolder)_parent, type_,
          forRes);

      // Todo: adds the base objects
      
      return result;
  } // copy()
  
  /**
   * UNUSED -
  * This method provides a copy of the instance variables.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy).
  * Copy the privileges only if it is possible.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @return the updated _result
  * @exception CreateError from
  **/
  protected VirtualFolderImpl copyOwnVariables(VirtualFolderMySQLImpl _result) throws CreateError {

      if (_result == null) {
          throw new CreateError(" in VirtualFolderMySQLImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      //
      // TODO: control the values from the ACS
      //

      return _result;
  } // copyOwnVariables(,,,)

} // VirtualFolderMySQLImpl
