//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface;

import ARoad0.gBaseInterface.*;

import java.util.Map;

/**
 * This interface is responsible for modeling a MySQL Directory to
 * manage the inherited AclEntries in the Directory children. <P>
 * To implement the MySQL rights selection algorithm, the inherited AclEntries feature
 * allows to put in each resource, the rights of its direct and indirect parents.
 * For instance, a column inherits the AclEntries from its table and from its base.
 * The core access path algorithms, as usual, search if a given account has
 * some AclEntries to the column, for instance, and what are the rights.
 * The updatings of the inherited AclEntries in the MySQL resource is
 * the responsibility of the resource parent. The core Access Road functions do not include
 * the management of the inherited AclEntries. This is the first reason to define
 * a MySQL AcsAddon. These rights are managed in the class DirectoryMySQLImpl.
 * BasesSetMySQL is not a DirectoryMySQL nor a ResourceMySQL.
 * Note: PrivilegeForLinks has not been choosen in place of AclEntry to simulate
 * the MySQL privilege from an account to a resource. The main justification is
 * that a conditional AclEntry allows the Access Road user to see and edit
 * more easily the rights for a given couple (account, host). <P>
 * @version 0.7.0
 */
public interface DirectoryMySQL extends Directory, ResourceMySQL {

  /**
    * Gets the map of the global rights from the root 'MySQL server'.
    * The map has lists of rights as values, and strongly contrained Name keys.
    * The rights may be from a real ACL or from a virtual ACL.
    * The AclEntries list is used to produce some processed rights, as virtual ACL into the map.
    * Only the real ACL are displayed by the beamer, as conditional or non-conditional ACL.
    * <P>
    * The global rights are necessary for the gWork classes, and the first user is
    * NodeRightsMySQLImpl.getConnectionPair().
    * To have some access rights on a MySQL Resource, a pair (user, host) has
    * to match on the root 'MySQL Server' a pair (user, host) or a pair
    * (user, accounts-for-hosts group) which has at least an USAGE right.
    * The FIRST matching pair in the 'MySQL server' list becomes the connection pair.
    * It is used to get the rights for the initial pair (user, host) at each parent level.
    * If the connection pair has only an USAGE right on the root and if, for a given Resource,
    * the connection pair has no other inherited ACL on the target, then no right is granted. <P>
    * Into the global ACL rights map,
    * the map key may have varied forms, where the parent order is always at the end.
    * The parent order tells the distance from the target to the parent which delivers
    * the referenced rights. For instance, the direct parent of a resource has the string '1'.
    * The direct parent of this parent has the string '2' as parent order, and so on.
    * The two key forms are listed hereinafter, where the parent order is always 0: <P>
    * - (user name @ host name last component) ::  parent order,
    * for instance 'one :: mysql :: jerry@localhost :: 0',<P>
    * - <Proc>:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '<Proc>:: one :: mysql :: jerry@192.168.1.% :: 0',<P>
    * '<Proc>' is a generic tag that means 'processed rights', to tell the key and
    * its rights are not get from the processing of a real ACL.
    * A key has exactly one origin, which is one real ACL.
    * The value in a rights map is always a non-null StringRight array.
    * @return the map of global rights from the root direct AclEntries.
    * This is a direct reference without protection, to be fast.
    * May be empty but never null.
    */
  public Map<ImmutableName, StringRight[]> getM_GlobalTargetACLRights();

} // DirectoryMySQL
