//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface;

import ARoad0.ErrB.UpDateError;

import ARoad0.gBaseInterface.*;

import java.util.Map;

/**
 * This interface is responsible for modeling a MySQL resource.
 * The Resource method getL_aclEntry is not overridden, but it provides
 * a sorted list of ACL from a ResourceMySQL.
 * @version 0.7.0
 */
public interface ResourceMySQL extends Resource, ACSObjectAddon {

    /**
    * Adds the inherited AclEntry. An inherited AclEntry may have a Resource
    * which is not this instance.
    * @param _acl inherited AclEntry to put in the list at the right order
    */
  public void addInheritedAclEntry(AclEntry _acl);
  
    /**
    * Removes the inherited AclEntries in the children.
    * @param _acl inherited AclEntry
    * @exception UpDateError if _acl is null or unknown
    */
  public void removeInheritedAclEntry(AclEntry _acl) throws UpDateError;

    /**
    * Resets the inherited AclEntry rights after an user edition of the ACL rights.
    * @param _acl inherited AclEntry with updated rights
    */
  public void resetInheritedAclEntryRights(AclEntry _acl);

  /**
    * Gets the internal map of the primary rights from each parent level of this resource.
    * The primary rights are displayed into the beamer. They help the user
    * to understand the current users, since it covers the root 'MySQL server' rights,
    * and to understand the complex database rights. Because this method overrides
    * the Resource method, its name is 'TargetInheritedACL' rather than 'TargetPrimaryACL'.
    * The map has lists of rights as values, and strongly contrained String keys.
    * The rights may be from a real ACL or from a virtual ACL.
    * The AclEntries list is used to produce some processed rights, as virtual ACL into the map.
    * Only the real ACL are displayed by the beamer, as conditional or non-conditional ACL.
    * A real ACL in the inherited AclEntries list produces rights directly into the map. 
    * <P>
    * The primary rights are not an official MySQL concept. It is an help
    * for the Access Road user and it is necessary for the gWork classes.
    * It describes both the inherited AclEntries and the direct AclEntries.
    * The root 'MySQL server' has primary rights only from its direct ACL,
    * and a database has primary rights from the root and its direct ACL.
    * <P>
    * To have some access rights on a MySQL Resource, a pair (user, host) has
    * to match on the root 'MySQL Server' a pair (user, host) or a pair 
    * (user, accounts-for-hosts group) which has at least an USAGE right.
    * The FIRST matching pair in the 'MySQL server' list becomes the current user.
    * It is used to get the rights for the initial pair (user, host) at each parent level.
    * The current user is returned by the method NodeRightsMySQLImpl.getConnectionPair().
    * If the current user has only an USAGE right and if, for a given Resource,
    * the current user has no direct ACL and no inherited ACL, then no right is granted.
    * Otherwise, the rights are the sum of the current user
    * rights from the FIRST matched direct ACL and from the FIRST matched inherited ACL
    * at EACH parent level. For these inherited ACL, the rights from each parent
    * are added to the final result.
    * <P>
    * This section is about the virtual ACL in the MySQL addon.
    * There are two types of virtual ACL: the virtual
    * 'user@accounts-for-hosts group' ACL on any Resource, which derives from
    * one 'accounts-for-hosts group' real ACL, and the mixed ACL which is only
    * on databases. The mixed ACL derives from two to-be-mixed real ACL.
    * The virtual user@accounts-for-hosts group ACL is so derived
    * from a non-conditional accounts-for-hosts group ACL, and the target may have any type.
    * There are processed rights for each account member of the accounts-for-hosts group.
    * If there are 3 members into the accounts-for-hosts group, there are 3 sets of
    * processed rights for three keys in the rights maps. All these rights are
    * simply the copy of the accounts-for-hosts group ACL rights.
    * <P>
    * The database level is the most complex case. It is the only parent level
    * to handle the virtual mixed ACL, the to-be-mixed real ACL and the real bases set ACLs.
    * The to-be-mixed real ACL are the only case where two real ACL are selected
    * at a level (this is indeed always the database level) to build up some rights
    * in a virtual mixed ACL. To-be-mixed and mixed ACL are not official MySQL concepts,
    * but they implement a MySQL algorithm.
    * Two non-conditional to-be-mixed ACL on a database-related target, respectively
    * for an user U and a host H, are combined to define the processed rights
    * for the pair (U, H). The processed rights are the union of the rights
    * from the two real ACL. This define a map key even if the union of rights is empty.
    * This is the only case where a key in the two maps, may have an empty set
    * of effective rights. 
    * <P>
    * This section is about the keys into the primary rights map.
    * Each parent level may produce a matched key for a given couple user/host and a target.
    * Into the primary ACL rights map,
    * the map key may have varied forms, where the parent order is always at the end.
    * The parent order tells the distance from the target to the parent which delivers 
    * the referenced rights. For instance, the direct parent of a resource has the string '1'.
    * The direct parent of this parent has the string '2' as parent order, and so on.
    * Into the two rights maps, the single difference between the keys is that the key for
    * the primary rights is a String form of a Name, and the global rights key is
    * the same Name. <P>
    * The ten key forms are listed hereinafter: <P>
    * - (user name @ host name last component) ::  parent order,
    * for instance 'one :: mysql :: jerry@localhost :: 2',<P>
    * - <Proc>:: (user name @ accounts-for-hosts group name last component) :: parent order,
    * for instance '<Proc>:: one :: mysql :: jerry@192.168.1.% :: 0',<P>
    * '<Proc>' is a generic tag that means 'processed rights', to tell the key and
    * its rights are not get from the direct reading of a real ACL.
    * These two first forms are the only used ones when the parent level is not a database.
    * These forms are also used for a database, but the followings are additional forms in this case:
    * - <Proc><from 'bases set name last component'>:: (user name @ host name last component):: parent order, <P>
    * - <Proc><from 'bases set name last component'>:: (user name @ accounts-for-hosts group name
    * last component) :: parent order, <P>
    * - <Proc><Mixed>:: (user name @ host name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'>:: (user name @ host name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'><from 'bases set name last component'>::
    * (user name @ host name last component) :: parent order, <P>
    * - <Proc><Mixed>:: (user name @ hosts group name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'>:: (user name @ hosts group name last component) :: parent order, <P>
    * - <Proc><Mixed><from 'bases set name last component'><from 'bases set name last component'>::
    * (user name @ hosts group name last component) :: parent order.
    * <P>
    * In this catalog of key forms for the database, '<Mixed>' means this is a virtual mixed ACL,
    * and '<from 'bases set name last component'>' means a bases set ACL is the real ACL
    * from which the virtual ACL is derived. The tag '<Mixed>' is proper to MySQL
    * but it may be reused in another ACS Addons. The tag '<from 'xxxx'>' is a generic form
    * another ACS Addon may use.
    * A key contains two times '<from 'bases set name last component'>' when there are
    * both a real bases set ACL for the user, and a real bases set ACL for the host
    * or the hosts group. These key forms are designed to help the Access Road user
    * to find the origin of a key into a rights map.
    * A key has exactly one origin, which is one real ACL in most of the cases,
    * except the virtual mixed ACL where two real to-be-mixed ACL are processed.
    * The rights map keys of a parent level are not sorted
    * following the ACL sorting, or following any other sorting.
    * The value in a rights map is always a non-null StringRight array.
    * This array may be empty only for a mixed virtual ACL.
   * Overrides the Resource method to describe its use by the MySQL addon.
   * @return unmodifiable map of primary rights from the direct and inherited AclEntries.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedACLRights();

    /**
    * Removes the direct AclEntry and the inherited AclEntries in the children, but
    * without event firing. The purpose of this method is to avoid an unuseful sequence
    * of firings when a lot of AclEntries are removed. It is called from the root
    * 'MySQL server' method removeAclEntry() when it deletes all the direct ACL of a pair
    * on the root children when the root ACL for this pair is deleted.
    * Calls removeInheritedAclEntry() on the children and also calls the overridden method,
    * and disables before all the firings for the 'InheritedTargetAclEntries', 'AclRights'
    * and 'EPsThroughAclEntry' events.
    * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl#addAclEntry(ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.StringRight[], java.lang.Boolean, ARoad0.gBaseInterface.ImmutableGroupID) 
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
  public void removeAclEntryWithoutFiring(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError;

} // ResourceMySQL
