//Title:        Access Road
//Version:      0.7.1
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gWork;

import ARoad0.gBaseInterface.*;
import ARoad0.gBase.*;

import ARoad0.AcsAddon.Accbee.MySQL.gBase.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;
import ARoad0.Pattern.AcsFactoryForAcsAddon;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

/**
* This class is an utility class for the creation of the MySQL 5 ACS on any parent ACS.
* It uses the packages ARoad0.AcsAddon.Accbee.MySQL.xxx to create the ACS in the base.
* The requirements for the MySQL access controls are described
* in http://dev.mysql.com/doc/refman/5.5/en/request-access.html and the near pages. <P>
* To simulate the MySQL server, the features are provided first by
* a configuration of the Access Road core functions in the generic gBase package.
* This is the responsibility of this class:<P>
* - to model the MySQL server, using UserID, GroupID, Resource, Directory, Actor,
* VirtualFolder, AclEntry, conditional AclEntry and inherited AclEntry, Bridge,
* hard alias, <P>
* - 31 MySQL rights, and no direct use of the metarights, <P>
* - 12 main types of Resources, 14 main types of EligibleParties, 3 main types of VirtualFolders, <P>
* - 42 ACS type policies for GroupID, Directory, Resource, VirtualFolder, Actor, <P>
* - 27 specialized rights for AclEntry on all the types of targets and sources, <P>
* - 4 automatic associations for new MySQL database. <P>
* The MySQL server is mainly simulated by the way of a MySQL AcsAddon,
* with the following specific features: <P>
* - inherited AclEntries and AclEntry sorting, in MySQL.gBase, <P>
* - enforcing exactly one host per account, in MySQL.gBase, <P>
* - UserID as parameter of the VirtualFolder constructor, in MySQL.gBase, <P>
* - VirtualFolder to simulate a MySQL view, enforcing to have
* elements only from the database, in MySQL.gBase, <P>
* - VirtualFolder alias as Actor running under the VirtualFolder UserID,
* in MySQL.gBase, <P>
* - '%' and '_' wildcard characters in the names of groups and directories, in MySQL.gBase, <P>
* - BasesSetMySQL to model a set of databases, in MySQL.gBase, <P>
* - ProxyMySQL to model a proxy account, in MySQL.gBase, <P>
* - conditional overlaying of the received AG context on the VirtualFolder alias,
* in MySQL.gWork, <P>
* - detection of the effective rights of an user, in MySQL.Work, <P>
* - access path deleting if there is another path having priority, <P>
* in MySQL.gWork. <P>
* No roots are created on the parent ACS.
* The limitations on the MySQL server simulation are described in the package documentation.
* To comply with the AcsAddon architectural pattern, this class offers the method
* create() called by ActionNewACSyst.create(), and the method configure()
* called by ActionNewACSyst.fillWithAcsObjects().
* @version 0.7.1
 **/
public class AcsFactoryMySQL implements AcsFactoryForAcsAddon {
      /**
       * the only instance of the class, following singleton pattern
       */
  private static AcsFactoryMySQL    instance__ = new AcsFactoryMySQL();
      /**
       * The descriptor of the AcsAddon to set in the menu for creating a new ACS.
       * It has to contain a '(' character.
       * ActionNewACSyst uses it to define the ACS type, as the beginning of the text
       * up to the first '(' character.
       */
  public static final String        PUBLIC_ACS_ADDON_NAME = "MySQL 5 server (by Accbee, France, EU)";

     // Constructors

  /**
   * with this public constructor for the subclasses.
  */
  public AcsFactoryMySQL() {
  }

    /**
    * Only method to obtain the unique instance.
    * Called by reflection from ActionNewACSyst.createACS().
    * @return an AcsFactoryMySQL instance
    */
  public static AcsFactoryForAcsAddon getInstance() {
    return instance__;
  } // getInstance()

  /**
   * Gets the descriptor of the AcsAddon, to set in the menu for creating a new ACS.
   * This method has to return a static variable, because it is called after a call
   * to the empty constructor.
   * Called by ActionNewACSyst.actionPerformed().
   */
  public String getPublicAcsAddonName() {
      return PUBLIC_ACS_ADDON_NAME;
  }

  /**
   * Creates a specific MySQL ACS.
   * The method configure() is called after this method by ActionNewACSyst.fillWithAcsObjects().
   * The limitations on the MySQL server simulation are described in the package documentation. <P>
   * Creates a specific MySQL ACS.
   * The NX ACS is detected when its type contains 'Linux'.
   * configure() is called after this method by ActionNewACSyst.fillWithAcsObjects().
   * This method makes most of the work, but the ACS vocabulary is set by
   * ACSFactoryMySQLImpl directly.
   * Called by ActionNewACSyst.create(). Calls configureMySQLonLinux().
   * @param _iSname the IS of the new ACS. May be null.
   * @param _name of the ACS to create or open.
   *   No test of null value nor empty.
   * @param _type must be the beginning of PUBLIC_ACS_ADDON_NAME
   * @param _parent is the parent node of the new ACS in the IS. Never null.
   * Have to be an ACS.
   * @exception BaseError if the ACS is known, the ACS is open,
   *   the core ACS is null when it is mandatory, or
   *   ClassNotFoundException, IOException with
   *   invalidClassException if ACSImpl HAS CHANGED,
   *   or if the type or the specified structure are not correct.
   * @return the new ACS
   */
  public ACSImpl create(ImmutableName _iSname,
          ImmutableName _name, String _type, ImmutableName _parent) throws BaseError {

        if (!PUBLIC_ACS_ADDON_NAME.startsWith(_type))
            throw new BaseError("The ACS type "+_type+"is not a MySQL 5 server, from "+
                    PUBLIC_ACS_ADDON_NAME);
        
        BaseObject parent = UtilityImpl.getFirstInstance().getBaseObject(_parent);
        if (parent == null || !(parent instanceof ACS))
            throw new BaseError("The parent node is not an ACS.");

        ACS core = (ACS)parent;
        ACSImpl acs = null;
        ACSFactoryImpl factory = null;
        StringRightImpl all = null;
        StringRightImpl grant = null;
        StringRightImpl proxy = null;
        StringRightImpl alter = null;
        StringRightImpl crea = null;
        StringRightImpl creavw = null;
        StringRightImpl creart = null;
        StringRightImpl crtt = null;
        StringRightImpl creatas = null;
        StringRightImpl creauser = null;
        StringRightImpl del = null;
        StringRightImpl drop = null;
        StringRightImpl ind = null;
        StringRightImpl ins = null;
        StringRightImpl sel = null;
        StringRightImpl showvw = null;
        StringRightImpl upd = null;
        StringRightImpl exe = null;

        StringRightImpl altrt = null;
        StringRightImpl fil = null;
        StringRightImpl trig = null;
        StringRightImpl lock = null;
        StringRightImpl proc = null;
        StringRightImpl rld = null;
        StringRightImpl evt = null;
        StringRightImpl rpl = null;
        StringRightImpl rpls = null;
        StringRightImpl showdb = null;
        StringRightImpl down = null;
        StringRightImpl sup = null;
        StringRightImpl no = null;

        String[] l_asso = null;

        // creates the ACS
        try {
            factory = new ACSFactoryMySQLImpl(core.getISName(), _name, _type, core.getName());

            // creates the acsrights
            grant = new StringRightImpl("GRANT OPTION",
                    BaseManagerImpl.getInstance().getMetaRight("|transfer_limited_rights|"),
                    980, null, null, factory );
            grant.setComment("The rights owner may give to or remove from a second owner "+
                  "all the rights he possesses himself, excluding this GRANT OPTION, "+
                  "and then the second owner cannot transfer the received rights to a third owner.");
            // privilege from MySQL 5.5.7
            proxy = new StringRightImpl("PROXY",
                    BaseManagerImpl.getInstance().getMetaRight("|modify_owner|"),
                    800, null, null, factory );
            proxy.setComment("Enable a proxy account to connect as another proxied account");
            creauser = new StringRightImpl("CREATE USER",
                    BaseManagerImpl.getInstance().getMetaRight("|modify_security_descriptor|"),
                    500, new StringRightImpl[] {proxy}, null, factory );
            creauser.setComment("Authorize to CREATE, DROP, RENAME, REVOKE all rights on USER");

            all = new StringRightImpl("ALL",
                    BaseManagerImpl.getInstance().getMetaRight("full_control"),
                    900, null, null, factory );
            all.setComment("All rights excepting GRANT OPTION");
            sup = new StringRightImpl("SUPER",
                    BaseManagerImpl.getInstance().getMetaRight("write_for_directory"),
                    800, new StringRightImpl[] {all}, null, factory );
            sup.setComment("Authorize to do some administrative tasks on the server");
            down = new StringRightImpl("SHUTDOWN",
                    BaseManagerImpl.getInstance().getMetaRight("write_for_directory"),
                    800, new StringRightImpl[] {all}, null, factory );
            down.setComment("Authorize to SHUTDOWN SERVER");
            rld = new StringRightImpl("RELOAD",
                    BaseManagerImpl.getInstance().getMetaRight("write_for_directory"),
                    800, new StringRightImpl[] {all}, null, factory );
            rld.setComment("Authorize to FLUSH, which is to clear or reload various internal caches");
            evt = new StringRightImpl("EVENT",
                    BaseManagerImpl.getInstance().getMetaRight("write_for_directory"),
                    800, new StringRightImpl[] {all}, null, factory );
            evt.setComment("Authorize to create, alter, or drop events of a database "+
                    "for the Event Scheduler");
            crea = new StringRightImpl("CREATE",
                    BaseManagerImpl.getInstance().getMetaRight("write"),
                    500, new StringRightImpl[] {all}, null, factory );
            crea.setComment("Authorize to CREATE database, TABLE, INDEX");
            creavw = new StringRightImpl("CREATE VIEW",
                    BaseManagerImpl.getInstance().getMetaRight("write"),
                    500, new StringRightImpl[] {all}, null, factory );
            creavw.setComment("Authorize to CREATE VIEW");
            creart = new StringRightImpl("CREATE ROUTINE",
                    BaseManagerImpl.getInstance().getMetaRight("write"),
                    500, new StringRightImpl[] {all}, null, factory );
            creart.setComment("Authorize to CREATE stored procedures and functions");
            creatas = new StringRightImpl("CREATE TABLESPACE",
                    BaseManagerImpl.getInstance().getMetaRight("write"),
                    500, new StringRightImpl[] {all}, null, factory );
            creatas.setComment("Authorize to CREATE, ALTER, or DROP tablespaces and log file groups");
            drop = new StringRightImpl("DROP",
                    BaseManagerImpl.getInstance().getMetaRight("deleteRenameCopy"),
                    500, new StringRightImpl[] {all}, null, factory );
            drop.setComment("Authorize to remove database, TABLE, VIEW");
            alter = new StringRightImpl("ALTER",
                    BaseManagerImpl.getInstance().getMetaRight("deleteRenameCopy"),
                    450, new StringRightImpl[] {drop}, null, factory );
            alter.setComment("Authorize to ALTER TABLE");
            del = new StringRightImpl("DELETE ROW",
                    BaseManagerImpl.getInstance().getMetaRight("delete"),
                    440, new StringRightImpl[] {alter}, null, factory );
            del.setComment("The officiel MySQL name is simply DELETE. Authorize to DELETE ROW in a TABLE.");
            altrt = new StringRightImpl("ALTER ROUTINE",
                    BaseManagerImpl.getInstance().getMetaRight("deleteRenameCopy"),
                    440, new StringRightImpl[] {drop}, null, factory );
            altrt.setComment("Authorize to ALTER stored procedures and functions");
            crtt = new StringRightImpl("CREATE TEMPORARY TABLE",
                    BaseManagerImpl.getInstance().getMetaRight("create"),
                    400, new StringRightImpl[] {crea}, null, factory );
            crtt.setComment("Authorize the use of CREATE TEMPORARY TABLE");
            ind = new StringRightImpl("INDEX",
                    BaseManagerImpl.getInstance().getMetaRight("modify"),
                    400, new StringRightImpl[] {crea}, null, factory );
            ind.setComment("Authorize to CREATE and DROP INDEX on TABLE");
            upd = new StringRightImpl("UPDATE",
                    BaseManagerImpl.getInstance().getMetaRight("modify"),
                    400, new StringRightImpl[] {crea}, null, factory );
            upd.setComment("Authorize to UPDATE rows on TABLE, or on COLUMN");
            exe = new StringRightImpl("EXECUTE",
                    BaseManagerImpl.getInstance().getMetaRight("execute"),
                    400, new StringRightImpl[] {crea}, null, factory );
            exe.setComment("Authorize to EXECUTE stored procedures and functions");
            proc = new StringRightImpl("PROCESS",
                    BaseManagerImpl.getInstance().getMetaRight("read_for_directory"),
                    400, new StringRightImpl[] {all}, null, factory );
            proc.setComment("Authorize to SHOW FULL PROCESSLIST, "+
                    "even the threads belonging to other accounts");
            rpl = new StringRightImpl("REPLICATION CLIENT",
                    BaseManagerImpl.getInstance().getMetaRight("read_for_directory"),
                    400, new StringRightImpl[] {all}, null, factory );
            rpl.setComment("Authorize to know the replication masters and slaves");
            rpls = new StringRightImpl("REPLICATION SLAVE",
                    BaseManagerImpl.getInstance().getMetaRight("read_for_directory"),
                    400, new StringRightImpl[] {all}, null, factory );
            rpls.setComment("Authorize slave to read the master binaries");
            showdb = new StringRightImpl("SHOW DATABASES",
                    BaseManagerImpl.getInstance().getMetaRight("read_for_directory"),
                    400, new StringRightImpl[] {all}, null, factory );
            showdb.setComment("Authorize to see all the BASES");
            ins = new StringRightImpl("INSERT",
                    BaseManagerImpl.getInstance().getMetaRight("insert_item"),
                    300, new StringRightImpl[] {drop}, null, factory );
            ins.setComment("Authorize to INSERT rows on TABLE, or on COLUMN");
            sel = new StringRightImpl("SELECT",
                    BaseManagerImpl.getInstance().getMetaRight("read"),
                    200, new StringRightImpl[] {all}, null, factory );
            sel.setComment("Authorize to SELECT rows on TABLE, COLUMN or VIEW");
            showvw = new StringRightImpl("SHOW VIEW",
                    BaseManagerImpl.getInstance().getMetaRight("execute"),
                    300, new StringRightImpl[] {all}, null, factory );
            showvw.setComment("Authorize to SHOW VIEW");
            trig = new StringRightImpl("TRIGGER",
                    BaseManagerImpl.getInstance().getMetaRight("execute"),
                    300, new StringRightImpl[] {all}, null, factory );
            trig.setComment("Authorize to create, drop, or execute triggers on TABLE");
            fil = new StringRightImpl("FILE",
                    BaseManagerImpl.getInstance().getMetaRight("read"),
                    200, new StringRightImpl[] {all}, null, factory );
            fil.setComment("Authorize to SELECT ... INTO OUTFILE and LOAD DATA INFILE");
            lock = new StringRightImpl("LOCK TABLES",
                    BaseManagerImpl.getInstance().getMetaRight("read"),
                    200, new StringRightImpl[] {all}, null, factory );
            lock.setComment("Authorize to LOCK on TABLE if SELECT");
            // 5 is the maximum number of upper rights
            no = new StringRightImpl("USAGE",
                    BaseManagerImpl.getInstance().getMetaRight("goto"),
                    100, new StringRightImpl[] {sel, del, upd, exe, ins},
                    null, factory );
            no.setComment("At the global level, USAGE means 'no rights at this level "+
                    "but allows rights at another level', while at the other levels, "+
                    "it simply means 'no rights at this level'.");

            factory.setManageAcsRights(true);

            factory.addCommentsOnProperties("All the rights",
                    new String[] {"There are 31 MySQL rights, including:",
                                    "- GRANT OPTION is the only",
                                    "  administrative right,",
                                    "  to give to or to remove from",
                                    "  its rights on a second owner,",
                                    "- ALL includes all rights",
                                    "  excepting GRANT OPTION,",
                                    "- the administrative tasks",
                                    "  are reserved to the server",
                                    "  administrators, and their",
                                    "  main rights are SUPER,",
                                    "  SHUTDOWN and RELOAD,",
                                    "- PROCESS, REPLICATION CLIENT,",
                                    "  REPLICATION SLAVE are other",
                                    "  rights for the server management",
                                    "- CREATE, DROP and ALTER",
                                    "  are the most powerful rights",
                                    "  on the bases and the tables,",
                                    "- SELECT, UPDATE, LOCK, INDEX,",
                                    "  INSERT and DELETE ROW are rights",
                                    "  on tables, indexes, rows,",
                                    "- CREATE and SHOW VIEW are reserved",
                                    "  to the view handling but do not",
                                    "  provide rights on the view items,",
                                    "- CREATE and ALTER ROUTINE,",
                                    "  EXECUTE on stored routines,",
                                    "- FILE and CREATE TEMPORARY TABLE",
                                    "  are secondary rights,",
                                    "- USAGE delivers no rights."
                                    });

            // creates the types ---------------------------------------------
            factory.setL_EPTypes(new String[] {
                "account",
                "proxy",
                "<byAcsCreator> connection", // only under 'User Connections'
                "view_delegate <byAcsCreator>",
                "trigger",
                "stored_function",
                "stored_procedure",
                // all the types containing 'host' and not 'hosts' are for simple hosts
                "host",
                    // ' <byAcsCreator>' means the type is editable while the instance
                    // cannot be finalized by the user
                "host <byAcsCreator>", // for 'localhost'
                "pattern-name_accounts-for-hosts group", // new
                "pattern-name_hosts group",
                //"Workbench administrative role",
                //"Workbench non-administrative role",

                "<immutable> account",
                "<immutable> proxy",
                "<immutable> <byAcsCreator> connection", // only under 'User Connections'
                "<immutable> view_delegate <byAcsCreator>",
                "<immutable> trigger",
                "<immutable> stored_function",
                "<immutable> stored_procedure",
                "<immutable> host",
                "<immutable> host <byAcsCreator>", // for 'localhost'
                "<immutable> pattern-name_accounts-for-hosts group", 
                "<immutable> pattern-name_hosts group",
                //"<immutable> Workbench administrative role",
                //"<immutable> Workbench non-administrative role",
                
                "<byAcsCreator>",
                "<byAcsCreator><immutable> ",

                // for Explorer.AcsGroupTree
                "<byAcsCreator> standard pattern-name_accounts-for-hosts group",
                "<byAcsCreator> standard pattern-name_hosts group",
                "<byAcsCreator> standard for simple hosts"
            });

            factory.setL_ResourceTypes(new String[] {
                "file",
                "trigger",
                "index",
                "table",
                "database",
                "view_delegate <byAcsCreator>",
                "table_column",
                "stored_function",
                "stored_procedure",
                "pattern-name_bases set",

                "<immutable> <byAcsCreator> connection",
                "<immutable> file",
                "<immutable> trigger",
                "<immutable> index",
                "<immutable> trigger",
                "<immutable> table",
                "<immutable> database",
                "<immutable> view_delegate <byAcsCreator>",
                "<immutable> table_column",
                "<immutable> stored_function",
                "<immutable> stored_procedure",
                "<immutable> pattern-name_bases set",
                
                "<byAcsCreator>",
                "MySQL server <byAcsCreator>",
                "<byAcsCreator><immutable> ",
                // parent of the bases sets
                "virtual_objects <byAcsCreator>",

                 // connection system
                "<byAcsCreator> connection"
            });

            // Actor types are the common values in the resource list and the EP list:
            // "view_delegate", "stored_function", "stored_procedure", "trigger",
            // "<byAcsCreator> connection"
            
            //factory.setL_ResourceVirtualFolderTypes(new String[] {});

            factory.setL_EPVirtualFolderTypes(new String[] {
                "view", 
                "<immutable> view",
                "<immutable> root for a set", // for database
                "<immutable> no-child node"}); // for MySQL server

            //factory.setL_PrivilegeForTypeTypes(new String[] {});
            //factory.setL_PrivilegeForLinksTypes(new String[] {});

            // creates the structure -------------------------------------------
            factory.setIncompleteStructureModeling(true);
            factory.setIncompleteBehaviorModeling(true);

            factory.setManageResources(true);
            factory.setManageResourceTrees(true);
            factory.setManageAccounts(true);
            factory.setManageGroups(true);
            factory.setManageAGORights(false);
            factory.setManageActors(true);
            factory.setManageInternalAcls(true);
            factory.setManageExternalAcls(true); // new in 0.7.1
            factory.setManageBridges(true);
            factory.setManageSecGroupsForActor(false);
            factory.setManageGroupTrees(true);
            factory.setManageAGORightsInheritance(false);
            factory.setManageACLRightsInheritance(true);
            factory.setManageLPRIRightsInheritance(false);
            factory.setManageInternalHardAlias(true);
            factory.setManageSoftAlias(false);
            factory.setManageFlexibleConditionalAcls(true);
            factory.setManageRightDefinedConditionalAcls(false);
                // VirtualFolders
            factory.setManageVirtualFolders(true);
            factory.setManageVirtualFolderTrees(true);
            factory.setManageOneTimeInVirtualFolderTreeForMember(false);
                // VirtualFolders for Resources
            factory.setManageResourceVirtualFolders(false);
            factory.setManageSimpleResourceVirtualFolders(false);
            factory.setManageDirectoryInResourceVirtualFolders(false);
                // VirtualFolders for EPs
            factory.setManageEPVirtualFolders(true);
            factory.setManageActorInEPVirtualFolders(true);
            factory.setManageDirectoryEPinEPVirtualFolders(false);
                //factory.setManageVirtualFolderRightsPropagationToMembers(true);
            // Typed and Linked Privileges
            factory.setManagePrivilegeRights(false);
            factory.setManageGrantingRightPrivileges(false);
            factory.setManageDenyingRightPrivileges(false);
            factory.setManageSimpleTypedPrivileges(false);
            factory.setManageSimpleLinkedPrivileges(false);

            // creates the SubAcsNodes and creates the relevant types and policy --------
            // (nothing for the key Explorer.SubAcsNode)
            // for Explorer.AcsGroupTree
            factory.addSubAndGroupPolicy("standard",
                    new String[] {"<GroupTree>"});
            factory.addSubAndGroupPolicy("connection",
                    new String[] {"<ResourceSubACS>"});

            // specialized types ------------------------------------------------
            // no NoAcsNode key to forbid some nodes in the explorer
            // CreationByBeamer.NoType to forbid creation of these types by the user
            factory.addTypesAndExplorer("CreationByBeamer.NoType",
                    new String[] {
                        // for Resource, only 'file', 'table_column', 'index' are allowed
                        "ResourceType.<undefined>",
                        "ResourceType.<byAcsCreator>",
                        "ResourceType.MySQL server <byAcsCreator>",
                        "ResourceType.<byAcsCreator><immutable> ",
                        "ResourceType.<byAcsCreator> connection",
                        "ResourceType.table",
                        "ResourceType.database",
                        "ResourceType.view_delegate <byAcsCreator>",
                        "ResourceType.trigger",
                        "ResourceType.stored_function",
                        "ResourceType.stored_procedure",
                        "ResourceType.pattern-name_bases set",
                        "ResourceType.<immutable> <byAcsCreator> connection",
                        "ResourceType.<immutable> table",
                        "ResourceType.<immutable> database",
                        "ResourceType.<immutable> view_delegate <byAcsCreator>",
                        "ResourceType.<immutable> trigger",
                        "ResourceType.<immutable> stored_function",
                        "ResourceType.<immutable> stored_procedure",
                        "ResourceType.<immutable> pattern-name_bases set",
                        "ResourceType.virtual_objects <byAcsCreator>",

                        // for Directory
                        "DirectoryType.<undefined>",
                        "DirectoryType.<byAcsCreator>",
                        "DirectoryType.MySQL server <byAcsCreator>",
                        "DirectoryType.<byAcsCreator><immutable> ",

                        "DirectoryType.file",
                        "DirectoryType.trigger",
                        "DirectoryType.index",
                        "DirectoryType.view_delegate <byAcsCreator>",
                        "DirectoryType.table_column",
                        "DirectoryType.<immutable> file",
                        "DirectoryType.<immutable> trigger",
                        "DirectoryType.<immutable> index",
                        "DirectoryType.<immutable> view_delegate <byAcsCreator>",
                        "DirectoryType.<immutable> table_column",
                        "DirectoryType.virtual_objects <byAcsCreator>",

                        // for Actor, only 'stored_function/procedure', 'trigger'
                        "ActorType.<undefined>",
                        "ActorType.<byAcsCreator>",
                        "ActorType.MySQL server <byAcsCreator>",
                        "ActorType.<byAcsCreator><immutable> ",
                        "ActorType.table",
                        "ActorType.database",
                        "ActorType.index",
                        "ActorType.view_delegate <byAcsCreator>",
                        "ActorType.<immutable> table",
                        "ActorType.<immutable> database",
                        "ActorType.<immutable> index",
                        "ActorType.<immutable> view_delegate <byAcsCreator>",

                        "ActorType.account",
                        "ActorType.<immutable> account",
                        "ActorType.proxy",
                        "ActorType.<immutable> proxy",
                        "ActorType.virtual_objects <byAcsCreator>",
                        "ActorType.pattern-name_bases set",
                        "ActorType.<immutable> pattern-name_bases set",
                        //"ActorType.general group",
                        //"ActorType.<immutable> general group",
                        "ActorType.view_delegate <byAcsCreator>",
                        "ActorType.<immutable> view_delegate <byAcsCreator>",
                        "ActorType.host",
                        "ActorType.<immutable> host",
                        
                        // for UserID, only 'account' and 'proxy'
                        "UserIDType.<byAcsCreator><immutable> ",
                        "UserIDType.<undefined>",
                        "UserIDType.<byAcsCreator>",

                        "UserIDType.host <byAcsCreator>",
                        "UserIDType.host",
                        "UserIDType.pattern-name_accounts-for-hosts group",
                        "UserIDType.pattern-name_hosts group",
                        "UserIDType.stored_function",
                        "UserIDType.stored_procedure",
                        "UserIDType.view_delegate <byAcsCreator>",
                        "UserIDType.trigger",

                        "UserIDType.<immutable> host",
                        "UserIDType.<immutable> host <byAcsCreator>",
                        "UserIDType.<immutable> pattern-name_accounts-for-hosts group",
                        "UserIDType.<immutable> pattern-name_hosts group",
                        "UserIDType.<immutable> stored_function",
                        "UserIDType.<immutable> stored_procedure",
                        "UserIDType.<immutable> view_delegate <byAcsCreator>",
                        "UserIDType.<immutable> trigger",

                        "UserIDType.<byAcsCreator> standard",
                        "UserIDType.<byAcsCreator> standard pattern-name_accounts-for-hosts group",
                        "UserIDType.<byAcsCreator> standard pattern-name_hosts group",
                        "UserIDType.<byAcsCreator> standard for simple hosts",

                        // for GroupID, only 'host', 'accounts-for-hosts group'
                        // and 'hosts group'
                        "GroupIDType.<byAcsCreator><immutable> ",
                        "GroupIDType.<undefined>",
                        "GroupIDType.<byAcsCreator>",

                        "GroupIDType.host <byAcsCreator>", // used only by configure() for '|localhost|'

                        "GroupIDType.stored_function",
                        "GroupIDType.stored_procedure",
                        "GroupIDType.view_delegate <byAcsCreator>",
                        "GroupIDType.trigger",
                        "GroupIDType.account",
                        "GroupIDType.proxy",

                        "GroupIDType.<immutable> host <byAcsCreator>",
                        "GroupIDType.<immutable> stored_function",
                        "GroupIDType.<immutable> stored_procedure",
                        "GroupIDType.<immutable> view_delegate <byAcsCreator>",
                        "GroupIDType.<immutable> trigger",
                        "GroupIDType.<immutable> account",
                        "GroupIDType.<immutable> proxy",

                        "GroupIDType.<byAcsCreator> standard",
                        "GroupIDType.<byAcsCreator> standard pattern-name_accounts-for-hosts group",
                        "GroupIDType.<byAcsCreator> standard pattern-name_hosts group",
                        "GroupIDType.<byAcsCreator> standard for simple hosts",

                        // for VirtualFolder, only 'view'
                        "VirtualFolderType.<immutable> no-child node",
                        "VirtualFolderType.root for a set",
                        "VirtualFolderType.<immutable> root for a set"
            });

            // GroupID.NoPrimaryGroup for all UserID
            factory.addTypesAndExplorer("GroupID.NoPrimaryGroup",
                    new String[] {
                        //"general group",
                        //"<immutable> general group",
                        "host",
                        "<immutable> host",
                        "host <byAcsCreator>",
                        "<immutable> host <byAcsCreator>",
                        "pattern-name_accounts-for-hosts group",
                        "pattern-name_hosts group",
                        "<byAcsCreator> standard",
                        "<byAcsCreator> standard pattern-name_accounts-for-hosts group",
                        "<byAcsCreator> standard pattern-name_hosts group",
                        "<byAcsCreator> standard for simple hosts"
            });
            // an Actor view_delegate, stored_function, stored_procedure,
            // trigger may have a current user and a current group
            // no Actor.NoCurrentAccount
            // no Actor.NoCurrentGroup

            // no GroupIDMember.NoMemberOf
            // no GroupID.MemberFromAdmin
            // no GroupID.NoMoreThanOneMember
            
            // no GroupID.NoUserIDAsMember for
            // 'pattern-name_accounts-for-hosts group <byAcsCreator>'
            factory.addTypesAndExplorer("GroupID.NoUserIDAsMember",
                    new String[] {
                        "pattern-name_hosts group",
                        "<byAcsCreator> standard",
            });

            // GroupID.NoGroupIDAsMember is for
            // 'pattern-name_accounts-for-hosts group <byAcsCreator>', but
            // 'GroupID.TypesOfMemberFor' already does the job.
            //factory.addTypesAndExplorer("GroupID.NoGroupIDAsMember",
              //      new String[] {
                //        "pattern-name_accounts-for-hosts group"
            //});

             // only one level of group-to-group children
            // 'GroupID.TypesOfMemberFor.' for 'host'
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "host",
                    new String[] {
                        "account",
                        "proxy"
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "host <byAcsCreator>",
                    new String[] {
                        "account",
                        "proxy"
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "pattern-name_accounts-for-hosts group",
                    new String[] {
                        "account"
            });
            // 'pattern-name_hosts group' is not allowed as child
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "pattern-name_hosts group",
                    new String[] {
                        "host",
                        "host <byAcsCreator>"
            });

            // for standard groups
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "<byAcsCreator> standard",
                    new String[] {
                        "<byAcsCreator> standard pattern-name_accounts-for-hosts group",
                        "<byAcsCreator> standard pattern-name_hosts group",
                        "<byAcsCreator> standard for simple hosts"
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "<byAcsCreator> standard pattern-name_accounts-for-hosts group",
                    new String[] {
                        "pattern-name_accounts-for-hosts group"
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "<byAcsCreator> standard pattern-name_hosts group",
                    new String[] {
                        "pattern-name_hosts group"
            });
            factory.addTypesAndExplorer("GroupID.TypesOfMemberFor."+
                    "<byAcsCreator> standard for simple hosts",
                    new String[] {
                        "host",
                        "host <byAcsCreator>"
            });

            // no 'GroupIDMember.NoMoreThanOneGroup.'
            // no 'GroupIDMember.RecommandedGroupsAsMember.'

            factory.addTypesAndExplorer("GroupID.ConditionalAclGroup",
                    new String[] {
                        "host",
                        "host <byAcsCreator>",
                        "pattern-name_hosts group"
            });

            // on Directories
            // for the root, only 'MySQL_server' and directories of Linux files
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "NULL",
                    new String[] {
                        "MySQL server <byAcsCreator>",
                        "<byAcsCreator>" // for 'connection' subACS
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "<byAcsCreator>",
                    new String[] {
                        "<byAcsCreator>",
                        "MySQL server <byAcsCreator>",
                        "<byAcsCreator><immutable> "
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "<byAcsCreator>",
                    new String[] {
                        "<byAcsCreator> connection"
            });
            // for the 'MySQL_server', only bases
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "MySQL server <byAcsCreator>",
                    new String[] {
                        "<byAcsCreator>",
                        "<byAcsCreator><immutable> ",
                        "database",
                        "virtual_objects <byAcsCreator>"
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "virtual_objects <byAcsCreator>",
                    new String[] {
                        "pattern-name_bases set"
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "pattern-name_bases set",
                    new String[0]);
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "database",
                    new String[] {
                        "<byAcsCreator>",
                        "<byAcsCreator><immutable> ",
                        "table",
                        "view_delegate <byAcsCreator>",
                        "stored_function", 
                        "stored_procedure" 
            });
            factory.addTypesAndExplorer("Directory.TypesOfChildFor."+
                    "table",
                    new String[] {
                        "<byAcsCreator>",
                        "<byAcsCreator><immutable> ",
                        "table_column",
                        "index",
                        "trigger"
            });

            // VirtualFolder children
            factory.addTypesAndExplorer("VirtualFolder.TypesOfChildFor."+
                    "NULL",
                    new String[] {
                        "no-child node"
            });
                // done in VirtualFolderMySQLImpl to avoid a wrong error message
            //factory.addTypesAndExplorer("VirtualFolder.TypesOfMemberFor."+
              //      "no-child node",
                //    new String[0]);
            factory.addTypesAndExplorer("VirtualFolder.TypesOfChildFor."+
                    "no-child node",
                    new String[] {
                        "root for a set"
            });
                // done in VirtualFolderMySQLImpl to avoid a wrong error message
            //factory.addTypesAndExplorer("VirtualFolder.TypesOfMemberFor."+
              //      "root for a set",
                //    new String[0]);
            factory.addTypesAndExplorer("VirtualFolder.TypesOfChildFor."+
                    "root for a set",
                    new String[] {
                        "view"
            });
            factory.addTypesAndExplorer("VirtualFolder.TypesOfMemberFor."+
                    "view",
                    new String[] {
                        "<byAcsCreator>",
                        "<byAcsCreator><immutable> ",
                        "table",
                        "table_column",
                        "stored_function",
                        "stored_procedure"
            });
            factory.addTypesAndExplorer("VirtualFolder.TypesOfChildFor."+
                    "view",
                    new String[0]);

            // 'Resource.NoAccount' for all the resources
            factory.addTypesAndExplorer("Resource.NoAccount",
                    new String[] {
                "<byAcsCreator> connection",
                "virtual_objects <byAcsCreator>",
                "pattern-name_bases set",
                "file",
                "index",
                "table",
                "database",
                "view_delegate <byAcsCreator>",
                "table_column",
                "stored_function",
                "stored_procedure",
                "trigger",

                "<byAcsCreator>",
                "MySQL server <byAcsCreator>",
                "<byAcsCreator><immutable> "
            });

            // 'index' as Resource, 'view_delegate' and 'trigger' as Actor
            // are targets having no direct ACL.

            // 'Resource.NoConditionalACL' for directory, file, index, trigger, 'view_delegate'
                    // inherited ACLs may be conditional ACLs,
                    // and they are not constrained by this command.
                    // 'EligibleParty.NoNonConditionalACL' does not exist,
                    // then it is done in the DirectoryMySQLImpl and BasesSetMySQLImpl classes.
            factory.addTypesAndExplorer("Resource.NoConditionalACL",
                    new String[] {
                "<byAcsCreator> connection",
                "virtual_objects <byAcsCreator>",
                "file",
                "index",
                "view_delegate <byAcsCreator>",
                "trigger",

                "<byAcsCreator>",
                "<byAcsCreator><immutable> "
            });

            // 'Resource.NoNonConditionalACL' for the same types
            factory.addTypesAndExplorer("Resource.NoNonConditionalACL",
                    new String[] {
                "<byAcsCreator> connection",
                "virtual_objects <byAcsCreator>",
                "file",
                "index",
                "view_delegate <byAcsCreator>",
                "trigger",

                "<byAcsCreator>",
                "<byAcsCreator><immutable> "
            });

            // 'Resource.OneConditionGroupInACL' for resources
            factory.addTypesAndExplorer("Resource.OneConditionGroupInACL",
                    new String[] {
                "table",
                "database",
                "table_column",
                "stored_function",
                "stored_procedure",
                "pattern-name_bases set",

                "<immutable> table",
                "<immutable> database",
                "<immutable> table_column",
                "<immutable> stored_function",
                "<immutable> stored_procedure",
                "<immutable> pattern-name_bases set",
                
                "MySQL server <byAcsCreator>"
            });

            // 'Resource.TypesOfSourceForConditionalACL.'
            factory.addTypesAndExplorer("Resource.TypesOfSourceForConditionalACL."+
                    "MySQL server <byAcsCreator>", new String[] {
                "account"
            });
            
            // 'Resource.TypesOfSourceForConditionalACL.'
            factory.addTypesAndExplorer("Resource.TypesOfSourceForConditionalACL."+
                    "database", new String[] {
                "account"
            });

            // 'Resource.TypesOfSourceForConditionalACL.'
            factory.addTypesAndExplorer("Resource.TypesOfSourceForConditionalACL."+
                    "table", new String[] {
                "account"
            });

            // 'Resource.TypesOfSourceForConditionalACL.'
            factory.addTypesAndExplorer("Resource.TypesOfSourceForConditionalACL."+
                    "stored_function", new String[] {
                "account"
            });

            // 'Resource.TypesOfSourceForConditionalACL.'
            factory.addTypesAndExplorer("Resource.TypesOfSourceForConditionalACL."+
                    "stored_procedure", new String[] {
                "account"
            });

            // 'Resource.TypesOfSourceForConditionalACL.'
            factory.addTypesAndExplorer("Resource.TypesOfSourceForConditionalACL."+
                    "pattern-name_bases set", new String[] {
                "account"
            });

            // 'Resource.TypesOfSourceForConditionalACL.'
            factory.addTypesAndExplorer("Resource.TypesOfSourceForConditionalACL."+
                    "table_column", new String[] {
                "account"
            });

            // no 'Actor.NoCurrentAccount'

            // 'Actor.NoNullCurrentAccount' 
            factory.addTypesAndExplorer("Actor.NoNullCurrentAccount",
                    new String[] {
                "trigger",
                "<byAcsCreator> connection"
                //"stored_function",
                //"stored_procedure",
                //"view_delegate <byAcsCreator>"
            });

            // 'Actor.IsNotBridgeTarget'
            factory.addTypesAndExplorer("Actor.IsNotBridgeTarget",
                    new String[] {
                "trigger",
                //"<byAcsCreator> connection"
                "stored_function",
                "stored_procedure",
                "view_delegate <byAcsCreator>"
            });

            // 'Actor.IsNotBridgeSource'
            // already forbidden by the ActorMySQL constructor from the GUI, but not from the code.
            factory.addTypesAndExplorer("Actor.IsNotBridgeSource",
                    new String[] {
                "trigger",
                "<byAcsCreator> connection",
                "stored_function",
                "stored_procedure",
                "view_delegate <byAcsCreator>"
            });
            
            // creates the rights sets for targets ---------------------------------------------

            // all the other rights are  specialized
            factory.setL_bridgeRightsSet(new StringRight[] {exe});
            
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.<byAcsCreator>",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.connections  <byAcsCreator>",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.<byAcsCreator> connection",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.virtual_objects <byAcsCreator>",
                    new StringRight[0]);

            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.file",
                    new StringRight[] {fil});
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.MySQL server <byAcsCreator>",
                    new StringRight[] {
                        // from http://dev.mysql.com/doc/refman/5.5/en/grant-table-structure.html
                        // not in Table 5.6. Permissible Privileges for GRANT and REVOKE
                        // in the MySQL table 'user'
                        all, no,
                        // in Table 5.6. Permissible Privileges, but without 'FILE'
                        grant, alter, crea, creavw, creart, crtt, creatas,
                        creauser, del, drop, ind, ins, sel, showvw, upd, exe, altrt,
                        trig, lock, proc, rld, evt, rpl, rpls, showdb, down, sup});

            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.pattern-name_bases set",
                    new StringRight[] {grant, alter, crea, creavw, creart, crtt,
                        del, drop, ind, ins, sel, showvw, upd, exe, altrt,
                        trig, lock, evt});

            // A MySQL privilege granted at the database level is applied to the database
            // and to all objects in the database, such as tables and stored programs.
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.database",
                    new StringRight[] {grant, alter, crea, creavw, creart, crtt,
                        del, drop, ind, ins, sel, showvw, upd, exe, altrt,
                        trig, lock, evt});
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.table",
                    new StringRight[] {grant, alter, crea, crtt,
                        del, drop, ind, ins, sel, upd, trig, lock});
            //factory.addSpecializedRights(
              //      "Target.AclRightsSet.Type.view_delegate <byAcsCreator>",
                //    new StringRight[] {creavw, sel, exe, drop, showvw});
            //factory.addSpecializedRights(
              //      "Target.AclRightsSet.Type.index",
                //    new StringRight[0]);
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.table_column",
                    new StringRight[] {ins, sel, upd});
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.stored_function",
                    new StringRight[] {
                        grant, exe, altrt});
            factory.addSpecializedRights(
                    "Target.AclRightsSet.Type.stored_procedure",
                    new StringRight[] {
                        grant, exe, altrt});

            // creates the rights sets for sources ------------------

            // the main list of rights, since for the accounts
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.account",
                    new StringRight[] {
                        all, no,
                        grant, alter, crea, creavw, creart, crtt, creatas,
                        creauser, del, drop, ind, ins, sel, showvw, upd, exe, altrt, fil,
                        trig, lock, proc, rld, evt, rpl, rpls, showdb, down, sup});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.pattern-name_accounts-for-hosts group",
                    new StringRight[] {
                        all, no,
                        grant, alter, crea, creavw, creart, crtt, creatas,
                        creauser, del, drop, ind, ins, sel, showvw, upd, exe, altrt, fil,
                        trig, lock, proc, rld, evt, rpl, rpls, showdb, down, sup});

                // ResourceMySQLImpl, DirectoryMySQLImpl and ActorMySQLImpl ensure that
                // 'host', 'host <byAcsCreator>', and 'pattern-name_hosts group'
                // can be used as ACL sources only for a database, a bases set.
                // The rights list is from the privilege columns in the db table (Table 5.7)
                // in the MySQL html documentation.
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.host",
                    new StringRight[] {
                        sel, ins, ind, upd, del,
                        alter, crea, drop, grant, creavw,
                        showvw, altrt, creart, exe, trig,
                        evt, crtt, lock});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.pattern-name_hosts group",
                    new StringRight[] {
                        sel, ins, ind, upd, del,
                        alter, crea, drop, grant, creavw,
                        showvw, altrt, creart, exe, trig,
                        evt, crtt, lock});
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.host <byAcsCreator>",
                    new StringRight[] {
                        sel, ins, ind, upd, del,
                        alter, crea, drop, grant, creavw,
                        showvw, altrt, creart, exe, trig,
                        evt, crtt, lock});

            // for Explorer.AcsGroupTree
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> standard",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> standard pattern-name_accounts-for-hosts group",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> standard pattern-name_hosts group",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.<byAcsCreator> standard for simple hosts",
                    new StringRight[0]);

            // these 3 Actors are not ACL sources, are without current UserID,
            // and they propagate the received UserID.
            // Actor with an alias relation from the view delegate to the view,
            // and propagation of the received userID or the indirect current userID
            // to the alias reference
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.view_delegate <byAcsCreator>",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.stored_function",
                    new StringRight[0]);
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.stored_procedure",
                    new StringRight[0]);
            // 'trigger' is the single Actor with a current UserID and a current GroupID.
            // it is no more a true ACL source.
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.trigger",
                    new StringRight[0]);
            // 'proxy' is no more a true ACL source.
            factory.addSpecializedRights(
                    "Source.AclRightsSet.Type.proxy",
                    new StringRight[0]);

            factory.addCommentsOnProperties("Specialized rights",
                    new String[] {"The MySQL server is modeled",
                                    "mostly as a tree of bases, tables ",
                                    "and columns (as resources),",
                                    "a virtual target for host mixed ACL,",
                                    "some bases set for bases ACL,",
                                    "a tree of database views",
                                    "(as virtual folders)",
                                    "user accounts (the UserIDs),",
                                    "groups (the GroupIDs)",
                                    "alias from a delegate to a view",
                                    "a set of ACLs and inherited ACLs",
                                    "(the AclEntries)",
                                    "from the accounts/groups to ",
                                    "the bases, tables, columns,",
                                    "and view delegates.",
                                    "",
                                    "There are specialized rights,",
                                    "without standard rights:",
                                    "- ACL is the unique type of rights,",
                                    "- each type of source or target",
                                    "  has a set of allowed rights,",
                                    "- if the target has the type",
                                    "  <byAcsACreator>, there is no right",
                                    "- the rights are defined both",
                                    "  with the source and the target",
                                    "  as criteria,",
                                    "- this implies 14 keys for",
                                    "  the targets",
                                    "  in the specialized rights",
                                    "- this implies 14 keys for",
                                    "  the sources",
                                    "  in the specialized rights",
                                    "- the view has no ACL rights, and",
                                    "  it is in the context of an account,",
                                    "- the view is accessible through ",
                                    "  its view delegate."
                                    });

            // associations ------------------------------------------------------

            // For the classes, the only types are 'Resource', 'Directory', 'Actor',
            // 'UserID', 'GroupID' and 'VirtualFolder' in a key as '<Class>.<Value>.UserID',
            // even in an AcsAddon. In the arguments, AcsAddon classes may be used.

            // a new UserID has a connection to allow a bridge from the true host account.
            // The bridge source has to be created by the user.
                // connection creation.
                // uses a constructor from the superclass ActorImpl to create
                // a 'connection' actor able to be a bridge target, but without
                // inheritance of the parent ACL (the current group is set by ActorImpl) .
            factory.addAssociationsToNewObject(
                    "<Class>.<Value>.UserID",
                    new String[] {
                        "<AssociateClass>.Actor",
                        "Name.<Value>.<MasterNameLastComponent>",
                        "ACS.<Value>.<MasterValue>",
                        "Parent Node.<Value>.<ACSName>+ <connection_subacs>::  ",
                        "CurrentUserID.<Master>",
                        "Resource Type.<Value>.<byAcsCreator> connection",
                        "CurrentGroupID.<Value>.<MasterValue>"});
            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                    get("<Class>.<Value>.UserID"));

                            // for testing and tracking
                            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                                throw new BaseError("AcsFactory.createACS(): error for "+
                                        "<Class>.<Value>.UserID "+l_asso[l_asso.length - 1]);

            // new directory comment
            factory.addAssociationsToNewObject(
                    "<Class>.<Value>.UserID",
                    new String[] {
                        "<CalledMethod>.<Action>.1.setComment",
                        "Comment.<Value>.This is normally reserved to"+
                                "the bridge connection "+
                                "of the same-name account or proxy from the right host."});

            // a new database has a VirtualFolder node 'base_name views',
            // with a null definer
            factory.addAssociationsToNewObject(
                    "Directory.ResourceType.<ValueContains>.database",
                    new String[] {
                        "<AssociateClass>.VirtualFolder",
                        "Name.<Value>.<MasterNameLastComponent>+ views",
                        "ACS.<Value>.<MasterValue>",
                        "Parent Node.<Value>.<ACSName>+ Views root::  ",
                        "Virtual Folder Type.<Value>.<immutable> root for a set",
                        "Definer.<Null>",
                        "Definer Host.<Null>"
            });
            
                            // for testing and tracking
                            l_asso = ((String[]) factory.getM_AssociationsOnNewObject().
                                    get("Directory.ResourceType.<ValueContains>.database"));
                            if ( l_asso != null && l_asso[l_asso.length - 1].contains("Error") )
                                throw new BaseError("AcsFactory.create(): error for "+
                                        "<AssociateClass>.VirtualFolder "+l_asso[l_asso.length - 1]);

            factory.addAssociationsToNewObject(
                    "Directory.ResourceType.<ValueContains>.database",
                    new String[] {
                        "<CalledMethod>.<Master>.setComment",
                        "Comment.<Value>.This database may have views in the virtual folders tree."
            });

            factory.addAssociationsToNewObject(
                    "Directory.ResourceType.<ValueContains>.database",
                    new String[] {
                        "<CalledMethod>.<Action>.1.setComment",
                        "Comment.<Value>.This is the root for the views of a given database"
            });

            factory.addCommentsOnProperties("Actions for association to a new object",
                    new String[] {"To understand this complex property,",
                                    " please read the documentation.",
                                    "",
                                    "-'GroupID.EPType.",
                                    " <ValueContains>.host'",
                                    " is a key to add a new host",
                                    " in the group 'AllHosts'.",
                                    "",
                                    "-'Directory.ResourceType.",
                                    " <ValueContains>.database'",
                                    " is a key which sets",
                                    " a script of 3 actions when",
                                    " a new database is created",
                                    " in MySQL.",
                                    "",
                                    "-The first action creates",
                                    " the single parent node",
                                    " for this-database views.",
                                    });

            // the image is under Accbee.MySQL.gBase.
            // caution: use the AcsAddon package editor and the AcsAddon name
            // in the ACS vocabulary.
            factory.setIcon("mysql.jpg");

            // sets the creation of the accounts, groups, directories and resources
            // roots of the MySQL ACS in its parent ACS
            //configureMySQLonLinux(_name, core, factory);

                    // for testing and tracking
                    //System.out.println("AcsFactory.createRBAC() "+
                      //"before the ACS creation of "+_name);

            // final check and creation
            factory.setTiedToParentACS(false); // may be changed
            acs = factory.createACS();

        } catch (CreateError c) {
            c.printStackTrace();
                // for testing and tracking
                throw new BaseError("From AcsFactoryMySQL.create(), exception: "+c);
        } catch (UpDateError u) {
            u.printStackTrace();
                // for testing and tracking
                throw new BaseError("From AcsFactoryMySQL.create(), exception: "+u);
        }

        factory = null;
        all = null;
        grant = null;
        alter = null;
        crea = null;
        creavw = null;
        creart = null;
        crtt = null;
        creatas = null;
        creauser = null;
        del = null;
        drop = null;
        ind = null;
        ins = null;
        sel = null;
        showvw = null;
        upd = null;
        exe = null;

        altrt = null;
        fil = null;
        trig = null;
        lock = null;
        proc = null;
        rld = null;
        evt = null;
        rpl = null;
        rpls = null;
        showdb = null;
        down = null;
        sup = null;
        no = null;

                    // for testing and tracking
                    //System.out.println("AcsFactoryMySQL.create() "+
                      //"before the ACS saving of "+_name);

        // saves the parent only, since the new ACS is saved in ActionNewACSyst
        // after the call to configure()
        //BaseManagerImpl.getInstance().saveOpenACS(core);
        return acs;
  } // create()
    
  /**
  * Configure MySQL server as an ACS.
  * The Resources first directory is 'MySQL server', with one database 'BASE One' and a
  * '|pattern-name_bases sets|' node for which 'BASES O%' is the single child.
  * The second directory without parent is the connection subACS, as parent of the user connections.
  * Under the database 'BASE One', two tables 'Customers table' and 'Products table' are created
  * with columns, indexes, triggers, stored functions and stored procedures.
  * The VirtualFolders root is 'Views root', and the associations create the passive
  * node 'database One views' as parent of all the 'database One' views.
  * The 3 standard GroupIDs to create under the 'standard_tree' are
  * '|pattern-name_accounts-for-hosts|', '|pattern-name_hosts|' and '|standard for simple hosts|'.
  * The only mandatory GroupID to create is 'localhost'. Creates also local%, '%',
  * and jerry§@llocal%.
  * Caution: only the type 'account' is correct for an UserID, even if the ACS policy allows more.
  * The created UserIDs are those found in the localhost, that is the parent ACS
  * if it is not a Linux ACS. Otherwise, only the accounts 'root', 'tom', 'jerry' and 'anonymous'
  * are created, with the proxy 'lp'.
  * 2 views 'Amounts view' and 'Prod view' are created.
  * Removes the default conditional AclEntry from 'root@localhost' to 'MySQL server',
  * and adds a new non-conditional AclEntry with the rights 'GRANT OPTION' and 'ALL'.
  * Compatibility note for an ACS derived from the version 0.7.0: no '<connection_subacs>' directory.
  * Called by ActionNewACSyst.fillWithAcsObjects().
  * @param _result to create
  * @return MySQL ACS with the initial and standard trees of resources and its
  *   eligible parties.
  */
  public ACSImpl configure(ACSImpl _result) {

                    // for testing and tracking
                    //System.out.println("call to AcsFactoryMySQL.configure()");

    ACS core = (ACS)UtilityImpl.getFirstInstance().getBaseObject(_result.getParentName());

    if (core == null) {
      throw new InternalError("AcsFactoryMySQL.configure() is called without any parent. ");
    }
    else if (!_result.getACSType().startsWith(
            PUBLIC_ACS_ADDON_NAME.substring(0, PUBLIC_ACS_ADDON_NAME.indexOf(" ("))) ) {

                                // for testing and tracking
                                System.out.println("AcsFactoryMySQL.configure() with error on "+
                                        "_result.getACSType(): "+_result.getACSType());

      throw new InternalError("AcsFactoryMySQL.configure() is not called for a MySQL 5 server. ");
    }

    StringRight[] l_right = new StringRight[] {
                                    core.getOneRight("|mdown|"),
                                    core.getOneRight("ctrl")
                                };

      
    // ---------------------------
    // creates the MySQL first components
    // --------------------------    

    //  creates MySQL directories, virtual folders, accounts and groups
    UserIDMySQLImpl     rootOne = null;
    UserIDImpl          tom = null;
    UserIDImpl          jerry = null;
    UserIDMySQLImpl     anon = null;
    ProxyMySQL          lp = null;
    GroupIDMySQLImpl    localhost = null;
    GroupIDMySQLImpl    host = null;
    GroupIDMySQLImpl    jerr = null;
    GroupIDMySQLImpl    localP = null;
    DirectoryMySQL      mysql_server = null;
    DirectoryMySQL      connections = null;
    VirtualFolderMySQL  views_root = null;
    AclEntry            acl = null;
    AclEntry            acl2 = null;

    ResourceMySQLImpl   rOne = null;
    ResourceMySQLImpl   purchases = null;
    ResourceMySQLImpl   sales = null;
    DirectoryMySQLImpl  dOne = null;
    DirectoryMySQLImpl  dTwo = null;
    DirectoryMySQLImpl  dThree = null;
    ActorMySQLImpl      aOne = null;
    VirtualFolderMySQL  base_root = null;
    VirtualFolderMySQL  vOne = null;
    
    try {
                // for testing and tracking
                //System.out.println("AcsFactoryMySQL.configure(): "+ "<connection_subacs>"

      mysql_server = new DirectoryMySQLImpl(
              ((Name)_result.getName()).extended("MySQL server"),
              _result, null, null, "MySQL server <byAcsCreator>");
      mysql_server.setComment("Its AclEntries define the global rights on the MySQL server"+
              " as a whole");

      // get <connection_subacs> directory, created through ACSFactoryImpl.createACS()
      connections = (DirectoryMySQL)_result.getEorM_OwnResources().get(
              ((Name) _result.getName()).extended("<connection_subacs>") );
      if (connections != null) // null if the ACS is derived from a 0.7.0 MySQL ACS
          connections.setComment("Parent of all the user connections to open a session "+
              "into the MySQL Server.");

      views_root = new VirtualFolderMySQLImpl(new NameImpl("Views root"),
              _result, null, "<immutable> no-child node", null, null);
      views_root.setComment("Parent of all the roots for database views; "+
              "such a root is set by the program when a database is created.");
    } // try
    catch (CreateError e) {
        e.printStackTrace();
        throw new InternalError(e.toString()+
              "AcsFactoryMySQL.configure(): "+e);
    }
    
    try {
        // creates the 3 standard groups

      // creates '|pattern-name_accounts-for-hosts|'
      host = new GroupIDMySQLImpl(((Name)_result.getName()).extended("|pattern-name_accounts-for-hosts|"),
              _result, "<byAcsCreator> standard pattern-name_accounts-for-hosts group", false);
      host.setComment("Set of all the pattern-name_accounts-for-hosts groups - Has no AclEntry.");
      host.setOrder(11);

      // creates '|pattern-name_hosts|'
      host = new GroupIDMySQLImpl(((Name)_result.getName()).extended("|pattern-name_hosts groups|"),
              _result, "<byAcsCreator> standard pattern-name_hosts group", false);
      host.setComment("Set of all the pattern-name_hosts groups - Has no AclEntry.");
      host.setOrder(14);

      // creates '|simple hosts|'
      host = new GroupIDMySQLImpl(((Name)_result.getName()).extended("|simple hosts|"),
              _result, "<byAcsCreator> standard for simple hosts", false);
      host.setComment("Set of the simple Hosts - Has no AclEntry.");
      host.setOrder(13);
      
        // creates the mandatory groups
      // creates the hosts group '%' to set a default ACL at the creation of an account
      localhost = new GroupIDMySQLImpl(((Name)_result.getName()).extended("%"),
              _result, "pattern-name_hosts group", false);
      localhost.setComment("All the simple hosts");
      localhost.setOrder(10);

      // creates localhost and local%
      localhost = new GroupIDMySQLImpl(((Name)_result.getName()).extended("localhost"),
              _result, "host <byAcsCreator>", false);
      localhost.setComment("Host for all the accounts from the parent server"+
              " which supports this MySQL server");
      localhost.setOrder(1);

      localP = new GroupIDMySQLImpl(((Name)_result.getName()).extended("local%"),
              _result, "pattern-name_hosts group", false);
      localP.setComment("All the hosts having a name that starts with 'local'.");

      // for a Linux parent, creates 'jerry§@local%'
      if (core.getACSType().indexOf("Linux Ubuntu") != -1) {
          jerr = new GroupIDMySQLImpl(((Name)_result.getName()).extended("jerry§@local%"),
                  _result, "pattern-name_accounts-for-hosts group", false);
          jerr.setComment("All the 'jerry' accounts in the 'local%' hosts. For demonstration purpose.");
      } // if

        // creates the users from the parent, with the right value of 'localhost',
      // if the parent is not a Linux Ubuntu
      int count = 5;
      if (core.getACSType().indexOf("Linux Ubuntu") == -1) {
          for (ImmutableEligibleParty ep : core.getEorM_EligibleParties().values()) {
              if (!(ep instanceof UserID))
                  continue;

              if (((UserID)ep).isAdmin()) {
                  rootOne = new UserIDMySQLImpl(((Name)_result.getName()).extended(ep.getNickName()),
                  _result, "account", true, localhost);
                  rootOne.setOrder(count);
                  AssociateUtilityImpl.addAssociatesAfterCreation(rootOne);
              }
              else {
                  tom = new UserIDMySQLImpl(((Name)_result.getName()).extended(ep.getNickName()),
                  _result, "account", false, localhost);
                  tom.setOrder(count);
                  AssociateUtilityImpl.addAssociatesAfterCreation(tom);
              } // else

              count++;
          } // for ep
      }
      else {
          
          rootOne = new UserIDMySQLImpl(((Name)_result.getName()).extended("root"),
                  _result, "account", true, localhost);
          AssociateUtilityImpl.addAssociatesAfterCreation(rootOne);
          
          // associates for the other accounts are made after the resources tree creation
          tom = new UserIDMySQLImpl(((Name)_result.getName()).extended("tom"),
                  _result, "account", false, localhost);
          tom.setOrder(5);
          jerry = new UserIDMySQLImpl(((Name)_result.getName()).extended("jerry"),
                  _result, "account", false, localhost);
          jerry.setOrder(6);
          anon = new UserIDMySQLImpl(((Name)_result.getName()).extended("anonymous"),
                  _result, "account", false, localhost);
          anon.setOrder(7);
          lp = new ProxyMySQLImpl(((Name)_result.getName()).extended("lp"),
                  _result, "proxy", false, localhost, (UserIDMySQLImpl)jerry);
          lp.setOrder(8);
      } // else if (core.getACSType().indexOf("Linux Ubuntu") == -1)

      // no action
      //AssociateUtilityImpl.addAssociatesAfterCreation(localhost);

    } catch (CreateError c) {
        c.printStackTrace();
      throw new InternalError("AcsFactoryMySQL.configure(): "+c);
    } catch (UpDateError u) {
        u.printStackTrace();
      throw new InternalError("AcsFactoryMySQL.configure(): "+u);
    } catch (BaseError b) {
        b.printStackTrace();
        throw new InternalError("After the creation of an account"+
                ", an association action throws an error: "+
                b.toString());
    } // catch
   
    try {

        // creates '|pattern-name_bases sets|' and 'BASE One', 'BASES O%'
      dOne = new DirectoryMySQLImpl(
              ((Name)mysql_server.getName()).extended("|pattern-name_bases sets|"),
              _result, mysql_server, null, "virtual_objects <byAcsCreator>");
      dOne.setComment("Directory of all the base sets, like 'BASES O%'.");

      // creates a bases set
      new BasesSetMySQLImpl(
              ((Name)dOne.getName()).extended("BASES O%"),
              _result, dOne, null, "pattern-name_bases set");

        // creates a database
      dOne = new DirectoryMySQLImpl(
              ((Name)mysql_server.getName()).extended("BASE One"),
              _result, mysql_server, null, "database");
      dOne.setComment("Default database, for demonstration purpose.");
      // creates the database groups and the views root
      AssociateUtilityImpl.addAssociatesAfterCreation(dOne);

      base_root = (VirtualFolderMySQL) _result.getOneVirtualFolder(
              ((Name)views_root.getName()).extended("BASE One views"));
      
      // dOne tables
      dTwo = new DirectoryMySQLImpl(
              ((Name)dOne.getName()).extended("Customers table"),
              _result, dOne, null, "table");
      dTwo.setComment("Default table 'customers', for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(dTwo);

      dThree = new DirectoryMySQLImpl(
              ((Name)dOne.getName()).extended("Products table"),
              _result, dOne, null, "table");
      dThree.setComment("Default table, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(dThree);

      // dTwo columns
      rOne = new ResourceMySQLImpl(
              ((Name)dTwo.getName()).extended("name column"),
              _result, dTwo, null, "table_column");
      rOne.setComment("Default column, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(rOne);

      rOne = new ResourceMySQLImpl(
              ((Name)dTwo.getName()).extended("address column"),
              _result, dTwo, null, "table_column");
      rOne.setComment("Default column, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(rOne);

      rOne = new ResourceMySQLImpl(
              ((Name)dTwo.getName()).extended("last_purchase_date column"),
              _result, dTwo, null, "table_column");
      rOne.setComment("Default column, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(rOne);

      purchases = new ResourceMySQLImpl(
              ((Name)dTwo.getName()).extended("purchases_amount column"),
              _result, dTwo, null, "table_column");
      purchases.setComment("Default column, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(purchases);

      // dThree columns
      rOne = new ResourceMySQLImpl(
              ((Name)dThree.getName()).extended("product_name column"),
              _result, dThree, null, "table_column");
      rOne.setComment("Default column, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(rOne);

      rOne = new ResourceMySQLImpl(
              ((Name)dThree.getName()).extended("index_one"),
              _result, dThree, null, "index");
      rOne.setComment("Default index, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(rOne);

      rOne = new ResourceMySQLImpl(
              ((Name)dThree.getName()).extended("product_price column"),
              _result, dThree, null, "table_column");
      rOne.setComment("Default column, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(rOne);

      sales = new ResourceMySQLImpl(
              ((Name)dThree.getName()).extended("sales_amount column"),
              _result, dThree, null, "table_column");
      sales.setComment("Default column, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(sales);

      // stored procedures and functions, triggers

      aOne = new ActorMySQLImpl(
              ((Name)dOne.getName()).extended("procedure_one"),
              _result, dOne, tom, "stored_procedure", localhost);
      aOne.setComment("Default stored procedure, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(aOne);

      if (core.getACSType().indexOf("Linux Ubuntu") != -1) {
          aOne = new ActorMySQLImpl(
                  ((Name)dOne.getName()).extended("procedure_two"),
                  _result, dOne, jerry, "stored_procedure", localhost);
          aOne.setComment("Default stored procedure, for demonstration purpose.");
          AssociateUtilityImpl.addAssociatesAfterCreation(aOne);

          aOne = new ActorMySQLImpl(
                  ((Name)dOne.getName()).extended("procedure_three"),
                  _result, dOne, anon, "stored_procedure",localhost);
          aOne.setComment("Default stored procedure, for demonstration purpose.");
          AssociateUtilityImpl.addAssociatesAfterCreation(aOne);

          aOne = new ActorMySQLImpl(
                  ((Name)dOne.getName()).extended("function_two"),
                  _result, dOne, null, "stored_function", null);
          aOne.setComment("Default stored function, for demonstration purpose.");
          AssociateUtilityImpl.addAssociatesAfterCreation(aOne);

          aOne = new ActorMySQLImpl(
                  ((Name)dThree.getName()).extended("trigger_one"),
                  _result, dThree, jerry, "trigger", localhost);
          aOne.setComment("Default trigger, for demonstration purpose.");
          AssociateUtilityImpl.addAssociatesAfterCreation(aOne);

          // creates the UserID associates after the default resources
          // creates the user connections under <connection_subacs>
          // uses the hosts group '%' in the created ACL
          AssociateUtilityImpl.addAssociatesAfterCreation(anon);
          AssociateUtilityImpl.addAssociatesAfterCreation(tom);
          AssociateUtilityImpl.addAssociatesAfterCreation(jerry);
          AssociateUtilityImpl.addAssociatesAfterCreation(lp);
      } // if (core.getACSType().indexOf("Linux Ubuntu") != -1) {

      aOne = new ActorMySQLImpl(
              ((Name)dOne.getName()).extended("function_one"),
              _result, dOne, tom, "stored_function", localhost);
      aOne.setComment("Default stored function, for demonstration purpose.");
      AssociateUtilityImpl.addAssociatesAfterCreation(aOne);

      // baseOne view 'Prod view' without definer
      vOne = new VirtualFolderMySQLImpl(new NameImpl("Prod view"),
              _result, base_root, "view", null, null);
      vOne.addEorBaseObject(rOne); // product_price column
      vOne.addEorBaseObject(sales); // sales_amount column

      // baseOne view 'Amounts view' with the last 'tom' as definer.
      // for a non-Linux parent, 'tom' may be null.
      vOne = new VirtualFolderMySQLImpl(new NameImpl("Amounts view"),
              _result, base_root, "view", tom, localhost);
      vOne.addEorBaseObject(purchases);
      vOne.addEorBaseObject(sales);

      // add a non-conditional AclEntry.
      // if the parent is not a Linux, only the last 'rootOne' value is processed
      acl = mysql_server.addAclEntry(_result, rootOne,
              new StringRight[] {
                    _result.getOneRight("GRANT OPTION"),
                    _result.getOneRight("ALL")},
              true, localhost);
      acl.setComment("This AclEntry delivers all rights to the local root. It is MANDATORY.");


      if (core.getACSType().indexOf("Linux Ubuntu") != -1) {

          // 2 default ACLs on 'BASE One'
          acl = dOne.addAclEntry(_result, jerry,
                  new StringRight[] {
                        _result.getOneRight("EXECUTE"),
                        _result.getOneRight("SELECT")},
                  true, localhost);
          acl.setComment("This default AclEntry is for demonstration purpose.");
          acl = dOne.addAclEntry(_result, jerr,
                  new StringRight[] {
                        _result.getOneRight("CREATE")},
                  true, null);
          acl.setComment("This default AclEntry is for demonstration purpose.");

          // ACLs for 'purchases_amount column', 'product_price column', 'sales_amount column'
          rOne = (ResourceMySQLImpl) _result.getEorM_Resources().get(
                  ((Name)dTwo.getName().clone()).extended("purchases_amount column"));
          acl = rOne.addAclEntry(_result, tom,
                  new StringRight[] {
                        _result.getOneRight("INSERT")},
                  true, localhost);
          acl.setComment("This default AclEntry is for demonstration purpose.");
          rOne = (ResourceMySQLImpl) _result.getEorM_Resources().get(
                  ((Name)dThree.getName().clone()).extended("product_price column"));
          acl = rOne.addAclEntry(_result, jerry,
                  new StringRight[] {
                        _result.getOneRight("INSERT")},
                  true, localhost);
          acl.setComment("This default AclEntry is for demonstration purpose.");
          rOne = (ResourceMySQLImpl) _result.getEorM_Resources().get(
                  ((Name)dThree.getName().clone()).extended("sales_amount column"));
          acl = rOne.addAclEntry(_result, jerry,
                  new StringRight[] {
                        _result.getOneRight("UPDATE")},
                  true, localhost);
          acl.setComment("This default AclEntry is for demonstration purpose.");

      } // if (core.getACSType().indexOf("Linux Ubuntu") != -1) {

      // a default ACL on 'BASE One' in all cases
      acl = dOne.addAclEntry(_result, localhost,
              new StringRight[] {
                    _result.getOneRight("EVENT")},
              true, null);
      acl.setComment("This default AclEntry is for demonstration purpose.");

    } catch (CreateError c) {
        c.printStackTrace();
      throw new InternalError("AcsFactoryMySQL.configure(): "+c);
    } catch (BaseError b) {
        b.printStackTrace();
        throw new InternalError("After the creation of an account"+
                ", an association action throws an error: "+
                b.toString());
    } // catch

        // after the Actor creation, to fill the no-proxy actors list
    BaseManagerImpl.getInstance().restoreACSInternalLinks(_result);

    rootOne = null;
    anon = null;
    tom = null;
    jerry = null;
    lp = null;
    localhost = null;
    host = null;
    jerr = null;
    localP = null;
    mysql_server = null;
    connections = null;
    views_root = null;
    acl = null;

    rOne = null;
    purchases = null;
    sales = null;
    dOne = null;
    dTwo = null;
    dThree = null;
    aOne = null;
    base_root = null;
    vOne = null;

    return _result;


    /*
     * THIS SET OF RESULTS IS INTENDED TO BE USED TO TEST.
     * LAST VERSION FROM 0.7.1.
     *
     * 1/5 - TEST OF THE PATHS BETWEEN
     * IO:: AA:: ubu:: mys:: <U>:: jerry::
     * IO:: AA:: ubu:: mys:: MySQL server:: BASE One::
     *
     *
FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: AA:: ubu:: mys:: <U>::  jerry
 TO  IO:: AA:: ubu:: mys:: MySQL server::  BASE One.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (EXECUTE SELECT Current User: (jerry, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
     *
     *
     * 2/5 - TESTS OF THE PATHS
     * IO:: svo:: lin:: mys:: <U>:: jerry::
     * IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Customers table:: purchases_amount column::
     * IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Products table:: product_price column::
     * IO:: svo:: lin:: mys:: MySQL server:: BASE One:: procedure_two::
     *
     *
FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Customers table::  purchases_amount column.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias
            Link number 2:
                  Is alias (Transfers the received rights <Sets Owner>: tom)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Amounts view
            Link number 3:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 4:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE
-- PATH NUMBER 4:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Products table::  product_price column.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (INSERT Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/EXECUTE/SELECT /USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 4:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Prod view alias
            Link number 2:
                  Is alias (Transfers the received rights)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Prod view
            Link number 3:
                  Virtual Member  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_two.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_two
 TO  IO:: four:: be:: sqyl3:: <U>::  jerry.
-- PATH NUMBER 1:
            Link number 1:
                  Runs Under  --  (no right)
     *
     *
     * 3/5 - TEST OF THE PATHS BETWEEN
 IO:: svo:: lin:: mys:: <G>:: localhost::
 IO:: svo:: lin:: mys:: <U>:: jerry::
 IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Amounts view alias::
 IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Customers table:: purchases_amount column::
 IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Prod view alias::
 IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Products table:: product_price column::
 IO:: svo:: lin:: mys:: Views root:: BASE One views:: Amounts view::
 IO:: svo:: lin:: mys:: Views root:: BASE One views:: Prod view::
     *
     * The view should generate the following SeeWhy text:
     *
FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: <G>::  localhost.
-- PATH NUMBER 1:
            Link number 1:
                  Is Member  --  (no right)
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_three
            Link number 2:
                  Runs Under  --  (no right)
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_two
            Link number 2:
                  Runs Under  --  (no right)
-- PATH NUMBER 4:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Products table::  trigger_one
            Link number 2:
                  Runs Under  --  (no right)
-- PATH NUMBER 5:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under  --  (no right)
-- PATH NUMBER 6:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under  --  (no right)
-- PATH NUMBER 7:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  Is Member  --  (no right)
-- PATH NUMBER 8:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  Is Member  --  (no right)
-- PATH NUMBER 9:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_three
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  anonymous
            Link number 3:
                  Is Member  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Customers table::  purchases_amount column.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Prod view alias.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Products table::  product_price column.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (INSERT Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/EXECUTE/SELECT /USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias
 TO  IO:: four:: be:: sqyl3:: <G>::  localhost.
-- PATH NUMBER 1:
            Link number 1:
                  Runs Under  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias
 TO  IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Amounts view.
-- PATH NUMBER 1:
            Link number 1:
                  Is alias (May sets Owner/Group: tom/localhost)  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Prod view alias
 TO  IO:: four:: be:: sqyl3:: <G>::  localhost.
-- PATH NUMBER 1:
            Link number 1:
                  May Run Under  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Prod view alias
 TO  IO:: four:: be:: sqyl3:: <U>::  jerry.
-- PATH NUMBER 1:
            Link number 1:
                  May Run Under  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Prod view alias
 TO  IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Prod view.
-- PATH NUMBER 1:
            Link number 1:
                  Is alias (Transfers the received rights)  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Amounts view
 TO  IO:: four:: be:: sqyl3:: <G>::  localhost.
-- PATH NUMBER 1:
            Link number 1:
                  May Run Under  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Amounts view
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Customers table::  purchases_amount column.
-- PATH NUMBER 1:
            Link number 1:
                  Virtual Member  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Prod view
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Products table::  product_price column.
-- PATH NUMBER 1:
            Link number 1:
                  Virtual Member  --  (no right)
     *
     *
     * 4/5 - TESTS OF THE PATHS
 IO:: svo:: lin:: mys:: <G>:: jerry§@local%::
 IO:: svo:: lin:: mys:: <U>:: jerry::
 IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Amounts view alias::
 IO:: svo:: lin:: mys:: Views root:: BASE One views:: Amounts view::
     *
     * note: jerry§@local% -> jerry via procedure_two and trigger_one
     * because CREATE contains EXECUTE
     *
FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <G>::  jerry§@local%
 TO  IO:: four:: be:: sqyl3:: <U>::  jerry.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (<inherited ACL> CREATE USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Products table::  trigger_one
            Link number 2:
                  Runs Under
-- PATH NUMBER 2:
            Link number 1:
                  ACL (<inherited ACL> CREATE USAGE <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_two
            Link number 2:
                  Runs Under

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <G>::  jerry§@local%
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (<inherited ACL> CREATE USAGE)  --  <WEAK>  -- All Rights: CREATE/USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: <G>::  jerry§@local%.
-- PATH NUMBER 1:
            Link number 1:
                  Is Member  --  (no right)
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Prod view alias
            Link number 2:
                  Is alias (Transfers the received rights)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Prod view
            Link number 3:
                  Runs Under  --   --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  --  <first rate link>  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias
 TO  IO:: four:: be:: sqyl3:: <G>::  jerry§@local%.
-- PATH NUMBER 1:
            Link number 1:
                  May Run Under  --  (no right)

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias
 TO  IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Amounts view.
-- PATH NUMBER 1:
            Link number 1:
                  Is alias (May sets Owner/Group: tom/localhost)  --  (no right)
     *
     *
     * 5/5 - TESTS OF THE PATHS
 IO:: svo:: lin:: mys:: <U>:: jerry::
 IO:: svo:: lin:: mys:: MySQL server:: BASE One:: Customers table:: purchases_amount column::
 IO:: svo:: lin:: mys:: MySQL server:: BASE One:: function_one::
     *
     * Note: 'tom' has no EXECUTE rights on 'Amounts view alias' that runs under 'tom'.
     * This is why 'function_one' running under 'tom' has no access to 'Amounts view alias'.
     * On the other hand, 'jerry' has access to 'Amounts view alias' through 'tom'.
     *
     *
FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Customers table::  purchases_amount column.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  Amounts view alias
            Link number 2:
                  Is alias (Transfers the received rights <Sets Owner>: tom)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: Views root:: BASE One views::  Amounts view
            Link number 3:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 4:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE
-- PATH NUMBER 3:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: <U>::  jerry
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one.
-- PATH NUMBER 1:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE)  -- All Rights: EXECUTE/SELECT/USAGE
-- PATH NUMBER 2:
            Link number 1:
                  ACL (Current User: (jerry, localhost) <inherited ACL> EXECUTE SELECT USAGE <Sets Group> <Sets Owner>)
            Intermediate node:
                   IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  procedure_one
            Link number 2:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 3:
                  ACL (Current User: (tom, localhost) <inherited ACL> USAGE)  -- All Rights: USAGE

FROM A RIGHT USER TO AN ACCESS TARGET
 FROM  IO:: four:: be:: sqyl3:: MySQL server:: BASE One::  function_one
 TO  IO:: four:: be:: sqyl3:: MySQL server:: BASE One:: Customers table::  purchases_amount column.
-- PATH NUMBER 1:
            Link number 1:
                  Runs Under
            Intermediate node:
                   IO:: four:: be:: sqyl3:: <U>::  tom
            Link number 2:
                  ACL (INSERT Current User: (tom, localhost) <inherited ACL> USAGE <conditional ACL> through localhost)  -- All Rights: INSERT/USAGE
     *
     */

  } // configure()
    
  /**
  * Currently unused. Creates the roots of the MySQL server in the parent ACS. They are only
  * opening and closing roots.
  * Called by create().
  * @param _name ImmutableName for the ACS to create.
  *   No test of null value nor empty.
  * @param _core is the ACS parent on which the MySQL server is building up, and
  * it is updated by the initial and standard trees of resources
  * and eligible parties of the MySQL server
  * @param _factory a ACSFactoryMySQLImpl instance
  */
  public static final void configureMySQLonLinux(
          ImmutableName _name, ACS _core, ACSFactoryImpl _factory) {

  } // configureMySQLonLinux()


} // AcsFactoryMySQL

