//Title:        Access Road
//Version:      0.7.1
//Copyright:    2011-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gWork;

import java.util.Map;
import java.util.HashSet;
import java.util.Arrays;

import ARoad0.CNot.*;
import ARoad0.ErrW.ProcessError;

import ARoad0.gBaseInterface.*;
import ARoad0.gWork.*;
import ARoad0.gWorkInterface.*;
import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;

/**
 * This class provides utilities about DisplayableLinkImpls on the MySQL server.
 * It operates on DisplayableLinkImpls and on the inherited rights.
 * An important service is the adding of a DisplayableLinkImpl
 * to the view map for these objects, and its filtering to follow the MySQL rules.
 * This is the responsibility of the method addDisplayableLink() to reject
 * 6 cases of DisplayableLinkImpls about the AclEntry sorting, the allowed targets
 * and the DEFINER mode of a VirtualFolder alias. <P>
 * The proper Actor AG context is required in the MySQL ACS, for any type of Actor
 * as intermediate node in an access path, each time there is a non-null current UserID
 * and there is an executing right on the Actor.
 * This is the responsability of the method setNextNodeAGrunningContext(). <P>
 * Each node in a view and each intermediate node has one LinkRights instance
 * for each view it belongs to. The implementation classes of this interface
 * have properties that depend only on the class of the node ACS,
 * then 2 nodes of the same view should use the same LinkRights instance,
 * even if it is not the actual behavior in all cases. But there is an exception
 * for the methods that detects the Account and AcsAddon priorities over all the links.
 * They use RightsMediatorImpl.getAllLinkRights() which returns
 * only one LinkRights instance per implied gWork package, and not one instance
 * per view node. <P>
 * Each instance of LinkRights is used by only one thread.
 * A node may be in several DisplayableLinks of the same view,
 * but it has only one LinkRights instance for the view.
 * The association between a node and its LinkRights
 * instance is managed by the RightsMediatorImpl instance of the view. <P>>
 * There is no direct call to the ACS.
 * This class is used by the gWork packages.
 * @see NodeRightsMySQLImpl
 * @version 0.7.1
 */
public class LinkRightsMySQLImpl extends LinkRightsImpl {

  /**
  * only one public constructor
  * @param _interpreter algorithm interpreter of this view
  */
  public LinkRightsMySQLImpl(AlgorithmInterpreter _interpreter) {
      super(_interpreter);
  }

  /**
  * This important method adds a DisplayableLinkImpl in the access paths map, if it does not exist,
  * or adds the rights of each AccessControlLink in the link, to the matched link in the map,
  * only if there is no collision between the new link and the previous ones.
  * Without this method, the rights from the wrong links would be added by the core algorithms
  * to the validated rights of the prioritary links.
  * This method overrides for MySQL the super method to reject a new link in 6 cases. <P>
  * The two first cases to detect is when the AclEntry sorting is not fulfilled.
  * First, because DisplayableLinkUtilities.getSecondEndAGcontext() returns
  * the accounts-for-hosts groups into the AG context of a link starting from an account,
  * it is necessary to forbid most of the links having such a group.
  * Secondly, it is necessary to forbid the adding of any link that contains the sequence
  * account/member/group/acl.../resource..., while a simple link account/member/group is correct.
  * The reason is because there are different types of links account/acl/resource
  * the NodeRightsMySQLImpl class may produce, implying varied membership relations
  * between the initial account and a group. Theses account/acl/resource links
  * are the only correct use of the membership relations to get MySQL rights on a resource. <P>
  * This is why an argument _dlink is rejected when one of the following conditions
  * are fulfilled: <P>
  * - _dLink contains, as intermediate node, an accounts-for-hosts group, <P>
  * - _dLink contains a sequence ...account/member/group/, then an ACL.  <P>
  * On the other hand, the initial sequence accounts-for-hosts group/acl/actor/run_under...
  * has to be kept since this is a weak but workable link.
  * A routine may call another routine. An argument _dlink is NOT rejected when
  * the following conditions are fulfilled: <P>
  * - _dLink contains a sequence (Actor routine)/aclink/UserID/, then an Actor. <P>
  * The third case to detect is for a link containing an Actor trigger
  * from a table T, and having as second end a resource in another table T, or
  * having as first end an actor not from the table T.
  * This is why an argument _dlink is rejected when the following conditions
  * are fulfilled: <P>
  * - _dLink is not simple, there is a node in _dLink that is a trigger from a table,
  *  and the second end is a Resource in another table or the first end is an Actor
  * not from the same table. <P>
  * The fourth case to detect is like the fourth one; it is for a link containing
  * an Actor view delegate having as first end a resource not in its view.
  * This is why an argument _dlink is rejected when the following conditions
  * are fulfilled: <P>
  * - _dLink is not simple, there is a node in _dLink that is a view delegate,
  *  and the second end is a Resource not in the view. <P>
  * The fifth case (new in 0.7.1) to detect is about the DEFINER mode of a VirtualFolder alias,
  * because an Actor view delegate having a current UserID has only its view as next node.
  * This is why an argument _dlink is rejected when the following conditions
  * are fulfilled: <P>
  * - _dLink is not simple, there is a node in _dLink that is a view delegate having a current UserID,
  *  and the next node is the delegate current UserID. <P>
  * The sixth case (new in 0.7.1) to detect is also about the DEFINER mode of a VirtualFolder alias,
  * because an Actor view delegate having a current UserID enforces its view to access
  * only through its current UserID, without direct 'virtual member' link.
  * The unique right access path is 'view delegate'/'view'/'DEFINER account'/'view member'.
  * Unfortunately, the reference to a virtual member relationship disappears in the comment.
  * This is why an argument _dlink is rejected when the following conditions
  * are fulfilled: <P>
  * - _dLink is not simple, there is a node in _dLink that is a view delegate having a current UserID,
  *  and the next two nodes are the view and one of its view element. <P>
  * Called by AgoRightsFactoryImpl, AclRightsFactoryImpl,
  * PrivilegeRightsFactoryImpl and CompoundRightsFactoryImpl.
  * Calls the overridden method at the end, if the new link is not rejected.
  * @see ARoad0.gWork.DisplayableLinkUtilities#getDisplayableLink(ARoad0.gBaseInterface.BaseObject, ARoad0.gBaseInterface.BaseObject, ARoad0.gBaseInterface.BaseObject[], java.util.Map) getDisplayableLink()
  * @param _dLink is the new link including some rights and to add to _upd_m_l_DisplayableLinks
  * @param _upd_m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair) to update.
  * @return DisplayableLinkImpl in _upd_m_l_DisplayableLinks,
  * which may be _dLink, or the similar link in _upd_m_l_DisplayableLinks,
  * after the rights merging but without merging of the properties. Null if the adding
  * is rejected.
  **/
  public DisplayableLinkImpl addDisplayableLink(
     DisplayableLinkImpl _dLink, Map _upd_m_l_DisplayableLinks) {

        if (_dLink.isSimple()) {
            return super.addDisplayableLink(_dLink, _upd_m_l_DisplayableLinks);
        } 

        // variables
        boolean         toForbid = false;
        boolean         withAccount = false;
        boolean         withHost = false;
        boolean         withTrigger = false;
        boolean         withViewDelegate = false;
        //boolean         withActor = false;
        BaseObject[]    l_nodes = _dLink.getEorAllNodes();
        boolean[]       l_Accounts = new boolean[l_nodes.length];
        boolean[]       l_Hosts = new boolean[l_nodes.length];
        boolean[]       l_Triggers = new boolean[l_nodes.length];
        boolean[]       l_ViewDelegates = new boolean[l_nodes.length];
        //boolean[]       l_Actors = new boolean[l_nodes.length];
        AccessControlLinkImpl       acLink = null;

                                            // for testing and tracking
                                            //System.out.println(
                                              //  "start of LinkRightsMySQLImpl.addDisplayableLink() for "+
                                                //_dLink);

        // set l_Triggers(, l_Actors), l_ViewDelegates, l_Accounts and l_Hosts
        for (int i = 0; i < l_nodes.length; i++) {
            // first case: contains, as intermediate node, an accounts-for-hosts group
            if (i != 0 && i != l_nodes.length - 1 &&
                    l_nodes[i] instanceof GroupIDMySQL &&
                    ((GroupIDMySQL)l_nodes[i]).getType().indexOf(
                        "accounts-for-hosts group") != -1) {

                                                    // for testing and tracking
                                                    //System.out.println(
                                                    //"LinkRightsMySQLImpl.addDisplayableLink() first case "+
                                                    //"at the node index: "+i+
                                                    //"- dLink to forbid "+_dLink);


                toForbid = true;
                break;
            }

            if (l_nodes[i] instanceof Actor &&
                    ((Actor)l_nodes[i]).getType().indexOf("trigger") != -1) {
                l_Triggers[i] = true;
                withTrigger = true;
                continue;
            }
            if (l_nodes[i] instanceof Actor) {
                //l_Actors[i] = true;
                //withActor = true;

                if (((Actor)l_nodes[i]).getType().indexOf("view_delegate") != -1) {
                    l_ViewDelegates[i] = true;
                    withViewDelegate = true;
                } // if
                
                continue;
            } // if

            if (!(l_nodes[i] instanceof GroupIDMember))
                continue;
            
            if (l_nodes[i] instanceof UserIDMySQL) { // also for ProxyMySQL
                l_Accounts[i] = true;
                withAccount = true;
                continue;
            } // if
            if (l_nodes[i] instanceof GroupID) {
                l_Hosts[i] = true;
                withHost = true;
            } // if
        } // for i


                                            // for testing and tracking
                                            //System.out.println(
                                              //  "LinkRightsMySQLImpl.addDisplayableLink() withAccount: "+
                                                //withAccount+ ", withHost: "+withHost);


        // second case: _dLink contains the sequence account/member/group/, then an ACL
        if (!toForbid && withAccount && withHost) {
            for (int i = 0; i < l_Accounts.length - 2; i++) { 
                if (!l_Accounts[i]) continue;

                                                // for testing and tracking
                                                //System.out.println(
                                                  //  "LinkRightsMySQLImpl.addDisplayableLink() first case "+
                                                    //"- from the account at : "+i);

                // acLink is after l_nodes[i]
                acLink = _dLink.getEorAllACLinks()[i];
                if (l_Hosts[i + 1] &&
                      (acLink.getLinkTypeValue(AccessControlLink.IS_MEMBER) ||
                        acLink.getLinkTypeValue(AccessControlLink.IS_INDIRECT_MEMBER))) {

                                                // for testing and tracking
                                                //System.out.println(
                                                  //  "LinkRightsMySQLImpl.addDisplayableLink() first case "+
                                                    //"- search ACL link from: "+
                                                    //_dLink.getEorAllACLinks()[i + 1]);

                            // search any ACL link after (i + 1)
                            for (int j = i+2; j < l_Accounts.length; j++) {
                                if (_dLink.getEorAllACLinks()[j - 1].getLinkTypeValue(
                                        AccessControlLink.ACL)) {

                                                // for testing and tracking
                                                //System.out.println(
                                                  //  "LinkRightsMySQLImpl.addDisplayableLink() second case "+
                                                    //"- dLink to forbid "+_dLink);

                                    toForbid = true;
                                    break;
                                } // if
                            } // for j

                            if (toForbid)
                                break;
                } // if (l_Hosts[i + 1]
            } // for i
        } // if (withAccount)

        // removed in 0.7.1: third case: _dLink contains, as initial sequence,
        // accounts-for-hosts group/acl/actor/run_under...
        //if (!toForbid && l_nodes.length > 2 &&
          //      ((ImmutableSource)l_nodes[0]).getType().indexOf(
            //            "accounts-for-hosts group") != -1 &&
              //  l_nodes[1] instanceof Actor) {

                //    if (_dLink.getEorAllACLinks()[0].getLinkTypeValue(
                  //          AccessControlLink.ACL) &&
                    //        _dLink.getEorAllACLinks()[1].getLinkTypeValue(
                      //      AccessControlLink.RUN_UNDER)) {
                        //toForbid = true;
                    //} // if
        //} // if (!toForbid)

        // third case:
        // _dLink is not simple, there is a node in _dLink that is a trigger from a table,
        // and the second end is not a child of this table or the first end is an Actor
        // not from the same table.
        if (!toForbid && withTrigger) {
            if (_dLink.getEorSecondEnd() instanceof Resource) {

                for (int i = 1; i < l_nodes.length - 1; i++) { // from 1 to l_nodes.length - 2
                    if (l_Triggers[i] &&
                            ((Actor)l_nodes[i]).getEorParent() !=
                                ((Resource)_dLink.getEorSecondEnd()).getEorParent()) {

                          toForbid = true;
                          break;
                    } // if
                 } // for i
            } // if (_dLink.getEorSecondEnd()
            if (!toForbid && _dLink.getEorFirstEnd() instanceof Actor) {

                for (int i = 1; i < l_nodes.length; i++) { // from 1 to l_nodes.length - 1
                    if (l_Triggers[i] &&
                            ((Actor)l_nodes[i]).getEorParent() !=
                                ((Actor)_dLink.getEorFirstEnd()).getEorParent()) {

                          toForbid = true;
                          break;
                    } // if
                 } // for i
            } // if (_dLink.getEorSecondEnd()
        } // if

        if (!toForbid && withViewDelegate) {
            // fourth case:
            // _dLink is not simple, there is a node in _dLink that is a MySQL view delegate,
            // and the second end is a Resource that is not in the MySQL view,
            // and it is not the view delegate.
            if (!toForbid && _dLink.getEorSecondEnd() instanceof Resource &&
                ((Resource)_dLink.getEorSecondEnd()).getType().indexOf("view_delegate") == -1) {

                if (!((Resource)_dLink.getEorSecondEnd()).getVirtualFolders().isEmpty()) {

                    for (int i = 1; i < l_nodes.length - 1; i++) { // from 1 to l_nodes.length - 2
                        if (l_ViewDelegates[i] &&
                                !((Resource)_dLink.getEorSecondEnd()).getVirtualFolders().contains(
                                    ((Actor)l_nodes[i]).getEorReferenceAsInternalHardAlias()) ) {

                                                      // for testing and tracking
                                                      //System.out.println(
                                                        //"LinkRightsMySQLImpl.addDisplayableLink(), "+
                                                        //"fourth case - to forbid because the view is not the VF of"+
                                                        //" the second end - view: "+
                                                        //((Actor)l_nodes[i]).getEorReferenceAsInternalHardAlias());


                              toForbid = true;
                              break;
                        } // if
                     } // for i
                } // if (_dLink.getEorSecondEnd()
                else {

                                                      // for testing and tracking
                                                      //System.out.println(
                                                        //"LinkRightsMySQLImpl.addDisplayableLink(), "+
                                                        //"fifth case - to forbid . ");

                    toForbid = true;
                } // else

            } // if (_dLink.getEorSecondEnd() instanceof Resource

            if (!toForbid ) {
                for (int i = 0; i < l_nodes.length - 1; i++) { // from 0 to l_nodes.length - 2
                    if (l_ViewDelegates[i] &&
                       ((Actor)l_nodes[i]).getEorCurrentUserID() != null) {
                        // fifth case:
                        // _dLink is not simple, there is a node in _dLink that is a view delegate
                        // having a current UserID,
                        //  and the nex node is the delegate current UserID
                            if (((Actor)l_nodes[i]).getEorCurrentUserID() == l_nodes[i+1]) {

                                                  // for testing and tracking
                                                  //System.out.println(
                                                    //"LinkRightsMySQLImpl.addDisplayableLink(), "+
                                                    //"sixth case - to forbid because "));


                                  toForbid = true;
                                  break;
                            } // if

                         // sixth case:
                        // _dLink is not simple, there is a node in _dLink that is a view delegate
                        // having a current UserID,
                        //  and the next two nodes are the view and one of its view element
                            if (i != l_nodes.length - 2 &&
                                    l_nodes[i+1] instanceof VirtualFolder &&
                                    ((Actor)l_nodes[i]).getEorReferenceAsInternalHardAlias() ==
                                        l_nodes[i+1] &&
                                    ((VirtualFolder)l_nodes[i+1]).containsBaseObject(
                                        l_nodes[i+2]) ) {

                                                  // for testing and tracking
                                                  //System.out.println(
                                                    //"LinkRightsMySQLImpl.addDisplayableLink(), "+
                                                    //"seventh case - to forbid because "));


                                  toForbid = true;
                                  break;

                            } // if (i != l_nodes.length - 2
                    } // if (l_ViewDelegates[i]
                 } // for i
            } // if (!toForbid)
        
        } // if (!toForbid && withViewDelegate

        l_nodes = null;
        l_Accounts = null;
        l_Hosts = null;
        acLink = null;
        l_ViewDelegates = null;
        
        if (toForbid) {

                                                // for testing and tracking
                                              //System.out.println(
                                                //    "in LinkRightsMySQLImpl.addDisplayableLink(), "+
                                                  //  "no adding of: "+_dLink);

            return null;
        }
        else {
            return super.addDisplayableLink(_dLink, _upd_m_l_DisplayableLinks);
        } // else
  } // addDisplayableLink()

  /**
   * This method increments the Account/Groups context position, updates the AG context set
   * in the DisplayableLinkImpl and returns this context.
   * The proper Actor AG context is required in the MySQL ACS, for any type of Actor
   * as intermediate node in an access path, each time there is a non-null current UserID
   * and there is an executing right on the Actor. There are two changes on the generic behavior. <P>
   * As a generic Access Road rule, the internal AclEntries propagate the direct AG context.
   * The received context overlays the AG context of the target, if it is not null.
   * The overlayed AG context is there the current context of an Actor, as an intermediate node
   * in an access path. In the MySQL ACS, the Actors are stored procedures and stored functions,
   * triggers and view delegates. They may have a current UserID and a current GroupID.
   * This method in this case does not the Actor AG context, and returns it
   * when the Actor is the second end.
   * In all these MySQL 4 cases, letting the Actor be a proxy
   * needs an AclEntry with an executing right on this Actor. <P>
   * There is a second change on the generic behavior.
   * If this Actor of _nodeIndex is the node before the second end with the same prerequisites,
   * this overriding method resets the proper AG context of the Actor.
   * Otherwise, this method calls the super method.
   * This process allows the generic paths search methods to find the new nodes
   * from the proper AG context of an Actor, then to give to these nodes their
   * complete AG context.
   * Called by LinkRightsImpl.updateAGrunningContext().
   * @see ARoad0.gWork.LinkRightsImpl#updateAGrunningContext(ARoad0.CNot.DisplayableLinkImpl)
   * @param _upd_dLink its second end is the node of this instance. Never null.
   * Its properties should be updated by this method,
   * and they are never null after the updating. The AG context position is set to
   * the node at _nodeIndex.
   * @param _nodeIndex index of the node to analyze; is the node of this instance.
   * Always greater than 0.
   * @return a clone of the updated AG context, that is a set of one applicalbe UserID and
   * some GroupIDs. May be null but never empty.
   * @exception ProcessError if the position of the node to analyze is lesser than 1,
   * and varied cases
   */
  protected HashSet<ImmutableGroupIDMember> setNextNodeAGrunningContext(
          DisplayableLinkImpl _upd_dLink, int _nodeIndex) throws ProcessError {

                                        // for testing and tracking
                                        //if (((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("Amounts view"))
                                          //  System.out.println(
                                            //    "LinkRightsMySQLImpl.setNextNodeAGrunningContext() "+
                                              //  " with _nodeIndex: "+_nodeIndex+
                                                //" for "+_upd_dLink);

    // reset the AG context if _nodeIndex is the index of the second end as Actor
      // having a current UserID
    if (_nodeIndex == _upd_dLink.noOfNodes() - 1 &&
            (_upd_dLink.getEorSecondEnd() instanceof Actor) &&
            ((Actor)_upd_dLink.getEorSecondEnd()).getEorCurrentUserID() != null ) {

                                    // for testing and tracking
                                    //System.out.println(
                                      //      "LinkRightsMySQLImpl.setNextNodeAGrunningContext()"+
                                        //    " ends with an Actor and a current UserID");


        AccessControlLinkImpl acLink = _upd_dLink.getEorLastLink();

        // an executing right on the second end is necessary to change the generic behavior
        if (!utility_.withDirectExecuteRight((StringRight[])acLink.getAllPositiveRights(false).
                toArray(new StringRight[0]))) {

                                        // for testing and tracking
                                        //System.out.println(
                                          //      "LinkRightsMySQLImpl.setNextNodeAGrunningContext()"+
                                            //    " has no 'execute': calls the super method ");

            acLink = null;
            return super.setNextNodeAGrunningContext(_upd_dLink, _nodeIndex);
        } // if

        Actor                       act = (Actor)_upd_dLink.getEorSecondEnd();
        HashSet<ImmutableGroupIDMember> l_newResult = null;
        ImmutableGroupID[]          l_Grps = null;

        l_newResult = new HashSet(10);

        l_newResult.add(act.getEorCurrentUserID());
        if (act.getEorCurrentGroupID() != null)
            l_newResult.add(act.getEorCurrentGroupID());

        l_Grps = act.getEorL_SecondaryGroup();
        if (l_Grps.length != 0)
            l_newResult.addAll(Arrays.asList(l_Grps));

        _upd_dLink.addProperty(
                DisplayableLinkImpl.AG_CONTEXT_SET, l_newResult.clone());

        _upd_dLink.addProperty(
                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION,
                new Integer(_nodeIndex));

                                        // for testing and tracking
                                        //System.out.println(
                                          //      "end of LinkRightsMySQLImpl.setNextNodeAGrunningContext()"+
                                            //    " with a new AG context: "+
                                              //  _upd_dLink.getProperties().
                                                //get(DisplayableLinkImpl.AG_CONTEXT_SET));

        act = null;
        l_Grps = null;
        acLink = null;
        return l_newResult;
    } // if (_nodeIndex == _upd_dLink.noOfNodes() - 1

    // may reset the AG context if there is an Actor just before the second end,
    // and the key AG_CONTEXT_NODE_POSITION is set
    if (_nodeIndex == _upd_dLink.noOfNodes() - 1 &&
            !_upd_dLink.isSimple() && _upd_dLink.getProperties() != null &&
            _upd_dLink.getProperties().get(DisplayableLinkImpl.AG_CONTEXT_SET) == null &&
            _upd_dLink.getEorNodeJustBeforeTheSecondEnd() instanceof Actor) {

                                        // for testing and tracking
                                        //System.out.println(
                                          //      "LinkRightsMySQLImpl.setNextNodeAGrunningContext()"+
                                            //    " with a null AG context. DLink to process.");


        Actor                       act = (Actor) _upd_dLink.getEorNodeJustBeforeTheSecondEnd();
        HashSet<ImmutableGroupIDMember> l_newResult = null;
        ImmutableGroupID[]          l_Grps = null;
        AccessControlLinkImpl       acLink = _upd_dLink.getEorAllACLinks()[_upd_dLink.noOfNodes() - 3];

        // control of 'execute' right on 'act'
        if (act.getEorCurrentUserID() != null &&
                utility_.withDirectExecuteRight((StringRight[])acLink.getAllPositiveRights(false).
                    toArray(new StringRight[0]))) {
            
            l_newResult = new HashSet(10);

            if (act.getEorCurrentUserID() != null)
                l_newResult.add(act.getEorCurrentUserID());
            if (act.getEorCurrentGroupID() != null)
                l_newResult.add(act.getEorCurrentGroupID());

            l_Grps = act.getEorL_SecondaryGroup();
            if (l_Grps.length != 0)
                l_newResult.addAll(Arrays.asList(l_Grps)); 

            _upd_dLink.addProperty(DisplayableLinkImpl.AG_CONTEXT_SET, l_newResult.clone());
        } // if (act.getEorCurrentUserID() != null)

        l_newResult = null;
        act = null;
        l_Grps = null;
        acLink = null;

        return super.setNextNodeAGrunningContext(_upd_dLink, _nodeIndex);

    }
    else {

                                    // for testing and tracking
                                    //System.out.println(
                                      //      "LinkRightsMySQLImpl.setNextNodeAGrunningContext()"+
                                        //    " directly calls the super method ");

        return super.setNextNodeAGrunningContext(_upd_dLink, _nodeIndex);
    } // else
  } // setNextNodeAGrunningContext()


} // LinkRightsMySQLImpl


