//Title:        Access Road
//Version:      0.7.2
//Copyright:    2010-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.MySQL.gWork;

import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import java.util.HashSet;
import java.util.Arrays;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.CNot.AccessControlLink;
import ARoad0.CNot.AccessControlLinkImpl;

import ARoad0.AcsAddon.Accbee.MySQL.gBaseInterface.*;
import ARoad0.gBaseInterface.*;
import ARoad0.gWork.*;
import ARoad0.gBase.NameImpl;

import ARoad0.AcsAddon.Accbee.MySQL.gBase.*;
import ARoad0.gWorkInterface.*;


/**
* This complex class is responsible for getting the rights for an MySQL node in a view,
* for the AcsAddon MySQL. It provides the proper handling of the ACL inherited rights. <P>
* The main methods of this class are presented now.
* This class overrides the method detectAddonFinalAllNodes() to add
* the properties AccessControlLink.C_FIST_RATE, AccessControlLink.C_WEAK_LINK and
* AccessControlLink.C_CONDITIONAL_ACL to some DisplayableLinkImpl. There are
* 3 cases, and the aim is to precise the effective rights in an Access Road view with MySQL nodes.
* The wrong access paths are handled by LinkRightsMySQLImpl.addDisplayablelink(),
* not by this class. <P>
* This class overrides the method detectL_aclEntryRights() to get
* all the rights of the relevant direct AclEntries managed by the MySQL Resource
* for a given EligibleParty, following the ACL sorting and the selection
* of the first matched ACL. <P>
* This class overrides the method getL_accessRightsThroughNodesTree()
* to return in an AccessControlLink the inherited rights
* that are given by the directory tree, as inherited rights on a resource.
* The result sets only the type AccessControlLink.ACL. <P>
* The new method getConnectionPair() returns the first current user, that is the pair
* (user, simple host or accounts-for-hosts group) under which the connection of _ep
* is done on the MySQL server. The requirements for the MySQL access controls are described
* in http://dev.mysql.com/doc/refman/5.6/en/request-access.html and the near pages. <P>
 * In the ACS MySQL, the effective rights of an user on a resource are selected
 * by a two-steps algorithm, derived from the MySQL specification, that is described hereinafter: <P>
 * At the first step of access control, only the ACL on the root 'MySQL server' are implied
 * to select the first pair (user, host or accounts-for-hosts group) that the user matches.
 * This pair (user, host or accounts-for-hosts group) is called the current user by MySQL.
 * It may be the initial pair (user, host) or not. For instance, all the effective rights
 * of a pair (jerry, localhost) may be provided by the current user ('jerry', '%').
 * One may notes that the MySQL term of current user includes the current group.
 * The AclEntries on the root 'MySQL server', for this first step of access control, may be: <P>
 * - conditional user/host ACL on the root 'MySQL server' EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on the root 'MySQL server'.  <P>
 * The ACL list is not totally ordered, since the accounts conditional ACL 
 * are not sorted if the accounts are not 'anomymous'.
 * If a right is granted at this first step (at least the 'USAGE' right),
 * some complementary rights may be added, for the SAME current user,
 * at each parent level from the database to the resource direct parent, as inherited ACL,
 * then other complementary rights may be added from the direct ACLs of the resource.
 * At each of these parent levels, and for the resource itself, into the sorted list
 * of inherited or direct ACLs, the FIRST ACL or the FIRST couple of ACL for
 * the current user, if any, defines the effective rights,
 * at this level, for the pair (user, host) on the resource. This is the sens of
 * the 'EXCLUSIVE OR' into each description of the two-steps algorithm.
 * This first-selected-single-output rule is also applied into each category of ACL,
 * like into the sorted list of the non-conditional 'accounts-for-hosts group' ACL 
 * for a resource. On the other hand, the current user may change along a given access path.<P>
 * As we have seen, if a right is granted at the first step on the root 'MySQL server',
 * some complementary rights may be added from the other levels: <P>
 * first from the database ACL or from an bases set ACL (both are called 'database-specific target'): <P>
 * - AND ((conditional user/host ACL on a database-specific target EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on a database-specific target)
 * EXCLUSIVE OR (user non-conditional ACL on a database-specific target AND
 * host non-conditional ACL on the database-specific target)), <P>
 * and secondly, some complementary rights may be also added, for the SAME current user
 * (user, host or group of hosts), from the rights on all the database children
 * which are parents of the resource or which is the resource itself: <P>
 * - AND (conditional user/host ACL on a table EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this table),  <P>
 * - AND (conditional user/host ACL on a column EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this column),  <P>
 * - AND (conditional user/host ACL on a routine EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this routine),  <P>
 * - AND (conditional user/host ACL on a view delegate EXCLUSIVE OR
 * non-conditional 'accounts-for-hosts group' ACL on this view delegate),  <P> <P>
 * Routines are the stored functions and the stored procedures.
 * Of course, the rights on a resource amy use the rights on itself and any
 * direct or indirect parent.
 * The rights on a database are more complex, since they use the ACL on the root,
 * on the related bases sets and on this database.
 * At each parent level (since there is the 'AND' logical operator between all of them),
 * including the root 'MySQL server' level for instance, only the first applicable ACL
 * is selected. This is why the ACL sorting into each ACL list is so important.
 * The only exception is at the database level. Two so-called to-be-mixed ACL
 * are used into the sentence (user non-conditional ACL on a database-specific target AND
 * host non-conditional ACL on the database-specific target) to produce the effective rights
 * from the database level. This is true only if the previous ACL at this level do not match,
 * because the ACL sorting puts first the conditional user/host ACL, then
 * the non-conditional 'accounts-for-hosts group' ACL, and so, the to-be-mixed ACL
 * are at the end of the sorted list. <P>
 * The following comments are about all the NodeRights in the AcsAddon framework.
 * The specialized methods of an AcsAddon are in a NodeRightsImpl subclass,
 * and they are called by an overriden NodeRights AcsAddon method;
 * there is no name pattern for these methods;
 * they use the generic properties of a node and its environment, but
 * they also handle the original properties of the AcsAddon node.
 * Since an AcsAddon may be based on another AcsAddon, a NodeRightsImpl
 * subclass may have another NodeRightsImpl subclass has super class. <P>
 * The only direct calls to the MySQL ACS in this class are ....
 * Upper-level methods begin by get...(), with...() and detect...(),
 * middle-level methods begin by select...().
 * @see ARoad0.gWork.NodeRightsImpl
 * @see ARoad0.AcsAddon.Accbee.MySQL.gWork.LinkRightsMySQLImpl#addDisplayableLink(ARoad0.CNot.DisplayableLinkImpl, java.util.Map)
 * @version 0.7.2
 */
public class NodeRightsMySQLImpl extends NodeRightsImpl {
  /**
   * The current user to add in the AccessControlLink comments.
   */
  public String                     CURRENT_USER = "Current User: (";
  /**
   * List of the accounts in the view
   */
  public List<UserIDMySQL>        l_accounts_ = new ArrayList<UserIDMySQL>(INITIAL_CAPACITY*5);
  /**
   * List of the actors in the view
   */
  public List<ActorMySQL>        l_actors_ = new ArrayList<ActorMySQL>(INITIAL_CAPACITY*5);
  /**
   * List of the hosts in the view
   */
  public List<GroupIDMySQL>         l_hosts_ = new ArrayList<GroupIDMySQL>(INITIAL_CAPACITY*5);
  /**
   * List of the groups in the view
   */
  public List<VirtualFolderMySQL>   l_vfolders_ = new ArrayList<VirtualFolderMySQL>(INITIAL_CAPACITY*5);

  /**
   * From the AlgorithmInterpreter
   */
  protected RightsFactory_Facade    facade_ = null;

  /**
   * For getL_accessRightsThroughNodesTree().
   */
  Set<StringRight>                  l_inheritedRightsResult_ = null;
  List<AclEntry>                    l_acl_for_one_level_ = null;

  /**
  * Only one constructor, called by RightsMediatorImpl. Calls the superclass constructor.
  * @param _interpreter algorithm interpreter of this view
  * @param _utility LinkRightsImpl or one of its subclass, to use for
  * processing this node
  */
  public NodeRightsMySQLImpl(AlgorithmInterpreter _interpreter, LinkRightsImpl _utility) {
      super(_interpreter, _utility);
      facade_ = (RightsFactory_Facade)interpreter_.getCoreAlgorithm("Facade");
      l_inheritedRightsResult_ = new HashSet(INITIAL_CAPACITY);
      l_acl_for_one_level_ = new ArrayList(INITIAL_CAPACITY);
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by
   * ThreeNodesRightsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilege().
   * @return true if the selectAddonLastViewNodeAfterGroupIDMember method provides some work 
   * in this class.
   */
  public boolean getSelectAddonLastViewNodeAfterGroupIDMember() {
      return true;
  }
  
   /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes, and by
   * ThreeNodesRightsFactoryImpl.addPathsFromGroupIDMemberAcsAddonRelationActor()
   * and addPathsFromGroupIDMemberAcsAddonRelationNoActor().
   * @return true since the selectAddonNewHiddenNodeForGroupIDMember method provides some work
   *  in this class.
   */
  public boolean getSelectAddonNewHiddenNodeForGroupIDMember() {
      return true;
  }

  /**
   * Called by RightsFactory_Facade.
   * @return true since the detectAddon2Nodes method provides some work in this class.
   */
  public boolean getDetectAddon2Nodes() {
      return true;
  }
  
   /**
   * Called by RightsFactory_Facade.
   * @return true since the detectAddonFinal3Nodes method provides some work in this class.
   */
  public boolean getDetectAddonFinal3Nodes() {
      return true;
  }

   /**
   * Called by RightsFactory_Facade.
   * @return true since the detectAddonFinalAllNodes method provides some work
   * in this class.
   */
  public boolean getDetectAddonFinalAllNodes() {
      return true;
  }

  /**
   * Detects the hidden actors that are linked to a GroupIDMember as access source,
   * to extend the current DisplayableLinkImpls this GroupIDMember ends.
   * Two lists are returned to separate the executing right-proxy nodes.
   * The last argument may be updated to provide the second returned value,
   * and it contains only the nodes on which the property changes have to be listened.<P>
   * This method covers two cases: <P>
   * (1) Extends AclRightsFactoryImpl.selectDirectHiddenAclBridgeLinksToActorForEP()
   * to the search of the inherited AclEntries, with the argument _withFinalExecute true.
   * With an internal ACL, the AG context is transferred by the core algorithms.
   * The method selectAddonNewHiddenNodeForGroupIDMember() is needed to detect
   * if the second end is running under its current UserID or GroupID.
   * MySQL actors are not ACL sources. There is no need to override
   * selectForGroupIDMemberItsHiddenActorsWithNextLinks() for detecting actors with
   * inherited AclEntries. <P>
   * (2) Adds an ALIAS link, when the GroupIDMember is a ProxyMySQL,
   * to its UserIDMySQL proxied account if it is not into the view.
   * To mimic the 'See Why' generic text, it adds the comment '-- All Rights: PROXY'. <P>
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * starting and central loops, to analyze the inherited ACLs.
   * Overrides and calls first the super method.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _node node of this instance. Never null.
   * @param _l_dLinks list of links with _node as second end. May be null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @param _upd_l_NoProxyOrNoExecuteNodes updated by the adding of the hidden nodes
   * without executing right or which are not right-proxy nodes, if any.
   * This list is only extended if necessary, as a complement of the returned value of the method.
   * This argument is usually empty at the call of this method, but this is not mandatory.
   * Never null.
   * @return new DisplayableLinkImpls that are extensions or updatings of some DisplayableLinkImpls
   * in _l_dLinks with an Actor or an UserID as the second end,
   * or null. If _l_dLinks is null, returns only simple links. May be null but never empty.
  * */
  public List<DisplayableLinkImpl> selectAddonNewHiddenNodeForGroupIDMember(
          EPRViewInBase _viewInBase, ImmutableGroupIDMember _node,
          Set<DisplayableLinkImpl> _l_dLinks, Map _m_l_DisplayableLinks,
          List<ACSObject> _upd_l_NoProxyOrNoExecuteNodes) {

    ActorMySQL          act = null;
    Set                 l_endOfLinks = null;
    DisplayableLinkImpl dLink = null;
    AccessControlLinkImpl   acLink = null;
    StringRight[]       l_rights = null;
    String              right = null;
    String              all = "ALL";
    String              create = "CREATE";
    String              execute = "EXECUTE";
    boolean             executing = false;
    List<DisplayableLinkImpl> l_result = super.selectAddonNewHiddenNodeForGroupIDMember(
          _viewInBase,  _node, _l_dLinks, _m_l_DisplayableLinks,
          _upd_l_NoProxyOrNoExecuteNodes);

    for (Iterator iti = _node.getEorACS().getEorM_OwnActors().values().iterator(); iti.hasNext() ; ) {
          act = (ActorMySQL) iti.next();


          if (_viewInBase.containsSource(act))
              continue;
          if (!act.getIsRightProxy()) {
              // act is already into _upd_l_NoProxyOrNoExecuteNodes
              continue;
          }

          // aclFact_.addAclBridgeRightsInSimpleDisplayableLink() is not useful there
          acLink = aclFact_.addAclInAccessControlLink(act, _node, acLink);

          if (acLink == null) {
              _upd_l_NoProxyOrNoExecuteNodes.add(act);
              continue;
          }

          l_rights = acLink.getRights(AccessControlLink.ACL);

          if (l_rights.length == 0) {
              _upd_l_NoProxyOrNoExecuteNodes.add(act);
              continue;
          }

          executing = false;
          for (int i = 0; i < l_rights.length; i++) {
              right = l_rights[i].getRight();

              // avoid utility_.withExecutingRights(), which it is more heavy
              if (right.equals(all) || right.equals(create) | right.equals(execute)) {
                  executing = true;
                  break;
              }
          } // for i
          if (!executing) {
              _upd_l_NoProxyOrNoExecuteNodes.add(act);
              continue;
          }

          dLink = new DisplayableLinkImpl(_node, act, acLink);

          // the hidden actor has always a next target: a view or a currentUser/Group.
          // the AG context has to be null, to allow the search of the next node
          // in CompoundRightsFactoryImpl.

          if (l_endOfLinks == null)
              l_endOfLinks = new HashSet(INITIAL_CAPACITY);
          l_endOfLinks.add(dLink);

                            // for testing and tracking
                            //System.out.println("in NodeRightsMySQLImpl."+
                              //"selectAddonNewHiddenNodeForGroupIDMember(), "+
                              //"detect the simple link as an extension for the current links:");
                            //System.out.println("----------- "+dLink);
                            //System.out.println("with a null AG context: "+
                              //          linkUtil_.getSecondEndAGcontext(dLink));

    } // for iti

    act = null;
    dLink = null;
    acLink = null;
    l_rights = null;
    right = null;

    // case 2: for a ProxyMySQL
    if (_node instanceof ProxyMySQL &&
            !_viewInBase.containsSource(((ProxyMySQL)_node).getProxiedAccount())) {

      StringRight proxy = _node.getEorACS().getOneRight("PROXY");

      acLink = new AccessControlLinkImpl();
      acLink.addRight(AccessControlLink.IS_ALIAS, proxy);
      acLink.addComment(
              AccessControlLink.IS_ALIAS,
              ((FactoryUtilities)interpreter_.getCoreAlgorithm("FactoryUtil")).
                getRightAcronym(proxy));

      dLink = new DisplayableLinkImpl(_node, ((ProxyMySQL)_node).getProxiedAccount(), acLink);

      if (l_endOfLinks == null)
          l_endOfLinks = new HashSet(INITIAL_CAPACITY);
      l_endOfLinks.add(dLink);

      proxy = null;
    } // if

    if (l_endOfLinks != null) {
        if (l_result == null) {
            if (_l_dLinks == null) {
                l_result = new ArrayList(l_endOfLinks);
            }
            else {
                l_result = new ArrayList(linkUtil_.
                        mergeTwoLinksCollection(_l_dLinks, l_endOfLinks));
            }
        } // if
        // unused, since l_result is null from the super method in this version
        else {
            if (_l_dLinks == null) {
                l_result.addAll(l_endOfLinks);
            }
            else {
                l_result.addAll(linkUtil_.
                        mergeTwoLinksCollection(_l_dLinks, l_endOfLinks));
            }
        } // else
    } // if (l_endOfLinks

    l_endOfLinks = null;
    return l_result;
  } // selectAddonNewHiddenNodeForGroupIDMember()

  /**
   * Detects the view target that is linked to a GroupIDMember, to extend
   * the current DisplayableLink ended by this GroupIDMember.
   * Adds an ALIAS link, when the GroupIDMember is a ProxyMySQL, 
   * if its UserIDMySQL proxied account is into the view.
   * To mimic the 'See Why' generic text, add the comment '-- All Rights: PROXY'.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by
   * ThreeNodesRighsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeAlias().
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _target view node, not the node of this instance. Never null.
   * @param _dLink with a GroupIDMember as second end and node of this instance. Never null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @return new DisplayableLinkImpl that is an extension of _dlink
   * with _target as the second node, or null. Never empty.
  * */
  public DisplayableLinkImpl selectAddonLastViewNodeAfterGroupIDMember(
          EPRViewInBase _viewInBase, ImmutableTarget _target,
          DisplayableLinkImpl _dLink, Map _m_l_DisplayableLinks) {

                        // for testing and tracking
                        //System.out.println(
                          // "Call to NodeRightsMySQLImpl.selectAddonLastViewNodeAfterGroupIDMember()");

      if (_target == null || !(_target instanceof UserIDMySQL) ||
              !_viewInBase.containsTarget(_target) || _dLink == null ||
              !(_dLink.getEorSecondEnd() instanceof ProxyMySQL) ||
              ((ProxyMySQL)_dLink.getEorSecondEnd()).getProxiedAccount() != _target)
                    return null;

      if (!((ACSObject)_dLink.getEorSecondEnd()).getAcsName().equals(_target.getAcsName()))
          return null;

      AccessControlLinkImpl acLink = new AccessControlLinkImpl();
      StringRight           proxy = _target.getEorACS().getOneRight("PROXY");

      acLink.addRight(AccessControlLink.IS_ALIAS, proxy);
      acLink.addComment(
              AccessControlLink.IS_ALIAS,
              ((FactoryUtilities)interpreter_.getCoreAlgorithm("FactoryUtil")).
                getRightAcronym(proxy));

      proxy = null;

      return DisplayableLinkImpl.valueOf(_dLink, _target, acLink);
  } // selectAddonLastViewNodeAfterGroupIDMember()

  /**
  * Adds and removes the AcsAddon specific rights: <P>
  * (1) adds a simple link to a MySQL proxied account
  * having one or several proxy accounts as a non-generic alias. <P>
  * (2) removes a simple link from an account when the current host for the connection
  * is an accounts-for-hosts group, and when this group is into the view,
  * since there is already a link from the account to this group, and another link
  * from this group to the target of the simple link to remove.<P>
  * Caution: if the current host for the connection is the simple host
  * of the account, this means the indirect path through the accounts-for-hosts group
  * is not an effective path. In this case, the method detectAddonFinalAllNodes()
  * sets an AccessControlLink.C_FIRST_RATE as GLOBAL comment on the direct path. <P>
  * The specific rights are searching for all sources in the view.
  * For a sketch view, this method may return new DisplayableLinks from _center to _node,
  * if any, but also, new DisplayableLinks from the view sources to _center
  * when _node is _center.
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @see #detectAddonFinal2Nodes
  * @param _viewInBase EPRViewInBase
  * @param _node node of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  *   with 'other' links associated to every pair (EP,Resource).
  * */
  public Map detectAddon2Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks, BaseObject _center) {

      ImmutableSource       sour = null;
      AccessControlLinkImpl acLink = null;
      StringRight           proxy = null;
      Set<DisplayableLinkImpl> l_dlinks = null;
      GroupIDMySQL          grp = null;
      List                  list = null;

      // case 1 on the proxied accounts
      if ((_node instanceof UserIDMySQL) &&
              ((UserIDMySQL)_node).getL_SoftAlias() != null) {

          proxy = _node.getEorACS().getOneRight("PROXY");
          for (Iterator ita = _viewInBase.getEorM_Sources().values().iterator() ; ita.hasNext() ; ) {
            sour = (ImmutableSource) ita.next();

            if (sour == _node || !(sour instanceof ProxyMySQL))
                continue;

            if (!((ProxyMySQL)sour).getProxiedAccount().getNickName().equals(_node.getNickName()))
                continue;

            if (_center != null) {
                if (_center != _node && _center != sour)
                    continue;
            } // if

            acLink = new AccessControlLinkImpl();

            acLink.addRight(AccessControlLink.IS_ALIAS, proxy);
            acLink.addComment(
                      AccessControlLink.IS_ALIAS,
                      ((FactoryUtilities)interpreter_.getCoreAlgorithm("FactoryUtil")).
                        getRightAcronym(proxy));

            link_.addDisplayableLink(
                     new DisplayableLinkImpl(sour, _node, acLink),
                     _m_l_DisplayableLinks);
          } // for ita
          sour = null;
          acLink = null;
          proxy = null;
      } // ((UserIDMySQL)_node).getL_SoftAlias() != null

      // case 2 on the accounts-for-host group connections
      if (_node instanceof ResourceMySQL) {
          l_dlinks = linkUtil_.selectDisplayableLinksForSecondEnd(
                                            _node, _m_l_DisplayableLinks);

          if (l_dlinks != null) {
              for (DisplayableLinkImpl link : l_dlinks) {
                    if (!link.isSimple() || !(link.getEorFirstEnd() instanceof UserIDMySQL))
                        continue;
                    grp = (GroupIDMySQL)getConnectionPair(
                            (UserIDMySQL)link.getEorFirstEnd() ).get(1);
                    if (grp.getType().indexOf("accounts-for-hosts group") == -1 ||
                            !_viewInBase.containsBaseObject(grp))
                        continue;

                                                // for testing and tracking
                                                //System.out.println(
                                                  //"NodeRightsMySQLImpl.detectAcsAddon2Nodes(), on "+
                                                  //_node.getName()+" to remove "+link);

                    // remove link
                    list = (List)_m_l_DisplayableLinks.get(
                            ((Name)((UserIDMySQL)link.getEorFirstEnd() ).
                                getName()).extended(_node.getName()));

                                                // for testing and tracking
                                                //for (Object ob : list) {
                                                  //  System.out.println("--- "+ob);
                                                //}
                                                //boolean rem = list.remove(link);
                    list.remove(link);

                    if (list.isEmpty())
                        _m_l_DisplayableLinks.remove(
                            ((Name)((UserIDMySQL)link.getEorFirstEnd() ).
                                getName()).extended(_node.getName()));

                                                // for testing and tracking
                                                //System.out.println(
                                                  //"NodeRightsMySQLImpl.detectAcsAddon2Nodes(), on "+
                                                  //_node.getName()+" after removing: "+rem);

              } // for link
          } // if (l_dlinks != null)
          l_dlinks = null;
          grp = null;
          list = null;
      } // if (_node instanceof ResourceMySQL)
      
      return _m_l_DisplayableLinks;
  } // detectAcsAddon2Nodes()


  /**
  * Removes some 3-nodes DisplayableLinkImpls: <P>
  * (1) from an Actor which is a view delegate (then has a VirtualFolder alias),
  * to a Resource the same VirtualFolder does not have as element. This case occurs
  * only when the Actor has a current UserID (the MySQL DEFINER account of the view). <P>
  * (2) from an Actor which is a trigger, to a Resource that is not a child
  * of the Actor parent, that is the MySQL table of the trigger. <P>
  * The stored procedures and the stored functions may have access
  * to any MySQL Resource outside their database.
  * Called by RightsFactory_Facade.
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @see #detectAddon3Nodes
  * @param _viewInBase EPRViewInBase
  * @param _node node of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @return Map of DisplayableLinks
  * @exception InterruptedException
   */
  public Map detectAddonFinal3Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks)  throws InterruptedException {

      if (!(_node instanceof Resource))
          return _m_l_DisplayableLinks;

      ImmutableSource       sour = null;
      Name                  pairKey = null;
      VirtualFolder         vFolder = null;

      for (Iterator ita = _viewInBase.getEorM_Sources().values().iterator() ; ita.hasNext() ; ) {
        sour = (ImmutableSource) ita.next();

        if (!(sour instanceof Actor))
            continue;
        
        if (((Actor)sour).getType().indexOf("view_delegate") != -1 &&
                ((Actor)sour).getEorCurrentUserID() != null) {

            vFolder = (VirtualFolder)((Actor)sour).getEorReferenceAsInternalHardAlias();

            if (vFolder.getEorM_BaseObjects().keySet().contains(_node.getName()))
                continue;

            if (!linkUtil_.containsDisplayableLinks(sour, _node, _m_l_DisplayableLinks))
                    continue;

            pairKey = ((Name)sour.getName()).extended(_node.getName());
            _m_l_DisplayableLinks.remove(pairKey);
        } // if (((Actor)sour).getType().indexOf("view_delegate")       
        else if (((Actor)sour).getType().indexOf("trigger") != -1) {

            if (Arrays.asList(
                    ((Directory)((Actor)sour).getEorParent()).getEorL_Children()).
                    contains((ImmutableResource)_node))
                continue;

            if (!linkUtil_.containsDisplayableLinks(sour, _node, _m_l_DisplayableLinks))
                    continue;

            pairKey = ((Name)sour.getName()).extended(_node.getName());
            _m_l_DisplayableLinks.remove(pairKey);
        } // if (((Actor)sour).getType().indexOf("trigger")

      } // for ita

      if (Thread.interrupted()) {
        throw new InterruptedException();
      }

      sour = null;
      pairKey = null;
      vFolder = null;
      return _m_l_DisplayableLinks;
  } // detectAddonFinal3Nodes()

  /**
   * This method adds 4 types of GLOBAL comments on some access paths. <P>
  * (1) Detecting when the user/target path 'overrides' or 'underrides' all the other paths,
  * this method puts AccessControlLink.C_FIRST_RATE as GLOBAL comment
  * in the simple DisplayableLinkImpl for the couple (user, target).
  * The aim is to display the user/target arrow with a double width,
  * while there is also a link for the couple (accounts-for-hosts group, target). <P>
  * (2) Detecting when the actor/target path 'overrides' or 'underrides' all the other paths,
  * this method puts AccessControlLink.C_FIRST_RATE as GLOBAL comment
  * of the last link into the non-simple DisplayableLinkImpl for the couple (actor, target).
  * The aim is to display the user/target arrow with a double width,
  * while there is also a link for the couple (accounts-for-hosts group, target). <P>
  * (3) This method puts AccessControlLink.C_WEAK_LINK as GLOBAL comment
  * in the first DisplayableLinkImpl for each couple (accounts-for-hosts group, target),
  * while there is no couple (host, target) or (hosts group, target). <P>
  * (4) This method puts AccessControlLink.C_WEAK_LINK as GLOBAL comment
  * in the first DisplayableLinkImpl for each couple (virtual folder, virtual member)
  * if the alias of the virtual folder runs under an account into the view.
  * This is the MySQL DEFINER mode, where the account defines the effective rights
  * on the virtual folder member. There is then a direct path from the account to the member,
  * and the path (virtual folder, member) is weak. <P>
  * The 3 cases are detailled hereinafter. <P>
  * For the MySQL ACS only, the simple access paths user/resource may be misunderstood.
  * The MySQL algorithm includes the access rule
  * '(one user privilege on database AND one host privilege on database)'.
  * The case to handle is when an Access Road view contains an UserID, some host GroupIDs
  * and, as target, a database or one of its child. The dadabase child inherits
  * the database ACLs. The path user/(database or database child)
  * may have to be processed if the user is (indirect or not) member of a host,
  * while the host has non-conditional ACL rights on the database.
  * The usual reading of the Access Road view tells us the user
  * has some rights on the target through the host(s). But this reading may be wrong,
  * since the host non-conditional ACL is fully applicable for the user only if there is
  * another non-conditional ACL, with at least the same rights, from the user to the target.
  * If it is not the case, it is necessary to indicate that the user/target arrow
  * 'overrides' the multiple-arrows path user/host(s)/target, in the Access Road view.
  * The user/target arrow must then be displayed with a double-width arrow.
  * This is the generic way to avoid a wrong reading of the view.
  * If this occurs for a MySQL target in an Access Road view, the method
  * detectAddonFinalAllNodes() provides the following actions: <P>
  * - if there is a simple DisplayablelinkImpl from the user to the target,
  * the property AccessControlLink.C_FIST_RATE is set on it,  <P>
  * - otherwise, this simple DisplayablelinkImpl is added with a desabled right,
  * that is the MySQL USAGE right. <P>
  * This method adds the constant AccessControlLink.C_WEAK_LINK
  * as GLOBAL comment in the first DisplayableLinkImpl for the couples (host, target)
  * and (user, target), when the rights are weak since they are not considered
  * in the context of an user or they are not effective MySQL rights. <P>
  * Note: to speed up this method, the argument _m_effectiveRights is set from
  * the initial argument _m_l_DisplayableLinks before the first call to this method,
  * and for a given EPR view. _m_effectiveRights is then immutable while
  * _m_l_DisplayableLinks may be updated by this method. This is not the case
  * for the NoThan views.  For the EPR views, this behavior implies this method
  * provides independant changes for each _node for which it is called,
  * so that the initial _m_effectiveRights remains usable.
  * Called by RightsFactory_Facade.
  * @param _viewInBase EPRViewInBase
  * @param _node node of this instance
  * @param _m_effectiveRights non-null Map where the keys are the _m_l_DisplayableLinks keys
  * where rights are activated, and the value is a set of StringRights. May be an empty set.
  * @return Map of DisplayableLinks
  * @exception InterruptedException
   */
  public Map detectAddonFinalAllNodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node, Map _m_l_DisplayableLinks,
        Map<ImmutableName, Set<StringRight>> _m_effectiveRights) throws InterruptedException {
                    
                                        // for testing and tracking
                                        //System.out.println("start of "+
                                          //"NodeRightsMySQLImpl.detectAddonFinalAllNodes(), on "+
                                          //_node.getName());

      if (!(_node instanceof ResourceMySQL))
          return _m_l_DisplayableLinks;

      l_accounts_.clear();
      l_actors_.clear();
      l_hosts_.clear();
      l_vfolders_.clear();

      // simple link from an account to _node
      DisplayableLinkImpl   simpleMainLink = null;
      // non-simple link from an actor to _node
      DisplayableLinkImpl   mainLink = null;
      ImmutableSource       sour = null;
      boolean               containsMainPairKey = false;
      boolean               containsPairKeyOne = false;
      boolean               containsPairKeyTwo = false;
      Set<StringRight>      l_mainRights = null;
      Set<StringRight>      l_hostRights = null;
      StringRight[]         array_hostRights = null;

                                    // for testing and tracking
                                    //System.out.println("call to "+
                                      //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes()"+
                                        //  " for _node "+_node.getNickName());
                                    //System.out.println("m_rights in "+
                                      //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes():");
                                    //for (Map.Entry entry : m_rights.entrySet()) {
                                      //  System.out.println("--- for the key "+entry.getKey());
                                        //for (StringRight right : (Set<StringRight>)entry.getValue()) {
                                          //  System.out.println("--------- "+right);
                                        //}
                                    //} // for entry


      // fills up l_accounts_, l_vfolders_ and l_hosts_. _node is never selected.
      for (Iterator ita = _viewInBase.getEorM_Sources().values().iterator() ; ita.hasNext() ; ) {
          sour = (ImmutableSource) ita.next();
          if (sour == _node  || sour.getEorACS() != _node.getEorACS())
              continue;

          if (sour instanceof UserIDMySQL) {
              l_accounts_.add((UserIDMySQL)sour);
          }
          else if (sour instanceof ActorMySQL) {
              l_actors_.add((ActorMySQL)sour);
          }
          else if (sour.getType().indexOf("accounts-for-hosts group") != -1) {
              l_hosts_.add((GroupIDMySQL)sour);
          }
          else if (sour instanceof VirtualFolderMySQL) {
              l_vfolders_.add((VirtualFolderMySQL)sour);
          }
      } // for ita

      // case 1: AccessControlLink.C_FIRST_RATE as GLOBAL comment
        // in the simple DisplayableLinkImpl for the couple (user, target)
        // if there is a path (user, accounts-for-hosts group)
        // and a path (accounts-for-hosts group, target)
      for (UserIDMySQL user : l_accounts_) {

                                            // for testing and tracking
                                        //System.out.println(
                                          //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                            //  "in the l_accounts_ loop for user: "+
                                              //user.getNickName());

            containsMainPairKey = linkUtil_.containsDisplayableLinks(
                user, _node, _m_l_DisplayableLinks);
            if (!containsMainPairKey) continue;

            // may be null
            simpleMainLink = linkUtil_.getDisplayableLink(
                  user, _node, null, _m_l_DisplayableLinks);
            if (simpleMainLink == null) continue;

                                    // for testing and tracking
                                    //if (simpleMainLink != null)
                                        //System.out.println(
                                          //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                            //  "for user, simpleMainLink.getEorFirstLink(): "+
                                              //simpleMainLink.getEorFirstLink());

            // may be null?
            l_mainRights = _m_effectiveRights.get(
                  ((Name)user.getName()).extended(_node.getName()) );
            if (l_mainRights == null) continue;

            for (GroupIDMySQL grp : l_hosts_) {

                containsPairKeyOne = linkUtil_.containsDisplayableLinks(
                    user, grp, _m_l_DisplayableLinks);
                if (!containsPairKeyOne) {
                  continue;
                } // else

                containsPairKeyTwo = linkUtil_.containsDisplayableLinks(
                    grp, _node, _m_l_DisplayableLinks);
                if (!containsPairKeyTwo) {
                  continue;
                } // else

                l_hostRights = _m_effectiveRights.get(
                      ((Name)grp.getName()).extended(_node.getName()) );
                array_hostRights = l_hostRights.toArray(new StringRight[0]);

                                    // for testing and tracking
                                    //if (array_hostRights != null) {
                                      //  System.out.println(
                                        //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                          //  " array_hostRights:");
                                        //for (int i = 0; i < array_hostRights.length; i++) {
                                          //  System.out.println("--- "+array_hostRights[i]);
                                        //}
                                    //}

                // never null
                //firstHostLink = (DisplayableLinkImpl)((List)_m_l_DisplayableLinks.get(
                  //    ((Name)grp.getName()).extended(_node.getName()))).iterator().next();

                                    // for testing and tracking
                                    //System.out.println(
                                      //"NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                      //"set a weak link from "+host.getName()+
                                      //" to "+_node.getName());

                if (array_hostRights != null) {
                    
                                        // for testing and tracking
                                        //System.out.println(
                                          //"NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                          //"set a first-rate link from "+user.getName()+
                                          //" to "+_node.getName());

                      //if (simpleMainLink != null) {
                          simpleMainLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                          simpleMainLink.getEorLastLink().addComment(
                                  AccessControlLink.GLOBAL, AccessControlLink.C_FIRST_RATE);
                          continue;
                      //}
                      //else {
                        //  acLink = new AccessControlLinkImpl();
                          //acLink.addComment(
                            //      AccessControlLink.GLOBAL,
                              //    AccessControlLink.C_FIRST_RATE);
                          //acLink.addRight(
                            //      AccessControlLink.GLOBAL,
                              //    _node.getEorACS().getOneRight("USAGE"));
                          //link_.addDisplayableLink(
                            //      new DisplayableLinkImpl(user, _node, acLink),
                              //    _m_l_DisplayableLinks);
                      //} // else
                } // if (array_hostRights != null

            } // for grp
      } // for user

      // case 2: AccessControlLink.C_FIRST_RATE as GLOBAL comment in the last link
        // into the non-simple DisplayableLinkImpl for the couple (actor, target)
        // if there is a path (actor, accounts-for-hosts group)
        // and a path (accounts-for-hosts group, target)
      for (ActorMySQL actor : l_actors_) {

                                            // for testing and tracking
                                        //System.out.println(
                                          //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                            //  "in the l_actors_ loop for user: "+
                                              //user.getNickName());

            containsMainPairKey = linkUtil_.containsDisplayableLinks(
                actor, _node, _m_l_DisplayableLinks);
            if (!containsMainPairKey) continue;

            mainLink = (DisplayableLinkImpl)((List)_m_l_DisplayableLinks.get(
                    ((Name)actor.getName()).extended(_node.getName()))).get(0);
            if (mainLink == null)
                throw new InternalError("NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                        "wrong second case on the actor "+actor.getName());

                                    // for testing and tracking
                                    //if (mainLink != null)
                                        //System.out.println(
                                          //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                            //  "for actor, mainLink.getEorFirstLink(): "+
                                              //mainLink.getEorFirstLink());

            // may be null?
            l_mainRights = _m_effectiveRights.get(
                  ((Name)actor.getName()).extended(_node.getName()) );
            if (l_mainRights == null) continue;

            for (GroupIDMySQL grp : l_hosts_) {

                containsPairKeyOne = linkUtil_.containsDisplayableLinks(
                    actor, grp, _m_l_DisplayableLinks);
                if (!containsPairKeyOne) {
                  continue;
                } // else

                containsPairKeyTwo = linkUtil_.containsDisplayableLinks(
                    grp, _node, _m_l_DisplayableLinks);
                if (!containsPairKeyTwo) {
                  continue;
                } // else

                l_hostRights = _m_effectiveRights.get(
                      ((Name)grp.getName()).extended(_node.getName()) );
                array_hostRights = l_hostRights.toArray(new StringRight[0]);

                                    // for testing and tracking
                                    //if (array_hostRights != null) {
                                      //  System.out.println(
                                        //    "NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                          //  " array_hostRights:");
                                        //for (int i = 0; i < array_hostRights.length; i++) {
                                          //  System.out.println("--- "+array_hostRights[i]);
                                        //}
                                    //}

                // never null
                //firstHostLink = (DisplayableLinkImpl)((List)_m_l_DisplayableLinks.get(
                  //    ((Name)grp.getName()).extended(_node.getName()))).iterator().next();

                                    // for testing and tracking
                                    //System.out.println(
                                      //"NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                      //"set a weak link from "+host.getName()+
                                      //" to "+_node.getName());

                if (array_hostRights != null) {
                    
                                        // for testing and tracking
                                        //System.out.println(
                                          //"NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                          //"set a first-rate link from "+actor.getName()+
                                          //" to "+_node.getName());

                          mainLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                          mainLink.getEorLastLink().addComment(
                                  AccessControlLink.GLOBAL, AccessControlLink.C_FIRST_RATE);
                          continue;
                } // if (array_hostRights != null

            } // for grp
      } // for actor


      // case 3 : AccessControlLink.C_WEAK_LINK as GLOBAL comment
        // in the simple DisplayableLinkImpl for each couple (accounts-for-hosts group, target)
        // if such simple link exists.
      if (!l_hosts_.isEmpty()) {

            for (GroupIDMySQL host : l_hosts_) {
                containsMainPairKey = linkUtil_.containsDisplayableLinks(
                    host, _node, _m_l_DisplayableLinks);
                if (!containsMainPairKey) continue;

                // may be null
                simpleMainLink = linkUtil_.getDisplayableLink(
                      host, _node, null, _m_l_DisplayableLinks);

                                        // for testing and tracking
                                        //if (simpleMainLink != null)
                                          //  System.out.println(
                                            //      "NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                              //    "for host, simpleMainLink.getEorFirstLink(): "+
                                                //  simpleMainLink.getEorFirstLink());

                if (simpleMainLink == null) continue;

                                        // for testing and tracking
                                        //System.out.println(
                                          //"NodeRightsMySQLImpl.detectAddonFinalAllNodes(), "+
                                          //"set a weak link from "+host.getName()+
                                          //" to "+_node.getName());

                  // set the (host, _node) link as a weak link
                simpleMainLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                simpleMainLink.getEorFirstLink().addComment(
                          AccessControlLink.GLOBAL,
                          AccessControlLink.C_WEAK_LINK);

            } // for host
      } // if (!l_hosts_.isEmpty())

      // case 4: AccessControlLink.C_WEAK_LINK as GLOBAL comment
        // in the simple DisplayableLinkImpl for each couple (virtual folder, virtual member)
        // if such simple link exists in the DEFINER mode with the current user in the view.
      for (VirtualFolderMySQL vf : l_vfolders_) {

          if (((Actor)vf.getEorAliasAsInternalHardReference()).getEorCurrentUserID() == null)
              continue;

          if (!_viewInBase.containsBaseObject(
                  ((Actor)vf.getEorAliasAsInternalHardReference()).getEorCurrentUserID() ))
              continue;
          
          // if it is the MySQL DEFINER mode, set a weak link
          containsMainPairKey = linkUtil_.containsDisplayableLinks(
                vf, _node, _m_l_DisplayableLinks);
          if (!containsMainPairKey) continue;

          simpleMainLink = linkUtil_.getDisplayableLink(
                  vf, _node, null, _m_l_DisplayableLinks);
          if (simpleMainLink == null) continue;

          // set the (virtual folder, _node) link as a weak link
          simpleMainLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
          simpleMainLink.getEorFirstLink().addComment(
                  AccessControlLink.GLOBAL,
                  AccessControlLink.C_WEAK_LINK);

      } // for vf

      sour = null;
      l_mainRights = null;
      l_hostRights = null;
      array_hostRights = null;
      simpleMainLink = null;
      mainLink = null;

      if (Thread.interrupted()) {
        throw new InterruptedException();
      }

                                        // for testing and tracking
                                        //System.out.println("end of "+
                                          //"NodeRightsMySQLImpl.detectAddonFinalAllNodes() ");
      
      return _m_l_DisplayableLinks;
  } // detectAddonFinalAllNodes()

  /**
  * Gets all the rights of the relevant direct AclEntries managed by the MySQL Resource
  * for a given EligibleParty, following the ACL sorting and the selection
  * of the first matched ACL. Inherited AclEntries are not covered.
  * For an accounts-for-hosts group, this method calls detectL_aclEntryRightsNOTForToBeMixedAcl()
  * with a null UserIDMySQL argument.
  * The selection of the ACL is explained in the documentation of DirectoryMySQLImpl.
  * These AclEntries cannot be managed by different ACSs, and
  * there is no negative rights in MySQL.
  * Note: this method is sensible since it is indirectly used for both the one-link method,
  * the two-link method and a multiple-links method.
  * Called by AclRightsFactoryImpl.addAclInAccessControlLink().
  * Calls getConnectionPair(), detectL_aclEntryRightsNOTForToBeMixedAcl(),
  * detectL_aclEntryRightsForDatabaseNOTForToBeMixedAcl() and
  * detectL_aclEntryRightsForDatabaseForToBeMixedAcl().
  * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl
  * @param _res of this instance, with some AclEntries to filter
  * @param _ep must be an account or an accounts-for-hosts group
  * @param _onlyNonConditionalACLs true to read only the non-conditional AclEntries.
  * Always false and unused into the MySQL AcsAddon.
  * @return list of positive or negative rights from different ACSs, or null.
  * */
  public List<StringRight> detectL_aclEntryRights(ImmutableResource _res,
    ImmutableEligibleParty _ep, boolean _onlyNonConditionalACLs) {

      if (!(_ep instanceof UserIDMySQL) && !(_ep instanceof GroupIDMySQL)) return null;
          //throw new InternalError("NodeRightsMySQLImpl.detectL_aclEntryRights() "+
            //      "with a wrong UserIDMySQL for "+_ep.getName());

                                      // for testing and tracking
                                      //System.out.println("start of NodeRightsMySQLImpl."+
                                        //      "detectL_aclEntryRights() "+
                                          //    "for _res "+_res+" and _ep "+
                                            //  _ep);

      List<StringRight>     l_result = null;
      List<GroupIDMember>   l_pair = null;

      if (_ep instanceof UserIDMySQL) {
          l_pair = getConnectionPair((UserIDMySQL)_ep);
      }
      else if (_ep instanceof GroupIDMySQL &&
              _ep.getEPType().indexOf("accounts-for-hosts group") != -1) {
          
          l_result = detectL_aclEntryRightsNOTForToBeMixedAcl(_res,
                    null, (GroupIDMySQL) _ep,
                    Arrays.asList(_res.getL_aclEntry()));

                                      // for testing and tracking
                                      //if (_res.getNickName().equals("MySQL server"))
                                        //  System.out.println("end of NodeRightsMySQLImpl."+
                                          //        "detectL_aclEntryRights() "+
                                            //  "for _res "+_res.getNickName()+" and _ep "+
                                              //_ep.getNickName()+"- result: "+l_result);

          return l_result;
      } // else

      if (l_pair == null)
          throw new InternalError("NodeRightsMySQLImpl.detectL_aclEntryRights() "+
                  "without connection pair for "+_ep.getName());

      l_result = detectL_aclEntryRightsNOTForToBeMixedAcl(_res,
                (UserIDMySQL) l_pair.get(0), (GroupIDMySQL) l_pair.get(1),
                Arrays.asList(_res.getL_aclEntry()));

      if (_res.getType().indexOf("database") != -1 && l_result == null)
        l_result = detectL_aclEntryRightsForDatabaseForToBeMixedAcl(_res, 
                (UserIDMySQL) l_pair.get(0), (GroupIDMySQL) l_pair.get(1),
                Arrays.asList(_res.getL_aclEntry()));

                                      // for testing and tracking
                                      //if (_res.getNickName().equals("MySQL server"))
                                        //  System.out.println("end of NodeRightsMySQLImpl."+
                                          //        "detectL_aclEntryRights() "+
                                            //  "for _res "+_res.getNickName()+" and _ep "+
                                              //_ep.getNickName()+"- result: "+l_result);

      l_pair = null;
      return l_result;
  } // detectL_aclEntryRights()

  /**
  * This method is designed to be fast, and to returns true if the resource tree
  * allows the source to access through the inherited ACL rights to the leaf.
  * The Resources root 'MySQL server' has no inherited AclEntries but it may have
  * some access. This is why this method overrides the super method.
  * The AGO and PRI inherited rights are not processed. <P>
  * In the ACS MySQL, the effective rights on a resource are selected
  * by a two-steps algorithm, derived from the MySQL specification.
  * Only the AclEntries of the root 'MySQL server' are considered by this method,
  * with the bases sets ACL. An Actor has no inherited rights. <P>
  * Called by AclRightsFactoryImpl.addAclInAccessControlLink(),
  * ThreeNodesRightsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeMemberOwnContain(),
  * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
  * Calls withAclAccessThroughNodesTree() to do the main job.
  * @see #getL_accessRightsThroughNodesTree(ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.ImmutableGroupID) 
  * @param _ep Actor, UserID or GroupID to test. If null, returns true.
  * If its ACS does not manage any rights inheritance, returns true.
  * @param _res Resource or VirtualFolder of this instance,
  * which may be accessed or not through its parent tree. Never null.
  * @param _ep_2 the group member which may be accessed by _ep. Never ridden there.
  * @return true if the resource parent tree allows _ep to go through down to the resource,
  * if _ep is null.
  * */
  public boolean withAccessThroughNodesTreeFromEP(
     ImmutableSource _ep, ImmutableLeaf _res, ImmutableGroupIDMember _ep_2) {

    boolean             result = true;

    // process the ACL inherited rights
    if (!(_res instanceof ImmutableResource)) return false;

                                       // for testing and tracking
                                       //if (_res.getNickName().equals("product_price column"))
                                         //   System.out.println("start of NodeRightsMySQLImpl."+
                                           //     "withAccessThroughNodesTreeFromEP(), "+
                                             //   "for ACL rights, from "+
                                               // _ep.getNickName()+" to "+_res.getNickName());

    if (_ep instanceof Actor) {
        return false;
    }
    else if (_ep instanceof ImmutableUserID) {
        // no primary group
        result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                    (ImmutableUserID)_ep, null);
    } else if (_ep instanceof ImmutableGroupID) {
        result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                    null, (ImmutableGroupID)_ep);
    } // else

                           // for testing and tracking
                          //System.out.println("end of NodeRightsMySQLImpl.withAccessThroughNodesTreeFromEP(), "+
                            //"for the inherited ACL rights, and from "+_ep.getNickName()+
                            //" to "+_res.getNickName()+" - result: "+
                            //result);

    return result;
  } // withAccessThroughNodesTreeFromEP()

  /**
  * This important method returns in an AccessControlLink the inherited rights
  * that are given by the directory tree, as inherited rights on a resource.
  * The result sets only the type AccessControlLink.ACL.
  * If there is at least one conditional ACL, the result contains in the comment
  * 'conditional ACL' for the relevant type. Furthermore, the conditional group names
  * are added in the comment, like in AclRightsFactoryImpl.addAclInAccessControlLink()
  * for the direct ACLs.
  * This method has to be called only on the NodeRightsImpl instance of the argument _res.
  * There is never bottom-up search in the resource tree to find the applicable rights.
  * The default algorithm in NodeRightsImpl is overridden and unused. <P>
  * The set of applicable rights on a target are derived from the direct AclEntries on this target,
  * or, as inherited AclEntries, from one of its (in)direct parent or from a matched bases set.
  * The database is a special parent, having the most complex rights. A first, but not unique,
  * difference is that it may have complementary copied ACL from the bases sets
  * for which this database matches. For instance, the bases set named 'on%' matches
  * the database named 'one', and all the 'on%' direct ACL are then copied
  * by the class BasesSetMySQLImpl into the 'one' direct ACL list. <P>
  * In the ACS MySQL, the effective rights of an user on a resource are selected
  * by a two-steps algorithm, derived from the MySQL specification, that is described
  * in the documentation of the class DirectoyMySQLImpl.
  * All the ACL sorting for a given parent level is the responsibility of the method
  * BaseUtilityMySQLImpl.addInSortedAclList() and the related methods in the same class,
  * like addAclNOTForDatabase(). <P>
  * If the argument _ep is a MySQL user, this method may return its inherited rights
  * through an accounts-for-hosts group or a host group the account belongs to. 
  * In the second case, to comment the returned AccessControlLink for the 'See Why' text,
  * this method adds is 'implies two database inherited to-be-mixed ACL'.
  * On the other hand, the comment 'implies two direct to-be-mixed ACL' is added by
  * getMergedInheritedAclPriRightsThroughNodesTree() for the direct ACLs. <P>
  * The user groups may have conditional ACLs, but they are never passed on as the argument _grp,
  * since the group to use is in the cuurent user for the argument _ep.
  * This is why the argument _grp is unused in this method. <P>
  * In all cases, the method getMergedInheritedAclPriRightsThroughNodesTree()
  * is responsible to set the effective ACL rights, from the inheritance or not.
  * To have effective rights, the array of inherited rights has to not be null, but it may be empty.
  * It is used to get the rights for the initial pair (user, host) at each parent level.
  * If the current user has only an USAGE right on the root and if, for a given Resource,
  * the current user has no other inherited ACL on the target, then no right is granted.
  * This method is strongly coupled to withAccessThroughNodesTreeFromEP() and
  * withAclAccessThroughNodesTree(). It is designed to be fast. <P>
  * Calls Resource.getEorImmutableL_TargetInheritedAclEntries(),
  * getConnectionPair() and, for doing the main job, detectL_aclEntryRightsNOTForToBeMixedAcl()
  * and detectL_aclEntryRightsForDatabaseForToBeMixedAcl().
  * Called only by AclRightsFactoryImpl.addAclInAccessControlLink().
  * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl
  * @see #withAccessThroughNodesTreeFromEP(ARoad0.gBaseInterface.ImmutableSource, ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableGroupIDMember)
  * @see #getMergedInheritedAclPriRightsAndComments(ARoad0.gBaseInterface.ImmutableSource, ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.CNot.AccessControlLinkImpl, ARoad0.gBaseInterface.StringRight[]) 
  * @param _res Resource or VirtualFolder of this instance
  * may be accessed or not through its parent tree. May be null.
  * @param _ep the UserID of _res or another UserID. Null only if _grp is not null.
  * @param _grp an accounts-for-hosts group. Null only if _ep is not null.
  * @return link with the AccessControlLink.ACL rights given by the parents tree, and the comments.
  * Null only when an error occurs, if _res or its ACS is null,
  * the ACS are not the same, the group is forbidden.
  * Empty if _res has no parent, no inherited rights, or if it is a bases set.
  * */
  public AccessControlLinkImpl getL_accessRightsThroughNodesTree(
          ImmutableLeaf _res, ImmutableEligibleParty _ep, ImmutableGroupID _grp) {

                                       // for testing and tracking
                                      //System.out.println("start of NodeRightsMySQLImpl."+
                                        //          "getL_accessRightsThroughNodesTree(), with _res: "+
                                          //        _res+", _ep: "+_ep+", _grp: "+_grp);

      if (_ep == null && _grp == null)
          throw new InternalError("NodeRightsMySQLImpl.getL_accessRightsThroughNodesTree() "+
                  "with 2 null arguments.");
      if (_ep != null && _grp != null)
          throw new InternalError("NodeRightsMySQLImpl.getL_accessRightsThroughNodesTree() "+
                  "with 2 non-null arguments.");
      if (!(_res instanceof ResourceMySQL))
          throw new InternalError("NodeRightsMySQLImpl.getL_accessRightsThroughNodesTree() "+
                  "with a first argument that is not a ResourceMySQL.");

      if (_res == null || _res.getEorACS() == null) {

                                       // for testing and tracking
                                      //System.out.println("NodeRightsMySQLImpl."+
                                        //          "getL_accessRightsThroughNodesTree(), with _res: "+
                                          //        _res+", _ep: "+_ep+", return null.");

          return null;
      }
      if (_ep != null && _res.getEorACS() != _ep.getEorACS() ) {

                                       // for testing and tracking
                                      //System.out.println("NodeRightsMySQLImpl."+
                                        //          "getL_accessRightsThroughNodesTree(), with _res: "+
                                          //        _res+", _ep: "+_ep+", return null.");

          return null;
      }
       if (_grp != null && _res.getEorACS() != _grp.getEorACS() ) {

                                       // for testing and tracking
                                      //System.out.println("NodeRightsMySQLImpl."+
                                        //          "getL_accessRightsThroughNodesTree(), with _res: "+
                                          //        _res+", _ep: "+_ep+", return null.");

          return null;
      }


      ResourceMySQL         res = (ResourceMySQL) _res;

      // for the 'MySQL server' root and the bases sets
      if (res.getEorParent() == null || res.getType().indexOf("bases set") != -1) {
          res = null;
          return new AccessControlLinkImpl();
      }

      List<AclEntry>        l_inherACL = res.getEorImmutableL_TargetInheritedAclEntries();

      if (l_inherACL == null || l_inherACL.isEmpty()) {
          l_inherACL = null;
          res = null;

                               //for testing and tracking
                             //System.out.println("in NodeRightsMySQLImpl."+
                               //     "getL_accessRightsThroughNodesTree(), for the resource "+
                                 //   _res.getName()+" with no inherited ACL. "+
                                   // "Returns an empty AccessControlLink.");

          return new AccessControlLinkImpl();
      } // if

                          // for testing and tracking
                          //System.out.println("NodeRightsMySQLImpl."+
                            //      "getL_accessRightsThroughNodesTree() "+
                              //    "with l_inherACL size = "+l_inherACL.size());

      AccessControlLinkImpl result = null;
      String                comment = null;
      List<GroupIDMember>   l_pair = null;
      UserIDMySQL           current_user = null;
      GroupIDMySQL          current_group = null;
      List<StringRight>     l_rights = null;
      int                   last_database_level_acl_index = -1;
      AclEntry              ac = null;
      String                current_target = null;
      boolean               onlyRootUSAGE = false;

      if (_ep != null && _ep instanceof UserIDMySQL) {
          l_pair = getConnectionPair((UserIDMySQL)_ep);
          current_user = (UserIDMySQL) l_pair.get(0);
          current_group = (GroupIDMySQL) l_pair.get(1);
      }
      else if (_grp != null &&
              _grp.getEPType().indexOf("accounts-for-hosts group") != -1) {
          // null current_user
          current_group = (GroupIDMySQL) _grp;

                                      // for testing and tracking
                                      //if (_res.getNickName().equals("MySQL server"))
                                        //  System.out.println("end of NodeRightsMySQLImpl."+
                                          //        "getL_accessRightsThroughNodesTree() "+
                                            //  "for _res "+_res.getNickName()+" and _ep "+
                                              //_ep.getNickName()+"- current_group "+
                                              //"is an accounts-for-hosts group: "+
                                              //current_group.getNickName());

      } // else


                                       // for testing and tracking
                                      //System.out.println("start of NodeRightsMySQLImpl."+
                                        //          "getL_accessRightsThroughNodesTree(), with _res: "+
                                          //        _res+", _ep: "+_ep+", _grp: "+_grp+
                                            //      ", current user: "+l_pair.get(0)+
                                              //    ", current user group: "+l_pair.get(1));

      // as a direct child 'MySQL server', a database uses l_acl without filtering
      if (res.getType().indexOf("database") != -1) {
        l_rights = detectL_aclEntryRightsNOTForToBeMixedAcl(
                res, current_user, current_group,
                l_inherACL);
        if (l_rights != null)
            l_inheritedRightsResult_.addAll(l_rights);
      }
      else {
          // for the 'MySQL server' parent level
        for (AclEntry acl : l_inherACL) {
            if (acl.getResName().lastComponent().equals("MySQL server")) {
                l_acl_for_one_level_.add(acl);
                last_database_level_acl_index++;
            } // if
        } // for acl
        l_rights = detectL_aclEntryRightsNOTForToBeMixedAcl(
                res, current_user, current_group,
                l_acl_for_one_level_);
        if (l_rights != null) {
            l_inheritedRightsResult_.addAll(l_rights);
            if (l_rights.size() == 1 &&
                    l_rights.get(0).equals(_res.getEorACS().getOneRight("USAGE"))) {
                onlyRootUSAGE = true;
            } // if
        }
        else //if (_ep != null)
            throw new InternalError("NodeRightsMySQLImpl.getL_accessRightsThroughNodesTree(),"+
                    " without rights for the root 'MySQL server'.");
        //else {
            // _grp is an accounts-for-hosts without root rights
          //l_inheritedRightsResult_.clear();
          //l_acl_for_one_level_.clear();

          //l_inherACL = null;
          //res = null;
          //l_pair = null;
          //l_rights = null;
          //comment = null;
          //current_user = null;
          //current_group = null;
          //return null;
        //} // else

        l_acl_for_one_level_.clear();

          // for the database-specific parent level
        for (AclEntry acl : l_inherACL) {
            if (acl.getResType().indexOf("database") != -1 ||
                    acl.getResType().indexOf("bases set") != -1) {
                l_acl_for_one_level_.add(acl);
                last_database_level_acl_index++;
            } // if
        } // for acl
        l_rights = detectL_aclEntryRightsNOTForToBeMixedAcl(
                res, current_user, current_group,
                l_acl_for_one_level_);
            // only for the database level, read the to-be-mixed ACLs
            // if _grp is not an accounts-for-hosts group
        if (l_rights == null) {
            if (current_user != null) {
                l_rights = detectL_aclEntryRightsForDatabaseForToBeMixedAcl(
                        res, current_user, current_group,
                        l_acl_for_one_level_);
                if (l_rights != null) {
                    l_inheritedRightsResult_.addAll(l_rights);
                    // for mixed non-conditional ACLs on the database
                    if (comment == null)
                        comment = "implies two database to-be-mixed ACL";
                } // if
            } // if
        }
        else {
            l_inheritedRightsResult_.addAll(l_rights);
        }
        l_acl_for_one_level_.clear();

                                   // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //      "getL_accessRightsThroughNodesTree() starts to process "+
                                      //    "the targets under the database, "+
                                        //  "with last_database_level_acl_index = "+
                                          //last_database_level_acl_index);

          // for each parent level under the database level
        if (last_database_level_acl_index != l_inherACL.size()) {
            current_target = l_inherACL.get(last_database_level_acl_index).getResName().
                    lastComponent();

                                   // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //      "getL_accessRightsThroughNodesTree() sets "+
                                      //    "the target: "+current_target);

            for (Iterator ite = l_inherACL.listIterator(last_database_level_acl_index); ite.hasNext(); ) {
                ac = (AclEntry)ite.next();

                if (current_target.equals(ac.getResName().lastComponent())) {
                    l_acl_for_one_level_.add(ac);
                }
                else {
                    // to process the previous parent level AclEntries
                    l_rights = detectL_aclEntryRightsNOTForToBeMixedAcl(
                            res, current_user, current_group,
                            l_acl_for_one_level_);
                    if (l_rights != null)
                        l_inheritedRightsResult_.addAll(l_rights);
                    l_acl_for_one_level_.clear();

                    // preparing the next parent level
                    current_target = ac.getResName().lastComponent();
                    l_acl_for_one_level_.add(ac);


                                   // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //      "getL_accessRightsThroughNodesTree() sets "+
                                      //    "the target: "+current_target+
                                        //  " from the ACL "+ac);
                } // else

                if (!ite.hasNext()) {

                                   // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //      "getL_accessRightsThroughNodesTree() ends the processing "+
                                      //    "for the target: "+current_target);

                    // to process the last parent level AclEntries
                    l_rights = detectL_aclEntryRightsNOTForToBeMixedAcl(
                            res, current_user, current_group,
                            l_acl_for_one_level_);
                    if (l_rights != null)
                        l_inheritedRightsResult_.addAll(l_rights);
                    //l_acl_for_one_level.clear();
                } // ite

            } // for ite
        } // if (last_database_level_acl_index
      } // else

      // there, l_inheritedRightsResult_ is never empty

        // If the current user has only an USAGE right on the root and if,
        // for a given Resource, the current user has no other inherited ACL
        // on the target, the USAGE right is granted to not forbid the direct ACL.
      //if (l_inheritedRightsResult_.size() == 1 && onlyRootUSAGE) {
        //  result = null;
      //}
      result = new AccessControlLinkImpl();
      result.setRights(
              AccessControlLink.ACL,
              l_inheritedRightsResult_.toArray(new StringRight[0]));

      if (comment != null)
          result.setLinkTypeComments(AccessControlLink.ACL, comment);

                                   // for testing and tracking
                                  //System.out.println("end of NodeRightsMySQLImpl."+
                                    //      "getL_accessRightsThroughNodesTree() with result "+
                                      //    result);

      l_inheritedRightsResult_.clear();
      l_acl_for_one_level_.clear();

      l_inherACL = null;
      res = null;
      l_pair = null;
      l_rights = null;
      ac = null;
      comment = null;
      current_user = null;
      current_group = null;
      return result;
  } // getL_accessRightsThroughNodesTree()

  /**
  * Gets the effective ACL rights of the source on the target when
  * there are non-null inherited rights through the nodes tree,
  * adds the current user as comment,
  * and adds the comment 'database mixed ACL' for the direct AclEntries
  * on databases. The argument _l_transientInheritedRights contains the inherited rights 
  * getL_accessRightsThroughNodesTree() has found. <P>
  * In this MySQL AcsAddon class, the inherited rights are simply added to 
  * the direct rights already set in the argument _upd_aclink.
  * This method adds the CURRENT_USER, calls the super method to do its work, then
  * adds 'conditional ACL' in the ACL comment, with the condition group names,
  * for each non-inherited conditional ACL when the leaf is a database. <P>
  * For a database, it adds also a second comment 'implies two direct to-be-mixed ACL'
  * which is proper to this MySQL AcsAddon, since it is used to give the group
  * in a couple of two non-conditional non-inherited ACLs on a database.
  * For a bases set, it adds a GLOBAL comment AccessControlLink.C_WEAK_LINK.
  * The comment 'implies two database inherited to-be-mixed ACL' is added
  * by getL_accessRightsThroughNodesTree(). <P>
  * This method is derived from detectL_aclEntryRights() and the super method.
  * Calls detectL_aclEntryRightsNOTForToBeMixedAcl() and
  * detectL_aclEntryRightsForDatabaseForToBeMixedAcl().
  * Called by AclRightsFactoryImpl.addAclInAccessControlLink().
  * @param _so source for which the rights on _targ has to be set
  * @param _targ Resource or VirtualFolder of this instance. May be null.
  * @param _upd_acLink contains the direct ACL or LPRI rights, if any. Never null.
  * May be empty.
  * @param _l_transientInheritedRights has to be from getL_accessRightsThroughNodesTree().
  * Never null. May be empty.
  * @return _upd_acLink with the effective ACL or LPRI rights of the source on the target.
  * May be null.
  * */
  public AccessControlLinkImpl getMergedInheritedAclPriRightsAndComments(
          ImmutableSource _so, ImmutableLeaf _targ,
          AccessControlLinkImpl _upd_acLink, StringRight[] _l_transientInheritedRights) {

        if (_so instanceof ActorMySQL)
              return null;
        else if (_so instanceof GroupIDMySQL) {
            if (((GroupIDMySQL)_so).getEPType().indexOf("accounts-for-hosts group") != -1) {
                super.getMergedInheritedAclPriRightsAndComments(_so, _targ,
                  _upd_acLink, _l_transientInheritedRights);

                if (((Resource)_targ).getType().indexOf("bases set") != -1) {
                    _upd_acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                    _upd_acLink.addComment(
                          AccessControlLink.GLOBAL,
                          AccessControlLink.C_WEAK_LINK);
                }

                return _upd_acLink;
            }
            else {
                return null;
            }
        } // else

        List<GroupIDMember>   l_pair = getConnectionPair((UserIDMySQL)_so);
        _upd_acLink.addComment(AccessControlLink.ACL, CURRENT_USER+
                ((GroupIDMember)l_pair.get(0)).getNickName()+", "+
                ((GroupIDMember)l_pair.get(1)).getNickName()+")");

                                        //for testing and tracking
                                    //System.out.println(
                                      //"NodeRightsMySQLImpl.getMergedInheritedAclPriRightsAndComments()"+
                                      //" ACLink after adding the current user as ACL comment:"+
                                      //_upd_acLink.getLinkTypeComments(AccessControlLink.ACL));

            // put _l_transientInheritedRights into _upd_acLink
        super.getMergedInheritedAclPriRightsAndComments(_so, _targ,
          _upd_acLink, _l_transientInheritedRights);

        if (((Resource)_targ).getType().indexOf("bases set") != -1) {
            _upd_acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
            _upd_acLink.addComment(
                  AccessControlLink.GLOBAL,
                  AccessControlLink.C_WEAK_LINK);
        }

        if (((ResourceMySQL)_targ).getType().indexOf("database") == -1) {
            return _upd_acLink;
        }

                                        //for testing and tracking
                                    //System.out.println(
                                      //"NodeRightsMySQLImpl.getMergedInheritedAclPriRightsAndComments()"+
                                      //" adding of 'database mixed ACL' comments");

        // the direct ACL are used only for the comments in the result and for databases only
        AclEntry[]          l_resAcl = ((Resource)_targ).getL_aclEntry();
        String              comment = null;

        List<StringRight> l_rights = detectL_aclEntryRightsNOTForToBeMixedAcl(
                (ImmutableResource) _targ,
                (UserIDMySQL) l_pair.get(0), (GroupIDMySQL) l_pair.get(1),
                Arrays.asList(l_resAcl));

        if (l_rights == null) {
            l_rights = detectL_aclEntryRightsForDatabaseForToBeMixedAcl(
            (ImmutableResource) _targ,
            (UserIDMySQL) l_pair.get(0), (GroupIDMySQL) l_pair.get(1),
            Arrays.asList(l_resAcl));

            if (l_rights != null) {
                // add for the to-be-mixed ACLs, the comment
                comment = "; implies two direct to-be-mixed ACL ";

                                            //for testing and tracking
                                        //System.out.println(
                                          //"NodeRightsMySQLImpl.getMergedInheritedAclPriRightsAndComments()"+
                                          //" for _res "+_targ.getNickName()+" and _so "+
                                          //_so.getNickName()+" adding of the comment "+
                                          //"'implies two direct to-be-mixed ACL'");

                _upd_acLink.addComment(AccessControlLink.ACL, comment);
            } // if
        } // if

        l_resAcl = null;
        comment = null;
        l_pair = null;
        l_rights = null;
        return _upd_acLink;
  } // getMergedInheritedAclPriRightsThroughNodesTree()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
      super.finalizeForProcess();
      facade_ = null;
      l_accounts_ = null;
      l_actors_ = null;
      l_hosts_ = null;
      l_vfolders_ = null;
      l_inheritedRightsResult_ = null;
      l_acl_for_one_level_ = null;
  } // finalizeForProcess()

  /**
   * @return a descriptor
   */
  public String toString() {
      return "NODERIGHTS_MYSQL_IMPL of the view "+interpreter_.getSource().getName();
  }

  /**
  * Gets the first current user, that is the pair (user, simple host or accounts-for-hosts group)
  * under which the connection of _ep is done on the MySQL server.
  * The proxy account is an exception. This method does not find its current user
  * that is always its proxied acccount.
  * The result depends uniquely on the direct ACL of the root 'MySQL server',
  * and their sorting in the relevant list.
  * The returned group cannot be an hosts group.
  * The direct ACL of the root 'MySQL server' are used to state if there is
  * an authorization for accessing the server, with at least the right 'USAGE'.
  * This conection pair is indeed the MySQL 'current user' for analyzing the rights
  * on any first node, in the access paths. If an intermediate node is an Actor
  * in a MySQL DEFINER mode, the 'current user' may change into the access path.
  * Called by detectL_aclEntryRights(), getL_accessRightsThroughNodesTree()
  * and, first of all, by withAclAccessThroughNodesTree().
  * @param _ep the userID for which there is a connection request to the MySQL Server.
  * May be null.
  * @return the current user (user, simple host or accounts-for-hosts group).
  * Null if there is no access to the server, or if _ep is null.
  * The returned account may be 'anonymous'.
  * */
  public List<GroupIDMember> getConnectionPair(UserIDMySQL _ep) {

                                                // for testing and tracking
                                              //System.out.println(
                                                //"start of NodeRightsMySQLImpl.getConnectionPair(), "+
                                                //"for _ep "+_ep.getName());

      if (_ep == null) return null;

      List<GroupIDMember>           l_result = null;
      DirectoryMySQL                server = (DirectoryMySQL)_ep.getEorACS().getEorM_Resources().get(
                                            ((Name)_ep.getEorACS().getName()).
                                                extended("MySQL server") );
      AclEntry[]                    l_acl = server.getL_aclEntry();
      Map<ImmutableName, StringRight[]>    m_prim = server.getM_GlobalTargetACLRights();
      Name                          keyName = null;
      // only to speed up
      Name                          procName = new NameImpl("<Proc>");
      ImmutableName                 epName = _ep.getName();
      String                        aclEpName = null;

      // sorted list of ACL, and the first matched ACL is used.
      // for the non-conditional ACL for an 'accounts-for-hosts group',
      // it is necessary to search if a primary key with _ep exists.
      for (AclEntry acl : l_acl) {
          aclEpName = acl.getEpName().lastComponent();

                                                // for testing and tracking
                                              //System.out.println(
                                                //"NodeRightsMySQLImpl.getConnectionPair() tests "+
                                                //"the ACL "+acl);

            // for conditional ACL of _ep or 'anonymous'
          if (acl.getEorEP() == _ep ||
                  (acl.getEorEP() instanceof UserIDMySQL &&
                  acl.getEorEP().getNickName().equals("anonymous"))) {

                keyName = ((Name)acl.getEpName().clone()).context().
                      extended(aclEpName+
                        "@"+
                        ((GroupID)acl.getL_ConditionSourceGroups().get(0)).
                            getName().lastComponent()).extended("0");

                                                // for testing and tracking
                                              //System.out.println(
                                                //"NodeRightsMySQLImpl.getConnectionPair() searches "+
                                                //"for the primary key "+keyName);

                if (m_prim.get(keyName) != null) {
                  l_result = new ArrayList(2);
                  l_result.add((GroupIDMember)acl.getEorEP());
                  l_result.add((GroupIDMember)acl.getL_ConditionSourceGroups().get(0));
                  break;
                } // if
          } // if (acl.getEorEP() == _ep

          else if (!acl.isConditionalAclEntry() &&
                    acl.getEorEP() instanceof GroupID &&
                    ((GroupID)acl.getEorEP()).containsAsMember(_ep)) {
              
                // for 'accounts-for-hosts group' ACL
                keyName = (Name) procName.clone();
                keyName.extended(epName.context()).
                      extended(epName.lastComponent()+
                        "@"+
                        aclEpName.substring(
                            aclEpName.indexOf("@")+1)).
                        extended("0");

                                                // for testing and tracking
                                              //System.out.println(
                                                //"NodeRightsMySQLImpl.getConnectionPair() searches "+
                                                //"for the primary key "+keyName);
                                              
                if (m_prim.get(keyName) != null) {
                  l_result = new ArrayList(2);
                  l_result.add((GroupIDMember)_ep);
                  l_result.add((GroupIDMember)acl.getEorEP());
                  break;
                } // if
          } // else

      } // for acl

                                            // for testing and tracking
                                          //if (_ep != null)
                                              //System.out.println(
                                                //"end of NodeRightsMySQLImpl.getConnectionPair(), "+
                                                //"for _ep "+_ep.getNickName()+
                                                //", result: "+l_result);
                                          //else
                                            //  System.out.println(
                                               // "end of NodeRightsMySQLImpl.getConnectionPair(), "+
                                                //"for _ep "+_ep.getNickName()+
                                                //", result: "+l_result);

      server = null;
      l_acl = null;
      keyName = null;
      procName = null;
      epName = null;
      m_prim = null;
      return l_result;
  } // getConnectionPair()

// ------------- protected methods ----------------------------------

  /**
  * Returns true if the AclEntry inherited rights in the resources tree allows
  * to access to the resource for the EligibleParty _ep.
  * For an account, this method calls simply getConnectionPair() to see if there is a connection.
  * For an accounts-for-hosts group, it returns true if the group has a non-null-rights ACL
  * on the root 'MySQL server', and at least one account as member.
  * Otherwise, it returns false for any group. <P>
  * A MySQL Actor cannot have some direct access on the MySQL resources through AclEntries.
  * An accounts-for-hosts group is not a true self-defined accessors (SDA)
  * and a bases set is not a true access target, since
  * the MySQL server does not deliver effective access in theses cases.
  * However, the properties of the MySQL objects has to be complete.
  * An Access Road view must display, as weak links, the access paths involving
  * an accounts-for-hosts group or a bases set. This method, for these two cases,
  * may return 'true'. It is 'true' for a bases set when a bases set ACL
  * produce a database ACL which is copied into the root 'MySQL server'.
  * If there is no database to match the bases set, this method returns 'false'
  * and the bases set ACL is not workable. If there is no ACL on the bases set,
  * it also returns 'false'. <P>
  * Caution: to follow the AcsAddon pattern, withAccessThroughNodesTreeFromEP()
  * should be the single caller of this method in the generic gWork package.
  * This method is not private to let the subclasses override it,
  * but it is necessary to call it through the same way, that is
  * to never jump the method withAccessThroughNodesTreeFromEP() which
  * chooses the arguments of this method. <P>
  * Called by withAccessThroughNodesTreeFromEP(). Calls getConnectionPair().
  * @param _res Resource of this instance; may be accessed or not through its parent tree
  * @param _ep an userID or an Actor. Null if _grp is not null.
  * @param _grp an accounts-for-hosts group or a simple host or a hosts group.
  * Null if _ep is not null. May return 'true' only for an accounts-for-hosts group.
  * @return true if the resource parent tree allows to go through down to the leaf.
  * */
  protected boolean withAclAccessThroughNodesTree(
    ImmutableResource _res, ImmutableEligibleParty _ep, ImmutableGroupID _grp) {

      // If there is no matched ACL on the bases set, it returns 'false'.
      if (_res.getType().indexOf("bases set") != -1) {
            boolean found = false;

            AclEntry[] l_acl = _res.getL_aclEntry();
            if (_grp != null &&
                    _grp.getType().indexOf("accounts-for-hosts group") != -1) {
                for (int i = 0; i < l_acl.length; i++) {
                      if (l_acl[i] != null && l_acl[i].getEorEP() == _grp &&
                              l_acl[i].getNoConditionalRight().length != 0) {
                        found = true;
                        break;
                      } // if
                } // for i
            } // if
            else if (_ep != null) {
                for (int i = 0; i < l_acl.length; i++) {
                      if (l_acl[i] != null && l_acl[i].getEorEP() == _ep &&
                              l_acl[i].getL_Rights().length != 0) {
                        found = true;
                        break;
                      } // if
                } // for i
            } // else

            if (!found) return false;
            // otherwise, the following code is applied
      } // if

      // process the accounts-for-hosts group
      if (_ep == null && _grp != null) {
          boolean found = false;

          if (_grp.getType().indexOf("accounts-for-hosts group") != -1) {
              
              if (((GroupIDMySQL)_grp).getEorL_Members().length == 0)
                  return false;

              AclEntry[] l_acl = null;
              
              for (ImmutableResource res : _grp.getEorL_AclConnectedResources()) {
                  // detect an ACL on the root, if non-null ACL rights
                if (res.getName().lastComponent().equals("MySQL server")) {
                    l_acl = res.getL_aclEntry();

                    for (int i = 0; i < l_acl.length; i++) {
                          if (l_acl[i] != null && l_acl[i].getEorEP() == _grp &&
                                  l_acl[i].getNoConditionalRight().length != 0) {
                            found = true;
                            break;
                          } // if
                    } // for i

                    if (found) break;
                } // if
              } // for res
              l_acl = null;

              if (found)
                  return true;
          } // if

          return false;
      } // if

      if (!(_ep instanceof UserIDMySQL) && !(_ep instanceof ActorMySQL))
          return false;

      List<GroupIDMember> l_connect = null;

      if (_ep instanceof UserIDMySQL)
          l_connect = getConnectionPair((UserIDMySQL)_ep);
      else
          l_connect = getConnectionPair((UserIDMySQL)((ActorMySQL)_ep).getEorCurrentUserID());

                                            // for testing and tracking
                                          //if (_ep != null)
                                            //  System.out.println(
                                              //  "end of NodeRightsMySQLImpl.withAclAccessThroughNodesTree(), "+
                                                //"for _res "+_res.getNickName()+" and _ep "+_ep.getNickName());

      if (l_connect != null) {
          l_connect = null;
          return true;
      }
      else {
          l_connect = null;
          return false;
      }
  } // withAclAccessThroughNodesTree()

  /**
  * Gets the applicable rights from the first applicable direct or inherited ACL
  * on a Resource which is not a database, or for the database rights
  * not from the to-be-mixed ACL. This is also the good method to use
  * when the Resource is a bases set. This method does not use
  * the _res rights maps on the inherited and primary ACL rights.
  * When the argument _res is the root 'MySQL server', this is the single case
  * where the right 'USAGE' may be returned if it is alone.
  * Called by detectL_aclEntryRights() and getL_accessRightsThroughNodesTree().
  * Calls BaseUtilityMySQLImpl.isAfterAsPatternNameHostsGroup().
  * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl
  * @param _res of this instance, with some direct AclEntries
  * @param _usr user of the current user. May be null only when _grp
  * is an accounts-for-hosts group.
  * @param _grp group of the current user. May be a simple host or an accounts-for-hosts group,
  * but not another group. Never null.
  * @param _l_resAcl immutable list of direct or inherited AclEntries
  * getL_accessRightsThroughNodesTree() has selected. Never null.
  * May be the instance variable l_acl_for_one_level_.
  * @return applicable direct-ACL rights, or null.
  * */
  protected List<StringRight> detectL_aclEntryRightsNOTForToBeMixedAcl(
    ImmutableResource _res, UserIDMySQL _usr, GroupIDMySQL _grp, List<AclEntry> _l_resAcl) {

                                      // for testing and tracking
                                      //System.out.println("start of NodeRightsMySQLImpl."+
                                        //      "detectL_aclEntryRightsNOTForToBeMixedAcl() "+
                                          //    "for _res "+_res+", _usr "+
                                            //  _usr+", _grp "+
                                              //_grp);

      List<StringRight>     l_result = null;
      StringRight[]         l_right = null;
      String                name = null;
      String                name2 = null;
      Pattern               pat = null;
      Matcher               matcher = null;

     for (AclEntry acl : _l_resAcl) {

                          // for testing and tracking
                          //System.out.println("NodeRightsMySQLImpl."+
                            //      "detectL_aclEntryRightsNOTForToBeMixedAcl() "+
                              //    "ACL to analyze "+acl);

          // to match, _grp must be the conditional group
          if (acl.isConditionalAclEntry()) {
              if (!acl.getEnabledRights())
                  continue;
              if (acl.getEorEP() != _usr)
                  continue;
              if (acl.getL_ConditionSourceGroups().get(0) != _grp)
                  continue;

              l_right = acl.getL_Rights();
              if (l_right.length == 0 ||
                      (l_right.length == 1 && l_right[0].getRight().equals("USAGE") &&
                      acl.getResType().indexOf("MySQL server") == -1))
                  continue;

              for (int j = 0; j < l_right.length; j++) {
                  if (l_result == null)
                      l_result = new ArrayList<StringRight>(INITIAL_CAPACITY);

                  l_result.add(l_right[j]);
              } // for j

                          // for testing and tracking
                          //System.out.println("NodeRightsMySQLImpl."+
                            //"detectL_aclEntryRightsNOTForToBeMixedAcl() "+
                              //    "selects the matched ACL "+acl);
              break;
          }
          // the other non-conditional ACL, as to-be-mixed ACL on a database, are not processed
          // if '_usr' member of the accounts-for-hosts group, including 'anonymous'
          else if (acl.getEorEP().getType().indexOf("accounts-for-hosts group") != -1 &&
                  ((_usr != null && ((GroupID)acl.getEorEP()).containsAsMember(_usr)) ||
                  _usr == null)) {

              // _grp may be an accounts-for-hosts group, and then,
              // the host part of its name cannot match the acl.getEorEP() name.
              // For instance, '192.168.1.%' does not match '192.168.%' there.
              // The single matching is when there are equals.
              if (_grp.getType().indexOf("accounts-for-hosts group") != -1) {
                  if (_grp != acl.getEorEP()) {

                                  // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //"detectL_aclEntryRightsNOTForToBeMixedAcl() jumps the ACL "+
                                    //acl);

                        continue;
                  } // if
              }
              else if (_usr == null) {
                  continue;
              }
              // otherwise, _grp is a simple host. It matches acl.getEorEP()
              // if the names matches, and this is always true since
              // acl.getEorEP()).containsAsMember(_usr).
              
              //else {
                //  name = _grp.getName().lastComponent();
                  //name2 = acl.getEorEP().getName().lastComponent().substring(
                    //        acl.getEorEP().getName().lastComponent().indexOf("@")+1);

                  //pat = Pattern.compile(
                    //    name2.replace("_", ".").replace("%", ".*"),
                      //  Pattern.CASE_INSENSITIVE);
                  //matcher = pat.matcher(name);
                  //if (!matcher.matches())
                    //continue;
              //} // else

                                  // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //"detectL_aclEntryRightsNOTForToBeMixedAcl() tests the ACL "+
                                    //acl);

              l_right = acl.getL_Rights();
              if (l_right.length == 0 ||
                      (l_right.length == 1 && l_right[0].getRight().equals("USAGE") &&
                      acl.getResType().indexOf("MySQL server") == -1))
                  continue;

              for (int j = 0; j < l_right.length; j++) {
                  if (l_result == null)
                      l_result = new ArrayList<StringRight>(INITIAL_CAPACITY);

                  l_result.add(l_right[j]);
              } // for j

                          // for testing and tracking
                          //System.out.println("NodeRightsMySQLImpl."+
                            //"detectL_aclEntryRightsNOTForToBeMixedAcl() "+
                              //    "selects the matched ACL "+acl);

              break;
          } // else
          // the non-conditional to-be-mixed ACL, at the end of the list for a database,
                          // are not processed
          else if (acl.getEorEP().getType().indexOf("accounts-for-hosts group") == -1)
              break;
     } // for i

                          // for testing and tracking
                          //System.out.println("NodeRightsMySQLImpl."+
                            //      "detectL_aclEntryRightsNOTForToBeMixedAcl() result, "+
                              //    "for _res "+_res.getNickName()+" and _usr "+
                                //  _usr.getNickName()+": ");
                          //if (l_result != null) {
                            //  for (StringRight r : l_result) {
                              //  System.out.println("----- "+r);
                              //}
                          //}
                          //else {
                            //  System.out.println("----- NULL ");
                          //}

     l_right = null;
     name = null;
     name2 = null;
     pat = null;
     matcher = null;

     return l_result;
  } // detectL_aclEntryRightsNOTForToBeMixedAcl()

  /**
  * Gets the applicable rights from the first applicable direct or inherited ACL
  * on a Resource which is a database, but without reading of the non-mixed ACL
  * of this database. Then, this method has to be called only if the near method
  * detectL_aclEntryRightsForDatabaseNOTForToBeMixedAcl() returns null, since
  * the mixed ACL are sorted after the non-mixed ACL. <P>
  * The argument _grp may be an accounts-for-hosts group, and then,
  * the host part of its name cannot match the hosts group of any non-conditional ACL.
  * Furthermore, there is no multi-level wildcard matching in MySQL,
  * since for instance, '192.168.1.%' never matches '192.168.%'.
  * The single matching in this case may be when the groups are not equals, but equivalent,
  * like the accounts-for-hosts group 'jerry§@192.168.1.%' as _grp argument,
  * and the hosts group '192.168.1.%' in a non-conditional to-be-mixed ACL. <P>
  * This method does not use the _res rights maps on the inherited and primary ACL rights.
  * Called by detectL_aclEntryRights() and getL_accessRightsThroughNodesTree().
  * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.DirectoryMySQLImpl
  * @param _res of this instance,as a database with some direct AclEntries
  * @param _usr user of the current user
  * @param _grp group of the current user. May be an accounts-for-hosts group,
  * but not an hosts group.
  * @param _l_resAcl immutable list of all the direct AclEntries or the AclEntries
  * of the database-specific target getL_accessRightsThroughNodesTree() has selected.
  * Never null.
  * May be the variable l_acl_for_one_level_.
  * @return applicable ACL rights, and they may be null, or also null if there
  * is no rights while there is a matching for two to-be-mixed ACL.
  * */
  protected List<StringRight> detectL_aclEntryRightsForDatabaseForToBeMixedAcl(
    ImmutableResource _res, UserIDMySQL _usr, GroupIDMySQL _grp, List<AclEntry> _l_resAcl) {

                                  // for testing and tracking
                                  //System.out.println("start of NodeRightsMySQLImpl."+
                                    //      "detectL_aclEntryRightsForDatabaseForToBeMixedAcl() "+
                                      //    "for _res "+_res.getNickName()+" and _ep "+
                                       //   _usr.getNickName());

      if (_l_resAcl.size() == 0) return null;

      List<StringRight>     l_result = null;
      AclEntry              acl2 = null;
      boolean               end = false;
      int                   aclIndex = -1;
      GroupID               grp = null;
      StringRight           right = null;
      String                name = null;
      String                name2 = null;
      Pattern               pat = null;
      Matcher               matcher = null;

      for (AclEntry acl : _l_resAcl) {

          // the ACL for simple hosts and hosts groups are at the end of the list,
          // and they are not processed alone
          if (acl.getEorEP() instanceof GroupID &&
                  acl.getEorEP().getType().indexOf("accounts-for-hosts group") == -1) {
              break;
          }

                              // for testing and tracking
                              //System.out.println("NodeRightsMySQLImpl."+
                                //  "detectL_aclEntryRightsForDatabaseForToBeMixedAcl() "+
                                  //"user ACL to analyze "+acl);
          
          // only the to-be-mixed ACL for an user is processed
          if (acl.isConditionalAclEntry())
              continue;
          if (acl.getEorEP() != _usr //&&
                  //!acl.getEorEP().getName().lastComponent().equals("anonymous")
                                      )
              continue;

          // search another non-condition ACL for a group,
          // in l_resAcl and after 'acl', for which _usr or 'anonymous' is a direct member.
          aclIndex = _l_resAcl.indexOf(acl) + 1;

          while (aclIndex != _l_resAcl.size()) {
                acl2 = _l_resAcl.get(aclIndex);

                              // for testing and tracking
                              //System.out.println("NodeRightsMySQLImpl."+
                                //  "detectL_aclEntryRightsForDatabaseForToBeMixedAcl() "+
                                  //"group ACL to analyze in the while loop "+acl2);

                if (!(acl2.getEorEP() instanceof GroupID)) {
                  aclIndex++;
                  continue; // in the while loop
                } // if

                grp = (GroupID) acl2.getEorEP();

                if (grp.getType().indexOf("accounts-for-hosts group") != -1)
                    throw new InternalError("NodeRightsMySQLImpl."+
                                  "detectL_aclEntryRightsForDatabaseForToBeMixedAcl() "+
                                  "detects a wrong order in the ACL list for the ACL "+
                                  acl2+" which has an accounts-for-hosts group,"+
                                  "during the searching on the ACL "+acl);

              // _grp may be an accounts-for-hosts group, and then,
              // the host part of its name cannot match the acl2.getEorEP() name.
              // Furthermore, there is no multi-level matching in MySQL,
              // since for instance, '192.168.1.%' never matches '192.168.%' there.
              // The single matching is when the groups are not equals, but equivalent,
              // like the accounts-for-hosts group _grp 'jerry§@192.168.1.%' and
              // the hosts group grp '192.168.1.%'.
                if (_grp.getType().indexOf("accounts-for-hosts group") != -1) {
                    name = _grp.getName().lastComponent();
                    name = name.substring(name.indexOf("§@") + 2);

                    if (!name.equals(grp.getName().lastComponent())) {
                            aclIndex++;
                            continue; // in the while loop
                    }
                } // if
              // otherwise, _grp is a simple host. It matches acl2.getEorEP()
              // if the names matches when 'grp' is a hosts group, or if 'grp' is equals.
                else if (_grp.getType().indexOf("accounts-for-hosts group") == -1 &&
                        _grp != grp) {
                    // 'grp' has to be a hosts group matching '_grp'
                  if (grp.getType().indexOf("hosts") == -1) {
                      aclIndex++;
                      continue; // in the while loop
                  }

                    // continue if _grp is a host and does not match grp as host group
                  name = _grp.getName().lastComponent();
                  name2 = grp.getName().lastComponent();

                  pat = Pattern.compile(
                        name2.replace("_", ".").replace("%", ".*"),
                        Pattern.CASE_INSENSITIVE);
                  matcher = pat.matcher(name);
                  if (!matcher.matches()) {
                      aclIndex++;
                      continue; // in the while loop
                  }
                } // else

                // if no rights in the user non-conditional ACL or the host non-conditional ACL,
                // or if there is no common rights,
                // there is a match and the method returns null
                // to forbid this host from the database-specific ACLs
                if (acl.getL_Rights().length == 0 ||
                        acl2.getL_Rights().length == 0) {
                    l_result = null;
                    end = true;
                    break; // in the while loop
                }
                else if (Collections.disjoint(
                        Arrays.asList(acl.getL_Rights()),
                        Arrays.asList(acl2.getL_Rights()))) {
                    l_result = null;
                    end = true;
                    break; // in the while loop
                }

                l_result = new ArrayList<StringRight>(INITIAL_CAPACITY);
                l_result.addAll(Arrays.asList(acl.getL_Rights()));

                // The rights of this mixed ACL are the intersection
                // of the rights in the two non-conditional ACLs
                for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                    right = (StringRight) ite.next();
                    if (!Arrays.asList(acl2.getL_Rights()).contains(right))
                        ite.remove();
                } // for ite

                                  // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //"detectL_aclEntryRightsForDatabaseForToBeMixedAcl() "+
                                      //    "processing for the GroupID ACL "+acl2);

                                  // for testing and tracking
                                  //System.out.println("NodeRightsMySQLImpl."+
                                    //"detectL_aclEntryRightsForDatabaseForToBeMixedAcl() result, "+
                                      //"for _res "+_res.getNickName()+" and _usr "+
                                       //_usr.getNickName()+": ");
                                  //if (l_result != null) {
                                    //  for (StringRight r : l_result) {
                                      //  System.out.println("----- "+r);
                                      //}
                                  //}
                                  //else {
                                    //  System.out.println("----- NULL ");
                                  //}

                end = true;
                break; // in the while loop
          } // while

          if (end)
              break;

      } // for i

      grp = null;
      right = null;
      name = null;
      name2 = null;
      pat = null;
      matcher = null;

      return l_result;
  } // detectL_aclEntryRightsForDatabaseForToBeMixedAcl()

} // NodeRightsMySQLImpl



