//Title:        Access Road
//Version:      0.7.2
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.AcsAddon.Accbee.Ubuntu.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.util.List;
import java.util.Map;
import java.util.ArrayList;
import java.util.SortedMap;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;

/**
 * This class is responsible for modeling a Linux Ubuntu as an access control system (ACS).
 * It ensures the modelization of a classic  Linux operating system, the Ubuntu authorizations
 * on the system and hardware resources, and it covers partially the capability system.
 * Most of the work is naturally done by the super class.
 * <P>
 * The ACS is created by ACSFactoryUbuntuImpl.createACS(), which set the booleans and string
 * parameters for the new ACS. They cannot be changed after the ACS creation.
 * They define the ACS general structure and comportement. At the ACS creation,
 * the conformance to the integrity rules linking the parameters is verified.
 * <P>
 * The standard rights include the specific AcsAddon key 'Resource.OtherRights'.
 * It is a javabean with the following bound properties:
 * 'OneActorFromOtherInCurrentContext'.
 * @see ARoad0.gBase.BaseManagerImpl
 * @see ARoad0.AcsAddon.Accbee.MySQL.gBase.ACSFactoryMySQLImpl
 * @version 0.7.2
 */
public class ACSUbuntuImpl extends ACSImpl implements AcsAddon {
    /**
    * List of own Linux/Unix right-proxy Actors which have a non-null currentUID or currentGID
    * and are executable from 'other' rights which exist only in Linux/Unix ACS.
    * These Actors are not accessible if their direct and indirect directories
    * have not the right 'r_nxdir', but this is not checked during the filling of this list.
    */
    protected transient List  l_FromOtherInCurrentContextActors_ = null;

    /**
    * Default constructor used by valueOf() and the subclasses
    */
    public ACSUbuntuImpl() {
    }

    /**
    * Factory method to use for creating an instance.
    * Called by ACSFactoryImplUbuntu.created().
    * Does not register the new instance in BaseManagerImpl.
    * Calls ACSImpl.valueOfForGenericBase() that makes the main work.
    * @param _iSname is the IS name. May be null.
    * @param _parentName is the name of the parent node in the IS nametree.
    *        No use if _iSname is null. May be null.
    * @param _name is the acs name
    * @param _type is ACSImpl.GENERIC (default value), LINUX, another ACSImpl constant
    *   or another user-defined value. Must be non-null. No use if _l_structure is null:
    *   the type is then set to LINUX.
    * @param _m_structure is the map of booleans which defines the ACS structure. Is null
    *   for LINUX-type ACS.
    * @param _m_constraintsForStringValues is the map where
    *   the key is a String of an ACSObject interface (like Resource) ended by
    *   a bound property name (like ResourceType), with '.' as separator,
    *   and the value is a String array of the authorized values for this bound property
    *   (like {file, directory, executable, script...} for ResourceType in Linux). May be null.
    * @param _m_constraintsForNewObjects is the map where
    *   the key is a String following one of the 7 possible formats,
    *   and the value is a String array of the authorized values among 17. May be null.
    * @param _m_constraintsForRightValues is the map where
    *   the key is a String of an ACSObject interface (like Resource) ended by
    *   a bound property name (like OthersRights), with '.' as separator,
    *   and the value is a StringRight array of the authorized values for this bound property
    *   (like {d, r, w, x} for OthersRights in Unix). Never null.
    * @param _m_SpecializedRights is the map where
    *   the key is a quartet (object.domain.key.subkey)
    *   and the value is is a StringRight array of the authorized values. Never null.
    * @param _l_Rights array of acsrights and/or metarights which are workable. Never null.
    * @param _l_ImmutableRules list of the immutable rules derivated from the structure.
    *    Never null.
    * @param _m_CommentsOnProperties comments on each ACS property, to explain the simulation.
    *    May be null.
    * @param _m_AcsVocabulary defines several independant domains, like
    * how to name the constants in Gui2.ACSTree.M_ACS_TREE_UPPER_NODES for this ACS,
    * as names of the ACS nodes in the proper ACS internal frame, and,
    * if this ACS is an AcsAddon, what are the subclassed classes in the generic gBase.
    * May be null.
    * @return the new ACSUbuntuImpl
    * @exception BaseError if _l_structure lenght is not 20, if there is a rule about
    *   _l_structure which is not applied, or if the ACS
    *   is already known by BaseManagerImpl.
    */
    protected static ACSImpl valueOf(ImmutableName _iSname, ImmutableName _parentName,
    ImmutableName _name, String _type,
    SortedMap<String,Boolean> _m_structure,
    HashMap<String,String[]>  _m_constraintsForStringValues,
    HashMap<String,String[]>  _m_constraintsForNewObjects,
    HashMap<String,StringRight[]>  _m_constraintsForRightValues,
    HashMap<String,StringRight[]>  _m_SpecializedRights,
    StringRight[] _l_Rights, List<String> _l_ImmutableRules,
    HashMap<String,String[]>  _m_CommentsOnProperties,
    HashMap<String,String[]>  _m_AcsVocabulary) throws BaseError {

      if (_name == null) {
           throw new BaseError("in the ACS constructor, "+
             "the ACS name is null.");
      } // if

      String message = null;

      if (_m_structure != null)
          message = ACSFactoryUtilityImpl.controlAcsStructure(_m_structure);

      if (_type == null) throw new BaseError("in the ACS constructor, "+
        " the ACS type must be defined.");
      if (message != null) throw new BaseError(message);

      ACSUbuntuImpl result = new ACSUbuntuImpl();

      return valueOfForGenericBase(
        result,
        _iSname, _parentName,
        _name, _type,
        _m_structure,
        _m_constraintsForStringValues,
        _m_constraintsForNewObjects,
        _m_constraintsForRightValues,
        _m_SpecializedRights,
        _l_Rights, _l_ImmutableRules,
        _m_CommentsOnProperties,
        _m_AcsVocabulary);
    } // valueOf() constructor

    /**
    * Registers a new own resource with a type control,
    * and fires the bound properties 'M_OwnResources' or
    * 'M_ExternalResources', 'OneResource' and 'OneBaseObject' only if it is not
    * an Actor.
    * Overrides the super method and call it
    * @param _res is just created in a Resource constructor, or
    *   external resource which is associated to an aclEntry owned by this ACS
    * @exception CreateError (and baseError) if the Name is already known
    *  in ACSCatalog, if the Name is null, or if the resource type is not allowed
    *  by this ACS
    */
    protected void newEorResource(ImmutableResource _res) throws CreateError {
      super.newEorResource(_res);
    } // newEorResource()

    /**
    * Unregisters the own resources.
    * Fires property change events on "M_OwnResources" or "M_ExternalResources",
    * and fire "OneResource", "OneBaseObject".
    * Overrides the super method and call it
    * @param       _res internal or external open ImmutableResource
    *                to be taken off in the ACS.
    * @exception   UpDateError (and baseError) if the Name is unknown
    *                in the ACS.
    */
    protected void removeEorResource(ImmutableResource _res) throws UpDateError {
      super.removeEorResource(_res);
    } // removeEorResource()

    /**
    * Registers a new own eligible party in the ACS and
    * controls that the eligible party type is allowed. <P>
    * Fires a property change on "M_OwnEligibleParties" or "M_ExternalEligibleParties",
    *  "OneEligibleParty" and "OneBaseObject".
    * Overrides the super method and call it
    * @param       _ep is just created in EligibleParty()
    *                or in ActorImpl(), or associated to a newEorAclEntry() call
    * @exception   CreateError (and baseError) if the Name is already known
    *                in ACSCatalog, if the Name is null or if the ep type is not allowed.
    */
    protected void newEorEligibleParty(ImmutableEligibleParty _ep) throws CreateError {
      super.newEorEligibleParty(_ep);
    }

    /**
    * Removes the EligibleParty in m_EligibleParties_.
    * May updates the closed bridge targets map.
    * Fires property change on "M_OwnEligibleParties" or "M_ExternalEligibleParties",
    *  "OneEligibleParty", and "OneBaseObject" only if it is not an Actor.
    * Overrides the super method and call it
    * @param _ep eligible party to remove.
    * @exception UpDateError (and baseError) if the Name is unknown in the ACS.
    */
    protected void removeEorEligibleParty(ImmutableEligibleParty _ep) throws UpDateError {
      super.removeEorEligibleParty(_ep);
    }

    /**
    * Gets the actors with an 'execute' right in 'other' rights,
    * and having a non-null current account/group. If the Actor is an Executable,
    * it may be set in this list by an enforced account/group.
    * Called in the access paths search, then may be called from a view worker thread.
    * @return unmodifiable List of actors. May be empty. Never null.
    */
    public List getEorL_FromOtherInCurrentContextActors() {
        if (l_FromOtherInCurrentContextActors_ == null)
            return Collections.EMPTY_LIST;
        else
            return Collections.unmodifiableList(l_FromOtherInCurrentContextActors_);
    } // getEorL_FromOtherInCurrentContextActors()

    /**
     * Adds a right-proxy actor with an 'execute' right in 'other' rights,
     * and having a non-null current account/group. If Executable, this current value
     * may be set by an enforced account/group.
     * Fires the event 'OneActorFromOtherInCurrentContext'.
     * Called by ActorUbuntuImpl.setEorCurrentUser/GroupID() which checks up
     * the actor state before the calling.
     * @param _act complies with the criterion
     */
    protected synchronized void addActorFromOtherInCurrentContext(ImmutableActor _act) {
 
                                    // for testing and tracking
                                    //System.out.println("ACSUbuntuImpl."+
                                      //      "addActorFromOtherInCurrentContext() start");

        if (l_FromOtherInCurrentContextActors_ == null)
            l_FromOtherInCurrentContextActors_ = new ArrayList(INITIAL_CAPACITY_25);

        List old = (List)((ArrayList)l_FromOtherInCurrentContextActors_).clone();

        if (!l_FromOtherInCurrentContextActors_.contains(_act)) {
            l_FromOtherInCurrentContextActors_.add(_act);

            try {
                  firePropertyChange("OneActorFromOtherInCurrentContext",
                          old, ((ArrayList)l_FromOtherInCurrentContextActors_).clone());
            } catch (Exception e) {
                ARoad0.Gui1.DialogBox.showAlarmMessage(
                  "There is an exception after the data update, from outside the database."+
                  "You should save the base and stop the program, then run it once again.");
                System.out.println("From ACSImplUbuntu.addActorFromOtherInCurrentContext(), "+
                        "firePropertChange() throws:");
                e.printStackTrace();
            }
        } // if
        old = null;
    } // addActorFromOtherInCurrentContext()

    /**
     * Removes an actor with an 'execute' right in 'other' rights,
     * and having a non-null current account/group. If Executable, this current value
     * may be set by an enforced account/group.
     * Called by ActorUbuntuImpl.setEorCurrentUser/GroupID().
     * Fires the event 'OneActorFromOtherInCurrentContext'.
     * @param _act does no more comply the criteria
     */
    protected synchronized void removeActorFromOtherInCurrentContext(ImmutableActor _act) {

                                    // for testing and tracking
                                    //System.out.println("ACSUbuntuImpl."+
                                      //      "removeActorFromOtherInCurrentContext() start");

        if (l_FromOtherInCurrentContextActors_ == null)
            l_FromOtherInCurrentContextActors_ = new ArrayList(INITIAL_CAPACITY_25);

        List old = (List)((ArrayList)l_FromOtherInCurrentContextActors_).clone();

        if (l_FromOtherInCurrentContextActors_.contains(_act)) {
            l_FromOtherInCurrentContextActors_.remove(_act);

            try {
                  firePropertyChange("OneActorFromOtherInCurrentContext",
                          old, ((ArrayList)l_FromOtherInCurrentContextActors_).clone());
            } catch (Exception e) {
                ARoad0.Gui1.DialogBox.showAlarmMessage(
                  "There is an exception after the data update, from outside the database."+
                  "You should save the base and stop the program, then run it once again.");
                System.out.println("From ACSImpl.addActorFromOtherInCurrentContext(), "+
                        "firePropertChange() throws:");
                e.printStackTrace();
            }

                    // for testing and tracking
                    //System.out.println("in ACSImpl.removeActorFromOtherInCurrentContext() for "+_act);

        } // if
        old = null;
    } // removeActorFromOtherInCurrentContext()

    /**
     * Fills the ACS internal property list, after an ACS deserialization.
     * The list contains the right-proxy actors with an 'execute' right
     * in 'other' rights, and having a non-null current account/group.
     * On the contrary of getEorM_Actors(), an Actor is tested as instance,
     * and not through Resource.getActorActivated().
     * Called by BaseManagerImpl.restoreACSInternalLinks().
     */
    protected synchronized void fillFromOtherInCurrentContextActors() {
        if (l_FromOtherInCurrentContextActors_ == null)
            l_FromOtherInCurrentContextActors_ = new ArrayList(INITIAL_CAPACITY_25);

        Resource res = null;
        ActorUbuntuImpl act = null;

                                    // for testing and tracking
                                    //System.out.println("ACSImpl.fillFromOtherInCurrentContextActors()");

        for (Iterator ite = m_Resources_.values().iterator() ; ite.hasNext() ; ) {
            res = (Resource) ite.next();
            if ( !(res instanceof ActorUbuntuImpl)) continue;
            act = (ActorUbuntuImpl)res;
            if ( !act.getIsRightProxy()) continue;
            if ( act.getEorCurrentUserID() == null && act.getEorCurrentGroupID() == null)
                continue;
            if (!UtilityImpl.getFirstInstance().withExecuteRight(
                    act.getL_otherRights(), false)) continue;

            l_FromOtherInCurrentContextActors_.add(act);

                                    // for testing and tracking
                                    //System.out.println("--------- ");

        } // for ite

        res = null;
        act = null;
    } // fillFromOtherInCurrentContextActors()

    /**
     * Fills the ACS internal property list, after an ACS deserialization.
     * Called by BaseManagerImpl.restoreACSInternalLinks().
     * Calls fillFromOtherInCurrentContextActors()
     * Overrides the super method and call it
     */
    protected synchronized void fillNoRightProxyActors() {
        super.fillNoRightProxyActors();
        fillFromOtherInCurrentContextActors();
    } // fillNoRightProxyActors()

  /**
   * Gets the specific events to listen on a Resource or a VirtualFolder,
   * its direct and indirect parents,
   * because these events should change the rights of the Resource or VirtualFolder in an ACS Addon.
   * Then, these events are listened to update the views containing these ACSObjects.
   * This method covers only the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBaseInterface.ACSObjectAddon.getAddonInnerPropertiesToListen()
   * for the other cases).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single basic object is the Resource or VirtualFolder, <P>
   * - 'ACS.method_name' to say that the list of Resources (or VirtualFolders) is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: PARENT (mandatory), <P>
   * - method_name: 'get' method name of each basic objet parent
   * that is associated, in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: where BBB is 'TRUE', 'FALSE', 'NULL' or 'NOT-NULL', to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL (option). <P>
   * When the map key starts with 'ACS', the ACS method to call has no argument.
   * When the map value starts with 'PARENT', the parent method to call has no argument.
   * Called by ARoad0.gBase.EPRViewInBaseImpl, NoThanViewInBaseImpl
   * and ARoad0.gWork.RightsMediatorImpl.
   * For instance, 'PARENT.getAGOAccessFromTheParentOther IF FALSE' is the value
   * the ACS Addon Ubuntu ACS returns for the key ACS.getEorL_FromOtherInCurrentContextActors,
   * to request the listening of the 'FromTheParentOther' event on all the indirect parents
   * of a ResourceUbuntu in a view.
   * @return map to define the specific ACS Addon events to listen. May be null.
   */
  public Map<String, String> getAddonParentPropertiesToListen() {
    Map<String, String> m_result = new HashMap();
    m_result.put(
            "<THIS>",
            "<PARENT>.getAGOAccessFromTheParentOther<IF><FALSE>");
    m_result.put(
            "<THIS>",
            "<PARENT>.getL_inheritedOtherRights");
    m_result.put(
            "<ACS>.getEorL_FromOtherInCurrentContextActors",
            "<PARENT>.getAGOAccessFromTheParentOther<IF><FALSE>");
    return m_result;
  } // getAddonInheritedPropertiesToListen()

} // ACSUbuntuImpl

