//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.AcsAddon.Accbee.Ubuntu.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.Ubuntu.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.BaseError;

import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.util.Arrays;

import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * This class is an embedded one, responsible for modeling the AGO 'other' rights
 * and the sticky bit on Ubuntu directories. It does not handle all the AGO rights.
 * It is used as a private attribute of a composite class instance (called the master)
 * which delegates the processing of some methods to this class.
 * This class responsability covers both the AGO 'other' rights and the inherited group rights,
 * since they depends on the sticky bit.
 * The master class is a ResourceUbuntu,
 * and the master transfers directly all the relevant actions to this class. <P>
 * It is a javabean with the following bound properties:
 * 'OtherRights', 'FromTheParentOther', 'InheritedOtherRights'.
 * The listeners are proper to the master instance.
 * Excepting for ViewInBaseEPRImpl, they are outside the gBase packages,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * This class is public to be overriden in the AcsAddons.
 * This class has no subclass in this package.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @version 0.7.0
 */
public class CAgoUserUbuntuImpl implements Serializable {

  // protected variables
  protected static final int  INITIAL_CAPACITY = 10; // the default value: 101 for map
  protected static final int  INITIAL_CAPACITY_2 = 4 ;  // the default value is 10
  protected static final int  INCREMENT_CAPACITY = 2;
  protected static final int  PRIME = 1000003;

   // Private variables
  private static final long   serialVersionUID = 0;
  /**
   * The master is the instance for which this class operates as an embedded one.
   */
  private ResourceUbuntu      master_;
  /**
   * The other rights of the Linux/Unix resource
   */
  private StringRight[]       l_otherRights_;
  /**
   * True if the master parent allows to go to the master for AGO other rigths.
   * This property complements the 2 booleans in ResourceImpl.l_parentTreeAGRightsAllowToGoAcross_.
   */
  private boolean             parentTreeORightsAllowToGoAcross = true;

  /* Out of the extended interfaces */

   /**
   * manages all the property change listeners and also for its subclasses
   */
  protected transient PropertyChangeSupport changeSupport_ = null;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  static final CAgoUserUbuntuImpl EMPTY_INSTANCE = new CAgoUserUbuntuImpl();

  /**
   * variable for the JDK 2 serialization
   * @serial ("serialVersionUID",Long.TYPE),
   * ("master_",ResourceUbuntu.class),
   * ("l_otherRights_",StringRight[].class),
   * ("parentTreeORightsAllowToGoAcross",Boolean.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("serialVersionUID",Long.TYPE),
      new ObjectStreamField("master_",ResourceUbuntu.class),
      new ObjectStreamField("l_otherRights_",StringRight[].class),
      new ObjectStreamField("parentTreeORightsAllowToGoAcross",Boolean.class)
  };

   /**
   * Default constructor. No registration in the master.
   */
  CAgoUserUbuntuImpl(){

                          // for testing purpose
                     //System.out.println("Call to CAgoUserUbuntuImpl constructor");

  }

   /**
   * Used by the master constructor before the registration in the ACS.
   * @param _master which owns this instance
   * @exception CreateError if _master is null.
   *   Does not control the ACS management.
   */
  CAgoUserUbuntuImpl(ResourceUbuntu _master) throws CreateError {
        if (_master == null)
         throw new CreateError("At the CAgoUserImplUbuntu construction, "+
            "the master is null.");

        master_ = _master;
        l_otherRights_ = new StringRight[0];

        if (master_.getEorParent() != null) {

                          // for testing purpose
                     //System.out.println("Call to CAgoUserUbuntuImpl constructor "+
                          //   "with the parent class: "+master_.getEorParent().getClass());

            parentTreeORightsAllowToGoAcross =
              BaseUtilityImpl.withAcrossToDirectoryAGRights(
                ((ResourceUbuntu)master_.getEorParent()).getL_otherRights());
        } // if master_ has a parent

                          // for testing purpose
                     //System.out.println("Call to CAgoUserUbuntuImpl constructor ");

  } // CAgoUserUbuntuImpl constructor

 //
 // -----------  Methods of the implemented interfaces --------------------
 //

  /**
  * This method provides a copy of the instance with another master.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * @param _masterCopy is the master of the new copy
  * @return a new instance of CAgoUserUbuntuImpl
  * @exception CreateError of the constructor with the same parameters.
  **/
  CAgoUserUbuntuImpl copy(ResourceUbuntu _masterCopy) throws CreateError {

      CAgoUserUbuntuImpl result = new CAgoUserUbuntuImpl(
        _masterCopy);

      result.l_otherRights_ = new StringRight[l_otherRights_.length];
      System.arraycopy(l_otherRights_, 0, result.l_otherRights_, 0, l_otherRights_.length);

      result.parentTreeORightsAllowToGoAcross = parentTreeORightsAllowToGoAcross;
      return result;
  } // copy(,,,)

  /**
   * Gets the access rights for the EligibleParties which are not
   * the UserID or the GroupID owners of this resource, nor the root UserID.
   * Does not take in account if the parent tree allows or not any access.
   * @return the AGO 'other' rights, following the Unix/Linux scheme. No null value.
   *  May be empty, or null only if the instance is finalized.
   */
  protected StringRight[] getL_otherRights()	{
      if (l_otherRights_ == null) return null;
      StringRight[] result = new StringRight[l_otherRights_.length];
      System.arraycopy(l_otherRights_, 0, result, 0, l_otherRights_.length);
      return result;
  }

  /**
   * Sets the rights for the EligibleParties which are not 'root'
   * nor this resource userID, and which are not members of this resource groupID.
   * The rights may have the two sens to implement Umask, and may be acsrights or metarights.
   * The ACS constraints on 'OtherRights' are applied only if they are not null,
   * otherwise all the arguments are validated.
   * The negative rights are subtracted from the positive rights.
   * No action if the instance has been finalized. If a new right
   * is not allowed by the ACS, it is removed and the other rights are kept.
   * Noop if the resource type contains 'byAcsCreator' or 'immutable'.
   * No adding of null or negative elements.
   * Calls userID.getUMask() to add the denyings rights of the account.
   * Calls BaseUtilityImpl.subtractNegativeFromPositiveInOneACS() and
   * UtilityImpl.selectHeaders().
   * Called by the master. If the master is an ActorUbuntuImpl, calls
   * ACS.addActorFromOtherInCurrentContext(). If the master is a DirectoryUbuntuImpl,
   * calls on its children for the methods setAGOAccessFromTheParentOther() and
   * fireL_inheritedOtherRights().
   * Fires a PropertyChange for 'OtherRights'.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the other rights are immutable,
   * or if there are more than 20 increasing chains to analyze from one lesser right, or
   * if the new rights contain a 'delete' or 'rename' right,
   * while its parent does not allow it through its deleting-renaming flag.
   */
  protected void setL_otherRights(StringRight[] _s) throws UpDateError {

                          // for testing and tracking purpose
                     //System.out.println("call to CAgoUserUbuntuImpl.setL_otherRights() for "+
                       //      master_.getName());

    if (l_otherRights_ == null) return;
                     
    if (master_.getType().startsWith("<byAcsCreator>") ||
            master_.getType().startsWith("<immutable> ") ||
            master_.getType().endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+master_.getName().lastComponent()+
            " sets it as immutable for its other rights.");

    // test the stickybit
    if (master_.getEorParent() instanceof DirectoryUbuntu &&
            ((DirectoryUbuntu)master_.getEorParent()).getStickyBit() && _s != null) {
       boolean sticky = false;
       for (int i = 0; i < _s.length; i++) {
           if (_s[i] != null &&
                   BaseUtilityUbuntuImpl.getDeleteRenameRights(
                    master_.getEorACS()).contains(_s[i])) {
               sticky = true;
               break;
           }
       } // for i
       if (sticky)
              throw new UpDateError(
                master_.getName().lastComponent()+
                " new rights contain a 'delete' or 'rename' right - "+
                "its parent does not allow it, through its deleting-renaming flag.");
    } // if (master_


                          // for testing and tracking purpose
                     //System.out.println("call to CAgoUserUbuntuImpl.setL_otherRights() for "+getName());

    // variables
    StringRight[]       l_oldRights = new StringRight[l_otherRights_.length];
    Set                 l_newRights = new HashSet(INITIAL_CAPACITY / 2);
    StringRight[]       l_restrictedRights = null;
    List<StringRight>   l_limitedRights = null;
    Set<StringRight>    l_posRights = new HashSet(INITIAL_CAPACITY / 2);
    Set<StringRight>    l_negRights = new HashSet(INITIAL_CAPACITY / 2);
    StringRight         right = null;
    boolean             toThrow = false;
    StringRight[]       l_oldInheritance = null;
    StringRight[]       l_newInheritance = null;
    ImmutableResource[] l_Children = null;
    boolean             allowToGoAcross = false;
    boolean             oldAllowToGoAcross = BaseUtilityImpl.withAcrossToDirectoryAGRights(
                                                    getL_otherRights());

    System.arraycopy(l_otherRights_, 0, l_oldRights, 0, l_otherRights_.length);

    if (master_ instanceof DirectoryUbuntu) {
        l_Children = ((DirectoryUbuntu)master_).getEorL_Children();

        if (l_Children != null) {
            l_oldInheritance = new StringRight[l_Children.length];
            
            for (int  i=0 ; i < l_Children.length ; i++) {
              if (l_Children[i] == null ) continue;
              // l_oldInheritance is the same for all the children
              l_oldInheritance = ((ResourceUbuntu)l_Children[i]).getL_inheritedOtherRights();
              break;
            } // for
        } // if
    } // if (master_ instanceof DirectoryUbuntu)

    try {
        // get the rights restriction in the ACS
        l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                   null, master_, "OtherRights", master_.getEorACS(), null);
    } catch (Exception c) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CAgoUserImplUbuntu.setL_otherRights(), "+
                "BaseUtilityImpl.getL_AcsRestrictedRights() throws:");
        c.printStackTrace();
    } // catch

    if (l_restrictedRights != null) {
        l_limitedRights = Arrays.asList( l_restrictedRights );


                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.setL_otherRights() in "+
                      //      master_.getName()+
                        //    " where the restricted rights from the ACS are:");
                    //for (Object obj : l_restrictedRights) {
                      //  System.out.println("------ "+obj);
                    //}

    } // if

    if (_s != null) {
      for (int i = 0; i < _s.length; i++) {
        if (_s[i] != null) {
            if (l_limitedRights == null ||
                    (l_limitedRights != null && l_limitedRights.contains(_s[i])))
                l_newRights.add(_s[i]);
        } // if
      } // for i
    } // if

    // filters the positive and negative rights
    if (!l_newRights.isEmpty()) {
        for (Iterator ite = l_newRights.iterator(); ite.hasNext(); ) {
            right = (StringRight)ite.next();
            if (right.isGrantingRight()) l_posRights.add(right);
            else l_negRights.add(right);
        }
    } // if

    // subtracts the negative rights to the positive ones
    if (!l_negRights.isEmpty())
        l_newRights = UtilityImpl.getFirstInstance().subtractNegativeNotRemovingInOneACS(
           l_posRights, l_negRights, master_.getEorACS(), this instanceof ImmutableDirectory);
    else l_newRights = l_posRights;

                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.setL_otherRights() in "+
                      //      master_.getName()+", l_newRights before selectHeaders():");
                    //for (Object obj : l_newRights) {
                      //  System.out.println("------------ "+obj);
                    //}


    try {
        UtilityImpl.getFirstInstance().selectHeaders(l_newRights);

                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.setL_otherRights() in "+
                      //      master_.getName()+", l_newRights after selectHeaders():");
                    //for (Object obj : l_newRights) {
                     //   System.out.println("------------ "+obj);
                    //}


        // update the other rights
        if (!l_newRights.isEmpty()) {
          l_otherRights_ = (StringRight[])
                  l_newRights.toArray(new StringRight[l_newRights.size()]);
        } else {
          l_otherRights_ = new StringRight[0];
        }

                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.setL_otherRights() in "+
                      //      master_.getName());
                    //for (int i = 0; i < l_otherRights_.length; i++) {
                      //  System.out.println("------------ "+l_otherRights_[i]);
                    //}

        if (l_oldRights.length != l_otherRights_.length )
            toThrow = true;
        else if (l_oldRights.length == l_otherRights_.length &&
                l_otherRights_.length == 0)
            toThrow = true;
        else {
            for (int i = 0; i < l_oldRights.length; i++) {
                if (!l_newRights.contains( l_oldRights[i] ) ) {
                    toThrow = true;
                    break;
                }
            } // for i
        } // else

        if (toThrow)
            firePropertyChange("OtherRights", l_oldRights, getL_otherRights());

    } catch (BaseError b) {
        throw new InternalError(b.toString());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CAgoUserImplUbuntu.setL_otherRights(), firePropertChange() throws:");
        e.printStackTrace();
    }

    // an Actor may have to be declared in its ACS through addActorFromOtherInCurrentContext()
    if (master_ instanceof ActorUbuntu && l_otherRights_ != null) {

                            // for testing and tracking
                            //System.out.println("CAgoUserUbuntuImpl.setL_otherRights() in "+
                              //      master_.getName());
                            //System.out.println("-- It is right-proxy: "+
                              //      ((ActorUbuntu)master_).getIsRightProxy());
                            //System.out.println("-- with a current UserID: "+
                              //      ((ActorUbuntu)master_).getEorCurrentUserID());
                            //System.out.println("-- with a current GroupID: "+
                              //      ((ActorUbuntu)master_).getEorCurrentGroupID());

            // add/remove to the ACS list if it is executable from 'other'
        if (((ActorUbuntu)master_).getIsRightProxy() &&
                (((ActorUbuntu)master_).getEorCurrentUserID() != null ||
                        ((ActorUbuntu)master_).getEorCurrentGroupID() != null)) {

                     if (UtilityImpl.getFirstInstance().withExecuteRight(getL_otherRights(), false))
                          ((ACSUbuntuImpl)master_.getEorACS()).
                                  addActorFromOtherInCurrentContext((ActorUbuntu)master_);
                     else ((ACSUbuntuImpl)master_.getEorACS()).
                             removeActorFromOtherInCurrentContext((ActorUbuntu)master_);

        } // if (((ActorUbuntu)master_).getIsRightProxy()
    } // if (master_ instanceof ActorUbuntu

    // a Directory should have to update the inherited other rights of its children
    if (master_ instanceof DirectoryUbuntu) {
        if (l_Children == null) {
            l_oldRights = null;
            l_newRights = null;
            l_limitedRights = null;
            l_restrictedRights = null;
            l_posRights = null;
            l_negRights = null;
            right = null;
            l_oldInheritance = null;
            l_Children = null;
            return;
        } // if (l_Children == null) {

        allowToGoAcross = BaseUtilityImpl.withAcrossToDirectoryAGRights(getL_otherRights());

        if ((oldAllowToGoAcross && !allowToGoAcross) ||
                (!oldAllowToGoAcross && allowToGoAcross)) {
                    // the boolean CAgoUserUbuntuImpl.parentTreeORightsAllowToGoAccess
                    //  must change for each child
                    for (int  i=0 ; i < l_Children.length ; i++) {
                      if (l_Children[i] == null ) continue;
                      ((ResourceUbuntu)l_Children[i]).
                              setAGOAccessFromTheParentOther( allowToGoAcross );
                    } // for
        } // if ((oldAllowToGoAcross &&

        for (int  i=0 ; i < l_Children.length ; i++) {
          if (l_Children[i] == null ) continue;
          // l_newInheritance is the same for all the children
          l_newInheritance = ((ResourceUbuntu)l_Children[i]).getL_inheritedOtherRights();
          break;
        } // for

        if (l_Children.length != 0 &&
                 !Arrays.asList(l_oldInheritance).equals(Arrays.asList(l_newInheritance)) ) {

                    // for testing purpose
                    //System.out.println("in CAgoUserUbuntuImpl.setL_otherRights(): "+
                      //"call to l_Children_.fireL_inheritedOtherRights()");

            // the other rights have changed for each child
            for (int  i=0 ; i < l_Children.length ; i++) {
              if (l_Children[i] == null ) continue;
              ((ResourceUbuntu)l_Children[i]).fireL_inheritedOtherRights(
                      l_oldInheritance, l_newInheritance);
            } // for
        } // if
    } // if (master_ instanceof DirectoryUbuntu) {

    l_oldRights = null;
    l_newRights = null;
    l_limitedRights = null;
    l_restrictedRights = null;
    l_posRights = null;
    l_negRights = null;
    right = null;
    l_oldInheritance = null;
    l_newInheritance = null;
    l_Children = null;
  } // setL_otherRights()

  /**
   * Gets the inherited rights from the parent of the resource group, including the
   * control by the parent sticky bit.
   * Calls BaseUtilityImpl.getChildRightsSetFromParentAGRights(),
   * BaseUtilityImpl.subtractNegativeNotRemovingInOneACS() and
   * BaseUtilityImpl.getDenyDeleteRenameRights().
   * @return an array of access rights for the GroupID which contains this resource,
   *   in the main ACS. May be empty, or null only if the instance is finalized.
   */
  protected StringRight[] getL_inheritedGroupRights() {
    if (master_.getEorParent() == null) return new StringRight[0];

                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.getL_inheritedGroupRights(): "+
                     //       "call to UtilityImpl.getChildRightsSetFromParentAGRights() "+
                      //      "from the parent "+parent_.getName());


    Set<StringRight> l_rights = BaseUtilityUbuntuImpl.getChildRightsSetFromParentAGORights(
            ((DirectoryImpl)master_.getEorParent()).getL_groupRights(), master_.getEorACS());

    if ( ((DirectoryUbuntuImpl)master_.getEorParent()).getStickyBit() ) {

                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.getL_inheritedOtherRights(): "+
                      //      "parent_.getStickyBit() true, and result of "+
                        //    "UtilityImpl.getChildRightsSetFromParentAGORights():");
                    //if (name_.lastComponent().equals("f_jerry")) {
                    //for (Iterator ite = l_rights.iterator(); ite.hasNext(); ) {
                      //  System.out.println("--------------- "+ite.next());
                    //}
                    //} // if

        // remove the 'delete', 'rename' rights
        StringRight[] l_res = UtilityImpl.getFirstInstance().subtractNegativeNotRemovingInOneACS(
                                l_rights,
                                BaseUtilityUbuntuImpl.getDenyDeleteRenameRights(master_.getEorACS()),
                                master_.getEorACS(),
                                this instanceof Node
                                ).toArray(new StringRight[0]);

                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.getL_inheritedGroupRights(): result of "+
                      //      "BaseUtilityImpl.subtractNegativeNotRemovingInOneACS():");
                    //if (name_.lastComponent().equals("f_jerry")) {
                    //for (int i = 0; i < l_res.length; i++) {
                      //  System.out.println("--------------- "+l_res[i]);
                    //}
                    //} // if

        l_rights = null;
        return l_res;

    } else {

                // for testing and tracking
                //if (name_.lastComponent().equals("f_jerry")) {
                  //  for (int i = 0; i < l_res.length; i++) {
                    //    System.out.println("for f_jerry, inherited group rights: "+l_res[i]);
                  //  }
                //} // if
      if (!l_rights.isEmpty()) return l_rights.toArray(new StringRight[0]);
      else {
          l_rights = null;
          return new StringRight[0];
      }
    } // else
  } // getL_inheritedGroupRights()

  /**
   * Gets the inherited rights from the parent of the resource for 'other' AGO rights.
   * Calls BaseUtilityUbuntuImpl.getChildRightsSetFromParentAGORights(),
   * BaseUtilityImpl.subtractNegativeNotRemovingInOneACS() and
   * BaseUtilityImpl.getDenyDeleteRenameRights().
   * @return an array of access rights for the 'other' AGO category,
   *   in the main ACS. May be empty, or null only if the instance is finalized.
   */
  protected StringRight[] getL_inheritedOtherRights() {
    if (master_.getEorParent() == null) return new StringRight[0];

                    /*
                      // for testing purpose
                  System.out.println("in CAgoUserUbuntuImpl.getL_inheritedOtherRights(), call from "+name_+
                        " to UtilityImpl.getChildRightsSetFromParentAGORights() with the parent rights: ");
                  StringRight[] l_right = ((DirectoryImpl)parent_).getL_otherRights();
                  for (int i = 0; i < l_right.length; i++) {
                      System.out.println("------ for "+i+", the value _l_parentRights[i] is "+l_right[i]+
                      " and the metaright is "+l_right[i].getEorMetaRight());
                  } // for i
                  l_right = null;
                     * */


    Set<StringRight> l_rights = BaseUtilityUbuntuImpl.getChildRightsSetFromParentAGORights(
            ((ResourceUbuntu)master_.getEorParent()).getL_otherRights(), master_.getEorACS());

    if ( ((DirectoryUbuntuImpl)master_.getEorParent()).getStickyBit() ) {


        // remove the 'delete', 'rename' rights
        StringRight[] l_res = UtilityImpl.getFirstInstance().subtractNegativeNotRemovingInOneACS(
                                l_rights,
                                BaseUtilityUbuntuImpl.getDenyDeleteRenameRights(master_.getEorACS()),
                                master_.getEorACS(),
                                this instanceof Node
                                ).toArray(new StringRight[0]);

                    // for testing and tracking
                    //System.out.println("CAgoUserUbuntuImpl.getL_inheritedOtherRights(): result of "+
                      //      "BaseUtilityImpl.subtractNegativeNotRemovingInOneACS():");
                    //if (name_.lastComponent().equals("f_jerry")) {
                    //for (int i = 0; i < l_res.length; i++) {
                      //  System.out.println("--------------- "+l_res[i]);
                    //}
                    //} // if

        l_rights = null;
        return l_res;

    } else {

                // for testing and tracking
                //if (name_.lastComponent().equals("f_jerry")) {
                  //  for (int i = 0; i < l_res.length; i++) {
                    //    System.out.println("for f_jerry, inherited group rights: "+l_res[i]);
                  //  }
                //} // if
      if (!l_rights.isEmpty()) return l_rights.toArray(new StringRight[0]);
      else {
          l_rights = null;
          return new StringRight[0];
      }
    } // else
  } // getL_inheritedOtherRights()

  /**
  * @return true if the parent authorizes its 'other' AGO category
  * to have access to this instance.
  */
  protected boolean getAGOAccessFromTheParentOther() {
      return parentTreeORightsAllowToGoAcross;
  }

  /**
  * Adds a listener to the bean, for all the bound properties.
  * Used by the master only.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  protected void addPropertyChangeListener(PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by the master for the properties:
  * 'OtherRights', 'FromTheParentOther', 'InheritedOtherRights'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  protected void addPropertyChangeListener(String _propertyName, PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}
  } // addPropertyChangeListener(String _propertyName,

  /**
  * Removes a listener to the bean, for all the bound properties.
  * Used by the master.
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("CAgoUserUbuntuImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_l);}
  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by the master for the properties:
  * 'OtherRights', 'FromTheParentOther', 'InheritedOtherRights'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("CAgoUserUbuntuImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()

 /**
  * Clone for embedded class, without registering the new object in the ACSs.
   * The external object references are not cloned (see "Eor" in get method).
   * Used by the master clone() method.
   * @param _master must be a clone of this instance master
   * @return instance with _master as master.
   */
  protected synchronized CAgoUserUbuntuImpl cloneC(ResourceUbuntu _master) {
    CAgoUserUbuntuImpl clon;
    try {
      clon = (CAgoUserUbuntuImpl) super.clone();

      clon.master_ = _master;

      clon.setL_otherRights(l_otherRights_);

    } catch (UpDateError u) {
      if (master_ == null) throw new InternalError("In CAgoUserImplUbuntu, the master is null");
      else {
        master_.getEorACS().getEorBaseError().addException(u.toString() +
          "CAgoUserImplUbuntu.clone() failed");
        throw new InternalError(u.toString());
      }
    } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
      if (master_ == null) throw new InternalError("In CAgoUserImplUbuntu, the master is null");
      else {
        master_.getEorACS().getEorBaseError().addException(e.toString() +
          "CAgoUserImplUbuntu.clone() failed");
        throw new InternalError(e.toString());
      }
    } // catch

    return clon;
  } // cloneC()

  /**
   * @return the hashcode value for this object
   */
  public int hashCode() {
    int result = 0;

    if (master_ != null) {
      result = "CAgoUserImplUbuntu".hashCode() + master_.hashCode();
    }

    return result;
  } // hashcode()

  /**
  * Works even if the instance is finalized.
  * @param _obj to compare
  * @return true if _obj is an instance of CAgoUserUbuntuImpl (or a derived class)
  *   equals to this instance (with the same class)
  */
  public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      CAgoUserUbuntuImpl obj = (CAgoUserUbuntuImpl) _obj;

      if (master_ != obj.master_ )
          return false;

      return true;
  } // equals

  /**
   * @return long
   */
  public long getSerialVersionUID()	{
      return serialVersionUID;
  }

  /**
   * Gets the name last component of each privilege. Never null.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("AGO_USER ");

    return " " + sb;
  } // toString()

  /**
  * @param _b true if the parent authorizes its 'other' AGO category 
   * to have access to this instance.
  */
  protected void setAGOAccessFromTheParentOther(boolean _b) {
      boolean old = parentTreeORightsAllowToGoAcross;
      parentTreeORightsAllowToGoAcross = _b;
      try {
        firePropertyChange("FromTheParentOther", old, _b);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the data update, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImplUbuntu.setAGOAccessFromTheParentOther(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
      }
  } // setAGOAccessFromTheParentOther()

  /**
   * Fires the event 'InheritedOtherRights', from the parent of the resource 'other'
   * category. Called by the parent DirectoryUbuntuImpl.setL_otherRights() method.
   * @param _l_old an array of the old access rights for the 'other' AGO category,
   *   in the main ACS. May be empty.
   * @param _l_new an array of the new access rights for the 'other' AGO category,
   *   in the main ACS. May be empty.
   */
  protected final void fireL_inheritedOtherRights(StringRight[] _l_old, StringRight[] _l_new) {
    if (master_.getEorParent() != null)
        firePropertyChange("InheritedOtherRights", _l_old, _l_new);
  } // fireL_inheritedOtherRights()

   /**
   * This method is called by finalize(), but finalize() may be not called.
   * An unused object has to be deleted to inform the others objects, and this method
   * HAS TO BE CALLED before deleting the instance.
   * This method does not raise PropertyChangeEvents since the BeanInfo for this class
   * specifies that its ACS informs the listeners when this instance is finalized.
   * To test if an instance has been finalized, uses (master_ == null).
   * @exception UpDateError is never thrown.
   */
  protected synchronized void finalizeForBase() throws UpDateError {

                //System.out.println("In CAgoUserUbuntuImpl, call of " + name_ +
                    //".finalizeForBase()");

    if (master_ == null) return;
    master_ = null;
    l_otherRights_ = null;
    parentTreeORightsAllowToGoAcross = true;
  } // finalizeForBase()

  /**
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called for EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {


                //System.out.println("In CAgoUserUbuntuImpl, call of " + name_ +
                    //".finalizeFromDeserialization()");

    if (master_ == null) return;
    master_ = null;
    l_otherRights_ = null;
    parentTreeORightsAllowToGoAcross = true;
  } // finalizeFromDeserialization()

  /**
  * Fires an event to every registered listener, in any order.
  * For the bound properties 'OtherRights', 'FromTheParentOther', 'InheritedOtherRights'.
  * Called by all the updating methods.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

                // for testing and tracking
                //System.out.println("in CAgoUserUbuntuImpl.firePropertyChange(), call to "+
                  //"propertyChange() of "+master_.getNickName()+" for the property "+
                  //_propertyName+" with the old value "+_oldValue);

      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
        {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
   * This method calls finalizeForBase(). There is no
   * guaranee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {u.printStackTrace();}

    super.finalize();
  } // finalize()


} // CAgoUserUbuntuImpl



