//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.Ubuntu.gBase;

import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;
import ARoad0.AcsAddon.Accbee.Ubuntu.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;

import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * This class is responsible for modeling a Linux Ubuntu resource.
 * It uses an embedded instance of CAgoUserUbuntuImpl to handle the AGO
 * 'Other' rights and sticky bit in a Linux/Unix operating system.
 * It has an umask from its UserIDUbuntu, to set the default AGO rights.
 * It is a javabean with the following bound properties:
 * 'OtherRights', 'FromTheParentOther', 'InheritedOtherRights'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched
 * in the 'set' method, and a dialog box is displayed to inform the user.
 * This class has no subclass in this package.
 * All variables are protected.
 * @version 0.7.0
 */
public class ResourceUbuntuImpl extends ResourceImpl implements ResourceUbuntu {
  // private variables
  private static final long   serialVersionUID__ = 0;
  // protected variables
  /*
   * 'Other' rights and sticky bit manager in a Linux/Unix operating system
   */
  protected CAgoUserUbuntuImpl agoHandler_;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ResourceUbuntuImpl EMPTY_INSTANCE = new ResourceUbuntuImpl();

  /**
   * variable for the JDK 2 serialization
   * @serial ("agoHandler_",CAgoUserUbuntuImpl.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("agoHandler_",CAgoUserUbuntuImpl.class)
  };
  
   /**
   * Default constructor. No registration in the ACS.
   */
  public ResourceUbuntuImpl() {
    
                          // for testing purpose
                     //System.out.println("Call to ResourceUbuntuImpl constructor");
  }

   /**
   * Used by the beamer editor for the ACS resources list, from the 'New' button,
   * and by the gWork.AcsFactory.
   * @param _name of the resource
   * @param _acs main ACS of the resource
   * @param _parent of the resource; may be null.
   * @param _owner owns the resource; may be null.
   * @param _resourceType the resource type which drives the rights and the immutability
   * of the instance. If null, the value is set to "<undefined>".
   * @exception CreateError if ACSImpl.newEorResource() throws it;
   * the instance has then to be deleted !
   */
  public ResourceUbuntuImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType) throws CreateError {

      this(_name, _acs, (Directory)_parent, _owner, _resourceType, true);
        
                          // for testing purpose
                     //System.out.println("Call to ResourceUbuntuImpl constructor("+
                       // "ImmutableName _name, ACS _acs, Node _parent, String _resourceType)"+
                       // " for "+_name);

  } // ResourceImpl constructor

   /**
   * Used by the other constructor.
   * If the argument _owner is not null, calls BaseUtilityUbuntuImpl.subtractUMask()
   * to apply the 'umask' to the AGO rights.
   * Calls first the super constructor.
   * @see BaseUtilityUbuntuImpl#subtractUMask
   * @param _name of the resource
   * @param _acs main ACS of the resource
   * @param _parent owns the resource; may be null.
   * @param _owner owns the resource; may be null.
   * @param _resourceType the resource type which drives the rights and the immutability
   * of the instance. If null, the value is set to "<undefined>".
   * @param _toRegisterInManager false if this method must not register
   *   the instance in its ACS.
   * @exception CreateError if ACSImpl.newEorResource() throws it;
   * the instance has then to be deleted !
   */
  public ResourceUbuntuImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType, boolean _toRegisterInManager) throws CreateError {

      super(_name, _acs, (Directory)_parent, _owner, _resourceType, false);

      agoHandler_ = new CAgoUserUbuntuImpl(this);

                          // for testing purpose
                     //System.out.println("Call to ResourceUbuntuImpl constructor"+
                       // "ImmutableName _name, ACS _acs, Node _parent, String _resourceType)"+
                       // " for "+_name);

      // updates the AGO rights if userID_ != null
      if (userID_ != null) {
          // changes type_ to avoid a noop, but does not change the Basic type
          String temp = null;
          if (type_.startsWith("<byAcsCreator>")) {
              temp = type_;
              type_ = "12345678901234567890";
          } else if (type_ .indexOf("<immutable> ") != -1) {
              temp = type_;
              type_ = "01234567891234567890";
          }
          try {
              Set<StringRight> l_Rights = new HashSet(INITIAL_CAPACITY);
              l_Rights.add(aCS_.getFullControlRight());

              // updates the AGO rights
              setL_ownerRights(
                      BaseUtilityUbuntuImpl.subtractUMask(l_Rights, aCS_, this,
                            "owner").toArray(new StringRight[0]));

              setL_groupRights(
                      BaseUtilityUbuntuImpl.subtractUMask(l_Rights, aCS_, this,
                            "group").toArray(new StringRight[0]));
              // updates the AGO other rights
              setL_otherRights(
                      BaseUtilityUbuntuImpl.subtractUMask(l_Rights, aCS_, this,
                            "other").toArray(new StringRight[0]));

          } catch (UpDateError u) {
            throw new InternalError(u.toString());
          }

          if (type_.equals("12345678901234567890") ||
             type_.equals("01234567891234567890")) type_ = temp;
          temp = null;

      } // if (userID_ != null
      // adds this to the map of resources in the main ACS
      // and sends an event "OneResource" to the GUI listeners.
      // may throw a CreateError
      try {
          if (_toRegisterInManager) ((ACSUbuntuImpl)aCS_).newEorResource(this);
      } catch (CreateError c) {
              try {
                  finalizeForBase();
              } catch (UpDateError r) {} 
              throw c;
      } // catch

  } // ResourceUbuntuImpl constructor

  /**
   * Changes the AGO rights from the _u umask, through the call to
   * setL_xxxRights().
   * Overrides the super method and calls it at the start of the method.
   * @param _u owns this resource, in the main ACS. May be null.
   * @exception UpDateError if _u does not belong to this resource main ACS, or
   *               if it is immutable for its userID.
   */
  public synchronized void setEorUserID(ImmutableUserID _u) throws UpDateError {

      super.setEorUserID(_u);

      // userID umask is used
      Set<StringRight> l_Rights = new HashSet(INITIAL_CAPACITY);
      l_Rights.add(aCS_.getFullControlRight());

      // updates the AGO rights
      if (_u != null) {
          setL_ownerRights(BaseUtilityUbuntuImpl.subtractUMask(l_Rights, aCS_, this,
                            "owner").toArray(new StringRight[0]));

          setL_groupRights(BaseUtilityUbuntuImpl.subtractUMask(l_Rights, aCS_, this,
                            "group").toArray(new StringRight[0]));
          setL_otherRights(BaseUtilityUbuntuImpl.subtractUMask(l_Rights, aCS_, this,
                            "other").toArray(new StringRight[0])); 
      }
      else {
          setL_ownerRights(null);
          setL_groupRights(null);
          setL_otherRights(null);               
      } // else;

      l_Rights = null;
  } // setEorUserID()

  /**
   * Sets the rights for the members of the groupID which contains this resource.
   * Fires a PropertyChange for 'GroupRights'.
   * Overrides the super method and call it
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the new rights contain a 'delete' or 'rename' right,
   * while its parent does not allow it through its deleting-renaming flag.
   */
  public synchronized void setL_groupRights(StringRight[] _s) throws UpDateError {

                          // for testing and tracking purpose
                        //System.out.println("call to ResourceUbuntuImpl.setL_groupRights() for "+name_);

    // tests the stickybit
    if (parent_ instanceof DirectoryUbuntuImpl &&
            ((DirectoryUbuntuImpl)parent_).getStickyBit() && _s != null) {
       boolean sticky = false;
       for (int i = 0; i < _s.length; i++) {
           if (_s[i] != null &&
                   BaseUtilityUbuntuImpl.getDeleteRenameRights(aCS_).contains(_s[i])) {
               sticky = true;
               break;
           }
       } // for i
       if (sticky)
              throw new UpDateError(
                getName().lastComponent()+" new rights contain a 'delete' or 'rename' right - "+
                    "its parent does not allow it, through its deleting-renaming flag.");
    } // if (parent_

    super.setL_groupRights(_s);
  } // setL_groupRights()

  /**
   * Gets the access rights for the EligibleParties which are not
   * the UserID or the GroupID owners of this resource, nor the root UserID.
   * Does not take in account if the parent tree does not allow any access.
   * @return the AGO 'other' rights, following the Unix/Linux scheme. No null value.
   *  May be empty, or null only if the instance is finalized.
   */
  public final synchronized StringRight[] getL_otherRights()	{
      if (agoHandler_ == null) return null;
      else return agoHandler_.getL_otherRights();
  }

  /**
   * Sets the rights for the EligibleParties which are not 'root'
   * nor this-resource UserID, and which are not members of this-resource GroupID.
   * The rights may have the two sens to implement the umask, and may be acsrights or metarights.
   * The ACS constraints on 'OtherRights' are applied only if they are not null,
   * otherwise all the arguments are validated.
   * The negative rights are subtracted from the positive rights.
   * No action if the instance has been finalized. If a new right
   * is not allowed by the ACS, it is removed and the other rights are kept.
   * Noop if the resource type contains 'byAcsCreator' or 'immutable'.
   * No adding of null or negative elements.
   * Calls directly the CAgoUserUbuntuImpl embedded instance.
   * Called by the constructor and the GUI.
   * Fires a PropertyChange for 'OtherRights'.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the other rights are immutable,
   * or if there are more than 20 increasing chains to analyze from one lesser right, or
   * if the new rights contain a 'delete' or 'rename' right,
   * while its parent does not allow it through its deleting-renaming flag.
   */
  public synchronized void setL_otherRights(StringRight[] _s) throws UpDateError {

                          // for testing and tracking purpose
                     //System.out.println("call to ResourceUbuntuImpl.setL_otherRights() for "+getName());

      if (agoHandler_ == null) return;
      else agoHandler_.setL_otherRights(_s);
  } // setL_otherRights()
  
  /**
   * Gets the inherited rights from the parent of the resource group, including the
   * parent sticky bit. 
   * Calls BaseUtilityImpl.getChildRightsSetFromParentRights(),
   * BaseUtilityImpl.subtractNegativeNotRemovingInOneACS() and
   * BaseUtilityImpl.getDenyDeleteRenameRights().
   * @return an array of access rights for the GroupID which contains this resource,
   *   in the main ACS. May be empty, or null only if the instance is finalized.
   */
  public synchronized StringRight[] getL_inheritedGroupRights() {
      if (agoHandler_ == null) return null;
      else return agoHandler_.getL_inheritedGroupRights();
  } // getL_inheritedGroupRights()
   
  /**
   * Gets the inherited rights from the parent of the resource for 'other' AGO rights.
   * Calls BaseUtilityImpl.getChildRightsSetFromParentRights(),
   * BaseUtilityImpl.subtractNegativeNotRemovingInOneACS() and
   * BaseUtilityImpl.getDenyDeleteRenameRights().
   * @return an array of access rights for the 'other' AGO category,
   *   in the main ACS. May be empty, or null only if the instance is finalized.
   */
  public synchronized StringRight[] getL_inheritedOtherRights() {
      if (agoHandler_ == null) return null;
      else return agoHandler_.getL_inheritedOtherRights();
  } // getL_inheritedOtherRights()

  /**
  * @return true if the parent authorizes its 'other' AGO category 
  * to have access to this instance. The default value is true.
  */
  public boolean getAGOAccessFromTheParentOther() {
      if (agoHandler_ == null) return true;
      else return agoHandler_.getAGOAccessFromTheParentOther();
  }

 /**
  * Clones without registering of the new object in the ACSs.
   * Caution: don't use any updating method on this clone,
   * because that may destroy the base integrity. A finalized instance returns
   * a clone with getEorACS() which returns null,
   * but getName() and getAcsName() operates.
   * @exception CloneNotSupportedException
   * @return Object that is not registred in the ACSs. The other rights are cloned.
   *  The external object references are not cloned
   */
  public synchronized Object clone() {

      ResourceUbuntuImpl clon = (ResourceUbuntuImpl) super.clone();

      clon.agoHandler_ = agoHandler_.cloneC(clon);

      return clon;
  } // clone()

  /**
  * Does not compare the comment. Works even if the instance is finalized.
  * @param _obj
  * @return true if _obj is an instance of ResourceUbuntuImpl (or a derived class)
  *   equals to this instance (with the same class)
  */
  public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      ResourceUbuntuImpl obj = (ResourceUbuntuImpl) _obj;

      // agoHandler_ may be null if the instance is finalized
      if (agoHandler_ == null || obj.agoHandler_ == null) {
          if (agoHandler_!= null || obj.agoHandler_ != null)
            return false;
      } // if
      else {
          if (!agoHandler_.equals(obj.agoHandler_))
            return false;
      } // else
      
      obj = null;
      return true;
  } // equals()
  
   /**
   * This method removes the resource from the ACS lists (main and controllers),
   * from its parent, its userID and its groupID. The other rights are updated.
   * @exception UpDateError if the main ACS do not known the resource (if
   * it is a clone without setName(), typically)
   */
  public synchronized void finalizeForBase() throws UpDateError {

                    // for testing and tracking
                //System.out.println("In ResourceUbuntuImpl, call of " + name_ + ".finalizeForBase()");

    if (folderMember_ == null) return;
    
    if (agoHandler_ != null)
        agoHandler_.finalizeForBase();
    
    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * Adds a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(PropertyChangeListener _l) {
        if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

        synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}

        if (agoHandler_ == null ) {
            try {
              agoHandler_ = new CAgoUserUbuntuImpl(this);
            } catch (CreateError c) {
                new InternalError("From "+name_+"- "+c.toString());
            }
        } // if
        agoHandler_.addPropertyChangeListener(_l);

                // for testing and tracking
                //System.out.println("ResourceUbuntuImpl.addPropertyChangeListener(), call from "+
                  //  _l+" to "+name_+" for all the properties ");

  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'OtherRights', 'FromTheParentOther', 'InheritedOtherRights'.
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      
      if (_propertyName.equals("OtherRights") ||
              _propertyName.equals("FromTheParentOther") ||
              _propertyName.equals("InheritedOtherRights") ) {
                    if (agoHandler_ == null ) {
                        try {
                          agoHandler_ = new CAgoUserUbuntuImpl(this);
                        } catch (CreateError c) {
                            new InternalError("From "+name_+"- "+c.toString());
                        }
                    } // if

                            // for testing and tracking
                            //if (_propertyName.equals("InheritedOtherRights"))
                              //System.out.println("ResourceUbuntuImpl.addPropertyChangeListener(), "+
                                //      "call to agoHandler_ for 'InheritedOtherRights'");

                    agoHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else super.addPropertyChangeListener(_propertyName, _l);

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("ResourceUbuntuImpl.addPropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the adding of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //System.out.println("listeners of all properties after the adding of "+_l+": ");
                    //l_list = changeSupport_.getPropertyChangeListeners();
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //l_list = null;
                //} // if


  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean, for all the bound properties.
  * Used by ViewInBaseImpl.
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(PropertyChangeListener _l){

      // no action if the transiant variable is null
      // It is null when a PropertyEditor calls back a finalized instance.
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_l);}

      agoHandler_.removePropertyChangeListener(_l);
  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'OtherRights', 'FromTheParentOther', 'InheritedOtherRights'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("BasicImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("OtherRights") ||
              _propertyName.equals("FromTheParentOther") ||
              _propertyName.equals("InheritedOtherRights") ) {
                    if (agoHandler_ == null ) {
                        try {
                          agoHandler_ = new CAgoUserUbuntuImpl(this);
                        } catch (CreateError c) {
                            new InternalError("From "+name_+"- "+c.toString());
                        }
                    } // if

                    agoHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else super.removePropertyChangeListener(_propertyName, _l);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * Gets the PropertyChangeEvent for 'OtherRights' to listen from an ACSObject,
   * because these events may change the rights of an ACSObject in an ACS Addon.
   * The events are 'OtherRights' without conditions.
   * These events are listened to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBase.AcsAddon.getAddonParentPropertiesToListen()).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single object to listen is the ACSObject, <P>
   * - 'ACS.method_name' to say that the list of objects to listen is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: THIS (mandatory), <P>
   * - method_name: 'get' method name of the ACSObject that is associated,
   * in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: is optional, where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL. <P>
   * When the map value starts with 'ACS', the ACS method to call has the ACSObject
   * as one argument. Otherwise, the ACS or the ACSObject method has no argument.
   * Called by ARoad0.gBase.UtilityImpl.add/removeAddonParentPropertyChangeListener().
   * @return map to define the specific ACS Addon events to listen. May be null.
   */
  public Map<String, String> getAddonInnerPropertiesToListen() {
    Map<String, String> m_result = new HashMap();
    m_result.put(
            "<THIS>",
            "<THIS>.getL_otherRights");
    return m_result;
  } // getAddonInnerPropertiesToListen()

  //
  // ------------------ protected methods -----------------------------------------
  //

  /**
  * Called indirectly by DirectoryUbuntuImpl.setL_otherRights().
  * Caution: This method is public to be in ResourceUbuntu. ResourceUbuntuImpl is
  * not the superclass of DirectoryUbuntuImpl or ActorUbuntuImpl, so it is not
  * possible to put this method only in ResourceUbuntuImpl, like the similar ones
  * in the generic gBase package.
  * But this method is reserved to calls from this Ubuntu gBase classes.
  * @param _b true if the parent authorizes its 'other' AGO category 
  * to have access to this instance.
  */
  public void setAGOAccessFromTheParentOther(boolean _b) {
      if (agoHandler_ == null) return;
      else agoHandler_.setAGOAccessFromTheParentOther(_b);
  } // setAGOAccessFromTheParentOther()

  /**
   * Fires the event 'InheritedOtherRights', from the parent of the resource 'other'
   * category. Called indirectly by the parent DirectoryUbuntuImpl.setL_otherRights() method.
  * Caution: This method is public to be in ResourceUbuntu. ResourceUbuntuImpl is
  * not the superclass of DirectoryUbuntuImpl or ActorUbuntuImpl, so it is not
  * possible to put this method only in ResourceUbuntuImpl, like the similar ones
  * in the generic gBase package.
  * But this method is reserved to calls from this Ubuntu gBase classes.
   * @param _l_old an array of the old access rights for the 'other' AGO category,
   *   in the main ACS. May be empty.
   * @param _l_new an array of the new access rights for the 'other' AGO category,
   *   in the main ACS. May be empty.
   */
  public final void fireL_inheritedOtherRights(StringRight[] _l_old, StringRight[] _l_new) {
      if (agoHandler_ == null) return;
      else agoHandler_.fireL_inheritedOtherRights(_l_old, _l_new);
  } // fireL_inheritedOtherRights()

  /**
  * This method provides a copy of the instance with another name.
  * Do not copy the own resources or the current actors.
  * The copy order is 3 times the source order.
  * _aliasRef may belong to another ACS than this instance ACS.
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ResourceUbuntuImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager)
    throws CreateError {

      if (_acs == null || _name == null  ) {
          throw new CreateError("in ResourceUbuntuImpl.copy(),"+
          " _name or _acs is null.");
      }

      ResourceUbuntuImpl result = new ResourceUbuntuImpl(
          _name, _acs, _parent, userID_, getType(),
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);

      return result;
  } // copy(,,,)
  
  /**
  * This method provides a copy of the instance variables.
  * Overrides the super method and call it
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return the updated _result
  * @exception CreateError of addAclEntry()
  **/
  protected ResourceUbuntuImpl copyOwnVariables(ResourceUbuntuImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in ResourceImplUbuntu.copyOwnVariables(),"+
          " _result is null.");
      }

      _result = (ResourceUbuntuImpl)super.copyOwnVariables(_result, _withoutInternalAcl);
      
      _result.agoHandler_ = agoHandler_.copy(_result);

      return _result;
  } // copyOwnVariables(,,,)

} // ResourceUbuntuImpl

