//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.Ubuntu.gBaseInterface;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.UpDateError;

/**
 * This interface is responsible for modeling a Linux Ubuntu resource.
 * It handles the AGO 'Other' rights and sticky bit in a Linux/Unix operating system.
 * It has an umask from its UserIDUbuntu, to set the default AGO rights.
 * It handles also the AGO inherited rights for 'Other' rights, and the
 * methods setAGOAccessFromTheParentOther() and fireL_inheritedOtherRights()
 * are reserved to calls from the Ubuntu gBase package.
 * This class has several subclasses in this package.
 * All variables are protected.
 * @version 0.7.0
 */
public interface ResourceUbuntu extends Resource, ACSObjectAddon {

  /**
   * Gets the access rights for the Eligible Parties which are not
   * the UserID or the GroupID owners of this resource, nor the root UserID.
   * Does not take in account if the parent tree does not allow any access.
   * @return the AGO 'other' rights, following the Unix/Linux scheme. No null value.
   *  May be empty, or null only if the instance is finalized.
   */
  public StringRight[] getL_otherRights();

  /**
   * Sets the rights for the EligibleParties which are not 'root'
   * nor this resource userID, and which are not members of this resource groupID.
   * The rights may have the two sens to implement Umask, and may be acsrights or metarights.
   * The ACS constraints on 'OtherRights' are applied only if they are not null,
   * otherwise all the arguments are validated.
   * The negative rights are subtracted from the positive rights.
   * No action if the instance has been finalized. If a new right
   * is not allowed by the ACS, it is removed and the other rights are kept.
   * Noop if the resource type contains 'byAcsCreator' or 'immutable'.
   * Fires a PropertyChange for 'OtherRights'.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the other rights are immutable,
   * or if there are more than 20 increasing chains to analyze from one lesser right, or
   * if the new rights contain a 'delete' or 'rename' right,
   * while its parent does not allow it through its deleting-renaming flag.
   */
  public void setL_otherRights(StringRight[] _s) throws UpDateError;

  /**
   * Gets the inherited rights from the parent of the resource group, including the
   * parent sticky bit.
   * @return an array of access rights for the GroupID which contains this resource,
   *   in the main ACS. May be empty, or null only if the instance is finalized.
   */
  public StringRight[] getL_inheritedGroupRights();

  /**
   * Gets the inherited rights from the parent of the resource for 'other' AGO rights.
   * @return an array of access rights for the 'other' AGO category,
   *   in the main ACS. May be empty, or null only if the instance is finalized.
   */
  public StringRight[] getL_inheritedOtherRights();

  /**
  * @return true if the parent authorizes its 'other' AGO category
  * to have access to this instance.
  */
  public boolean getAGOAccessFromTheParentOther();

  /**
  * Called indirectly by DirectoryUbuntuImpl.setL_otherRights().
  * Caution: This method is public to be in ResourceMySQL. ResourceUbuntuImpl is
  * not the superclass of DirectoryUbuntuImpl or ActorUbuntuImpl, so it is not
  * possible to put this method only in ResourceUbuntuImpl, like the similar ones
  * in the generic gBase package. 
  * But this method is reserved to calls from this Ubuntu gBase classes.
  * @param _b true if the parent authorizes its 'other' AGO category 
  * to have access to this instance.
  */
  public void setAGOAccessFromTheParentOther(boolean _b);

  /**
   * Fires the event 'InheritedOtherRights', from the parent of the resource 'other'
   * category. Called indirectly by the parent DirectoryUbuntuImpl.setL_otherRights() method.
  * Caution: This method is public to be in ResourceMySQL. ResourceUbuntuImpl is
  * not the superclass of DirectoryUbuntuImpl or ActorUbuntuImpl, so it is not
  * possible to put this method only in ResourceUbuntuImpl, like the similar ones
  * in the generic gBase package.
  * But this method is reserved to calls from this Ubuntu gBase classes.
   * @param _l_old an array of the old access rights for the 'other' AGO category,
   *   in the main ACS. May be empty.
   * @param _l_new an array of the new access rights for the 'other' AGO category,
   *   in the main ACS. May be empty.
   */
  public void fireL_inheritedOtherRights(StringRight[] _l_old, StringRight[] _l_new);

} // ResourceMySQL
