//Title:        Access Road
//Version:      0.7.2
//Copyright:    2011-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.Ubuntu.gWork;

import java.util.Map;
import java.util.List;
import java.util.Iterator;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Set;

import ARoad0.CNot.*;

import ARoad0.gWork.*;

import ARoad0.gBaseInterface.*;
import ARoad0.gWorkInterface.*;

/**
 * This class provides utilities about rights and links for Linux Ubuntu.
 * It adds to its superclass the handling of Linux AGO Other rights.
 * It removes for Ubuntu any DisplayableLinkImpl containing an AGO Other right
 * while there is a superior right in another DisplayableLink, following the rule
 * 'AGO right: AG rights deletes O rights'. Some unuseful Displayablelinks are
 * removed, when the end has the form Actor/run_under/GroupIDMember/AGO Other/Resource,
 * while another DisplayableLink ends with the form (same Actor)/AGO Other/(same Resource). <P>
 * This class adds also the search of the 2 Linux Ubuntu capacity groups that are simulated
 * in this Access Road version.
 * For the AGO rights, the AccessControlLink.GLOBAL comments use in this Addon
 * the constant ACLINK_C_AGO_OTHER to comment the AGO Other rights. <P>
 * The paths search covers fully the ACL rights, even if in Ubuntu for
 * this version, there is no management of the Linux Access Control Lists.
 * An ACL transmits the Acount/Group context of its source to its target if they
 * are from the same ACS.
 * This is never true for a privilege (not managed in Ubuntu) or a bridge, for which
 * the target has its proper AG context, even if sometimes this context is empty.
 * This class is used by the gWork package.
 * @see ARoad0.CNot.AccessControlLinkImpl
 * @version 0.7.2
 */
public class LinkRightsUbuntuImpl extends LinkRightsImpl {

  /**
  * only one public constructor
  * @param _interpreter algorithm interpreter of this view
  */
  public LinkRightsUbuntuImpl(AlgorithmInterpreter _interpreter) {
      super(_interpreter);
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop.
   * Overrides the super class method.
   * @return true if the method detectAddonPriorityInAllLinksAGORights provides
   * some work in this class.
   */
  public boolean getDetectAddonPriorityInAllLinksAGORights() {
      return true;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the central loops, 
   * and by endsPathsFromGroupIDMemberWithAclPrivilegeMemberOwnContain() on
   * ThreeNodeRightsFactoryImpl.
   * @return true if the detectAddonPriorityInLastLinksAGORights method provides
   * some work in this class.
   */
  public boolean getDetectAddonPriorityInLastLinksAGORights() {
      return true;
  }

  /**
  * Removes any DisplayableLink which has an Ubuntu object, and contains an AGO right
  * while there is a superior right in another DisplayableLink, following
  * the applicable ACS rules of a given AcsAddon ACS.
  * Removes for Ubuntu any DisplayableLink which contains an AGO Other right
  * while there is a superior right in another DisplayableLink, following the rule
  * 'AGO right: AG rights deletes O rights', or simply a more direct AGO Other right.
  * The OWNER or CONTAIN removing relation may be without workable rights, but
  * the CONTAIN relation from a capacity group is not a criterion for a removing.
  * The DisplayableLinkImpl analysis is based on the identity of the nodes
  * in 2 links (with the operator '=='). The second end is not the only node to process.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the final
  * search if getDetectAddonPriorityInAllLinksNonAGORights() returns true.
  * Calls removeSomeOtherRightsInAllLinksAGORights().
  * @see #removeSomeOtherRightsInAllLinksAGORights(java.util.Collection) 
  * @param _l_dLinks DisplayableLinkImpls to analyze, all with the same two ends or not,
  * where this LinkRights is for one of the node in at least one DisplayableLinkImpl.
  * Not null.
  * @return the updated argument
  **/
  public Collection detectAddonPriorityInAllLinksAGORights(
          Collection<DisplayableLinkImpl> _l_dLinks) {

                      // for testing and tracking
                      //System.out.println("");
                      //System.out.println("LinkRightsUbuntuImpl."+
                        //      "detectPriorityInAllLinksAGORights() dLinks in the argument:");
                      //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                        // System.out.println("----- "+ite.next());
                      //}

      _l_dLinks = removeSomeOtherRightsInAllLinksAGORights(_l_dLinks);

                      // for testing and tracking
                      //System.out.println("LinkRightsUbuntuImpl.detectPriorityInAllLinksAGORights() dLinks at the end:");
                      //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}

      return _l_dLinks;
  } // detectAddonPriorityInAllLinksAGORights()

  /**
  * Applies the ACS priority rules to a DisplayableLink
  * that ends with an access target, for a given Addon ACS that owns this target,
  * through the analysis of a list of DisplayableLinkImpls.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the central
  * loop if getDetectAddonPriorityInLastLinksAGORights() returns true,
  * for the central iteration and the final search,
  * and for a collection of DisplayableLinkImpls having more than 2 nodes.
  * Called by the 4 NodeRightsUbuntuImpl.selectAddon...() methods.
  * Calls removeSomeOtherRightsInLastLinkAGORights().
  * @see #removeSomeOtherRightsInLastLinkAGORights(ARoad0.CNot.DisplayableLinkImpl, java.util.Collection)
  * @param _dLink where the second end is the node of this instance.
  * @param _upd_l_dLinks to analyze and update, all with the same two ends of _dLink.
  * Returns false if it is null or empty. It may be updated. May contain
  * the argument _dLink that is then ignored. May be null.
  * @return true if _dLink is not valid
  * @exception InternalError with some ACS-specific errors
  **/
  public boolean detectAddonPriorityInLastLinkAGORights(
    DisplayableLinkImpl _dLink, Collection<DisplayableLinkImpl> _upd_l_dLinks) {

    if (_upd_l_dLinks == null) return false;
    return removeSomeOtherRightsInLastLinkAGORights(_dLink, _upd_l_dLinks);
  } // detectAddonPriorityInLastLinkAGORights()

  /**
  * Removes for Ubuntu any DisplayableLink which contains an AGO Other right
  * while there is a more direct AGO Other right.
  * AGO Other rights are specific to the ACS addon Ubuntu nodes, which may be simply
  * intermediate nodes in the DisplayableLinks.
  * The DisplayableLinkImpl analysis is based on the identity of the nodes
  * in 2 links (with the operator '=='). The second end is not the only node to process.
  * The single type of DisplayableLinkImpl to remove is listed hereafter: <P>
  * in dLink to remove, there is the form
  * (same node1)-any link-(ep)-IS_MEMBER/RUN_UNDER-(groupidmember)-other link-(resource),
  * and in some dLink2, there is a triplet (same node1)-any link-(same ep)-other link-(same resource). <P>
  * A DisplayableLinkImpl may do not contain any Linux node, and then, it is not updated.
  * Called by detectAddonPriorityInAllLinksAGORights().
  * @see #detectAddonPriorityInAllLinksAGORights(java.util.Collection) 
  * @param _l_dLinks DisplayableLinkImpls to analyze, all with the same two ends or not,
  * where this LinkRights is for one of the node in at least one DisplayableLinkImpl.
  * Not null.
  * @return the updated argument
  **/
  public Collection removeSomeOtherRightsInAllLinksAGORights(Collection _l_dLinks) {

      /*
  * Previous version:
  * The OWNER or CONTAIN removing relation may be without workable rights, but
  * the CONTAIN relation from a capacity group is not a criterion for a removing.
  * The 3 types of DisplayableLinkImpl to remove are listed hereafter: <P>
  * (1) in dLink to remove, there is a pair (ep)-...-other link-(resource),
  * and in some dLink2, there is a node couple (ep)-OWNER-(resource) or
  * a node couple (ep)-CONTAIN-(resource); <P>
  * (2) in dLink to remove, there is a pair (ep)-... -other link-(resource)
  * and in some dLink2, there is a node triplet (ep)-RUN_UNDER-(userid)-OWNER-(resource); <P>
  * (3) in dLink to remove, there is a node pair (ep)-... other link-(resource),
  * and in some dLink2, there is a node triplet (ep)-IS_MEMBER-
  * (groupid)-CONTAIN-(resource) or a node triplet (ep)-RUN_UNDER-(groupid)-CONTAIN-(resource). <P>
       */

                          // for testing and tracking
                          //System.out.println(
                            //  "LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight()"+
                              //" the argument is:");
                          //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                            // System.out.println("--- "+ite.next());
                          //}

      // variables
    List                    l_dLinks2 = new ArrayList(_l_dLinks);
    //Map                     m_ownerLinkTypePositions = new HashMap(INITIAL_CAPACITY);
    //Map                     m_containLinkTypePositions = new HashMap(INITIAL_CAPACITY);
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink2 = null;
    BaseObject[]            l_obj = null;
    BaseObject[]            l_obj2 = null;
    boolean                 toRemove = false;
    //List                    l_owner = null;
    //List                    l_contain = null;
    List<Integer>           l_other = null;
    List<Integer>           l_other2 = null;
    //int                     ownIndex = -1;
    //int                     conIndex = -1;
    //int                     othIndex = -1;

    // Previous version: initialize
    //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
      //  dLink = (DisplayableLinkImpl)ite.next();
        //m_ownerLinkTypePositions.put(dLink.hashCode(),
          //      dLink.getOwnerLinkTypePositions());
        //m_containLinkTypePositions.put(dLink.hashCode(),
          //      dLink.getContainLinkTypePositions());
    //} // for ite

                  // for testing and tracking
                  //System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() m_ownerLinkTypePositions:");
                  //for (Iterator ite = m_ownerLinkTypePositions.entrySet().iterator(); ite.hasNext(); ) {
                    //  System.out.println("----- "+ite.next());
                  //}

                  // for testing and tracking
                  //System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() m_containLinkTypePositions:");
                  //for (Iterator ite = m_containLinkTypePositions.entrySet().iterator(); ite.hasNext(); ) {
                    //  System.out.println("----- "+ite.next());
                  //}

    for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
        dLink = (DisplayableLinkImpl)ite.next();
        toRemove = false;

        // processing even if one node ACS does not apply the AGO priorities
        // or if is not from the AcsAddon Ubuntu.
        // dLink.getEorAllNodes()[index of l_other] and
        // dLink.getEorAllNodes()[index of l_other + 1] are from an ACS Ubuntu.
        l_other = getLinuxOtherLinkTypePositions(dLink);
        if (l_other == null) {
            continue;
        }

        for (Iterator ita = l_dLinks2.iterator(); ita.hasNext(); ) {
             dLink2 = (DisplayableLinkImpl)ita.next();
             if (dLink2 == dLink) continue;

             // no looking for the common Resource in dLink and dLink2

             /*
              * Previous version:
             l_owner = (List)m_ownerLinkTypePositions.get( dLink2.hashCode() );

             // to remove dLink, detect an owner relation in dLink2.
             // rule 'AGO right: AG rights overlay O rights' for A rights
             if (l_owner != null) {
                 l_obj = dLink.getEorAllNodes();
                 l_obj2 = dLink2.getEorAllNodes();

                                            // for testing and tracking
                                      //System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() "+
                                        //      "at the start of the ita loop on l_dLinks2, l_obj2 is not null: ");
                                      //for (int k = 0; k < l_obj2.length; k++) {
                                        //  System.out.println("-------- "+l_obj2[k]);
                                      //}

                 // do not analyze the first owner relation.
                 // the criteria is comparing 3-nodes dLinks. The cases are:
                 // in dLink, triplet (ep)-RUN_UNDER-
                 //                 (userid)-other link-(resource)
                 // in dLink2, triplet (ep)-RUN_UNDER-(userid)-OWNER-(resource)

                 // analyze the second owner relation.
                 // the way is comparing a 3-nodes dLink and a 2-nodes dLink. The cases are:
                 // in dLink, pair (ep)-other link-(resource)
                 // in dLink2, triplet (ep)-RUN_UNDER-(userid)-OWNER-(resource)
                 for (Iterator ito = l_owner.iterator(); ito.hasNext(); ) {
                     ownIndex = ((Integer)ito.next()).intValue();

                     for (Iterator itc = l_other.iterator(); itc.hasNext(); ) {
                         othIndex = ((Integer)itc.next()).intValue();
                         // The value at the index i of l_contain = dLink.getContainLinkTypePositions()
                         // is connected to the nodes getEorAllNodes()[i] and getEorAllNodes()[i+1].
                         if (ownIndex > 0 &&
                                 // l_obj[othIndex] and l_obj[othIndex+1] are from an ACS Ubuntu
                                 l_obj[othIndex+1] instanceof ImmutableResource &&
                                 l_obj[othIndex] == l_obj2[ownIndex - 1] &&
                                 l_obj[othIndex+1] == l_obj2[ownIndex+1]) {
                             // dLink2 has a owner relation where dLink has for an other link
                             // for the same ep and resource: remove dLink
                             toRemove = true;

                                  // for testing and tracking
                                  System.out.println("RightsFactoryUtilitiesUbuntuUbuntu.removeSomeOtherRightsInAllLinksAGORight() "+
                                    "first test on an other link at "+othIndex);
                                  System.out.println("--- implies the removing of the dLink "+dLink);
                                  System.out.println("--- since there is a similar owner link at "+ownIndex+" in "+dLink2);

                         }
                         if (toRemove) break;
                     } // for itc
                     if (toRemove) break;
                 } // for ito
             } // if (l_owner != null)

             if (toRemove) {
                 l_obj = null;
                 l_obj2 = null;
                 break; // on the ita loop on l_dLinks2
             }

             l_contain = (List)m_containLinkTypePositions.get( dLink2.hashCode() );
             // detect contain and other relations in 2 dLinks
             // rule 'AGO right: AG rights overlay O rights' for G rights
             if (l_contain != null) {
                 if (l_obj == null)
                     l_obj = dLink.getEorAllNodes();

                                        // for testing and tracking
                                        //System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() "+
                                           //     "before dLink2.getEorAllNodes() in thread "+Thread.currentThread().getName()+
                                             //   "dLink2.hashCode(): "+dLink2.hashCode()+
                                              //" number of nodes: "+dLink2.noOfNodes()+" and dLink2: "+dLink2);

                 if (l_obj2 == null) {
                     
                                        // for testing and tracking
                                        //System.out.println("l_obj2 is null, so call dLink2.getEorAllNodes()");

                    l_obj2 = dLink2.getEorAllNodes();
                 } // if

                 // analyze the first contain relation.
                 // the way is comparing 3-nodes dLinks. The cases are:
                 // in dLink, triplet (ep)-RUN_UNDER or IS_MEMBER-
                 //                 (groupid)-other link-(resource)
                 // in dLink2, triplet (ep)-RUN_UNDER-(groupid)-CONTAIN-(resource)
                 for (Iterator itc = l_contain.iterator(); itc.hasNext(); ) {
                     conIndex = ((Integer)itc.next()).intValue();
                     // The value at the index i of l_contain = dLink.getContainLinkTypePositions()
                     // is connected to the nodes getEorAllNodes()[i] and getEorAllNodes()[i+1].
                     if (conIndex > 0 && othIndex > 0 &&
                             // l_obj[othIndex] and l_obj[othIndex+1] are from an ACS Ubuntu
                             l_obj2[conIndex+1] instanceof ImmutableResource &&
                             l_obj2[conIndex - 1] == l_obj[othIndex - 1] &&
                             l_obj2[conIndex+1] == l_obj[othIndex+1]  &&
                             // no removing if the central groupid is a capacity group
                             !getLinuxCapacityGroups(
                                ((ACSObject)l_obj2[conIndex]).getEorACS()).
                                contains((GroupID)l_obj2[conIndex]) ) {
                         // dLink2 has a contain relation on the same nodes pair
                         // that dLink has for an other link for the same ep and resource:
                         // remove dLink
                         toRemove = true;

                              // for testing and tracking
                              System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORights() "+
                                "analyze the first contain relation on an other link at "+othIndex);
                              System.out.println("--- implies the removing of the dLink "+dLink);
                              System.out.println("--- since there is a similar contain link at "+conIndex+" in "+dLink2);

                     }
                     if (toRemove) break;
                 } // for itc

                 // analyze the second contain relation.
                 // the way is comparing a 3-nodes dLink and a 2-nodes dLink. The cases are:
                 // in dLink, pair (ep)-other link-(resource)
                 // in dLink2, triplet (ep)-RUN_UNDER or IS_MEMBER-
                 //                 (groupid)-CONTAIN-(resource)
                 for (Iterator ito = l_contain.iterator(); ito.hasNext(); ) {
                     conIndex = ((Integer)ito.next()).intValue();
                     for (Iterator itc = l_other.iterator(); itc.hasNext(); ) {
                         othIndex = ((Integer)itc.next()).intValue();


                                      // for testing and tracking
                                      //System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() "+
                                        //"analyzes an other link at "+othIndex+" for the removing of the dLink "+
                                        //"and compared to dLink.hashCode(): "+dLink.hashCode()+" dLink "+
                                        //dLink);

                         // The value at the index i of l_contain = dLink.getContainLinkTypePositions()
                         // is connected to the nodes getEorAllNodes()[i] and getEorAllNodes()[i+1].
                         if (conIndex > 0 &&
                                 // l_obj[othIndex] and l_obj[othIndex+1] is from an ACS Ubuntu
                                 l_obj[othIndex+1] instanceof ImmutableResource &&
                                 l_obj[othIndex] == l_obj2[conIndex - 1] &&
                                 l_obj[othIndex+1] == l_obj2[conIndex+1] &&
                                 // no removing if the central groupid is a capacity group
                                 !getLinuxCapacityGroups(
                                    ((ACSObject)l_obj2[conIndex]).getEorACS()).
                                    contains((GroupID)l_obj2[conIndex]) ) {
                             // dLink2 has a contain relation where dLink has for an other link
                             // for the same ep and resource: remove dLink
                             toRemove = true;

                                      // for testing and tracking
                                      System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() "+
                                        "analyze the second contain relation on an other link at "+othIndex);
                                      System.out.println("--- implies the removing of the dLink "+dLink);
                                      System.out.println("--- since there is a similar contain link at "+conIndex+" in "+dLink2);

                         }
                         if (toRemove) break;
                     } // for itc
                     if (toRemove) break;
                 } // for ito

                 // do not analyze the third contain relation.
                 // the way is comparing some 2-nodes dLink. The cases are:
                 // in dLink, pair (ep)-other link-(resource)
                 // in dLink2, triplet (ep)-CONTAIN-(resource)

             } // if (l_contain
             */

             // only case
             // dLink has the form (node1)-any link-(ep1)-IS_MEMBER/RUN_UNDER-(groupidmember)-other link-(resource),
             // and in some dLink2, there is the triplet (same node1)-any link-(same ep1)-other link-(same resource)
             if (dLink.noOfNodes() > 3 && dLink2.noOfNodes() > 2) {

                                      // for testing and tracking
                                      //System.out.println(
                                        //      "LinkRightsUbuntuImpl."+
                                          //    "removeSomeOtherRightsInAllLinksAGORight() "+
                                            //  "for the fourth IS_MEMBER/RUN_UNDER case ");
                                      //System.out.println("--- "+dLink);
                                      //System.out.println("--- "+dLink2);

                 for (Integer in : l_other) {
                     if (in - 2 < 0)
                         continue;
                     
                     l_obj = dLink.getEorAllNodes();
                     if (!(l_obj[in] instanceof GroupIDMember) ||
                        !(l_obj[in - 1] instanceof EligibleParty))
                         continue;

                     l_other2 = getLinuxOtherLinkTypePositions(dLink2);
                     if (l_other2 == null) {
                            continue;
                     }

                     for (Integer in2 : l_other2) {
                         if (in2 - 1 < 0)
                             continue;

                                      // for testing and tracking
                                      //System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() "+
                                        //"for the fourth IS_MEMBER/RUN_UNDER case, with other link at "+in);
                                      //System.out.println(
                                        //      "--- analyzes the removing of the dLink "+dLink);
                                      //System.out.println(
                                        //      "--- with in2 = "+in2+" in "+dLink2);


                         l_obj2 = dLink2.getEorAllNodes();
                         // l_obj[in + 1] is from an ACS Ubuntu
                         if (l_obj[in + 1] != l_obj2[in2 + 1])
                             continue;
                         if (l_obj[in - 1] != l_obj2[in2])
                             continue;
                         if (l_obj[in - 2] != l_obj2[in2 - 1])
                             continue;

                         toRemove = true;
                         break;
                     } // for in2
                     if (toRemove) break;
                 } // for in
             } // if (dLink.noOfNodes() > 3)

             l_obj = null;
             l_obj2 = null;
             l_other2 = null;
             if (toRemove) break;
        } // for ita

        if (toRemove) {
            ite.remove();
            // l_dLinks2 is also cleared to speed up the work
            l_dLinks2.remove(dLink);
        } // if
    } // for ite

    l_dLinks2 = null;
    dLink = null;
    dLink2 = null;
    l_obj = null;
    l_obj2 = null;
    l_other = null;
    // Previous version:
    //l_owner = null;
    //l_contain = null;
    //m_ownerLinkTypePositions = null;
    //m_containLinkTypePositions = null;

                      // for testing and tracking
                      //System.out.println("LinkRightsUbuntuImpl.removeSomeOtherRightsInAllLinksAGORight() dLinks at the end:");
                      //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}

    return _l_dLinks;
  } // removeSomeOtherRightsInAllLinksAGORights()

  /**
  * Removes any DisplayableLinkImpl which contains an AGO Other right
  * in the last link. There are two cases: <P>
  * - there is a superior right in another DisplayableLink, following the rule
  * 'AGO right: AG rights deletes O rights', <P>
  * - the end has the form Actor/run_under/GroupIDMember/AGO Other/Resource,
  * while another DisplayableLink ends with the form (same Actor)/AGO Other/(same Resource). <P>
  * If _dLink has no 'AGO Other' last link, it is always valid. <P>
  * In the first case, the OWNER or CONTAIN removing relation may be without workable rights, but
  * the CONTAIN relation from a capacity group is not a criterion for a removing.
  * The _dLink is not validated if: <P>
  * - in _dLink, there is a pair (EligibleParty)-...-other link-(_dLink second end),
  * whatever the number of intermediate nodes between the EligibleParty and the Resource, <P>
  * - in a _upd_l_dLinks link, there is the same pair with a path
  * (EligibleParty)-...-OWNER or CONTAIN-(_dLink second end), with the Account/Primary Group
  * context of the _dLink second end for this link second end, whatever the number of
  * intermediate nodes between the EligibleParty and the link second end, and
  * if the CONTAIN relation is not from a capacity group.
  * This first case does not take account of the non-primary groups in the AG context
  * of the second end. <P>
  * In the second case, a DisplayableLink is not false but it is unuseful, as _dLink or not:
  * - _dLink end has the form EligibleParty/run_under or is_member/GroupIDMember/AGO Other/Resource, and
  * it is removed because, <P>
  * - in a _upd_l_dLinks link, there is the ending form (same EligibleParty)/AGO Other/(same Resource). <P>
  * OR, <P>
  * - a _upd_l_dLinks link has the form EligibleParty/run_under or is_member/GroupIDMember/AGO Other/Resource,
  * and it is removed because, <P>
  * - _dlink has the ending form (same EligibleParty)/AGO Other/(same Resource). <P>
  * This last subcase is handled by removeSomeOtherRightsInAllLinksAGORights(),
  * because the shorter link may not be known when this method is called for the longer link.<P>
  * Called by detectAddonPriorityInLastLinkAGORights().
  * @see #detectAddonPriorityInLastLinkAGORights(ARoad0.CNot.DisplayableLinkImpl, java.util.Collection)
  * @param _dLink where the second end is the node of this instance.
  * @param _upd_l_dLinks to analyze and update, all with the same two ends of _dLink.
  * Returns false if it is null or empty. It may be updated. May contain
  * the argument _dLink that is then ignored.
  * @return true if _dLink is not valid
  **/
  public boolean removeSomeOtherRightsInLastLinkAGORights(
          DisplayableLinkImpl _dLink, Collection<DisplayableLinkImpl> _upd_l_dLinks) {

      // the _dLink last link has to be an other link from the ACS addon Ubuntu
      if (_dLink.getEorLastLink().getLinkTypeComments(AccessControlLink.GLOBAL).indexOf(
            NodeRightsUbuntuImpl.ACLINK_C_AGO_OTHER) == -1)
                return false;

                            // for testing and tracking
                          //System.out.println(
                            //    "LinkRightsUbuntuImpl.removeSomeOtherRightsInLastLinkAGORight()"+
                              //  " test the first case on "+_dLink);

      Set<ImmutableGroupIDMember> l_mainAGcontext = linkUtil_.getSecondEndAGcontext(_dLink);
      Set<ImmutableGroupIDMember> l_AGcontext = null;
      UserID                      user = null;
      List<GroupID>               l_capacities = getLinuxCapacityGroups(
                                        ((ACSObject)_dLink.getEorSecondEnd()).getEorACS());
      boolean                     result = false;

      // first case
      // l_mainAGcontext never null
      for (ImmutableGroupIDMember memb : l_mainAGcontext) {
          if (memb instanceof UserID) {
              user = (UserID)memb;
              break;
          }
      } // for memb

      for (DisplayableLinkImpl link : _upd_l_dLinks) {
            
            if (!link.getEorLastLink().getLinkTypeValue(AccessControlLink.OWNER) &&
                    !link.getEorLastLink().getLinkTypeValue(AccessControlLink.CONTAIN))
                        continue;
            if (l_capacities.contains(link.getEorNodeJustBeforeTheSecondEnd()))
                        continue;

            l_AGcontext = linkUtil_.getSecondEndAGcontext(link);
            
            if (l_AGcontext.contains(user) || l_AGcontext.contains(user.getEorPrimGroup())) {
                result = true;
                break;
            }          
            
      } // for link


                      // for testing and tracking
                      //if (!result) {
                        //  System.out.println(
                          //      "LinkRightsUbuntuImpl.removeSomeOtherRightsInLastLinkAGORight()");
                          //System.out.println("----- with removing of _dlink "+_dLink);
                          //System.out.println("----- from the dLink "+dLink);
                          //for (Iterator ite = _upd_l_dLinks.iterator(); ite.hasNext(); ) {
                            //  System.out.println("----- "+ite.next());
                          //}
                      //} // if

                            // for testing and tracking
                            //if (!result  && _dLink.noOfNodes() > 2) {
                              //System.out.println(
                                //    "LinkRightsUbuntuImpl.removeSomeOtherRightsInLastLinkAGORight()"+
                                  //  " test the second case on "+_dLink);
                              //System.out.println(
                                //    "--- _dLink.getEorAllACLinks()[_dLink.noOfNodes() - 2]:"+
                                  //  _dLink.getEorAllACLinks()[_dLink.noOfNodes() - 3]);
                            //}

      // second case
      // if the _dLink end has the form 
      // EligibleParty/run_under or is_member/GroupIDMember/AGO Other/Resource.
      // the test of the last link at the start of the method is already done.
      if (!result && _dLink.noOfNodes() > 2 &&
              _dLink.getEorNodeJustBeforeTheSecondEnd() instanceof GroupIDMember &&
              _dLink.getEorAllNodes()[_dLink.noOfNodes() - 3] instanceof EligibleParty) {

          for (DisplayableLinkImpl link : _upd_l_dLinks) {

                if (link == _dLink || link.getEorSecondEnd() != _dLink.getEorSecondEnd())
                            continue;
                if (link.getEorLastLink().getLinkTypeComments(AccessControlLink.GLOBAL).indexOf(
                        NodeRightsUbuntuImpl.ACLINK_C_AGO_OTHER) == -1)
                            continue;

                if (link.getEorNodeJustBeforeTheSecondEnd() ==
                        _dLink.getEorAllNodes()[_dLink.noOfNodes() - 3]) {
                    result = true;

                            // for testing and tracking
                          //System.out.println(
                            //    "LinkRightsUbuntuImpl.removeSomeOtherRightsInLastLinkAGORight()"+
                              //  " applies the second case to "+_dLink);

                    break;
                }

          } // for link

      } // if (result &&

      l_mainAGcontext = null;
      l_AGcontext = null;
      user = null;
      l_capacities = null;

      return result;
  } // removeSomeOtherRightsInLastLinkAGORights()

  /**
  * Gets the capacity groups in a Linux system (not for an Unix system).
  * The list does not vary upon the ACS, but the argument is used for future evolutions.
  * @param    _acs to process. May not be a Linux system.
  * @return   the groups with the type '<byAcsCreator> capability'.
  * Not null. Empty for a non-Linux aCS.
  */
  public List<GroupID> getLinuxCapacityGroups(ImmutableACS _acs) {

    List<GroupID>           l_result = new ArrayList(INITIAL_CAPACITY);

    if (!_acs.getACSType().startsWith("Linux Ubuntu"))
        return l_result;
    
    Map                     m_ep = _acs.getEorM_OwnEligibleParties();
    EligibleParty           ep = null;

    for (Iterator ite = m_ep.values().iterator(); ite.hasNext(); ) {
      ep = (EligibleParty) ite.next();
      if (ep instanceof GroupID && ((GroupID)ep).getEPType().
              equals("<byAcsCreator> capability"))
          l_result.add((GroupID)ep);
    } // for ite

    m_ep = null;
    ep = null;
    return l_result;
  } // getLinuxCapacityGroups()

  /**
  * Gets the positions in getEorAllACLinks() having the types, if any,
  * of the AccessControlLinks having the type GLOBAL which is commented by
  * NodeRightsUbuntuImpl.ACLINK_C_AGO_OTHER to indicate that there is
  * a Linux AGO Other right. This works only if NodeRightsUbuntuImpl.ACLINK_C_AGO_OTHER
  * is reserved to the AcsAddon Ubuntu.
  * No control on the ACS type of the nodes in the argument _dLink, since only
  * NodeRightsUbuntuImpl and its subclasses use these constants.
  * Called by removeSomeOtherRightsInAllLinksAGORights().
  * @param _dLink may contain nodes not from the AcsAddon Ubuntu
  * @return list of indexes from getEorAllACLinks(), for the AccessControlLinks
  * which have a Linux AGO Other right. Never empty. Null if there is no such a relation
  * in the argument, including when the link does not contain any Linux node.
  */
  public List<Integer> getLinuxOtherLinkTypePositions(DisplayableLinkImpl _dLink) {
    AccessControlLinkImpl[] l_ACLinks = _dLink.getEorAllACLinks();
    List result = null;
    if (l_ACLinks == null || l_ACLinks.length == 0) return null;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (l_ACLinks[i].getLinkTypeValue(AccessControlLink.GLOBAL) &&
                    l_ACLinks[i].getLinkTypeComments(AccessControlLink.GLOBAL).
                        indexOf(NodeRightsUbuntuImpl.ACLINK_C_AGO_OTHER) != - 1 ) {

          if (result == null) result = new ArrayList(INITIAL_CAPACITY);
          result.add(i);
      } // if
    } // for (int i = 0
    l_ACLinks = null;

                    // for testing and tracking
                    //if (result != null) {
                      //  System.out.println("RightsFactoryUtilitiesUbuntuUbuntu."+
                        //        "getLinuxOtherLinkTypePositions() "+
                          //      " non-null result for "+_dLink+" :");
                        //for (Iterator ite = result.iterator(); ite.hasNext(); ) {
                          // System.out.println("--------"+ite.next());
                        //}
                    //} // if

    return result;
  } // getLinuxOtherLinkTypePositions()

  /**
   * Adds to the super method, the transfering capacity when
   * the type is GLOBAL with the comment NodeRightsUbuntuImpl.ACLINK_C_AGO_OTHER.
   * This is for the AGO Other rights.
   * Called by setNextNodeAGrunningContext(). Calls the super method.
   * @param _acLink never null
   * @param _firstACS ACS of the _acLink first node. May be null.
   * @param _secondACS ACS of the _acLink second node May be null.
   * @return true if the Account/Groups context of the starting node is transfered
   * to the ending node of the AccessControlLinkImpl
   */
  public boolean isTransferingAGcontext(AccessControlLinkImpl _acLink,
          ImmutableACS _firstACS, ImmutableACS _secondACS) {
      return (super.isTransferingAGcontext(_acLink, null, null) ||
            _acLink.getLinkTypeComments(AccessControlLink.GLOBAL).indexOf(
                NodeRightsUbuntuImpl.ACLINK_C_AGO_OTHER) != -1);
  } // isTransferingAGcontext()

} // LinkRightsUbuntuImpl

