//Title:        Access Road
//Version:      0.7.2
//Copyright:    2010-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.AcsAddon.Accbee.Ubuntu.gWork;

import ARoad0.gWorkInterface.AlgorithmInterpreter;
import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Collection;

import ARoad0.ErrB.BaseError;
import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.CNot.AccessControlLink;
import ARoad0.CNot.AccessControlLinkImpl;

import ARoad0.AcsAddon.Accbee.Ubuntu.gBase.*;
import ARoad0.AcsAddon.Accbee.Ubuntu.gBaseInterface.*;
import ARoad0.gBaseInterface.*;
import ARoad0.gWork.*;

import ARoad0.gBase.ACSImpl;
import ARoad0.gBase.UtilityImpl;
import ARoad0.gWorkInterface.LinkRights;

/**
 * This class is responsible for getting the rights for an Ubuntu node in a view,
 * for the AcsAddon Ubuntu. It adds to its superclass the handling of Linux AGO Other rights.
 * It adds also the search of the 2 Linux Ubuntu capacity groups that are simulated
 * in this Access Road version. To provide these tasks, the methods detectAddon2Nodes()
 * and detectAddon3Nodes() of the superclass are overridden. <P>
 * For the AGO rights, the AccessControlLink.GLOBAL comments use in this Addon
 * the constant ACLINK_C_AGO_OTHER to comment the AGO Other rights. <P>
 * The paths search covers fully the ACL rights, even if in Ubuntu for
 * this version, there is no management of the Linux Access Control Lists.
 * An ACL transmits the direct Account/Group context of its source to its target if they
 * are from the same ACS. It does not transmit the indirect like an alias relation.
 * The Account/Group context is never propagated for a privilege (not managed in Ubuntu)
 * or a bridge, for which the target has its proper AG context, even if sometimes
 * this context is empty.
 * <P>
 * For the Linux/Unix ACS only, there are AGO Other rights on resources and actors, but
 * no privileges nor virtual folders to handle. Some 3-nodes paths are specific
 * to the Linux/Unix ACS, and they are listed hereinafter: <P>
 * - ep/acl/actor/other-ridden resource in Linux-Unix, <P>
 * - actor started from xid/run under/account or group or secondary group/other-ridden resource in Linux-Unix, <P>
 * - actor started from xid/other-executed actor/acl/resource, <P>
 * - actor started from xid/other-executed actor/bridge/actor, <P>
 * - groupidmember/group/other-ridden resource in Linux-Unix, <P>
 * - NO PRI, so no: ep/pri/actor/other-ridden resource in Linux-Unix, <P>
 * - actor/bridge/actor/other-ridden resource in Linux-Unix, <P>
 * - NO VF, so no: ep virtual folder/virtual member/actor/other-ridden resource in Linux-Unix, <P>
 * - groupidmember/other-executed actor in Linux-Unix/acl/resource, <P>
 * - groupidmember/other-executed actor in Linux-Unix/bridge/actor, <P>
 * - NO PRI, so no: groupidmember/other-executed actor in Linux-Unix/pri/resource or virtual folder, <P>
 * <P>
 * Here are the 2 Linux Ubuntu capacity groups which are simulated:
 * - '&lt;CAP_DAC_OVERRIDE&gt;': overrides all read/write/execute AGO rights,
 * including ACL execute access if [_POSIX_ACL] is defined. Excluding DAC access
 * covered by CAP_LINUX_IMMUTABLE. The executing permission is set only when at least
 * one of the file's three AGO execute permission bits is set. <P>
 * - '&lt;CAP_DAC_READ_SEARCH&gt;': overrides all read/write/execute AGO rights
 * regarding read and search on files and directories, including ACL restrictions
 * if [_POSIX_ACL] is defined. Excluding DAC access covered by CAP_LINUX_IMMUTABLE. <P>
 * The following comments are about all the NodeRights in the AcsAddon framework.
 * The specialized methods of an AcsAddon are in a NodeRightsImpl subclass,
 * and they are called by an overriden NodeRights AcsAddon method;
 * there is no name pattern for these methods;
 * they use the generic properties of a node and its environment, but
 * they also handle the original properties of the AcsAddon node.
 * Since an AcsAddon may be based on another AcsAddon, a NodeRightsImpl
 * subclass may have another NodeRightsImpl subclass has super class. <P>
 * The direct calls to the Ubuntu ACS in this class are calls to
 * getEorL_FromOtherInCurrentContextActors() and getEorM_EligibleParties().
 * Upper-level methods begin by get...(), with...() and detect...(),
 * middle-level methods begin by select...().
 * @see ARoad0.gWork.NodeRightsImpl
 * @version 0.7.2
 */
public class NodeRightsUbuntuImpl extends NodeRightsImpl {

  /**
   * For the AGO types, this is a specialized comment for Ubuntu to say that a resource
   * is targeted through Linux AGO Other rights. It is an addition to the
   * ARoad0.CNot.AccessControlLink constants. It is used only by the AcsAddon Ubuntu
   * classes, like LinkRightsUbuntuImpl for example.
   */
  public static final String ACLINK_C_AGO_OTHER = "AGO Other rights";

  protected AgoRightsFactoryImpl    ago_ = null;
  protected UtilityImpl             util_ = null;
  protected DisplayableLinkUtilities      linkUtil_ = null;

  /**
  * Only one constructor, called by RightsMediatorImpl. Calls the superclass constructor.
  * @param _interpreter algorithm interpreter of this view
  * @param _utility LinkRightsImpl or one of its subclass, to use for
  * processing this node
  */
  public NodeRightsUbuntuImpl(AlgorithmInterpreter _interpreter, LinkRightsImpl _utility) {
      super(_interpreter, _utility);
      ago_ = (AgoRightsFactoryImpl)interpreter_.getCoreAlgorithm("Ago");
      util_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
  }

  /**
   * Called by RightsFactory_Facade.
   * Overrides the superclass.
   * @return true if the detectAddon2Nodes method provides some work in this class.
   */
  public boolean getDetectAddon2Nodes() {
      return true;
  }

   /**
   * Called by RightsFactory_Facade.
   * Overrides the superclass.
   * @return true if the detectAddon3Nodes method provides some work in this class.
   */
  public boolean getDetectAddon3Nodes() {
      return true;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by detectOneHiddenNodeCompoundRights().
   * Overrides the superclass.
   * @return true if the selectAddonLastViewNodeAfterActor method provides some work
   * in this class.
   */
  public boolean getSelectAddonLastViewNodeAfterActor() {
      return true;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by detectOneHiddenNodeCompoundRights().
   * Overrides the superclass.
   * @return true if the selectAddonLastViewNodeAfterGroupIDMember method provides some work
   * in this class.
   */
  public boolean getSelectAddonLastViewNodeAfterGroupIDMember() {
      return true;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by detectOneHiddenNodeCompoundRights().
   * Overridden by the subclasses.
   * @return true if the selectAddonLastViewNodeAfterVirtualFolder method provides some work
   * in this class.
   */
  //static public boolean getSelectAddonLastViewNodeAfterVirtualFolder() {
    //  return false;
  //}

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes. Overrides the supersubclass.
   * @return true if the selectAddonNewHiddenNodeForActor method provides some work
   *  in this class.
   */
  public boolean getSelectAddonNewHiddenNodeForActor() {
      return true;
  }

   /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes. Overrides the supersubclass.
   * @return true if the selectAddonNewHiddenNodeForGroupIDMember method provides some work
   *  in this class.
   */
  public boolean getSelectAddonNewHiddenNodeForGroupIDMember() {
      return true;
  }
  
   /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes. Overridden by the subclasses.
   * @return true if the selectAddonNewHiddenNodeForVirtualFolder method provides some work
   *  in this class.
   */
  //static public boolean getSelectAddonNewHiddenNodeForVirtualFolder() {
    //  return false;
  //}

  /**
  * Detects the capacity groups '<CAP_DAC_OVERRIDE>' and '<CAP_DAC_READ_SEARCH>'
  * in Linux Ubuntu for the DisplayableLinks of 2 nodes.
  * Analyzes the AGO Other rights of the view Actors, UserIDs and GroupIDs on _node.
  * For a sketch view, this method returns the DisplayableLink from _center to _node,
  * if any, but also, all the DisplayableLinks from _node to _center.
  * Called by RightsFactory_Facade. Calls select2NodesCapacityRights() and
  * addOtherRightsToAccessControlLink().
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @see #select2NodesCapacityRights
  * @param _viewInBase EPRViewInBase
  * @param _node is the node to analyze, as node of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  * with 'other' links associated to every pair (EP, Resource).
  * */
  public Map detectAddon2Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks, BaseObject _center) {

      ImmutableSource       so = null;
      GroupIDMember         gm = null;
      Actor                 act = null;
      List                  l_context = new ArrayList(INITIAL_CAPACITY);
      Set                   l_hiddenGroups = null;
      AccessControlLinkImpl acLink = null;
      LinkRights            linkRights = null;

      _m_l_DisplayableLinks = select2NodesCapacityRights(
              _viewInBase, _node, _m_l_DisplayableLinks, _center);

      if (!(_node instanceof ResourceUbuntu)) {
          l_context = null;
          return _m_l_DisplayableLinks;
      }

      l_hiddenGroups = selectDirectOwnerContainForTarget((ImmutableResource) _node);

      // iteration ITI on view Actors, UserIDs and GroupIDs
      for (Iterator iti = _viewInBase.getEorM_Sources().values().iterator() ;
        iti.hasNext() ; ) {
            so = (ImmutableSource) iti.next();
            if (so == null || so.getEorACS() != _node.getEorACS()) continue;

            // for the sketch view
            if (_center != null && _node != _center && so != _center) continue;
            
            if (_node == so) continue;

            if (so instanceof GroupIDMember) {
                gm = (GroupIDMember)so;

                l_context.clear();
                l_context.add( gm );
                l_context.addAll(Arrays.asList(gm.getEorL_GroupID()));

                                          // for testing and tracking
                                          //System.out.println("In NodeRightsUbuntuImpl."+
                                            //"detectAddon2Nodes(), gm: "+gm.getName()
                                            //+" and the node "+_node.getName()+
                                            //" with hidden User and Groups:");
                                          //if (l_hiddenGroups != null) {
                                            //  for (Object obj : l_hiddenGroups) {
                                              //    System.out.println("---- "+((GroupIDMember)obj).getName());
                                              //}
                                          //}
                                          //else
                                            //  System.out.println("----- null");
                                          //System.out.println("-- and gm context:");
                                          //for (Object obj : l_context) {
                                            //  System.out.println("---- "+
                                              //        ((ACSObject)obj).getName());
                                          //} // for obj


                // add the other rights to 'gm'
                // if the AG context of _node has no common entities
                // with the AG context of 'gm'
                if (l_hiddenGroups != null && Collections.disjoint(l_context, l_hiddenGroups)) {

                      acLink = addOtherRightsToAccessControlLink(gm, (ResourceUbuntu) _node);

                                          // for testing and tracking
                                          //System.out.println("In NodeRightsUbuntuImpl."+
                                            //"detectAddon2Nodes(), new acLink: "+acLink);

                      if (acLink != null) {
                          linkRights = interpreter_.getLinkRightsInstances().get(
                              _node.getName());
                          // if 'gm' is a group, this is the first link so
                          // the 'weak' property is visible, if any
                          linkRights.addDisplayableLink(
                                  new DisplayableLinkImpl(gm, _node, acLink),
                                  _m_l_DisplayableLinks);
                      } // if
                } // if (!Collections.disjoint(l_context,

            } // if (so instanceof GroupIDMember
            else if (so instanceof Actor) {
                act = (Actor)so;

                l_context.clear();
                if (act.getEorCurrentUserID() != null) {
                      l_context.add(act.getEorCurrentUserID());
                      l_context.addAll(Arrays.asList(act.getEorL_SecondaryGroup()));
                }
                if (act.getEorCurrentGroupID() != null) {
                      l_context.add(act.getEorCurrentGroupID());
                }

                if (l_context.isEmpty()) continue;

                                          // for testing and tracking
                                          //System.out.println("In NodeRightsUbuntuImpl."+
                                            //"detectAddon2Nodes(), "+act.getName()+" and "+
                                            //"node "+_node.getName()+" with l_hiddenGroups:");
                                          //if (l_hiddenGroups != null) {
                                            //  for (Object obj : l_hiddenGroups) {
                                              //    System.out.println("---- "+((GroupIDMember)obj).getName());
                                              //}
                                          //} // if
                                          //else
                                            //  System.out.println("----- null");

                if (l_hiddenGroups == null ||
                        !Collections.disjoint(l_context, l_hiddenGroups)) {
                    continue;
                } // if (!Collections.disjoint(l_context,

                acLink = null;

                // add the other rights to 'act'
                // if the AG context of _node has no common entities
                // with the AG context of 'act'
                if (act.getEorCurrentUserID() != null) {
                    acLink = addOtherRightsToAccessControlLink(
                                    act.getEorCurrentUserID(),
                                    (ResourceUbuntu) _node);
                } // if
                else {
                    acLink = addOtherRightsToAccessControlLink(
                                    (ImmutableGroupIDMember)l_context.iterator().next(),
                                    (ResourceUbuntu) _node);
                } // else

                if (acLink != null) {

                                          // for testing and tracking
                                          //System.out.println("In NodeRightsUbuntuImpl."+
                                            //"detectAddon2Nodes(), for "+act.getName()+" and "+
                                            //"node "+_node.getName()+", acLink: "+acLink);

                    linkRights = interpreter_.getLinkRightsInstances().get(
                                        _node.getName());
                    linkRights.addDisplayableLink(
                                  new DisplayableLinkImpl(act, _node, acLink),
                                  _m_l_DisplayableLinks);
                } // if

            } // else if (so instanceof Actor)

      } // for iti on Sources

      so = null;
      gm = null;
      l_context = null;
      l_hiddenGroups = null;
      acLink = null;
      act = null;
      linkRights = null;

      return _m_l_DisplayableLinks;
  } // detectAcsAddon2Nodes()

  /**
  * For an AcsAddon, filters the incompatible rights, if any, for a node,
  * and adds the AcsAddon specific rights, if any.
  * The specific rights are searching for any source in the view.
  * Analyzes the capacity groups '<CAP_DAC_OVERRIDE>' and '<CAP_DAC_READ_SEARCH>'
  * in Linux Ubuntu for the DisplayableLinks of 3 nodes.
  * Called by RightsFactory_Facade.
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @param _viewInBase EPRViewInBase
  * @param _node is the node to analyze, as node of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @return Map of DisplayableLinks
  * @exception InterruptedException
   */
  public Map detectAddon3Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks)  throws InterruptedException {

      _m_l_DisplayableLinks = select3NodesCapacityRights(
              _viewInBase, _node, _m_l_DisplayableLinks);

      return _m_l_DisplayableLinks;
  } // detectAcsAddon3Nodes()

  /**
  * Returns true if the directory tree or the virtual folder tree allows to access
  * to the leaf. If the leaf is a resource, it is true if the parent UserID/GroupID are
  * the leaf UserID/GroupID, and if the parent rights allow the access to children.
  * If the AG rights inheritance is applicable for the ImmutableSource ACS,
  * this method applies the following generic rules: <P>
  * 'AGO right inheritance 1: if no access from the  inherited rights,
  * no AGO access to the child if and only if the account or the group is not null' <P>
  * 'AGO right: A rights overlay G rights' <P>
  * and the Ubuntu rule: <P>
  * 'AGO right: AG rights overlay O rights' <P>
  * Called only by super.withAccessThroughNodesTreeFromEP(), to follow the
  * AcsAddon pattern. Overrides the super method, to add the AGO Other rights analysis
  * when the super method returns false. Does not call the super method.
  * @param _res ResourceUbuntu which may be accessed or not through its parent tree
  * @param _acc the _res userID. May be null. If null and if all the _res parents have no UserID,
  * returns true.
  * @param _grp the _res groupID. May be null. If null and if all the _res parents have no GroupID,
  * returns true.
  * @return true if the resource parent tree allows to go through down to the resource.
  * @exception InternalError if _res is not a ResourceUbuntu
  * */
  protected boolean withAgoAccessThroughNodesTree(
    ImmutableResource _res, ImmutableUserID _acc, ImmutableGroupID _grp) {

      if (!(_res instanceof ResourceUbuntu))
          throw new InternalError("NodeRightsUbuntuImpl.withAccessThroughNodesTree() "+
                  "called for a resource that is not a ResourceUbuntu.");

      if (_res == null || _res.getEorACS() == null) return false;
      if (_res.getEorParent() == null) return true;

      boolean result = false;

      ImmutableLeaf res = _res;

      while (res.getEorParent() != null) {
          if (res instanceof ResourceUbuntu &&
                  _acc == ((Directory)res.getEorParent()).getEorUserID()) {

                if (((Resource)res).getAGOAccessFromTheParentAccount()) result = true;
                else result = false;

          } else if (res instanceof ResourceUbuntu &&
                  _grp == ((Directory)res.getEorParent()).getEorGroupID()) {

                if (((Resource)res).getAGOAccessFromTheParentGroup()) result = true;
                else result = false;
          } else if (res instanceof ResourceUbuntu &&
                  ((ResourceUbuntu)res).getAGOAccessFromTheParentOther() ) {

                result = true;

                       //for testing and tracking
                      //System.out.println("in NodeRightsUbuntuImpl.withAccessThroughNodesTree(), "+
                        //" for res "+res.getName()+" and 'other', result: "+result);
          } else {
                result = false;

                       //for testing and tracking
                      //System.out.println("in NodeRightsUbuntuImpl.withAccessThroughNodesTree(), "+
                        //" for 'other', result: "+result+" and res.getAGOAccessFromTheParentOther(): "+
                        //res.getAGOAccessFromTheParentOther());
          } // else

          if (!result) break;
          else res = (ImmutableLeaf)res.getEorParent();
      } // while

                       //for testing and tracking
                      //if (_acc != null && _grp != null)
                          //System.out.println("in NodeRightsUbuntuImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _acc "+_acc.getName()+" _grp "+_grp.getName());
                      //else if (_acc != null)
                          //System.out.println("in NodeRightsUbuntuImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _acc "+_acc.getName()+" _grp null");
                      //else if (_grp != null)
                          //System.out.println("in NodeRightsUbuntuImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _acc null and _grp "+_grp.getName());
                      //else
                          //System.out.println("in NodeRightsUbuntuImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _acc null and _grp null");

      return result;
  } // withAgoAccessThroughNodesTree()

  /**
  * Returns the inherited rights given by the directory tree, as AGO rights.
  * Returns null if _res is null, if the _res parent is null, if _acc and _grp are null.
  * Called by addOtherRightsToAccessControlLink() and
  * AgoRightsFactoryImpl.detectAGRights(), and by the unused selectOtherRights().
  * Overrides the super method, calls it first, and adds a complementary process
  * to set the inherited AGO Other rights in the GLOBAL type.
  * The comment AccessControlLink.ACLINK_C_AGO_OTHER is added on the GLOBAL type
  * if necessary.
  * @param _res Resource of this instance;
  * may be accessed or not through its parent tree. May be null.
  * @param _ep the _res userID. May be null.
  * @param _grp the _res groupID. May be null.
  * @return link with the rights given by the parent in the GLOBAL type.
   * May be null but not empty.
  * */
  public AccessControlLinkImpl getL_accessRightsThroughNodesTree(
          ImmutableLeaf _res, ImmutableEligibleParty _ep, ImmutableGroupID _grp) {

                                   //for testing and tracking
                                  //if (_res != null && _ep != null)
                                    //  System.out.println(
                                    //"start of NodeRightsUbuntuImpl.getL_accessRightsThroughNodesTree(), "+
                                    //" for _res "+_res.getName()+
                                    //" and for _acc "+_ep.getName());

      if (_res == null || !(_res instanceof Resource) ||
            (_res.getEorACS().getACSType().indexOf(ACSImpl.LINUX) == -1 ))
          throw new InternalError("in NodeRightsUbuntuImpl.getL_accessRightsThroughNodesTree(), "+
                  "the leaf argument is not a resource or from a NX ACS.");
      
      AccessControlLinkImpl result = super.getL_accessRightsThroughNodesTree(
                                            _res, _ep, _grp);

      if (result == null ) {
          if (((ResourceUbuntu)_res).getAGOAccessFromTheParentOther()) {

                    result = new AccessControlLinkImpl();
                    // to activate even if getL_inheritedOtherRights() is empty
                    result.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                    result.setRights(AccessControlLink.GLOBAL,
                            ((ResourceUbuntu)_res).getL_inheritedOtherRights());
                    result.addComment(AccessControlLink.GLOBAL, ACLINK_C_AGO_OTHER);

                                   //for testing and tracking
                                  //System.out.println(
                                    //"in NodeRightsUbuntuImpl.getL_accessRightsThroughNodesTree(), "+
                                    //" for 'other' ");
          } // if
      } // if (l_result == null)


                                   //for testing and tracking
                                  //System.out.println(
                                    //"in NodeRightsUbuntuImpl.getL_accessRightsThroughNodesTree(), "+
                                    //"result: "+result);

      if (result == null || result.isEmpty())
          return null;
      else return result;
  } // getL_accessRightsThroughNodesTree()

  /**
   * Detects the view target that is linked to an Ubuntu Actor, to extend
   * the current DisplayableLinks ended by this actor.
   * Adds the AGO Other rights if the actor has no AGO context.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by detectOneHiddenNodeCompoundRights().
   * Calls LinkRights.updateAGrunningContext() and addOtherRightsToExtendDisplayableLink().
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _target ResourceUbuntu to analyze. Never null.
   * @param _dLink with an actor as second end and node of this instance. Never null.
   * @param _m_l_DisplayableLinks immutable Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @return new DisplayableLinkImpl that is an extension of _dlink
   * with _target as the second node, or null. Never empty.
  * */
  public DisplayableLinkImpl selectAddonLastViewNodeAfterActor(
          EPRViewInBase _viewInBase, ImmutableTarget _target,
          DisplayableLinkImpl _dLink, Map _m_l_DisplayableLinks) {

                        // for testing and tracking
                        //System.out.println(
                          //      "NodeRightsUbuntuImpl.selectAddonLastViewNodeAfterActor() starts");

        if (_target == null || !(_target instanceof ResourceUbuntu) ||
              !_viewInBase.containsTarget(_target) || _dLink == null ||
              !(_dLink.getEorSecondEnd() instanceof ImmutableActor))
                    return null;

        if (!((ACSObject)_dLink.getEorSecondEnd()).getAcsName().equals(_target.getAcsName()))
          return null;

        Set<ImmutableGroupIDMember> l_actorContext = 
                                            linkUtil_.getSecondEndAGcontext(_dLink);
        Set                         l_hiddenGroups =
                                            selectDirectOwnerContainForTarget(
                                                (ImmutableResource) _target);

        if (l_actorContext == null) {
            
                        // for testing and tracking
                        //if (((ACSObject)_dLink.getEorSecondEnd()).getNickName().equals("e_jerry"))
                          //  System.out.println(
                            //    "in NodeRightsUbuntuImpl.selectAddonLastViewNodeAfterActor()"+
                              //  " for 'e_jerry', l_actorContext is null");

            return null;
        }

        if (l_actorContext.contains(((ImmutableActor)_dLink.getEorSecondEnd()).
              getEorCurrentUserID()))
                l_actorContext.addAll(Arrays.asList(
                  ((ImmutableActor)_dLink.getEorSecondEnd()).getEorL_SecondaryGroup()));

        // add the other rights to the _dLink second end
        // if the AG context of _target has no common entities
        // with the AG context of the _dLink second end actor
        if (l_hiddenGroups == null ||
                !Collections.disjoint(l_actorContext, l_hiddenGroups)) {

                        // for testing and tracking
                        //if (((ACSObject)_dLink.getEorSecondEnd()).getNickName().equals("e_jerry")) {
                          //  System.out.println(
                            //    "Call to NodeRightsUbuntuImpl.selectAddonLastViewNodeAfterActor()"+
                              //  "for 'e_jerry':");
                            //System.out.println("--- the AG context of "+_target.getNickName()+
                              //      " has common entities with 'e_jerry'. AG context: ");
                            //for (ImmutableGroupIDMember mem : l_actorContext) {
                              //  System.out.println("------ "+mem.getNickName());
                            //} // for mem
                        //} // if

          l_actorContext = null;
          return null;
        } // if

        DisplayableLinkImpl result = addOtherRightsToExtendDisplayableLink(
              (ResourceUbuntu)_target, _dLink, l_actorContext);

        Name pairKey = ((Name)((BaseObject)_dLink.getEorFirstEnd()).
                              getName()).extended(_target.getName());

        if (result != null &&
                link_.detectAddonPriorityInLastLinkAGORights(
                    result,
                    (Collection<DisplayableLinkImpl>)_m_l_DisplayableLinks.get(pairKey))) {

            result = null;

                        // for testing and tracking
                        //if (((ACSObject)_dLink.getEorSecondEnd()).getNickName().equals("e_jerry")) {
                          //  System.out.println(
                            //    "Call to NodeRightsUbuntuImpl.selectAddonLastViewNodeAfterActor()"+
                              //  "for 'e_jerry':");
                            //System.out.println("--- the new link to "+_target.getNickName()+
                              //      " is deleted by detectAddonPriorityInLastLinkAGORights()."+
                                //    " AG context: ");
                            //for (ImmutableGroupIDMember mem : l_actorContext) {
                              //  System.out.println("------ "+mem.getNickName());
                            //} // for mem
                        //} // if
        } // if
            
                        // for testing and tracking
                        //if (((ACSObject)_dLink.getEorSecondEnd()).getNickName().equals("e_jerry")) {
                          //  System.out.println(
                            //    "NodeRightsUbuntuImpl.selectAddonLastViewNodeAfterActor() ends"+
                              //  " for 'e_jerry', with a true result: ");
                            //System.out.println(
                              //  "--- "+result);
                        //}

        pairKey = null;
        l_actorContext = null;
        l_hiddenGroups = null;
        return result;
  } // selectAddonLastViewNodeAfterActor()

  /**
   * Detects the view target that is linked to an Ubuntu GroupIDMember, to extend
   * the current DisplayableLink ended by this GroupIDMember.
   * Adds the Linux AGO Other rights and, if the groupIDMember is a capacity group,
   * adds the specific capacity rights.
   * Does not add the AGO Other rights to the capacity groups.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by
   * ThreeNodesRighsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeAlias().
   * Calls selectDirectOwnerContainForTarget(),
   * addOtherRightsToExtendDisplayableLink(),
   * addCapacityRightsToDisplayableLink(),
   * LinkRightsUbuntuImpl.getLinuxCapacityGroups().
   * Overrides the superclass.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _target view node to analyze. Never null.
   * @param _dLink with a groupIDMember as second end and node of this instance. Never null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @return new DisplayableLinkImpl that is an extension of _dlink
   * with _target as the second node, or null. Never empty.
  * */
  public DisplayableLinkImpl selectAddonLastViewNodeAfterGroupIDMember(
          EPRViewInBase _viewInBase, ImmutableTarget _target,
          DisplayableLinkImpl _dLink, Map _m_l_DisplayableLinks) {

                        // for testing and tracking
                        //System.out.println(
                          // "Call to NodeRightsUbuntuImpl.selectAddonLastViewNodeAfterGroupIDMember()");

      if (_target == null || !(_target instanceof ResourceUbuntu) ||
              !_viewInBase.containsTarget(_target) || _dLink == null ||
              !(_dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember))
                    return null;

      if (!((ACSObject)_dLink.getEorSecondEnd()).getAcsName().equals(_target.getAcsName()))
          return null;

      DisplayableLinkImpl       result = null;
      ImmutableGroupIDMember    gm = (ImmutableGroupIDMember)_dLink.getEorSecondEnd();

          // process the capacity groups
      if (((LinkRightsUbuntuImpl)link_).getLinuxCapacityGroups(gm.getEorACS()).contains(gm)) {

          // add gm/_target with gm as Linux capacity group
          result = DisplayableLinkImpl.valueOf(_dLink, _target, new AccessControlLinkImpl());

          result = addCapacityRightsToDisplayableLink(result);

                        // for testing and tracking
                        //System.out.println("NodeRightsUbuntuImpl."+
                          //      "selectAddonLastViewNodeAfterGroupIDMember() for "+_dLink+
                            //    "and a group that is a capacity group: "+gm.getName());

          gm = null;
          return result;
      } // if link_.getLinuxCapacityGroups(

                        // for testing and tracking
                        //System.out.println("NodeRightsUbuntuImpl."+
                          //      "selectAddonLastViewNodeAfterGroupIDMember() for "+_dLink+
                            //    "and a group that is NOT a capacity group: "+gm.getName());


      Set                       l_context = new HashSet(INITIAL_CAPACITY);
      Set                       l_hiddenGroups = selectDirectOwnerContainForTarget(
                                        (ImmutableResource) _target);

      l_context.add( gm );
      l_context.addAll(Arrays.asList(gm.getEorL_GroupID()));

      // add the other rights to the _dLink second end
      // if the AG context of _target has no common entities
      // with the AG context of the _dLink second end
      if (l_hiddenGroups != null && Collections.disjoint(l_context, l_hiddenGroups)) {

          result = addOtherRightsToExtendDisplayableLink(
                  (ResourceUbuntu)_target, _dLink, l_context);
      } // if (!Collections.disjoint(l_context,

      Name pairKey = ((Name)((BaseObject)_dLink.getEorFirstEnd()).
                              getName()).extended(_target.getName());

      if (result != null &&
                link_.detectAddonPriorityInLastLinkAGORights(
                    result, (Collection<DisplayableLinkImpl>)_m_l_DisplayableLinks.get(pairKey)))
                                result = null;

      pairKey = null;
      l_context = null;
      gm = null;
      l_hiddenGroups = null;
      return result;
  } // selectAddonLastViewNodeAfterGroupIDMember()

  /**
   * Detects the hidden actors that are linked to an Ubuntu Actor as access source,
   * to extend the current DisplayableLinks ended by this actor.
   * Two lists are returned to separate the executing right-proxy nodes.
   * The last argument may be updated to provide the second returned value,
   * and it contains only the nodes on which the property changes have to be listened.
   * Adds two forms of links: <P>
   * - actor _node with Linux setuid or setgid, that enforces in its Account/Groups
   * context of execution, its current account and/or current group,
   * and new pathes to them are added to each link in _l_dLinks <P>
   * - actor _node started from xid/AGO Other rights/actor 'B', IF there is
   * some B/acl/actor, B/bridge/actor or B/privilege/actor links;
   * these AGO Other rights imply that _node has no AG running context that matches
   * to the 'B' Account and Group (otherwise, it is not other-executed). <P>
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes and in the starting search, and
   * by ThreeNodesRightsFactoryImpl.addPathsFromActorAcsAddonRelationActor(),
   * addPathsFromActorAcsAddonRelationNoActor().
   * Overrides and calls first the super method.
   * Calls addOtherRightsToExtendDisplayableLink(), LinkRights.updateAGrunningContext()
   * and ACSUbuntuImpl.getEorL_FromOtherInCurrentContextActors().
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _node actor to analyze. Never null.
   * @param _l_dLinks list of links with _node as second end and node of this instance.
   * May be null.
   * @param _m_l_DisplayableLinks immutable Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @param _upd_l_NoProxyOrNoExecuteNodes updated by the adding of the hidden nodes
   * without executing right or which are not right-proxy nodes, if any.
   * A RUN_UNDER relation on a GroupIDMember is considered there as an executing
   * right on a right-proxy node.
   * This list is only extended if necessary, as a complement of the returned value of the method.
   * This argument is usually empty at the call of this method, but this is not mandatory.
   * Never null.
   * @return new DisplayableLinkImpls that are extensions or updatings
   * of some DisplayableLinkImpls in _l_dLinks with a GroupIDMember or an Actor
   * as the second end, or null.
   * If _l_dLinks is null, returns only simple links. May be null but never empty.
   * @exception InternalError if _l_dLinks contains a link where _node is not the second end
  * */
  public List<DisplayableLinkImpl> selectAddonNewHiddenNodeForActor(
          EPRViewInBase _viewInBase, ImmutableActor _node,
          Set<DisplayableLinkImpl> _l_dLinks, Map _m_l_DisplayableLinks,
          List<ACSObject> _upd_l_NoProxyOrNoExecuteNodes) {

        // TODO: AG context in the initial path to update for all its consumers
        // change the order of calls?

                                    // for testing and tracking
                                    //System.out.println(
                                      // "call to NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor() for "+
                                       //"_l_dLinks:");
                                    //if (_l_dLinks != null) {
                                      //  for (Object ob : _l_dLinks) {
                                        //    System.out.println("---- "+ob);
                                        //}
                                    //} // if

      if (_node == null || !(_node instanceof ActorUbuntu) ||
              _viewInBase.containsTarget(_node) || !((ActorUbuntu)_node).getIsRightProxy())
                    return null;

      if (_node.getEorCurrentUserID() == null && _node.getEorCurrentGroupID() == null)
          return null;

    List<DisplayableLinkImpl> l_result = super.selectAddonNewHiddenNodeForActor(
          _viewInBase,  _node, _l_dLinks, _m_l_DisplayableLinks,
          _upd_l_NoProxyOrNoExecuteNodes);
      ActorUbuntu               act = null;
      DisplayableLinkImpl       dLink = null;
      Set<ImmutableGroupIDMember>  l_actorContext = null;
      ImmutableGroupIDMember    mem = null;
      boolean                   exception = false;
      AccessControlLinkImpl     acLink = null;

      // first search: analyze _node with Linux setuid or setgid --------------------------
      if (_node instanceof ExecutableUbuntu &&
              (((ExecutableUbuntu)_node).getEnforceUID() ||
              ((ExecutableUbuntu)_node).getEnforceGID() ) ) {

          for (DisplayableLinkImpl link : _l_dLinks) {

                if (link.getEorSecondEnd() != _node) {
                  exception = true;
                  break;
                }

                            // for testing and tracking
                            //System.out.println(
                              // "in NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor() for "+
                               //"link "+link);

                l_actorContext = linkUtil_.getSecondEndAGcontext(link);

                            // for testing and tracking
                            //System.out.println(
                              // "in NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor() end "+
                               //"of the loop on l_nodes for link "+link);

                if (l_actorContext == null) {
                    l_actorContext = new HashSet(10);
                } // if (l_actorContext == null)
                else {
                    for (Iterator ite = l_actorContext.iterator(); ite.hasNext();) {
                        mem = (ImmutableGroupIDMember)ite.next();
                        // mem is removed even if _node.getEorCurrentUserID() is null
                        if (mem instanceof UserID && ((ExecutableUbuntu)_node).getEnforceUID())
                            ite.remove();
                        if (mem instanceof GroupID && ((ExecutableUbuntu)_node).getEnforceGID())
                            ite.remove();
                    } // for ite
                } // else if (l_actorContext == null)

                if (((ExecutableUbuntu)_node).getEnforceUID() && _node.getEorCurrentUserID() != null) {
                    l_actorContext.add(_node.getEorCurrentUserID());
                    l_actorContext.addAll(Arrays.asList(_node.getEorL_SecondaryGroup()));
                    // the previous groups are kept if _node.getEnforceGID() false
                    link.getEorLastLink().addComment(
                              AccessControlLink.OWNER, AccessControlLink.C_SET_USER);
                } // if

                if (((ExecutableUbuntu)_node).getEnforceGID() && _node.getEorCurrentGroupID() != null) {
                    l_actorContext.add(_node.getEorCurrentGroupID());
                    link.getEorLastLink().addComment(
                              AccessControlLink.CONTAIN, AccessControlLink.C_SET_GROUP);
                } // if

                // clears the AG context of link, and replaces it with l_actorContext
                // l_actorContext may be empty
                link.addProperty(
                    DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION,
                    new Integer(link.noOfNodes() - 1));
                link.addProperty(
                    DisplayableLinkImpl.AG_CONTEXT_SET,
                    ((HashSet)l_actorContext).clone());

                if (l_result == null)
                  l_result = new ArrayList(INITIAL_CAPACITY);

                for (ImmutableGroupIDMember memb : l_actorContext) {

                    if (memb instanceof GroupID)
                      dLink.getEorLastLink().addComment(
                              AccessControlLink.CONTAIN, AccessControlLink.C_SET_GROUP);
                    else if (memb instanceof UserID)
                      dLink.getEorLastLink().addComment(
                              AccessControlLink.OWNER, AccessControlLink.C_SET_USER);

                    dLink = DisplayableLinkImpl.valueOf(link, memb,
                            new AccessControlLinkImpl());
                    if (dLink == null) continue;

                    dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                    if (memb instanceof GroupID &&
                            _node.isMemberToSecondaryGroup((GroupID)memb)) {
                        dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink.getEorLastLink().addComment(
                            AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                    } // if

                    l_result.add(dLink);
                } // for memb
          } // for link
      } // if ((ExecutableUbuntu)_node).getEnforceUID()

      if (exception)
          throw new InternalError("in NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor(),"+
                  " the list of DisplayableLinkImpl as argument "+
                  "contains a link where the argument _node is not the second end.");

      List<ActorUbuntu> l_otherExecutedActorsWithAclPrivOrBridge = new ArrayList(INITIAL_CAPACITY);
      boolean           isForActors = false;
      List              l_AclConnectedResources = null;
      Set<PrivilegeTarget> l_TargetsOfSourcePrivilege = null;

      // second search: _node started from xid/AGO Other rights/actor 
      // test if there is some other-executed actor/acl or privilege or /bridge/actor links
      for (Object obj : ((ACSUbuntuImpl)_node.getEorACS()).getEorL_FromOtherInCurrentContextActors()) {
            act = (ActorUbuntu)obj;

            if (_viewInBase.containsBaseObject(act) || act == _node)
                continue;
            if (!act.getIsRightProxy()) {
                // act is already into _upd_l_NoProxyOrNoExecuteNodes
                continue;
            }

            if (l_otherExecutedActorsWithAclPrivOrBridge == null)
                l_otherExecutedActorsWithAclPrivOrBridge = new ArrayList(INITIAL_CAPACITY);

            l_AclConnectedResources = act.getEorL_AclConnectedResources();
            if (l_AclConnectedResources != null) {
                // searching for actors
                isForActors = false;
                for (Object ob : l_AclConnectedResources) {
                    if (ob instanceof Actor) {
                        isForActors = true;
                        break;
                    } // if
                } // for ob
                if (isForActors)
                    l_otherExecutedActorsWithAclPrivOrBridge.add(act);
            } // if (l_AclConnectedResources != null)

            if (act.getEorACSBridgeTarget() != null)
                l_otherExecutedActorsWithAclPrivOrBridge.add(act);

            // even if there are no privilege in Ubuntu...
            l_TargetsOfSourcePrivilege = ((PrivilegeRightsFactoryImpl)interpreter_.
                    getCoreAlgorithm("Pri")).getL_TargetsOfSourcePrivilege(act);
             if (l_TargetsOfSourcePrivilege != null) {
                // searching for actors
                isForActors = false;
                for (Object ob : l_TargetsOfSourcePrivilege) {
                    if (ob instanceof Actor) {
                        isForActors = true;
                        break;
                    } // if
                } // for ob
                if (isForActors)
                    l_otherExecutedActorsWithAclPrivOrBridge.add(act);
            } // if (l_TargetsOfSourcePrivilege != null)
      } // for obj

      if (l_otherExecutedActorsWithAclPrivOrBridge == null ||
              l_otherExecutedActorsWithAclPrivOrBridge.isEmpty()) {

          //l_result is null from the super method in this version, but this may change
          dLink = null;
          l_actorContext = null;
          acLink = null;
          l_otherExecutedActorsWithAclPrivOrBridge = null;
          act = null;
          l_AclConnectedResources = null;
          l_TargetsOfSourcePrivilege = null;

          return l_result;
      } // if

      Name pairKey = null;
      Set  l_hiddenGroups = null;

      for (Object obj : l_otherExecutedActorsWithAclPrivOrBridge) {
          act = (ActorUbuntu)obj;
          l_hiddenGroups = selectDirectOwnerContainForTarget(act);

                                    // for testing and tracking
                                    //System.out.println(
                                      // "in NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor() for "+
                                       //"the l_otherExecutedActorsWithAclPrivOrBridge item "+act.getName());

          if (_l_dLinks != null) {
              for (Object link : _l_dLinks) {
                    dLink = (DisplayableLinkImpl)link;

                    if (dLink.getEorSecondEnd() != _node) {
                      exception = true;
                      break;
                    }

                                // for testing and tracking
                                //System.out.println(
                                  // "in NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor() for "+
                                   //"dLink "+dLink);

                    l_actorContext = linkUtil_.getSecondEndAGcontext(dLink);

                                // for testing and tracking
                                //System.out.println(
                                  // "in NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor() end "+
                                   //"of the loop on l_nodes for dLink "+dLink);

                    if (l_actorContext == null) continue;

                    if (l_actorContext.contains(_node.getEorCurrentUserID()))
                      l_actorContext.addAll(Arrays.asList(_node.getEorL_SecondaryGroup()));

                    

                    // do not add a new dLink
                    // if the AG context of 'act' has a common entity
                    // with the AG context of the _dLink second end actor
                    if (l_hiddenGroups == null ||
                            !Collections.disjoint(l_actorContext, l_hiddenGroups)) {
                      continue;
                    } // if

                    dLink = addOtherRightsToExtendDisplayableLink(
                            act, dLink, l_actorContext);

                    if (dLink == null) continue;

                    pairKey = ((Name)((BaseObject)dLink.getEorFirstEnd()).
                                          getName()).extended(act.getName());

                    if (link_.detectAddonPriorityInLastLinkAGORights(
                                dLink,
                                (Collection<DisplayableLinkImpl>)_m_l_DisplayableLinks.get(pairKey)))
                                            continue;

                    if (l_result == null)
                      l_result = new ArrayList(INITIAL_CAPACITY);
                    l_result.add(dLink);
              } // for link

              if (exception) break;
              
          } // if (_l_dLinks != null)
          else {

              if (l_actorContext == null)
                  l_actorContext = new HashSet(INITIAL_CAPACITY);
              else
                  l_actorContext.clear();

              if (_node.getEorCurrentUserID() != null) {
                  l_actorContext.add(_node.getEorCurrentUserID());
                  l_actorContext.addAll(Arrays.asList(_node.getEorL_SecondaryGroup()));
              } 
              if (_node.getEorCurrentGroupID() != null) {
                  l_actorContext.add(_node.getEorCurrentGroupID());
              }

              if (l_actorContext.isEmpty()) continue;

              // do not add a new dLink
              // if the AG context of 'act' has a common entity
              // with the AG context of _node
              if (l_hiddenGroups == null || 
                            !Collections.disjoint(l_actorContext, l_hiddenGroups)) {
                  _upd_l_NoProxyOrNoExecuteNodes.add(act);
                  continue;
              } // if

              acLink = null;

              if (_node.getEorCurrentUserID() != null) {
                    acLink = addOtherRightsToAccessControlLink(
                                    _node.getEorCurrentUserID(), 
                                    (ResourceUbuntu) act);
              } // if                      
                else {
                    // check up and add the AGO inherited rights
                    acLink = addOtherRightsToAccessControlLink(
                                    (ImmutableGroupIDMember)l_actorContext.iterator().next(),
                                    (ResourceUbuntu) act);
              } // else


              if (acLink == null) {
                  _upd_l_NoProxyOrNoExecuteNodes.add(act);
                  continue;
              }

              dLink = new DisplayableLinkImpl(_node, act, acLink);

              pairKey = ((Name)_node.getName()).extended(act.getName());

              if (link_.detectAddonPriorityInLastLinkAGORights(
                            dLink,
                            (Collection<DisplayableLinkImpl>)_m_l_DisplayableLinks.get(pairKey)))
                                        continue;

              if (l_result == null)
                    l_result = new ArrayList(INITIAL_CAPACITY);
              
              l_result.add(dLink);

          } // else _l_dLinks == null
      } // for obj

      if (exception)
          throw new InternalError("in NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForActor(),"+
                  " the list of DisplayableLinkImpl as argument "+
                  "contains a link where the argument _node is not the second end.");

      l_otherExecutedActorsWithAclPrivOrBridge = null;
      act = null;
      l_actorContext = null;
      dLink = null;
      acLink = null;
      pairKey = null;
      l_hiddenGroups = null;

      return l_result;
  } // selectAddonNewHiddenNodeForActor()

  /**
   * Detects the hidden nodes that are linked to an Ubuntu GroupIDMember as access source,
   * to extend the current DisplayableLinks ended by this GroupIDMember.
   * Two lists are returned to separate the executing right-proxy nodes.
   * The last argument may be updated to provide the second returned value,
   * and it contains only the nodes on which the property changes have to be listened.<P>
   * Adds the links: <P>
   * - groupidmember/other-executed actor 'B', IF there is
   * some B/acl/resource or an B/bridge/actor links,
   * and IF groupidmember is not a capacity group nor in the AG running context that matches
   * to the 'B' Account and Group (otherwise, it is not other-executed). <P>
   * - capacity group '<CAP_DAC_OVERRIDE>'/executed actor 'B', IF there is
   * some B/acl/resource or an B/bridge/actor links. <P>
   * Does not add the AGO Other rights to the capacity groups.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes. Overridden by the subclasses.
   * Overrides and calls first the super method.
   * Calls addOtherRightsToExtendDisplayableLink() and addOtherRightsToAccessControlLink(),
   * LinkRightsImpl.withAccessThroughNodesTreeFromEP(), UtilityImpl.withExecuteRight(),
   * ACS.getEorL_FromOtherInCurrentContextActors(),
   * AgoRightsFactoryImpl.addOwnerContainRightsToDisplayableLink().
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _node groupIDMember to analyze. Never null.
   * @param _l_dLinks list of links with _node as second end and node of this instance.
   * May be null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @param _upd_l_NoProxyOrNoExecuteNodes updated by the adding of the hidden nodes
   * without executing right or which are not right-proxy nodes, if any.
   * This list is only extended if necessary, as a complement of the returned value of the method.
   * This argument is usually empty at the call of this method, but this is not mandatory.
   * Never null.
   * @return new DisplayableLinkImpls that are extensions or updatings
   * of some DisplayableLinkImpls in _l_dLinks with an Actor or a GroupIDMember
   * as the second end, or null.
   * If _l_dLinks is null, returns only simple links. May be null but never empty.
  * */
  public List<DisplayableLinkImpl> selectAddonNewHiddenNodeForGroupIDMember(
          EPRViewInBase _viewInBase, ImmutableGroupIDMember _node,
          Set<DisplayableLinkImpl> _l_dLinks, Map _m_l_DisplayableLinks,
          List<ACSObject> _upd_l_NoProxyOrNoExecuteNodes) {

                        // for testing and tracking
                        //System.out.println(
                          // "Call to NodeRightsUbuntuImpl.selectAddonNewHiddenNodeForGroupIDMember()");

      if (_node == null || !(_node instanceof ImmutableGroupIDMember) ||
              _viewInBase.containsSource(_node))
                    return null;

    List<DisplayableLinkImpl> l_result = super.selectAddonNewHiddenNodeForGroupIDMember(
          _viewInBase,  _node, _l_dLinks, _m_l_DisplayableLinks,
          _upd_l_NoProxyOrNoExecuteNodes);
      List<ActorUbuntu>             l_ActorsWithAclOrBridge = null;
      DisplayableLinkImpl           dLink = null;
      ActorUbuntu                   act = null;
      boolean                       exception = false;
      ImmutableResource[]           l_res = null;

          // process the capacity groups, and return because
            // the AGO Other rights are not applicable
      if (((LinkRightsUbuntuImpl)link_).getLinuxCapacityGroups(_node.getEorACS()).
              contains(_node)) {


          // returns if _node is not the <CAP_DAC_OVERRIDE> group
          if (!_node.getName().equals(((Name)_node.getEorACS().getName()).extended("<G>").
                       extended("<CAP_DAC_OVERRIDE>"))) {

              return l_result;
          } // if (!_node is not '<CAP_DAC_OVERRIDE>'

          l_res = ((ImmutableGroupID)_node).getEorL_GroupIDOwnResources();

          // test if there are some capacity group '<CAP_DAC_OVERRIDE>' actors
          // as ACL or BRIDGE sources
          for (int i = 0; i < l_res.length; i++) {
              if (!(l_res[i] instanceof ActorUbuntu)) continue;
              act = (ActorUbuntu)l_res[i];

              if (_viewInBase.containsBaseObject(act))
                    continue;
              if (!act.getIsRightProxy()) {
                  // act is already into _upd_l_NoProxyOrNoExecuteNodes
                  continue;
              }

              if (l_ActorsWithAclOrBridge == null)
                  l_ActorsWithAclOrBridge = new ArrayList(INITIAL_CAPACITY);

              if (act.getEorL_AclConnectedResources() != null)
                    l_ActorsWithAclOrBridge.add(act);
              if (act.getEorACSBridgeTarget() != null)
                    l_ActorsWithAclOrBridge.add(act);
          } // for obj

          // searching for
          // groupidmember/executed actor 'B'
          if (l_ActorsWithAclOrBridge != null && !l_ActorsWithAclOrBridge.isEmpty()) {

              for (Object obj : l_ActorsWithAclOrBridge) {
                  act = (ActorUbuntu)obj;

                  if (!withAccessThroughNodesTreeFromEP(_node, act, null)) {
                      _upd_l_NoProxyOrNoExecuteNodes.add(act);
                      continue;
                  }
                  // never executing rights in inherited rights
                  if (!UtilityImpl.getFirstInstance().
                          withExecuteRight(act.getL_groupRights(), false) ){
                      _upd_l_NoProxyOrNoExecuteNodes.add(act);
                      continue;
                  }

                  if (_l_dLinks != null) {

                      for (Object link : _l_dLinks) {
                          dLink = (DisplayableLinkImpl)link;

                          if (dLink.getEorSecondEnd() != _node) {
                              exception = true;
                              break;
                          }

                          if (l_result == null)
                              l_result = new ArrayList(INITIAL_CAPACITY);
                          l_result.add(
                                  ago_.addOwnerContainRightsToDisplayableLink(
                                        DisplayableLinkImpl.valueOf(
                                            dLink, act, new AccessControlLinkImpl()) ));
                      } // for link
                      if (exception) break;
                  } // if (_l_dLinks != null)
                  else {

                      if (l_result == null)
                          l_result = new ArrayList(INITIAL_CAPACITY);
                      l_result.add(
                              ago_.addOwnerContainRightsToDisplayableLink(
                                    new DisplayableLinkImpl(
                                        _node, act, new AccessControlLinkImpl()) ));
                  } // else _l_dLinks == null
              } // for obj

              if (exception)
                  throw new InternalError("in NodeRightsUbuntuImpl."+
                          "selectAddonNewHiddenNodeForGroupIDMember(),"+
                          " the list of DisplayableLinkImpl as argument "+
                          "contains a link where the argument _node is not the second end.");

          } // for (int i = 0; i < l_res.length; i++)
          
          l_res = null;
          l_ActorsWithAclOrBridge = null;
          act = null;      
          dLink = null;
          l_res = null;

          return l_result;        
      } // link_.getLinuxCapacityGroups()

      // process the AGO Other rights
      Set                   l_context = new HashSet(INITIAL_CAPACITY);
      AccessControlLinkImpl acLink = null;
      Name                  pairKey = null;
      Set                   l_hiddenGroups = null;

      // test if there is some other-executed actor/acl/resource or /bridge/actor links
      for (Object obj : ((ACSUbuntuImpl)_node.getEorACS()).getEorL_FromOtherInCurrentContextActors()) {
            act = (ActorUbuntu)obj;

            if (_viewInBase.containsBaseObject(act))
                continue;
            if (!act.getIsRightProxy()) {
                _upd_l_NoProxyOrNoExecuteNodes.add(act);
                continue;
            }

            if (l_ActorsWithAclOrBridge == null)
                l_ActorsWithAclOrBridge = new ArrayList(INITIAL_CAPACITY);

            if (act.getEorL_AclConnectedResources() != null)
                l_ActorsWithAclOrBridge.add(act);
            if (act.getEorACSBridgeTarget() != null)
                l_ActorsWithAclOrBridge.add(act);
      } // for obj

      l_context.add( _node );
      l_context.addAll(Arrays.asList(_node.getEorL_GroupID()));

      // searching for
      // groupidmember/other-executed actor 'B'
      if (l_ActorsWithAclOrBridge != null && !l_ActorsWithAclOrBridge.isEmpty()) {

          for (Object obj : l_ActorsWithAclOrBridge) {
              act = (ActorUbuntu)obj;
              l_hiddenGroups = selectDirectOwnerContainForTarget(act);

              if (_l_dLinks != null) {

                  for (Object link : _l_dLinks) {
                      dLink = (DisplayableLinkImpl)link;

                      if (dLink.getEorSecondEnd() != _node) {
                          exception = true;
                          break;
                      }

                      // do not add a new dLink
                      // if the AG context of 'act' has a common entity
                      // with the AG context of the _dLink second end actor
                      if (l_hiddenGroups != null && 
                              !Collections.disjoint(l_context, l_hiddenGroups)) {
                          continue;
                      } // if

                      dLink = addOtherRightsToExtendDisplayableLink(act, dLink, l_context);

                      if (dLink == null) {
                          _upd_l_NoProxyOrNoExecuteNodes.add(act);
                          continue;
                      }
                    
                      pairKey = ((Name)((BaseObject)dLink.getEorFirstEnd()).
                                          getName()).extended(act.getName());

                      if (link_.detectAddonPriorityInLastLinkAGORights(
                                dLink, 
                                (Collection<DisplayableLinkImpl>)_m_l_DisplayableLinks.get(pairKey)))
                                            continue;

                      if (l_result == null)
                          l_result = new ArrayList(INITIAL_CAPACITY);
                      l_result.add(dLink);

                  } // for link
                  if (exception) break;
              } // if (_l_dLinks != null)
              else {

                  // do not add a new dLink
                  // if the AG context of 'act' has a common entity
                  // with the AG context of the _dLink second end groupidmember
                  if (l_hiddenGroups != null && 
                              !Collections.disjoint(l_context, l_hiddenGroups)) {
                      continue;
                  } // if

                  acLink = addOtherRightsToAccessControlLink(
                            (GroupIDMember)_node, act);

                  if (acLink == null) {
                      _upd_l_NoProxyOrNoExecuteNodes.add(act);
                      continue;
                  }
                  
                  dLink = new DisplayableLinkImpl(_node, act, acLink);

                  pairKey = ((Name)_node.getName()).extended(act.getName());

                  if (link_.detectAddonPriorityInLastLinkAGORights(
                            dLink,
                            (Collection<DisplayableLinkImpl>)_m_l_DisplayableLinks.get(pairKey)))
                                        continue;

                  if (l_result == null)
                      l_result = new ArrayList(INITIAL_CAPACITY);

                  l_result.add(dLink);
              } // else _l_dLinks == null
          } // for obj

          if (exception)
              throw new InternalError("in NodeRightsUbuntuImpl."+
                      "selectAddonNewHiddenNodeForGroupIDMember(),"+
                      " the list of DisplayableLinkImpl as argument "+
                      "contains a link where the argument _node is not the second end.");

      } // if (l_ActorsWithAclOrBridge != null

      l_ActorsWithAclOrBridge = null;
      act = null;      
      dLink = null;
      l_res = null;
      l_context = null;
      acLink = null;
      pairKey = null;
      l_hiddenGroups = null;

      return l_result;
  } // selectAddonNewHiddenNodeForGroupIDMember()

  /**
  * Detects the UserID and the GroupID in the _res ACS,
  * that owns or contains _res directly, and that are in _viewInBase or not.
  * The link type is given in that order: <P>
  * - the userID that owns the resource, <P>
  * - the groupID that contains the resource. <P>
  * Called by detectAddon2Nodes().
  * @param _res node of this instance; is in _viewInBase
  * @return Set of UserID and GroupIDs that owns/contain _res and are not
  * in the view. May be null but not empty.
  **/
  protected Set selectDirectOwnerContainForTarget(ImmutableResource _res) {
    // variables
    Set l_result = null;

    if (_res.getEorUserID() != null) {
        if (l_result == null)
            l_result = new HashSet(2);
        l_result.add(_res.getEorUserID() );

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectDirectOwnerContainForTarget(), "+
                  //  _res.getName().lastComponent()+" is a Resource owned by: " +
                  //  _res.getEorUserID().getName().lastComponent());

    } // if (_res.getEorUserID() != null

    if (_res.getEorGroupID() != null ) {
        if (l_result == null)
            l_result = new HashSet(2);
        l_result.add(_res.getEorGroupID() );

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectDirectOwnerContainForTarget(), "+
                  //  _res.getName().lastComponent()+" is a Resource contained in: " +
                  //  _res.getEorGroupID().getName().lastComponent());

    } // if (_res.getEorGroupID() != null

    return l_result;
  } // selectDirectOwnerContainForTarget()

  /**
  * This specialized AcsAddon method defines the capacity groups '<CAP_DAC_OVERRIDE>' 
  * and '<CAP_DAC_READ_SEARCH>' in Linux Ubuntu for the DisplayableLinks of 2 nodes.
  * For a sketch view, if _center is a capacity group, this method does not return
  * the DisplayableLink from _center to _node  since the targets are too numerous,
  * and it does not return the links from _node as GroupIDMember that is member
  * of the capacity group, since this is the work of another method.
  * If _center is a Resource, this method returns
  * the DisplayableLinks from each _node that is a view capacity group to _center.
  * Called by detectAddon2Nodes(). Calls addCapacityRightsToDisplayableLink().
  * If _center is not null but is not a Resource, there is no operation.
  * @param _viewInBase EPRViewInBase
  * @param _node is the node to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks
  * */
  protected Map select2NodesCapacityRights(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks, BaseObject _center) {

      if (_center == _node)
          return _m_l_DisplayableLinks;

      ImmutableSource       so;
      EligibleParty         ep;
      DisplayableLinkImpl   dLink;
      LinkRights            linkRights = null;

      if (_center != null) {
          if (!(_center instanceof ImmutableResource))
              return _m_l_DisplayableLinks;

          if (_node instanceof ImmutableGroupID &&
                  _node.getEorACS() == ((ACSObject)_center).getEorACS() &&
                  ((LinkRightsUbuntuImpl)link_).getLinuxCapacityGroups(
                        ((ACSObject)_node).getEorACS()).contains((GroupID)_node)) {

                dLink = new DisplayableLinkImpl(_node, _center, (BaseObject[])null);
                if (dLink != null) {
                    dLink = addCapacityRightsToDisplayableLink(dLink);
                } // if (dLink
                if (dLink != null) {
                    linkRights = interpreter_.getLinkRightsInstances().get(
                              _center.getName());
                    linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);
                } // if (dLink

          } // if

          dLink = null;
          return _m_l_DisplayableLinks;
      } // if (_center != null)

      for (Iterator ita = _viewInBase.getEorM_Sources().values().iterator() ; ita.hasNext() ; ) {
          so = (ImmutableSource) ita.next();
          if (so == null || !(so instanceof EligibleParty)) continue;
          ep = (EligibleParty)so;

          if (ep == _node || !(_node instanceof Resource) ||
                    _node.getEorACS() != ep.getEorACS()) continue;

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"select2NodesCapacityRights(), for the target "+_node.getName()+
                            //", process an ep in the view: "+ep.getName());

            // detect <CAP_DAC_OVERRIDE> and <CAP_DAC_READ_SEARCH> 2-nodes relation
           if (!((LinkRightsUbuntuImpl)link_).getLinuxCapacityGroups(
                    ep.getEorACS()).contains(ep))
                           continue;

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"select2NodesCapacityRights(), "+
                            //"process an ep: "+ep.getName());

            dLink = new DisplayableLinkImpl(ep, _node, (BaseObject[])null);
            if (dLink != null) {
                dLink = addCapacityRightsToDisplayableLink(dLink);
            } // if (dLink
            if (dLink != null) {
                linkRights = interpreter_.getLinkRightsInstances().get(
                                _node.getName());
                linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);
            } // if (dLink

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"select2NodesCapacityRights(), "+
                            //"add a DisplayableLinkImpl: "+dLink);
      } // for ita
      
      so = null;
      ep = null;
      dLink = null;

      return _m_l_DisplayableLinks;
  } // select2NodesCapacityRights()

  /**
  * This specialized AcsAddon method defines the capacity groups '<CAP_DAC_OVERRIDE>'
  * and '<CAP_DAC_READ_SEARCH>' in Linux Ubuntu for the DisplayableLinks of 3 nodes
  * where the intermediate node is a capacity group.
  * Called by detectAddon3Nodes().
  * @param _viewInBase EPRViewInBase
  * @param _node is not used
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update
  * @return Map of DisplayableLinks
  * @exception InterruptedException
   */
  protected Map select3NodesCapacityRights(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks)  throws InterruptedException {

      // for Ubuntu
      ImmutableSource       so;
      EligibleParty         ep;
      DisplayableLinkImpl   dLink;
      GroupID               capacity = null;
      LinkRights            linkRights = null;

      for (Iterator ita = _viewInBase.getEorM_Sources().values().iterator() ; ita.hasNext() ; ) {
          so = (ImmutableSource) ita.next();
          if (so == null || !(so instanceof EligibleParty)) continue;
          ep = (EligibleParty)so;

          if (ep == _node || !(_node instanceof Resource)) continue;

          // detect <CAP_DAC_OVERRIDE> 3-nodes relation in Linux
          capacity = (GroupID) ep.getEorACS().getEorM_EligibleParties().get(
                      ((Name)ep.getEorACS().getName()).extended("<G>").
                      extended("<CAP_DAC_OVERRIDE>") );
          if (capacity != null && !_viewInBase.containsBaseObject( capacity ) &&
                  ep.getEorACS().getACSType().indexOf(ACSImpl.LINUX) != -1 &&
            ((ep instanceof UserID) &&
                Arrays.asList(((UserID)ep).getEorL_GroupID()).
                   contains(capacity) ) //|| this case is processed by ThreeNodesRightsFactoryImpl
            //((ep instanceof Actor) &&
            //((Actor)ep).getEorCurrentUserID() != null &&
            //Arrays.asList(((Actor)ep).getEorCurrentUserID().getEorL_GroupID()).
              // contains(capacity ) )
          ) {

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"selectAcsAddon3NodesCapacityRightsOnTarget(), "+
                            //"process an ep which is a member of '<CAP_DAC_OVERRIDE': "+ep.getName());

                dLink = new DisplayableLinkImpl(ep, _node, new EligibleParty[] {capacity});
                if (dLink != null) {
                    dLink = addCapacityRightsToDisplayableLink(dLink);
                } // if (dLink
                if (dLink != null) {
                    linkRights = interpreter_.getLinkRightsInstances().get(
                                _node.getName());
                    linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);
                } // if (dLink

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"selectAcsAddon3NodesCapacityRightsOnTarget(), "+
                            //"add a DisplayableLinkImpl: "+dLink);


          } // if (capacity != null &&


          // detect <CAP_DAC_READ_SEARCH> 3-nodes relation in Linux
          capacity = (GroupID) ep.getEorACS().getEorM_EligibleParties().get(
                      ((Name)ep.getEorACS().getName()).extended("<G>").
                      extended("<CAP_DAC_READ_SEARCH>") );

                          // for testing and tracking
                          //if (ep instanceof UserID)
                            //  System.out.println("In AgoRightsFactoryImpl."+
                            //"detectRootRights(), "+
                            //"capacity from '<CAP_DAC_READ_SEARCH>': "+capacity.getName()+
                            //", and ep is member of capacity: "+Arrays.asList(((UserID)ep).getEorL_GroupID()).
                            //contains(capacity));

          if (capacity != null && !_viewInBase.containsBaseObject( capacity ) &&
                  ep.getEorACS().getACSType().indexOf(ACSImpl.LINUX) != -1 &&
            ((ep instanceof UserID) &&
                Arrays.asList(((UserID)ep).getEorL_GroupID()).
                   contains(capacity) ) //|| this case is processed by CompoundRightsFactoryImpl
            //((ep instanceof Actor) &&
            //((Actor)ep).getEorCurrentUserID() != null &&
            //Arrays.asList(((Actor)ep).getEorCurrentUserID().getEorL_GroupID()).
              // contains(override ) )
          ) {

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"selectAcsAddon3NodesCapacityRightsOnTargets(), "+
                            //"process an ep which is member of '<CAP_DAC_READ_SEARCH>': "+ep.getName());

                dLink = new DisplayableLinkImpl(ep, _node, new EligibleParty[] {capacity});
                if (dLink != null) {
                    dLink = addCapacityRightsToDisplayableLink(dLink);
                } // if (dLink
                if (dLink != null) {
                    linkRights = interpreter_.getLinkRightsInstances().get(
                              _node.getName());
                    linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);
                } // if (dLink

                              // for testing and tracking
                              //System.out.println("In NodeRightsUbuntuImpl."+
                                //"selectAcsAddon3NodesCapacityRightsOnTarget(), "+
                                //"add a DisplayableLinkImpl: "+dLink);

          } // if (capacity != null &&

          if (Thread.interrupted()) {
            throw new InterruptedException();
          }
      } // for ita

      so = null;
      ep = null;
      dLink = null;
      capacity = null;
      linkRights = null;

      return _m_l_DisplayableLinks;
  } // select3NodesCapacityRights()

  /**
   * Defines the AGO other rights to apply to a Resource to extend a DisplayableLink.
   * In the AG context, the UserID is used to get the AGO inherited rights.
   * If the UserID is not found, the first GroupID is taken.
   * Called by selectAddonLastViewNodeForActor() for ThreeNodesRightsFactoryImpl,
   * and by selectAddonNewHiddenNodeForActor() for
   * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
   * Does not filter the capacity groups, so it has to be done by the caller.
   * Calls addOtherRightsToAccessControlLink().
   * @param _target node to analyze. Never null.
   * @param _dLink simple link with ACL rights and an EligibleParty as second end.
   * Never null. 
   * @param _l_epContext AG context of the _dLink second end.
   * Never null nor empty.
   * @return new DisplayableLinkImpl that are an extension of _dlink
   * with _target as the second node, or null. Never empty.
  * */
  protected DisplayableLinkImpl addOtherRightsToExtendDisplayableLink(
          ResourceUbuntu _target, DisplayableLinkImpl _dLink,
          Set<ImmutableGroupIDMember> _l_epContext) {

      if (_l_epContext.isEmpty() || _dLink == null)
          throw new InternalError(
                  "in NodeRightsUbuntuimpl.addOtherRightsToExtendDisplayableLink(), "+
                  "the argument _l_eoContext is empty or _dLink is null.");

                              // for testing and tracking
                              //System.out.println("Call to NodeRightsUbuntuImpl."+
                                //"addOtherRightsToExtendDisplayableLink() for "+_dLink);

      DisplayableLinkImpl           dLink = null;
      AccessControlLinkImpl         acLink = null;
      GroupIDMember                 gm = null;

      // try to find the UserID, or take the GroupID otherwise
      for (Object ob : _l_epContext) {
        if (ob instanceof UserID) {
            gm = (GroupIDMember)ob;
            break;
        }
      } // for ob
      if (gm == null)
          gm = (GroupIDMember)_l_epContext.iterator().next();

        // check up and add the AGO inherited rights
      acLink = addOtherRightsToAccessControlLink(gm, _target);

      if (acLink != null) {
          dLink = DisplayableLinkImpl.valueOf(_dLink, _target, acLink);
      } // if (acLink

      acLink = null;
      gm = null;
      return dLink;
  } // addOtherRightsToExtendDisplayableLink()


  /**
   * UNUSED
  * This specialized AcsAddon method defines the effective applicable 'other' rights
  * for any view Resource from a view EligibleParty.
  * Does not set these rights if the first node does not belong to the same ACS,
  * or if the userID or groupID rights are already applied on the resource.
  * Removes the previous 'other' rights if there is a own/contain right, following the rules
  * 'AGO right: A rights overlay G rights' and 'AGO right: AG rights overlay O rights'.
  * Calls getL_accessRightsThroughNodesTree().
  * @param _viewInBase EPRViewInBase
  * @param _node view node to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  *   with 'other' links associated to every pair (EligibleParty, Resource).
  * */
  protected Map selectOtherRights(EPRViewInBase _viewInBase, 
          ResourceUbuntu _node, Map _m_l_DisplayableLinks, BaseObject _center)  {

      ImmutableSource       so;
      EligibleParty         ep;
      StringRight[]         l_rights;
      DisplayableLinkImpl   dLink;
      List                  l_links;
      boolean               toContinue = false;
      Name                  pairKey;
      LinkRights            linkRights = null;
      AccessControlLinkImpl inhLinkRights = null;

      Map m_result = _m_l_DisplayableLinks;

      for (Iterator ita = _viewInBase.getEorM_Sources().values()
        .iterator() ; ita.hasNext() ; ) {
          so = (ImmutableSource) ita.next();
          if (so == null || !(so instanceof EligibleParty)) continue;
          ep = (EligibleParty)so;

          // ep and res must belong to the same Linux/Unix ACS
          if (_node == ep || _node.getEorACS() != ep.getEorACS()) continue;

          // detect all root relations, but not the capability groups
          if (((ep instanceof UserID) && ((UserID)ep).getOrder() == 0))
            continue;

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"selectOtherRights(), "+
                            //"process an ep: "+ep.getName());

          if (_node.getEorACS().getACSType().indexOf(ACSImpl.LINUX) == -1 ) continue;
          // for a sketch view: only links to _center are processing
          else if (_center != null && _center != _node && _center != ep) continue;

          pairKey = ((Name)ep.getName()).extended(_node.getName());
          l_links = (List) m_result.get(pairKey);

          if (l_links != null) {
                    // unuseful: rules 'AGO right: A rights overlay G rights' and
                    // 'AGO right: AG rights overlay O rights'
                    //l_links = (List)NodeRightsUbuntuImpl.detectPriorityInAllLinksAGORights(l_links);
                    //m_result.put(pairKey, l_links);

            // detect all user or group relations
            // jump if it is the last link has the type OWNER or CONTAIN
            toContinue = false;
            for (Iterator itu = l_links.iterator() ; itu.hasNext() ; ) {
              dLink = (DisplayableLinkImpl)itu.next();
              if (dLink == null) continue;

                    // for testing and tracking
                    //System.out.println("In NodeRightsUbuntuImpl."+
                      //"selectOtherRights(), "+
                      //"analyze a DisplayableLinkImpl: "+dLink);

              // if dLink contain a current account/group relation,
              // it contains 4 nodes, and
              // a new dLink with an 'other' relation is still possible
              if ( dLink.noOfComponents() < 4 &&
                (dLink.getEorLastLink().isTrueForOwnerOrContainLinkType())
              ) {
                toContinue = true;
                break;
              }
            } // for itu
            if (toContinue) continue;
          } // if (l_links != null

                    // for testing and tracking
                    //System.out.println("In NodeRightsUbuntuImpl."+
                      //"selectOtherRights(), "+
                      //"add the AGO Other rights to: "+_node.getName());


          l_rights = _node.getL_otherRights();

          // updates m_l_DisplayableLinks with the new 'other' rights
          dLink = linkUtil_.getDisplayableLink(
                  ep, _node, null, m_result);
          if (dLink == null) {
              dLink = new DisplayableLinkImpl(
                      ep, _node, (BaseObject[])null);
              linkRights = interpreter_.getLinkRightsInstances().get(
                              _node.getName());
              dLink = linkRights.addDisplayableLink(dLink, m_result);
          } // dLink == null

          if (dLink == null) continue;

          // never null
          dLink.getEorFirstLink().setRights(AccessControlLink.GLOBAL, l_rights);

          if (ep instanceof ImmutableUserID)
              inhLinkRights = getL_accessRightsThroughNodesTree(
                _node, (ImmutableUserID)ep, null);
          else if (ep instanceof ImmutableGroupID)
              inhLinkRights = getL_accessRightsThroughNodesTree(
                _node, null, (ImmutableGroupID)ep);
          l_rights = inhLinkRights.getRights(AccessControlLink.GLOBAL);

          if (l_rights != null) {
            for (int j = 0; j < l_rights.length; j++) {
                dLink.getEorFirstLink().addRight(
                        AccessControlLink.GLOBAL, l_rights[j]);
            } // for j
            dLink.getEorFirstLink().addComment(
                            AccessControlLink.GLOBAL,
                            AccessControlLink.C_INHERITED_AG);
          } // if

          dLink.getEorFirstLink().addComment(AccessControlLink.GLOBAL,
                  inhLinkRights.getLinkTypeComments(AccessControlLink.GLOBAL));

          if (ep instanceof ImmutableGroupID)
            dLink.getEorFirstLink().addComment(
                    AccessControlLink.CONTAIN, AccessControlLink.C_WEAK_LINK);

                        // for testing and tracking
                    //System.out.println("In NodeRightsUbuntuImpl."+
                      //"selectOtherRights(), "+
                      //"add a DisplayableLinkImpl: "+dLink);

      } // for ita

      ep = null;
      l_rights = null;
      dLink = null;
      l_links = null;
      so = null;
      linkRights = null;
      inhLinkRights = null;

      return m_result;
  } // selectOtherRights()

  /**
  * This specialized AcsAddon method defines the effective applicable 'other' rights
  * for any pair of nodes in a DisplayableLink.
  * Does not control if the previous node does not belong to the same ACS,
  * or if the userID or groupID rights are applicable.
  * No adding of AGO Other rights if there are no AGO inherited rights.
  * Caution: the 'other' Linux/Unix rights have to be applied
  * when 'owner' and 'contain' relations are not set.
  * This is the responsability of the calling method.
  * If there are non-null 'other' rights, any actor
  * of the same Linux/Unix have these rights on these numerous objects.
  * It should be necessary to track the 'other' rights on actors
  * when the right 'executable' is applicable. That defines numerous access paths,
  * but it adds no more rights when the actor is executed in the account/group
  * context of the caller. So theses cases are not searched. However, there is
  * a most useful case, when an actor has a non-null 'other' rights, and when
  * it is executed under a current account/group, since it is then executable
  * from any eligible party and may open new paths and new righs to its caller.
  * Called by addOtherRightsToExtendDisplayableLink(), selectAddonNewHiddenNodeForActor(),
  * selectAddonNewHiddenNodeForGroupIDMember(), detectAddon2Nodes().
  * @param _gm first node of the returned AccessControlLinkImpl
  * @param _node second node of the returned AccessControlLinkImpl
  * @return a link with the AGO 'Other' rights or null if these rights are null
  * */
  protected AccessControlLinkImpl addOtherRightsToAccessControlLink(
    ImmutableGroupIDMember _gm, ResourceUbuntu _node) {

                          // for testing and tracking
                          //System.out.println("start of NodeRightsUbuntuImpl."+
                            //"addOtherRightsToAccessControlLink(), for "+_gm.getNickName()+
                            //" on "+_node.getNickName());

      if (!withAccessThroughNodesTreeFromEP(_gm, _node, null)) {

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addOtherRightsToAccessControlLink(), for "+_gm.getNickName()+
                            //", no access through the targets tree on "+_node.getNickName());

          return null;
      } // if

      AccessControlLinkImpl result = new AccessControlLinkImpl();
      StringRight[]         l_rights = null;
      AccessControlLinkImpl inhLinkRights = null;

      if (_gm instanceof ImmutableUserID)
          inhLinkRights = getL_accessRightsThroughNodesTree(
            _node, (ImmutableUserID)_gm, null);
      else if (_gm instanceof ImmutableGroupID)
          inhLinkRights = getL_accessRightsThroughNodesTree(
            _node, null, (ImmutableGroupID)_gm);
      if (inhLinkRights != null)
          l_rights = inhLinkRights.getRights(AccessControlLink.GLOBAL);

      if (l_rights != null) {
        for (int j = 0; j < l_rights.length; j++) {
            result.addRight(
                    AccessControlLink.GLOBAL, l_rights[j]);
        } // for j

        l_rights = _node.getL_otherRights();

          // getL_otherRights() may be null or empty, while there is non-null inherited rights
        if (l_rights != null) {
            for (int j = 0; j < l_rights.length; j++) {
                // activate AccessControlLink.GLOBAL
                result.addRight(
                        AccessControlLink.GLOBAL, l_rights[j]);
            } // for j
            if (inhLinkRights != null) {
                  result.addComment(AccessControlLink.GLOBAL,
                            AccessControlLink.C_INHERITED_AG);
                  result.addComment(
                          AccessControlLink.GLOBAL,
                          inhLinkRights.getLinkTypeComments(AccessControlLink.GLOBAL));
            } // if
        } // if
      } // if
      else {

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addOtherRightsToAccessControlLink(), for "+_gm.getNickName()+
                            //", no inherited nor AGO Other rights on "+_node.getNickName());

          result = null;
          return null;
      } // else

      if (_gm instanceof GroupID)
        result.addComment(AccessControlLink.GLOBAL, AccessControlLink.C_WEAK_LINK);

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addOtherRightsToAccessControlLink(), for "+_gm.getNickName()+
                            //", result: "+result);

      inhLinkRights = null;
      return result;
  } // addOtherRightsToAccessControlLink()

  /**
  * This specialized AcsAddon method updates a DisplayableLinkImpl with capacity rights,
  * for the '&lt;files_tree&gt;' resources only.
  * The rights on the last node are updated following the capacity
  * connected to the resource as the previous node: '&lt;CAP_DAC_OVERRIDE&gt;' or
  * '&lt;CAP_DAC_READ_SEARCH&gt;'. Other Linux capacities are not processed. <P>
  * - '&lt;CAP_DAC_OVERRIDE&gt;': override all DAC access, including ACL execute access if
  * [_POSIX_ACL] is defined. Excluding DAC access covered by CAP_LINUX_IMMUTABLE. <P>
  * - '&lt;CAP_DAC_READ_SEARCH&gt;': overrides all DAC restrictions regarding read and search
  * on files and directories, including ACL restrictions if [_POSIX_ACL] is
  * defined. Excluding DAC access covered by CAP_LINUX_IMMUTABLE. <P>
  * Called by select2NodesCapacityRights(), select3NodesCapacityRights(),
  * AgoRightsFactoryImpl.getRootRights() and
  * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()
  * through detectAcsAddonPriorityRightsOnTarget(),
  * selectAddonLastViewNodeAfterGroupIDMember(),
  * selectAddonNewHiddenNodeForGroupIDMember().
  * Calls UtilityImpl.withDirectExecuteRight().
  * @param _dLink is a link having a Resource as second end,
  *     and a GroupID capacity as previous node. May be null.
  * @return the updated _dLink, or null if the rights are not applicable
  * */
  protected DisplayableLinkImpl addCapacityRightsToDisplayableLink(
     DisplayableLinkImpl _dLink) {

                          // for testing and tracking
                          //System.out.println("Call to NodeRightsUbuntuImpl."+
                            //"addCapacityRightsToDisplayableLink() for "+_dLink);

        if (_dLink == null) return null;
        if (!(_dLink.getEorSecondEnd() instanceof Resource) ) {

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addCapacityRightsToDisplayableLink(), first test returns null");

            return null;
        }

        BaseObject[]            l_nodes = _dLink.getEorAllNodes();
        if (!(l_nodes[l_nodes.length - 2] instanceof ImmutableGroupID)) {
            l_nodes = null;

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addCapacityRightsToDisplayableLink(), second test returns null");
            return null;
        }

        ImmutableGroupID        ep = (ImmutableGroupID) l_nodes[l_nodes.length - 2];
        Resource                res = (Resource)_dLink.getEorSecondEnd();

        if ( res.getEorACS() != ep.getEorACS() ||
                !res.getName().immutableComponents().contains("<files_tree>")) {
            l_nodes = null;
            ep = null;
            res = null;

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addCapacityRightsToDisplayableLink(), fourth test returns null");
            return null;
        }

        AccessControlLinkImpl[] l_links = _dLink.getEorAllACLinks();
        AccessControlLinkImpl   beforeLastACLink = null;
        if (l_nodes.length > 2)
                beforeLastACLink = l_links[l_links.length - 2];

        String                  rightType = AccessControlLink.CONTAIN;
        StringRight[]           l_rights = null;

         // detect <CAP_DAC_OVERRIDE> relation in Linux
        if (ep.equals( (GroupID) ep.getEorACS().getEorM_EligibleParties().get(
               ((Name)ep.getEorACS().getName()).extended("<G>").
                   extended("<CAP_DAC_OVERRIDE>") )) ) {

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addCapacityRightsToDisplayableLink(), "+
                            //"process a capacity: "+ep.getName());

              StringRight[] l_owner = null;
              StringRight[] l_group = null;
              StringRight[] l_other = null;
              StringRight[] l_sum = null;

              try {

                    l_owner = ((Resource)res).getL_ownerRights();
                    l_group = ((Resource)res).getL_groupRights();
                    l_other = ((ResourceUbuntu)res).getL_otherRights();
                    l_sum = new StringRight[l_owner.length+l_group.length+l_other.length];
                    System.arraycopy(l_owner, 0, l_sum, 0, l_owner.length);
                    System.arraycopy(l_group, 0, l_sum, l_owner.length, l_group.length);
                    System.arraycopy(l_other, 0, l_sum, l_owner.length+l_group.length, l_other.length);

                              // for testing and tracking
                              //System.out.println("In NodeRightsUbuntuImpl."+
                                //"addCapacityRightsToDisplayableLink(), "+
                                //"l_sum: ");
                              //for (int i = 0; i < l_sum.length; i++) {
                                //  System.out.println("---- for "+i+": "+l_sum[i]);
                              //} // for i

                    // <CAP_DAC_OVERRIDE> grants read and write, and
                    // the executing permission only when at least
                    // one of the file's three execute permission bits is set
                    if (util_.withDirectExecuteRight(l_sum)) {
                        l_rights = new StringRight[] {
                            ep.getEorACS().getFirstRightFromMeta("full_control") };
                    } else {
                        Set<StringRight> set = util_.
                            subtractNotRemovingCommonLowerNodes(
                                ep.getEorACS().getFirstRightFromMeta("full_control"),
                                ep.getEorACS().getFirstRightFromMeta("execute"),
                                ep.getEorACS(),
                                res instanceof Node);
                        l_rights = (StringRight[]) set.toArray(new StringRight[0]);
                        set = null;
                    } // else

                    if (beforeLastACLink != null) {
                        if (l_nodes[l_nodes.length - 2] instanceof Actor)
                            beforeLastACLink.setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                        else if (l_nodes[l_nodes.length - 2] instanceof GroupIDMember)
                            beforeLastACLink.setLinkTypeToTrue(AccessControlLink.IS_MEMBER);
                    } // if

                    for (int i = 0; i < l_rights.length; i++) {
                        _dLink.getEorLastLink().addRight(
                            rightType, l_rights[i]);
                    } // for i

                    _dLink.getEorLastLink().addComment(rightType,
                             "Capability subACS");

                              // for testing and tracking
                              //System.out.println("In NodeRightsUbuntuImpl."+
                                //"addCapacityRightsToDisplayableLink(), "+
                                //"add a DisplayableLinkImpl: "+_dLink);

              } catch (BaseError b) {
                  throw new InternalError("In AgoRightsFactoryImpl."+
                                "addCapacityRightsToDisplayableLink(), "+b.toString());
              }
              l_owner = null;
              l_group = null;
              l_other = null;
              l_sum = null;
        } // if (ep.equals(<CAP_DAC_OVERRIDE>

          // detect <CAP_DAC_READ_SEARCH> relation in Linux
        else if (ep.equals( (GroupID) ep.getEorACS().getEorM_EligibleParties().get(
              ((Name)ep.getEorACS().getName()).extended("<G>").
                    extended("<CAP_DAC_READ_SEARCH>") )) ) {

                          // for testing and tracking
                          //System.out.println("In NodeRightsUbuntuImpl."+
                            //"addCapacityRightsToDisplayableLink(), "+
                            //"process a capacity: "+ep.getName());

                    if (res instanceof Node) {
                        // <CAP_DAC_READ_SEARCH> grants read and search permission
                        // on directories
                        l_rights = new StringRight[] {
                                ep.getEorACS().getFirstRightFromMeta("read"),
                                ep.getEorACS().getFirstRightFromMeta("read_for_directory")};
                    } else if (res instanceof Resource) {
                        // <CAP_DAC_READ_SEARCH> grants read permission on ordinary files
                        l_rights = new StringRight[] {
                                ep.getEorACS().getFirstRightFromMeta("read") };
                    }

                    if (beforeLastACLink != null) {
                        if (l_nodes[l_nodes.length - 2] instanceof Actor)
                            beforeLastACLink.setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                        else if (l_nodes[l_nodes.length - 2] instanceof GroupIDMember)
                            beforeLastACLink.setLinkTypeToTrue(AccessControlLink.IS_MEMBER);
                    }

                    for (int i = 0; i < l_rights.length; i++) {
                        _dLink.getEorLastLink().addRight(
                            rightType, l_rights[i]);
                    } // for i

                    _dLink.getEorLastLink().addComment(rightType,
                             "Capability subACS");

                              // for testing and tracking
                              //System.out.println("In NodeRightsUbuntuImpl."+
                                //"addCapacityRightsToDisplayableLink(), "+
                                //"add a DisplayableLinkImpl: "+_dLink);

        } // if (ep.equals(<CAP_DAC_READ_SEARCH>

                          // for testing and tracking
                          //System.out.println("in NodeRightsUbuntuImpl."+
                            //"addCapacityRightsToDisplayableLink(), "+
                            //"set rights "+rightType+
                            //" for the link: "+_dLink);
                          //for (int i = 0; i < l_rights.length; i++) {
                            //System.out.println("----- for "+i+": "+l_rights[i]);
                          //}

        ep = null;
        res = null;
        l_rights = null;
        l_nodes = null;
        l_links = null;
        l_rights = null;
        rightType = null;
        beforeLastACLink = null;

        return _dLink;
  } // addCapacityRightsToDisplayableLink()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
      super.finalizeForProcess();
      ago_ = null;
      util_ = null;
      linkUtil_ = null;
  } // finalizeForProcess()

  /**
   * @return a descriptor
   */
  public String toString() {
      return "NODERIGHTS_UBUNTU_IMPL of the view "+interpreter_.getSource().getName();
  }

} // NodeRightsUbuntuImpl


