//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.CNot;


 /**
  * This interface is responsible for defining the 20 link types used
  * for qualifying the access control link between two base objects.
  * Only AclEntries and Privileges may define negative rights, in other words,
  * provide some access forbiddings. These negative rights may be propagated in
  * GLOBAL links. SPREAD is a type without any rights, since its comment
  * is used for the 'See why' text. Another types have no rights.
  * Caution: a MANDATORY RULE to comply with is that an AccessControlLink type
  * must have the lenght close to its name lenght, since the AccessControlLinkImpl
  * methods use the variable lenght as criterion. There are 4 criterion in the tests:
  * 'greater than 5', 'lesser than 6', 'lesser than 9' and 'greater than 11'.
  * Beside the link types, there are 12 comments which are started with 'C_'.
  * They are used in the 'See why' texts of the views
  * @see ARoad0.CNot.AccessControlLinkImpl
  */
public interface AccessControlLink {

  /**
  * link type 'global', used for an actor which is running under a secondary group,
  * or for a first rate link with the comment C_FIRST_RATE,
  * or by some AcsAddon for comments on specific cases. May have negative rights.
  * @see #C_SECONDARY_GROUP
  */
  public static final String GLOBAL = "global";

  /**
  * pseudo-structural link type 'bridge', used for inter-ACS bridge relations.
  * '[]' is used for structural and pseudo-structural links when there are rights.
  */
  public static final String BRIDGE = "[brid]";
  
  /**
  * link type 'acl', associated to an AclEntry between 2 base objects
  * which are the access control link ends. Used for inner-ACS and inter-ACS relations.
  */
  public static final String ACL = "acl";

  /**
  * link type 'negativeAcl', associated to a negative AclEntry between
  * 2 base objects.
  */
  public static final String NEG_ACL = "neg acl";

  /**
  * pseudo-structural link type 'owner', associated to an owning relation between
  * an UserID and a Resource. '[]' is used for structural and pseudo
   * structural links when there are rights.
  */
  public static final String OWNER = "[own]";

  /**
  * Pseudo-structural link type 'contain', associated to a GroupID
  * which contains a Resource. This may be a weak link when the ACS policy specifies
  * that other links are stronger, like for instance OWNER in a Linux-Unix operating system.
  * This link is not applicable from a GroupID to a GroupIDMember, since there are
  * no actions is that case (see IS_MEMBER).
  * '[]' is used for structural and pseudo-structural links when there are rights.
  */
  public static final String CONTAIN = "[conta]";

  /**
  * link type 'spread', associated to an applicable relation implying two base objects.
  * The comment for the 'See why' text is set in the SPREAD comments of the last
  * AccessControlLink of each DisplayableLink. This is the only type which has no rights.
  */
  public static final String SPREAD = "spread";

  /**
  * link type 'negSpread', associated to an applicable
  * relation implying two base objects. Used onky for the propagated rights on
  * several AccessControlLinks.
  */
  //public static final String NEG_SPREAD = "negSpread";

  /**
  * structural link type 'member of', from a GroupIDMember to a GroupID.
  */
  public static final String IS_MEMBER = "member of";

  /**
  * structural link type 'is indirect member', from a GroupIDMember to a GroupID
  * like 'is member', but through some intermediate GroupID(s).
  */
  public static final String IS_INDIRECT_MEMBER = "indirect member";

  /**
  * structural link type 'has child', from a Node
  * to one child which is directly under this node.
  */
  public static final String HAS_CHILD = "has child";


  /**
   * For the NoThan view rights, to indicate when the No-More-Than
   * right criteria is not fullfilled by the Eligible Parties set.
   */
  public static final String NO_MORE_COMPLIANCE = "no compliance with the No-More-Than criterion";

  /**
   * For the NoThan view rights, to indicate when the No-Less-Than
   * right criteria is not fullfilled by the Eligible Parties set.
   */
  public static final String NO_LESS_COMPLIANCE = "no compliance with the No-Less-Than criterion";


  // new in 0.7.0

  /**
  * structural link type 'has virtual member', from a VirtualFolder
  * to one ACSObject which belongs to this folder.
  * @since 0.7.0
  */
  public static final String HAS_VIRTUAL_MEMBER = "has virt. member";

  /**
  * link type 'typedPriv', associated to a Privilege between 2 base objects.
  * @since 0.7.0
  */
  public static final String TPRI = "Tpri";

  /**
  * link type 'negativeTypedPriv', associated to a negative Privilege between
  * 2 base objects.
  * @since 0.7.0
  */
  public static final String NEG_TPRI = "neg Tpri";

  /**
  * link type 'linkedPriv', associated to a Privilege between 2 base objects.
  * Used for inner-ACS and inter-ACS relations, like the ACL types.
  * @since 0.7.0
  */
  public static final String LPRI = "Lpri";

  /**
  * link type 'negativeLinkedPriv', associated to a negative Privilege between
  * 2 base objects. Used for inner-ACS and inter-ACS relations,
  * like the ACL types.
  * @since 0.7.0
  */
  public static final String NEG_LPRI = "neg Lpri";

 /**
  * structural link type 'alias' associated to an Alias link from an alias to its reference.
  * '[]' is used for structural and pseudo-structural links when there are rights.
  * @since 0.7.0
  */
  public static final String IS_ALIAS = "alias of";

  /**
  * structural link type 'runs under', associated to a process an Actor
  * which starts in the context of an account, a group or a secondary group,
  * taking the account/group rights through this context.
  * @since 0.7.0
  */
  public static final String RUN_UNDER = "init contxt";

  /**
  * structural link type 'soft runs under', associated to a process 
  * (an Actor, a VirtualFolder) which receives the context of
  * an account, a group or a secondary group through an access path,
  * taking the account/group rights through this context.
  * This link type requires there is no RUN_UNDER type.
  * @since 0.7.0
  */
  public static final String SOFT_RUN_UNDER = "recv contxt";

  /**
  * pseudo-structural link type 'data exchange', between 2 processes which may exchange some data.
  * An example is between a IP network socket and a running process having some network exchanges.
  * This is a weak link since it means that is not possible to known the impact
  * the exchanged data may have on the receiver behavior. To manage that,
  * the ACS modelizes for each data_exchange link, the own-ACS target rights which are propagated
  * through the link, and it uses the metaright 'get_target_rights'.
  * '[]' is used for structural and pseudo-structural links when there are rights.
  * @since 0.7.0
  */
  public static final String DATA_EXCHANGE = "[data exchange]";

 /**
  * structural link type 'launch' from a launcher to an active process.
  * This provides an stronger equivalent to the executing right.
  * @since 0.7.0
  */
  public static final String LAUNCH = "launch";

  /**
   * For the GLOBAL type, this is a comment to say that is for a secondary group.
   * @since 0.7.0
   */
  public static final String C_SECONDARY_GROUP = "secondary group";

  /**
   * For the GLOBAL and the CONTAIN types, this is a comment to say that the first link,
   * for a couple of nodes, is found with some restricted data. The path for this couple
   * has to be drawn differently in the view. This is used for instance when there is a link
   * directly from a GroupID, then the UserID of the AG context is unknown. In this case,
   * the Account rights, if any, delete the Group rights to comply with the ACS rule
   * 'AGO right: A rights deletes GO rights'. If this is a GLOBAL comment while
   * the GLOBAL type remains false, the path is drawn differently, but the 'See Why'
   * text does not contain this property as comment.
   * @since 0.7.0
   * @see AccessControlLinkImpl#isWeak() 
   */
  public static final String C_WEAK_LINK = "<WEAK>";

  /**
   * For the OWNER or the ALIAS type, this is a comment to say that the AG context of execution
   * is changed for the UserID.
   * @since 0.7.0
   */
  public static final String C_SET_USER = "<Sets Owner>";

  /**
   * For the CONTAIN or the ALIAS type, this is a comment to say that the AG context of execution
   * is changed for the UserID.
   * @since 0.7.0
   */
  public static final String C_SET_GROUP = "<Sets Group>";

  /**
   * For the ALIAS type, this is a comment to say that the AG context of execution
   * is may have been changed for the UserID and the GroupID.
   * @since 0.7.0
   */
  public static final String C_MAY_SET_USER_GROUP = "May sets Owner/Group: ";

  /**
   * For the ALIAS type, this is a comment to say that the received rights of the alias
   * are transfered to the alias reference. It is used only in the views for ACS nodes.
   * @since 0.7.0
   */
  public static final String C_RECEIVE_ALIAS_RIGHTS = "Transfers the received rights";

  /**
   * For the ACL and Privilege types, this is a comment to say that an ACS
   * controls another ACS. It is used only in the ACS views.
   * @since 0.7.0
   */
  public static final String C_CONTROL_OTHER_ACS = "<controls another ACS>";

  /**
   * For the ACL types, this is a comment to say that there are rights
   * from inherited ACLs. It is used only in the views for ACS nodes.
   * @since 0.7.0
   */
  public static final String C_INHERITED_ACL = "<inherited ACL>";

  /**
   * This is a comment to say that this is there are inherited AG rights.
   * It is used only in the views for ACS nodes.
   * @since 0.7.0
   */
  public static final String C_INHERITED_AG = "<inherited AG rights>";

  /**
   * For the Privilege types, this is a comment to say that there are rights
   * from inherited Privileges. It is used only in the views for ACS nodes.
   * @since 0.7.0
   */
  public static final String C_INHERITED_PRI = "<inherited PRI>";

  /**
   * For the Global type, this is a comment to say that this is the link
   * to display with a double with. It is used only in the views for ACS nodes.
   * @since 0.7.0
   */
  public static final String C_FIRST_RATE = "<first rate link>";

  /**
   * For the Global type, this is a comment to say that the link has negative rights
   * which are copied from another DisplayableLinkImpl(s).
   * It is used only in the views for ACS nodes.
   * @since 0.7.0
   */
  public static final String C_COPIED_FROM_LINK = "<includes negative rights from other paths>";

  /**
   * For the ACL types, this is a comment to say that there is at least one
   * conditional ACL. It is used only in the views for ACS nodes.
   * @since 0.7.0
   */
  public static final String C_CONDITIONAL_ACL = "<conditional ACL>";

} // AccessControlLink