//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.CNot;

import ARoad0.gBaseInterface.*;

import java.util.Set;
import java.util.List;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.Arrays;

 /**
  * This class is responsible for modeling the rights (and their justications)
  * between two directly connected base objects.
  * An access control link may be viewed as a set of rights with one or several types
  * to say where they come from. The link state is defined by the gWork classes
  * when they analyze the access paths in a view. The two base objects
  * are not referenced in the access control link, but in the DisplayableLinkImpl class
  * which manages a path of base objects and a list of access control links. <P>
  * There are 20 link types, defined in the AccessControlLink interface.
  * The link type value is typically true when there is a right which is defined
  * for this link type, but the right is not always needed nor known when the value is true.
  * The link type value may be true if it is a structural or pseudo-structural type,
  * where there is a default static right or no right at all. However, the rights
  * of all the types are managed in this class, except for the SPREAD type.
  * Only the ACL and Privilege types may define negative rights, which are
  * access forbiddings. <P>
  * There are comments for each type link, and they may be set and ridden even
  * when the link type value is false.
  * Some comments are interpreted by the gWork package to display the access paths.
  * They are AccessControlLink constants. Their roles are presented hereinafter: <P>
  * - for GLOBAL, C_SECONDARY_GROUP tells it is a path link from an Actor to
  * one of its secondary GroupID, with a context similar to the 'current account/group' context, <P>
  * - for the OWNER and CONTAIN types, C_SET_USER, C_SET_GROUP and C_SET_USER_GROUP
  * say that the AG context of execution is changed. <P>
  * - for CONTAIN or GLOBAL, C_WEAK_LINK tells it is a path link from a GroupID without
  * priority control from an UserID, and this produces a dashed-line arrow
  * to display the access path in the view, <P>
  * - for ACL and the Privilege types, C_CONTROL_OTHER_ACS tells an ACS controls
  * another ACS, in the ACS views. <P>
  * - the SPREAD comment is used to keep the summary of all the effective rights
  * in the last AccessControlLink of each DisplayableLinkImpl. This is for the 'See Why' text.
  * It is the responsibility of the method RightsFactory_Facade.getEffectiveRightsForPair(). <P>
  * - the GLOBAL comment is used to add the comment C_FIRST_RATE to drawn
  * a double-with arrow in the view. <P>
  * - the GLOBAL comment is used to add the comment C_COPIED_FROM_LINK to indicate,
  * in the 'See Why' text, that some negative rights from other link(s) have been copied in this link. <P>
  * There are 2 main methods to get a synthesis of all the rights in an AccessControlLink.
  * The getRedondantRights method returns simply the concatenation of all the applicable rights,
  * that is for all the activated link types.
  * The getFilteredRights method returns also all the applicable rights, 
  * but it may provide a filtering of the negative and the positive rights,
  * with some specific sets for the immutable rights. <P>
  * Some AccessControlLinks are able to transfer the Account/Groups context.
  * A first response is returned by the method isTransferingAGcontextWithoutACL(),
  * that returns true when one of the following types is true: OWNER, CONTAIN, IS_MEMBER,
  * IS_INDIRECT_MEMBER, HAS_VIRTUAL_MEMBER, IS_ALIAS, RUN_UNDER, SOFT_RUN_UNDER, LAUNCH.
  * All the ACL provide such a transfert, in the generic algorithms. An ACS addon
  * may change this behavior through its gWork.LinkRight implementation.
  * This class is designed to provide fast methods for the gWork packages.
  * The DisplayableLinkImpl constructors do not allow to add 2 objects
  * that are equals. The equals() method is not implemented in this class.
  * @see ARoad0.CNot.AccessControlLink
  * @see ARoad0.CNot.DisplayableLinkImpl
  */
public final class AccessControlLinkImpl implements AccessControlLink, Cloneable {

  /*
  * protected variables
  */
  private static final long serialVersionUID__ = 0;
  private static final int PRIME = 1000003;
  private static final int INITIAL_CAPACITY = 5;

  protected boolean global_;
  protected boolean bridge_;
  protected boolean acl_;
  protected boolean negAcl_;
  protected boolean owner_;
  protected boolean contain_;
  protected boolean spread_;
  protected boolean isMember_;
  protected boolean hasChild_;
  // new in 0.7.0
  protected boolean isIndirectMember_;
  protected boolean hasVirtualMember_;
  protected boolean typedPriv_;
  protected boolean negTypedPriv_;
  protected boolean linkedPriv_;
  protected boolean negLinkedPriv_;
  protected boolean isAlias_;
  protected boolean runUnder_;
  protected boolean softRunUnder_;
  protected boolean dataExchange_;
  protected boolean launch_;


  // caution: any update should also be done in setEmpty()
  protected StringRight[] l_globalRights_;
  protected StringRight[] l_bridgeRights_;//{
    //BaseManagerImpl.getInstance().getMetaRight("execute")};
  protected StringRight[] l_aclRights_;
  protected StringRight[] l_negAclRights_;
  // "run under" is used for actors running under an UserID or GroupID
  protected StringRight[] l_ownerRights_;//{
        //BaseManagerImpl.getInstance().getMetaRight("|run_under|")}; //String[]{"|is_owner|"};
  protected StringRight[] l_containRights_;//{
        //BaseManagerImpl.getInstance().getMetaRight("|run_under|")}; //String[0];
  // The SPREAD rights array does not exist
  protected StringRight[] l_isMemberRights_; //{
      //BaseManagerImpl.getInstance().getMetaRight("|has_member|")};
  protected StringRight[] l_hasChildRights_;
  // new in 0.7.0
  protected StringRight[] l_isIndirectMemberRights_;
  protected StringRight[] l_hasVirtualMemberRights_;
  protected StringRight[] l_typedPrivRights_;
  protected StringRight[] l_negTypedPrivRights_;
  protected StringRight[] l_linkedPrivRights_;
  protected StringRight[] l_negLinkedPrivRights_;
  protected StringRight[] l_isAliasRights_;
  protected StringRight[] l_runUnderRights_;
  protected StringRight[] l_softRunUnderRights_;
  protected StringRight[] l_dataExchangeRights_;
  protected StringRight[] l_launchRights_;


  protected String globalComments_ = "";
  protected String bridgeComments_ = "";
  protected String aclComments_ = "";
  protected String negAclComments_ = "";
  protected String ownerComments_ = "";
  protected String containComments_ = "";
  /*
   * the SPREAD comments are reserved for the 'See Why' text
   * in the last AccessControlLinkImpl of a DisplayableLinkImpl.
   */
  protected String spreadComments_ = "";
  protected String isMemberComments_ = "";
  protected String hasChildComments_ = "";
  // new in 0.7.0
  protected String isIndirectMemberComments_ = "";
  protected String hasVirtualMemberComments_ = "";
  protected String typedPrivComments_ = "";
  protected String negTypedPrivComments_ = "";
  protected String linkedPrivComments_ = "";
  protected String negLinkedPrivComments_ = "";
  protected String isAliasComments_ = "";
  protected String runUnderComments_ = "";
  protected String softRunUnderComments_ = "";
  protected String dataExchangeComments_ = "";
  protected String launchComments_ = "";

  /*
  * contructors.
  * /

  /**
  * @return empty instance, with all components set to false
  */
  public AccessControlLinkImpl() {
      
          // for testing and tracking
          //System.out.println("AccessControlLinkImpl empty constructor");
         
  } 

//
//  ------------- static methods -----------------------------------------
//
  
   /**
   * Called by setRights().
   * @return the types GLOBAL, NEG_ACL, NEG_LPRI and NEG_TPRI.
   */
  public static List<String> getAllNegativeRightsTypes() {
      List<String> result = new ArrayList(3);
      result.add(AccessControlLink.GLOBAL);
      result.add(AccessControlLink.NEG_ACL);
      result.add(AccessControlLink.NEG_TPRI);
      result.add(AccessControlLink.NEG_LPRI);
      return result;
  } // getAllNegativeRightsTypes()

//
//  ------------- non-static methods -----------------------------------------
//

   /**
   * Activates a link type. Do nothing if the argument is not a link type.
   * If the rights array is null, sets it as an empty array.
   * Caution: the link type rights or comments are still empty.
   * @see ARoad0.CNot.AccessControlLink for the list of the link types
   * @param _linkType link type in AccessControlLink
   */
  public void setLinkTypeToTrue(String _linkType) {

    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (NEG_ACL.equals(_linkType) ) {
                negAcl_ = true;
                if (l_negAclRights_ == null)
                    l_negAclRights_ = new StringRight[0];
            }
            else if (CONTAIN.equals(_linkType) ) {
                contain_ = true;
                if (l_containRights_ == null)
                    l_containRights_ = new StringRight[0];
            }
            else if (NEG_TPRI.equals(_linkType) ) {
                negTypedPriv_ = true;
                if (l_negTypedPrivRights_ == null)
                    l_negTypedPrivRights_ = new StringRight[0];
            }
            else if (NEG_LPRI.equals(_linkType) ) {
                negLinkedPriv_ = true;
                if (l_negLinkedPrivRights_ == null)
                    l_negLinkedPrivRights_ = new StringRight[0];
            }
            else if (IS_ALIAS.equals(_linkType) ) {
                isAlias_ = true;
                if (l_isAliasRights_ == null)
                    l_isAliasRights_ = new StringRight[0];
            }
        } // if (_linkType.length() < 9)
        if (_linkType.length() > 11) {
            if (HAS_VIRTUAL_MEMBER.equals(_linkType) ) {
                hasVirtualMember_ = true;
                if (l_hasVirtualMemberRights_ == null)
                    l_hasVirtualMemberRights_ = new StringRight[0];
            }
            else if (DATA_EXCHANGE.equals(_linkType) ) {
                dataExchange_ = true;
                if (l_dataExchangeRights_ == null)
                    l_dataExchangeRights_ = new StringRight[0];
            }
            else if (IS_INDIRECT_MEMBER.equals(_linkType) ) {
                isIndirectMember_ = true;
                if (l_isIndirectMemberRights_ == null)
                    l_isIndirectMemberRights_ = new StringRight[0];
            }
        } // if (_linkType.length() > 11)
        else {
            if (IS_MEMBER.equals(_linkType) ) {
                isMember_ = true;
                if (l_isMemberRights_ == null)
                    l_isMemberRights_ = new StringRight[0];
            }
            else if (RUN_UNDER.equals(_linkType) ) {
                runUnder_ = true;
                if (l_runUnderRights_ == null)
                    l_runUnderRights_ = new StringRight[0];
            }
            else if (SOFT_RUN_UNDER.equals(_linkType) ) {
                softRunUnder_ = true;
                if (l_softRunUnderRights_ == null)
                    l_softRunUnderRights_ = new StringRight[0];
            }
            else if (HAS_CHILD.equals(_linkType) ) {
                hasChild_ = true;
                if (l_hasChildRights_ == null)
                    l_hasChildRights_ = new StringRight[0];
            }
        } // if (_linkType.length() > 8)
    } // if _linkType.length() > 6

    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) ) {
                global_ = true;
                if (l_globalRights_ == null)
                    l_globalRights_ = new StringRight[0];
            }
            else if (BRIDGE.equals(_linkType) ) {
                bridge_ = true;
                if (l_bridgeRights_ == null)
                    l_bridgeRights_ = new StringRight[0];
            }

            else if (SPREAD.equals(_linkType) ) {
                spread_ = true;
            }

            else if (LAUNCH.equals(_linkType) ) {
                launch_ = true;
                if (l_launchRights_ == null)
                    l_launchRights_ = new StringRight[0];
            }
        } // if (_linkType.length() > 5) {
        else {
            if (ACL.equals(_linkType) ) {
                acl_ = true;
                if (l_aclRights_ == null)
                    l_aclRights_ = new StringRight[0];
            }
            else if (OWNER.equals(_linkType) ) {
                owner_ = true;
                if (l_ownerRights_ == null)
                    l_ownerRights_ = new StringRight[0];
            }
            else if (TPRI.equals(_linkType) ) {
                typedPriv_ = true;
                if (l_typedPrivRights_ == null)
                    l_typedPrivRights_ = new StringRight[0];
            }
            else if (LPRI.equals(_linkType) ) {
                linkedPriv_ = true;
                if (l_linkedPrivRights_ == null)
                    l_linkedPrivRights_ = new StringRight[0];
            }
        } // else if (_linkType.length() > 5) {
    } // else if _linkType.length() > 6

  } // setLinkTypeToTrue()

   /**
   * Desactivates a link type. Does nothing if the argument
   * is not a link type. Is used for invalidating link type rights.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink
   */
  public void setLinkTypeToFalse(String _linkType) {

    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (NEG_ACL.equals(_linkType) ) {
                negAcl_ = false;
            }
            else if (CONTAIN.equals(_linkType) ) {
                contain_ = false;
            }
            else if (NEG_TPRI.equals(_linkType) ) {
                negTypedPriv_ = false;
            }
            else if (NEG_LPRI.equals(_linkType) ) {
                negLinkedPriv_ = false;
            }
            else if (IS_ALIAS.equals(_linkType) ) {
                isAlias_ = false;
            }
        } // if (_linkType.length() < 9)
        if (_linkType.length() > 11) {
            if (HAS_VIRTUAL_MEMBER.equals(_linkType) ) {
                hasVirtualMember_ = false;
            }
            else if (DATA_EXCHANGE.equals(_linkType) ) {
                dataExchange_ = false;
            }
            else if (IS_INDIRECT_MEMBER.equals(_linkType) ) {
                isIndirectMember_ = false;
            }
        } // if (_linkType.length() > 8)
        else {
            if (IS_MEMBER.equals(_linkType) ) {
                isMember_ = false;
            }
            else if (RUN_UNDER.equals(_linkType) ) {
                runUnder_ = false;
            }
            else if (SOFT_RUN_UNDER.equals(_linkType) ) {
                softRunUnder_ = false;
            }
            else if (HAS_CHILD.equals(_linkType) ) {
                hasChild_ = false;
            }
        } // if (_linkType.length() > 8)
    } // if _linkType.length() > 6

    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) ) {
                global_ = false;
            }
            else if (BRIDGE.equals(_linkType) ) {
                bridge_ = false;
            }

            else if (SPREAD.equals(_linkType) ) {
                spread_ = false;
            }

            else if (LAUNCH.equals(_linkType) ) {
                launch_ = false;
            }
        } // if (_linkType.length() > 5) {
        else {
            if (ACL.equals(_linkType) ) {
                acl_ = false;
            }
            else if (OWNER.equals(_linkType) ) {
                owner_ = false;
            }
            else if (TPRI.equals(_linkType) ) {
                typedPriv_ = false;
            }
            else if (LPRI.equals(_linkType) ) {
                linkedPriv_ = false;
            }
        } // else if (_linkType.length() > 5) {
    } // else if _linkType.length() > 6
  } // setLinkTypeToFalse()

   /**
   * Gets the state of activation for one link type.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink
   * @return the activation state for the link type (if the argument
   * is not a link type, return false).
   */
  public boolean getLinkTypeValue(String _linkType) {
      
                  // for testing and tracking
                  //System.out.println("AccessControlLinkImpl.getLinkTypeValue()
                    //for _linkType: "+_linkType);


    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (NEG_ACL.equals(_linkType) ) {
                return negAcl_;
            }
            else if (CONTAIN.equals(_linkType) ) {
                return contain_;
            }
            else if (NEG_TPRI.equals(_linkType) ) {
                return negTypedPriv_;
            }
            else if (NEG_LPRI.equals(_linkType) ) {
                return negLinkedPriv_;
            }
            else if (IS_ALIAS.equals(_linkType) ) {
                return isAlias_;
            }
        } // else if (_linkType.length() < 9)
        if (_linkType.length() > 11) {
            if (HAS_VIRTUAL_MEMBER.equals(_linkType) ) {
                return hasVirtualMember_;
            }
            else if (DATA_EXCHANGE.equals(_linkType) ) {
                return dataExchange_;
            }
            else if (IS_INDIRECT_MEMBER.equals(_linkType) ) {
                return isIndirectMember_;
            }
        } // if (_linkType.length() > 11)
        else {
            if (IS_MEMBER.equals(_linkType) ) {
                return isMember_;
            }
            else if (RUN_UNDER.equals(_linkType) ) {
                return runUnder_;
            }
            else if (SOFT_RUN_UNDER.equals(_linkType) ) {
                return softRunUnder_;
            }
            else if (HAS_CHILD.equals(_linkType) ) {
                return hasChild_;
            }
        } // else
    } // if _linkType.length() > 6

    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) ) {
                return global_;
            }
            else if (BRIDGE.equals(_linkType) ) {
                return bridge_;
            }

            else if (SPREAD.equals(_linkType) ) {
                return spread_;
            }

            else if (LAUNCH.equals(_linkType) ) {
                return launch_;
            }
        } // if (_linkType.length() > 5) {
        else {
            if (ACL.equals(_linkType) ) {
                return acl_;
            }
            else if (OWNER.equals(_linkType) ) {
                return owner_;
            }
            else if (TPRI.equals(_linkType) ) {
                return typedPriv_;
            }
            else if (LPRI.equals(_linkType) ) {
                return linkedPriv_;
            }
        } // else if (_linkType.length() > 5) {
    } // else if _linkType.length() > 6

    return false;
  } // getLinkTypeValue()

   /**
    * NO USE.
   * @return true if at least one link type value is set to 'true'.
   */
  public boolean isTrueForOneLinkType() {
    return (global_ || bridge_ || acl_ || negAcl_ || owner_ || 
            contain_ || spread_ || isMember_ || hasChild_ || isIndirectMember_ ||
            hasVirtualMember_ ||
            typedPriv_ || negTypedPriv_ || linkedPriv_ || negLinkedPriv_ ||
            isAlias_ || runUnder_ || softRunUnder_ || dataExchange_ || launch_);
  } // isTrueForOneLinkType()

   /**
    * Called by Linux and CompoundRightsFactoryImpl.
   * @return true if at least one ACL link type value (ACL, NEG_ACL)
   * is activated.
   */
  public boolean isTrueForOneAclLinkType() {
    return (acl_ || negAcl_);
  } // isTrueForOneAclLinkType()

   /**
    * Called by RightsFactory_Facade.getEffectiveRightsForPair()..
   * @return true if at least one negative ACL link type value (NEG_ACL,
   * NEG_TPRI, NEG_LPRI) is activated.
   */
  public boolean isTrueForOneNegativeLinkType() {
    return (negAcl_|| negTypedPriv_ || negLinkedPriv_);
  } // isTrueForOneNegativeLinkType()

   /**
   * Called by GraphicEPRView.hasFirstRateLinkInOneDLink().
   * The GLOBAL type without activation is used to add the comment C_FIRST_RATE.
   * This is used in a last link of a DisplayableLinkImpl in two cases: <P>
   * - in the generic access path search, if there is a negative-right path
   * (A to B) while there are two paths (A to C) and (C to B) which have positive rights,
   * the path (A to B) is set as 'first-rate' to say the negative right is a prioritary one,
   * and A cannot access to B through C. <P>
   * - outside the generic access path search, in any AcsAddon where specific priorities
   * are managed, like in the MySQL server AcsAddon.
   * @return true if the GLOBAL type comment contains C_FIRST_RATE.
   */
  public boolean isFirstRateLink() {
    if (globalComments_.indexOf(AccessControlLink.C_FIRST_RATE) != -1)
        return true;
    else
        return false; //(negAcl_|| negTypedPriv_ || negLinkedPriv_);
  } // isFirstRateLink()

   /**
   * Called by DisplayableLinkImpl.containsOwnerContainLinkTypes() and
   * LinuxRightsFactoryImpl.addOtherRights()/detectOtherRights().
   * @return true if at least one link type value (CONTAIN,
   * OWNER) is activated.
   */
  public boolean isTrueForOwnerOrContainLinkType() {
    return (contain_ || owner_);
  } // 

   /**
   * Gets the state of activation for (OWNER, IS_MEMBER, HAS_CHILD, IS_INDIRECT_MEMBER,
   * HAS_VIRTUAL_MEMBER, IS_ALIAS, RUN_UNDER, LAUNCH).
   * They include the strengthen types, then the structural types.
   * Called by RightsFactoryUtilities.isActivePath() and
   * RightsFactory_Facade.getEffectiveRightsForPair().
   * @return true if at least one link type is activated
   */
  public boolean isTrueForOnePseudoStructuralLinkType() {
    return ( launch_ || hasVirtualMember_ || softRunUnder_ || isTrueForOneStrengthenLinkType());
  } // 

   /**
   * Gets the state of activation for (IS_MEMBER, HAS_CHILD, IS_INDIRECT_MEMBER,
   * IS_ALIAS). They are also pseudo-structural and strengthen types.
   * Called by DisplayableLinkImpl.containsStructureLinkTypes() which is not used.
   * Called by isTrueForOneStrengthenLinkType().
   * @return true if at least one link type value is set to 'true'.
   */
  public boolean isTrueForOneStructuralLinkType() {
    return (isMember_ || hasChild_ || isIndirectMember_ || isAlias_ );
  } // 

   /**
   * Gets the state of activation for (OWNER, IS_MEMBER, HAS_CHILD, IS_INDIRECT_MEMBER,
   * IS_ALIAS, RUN_UNDER). They include the structural types.
   * SOFT_RUN_UNDER is not a strengthen type.
   * Called by DisplayableLinkImpl.containsStrengthenLinkTypes().
   * @return true if at least one link type value is set to 'true'.
   */
  public boolean isTrueForOneStrengthenLinkType() {
    return (owner_ || runUnder_ || isTrueForOneStructuralLinkType());
  } //

   /**
   * The Account/Group context of a node may be transfered along the next nodes
   * of a DisplayableLinkImpl if the next AccessControlLinks are activated
   * for some given types. To allow this transfer, one of the following types
   * has to be set to true: OWNER, CONTAIN, IS_MEMBER,
   * IS_INDIRECT_MEMBER, HAS_VIRTUAL_MEMBER, IS_ALIAS, RUN_UNDER, SOFT_RUN_UNDER, LAUNCH.
   * The relation ACL tranfers the AG context only for the internal ACL,
   * and this cannot be set by this class.
   * The privilege types do not transfer the AG context in this core method,
   * but an AcsAddon implementation of LinkRight may change this behavior.
   * Called by LinkRightsImpl.isTransferingAGcontext().
   * @see ARoad0.gWork.LinkRightsImpl#isTransferingAGcontext(ARoad0.CNot.AccessControlLinkImpl, ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableACS) 
   * @return true if at least one of the relevant link type value is activated.
   */
  public boolean isTransferingAGcontextWithoutACL() {
    return (//global_ || bridge_ || 
            //acl_ || negAcl_ ||
            owner_ ||
            contain_ || 
            //spread_ || 
            isMember_ || 
            //hasChild_ || 
            isIndirectMember_ ||
            hasVirtualMember_ ||
            //typedPriv_ || negTypedPriv_ || linkedPriv_ || negLinkedPriv_ ||
            isAlias_ || 
            runUnder_ ||
            softRunUnder_ ||
            //dataExchange_ ||
            launch_);
  } // isTransferingAGcontextWithoutACL()
  
   /**
   * Gets all the link type rights where the link type value is activated.
   * The rights are strings and are not redundant for each link type,
   * but may be redundant over the link types of the same DisplayableLink.
   * In other words, there are never two 'execute' rights for one same type,
   * but it is possible to has it both for ACL and CONTAIN for instance.
   * Calls getAllNegativeRights() and getAllPositiveRights().
   * @see ARoad0.CNot.AccessControlLink for link types
   * @return Set of StringRights. May be empty but not null.
   */
  public Set getRedondantRights() {
    Set result = getAllNegativeRights(false);
    result.addAll(getAllPositiveRights(false));
    return result;
  } // getRedondantRights()

   /**
   * Gets all the applicable rights, after a filtering of the negative and
   * the positive rights, with some specific returned sets for the immutable rights.
   * For each returned set, there is no filtering of the headers.
   * There is no SPREAD rights.
   * The 'Other rights' in AGO rights are managed by GLOBAL rights, and GLOBAL
   * that may be also used for 'current account/group', 'ext-acl' or 'bridge'.
   * Called by DisplayableLinkUtilities.getEffectiveRightsAndCommentOnRights().
   * @param _immutableRange is 0 if there is no right to put in the two immutable sets (unused), <P>
   *     1 if only owner rights are to put in the two immutable sets, <P>
   *     2 if owner and structural rights are to be get as immutable (unused),<P>
   *     3 if owner, structural, bridge and data_exchange rights are to be get as immutable (unused).
   * @param _withoutGroupInAGOrights true if the CONTAIN rights are not returned,
   *        and false to return it
   * @param _withoutOtherInAGOrights true if the 'Other rights' in AGO rights
    *         are not returned, and false to return it.
   * @return four Set of StringRights: the immutable positive ones at the index 0,
    *     the immutable negative ones at the index 1, the other positive rights at 2, 
    *     and the other negative ones at 3. May contain empty sets, but the array
    *     and its 4 values are never null.
   */
  public Set[] getFilteredRights(int _immutableRange, boolean _withoutGroupInAGOrights,
          boolean _withoutOtherInAGOrights) {
    Set<StringRight>    l_result = new HashSet(
                                DisplayableLinkImpl.INITIAL_CAPACITY*3); // capacity == 15
    Set<StringRight>    l_negResult = new HashSet(
                                DisplayableLinkImpl.INITIAL_CAPACITY*3); // capacity == 15
    Set<StringRight>    l_immutableResult = new HashSet(
                                DisplayableLinkImpl.INITIAL_CAPACITY*3); // capacity == 15
    Set<StringRight>    l_immutableNegResult = new HashSet(
                                DisplayableLinkImpl.INITIAL_CAPACITY*3); // capacity == 15


    if (bridge_ ) {
      for (int i = 0; i < l_bridgeRights_.length; i++) {
        if (_immutableRange > 2) l_immutableResult.add(l_bridgeRights_[i]);
        else l_result.add(l_bridgeRights_[i]);
      } // for i
    }

    if (acl_ ) {
      for (int i = 0; i < l_aclRights_.length; i++) {
        l_result.add(l_aclRights_[i]);
      } // for i
    }

    if (negAcl_ ) {
      for (int i = 0; i < l_negAclRights_.length; i++) {
        l_negResult.add(l_negAclRights_[i]);
      } // for i
    }        

    if (owner_ ) {
      for (int i = 0; i < l_ownerRights_.length; i++) {
        //if (!l_ownerRights_[i].isGrantingRight()) {
          //  if (_immutableRange > 0) l_immutableNegResult.add(l_ownerRights_[i]);
            //else l_negResult.add(l_ownerRights_[i]);
        //} else {
            if (_immutableRange > 0) l_immutableResult.add(l_ownerRights_[i]);
            else l_result.add(l_ownerRights_[i]);
        //}
      } // for i
    }
    
    if (contain_ && !_withoutGroupInAGOrights) {
      for (int i = 0; i < l_containRights_.length; i++) {
        //if (!l_containRights_[i].isGrantingRight()) {
          //  if (_immutableRange > 1) l_immutableNegResult.add(l_containRights_[i]);
            //else l_negResult.add(l_containRights_[i]);
        //} else {
            //if (_immutableRange > 1) l_immutableResult.add(l_containRights_[i]);
            //else
                l_result.add(l_containRights_[i]);
        //}
      } // for i
    }
    
    if (!_withoutOtherInAGOrights ) {
        if (global_ ) {
          for (int i = 0; i < l_globalRights_.length; i++) {
            if (!l_globalRights_[i].isGrantingRight()) {
                l_negResult.add(l_globalRights_[i]);
            } else l_result.add(l_globalRights_[i]);
          } // for i
        } // if global
    } // if (!_withoutOtherInAGO
    else if (_withoutOtherInAGOrights && 
            (globalComments_.indexOf("current account/group") != -1 ||
            globalComments_.indexOf("ext-acl") != -1 ||
            globalComments_.indexOf("bridge") != -1)) {
        if (global_ ) {
          for (int i = 0; i < l_globalRights_.length; i++) {
            if (!l_globalRights_[i].isGrantingRight()) {
                l_negResult.add(l_globalRights_[i]);
            } else l_result.add(l_globalRights_[i]);
          } // for i
        } // if global
    } // if (!_withoutOtherInAGO

          // for testing and tracking
          //if (!_withoutGroupInAGOrights && !_withoutOtherInAGOrights) {
            //  System.out.println("AccessControlLinkImpl.getRights() with two false boolean arguments. l_result:");
            //  for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
              //    System.out.println("--------"+ite.next());
           //   }
             // System.out.println("--- and l_negResult: ");
             // for (Iterator ite = l_negResult.iterator(); ite.hasNext(); ) {
               //   System.out.println("--------"+ite.next());
             // }
          //} // if

    if (isMember_ ) {
      for (int i = 0; i < l_isMemberRights_.length; i++) {
        //if (!l_isMemberRights_[i].isGrantingRight()) {
          //  if (_immutableRange > 1) l_immutableNegResult.add(l_isMemberRights_[i]);
            //else l_negResult.add(l_isMemberRights_[i]);
        //} else {
            if (_immutableRange > 1) l_immutableResult.add(l_isMemberRights_[i]);
            else l_result.add(l_isMemberRights_[i]);
        //}
      } // for i
    }

    if (hasChild_ ) {
      for (int i = 0; i < l_hasChildRights_.length; i++) {
        //if (!l_hasChildRights_[i].isGrantingRight()) {
          //  if (_immutableRange > 1) l_immutableNegResult.add(l_hasChildRights_[i]);
            //else l_negResult.add(l_hasChildRights_[i]);
        //} else {
            if (_immutableRange > 1) l_immutableResult.add(l_hasChildRights_[i]);
            else l_result.add(l_hasChildRights_[i]);
        //}
      } // for i
    }

    if (isIndirectMember_ ) {
      for (int i = 0; i < l_isIndirectMemberRights_.length; i++) {
        //if (!l_isIndirectMemberRights_[i].isGrantingRight()) {
          //  if (_immutableRange > 1) l_immutableNegResult.add(l_isIndirectMemberRights_[i]);
            //else l_negResult.add(l_isIndirectMemberRights_[i]);
        //} else {
            if (_immutableRange > 1) l_immutableResult.add(l_isIndirectMemberRights_[i]);
            else l_result.add(l_isIndirectMemberRights_[i]);
        //}
      } // for i
    }

    if (hasVirtualMember_ ) {
      for (int i = 0; i < l_hasVirtualMemberRights_.length; i++) {
        //if (!l_hasVirtualMemberRights_[i].isGrantingRight()) {
          //  if (_immutableRange > 1) l_immutableNegResult.add(l_hasVirtualMemberRights_[i]);
            //else l_negResult.add(l_hasVirtualMemberRights_[i]);
        //} else {
            if (_immutableRange > 1) l_immutableResult.add(l_hasVirtualMemberRights_[i]);
            else l_result.add(l_hasVirtualMemberRights_[i]);
        //}
      } // for i
    }

    if (typedPriv_ ) {
      for (int i = 0; i < l_typedPrivRights_.length; i++) {
        l_result.add(l_typedPrivRights_[i]);
      } // for i
    }

    if (negTypedPriv_ ) {
      for (int i = 0; i < l_negTypedPrivRights_.length; i++) {
        l_negResult.add(l_negTypedPrivRights_[i]);
      } // for i
    }

    if (linkedPriv_ ) {
      for (int i = 0; i < l_linkedPrivRights_.length; i++) {
        l_result.add(l_linkedPrivRights_[i]);
      } // for i
    }

    if (negLinkedPriv_ ) {
      for (int i = 0; i < l_negLinkedPrivRights_.length; i++) {
        l_negResult.add(l_negLinkedPrivRights_[i]);
      } // for i
    }

    if (isAlias_ ) {
      for (int i = 0; i < l_isAliasRights_.length; i++) {
        if (_immutableRange > 1) l_immutableResult.add(l_isAliasRights_[i]);
        else l_result.add(l_isAliasRights_[i]);
      } // for i
    }

    if (runUnder_ ) {
      for (int i = 0; i < l_runUnderRights_.length; i++) {
        if (_immutableRange > 1) l_immutableResult.add(l_runUnderRights_[i]);
        else l_result.add(l_runUnderRights_[i]);
      } // for i
    }

    if (softRunUnder_ ) {
      for (int i = 0; i < l_softRunUnderRights_.length; i++) {
        if (_immutableRange > 1) l_immutableResult.add(l_softRunUnderRights_[i]);
        else l_result.add(l_softRunUnderRights_[i]);
      } // for i
    }

    if (dataExchange_ ) {
      for (int i = 0; i < l_dataExchangeRights_.length; i++) {
        if (_immutableRange > 2) l_immutableResult.add(l_dataExchangeRights_[i]);
        else l_result.add(l_dataExchangeRights_[i]);
      } // for i
    }

    if (launch_ ) {
      for (int i = 0; i < l_launchRights_.length; i++) {
        if (_immutableRange > 1) l_immutableResult.add(l_launchRights_[i]);
        else l_result.add(l_launchRights_[i]);
      } // for i
    }

    return new Set[] {l_immutableResult, l_immutableNegResult, l_result, l_negResult};
  } // getFilteredRights()

   /**
   * Gets the types (OWNER, IS_MEMBER, HAS_CHILD, IS_INDIRECT_MEMBER,
   * IS_ALIAS, RUN_UNDER) if they are activated, without any rights.
   * Called by GraphicEPRView.buildUpArrowsAndRights().
   * @see ARoad0.CNot.AccessControlLink for the link types
   * @return the strengthen types. May be empty but not null.
   */
  public String getActivatedStrengthenTypes() {
    StringBuilder result = new StringBuilder();

    if (owner_ ) {
      result.append(OWNER+" ");
    }
    if (isMember_ ) {
      result.append(IS_MEMBER+" ");
    }
    if (hasChild_ ) {
      result.append(HAS_CHILD+" ");
    }
    if (isIndirectMember_ ) {
      result.append(IS_INDIRECT_MEMBER+" ");
    }
    if (isAlias_ ) {
      result.append(IS_ALIAS+" ");
    }
    if (runUnder_ ) {
      result.append(RUN_UNDER+" ");
    }

    return result.toString();
  } // getActivatedStrengthenTypes()

   /**
   * Gets the AccessControlLink types (OWNER, IS_MEMBER, HAS_CHILD, IS_INDIRECT_MEMBER,
   * IS_ALIAS, RUN_UNDER) if they are activated, without any rights.
   * Called by getL_ActivatedPseudoStructuralTypes(),
   * GraphicEPRView.buildUpArrowsAndRights()
   * DisplayableLinkImpl.containsStrengthenLinkTypes() and
   * RightsFactory_Facade.getEffectiveRightsForPair().
   * @see ARoad0.CNot.AccessControlLink for the link types
   * @return the activated strengthen types. May be empty but not null.
   */
  public List<String> getL_ActivatedStrengthenTypes() {
    List<String> result = new ArrayList(INITIAL_CAPACITY*6);

    if (owner_ ) {
      result.add(OWNER);
    }
    if (isMember_ ) {
      result.add(IS_MEMBER);
    }
    if (hasChild_ ) {
      result.add(HAS_CHILD);
    }
    if (isIndirectMember_ ) {
      result.add(IS_INDIRECT_MEMBER);
    }
    if (isAlias_ ) {
      result.add(IS_ALIAS);
    }
    if (runUnder_ ) {
      result.add(RUN_UNDER);
    }

    return result;
  } // getL_ActivatedStrengthenTypes()

   /**
   * Gets the AccessControlLink types (OWNER, IS_MEMBER, HAS_CHILD, IS_INDIRECT_MEMBER,
   * HAS_VIRTUAL_MEMBER, IS_ALIAS, RUN_UNDER, SOFT_RUN_UNDER, LAUNCH) if they are activated,
   * without any rights.
   * Calls getL_ActivatedStrengthenTypes().
   * Called by RightsFactory_Facade.getEffectiveRightsForPair().
   * @see ARoad0.CNot.AccessControlLink for the link types
   * @return the activated types. May be empty but not null.
   */
  public List<String> getL_ActivatedPseudoStructuralTypes() {
    List<String> result = getL_ActivatedStrengthenTypes();
    
    if (hasVirtualMember_ ) {
      result.add(HAS_VIRTUAL_MEMBER);
    }
    if (softRunUnder_ ) {
      result.add(SOFT_RUN_UNDER);
    }
    if (launch_ ) {
      result.add(LAUNCH);
    }

    return result;
  } // getL_ActivatedPseudoStructuralTypes()

   /**
   * Gets all the AccessControlLink types that are activated, without any rights.
   * Calls getL_ActivatedPseudoStructuralTypes().
   * Called by CompoundRightsFactoryImpl.
   * @see ARoad0.CNot.AccessControlLink for the link types
   * @return the activated types. May be empty but not null.
   */
  public List<String> getL_AllActivatedTypes() {
    List<String> result = getL_ActivatedPseudoStructuralTypes();
    
    if (global_ ) {
      result.add(GLOBAL);
    }
    if (bridge_ ) {
      result.add(BRIDGE);
    }  
    if (acl_ ) {
      result.add(ACL);
    }
    if (contain_ ) {
      result.add(CONTAIN);
    }  
    if (typedPriv_ ) {
      result.add(TPRI);
    }
    if (linkedPriv_ ) {
      result.add(LPRI);
    }    
    if (dataExchange_) {
      result.add(DATA_EXCHANGE);
    }
    if (negAcl_ ) {
      result.add(NEG_ACL);
    }
    if (negTypedPriv_ ) {
      result.add(NEG_TPRI);
    }
    if (negLinkedPriv_ ) {
      result.add(NEG_LPRI);
    }
    if (spread_ ) {
      result.add(SPREAD);
    }    
    
    return result;
  } // getL_AllActivatedTypes()

   /**
   * Gets all the positive rights for which the link type is activated.
   * The rights are not redundant for each link type,
   * but may be redundant over the link types of the same DisplayableLink.
   * In other words, there are never two executing rights for one same type,
   * but it is possible to have it both for ACL and for CONTAIN.
   * There is no SPREAD rights.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _withType true to have the link type at the beginning of the rights list
   *  for each type, in the returned Set
   * @return Set of StringRights, in any order. May be empty but not null.
   */
  public Set getAllPositiveRights(boolean _withType) {
    Set result = new HashSet(INITIAL_CAPACITY*2);

    if (global_) {
      if (_withType) result.add(GLOBAL);
      for (int i = 0; i < l_globalRights_.length; i++) {
        if (l_globalRights_[i].isGrantingRight())
            result.add(l_globalRights_[i]);
      } // for i
    }
    if (bridge_) {
      if (_withType) result.add(BRIDGE);
      for (int i = 0; i < l_bridgeRights_.length; i++) {
        result.add(l_bridgeRights_[i]);
      }
    }
    if (acl_ ) {
      if (_withType) result.add(ACL);
      for (int i = 0; i < l_aclRights_.length; i++) {
        result.add(l_aclRights_[i]);
      }
    }
    if (owner_) {
      if (_withType) result.add(OWNER);
      for (int i = 0; i < l_ownerRights_.length; i++) {
        result.add(l_ownerRights_[i]);
      }
    }
    if (contain_ ) {
      if (_withType) result.add(CONTAIN);
      for (int i = 0; i < l_containRights_.length; i++) {
        result.add(l_containRights_[i]);
      }
    }
    if (isMember_ ) {
      if (_withType) result.add(IS_MEMBER);
      for (int i = 0; i < l_isMemberRights_.length; i++) {
        result.add(l_isMemberRights_[i]);
      }
    }
    if (hasChild_ ) {
      if (_withType) result.add(HAS_CHILD);
      for (int i = 0; i < l_hasChildRights_.length; i++) {
        result.add(l_hasChildRights_[i]);
      }
    }

    // new in 0.7.0
    if (isIndirectMember_ ) {
      if (_withType) result.add(IS_INDIRECT_MEMBER);
      for (int i = 0; i < l_isIndirectMemberRights_.length; i++) {
        result.add(l_isIndirectMemberRights_[i]);
      }
    }
    if (hasVirtualMember_ ) {
      if (_withType) result.add(HAS_VIRTUAL_MEMBER);
      for (int i = 0; i < l_hasVirtualMemberRights_.length; i++) {
        result.add(l_hasVirtualMemberRights_[i]);
      }
    }
    if (typedPriv_  ) {
      if (_withType) result.add(TPRI);
      for (int i = 0; i < l_typedPrivRights_.length; i++) {
        result.add(l_typedPrivRights_[i]);
      }
    }
    if (linkedPriv_ ) {
      if (_withType) result.add(LPRI);
      for (int i = 0; i < l_linkedPrivRights_.length; i++) {
        result.add(l_linkedPrivRights_[i]);
      }
    }
    if (isAlias_) {
      if (_withType) result.add(IS_ALIAS);
      for (int i = 0; i < l_isAliasRights_.length; i++) {
        result.add(l_isAliasRights_[i]);
      }
    }
    if (runUnder_ ) {
      if (_withType) result.add(RUN_UNDER);
      for (int i = 0; i < l_runUnderRights_.length; i++) {
        result.add(l_runUnderRights_[i]);
      }
    }
    if (softRunUnder_ ) {
      if (_withType) result.add(SOFT_RUN_UNDER);
      for (int i = 0; i < l_softRunUnderRights_.length; i++) {
        result.add(l_softRunUnderRights_[i]);
      }
    }
    if (dataExchange_ ) {
      if (_withType) result.add(DATA_EXCHANGE);
      for (int i = 0; i < l_dataExchangeRights_.length; i++) {
        result.add(l_dataExchangeRights_[i]);
      }
    }
    if (launch_ ) {
      if (_withType) result.add(LAUNCH);
      for (int i = 0; i < l_launchRights_.length; i++) {
        result.add(l_launchRights_[i]);
      }
    }

    return result;
  } // getAllPositiveRights()

   /**
   * Gets all the link type negative rights where the link type value is activated.
   * The rights are not redundant for each link type,
   * but may be redundant over the link types of the same DisplayableLink.
   * In other words, there are never two 'no executing' rights for one same type,
   * but it is possible to have it both for NEG_ACL and for NEG_LPRI.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _withType true to have the link type at the beginning of the rights list
   *  for each type, in the returned Set
   * @return Set of StringRights, in any order. May be empty but not null.
   */
  public Set getAllNegativeRights(boolean _withType) {
    Set result = new HashSet(INITIAL_CAPACITY*2);
     if (global_) {
      if (_withType) result.add(GLOBAL);
      for (int i = 0; i < l_globalRights_.length; i++) {
        if (!l_globalRights_[i].isGrantingRight())
            result.add(l_globalRights_[i]);
      } // for i
    }
    if (negAcl_ ) {
      if (_withType) result.add(NEG_ACL);
      for (int i = 0; i < l_negAclRights_.length; i++) {
        result.add(l_negAclRights_[i]);
      }
    }
    if (negTypedPriv_ ) {
      if (_withType) result.add(NEG_TPRI);
      for (int i = 0; i < l_negTypedPrivRights_.length; i++) {
        result.add(l_negTypedPrivRights_[i]);
      }
    }
    if (negLinkedPriv_ ) {
      if (_withType) result.add(NEG_LPRI);
      for (int i = 0; i < l_negLinkedPrivRights_.length; i++) {
        result.add(l_negLinkedPrivRights_[i]);
      }
    }
    return result;
  } // getAllNegativeRights()


   /**
   * Gets all the link type comments where the link type value is 'true'.
   * If addLinkTypeComments() has been used, the comments are separated
   * by spaces in each comments type.
   * Note: The comments are strings and are not redundant for each link type
   * and over all the link types.
   * This a quite slow method.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @return Set of String comments. May be empty.
   */
  public Set getAllComments() {
    Set result = new HashSet();

    if (global_ && globalComments_.length() != 0) {
      result.add(globalComments_);
    }
    if (bridge_ && bridgeComments_.length() != 0) {
      result.add(bridgeComments_);
    }
    if (acl_ && aclComments_.length() != 0) {
      result.add(aclComments_);
    }
    if (negAcl_ && negAclComments_.length() != 0) {
      result.add(negAclComments_);
    }
    if (owner_ && ownerComments_.length() != 0) {
      result.add(ownerComments_);
    }
    if (contain_ && containComments_.length() != 0) {
      result.add(containComments_);
    }
    if (spread_ && spreadComments_.length() != 0) {
      result.add(spreadComments_);
    }
    if (isMember_ && isMemberComments_.length() != 0) {
      result.add(isMemberComments_);
    }
    if (hasChild_ && hasChildComments_.length() != 0) {
      result.add(hasChildComments_);
    }

    // new in 0.7.0
    if (isIndirectMember_ && isIndirectMemberComments_.length() != 0) {
      result.add(isIndirectMemberComments_);
    }
    if (hasVirtualMember_ && hasVirtualMemberComments_.length() != 0) {
      result.add(hasVirtualMemberComments_);
    }
    if (typedPriv_ && typedPrivComments_.length() != 0) {
      result.add(typedPrivComments_);
    }
    if (negTypedPriv_ && negTypedPrivComments_.length() != 0) {
      result.add(negTypedPrivComments_);
    }
    if (linkedPriv_ && linkedPrivComments_.length() != 0) {
      result.add(linkedPrivComments_);
    }
    if (negLinkedPriv_ && negLinkedPrivComments_.length() != 0) {
      result.add(negLinkedPrivComments_);
    }
    if (isAlias_ && isAliasComments_.length() != 0) {
      result.add(isAliasComments_);
    }
    if (runUnder_ && runUnderComments_.length() != 0) {
      result.add(runUnderComments_);
    }
    if (softRunUnder_ && softRunUnderComments_.length() != 0) {
      result.add(softRunUnderComments_);
    }
    if (dataExchange_ && dataExchangeComments_.length() != 0) {
      result.add(dataExchangeComments_);
    }
    if (launch_ && launchComments_.length() != 0) {
      result.add(launchComments_);
    }

    return result;
  } // getAllComments()

   /**
   * Gets the rights associated to one link type, if the link type value is true.
   * The SPREAD rights are not ridden by this method.
   * Called by AclRightsFactoryImpl.addACLRightsInDisplayableLink() and
   * CompoundRightsFactoryImpl.detectOneHiddenNodeCompoundRights() for GLOBAL and ACL.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink
   * @return a copy of the rights or an empty array. Never null.
   */
  public StringRight[] getRights(String _linkType) {
    StringRight[] result;

    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (CONTAIN.equals(_linkType) && contain_) {
              result = new StringRight[l_containRights_.length];
              System.arraycopy(l_containRights_, 0, result, 0, result.length);
              return result;
            }
            else if (NEG_ACL.equals(_linkType) && negAcl_) {
              result = new StringRight[l_negAclRights_.length];
              System.arraycopy(l_negAclRights_, 0, result, 0, result.length);
              return result;
            }
            else if (NEG_TPRI.equals(_linkType) && negTypedPriv_) {
              result = new StringRight[l_negTypedPrivRights_.length];
              System.arraycopy(l_negTypedPrivRights_, 0, result, 0, result.length);
              return result;
            }
            else if (NEG_LPRI.equals(_linkType) && negLinkedPriv_) {
              result = new StringRight[l_negLinkedPrivRights_.length];
              System.arraycopy(l_negLinkedPrivRights_, 0, result, 0, result.length);
              return result;
            }
            else if (IS_ALIAS.equals(_linkType) && isAlias_) {
              result = new StringRight[l_isAliasRights_.length];
              System.arraycopy(l_isAliasRights_, 0, result, 0, result.length);
              return result;
            }
        } // if (_linkType.length() < 9)
        else if (_linkType.length() > 11) {
            if (IS_INDIRECT_MEMBER.equals(_linkType) && isIndirectMember_) {
              result = new StringRight[l_isIndirectMemberRights_.length];
              System.arraycopy(l_isIndirectMemberRights_, 0, result, 0, result.length);
              return result;
            }
            else if (HAS_VIRTUAL_MEMBER.equals(_linkType) && hasVirtualMember_) {
              result = new StringRight[l_hasVirtualMemberRights_.length];
              System.arraycopy(l_hasVirtualMemberRights_, 0, result, 0, result.length);
              return result;
            }
            else if (DATA_EXCHANGE.equals(_linkType) && dataExchange_) {
              result = new StringRight[l_dataExchangeRights_.length];
              System.arraycopy(l_dataExchangeRights_, 0, result, 0, result.length);
              return result;
            }
        } // else if (_linkType.length() > 11)
        else {
            if (IS_MEMBER.equals(_linkType) && isMember_) {
              result = new StringRight[l_isMemberRights_.length];
              System.arraycopy(l_isMemberRights_, 0, result, 0, result.length);
              return result;
            }
            else if (HAS_CHILD.equals(_linkType) && hasChild_) {
              result = new StringRight[l_hasChildRights_.length];
              System.arraycopy(l_hasChildRights_, 0, result, 0, result.length);
              return result;
            }
            if (RUN_UNDER.equals(_linkType) && runUnder_) {
              result = new StringRight[l_runUnderRights_.length];
              System.arraycopy(l_runUnderRights_, 0, result, 0, result.length);
              return result;
            }
            if (SOFT_RUN_UNDER.equals(_linkType) && softRunUnder_) {
              result = new StringRight[l_softRunUnderRights_.length];
              System.arraycopy(l_softRunUnderRights_, 0, result, 0, result.length);
              return result;
            }
        } // else
    } // if (_linkType.length() > 6)

    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) && global_) {
              result = new StringRight[l_globalRights_.length];
              System.arraycopy(l_globalRights_, 0, result, 0, result.length);
              return result;
            }
            else if (BRIDGE.equals(_linkType) && bridge_) {
              result = new StringRight[l_bridgeRights_.length];
              System.arraycopy(l_bridgeRights_, 0, result, 0, result.length);
              return result;
            }
            else if (LAUNCH.equals(_linkType) && launch_) {
              result = new StringRight[l_launchRights_.length];
              System.arraycopy(l_launchRights_, 0, result, 0, result.length);
              return result;
            }
        } // if (_linkType.length() > 5)
        else {
            if (ACL.equals(_linkType) && acl_) {
              result = new StringRight[l_aclRights_.length];
              System.arraycopy(l_aclRights_, 0, result, 0, result.length);
              return result;
            }
            else if (OWNER.equals(_linkType) && owner_) {
              result = new StringRight[l_ownerRights_.length];
              System.arraycopy(l_ownerRights_, 0, result, 0, result.length);
              return result;
            }
            else if (TPRI.equals(_linkType) && typedPriv_) {
              result = new StringRight[l_typedPrivRights_.length];
              System.arraycopy(l_typedPrivRights_, 0, result, 0, result.length);
              return result;
            }
            else if (LPRI.equals(_linkType) && linkedPriv_) {
              result = new StringRight[l_linkedPrivRights_.length];
              System.arraycopy(l_linkedPrivRights_, 0, result, 0, result.length);
              return result;
            }
        } // else if (_linkType.length() > 5)
    } // else if (_linkType.length() > 6)
   
    return new StringRight[0];
  } // getRights()
  
   /**
   * Gets the comments associated to one link type, if the link type value is true.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink. If not, return "".
   * @return comments or ""
   */
  public String getLinkTypeComments(String _linkType) {
    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (CONTAIN.equals(_linkType) && contain_)
              return containComments_;
            else if (NEG_ACL.equals(_linkType) && negAcl_)
                return negAclComments_;
            else if (NEG_TPRI.equals(_linkType) && negTypedPriv_)
                return negTypedPrivComments_;
            else if (NEG_LPRI.equals(_linkType) && negLinkedPriv_)
                return negLinkedPrivComments_;
            else if (IS_ALIAS.equals(_linkType) && isAlias_)
              return isAliasComments_;
        }
        else if (_linkType.length() > 11) {
            if (IS_INDIRECT_MEMBER.equals(_linkType) && isIndirectMember_)
              return isIndirectMemberComments_;
            else if (HAS_VIRTUAL_MEMBER.equals(_linkType) && hasVirtualMember_)
                return hasVirtualMemberComments_;
            else if (DATA_EXCHANGE.equals(_linkType) && dataExchange_)
              return dataExchangeComments_;
        } // if (_linkType.length() > 8)
        else {
            if (IS_MEMBER.equals(_linkType) && isMember_)
              return isMemberComments_;
            else if (HAS_CHILD.equals(_linkType) && hasChild_)
              return hasChildComments_;
            else if (RUN_UNDER.equals(_linkType) && runUnder_)
              return runUnderComments_;
            else if (SOFT_RUN_UNDER.equals(_linkType) && softRunUnder_)
              return softRunUnderComments_;
        } // if (_linkType.length() > 8)
    } // if (_linkType.length() > 6)
    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) && global_)
                return globalComments_;
            else if (BRIDGE.equals(_linkType) && bridge_)
                return bridgeComments_;
            else if (SPREAD.equals(_linkType) && spread_)
              return spreadComments_;
            else if (LAUNCH.equals(_linkType) && launch_)
              return launchComments_;
        } // if (_linkType.length() > 5)
        else {
            if (ACL.equals(_linkType) && acl_)
                return aclComments_;
            else if (OWNER.equals(_linkType) && owner_)
                return ownerComments_;
            else if (TPRI.equals(_linkType) && typedPriv_)
                return typedPrivComments_;
            else if (LPRI.equals(_linkType) && linkedPriv_)
                return linkedPrivComments_;

        } // else if (_linkType.length() > 5)
    } // else if (_linkType.length() > 6)

    return "";
  } // getLinkTypeComments()
   /**
   * Sets the comments associated to one link type, whatever the link type value.
   * No action if _linkType is not a link type, or if _comments is null.
   * Operates even if the link type is not activated, but
   * to be in the getAllComments() returned value, the link type value must be activated.
   * The SPREAD type is reserved to the last AccessControlLinkImpl of a DisplayableLinkImpl.
   * The GLOBAL type without activation is used to add the comment C_FIRST_RATE.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink
   * @param _comments may be ""
   */
  public void setLinkTypeComments(String _linkType, String _comments) {
    if (_comments == null) return;
    
    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (CONTAIN.equals(_linkType))
              containComments_ = _comments;
            else if (NEG_ACL.equals(_linkType))
              negAclComments_ = _comments;
            else if (NEG_TPRI.equals(_linkType) )
              negTypedPrivComments_ = _comments;
            else if (NEG_LPRI.equals(_linkType) )
              negLinkedPrivComments_ = _comments;
            else if (IS_ALIAS.equals(_linkType))
              isAliasComments_ = _comments;
        }
        else if (_linkType.length() > 11) {
            if (IS_INDIRECT_MEMBER.equals(_linkType))
              isIndirectMemberComments_ = _comments;
            else if (HAS_VIRTUAL_MEMBER.equals(_linkType))
              hasVirtualMemberComments_ = _comments;
            else if (DATA_EXCHANGE.equals(_linkType))
              dataExchangeComments_ = _comments;
        } // if (_linkType.length() > 8)
        else {
            if (IS_MEMBER.equals(_linkType))
              isMemberComments_ = _comments;
            else if (HAS_CHILD.equals(_linkType))
              hasChildComments_ = _comments;
            else if (RUN_UNDER.equals(_linkType) )
              runUnderComments_ = _comments;
            else if (SOFT_RUN_UNDER.equals(_linkType) )
              softRunUnderComments_ = _comments;
        } // if (_linkType.length() > 8)
    } // if (_linkType.length() > 6)
    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) )
                globalComments_ = _comments;
            else if (BRIDGE.equals(_linkType))
                bridgeComments_ = _comments;
            else if (SPREAD.equals(_linkType) )
                spreadComments_ = _comments;
            else if (LAUNCH.equals(_linkType) )
                launchComments_ = _comments;
        } // if (_linkType.length() > 5)
        else {
            if (ACL.equals(_linkType))
                aclComments_ = _comments;
            else if (OWNER.equals(_linkType) )
                ownerComments_ = _comments;
            else if (TPRI.equals(_linkType) )
                typedPrivComments_ = _comments;
            else if (LPRI.equals(_linkType) )
                linkedPrivComments_ = _comments;
        } // else if (_linkType.length() > 5)
    } // else if (_linkType.length() > 6)
  } // setLinkTypeComments()

   /**
   * Adds the rights associated to one link type, and )activates the link type if necessary.
   * No action on comments. The null, empty or desactivated rights are not added.
   * No action if _linkType is not a link type, if _rights is null or empty, and
   * for the structure links. Controls the redondances among the rights.
   * Only the types GLOBAL, NEG_ACL, NEG_LPRI and NEG_TPRI may have negative rights,
   * and the third last ones cannot have positive rights.
   * The SPREAD type is forbidden, since it is reserved to comments only.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink, excepted SPREAD
   * @param _l_rights may be null, new StringRight[0], or an array of non-null elements.
   * @exception InternalError if the sens of a right does not comply to the link type policy,
   * or if SPREAD is the link type
   */
  public void addRightsArray(String _linkType, StringRight[] _l_rights) {

    if (_l_rights == null || _l_rights.length == 0) return;

    if (!getLinkTypeValue(_linkType)) {
        setRights(_linkType, _l_rights);
        return;
    }

    Set<StringRight> result = new HashSet(INITIAL_CAPACITY*4);
    result.addAll(Arrays.asList(_l_rights));

    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (CONTAIN.equals(_linkType) ) {
              result.addAll(Arrays.asList(l_containRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (NEG_ACL.equals(_linkType)) {
              result.addAll(Arrays.asList(l_negAclRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (NEG_TPRI.equals(_linkType)) {
              result.addAll(Arrays.asList(l_negTypedPrivRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (NEG_LPRI.equals(_linkType)) {
              result.addAll(Arrays.asList(l_negLinkedPrivRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (IS_ALIAS.equals(_linkType)) {
              result.addAll(Arrays.asList(l_isAliasRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
        } // if (_linkType.length() < 9)
        else if (_linkType.length() > 11) {
            if (IS_INDIRECT_MEMBER.equals(_linkType)) {
              result.addAll(Arrays.asList(l_isIndirectMemberRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (HAS_VIRTUAL_MEMBER.equals(_linkType)) {
              result.addAll(Arrays.asList(l_hasVirtualMemberRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (DATA_EXCHANGE.equals(_linkType)) {
              result.addAll(Arrays.asList(l_dataExchangeRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
        } // else if (_linkType.length() > 11)
        else {
            if (IS_MEMBER.equals(_linkType)) {
              result.addAll(Arrays.asList(l_isMemberRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (HAS_CHILD.equals(_linkType)) {
              result.addAll(Arrays.asList(l_hasChildRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            if (RUN_UNDER.equals(_linkType)) {
              result.addAll(Arrays.asList(l_runUnderRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            if (SOFT_RUN_UNDER.equals(_linkType)) {
              result.addAll(Arrays.asList(l_softRunUnderRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
        } // else
    } // if (_linkType.length() > 6)

    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType)) {
              result.addAll(Arrays.asList(l_globalRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (BRIDGE.equals(_linkType)) {
              result.addAll(Arrays.asList(l_bridgeRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (LAUNCH.equals(_linkType) ) {
              result.addAll(Arrays.asList(l_launchRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
        } // if (_linkType.length() > 5)
        else {
            if (ACL.equals(_linkType)) {
              result.addAll(Arrays.asList(l_aclRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (OWNER.equals(_linkType)) {
              result.addAll(Arrays.asList(l_ownerRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (TPRI.equals(_linkType)) {
              result.addAll(Arrays.asList(l_typedPrivRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
            else if (LPRI.equals(_linkType)) {
              result.addAll(Arrays.asList(l_linkedPrivRights_));
              setRights(_linkType, result.toArray(new StringRight[0]));
            }
        } // else if (_linkType.length() > 5)
    } // else if (_linkType.length() > 6)

    result = null;
  } // addRightsArray()

   /**
   * Sets the rights associated to one link type, and (des)activates the link type.
   * If the rights argument is null or empty, the link type is desactivated.
   * No action on comments. The null, empty or desactivated rights are not added.
   * No action if _linkType is not a link type, if _rights is null or empty, and
   * for the structure links. No control of the redondances among the rights.
   * Only the types GLOBAL, NEG_ACL, NEG_LPRI and NEG_TPRI may have negative rights,
   * and the third last ones cannot have positive rights.
   * The SPREAD type is forbidden, since it is reserved to comments only.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink, excepted SPREAD
   * @param _rights may be null, new StringRight[0], or an array of non-null elements.
   * @exception InternalError if the sens of a right does not comply to the link type policy,
   * or if SPREAD is the link type
   */
  public void setRights(String _linkType, StringRight[] _rights) {

    if (SPREAD.equals(_linkType) )
        throw new InternalError("AccessControlLinkImpl.setRights() is "+
                    "not applicable to "+_linkType);

    // variables
    short           shiftingBack = 0;
    StringRight[]   newArray;
    StringRight[]   nullArray = new StringRight[0];
    boolean         error = false;

          // for testing and tracking
          //System.out.println("Call to AccessControlLinkImpl.setRights( , )");

    // newArray initialization
    if (_rights == null || _rights.length == 0) {
        newArray = nullArray;
    }
    else {
      newArray = new StringRight[_rights.length];
      for (int i = 0; i < _rights.length; i++) {
        if (_rights[i] == null || _rights[i].isEmpty() || _rights[i].isDesactivated()) {
          shiftingBack++;
          continue;
        }
        else {
          if (getAllNegativeRightsTypes().contains(_linkType) ) {
              if (_rights[i].isGrantingRight() &&
                      !AccessControlLink.GLOBAL.equals(_linkType)) {
                  error = true;
                  break;
              } // if (_rights[i].isGrantingRight()
          }
          else if (!_rights[i].isGrantingRight()) {
              error = true;
              break;
          }

          newArray[i - shiftingBack] = _rights[i];
        }
      } // for
    } // else

    if (error) {
        newArray = null;
        nullArray = null;
        throw new InternalError("AccessControlLinkImpl.setRights() with "+
                    "a right having a sens which does not comply with the link type "+
                    _linkType);
    } // if (error)

    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (CONTAIN.equals(_linkType) ) {
              if (newArray.length != 0) contain_ = true;
              else contain_ = false;
              l_containRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_containRights_, 0, newArray.length - shiftingBack);

                  // for testing and tracking
                  //if (l_containRights_ != null && l_containRights_.length != 0)
                      //System.out.println("Call to AccessControlLinkImpl.setRights( , ) for CONTAIN,"+
                         //" with at 0: "+l_containRights_[0]);

            } // if (CONTAIN
            else if (NEG_ACL.equals(_linkType) ) {
              if (newArray.length != 0) negAcl_ = true;
              else negAcl_ = false;
              l_negAclRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_negAclRights_, 0, newArray.length - shiftingBack);
            } // if (NEG_ACL
            else if (NEG_TPRI.equals(_linkType)) {
              if (newArray.length != 0) negTypedPriv_ = true;
              else negTypedPriv_ = false;
              l_negTypedPrivRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_negTypedPrivRights_, 0, newArray.length - shiftingBack);
            } // if (negTypedPriv
            else if (NEG_LPRI.equals(_linkType)) {
              if (newArray.length != 0) negLinkedPriv_ = true;
              else negLinkedPriv_ = false;
              l_negLinkedPrivRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_negLinkedPrivRights_, 0, newArray.length - shiftingBack);
            } // if (negLinkedPriv
            else if (IS_ALIAS.equals(_linkType)) {
              if (newArray.length != 0) isAlias_ = true;
              else isAlias_ = false;
              l_isAliasRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_isAliasRights_, 0, newArray.length - shiftingBack);
            } // if (isAlias
        } // if (_linkType.length() < 9)
        else if (_linkType.length() > 11) {
            if (IS_INDIRECT_MEMBER.equals(_linkType)) {
              if (newArray.length != 0) isIndirectMember_ = true;
              else isIndirectMember_ = false;
              l_isIndirectMemberRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_isIndirectMemberRights_, 0, newArray.length - shiftingBack);
            } // if (IS_INDIRECT_MEMBER
            else if (HAS_VIRTUAL_MEMBER.equals(_linkType)) {
              if (newArray.length != 0) hasVirtualMember_ = true;
              else hasVirtualMember_ = false;
              l_hasVirtualMemberRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_hasVirtualMemberRights_, 0, newArray.length - shiftingBack);

                  // for testing and tracking
                  //if (l_hasVirtualMemberRights_ != null && l_hasVirtualMemberRights_.length != 0)
                    //  System.out.println("Call to AccessControlLinkImpl.setRights( , ) for VIRTUALMEMBER,"+
                      //   " with at 0: "+l_hasVirtualMemberRights_[0]);

            } // if (HAS_VIRTUAL_MEMBER
            else if (DATA_EXCHANGE.equals(_linkType)) {
              if (newArray.length != 0) dataExchange_ = true;
              else dataExchange_ = false;
              l_dataExchangeRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_dataExchangeRights_, 0, newArray.length - shiftingBack);
            } // if (dataExchange
        } // if (_linkType.length() > 11)
        else {
            if (IS_MEMBER.equals(_linkType)) {
              if (newArray.length != 0) isMember_ = true;
              else isMember_ = false;
              l_isMemberRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_isMemberRights_, 0, newArray.length - shiftingBack);
            } // if (IS_MEMBER
            else if (HAS_CHILD.equals(_linkType)) {
              if (newArray.length != 0) hasChild_ = true;
              else hasChild_ = false;
              l_hasChildRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_hasChildRights_, 0, newArray.length - shiftingBack);
            } // if (HAS_CHILD
            else if (RUN_UNDER.equals(_linkType)) {
              if (newArray.length != 0) runUnder_ = true;
              else runUnder_ = false;
              l_runUnderRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_runUnderRights_, 0, newArray.length - shiftingBack);
            } // if (RUN_UNDER
            else if (SOFT_RUN_UNDER.equals(_linkType)) {
              if (newArray.length != 0) softRunUnder_ = true;
              else softRunUnder_ = false;
              l_softRunUnderRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_softRunUnderRights_, 0, newArray.length - shiftingBack);
            } // if (RUN_UNDER
        } // else
    } // if (_linkType.length() > 6)
    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) ) {
              if (newArray.length != 0) global_ = true;
              else global_ = false;
              l_globalRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_globalRights_, 0, newArray.length - shiftingBack);
            } // if (global
            else if (BRIDGE.equals(_linkType) ) {
              if (newArray.length != 0) bridge_ = true;
              else bridge_ = false;
              l_bridgeRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_bridgeRights_, 0, newArray.length - shiftingBack);
            } // if (bridge
            else if (LAUNCH.equals(_linkType)) {
              if (newArray.length != 0) launch_ = true;
              else launch_ = false;
              l_launchRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_launchRights_, 0, newArray.length - shiftingBack);
            } // if (launch
        } // if (_linkType.length() > 5)
        else {
            if (ACL.equals(_linkType) ) {
              if (newArray.length != 0) acl_ = true;
              else acl_ = false;
              l_aclRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_aclRights_, 0, newArray.length - shiftingBack);

                  // for testing and tracking
                  //if (l_aclRights_ != null && l_aclRights_.length != 0)
                      //System.out.println("Call to AccessControlLinkImpl.setRights( , ) for ACL,"+
                         //" with at 0 the right: "+l_aclRights_[0]);

            } // if (ACL
            else if (OWNER.equals(_linkType) ) {
              if (newArray.length != 0) owner_ = true;
              else owner_ = false;
              l_ownerRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_ownerRights_, 0, newArray.length - shiftingBack);
            } // if (OWNER
            else if (TPRI.equals(_linkType)) {
              if (newArray.length != 0) typedPriv_ = true;
              else typedPriv_ = false;
              l_typedPrivRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_typedPrivRights_, 0, newArray.length - shiftingBack);
            } // if (TPRI
            else if (LPRI.equals(_linkType)) {
              if (newArray.length != 0) linkedPriv_ = true;
              else linkedPriv_ = false;
              l_linkedPrivRights_ = new StringRight[newArray.length - shiftingBack];
              System.arraycopy(newArray, 0, l_linkedPrivRights_, 0, newArray.length - shiftingBack);
            } // if (linkedPriv
        } // else if (_linkType.length() > 5)
    } // else if (_linkType.length() > 6)

                  // for testing and tracking
                  //if (l_aclRights_ != null && l_aclRights_.length != 0) {
                      //System.out.println("End of AccessControlLinkImpl.setRights( , ) for ACL");
                      //for (int i = 0; i < l_aclRights_.length; i++) {
                          //System.out.println("------ "+l_aclRights_[i]);
                      //}
                      //System.out.println("-- and result from getFilteredRights():");
                      //Set array = getFilteredRights(1, false, false)[2];
                      //for (Object obj : array) {
                          //System.out.println("------ "+obj);
                      //}
                      //System.out.println("-- and result from getRights(AccessControlLink.ACL):");
                      //newArray = getRights(AccessControlLink.ACL);
                      //for (int i = 0; i < newArray.length; i++) {
                          //System.out.println("------ "+newArray[i]);
                      //}
                      //array = null;
                  //} // if

    newArray = null;
    nullArray = null;
  } // setRights()

  /**
   * NO USE.
  * Merges the current rights (which may be empty) and the new rights.
  * No redondance in the result.
  * Quite slow method which calls getRights() and setRights().
  * @see ARoad0.CNot.AccessControlLink for link types
  * @param _linkType is a link type in AccessControlLink.
  * @param _l_rights are the rights to add. May be empty, but not null.
  **/
  public void mergeRightsForOneLinkType(
     String _linkType, StringRight[] _l_rights) {

                        // for testing and tracking
                        if (_l_rights != null)
                            throw new InternalError("From AccessControlLinkImpl.mergeRightsForOneLinkType() "+
                                    "that is an unused method.");

        // variables
        Set    l_result = new HashSet(INITIAL_CAPACITY);
        StringRight[] l_rights = getRights(_linkType);

        for (int j = 0; j < l_rights.length; j++) {
          l_result.add(l_rights[j]);
        } // for j
        for (int j = 0; j < _l_rights.length; j++) {
          l_result.add(_l_rights[j]);
        } // for j

        setRights(_linkType,
          (StringRight[])l_result.toArray(new StringRight[l_result.size()]));  
        
        l_result = null;
        l_rights = null;
  } // mergeRightsForOneLinkType()

   /**
   * Adds a new right to the current rights associated to a link type,
   * and activates the link type if it is false.
   * No action if _linkType is not a link type,
   * if _right is null, empty or already known.
   * The link type state is activated.
   * Optimized to be a fast method.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink
   * @param _right to add
   * @return true if the argument has been addded to, and false otherwise,
   * including when the right is null or has a sens which does not comply to the link type policy
   */
  public boolean addRight(String _linkType, StringRight _right) {
    if (_right == null || _right.isEmpty()) {

                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addRight() "+
                            //      "the right is null or empty for the type "+_linkType);

        return false;
    }

    boolean error = false;

    if (getAllNegativeRightsTypes().contains(_linkType) ) {
      if (_right.isGrantingRight() &&
                      !AccessControlLink.GLOBAL.equals(_linkType)) {

                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addRight() "+
                            //      "error with "+_right+" where negative ones are required for"+
                              //    _linkType);

          error = true;
      } // if (_right.isGrantingRight()
    }
    else if (!_right.isGrantingRight()) {


                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addRight() "+
                           //       "error with "+_right+" where positive ones are required for"+
                            //      _linkType);

      error = true;
    }

    if (error) return false;

    StringRight[] l_rights = null;


                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addRight() "+
                            //      "normal operation for the type "+_linkType+" with the length "+
                              //    _linkType.length());

    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {

            if (CONTAIN.equals(_linkType) ) {
              l_rights = l_containRights_;
              if (l_rights == null || l_rights.length == 0) {
                  contain_ = true;
                  l_containRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {

                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addRight() for CONTAIN"+
                            //      "_right "+_right+"is already known");

                      l_rights = null;
                      return false;
                  }
                  l_containRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_containRights_, 0, l_rights.length);
              }
              l_containRights_[l_containRights_.length - 1] = _right;


                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addRight() for CONTAIN"+
                            //      "- result:");
                          //for (int i = 0; i < l_containRights_.length; i++) {
                            //  System.out.println("--------- "+l_containRights_[i]);
                          //} // for i

            } // if (CONTAIN

            else if (NEG_ACL.equals(_linkType) ) {
              l_rights = l_negAclRights_;
              if (l_rights == null || l_rights.length == 0) {
                  negAcl_ = true;
                  l_negAclRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_negAclRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_negAclRights_, 0, l_rights.length);
              }
              l_negAclRights_[l_negAclRights_.length - 1] = _right;
            } // if (NEG_ACL

            else if (NEG_TPRI.equals(_linkType)) {
              l_rights = l_negTypedPrivRights_;
              if (l_rights == null || l_rights.length == 0) {
                  negTypedPriv_ = true;
                  l_negTypedPrivRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_negTypedPrivRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_negTypedPrivRights_, 0, l_rights.length);
              }
              l_negTypedPrivRights_[l_negTypedPrivRights_.length - 1] = _right;
            } // if (negTypedPriv

            else if (NEG_LPRI.equals(_linkType)) {
              l_rights = l_negLinkedPrivRights_;
              if (l_rights == null || l_rights.length == 0) {
                  negLinkedPriv_ = true;
                  l_negLinkedPrivRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_negLinkedPrivRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_negLinkedPrivRights_, 0, l_rights.length);
              }
              l_negLinkedPrivRights_[l_negLinkedPrivRights_.length - 1] = _right;
            } // if (negLinkedPriv

            else if (IS_ALIAS.equals(_linkType)) {
              l_rights = l_isAliasRights_;
              if (l_rights == null || l_rights.length == 0) {
                  isAlias_ = true;
                  l_isAliasRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_isAliasRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_isAliasRights_, 0, l_rights.length);
              }
              l_isAliasRights_[l_isAliasRights_.length - 1] = _right;
            } // if (isAlias
        } // if (_linkType.length() < 9)
        else if (_linkType.length() > 11) {

            if (IS_INDIRECT_MEMBER.equals(_linkType)) {
              l_rights = l_isIndirectMemberRights_;
              if (l_rights == null || l_rights.length == 0) {
                  isIndirectMember_ = true;
                  l_isIndirectMemberRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_isIndirectMemberRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_isIndirectMemberRights_, 0, l_rights.length);
              }
              l_isIndirectMemberRights_[l_isIndirectMemberRights_.length - 1] = _right;
            } // if (IS_INDIRECT_MEMBER

            else if (HAS_VIRTUAL_MEMBER.equals(_linkType)) {
              l_rights = l_hasVirtualMemberRights_;
              if (l_rights == null || l_rights.length == 0) {
                  hasVirtualMember_ = true;
                  l_hasVirtualMemberRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_hasVirtualMemberRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_hasVirtualMemberRights_, 0, l_rights.length);
              }
              l_hasVirtualMemberRights_[l_hasVirtualMemberRights_.length - 1] = _right;
            } // if (HAS_VIRTUAL_MEMBER

            else if (DATA_EXCHANGE.equals(_linkType)) {
              l_rights = l_dataExchangeRights_;
              if (l_rights == null || l_rights.length == 0) {
                  dataExchange_ = true;
                  l_dataExchangeRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_dataExchangeRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_dataExchangeRights_, 0, l_rights.length);
              }
              l_dataExchangeRights_[l_dataExchangeRights_.length - 1] = _right;
            } // if (dataExchange

        } // if (_linkType.length() > 11)
        else {

            if (IS_MEMBER.equals(_linkType)) {
              l_rights = l_isMemberRights_;
              if (l_rights == null || l_rights.length == 0) {
                  isMember_ = true;
                  l_isMemberRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_isMemberRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_isMemberRights_, 0, l_rights.length);
              }
              l_isMemberRights_[l_isMemberRights_.length - 1] = _right;
            } // if (IS_MEMBER

            else if (HAS_CHILD.equals(_linkType)) {
              l_rights = l_hasChildRights_;
              if (l_rights == null || l_rights.length == 0) {
                  hasChild_ = true;
                  l_hasChildRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_hasChildRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_hasChildRights_, 0, l_rights.length);
              }
              l_hasChildRights_[l_hasChildRights_.length - 1] = _right;
            } // if (HAS_CHILD

            else if (RUN_UNDER.equals(_linkType)) {
              l_rights = l_runUnderRights_;
              if (l_rights == null || l_rights.length == 0) {
                  runUnder_ = true;
                  l_runUnderRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_runUnderRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_runUnderRights_, 0, l_rights.length);
              }
              l_runUnderRights_[l_runUnderRights_.length - 1] = _right;
            } // if (RUN_UNDER

            else if (SOFT_RUN_UNDER.equals(_linkType)) {
              l_rights = l_softRunUnderRights_;
              if (l_rights == null || l_rights.length == 0) {
                  softRunUnder_ = true;
                  l_softRunUnderRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_softRunUnderRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_softRunUnderRights_, 0, l_rights.length);
              }
              l_softRunUnderRights_[l_softRunUnderRights_.length - 1] = _right;
            } // if (RUN_UNDER

        } // else

    } // if (_linkType.length() > 6)
    else {
        if (_linkType.length() > 5) {

            if (GLOBAL.equals(_linkType) ) {
              l_rights = l_globalRights_;
              if (l_rights == null || l_rights.length == 0) {
                  global_ = true;
                  l_globalRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_globalRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_globalRights_, 0, l_rights.length);
              }
              l_globalRights_[l_globalRights_.length - 1] = _right;

                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addRight() for GLOBAL"+
                            //      "- result:");
                          //for (int i = 0; i < l_globalRights_.length; i++) {
                            //  System.out.println("--------- "+l_globalRights_[i]);
                          //} // for i

            } // if (global

            else if (BRIDGE.equals(_linkType) ) {
              l_rights = l_bridgeRights_;
              if (l_rights == null || l_rights.length == 0) {
                  bridge_ = true;
                  l_bridgeRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_bridgeRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_bridgeRights_, 0, l_rights.length);
              }
              l_bridgeRights_[l_bridgeRights_.length - 1] = _right;
            } // if (bridge

            else if (LAUNCH.equals(_linkType)) {
              l_rights = l_launchRights_;
              if (l_rights == null || l_rights.length == 0) {
                  launch_ = true;
                  l_launchRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_launchRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_launchRights_, 0, l_rights.length);
              }
              l_launchRights_[l_launchRights_.length - 1] = _right;
            } // if (launch

        } // if (_linkType.length() > 5)
        else {

            if (ACL.equals(_linkType) ) {
              l_rights = l_aclRights_;
              if (l_rights == null || l_rights.length == 0) {
                  acl_ = true;
                  l_aclRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_aclRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_aclRights_, 0, l_rights.length);
              }
              l_aclRights_[l_aclRights_.length - 1] = _right;


                  // for testing and tracking
                  //if (l_aclRights_ != null && l_aclRights_.length != 0)
                    //  System.out.println("Call to AccessControlLinkImpl.addRight( , ) for ACL,"+
                      //   " with at 0 the right: "+l_aclRights_[0]);

            } // if (ACL

            else if (OWNER.equals(_linkType) ) {
              l_rights = l_ownerRights_;
              if (l_rights == null || l_rights.length == 0) {
                  owner_ = true;
                  l_ownerRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_ownerRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_ownerRights_, 0, l_rights.length);
              }
              l_ownerRights_[l_ownerRights_.length - 1] = _right;
            } // if (OWNER

            else if (TPRI.equals(_linkType)) {
              l_rights = l_typedPrivRights_;
              if (l_rights == null || l_rights.length == 0) {
                  typedPriv_ = true;
                  l_typedPrivRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_typedPrivRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_typedPrivRights_, 0, l_rights.length);
              }
              l_typedPrivRights_[l_typedPrivRights_.length - 1] = _right;
            } // if (TPRI

            else if (LPRI.equals(_linkType)) {
              l_rights = l_linkedPrivRights_;
              if (l_rights == null || l_rights.length == 0) {
                  linkedPriv_ = true;
                  l_linkedPrivRights_ = new StringRight[1];
              }
              else {
                  if (Arrays.asList(l_rights).contains(_right)) {
                      l_rights = null;
                      return false;
                  }
                  l_linkedPrivRights_ = new StringRight[l_rights.length + 1];
                  System.arraycopy(l_rights, 0, l_linkedPrivRights_, 0, l_rights.length);
              }
              l_linkedPrivRights_[l_linkedPrivRights_.length - 1] = _right;
            } // if (linkedPriv

        } // else if (_linkType.length() > 5)
    } // else if (_linkType.length() > 6)

    l_rights = null;
    return true;
  } // addRight()

   /**
   * Adds a new comment to the current comments associated to a link type,
   * if it is not already there.
   * The new comments are added to the current ones after a space.
   * No action if _linkType is not a link type,
   * if _comments is null, empty, or already known.
   * Caution: the link type state is not controlled, and it is not activated
   * by this method.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink
   * @param _comments to add
   */
  public void addComment(String _linkType, String _comments) {
    if (_comments == null || _comments.length() == 0) return;

    if (_linkType.length() > 6) {
        if (_linkType.length() < 9) {
            if (CONTAIN.equals(_linkType) ) {
              if (containComments_.indexOf(_comments) == -1) {
                if (containComments_.length() != 0)
                    containComments_ = containComments_ + " " +_comments;
                else
                    containComments_ = _comments;
                containComments_.trim();
              }
            }
            else if (NEG_ACL.equals(_linkType) ) {
              if (negAclComments_.indexOf(_comments) == -1) {
                if (negAclComments_.length() != 0)
                    negAclComments_ = negAclComments_ + " " +_comments;
                else
                    negAclComments_ = _comments;
                negAclComments_.trim();
              }
            }
            else if (NEG_TPRI.equals(_linkType) ) {
              if (negTypedPrivComments_.indexOf(_comments) == -1) {
                if (negTypedPrivComments_.length() != 0)
                    negTypedPrivComments_ = negTypedPrivComments_ + " " +_comments;
                else
                    negTypedPrivComments_ = _comments;
                negTypedPrivComments_.trim();
              }
            }
            else if (NEG_LPRI.equals(_linkType) ) {
              if (negLinkedPrivComments_.indexOf(_comments) == -1) {
                if (negLinkedPrivComments_.length() != 0)
                    negLinkedPrivComments_ = negLinkedPrivComments_ + " " +_comments;
                else
                    negLinkedPrivComments_ = _comments;
                negLinkedPrivComments_.trim();
              }
            }
            else if (IS_ALIAS.equals(_linkType) ) {
              if (isAliasComments_.indexOf(_comments) == -1) {
                if (isAliasComments_.length() != 0)
                    isAliasComments_ = isAliasComments_ + " " +_comments;
                else
                    isAliasComments_ = _comments;
                isAliasComments_.trim();
              }
            }
        } // if (_linkType.length() < 9)
        else if (_linkType.length() > 11) {
            if (IS_INDIRECT_MEMBER.equals(_linkType) ) {
              if (isIndirectMemberComments_.indexOf(_comments) == -1) {
                if (isIndirectMemberComments_.length() != 0)
                    isIndirectMemberComments_ = isIndirectMemberComments_ + " " +_comments;
                else
                    isIndirectMemberComments_ = _comments;
                isIndirectMemberComments_.trim();
              }
            }
            else if (HAS_VIRTUAL_MEMBER.equals(_linkType) ) {
              if (hasVirtualMemberComments_.indexOf(_comments) == -1) {
                if (hasVirtualMemberComments_.length() != 0)
                    hasVirtualMemberComments_ = hasVirtualMemberComments_ + " " +_comments;
                else
                    hasVirtualMemberComments_ = _comments;
                hasVirtualMemberComments_.trim();
              }
            }
            else if (DATA_EXCHANGE.equals(_linkType) ) {
              if (dataExchangeComments_.indexOf(_comments) == -1) {
                if (dataExchangeComments_.length() != 0)
                    dataExchangeComments_ = dataExchangeComments_ + " " +_comments;
                else
                    dataExchangeComments_ = _comments;
                dataExchangeComments_.trim();
              }
            }
        } // if (_linkType.length() > 11)
        else {
            if (IS_MEMBER.equals(_linkType) ) {
              if (isMemberComments_.indexOf(_comments) == -1) {
                if (isMemberComments_.length() != 0)
                    isMemberComments_ = isMemberComments_ + " " +_comments;
                else
                    isMemberComments_ = _comments;
                isMemberComments_.trim();
              }
            }
            else if (HAS_CHILD.equals(_linkType) ) {
              if (hasChildComments_.indexOf(_comments) == -1) {
                if (hasChildComments_.length() != 0)
                    hasChildComments_ = hasChildComments_ + " " +_comments;
                else
                    hasChildComments_ = _comments;
                hasChildComments_.trim();
              }
            }
            else if (RUN_UNDER.equals(_linkType) ) {
              if (runUnderComments_.indexOf(_comments) == -1) {
                if (runUnderComments_.length() != 0)
                    runUnderComments_ = runUnderComments_ + " " +_comments;
                else
                    runUnderComments_ = _comments;
                runUnderComments_.trim();
              }
            }
            else if (SOFT_RUN_UNDER.equals(_linkType) ) {
              if (softRunUnderComments_.indexOf(_comments) == -1) {
                if (softRunUnderComments_.length() != 0)
                    softRunUnderComments_ = softRunUnderComments_ + " " +_comments;
                else
                    softRunUnderComments_ = _comments;
                softRunUnderComments_.trim();
              }
            }
        } // else
    } // if (_linkType.length() > 6)
    else {
        if (_linkType.length() > 5) {
            if (GLOBAL.equals(_linkType) ) {
              if (globalComments_.indexOf(_comments) == -1) {
                if (globalComments_.length() != 0)
                    globalComments_ = globalComments_ + " " +_comments;
                else
                    globalComments_ = _comments;
                globalComments_.trim();
              }

                          // for testing and tracking
                          //System.out.println("Call to AccessControlLinkImpl.addComment()"+
                            //      "for GLOBAL"+
                              //    "- result: "+globalComments_);

            }
            else if (BRIDGE.equals(_linkType) ) {
              if (bridgeComments_.indexOf(_comments) == -1) {
                if (bridgeComments_.length() != 0)
                    bridgeComments_ = bridgeComments_ + " " +_comments;
                else
                    bridgeComments_ = _comments;
                bridgeComments_.trim();
              }
            }
            else if (SPREAD.equals(_linkType) ) {
              if (spreadComments_.indexOf(_comments) == -1) {
                if (spreadComments_.length() != 0)
                    spreadComments_ = spreadComments_ + " " +_comments;
                else
                    spreadComments_ = _comments;
                spreadComments_.trim();
              }
            }
            else if (LAUNCH.equals(_linkType) ) {
              if (launchComments_.indexOf(_comments) == -1) {
                if (launchComments_.length() != 0)
                    launchComments_ = launchComments_ + " " +_comments;
                else
                    launchComments_ = _comments;
                launchComments_.trim();
              }
            }
        } // if (_linkType.length() > 5)
        else {
            if (ACL.equals(_linkType) ) {
              if (aclComments_.indexOf(_comments) == -1) {
                if (aclComments_.length() != 0)
                    aclComments_ = aclComments_ + " " +_comments;
                else
                    aclComments_ = _comments;
                aclComments_.trim();
              }
            }
            else if (OWNER.equals(_linkType) ) {
              if (ownerComments_.indexOf(_comments) == -1) {
                if (ownerComments_.length() != 0)
                    ownerComments_ = ownerComments_ + " " +_comments;
                else
                    ownerComments_ = _comments;
                ownerComments_.trim();
              }
            }
            else if (TPRI.equals(_linkType) ) {
              if (typedPrivComments_.indexOf(_comments) == -1) {
                if (typedPrivComments_.length() != 0)
                    typedPrivComments_ = typedPrivComments_ + " " +_comments;
                else
                    typedPrivComments_ = _comments;
                typedPrivComments_.trim();
              }
            }
            else if (LPRI.equals(_linkType) ) {
              if (linkedPrivComments_.indexOf(_comments) == -1) {
                if (linkedPrivComments_.length() != 0)
                    linkedPrivComments_ = linkedPrivComments_ + " " +_comments;
                else
                    linkedPrivComments_ = _comments;
                linkedPrivComments_.trim();
              }
            }
        } // else if (_linkType.length() > 5)
    } // else if (_linkType.length() > 6)
  } // addLinkTypeComments()

   /**
   * Removes a right associated to one link type if it exists,
   * and desactivates the link type if it is the last right.
   * No action if _linkType is not a link type,
   * if _right is null, empty or unknown.
   * Not optimized to be a fast method.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink
   * @param _right to remove
   */
  public void removeRight(String _linkType, StringRight _right) {

    if (_right == null || _right.isEmpty()) return;

          // for testing and tracking
          //System.out.println("Call to AccessControlLinkImpl.removeRights( , )");
    
    StringRight[] l_rights = getRights(_linkType);

    if (l_rights.length == 0 || !Arrays.asList(l_rights).contains(_right)) {
        l_rights = null;
        return;
    }
    else if (l_rights.length == 1) {
        setRights(_linkType, null);
        l_rights = null;
        return;
    }

    StringRight[]   l_newRights = new StringRight[l_rights.length - 1];
    int             shiftingBack = 0;

    for (int i = 0; i < l_rights.length; i++) {
        if (!l_rights[i].equals(_right))
            l_newRights[i - shiftingBack] = l_rights[i];
        else
            shiftingBack++;
    } // for i

    setRights(_linkType, l_newRights);
    l_rights = null;
    l_newRights = null;
  } // removeRight()

   /**
   * Removes all the instances of the comments associated to one link type if they exist.
   * The link type state is not controlled, and it is never changed.
   * Not optimized to be a fast method.
   * @see ARoad0.CNot.AccessControlLink for link types
   * @param _linkType is a link type in AccessControlLink;
   *   No action if _linkType is not a link type, if _commenst is unknown.
   * @param _comments to remove one or several times; no action if null or empty.
   */
  public void removeComment(String _linkType, String _comments) {
    if (_comments == null || _comments.length() == 0) return;

    // variables
    int index;
    String com = getLinkTypeComments(_linkType);

    if (com.indexOf(_comments) == -1) {
        com = null;
        return;
    }

    while (com.indexOf(_comments) != -1) {
        index = com.lastIndexOf(_comments);
        com = com.substring(0, index) +
          com.substring(index+ _comments.length(), com.length());
        com.trim();
    } // while

    setLinkTypeComments(_linkType, com);
    com = null;
  } // removeComment()

   /**
   * Adds all the argument rights and comments in the rights and comments of this instance,
   * for each activated type in the argument.
   * If the link type is desactivated in the argument, there is no adding of rights nor comments.
   * If the link type value is activated and the link type rights are empty,
   * there is generally no adding of comments, but the exception is for BRIDGE,
   * where the comments are then added.
   * Optimized to be a fast method.
   * Called by DisplayableLinkImpl.mergeRightsWithoutFiltering() and
   * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
   * @param _acLink to merge
   * @return boolean true if one _acLink right has been added to (comment addings are not
   *    taken in account)
   */
  public boolean mergeAccessControlLinkWithoutFiltering(AccessControlLinkImpl _acLink) {
    boolean         result = false;
    StringRight[]   l_rights;
    Set             l_ri = new HashSet(INITIAL_CAPACITY);


    if (_acLink.contain_ || _acLink.negAcl_ || _acLink.negTypedPriv_ ||
            _acLink.negLinkedPriv_ || _acLink.isAlias_) {
        if (_acLink.contain_) {
            l_rights = l_containRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_containRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              contain_ = true;
              l_containRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.contain_) contain_ = true;
              else contain_ = false;
              l_containRights_ = new StringRight[0];
            }

            if (contain_ && containComments_.indexOf(_acLink.containComments_) == -1) {
                containComments_ = containComments_ + " " +_acLink.containComments_;
                containComments_.trim();
            }
        } // if contain_

        if (_acLink.negAcl_) {
            l_rights = l_negAclRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_negAclRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              negAcl_ = true;
              l_negAclRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.negAcl_) negAcl_ = true;
              else negAcl_ = false;
              l_negAclRights_ = new StringRight[0];
            }

            if (negAcl_ && negAclComments_.indexOf(_acLink.negAclComments_) == -1) {
                negAclComments_ = negAclComments_ + " " +_acLink.negAclComments_;
                negAclComments_.trim();
            }
        } // if negAcl_

        if (_acLink.negTypedPriv_) {
            l_rights = l_negTypedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_negTypedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              negTypedPriv_ = true;
              l_negTypedPrivRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.negTypedPriv_) negTypedPriv_ = true;
              else negTypedPriv_ = false;
              l_negTypedPrivRights_ = new StringRight[0];
            }

            if (negTypedPriv_ && negTypedPrivComments_.indexOf(_acLink.negTypedPrivComments_) == -1) {
                negTypedPrivComments_ = negTypedPrivComments_ + " " +_acLink.negTypedPrivComments_;
                negTypedPrivComments_.trim();
            }
        } // if negTypedPriv_

        if (_acLink.negLinkedPriv_) {
            l_rights = l_negLinkedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_negLinkedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              negLinkedPriv_ = true;
              l_negLinkedPrivRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.negLinkedPriv_) negLinkedPriv_ = true;
              else negLinkedPriv_ = false;
              l_negLinkedPrivRights_ = new StringRight[0];
            }

            if (negLinkedPriv_ && negLinkedPrivComments_.indexOf(_acLink.negLinkedPrivComments_) == -1) {
                negLinkedPrivComments_ = negLinkedPrivComments_ + " " +_acLink.negLinkedPrivComments_;
                negLinkedPrivComments_.trim();
            }
        } // if negLinkedPriv_

        if (_acLink.isAlias_) {
            l_rights = l_isAliasRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_isAliasRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              isAlias_ = true;
              l_isAliasRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.isAlias_) isAlias_ = true;
              else isAlias_ = false;
              l_isAliasRights_ = new StringRight[0];
            }

            if (isAlias_ && isAliasComments_.indexOf(_acLink.isAliasComments_) == -1) {
                isAliasComments_ = isAliasComments_ + " " +_acLink.isAliasComments_;
                isAliasComments_.trim();
            }
        } // if isAlias_
    } // if (_acLink.contain_

    if (_acLink.isIndirectMember_ || _acLink.hasVirtualMember_ ||
            _acLink.dataExchange_) {

        if (_acLink.isIndirectMember_) {
            l_rights = l_isIndirectMemberRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_isIndirectMemberRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              isIndirectMember_ = true;
              l_isIndirectMemberRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.isIndirectMember_) isIndirectMember_ = true;
              else isIndirectMember_ = false;
              l_isIndirectMemberRights_ = new StringRight[0];
            }

            if (isIndirectMember_ && isIndirectMemberComments_.indexOf(_acLink.isIndirectMemberComments_) == -1) {
                isIndirectMemberComments_ = isIndirectMemberComments_ + " " +_acLink.isIndirectMemberComments_;
                isIndirectMemberComments_.trim();
            }
        } // if isIndirectMember_

        if (_acLink.hasVirtualMember_) {
            l_rights = l_hasVirtualMemberRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_hasVirtualMemberRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              hasVirtualMember_ = true;
              l_hasVirtualMemberRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.hasVirtualMember_) hasVirtualMember_ = true;
              else hasVirtualMember_ = false;
              l_hasVirtualMemberRights_ = new StringRight[0];
            }

            if (hasVirtualMember_ &&
                    hasVirtualMemberComments_.indexOf(_acLink.hasVirtualMemberComments_) == -1) {
                hasVirtualMemberComments_ = hasVirtualMemberComments_ + " " +
                        _acLink.hasVirtualMemberComments_;
                hasVirtualMemberComments_.trim();
            }
        } // if hasVirtualMember_

        if (_acLink.dataExchange_) {
            l_rights = l_dataExchangeRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_dataExchangeRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              dataExchange_ = true;
              l_dataExchangeRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.dataExchange_) dataExchange_ = true;
              else dataExchange_ = false;
              l_dataExchangeRights_ = new StringRight[0];
            }

            if (dataExchange_ && dataExchangeComments_.indexOf(_acLink.dataExchangeComments_) == -1) {
                dataExchangeComments_ = dataExchangeComments_ + " " +_acLink.dataExchangeComments_;
                dataExchangeComments_.trim();
            }
        } // if dataExchange_

    } // if (_acLink.isIndirectMember_

    if (_acLink.isMember_ || _acLink.hasChild_ || _acLink.runUnder_ || _acLink.softRunUnder_) {

        if (_acLink.isMember_) {
            l_rights = l_isMemberRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_isMemberRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              isMember_ = true;
              l_isMemberRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.isMember_) isMember_ = true;
              else isMember_ = false;
              l_isMemberRights_ = new StringRight[0];
            }

            if (isMember_ && isMemberComments_.indexOf(_acLink.isMemberComments_) == -1) {
                isMemberComments_ = isMemberComments_ + " " +_acLink.isMemberComments_;
                isMemberComments_.trim();
            }
        } // if isMember_

        if (_acLink.hasChild_) {
            l_rights = l_hasChildRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_hasChildRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              hasChild_ = true;
              l_hasChildRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.hasChild_) hasChild_ = true;
              else hasChild_ = false;
              l_hasChildRights_ = new StringRight[0];
            }

            if (hasChild_ && hasChildComments_.indexOf(_acLink.hasChildComments_) == -1) {
                hasChildComments_ = hasChildComments_ + " " +_acLink.hasChildComments_;
                hasChildComments_.trim();
            }
        } // if hasChild_

        if (_acLink.runUnder_) {
            l_rights = l_runUnderRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_runUnderRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              runUnder_ = true;
              l_runUnderRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.runUnder_) runUnder_ = true;
              else runUnder_ = false;
              l_runUnderRights_ = new StringRight[0];
            }

            if (runUnder_ && runUnderComments_.indexOf(_acLink.runUnderComments_) == -1) {
                runUnderComments_ = runUnderComments_ + " " +_acLink.runUnderComments_;
                runUnderComments_.trim();
            }
        } // if runUnder_

        if (_acLink.softRunUnder_) {
            l_rights = l_softRunUnderRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_softRunUnderRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              softRunUnder_ = true;
              l_softRunUnderRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.softRunUnder_) softRunUnder_ = true;
              else softRunUnder_ = false;
              l_softRunUnderRights_ = new StringRight[0];
            }

            if (softRunUnder_ && softRunUnderComments_.indexOf(_acLink.softRunUnderComments_) == -1) {
                softRunUnderComments_ = softRunUnderComments_ + " " +_acLink.softRunUnderComments_;
                softRunUnderComments_.trim();
            }
        } // if softRunUnder_

    } // if (_acLink.isMember_

    if (_acLink.global_ || _acLink.bridge_ || _acLink.spread_ || _acLink.launch_) {
        if (_acLink.global_) {
            // merges the _acLink rights in the current rights if they do not exist
            l_rights = l_globalRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_globalRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;

                      // for testing and tracking
                      //System.out.println("in AccessControlLinkImpl."+
                        //"mergeAccesControlLinkWithoutFiltering() for GLOBAL in "+this+
                        //", add the right: "+l_acLink_rights[j]);

                } // for j
            } // if

            if (l_ri.size() != 0) {
              global_ = true;
              l_globalRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.global_) global_ = true;
              else global_ = false;
              l_globalRights_ = new StringRight[0];
            }

            if (global_ && globalComments_.indexOf(_acLink.globalComments_) == -1) {
                globalComments_ = globalComments_ + " " +_acLink.globalComments_;
                globalComments_.trim();
            }
        } // if global_

        if (_acLink.bridge_) {
            // merges the _acLink rights in the current rights if they do not exist
            l_rights = l_bridgeRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_bridgeRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              bridge_ = true;
              l_bridgeRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.bridge_) bridge_ = true;
              else bridge_ = false;
              l_bridgeRights_ = new StringRight[0];
            }

            if (bridgeComments_.indexOf(_acLink.bridgeComments_) == -1) {
                bridgeComments_ = bridgeComments_ + " " +_acLink.bridgeComments_;
                bridgeComments_.trim();
            }
        } // if bridge_

        if (_acLink.spread_) {

            if (spreadComments_.indexOf(_acLink.spreadComments_) == -1) {
                spreadComments_ = spreadComments_ + " " +_acLink.spreadComments_;
                spreadComments_.trim();
            }
        } // if spread_

        if (_acLink.launch_) {
            // merges the _acLink rights in the current rights if they do not exist
            l_rights = l_launchRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_launchRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              launch_ = true;
              l_launchRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.launch_) launch_ = true;
              else launch_ = false;
              l_launchRights_ = new StringRight[0];
            }

            if (launchComments_.indexOf(_acLink.launchComments_) == -1) {
                launchComments_ = launchComments_ + " " +_acLink.launchComments_;
                launchComments_.trim();
            }
        } // if launch_

    } // if (_acLink.global_

    if (_acLink.acl_ || _acLink.owner_ || _acLink.typedPriv_ || _acLink.linkedPriv_) {
        if (_acLink.acl_) {
            l_rights = l_aclRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_aclRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              acl_ = true;
              l_aclRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.acl_) acl_ = true;
              else acl_ = false;
              l_aclRights_ = new StringRight[0];
            }

            if (acl_ && aclComments_.indexOf(_acLink.aclComments_) == -1) {
                aclComments_ = aclComments_ + " " +_acLink.aclComments_;
                aclComments_.trim();
            }
        } // if acl_

        if (_acLink.owner_) {
            l_rights = l_ownerRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_ownerRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              owner_ = true;
              l_ownerRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.owner_) owner_ = true;
              else owner_ = false;
              l_ownerRights_ = new StringRight[0];
            }

            if (owner_ && ownerComments_.indexOf(_acLink.ownerComments_) == -1) {
                ownerComments_ = ownerComments_ + " " +_acLink.ownerComments_;
                ownerComments_.trim();
            }
        } // if owner_

        if (_acLink.typedPriv_) {
            l_rights = l_typedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_typedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              typedPriv_ = true;
              l_typedPrivRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.typedPriv_) typedPriv_ = true;
              else typedPriv_ = false;
              l_typedPrivRights_ = new StringRight[0];
            }

            if (typedPriv_ && typedPrivComments_.indexOf(_acLink.typedPrivComments_) == -1) {
                typedPrivComments_ = typedPrivComments_ + " " +_acLink.typedPrivComments_;
                typedPrivComments_.trim();
            }
        } // if typedPriv_

        if (_acLink.linkedPriv_) {
            l_rights = l_linkedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                } // for j
            } // if
            l_rights = _acLink.l_linkedPrivRights_;
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                  l_ri.add(l_rights[j]);
                  result = true;
                } // for j
            } // if

            if (l_ri.size() != 0) {
              linkedPriv_ = true;
              l_linkedPrivRights_ = (StringRight[])l_ri.toArray(new StringRight[l_ri.size()]);
              l_ri.clear();
            }
            else {
              if (_acLink.linkedPriv_) linkedPriv_ = true;
              else linkedPriv_ = false;
              l_linkedPrivRights_ = new StringRight[0];
            }

            if (linkedPriv_ && linkedPrivComments_.indexOf(_acLink.linkedPrivComments_) == -1) {
                linkedPrivComments_ = linkedPrivComments_ + " " +_acLink.linkedPrivComments_;
                linkedPrivComments_.trim();
            }
        } // if linkedPriv_

    } // if (_acLink.acl_
    
    return result;
  } // mergeAccessControlLinkWithoutFiltering()

   /**
   * Adds all the argument comments to this instance comments,
   * whatever the link type value in the argument and this instance.
   * Optimized to be a fast method.
   * Called by DisplayableLinkImpl.mergeCommentsWithoutFiltering().
   * @param _acLink a link with comments to put in this link
   */
  public void mergeCommentsWithoutFiltering(AccessControlLinkImpl _acLink) {
    if (_acLink.contain_ || _acLink.negAcl_ || _acLink.negTypedPriv_ ||
            _acLink.negLinkedPriv_ || _acLink.isAlias_) {

            if (containComments_.indexOf(_acLink.containComments_) == -1) {
                containComments_ = containComments_ + " " +_acLink.containComments_;
                containComments_.trim();
            }

            if (negAclComments_.indexOf(_acLink.negAclComments_) == -1) {
                negAclComments_ = negAclComments_ + " " +_acLink.negAclComments_;
                negAclComments_.trim();
            }

            if (negTypedPrivComments_.indexOf(_acLink.negTypedPrivComments_) == -1) {
                negTypedPrivComments_ = negTypedPrivComments_ + " " +_acLink.negTypedPrivComments_;
                negTypedPrivComments_.trim();
            }

            if (negLinkedPrivComments_.indexOf(_acLink.negLinkedPrivComments_) == -1) {
                negLinkedPrivComments_ = negLinkedPrivComments_ + " " +_acLink.negLinkedPrivComments_;
                negLinkedPrivComments_.trim();
            }

            if (isAliasComments_.indexOf(_acLink.isAliasComments_) == -1) {
                isAliasComments_ = isAliasComments_ + " " +_acLink.isAliasComments_;
                isAliasComments_.trim();
            }
    } // if (_acLink.contain_

    if (_acLink.isIndirectMember_ ||  _acLink.hasVirtualMember_ ||
            _acLink.dataExchange_ ) {

            if (isIndirectMemberComments_.indexOf(_acLink.isIndirectMemberComments_) == -1) {
                isIndirectMemberComments_ = isIndirectMemberComments_ + " " +_acLink.isIndirectMemberComments_;
                isIndirectMemberComments_.trim();
            }

            if (hasVirtualMemberComments_.indexOf(_acLink.hasVirtualMemberComments_) == -1) {
                hasVirtualMemberComments_ = hasVirtualMemberComments_ + " " +
                        _acLink.hasVirtualMemberComments_;
                hasVirtualMemberComments_.trim();
            }

            if (dataExchangeComments_.indexOf(_acLink.dataExchangeComments_) == -1) {
                dataExchangeComments_ = dataExchangeComments_ + " " +_acLink.dataExchangeComments_;
                dataExchangeComments_.trim();
            }

    } // if (_acLink.isIndirectMember_

    if (_acLink.isMember_ || _acLink.hasChild_ || _acLink.runUnder_ || _acLink.softRunUnder_) {

            if (isMemberComments_.indexOf(_acLink.isMemberComments_) == -1) {
                isMemberComments_ = isMemberComments_ + " " +_acLink.isMemberComments_;
                isMemberComments_.trim();
            }

            if (hasChildComments_.indexOf(_acLink.hasChildComments_) == -1) {
                hasChildComments_ = hasChildComments_ + " " +_acLink.hasChildComments_;
                hasChildComments_.trim();
            }

            if (runUnderComments_.indexOf(_acLink.runUnderComments_) == -1) {
                runUnderComments_ = runUnderComments_ + " " +_acLink.runUnderComments_;
                runUnderComments_.trim();
            }

            if (softRunUnderComments_.indexOf(_acLink.softRunUnderComments_) == -1) {
                softRunUnderComments_ = softRunUnderComments_ + " " +_acLink.softRunUnderComments_;
                softRunUnderComments_.trim();
            }
    } // if (_acLink.isMember_

    if (_acLink.global_ || _acLink.bridge_ || _acLink.spread_ || _acLink.launch_) {

            if (globalComments_.indexOf(_acLink.globalComments_) == -1) {
                globalComments_ = globalComments_ + " " +_acLink.globalComments_;
                globalComments_.trim();
            }

            if (bridgeComments_.indexOf(_acLink.bridgeComments_) == -1) {
                bridgeComments_ = bridgeComments_ + " " +_acLink.bridgeComments_;
                bridgeComments_.trim();
            }

            if (spreadComments_.indexOf(_acLink.spreadComments_) == -1) {
                spreadComments_ = spreadComments_ + " " +_acLink.spreadComments_;
                spreadComments_.trim();
            }

            if (launchComments_.indexOf(_acLink.launchComments_) == -1) {
                launchComments_ = launchComments_ + " " +_acLink.launchComments_;
                launchComments_.trim();
            }

    } // if (_acLink.global_

    if (_acLink.acl_ || _acLink.owner_ || _acLink.typedPriv_ || _acLink.linkedPriv_) {

            if (aclComments_.indexOf(_acLink.aclComments_) == -1) {
                aclComments_ = aclComments_ + " " +_acLink.aclComments_;
                aclComments_.trim();
            }

            if (ownerComments_.indexOf(_acLink.ownerComments_) == -1) {
                ownerComments_ = ownerComments_ + " " +_acLink.ownerComments_;
                ownerComments_.trim();
            }

            if (typedPrivComments_.indexOf(_acLink.typedPrivComments_) == -1) {
                typedPrivComments_ = typedPrivComments_ + " " +_acLink.typedPrivComments_;
                typedPrivComments_.trim();
            }

            if (linkedPrivComments_.indexOf(_acLink.linkedPrivComments_) == -1) {
                linkedPrivComments_ = linkedPrivComments_ + " " +_acLink.linkedPrivComments_;
                linkedPrivComments_.trim();
            }

    } // if (_acLink.acl_
  } // mergeCommentsWithoutFiltering

   /**
   * Sets all the link types with the _acLink link types.
   * @param _acLink is an AccessControlLinkImpl to copy
   */
  public void setLinkTypesWithoutFiltering(AccessControlLinkImpl _acLink) {

    isMember_ = _acLink.isMember_;
    hasChild_ = _acLink.hasChild_;
    isIndirectMember_ = _acLink.isIndirectMember_;
    hasVirtualMember_ = _acLink.hasVirtualMember_;
    dataExchange_ = _acLink.dataExchange_;
    runUnder_ = _acLink.runUnder_;
    softRunUnder_ = _acLink.softRunUnder_;

    contain_ = _acLink.contain_;
    negAcl_ = _acLink.negAcl_;
    negTypedPriv_ = _acLink.negTypedPriv_;
    negLinkedPriv_ = _acLink.negLinkedPriv_;
    isAlias_ = _acLink.isAlias_;

    global_ = _acLink.global_;
    bridge_ = _acLink.bridge_;
    spread_ = _acLink.spread_;
    launch_ = _acLink.launch_;

    acl_ = _acLink.acl_;
    owner_ = _acLink.owner_;
    typedPriv_ = _acLink.typedPriv_;
    linkedPriv_ = _acLink.linkedPriv_;

  } // setLinkTypesWithoutFiltering

  /**
   * True if this is the first link of a DisplayableLink from a GroupID to a Resource,
   * to say that the effective rights depend on the AGO rights Actor which uses
   * the link to this GroupID, and so this is a weak DisplayableLink.
   * A link which is based on non-AGO rights is never weak.
   * Note: the link may be weak through an AGO other relation from a group,
   * and then there is no contain relation.
   * @return true if the contain or the global comments include 
   * AccessControlLink.WEAK_LINK, even if the type is not activated.
   */
  public boolean isWeak() {
    if ( containComments_.indexOf(AccessControlLink.C_WEAK_LINK) != -1 ||
        globalComments_.indexOf(AccessControlLink.C_WEAK_LINK) != -1
      ) return true;
    else return false;
  }
  
  /**
   * Processes the link type values only, not the comments nor
   *  the rights.
   * @return true if all link type values are desactivated
   */
  public boolean isEmpty() {
    if ( 
        !isMember_ &&
        !hasChild_ &&
        !isIndirectMember_ &&
        !hasVirtualMember_ &&
        !dataExchange_ &&
        !runUnder_ &&
        !softRunUnder_ &&

        !contain_ &&
        !negAcl_ &&
        !negTypedPriv_ &&
        !negLinkedPriv_&&
        !isAlias_ &&

        !global_ &&
        !bridge_ &&
        !spread_ &&
        !launch_ &&

        !acl_ &&
        !owner_ &&
        !typedPriv_ &&
        !linkedPriv_
      ) return true;
    else return false;
  } // isEmpty()

  /**
   * Gets a clone of the link. Called by DisplayableLinkImpl.valueOf().
   * @exception CloneNotSupportedException
   * @return a cloned and no-backed instance.
   */
  public Object clone() {
    try {
      AccessControlLinkImpl clon = (AccessControlLinkImpl) super.clone();

      clon.isMember_ = isMember_;
      clon.hasChild_ = hasChild_;
      clon.isIndirectMember_ = isIndirectMember_;
      clon.hasVirtualMember_ = hasVirtualMember_;
      clon.dataExchange_ = dataExchange_;
      clon.runUnder_ = runUnder_;
      clon.softRunUnder_ = softRunUnder_;

      clon.contain_ = contain_;
      clon.negAcl_ = negAcl_;
      clon.negTypedPriv_ = negTypedPriv_;
      clon.negLinkedPriv_ = negLinkedPriv_;
      clon.isAlias_ = isAlias_;

      clon.global_ = global_;
      clon.bridge_ = bridge_;
      clon.spread_ = spread_;
      clon.launch_ = launch_;

      clon.acl_ = acl_;
      clon.owner_ = owner_;
      clon.typedPriv_ = typedPriv_;
      clon.linkedPriv_ = linkedPriv_;


      if (l_isMemberRights_ != null) {
          clon.l_isMemberRights_ = new StringRight[l_isMemberRights_.length];
          System.arraycopy(
                  l_isMemberRights_, 0, clon.l_isMemberRights_, 0, l_isMemberRights_.length);
      }

      if (l_hasChildRights_ != null) {
          clon.l_hasChildRights_ = new StringRight[l_hasChildRights_.length];
          System.arraycopy(
                  l_hasChildRights_, 0, clon.l_hasChildRights_, 0, l_hasChildRights_.length);
      }

      if (l_isIndirectMemberRights_ != null) {
          clon.l_isIndirectMemberRights_ = new StringRight[l_isIndirectMemberRights_.length];
          System.arraycopy(
                  l_isIndirectMemberRights_, 0, clon.l_isIndirectMemberRights_, 0, l_isIndirectMemberRights_.length);
      }

      if (l_hasVirtualMemberRights_ != null) {
          clon.l_hasVirtualMemberRights_ = new StringRight[l_hasVirtualMemberRights_.length];
          System.arraycopy(
                  l_hasVirtualMemberRights_, 0, clon.l_hasVirtualMemberRights_, 0, l_hasVirtualMemberRights_.length);
      }

      if (l_dataExchangeRights_ != null) {
          clon.l_dataExchangeRights_ = new StringRight[l_dataExchangeRights_.length];
          System.arraycopy(
                  l_dataExchangeRights_, 0, clon.l_dataExchangeRights_, 0, l_dataExchangeRights_.length);
      }

      if (l_runUnderRights_ != null) {
          clon.l_runUnderRights_ = new StringRight[l_runUnderRights_.length];
          System.arraycopy(
                  l_runUnderRights_, 0, clon.l_runUnderRights_, 0, l_runUnderRights_.length);
      }

      if (l_softRunUnderRights_ != null) {
          clon.l_softRunUnderRights_ = new StringRight[l_softRunUnderRights_.length];
          System.arraycopy(
                  l_softRunUnderRights_, 0, clon.l_softRunUnderRights_, 0, l_softRunUnderRights_.length);
      }

      if (l_containRights_ != null) {
          clon.l_containRights_ = new StringRight[l_containRights_.length ];
          System.arraycopy(
                  l_containRights_, 0, clon.l_containRights_, 0, l_containRights_.length);
      }

      if (l_negAclRights_ != null) {
          clon.l_negAclRights_ = new StringRight[l_negAclRights_.length];
          System.arraycopy(
                  l_negAclRights_, 0, clon.l_negAclRights_, 0, l_negAclRights_.length);
      }

      if (l_negTypedPrivRights_ != null) {
          clon.l_negTypedPrivRights_ = new StringRight[l_negTypedPrivRights_.length ];
          System.arraycopy(
                  l_negTypedPrivRights_, 0, clon.l_negTypedPrivRights_, 0, l_negTypedPrivRights_.length);
      }

      if (l_negLinkedPrivRights_ != null) {
          clon.l_negLinkedPrivRights_ = new StringRight[l_negLinkedPrivRights_.length];
          System.arraycopy(
                  l_negLinkedPrivRights_, 0, clon.l_negLinkedPrivRights_, 0, l_negLinkedPrivRights_.length);
      }

      if (l_isAliasRights_ != null) {
          clon.l_isAliasRights_ = new StringRight[l_isAliasRights_.length ];
          System.arraycopy(
                  l_isAliasRights_, 0, clon.l_isAliasRights_, 0, l_isAliasRights_.length);
      }

      if (l_globalRights_ != null) {
          clon.l_globalRights_ = new StringRight[l_globalRights_.length];
          System.arraycopy(
                  l_globalRights_, 0, clon.l_globalRights_, 0, l_globalRights_.length);
      }

      if (l_bridgeRights_ != null) {
          clon.l_bridgeRights_ = new StringRight[l_bridgeRights_.length];
          System.arraycopy(
                  l_bridgeRights_, 0, clon.l_bridgeRights_, 0, l_bridgeRights_.length);
      }

      if (l_launchRights_ != null) {
          clon.l_launchRights_ = new StringRight[l_launchRights_.length];
          System.arraycopy(
                  l_launchRights_, 0, clon.l_launchRights_, 0, l_launchRights_.length);
      }

      if (l_aclRights_ != null) {
          clon.l_aclRights_ = new StringRight[l_aclRights_.length];
          System.arraycopy(
                  l_aclRights_, 0, clon.l_aclRights_, 0, l_aclRights_.length);
      }

      if (l_ownerRights_ != null) {
          clon.l_ownerRights_ = new StringRight[l_ownerRights_.length];
          System.arraycopy(
                  l_ownerRights_, 0, clon.l_ownerRights_, 0, l_ownerRights_.length);
      }

      if (l_typedPrivRights_ != null) {
          clon.l_typedPrivRights_ = new StringRight[l_typedPrivRights_.length];
          System.arraycopy(
                  l_typedPrivRights_, 0, clon.l_typedPrivRights_, 0, l_typedPrivRights_.length);
      }

      if (l_linkedPrivRights_ != null) {
          clon.l_linkedPrivRights_ = new StringRight[l_linkedPrivRights_.length];
          System.arraycopy(
                  l_linkedPrivRights_, 0, clon.l_linkedPrivRights_, 0, l_linkedPrivRights_.length);
      }

      clon.isMemberComments_ = isMemberComments_;
      clon.hasChildComments_ = hasChildComments_;
      clon.isIndirectMemberComments_ = isIndirectMemberComments_;
      clon.hasVirtualMemberComments_ = hasVirtualMemberComments_;
      clon.dataExchangeComments_ = dataExchangeComments_;
      clon.runUnderComments_ = runUnderComments_;
      clon.softRunUnderComments_ = softRunUnderComments_;

      clon.containComments_ = containComments_;
      clon.negAclComments_ = negAclComments_;
      clon.negTypedPrivComments_ = negTypedPrivComments_;
      clon.negLinkedPrivComments_ = negLinkedPrivComments_;
      clon.isAliasComments_ = isAliasComments_;

      clon.globalComments_ = globalComments_;
      clon.bridgeComments_ = bridgeComments_;
      clon.spreadComments_ = spreadComments_;
      clon.launchComments_ = launchComments_;

      clon.aclComments_ = aclComments_;
      clon.ownerComments_ = ownerComments_;
      clon.typedPrivComments_ = typedPrivComments_;
      clon.linkedPrivComments_ = linkedPrivComments_;

      return clon;
    } catch (CloneNotSupportedException e) {
      // this shouldn't happen, since we are Cloneable
      throw new InternalError(e.toString() + "AccessControlLinkImpl.clone() failed");
    } // catch
  } // clone()

    /**
     * The DisplayableLinkImpl constructor does not allow to add 2 objects
     * that are equals, then this method is not implemented.
   * @param _obj the object to compare
   * @return true if the lists of link type value, rights and comments are equals
   */
  /*
  public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      AccessControlLinkImpl obj = (AccessControlLinkImpl)_obj;

      if (isMember_ != obj.isMember_ ) return false;
      if (hasChild_ != obj.hasChild_ ) return false;
      if (isIndirectMember_ != obj.isIndirectMember_ ) return false;
      if (hasVirtualMember_ != obj.hasVirtualMember_ ) return false;
      if (dataExchange_ != obj.dataExchange_ ) return false;
      if (runUnder_ != obj.runUnder_ ) return false;
      if (softRunUnder_ != obj.softRunUnder_ ) return false;
   *
      if (contain_ != obj.contain_ ) return false;
      if (negAcl_ != obj.negAcl_ ) return false;
      if (negTypedPriv_ != obj.negTypedPriv_ ) return false;
      if (negLinkedPriv_ != obj.negLinkedPriv_ ) return false;
      if (isAlias_ != obj.isAlias_ ) return false;

      if (global_ != obj.global_ ) return false;
      if (bridge_ != obj.bridge_ ) return false;
      if (spread_ != obj.spread_ ) return false;
      if (launch_ != obj.launch_ ) return false;

      if (acl_ != obj.acl_ ) return false;
      if (owner_ != obj.owner_ ) return false;
      if (typedPriv_ != obj.typedPriv_ ) return false;
      if (linkedPriv_ != obj.linkedPriv_ ) return false;


      if (l_isMemberRights_ == null && obj.l_isMemberRights_ != null) return false;
      else if (l_isMemberRights_ != null && obj.l_isMemberRights_ == null) return false;
      else if (l_isMemberRights_ != null && obj.l_isMemberRights_ != null &&
              !l_isMemberRights_.equals(obj.l_isMemberRights_) )
          return false;
      if (l_hasChildRights_ == null && obj.l_hasChildRights_ != null) return false;
      else if (l_hasChildRights_ != null && obj.l_hasChildRights_ == null) return false;
      else if (l_hasChildRights_ != null && obj.l_hasChildRights_ != null &&
              !l_hasChildRights_.equals(obj.l_hasChildRights_) )
          return false;
      if (l_isIndirectMemberRights_ == null && obj.l_isIndirectMemberRights_ != null) return false;
      else if (l_isIndirectMemberRights_ != null && obj.l_isIndirectMemberRights_ == null) return false;
      else if (l_isIndirectMemberRights_ != null && obj.l_isIndirectMemberRights_ != null &&
              !l_isIndirectMemberRights_.equals(obj.l_isIndirectMemberRights_) ) 
          return false;
      if (l_hasVirtualMemberRights_ == null && obj.l_hasVirtualMemberRights_ != null) return false;
      else if (l_hasVirtualMemberRights_ != null && obj.l_hasVirtualMemberRights_ == null) return false;
      else if (l_hasVirtualMemberRights_ != null && obj.l_hasVirtualMemberRights_ != null &&
              !l_hasVirtualMemberRights_.equals(obj.l_hasVirtualMemberRights_) ) 
          return false;
      if (l_dataExchangeRights_ == null && obj.l_dataExchangeRights_ != null) return false;
      else if (l_dataExchangeRights_ != null && obj.l_dataExchangeRights_ == null) return false;
      else if (l_dataExchangeRights_ != null && obj.l_dataExchangeRights_ != null &&
              !l_dataExchangeRights_.equals(obj.l_dataExchangeRights_) )
          return false;
      if (l_runUnderRights_ == null && obj.l_runUnderRights_ != null) return false;
      else if (l_runUnderRights_ != null && obj.l_runUnderRights_ == null) return false;
      else if (l_runUnderRights_ != null && obj.l_runUnderRights_ != null &&
              !l_runUnderRights_.equals(obj.l_runUnderRights_) )
          return false;
      if (l_softRunUnderRights_ == null && obj.l_softRunUnderRights_ != null) return false;
      else if (l_softRunUnderRights_ != null && obj.l_softRunUnderRights_ == null) return false;
      else if (l_softRunUnderRights_ != null && obj.l_softRunUnderRights_ != null &&
              !l_softRunUnderRights_.equals(obj.l_softRunUnderRights_) )
          return false;
      if (l_containRights_ == null && obj.l_containRights_ != null) return false;
      else if (l_containRights_ != null && obj.l_containRights_ == null) return false;
      else if (l_containRights_ != null && obj.l_containRights_ != null &&
              !l_containRights_.equals(obj.l_containRights_) )
          return false;
      if (l_negAclRights_ == null && obj.l_negAclRights_ != null) return false;
      else if (l_negAclRights_ != null && obj.l_negAclRights_ == null) return false;
      else if (l_negAclRights_ != null && obj.l_negAclRights_ != null &&
              !l_negAclRights_.equals(obj.l_negAclRights_) ) 
          return false;
      if (l_negTypedPrivRights_ == null && obj.l_negTypedPrivRights_ != null) return false;
      else if (l_negTypedPrivRights_ != null && obj.l_negTypedPrivRights_ == null) return false;
      else if (l_negTypedPrivRights_ != null && obj.l_negTypedPrivRights_ != null &&
              !l_negTypedPrivRights_.equals(obj.l_negTypedPrivRights_) )
          return false;
      if (l_negLinkedPrivRights_ == null && obj.l_negLinkedPrivRights_ != null) return false;
      else if (l_negLinkedPrivRights_ != null && obj.l_negLinkedPrivRights_ == null) return false;
      else if (l_negLinkedPrivRights_ != null && obj.l_negLinkedPrivRights_ != null &&
              !l_negLinkedPrivRights_.equals(obj.l_negLinkedPrivRights_) ) 
          return false;
      if (l_isAliasRights_ == null && obj.l_isAliasRights_ != null) return false;
      else if (l_isAliasRights_ != null && obj.l_isAliasRights_ == null) return false;
      else if (l_isAliasRights_ != null && obj.l_isAliasRights_ != null &&
              !l_isAliasRights_.equals(obj.l_isAliasRights_) )
          return false;
      if (l_globalRights_ == null && obj.l_globalRights_ != null) return false;
      else if (l_globalRights_ != null && obj.l_globalRights_ == null) return false;
      else if (l_globalRights_ != null && obj.l_globalRights_ != null &&
              !l_globalRights_.equals(obj.l_globalRights_) ) 
          return false;
      if (l_bridgeRights_ == null && obj.l_bridgeRights_ != null) return false;
      else if (l_bridgeRights_ != null && obj.l_bridgeRights_ == null) return false;
      else if (l_bridgeRights_ != null && obj.l_bridgeRights_ != null &&
              !l_bridgeRights_.equals(obj.l_bridgeRights_) )
          return false;
      if (l_launchRights_ == null && obj.l_launchRights_ != null) return false;
      else if (l_launchRights_ != null && obj.l_launchRights_ == null) return false;
      else if (l_launchRights_ != null && obj.l_launchRights_ != null &&
              !l_launchRights_.equals(obj.l_launchRights_) ) 
          return false;
      if (l_aclRights_ == null && obj.l_aclRights_ != null) return false;
      else if (l_aclRights_ != null && obj.l_aclRights_ == null) return false;
      else if (l_aclRights_ != null && obj.l_aclRights_ != null &&
              !l_aclRights_.equals(obj.l_aclRights_) )
          return false;
      if (l_ownerRights_ == null && obj.l_ownerRights_ != null) return false;
      else if (l_ownerRights_ != null && obj.l_ownerRights_ == null) return false;
      else if (l_ownerRights_ != null && obj.l_ownerRights_ != null &&
              !l_ownerRights_.equals(obj.l_ownerRights_) ) 
          return false;
      if (l_typedPrivRights_ == null && obj.l_typedPrivRights_ != null) return false;
      else if (l_typedPrivRights_ != null && obj.l_typedPrivRights_ == null) return false;
      else if (l_typedPrivRights_ != null && obj.l_typedPrivRights_ != null &&
              !l_typedPrivRights_.equals(obj.l_typedPrivRights_) )
          return false;
      if (l_linkedPrivRights_ == null && obj.l_linkedPrivRights_ != null) return false;
      else if (l_linkedPrivRights_ != null && obj.l_linkedPrivRights_ == null) return false;
      else if (l_linkedPrivRights_ != null && obj.l_linkedPrivRights_ != null &&
              !l_linkedPrivRights_.equals(obj.l_linkedPrivRights_) )
          return false;


      if (!isMemberComments_.equals(obj.isMemberComments_) ) return false;
      if (!hasChildComments_.equals(obj.hasChildComments_) ) return false;
      if (!isIndirectMemberComments_.equals(obj.isIndirectMemberComments_) ) return false;
      if (!hasVirtualMemberComments_.equals(obj.hasVirtualMemberComments_) ) return false;
      if (!dataExchangeComments_.equals(obj.dataExchangeComments_) ) return false;
      if (!runUnderComments_.equals(obj.runUnderComments_) ) return false;
      if (!softRunUnderComments_.equals(obj.softRunUnderComments_) ) return false;

      if (!containComments_.equals(obj.containComments_) ) return false;
      if (!negAclComments_.equals(obj.negAclComments_) ) return false;
      if (!negTypedPrivComments_.equals(obj.negTypedPrivComments_) ) return false;
      if (!negLinkedPrivComments_.equals(obj.negLinkedPrivComments_) ) return false;
      if (!isAliasComments_.equals(obj.isAliasComments_) ) return false;

      if (!globalComments_.equals(obj.globalComments_) ) return false;
      if (!bridgeComments_.equals(obj.bridgeComments_) ) return false;
      if (!launchComments_.equals(obj.launchComments_) ) return false;

      if (!aclComments_.equals(obj.aclComments_) ) return false;
      if (!ownerComments_.equals(obj.ownerComments_) ) return false;
      if (!typedPrivComments_.equals(obj.typedPrivComments_) ) return false;
      if (!linkedPrivComments_.equals(obj.linkedPrivComments_) ) return false;
      
      return true;
  } // equals(Object)
   * *
   */

  /**
   * Gets some right hashcodes without including the rights, nor the comments.
   * Uses only 10 types upon 20.
   * @return the hashcode value for this instance. May be 3768901 if empty.
   */
  public int hashCode() {
    int result = 3768901;

    if (global_ && l_globalRights_ != null)
        result = PRIME*result + l_globalRights_.hashCode();
    
    if (bridge_) result = PRIME*result + l_bridgeRights_.hashCode();
    if (acl_) result = PRIME*result + l_aclRights_.hashCode();
    if (negAcl_) result = PRIME*result + l_negAclRights_.hashCode();
    if (owner_) result = PRIME*result + l_ownerRights_.hashCode();
    if (contain_) result = PRIME*result + l_containRights_.hashCode();
    if (isMember_) result = PRIME*result + l_isMemberRights_.hashCode();
    if (hasChild_) result = PRIME*result + l_hasChildRights_.hashCode();
    if (typedPriv_) result = PRIME*result + l_typedPrivRights_.hashCode();
    if (linkedPriv_) result = PRIME*result + l_linkedPrivRights_.hashCode();

    return result;
  } // hashcode()

  /**
   * Sets the rights to null.
   */
  public synchronized void finalizeForGui() {

          // for testing and tracking
          //System.out.println("Call of AccessControlLinkImpl.finalizeForGui()");

      isMember_ = false;
      hasChild_ = false;
      isIndirectMember_ = false;
      hasVirtualMember_ = false;
      dataExchange_ = false;
      runUnder_ = false;
      softRunUnder_ = false;

      contain_ = false;
      negAcl_ = false;
      negTypedPriv_ = false;
      negLinkedPriv_ = false;
      isAlias_ = false;

      global_ = false;
      bridge_ = false;
      spread_ = false;
      launch_ = false;

      acl_ = false;
      owner_ = false;
      typedPriv_ = false;
      linkedPriv_ = false;

      l_isMemberRights_ = null;
      l_hasChildRights_ = null;
      l_isIndirectMemberRights_ = null;
      l_hasVirtualMemberRights_ = null;
      l_dataExchangeRights_ = null;
      l_runUnderRights_ = null;
      l_softRunUnderRights_ = null;

      l_containRights_ = null;
      l_negAclRights_ = null;
      l_negTypedPrivRights_ = null;
      l_negLinkedPrivRights_ = null;
      l_isAliasRights_ = null;

      l_isAliasRights_ = null;
      l_globalRights_ = null;
      l_bridgeRights_ = null;
      l_launchRights_ = null;
      
      l_aclRights_ = null;
      l_ownerRights_ = null;
      l_typedPrivRights_ = null;
      l_linkedPrivRights_ = null;

  } // finalizeForGui()

  /**
   * Sets the link type values to false, the rights to null and the comments to empty values.
   */
  public synchronized void setEmpty() {

          // for testing and tracking
          //System.out.println("Call of AccessControlLinkImpl.setEmpty()");

      isMember_ = false;
      hasChild_ = false;
      isIndirectMember_ = false;
      hasVirtualMember_ = false;
      dataExchange_ = false;
      runUnder_ = false;
      softRunUnder_ = false;

      contain_ = false;
      negAcl_ = false;
      negTypedPriv_ = false;
      negLinkedPriv_ = false;
      isAlias_ = false;

      global_ = false;
      bridge_ = false;
      spread_ = false;
      launch_ = false;

      acl_ = false;
      owner_ = false;
      typedPriv_ = false;
      linkedPriv_ = false;

      l_isMemberRights_ = null;
      l_hasChildRights_ = null;
      l_isIndirectMemberRights_ = null;
      l_hasVirtualMemberRights_ = null;
      l_dataExchangeRights_ = null;
      l_runUnderRights_ = null;
      l_softRunUnderRights_ = null;

      l_containRights_ = null;
      l_negAclRights_ = null;
      l_negTypedPrivRights_ = null;
      l_negLinkedPrivRights_ = null;
      l_isAliasRights_ = null;

      l_isAliasRights_ = null;
      l_globalRights_ = null;
      l_bridgeRights_ = null;
      l_launchRights_ = null;

      l_aclRights_ = null;
      l_ownerRights_ = null;
      l_typedPrivRights_ = null;
      l_linkedPrivRights_ = null;
    
      isMemberComments_ = "";
      hasChildComments_ = "";
      isIndirectMemberComments_ = "";
      hasVirtualMemberComments_ = "";
      dataExchangeComments_ = "";
      runUnderComments_ = "";
      softRunUnderComments_ = "";

      containComments_ = "";
      negAclComments_ = "";
      negTypedPrivComments_ = "";
      negLinkedPrivComments_ = "";
      isAliasComments_ = "";

      globalComments_ = "";
      bridgeComments_ = "";
      launchComments_ = "";

      aclComments_ = "";
      ownerComments_ = "";
      typedPrivComments_ = "";
      linkedPrivComments_ = "";
  } // setEmpty()
  

  /**
   * @return long serial version UID for a future serialization of the class
   */
  public long getSerialVersionUID(){
    return serialVersionUID__;
  }

   /**
   * Returns the comments for all the type link values which are true,
   * but not the rights. 
   * In RightsFactory_Facade.getEffectiveRightsForPair(), the SPREAD comment
   * (starting with 'All Rights:') is used in the last link of a DisplayableLink,
   * to show the list of all the effective rights before the rights subtraction.
   * This is why the SPREAD comment is returned after the other comments.
   * The GLOBAL comment is used for delivering general comments, like for an AclEntry
   * from another ACS or a conditional AclEntry. It is put just before the SPREAD comment.
   * Optimized to be a fast method.
   * Called by gWork.FactoryUtilities.getXXXWhyText() for each AccesControlLink in
   * each DisplayableLink in the 'See Why' text.
   * @return text for the 'See why' information
   */
  public String toSeeWhyText() {
    StringBuilder   sb = new StringBuilder();
    boolean         write = false;
    int             index = -1;

    if (contain_ || negAcl_ || negTypedPriv_ || negLinkedPriv_ || isAlias_) {
            if (contain_) {
              if (write) sb.append(" -- ");
              sb.append("Contains");
              //for (int i = 0; i < l_containRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_containRights_[i].getNickName()+"; ");
              //}
              if (!containComments_.equals("")) {
                sb.append(" (");
                sb.append(containComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (negAcl_) {
              if (write) sb.append(" -- ");
              sb.append("Negative ACL");
              //for (int i = 0; i < l_negAclRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_negAclRights_[i].getNickName()+"; ");
              //}
              if (!negAclComments_.equals("")) {
                sb.append(" (");
                sb.append(negAclComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (negTypedPriv_) {
              if (write) sb.append(" -- ");
              sb.append("Negative Typed PRI");
              //for (int i = 0; i < l_negTypedPrivRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_negTypedPrivRights_[i].getNickName()+"; ");
              //}
              if (!negTypedPrivComments_.equals("")) {
                sb.append(" (");
                sb.append(negTypedPrivComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (negLinkedPriv_) {
              if (write) sb.append(" -- ");
              sb.append("Negative Linked PRI");
              //for (int i = 0; i < l_negLinkedPrivRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_negLinkedPrivRights_[i].getNickName()+"; ");
              //}
              if (!negLinkedPrivComments_.equals("")) {
                sb.append(" (");
                sb.append(negLinkedPrivComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (isAlias_) {
              if (write) sb.append(" -- ");
              sb.append("Is alias");
              //for (int i = 0; i < l_isAliasRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_isAliasRights_[i].getNickName()+"; ");
              //}
              if (!isAliasComments_.equals("")) {
                sb.append(" (");
                sb.append(isAliasComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }
    } // if (_acLink.contain_

    if (isIndirectMember_ || hasVirtualMember_ || dataExchange_) {
            if (isIndirectMember_) {
              if (write) sb.append(" -- ");
              sb.append("Indirect Member");
              //for (int i = 0; i < l_isIndirectMemberRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_isIndirectMemberRights_[i].getNickName()+"; ");
              //}
              if (!isIndirectMemberComments_.equals("")) {
                sb.append(" (");
                sb.append(isIndirectMemberComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (hasVirtualMember_) {
              if (write) sb.append(" -- ");
              sb.append("Virtual Member");
              //for (int i = 0; i < l_hasVirtualMemberRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_hasVirtualMemberRights_[i].getNickName()+"; ");
              //}
              if (!hasVirtualMemberComments_.equals("")) {
                sb.append(" (");
                sb.append(hasVirtualMemberComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (dataExchange_) {
              if (write) sb.append(" -- ");
              sb.append("Data Exchange");
              //for (int i = 0; i < l_dataExchangeRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_dataExchangeRights_[i].getNickName()+"; ");
              //}
              if (!dataExchangeComments_.equals("")) {
                sb.append(" (");
                sb.append(dataExchangeComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }
    } // if (_acLink.isIndirectMember_

    if (isMember_ || hasChild_ || runUnder_ || softRunUnder_) {
            if (isMember_) {
              if (write) sb.append(" -- ");
              sb.append("Is Member");
              //for (int i = 0; i < l_isMemberRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_isMemberRights_[i].getNickName()+"; ");
              //}
              if (!isMemberComments_.equals("")) {
                sb.append(" (");
                sb.append(isMemberComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (hasChild_) {
              if (write) sb.append(" -- ");
              sb.append("Has Child");
              //for (int i = 0; i < l_hasChildRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_hasChildRights_[i].getNickName()+"; ");
              //}
              if (!hasChildComments_.equals("")) {
                sb.append(" (");
                sb.append(hasChildComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (runUnder_) {
              if (write) sb.append(" -- ");
              sb.append("Runs Under");
              //for (int i = 0; i < l_runUnderRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_runUnderRights_[i].getNickName()+"; ");
              //}
              if (!runUnderComments_.equals("")) {
                sb.append(" (");
                sb.append(runUnderComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (softRunUnder_) {
              if (write) sb.append(" -- ");
              sb.append("May Run Under");
              //for (int i = 0; i < l_softRunUnderRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_softRunUnderRights_[i].getNickName()+"; ");
              //}
              if (!softRunUnderComments_.equals("")) {
                sb.append(" (");
                sb.append(softRunUnderComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }
    } // if (_acLink.isMember_

    if (bridge_ || launch_) {

            if (bridge_) {
              if (write) sb.append(" -- ");
              sb.append("Bridge");
              //for (int i = 0; i < l_bridgeRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_bridgeRights_[i].getNickName()+"; ");
              //}
              if (!bridgeComments_.equals("")) {
                sb.append(" (");
                sb.append(bridgeComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (launch_) {
              if (write) sb.append(" -- ");
              sb.append("Launches");
              //for (int i = 0; i < l_launchRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_launchRights_[i].getNickName()+"; ");
              //}
              if (!launchComments_.equals("")) {
                sb.append(" (");
                sb.append(launchComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }
    } // if (_acLink.global_

    if (acl_ || owner_ || typedPriv_ || linkedPriv_) {
            if (acl_) {
              if (write) sb.append(" -- ");
              sb.append("ACL");
              //for (int i = 0; i < l_aclRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_aclRights_[i].getNickName()+"; ");
              //}
              if (!aclComments_.equals("")) {
                sb.append(" (");
                sb.append(aclComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (owner_) {
              if (write) sb.append(" -- ");
              sb.append("Owns");
              //for (int i = 0; i < l_ownerRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_ownerRights_[i].getNickName()+"; ");
              //}
              if (!ownerComments_.equals("")) {
                sb.append(" (");
                sb.append(ownerComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (typedPriv_) {
              if (write) sb.append(" -- ");
              sb.append("Typed PRI");
              //for (int i = 0; i < l_typedPrivRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_typedPrivRights_[i].getNickName()+"; ");
              //}
              if (!typedPrivComments_.equals("")) {
                sb.append(" (");
                sb.append(typedPrivComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }

            if (linkedPriv_) {
              if (write) sb.append(" -- ");
              sb.append("Linked PRI");
              //for (int i = 0; i < l_linkedPrivRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_linkedPrivRights_[i].getNickName()+"; ");
              //}
              if (!linkedPrivComments_.equals("")) {
                sb.append(" (");
                sb.append(linkedPrivComments_);
                sb.append(") ");
              } else sb.append(" ");
              write = true;
            }
    } // if (_acLink.acl_

    // comments from the gWork.RightsFactory classes
    if (global_) {

              if (write) sb.append(" -- ");
              //sb.append("Global");
              //for (int i = 0; i < l_globalRights_.length; i++){
                //if (i == 0) sb.append(": ");
                //sb.append(l_globalRights_[i].getNickName()+"; ");
              //}
              if (!globalComments_.equals("")) {
                sb.append(" ");
                sb.append(globalComments_);
                sb.append(" ");
              } else sb.append(" ");
              write = true;

    } // if global_

    // Spread comments are edited by RightsFactory_Facade.getEffectiveRightsForPair()
    // for keeping the global text that is for the SeeWhy text
    if (spread_) {
      if (write) sb.append(" -- ");
      if (!spreadComments_.equals("")) {
        sb.append("All Rights: ");
        if (spreadComments_.length() <= 20) sb.append(spreadComments_);
        else {
            if (spreadComments_.indexOf(" ") == -1 ||
                    spreadComments_.indexOf(" ") > 20) {

                                      // for testing and tracking
                                      //System.out.println("AccessControlLinkImpl.toSeeWhyText() "+
                                        //      "spreadComments_: "+spreadComments_);

                index = spreadComments_.indexOf("/", 15);
                if (index != -1) {
                    sb.append(spreadComments_.substring(0, index));
                    sb.append(" /");
                    sb.append(spreadComments_.substring(index + 1));
                }
                else if (spreadComments_.indexOf("/", 11) != -1) {
                    sb.append(spreadComments_.substring(0, spreadComments_.indexOf("/", 11)));
                    sb.append(" /");
                    sb.append(spreadComments_.substring(spreadComments_.indexOf("/", 11) + 1));
                }
                else {
                    sb.append(spreadComments_.substring(0, spreadComments_.indexOf("/", 7)));
                    sb.append(" /");
                    sb.append(spreadComments_.substring(spreadComments_.indexOf("/", 7) + 1));
                } // else
            } else
                sb.append(spreadComments_);
        } // else
        //sb.append(") ");
      }
      else sb.append(" (no right)");
      write = true;
    }

    if (!write) sb.append("empty AccessControlLink");
    return sb.toString();
  } // toSeeWhyText()
  
   /**
   * Returns the comments for all the type link values which are true,
   * but not the rights. 
   * In RightsFactory_Facade.getEffectiveRightsForPair(), the SPREAD comment
   * is used in the last link of a DisplayableLink to get the list of 
   * the effective rights in the DisplayableLinkImpl.
   * Called by DisplayableLinkImpl.toString().
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    boolean write = false;

    if (contain_ || negAcl_ || negTypedPriv_ || negLinkedPriv_ || isAlias_) {
        if (contain_) {
          if (write) sb.append(" -- ");
          sb.append("Contains");
          for (int i = 0; i < l_containRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_containRights_[i].getNickName()+"; ");
          }
          if (!containComments_.equals("")) {
            sb.append(" (");
            sb.append(containComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (negAcl_) {
          if (write) sb.append(" -- ");
          sb.append("Negative ACL");
          for (int i = 0; i < l_negAclRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_negAclRights_[i].getNickName()+"; ");
          }
          if (!negAclComments_.equals("")) {
            sb.append(" (");
            sb.append(negAclComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (negTypedPriv_) {
          sb.append("Negative Typed PRIV");
          for (int i = 0; i < l_negTypedPrivRights_.length; i++){
                if (i == 0) sb.append(": ");
                sb.append(l_negTypedPrivRights_[i].getNickName()+"; ");
          }
          if (!negTypedPrivComments_.equals("")) {
            sb.append(" (");
            sb.append(negTypedPrivComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (negLinkedPriv_) {
          if (write) sb.append(" -- ");
          sb.append("Negative Linked PRIV");
          for (int i = 0; i < l_negLinkedPrivRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_negLinkedPrivRights_[i].getNickName()+"; ");
          }
          if (!negLinkedPrivComments_.equals("")) {
            sb.append(" (");
            sb.append(negLinkedPrivComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (isAlias_) {
          if (write) sb.append(" -- ");
          sb.append("Is alias");
          for (int i = 0; i < l_isAliasRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_isAliasRights_[i].getNickName()+"; ");
          }
          if (!isAliasComments_.equals("")) {
            sb.append(" (");
            sb.append(isAliasComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }
    } // if (_acLink.contain_

    if (isIndirectMember_ || hasVirtualMember_ || dataExchange_ ) {
        if (isIndirectMember_) {
          if (write) sb.append(" -- ");
          sb.append("Indirect Member");
          for (int i = 0; i < l_isIndirectMemberRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_isIndirectMemberRights_[i].getNickName()+"; ");
          }
          if (!isIndirectMemberComments_.equals("")) {
            sb.append(" (");
            sb.append(isIndirectMemberComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (hasVirtualMember_) {
          if (write) sb.append(" -- ");
          sb.append("Virtual Member");
          for (int i = 0; i < l_hasVirtualMemberRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_hasVirtualMemberRights_[i].getNickName()+"; ");
          }
          if (!hasVirtualMemberComments_.equals("")) {
            sb.append(" (");
            sb.append(hasVirtualMemberComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (dataExchange_) {
          if (write) sb.append(" -- ");
          sb.append("Data Exchange");
          for (int i = 0; i < l_dataExchangeRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_dataExchangeRights_[i].getNickName()+"; ");
          }
          if (!dataExchangeComments_.equals("")) {
            sb.append(" (");
            sb.append(dataExchangeComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }
    } // if (_acLink.isIndirectMember_

    if (isMember_ || hasChild_ || runUnder_ || softRunUnder_) {
        if (isMember_) {
          if (write) sb.append(" -- ");
          sb.append("Is Member");
          for (int i = 0; i < l_isMemberRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_isMemberRights_[i].getNickName()+"; ");
          }
          if (!isMemberComments_.equals("")) {
            sb.append(" (");
            sb.append(isMemberComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (hasChild_) {
          if (write) sb.append(" -- ");
          sb.append("Has Child");
          for (int i = 0; i < l_hasChildRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_hasChildRights_[i].getNickName()+"; ");
          }
          if (!hasChildComments_.equals("")) {
            sb.append(" (");
            sb.append(hasChildComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (runUnder_) {
          if (write) sb.append(" -- ");
          sb.append("Runs Under");
          for (int i = 0; i < l_runUnderRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_runUnderRights_[i].getNickName()+"; ");
          }
          if (!runUnderComments_.equals("")) {
            sb.append(" (");
            sb.append(runUnderComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (softRunUnder_) {
          if (write) sb.append(" -- ");
          sb.append("May Run Under");
          for (int i = 0; i < l_softRunUnderRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_softRunUnderRights_[i].getNickName()+"; ");
          }
          if (!softRunUnderComments_.equals("")) {
            sb.append(" (");
            sb.append(softRunUnderComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }
    } // if (_acLink.isMember_

    if (global_ || bridge_ || spread_ || launch_) {
        if (global_) {
          sb.append("Global");
          for (int i = 0; i < l_globalRights_.length; i++){
                if (i == 0) sb.append(": ");
                sb.append(l_globalRights_[i].getNickName()+"; ");
          }
          if (!globalComments_.equals("")) {
            sb.append(" (");
            sb.append(globalComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (bridge_) {
          if (write) sb.append(" -- ");
          sb.append("Bridge");
          for (int i = 0; i < l_bridgeRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_bridgeRights_[i].getNickName()+"; ");
          }
          if (!bridgeComments_.equals("")) {
            sb.append(" (");
            sb.append(bridgeComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (spread_) {
          sb.append("Spread");
          if (!spreadComments_.equals("")) {
            sb.append(" (");
            sb.append(spreadComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (launch_) {
          if (write) sb.append(" -- ");
          sb.append("Launches");
          for (int i = 0; i < l_launchRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_launchRights_[i].getNickName()+"; ");
          }
          if (!launchComments_.equals("")) {
            sb.append(" (");
            sb.append(launchComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }
    } // if (_acLink.global_

    if (acl_ || owner_ || typedPriv_ || linkedPriv_) {
        if (acl_) {
          if (write) sb.append(" -- ");
          sb.append("ACL");
          for (int i = 0; i < l_aclRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_aclRights_[i].getNickName()+"; ");
          }
          if (!aclComments_.equals("")) {
            sb.append(" (");
            sb.append(aclComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (owner_) {
          if (write) sb.append(" -- ");
          sb.append("Owns");
          for (int i = 0; i < l_ownerRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_ownerRights_[i].getNickName()+"; ");
          }
          if (!ownerComments_.equals("")) {
            sb.append(" (");
            sb.append(ownerComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (typedPriv_) {
          sb.append("Typed PRIV");
          for (int i = 0; i < l_typedPrivRights_.length; i++){
                if (i == 0) sb.append(": ");
                sb.append(l_typedPrivRights_[i].getNickName()+"; ");
          }
          if (!typedPrivComments_.equals("")) {
            sb.append(" (");
            sb.append(typedPrivComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }

        if (linkedPriv_) {
          if (write) sb.append(" -- ");
          sb.append("Linked PRIV");
          for (int i = 0; i < l_linkedPrivRights_.length; i++){
            if (i == 0) sb.append(": ");
            sb.append(l_linkedPrivRights_[i].getNickName()+"; ");
          }
          if (!linkedPrivComments_.equals("")) {
            sb.append(" (");
            sb.append(linkedPrivComments_);
            sb.append(") ");
          } else sb.append(" ");
          write = true;
        }
    } // if (_acLink.acl_
    
    if (!write) sb.append("empty AccessControlLink");
    return sb.toString();
  } // toString()

} // AccessControlLinkImpl