//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.CNot;

import ARoad0.gBaseInterface.StringRight;
import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.BaseObject;
import ARoad0.gBaseInterface.ImmutableActor;
import ARoad0.gBaseInterface.ImmutableVirtualFolder;
import ARoad0.gBaseInterface.ImmutableGroupIDMember;

import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Collections;
import java.util.Map;
import java.util.HashMap;

import java.io.Serializable;
import java.io.ObjectStreamField;

 /**
  * This class is responsible for modeling a displayable link
  * that manages a path of base objects connected through AccessControlLinkImpls.
  * Typically, it is associated to an item in a graphical view
  * (an arrow or a line for instance) between two view objects.
  * A displayable link manages a chain of nodes where there is one node for one base object.
  * It manages also the access control links between nodes, defining the paths.
  * The results in the right searching for a view are put by the gWork classes
  * in DisplayableLinkImpl instances. They may be seen as a dynamic data base 
  * of the access paths in the open views. They are ridden both to get
  * the effective rights and to directly generate the 'See Why' text of the view.
  * <P>
  * A displayable link may be simple - it contains then only the 2 base objects
  * as ends and one link between them - or it may be a chain of n base objects
  * (or nodes) with n-1 links. In this last case, the intermediate base objects
  * are often hidden, if it is for a view.
  * The two ends are always immutable. They can't be changed in the displayable link,
  * but it is possible to create a new link from a previous link
  * with the static method valueOf().
  * <P>
  * An internal list contains the chain of BaseObjects and AccessControlLinks
  * which defines the DisplayableLink. Usually, the intermediate BaseObjects
  * are ImmutableSources: Actors, GroupIDMembers, VirtualFolders.
  * The possible sizes of the internal vector is 3, 5, 7... Its structure is designed
  * to add easily a link or a node without changing of the ends. It is:<P>
  * - 0: first end,<P>
  * - 1: second end,<P>
  * - 2: first access control link between the first end and its next node
  *   (which is the second end if the displayable link is simple),<P>
  * - 3: first intermediate node if the displayable link is not simple,<P>
  * - 4: second access control link between the first
  * intermediate node and the second intermediate node, if it exists,
  * or the second end if the path has 3 nodes.<P>
  * - ...<P>
  * <P>
  * Some general properties are defined for a DisplayableLinkImpl in a map.
  * The first use is the 2 map keys AG_CONTEXT_NODE_POSITION and AG_CONTEXT_SET,
  * for the Account/Groups context of the link.
  * The AG context is a set of GroupIDMembers that may contain, first,
  * an optional couple (UserID, GroupID), and, second, a list of optional GroupIDs.
  * AG_CONTEXT_NODE_POSITION is the index of a node in the DislayableLinkImpl
  * for which there is a defined Account/Groups context.
  * The key AG_CONTEXT_SET returns a couple (UserID, GroupID)
  * the node at AG_CONTEXT_NODE_POSITION may use as Account/Main_Group context.
  * The key AG_CONTEXT_SET may also return complementary groups. <P>
  * The links depend on the varied ACS which are implied in the view,
  * and also may depend on some ACS not implied in the view
  * but which control some intermediate base objects.
  * A displayable link is not always displayed in a view. For instance,
  * it may contain a right between two objects which is deleted by a stronger right
  * in another displayable link containing the same objects.
  * @see ARoad0.CNot.AccessControlLinkImpl
  * @see ARoad0.gWork.DisplayableLinkUtilities#getEffectiveRightsAndCommentOnRights(ARoad0.CNot.DisplayableLinkImpl, boolean) 
  * @version 0.7.0
  */
public final class DisplayableLinkImpl implements Cloneable, Serializable {

  // private static variables
  private static final long serialVersionUID__ = 0;
  private static final int PRIME = 1000003;

  // protected static variables
      // used by AccessControlLinkImpl.getRightsToString()
  protected static final int INITIAL_CAPACITY = 20 ;  // the default value is 10

  // public static variables
  /**
   * Key of l_properties_ where the value is the Set of (O to 1) UserId and (0 to N) GroupIDs
   * that defines the Account/Groups context of the node given by the key AG_CONTEXT_NODE_POSITION.
   */
  public static final String AG_CONTEXT_SET =
          "Account/Groups context as set of one UserId and some GroupIDs";

  /**
   * Key of l_properties_ where the value is the position of the node
   * having the AG context set in AG_CONTEXT_SET.
   */
  public static final String AG_CONTEXT_NODE_POSITION =
          "node having the AG context set by the key AG_CONTEXT_SET";

  /**
  * This list contains the chain of BaseObjects and AccessControlLinks
  * which defines the DisplayableLink. Usually, all intermediate BaseObjects
  * are Actors, UserIDs or GroupIDs.
  * The possible sizes of this vector is 3, 5, 7, 9... Its structure is designed
  * to easily add a link without change the ends. It is:<P>
  * - 0: first end,<P>
  * - 1: second end,<P>
  * - 2: first access control link between the first end and its next node
  *   (which is the second end if the displayable link is simple),<P>
  * - 3: first intermediate node if the displayable link is not simple,<P>
  * - 4: second access control link between the first
  * intermediate node and the second intermediate node, if it exists,
  * or the second end if the path has 3 nodes.<P>
  * - ...
  */
  protected ArrayList vector_ = null;

  /**
   * This map describes the general properties of the DisplayableLinkImpl.
   * The first use is the 2 keys AG_CONTEXT_NODE_POSITION and AG_CONTEXT_SET
   * for the Account/Groups context of the link.
   */
  protected Map<String, Object> m_properties_;

  /** standardized variable for the JDK 2 serialization
  * @serial ("vector_", Vector.class), ("l_properties_", HashMap.class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("vector_", ArrayList.class),
      new ObjectStreamField("l_properties_", HashMap.class)
  };

  /*
  * Default contructor.
  * The arguments when they exist are copied. So, the initial object
  * referenced by an argument constructor is never updated.
  */

  /**
  */
  public DisplayableLinkImpl() {
  }

  /**
   * Used by valueOf().
   * @param _components is copied without detailled controls on its structure
   * @exception InternalError if _components is null, or if one node or one aclink is null
   *   or duplicated
   */
  public DisplayableLinkImpl(ArrayList _components) {
    if (_components == null || _components.size() < 3) throw new InternalError(
      "In the DisplayableLinkImpl constructor, the parameter is null or too small");
    
    vector_ = new ArrayList(_components.size());
    boolean error = false;
    boolean error2 = false;

    for (int i = 0 ; i < _components.size() ; i++) {
      if (_components.get(i) == null ) {
          error = true;
          break;
      }
      if (vector_.contains( _components.get(i) )) {
          error2 = true;
          break;
      }
      else
          vector_.add( _components.get(i) );
    } // for i

    if (error)
        throw new InternalError("In the DisplayableLinkImpl constructor, "+
            "one node or one aclink is null");
    if (error2)
        throw new InternalError("In the DisplayableLinkImpl constructor, "+
            "one node or one aclink is duplicated");
  } // constructor

  /**
  * @param _firstEnd is a non-null BaseObject as the first end of the displayable link.
  * It is not copied.
  * @param _secondEnd is a non-null BaseObject as the second end of the displayable link.
  * It is not copied.
  * @param _link is a non-null first AccessControlLinkImpl
  * @exception InternalError if one parameter is null or a duplicate
  */
  public DisplayableLinkImpl(BaseObject _firstEnd, BaseObject _secondEnd,
    AccessControlLinkImpl _link) {
      if (_firstEnd == null || _secondEnd == null || _link == null || 
              _firstEnd == _secondEnd || _firstEnd.equals(_secondEnd))
        throw new InternalError(
                "In the DisplayableLinkImpl constructor, null or duplicate item");

      vector_ = new ArrayList(INITIAL_CAPACITY);
      vector_.add(_firstEnd);
      vector_.add(_secondEnd);
      vector_.add(_link);
  } // DisplayableLinkImpl()

  /**
   * Returns instance of a displayable link with empty AccessControlLinks.
  * @param _firstEnd is a non-null BaseObject as the first end of the displayable link.
  *   It is not copied.
  * @param _secondEnd is a non-null BaseObject as the second end of the displayable link.
  *   It is not copied.
  * @param _path is an array of BaseObjects which are the intermediate nodes
  *   of the displayable link. May be null, but if not, each _path item must be non-null
  *   without duplication over the nodes.
  * @exception InternalError if a BaseObject is null or a duplicate
  */
  public DisplayableLinkImpl(BaseObject _firstEnd, BaseObject _secondEnd,
    BaseObject[] _path) {

      this(_firstEnd, _secondEnd, new AccessControlLinkImpl());
      boolean exception = false;
      if (_path != null) {
        Set set = new HashSet(25);
        set.add(_firstEnd);
        set.add(_secondEnd);

        for (int i = 0; i < _path.length; i++) {
          if (_path[i] == null || set.contains(_path[i])) {
            exception = true;
            break;
          }
          // main work
          vector_.add( _path[i] );
          vector_.add(new AccessControlLinkImpl());

          set.add(_path[i]);
        } // for
        set.clear();
        set = null;
      } // if
      
      if (exception) throw new InternalError(
        "In the DisplayableLinkImpl constructor, one null or duplicate node");
  } // DisplayableLinkImpl()

  //
  // ---------------- static methods ----------------------------------
  //

  /**
   * To create a new DisplayableLinkImpl, extends a current DisplayableLinkImpl _dLink
   * to which, after a deep copy, a new node is added as the new second end.
   * So, the returned value and _dLink have the same nodes excepting the second end,
   * but with different AccessControlLinks since these links are cloned.
   * The returned value has a copy of the _dLink properties where the values 
   * are cloned if they are ImmutableName, Integer, HashSet, ArrayList or anything else,
   * but in the latter case, the value is not cloned.
   * Returns a null result if the argument _secondEnd is already known in _dLink.
   * @param _secondEnd to add at the end of _dLink to define a new second end. Never null.
   * @param _lastLink to put before _secondEnd in the new DisplayableLinkImpl. Never null.
   * @exception InternalError if one parameter is null.
   * @return the new displayable link, no-backed to _dLink, or null if
   *   _secondEnd is known in _dLink.
   */
   public static final DisplayableLinkImpl valueOf(
    DisplayableLinkImpl _dLink, BaseObject _secondEnd, AccessControlLinkImpl _lastLink) {

      if (_dLink == null)
        throw new InternalError("In DisplayableLinkImpl.valueOf(), the DisplayableLink is null");
      if (_secondEnd == null)
        throw new InternalError("In DisplayableLinkImpl.valueOf(), the second end is null");
      if (_lastLink == null)
        throw new InternalError("In DisplayableLinkImpl.valueOf(), the last link is null");
      if (_dLink.containsBaseObject(_secondEnd)) return null;

      int index = 0;
      Object obj;
      ArrayList vector = new ArrayList(INITIAL_CAPACITY);
      vector.ensureCapacity(_dLink.vector_.size()+3);
      DisplayableLinkImpl result;
      Map.Entry entry = null;

                            // for testing and tracking
                            //System.out.println("in DisplayableLinkImpl.valueOf(), _dLink.vector_: "
                              //+_dLink.vector_+" and second end is: "+_secondEnd);

      for (Iterator ite = _dLink.vector_.iterator(); ite.hasNext() ; ) {
        obj = ite.next();

                            // for testing and tracking
                            //System.out.println("in DisplayableLinkImpl.valueOf(),"
                            //  +" index: "+index+", obj: "+obj);

        if (index == 0) {
          vector.add(index, obj);
        }
        // index == 1 is for _secondEnd
        else if (index == 1) {
          vector.add(index, _secondEnd);
        }
        // deep copy of all the Access Control Links, including the last one
        else if (index%2 == 0) {
          vector.add(index, ((AccessControlLinkImpl)obj).clone() );
        }
        // shallow copy of all the nodes, but not the second end
        else if (index != 1 && index%2 != 0) {
          vector.add(index, obj );
        }
        index++;
      } // for ite

      // add the dLink second end at the end of vector
      vector.add(_dLink.vector_.get(1));
      // add the last link
      vector.add(_lastLink);

                            // for testing and tracking
                            //System.out.println("in DisplayableLinkImpl.valueOf(), vector: "
                              //+vector);

      result = new DisplayableLinkImpl(vector);

      Map<String, Object> m_properties = _dLink.getProperties();

      if (m_properties != null) {
          for (Object ob : m_properties.entrySet()) {
              entry = (Map.Entry)ob;
              if (entry.getValue() instanceof ImmutableName)
                  result.addProperty(
                          (String)entry.getKey(),
                          ((ImmutableName)entry.getValue()).clone());
              if (entry.getValue() instanceof Integer)
                  result.addProperty(
                          (String)entry.getKey(),
                          new Integer(((Integer)entry.getValue()).intValue()));
              if (entry.getValue() instanceof HashSet)
                  result.addProperty(
                          (String)entry.getKey(),
                          ((HashSet)entry.getValue()).clone());
              if (entry.getValue() instanceof ArrayList)
                  result.addProperty(
                          (String)entry.getKey(),
                          ((ArrayList)entry.getValue()).clone());
              else 
                  result.addProperty(
                          (String)entry.getKey(),
                          entry.getValue());
          } // for
      } // if (m_properties != null)

      obj = null;
      vector = null;

                            // for testing and tracking
                            //if (result != null && m_properties != null)
                              //  System.out.println("DisplayableLinkImpl.valueOf() end, "+
                                //    " AG context position in result: "
                                  //  +result.getProperties().
                                    //get(DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION));

      return result;
  } // valueOf()

  //
  // ---------------- public methods ----------------------------------
  //

  /**
   * @return true if the displayable link  has no component, or if all components
   * are null.
   */
   public boolean isEmpty() {
    if (vector_ == null) return true;
    boolean result = true;
    for (int i = 0 ; i < vector_.size() ; i++) {
      if (vector_.get(i) != null ) {
          result = false;
          break;
      }
    } // for
    return result;
  }

  /**
   * @return true if the displayable link contains only the two ends and one
   * direct link between us.
   */
  public boolean isSimple() {
    if (vector_ != null && vector_.size() == 3) return true;
    else return false;
  } // isSimple()

  /**
   * @return true if an AccessControlLink is weak, whatever its position.
   */
  public boolean isWeak() {
      boolean result = false;
      AccessControlLinkImpl[] l_links = getEorAllACLinks();
      for (int i = 0; i < l_links.length; i++) {
          if (l_links[i].isWeak()) {
            result = true;
            break;
          }
      } // for i
      l_links = null;
      return result;
  } // isWeak()
  
  /**
  * @return the first end of the displayable link.
  * May be null.
  */
  public BaseObject getEorFirstEnd() {
    if (vector_ != null) {
        return (BaseObject)vector_.get(0);
    } else return null;
  } // getFirstEnd()

  /**
  * @return the second end of the displayable link
  * May be null.
  */
  public BaseObject getEorSecondEnd() {
    if (vector_ != null) {
        return (BaseObject)vector_.get(1);
    } else return null;
  } // getSecondEnd()

  /**
  * @return the node before the second end of the displayable link
  * May be null. Is the first end if the displayable link is simple.
  */
  public BaseObject getEorNodeJustBeforeTheSecondEnd() {
    if (vector_ != null && vector_.size() < 4) {
        return (BaseObject)vector_.get( 0 );
    } else if (vector_ != null) {
        return (BaseObject)vector_.get( vector_.size() - 2 );
    } else return null;
  } // getEorNodeJustBeforeTheSecondEnd()
  
  /**
  * If the instance is simple, return the single aclink.
  * @return AccessControlLinkImpl the first link of the displayable link.
  *   May be null.
  */
  public AccessControlLinkImpl getEorFirstLink() {
    if (vector_ != null) {
        return (AccessControlLinkImpl)vector_.get(2);
    } else return null;
  } // getEorFirstLink()

  /**
  * If the instance is simple, returns the single aclink.
  * @return AccessControlLinkImpl the last link of the displayable link.
  *   May be null.
  */
  public AccessControlLinkImpl getEorLastLink() {
    if (vector_ != null) {
        return (AccessControlLinkImpl)vector_.get(vector_.size() - 1);
    } else return null;
  } // getEorLastLink()

  /**
  * Returns a sorted array of all the nodes, including the first end at the index 0
  * and the second end at the end of the array.
  * The value at the index i is connected to the acLinks
  * getEorAllACLinks()[i-1] and getEorAllACLinks()[i], if it is not 0 or (noOfNodes() - 1).
  * @return BaseObject[] array of all the nodes. May be null.
  */
  public BaseObject[] getEorAllNodes() {
    BaseObject[] result = new BaseObject[noOfNodes()];
    if (vector_ != null) {
      synchronized (vector_) {
        result[0] = (BaseObject)vector_.get(0);
        result[result.length - 1] = (BaseObject)vector_.get(1);
        if (result.length > 2) {
          // count set to the index of the first intermediate node
          int count = 3;
          for (int i = 1 ; i < result.length - 1 ; i++) {
            result[i] = (BaseObject)vector_.get(count);
            count += 2;

                      //if (result.length > 10 && count > 7) {
                         // System.out.println("DisplayableLinkImpl.getEorAllNodes() result:");
                          //for (int i = 0; i < result.length; i++) {
                           //   System.out.println("for i = "+i+" ----------- "+result[i]);
                           //   System.out.println("count = "+count);
                          //}
                      ////} // if

          } // for (int i = 1
        } // if (result.length > 2)
      } // synchro

                      //if (noOfNodes() > 10) {
                        //  System.out.println("DisplayableLinkImpl.getEorAllNodes() result at the end:");
                          //for (int i = 0; i < result.length; i++) {
                          //    System.out.println("result.length: "+result.length+
                            //          " in thread "+Thread.currentThread().getName());
                          //}
                      //} // if

      return result;
    } else return null;
  } // getEorAllNodes()

  /**
   * Note: getEorAllNodes() is a little bit faster.
  * @return BaseObject[] array of all the nodes, but not the ends.
  *   Is null if the link is empty or simple.
  */
  public BaseObject[] getEorNodesPath() {
    if (vector_ == null || isSimple()) return null;
    BaseObject[] l_nodes = getEorAllNodes();
    BaseObject[] result = new BaseObject[l_nodes.length - 2];
    System.arraycopy(l_nodes, 1, result, 0, result.length);
    return result;
  } // getEorNodesPath()

  /**
  * Gets an array of all the AccessControlLinkImpl in this DisplayableLinkImpl.
  * Internally synchronized on the vector. The value at the index i is connected to the nodes
  * getEorAllNodes()[i] and getEorAllNodes()[i+1].
  * @return AccessControlLinkImpl[] array of all the links, including the first one
  * at the index 0. May be null.
  */
  public AccessControlLinkImpl[] getEorAllACLinks() {
    if (vector_ != null) {
      synchronized (vector_) {
        int shift = 0;
        if (vector_.size()%2 != 0) shift = (int)0.5;
        AccessControlLinkImpl[] result = new AccessControlLinkImpl[
          (int)(vector_.size()/2) - shift];
        result[0] = (AccessControlLinkImpl)vector_.get(2);
        if (result.length > 1) {
          int count = 4;
          for (int i = 1 ; i < result.length ; i++) {
            result[i] = (AccessControlLinkImpl)vector_.get(count);
            count += 2;
          }  // for i
        } // if (result.length > 1)
        return result;
      } // synchro
    } else return null;
  } // getAllACLinks()

  /**
   * The path contains all the intermediate nodes between the two ends.
   * Called by RightsFactory_Facade.getEffectiveRightsForPair() to add '<via>'
   * to the String for an arrow in the view.
  * @return the last Actor, if any, inthe path. May be null.
  */
  public ImmutableActor getEorLastActorInPath() {
    ImmutableActor result = null;
    BaseObject[] l_path;
    if (vector_ != null) {
      synchronized (vector_) {
        l_path = getEorNodesPath();
        if (l_path != null) {
            
          for (int i = l_path.length - 1 ; i > -1  ; i--) {
            if (l_path[i] instanceof ImmutableActor) {
                result = (ImmutableActor)l_path[i];
          
                            // for testing and tracking
                            //System.out.println("DisplayableLinkImpl.getEorLastActorInPath() "+
                              //      "the last actor is "+result+" in "+this); 
                
                break;
            }
          } // for (int i = 1
          l_path = null;
          return result;
          
        } // if (l_path !=)
        else return null;
      } // synchro
    } else return null;
  } // getEorLastActorInPath()

  /**
   * The path contains all the intermediate nodes between the two ends.
   * Called by DisplayableLinkImpl.getContainVirtualFolderWithoutMember().
  * @return the last VirtualFolder, if any. May be null.
  */
  public ImmutableVirtualFolder getEorLastVirtualFolderInPath() {
    ImmutableVirtualFolder result = null;
    BaseObject[] l_path;
    if (vector_ != null) {
      synchronized (vector_) {
        l_path = getEorNodesPath();
        if (l_path != null) {
            
          for (int i = l_path.length - 1 ; i > -1  ; i--) {
            if (l_path[i] instanceof ImmutableVirtualFolder) {
                result = (ImmutableVirtualFolder)l_path[i];
          
                            // for testing and tracking
                            //System.out.println("DisplayableLinkImpl.getEorLastVirtualFolderInPath() "+
                              //      "the last actor is "+result+" in "+this); 
                
                break;
            }
          } // for (int i = 1
          l_path = null;
          return result;
          
        } // if (l_path !=)
        else return null;
      } // synchro
    } else return null;
  } // getEorLastVirtualFolderInPath()
  
  /**
   * @return List of all the components in the displayable link (that is,
   *   BaseObjects and AccessControlLinkImpls). Is immutable. May be null.
   */
  public List immutableComponents() {
    if (vector_ == null) return null;
    return Collections.unmodifiableList(vector_);
  } // immutableComponents()

  /**
   * @return int number of components in the instance, including
   *   all the AccessControlLinks and all the nodes
   */
  public int noOfComponents() {
    if (vector_ != null) return vector_.size();
    else return 0;
  } // noOfComponents()

  /**
   * Returns the number of nodes, including the first and
   * the second ends.
   * @return number of nodes in the instance
   */
  public int noOfNodes() {
    if (vector_ != null)
      return (int)vector_.size()%2 + (int)vector_.size()/2;
    else return 0;
  } // noOfNodes()

   /**
   * @param _obj BaseObject
   * @return true if _obj is a node of the DisplayableLinkImpl
   */
  public boolean containsBaseObject(BaseObject _obj) {
    if (vector_ == null) return false;
    return vector_.contains(_obj);
  } // containsBaseObject()

   /**
   * Returns the position of the BaseObject as node returned in getEorAllNodes().
   * It is not the position in immutableComponents().
   * Uses getEorAllNodes().
   * @param _obj BaseObject
   * @return the position of the BaseObject, or -1 if it is not a node.
   */
  public int indexOfBaseObject(BaseObject _obj) {
    if (!containsBaseObject(_obj)) return -1;
    int result = -1;
    BaseObject[] l_obj = getEorAllNodes(); // never null here
    for (int i = 0; i < l_obj.length ; i++) {
      if (l_obj[i] == _obj) {
        result = i;
        break;
      }
    } // for i
    l_obj = null;
    return result;
  } // containsBaseObject()

  /**
  * Detects the presence of a sequence of 3 nodes in this instance, in the same order.
  * Called by DisplayableLinkUtilities.selectDisplayableLinksWith3NodesSequence().
  * @param _l_sequence array of 3 non-null nodes. Never null.
  * @return true if the argument is a portion of the nodes sequence in this instance
  */
  public boolean contains3NodesSequence(BaseObject[] _l_sequence) {
    boolean result = false;
    BaseObject[] l_nodes = null;
    if (vector_ != null && vector_.size() > 4 && _l_sequence.length == 3) {
      synchronized (vector_) {
          
          l_nodes = getEorAllNodes();
          for (int i = 0 ; i < l_nodes.length - 2 ; i++) {
            if (  _l_sequence[0] == l_nodes[i] && 
                  _l_sequence[1] == l_nodes[i+1] &&
                  _l_sequence[2] == l_nodes[i+2] ) {
                result = true;
                break;
            } // if
          } // for (int i = 1
          l_nodes = null;
          return result;
          
      } // synchro
    } else return false;
  } // contains3NodesSequence()

  /**
  * Detects the presence of a sequence of 2 nodes in this instance, in the same order,
  * for a given first end. Return true if the first end is the first node of the nodes sequence.
  * Called by DisplayableLinkUtilities.selectDisplayableLinksWith2NodesSequenceAndFirstEnd().
  * @param _l_sequence array of 2 non-null nodes. Never null. Return false if the array
  *     has not a lenght of 2.
  * @param _firstEnd first end
  * @return true if the first end is correct, and _l_sequence is a portion 
  *     of the nodes sequence in this instance
  */
  public boolean contains2NodesSequenceAndFirstEnd(
          BaseObject[] _l_sequence, BaseObject _firstEnd) {
    boolean result = false;
    BaseObject[] l_nodes = null;
    if (vector_ != null && vector_.size() > 4 && _l_sequence.length == 2) {
      synchronized (vector_) {
          if (vector_.get(0) != _firstEnd) return false;
          
          l_nodes = getEorAllNodes();
          for (int i = 0 ; i < l_nodes.length - 1 ; i++) {
            if (  _l_sequence[0] == l_nodes[i] && 
                  _l_sequence[1] == l_nodes[i+1] ) {
                result = true;
                break;
            } // if
          } // for (int i = 1
          l_nodes = null;
          return result;
          
      } // synchro
    } else return false;
  } // contains2NodesSequenceAndFirstEnd()
  
   /**
   * Tests if a right is in any AccessControlLink in this instance, through
   * the use of AccessControlLinkImpl.getRedondantRights().
   * Caution: the right may be invalidated by other rights in other
   * AccessControlLinks, or may be invalidated by some negative rights
   * in this instance.
   * @param _right to test
   * @return true if _right is there
   */
  public boolean containsRight(StringRight _right) {
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    if (l_ACLinks == null) return false;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      for (Iterator ite = l_ACLinks[i].getRedondantRights().iterator() ;
        ite.hasNext() ; ) {
          if (((StringRight)ite.next()).equals(_right) ) return true;
      }
    } // for (int i = 0
    return false;
  } // containsRight()

   /**
    * No use.
   * Tests if all AccessControlLinks in this instance contain
   * a pseudo structure relation: CONTAIN, OWNER, or BRIDGE.
   * @return true if there is a pseudo structure relation between
   *  each nodes pair.
   */
  public boolean containsPseudoStructuralLinkTypes() {
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    boolean result = true;
    if (l_ACLinks == null || l_ACLinks.length == 0) return false;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (!l_ACLinks[i].isTrueForOnePseudoStructuralLinkType()) {
          result = false;
          break;
      }
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // containsPseudoStructuralLinkTypes()

   /**
    * NO USE.
   * Tests if all AccessControlLinks in this instance contain
   * a structure relation among IS_MEMBER, HAS_CHILD, IS_INDIRECT_MEMBER,
   * HAS_VIRTUAL_MEMBER, IS_ALIAS.
   * @return true if there is a structure relation between
   *  each nodes pair.
   */
  public boolean containsStructureLinkTypes() {
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    boolean result = true;
    if (l_ACLinks == null || l_ACLinks.length == 0) return false;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (!l_ACLinks[i].isTrueForOneStructuralLinkType()) {
          result = false;
          break;
      }
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // containsStructureLinkTypes()

   /**
    * GraphicEPRView displays a strengthen relation as a line, not an arrow.
  * Checks if all the AccessControlLinks have at least one same strengthen relation.
  * Analyzes the state of activation for the types OWNER, IS_MEMBER, HAS_CHILD,
  * IS_INDIRECT_MEMBER, IS_ALIAS and RUN_UNDER.
  * Called by Gui2.GraphicEPRView.hasStrengthenLinksInOneDLink() and
  * RightsFactory_Facade.getEffectiveRightsForPair().
  * @return true if there is the same strengthen relation between each nodes pair.
   */
  public boolean containsStrengthenLinkTypes() {
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    boolean                 result = true;

    if (l_ACLinks == null || l_ACLinks.length == 0) return false;

    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (!l_ACLinks[i].isTrueForOneStrengthenLinkType()) {
          result = false;
          
                    // for testing and tracking
                    //System.out.println("in DisplayableLinkImpl.containsStrengthenLinkTypes(), "+
                      //      " it does not contain strengthen link types - dLink: "+
                      //      this);
          
          break;
      } // if
      if (i != l_ACLinks.length - 1 &&
              !l_ACLinks[i].getL_ActivatedStrengthenTypes().containsAll(
                l_ACLinks[l_ACLinks.length - 1].getL_ActivatedStrengthenTypes())) {
          result = false;
          break;
      } // if (i !=
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // containsStrengthenLinkTypes()

   /**
  * Tests if each of its AccessControlLinks has a OWNER or CONTAIN relation.
   * Called by Gui2.GraphicEPRView.hasOwnerContainLinkToGroupIDMember().
   * @return true if there is a owner or contain relation between
   *  each nodes pair.
   */
  public boolean containsOwnerContainLinkTypes() {      
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    boolean result = true;
    if (l_ACLinks == null || l_ACLinks.length == 0) return false;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (!l_ACLinks[i].isTrueForOwnerOrContainLinkType() ) {
          result = false;
          break;
      }
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // containsOwnerContainLinkTypes()

   /**
  * Tests if each of its AccessControlLinks has a ACL or NEG_ACL relation.
  * @return true if there is an acl relation between each nodes pair.
  */
  public boolean containsAclLinkTypes() {      
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    boolean result = true;
    if (l_ACLinks == null || l_ACLinks.length == 0) return false;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (!l_ACLinks[i].isTrueForOneAclLinkType() ) {
          result = false;
          break;
      }
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // containsAclLinkTypes()
  
  /**
  * Gets the positions in getEorAllACLinks(), if any, of the AccessControlLinks
  * having the type OWNER. 
  * @return list of indexes. Null if there is no such relation between any nodes pair.
  */
  public List<Integer> getOwnerLinkTypePositions() {      
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    List result = null;
    if (l_ACLinks == null || l_ACLinks.length == 0) return null;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (l_ACLinks[i].getLinkTypeValue(AccessControlLink.OWNER) ) {
          if (result == null) result = new ArrayList(INITIAL_CAPACITY);
          result.add(i);
      }
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // getOwnerLinkTypePositions()
  
  /**
  * Gets the positions in getEorAllACLinks(), if any, of the AccessControlLinks
  * having the type CONTAIN.
  * @return list of indexes. Null if there is no such relation between any nodes pair.
  */
  public List<Integer> getContainLinkTypePositions() {      
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    List result = null;
    if (l_ACLinks == null || l_ACLinks.length == 0) return null;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (l_ACLinks[i].getLinkTypeValue(AccessControlLink.CONTAIN) ) {
          if (result == null) result = new ArrayList(INITIAL_CAPACITY);
          result.add(i);
      } // if
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // getContainLinkTypePositions()
  
  /**
  * Gets the positions in getEorAllACLinks(), if any, of the AccessControlLinks
  * having the types ACL or NEG_ACL.
  * @return list of indexes. Null if there is no such relation between any nodes pair.
  */
  public List<Integer> getAclLinkTypePositions() {      
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    List result = null;
    if (l_ACLinks == null || l_ACLinks.length == 0) return null;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (l_ACLinks[i].isTrueForOneAclLinkType() ) {
          if (result == null) result = new ArrayList(INITIAL_CAPACITY);
          result.add(i);
      } // if
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // getAclLinkTypePositions()
  
  /**
  * Gets the positions in getEorAllACLinks(), if any, of the AccessControlLinks
  * having the type NEG_ACL.
  * Called by DisplayableLinkUtilities.mergeNegativeACLPaths().
  * @return list of indexes. Null if there is no such relation between any nodes pair.
  */
  public List<Integer> getNegativeAclLinkTypePositions() {      
    AccessControlLinkImpl[] l_ACLinks = getEorAllACLinks();
    List result = null;
    if (l_ACLinks == null || l_ACLinks.length == 0) return null;
    for (int i = 0 ; i < l_ACLinks.length ; i++) {
      if (l_ACLinks[i].getLinkTypeValue(AccessControlLink.NEG_ACL) ) {
          if (result == null) result = new ArrayList(INITIAL_CAPACITY);
          result.add(i);
      } // if
    } // for (int i = 0
    l_ACLinks = null;
    return result;
  } // getNegativeAclLinkTypePositions()
  
   /**
   * Gets the 2 AccessControlLinks connected to a BaseObject
   * which this DisplayableLinkImpl contains.
   * Caution: getEorConnectedLinks(_obj)[0] always exists, but not getEorConnectedLinks(_obj)[1].
   * @param _obj is  in this displayable link
   * @return AccessControlLinkImpl[] array of references to the two links
   *   directly connected to _obj, or one link at index 0 or 1, and a null value
   *   if _obj is an end. Is null if _obj is not in this displayable link.
   */
  public AccessControlLinkImpl[] getEorConnectedLinks(BaseObject _obj) {
    int index = vector_.indexOf(_obj);
    if (index == - 1) return null;
    AccessControlLinkImpl[] result = new AccessControlLinkImpl[2];
    if (index > 1) {
      result[0] = (AccessControlLinkImpl)vector_.get(index - 1);
      result[1] = (AccessControlLinkImpl)vector_.get(index + 1);
    } else if (index == 0) {
      result[0] = (AccessControlLinkImpl)vector_.get(2);
      result[1] = null;
    } else if (index == 1) {
      result[0] = (AccessControlLinkImpl)vector_.get(vector_.size() - 1 );
      result[1] = null;
    }
    return result;
  } // getEorConnectedLinks()

   /**
   *
   * @param _link is in this displayable link
   * @return BaseObject[] array of references to the two nodes
   *   directly connected to _link. The first node is at the index 0.
   *   Is null if _link is not in this displayable link.
   */
  public BaseObject[] getEorConnectedNodes(AccessControlLinkImpl _link) {
    int index = vector_.indexOf(_link);
    if (index == - 1) return null;
    BaseObject[] result = new BaseObject[2];
    if (vector_.size() == 3 && index == 2) {
      result[0] = (BaseObject)vector_.get(0);
      result[1] = (BaseObject)vector_.get(1);
    }
    // vector_.size() > 3
    else if (index > 2 && index != vector_.size() - 1) {
      result[0] = (BaseObject)vector_.get(index - 1);
      result[1] = (BaseObject)vector_.get(index + 1);
    } else if (index == 2) {
      result[0] = (BaseObject)vector_.get(0);
      result[1] = (BaseObject)vector_.get(3);
    } else if (index == vector_.size() - 1) {
      result[0] = (BaseObject)vector_.get(vector_.size() - 2 );
      result[1] = (BaseObject)vector_.get(1);
    }
    return result;
  } // getEorConnectedNodes()

   /**
   * Adds a node at the position _index with its two AccessControlLinkImpl.
   * Removes before the AccessControlLinkImpl between the current node at (_index - 2)
   * and the current node at _index, and move this last node to _index + 2.
   * But if _index is 1, the second end is NOT removed, _newNode is put just
   * before the second end, and the last link is removed.
   * No action if the instance is empty. The AccessControlLinkImpls
   * may be known.
   * @param _index int where to add _newNode. Can't be 0 nor pair.
   * @param _newNode BaseObject to add to this displayable link
   * @param _firstLink AccessControlLinkImpl to put between _newNode and the node
   * at _index - 2
   * @param _secondLink AccessControlLinkImpl to put between _newNode and the node
   * at _index + 2, which was at _index before the add operation
   * @exception InternalError if _index is null or pair, if _newNode is known.
   */
  public synchronized void addEorNode(int _index, BaseObject _newNode,
    AccessControlLinkImpl _firstLink, AccessControlLinkImpl _secondLink) {
    if (vector_ == null || vector_.size() == 0) return;
    synchronized (vector_) {
      if (vector_.contains(_newNode) ) {
        throw new InternalError("in DisplayableLinkImpl.addEorNode(),"+
          " the new node is already in the link.");
      }
      // test if _index is 0 or is associated to a node
      if (_index == 0 || _index%2 == 0 ) {
        throw new InternalError("in DisplayableLinkImpl.addEorNode(),"+
          " the index is null or pair.");
      }
      if (_index == 1) {
        // remove the last link
        vector_.remove(vector_.size() - 1);
        // update
        vector_.add(_firstLink);
        vector_.add( _newNode);
        vector_.add(_secondLink);
      }
      if (_index > 2) {
        List afterIndex = vector_.subList(_index, vector_.size() );
        ArrayList result = new ArrayList(vector_.size() + 2);
        result.addAll(vector_.subList(0, _index - 2 ));
        result.set(_index - 1, _firstLink); // use add?
        result.set(_index, _newNode);
        result.set(_index + 1, _secondLink);
        result.addAll(_index + 2, afterIndex);
        vector_ = result;
      }
    } // synchro

                            // for testing and tracking
                            //System.out.println("in DisplayableLinkImpl.addEorNode()"+
                              //" the link at the end is "+this);
  } // addEorNode()

   /**
   * If the argument has the same ends and the same path, and for the activated
   * AccessControlLink types, adds all the rights and the comments of the argument
   * to the relevant AccessControlLinks of this instance.
   * Caution: the DisplayableLinkImpl properties are not merged.
   * @see AccessControlLinkImpl#mergeAccessControlLinkWithoutFiltering
   * @param _dLink to merge
   * @exception InternalError in _dLink has not the same path,
   *  if an end is not the same.
   * @return boolean true if the operation has been processed. If false,
   *   the last merging operation on AccessControlLinkImpl which has been
   *   processed is the false operation.
   */
  public boolean mergeRightsWithoutFiltering(DisplayableLinkImpl _dLink) {
    boolean result = true;
    if (_dLink.getEorFirstEnd() != getEorFirstEnd()) result = false;
    if (_dLink.getEorSecondEnd() != getEorSecondEnd()) result = false;
    if (vector_.size() != _dLink.vector_.size()) result = false;
    if (!result) {
        throw new InternalError("In DisplayableLinkImpl."+
            "mergeRightsWithoutFiltering(), the ends are different or "+
            "the paths have not the same size for "+this+" and for "+_dLink);
    }
    if (!isSimple()) {
        BaseObject[] l_thisNodes = getEorAllNodes();
        BaseObject[] l_dLinkNodes = _dLink.getEorAllNodes();    
        for (int i = 1; i < l_thisNodes.length - 1; i++) {
          if(l_thisNodes[i] != l_dLinkNodes[i]) result = false;
        } // for
        if (!result) {
            throw new InternalError("In DisplayableLinkImpl."+
                "mergeRightsWithoutFiltering(), the paths are not the same for "+
                this+" and for "+_dLink);
        } // if
        l_thisNodes = null;
        l_dLinkNodes = null;
    } // if

    int shift = 0;
    int count = 4;
    if (vector_.size()%2 != 0) shift = (int)0.5;
    AccessControlLinkImpl[] l_links_2 = new AccessControlLinkImpl[
      (int)(vector_.size()/2) - shift];

    result = ((AccessControlLinkImpl)vector_.get(2)).
          mergeAccessControlLinkWithoutFiltering(
           (AccessControlLinkImpl)_dLink.vector_.get(2));
    if (!result) {
      l_links_2 = null;
      return false;
    }
    if (l_links_2.length > 1) {
      for (int i = 1 ; i < l_links_2.length ; i++) {
        result = ((AccessControlLinkImpl)vector_.get(count)).
            mergeAccessControlLinkWithoutFiltering(
               (AccessControlLinkImpl)_dLink.vector_.get(count));
        if (!result) break;
        count += 2;
      }  // for i
    } // if (l_links_2.length > 1)

    l_links_2 = null;
    return result;
  } // mergeRightsWithoutFiltering() 
  
/**
   * If the argument has the same ends and the same path, adds all the comments
   * in the _dLink AccessControlLinks to the relevant AccessControlLinks
   * of this instance, whatever the link type values in each AccessControlLink.
   * Caution: the DisplayableLinkImpl properties are not merged.
   * Called by DisplayableLinkUtilities.addDisplayableLink().
   * @param _dLink
   * @exception InternalError in _dLink has not the same path,
   *  if an end is not the same.
   * @return boolean true in all cases.
   */
  public boolean mergeCommentsWithoutFiltering(DisplayableLinkImpl _dLink) {
    boolean result = true;
    if (_dLink.getEorFirstEnd() != getEorFirstEnd()) result = false;
    if (_dLink.getEorSecondEnd() != getEorSecondEnd()) result = false;
    if (vector_.size() != _dLink.vector_.size()) result = false;
    if (!result) {
        throw new InternalError("In DisplayableLinkImpl."+
            "mergeCommentsWithoutFiltering(), the ends are different or "+
            "the paths have not the same size for "+this+" and for "+_dLink);
    }
    if (!isSimple()) {
        BaseObject[] l_thisNodes = getEorAllNodes();
        BaseObject[] l_dLinkNodes = _dLink.getEorAllNodes();    
        for (int i = 1; i < l_thisNodes.length - 1; i++) {
          if(l_thisNodes[i] != l_dLinkNodes[i]) result = false;
        } // for
        if (!result) {
            throw new InternalError("In DisplayableLinkImpl."+
                "mergeCommentsWithoutFiltering(), the paths are not the same for "+
                this+" and for "+_dLink);
        } // if
        l_thisNodes = null;
        l_dLinkNodes = null;
    } // if

    int shift = 0;
    int count = 4;
    if (vector_.size()%2 != 0) shift = (int)0.5;
    AccessControlLinkImpl[] l_links_2 = new AccessControlLinkImpl[
      (int)(vector_.size()/2) - shift];

    ((AccessControlLinkImpl)vector_.get(2)).
          mergeCommentsWithoutFiltering(
           (AccessControlLinkImpl)_dLink.vector_.get(2));

    if (l_links_2.length > 1) {
      for (int i = 1 ; i < l_links_2.length ; i++) {
        ((AccessControlLinkImpl)vector_.get(count)).
            mergeCommentsWithoutFiltering(
               (AccessControlLinkImpl)_dLink.vector_.get(count));
        count += 2;
      }  // for i
    } // if (l_links_2.length > 1)

    l_links_2 = null;
    return result;
  } // mergeCommentsWithoutFiltering()
  
/**
   * Merges all the GLOBAL comments in the AccessControlLinks of this instance.
   * Calls AccesControlLinkImpl.getLinkTypeComments().
   * @return resulting GLOBAL comments. May be empty but never null.
   */
  public String mergeGlobalComments() {
    StringBuilder buffer = new StringBuilder();
    int         shift = 0;
    int         count = 4;
    if (vector_.size()%2 != 0) 
        shift = (int)0.5;
    
    AccessControlLinkImpl[] l_links = new AccessControlLinkImpl[(int)(vector_.size()/2) - shift];

    buffer.append(((AccessControlLinkImpl)vector_.get(2)).
          getLinkTypeComments(AccessControlLink.GLOBAL));

    if (l_links.length > 1) {
      for (int i = 1 ; i < l_links.length ; i++) {
        buffer.append(" ");

        buffer.append(((AccessControlLinkImpl)vector_.get(count)).
          getLinkTypeComments(AccessControlLink.GLOBAL));

        count += 2;
      }  // for i
    } // if (l_links_2.length > 1)

    String result = buffer.toString();

    l_links = null;
    buffer = null;
    return result;
  } // mergeGlobalComments()
  
/**
   * Gets the general properties of this DisplayableLinkImpl.
   * The first use is the 2 keys AG_CONTEXT_NODE_POSITION and AG_CONTEXT_SET
   * for the Account/Groups context of this link.
   * The AG context is a set of GroupIDMembers that may contain, first,
   * an optional couple (UserID, GroupID), and, second, a list of optional GroupIDs.<P>
   * AG_CONTEXT_NODE_POSITION is the index of a node in the DislayableLinkImpl
   * for which there is a defined Account/Groups context.
   * The key AG_CONTEXT_SET returns a couple (UserID, GroupID)
   * the node at AG_CONTEXT_NODE_POSITION may use as Account/Main_Group context.
   * The key AG_CONTEXT_SET may also return complementary groups.
   * The UserID is generally, but not only, given by the last UserID node
   * in the DisplayableLinkImpl that is the source for an AccessControlLink with
   * the types OWNER. The UserID is get through a chain of links that transfers the AG context.
   * The GroupID is given by the last GroupID node in the DisplayableLinkImpl
   * that is the source for an AccessControlLink with the types CONTAIN,
   * and through a chain of links that transfers the AG context. <P>
   * The list of GroupIDs depends on the node giving its proper context,
   * if the chain of links transfers the AG context.
   * If the first end is a GroupIDMember, the list contains the groups to which
   * this GroupIDMember is member, if any. If the first end is an Actor, the list
   * may contain the secondary groups of this Actor, if any.
   * If the chain of links does not transfer the first end AG context,
   * the list of GroupIDs is the secondary groups of the last intermediate Actor
   * starting from no previous transfered AG context.
   * The values in the retured map may be an ImmutableName, an Integer, a HashSet,
   * an ArrayList or anything else.
   * @see ARoad0.gWork.DisplayableLinkUtilities#getSecondEndAGcontext(ARoad0.CNot.DisplayableLinkImpl)
   * @see ARoad0.gWork.LinkRightsImpl#setNextNodeAGrunningContext(ARoad0.CNot.DisplayableLinkImpl, int) 
   * @since 0.7.0
   * @return the immutable DisplayableLinkImpl properties, or null
   */
  public Map<String, Object> getProperties() {
      if (m_properties_ != null)
          return Collections.unmodifiableMap(m_properties_);
      else
          return null;
  } // getProperties()

/**
 * Sets a new value for a known key or a new key.
 * The first use is the 2 keys AG_CONTEXT_NODE_POSITION and AG_CONTEXT_SET
 * for the Account/Groups context of the link.
 * @see ARoad0.gWork.DisplayableLinkUtilities#getSecondEndAGcontext(ARoad0.CNot.DisplayableLinkImpl)
 * @since 0.7.0
 * @param _key the key for the new property
 * @param _value the new value may be a Name, a Set or another object
 * @return the previous value associated with _key, or null if there was no mapping for _key
 * or if an argument is null, or if the value for the key AG_CONTEXT_SET is not a set
 * of ImmutableGroupIDMembers, or if the value for the AG_CONTEXT_NODE_POSITION is
 * not an Integer as position of the node having the AG context, or if it is a value
 * lesser than the previous Integer value
 */
  public Object addProperty(String _key, Object _value) {
      if (_key == null || _value == null)
          return null;
      if (_key.equals(AG_CONTEXT_NODE_POSITION)) {
          if (!(_value instanceof Integer))
            return null;

          BaseObject[] l_nodes = getEorAllNodes();
          boolean found = false;
          boolean previousNodeFound = false;
          Integer previousNode = null;

          if (m_properties_ != null)
              previousNode = (Integer) m_properties_.get(AG_CONTEXT_NODE_POSITION);
          
          for (int i = 0; i < l_nodes.length; i++) {
              // previousNode may be _value
              if (previousNode != null && i == previousNode.intValue()) {
                  previousNodeFound = true;
              }
              if (i == ((Integer)_value).intValue()) {
                  found = true;
                  break;
              }
          } // for i
          
          if (!found)
              return null;
          if (previousNode != null && !previousNodeFound)
              return null;

          l_nodes = null;
          previousNode = null;
      } // if
      else if (_key.equals(AG_CONTEXT_SET)) {
          boolean error = false;

          if (!(_value instanceof Set))
            return null;
          
          for (Object ob : (Set)_value) {
              if (!(ob instanceof ImmutableGroupIDMember)) {
                  error = true;
                  break;
              }
          } // for ob
          
          if (error)
              return null;
      } // else

      if (m_properties_ == null)
          m_properties_ = new HashMap(INITIAL_CAPACITY);
      Object ret = m_properties_.put(_key, _value);
      return ret;
  } // addProperty()
  
/**
 * Removes a known key.
 * The first use is the 2 keys AG_CONTEXT_NODE_POSITION and AG_CONTEXT_SET
 * for the Account/Groups context of the link.
 * @since 0.7.0
 * @param _key the key for the property to remove
 * @return the previous value associated with _key, or null if there was no mapping for _key
 */
  public Object removeProperty(String _key) {
      if (m_properties_ == null)
          return null;
      Object ret = m_properties_.remove(_key);
      return ret;
  }

 /**
 * Gets a clone of a displayable link, with a shallow copy of each node and
 * value in the properties map, and a deep copy of each AccessControlLink.
 * @exception CloneNotSupportedException
 * @return a cloned and no-backed instance of displayable link. May be null.
 */
  public Object clone() {
    try {
      DisplayableLinkImpl clon = (DisplayableLinkImpl) super.clone();
      synchronized (vector_) {
        if (vector_ == null) clon.vector_ = null;
        else {

          int index = 0;
          Object obj;
          Map.Entry entry = null;

          clon.vector_ = new ArrayList(vector_.size());
          for (Iterator ite = vector_.iterator(); ite.hasNext() ; ) {
            obj = ite.next();
            if (index == 0) {
              clon.vector_.add(index, obj);
            }
            // deep copy of all the Access Control Links
            else if (index%2 == 0) {
              clon.vector_.add(index, ((AccessControlLinkImpl)obj).clone() );
            }
            // shallow copy of all the nodes after the first end
            else if (index%2 != 0) {
              clon.vector_.add(index, obj );
            }
            index++;
          } // for ite

          if (m_properties_ != null) {
              clon.m_properties_ = new HashMap(20);
              for (Object ob : m_properties_.entrySet()) {
                  entry = (Map.Entry)ob;
                  clon.m_properties_.put(
                          (String)entry.getKey() ,
                          entry.getValue());
              } // for ob
          } // if

        } // else
      } // synchro
      return clon;
    } catch (CloneNotSupportedException e) {
      // this shouldn't happen, since we are Cloneable
      throw new InternalError(e.toString() + "DisplayableLinkImpl.clone() failed");
    } // catch
  } // clone()

  /**
   * Two DisplayableLinkImpls are equals if and only if
   * the sequences of nodes are identical, and all the nodes are identical,
   * and not simply equal.
   * The AccessControlLinks and the properties are not implied.
   * @param _obj the object to compare
   * @return true if the argument is equal
   */
  public synchronized boolean equals(Object _obj) {
    // variables
    boolean result = true;

    if (this == _obj)
      return true;
    if (_obj == null)
      return false;
    if (_obj.getClass() != getClass())
      return false;

    if (vector_ == null || ((DisplayableLinkImpl)_obj).vector_ == null) {
    if (vector_ != null || ((DisplayableLinkImpl)_obj).vector_ != null )
      return false;
    else return true; // if the two are null
    } // if
    else if (vector_.size() != ((DisplayableLinkImpl)_obj).vector_.size())
      return false;

    // derived from the getEorAllNodes() code
    if (vector_.get(0) != ((DisplayableLinkImpl)_obj).vector_.get(0))
        return false;
    if (vector_.get(1) != ((DisplayableLinkImpl)_obj).vector_.get(1))
        return false;

    BaseObject[] l_nodes = new BaseObject[(int)vector_.size()%2 +
      (int)vector_.size()/2];

    if (l_nodes.length > 2) {
      // count set to the index of the first intermediate node
      int count = 3;
      for (int i = 1 ; i < l_nodes.length - 1 ; i++) {
        if (vector_.get(count) != ((DisplayableLinkImpl)_obj).vector_.get(count)) {
            result = false;
            break;
        }
        count += 2;
      } // for (int i = 1
    } // if (l_nodes.length > 2)
    l_nodes = null;

    //if (!result) return false;

    // derived from the getEorAllACLinks() code
    //int shift = 0;
    //if (vector_.size()%2 != 0) shift = (int)0.5;
    //AccessControlLinkImpl[] l_links = new AccessControlLinkImpl[
      //(int)(vector_.size()/2) - shift];

    //if (!((AccessControlLinkImpl)vector_.get(2)).equals(
      //     (AccessControlLinkImpl)((DisplayableLinkImpl)_obj).vector_.get(2) ))
        //result = false;

    //if (l_links.length > 1 && result) {
      //int count = 4;
      //for (int i = 1 ; i < l_links.length ; i++) {
        //l_links[i] = (AccessControlLinkImpl)vector_.get(count);
        //if (!((AccessControlLinkImpl)vector_.get(count)).equals(
           //(AccessControlLinkImpl)((DisplayableLinkImpl)_obj).vector_.get(count) )) {
             //   result = false;
               // break;
        //} // if
        //count += 2;
      //}  // for i
      //l_links = null;
    //} // if (result.length > 1)

    return result;
  } // equals(Object)

  /**
   * Based on the hashCode nodes.
   * @return the hashcode value for this instance. May be 0.
   */
  public int hashCode() {
    int result = 0;
    BaseObject[] l_nodes = getEorAllNodes();
    if (l_nodes != null) {
      for (int i = 0; i < l_nodes.length; i++) {
        result = result*PRIME + l_nodes[i].hashCode();
      }
    } // if
    return result;
  } // hashcode()


  /**
   * Finalizes the AccessControlLinkImpl associated to this instance,
   * and clears the list vector_.
   */
  public synchronized void finalizeForGui() {
    AccessControlLinkImpl[] l_acLinks = getEorAllACLinks();
    if (l_acLinks != null) {
      for (int i = 0; i < l_acLinks.length; i++) {
        l_acLinks[i].finalizeForGui();
      }
    } // if
    if (vector_ != null) {
      vector_.clear();
      vector_ = null;
    } // if
  } // finalizeForGui()

  /**
   * @return long serial version UID for serialization of the class
   */
  public long getSerialVersionUID() {
    return serialVersionUID__;
  }

   /**
   * Returns the list of the nodes and links between them, where the two first
   * objects are the first end and the second end.
   * An exemple of returned String, from the MySQL AcsAddon, is: <P>
   * tom  |localhost|  ACL: EXECU;  (<inherited ACL> EXECUTE conditional ACL via
   * |AllSimpleHosts| sets Group)    Amounts view alias  Runs Under, <P>
   * to desccribe a link where the suite of 3 nodes is ('tom', 'Amounts view alias', '|localhost|'),
   * where the first AccessControlLink from 'tom' to 'Amounts view alias' is
   * described by 'ACL: EXECU;  (<inherited ACL> EXECUTE conditional ACL via
   * |AllSimpleHosts| sets Group)', and the last AccessControlLink is, as always,
   * at the end: 'Runs Under' from 'Amounts view alias' to '|localhost|'.
   * Calls AccessControlLinkImpl.toString().
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    if (vector_ == null) {
        sb.append(" null ");
    } 
    else {
        int p = vector_.size();
        sb.append( vector_.get(0) ); //((BaseObject)vector_.get(0)).getName() );

                            // for testing and tracking
                            //System.out.println("in DisplayableLinkImpl.toString(), vector size is"+
                              //      p+" and vector_.get(0): "+vector_.get(0));
        
        sb.append("  ");
        sb.append( vector_.get(1) );

                            // for testing and tracking
                            //System.out.println("in DisplayableLinkImpl.toString(), vector size is"+
                              //      p+" and vector_.get(0): "+vector_.get(0));
        
        sb.append("  ");
        for (int i = 2; i < p; i++) {
            sb.append(vector_.get(i));
            
                            // for testing and tracking
                            //if (vector_.get(i) instanceof AccessControlLinkImpl) {
                              //  sb.append("NEG_ACL: ");
                                //sb.append(((AccessControlLinkImpl)vector_.get(i)).
                                  //      getLinkTypeValue(AccessControlLink.NEG_ACL));
                            //}

                            // for testing and tracking
                            //System.out.println("in DisplayableLinkImpl.toString(), vector_.get(i): "
                              //+vector_.get(i)+" for i = "+i);
            
            sb.append("  "); // adds each node and link in sb
        } //for        
    } // else
    return sb.toString();
  } // toString()


} // DisplayableLinkImpl