//Title:        Access Road
//Version:      0.7.0
//Copyright:    2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui1;

import java.beans.*;
import javax.swing.*;
import java.util.List;
import java.util.ArrayList;

import ARoad0.ErrW.ProcessError;
import ARoad0.ErrW.SmallError;
import ARoad0.gBaseInterface.BaseObject;

import java.lang.reflect.InvocationTargetException;

/**
 * This class is responsible for creating, by the mean of panels, associations
 * to do as calls to constructors and methods in gBase, with all the required arguments.<P>
 * It is used by ACSFactoryImpl to define the list of associations to perform
 * when a new object is created in an ACS, and to define a list of creations
 * when a child ACS is opening, to update its ACSObject roots in its parent ACS.<P>
 * This class reads and updates a modifiable property map
 * of a gBase JavaBean, for which the type is a Map<String,String[]>,
 * and where the map values are the associations. <P>
 * The map keys have the format 'keyStarting.subkey' as criteria for defining
 * the new objects which comply to the required associations by the acs designer. <P>
 * This class works with gBase.AssociateUtilityImpl strongly for the values updating.
 * The map updating is never done through a write method or a 'choiceMethodForNew' in
 * the property descriptor, but through 'add/remove' methods defined in the property
 * descriptor.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 */
public class AssociationMapPropertyEditor extends StringTwoKeysMapPropertyEditor {

  /**
   * The key to use as argument in the call of the limited values method.
   * On the contrary of the super class, the argument is not the currently selected
   * key in the map. The argument is static, and the limited values  method returns
   * only the first value of the values array. The set of the first values is smaller
   * when this class is the property editor of a child ACS. To detect this case,
   * the BeanInfo must have a 'ForChildACS' property which initializes this property.
   * In other cases, this property should remains null.
   */
  protected String                                 limitedValuesMethodKey_ = null;
  
  //-----------------------------------------------------------------------

  /**
   * Constructor FOR USE ONLY through the valueOf() method,
   * EMPTY_INSTANCE and the subclasses.
   */
  AssociationMapPropertyEditor() {
    displayName_ = "";
  }

  /**
   * Factory method to use when a AssociationMapPropertyEditor is delegating to us.
   * It constructs the read method for the property, from the BeanInfo information.
   * The property must be a map of Strings as keys, and String arrays as values.
   * Overrides the super method and calls it.
   * Called by BeamerManager.addPropertyEditor().
   * @param _source owning the property.
   * @param _propertyDescriptor is the descriptor of the property
   * @exception NullPointerException if _source or _displayName is null
   * @exception ProcessError if BeanInfo and PropertyDecriptor calls fail
   * @return a AssociationMapPropertyEditor instance
   */
  public static CommonPropertyEditor valueOf(BaseObject _source,
    PropertyDescriptor _propertyDescriptor)
    throws NullPointerException, ProcessError {

      AssociationMapPropertyEditor result =
        (AssociationMapPropertyEditor) StringTwoKeysMapPropertyEditor.valueOf(
          _source, _propertyDescriptor);

            // for testing purpose
             //System.out.println("AssociationMapPropertyEditor.valueOf() "+
               //      "after StringTwoKeysMapPropertyEditor.valueOf(), for: "+
                 //       result.getDisplayName());
            

      if (result.propertyDescriptor_.getValue("limitedValuesMethodKey") != null) {
          result.limitedValuesMethodKey_ = (String)result.propertyDescriptor_.
                  getValue("limitedValuesMethodKey");
      }
      return result;
  } // valueOf()

  // ----------------------- PUBLIC METHODS ---------------------------------

  /**
   * Get the values of variables for the instance.
   *
   * @return source, propertyName and listeners (if they have toString() methods).
   */
  public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("  ASSOCIATION_MAP_PROPERTY_EDITOR.TOSTRING():");

        sb.append(" source_ = ");
        sb.append(source_);

        sb.append("; display name = ");
        sb.append(displayName_);

        sb.append("; listeners list = ");
        sb.append(listeners_);

        return " [" + sb + "] ";
  } // toString()

  // -------------------- PROTECTED METHODS ----------------------------------- 

  /**
   * Called by actionPerformed() when the user clicks on the button 'NewV' to
   * add new values in the map.
   * The allowed String map values depends lightly on the map key.
   * The values define one or several actions to do if the key matches.
   * The possible formats for the first value of the action N, are get
   * from the method limitedValuesMethod_ among this set: <P>
   * - <AssociateClass>.XXX which specifies the class name XXX in gBase, without the Impl String,<P>
   *    of an associate object to build up or to finalize when the master is build up <P>
   *    or finalized <P>
   * - <CalledMethod>.<Master>.YYY which specifies the method YYY to call on the master, <P>
   * - <CalledMethod>.<ACS>.YYY which specifies the method YYY to call on the master ACS, <P>
   * - <CalledMethod>.<Action>.C.ZZZ which specifies the method ZZZ <P>
   *    to call on the associate previously created by the action number C. <P>
   * - <CalledMethod>.<Object>.XXX+<Value>+ZZZ which specifies the method ZZZ <P>
   *    to call on the object having the name XXX, where the parameters must
   *    be the master or an associate to set the reverse operation when it is finalized. <P>
   * After each first value, there is the optional list of the constraints on the arguments
   * of the constructor (case 1) or the method (cases 2 to 5),
   * and there is one constraint specification per argument.
   * Each argument constraint specification must belong to a set of 10 allowed formats,
   * presented herein without the number 'N.' at the start, for the action N: <P>
   * - 'PropertyName'.<Null>: null value for the argument having the name 'PropertyName' <P>
   * - 'PropertyName'.<Master>: the master is the argument <P>
   * - 'PropertyName'.<MasterCreation>: boolean true if the master is creating, 
   *        false if it is finalizing <P>
   * - 'PropertyName'.<ArgumentValue>.C: the associate or argument number C object <P>
   * - 'PropertyName'.<Value>.XXX: the argument is XXX as String or ImmutableName if it is 
   *        the property type, or any BaseObject get from its name <P>
   * - 'PropertyName'.<Value>.XXX+<MasterValue>+YYY: the master value <P>
   *        for the same property in its BeanInfo (the read method has to not return an array),
   *        with some optional pre and post String values
   *        if the property type is String <P>
   * - 'PropertyName'.<Value>.XXX+<MasterNameLastComponent>+YYY: <P>
   *        the last component in the master name, with some optional  <P>
   *        pre and post Name values as Strings <P>
   * - 'PropertyName'.<Value>.<MasterNameContext>+YYY <P>
   *        the context in the master name, with a post Name value as String <P>
   * - 'PropertyName'.<Value>.<ACSName>+YYY: the master ACS name, <P>
   *        with a post Name value as String <P>
   * - 'PropertyName'.<Value>.XXX+<ChildNameLastComponent>+YYY: <P>
   *        the last component in the child ACS name, with some optional
   *        pre and post Name values as Strings. <P>
   * If the last value is the gBase message 'Error: lack of arguments', allow
   * the adding of value(s). If it starts with 'Error to delete', request to the user
   * a removing of the current values. Detect the cases when a first-format value
   * has to be added, and when there is a lack of arguments to add.
   * Calls StringMapPropertyEditor.addValueElements().
   * Overrides the StringMapPropertyEditor method.
   * @see ARoad0.gBase.AssociateUtilityImpl#addAssociatesAfterCreation
   * @param _key current selected key 
   */
  protected void addRestrictedValueElements(String _key) {

                          // for testing and tracking
                     //System.out.println("in AssociationMapPropertyEditor.addValueElements(), "+
                       //     "the property type is "+propertyTypeInGroup_);

    // this method is reserved to the constrained values
    if (limitedValuesMethod_ == null) return;

    List<String>    l_limitedFirstValues = null;

    try {
      l_limitedFirstValues = (List)limitedValuesMethod_.invoke(
              source_, new Object[] {limitedValuesMethodKey_} );
    } catch (InvocationTargetException i) {
          // show the message to the user
          int lastIndex = ((InvocationTargetException)i).getTargetException().
                  toString().lastIndexOf(": ");
          if (lastIndex != 0)
            DialogBox.showErrorMessage(
              ((InvocationTargetException)i).getTargetException().toString().
                  substring(lastIndex+1));
          else DialogBox.showErrorMessage(
            ((InvocationTargetException)i).getTargetException().toString());

          System.out.println("in AssociationMapPropertyEditor.addValueElements(), "+
            "the exception is an InvocationTargetException: ");
          ((InvocationTargetException)i).printStackTrace();

          // restore the current value in the editor
          propertyChange(new PropertyChangeEvent(source_,
                  propertyDescriptor_.getName(), null, null));
          return;
    } catch (Exception e) {
          System.out.println("AssociationpPropertyEditor.addValueElements(), exception: "+e);
    } // catch

    String[]        l_formats = l_limitedFirstValues.toArray(new String[0]); 
    List<String>    l_values = new ArrayList(INITIAL_CAPACITY);    

                          // for testing and tracking purpose
                     //System.out.println("AssociationMapPropertyEditor.addValueElement() for "+_key+
                       //     ", the limited values are:");
                     //for (Iterator ite = l_values.iterator(); ite.hasNext();) {
                       //  System.out.println("       "+ite.next());
                     //} // for ite
    
                     
    ListModel       l_currentValues = ((JList)boundComponentTwo_).getModel();

    String          text = "";
    int             firstIndex = -1;
    boolean         error = true;
    boolean         firstValueFormatInLastValue = false;
    String          lastValue = null;
    
    if (l_currentValues.getSize() > 0) {
        lastValue = (String)l_currentValues.getElementAt(l_currentValues.getSize() - 1);
    }
    if (lastValue != null && !lastValue.startsWith("Error to delete") && 
            l_currentValues.getSize() > 1) {    
        for (int i = 0; i < l_formats.length; i++) {
            if ( ((String)l_currentValues.getElementAt(l_currentValues.getSize() - 2)).
                    startsWith(l_formats[i])) {
                firstValueFormatInLastValue = true;
                break;
            } // if
        } // for i
    } // if
    
    // the last current value is null or numbered, so add a first value among
    // the five allowed formats
    if (lastValue == null || 
            (lastValue.length() > 2 && lastValue.substring(1, 2).equals(".")) ||
            (lastValue.length() > 3 && lastValue.substring(2, 3).equals(".")) ) {

        try {

                          // for testing and tracking
                     //System.out.println("in AssociationMapPropertyEditor.addValueElements(), "+
                       //     "for "+propertyTypeInGroup_+", call DialogBox.showOneSelectionOneDialog() with"+
                         //   " the key "+_key);
                     
            firstIndex = DialogBox.showOneSelectionOneDialog(
                "value type for "+_key, l_formats);
            
            if (firstIndex == -1) {
                l_currentValues = null;
                l_values = null;
                l_formats = null;
                return;
            }
            
            if (firstIndex == 0)
                text = DialogBox.showSingleEditor("Enter the class name");
            else if (firstIndex == 3)
                text = DialogBox.showSingleEditor("Enter action_number.method");
            else if (firstIndex == 4)
                text = DialogBox.showSingleEditor("Enter object+<Class>+object_class+<Value>+method");
            else
                text = DialogBox.showSingleEditor("Enter the method name");
            
            if (text == null) {
                l_currentValues = null;
                l_values = null;
                l_formats = null;
                return;
            }
            
            l_values.add(l_formats[firstIndex]+text); 
        } catch (SmallError s) {
            throw new InternalError("ActionMapPropertyEditor.addValueElements()"+s.toString());
        } // catch          
        
    } // if (l_currentValues_
    else if (firstValueFormatInLastValue || lastValue.equals("Error: lack of arguments")) {
        // the value before the last current value is not numbered and is a first value, so
        // add the matched set of arguments through a suite of user enterings,
        // or add the other arguments
        while (text != null ) {
            text = DialogBox.showSingleEditor("Enter a new argument, or Close");
            error = true;
            if (text != null &&
                    (text.endsWith(".<Null>") ||
                    text.endsWith(".<Master>") ||
                    text.endsWith(".<MasterCreation>") ||
                    text.indexOf(".<Value>.") != -1 )) 
                            error = false; 
            if (!error)
                l_values.add(text);
            else if (text != null)
                DialogBox.showErrorMessage("The last value is wrong.");
        } // while
    } // else
    else if (lastValue != null && lastValue.startsWith("Error to delete:")) {
        DialogBox.showErrorMessage("The cause of error should be removed.");
        l_currentValues = null;
        l_values = null;
        l_formats = null;
        lastValue = null;
        return;
    }

                          // for testing and tracking purpose
                     //System.out.println("AssociationMapPropertyEditor.addValueElements() for "+
                       //     _key+", the part values are:");
                     //for (Iterator ite = l_values.iterator(); ite.hasNext();) {
                       //  System.out.println("       "+ite.next());
                     //} // for ite
                     
    addValueElements(l_values.toArray());
    
    l_values = null;
    l_currentValues = null;
    l_formats = null;
    lastValue = null;
    l_limitedFirstValues = null;
 } // addRestrictedValueElements()

  
} // AssociationMapPropertyEditor


