//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui1;

import java.beans.*;
import javax.swing.*;
import java.util.Map;
import java.util.List;
import java.util.Iterator;
import java.util.HashSet;
//import java.util.Collections;
//import java.util.Arrays;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Dimension;
import java.awt.Component;
import javax.swing.border.BevelBorder;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import ARoad0.gWork.ACS_Facade;
import ARoad0.Pattern.BeanInfoPattern;
import ARoad0.Pattern.MVC_Controller;
import ARoad0.Pattern.MVC_View;
import ARoad0.gBaseInterface.*;
import ARoad0.gBase.ResourceImpl;
import ARoad0.gBase.UtilityImpl;
import ARoad0.gBase.StringRightImpl;
import ARoad0.gBase.BaseUtilityImpl;
import ARoad0.gBase.PrivilegeForTypeImpl;

import ARoad0.gDMak.ActionNewACSyst;

import ARoad0.ErrW.SmallError;
import ARoad0.ErrW.ProcessError;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

/**
 * This important class is responsible for creating or copying
 * a BaseObject instance, but not an ACS nor a view.
 * In the MVC pattern, this is the controler that handles all user dialogs
 * to define the relevant parameters.
 * For creating or copying, the BaseObject class methods to use
 * are declared in the BeanInfo class (the model in the MVC pattern)
 * which is associated to the gBase class in the gBaseBeanInfo package.
 * User dialogs are handled by DialogBox (the View in MVC).
 * It allows to create a base object both:
 * - through its constructor,
 * - through the 'add' method of another object (for example, AclEntry from Resource),
 * - or through its copy method.
 * The selection depends on the valueOf method which is used, among three,
 * to build up the correct BaseObjectHandler instance. The BaseObjectHandler
 * constructor is to be avoided. <P>
 * The general logic of this class is a mix of BeamerManager and SimplePropertyEditor.
 * The creator of the BaseObjectHandler instance has to be 
 * a NewBaseObjectListener for using a constructor or a copy method,
 * or the subinterface AddRemoveBaseObjectListener for using an 'add' method.
 * In the first case, a handleNewBaseObject() method is called by this instance
 * after the new BaseObject creation or copying. In the second case,
 * an addOneBaseObject method is called back.
 * It handles preset, constrained, optional, uncomplete or mandatory arguments,
 * and checks the ACS properties of an ACSObject to control the entered values. <P>
 * This class is used by SimplePropertyEditor.actionPerformed(),
 * CollectionPropertyEditor.addElement(), ActionNewResource,
 * ActionNewEligibleParty, ActionCopy.
 * This class displays and manages the panels which are required to
 * enter all the parameters of the new BaseObject.
 * The number of parameters has to be strictly inferior to 8.
 * It allows the user to choose a parameter:
 *  - among any open base objects, through the explorer (the second-level dialog
 *    to handle is then controlled by this class),
 *  - among the allowed values, if the parameter is constrained by the ACS or the
 *    base object BeanInfo class,
 *  - or to create it, if it is a boolean, a string, a name.
 * It does not allow to create a parameter from scratch if it is a BaseObject;
 * this type of parameter has to be created before the call to BaseObjectHandler.
 * When the argument is a relative name to concatenate with the name of
 * another argument (the ACS or the directory name, for example), the root argument
 * must be a BaseObject.
 * The panel for a parameter may be dynamically changed in the user windows
 * when the presence of a constraint on the values for a parameter
 * depends on the value of the ACS, and when the ACS value is a parameter
 * the user has to choice (example: 'Acl Rights' at the creation of the AclEntry). 
 * <P>
 * Each parameter may be constrained through the ACS, with
 * a limited set of values, including the control of a null argument.
 * Each parameter may be optional, and then, it is not used to create
 * the base object instance if it is null.
 * When the parameter is an array, only one-value array is created as parameter
 * of the constructing method.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.Gui1.SimplePropertyEditor
 * @see ARoad0.Gui1.CollectionPropertyEditor
 */
public class BaseObjectHandler implements MVC_Controller, MVC_View, ActionListener,
        SelectionDialogListener, CreationDialogListener {

  /**
   * Default capacity for collections and maps
   */
  protected static final int        INITIAL_CAPACITY = 10;
  /**
   * Name dimension for all arguments
   */
  protected static final Dimension NAME_DIMENSION = new Dimension(250, 15);
  /**
   * This index is used for the buttons in the creation dialog
   */
  protected final static String[]   L_INDEX =
    new String[] {"1", "2", "3", "4", "5", "6", "7"};

  /**
   * This is the creator of this instance. It is called back by this instance.
   * It is the subinterface AddRemoveBaseObjectListener if an 'add/remove' method is used.
   */
  protected NewBaseObjectListener   caller_;
  /**
   * When it is new BaseObject to create through the call of an 'add' method,
   *   this is the descriptor of the property. Otherwise, it is always null.
   */
  protected  PropertyDescriptor     propertyDescriptor_;
  /**
   * The class of the new BaseObject value to create. Cannot be an ACS.
   */
  protected  Class                  newBaseObjectClass_;
  /**
   * The beanInfo instance associated to newBaseObjectClass_
   */
  protected BeanInfoPattern         newBaseObjectBeanInfo_;
  /**
   * The arguments values array to use for creating 
   *  the new instance of a BaseObject.
   */
  protected Object[]                l_argumentsValues_;
  /**
   * The classes array of the arguments to use for creating the new instance
   * of a BaseObject. To select an item in the explorer when the argument
   * is an array of items (like StringRight[] for new Privileges),
   * the class is not the array but it is the item class (like StringRight).
   */
  protected  Class[]                l_argumentsClass_;
  /**
   * The names array of the arguments
   *  for the constructor of the new instance, in the same order used
   *  in l_argumentsClass_. These names are properties of the BaseObject
   *  to create.
   */
  protected String[]                l_argumentsNames_;
  /**
   * The JComponent (JTextField, JComboBox) array
   *  to use for displaying the arguments.
   * May be larger than l_argumentsClass_ if the source
   * is displayed when an 'add' method is used.
   */
  protected JComponent[]            l_argumentsJComponent_;
  /**
   * The JPanel array to use for displaying the arguments.
   * May be larger than l_argumentsClass_ if the source
   * is displayed when an 'add' method is used.
   * An item may be null.
   */
  protected JPanel[]                l_argumentsJPanel_;
  /**
   * The Boolean array
   *  to use for selecting or not the value, from the explorer or through a special dialog.
   */
  protected Boolean[]               l_argumentsFromSelect_;
  /**
   * The Int array of the name arguments which are relative and are
   * to concatenate with names from another arguments
   */
  protected Integer[]               l_argumentsConcatening_;
  /**
   * The Boolean array to allow or not a null argument.
   */
  protected Boolean[]               l_nullArgumentPossible_;
  /**
   * The Boolean array of optional argument. If true,
   *  do not use an argument in the constructor when this argument is null.
   */
  protected Boolean[]               l_optionalArguments_;
  /**
   * The Boolean array of restricted argument. If true,
   *  allows only the restricted values.
   */
  protected Boolean[]               l_restrictedArguments_;
  /**
   * The Boolean array of non visible arguments. If true,
   *  the argument has to be not displayed and it is set to 'null' in the
   *  constructor. Non null only for constructors through the first valueOf().
   */
  protected Boolean[]               l_nonVisibleArguments_ = null;
  /**
   * The String array of user messages for restricted arguments. If not null,
   * the message is displayed in the property editor at its opening.
   * Exemple of value: 'ACS', and displayed message: <Set the ACS first>.
   * 'ACS' and 'Strength' are the two possible values.'
   * Not null only for constructors through the first valueOf().
   */
  protected String[]               l_restrictedFromOtherArgumentMessage_ = null;
  /**
   * The index in l_argumentsValues_ of the argument currently processed,
   * between '0' and l_argumentsValues_.length - 1.
   * Equal to '99' if no argument is processed.
   */
  protected int                     argIndex_ = 99;
  /**
   * The BaseObject instance to copy. Null if the copy method is not used.
   */
  protected Copyable                objectToCopy_;
  /**
   * The property names displayed in the panel.
  */
  protected String[]                l_displayNames_;
  /**
   * True when getBaseObjectEditor() has called DialogBox for entering the user
   * choices, until the call back to handleCreationDialog().
  */
  protected boolean                onGoingCreationDialog_ =false;  
  
   /**
   * Constructor FOR USING ONLY through the valueOf() methods.
   */
  private BaseObjectHandler() {
  }

  /**
   * First of the 3 valueOf methods, to create a new BaseObject through its CONSTRUCTOR,
   * as it is described in the bean descriptor in the relevant BeanInfo class.
   * Cannot create ACS and View instances.
   * Called by SimplePropertyEditor, CollectionPropertyEditor.actionPerformed(),
   * ActionNewEligibleParty, ActionNewResource.
   * With this method, propertyDescriptor_ is always null.
   * @param _caller of this method is a NewBaseObjectListener with a
   *  handleNewBaseObject() method that is called back after the instance creation.
   *  Cannot be null.
   * @param _baseObjectClass is the class of the new instance of BaseObject to create.
   *  Used to find the right BeanInfo class. Cannot be null.
   * @return the BaseObjectHandler instance with the correct initialization
   * @exception if _caller is null or if _baseObjectClass is not correct
   */
  public static BaseObjectHandler valueOf(NewBaseObjectListener _caller,
    Class _baseObjectClass) {

      // argument control
      if (_caller == null || _baseObjectClass == null) {
         throw new NullPointerException();
      }
      if (ImmutableACS.class.isAssignableFrom(_baseObjectClass)) {
        throw new InternalError("In BaseObjectHandler,"+
          " it is not possible to create an ACS.");
      } // if
      if (ViewInBase.class.isAssignableFrom(_baseObjectClass)) {
        throw new InternalError("In BaseObjectHandler,"+
          " it is not possible to create a view.");
      } // if
      
      // initialization of the BaseObject BeanInfo
      BaseObjectHandler result = new BaseObjectHandler();
      // common intialization from the BeanInfo
      result = commonValueOf(result, _baseObjectClass, true);

      // initialization
      result.caller_ = _caller;
      
      String[] l_nonVisible = (String[]) result.newBaseObjectBeanInfo_.
              getBeanDescriptor().getValue("constrNonVisibleArgument");
      if (l_nonVisible != null) {
          result.l_nonVisibleArguments_ = new Boolean[l_nonVisible.length];
          for (int i = 0; i< l_nonVisible.length; i++) {
              if (l_nonVisible[i] == null)
                  result.l_nonVisibleArguments_[i] = Boolean.FALSE;
              else if (l_nonVisible[i] instanceof String) {
                  if (_caller.getSource() instanceof ACS)
                      result.l_nonVisibleArguments_[i] = !BeamerManager.getACSboolean(
                          (ACS)_caller.getSource(), l_nonVisible[i]);
                  else if (_caller.getSource() instanceof ACSObject)
                      result.l_nonVisibleArguments_[i] = !BeamerManager.getACSboolean(
                         (ACS)ACS_Facade.getOpenACS().get(
                            ((ACSObject)_caller.getSource()).getAcsName()), 
                         l_nonVisible[i]);
                  
                      // for testing and tracking
                  //System.out.println("in BaseObjectHandler.valueOf() for new, "+
                    //"l_nonVisibleArguments_[i] is set for "+i+" with the value "+
                    //result.l_nonVisibleArguments_[i]);
                  
              } // else
          } // for i
          l_nonVisible = null;
      } // if l_nonVisible

      result.l_argumentsValues_ = new Object[result.l_argumentsClass_.length];

      // for testing and tracking
      result.l_argumentsValues_[0] = "<uninitialized>";

      result.l_argumentsJComponent_ =
        new JComponent[result.l_argumentsClass_.length];
      
      result.l_argumentsJPanel_ = new JPanel[result.l_argumentsValues_.length];

      if (result.l_argumentsClass_.length != result.l_argumentsNames_.length) {
        throw new InternalError("In BaseObjectHandler,"+
          " l_argumentsNames_ has not the l_argumentsClass_ length.");
      } // if
      if (result.l_argumentsClass_.length > 7) {
        throw new InternalError("In BaseObjectHandler,"+
          " the number of arguments is superior to 7.");
      } // if

                  // for testing and tracking
              //System.out.println("in BaseObjectHandler.valueOf() for new, "+
                //"the instance is: "+result);
              //System.out.println("in BaseObjectHandler.valueOf() for new, "+
               //"l_argumentsNames_ is: ");
              //for (int i = 0 ; i < result.l_argumentsNames_.length; i++) {
                  //System.out.println(" -------- "+result.l_argumentsNames_[i]);
              //}
                
      return result;
  } // valueOf()

  /**
   * Second of the 3 valueOf methods, to create a new BaseObject through
   * an 'ADD' method which is described in a property descriptor.
   * Intializes the 'add' method arguments to use for adding an instance
   * to the collection in invoking a method on an another BaseObject instance.
   * Cannot create ACS and View instances.
   * On the contrary of the other valueOf methods,
   * propertyDescriptor_ is not null, and the instance creation is done in _caller,
   * not in this class. The object on which the creation is done is get with
   * AddRemoveBaseObjectListener.getSource().
   * May control the add arguments values after the user choice of the ACS as argument,
   * or if the ACS is defined and immutable.
   * Called by CollectionPropertyEditor.addElement().
   * @param _caller of this method is a AddRemoveBaseObjectListener with an
   *  handleAddBaseObject() method that is called before the instance creation.
   *  Cannot be null.
   * @param _propertyDescriptor is the descriptor of the property
   *  to update through an 'add' method. Cannot be null.
   * @param _addMethod is the source method to use for adding a new instance.
   *  Cannot be null.
   * @return the BaseObjectHandler instance with the correct initialization
   */
  public static BaseObjectHandler valueOf(AddRemoveBaseObjectListener _caller,
    PropertyDescriptor _propertyDescriptor, Method _addMethod) {

      // argument control
      if (_caller == null || _propertyDescriptor == null || _addMethod == null) {
         throw new NullPointerException();
      }

      // initialization
      BaseObjectHandler   result = new BaseObjectHandler();
      result.newBaseObjectClass_ = _propertyDescriptor.getPropertyType(); // may be an array
      
      // common initialization from the BeanInfo.
      // updates result.newBaseObjectClass_
      result = commonValueOf(result, result.newBaseObjectClass_, true);

      // 'addMethod' defines the 'add' method for the property.
      // The returned array defines:
      // - the name of the add method, at the index 0
                // - the name of the BeanInfo class where the add method arguments
                //      are defined, at the index 1
                // - the property names of the add method arguments, in the BeanInfo class,
                //      in the correct order
      // These elements allow the PropertyEditor to display a window to create
      // all the required arguments and to call the add method of the property.      

      result.caller_ = _caller;
      result.propertyDescriptor_ = _propertyDescriptor;      

      // initializes l_argumentsClass_; not like the other valueOf()
      result.l_argumentsClass_ = _addMethod.getParameterTypes();      
      result.l_argumentsValues_ = new Object[result.l_argumentsClass_.length];

      result.l_argumentsValues_[0] = "uninitialized";
      result.l_argumentsJComponent_ =
        new JComponent[result.l_argumentsClass_.length];     
      result.l_argumentsJPanel_ = new JPanel[result.l_argumentsValues_.length];
      
      if (result.l_argumentsClass_.length > 7) {
        throw new InternalError("In BaseObjectHandler,"+
          " the number of arguments is superior to 7.");
      } // if

              // for testing and tracking
          //System.out.println("in BaseObjectHandler.valueOf() for add, "+
          //  "the instance is: "+result);
          //System.out.println("in BaseObjectHandler.valueOf() for add, "+
          //  "l_argumentsNames_ is: "+result.l_argumentsNames_);

      return result;
  } // valueOf()

  /**
   * Third of the 3 valueOf methods, to create a new BaseObject
   * through the COPY method, as it is described in
   * the bean descriptor of the relevant BeanInfo class.
   * Cannot copy ACS and ViewInBase instances.
   * Called by ActionCopy, which adds the last parameter as Boolean.TRUE
   * for the copy method.
   * With this method, propertyDescriptor_ is always null.
   * @param _caller of this method is a NewBaseObjectListener with a
   *  handleNewBaseObject() method that is called when _toCopy has been copied.
   *  Cannot be null.
   * @param _toCopy is the base object to copy. Cannot be an ACS or a ViewInBase,
       nor null.
   * @param _toCopy2 is not used. Only to be different to the first valueOf method.
   * @return the BaseObjectHandler instance with the correct initialization
   */
  public static BaseObjectHandler valueOf(NewBaseObjectListener _caller,
    Copyable _toCopy, Copyable _toCopy2) {

      // argument control
      if (_caller == null || _toCopy == null) {
         throw new NullPointerException();
      }
      if (ImmutableACS.class.isAssignableFrom(_toCopy.getClass())) {
        throw new InternalError("In BaseObjectHandler,"+
          " it is not possible to create an ACS.");
      } // if

      // initialization of the BaseObject BeanInfo
      BaseObjectHandler result = new BaseObjectHandler();
      // common intialization from the BeanInfo
      result = commonValueOf(result, _toCopy.getClass(), false);
      
      // initialization of the other internal variables
      result.caller_ = _caller;
      result.objectToCopy_ = _toCopy;

      // l_argumentsClass_ never null there
      result.l_argumentsValues_ = new Object[result.l_argumentsClass_.length];

      // for testing and tracking
      result.l_argumentsValues_[0] = "uninitialized";

      result.l_argumentsJComponent_ =
        new JComponent[result.l_argumentsClass_.length];

      result.l_argumentsJPanel_ = new JPanel[result.l_argumentsValues_.length];
      
      if (result.l_argumentsClass_.length != result.l_argumentsNames_.length) {
        throw new InternalError("In BaseObjectHandler,"+
          " l_argumentsNames_ has not the l_argumentsClass_ length.");
      } // if
      if (result.l_argumentsClass_.length > 7) {
        throw new InternalError("In BaseObjectHandler,"+
          " the number of arguments is superior to 7.");
      } // if

              // for testing and tracking
          //System.out.println("in BaseObjectHandler.valueOf() for copy, "+
            //"the instance is: "+result);
          //System.out.println("in BaseObjectHandler.valueOf() for copy, "+
            //"l_argumentsNames_ is: "+result.l_argumentsNames_);

      return result;
  } // valueOf()

  /**
   * Sets most of the _result properties from its BeanInfo class.
   * Called by the valueOf() methods.
   * @param _result is the BaseObjectHandler to set up
   * @param _baseObjectClass is the class of the new instance of BaseObject to create.
   *  Cannot be an ImmutableACS or an ImmutableView.
   *  Used to find the right BeanInfo class. Cannot be null.
   * @param _fromConstructor true to build up the new object from its
   *  constructor, and false to do it from its copy
   * @return the BaseObjectHandler instance with the correct initialization
   */
  private static BaseObjectHandler commonValueOf(BaseObjectHandler _result,
          Class _baseObjectClass, boolean _fromConstructor) {

                          // for testing and tracking
                      //System.out.println("in BaseObjectHandler.commonValueOf(), "+
                        //"_baseObjectClass is: "+_baseObjectClass);

      // argument control
      if (_result == null || _baseObjectClass == null) {
         throw new NullPointerException();
      }
      // initialization of the property BeanInfo
      try {
          // className is never null since
              // it is never a generic interface like BaseObject or Node
        String className = UtilityImpl.getFirstInstance().
                            getClassNameFromArrayOrInterface(
                                UtilityImpl.getFirstInstance().
                                    getACSVocabularyFromArrayOrInterface(
                                         _baseObjectClass.getName()),
                                _baseObjectClass.getName());
        String beanName =  UtilityImpl.getFirstInstance().
                                getBeanInfoNameFromClass(className);
        Class beanInfoClass = Class.forName(beanName);
        _result.newBaseObjectBeanInfo_ = (BeanInfoPattern) beanInfoClass.newInstance();
        _result.newBaseObjectClass_ = Class.forName(className);
        beanInfoClass = null;
        className = null;
      } catch (Exception c) {
          c.printStackTrace();
          throw new InternalError("In BaseObjectHandler.commonValueOf(): "+
            c.toString());
      } // initialization of the property BeanInfo

      // initializes from the BeanInfo
      BeanDescriptor beanDescriptor =
        _result.newBaseObjectBeanInfo_.getBeanDescriptor();

      if (_fromConstructor) {
          _result.l_argumentsClass_ = (Class[]) beanDescriptor.getValue(
            "constrArgumentClasses"); // may be rewritten in the calling valueOf() method
          _result.l_argumentsNames_ = (String[]) beanDescriptor.getValue(
            "constrArgumentPropertyNames");
          _result.l_argumentsFromSelect_ = (Boolean[]) beanDescriptor.getValue(
            "constrArgumentFromSelect");
          _result.l_argumentsConcatening_ = (Integer[]) beanDescriptor.getValue(
            "constrArgumentConcatenedNameSource");
          _result.l_nullArgumentPossible_ = (Boolean[]) beanDescriptor.getValue(
            "constrNullArgumentPossible");
          _result.l_optionalArguments_ = (Boolean[]) beanDescriptor.getValue(
            "constrOptionalArgument");
          _result.l_restrictedArguments_ = (Boolean[]) beanDescriptor.getValue(
            "constrRestrictedArgument"); // may be null 
          _result.l_restrictedFromOtherArgumentMessage_ = (String[]) beanDescriptor.getValue(
            "constrRestrictedFromOtherArgumentMessage"); // may be null           
      } else {
          _result.l_argumentsClass_ = (Class[]) beanDescriptor.getValue(
            "copyArgumentClasses"); // may be rewritten in the calling valueOf() method
          _result.l_argumentsNames_ = (String[]) beanDescriptor.getValue(
            "copyArgumentPropertyNames");
          _result.l_argumentsFromSelect_ = (Boolean[]) beanDescriptor.getValue(
            "copyArgumentFromExplorer");
          _result.l_argumentsConcatening_ = (Integer[]) beanDescriptor.getValue(
            "copyArgumentConcatenedNameSource");
          _result.l_nullArgumentPossible_ = (Boolean[]) beanDescriptor.getValue(
            "copyNullArgumentPossible");
          _result.l_optionalArguments_ = (Boolean[]) beanDescriptor.getValue(
            "copyOptionalArgument");
          _result.l_restrictedArguments_ = (Boolean[]) beanDescriptor.getValue(
            "copyRestrictedArgument"); // may be null
      } // else

              // for testing and tracking
          //System.out.println("in BaseObjectHandler.commonValueOf(), "+
            //"the instance is: "+_result);
          //System.out.println("in BaseObjectHandler.commonValueOf(), "+
            //"l_argumentsNames_ is: "+_result.l_argumentsNames_);

      return _result;
  } // commonValueOf()  
  
  /**
   * Manages the events fired by all the Select buttons in the JPanels which are
   * created by getBaseObjectEditor().
   * Updates the correct panel when a base object is selected in the explorer, and
   * displays the restricted values if any. The restricted values may be Strings or
   * StringRights, and they may depend on the value of a previous argument
   * (typically, an ACS selection, since the ACS policy may define them).
   * No action for other events, for exemple from a ComboBox set in getArgumentEditor().
   * If the property is 'Upper Rights' for a StringRight, a dedicated method has to be called.
   * If there is no rights to select, this method produces an error message to the user,
   * and the creation process is ended through a call to the Cancel button on the
   * parent JDialog.
   * @param _evt is the event fired by a JButton class
   */
  public void actionPerformed(ActionEvent _evt) {

    // no action if the explorer is currently used for an object selection
    // after a click on a 'Select' button
    if (DialogBox.waitClickInExplorer__) return;
    
    // this method is for 'Select_' JButtons only
    if (!_evt.getSource().getClass().equals(JButton.class) ||
      ((JButton)_evt.getSource()).getText().indexOf("Select_") == -1) return;

    try {
        // initialization
        argIndex_ = -1;
        
        for (int i = 0 ; i < L_INDEX.length ; i++) {
            if ( !((JButton)_evt.getSource()).getText().endsWith(L_INDEX[i]) ) continue;
            // selects the index of the right argument
            else {
                argIndex_ = i;
                break;
            }
        } // for i

        if (argIndex_ == -1) return;

        // if it is a base object (not an array), selection from the explorer
        if (BaseObject.class.isAssignableFrom(l_argumentsClass_[argIndex_]) ) {
            
            // select the correct argument class
            HashSet l_type = new HashSet(INITIAL_CAPACITY);
            l_type.add(l_argumentsClass_[argIndex_]);

            // for testing purpose
            //System.out.println("in BaseObjectHandler.ActionPerformed() "+
              //"the class to select in the explorer is: "+l_argumentsClass_[argIndex_] );

            // DialogBox.showExplorerSelectionDialog() calls handleSelectionDialog()
              // handles explorer selection 
            DialogBox.getInstance().showExplorerSelectionDialog(this, null, l_type, null, true);
            l_type = null;
        } // if BaseObject
        
        // else, if the property is an array of String or StringRight restricted values,
        // displays the list of these values for the user (through getName().toString() for rights).
        // The restricted values may be defined by the ACS.
        // The ACS must be known when the 'Select' button has been displayed.
        else if (l_restrictedArguments_[argIndex_].booleanValue()) {
            ACSRun acs = null;
            // searching for the 'ACS' argument if l_restrictedFromOtherArgumentMessage_[_argIndex] is not null,
            // equals to 'ACS' or 'Main Type'
            if (l_restrictedFromOtherArgumentMessage_ != null &&
                    l_restrictedFromOtherArgumentMessage_[argIndex_] != null ) {
                for (int j = 0; j < l_argumentsClass_.length; j++) {
                  if (ACSRun.class.isAssignableFrom( l_argumentsClass_[j]) &&
                    l_argumentsNames_[j].equals("ACS")) {

                       if (!(l_argumentsValues_[j] instanceof ACSRun)) break; 
                       acs = (ACSRun)l_argumentsValues_[j];

                            // for testing and tracking
                            //System.out.println("in BaseObjectHandler.actionPerformed(), "+
                              //", the ACS is found: "+acs);

                       break;
                  } // if
                } // j
                // acs may be null if the user clicks on the button
                // before the choice of ACS
                if (acs == null) {
                    DialogBox.showErrorMessage("Select the ACS first.");
                    return;
                }
            } // if 
            
            Object[] l_ToBeSelectedValues = null;
            
            // if 'Upper Rights' for a StringRight, a dedicated method has to be called
            if (l_argumentsNames_[argIndex_].equals("Upper Rights") &&
                    newBaseObjectClass_.isAssignableFrom(StringRightImpl.class)) {
                int keyIndex = -1;
                int rightIndex = -1;
                for (int j = 0; j < l_argumentsNames_.length; j++) {
                    if (l_argumentsNames_[j].equals("Internal Key")) {
                        keyIndex = j;
                    } // if
                    else if (l_argumentsNames_[j].equals("Generic Right")) {
                        rightIndex = j;
                    } // if
                } // for j
                if (keyIndex != -1) {
                    // get the entered internal key for having the upper rights
                    try {
                        String value = ((JTextField)l_argumentsJComponent_[keyIndex]).getText();
                        // metaright may be null
                        StringRight metaright = (StringRight)l_argumentsValues_[keyIndex];

                        if (value != null && value.trim().length() != 0)
                            l_ToBeSelectedValues = ActionNewACSyst.getInstance().
                                getRestrictedUpperRightsForRight(new Integer(value), metaright);         

                            // for testing and tracking
                            //System.out.println("in BaseObjectHandler.actionPerformed(),"+
                              //  " the entered internal key: "+value);
                        value = null;
                    } catch (NumberFormatException n) {
                              DialogBox.showErrorMessage(
                                      "Internal Key must be a number, "+
                                      "to allow the Upper Rights selection");
                              acs = null;
                              return;
                    } // catch                                 
                } // if keyIndex
            // if it is about acl rights to select, for selecting the restricted rights,
            // the EP type of the previous argument has to be known  
            } else if (l_argumentsNames_[argIndex_].equals("AclRights") &&
                    l_argumentsValues_[argIndex_ - 1] instanceof EligibleParty) {
                l_ToBeSelectedValues = BaseUtilityImpl.getRestrictedValues(
                                                newBaseObjectBeanInfo_,
                                                l_argumentsNames_[argIndex_],
                                                acs,
                                                caller_.getSource(),
                                                ((EligibleParty)l_argumentsValues_[argIndex_ - 1])
                                                    .getEPType());
            // if it is about privilege rights to select, for selecting the restricted rights,
            // the privilege main type may be used; it is the previous argument
            } else if (l_argumentsNames_[argIndex_].equals("PrivilegeRights") ) {
                l_ToBeSelectedValues = BaseUtilityImpl.getRestrictedValues(
                                                newBaseObjectBeanInfo_,
                                                l_argumentsNames_[argIndex_],
                                                acs,
                                                // to find the interface 'Privilege'
                                                PrivilegeForTypeImpl.EMPTY_INSTANCE, 
                                                (String)l_argumentsValues_[argIndex_ - 1]);
            // if it is about a right to select, for selecting the restricted rights,
            // the source type in the beamer has to be known  
            } else l_ToBeSelectedValues = BaseUtilityImpl.getRestrictedValues(
                                                newBaseObjectBeanInfo_,
                                                l_argumentsNames_[argIndex_],
                                                acs,
                                                caller_.getSource(),
                                                null);

            if (l_ToBeSelectedValues == null || l_ToBeSelectedValues.length == 0) {

                DialogBox.showErrorMessage("No value to select. No creation is allowed "+
                        "by the ACS for this couple (source, target).");
                
                // searching for the Cancel button in the grand parent of the argument panels.
                // this grand parent is a JPanel set by DialogBox.showBackedJComponents().
                JPanel      panel = (JPanel)l_argumentsJPanel_[0].getParent().getParent();
                Component[] l_compo = panel.getComponents();
                JButton     cancel = null;

                                        // for testing purpose
                                        //System.out.println("in BaseObjectHandler.actionPerformed(),"+
                                          //      "JDialog components:");

                for (int i = 0; i < l_compo.length; i++) {

                                        // for testing purpose
                                        //System.out.println("----- : "+l_compo[i].getClass() );
                                        //if (((JPanel)l_compo[i]).getComponent(0) instanceof JButton)
                                          //  System.out.println("------------ component 0 is a JButton: "+
                                            //        ((JButton)(((JPanel)l_compo[i]).getComponent(0))).getText());


                    if (((JPanel)l_compo[i]).getComponent(0) instanceof JButton
                            && ((JButton)(((JPanel)l_compo[i]).getComponent(1))).getText().
                                equals("   Cancel   ") ) {
                        // in a component that is a child of 'panel',
                        // 'OK' is the first button, and 'Cancel' the second one
                        cancel = (JButton)((JPanel)l_compo[i]).getComponent(1);
                        break;
                    } // if
                } // for i

                if (cancel != null)
                    // indirect call to finalizeForProcess()
                    cancel.doClick(100);
                else
                    throw new InternalError("BaseObjectHandler.actionPerformed() error "+
                            "to finalize the dialog when there is no value to select.");
                
                panel = null;
                l_compo = null;
                cancel = null;
                acs = null;
                return;
            } // if (l_ToBeSelectedValues == null
            
            List<String> l_selectedValues = DialogBox.showSimpleSelectionsInOneDialog(
                                                l_displayNames_[argIndex_],
                                                l_ToBeSelectedValues);
                   
            if (l_selectedValues.size() != 0) {
                // the combobox is the second component, after the name component
                JComboBox box = (JComboBox) l_argumentsJPanel_[argIndex_].getComponent(1);
                box.removeAllItems();
                String item = null;
                for (Iterator ite = l_selectedValues.iterator(); ite.hasNext();) {   
                    item = (String)ite.next();
                    box.addItem( item );
                } // for
                item = null;
                box.setBackground(InternalFrame.NON_EDITABLECOLOR);

                l_argumentsJPanel_[argIndex_].repaint();
                box = null;
            } // if
            else {
                DialogBox.showOkMessage("No selected value");
            } // else

            acs = null;
            l_ToBeSelectedValues = null;
            l_selectedValues = null;
        } // else

    } catch (SmallError s) {
      throw new InternalError("In BaseObjectHandler.actionPerformed()"+
        ", DialogBox.showExplorerSelectionDialog() throws: "+s.toString());
    } // catch  

    /*
    // for JTextField processing: no update of the panel since it is done
    else if (_evt.getSource().getClass().equals(JTextField.class)) {
      return;
    } // else if

    // for JComboBox processing: no update of the panel since it is done
    else if (_evt.getSource().getClass().equals(JComboBox.class)) {
        return;
    } // else if
    */
  } // actionPerformed()

  /**
  * Called by DialogBox.showExplorerSelectionDialog() for actionPerformed(),
  * when an argument is a BaseObject which has to be selected in the explorer.
  * The right argument is found from the current value of argIndex_. 
  * If the acs has been selected, updates the argument editor for
  * any string restricted argument,
  * to take into account the value restrictions from the selected ACS.
  * For aclentry addings only, after the selection of an acs for the aclentry,
  * this method checks the ability of the selected acs to manage aclentries,
  * displays an user message if not, and otherwise updates the restricted properties 
  * in comboboxes.
  * @param _validatedObjectsSetMap with one element, where
  * - the key is l_argumentsClass_[argIndex_],
  * - the value is one selected object with the right type.
  * May be null or empty if the user cancels the selection operation.
  */
  public void handleSelectionDialog(Map _validatedObjectsSetMap) {
    // stop if the argument value is null or empty
    if (_validatedObjectsSetMap == null || _validatedObjectsSetMap.isEmpty()) {
        return;
    } // if  

            // for testing purpose
            //System.out.println("in BaseObjectHandler.handleSelectionDialog() "+
              //"the selected class in the explorer is: "+l_argumentsClass_[argIndex_] );    
    
    
    List list = (List)_validatedObjectsSetMap.get(l_argumentsClass_[argIndex_]);

    if (list != null) {
      // stores the new parameter without any check
      l_argumentsValues_[argIndex_] = (BaseObject)list.get(0);
      // displays the new parameter in the right JTextField of the panel
      ((JTextField)l_argumentsJComponent_[argIndex_]).setText(
        ((BaseObject)list.get(0)).getName().toString() );
      ((JTextField)l_argumentsJComponent_[argIndex_]).
              setBackground(InternalFrame.NON_EDITABLECOLOR);
    } // if (list != null)

        // for testing purpose
        //System.out.println("handleSelectionDialog call in BaseObjectHandler");

    // If the acs has been selected, checks and updates
    if (l_restrictedArguments_ != null &&
      ImmutableACS.class.isAssignableFrom( l_argumentsClass_[argIndex_]) ) {
          
          // For aclentry addings only, after the selection of an acs for the aclentry,
          // checks the ability of the selected acs to manage aclentries
          // and inform the user
        ImmutableACS acs = (ImmutableACS) l_argumentsValues_[argIndex_];
        if ( caller_ instanceof AddRemoveBaseObjectListener &&
            ((AddRemoveBaseObjectListener)caller_).getSource() instanceof ACSObject) {
                ACSObject obj = (ACSObject) ((AddRemoveBaseObjectListener)caller_).getSource();
                // information for external aclentries
                if (!acs.getManageInternalAcls() && acs.getName().equals(obj.getAcsName())) {
                    DialogBox.showOkMessage("Please select a right user in another ACS.");
                } // if (acs.
                // information for internal aclentries
                //if (!acs.getManageExternalAcls() && acs.getName().equals(obj.getAcsName())) {
                  //  DialogBox.showOkMessage("Please select a rights user in the same ACS.");
                //} // if (acs.
                // return for all aclentries
                if (!acs.getManageInternalAcls() && !acs.getManageExternalAcls()) {
                    DialogBox.showErrorMessage("This selected ACS does not manage AclEntries. "+
                            "Please select another ACS.");
                    list = null;
                    argIndex_ = 99;
                    acs = null;
                    obj = null;
                    return;
                } // if (acs.
                // return for external aclentries
                if (!acs.getManageExternalAcls() && !acs.getName().equals(obj.getAcsName())) {
                    DialogBox.showErrorMessage("This selected ACS does not manage External AclEntries. "+
                            "Please select another ACS.");
                    list = null;
                    argIndex_ = 99;
                    acs = null;
                    obj = null;
                    return;
                } // if (acs.                        
        } // if (caller_
        
          // updates the restricted properties in comboboxes
        for (int i = 0; i < l_restrictedArguments_.length; i++) {
            // clears the argument panel from the ACS
            // only if there is a restriction from the ACS
            if (l_restrictedArguments_[i].booleanValue() &&
                    l_restrictedFromOtherArgumentMessage_ != null &&
                    l_restrictedFromOtherArgumentMessage_[i] != null // equals to 'ACS' or 'Main Type' for Privileges
              ) {

                        // for testing purpose
                        //System.out.println("in BaseObjectHandler.handleSelectionDialog,"+
                          //  " there are some restricted arguments,"+
                          //  " and the index is:"+i+" before the repainting");
                        //System.out.println("the components in l_argumentsJPanel_[argIndex_] are:");
                        //for (int j = 0; j < l_argumentsJPanel_[argIndex_].getComponentCount(); j++) {
                          //  System.out.println("   for j:"+j+", "+
                          //    l_argumentsJPanel_[i].getComponent(j).getClass());
                        //} // for

                // the combobox is the second component, after the name component
                JComboBox box = (JComboBox) l_argumentsJPanel_[i].getComponent(1);
                // sets an empty box, to fill through the associated Select button
                box.removeAllItems();
                box.addItem( "" );
                l_argumentsJPanel_[i].repaint();
                box = null;
             } // if
        } // for
          
    } // if
    
    list = null;
    argIndex_ = 99;
  } // handleSelectionDialog()

  /**
   * Displays the arguments panels for creating the new BaseObject instance,
   * then creates it after the user had entered all the values.
   * Calls getBaseObjectEditor(new Object[1]).
   */
  public void getBaseObjectEditor() {
    getBaseObjectEditor(new Object[1]);
  }

  /**
   * This is the unique method to call on a BaseObjectHandler instance, which
   * displays the arguments panels for creating the new BaseObject instance,
   * then creates it after the user had entered all the values.
   * This method is used whatever the valueOf method which has been invoked
   * to build up the BaseObjectHandler instance.
   * When a copy method is used, a last Boolean.TRUE argument is always added
   * to the values which are entered by the user, following the copy() pattern.
   * The typical arguments are then (ImmutableName _name, ACS _acs, Directory _parent,
   * Boolean _toRegisterInManager) from the gBaseInterface.Copyable method.
   * Calls getArgumentEditor() to build up the panel for each argument.
   * Calls DialogBox.showBackedJPanelsList() to display the dialog window to the user,
   * which calls back handleCreationDialog().
   * @param _l_knownValue array of arguments for the object creation.
   *  They may be Booleans, Strings, ImmutableNames or BaseObjects,
   *  as arguments which has been chosen by the user before the call of this method.
   *  The value at the index 0 is the chosen value for the first argument, and so on.
   *  A value in the array is null if the argument for this index is not set in the BeanInfo.
   *  At the index equal or superior to the number of arguments,
   *  any fixed BaseObject may be put for displaying without taking it as an argument
   *  (typically, the source).
   *  So, the size of _l_knownValue may be inferior, or equal,
   *  or superior to the contructor arguments number.
   *  4 is the number of arguments for the instance creation of most of the base objects,
   *  excepting for actors, executables and acsrights for which there are 7 arguments.
   */
  public void getBaseObjectEditor(Object[] _l_knownValue) {

    // no action if the explorer is currently used for an object selection
    // or if there is an on-going user dialog
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getBaseObjectEditor(), "+
                  //", the incoming fixed values of the parameters are: ");
                  //for (int i = 0; i < _l_knownValue.length; i++) {
                    //System.out.println("for "+i+": "+_l_knownValue[i]);
                  //} // for
    
    // initialization
    BeanDescriptor        beanDescriptor = newBaseObjectBeanInfo_.getBeanDescriptor();
    //PropertyDescriptor[]  l_propertyDescriptor =
      //  newBaseObjectBeanInfo_.getPropertyDescriptors();

    JPanel                argumentPanel = null;
    JPanel                mainPanel = new JPanel();
    Object[]              l_value = new Object[l_argumentsValues_.length];
    ACSRun                acs = null;

    // _l_knownValue may be superior of 1 to l_argumentsValues_, equal in size or inferior
    // Set l_value values with _l_knownValue values
    if (_l_knownValue.length <= l_value.length)
      System.arraycopy(_l_knownValue, 0, l_value, 0, _l_knownValue.length);
    else {
      if (l_value.length + 1 != _l_knownValue.length) {
        throw new InternalError("in BaseObjectHandler."
          +"getBaseObjectEditor(Object[] _l_knownValue), _l_knownValue is more than 1 + "+
          l_value.length);
      } // if
      l_argumentsJPanel_ = new JPanel[_l_knownValue.length];
      l_value = new Object[_l_knownValue.length];
      System.arraycopy(_l_knownValue, 0, l_value, 0, _l_knownValue.length);
    } // else

    l_displayNames_ = new String[l_value.length];
    
    // searching for the ACS if it is known 
        // searching among the defined arguments
    for (int j = 0; j < l_argumentsClass_.length; j++) {
      if (ACSRun.class.isAssignableFrom( l_argumentsClass_[j]) &&
        (j < _l_knownValue.length && _l_knownValue[j] != null) &&
        l_argumentsNames_[j].equals("ACS") ) {
            
           acs = (ACSRun)_l_knownValue[j];

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getBaseObjectEditor(), "+
                  //", the ACS is found: "+acs);
                
           break;
      } // if
    } // j
    // else, the ACS may have still not be defined by the user, and its value will be
    // observed by any restricted-value property, in actionPerformed().
    // acs may be still null.


                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getBaseObjectEditor(), "+
                  //", the l_value length is: "+l_value.length);

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getBaseObjectEditor(), "+
                  //", the l_displayNames_ length is: "+l_displayNames_.length);
               
    
    try {
       for (int i = 0; i < l_value.length; i++) {

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getBaseObjectEditor(), "+
                  //"l_value[i] before the call to getArgumentEditor(), "+
                  //"for "+i+": "+l_value[i]);
           
          // puts the right argument editor in argumentPanel
          // when the value is known, and returns null if
          // l_argumentsNames_[i] is "" (for Directory argument of EP copying)
          if (i < l_argumentsValues_.length && i < _l_knownValue.length &&
            _l_knownValue[i] != null) {            
                l_argumentsValues_[i] = _l_knownValue[i];
                // creates the argument panel
                argumentPanel = (JPanel) getArgumentEditor(
                  new JPanel(),
                  l_argumentsNames_[i],
                  false,
                  i,
                  true,
                  _l_knownValue[i],
                  acs);
          } // if
          // no display of an argument which has to be null since the ACS does not
          // manage it. Example: Directory if !acs.getManageResourceTrees()
          else if (l_nonVisibleArguments_ != null && i < l_nonVisibleArguments_.length &&
                  l_nonVisibleArguments_[i]) {

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getBaseObjectEditor(), "+
                  //"the argument is not visible for i = "+i);
              
              argumentPanel = null;
          }
          // standard case: puts the right argument editor in argumentPanel
          // with a value to be choosen by the user (among a list or not)
          else if (i < l_argumentsValues_.length)
              // creates the argument panel
              argumentPanel = (JPanel) getArgumentEditor(
                  new JPanel(),
                  l_argumentsNames_[i],
                  l_argumentsFromSelect_[i].booleanValue(),
                  i,
                  false,
                  null,
                  acs);
          // outside the range of l_argumentsValues_,
          // displays the known value of a BaseObject (typically the source)
          // when it is not an argument to build up the new object
          else if (i < _l_knownValue.length && (_l_knownValue[i] instanceof BaseObject))
              // creates the argument panel
              argumentPanel = (JPanel) getArgumentEditor(
                  new JPanel(),
                  UtilityImpl.getFirstInstance().
                    selectBeanInfo((BaseObject)_l_knownValue[i]).
                        getBeanDescriptor().getDisplayName(),
                  false,
                  i,
                  true,
                  _l_knownValue[i],
                  null);

          // updates for handleSelectionDialog()
          // and actionPerformed()
          if (argumentPanel != null)
              argumentPanel.setMaximumSize(new Dimension(1200,75)); //1200 does not matter
          l_argumentsJPanel_[i] = argumentPanel;
          
        } // for (int i =
    } catch (Exception e) {

      e.printStackTrace();
      throw new InternalError("in BaseObjectHandler.getBaseObjectEditor(), "+
                  "exception from getArgumentEditor(): "+e.toString());    
                
    } // catch  
    
    
    mainPanel.setBackground(InternalFrame.BACKCOLOR_2);
    mainPanel.setForeground(InternalFrame.FORECOLOR);
    mainPanel.setLayout(new BoxLayout(mainPanel,BoxLayout.Y_AXIS));
    for (int i = 0; i < l_argumentsJPanel_.length; i++) {
        if (l_argumentsJPanel_[i] != null) { // null if l_nonVisibleArguments_[i] == TRUE
            l_argumentsJPanel_[i].setBackground(InternalFrame.BACKCOLOR_2);
            l_argumentsJPanel_[i].setForeground(InternalFrame.FORECOLOR);
            mainPanel.add(l_argumentsJPanel_[i]);
        }
    } // for i

    // displays l_argumentsJPanel_ which contains the required editor for each argument.
      // if the user enters a "return" in an editable field in l_argumentsJPanel_,
      // it is processed in actionPerformed() and do nothing.
      // calls handleCreationDialog() when the user clicks on 'OK'.
    if (objectToCopy_ == null) DialogBox.showBackedJComponents(this,
            "Create a new "+ beanDescriptor.getDisplayName(),
            mainPanel);  
    else DialogBox.showBackedJComponents(this,
            "Copy "+ objectToCopy_.getName().toString(),
            mainPanel);

    beanDescriptor = null;
    argumentPanel = null;
    l_value = null;
    mainPanel = null;
    onGoingCreationDialog_ = true;
  } // getBaseObjectEditor()

/**
 * After an user click on 'OK', reads the data entered by the user in the JPanels,
 * and creates the new BaseObject.
 * Called from DialogBox.showBackedJComponents() for getBaseObjectEditor().
 * Calls checkNullArguments(), then setRelativeArguments(),
 * setFinalArguments(), and createBaseObject() at the end of operations.
 * Calls finalizeForProcess() to finalize the BaseObjectHandler instance,
 * whatever the result.
 * @param _data true if the user has entered the data and false if the user has closed the dialog 
 */
  public void handleCreationDialog(boolean _data) {
    
            // for testing purpose
            //System.out.println("Call to BaseObjectHandler.handleCreationDialog() "); 
      
    if (!onGoingCreationDialog_) throw new InternalError("in BaseObjectHandler."+
            "handleCreationDialog(), error in the user dialog handling.");
    
    onGoingCreationDialog_ = false;
      
    if (!_data) {
      
                // for testing and tracking
                //System.out.println("in BaseObjectHandler.handleCreationDialogr(), "+
                  //"the dialog is aborted by the user. ");       
      
      finalizeForProcess();
      return;
    } // if
    
    
    if (!checkNullArguments()) {
      
                // for testing and tracking
                //System.out.println("in BaseObjectHandler.handleCreationDialogr(), "+
                  //"null arguments control is positive; the object creation is aborted. ");       
      
      finalizeForProcess();
      return;
    } // if

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.handleCreationDialog(), "+
                  //", before setRelativeArguments() ");
    
    try {
        
                      // for testing and tracking
                      //System.out.println("In BaseObjectHandler.handleCreationDialog()"+
                        //" before the call to setRestrictedSelectedArguments(), the arguments are:");
                      //for (int i = 0; i < l_argumentsValues_.length; i++) {
                        // if (l_argumentsValues_[i] != null) 
                          //   System.out.println("----- true class is "+l_argumentsValues_[i].getClass()+
                            //         " and value is "+l_argumentsValues_[i]+"|"); 
                         //else System.out.println("----- l_argumentsValues_ is null for the required "
                           //      +l_argumentsClass_[i]+" at "+i);
                         //if (l_argumentsValues_[i] != null && l_argumentsValues_[i].getClass().isArray() &&
                           //      l_argumentsValues_[i] != null &&
                           //      l_argumentsValues_[i] instanceof Object[]) {
                             //System.out.println("----- for the array "+l_argumentsValues_[i].getClass()+":");
                             //Object[] l_obj = (Object[])l_argumentsValues_[i];
                             //for (int j = 0; j < l_obj.length; j++) {
                               //  System.out.println("----------- for "+j+", value: "+l_obj[j]);
                             //} // for
                         //} // if
                    //} // for i
        
        // updates the restricted selected arguements 
        // for 'upper rights' in new StringRightImpl, and for all rights in an ACSObject
     setRestrictedSelectedArguments();
        
                      // for testing and tracking
                      //System.out.println("In BaseObjectHandler.handleCreationDialog()"+
                        //" before the call to setRelativeArguments(), the arguments are:");
                      //for (int i = 0; i < l_argumentsValues_.length; i++) {
                         //if (l_argumentsValues_[i] != null) 
                           //  System.out.println("----- true class is "+l_argumentsValues_[i].getClass()+
                             //        " and value is "+l_argumentsValues_[i]+"|"); 
                         //else System.out.println("----- l_argumentsValues_ is null for the required "
                           //      +l_argumentsClass_[i]+" at "+i);
                         //if (l_argumentsValues_[i] != null && l_argumentsValues_[i].getClass().isArray() &&
                           //      l_argumentsValues_[i] != null &&
                           //      l_argumentsValues_[i] instanceof Object[]) {
                             //System.out.println("----- for the array "+l_argumentsValues_[i].getClass()+":");
                             //Object[] l_obj = (Object[])l_argumentsValues_[i];
                             //for (int j = 0; j < l_obj.length; j++) {
                               //  System.out.println("----------- for "+j+", value: "+l_obj[j]);
                             //} // for
                         //} // if
                      //} // for i
                      
      // updates l_argumentsValues_ from the current values in the panels
      setRelativeArguments();

                      // for testing and tracking
                      //System.out.println("In BaseObjectHandler.handleCreationDialog()"+
                        //" after the call to setRelativeArguments(), the arguments are:");
                      //for (int i = 0; i < l_argumentsValues_.length; i++) {
                        // if (l_argumentsValues_[i] != null) 
                          //   System.out.println("----- true class is "+l_argumentsValues_[i].getClass()+
                            //         " and value is "+l_argumentsValues_[i]+"|"); 
                         //else System.out.println("----- l_argumentsValues_ is null for the required "
                           //      +l_argumentsClass_[i]+" at "+i);
                         //if (l_argumentsValues_[i] != null && l_argumentsValues_[i].getClass().isArray() &&
                           //      l_argumentsValues_[i] != null &&
                           //      l_argumentsValues_[i] instanceof Object[]) {
                             //System.out.println("----- for the array "+l_argumentsValues_[i].getClass()+":");
                             //Object[] l_obj = (Object[])l_argumentsValues_[i];
                             //for (int j = 0; j < l_obj.length; j++) {
                               //  System.out.println("----------- for "+j+", value: "+l_obj[j]);
                             //} // for
                         //} // if
                      //} // for i
      
      
      setFinalArguments();
    
    } catch (BaseError b) {
      // The user has entered a non-Integer in a textfield.
        // He has ridden an error message and the frame stays open.
      b.printStackTrace();
      return;     
    } catch (Exception e) {
        
      e.printStackTrace();
      throw new InternalError("in BaseObjectHandler.handleCreationDialogr(), "+
        "exception before the creation of "+newBaseObjectClass_+": ");      
    } // catch
      
              // for testing and tracking
              //System.out.println("in BaseObjectHandler.handleCreationDialog(), "+
                //"after setFinalArguments(), for "+newBaseObjectClass_);
              //for (int i = 0; i < l_argumentsValues_.length; i++) {
                //System.out.println("       l_argumentsValues_Class["+i+"]: "+l_argumentsClass_[i]);
                //System.out.println("       l_argumentsValues_["+i+"]: "+l_argumentsValues_[i]);
                  //       if (l_argumentsValues_[i] != null && l_argumentsValues_[i].getClass().isArray() &&
                    //             l_argumentsValues_[i] != null &&
                    //             l_argumentsValues_[i] instanceof Object[]) {
                      //       System.out.println("----- for the array "+l_argumentsValues_[i].getClass()+":");
                      //       Object[] l_obj = (Object[])l_argumentsValues_[i];
                      //       for (int j = 0; j < l_obj.length; j++) {
                      //           System.out.println("----------- for "+j+", value: "+l_obj[j]);
                      //       } // for
                      //   } // if
              //} // for

    // do the main job and call finalizeForProcess() at the end, whatever the result
    createBaseObject();
      
  } // handleCreationDialog()
  
  //
  // --------------- PRIVATE METHODS ----------------------------------
  //

  /**
   * Gets a graphical editor for an argument or for the source,
   * when the type is a boolean, string, name, BaseObject (including StringRight),
   * or a String list.
   * It may display first a null-value editor, or an editor with
   * a pre-defined value for the argument, or a comboBox of Strings.
   * A button 'Select' is added at the right of the text, when:
   * - a BaseObject has to be selected in the explorer,
   * - or the type is a String list, and the list has to be selected
   *    from some restricted values.
   * If the type is a String with restricted values, a comboBox is displayed.
   * But the 'Select' button or the comboBox is displayed only if the
   * rectricted values are known. For example, when the ACS defines the restriction
   * and is not still choosen by the user for the new Base Object,
   * a non-editable text is displayed: "<Set the ACS first>". 'ACS' word, or any other
   * word, is provided there by l_restrictedFromOtherArgumentMessage_.
   * Called by getBaseObjectEditor().
   * Calls actionPeformed() when the button 'Select' is clicked on, and getBooleanEditor()
   * for the booleans.
   * Calls BaseUtilityImpl.getRestrictedValues().
   * Initializes l_displayNames_ with the property display name in most of the cases.
   * @param _result is the returned JPanel.
   * @param _propertyName of the argument. Never null. Has to be inferior to 50.
   *   If it is equals to "", returns an empty panel. This is not always the name
   *   displayed to the user in the panel.
   * @param _isSelected true if the argument is selected in the explorer or through
   *   another mean which needs the 'Select' button.
   * @param _argIndex is the index of the argument in all the argument list.
   * @param _isDefined true when the argument has been chosen by the user
   *    before the call of this method.
   * @param _knownValue of argument only if the argument is a String,
   *  an ImmutableName or a BaseObject (not Boolean),
   *  which has been chosen by the user before the call to this class.
   *  May be null if _isDefined is false.
   * @param _acs is null when the method is called by getBaseObjectEditor()
   *   without knowledge of the acs,
   *   and non null when the acs is known or when it is called
   *   by handleSelectionDialog() after the selection of the acs by the user.
   * @return JPanel to be displayed
   */
  private JPanel getArgumentEditor(JPanel _result, String _propertyName,
    boolean _isSelected, int _argIndex, boolean _isDefined,
    Object _knownValue, ACSRun _acs) {
        
      if (_propertyName.equals("")) return _result;
 
                            // for testing and tracking
                            //System.out.println("starting BaseObjectHandler.getArgumentEditor()"+
                              //      " for _propertyName: "+_propertyName);

      // initialization
      JLabel                nameLabel;
      JButton               selectButton;
      JTextField            field = null;
      PropertyDescriptor[]  l_propertyDescriptor = newBaseObjectBeanInfo_.getPropertyDescriptors();

      _result.setFont(BeamerManager.plainButtonFont);
      
      //index of the property in the BeanInfo, not in the final BaseObjectHandler main panel.
      // this method will throw an OutOfBoundsArray exception if index is not updated.
      int beanInfoIndex = -1;
      // searching for the right property,
      // unless it is a defined value for which _propertyName is the displayed name
      for (int j = 0; j < l_propertyDescriptor.length; j++) {

                            // for testing and tracking
                          //System.out.println("-- in the search of _propertyName, for j: "+j);
                          //System.out.println("------ l_propertyDescriptor[j].getName(): "+
                            //      l_propertyDescriptor[j].getName());
          
          if ( l_propertyDescriptor[j].getName().equals( 
            BaseUtilityImpl.removeSpacesInString(_propertyName ) ) ) {
              beanInfoIndex = j;
              
              break;
          } // if
      } // for j 

                            // for testing and tracking
                            //System.out.println("in BaseObjectHandler.getArgumentEditor(),"+
                              //"the "+_propertyName+" index in l_propertyDescriptor is: "+
                              //beanInfoIndex+ " and the array lenght is "+
                              //l_propertyDescriptor.length);

      // Set the display name
      // first case: it is defined as an argument or
      // as the source name for an 'add' method
      if (_isDefined )
          l_displayNames_[_argIndex] = _propertyName;

      // second case: the argument may be null and it is not set by the caller
      else if (_argIndex < l_argumentsValues_.length &&
        l_nullArgumentPossible_[_argIndex].booleanValue() && !_isDefined) {

          //System.out.println("in BaseObjectHandler.getArgumentEditor(), null argument is possible, and"+
            //          " l_displayNames_ size is: "+l_displayNames_.length);          
          //System.out.println("in BaseObjectHandler.getArgumentEditor(),"+
            //          " _argIndex = "+_argIndex);
          //System.out.println("in BaseObjectHandler.getArgumentEditor(),"+
            //          " l_propertyDescriptor size is: "+l_propertyDescriptor.length); 
          //System.out.println("in BaseObjectHandler.getArgumentEditor(),"+
            //          " index = "+index);
          
          l_displayNames_[_argIndex] = 
                  l_propertyDescriptor[beanInfoIndex].getDisplayName() + " (may be null)";
          
      // third case: the most common one
      } else
          l_displayNames_[_argIndex] = l_propertyDescriptor[beanInfoIndex].getDisplayName();

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getArgumentEditor(), "+
                  //"displayName: "+l_displayNames_[_argIndex]);
                //if (_argIndex < l_argumentsValues_.length)
                     //System.out.println("   argument class: "+l_argumentsClass_[_argIndex]);
                
      
      nameLabel = new JLabel(l_displayNames_[_argIndex]);
      nameLabel.setMaximumSize(NAME_DIMENSION);
      nameLabel.setMinimumSize(NAME_DIMENSION);
      nameLabel.setPreferredSize(NAME_DIMENSION);

      // if the argument is a BOOLEAN
      if ((_argIndex < l_argumentsClass_.length) &&
        Boolean.class.isAssignableFrom(l_argumentsClass_[_argIndex])) {

          _result = getBooleanEditor(_result, _propertyName, _argIndex, _isDefined,
                  _knownValue, l_propertyDescriptor,  beanInfoIndex);

          nameLabel = null;
          l_propertyDescriptor = null;
          return _result;
      } // if the argument is a boolean

      // else, detects the restricted values if they exist
      String[] l_RestrictedValues = null;
      Object[] l_RestrictedBuffer = null;
      String[] l_RestrictedValuesTwo = null;
      String   maxValue = null;
      int      maxValueLength = 0;
      

      //try {
        if (l_restrictedArguments_ != null &&
          _argIndex < l_restrictedArguments_.length &&
          l_restrictedArguments_[_argIndex].booleanValue()) {


                        // for testing and tracking
                        //System.out.println("in BaseObjectHandler.getArgumentEditor(), "+
                          //"the main variables are for a restricted argument:");
                        //System.out.println("   newBaseObjectBeanInfo_: "+newBaseObjectBeanInfo_);
                        //System.out.println("   _propertyName: "+_propertyName);
                        //System.out.println("   _isExplorerSelected: "+_isSelected);
                        //System.out.println("   _argIndex: "+_argIndex);
                        //System.out.println("   _isDefined: "+_isDefined);
                        //System.out.println("   _knownValue: "+_knownValue);
                        //System.out.println("   _acs: "+_acs);
                        //System.out.println("Argument class: "+l_argumentsClass_[_argIndex]);


              // if the argument is a restricted String or StringRight
              // from the BeanInfo or from the acs when it is known,
              // initializes l_RestrictedValues. If it is StringRight,
              // the last null argument implies that the values may be
              // larger than the true restricted values, but it does not
              // matter since l_restrictedFromOtherArgumentMessage_[_argIndex]
              // is not null
              l_RestrictedBuffer = BaseUtilityImpl.getRestrictedValues(
                    newBaseObjectBeanInfo_,
                    l_argumentsNames_[_argIndex],
                    ((ACSRun)_acs), caller_.getSource(),
                    null);
              
              if (l_RestrictedBuffer != null &&
                      (StringRight.class.isAssignableFrom( l_argumentsClass_[_argIndex]) ||
                      StringRight[].class.isAssignableFrom( l_argumentsClass_[_argIndex])) ) {

                  l_RestrictedValues = new String[l_RestrictedBuffer.length];
                  for (int i = 0; i < l_RestrictedBuffer.length; i++) {
                      l_RestrictedValues[i] = 
                              ((StringRight)l_RestrictedBuffer[i]).getName().toString();
                  } // for i
                  
              } 
              else l_RestrictedValues = (String[])l_RestrictedBuffer;
              
              // initializes maxValue, maxValueLength
              if (l_RestrictedValues != null) {                
                  // if 'null' is the first value in the array set by the property descriptor,
                  // removes it
                  if (l_RestrictedValues[0] == null) {
                      l_RestrictedValuesTwo = new String[l_RestrictedValues.length - 1];
                      System.arraycopy(l_RestrictedValues, 1, 
                              l_RestrictedValuesTwo, 0, l_RestrictedValuesTwo.length);
                  }
                  else l_RestrictedValuesTwo = l_RestrictedValues;

                                // for testing and tracking
                                //System.out.println("in BaseObjectHandler.getArgumentEditor(), "+
                                  //"after the restricted values detection, their values: ");
                  
                  for (int i = 0; i < l_RestrictedValuesTwo.length; i++) {

                                // for testing and tracking
                                //System.out.println(" --------- at "+i+": "+l_RestrictedValuesTwo[i]);

                        if (l_RestrictedValuesTwo[i].length() > maxValueLength) {
                            maxValue = l_RestrictedValuesTwo[i];
                            maxValueLength = l_RestrictedValuesTwo[i].length();
                        } // if
                  } // for i          
                  //maxValue = Collections.max( Arrays.asList(l_RestrictedValuesTwo) ); 
                  maxValue = maxValue+"-------";
 
                        // for testing and tracking
                        //System.out.println("in BaseObjectHandler.getArgumentEditor(), maxValue: "+maxValue); 
                  
              } // if (l_RestrictedValues != null)

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getArgumentEditor(), for "+
                  //_propertyName+", after the restricted values detection, their values are: "+
                  //l_RestrictedValues);

        } // if l_restrictedArguments_[_argIndex].booleanValue()

      //} catch (Exception e) {
        //   System.out.println("in BaseObjectHandler.getArgumentEditor(), "+
          //     "exception:"+ e.toString());          
      //} // catch

      // if there are restricted values (sometimes still unknown), displays a comboBox 
      if (_argIndex < l_argumentsValues_.length && l_restrictedArguments_ != null &&
        l_restrictedArguments_[_argIndex].booleanValue() ) {
             
              JComboBox  box = null;
              // for a restricted-values property which is an array,
              // the following states 1 - 3 - 4 of 'box'
              // are its successive states when the user selects the property, then the values
              if (l_RestrictedValues == null && l_restrictedFromOtherArgumentMessage_ != null &&
                      l_restrictedFromOtherArgumentMessage_[_argIndex] != null) {
                  box = new JComboBox(
                          new String[] {"<Set the "+
                                  l_restrictedFromOtherArgumentMessage_[_argIndex]+
                                  " first>"});
                  box.setBackground(InternalFrame.SOFTGREY);
                  box.setPrototypeDisplayValue("----------------------------------------");
                  box.setFont(BeamerManager.plainFont);
              } else if (l_RestrictedValues != null && !l_argumentsClass_[_argIndex].isArray()) {
                  box = new JComboBox(l_RestrictedValues);
                  box.setBackground(Color.white);
                  if (maxValue != null) box.setPrototypeDisplayValue(maxValue);
                  box.setFont(BeamerManager.plainFont);
                  
                  // to display the greater values correctly in the combo box
                  //String greater = null;
                  //for (int i = 0; i < l_RestrictedValues.length; i++) {
                    //  if (greater == null) greater = l_RestrictedValues[i];
                    //  else if (l_RestrictedValues[i].length() > greater.length())
                    //      greater = l_RestrictedValues[i];
                  //} // for i
                  //box.setPrototypeDisplayValue(greater);
                  //greater = null;
              } else if (!_isDefined) {// no use since handleSelectionDialog() does it
                  box = new JComboBox(new String[] {""});
                  box.setBackground(Color.white);
                  box.setPrototypeDisplayValue("----------------------------------------");
                  box.setFont(BeamerManager.plainFont);
              } else { // no use since actionPerformed() does it
                  box = new JComboBox((String[])_knownValue);
                  box.setBackground(Color.white);
                  box.setPrototypeDisplayValue("----------------------------------------");
                  box.setFont(BeamerManager.plainFont);
              }

              if (box != null) {
                  DefaultListCellRenderer renderer = null;
                  try {
                      l_argumentsJComponent_[_argIndex] = box;
                      box.setActionCommand(Integer.toString(_argIndex));
                      renderer = new DefaultListCellRenderer();
                      //box.setMaximumRowCount(10);
                      //box.setMaximumSize(new Dimension(200, 22));
                      //box.setMinimumSize(new Dimension(200, 22));
                      //box.setPreferredSize(new Dimension(200, 22));
                  } catch (Exception e) {
                            // for testing and tracking
                            System.out.println("in BaseObjectHandler.getArgumentEditor(), "+
                              "exception from JComboBox processing:"+ e.toString());
                  }

                  renderer.setHorizontalAlignment(SwingConstants.RIGHT);
                  box.setRenderer(renderer);
                  box.setEditable(false);
                  box.addActionListener(this);

                  nameLabel.setLabelFor(box);
                  _result.setLayout(new FlowLayout(FlowLayout.LEFT));

                  _result.add(nameLabel);
                  _result.add(box);
                  _result.setBackground(Color.white);

                  // creates a button 'Select_i'
                  // with this instance as listener
                  // if the property is an array
                  if (l_argumentsClass_[_argIndex].isArray()) {
                      selectButton = new JButton("Select_"+L_INDEX[_argIndex]);
                      selectButton.setBackground(Color.white);
                      selectButton.setFont(BeamerManager.plainButtonFont);
                      selectButton.setBorder(new BevelBorder(
                        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR) );
                      selectButton.addActionListener(this);
                      _result.add( selectButton);
                  }

                  l_RestrictedValuesTwo = null;
                  l_RestrictedBuffer = null;
                  l_RestrictedValues = null;
                  renderer = null;
                  box = null;
                  nameLabel = null;
                  l_propertyDescriptor = null;
                  selectButton = null;
                  return _result;
              } // if (box != null) {
      
      } // if (l_restrictedArguments_[_argIndex].booleanValue()
      
      // the argument is not a boolean and there are not restricted values;
      // a text is displayed, even if the argument is an Integer,
      // and sometimes a 'Select' button to select from the explorer
      // displays the defined value
      field = new JTextField(10);
      
      if (_isDefined) {
        if (_knownValue == null) {
            nameLabel = null;
            l_propertyDescriptor = null;
            l_RestrictedValuesTwo = null;
            l_RestrictedBuffer = null;
            l_RestrictedValues = null;
            l_displayNames_[_argIndex] = null;
            field = null;

            return _result; // empty panel
        }
        else if (_knownValue instanceof String) 
          field = new JTextField((String)_knownValue);
        else if (_knownValue instanceof ImmutableName)
          field = new JTextField(((ImmutableName)_knownValue).toString());
        else if (_knownValue instanceof BaseObject)
          field = new JTextField(((BaseObject)_knownValue).toString()); 
      } // if      

      if (_argIndex < l_argumentsJComponent_.length)
        l_argumentsJComponent_[_argIndex] = field;

      field.setHorizontalAlignment(JTextField.RIGHT); // fires an event
      field.setFont(BeamerManager.plainFont);

      nameLabel.setLabelFor(field);
      _result.setLayout(new FlowLayout(FlowLayout.LEFT));

      _result.add(nameLabel);
      _result.add(field);
      _result.setBackground(Color.white);

      // for BaseObject argument
      if (_isDefined) {
          field.setEditable(false);

          // through the ToolTipText analysis
          field.setToolTipText(_propertyName);
          field.setBackground(InternalFrame.NON_EDITABLECOLOR);
      } // if (_isDefined)
      // displays a 'Select' button for choosing a base object
      else if (_isSelected ) {
            
            field.setEditable(false);

            // through the ToolTipText analysis
            field.setToolTipText(_propertyName);
            field.setBackground(InternalFrame.SOFTGREY);

            // creates a button 'Select_i'
            // with this instance as listener
            selectButton = new JButton("Select_"+L_INDEX[_argIndex]);
            selectButton.setBackground(Color.white);
            selectButton.setFont(BeamerManager.plainButtonFont);
            selectButton.setBorder(new BevelBorder(
              BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR) );

            selectButton.addActionListener(this);

            //_result.setBorder(new EtchedBorder() );
            _result.add( selectButton );

      } // if (_isSelected
      else {
          field.setEditable(true);

          // if 'enter' calls actionPerformed(), to NOT take it into account
          // since the value is ridden only when the user says OK
          field.addActionListener(this);

          // through the ToolTipText analysis
          field.setToolTipText(_propertyName);
          field.setBackground(Color.white);
      } // else

          // for testing and tracking
          //System.out.println();

      field = null;
      nameLabel = null;
      l_propertyDescriptor = null;
      l_RestrictedValues = null;
      l_RestrictedValuesTwo = null;
      l_RestrictedBuffer = null;
      selectButton = null;
      maxValue = null;
      
      return _result;
  } // getArgumentEditor()

  /**
   * Gets a graphical editor for an argument or for the source, when the type is a boolean.
   * Called by getArgumentEditor().
   * Calls actionPeformed() when the button 'Select' is clicked on.
   * Calls BaseUtilityImpl.getRestrictedValues().
   * Initializes l_displayNames_ with the property display name in most of the cases.
   * @param _result is the returned JPanel.
   * @param _propertyName of the argument. Never null. Has to be inferior to 50.
   *   If it is equals to "", returns an empty panel. This is not always the name
   *   displayed to the user in the panel.
   * @param _argIndex is the index of the argument in all the argument list.
   * @param _isDefined true when the argument has been chosen by the user
   *    before the call of this method.
   * @param _knownValue of argument only if the argument is a String,
   *  an ImmutableName or a BaseObject (not Boolean),
   *  that has been chosen by the user before the call to this class.
   *  May be null if _isDefined is false.
   * @param _l_propertyDescriptor descriptors from the BeanInfo
   * @param _beanInfoIndex index of the right PropertyDescriptor
   * @return JPanel to be displayed
   */
  private JPanel getBooleanEditor(JPanel _result, String _propertyName,
    int _argIndex, boolean _isDefined, Object _knownValue,
    PropertyDescriptor[]  _l_propertyDescriptor,  int _beanInfoIndex) {
      
      // initialization
      JLabel                nameLabel;
      JTextField            field = null;

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.getBooleanEditor(), "+
                  //"displayName: "+l_displayNames_[_argIndex]);
                //if (_argIndex < l_argumentsValues_.length)
                     //System.out.println("   argument class: "+l_argumentsClass_[_argIndex]);
                
      
      nameLabel = new JLabel(l_displayNames_[_argIndex]);
      nameLabel.setMaximumSize(NAME_DIMENSION);
      nameLabel.setMinimumSize(NAME_DIMENSION);
      nameLabel.setPreferredSize(NAME_DIMENSION);

      // if the value is already known - new in 0.7.0
      if (_isDefined) {


                    // for testing and tracking
                    //System.out.println("in BaseObjectHandler.getBooleanEditor(), "+
                      //"the argument is a known boolean ");

          // displays the defined value
          if (_knownValue == null || !(_knownValue instanceof Boolean)) {
                nameLabel = null;
                l_displayNames_[_argIndex] = null;
                field = null;

                return _result; // empty panel
          }
          else {
              if (((Boolean)_knownValue).booleanValue())
                  field = new JTextField("TRUE");
              else
                  field = new JTextField("False");
          }

          if (_argIndex < l_argumentsJComponent_.length)
            l_argumentsJComponent_[_argIndex] = field;

          field.setHorizontalAlignment(JTextField.RIGHT); // fires an event
          field.setFont(BeamerManager.plainFont);

          nameLabel.setLabelFor(field);
          _result.setLayout(new FlowLayout(FlowLayout.LEFT));

          _result.add(nameLabel);
          _result.add(field);
          nameLabel.setBackground(Color.white);

          field.setEditable(false);

          // through the ToolTipText analysis
          field.setToolTipText(_propertyName);
          field.setBackground(InternalFrame.NON_EDITABLECOLOR);

          field = null;
      }
      else { // no defined BOOLEAN value

          // initializes the JComboBox with the tags, and selects the TRUE value
            // as default value
          String[] l_tags;
          if (_l_propertyDescriptor[_beanInfoIndex].getValue("tags") != null) {
            l_tags = (String[]) _l_propertyDescriptor[_beanInfoIndex].getValue("tags");
          } else l_tags = new String[] {"yes", "no"};

          JComboBox  box = new JComboBox(l_tags);
          if (_argIndex < l_argumentsJComponent_.length)
              l_argumentsJComponent_[_argIndex] = box;
          box.setActionCommand(Integer.toString(_argIndex));
          DefaultListCellRenderer renderer = new DefaultListCellRenderer();

          renderer.setHorizontalAlignment(SwingConstants.RIGHT);
          ((JComboBox)box).setRenderer(renderer);
          ((JComboBox)box).setEditable(false);
          ((JComboBox)box).addActionListener(this);
          box.setBackground(Color.white);
          box.setFont(BeamerManager.plainFont);

          nameLabel.setLabelFor(box);
          _result.setLayout(new FlowLayout(FlowLayout.LEFT));

          _result.add(nameLabel);
          _result.add(box);
          _result.setBackground(Color.white);

          l_tags = null;
          renderer = null;
          box = null;
      } // else

      nameLabel = null;

      return _result;
  } // getBooleanEditor()


  /**
   * Checks in l_argumentsJComponent_ if there is a null or empty argument
   * while it is forbidden.
   * If the checking is negative, a message is displayed to the user,
   * this method returns 'FALSE' and the user should enter new arguments.
   * Called by handleCreationDialog().
   * @return true if ok
   */
  private boolean checkNullArguments() {
    // initialization
    boolean                 check = true;
    
    // checks that each argument has been set if it must not be null
    for (int i = 0; i < l_argumentsValues_.length; i++) {
        
          // for testing and tracking
          //System.out.println("in BaseObjectHandler.checkNullArguments(), "+
            //"the argument number"+i+" is: "+((JTextField or
                        //ComboBox if restricted...)l_argumentsJComponent_[i]).getText()+"|");
          
        if (Boolean.class.isAssignableFrom(l_argumentsClass_[i]) ||
                l_nullArgumentPossible_[i].booleanValue())
                        continue;
        
        else if (l_argumentsJComponent_[i] instanceof JTextField &&
          (((JTextField)l_argumentsJComponent_[i]).getText() == null
            || ((JTextField)l_argumentsJComponent_[i]).getText().trim().length() == 0)) {

              DialogBox.showErrorMessage(
                l_argumentsNames_[i]+" can't be null or empty. ");
              check = false;
              break;
        } // else if

        else if (l_argumentsJComponent_[i] instanceof JComboBox &&
                ((JComboBox)l_argumentsJComponent_[i]).getItemAt(0) == null ) {

              DialogBox.showErrorMessage(
                l_argumentsNames_[i]+" can't have a null first item. ");
              check = false;
              break;
        } // else if
        
    } // for i

    return check;
  } // checkNullArguments()
  
  /**
   * Sets l_argumentsValues_ when an argument is a StringRight array argument
   * in a combox, from a restricted set proposed to the user.
   * Called by handleCreationDialog().
   * Calls StringRightImpl.getRightFromNameAsString() to get each StringRight, which
   * requires to select the ACSRun from the value of caller_ directly when it is an ACSRun,
   * or indirectly, through the ACS argument, when caller_ is an ACSObject. <P>
   * If there are some rights to select but without user selection, the rights array
   * is set to null, and there is no exception from the AclEntry or the Privilege.
   * @exception BaseError if the convertion of String to StringRight does not work
   */
  private void setRestrictedSelectedArguments() throws BaseError {
    
    if (l_restrictedArguments_ == null ) return; 
    
    for (int i = 0; i < l_argumentsValues_.length; i++) {

        // caution: l_argumentsFromSelect_[i].booleanValue() may be false, since 'Select'
        // is set when this property is true, or when l_restrictedArguments_ is true.
       
        if (l_restrictedArguments_ != null &&
            !l_restrictedArguments_[i].booleanValue() ) continue;
        
        if (!l_argumentsClass_[i].isArray()) continue;
        // the argument is an array

        if (!StringRight.class.isAssignableFrom( l_argumentsClass_[i].getComponentType() )) 
          continue;
        
        // it is a StringRight array argument in a combox   
        if (!(l_argumentsJComponent_[i] instanceof JComboBox)) continue;
        if (((JComboBox)l_argumentsJComponent_[i]).getItemAt(0).equals( "<Set the "+
              l_restrictedFromOtherArgumentMessage_[i]+" first>" )) continue;
        
        if (l_restrictedFromOtherArgumentMessage_ != null ) {

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.setRestrictedSelectedArguments(),"+
                  //  " it is a JCombobox at the index "+i+", and the values are:");

          if (!StringRightImpl.class.isAssignableFrom( l_argumentsClass_[i].getComponentType() ))
              l_argumentsValues_[i] = new StringRight[
                ((JComboBox)l_argumentsJComponent_[i]).getItemCount()];
          else
              l_argumentsValues_[i] = new StringRightImpl[
                ((JComboBox)l_argumentsJComponent_[i]).getItemCount()];
          
          StringRight[] l_newValues = new StringRight[
            ((JComboBox)l_argumentsJComponent_[i]).getItemCount()];

          ACSRun acs = null;
          String text = null;
          if (caller_.getSource() instanceof ACSRun) 
              acs = (ACSRun)caller_.getSource();
          else if (caller_.getSource() instanceof ACSObject) {
              // the single case is when an AclEntry or a Privilege is created,
              // and the ACS may be any open ACS
                for (int j = 0; j < l_argumentsClass_.length; j++) {
                  if (ImmutableACS.class.isAssignableFrom( l_argumentsClass_[j]) &&
                    l_argumentsNames_[j].equals("ACS")) {

                       if (!(l_argumentsValues_[j] instanceof ACSRun)) break; 
                       acs = (ACSRun)l_argumentsValues_[j];

                            // for testing and tracking
                            //System.out.println("in BaseObjectHandler.actionPerformed(), "+
                              //", the ACS is found: "+acs);

                       break;
                  } // if
                } // j
                // acs may be null if the user clicks on the button
                // before the choice of ACS
                if (acs == null) {
                    DialogBox.showErrorMessage("Select the ACS first");
                    return;
                }  
          } // else

                    // for testing and tracking
                    //System.out.println("BaseObjectHandler.setRestrictedSelectedArguments(), acs:"+acs);

          if (acs != null) {
              // ACS_Facade.getEorBaseObject() does not work for ACSFactoryImpl rights
              for (int j = 0; j < l_newValues.length; j++) {
                  text = (String)((JComboBox)l_argumentsJComponent_[i]).getItemAt(j);
                  // if text is null, there is no selected value
                  // and the right at 0 is null
                  if (text != null && text.length() != 0)
                      l_newValues[j] = BaseUtilityImpl.getRightFromNameAsString(text, acs);

                        // for testing and tracking
                        //System.out.println("------ for j: "+j+"- new value class: ");
                        //if (l_newValues[j] != null)
                          //  System.out.println(l_newValues[j].getClass()+
                            //    "- new value: "+l_newValues[j]);

              } // for j
                        // for testing and tracking
                        //System.out.println("l_argumentsValues_[i] length: "+
                          //      ((StringRight[])l_argumentsValues_[i]).length);

              if (l_newValues[0] != null)
                  System.arraycopy(l_newValues, 0, (StringRight[])l_argumentsValues_[i], 0,
                    l_newValues.length);
              else // the AclEntry or the Privilege has no right
                  l_argumentsValues_[i] = null;
          } // if

          l_newValues = null;
          acs = null;
          text = null;
      } // if (l_restrictedFromOtherArgumentMessage_ != null
      else {

                    // for testing and tracking
                    //System.out.println("BaseObjectHandler.setRestrictedSelectedArguments(), "+
                      //      "it is not an array of StringRight");                  
      }

    } // for i
  } // setRestrictedSelectedArguments()
  
  /**
   * Sets in l_argumentsValues_, from l_argumentsJComponent_,
   * the right arguments to use, but without handling of the absolute names.
   * Converts String in Integer if necessary. This is the complement of
   * setRestrictedSelectedArguments() which processes only selected arguments
   * never handled here.
   * Caution: an empty JTextField returns through getText() the value "", not null.
   * Called by handleCreationDialog().
   * @exception BaseError if Convertion from String to Integer does not work
   */
  private void setRelativeArguments() throws BaseError {
    Name  name = null;
    String error = null;

    for (int i = 0; i < l_argumentsValues_.length; i++) {
        // sets in l_argumentsValues_ the displayed text
        // in editable JTextFields and JComboBoxes,
        // but not in the non editable JTextFields, for which the argument
        // is a BaseObject instance
        if (l_argumentsFromSelect_[i].booleanValue() ) continue;

        if (!l_argumentsClass_[i].isArray()){
            // first case: if it is string argument
            if (String.class.isAssignableFrom(l_argumentsClass_[i])) {

              if (l_argumentsJComponent_[i] instanceof JTextField) {
                  l_argumentsValues_[i] = 
                    ((JTextField)l_argumentsJComponent_[i]).getText();
              } else if (l_argumentsJComponent_[i] instanceof JComboBox)
                  l_argumentsValues_[i] = 
                    (String)((JComboBox)l_argumentsJComponent_[i]).getSelectedItem();  

            // second case: if it is a boolean, l_argumentsJComponent_[i] may has been updated
            } else if (Boolean.class.isAssignableFrom(l_argumentsClass_[i])) {

                // for testing and tracking
                //System.out.println("in BaseObjectHandler.setRelativeArguments(),"+
                  //  " for i="+i+    
                    //" the JComponent is: "+l_argumentsJComponent_[i].getClass());

              // if the value is known, the component is a JTextField
              if (!(l_argumentsJComponent_[i] instanceof JComboBox))
                  continue;

              if (((JComboBox)l_argumentsJComponent_[i]).getSelectedIndex() == 0 )
                  l_argumentsValues_[i] = Boolean.TRUE;
              else l_argumentsValues_[i] = Boolean.FALSE;
            
            // third case: it is a Integer argument
            } else if (Integer.class.isAssignableFrom(l_argumentsClass_[i]) ) {
                // conversion to the name type, with addition of the required
                // caracters (but it may be a relative name, that is processed after
                // to add the first components)
                try {
                    String value = ((JTextField)l_argumentsJComponent_[i]).getText();
                    
                    // value != null && value.trim().length() != 0 (see checkNullArgument())
                    l_argumentsValues_[i] = new Integer(value);         

                        // for testing and tracking
                        //System.out.println("in BaseObjectHandler.setRelativeArguments(),"+
                          //  " the value is an Integer: "+((Integer)l_argumentsValues_[i]).intValue());
                        
                } catch (NumberFormatException n) {
                          DialogBox.showErrorMessage(
                                  ((JTextField)l_argumentsJComponent_[i]).getText()+
                                  " is not convertible in a number; ");
                          error = ((JTextField)l_argumentsJComponent_[i]).getText()+
                                  " is not convertible in a number ";
                          break;
                } // catch
            
            // fourth case: it is a free name argument
            } else if (!l_restrictedArguments_[i].booleanValue() &&
                    ImmutableName.class.isAssignableFrom(l_argumentsClass_[i])) {
                // conversion to the name type, with addition of the required
                // caracters (but it may be a relative name, that is processed after
                // to add the first components)
                l_argumentsValues_[i] = ACS_Facade.getNameFromString(
                    " "+((JTextField)l_argumentsJComponent_[i]).getText()+"::  " );
            }
                // if the argument is a BaseObject,
                // it has been set in l_argumentsValues_[i]
                // by getArgumentEditor()
  
        } // if (!l_argumentsClass_[i].isArray())

        else { // the argument is an array
            // if it is a string argument
            if (l_argumentsClass_[i].getComponentType() == String.class) {
              if (l_argumentsJComponent_[i] instanceof JTextField)
                  l_argumentsValues_[i] = new String[] 
                    {((JTextField)l_argumentsJComponent_[i]).getText()};
              else if (l_argumentsJComponent_[i] instanceof JComboBox) {

                        // for testing and tracking
                        //System.out.println("in BaseObjectHandler.setRelativeArguments(),"+
                          //  " it is a JCombobox, and the values are:");

                  l_argumentsValues_[i] = new String[
                    ((JComboBox)l_argumentsJComponent_[i]).getItemCount()];
                  String[] l_newValues = new String[
                    ((JComboBox)l_argumentsJComponent_[i]).getItemCount()];

                  for (int j = 0; j < l_newValues.length; j++) {
                      l_newValues[j] = (String)
                        ((JComboBox)l_argumentsJComponent_[i]).getItemAt(j);

                        // for testing and tracking
                        //System.out.println("      for j"+j+": "+l_newValues[j]);

                  } // for j
                  System.arraycopy(l_newValues, 0, (String[])l_argumentsValues_[i], 0,
                    l_newValues.length);
                  l_newValues = null;
              } // else if (l_argumentsJComponent_[i] instanceof JComboBox)
            } else {  // else, it is not a String type
                if (ImmutableName.class.isAssignableFrom(l_argumentsClass_[i].getComponentType())) {
                    name = ACS_Facade.getNameFromString(
                        " "+((JTextField)l_argumentsJComponent_[i]).getText()+"::  " );
                    l_argumentsValues_[i] = new Name[] {name};
                } // else it is a name argument
            } // else
        } // if (l_argumentsClass_[i] is an Array

    } // for i
    
    if (error != null) throw new BaseError(error);
    
    name = null;
  } // setRelativeArguments()

  /**
   * Sets in l_argumentsValues_ the absolute arguments to use,
   * through an analyse of the relative names to concatenate.
   * The relative name is a String or an ImmutableName, and the argument
   * to put as context is the name of a BaseObject.
   * Caution: any change in this method should be also done in AssociateUtilityImpl.
   * @see ARoad0.gBase.AssociateUtilityImpl#setFinalArgumentsForAssociateCreation
   */
  private void setFinalArguments() {
    // variables
    int startIndex = -1;

    // it may exist some relative names that need to use an other argument
    // to set the first components of their name
    for (int i = 0; i < l_argumentsValues_.length; i++) {
      // if it is a relative name to concatenate, process it,
      // test if the argument to concatenate is a BaseObject and is not null
      if (l_argumentsConcatening_[i] == null) continue;

        if (l_argumentsValues_[l_argumentsConcatening_[i].intValue()] != null &&
          l_argumentsValues_[l_argumentsConcatening_[i].intValue()] instanceof BaseObject) {


                           // for testing and tracking purpose
                           //System.out.println("call to BaseObjectHandler.setFinalArguments(); ");

                           //System.out.println("        the class is "+l_argumentsClass_[i]);
                           //System.out.println("        the root name to add is "+
                              //(Name)((BaseObject)
                                    //l_argumentsValues_[l_argumentsConcatening_[i].intValue()]).
                                    //getName());


            if (!l_argumentsClass_[i].isArray()){

                    // l_argumentsValues_[i] is a String or an ImmutableName,
                    // and the context is given by the name of BaseObject
              if (l_argumentsValues_[i] instanceof ImmutableName)
                  l_argumentsValues_[i] = ((Name)((BaseObject)
                    l_argumentsValues_[l_argumentsConcatening_[i].intValue()]).
                    getName()).extended( (ImmutableName)l_argumentsValues_[i] );
              else if (l_argumentsValues_[i] instanceof String)
                  l_argumentsValues_[i] = ((Name)((BaseObject)
                    l_argumentsValues_[l_argumentsConcatening_[i].intValue()]).
                    getName()).extended( (String)l_argumentsValues_[i] );

            } // if (!l_argumentsClass_[i].isArray())
            else {
                    // l_argumentsValues_[i] is a String or an ImmutableName
                    // and the context is given by the name of BaseObject
              if (l_argumentsValues_[i] instanceof ImmutableName)
                l_argumentsValues_[i] = new Name[] {((Name)((BaseObject)
                  l_argumentsValues_[l_argumentsConcatening_[i].intValue()]).
                  getName()).extended( (ImmutableName)l_argumentsValues_[i] )};
              else if (l_argumentsValues_[i] instanceof String)
                l_argumentsValues_[i] = new Name[] {((Name)((BaseObject)
                  l_argumentsValues_[l_argumentsConcatening_[i].intValue()]).
                  getName()).extended( (String)l_argumentsValues_[i] )};

            } // else
        }
        else { // the start of the name is null, so add the ACS/view name at the start

          // searches for the ACS/view name in the arguments
          for (int j = 0; j < l_argumentsNames_.length; j++) {
            if (l_argumentsNames_[j].equals("ACS") ||
              l_argumentsNames_[j].equals("View")) {
                startIndex = j;
                break;
            } // if
          } // for

          // the ACS/view name is non null; it has been checked in checkNullArguments()
          if (!l_argumentsClass_[i].isArray()){

                    // l_argumentsValues_[i] is a String or an ImmutableName
              if (l_argumentsValues_[i] instanceof ImmutableName)
                  l_argumentsValues_[i] = ((Name)((BaseObject)
                    l_argumentsValues_[startIndex]).
                    getName()).extended( (ImmutableName)l_argumentsValues_[i] );
              else if (l_argumentsValues_[i] instanceof String)
                  l_argumentsValues_[i] = ((Name)((BaseObject)
                    l_argumentsValues_[startIndex]).
                    getName()).extended( (String)l_argumentsValues_[i] );

          } // if (!l_argumentsClass_[i].isArray())
          else {
                    // l_argumentsValues_[i] is a String or an ImmutableName
              if (l_argumentsValues_[i] instanceof ImmutableName)
                l_argumentsValues_[i] = new Name[] {((Name)((BaseObject)
                  l_argumentsValues_[startIndex]).
                  getName()).extended( (ImmutableName)l_argumentsValues_[i] )};
              else if (l_argumentsValues_[i] instanceof String)
                l_argumentsValues_[i] = new Name[] {((Name)((BaseObject)
                  l_argumentsValues_[startIndex]).
                  getName()).extended( (String)l_argumentsValues_[i] )};

          } // else
        } // else

    } // for
    return;
  } // setFinalArguments()

  /**
   * Removes the null arguments which are optional, to select the right constructor.
   * Based on 'optionalArgument' in the bean descriptor.
   */
  private void removeOptionalArguments() {
    // initialization
    Class[] l_newArgumentsClass = new Class[l_argumentsClass_.length];
    Object[] l_newObjectArguments = new Object[l_argumentsValues_.length];
    int count = 0;

    // for testing purpose
    //System.out.println("call to BaseObjectHandler.removeOptionalArguments(); "+
    //  "l_optionalArguments_.length: "+l_optionalArguments_.length+
    //  "; l_argumentsClass_.length: "+l_argumentsClass_.length+
    //  "; l_argumentsValues_.length: "+l_argumentsValues_.length);

    for (int i = 0; i < l_argumentsValues_.length; i++) {
      // initialize the arrays, and remove the null and optional argument in i
      if (l_optionalArguments_[i].booleanValue() && l_argumentsValues_[i] == null) {
        for (int j = i; j < l_argumentsValues_.length; j++) {
          if (j != l_argumentsValues_.length - 1) {
            l_newArgumentsClass[j] = l_argumentsClass_[j+1];
            l_newObjectArguments[j] = l_argumentsValues_[j+1];
          } else {
            l_newArgumentsClass[j] = null;
            l_newObjectArguments[j] = null;
          } // else
        } // for
        count++;
      } // if
      // initializes the arrays
      else {
        l_newArgumentsClass[i] = l_argumentsClass_[i];
        l_newObjectArguments[i] = l_argumentsValues_[i];
      }
    } //for

    // updates l_argumentsClass_ and l_argumentsValues_
    l_argumentsClass_ = new Class[l_newArgumentsClass.length - count];
    System.arraycopy(l_newArgumentsClass, 0, l_argumentsClass_, 0,
      l_newArgumentsClass.length - count);
    l_argumentsValues_ = new Object[l_newObjectArguments.length - count];
    System.arraycopy(l_newObjectArguments, 0, l_argumentsValues_, 0,
      l_newObjectArguments.length - count);

    l_newArgumentsClass = null;
    l_newObjectArguments = null;

    // for testing purpose
    //System.out.println("call to BaseObjectHandler.removeOptionalArguments(). l_argumentsValues_ is: ");
    //for (int i = 0; i < l_argumentsValues_.length; i++) {
    //  System.out.println("for ["+i+"]: "+l_argumentsValues_[i].toString());
    //}

    return;
  } // removeOptionalArguments()

  /**
   * Called by handleCreationDialog() to create the base object.
   * Calls caller_.addOneBaseObject() if caller_ is a AddRemoveBaseObjectListener,
   * or calls caller_.handleNewBaseObject() when caller_ is a NewBaseObjectListener.
   * The test order is add method, then copy, then constructor method.
   * The exception handling is extended to provide full information.
   */
  private void createBaseObject() {

                    // for testing purpose
                    //System.out.println("BaseObjectHandler.createBaseObject() starts ");

    BaseObject result = null;
    
    try {
      // creates the arguments array for the 'add' method
      if (caller_ instanceof AddRemoveBaseObjectListener && propertyDescriptor_ != null &&
        propertyDescriptor_.getValue("addMethod") != null) {

          ((AddRemoveBaseObjectListener)caller_).
              addOneBaseObject(l_argumentsValues_);

                      // for testing and tracking
                      //System.out.println("In BaseObjectHandler.createBaseObject()"+
                        //" with the add method for the object class: "+newBaseObjectClass_ +
                        //" and the caller class: "+caller_.getClass());

      } // if
      else if (objectToCopy_ != null) {

                      // for testing and tracking
                      //System.out.println("In BaseObjectHandler.createBaseObject()"+
                        //" creates the copy with: ");
                      //for (int i = 0; i < l_argumentsValues_.length; i++) {
                          //System.out.println("at "+i+": "+l_argumentsValues_[i]);
                      //} // for i
          
        // creates the BaseObject copy
        result = (BaseObject)objectToCopy_.copy(
          (ImmutableName)l_argumentsValues_[0],
          (ACS)l_argumentsValues_[1],
          (Directory)l_argumentsValues_[2],
          Boolean.TRUE);

        // returns the new BaseObject to the caller
        caller_.handleNewBaseObject(result);

                      // for testing and tracking
                      //System.out.println("In BaseObjectHandler.createBaseObject()"+
                        //" the final state of the copy is: "+result.toString());
        
      } // else
      // most current case: propertyDescriptor_ and objectToCopy_ are null.
      // caller_ is a NewBaseObjectListener.
      // uses the constructor to create the instance
      else {

        // removes the null arguments to select the right constructor
        removeOptionalArguments();

      
              // for testing and tracking
              //System.out.println("in BaseObjectHandler.createBaseObject(), "+
                //"for "+newBaseObjectClass_);
              //System.out.println("   and the constructor "+
                //      newBaseObjectClass_.getConstructor(l_argumentsClass_));
              //for (int i = 0; i < l_argumentsValues_.length; i++) {
                // System.out.println("       l_argumentsValues_Class["+i+"]: "+l_argumentsClass_[i]);
                 //System.out.println("       l_argumentsValues_["+i+"]: "+l_argumentsValues_[i]);
                 //if (l_argumentsValues_[i] != null && l_argumentsValues_[i].getClass().isArray() &&
                   //      l_argumentsValues_[i] != null &&
                     //    l_argumentsValues_[i] instanceof Object[]) {
                     //System.out.println("----- for the array "+l_argumentsValues_[i].getClass()+":");
                     //Object[] l_obj = (Object[])l_argumentsValues_[i];
                     //for (int j = 0; j < l_obj.length; j++) {
                       //  System.out.println("----------- for "+j+", value: "+l_obj[j]);
                     //} // for
                 //} // if
              //} // for
        
        // creates the BaseObject
        result = (BaseObject) newBaseObjectClass_.
          getConstructor(l_argumentsClass_).newInstance(l_argumentsValues_);

        // returns the new BaseObject to the caller
        caller_.handleNewBaseObject(result);

                      // for testing and tracking
                      //System.out.println("In BaseObjectHandler.createBaseObject()"+
                        //" the final result is: "+result.toString());
        
      } // else
      
      // extended exception handling to understand all the cases
      // and finalizeForProcess() for the cases where the user may produce the exception
      // exemple: enter an internal bridge target for creating an Actor
    } // try
    catch (NoSuchMethodException b) {
      if (result != null && result != ResourceImpl.EMPTY_INSTANCE) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError u) {
            throw new InternalError("NoSuchMethodException in the "+
              "creation of "+newBaseObjectClass_.getName());
          } // catch
      } // if
 
              // for testing and tracking
              //System.out.println("In BaseObjectHandler.createBaseObject(), NoSuchMethodException "+
                //      "for the class "+newBaseObjectClass_+
                  //    " and the possible constructors are:");
              //Constructor[] l_const = newBaseObjectClass_.getConstructors();
              //for (int i = 0; i < l_const.length; i++) {
                //  System.out.println("------------ at "+i+": "+l_const[i]);
              //} // for i
      
      DialogBox.showErrorMessage("NoSuchMethodException in the "+
        "creation of "+newBaseObjectClass_.getName()+
        ", the exception is: "+b.toString());
    }
    catch (InstantiationException b) {
      if (result != null && result != ResourceImpl.EMPTY_INSTANCE) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError u) {
            throw new InternalError("InstantiationException in the "+
              "creation of "+newBaseObjectClass_.getName());
          } // catch
      } // if
      DialogBox.showErrorMessage("InstantiationException in the "+
        "creation of "+newBaseObjectClass_.getName()+" with the error: "+b);
      System.out.println("in BaseObjectHandler.createBaseObject(), "+
        "the exception is an InstantiationException: ");
        b.printStackTrace();
    }
    catch (InvocationTargetException t) {
      if (result != null && !result.equals(ResourceImpl.EMPTY_INSTANCE)) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError u) {
            System.out.println("In BaseObjectHandler.createBaseObject(), UpDateError: "+
              u.toString()+" in InvocationTargetException: "+t.toString() );
          }
      } // if

      // show the message to the user
      int lastIndex = ((InvocationTargetException)t).getTargetException().toString().lastIndexOf(": ");
      if (lastIndex != 0)
        DialogBox.showErrorMessage(
          ((InvocationTargetException)t).getTargetException().toString().substring(lastIndex+2));
      else DialogBox.showErrorMessage(
        ((InvocationTargetException)t).getTargetException().toString());

      // CreateError is for the user errors
      if (!(t.getTargetException() instanceof CreateError)) {
          System.out.println("in BaseObjectHandler.createBaseObject(), "+
            "the exception is an InvocationTargetException: ");
          ((InvocationTargetException)t).printStackTrace();
      } // if

    }
    catch (IllegalAccessException b) {
      if (result != null && result != ResourceImpl.EMPTY_INSTANCE) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError u) {
            throw new InternalError("IllegalAccessException in the "+
              "creation of "+newBaseObjectClass_.getName());
          } // catch
      } // if
      DialogBox.showErrorMessage("IllegalAccessException in the "+
        "creation of "+newBaseObjectClass_.getName()+" with the error: "+b);
    }
    catch (ProcessError p) {
      if (result != null && result != ResourceImpl.EMPTY_INSTANCE) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError u) {}
      } // if
      // show the message to the user
      int lastIndex = p.toString().lastIndexOf(": ");
      if (lastIndex != 0)
        DialogBox.showErrorMessage(p.toString().substring(lastIndex+2));
      else DialogBox.showErrorMessage(p.toString());

                          // for testing and tracking
                          //System.out.println("in BaseObjectHandler.createBaseObject(), creation of a : "+
                            //newBaseObjectClass_+", the exception is: "+p.toString());
                          //for (int i = 0; i < l_argumentsValues_.length; i++) {
                            //System.out.println("l_argumentsClass_["+i+"]: "+l_argumentsClass_[i]);
                            //System.out.println("l_argumentsValues_["+i+"]: "+l_argumentsValues_[i]);
                          //} // for
                          //p.printStackTrace();

    } // catch
    catch (CreateError c) {
      if (result != null && result != ResourceImpl.EMPTY_INSTANCE) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError u) {}
      } // if
      // show the message to the user
      int lastIndex = c.toString().lastIndexOf(": ");
      if (lastIndex != 0)
        DialogBox.showErrorMessage(c.toString().substring(lastIndex+2));
      else DialogBox.showErrorMessage(c.toString());

                          // for testing and tracking
                          //System.out.println("in BaseObjectHandler.createBaseObject(), "+
                            //"the exception is a CreateError: ");
                          //c.printStackTrace();

    } // catch
    /*
    catch (UpDateError u) {
      if (result != ResourceImpl.EMPTY_INSTANCE) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError v) {}
      } // if
      // show the message to the user
      DialogBox.getInstance().showErrorMessage("UpdateError in the "+
        "creation of a "+newBaseObjectClass_.getName()+
        ", the exception is: "+u.toString());
    } // catch
     **/
    catch (Exception e) {
      if (result != null && result != ResourceImpl.EMPTY_INSTANCE) {
          try {
            ((FinalizedObjectForUser)result).finalizeForUser();
          } catch (UpDateError u) {}
      } // if
      // show the message to the user
      DialogBox.showErrorMessage("Exception in the "+
        "creation of a "+newBaseObjectClass_.getName()+
        ", the exception is: "+e.toString());

                          // for testing and tracking
                          //System.out.println("in BaseObjectHandler.createBaseObject(), "+
                            //"the exception is an "+e.getClass());
                          //e.getMessage();
                          //e.printStackTrace();

              // for testing and tracking
              //System.out.println("Exception in the creation of a : "+
                //newBaseObjectClass_+", the exception is: "+e.toString());
              //for (int i = 0; i < l_argumentsValues_.length; i++) {
                //System.out.println("l_argumentsClass_["+i+"]: "+l_argumentsClass_[i]);
                //System.out.println("l_argumentsValues_["+i+"]: "+l_argumentsValues_[i]);
              //} // for
              //e.printStackTrace();
    } // catch
    
    finalizeForProcess();
    result = null;

                          // for testing and tracking
                          //System.out.println("BaseObjectHandler.createBaseObject() ends. ");
  } // createBaseObject()
  
  
  /**
   * Sets to null all the variables of the instance
   */
  private void finalizeForProcess() {

    // for testing purpose
    //System.out.println("call to BaseObjectHandler.finalizeForProcess() "); 

    caller_ = null;
    propertyDescriptor_ = null;
    newBaseObjectClass_ = null;
    newBaseObjectBeanInfo_ = null;
    l_argumentsValues_ = null;
    l_argumentsClass_ = null;
    l_argumentsNames_ = null;

    if (l_argumentsJComponent_ != null) {
        for (int i = 0; i < l_argumentsJComponent_.length; i++) {
            if (l_argumentsJComponent_[i] != null) 
                l_argumentsJComponent_[i].setVisible(false);
        } // for
    } // if
    l_argumentsJComponent_ = null;
    l_argumentsFromSelect_ = null;
    l_argumentsConcatening_ = null;
    l_nullArgumentPossible_ = null;
    l_optionalArguments_ = null;
    l_restrictedArguments_ = null;
    
    if (l_argumentsJPanel_ != null) {
        for (int i = 0; i < l_argumentsJPanel_.length; i++) {
            if (l_argumentsJPanel_[i] != null) l_argumentsJPanel_[i].removeAll();
        } // for
    } // if
    if (l_argumentsJPanel_ != null && l_argumentsJPanel_[0] != null)
        l_argumentsJPanel_[0].invalidate(); // invalidate the parent
    l_argumentsJPanel_ = null;
    argIndex_ = -1;

    try {
        super.finalize();
    } catch (Throwable t) {
        throw new InternalError("in BaseObjectHandler.finalizeForProcess(), "+
          "super.finalize throws: "+t.toString());
    }
  } // finalizeForProcess()

} // BaseObjectHandler