//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui1;

import javax.swing.plaf.basic.BasicButtonUI;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JToolBar;
import javax.swing.border.TitledBorder;
import javax.swing.border.BevelBorder;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.Box;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreePath;
import javax.swing.BorderFactory;
import java.net.MalformedURLException;
import java.awt.Color;
import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.BasicStroke;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.ArrayList;
import java.util.List;
import java.util.Collection;
import java.beans.*;
import java.lang.reflect.Method;
import java.util.Collections;

import ARoad0.Gui2.ACSTreeUtilities;
import ARoad0.Gui2.TreeManager;
import ARoad0.Gui2.ViewModel;
import ARoad0.Pattern.Singleton;
import ARoad0.Pattern.BeanInfoPattern;
import ARoad0.Pattern.MVC_Controller;
import ARoad0.Pattern.MVC_View;
import ARoad0.ErrW.ProcessError;
import ARoad0.ErrB.BaseError;
import ARoad0.gBase.*;
import ARoad0.gBaseInterface.*;

import ARoad0.gWork.ACS_Facade;
import ARoad0.gWork.ViewInBase_Facade;
import ARoad0.gDMak.ActionOpenArchitecture;
import ARoad0.gDMak.ActionUtilities;

/**
 * This important class displays and updates the properties of any base object
 * in matched tabbed panes. It manages customizable CRUD (create/read/update/delete) panels
 * from the entities both from the generic gBase package, from any ACS addon gBase,
 * and from some virtual entities the gDMak package handles. For this purpose,
 * it is responsible for managing the ARoad0.gBaseInterface.BaseObject
 * property editors and their layout. It works both for the beamer,
 * the ACSFactoryImpl frame, the structure frame, the 'All full views' and
 * 'All NoThan views' frames (see ActionOpenAllFullViews and ActionOpenAllNoThanViews).
 * <P>
 * Each property to display for a base object is analyzed
 * in the static method updateBeamerPane(), that
 * gets the right beamer pane, that is always a JPanel so far,
 * and initialize it with all the panels associated to the
 * properties of the displayed object. For this,
 * it uses the BeanDescriptor of each property class.
 * Group the property editors with the same panel group name
 * in the same group panel.
 * To display the tabbed panes in the order (1, 2, 3),
 * the first set (one or several) of properties has to be from the pane 1,
 * then the following property has to be from the pane 2,
 * and it works even if another pane 1 property is at the end
 * of the properties array.
 * <P>
 * The first method to call on a BeamerManager instance is updateBeamerPane(), which
 * returns a JComponent to use as an argument for the method InternalFrame.openBeamer().
 * The single instance listens all the closing actions from the base object
 * shown in the beamer.
 * It handles a list for back and forward operations, up to 30 updatings.
 * <P>
 * For the 'allViews' window, the setJTabbedPane() method is called by ActionOpenAllFullViews
 * and ActionOpenAllNoThanViews to populate a JPanel with the GraphicViewPropertyEditors,
 * from property descriptors which are dynamically constructed 
 * by ActionOpenUserV.getViewPropertyDescriptor(). This is the same way for
 * ActionNewACSyst.createUserDefinedACSrights() to define the properties of a new ACS.
 */
public class BeamerManager implements Singleton, MVC_Controller, MVC_View,
        PropertyChangeListener {

  /**
   * The maximal size of the back list.
   */
  private static final int            BACK_SIZE_LIMIT = 30;
  /**
   * List of the backing base objects displayed in the beamer.
   */
  private static final List           L_BACK_BASEOBJECTS = new ArrayList(BACK_SIZE_LIMIT);
  /**
   * List of the forwarding base objects displayed in the beamer.
   */
  private static final List           L_FORWARD_BASEOBJECTS = new ArrayList(BACK_SIZE_LIMIT);
  /**
   * Center tabbed pane in the beamer
   */
  private static final JTabbedPane    CENTER_JTABBEDPANE = new JTabbedPane();
  /**
   * Center panel in the beamer
   */
  private static final JPanel         CENTER_JPANEL = new JPanel();
  /**
   * The tool bar
   */
  private static final JToolBar       TOOLBAR = new JToolBar();
  /**
   * General font size
   */
  public static final int             VIEW_FONT_SIZE = 9;
  /**
   * General font size
   */
  public static final int             BEAMER_FONT_SIZE = 10;
  /**
   * Font size for buttons
   */
  public static final int             BUTTON_FONT_SIZE = 9;
  /**
   * Font size for large texts at 12
   */
  public static final int             LARGE_TEXT_FONT_SIZE = 12;
  /**
   * Italic font
   */
  public static final Font            italicFont = new Font(
    "Serif", Font.ITALIC, BEAMER_FONT_SIZE);
  /**
   * Plain font. Used by GraphicNoThanView.drawArrowsAndRights().
   */
  public static final Font            plainFont = new Font(
    "Serif", Font.PLAIN, BEAMER_FONT_SIZE);
  /**
   * Bold font with BEAMER_FONT_SIZE as size
   */
  public static final Font            boldUpFont = new Font(
    "Serif", Font.BOLD, BEAMER_FONT_SIZE );
  /**
   * Plain font for button titles with BUTTON_FONT_SIZE as size
   */
  public static final Font            plainButtonFont = new Font(
    "Serif", Font.PLAIN, BUTTON_FONT_SIZE );
  /**
   * Bold font with VIEW_FONT_SIZE as size
   */
  public static final Font            smallBoldUpFont = new Font(
    "Serif", Font.BOLD, VIEW_FONT_SIZE );
  /**
   * Bold font. Used by GraphicText with LARGE_TEXT_FONT_SIZE as size.
   */
  public static final Font            boldButtonFont = new Font(
    "Serif", Font.BOLD, LARGE_TEXT_FONT_SIZE );
  
  /**
   * Three buttons in the tool bar
   */
  private static JButton              backButton__;
  private static JButton              forwardButton__;
  private static JButton              deleteButton__;
  protected static ImageIcon          backButtonIcon__;
  protected static ImageIcon          forwardButtonIcon__;
  /**
   * This boolean is true when a click on the back or forward button
   * provides a new selection in the explorer, and then updates the beamer.
   */
  private static boolean              updateFromButtons__;

  /**
   * Pane of the beamer; this component is a JPanel with a BorderLayout.
   */
  private static JComponent           beamerPane__;
  /**
   * Central panel in beamerPane__; may be a JPanel or a JTabbedPane.
   */
  private static JComponent           centerPanel__;
  /**
   * Object displayed in the beamer
   */
  private static BaseObject           objectInBeamer__;
  /**
   * Name of the object displayed in the beamer
   */
  private static ImmutableName        objectInBeamerName__;
  /**
   * Base object manager which fire propertyChangeEvent when objectInBeamer__
   * is deleted (not when it changes; these changes are listened
   * by the property editors)
   */
  private static BaseBeanManager      objectInBeamerManager__;
  /**
   * BeanInfo of the object displayed in the beamer
   */
  private static BeanInfoPattern      objectInBeamerBeanInfo__;
  /**
   * Current CommonPropertyEditors in the beamer. The key is the display name.
   * Used only to finalize the properties.
   */
  private static Map                  m_currentPropertyEditors__;
  /**
   * The title of the current pane selected in the JTabbedPane
   */
  private static String               previousTabbedPaneTitle__;

  // --------------------- static initialization -----------------------------
  static {
    beamerPane__ = new JPanel();
    centerPanel__ = CENTER_JPANEL;
    beamerPane__.setBackground(Color.white);
    beamerPane__.setLayout(new BorderLayout());
    try { //  - arFolder__ replaced
      backButtonIcon__ = new ImageIcon (
              Desktop.AR_URI.resolve("Gui1"+Desktop.URL_SEPAR+"images"+
              Desktop.URL_SEPAR+"back.jpg").toURL());
      forwardButtonIcon__ = new ImageIcon (
              Desktop.AR_URI.resolve("Gui1"+Desktop.URL_SEPAR+"images"+
              Desktop.URL_SEPAR+"forward.jpg").toURL());
    } catch (MalformedURLException m){
      throw new InternalError("in BeamerManager constructor, exception"+
        " on an icon: "+m.toString() );
    }
    //backButtonIcon__.setDescription("to call back the previous object in the beamer");
    backButton__ = new JButton(backButtonIcon__);
    forwardButton__ = new JButton(forwardButtonIcon__);
    deleteButton__ = new JButton("Delete It ");
    updateFromButtons__ = false;

    // initialize the buttons and the toolBar
    TOOLBAR.setFloatable(false);
    TOOLBAR.setBackground(Color.white);
    backButton__.setEnabled(false);
    forwardButton__.setEnabled(false);
    forwardButton__.setAlignmentY(java.awt.Component.CENTER_ALIGNMENT);
    backButton__.setAlignmentY(java.awt.Component.CENTER_ALIGNMENT);
    forwardButton__.setBackground(Color.white);
    backButton__.setBackground(Color.white);
    forwardButton__.setToolTipText("forward to the next displayed object");
    backButton__.setToolTipText("back to the previous displayed object");
    
    deleteButton__.setBackground(Color.white);
    deleteButton__.setBorder(new BevelBorder(
      BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));

    TOOLBAR.addSeparator();
    TOOLBAR.add(backButton__);
    TOOLBAR.addSeparator();
    TOOLBAR.add(forwardButton__);
    TOOLBAR.addSeparator(new Dimension(250, 5)); // to have the Delete button at right
    beamerPane__.add( TOOLBAR, BorderLayout.NORTH );

    CENTER_JTABBEDPANE.setBackground(Color.white);
    CENTER_JPANEL.setBackground(Color.white);
    CENTER_JPANEL.setLayout(new BoxLayout(CENTER_JPANEL, BoxLayout.Y_AXIS));

    // set the listeners in back, forward and delete buttons in the toolbar
    backButton__.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {

        // no action if the explorer is currently used for an object selection
        // or if there is an user dialog currently
        if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
        
        if (L_BACK_BASEOBJECTS.isEmpty()) return;

        // deplace objectInBeamer__ in lists
        // and enable the right button
        BaseObject selectedObject = (BaseObject)
            L_BACK_BASEOBJECTS.get(L_BACK_BASEOBJECTS.size() - 1);
        L_FORWARD_BASEOBJECTS.add(objectInBeamer__);
        forwardButton__.setEnabled(true);
        L_BACK_BASEOBJECTS.remove(L_BACK_BASEOBJECTS.size() - 1);
        if (L_BACK_BASEOBJECTS.isEmpty()) backButton__.setEnabled(false);

                // for testing and tracking purpose
              //if (objectInBeamer__ != null)
              //  System.out.println("in BeamerManager back button listener,"+
              //    " store the current object in the forward list:"+
              //    objectInBeamer__.toString());

        updateFromButtons__ = true;
        
        TreeManager.selectNodeInExplorer(selectedObject);
        selectedObject = null;
        return;
      } // actionPerformed()
    }); // backButton__.addActionListener

    forwardButton__.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {

        // no action if the explorer is currently used for an object selection
        // or if there is an user dialog currently
        if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
          
        // variables
        DefaultMutableTreeNode node = TreeManager.EMPTY_TREE_NODE;
        // deplaces objectInBeamer__ in lists
        // and enables the right button
        if (!L_FORWARD_BASEOBJECTS.isEmpty()) {
          BaseObject selectedObject = (BaseObject)L_FORWARD_BASEOBJECTS.
            get(L_FORWARD_BASEOBJECTS.size() - 1);
          L_BACK_BASEOBJECTS.add(objectInBeamer__);
          // limit the back list size to BACK_SIZE_LIMIT
          if (L_BACK_BASEOBJECTS.size() >= BACK_SIZE_LIMIT) {
            L_BACK_BASEOBJECTS.subList(0, L_BACK_BASEOBJECTS.size() - BACK_SIZE_LIMIT).clear();
          }
          backButton__.setEnabled(true);
          L_FORWARD_BASEOBJECTS.remove(L_FORWARD_BASEOBJECTS.size() - 1);
          if (L_FORWARD_BASEOBJECTS.isEmpty()) forwardButton__.setEnabled(false);

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager forward button listener,"+
                  //  " store the current object in the back list:"+objectInBeamer__.toString());

          updateFromButtons__ = true;
          // changes the selection in the explorer
          if (selectedObject instanceof DetailledName && selectedObject instanceof ACSObject) {
            node = ACSTreeUtilities.getDetailledNameNodeInExplorer(
              (DetailledName) selectedObject, ((ACSObject)selectedObject).getAcsName());
          } // if
          else if (selectedObject instanceof BaseObject) {
            node = ACSTreeUtilities.getCurrentObjectNode(
              (DefaultMutableTreeNode) InternalFrameModel.getTreeInExplorer().getModel().getRoot(),
              selectedObject);
          } // if
          if (node == null) return;
          TreePath path = new TreePath(node.getPath());
          InternalFrameModel.getTreeInExplorer().setSelectionPath(path);

          selectedObject = null;
          node = null;
          path = null;
        } // if
        return;
      } // actionPerformed()
    }); // forwardButton__.addActionListener

  } // static initialization

      /* the only instance of the class, following singleton pattern */
  private static final BeamerManager  INSTANCE = new BeamerManager();

  // ------------------------- Constructors -------------------------------

  /**
   * With this private constructor, the compiler won't generate
  * a default public constructor.
  */
  private BeamerManager() {
    initializeDeleteButton();
  } // BeamerManager()

  /**
  * The single method to obtain the unique instance of BeamerManager.
  * @return BeamerManager
  */
  public static BeamerManager getInstance() {
    return INSTANCE;
  } // getInstance()


  /**
  * The method to obtain the initialization of the delete button,
  * before any use of setDeleteButton(). Used only if
  * beanDescriptor.getValue("withDeleteMethod") is true.
  * Calls UtilityImpl.selectBeanInfo(),
  * CollectionPropertyEditor.valueOf() and removeWithSpecificMethod().
  */
  public static void initializeDeleteButton() {

    deleteButton__.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {

        // no action if the explorer is currently used for an object selection
        // or if there is an user dialog currently
        if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
          
        if (!DialogBox.showConfirmDialog(
          " Do you want to delete this element?", null)) return;

        try {
                // for AclEntry only, into the genric gBase
          if (objectInBeamer__ != null && objectInBeamerBeanInfo__ != null &&
            objectInBeamerManager__ != null &&
            objectInBeamerBeanInfo__.getBeanDescriptor().
            getValue("externalFinalizeProperty") != null ) {

              // 'externalFinalizeProperty' is the string name of the property,
              // if it exists (for exemple, AclRights),
              // under which the finalize method for this object is described
              // in the BeanInfo of the BaseBeanManager.
              String propertyString =
                ((String[])objectInBeamerBeanInfo__.getBeanDescriptor().
                  getValue("externalFinalizeProperty"))[0];
              // searches for the right property descriptor
              PropertyDescriptor[]  l_propertyDescriptor =
                                      UtilityImpl.getFirstInstance().
                                        selectBeanInfo((BaseObject)objectInBeamerManager__).
                                            getPropertyDescriptors();
              int propertyIndex = 0;
              for (int i = 0; i < l_propertyDescriptor.length; i++) {
                if (propertyString.equals(
                  (String)l_propertyDescriptor[i].getName() )) {
                      propertyIndex = i;
                      break;
                } // if
              } // for
                   // @exception NullPointerException if _source or _displayName is null
                   // @exception BaseError if the property type is not equal to _propertyType
                   // @exception ProcessError if BeanInfo and PropertyDecriptor calls fail
                   // @exception ClassNotFoundException, IllegalAccessException, InvocationTargetException,
                   //   InstantiationException if removeWithSpecificMethod() fails
              CollectionPropertyEditor propertyEditor = (CollectionPropertyEditor)
                CollectionPropertyEditor.valueOf(
                  (BaseObject) objectInBeamerManager__,
                  l_propertyDescriptor[propertyIndex]);
              propertyEditor.removeWithSpecificMethod(objectInBeamer__);

              propertyString = null;
              l_propertyDescriptor = null;
              propertyEditor.finalizeForProcess();
              propertyEditor = null;

                // clear the beamer
              finalizePropertyEditors();
              return;          
          } // if
          else if (FinalizedObjectForUser.class.isInstance(objectInBeamer__) ) {

              ((FinalizedObjectForUser)objectInBeamer__).finalizeForUser();

                        // for testing and tracking purpose
                      //System.out.println("in BeamerManager.initializeDeleteButton(), "+
                        //"the object is finalized " );

                // clear the beamer
              finalizePropertyEditors();
              return;
          } // if
          else {
            new ProcessError("in BeamerManager, "+
              "the object in the beamer can't be deleted.");
            DialogBox.showErrorMessage(
              "The element in the beamer can't be deleted.");
            return;
          } // else

        } catch (Exception u) {
          new ProcessError("in BeamerManager, "+
            "the object in the beamer can't be deleted, and the error is: ");
          u.printStackTrace();
          DialogBox.showErrorMessage(
            "The element in the beamer can't be deleted, and the error is: "+
            u.toString() );
        }

      } // actionPerformed()
    }); // deleteButton__.addActionListener
    
    deleteButton__.setToolTipText("Delete the object which is visible in the beamer");
  } // initializeDeleteButton()

  //
  // ------------------------- Public Methods -------------------------------
  //

  /**
  * The method to obtain the beamer pane, which may be a JPanel or a JTabbedPane.
  * @return the JComponent
  */
  public static JComponent getBeamerPane() {
    return beamerPane__;
  } // getBeamerPane()

  /**
  * This is the only method to call just after getInstance() to
  * set the right beamer pane, that is always a JPanel so far,
  * and to initialize it with all the panels associated to the
  * properties of the base object to display.
  * Uses the BeanDescriptor of each property class.
  * Called by ActionOpenBeamer.openSelection().
  * Calls finalizePropertyEditors(), reinitializePanes() and initializeObjectClosing(),
  * and UtilityImpl.selectBeanInfo().
  * @param _objectInBeamer is the object to display in the beamer
  * @exception ProcessError in initializeObjectClosing()
  *   if BeanInfo.getBaseBeanManagerGetMethod() returns a wrong method
  * @return the beamer
  */
  public JComponent updateBeamerPane(BaseObject _objectInBeamer)
    throws ProcessError {

    if (_objectInBeamer == null) throw new ProcessError(
      "In BeamerManager.updateBeamerPane(), the parameter is null.");

    // stores the last objectInBeamer__ if necessary
    setBFButtons();

    // stores the previous JTabbedPane title to display the same pane
    // in the next JTabbedPane, if it exists
    if (centerPanel__ instanceof JTabbedPane &&
      ((JTabbedPane)centerPanel__).getSelectedIndex() != -1)
        previousTabbedPaneTitle__ = ((JTabbedPane)centerPanel__).getTitleAt(
          ((JTabbedPane)centerPanel__).getSelectedIndex());

    // finalizes the old elements, set objectInBeamer__, centerPanel__ to null
    finalizePropertyEditors();

    // prepares the beamer
    objectInBeamer__ = _objectInBeamer;
    objectInBeamerName__ = _objectInBeamer.getName();
    objectInBeamerBeanInfo__ = UtilityImpl.getFirstInstance().
                                selectBeanInfo(_objectInBeamer);

    // initializes centerPanel__
    reinitializePanes(objectInBeamerBeanInfo__, _objectInBeamer);

    // initializes the event handling for deleting or closing operations
    // on the displayed object
    initializeObjectClosing(_objectInBeamer);

    return beamerPane__;
  } // updateBeamerPane()

  /**
   * Set an empty beamer when an event is fired by the closing JInternalFrame
   * of the beamer (event 'beamerClosing'), or by an Action class when an ACS or
   * a view is closing (event 'objectClosing'), or by the BaseBeanManager 
   * of the displayed object when this object is closed or deleted.
   * When it is a closing or a deleting, the new value in the fired event has to be
   * a map where the name is the key, a collection of names,
   * or an array of names or ACSObject instances. The new value may be null or empty,
   * and this produces the finalization of the beamer panel.
   * For instance, AclEntries and Privileges are deleted with the firing of 
   * an ACSObject array as the new value.
   * No control of the bound property name, since there is only one. 
   * The current displayed name or ACSObject is searched for to finalize the beamer panel
   * if it is no more there.
   */
  public void propertyChange(PropertyChangeEvent _evt) {
    // initialization
    boolean emptyBeamer = false;
    boolean containsObject = false;

    // if the beamer displays an object managed by the source of the event
    if (_evt.getSource() == objectInBeamerManager__) {

      // if the beamer displays a closing object which is not in the map
      // send as the new value
      if ((_evt.getNewValue() instanceof Map)
        && !((Map)_evt.getNewValue()).containsKey(objectInBeamerName__)) {

          emptyBeamer = true;

      } // if ((_evt.getNewValue() instanceof Map)

      // if the beamer displays a closing object which is not in the collection
      // send as the new value
      if ((_evt.getNewValue() instanceof Collection)
        && !((Collection)_evt.getNewValue()).contains(objectInBeamerName__)) {

          emptyBeamer = true;

      } // if ((_evt.getNewValue() instanceof Collection)

      // if the beamer displays a closing object which is not in the array
      // of names or ACSObjects
      else if (_evt.getNewValue() instanceof Object[]) {

        if ( ((Object[])_evt.getNewValue()).length != 0) {
            boolean isNamesArray = false;
            if (((Object[])_evt.getNewValue())[0] instanceof Name)
                isNamesArray = true;

            for (int i = 0; i < ((Object[])_evt.getNewValue()).length; i++) {
                if (isNamesArray &&
                        ((Name)((Object[])_evt.getNewValue())[i]).equals(objectInBeamerName__)) {
                    containsObject = true;
                    break;
                } // if
                else if (((Object[])_evt.getNewValue())[i] == objectInBeamer__) {
                    containsObject = true;
                    break;
                } // if
            } // for (int i

        } // if _evt.getNewValue() != null

        if (!containsObject) emptyBeamer = true;

      } // if (_evt.getNewValue() instanceof Object[])

      else if (_evt.getNewValue() == null)
          emptyBeamer = true;

      // clears the beamer
      if (emptyBeamer) {

            // for testing and tracking purpose
          //System.out.println("in BeamerManager.propertyChange(), "+
            //"the beamer is cleared " );

        //InternalFrame.openBeamer(new JPanel(), null);
        finalizePropertyEditors();
      }

    } // if (_evt.getSource() == objectInBeamerManager__

    // if the beamer is closing
    else if (_evt.getPropertyName().equals("beamerClosing") ) {
      finalizePropertyEditors();
      finalizeForUser();
    }

    // if the ACS or the view is closing
    else if (_evt.getPropertyName().equals("objectClosing") ) {
      finalizePropertyEditors();
      finalizeForUser();
    }
  } // propertyChange()

  /**
  * Set centerPanel__ when it is a JPanel.
  * Called by reinitializePanes().
  * @param _objectInBeamer to display in the beamer
  * @param _l_propertyDescriptor is array of the BeanInfo property descriptors
  * @param _result is the tabbed pane where the property editors are added
  * @param _m_currentPropertyEditors is the map of the current property editors
  * @exception ProcessError if a CommonPropertyEditor throws an exception
  *   in addPropertyEditor()
  */
  public static void setJPanel(BaseObject _objectInBeamer,
    PropertyDescriptor[] _l_propertyDescriptor, JPanel _result,
    Map _m_currentPropertyEditors) throws ProcessError {
      // initialization
      JPanel                  groupPanel;
      Box                     groupBox;
      JPanel                  propertyPanel;
      String[]                l_panelGroupNames = getOrderedNames(
                                 _objectInBeamer, _l_propertyDescriptor, "panelGroupName");
      CommonPropertyEditor    propertyEditor = null;
      
      // group the property editors with the same panel group name
      // in the same groupPanel
      for (int j = 0; j < l_panelGroupNames.length; j++) {
        groupPanel = new JPanel();
        groupPanel.setBackground(InternalFrame.SOFTGREY);
        groupPanel.setBorder(new TitledBorder(l_panelGroupNames[j]));
        groupPanel.setFont(plainFont);
        groupBox = Box.createVerticalBox();
        groupBox.add(Box.createVerticalStrut(10));
        groupPanel.add(groupBox);

        for (int i = 0; i < _l_propertyDescriptor.length; i++) {
          if (l_panelGroupNames[j].
            equals((String)_l_propertyDescriptor[i].getValue("panelGroupName"))) {

              propertyPanel = new JPanel();
              // put the right property editor in propertyPanel
              propertyEditor = addPropertyEditor(
                _objectInBeamer, _l_propertyDescriptor[i], propertyPanel, false);
              _m_currentPropertyEditors.put(_l_propertyDescriptor[i].getDisplayName(),
                propertyEditor);
              
              propertyPanel.setBackground(InternalFrame.BACKCOLOR_3);

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.updateBeamerPane(),"+
                  //" for the object:"+beanDescriptor.getDisplayName()+
                  //" add the editor for the property descriptor: "+
                  //_l_propertyDescriptor[i].getShortDescription());

              groupBox.add(propertyPanel);
              //propertyPanel.setLocation(0,0);
          } // if (l_panelGroupNames[j].
        } // for (int i

        // adds each panel horizontally
        _result.add(groupPanel);
        _result.setFont(plainButtonFont);
      } // for j

      groupBox = null;
      groupPanel = null;
      propertyPanel = null;
      l_panelGroupNames = null;
      return;
  } // setJPanel()

  /**
  * Set the properties of a BaseObject in a JTabbedPane, based on the property descriptors
  * and for a given JTabbedPane. Tabbed panes are the default configuration of the beamer.
  * Display only the tabbed panes and properties the ACS authorizes.
  * For adding to a JTabbedPane some new tabs, use the current _result and
  * _m_currentPropertyEditors as arguments, and the new tabs will be added 
  * at the end without updating the current tabs. <P>
  * Calls getOrderedNames(), getOrderedPropertyDescriptors()
  * and addPropertyEditor().
  * When called by reinitializePanes(), reads previousTabbedPaneTitle__
  * without updating, to select the same pane if possible.
  * Add a closing button on each pane when it is called by a TabClosingHandler. <P>
  * Called by reinitializePanes() to sets centerPanel__, by ActionNewACSyst
  * to define the ACS properties, by ActionOpenAllFullViews, ActionOpenGlobalStructure,
  * ActionOpenAllNoThanViews and ActionOpenGraphicView to display the views.
  * @param _objectInPane to display in the tabbed pane. Used
  *    if it is an ACSObject or an ACS. May be null otherwise.
  * @param _l_propertyDescriptor is an array of the BeanInfo property descriptors.
   *    For a GraphicViewPropertEditor, most of the values are not used.
  * @param _result is the tabbed pane where the property editors are added
  * @param _m_currentPropertyEditors is the map to update with the property editors.
  *    Used only to finalize them.
  * @param _handler may be null. Have to be null for the calls from this class.
  * If it is a TabClosingHandler as ActionClose, the tab closing is initialized.
  * Noop otherwise.
  * @exception ProcessError if a CommonPropertyEditor throws an exception
  *   in addPropertyEditor()
  */
  public static void setJTabbedPane(BaseObject _objectInPane,
    PropertyDescriptor[] _l_propertyDescriptor, JTabbedPane _result,
    Map _m_currentPropertyEditors, Object _handler) throws ProcessError {
      // initialization
      JPanel                groupPanel = null;
      JPanel                propertyPanel;
      // select the pane names following the structure
      String[]              l_orderedName = getOrderedNames(
                              _objectInPane, _l_propertyDescriptor, "tabbedPaneName");
      PropertyDescriptor[]  l_OrderedPropertyDescriptor =
                              getOrderedPropertyDescriptors(_objectInPane,
                              _l_propertyDescriptor, l_orderedName);
      JPanel[]              l_tabbedPanels = new JPanel[l_orderedName.length];
      Box[]                 l_groupBox = new Box[l_orderedName.length];
      Box                   panelBox = l_groupBox[0]; // no-sense init
      int                   index = 0;
      String                panelGroupName;
      TitledBorder          titledBorder;
      boolean               hasBeenProcessed = false;
      CommonPropertyEditor  propertyEditor = null;
      int                   tabCount = _result.getTabCount();
      

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.setJTabbedPanee(),"+
                  //" for the object "+_objectInPane.getName()+
                  //", l_orderedName is: ");
                  //for (int i = 0; i < l_orderedName.length; i++) {
                    //System.out.println("index "+i+": "+l_orderedName[i]);
                  //}
                  //System.out.println("and l_OrderedPropertyDescriptor is: ");
                  //for (int i = 0; i < l_OrderedPropertyDescriptor.length; i++) {
                    //System.out.println("index "+i+": "+l_OrderedPropertyDescriptor[i].getDisplayName());
                  //}

      // builds up the map of tabbed panels and the group boxes
      for (int j = 0; j < l_orderedName.length; j++) {
        l_tabbedPanels[j] = new JPanel();
        l_tabbedPanels[j].setBackground(InternalFrame.SOFTGREY);
        l_tabbedPanels[j].setForeground(Color.black);
        l_tabbedPanels[j].setFont(plainFont);
        l_groupBox[j] = Box.createVerticalBox();
        l_tabbedPanels[j].add(l_groupBox[j]);
      } // for j

      // builds up the property descriptors in the group boxes
      for (int j = 0; j < l_orderedName.length; j++) {
        for (int i = index; i < l_OrderedPropertyDescriptor.length; i++) {

          if (!l_orderedName[j].equals(l_OrderedPropertyDescriptor[i].
            getValue("tabbedPaneName"))) break;

          // incrementes index to start the previous test at the right position
          // at the next run of the i loop
          index++;
          panelGroupName = (String)l_OrderedPropertyDescriptor[i].
            getValue("panelGroupName");

          // sets all the property descriptors without panel group names
          if (panelGroupName == null ) {

              // puts the right property editor in the tabbed pane
              if (!hasBeenProcessed) {
                  propertyPanel = new JPanel();
              
                  if (l_OrderedPropertyDescriptor[i].getReadMethod().
                          getReturnType() != int.class &&
                    l_OrderedPropertyDescriptor[i].getReadMethod().
                            getReturnType() != boolean.class
                            ) {
                        
                          propertyEditor = addPropertyEditor(
                            _objectInPane, l_OrderedPropertyDescriptor[i], propertyPanel, false);
                          _m_currentPropertyEditors.put(
                                      l_OrderedPropertyDescriptor[i].getDisplayName(),
                                      propertyEditor);
                      
                  } else {

                      // if there is a second boolean (or integer) property after
                      // this boolean (or integer) property,
                      // adds the second property in the same property panel
                      if (i+1 < l_OrderedPropertyDescriptor.length &&
                              l_orderedName[j].equals(l_OrderedPropertyDescriptor[i+1].
                                    getValue("tabbedPaneName")) &&
                        (l_OrderedPropertyDescriptor[i+1].getReadMethod().getReturnType() == int.class ||
                        l_OrderedPropertyDescriptor[i+1].getReadMethod().getReturnType() == boolean.class) ) {

                              propertyEditor = addPropertyEditor(
                                _objectInPane, l_OrderedPropertyDescriptor[i],
                                propertyPanel, false);
                              _m_currentPropertyEditors.put(
                                          l_OrderedPropertyDescriptor[i].getDisplayName(),
                                          propertyEditor);
                              propertyEditor = addPropertyEditor(
                                _objectInPane, l_OrderedPropertyDescriptor[i+1],
                                propertyPanel, false);
                              _m_currentPropertyEditors.put(
                                          l_OrderedPropertyDescriptor[i].getDisplayName(),
                                          propertyEditor);
                              
                              hasBeenProcessed = true;   

                      // there is only one boolean property to add with a larger panel
                      } else {
                          propertyEditor = addPropertyEditor(
                            _objectInPane, l_OrderedPropertyDescriptor[i],
                            propertyPanel, true);
                          _m_currentPropertyEditors.put(
                                    l_OrderedPropertyDescriptor[i].getDisplayName(),
                                    propertyEditor);
                      } // else
                  
                  } // else
                        
                                // for testing and tracking purpose
                              //System.out.println("in BeamerManager.setJTabbedPane(),"+
                              //" for the object:"+_objectInPane.getName()+" at"+i+
                              //" add the editor for the property descriptor: "+
                              //l_OrderedPropertyDescriptor[i].getDisplayName());
                  
                  l_groupBox[j].add(propertyPanel);

              } else hasBeenProcessed = false; // reinit in the second boolean property loop         
              
          } // if panelGroupName == null

          // sets the property descriptor with panel group names
          else {
              if (groupPanel == null) {
                groupPanel = new JPanel();
                panelBox = Box.createVerticalBox();
                groupPanel.add(panelBox);
                titledBorder = new TitledBorder(panelGroupName);
                titledBorder.setTitleFont(italicFont);
                groupPanel.setBorder( titledBorder );
                groupPanel.setBackground(Color.white);
                groupPanel.setForeground(Color.black);
              } // if

              // puts the right property editor in the tabbed pane
              if (!hasBeenProcessed) {
                  propertyPanel = new JPanel();
              
                  if (l_OrderedPropertyDescriptor[i].getReadMethod().getReturnType() != Boolean.class &&
                    l_OrderedPropertyDescriptor[i].getReadMethod().getReturnType() != boolean.class) {
                        
                          propertyEditor = addPropertyEditor(
                            _objectInPane, l_OrderedPropertyDescriptor[i], propertyPanel, false);
                          _m_currentPropertyEditors.put(
                                      l_OrderedPropertyDescriptor[i].getDisplayName(),
                                      propertyEditor);
                      
                  } else {

                      // if there is a second boolean property after this boolean property,
                      // adds the second property in the same property panel
                      if (i+1 < l_OrderedPropertyDescriptor.length &&
                              l_orderedName[j].equals(l_OrderedPropertyDescriptor[i+1].
                                    getValue("tabbedPaneName")) &&
                              (l_OrderedPropertyDescriptor[i+1].getReadMethod().getReturnType() == Boolean.class ||
                        l_OrderedPropertyDescriptor[i+1].getReadMethod().getReturnType() == boolean.class) ) {

                              propertyEditor = addPropertyEditor(
                                _objectInPane, l_OrderedPropertyDescriptor[i], propertyPanel, false);
                              _m_currentPropertyEditors.put(
                                          l_OrderedPropertyDescriptor[i].getDisplayName(),
                                          propertyEditor);
                              propertyEditor = addPropertyEditor(
                                _objectInPane, l_OrderedPropertyDescriptor[i+1], propertyPanel, false);
                              _m_currentPropertyEditors.put(
                                          l_OrderedPropertyDescriptor[i].getDisplayName(),
                                          propertyEditor);
                              
                              hasBeenProcessed = true;   

                      // there is only one boolean property to add with a larger panel
                      } else {
                          propertyEditor = addPropertyEditor(
                            _objectInPane, l_OrderedPropertyDescriptor[i], propertyPanel, true);
                          _m_currentPropertyEditors.put(l_OrderedPropertyDescriptor[i].getDisplayName(),
                            propertyEditor);
                      } // else
                  
                  } // else
                        
                        // for testing and tracking purpose
                      //System.out.println("in BeamerManager.setJTabbedPane(),"+
                      //" for the object:"+_objectInPane.getName()+
                      //" add the editor for the property descriptor: "+
                      //_l_propertyDescriptor[i].getShortDescription());

                  panelBox.add(propertyPanel);
              } else hasBeenProcessed = false; // reinit in the second boolean property loop
              

              // if it is the last descriptor, puts the group panel in the group box
              if (i < l_OrderedPropertyDescriptor.length - 1 &&
                !panelGroupName.equals(
                  l_OrderedPropertyDescriptor[i+1].getValue("panelGroupName"))) {
                    l_groupBox[j].add(groupPanel);
                    groupPanel = null;
                    panelBox = null;
              } // if
              else if (i == l_OrderedPropertyDescriptor.length - 1) {
                    l_groupBox[j].add(groupPanel);
                    groupPanel = null;
                    panelBox = null;
              } // else
          } // if "panelGroupName") != null

        } // for (int i = 0
      } // for j

      _result.setFont(plainFont);
      
      // builts up the JTabbedPane
      for (int j = 0; j < l_orderedName.length; j++) {
        _result.addTab(l_orderedName[j], l_tabbedPanels[j]);
        
        // to have a special icon when the GraphicNoThanView has no_compliance state
        // insertTab(String title,
                    //  Icon icon,
                    //  Component component,
                    //  String tip,
                    //  int index)
        

        if (_handler instanceof TabClosingHandler) {
            // for ActionOpenAllFullViews, add a closing button
            _result.setTabComponentAt(tabCount+j, initializeViewTabComponent(
                    _result, l_orderedName[j], (TabClosingHandler)_handler));
            // and register the JPanel for the ViewInBase
            InternalFrameModel.addBaseObjectJComponent(
                    (BaseObject)ViewInBase_Facade.getAllStandardOpenViews().get(
                           ViewModel.getViewNameFromTabTitle(l_orderedName[j])),
                    l_tabbedPanels[j]);
        // for the Beamer, select the previous pane if possible
        } else if (l_orderedName[j].equals(previousTabbedPaneTitle__))
            _result.setSelectedIndex(tabCount+j); 

        _result.setBackgroundAt(tabCount+j, InternalFrame.BACKCOLOR_5);
      } // for


                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.setJTabbedPane(),"+
                  //  " centerPanel__ is: "+centerPanel__.toString());

      propertyPanel = null;
      l_OrderedPropertyDescriptor = null;
      groupPanel = null;
      propertyPanel = null;
      l_orderedName = null;
      l_tabbedPanels = null;
      l_groupBox = null;
      titledBorder = null;
      return;
  } // setJTabbedPane()  
  
  /**
  * Detect the type for a property type from a property descriptor, since
  * the returned value by getPropertyType() may be an elementary BaseObject,
  * or an array, a map or collection of BaseObjects, Strings or booleans.
  * For an array, the returned type is the type of the array elements, or
  * is given through the key 'valueTypeInGroup'.
  * For a collection or a map, the returned type is given through the key 'valueTypeInGroup'.
  * For a map, it is recommanded to set this key with the type of the values, 
  * if the value is not a collection or an array, and otherwise, to use the type
  * of the elements in this collection or array which is the map value.
  * Called by CollectionPropertyEditor.valueOf().
  * Calls UtilityImpl.getClassNameFromArrayOrInterface() and
  * getACSVocabularyFromArrayOrInterface().
  * @param _propertyDescriptor is the descriptor of the property
  * @return the elementary property type. May be the native property type.
  */
  public static Class<?> getElementaryPropertyType(PropertyDescriptor _propertyDescriptor) {
    Class<?> result = null;
    Class<?> propertyType = _propertyDescriptor.getPropertyType();
    Class<?> element = null;
    

                          // for testing purpose
                          //if (_propertyDescriptor.getName().equals("OneEPVirtualFolder"))
                            //  System.out.println("in BeamerManager.getElementaryPropertyType(),"+
                              //        " the property is 'OneEPVirtualFolder' "+
                                //      " and the propertyType is: "+propertyType);
    
    // initialization if the property is an array
    if (propertyType.isArray()) {
       try { 
           // for VirtualFolderMySQLImpl which has a specific constructor described by
           // the VirtualFolderMySQL.Children property descriptor
           if (_propertyDescriptor.getValue("valueTypeInGroup") != null) {
                element = (Class<?>)((Object[])_propertyDescriptor.
                  getValue("valueTypeInGroup"))[0];
           }
            // it is an array of Java classes or interfaces, so we have to avoid '[L' and ';'
            // at the beginning and at the end of the name
           else {
                element = Class.forName(propertyType.getName().
                    substring(2, propertyType.getName().length() - 1) );
           }

            // if propertyType_ is an array of the gBase classes or interfaces,
            // from the generic gBase or from an AcsAddon package,
            // get the direct class implementing the interface, excepted for StringRight
            // since there are 2 different implementations (StringRightImpl and StringMetaRightImpl)
            // and it is necessary to select simply StringRight if 'valueTypeInGroup' is not used,
          if (BaseObject.class.isAssignableFrom(element) &&
                   ((element.getName().indexOf("StringRight") == -1 &&
                   element.getName().indexOf("StringMetaRight") == -1) ||
                    _propertyDescriptor.getValue("valueTypeInGroup") != null)) {
              // result is never null since it is never an interface without an
              // implementation by a direct class, like BaseObject or Node
            result = Class.forName(
              UtilityImpl.getFirstInstance().getClassNameFromArrayOrInterface(
                UtilityImpl.getFirstInstance().
                    getACSVocabularyFromArrayOrInterface(element.getName()),
                element.getName()));
          } // if
          else {
            // the case when 'valueTypeInGroup' is DetailledName.
              result = element;
          } // else

                          // for testing purpose
                          //if (_propertyDescriptor.getName().equals("Children"))
                            //  System.out.println("in BeamerManager.getElementaryPropertyType()"+
                              //    "for an array, BaseObject.class.isAssignableFrom(propertyType): "+
                                //  BaseObject.class.isAssignableFrom(element)+
                                  //" - the result is: "+result);

       } catch (ClassNotFoundException c) {
           throw new InternalError("in BeamerManager.getElementaryPropertyType(), "+
                   "the BeanInfo has a bug since the class is not found.");   
       }
    } // if (_propertyType.isArray())

    // initialization if the property is a collection or a map
    else if (Map.class.isAssignableFrom( propertyType) || 
            Collection.class.isAssignableFrom(propertyType)) {

                              // for testing purpose
                              //System.out.println("in BeamerManager.getElementaryPropertyType(), "+
                                //      "Map or Collection for the property "+
                                  //    _propertyDescriptor.getName());

                result = (Class<?>)((Object[])_propertyDescriptor.
                                            getValue("valueTypeInGroup"))[0];

    } // if (Collection or Map.class.isAssignableFrom(result.propertyType_)
    else result = propertyType;

                              // for testing purpose
                              //if (_propertyDescriptor.getName().equals("OneEPVirtualFolder"))
                                //  System.out.println("in BeamerManager.getElementaryPropertyType(), "+
                                  //    "for 'OneEPVirtualFolder', result "+result);
              
    propertyType = null;
    element = null;
    return result;
  } // getElementaryPropertyType()
  
  /**
  * Detect the subtype which is allowed by the acs, if there are some restrictions.
  * The only restriction is for GroupIDMember which has to be an UserID for
  * the ACS where getManageGroupTrees() returns false. 
  * Called by CollectionPropertyEditor.actionPerformed().
  * <P>
  * @param _object is the object with _type as type
  * @param _type is the general type of _object
  * @return the subtype which is allowed by the acs. May be _type.
  */
  public static Class<?> getACSAllowedType(ACSObject _object, Class<?> _type) {
   
    if (!ImmutableGroupIDMember.class.equals(_type) &&
            !GroupIDMember.class.equals(_type) &&
            !GroupIDMemberImpl.class.equals(_type)) return _type;
    else if ( !((ACS)ACS_Facade.getOpenACS().get( _object.getAcsName() )).
        getManageGroupTrees()) {

              // for testing purpose
              //System.out.println("in BeamerManager.getACSAllowedType(), return userID.");        
        
        return ImmutableUserID.class;
    }
    else return _type;
    
  } // getACSAllowedPropertyType()  
  

  /**
  * Gets the ACS configuration through the call of a boolean method.
  * @param _acs of the object to display
  * @param _method is the name of an ACS method which returns a boolean
  * @return the returned value by the ACS method
  */
  public static boolean getACSboolean(ACS _acs, String _method) {
      
      Boolean result = false; 
      Method[] l_methods = ACS.class.getMethods();
      
      for (int k = 0; k < l_methods.length; k++){

                    //System.out.println("in BeamerManager.getACSboolean(),"+
                    //  " ACS methods: "+l_methods[k].getName());

            try {
              if (l_methods[k].getName().equals(_method)) {

                        //System.out.println("in BeamerManager.getACSboolean(),"+
                          //" right source method: "+l_methods[k].getName());  

                  result = (Boolean)l_methods[k].invoke(_acs, new Object[] {});
                  break;
              } // if (l_methods[k]
            } catch (Exception e) {
                  throw new InternalError("In BeamerManager.getACSboolean(): "
                    +e.toString());
            } // catch
      } // for k
      l_methods = null;
      return result.booleanValue();
  }  // getACSboolean()
  
  //
  // ------------------------- PACKAGE METHODS ------------------------------
  // 

  /**
  * Unused - this method forbids the display of the 'Select' and 'Remove' buttons for the
  * members of the 'Architecture' view, and forbids the 'ConnectedACS' property. 
  * Called by SimplePropertyEditor.getButtonsDisplaying() and
  * readDescriptorIndexesForTabbedNames().
  * @param _source is the object to display in the property editor
  * @param _propertyDescriptor is one property descriptor from the BeanInfo class
  * @param _buttonType is the property value for checking the right to display
  *   the button 
  * @return true only for the 'Architecture' view, to forbid the updating of members and
  *  the display of the ConnectedACS property; false otherwise.
  */
  static boolean getForbidsForArchitecture(BaseObject _source,
          PropertyDescriptor _propertyDescriptor, String _buttonType) {
      
      boolean result = false; 
      if (_source instanceof ViewInBase &&
              _source.getName().equals(ActionOpenArchitecture.ARCHI_VIEW_NAME)) {
                if (_propertyDescriptor.getName().equals("OneBaseObjectInFolder") &&
                  _buttonType != null ) {
                    if (_buttonType.equals("withSelecButton")) result = true;
                    else if (_buttonType.equals("withRemoveButton")) result = true;
                }
                else if (_propertyDescriptor.getName().equals("ConnectedACS") )
                    result = true;
      } // if
      return result;
  }  // getForbidsForArchitecture()  

  /**
  * This is wrapper of ACS_Facade.getEorBaseObject() to get a StringRight
  * while there is an on-going entering of the properties for an ACSFactory.
  * Called by SimplePropertyEditor.viewOneElement(), CollectionPropertyEditor.removeElements().
  * Calls ACS_Facade.getEorBaseObject(), or BaseUtilityImpl.getRightFromNameAsString()
  * only if _type is an instance of StringRight or an array of StringRight.
  * @param _name name of the BaseObject
  * @param _type class of the BaseObject or array class. It is ridden only if _type
  * is an instance of StringRight or an array of StringRight. Never null.
  * @return the BaseObject. May be null.
  */
  static BaseObject getBaseObject(ImmutableName _name, Class _type) {
      
    BaseObject  result = null;
    Class       type = null;

    if (_type.isArray())
        type = _type.getComponentType();
    else
        type = _type;

    try {
        if (StringRight.class.isAssignableFrom(type) &&
                ActionUtilities.getInstance().getACSFactory() != null &&
                BaseUtilityImpl.isAcsRightName(_name) &&
                BaseUtilityImpl.getAcsNameFromName(_name).equals(
                    ActionUtilities.getInstance().getACSFactory().getName() ) ) {

            result = BaseUtilityImpl.getRightFromNameAsString(
                        _name.toString(),
                        ActionUtilities.getInstance().getACSFactory());            
        }
        else { 

                            // for testing and tracking purpose
                          //System.out.println("in BeamerManager.getBaseObject(),"+
                            //" for the base object "+_name);
                          //System.out.println("----- and _type: "+
                            //_type);
                          //System.out.println("----- BaseUtilityImpl.isAcsRightName(_name): "+
                            //BaseUtilityImpl.isAcsRightName(_name));
                          //System.out.println("----- BaseUtilityImpl.getAcsNameFromName(_name): "+
                            //BaseUtilityImpl.getAcsNameFromName(_name));
                          //System.out.println("----- ActionUtilities.getInstance().getACSFactory().getName(): "+
                            //ActionUtilities.getInstance().getACSFactory().getName());

            // it is the name which is displayed in the object list
            result = ACS_Facade.getEorBaseObject(_name);
        } // else             
    } catch (BaseError b) {
        DialogBox.showErrorMessage("This element cannot be displayed in the beamer.");
        b.printStackTrace();
        throw new InternalError("BeamerManager.getBaseObject(): "+
                "This element cannot be displayed in the beamer. "+b.toString());
    }
    return result;
  }  // getBaseObject() 
  
  //
  // ------------------------- PRIVATE METHODS -------------------------------
  //

  /**
  * Updates the back and forward buttons states.
  * Stores the last base object in the back list,
  * if it is not send from the back or the forward list.
  */
  private static void setBFButtons() {

    if (objectInBeamer__ != null && !updateFromButtons__) {
        backButton__.setEnabled(true);
        forwardButton__.setEnabled(false);
          //L_BACK_BASEOBJECTS.addAll(L_FORWARD_BASEOBJECTS);
        // clear the forward list
        L_FORWARD_BASEOBJECTS.clear();
        // store the current object in the back list
        L_BACK_BASEOBJECTS.add(objectInBeamer__);

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.updateBeamerPane(),"+
                  //  " store the current object in the back list:"+objectInBeamer__.toString());

        // limit the back list size to BACK_SIZE_LIMIT
        if (L_BACK_BASEOBJECTS.size() >= BACK_SIZE_LIMIT) {
          L_BACK_BASEOBJECTS.subList(0, L_BACK_BASEOBJECTS.size() - BACK_SIZE_LIMIT).clear();
        }
    } // if objectInBeamer__ != null
    else if (updateFromButtons__) updateFromButtons__ = false;
    // else, the operations has been done in the button listener

    return;
  } // setBFButtons()

  /**
  * Analyzes the BeanInfo to set the internal static variables beamerPane__
  * and centerPanel__ (new JTabbedPane or new JPanel).
  * Called by updateBeamerPane().
  * Calls setJTabbedPane() or setJPanel().
  * @param _info bean info for the object to put in the beamer
  * @param _objectInBeamer is the object to display in the beamer
  * @exception ProcessError if a CommonPropertyEditor throws an exception
  *   in addPropertyEditor()
  */
  private static void reinitializePanes(BeanInfo _info,
    BaseObject _objectInBeamer) throws ProcessError {

      // initialization
      BeanDescriptor        beanDescriptor = _info.getBeanDescriptor();
      PropertyDescriptor[]  l_propertyDescriptor = _info.getPropertyDescriptors();

                                  // for testing and tracking purpose
                                //System.out.println("in BeamerManager.reinitializePanes(),"+
                                  //" for the object:"+_objectInBeamer.getName()+
                                  //" beanDescriptor.getValue(withTabbedPanes): "+
                                  //((Boolean)beanDescriptor.getValue("withTabbedPanes")).toString()
                                  //);


      // tests if the beamer has to handle a JTabbedPane (the default configuration)
      if ((Boolean)beanDescriptor.getValue("withTabbedPanes")) {
          centerPanel__ = CENTER_JTABBEDPANE;
          if (m_currentPropertyEditors__ == null)
              m_currentPropertyEditors__ = new HashMap(40);
          setJTabbedPane(_objectInBeamer, l_propertyDescriptor, (JTabbedPane)centerPanel__,
            m_currentPropertyEditors__, null);
      } // if (((Boolean)beanDescriptor
      // the beamer has to handle a JPanel
      else {
          centerPanel__ = CENTER_JPANEL;
          if (m_currentPropertyEditors__ == null) m_currentPropertyEditors__ = new HashMap(40);
          setJPanel(_objectInBeamer, l_propertyDescriptor, (JPanel)centerPanel__,
            m_currentPropertyEditors__);
      } // else

      // centerPanel__ is opaque if the next element has been closed
      centerPanel__.setOpaque(false);
      // puts centerPanel__ in the beamer
      beamerPane__.add( centerPanel__, BorderLayout.CENTER );
      // sets the Delete button if necessary
      setDeleteButton(beanDescriptor);
      // shows the component display name in the beamer title
      if (_objectInBeamer instanceof AclEntry) 
          InternalFrameModel.getBeamer().setTitle("Beamer: "+
            beanDescriptor.getDisplayName()+" "+
            ((AclEntry)_objectInBeamer).getEpName().lastComponent()+
            " >> "+((AclEntry)_objectInBeamer).getResName().lastComponent()+"  ");
      else if (_objectInBeamer instanceof Privilege) 
          InternalFrameModel.getBeamer().setTitle("Beamer: "+
            beanDescriptor.getDisplayName()+" "+
            ((Privilege)_objectInBeamer).getNickName());
      else InternalFrameModel.getBeamer().setTitle("Beamer: "+
            beanDescriptor.getDisplayName()+" "+
            objectInBeamer__.getName().lastComponent()+"   ");

      beanDescriptor = null;
      l_propertyDescriptor = null;
  } // reinitializePanes()

  /**
  * Set the delete button if it is required in the bean descriptor.
  * @param _beanDescriptor
  */
  private static void setDeleteButton(BeanDescriptor _beanDescriptor) {

    if (_beanDescriptor.getValue("withDeleteMethod") == Boolean.TRUE ){
      TOOLBAR.add(deleteButton__);
    } // if
    else {
      TOOLBAR.remove(deleteButton__);
    }
  } // setDeleteButton()

  /**
  * Read the names for the panel and tabbed pane groups.
  * Called by setJTabbedPane().
  * If _objectInPane is an ACS or an ACSObject and _name is "tabbedPaneName",
  * remove the panes for the properties which are not allowed by the ACS.
  * call to getNoDisplayThisProperty().
  * If _objectInPane is the 'Architecture' view, remove the 'ConnectedACS' property.
  * @param _objectInPane to display in the tabbed pane
  * @param _l_descriptors from a BeanInfo
  * @param _name of the property, like "tabbedPaneName" and "panelGroupName"
  * @return array of the names. May be empty.
  */
  private static String[] getOrderedNames(BaseObject _objectInPane,
    PropertyDescriptor[] _l_descriptors, String _name) {
      // initialization
      String[]  l_temp = new String[_l_descriptors.length];
      String[]  l_result;
      Object    object = null;
      boolean   tabbedPaneNeedChecking = false;
      
      if ((_objectInPane instanceof ACSObject || _objectInPane instanceof ImmutableACS
              || _objectInPane instanceof StringMetaRight) 
              && _name.equals("tabbedPaneName"))
                   tabbedPaneNeedChecking = true;

      // count: number of null panelGroupNames
      int       count = 0;
      boolean   toJump = false;

      for (int i = 0; i < _l_descriptors.length; i++) {

                                  // for testing and tracking purpose
                                //if (_l_descriptors[i] != null)
                                    //System.out.println("in BeamerManager.getOrderedNames(),"+
                                  //" for the object:"+_objectInPane.getName()+
                                  //" processing of the property: "+_l_descriptors[i].getName());
                                //else
                                    //System.out.println("in BeamerManager.getOrderedNames(),"+
                                  //" for the object:"+_objectInPane.getName()+
                                  //" for i = "+i+", the property is null.");

        object = _l_descriptors[i].getValue(_name);
        if (object != null) {

          toJump = false;        
          // searches if object is already known          
          for (int j = 0; j < i - count; j++) {
            if (object.equals(l_temp[j])) {
              toJump = true;
              break;
            }
          }          
          if (toJump) {
            count++;
            continue;
          }          
          // searches if the tabbed pane is allowed by the acs
          if (tabbedPaneNeedChecking) {
              if (!getNotDisplayThisProperty(
                      _objectInPane, _l_descriptors[i])) {
                            count++;

                                  // for testing and tracking purpose
                                //System.out.println("in BeamerManager.getOrderedNames(),"+
                                  //" for the object:"+_objectInPane.getName()+
                                  //" no display of the property: "+_l_descriptors[i].getName());
                            
                            continue;                  
              }
          } // if
          // sets object in the temporary array  
          l_temp[i - count] = (String) object;
        } // (object != null)
        else count++;
      } // for i
      // removes null objects in l_temp
      l_result = new String[_l_descriptors.length - count];
      System.arraycopy(l_temp, 0, l_result, 0, _l_descriptors.length - count);


                      // for testing and tracking purpose
                    //System.out.println("in BeamerManager.getOrderedNames(),"+
                      //" for the object:"+_objectInPane.getName()+
                      //" the result is: ");
                   //for (int i = 0; i < l_result.length; i++) {
                     //System.out.println("in "+i+", the value is: "+
                       // l_result[i]);
                   //} // for

      return l_result;
  } // getOrderedNames()

  /**
  * For each panel group name, return the indexes of the descriptors
  * which are associated to.
  * Called by getOrderedPropertyDescriptors().
  * @param _l_descriptors array of descriptors from a BeanInfo, where some values may be null.
  * @return map where
  *   - the panel group name is the key,
  *   - the value is a List of all the indexes in _l_descriptors
  *     where the panel names are used. Is never empty.
  *   May be empty.
  */
  private static Map readDescriptorIndexesForPanelNames(PropertyDescriptor[] _l_descriptors) {
    // initialization
    Map       m_result = new HashMap(10);
    List      l_descriptorsForPanel;

    for (int i = 0; i < _l_descriptors.length; i++) {
      // searches all the descriptors with a panel group name
      if (_l_descriptors[i] != null && _l_descriptors[i].getValue("panelGroupName") != null) {

        // sets m_result with a new list if the panel group name is not registered
        if (!m_result.containsKey(_l_descriptors[i].getValue("panelGroupName"))) {
          l_descriptorsForPanel = new ArrayList(10);
          l_descriptorsForPanel.add(new Integer(i) );
          m_result.put(_l_descriptors[i].getValue("panelGroupName"),l_descriptorsForPanel);
        }
        // adds the index i in the right list
        else {
          ((List)m_result.get(_l_descriptors[i].getValue("panelGroupName"))).
            add(new Integer(i));
        }

      } // if "panelGroupName") != null
    } // for i

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.readDescriptorIndexesForPanelNames(),"+
                  //  " for the object:"+objectInBeamerBeanInfo__.getBeanDescriptor().getDisplayName()+
                  //  ", the result is: ");
                  //Map.Entry set;
                  //int count = 0;
                  //for (Iterator ite_1 = m_result.entrySet().iterator() ; ite_1.hasNext() ;) {
                  //  set = (Map.Entry)ite_1.next();
                  //  System.out.println("in "+count+", the key is: "+
                  //    set.getKey().toString()+", and the value is: "+
                  //    set.getValue().toString());
                  //  count++;
                  //} // for

    l_descriptorsForPanel = null;
    return m_result;
  } // readDescriptorIndexesForPanelNames()

  /**
  * For each tabbed panel name, return the indexes for the descriptors
  * which are associated to.
  * Called by getOrderedPropertyDescriptors().
  * Call getNoDisplayThisProperty() and getForbidsForArchitecture(). 
  * @param _objectInPane to display in the tabbed pane. May be null.
  * @param _l_descriptors array of descriptors from a BeanInfo
  * @return map where
  *   - the tab group name is the key,
  *   - the value is a List of all the indexes in _l_descriptors
  *     where the panel names are used. Is never empty.
  *   May be empty.
  */
  private static Map readDescriptorIndexesForTabbedNames(
          BaseObject _objectInPane, PropertyDescriptor[] _l_descriptors) {
    // initialization
    Map       m_result = new HashMap(10);
    List      l_descriptorsForPanel;

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.readDescriptorIndexesForTabbedNames(),"+
                    //" for the object:"+_objectInPane.getName()+
                    //", the descriptor array, as parameter, is: ");
                  //for (int i = 0; i < _l_descriptors.length ; i++) {
                    //System.out.println("-------- in "+i+", the property descriptor is: ");
                      //if (_l_descriptors[i] != null) System.out.println("-------------- "+_l_descriptors[i].getName());
                      //else System.out.println("-------------- null");
                  //} // for
    
    for (int i = 0; i < _l_descriptors.length; i++) {
      if (!getNotDisplayThisProperty(
              _objectInPane, _l_descriptors[i]) )  {

                                  // for testing and tracking purpose
                                //System.out.println("in BeamerManager.getOrderedNames(),"+
                                  //" for the object:"+_objectInPane.getName()+
                                  //" no display of the property: "+_l_descriptors[i].getName());          
          continue;
      }
      // searches if it is the 'ConnectedACS' to remove for the 'Architecture' view
      if (getForbidsForArchitecture(_objectInPane, _l_descriptors[i], null)) {
                        continue;                  
      } // if
      
      // sets m_result with a new list if the tabbed panel name is not registered
      if (!m_result.containsKey(_l_descriptors[i].getValue("tabbedPaneName"))) {
        l_descriptorsForPanel = new ArrayList(10);
        l_descriptorsForPanel.add(new Integer(i));
        m_result.put(_l_descriptors[i].getValue("tabbedPaneName"),l_descriptorsForPanel);
      }
      // adds the index i int the right list
      else {
        ((List)m_result.get(_l_descriptors[i].getValue("tabbedPaneName"))).add(new Integer(i));
      }

    } // for i


                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.readDescriptorIndexesForTabbedNames(),"+
                    //" for the object:"+_objectInPane.getName()+
                    //", the result is: ");
                  //Map.Entry set;
                  //int count = 0;
                  //for (Iterator ite_1 = m_result.entrySet().iterator() ; ite_1.hasNext() ;) {
                    //set = (Map.Entry)ite_1.next();
                    //System.out.println("in "+count+", the key is: "+
                      //set.getKey().toString()+", and the value is: "+
                      //set.getValue().toString());
                    //count++;
                  //} // for

    l_descriptorsForPanel = null;
    return m_result;
  } // readDescriptorIndexesForTabbedNames()

  /**
  * Return the indexes of the property descriptors in the right order,
  * where the first property to display in the first tabbed pane is at the index 0.
  * The alphabethic order of the panel group names is used to define their sequence
  * in the same tabbed pane.
  * Called by setJTabbedPane(). Use readDescriptorIndexesForPanelNames(),
  * readDescriptorIndexesForTabbedNames().
  * @param _objectInPane to display in the tabbed pane 
  * @param _l_descriptors is an array of descriptors from a BeanInfo
  * @param _l_orderedName is an array of the names for the tabbed panes 
  * @return ordered array of all the property descriptors of the beanInfo
  */
  private static PropertyDescriptor[] getOrderedPropertyDescriptors(
          BaseObject _objectInPane,
          PropertyDescriptor[] _l_descriptors,
          String[] _l_orderedName) {

    // initialization
    PropertyDescriptor[]  l_temp = new PropertyDescriptor[_l_descriptors.length];
    PropertyDescriptor[]  l_tempDescriptor = l_temp;
    Map                   m_tempIndexesForPanelNames;
    List                  list;
    List                  list_2;
    int                   index = 0;
    int                   resultIndex = 0;
    //int                   tempIndex = 0;
    String                tabbedPaneName;
    Map                   m_indexesForTabbedNames =
                              readDescriptorIndexesForTabbedNames(_objectInPane,
                                _l_descriptors);

    for (int j = 0; j < _l_orderedName.length ; j++) {

        list = (List)m_indexesForTabbedNames.get(_l_orderedName[j]) ;

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.getOrderedPropertyDescriptor(),"+
                    //" before the first loop in:"+_objectInPane.getName()+" for "+tabbedPaneName+
                    //", the list is: ");
                  //for (int i = 0; i < list.toArray().length ; i++) {
                    //System.out.println("at "+i+
                    //", the value is: "+list.toArray()[i].toString());
                  //}

        tabbedPaneName = (String) _l_descriptors[
          ((Integer)list.get(0)).intValue()].getValue("tabbedPaneName");
        
        // first loop for properties without panel
        for (ListIterator ite_2 = list.listIterator() ; ite_2.hasNext() ;) {
          index = ((Integer)ite_2.next()).intValue();
          if (_l_descriptors[index].getValue("panelGroupName") == null) {
            l_temp[resultIndex] = _l_descriptors[index];
            resultIndex++;
            ite_2.remove();
          }
        } // for ite_2
        if (list.size() == 0) continue;

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.getOrderedPropertyDescriptor(),"+
                    //" after the first loop in:"+_objectInPane.getName()+" for "+tabbedPaneName+
                    //", the list is: ");
                  //for (int i = 0; i < list.toArray().length ; i++) {
                    //System.out.println("at "+i+
                    //", the value is: "+list.toArray()[i].toString());
                  //}

        // second loop through a search on the panel names for this tabbed pane.
          // put in l_tempDescriptor all the remained descriptors of the tabbed pane
          // and have a null value when it is a property for another tabbed pane or without panel
        l_tempDescriptor = new PropertyDescriptor[_l_descriptors.length];
        for (ListIterator ite_2 = list.listIterator() ; ite_2.hasNext() ;) {
          index = ((Integer)ite_2.next()).intValue();
          l_tempDescriptor[index] = _l_descriptors[index];
        } // for ite_2

        m_tempIndexesForPanelNames = readDescriptorIndexesForPanelNames(l_tempDescriptor);
        list = new ArrayList(m_tempIndexesForPanelNames.keySet());

        // uses the alphabethic order on panel group names to locate the editors - new in 0.7.0
        Collections.sort(list);
        for (Iterator ite_2 = list.iterator() ; ite_2.hasNext() ;) {
            list_2 = (List) m_tempIndexesForPanelNames.get(ite_2.next());

            // adds all the descriptors associated to the list_2 panel group name
            for (ListIterator ite_3 = list_2.listIterator() ; ite_3.hasNext() ;) {
                index = ((Integer)ite_3.next()).intValue();
                l_temp[resultIndex] = _l_descriptors[index];

                        // for testing and tracking purpose
                      //System.out.println("in BeamerManager.getOrderedPropertyDescriptor(),"+
                        //" under the second loop in:"+_objectInPane.getName()+" for "+tabbedPaneName+
                        //", adds the editor for : "+l_temp[resultIndex].getName());

                resultIndex++;
            } // for ite_3

        } // for ite_2
    } // for j
   // removes null objects in l_temp
   PropertyDescriptor[] l_result = new PropertyDescriptor[resultIndex];
   System.arraycopy(l_temp, 0, l_result, 0, resultIndex);

                    // for testing and tracking purpose
                  //System.out.println("in BeamerManager.getOrderedPropertyDescriptor(),"+
                    //" for the object:"+_objectInPane.getName()+
                    //", the result is: ");
                  //for (int i = 0; i < l_result.length; i++) {
                    //System.out.println("at "+i+", the value is: "+l_result[i].getName());
                  //} // for

    l_temp = null;
    l_tempDescriptor = null;
    m_tempIndexesForPanelNames = null;
    m_indexesForTabbedNames = null;
    list = null;
    list_2 = null;
    tabbedPaneName = null;

    return l_result;
  } // getOrderedPropertyDescriptor()


  /**
  * Adds the property custom editor to a container, through
  * PropertyEditor.getCustomEditor().
  * Called by setJPanel() and setJTabbedPane().
  * Calls SimplePropertyEditor, StringMapPropertyEditor, RightMapPropertyEditor,
  * CollectionPropertyEditor, AssociationMapPropertyEditor, GraphicViewPropertyEditor
  * or BaseObjectMapPropertyEditor.
  * @param _objectInBeamer is the object to display in the beamer
  * @param _descriptor of a _objectInBeamer property
  * @param _component which has to handle the property editor
  * @param _largerContainer is true to have a larger returned container.
  *  Used only when the property is boolean, and adds an empty label
  *  with a size get from CommonPropertyEditor.EMPTY_FIELD_DIMENSION.
  * @exception ProcessError if a CommonPropertyEditor throws an exception
  * @return CommonPropertyEditor which displays an added JPanel for the property.
  */
  private static CommonPropertyEditor addPropertyEditor(BaseObject _objectInBeamer,
    PropertyDescriptor _descriptor, Container _component, boolean _largerContainer)
    throws ProcessError {

                      // for testing purpose
                      //if (_descriptor.getName().contains("InheritedTargetAclEntries"))
                        //      System.out.println("BeamerManager.addPropertyEditor() for "+
                          //            _descriptor.getName());

      // initialization
      CommonPropertyEditor        propertyEditor = null;

      try {
        if (_descriptor.getValue("isCollection") == Boolean.TRUE ) {
          propertyEditor = CollectionPropertyEditor.valueOf(_objectInBeamer,
            _descriptor);
        } else if (_descriptor.getValue("isStringMap") == Boolean.TRUE ) {
          propertyEditor = StringMapPropertyEditor.valueOf(_objectInBeamer,
            _descriptor);
        } else if (_descriptor.getValue("isRightMap") == Boolean.TRUE ) {
          propertyEditor = RightMapPropertyEditor.valueOf(_objectInBeamer,
            _descriptor);
        } else if (_descriptor.getValue("isView") == Boolean.TRUE ) {
            // for the opening of a 'Full views' or 'NoThan views' frame, 
            // get the opened graphic view from ViewModel
          propertyEditor = GraphicViewPropertyEditor.valueOf(_objectInBeamer,
                        _descriptor);
        } else if (_descriptor.getValue("isActionMap") == Boolean.TRUE ) {
          propertyEditor = AssociationMapPropertyEditor.valueOf(_objectInBeamer,
            _descriptor);
        } else if (_descriptor.getValue("isBaseObjectMap") == Boolean.TRUE ) {
          propertyEditor = BaseObjectMapPropertyEditor.valueOf(_objectInBeamer,
            _descriptor);
        } else if (_descriptor.getValue("isStringTwoKeysMap") == Boolean.TRUE ) {
          propertyEditor = StringTwoKeysMapPropertyEditor.valueOf(_objectInBeamer,
            _descriptor);        }
        else {
          propertyEditor = SimplePropertyEditor.valueOf(_objectInBeamer,
            _descriptor);
        } // else

        // adds each property panel in the component,
        // and the Property Editor does its job
        _component.add( (JPanel)propertyEditor.getCustomEditor() );
        
        if (_largerContainer) {
          JPanel      emptyPanel = new JPanel();
          JLabel      emptyLabel = new JLabel("           ");
          emptyLabel.setMaximumSize(CommonPropertyEditor.EMPTY_FIELD_DIMENSION);
          emptyLabel.setMinimumSize(CommonPropertyEditor.EMPTY_FIELD_DIMENSION);
          emptyLabel.setPreferredSize(CommonPropertyEditor.EMPTY_FIELD_DIMENSION);
          emptyLabel.setBackground(Color.WHITE);
          emptyPanel.add(emptyLabel);
          _component.add(emptyPanel);
          emptyLabel = null;
          emptyPanel = null;
        } // if
        
        _component.setBackground(InternalFrame.BACKCOLOR_3);

      } catch (BaseError b) {
            throw new InternalError("in BeamerManager.addPropertyEditor(), throw BaseError - "+
              "descriptor is: "+_descriptor.getName()+"; and error: "+b.toString());//+
             // " and property custom editor is: "+propertyEditor.getCustomEditor() );          
      }

                      // for testing purpose
                      //if (_descriptor.getName().contains("InheritedTargetAclEntries"))
                        //      System.out.println("BeamerManager.addPropertyEditor() end. "+
                          //            "PropertyEditor as result: "+propertyEditor);

      return propertyEditor;
  } // addPropertyEditor()

  /**
  * Detects the need to not display a property in the beamer, following the ACS policy
  * about its allowed properties and tabbed panes, and also some more general
  * requirements about metarights.
  * Called by getOrderedNames(), getOrderedPropertyDescriptors() and
  * readDescriptorIndexesForTabbedNames().
  * For right properties, call BaseUtilityImpl.getNoDisplayThisRightProperty().
  * @param _objectInPane to display in the tabbed pane. May be null.
  * @param _propertyDescriptor is a property descriptor from a BeanInfo
  * @return false if the tabbed pane or the property has to NOT be displayed
  *   for this acs. Always true if _objectInPane is null.
  */
  private static boolean getNotDisplayThisProperty(
    BaseObject _objectInPane, PropertyDescriptor _propertyDescriptor) {

      if (_objectInPane == null) return true;
      
      // initialization
      boolean result = false;
      //Class<?> propertyType = getElementaryPropertyType(_propertyDescriptor );
      String paneName = (String)_propertyDescriptor.getValue("tabbedPaneName");
      
      // _objectInPane is an ACS
      ACS acs = null;
      if (_objectInPane instanceof ACS) {
         acs = (ACS)_objectInPane;

              // for testing purpose
              //System.out.println("in BeamerManager.getNoDisplayThisProperty(),"+
                //" for the acs"+acs.getName()+
                //", the property is: "+_propertyDescriptor.getName()+
                //", and the pane name is: "+paneName);             

         if (
             // checks up the tabbed pane
                 // resource in an acs 
             (paneName.equals("  Access Targets  ") 
                      && !acs.getManageResources() 
                      && !acs.getManageExternalAcls()
                      && !acs.getManageResourceVirtualFolders() ) ||
                 // eligibleparty in an acs
             (paneName.equals("  Right Users  ") 
                      && !acs.getManageAccounts()
                      && !acs.getManageGroups() 
                      && !acs.getManageActors() 
                      && !acs.getManageExternalAcls()
                      && !acs.getManageEPVirtualFolders() )  ||
                  // roots in an acs
             ((paneName.equals("  Roots  ") || paneName.equals("  Root Actions  "))
                      && !acs.getContainsRootActions() ) ||
                  // proxies in an acs
             (paneName.equals("  Proxies  ")
                      && !acs.getManageActors() ) ||
                  // actions in an acs
             (paneName.equals("  Actions  ")
                      && acs.getM_AssociationsOnNewObject().isEmpty() ) ||
                  // structure 3 in an acs
             (paneName.equals("  Structure 3  ")
                      && !acs.getManagePrivilegeRights()
                      && !acs.getManageVirtualFolders()) ||
                  // see why in an acs
             (paneName.equals("  See Why  ") 
                      && acs.getM_CommentsOnProperties().isEmpty() ) ||

             // checks up the property
                      // resources in an acs     
             (_propertyDescriptor.getName().equals("M_OwnResources") 
                      && !acs.getManageResources())  ||
                 // eligible parties in an acs     
             (_propertyDescriptor.getName().equals("M_OwnEligibleParties") 
                      && !acs.getManageAccounts()
                      && !acs.getManageGroups()
                      && !acs.getManageActors() )  ||
                      // resources root in an acs     
             (_propertyDescriptor.getName().equals("OwnResourcesRootName") 
                      && !acs.getManageResources())  ||
                 // 'external-oreinted' in an acs     
             (_propertyDescriptor.getName().equals("ExternalOriented")
                      && !acs.getManageExternalAcls()
                      && !acs.getManageExternalForOneToOneLinkedPrivileges())  ||
                // 'controlled' in an acs (also for 'getL_AcsControllers')
             (_propertyDescriptor.getName().equals("Controlled")
                      && !acs.getManageExternalAcls()
                      && !acs.getManageExternalForOneToOneLinkedPrivileges())  ||
                // external EPs in an acs     
             (_propertyDescriptor.getName().equals("OneExternalEligibleParty")
                      && !acs.getManageExternalAcls()
                      && !acs.getManageExternalForOneToOneLinkedPrivileges())  ||
                // external resources in an acs
             (_propertyDescriptor.getName().equals("OneExternalResource")
                      && !acs.getManageExternalAcls()
                      && !acs.getManageExternalForOneToOneLinkedPrivileges())  ||
                // bridge sources in an acs
             (_propertyDescriptor.getName().equals("BridgeSources") 
                      && !acs.getManageBridges())  ||
                // bridge targets in an acs
             (_propertyDescriptor.getName().equals("BridgeTargets") 
                      && !acs.getManageBridges())  ||

                      // specialized rights in an acs
             (_propertyDescriptor.getName().equals("SpecializedRightsSet") 
                      && !acs.getHaveSpecializedRights())   ||

                 // EP-VF root
             (_propertyDescriptor.getName().equals("OwnEPFoldersRootName")
               && !acs.getManageEPVirtualFolders()  )  ||

                 // Res-VF root
             (_propertyDescriptor.getName().equals("OwnResourceFoldersRootName")
               && !acs.getManageResourceVirtualFolders()  )  ||

                 // Res-VF
             (_propertyDescriptor.getName().equals("OneResourceVirtualFolder")
               && !acs.getManageResourceVirtualFolders()  )  ||

                 // EP-VF
             (_propertyDescriptor.getName().equals("OneEPVirtualFolder")
               && !acs.getManageEPVirtualFolders()  ) ||

                 // Typed privileges
             (_propertyDescriptor.getName().equals("OneTypedPrivilege")
               && !acs.getManageSimpleTypedPrivileges()  ) ||

                 // Linked privileges
             (_propertyDescriptor.getName().equals("OneLinkedPrivilege")
               && !acs.getManageSimpleLinkedPrivileges()
               && !acs.getManageInternalForOneToOneLinkedPrivileges()
               && !acs.getManageExternalForOneToOneLinkedPrivileges()  )) {


                // for testing purpose   
                //System.out.println("in BeamerManager.getNoDisplayThisProperty(),"+
                  //" no display for "+_propertyDescriptor.getName()+" in "+
                  //acs.getName());                  

                 result = false;
         } // if
         else result = true;
      } // if ACS
      
      // _objectInPane is an ACSObject
      else if (_objectInPane instanceof ACSObject ) {
         acs = (ACS)ACS_Facade.getOpenACS().get( ((ACSObject)_objectInPane).getAcsName() );
         
         if (  
             // checks up the tabbed pane
                 // Node in a VirtualFolder
             (_objectInPane instanceof VirtualFolder
                      && paneName.equals("  As Node  ") 
                      && !acs.getManageVirtualFolderTrees() ) ||

             // checks up the property
                 // primary group in userID
             (_propertyDescriptor.getName().equals("PrimGroup")
               && !acs.getManageGroups()  )  ||
               
                 // primary group in userID
             (_propertyDescriptor.getName().equals("PrimGroup")
               && !acs.getL_ImmutableRules().contains(
                      "EP 2: primary group for account")  )  ||
                      
                 // direct member of group in groupIDMember 
             (_propertyDescriptor.getName().equals("AssociatedGroups")
               && !acs.getManageGroups()  )  ||
               
                 // direct member of group in groupIDMember for a GroupID
             (_propertyDescriptor.getName().equals("AssociatedGroups")
               && !acs.getManageGroupTrees()
               && (_objectInPane instanceof GroupID ) )  ||
               
                 // own resource in userID
             (_propertyDescriptor.getName().equals("l_UserIDOwnResources")
               && !acs.getManageResources()  )  ||
               
                 // current actor in userID 
             (_propertyDescriptor.getName().equals("l_UserIDCurrentActors")
               && !acs.getManageActors()  )  ||
               
                 // own resource in groupID
             (_propertyDescriptor.getName().equals("l_GroupIDOwnResources")
               && !acs.getManageResources()  )  ||
               
                 // current actor in groupID 
             (_propertyDescriptor.getName().equals("l_GroupIDCurrentActors")
               && !acs.getManageActors()  )  || 
                 
                 // parent node in resource
             (_propertyDescriptor.getName().equals("ParentNode")
               && !acs.getManageResourceTrees()  )  ||
               
                 // user in resource
             (_propertyDescriptor.getName().equals("UserID")
               && !acs.getManageAGORights()  )  ||
               
                 // group in resource
             (_propertyDescriptor.getName().equals("GroupID")
               && !acs.getManageAGORights()  )  ||
               
                 // inheritance attributes in resource
             (!acs.getL_ImmutableRules().contains(
                     "right access 1: child/node access only if "+
                     "the current parent tree rights allow to go across it") &&
                     (_propertyDescriptor.getName().equals("AccountAndGroupOfParent") ||
                     _propertyDescriptor.getName().equals("FromTheParentAccount") ||
                     _propertyDescriptor.getName().equals("FromTheParentGroup") )  )  ||
               
                 // child group enforcing in directory 
             (_propertyDescriptor.getName().equals("EnforceChildGID")
               && !acs.getManageGroups()  )  ||               
               
                 // child group enforcing in directory 
             (_propertyDescriptor.getName().equals("EnforceChildGID")
               && !acs.getL_ImmutableRules().contains(
                     "resource 1: new child/node inherits the current G, if activated,"+
                     " and G is not the account primary group")  )  ||  
               
                 // current user in actor 
             (_propertyDescriptor.getName().equals("CurrentUserID")
               && !acs.getManageAccounts()  )  ||
               
                 // current group in actor 
             (_propertyDescriptor.getName().equals("CurrentGroupID")
               && !acs.getManageGroups()  )  ||
               
                 // secondary groups in actor 
             (_propertyDescriptor.getName().equals("SecondaryGroups")
               && !acs.getManageSecGroupsForActor()  )  ||

                 // conditional groups in ACL
             (_propertyDescriptor.getName().equals("ConditionSourceGroups")
               && !acs.getManageFlexibleConditionalAcls()
               && !acs.getManageRightDefinedConditionalAcls() )  ||

                 // enabled rights in ACL
             (_propertyDescriptor.getName().equals("EnabledRights")
               && !acs.getManageFlexibleConditionalAcls()
               && !acs.getManageRightDefinedConditionalAcls() )  ||

                 // aliases in Basic and VirtualFolder,
              (_propertyDescriptor.getName().contains("HardAlias")
              && !acs.getManageInternalHardAlias() )  ||

              (_propertyDescriptor.getName().contains("HardAliasReference")
              && !acs.getManageInternalHardAlias())  ||

                 // soft aliases in Basic and VirtualFolder
               // but it is possible, as in MySQL addon, to divert
               // the use of gBase.CAlias for 'SoftAlias' if at least
               // one of the ACS manages the hard aliasing
              (_propertyDescriptor.getName().equals("SoftAlias")
              && !acs.getManageSoftAlias()
              && !acs.getManageInternalHardAlias())  ||

                 // ACS of soft references in Basic and VirtualFolder
              (_propertyDescriptor.getName().equals("AcsOfSoftReferences")
              && !acs.getManageSoftAlias())  ||

                 // bridges in Actor
              (_propertyDescriptor.getName().equals("ACSBridgeTarget")
              && !acs.getManageBridges() )  ||

                 // bridges in Actor
              (_propertyDescriptor.getName().equals("l_bridgeSources")
              && !acs.getManageBridges() )  ||

                 // VirtualFolders in Basic
              (_propertyDescriptor.getName().equals("VirtualFolders")
              && !acs.getManageVirtualFolders() )  ||

                 // all right properties, including ACL and AG rights
                 // and control of the panes
             !BaseUtilityImpl.getNoDisplayThisRightProperty(
                 (ACSObject)_objectInPane, _propertyDescriptor, acs)
               
               ) {              

                        // for testing purpose
                        //System.out.println("in BeamerManager.getNoDisplayThisProperty(),"+
                          //" no display for "+_propertyDescriptor.getName()+" in "+
                          //acs.getNickName());
             
                 result = false;
         } // if 
         else result = true;
      } // if      
      else if (_objectInPane instanceof StringMetaRight) {          

         if (
             // no metaright property to display
             _propertyDescriptor.getName().equals("MetaRight") ) {

                 result = false;
         } // if
         else result = true;
      } else result = true;
      
      acs = null; 
      paneName= null;
    
      return result;
  } // getNotDisplayThisProperty()  
  
  /**
  * Analyzes the BeanInfo of the displayed object to set the event handling
  * when the object is closing or deleting, and to set objectInBeamerManager__.
  * Called by updateBeamerPane().
  * @exception ProcessError if BeanInfo.getBaseBeanManagerGetMethod()
  *   returns a wrong method
  * @see ARoad0.Pattern.BeanInfoPattern
  * @param _objectInBeamer is the object to display in the beamer
  */
  private static void initializeObjectClosing(BaseObject _objectInBeamer)
    throws ProcessError {

            // for testing and tracking purpose
            // System.out.println("in BeamerManager.initializeObjectClosing(): "+
            //  "_objectInBeamer: "+
            //  _objectInBeamer.toString()
            // );

    // tries the first way to obtain the manager for closing/deleting operations
    // following ARoad0.Pattern.BeanInfoPattern
    if (objectInBeamerBeanInfo__.getBaseBeanManagerInstance() != null) {
      objectInBeamerManager__ = (BaseBeanManager)
        objectInBeamerBeanInfo__.getBaseBeanManagerInstance();
    }
    // tries the second way to obtain the manager for closing/deleting operations
    // following ARoad0.Pattern.BeanInfoPattern
    else if (objectInBeamerBeanInfo__.getBaseBeanManagerGetMethod() != null) {
      try {
        objectInBeamerManager__ = (BaseBeanManager) objectInBeamerBeanInfo__.
          getBaseBeanManagerGetMethod().invoke(_objectInBeamer,(Object[]) null);
      } catch (Exception e) {
        throw new ProcessError("in BeamerManager.initializeObjectClosing(): "+
          e.toString());
      }
    } // else if (

            // for testing and tracking purpose
            // System.out.println("in BeamerManager.initializeObjectClosing(): "+
            //  "objectInBeamerBeanInfo__: "+objectInBeamerBeanInfo__+
            //  "; objectInBeamerManager__: "+objectInBeamerManager__);

    // objectInBeamerManager__ is null for StringMetaRightImpls which 
    // are never finalized by the user.
    if (objectInBeamerManager__ != null)
        objectInBeamerManager__.
            addPropertyChangeListener(objectInBeamerBeanInfo__.
                getBaseBeanManagerBoundProperty(), INSTANCE);

  } // initializeObjectClosing()

  
  /**
  * Sets the component to put in the tab title, with a JLabel and a JButton
  * to close the tab. This closing finalizes the GraphicViewPropertyEditor
  * in the pane.
  * Called by setJTabbedPane() when it is called by ActionOpenAllFullViews and
  * ActionOpenNoThanViews.
  * The closing of a tab calls ActionClose.tabClosingPerformed().
  * Updates ViewModel.m_closingTabButtons__
  * @see ARoad0.Gui1.TabClosingHandler
  * @param _pane is the JTabbedPane where a new tab has been added
  * @param _title is the current title of the tab
  * @param _caller is the caller of the setJTabbedPane method (ActionClose indeed)
  * @return the component to put as tab component in _pane for the new tab.
  *     Contains a JLabel to show the text and a JButton to close the tab it belongs to.
  */
  static private JPanel initializeViewTabComponent(JTabbedPane _pane, String _title,
          TabClosingHandler _caller) {

    if (_pane == null) {
        throw new NullPointerException("in BeamerManager.initializeViewTabComponent(),"+
                " the tabbedPane is null");
    }      
    if (_title == null) {
        throw new NullPointerException("in BeamerManager.initializeViewTabComponent(),"+
                " the title is null");
    }  
    //unset default gaps
    final JPanel result = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
    
    result.setOpaque(false);

    JLabel label = new JLabel();
    label.setText(_title);

    result.add(label);
    //add more space between the label and the button
    label.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 5));
    
    class TabButton extends JButton implements ActionListener {
        JTabbedPane           pane_;
        String                title_;
        TabClosingHandler     caller_;
        JPanel                result_;
        
        public TabButton(JTabbedPane _pane, String _title,
          TabClosingHandler _caller, JPanel _result) {
            super();
            pane_ = _pane;
            title_ = _title;
            caller_ = _caller;
            result_ = _result;
            
            int size = 17;
            setPreferredSize(new Dimension(size, size));
            setToolTipText("close this tab");

            setUI(new BasicButtonUI());
            //Make it transparent
            setContentAreaFilled(false);
             //No need to be focusable
            setFocusable(false);
            setBorder(BorderFactory.createEtchedBorder());
            setBorderPainted(false);            
            
            // register the closing button in ViewModel, to call it for any closing
            ViewModel.addClosingTabButtonInTabbedPane(
                    ViewModel.getViewNameFromTabTitle(title_), this);
            addActionListener(this);
            
            
                    // for testing and tracking
                    //System.out.println("in BeamerManager.initializeViewTabComponent(), "+
                      //"create the TabButton for the tab: "+title);          
            
        } // TabButton(

        public void actionPerformed(ActionEvent e) {
            int i = pane_.indexOfTabComponent(result_);
            if (i == -1) return;

            // extract the name only if it is a view
            ImmutableName name = ViewModel.getViewNameFromTabTitle(title_);
            if (name == null) return;
            
            // fire an event to ActionClose if it is not implied,
            // to request to the user if he wants a ViewInBase closing.
            // If ActionClose is implied, the tab closing has been
            // done on a request from ActionClose
            if (!ViewModel.getOnGoingClosingOnView().contains(name)) {

                  // for testing and tracking
                  //System.out.println("in BeamerManager.initializeViewTabComponent(),"+
                    //" the class TabButton.actionPerformed() calls "+
                    //"ActionClose.tabClosingPerformed(title)"+
                    //" for the title: "+title_+" and the tab index: "+i);

                caller_.tabClosingPerformed(name);
                name = null;
                return;
            } // else, clear the list
            else ViewModel.removeOnGoingClosingOnView(name);

            // remove the GraphicViewPropertyEditor in ViewModel.
            // GraphicView and GraphicViewBaseListener are finalized in ActionClose.
            GraphicViewPropertyEditor editor = ViewModel.
                    getGraphicViewPropertyEditor(name.toString());
            if (editor != null) {
                editor.finalizeForProcess();
            }
            
            // unregister the closing button in ViewModel
            ViewModel.removeClosingTabButtonInTabbedPane(
                    ViewModel.getViewNameFromTabTitle(title_)); 

                  // for testing and tracking
                  //System.out.println("in BeamerManager.initializeTabComponent(),"+
                    //" the class TabButton.actionPerformed() calls "+
                    //"removeTabAt(i)"+
                    //" for the title: "+title_+" and the tab index: "+i);            
           
            // close the tab
            pane_.removeTabAt(i);
            
            title_ = null;
            name = null; 
            editor = null;
        } // actionPerformed()

        public void updateUI() {
        }

        //paint the cross
        protected void paintComponent(Graphics g) {
            super.paintComponent(g);
            Graphics2D g2 = (Graphics2D) g.create();
            //shift the image for pressed buttons
            if (getModel().isPressed()) {
                g2.translate(1, 1);
            }
            g2.setStroke(new BasicStroke(2));
            g2.setColor(Color.BLACK);
            if (getModel().isRollover()) {
                g2.drawRect(4, 4, getWidth() - 4, getHeight() - 4);
                g2.translate(1, 1);
            }
            int delta = 6;
            g2.drawLine(delta, delta, getWidth() - delta - 1, getHeight() - delta - 1);
            g2.drawLine(getWidth() - delta - 1, delta, delta, getHeight() - delta - 1);
            g2.dispose();
        } // paintComponent()
    }; // class TabButton()
    
    //closing button is defined in an internal class
    TabButton button = new TabButton(_pane, _title, _caller, result);    
    
    result.add(button);
    // Tab component title is used in ActionUtilities.analyzeActionEvent()
    result.setName(_title);
    //add more space to the top of the component
    result.setBorder(BorderFactory.createEmptyBorder(2, 0, 0, 0));    
      
            // for testing and tracking purpose
            // System.out.println("in BeamerManager.initializeTabClosing(): "+

    label = null;
    button = null;
    return result;
  } // initializeViewTabComponent()  
  
  /**
  * Clears the center panel and removes it in the beamer.
  * Finalizes the current property editors in the beamer. Unregisters
  * the base object manager if it exists. Resets the beamer title.
  * The calling method has finalized the object displayed by the beamer.
  * Called in the static initializer on the delete button, called
  * by propertyChange() and updateBeamerPane().
  */
  private synchronized static void finalizePropertyEditors() {

    // tests null values because a double click to open the beamer calls this method twice
    if (beamerPane__ != null)
    beamerPane__.remove(centerPanel__);
    if (centerPanel__ != null)
    centerPanel__.removeAll();
    centerPanel__.setOpaque(true);

    // removes the old manager
    if (objectInBeamerManager__ != null) {
      objectInBeamerManager__.
        removePropertyChangeListener(objectInBeamerBeanInfo__.
          getBaseBeanManagerBoundProperty(), INSTANCE);
    }
    // deletes the property editors
    if (m_currentPropertyEditors__ != null) {
      for (Iterator ite = m_currentPropertyEditors__.values().iterator() ; ite.hasNext() ;) {
        ((CommonPropertyEditor)ite.next()).finalizeForProcess();
      }
    }
    // resets the beamer title
    InternalFrameModel.getBeamer().setTitle(
      "Beamer on a closed element");

    objectInBeamerManager__ = null;
    objectInBeamer__ = null;
    objectInBeamerName__ = null;
    m_currentPropertyEditors__ = null;
  } // finalizePropertyEditors()

  /**
  * Finalizes the static variables of the beamer when it is closing,
  * but reinitializes the panels in the beamer and the buttons in the toolbar.
  */
  private synchronized static void finalizeForUser() {
    /*
    backButton__ = null;
    forwardButton__ = null;
    deleteButton__ = null;
    backButtonIcon__ = null;
    forwardButtonIcon__ = null;
    */

    //beamerPane__ = null;
    //centerPanel__ = null;
    beamerPane__ = new JPanel();
    centerPanel__ = CENTER_JPANEL;
    beamerPane__.setBackground(Color.white);
    beamerPane__.setLayout(new BorderLayout());
    beamerPane__.add( TOOLBAR, BorderLayout.NORTH );

    backButton__.setEnabled(false);
    forwardButton__.setEnabled(false);

    objectInBeamer__ = null;
    objectInBeamerName__ = null;
    objectInBeamerManager__ = null;
    objectInBeamerBeanInfo__ = null;
    m_currentPropertyEditors__ = null;
  } // finalizeForUser()

} // BeamerManager