//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui1;

import java.beans.*;
import javax.swing.*;
import java.lang.reflect.Method;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import java.util.ArrayList;
import java.util.List;
import javax.swing.border.LineBorder;
import javax.swing.text.Document;
import javax.swing.text.BadLocationException;

import ARoad0.Gui2.GraphicView;
import ARoad0.Gui2.GraphicNode;
import ARoad0.Gui2.ViewModel;
import ARoad0.ErrW.ProcessError;
import ARoad0.ErrB.BaseError;
import ARoad0.gWork.ViewInBase_Facade;
import ARoad0.gBase.NameImpl;
import ARoad0.gBaseInterface.ViewInBase;
import ARoad0.gBaseInterface.BaseObject;
import ARoad0.gDMak.ActionOpenArchitecture;

/**
 * This class is responsible for displaying, without editing, the
 * GraphicView and the text properties of a ViewInBase, by the mean of 
 * a simple panel without buttons.
 * The getCustomEditor() method returns the final result.
 * This class is returned by Gui1.ActionOpenAllFullViews or ActionOpenAllNoThanViews,
 * which generates the descriptors,
 * and it is so independant to the gBase package, on the contrary of other
 * CommonPropertyEditor subclasses. getSource() returns null.
 * It is used by Gui1.BeamerManager.addPropertyEditor() when BeamerManager is called
 * by ActionOpenAllFullViews or ActionOpenAllNoThanViews.
 * <P>
 * This class displays two properties derived from the view, and their types
 * are ARoad0.Gui2.GraphicView and javax.swing.JTextPane.
 * This class is a PropertyEditor that gets the property and does NOT listen the changes. 
 * It is a custom editor that provides getCustomEditor(),
 * and specifies the prefered size of the panel to be displayed.
 * <P>
 * When the user updatings make some changes in the view, there are
 * listen by a GraphicViewBaseListenerImpl instance dedicated to the view,
 * then this instance updates directly the GraphicView and the JTextPane
 * in this property editor by the way of propertyChange().
 * The GUI displays always the right data.
 * <P>
 * As a PropertyEditor, it must be able to register PropertyChangeListeners
 * but this is not a feature to use so far. Use only registration
 * on GraphicViewBaseListenerImpl, that handle PropertyChangeListeners.
 * <P>
 * An instance is constructed by the factory method valueOf(),
 * and it is possible to get a reusable empty instance.
 * java.beans.PropertyEditorManager is not used.
 * <P>
 * For thread management, this class runs a worker thread in propertyChange(). It calls
 * GraphicView.resetGraphicView() or resetRights(), then call GraphicView.setWhyText()
 * in the Event Dispatch Thread. propertyChange() processes the calls
 * in their calling order, as asynchronous tasks in the worker thread.
 * @version 0.7.0
 */
public class GraphicViewPropertyEditor implements CommonPropertyEditor,
  ActionListener, PropertyChangeListener {

  protected static final int    INITIAL_CAPACITY = 3;

  /**
   * It is not the display name of the property, given by the property
   * descriptor.
   */
  protected String displayName_;

  /**
   * The instance owning the property. Always null.
   */
  protected BaseObject source_;

  /**
   * The property descriptor
   */
  protected PropertyDescriptor propertyDescriptor_;
  
  /**
   * The property is a GraphicView there. It is NOT a BaseObject.
   */
  protected GraphicView property_;
  
  /**
   * The property type. A GraphicView there. It is NOT a BaseObject subclass.
   */
  protected Class propertyType_;

  /**
   * The subproperty is a text from the GraphicView. It is NOT a BaseObject.
   */
  protected JTextPane subproperty_;
  
  /**
   * The subproperty type. A JTextPane there. It is NOT a BaseObject subclass.
   */
  protected Class subpropertyType_ = JTextPane.class;  
  
  /**
   * True when the 'See Why' text is displayed; then, subproperty_ is not null.
   */
  protected boolean withSeeWhy_ = true;
  
  /**
   * The listeners for property updates
   */
  protected ArrayList listeners_;
  
  /**
   * An empty object for initialization
   */
  protected static final Object EMPTY_OBJECT = new Object();

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final GraphicViewPropertyEditor EMPTY_INSTANCE =
    new GraphicViewPropertyEditor();


  //-----------------------------------------------------------------------

  /**
   * Constructor FOR USE ONLY through the valueOf() method,
   * EMPTY_INSTANCE and the subclasses.
   */
  GraphicViewPropertyEditor() {
    displayName_ = "";
  }

  /**
   * Factory method to use when a CommonPropertyEditor is delegating to us.
   * The graphic view associated to the ViewInBase has to be created and registered
   * before.
   * Called by BeamerManager.addPropertyEditor().
   * @param _source the ViewInBase - no use there
   * @param _propertyDescriptor for the property, where getDisplayName() gives
   *    the ViewInBase name.toString()
   * @exception NullPointerException if _source or _propertyDescriptor is null
   * @exception BaseError if the display name or the graphic view are not found
   * @exception ProcessError if BeanInfo and PropertyDecriptor calls fail
   * @return a GraphicViewPropertyEditor instance
   */
  public static CommonPropertyEditor valueOf(BaseObject _source,
    PropertyDescriptor _propertyDescriptor)
    throws NullPointerException, BaseError, ProcessError {

      if (_source == null || _propertyDescriptor == null) {
         throw new NullPointerException();
      }
      if (!(_source instanceof ViewInBase)) {
         throw new ProcessError("in GraphicViewPropertyEditor.valueOf(), "+
                 "the source is not a view.");
      }
      // initialization
      GraphicViewPropertyEditor result;
      // set result as a new instance of
      // GraphicViewPropertyEditor
      if (_propertyDescriptor.getValue("isView") == Boolean.TRUE)
        result = new GraphicViewPropertyEditor();
      else throw new ProcessError("in GraphicViewPropertyEditor.valueOf(), "+
              "the descriptor returns a wrong value.");

      // this initialization is not common with other CommonPropertyEditors

      // because _source is not the right ViewInBase
      result.source_ = (ViewInBase) ViewInBase_Facade.getAllStandardOpenViews().get(
              NameImpl.getNameFromString(_propertyDescriptor.getDisplayName()));
      if (result.source_ == null &&
              ActionOpenArchitecture.ARCHI_VIEW_NAME.equals(
                NameImpl.getNameFromString(_propertyDescriptor.getDisplayName())))
          result.source_ = ViewInBase_Facade.getArchitectureView();
      result.propertyDescriptor_ = _propertyDescriptor;
      result.displayName_ = _propertyDescriptor.getDisplayName();
      result.propertyType_ = ARoad0.Gui2.GraphicView.class;
      result.property_ = ViewModel.getGraphicViewOfView(result.source_.getName());

      if (result.property_ == null) {
         throw new BaseError("In the GraphicViewPropertyEditor.valueOf(), "+
                 "the property is null for the descriptor:"+
                 _propertyDescriptor.getDisplayName());
      }      
      
      result.subproperty_ = new JTextPane();
      Document doc = result.subproperty_.getDocument();
      try{
            doc.insertString(0,
                             result.property_.setWhyText().getText(),
                             null);
      } catch (BadLocationException b) {
            throw new BaseError("in GraphicViewPropertyEditor.valueOf(): "+b);
      } // catch    
      result.subproperty_.setSize(
               GraphicNode.TAB_SIZE.width, 
               result.property_.setWhyText().getHeight());
      
      ViewModel.addGraphicViewPropertyEditor(
              _propertyDescriptor.getDisplayName(),
              result); 

      return result;
  } // valueOf()

  // ----------------------- PUBLIC METHODS ---------------------------------


  /**
   * NOOP - Set (or change) the object that is to be edited.
   * @param _value is the new target object to be edited. This
   *     object is not modified by the PropertyEditor. Never null.
   */
  public void setValue(Object _value) {

    return;
  } // setValue(Object value)


  /**
   * Get the value of the GraphicView.
   * @return the value of the property. May be null.
   */
  public Object getValue() {
    return property_;
  } // getValue()


  // ---------------------------------------------------------------------

  /**
   * Gets the property value as a string suitable for presentation
   * to a human to edit.
   * @return the GraphicView.setWhyText() or null.
   */
  public String getAsText() {

                  // for testing and tracking
            //System.out.println("The propertyType_ in "+
            //  "GraphicViewPropertyEditor.getAsText() is: "+
            //  propertyType_);


    return new String("no readable value");
  } // getAsText()

  /**
   * Raise java.lang.IllegalArgumentException since this kind of property can't
   * be expressed as text.
   * <P>
   * @param _text is the string to be parsed.
   * @exception in all cases.
   */
  public void setAsText(String _text) throws java.lang.IllegalArgumentException {
    throw new java.lang.IllegalArgumentException(_text);
  } // setAsText(String _text)


  // ----------------------------------------------------------------------

  /**
   * Gets the value of the property source, which is null there.
   *
   * @return the source of the property.
   */
  public BaseObject getSource() {
    return source_;
  }  // getSource()

  /**
   * Gets the value of the display property name (not the true name, given by
   * the property descriptor).
   *
   * @return the display name of the property.
   */
  public String getDisplayName() {
    return displayName_;
  }  // getDisplayName()

  /**
   * Gets the value of the property type, that is the Class of the returned
   * object in getValue(): GraphicView, and not ViewInBase.
   * @return The type of the property.
   */
  public Class getPropertyType() {
    return propertyType_;
  }  // getPropertyType()

  /**
   * Gets the descriptor of the property.
   *
   * @return The descriptor of the property.
   */
  public PropertyDescriptor getPropertyDescriptor() {
    return propertyDescriptor_;
  } // getPropertyDescriptor()

  /**
   * Gets the method on the source for updates.
   *
   * @return null.
   */
  public Method getWriteMethod() {
    return null;
  } // getWriteMethod

  /**
   * When the value is a boolean, the tages are yes/no, excepted if a couple
   * of tag is set in the property descriptor with the key 'tags'.
   * @return null.
   *
   */
  public String[] getTags() {
    return null;
  } // getTags()

  // -------------------------------------------------------------------------

  /**
   * Determines whether the propertyEditor can provide a custom editor.
   *
   * @return true.
   */
  public boolean supportsCustomEditor() {
      return true;
  }


  /**
   * This method is intended for use when generating Java code to set
   * the value of the property.  It should return a fragment of Java code
   * that can be used to initialize a variable with the current property
   * value.
   * <p>
   * Example results are "2", "new Color(127,127,34)", "Color.orange", etc.
   *
   * @return "???".
   */
  public String getJavaInitializationString() {
      return "???";
  }

  /**
   * Get the prefered size for the rectangle into which the value is displayed,
   * in the component returned by getCustomEditor().
   *
   * @return prefered size.
   */
  public Dimension getPreferedSizeForEditing() {
    return new Dimension(400, HEIGHT_DIMENSION); //400 is an estimation from CommonPropertyEditor values
  }

    //----------------------------------------------------------------------


    /**
     * Determines whether the class will honor the paintValue method.
     *
     * @return  false.
     */
    public boolean isPaintable() {
	return false;
    }

    /**
     * Paint a representation of the value into a given area of screen
     * real estate.  Note that the propertyEditor is responsible for doing
     * its own clipping so that it fits into the given rectangle.
     * <p>
     * If the PropertyEditor doesn't honor paint requests (see isPaintable)
     * this method should be a silent noop.
     *
     * @param gfx  Graphics object to paint into.
     * @param box  Rectangle within graphics object into which we should paint.
     */
    public void paintValue(java.awt.Graphics gfx, java.awt.Rectangle box) {
    }

    //----------------------------------------------------------------------

  /**
   * Register a listener for the PropertyChange event - NO USE SO FAR.
   * The class will fire a PropertyChange value whenever the value is updated.
   * @param _listener  An object to be invoked when a PropertyChange
   *		event is fired.
   */
  public synchronized void addPropertyChangeListener(
    PropertyChangeListener _listener) {
      if (listeners_ == null) {
          listeners_ = new ArrayList(INITIAL_CAPACITY);
      }
      listeners_.add(_listener);
  } // addPropertyChangeListener()

  /**
   * Remove a listener for the PropertyChange event - NO USE SO FAR.
   * @param _listener  The PropertyChange listener to be removed.
   */
  public synchronized void removePropertyChangeListener(
    PropertyChangeListener _listener) {
      if (listeners_ == null) {
          return;
      }
      listeners_.remove(_listener);
  } // removePropertyChangeListener()

  /**
   * Manages the event fired by the GraphicViewBaseListener when the GraphicView
   * has to be changed (then, _evt.getNewValue() is not null) or to be reset only
   * for updating the rights.
   * This method is called in the event dispatch thread, and uses a worker thread pool
   * which processes the calls in their calling order.
   * @param _evt the event
   */
  public synchronized void propertyChange(PropertyChangeEvent _evt) {
      
    if (_evt.getSource() != ViewModel.getGraphicViewBaseListener((ViewInBase)source_))
        throw new InternalError("in GraphicViewPopertyEditor.propertyChange(), "+
                "the event is not from the GraphicViewBaseListener");

    final PropertyChangeEvent evt = _evt;

    // waiting message for the rights changings.
    // when the nodes are changing, property_.resetGraphicView() does the work.
    if (evt.getNewValue() == null) {

                // for testing and tracking
                //System.out.println("call to "+
                  //"GraphicViewPropertyEditor.propertyChange() for the view: "+
                  //property_.getEorDataModel().getName()+"; call addWaitingMessage().");

        property_.addWaitingMessage();
    }

                // for testing and tracking
                //System.out.println("call to "+
                  //"GraphicViewPropertyEditor.propertyChange() for the view: "+
                  //property_.getEorDataModel().getName()+"; evt.getPropertyName():"+
                  //evt.getPropertyName());

    ViewModel.getViewsThreadsExecutor().execute(new Runnable() {
        @Override
        public void run() {
            // call in a worker Thread

                                    //System.out.println("call to "+
                                    //"GraphicViewPropertyEditor.propertyChange() with seewhy_size: "+
                                      //  seewhy_size);

            try {
                  // no synchronization on property_ to let the EDT handles this object.
                  // it is the same graphic view with a different set of elements.
                  // reset all the graphic view
                  if (evt.getNewValue() != null )
                      property_.resetGraphicView();
                  // else, property_ elements are not changed; its rights are updated
                  else
                      property_.resetRights();

            } catch (ProcessError p) {
                String why = null;
                Throwable cause = p.getCause();
                if (cause != null) {
                    why = cause.getMessage();
                } else {
                    why = p.getMessage();
                }
                System.err.println("In GraphicViewPropertyEditor.propertyChange(), "+
                        "error searching access paths in "+property_.getEorDataModel().getName()+
                        ", and the cause is: " + why);
                p.printStackTrace();
            } catch (InterruptedException ignore) {}

            // no InterruptedException handling there
            SwingUtilities.invokeLater(new Runnable(){
                public void run(){
                        // in the Event Dispatch Thread

                                // for testing and tracking
                                //System.out.println("call to GraphicView.removeWaitingMessage() in "+
                                  //"GraphicViewPropertyEditor.propertyChange(), for the view: "+
                                  //property_.getEorDataModel().getName());


                    property_.removeWaitingMessage();
                    property_.repaint();

                    // update the 'See why' text without style
                    Document doc = subproperty_.getDocument();
                    try{
                        doc.remove(0, doc.getLength());
                        doc.insertString(0,
                                         property_.setWhyText().getText(),
                                         null);
                    } catch (BadLocationException b) {
                        throw new InternalError(
                                "in GraphicViewPropertyEditor.propertyChange(): "+b);
                    }
                    subproperty_.setSize(
                            GraphicNode.TAB_SIZE.width,
                            property_.setWhyText().getHeight());
                    // repaint() is a thread-safe method, but not doc.insertString(),
                    // so it is necessary to keep this in run()
                    subproperty_.repaint();
                    doc = null;

                } // run
            }); // invokeLater()

                    // for testing and tracking
                    //System.out.println("in GraphicViewPropertyEditor.propertyChange(): "+
                      //      "end of run() for "+
                        //    property_.getEorDataModel().getName());

        } // run()

    }); // new Thread

  } // propertyChange()

    //----------------------------------------------------------------------

  /**
   * Get the values of variables for the instance.
   * @return source, propertyName and listeners (if they have toString() methods).
   */
  public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("  GraphicViewPropertyEditor.TOSTRING():");

        sb.append(" source_ = ");
        sb.append(source_);

        sb.append("; display name = ");
        sb.append(displayName_);

        sb.append("; listeners list = ");
        sb.append(listeners_);

        return " [" + sb + "] ";
  } // toString()

  /**
   * Finalizes the instance; the call is reserved to BeamerManager.
   */
  public void finalizeForProcess() {

                // for testing and tracking
                //System.out.println("call to "+
                  //"GraphicViewPropertyEditor.finalizeForProcess() for the view "+
                  //property_.getEorDataModel().getName());

    // listeners_ are not used so far
    if (listeners_ != null) {
      synchronized (listeners_) {
          // Tell our listeners that "everything" has changed.
          PropertyChangeEvent evt = new PropertyChangeEvent(
            this, propertyDescriptor_.getName(), null, null);

          for (int i = 0; i < listeners_.size(); i++) {
              PropertyChangeListener target =
                (PropertyChangeListener)listeners_.get(i);
              target.propertyChange(evt);

        // for testing and tracking
        //System.out.println("a listener to inform in "+
        //  "GraphicViewPropertyEditor.finalizeForProcess() for "+
        //  displayName_+" is "+target );

          } // for
        } // synchronized (listeners_)
    } // if (listeners_ != null)


    ViewModel.removeGraphicViewPropertyEditor(displayName_);
    
    displayName_ = null;
    source_ = null;
    propertyDescriptor_ = null;
    listeners_ = null;
    property_ = null;
    subproperty_ = null;
    return;
  } // finalizeForProcess()


  /**
   * NOOP
   * @param _evt is the event fired by a JComponent displayed by getCustomEditor()
   */
  public void actionPerformed(ActionEvent _evt) {

    // no action if the explorer is currently used for an object selection
    // or if there is an user dialog currently
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
      

    // for 
    else if (_evt.getSource().getClass().equals(JButton.class)) {

      // for
      if ( ((JButton)_evt.getSource()).getText().equals(" See ")) {



      } // else if (((JButton) == "See"

    } // if (_evt.getSource().getClass().equals(JButton.class))

  } // actionPerformed()  
  
  /**
   * Returns a full custom Component that displays the GraphicView and the
   * associated 'See why' text.
   * Called by BeamerManager.addPropertyEditor().
   * @return a JPanel which displays the GraphicView and the 'See why' text
   */
  public java.awt.Component getCustomEditor() {

    JPanel result = new JPanel();
    result.setLayout(new BoxLayout(result, BoxLayout.Y_AXIS));
    /*
    JLabel nameLabel = new JLabel(displayName_); //this is the ViewInBase name
    nameLabel.setMaximumSize(NAME_DIMENSION);
    nameLabel.setMinimumSize(NAME_DIMENSION);
    nameLabel.setPreferredSize(NAME_DIMENSION);
    nameLabel.setFont(BeamerManager.plainFont);
    nameLabel.setToolTipText(propertyDescriptor_.getShortDescription());
    result.add(nameLabel);
     * */
    
    JScrollPane               scrollPane;              
    JPanel                    panel; 
    //Box                       groupBox = Box.createVerticalBox();
    //result.add(groupBox);
    Dimension                 dimension = new Dimension(
                                (int)GraphicNode.TAB_SIZE.getWidth(),
                                710);
    
    // add the GraphicView in a scrollpane
    scrollPane = new JScrollPane(property_);
    //scrollPane.setPreferredSize(property_.getPreferredSize()); // new line
    //scrollPane.getViewport().add(property_);   
    
    // put the scrollPane in result
    panel = new JPanel(true);// doublebuffered panel
    panel.setLayout(new BorderLayout()); // does not use getPreferredSize()
    //panel.setSize(
      //new Dimension(GraphicNode.VIEW_INTAB_SIZE));
    panel.setMinimumSize(
      new Dimension(GraphicNode.VIEW_INTAB_SIZE));
    panel.setMaximumSize(
      new Dimension(GraphicNode.VIEW_INTAB_SIZE));     

    panel.setBorder(new LineBorder(InternalFrame.FORECOLOR));    
    panel.add(scrollPane);//panel.add("Center", scrollPane);
    result.add(panel);    
    
    // add the see why text
    // use a variable scrollpane, with test of the subproperty_ size
    scrollPane = new JScrollPane(subproperty_);

    // put the scrollPane in result
    panel = new JPanel();
    panel.setLayout(new BorderLayout());     
    //panel.setSize(
      //new Dimension(GraphicNode.VIEW_INTAB_SIZE));
    panel.setMinimumSize(
      new Dimension(GraphicNode.VIEW_INTAB_SIZE));
    panel.setMaximumSize(
      new Dimension(GraphicNode.VIEW_INTAB_SIZE));
    panel.add(scrollPane);
    scrollPane.getVerticalScrollBar().setValue(
      scrollPane.getVerticalScrollBar().getMinimum() );  // does not work    
    result.add(panel);    
    
    result.setBackground(Color.white);
    result.setPreferredSize(dimension);
    // the JPanel has the ViewInBase name
    result.setName(property_.getEorDataModel().getName().toString()); 
    
    dimension = null;
    scrollPane = null;              
    panel = null; 
    //groupBox = null;    
    return result;
  } // getCustomEditor()

  
  // -------------------- PRIVATE METHODS -----------------------------------
  
  /**
   * Reports that we have been modified to any interested listeners - NOT USED SO FAR.
   */
  private void firePropertyChange() {
    List targets;
    synchronized (this) {
        if (listeners_ == null) {
          return;
        }
        targets = (List) listeners_.clone();
    }
    // Tell our listeners that "everything" has changed. Caution: source_ is null.
    PropertyChangeEvent evt = new PropertyChangeEvent(source_, null, null, null);

    for (int i = 0; i < targets.size(); i++) {
        PropertyChangeListener target =
          (PropertyChangeListener)targets.get(i);
        target.propertyChange(evt);
    }
  } // firePropertyChange()


} // GraphicViewPropertyEditor
