//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui1;

import javax.swing.*;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

import ARoad0.Pattern.*;
import ARoad0.gDMak.ActionOpenArchitecture;
import ARoad0.Gui2.GraphicSketchView;
import ARoad0.ErrB.BaseError;

import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.BaseObject;
import ARoad0.gBase.NameImpl;
import ARoad0.gBaseInterface.Name;

/**
 * This class is responsible to manage the internal frames data and the trees data,
 * following the MVC pattern. This includes Explorer, Beamer, Sketcher, IS structure,
 * Architecture, ACS main frames and varied view frames.
 * It has maps of the base objects and the GraphicViews
 * displayed in main frames.
 * @version 0.7.0
 */
public class InternalFrameModel implements Singleton, MVC_Model {

  /**
   * the only instance of this class, following the singleton pattern
   */
  private static InternalFrameModel      instance__ = new InternalFrameModel();

  // variables for the frames
  private static final int          INITIAL_CAPACITY = 10;//the default value:101
    /**
     * current base objects displayed in dedicated main frames ; getName() gets the key
     *   do not include the object displayed in the beamer.
     */
  private static Map                m_BaseObjectsInMainFrames__  =
                                    new HashMap(INITIAL_CAPACITY);
    /**
     * current main frames associated to base objects ; the object name is the key
     *  does not include the beamer nor the 'EPR views' frame.
     */
  private static Map                m_MainFramesOfBaseObjects__  =
                                    new HashMap(INITIAL_CAPACITY);
    /**
     * current JComponents (JTree, JPanel) displayed in dedicated main frames
     * or JTabbedPanes, and associated to base objects.
     * getName() gets the key from the base object.
     *   do not include the object displayed in the beamer.
     *   contain JPanels in 'EPR views' JTabbedPane tabs.
     */
  private static Map                m_JComponentsOfBaseObjects__  =
                                    new HashMap(INITIAL_CAPACITY);
    /**
     * dynamic list of names for the base objects for which ActionClose 
     * is closing the main frame.
     * The closing has not to produce the call of the
     * relevant ActionClose class. Used by an inner class in InternalFrameUtilities
     * and by gDMak.ActionClose.
     */
  private static List               l_onGoingClosingOnBaseObjects__ =
                                    new ArrayList(INITIAL_CAPACITY);
  public static final JInternalFrame      EMPTY_FRAME = new JInternalFrame();
  private static JInternalFrame     activeFrame__ = EMPTY_FRAME;

  // variables for the explorer
  private static JInternalFrame     explorer__ = new JInternalFrame();
  private static JTree              explorerTree__ = new JTree();
  private static boolean            withExplorer__;

  // variables for the beamer
  private static JInternalFrame     beamer__ = new JInternalFrame();
  private static JComponent         beamerPane__;
  private static boolean            withBeamer__;
  
  // variables for the view frames
  private static JInternalFrame     FullViewsframe__;
  private static JInternalFrame     NoThanViewsframe__;

  // variables for the global structure
  private static JInternalFrame     structureframe__;
  private static JTree              structureTree__ = new JTree();
  
  // variables for the sketcher
  private static JInternalFrame     sketcher__ = new JInternalFrame();
  private static GraphicSketchView  sketcherPane__;
  private static boolean            withSketcher__;
  
  /**
  * only one private constructor
  */
  public InternalFrameModel() {
  }

  /**
  * Only method to get an instance.
  * @return InternalFrameModel
  */
  public static synchronized InternalFrameModel getInstance(){
    return instance__;
  } // getInstance()


  /**
  * True if the explorer is displayed in the main window.
  * It is always true after the first ACS opening.
  * @return boolean
  */
  public static boolean withExplorer(){
    return withExplorer__;
  }

  /**
  * Gets the explorer.
  * @return JInternalFrame explorer displayed in the desktop
  */
  public static JInternalFrame getExplorer(){
    return explorer__;
  }

  /**
  * Get the tree in the explorer.
  * @return JTree tree of open objects displayed in the desktop explorer. Never null.
  */
  public static JTree getTreeInExplorer(){
    return explorerTree__;
  }
  
  /**
  * True if the beamer is displayed in the main window.
  * @return boolean 
  */
  public static boolean withBeamer(){
    return withBeamer__;
  }
  
  /**
  * Gets the beamer.
  * @return the beamer displayed in the desktop. May be empty.
  */
  public static JInternalFrame getBeamer() {
    return beamer__;
  }

  /**
  * Get the pane in the beamer.
  * @return the pane displayed in the beamer. May be null,
  *   a JPanel or a JTabbedPane.
  */
  public static JComponent getBeamerPane() {
    return beamerPane__;
  }

  /**
  * True if the sketcher is displayed in the main window.
  * @return boolean
  */
  public static boolean withSketcher(){
    return withSketcher__;
  }
  
  /**
  * Gets the sketcher.
  * @return the sketcher displayed in the desktop. May be empty.
  */
  public static JInternalFrame getSketcher() {
    return sketcher__;
  }

  /**
  * Gets the sketch view in the sketcher.
  * @return the pane displayed in the sketcher. May be null.
  */
  public static GraphicSketchView getSketcherPane() {
    return sketcherPane__;
  }
  
  /**
  * Get the 'All full views' frame, if it exists.
  * @return the 'All full views' frame. May be null.
  */
  public static JInternalFrame getAllFullViewsFrame(){
    return FullViewsframe__;
  }
 
  /**
  * Sets the 'All full views' frame.
  * @param _frame the 'All full views' frame. May be null.
  */
  public static void setAllFullViewsFrame(JInternalFrame _frame){
    FullViewsframe__= _frame;
  }
  
  /**
  * True if the 'All full views' frame is displayed in the main window.
  * @return boolean
  */
  public static boolean withAllFullViews(){
    if (FullViewsframe__ != null) return true;
    else return false;
  }
  
  /**
  * Gets the 'All NoThan views' frame, if it exists.
  * @return the 'All NoThan views' frame. May be null.
  */
  public static JInternalFrame getAllNoThanViewsFrame(){
    return NoThanViewsframe__;
  }
 
  /**
  * Sets the 'All NoThan views' frame.
  * @param _frame the 'All NoThan views' frame. May be null.
  */
  public static void setAllNoThanViewsFrame(JInternalFrame _frame){
    NoThanViewsframe__= _frame;
  }
  
  /**
  * True if the 'All NoThan views' frame is displayed in the main window.
  * @return boolean
  */
  public static boolean withAllNoThanViews(){
    if (NoThanViewsframe__ != null) return true;
    else return false;
  }
  
  /**
  * Gets the 'IS structure' frame, if it exists.
  * @return the frame. May be null.
  */
  public static JInternalFrame getStructureFrame(){
    return structureframe__;
  }

  /**
  * Gets the tree in the IS structure.
  * @return JTree of open or closed objects displayed in the IS structure. Never null.
  */
  public static JTree getTreeInStructure(){
    return structureTree__;
  }

  /**
  * Sets the 'IS structure' frame.
  * @param _frame the frame. May be null.
  * @param _tree the global structure tree. Never null.
  */
  public static void setStructureFrame(JInternalFrame _frame, JTree _tree){
    structureframe__= _frame;
    structureTree__ = _tree;
  }
  
  /**
  * @return true if the 'IS structure' frame is displayed in the main window
  */
  public static boolean withStructure(){
    if (structureframe__ != null) return true;
    else return false;
  }
  
  /**
  * True if the 'Architecture' frame is displayed in the main window. May be
  * false with an 'Architecture' node in the explorer, which is tested through
  * ViewInBase_Facade.getArchitectureView() null or not.
  * @return boolean 
  */
  public static boolean withArchitecture(){
    if (m_BaseObjectsInMainFrames__.get(ActionOpenArchitecture.ARCHI_VIEW_NAME) != null) 
        return true;
    else return false;
  }  
  
  /**
  * Gets the objects in the base that are associated to main internal frame(s).
  * The object name is the map key. The object viewed in the beamer is not
  * in this map.
  * @return unmodifiableMap of base objects displayed in the main window
  */
  public static Map getBaseObjectsInMainFrames(){
    return Collections.unmodifiableMap(m_BaseObjectsInMainFrames__);
  }

  /**
  * Gets the main frames associated to the base objects.
  * The object name is the map key. The beamer, the explorer, the Sketcher are never in this map.
  * The Architecture view is included.
  * @return unmodifiableMap of frames displayed in the main window
  */
  public static Map getMainFramesOfBaseObjects(){

                          // for testing and tracking
                          //System.out.println("in InternalFrameModel.getMainFramesOfBaseObjects(), "+
                            //" m_MainFramesOfBaseObjects__: ");
                          //for (Iterator ite = m_MainFramesOfBaseObjects__.entrySet().iterator();
                            //ite.hasNext(); ) {
                              //System.out.println("---- "+ite.next());
                          //} // for ite


    return Collections.unmodifiableMap(m_MainFramesOfBaseObjects__);
  }

  /**
  * Gets the JComponent (JTree, JPanel or GraphicView) that is associated to the base object
  * in its main frame or tab in 'All view' tabbed pane. No JComponent if the base object
  * is a view opened in the GUI whithout main frame ou tab. <P>
  * The object name is the map key. The object shown in the beamer is not
  * in this map.
  * @return unmodifiableMap of JComponents displayed in the main frames
  */
  public static Map getJComponentsOfBaseObjects(){
    return Collections.unmodifiableMap(m_JComponentsOfBaseObjects__);
  }

  /**
   * Gets the list of names for the base objects that ActionClose is closing.
   * @return unmodifiableList
   */
  static List getOnGoingClosingFromActionClass() {
    return Collections.unmodifiableList(l_onGoingClosingOnBaseObjects__);
  } // getOnGoingClosingOnBaseObjects()

  /**
  * Remove an object in the base. It is associated to one internal frame
  * or more.
  * @param ImmutableName of the object
  * @return void
  */
  //public static synchronized void removeBaseObjectInFrames(ImmutableName _name){
  //  m_BaseObjectsInMainFrames__.remove(_name);
  //  return;
  //} // removeBaseObjectInFrames()

  /**
  * Gets the 'See why' frame of a view, if it is opened.
  * Used by ActionSave.
  * @param _viewName is a view name for a view placed in a main frame or not,
  * with an open 'See why' frame or not. Never null.
  * @return the 'See why' frame or null
  */
  public static JInternalFrame getOpenedSeeWhyFrame(ImmutableName _viewName) {
    // variables
    JInternalFrame          result = null;
    JInternalFrame[]        allframes = ARoad0.Gui1.Desktop.getCurrentFrames();
    Name                    name = null;
    
    for (int i = 0; i < allframes.length; i++) {
        if (allframes[i].getTitle().indexOf(" 'See why' text") != -1) {

                              // for testing and tracking
                        //System.out.println("call to InternalFrameModel.getSeeWhyFrame(), for the frame "+
                          //      allframes[i].getTitle());

            try {
                  name = NameImpl.getNameFromString(allframes[i].getTitle().
                          substring(0, allframes[i].getTitle().indexOf(" 'See why' text")) );
              } catch (BaseError b) {
                  throw new InternalError("in InternalFrameModel.getSeeWhyFrame(), "+b);
              } // catch

              if (_viewName.equals(name) ) {
                  result = allframes[i];
                  break;              
              } // if
        } // if  (allframes
    } // for  

    allframes = null;
    name = null;
    return result;
  } // getOpenedSeeWhyFrame()

  /**
   * Get the frame which is active in the desktop.
   * Used by ActionXXX.
   * @return JInternalFrame which is currently active
   */
  public static JInternalFrame getActiveFrame() {
      synchronized (activeFrame__) {return activeFrame__;}
  } // getActiveFrame()

  /**
  * Set the beamer.
  * @param _frame is the beamer. Has to be equals to new JInternalFrame
  *   if _withBeamer is false. Never null.
  * @param _pane is the JComponent in the beamer. Has to be equals to
  *   a new JComponent if _withBeamer is false. May be null.
  * @param _withBeamer has to be false if the beamer is closed.
  */
  static void setBeamer(JInternalFrame _frame, JComponent _pane, boolean _withBeamer){
    withBeamer__ = _withBeamer;
    beamer__ = _frame;
    beamerPane__ = _pane;
  }

  /**
  * Set the explorer.
  * @param _frame is the explorer. Has to be equals to new JInternalFrame
  *   if _withExplorer is false.
  * @param _tree is the JTree in the explorer. Never null. Has to be equals to new JTree
  *   if _withExplorer is false.
  * @param _withExplorer has to be false if the explorer is closed.
  */
  static void setExplorer(JInternalFrame _frame, JTree _tree, boolean _withExplorer){
    withExplorer__ = _withExplorer;
    explorer__ = _frame;
    explorerTree__ = _tree;
  }

  /**
  * Set the sketcher.
  * @param _frame is the sketcher. Has to be equals to new JInternalFrame
  *   if _withSketcher is false. Never null.
  * @param _pane is the sketch view in the sketcher. This method sets the internal value to
  *   null if _withSketcher is false.
  * @param _withSketcher has to be false if the sketcher is closed.
  */
  static void setSketcher(JInternalFrame _frame, GraphicSketchView _pane, boolean _withSketcher){
    withSketcher__ = _withSketcher;
    sketcher__ = _frame;
    if (_withSketcher) sketcherPane__ = _pane;
    else sketcherPane__ = null;
  }
  
  /**
   * Update the dynamic list of names for the base objects
   * which ActionCloseXXX is closing.
   * Used by gDMak.ActionCloseUtilities.
   * @param _name ImmutableName of the base object which is closing
   */
  public static void addOnGoingClosingFromActionClass(ImmutableName _name) {
      synchronized (l_onGoingClosingOnBaseObjects__)
        {l_onGoingClosingOnBaseObjects__.add(_name);}
  } // addOnGoingClosingOnBaseObjects()

  /**
   * Updates m_BaseObjectsInMainFrames__, m_MainFramesOfBaseObjects__
   * and m_JComponentsOfBaseObjects__ with a lock on them.
   * Used by InternalFrame.
   * @param _object to declare
   * @param _frame main frame associated to _object
   * @param _component is the JTree, JPanel or the GraphicView associated to the ACS
   *  or to the view
   */
  static void addMainFrameBaseObject(BaseObject _object, JInternalFrame _frame,
    JComponent _component) {
      synchronized (m_BaseObjectsInMainFrames__) {
        synchronized (m_MainFramesOfBaseObjects__) {
            m_BaseObjectsInMainFrames__.put(_object.getName(), _object);
            m_MainFramesOfBaseObjects__.put(_object.getName(), _frame);

                          // for testing and tracking
                          //System.out.println("in InternalFrameModel.addMainFrameBaseObject(), "+
                            //" m_MainFramesOfBaseObjects__ after the updating: ");
                          //for (Iterator ite = m_MainFramesOfBaseObjects__.keySet().iterator();
                            //ite.hasNext(); ) {
                            //System.out.println("---- "+ite.next());
                          //} // for ite

        }
      }
      addBaseObjectJComponent(_object, _component);
  } // addMainFrameBaseObject()

  /**
   * Updates m_BaseObjectsInMainFrames__, m_MainFramesOfBaseObjects__
   * and m_JComponentsOfBaseObjects__ with a lock on them.
   * Used by InternalFrameUtilities.
   * @param _name of the object to remove
   */
  static void removeMainFrameBaseObject(ImmutableName _name) {
      synchronized (m_BaseObjectsInMainFrames__) {
        synchronized (m_MainFramesOfBaseObjects__) {
            m_BaseObjectsInMainFrames__.remove(_name);
            m_MainFramesOfBaseObjects__.remove(_name);

                          // for testing and tracking
                          //System.out.println("in InternalFrameModel.removeMainFrameBaseObject(), "+
                            //" m_MainFramesOfBaseObjects__ after the updating: ");
                          //for (Iterator ite = m_MainFramesOfBaseObjects__.keySet().iterator();
                            //ite.hasNext(); ) {
                            //System.out.println("---- "+ite.next());
                          //} // for ite

        }
      }
      removeBaseObjectJComponent(_name);
  } // removeMainFrameBaseObject()


  /**
   * Updates m_JComponentsOfBaseObjects__.
   * Used by addMainFrameBaseObject().
   * @param _object to declare
   * @param _component is the JTree, JPanel or GraphicView associated to _object
   */
  public static void addBaseObjectJComponent(BaseObject _object,
    JComponent _component) {
      synchronized (m_JComponentsOfBaseObjects__) {
            m_JComponentsOfBaseObjects__.put(_object.getName(), _component);

                          // for testing and tracking
                          //System.out.println("in InternalFrameModel.addBaseObjectJComponent(), "+
                            //" m_JComponentsOfBaseObjects__ after the updating: ");
                          //for (Iterator ite = m_JComponentsOfBaseObjects__.keySet().iterator();
                            //ite.hasNext(); ) {
                            //System.out.println("---- "+ite.next());
                          //} // for ite
      }
  } // addeBaseObjectJComponent()

  /**
   * Updates m_JComponentsOfBaseObjects__.
   * Called by ActionClose.closeOneEPRviewsTab().
   * @param _name of the BaseObject to remove
   */
  public static void removeBaseObjectJComponent(ImmutableName _name) {
      synchronized (m_JComponentsOfBaseObjects__) {
            m_JComponentsOfBaseObjects__.remove(_name);


                          // for testing and tracking
                          //System.out.println("in InternalFrameModel.removeBaseObjectJComponent(), "+
                            //" m_JComponentsOfBaseObjectss__ after the updating: ");
                          //for (Iterator ite = m_JComponentsOfBaseObjects__.keySet().iterator();
                            //ite.hasNext(); ) {
                            //System.out.println("---- "+ite.next());
                          //} // for ite
      }
  } // removeBaseObjectJComponent()  
  
  /**
   * Removes a name for the base objects which are closing.
   * @param _name to remove
   */
  static void removeOnGoingClosingFromActionClass(ImmutableName _name) {
    l_onGoingClosingOnBaseObjects__.remove(_name);
  } // removeOnGoingClosingOnBaseObjects()

  /**
   * Sets the frame which is active in the desktop.
   * @param _frame which is currently active. May be EMPTY_FRAME.
   */
  static void setActiveFrame(JInternalFrame _frame) {
    synchronized (activeFrame__) {
      if (_frame != null) activeFrame__ = _frame;
      else activeFrame__ = InternalFrameModel.EMPTY_FRAME;
    }

                          // for testing and tracking
                          //System.out.println("InternalFrameModel.setActiveFrame(), "+
                            //" activeFrame__: "+activeFrame__.getTitle());
  } // setActiveFrame()
  
} // InternalFrameModel