//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui1;

import java.beans.*;
import javax.swing.*;
import javax.swing.border.BevelBorder;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.util.Vector;
import java.util.Map;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import ARoad0.Gui2.TreeManager;
import ARoad0.ErrW.ProcessError;
import ARoad0.ErrW.SmallError;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.gWork.ACS_Facade;

import ARoad0.Pattern.BeanInfoPattern;
import ARoad0.Pattern.MVC_Controller;
import ARoad0.Pattern.MVC_View;

import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.ImmutableACS;
import ARoad0.gBaseInterface.BoundBean;
import ARoad0.gBaseInterface.BaseObject;
import ARoad0.gBaseInterface.ACSObject;
import ARoad0.gBaseInterface.ACS;
import ARoad0.gBaseInterface.ImmutableNode;
import ARoad0.gBaseInterface.ImmutableResource;
import ARoad0.gBaseInterface.ImmutableEligibleParty;
import ARoad0.gBaseInterface.StringRight;

import ARoad0.gBase.BaseUtilityImpl;
import ARoad0.gBase.UtilityImpl;
import ARoad0.gBase.AclEntryImpl;
import ARoad0.gBase.PrivilegeAbst;
import ARoad0.gBase.NameImpl;

/**
 * This class is responsible for displaying and editing the more simple
 * properties in the gBase classes, by the mean of panels.
 * The getCustomEditor() method is used then.
 * This class is selected by the numerous XXXBeanInfo classes in gBase,
 * in their property descriptors.
 * It is used by Gui1.BeamerManager.addPropertyEditor().
 * <P>
 * The possible property types are boolean, int, String,
 * ARoad0.gBase.ImmutableName and any ARoad0.gBase.BaseObject.
 * Object is not an allowed type.
 * This class is a PropertyEditor that gets, updates and listens the property
 * in the relevant gBase instance. It has several subclasses.
 * It is also a custom editor that provides getCustomEditor(),
 * and specifies the prefered size of the panel to be displayed.
 * <P>
 * This instance listens actions from the JComponent property when the user enters data,
 * and from gBase objects when the program updates the property.
 * When the user updates make changes in the data base, this instance
 * updates the base, then these changes
 * are fired to this instance to update the JComponent in the GUI.
 * So, the GUI always DISPLAYS THE RIGHT DATA which is in the base.
 * After an user update, the sequence of calls is:
 * - actionPerformed() from the JComponent,
 * - setAsText() or setValue(), which updates the data base object,
 * - propertyChange() from the object, with firing to the JComponent and before,
 *    unregisters this PropertyEditor as a JComponent listener, to avoid a loop.
 * The updating handles all the exceptions in the same way.
 * A message is displayed to the user from the exception, and the editor data
 * is reset through the call to the current value. So, the called gBase class for
 * the updating method may choose how the exception impacts the current value.
 * Caution: the property sourceBeanInfo_ is initialized only when the property
 * values are constrained.
 * <P>
 * As a PropertyEditor, it must be able to register PropertyChangeListeners
 * but this is not a feature to use so far. Use only registration
 * on gBase BoundBeans, that handle PropertyChangeListeners. In the same way,
 * do not use the ActionEvents handling - add/removeActionListener() -
 * in the JComponent property.
 * <P>
 * An instance is constructed through the factory method 'valueOf()'.
 * This static constructor has to be called by all the subclasses, which have to
 * be build up by an overriding version of this method.
 * It is possible to get a reusable empty instance.
 * java.beans.PropertyEditorManager is not used.
 * When a String is entered in a JTextField, its length is limited to 200,
 * to prevent any unauthorized input.
 * A space is always added to the end of the string.
 * For a boolean property, the editor may be a combobox with a specified text,
 * or a checkbox to select to select yes/no simply. The right editor is get
 * from the property descriptor in the BeanInfo class.
 * @version 0.7.0
 * @since 0.5.1
 */
public class SimplePropertyEditor implements MVC_Controller, MVC_View,
  CommonPropertyEditor, ActionListener, PropertyChangeListener,
  SelectionDialogListener, NewBaseObjectListener {

  protected static final int    INITIAL_CAPACITY = 3;

  /**
   * The type of the property. If the property is an array in CollectionPropertyEditor,
   * it is the type of the array elements.
   */
  protected Class propertyType_;

  /**
   * It is not the display name of the property, given by the property descriptor.
   */
  protected String displayName_;

  /**
   * The instance owning the property
   */
  protected BaseObject source_;

  /**
   * The property descriptor
   */
  protected PropertyDescriptor propertyDescriptor_;

  /**
   * The component that displays a bound property in the beamer.
   * That is the very component which displays it, not the panel arround.
   * It is a JComboBox if the property is a Boolean,
   * a JTextField if the property is a String, an ImmutableName or a BaseObject,
   * or a JList in a CollectionPropertyEditor or RightMapPropertEditor for the keys list.
   * No use in StringMapPropertyEditor.
   */
  protected JComponent boundComponent_;

  /**
   * The read method if it exists, for the property
   */
  protected Method readMethod_;

  /**
   * The write method if it exists, for the property
   */
  protected Method writeMethod_;

  /**
   * The listeners for property updates
   */
  protected Vector listeners_;
  
  /**
   * Is true if it is a constrained property that is not
   * in the constructor.
   * The restricted values are get in the property descriptor 
   * from 'staticSetOfItems', for Strings only, or from 'choiceMethodForNew',
   * or through a search in the ACS by this PropertyEditor.
   */
  protected boolean isRestrictedArgument_ = false;

  /**
   * The BeanInfo instance associated to the source_.
   * caution: it is initialized only if isRestrictedArgument_ is true, to select
   * the restricted keys or values.
   */
  protected BeanInfoPattern sourceBeanInfo_;
  
  /**
   * An empty object for initialization
   */
  protected static final Object EMPTY_OBJECT = new Object();


  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final SimplePropertyEditor EMPTY_INSTANCE =
    new SimplePropertyEditor();

  //-----------------------------------------------------------------------

  /**
   * Constructor FOR USE ONLY through the valueOf() method,
   * EMPTY_INSTANCE and the subclasses.
   */
  SimplePropertyEditor() {
    displayName_ = "";
  }

  /**
   * Factory method to use when a CommonPropertyEditor is delegating to us.
   * It selects the class to use for building up the object.
   * It constructs the read and write methods for the property,
   * if they exist, and all the common properties.
   * Called by BeamerManager.addPropertyEditor().
   * @param _source owning the property, to update through setValue()
   *  and to use for any events we fire.
   * @param _propertyDescriptor is the descriptor of the property
   * @exception NullPointerException if _source or _displayName is null
   * @exception ProcessError if BeanInfo and PropertyDecriptor calls fail
   * @return a SimplePropertyEditor or a subclass instance
   */
  public static CommonPropertyEditor valueOf(BaseObject _source,
    PropertyDescriptor _propertyDescriptor)
    throws NullPointerException, ProcessError {

                  // for testing purpose
                  //if ( _propertyDescriptor.getName().equals("SoftAlias") )
                    //  System.out.println("starting of SimplePropertyEditor.valueOf() for 'SoftAlias'");


      if (_source == null || _propertyDescriptor == null) {
         throw new NullPointerException();
      }

      // initialization
      SimplePropertyEditor result;
      String beanName = null;
      // set result as a new instance of
      // SimplePropertyEditor or one subclass
      if (_propertyDescriptor.getValue("isCollection") == Boolean.TRUE)
        result = new CollectionPropertyEditor();
      else if (_propertyDescriptor.getValue("isStringMap") == Boolean.TRUE)
        result = new StringMapPropertyEditor();
      else if (_propertyDescriptor.getValue("isRightMap") == Boolean.TRUE)
        result = new RightMapPropertyEditor();
      else if (_propertyDescriptor.getValue("isActionMap") == Boolean.TRUE)
        result = new AssociationMapPropertyEditor();
      else if (_propertyDescriptor.getValue("isBaseObjectMap") == Boolean.TRUE)
        result = new BaseObjectMapPropertyEditor();
      else if (_propertyDescriptor.getValue("isStringTwoKeysMap") == Boolean.TRUE)
        result = new StringTwoKeysMapPropertyEditor();
      else result = new SimplePropertyEditor();

                  // for testing purpose
                  //System.out.println("SimplePropertyEditor.valueOf() after the editor init");

      //
      // --------------- common initialization for  --------------------------
      // ------------- the SimplePropertyEditor class and its subclasses -----
      //
      result.source_ = _source;
      result.propertyDescriptor_ = _propertyDescriptor;
      result.displayName_ = _propertyDescriptor.getDisplayName();

                  // for testing purpose
                  //if ( result.propertyDescriptor_.getName().equals("InheritedGroupRights") )
                    //  System.out.println("in SimplePropertyEditor.valueOf(), propertyDescriptor_.getName()"+
                      //  " is InheritedGroupRights"); 
      
      try {
        // initializes readMethod_
        if (result.propertyDescriptor_.getReadMethod() != null) {
          result.readMethod_ = result.propertyDescriptor_.getReadMethod();

                  // for testing purpose
                  //if ( result.propertyDescriptor_.getName().equals("SoftAlias") )
                    //  System.out.println("in SimplePropertyEditor.valueOf(), propertyDescriptor_.getReadMethod()"+
                      //  " is "+result.propertyDescriptor_.getReadMethod());

        } // if

        // initializes writeMethod_
        if (result.propertyDescriptor_.getWriteMethod() != null) {
          result.writeMethod_ = result.propertyDescriptor_.getWriteMethod();
        } // if

        // initializes the property class.
        // it is overidden by the subclasses.
        result.propertyType_ = result.propertyDescriptor_.getPropertyType();
        
                  // for testing purpose
                  //System.out.println("in SimplePropertyEditor.valueOf(), result.propertyType_: "+
                    //result.propertyType_);
        
        if (result.propertyDescriptor_.getValue("isRestrictedArgument") == Boolean.TRUE) {
            result.isRestrictedArgument_ = true; 
            
            // initializes the BeanInfo
                // gName is never null, since it is never a generic interface
                // without a direct implementation in gBase
            String gName = UtilityImpl.getFirstInstance().getClassNameFromArrayOrInterface(
                                            UtilityImpl.getFirstInstance().
                                                getACSVocabularyFromArrayOrInterface(
                                                    _source.getClass().getName()),
                                                    _source.getClass().getName());
            beanName = UtilityImpl.getFirstInstance().getBeanInfoNameFromClass(gName);
            Class beanInfoClass = Class.forName(beanName);
            result.sourceBeanInfo_ = (BeanInfoPattern) beanInfoClass.newInstance();

            gName = null;
            beanName = null;
            beanInfoClass = null;
        } // if 'isRestrictedArgument'
        
                  // for testing purpose
                  //System.out.println("in SimplePropertyEditor.valueOf(), result.isRestrictedArgument_: "+
                    //result.isRestrictedArgument_);

                  // for testing purpose
                  //System.out.println("in SimplePropertyEditor.valueOf(), result.propertyDescriptor_.getName(): "+
                    //result.propertyDescriptor_.getName());
              
        // registers result as a listener
            // if it is a bound property
        if (result.propertyDescriptor_.isBound()) {
            // _source has to be a BoundBean
            ((BoundBean)_source).addPropertyChangeListener(
              result.propertyDescriptor_.getName(), result);

                    //if (result.propertyDescriptor_.getName().equals("OneTargetLinkedPrivilege"))
                      //  System.out.println("SimplePropertyEditor.valueOf(), "+
                        //        "listen 'OneTargetLinkedPrivilege' in "+_source);
        } // if

      } catch (Exception l) {
          l.printStackTrace();
          throw new ProcessError("from SimplePropertyEditor.valueOf(): "+
            l.toString()+" for "+_propertyDescriptor.getDisplayName());
      } // catch
      return result;
  } // valueOf()

  // ----------------------- PUBLIC METHODS ---------------------------------


  /**
   * Sets a new value. No action if the write method is unknown.
   * Works when the property is a primitive: boolean, which is
   * set with Boolean, and int which is set with Integer.
   * Noop if the value Class is not the property type, or if there is no write method,
   * or if the invocation to the write method fires an exception,
   * or if the String value is superior to 200 - then,
   * a user error dialog is displayed, and
   * a ProcessError is created but not thrown, since the method signature
   * does not allow it.
   * @param _value object to be edited. This
   *     object is not modified by the PropertyEditor. May be null.
   */
  public void setValue(Object _value) {
    if (writeMethod_ == null) return;
    if (_value != null &&
      !propertyType_.isAssignableFrom(_value.getClass()) &&
      !propertyType_.isPrimitive()) {
        new ProcessError("From SimplePropertyEditor.setValue(),"+
          "no operation because the value class "+_value.getClass()+
          " is not the property type "+propertyType_);
        return;
    }
    if (_value != null && _value instanceof String && ((String)_value).length() > 200) {
        new ProcessError("From SimplePropertyEditor.setValue(),"+
          "no operation because the value class "+
          " is String, and the value length is superior to 200.");
        DialogBox.showErrorMessage(
            "The text is too long. 200 characters is the limit.");
        return;
    }
              // for testing and tracking
              //System.out.println("In SimplePropertyEditor.setValue(), for "+
              //  source_.getName()+" value for the property "+
              //  propertyDescriptor_.getDisplayName()+" is: "+_value);

    Object[] l_argument = new Object[1];
    
    if (propertyDescriptor_.getValue("callerIndexInWriteArgument") != null) {
      if (propertyDescriptor_.getValue("callerIndexInWriteArgument").equals(
        new Integer(1)) ) l_argument = new Object[] {_value, this};

      else if (propertyDescriptor_.getValue("callerIndexInWriteArgument").equals(
        new Integer(0)) ) l_argument = new Object[] {this, _value};
    }
    else l_argument = new Object[] {_value};

    try {

              // for testing and tracking
              //System.out.println("In SimplePropertyEditor.setValue(), writeMethod_: "
                //+writeMethod_+", and argument is: "+_value);

      writeMethod_.invoke(source_, l_argument);

    } catch (InvocationTargetException i) {
      // shows the message to the user
      int lastIndex = ((InvocationTargetException)i).getTargetException().toString().lastIndexOf(": ");
      if (lastIndex != 0)
        DialogBox.showErrorMessage(
          ((InvocationTargetException)i).getTargetException().toString().substring(lastIndex+1));
      else DialogBox.showErrorMessage(
        ((InvocationTargetException)i).getTargetException().toString());

      // CreateError is for the user errors
      //if (!(i.getTargetException() instanceof CreateError)) {
        //  System.out.println("in SimplePropertyEditor.setValue(), "+
          //  "the exception is an InvocationTargetException: ");
          //i.getCause().printStackTrace();
      //}
      
      // restores the base value in the editor
      propertyChange(new PropertyChangeEvent(source_, propertyDescriptor_.getName(), null, null));
      
    } catch (Exception e) {
        if (e instanceof CreateError || e instanceof UpDateError || e instanceof BaseError) {
          DialogBox.showErrorMessage(e.getCause().getMessage());
          //System.out.println("in CollectionPropertyEditor.setValue(), exception: ");
          //i.getCause().printStackTrace();
          return;
        }

        new ProcessError("in SimplePropertyEditor.setValue(), exception: "+e.toString());
        e.getCause().printStackTrace();
    } // catch
    l_argument = null;
  } // setValue(Object value)

  /**
   * Gets the value of the property, from the gBase instance,
   * through the property descriptor in the relevant BeanInfo class.
   * Caution: if the property type is boolean,
   *  the returned value is Boolean.TYPE, and it is Integer.TYPE for int.
   * Create a ProcessError without thrown it, if there is an exception
   * fired by java.lang.reflect.Method.invoke().
   * Called by StringMapPropertyEditor and RightMapPropertyEditor.
   * @return The value of the property. Is an empty object if there is 
   *   no read method, and null if it is the returned value.
   */
  public Object getValue() {
	  // initialization
    Object value = EMPTY_OBJECT;
    if (readMethod_ == null) return value;
    try {
      value = readMethod_.invoke(source_, (Object[])null);
    } catch (Exception l) {
      new ProcessError("from SimplePropertyEditor.getValue() for "+displayName_+
              //" and read method "+readMethod_+" and source "+source_+
              ", exception "+l.toString()+" for read method "+readMethod_+
              " on the source "+source_);
      return value;
    }
            // for testing and tracking
            //if (value == null)
              //System.out.println("SimplePropertyEditor.getValue(), value is null "+
                //" for the property: "+displayName_);
            //else System.out.println("SimplePropertyEditor.getValue(), value: "+value+
              //" with the type: "+value.getClass()+//", for the property type: "+propertyType_+
              //" and the property: "+displayName_);

    return value;
  } // getValue()


  // ---------------------------------------------------------------------

  /**
   * Gets the property value as a string suitable for presentation
   * to a human to edit.
   * @return The property value as a string suitable for presentation
   *       to a human to edit. Is " " if the value is null.
   * <p>   Returns "<no_readable_value>" if the value can't be expressed as a string.
   */
  public String getAsText() {

                  // for testing and tracking
            //System.out.println("The propertyType_ in "+
            //  "SimplePropertyEditor.getAsText() is: "+
            //  propertyType_);

    if (propertyType_ == Boolean.TYPE) {
        if (getTags().length != 0) {
            if (getValue().equals(Boolean.TRUE)) return getTags()[0];
            else return getTags()[1];
        } else {
            if (getValue().equals(Boolean.TRUE)) return "TRUE";
            else return "False";
        }
    }
    else if (propertyType_ == Integer.TYPE) {
        if (getValue() == null) return new String(" ");
        return ((Integer)getValue()).toString();
    }
    else if (propertyType_ == String.class) {
        if (getValue() == null) return new String(" ");
        return (String)getValue();
    }
    else if (ImmutableName.class.isAssignableFrom(propertyType_)) {
        // returns the last component of the name
        if (getValue() == null) return new String(" ");
        else return ((ImmutableName)getValue()).toString();
    }
    else if (BaseObject.class.isAssignableFrom(propertyType_)) {
        // returns the textual form of the name
        if (getValue() == null) {

                  // for testing and tracking
            //System.out.println("The text in "+
            //  "SimplePropertyEditor.getAsText() is one space"+
            //  " and propertyType_ is a BaseObject.");

          return new String(" ");
        }
        else {

                                      // for testing and tracking
                                //System.out.println("The text in "+
                                  //"SimplePropertyEditor.getAsText() is a BaseObject name for the type "+
                                  //propertyType_+" and getValue() "+getValue().getClass());

          return ((BaseObject)getValue()).getName().toString();
        } // else
    } // else

                  // for testing and tracking
            //System.out.println("The text in "+
            //  "SimplePropertyEditor.getAsText() is not matched to a case.");

    return new String("<no_readable_value>");
  } // getAsText()

  /**
   * Sets the property value by parsing a given String.  May raise
   * java.lang.IllegalArgumentException if either the String is
   * badly formatted or if this kind of property can't be expressed
   * as text.
   * If the property type is a name, suppose that _text is conform
   * to the standard form with ' :: ' between name components.
   * Caution: getAsText() returns all the components of the name to display,
   * and the name to enter is also the full one.
   * If the property type is a BaseObject, it cannot be an editable
   * property so far.
   * @param _text is the string to be parsed.
   * @exception if the argument is null, empty or not well matched.
   */
  public void setAsText(String _text) throws java.lang.IllegalArgumentException {

    if (propertyType_ == Boolean.TYPE) {
        if (getTags().length != 0) {
            // compare to the first tag (an equivallent of "yes")
            if (getTags()[0].equalsIgnoreCase(_text)) {
              setValue(Boolean.TRUE);
              return;
            }
            else setValue(Boolean.FALSE);
            return;
        }
        else {
            if (_text.equals("TRUE"))
                    setValue(Boolean.TRUE);
            else setValue(Boolean.FALSE);
        }
    } // if (propertyType_

    else if (propertyType_ == String.class) {
        setValue(_text);
        return;
    }

    else if (propertyType_ == Integer.TYPE) {
        setValue(Integer.valueOf(_text));
        return;
    }

    // if propertyType_ is a BaseObject
    else if ( BaseObject.class.isAssignableFrom(propertyType_) ) {
        ImmutableName name;
        try {
          name = ACS_Facade.getNameFromString(_text);
        } catch (BaseError b) {
          resetUpdate(", the new value is not a correct name");
          return;
        } // catch (BaseError b) {

        if (ACSObject.class.isAssignableFrom(propertyType_) ||
                StringRight.class.isAssignableFrom(propertyType_) ) {
          BaseObject object;
          try {
            object = ACS_Facade.getEorBaseObject(name);
          } catch (BaseError b) {
            resetUpdate(", the text is not the name of a known object");
            object = null;
            name = null;
            return;
          } // catch

          if (propertyType_.isAssignableFrom(object.getClass()) ) {
            setValue( object );
            object = null;
            name = null;
            return;
          } else {
            resetUpdate(", the new value type is not a "+propertyType_.toString() );
            object = null;
            name = null;
            return;
          } // else
        } // if (ACSObject.class.isAssignableFrom(propertyType_))
        else throw new java.lang.IllegalArgumentException("FromSimplePropertyEditor"
          +".setAsText(), it is the name of an unknown BaseObject");

    } // if ( BaseObject.class.isAssignableFrom(propertyType_) )

    // if propertyType_ is a name
    // suppose that _text is conform to the standard form with ' :: '
    // between name components
    else if ( ImmutableName.class.isAssignableFrom(propertyType_) ) {
        try {
          setValue(ACS_Facade.getNameFromString(_text));
        } catch (BaseError b) {
          throw new java.lang.IllegalArgumentException("FromSimplePropertyEditor"
            +".setAsText(), exception: "+b.toString());
        }
        return;
    }

    throw new java.lang.IllegalArgumentException(_text);
  } // setAsText(String _text)


  // ----------------------------------------------------------------------

  /**
   * Gets the value of the property source.
   *
   * @return the source of the property.
   */
  public BaseObject getSource() {
    return source_;
  }  // getSource()

  /**
   * Gets the value of the display property name (not the true name, given by
   * the property descriptor).
   *
   * @return the display name of the property.
   */
  public String getDisplayName() {
    return displayName_;
  }  // getDisplayName()

  /**
   * Gets the value of the property type, that is the Class of the returned
   * object in getValue(): boolean (not Boolean here), String, ImmutableName,.
   * or a BaseObject instance.
   * @return The type of the property.
   */
  public Class getPropertyType() {
    return propertyType_;
  }  // getPropertyType()

  /**
   * Gets the component displayed in the beamer that handles the property
   * directly.
   * That is the very component which displays it, not the panel arround.
   * It is a JComboBox if the property is a Boolean,
   * and a JTextField if the property is a String, an ImmutableName or a BaseObject,
   * or a JList in a CollectionPropertyEditor.
   *
   * @return The component displayed in the beamer.
   */
/*  public JComponent getBoundComponent() {
    return boundComponent_;
  }  // getBoundComponent()
*/
  /**
   * Gets the descriptor of the property.
   *
   * @return The descriptor of the property.
   */
  public PropertyDescriptor getPropertyDescriptor() {
    return propertyDescriptor_;
  } // getPropertyDescriptor()

  /**
   * Gets the method on the source for updates.
   *
   * @return The method on the source for updates.
   */
  public Method getWriteMethod() {
    return writeMethod_;
  } // getWriteMethod

  /**
   * When the value is a boolean, the tags are empty to indicate that
   * a checkbox is required, excepted if a couple of tags is set
   * in the property descriptor with the key 'tags', to indicate
   * that a combobox with the tags is required.
   * @return the tags or an empty array (length == 0) for a boolean, and null else.
   *
   */
  public String[] getTags() {
    if (propertyType_ == Boolean.TYPE) {

      if (propertyDescriptor_.getValue("tags") != null) {
        return (String[]) propertyDescriptor_.getValue("tags");
      }

      else return new String[0];

    } // if (propertyType_ == Boolean.TYPE)
    return null;
  } // getTags()

  // -------------------------------------------------------------------------

  /**
   * Determines whether the propertyEditor can provide a custom editor.
   *
   * @return true.
   */
  public boolean supportsCustomEditor() {
      return true;
  }

  /**
   * Returns a full custom Component that edits its property value,
   * with the relevant buttons following the property descriptor.
   * Called by BeamerManager.addPropertyEditor().
   * @return a JPanel that will allow a human to directly
   *      edit the current property value.
   */
  public java.awt.Component getCustomEditor() {

    Component result = InternalFrameUtilities.getEmptyPanel();

    if (BaseObject.class.isAssignableFrom(propertyType_) ) {
      // initialization
      boolean withWriteMethod = false;
      if (writeMethod_ != null) withWriteMethod = true;

      result = getGeneralEditor(displayName_, getAsText(),
        withWriteMethod, propertyDescriptor_.isBound(), true);
      //System.out.println("SimplePropertyEditor.getCustomEditor(): "+propertyType_);
    }
    else if (propertyType_ == Boolean.TYPE) {
      result = getBooleanEditor();
      //System.out.println("SimplePropertyEditor.getCustomEditor(): "+propertyType_);
    }
    else if ((propertyType_ == String.class) ||
      (propertyType_ == Integer.TYPE) ||
      (ImmutableName.class).isAssignableFrom(propertyType_) ) {

        // initialization
        boolean withWriteMethod = false;
        if (writeMethod_ != null) withWriteMethod = true;

                // for testing and tracking
                //if (displayName_.endsWith("Comments: ")) {
                  //   System.out.println("SimplePropertyEditor.getCustomEditor() for a comment "+
                    //   displayName_+" is "+getAsText()+ " and bound or not: "+propertyDescriptor_.isBound());
                //} // if
        
        result = getGeneralEditor(displayName_, getAsText(),
          withWriteMethod, propertyDescriptor_.isBound(), false);
        //System.out.println("SimplePropertyEditor.getCustomEditor() for a string, an integer or a name");

    } // else

    return result;
  } // getCustomEditor()

  /**
   * This method is intended for use when generating Java code to set
   * the value of the property.  It should return a fragment of Java code
   * that can be used to initialize a variable with the current property
   * value.
   * <p>
   * Example results are "2", "new Color(127,127,34)", "Color.orange", etc.
   *
   * @return "???".
   */
  public String getJavaInitializationString() {
      return "???";
  }

  /**
   * Get the prefered size for the rectangle into which the value is displayed,
   * in the component returned by getCustomEditor().
   *
   * @return prefered size.
   */
  public Dimension getPreferedSizeForEditing() {
    return new Dimension(400, HEIGHT_DIMENSION); //400 is an estimation from CommonPropertyEditor values
  }

    //----------------------------------------------------------------------


    /**
     * Determines whether the class will honor the paintValue method.
     * @return  false.
     */
    public boolean isPaintable() {
	return false;
    }

    /**
     * Paint a representation of the value into a given area of screen
     * real estate.  Note that the propertyEditor is responsible for doing
     * its own clipping so that it fits into the given rectangle.
     * If the PropertyEditor doesn't honor paint requests (see isPaintable)
     * this method should be a silent noop.
     * @param gfx  Graphics object to paint into.
     * @param box  Rectangle within graphics object into which we should paint.
     */
    public void paintValue(java.awt.Graphics gfx, java.awt.Rectangle box) {
    }

    //----------------------------------------------------------------------

  /**
   * Register a listener for the PropertyChange event - NO USE SO FAR.
   * The class will fire a PropertyChange value whenever the value is updated.
   * @param _listener  An object to be invoked when a PropertyChange
   *		event is fired.
   */
  public synchronized void addPropertyChangeListener(
    PropertyChangeListener _listener) {
      if (listeners_ == null) {
          listeners_ = new java.util.Vector(2);
      }
      listeners_.addElement(_listener);
  } // addPropertyChangeListener()

  /**
   * Remove a listener for the PropertyChange event - NO USE SO FAR.
   * @param _listener  The PropertyChange listener to be removed.
   */
  public synchronized void removePropertyChangeListener(
    PropertyChangeListener _listener) {
      if (listeners_ == null) {
          return;
      }
      listeners_.removeElement(_listener);
  } // removePropertyChangeListener()

  /**
   * Manages the events fired by the components in the JPanel
   * displayed in the beamer, returned by getCustomEditor().
   * The components may be JComboBox, JCheckBox, JButton or JTextField.
   * The buttons are 'See', 'Select', 'Clear'.
   * For a BaseObject property, use the DialogBox class for the 'Select',
   * and use the setAsText() method for the direct updates in the JTextField.
   * DialogBox.showExplorerSelectionDialog() calls handleSelectionDialog().
   * <p>
   * When a string is entered in a JTextField, its length is limited to 200.
   * Then, after the user text, a space is added to the end of the string.
   * When an integer is entered in a JTextField, its length is limited to 50.
   * When a name is entered, this method checks its format and if it is wrong,
   * it displays an error message and returns.
   * @param _evt is the event fired by a JComboBox, a JCheckBox, a JTextField
   *  or a JButton class
   */
  public void actionPerformed(ActionEvent _evt) {

    // no action if the explorer is currently used for an object selection 
    // or if there is an user dialog currently
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
      
    String userText = "";

    // for boolean properties
    if (_evt.getSource().getClass() == JCheckBox.class) {
      if ( ((JCheckBox)_evt.getSource()).isSelected() )
            setValue(Boolean.TRUE);
      else setValue(Boolean.FALSE);
      userText = null;
      return;
    }    
    
    // for boolean properties and String restricted values
    if (_evt.getSource().getClass().equals(JComboBox.class)) {
      userText = (String) ((JComboBox)_evt.getSource()).getSelectedItem();
      setAsText(userText);
      userText = null;
      return;
    }

    // for String name and int properties
    else if (_evt.getSource().getClass().equals(JTextField.class)) {
      userText = ((JTextField)_evt.getSource()).getText();
      // for Name property
      if (ImmutableName.class.isAssignableFrom(propertyType_)) {
        userText = userText.trim();
        // checks the Name for the user
        try {
          ACS_Facade.getNameFromString(" "+userText+"  ");
        } catch (BaseError b) {
          DialogBox.showErrorMessage("Sorry, the name format is not fulfilled. "+
                  "Each name component has to end with ':: '.");
          userText = null;
          return;
        }

        // the String form of a Name has leading and trailing whitespaces
        setAsText(" "+userText+"  ");
      } // if (ImmutableName.class

      // for int
      else if (propertyType_ != Integer.TYPE) {
          if (userText.length() < 200) setAsText(userText+" ");
      } else {
          if (userText.length() < 50) setAsText(userText);
      }
      userText = null;
      return;
    }

    // for BaseObject properties
    else if (_evt.getSource().getClass().equals(JButton.class)) {

      // for viewing an object
      if ( ((JButton)_evt.getSource()).getText().equals(" See ")) {

        viewOneElement( getValue() ); // finalize this instance, if in the beamer

      // for selecting a present object
      } else if( ((JButton)_evt.getSource()).getText().equals("Select")) {

        HashSet l_type = new HashSet(INITIAL_CAPACITY);
        l_type.add(propertyType_);
        // DialogBox.showExplorerSelectionDialog() calls handleSelectionDialog()
        try {
          DialogBox.getInstance().showExplorerSelectionDialog(this, null, l_type, null, true);
        } catch (SmallError s) {
          throw new InternalError("In SimplePropertyEditor.actionPerformed()"+
            ", DialogBox.showExplorerSelectionDialog() throws: "+s.toString());
        } // catch

        l_type = null;
        return;
      } // if

      // for setting the value to null
      else if( ((JButton)_evt.getSource()).getText().equals(" Clear")) {

        setValue(null);

      } // else if == "Clear"
      
      // for creating a new value. NOT USED.
      else if( ((JButton)_evt.getSource()).getText().equals(" New  ")) {

        if ( !ImmutableACS.class.isAssignableFrom(propertyType_) ) {
          // create any new BaseObject, except new ACS, with some fixed arguments
          Object[] l_fixedValue = getKnownValuesForNewElement();

            // use the constructor method, and
            // call back the handleNewBaseObject() method
          BaseObjectHandler.valueOf(this, propertyType_).
            getBaseObjectEditor(l_fixedValue);
          return;
        } // if
        else {
          // noop
          return;
        } // else !ImmutableACS.class.isAssignableFrom(propertyType_)

      } // else if == "New"

    } // if (_evt.getSource().getClass().equals(JButton.class))

  } // actionPerformed()

  /**
   * Manages the events fired by the base object source_ displayed in the beamer,
   * if it is a bound property and when the property value changes.
   * The main behavior is that the old and new property values in the argument
   * are not ridden to update the values. This allow to process a PropertyChangeEvent
   * that does not describe these values, but is fired by the gBase package when
   * the relevant values are changed in practice.
   * Before updating the component in the GUI, unregisters this instance
   * as ActionListener to avoid loops.
   * The properties of _evt are never used there, since this method reads directly
   * the new value in the base.
   * @param _evt only source and property name in _evt are used in this method
   */
  public void propertyChange(PropertyChangeEvent _evt) {

    if (_evt.getSource() == source_ &&
      _evt.getPropertyName().equals(propertyDescriptor_.getName()) &&
      boundComponent_ != null) {

                            // for testing purpose
                            //System.out.println("SimplePropertyEditor.PropertyChangeEvent() for "+
                              //_evt.getPropertyName()+" for the source "+((BaseObject)_evt.getSource()).getName()+
                              //" and the boundComponent_ class "+boundComponent_.getClass());

        if (boundComponent_ instanceof JTextField) {
          // for BaseObject property, this instance is not a listener
          // of the JTextField
          if (propertyType_ != String.class && propertyType_ != Integer.TYPE &&
            !ImmutableName.class.isAssignableFrom(propertyType_) ) {
              ((JTextField)boundComponent_).setEditable(true);
              ((JTextField)boundComponent_).setText( getAsText() );
              ((JTextField)boundComponent_).setEditable(false);
              // user message:
              // DialogBox.getInstance().showOkMessage(displayName_+
              //  " is updated.");
          }
          else {
            ((JTextField)boundComponent_).removeActionListener(this);
            ((JTextField)boundComponent_).setText( getAsText() );
            ((JTextField)boundComponent_).addActionListener(this);
          }
        } // if (boundComponent_ instanceof JTextField)

        else if (boundComponent_ instanceof JComboBox) {
          ((JComboBox)boundComponent_).removeActionListener(this);
          ((JComboBox)boundComponent_).getModel().
            setSelectedItem( getAsText() );
          ((JComboBox)boundComponent_).addActionListener(this);
        } // else if

        else if (boundComponent_ instanceof JCheckBox) {
           ((JCheckBox)boundComponent_).removeActionListener(this);
           ((JCheckBox)boundComponent_).setSelected( ((Boolean)getValue()).booleanValue() );
           ((JCheckBox)boundComponent_).addActionListener(this);
        } // else if        
    } // _evt.getSource() == source_

  } // propertyChange()

    //----------------------------------------------------------------------

  /**
   * Get the values of variables for the instance.
   *
   * @return source, propertyName and listeners (if they have toString() methods).
   */
  public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("  SIMPLEPROPERTYEDITOR.TOSTRING():");

        sb.append(" source_ = ");
        sb.append(source_);

        sb.append("; display name = ");
        sb.append(displayName_);

        sb.append("; listeners list = ");
        sb.append(listeners_);

        return " [" + sb + "] ";
  } // toString()

  /**
  * Called by DialogBox.showExplorerSelectionDialogt() for actionPerformed(),
  * when the property is a BaseObject which has been selected in the explorer.
  * Overridden method by RightMapPropertyEditor.
  * @param _validatedObjectsSetMap with one element, where
  * - the key is propertyType_,
  * - the value is a List of one selected object.
  * May be null or empty if the user cancels the operation.
  */
  public void handleSelectionDialog(Map _validatedObjectsSetMap) {

    // stop if the argument value is null or empty
    if (_validatedObjectsSetMap == null || _validatedObjectsSetMap.isEmpty()) {
        return;
    } // if
      
    List list = (List)_validatedObjectsSetMap.get(propertyType_);

    //String userText = "";
    if (list != null)
      setAsText( ((BaseObject)list.get(0)).getName().toString() );

        // for testing purpose
        //System.out.println("handleSelectionDialog called in SimplePropertyEditor");

    list = null;
  } // handleSelectionDialog()

  /**
  * NO USE SO FAR.
  * Called by BaseObjectHandler for actionPerformed(),
  * when the property is a BaseObject
  * which has been created from scratch, and then set as a property.
  * @param _obj is the BaseObject instance to set for this property
  */
  public void handleNewBaseObject(BaseObject _obj) {

        // for testing and tracking
        //System.out.println("SimplePropertyEditor.handleNewBaseObject() call for "+
        //  displayName_+", the new object is:"+_obj);

    setValue( _obj);
  } // handleNewBaseObject()

  /**
   * Finalizes the instance. The call is reserved to BeamerManager.
   */
  public void finalizeForProcess() {

    // listeners_ are not used so far
    if (listeners_ != null) {
      synchronized (listeners_) {
          // Tell our listeners that "everything" has changed.
          PropertyChangeEvent evt = new PropertyChangeEvent(
            this, propertyDescriptor_.getName(), null, null);

          for (int i = 0; i < listeners_.size(); i++) {
              PropertyChangeListener target =
                (PropertyChangeListener)listeners_.elementAt(i);
              target.propertyChange(evt);

        // for testing and tracking
        //System.out.println("a listener to inform in "+
        //  "SimplePropertyEditor.finalizeForProcess() for "+
        //  displayName_+" is "+target );

          } // for
        } // synchronized (listeners_)
    } // if (listeners_ != null)

    // unregisters as a listener
        // if it is a bound property
    if (propertyDescriptor_.isBound()) {
      // then _source has to be a BoundBean
      ((BoundBean)source_).removePropertyChangeListener(
        propertyDescriptor_.getName(), this);
    }

    propertyType_ = null;
    displayName_ = null;
    source_ = null;
    propertyDescriptor_ = null;
    if (boundComponent_ != null) boundComponent_.setVisible(false);
    boundComponent_ = null;
    readMethod_ = null;
    writeMethod_ = null;
    listeners_ = null;
  } // finalizeForProcess()

  /**
   * Gets the restricted String values for the argument or the key.
   * When a value is restricted, the allowed values are get:
   * - 1 - from the property descriptor if 'staticSetOfItems' is not null, <P>
   * - 2 - uses the 'choiceMethodForNew' to call a method which returns
   *    a String array, a List or a Set of Strings, and this method is from
   *    a specified objet or from the source, <P>
   * - 3 - from BaseUtilityImpl.getL_AcsRestrictedTypes() if there is a matching key. <P>
   * For the subclasses of this class where the property is a map, this method
   * is used only for the map keys, not for the map values.
   * Called by actionPerformed() when the value has been entered by the user,
   * CollectionPropertyEditor.addOneStringOrNameElements() and
   * StringMapPropertyEditor.addKeyElement().
   * Calls BaseUtilityImpl.getL_AcsRestrictedStrings().
   * @param _acs is the new object ACS. Not null only for the third case with
   * BaseUtilityImpl.getL_AcsRestrictedTypes().
   * @return the restricted values, or null if they are not found
   */
  protected String[] getRestrictedValues(ACS _acs) {

                    // for testing and tracking
                    //System.out.println("in SimplePropertyEditor.getRestrictedValues(), "+
                      //"isRestrictedArgument_ is :"+isRestrictedArgument_);
      
    if (!isRestrictedArgument_) return null;
      
      // initialization
    String[]                result = null;
    boolean                 found = false;
    String                  propertyName = propertyDescriptor_.getName();

                                    // for testing and tracking
                                    //System.out.println("in SimplePropertyEditor.getRestrictedValues(), "+
                                      //"for ACS:"+_acs);

    // first search of the restricted values in the property descriptor
    if (propertyDescriptor_.getValue("staticSetOfItems") != null) {

                // for testing and tracking
                //System.out.println("in SimplePropertyEditor.getRestrictedValues(), for "+displayName_+
                  //", l_restrictedValues is set by 'staticSetOfItems' in the BeanInfo. ");
                
        found = false;
        Object[] temp = (Object[])propertyDescriptor_.getValue("staticSetOfItems");
        if (temp instanceof String[]) {
            found = true;
            result = (String[])temp;
        }
        temp = null;
    } // if "staticSetOfItems"
    else { // second search in the property descriptor
        
        Object[] l_choice;
        Method[] l_methods;
        if (propertyDescriptor_.getValue("choiceMethodForNew") != null) {

                            // for testing and tracking
                            //System.out.println("in SimplePropertyEditor.getRestrictedValues(), for "+
                              //      displayName_+
                              //", l_restrictedValues is set by 'choiceMethodForNew' in the BeanInfo. ");

          l_choice = (Object[]) propertyDescriptor_.getValue("choiceMethodForNew");
          // example of value for l_choice:
          //      pd_one.setValue("choiceMethodForNew", new Object[] {
          //        ActionNewResource.getInstance(), "createResource"} );
          // or to call the source:  pd_seventyTwo.setValue("choiceMethodForNew", new Object[] {
          //        null, "getL_MapKeysStartsForTypePolicies"} );        
          if (l_choice[0] != null) 
              l_methods = l_choice[0].getClass().getMethods();
          else
              l_methods = source_.getClass().getMethods();

          for (int k = 0; k < l_methods.length; k++){

                        //System.out.println("in SimplePropertyEditor.getRestrictedValues(),"+
                          //" source methods: "+l_methods[k].getName());

            try {
              if (l_methods[k].getName().equals(l_choice[1])) {

                        // for testing and tracking
                        //System.out.println("in SimplePropertyEditor.getRestrictedValues(),"+
                          //" right source method is found for the object (may be null): "+
                          //l_choice[0]);
                          //" with the arguments:");
                        //for (int i = 0; i < l_fixedValue.length; i++) {
                          //  System.out.println("-- the fixed value at "+i+" is "+l_fixedValue[i]);
                        //}

                  // null argument in all cases
                  if (l_choice[0] != null) {
                    if (l_methods[k].invoke(l_choice[0], new Object[0]) instanceof String[])
                        result = (String[])l_methods[k].invoke(l_choice[0], new Object[0]);
                    else if (l_methods[k].invoke(l_choice[0], new Object[0]) instanceof List)
                        result = (String[]) ((List<String>)l_methods[k].invoke(l_choice[0], 
                                new Object[0])).toArray(new String[0]);
                    else if (l_methods[k].invoke(l_choice[0], new Object[0]) instanceof Set)
                        result = (String[]) ((Set<String>)l_methods[k].invoke(l_choice[0], 
                                new Object[0])).toArray(new String[0]);
                  }
                  else {
                    if (l_methods[k].invoke(source_, new Object[0]) instanceof String[])
                        result = (String[])l_methods[k].invoke(source_, new Object[0]);
                    else if (l_methods[k].invoke(source_, new Object[0]) instanceof List)
                        result = (String[]) ((List<String>)l_methods[k].invoke(source_, 
                                new Object[0])).toArray(new String[0]);                      
                    else if (l_methods[k].invoke(source_, new Object[0]) instanceof Set)
                        result = (String[]) ((Set<String>)l_methods[k].invoke(source_, 
                                new Object[0])).toArray(new String[0]);
                  } // else
                  found = true;
                  break;
              } // if (l_methods[k]
            } catch (Exception e) {
                e.printStackTrace();
                throw new InternalError("In SimplePropertyEditor.getRestrictedValues(), "+
                "for 'choiceMethodForNew': "+e.toString());
            } // catch
          } // for k

          if (!found) {
              throw new InternalError("SimplePropertyEditor.getRestrictedValues(): "+
                "for 'choiceMethodForNew' there is no method");
          }
        } // if "choiceMethodForNew"            
        l_choice = null;
        l_methods = null;  
    } // else
    
    if (found) {
        propertyName = null;
        return result;
    }
    
    if (_acs == null) {
                    // for testing and tracking
                    System.out.println("in SimplePropertyEditor.getRestrictedValues(), for "+displayName_+
                      ", _acs is null and the returned value is null");
                    
        propertyName = null;          
        return null;
    }

                // for testing and tracking
                //System.out.println("in SimplePropertyEditor.getRestrictedValues(), for "+displayName_+
                  //", l_restrictedValues is set by BaseUtilityImpl.getL_AcsRestrictedStrings(). ");  
                // for testing and tracking
                //System.out.println("in SimplePropertyEditor.getRestrictedValues(),"+
                  //" the common bean class is: "+
                  //((Class)sourceBeanInfo_.getBeanDescriptor().getBeanClass()
    
    // third search of the restricted types in the ACS,
    // if the search in the property descriptor has not succeed    
    result = BaseUtilityImpl.getL_AcsRestrictedTypes(sourceBeanInfo_,
                    propertyName, _acs);

                
                // for testing and tracking
                //System.out.println("in SimplePropertyEditor.getRestrictedValues(),"+
                  //" l_restrictedValues is set by the ACS. "+
                  //"The restricted values are: ");
                //if (result != null) {
                  //  for (int t = 0; t < result.length; t++) {
                    //   System.out.println(result[t]); 
                    //} // for t
                //}

    propertyName = null;
    return result;
  } // getRestrictedValues()

  /**
   * Gets the status of buttons: to display or not in the property editor,
   * through a reading of the property descriptor.
   * Generally speaking, return the buttons 'New'/ 'Delete' when
   * they are not external reference objects, or 'Select'/ Remove'
   * when they are external reference objects, or 'Delete' only, or nothing.
   * The 'See' (or 'View') button may be always added if it is not a string.
   * The 'Previous' button may be added if the previous property editor
   * handles an array with the same type.
   * 'Delete' goes generally with 'New'.
   * On the contrary, 'Remove' goes generally with 'Select'.
   * The buttons 'NewV'/ 'DeletV' are used for handling the values in a map, as
   * an alternative to 'Select'/ 'Remove'. For a map, 'New'/ 'Delete' is reserved
   * to the keys.
   * @see #setVerticalButtonsInEditor
   * @return an array of 9 booleans: withPreviousButton at 0, withViewButton,
   *   withSelectButton, withNewButton, withDeleteButton, withRemoveButton,
   *   withClearButton at 6, withNewVButton, withDeletVButton at 8
   */
  protected boolean[] getButtonsDisplaying() {

                    // for testing and tracking
                    //System.out.println("start of SimplePropertyEditor.getButtonsDisplaying()");
      
    boolean[] l_result = new boolean[9];

    if (propertyDescriptor_.getValue("withPreviousButton") == Boolean.TRUE)
      l_result[0] = true;
    
    if (propertyDescriptor_.getValue("withViewButton") == Boolean.TRUE)
      l_result[1] = true;
    
    // Select may be forbidden, in general or for the Architecture view
    if ( 
            !BeamerManager.getForbidsForArchitecture(source_, propertyDescriptor_, 
            "withSelecButton") && 
            propertyDescriptor_.getValue("withSelectButton") == Boolean.TRUE) {
                    
                l_result[2] = true;
             
                  // for testing purpose
                  //System.out.println("in SimplePropertyEditor.getButtonsDisplaying(),"+
                    //" the property is: "+propertyDescriptor_.getName()+
                    //" and the 'Select' button is set."); 
    } // if

    if (propertyDescriptor_.getValue("withNewButton") == Boolean.TRUE ) {
          // get the acs to check the right to create this object
          ACS acs = null;
          if (source_ instanceof ACS) {
             acs = (ACS)source_;
             
                  // for testing purpose
                  //System.out.println("in SimplePropertyEditor.getButtonsDisplaying(),"+
                    //" the property type is: "+propertyType_);             
             
             if (
                     // create resource in an acs
                 (ImmutableResource.class.isAssignableFrom(propertyType_) 
                          && !acs.getManageResources() ) ||
                     // create eligibleparty in an acs
                 (ImmutableEligibleParty.class.isAssignableFrom(propertyType_) 
                          && !acs.getManageAccounts()
                          && !acs.getManageGroups() && !acs.getManageActors() )  ) {

             
                    // for testing purpose
                    //System.out.println("in SimplePropertyEditor.getButtonsDisplaying(),"+
                      //" no new button for the acs");                  
                 
                     l_result[3] = false;
             } // if
             else l_result[3] = true;
          } else if (source_ instanceof ACSObject ) {
             acs = (ACS)ACS_Facade.getOpenACS().get( ((ACSObject)source_).getAcsName() );
             if (
                     // create resource under a directory
                 source_ instanceof ImmutableNode     
                   && propertyDescriptor_.getName().equals("Children") 
                   && !acs.getManageResourceTrees()  ) {              
                     // for creating aclentry under a resource, it depends on the aclentry acs,
                     // not on the resource acs - see BaseObjectHandler, called by addElement()
                     // after a check of open ACS.

                     l_result[3] = false;
             } // if 
             else l_result[3] = true;
          } // if      
          else l_result[3] = true;
          acs = null;
    }  // if "withNewButton"

    if (propertyDescriptor_.getValue("withDeleteButton") == Boolean.TRUE)
      l_result[4] = true;

    if (
         !BeamerManager.getForbidsForArchitecture(source_, propertyDescriptor_, 
            "withRemoveButton") && 
            propertyDescriptor_.getValue("withRemoveButton") == Boolean.TRUE)
                l_result[5] = true;
    
    if (propertyDescriptor_.getValue("withClearButton") == Boolean.TRUE) {
      if (ImmutableACS.class.isAssignableFrom(propertyType_)) l_result[6] = false;
      else l_result[6] = true;
    } // if

                    // for testing and tracking
                    //System.out.println("end of SimplePropertyEditor.getButtonsDisplaying()"+
                      //      " the property is: "+propertyDescriptor_.getName()+
                      //      " and the result is: ");
                    //String[] l_ = new String[] {"Previous", "See", "Select", "New", "Delete", "Remove", "Clear", "NewV", "DeletV"};
                    //for (int i = 0; i < l_result.length; i++) {
                      //  System.out.println("---------------- "+l_[i]+" at "+i+": "+l_result[i]);
                    //}
                    //l_ = null;

    if (propertyDescriptor_.getValue("withNewVButton") == Boolean.TRUE)
      l_result[7] = true;

    if (propertyDescriptor_.getValue("withDeletVButton") == Boolean.TRUE)
      l_result[8] = true;
    
    return l_result;
  } // getButtonsDisplaying()  

  /**
   * Creates a vertical panel with the relevant buttons.
   * The current instance is a button listener.
   * The integrity rules are, as described following: <P>
   * - _forMapKey and _forMapValue are not both true <P>
   * - for collections, Select/Remove or New/Delete, but not together <P>
   * - only New and Delete for map keys <P>
   * - for map values, Select/Remove or NewV/DeletV, but not together <P>
   * The returned value is sometimes used only to test if it is null.
   * Called by CollectionPropertyEditor, StringMapPropertyEditor, RightMapPropertyEditor.
   * @see #getButtonsDisplaying
   * @param _l_buttons is an array of 9 booleans from getButtonsDisplaying()
   * @param _forMapKey is true when the panel is for the buttons of the key 
   *   in a MapPropertyEditor. Only 'New' and 'Delete' buttons are relevant then.
   * @param _forMapValue is true when the panel is for the buttons of the value 
   *   in a MapPropertyEditor. 'New' and 'Delete' buttons are never relevant then.
   * @return the panel with the right buttons and BoxLayout.Y_AXIS, or null 
   *    if there is no button to display, or if the integrity rules are not followed
   */
  protected JPanel setVerticalButtonsInEditor(boolean[] _l_buttons, 
          boolean _forMapKey, boolean _forMapValue) {
    
              // for testing purpose
         //System.out.println("start of SimplePropertyEditor.setVerticalButtonsInEditor()");
         
    if (_l_buttons.length != 9) {
        System.out.println("in SimplePropertyEditor.setVerticalButtonsInEditor(), "+
                "it is not a 9-lenght array argument for"+propertyDescriptor_.getName());
        return null;
    }
    if (_forMapKey && _forMapValue) return null;
    // for CollectionPropertyEditor
    if (!_forMapKey && !_forMapValue && 
                        !_l_buttons[0] && !_l_buttons[1] && !_l_buttons[2] &&
                        !_l_buttons[3] && !_l_buttons[4] && !_l_buttons[5] && 
                        !_l_buttons[6] && !_l_buttons[7] && !_l_buttons[8]) {

                        // for testing purpose
                    //System.out.println("in SimplePropertyEditor.setVerticalButtonsInEditor(), "+
                      //      "not button is required for "+propertyDescriptor_.getName());

        return null;
    }
    // for StringMapPropertyEditor, RightMapPropertyEditor, AssociationMapPropertyEditor
    if (_forMapKey && !_l_buttons[3] && !_l_buttons[4]) {
        //System.out.println("in SimplePropertyEditor.setVerticalButtonsInEditor(), "+
          //      "not button is required for "+propertyDescriptor_.getName());        
        return null;
    }
    // for StringMapPropertyEditor, RightMapPropertyEditor, AssociationMapPropertyEditor
    if (_forMapValue && !_l_buttons[0] && !_l_buttons[1] && !_l_buttons[2] &&
                         !_l_buttons[5] && !_l_buttons[6] && !_l_buttons[7] && !_l_buttons[8]) {
        //System.out.println("in SimplePropertyEditor.setVerticalButtonsInEditor(), "+
          //      "not button is required for "+propertyDescriptor_.getName());        
        return null;
    }
    // for StringMapPropertyEditor, RightMapPropertyEditor, AssociationMapPropertyEditor
    // Select/Remove or NewV/DeletV for values
    if (_forMapValue && (_l_buttons[2] || _l_buttons[5]) && 
            (_l_buttons[7] || _l_buttons[8])) {
        //System.out.println("in SimplePropertyEditor.setVerticalButtonsInEditor(), "+
          //      "not button is required for "+propertyDescriptor_.getName());        
        return null;
    }
    
    JPanel result = new JPanel();
    result.setLayout(new BoxLayout(result, BoxLayout.Y_AXIS));

    JButton button = null;
    JLabel label = null;

    // add the 'for key' if it is required
    if (_forMapKey && (_l_buttons[3] || _l_buttons[4])) {
      label = new JLabel("  keys:");
      label.setBackground(Color.white);
      label.setFont(BeamerManager.plainButtonFont);
      label.setMaximumSize(BUTTON_DIMENSION_2);
      label.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( label );
      label = new JLabel(" ");
      label.setMaximumSize(BUTTON_DIMENSION_3);
      label.setMinimumSize(BUTTON_DIMENSION_3);
      result.add( label );
    }    
    
    // add the See button if it is required
    if (_l_buttons[1] && !_forMapKey) {
      button = new JButton(" See ");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    // add the Previous button if it is required
    if (_l_buttons[0] && !_forMapKey) {
      button = new JButton("Prev.");
      button.setToolTipText("Previous");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }    
    if (_l_buttons[2] && !_forMapKey) {
             
                  // for testing purpose
                  //System.out.println("in SimplePropertyEditor.setVerticalButtonsInEditor(),"+
                    //" display of the Select button for "+propertyDescriptor_.getName()); 
        
      button = new JButton("Select");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    if (_l_buttons[3] && !_forMapValue) {
        
             
                  // for testing purpose
                  //System.out.println("in CollectionPropertyEditor.setButtonsInEditor(),"+
                    //" display of the New button");         
        
      button = new JButton("  New  ");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    if (_l_buttons[4] && !_forMapValue ) {
      button = new JButton("Delete");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    if (_l_buttons[5] && !_forMapKey) {
      button = new JButton("Remove");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    if (_l_buttons[6] && !_forMapKey) {
      button = new JButton(" Clear ");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    if (_l_buttons[7] && !_forMapKey && _forMapValue) {
      button = new JButton("NewV");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    if (_l_buttons[8] && !_forMapKey && _forMapValue) {
      button = new JButton("DeletV");
      button.setBackground(Color.white);
      button.setBorder(new BevelBorder(
        BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
      button.setFont(BeamerManager.plainButtonFont);
      button.addActionListener(this);
      button.setMaximumSize(BUTTON_DIMENSION_2);
      button.setMinimumSize(BUTTON_DIMENSION_2);
      result.add( button);
    }
    
    result.setBackground(Color.white);
    
              // for testing purpose
         //System.out.println("end of SimplePropertyEditor.setVerticalButtonsInEditor(), ");
    
    button = null;
    label = null;

    return result;
  } // setVerticalButtonsInEditor()

  /**
   * Called when the user clicks on the button 'See'.
   * This method selects a BaseObject in the explorer, and then it is displayed
   * in the beamer. Displays an user message if the item is not found.
   * This method finalizes this instance if it is in the beamer, through a call
   * to TreeManager which fires a property change to the BeamerManager.
   * Called from actionPerformed(), CollectionPropertyEditor and RightMapPropertyEditor.
   * Calls TreeManager.selectNodeInExplorer() for a BaseObject, and
   * BeamerManager.getBaseObject() when the argument is a name. In this latter case,
   * propertyType_ is passed on, but it is ridden only if it is an instance of StringRight
   * or an array of StringRight.
   * When the element is displayed as a String, it may be because it is trully a String,
   * or because it is the detailled name of an AclEntry or a Privilege. But for this method,
   * the true element cannot be a simple String, since it is viewable as a BaseObject.
   * Caution: for the DetailledName of an AclEntry or a Privilege, it is necessary to use
   * the ACS of the property source_ to get the object from the name. This method
   * does not make the job if the property displays an external AclEntry or Privilege
   * as a property of another ACS.
   * @param _toView BaseObject to view, or ImmutableName of the BaseObject to view,
   *   or String of DetailledName for an AclEntry or Privilege to view in the beamer
   */
  protected void viewOneElement(Object _toView) {

                // for testing and tracking
                //System.out.println("SimplePropertyEditor.viewOneElement() for a "+_toView.getClass());

    // initialization
    BaseObject      selectedObject = null;
    ImmutableName   name = null;
    ImmutableACS    acs = null;

    // gets the selected object from _toView
    if (_toView instanceof ImmutableName) {
        // used by CollectionPropertyEditor and StringMapPropertyEditor.viewElement()
        selectedObject = BeamerManager.getBaseObject((ImmutableName)_toView, propertyType_);
    }
    else if (_toView instanceof BaseObject) {
        selectedObject = (BaseObject) _toView;
    }
    else if (_toView instanceof String) {
        // only from CollectionPropertyEditor, for a DetailledName.
        name = getNameFromDetailledName((String)_toView);

                // for testing and tracking
                //System.out.println("SimplePropertyEditor.viewOneElement() name for the detailled name: "+
                  //      name);

        selectedObject = BeamerManager.getBaseObject(name, propertyType_);
    }
    else return;

                // for testing and tracking
                //System.out.println("SimplePropertyEditor.viewOneElement() selectedObject: "+
                  //      selectedObject);

    // finalizes this instance if it is in the beamer
    int result = TreeManager.selectNodeInExplorer(selectedObject);
    if (result == -1)
            DialogBox.showErrorMessage("Sorry, the object "+_toView+
                    " is not found in the Explorer. It should be closed.");

    selectedObject = null;
    name = null;
    acs = null;
  } // viewOneElement()

  /**
   * Some constructor/add/copy arguments may be set as mandatory by the BeanInfo,
   * and this method returns an array of these known values, at the right position
   * for the arguments array to use.  <P>
   * In the property descritor, 'withKnownArgumentsForNew' allows to define
   * the constructor arguments which has to be fixed
   * - with 'choiceMethodForNew' or not, add method or not -
   * when a new instance is created through the 'New' button. <P>
   * First case for the fixed argument: the string 'this' is used
   * to point to the object in the beamer;
   * but this object is NOT get as argument if the 'this' index in the array
   * is after the last specified argument for the constructor or the add method.
   * Second case: search the right property descriptor in the beamer,
   * and call the read method to initialize a l_fixedValue from this object.
   * Third case for the fixed argument: the string 'thisname' is used
   * to point to the value returned from getName() by the object in the beamer.
   * Fourth case: the string 'true' or 'false' is used to define a boolean value.
   * Fifth case, if none of the previous cases matches: takes directly
   * the fixed argument if it is a String. <P>
   * When the value of an argument is not determined,
   * the returned array sets 'null'.
   * The returned array may be shorter than the argument arrays for the new object.
   * Called by actionPerformed() and CollectionPropertyEditor.addElements().
   * @return array of known values for the arguments of the new element, in the
   *   right order for their use. May be equal to new Object[1].
   */
  protected Object[] getKnownValuesForNewElement() {

    Object[] l_fixedValue = new Object[1];

    // example for the "withKnownArgumentsForNew" value:
    // pd_one.setValue("withKnownArgumentsForNew", new Object[] {null, "ACS", "ParentNode"} );

    if (propertyDescriptor_.getValue("withKnownArgumentsForNew") != null) {
      Object[] l_fixedProperty = (Object[]) propertyDescriptor_.getValue(
        "withKnownArgumentsForNew");
      l_fixedValue = new Object[l_fixedProperty.length];
      PropertyDescriptor[]  l_propertyDescriptor =
            UtilityImpl.getFirstInstance().
                selectBeanInfo(source_).getPropertyDescriptors();
      try {
        for (int i = 0; i < l_fixedProperty.length; i++) {
            if (l_fixedProperty[i] == null) continue;

            // first case: the object in the beamer
            if (((String)l_fixedProperty[i]).equals("this")) {
              l_fixedValue[i] = source_;
            }
            // third case: the name of the object in the beamer
            else if (((String)l_fixedProperty[i]).equals("thisname")) {
              l_fixedValue[i] = source_.getName();
            }
            // fourth case: a boolean
            else if (((String)l_fixedProperty[i]).equals("true") ) {
              l_fixedValue[i] = Boolean.TRUE;
            }
            // fourth case: a boolean
            else if (((String)l_fixedProperty[i]).equals("false")) {
              l_fixedValue[i] = Boolean.FALSE;
            }
            else {
                for (int j = 0; j < l_propertyDescriptor.length; j++) {
                  // second case: searches for the right property descriptor of the source_
                  // and calls the read method
                  // to initialize a l_fixedValue from the source_
                  if (l_propertyDescriptor[j].getName().equals((String)l_fixedProperty[i])) {
                    l_fixedValue[i] = l_propertyDescriptor[j].getReadMethod().
                      invoke(source_, (Object[])null);
                    break;
                  } // if
                } // for j
            } // else

            if (l_fixedValue[i] != null)
                continue;
            // fifth case: takes directly the String value
            else if (l_fixedProperty[i] instanceof String)
                l_fixedValue[i] = l_fixedProperty[i];
        } // for i
      } catch (Exception e) {
        throw new InternalError("In SimplePropertyEditor.getKnownValuesForNewElement(), "+
          "for 'withKnownArgumentsForNew': "+e.toString());
      } // catch
      l_fixedProperty = null;
      l_propertyDescriptor = null;
    } // if "withKnownArgumentsForNew"

    return l_fixedValue;
  } // getKnownValuesForNewElement()

  /**
   * Called by RightMapPropertyEditor, CollectionPropertyEditor,
   * StringMapPropertyEditor, AssociationMapPropertyEditor,
   * to set the dimension of the list. The tresholds for the height are 3, 7, 15,
   * and "limitedSizeForPresentation" from the property descriptor, if this value
   * is greater than 4 and if _listSize is not lesser than 4.
   * @param _listSize height of the list in _component
   * @param _component is the scrollpane which contains the JList to display
   * @param _largeWidth true to have a width equal to the beamer width,
   *          as for CollectionPropertyEditor
   * @param _buttonPanel true if there is a vertical panel for buttons,
   *          for which the width is from CommonPropertyEditor.BUTTON_DIMENSION_2.
   * @param _namePanel false if the name of the property is not displayed,
   *          and the panel is then larger.
   */
  protected void setListPanelDimension(int _listSize, JScrollPane _component,
          boolean _largeWidth, boolean _buttonPanel, boolean _namePanel) {

    int width = 0;
    if (_buttonPanel) {
        if (_largeWidth && _namePanel) {

                // for testing and tracking
                //System.out.println("SimplePropertyEditor.setListPanelDimension() "+
                  //      "_namePanel: "+_namePanel);

            // for CollectionPropertyEditor with names and buttons
            width = LIST_WIDTH_DIMENSION*97/100;
        }
        else if (_largeWidth && !_namePanel) {

                // for testing and tracking
                //System.out.println("SimplePropertyEditor.setListPanelDimension() "+
                  //      "_namePanel: "+_namePanel);

            // true, true , false
            // StringMapPropertyEditor for values with and without buttons
            // AssociationMapPropertyEditor for keys without buttons
            // RightMapPropertyEditor for values with buttons
            width = LIST_WIDTH_DIMENSION*63/100; // 3/5
        }
        else if (!_largeWidth && _namePanel) {

            // StringMapPropertyEditor for keys with buttons
            // AssociationMapPropertyEditor for keys with buttons
            width = LIST_WIDTH_DIMENSION*7/10; // 2/5
        }
        else {
            // RightMapPropertyEditor for keys with buttons
            // AssociationMapPropertyEditor for values with buttons
            width = LIST_WIDTH_DIMENSION*7/10;
                //- CommonPropertyEditor.BUTTON_DIMENSION_2.width;
        }
    }
    else {
        if (_largeWidth) {
            if (_namePanel) {

                // CollectionPropertyEditor without buttons and with display name
                width = CommonPropertyEditor.BUTTON_DIMENSION_2.width*3/5+
                    LIST_WIDTH_DIMENSION*9/10;
            }
            else {
                // true, false, false
                // CollectionPropertyEditor without buttons nor display name
                width = 2*(CommonPropertyEditor.BUTTON_DIMENSION_2.width +
                        LIST_WIDTH_DIMENSION*3/5);
            }
        } // else
        else {
            if (_namePanel) {

                        // for testing and tracking
                        //System.out.println("SimplePropertyEditor.setListPanelDimension() "+
                          //      "with the arguments 'false, false, true' for "+
                            //    propertyDescriptor_.getDisplayName());

                // false, false, true
                // CollectionPropertyEditor with buttons and without name
                width = 345;
                        //4*CommonPropertyEditor.BUTTON_DIMENSION_2.width +
                        //NAME_WIDTH_DIMENSION_2;
                        //2*CommonPropertyEditor.BUTTON_DIMENSION_2.width +
                        //LIST_WIDTH_DIMENSION;

            }
            // false, false, false
            // AssociationMapPropertyEditor for values without buttons
            // StringMapPropertyEditor for keys without buttons
            // RightMapPropertyEditor for keys without buttons
            else width = CommonPropertyEditor.BUTTON_DIMENSION_2.width+
                    LIST_WIDTH_DIMENSION*13/20;
        } // else
    } // else

    Integer limitedSize = (Integer)propertyDescriptor_.getValue("limitedSizeForPresentation");

    if (_listSize < 3) {
            _component.setMaximumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION/3));
            _component.setMinimumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION/3));
            _component.setPreferredSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION/3));
    } else if (limitedSize != null && limitedSize.intValue() > 8 && _listSize < 6) {
        // for limitedSize == 3, the height is almost like for the case '_listSize == 2'
            _component.setMaximumSize(
              new Dimension(width, limitedSize.intValue()*LIST_HEIGHT_DIMENSION/20));
            _component.setMinimumSize(
              new Dimension(width, limitedSize.intValue()*LIST_HEIGHT_DIMENSION/20));
            _component.setPreferredSize(
              new Dimension(width, limitedSize.intValue()*LIST_HEIGHT_DIMENSION/20));
    } else if (limitedSize != null && limitedSize.intValue() < 9) {
        // for limitedSize == 3, the height is almost like for the case '_listSize == 2'
            _component.setMaximumSize(
              new Dimension(width, limitedSize.intValue()*LIST_HEIGHT_DIMENSION/10));
            _component.setMinimumSize(
              new Dimension(width, limitedSize.intValue()*LIST_HEIGHT_DIMENSION/10));
            _component.setPreferredSize(
              new Dimension(width, limitedSize.intValue()*LIST_HEIGHT_DIMENSION/10));
    } else if (_listSize < 7 && limitedSize != null && limitedSize.intValue() < 11) {
            _component.setMaximumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION/2));
            _component.setMinimumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION/2));
            _component.setPreferredSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION/2));
    } else if (_listSize > 15  && limitedSize != null && limitedSize.intValue() > 16) {


                        // for testing and tracking
                        //System.out.println("SimplePropertyEditor.setListPanelDimension() for "+
                          //      propertyDescriptor_.getName()+" and limitedSize = "+limitedSize+
                            //    " with _listSize > 15  && limitedSize != null && limitedSize.intValue() > 16");

            _component.setMaximumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION*4));
            _component.setMinimumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION*4));
            _component.setPreferredSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION*4));
    } else if (_listSize > 15  && limitedSize != null && limitedSize.intValue() > 10) {


                        // for testing and tracking
                        //System.out.println("SimplePropertyEditor.setListPanelDimension() for "+
                          //      propertyDescriptor_.getName()+" and limitedSize = "+limitedSize+
                            //    " with _listSize > 15  && limitedSize != null && limitedSize.intValue() > 10");

            _component.setMaximumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION*2));
            _component.setMinimumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION*2));
            _component.setPreferredSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION*2));
    } else {

                        // for testing and tracking
                        //System.out.println("SimplePropertyEditor.setListPanelDimension() for "+
                          //      propertyDescriptor_.getName()+
                            //    " with the default case");


            _component.setMaximumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION));
            _component.setMinimumSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION));
            _component.setPreferredSize(
              new Dimension(width, LIST_HEIGHT_DIMENSION));
    } // else

    limitedSize = null;
  } // setListPanelDimension()  

  /**
   * Uses the property source_ to get the name of an ACSObject from its detailled name.
   * Only an AclEntry or a Privilege has a DetailledName.
   * An external AclEntry or Linked Privilege detailled name starts with the name of its ACS
   * ('ACS: ACS_name |'). For an internal AclEntry or Privilege, the ACS of the source is used.
   * Called by viewOneElement() and CollectionPropertyEditor.
   * @see ARoad0.gBaseInterface.DetailledName
   * @param _detailledName
   * @return the name
   */
  protected ImmutableName getNameFromDetailledName(String _detailledName) {
        // a detailled name contains at least one sequence '||'.
        if (!_detailledName.contains("||"))
            throw new InternalError("SimplePropertyEditor.viewOneElement() cannot handle "+
                    "a wrong detailled name: "+_detailledName);

        ImmutableName   result = null;
        ImmutableACS    acs = null;

        if (source_ instanceof ImmutableACS)
            acs = (ImmutableACS)source_;
        else if (source_ instanceof ACSObject)
            acs = ((ACSObject)source_).getEorACS();
        else
            throw new InternalError("SimplePropertyEditor.getNameFromDetailledName() cannot find "+
                    "the ACS for "+_detailledName);

        try {
          // _detailledName has to be an AclEntry or Privilege detailled name,
            // and the DetailledName rule is applied there to tell what class to call
          if (!_detailledName.contains(">>") ||
                (_detailledName.indexOf(">>") > _detailledName.lastIndexOf("||"))) {
            if (!_detailledName.startsWith("ACS: "))
                result = PrivilegeAbst.EMPTY_INSTANCE.getNameFromDetailledName(
                  acs, _detailledName);
            else {
                acs = (ImmutableACS) ACS_Facade.getEorBaseObject(
                        NameImpl.getNameFromString(
                            _detailledName.substring(4, _detailledName.indexOf("|"))));
                result = PrivilegeAbst.EMPTY_INSTANCE.getNameFromDetailledName(
                  acs, _detailledName);
            } // else
          }
          else {
            if (!_detailledName.startsWith("ACS: "))
                result = AclEntryImpl.EMPTY_INSTANCE.getNameFromDetailledName(
                  acs, _detailledName);
            else {
                acs = (ImmutableACS) ACS_Facade.getEorBaseObject(
                        NameImpl.getNameFromString(
                            _detailledName.substring(4, _detailledName.indexOf("|"))));
                result = AclEntryImpl.EMPTY_INSTANCE.getNameFromDetailledName(
                  acs, _detailledName);
            } // else
          } // else

        } catch (BaseError b) {
          throw new InternalError("SimplePropertyEditor.getNameFromDetailledName() cannot "+
            "parse the detailled name of an AclEntry or Privilege.");
        }
        acs = null;
        return result;
  } // getNameFromDetailledName()
  
  // -------------------- PRIVATE METHODS -----------------------------------

  /**
   * Gets the boolean editor with the property name.
   * The editor is a JComboBox defined with getTags(), if the property is editable,
   * or a JCheckBox if there is no tags, and else, the editor is a JTextField. 
   * If it is a bound property, the editor is updated through propertyChange().
   * The size of the panel is shorted than the other returned panels.
   * Called by getCustomEditor() when getGeneralEditor() does not match.
   * @return the panel for editing a boolean value.
   */
  private JPanel getBooleanEditor() {
    JPanel      result = new JPanel();
    JLabel      nameLabel = new JLabel(displayName_);
    JComponent              box = null;
    DefaultListCellRenderer renderer;

    // editable JComboBox or JCheckBox, and listen user actions,
    // only if there is a write method
    if (writeMethod_ != null) {

        if (getTags() != null && getTags().length != 0) {
            box = new JComboBox(getTags());
            renderer = new DefaultListCellRenderer();
            //renderer.setHorizontalAlignment(SwingConstants.RIGHT);
            //((JComboBox)box).setRenderer(renderer);
            ((JComboBox)box).getModel().setSelectedItem(getAsText() );
            //((JComboBox)box).setEditable(false);
            ((JComboBox)box).addActionListener(this);
            box.setBackground(Color.white);
            renderer = null;

            nameLabel.setMaximumSize(NAME_DIMENSION);
            nameLabel.setMinimumSize(NAME_DIMENSION);
            nameLabel.setPreferredSize(NAME_DIMENSION);
            box.setMaximumSize(COMBOBOX_FIELD_DIMENSION);
            box.setMinimumSize(COMBOBOX_FIELD_DIMENSION);
            box.setPreferredSize(COMBOBOX_FIELD_DIMENSION);
        } // if
        else { // use  a simple checkbox

            box = new JCheckBox();
            // see java-6-sun/docs/api/javax/swing/Action.html#buttonActions
            ((JCheckBox)box).setActionCommand(propertyDescriptor_.getShortDescription()); 
            ((JCheckBox)box).setAlignmentX( JCheckBox.RIGHT_ALIGNMENT );
            ((JCheckBox)box).setSelected( ((Boolean)getValue()).booleanValue() );
            ((JCheckBox)box).addActionListener(this);
            ((JCheckBox)box).setBackground(Color.white);

            nameLabel.setMaximumSize(BOOLEAN_NAME_DIMENSION);
            nameLabel.setMinimumSize(BOOLEAN_NAME_DIMENSION);
            nameLabel.setPreferredSize(BOOLEAN_NAME_DIMENSION);
            box.setMaximumSize(BOOLEAN_FIELD_DIMENSION);
            box.setMinimumSize(BOOLEAN_FIELD_DIMENSION);
            box.setPreferredSize(BOOLEAN_FIELD_DIMENSION);
        } // else

    } else {

        box = new JTextField(getAsText());
        ((JTextField)box).setEditable(false);
        box.setBackground(InternalFrame.NON_EDITABLECOLOR);
        ((JTextField)box).setHorizontalAlignment(JTextField.RIGHT); // fire an event

        nameLabel.setMaximumSize(BOOLEAN_NAME_DIMENSION);
        nameLabel.setMinimumSize(BOOLEAN_NAME_DIMENSION);
        nameLabel.setPreferredSize(BOOLEAN_NAME_DIMENSION);
        box.setMaximumSize(BOOLEAN_FIELD_DIMENSION);
        box.setMinimumSize(BOOLEAN_FIELD_DIMENSION);
        box.setPreferredSize(BOOLEAN_FIELD_DIMENSION);
    } // else

              // for testing and tracking
              //System.out.println("From SimplePropertyEditor.getBooleanEditor(), "+
              //  "for "+propertyDescriptor_.getDisplayName()+
              //  ", the box width is: "+box.getWidth());

    nameLabel.setFont(BeamerManager.plainFont);
    nameLabel.setToolTipText(propertyDescriptor_.getShortDescription());
    box.setFont(BeamerManager.plainFont);

    if (propertyDescriptor_.isBound()) boundComponent_ = box;

    nameLabel.setLabelFor(box);
    result.add(nameLabel);
    result.add(box);

    result.setBackground(Color.white);

    renderer = null;
    box = null;
    nameLabel = null;
    return result;
  } // getBooleanEditor()

  /**
   * Gets the editor with the property name
   * when the property type is not boolean: int, String, ImmutableName or BaseObject.
   * Calls actionPerformed() when the button "View", "Select" or "Clear" is actionned,
   * if the property type is a BaseObject.
   * Called by getCustomEditor() when getBooleanEditor() does not match.
   * @param _displayName of the property
   * @param _text is the current value of the property, in a textual form
   * @param _withWriteMethod true if the property may be updated through this editor
   * @param _isBound true if the property may change (even if _withWriteMethod
   *  is false)
   * @param _isBaseObject true if the property type is a BaseObject. The 'View' button
   *  is then displayed.
   * @return the JPanel for editing the property.
   */
  private JPanel getGeneralEditor(
    String _displayName,
    String _text,
    boolean _withWriteMethod,
    boolean _isBound,
    boolean _isBaseObject) {

    JPanel result = new JPanel();
    JLabel nameLabel = new JLabel(_displayName);
    JComboBox box = null; // for String restricted values
    // an array of 9 booleans: withPreviousButton at 0, withViewButton,
    // withSelectButton, withNewButton, withDeleteButton, withRemoveButton,
    // withClearButton at 6 (others are not for this editor)
    boolean[] l_buttons = getButtonsDisplaying();

    nameLabel.setMaximumSize(NAME_DIMENSION);
    nameLabel.setMinimumSize(NAME_DIMENSION);
    nameLabel.setPreferredSize(NAME_DIMENSION);
    nameLabel.setFont(BeamerManager.plainFont);
    nameLabel.setToolTipText(propertyDescriptor_.getShortDescription());

    JTextField field = new JTextField();
    field.setFont(BeamerManager.plainFont);
    JButton selectButton;
    JButton clearButton;
    JButton viewButton;
    field.setHorizontalAlignment(JTextField.RIGHT); // fire an event
    field.setText(_text);

    nameLabel.setLabelFor(field);
    result.add(nameLabel);
    result.setBackground(Color.white);

    if (!_withWriteMethod) {
      result.add(field);
      field.setEditable(false);
      field.setBackground(InternalFrame.NON_EDITABLECOLOR);

      if (_isBaseObject ||
        (ImmutableName.class.isAssignableFrom(propertyType_) && l_buttons[1])) {
          viewButton = new JButton(" See ");
          viewButton.setFont(BeamerManager.plainButtonFont);
          viewButton.setBackground(Color.white);
          viewButton.setBorder(new BevelBorder(
            BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
          viewButton.addActionListener(this);
          viewButton.setMaximumSize(BUTTON_DIMENSION);
          viewButton.setMinimumSize(BUTTON_DIMENSION);
          viewButton.setPreferredSize(BUTTON_DIMENSION);
          result.add( viewButton );

          field.setMaximumSize(ONE_BUTTON_FIELD_DIMENSION);
          field.setMinimumSize(ONE_BUTTON_FIELD_DIMENSION);
          field.setPreferredSize(ONE_BUTTON_FIELD_DIMENSION);
      }
      else if (ImmutableName.class.isAssignableFrom(propertyType_) ) {
        field.setMaximumSize(NO_BUTTON_FIELD_DIMENSION);
        field.setMinimumSize(NO_BUTTON_FIELD_DIMENSION);
        field.setPreferredSize(NO_BUTTON_FIELD_DIMENSION);
      }
      else if (propertyType_ == Integer.TYPE) {
        field.setMaximumSize(INT_FIELD_DIMENSION);
        field.setMinimumSize(INT_FIELD_DIMENSION);
        field.setPreferredSize(INT_FIELD_DIMENSION);

        //JLabel      emptyLabel = new JLabel();
        //emptyLabel.setMaximumSize(INT_EMPTY_FIELD_DIMENSION);
        //emptyLabel.setMinimumSize(INT_EMPTY_FIELD_DIMENSION);
        //emptyLabel.setPreferredSize(INT_EMPTY_FIELD_DIMENSION);

        //result.add(emptyLabel);
        //emptyLabel = null;
      }
      else if (String.class.isAssignableFrom(propertyType_)) {
        field.setMaximumSize(NO_BUTTON_FIELD_DIMENSION);
        field.setMinimumSize(NO_BUTTON_FIELD_DIMENSION);
        field.setPreferredSize(NO_BUTTON_FIELD_DIMENSION);
      }
      // for name property
      else if (ImmutableName.class.isAssignableFrom(propertyType_)) {
        field.setMaximumSize(NO_BUTTON_FIELD_DIMENSION);
        field.setMinimumSize(NO_BUTTON_FIELD_DIMENSION);
        field.setPreferredSize(NO_BUTTON_FIELD_DIMENSION);            
      }
    } // if
    else {
      //there is a write method, and a field or box is used

      // for BaseObject property
      if (_isBaseObject) {
            result.add(field);
            field.setEditable(false);
            field.setBackground(InternalFrame.NON_EDITABLECOLOR);

            // creates two buttons 'Select' and 'See'
            // with this instance as listener
            if (l_buttons[1]) {
                viewButton = new JButton(" See ");
                viewButton.setBackground(Color.white);
                viewButton.setBorder(new BevelBorder(
                  BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
                viewButton.setFont(BeamerManager.plainButtonFont);
                viewButton.addActionListener(this);
                viewButton.setMaximumSize(BUTTON_DIMENSION);
                viewButton.setMinimumSize(BUTTON_DIMENSION);
                viewButton.setPreferredSize(BUTTON_DIMENSION);
                result.add( viewButton );
            } // if View
            
            if (l_buttons[2]) {
                selectButton = new JButton("Select");
                selectButton.setBackground(Color.white);
                selectButton.setBorder(new BevelBorder(
                  BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
                selectButton.setFont(BeamerManager.plainButtonFont);
                selectButton.addActionListener(this);
                selectButton.setMaximumSize(BUTTON_DIMENSION);
                selectButton.setMinimumSize(BUTTON_DIMENSION);
                selectButton.setPreferredSize(BUTTON_DIMENSION);
                result.add( selectButton );
            }
            
            // creates a button 'Clear'
            // with this instance as listener
            if (l_buttons[6]) {
              clearButton = new JButton(" Clear");
              clearButton.setBackground(Color.white);
              clearButton.setBorder(new BevelBorder(
                BevelBorder.RAISED, Color.white, InternalFrame.FORECOLOR));
              clearButton.setFont(BeamerManager.plainButtonFont);
              clearButton.addActionListener(this);
              clearButton.setMaximumSize(BUTTON_DIMENSION);
              clearButton.setMinimumSize(BUTTON_DIMENSION);
              clearButton.setPreferredSize(BUTTON_DIMENSION);
              result.add( clearButton );

              field.setMaximumSize(THREE_BUTTONS_FIELD_DIMENSION);
              field.setMinimumSize(THREE_BUTTONS_FIELD_DIMENSION);
              field.setPreferredSize(THREE_BUTTONS_FIELD_DIMENSION);
            }
            else {
              field.setMaximumSize(TWO_BUTTONS_FIELD_DIMENSION);
              field.setMinimumSize(TWO_BUTTONS_FIELD_DIMENSION);
              field.setPreferredSize(TWO_BUTTONS_FIELD_DIMENSION);
            } // if (!_withClearButton) {


      } // if (_isBaseObject)
      else if (propertyType_ == Integer.TYPE) {
        result.add(field);
        field.setEditable(true);
        field.addActionListener(this);
        field.setBackground(Color.white);

        field.setMaximumSize(INT_FIELD_DIMENSION);
        field.setMinimumSize(INT_FIELD_DIMENSION);
        field.setPreferredSize(INT_FIELD_DIMENSION);

        JLabel      emptyLabel = new JLabel();
        emptyLabel.setMaximumSize(INT_EMPTY_FIELD_DIMENSION);
        emptyLabel.setMinimumSize(INT_EMPTY_FIELD_DIMENSION);
        emptyLabel.setPreferredSize(INT_EMPTY_FIELD_DIMENSION);

        result.add(emptyLabel);
        emptyLabel = null;
      }
      // for String property
      else if (String.class.isAssignableFrom(propertyType_)) {
        ACS acs = null;
        if (source_ instanceof ACSObject)
            acs = (ACS)ACS_Facade.getOpenACS().get(
              ((ACSObject)source_).getAcsName() );

                    // for testing and tracking
                    //System.out.println("Call to getRestrictedValues(acs) in "+
                      //"SimplePropertyEditor.getGeneralEditor(). " );
        
        String[] l_restrictedValues = getRestrictedValues(acs);


                    // for testing and tracking
                    //System.out.println("After the call to getRestrictedValues(acs) in "+
                      //"SimplePropertyEditor.getGeneralEditor(). " );
        
        // if there are restrictions on values, uses a JComboBox
        if (l_restrictedValues != null ) {
            box = new JComboBox(l_restrictedValues);
            box.setFont(BeamerManager.plainFont);
            //DefaultListCellRenderer renderer = new DefaultListCellRenderer();
            //renderer.setHorizontalAlignment(SwingConstants.RIGHT);
            //box.setRenderer(renderer);
            box.getModel().setSelectedItem(_text ); // must be among l_restrictedValues
            
            box.setEditable(true);
            box.addActionListener(this);
            box.setBackground(Color.white);
            box.setMaximumSize(NO_BUTTON_FIELD_DIMENSION);
            box.setMinimumSize(NO_BUTTON_FIELD_DIMENSION);
            box.setPreferredSize(NO_BUTTON_FIELD_DIMENSION);
            
            result.add(box);
            acs = null;
            l_restrictedValues = null;
            //renderer = null;
        } else {
            result.add(field);
            field.setEditable(true);
            field.addActionListener(this);
            field.setBackground(Color.white);
            field.setMaximumSize(NO_BUTTON_FIELD_DIMENSION);
            field.setMinimumSize(NO_BUTTON_FIELD_DIMENSION);
            field.setPreferredSize(NO_BUTTON_FIELD_DIMENSION);
        } // else
      } // else if for String

      // for Name property
      else if (ImmutableName.class.isAssignableFrom(propertyType_)) {
        ACS acs = null;
        if (source_ instanceof ACSObject)
            acs = (ACS)ACS_Facade.getOpenACS().get(
              ((ACSObject)source_).getAcsName() );

                    // for testing and tracking
                    //System.out.println("Call to getRestrictedValues(acs) in "+
                      //"SimplePropertyEditor.getGeneralEditor(). " );

        // returns the String format of names
        String[] l_restrictedValues = getRestrictedValues(acs);


                    // for testing and tracking
                    //System.out.println("After the call to getRestrictedValues(acs) in "+
                      //"SimplePropertyEditor.getGeneralEditor(). " );

        // if there are restrictions on values, uses a JComboBox
        if (l_restrictedValues != null ) {
            box = new JComboBox(l_restrictedValues);
            box.setFont(BeamerManager.plainFont);
            //DefaultListCellRenderer renderer = new DefaultListCellRenderer();
            //renderer.setHorizontalAlignment(SwingConstants.RIGHT);
            //box.setRenderer(renderer);
            box.getModel().setSelectedItem(_text ); // must be among l_restrictedValues

            box.setEditable(true);
            box.addActionListener(this);
            box.setBackground(Color.white);
            box.setMaximumSize(NO_BUTTON_FIELD_DIMENSION);
            box.setMinimumSize(NO_BUTTON_FIELD_DIMENSION);
            box.setPreferredSize(NO_BUTTON_FIELD_DIMENSION);

            result.add(box);
            acs = null;
            l_restrictedValues = null;
            //renderer = null;
        } else {
            result.add(field);
            field.setEditable(true);
            field.addActionListener(this);
            field.setBackground(Color.white);
            field.setMaximumSize(NO_BUTTON_FIELD_DIMENSION);
            field.setMinimumSize(NO_BUTTON_FIELD_DIMENSION);
            field.setPreferredSize(NO_BUTTON_FIELD_DIMENSION);
        } // else
      } // else if for Name
    } // if (writeMethod_ != null)

    if (_isBound) {
        if (box == null) boundComponent_ = field;
        else boundComponent_ = box;
    }

        // for testing and tracking
        //System.out.println("The text length in "+
        //  "SimplePropertyEditor.getGeneralEditor() is: "+
        //  _text.length() );
    
    nameLabel = null;
    selectButton = null;
    clearButton = null;
    viewButton = null;
    field = null;
    box = null;

    return result;
  } // getGeneralEditor()


  /**
   * Reports that we have been modified to any interested listeners - NOT USED SO FAR.
   */
  private void firePropertyChange() {
    java.util.Vector targets;
    synchronized (this) {
        if (listeners_ == null) {
          return;
        }
        targets = (java.util.Vector) listeners_.clone();
    }
    // Tell our listeners that "everything" has changed.
    PropertyChangeEvent evt = new PropertyChangeEvent(source_, null, null, null);

    for (int i = 0; i < targets.size(); i++) {
        PropertyChangeListener target =
          (PropertyChangeListener)targets.elementAt(i);
        target.propertyChange(evt);
    }
  } // firePropertyChange()

  /**
   * Resets the update when the user enter is not correct,
   * and display an error message.
   * @param _text message that explains the error to the user
   */
  private void resetUpdate(String _text) {
    if (boundComponent_ instanceof JTextField) {
      ((JTextField)boundComponent_).removeActionListener(this);

      if (getValue() != null && ((BaseObject)getValue()).getName() != null)
        ((JTextField)boundComponent_).setText(
          ((BaseObject)getValue()).getName().toString() );
      else ((JTextField)boundComponent_).setText("");

      //((JTextField)boundComponent_).repaint();
      ((JTextField)boundComponent_).addActionListener(this);
    } // if
    DialogBox.showErrorMessage("Error in the update of "+displayName_+" for "+_text);
    return;
  }  // resetUpdate()

  /**
   * Called by actionPerformed().
   * The class name must not contain "Immutable" twice, or at the end of the name.
   * @return the BeanDescriptor of the property
   */
  /*
  private BeanDescriptor getBeanDescriptor() {

    // initialization
    Class                 beanInfoClass;
    BeanDescriptor        result;
    String                matchedPropertyName = propertyType_.getName();
    if (matchedPropertyName.indexOf("Immutable") != -1) {
      // remove 'Immutable' in the name
      StringBuilder str = new StringBuilder();
      str.append(matchedPropertyName.substring(
        0, matchedPropertyName.indexOf("Immutable") ));
      matchedPropertyName = str.append(matchedPropertyName.substring(
        matchedPropertyName.indexOf("Immutable")+9, matchedPropertyName.length() )).
        toString();
      str = null;
    }
    if (!matchedPropertyName.endsWith("Impl")) {
      // add 'Impl' to the name
      matchedPropertyName = matchedPropertyName + "Impl";
    }

    try {
      beanInfoClass = Class.forName(matchedPropertyName+"BeanInfo");
      result =
        ((BeanInfo) beanInfoClass.newInstance()).getBeanDescriptor();
    } catch (Exception l) {
      throw new InternalError("from SimplePropertyEditor.getBeanDescriptor(): "+
        l.toString());
    }
    return result;
  }  // getBeanDescriptor()
   * */

  
} // SimplePropertyEditor
