//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui1;

import java.beans.*;
import javax.swing.*;
import javax.swing.border.EtchedBorder;
import javax.swing.border.LineBorder;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Array;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.util.Map;
import java.util.List;
import java.util.Iterator;
import java.util.Collection;

import ARoad0.ErrW.ProcessError;
import ARoad0.ErrW.SmallError;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.gWork.ACS_Facade;

import ARoad0.gBase.UtilityImpl;
import ARoad0.gBase.NameImpl;

import ARoad0.gBaseInterface.BaseObject;
import ARoad0.gBaseInterface.FinalizedObjectForUser;
import ARoad0.gBaseInterface.ACSObject;
import ARoad0.gBaseInterface.ACS;
import ARoad0.gBaseInterface.ImmutableName;
        
/**
 * This class is responsible for displaying, by the mean of panels,
 * a property map of a gBase BaseObject, for which the type is a map
 * where keys are Strings and the values are only String arrays. <P>
 * The map updating is never done through a write method or a 'choiceMethodForNew'
 * in the property descriptor, but only through 'add/remove' methods set in the property
 * descriptor. On the other hand, 'choiceMethodForNew' is used in 
 * super.getRestrictedValues() to get the constrained keys, if any.<P>
 * The keys updating may be forbidden or constrained, and if this updating is
 * allowed, the deleting of a key is always through the call of the removing method 
 * with a null value for each key to delete. The maximal number of constrained keys is 84. <P>
 * The values updating may be constrained by a specific method which declares
 * the allowed values. A maximal value may be defined for the number of map keys,
 * through the property 'MaxSize'. Through the property 'MaxValueSize',
 * a maximal value may be defined for the number of values in the map.
 * The default maximal size for the keys set is 50, and 100 for the values set. <P>
 * The updating of the keys and the values handles all the exceptions in the same way,
 * like the superclass. A message is displayed to the user from the exception, and
 * the map is reset through the call to the current value. So, the called gBase class
 * may set in the updating methods how the exception impacts the current value.
 * This class complements CollectionPropertyEditor, which handles more complex types
 * but manages maps only as collections of values. <P>
 * StringTwoKeysMapPropertyEditor is a subclass for handling two-parts, structured keys.
 * AssociationMapPropertyEditor is its subclass and it creates associations
 * to use as scripts calling the constuctors and methods in any gBase class. <P>
 * BaseObjectMapPropertyEditor is a subclass for BaseObjects or ImmutableNames 
 * in the values. RightMapPropertyEditor is a subclass for StringRights
 * in the values. <P> 
 * The getCustomEditor() method is the main method.
 * This class is specified by some BeanInfo classes in gBase, for instance 
 * in their property descriptors for ACSFactory.SpecializedTypesAndExplorerSet,
 * ACSImpl.m_TypesAndGUIPolicy and M_ContainerPolicy.
 * This class is called by Gui1.BeamerManager.addPropertyEditor(). <P>
 * The unique property type is String, both for the keys and the values, but some
 * methods are more general and they are used by the subclasses.
 * This class is a PropertyEditor that gets, and listens the property
 * in the relevant gBase instance, and edits the keys and the values.
 * It is also a custom editor that provides getCustomEditor(),
 * and specifies the prefered size of the panel to be displayed.
 * This instance listens actions from gBase objects when the program updates the property.
 * So, the GUI always DISPLAYS THE RIGHT DATA which is in the base.
 * When a string is entered in a JTextField, its length is limited to 200,
 * to prevent any unauthorized input. A space is always added to the end of the string. <P>
 * As a PropertyEditor, it must be able to register PropertyChangeListeners
 * but this is not a feature to use so far. It uses only registration
 * on gBase BoundBeans, that handle PropertyChangeListeners. In the same way,
 * does not use the ActionEvents handling - add/removeActionListener() -
 * in the JComponent property.
 * An instance is constructed through the factory method valueOf() which calls
 * the super valueOf() method, and it is possible to get a reusable empty instance.
 * java.beans.PropertyEditorManager is not used.
 * setValue(), setAsText() and getAsText() are not operational.
 * @version 0.7.0
 */
public class StringMapPropertyEditor extends SimplePropertyEditor
  implements ListSelectionListener {

  /**
   * When the property type is a map (Map<String, String[]> there),
   * propertyTypeInGroup_ is the type of the elements in the arrays which are the map values.
   */
  protected Class                       propertyTypeInGroup_ = null;
  /**
   * When the property type is a map (Map<String, String[]> there),
   * selectedTypeInGroup_ is the type of the elements to select in the explorer.
   * Generally speaking, it is propertyTypeInGroup_, except for BaseObjectPropertyEditor
   * when a BaseObject is selected to put its name in an array map value. This
   * property is then equal to ImmutableName.
   */
  protected Class                       selectedTypeInGroup_ = null;
  /**
   * The ListModel of Strings, in the JList which displays the keys
   */
  protected DefaultListModel            keysModel_;

  /**
   * The ListModel of Strings, in the JList which displays the values
   */
  protected DefaultListModel            valuesModel_;

  /**
   * The values list that displays a bound property.
   * Used as boundComponent_ which is the keys list.
   */
  protected JComponent                  boundComponentTwo_;

  /**
   * The scrollpane that displays the keys list.
   */
  protected JScrollPane                 keyScrollPane_;
  
    /**
   * The scrollpane that displays the values list.
   */
  protected JScrollPane                 valueScrollPane_;  
  
  /**
   * The add method if it exists, for the property.
   * If it is non null, the writeMethod in 'super' may be still non null.
   */
  protected Method                      addMethod_;

  /**
   * The remove method if it exists, for the property
   * If it is non null, the writeMethod in 'super' may be still non null.
   */
  protected Method                      removeMethod_;  

  /**
   * The limited values method if it exists, for the property.
   */
  protected Method                      limitedValuesMethod_;  
  
  /**
   * True if the key list may be updated. If it is false,
   * and if there is some updating methods, only the value list in the map
   * may be updated for any key.
   */
  protected boolean                     keysUpdateInMap_ = false; 
  
  /**
   * The maximal size which is allowed for the keys set.
   */
  protected int                         maxKeySize_;

  /**
   * The maximal size which is allowed for the values set.
   */
  protected int                         maxValueSize_;
  
  //-----------------------------------------------------------------------

  /**
   * Constructor FOR USE ONLY through the valueOf() method,
   * EMPTY_INSTANCE and the subclasses.
   */
  StringMapPropertyEditor() {
    displayName_ = "";
  }

  /**
   * Factory method to use when a StringMapPropertyEditor is delegating to us.<P>
   * It constructs the adding and removing methods for the property, from the BeanInfo information.
   * From the same information, it uses 'KeysUpdateInMap' to allow the updating of keys.
   * It uses also 'limitedValuesMethod' to construct the limited-values method 
   * which defines the set of the allowed values in any map value.
   * This is the equivalent for values, of the SimplePropertyEditor work to set
   * the limited-keys method from the property descriptor, with 'choiceMethodForNew'.
   * The property must be a map. These 2 'limit' methods are also used by the subclasses
   * of this class. <P>
   * Called by BeamerManager.addPropertyEditor().
   * Overrides the SimplePropertyEditor method and calls it.
   * @param _source owning the property.
   * @param _propertyDescriptor is the descriptor of the property
   * @exception NullPointerException if _source or _displayName is null
   * @exception ProcessError if BeanInfo and PropertyDecriptor calls fail
   * @return a StringMapPropertyEditor instance
   */
  public static CommonPropertyEditor valueOf(BaseObject _source,
    PropertyDescriptor _propertyDescriptor)
    throws NullPointerException, ProcessError {

      StringMapPropertyEditor result =
        (StringMapPropertyEditor) SimplePropertyEditor.valueOf(
          _source, _propertyDescriptor);

            // for testing purpose
            // System.out.println("end of StringMapPropertyEditor.valueOf() for: "+
            //  result.getDisplayName());

      result.propertyTypeInGroup_ = BeamerManager.getElementaryPropertyType(_propertyDescriptor);
      result.selectedTypeInGroup_ = result.propertyTypeInGroup_;
      result.keysModel_ = new DefaultListModel();
      result.valuesModel_ = new DefaultListModel();

      String      meth;
      Method[]    l_methods;    
      
      try {
            // for testing purpose
            //System.out.println("in StringMapPropertyEditor.valueOf(), "+
            // " the property type name is: "+result.propertyTypeInGroup_.getName()+", and "+
            // "the property descriptor is:"+
            // result.propertyDescriptor_.getShortDescription());

        // if there is an add method
        if (result.propertyDescriptor_.getValue("addMethod") != null) {
            meth = (String) result.propertyDescriptor_.getValue("addMethod");

              // for testing and tracking purpose
               //System.out.println("in StringMapPropertyEditor.valueOf(), "+
                //" the add method is: "+meth+
                //", the source class is:"+result.source_.getClass());

            l_methods = result.source_.getClass().getMethods();
            for (int k = 0; k < l_methods.length; k++){
              //System.out.println("in RightMapPropertyEditor.valueOf(),"+
              //  " source methods: "+l_methods[k].getName());
              if (l_methods[k].getName().equals(meth)) {
                result.addMethod_ = l_methods[k];

                    // for testing and tracking purpose
                    //System.out.println("in StringMapPropertyEditor.valueOf(),"+
                      //" the add method for the property "+
                      //_propertyDescriptor.getName()+" is: "+result.addMethod_);
                    //  +", and the parameter types are: ");
                    //for (int m = 0; m < l_methods[k].getParameterTypes().length; m++){
                    //  System.out.println(
                    //    "For m="+m+": "+l_methods[k].getParameterTypes()[m].getName());
                    //} // for m

                break;
              } // if (l_methods[k]
            } // for k
            
            if (result.addMethod_ == null)
                throw new ProcessError("Add method not found for "+result.displayName_);
        } // if "addMethod

        // if there is a remove method
        // this works to finalize a FinalizedObjectForUser,
        // and also for other cases like a reference to a BaseObject
        // in BaseObjectMapPropertyEditor.
        if (result.propertyDescriptor_.getValue("removeMethod") != null) {
            // there is only one element in the array "removeMethod"
            meth = (String) result.propertyDescriptor_.getValue("removeMethod");
            if (meth.equals("finalizeForUser")) {

              if (FinalizedObjectForUser.class.isAssignableFrom(result.propertyTypeInGroup_))
                l_methods = result.propertyTypeInGroup_.getMethods();
              else {

                  String typeName = UtilityImpl.getFirstInstance().getClassNameFromArrayOrInterface(
                        UtilityImpl.getFirstInstance().getACSVocabularyFromArrayOrInterface(
                            result.propertyTypeInGroup_.getName()),
                        result.propertyTypeInGroup_.getName());
                  // if it is a generic interface like BaseObject or Node
                  if (typeName == null) {

                                    // for testing and tracking
                                    //System.out.println("call to StringMapPropertyEditor.valueOf():");
                                    //System.out.println("--------- with a generic interface type: "+
                                      //      result.propertyTypeInGroup_.getName());

                      typeName = result.propertyTypeInGroup_.getName().substring(
                          result.propertyTypeInGroup_.getName().indexOf("gBaseInterface.") + 15,
                          result.propertyTypeInGroup_.getName().length() );
                  } // if

                  l_methods =  Class.forName(typeName).getMethods();
                  typeName = null;

              } // else

            } // if  (l_meth[0]
            else { // "!finalizeForUser"
              l_methods = result.source_.getClass().getMethods();
            } // else

                  // BUG from Java:
                  //l_addClasses = getClassesForMethod(l_meth);
                  //result.removeMethod_ = result.source_.getClass().getMethod(
                  //  l_meth[0], l_addClasses);

            for (int k = 0; k < l_methods.length; k++){
              if (l_methods[k].getName().equals(meth)) {
                result.removeMethod_ = l_methods[k];

                    // for testing and tracking
                    //System.out.println("in StringMapnPropertyEditor.valueOf(),"+
                    //  " the remove method is: "+
                    //  result.removeMethod_+", and the parameter types are: ");
                    //for (int m = 0; m < l_methods[k].getParameterTypes().length; m++){
                    //  System.out.println(
                    //    "For m="+m+": "+l_methods[k].getParameterTypes()[m].getName());
                    //} // for m

                break;
              } // if (l_methods[k]
            } // for k
            
            if (result.removeMethod_ == null)
                throw new ProcessError("Remove method not found for "+result.displayName_);
        } // if ("removeMethod")!= null "limitedValuesMethod"

        // if there is a limited values method.
        if (result.propertyDescriptor_.getValue("limitedValuesMethod") != null) {
            // there is only one element in the array "limitedValuesMethod"
            meth = (String) result.propertyDescriptor_.getValue("limitedValuesMethod");
            l_methods = result.source_.getClass().getMethods();

                  // BUG from Java:
                  //l_addClasses = getClassesForMethod(l_meth);
                  //result.removeMethod_ = result.source_.getClass().getMethod(
                  //  l_meth[0], l_addClasses);

            for (int k = 0; k < l_methods.length; k++){
              if (l_methods[k].getName().equals(meth)) {
                result.limitedValuesMethod_ = l_methods[k];

                    // for testing and tracking
                    //System.out.println("in StringMapnPropertyEditor.valueOf(),"+
                    //  " the remove method is: "+
                    //  result.removeMethod_+", and the parameter types are: ");
                    //for (int m = 0; m < l_methods[k].getParameterTypes().length; m++){
                    //  System.out.println(
                    //    "For m="+m+": "+l_methods[k].getParameterTypes()[m].getName());
                    //} // for m

                break;
              } // if (l_methods[k]
            } // for k
            
            if (result.limitedValuesMethod_ == null)
                throw new ProcessError("Limited-values method not found for "+result.displayName_);
        } // if ("limitedValuesMethod")!= null 
        
        // if there is a maximum for the number of keys
        if (result.propertyDescriptor_.getValue("MaxKeySize") != null) {
            
          result.maxKeySize_ = ((Integer)result.propertyDescriptor_.
            getValue("MaxKeySize")).intValue();
                  // for testing purpose
                   //System.out.println("in StringMapPropertyEditor.valueOf(), "+
                    //" for the property "+result.propertyDescriptor_.getName()+
                    //", there is a maximum key size: "+result.maxSize_);
        } // if
        else
            result.maxKeySize_ = 50;
        
        // if there is a maximum for the number of values, or sets it to 100
        if (result.propertyDescriptor_.getValue("MaxValueSize") != null) {
            
          result.maxValueSize_ = ((Integer)result.propertyDescriptor_.
            getValue("MaxValueSize")).intValue();
                    // for testing purpose
                   //System.out.println("in StringMapPropertyEditor.valueOf(), "+
                    //" for the property "+result.propertyDescriptor_.getDisplayName()+
                    //", there is a maximum value size: "+result.maxValueSize_);
        } // if
        else
            result.maxValueSize_ = 100;

        // if keys may be updated 
        if (result.propertyDescriptor_.getValue("KeysUpdateInMap") != null) {
            result.keysUpdateInMap_ = ((Boolean)result.propertyDescriptor_.
                    getValue("KeysUpdateInMap")).booleanValue();
        }        
        
      } catch (SecurityException s) {
        throw new InternalError("in StringMapPropertyEditor.valueOf for "+
                result.displayName_+": "+s.toString());
      } catch (Exception l) {
        new ProcessError("from StringMapPropertyEditor.valueOf() for "
          +result.displayName_+": "+l.toString());
      }

            // for testing purpose
             //System.out.println("end of StringMapPropertyEditor.valueOf() for: "+
              //result.getDisplayName());

      meth = null;
      l_methods = null;   
      
      return result;
  } // valueOf()

  // ----------------------- PUBLIC METHODS ---------------------------------


  /**
   * No use. Sets the map value.
   * @param _value The new target object to be edited.  Note that this
   *     object is not modified by the PropertyEditor. May be null.
   */
  public void setValue(Object _value) {
    if (writeMethod_ == null) return;

    return;
  } // setValue(Object value)

  /**
   * Gets the property value as a string suitable for presentation
   * to a human to edit. Return null.
   * @return 'null'
   */
  public String getAsText() {
    return ("null");
  } // getAsText()

  /**
   * Sets the property value by parsing a given String. Raise
   * java.lang.IllegalArgumentException, because this kind of property
   * can't be expressed as text.
   * @param _text  The string to be parsed.
   */
  public void setAsText(String _text) throws java.lang.IllegalArgumentException {
    throw new java.lang.IllegalArgumentException("setAsText(String) is not implemented"+
      " in ARoda0.Gui1.StringMapPropertyEditor");
  } // setAsText(String _text)


  // ----------------------------------------------------------------------

  /**
   * Return a full custom Component that edits its property value,
   * with the relevant buttons following the property descriptor.
   * Override the super method. Overridden in the RightMapPropertyEditor, but not
   * in AssociationMapPropertyEditor and BaseObjectMapPropertyEditor. 
   * Called by BeamerManager.addPropertyEditor().
   * @return a JPanel which contains one panel for the keys and their buttons,
   *   and one panel for the values and their buttons.
   */
  public java.awt.Component getCustomEditor() {

    JPanel result = new JPanel();
    try {
      if (readMethod_.invoke(source_, (Object[])null) == null ) {
          
            // for testing and tracking
            //System.out.println("in StringMapPropertyEditor.getCustomEditor(), "+
              //"read method is null for "+readMethod_+" on "+source_);
        
          return result;
      } // if
    } catch (InvocationTargetException j) {
      new ProcessError("from StringMapPropertyEditor.getCustomEditor(): "+
        j.getTargetException());
      j.printStackTrace();
      return result;
    } catch (Exception l) {
      new ProcessError("from StringMapPropertyEditor.getCustomEditor(): "+
        l.toString());
      l.printStackTrace();
      return result;
    }

    // initialization
    //JLabel                      keyLabel = new JLabel("Key" );
    //JLabel                      valueLabel = new JLabel(getDisplayName()+"Values");
    JList                       keyList;
    JList                       valueList;
    JPanel                      keyListPanel = new JPanel();
    JPanel                      valueListPanel = new JPanel();
    JPanel                      keyButtonPanel = null;
    JPanel                      valueButtonPanel = null;
    JPanel                      mapPanel = new JPanel();
    boolean                     keyButtons = false;
    boolean                     valueButtons = false;
    // an array of 9 booleans: withPreviousButton at 0, withViewButton,
    // withSelectButton, withNewButton, withDeleteButton, withRemoveButton,
    // withClearButton at 6, withNewVButton, withDeletVButton,  
    boolean[]                   l_buttonDisplay = null;   
    
    try {
        setKeysModel();
    } catch (Exception e) {
        System.out.println("in StringMapPropertyEditor.getCustomEditor(), "+
          "exception from the intialize keys method: "+e);
        e.printStackTrace();
    }

    try {
        if (!keysModel_.isEmpty())
            setValuesModel( (String)keysModel_.getElementAt(0) );
    } catch (Exception e) {
        System.out.println("in StringMapPropertyEditor.getCustomEditor(), "+
          "exception from the initialize values method: "+e);
        e.printStackTrace();
    }
          // for testing purpose
          //System.out.println("in StringMapPropertyEditor.getCustomEditor(),"+
            //" the keys list size is: "+keysModel_.getSize());
          //if (!keysModel_.isEmpty()) System.out.println("the first key is:"+
            //keysModel_.get(0)+".");
          //System.out.println("in StringMapPropertyEditor.getCustomEditor(),"+
            //" the values list size is: "+valuesModel_.getSize());
          //if (!valuesModel_.isEmpty()) System.out.println("the first value is:"+
            //valuesModel_.get(0)+".");
          
    //keyLabel.setMaximumSize(NAME_DIMENSION);
    //keyLabel.setMinimumSize(NAME_DIMENSION);
    //keyLabel.setPreferredSize(NAME_DIMENSION);
    //keyLabel.setFont(BeamerManager.plainFont);
    //keyLabel.setToolTipText(propertyDescriptor_.getShortDescription());

    keyList = new JList(keysModel_);
    // valuesModel_ contains the values for the first key
    if (keysModel_.size() != 0) keyList.setSelectedIndex(0);
    keyList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    // fire events to valueChanged()
    keyList.addListSelectionListener(this);
    
    keyList.setBackground(InternalFrame.NON_EDITABLECOLOR);
    ((DefaultListCellRenderer)keyList.getCellRenderer()).
      setHorizontalAlignment(JLabel.RIGHT);
    ((DefaultListCellRenderer)keyList.getCellRenderer()).
      setHorizontalTextPosition(JLabel.RIGHT);
    keyList.setFont(BeamerManager.plainFont);
    keyScrollPane_ = new JScrollPane(keyList);
    
    valueList = new JList(valuesModel_);
    valueList.setBackground(InternalFrame.NON_EDITABLECOLOR);

    if (propertyDescriptor_.getValue("horizontalTextPositionToLeft") != null &&
            (Boolean)propertyDescriptor_.getValue("horizontalTextPositionToLeft")) {
        ((DefaultListCellRenderer)valueList.getCellRenderer()).
            setHorizontalAlignment(JLabel.LEFT);
        ((DefaultListCellRenderer)valueList.getCellRenderer()).
            setHorizontalTextPosition(JLabel.LEFT);
    } else {
        ((DefaultListCellRenderer)valueList.getCellRenderer()).
            setHorizontalAlignment(JLabel.RIGHT);
        ((DefaultListCellRenderer)valueList.getCellRenderer()).
            setHorizontalTextPosition(JLabel.RIGHT);
    }    
    
    valueList.setFont(BeamerManager.plainFont);
    // valueList has not this instance as ListSelectionListener
    valueScrollPane_ = new JScrollPane(valueList);
    
    l_buttonDisplay = getButtonsDisplaying(); 
    keyButtonPanel = setVerticalButtonsInEditor(l_buttonDisplay, true, false);
    if (keyButtonPanel != null) keyButtons = true;
    valueButtonPanel = setVerticalButtonsInEditor(l_buttonDisplay, false, true);
    if (valueButtonPanel != null) valueButtons = true;  
    
    // initializes the keys and values panel dimensions
    setPanelDimension(keyButtons, valueButtons);
    
    keyScrollPane_.getHorizontalScrollBar().setValue(
      keyScrollPane_.getHorizontalScrollBar().getMaximum() );  // does not work well
    //keyListPanel.add(keyTitle);
    keyListPanel.add(keyScrollPane_);
    keyListPanel.setBackground(Color.WHITE);
    keyListPanel.setBorder(new LineBorder(InternalFrame.FORECOLOR));   

          // for testing purpose
          //System.out.println("in StringMapPropertyEditor.getCustomEditor(),"+
            //" the font cell renderer in list is a: "+
            //((DefaultListCellRenderer)keyList.getCellRenderer()).getFont() );

    
    valueScrollPane_.getHorizontalScrollBar().setValue(
      valueScrollPane_.getHorizontalScrollBar().getMaximum() ); 
    valueListPanel.add(valueScrollPane_);
    valueListPanel.setBackground(Color.WHITE);
    valueListPanel.setBorder(new LineBorder(InternalFrame.FORECOLOR)); 
    
    // builds up the result
    if (keyButtonPanel != null) keyListPanel.add(keyButtonPanel);
    if (valueButtonPanel != null) valueListPanel.add(valueButtonPanel);
    
    mapPanel.add(keyListPanel);
    mapPanel.add(valueListPanel);
    
    result.add(Box.createVerticalBox());
    //result.add(titlePanel);    
    result.setBackground(Color.white);
    result.setBorder(new EtchedBorder());
    result.add(mapPanel);

    boundComponent_ = keyList;
    boundComponentTwo_ = valueList;


    //if (propertyDescriptor_.isBound() ) boundComponent_ = keyList;

              // for testing purpose
         //System.out.println("in StringMapPropertyEditor.getCustomEditor(), "+
         // "the displayed list for "+propertyDescriptor_.getName()+" is:");
         // for (int i = 0; i < keysModel_.getSize(); i++) {
         //   System.out.println("for "+i+": "+keysModel_.getElementAt(i));
         // }

    mapPanel = null;
    //keyTitle = null;
    keyList = null;
    keyListPanel = null;
    //valueTitle = null;
    valueList = null;
    valueListPanel = null;
    keyButtonPanel = null;
    valueButtonPanel = null; 
    
    return result;
  } // getCustomEditor()

  /**
   * Manages the events fired by the buttons 'New' and 'Delete' for the map keys,
   * and 'Select', 'Remove' for the map values, in the JPanel displayed 
   * by getCustomEditor().
   * Overrides the super method. Overridden and called by the subclasses method.
   * @param _evt is the event fired by a button
   */
  public void actionPerformed(ActionEvent _evt) {

    // no action if the explorer is currently used for an object selection
    // or if there is an user dialog currently
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
      
    String buttonText = ((JButton)_evt.getSource()).getText();

    if (!_evt.getSource().getClass().equals(JButton.class)) {
      throw new InternalError("actionPerformed received "+
        "in StringMapPropertyEditor from a component which is not a button");
    }

    // for testing purpose
    // System.out.println("actionPerformed received in StringMapPropertyEditor from "+
    //  "the button "+((JButton)_evt.getSource()).getText());

    // only for a new key, not a new value
    if ( buttonText.equals( "  New  ")) {
      addKeyElement(); 
    } // if (((JButton)
    // Delete only for a key, not a map value
    else if ( buttonText.equals("Delete")) {

                  // for testing purpose
                  //System.out.println("actionPerformed() in RightMapPropertyEditor:"+
                   // "selection: "+((JList)boundComponent_).getSelectedValues().length);


      // user message if no selection in the keys JList
      if ( ((JList)boundComponent_).getSelectedValues().length == 0) {
        DialogBox.showErrorMessage("No selection in the keys list.");
      }
      else deleteKeyElements( ((JList)boundComponent_).getSelectedValues());
    }
    // selects only for a value, not for a key
    else if ( buttonText.equals("Select")) {
      // user message if no selection in the keys JList
      if ( ((JList)boundComponent_).getSelectedValues().length == 0) {
        DialogBox.showErrorMessage("No selection in the keys list.");
      }
      else {
        addRestrictedValueElements( (String)((JList)boundComponent_).getSelectedValue() );
      } // else
    } // else if 'Select' addValueElements(Object[] _l_values)
    // selects only for a value, not for a key
    else if ( buttonText.equals("NewV")) {
      // user message if no selection in the keys JList
      if ( ((JList)boundComponent_).getSelectedValues().length == 0) {
        DialogBox.showErrorMessage("No selection in the keys list.");
      }
      else {
        String value = DialogBox.showSingleEditor("Enter a new line of text");
        if (value != null) addValueElements(new String[] {value});
        value = null;
      } // else
    } // else if 'NewV'
    else if ( buttonText.equals("Remove" )) {
      // user message if no selection in the values JList
      if ( ((JList)boundComponentTwo_).getSelectedValues().length == 0) {
        DialogBox.showErrorMessage("No selection in the values list.");
      }
      else removeValueElements(((JList)boundComponentTwo_).getSelectedValues(), true);
    } // else if ( buttonText == "Remove"
    else if ( buttonText.equals("DeletV" )) {
      // user message if no selection in the values JList
      if ( ((JList)boundComponentTwo_).getSelectedValues().length == 0) {
        DialogBox.showErrorMessage("No selection in the values list.");
      }
      else removeValueElements(((JList)boundComponentTwo_).getSelectedValues(), true);
    } // else if ( buttonText == "DeletV"
    buttonText = null;
  } // actionPerformed()

  /**
   * Manages the events fired by the BaseObject source_,
   * if it is a bound property and when the property value changes.
   * The main behavior is that the old and new property values in the argument
   * are not ridden to update the values. This allow to process a PropertyChangeEvent
   * that does not describe these values, but is fired by the gBase package when
   * the relevant values are trully changed.
   * The selected key is kept if possible. Manages the different pane dimensions
   * which are required. Called by addRestrictedValueElements() when an
   * exception is thrown by the method which gives the restricted values.
   * Not overridden in the subclasses.
   * @param _evt only source and property name in _evt are used in this method
   */
  public void propertyChange(PropertyChangeEvent _evt) {

            // for testing purpose
            //System.out.println("StringMapPropertyEditor.propertyChange() called for "+
              //_evt.getPropertyName()+" by "+((BaseObject)_evt.getSource()).getName());
            
    if (_evt.getSource() != source_ ||
      !_evt.getPropertyName().equals(propertyDescriptor_.getName()) ||
      boundComponent_ == null) {
        
            // for testing purpose
            System.out.println("StringMapPropertyEditor.propertyChange() noop");

        return;
    }

    // to restore the previous key in the new map, if possible
    String selectedKey = null;
    if (((JList)boundComponent_).getSelectedIndex() != -1) 
        selectedKey = (String)keysModel_.getElementAt(
                ((JList)boundComponent_).getSelectedIndex() );    
    
    synchronized (keysModel_) { 
        ((JList)boundComponent_).removeListSelectionListener(this); // no call to valueChanged()               
        setKeysModel();
        
        if (selectedKey != null && keysModel_.contains(selectedKey)) { 
            setValuesModel( selectedKey );
            ((JList)boundComponent_).setSelectedValue(selectedKey, true);
        }
        else if (!keysModel_.isEmpty()) {
                setValuesModel((String)keysModel_.getElementAt(0) );
                ((JList)boundComponent_).setSelectedValue(keysModel_.getElementAt(0), true);
        }
        else {
            setValuesModel(null);
        }
        ((JList)boundComponent_).addListSelectionListener(this);        
    } // synchro

    changeDimensions();
    
    selectedKey = null;   
  } // propertyChange()
  
    /**
   * Updates the dimensions of the key and value panels.
   * Called by propertyChange().
   * Overridden by the RightMapPropertyEditor subclass. 
   */
  public void changeDimensions() {
    JPanel                      keyButtonPanel = null;
    boolean                     keyButtons = false; 
    boolean[]                   l_buttonDisplay = getButtonsDisplaying();    
     
    keyButtonPanel = setVerticalButtonsInEditor(l_buttonDisplay, true, false);
    if (keyButtonPanel != null) keyButtons = true;
    if (keyButtons)
      setListPanelDimension(keysModel_.getSize(), keyScrollPane_,
              false, true, true);
    else
      setListPanelDimension(keysModel_.getSize(), keyScrollPane_,
              false, false, false);

     setVerticalButtonsInEditor(l_buttonDisplay, false, true);
    setListPanelDimension(valuesModel_.getSize(), valueScrollPane_, true, true, false);
   
    keyButtonPanel = null;
    l_buttonDisplay = null; 
  }  // changeDimensions()
  
    /**
   * When a key is selected by the user in the beamer, displays the right values list.
   * @param _e from the key list
   */
  public void valueChanged(ListSelectionEvent _e) {
                          
                    // for testing and tracking         
                    //System.out.println("in StringMapPropertyEditor.valueChanged(),"+
                      //  " event: "+_e);

      setValuesModel( (String)keysModel_.getElementAt(
        ((JList)_e.getSource()).getSelectedIndex() ) );
  }
  
    /**
   * Sets the two panels for the keys and the values lists.
   * Called by getCustomEditor(). Calls setListPanelDimension().
   * Overridden by AssociationMapPropertyEditor, but not by RightMapPropertyEditor.
   * @param _keyButtons true if there is a vertical panel for key buttons.
   * @param _valueButtons true if there is a vertical panel for value buttons.
   */
  public void setPanelDimension(boolean _keyButtons, boolean _valueButtons) {
                          
      if (_keyButtons)
          setListPanelDimension(keysModel_.getSize(), keyScrollPane_,
                  false, true, true);
      else
          setListPanelDimension(keysModel_.getSize(), keyScrollPane_, 
                  false, false, false);
      
        setListPanelDimension(getMaximumValuesListSize(), valueScrollPane_,
               true, true, false); 
  } // setPanelDimension()
  
    //----------------------------------------------------------------------

  /**
   * Get the values of variables for the instance.
   * @return source, propertyName and listeners (if they have toString() methods).
   */
  public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("  STRING_MAP_PROPERTY_EDITOR.TOSTRING():");

        sb.append(" source_ = ");
        sb.append(source_);

        sb.append("; display name = ");
        sb.append(displayName_);

        sb.append("; listeners list = ");
        sb.append(listeners_);

        return " [" + sb + "] ";
  } // toString()

  /**
   * Finalizes the instance; the call is reserved to BeamerManager.
   */
  public void finalizeForProcess() {

    super.finalizeForProcess();
  } // finalizeForProcess()

  // -------------------- PROTECTED METHODS -----------------------------------

  /**
   * Called by getCustomEditor(), propertyChange(),  and the subclasses,
   * to set the keys list, without database update. Controls the size. 
   * Caution: unregister this instance as ListSelectionListener of boundComponent_ 
   * before the call of this method, if it is outside getCustomEditor().
   */
  protected void setKeysModel() {

                // for testing and tracking
                //System.out.println("in StringMapPropertyEditor.setKeysModel(),"+
                  //      " keysModel_ at the start: "+keysModel_);      
      
    keysModel_.clear();
    Map map = (Map)getValue();
    String key = null;
    
    if (map == null || map.equals(EMPTY_OBJECT) ) {
        map = null;
        
                // for testing and tracking
                //System.out.println("in StringMapPropertyEditor.setKeysModel(),"+
                  //      " the map is null.");
        
        return;
    }
    try {
      // sets keysModel_ with the read method if it is possible.
      // else keysModel_ is empty
          
                          // for testing and tracking
                //if (propertyDescriptor_.getName().equals("ImpliedOpenViews")) {           
                    //System.out.println("in StringMapPropertyEditor.setKeystModel(),"+
                        //" for the property 'ImpliedOpenViews', readMethod_ is: "+
                        //readMethod_);
                //} // if

      // for Strings, puts them in the keysModel_
        for (Iterator ite = map.keySet().iterator() ; ite.hasNext() ; ) {
            key = (String) ite.next();
            keysModel_.addElement(key);
        
                // for testing and tracking
                //System.out.println("in StringMapPropertyEditor.setKeysModel(),"+
                  //      " add the key: "+key);
                
        }
        if (keysModel_.size() > maxKeySize_) {
            new ProcessError("from StringMapPropertyEditor.setKeysModel(), "+
                    "for the property "+propertyDescriptor_.getName()+
                    "the key size is superior to the maximum: "+maxKeySize_);
            keysModel_.clear();
        }


    } catch (Exception l) {
      new ProcessError("from StringMapPropertyEditor.setKeysModel(): ");
      l.printStackTrace();
    }
    map = null;
    key = null;
  } // setKeysModel()
  
  /**
   * Sets the values list in the JList of Objects, without database update.
   * Controls the maximal size. Sets the scroll bar to right if it is not an
   * AssociationMapPropertyEditor instance, for which the scroll bar is set to left.
   * Called by getCustomEditor(), propertyChange(), valueChanged() and the subclasses.
   * @param _key is a map key for the property. Clears the values
   *  if it is not a current key or if _key is null.
   */
  protected void setValuesModel(String _key) {

    valuesModel_.clear();
    if (_key == null) return;
    
    Map map = (Map)getValue();
    if (map == null || map.equals(EMPTY_OBJECT) || 
      map.isEmpty() ) {
        map = null;
        return;
    }
    
                          // for testing and tracking
                        //System.out.println("call to StringMapPropertyEditor.setValuesModel()");


    try {
        // sets valuesModel_ with the read method if it is possible.
        // else valuesModel_ is empty

        // the value is a String array, or an ImmutableName for a BaseObjectMapPropertyEditor,
        // or StringRight array for a RightMapPropertyEditor
        Object value = (Object) map.get(_key ); 
        
                          // for testing and tracking
                //if (propertyDescriptor_.getName().equals("ImpliedOpenViews")) {           
                    //System.out.println("in StringMapPropertyEditor.inializeValuesModel(),"+
                        //" for the property 'ImpliedOpenViews', readMethod_ is: "+
                        //readMethod_);
                //} // if
        
        if (value instanceof Object[]) {

            Object[] l_values = (Object[]) value;

            // puts them in the valuesModel_ in the same order
            if (l_values != null) {
                for (int i = 0; i < l_values.length; i++ ) {
                   valuesModel_.addElement(l_values[i] );
                } // for
            } // if
            if (valuesModel_.size() > maxValueSize_) {
                new ProcessError("from StringMapPropertyEditor.setValuesModel(),"+
                    " for "+displayName_+
                    ", the number of values is greater than "+maxValueSize_);
                valuesModel_.clear();
            }
            l_values = null;
        }
        else {
            valuesModel_.addElement(value );
        }

        value = null;
    } catch (Exception l) {
      new ProcessError("from StringMapPropertyEditor.setValuesModel(): ");
      l.printStackTrace();
    }
    
    // sets the scroll bar to right or left, if initialized
    if (valueScrollPane_ != null) {
        if (!(this instanceof AssociationMapPropertyEditor))
            valueScrollPane_.getHorizontalScrollBar().setValue(
            valueScrollPane_.getHorizontalScrollBar().getMaximum() );
        else valueScrollPane_.getHorizontalScrollBar().setValue(
            valueScrollPane_.getHorizontalScrollBar().getMinimum() );
    } // if
    
    map = null;
  } // setValuesModel()


  /**
   * Remove some selected values in the map, from an user click on 'Remove'.
   * Called by actionPerformed() and the same methods in the subclasses.
   * Uses the 'remove' method to update the values for the current selected key.
   * @param _l_valuesToRemove String or StringRight values to remove 
   *    for the current selected key. Never null.
   * @param _toConfirm true if the user has to confirm the removing
   */
  protected void removeValueElements(Object[] _l_valuesToRemove, boolean _toConfirm) {

    if (_toConfirm && !DialogBox.showConfirmDialog(
          " Do you want to delete these selected value(s)?", null)) return;

    // updates the values for the right key
    String keyElement = (String)keysModel_.getElementAt(
        ((JList)boundComponent_).getSelectedIndex() );
    
    if (keyElement == null) {
        DialogBox.showErrorMessage("Sorry, a key must be selected in the key list");
        return;
    }
    
    // uses propertyTypeInGroup_ to create an array from the correct class:
    // String[], StringRight[], subclass of BaseObjet[]
    Object[] l_propertyTypeArray = (Object[])
        Array.newInstance(propertyTypeInGroup_, _l_valuesToRemove.length);   
    System.arraycopy(_l_valuesToRemove, 0, l_propertyTypeArray, 0, _l_valuesToRemove.length);
    
    try {
        
          removeMethod_.invoke(source_, new Object[] {keyElement, l_propertyTypeArray} );
          
    } catch (InvocationTargetException i) {
          // show the message to the user
          int lastIndex = ((InvocationTargetException)i).getTargetException().
                  toString().lastIndexOf(": ");
          if (lastIndex != 0)
            DialogBox.showErrorMessage(
              ((InvocationTargetException)i).getTargetException().toString().
                  substring(lastIndex+1));
          else DialogBox.showErrorMessage(
            ((InvocationTargetException)i).getTargetException().toString()); 

          //System.out.println("in StringMapPropertyEditor.removeValueElements(), "+
            //"the exception is an InvocationTargetException: ");
          //((InvocationTargetException)i).printStackTrace();
      
          // restore the base value in the editor
          propertyChange(new PropertyChangeEvent(source_, 
                  propertyDescriptor_.getName(), null, null));
          
    } catch (Exception e) {
        if (e instanceof CreateError || e instanceof UpDateError || e instanceof BaseError) {
          DialogBox.showErrorMessage(e.getCause().getMessage());
          //System.out.println("in CollectionPropertyEditor.setValue(), exception: ");
          //i.getCause().printStackTrace();
          return;
        }

        System.out.println("in StringMapPropertyEditor.removeValueElements(), exception: "+e);
        e.printStackTrace();
    } // catch 
    keyElement = null;
    l_propertyTypeArray = null;
  } // removeValueElements()   
  
  /**
   * Called by the setPanelDimension method for the subclasses.
  * @return the maximum size of the values list among the current keys
  */
  protected int getMaximumValuesListSize() {
    int         result = 0;
    Map         map = (Map)getValue();
    Object[]    l_values = null;
    Object      value = null;
    Collection  list = null;
    
    if (map == null || map.equals(EMPTY_OBJECT) ) {
        map = null;
        
                // for testing and tracking
                //System.out.println("StringMapPropertyEditor.getMaximumValuesListSize(),"+
                  //      " the map is null.");
        
        return 0;
    }
    try {
        for (Iterator ite = map.values().iterator() ; ite.hasNext() ; ) {
            value = (Object) ite.next();
            if (value instanceof Object[]) {
                l_values = (Object[]) value;
                if (l_values.length > result) result = l_values.length;
            }
            if (value instanceof Collection) {
                list = (Collection) value;
                if (list.size() > result) result = list.size();
            }
            else {
                result++;
            }
        } // for ite
    } catch (Exception l) {
          new ProcessError("from StringMapPropertyEditor.getMaximumValuesListSize(): ");
          l.printStackTrace();
    }
    map = null;
    l_values = null;

                // for testing and tracking
                //System.out.println("in StringMapPropertyEditor.getMaximumValuesListSize(),"+
                  //      " result = "+result);
      
    return result;
  } // getMaximumValuesListSize()  
  
  /**
   * Called by actionPerformed() when the user clicks on the button 'New' to
   * add a new key in the map. Noop if keysUpdateInMap_ is false.
   * The key may be restricted or not, and to get it, it calls getRestrictedValues().
   * The maximal number of restricted keys is 111, since it is possible to display
   * up to 4 lists of 28 items.
   * There is a null argument if the source is not an ACSObject.
   * The ACS of the source is used only if the source is an ACSObject,
   * not for an ACSFActoryImpl for instance.
   * Overridden in RightMapPropertyEditor and StringTwoKeysMapPropertyEditor, but
   * not in BaseObjectMapPropertyEditor.
   */
  protected void addKeyElement() {

                          // for testing and tracking purpose
                     //System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
                       //     "the property type is "+propertyTypeInGroup_);
      
    if (!keysUpdateInMap_) {
        DialogBox.showErrorMessage("Keys cannot be changed in this property");
        return;
    }
                     
    ACS         acs = null;
    String[]    l_restrictedValues = null;
    String      newValue = null;
    // l_valuesArray is a StringRight array, or an ImmutableName
    // for the subclass BaseObjectPropertyEditor
    Object[]    l_valuesArray = (Object[]) Array.newInstance(propertyTypeInGroup_, 0);
    
    if (source_ instanceof ACSObject)
            acs = (ACS)ACS_Facade.getOpenACS().get(((ACSObject)source_).getAcsName() );
    
    // the ACSFactoryImpl properties do not use 'acs'
    l_restrictedValues = getRestrictedValues(acs);
    
    if (l_restrictedValues == null) {
        newValue = DialogBox.showSingleEditor("Enter the new key:");

        if (newValue == null) {
            acs = null;
            return;
        } // if

                          // for testing and tracking purpose
                     //System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
                       //      "entered value: "+newValue);
    }
    else {
        boolean[]       l_currentValuesIndex = null;
        short           selectionCount = 0;
        String[]        l_ToBeSelectedValues;
        String[]        l_ToBeSelectedPart = null;
        boolean         singleAnswer = false;

        // removes in the displayed list, all the values which are already selected
        // and returns if all of them are selected
        l_currentValuesIndex = new boolean[l_restrictedValues.length];
        for (int i = 0; i < l_restrictedValues.length; i++) {
            if (keysModel_.contains(l_restrictedValues[i])) {
              l_currentValuesIndex[i] = true;
              selectionCount++;
            }
        }   // for i
        if (l_restrictedValues.length == selectionCount) {
          DialogBox.showErrorMessage(
              "All the allowed keys are already selected." );
          acs = null;
          l_restrictedValues = null;
          l_currentValuesIndex = null;
          return;
        }
        l_ToBeSelectedValues = new String[l_restrictedValues.length - selectionCount];
        selectionCount = 0;
        for (int i = 0; i < l_restrictedValues.length; i++) {
          if (!l_currentValuesIndex[i]) {
            l_ToBeSelectedValues[selectionCount] = l_restrictedValues[i];
            selectionCount++;
          }
        } // for

                                      // for testing and tracking purpose
                                 //System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
                                   //      "l_ToBeSelectedValues.length: "+l_ToBeSelectedValues.length);


        int index = -1;
        if (l_ToBeSelectedValues.length == 1) {
            DialogBox.showOkMessage("There is only one authorized key.");
            index = 0;
            l_ToBeSelectedPart = l_ToBeSelectedValues;
        }
        else {
            try {
                if (l_ToBeSelectedValues.length < 30) {
                    l_ToBeSelectedPart = l_ToBeSelectedValues;
                    index = DialogBox.showOneSelectionOneDialog(
                        "new key", l_ToBeSelectedPart);
                } else if (l_ToBeSelectedValues.length < 57) {

                                      // for testing and tracking purpose
                                 //System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
                                   //      "l_ToBeSelectedValues.length: "+l_ToBeSelectedValues.length);

                    l_ToBeSelectedPart = new String[28];
                    System.arraycopy(l_ToBeSelectedValues, 0, l_ToBeSelectedPart, 0, 28);

                    index = DialogBox.showOneSelectionOneDialog(
                        "item, or close it to skip to the second list", l_ToBeSelectedPart);
                    if (index == -1 && l_ToBeSelectedValues.length != 29) {
                        l_ToBeSelectedPart = new String[l_ToBeSelectedValues.length - 28];
                        System.arraycopy(l_ToBeSelectedValues, 28, l_ToBeSelectedPart,
                                0, l_ToBeSelectedValues.length - 28);
                        index = DialogBox.showOneSelectionOneDialog(
                            "new key", l_ToBeSelectedPart);
                    } // if
                    if (index == -1 && l_ToBeSelectedValues.length == 29) {
                        singleAnswer = DialogBox.showConfirmDialog("Do you choose the key "+
                                l_ToBeSelectedValues[29]+"?",
                                "This key is the single key in the list 2/2");
                        if (singleAnswer) index = 29;
                    }
                } // else
                else if (l_ToBeSelectedValues.length < 85) {

                                      // for testing and tracking purpose
                                 //System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
                                   //      "l_ToBeSelectedValues.length: "+l_ToBeSelectedValues.length);

                    l_ToBeSelectedPart = new String[28];
                    System.arraycopy(l_ToBeSelectedValues, 0, l_ToBeSelectedPart, 0, 28);

                    index = DialogBox.showOneSelectionOneDialog(
                        "item, or close it to skip to the list 2/3", l_ToBeSelectedPart);
                    if (index == -1) {
                        l_ToBeSelectedPart = new String[28];
                        System.arraycopy(l_ToBeSelectedValues, 28, l_ToBeSelectedPart,
                                0, 28);
                        index = DialogBox.showOneSelectionOneDialog(
                            "item, or close it to skip to the list 3/3", l_ToBeSelectedPart);
                    } // if
                    if (index == -1 && l_ToBeSelectedValues.length != 57) {
                        l_ToBeSelectedPart = new String[l_ToBeSelectedValues.length - 56];
                        System.arraycopy(l_ToBeSelectedValues, 56, l_ToBeSelectedPart,
                                0, l_ToBeSelectedValues.length - 56);
                        index = DialogBox.showOneSelectionOneDialog(
                            "new key", l_ToBeSelectedPart);
                    } // if
                    if (index == -1 && l_ToBeSelectedValues.length == 57) {
                        singleAnswer = DialogBox.showConfirmDialog("Do you choose the key "+
                                l_ToBeSelectedValues[57]+"?",
                                "This key is the single key in the list 3/3");
                        if (singleAnswer) index = 57;
                    }
                } // else
                else if (l_ToBeSelectedValues.length < 112) {

                                      // for testing and tracking purpose
                                 //System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
                                   //      "l_ToBeSelectedValues.length: "+l_ToBeSelectedValues.length);

                    l_ToBeSelectedPart = new String[28];
                    System.arraycopy(l_ToBeSelectedValues, 0, l_ToBeSelectedPart, 0, 28);

                    index = DialogBox.showOneSelectionOneDialog(
                        "item, or close it to skip to the list 2/4", l_ToBeSelectedPart);
                    if (index == -1) {
                        l_ToBeSelectedPart = new String[28];
                        System.arraycopy(l_ToBeSelectedValues, 28, l_ToBeSelectedPart,
                                0, 28);
                        index = DialogBox.showOneSelectionOneDialog(
                            "item, or close it to skip to the list 3/4", l_ToBeSelectedPart);
                    } // if
                    if (index == -1) {
                        l_ToBeSelectedPart = new String[28];
                        System.arraycopy(l_ToBeSelectedValues, 56, l_ToBeSelectedPart,
                                0, 28);
                        index = DialogBox.showOneSelectionOneDialog(
                            "item, or close it to skip to the list 4/4", l_ToBeSelectedPart);
                    } // if
                    if (index == -1 && l_ToBeSelectedValues.length != 85) {
                        l_ToBeSelectedPart = new String[l_ToBeSelectedValues.length - 84];
                        System.arraycopy(l_ToBeSelectedValues, 84, l_ToBeSelectedPart,
                                0, l_ToBeSelectedValues.length - 84);
                        index = DialogBox.showOneSelectionOneDialog(
                            "new key", l_ToBeSelectedPart);
                    } // if
                    if (index == -1 && l_ToBeSelectedValues.length == 85) {
                        singleAnswer = DialogBox.showConfirmDialog("Do you choose the key "+
                                l_ToBeSelectedValues[84]+"?",
                                "This key is the single key in the list 4/4");
                        if (singleAnswer) index = 84;
                    }
                } // else if (l_ToBeSelectedValues.length < 112)

            } catch (SmallError s) {
              acs = null;
              l_restrictedValues = null;
              l_currentValuesIndex = null;
              l_ToBeSelectedValues = null;
              return;
            }
        } // else

        if (index == -1) {
          acs = null;
          l_restrictedValues = null;
          l_currentValuesIndex = null;
          l_ToBeSelectedValues = null;
          return;
        }

        if (!singleAnswer)
            newValue = l_ToBeSelectedPart[index];
        else
            newValue = l_ToBeSelectedValues[index];

                          // for testing and tracking purpose
                     //System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
                       //      "entered value: "+newValue);

        l_currentValuesIndex = null;
        l_ToBeSelectedValues = null;
        l_ToBeSelectedPart = null;
    } // else of if (l_restrictedValues == null
 
    try {
        
      addMethod_.invoke(source_, new Object[] {newValue, l_valuesArray } );
    
    } catch (InvocationTargetException i) {
          // show the message to the user
          int lastIndex = ((InvocationTargetException)i).getTargetException().
                  toString().lastIndexOf(": ");
          if (lastIndex != 0)
            DialogBox.showErrorMessage(
              ((InvocationTargetException)i).getTargetException().toString().
                substring(lastIndex+1));
          else DialogBox.showErrorMessage(
            ((InvocationTargetException)i).getTargetException().toString()); 

          // CreateError is for the user errors
          //if (!(i.getTargetException() instanceof CreateError)) {
            //  System.out.println("in StringMapPropertyEditor.addKeyElement(), "+
              //  "the exception is an InvocationTargetException: ");
              //((InvocationTargetException)i).printStackTrace();
          //} // if
      
          // restores the current value in the editor
          propertyChange(new PropertyChangeEvent(source_, 
                  propertyDescriptor_.getName(), null, null));
          
          return;
    } catch (Exception e) {
        if (e instanceof CreateError || e instanceof UpDateError || e instanceof BaseError) {
          DialogBox.showErrorMessage(e.getCause().getMessage());
          //System.out.println("in CollectionPropertyEditor.setValue(), exception: ");
          //i.getCause().printStackTrace();
          return;
        }

        System.out.println("in StringMapPropertyEditor.addKeyElement(), exception: "+e);
        e.printStackTrace();
    } // catch 
        
    // selects the new key
    ((JList)boundComponent_).setSelectedValue(newValue, true);
    
    newValue = null;
    acs = null;
    l_restrictedValues = null;
 } // addKeyElement()

  /**
   * Called by a subclass actionPerformed() method, when the user clicks on the button 'See'.
   * This is for the subclasses which handle BaseObjects and ImmutableNames
   * as elementary items in the map values.
   * This method selects a BaseObject in the explorer, and then it is displayed
   * in the beamer. Copy of the same method in CollectionPropertyEditor.
   * Call super.viewOneElement().
   * Called by RightMapPropertyEditor and BaseObjectPropertyEditor.
   */
  protected void viewElement() {

                      // for testing purpose
                 //System.out.println("StringMapPropertyEditor.viewElement()");
      
    // initialization
    JList       jlist = (JList) boundComponentTwo_;
    Object[]    l_selectedArray = jlist.getSelectedValues();
    Object      valueInList;

    if (jlist.getModel().getSize() == 0) {
      DialogBox.showErrorMessage("The list is empty.");
      jlist = null;
      l_selectedArray = null;
      return;
    } // if

    if (jlist.getModel().getSize() > 1 && l_selectedArray.length != 1) {
      DialogBox.showErrorMessage("Select only one element to view.");
      jlist = null;
      l_selectedArray = null;
      return;
    } // if

    if (jlist.getModel().getSize() == 1) 
        valueInList = jlist.getModel().getElementAt(0);
    else valueInList = l_selectedArray[0];

    // sets the selected object in the beamer
    if (BaseObject.class.isAssignableFrom(propertyTypeInGroup_) && valueInList != null) {
        viewOneElement( valueInList ); // finalizes this instance, if in the beamer
    }    
    if (ImmutableName.class.isAssignableFrom(propertyTypeInGroup_) && valueInList != null &&
            BaseObject.class.isAssignableFrom(selectedTypeInGroup_)) {
        //try {
          //  valueInList = ACS_Facade.getEorBaseObject((ImmutableName)valueInList);
        //}
        //catch (BaseError b) {
          //DialogBox.showErrorMessage("The selected name is unknown.");
          //jlist = null;
          //l_selectedArray = null;
          //valueInList = null;
          //return;
        //}
        jlist = null;
        l_selectedArray = null;
        viewOneElement( valueInList ); // finalizes this instance, if in the beamer
    }
    else {
                      // for testing purpose
                 //System.out.println("StringMapPropertyEditor.viewElement(), "+
                   //      "no call to viewOneElement() for "+displayName_+
                     //    " with selectedTypeInGroup_: "+selectedTypeInGroup_+
                       //  " and with propertyTypeInGroup_: "+propertyTypeInGroup_);
    }

    jlist = null;
    l_selectedArray = null;
    valueInList = null;
  } // viewElement()

  /**
   * Adds some values in the map, from an user click on 'Select', 'NewV' or 'Prev.'.
   * Controls the resulting array size and the current selection of a key.
   * Called directly only by the subclasses BaseObjectMapPropertyEditor and RightMapPropertyEditor.
   * Uses the 'add' method to update the values for the current selected key,
   * and propertyTypeInGroup_ to define the correct type to handle.
   * Called by addRestrictedValueElements() and BaseObjectMapPropertyEditor.actionPerformed().
   * @param _l_values to add for the current selected key. Never null.
   */
  protected void addValueElements(Object[] _l_values) {

    int currentSize = 0;
    if (valuesModel_ != null) currentSize = valuesModel_.size();
    if (maxValueSize_ > 0 && maxValueSize_ < currentSize + _l_values.length) {
      DialogBox.showErrorMessage("No operation since "+
        maxValueSize_+" is the maximum size for the values list.");
      return;
    } // if
      
    // updates the values for the right key
    String keyElement = (String)keysModel_.getElementAt(
        ((JList)boundComponent_).getSelectedIndex() );
    
    if (keyElement == null) {
        DialogBox.showErrorMessage("Sorry, a key must be selected in the key list");
        return;
    }
    
    // uses propertyTypeInGroup_ to create an array from the correct class:
    // String[], StringRight[], subclass of BaseObjet[]
    Object[] l_propertyTypeArray = (Object[])
        Array.newInstance(propertyTypeInGroup_, _l_values.length);   
    System.arraycopy(_l_values, 0, l_propertyTypeArray, 0, _l_values.length);
    
    try {
        
          addMethod_.invoke(source_, new Object[] {keyElement, l_propertyTypeArray} );
          
    } catch (InvocationTargetException i) {
          // shows the message to the user
          int lastIndex = ((InvocationTargetException)i).getTargetException().
                  toString().lastIndexOf(": ");
          if (lastIndex != 0)
            DialogBox.showErrorMessage(
              ((InvocationTargetException)i).getTargetException().toString().
                  substring(lastIndex+1));
          else DialogBox.showErrorMessage(
            ((InvocationTargetException)i).getTargetException().toString()); 

          // CreateError is for the user errors
          //if (!(i.getTargetException() instanceof CreateError)) {
            //      System.out.println("in StringMapPropertyEditor.addValueElements(), "+
              //      "the exception is an InvocationTargetException: ");
                //  ((InvocationTargetException)i).printStackTrace();
          //} // if
      
          // restores the base value in the editor
          propertyChange(new PropertyChangeEvent(source_, 
                  propertyDescriptor_.getName(), null, null));

          keyElement = null;
          l_propertyTypeArray = null;
          return;
    } catch (Exception e) {
        if (e instanceof CreateError || e instanceof UpDateError || e instanceof BaseError) {
          DialogBox.showErrorMessage(e.getCause().getMessage());
          //System.out.println("in CollectionPropertyEditor.setValue(), exception: ");
          //i.getCause().printStackTrace();
          return;
        }
        
        System.out.println("in StringMapPropertyEditor.addValueElements(), exception: "+e);
        e.printStackTrace();
    } // catch

    keyElement = null;
    l_propertyTypeArray = null;
  } // addValueElements()  
  
  
  // -------------------- PRIVATE METHODS -----------------------------------


  /**
   * Deletes some selected keys in the map, from an user click on 'Delete'.
   * Calls the removing method with a null value for each key to delete.
   * Called by actionPerformed(), which is called by the subclasses. 
   * Noop if keysUpdateInMap_ is false.
   * @param _l_keysToRemove map keys to remove with their associated values.
   */
  private void deleteKeyElements(Object[] _l_keysToRemove) {
    if (!keysUpdateInMap_) {
        DialogBox.showErrorMessage("Keys cannot be changed in this property");
        return;
    }
    if (!DialogBox.showConfirmDialog(
          " Do you want to delete these selected key(s)?", null)) return;
    
    try {
      for (int i = 0; i < _l_keysToRemove.length; i++) {
         removeMethod_.invoke(source_, new Object[] {_l_keysToRemove[i], null} );
      } // for i
    } catch (InvocationTargetException i) {
          // show the message to the user
          int lastIndex = ((InvocationTargetException)i).getTargetException().
                  toString().lastIndexOf(": ");
          if (lastIndex != 0)
            DialogBox.showErrorMessage(
              ((InvocationTargetException)i).getTargetException().toString().
                  substring(lastIndex+1));
          else DialogBox.showErrorMessage(
            ((InvocationTargetException)i).getTargetException().toString()); 

          System.out.println("in StringMapPropertyEditor.deleteKeyElements(), "+
            "the exception is an InvocationTargetException: ");
          ((InvocationTargetException)i).printStackTrace();
      
          // restores the base value in the editor
          propertyChange(new PropertyChangeEvent(source_, 
                  propertyDescriptor_.getName(), null, null));
          
    } catch (Exception e) {
          System.out.println("in StringMapPropertyEditor.deleteKeyElements(), exception: "+e);
    } // catch 

    return;
  } // deleteKeyElements()  

  /**
   * Called when the user clicks on the button 'Select' or 'NewV' to
   * add a new value in the map, among the constrained values given by the property
   * limitedValuesMethod_.
   * The type of the values is given by the BeanInfo.
   * The values are get from the method sets by the property editor key 'limitedValuesMethod'.
   * The values are limited to 66.
   * There is no control of order in the selected values, until the call to the
   * source from which an error message may be returned and displayed to the user.
   * Called by actionPerformed() in this class, and in StringTwoKeysMapPropertyEditor,
   * BaseObjectMapPropertyEditor. Overridden by AssociationMapPropertyEditor.
   * Calls addValueElements().
   * @param _key current selected key 
   */
  protected void addRestrictedValueElements(String _key) {

                          // for testing and tracking purpose
                     //System.out.println("in StringMapPropertyEditor.addRestrictedValueElements(), "+
                       //     "the property type is "+propertyTypeInGroup_);

    // this method is reserved to the constrained values
    if (limitedValuesMethod_ == null) {

                          // for testing and tracking purpose
                     //System.out.println("in StringMapPropertyEditor.addRestrictedValueElements(), "+
                       //     "there is no limited values method. ");

        return;
    }

    List<String>    l_values = null;
    ListModel       l_currentValues = null;
    List<String>    l_choice = null;
    String          choice = "";
    Object[]        l_valuesArray = null;
    int             index = 0;
    
    try {
      l_values = (List)limitedValuesMethod_.invoke(source_, new Object[] {_key} );
    } catch (InvocationTargetException i) {
          // shows the message to the user
          int lastIndex = ((InvocationTargetException)i).getTargetException().
                  toString().lastIndexOf(": ");
          if (lastIndex != 0)
            DialogBox.showErrorMessage(
              ((InvocationTargetException)i).getTargetException().toString().
                  substring(lastIndex+1));
          else DialogBox.showErrorMessage(
            ((InvocationTargetException)i).getTargetException().toString()); 

          System.out.println("in StringMapPropertyEditor.addRestrictedValueElements(), "+
            "the exception is an InvocationTargetException: ");
          ((InvocationTargetException)i).printStackTrace();
      
          // restores the current value in the editor
          propertyChange(new PropertyChangeEvent(source_, 
                  propertyDescriptor_.getName(), null, null));
          return;
    } catch (Exception e) {
        if (e instanceof CreateError || e instanceof UpDateError || e instanceof BaseError) {
          DialogBox.showErrorMessage(e.getCause().getMessage());
          //System.out.println("in CollectionPropertyEditor.setValue(), exception: ");
          //i.getCause().printStackTrace();
          return;
        }

        System.out.println("StringMapPropertyEditor.addRestrictedValueElements(), exception: "+e);
        e.printStackTrace();
    } // catch     

                                      // for testing and tracking purpose
                                 //System.out.println("StringMapPropertyEditor.addRestrictedValueElements() for "+
                                   //      _key+", the limited values are:");
                                 //for (Iterator ite = l_values.iterator(); ite.hasNext();) {
                                   //  System.out.println("       "+ite.next());
                                 //} // for ite
    
                     
    l_currentValues = ((JList)boundComponentTwo_).getModel();
    for (int i = 0; i < l_currentValues.getSize(); i++) {
        l_values.remove(l_currentValues.getElementAt(i));
    } // for i

                          // for testing and tracking purpose
                     //System.out.println("StringMapPropertyEditor.addRestrictedValueElements() for "+_key+
                       //     ", the possible new values are:");
                     //for (Iterator ite = l_values.iterator(); ite.hasNext();) {
                       //  System.out.println("       "+ite.next());
                     //} // for ite
    
    if (l_values.isEmpty()) {
        l_values = null;
        l_currentValues = null;
        DialogBox.showOkMessage("All the allowed values are selected.");
        return;
    }
    else if (l_values.size() > 66) {
        l_values = null;
        l_currentValues = null;
        DialogBox.showOkMessage("There are more than 66 allowed values.");
        return;
    }
    else if (l_values.size() > 22 && l_values.size() < 45) {
        choice = DialogBox.showSingleEditor("Select in part '1' or '2'?");
        if (choice != null && choice.equals("1"))
            l_values.subList(22, l_values.size()).clear();
        else if (choice != null)
            l_values.subList(0, 22).clear();
    }
    else if (l_values.size() > 44) {
        choice = DialogBox.showSingleEditor("Select in part '1', '2' or '3'?");
        if (choice != null && choice.equals("1"))
            l_values.subList(22, l_values.size()).clear();
        else if (choice != null && choice.equals("2")) {
            l_values.subList(44, l_values.size()).clear();
            l_values.subList(0, 22).clear();
        } else if (choice != null)
            l_values.subList(0, 44).clear();
    } // else
 
                          // for testing and tracking purpose
                     //System.out.println("StringMapPropertyEditor.addRestrictedValueElements() for "+
                       //     _key+", the part values are:");
                     //for (Iterator ite = l_values.iterator(); ite.hasNext();) {
                       //  System.out.println("       "+ite.next());
                     //} // for ite

    // default value is ''
    if (choice != null) {
        l_choice = DialogBox.showSimpleSelectionsInOneDialog(
            "the new values", l_values.toArray(new Object[0]));
    }
    else {
        l_values = null;
        l_currentValues = null;
        l_choice = null;
        choice = null;
        l_valuesArray = null;
        return;
    } // else if (choice != null)

    if (l_choice.isEmpty()) {
        l_values = null;
        l_currentValues = null;
        l_choice = null;
        choice = null;
        l_valuesArray = null;
        return;
    } // if (l_choice.isEmpty())

    l_valuesArray = (Object[]) Array.newInstance(
                                        propertyTypeInGroup_, l_choice.size());

    try {
        for (String value : l_choice) {
            
            if (propertyTypeInGroup_ == String.class)
                l_valuesArray[index] = value;
            else if (BaseObject.class.isAssignableFrom(propertyTypeInGroup_))
                l_valuesArray[index] = ACS_Facade.getEorBaseObject(
                        NameImpl.getNameFromString(value));
            else if (ImmutableName.class.isAssignableFrom(propertyTypeInGroup_))
                l_valuesArray[index] = NameImpl.getNameFromString(value);

            index++;
        } // for value
    }
    catch (BaseError b) {
        throw new InternalError("StringMapPropertyEditor.addRestrictedValueElements(): "+
                b.toString());
    } // catch

    addValueElements(l_valuesArray);
    
    l_values = null;
    l_currentValues = null;
    l_choice = null;
    choice = null;
    l_valuesArray = null;
 } // addRestrictedValueElements()
  
  
} // StringMapPropertyEditor
