//Title:        Access Road
//Version:      0.7.1
//Copyright:    Copyright 2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;


import javax.swing.*;
import javax.swing.tree.*;
import java.beans.*;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;
import java.util.Map;

import ARoad0.ErrW.ProcessError;
import ARoad0.ErrB.BaseError;
import ARoad0.gWork.ACS_Facade;
import ARoad0.Gui1.InternalFrameModel;
import ARoad0.gDMak.ActionClose;

import ARoad0.gBase.ACSImpl;
import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.VirtualFolderImpl;
import ARoad0.gBase.ACSFactoryUtilityImpl;

import ARoad0.gBaseInterface.*;

/**
 * This class is responsible for listening the events
 * fired by one ACSRun - and also the events fired by the ACSRun objects -
 * for updating the unique JTree which is associated
 * to this TreeBaseListener, and which displays the ACSRun objects.
 * Called by Gui2.ACSTree for the explorer and the own acs tree frame.
 * Handles 'OneResource', 'OneEligibleParty', 'OneExternalEligibleParty',
 * 'OneExternalResource',
 * 'l_Members', 'AclEntry', 'OneAcsRight', 'OnePrivilege', 'OneResourceVirtualFolder',
 * 'OneEPVirtualFolder', 'OneBaseObjectInFolder', 'OneTypedPrivilege', 'OneLinkedPrivilege',
 * 'OneLinkedPrivilegeSource', 'OneLinkedPrivilegeTarget' events from the base.
 * When the ACSRun is a partial ACSFactory, only 'OneAcsRight' is operational.
 * If two event changes are fired for the same removing event, this class produces
 * a NullPointException in ACSTreeUtilities. It is a good way to detect bugs! <P>
 * The creation of a new Resource which is a Directory child
 * produces 2 events 'OneResource' and 'Children'. These 2 events are fired
 * in any order. This is depending on the constructor or the copy method which is used,
 * and 'Children' is asynchronously used by the CollectionPropertyEditor.
 * It is the same logic with 'OneResourceVirtualFolder' and 'OneEPVirtualFolder',
 * which may be associated to 'Children' events when the virtual folder is deleted
 * or created. <P>
 * Each event source is registered only once.
 * The ACSRun closing operations are not managed,
 * since these operations are handled in Gui1/Gui2, in the gDMak classes
 * ActionClose and ActionOpen, and by TreeManager in Gui2.
 * The ACSRun opening and closing operations are indirectly listened
 * if it adds or removes some external oriented AclEntry.
 * @version 0.7.1
 */
public class ACSTreeBaseListenerImpl implements TreeBaseListener {
    // variables
  private static final int          PRIME = 1000003;
  private static final int          INITIAL_CAPACITY = 5; // the default value: 101
      /**
      * list of ACSObjects firing PropertyChangeEvents
      */
  private List                      l_eventSources_ = new ArrayList(INITIAL_CAPACITY);
      /**
       *  ACS associated to the JTree
       */
  private ACSRun                    acs_;
      /**
       * the JTree
       */
  private JTree                     tree_;
      /**
       * tree model of the JTree
       */
  private DefaultTreeModel          treeModel_;
      /**
       * root node of the JTree
       */
  private DefaultMutableTreeNode    root_;
  private boolean                   isExplorer_;
      /*
       * To listen or not the events in propertyChange(). Not used here.
       */
  private boolean                   listenEvents_ = true;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ACSTreeBaseListenerImpl EMPTY_INSTANCE = new ACSTreeBaseListenerImpl();

    // constructors

  /**
   * Public constructor.
  */
  public ACSTreeBaseListenerImpl() {
    acs_ = ACSImpl.EMPTY_INSTANCE;
    tree_ = new JTree();
    root_ = new DefaultMutableTreeNode();
    treeModel_ = new DefaultTreeModel(root_);
  }

 /**
 * With this constructor, the ACSTreeBaseListenerImpl is registered
 * as a listener for an ACSRun firing PropertyChangeEvents,
 * and is associated to one JTree.
 * For each GroupID, registers this instance as a listener of the member changes.
 * For each VirtualFolder, listens the 'OneBaseObjectInFolder' events.
 * For each Privilege, listens 'OneTypedPrivilege', 'OneLinkedPrivilege',
 * 'OneLinkedPrivilegeSource' and 'OneLinkedPrivilegeTarget' events.
 * @param   _source of PropertyChangeEvents; an ACS or a partial ACSFactory
 * @param   _tree associated to the instance
 * @param   _root associated to _source in _tree
 * @exception ProcessError if _source is null
 */
  public ACSTreeBaseListenerImpl(ACSRun _source, JTree _tree,
    MutableTreeNode _root) throws ProcessError {
      if (_source != null) acs_ = _source;
      else throw new ProcessError("in the ACSTreeBaseListenerImpl constructor, "+
        "the ACSRun argument is null.");
      
      //System.out.println("ACSTreeBaseListenerImpl constructor: "+acs_.getName() );
      
      tree_ = _tree;
      treeModel_ = (DefaultTreeModel) _tree.getModel();
      root_ = (DefaultMutableTreeNode)_root;
      l_eventSources_.add(acs_ );
      if (((DefaultMutableTreeNode)treeModel_.getRoot()).getUserObject().
        equals("Open Entities")) isExplorer_ = true;

      acs_.addPropertyChangeListener("OneExternalEligibleParty", this);
      acs_.addPropertyChangeListener("OneExternalResource", this);
      acs_.addPropertyChangeListener("OneResource", this);
      acs_.addPropertyChangeListener("OneEligibleParty", this);
      acs_.addPropertyChangeListener("AclEntry", this);

      // new 0.7.0
      acs_.addPropertyChangeListener("OneTypedPrivilege", this);
      acs_.addPropertyChangeListener("OneLinkedPrivilege", this);
      acs_.addPropertyChangeListener("OneResourceVirtualFolder", this);
      acs_.addPropertyChangeListener("OneEPVirtualFolder", this);

      // the single efficient property in a partial ACSFactory
      acs_.addPropertyChangeListener("OneAcsRight", this);
      
      BaseManagerImpl.getInstance().addPropertyChangeListener("OpenACS", this);

      // registers this for 'l_Members' for GroupIDs,
      // 'OneBaseObjectInFolder' for VirtualFolders,
      // 'OneLinkedPrivilegeSource' and 'OneLinkedPrivilegeTarget' for LinkedPrivileges
      if (acs_ instanceof ImmutableACS) {
          
          if (ACSTreeUtilities.getACSTreeMainNode(
            treeModel_, acs_, ACSFactoryUtilityImpl.GROUPS, isExplorer_) != null) {
            ImmutableEligibleParty ep;
            for(Iterator ite = ((ImmutableACS)acs_).getEorM_OwnEligibleParties().values().iterator();
              ite.hasNext(); ){
                ep = (ImmutableEligibleParty)ite.next();
                if (!(ep instanceof ImmutableGroupID)) continue;
                ep.addPropertyChangeListener("l_Members", this);

                     //System.out.println("ACSTreeBaseListenerImpl constructor - "+
                       //"listen 'l_Members' from the group: "+ep.getName() );

            } // for ite
            ep = null;
          } // if
          if (ACSTreeUtilities.getACSTreeMainNode(
            treeModel_, acs_, ACSFactoryUtilityImpl.RESVFOLDERS, isExplorer_) != null ||
            ACSTreeUtilities.getACSTreeMainNode(
            treeModel_, acs_, ACSFactoryUtilityImpl.EPVFOLDERS, isExplorer_) != null) {
              VirtualFolder[] l_vf = ((ImmutableACS)acs_).getEorL_VirtualFolders();
              for (int i = 0; i < l_vf.length; i++ ) {
                 l_vf[i].addPropertyChangeListener("OneBaseObjectInFolder", this);
              } // fori
              l_vf = null;
          } // if
          if (ACSTreeUtilities.getACSTreeMainNode(
            treeModel_, acs_, ACSFactoryUtilityImpl.PRIVILEGES, isExplorer_) != null ) {
              List l_pri = ((ImmutableACS)acs_).getEorL_PrivilegeForLinks();
              for (Object obj : l_pri) {
                 ((Privilege)obj).addPropertyChangeListener("OneLinkedPrivilegeSource", this);
                 ((Privilege)obj).addPropertyChangeListener("OneLinkedPrivilegeTarget", this);
              } // obj
              l_pri = null;
          } // if

      } // if (_acs instanceof

  } // ACSTreeBaseListenerImpl

    //
    // --------------- METHODS ---------------------
    //

 /**
 * Gets the main event source associated to the instance in its construction.
 * @return name of the main source for PropertyChangeEvents. May be null.
 */
  public ImmutableName getEorMainEventSource() {
    if (acs_ != null) return acs_.getName();
    else return null;
  } // getEorMainEventSource()

 /**
 * To listen or not the events in propertyChange(). No use.
 * @param _listen true to listen the events in this instance, through propertyChange()
 * @return true if the previous value was different to this argument
 */
  public synchronized boolean setListenEvents(boolean _listen) {
    boolean previous = listenEvents_;
    listenEvents_ = _listen;

    if (previous && listenEvents_) return false;
    else if (!previous && !listenEvents_) return false;
    else return true;
  } // setListenEvents()

 /**
 * Gets the TreeModel associated to the ACS tree instance in its construction.
 * @return DefaultTreeModel associated to the instance and to a JTree
 */
  public DefaultTreeModel getEorTreeModel() {
    return treeModel_;
  } // getEorTreeModel()

 /**
 * Gets the TreeModel root associated to the instance in its construction.
 * @return MutableTreeNode associated to the instance in the TreeModel
 */
  public MutableTreeNode getEorRoot() {
    return root_;
  } // getEorRoot()

 /**
 * Registers the instance as a listener for an ACSObject
 * firing PropertyChangeEvents.
 * @param _source BaseObject source of PropertyChangeEvents
 * @exception SmallError if _source is not an ACSObject or
 * doesn't belong to the right ACS, or if it is already known.
 * @return void
 */
/*  public void addEventSource(BaseObject _source) throws SmallError {
      if (!(_source instanceof ACSObject) ||
        !((ACSObject)_source).getAcsName().equals(acs_.getName()) )
          throw new SmallError("in ACSTreeBaseListenerImpl.addEventSource(_source)"+
            "_source is not an ACSObject or doesn't belong to the right ACS.");
      if (l_eventSources_.contains(_source) )
        throw new SmallError("in ACSTreeBaseListenerImpl.addEventSource(_source)"+
          "_source is already known.");
      l_eventSources_.add(_source);
      // TO DO: change the cast to ACSObject later
      ((ACSImpl)_source).addPropertyChangeListener(this);
  } // addEventSource()
*/

 /**
 * Unregister the instance as a listener for a source.
 * @param _source BaseObject registered source of PropertyChangeEvents
 * @exception SmallError if _source is unknown.
 * @return void
 */
/*  public void removeEventSource(BaseObject _source) throws SmallError {
      if (!l_eventSources_.contains(_source) )
        throw new SmallError("in ACSTreeBaseListenerImpl.removeEventSource(_source)"+
          "_source is unknown.");
      l_eventSources_.remove(_source);
      // TO DO: change the cast to ACSObject later
      ((ACSImpl)_source).removePropertyChangeListener(this);
  } // removeEventSource()
*/

  /**
   * Receives the source events about property changes on 
   * 'OneExternalEligibleParty', 'OneExternalResource',
   * 'OneResource', 'OneEligibleParty', 'l_Members', 'AclEntry', 'OneAcsRight',
   * 'OnePrivilege', 'OneResourceVirtualFolder', 'OneEPVirtualFolder',
   * 'OneBaseObjectInFolder',  'OneTypedPrivilege', 'OneLinkedPrivilege',
   * 'OneLinkedPrivilegeSource', 'OneLinkedPrivilegeTarget'.
   * Throw InternalError if both the old and the new values are non null,
   * or if the two are null.
   * Calls the appropriate private method.
   * @param _evt PropertyChangeEvent
   */
  public synchronized void propertyChange(PropertyChangeEvent _evt) {

        // for testing and tracking
        //System.out.println("ACSTreeBaseListenerImpl: "+
            //"call to propertyChange() of "+acs_+" by the source: "+_evt.getSource().toString()+
            //" for the property "+_evt.getPropertyName() );

    // to avoid the processing if this instance is empty or has been finalized
    if (acs_ == null || acs_ == ACSImpl.EMPTY_INSTANCE) return;

    // the two values may be non null when there are arrays, collections or maps
    if (_evt.getOldValue() == null && _evt.getNewValue() == null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to propertyChange() - both old and new values are null.");
    }

    // for 'OneExternalEligibleParty'
    if (_evt.getPropertyName().equals("OneExternalEligibleParty") ) {
          changeExternalEligibleParty(_evt);
          return;
    } 

    // for 'OneExternalResource'
    if (_evt.getPropertyName().equals("OneExternalResource") ) {
          changeExternalResource(_evt);
          return;
    }

    // for 'OneResource'
    else if (_evt.getPropertyName().equals("OneResource") ) {
          changeOneResource(_evt);
          return;
    } // if "OneResource"

    // for 'OneEligibleParty'
    else if (_evt.getPropertyName().equals("OneEligibleParty") ) {
          changeOneEligibleParty(_evt);
          return;
    } // if "OneEligibleParty"

    // for 'l_Members'
    else if (_evt.getPropertyName().equals("l_Members") ) {
          changeGroupIDMembers(_evt);
          return;
    } // if "OneEligibleParty"

    // for 'OpenACS' is for closing ACS only
    else if (_evt.getPropertyName().equals("OpenACS") ) {
          changeOpenACS(_evt);
          return;
    }

    // for 'AclEntry'
    else if (_evt.getPropertyName().equals("AclEntry") ) {
          changeOneAclEntry(_evt);
          return;
    } 

    // for 'OneAcsRight'
    else if (_evt.getPropertyName().equals("OneAcsRight") ) {
          changeOneAcsRight(_evt);
          return;
    } 

    // for 'OneTypedPrivilege', 'OneLinkedPrivilege'
    else if (_evt.getPropertyName().equals("OneTypedPrivilege") ||
                _evt.getPropertyName().equals("OneLinkedPrivilege")) {
          changeOnePrivilege(_evt);
          return;
    }

    // 'OneResourceVirtualFolder'
    else if (_evt.getPropertyName().equals("OneResourceVirtualFolder") ) {
          changeOneResourceVirtualFolder(_evt);
          return;
    }

    // 'OneEPVirtualFolder' and 'OneBaseObjectInFolder' for members
    else if (_evt.getPropertyName().equals("OneEPVirtualFolder") ) {
          changeOneEPVirtualFolder(_evt);
          return;
    }

    // 'OneBaseObjectInFolder' for VirtualFolder members
    else if (_evt.getPropertyName().equals("OneBaseObjectInFolder") ) {
          changeOneBaseObjectInFolder(_evt);
          return;
    }

  } // propertyChange()

  /**
   * Processes event from ACS about 'OneExternalEligibleParty' updatings
   * for objects which belong to another ACS than acs_.
   * @param _evt PropertyChangeEvent
   */
  private void changeExternalEligibleParty(PropertyChangeEvent _evt){
      
                    // for testing and tracking
                    //System.out.println("ACSTreeBaseListenerImpl: "+
                      //  "call to changeExternalEligibleParty() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeExternalEligibleParty() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    ImmutableEligibleParty ep;

    // if the new value is non-null,
    // it is a new EligibleParty to insert in the tree.
    if (_evt.getNewValue() != null) {
      ep = (ImmutableEligibleParty)_evt.getNewValue();
      // no action if the event is for acs_
      if (acs_.getName().equals(ep.getAcsName())) return;

               // for testing and tracking
              // System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
              //  ", ep to insert: "+ep);

      modifyExternalEpInTree(ep, true);
    }// if (_evt.getNewValue() != null)

    // if the old value is non-null,
    // it is an aclEntry to remove from the tree.
    else if (_evt.getOldValue() != null) {
      ep = (ImmutableEligibleParty)_evt.getOldValue();
      // no action if the event is for acs_
      if (acs_.getName().equals(ep.getAcsName())) return;

        // for testing and tracking
       //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
       // ", external EligibleParty to remove: "+ep);

      modifyExternalEpInTree(ep, false);
    }// if (_evt.getOldValue() != null)

    ep = null;
  } // changeExternalEligibleParty

  /**
   * Processes event from ACS about 'OneExternalResource' updatings
   * for objects which belong to another ACS than acs_.
   * @param _evt PropertyChangeEvent
   */
  private void changeExternalResource(PropertyChangeEvent _evt){

                        // for testing and tracking
                        //System.out.println("ACSTreeBaseListenerImpl: "+
                          //  "call to changeExternalResource() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeExternalResource() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    ImmutableResource       res;

    // if the new value is non-null,
    // it is a new aclEntry to insert in the tree.
    if (_evt.getNewValue() != null) {
      res = (ImmutableResource)_evt.getNewValue();
      // no action if the event is for acs_
      if (acs_.getName().equals(res.getAcsName())) return;

               // for testing and tracking
              // System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
              //  ", resource to insert: "+res);

      modifyExternalResourceInTree(res, true);
    }// if (_evt.getNewValue() != null)

    // if the old value is non-null,
    // it is an aclEntry to remove from the tree.
    else if (_evt.getOldValue() != null) {
      res = (ImmutableResource)_evt.getOldValue();
      // no action if the event is for acs_
      if (acs_.getName().equals(res.getAcsName())) return;

                // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
               // ", external resource to remove: "+res);

      modifyExternalResourceInTree(res, false);
    }// if (_evt.getOldValue() != null)

    res = null;
  } // changeExternalResource

  /**
   * Processes event from ACSRun about 'OneResource' updates.
   * Doesn't control the new value.
   * Processes actor creation, but do not process actor deletion which is
   *  the responsability of changeOneEligibleParty(), before this method call.
   * Does not process the external resource updates, which is processed
   *  through the 'ExternalOrientedAclEntry' event.
   * @see #changeOneEligibleParty(java.beans.PropertyChangeEvent)
   * @see ARoad0.Gui2.ACSTreeUtilities
   * @param   _evt PropertyChangeEvent
   */
  private void changeOneResource(PropertyChangeEvent _evt){
    // for testing and tracking
    //if (_evt.getOldValue() != null)
      //System.out.println("ACSTreeBaseListenerImpl: "+
        //"call to changeOneResource() for a removing of: "+_evt.getOldValue());

    //if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      //throw new InternalError("ACSTreeBaseListenerImpl: "+
        //"call to changeOneResource() with both old and new values"+
        //" which are not null.");
    //} // if

    // variables
    ImmutableResource       res = (ImmutableResource)_evt.getNewValue();
    DefaultMutableTreeNode  node;
    DefaultMutableTreeNode  rootNode;
    DefaultMutableTreeNode  resourceNode;

    // if the new value is non-null,
    // it is a new resource to insert in the tree.
    if (res != null) {

       // for testing and tracking
       //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
        //", resource to insert: "+res);

      // create the resource in ACSTree.RESOURCES
      if (res.getEorACS() == acs_) {
        rootNode = ACSTreeUtilities.getACSTreeMainNode(
          treeModel_, acs_, ACSFactoryUtilityImpl.RESOURCES, isExplorer_);

                  // for testing and tracking
                //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //  ", ACS tree main node for res: "+rootNode);

        if (rootNode != null) {
          node = ACSTreeUtilities.getParentOfResource(
            (ImmutableACS)acs_, treeModel_, res, isExplorer_);
          if (node != null) treeModel_ = ACSTreeUtilities.
            modifyObjectUnderNode(treeModel_, node, res, true);
        } // if
      } // if
      // else, do not create the external resource in ACSTree.ECOTHER_RESOURCES
      // since a 'ExternalOrientedAclEntry' event will be fired

    }// if (_evt.getNewValue() != null)

    // if the old value is non-null,
    // it is a resource to remove from the tree.
    else {
      res = (ImmutableResource)_evt.getOldValue();

        // for testing and tracking
      //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
        //", res to remove: "+res);

      // Actor finalization is declared by its ACS, but not when the actor
      // has not been properly built up by its constructor which has finalized the instance.
      if (res instanceof ImmutableActor &&
              res.getActorActivated()) {
        res = null;
        node = null;
        rootNode = null;
        return;
      }

      // deletes the resource in ACSTree.RESOURCES
      // can't use res.getEorACS() because the resource has been finalized
      if (res.getAcsName().equals( acs_.getName() )) {
        rootNode = ACSTreeUtilities.getACSTreeMainNode(
          treeModel_, acs_, ACSFactoryUtilityImpl.RESOURCES, isExplorer_);

                  // for testing and tracking
                //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //  ", ACS tree main node for res: "+rootNode);

        if (rootNode != null) {
          node = ACSTreeUtilities.getCurrentParent(rootNode, res);

                  // for testing and tracking
                //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //  ", node of res: "+node);

          treeModel_ = ACSTreeUtilities.
          modifyObjectUnderNode(treeModel_, node, res, false);
        } // if
      } // if

    }// if (_evt.getOldValue() != null)

    res = null;
    node = null;
    rootNode = null;
  } // changeOneResource

  /**
   * Processes event from ACSRun about 'OneEligibleParty' updates.
   * Doesn't control the new value.
   * Processes also Actors updates, but does not process the external EligibleParty
   * updatings.
   * If it is a GroupID to delete, unregisters this instance as a listener
   * of member changes. The registration is done by ActionNewEligibleParty.
   * If it is a to-display typed GroupID, manages its position under the right second node.
   * When the EligibleParty ep is an userID or a groupID to delete or to add,
   * this method used ep.getEorL_GroupID() to define the groups to update,
   * then calls changeOwnEPasMemberOfGroupIDs() which removes all these groups in
   * the EligibleParty. No action if a relevant main node (as ACSTree.GROUPS)
   * is not displayed.
   * @see     ACSTreeBaseListenerImpl#changeOneResource
   * @param   _evt about an 'OneEligibleParty' update
   */
  private void changeOneEligibleParty(PropertyChangeEvent _evt){
              
                // for testing and tracking
              //System.out.println("ACSTreeBaseListenerImpl in "+toString()+
                //  " - call to changeOneEligibleParty() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeOneEligibleParty() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    ImmutableEligibleParty  ep;
    DefaultMutableTreeNode  node;
    DefaultMutableTreeNode  rootNode;
    boolean                 toInsert = false;
    Object                  object = null;
    String                  toDisplayGroupType = null;
    String                  groupType = null;
    
    // if the new value is non-null,
    // it is a new eligible party to insert in the tree.
    if (_evt.getOldValue() == null) {
      toInsert = true;
      ep = (ImmutableEligibleParty)_evt.getNewValue();
      //if (ep instanceof ImmutableGroupID) {

              
                // for testing and tracking
              //System.out.println("ACSTreeBaseListenerImpl: in "+toString()+
                  //" - call to addPropertyChangeListener('l_Members', this) for "+ep.getName());
          
        // this registration is not reliable, because sometimes,
        // ep has not built up its transient variables
        // since its contruction is not finished.
        // This is due to the fact that ACSImpl fires a property change
        // when the EligiblePartyImpl constructor calls the ACSImpl newEligibleParty() method.
        // So, the true registration is done by ActionNewEligibleParty
        // when the new instance is fully created.
        //ep.addPropertyChangeListener("l_Members", this);
      //}
    }
    else {
      ep = (ImmutableEligibleParty)_evt.getOldValue();
      if (ep instanceof ImmutableGroupID) {
              
                // for testing and tracking
              //System.out.println("ACSTreeBaseListenerImpl: in "+toString()+
                  //" - call to removePropertyChangeListener('l_Members', this) for "+ep.getName());
          
          ep.removePropertyChangeListener("l_Members", this);
      }
    } // else

       // for testing and tracking
       //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
       // ", eligible party to insert or remove: "+ep);

    // can't use ep.getEorACS() because sometimes the ep has been finalized
    if (!ep.getAcsName().equals( acs_.getName() )) {
        rootNode = null;
        node = null;
        object = null;
        toDisplayGroupType = null;
        groupType = null;

                    // for testing and tracking
                //DialogBox.showErrorMessage("ACSTreeBaseListenerImpl.changeOneEligibleParty()"+
                  //      " with an EP: "+ep.getName()+" from another ACS than: "+acs_.getName());

        ep = null;
        return;        
    } // if

    // for actor
    if (ep instanceof ImmutableActor)
        changeOneOwnActor((ImmutableActor)ep, toInsert);

    // an userID has to be updated under ACSTree.USERS
    // and in all the groupIDs which own it
    else if (ep instanceof ImmutableUserID ) {
        rootNode = ACSTreeUtilities.getACSTreeMainNode(
          treeModel_, acs_, ACSFactoryUtilityImpl.USERS, isExplorer_);
        if (rootNode != null) {
          treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
            treeModel_, rootNode, ep, toInsert);
        } // if
        changeOwnEPasMemberOfGroupIDs(ep, toInsert);
    } // if

    // a groupID has to be updated as a child of ACSTree.GROUPS
    // and in all the groupIDs where it is member
    else if (ep instanceof ImmutableGroupID) {
        rootNode = ACSTreeUtilities.getACSTreeMainNode(
          treeModel_, acs_, ACSFactoryUtilityImpl.GROUPS, isExplorer_);
        if (rootNode != null) {

              // sets node in testing sub-or-tree-typed group nodes under rootNode
            if (ep.getEPType().startsWith("<byAcsCreator> ")) {
                  node = ACSTreeUtilities.getParentOfSubOrTreeTypedGroupID(
                            rootNode, (ImmutableGroupID)ep);
                  /*
                  for (int j = 0; j < rootNode.getChildCount(); j++) {

                      object = ((DefaultMutableTreeNode)rootNode.getChildAt(j)).getUserObject();
                      if (!(object instanceof String)) continue;
                      toDisplayGroupType = (String) object;
                      if (toDisplayGroupType.length() < 9 || !toDisplayGroupType.endsWith("_tree>")) continue;
                            // '1' for "<" at the start
                      groupType = toDisplayGroupType.substring(1, toDisplayGroupType.indexOf("_tree>"));
                      if (ep.getEPType().length() >= 24 + groupType.length() &&
                              ep.getEPType().startsWith("<byAcsCreator> "+groupType)) {
                        node = (DefaultMutableTreeNode) rootNode.getChildAt(j);
                        break;
                      } // if          

                  } // for j
                   * */
                  if (node != null)
                      treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                                    treeModel_, node, ep, toInsert);
            } // if

            else treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                                    treeModel_, rootNode, ep, toInsert);
        } // if
        
        if (!toInsert) changeOwnEPasMemberOfGroupIDs(ep, false);
    } // if

    ep = null;
    rootNode = null;
    node = null;
    object = null;
    toDisplayGroupType = null;
    groupType = null;
  } // changeOneEligibleParty

  /**
   * Updates the ACS tree when one actor is created or removed.
   * The update for a creation in the 'Resources' node has already be done
   * through a 'OneResource' event.
   * The actor can't be a direct member of several groupIDs.
   * Called by changeOneEligibleParty().
   * No action if ACSTree.RESOURCES and ACSTree.ACTORS
   * are not displayed.
   * For this method, acs_ may be cast as ImmutableACS.
   * <p>
   * @param _ep the instance created or removed.
   * @param _toInsert true if the instance is created
   */
  private void changeOneOwnActor(ImmutableActor _ep, boolean _toInsert){

    // variables
    DefaultMutableTreeNode  node;
    DefaultMutableTreeNode  rootNode;

           // for testing and tracking
          //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
          //", call to changeOneOwnActor() to insert/delete ("+_toInsert+") the ep: "+_ep+
          //"with the acs: "+_ep.getEorACS().getName());

    // it is an actor, then create or remove _ep in ACSTree.RESOURCES.
    // if it already exists and it is a creation, no operation
    if ((ACSTreeUtilities.getACSTreeMainNode(treeModel_, acs_, ACSFactoryUtilityImpl.RESOURCES,
      isExplorer_) != null)) {

          if (_toInsert) {
            node = ACSTreeUtilities.getParentOfResource(
              (ImmutableACS)acs_, treeModel_, (ImmutableResource)_ep, isExplorer_);
            if (ACSTreeUtilities.getCurrentObjectNode( node, _ep) == null) {
              treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                treeModel_, node, _ep, _toInsert);
            }
            // else, no operation since _ep is there
          } else {
            node = ACSTreeUtilities.getCurrentParent(
              ACSTreeUtilities.getACSTreeMainNode(
                treeModel_, acs_, ACSFactoryUtilityImpl.RESOURCES, isExplorer_),
                _ep);

            if (node != null) treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                treeModel_, node, _ep, _toInsert);

                   // for testing and tracking
                  //System.out.println("ACSTreeBaseListenerImpl, for the acs "+acs_+
                  //", call to changeOneOwnActor() to delete the ep: "+_ep+" ...");
                  //if (node != null) System.out.println("-------------------- ... under the node:"+node.getUserObject());
                  
          } // else
          
    } // if

    // returns if ACSTree.ACTORS is not displayed
    if (
      ACSTreeUtilities.getACSTreeMainNode(treeModel_, acs_, ACSFactoryUtilityImpl.ACTORS, isExplorer_)
        == null
    ) {
      node = null;
      rootNode = null;
      return;
    }

    // as an actor, creates or removes _ep in ACSTree.ACTORS
    if (_toInsert) rootNode = ACSTreeUtilities.getACSTreeMainNode(
      treeModel_, acs_, ACSFactoryUtilityImpl.ACTORS, isExplorer_);
    else rootNode = ACSTreeUtilities.getCurrentParent(
      ACSTreeUtilities.getACSTreeMainNode(
        treeModel_, acs_, ACSFactoryUtilityImpl.ACTORS, isExplorer_),
      _ep);
    if (rootNode != null) treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
      treeModel_, rootNode, _ep, _toInsert);

    node = null;
    rootNode = null;
  } // changeOneOwnActor

  /**
   * Processes event from ACSRun about 'AclEntry' updates.
   * May be an internal or an external-oriented AclEntry.
   * Doesn't control the value of a new AclEntry.
   * No action if it is not an acs_ AclEntry.
   * For removing, the AclEntry in _evt is a clone of the AclEntry to remove.
   * @param _evt PropertyChangeEvent
   */
  private void changeOneAclEntry(PropertyChangeEvent _evt){
    // for testing and tracking
    //System.out.println("in ACSTreeBaseListenerImpl: "+
      //  "call to changeOneAclEntry() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeOneAclEntry() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    ImmutableAclEntry       acl;
    boolean                 toInsert;
    DefaultMutableTreeNode  aclNode = ACSTreeUtilities.getACSTreeMainNode(
                                treeModel_, acs_, 
                                (String)ACSTree.M_ACS_TREE_UPPER_NODES.get("ACLENTRIES"), 
                                isExplorer_);
    DefaultMutableTreeNode  node = null;
    Object                  object = null;
    String                  toDisplayType = null;
    String                  type = null;
    String                  resType = null;  
    
    
    if (aclNode == null) {
        
                // for testing and tracking
                //System.out.println("in ACSTreeBaseListenerImpl.changeOneAclEntry(), "+
                  //"the acl node in the explorer is null");
        return;
    }

    // if the new value is non-null,
    // it is a new AclEntry to insert in the tree.
    if (_evt.getNewValue() != null) {

      acl = (ImmutableAclEntry)_evt.getNewValue();
      // no action if the event is not for acs_, and use getAcsName() since acl is a clone
      if (!acs_.getName().equals(acl.getAcsName())) {
        
                // for testing and tracking
                //System.out.println("in ACSTreeBaseListenerImpl.changeOneAclEntry(), "+
                  //"with toInsert = true, the acl is not from "+acs_);
          
          aclNode = null;
          return;
      }

      toInsert = true;
    } // if
    else {

      acl = (ImmutableAclEntry)_evt.getOldValue();
      // no action if the event is not for acs_
      if (!acs_.getName().equals(acl.getAcsName())) {
        
                // for testing and tracking
                //System.out.println("in ACSTreeBaseListenerImpl.changeOneAclEntry(), "+
                  //"with toInsert = false, the acl is not from "+acs_);
          
          aclNode = null;
          return;
      }

      toInsert = false;
    } // else

    // for testing and tracking
    //System.out.println("in ACSTreeBaseListenerImpl."+
      //  "changeOneAclEntry(), call to ACSTreeUtilities.modifyObjectUnderNode() "+
        //"with toInsert: "+toInsert);
    
    resType = acl.getEorResource().getResourceType();

      // searches node, if any, in testing typed ACL nodes under the ACL node
    if (resType.startsWith("<byAcsCreator> ")) {
          node = null;
          for (int j = 0; j < aclNode.getChildCount(); j++) {

              object = ((DefaultMutableTreeNode)aclNode.getChildAt(j)).getUserObject();
              if (!(object instanceof String)) continue;
              toDisplayType = (String) object;
              if (toDisplayType.length() < 9 || !toDisplayType.endsWith("_subacs>")) continue;
                    // '1' for "<" at the start
              type = toDisplayType.substring(1, toDisplayType.indexOf("_subacs>"));

                    // for testing and tracking
                    //System.out.println("in ACSTreeBaseListenerImpl.changeOneAclEntry(): comparison on ");
                    //System.out.println("resType: "+resType);
                    //System.out.println("type: "+type);

              if (//resType.length() >= 24 + type.length() &&
                      resType.startsWith("<byAcsCreator> "+type)) {
                node = (DefaultMutableTreeNode) aclNode.getChildAt(j);
                break;
              } // if

          } // for j 
          if (node != null) aclNode = node;
    } // if    
    // handles the 'AclEntry' flat list in the tree;
    // inserts (following the order of the AclEntry name, not the detailled name order)
    // or removes _acl at the fifth position
    if (!toInsert)
        treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
                        aclNode, acl.getDetailledName(), toInsert);
    else
        treeModel_ = ACSTreeUtilities.addDetailledNameUnderNode(treeModel_,
                        aclNode, acl, (ACS) acs_);

    aclNode = null;
    acl = null;
    node = null;
    object = null;
    toDisplayType = null;
    type = null;
    resType = null;
  } // changeOneAclEntry

  /**
   * Processes event from ACSRun about 'OneAcsRight' updates.
   * Doesn't control the new value.
   * No action if a relevant main node (as ACSTree.ACSRIGHTS)
   * is not displayed.
   * @see     ACSTreeBaseListenerImpl#changeOneResource
   * @param   _evt PropertyChangeEvent
   */
  private void changeOneAcsRight(PropertyChangeEvent _evt){
              
                // for testing and tracking
              //System.out.println("ACSTreeBaseListenerImpl: in "+toString()+
                  //" - call to changeOneAcsRight() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeOneAcsRight() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    StringRight             right;
    DefaultMutableTreeNode  rootNode;
    boolean                 toInsert = false;

    // if the new value is non-null,
    // it is a new eligible party to insert in the tree.
    if (_evt.getOldValue() == null) {
      toInsert = true;
      right = (StringRight)_evt.getNewValue();
    } 
    else {
      right = (StringRight)_evt.getOldValue();
    } // else

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", right "+right+" to insert: "+toInsert);


    if (!right.getAcsName().equals( acs_.getName() )) {
        right = null;
        rootNode = null;
        
               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", right has not the correct ACS: "+right.getAcsName());
        
        return;
    }

    // a right has to be updated under ACSTree.ACSRIGHTS
        // if the right has been finalized, the metaright has unregistered it
    rootNode = ACSTreeUtilities.getACSTreeMainNode(
      treeModel_, acs_, ACSFactoryUtilityImpl.ACSRIGHTS, isExplorer_);
    if (rootNode != null) {
      treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
        treeModel_, rootNode, right, toInsert);
    } // if
    else {
               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", NOOP since rootNode is null ");
    }

    right = null;
    rootNode = null;
  } // changeOneAcsRight
  
  /**
   * Updates the ACS tree when one userID or groupID is created or removed,
   * and when it is a direct member of one or several groupIDs.
   * Each of these groupIDs may be associated to several nodes in the tree,
   * so they are all removed then built up.
   * Called by the changeOneEligibleParty method.
   * Calls ACSTreeUtilities.buildUpGroupTree().
   * No action if ACSTree.GROUPS is not displayed.
   * @param   _ep the instance created or removed
   * @param _toInsert true if the instance is created
   */
  private void changeOwnEPasMemberOfGroupIDs(ImmutableEligibleParty _ep,
    boolean _toInsert){
    // variables
    DefaultMutableTreeNode  rootNode;
    DefaultMutableTreeNode  rootNodeTwo;
    DefaultMutableTreeNode  groupsNode;
    GroupID                 groupID;
    Enumeration             enumeration;
    Enumeration             enumerationTwo;
    
    // tests the ACSTree.GROUPS node
    groupsNode = ACSTreeUtilities.getACSTreeMainNode(
      treeModel_, acs_, ACSFactoryUtilityImpl.GROUPS, isExplorer_);
    if (groupsNode == null) return;

    // first enumeration to select the nodes of groupID which contains
    // the ep as indirect member, and to remove all their children
    enumeration = groupsNode.children();
    while (enumeration.hasMoreElements()) {

      rootNode = (DefaultMutableTreeNode) enumeration.nextElement();
      if (rootNode.getUserObject() instanceof GroupID) {
          groupID = (GroupID)rootNode.getUserObject();

                   // for testing and tracking
                  //System.out.println("in ACSTreeBaseListenerImplchangeOwnEPasMemberOfGroupIDs(), for"+
                  //  acs_.getName()+", remove children under the node of: "+groupID);

          // removes rootNode children, whatever _toInsert, then build up the children
          ACSTreeUtilities.removeAllChildren(treeModel_, rootNode);
          treeModel_ = ACSTreeUtilities.buildUpGroupTree((ACS)acs_,
                treeModel_, (DefaultMutableTreeNode)rootNode, groupID );
      } else if (rootNode.getUserObject() instanceof String) { 
          // second enumeration to select the groupIDs under the second nodes
          enumerationTwo = rootNode.children();
          while (enumerationTwo.hasMoreElements()) {
              rootNodeTwo = (DefaultMutableTreeNode) enumerationTwo.nextElement();
              groupID = (GroupID)rootNodeTwo.getUserObject();

                       // for testing and tracking
                      //System.out.println("in ACSTreeBaseListenerImplchangeOwnEPasMemberOfGroupIDs(), for"+
                      //  acs_.getName()+", remove children under the node of: "+groupID);

              // removes rootNodeTwo children, whatever _toInsert, then build up the children
              ACSTreeUtilities.removeAllChildren(treeModel_, rootNodeTwo);
              treeModel_ = ACSTreeUtilities.buildUpGroupTree((ACS)acs_,
                    treeModel_, (DefaultMutableTreeNode)rootNodeTwo, groupID );
          } // while (enumerationTwo       
      } // else

    } // while (enumeration

    rootNode = null;
    rootNodeTwo = null;
    groupsNode = null;
    groupID = null;
    enumeration = null;
    enumerationTwo = null;
  } // changeOwnEPasMemberOfGroupIDs

  /**
   * Processes events from BaseManagerImpl about 'openACS' updates only for ACS closings.
   * Used if there is some ACS closings after an user command to close an ancestor ACS.
   * @param _evt PropertyChangeEvent
   */
  private void changeOpenACS(PropertyChangeEvent _evt){
    Map             newOpenACS = (Map)_evt.getNewValue();
    Map             oldOpenACS = (Map)_evt.getOldValue();

               // for testing and tracking
              //System.out.println("Call of ACSTreeBaseListenerImpl.changeOpenACS(), for"+
              //  acs_.getName());

    JInternalFrame  frame;
    ImmutableName   name;

    for (Object obj : oldOpenACS.keySet()) {
        if (newOpenACS.keySet().contains(obj)) continue;
        name = (ImmutableName) obj;

                      // for testing and tracking
                    //System.out.println("in ACSTreeBaseListenerImpl.changeOpenACS(),"+
                      //      " call ActionClose.closeInGUI() for "+name);

        frame = (JInternalFrame)InternalFrameModel
                                      .getMainFramesOfBaseObjects().get(name);

        // closes in the GUI
        if (frame != null) ActionClose.getInstance().closeInGUI(name, frame);
        else ActionClose.getInstance().closeInGUI(name, null); // if it is open only in the explorer
    } // for name

    frame = null;
    name = null;
    newOpenACS = null;
    oldOpenACS = null;
  }  // changeOpenACS()

  /**
   * Inserts ou removes an external EligibleParty in treeModel_,
   * under the 3 subtrees which handle the external eligible parties.
   * Does not modify the flat list of AclEntries or Privileges in the tree.
   * No operation under a subtree if the 'External Eligible Parties' subtree
   * is not displayed.
   * Called by changeExternalEligibleParty().
   * @param _ep to insert or to remove
   * @param _toInsert true to insert, false to remove
   */
  private void modifyExternalEpInTree(ImmutableEligibleParty _ep,
      boolean _toInsert) {
    // variables
    DefaultMutableTreeNode node = null;
    DefaultMutableTreeNode eepNode;

                // for testing and tracking
             //System.out.println("in ACSTreeBaseListenerImpl.modifyExternalEpInTree() -"+
               // " the EP is "+_ep.getName());

    // handles the 'External Eligible Parties' subtree, if it exists
    eepNode = ACSTreeUtilities.getACSTreeMainNode(
        treeModel_, acs_, ACSFactoryUtilityImpl.EEPARTIES, isExplorer_);

    if (eepNode == null) {
      return;
    }

    if (_toInsert) {

      // if the 3 subtrees doesn't exist, create them
      if (eepNode.isLeaf() )
          treeModel_ = ACSTree.getExtEPSubTreeRoots(treeModel_, eepNode);
      // insert the external eligible party under the right subtree
      if (_ep instanceof ImmutableActor) {
        node = (DefaultMutableTreeNode)treeModel_.getChild(eepNode, 0);
        treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
          node, _ep.getName(), true);
      }
      else if (_ep instanceof ImmutableUserID) {
        node = (DefaultMutableTreeNode)treeModel_.getChild(eepNode, 1);
        treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
          node, _ep.getName(), true);
      }
      else if (_ep instanceof ImmutableGroupID) {
        node = (DefaultMutableTreeNode)treeModel_.getChild(eepNode, 2);
        treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
          node, _ep.getName(), true);
      }

    } // if (_toInsert)
    else {
        
      if (_ep instanceof ImmutableActor) {
        node = (DefaultMutableTreeNode)treeModel_.getChild(eepNode, 0);;
        treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
          node, _ep.getName(), false);
      }
      else if (_ep instanceof ImmutableUserID) {
        node = (DefaultMutableTreeNode)treeModel_.getChild(eepNode, 1);
        treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
          node, _ep.getName(), false);
      }
      else if (_ep instanceof ImmutableGroupID) {
        node = (DefaultMutableTreeNode)treeModel_.getChild(eepNode, 2);
        treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
          node, _ep.getName(), false);
      }

    } // if (!_toInsert)

    eepNode = null;
    node = null;
} // modifyExternalEpInTree()

  /**
   * Inserts ou removes an external Resource in treeModel_,
   * under the 5 subtrees which handle it the external resources.
   * Does not modify the flat list of AclEntries or Privileges in the tree.
   * No operation under a subtree if the 'External Controlled Resources' subtree
   * is not displayed in the ACS tree.
   * Called by changeExternalResource().
   * @param _res external resource to insert or to remove
   * @param _toInsert true to insert, false to remove
   */
  private void modifyExternalResourceInTree(ImmutableResource _res,
      boolean _toInsert) {
    // variables
    DefaultMutableTreeNode node = null;
    DefaultMutableTreeNode ecrNode;
    

                // for testing and tracking
             //System.out.println("in ACSTreeBaseListenerImpl.modifyExternalResourceInTree() -"+
               // " the resources name is "+_res.getName());

      // handles the 'External Controlled Resources' subtree, if it exists
      ecrNode = ACSTreeUtilities.getACSTreeMainNode(
        treeModel_, acs_, ACSFactoryUtilityImpl.ECRESOURCES, isExplorer_);

      if (ecrNode == null ) {
        return;
      } // if

      if (_toInsert) {

          // if the children doesn't exist, create them
        if (ecrNode.isLeaf())
            treeModel_ = ACSTree.getExtContResourceSubTreeRoots(
                treeModel_, ecrNode);
          // handle the 'External Controlled Actors' subtree
          // get the node and insert the actor name
        if (_res instanceof ImmutableActor) {
          node = (DefaultMutableTreeNode)treeModel_.getChild(ecrNode, 0);
          treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
            node, _res.getName(), true);
        }
          // handle the 'Other External Controlled Resources' subtree
        else {
          node = (DefaultMutableTreeNode)treeModel_.getChild(ecrNode, 1);
          treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
            node, _res.getName(), true);
        }

      } // if (_toInsert)
      else {

          if (_res instanceof ImmutableActor) {
            node = (DefaultMutableTreeNode)treeModel_.getChild(ecrNode, 0);
            treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
              node, _res.getName(), false);
          }
          // handles the 'Other External Controlled Resources' subtree
          else {
            node = (DefaultMutableTreeNode)treeModel_.getChild(ecrNode, 1);
            treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(treeModel_,
              node, _res.getName(), false);
          } //else          

      } // else

    ecrNode = null;
    node = null;
} // modifyExternalResourceInTree()

  /**
   * Processes event from ACSRun about 'l_Members' updates for one GroupID.
   * Calls modifyGroupIDMembers().
   * No action if a relevant main node (as ACSTree.GROUPS)
   * is not displayed.
   * @see     ACSTreeBaseListenerImpl#changeOneResource
   * @param   _evt PropertyChangeEvent
   */
  private void changeGroupIDMembers(PropertyChangeEvent _evt){

                // for testing and tracking
                //System.out.println("ACSTreeBaseListenerImpl: "+
                  //  "call to changeGroupIDMembers() ");

    // variables
    Object                  obj;
    ImmutableGroupID        grp;
    ImmutableGroupID[]      l_groups;
    Set                     l_parentGroups = new HashSet(INITIAL_CAPACITY);
    Set                     l_newParentGroups = new HashSet(INITIAL_CAPACITY);
    Set                     l_processedGroups = new HashSet(INITIAL_CAPACITY);

    obj = _evt.getSource();
    if (obj instanceof ImmutableGroupID) {
      grp = (ImmutableGroupID)obj;
      l_groups = grp.getEorL_GroupID();
    }
    else throw new InternalError("in ACSTreeBaseListenerImpl.changeGroupIDMembers(),"+
      " the event source is not a GroupID.");

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", GroupID with new members: "+grp);

    modifyGroupIDMembers(grp);

    if (l_groups.length == 0) {
      obj = null;
      grp = null;
      l_groups = null;
      return;
    } else {
      for (int i = 0; i < l_groups.length; i++) {
        if (l_groups[i] != null) l_parentGroups.add(l_groups[i]);
      } // for i
    } // else

    while (!l_parentGroups.isEmpty()) {

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", l_parentGroups.isEmpty() is true");
       
      for (Iterator ite = l_parentGroups.iterator(); ite.hasNext(); ) {
        grp = (ImmutableGroupID) ite.next();
        modifyGroupIDMembers(grp);
        l_processedGroups.add(grp);

        l_groups = grp.getEorL_GroupID();
        if (l_groups.length != 0) {
          for (int i = 0; i < l_groups.length; i++) {
            if (l_groups[i] != null && !l_processedGroups.contains(l_groups[i]))
              l_newParentGroups.add(l_groups[i]);
          } // for i
        }
      } // for ite

      l_parentGroups.clear();
      l_parentGroups = (Set)((HashSet)l_newParentGroups).clone();
      l_newParentGroups.clear();
    } // while

    obj = null;
    grp = null;
    l_groups = null;
    l_parentGroups = null;
    l_newParentGroups = null;
    l_processedGroups = null;
  } // changeGroupIDMembers

  /**
   * Updates the ACS tree when a groupID has members, through the selection
   * of the first node that has the argument as user object.
   * Does not process all the groupID tree under the node that is associated to _grp.
   * If it is a to-display typed GroupID, manage its position under the right second node.
   * Called by changeGroupIDMembers().
   * Calls ACSTreeUtilities.getACSTreeMainNode() and buildUpGroupTree().
   * No action if ACSTree.GROUPS is not displayed.
   * @param _grp the instance with some members
   */
  private void modifyGroupIDMembers(ImmutableGroupID _grp){
    // variables
    DefaultMutableTreeNode  grpNode = null;
    DefaultMutableTreeNode  groupsNode;
    Enumeration             enumeration;
    DefaultMutableTreeNode  node = null;

    // test the ACSTree.GROUPS node
    groupsNode = ACSTreeUtilities.getACSTreeMainNode(
      treeModel_, acs_, ACSFactoryUtilityImpl.GROUPS, isExplorer_);
    if (groupsNode == null || _grp == null) {

                  // for testing and tracking
                  //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                    //", ACSTree.GROUPS is not detected, or the argument is null "+
                    //"in modifyGroupIDMembers().");

      return;
    }
 
                  // for testing and tracking
                  //System.out.println("ACSTreeBaseListenerImpl.modifyGroupIDMembers() "+
                    //      " for the argument: "+_grp.getNickName());

      // changes groupsNode if it is a sub-or-tree-typed group
    if (_grp.getEPType().startsWith("<byAcsCreator> ")) {
          groupsNode = ACSTreeUtilities.getParentOfSubOrTreeTypedGroupID(groupsNode, _grp);
    } // if
 
                  // for testing and tracking
                  //System.out.println("ACSTreeBaseListenerImpl.modifyGroupIDMembers() "+
                    //      " with groupsNode: "+groupsNode);

    // enumeration to select the FIRST node which contains
    // _grp as user object, and change all its children
    enumeration = groupsNode.children();
    while (enumeration.hasMoreElements()) {
      grpNode = (DefaultMutableTreeNode) enumeration.nextElement();
      if (grpNode.getUserObject() == _grp) break;
    } // while    
    
    if (grpNode == null) throw new InternalError("in ACSTreeBaseListenerImpl."+
      "modifyGroupIDMembers(), the GroupID node is not found in the explorer.");

             // for testing and tracking
            //System.out.println("in ACSTreeBaseListenerImpl.modifyGroupIDMembers(), for"+
              //acs_.getName()+", change the children for the group: "+_grp);

    // removes grpNode children, then builds up the children
    ACSTreeUtilities.removeAllChildren(treeModel_, grpNode);
    treeModel_ = ACSTreeUtilities.buildUpGroupTree((ACS)acs_,
          treeModel_, grpNode, (GroupID)_grp );

    groupsNode = null;
    grpNode = null;
    enumeration = null;
    node = null;   
  } // modifyGroupIDMembers


  /**
   * Processes an event from ACSRun about 'OneTypedPrivilege' and 'OneLinkedPrivilege' 
   * updates. Doesn't control the new value.
   * No action if a relevant main node (as ACSTree.PRIVILEGES) is not displayed.
   * @param   _evt PropertyChangeEvent
   */
  private void changeOnePrivilege(PropertyChangeEvent _evt){

                                    // for testing and tracking
                                  //if (_evt.getOldValue() != null)
                                    //  System.out.println("ACSTreeBaseListenerImpl"+
                                      //".changeOnePrivilege()  for the old value "+
                                      //_evt.getOldValue());
                                  //else
                                    //  System.out.println("ACSTreeBaseListenerImpl"+
                                      //".changeOnePrivilege()  for the new value "+
                                      //_evt.getNewValue());

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeOnePrivilege() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    Privilege               priv = null;
    DefaultMutableTreeNode  rootNode;
    boolean                 toInsert = false;

    // if the new value is non-null,
    // it is a new privilege to insert in the tree.
    if (_evt.getOldValue() == null) {
      toInsert = true;
      priv = (Privilege)_evt.getNewValue();
    }
    else {
      priv = (Privilege)_evt.getOldValue();
    } // else

                       // for testing and tracking
                       //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                        //", privilege "+priv+" to insert: "+toInsert);


    if (!priv.getAcsName().equals( acs_.getName() )) {
        priv = null;
        rootNode = null;

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", priv has not the correct ACS: "+priv.getAcsName());

        return;
    }

    // a privilege has to be updated under ACSTree.PRIVILEGES
    rootNode = ACSTreeUtilities.getACSTreeMainNode(
      treeModel_, acs_, ACSFactoryUtilityImpl.PRIVILEGES, isExplorer_);
    if (rootNode != null) {

        // handles the 'Privilege' flat list in the tree;
        // inserts (following the order of the Privilege name, not the detailled name order)
        // or removes 'priv' at the fifth position
        if (!toInsert) {
            if (priv instanceof PreviousDetailledName) {

                           // for testing and tracking
                           //System.out.println("ACSTreeBaseListenerImpl.changeOnePrivilege() "+
                            //", previous detailled name of the old privilege: "+
                            //((PreviousDetailledName)priv).getPreviousDetailledName());


                treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                    treeModel_, rootNode,
                    ((PreviousDetailledName)priv).getPreviousDetailledName(), toInsert);
            }
            else {
                treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                    treeModel_, rootNode,
                    priv.getDetailledName(), toInsert);
            } // else
        }
        else {

                           // for testing and tracking
                           //System.out.println("ACSTreeBaseListenerImpl.changeOnePrivilege() "+
                            //", detailled name of the new privilege: "+
                            //((PreviousDetailledName)priv).getDetailledName());

                treeModel_ = ACSTreeUtilities.addDetailledNameUnderNode(
                    treeModel_, rootNode,
                    priv, (ACS)acs_);
        } // else

    } // if
    else {
       throw new InternalError(
               "ACSTreeBaseListenerImpl.changeOnePrivilege(): in "+
               acs_.getName()+", NOOP since rootNode is null ");
    }

    priv = null;
    rootNode = null;
  } // changeOnePrivilege


  /**
   * Processes event from ACS about 'OneResourceVirtualFolder' updates.
   * Doesn't control the new value. Builds up the member nodes if necessary.
   * Manage the VirtualFolders tree.
   * No action if a relevant main node (as ACSTree.ACSRIGHTS)
   * is not displayed.
   * @see #changeOneResource(java.beans.PropertyChangeEvent)
   * @param   _evt PropertyChangeEvent
   */
  private void changeOneResourceVirtualFolder(PropertyChangeEvent _evt){

                // for testing and tracking
              //System.out.println("ACSTreeBaseListenerImpl: in "+toString()+
                  //" - call to changeOneResourceVirtualFolder() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeOneResourceVirtualFolder() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    VirtualFolder           vf;
    DefaultMutableTreeNode  rootNode;
    boolean                 toInsert = false;

    // if the new value is non-null,
    // it is a new eligible party to insert in the tree.
    if (_evt.getOldValue() == null) {
      toInsert = true;
      vf = (VirtualFolder)_evt.getNewValue();
    }
    else {
      vf = (VirtualFolder)_evt.getOldValue();
    } // else

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", vf "+vf+" to insert: "+toInsert);


    if (!vf.getAcsName().equals( acs_.getName() )) {
        vf = null;
        rootNode = null;

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", right has not the correct ACS: "+right.getAcsName());

        return;
    }

    // a virtual folder has to be updated under ACSTree.RESVFOLDERS
        // if the right has been finalized, the metaright has unregistered it
    rootNode = ACSTreeUtilities.getACSTreeMainNode(
      treeModel_, acs_, ACSFactoryUtilityImpl.RESVFOLDERS, isExplorer_);


    if (rootNode == null) {
               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", NOOP since rootNode is null ");

        vf = null;
        rootNode = null;
        return;
    } // if

    if (vf.getEorParent() != null) {
        rootNode = ACSTreeUtilities.getCurrentParentOfVirtualFolderInTree(rootNode, vf);
               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", NOOP since rootNode is null ");

        if (rootNode == null) {
            vf = null;
            rootNode = null;
            return;
        } // if
    } // if

    // if toInsert false, deletes all the subnodes of the vf node
    treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                    treeModel_, rootNode, vf, toInsert);
    // does not build up the subtree under the vf node, since it has no child
    // at its creation
    // uses rootNode.getFirstLeaf() since the vf node has been put at the index 0
    if (toInsert) {
      ACSTreeUtilities.buildUpVirtualFolderMembers(
              treeModel_, rootNode.getFirstLeaf());
      vf.addPropertyChangeListener("OneBaseObjectInFolder", this);
    } // if

    vf = null;
    rootNode = null;
  } // changeOneResourceVirtualFolder

  /**
   * Processes event from ACS about 'OneEPVirtualFolder' updates.
   * Doesn't control the new value. Builds up the member nodes if necessary.
   * Manage the VirtualFolders tree.
   * No action if a relevant main node (as ACSTree.ACSRIGHTS)
   * is not displayed.
   * @see     ACSTreeBaseListenerImpl#changeOneResource
   * @param   _evt PropertyChangeEvent
   */
  private void changeOneEPVirtualFolder(PropertyChangeEvent _evt){

                // for testing and tracking
              //System.out.println("ACSTreeBaseListenerImpl: in "+toString()+
                //  " - call to changeOneEPVirtualFolder() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeOneEPVirtualFolder() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    VirtualFolder           vf;
    DefaultMutableTreeNode  rootNode;
    boolean                 toInsert = false;

    // if the new value is non-null,
    // it is a new eligible party to insert in the tree.
    if (_evt.getOldValue() == null) {
      toInsert = true;
      vf = (VirtualFolder)_evt.getNewValue();
    }
    else {
      vf = (VirtualFolder)_evt.getOldValue();
    } // else

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl.changeOneEPVirtualFolder(): in "+
                //acs_.getName()+", vf "+vf+" to insert or remove: "+toInsert);


    if (!vf.getAcsName().equals( acs_.getName() )) {
        vf = null;
        rootNode = null;

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", right has not the correct ACS: "+right.getAcsName());

        return;
    }

    // a virtual folder has to be updated under ACSTree.EPVFOLDERS
    rootNode = ACSTreeUtilities.getACSTreeMainNode(
      treeModel_, acs_, ACSFactoryUtilityImpl.EPVFOLDERS, isExplorer_);

    if (rootNode == null) {
               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", NOOP since rootNode is null ");

        vf = null;
        rootNode = null;
        return;
    } // if

    if (vf.getEorParent() != null) {
        rootNode = ACSTreeUtilities.getCurrentParentOfVirtualFolderInTree(rootNode, vf);
               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", NOOP since rootNode is null ");

        if (rootNode == null) {
            vf = null;
            rootNode = null;
            return;
        } // if
    } // if

    // if toInsert false, deletes all the subnodes of the vf node
    treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                    treeModel_, rootNode, vf, toInsert);
    // does not build up the subtree under the vf node, since it has no child
    // at its creation
    // uses rootNode.getFirstLeaf() since the vf node has been put at the index 0
    if (toInsert) {
      ACSTreeUtilities.buildUpVirtualFolderMembers(
              treeModel_, rootNode.getFirstLeaf());
      vf.addPropertyChangeListener("OneBaseObjectInFolder", this);
    } // if

    vf = null;
    rootNode = null;
  } // changeOneEPVirtualFolder()

  /**
   * Processes event from VirtualFolder about 'OneBaseObjectInFolder' updates.
   * Doesn't control the new value.
   * No action if a relevant main node (as ACSTree.ACSRIGHTS) is not displayed.
   * @see     ACSTreeBaseListenerImpl#changeOneResource
   * @param   _evt PropertyChangeEvent
   */
  private void changeOneBaseObjectInFolder(PropertyChangeEvent _evt){

                // for testing and tracking
              //System.out.println("ACSTreeBaseListenerImpl: in "+toString()+
                //  " - call to changeOneBaseObjectInFolder() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "call to changeOneBaseObjectInFolder() with both old and new values"+
        " which are not null.");
    } // if
    if (!(_evt.getSource() instanceof VirtualFolder)) {
      throw new InternalError("ACSTreeBaseListenerImpl: "+
        "changeOneBaseObjectInFolder() but the event source is not a VirtualFolder.");
    } // if

    // variables
    VirtualFolder           vf = (VirtualFolder)_evt.getSource();
    DefaultMutableTreeNode  rootNode;
    DefaultMutableTreeNode  vfNode = null;
    DefaultMutableTreeNode  node = null;
    boolean                 toInsert = false;
    boolean                 found = false;
    ACSObject               obj = null;

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl: in "+acs_.getName()+
                //", vf "+vf+" to insert: "+toInsert);


    if (!vf.getAcsName().equals( acs_.getName() )) {
        vf = null;

               // for testing and tracking
               //System.out.println("ACSTreeBaseListenerImpl.changeOneBaseObjectInFolder(): in "+
                    //acs_.getName()+", right has not the correct ACS: "+right.getAcsName());

        return;
    }

    // if the new value is non-null,
    // it is an ACSObject to insert in the tree
    if (_evt.getOldValue() == null) {
      toInsert = true;
      obj = (ACSObject)_evt.getNewValue();
    }
    else {
      obj = (ACSObject)_evt.getOldValue();
    } // else

    // a virtual folder member has to be updated under one of the two main nodes
    if (vf.getMainVFType().startsWith(VirtualFolderImpl.EP_VFOLDER))
        rootNode = ACSTreeUtilities.getACSTreeMainNode(
              treeModel_, acs_, ACSFactoryUtilityImpl.EPVFOLDERS, isExplorer_);
    else
        rootNode = ACSTreeUtilities.getACSTreeMainNode(
              treeModel_, acs_, ACSFactoryUtilityImpl.RESVFOLDERS, isExplorer_);

    if (rootNode != null) {

        if (vf.getEorParent() != null) {
            rootNode = ACSTreeUtilities.getCurrentParentOfVirtualFolderInTree(rootNode, vf);

            if (rootNode == null) {

                           // for testing and tracking
                           //System.out.println("ACSTreeBaseListenerImpl.changeOneBaseObjectInFolder(): in "+
                            //acs_.getName()+", the parent node is not found. ");

                vf = null;
                obj = null;
                rootNode = null;
                return;
            } // if
        } // if

        // find the vf node under rootNode
        for (Enumeration e = rootNode.children() ; e.hasMoreElements() ;) {
            vfNode = (DefaultMutableTreeNode)e.nextElement();

                       // for testing and tracking
                       //System.out.println("ACSTreeBaseListenerImpl.changeOneBaseObjectInFolder(): "+
                         //      "test VF: "+vfNode.getUserObject());

            if (vfNode.getUserObject().equals(vf)) {
                found = true;
                break;
            } // if
        } // for e
        if (!found) {
            throw new InternalError("ACSTreeBaseListenerImpl: "+
                "changeOneBaseObjectInFolder()"+
                "the virtual folder "+vf.getNickName()+" is not found in the explorer "+
                "under the node "+rootNode.getUserObject());
        } // if

        if (toInsert) {
            // do not use ACSTreeUtilities.modifyObjectUnderNode()
            // to add the new node at the end of the current children
            node = new DefaultMutableTreeNode(obj, false);
            treeModel_.insertNodeInto(node, vfNode, vfNode.getChildCount());
        }
        else {
             // remove the obj node under vfNode
            treeModel_ = ACSTreeUtilities.modifyObjectUnderNode(
                treeModel_, vfNode, obj, false);
        } // else
    } // if
    else {
           throw new InternalError("ACSTreeBaseListenerImpl.changeOneBaseObjectInFolder() in "+
                   acs_.getName()+", NOOP since rootNode is null ");
    } // else

    vf = null;
    rootNode = null;
    vfNode = null;
    node = null;
    obj = null;
  } // changeOneBaseObjectInFolder()

 /**
 * Unregisters the instance as a listener for all registered sources.
 * Sets the variables to null.
 */
  public void finalizeForUser() {

                // for testing and tracking
            //System.out.println("call to ACSTreeBaseListenerImpl.finalizeForUser()"+
              //" for the ACS "+acs_+ ". In the explorer?: "+ isExplorer_);

   acs_.removePropertyChangeListener("OneExternalEligibleParty", this);
   acs_.removePropertyChangeListener("OneExternalResource", this);
   acs_.removePropertyChangeListener("OneResource", this);
   acs_.removePropertyChangeListener("OneEligibleParty", this);
   acs_.removePropertyChangeListener("AclEntry", this);
   
   // new in 0.7.0
   acs_.removePropertyChangeListener("OneTypedPrivilege", this);
   acs_.removePropertyChangeListener("OneLinkedPrivilege", this);
   acs_.removePropertyChangeListener("OneResourceVirtualFolder", this);
   acs_.removePropertyChangeListener("OneEPVirtualFolder", this);

   acs_.removePropertyChangeListener("OneAcsRight", this);
   
   //for (ListIterator ite = l_eventSources_.listIterator() ; ite.hasNext() ; ) {
    //obj = ite.next();
    //if (obj instanceof ACSRun) ((ACSRun)obj).removePropertyChangeListener(
    //  "ExternalOrientedAclEntry", this);
      // to complete later
      // if (obj instanceof ACSObject) ((ACSObject)obj).removePropertyChangeListener(this);
      // if (obj instanceof ACSManager)
   //} // for

   BaseManagerImpl.getInstance().removePropertyChangeListener("OpenACS", this);

   if (acs_ instanceof ImmutableACS) {

        if (ACSTreeUtilities.getACSTreeMainNode(
          treeModel_, acs_, ACSFactoryUtilityImpl.GROUPS, isExplorer_) != null) {
              ImmutableEligibleParty ep;
              for(Iterator ite = ((ImmutableACS)acs_).getEorM_OwnEligibleParties().values().iterator();
                ite.hasNext(); ){
                  ep = (ImmutableEligibleParty)ite.next();
                  if (!(ep instanceof ImmutableGroupID)) continue;
                  ep.removePropertyChangeListener("l_Members", this);
              } // for ite
              ep = null;
        }
        if (ACSTreeUtilities.getACSTreeMainNode(
            treeModel_, acs_, ACSFactoryUtilityImpl.RESVFOLDERS, isExplorer_) != null ||
            ACSTreeUtilities.getACSTreeMainNode(
            treeModel_, acs_, ACSFactoryUtilityImpl.EPVFOLDERS, isExplorer_) != null) {
              VirtualFolder[] l_vf = ((ImmutableACS)acs_).getEorL_VirtualFolders();
              for (int i = 0; i < l_vf.length; i++ ) {
                 l_vf[i].removePropertyChangeListener("OneBaseObjectInFolder", this);
              } // fori
              l_vf = null;
          } // if
          if (ACSTreeUtilities.getACSTreeMainNode(
            treeModel_, acs_, ACSFactoryUtilityImpl.PRIVILEGES, isExplorer_) != null ) {
              List l_pri = ((ImmutableACS)acs_).getEorL_PrivilegeForLinks();
              for (Object obj : l_pri) {
                 ((Privilege)obj).removePropertyChangeListener("OneLinkedPrivilegeSource", this);
                 ((Privilege)obj).removePropertyChangeListener("OneLinkedPrivilegeTarget", this);
              } // obj
              l_pri = null;
          } // if

   } // if (acs_ instanceof

   l_eventSources_.clear();
   acs_ = null;
   treeModel_ = null;
   root_ = null;
   tree_ = null;
  } // finalizeForUser()

 /**
 * HashCode of the instance.
 * It is used by TreeManager, ACSImpl, GroupIDImpl.
 * Don't use the event source list.
 * @return int
 */
  /*public int hashCode(){
        int result = 0;

        result = PRIME*result + acs_.getName().hashCode();
        // no use in the ACS main frame
        if (root_.getParent() != null) result =
            PRIME*result + ((DefaultMutableTreeNode)root_.getParent()).
                getUserObject().hashCode();

        return result;
  } // hashCode()
   */

  /**
   * no equals: each instance is equals only to itself.  
   * DefaultTreeModel and DefaultMutaleTreeNode have no equals() methods.
   * @param _obj: object to compare
   * @return true if the two objects are identical instances
   * of identical classes with the same acs and the same root
   */
   /*public synchronized boolean equals(Object _obj) {
    if (this == _obj)
        return true;
    if (_obj == null)
        return false;
    if (_obj.getClass() != getClass())
        return false;

    ACSTreeBaseListenerImpl obj = (ACSTreeBaseListenerImpl) _obj;

    if (acs_ == null || obj.getEorMainEventSource() == null) {
        if (acs_!= null || obj.getEorMainEventSource() != null)
          return false;  // if
    } // if
    else {
      if (!acs_.getName().equals( obj.getEorMainEventSource() )) return false; }

    
    if (root_.getParent() == null || obj.getEorRoot().getParent() == null) {
        if (root_.getParent() != null || obj.getEorRoot().getParent() != null)
          return false;  // if
    } // if
    else {
      if (!((DefaultMutableTreeNode)root_.getParent()).getUserObject().equals(
        ((DefaultMutableTreeNode)obj.getEorRoot().getParent()).getUserObject() ))
          return false; }

    return true;
  } //equals()
    */
   
 /**
 * @return String
 */
  public String toString(){
        StringBuilder sb = new StringBuilder();
        sb.append("  ACSTREE_BASELISTENERIMPL.TOSTRING():");

        sb.append(" acs name = ");
        if (acs_ != null) sb.append(acs_.getName());
        else sb.append("null acs");

        sb.append(" user object root = ");
        if (root_ != null && root_.getUserObject() != null)
          sb.append( root_.getUserObject().toString() );
        else sb.append("null");

        return " " + sb;
  }

} // ACSTreeBaseListenerImpl