//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard 
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;

import ARoad0.gWorkInterface.RightsMediator;
import java.awt.*;
import java.awt.geom.*;
import java.awt.font.*;
import javax.swing.*;
import javax.swing.border.*;

import java.util.Map;
import java.util.List;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import ARoad0.Gui1.*;
import ARoad0.ErrW.*;
import ARoad0.ErrB.BaseError;
import ARoad0.gWork.*;
import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.gBaseInterface.*;

/**
 * This class is responsible for displaying an EPRViewInBase and its graphical
 * objects (nodes, links, arrows, rights) in the desktop.
 * The BaseObjects map is a map of (Source, Target) pairs, that is in fact
 * (EligibleParty/VirtualFolder, Resource/VirtualFolder) pairs,
 * and the GraphicNodes are GraphicAccount/Group/Resource/Actor/VirtualFolder.
 * Study the parent GraphicView class for a good understanding of this class.
 * Its subclasses are GraphicSketchView and GraphicNoThanView.
 * @see ARoad0.gBase.EPRViewInBaseImpl
 * @version 0.7.0
 */
public class GraphicEPRView extends GraphicView {

  // variables
    /**
     * List of Targets in viewInBase_ that are not ImmutableSources, then this is
     * the list of Resources which are not EligibleParties.
     * Caution: in the subclass GraphicSketchView, it never contains the center node.
     */
  protected List                          l_NoSourceTargets_ =
                                            new ArrayList(INITIAL_CAPACITY);
    /**
     * After the search of the DisplayableLinkImpls,
     * list of EligibleParties that are access targets for
     * other Sources in viewInBase_. Do not contain the VirtualFolders.
     * Caution: in the subclass GraphicSketchView, it never contains the center node.
     */
  protected final List                    l_AccessTargetSources_ =
                                            new ArrayList(INITIAL_CAPACITY);
    /**
     *  l_NoSourceTargets_.size().
     */
  protected int                           noSourceTargetsCount_;
    /**
     *  number of Sources in viewInBase_
     */
  protected int                           sourcesCount_;
  /*
   * These booleans are used to accelerate the rights processing.
   * False is the default value.
   */
  private boolean                                 detectAcl_ = false;
  private boolean                                 detectUserGroupLinks_ = false;

  //
  // ------------- constructors ------------------------------
  //
  /**
   * This constructor is only for the subclasses.
   */
  public GraphicEPRView() {} 
  
  /**
  * This constructor adds all the _view objects in this panel,
  * each associated to a GraphicNode.
  * Builds up the map of (Source, Target) pair
  * linked by displayable links, then sets the JPanel size, its GraphicNodes,
  * constructs the arrows and the visible rights shapes
  * to be drawn in paintComponents().
  * The main work is done there by buildUpArrowsAndRights(),
  * so that the paint action is fast. 
  * Any BaseObject updating is done by GraphicViewBaseListenerImpl,
  * which resets the instance through resetRights() and resetGraphicView().
  * Called by ActionOpenGraphicView.openViewInGui() in the event dispatch thread.
  * @param _view EPRViewInBase to associate to this instance.
  * @param _m_positionIndex is the map (name, Integer) returned
  *   by getIndexesOfNodeLocations() at the previous view opening. May be null.
  * @exception ProcessError if there is a loop in the rights searching
  **/
  public GraphicEPRView(EPRViewInBase _view, Map _m_positionIndex) throws ProcessError {
    viewInBase_ = _view;

    // no layout, because they are not able to return
    // the absolute coordinates of a component.
    this.setLayout(null);

    this.setBorder(new EmptyBorder(BORDER*2,BORDER,BORDER*2,BORDER));

    // sets the GraphicNodes in the JPanel
    try {
        initialize();
    } catch (InterruptedException i) {
        throw new ProcessError("GraphicEPRView constructor: "+i.toString());
    }
    
    // resets the node positions
    if (_m_positionIndex != null) setIndexesOfNodeLocations(_m_positionIndex);
    
    // constructs the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    l_LinksAndRights_ = buildUpArrowsAndRights(m_l_DisplayableLinks_, m_GraphicNodesBounds_);

    // sets the color
    setBackground(Color.white);

    addMouseListener(this);
    addMouseMotionListener(this);
  } // GraphicEPRView(EPRViewInBase _view)

  /**
   * This central method is the main manager of the instance.
  * It initializes nodesNumber_, m_l_DisplayableLinks_, l_allNodesInLinks_, and
  * set the Graphicnodes locations, and initializes l_NoSourceTargets_,
  * l_AccessTargetSources_, noSourceTargetsCount_, sourcesCount_,
  * detectAcl_, detectUserGroupLinks_, detectPriv_, detectCompoundLinks_.
  * Called by the constructor and resetGraphicView() to do the main work.
  * Calls RightsFactory_Facade.detectEPRRights().
  * Synchronized on viewInBase_.
  * Overrides the GraphicView method.
  * @exception InterruptedException, ProcessError if there is a loop in the rights searching
  * */
  protected void initialize() throws ProcessError, InterruptedException {
    // variables
      // one (EP, Resource) pair
    List                  l_link;
    BaseObject            secondEnd;
    BaseObject[]          l_node;
    DisplayableLinkImpl   displayableLink;
    ImmutableName         acsname;
    ImmutableACS          acs;
    List                  l_acsnames = new ArrayList(30);
    RightsMediator        mediator = ViewModel.getViewRightsMediator(viewInBase_);
    RightsFactory_Facade  facade = null;

    if (mediator != null)
        facade = (RightsFactory_Facade) mediator.getCoreAlgorithm("Facade");

    if (mediator == null || facade == null)
        throw new InternalError("GraphicEPRView.initialize() does not found the "+
                "RightsFactory_Facade instance for the view "+viewInBase_.getNickName());

    synchronized (viewInBase_) {

      nodesNumber_ = viewInBase_.getEorM_BaseObjects().size();

      l_NoSourceTargets_ = new ArrayList(
        ((EPRViewInBase)viewInBase_).getEorM_NoSourceTargets().values());

                        // for testing and tracking
                    //System.out.println("In GraphicEPRView.initialize(), the ep are:");
                    //for (Iterator ite = eprnodesNumber_.getEorM_EligibleParties().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
                    //System.out.println("In GraphicEPRView.initialize(), the non-ep resources are:");
                    //for (Iterator ite = l_NoSourceTargets_.iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
  
      // initializes detectAcl_, detectUserGroupLinks_
      try {
          for (Object obj : viewInBase_.getEorM_BaseObjects().values()) {
              acsname = ((ACSObject)obj).getAcsName();
              
              if (l_acsnames.contains(acsname)) continue;
              else {
                  l_acsnames.add(acsname);
                  acs = ACS_Facade.openOrGetACS(acsname);

                  if (!detectAcl_ &&
                          (acs.getManageInternalAcls() || acs.getManageExternalAcls()))
                      detectAcl_ = true;
                  if (!detectUserGroupLinks_ &&
                          acs.getManageAccounts())
                      detectUserGroupLinks_ = true;
              } // else
          } // for obj
      }
      catch (BaseError b) {
          throw new InternalError("GraphicEPRView.initialize(): "+b.toString());
      }

      // builds up the map of (Source, Target) pair linked by displayable links
      m_l_DisplayableLinks_ = facade.
        detectEPRRights(this, ((EPRViewInBase)viewInBase_),
        detectAcl_, detectUserGroupLinks_, true, true);

      // updates l_AccessTargetSources_, l_allNodesInLinks_.
      // m_l_DisplayableLinks_ may be empty.
      l_AccessTargetSources_.clear();
      l_allNodesInLinks_.clear();
      for (Iterator ita = m_l_DisplayableLinks_.values().iterator() ;
        ita.hasNext() ; ) {
          l_link = (List)ita.next();

                  // for testing and tracking
                  //System.out.println("l_link in initialize(): "+l_link);

          secondEnd = ((DisplayableLinkImpl)l_link.get(0)).getEorSecondEnd();
          if (secondEnd instanceof ImmutableSource
            && !l_AccessTargetSources_.contains(secondEnd) )
              l_AccessTargetSources_.add(secondEnd);

          for (Iterator ite = l_link.iterator() ;ite.hasNext() ; ) {
            displayableLink = (DisplayableLinkImpl)ite.next();
            l_node = displayableLink.getEorAllNodes();
            for (int i = 0; i < l_node.length; i++) {
              l_allNodesInLinks_.add(l_node[i] );
            } // for i
          } // for ite
      } // for (Iterator ita = m_l_DisplayableLinks_

                    // for testing and tracking
                    //System.out.println("l_AccessTargetSources_: "+l_AccessTargetSources_);

      // updates noSourceTargetsCount_ and sourcesCount_ before the call to setGraphicNodes...()
      noSourceTargetsCount_ = l_NoSourceTargets_.size();
      sourcesCount_ = ((EPRViewInBase)viewInBase_).getEorM_Sources().values().size();

      if (nodesNumber_ < 48) setGraphicNodes();
      else {
        // displays a message
        DialogBox.showOkMessage(
          "This view has more than 47 elements. It is not displayed in a diagram"+
            ", while the display of all the access paths is working.");
      }
    } // synchro
;
    l_link = null;
    secondEnd = null;
    l_node = null;
    displayableLink = null;
    l_acsnames.clear();
    l_acsnames = null;
    mediator = null;
    facade = null;
  } // initialize()

  /**
  * Sets the GraphicNodes in a 16, 25, 49 or 100-positions layout in the JPanel,
  * with their absolute coordinates. Called by intialize().
  * Calls super.setGraphicNodesPositions().
  * Clears then updates m_GraphicNodes_, updates m_GraphicNodesBounds_.
  * Overridden by the GraphicSketchView and the GraphicNoThanView methods.
  * Overrides the GraphicView method.
  * */
  protected void setGraphicNodes() {
    //variables
    GraphicNode       gep;
    GraphicNode       gact;
    GraphicNode       gres;
    ImmutableSource   sour;
    ImmutableTarget   targ;
    int               index = 0;
    List<ImmutableSource>  l_noAccessTargetSources =
      new ArrayList(  ((EPRViewInBase)viewInBase_).getEorM_Sources().values());

                        // for testing and tracking
                    //System.out.println("GraphicEPRView.setGraphicNodes():");
                    //System.out.println("for "+viewInBase_.getName()+", nodesNumber_: "+nodesNumber_);
                    //System.out.println("l_noAccessTargetSources before the removing of l_AccessTargetSources_:");
                    //for (Iterator itetest = l_noAccessTargetSources.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_AccessTargetSources_:");
                    //for (Iterator itetest = l_AccessTargetSources_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}

    //m_GraphicNodes_.clear(); // remove the WAITING_TEXT if any
    l_noAccessTargetSources.removeAll(l_AccessTargetSources_);

    // set a fixed size
    setViewSize(nodesNumber_);

    // set the nodes positions
    Point[]           l_positions = setGraphicNodesPositions(nodesNumber_, true);

                        // for testing and tracking    
                    //System.out.println("GraphicEPRView.setGraphicNodes():");
                    //System.out.println("for "+viewInBase_.getName()+", nodesNumber_: "+nodesNumber_);
                    //System.out.println("l_positions:");
                    //for (int i = 0; i < l_positions.length; i++ ) {
                      //if (l_positions[i] != null)
                        //  System.out.println("--"+i+": "+ l_positions[i].toString() );
                    //}
                    //System.out.println("l_noAccessTargetSources after the removing of l_AccessTargetSources_:");
                    //for (Iterator itetest = l_noAccessTargetSources.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_AccessTargetSources_:");
                    //for (Iterator itetest = l_AccessTargetSources_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_NoSourceTargets_:");
                    //for (Iterator itetest = l_NoSourceTargets_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}

    for (Iterator ite = l_noAccessTargetSources.iterator() ; ite.hasNext(); ) {

        sour = (ImmutableSource) ite.next();
        if (sour instanceof Actor)
            gep = new GraphicActor(this, (Actor)sour );
        else if (sour instanceof GroupID)
            gep = new GraphicGroup(this, (GroupID)sour );
        else if (sour instanceof UserID)
            gep = new GraphicAccount(this, (UserID)sour );
        else
            gep = new GraphicVirtualFolder(this, (VirtualFolder)sour);
        // adds to this JPanel
        add(gep );
        // puts it at the right place
        gep.setBounds(
          0,
          0,
          (int)gep.getShape().getBounds2D().getWidth(),
          (int)gep.getShape().getBounds2D().getHeight());
        gep.setLocation(l_positions[index]);
        
        // adds nodeBounds to its map
        m_GraphicNodesBounds_.put(sour.getName(), gep.getBounds());

                            // for testing and tracking
                            //System.out.println("GraphicEPRView.setGraphicNodes():");
                            //System.out.println("l_noAccessTargetSources: node bounds "+
                              //      gep.getBounds()+" for "+ep.getName());
        
        // adds the node to its map
        m_GraphicNodes_.put(sour.getName(), gep);
        gep.setVisible(true);
        index++;

    } // for (Iterator ite

    for (Iterator ite = l_AccessTargetSources_.iterator() ; ite.hasNext(); ) {

        sour = (ImmutableSource) ite.next();
        if (sour instanceof Actor)
            gact = new GraphicActor(this, (Actor)sour );
        else if (sour instanceof GroupID)
            gact = new GraphicGroup(this, (GroupID)sour );
        else if (sour instanceof UserID)
            gact = new GraphicAccount(this, (UserID)sour );
        else
            gact = new GraphicVirtualFolder(this, (VirtualFolder)sour);
        // adds to this JPanel
        add(gact );

        // puts it at the right place
        gact.setBounds(
          0,
          0,
          (int)gact.getShape().getBounds2D().getWidth(),
          (int)gact.getShape().getBounds2D().getHeight());

                            // for testing and tracking
                            //System.out.println("GraphicEPRView.setGraphicNodes():");
                            //System.out.println("----- l_positions[index]: "+l_positions[index]+
                              //      " at index "+index);


        gact.setLocation(l_positions[index]);

        // adds nodeBounds to its map
        m_GraphicNodesBounds_.put(sour.getName(), gact.getBounds());
        // adds nodes to its map
        m_GraphicNodes_.put(sour.getName(), gact);
        gact.setVisible(true);
        index++;

    } // for (Iterator ite

    for (Iterator ite = l_NoSourceTargets_.iterator() ; ite.hasNext(); ) {

        targ = (ImmutableTarget) ite.next();
        //if (targ instanceof Resource)
        gres = new GraphicResource(this, (Resource)targ);
        // adds to this JPanel
        add(gres );

        // puts it at the right place
        gres.setBounds(
          0,
          0,
          (int)gres.getShape().getBounds2D().getWidth(),
          (int)gres.getShape().getBounds2D().getHeight());
        
        gres.setLocation(l_positions[index]);

        // adds nodeBounds and nodes to their maps
        m_GraphicNodesBounds_.put(targ.getName(), gres.getBounds());
        m_GraphicNodes_.put(targ.getName(), gres );
        gres.setVisible(true);
        index++;

    } // for (Iterator ite

                            // for testing and tracking
                            //System.out.println("GraphicEPRView.setGraphicNodes(), nodes in m_GraphicNodes_: ");
                            //for (ImmutableName na : m_GraphicNodes_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na
                            //System.out.println("nodes in m_GraphicNodesBounds_: ");
                            //for (ImmutableName na : m_GraphicNodesBounds_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na


    gep = null;
    gact = null;
    gres = null;
    sour = null;
    targ = null;
    l_noAccessTargetSources = null;
    l_positions = null;
  } // setGraphicNodes()
  
  /**
   * This why text depends on the comments.
   * This method is used to test if the instance has been finalized.
   * Its height is used by the GraphicViewBaseListener.
   * Overridden by the GraphicNoThanView.
   * Overrides the GraphicView method
   * @return the text which explains the rights in the view. Null if finalized.
   *   Empty at the GraphicEPRView initialization.
   */
   public final JTextPane setWhyText() {
        if (whyText_ == null) return null;

        RightsMediator        mediator = ViewModel.getViewRightsMediator(viewInBase_);
        RightsFactory_Facade  facade = null;

        if (mediator != null)
            facade = (RightsFactory_Facade) mediator.getCoreAlgorithm("Facade");

        if (mediator == null || facade == null)
            throw new InternalError("GraphicEPRView.setWhyText() does not found the "+
                    "RightsFactory_Facade instance for the view "+viewInBase_.getNickName());

        // construct the why text
        whyText_ = facade.getEPRWhyText(this, m_l_DisplayableLinks_,
               ((EPRViewInBase)viewInBase_));  

        mediator = null;
        facade = null;
        return whyText_;
   } // setWhyText()

  /**
   * Invoked by resetGraphicView() and ViewTreeBaseListenerImpl to finalize the instance
   * and set it as a new object before the initialization. There is the exception of
   * whyText_ that is set to null to indicate the finalizing operation.
   * Overrides the GraphicView method.
   */
   public synchronized void finalizeForGui() {

          // for testing and tracking
          //System.out.println("Call of GraphicEPRView.finalizeForGui()");

      if (l_NoSourceTargets_ != null) l_NoSourceTargets_.clear();
      l_NoSourceTargets_ = null;
      if (l_AccessTargetSources_ != null) l_AccessTargetSources_.clear();
      super.finalizeForGui();
   } // finalizeForGui()

  /**
  * Sets the property l_LinksAndRights_ which contains the links and rights as Shapes,
  * to be painted by drawArrowsAndRights() in this GraphicView..
  * This complex method has to be fast. <P>
  * There are 4 elements for each typical arrow: 
  * 3 for the arrow and 1 text for the description,
  * which are often the rights to display. <P>
  * The strengthen path is associated to a strengthen relation
  * including the OWNER, HAS_MEMBER, HAS_CHILD, IS_ALIAS, RUN_UNDER links.
  * When it is a strengthen relation in the last link between 2 GroupIDMembers,
  * to draw a line and not an arrow, there is one element for a line,
  * two empty elements, and one text for the comments and rights. <P>
  * Otherwise, there are plain arrows, but not from a starting node with a dashed_stroke
  * arrow to show that this link is weak (marked by a 'WEAK' property of the first link). <P>
  * The first-rate path is associated to a direct path which is superior to the
  * other indirect paths between the two nodes, and its arrow is drawn with a larger width. <P>
  * In GraphicSketchView, when a BaseObject is drawn twice in the graphic view,
  * the second node is put on the bottom line of the access targets, and its key 
  * in m_GraphicNodes_ and m_GraphicNodesBounds_ is its name extended by '<!!§§!!%%!!%%!!§§>'. <P>
  * This method selects only the first DisplayableLinkImpl to get the two ends,
  * and to call link.getEorLastLink().getActivatedStrengthenTypes() and link.isWeak()
  * The view Shapes and Strings are actually drawn in drawArrowsAndRights().
  * Called by the contructor, by resetRights() in this class and its subclass, 
  * and by the super class method GraphicView.mouseReleased().
  * Calls RightsFactory_Facade.getEffectiveRightsForPair().
  * Calls the super class method setPoint2DForArrowAndRights(), setArrowHead().
  * Synchronized on m_l_DisplayableLinks_. Puts the absolute positionning
  * in l_currentArrowsAndRightsStarts_.
  * Overridden by GraphicNoThanView, but not by GraphicSketchView.
  * Overrides the GraphicView method.
  * @param _m_l_DisplayableLinks is the map of DisplayableLinks lists (one per pair)
  * associated to the view.
  * @param _m_gNodeBoundsModel is the map of the node bounds for the shapes
  * in the GraphicEPRView.
  * @return Object[] to draw in paintComponent(), with both the lines/arrows
  * at absolute positionning (1 Shape for the line and 2 Shapes for the head)
  * and the access rights and comments (1 String).
  * */
  protected Object[] buildUpArrowsAndRights(Map _m_l_DisplayableLinks,
    Map _m_gNodeBoundsModel) {

    // works on _m_l_DisplayableLinks, set in initialize()
    if (_m_l_DisplayableLinks.isEmpty()) return NULL_SHAPE_ARRAY;

                            // for testing and tracking
                            //System.out.println("GraphicEPRView.buildUpArrowsAndRights() on"+
                              //      " the view "+viewInBase_.getName()+
                                //" getRightsRelatedBaseObjects() contains; ");
                            //for (Object obj : getRightsRelatedBaseObjects()) {
                              //  System.out.println("--- "+obj);
                            //}

    // variables
    Object[]         result;
      // list of pair (EP,Resource)
      // with one or several access control links associated to, or one or several
      // ACS-specific access control right
    List            l_links = new ArrayList(INITIAL_CAPACITY);
    DisplayableLinkImpl link;
    BaseObject[]    l_entities = new BaseObject[2];
    Point2D.Float[] l_points = new Point2D.Float[3];
    int             result_index = 0;
    String          rights;
    Object[]        newResult;
    Object[]        newCurrentArrowsAndRightsStarts;
    int             epArrowsCount = 0;
    Object          lastEpEntity = null;
    boolean         hasStrengthenLinks;
    boolean         hasFirstRateLastLink;
    RightsMediator  mediator = ViewModel.getViewRightsMediator(viewInBase_);
    RightsFactory_Facade  facade = null;

    if (mediator != null)
        facade = (RightsFactory_Facade) mediator.getCoreAlgorithm("Facade");

    if (mediator == null || facade == null)
        throw new InternalError("GraphicEPRView.buildUpArrowsAndRights() does not found the "+
                "RightsFactory_Facade instance of the view "+viewInBase_.getNickName());
    
      // shallow copy used by getCurrentEffectiveRightsForPair()
    Map             m_l_DisplayableLinks_Copy = (Map)((HashMap)_m_l_DisplayableLinks).clone();

                                  // for testing and tracking
                                  //Map.Entry entry = null;
                                  //System.out.println("GraphicEPRView.buildUpArrowsAndRights() at the start, _m_gNodeBoundsModel contains the value:");
                                  //for (Iterator ite_test = _m_gNodeBoundsModel.entrySet().iterator() ; ite_test.hasNext() ; ) {
                                    //  entry = (Map.Entry)ite_test.next();
                                    //  System.out.println("---------------- "+ entry.getValue());
                                  //} // ite_test    
    

    synchronized (_m_l_DisplayableLinks) { // may be updated

      // the result length is four time the _m_l_DisplayableLinks size
      // since each access control link may have an arrow with 3 lines
      // (main line + 2 lines for the head)
      // plus a String to draw
      result = new Object[4*_m_l_DisplayableLinks.size()];
      l_currentArrowsAndRightsStarts_ = new Object[4*_m_l_DisplayableLinks.size()];

      // handles the matched relation
      for (Iterator ite = m_l_DisplayableLinks_Copy.values().iterator() ; ite.hasNext() ; ) {
        
        l_links = (List)ite.next();
        
        // selects only the first DisplayableLinkImpl, since it gives the two ends,
        // and to call link.getEorLastLink().getActivatedStrengthenTypes() and link.isWeak()
        link = (DisplayableLinkImpl) l_links.get(0);
        l_entities[0] = link.getEorFirstEnd();
        l_entities[1] = link.getEorSecondEnd();

        // to draw several arrows around a node, count them with epArrowsCount
        if (l_entities[0] != lastEpEntity) {
          lastEpEntity = l_entities[0];
          epArrowsCount = 0;
        }
        
                      // for testing and tracking
                      //System.out.println("in GraphicEPRView.buildUpArrowsAndRights(), "+
                        //"_m_l_DisplayableLinks iteration, result_index: "+result_index+
                        //", and result.length: "+result.length);
        
                      //for testing and tracking
                      //int e = 0;
                      //System.out.println("in GraphicEPRView.buildUpArrowsAndRights() iteration, one dLink list is:");
                      //for (Iterator ite_test = l_links.iterator() ; ite_test.hasNext() ; ) {
                        //link = (DisplayableLinkImpl) ite_test.next();
                        //System.out.println("        at ["+e+"], from "+ link.getEorFirstEnd().getName() +
                          //      " to "+ link.getEorSecondEnd().getName());
                        //e++;
                      //} // ite_test

        hasStrengthenLinks = hasStrengthenLinksInOneDLink(l_links);
        hasFirstRateLastLink = hasFirstRateLinkInOneDLink(l_links);

        // construct the effective access rights string
        // in handling all the relevant DisplayableLinks in _m_l_DisplayableLinks,
        // and puts it in rights. 
        rights = facade.getEffectiveRightsForPair(
          l_entities[0], l_entities[1], _m_l_DisplayableLinks);
        
                      // for testing and tracking
                      //System.out.println("in GraphicEPRView.buildUpArrows(), "+
                        //"_m_l_DisplayableLinks iteration, result_index: "+result_index);
                      //System.out.println("----- rights: "+rights+", hasStrengthenLinks: "+
                        //hasStrengthenLinks+", hasFirstRateLastLink: "+hasFirstRateLastLink);
        
        // if rights is empty and if it is not about a strengthen/first-rate relation
        // in the last acLink, it is not processed: then
        // reduce the result length and delete the current arrow
        if ((rights == null || rights.length() == 0 || rights.equals(" ")) && 
          !hasStrengthenLinks ) {

                          // for testing and tracking
                          //System.out.println("in GraphicEPRView.buildUpArrows(), rights empty with "+
                            //"result.length: "+result.length+" at result_index: "+result_index);

            if (result.length <= 4) {
              result = NULL_SHAPE_ARRAY;
              break;   
            } // if
            
            newResult = new Object[result.length - 4];
            newCurrentArrowsAndRightsStarts = new Object[result.length - 4];

            if (result_index > 3) {
              System.arraycopy(result, 0, newResult, 0, result_index);
              System.arraycopy(l_currentArrowsAndRightsStarts_, 0,
                newCurrentArrowsAndRightsStarts, 0, result_index);
            } // if  (result_index > 4) 
            result = newResult;
            l_currentArrowsAndRightsStarts_ = newCurrentArrowsAndRightsStarts;

                        // for testing and tracking    
                        //System.out.println("in GraphicEPRView.buildUpArrows(), rights empty with "+
                          //"new result:");
                        //for (int i = 0 ; i < newResult.length ; i++) {
                          //System.out.println("new result["+i+"]: "+newResult[i]);
                        //}

            continue;

        } // if (rights.length() == 0
        else epArrowsCount++;

                                  // for testing and tracking
                                  //Map.Entry entry = null;
                                  //System.out.println("GraphicEPRView.buildUpArrowsAndRights() "+
                                    //      "before the setPoint2DForArrowAndRights(), _m_gNodeBoundsModel contains the value:");
                                  //for (Iterator ite_test = _m_gNodeBoundsModel.entrySet().iterator() ; ite_test.hasNext() ; ) {
                                    //  entry = (Map.Entry)ite_test.next();
                                    //  System.out.println("---------------- "+ entry.getValue()+ " for the key "+entry.getKey());
                                  //} // ite_test 
                                  //System.out.println(
                                    //  " and from l_entities[0].getName(): "+ l_entities[0].getName() );
        
        // compute points and rights Point2D.Float associated to this access path.
        // test if it is the second drawn of a BaseObject in a GraphicSketchView.
        if (  _m_gNodeBoundsModel.containsKey(((Name)l_entities[1].getName()).
                extended("<!!§§!!%%!!%%!!§§>")  ) )
            l_points = setPoint2DForArrowAndRights(
               l_entities[0].getName(),
               ((Name)l_entities[1].getName()).extended("<!!§§!!%%!!%%!!§§>"),
               _m_gNodeBoundsModel, epArrowsCount);
        else l_points = setPoint2DForArrowAndRights(
                l_entities[0].getName(),
                l_entities[1].getName(), 
                 _m_gNodeBoundsModel, epArrowsCount);
        
        // to have a clear diagram, make a jump on the arrow, up or down:
        // - up if its sens is from left to right,
        if (l_points[0].getX() < l_points[1].getX()+30) {
            l_points[0] = new Point2D.Float(
                    (float)l_points[0].getX(), (float)l_points[0].getY() - 10);
            l_points[1] = new Point2D.Float(
                    (float)l_points[1].getX(), (float)l_points[1].getY() - 10);
            l_points[2] = new Point2D.Float(
                    (float)l_points[2].getX(), (float)l_points[2].getY() - 10);
        }    
        // - down if its sens is from right to left
        else if (l_points[0].getX() > l_points[1].getX()+30) {
            l_points[0] = new Point2D.Float(
                    (float)l_points[0].getX(), (float)l_points[0].getY() + 10);
            l_points[1] = new Point2D.Float(
                    (float)l_points[1].getX(), (float)l_points[1].getY() + 10);
            l_points[2] = new Point2D.Float(
                    (float)l_points[2].getX(), (float)l_points[2].getY() + 10);
        }
        // - no action if it is a vertical arrow (+-30)

        // construct each line and arrow without head
        result[result_index] = new Line2D.Float(l_points[0], l_points[1]);
        l_currentArrowsAndRightsStarts_[result_index] = result[result_index];

                        // for testing and tracking
                        //System.out.println("   a line to draw for the index: "+result_index);
                        //System.out.println("      start point: x: "+((Line2D.Float)result[result_index]).getP1().getX()+
                          //"; y: "+((Line2D.Float)result[result_index]).getP1().getY());
                        //System.out.println("      end point: x: "+((Line2D.Float)result[result_index]).getP2().getX()+
                          //"; y: "+((Line2D.Float)result[result_index]).getP2().getY());
        
        
        result_index++;

        // compute the 2-lines arrow head.
        // if it is a strengthen relation, sets two empty elements.
        if (!hasStrengthenLinks ) {

                                // for testing and tracking
                                //System.out.println("arrowAngle: "+arrowAngle);

                // line above the path
                result[result_index] = setArrowHead(l_points[0], l_points[1], true);
                l_currentArrowsAndRightsStarts_[result_index] = result[result_index];
                result_index++;
                // line below the path
                result[result_index] = setArrowHead(l_points[0], l_points[1], false);
                l_currentArrowsAndRightsStarts_[result_index] = result[result_index];
                result_index++;
        } else {
            result_index++;
            result_index++;
        }

                // for testing and tracking
                //System.out.println("   one arrow to draw: ");
                //System.out.println("   start point: x: "+((Line2D.Float)result[result_index]).getP1().getX()+
                  //"; y: "+((Line2D.Float)result[result_index]).getP1().getY());
                //System.out.println("   end point: x: "+((Line2D.Float)result[result_index]).getP2().getX()+
                  //"; y: "+((Line2D.Float)result[result_index]).getP2().getY());

                // for testing and tracking
                //System.out.println("in GraphicEPRView.buildUpArrows(), rights:"+rights);

        // Set the non-null String for rights.
        // if they are about a strengthen relation,
        // set the redondant rights in the last acLink as a text in result.
                // deprecated: for the hasOCLinks case, the text begins by '<>' 
                // for having a blue dashed line,
                // and these 2 caracters are not displayed.
        if (hasStrengthenLinks) {
                
            // link is the first DisplayableLinkImpl
            if (rights != null) result[result_index] = link.getEorLastLink().
                    getActivatedStrengthenTypes()+rights;
            else result[result_index] = link.getEorLastLink().
                    getActivatedStrengthenTypes();

                        // for testing and tracking
                        //System.out.println("GraphicEPRView.buildUpArrows(), "+
                          //      "the dLink has strengthen rights: "+result[result_index]);
        }
        // doublewidth_stroke arrow
        else if (hasFirstRateLastLink && !link.isWeak()) {

                        // for testing and tracking
                        //System.out.println("GraphicEPRView.buildUpArrows(), "+
                          //      "the dLink is first-rate");
            
            result[result_index] = "><"+rights;
        // dashed_stroke arrow from a group
        } else if (!hasFirstRateLastLink && link.isWeak()) {

                        // for testing and tracking
                        //System.out.println("GraphicEPRView.buildUpArrows(), "+
                          //      "the dLink is weak");
            
            result[result_index] = "<>"+rights;
        // doublewidth_stroke and dashed_stroke arrow
        } else if (hasFirstRateLastLink && link.isWeak()) { 

                        // for testing and tracking
                        //System.out.println("GraphicEPRView.buildUpArrows(), "+
                          //      "the dLink is first-rate and weak");
            
            result[result_index] = ">>"+rights;
        } else result[result_index] = rights;
        
        // this is the only element for which l_currentArrowsAndRightsStarts_
        // and result is different.
        l_currentArrowsAndRightsStarts_[result_index] = l_points[2];

                        // for testing and tracking
                        //System.out.println("in GraphicEPRView.buildUpArrows(), "+
                          //    "rights or description:"+result[result_index]+
                            //    " and l_points[2]: "+l_points[2]);
        
        result_index++;

      } // for (Iterator ite = m_l_DisplayableLinks_Copy

    } // synchro

    l_links = null;
    link = null;
    l_entities = null;
    l_points = null;
    rights = null;
    newResult = null;
    m_l_DisplayableLinks_Copy = null;
    newCurrentArrowsAndRightsStarts = null;
    mediator = null;
    facade = null;

    return result;
  } // buildUpArrowsAndRights()

  /**
  * Draws the new arrows and rights after a change in the rights,
  * without changing in GraphicNodes.
  * Called in a worker thread by GraphicViewBaseListener.changeFrameView() and
  * GraphicViewPropertyEditor.propertyChange().
  * Calls RightsFactory_Facade.detectEPRRights(), buildUpArrowsAndRights().
  * This method call must be preceded by addWaitingMessage() and
  * must be followed by a call to removeWaitingMessage() and repaint(),
  * both in the event dispatch thread.
  * Overridden by GraphicSketchView. Overrides the super class.
  * @exception ProcessError if there is a loop in the rights search 
  * */
  public synchronized void resetRights() throws ProcessError, InterruptedException {
      
    RightsMediator        mediator = ViewModel.getViewRightsMediator(viewInBase_);
    RightsFactory_Facade  facade = null;

    if (mediator != null)
        facade = (RightsFactory_Facade) mediator.getCoreAlgorithm("Facade");

    if (mediator == null || facade == null)
        throw new InternalError("GraphicEPRView.resetRights() does not found the "+
                "RightsFactory_Facade instance for the view "+viewInBase_.getNickName());

                    // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("call to GraphicEPRView.resetRights() in "+
                      //      viewInBase_.getName()+" within the thread :"+
                        //    Thread.currentThread().getName());
      
      // clear m_l_DisplayableLinks_ to cut any link to the current dLinks
    for (Iterator ite = m_l_DisplayableLinks_.values().iterator() ; ite.hasNext() ; ) {
        for (Iterator ite2 = ((List) ite.next()).iterator() ; ite2.hasNext() ; ) {
          ((DisplayableLinkImpl)ite2.next()).finalizeForGui();
        }
    } // for ite
    m_l_DisplayableLinks_.clear();
    l_rightsRelatedBaseObjects_ = null;
      
    // build up the map of (EP, Resource) pair linked by displayable links
    // throws InterruptedException
    m_l_DisplayableLinks_ = facade.detectEPRRights(
          this, (EPRViewInBase)viewInBase_,
          detectAcl_, detectUserGroupLinks_, true, true);

    // construct the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    Object[] l_LinksAndRights = buildUpArrowsAndRights(
            m_l_DisplayableLinks_, m_GraphicNodesBounds_);
    
    // if the event dispatch thread tries to repaint this instance, it waits
    // on drawArrowsAndRights() which uses l_LinksAndRights_
    synchronized (l_LinksAndRights_) {
        l_LinksAndRights_ = l_LinksAndRights;
    }


                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("in GraphicEPRView.resetRights(), "+
                      //"the DisplayableLinkImpls before repaint() are:");
                    //for (Iterator ite = m_l_DisplayableLinks_.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)m_l_DisplayableLinks_.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

    l_LinksAndRights = null;
    mediator = null;
    facade = null;
  } // resetRights()  

  /**
   * These BaseObjects are provided by the gWork classes, and cannot be view objects.
   * Called by the gWork core algorithms.
   * Overrides the super method to remove the view BaseObjects
   * @param _l_baseObjects BaseObject set. May be null.
   */
   public final void setRightsRelatedBaseObjects(HashSet _l_baseObjects) {
       if (_l_baseObjects == null) 
           l_rightsRelatedBaseObjects_ = new HashSet(INITIAL_CAPACITY*10);
       else {
           l_rightsRelatedBaseObjects_ = _l_baseObjects;
           l_rightsRelatedBaseObjects_.removeAll(viewInBase_.getEorM_BaseObjects().values());
       } // else

                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("GraphicEPRView.setRightsRelatedBaseObjects() with: ");
                    //for (BaseObject obj : l_rightsRelatedBaseObjects_ ) {
                      // System.out.println("------------"+
                        //  obj.toString() );
                    //} // for obj
     
   } // setRightsRelatedBaseObjects()
  
  /**
  * Check if there is at most one DisplayableLink which has all of 
  * its AccessControlLinks with a strengthen relation (OWNER, HAS_MEMBER,
  * HAS_CHILD, IS_ALIAS, RUN_UNDER).
  * The typical case is a simple dLink with one
  * strengthen relation betwen the two nodes.
  * Called by buildUpArrowsAndRights().
  * @param _l_links list of DisplayableLinkImpl
  * @return true if has one strengthen links in one dLink.
  * */
  public final boolean hasStrengthenLinksInOneDLink(List _l_links) {
    boolean result = false;
    int count = 0;
    for (Iterator ite = _l_links.iterator(); ite.hasNext(); ) {
        if (((DisplayableLinkImpl)ite.next()).containsStrengthenLinkTypes() ) {
            result = true;
            
            // for testing and tracking
            //System.out.println("in GraphicEPRView.hasStrengthenLinksInOneDLink(),"+
              //      " this is a strengthen dLink "+(DisplayableLinkImpl)_l_links.get(count));            
            
            break;
        }
        else count++;
    } // for
    
    return result;
  } // hasStrengthenLinksInOneDLink()

  /**
  * Checks up if there is at most one DisplayableLink which has 
  * a first rate link as last AccessControlLink.
  * The arrow for this couple of nodes is then drawn with a double width by GraphicView.
  * The typical case is a simple dLink with one NEG_ACL relation
  * betwen the two nodes. This is a first-rate relation because it defines
  * the true rights between the ends, overriding the meaning of the other
  * indirect paths through other nodes in the view.
  * These indirect paths are also drawn.
  * Called by buildUpArrowsAndRights().
  * Calls AccessControlLinkImpl.isFirstRateLink().
  * See gWork.FactoryUtilities.getDLinksFromNegativePaths()
  * as an exemple of the creation of such first-rate links.
  * @param _l_links list of DisplayableLinkImpl
  * @return true if there is a first-rate link in the argument.
  * */
  public final boolean hasFirstRateLinkInOneDLink(List _l_links) {
    boolean result = false;
    int count = 0;
    DisplayableLinkImpl dLink;
    
    for (Iterator ite = _l_links.iterator(); ite.hasNext(); ) {
        dLink = (DisplayableLinkImpl)ite.next();
        if (dLink.getEorLastLink().isFirstRateLink()) {
            result = true;
            
                        // for testing and tracking
                        //System.out.println("in GraphicEPRView.hasFirstRateLinkInOneDLink(),"+
                          //      " this is a first-rate dLink from \n"+
                            //    ((DisplayableLinkImpl)_l_links.get(count)).getEorFirstEnd()+" to "+
                              //  ((DisplayableLinkImpl)_l_links.get(count)).getEorSecondEnd() );
                        
            break;
        }
        else count++;
    } // for
    return result;
  } // hasFirstRateLinkInOneDLink()
  
  /**
  * Checks if there is at most one DisplayableLink which has all of
  * its AccessControlLinks with a OWNER or CONTAIN relation.
  * The typical case is a simple dLink with one
  * CONTAIN relation betwen the two nodes, without a Resource as second end since
  * they may use these links from GroupIDMembers or Actors.
  * Return true for a link typically from an Actor to its current UserID or GroupID.
  * Called by buildUpArrowsAndRights().
  * @param _l_links list of DisplayableLinkImpl
  * @return true if contain links in one dLink.
  * */
  public final boolean hasOwnerContainLinkToGroupIDMember(List _l_links) {
    boolean result = false;
    int count = 0;
    DisplayableLinkImpl dLink;
    
    for (Iterator ite = _l_links.iterator(); ite.hasNext(); ) {
        dLink = (DisplayableLinkImpl)ite.next();
        if (dLink.containsOwnerContainLinkTypes() &&
                !(dLink.getEorSecondEnd() instanceof Resource)) {
            result = true;
            
            // for testing and tracking  
            //System.out.println("in GraphicEPRView.hasOwnerContainLinkToGroupIDMember(),"+
                    //" this is a contain dLink from \n"+
                    //((DisplayableLinkImpl)_l_links.get(count)).getEorFirstEnd()+" to "+
                    //((DisplayableLinkImpl)_l_links.get(count)).getEorSecondEnd() );            
                        
            break;
        }
        else count++;
    } // for
    return result;
  } // hasOwnerContainLinkToGroupIDMember()

  //
  //  --------------- private methods ------------------------------------
  //
  
  /**
  * Sets each side of an arrow head as a Line2D at the right position.
  * Called by buildUpArrowsAndRights() from this class and GraphicNoThanView.
  * @param _pointOne starting point of the arrow
  * @param _pointTwo ending point of the arrow
  * @param _aboveLine true to set the superior side of the arrow head, else set
  * the inferior side
  * @return one of the two sides of the arrow head
  * */
  protected static final Line2D.Float setArrowHead(Point2D.Float _pointOne,
    Point2D.Float _pointTwo, boolean _aboveLine) {
      // variables
      Line2D.Float  result = null;
      double        arrowAngle = Math.atan2(
                  _pointTwo.getY() - _pointOne.getY() ,
                  _pointTwo.getX() - _pointOne.getX() );

      if (_aboveLine)
          result = new Line2D.Float((float)_pointTwo.getX()
                  - ARROW_HEAD_LENGTH*((float)Math.cos(- arrowAngle + ARROW_HEAD_ANGLE)),
                 (float)_pointTwo.getY()
                 + ARROW_HEAD_LENGTH*((float)Math.sin(- arrowAngle + ARROW_HEAD_ANGLE)),
                 (float)_pointTwo.getX(),
                 (float)_pointTwo.getY());
      else
          result = new Line2D.Float((float)_pointTwo.getX()
                  - ARROW_HEAD_LENGTH*((float)Math.cos(- arrowAngle - ARROW_HEAD_ANGLE)),
                 (float)_pointTwo.getY()
                 + ARROW_HEAD_LENGTH*((float)Math.sin(- arrowAngle - ARROW_HEAD_ANGLE)),
                 (float)_pointTwo.getX(),
                 (float)_pointTwo.getY());

      return result;
  } //setArrowHead()
  
  /**
  * NO USE.
  * Forms the true shape of the rights for an arrow, on one line.
  * If rights length is superior to 14, cuts in two lines at the next
  * space character after the index 10.
  * Synchronized on m_l_DisplayableLinks_.
  * @param _g2
  * @param _rights String.
  * @param _point Point2D.Float where the rights have to be drawn.
  * @return Shape of the rights.
  * */
  private static final Shape formRightsShape(Graphics2D _g2, String _rights,
    Point2D.Float _point) {
      // variables
      GeneralPath     doubleRights;
      int             index;
      Font            font = new Font("Serif", Font.BOLD, 13);//Desktop.getInstance().getGraphics().getFont();
      TextLayout      text;
      AffineTransform transf;
      Shape           rightsShape;
      String          rights;

      /*
      if (_rights.length() > 14) {
        index = _rights.indexOf(" ", 10);
        doubleRights = new GeneralPath();

        text = new TextLayout(_rights.substring(0,index), font, _g2.getFontRenderContext());
            //new FontRenderContext(new AffineTransform(), true, false));
        transf = new AffineTransform();
        transf.translate( _point.getX(), _point.getY());
        rightsShape = text.getOutline( transf );
        doubleRights.append(rightsShape , false);

        text = new TextLayout(_rights.substring(index,_rights.length()), font,
            _g2.getFontRenderContext());
            //new FontRenderContext(new AffineTransform(),true, false));
        transf = new AffineTransform();
        transf.translate( _point.getX(), _point.getY() + 10);
        rightsShape = text.getOutline( transf );
        doubleRights.append(rightsShape , false);

        rightsShape = (Shape)doubleRights;
      }
      // else, _rights in one line
      else {
      */
      if (_rights.length() > 14) {
        index = _rights.indexOf(" ", 10);
        rights = _rights.substring(0, index)+" ...";
      }
      else rights = _rights;

      text = new TextLayout(rights, font, _g2.getFontRenderContext());
          //new FontRenderContext(new AffineTransform(), true, false));
      transf = new AffineTransform();
      transf.translate( _point.getX(), _point.getY());
      rightsShape = text.getOutline( transf );
      //}
      doubleRights = null;
      index = 0;
      font = null;
      text = null;
      transf = null;
      rights = null;

      return rightsShape;
  } //formRightsShape()

} // GraphicEPRView


