//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;

import ARoad0.gWorkInterface.RightsMediator;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.border.*;

import java.util.Map;
import java.util.List;
import java.util.Set;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.HashSet;

import ARoad0.Gui1.*;
import ARoad0.ErrW.*;
import ARoad0.gWork.*;
import ARoad0.CNot.*;
import ARoad0.gBaseInterface.*;
import ARoad0.gBase.NameImpl;

/**
 * This class is responsible for displaying a set of Sources and one
 * access Target, and to show if all the rights of all the sources on this target are
 * above or not a No-More-Than right criterion, and below or not a No-Less-Than
 * right criterion. Using StringRight.compareTo() between the rights, the
 * right internal key is the first compared property if the rights have the same sens.
 * Thus, this method may detect the criterion right is not fulfilled while
 * the effective right has no direct path to the criterion right, in the network of rights.
 * This is a defensive behavior, which may be considered too restrictive in some cases.
 * On the other hand, the Access Road user gets the effective right
 * in the 'See why' text. He may so decide the gravity to get no fulfillment. <P>
 * The target is displayed at the top of the view. It is processed as another
 * BaseObject to show in a GraphicNode, as usual for GraphicViews. If it is null,
 * the criteria rights are not displayed.
 * The two criteria rights use two GraphicTexts to switch the color between green and
 * red, following the result of each rights analysis. <P>
 * This class overrides some GraphicView methods: resetGraphicView(), drawArrowsAndRights().
 * It overrides several methods of its parent GraphicEPRView: initialize(),
 * buildUpArrowsAndRights(), setGraphicNodes(), resetRights(), setWhyText().<P>
 * It has some specific methods for allowing its GraphicViewListener to detect any change
 * in the criteria states. This is used for instance to update the frame icon
 * of the GraphicNoThanView frame.
 * It has also a specific method to detect when an Source has an excluded Source,
 * to draw this Source node in gray rather than in blue.
 * See the superclass GraphicView to understand this class.
 * @see ARoad0.gBase.NoThanViewInBaseImpl
 * @see GraphicView
 * @see GraphicEPRView
 * @version 0.7.0
 */
public class GraphicNoThanView extends GraphicView {
    
  /**
   * The GraphicText which displays the more criterion right in green or red.
   * It is a Graphicnode but is not in m_GraphicNodes_, m_GraphicNodesBounds_
   * since it cannot be moved.
   */
  protected GraphicText                     moreCriterion_;
  
  /**
   * The GraphicText which displays the less criterion right in green or red.
   * It is a Graphicnode but is not in m_GraphicNodes_, m_GraphicNodesBounds_
   * since it cannot be moved.
   */
  protected GraphicText                     lessCriterion_;
  
  /**
   * The key in GraphicView.m_GraphicNodesBounds_ for the GraphicText of
   * the no-more-than right.
   */
  protected static final String             NO_MORE_RIGHT_GRAPHICTEXT_BOUNDS_KEY = 
                                                "<No-More-Than right>";

  /**
   * The key in GraphicView.m_GraphicNodesBounds_ for the GraphicText of
   * the no-less-than right.
   */
  protected static final String             NO_LESS_RIGHT_GRAPHICTEXT_BOUNDS_KEY =
                                                "<No-Less-Than right>";

  // for the Sources rectangle of at most 9 Sources
  protected static final int                      EP_ORIGINE_WIDTH = 20;
          //(int)(GraphicNode.COLUMN_SPACE + GraphicNode.NODE_SIZE);
  protected static final int                      EP_ORIGINE_HEIGHT =
          (int)(GraphicNode.ROW_SPACE*2 + GraphicNode.NODE_SIZE*2);
  // for the unique target
  protected static final int                      TARGET_ORIGINE_WIDTH =
          (int)(GraphicNode.COLUMN_SPACE + GraphicNode.NODE_SIZE + EP_ORIGINE_WIDTH - 5);
  protected static final int                      TARGET_ORIGINE_HEIGTH = 10;
  // for the double vertical arrow
  protected static Point2D.Float            ARROW_END = new Point2D.Float(
            GraphicNode.COLUMN_SPACE + GraphicNode.NODE_SIZE*23/16 + EP_ORIGINE_WIDTH,
            GraphicNode.ROW_SPACE*2 + 2*GraphicNode.NODE_SIZE/3); // +
  protected static final Point2D.Float            ARROW_START = new Point2D.Float(
            GraphicNode.COLUMN_SPACE + GraphicNode.NODE_SIZE*23/16 + EP_ORIGINE_WIDTH,
            GraphicNode.ROW_SPACE*2 + GraphicNode.NODE_SIZE*2);
  protected static final int                      ARROW_WIDTH_SHIFT = 10;
  // for the rights
  protected static final int                      RIGHTS_ORIGINE_WIDTH =
          (int)(GraphicNode.NODE_SIZE + EP_ORIGINE_WIDTH/2);
  protected static final int                      RIGHTS_ORIGINE_HEIGTH =
          (int)(GraphicNode.ROW_SPACE*2 + GraphicNode.NODE_SIZE);
          
  //
  // ------------- constructors ------------------------------
  //
  /**
   * This constructor is only for the subclasses, if any.
   */
  public GraphicNoThanView() {} 
  
  /**
  * This constructor adds all the _view objects in this panel,
  * each associated to a GraphicNode.
  * Set the JPanel size, its GraphicNodes, the two Sources and excluded Sources rectangles,
  * construct the single arrow from the Sources rectangle to the target,
  * and the No-More-Than/No-Less-Than criteria states to be drawn in paintComponents().
  * The main work is done there by buildUpArrowsAndRights(),
  * so that the paint action is fast. 
  * Any BaseObject updating is done by GraphicViewBaseListenerImpl,
  * which resets the instance through resetRights() and resetGraphicView().
  * Does not call the parent constructor.
  * Called by ActionOpenGraphicView.openViewInGui() in the event dispatch thread.
  * @param _view NoThanViewInBase to associate to this instance.
  * @param _m_positionIndex is the map (name, Integer) returned
  *   by getIndexesOfNodeLocations() at the previous view opening. May be null.
  * @exception ProcessError if there is a loop in the rights searching
  **/
  public GraphicNoThanView(NoThanViewInBase _view, Map _m_positionIndex) 
          throws ProcessError, InterruptedException {
    viewInBase_ = _view;

    // no Layout, because they are not able to return
    // the absolute coordinates of a component.
    this.setLayout(null);

    this.setBorder(new EmptyBorder(BORDER*2,BORDER,BORDER*2,BORDER));

    // sets the GraphicNodes and GraphicTexts in the JPanel
    initialize();
    
    // resets the node positions
    if (_m_positionIndex != null) setIndexesOfNodeLocations(_m_positionIndex);
    
    // constructs the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    l_LinksAndRights_ = buildUpArrowsAndRights(m_l_DisplayableLinks_, m_GraphicNodesBounds_);

    // sets the color
    setBackground(Color.white);

    addMouseListener(this);
    addMouseMotionListener(this);
  } // GraphicNoThanView(EPRViewInBase _view)

  /**
   * The green color for the drawn text is InternalFrame.FORECOLOR_3, else
   * it is a red color: InternalFrame.FORECOLOR_4. A null target does not
   * change the result.
   * @return _true if the current text color is green for the more criterion,
   * false if it is red
   */
   public final boolean isGreenForMore() {
    return moreCriterion_.isGreenCriterionColor();
   } // isGreenForMore() 

  /**
   * The green color for the drawn text is InternalFrame.FORECOLOR_3, else
   * it is a red color: InternalFrame.FORECOLOR_4. A null target does not
   * change the result.
   * @return true if the current text color is green for the less criterion,
   * false if it is red
   */
   public final boolean isGreenForLess() {
    return lessCriterion_.isGreenCriterionColor();
   } // isGreenForLess() 

  
  /**
   * Provides the comments that the program generates in this method
   * to explain the view rights. It explains the first access path which does
   * not fulfill one of the view criteria.
   * This method is also used to test if the instance has been finalized.
   * Its height is used by the GraphicViewBaseListener.
   * Overrides the GraphicEPRView method
   * @return the text which explains the rights in the view. Null if finalized.
   *   Empty at the GraphicNoThanView initialization.
   */
   public final JTextPane setWhyText() {
        if (whyText_ == null) return null;

        RightsMediator        mediator = ViewModel.getViewRightsMediator(viewInBase_);
        RightsFactory_Facade  facade = null;

        if (mediator != null)
            facade = (RightsFactory_Facade) mediator.getCoreAlgorithm("Facade");

        if (mediator == null || facade == null)
            throw new InternalError("GraphicNoThanView.setWhyText() does not found the "+
                    "RightsFactory_Facade instance for the view "+viewInBase_.getNickName());

       // construct the why text
       whyText_ = facade.getNoThanWhyText(this, m_l_DisplayableLinks_,
                    ((NoThanViewInBase)viewInBase_));

       mediator = null;
       facade = null;
       return whyText_;
   } // setWhyText()

  /** 
  * No operation in any case.
  * Overrides the GraphicView method
  * @param _e is the event
  * */
  public void mouseReleased(MouseEvent _e) {

                    //System.out.println("GraphicNoThanView.mouseReleased(): "
  } // mouseReleased

  /**
  * Draws the two right colors after a change in the rights,
  * without changing in GraphicNodes and the arrow.<P>
  * This class is designed to be updated, as any JComponent, by the event dispatch thread
  * for most of the operations, but it allows the rights updating in a worker thread, through
  * the call of this method.<P>
  * Called in a worker thread by GraphicViewBaseListener.changeTabView()
  * and changeFrameView(), GraphicViewPropertyEditor.propertyChange().
  * Calls RightsFactory_Facade.detectNoThanRights().
  * This method must be followed by a call to repaint() in the event dispatch thread.
  * Overrides the super method.
  * @exception ProcessError if there is a loop in the rights search 
  * */
  public synchronized final void resetRights() throws ProcessError, InterruptedException {
    RightsMediator        mediator = ViewModel.getViewRightsMediator(viewInBase_);
    RightsFactory_Facade  facade = null;

    if (mediator != null)
        facade = (RightsFactory_Facade) mediator.getCoreAlgorithm("Facade");

    if (mediator == null || facade == null)
        throw new InternalError("GraphicNoThanView.resetRights() does not found the "+
                "RightsFactory_Facade instance for the view "+viewInBase_.getNickName());
    
                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("call to GraphicNoThanView.resetRights() "); 
      
      // clear m_l_DisplayableLinks_ to cut any link to the current dLinks
    for (Iterator ite = m_l_DisplayableLinks_.values().iterator() ; ite.hasNext() ; ) {
        for (Iterator ite2 = ((List) ite.next()).iterator() ; ite2.hasNext() ; ) {
          ((DisplayableLinkImpl)ite2.next()).finalizeForGui();
        }
    } // for ite
    m_l_DisplayableLinks_.clear();
    l_rightsRelatedBaseObjects_ = null;
      
    // build up the access paths
    m_l_DisplayableLinks_ = facade.detectNoThanRights(this, (NoThanViewInBase)viewInBase_);

    // do not construct the arrow shapes and l_currentArrowsAndRightsStarts_
    // since there is no change, but reset the criteria rights
    // with a lock on the updated criterion to avoid a conflict with the event dispatch thread
    synchronized (moreCriterion_) {
        synchronized (lessCriterion_) {
            resetCriteria(m_l_DisplayableLinks_);
        }
    }

                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("GraphicNoThanView.resetRights(), "+ 
                      //  "the DisplayableLinkImpls before repaint() are: ");
                    //for (Iterator ite = m_l_DisplayableLinks_.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)m_l_DisplayableLinks_.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite                        

    mediator = null;
    facade = null;
  } // resetRights()    

  /**
   * These BaseObjects are provided by the gWork classes, and cannot be view objects.
   * Called by the gWork core algorithms.
   * Overrides the super method to remove the view BaseObjects.
   * @param _l_baseObjects BaseObject set. May be null.
   */
   public final void setRightsRelatedBaseObjects(HashSet _l_baseObjects) {
       if (_l_baseObjects == null) 
           l_rightsRelatedBaseObjects_ = new HashSet(INITIAL_CAPACITY*10);
       else {
           l_rightsRelatedBaseObjects_ = _l_baseObjects;
           l_rightsRelatedBaseObjects_.removeAll(viewInBase_.getEorM_BaseObjects().values());
           l_rightsRelatedBaseObjects_.remove(((NoThanViewInBase)viewInBase_).
                   getEorFirstAccessTarget());
       } // else

                        // for testing and tracking
                    //ImmutableName key = null;       
                    //System.out.println("GraphicEPRView.setRightsRelatedBaseObjects() with: ");
                    //for (Iterator ite = l_rightsRelatedBaseObjects_.iterator() ; ite.hasNext() ; ) {
                      // System.out.println("------------"+
                        //  ((BaseObject)ite.next()).toString() );
                    //} // for ite       
     
   } // setRightsRelatedBaseObjects()

  /**
   * Invoked by resetGraphicView() and ViewTreeBaseListenerImpl to finalize the instance
   *  and set is a a new object before the initialization, with the exception of
   * whyText_ that is set to null to indicate the finalizing operation.
   * Overrides the super method and calls it.
   */
   public synchronized final void finalizeForGui() {

          // for testing and tracking
          //System.out.println("Call of GraphicNoThanView.finalizeForGui()");

      remove(moreCriterion_);
      remove(lessCriterion_);
      moreCriterion_ = null;
      lessCriterion_ = null;
      super.finalizeForGui();;
   } // finalizeForGui()
   
  //
  // --------------------- protected methods ------------------------------------------
  //
  
  /**
  * Build up the links and rights to put in l_LinksAndRights_,
  * to be painted by GraphicView.drawArrowsAndRights().
  * There are 3 Line2Ds for the arrow, one String for each of the two titles and
  * one GraphicText for each of the two right criteria. The GraphicTexts are not used
  * in drawArrowsAndRights().
  * Called by the contructor in this class and by GraphicView.resetGraphicView().
  * Not called by resetRights() and the super class method GraphicView.mouseReleased()
  * which are overridden.
  * Calls the static method GraphicEPRView.setArrowHead(). The absolute position of the
  * GraphicText is set in the initialize() method, and not there like for the other objects.
  * Overrides the super method.
  * @param _m_l_DisplayableLinks is the map of DisplayableLinks lists (one per pair)
  * associated to the view.
  * @param _m_gNodeBoundsModel is the map of the node bounds for the shapes
  * in the GraphicNoThanView. No use there.
  * @return Object[] to draw in paintComponent(), with both the arrow
  * at absolute positionning (1 Shape for the line and 2 Shapes for the head)
  * and the titles and criteria rights (1 String and 1 GraphicText).
  * */
  protected final Object[] buildUpArrowsAndRights(Map _m_l_DisplayableLinks,
    Map _m_gNodeBoundsModel) {
          
                            // for testing and tracking
                        //System.out.println("call to GraphicNoThanView.buildUpArrowsAndRights()");

    // variables
    l_currentArrowsAndRightsStarts_ = new Object[7];
    Object[]            result = new Object[7];
    
    synchronized (_m_l_DisplayableLinks) { // may be updated

        result[0] = new Line2D.Float(ARROW_START, ARROW_END);
        l_currentArrowsAndRightsStarts_[0] = result[0];

        result[1] = GraphicEPRView.setArrowHead(ARROW_START, ARROW_END, true);
        l_currentArrowsAndRightsStarts_[1] = result[1];

        result[2] = GraphicEPRView.setArrowHead(ARROW_START, ARROW_END, false);
        l_currentArrowsAndRightsStarts_[2] = result[2];

        result[3] = "No-More-Than Right:";
        l_currentArrowsAndRightsStarts_[3] = new Point2D.Float(
                RIGHTS_ORIGINE_WIDTH, RIGHTS_ORIGINE_HEIGTH);

        result[4] = "No-Less-Than Right:";
        l_currentArrowsAndRightsStarts_[4] = new Point2D.Float(
                RIGHTS_ORIGINE_WIDTH, RIGHTS_ORIGINE_HEIGTH + GraphicNode.ROW_SPACE*3);

        // no use in drawArrowsAndRights()
        result[5] = moreCriterion_;
        // no use in drawArrowsAndRights()
        result[6] = lessCriterion_; 

        resetCriteria(_m_l_DisplayableLinks);
    } // synchro    

    return result;
  } // buildUpArrowsAndRights()

  /**
   * Called only by resetGraphicView(). Overrides GraphicView to do nothing
   * since the initialize() method does the job.
   * Overrides the super method.
   * @param _newNodeName name of the base object to add. May be null.
   * @param _m_GraphicNodesPosition is the map of the old locations. May be empty.
   */
  protected final void setGraphicNodeLocations(ImmutableName _newNodeName,
          Map _m_GraphicNodesPosition) {
       
  } // setGraphicNodeLocations()
  
  /** 
  * Sets the criterion right and the green/red color on each right criteria.
  * If the view target is null or if there is no view Sources,
  * the GraphicText BaseObjects are set to null and
  * the criteria rights are not visible in the view.
  * Called by buildUpArrowsAndRights() and resetRights().
  * @param _m_l_DisplayableLinks is the map of DisplayableLinks lists (one per pair)
  * associated to the view. If empty, the colors are green.
  * */
  protected final void resetCriteria(Map _m_l_DisplayableLinks) {

                    //System.out.println("GraphicNoThanView.resetCriteria(): "
        
    if (((NoThanViewInBase)viewInBase_).getEorFirstAccessTarget() != null &&
            !viewInBase_.getEorM_BaseObjects().isEmpty()) {

                                // for testing and tracking
                            //System.out.println("In GraphicNoThanView.resetCriteria(), "+
                              //      "GraphicText.setEorBaseObject() for MORE with "+
                                //    ((NoThanViewInBase)viewInBase_).getMoreCriterion());


        moreCriterion_.setEorBaseObject(((NoThanViewInBase)viewInBase_).getMoreCriterion());
        lessCriterion_.setEorBaseObject(((NoThanViewInBase)viewInBase_).getLessCriterion());

    } else {

                                // for testing and tracking
                            //System.out.println("In GraphicNoThanView.resetCriteria(), "+
                              //      "GraphicText.setEorBaseObject(null)");

        moreCriterion_.setEorBaseObject(null);
        lessCriterion_.setEorBaseObject(null);
        return;
    } // else
    
    moreCriterion_.setCriterionColorInGreen(true); // not always drawn if _m_l_DisplayableLinks empty
    lessCriterion_.setCriterionColorInGreen(true);

    List                l_links = new ArrayList(INITIAL_CAPACITY);
    DisplayableLinkImpl link;

    for (Iterator ite = _m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {

        l_links = (List)ite.next();

        // selects only the first DisplayableLinkImpl, since it gives the two ends
        link = (DisplayableLinkImpl) l_links.get(0);

                        // for testing and tracking
                    //System.out.println("In GraphicNoThanView.resetCriteria(), the comments are:");
                    //for (Iterator ita = link.getEorLastLink().getAllComments().iterator() ; ita.hasNext() ; ) {
                      //System.out.println( ita.next().toString() );
                    //}
        
        if (link.getEorLastLink().getLinkTypeComments(AccessControlLink.GLOBAL).
                            contains(AccessControlLink.NO_MORE_COMPLIANCE)) {
            moreCriterion_.setCriterionColorInGreen(false);
        }
        if (link.getEorLastLink().getLinkTypeComments(AccessControlLink.GLOBAL).
                            contains(AccessControlLink.NO_LESS_COMPLIANCE)) {
            lessCriterion_.setCriterionColorInGreen(false);
        }
    } // for ite
    
    l_links = null;  
    link = null;    
  } // resetCriteria()
   
  /**
  * Initializes nodesNumber_ (not used in this clss), m_l_DisplayableLinks_,
  * l_allNodesInLinks_, nodesNumber_, moreCriterion_ and lessCrietria_
  * and set the GraphicNodes locations. In m_GraphicNodesBounds_, moreCriterion_
  * and lessCriterion_ are put with the keys NameImpl("<No-More/Less-Than right>")
  * to be activated through a mouse click.
  * Does not initialize l_NoSourceTargets_, l_accessTargetEP_, noSourceTargetsCount_,
  * sourcesCount_. Called by the constructor and resetGraphicView().
  * Calls RightsFactory_Facade.detectNoThanRights().
  * Synchronized on viewInBase_.
  * Overrides the super method.
  * @exception InterruptedException, ProcessError if there is a loop in the rights search
  * */
  protected final void initialize() throws ProcessError, InterruptedException {
    RightsMediator        mediator = ViewModel.getViewRightsMediator(viewInBase_);
    RightsFactory_Facade  facade = null;

    if (mediator != null)
        facade = (RightsFactory_Facade) mediator.getCoreAlgorithm("Facade");

    if (mediator == null || facade == null)
        throw new InternalError("GraphicNoThanView.initialize() does not found the "+
                "RightsFactory_Facade instance for the view "+viewInBase_.getNickName());

    synchronized (viewInBase_) {

        nodesNumber_ = viewInBase_.getEorM_BaseObjects().size() + 1;

        //l_NoSourceTargets_ = new ArrayList(
        //((EPRViewInBase)viewInBase_).getEorM_NoActorResources().values());

                        // for testing and tracking
                    //System.out.println("In GraphicNoThanView.initialize(), set the more and less criteria");
                    //for (Iterator ite = eprnodesNumber_.getEorM_EligibleParties().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
                    //System.out.println("In GraphicNoThanView.initialize(), the non-ep resources are:");
                    //for (Iterator ite = l_NoSourceTargets_.iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
    
        // sets the more criteria right
        moreCriterion_ = new GraphicText(this, null);
        add(moreCriterion_);
        moreCriterion_.setBounds(
          0,
          0,
          (int)moreCriterion_.getShape().getBounds2D().getWidth(),
          (int)moreCriterion_.getShape().getBounds2D().getHeight());
        moreCriterion_.setLocation(RIGHTS_ORIGINE_WIDTH + 130, RIGHTS_ORIGINE_HEIGTH - 10);
        moreCriterion_.setToolTipText("In RED, when at least ONE right for ONE checked source"+
                " is strictly GREATER THAN this right.");
        // adds to nodeBounds for mouse clickings - does not work
        m_GraphicNodesBounds_.put(
                new NameImpl("NO_MORE_RIGHT_GRAPHICTEXT_BOUNDS_KEY"),
                moreCriterion_.getBounds());
        moreCriterion_.setVisible(true);

        // sets the less criterion right
        lessCriterion_ = new GraphicText(this, null);
        add(lessCriterion_);
        lessCriterion_.setBounds(
          0,
          0,
          (int)lessCriterion_.getShape().getBounds2D().getWidth(),
          (int)lessCriterion_.getShape().getBounds2D().getHeight());
        lessCriterion_.setLocation(RIGHTS_ORIGINE_WIDTH + 130,
                (int)(RIGHTS_ORIGINE_HEIGTH + GraphicNode.ROW_SPACE*3) - 10);
        lessCriterion_.setToolTipText("In RED, when ALL rights for ONE checked source"+
                " are strictly LESSER THAN this right.");
        // adds to nodeBounds for mouse clickings - does not work
        m_GraphicNodesBounds_.put(
                new NameImpl("NO_LESS_RIGHT_GRAPHICTEXT_BOUNDS_KEY"),
                lessCriterion_.getBounds());
        lessCriterion_.setVisible(true);

          // builds up the list of 2 displayable links, for the more and the less criterion,
          // which are the first detected path links to be not compliant to.
          // if empty, this means that the two criteria are followed.
          // calls back setRightsRelatedBaseObjects().
        m_l_DisplayableLinks_ = facade.detectNoThanRights(this, ((NoThanViewInBase)viewInBase_));

                        // for testing and tracking
                        //System.out.println("l_accessTargetEP_: "+l_accessTargetEP_);

          // draws the target and the sources, with source node in grey if
          // it contains an excluded source
        if (nodesNumber_ < 47) setGraphicNodes();
        else {
            // displays a message
            DialogBox.showOkMessage(
              "This view has more than 46 members. It is not displayed in a diagram"+
                ", but you may see all the access pathes in the 'See why' text.");
        }
    } // synchro

    mediator = null;
    facade = null;
  } // initialize()
  
  /**
  * The GraphicNodes there are for the target and the sources,
  * but not for the GraphicTexts which are updated with the rights.
  * Sets the GraphicNodes in a 9-positions layout for the Eligible Parties,
  * and sets the target GraphicNode in the JPanel, with their absolute coordinates.
  * The nodes for GroupIDs are drawn first. GraphicNodes for GroupID or UserID are
  * drawn in grey rather than blue if they are associated to an excluded EP.
  * The excluded sources are not drawn as GraphicNodes; they are in the 'See why' text.
  * Calls super.setGraphicNodesPositions().
  * Called by initialize().
  * Clears then updates m_GraphicNodes_, updates m_GraphicNodesBounds_.
  * Overrides the super method.
  * */
  protected final void setGraphicNodes() {
    //variables
    GraphicNode       gep;
    GraphicNode       gact;
    GraphicNode       gres;
    ImmutableSource   sour;
    int               index = 0;
    ImmutableTarget   targ = ((NoThanViewInBase)viewInBase_).getEorFirstAccessTarget();

    //m_GraphicNodes_.clear(); // remove the WAITING_TEXT if any
    // set a fixed size
    setSize(GraphicNode.TWENTY_FOUR_NODES_VIEW_SIZE);

    // set a grid of 25 nodes for the global positionning - usefull?
    setGraphicNodesPositions(10, false);    
    
    // set a grid of 9 nodes positions with their coordinates from the view origine point,
    // to use for the Sources, not for the global positionning
    Point[]           l_sourcePositions = getGraphicNodesPositionsForSources();

                        // for testing and tracking    
                    //System.out.println("GraphicNoThanView.setGraphicNodes():");
                    //System.out.println("for "+viewInBase_.getName()+", nodesNumber_: "+nodesNumber_);
                    //System.out.println("l_positions:");
                    //for (int i = 0; i < l_positions.length; i++ ) {
                      //if (l_positions[i] != null) System.out.println("--"+i+": "+ l_positions[i].toString() );
                    //}
                    //System.out.println("l_noAccessTargetSources:");
                    //for (Iterator itetest = l_noAccessTargetSources.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_accessTargetEP_:");
                    //for (Iterator itetest = l_accessTargetEP_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_NoSourceTargets_:");
                    //for (Iterator itetest = l_NoSourceTargets_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}

    // first loop only for GroupID
    for (Iterator ite = viewInBase_.getEorM_BaseObjects().values().iterator() ; ite.hasNext(); ) {

        sour = (ImmutableSource) ite.next();
        if (sour instanceof GroupID) gact = new GraphicGroup(this, (GroupID)sour );
        else continue;
        // adds to this JPanel
        add(gact );

        // puts it at the right place, in the Sources rectangle which has its origine
        // at (epOrigineWidth, epOrigineHeight)
        gact.setBounds(
          0,
          0,
          (int)gact.getShape().getBounds2D().getWidth(),
          (int)gact.getShape().getBounds2D().getHeight());
        
        l_sourcePositions[index].translate(EP_ORIGINE_WIDTH, EP_ORIGINE_HEIGHT);
        l_sourcePositions[index].translate((int)GraphicNode.COLUMN_SPACE/4, 0);
        gact.setLocation(l_sourcePositions[index]);

        // adds nodeBounds to its map
        m_GraphicNodesBounds_.put(sour.getName(), gact.getBounds());
        // adds nodes to its map
        m_GraphicNodes_.put(sour.getName(), gact);
        gact.setVisible(true);
        index++;

    } // for (Iterator ite
    
    // second loop for no-GroupIDs
    for (Iterator ite = viewInBase_.getEorM_BaseObjects().values().iterator() ; ite.hasNext(); ) {

        sour = (ImmutableSource) ite.next();
        if (sour instanceof GroupID) continue;
        else if (sour instanceof Actor) gact = new GraphicActor(this, (Actor)sour );
        else if (sour instanceof UserID) gact = new GraphicAccount(this, (UserID)sour );
        else gact = new GraphicVirtualFolder(this, (VirtualFolder)sour);

        // add to this JPanel
        add(gact );

        // puts it at the right place, in the Sources rectangle which has its origine
        // at (epOrigineWidth, epOrigineHeight)
        gact.setBounds(
          0,
          0,
          (int)gact.getShape().getBounds2D().getWidth(),
          (int)gact.getShape().getBounds2D().getHeight());
        
        l_sourcePositions[index].translate(EP_ORIGINE_WIDTH, EP_ORIGINE_HEIGHT);
        l_sourcePositions[index].translate((int)GraphicNode.COLUMN_SPACE/4, 0);
        gact.setLocation(l_sourcePositions[index]);

        // adds nodeBounds to its map
        m_GraphicNodesBounds_.put(sour.getName(), gact.getBounds());
        // adds nodes to its map
        m_GraphicNodes_.put(sour.getName(), gact);
        gact.setVisible(true);
        index++;

    } // for (Iterator ite
    
    if (((NoThanViewInBase)viewInBase_).getEorFirstAccessTarget() == null) {
        gep = null;
        gact = null;
        gres = null;
        sour = null;
        l_sourcePositions = null;

        return;        
    } // if

                        // for testing and tracking    
                    //System.out.println("GraphicNoThanView.setGraphicNodes() with a target, in "+
                      //      viewInBase_.getName());

    // sets the access target at the top of the view -----------------------------
    if (targ instanceof Resource)
        gres = new GraphicResource(this, (Resource)targ);
    else
        gres = new GraphicVirtualFolder(this, (VirtualFolder)targ);
    add(gres );
    gres.setBounds(
      0,
      0,
      (int)gres.getShape().getBounds2D().getWidth(),
      (int)gres.getShape().getBounds2D().getHeight());
    gres.setLocation(TARGET_ORIGINE_WIDTH, TARGET_ORIGINE_HEIGTH);
    //gres.setToolTipText("The target");
    // adds nodeBounds and nodes to their maps
    m_GraphicNodesBounds_.put(
            ((NoThanViewInBase)viewInBase_).getEorFirstAccessTarget().getName(), 
            gres.getBounds());
    m_GraphicNodes_.put(
            ((NoThanViewInBase)viewInBase_).getEorFirstAccessTarget().getName(), 
            gres );
    gres.setVisible(true);
    
    // sets the Sources rectangle ----------------------------------------------
    gep = new GraphicContainer();
    ((GraphicContainer)gep).SHAPE = new Rectangle(
            (int)(GraphicNode.COLUMN_SPACE*3 + GraphicNode.NODE_SIZE*2), 
            5);
    add(gep);
    gep.setBounds(
      0,
      0,
      (int)gep.getShape().getBounds2D().getWidth(),
      (int)gep.getShape().getBounds2D().getHeight());
    gep.setLocation(EP_ORIGINE_WIDTH, EP_ORIGINE_HEIGHT);
    gep.setToolTipText("These user rights are checked,"+
            " and their members/running-under actors too, "+
            "but not the excluded users.");
    gep.setVisible(true);
        // second line
    gep = new GraphicContainer();
    ((GraphicContainer)gep).SHAPE = new Rectangle(
            5, 
            (int)(GraphicNode.ROW_SPACE*3 + GraphicNode.NODE_SIZE*3));
    add(gep);
    gep.setBounds(
      0,
      0,
      (int)gep.getShape().getBounds2D().getWidth(),
      (int)gep.getShape().getBounds2D().getHeight());
    gep.setLocation(EP_ORIGINE_WIDTH, EP_ORIGINE_HEIGHT);
    gep.setToolTipText("These user rights are checked,"+
            " and their members/running-under actors too, "+
            "but not the excluded users.");
    gep.setVisible(true);
        // third line
    gep = new GraphicContainer();
    ((GraphicContainer)gep).SHAPE = new Rectangle(
            5, 
            (int)(GraphicNode.ROW_SPACE*3 + GraphicNode.NODE_SIZE*3));
    add(gep);
    gep.setBounds(
      0,
      0,
      (int)gep.getShape().getBounds2D().getWidth(),
      (int)gep.getShape().getBounds2D().getHeight());
    gep.setLocation(
            EP_ORIGINE_WIDTH + (int)(GraphicNode.COLUMN_SPACE*3 + GraphicNode.NODE_SIZE*2), 
            EP_ORIGINE_HEIGHT);
    gep.setToolTipText("These user rights are checked,"+
            " and their members/running-under actors too, "+
            "but not the excluded users.");
    gep.setVisible(true);
        // fourth line
    gep = new GraphicContainer();
    ((GraphicContainer)gep).SHAPE = new Rectangle(
            (int)(GraphicNode.COLUMN_SPACE*3 + GraphicNode.NODE_SIZE*2) + 5, 
            5);
    add(gep);
    gep.setBounds(
      0,
      0,
      (int)gep.getShape().getBounds2D().getWidth(),
      (int)gep.getShape().getBounds2D().getHeight());
    gep.setLocation(
            EP_ORIGINE_WIDTH, 
            EP_ORIGINE_HEIGHT + (int)(GraphicNode.ROW_SPACE*3 + GraphicNode.NODE_SIZE*3));
    gep.setToolTipText("These user rights are checked,"+
            " and their members/running-under actors too, "+
            "but not the excluded users.");
    gep.setVisible(true);
    
    gep = null;
    gact = null;
    gres = null;
    sour = null;
    targ = null;
    l_sourcePositions = null;
  } // setGraphicNodes()
 
  /**
   * Called by GraphicGroup, GraphicAccount, GraphicActor.
   * @param _sour an access source
   * @return true if _sour is a reduced EP in the NoThanViewInBase
   */
  protected final boolean isReducedSource(ImmutableSource _sour) {
    Set<ImmutableSource> l_reduced = ((NoThanViewInBase)viewInBase_).
            getReducedSourcesInViewSources();
    boolean result = false;
    
    result = l_reduced.contains(_sour);
    l_reduced = null;
    return result;  
  } // isReducedEP()
  
  //
  // --------------------- private methods ------------------------------------------
  //
  
  /**
  * Draws the single arrow, the comments and rights with BeamerManager.plainFont
  * from the values in l_LinksAndRights_ and l_currentArrowsAndRightsStarts_.
  * The GraphicNodes are painted by paintComponent() and getExportImage() 
  * and they call this method.
  * This method is designed to be fast, since Shapes and Strings to draw has been
  * put in the l_LinksAndRights_ array by buildUpArrowsAndRights() before the call 
  * of this method.
  * The two titles ('No-More-Than right:' and 'No-Less-Than right:') are in dark grey.
  * The two right texts are in green or red, following the compliance with the criterion,
  * but they are not in l_LinksAndRights_, since they are added to the GraphicView by
  * initialize().
  * Uses antialiasing on shapes and texts.
  * Caution: Swing calls always GraphicView.drawArrowsAndRights() before this method,
  * and l_LinksAndRights_ is often null at this moment.
  * Called by paintComponent(). Overrides the super method.
  * Overrides the ViewInBase method.
  * @param _g2 is the Graphics2D to draw
  * */
  protected final void drawArrowsAndRights(Graphics2D _g2) {
    String     text;
    //int        index;
    //Point2D[]  l_points = new Point2D[2];
    //Point2D[]  l_newPoints = new Point2D[2];
    
    _g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING ,
      RenderingHints.VALUE_ANTIALIAS_ON);
    _g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING ,
      RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    _g2.setFont(BeamerManager.plainFont);
    _g2.setStroke(BASIC_STROKE);
    _g2.setPaint(Color.DARK_GRAY);
    
                            // for testing and tracking
                        //System.out.println(" ");
                        //System.out.println("call to GraphicNoThanView.drawArrowsAndRights()");
                        //if (l_LinksAndRights_ != null)
                          //  System.out.println("l_LinksAndRights_.length: "+l_LinksAndRights_.length);
         
    // draws the single vertical arrow from the Sources rectangle to the target
    // with the 3 first values in l_LinksAndRights_, then
    // draws the two titles ('No-More-Than right:' and 'No-Less-Than right:') in grey,
    // and draws the two GraphicTexts for the right criterion
    if (l_LinksAndRights_ != null && l_LinksAndRights_.length == 7) {
      for (int i = 0 ; i < l_LinksAndRights_.length ; i++) {

                            // for testing and tracking             
                        //System.out.println("--------- i = "+i+", type of l_LinksAndRights_[i]: ");
                          //if (l_LinksAndRights_[i] != null) System.out.println("     is "+
                            //  l_LinksAndRights_[i].getClass());
                          //else System.out.println("      l_LinksAndRights_[i] is null ");
          
        if (l_LinksAndRights_[i] instanceof Line2D.Float) {
            
              // draws the two sides of the arrow head,
              if (l_LinksAndRights_.length > i+1 &&
                           l_LinksAndRights_[i+1] instanceof String ) {
                  _g2.draw( (Shape)l_LinksAndRights_[i] );                 
              } else if (l_LinksAndRights_.length > i+2 &&
                           l_LinksAndRights_[i+2] instanceof String ) {
                  _g2.draw( (Shape)l_LinksAndRights_[i] );             
              } else { // draw a line for an arrow

                            // for testing and tracking
                        //System.out.println("    third case of Line2D.Float for i = "+i);
                        //System.out.println(" --- i+1: "+l_LinksAndRights_[i+1]);
                        //System.out.println(" --- i+2: "+l_LinksAndRights_[i+2]);
                        //System.out.println(" --- i+3: "+l_LinksAndRights_[i+3]);
                  
                  _g2.draw( (Shape)l_LinksAndRights_[i] );
              } // else
          
        } // if (l_LinksAndRights_[i]
                        
        else if (l_LinksAndRights_[i] instanceof String) {

          text = (String)l_LinksAndRights_[i];
          
                            // for testing and tracking
                        //System.out.println("call to GraphicNoThanView.drawArrowsAndRights() for String: ");
                        //System.out.println(" --- i: "+i+" rights: "+text);
          
          _g2.drawString(
            text,
            (new Double(((Point2D.Float)l_currentArrowsAndRightsStarts_[i]).getX()))
              .intValue(),
            (new Double(((Point2D.Float)l_currentArrowsAndRightsStarts_[i]).getY()))
              .intValue());
          
        } // if (l_LinksAndRights_[i] instanceof String)
                        
        //else if (l_LinksAndRights_[i] instanceof GraphicText) {
          
                            // for testing and tracking
                        //System.out.println("call to GraphicView.drawArrowsAndRights() for GraphicText: ");
                        //System.out.println(" --- i: "+i+" rights: "+rights);
          
          //_g2.draw(((GraphicText)l_LinksAndRights_[i]).getShape());

        //} // if (l_LinksAndRights_[i] instanceof GraphicText)
                        
      } // for i
    } //if (l_LinksAndRights_ != null)
                        
    text = null;
    //l_points = null;
    //l_newPoints = null;
  } // drawArrowsAndRights()

  /**
  * Gets an array of 9-positions for the layout of the GraphicNoThanView Eligible Parties,
  * with their coordinates from the GraphicView origine point.
  * Called by setGraphicNodes().
  * @return the array of node positions, in the correct order.
   */
  private final Point[] getGraphicNodesPositionsForSources() {
      Point[] l_result = new Point[9];
      for (int i = 0; i < 9; i++) { // 9 positions of the grid
        l_result[i] = new Point(
          (new Double(getInsets().left + GraphicNode.L_NOTHAN_NINE_GRID[i].getX())).intValue(),
          (new Double(getInsets().top + GraphicNode.L_NOTHAN_NINE_GRID[i].getY())).intValue());          
      } // for i 
      
      return l_result;
  } // getGraphicNodesPositionsForSources()
  
} // GraphicNoThanView



