//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.geom.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;

import ARoad0.ErrB.BaseError;
import ARoad0.Gui1.*;
import ARoad0.ErrW.*;
import ARoad0.gWork.*;
import ARoad0.CNot.DisplayableLinkImpl;

import ARoad0.gBaseInterface.*;

/**
 * This class is responsible for displaying an EPRViewInBase in the sketch view
 * of any BaseObject selected in the desktop, which is named the view center.
 * The sketch view displays all the direct access sources and direct access targets
 * of the center, in a diagram where the center sources on the
 * top line, and the center targets at the bottom line under the center node. <P>
 * Sees the parent GraphicEPRView, for a general understanding of this class work.
 * In GraphicSketchView, GraphicEPRView.l_AccessTargetSources_ and l_NoSourceTargets_
 * never contain the center object. GraphicView.l_allNodesInLinks_ is never used
 * since there is no extended view.
 * l_AccessTargetSources_ and l_SourcesForCenter_ are set from the links
 * ViewInBase_Facade.populateSketchView() discovers, while l_NoSourceTargets_
 * is set from the ViewInBase. This means there is never a source without a link
 * to the center, and it is possible to have a target without link from the center.
 * The GraphicView function of node moving and switching is disabled. 
 * As resetGraphicView() is called after a full change of the view nodes,
 * the node storing during a node adding is also disabled. <P>
 * The method setGraphicNodesPositions() produces an empty sketcher in grey, if
 * the number of nodes around the center is superior to 21, or if there are more
 * than 10 sources or more than 10 targets.
 * @see ARoad0.gBase.EPRViewInBaseImpl
 * @version 0.7.0
 */
public class GraphicSketchView extends GraphicEPRView {

  // variables
    /**
     * This is the central object for which the access sources and targets
     * are displayed in the sketch view. The single node at the second line of the view.
     */
  protected BaseObject  center_;
    /**
     * List of sources that are access sources for
     * the center, to put above the center node.
     */
  protected final List  l_SourcesForCenter_ = new ArrayList(INITIAL_CAPACITY);
    /**
     * The size of GraphicEPRView.l_AccessTargetSources_, which does not include the center.
     */
  protected int         accessTargetSourcesCount_;
    /**
     * The size of l_SourcesForCenter_
     */
  protected int         sourcesForCenterCount_;
  
  //
  // ------------- constructors ------------------------------
  //

  /**
  * This constructor prepare a sketch view where a central object is displayed
  * with all its near-access objects around it.
  * On the contrary of the parent classes, this constructor does not provide
  * an operational instance, since a resetGraphicView() call has to come after
  * for defining the center and for obtaining a workable graphic view.
  * No right updating is done by a GraphicViewBaseListenerImpl, and this is
  * a major difference from the super classes.
  * @param _view to associate to this instance.
  **/
  public GraphicSketchView(EPRViewInBase _view) {
    viewInBase_ = _view;

    // no Layout, because they are not able to return
    // the absolute coordinates of a component.
    this.setLayout(null);

    this.setBorder(new EmptyBorder(BORDER*2,BORDER,BORDER*2,BORDER));

    // set the color
    setBackground(InternalFrame.BACKCOLOR_8);
    
    // set the size to prepare the InternalFrame scrollpane
    setPreferredSize(GraphicNode.TWENTYONE_NODES_SKETCH_VIEW_SIZE);
    
    addMouseListener(this);
    //addMouseMotionListener(this);    
  } // GraphicSketchView(EPRViewInBase _view)

  /**
   * This method must be called through resetGraphicView() to have an operational instance.
  * Initializes nodesNumber_, m_l_DisplayableLinks_, set
  * the GraphicNode locations, and initializes l_NoSourceTargets_,
  * l_AccessTargetSources_, l_SourcesForCenter_ and the counts.
  * l_AccessTargetSources_ and l_SourcesForCenter_ are set from the links in
  * _m_l_DisplayableLinks, while l_NoSourceTargets_ is set from the ViewInBase.
  * l_allNodesInLinks_ is not initialized there since it is not used in this class
  * to construct an extended view.
  * Called by resetGraphicView(). This method is derived from the super.initialize()
  * method, but with different arguments.
  * Synchronized on viewInBase_.
  * @param _m_l_DisplayableLinks is the map of DisplayableLinkImpls in this view.
  *   May be null only if center_ is null and viewInBase_ empty.
  * @exception ProcessError if there is a loop in the rights search
  * */
  protected final void initialize(Map _m_l_DisplayableLinks) throws ProcessError {
    // variables
      // one (EP, Resource) pair
    List                  l_link;
    BaseObject            end;
    
                                // for testing and tracking
                            //ImmutableName key = null;
                            //System.out.println("GraphicSketchView.initialize(), "+
                              //"the DisplayableLinkImpls in the argument are:");
                            //if (_m_l_DisplayableLinks != null) {
                              //  for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                                //   key = (ImmutableName) ite.next();
                                  // System.out.println("--- for key: "+key);
                                   //for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                                    //System.out.println("------------"+
                                      //((DisplayableLinkImpl)ita.next()).toString() );
                                  //} // for ita
                                //} // for ite
                            //}
                            //else {
                              //  System.out.println("-- null map ");
                            //}
    
    l_rightsRelatedBaseObjects_ = null;

    synchronized (viewInBase_) {

      nodesNumber_ = viewInBase_.getEorM_BaseObjects().size();

      l_NoSourceTargets_ = new ArrayList(
        ((EPRViewInBase)viewInBase_).getEorM_NoSourceTargets().values());
      // remove center_
      if (l_NoSourceTargets_.contains(center_)) l_NoSourceTargets_.remove(center_);

                        // for testing and tracking
                    //System.out.println("In GraphicSketchView.initialize(), the ep are:");
                    //for (Iterator ite = eprnodesNumber_.getEorM_EligibleParties().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
                    //System.out.println("In GraphicSketchView.initialize(), "+
                      //      "l_NoSourceTargets_ contains:");
                    //for (Iterator ite = l_NoSourceTargets_.iterator() ; ite.hasNext() ; ) {
                      //System.out.println( ite.next().toString() );
                    //}
      
      // build up the map of (BaseObject, center_) pairs linked by displayable links
      if (center_ != null) m_l_DisplayableLinks_ = _m_l_DisplayableLinks;
      else m_l_DisplayableLinks_.clear();

      // updates l_AccessTargetSources_, l_SourcesForCenter_, 
      // and not GraphicView.l_allNodesInLinks_ as in GraphicEPRView.
      // m_l_DisplayableLinks_ may be empty.
      l_AccessTargetSources_.clear();
      l_SourcesForCenter_.clear();
      //l_allNodesInLinks_.clear();
      for (Iterator ita = m_l_DisplayableLinks_.values().iterator() ; ita.hasNext() ; ) {
          l_link = (List)ita.next();

                     // for testing and tracking
                   //System.out.println("----------- for "+center_+
                     //      ", l_link in GraphicSketchView.initialize(): "+l_link);  
                   
          end = ((DisplayableLinkImpl)l_link.get(0)).getEorSecondEnd();
          if (end instanceof ImmutableSource && !l_AccessTargetSources_.contains(end) )
              l_AccessTargetSources_.add(end);

          end = ((DisplayableLinkImpl)l_link.get(0)).getEorFirstEnd();
          if (end instanceof ImmutableSource && !l_SourcesForCenter_.contains(end) )
              l_SourcesForCenter_.add(end);
      } // for (Iterator ita = m_l_DisplayableLinks_

                    // for testing and tracking
                    //System.out.println("l_AccessTargetSources_: "+l_AccessTargetSources_);

      // updates the counts before the call to setGraphicNodes()
      noSourceTargetsCount_ = l_NoSourceTargets_.size();
      sourcesCount_ = ((EPRViewInBase)viewInBase_).getEorM_Sources().values().size();
      accessTargetSourcesCount_ = l_AccessTargetSources_.size();
      
      if (l_AccessTargetSources_.contains(center_)) {
          l_AccessTargetSources_.remove(center_);
          accessTargetSourcesCount_--;
      } // if
      sourcesForCenterCount_ = l_SourcesForCenter_.size();
      if (l_SourcesForCenter_.contains(center_)) {
          l_SourcesForCenter_.remove(center_);
          sourcesForCenterCount_--;
      } // if

                        // for testing and tracking
                    //System.out.println("In GraphicSketchView.initialize(), nodesNumber_ "+
                      //      "before setGraphicNodes(): "+nodesNumber_);
      
      // the sketch view is smaller than the EPR view
      if (nodesNumber_ != 0 && nodesNumber_ < 22) setGraphicNodes(); 
      else nodesNumber_ = -1; //  used by resetGraphicView() for avoiding some operations

    } // synchro

    l_link = null;
    end = null;   
  } // initialize()
  
  //
  // ------------------public methods -----------------------------------------
  //

  /**
   * Called by InternalFrame to finalize the instance and to set is a
   * a new object before the initialization, with the exception of
   * whyText_ that is set to null to indicate the finalizing operation.
   * Overrides the GraphicEPRView method and calls it.
   */
   public synchronized void finalizeForGui() {

                  // for testing and tracking
                  //System.out.println("Call to GraphicSketchView.finalizeForGui()");

      l_SourcesForCenter_.clear();
      center_ = null;
      super.finalizeForGui();
   } // finalizeForGui()

  /**
  * UNUSED - Draws the new arrows and rights after a change in the rights,
  * without changing in GraphicNodes.
  * Not called by GraphicViewBaseListener, since there is no right update.
  * Calls RightsFactory_Facade.detectSketchRights(), buildUpArrowsAndRights().
  * This method call must be followed by a call to repaint() in the event dispatch thread.
  * Override the super method.
  * @exception ProcessError if there is a loop in the rights search 
  * */
  public final void resetRights() throws ProcessError, InterruptedException {
    // clear m_l_DisplayableLinks_ to cut any link to the current dLinks
    for (Iterator ite = m_l_DisplayableLinks_.values().iterator() ; ite.hasNext() ; ) {
        for (Iterator ite2 = ((List) ite.next()).iterator() ; ite2.hasNext() ; ) {
          ((DisplayableLinkImpl)ite2.next()).finalizeForGui();
        }
    } // for ite
    m_l_DisplayableLinks_.clear();
    l_rightsRelatedBaseObjects_ = null;
      
    // build up the map of (BaseObject, center_) pair linked by displayable links
    if (center_ != null) m_l_DisplayableLinks_ = RightsFactory_Facade.
        detectSketchRights(((EPRViewInBase)viewInBase_), center_, true, true);
    else m_l_DisplayableLinks_.clear();

    // construct the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    Object[] l_LinksAndRights = null;
    if (center_ != null) 
        l_LinksAndRights = buildUpArrowsAndRights(m_l_DisplayableLinks_, m_GraphicNodesBounds_);

    // if the event dispatch thread tries to repaint this instance, it waits
    // on drawArroawsAndRights() which uses l_LinksAndRights_
    synchronized (l_LinksAndRights_) {
        l_LinksAndRights_ = l_LinksAndRights;
    }


                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("in GraphicSketchView.resetRights(), "+
                      //"the DisplayableLinkImpls before repaint() are:");
                    //for (Iterator ite = m_l_DisplayableLinks_.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)m_l_DisplayableLinks_.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

    l_LinksAndRights = null;
  } // resetRights()  
  
  /**
   * This method must be called after the constructor to have an operational instance.
  * Builds up the map of (BaseObject, center_) pair
  * linked by DisplayableLinks, then sets the JPanel size, its GraphicNodes,
  * constructs the arrows and the visible rights shapes
  * to be drawn in paintComponents().
  * The main work is done there by buildUpArrowsAndRights(),
  * so that the paint action is fast. 
  * Called by ActionOpenSketcher.openSelection() in the main thread.
  * Calls finalizeForGui(), initialize(), buildUpArrowsAndRights() and repaint().
  * Derived from GraphiView.resetGraphicView().
  * @param _center is the central object of the view for which this sketch view 
  *   is displayed. May be null if the EPRViewInBase has currently no elements.
  * @param _m_l_DisplayableLinks is the map of DisplayableLinkImpls in this view.
  *   No use if _center is null. May be null.
  * @return -1 if the view is too complex to be drawn, or return nodesNumber_
  * @exception ProcessError if there is a loop in the rights search, or InternalError
  *   if the center is not in the view.
  * */
  public final int resetGraphicView(BaseObject _center, Map _m_l_DisplayableLinks)
          throws ProcessError {

                        // for testing and tracking
                        //System.out.println("GraphicSketchView.resetGraphicView(), size of the view:"+
                          //      " "+viewInBase_.getEorM_BaseObjects().size());

                            // for testing and tracking
                        //System.out.println("GraphicSketchView.resetGraphicView() "+
                          //" for the thread "+Thread.currentThread().getName()+
                          //" with the priority: "+Thread.currentThread().getPriority());

    if (_center == null && viewInBase_.getEorM_BaseObjects().size() != 0) 
        throw new InternalError("In the GraphicSketchView.resetGraphicView(),"+
            "the center is null in the non-empty view.");
    else if (_center != null && !viewInBase_.containsBaseObject(_center)) {
    
                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("in GraphicSketchView.resetGraphicView(), "+
                      //"the view objects are:");
                    //for (Iterator ite = viewInBase_.getEorM_BaseObjects().keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                       //System.out.println("------- : "+key);
                    //} // for ite
        
        throw new InternalError("In the GraphicSketchView.resetGraphicView(),"+
            "the center is not in the view.");
    } else if (_center instanceof UserID && ((UserID)_center).getOrder() == 0)
        return -1;
        //throw new ProcessError("In the GraphicSketchView.resetGraphicView(),"+
          //  "the center is an administrative account and it cannot be displayed."); // for actor under 'root'
        

    // initialize
    ViewInBase      view = viewInBase_;
    finalizeForGui();
    // these are the differences between the finalized state and the constructor state
    whyText_ = new JTextPane(); 
    viewInBase_ = view;
    center_ = _center;
    
    // build up the GraphicNodes and the map of pairs, or set 'nodesNumber_ = -1'
    initialize(_m_l_DisplayableLinks);
    
      // construct the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    if (center_ != null && nodesNumber_ != -1)
        l_LinksAndRights_ = buildUpArrowsAndRights(m_l_DisplayableLinks_, m_GraphicNodesBounds_);
    else l_LinksAndRights_ = null;

    
                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("in GraphicSketchView.resetGraphicView(), "+
                      //"the DisplayableLinkImpls before repaint() are:");
                    //for (Iterator ite = m_l_DisplayableLinks_.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)m_l_DisplayableLinks_.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
                    //System.out.println("-- and nodesNumber_: "+nodesNumber_);
    
    repaint();
    
    view = null;
    return nodesNumber_;
  } // resetGraphicView()   
  
  /**
  * Called by mouseClicked() to find the BaseObject in the view on which the user
  * has clicked. In setGraphicNodes(), m_GraphicNodesBounds_ uses a special key
  * which is not the BaseObject name when the BaseObject is two times in the view.
  * The key of the second display in m_GraphicNodes_
  * and m_GraphicNodesBounds_ is BaseObject name extended by '<!!§§!!%%!!%%!!§§>'.
  * So, after a click on the second display of a BaseObject, this method converts
  * the node name to find the right object. Overrides the GraphicView method.
  * @since 0.7.0
  * @param _e is the event
  * @return the selected BaseObject or null
  * */
  protected BaseObject getSelectedObjectOnMouseClicked(MouseEvent _e) {
    BaseObject              selectedObject = null;
    Name                    name = null;
    Rectangle2D             rect;
    boolean                 found = false;

                      // for testing and tracking
                      //System.out.println("GraphicSketchView.getSelectedObjectOnMouseClicked() start");

    // test the current nodes as target
    for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
      name = (Name) ite.next();
      rect = (Rectangle2D)m_GraphicNodesBounds_.get(name);
      //if ( _e.getPoint().distance(rect.getX(), rect.getY()) < GraphicNode.NODE_SIZE) {
      if ( rect.contains(_e.getPoint()) ) {
          found = true;
          break;
      }
    } // for ite

    if (!found) {
        name = null;
        rect = null;

                      // for testing and tracking
                      //System.out.println("in GraphicSketchView.getSelectedObjectOnMouseClicked(), "+
                        //" no selection with the nodes:");
                      //for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
                        //  System.out.println((ImmutableName) ite.next());
                      //}

        return null;
    }

    // the main work in this overriding method
    if (name.lastComponent().equals("<!!§§!!%%!!%%!!§§>"))
        name = name.context();

                      // for testing and tracking
                      //System.out.println("in GraphicSketchView.getSelectedObjectOnMouseClicked(), "+
                        //" selection of the node "+name+" among the nodes");
                      //for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
                        //  System.out.println((ImmutableName) ite.next());
                      //}

    try {
        selectedObject = ACS_Facade.getEorBaseObject(name);
    } catch (BaseError b) {

                      // for testing and tracking
                      //System.out.println("in GraphicSketchView.getSelectedObjectOnMouseClicked(), "+
                        //" the selection is not a known BaseObject. The name is: "+name);

        // display a message
        DialogBox.showErrorMessage(
          "Sorry, the object "+name+" is not found in the base.");
        
        selectedObject = null;
        name = null;
        rect = null;
        return null;
    } // catch

    name = null;
    rect = null;
    return selectedObject;
  } // getSelectedObjectOnMouseClicked()
  
  /** 
  * This method deletes the GraphicView process of switching between nodes.
  * Override the GraphicView method 
  * @param _e is the event
  * */
  public final void mouseReleased(MouseEvent _e) {

  } // mouseReleased

  /**
  * This method deletes the GraphicView process of switching between nodes.
  * Override the GraphicView method 
  * @param _e is the event
  * */
  public final void mouseDragged(MouseEvent _e) {

  } // mouseDragged  
  
  //
  // -------------------- protected methods ------------------------------------
  //
  
  /**
  * Sets the size of the JPanel when the view contains less than 21 elements.
  * The size has three values if _nodesNumber is inferior to 8, to 12 or not. 
  * No operation if it is superior to twenty one or negative, and nodesNumber_
  * is then set to '-1'.
  * Called by setGraphicNodes(). 
  * Overrides the super method and oes not call it.
  * @param _nodesNumber is the number of BaseObjects in the view.
  * */
  protected void setViewSize(int _nodesNumber) {
    if (_nodesNumber < 0 || _nodesNumber > 21) {
        nodesNumber_ = -1;
        return;
    }
    else if (_nodesNumber < 8) {
        setSize(GraphicNode.SEVEN_NODES_SKETCH_VIEW_SIZE);;
    }
    else if (_nodesNumber < 12) {
        setSize(GraphicNode.ELEVEN_NODES_SKETCH_VIEW_SIZE);
    }
    else {
        setSize(GraphicNode.TWENTYONE_NODES_SKETCH_VIEW_SIZE);
    }
  } // setViewSize()  

  /**
  * Sets the GraphicNodes in a 7, 11 or 21 nodes layout
  * in the JPanel, with their absolute coordinates in 3 levels.
  * The center node is the single node at the second level.
  * Called by intialize(). No operation if center_ is null.
  * No operation if the view cannot be processed, and nodesNumber_
  * is then set to '-1'.
  * Updates m_GraphicNodes_, m_GraphicNodesBounds_ from the values in
  * l_SourcesForCenter_, l_AccessTargetSources_, l_NoSourceTargets_.
  * When a BaseObject is drawn twice in the graphic view, the second node
  * is put on the bottom line of the access targets, and its key in m_GraphicNodes_
  * and m_GraphicNodesBounds_ is its name extended by '<!!§§!!%%!!%%!!§§>'.
  * Overrides the GraphicEPRView method.
  * */
  protected final void setGraphicNodes() {

                        // for testing and tracking
                    //System.out.println("GraphicSketchView.setGraphicNodes() start for center: "+
                      //      center_.getName());

    if (center_ == null) return;  
      
    //variables
    GraphicNode       gep;
    GraphicNode       gact;
    GraphicNode       gres;
    ImmutableSource   sour;
    ImmutableTarget   targ;
    int               index = 0;

    // set a fixed size, or set nodesNumber_ to -1
    setViewSize(nodesNumber_);

    // set the nodes positions, if it is possible
    Point[]           l_positions = setGraphicNodesPositions();
    if (l_positions.length == 0) {
        nodesNumber_ = -1; //  used by resetGraphicView() for avoiding some operations
    }
    
    if (nodesNumber_ == -1) {
        return;
    }    
    
                        // for testing and tracking    
                    //System.out.println("In GraphicSketchView.setGraphicNodes() for center: "+
                      //      center_.getName());
                    //System.out.println("nodesNumber_: "+nodesNumber_);
                    //System.out.println("l_positions:");
                    //for (int i = 0; i < l_positions.length; i++ ) {
                      //System.out.println("--"+i+": "+ l_positions[i].toString() );
                    //}
                    //System.out.println("l_SourcesForCenter_:");
                    //for (Iterator itetest = l_SourcesForCenter_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_AccessTargetSources_:");
                    //for (Iterator itetest = l_AccessTargetSources_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_NoSourceTargets_:");
                    //for (Iterator itetest = l_NoSourceTargets_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}

    for (Iterator ite = l_SourcesForCenter_.iterator() ; ite.hasNext(); ) {

        sour = (ImmutableSource) ite.next();
        if (sour instanceof Actor)
            gep = new GraphicActor(this, (Actor)sour );
        else if (sour instanceof GroupID)
            gep = new GraphicGroup(this, (GroupID)sour );
        else if (sour instanceof UserID)
            gep = new GraphicAccount(this,(UserID) sour );
        else
            gep = new GraphicVirtualFolder(this, (VirtualFolder)sour);
        // add to this JPanel
        add(gep );
        // put it at the right place
        gep.setBounds(
          0,
          0,
          (int)gep.getShape().getBounds2D().getWidth(),
          (int)gep.getShape().getBounds2D().getHeight());
        gep.setLocation(l_positions[index]);
        
        // add nodeBounds to its map, with the name,
        // even if the same ep is in l_AccessTargetSources_
        m_GraphicNodesBounds_.put((Name)sour.getName(), gep.getBounds());
        // add the node to its map
        m_GraphicNodes_.put((Name)sour.getName(), gep);
        
        gep.setVisible(true);
        index++;

    } // for (Iterator ite

    // handle center_ and put it at a given location
    if (center_ instanceof Actor)
        gep = new GraphicActor(this, (Actor)center_ );
    else if (center_ instanceof GroupID)
        gep = new GraphicGroup(this, (EligibleParty)center_ );
    else if (center_ instanceof EligibleParty)
        gep = new GraphicAccount(this, (EligibleParty)center_ );
    else if (center_ instanceof Resource)
        gep = new GraphicResource(this, (Resource)center_);
    else if (center_ instanceof VirtualFolder) 
        gep = new GraphicVirtualFolder(this, (VirtualFolder)center_);
    else gep = null;
    // add to this JPanel
    add(gep );
    // put it at the right place
    gep.setBounds(
      0,
      0,
      (int)gep.getShape().getBounds2D().getWidth(),
      (int)gep.getShape().getBounds2D().getHeight());
    // put center_ at the second line, whatever the view size
    gep.setLocation(l_positions[ (l_positions.length - 1)/2 ]);
    index = (l_positions.length - 1)/2 + 1; // for the following nodes
    // add nodeBounds to its map
    m_GraphicNodesBounds_.put(center_.getName(), gep.getBounds());
    // add the node to its map
    m_GraphicNodes_.put(center_.getName(), gep);
    gep.setVisible(true);    
    
    for (Iterator ite = l_AccessTargetSources_.iterator() ; ite.hasNext(); ) {

        sour = (ImmutableSource) ite.next();
        if (sour instanceof Actor)
            gact = new GraphicActor(this, (Actor)sour );
        else if (sour instanceof GroupID)
            gact = new GraphicGroup(this, (GroupID)sour );
        else if (sour instanceof VirtualFolder)
            gact = new GraphicVirtualFolder(this, (VirtualFolder)sour);
        else 
            gact = new GraphicAccount(this, (UserID)sour );
        // add to this JPanel
        add(gact );

        // put it at the right place
        gact.setBounds(
          0,
          0,
          (int)gact.getShape().getBounds2D().getWidth(),
          (int)gact.getShape().getBounds2D().getHeight());
        gact.setLocation(l_positions[index]);

        // add nodeBounds and nodes to their map, with '<!!§§!!%%!!%%!!§§>' if the same ep 
        // is already in l_SourcesForCenter_
        if (l_SourcesForCenter_.contains(sour) ) {
            m_GraphicNodesBounds_.put(
                    ((Name)sour.getName()).extended("<!!§§!!%%!!%%!!§§>"), gact.getBounds());
            m_GraphicNodes_.put(
                    ((Name)sour.getName()).extended("<!!§§!!%%!!%%!!§§>"), gact);
            
                        // for testing and tracking    
                    //System.out.println("In GraphicSketchView.setGraphicNodes() for center: "+
                      //      center_.getName()+", add the target node "+
                      //      ((Name)ep.getName()).extended("<!!§§!!%%!!%%!!§§>"));            
        } else {
            m_GraphicNodesBounds_.put(
                    (Name)sour.getName(), gact.getBounds());
            m_GraphicNodes_.put(
                    (Name)sour.getName(), gact);
        } // else
        
        gact.setVisible(true);
        index++;

    } // for (Iterator ite

    for (Iterator ite = l_NoSourceTargets_.iterator() ; ite.hasNext(); ) {

        targ = (ImmutableTarget) ite.next();
        //if (targ instanceof Resource)
        gres = new GraphicResource(this, (Resource)targ);
        // add to this JPanel
        add(gres );

        // put it at the right place
        gres.setBounds(
          0,
          0,
          (int)gres.getShape().getBounds2D().getWidth(),
          (int)gres.getShape().getBounds2D().getHeight());
        gres.setLocation(l_positions[index]);

        // add nodeBounds and nodes to their map, with '<!!§§!!%%!!%%!!§§>' if the same res 
        // is already in l_SourcesForCenter_
        if (l_SourcesForCenter_.contains(targ) ) {
            m_GraphicNodesBounds_.put(
                    ((Name)targ.getName()).extended("<!!§§!!%%!!%%!!§§>"), gres.getBounds());
            m_GraphicNodes_.put(
                    ((Name)targ.getName()).extended("<!!§§!!%%!!%%!!§§>"), gres);
            
                        // for testing and tracking    
                    //System.out.println("In GraphicSketchView.setGraphicNodes() for center: "+
                      //      center_.getName()+", add the target node "+
                      //      ((Name)res.getName()).extended("<!!§§!!%%!!%%!!§§>"));            
        } else {
            m_GraphicNodesBounds_.put(
                    (Name)targ.getName(), gres.getBounds());
            m_GraphicNodes_.put(
                    (Name)targ.getName(), gres);
        } // else
        
        gres.setVisible(true);
        index++;

    } // for (Iterator ite

    gep = null;
    gact = null;
    gres = null;
    sour = null;
    targ = null;
    l_positions = null;
  } // setGraphicNodes()    
  
  /**
  * Sets the allowed positions of the node in the view.
  * There are 3 sets of node positions: one if nodesNumber_
  * is inferior to 8, one between 8 and 11,
  * and one for the values from 12 to 21.
  * Each position is at the top left corner of the node.
  * This method must be called after setViewSize().
  * Called by setGraphicNodes().
  * Does not override the GraphicView method since the parameters are not used there.
  * @return the array of node positions, in the correct order;
  *   equal to Point[0] if nodesNumber_ is superior to 21, or more than 10 sources
  *   or more than 10 targets.
  * */
  protected final Point[] setGraphicNodesPositions() {
      
    if (nodesNumber_ > 21 || sourcesForCenterCount_ > 10 ||
            (accessTargetSourcesCount_ + noSourceTargetsCount_) > 10) {
            
                        // for testing and tracking    
                    //System.out.println("In GraphicSketchView.setGraphicNodesPositions() for center: "+
                      //      center_.getName()+", returns an empty array ");
        
        
        return new Point[0];
    } // if

    Point[] l_result;
    
    if (nodesNumber_ < 5 || 
           (nodesNumber_ < 8 && sourcesForCenterCount_ < 4 &&
                (accessTargetSourcesCount_ + noSourceTargetsCount_) < 4) ) {

      l_result = new Point[7];
      for (int i = 0; i < 7; i++) { // 7 positions of the grid
        l_result[i] = new Point(
          (new Double(getInsets().left +
            GraphicNode.L_SEVEN_NODES_SKETCH_GRID[i].getX())).intValue(),
          (new Double(getInsets().top +
            GraphicNode.L_SEVEN_NODES_SKETCH_GRID[i].getY())).intValue());
      } // for i          
      
    } // if (nodesNumber_ < 5) 
    else if (nodesNumber_ < 7 ||
           (nodesNumber_ < 12 && sourcesForCenterCount_ < 6 &&
                (accessTargetSourcesCount_ + noSourceTargetsCount_) < 6) ) {

      l_result = new Point[11];
      for (int i = 0; i < 11; i++) { // 11 positions of the grid
        l_result[i] = new Point(
          (new Double(getInsets().left +
            GraphicNode.L_ELEVEN_NODES_SKETCH_GRID[i].getX())).intValue(),
          (new Double(getInsets().top +
            GraphicNode.L_ELEVEN_NODES_SKETCH_GRID[i].getY())).intValue());
      } // for i          
      
    } // if (nodesNumber_ < 7) 
    else {

      l_result = new Point[21];
      for (int i = 0; i < 21; i++) { // 21 positions of the grid
        l_result[i] = new Point(
          (new Double(getInsets().left +
            GraphicNode.L_TWENTYONE_NODES_SKETCH_GRID[i].getX())).intValue(),
          (new Double(getInsets().top +
            GraphicNode.L_TWENTYONE_NODES_SKETCH_GRID[i].getY())).intValue());
      } // for i  
      
    } // else

    return l_result;
  } // setGraphicNodesPositions()  

  
} // GraphicSketchView



