//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.border.*;

import java.util.Map;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import java.lang.reflect.InvocationTargetException;

import ARoad0.Gui1.*;
import ARoad0.ErrW.*;
import ARoad0.ErrB.BaseError;
import ARoad0.gWork.*;

import ARoad0.Pattern.MVC_View;

import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.gBaseInterface.BaseObject;
import ARoad0.gBaseInterface.ViewInBase;
import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.ACSObject;
import ARoad0.gBaseInterface.DetailledName;
import ARoad0.gBase.NameImpl;

/**
 * This class is responsible for displaying a view and its graphical
 * objects (nodes, links, arrows, rights) in the desktop. <P>
 * The main work is done at the initialization of the instance, so that
 * painting actions are very fast.
 * The constructor calls initialize() which builds up the map of node pairs, 
 * each linked by none, one or many displayable links, 
 * in calling gWork.RightsFactory_Facade.detectXXXRights().
 * Then, initialize() sets the JPanel size, its GraphicNodes,
 * sets its columns and defines the GraphicNode locations
 * to obtain a clearer graphic representation.
 * The constructor then calls buildUpArrowsAndRights() which sets the links
 * and the visible rights in the view, as shapes or as strings
 * which will be drawn in paintComponents() when the GraphicView is displayed.
 * <P>
 * The user may provide a drag&drop operation for moving a node in the view, and/or
 * to permute two nodes. The user may also click on a node to select
 * its BaseObject in the explorer. When he adds a new node in the view,
 * it is put on the top left coin of the graphic view.
 * Any updating is done by resetGraphicView and resetRights methods.
 * When the nodes set changes, the method resetGraphicView() is called.
 * When a right changes, the method resetRights() is called.
 * No layout manager is used, because they do not allow to get the
 * absolute position of a GraphicNode, that is necessary to draw the links
 * between the nodes.<P>
 * This class is directly used only for the 'Architecture' view. Its subclasses
 * are GraphicEPRView and GraphicNoThanView, and they provide most of the work.
 * However, this class is a strong support for all the subclasses methods, and
 * should be well understood. <P>
 * This class is designed to be updated, as any JComponent, by the event dispatch thread
 * for most of the operations, but it allows the rights updating in a worker thread, through
 * the call of resetRights(). For a presentation of the view  threads, see
 * GraphicViewBaseListernerImpl. A main design principle is that each view is
 * independant, so it can be processed in any thread. On the other hand, it is
 * not possible to reuse the access paths of a view in another view.
 * @see ARoad0.gBase.ViewInBaseImpl
 * @see ARoad0.Gui2.GraphicViewBaseListenerImpl
 */
public class GraphicView extends JPanel implements MVC_View, MouseListener,
  MouseMotionListener {

  // variables
  protected static final int              PRIME = 1000003;
    /**
     * initial capacity used for List and Map initialization
     * the default value for Map is 101
     */
  protected static final int              INITIAL_CAPACITY = 15;
    /**
     * shift along the y axis to display a second line for rights. No use.
     */
  protected static final int              RIGHTS_SHIFT = 6;
    /**
     * border of the graphic view panel
     */
  protected static final int              BORDER = 10;
    /**
     * length of the arrow head
     */
  protected static final float            ARROW_HEAD_LENGTH = 10;
    /**
     * angle in the arrow head
     */
  protected static final float            ARROW_HEAD_ANGLE = (float)Math.PI / 8;

    /**
     * dashed stroke for arrows or lines
     */
  protected static final BasicStroke      DASHED_STROKE = new BasicStroke(1,
                                            BasicStroke.CAP_BUTT,
                                            BasicStroke.JOIN_BEVEL, 1.0f,
                                            new float[] {6,6}, 0);
    /**
     * double-width stroke for arrows
     */
  protected static final BasicStroke      DOUBLEWIDTH_STROKE = new BasicStroke(2);
    /**
     * dashed stroke for arrows or lines
     */
  protected static final BasicStroke      DASHED_DOUBLEWIDTH_STROKE = new BasicStroke(2,
                                            BasicStroke.CAP_BUTT,
                                            BasicStroke.JOIN_BEVEL, 1.0f,
                                            new float[] {6,6}, 0);
  /**
     * stroke generally used for drawing and filling the shapes
     */
  protected static final BasicStroke      BASIC_STROKE = new BasicStroke();
    /**
     * null shape array used to test the return value from buildUpArrows()
     */
  protected static final Object[]         NULL_SHAPE_ARRAY = new Object[0];
    /**
     * distance to select the arrow construction algorithm in setPoint2DForArrowAndRights()
     */
  protected static final int              NODE_DISTANCE = 30;
   /*
    * The text displayed to tell the user the processing is going on
    */
  protected static final String           WAITING_TEXT = "<Please wait...>";

    /**
     * map of GraphicNodes, the associated BaseObject name is the key, sometimes
     * extended by <DOWN> in GraphiSketchView, or the key is WAITING_TEXT in a name
     * */
  protected final Map<ImmutableName, GraphicNode>   m_GraphicNodes_ = new HashMap(INITIAL_CAPACITY);
    /**
    * map of the GraphicNode absolute cordinates in this panel Graphics2D,
    * the associated BaseObject name is the key,
    * Rectangle2D is the value type; get through GraphicNode.getShape().
    */
  protected final Map<ImmutableName, Rectangle2D>   m_GraphicNodesBounds_ =
                                            new HashMap(INITIAL_CAPACITY);
  /**
   * map of DisplayableLinkImpls; the compound name constructed with
   * the two base objects names (eligible party first) is the key;
   * the value is a List of DisplayableLinkImpls for getting all the
   * visible or hiden links between two view base objects.
   * All links in a list have the SAME 2 ENDS.
   * */
  protected Map<ImmutableName, List<DisplayableLinkImpl>>  m_l_DisplayableLinks_ =
                                              new HashMap(INITIAL_CAPACITY);
    /**
     * array of objects ready to be drawn, as Shapes and associated Strings.
     * Updated directly by some subclasses.
     */
  protected Object[]                       l_LinksAndRights_;
    /**
     * image of l_LinksAndRights_, but with a Point2D at the rights place,
     * to give to drawArrowsAndRights() the position of the rights to draw
     */
  protected Object[]                       l_currentArrowsAndRightsStarts_;
    /**
     * ViewInBase associated to this GraphicView
     */
  protected ViewInBase                     viewInBase_;
    /**
     *  JTextPane displayed through the button "See why" in the view frame
     */
  protected JTextPane                     whyText_ = new JTextPane();
    /**
     * Set of all the implied BaseObjects that are nodes, displayable or not,
     * in the GraphicView DisplayableLinkImpls
     */
  protected Set<BaseObject>               l_allNodesInLinks_ =
                                            new HashSet(INITIAL_CAPACITY);
    /**
     * Set of all the BaseObjects implied in gWork to define the rights.
     * Updated by gWork classes.
     */
  protected HashSet<BaseObject>           l_rightsRelatedBaseObjects_ =
                                             new HashSet(INITIAL_CAPACITY*10);

  /**
   * number of nodes in the view.
   */
  protected int                           nodesNumber_ = 0;

  /**
   * in a drag&drop operation, the name of the source node to move.
   */
  protected ImmutableName                 dragSourceNodeName_;

  /**
   * in a drag&drop operation to move a node, the target node name, if
   * the target is not empty.
   */
  protected ImmutableName                 dragTargetNodeName_;

  /**
   * true when there is a drag&drop operation.
   */
  protected boolean                       mouseDragged_ = false;

  //
  // ------------- constructors ------------------------------
  //
  /**
   * This constructor is only for the subclasses.
   */
  public GraphicView() {}  
  
  /**
  * This constructor add all the _view objects in this panel,
  * each associated to a GraphicNode.
  * Build up the map of (EP, Resource) pair
  * linked by displayable links, then set the JPanel size, its GraphicNodes
  * and its columns, construct the arrows and the visible rights shapes
  * to be drawn in paintComponents().
  * The main work is done there by buildUpArrowsAndRights(),
  * so that the paint action is fast. But
  * an important condition is to have a stable set of BaseObjects in the view,
  * without updating. Any BaseObject updating is done by GraphicViewBaseListenerImpl,
  * which finalizes the previous GraphicView instance and then, creates a new one.
  * Mantain an External Object Reference to _view.
  * This constructor is not called by the constructors of the subclasses.
  * @param _view EPRViewInBase to associate to this instance.
  * @param _m_positionIndex is the map (name, Integer) of the map returned
  *   by getIndexesOfNodeLocations() at the previous view opening. May be null.
  * @exception ProcessError if there is a loop in the access paths searching
  **/
  public GraphicView(ViewInBase _view, Map _m_positionIndex) throws ProcessError {
    viewInBase_ = _view;

    // no Layout, because they are not able to return
    // the absolute coordinates of a component.
    // other way: use Container.getComponentAt(Point), but more complex
    this.setLayout(null);

    this.setBorder(new EmptyBorder(BORDER*2,BORDER,BORDER*2,BORDER));

    // set the GraphicNodes in the JPanel
    try {
        initialize();
    } catch (InterruptedException i) {}
    
    // reset the node positions
    if (_m_positionIndex != null) setIndexesOfNodeLocations(_m_positionIndex);
    
    // construct the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    l_LinksAndRights_ = buildUpArrowsAndRights(m_l_DisplayableLinks_, m_GraphicNodesBounds_);

    // set the color
    setBackground(Color.white);

    addMouseListener(this);
    addMouseMotionListener(this);
  } // GraphicView(EPRViewInBase _view)

  /**
  * Initializes nodesNumber_, m_l_DisplayableLinks_, l_allNodesInLinks_, and set
  * the GraphicNodes locations. Display an user message if the view size is superior to 46.
  * Called by the constructor and resetGraphicView().
  * Calls RightsFactory_Facade.detectRights().
  * Synchronized on viewInBase_. 
  * Overridden method.
  * @exception InterruptedException, ProcessError if there is a loop in the rights search
  * */
  protected void initialize() throws ProcessError, InterruptedException {

                    //System.out.println("In GraphicView.initialize(), the non-ep resources are:");
                    //for (Iterator ite = l_nonEpResources_.iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
    // variables
    List                  l_link;
    BaseObject[]          l_node;
    DisplayableLinkImpl   displayableLink;
    
    l_rightsRelatedBaseObjects_.clear();

    synchronized (viewInBase_) {

      nodesNumber_ = viewInBase_.getEorM_BaseObjects().size();

                    //System.out.println("In GraphicView.initialize(), the ep are:");
                    //for (Iterator ite = viewInBase_.getEorM_EligibleParties().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}

      // build up the map of (ACS, ACS) pair linked by displayable links
      m_l_DisplayableLinks_ = RightsFactory_Facade.
        detectRights(this, viewInBase_);

                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("in GraphicView.initialize(), "+ 
                      //"the DisplayableLinkImpls are:");
                    //for (Iterator ite = m_l_DisplayableLinks_.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)m_l_DisplayableLinks_.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite 
                    //key = null;
      
      // update l_allNodesInLinks_.
      // m_l_DisplayableLinks_ may be empty.
      l_allNodesInLinks_.clear();
      for (Iterator ita = m_l_DisplayableLinks_.values().iterator() ;
        ita.hasNext() ; ) {
          l_link = (List)ita.next();

                  // for testing and tracking
                  // System.out.println("l_link in initialize(): "+l_link);

          for (Iterator ite = l_link.iterator() ;ite.hasNext() ; ) {
            displayableLink = (DisplayableLinkImpl)ite.next();
            l_node = displayableLink.getEorAllNodes();
            for (int i = 0; i < l_node.length; i++) {
              l_allNodesInLinks_.add(l_node[i] );
            } // for i
          } // for ite
      } // for (Iterator ita = m_l_DisplayableLinks_

                    // for testing and tracking
                    //System.out.println("l_accessTargetEP_: "+l_accessTargetEP_);

      if (nodesNumber_ < 47) setGraphicNodes();
      else {
        // display a message
        DialogBox.showOkMessage(
          "This view has more than 46 members. It is not displayed in a diagram.");
      }
      //else setGraphicNodesInThreeColumns(); //bug when a resource number 13 is added
    } // synchro

    l_link = null;
    l_node = null;
    displayableLink = null;    
  } // initialize()

  /**
  * Provides the image to be exported, as a PNG image for instance. Not used
  * by ActionExportPNG, since it does not provide the true colors.
  * @return the image
  * */
  public final BufferedImage getExportImage() {
    GraphicsEnvironment local = GraphicsEnvironment.getLocalGraphicsEnvironment();
    GraphicsDevice screen = local.getDefaultScreenDevice();
    GraphicsConfiguration configuration = screen.getDefaultConfiguration();
    BufferedImage result = configuration.createCompatibleImage(getWidth(), getHeight());

    Graphics2D graphics = result.createGraphics();

    graphics.setBackground(Color.white);

    // draw the nodes
    GraphicNode node;
    ImmutableName name;
    Rectangle2D rect2D;
    for (Iterator ite = m_GraphicNodes_.keySet().iterator(); ite.hasNext(); ) {
      name = (ImmutableName)ite.next();

      if (name.equals(new NameImpl(WAITING_TEXT))) continue;

      node = (GraphicNode)m_GraphicNodes_.get(name);
      rect2D = (Rectangle2D)m_GraphicNodesBounds_.get(name);
      graphics.translate( rect2D.getBounds().getX(), rect2D.getBounds().getY() );
      node.paintComponentForExport(graphics);
      graphics.translate(- rect2D.getBounds().getX(), - rect2D.getBounds().getY() );
    }

    // draw the arrows associated to visible rights
    drawArrowsAndRights(graphics);

    graphics.drawString("Generated by the Access Road software,"+
      " v0.7, http://accessroad.sourceforge.net/home.html", 5, getHeight() - 20);

    graphics = null;
    node = null;
    name = null;
    rect2D = null;
    return result;
  } // getExportImage()


  /**
  * Draws the new arrows and rights after a change in the rights,
  * with no change in GraphicNodes.
  * Called in a worker thread by GraphicViewBaseListener.updateFrameViewRights()
  * and GraphicViewPropertyEditor.propertyChange().
  * Calls RightsFactory_Facade.detectRights(), buildUpArrowsAndRights().
  * This method call must be preceded by addWaitingMessage() and
  * must be followed by a call to removeWaitingMesage() and repaint(),
  * both in the event dispatch thread.
  * Overridden by GraphicEPRView and GraphicNoThanView.
  * @exception ProcessError if there is a loop in the rights search 
  * */
  public synchronized void resetRights() throws ProcessError, InterruptedException {

    // clear m_l_DisplayableLinks_ to cut any link to the current dLinks
    for (Iterator ite = m_l_DisplayableLinks_.values().iterator() ; ite.hasNext() ; ) {
        for (Iterator ite2 = ((List) ite.next()).iterator() ; ite2.hasNext() ; ) {
          ((DisplayableLinkImpl)ite2.next()).finalizeForGui();
        }
    } // for ite
    m_l_DisplayableLinks_.clear();
    l_rightsRelatedBaseObjects_.clear();
      
    // build up the map of (ACS, ACS) pair linked by displayable links
    m_l_DisplayableLinks_ = RightsFactory_Facade.detectRights(this, viewInBase_);

    // construct the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    // throws InterruptedException
    Object[] l_LinksAndRights = buildUpArrowsAndRights(
            m_l_DisplayableLinks_, m_GraphicNodesBounds_);

    // if the event dispatch thread tries to repaint this instance, it waits
    // on drawArrowsAndRights() which uses l_LinksAndRights_
    synchronized (l_LinksAndRights_) {
        l_LinksAndRights_ = l_LinksAndRights;
    }


                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("in GraphicView.resetRights(), "+
                      //"the DisplayableLinkImpls before repaint() are:");
                    //for (Iterator ite = m_l_DisplayableLinks_.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)m_l_DisplayableLinks_.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
    
    l_LinksAndRights = null;
  } // resetRights()  
  
  /**
  * This central method resets the view after a view nodes changing.
  * It finalizes the previous graphic nodes which are no more there.
  * Called in a worker thread from GraphicViewBaseListener.changeFrameView() and
  * GraphicViewPropertyEditor.propertyChange().
  * Calls addWaitingMessage() and repaint() in the Event Dispatch Thread, before
  * the reinitialization of the GraphicNodes.
  * Calls finalizeForGui(), initialize(),
  * addNewGraphicNodeLocation(), buildUpArrowsAndRights().
  * There is no removeGraphicNodeLocation() method.
  * This method is used directly by GraphicEPRView and GraphicNoThanView when
  * a node has been added or removed. It is overridden by GraphicSketchView.
  * @exception ProcessError if there is a loop in the rights search
  * */
  public synchronized void resetGraphicView() throws ProcessError, InterruptedException {
    ViewInBase      view = viewInBase_;
    boolean         newNode = false;
    Map             m_GraphicNodesPosition = new HashMap(INITIAL_CAPACITY);
    ImmutableName   newNodeName = null;
    ImmutableName   name = null;
    
                                // for testing and tracking
                            //System.out.println("GraphicView.resetGraphicView() starting for "+
                              //      view.getName());
    
    // stores the old node locations to restore them on the new view
    for (Iterator ite = m_GraphicNodes_.entrySet().iterator(); ite.hasNext(); ) {
        Map.Entry entry = (Map.Entry)ite.next();
        if ( !((ImmutableName)entry.getKey()).equals(new NameImpl(WAITING_TEXT)) )
            m_GraphicNodesPosition.put(
                entry.getKey(),
                ((GraphicNode)entry.getValue()).getLocation() );
    } // for ite
    
    if (view.getEorM_BaseObjects().keySet().containsAll(m_GraphicNodes_.keySet()))
        newNode = true;
    
    if (newNode) {
        // searching for the new node - the method runs even if it is not found
        for (Iterator ite = view.getEorM_BaseObjects().keySet().iterator(); ite.hasNext(); ) {
            name = (ImmutableName)ite.next();
            if (!m_GraphicNodes_.keySet().contains(name)) {
                newNodeName = name;
                break;
            }
        } // for
        // the target in a NoThanViewInBase is not in getEorM_BaseObjects()
        // and then newNodeName is not used:
        // initialize() is overridden to do all the work, and
        // setGraphicNodeLocations() is overridden to become empty
    } // if
    
    finalizeForGui();
    // these are the differences between the finalized state and the constructor state
    whyText_ = new JTextPane(); 
    viewInBase_ = view;

    // no InterruptedException handling there
    try {
        SwingUtilities.invokeAndWait(new Runnable() {
            public void run(){

                                // for testing and tracking
                            //System.out.println("in GraphicView.resetGraphicView(), "+
                              //      "for the view "+viewInBase_.getNickName()+
                                //    "call to addWaitingMessage() in the EVT");

                    // in the Event Dispatch Thread
               addWaitingMessage();

                                // for testing and tracking
                            //System.out.println("in GraphicView.resetGraphicView(), "+
                              //      "for the view "+viewInBase_.getNickName()+
                                //    "call to repaint() in the EVT");

               repaint();


                                // for testing and tracking
                            //System.out.println("in GraphicView.resetGraphicView(), "+
                              //      "for the view "+viewInBase_.getNickName()+
                                //    "after addWaitingMessage() and repaint() in the EVT");
            } // run
        }); // invokeAndWait()
    }
    catch (InvocationTargetException i) {
        throw new InternalError("From GraphicView.resetGraphicView() at the adding of the "+
                "waiting message: "+i.toString());
    }
    
      // builds up the GraphicNodes and the map of pairs
    // overriden method
    initialize();

    // sets the new node location, if non-null node,
    // and restores the previous node locations
    setGraphicNodeLocations(newNodeName, m_GraphicNodesPosition);

    // constructs the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    // overriden method- throws InterruptedException
    Object[] l_LinksAndRights = buildUpArrowsAndRights(
            m_l_DisplayableLinks_, m_GraphicNodesBounds_);
    
                        // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println("in GraphicView.resetGraphicView(), "+ 
                      //"the DisplayableLinkImpls before repaint() are:");
                    //for (Iterator ite = m_l_DisplayableLinks_.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)m_l_DisplayableLinks_.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite                        

    if (l_LinksAndRights_ != null) {
        synchronized (l_LinksAndRights_) {
            l_LinksAndRights_ = l_LinksAndRights;
        }
    }
    else
        l_LinksAndRights_ = l_LinksAndRights;

                                // for testing and tracking
                            //System.out.println("in GraphicView.resetGraphicView(), "+
                              //      "for the view "+view.getNickName()+
                                //    "end in a worker thread");

    m_GraphicNodesPosition.clear();
    m_GraphicNodesPosition = null;
    name = null;
    newNodeName = null;
    view = null;
    l_LinksAndRights = null;
  } // resetGraphicView()   
  
  /**
  * Through a click on a view object, select the object in the explorer and the beamer.
  * Use m_GraphicNodesBounds_ to find the right object on which the user clicks on.
  * In GraphicNoThanView, it works also for the two GraphicTexts for right criteria,
  * where the keys in m_GraphicNodesBounds_ are the two values
  * GraphicNoThanView.NO_MORE or LESS_RIGHT_GRAPHICTEXT_BOUNDS_KEY.
  * Calls getSelectedObjectOnMouseClicked().
  * @param _e is the event
  * */
  public final void mouseClicked(MouseEvent _e) {
    // variables
    DefaultMutableTreeNode  node = null;
    BaseObject              selectedObject = getSelectedObjectOnMouseClicked(_e);
    TreePath                path;

                      // for testing and tracking
                      //System.out.println("GraphicView.mouseClicked() start");

    if (selectedObject == null) {
        return;
    } // if


    // change the selection in the explorer
    if (selectedObject instanceof DetailledName && selectedObject instanceof ACSObject) {
    node = ACSTreeUtilities.getDetailledNameNodeInExplorer(
      (DetailledName) selectedObject, ((ACSObject)selectedObject).getAcsName());
    } // if
    else if (selectedObject instanceof BaseObject) {
    node = ACSTreeUtilities.getCurrentObjectNode(
      (DefaultMutableTreeNode) InternalFrameModel.getTreeInExplorer().getModel().getRoot(),
      selectedObject);
    } // if
    if (node == null) {
        // for testing and tracking
        //System.out.println("in GraphicView.mouseClicked() "+
          //"with null node to display ");
        return;
    }
    path = new TreePath(node.getPath());
    InternalFrameModel.getTreeInExplorer().setSelectionPath(path);

      // for testing and tracking
      //System.out.println("in the explorer, the selected object is"+
      //  selectedObject.toString()+" and the new selected path is: "+path.toString());

    selectedObject = null;
    node = null;
    path = null; 
  } // mouseClicked()
  
  /**
  * Called by mouseClicked() to find the BaseObject in the view on which the user
  * has clicked. Overidden method by GraphicSketchView.
  * @since 0.7.0
  * @param _e is the event
  * @return the selected BaseObject or null
  * */
  protected BaseObject getSelectedObjectOnMouseClicked(MouseEvent _e) {
    BaseObject              selectedObject = null;
    ImmutableName           name = null;
    Rectangle2D             rect;
    boolean                 found = false;

                      // for testing and tracking
                      //System.out.println("GraphicView.getSelectedObjectOnMouseClicked() start"); 
    
    // test the current nodes as target    
    for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
      name = (ImmutableName) ite.next();
      rect = (Rectangle2D)m_GraphicNodesBounds_.get(name);
      //if ( _e.getPoint().distance(rect.getX(), rect.getY()) < GraphicNode.NODE_SIZE) {
      if ( rect.contains(_e.getPoint()) ) {
          found = true;  
          break;
      }
    } // for ite

    if (!found) {
        name = null;
        rect = null;

                      // for testing and tracking
                      //System.out.println("in GraphicView.getSelectedObjectOnMouseClicked(), "+
                        //" no selection with the nodes:");
                      //for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
                        //  System.out.println((ImmutableName) ite.next());
                      //}
        
        return null;
    }


                      // for testing and tracking
                      //System.out.println("in GraphicView.getSelectedObjectOnMouseClicked(), "+
                        //" selection of the node "+name+" among the nodes"); 
                      //for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
                        //  System.out.println((ImmutableName) ite.next());
                      //}
    
    try {
        selectedObject = ACS_Facade.getEorBaseObject(name);
    } catch (BaseError b) {

                      // for testing and tracking

                      // for testing and tracking
                      System.out.println("in GraphicView.getSelectedObjectOnMouseClicked(), "+
                        " the selection is not a known BaseObject. The name is: "+name);
        
        // display a message
        DialogBox.showErrorMessage(
          "Sorry, the object "+name+" is not found in the base.");

        selectedObject = null;
        name = null;
        rect = null;
        return null;        
    } // catch
    
    name = null;
    rect = null;    
    return selectedObject;
  } // getSelectedObjectOnMouseClicked()

  /**
  * No action.
  * @param _e is the event
  * */
  public final void mouseEntered(MouseEvent _e) {

  } // mouseMoved

  /**
  * No action.
  * @param _e is the event
  * */
  public final void mouseExited(MouseEvent _e) {

  } // mouseMoved

  /**
  * No action.
  * @param _e is the event
  * */
  public final void mousePressed(MouseEvent _e) {

                    //System.out.println("GraphicView.mousePressed() "+_e);
                    //for (Iterator ite = viewInBase_.getEorM_EligibleParties().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
    return;
  } // mouseMoved

  /**
  * No action.
  * @param _e is the event
  * */
  public final void mouseMoved(MouseEvent _e) {

  } // mouseMoved  

  /** 
  * If the mouse has been dragged from a view node position to another view node position,
  * the two nodes are permutted when the mouse is released.
  * If the mouse is dragged and released to an empty node position,
  * the first node is moved to this empty position if it is not the index 0
  * in the positions array. Noop on the WAITING_TEXT node.<P>
  * Caution: dragSourceNodeName_ is set by mouseDragged() that searches among the node
  * positions. The user has to not be too fast, if there are a big number of nodes. <P>
  * Overridden method in GraphicSketchView and GraphicNoThanView.
  * @param _e is the event
  * */
  public void mouseReleased(MouseEvent _e) {

                    //System.out.println("GraphicView.mouseReleased(): "
                      //+_e);

    if (!mouseDragged_) {
      return;
    }

    // test the current nodes as target
    ImmutableName name;
    Rectangle2D rect;
    boolean drag_drop = false;
    for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
      name = (ImmutableName) ite.next();
      rect = (Rectangle2D)m_GraphicNodesBounds_.get(name);
      if ( _e.getPoint().distance(rect.getX(), rect.getY()) < GraphicNode.NODE_SIZE) {
            dragTargetNodeName_ = name;
            drag_drop = true;
            break;
      }
    } // for ite

    // test the empty locations as target
    Point targetLocation = null;
    if (!drag_drop) {
      Point[] l_points = setGraphicNodesPositions(nodesNumber_, false);
      // the index 0 is never a target, to let it for a new node
      for (int i = 1; i < l_points.length; i++) {
        if ( l_points[i].distance(_e.getPoint()) < GraphicNode.NODE_SIZE ) {
          targetLocation = l_points[i];
          drag_drop = true;
          break;
        }
      } // for i
      l_points = null;
    } // if (!mouseDragged_)

    if (!drag_drop) {
      mouseDragged_ = false;
      return;
    }

                            // for testing and tracking
                            //System.out.println("GraphicView.mouseReleased(), nodes in m_GraphicNodes_: ");
                            //for (ImmutableName na : m_GraphicNodes_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na
                            //System.out.println("dragSourceNodeName_: "+dragSourceNodeName_);
                            //System.out.println("dragTargetNodeName_: "+dragTargetNodeName_);
                            //System.out.println("nodes in m_GraphicNodesBounds_: ");
                            //for (ImmutableName na : m_GraphicNodesBounds_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na


    Point sourceLocation = null;
    if (dragTargetNodeName_ != null) {
      sourceLocation = ((GraphicNode)m_GraphicNodes_.get(dragSourceNodeName_)).
        getLocation();
        // set the new node at the target
      ((GraphicNode)m_GraphicNodes_.get(dragSourceNodeName_)).setLocation(
        ((GraphicNode)m_GraphicNodes_.get(dragTargetNodeName_)).getLocation() );
      
        // set the new node at the source
      ((GraphicNode)m_GraphicNodes_.get(dragTargetNodeName_)).
        setLocation(sourceLocation);
      m_GraphicNodesBounds_.put(dragTargetNodeName_,
        ((GraphicNode)m_GraphicNodes_.get(dragTargetNodeName_)).getBounds() );
    } else {
        // set the new node at the target
      ((GraphicNode)m_GraphicNodes_.get(dragSourceNodeName_)).
         setLocation(targetLocation);
    } // else
    // set the new node at the target
    m_GraphicNodesBounds_.put(dragSourceNodeName_,
      ((GraphicNode)m_GraphicNodes_.get(dragSourceNodeName_)).getBounds() );

    // construct the arrow shapes, the rights, and l_currentArrowsAndRightsStarts_
    l_LinksAndRights_ = buildUpArrowsAndRights(m_l_DisplayableLinks_, m_GraphicNodesBounds_);

    mouseDragged_ = false;
    dragTargetNodeName_ = null;
    dragSourceNodeName_ = null;
    targetLocation = null;
    sourceLocation = null;
    name = null;

    repaint();
  } // mouseReleased

  /**
  * If the mouse is dragged from a view node position to another view node position,
  * the two nodes are permutted. If it is dragged to an empty node position,
  * the first node is moved to this position.
  * For large views, the search of the right node is quite slow.
  * Overridden method in GraphicSketchView. 
  * @see #mouseReleased
  * @param _e is the event
  * */
  public void mouseDragged(MouseEvent _e) {
    if (mouseDragged_) {
      /*
      int currentX = _e.getX();
      int currentY = _e.getY();
      if (previousMouseDragX_ != 0 &&
        Point2D.distance((double)currentX, (double)currentY,
          (double)previousMouseDragX_, (double)previousMouseDragY_)
          < (double)GraphicNode.NODE_SIZE) return;
      // draw a stroked rectangle to show the current drag
      if (graphics_ != null) {
        DASHED_RECTANGLE.setLocation(currentX, currentY);
        graphics_.setStroke(DASHED_STROKE);
        graphics_.draw( DASHED_RECTANGLE );
        graphics_.setStroke(BASIC_STROKE);
      }
      repaint();
      previousMouseDragX_ = currentX;
      previousMouseDragY_ = currentY;
      */
      return;
    } // if (mouseDragged_) {
    ImmutableName name = null;

                    //System.out.println("GraphicView.mouseDragged() "+_e);
                    //for (Iterator ite = viewInBase_.getEorM_EligibleParties().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ite.next().toString() );
                    //}
    // && isFocusOwner() && getVisibleRect().contains(_e.getPoint())
    if (nodesNumber_ < 48 ) {
      for (Iterator ite = m_GraphicNodesBounds_.keySet().iterator(); ite.hasNext(); ) {
        name = (ImmutableName) ite.next();
        if ( ((Rectangle2D)m_GraphicNodesBounds_.get(name)).contains(_e.getPoint()) ) {
          dragSourceNodeName_ = name;
          mouseDragged_ = true;

                    //System.out.println("GraphicView.mouseDragged(), dragSourceNodeName_: "
                      //+dragSourceNodeName_);

          break;
        }
      } // for ite
    } // if
    name = null;
  } // mouseDragged

  /**
  * Get the relative positions of nodes, for the GUI saving.
  * The position is an index in the Points array which is applicable to this
  * view, according to its size, which is returned by setGraphicNodesPositions().
  * Do not process the WAITING_TEXT node.
  * No use for the sketch views, and usuful for NoThan views.
  * @return Map of indexes in the GraphicNode Points array, where the node name is the key
  *  and the value is an Integer
  * */
  public final Map getIndexesOfNodeLocations() {
    Map             m_result = new HashMap(INITIAL_CAPACITY);
    Point           point = null;
    Map.Entry       entry = null;
    Point[]         l_points = setGraphicNodesPositions(nodesNumber_, false);    

    // get the node locations
    for (Iterator ite = m_GraphicNodes_.entrySet().iterator(); ite.hasNext(); ) {
        entry = (Map.Entry)ite.next();

        if (((NameImpl)entry.getKey()).equals(new NameImpl(WAITING_TEXT)) )
            continue;

        point = ((GraphicNode)entry.getValue()).getLocation();
        // search the location in the points array
        for (int i = 0; i < l_points.length; i++) {
            if ( l_points[i].equals(point)  ) {
              m_result.put(entry.getKey(), new Integer(i));
              break;
            }
        } // for i        
    } // for ite
    
    entry = null;
    point = null;
    l_points = null;    
    return m_result;
  } // getIndexesOfNodeLocations()

  /**
  * Sets the relative positions of nodes, after a GUI saving.
  * The argument is typically the returned value from getIndexesOfNodeLocations() at
  * the previous opening of the view.
  * As any new node changes the view size, the array of points may then be not relevant
  * for the locations which are set by _m_positionIndex. So this method does not operate
  * if the number of nodes is not the previous number at the saving time.
  * Called by the constructor. No use for the sketch views. For the NoThan views,
  * noop since the number of nodes is not the previous number at the saving time.
  * Caution: view nodes added after the saving are not processed. Rights may change.
  * @param _m_positionIndex Map of indexes in the GraphicNode Points array, 
   * where the node name is the key. Never null.
  * */
  public final void setIndexesOfNodeLocations(Map<ImmutableName, Integer> _m_positionIndex) {
    Map.Entry       entry = null;
    Point[]         l_points = setGraphicNodesPositions(nodesNumber_, false);    

    if (m_GraphicNodes_.size() != _m_positionIndex.size()) {

                        // for testing and tracking
                    //System.out.println("GraphicView.setIndexesOfNodeLocations(() for"+
                      //      viewInBase_.getName()+", "+
                        //    "in _m_positionIndex.size(): "+_m_positionIndex.size()+
                          //  " and m_GraphicNodes_.size(): "+m_GraphicNodes_.size() );
                    
        return;
    }
    
    for (Iterator ite = _m_positionIndex.entrySet().iterator(); ite.hasNext(); ) {
        entry = (Map.Entry)ite.next();

                        // for testing and tracking
                    //System.out.println("GraphicView.setIndexesOfNodeLocations((), "+
                      //      "in _m_positionIndex, key: "+entry.getKey() );
        
        // set the location from the l_points value
        if ( m_GraphicNodes_.get(entry.getKey()) != null ) {
            
            ((GraphicNode)m_GraphicNodes_.get(entry.getKey())).setLocation(
                l_points[ ((Integer)entry.getValue()).intValue() ]);
            
            m_GraphicNodesBounds_.put((ImmutableName)entry.getKey(),
                ((GraphicNode)m_GraphicNodes_.get(entry.getKey())).getBounds() );            

                        // for testing and tracking
                    //System.out.println("GraphicView.setIndexesOfNodeLocations((), "+
                      //      "in m_GraphicNodesBounds_, for "+entry.getKey()+
                        //    " put the bounds "+((GraphicNode)m_GraphicNodes_.get(entry.getKey())).getBounds() );
        
        } // if
    } // for ite
    
    entry = null;
    l_points = null;    
  } // setIndexesOfNodeLocations()

  /**
   * @return ViewInBase drawn by the GraphicView
   */
   public ViewInBase getEorDataModel() {
    return viewInBase_;
   }

  /**
   * @return Set of all the implied BaseObjects that are nodes, displayable or not,
   * in the GraphicView DisplayableLinkImpls
   */
   public Set getAllNodesInLinks() {
     return (Set)((HashSet)l_allNodesInLinks_).clone();
   }

  /**
   * Called by InternalFrame.showViewFrame() for building up the extended view.
   * @return Set of the implied BaseObjects that are hidden, not displayable
   * in the GraphicView DisplayableLinkImpls since they are not in the view
   */
   public Set getHiddenNodesInLinks() {
    Set result = (Set)((HashSet)l_allNodesInLinks_).clone();
    if (viewInBase_ == null) return result;
                  // for testing purpose
              //System.out.println("call to GraphicView.getHiddenNodesInLinks; viewInBase_: "+
                //viewInBase_);
    result.removeAll( viewInBase_.getEorM_BaseObjects().values() );
    return result;
   }

  /**
   * Gets the set of all the BaseObjects implied in gWork to define the rights.
   * These BaseObjects are provided by the gWork classes (CompoundRightsFactoryImpl).
   * They include l_allNodesInLinks_.
   * Called by ViewModel, GraphicViewBaseListenerImpl.changeView(),
   * gWork.RightsFactoryUtilities.setWhyText().
   * @return the rights-related BaseObjects or ImmutableNames. Never null.
   */
   public HashSet getRightsRelatedBaseObjects() {
     if (l_rightsRelatedBaseObjects_ == null)
         l_rightsRelatedBaseObjects_ = new HashSet(INITIAL_CAPACITY*10);
     return l_rightsRelatedBaseObjects_;
   }

  /**
   * These BaseObjects are provided by the gWork classes.
   * Called by the gWork core algorithms. Overridden method.
   * @param _l_baseObjects BaseObject set. May be null.
   */
   public void setRightsRelatedBaseObjects(HashSet _l_baseObjects) {
       if (_l_baseObjects == null)
           l_rightsRelatedBaseObjects_ = new HashSet(INITIAL_CAPACITY*10);
       else
           l_rightsRelatedBaseObjects_ = _l_baseObjects;
   }

  /**
   * @return the hashCode from the DisplayableLinks map
   */
   public int getLinksHashCode() {
     int result = PRIME*17;
     for (Iterator ite = m_l_DisplayableLinks_.values().iterator() ; ite.hasNext() ; ) {
       for (Iterator ita = ((List)ite.next()).iterator() ; ita.hasNext() ; ) {
              result = PRIME*result + ((DisplayableLinkImpl)ita.next()).hashCode();
      } // for ita
     } // for ite
     return result;
   }

  /**
   * Get the current comments that the program has generated to explain the current view rights.
   * @return the text which explains the rights in the view. Null if finalized or
   *   at the GraphicView initialization.
   */
   public synchronized String getWhyText() {
       if (whyText_ == null) return null;    
       
       return whyText_.getText();
   } // getWhyText()

  /**
   * Sets the comments that the gWork package generates to explain the view rights. No use.
   * This method is operating only in the subclasses.
   * Also used to test if this instance has been finalized.
   * Its height is used by the GraphicViewBaseListener.
   * @return the text which explains the rights in the view. Null if finalized.
   *   Empty at the GraphicView initialization.
   */
   public synchronized JTextPane setWhyText() {
       if (whyText_ == null) return null;
         
       // construct the 'See why' text
       whyText_ = RightsFactory_Facade.getWhyText(this, m_l_DisplayableLinks_, viewInBase_);   
       
       return whyText_;
   } // setWhyText()

  /**
   * Invoked by resetGraphicView() and ViewTreeBaseListenerImpl to finalize
   * the instance, or to set it as a new object before the initialization, 
   * with the exception of whyText_ that is set to null to indicate the finalizing operation.
   * viewInBase_ may be already finalized when this method is called
   * (see InternalFrame.openSketcher()).
   * Overridden by GraphicEPRView and GraphicNoThanView.
   */
   public synchronized void finalizeForGui() {

          // for testing and tracking
          //System.out.println("Call of GraphicView.finalizeForGui()");

     m_GraphicNodesBounds_.clear();
     l_LinksAndRights_ = null;
     whyText_.removeAll();
     whyText_ = null;
     l_currentArrowsAndRightsStarts_ = null;
     nodesNumber_ = 0;
     dragSourceNodeName_ = null;
     dragTargetNodeName_ = null;
     l_allNodesInLinks_.clear();

     GraphicNode node;
     // GraphicNodes are cleared
      for (Iterator ite = m_GraphicNodes_.values().iterator() ; ite.hasNext() ; ) {
          node = (GraphicNode)ite.next();
          remove(node);
      } // for ite
      validate();
      m_GraphicNodes_.clear();
      removeAll();
      node = null;

      for (Iterator ite = m_l_DisplayableLinks_.values().iterator() ; ite.hasNext() ; ) {
        for (Iterator ite2 = ((List) ite.next()).iterator() ; ite2.hasNext() ; ) {
          ((DisplayableLinkImpl)ite2.next()).finalizeForGui();
        }
      } // for ite
      m_l_DisplayableLinks_.clear();

      viewInBase_ = null;
   } // finalizeForGui()
  
  /**
  * @return Dimension of the minimum GraphicView size
  * */
  public Dimension getMinimumSize(  ) {
    return getSize();
  }

  /**
  * @return Dimension of the maximum GraphicView size
  * */
  public Dimension getMaximumSize(  ) {
    return getSize();
  }

  /**
  * @return Dimension of the preferred GraphicView size
  * */
  public Dimension getPreferredSize(  ) {
    return getSize();
  }

  /**
   * Display at the top left corner of the view, the user message
   * 'Please wait...' in green, as a GraphicText. <P>
   * The map m_GraphicNodes_ is updated, but not m_GraphicNodesBounds_. <P>
   * Called by GraphicViewBaseListenerImpl.updateFrameViewRights(),
   * GraphicViewPropertyEditor.propertyChange() just before the call of resetRights(),
   * and resetGraphicView() just before a call to repaint().
   * After the call of one of this method, the user message has to be removed
   * through a call to removeWaitingMessage(). No use with resetGraphicView(). <P>
   * This method has to be called only in the event dispatch thread, to avoir any thread collision.
   */
  public void addWaitingMessage() {

    if (m_GraphicNodes_.get(new NameImpl(WAITING_TEXT)) != null) return;

    GraphicText gep = new GraphicText(this, WAITING_TEXT.substring(1, WAITING_TEXT.length() - 1) );


                                // for testing and tracking
                            //System.out.println("GraphicView.addWaitingMessage() starting for "+
                              //      viewInBase_.getName());

    // lock to avoid a conflict with other threads
    //synchronized (this) {
        // add to this JPanel
        add(gep );
        // put it at the left top corner
        gep.setBounds(
          0,
          0,
          (int)gep.getShape().getBounds2D().getWidth(),
          (int)gep.getShape().getBounds2D().getHeight());

        gep.setLocation(
            new Point(
              (new Double(getInsets().left + GraphicNode.L_NINE_GRID[0].getX())).intValue(),
              (new Double(getInsets().top + GraphicNode.L_NINE_GRID[0].getY())).intValue())
        );

        gep.setVisible(true);
    //} // synchronized

        // DO NOT add nodeBounds to its map
        //m_GraphicNodesBounds_.put(new NameImpl(WAITING_TEXT), gep.getBounds());

        // add the node to its map
        m_GraphicNodes_.put(new NameImpl(WAITING_TEXT), gep);

                            // for testing and tracking
                            //System.out.println("GraphicView.addWaitingMessage(), nodes in m_GraphicNodes_: ");
                            //for (ImmutableName na : m_GraphicNodes_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na
                            //System.out.println("nodes in m_GraphicNodesBounds_: ");
                            //for (ImmutableName na : m_GraphicNodesBounds_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na


        gep = null;
  } // addWaitingMessage

  /**
   * Removes at the top left corner of the view, the user message
   * 'Please wait...' in green. <P>
   * The map m_GraphicNodes_ is updated, but not m_GraphicNodesBounds_. <P>
   * Called by GraphicViewBaseListenerImpl.changeFrameView() and updateFrameViewRights(),
   * and by GraphicViewPropertyEditor.propertyChange() just before the call to repaint().<P>
   * Called only in the event dispatch thread, but the method is synchronized
   * to detect any thread collision.
   * No operation if there is no waiting message in the view.
   */
  public synchronized void removeWaitingMessage() {

    NameImpl waiting = new NameImpl(WAITING_TEXT);

    if (m_GraphicNodes_.get(waiting) == null) {
        waiting = null;
        return;
    }

    GraphicText gep = (GraphicText)m_GraphicNodes_.get(waiting);
    remove(gep);
    m_GraphicNodes_.remove(waiting);

    waiting = null;
    gep = null;
  } // removeWaitingMessage()

  //
  //--------------------- protected methods -----------------------------------
  //

  /**
  * Paint the GraphicView.
  * This method is designed to be fast. It draws Shapes and Strings.
  * Overrides the super method and calls it.
  * @param _g Graphics.
  * */
  protected void paintComponent(Graphics _g) {

                // for testing and tracking
                //System.out.println("call to GraphicView.paintComponent() ");

    // variables
    Graphics2D g2 = (Graphics2D)_g;

    // paint the JPanel components: each GraphicNode
    super.paintComponent(_g);

    drawArrowsAndRights(g2);

    g2 = null;
  } // paintComponent()

  /**
  * Sets the GraphicNodes in a 16, 25, 49 or 100 nodes layout in the JPanel,
  * with their absolute coordinates. Called by intialize().
  * Updates m_GraphicNodes_, m_GraphicNodesBounds_.
  * Overridden method.
  * */
  protected void setGraphicNodes() {
    //variables
    GraphicNode       gep;
    BaseObject        obj;
    int               index = 0;

    // sets a fixed size
    setViewSize(nodesNumber_);

    // sets the nodes positions
    Point[]           l_positions = setGraphicNodesPositions(nodesNumber_, true);

                        // for testing and tracking
                    //System.out.println("In GraphicView.setGraphicNodes()");
                    //System.out.println("nodesNumber_: "+nodesNumber_);
                    //System.out.println("l_positions:");
                    //for (int i = 0; i < l_positions.length; i++ ) {
                      //System.out.println("--"+i+": "+ l_positions[i].toString() );
                    //}
                    //System.out.println("l_noAccessTargetEP:");
                    //for (Iterator itetest = l_noAccessTargetEP.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_accessTargetEP_:");
                    //for (Iterator itetest = l_accessTargetEP_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}
                    //System.out.println("l_nonEpResources_:");
                    //for (Iterator itetest = l_nonEpResources_.iterator() ; itetest.hasNext() ; ) {
                      //System.out.println(" --- "+ itetest.next().toString() );
                    //}

    for (Iterator ite = viewInBase_.getEorM_BaseObjects().values().iterator() ; ite.hasNext(); ) {

        obj = (BaseObject) ite.next();
        gep = new GraphicContainer(this, obj );
        // add to this JPanel
        add(gep );
        // put it at the right place
        gep.setBounds(
          0,
          0,
          (int)gep.getShape().getBounds2D().getWidth(),
          (int)gep.getShape().getBounds2D().getHeight());

        gep.setLocation(l_positions[index]);

        // add nodeBounds to its map
        m_GraphicNodesBounds_.put(obj.getName(), gep.getBounds());
        // add the node to its map
        m_GraphicNodes_.put(obj.getName(), gep);
        gep.setVisible(true);
        index++;

    } // for (Iterator ite

    gep = null;
    obj = null;
    l_positions = null;
  } // setGraphicNodes()


  /**
  * Sets the space of all the allowed positions for the nodes in the view,
  * while the method setGraphicNodeLocations provides the effective location of each node.
  * There are 4 sets of node positions: one if _nodesNumber
  * is inferior to 8 (offering 16 possible positions),  <P>
  * one between 8 and 12 nodes (25 positions); <P>
  * one between 13 and 20 nodes (49 possible positions), <P>
  * and one for the values from 21 to 46 nodes (100 possible positions). <P>
  * Each node position is at the top left corner of the node.
  * Must be called always after setViewSize().
  * Called by setGraphicNodes(), mouseReleased(), setGraphicNodeLocations(),
  * GraphicEPRView.setGraphicNodes() and GraphicNoThanView.setGraphicNodes().
  * Overridden by the GraphicSketchView method.
  * @param _nodesNumber is the number of BaseObjects in the view. May be 0.
  *   No operation if it is superior to 46 or negative.
  * @param _initial true if it is for the initial view where the nodes are put
  * on a circle or a scare
  * @return the array of node positions, in the correct order;
  *   equal to Point[0] if no operation. No null item in the array.
  * */
  protected final Point[] setGraphicNodesPositions(int _nodesNumber, boolean _initial) {
    if (_nodesNumber < 0 || _nodesNumber > 46) return new Point[0];
    Point[] l_result;
    int insertLeft = getInsets().left;
    int insertTop = getInsets().top;

                                // for testing and tracking
                            //System.out.println("In GraphicView.setGraphicNodesPositions(), "+
                              //      "insertLeft = "+insertLeft+
                                //    " and insertTop = "+insertTop);


    if (_nodesNumber < 8) {

      if (_initial) {

                                // for testing and tracking
                            //System.out.println("In GraphicView.setGraphicNodesPositions(), "+
                              //      "for initial and less than 8 nodes ");
                            //System.out.println("--- use GraphicNode.L_SIXTEEN_GRID:");
                            //for (int i = 0; i < GraphicNode.L_SIXTEEN_GRID.length; i++ ) {
                              //if (GraphicNode.L_SIXTEEN_GRID[i] != null)
                                //  System.out.println("--"+i+": "+
                                  //        GraphicNode.L_SIXTEEN_GRID[i].toString() );
                            //} // for i

          l_result = new Point[8]; // values after _nodesNumber are not null
          // 8 node positions among the 16 grid locations
          Integer[] l_position = new Integer[] {1, 2, 4, 7, 8, 11, 13, 14};
          for (int i = 0; i < 8; i++) {

                                // for testing and tracking
                            //System.out.println("In GraphicView.setGraphicNodesPositions(), "+
                              //      "use GraphicNode.L_SIXTEEN_GRID at the index "+
                                //    l_position[i].intValue()+" and "+
                                  //  "GraphicNode.L_SIXTEEN_GRID[ l_position[i].intValue() ].getY():"+
                                    //GraphicNode.L_SIXTEEN_GRID[ l_position[i].intValue() ].getY());

            l_result[i] = new Point(
              (new Double(insertLeft +
              GraphicNode.L_SIXTEEN_GRID[ l_position[i].intValue() ].getX())).intValue(),
              (new Double(insertTop +
              GraphicNode.L_SIXTEEN_GRID[ l_position[i].intValue() ].getY())).intValue()
            );
          } // for i
      } else {

                                // for testing and tracking
                            //System.out.println("In GraphicView.setGraphicNodesPositions(), "+
                              //      "NOT for initial, and less than 8 nodes ");

          l_result = new Point[16];
          for (int i = 0; i < 16; i++) { // 16 positions of the grid
            l_result[i] = new Point(
              (new Double(insertLeft + GraphicNode.L_SIXTEEN_GRID[i].getX())).intValue(),
              (new Double(insertTop + GraphicNode.L_SIXTEEN_GRID[i].getY())).intValue());
          } // for i
      } // else

    } // if (_nodesNumber < 8)

    else if (_nodesNumber < 13) {

      if (_initial) {
          l_result = new Point[12]; // values after _nodesNumber are not null
          // 12 node positions among the 25 grid locations
          Integer[] l_position = new Integer[] {1, 3, 5, 7, 9, 10, 14, 15, 17, 19, 20, 24};
          for (int i = 0; i < 12; i++) {
            l_result[i] = new Point(
              (new Double(insertLeft +
              GraphicNode.L_TWENTY_FIVE_GRID[ l_position[i].intValue() ].getX())).intValue(),
              (new Double(insertTop +
              GraphicNode.L_TWENTY_FIVE_GRID[ l_position[i].intValue() ].getY())).intValue()
            );
          } // for i
      } else {
          l_result = new Point[25];
          for (int i = 0; i < 25; i++) { // 25 positions of the grid
            l_result[i] = new Point(
              (new Double(insertLeft + GraphicNode.L_TWENTY_FIVE_GRID[i].getX())).intValue(),
              (new Double(insertTop + GraphicNode.L_TWENTY_FIVE_GRID[i].getY())).intValue());
          } // for i
      } // else

    } // if (_nodesNumber < 12)

    else if (_nodesNumber < 21) {
      if (_initial) {
          l_result = new Point[24]; // values after _nodesNumber are not null
          // 24 node positions in a double circle among the 49 grid locations.
          // the 8-nodes internal circle is 16, 17, 18, 23, 25, 30, 31, 32.
          // the nodes are drawn first on the internal circle, the left and the top.
          Integer[] l_position = new Integer[] {16, 17, 18, 23, 25, 30, 31, 32,
            14, 21, 3, 8, 9, 11, 12, 20, 27, 28, 34, 36, 37, 39, 40, 45};
          for (int i = 0; i < 24; i++) {
            l_result[i] = new Point(
              (new Double(insertLeft +
              GraphicNode.L_FOURTY_NINE_GRID[ l_position[i].intValue() ].getX())).intValue(),
              (new Double(insertTop +
              GraphicNode.L_FOURTY_NINE_GRID[ l_position[i].intValue() ].getY())).intValue()
            );
          } // for i
      } else {
          l_result = new Point[49];
          for (int i = 0; i < 49; i++) { // 49 positions of the grid
            l_result[i] = new Point(
              (new Double(insertLeft + GraphicNode.L_FOURTY_NINE_GRID[i].getX())).intValue(),
              (new Double(insertTop + GraphicNode.L_FOURTY_NINE_GRID[i].getY())).intValue());
          } // for i
      } // else
    }

    else { // 21 to 46 nodes
      if (_initial) {
          l_result = new Point[46]; // values after _nodesNumber are not null
          // 46 node positions in a double scare among the 100 grid locations.
          // the 14-nodes internal scare is 23, 24, 25, 32, 36, 42, 46, 52, 56, 62, 66,
          // 73, 74, 75.
          // the nodes are drawn first on the internal scare, the left and the top.
          Integer[] l_position = new Integer[] {23, 24, 25, 32, 36, 42, 46, 52, 56, 62, 66,
            73, 74, 75, 10, 20, 30, 40, 50, 60, 70, 80, 1, 2, 3, 4, 5, 6, 7, 8,
            91, 92, 93, 94, 95, 96, 97, 98, 19, 29, 39, 49, 59, 69, 79, 89};
          for (int i = 0; i < 46; i++) {
            l_result[i] = new Point(
              (new Double(insertLeft +
              GraphicNode.L_ONE_HUNDRED_GRID[ l_position[i].intValue() ].getX())).intValue(),
              (new Double(insertTop +
              GraphicNode.L_ONE_HUNDRED_GRID[ l_position[i].intValue() ].getY())).intValue()
            );
          } // for i
      } else {
          l_result = new Point[100];
          for (int i = 0; i < 100; i++) { // 100 positions of the grid
            l_result[i] = new Point(
              (new Double(insertLeft + GraphicNode.L_ONE_HUNDRED_GRID[i].getX())).intValue(),
              (new Double(insertTop + GraphicNode.L_ONE_HUNDRED_GRID[i].getY())).intValue());
          } // for i
      } // else
    } // else

    return l_result;
  } // setGraphicNodesPositions()

  /**
   * Add/remove to the graphic view a base object and its location, if non-null, and
   * restores the old locations for the old objects. Adapts the locations to the new grid
   * if necessary. Works even if there are more than one new node, for instance with
   * an ACSObject and its alias EPRViewInBase adds to itself. When the ViewInBase
   * contains a new object that is not named by _newNodeName, this object is unknown
   * in m_GraphicNodesBounds_, and it is put at the end of the possible node locations.
   * Called only by resetGraphicView() and after initialize().
   * Overridden by GraphicNoThanView which do nothing,
   * since its initialize() method makes the job.
   * @param _newNodeName name of the base object to add. May be null.
   * @param _m_GraphicNodesPosition is the map of the old locations. May be empty.
   */
  protected void setGraphicNodeLocations(ImmutableName _newNodeName, Map _m_GraphicNodesPosition) {
    ImmutableName   name = null;
    Point[]         l_points = setGraphicNodesPositions(nodesNumber_, false);

    if (_newNodeName != null) {
        
        ((GraphicNode)m_GraphicNodes_.get(_newNodeName)).setLocation(
            new Point(
              (new Double(getInsets().left + GraphicNode.L_SIXTEEN_GRID[0].getX())).intValue(),
              (new Double(getInsets().top + GraphicNode.L_SIXTEEN_GRID[0].getY())).intValue())
            );
        m_GraphicNodesBounds_.put(_newNodeName,
            ((GraphicNode)m_GraphicNodes_.get(_newNodeName)).getBounds() );
    } // if

     // restores the previous node locations, if the new node does not push
    // the view to a new grid of node locations
    if ( (_newNodeName != null && nodesNumber_ != 7 && nodesNumber_ != 11 && nodesNumber_ != 20) ||
            (_newNodeName == null && nodesNumber_ != 8 && nodesNumber_ != 12 && nodesNumber_ != 21)
            ) {
        for (Iterator ite = m_GraphicNodes_.keySet().iterator(); ite.hasNext(); ) {
            name = (ImmutableName)ite.next();
            if (_m_GraphicNodesPosition.get(name) != null) {
                ((GraphicNode)m_GraphicNodes_.get(name)).setLocation(
                            (Point)_m_GraphicNodesPosition.get(name));
                m_GraphicNodesBounds_.put(name,
                            ((GraphicNode)m_GraphicNodes_.get(name)).getBounds() );
            } // if
            else if (!name.equals(new NameImpl(WAITING_TEXT)) &&
                    _newNodeName != null && !_newNodeName.equals(name)) {
                // this is an alias or reference just added by EPRViewInBase,
                // and to put at the last place
                ((GraphicNode)m_GraphicNodes_.get(name)).setLocation(
                    l_points[l_points.length - 1]);
                m_GraphicNodesBounds_.put(name,
                    ((GraphicNode)m_GraphicNodes_.get(name)).getBounds() );
            } // else
        } // for ite
    }
    else { // try to adapt the old locations to the new grid

                            // for testing and tracking
                            //System.out.println("GraphicView.setGraphicNodeLocations(), "+
                              //      "try to match the old locations in m_GraphicNodes_: ");

        for (Iterator ite = m_GraphicNodes_.keySet().iterator(); ite.hasNext(); ) {
            name = (ImmutableName)ite.next();

                            //System.out.println("--- for "+name);

            if (name.equals(new NameImpl(WAITING_TEXT)))
                continue;

                            // for testing and tracking
                            //System.out.println("GraphicView.setGraphicNodeLocation(), nodes in m_GraphicNodes_: ");
                            //for (ImmutableName na : m_GraphicNodes_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na
                            //System.out.println("nodes in m_GraphicNodesBounds_: ");
                            //for (ImmutableName na : m_GraphicNodesBounds_.keySet()) {
                              //   System.out.println("---- "+na);
                            //} // for na

            if (_m_GraphicNodesPosition.get(name) == null) {

                            //System.out.println("--- setGraphicNodesPositions() is null for "+
                              //                              name);

                if (_newNodeName != null && !_newNodeName.equals(name)) {
                    // this is a node just added by EPRViewInBase (an alias or reference),
                    // and to put at the last place (where a current node may already be)
                    ((GraphicNode)m_GraphicNodes_.get(name)).setLocation(
                        l_points[l_points.length - 1]);
                    m_GraphicNodesBounds_.put(name,
                        ((GraphicNode)m_GraphicNodes_.get(name)).getBounds() );
                    continue;
                }
                else if (_newNodeName != null) {
                    // the new node
                    continue;
                } // else
            } // if

                            //System.out.println("--- from setGraphicNodesPositions() with "+
                              //                      "nodesNumber_ = "+nodesNumber_+
                                //                    " for "+name);

            // the index 0 is never a target, to let it for the new node ot the waiting text
            for (int i = 1; i < l_points.length; i++) {

                            //System.out.println("------ from setGraphicNodesPositions(), index "+i+
                              //                              ", l_points[i]: "+l_points[i]);

                if ( l_points[i].distance((Point)_m_GraphicNodesPosition.get(name)) <
                        GraphicNode.NODE_SIZE ) {

                    ((GraphicNode)m_GraphicNodes_.get(name)).setLocation(
                        l_points[i]);
                    m_GraphicNodesBounds_.put(name,
                        ((GraphicNode)m_GraphicNodes_.get(name)).getBounds() );
                    break;
                } // if
            } // for i
        } // for ite
    } // else

    l_points = null;
    name = null;
  } // setGraphicNodeLocations()

  /**
  * Set the size of the JPanel when the view contains less than 47 elements.
  * If _nodesNumber is inferior to 8, the size is GraphicNode.EIGHT_NODES_VIEW_SIZE.
  * Else if _nodesNumber is inferior to 13, the size is GraphicNode.TWELVE_NODES_VIEW_SIZE.
  * Else if _nodesNumber is inferior to 21, the size is GraphicNode.TWENTY_FOUR_NODES_VIEW_SIZE.
  * Else the size is GraphicNode.FORTY_SIX_NODES_VIEW_SIZE.
  * Called by setGraphicNodes(). Overridden method in GraphicSketchView.
  * @param _nodesNumber is the number of BaseObjects in the view.
  *   No operation if it is superior to 47 or negative.
  * */
  protected void setViewSize(int _nodesNumber) {
    if (_nodesNumber < 0 || _nodesNumber > 47) return;

    if (_nodesNumber < 8) {

                        // for testing and tracking
                    //System.out.println("In GraphicView.setViewSize(), "+
                      //      "set GraphicNode.EIGHT_NODES_VIEW_SIZE ");

      setSize(GraphicNode.EIGHT_NODES_VIEW_SIZE);
    }
    else if (_nodesNumber < 13) {
      setSize(GraphicNode.TWELVE_NODES_VIEW_SIZE);
    }
    else if (_nodesNumber < 21) {
      setSize(GraphicNode.TWENTY_FOUR_NODES_VIEW_SIZE);
    }
    else {
      setSize(GraphicNode.FORTY_SIX_NODES_VIEW_SIZE);
    }
  } // setViewSize()


  /**
  * Sets the property l_LinksAndRights_ which contains the links and rights as Shapes,
  * to be painted by drawArrowsAndRights() in this GraphicView.
  * There are 4 elements for each typical arrow:
  * 3 for the arrow and 1 text for the description,
  * which are often the rights to display.
  * Only structural path is drawn; it is associated to a BRIDGE relation.
  * These Shapes and Strings are actually drawn in drawArrowsAndRights().
  * Called by the contructor after initialize(), by resetRights() and mouseReleased().
  * Calls setPoint2DForArrowAndRights().
  * Synchronized on m_l_DisplayableLinks_. Overridden method.
  * @param _m_l_DisplayableLinks is the Map of DisplayableLinks lists (one per pair)
  * associated to the view.
  * @param _m_gNodeBoundsModel is the Map of the node bounds for the shapes
  * in the GraphicView.
  * @return Object[] to draw in paintComponent(), with both the arrows
  * at absolute positionning (1 Shape for the line and 2 empty Shapes for the head)
  * and the access rights and comments (1 String).
  * */
  protected Object[] buildUpArrowsAndRights(Map _m_l_DisplayableLinks,
    Map _m_gNodeBoundsModel) {

    // works on _m_l_DisplayableLinks, set in initialize()
    if (_m_l_DisplayableLinks.isEmpty()) return NULL_SHAPE_ARRAY;

    // variables
    Object[]         result;
      // list of pairs (ACS, ACS)
    List            l_links = new ArrayList(INITIAL_CAPACITY);
    DisplayableLinkImpl link;
    BaseObject[]    l_entities = new BaseObject[2];
    Point2D.Float[] l_points = new Point2D.Float[3];
    int             result_index = 0;
    int             epArrowsCount = 0;
    Object          lastEpEntity = null;

    synchronized (_m_l_DisplayableLinks) {

      // the result length is four time the _m_l_DisplayableLinks size
      // since each access control link has a arrow with 3 lines
      // (main line + 2 lines for head)
      // plus a rights string to draw
      result = new Object[4*_m_l_DisplayableLinks.size()];
      l_currentArrowsAndRightsStarts_ = new Object[4*_m_l_DisplayableLinks.size()];

      // handles the matched relation
      for (Iterator ite = _m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {

        l_links = (List)ite.next();

        // selects only the first DisplayableLinkImpl, since it gives the two ends
        link = (DisplayableLinkImpl) l_links.get(0);
        l_entities[0] = link.getEorFirstEnd();
        l_entities[1] = link.getEorSecondEnd();

        // to draw several arrows around a node, count them with epArrowsCount
        if (l_entities[0] != lastEpEntity) {
          lastEpEntity = l_entities[0];
          epArrowsCount = 0;
        }

                      // for testing and tracking
                      //System.out.println("in GraphicView.buildUpArrows(), "+
                        //"_m_l_DisplayableLinks iteration, result_index: "+result_index+
                        //", and result.length: "+result.length);

                      //for testing and tracking
                      //int e = 0;
                      //System.out.println("in GraphicView.buildUpArrows() iteration, one dLink list is:");
                      //for (Iterator ite_test = l_links.iterator() ; ite_test.hasNext() ; ) {
                        //link = (DisplayableLinkImpl) ite_test.next();
                        //System.out.println("        at ["+e+"], from "+ link.getEorFirstEnd().getName() +
                          //      " to "+ link.getEorSecondEnd().getName());
                        //e++;
                      //} // ite_test

                      // for testing and tracking
                      //System.out.println("in GraphicView.buildUpArrows(), "+
                        //"_m_l_DisplayableLinks iteration, result_index: "+result_index+
                        //", rights: "+rights+" and hasOnlyStructuralLinks: "+hasOnlyStructuralLinks);

        epArrowsCount++;

        // computes link points and rights Point2D.Float
        // associated to this access path
        l_points = setPoint2DForArrowAndRights(
          l_entities[0].getName(), l_entities[1].getName(), _m_gNodeBoundsModel, epArrowsCount);
        // to have a clear diagram, make a jump on the arrow, up or down:
        // - up if its sens is from left to right,
        if (l_points[0].getX() < l_points[1].getX()+30) {
            l_points[0] = new Point2D.Float(
                    (float)l_points[0].getX(), (float)l_points[0].getY() - 10);
            l_points[1] = new Point2D.Float(
                    (float)l_points[1].getX(), (float)l_points[1].getY() - 10);
            l_points[2] = new Point2D.Float(
                    (float)l_points[2].getX(), (float)l_points[2].getY() - 10);
        }
        // - down if its sens is from right to left
        else if (l_points[0].getX() > l_points[1].getX()+30) {
            l_points[0] = new Point2D.Float(
                    (float)l_points[0].getX(), (float)l_points[0].getY() + 10);
            l_points[1] = new Point2D.Float(
                    (float)l_points[1].getX(), (float)l_points[1].getY() + 10);
            l_points[2] = new Point2D.Float(
                    (float)l_points[2].getX(), (float)l_points[2].getY() + 10);
        }
        // - no action if it is a vertical arrow (+-30)

        // constructs each line and arrow without head
        result[result_index] = new Line2D.Float(l_points[0], l_points[1]);
        l_currentArrowsAndRightsStarts_[result_index] = result[result_index];

                // for testing and tracking
                //System.out.println("   a line to draw for the index: "+result_index);
                //System.out.println("      start point: x: "+((Line2D.Float)result[result_index]).getP1().getX()+
                  //"; y: "+((Line2D.Float)result[result_index]).getP1().getY());
                //System.out.println("      end point: x: "+((Line2D.Float)result[result_index]).getP2().getX()+
                  //"; y: "+((Line2D.Float)result[result_index]).getP2().getY());


        result_index++;

        // sets to null the 2-lines arrow head.
        result_index++;
        result_index++;

                // for testing and tracking
                //System.out.println("   one arrow to draw: ");
                //System.out.println("   start point: x: "+((Line2D.Float)result[result_index]).getP1().getX()+
                  //"; y: "+((Line2D.Float)result[result_index]).getP1().getY());
                //System.out.println("   end point: x: "+((Line2D.Float)result[result_index]).getP2().getX()+
                  //"; y: "+((Line2D.Float)result[result_index]).getP2().getY());

                // for testing and tracking
                //System.out.println("in GraphicView.buildUpArrows(), rights:"+rights);

        // the text begins by '<>' for having a blue dashed line,
        // and these 2 caracters are not displayed.
        result[result_index] = "<> ";

        // this is the only element for which l_currentArrowsAndRightsStarts_
        // and result is different.
        l_currentArrowsAndRightsStarts_[result_index] = l_points[2];

                  // for testing and tracking
                  //System.out.println("in GraphicView.buildUpArrows(), rights or description:"+result[result_index]+
                    //" and l_points[2]: "+l_points[2]);

        result_index++;

      } // for (Iterator ite = _m_l_DisplayableLinks

    } // synchro

    l_links = null;
    link = null;
    l_entities = null;
    l_points = null;

    return result;
  } // buildUpArrowsAndRights()

  /**
  * Sets the absolute positions of the line or arrow for a pair of GraphicNodes,
  * by the mean of the shape positions for the 2 GraphicRects that are associated to the pair.
  * Defines the default value for the rights string start,
  * which is near the start of the arrow, with a shift that is RIGHTS_SHIFT.
  * Takes account of the relative positions of the two base objects.
  * Called by buildUpArrowsAndRights(), in this class and in the subclasses,
  * but not by GraphicNoThanView.
  * Synchronized on m_l_DisplayableLinks_ through buildUpArrowsAndRights().
  * @param _end_one is the name of a GraphicNode to display
  * @param _end_two is the name of a GraphicNode to display
  * @param _m_gNodeBoundsModel is the Map of the node bounds in the GraphicView,
  *   where the GraphiNode name is the key.
  * @param _epArrowsCount is the number of arrows already drawn for the node,
  *   + 1.
  * @return Point2D.Float[] array of 3 items: arrowStartPoint2D and arrowEndPoint2D
  *   defines the line (or arrrow) start and end points, and defaultRightsStartPoint2D
  *   defines the start point where the String is drawn.
  * */
  protected static final Point2D.Float[] setPoint2DForArrowAndRights(
    ImmutableName _end_one, ImmutableName _end_two, Map _m_gNodeBoundsModel,
    int _epArrowsCount) {
    // variables
    Rectangle2D     bounds_ep;
    Rectangle2D     bounds_res;
    Point2D.Float   center_ep;
    Point2D.Float   center_res;
    Point2D.Float   arrowStartPoint = new Point2D.Float();
    Point2D.Float   arrowEndPoint = new Point2D.Float();
    Point2D.Float   stringStartPoint = new Point2D.Float();
    float           yshift;
    float           yStringStartPoint;
    float           xshift;
    float           xStringStartPoint;
    float           dirFactor = 0;
    float           constant = 0;
    Point2D.Float[] result = new Point2D.Float[3];

    // gets the rectangular bounds of the 2 implied GraphicNodes
    bounds_ep = (Rectangle2D)_m_gNodeBoundsModel.get( _end_one );
    bounds_res = (Rectangle2D)_m_gNodeBoundsModel.get( _end_two );

                        // for testing and tracking
                        //System.out.println("GraphicView.setPoint2DForArrowAndRights(): ");
                        //System.out.println("   bounds_ep: "+bounds_ep+" for"+_end_one);
                        //System.out.println("   bounds_res: "+bounds_res+" for"+_end_two);

    // error
    //if (bounds_ep == null)
      //  throw new InternalError("GraphicView.setPoint2DForArrowAndRights(): bounds_ep null");
    if (bounds_ep == null || bounds_res == null)
      return new Point2D.Float[] {
        new Point2D.Float(), new Point2D.Float(), new Point2D.Float() };

    // gets the center of the 2 implied GraphicEntities
    center_ep = new Point2D.Float( (float)bounds_ep.getCenterX(),
      (float)bounds_ep.getCenterY());
    center_res = new Point2D.Float( (float)bounds_res.getCenterX(),
      (float)bounds_res.getCenterY());

    // computes the director factor of the arrow straight line
    // that is defined by the 2 centers.
    if (center_ep.getX() != center_res.getX()) {
      dirFactor = (float)( (float)center_ep.getY() - (float)center_res.getY()) /
        ((float)center_ep.getX() - (float)center_res.getX());
      // and the constant
      constant = (float)((float)center_ep.getY() - (float)dirFactor*center_ep.getX());
    }

    if (center_res.getY() < center_ep.getY() &&
      Math.abs(center_res.getX() - center_ep.getX()) < NODE_DISTANCE ) {
        // the Resource is in the same X, above
        arrowStartPoint = new Point2D.Float( (float)center_ep.getX(),
            (float)bounds_ep.getMinY() + 5);
        arrowEndPoint = new Point2D.Float(
            (float)center_res.getX() + 5,
            (float)bounds_res.getMaxY() + GraphicNode.NODE_SIZE/8 - 5);

        yshift = (float)arrowEndPoint.getY() - (float)arrowStartPoint.getY();
        yStringStartPoint = (float)arrowStartPoint.getY() +
          _epArrowsCount*(yshift/10);
        stringStartPoint = new Point2D.Float(
            (float)center_ep.getX() + RIGHTS_SHIFT ,
            yStringStartPoint);
    }
    else if (center_res.getY() > center_ep.getY() &&
      Math.abs(center_res.getX() - center_ep.getX()) < NODE_DISTANCE ) {
        // the Resource is in the same X, below
        arrowStartPoint = new Point2D.Float( (float)center_ep.getX(),
            (float)bounds_ep.getMaxY() - 5);
        arrowEndPoint = new Point2D.Float(
            (float)center_res.getX(),
            (float)bounds_res.getMinY() - GraphicNode.NODE_SIZE/8 + 10);

        yshift = (float)arrowEndPoint.getY() - (float)arrowStartPoint.getY();
        yStringStartPoint = (float)arrowStartPoint.getY() +
          _epArrowsCount*(yshift/10) + 8*yshift/NODE_DISTANCE; // new: + 8*yshift/NODE_DISTANCE
        stringStartPoint = new Point2D.Float(
            (float)center_ep.getX() + RIGHTS_SHIFT ,
            yStringStartPoint);
    }
    else if (Math.abs(center_res.getY() - center_ep.getY()) < NODE_DISTANCE &&
      center_res.getX() > center_ep.getX() ) {
        // the Resource is in the same Y, at right
        arrowStartPoint = new Point2D.Float( (float)bounds_ep.getMaxX(),
            (float)center_ep.getY() + 5);
        arrowEndPoint = new Point2D.Float(
            (float)bounds_res.getMinX() - GraphicNode.NODE_SIZE/8,
            (float)center_res.getY() + 5);  // previous: - 5

        xshift = (float)arrowEndPoint.getX() - (float)arrowStartPoint.getX();
        xStringStartPoint = (float)arrowStartPoint.getX() +
          _epArrowsCount*(xshift/10);
        stringStartPoint = new Point2D.Float(xStringStartPoint,
            (float)center_ep.getY() - RIGHTS_SHIFT );
    }
    else if (Math.abs(center_res.getY() - center_ep.getY()) < NODE_DISTANCE &&
      center_res.getX() < center_ep.getX() ) {
        // the Resource is in the same Y, at left
        arrowStartPoint = new Point2D.Float( (float)bounds_ep.getMinX(),
            (float)center_ep.getY() - 5);
        arrowEndPoint = new Point2D.Float(
            (float)bounds_res.getMaxX() + GraphicNode.NODE_SIZE/8,
            (float)center_res.getY() - 5); // previous: + 5

        xshift = (float)arrowEndPoint.getX() - (float)arrowStartPoint.getX();
        xStringStartPoint = (float)arrowStartPoint.getX() +
          _epArrowsCount*(xshift/10) - 2*NODE_DISTANCE + 3*xshift/NODE_DISTANCE; // - NODE_DISTANCE/2 + xshift/NODE_DISTANCE
        stringStartPoint = new Point2D.Float(xStringStartPoint,
            (float)center_ep.getY() - RIGHTS_SHIFT - 5);
    }
    else if (center_res.getX() + NODE_DISTANCE < center_ep.getX() ) {
        // the Resource is in the third or fourth quarter
        arrowStartPoint = new Point2D.Float( (float)bounds_ep.getMinX(),
            constant+dirFactor*( (float)bounds_ep.getMinX() + 30) );
        arrowEndPoint = new Point2D.Float( (float)bounds_res.getMaxX() + 5,
            constant+dirFactor*( (float)bounds_res.getMaxX() + 5) - 5);

        xshift = (float)arrowEndPoint.getX() - (float)arrowStartPoint.getX();
        xStringStartPoint = (float)arrowStartPoint.getX() +
          _epArrowsCount*(xshift/10) - NODE_DISTANCE/2 + xshift/NODE_DISTANCE; // new: + xshift/NODE_DISTANCE
        stringStartPoint = new Point2D.Float(xStringStartPoint,
            constant+dirFactor*( xStringStartPoint ) ); // -RIGHTS_SHIFT
    }
    else {
        // the Resource is in the first or second quarter
        arrowStartPoint = new Point2D.Float( (float)bounds_ep.getMaxX(),
            constant+dirFactor*( (float)bounds_ep.getMaxX() - 5) );
        arrowEndPoint = new Point2D.Float( (float)bounds_res.getMinX() - 5,
            constant+dirFactor*( (float)bounds_res.getMinX() + 15) ); // MinX() - 5) + 5

        xshift = (float)arrowEndPoint.getX() - (float)arrowStartPoint.getX();
        xStringStartPoint = (float)arrowStartPoint.getX() +
          _epArrowsCount*(xshift/10);
        stringStartPoint = new Point2D.Float(xStringStartPoint,
            constant+RIGHTS_SHIFT+dirFactor*( xStringStartPoint ) ); // -RIGHTS_SHIFT
    } //else

    center_ep = null;
    center_res = null;
    bounds_ep = null;
    bounds_res = null;

    result[0] = arrowStartPoint;
    result[1] = arrowEndPoint;
    result[2] = stringStartPoint;

                        // for testing and tracking
                        //System.out.println("GraphicView.setPoint2DForArrowAndRights(), result: ");
                        //for (int i = 0; i < result.length; i++) {
                          //  System.out.println("------- "+result[i]);
                        //}

    return result;
  } // setPoint2DForArrowAndRights()

//
// ---------------- private methods ---------------------------------------
//

  /**
  * Draws the arrows and the rights in Serif 11 from the values in l_LinksAndRights_
  * and l_currentArrowsAndRightsStarts_.
  * The GraphicNodes are painted by paintComponent() and getExportImage()
  * which call this method.
  * This method is designed to be fast, since Shapes and Strings to draw has been
  * put in the l_LinksAndRights_ array by buildUpArrowsAndRights()
  * before the call of this method.
  * The links are in black, dark grey or blue, dashed or not, as lines or arrows.
  * The texts are in black or blue. Arrows are in dark grey and may be dashed.
  * Only simple lines are in blue.
  * Simple lines and arrows are dashed only if the text begins with '<>',
  * which is then removed.
  * Arrows have double lenght only if the text begins with '><', (>) which is then removed.
  * Arrows are dashed with double lenght only if the text begins with '>>',
  * which is then removed.
  * The text is in blue only if it contains more than one '!', which are not removed.
  * Uses antialiasing on shapes and texts.
  * Called by paintComponent(), getExportImage() and the subclasses, excepted in
  * GraphicNoThanView that Overrides it.
  * @param _g2 is the Graphics2D to draw
  * */
  protected void drawArrowsAndRights(Graphics2D _g2) {
    String     rights;
    int        index;

    _g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING ,
      RenderingHints.VALUE_ANTIALIAS_ON);
    _g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING ,
      RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    _g2.setFont(BeamerManager.plainFont);
    _g2.setPaint(Color.DARK_GRAY);
    _g2.setStroke(BASIC_STROKE);

                            // for testing and tracking
                        //System.out.println(" ");
                        //System.out.println("call to GraphicView.drawArrowsAndRights()");
                        //if (l_LinksAndRights_ != null)
                          //  System.out.println("l_LinksAndRights_.length: "+l_LinksAndRights_.length);

    if (l_LinksAndRights_ != null) {
      for (int i = 0 ; i < l_LinksAndRights_.length ; i++) {

                            // for testing and tracking
                        //System.out.println("    for i = "+i+", type of l_LinksAndRights_[i] : ");
                          //if (l_LinksAndRights_[i] != null) System.out.println("     is "+
                            //  l_LinksAndRights_[i].getClass());
                          //else System.out.println("      l_LinksAndRights_[i] is null ");

        if (l_LinksAndRights_[i] instanceof Line2D.Float) {

              // when it is not an arrow, but a simple line with a text
              if (l_LinksAndRights_.length > i+3 &&
              l_LinksAndRights_[i+1] == null &&
              (l_LinksAndRights_[i+3] instanceof String) ) {

                            // for testing and tracking
                        //System.out.println("    first case of Line2D.Float for i = "+i);
                        //System.out.println(" --- i+1: "+l_LinksAndRights_[i+1]);
                        //System.out.println(" --- i+2: "+l_LinksAndRights_[i+2]);
                        //System.out.println(" --- i+3: "+l_LinksAndRights_[i+3]);

                  _g2.setPaint(Color.blue);

                  if (((String)l_LinksAndRights_[i+3]).startsWith("<>") ) {
                      // for !O!, !ext-acl-C!...
                      _g2.setStroke(DASHED_STROKE);
                      _g2.draw( (Shape)l_LinksAndRights_[i] );
                      _g2.setStroke(BASIC_STROKE);
                  } else _g2.draw( (Shape)l_LinksAndRights_[i] ); // for BRIDGE, Member...

                  _g2.setPaint(Color.DARK_GRAY);

              } // if (l_LinksAndRights_.length > i+3
              // when it is an dashed_stroke arrow
              else if (l_LinksAndRights_.length > i+3 &&
              l_LinksAndRights_[i+1] != null &&
              (l_LinksAndRights_[i+3] instanceof String) &&
              ((String)l_LinksAndRights_[i+3]).startsWith("<>") ) {

                            // for testing and tracking
                        //System.out.println("    link with '><' = "+i);
                        //System.out.println(" --- i+1: "+l_LinksAndRights_[i+1]);
                        //System.out.println(" --- i+2: "+l_LinksAndRights_[i+2]);
                        //System.out.println(" --- i+3: "+l_LinksAndRights_[i+3]);

                  _g2.setStroke(DASHED_STROKE);
                  _g2.draw( (Shape)l_LinksAndRights_[i] );
                  _g2.setStroke(BASIC_STROKE);

              } // if (l_LinksAndRights_.length > i+3
              // when it is an dashed_stroke and double_with arrow
              else if (l_LinksAndRights_.length > i+3 &&
              l_LinksAndRights_[i+1] != null &&
              (l_LinksAndRights_[i+3] instanceof String) &&
              ((String)l_LinksAndRights_[i+3]).startsWith(">>") ) {

                            // for testing and tracking
                        //System.out.println("    link with '>>' = "+i);
                        //System.out.println(" --- i+1: "+l_LinksAndRights_[i+1]);
                        //System.out.println(" --- i+2: "+l_LinksAndRights_[i+2]);
                        //System.out.println(" --- i+3: "+l_LinksAndRights_[i+3]);

                  _g2.setStroke(DASHED_DOUBLEWIDTH_STROKE);
                  _g2.draw( (Shape)l_LinksAndRights_[i] );
                  _g2.setStroke(BASIC_STROKE);

              } // if (l_LinksAndRights_.length > i+3
              // when it is an doublewidth_stroke arrow
              else if (l_LinksAndRights_.length > i+3 &&
              l_LinksAndRights_[i+1] != null &&
              (l_LinksAndRights_[i+3] instanceof String) &&
              ((String)l_LinksAndRights_[i+3]).startsWith("><") ) {

                            // for testing and tracking
                        //System.out.println("    link with '><' = "+i);
                        //System.out.println(" --- i+1: "+l_LinksAndRights_[i+1]);
                        //System.out.println(" --- i+2: "+l_LinksAndRights_[i+2]);
                        //System.out.println(" --- i+3: "+l_LinksAndRights_[i+3]);

                  _g2.setStroke(DOUBLEWIDTH_STROKE);
                  _g2.draw( (Shape)l_LinksAndRights_[i] );
                  _g2.setStroke(BASIC_STROKE);

              } // if (l_LinksAndRights_.length > i+3
              // draw the two sides of the arrow head,
              else if (l_LinksAndRights_.length > i+1 &&
                           l_LinksAndRights_[i+1] instanceof String ) {
                  //_g2.setPaint(Color.DARK_GRAY);
                  _g2.draw( (Shape)l_LinksAndRights_[i] );
              } else if (l_LinksAndRights_.length > i+2 &&
                           l_LinksAndRights_[i+2] instanceof String ) {
                  //_g2.setPaint(Color.DARK_GRAY);
                  _g2.draw( (Shape)l_LinksAndRights_[i] );
              // draw a normal line for an arrow
              } else {

                            // for testing and tracking
                        //System.out.println("    third case of Line2D.Float for i = "+i);
                        //System.out.println(" --- i+1: "+l_LinksAndRights_[i+1]);
                        //System.out.println(" --- i+2: "+l_LinksAndRights_[i+2]);
                        //System.out.println(" --- i+3: "+l_LinksAndRights_[i+3]);

                  _g2.draw( (Shape)l_LinksAndRights_[i] );

              } // else

        } // if (l_LinksAndRights_[i]

        else if (l_LinksAndRights_[i] instanceof String) {

          rights = (String)l_LinksAndRights_[i];

                            // for testing and tracking
                        //System.out.println("call to GraphicView.drawArrowsAndRights() for String: ");
                        //System.out.println(" --- i: "+i+" rights: "+rights);

          if (rights.length() > GraphicNode.RIGHTS_PART+(GraphicNode.RIGHTS_PART/3)) {
            index = rights.indexOf(" ", GraphicNode.RIGHTS_PART);
            if (index != -1) rights = rights.substring(0, index)+" ...";
            else rights = rights.substring(0, GraphicNode.RIGHTS_PART)+" ...";
          } //if
          
          if (rights.startsWith("<>") ) rights = rights.substring(2); // for dashed_stroke line
          if (rights.startsWith("><") ) rights = rights.substring(2); 
          if (rights.startsWith(">>") ) rights = rights.substring(2); 

                            // for testing and tracking
                        //System.out.println("call to GraphicView.drawArrowsAndRights() for String: ");
                        //System.out.println(" --- i: "+i+" rights: "+rights);

          // for !O!, !ext-acl-C!...
          if (rights.indexOf("!") != rights.lastIndexOf("!"))
              _g2.setPaint(Color.blue);

          _g2.drawString(
            rights,
            (new Double(((Point2D.Float)l_currentArrowsAndRightsStarts_[i]).getX()))
              .intValue(),
            (new Double(((Point2D.Float)l_currentArrowsAndRightsStarts_[i]).getY()))
              .intValue());

          if (rights.indexOf("!") != rights.lastIndexOf("!"))
              _g2.setPaint(Color.BLACK);
        } // else if (l_LinksAndRights_

      } // for i
    } //if (l_LinksAndRights_ != null)

    rights = null;
  } // drawArrowsAndRights()


} // GraphicView

