//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;

import javax.swing.*;
import java.beans.PropertyChangeEvent;
import java.util.Iterator;
import java.util.List;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import ARoad0.gDMak.ActionOpenArchitecture;
import ARoad0.Gui1.*;
import ARoad0.Gui1.Desktop;
import ARoad0.ErrW.ProcessError;
import ARoad0.gWorkInterface.RightsMediator;
import ARoad0.gWork.ViewInBase_Facade;

import ARoad0.gBaseInterface.*;
import ARoad0.gBase.ViewInBaseImpl; // used only for init
import ARoad0.gBase.NameImpl; // used only for init
import ARoad0.gBase.AclEntryImpl;
import ARoad0.gBase.GroupIDImpl;

/**
 * This important class manages the GraphicView associated in the GUI to a ViewInBase,
 * and it is the main manager of a specialized thread to update the view and its rights 
 * at every relevant base changing. It is created by ViewModel.
 * Its structure is derived from ViewTreeBaseListenerImpl, but it is more complex.
 * It is responsible for listening the events:<P>
 * - fired after any relevant change in a ViewInBase object or in its ACS,
 * through the ViewInBase or directly,<P>
 * - fired by the RightsNetworkObserver associated to the view, and this includes
 * the listening of the specific events from any AcsAddon ACS or any of its objects.<P>
 * Each event source is registered only one time. EPRviewInBase and NoThanViewInBase
 * are handled for a dedicated listening.
 * The ViewInBase updatings activate all the changings in the GraphicView.
 * The opening and closing of views are managed only by ActionClose. This class
 * does not listen the relevant events, and it is finalized indirectly by ActionClose
 * through ViewModel. To avoid an useless processing of view rights, ActionSave calls
 * setListenEvents(false) before any saving. It restores the listening after
 * the saving. <P>
 * The thread management for the views is described hereinafter. <P>
 * Using the view threads, the first aim is to update quickly the access paths of a view
 * when a property has changed in a view node. This is why it is done outside
 * the event dispatch thread and the main thread, in a pool of worker threads
 * managed by Gui2.ViewModel. The worker thread works also at
 * the view creation or opening. In all cases, after the paths search,
 * the graphic view updating is done in the event dispatch thread.
 * All the worker threads have the same priority. <P>
 * The second aim of the thread management is to process the paths search only once,
 * while a database editing may generate several property changes for the same user action.
 * A TIMER_DELAY of 150 ms is set to not process immediatly the first property change.
 * After the TIMER_DELAY, even if several change events have been fired, the
 * access path search reads the current state of the base.
 * The maximal number of worker threads is set to 3 in Gui2.ViewModel. <P>
 * This class is running under the event dispatch thread (EVT), and the calling sequence is:<P>
 * - calls to propertyChange() one or several times, by one or several objects in gBase or gWork,
 *      and updates the relevant listeners in the base for each base change <P>
 * - propertyChange() updates the 2 booleans toChange_ and toUpdated_, starts a timer at the first
 *      firing, and restarts it at each firing, <P>
 * - the timer calls actionPerformed() when the TIMER_DELAY is expired, after 150 milliseconds, <P>
 * - actionPerformed() reads toChange_ and toUpdate_, and it calls changeFrameView(),
 *      changeTabView() or updateFrameViewRights(); then it stops the timer, <P>
 * - changeTabView() and sometimes actionPerformed() calls
 *      GraphicViewPropertyEditor.PropertyChangeEvent() if the view is in a JTabbedPane,
 *      which uses a worker thread to update the view, then run a task
 *      in the EVT to terminate the GUI updating through a call to repaint(), <P>
 * - changeFrameView() and updateFrameViewRights() use also a worker thread
 *      to update the view, then to run a task in the EVT; this task terminates
 *      the GUI updating by the way of a call to repaint(), <P>
 * - actionPerformed() resets the instance for the next firings on propertyChange().<P>
 * In all cases after the view creation, the worker thread is run to call
 * GraphicView.resetRights() or resetGraphicView(), and there is a call
 * to GraphicView.setWhyText() in the EVT. The task in the worker thread is then ended.
 * Since the worker thread is in a pool which is provided by ViewModel,
 * the same thread is able to process a new task for
 * another GraphicViewBaseListenerImpl instance, or for the same instance.
 * @version 0.7.0
 */
public class GraphicViewBaseListenerImpl implements BaseListener, ActionListener {
  /*
   * Sets 150 milliseconds for the delay of timer_
   */
  private static int TIMER_DELAY = 150;

  // non-static variables
  /*
       * To listen or not the events in propertyChange(). Used to stop the listening
       * before a saving of the relevant ACS, since the listening produces
       * a reset of the view rights through a process which may be heavy.
       * A saving implies a switch of the AclEntries and the firing of
       * numerous events.
       */
  private boolean                   listenEvents_ = true;
      /*
       * The view managed by this listener
       */
  private ViewInBase                view_ = ViewInBaseImpl.EMPTY_INSTANCE;
  private ImmutableName             viewName_ = NameImpl.EMPTY_INSTANCE;
      /*
       * the network observer associated to the view and created by ViewModel
       */
  private RightsMediator            rightsNetworkObserver_;

  /**
   * This is used by propertyChange to create a worker thread which resets the view.
   */
  private Thread worker_ = null;

  /*
   * This is used to cancel any near received firings from
   * gBase and gWork, and to create a worker thread only after the last firing and a delay.
   * This instance is its single listener.
   */
  private Timer timer_ = null;

  /*
   * True if the view nodes have changed. It is stronger than toUpdated_.
   */
  private boolean toChange_ = false;
  
  /*
   * True if the view nodes have not changed and rights may have changed.
   * It is weaker than toChange_.
   */
  private boolean toUpdate_ = false;

    // constructors

  /**
   * Public constructor.
  */
  public GraphicViewBaseListenerImpl() {
  }

 /**
 * With this constructor, the GraphicViewBaseListenerImpl is registered
 * as a listener for the view and a RightsNetworkObserver. There is a specific
 * listening if _source is a NoThanViewInBase.
 * @param _source is the view to manage
 */
  public GraphicViewBaseListenerImpl(ViewInBase _source) {
      view_ = _source;
      viewName_ = _source.getName();
      timer_ = new Timer(TIMER_DELAY, this);
      timer_.setRepeats(false);
      timer_.setActionCommand(null);
      
      ImmutableTarget       targ = null;
      ImmutableSource       sour = null;
      AclEntry[]            l_acl = null;
      ImmutableGroupID[]    l_grps = null;

      // note: if the view contains a VirtualFolder, it fires also 'OneBaseObjectInFolder'
      // but this event is then wrapped in a 'OnePropertyInView' event
      view_.addPropertyChangeListener("OneBaseObjectInFolder", this);
      
      if (view_ instanceof EPRViewInBase) {
        view_.addPropertyChangeListener("OnePropertyInView", this);
        
        for (Iterator ite = ((EPRViewInBase)view_).getEorM_Targets().values().iterator(); ite.hasNext(); ) {
              targ = (ImmutableTarget)ite.next();
              if (targ instanceof Resource) {
                  // to follow the ACLs
                  targ.addPropertyChangeListener("AclRights", this);

                  if (((Resource)targ).getL_aclEntry() != null) {
                      l_acl = ((Resource)targ).getL_aclEntry();
                      for (int i = 0; i < l_acl.length; i++) {
                          l_acl[i].addPropertyChangeListener("AclRights", this);
                      } // for i
                      
                      if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                              targ.getEorACS().getManageFlexibleConditionalAcls()) {
                          for (int i = 0; i < l_acl.length; i++) {
                              l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                              // For a conditional right, the ACL is deleted if a condition group is deleted.
                              // For a conditional ACL, registers the intermediate groups
                              l_grps = GroupIDImpl.
                                      getL_IntermediateGroupsOfAclConditionalGroups(
                                        (AclEntryImpl)l_acl[i]);
                              if (l_grps != null) {
                                  for (int j = 0; j < l_grps.length; j++) {
                                    l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                  } // for j
                              } // if (l_grps
                          } // for i
                      } // if (getManageRightDefinedConditionalAcls()
                  } // if
              } // if (targ instanceof Resource)

              if (targ instanceof VirtualFolder) { 
                  targ.addPropertyChangeListener("OneBaseObjectInFolder", this);//useful?
                  targ.addPropertyChangeListener("OneTargetTypedPrivilege", this);
                  targ.addPropertyChangeListener("OneTargetLinkedPrivilege", this);
                  targ.addPropertyChangeListener("Children", this);//useful?
              } // if (targ instanceof VirtualFolder)

              // to follow the sources of the linked privileges
              if (targ instanceof PrivilegeTarget) {
                  targ.addPropertyChangeListener("OneTargetLinkedPrivilege", this);

                  for (Object obj : ((PrivilegeTarget)targ).getEorL_TargetPrivilegeForLinks()) {
                    ((PrivilegeForLinks)obj).addPropertyChangeListener("OneLinkedPrivilegeSource", this);
                  } // for obj
              } // if (targ instanceof PrivilegeTarget)
        } // for ite 
        for (Iterator ite = ((EPRViewInBase)view_).getEorM_Sources().values().iterator();
          ite.hasNext(); ) {
            sour = (ImmutableSource)ite.next();
            // not for VirtualFolders, which are also Targets
            if (sour instanceof EligibleParty) { 
                sour.addPropertyChangeListener("AclConnectedResources", this);
                for (Iterator ita = ((EligibleParty)sour).getEorL_AclConnectedResources().
                        iterator(); ita.hasNext(); ) {
                  targ = (ImmutableTarget)ita.next();
                  l_acl = ((Resource)targ).getL_aclEntry();
                  if (l_acl == null || l_acl.length == 0) continue;

                  for (int i = 0; i < l_acl.length; i++) {
                    if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                          l_acl[i].addPropertyChangeListener("AclRights", this);
                    } // if l_acl[i].getEorEP() == sour
                  } // for

                  if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                          targ.getEorACS().getManageFlexibleConditionalAcls()) {
                      for (int i = 0; i < l_acl.length; i++) {
                          if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                              l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                              // For a conditional right, the ACL is deleted if a condition group is deleted.
                              // For a conditional ACL, registers the intermediate groups
                              l_grps = GroupIDImpl.
                                      getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                              if (l_grps != null) {
                                  for (int j = 0; j < l_grps.length; j++) {
                                    l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                  } // for j
                              } // if (l_grps
                          } // if l_acl[i].getEorEP() == sour
                      } // for i
                  } // if (getManageRightDefinedConditionalAcls()

                } // for ita
            } // // if (sour

            // to follow the targets of the linked privileges
            if (sour instanceof PrivilegeSource) {
              sour.addPropertyChangeListener("OneSourceLinkedPrivilege", this);

              for (Object obj : ((PrivilegeSource)sour).getEorL_SourcePrivilegeForLinks()) {
                ((PrivilegeForLinks)obj).addPropertyChangeListener("OneLinkedPrivilegeTarget", this);
              } // for obj
            } // if (sour instanceof PrivilegeSource)
        } // for ite
      } // if  EPRViewInBase
      
      if (view_ instanceof NoThanViewInBase) {
          view_.addPropertyChangeListener("FirstAccessTarget", this);
          //view_.addPropertyChangeListener("OneExcludedEP", this);
          view_.addPropertyChangeListener("OneMoreCriterion", this);
          view_.addPropertyChangeListener("OneLessCriterion", this);
          
          targ = ((NoThanViewInBase)view_).getEorFirstAccessTarget();
          if (targ != null) {
              if (targ instanceof Resource) {
                  // to follow the ACLs
                  targ.addPropertyChangeListener("AclRights", this);

                  if (((Resource)targ).getL_aclEntry() != null) {
                      l_acl = ((Resource)targ).getL_aclEntry();
                      for (int i = 0; i < l_acl.length; i++) {
                          l_acl[i].addPropertyChangeListener("AclRights", this);
                      } // for i

                      if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                              targ.getEorACS().getManageFlexibleConditionalAcls()) {
                          for (int i = 0; i < l_acl.length; i++) {
                                  l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                                  // For a conditional right, the ACL is deleted if a condition group is deleted.
                                  // For a conditional ACL, registers the intermediate groups
                                  l_grps = GroupIDImpl.
                                          getL_IntermediateGroupsOfAclConditionalGroups(
                                            (AclEntryImpl)l_acl[i]);
                                  if (l_grps != null) {
                                      for (int j = 0; j < l_grps.length; j++) {
                                        l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                      } // for j
                                  } // if (l_grps
                          } // for i
                      } // if (getManageRightDefinedConditionalAcls()

                  } // if
              } // if (targ instanceof Resource)

              if (targ instanceof VirtualFolder) {
                  targ.addPropertyChangeListener("OneTargetTypedPrivilege", this);
                  targ.addPropertyChangeListener("OneTargetLinkedPrivilege", this);
              } // if (targ instanceof Resource)

              // to follow the sources of the linked privileges
              if (targ instanceof PrivilegeTarget) {
                  targ.addPropertyChangeListener("OneTargetLinkedPrivilege", this);

                  for (Object obj : ((PrivilegeTarget)targ).getEorL_TargetPrivilegeForLinks()) {
                    ((PrivilegeForLinks)obj).addPropertyChangeListener(
                            "OneLinkedPrivilegeSource", this);
                  } // for obj
              } // if (targ instanceof PrivilegeTarget)
          } // if (res
          
          view_.addPropertyChangeListener("l_ObjectChangesInListeningPerimeter", this);
          // 'AclRights' events in the checking perimeter are listened through this event,
          // such as all the property changes in the perimeter which are handled by NoThanViewInBase
          view_.addPropertyChangeListener("OnePropertyInAListeningPerimeterObject", this); 
      } // if NoThanViewInBase

      rightsNetworkObserver_ = ViewModel.getViewRightsMediator(_source);
      rightsNetworkObserver_.addPropertyChangeListener(this);
      
                        // for testing and tracking
                      //System.out.println("call to the GraphicViewBaseListenerImpl constructor "+
                        //      "for the view "+view_.getName());
      sour = null;
      targ = null;
      l_acl = null;
      l_grps = null;
  } // GraphicViewBaseListenerImpl constructor

    //
    // -------------------  METHODS  ----------------------------
    //

 /**
 * Get the main event source associated to the instance
 * in its construction.
 * @return name of the main source for PropertyChangeEvents. May be null.
 */
  public ImmutableName getEorMainEventSource() {
    if (view_ != null) return viewName_;
    else return null;
  } // getEorMainEventSource()

 /**
 * To listen or not the events in propertyChange(). Used to stop the listening
 * before a saving of the relevant ACS, since the listening produces
 * a reset of the view rights through a process which may be heavy.
 * @param _listen true to listen the events in this instance, through propertyChange()
 * @return true if the previous value was different to this argument
 */
  public synchronized boolean setListenEvents(boolean _listen) {
    boolean previous = listenEvents_;
    listenEvents_ = _listen;

    if (previous && listenEvents_) return false;
    else if (!previous && !listenEvents_) return false;
    else return true;
  } // setListenEvents()

  /**
   * Receives the event from timer_ when the last event has been received
   * by propertyChange() at TIMER_DELAY, and calls an updating method for the view.
   * Calls GraphicViewPropertyEditor.propertyChange() with a null new value,
   * RightsNetworkObserver.setRightsRelatedBaseObjects(),
   * changeTabView(), updateFrameViewRights(), changeFrameView().
   * @param _evt contains a null command
   */
  public synchronized void actionPerformed(ActionEvent _evt) {
    // useful?
    timer_.stop();

                            // for testing and tracking
                        //System.out.println("in GraphicViewBaseListenerImpl, "+
                          //"call to actionPerformed() for the view: "+view_+
                          //" for the thread "+Thread.currentThread().getName()+
                          //" with the priority: "+Thread.currentThread().getPriority());

    // process the Architecture view that is in a main frame
    if (viewName_.equals(ActionOpenArchitecture.ARCHI_VIEW_NAME)) {
        if (toChange_) changeFrameView();
        // update the view
        else if (toUpdate_) updateFrameViewRights();
    }
    // process a full view that is in a main frame
    else if (!InternalFrameModel.withAllFullViews() &&
            !ViewInBase_Facade.isNoThanViewName(viewName_)) {
        if (toChange_) changeFrameView();
        // update the view
        else if (toUpdate_) {
            updateFrameViewRights();
        }
    }
    // process a NoThan view that is in a main frame
    else if (!InternalFrameModel.withAllNoThanViews() &&
            ViewInBase_Facade.isNoThanViewName(viewName_)) {
        if (toChange_) changeFrameView();
        // update the view
        else if (toUpdate_) updateFrameViewRights();
    // process the Full or NoThan view that is in a tabbedpane tab
    } else {
        // change the view
        if (toChange_) changeTabView();
        // update the view
        else if (toUpdate_) {
             // update the rights in the GraphicViewPropertyEditor with a null new value
            ViewModel.getGraphicViewPropertyEditor(viewName_.toString()).propertyChange(
                new PropertyChangeEvent(this, "", null, null));

            // update the RightsNetworkObserver
            ViewModel.getViewRightsMediator(view_).setRightsNetwork(
                    ViewModel.getGraphicViewOfView(view_.getName()).
                        getRightsRelatedBaseObjects());
        } // else if (toUpdate)
    } // else

    toChange_ = false;
    toUpdate_ = false;
  } // actionPerformed(ActionEvent e)

  /**
   * Receives all the events about the view property changes, adds and removes
   * this instance as property listener in the changing base objects, and calls
   * actionPerformed() after a delay to avoid multiple unuseful calls for the
   * same basic event.
   * No processing if the property listenEvents_ is false.
   * Listens 'OnePropertyInView' which is a wrapper for
   * any property change event from an EPRViewInBase ACSObject or an AclEntry.
   * Listens all the RightsNetworkObserver events for updating the view.
   * Listens 'OneBaseObjectInView' (and not 'ConnectedACS') from the ViewInBase,
   * 'OnePropertyInView' and 'AclConnectedResources' from the EPRViewInBase, 
   * and, from a NoThanViewInBase, 'FirstAccessTarget', 'OneMoreCriterion', 
   * 'OneLessCriterion', 'ObjectChangesInListeningPerimeter', 
   * 'OnePropertyInAListeningPerimeterObject'. Listens for the AclEntries,
   * 'AclConnectedResources' (for EligibleParty), 'AclRights', 'ConditionSourceGroups'
   * and 'AssociatedGroups'.
   * 'OnePropertyInView' is in fact an event wrapper for 
   * any property change event from an EPRViewInBase ACSObject,
   * so it is processed for each event where the declared source is a BaseObject.
   * 'OnePropertyInAListeningPerimeterObject' is also an event wrapper for
   * any property change event from the listening perimeter of a NoThanViewInBase.
   * Calls actionPerformed() after TIMER_DELAY milliseconds.
   * @param _evt PropertyChangeEvent
   */
  public synchronized void propertyChange(PropertyChangeEvent _evt){
    // variables
    boolean             toUpdate = false;
    boolean             toChange = false;
    Resource            res = null;
    EligibleParty       ep = null;
    AclEntry[]          l_acl = null;
    AclEntry[]          l_old = null;
    List<ImmutableSource> l_sour = null;
    ImmutableSource     sour = null;
    ImmutableGroupID[]  l_grps = null;
    ImmutableTarget     targ = null;
    
                    // for testing and tracking
                //if (view_ != null)
                  //  System.out.println("GraphicViewBaseListenerImpl.propertyChange() "+
                    //    "for the property: "+_evt.getPropertyName()+
                      //  " from the source "+_evt.getSource()+" for the view: "+view_.getName());
                //if (_evt.getPropertyName().equals("OneBaseObjectInFolder"))
                    //throw new InternalError();

    if (view_ == null || !listenEvents_) {

                     // for testing and tracking
                //System.out.println("GraphicViewBaseListenerImpl.propertyChange(), "+
                  //"for the property: "+_evt.getPropertyName()+
                  //" for the view: "+view_+ ": null view or no listening of events.");

        return;
    }

    // test the worker thread
    if (worker_ != null) {
                // for testing and tracking
                //System.out.println("call to "+
                  //"GraphicViewPropertyEditor.propertyChange() with a non-null worker, for the view: "+
                  //viewName_);
    } // if

        // start of the 'if else' tests
    // to listen the ACLs outside the listening perimeter
    if (_evt.getSource() instanceof ViewInBase &&
            (_evt.getPropertyName().equals("OneBaseObjectInFolder") || // from ViewInBase
            _evt.getPropertyName().equals("FirstAccessTarget") ) )   { // from NoThanViewInBase
        
              // to follow the ACLs if it is a new resource
              if (_evt.getNewValue() != null && _evt.getNewValue() instanceof Resource) {
                  res = (Resource)_evt.getNewValue();
                  res.addPropertyChangeListener("AclRights", this);
                  if (res.getL_aclEntry() != null) {
                      l_acl = res.getL_aclEntry();
                      for (int i = 0; i < l_acl.length; i++) {
                          l_acl[i].addPropertyChangeListener("AclRights", this);
                      } // for i

                      if (res.getEorACS().getManageRightDefinedConditionalAcls() ||
                              res.getEorACS().getManageFlexibleConditionalAcls()) {
                          for (int i = 0; i < l_acl.length; i++) {

                                  l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                                  // For a conditional right, the ACL is deleted if a condition group is deleted.
                                  // For a conditional ACL, registers the intermediate groups
                                  l_grps = GroupIDImpl.
                                          getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                                  if (l_grps != null) {
                                      for (int j = 0; j < l_grps.length; j++) {
                                        l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                      } // for j
                                  } // if (l_grps

                          } // for i
                      } // if (getManageRightDefinedConditionalAcls()

                  } // if (res.getL_aclEntry()  
              } // if (_evt.getNewValue()
              else if (_evt.getOldValue() != null && _evt.getOldValue() instanceof Resource) {
                  res = (Resource)_evt.getOldValue();
                  res.removePropertyChangeListener("AclRights", this);
                        //no action since the resource is deleted
                  //if (res.getL_aclEntry() != null) {
                    //  l_acl = res.getL_aclEntry();
                      //for (int i = 0; i < l_acl.length; i++) {
                        //  l_acl[i].removePropertyChangeListener("AclRights", this);
                      //} // for i

                      //if (res.getEorACS().getManageRightDefinedConditionalAcls() ||
                        //      res.getEorACS().getManageFlexibleConditionalAcls()) {
                          //for (int i = 0; i < l_acl.length; i++) {

                                  //l_acl[i].removePropertyChangeListener("ConditionSourceGroups", this);

                                  // For a conditional right, the ACL is deleted if a condition group is deleted.
                                  // For a conditional ACL, does not unregister the intermediate groups
                                  //l_grps = GroupIDImpl.
                                    //      getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                                  //if (l_grps != null) {
                                    //  for (int j = 0; j < l_grps.length; j++) {
                                      //  l_grps[j].removePropertyChangeListener("AssociatedGroups", this);
                                      //} // for j
                                  //} // if (l_grps

                          //} // for i
                      //} // if (getManageRightDefinedConditionalAcls()
                  //} // if (res.getL_aclEntry()
              } // if (_evt.getNewValue()

              // to follow the ACLs if it is an EligibleParty
              if (_evt.getNewValue() != null && _evt.getNewValue() instanceof EligibleParty) {
                    ep = (EligibleParty)_evt.getNewValue();
                    ep.addPropertyChangeListener("AclConnectedResources", this);
                    for (Iterator ita = ep.getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                      targ = (ImmutableTarget)ita.next();
                      l_acl = ((Resource)targ).getL_aclEntry();
                      if (l_acl == null || l_acl.length == 0) continue;

                      for (int i = 0; i < l_acl.length; i++) {
                        if (l_acl[i] != null && l_acl[i].getEorEP() == ep ) {
                              l_acl[i].addPropertyChangeListener("AclRights", this);
                        } // if
                      } // for

                      if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                              targ.getEorACS().getManageFlexibleConditionalAcls()) {
                          for (int i = 0; i < l_acl.length; i++) {

                                  l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                                  // For a conditional right, the ACL is deleted if a condition group is deleted.
                                  // For a conditional ACL, registers the intermediate groups
                                  l_grps = GroupIDImpl.
                                          getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                                  if (l_grps != null) {
                                      for (int j = 0; j < l_grps.length; j++) {
                                        l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                      } // for j
                                  } // if (l_grps

                          } // for i
                      } // if (getManageRightDefinedConditionalAcls()

                    } // for ita
              } // if (_evt.getNewValue()
              else if (_evt.getOldValue() != null && _evt.getOldValue() instanceof EligibleParty) {
                    ep = (EligibleParty)_evt.getOldValue();
                    ep.removePropertyChangeListener("AclConnectedResources", this);
                    for (Iterator ita = ep.getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                      targ = (ImmutableTarget)ita.next();
                      l_acl = ((Resource)targ).getL_aclEntry();
                      if (l_acl == null || l_acl.length == 0) continue;

                            // the ACLs from the right EP are deleted, so no unregistering
                      //for (int i = 0; i < l_acl.length; i++) {
                        //if (l_acl[i] != null && l_acl[i].getEorEP() == ep ) {
                          //    l_acl[i].removePropertyChangeListener("AclRights", this);
                        //} // if
                      //} // for i

                      //if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                            //  targ.getEorACS().getManageFlexibleConditionalAcls()) {
                          //for (int i = 0; i < l_acl.length; i++) {

                                //  l_acl[i].removePropertyChangeListener("ConditionSourceGroups", this);

                                  // For a conditional right, the ACL is deleted if a condition group is deleted.
                                  // For a conditional ACL, does not register the intermediate groups
                                //  l_grps = GroupIDImpl.
                                  //        getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                                //  if (l_grps != null) {
                                //      for (int j = 0; j < l_grps.length; j++) {
                              //          l_grps[j].removePropertyChangeListener("AssociatedGroups", this);
                            //          } // for j
                          //        } // if (l_grps

                        //  } // for i
                      //} // if (getManageRightDefinedConditionalAcls()

                    } // for ita
              } // if (_evt.getNewValue()
              
              toChange = true;
    } // if (_evt.getSource() instanceof ViewInBase

    // to listen the ACLs and all changes in the NoThanViewInBase listening perimeter
    else if (_evt.getSource() instanceof ViewInBase &&
            _evt.getPropertyName().equals("l_ObjectChangesInListeningPerimeter") )  { 

              // to follow the ACLs if it is an EligibleParty
              if (!((List)_evt.getNewValue()).isEmpty()) {
                    l_sour = (List)_evt.getNewValue();
                    for (Iterator iti = l_sour.iterator(); iti.hasNext(); ) {
                        sour = (ImmutableSource)iti.next();
                        sour.addPropertyChangeListener("AclConnectedResources", this);
                        for (Iterator ita = ((EligibleParty)sour).
                                getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                          targ = (ImmutableTarget)ita.next();
                          l_acl = ((Resource)targ).getL_aclEntry();
                          if (l_acl == null || l_acl.length == 0) continue;

                          for (int i = 0; i < l_acl.length; i++) {
                            if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                                  l_acl[i].addPropertyChangeListener("AclRights", this);
                            } // if
                          } // for i

                          if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                                  targ.getEorACS().getManageFlexibleConditionalAcls()) {
                              for (int i = 0; i < l_acl.length; i++) {
                                      l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                                      // For a conditional right, the ACL is deleted if a condition group is deleted.
                                      // For a conditional ACL, registers the intermediate groups
                                      l_grps = GroupIDImpl.
                                              getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                                      if (l_grps != null) {
                                          for (int j = 0; j < l_grps.length; j++) {
                                            l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                          } // for j
                                      } // if (l_grps
                              } // for i
                          } // if (getManageRightDefinedConditionalAcls()

                        } // for ita
                    } // for iti
              } // if (_evt.getNewValue()
              else if (!((List)_evt.getOldValue()).isEmpty()) {
                    l_sour = (List)_evt.getOldValue();
                    for (Iterator iti = l_sour.iterator(); iti.hasNext(); ) {
                        sour = (ImmutableSource)iti.next();
                        sour.removePropertyChangeListener("AclConnectedResources", this);
                        for (Iterator ita = ((EligibleParty)sour).
                                getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                          targ = (ImmutableTarget)ita.next();
                          l_acl = ((Resource)targ).getL_aclEntry();
                          if (l_acl == null || l_acl.length == 0) continue;

                          for (int i = 0; i < l_acl.length; i++) {
                            if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                                  l_acl[i].removePropertyChangeListener("AclRights", this);
                            } // if
                          } // for i

                          if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                                  targ.getEorACS().getManageFlexibleConditionalAcls()) {
                              for (int i = 0; i < l_acl.length; i++) {

                                      l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                                      // For a conditional right, the ACL is deleted if a condition group is deleted.
                                      // For a conditional ACL, registers the intermediate groups
                                      l_grps = GroupIDImpl.
                                              getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                                      if (l_grps != null) {
                                          for (int j = 0; j < l_grps.length; j++) {
                                            l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                          } // for j
                                      } // if (l_grps

                              } // for i
                          } // if (getManageRightDefinedConditionalAcls()

                        } // for ita
                    } // for iti
              } // if (_evt.getNewValue()
              
              toChange = true;
    } // if (_evt.getSource() instanceof ViewInBase
 
    // to listen the property changes in the NoThanViewInBase listening perimeter
    else if (_evt.getSource() instanceof ViewInBase &&
            _evt.getPropertyName().equals("OnePropertyInAListeningPerimeterObject") )  { 
        toUpdate = true;
    }
 
    // to listen the property changes in a new AclEntry, only 'AclRights' from
    // resource is processed. An 'AclRights' event from AclEntryImpl is not handled.
    else if (_evt.getSource() instanceof Resource &&
            _evt.getPropertyName().equals("AclRights") )  {
        
        boolean newACL = false;
        l_old = (AclEntry[])_evt.getOldValue();
        l_acl = (AclEntry[])_evt.getNewValue();
        if (l_acl.length > l_old.length) newACL = true;

        // the new ACL is at the end of the list, if there is a new one
        if (newACL) {

                                    // for testing and tracking
                                //System.out.println("in GraphicViewBaseListenerImpl, "+
                                  //"call to propertyChange() for the property: "+
                                  //_evt.getPropertyName()+
                                  //", to listen the new ACL: "+l_acl[l_acl.length - 1]);

              l_acl[l_acl.length - 1].addPropertyChangeListener("AclRights", this);

              if (l_acl[l_acl.length - 1].getEorACS().getManageRightDefinedConditionalAcls() ||
                      l_acl[l_acl.length - 1].getEorACS().getManageFlexibleConditionalAcls()) {
                  l_acl[l_acl.length - 1].addPropertyChangeListener("ConditionSourceGroups", this);

                  // For a conditional right, the ACL is deleted if a condition group is deleted.
                  // For a conditional ACL, registers the intermediate groups
                  l_grps = GroupIDImpl.
                          getL_IntermediateGroupsOfAclConditionalGroups(
                                (AclEntryImpl)l_acl[l_acl.length - 1]);
                  if (l_grps != null) {
                      for (int j = 0; j < l_grps.length; j++) {
                        l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                      } // for j
                  } // if (l_grps
              } // if (getManageRightDefinedConditionalAcls()

        } // if (newACL)

        toUpdate = true;
    } // else if (_evt.getSource() instanceof Resource
                
    // for 'OnePropertyInView', which is in fact a wrapper for 
    // any property change event from an EPRViewInBase ACSObject,
    // for 'AclRights' from AclEntry, 'AclConnectedResources' among other events
    // from an EPRViewInBase.
    // for NoThanViewInBase, 'OneMoreCriterion', 'OneLessCriterion' events.
    // no filtering of a 'OnePropertyInView' event which is the second-order event
    // for the same basic operation, like for instance when a GroupIDMember relation changes,
    // and it is fired by both the group and the groupmember.
    else if (_evt.getSource() instanceof BaseObject ) {
        
                        // for testing and tracking
                    //System.out.println("in GraphicViewBaseListenerImpl.propertyChange(), "+
                      //  "update from a BaseObject with the event: "+_evt+
                        //" for the view: "+view_.getName());
                
          toUpdate = true;
    } // if (_evt.getSource() == BaseManagerImpl

    // for any change in the rights network of the view
    else if (_evt.getSource() == rightsNetworkObserver_ ) {
        
                        // for testing and tracking
                    //System.out.println("in GraphicViewBaseListenerImpl.propertyChange(), "+
                      //  "update from the RightsNetworkObserver "+_evt+
                        //" for the view: "+view_.getName());
                
          toUpdate = true;
    } // change in the rights network

    if (!toChange && !toUpdate) {
        res = null;
        ep = null;
        l_acl = null;
        l_sour = null;
        sour = null;

                    //System.out.println("end of "+
                      //  "GraphicViewBaseListener.propertyChange() with toChange_: "+
                        //toChange_+" and toUpdate_: "+toUpdate_);

        return;
    } // if

    if (toChange) {
        toChange_ = true;
        toUpdate_ = false;
    } else if (toUpdate && !toChange_) {
        toChange_ = false;
        toUpdate_ = true;
    }

    // timer_ calls actionPerformed() after TIMER_DELAY milliseconds
    timer_.restart();

                // for testing and tracking
                //if (worker_ != null)
                    //System.out.println("end of "+
                      //  "GraphicViewBaseListener.propertyChange() with toChange_: "+
                        //toChange_+" and toUpdate_: "+toUpdate_);

    
    res = null;
    l_acl = null;
    ep = null;
    l_sour = null;
    sour = null;
  } // propertyChange()

  /**
   * Resets the graphic view which has a different set of nodes, and
   * the 'See why' frame if it is open, but not the explorer.
   * Updates the RightsMediator.
   * Runs a worker thread to call GraphicView.resetRights(), then calls
   * GraphicView.setWhyText() in the Event Dispatch Thread.
   * This method calls the worker thread pool returned by ViewModel.getViewsThreadsExecutor()
   * which processes the calls in their calling order, as asynchronous tasks.
   * Called by propertyChange().
   */
  private void changeFrameView() {

    // variables
    JInternalFrame  mainFrame;
    // get the GraphicView
    final GraphicView gview = (GraphicView) InternalFrameModel.
        getJComponentsOfBaseObjects().get(viewName_);
    // get the view main frame
    mainFrame = (JInternalFrame)InternalFrameModel.
      getMainFramesOfBaseObjects().get(viewName_);

                                                // for testing and tracking
                                        //System.out.println("in GraphicViewBaseListenerImpl.changeFrameView(), "+
                                          //  "viewName_: "+viewName_+"; mainFrame title: "+mainFrame.getTitle() );

    if (mainFrame == null) return;

    ViewModel.getViewsThreadsExecutor().execute(new Runnable() {
        @Override
        public void run() {
            // call in a worker Thread

                                    // for testing and tracking
                                    //System.out.println("call to "+
                                    //"GraphicViewBaseListenerImpl.changeFrameView() for "+
                                      //  " calling gview.resetGraphicView() for "+
                                        //gview.getEorDataModel().getName());

            try {
                // no synchronization on property_ to let the EDT handles this object.
                //synchronized (gview) {
                    // update the node and link algorithms
                    ViewModel.changeNonCoreAlgorithms(gview.viewInBase_);

                    // update the view
                    gview.resetGraphicView();
                //}

            } catch (ProcessError p) {
                String why = null;
                Throwable cause = p.getCause();
                if (cause != null) {
                    why = cause.getMessage();
                } else {
                    why = p.getMessage();
                }
                System.err.println("In GraphicViewBaseListenerImpl.changeFrameView(), "+
                        "error searching access paths in "+gview.getEorDataModel().getName()+
                        ", and the cause is: " + why);
                p.printStackTrace();
            } catch (InterruptedException ignore) {}

            // no InterruptedException handling there
            SwingUtilities.invokeLater(new Runnable(){
                public void run(){
                        // in the Event Dispatch Thread

                            // for testing and tracking
                            //System.out.println("In GraphicViewBaseListenerImpl.changeFrameView(), "+
                              //      "call to gview.removeWaitingMessage() then repaint() for "+
                                //    gview.getEorDataModel().getName());

                    gview.removeWaitingMessage();
                    gview.repaint();

                    boolean open = false;

                    // test if the view justification frame does already exist
                    if (viewName_.noOfComponents() > 1) {
                        JInternalFrame iframe = null;
                        for (int i = 0; i < Desktop.getCurrentFrames().length; i++) {
                            iframe = Desktop.getCurrentFrames()[i];
                            if (iframe.getTitle().equals(viewName_.toString()+ " 'See why' text")) {
                                open = true;
                                break;
                            } // if
                        } // for
                        // if it exists, rebuild up the 'See why' frame at its current position
                        if (open) {
                            InternalFrame.updateTextFrame(gview.setWhyText(), iframe);
                        }
                        iframe = null;
                    } // if (viewName_

                    // this method keeps this GraphicViewBaseListenerImpl instance
                    ViewModel.changeRightsNetworkObserver(view_, gview);

                } // run
            }); // invokeLater()

                    // for testing and tracking
                    //System.out.println("In GraphicViewBaseListenerImpl.changeFrameView(), "+
                      //      "end of run() for "+
                        //    gview.getEorDataModel().getName());

        } // run()

    }); // new Thread

    mainFrame = null;
  }  // changeFrameView()

  /**
   * Resets the rights without node changing.
   * When the 'all views' frame is closed and a right has been changed,
   * reset the rights in the main frame and reset the 'See why' frame.
   * Do not change the explorer.
   * Reset the 'See why' frame, if it is open, to its initial position.
   * Called by the RightsNetworkObserver.
   * Used also for the 'Architecture' view.
   * Run a worker thread to call GraphicView.resetRights(), then call
   * GraphicView.setWhyText() in the Event Dispatch Thread.
   * This method calls the worker thread pool returned by ViewModel.getViewsThreadsExecutor()
   * which processes the calls in their calling order, as asynchronous tasks.
   * Called by propertyChange().
   */
  private void updateFrameViewRights() {
    // variables
    JInternalFrame  mainFrame = null;
    // get the GraphicView
    final GraphicView gview = (GraphicView) InternalFrameModel.
            getJComponentsOfBaseObjects().get(viewName_);

    // get the view main frame
    mainFrame = (JInternalFrame)InternalFrameModel.
      getMainFramesOfBaseObjects().get(viewName_);

    if (mainFrame == null) return;

                                    //System.out.println("call to "+
                                    //"GraphicViewBaseListenerImpl.updateFrameViewRights() "+
                                      //      "for the view "+viewName_+
                                        //    " before the worker thread "+
                                          //  "for GraphicView: "+
                                            //gview.getEorDataModel().getNickName());

    gview.addWaitingMessage();

    ViewModel.getViewsThreadsExecutor().execute(new Runnable() {
        @Override
        public void run() {
            // call in a worker thread

                                        // for testing and tracking
                                    //System.out.println("call to "+
                                      //  "GraphicViewBaseListenerImpl.updateFrameViewRights() "+
                                        //"for the thread "+Thread.currentThread().getName()+
                                        //" with the priority: "+Thread.currentThread().getPriority()+
                                        //" for GraphicView: "+
                                        //gview.getEorDataModel().getNickName());

            try {
                // no synchronization on property_ to let the EDT handles this object.
                //synchronized (gview) {
                    // update the view
                    gview.resetRights();
                //}

            } catch (ProcessError p) {
                String why = null;
                Throwable cause = p.getCause();
                if (cause != null) {
                    why = cause.getMessage();
                } else {
                    why = p.getMessage();
                }
                System.err.println("In GraphicViewPropertyEditor.updateFrameViewRights(), "+
                        "error in searching access paths for "+gview.getEorDataModel().getName()+
                        "; the cause is: " + why);
                p.printStackTrace();
            } catch (InterruptedException ignore) {}

            SwingUtilities.invokeLater(new Runnable(){
                public void run(){
                        // in the Event Dispatch Thread
                    boolean open = false;

                    gview.removeWaitingMessage();
                    gview.repaint();

                    // test if the view justification frame does already exist
                    if (viewName_.noOfComponents() > 1) {
                        JInternalFrame iframe = null;
                        for (int i = 0; i < Desktop.getCurrentFrames().length; i++) {
                            iframe = Desktop.getCurrentFrames()[i];
                            if (iframe.getTitle().equals(viewName_.toString()+ " 'See why' text")) {
                                open = true;
                                break;
                            } // if
                        } // for
                        // if it exists, rebuild up the 'See why' frame at its current position
                        if (open) {
                            InternalFrame.updateTextFrame(gview.setWhyText(), iframe);
                        }
                        iframe = null;
                    } // if (viewName_

                    // this method keeps this GraphicViewBaseListenerImpl instance
                    ViewModel.changeRightsNetworkObserver(view_, gview);

                } // run
            }); // invokeLater()

                    // for testing and tracking
                    //System.out.println("In GraphicViewPropertyEditor.changeFrameView(), "+
                      //      "end of run() for "+
                        //    gview.getEorDataModel().getName());

        } // run()

    }); // new Thread
    
    mainFrame = null;
  }  // updateFrameViewRights()  


  /**
   * Updates the RightsMediator and the view tab (GraphicView and 'See why').
   * Does not change the explorer.
   * On the contrary of changeFrameView(), no graphic view creation.
   * Calls GraphicViewPropertyEditor.PropertyChangeEvent() which uses a worker thread
   * to update the view
   * Called by propertyChange().
   */
  private void changeTabView() {

        // get the GraphicView to finalize it
        GraphicView gview = (GraphicView) ViewModel.getGraphicViewOfView(viewName_);

        // update the node and link algorithms of the views
        ViewModel.changeNonCoreAlgorithms(view_);

        // update the GraphicViewPropertyEditor with a non-null gview
        // to have a call to gview.resetGraphicView()
        ViewModel.getGraphicViewPropertyEditor(viewName_.toString()).
                propertyChange(
                    new PropertyChangeEvent(this, "", null, gview));
        
        // this method keeps this GraphicViewBaseListenerImpl instance
        // since in fact, gview is the same.
        // update the RelatedBaseObjects of the RightsNetworkObserver
        ViewModel.changeRightsNetworkObserver(view_, gview);

        gview = null;
  }  // changeTabView()  
  
 /**
 * Unregisters the instance as a listener for all registered sources.
 * Sets the variables to null. Called by ViewModel.removeViewMainObservers().
 */
  public void finalizeForUser() {
    ImmutableTarget targ = null;
    ImmutableSource sour = null;
    AclEntry[]      l_acl = null;
    ImmutableGroupID[] l_grps = null;
     
                        // for testing and tracking
                        //System.out.println("call of GraphicViewBaseListenerImpl."+
                          //"finalizeForUser() for "+view_.getName());


    view_.removePropertyChangeListener("OneBaseObjectInFolder", this);
    
    if (view_ instanceof EPRViewInBase) {
        view_.removePropertyChangeListener("OnePropertyInView", this);
        
        for (Iterator ite = ((EPRViewInBase)view_).getEorM_Targets().values().iterator();
          ite.hasNext(); ) {
              targ = (ImmutableTarget)ite.next();
              // targ is always a Resource, and the ACLs are followed
              if (targ instanceof Resource) {
                  // to follow the ACLs
                  targ.removePropertyChangeListener("AclRights", this);

                  if (((Resource)targ).getL_aclEntry() != null) {
                      l_acl = ((Resource)targ).getL_aclEntry();
                      for (int i = 0; i < l_acl.length; i++) {
                          l_acl[i].removePropertyChangeListener("ConditionSourceGroups", this);

                          // For a conditional right, the ACL is deleted if a condition group is deleted.
                          // For a conditional ACL, registers the intermediate groups
                          l_grps = GroupIDImpl.
                                  getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                          if (l_grps != null) {
                              for (int j = 0; j < l_grps.length; j++) {
                                l_grps[j].removePropertyChangeListener("AssociatedGroups", this);
                              } // for j
                          } // if (l_grps
                      } // for i
                  } // if
              } // if (targ instanceof Resource)

              if (targ instanceof VirtualFolder) {
                  targ.removePropertyChangeListener("OneBaseObjectInFolder", this);
                  targ.removePropertyChangeListener("OneTargetTypedPrivilege", this);
                  targ.removePropertyChangeListener("OneTargetLinkedPrivilege", this);
                  targ.removePropertyChangeListener("Children", this);                  
              } // if

              // to not follow the sources of the linked privileges
              if (targ instanceof PrivilegeTarget) {
                  targ.removePropertyChangeListener("OneTargetLinkedPrivilege", this);

                  for (Object obj : ((PrivilegeTarget)targ).getEorL_TargetPrivilegeForLinks()) {
                    ((PrivilegeForLinks)obj).removePropertyChangeListener("OneLinkedPrivilegeSource", this);
                  } // for obj
              } // if (targ instanceof PrivilegeTarget)
        } // for ite
        for (Iterator ite = ((EPRViewInBase)view_).getEorM_Sources().values().iterator();
          ite.hasNext(); ) {
            sour = (ImmutableSource)ite.next();
            if (sour instanceof EligibleParty) {
                sour.removePropertyChangeListener("AclConnectedResources", this);
                for (Iterator ita = ((EligibleParty)sour).
                        getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                  targ = (ImmutableTarget)ita.next();
                  l_acl = ((Resource)targ).getL_aclEntry();
                  if (l_acl == null || l_acl.length == 0) continue;

                  for (int i = 0; i < l_acl.length; i++) {
                    if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                          l_acl[i].removePropertyChangeListener("AclRights", this);
                    } // if l_acl[i].getEorEP() == sour
                  } // for i

                  if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                          targ.getEorACS().getManageFlexibleConditionalAcls()) {
                      for (int i = 0; i < l_acl.length; i++) {
                          if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                              l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                              // For a conditional right, the ACL is deleted if a condition group is deleted.
                              // For a conditional ACL, registers the intermediate groups
                              l_grps = GroupIDImpl.
                                      getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                              if (l_grps != null) {
                                  for (int j = 0; j < l_grps.length; j++) {
                                    l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                  } // for j
                              } // if (l_grps
                          } // if l_acl[i].getEorEP() == sour
                      } // for i
                  } // if (getManageRightDefinedConditionalAcls()

                } // for ita
            } // if sour EligibleParty

            // to not follow the targets of the linked privileges
            if (sour instanceof PrivilegeSource) {
              sour.removePropertyChangeListener("OneSourceLinkedPrivilege", this);

              for (Object obj : ((PrivilegeSource)sour).getEorL_SourcePrivilegeForLinks()) {
                ((PrivilegeForLinks)obj).removePropertyChangeListener("OneLinkedPrivilegeTarget", this);
              } // for obj
            } // if (sour instanceof PrivilegeSource)
        } // for ite    
    } // if EPRViewInBase
    
    if (view_ instanceof NoThanViewInBase) {
      view_.removePropertyChangeListener("FirstAccessTarget", this);
      //view_.removePropertyChangeListener("OneExcludedEP", this);
      view_.removePropertyChangeListener("OneMoreCriterion", this);
      view_.removePropertyChangeListener("OneLessCriterion", this); 
          
      targ = ((NoThanViewInBase)view_).getEorFirstAccessTarget();
      if (targ != null) {
          if (targ instanceof Resource) {
              // to not follow the ACLs
              targ.removePropertyChangeListener("AclRights", this);

              if (((Resource)targ).getL_aclEntry() != null) {
                  l_acl = ((Resource)targ).getL_aclEntry();
                  for (int i = 0; i < l_acl.length; i++) {
                      l_acl[i].removePropertyChangeListener("AclRights", this);
                  } // for i

                  if (targ.getEorACS().getManageRightDefinedConditionalAcls() ||
                          targ.getEorACS().getManageFlexibleConditionalAcls()) {
                      for (int i = 0; i < l_acl.length; i++) {

                              l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                              // For a conditional right, the ACL is deleted if a condition group is deleted.
                              // For a conditional ACL, registers the intermediate groups
                              l_grps = GroupIDImpl.
                                      getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)l_acl[i]);
                              if (l_grps != null) {
                                  for (int j = 0; j < l_grps.length; j++) {
                                    l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                  } // for j
                              } // if (l_grps

                      } // for i
                  } // if (getManageRightDefinedConditionalAcls()

              } // if
          } // if (targ instanceof Resource)

          if (targ instanceof VirtualFolder) {
              targ.removePropertyChangeListener("OneTargetTypedPrivilege", this);
              targ.removePropertyChangeListener("OneTargetLinkedPrivilege", this);
          } // if (targ instanceof Resource)

          // to not follow the sources of the linked privileges
          if (targ instanceof PrivilegeTarget) {
              targ.removePropertyChangeListener("OneTargetLinkedPrivilege", this);

              for (Object obj : ((PrivilegeTarget)targ).getEorL_TargetPrivilegeForLinks()) {
                ((PrivilegeForLinks)obj).removePropertyChangeListener("OneLinkedPrivilegeSource", this);
              } // for obj
          } // if (targ instanceof PrivilegeTarget)
      } // if (targ != null)
      
      view_.removePropertyChangeListener("l_ObjectChangesInListeningPerimeter", this);
      view_.removePropertyChangeListener("OnePropertyInAListeningPerimeterObject", this); 
    } // if NoThanViewInBase

    rightsNetworkObserver_.removePropertyChangeListener(this);
    rightsNetworkObserver_ = null;
    
            // for testing and tracking
            //System.out.println("end of GraphicViewBaseListenerImpl."+
              //"finalizeForUser() for "+view_.getName());
    
    view_ = null;
    timer_ = null;
    targ = null;
    sour = null;
    l_acl = null;
  }// finalizeForUser()

 /**
 * HashCode of the instance. It is used by TreeManager to remove
 * a TreeBaseListener, and by BaseManagerImpl to fire events to this instance.
 * Don't use the event source list.
 * @return int
 */
/*  public int hashCode(){
        int result = 0;

        result = PRIME*result + view_.hashCode();
        result = PRIME*result + treeModel_.getRoot().hashCode();

        if (root_.getParent() != null) result =
            PRIME*result + ((DefaultMutableTreeNode)root_.getParent()).
                getUserObject().hashCode();

        return result;
  } // hashCode()
*/

  /**
   * @param _obj: object to compare
   * @return true if the two objects are identical instances
   * of identical classes that implemente ViewTreeBaseListener
   */
/*   public synchronized boolean equals(Object _obj) {
    if (this == _obj)
        return true;
    if (_obj == null)
        return false;
    if (_obj.getClass() != getClass())
        return false;

    GraphicViewBaseListenerImpl obj = (GraphicViewBaseListenerImpl) _obj;

    if (view_ == null || obj.getEorMainEventSource() == null) {
        if (view_!= null || obj.getEorMainEventSource() != null)
          return false;  // if
    } // if
    else {
      if (!viewName_.equals( obj.getEorMainEventSource() )) return false; }

    if (treeModel_ == null || obj.getEorTreeModel() == null) {
        if (treeModel_ != null || obj.getEorTreeModel() != null)
          return false;  // if
    } // if
    else {
      if (!treeModel_.equals( obj.getEorTreeModel() )) return false; }

    if (root_ == null || obj.getEorRoot() == null) {
        if (root_ != null || obj.getEorRoot() != null)
          return false;  // if
    } // if
    else {
      if (!root_.equals( obj.getEorRoot() )) return false; }

    return true;
  } //equals()
*/

 /**
 * @return String
 */
  public String toString(){
        StringBuilder sb = new StringBuilder();
        sb.append("  GRAPHICVIEW_BASELISTENERIMPL.TOSTRING():");

        sb.append(" view name = ");
        sb.append(viewName_);

        sb.append(" hashCode() = ");
        sb.append( Integer.toString(hashCode()) );

        return " " + sb;
  } // toString()

} // GraphicViewBaseListenerImpl