//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;


import javax.swing.*;
import javax.swing.tree.*;
import java.beans.*;
import java.util.List;
import java.util.ArrayList;
import javax.swing.event.TreeSelectionEvent;

import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gDMak.ActionOpenGlobalStructure;

import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.NameImpl;

/**
 * This class is responsible for listening the events fired by BaseManagerImpl
 * for updating the JTree which is associated to this TreeBaseListener,
 * managing the tree of the IS Structure.
 * Handles 'OneInformationSystem', 'OneInformationSystemNode',
 * 'OneInformationSystemNodeOpening' change properties.
 * The single source is the single instance of gBase.BaseManagerImpl.
 * The closing and deleting operations are not managed.
 * These operations are handled in Gui1/Gui2, in the gDMak classes
 * ActionClose and ActionOpen, and by TreeManager in Gui2.
 * @version 0.7.0
 */
public class ISTreeBaseListenerImpl implements TreeBaseListener {
    // variables
  private static final int          PRIME = 1000003;
  private static final int          INITIAL_CAPACITY = 5; // the default value: 101
      /**
      * contains only BaseManagerImpl firing PropertyChangeEvents
      */
  private List                      l_eventSources_ = new ArrayList(INITIAL_CAPACITY);
      /**
       * the JTree
       */
  private JTree                     tree_;
      /**
       * tree model of the JTree
       */
  private DefaultTreeModel          treeModel_;
      /**
       * root node of the JTree, which is not an IS node
       */
  private DefaultMutableTreeNode    root_;
      /**
       * name of the source for TreeManager
       */
  private ImmutableName             sourceName_ = new NameImpl("<BaseManager>");
      /*
       * To listen or not the events in propertyChange(). Not used here.
       */
  private boolean                   listenEvents_ = true;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ISTreeBaseListenerImpl EMPTY_INSTANCE = new ISTreeBaseListenerImpl();

    // constructors

  /**
   * Public constructor.
  */
  public ISTreeBaseListenerImpl() {
    tree_ = new JTree();
    root_ = new DefaultMutableTreeNode();
    treeModel_ = new DefaultTreeModel(root_);
  }

 /**
 * With this constructor, the ISTreeBaseListenerImpl is registered
 * as a listener for BaseManagerImpl firing PropertyChangeEvents,
 * and is associated to the global structure JTree.
 * Called by Gui2.ISTree for the IS structure frame.
 * @param   _tree associated to the instance
 * @param   _root associated to _source in _tree
 * @exception ProcessError if _source is null
 */
  public ISTreeBaseListenerImpl(JTree _tree, MutableTreeNode _root) {

      //System.out.println("ISTreeBaseListenerImpl constructor ");

      tree_ = _tree;
      treeModel_ = (DefaultTreeModel) _tree.getModel();
      root_ = (DefaultMutableTreeNode)_root;
      l_eventSources_.add(BaseManagerImpl.getInstance());

      BaseManagerImpl.getInstance().addPropertyChangeListener(
              "OneInformationSystem", this);
      BaseManagerImpl.getInstance().addPropertyChangeListener(
              "OneInformationSystemNode", this);
      BaseManagerImpl.getInstance().addPropertyChangeListener(
              "OneInformationSystemNodeOpening", this);
  } // ISTreeBaseListenerImpl

    //
    // --------------- METHODS ---------------------
    //

 /**
 * Get the main event source associated to the instance
 * in its construction, which is sourceName_.
 * @return name of the main source for PropertyChangeEvents. May be null.
 */
  public ImmutableName getEorMainEventSource() {
    return sourceName_;
  } // getEorMainEventSource()

 /**
 * To listen or not the events in propertyChange(). No use.
 * @param _listen true to listen the events in this instance, through propertyChange()
 * @return true if the previous value was different to this argument
 */
  public synchronized boolean setListenEvents(boolean _listen) {
    boolean previous = listenEvents_;
    listenEvents_ = _listen;

    if (previous && listenEvents_) return false;
    else if (!previous && !listenEvents_) return false;
    else return true;
  } // setListenEvents()

 /**
 * Get the TreeModel associated to the ACS tree instance
 * in its construction.
 * @return DefaultTreeModel associated to the instance and to a JTree
 */
  public DefaultTreeModel getEorTreeModel() {
    return treeModel_;
  } // getEorTreeModel()

 /**
 * Get the TreeModel root associated to the instance
 * in its construction.
 * @return MutableTreeNode associated to the instance in the TreeModel
 */
  public MutableTreeNode getEorRoot() {
    return root_;
  } // getEorRoot()

  /**
   * Receives the source events about property changes on 'OneInformationSystem',
   * 'OneInformationSystemNode', 'OneInformationSystemNodeOpening'.
   * Throw InternalError if both the old and the new values are non null,
   * or if the two are null. Calls the appropriate private method.
   * @param _evt PropertyChangeEvent
   */
  public synchronized void propertyChange(PropertyChangeEvent _evt){

                // for testing and tracking
                //System.out.println("ISTreeBaseListenerImpl: "+
                  //  "call to propertyChange() from the source "+_evt.getSource().toString()+
                    //" for the property "+_evt.getPropertyName() );

    // to avoid the processing if this instance is empty or has been finalized
    if (sourceName_ == null) return;

    // the two values may be non null when there are arrays, collections or maps
    if (_evt.getOldValue() == null && _evt.getNewValue() == null) {
      throw new InternalError("ISTreeBaseListenerImpl: "+
        "call to propertyChange() - both old and new values are null.");
    }

    // for 'OneInformationSystem'
    if (_evt.getPropertyName().equals("OneInformationSystem") ) {
          changeOneInformationSystem(_evt);
    } // if

    // for 'OneInformationSystemNode'
    else if (_evt.getPropertyName().equals("OneInformationSystemNode") ) {
          changeOneInformationSystemNode(_evt);
    } // if 

    // for 'OneInformationSystemNodeOpening'
    else if (_evt.getPropertyName().equals("OneInformationSystemNodeOpening") ) {
          changeOneInformationSystemNodeOpening(_evt);
    } // if 

    return;
  } // propertyChange()

  /**
   * Process event from BaseManagerImpl about 'OneInformationSystem' updatings.
   * Doesn't control the value of a new IS. No action if it is not
   * a registered IS.
   * For removing, the IS name in _evt is a clone,
   * and the IS has been already unregistered in BaseManagerImpl.
   * @param _evt PropertyChangeEvent
   */
  private void changeOneInformationSystem(PropertyChangeEvent _evt){

                    // for testing and tracking
                    //System.out.println("ISTreeBaseListenerImpl: "+
                      //  "call to changeOneInformationSystem() with "+
                        //"old value "+_evt.getOldValue());

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ISTreeBaseListenerImpl: "+
        "call to changeOneInformationSystem() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    ImmutableName           iSname;
    DefaultMutableTreeNode  objNode = null;

    // if the new value is non-null,
    // it is a new IS to insert in the tree.
    if (_evt.getNewValue() != null) {
      iSname = (ImmutableName)_evt.getNewValue();

      ISTree.createISnodeTree(tree_, iSname, iSname, root_);
    } // if (_evt.getNewValue() != null)

    // if the old value is non-null,
    // it is an IS to remove from the tree.
    else if (_evt.getOldValue() != null) {
        iSname = (ImmutableName)_evt.getOldValue();

        objNode = ISTree.getTreeNodeInISTree(iSname);
        if (objNode == null)
          throw new InternalError("ISTreeBaseListenerImpl: "+
            "call to changeOneInformationSystem(), unknown node for "+iSname);

        objNode.setUserObject(null);
        treeModel_.removeNodeFromParent(objNode );
        
        tree_.repaint();
    }// if (_evt.getOldValue() != null)

    iSname = null;
    objNode = null;
  } // changeOneInformationSystem()

  /**
   * Process event from BaseManagerImpl about 'OneInformationSystemNode' updatings.
   * For adding, the IS node may have already children.
   * For removing, the name in _evt is a clone,
   * and the IS node has been already unregistered in BaseManagerImpl.
   * Repaints the 'IS structure' tree and updates the displayed properties
   * @param _evt PropertyChangeEvent
   */
  private void changeOneInformationSystemNode(PropertyChangeEvent _evt){
      
                        // for testing and tracking
                        //System.out.println("ISTreeBaseListenerImpl: "+
                          //  "call to changeOneInformationSystemNode() with "+
                            //"old value: "+_evt.getOldValue());

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ISTreeBaseListenerImpl: "+
        "call to changeOneInformationSystemNode() with both old and new values"+
        " which are not null.");
    } // if

    // variables
    ImmutableName           iSname;
    ImmutableName           name;
    DefaultMutableTreeNode  objNode = null;
    TreePath                path = null;

    // if the new value is non-null,
    // it is a new node to insert in the tree.
    if (_evt.getNewValue() != null) {
      name = (ImmutableName)_evt.getNewValue();
      iSname = BaseManagerImpl.getInstance().getISofName(name);
      if (name.equals(iSname))
          throw new InternalError("ISTreeBaseListenerImpl: "+
            "call to changeOneInformationSystemNode(), the node is an IS.");

      objNode = ISTree.getTreeNodeInISTree(
              BaseManagerImpl.getInstance().getISTree(iSname).getParent(name));
      if (objNode == null)
          throw new InternalError("ISTreeBaseListenerImpl: "+
            "call to changeOneInformationSystemNode(), unknown parent");

      ISTree.createISnodeTree(tree_, iSname, name, objNode);
    } // if (_evt.getNewValue() != null)

    // if the old value is non-null,
    // it is a node to remove from the tree.
    else if (_evt.getOldValue() != null) {
      name = (ImmutableName)_evt.getOldValue();

      objNode = ISTree.getTreeNodeInISTree(name);
      if (objNode == null)
          throw new InternalError("ISTreeBaseListenerImpl: "+
            "call to changeOneInformationSystemNode(), unknown node for "+name);

      objNode.setUserObject(null);
      treeModel_.removeNodeFromParent(objNode );
    }// if (_evt.getOldValue() != null)

    path = tree_.getSelectionModel().getSelectionPath();

    tree_.repaint();

    // updates the properties through a TreeSelectionEvent to valueChanged()
    // which analyzes the current selected node, so the TreeSelectionEvent
    // constructor arguments does not matter
    tree_.getSelectionModel().setSelectionPath(path); // path should have been removed

                // for testing and tracking
                //System.out.println("ISTreeBaseListenerImpl: "+
                  //  "call to changeOneInformationSystemNode() with path: "+path+
                    //", and tree selection is empty: "+
                    //tree_.getSelectionModel().isSelectionEmpty());

    if (!tree_.getSelectionModel().isSelectionEmpty())
       ActionOpenGlobalStructure.getInstance().valueChanged(
            new TreeSelectionEvent(this, path, true, path, path));

    path = null;
    iSname = null;
    objNode = null;
    name = null;
  } // changeOneInformationSystemNode()

  /**
   * Process event from BaseManagerImpl about 'OneInformationSystemNodeOpening' updatings.
   * Repaints the 'IS structure' tree and updates the displayed properties.
   * @param _evt PropertyChangeEvent
   */
  private void changeOneInformationSystemNodeOpening(PropertyChangeEvent _evt){
    // for testing and tracking
    //System.out.println("ISTreeBaseListenerImpl: "+
      //  "call to changeOneInformationSystemNodeOpening() ");

    if (_evt.getOldValue() != null && _evt.getNewValue() != null) {
      throw new InternalError("ISTreeBaseListenerImpl: "+
        "call to changeOneInformationSysteNodeOpening() with both old and new values"+
        " which are not null.");
    } // if

    TreePath path = tree_.getSelectionModel().getSelectionPath();

    tree_.repaint();

    // updates the 'isOpen' property through a TreeSelectionEvent to valueChanged()
    // which analyzes the current selected node, so the TreeSelectionEvent
    // constructor arguments does not matter
    tree_.setSelectionPath(path); // path should have been removed
    if (tree_.isPathSelected(path))
       ActionOpenGlobalStructure.getInstance().valueChanged(
            new TreeSelectionEvent(this, path, true, path, path));

    path = null;
  } // changeOneInformationSystemNodeOpening()

 /**
 * Unregister the instance as a listener for all registered sources.
 * Set the variables to null.
 */
  public void finalizeForUser() {

                // for testing and tracking
            //System.out.println("call to ISTreeBaseListenerImpl.finalizeForUser()");

    BaseManagerImpl.getInstance().removePropertyChangeListener(
          "OneInformationSystem", this);
    BaseManagerImpl.getInstance().removePropertyChangeListener(
          "OneInformationSystemNode", this);
    BaseManagerImpl.getInstance().removePropertyChangeListener(
          "OneInformationSystemNodeOpening", this);

    l_eventSources_.clear();
    sourceName_ = null;
    treeModel_ = null;
    root_ = null;
    tree_ = null;
  } // finalizeForUser()

 /**
 * @return String
 */
  public String toString(){
        StringBuilder sb = new StringBuilder();
        sb.append("  ISTREE_BASELISTENERIMPL.TOSTRING():");

        sb.append(" user object root = ");
        if (root_ != null && root_.getUserObject() != null)
          sb.append( root_.getUserObject().toString() );
        else sb.append("null");

        return " " + sb;
  }

} // ISTreeBaseListenerImpl
