//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;

import javax.swing.*;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

import ARoad0.Gui1.*;
import ARoad0.ErrW.*;
import ARoad0.ErrB.BaseError;
import ARoad0.Pattern.StaticClass;
import ARoad0.Pattern.MVC_Model;

import ARoad0.gWork.ViewInBase_Facade;
import ARoad0.gWorkInterface.RightsMediator;
import ARoad0.gWork.RightsMediatorImpl;

import ARoad0.gBaseInterface.ViewInBase;
import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBase.NameImpl;

/**
 * This class is responsible for managing the gWork.RightsMediator instances,
 * the GraphicViewBaseListeners, the closing tab buttons which may be
 * associated to views, and to manage the pool of view worker threads.
 * It handles both Architecture, Full and NoThan views.
 * When a dedicated view frame is used, the GraphicViewPropertyEditor
 * and the GraphicView are also managed. The Architecture view is based on a ViewInBaseImpl,
 * the 'NoThan' view on a NoThanViewInBaseImpl, and the 'full' view on a EPRViewInBaseImpl.
 * The 'NoThan' view is put in a tab with '<NMNL> ' at the beginning of the title,
 * and all the views use their name last component in the title.
 * The Sketch view is not managed there (see InternalFrameModel). <P>
 * A worker threads pool is provided by this class.
 * It is used by GraphicViewBaseListenerImpl and GraphicViewPropertyEditor
 * to process the view updatings. The maximal number of worker threads is 3,
 * and this allows to use efficiently a modern processor having from 2 to 4 cores. 
 * For a given computer, the number of worker threads may be increased or decreased
 * in the code to speed up the updating of the views. <P>
 * This class is invoked by several Action classes.<P>
 * All methods are static, following the StaticClass pattern.
 * @see GraphicViewBaseListenerImpl
 * @see InternalFrameModel
 * @version 0.7.0
 */
public class ViewModel implements MVC_Model, StaticClass {

  // the only instance of the class, following singleton pattern
  //private static ViewModel            instance__ = new ViewModel();

  private static final int              INITIAL_CAPACITY = 10; // the default value: 101

  /**
   *  list of RightsMediators to control
   */
  private static List                   l_rightsMediators__ =
                                        new ArrayList(INITIAL_CAPACITY);

  /**
   * list of GraphicViewBaseListeners to control
   */ 
  private static List                   l_graphicViewBaseListener__ =
                                        new ArrayList(INITIAL_CAPACITY);
  
  /**
   * current GraphicViews associated to views; the view name is the key
   *  does not include the beamer. 
   */
  private static Map                    m_graphicViewOfViews__  =
                                        new HashMap(INITIAL_CAPACITY); 
  
  /**
   * current GraphicViewPropertyEditors associated to views when one of the two
   *  'All views' JTabbedPane is used; the view name.toString() is the key
   *  does not include the beamer property editors used for views. 
   */
  private static Map                    m_graphicViewPropertyEditors__  =
                                        new HashMap(INITIAL_CAPACITY);  
  
  /**
   * current closing tab JButtons associated to views when one of the 'All full views'
   * JTabbedPane is used; the view name is the key.
   *  does not include the beamer property editors used for views. 
   */
  private static Map                    m_closingFullViewTabButtons__  =
                                        new HashMap(INITIAL_CAPACITY); 
  
  /**
   * current closing tab JButtons associated to views when the 'All NoThan views'
   * JTabbedPane is used; the view name is the key.
   *  does not include the beamer property editors used for views. 
   */
  private static Map                    m_closingNoThanViewTabButtons__  =
                                        new HashMap(INITIAL_CAPACITY); 
  
  /**
   * the 'All Full views' JTabbedPane. May be null. Updated only by
   * ActionOpenAllFullViews.openAllFullViews().
   */
  private static JTabbedPane            FullViewsJTabbedPane__ = new JTabbedPane();

  /**
   * the 'All NoThan views' JTabbedPane. May be null.
   */
  private static JTabbedPane            NoThanViewsJTabbedPane__ = new JTabbedPane();

  /**
 * dynamic list of names for the views for which ActionClose 
 * is closing the tab in the 'All Full views' JTabbedPane.
 * The closing has not to produce the call of the
 * relevant ActionClose class. Used by an inner class in ViewModel
 * and by gDMak.ActionClose.
 */
  private static List                   l_onGoingClosingOnAllFullViews__ =
                                            new ArrayList(INITIAL_CAPACITY);

  /**
 * dynamic list of names for the views for which ActionClose 
 * is closing the tab in the 'All NoThan views' JTabbedPane.
 * The closing has not to produce the call of the
 * relevant ActionClose class. Used by an inner class in ViewModel
 * and by gDMak.ActionClose.
 */
  private static List                   l_onGoingClosingOnAllNoThanViews__ =
                                            new ArrayList(INITIAL_CAPACITY);
  
  /**
   * This is a pool of worker threads which reset the views, and it is used by
   * GraphicViewBaseListenerImpl and GraphicViewPropertyEditor.
   */
  private static ExecutorService        executor__ = null;
  
  /*
   * The number of worker threads to process the updating of the open views,
   * whatever the number of open views. The current value is 3.
   * If the JVM runs on more than 3 processors, the developper should
   * increase this number to try to speed up the updating of the complex views.
   * If the JVM runs on one processor, it is possible to
   * set this number to 1 to try to speed up the updating of all the views, 
   * since the threads handling uses some processing time.
   */
  private static int                    numberOfViewThreads__ = 3;

  static {
      executor__ = Executors.newFixedThreadPool(
              numberOfViewThreads__,
              new SimpleThreadFactory());
  } // static

  /*
   * The private class to set the thread priority to the value Thread.NORM-PRIORITY - 2
   */
  private static class SimpleThreadFactory implements ThreadFactory {
   public Thread newThread(Runnable r) {
       Thread result = new Thread(r);
       result.setPriority(Thread.NORM_PRIORITY - 2);
       return result;
   }
  } // SimpleThreadFactory

  /**
  * only one private constructor
  */
  private ViewModel() {
  }

  /**
  * Only method to get an instance.
  */
//  public static synchronized ViewModel getInstance(){
//    return instance__;
//  } // getInstance()

  /**
  * Creates and registers the RightsMediator associated to the view.
  * Called before the GraphicView creation by ActionOpenGraphicView, ActionOpenArchitecture
  * ActionNewView and ActionOpenSketcher.
  * @param _view to process
  */
  public static final synchronized void createViewRightsMediator(ViewInBase _view) {

                        // for testing and tracking
                        //System.out.println("ViewModel.createViewRightsMediator(): ");
                        //System.out.println("view to add: "+_view.getName());
      
      
        if (getViewRightsMediator(_view) != null) {

                    // for testing and tracking
                System.out.println("error: in ViewModel.createViewRightsMediator(), "+
                    "the view "+_view.getName()+" is already known");

          return;
        }

        RightsMediator mediator = new RightsMediatorImpl(_view);

        synchronized (l_rightsMediators__) {
            l_rightsMediators__.add(mediator);
        }

        mediator = null;
  } // createViewRightsMediator()

  /**
  * Creates and registers the GraphicViewBaseListener.
  * Registers the graphic view and updates the RightsMediator
  * from _gview.getRightsRelatedBaseObjects().
  * Called after the GraphicView creation by ActionOpenGraphicView, ActionOpenArchitecture
  * and ActionNewView. Not called by ActionOpenSketcher.
  * @param _view to process
  * @param _gview is used only for calling _gview.getRightsRelatedBaseObjects()
  *   and for updating the RightsMediator
  */
  public static final synchronized void createGraphicViewBaseListener(
    ViewInBase _view, GraphicView _gview) {

                        // for testing and tracking
                        //System.out.println("ViewModel.createGraphicViewBaseListener(): ");
                        //System.out.println("view to add: "+_view.getName());
      
      
        if (getViewRightsMediator(_view) == null) {

                    // for testing and tracking
                //System.out.println("in ViewModel.createGraphicViewBaseListener(), "+
                  //  "the view "+_view.getName()+" is unknown");

          return;
        }

        getViewRightsMediator(_view).setRightsNetwork( _gview.getRightsRelatedBaseObjects());

        BaseListener listen = new GraphicViewBaseListenerImpl(_view);

        synchronized (l_graphicViewBaseListener__) {l_graphicViewBaseListener__.
          add(listen);}
        synchronized (m_graphicViewOfViews__) {
            m_graphicViewOfViews__.put(_view.getName(), _gview);}

        listen = null;
  } // createGraphicViewBaseListener()

  /**
  * Register a new GraphicView, and reset the main RightsMediator rights network
  * when the paths may have changed, from _gview.getRightsRelatedBaseObjects().
  * No GraphicViewBaseListener changing since this method is called by that.
  * @param _view to process
  * @param _gview is a new or updated graphic view
  */
  public static final synchronized void changeRightsNetworkObserver(
    ViewInBase _view, GraphicView _gview) {
 
        if (getViewRightsMediator(_view) == null) {

                    // for testing and tracking
                //System.out.println("in ViewModel.changeGraphicViewFromListener(), "+
                  //  "the view "+_view+" is not known");

          return;
        }

        getViewRightsMediator(_view).setRightsNetwork(
                _gview.getRightsRelatedBaseObjects());

        synchronized (m_graphicViewOfViews__) {
            m_graphicViewOfViews__.put(_view.getName(), _gview);
        }
  } // changeRightsNetworkObserver()

  /**
  * Change the instances of NodeRightsImpl and LinkRightsImpl of the nodes,
  * when theses nodes have changed. This method has to be called before any
  * updating in the graphic view.
  * Called by GraphicViewBaseListener.changeTabView() and changeFrameView().
  * Calls RightsMediatorImpl.updateNonCoreAlgorithmsForViewNodes().
  * @param _view to process
  */
  public static final synchronized void changeNonCoreAlgorithms(
    ViewInBase _view) {
 
        if (getViewRightsMediator(_view) == null) {

                    // for testing and tracking
                //System.out.println("in ViewModel.changeNonCoreAlgorithms(), "+
                  //  "the view "+_view+" is not known");

          return;
        }

        ((RightsMediator)getViewRightsMediator(_view)).updateNonCoreAlgorithmsForViewNodes(false);

  } // changeNonCoreAlgorithms()

  /**
  * Gets the list of GraphicViewBaseListener associated to all the views in the GUI,
  * excepted for the sketch view that has not any GraphicViewBaseListener.
  * Used by ActionSave.
  * @return unmodifiable list. May be empty, not null.
  */
  public static synchronized final List getL_GraphicViewBaseListener() {
    return Collections.unmodifiableList( l_graphicViewBaseListener__ );
  } // getL_GraphicViewBaseListener()

  /**
  * Gets the main RightsNetworkObserver associated to a view in the GUI.
  * Used by GraphicViewBaseListenerImpl, ActionSave, ActionOpenMainFrame,
  * and numerous classes in the packages gWork and AcsAddon gWork.
  * @param _view the view
  * @return a RightsMediator or null
  */
  public static synchronized final RightsMediator getViewRightsMediator(
    ViewInBase _view) {
        RightsMediator result = null;
        boolean found = false;
        for(Iterator ite = l_rightsMediators__.iterator(); ite.hasNext(); ) {
          result = (RightsMediator)ite.next();
          if (result.getSource().getName().equals(_view.getName())) {
            found = true;
            break;
          } // if
        } // for
        if (found) return result;
        else return null;
  } // getViewRightsMediator()

  /**
  * Gets the main GraphicViewBaseListener associated to a view in the GUI.
  * @param _view the view
  * @return a GraphicViewBaseListener or null
  */
  public static final BaseListener getGraphicViewBaseListener(
    ViewInBase _view) {
        BaseListener result = null;
        boolean found = false;
        for(Iterator ite = l_graphicViewBaseListener__.iterator(); ite.hasNext(); ) {
          result = (BaseListener)ite.next();
          if (result.getEorMainEventSource().equals(_view.getName())) {
            found = true;
            break;
          } // if
        } // for
        if (found) return result;
        else return null;
  } // getGraphicViewBaseListener()

  /**
  * Removes the view RightsMediator and the GraphicView which are associated to a view.
  * Used also for the sketch view. Called by ActionClose.
  * @param _source is the ViewInBase
  * @exception SmallError if the observer is unknown
  */
  public static final void removeViewRightsMediator(ViewInBase _source)
    throws SmallError {
      // variables
      RightsMediator    mediator = null;
      boolean           find = false;

                        // for testing and tracking
                        //System.out.println("ViewModel.removeViewRightsMediator() for the view: "+
                          //  _source.getName());

      // search the right baseListeners
      synchronized (l_rightsMediators__) {
        for (ListIterator ite = l_rightsMediators__.listIterator() ; ite.hasNext() ; ) {
          mediator = (RightsMediator) ite.next();

                        // for testing and tracking
                        //System.out.println("--- a registered mediator for: "+mediator.getSource());

          if (mediator.getSource().getName().equals(_source.getName())) {
              find = true;
              break;
          } // if (listener.
        } // for

        // exception
        if ( !find ) throw new SmallError(
          "in ViewModel.removeViewRightsMediator(), the mediator is unknown.");
        else {
          l_rightsMediators__.remove(mediator);
          mediator.finalizeForProcess();
        }
      } // synchro

      mediator = null;
  } // removeViewRightsMediator()

  /**
  * Removes the view the GraphicViewBaseListenerImpl and
  * the GraphicView which are associated to a view.
  * No use for the sketch view. Called by ActionClose.
  * @param _source is the ViewInBase
  * @exception SmallError if the graphic view is unknown
  */
  public static final void removeGraphicViewBaseListener(ViewInBase _source)
    throws SmallError {
      // variables
      BaseListener            listener = null;
      boolean                 find = false;

                        // for testing and tracking
                        //System.out.println("ViewModel.removeGraphicViewBaseListener() for the view: "+
                          //  _source.getName());

      // search the right baseListeners
      synchronized (l_graphicViewBaseListener__) {

        for (ListIterator ite = l_graphicViewBaseListener__.listIterator() ; ite.hasNext() ; ) {
          listener = (BaseListener) ite.next();

                        // for testing and tracking
                        //System.out.println("a registered listener: "+listener);

          if (listener.getEorMainEventSource().equals(_source.getName())) {
              find = true;
              break;
          } // if (listener.
        } // for

        // exception
        if ( !find ) throw new SmallError(
          "in ViewModel.removeGraphicViewBaseListener(), the listener is unknown.");
        else {

          l_graphicViewBaseListener__.remove(listener);
          listener.finalizeForUser();
          
          synchronized (m_graphicViewOfViews__) {
            m_graphicViewOfViews__.remove(_source.getName());}
        }
      } // synchro

      listener = null;
  } // removeGraphicViewBaseListener()
  
  /**
   * Update m_GraphicViewPropertyEditor__.
   * Used by GraphicViewPropertyEditor.
   * @param _viewName is ViewInBase.getName().toString()
   * @param _component is associated to _object
   */
  public static void addGraphicViewPropertyEditor(String _viewName,
    GraphicViewPropertyEditor _component) {
      
      if (_viewName == null || _component == null)
        throw new NullPointerException();
      synchronized (m_graphicViewPropertyEditors__) {
            m_graphicViewPropertyEditors__.put(_viewName, _component);

                          // for testing and tracking
                          //System.out.println("in ViewModel.addGraphicViewPropertyEditor(), "+
                            //" m_graphicViewPropertyEditor__ after the updating: ");
                          //for (Iterator ite = m_graphicViewPropertyEditor__.keySet().iterator();
                            //ite.hasNext(); ) {
                            //System.out.println("---- "+ite.next());
                          //} // for ite
      }
  } // addGraphicViewPropertyEditor()

  /**
   * Update m_GraphicViewPropertyEditor__.
   * Used by GraphicViewPropertyEditor.
   * @param _viewName is ViewInBase.getName().toString()
   */
  public static void removeGraphicViewPropertyEditor(String _viewName) {
      synchronized (m_graphicViewPropertyEditors__) {
            m_graphicViewPropertyEditors__.remove(_viewName);


                          // for testing and tracking
                          //System.out.println("in InternalFrameModel.removeGraphicViewPropertyEditor(), "+
                            //" m_graphicViewPropertyEditors__ after the updating: ");
                          //for (Iterator ite = m_graphicViewPropertyEditor__.keySet().iterator();
                            //ite.hasNext(); ) {
                            //System.out.println("---- "+ite.next());
                          //} // for ite
      }
      //if (m_graphicViewPropertyEditors__.isEmpty()) EPRviewsIsOpen__ = false;      
  } // removeGraphicViewPropertyEditor()    

  /**
   * Used by BeamerManager.
   * @param _viewName is ViewInBase.getName().toString()
   * @return the GraphicViewPropertyEditor or null
   */
  public static GraphicViewPropertyEditor getGraphicViewPropertyEditor(String _viewName) {
      return (GraphicViewPropertyEditor)m_graphicViewPropertyEditors__.get(_viewName);
  } // getGraphicViewPropertyEditor() 
  
  /**
   * Used by GraphicViewPropertyEditor.valueOf().
   * @param _name of a ViewInBase
   * @return the GraphicView or null
   */ 
  public static GraphicView getGraphicViewOfView(ImmutableName _name) {
      return (GraphicView)m_graphicViewOfViews__.get(_name);
  } // getGraphicViewOfView()  
  
  //
  // ---------- start of the Full and NoThan views management ----------------
  //
  
  /**
   * Update the dynamic list of names for the Full and NoThan views
   * which ActionClose is closing.
   * Used by gDMak.ActionClose.
   * @param _name of the view which is closing
   */
  public static void addOnGoingClosingOnView(ImmutableName _name) {
      if (ViewInBase_Facade.isNoThanViewName(_name)) {
          synchronized (l_onGoingClosingOnAllNoThanViews__)
            {l_onGoingClosingOnAllNoThanViews__.add(_name);}          
      }
      else synchronized (l_onGoingClosingOnAllFullViews__)
            {l_onGoingClosingOnAllFullViews__.add(_name);}
  } // addOnGoingClosingOnView()  

  
  /**
   * Remove a name among the Full and NoThan views which are closing.
   * @param _name to remove
   */
  public static void removeOnGoingClosingOnView(ImmutableName _name) {
      if (ViewInBase_Facade.isNoThanViewName(_name)) {
          synchronized (l_onGoingClosingOnAllNoThanViews__)
            {l_onGoingClosingOnAllNoThanViews__.remove(_name);}          
      }
      else synchronized (l_onGoingClosingOnAllFullViews__)
            {l_onGoingClosingOnAllFullViews__.remove(_name);}
  } // removeOnGoingClosingOnView()  

  /**
   * Used by BeamerManager.
   * @return unmodifiableList of the Full and NoThan views which are closing
   */
  public static List getOnGoingClosingOnView() {
    List result = (List)((ArrayList)l_onGoingClosingOnAllFullViews__).clone();
    result.addAll(l_onGoingClosingOnAllNoThanViews__);
    return result;
  } // getOnGoingClosingOnView()
  
  /**
   * Update m_closingFullViewTabButtons__ or m_closingNoThanViewTabButtons__.
   * Check if it is a full view name or a NoThan view name, to update the right map.
   * Called by BeamerManager.initializeViewTabComponent().
   * @param _name of the ViewInBase
   * @param _component is the closing button in the tab associated 
   *    to the ViewInBase
   */
  public static void addClosingTabButtonInTabbedPane(ImmutableName _name,
    JButton _component) {
      
      if (_component == null || _name == null) throw new NullPointerException();
      if (FullViewsJTabbedPane__ == null && NoThanViewsJTabbedPane__ == null) 
          throw new InternalError("ViewModel.addClosingTabButtonInTabbedPane() with null JTabbed panes");
      // the next test does not work since the button is not still put in its ancestor
      //if (!FullViewsJTabbedPane__.isAncestorOf(_component))
        //  throw new InternalError("in ViewModel.addClosingTabButtonInTabbedPane(), "+
          //        "the button is not in the 'All views' tabbed pane");
      if (ViewInBase_Facade.isNoThanViewName(_name)) {
          synchronized (m_closingNoThanViewTabButtons__) {
                m_closingNoThanViewTabButtons__.put(_name, _component);
          }
      }
      else {
          synchronized (m_closingFullViewTabButtons__) {
                m_closingFullViewTabButtons__.put(_name, _component);
          }          
      }
  } // addClosingTabButtonInTabbedPane()

  /**
   * Update m_closingFullViewTabButtons__ or m_closingNoThanViewTabButtons__
   * without any check. Used by ActionClose.
   * @param _name of the ViewInBase for which the tab is removed
   */
  public static void removeClosingTabButtonInTabbedPane(ImmutableName _name) {
      if (ViewInBase_Facade.isNoThanViewName(_name)) {
          synchronized (m_closingNoThanViewTabButtons__) {
                m_closingNoThanViewTabButtons__.remove(_name);
          }
      }
      else {
          synchronized (m_closingFullViewTabButtons__) {
                m_closingFullViewTabButtons__.remove(_name);
          }          
      }   
  } // removeClosingTabButtonInTabbedPane()    

  /**
   * Used by ActionClose.
   * @param _name of a full or NoThan view
   * @return the JButton which closes the view tab, or null
   */
  public static JButton getClosingTabButtonInTabbedPane(ImmutableName _name) {
      if (ViewInBase_Facade.isNoThanViewName(_name))
          return (JButton)m_closingNoThanViewTabButtons__.get(_name);
      else
          return (JButton)m_closingFullViewTabButtons__.get(_name);
  } // getClosingFullViewTabButtonInTabbedPane()      
  
  /**
   * Gets the tabbed pane in the 'All full views' window.
   * @return the tabbed pane
   */
  public static JTabbedPane getAllFullViewsJTabbedPane() {
      return FullViewsJTabbedPane__;
  } // getAllEPRviewsJTabbedPane()   
  
  /**
   * Gets the tabbed pane in the 'All NoThan views' window.
   * @return the tabbed pane
   */
  public static JTabbedPane getAllNoThanViewsJTabbedPane() {
      return NoThanViewsJTabbedPane__;
  } // getAllNoThanViewsJTabbedPane() 

  /**
   * Gets the tab title in the 'All views' or 'NoThan views' JTabbedPane,
   * as two or three name components with ':: ' as separator. For a NoThanViewInBase
   * where the name has 3 components, the title contains
   * the component ViewInBase_Facade.NOTHAN_VIEW_INDICATOR.
   * The title is like the String form of a Name, without the start and the end.
   * @param _name of a ViewInBase
   * @return the title. Null if _name is null.
   */
  public static String getTabTitleFromViewName(ImmutableName _name) {
       if (_name == null)
           return null;
       else if (_name.noOfComponents() == 1)
           return _name.lastComponent();
       else if (_name.noOfComponents() == 2) {
           return _name.component(_name.noOfComponents() - 2) + ":: " +
                   _name.component(_name.noOfComponents() - 1);
       }
       else if (_name.noOfComponents() == 3) {
           return _name.component(_name.noOfComponents() - 3) + ":: " +
                   _name.component(_name.noOfComponents() - 2) + ":: " +
                   _name.component(_name.noOfComponents() - 1);
       }
       else return null;
  } // getTabTitleFromViewName()

  /**
   * Gets the ViewInBase name from the tab title in the 'All full views' or from the
   * 'All NoThan views' JTabbedPane. Called by BeamerManager.setJTabbedPane(),
   * initializeViewTabComponent(), ActionOpenAllFull/NoThanViews.stateChanged().
   * Calls ViewInBase_Facade.getAllStandardOpenViews() and isNoThanViewName().
   * @exception InternalError if the name is not found with a non-null argument
   * @param _title of the tab
   * @return ImmutableName of the view. Null only if a null argument.
   */
  public static ImmutableName getViewNameFromTabTitle(String _title) {
      if (_title == null) return null;

      ImmutableName name = null;
      ImmutableName viewName = null;
      boolean found = false;

      try {
          viewName = NameImpl.getNameFromString(" "+_title+"::  ");
      }
      catch (BaseError b) {
          throw new InternalError("Wrong name in ViewModel.getViewNameFromTabTitle() "+
              "for the title: "+_title);
      }

      for (Iterator ite = ViewInBase_Facade.getAllStandardOpenViews().keySet().iterator(); 
         ite.hasNext();) {
             name = (ImmutableName) ite.next();
             if (name.equals(viewName) &&
                     !ViewInBase_Facade.isNoThanViewName(name)) {
                 found = true;
                 break;
             }
             else if (name.equals(viewName) &&
                     ViewInBase_Facade.isNoThanViewName(name)) {
                 found = true;
                 break;
             }
      } // for
      viewName = null;
      if (found) return name;
      else throw new InternalError("Name not found in ViewModel.getViewNameFromTabTitle() "+
              "for the title: "+_title);
  } // getViewNameFromTabTitle()   

  /**
   * Gets the pool of worker threads which reset the views. It is used by
   * GraphicViewBaseListenerImpl and GraphicViewPropertyEditor.
   * @return the thread pool
   */
  public static ExecutorService getViewsThreadsExecutor() {
      return executor__;
  } // getAllNoThanViewsJTabbedPane()

} // class ViewModel