//Title:        Access Road
//Version:      0.6.5
//Copyright:    Copyright 2002, 2004, 2007, 2008, 2009 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.Gui2;

import javax.swing.*;
import javax.swing.tree.*;
import java.beans.*;
import java.util.Map;

import ARoad0.Gui1.*;
import ARoad0.gDMak.ActionClose;
import ARoad0.gBaseInterface.ViewInBase;
import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.NoThanViewInBase;
import ARoad0.gBase.BaseManagerImpl;

/**
 * This class manages one tree view in the explorer.
 * It is responsible for listening the events: <P>
 * - 'OpenViews' fired BaseManager when the view is closed, <P>
 * - 'ConnectedACS' fired by the ViewInBase when an implied ACS is closing, <P>
 * - events fired by the changing ViewInBase objects, through the ViewInBase,
 *    for updating the JTree which displays them in the explorer <P>
 * Each event source is registered only one time.
 * The automatic closing of views are managed only by them, but
 * this class detects the closing by the base, and call ActionClose to
 * call the relevant GUI objects.
 * @version 0.6.5
 */
public class ViewTreeBaseListenerImpl implements TreeBaseListener {
    // variables
  private static final int  PRIME = 1000003;
  private static final int  INITIAL_CAPACITY = 5; // the default value: 101
      /*
       * The view managed by this listener
       */
  private ViewInBase                view_ = null;
  private ImmutableName             viewName_ = null;
  private DefaultMutableTreeNode    root_ = null;
  private DefaultTreeModel          treeModel_ = null;
      /*
       * To listen or not the events in propertyChange(). Not used here.
       */
  private boolean                   listenEvents_ = true;
    // constructors

  /**
   * Public constructor.
  */
  public ViewTreeBaseListenerImpl() {
  }

 /**
 * With this constructor, the ViewTreeBaseListenerImpl is registered
 *  as a listener for the view.
 * It is associated to the TreeModel of the explorer.
 * @param   _source is the view
 */
  public ViewTreeBaseListenerImpl(ViewInBase _source) {
      view_ = _source;
      viewName_ = _source.getName();

        //System.out.println("ViewTreeBaseListenerImpl constructor for "+viewName_ );

      treeModel_ = (DefaultTreeModel)InternalFrameModel.getTreeInExplorer().getModel();

      DefaultMutableTreeNode viewsNode = (DefaultMutableTreeNode)
        ((DefaultMutableTreeNode)treeModel_.getRoot()).getChildAt(1);
      // works for all indirect children of viewsNode
      root_ = ACSTreeUtilities.getCurrentObjectNode(viewsNode, _source);
      
      BaseManagerImpl.getInstance().addPropertyChangeListener("OpenViews", this); 

      view_.addPropertyChangeListener("OneBaseObjectInFolder", this);
      view_.addPropertyChangeListener("ConnectedACS", this);
      if (_source instanceof NoThanViewInBase)
          view_.addPropertyChangeListener("FirstAccessTarget", this);
  } // ViewTreeBaseListenerImpl()

    //
    // ----------------- METHODS -------------------------------
    //

 /**
 * Get the main event source associated to the instance
 * in its construction.
 * @return name of the main source for PropertyChangeEvents. May be null.
 */
  public ImmutableName getEorMainEventSource() {
    return viewName_;
  } // getEorMainEventSource()

 /**
 * To listen or not the events in propertyChange(). No use.
 * @param _listen true to listen the events in this instance, through propertyChange()
 * @return true if the previous value was different to this argument
 */
  public synchronized boolean setListenEvents(boolean _listen) {
    boolean previous = listenEvents_;
    listenEvents_ = _listen;

    if (previous && listenEvents_) return false;
    else if (!previous && !listenEvents_) return false;
    else return true;
  } // setListenEvents()

 /**
 * Get the TreeModel associated to the instance
 * in its construction.
 * @return DefaultTreeModel associated to the instance and to a JTree
 */
  public DefaultTreeModel getEorTreeModel() {
    return treeModel_;
  } // getEorTreeModel()

 /**
 * Get the TreeModel root associated to the instance
 * in its construction.
 * @return MutableTreeNode associated to the instance in the TreeModel
 */
  public MutableTreeNode getEorRoot() {
    return root_;
  } // getEorRoot()

 /**
 * Register the instance as a listener for a BaseObject
 * firing PropertyChangeEvents. No use.
 * @param _source BaseObject source of PropertyChangeEvents
 * @exception SmallError if _source is already known.
 */
/*  public void addEventSource(BaseObject _source) throws SmallError {
      if (!(_source instanceof BaseObject) )
        throw new SmallError("in ViewTreeBaseListenerImpl.addEventSource(_source)"+
          "_source is not a BaseObject.");
      if (l_eventSources_.contains(_source) )
        throw new SmallError("in ViewTreeBaseListenerImpl.addEventSource(_source)"+
          "_source is already known.");
      l_eventSources_.add(_source);
      // TO DO: change the cast to ACSObject later
      //((ACSImpl)_source).addPropertyChangeListener(this);
  } // addEventSource()
*/

 /**
 * Unregister the instance as a listener for a source. No use.
 * @param _source BaseObject registered source of PropertyChangeEvents
 * @exception SmallError if _source is unknown.
 */
/*  public void removeEventSource(BaseObject _source) throws SmallError {
      if (!l_eventSources_.contains(_source) )
        throw new SmallError("in ViewTreeBaseListenerImpl.removeEventSource(_source)"+
          "_source is unknown.");
      l_eventSources_.remove(_source);
      // TO DO: change the cast to ACSObject later
      //((ACSImpl)_source).removePropertyChangeListener(this);
  } // removeEventSource()
*/

  /**
   * Receive the source events about property changes.
   * Listen 'OpenViews' from the base manager,
   * 'ConnectedACS' and 'OneBaseObjectInView' from the ViewInBase,
   * and 'FirstAccessTarget' from the NoThanViewInBase.
   * When the view fires a 'ConnectedACS' event with a null value, 
   * since this means that the view closing by the base is the next step,
   * inform the user and call ActionClose.closeInGUI().
   * @param _evt PropertyChangeEvent 
   */
  public void propertyChange(PropertyChangeEvent _evt){

                    // for testing and tracking
                    //System.out.println("in ViewTreeBaseListenerImpl, for "+viewName_+" and "+
                      //  _evt.getPropertyName()+" and the new value map is: "); 
                    //for (Iterator ite = ((Map)_evt.getNewValue()).keySet().iterator(); ite.hasNext(); ) {
                      // System.out.println("-------- "+ite.next()); 
                    ///}

    if (view_ == null) return;

    boolean toClose = false;

    // for 'OpenViews'
    if ( _evt.getPropertyName().equals("OpenViews") ) {
          // test if view_ is still open - _evt.getNewValue() is NOT the current base map of open views
          if (_evt.getNewValue() != null && !((Map)_evt.getNewValue()).containsKey(viewName_)) {
            toClose = true;
            
                    // for testing and tracking
                //System.out.println("in ViewTreeBaseListenerImpl, closing detected for "+viewName_);
          
          } // if
    } // if (_evt.getPropertyName()
    // for 'ConnectedACS'
    else if ( _evt.getPropertyName().equals("ConnectedACS") ) {
          // test if view_ is still open
          if (_evt.getNewValue() == null) {
            toClose = true;
            DialogBox.showOkMessage("The view "+viewName_+
              " has to be closed, since"+
              " one of its ACS is closing.");
          } // if
    } // if (_evt.getPropertyName()
    // for 'OneBaseObjectInView'
    else if ( _evt.getPropertyName().equals("OneBaseObjectInFolder") ) {

          // test if it is a deleted element
          if (_evt.getNewValue() == null) {
            changeViewTree();
          }
          // else it is a new element
          else {
            changeViewTree();
          }

    } // if (_evt.getPropertyName()
    else if ( _evt.getPropertyName().equals("FirstAccessTarget") ) {

          // test if it is a deleted element
          if (_evt.getNewValue() == null) {
            changeViewTree();
          }
          // else it is a new element
          else {
            changeViewTree();
          }

    } // if (_evt.getPropertyName()
    
    if (toClose) {
          // close the view in explorer, the view main frame and the 'See why' frame if any.
          // closing the view in the base is not done there.
          // call back this.finalizeForUser() through ViewModel.
        JInternalFrame          frame = null;
        JPanel                  pane = null;

                          // for testing and tracking
                        //System.out.println("ViewTreeBaseListernerImpl.propertyChange() closing for "+viewName_);

        frame = (JInternalFrame)InternalFrameModel
                                      .getMainFramesOfBaseObjects().get(viewName_);
        // if a view is not in a frame, it may be in a tab of the 'All views' frame
        if (frame == null) pane = (JPanel)InternalFrameModel
                                            .getJComponentsOfBaseObjects().get(viewName_);

        // close in the GUI
        if (frame != null) ActionClose.getInstance().closeInGUI(viewName_, frame);
        else if (pane != null) ActionClose.getInstance().closeInGUI(viewName_, pane);
        else ActionClose.getInstance().closeInGUI(viewName_, null); // if it is open only in the explorer
        
        frame = null;
        pane = null;
    } // if (toClose) 
                
    return;
  } // propertyChange()

  /**
   * Update the view tree in the explorer.
   * Call ViewTree.fillViewNodeInExplorer().
   */
  private void changeViewTree(){

    DefaultMutableTreeNode objNode = null;

    if (root_ == null) return;

    // does root_.removeAllChildren()
    while (!root_.isLeaf()) {
        objNode = (DefaultMutableTreeNode)treeModel_.getChild(root_, 0);
        objNode.setUserObject(null);
        treeModel_.removeNodeFromParent(objNode );
    } // while

            // for testing and tracking
            //System.out.println("ViewTreeBaseListenerImpl."+
              //"changeViewTree() after the node clearing, for "+view_);

    if (!view_.getEorM_BaseObjects().isEmpty())
          ViewTree.fillViewNodeInExplorer(
                view_, InternalFrameModel.getTreeInExplorer(), root_);
    
    objNode = null;
    return;
  }  // changeViewTree()

 /**
 * Unregister the instance as a listener for the view.
 * Set the variables to null.
 */
  public void finalizeForUser() {
    // remove as a listener
    BaseManagerImpl.getInstance().removePropertyChangeListener("OpenViews", this);
    view_.removePropertyChangeListener("OneBaseObjectInFolder", this);
    view_.removePropertyChangeListener("ConnectedACS", this);
   if (view_ instanceof NoThanViewInBase)
      view_.removePropertyChangeListener("FirstAccessTarget", this);
    
            // for testing and tracking
            //System.out.println("call of ViewTreeBaseListenerImpl."+
              //"finalizeForUser() for "+view_);

   /*
   Object obj;
   for (ListIterator ite = l_eventSources_.listIterator() ; ite.hasNext() ; ) {
    obj = ite.next();
    //if (obj instanceof ViewInBase) ((ViewInBase)obj).removePropertyChangeListener(
    //  (PropertyChangeListener)this);
    // to complete later
    // if (obj instanceof ACSObject) ((ACSObject)obj).removePropertyChangeListener(this);
    // if (obj instanceof ACSManager)
   } // for
   l_eventSources_.clear();
   */
   view_ = null;
   viewName_ = null;
   treeModel_ = null;
   root_ = null;
  }// finalizeForUser()

 /**
 * HashCode of the instance.
 * It is used by TreeManager to remove a TreeBaseListener,
 * by BaseManagerImpl and ViewInBase to fire events to this instance.
 * Don't use the event source list.
 * @return int
 */
  /*public int hashCode(){
        int result = 0;

        result = view_.getName().hashCode();
        result = PRIME*result + ((DefaultMutableTreeNode)root_.getParent()).
                getUserObject().hashCode();

        return result;
  } // hashCode()
  */
 
  /**
   * no equals: each instance is equals only to itself.
   * DefaultTreeModel and DefaultMutaleTreeNode have no equals() methods.
   * @param _obj: object to compare
   * @return true if the two objects are identical instances
   * of identical classes that implement ViewTreeBaseListener
   */
  /*public synchronized boolean equals(Object _obj) {
    if (this == _obj)
        return true;
    if (_obj == null)
        return false;
    if (_obj.getClass() != getClass())
        return false;

    ViewTreeBaseListenerImpl obj = (ViewTreeBaseListenerImpl) _obj;

    if (view_ == null || obj.getEorMainEventSource() == null) {
        if (view_!= null || obj.getEorMainEventSource() != null)
          return false;  // if
    } // if
    else {
      if (!viewName_.equals( obj.getEorMainEventSource() )) return false; }

    if (root_.getParent() == null || obj.getEorRoot().getParent() == null) {
        if (root_.getParent() != null || obj.getEorRoot().getParent() != null)
          return false;  // if
    } // if
    else {
      if (!((DefaultMutableTreeNode)root_.getParent()).getUserObject().equals(
        ((DefaultMutableTreeNode)obj.getEorRoot().getParent()).getUserObject() ))
          return false; }

    return true;
  } //equals()
  */

 /**
 * @return String
 */
  public String toString(){
        StringBuilder sb = new StringBuilder();
        sb.append("  VIEWTREE_BASELISTENERIMPL.TOSTRING():");

        sb.append(" view name = ");
        sb.append(viewName_);

        sb.append(" root = ");
        sb.append( root_ );

        sb.append(" hashCode() = ");
        sb.append( Integer.toString(hashCode()) );

        return " " + sb;
  } // toString()

} // ViewTreeBaseListenerImpl