//Title:        Access Road
//Version:      0.7.2
//Copyright:    Copyright 2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.Pattern.Singleton;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.Arrays;
import java.util.Iterator;
import java.util.HashMap;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Collection;

/**
* This class is an important utility class for ACSFactoryImpl, with only static methods.
* It cannot be subclassed in an AcsAddon gBase package. Each static method is documented
* independently. This class contains also some important static properties
* which are presented hereinafter. <P>
* This class defines around 60 granting metarights, and their denying 'deny_' opposites.
* '|unknown_rights|' is the less powerful right, with a strenght at 1 and no opposite right.
* '|is_superadmin|' is the more powerful right, with a strenght at 90900 and no opposite right.
* The pattern 'directory/child' in names is used for the AGO inherited rights in
* BaseUtilityImpl.getChildRightsSetFromParentAGORights().<P>
* The generic names of the ACS upper nodes are in L_ACS_TREE_GENERIC_UPPER_NODES.
* They are used by a ARoad0.Gui2.ACSTreeBaseListener for adding or removing an ACSObject. <P>
* Some tree ACS upper node names may be replaced for a given AcsAddon ACS.
* Their default values are in L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.
* The getM_AcsVocabulary() method returns the effective values for an ACS.
* They are never replaced in the explorer ACS tree, in which the generic vocabulary
* of Access Road is mandatory to ensure a coherent view. They should be used
* in the internal frame of the ACS, to describe the ACS with its own public vocabulary.
* For instance, the node name 'Resources' should be replaced by the node name 'IP addresses'
* in the ACS tree of a firewall. <P>
* The AcsAddon framework is based on the following properties: <P>
* Some classes in the generic gBase package may be subclassed in an AcsAddon package,
* and being visible for all the generic packages. These subclasses are called
* classic AcsAddon gBase classes. They are accessible for all the Access Road program
* through the getM_AcsVocabulary() method of the ACS. The default values of these classes
* are the generic gBase classes, in M_GENERIC_ACS_VOCABULARY, with the editor name and the version.
* Other gBase classes like BaseUtilityImpl are not registered if they are subclassed
* in an AcsAddon, like in Ubuntu. In that case, only the classes of their AcsAddon gBase package,
* or a derived package, may call them. The classic classes do not include
* the embedded classes like CAliasImpl. They may be subclassed in the AcsAddons
* but are not used outside the AcsAddon gBase package. <P>
* An AcsAddon may have gBase classes which are not classic classes.
* They are called AcsAddon extension classes. For an AcsAddon vocabulary map,
* ACSADDON_EXTENSION_CLASSES is the key to define the extension classes in gBase.
* The extension class descriptions are separated by one space,
* like in 'Executable/is/ACTOR_CLASS XXX/is/RESOURCE_CLASS'. This says that
* there is a class Executable'AcsAddon_Name'Impl under the AcsAddon gBase package, and
* that this class is a direct subclass of the ACTOR_CLASS. This ACTOR_CLASS may
* be the generic ARoad0.gBase.Actor class, or may be the classic class in
* the same AcsAddon package that replaces the ARoad0.gBase.Actor class.
* For instance, in the AcsAddon Ubuntu, the classic class ActorUbuntuImpl replaces
* the generic ARoad0.gBase.ActorImpl class, and the extension class
* ExecutableUbuntuImpl is a subclass of the classic class ActorUbuntuImpl.
* ACSADDON_EXTENSION_CLASSES is used by Gui1.BaseObjectHandler and some gDMak classes
* to set the list of the allowed subclasses for creating an ACSObject in an AcsAddon. <P>
* The domain of a specialized right is one of the following categories:
* 'AccRightsSet' and 'GpORightsSet' for Target only, 'AclRightsSet', 'PrvRightsSet', 
* 'BdgRightsSet', 'RooRightsSet' for Source and Target. It is not possible to change
* this set for an AcsAddon, while it is necessary sometimes to have specialized
* rights on a specific right property.
* In the AcsAddon vocabulary map, ACSADDON_SPECIALIZED_RIGHT_DOMAINS is the key
* to define the specialized right domains. The associated values set the non-generic rights
* in the ACSObjects on which the ACS constraints of the specialized rights
* are applicable, and in that case, what domain is applicable. For instance,
* the domain 'GpORightsSet' is applicable to the specific property 'OtherRights' in Ubuntu,
* to find the specialized rights on the 'OtherRights', if any.
* The value must have the format 'PropertyName.SpecializedRightDomainName',
* and it is ridden by the method getL_MapKeysForSpecializedRights().
* Caution: the current version of Access Road has to be put into M_GENERIC_ACS_VOCABULARY.
* @version 0.7.2
*/
public class ACSFactoryUtilityImpl implements Singleton {
      /* the only instance of the class, following singleton pattern */
  private static ACSFactoryUtilityImpl  instance__ = new ACSFactoryUtilityImpl();

  private static final int        INITIAL_CAPACITY = 10; // the default value: 101

  // these are the generic names of the ACS nodes. They are used
  // by an Gui2.ACSTreeBaseListener for adding or removing an ACS object.
  public static final String  ACTORS = "Actors (right user & target)";

  public static final String  USERS = "UserIDs (right user)";

  public static final String  GROUPS = "GroupIDs (right user)";

  public static final String  RESOURCES = "Resources (access target)";

  public static final String  ACLENTRIES = "AclEntries"+
                              ": right user >> access target";
  public static final String  EEPARTIES = "External Eligible Parties"+
                              " (right user)";
  public static final String  EEACTORS = "External Eligible Actors (right user & target)";

  public static final String  EGROUPS = "External GroupIDs";

  public static final String  EUSERS = "External UserIDs";

  public static final String  ECRESOURCES = "External Controlled Resources"+
                              " (access target)";
  public static final String  ECACTORS = "External Controlled Actors (right user & target)";

  public static final String  ECOTHER_RESOURCES = "Other External Controlled Resources";

  public static final String  ACSRIGHTS = "Types of ACS Right";

  public static final String  RESVFOLDERS = "Virtual Folders for resources (access target)";

  public static final String  EPVFOLDERS = "Virtual Folders for all (access target)";

  public static final String  PRIVILEGES = "Privileges: right user >> access target";

  /*
   * List of the tree ACS node names that are the generic names, to use in all cases in the explorer.
   * They may be replaced in the proper internal frame of an ACS.
   */
  public static final List    L_ACS_TREE_GENERIC_UPPER_NODES = new ArrayList(20);


  // these are the Addon names of the ACS nodes. They are used as nodes
  // in the specific tree of the AcsAddon ACS.
  public static final String  ACTORS_FOR_THIS_ACS = "Actors for this ACS (right user target)";

  public static final String  USERS_FOR_THIS_ACS = "UserIDs for this ACS (right user)";

  public static final String  GROUPS_FOR_THIS_ACS = "GroupIDs for this ACS (right user)";

  public static final String  RESOURCES_FOR_THIS_ACS = "Resources for this ACS (access target)";

  public static final String  ACLENTRIES_FOR_THIS_ACS = "AclEntries for this ACS"+
                              ": right user >> access target";
  public static final String  EEPARTIES_FOR_THIS_ACS = "External Eligible Parties for this ACS"+
                              " (right user)";
  public static final String  EEACTORS_FOR_THIS_ACS = "External Eligible Actors for this ACS";

  public static final String  EGROUPS_FOR_THIS_ACS = "External GroupIDs for this ACS";

  public static final String  EUSERS_FOR_THIS_ACS = "External UserIDs for this ACS";

  public static final String  ECRESOURCES_FOR_THIS_ACS = "External Controlled Resources for this ACS"+
                              " (access target)";
  public static final String  ECACTORS_FOR_THIS_ACS = "External Controlled Actors for this ACS";

  public static final String  ECOTHER_RESOURCES_FOR_THIS_ACS = "Other External Controlled Resources"+
                                " for this ACS";

  public static final String  ACSRIGHTS_FOR_THIS_ACS = "Types of ACS Right for this ACS";

  public static final String  RESVFOLDERS_FOR_THIS_ACS = "Virtual Folders of targets (access target)"+
                                " for this ACS";

  public static final String  EPVFOLDERS_FOR_THIS_ACS = "Virtual Folders of right user targets "+
                                " for this ACS (access target)";

  public static final String  PRIVILEGES_FOR_THIS_ACS = "Privileges for this ACS:"+
                                "right user >> access target";

  /*
   * List of the tree ACS node names that may be replaced for a given AcsAddon ACS.
   * They are never used in the explorer, and they are reserved to
   * the proper internal frame of the ACS.
   */
  public static final List    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS = new ArrayList(20);


  public static final String  ACTOR_CLASS = "Actor";

  public static final String  USER_CLASS = "UserID";

  public static final String  GROUP_CLASS = "GroupID";

  public static final String  RESOURCE_CLASS = "Resource";

  public static final String  ACLENTRY_CLASS = "AclEntry";
          
  public static final String  DIRECTORY_CLASS = "Directory";
  
  public static final String  GROUPMEMBER_CLASS = "GroupIDMember";

  public static final String  BASIC_CLASS = "Basic";

  public static final String  ELIGIBLEPARTY_CLASS = "EligibleParty";
          
  public static final String  ACS_CLASS = "ACS";

  public static final String  ACSFACTORY_CLASS = "ACSFactory";

  public static final String  VIRTUALFOLDER_CLASS = "VirtualFolder";

  public static final String  PRIVILEGEABST_CLASS = "PrivilegeAbst";

  public static final String  PRIVILEGEFORTYPE_CLASS = "PrivilegeForType";

  public static final String  PRIVILEGEFORLINKS_CLASS = "PrivilegeForLinks";
          
  public static final String  DIRECTORYEP_CLASS = "DirectoryEP";

  /*
   * List of the generic classes in ARoad0.gBase, like 'Resource' or 'ACS', without
   * the 'Impl' term, that may be used in an AcsAddon as keys to define the
   * classic classes which replace these generic classes. Used in the ACS vocabulary.
   * The classic classes do not include the embedded classes like CAliasImpl,
   * since they may be subclassed in the AcsAddons but are not used outside the AcsAddon
   * gBase package.
   */
  public static final List    L_CLASSIC_CLASS_KEYS = new ArrayList(20);

  /*
  * For an AcsAddon vocabulary, this is the key to define in the ACS vocabulary
  * the list of the extension classes in the AcsAddon gBase package. This list complements
  * the classic classes which simply replace some generic gBase classes.
  * The class descriptions are separated by one space,
  * like in 'Executable/is/ACTOR_CLASS XXX/is/RESOURCE_CLASS'. This says that
  * there is a class Executable'AcsAddon_Name'Impl under the AcsAddon gBase package, and
  * that this class is a direct subclass of the ACTOR_CLASS. This ACTOR_CLASS may
  * be the generic Actor class, or may be the classic class in the same AcsAddon package
  * that replaces the generic Actor class. For instance, in the AcsAddon Ubuntu, the classic class
  * ActorUbuntuImpl replaces the generic ActorImpl class, and the extension class
  * ExecutableUbuntuImpl is a subclass of the classic class ActorUbuntuImpl.
  */
  public static final String  ACSADDON_EXTENSION_CLASSES = "AcsAddon new classes";

  /*
  * For an Acs vocabulary, this is the key to define the AcsAddon name.
  */
  public static final String  ACSADDON_NAME = "AcsAddon name";

  /*
  * For an Acs vocabulary, this is the key to define the AcsAddon editor.
  * The recommanded form for the value is 'ACCBEE, FRANCE, EU'.
  */
  public static final String  ACSADDON_EDITOR = "ACS editor";

  /*
  * For an Acs vocabulary, this is the key to define the AcsAddon package editor,
  * that is derived from the ACS editor. The recommanded form for the value is 'Accbee'.
  */
  public static final String  ACSADDON_PACKAGE_EDITOR = "ACS package editor";

  /*
  * For an Acs vocabulary, this is the key to define the AcsAddon version.
  * The form of the value is 'a.b.cd', like '0.7.1'.
  */
  public static final String  ACSADDON_VERSION = "ACS version";

  /*
  * For an AcsAddon vocabulary, this property is the key to define the specialized right domains.
  * The values set the non-generic rights in the ACSObjects on which 
  * the ACS constraints of the specialized rights are applicable,
  * and in that case, what domain is applicable,
  * like the domain 'GpORightsSet' for the property 'OtherRights' in Ubuntu.
  * The value must have the format 'PropertyName.SpecializedRightDomainName'.
  * and it is ridden by getL_MapKeysForSpecializedRights().
  */
  public static final String  ACSADDON_SPECIALIZED_RIGHT_DOMAINS = 
          "AcsAddon right domains";

  /*
   * List of the domains for the specialized rights:
   * 'AccRightsSet', 'GpORightsSet', 
   * 'AclRightsSet', 'PrvRightsSet', 'BdgRightsSet', 'RooRightsSet'.
   * Used by ACSFactoryImpl.addAcsVocabulary().
   */
  public static final List    L_SPECIALIZED_RIGHT_DOMAINS = new ArrayList(10);

  /**
   * These are the default values for the classes in the generic gBase package
   * that may be subclassed in an AcsAddon package, and are then accessible for all
   * the Access Road program through ARoad0.gBaseInterface.ACSRun.getM_AcsVocabulary().
   * Other gBase classes like BaseUtilityImpl are not registered if they are subclassed
   * in an AcsAddon, and only the classes of their package or a derived package may call them.
   */
  public static final HashMap<String,String[]> M_GENERIC_ACS_VOCABULARY = new HashMap(25);

  // --------------------- static initialization -----------------------------
  static {
    L_ACS_TREE_GENERIC_UPPER_NODES.add(ACTORS);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(USERS);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(GROUPS);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(RESOURCES);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(ACLENTRIES);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(EEPARTIES);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(ECRESOURCES);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(ACSRIGHTS);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(RESVFOLDERS);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(EPVFOLDERS);
    L_ACS_TREE_GENERIC_UPPER_NODES.add(PRIVILEGES);

    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(ACTORS_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(USERS_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(GROUPS_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(RESOURCES_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(ACLENTRIES_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(EEPARTIES_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(ECRESOURCES_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(ACSRIGHTS_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(RESVFOLDERS_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(EPVFOLDERS_FOR_THIS_ACS);
    L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS.add(PRIVILEGES_FOR_THIS_ACS);

    L_CLASSIC_CLASS_KEYS.add(ACTOR_CLASS);
    L_CLASSIC_CLASS_KEYS.add(USER_CLASS);
    L_CLASSIC_CLASS_KEYS.add(GROUP_CLASS);
    L_CLASSIC_CLASS_KEYS.add(RESOURCE_CLASS);
    L_CLASSIC_CLASS_KEYS.add(ACLENTRY_CLASS);
    L_CLASSIC_CLASS_KEYS.add(DIRECTORY_CLASS);
    L_CLASSIC_CLASS_KEYS.add(GROUPMEMBER_CLASS);
    L_CLASSIC_CLASS_KEYS.add(BASIC_CLASS);
    L_CLASSIC_CLASS_KEYS.add(ELIGIBLEPARTY_CLASS);
    L_CLASSIC_CLASS_KEYS.add(ACS_CLASS);
    L_CLASSIC_CLASS_KEYS.add(ACSFACTORY_CLASS);
    L_CLASSIC_CLASS_KEYS.add(VIRTUALFOLDER_CLASS);
    L_CLASSIC_CLASS_KEYS.add(PRIVILEGEABST_CLASS);
    L_CLASSIC_CLASS_KEYS.add(PRIVILEGEFORTYPE_CLASS);
    L_CLASSIC_CLASS_KEYS.add(PRIVILEGEFORLINKS_CLASS);
    L_CLASSIC_CLASS_KEYS.add(DIRECTORYEP_CLASS);

    M_GENERIC_ACS_VOCABULARY.put(ACTOR_CLASS, new String[] {
        "ARoad0.gBase.ActorImpl"});
    M_GENERIC_ACS_VOCABULARY.put(USER_CLASS, new String[] {
        "ARoad0.gBase.UserIDImpl"});
    M_GENERIC_ACS_VOCABULARY.put(GROUP_CLASS, new String[] {
        "ARoad0.gBase.GroupIDImpl"});
    M_GENERIC_ACS_VOCABULARY.put(RESOURCE_CLASS, new String[] {
        "ARoad0.gBase.ResourceImpl"});
    M_GENERIC_ACS_VOCABULARY.put(ACLENTRY_CLASS, new String[] {
        "ARoad0.gBase.AclEntryImpl"});
    M_GENERIC_ACS_VOCABULARY.put(DIRECTORY_CLASS, new String[] {
        "ARoad0.gBase.DirectoryImpl"});
    M_GENERIC_ACS_VOCABULARY.put(GROUPMEMBER_CLASS, new String[] {
        "ARoad0.gBase.GroupIDMemberImpl"});
    M_GENERIC_ACS_VOCABULARY.put(BASIC_CLASS, new String[] {
        "ARoad0.gBase.BasicImpl"});
    M_GENERIC_ACS_VOCABULARY.put(ELIGIBLEPARTY_CLASS, new String[] {
        "ARoad0.gBase.EligiblePartyImpl"});
    M_GENERIC_ACS_VOCABULARY.put(ACS_CLASS, new String[] {
        "ARoad0.gBase.ACSImpl"});
    M_GENERIC_ACS_VOCABULARY.put(ACSFACTORY_CLASS, new String[] {
        "ARoad0.gBase.ACSFactoryImpl"});
    M_GENERIC_ACS_VOCABULARY.put(VIRTUALFOLDER_CLASS, new String[] {
        "ARoad0.gBase.VirtualFolderImpl"});
    M_GENERIC_ACS_VOCABULARY.put(PRIVILEGEABST_CLASS, new String[] {
        "ARoad0.gBase.PrivilegeAbstImpl"});
    M_GENERIC_ACS_VOCABULARY.put(PRIVILEGEFORTYPE_CLASS, new String[] {
        "ARoad0.gBase.PrivilegeForTypeImpl"});
    M_GENERIC_ACS_VOCABULARY.put(PRIVILEGEFORLINKS_CLASS, new String[] {
        "ARoad0.gBase.PrivilegeForLinksImpl"});
    M_GENERIC_ACS_VOCABULARY.put(DIRECTORYEP_CLASS, new String[] {
        "ARoad0.gBase.DirectoryEPImpl"});

    M_GENERIC_ACS_VOCABULARY.put(ACSADDON_NAME, new String[] {"<NONE>"});
    M_GENERIC_ACS_VOCABULARY.put(ACSADDON_EDITOR, new String[] {"ACCBEE, FRANCE, EU"});
    M_GENERIC_ACS_VOCABULARY.put(ACSADDON_PACKAGE_EDITOR, new String[] {"Accbee"});
    M_GENERIC_ACS_VOCABULARY.put(ACSADDON_VERSION, new String[] {"0.7.2"});

    L_SPECIALIZED_RIGHT_DOMAINS.add("AccRightsSet");
    L_SPECIALIZED_RIGHT_DOMAINS.add("GpORightsSet");
    L_SPECIALIZED_RIGHT_DOMAINS.add("AclRightsSet");
    L_SPECIALIZED_RIGHT_DOMAINS.add("PrvRightsSet");
    L_SPECIALIZED_RIGHT_DOMAINS.add("BdgRightsSet");
    L_SPECIALIZED_RIGHT_DOMAINS.add("RooRightsSet");
  } // static


    /* Constructors
     */

  /** with this private constructor, the compiler won't generate
  * a default public constructor.
  */
  private ACSFactoryUtilityImpl() {
  }
  
    /**
    * Only method to obtain the unique instance of ACSFactoryUtilityImpl.
    * @return ACSFactoryUtilityImpl
    */
  public static ACSFactoryUtilityImpl getInstance() {
    return instance__;
  } // getInstance()
  
  /**
   * Gets the structure keys for the creation of an ACS, to display to the user.
   * Called by ActionNewACSyst.
   * @deprecated
   * @return a 23-length array defining the acs structure keys.
   */
  public static String[] getAcsStructureKeys()	{
      
      String[] l_keys = new String[22];
      
      l_keys[0] = "resources";//"resource: no resources"; 
      l_keys[1] = "resources trees";//"resource: directories of resources trees";
      l_keys[2] = "accounts";//"EP: accounts";
      l_keys[3] = "groups of accounts";//"EP: groups of accounts";
      l_keys[4] = "groups trees";//"EP: group of groups";
      l_keys[5] = "internal ACLs";//"ACL right: internal ACL";
      l_keys[6] = "external ACLs from/to it";//"ACL right: external ACL";
      l_keys[7] = "actors";//"resource: no actors";
      l_keys[8] = "bridge rights";//"bridge right: no BriDGes";
      l_keys[9] = "second. groups for actors";//"EP 1: secondary groups for actors as right users";   
      l_keys[10] = "some Acc-Group-Other rights";//"AGO right: Account and Group rights";
      l_keys[11] = "NX executables";//"resource: no NX executables";   
      
      l_keys[12] = "EP directories";//"resource: eligible party directories";
      l_keys[13] = "Other rights in AGO rights";//"AGO right: no other rights";
      l_keys[14] = "privilege rights";//"PRI right: no PRI rights";
      l_keys[15] = "virtual folders";//"virtual folder: no virtual folders";
      l_keys[16] = "granting rights";//"access metarule: no granting rights";      
      l_keys[17] = "denying rights";//"access metarule: denying rights"; 
      l_keys[18] = "running on its parent";//"embedded: this ACS is running on the parent ACS"; 
      l_keys[19] = "embedded in its parent";//"embedded: this ACS is running on and is embedded in the parent ACS";     
      l_keys[20] = "acsrights";//"acs: acsrights"; 
      l_keys[21] = "metarights";//"acs: metarights";
      l_keys[22] = "right server";//"acs: this ACS is an authorization store";
      
      return l_keys;
  } // getAcsStructureKeys()  
  
  /**
   * Gets the default structure for the ACS.
   * This is the reference method for the ACS structure, with 36 booleans.
   * Almost all booleans are all set to false.
   * Only 5 properties are activated by default. They are resources, metarights,
   * granting rights, incomplete modelings of the ACS structure and behavior.
   * Called by the ACSFactoryImpl constructor.
   * @return the ACS structure as a map with standardized keys and default values.
   */
  public static SortedMap<String,Boolean> getDefaultAcsStructure() {

      SortedMap<String,Boolean> m_Structure = new TreeMap<String,Boolean>();

      // activated by default
      m_Structure.put("resource: no resources", new Boolean(false) );
      m_Structure.put("resource: directories of resources trees",
              new Boolean(false));
      m_Structure.put("EP: accounts", new Boolean(false));
      m_Structure.put("EP: groups of accounts", new Boolean(false));
      m_Structure.put("EP: group of groups", new Boolean(false));
      m_Structure.put("ACL right: internal ACL", new Boolean(false));
      m_Structure.put("ACL right: external ACL with entity in this acs",
              new Boolean(false));
      m_Structure.put("ACL right: pure external ACL without entity in this acs",
              new Boolean(false));
      m_Structure.put("resource: no actors", new Boolean( true));
      m_Structure.put("bridge right: no BriDGes", new Boolean(true));
      m_Structure.put("EP 1: secondary groups for actors as right users",
              new Boolean(false));
      m_Structure.put("EP: secondary groups for actors, excluding area groups",
              new Boolean(false)); 
      m_Structure.put("AGO right: Account and Group rights",
              new Boolean(false));
      m_Structure.put("resource: eligible party directories", new Boolean(false));
      m_Structure.put("PRI right: no PRI rights", new Boolean(true));
      m_Structure.put("virtual folder: no virtual folders", new Boolean(true));
      // activated by default
      m_Structure.put("access metarule: no granting rights", new Boolean(false));
      m_Structure.put("access metarule: denying rights", new Boolean(false));
      m_Structure.put("embedded: this ACS is running on and is tied to the parent ACS",
              new Boolean(false));
      m_Structure.put("acs: acsrights", new Boolean(false));
      // activated by default
      m_Structure.put("acs: metarights", new Boolean(true));
      m_Structure.put("acs: this ACS is an authorization store", new Boolean(false));

      // new in 0.7.0
      // activated by default
      m_Structure.put("access metarule: incomplete modeling of the ACS structure",
              new Boolean(true));
      // activated by default
      m_Structure.put("access metarule: incomplete modeling of the ACS behavior",
              new Boolean(true));

      // caution: the other types of PRI are not initialized
      m_Structure.put("PRI right: simple LPRI", new Boolean(false));
      m_Structure.put("PRI right: simple TPRI", new Boolean(false));
      m_Structure.put("PRI right: no granting rights", new Boolean(true));
      m_Structure.put("PRI right: no denying rights", new Boolean(true));
      m_Structure.put("PRI right: for VirtualFolder", new Boolean(false));
      m_Structure.put("PRI right: for Resource", new Boolean(false));
      m_Structure.put("PRI right: external One-To-One LPRI", new Boolean(false));

      m_Structure.put("virtual folder: no virtual folders",
              new Boolean(true));
      m_Structure.put("virtual folder: no virtual folders of Resources",
              new Boolean(true));
      m_Structure.put("virtual folder: no virtual folders of EP",
              new Boolean(true));
      m_Structure.put("virtual folder: no simple resources in virtual folders of Resources",
              new Boolean(true));
      m_Structure.put("virtual folder: no directories in virtual folders of Resources",
              new Boolean(true));
      m_Structure.put("virtual folder: no directoryEPs in virtual folders of EP",
              new Boolean(true));
      m_Structure.put("virtual folder: no actors in virtual folders of EP",
              new Boolean(true));
      m_Structure.put("virtual folder: no virtual folder trees", 
              new Boolean(true));
      m_Structure.put("virtual folder: a member is no more than one time"+
              "in each virtual folder tree",
              new Boolean(false));
      //m_Structure.put("virtual folder: propagation of target rights to all the members",
        //      new Boolean(false));

      m_Structure.put("ACL inheritance 1: ACL inheritance",
              new Boolean(false));
      m_Structure.put("AGO inheritance 1: AG inheritance",
              new Boolean(false));
      m_Structure.put("LPRI inheritance 1: Linked PRI inheritance",
              new Boolean(false));

      m_Structure.put("ACL right: flexible conditional ACL with "+
          "condition groups the groupIDmember has to be member",
              new Boolean(false));
      m_Structure.put("ACL right: right-defined conditional ACL with "+
            "predefined condition groups the groupIDmember has to be member",
              new Boolean(false));

      return m_Structure;
  } // getDefaultAcsStructure()
  
    /**
    * Creates the set of rules which are derived from the ACS structure.
    * More than 50 rules may be added, and the AcsAddons may complement them.
    * The order of the rules is not important, since the returned list is sorted
    * in ACSFactoryImpl. Used by ACSFactoryImpl.
    * @param _m_Structure is the ACS structure, including the values choosen by the user
    * @param _acsType the ACS type, used for the Linux type
    * @return the list of rules which are applied in the ACS.
    */
  public static List<String> getAcsImmutableRules(SortedMap<String,Boolean> _m_Structure,
          String _acsType) {
      List<String> result = new ArrayList(INITIAL_CAPACITY*20);
      Boolean rule = false;
      Boolean rule2 = false;
      Boolean rule3 = false;

      result.add("* sens metarule 1: no-parent-ACS denying rights and access-deficiencies "+
              "do not delete this ACS granting rights (no exception, no user update)");
      result.add("* sens metarule 1: parent-ACS denying rights and access-deficiencies "+
              "delete this embedded-ACS granting rights (no exception, no user update)");
      result.add("* sens metarule 1: this-ACS denying rights are superior "+
              "to this-ACS granting rights (there are + and – operations), "+
              "for each right type (no exception, no user update)");
      
      rule = _m_Structure.get("ACL right: internal ACL");
      rule2 = _m_Structure.get("ACL right: external ACL with entity in this acs");
      if (rule.booleanValue() || rule2.booleanValue())
          result.add("* sens metarule 2: metarules are applied to ACL rights "+
                  "(no exception, no user update)");
      if (rule.booleanValue())
          result.add("* ACL right: on actors as targets, and for internal ACLs, "+
                  "the received Account/Groups context overlays the current Account/Groups "+
                  "context (no exception, no user update)");
      if (rule2.booleanValue())
          result.add("* ACL right: on actors as targets, and for ACLs with external sources, "+
                  "actors run under their current Account/Groups context "+
                  "(no exception, no user update)");
      if (rule.booleanValue()) {
          result.add("* ACL right: internal AclEntry propagates the Account/Groups context "+
                  "of the ACL source");
      }
      if (rule2.booleanValue()) {
          result.add("* ACL right: external AclEntry propagates the Account/Groups context "+
                  "of the ACL source only if source and target are from the same ACS");
      }
      rule2 = _m_Structure.get("ACL right: pure external ACL without entity in this acs");
      if (rule2.booleanValue()) {
          result.add("* ACL right: external AclEntry propagates the Account/Groups context "+
                  "of the ACL source only if source and target are from the same ACS");
      }
      
      rule = _m_Structure.get("AGO right: Account and Group rights");
      if (rule.booleanValue())
          result.add("* sens metarule 2: metarules are applied to AGO rights "+
                  "(no exception, no user update)");
      result.add("* right type order 1: use a right type order for defining "+
              "the applied rights (no exception, no user update)");
      result.add("* right type order 1: merge all granting rights over all right types,"+
              "and merge all denying rights too, before their comparison");
      
      // reuse 'rule'
      if (rule.booleanValue()) {
          result.add("* right type order 2: for AGO rights, super-admin, "+
                  "then A in AGO, then all other right types (no exception, no user update)");
          result.add("* right type rule 1: for AGO rights, root rights and A rights "+
                  "are not changed by other-right-type rights (no exception, no user update)");
      }
      
      rule = _m_Structure.get( "bridge right: no BriDGes" );
      if ( !rule.booleanValue()) {
          result.add("* bridge right: BriDGes for actors as right users and targets");
          result.add("* bridge right: on actors as targets, actors run under their "+
                  "current account and group, if any (no exception, no user update)");
          result.add("bridge right 1: BriDGes only for actors");
      }

      result.add("EP: user changes root rights");
      
      rule = _m_Structure.get("EP: accounts");
      rule2 = _m_Structure.get("resource: no actors");
      if (rule.booleanValue() && !rule2.booleanValue()) {
          result.add("* EP 1: current account for actors as right users");
          result.add("* EP 2: current account updating changes the current group "+
                  "and the secondary groups for actors as right users");
      }

      rule = _m_Structure.get("EP: groups of accounts");
      rule2 = _m_Structure.get("resource: no actors");
      if (rule.booleanValue() && !rule2.booleanValue())
          result.add("* EP 1: current group for actors as right users");
      
      rule = _m_Structure.get("AGO right: Account and Group rights");
      if (rule.booleanValue()) {
          result.add("resource 1: a resource has only one account (no exception, no user update)");
          result.add("right mask 1: denying rights masks (no exception, no user update)");
          result.add("right mask 2: new child/node applies account denying rights umask "+
                  "on AGO rights (no exception, no user update)");
          result.add("AGO right 1: A rights overlay G rights (no exception, no user update)");
      }

      // reuse 'rule'
      rule2 = _m_Structure.get("resource: directories of resources trees");
      rule3 = _m_Structure.get("AGO inheritance 1: AG inheritance");
      if (rule.booleanValue() && rule2.booleanValue() && rule3.booleanValue()) {
          result.add("right access 1: child/node access only if "+
                     "the current parent tree rights allow to go across it");

          result.add("resource 1: new child/node inherits the current G, if activated,"+
                         " and G is not the account primary group");
      }

      rule2 = _m_Structure.get("resource: no actors");
      if (rule.booleanValue() && !rule2.booleanValue()) {
          result.add("EP 2: primary group for account");
      }

      rule = _m_Structure.get("EP: group of groups");
      if (rule.booleanValue())
          result.add("EP: group of groups");

      rule = _m_Structure.get("EP 1: secondary groups for actors as right users");
      if (rule.booleanValue())
          result.add("EP 1: secondary groups for actors as right users");

      rule = _m_Structure.get("virtual folder: a member is no more than one time"+
              "in each virtual folder tree");
      if (rule.booleanValue())
          result.add("virtual folder: a member is no more than one time"+
              "in each virtual folder tree");

      rule = _m_Structure.get("ACL inheritance 1: ACL inheritance");
      rule2 = _m_Structure.get("EP: group of groups");
      if (rule.booleanValue()) {
          result.add("* right inheritance in the Resources tree: no more than one of "+
                  "the three ACL/AGO/LPRI rights inheritance may be activated");
          result.add("ACL right inheritance 1: child/node inherits rights from "+
                  "the direct and indirect parent");
          result.add("ACL right inheritance 1: if activated and no access from the inherited rights, "+
                  "no ACL access to the child");
          result.add("ACL right inheritance 2: at each access of an account, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  "first the account, second its primary group, and third the account groups");
          result.add("ACL right inheritance 2: at each access of a group, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  "first the group, second the group groups");
          if (rule2.booleanValue())
                  result.add("EP: group of groups on N levels, with search of the "+
                          "group inherited rights only for the first is_member level");
          result.add("ACL right inheritance 2: at each access of an actor, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  " first the actor, second the account of its AG context,"+
                  " third the group of its AG context,"+
                  " fourth the secondary groups of the actor");
      }

      rule = _m_Structure.get("AGO inheritance 1: AG inheritance");
      if (rule.booleanValue()) {
          result.add("* right inheritance in the Resources tree: no more than one of "+
                  "the three ACL/AGO/LPRI rights inheritance may be activated");
          result.add("AGO right inheritance 1: child/node inherits rights from "+
                  "the direct and indirect parents");
          result.add("AGO right inheritance 1: if activated and no access from the inherited rights, "+
                  "no AGO access to the child if and only if the account or the group is not null");
          result.add("AGO right inheritance 2: at each access of an account, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  "first the account, second its primary group, and third the account groups");
          result.add("AGO right inheritance 2: at each access of a group, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  "first the group, second the group groups");
          if (rule2.booleanValue())
                  result.add("EP: group of groups on N levels, with search of the "+
                          "group inherited rights only for the first is_member level");
          result.add("AGO right inheritance 2: at each access of an actor, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  " first the account of its AG context, second the group of its AG context,"+
                  " third the secondary groups of the actor");
      }

      rule = _m_Structure.get("PRI right: no PRI rights");
      if (!rule.booleanValue()) {
          result.add("* PRI right: privilege does not propagate the Account/Groups context "+
                  "of the privilege source");
      }

      rule = _m_Structure.get("LPRI inheritance 1: Linked PRI inheritance");
      if (rule.booleanValue()) {
          result.add("* right inheritance in the Resources tree: no more than one of "+
                  "the three ACL/AGO/LPRI rights inheritance may be activated");
          result.add("LPRI right inheritance 1: child/node inherits rights from "+
                  "the direct and indirect parents");
          result.add("LPRI right inheritance 1: if activated and no access from the inherited rights, "+
                  "no LPRI access to the child");
          result.add("LPRI right inheritance 2: at each access of an account, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  "first the account, second its primary group, and third the account groups");
          result.add("LPRI right inheritance 2: at each access of a group, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  "first the group, second the group groups");
          if (rule2.booleanValue())
                  result.add("EP: group of groups on N levels, with search of the "+
                          "group inherited rights only for the first is_member level");
          result.add("LPRI right inheritance 2: at each access of an actor, the first"+
                  " eligible party having inherited rights set them, and the search order is,"+
                  " first the actor, second the account of its AG context,"+
                  " third the group of its AG context,"+
                  " fourth the secondary groups of the actor");
      }

      rule = _m_Structure.get("virtual folder: no virtual folders");
      rule2 = _m_Structure.get("virtual folder: propagation of target rights to all the members");
      if (!rule.booleanValue() && rule2 != null && rule2.booleanValue()) {
          result.add("virtual folder: propagation of target rights to all the members");
      }
      else if (!rule.booleanValue()) {
          result.add("virtual folder: no propagation of target rights to the members");
      }

      rule2 = _m_Structure.get("virtual folder: no virtual folders of EP");
      if (!rule.booleanValue() && !rule2.booleanValue()) {
          result.add("* virtual folder: as source, an access to a Resource that is not "+
                  "a virtual member, is always through a virtual member");
      }

      return result;
  } // getAcsImmutableRules() 
  
    /**
    * This important method contains all the basic integrity rules
    * the ACS structure has to comply with in all cases.
    * It controls the ACS integrity by the mean of comparisons of the ACS properties,
    * to validate the choices before the ACS creation.
    * An user-friendly message is returned when an error is found.
    * About 65 controls are provided, each with its user messsage.
    * Used by the ACS constructor and by ACSFactoryImpl.
    * @param _m_structure is the ACS structure, including the values choosen by the user.
    *   Never null.
    * @return null if the integrity rules are followed, or
    *   an user message which explains the first integrity error.
    */
  public static String controlAcsStructure(SortedMap<String,Boolean> _m_structure) {
      String result = null;
      
      if (
          _m_structure.get("resource: no resources") &&
          _m_structure.get("resource: directories of resources trees")              
          // resources if resource trees      
      ) {
          result = "in the ACS constructor, the resources must be"+
            " managed when the resource trees are.";
      }

      else if (
          !_m_structure.get("EP: accounts") &&
          _m_structure.get("EP: groups of accounts")
          // accounts if groups     
      ) {
          result = "in the ACS constructor, the accounts must be"+
            " managed when the groups are.";
      }

      else if (
          !_m_structure.get("EP: groups of accounts") &&
          _m_structure.get("EP: group of groups")
          // groups if group trees      
      ) {
          result = "in the ACS constructor, the groups must be"+
            " managed when the group trees are.";
      }

      else if (
          _m_structure.get("resource: no resources") &&
          _m_structure.get("ACL right: internal ACL")             
          // resources if acls      
      ) {
          result = "in the ACS constructor, the resources must be"+
            " managed when the internal AclEntries are.";
      }

      else if (
          (!_m_structure.get("EP: accounts") &&
          !_m_structure.get("EP: groups of accounts") &&
          _m_structure.get("resource: no actors")) &&
          _m_structure.get("ACL right: internal ACL")               
              // users or groups or actors if internal acls       
      ) {
          result = "in the ACS constructor, the users, the groups or the actors must be"+
            " managed when the internal AclEntries are.";
      }

      else if (
          _m_structure.get("ACL inheritance 1: ACL inheritance") &&
          !_m_structure.get("resource: directories of resources trees")
          // resources tree if acl inheritance
      ) {
          result = "in the ACS constructor, the resources tree must be"+
            " managed when the ACL inheritance is used.";
      }

      else if (
          _m_structure.get("ACL inheritance 1: ACL inheritance") &&
          !_m_structure.get("ACL right: internal ACL")
          // acl if acl inheritance
      ) {
          result = "in the ACS constructor, the internal ACL must be"+
            " managed when the ACL inheritance is used.";
      }

      else if (
          _m_structure.get("resource: no resources") &&
          !_m_structure.get("resource: no actors")          
              // resources if actors      
      ) {
          result = "in the ACS constructor, the resources must be"+
            " managed when actors are.";
      }

      else if (
          !_m_structure.get("EP: groups of accounts") &&
          (_m_structure.get("EP 1: secondary groups for actors as right users") ||       
          _m_structure.get("EP: secondary groups for actors, excluding area groups"))
              // groups if secondary groups for actors     
      ) {
          result = "in the ACS constructor, the groups must be"+
            " managed when there are secondary groups.";
      }

      else if (
          _m_structure.get("resource: no actors") &&
          (_m_structure.get("EP 1: secondary groups for actors as right users") ||       
          _m_structure.get("EP: secondary groups for actors, excluding area groups"))
              // actors if secondary groups for actors     
      ) {
          result = "in the ACS constructor, the actors must be"+
            " managed when they may have secondary groups.";
      }

      else if (
          _m_structure.get("resource: no actors") &&
          !_m_structure.get("bridge right: no BriDGes")            
              // actors if bridges       
      ) {
          result = "in the ACS constructor, the actors must be"+
            " managed when there are bridges.";
      }

      else if (
          (!_m_structure.get("EP: accounts") &&
          !_m_structure.get("EP: groups of accounts")) &&
          _m_structure.get("AGO right: Account and Group rights")
              // users or groups if AGO rights       
      ) {
          result = "in the ACS constructor, the users or the groups must be"+
            " managed when there are AGO rights.";
      }

      else if (
          _m_structure.get("AGO inheritance 1: AG inheritance") &&
          !_m_structure.get("resource: directories of resources trees")
      ) {
          result = "in the ACS constructor, the resources tree must be"+
            " managed when the Account/Group inheritance is used.";
      }

      else if (
          _m_structure.get("AGO inheritance 1: AG inheritance") &&
          !_m_structure.get("AGO right: Account and Group rights")
          // AGO if AGO inheritance
      ) {
          result = "in the ACS constructor, the AG rights must be"+
            " managed when the Account/Group inheritance is used.";
      }

      else if ( 
          !_m_structure.get("acs: acsrights") &&
          !_m_structure.get("acs: metarights")           
              // acs or meta rights       
      ) {
          result = "in the ACS constructor, the acs must manage"+
            " metarights or manage acsrights.";
      }

      // new in 0.7.0
      else if (_m_structure.get("PRI right: no granting rights") &&
              _m_structure.get("PRI right: no denying rights") &&
              !_m_structure.get("PRI right: no PRI rights")) {
          result = "in the ACS constructor, the granting and the denying rights in privileges"+
                  "are not set, while the privileges are managed.";
      }

      else if (!_m_structure.get("PRI right: no granting rights") &&
              _m_structure.get("PRI right: no PRI rights")) {
          result = "in the ACS constructor, the granting rights in privileges are set,"+
                      " while the privileges are not managed.";
      }
      
      else if (!_m_structure.get("PRI right: no denying rights") &&
              _m_structure.get("PRI right: no PRI rights")) {
          result = "in the ACS constructor, the denying rights in privileges are set,"+
                      " while the privileges are not managed.";
      }
      
      else if (_m_structure.get("PRI right: simple TPRI") &&
              _m_structure.get("PRI right: no PRI rights")) {
          result = "in the ACS constructor, the simple typed privileges are set,"+
                      " while the privileges are not managed.";
      }
      
      else if (_m_structure.get("PRI right: simple LPRI") &&
              _m_structure.get("PRI right: no PRI rights")) {
          result = "in the ACS constructor, the simple linked privileges are set,"+
                      " while the privileges are not managed.";
      }
      
      // caution: the other types of TPRI and LPRI are not managed nand not initialized
      // in getDefaultAcsStructure()
      else if (!_m_structure.get("PRI right: simple TPRI") &&
              !_m_structure.get("PRI right: simple LPRI") &&
              !_m_structure.get("PRI right: no PRI rights")) {
          result = "in the ACS constructor, the simple linked and simple typed privileges are not set,"+
                      " while the privileges are managed. This version does not handle"+
                      " the other types of privileges.";
      }

      else if (
           _m_structure.get("PRI right: no PRI rights") &&
          _m_structure.get("PRI right: for VirtualFolder")
      ) {
           result = "in the ACS constructor, the privileges are not managed,"+
                      " while the privileges for virtual folders are used.";
      }

      else if (
           _m_structure.get("PRI right: no PRI rights") &&
          _m_structure.get("PRI right: for Resource")
      ) {
           result = "in the ACS constructor, the privileges are not managed,"+
                      " while the privileges for resources are used.";
      }

      else if (
           _m_structure.get("resource: no resources") &&
          _m_structure.get("PRI right: for Resource")
      ) {
           result = "in the ACS constructor, the resources are not managed,"+
                      " while the privileges for resources are used.";
      }

      else if (
          !_m_structure.get("PRI right: no PRI rights") &&
          !_m_structure.get("PRI right: for Resource") &&
          !_m_structure.get("PRI right: for VirtualFolder") &&
          !_m_structure.get("PRI right: external One-To-One LPRI")
      ) {
           result = "in the ACS constructor, the privileges are managed,"+
                      " while they are not used for the virtual folders, for the resources "+
                      "nor for the external One-To-One LPRI.";
      }

      else if (
          (_m_structure.get("ACL right: external ACL with entity in this acs") ||
          _m_structure.get("ACL right: pure external ACL without entity in this acs")) &&
          _m_structure.get("PRI right: external One-To-One LPRI")
      ) {
           result = "in the ACS constructor, the external one-to-one privileges and"+
                      " the external AclEntries cannot be managed together. ";
      }

      else if (
          _m_structure.get("LPRI inheritance 1: Linked PRI inheritance") &&
          _m_structure.get("PRI right: for Resource") &&
          !_m_structure.get("resource: directories of resources trees")
      ) {
          result = "in the ACS constructor, the resources tree must be"+
            " managed when the linked privilege inheritance is used for resources.";
      }

      else if (
          _m_structure.get("LPRI inheritance 1: Linked PRI inheritance") &&
          !_m_structure.get("PRI right: simple LPRI")
      ) {
          result = "in the ACS constructor, the linked privilege rights must be"+
            " managed when the linked privilege inheritance is used.";
      }

      else if (
           _m_structure.get("virtual folder: no virtual folders") &&
           _m_structure.get("PRI right: for VirtualFolder")
      ) {
           result = "in the ACS constructor, the virtual folders are not used,"+
                      " while the privileges for virtual folders are managed.";
      }

      else if (_m_structure.get("virtual folder: no virtual folders of EP") &&
              _m_structure.get("virtual folder: no virtual folders of Resources") &&
           !_m_structure.get("virtual folder: no virtual folders")
      ) {
           result = "in the ACS constructor, the virtual folders of Righ User Resources"+
                   "and of simple Resources are not used,"+
                      " while the virtual folders are used.";
      }

      else if (!_m_structure.get("virtual folder: no virtual folders of EP") &&
              _m_structure.get("virtual folder: no virtual folders")) {
            result = "in the ACS constructor, the EP virtual folders are used,"+
                      " while the virtual folders are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no virtual folders of EP") &&
              !_m_structure.get("EP: accounts") &&
              !_m_structure.get("EP: groups of accounts") &&
              _m_structure.get("resource: no actors")) {
            result = "in the ACS constructor, the EP virtual folders are used,"+
                      " while the eligible parties are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no virtual folders of Resources") &&
              _m_structure.get("virtual folder: no virtual folders")) {
             result = "in the ACS constructor, the resource virtual folders are used,"+
                      " while the virtual folders are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no virtual folders of Resources") &&
              _m_structure.get("resource: no resources")) {
             result = "in the ACS constructor, the resource virtual folders are used,"+
                      " while the resources are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no simple resources in virtual folders of Resources") &&
              _m_structure.get("virtual folder: no virtual folders of Resources")) {
             result = "in the ACS constructor, the simple resource virtual folders are used,"+
                      " while the resource virtual folders are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no simple resources in virtual folders of Resources") &&
              _m_structure.get("resource: no resources")) {
             result = "in the ACS constructor, the simple resource virtual folders are used,"+
                      " while the resources are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no directories in virtual folders of Resources") &&
              _m_structure.get("virtual folder: no virtual folders of Resources")) {
             result = "in the ACS constructor, the directories in resource virtual folders are used,"+
                      " while the resource virtual folders are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no directories in virtual folders of Resources") &&
              !_m_structure.get("resource: directories of resources trees")) {
             result = "in the ACS constructor, the directories in resource virtual folders are used,"+
                      " while the directories are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no directoryEPs in virtual folders of EP") &&
              _m_structure.get("virtual folder: no virtual folders of EP")) {
             result = "in the ACS constructor, the rights user directories in EP virtual folders are used,"+
                      " while the EP virtual folders are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no directoryEPs in virtual folders of EP") &&
              !_m_structure.get("resource: eligible party directories")) {
             result = "in the ACS constructor, the rights user directories in EP virtual folders are used,"+
                      " while the rights user directories are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no actors in virtual folders of EP") &&
              _m_structure.get("virtual folder: no virtual folders of EP")) {
             result = "in the ACS constructor, the actors in EP virtual folders are used,"+
                      " while the EP virtual folders are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no actors in virtual folders of EP") &&
              _m_structure.get("resource: no actors")) {
             result = "in the ACS constructor, the actors in EP virtual folders are used,"+
                      " but the actors are not managed.";
      }
      
      else if (!_m_structure.get("virtual folder: no virtual folder trees") &&
              _m_structure.get("virtual folder: no virtual folders")) {
             result = "in the ACS constructor, the trees of virtual folders are used,"+
                      " while the virtual folders are not managed.";
      }

      else if (_m_structure.get("virtual folder: a member is no more than one time"+
                        "in each virtual folder tree") &&
              _m_structure.get("virtual folder: no virtual folder trees")) {
             result = "in the ACS constructor, the number of a member in a tree"+
                     "of virtual folders is controlled,"+
                      " while the virtual folder trees are not managed.";
      }
      
      else if (_m_structure.get("virtual folder: propagation of target rights to all the members") != null &&
              _m_structure.get("virtual folder: propagation of target rights to all the members") &&
              _m_structure.get("virtual folder: no virtual folders")) {
             result = "in the ACS constructor, the propagation of target rights is used,"+
                      " while the virtual folders are not managed.";
      } 

      else if (
          _m_structure.get("LPRI inheritance 1: Linked PRI inheritance") &&
          _m_structure.get("PRI right: for VirtualFolder") &&
          _m_structure.get("virtual folder: no virtual folder trees")
      ) {
          result = "in the ACS constructor, the virtual folders tree must be"+
            " managed when the linked privilege inheritance is used for virtual folders.";
      }

      else if (
          _m_structure.get("LPRI inheritance 1: Linked PRI inheritance") &&
          _m_structure.get("PRI right: for Resource") &&
          _m_structure.get("AGO inheritance 1: AG inheritance")
      ) {
          result = "in the ACS constructor, it is not possible to manage both"+
            " the Account/Group inheritance and the linked privilege inheritance"+
            " in the Resources tree.";
      }
      
      else if (
          _m_structure.get("ACL inheritance 1: ACL inheritance") &&
          _m_structure.get("AGO inheritance 1: AG inheritance")
      ) {
          result = "in the ACS constructor, it is not possible to manage both"+
            " the ACL inheritance and the Account/Group inheritance"+
            " in the Resources tree.";
      }
      
      else if (
          _m_structure.get("ACL inheritance 1: ACL inheritance") &&
          _m_structure.get("LPRI inheritance 1: Linked PRI inheritance") &&
          _m_structure.get("PRI right: for Resource")
      ) {
          result = "in the ACS constructor, it is not possible to manage both"+
            " the ACL inheritance and the linked privilege inheritance"+
            " in the Resources tree.";
      }

      else if (
          !_m_structure.get("ACL right: internal ACL") &&
          !_m_structure.get("ACL right: external ACL with entity in this acs") &&
          _m_structure.get("ACL right: flexible conditional ACL with "+
          "condition groups the groupIDmember has to be member")
      ) {
          result = "in the ACS constructor, it is not possible to manage"+
            " flexible conditional AclEntries if the required AclEntries"+
            " are not managed.";
      }

      else if (
          !_m_structure.get("ACL right: internal ACL") &&
          !_m_structure.get("ACL right: external ACL with entity in this acs") &&
          _m_structure.get("ACL right: right-defined conditional ACL with "+
            "predefined condition groups the groupIDmember has to be member")
      ) {
          result = "in the ACS constructor, it is not possible to manage"+
            " right-defined conditional AclEntries if the required AclEntries"+
            " are not managed.";
      }


      else if (
          _m_structure.get("ACL right: flexible conditional ACL with "+
          "condition groups the groupIDmember has to be member") &&
          _m_structure.get("ACL right: right-defined conditional ACL with "+
            "predefined condition groups the groupIDmember has to be member")
      ) {
          result = "in the ACS constructor, it is not possible to manage both"+
            " right-defined conditional AclEntries and flexible conditional AclEntries.";
      }

      return result;
  } // controlAcsStructure() 

  /**
  * Gets the types of the resources or groups subACS,
  * with the values from the key 'Explorer.SubAcsNode'
  * in the map of ACSImpl.getM_TypesAndGUIPolicy().
  * @param    _acs to process
  * @return   the types with the format 'ResourceType.byAcsCreator XXX' or
  * 'GroupType.byAcsCreator XXX'. May be null but not empty.
  */
  public static final List getL_SubAcsToDisplay(ACS _acs) {
    String[]                l_val = null;
    List                    l_result = null;
    Map                     map = _acs.getM_TypesAndGUIPolicy();

    if (map != null ) {
          l_val = (String[]) map.get("Explorer.SubAcsNode");
          if (l_val != null) {
              l_result = new ArrayList(INITIAL_CAPACITY);
              for (int i = 0; i < l_val.length; i++) {
                  l_result.add(l_val[i]);
              } // for i
          } // if
    } // if

    l_val = null;
    map = null;
    return l_result;
  } // getL_SubAcsToDisplay()

  /**
  * Gets the types of the groups tree nodes,
  * with the values from the key 'Explorer.AcsGroupTree'
  * in the map of ACSImpl.getM_TypesAndGUIPolicy().
  * @param _acs to process
  * @return the types having the format 'GroupType.&lt;byAcsCreator&gt; XXX'.
   * May be null but not empty.
  */
  public static final List getL_GroupTreesToDisplay(ACS _acs) {
    String[]                l_val = null;
    List                    l_result = null;
    Map                     map = _acs.getM_TypesAndGUIPolicy();

    if (map != null ) {

                    // for testing and tracking
                    //System.out.println("ACSFactoryUtilityImpl.getL_GroupTreesToDisplay(), "+
                      //      "for the ACS "+_acs.getName()+
                        //    ", use of _acs.getM_TypesAndGUIPolicy() for 'Explorer.AcsGroupTree'");

          l_val = (String[]) map.get("Explorer.AcsGroupTree");
          if (l_val != null) {
              if (l_result == null)
                  l_result = new ArrayList(INITIAL_CAPACITY);
              for (int i = 0; i < l_val.length; i++) {
                  l_result.add(l_val[i]);

                        // for testing and tracking
                        //System.out.println("         adds: "+l_val[i]);

              } // for i
          } // if
          else {

                    // for testing and tracking
                    //System.out.println("         no value for 'Explorer.AcsGroupTree' ");
          }
    } // if
    else {

                    // for testing and tracking
                    //System.out.println("ACSFactoryUtilityImpl.getL_GroupTreesToDisplay() "+
                      //      "without _acs.getM_TypesAndGUIPolicy() in the ACS "+_acs.getName());
    }

    l_val = null;
    map = null;
    return l_result;
  } // getL_GroupTreesToDisplay()
    
      /**
    * Called by BaseManagerImpl during the base initialization 
    * to create the immutable set of Access Road metarights.
    * Defines around 60 granting metarights, and their denying 'deny_' opposites, if any.
    * '|unknown_rights|' is the less powerful right, with a strenght at 1 and no opposite right.
    * '|is_superadmin|' is the more powerful right, with a strenght at 90900 and no opposite right. <P>
    * The metaright strengths are from 10000 to 99999, excepted for '|unknown_rights|' with
    * a strength of 1.
    * Only AclEntries and Privileges may use negative rights. Then, the getPositive() or
    * the getPositiveRight() method allows to know if the object grants or denies its rights.
    * An alternative of the denying rights is the negative right management,
    * like for instance the 'umask' function in a Linux Ubuntu AcsAddon. <P>
    * The power and the field of each metaright are defined.
    * The upper metarights have never a target class more restricted than the created right.
    * Two metarights are never identical with different target classes.
    * NFS4 SYNCHRONIZE right (it allows synchronous readings and writings) is not defined
    * since it is not an access right. <P>
    * The metarights list includes the six general 'do_action_xxx' with their denying rights.
    * This method defines the administrative right |read_security_descriptor| and its opposite.
    * The pattern 'deny_' for names is not followed for the 11 administrative right
    * '|modify_from_parent_acs|', '|modify_security_descriptor|', '|read_from_parent_acs|',
    * '|is_superadmin|', '|unknown_rights|', '|get_target_rights|', '|unverified_rights|',
    * |transfer_all_rights_to_third_party|, |transfer_limited_rights_to_third_party|,
    * |transfer_all_rights|, |transfer_limited_rights|
    * which have no opposite rights, and for 'deny_all' that is the opposite of 'goto'. <P>
    * BaseUtilityImpl.withAcrossToDirectoryAGORights() returns true if
    * the Directory rights contain an equivalent to the metarights
    * 'gothrough_for_directory', 'read_for_directory' or 'full_control'. In an AcsAddon,
    * this method may be overriden. <P>
    * Caution: gWork.FactoryUtilities.M_RIGHTS_ACRONYMS should be updated after
    * any change in this list of administrative metarights, and
    * gBase.BaseUtilityImpl.withAcrossToDirectoryAGORights() uses some metarights. <P>
    * The pattern 'directory/child' in names is used for the AGO inherited rights in
    * BaseUtilityImpl.getChildRightsSetFromParentAGORights(). The rules are: <P>
    * - if a metaright is 'Rdirectory', this is a right for Nodes, and it must
    * have an image 'Rchild', <P>
    * - if the metaright 'Rchild' has the lower rights 'Achild' and 'Bchild'
    * ending with 'child', then the metarights 'A' and 'B' must be defined. <P>
    * - then, BaseUtilityImpl.getChildRightsSetFromParentAGORights() detects
    * the rights 'A' and 'B' as AGO inherited rights of the leaf, if its parent
    * has the right 'Rdirectory'. <P>
    * - 'Rchild, 'Achild' and 'Bchild' have to be desactivated rights,
    * and they cannot never be used directly in an ACS. <P>
    * - if 'A' is a metaright having an ACS right, getChildRightsSetFromParentAGORights()
    * returns this ACS right in place of 'A'.
    * BaseUtilityImpl.getChildRightsSetFromParentAGORights() may be overriden in an AcsAddon.
    * @return a set of granting and denying metarights
    */
  static final Collection initializeBasicMetaRights() {
      Map m_result = new HashMap(INITIAL_CAPACITY*10);
      StringMetaRightImpl right;

      try {
          // as unALLOWED in NFS4
          right = new StringMetaRightImpl("|unknown_rights|", null, 1, 97, 99, true);
          right.setComment("when there is no information about the effective rights; "+
                  "this is the less powerful right; this is a metadata added to the declared rights");
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("|unverified_rights|", null, 99999, 97, 99, true);
          right.setComment("when the declared rights are unverified; "+
                  "this is a metadata added to the declared rights");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("|is_superadmin|", null, 90900, 95, 99, true);
          right.setComment("as super-administrator, may perform all the possible operations on the target");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("|modify_from_parent_acs|", null, 90900, 95, 99, true);
          right.setComment("may perform updatings on the target from the parent ACS of"+
                  " the target ACS");
          m_result.put(right.getRight(), right);  

          // as WRITE_OWNER in NFS4, PROXY in MySQL
          right = new StringMetaRightImpl("|modify_owner|", null, 90500, 93, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("|is_superadmin|") );
          right.addUpperRight( (StringMetaRight) m_result.get("|modify_from_parent_acs|") );
          right.setComment("may change the target owner in its ACS");
          m_result.put(right.getRight(), right); 

          // as WRITE_ACL in NFS4          
          right = new StringMetaRightImpl("|modify_security_descriptor|", null, 90400, 92, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("|modify_owner|") );
          right.setComment("may modify all the right descriptors, but not the owner "+
                  "nor the super-administrator.");
          m_result.put(right.getRight(), right);

          // for AccessControlLinkImpl.DATA_EXCHANGE
          right = new StringMetaRightImpl("|get_target_rights|", null, 90300, 92, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("|modify_security_descriptor|") );
          right.setComment("get all the non-administrative rights of the target, in its proper context "+
                  "of execution as an Actor.");
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("|transfer_all_rights_to_third_party|", null, 90300, 92, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("|modify_security_descriptor|") );
        right.setComment("the rights owner may give to or remove from a second owner "+
                  "the right to give to or remove from a third owner "+
                  "all the delegated rights the second owner has received, "+
                  "including the administrative rights like this one");
          m_result.put(right.getRight(), right);

          // as GRANT in SQL
          right = new StringMetaRightImpl("|transfer_limited_rights_to_third_party|", null, 90200, 92, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("|transfer_all_rights_to_third_party|") );
          right.setComment("the rights owner may give to or remove from a second owner "+
                  "the right to give to or remove from a third owner "+
                  "all the delegated non-administrative rights the second owner has received, "+
                  "excluding the administrative rights like this one");
          m_result.put(right.getRight(), right);
 
          right = new StringMetaRightImpl("|transfer_all_rights|", null, 90300, 92, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("|modify_security_descriptor|") );
          right.setComment("the rights owner may delegate to a second owner "+
                  "all the rights he gets, "+
                  "including his administrative rights like this one");
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("|transfer_limited_rights|", null, 90200, 92, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("|transfer_all_rights|") );
          right.setComment("the rights owner may delegate to a second owner "+
                  "all the non-administrative rights he gets, "+
                  "excluding his administrative rights like this one");
          m_result.put(right.getRight(), right);

          // No links between these upper rights and these lower rights,
          // so a subtraction of any negative right to one of these upper rights
          // do not change the upper right. The limit is the strenght of 90000.
          //
          // |read_from_parent_acs| is an exception: it is an administrative right
          // with a strenght inferior to 90000 and no upper/lower link;
          // so, it cannot be reduced.
          //
          // Start of the double network granting/denying rights. 

          // used by gBase.BaseUtilityImpl.withAcrossToDirectoryAGORights()
          right = new StringMetaRightImpl("full_control", null, 90000, 90, 95, true);
          right.setComment("get all the operational, non-administrative rights");
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_full_control", null, 90000, 90, 95, false);
          right.setComment("forbid the full control and allow the lower rights;"+
                  "to forbid all the operational, non-administrative rights, use deny_all");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("full_controlchild", null, 90000, 90, 95, true);
          right.setComment("Desactivated and derived from a parent right by the program, "+
                  "this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_allchild", null, 90000, 90, 95, false);
          right.setComment("Desactivated and derived from a parent right by the program, "+
                  "this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("full_controlnxchild", null, 90000, 90, 95, true);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_allnxchild", null, 90000, 90, 95, false);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("modify_group", 
                  "ARoad0.gBaseInterface.Resource", 85000, 85, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("deny_modify_group", 
                  "ARoad0.gBaseInterface.Resource", 85000, 85, 95, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid this right inverse");
          m_result.put(right.getRight(), right); 
          
          // as PERMIT in Cisco IOS ACL
          right = new StringMetaRightImpl("access_to", 
                  "ARoad0.gBaseInterface.Actor", 70000, 70, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_access_to", 
                  "ARoad0.gBaseInterface.Actor", 70000, 70, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid this right inverse");
          m_result.put(right.getRight(), right); 
          
          // for open metarights.
          // propose up to 6 levels of no-meaning rights if necessary.
          // create 'do_action_one' (lesser right) to 'do_action_six'
          right = new StringMetaRightImpl("do_action_six", 
                  null, 70000, 70, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_do_action_six", 
                  null, 70000, 70, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "this right does not forbid do_action_five");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("do_action_five", 
                  null, 69000, 69, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("do_action_six") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_do_action_five", 
                  null, 69000, 69, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_do_action_six") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "this right does not forbid do_action_four");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("do_action_four", 
                  null, 68000, 68, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("do_action_five") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_do_action_four", 
                  null, 68000, 68, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_do_action_five") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "this right does not forbid do_action_three");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("do_action_three", 
                  null, 67000, 67, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("do_action_four") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_do_action_three", 
                  null, 67000, 67, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_do_action_four") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "this right does not forbid do_action_two");
          m_result.put(right.getRight(), right);          
          
          right = new StringMetaRightImpl("do_action_two", 
                  null, 66000, 66, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("do_action_three") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_do_action_two", 
                  null, 66000, 66, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_do_action_three") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "this right does not forbid do_action_one");
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("do_action_one", 
                  null, 65000, 65, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("do_action_two") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_do_action_one", 
                  null, 65000, 65, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_do_action_two") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right);           
          
          right = new StringMetaRightImpl("execute_&_write", 
                  "ARoad0.gBaseInterface.Actor", 70000, 65, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_execute_&_write",
                  "ARoad0.gBaseInterface.Actor", 70000, 65, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("read_&_write", null, 65000, 65, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_read_&_write", null, 65000, 65, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right); 
          
          // as WRITE_DATA in NFS4
          right = new StringMetaRightImpl("write", null, 60000, 60, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("read_&_write") );
          right.addUpperRight( (StringMetaRight) m_result.get("execute_&_write") );
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_write", null, 60000, 60, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read_&_write") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_execute_&_write") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right);

          // as READ_ACL in NFS4          
          right = new StringMetaRightImpl("|read_security_descriptor|", null, 60000, 60, 99, true);
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("|deny_read_security_descriptor|", null, 60000, 60, 99, false);
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("shutdown", null, 55500, 55, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_shutdown", null, 55500, 55, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right);           
          
          right = new StringMetaRightImpl("write_for_directory", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_write_for_directory", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("write_for_child", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, true);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_write_for_child", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, false);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("execute_for_directory", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_execute_for_directory", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("execute_for_child", 
                  null, 55000, 55, 50, true);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
         
          right = new StringMetaRightImpl("deny_execute_for_child", 
                  null, 55000, 55, 50, false);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("write_for_nxdirectory", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_write_for_nxdirectory", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("write_for_nxchild", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, true);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_write_for_nxchild", 
                  "ARoad0.gBaseInterface.Node", 55000, 55, 50, false);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);
         
          right = new StringMetaRightImpl("viewDeleteRenameCopychild", 
                  null, 53000, 53, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );   
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
         
          right = new StringMetaRightImpl("deny_viewDeleteRenameCopychild", 
                  null, 53000, 53, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
         
          right = new StringMetaRightImpl("viewDeleteRenameCopy", 
                  null, 53000, 53, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );          
          m_result.put(right.getRight(), right); 
         
          right = new StringMetaRightImpl("deny_viewDeleteRenameCopy", 
                  null, 53000, 53, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("deleteRenameCopychild", 
                  null, 52000, 52, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("execute_for_child") );          
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlnxchild") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
         
          right = new StringMetaRightImpl("deny_deleteRenameCopychild", 
                  null, 52000, 52, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_execute_for_child") );          
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allnxchild") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
         
          right = new StringMetaRightImpl("deleteRenameCopy", 
                  null, 52000, 52, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );          
          right.addUpperRight( (StringMetaRight) m_result.get("viewDeleteRenameCopy") );
          m_result.put(right.getRight(), right); 
         
          right = new StringMetaRightImpl("deny_deleteRenameCopy", 
                  null, 52000, 52, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_viewDeleteRenameCopy") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("create", null, 50000, 50, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_create", null, 50000, 50, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right); 

          // as DELETE in NFS4          
          right = new StringMetaRightImpl("delete", null, 50000, 50, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("deleteRenameCopy") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_delete", null, 50000, 50, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_deleteRenameCopy") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right); 
          
          // as ADD_SUBDIRECTORY in NFS4
          right = new StringMetaRightImpl("create_subdir", 
                  "ARoad0.gBaseInterface.Node" , 50000, 50, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("write_for_directory") );
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("deny_create_subdir", 
                  "ARoad0.gBaseInterface.Node" , 50000, 50, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write_for_directory") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right);
          
          // as ADD_FILE in NFS4
          right = new StringMetaRightImpl("createchild", 
                  "ARoad0.gBaseInterface.Node" , 50000, 50, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("write_for_nxchild") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("write_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlnxchild") );
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);   

          right = new StringMetaRightImpl("deny_createchild", 
                  "ARoad0.gBaseInterface.Node" , 50000, 50, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write_for_nxchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allnxchild") );
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
          
          // as DELETE_CHILD in NFS4          
          right = new StringMetaRightImpl("deletechild", 
                  "ARoad0.gBaseInterface.Node", 50000, 50, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("deleteRenameCopychild") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("write_for_nxchild") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("write_for_child") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlnxchild") );
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_deletechild", 
                  "ARoad0.gBaseInterface.Node", 50000, 50, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_deleteRenameCopychild") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write_for_nxchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allnxchild") );
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("delete_tree", 
                  "ARoad0.gBaseInterface.Node", 50000, 50, 60, true);
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_delete_tree", 
                  "ARoad0.gBaseInterface.Node", 50000, 50, 60, false);
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right); 
           
          right = new StringMetaRightImpl("read_&_execute", 
                  "ARoad0.gBaseInterface.Actor", 50000, 35, 30, true);
          right.addUpperRight( (StringMetaRight) m_result.get("execute_&_write") );
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("modify", null, 48000, 45, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("write") );
          m_result.put(right.getRight(), right);   

          right = new StringMetaRightImpl("deny_modify", null, 48000, 45, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right);
          
          // as INSERT in SQL, APPEND_DATA in NFS4
          right = new StringMetaRightImpl("insert_item", 
                  "ARoad0.gBaseInterface.Node", 48000, 45, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("create") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_insert_item", 
                  "ARoad0.gBaseInterface.Node", 48000, 45, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_create") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right); 
          
          // as DELETE in SQL          
          right = new StringMetaRightImpl("delete_item", 
                  "ARoad0.gBaseInterface.Node", 48000, 45, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("delete") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_delete_item", 
                  "ARoad0.gBaseInterface.Node", 48000, 45, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_delete") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("rename", null, 45000, 40, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("deleteRenameCopy") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_rename", null, 45000, 40, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_deleteRenameCopy") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("copy", null, 45000, 40, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("deleteRenameCopy") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_copy", null, 45000, 40, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_deleteRenameCopy") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("copychild", 
                  "ARoad0.gBaseInterface.Node", 45000, 40, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("createchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deleteRenameCopychild") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("write_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlchild") );
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_copychild", 
                  "ARoad0.gBaseInterface.Node", 45000, 40, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_createchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_execute_for_child") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allchild") );
          right.setComment("derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 
          
          // as UPDATE in SQL          
          right = new StringMetaRightImpl("update_item", 
                  "ARoad0.gBaseInterface.Node", 45000, 40, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("insert_item") );
          right.addUpperRight( (StringMetaRight) m_result.get("delete_item") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_update_item", 
                  "ARoad0.gBaseInterface.Node", 45000, 40, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_insert_item") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_delete_item") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right); 
          
          right = new StringMetaRightImpl("renamechild", 
                  "ARoad0.gBaseInterface.Node", 45000, 40, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("createchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deleteRenameCopychild") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("write_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlchild") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_renamechild", 
                  "ARoad0.gBaseInterface.Node", 45000, 40, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_createchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_deleteRenameCopychild") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("deny_write_for_child") ); 
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allchild") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_read_&_execute", 
                  "ARoad0.gBaseInterface.Actor", 35000, 35, 30, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_execute_&_write") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right);

          // as WRITE_ATTIBUTES and WRITE_NAMED_ATTRS in NFS4          
          right = new StringMetaRightImpl("modify_attributes", null, 35000, 35, 95, true);
          right.addUpperRight( (StringMetaRight) m_result.get("modify") );
          right.addUpperRight( (StringMetaRight) m_result.get("renamechild") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_modify_attributes", null, 35000, 35, 95, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_modify") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_renamechild") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right); 

          // as EXECUTE in NFS4, SHOW_VIEW in MySQL
          right = new StringMetaRightImpl("execute", 
                  "ARoad0.gBaseInterface.Actor", 30000, 30, 30, true);
          right.addUpperRight( (StringMetaRight) m_result.get("read_&_execute") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_execute", 
                  "ARoad0.gBaseInterface.Actor", 30000, 30, 30, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read_&_execute") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right);

            // right for gBase.BaseUtilityImpl.withAcrossToDirectoryAGORights()
          right = new StringMetaRightImpl("read_for_directory", 
                  "ARoad0.gBaseInterface.Node", 30000, 55, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_read_for_directory", 
                  "ARoad0.gBaseInterface.Node", 30000, 55, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("read_for_child", 
                  "ARoad0.gBaseInterface.Node", 30000, 55, 50, true);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_read_for_child", 
                  "ARoad0.gBaseInterface.Node", 30000, 55, 50, false);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);

          // as READ_ATTRIBUTES and READ_NAMED_ATTRS in NFS4
          right = new StringMetaRightImpl("read_attributes", null, 25000, 25, 20, true);
          right.addUpperRight( (StringMetaRight) m_result.get("modify_attributes") );
          m_result.put(right.getRight(), right); 

          right = new StringMetaRightImpl("deny_read_attributes", null, 25000, 25, 20, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_modify_attributes") );
          right.setComment("the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right); 
          
          // as REFERENCES in SQL
          right = new StringMetaRightImpl("link_to_other", null, 25000, 25, 40, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_link_to_other", null, 25000, 25, 40, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          m_result.put(right.getRight(), right);
          
          // as SELECT in SQL, READ_DATA in NFS4
          right = new StringMetaRightImpl("read", null, 20000, 20, 20, true);
          right.addUpperRight( (StringMetaRight) m_result.get("read_&_write") );
          right.addUpperRight( (StringMetaRight) m_result.get("read_&_execute") );
          m_result.put(right.getRight(), right);
          
          // |read_from_parent_acs| has no upper/lower link and no opposite; it cannot be reduced
          right = new StringMetaRightImpl("|read_from_parent_acs|", null, 20000, 20, 20, true);
          right.setComment("indicate a global reading ability from the parent ACS; has no upper/lower link and no opposite.");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("deny_read", null, 20000, 20, 20, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read_&_write") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read_&_execute") );
          right.setComment("one of the most powerful metaright, since"+
                  " the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right);
          
          // as LIST_DIRECTORY in NFS4
          right = new StringMetaRightImpl("viewchild", 
                  "ARoad0.gBaseInterface.Node", 15000, 10, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("read") );
          right.addUpperRight( (StringMetaRight) m_result.get("read_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlnxchild") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern"); 
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_viewchild", 
                  "ARoad0.gBaseInterface.Node", 15000, 10, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allnxchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("view", null, 15000, 10, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("read") );
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_view", null, 15000, 10, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read") );
          right.setComment("one of the most powerful metaright, since"+
                  " the strength sens is inverse for negative rights.");
          m_result.put(right.getRight(), right);

            // right for gBase.BaseUtilityImpl.withAcrossToDirectoryAGORights()
          right = new StringMetaRightImpl("gothrough_for_directory", 
                  "ARoad0.gBaseInterface.Node",12000, 55, 50, true);
          right.addUpperRight( (StringMetaRight) m_result.get("full_control") );
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_gothrough_for_directory", 
                  "ARoad0.gBaseInterface.Node", 12000, 55, 50, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_full_control") );
          right.setComment("the strength sens is inverse for negative rights, since for instance "+
                  "deny_full_control does not forbid to read or write");
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("gothrough_for_child", 
                  "ARoad0.gBaseInterface.Node", 12000, 55, 50, true);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);      

          right = new StringMetaRightImpl("deny_gothrough_for_child", 
                  "ARoad0.gBaseInterface.Node", 12000, 55, 50, false);
          right.setComment("Desactivated and derived from a parent right by the program, this is the upper image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("gotochild", null, 11000, 5, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("execute_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("read_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlchild") );          
          right.addUpperRight( (StringMetaRight) m_result.get("full_controlnxchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("gothrough_for_child") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_gotochild", null, 11000, 5, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_execute_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_read_for_child") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allchild") );          
          right.addUpperRight( (StringMetaRight) m_result.get("deny_allnxchild") );
          right.addUpperRight( (StringMetaRight) m_result.get("deny_gothrough_for_child") );
          right.setComment("Desactivated and derived from a parent right, this is the lower image of "+
                  "the children rights, in the AGO inheritance pattern");
          right.setDesactivated(true);
          m_result.put(right.getRight(), right);
          
          right = new StringMetaRightImpl("goto", null, 11000, 5, 60, true);
          right.addUpperRight( (StringMetaRight) m_result.get("view") );
          right.setComment("the lesser positive metaright, since its opposite is the most powerful" +
                  " negative metaright.");
          m_result.put(right.getRight(), right);

          right = new StringMetaRightImpl("deny_all", null, 11000, 5, 60, false);
          right.addUpperRight( (StringMetaRight) m_result.get("deny_view") );
          right.setComment("the most powerful negative metaright, that forbids all"+
                  " the positive rights.");
          m_result.put(right.getRight(), right);
          
      } catch (BaseError b) { 
          throw new InternalError("ACSFactoryUtilityImpl.initializeBasicMetaRights(): "+b.toString());
      } catch (CreateError c) {
          throw new InternalError("ACSFactoryUtilityImpl.initializeBasicMetaRights(): "+c.toString());
      }         
      
      right = null;
      Collection result = new ArrayList( m_result.values() );
      m_result.clear();
      m_result = null;

                  // for testing purpose
            //System.out.println("end of ACSFactoryUtilityImpl.initializeBasicMetaRights(), "+
              //"the result size is: "+result.size());
      
      return result;
  } // initializeBasicMetaRights() 

    /**   
   * Gets the map keys in any ACS which should be used to defined the specialized rights
   * associated to a given ACSObject right property.<P>
   * As map keys, examples of quartets (object.domain.key.subkey) are: <P>
   *     - 'Target.AclRights.Interface.Directory', for ACL rights when the target is a Directory instance <P>
   *     - 'Target.AccRights.Type.script', for Account rights when the target has the type 'script' <P>
   *     - 'Source.AccRights.Type.script', for Account rights when the source has the type 'script' <P>
   * The object is Target or Source to indicate that these rights are for the target
   * (or source) which has the correct interface or type. <P>
   * The domain is one specific right category:
   * 'AccRightsSet' and 'GpORightsSet' for Target only, 
   * 'AclRightsSet', 'PrvRightsSet', 'BdgRightsSet', 'RooRightsSet' 
   * for Source and Target. <P>
   * The key is 'Interface' or 'Type' to indicate that the subkey belongs to one
   * of these categories. <P>
   * The subkey is an interface name or an EligibleParty type (or Resource type if Target),
   * for which the specialized rights have to be used. <P>
   * The order of search is: <P>
   *    - 'Target' keys then 'Source' keys, <P>
   *    - one specific right category keys, <P>
   *    - resource type keys, then EP type keys, then all interfaces with an order
   *      setting by Class.getInterfaces(). <P>
   * Called by BaseUtilityImpl.getL_AcsSpecializedRights().
   * Calls getAcsAddonExtensionSpecializedRightDomain().
   * @param _acs ACS which manages these specialized rights. Never null.
   * @param _l_interfaces is the set of interfaces and super-interfaces
   *    of a BaseObject class. Never null.
   * @param _propertyName the name of the property.Never null.
   * Has to contain 'Rights' since it is a right property.
   * @param _resourceType is the type of the BaseObject as Resource. May be null.
   * @param _EPType is the type of the BaseObject as EligibleParty. May be null.
   * @return the array of the restricted values for the property. Null if there is no
   *   restricted values in _acs for these arguments.
   */
  public static String[] getL_MapKeysForSpecializedRights(
    ImmutableACS _acs, Collection _l_interfaces,
    String _propertyName, String _resourceType, String _EPType) {

      List       l_grounds = new ArrayList(INITIAL_CAPACITY);
      List       l_keys = new ArrayList(INITIAL_CAPACITY);
      List       l_domains = new ArrayList(INITIAL_CAPACITY);
      List       l_result = new ArrayList(INITIAL_CAPACITY);
      //int        index = 0;
      StringBuilder buffer = new StringBuilder(200);
      String     domain = null;

      // construct for Source
      if (_l_interfaces.contains("EligibleParty") || _l_interfaces.contains("AclEntry")) {
          l_grounds.add("Source");

          // construct the domains
          if (_propertyName.equals("RootRights")) l_domains.add("RooRightsSet"); 
          else if (_propertyName.equals("PrivilegeRights") ||
                  _propertyName.equals("OneSourceLinkedPrivilege") ||
                  _propertyName.equals("OneSourceTypedPrivilege"))
                        l_domains.add("PrvRightsSet");
          else if (_propertyName.equals("BridgRights")) l_domains.add("BdgRightsSet");
          else if (_propertyName.equals("AclRights")) l_domains.add("AclRightsSet");
          else {
              domain = getAcsAddonExtensionSpecializedRightDomain(
                                    _acs, _propertyName);
              if (domain != null) l_domains.add(domain);
          } // else

          // construct the keys and subkeys
          if (_EPType != null) {
              l_keys.add("Type."+_EPType);
              //index++;          
          }
          for (Iterator ite = _l_interfaces.iterator(); ite.hasNext();) {
              l_keys.add( "Interface."+(String)ite.next());
              //index++;          
          } // for ite

                            // for testing and tracking
                    //System.out.println("in ACSFactoryUtilityImpl.getL_MapKeysForSpecializedRights(), for Source, l_grounds: ");
                      //for (int i = 0; i < 2; i++) { // Source'
                        //    System.out.println("-------------- "+l_grounds.get(i));
                      //} // for i                
                    //System.out.println("in ACSFactoryUtilityImpl.getL_MapKeysForSpecializedRights(), l_domains: ");
                    //for (int i = 0; i < 2; i++) { // a specific right key
                      //      System.out.println("-------------- "+l_domains.get(i));
                      //} // for i       

          // concatenate the 3 members of the map keys
          for (int i = 0; i < l_grounds.size(); i++) { // 'Target' then 'Resource'
              for (int j = 0; j < l_domains.size(); j++) { // a specific right key
                  for (Iterator ite = l_keys.iterator(); ite.hasNext(); ) { 
                      buffer.append(l_grounds.get(i));
                      buffer.append(".");
                      buffer.append(l_domains.get(j));
                      buffer.append(".");
                      buffer.append((String)ite.next());

                      l_result.add(buffer.toString());

                      buffer.delete(0, buffer.length());
                  } // for ite
              } // for j
          } // for i   
      } // for Source
      
      l_grounds.clear();
      l_keys.clear();
      l_domains.clear();
      
      // construct the grounds for Target
      if (_l_interfaces.contains("Resource") || _l_interfaces.contains("AclEntry") ) {
          l_grounds.add("Target");

          // construct the domains
          if (_propertyName.equals("OwnerRights")) l_domains.add("AccRightsSet");
          else if (_propertyName.equals("GroupRights")) l_domains.add("GpORightsSet");
          else if (_propertyName.equals("AclRights")) l_domains.add("AclRightsSet");
          else if (_propertyName.equals("RootRights")) l_domains.add("RooRightsSet"); 
          else if (_propertyName.equals("PrivilegeRights")  ||
                  _propertyName.equals("OneTargetLinkedPrivilege") ||
                  _propertyName.equals("OneTargetTypedPrivilege"))
                        l_domains.add("PrvRightsSet");
          else if (_propertyName.equals("BridgRights")) l_domains.add("BdgRightsSet");
          else {
              domain = getAcsAddonExtensionSpecializedRightDomain(
                                    _acs, _propertyName);
              if (domain != null) l_domains.add(domain);
          } // else

          // constructs the keys and subkeys
          if (_resourceType != null) {
              l_keys.add("Type."+_resourceType);
              //index++;
          }
          for (Iterator ite = _l_interfaces.iterator(); ite.hasNext();) {
              l_keys.add("Interface."+(String)ite.next());
              //index++;          
          } // for ite

                            // for testing and tracking
                            //if (_propertyName.equals("OneTargetTypedPrivilege")) {
                              //  System.out.println("in ACSFactoryUtilityImpl.getL_MapKeysForSpecializedRights(), for Target, l_grounds: ");
                                //  for (int i = 0; i < l_grounds.size(); i++) { // 'Target'
                                  //      System.out.println("-------------- "+l_grounds.get(i));
                                  //} // for i
                                //System.out.println("in ACSFactoryUtilityImpl.getL_MapKeysForSpecializedRights(), l_domains: ");
                                //for (int i = 0; i < l_domains.size(); i++) { // a specific right key
                                  //      System.out.println("-------------- "+l_domains.get(i));
                                //} // for i
                            //} // if

          // concatenate the 3 members of the map keys
          for (int i = 0; i < l_grounds.size(); i++) { // 'Target'
              for (int j = 0; j < l_domains.size(); j++) { // a specific right key
                  for (Iterator ite = l_keys.iterator(); ite.hasNext(); ) { 
                      buffer.append(l_grounds.get(i));
                      buffer.append(".");
                      buffer.append(l_domains.get(j));
                      buffer.append(".");
                      buffer.append((String)ite.next());

                      l_result.add(buffer.toString());

                      buffer.delete(0, buffer.length());
                  } // for ite
              } // for j
          } // for i
      } // for Target
      
      l_grounds = null;
      l_keys = null;
      l_domains = null;
      buffer = null;
      domain = null;

                                // for testing and tracking
                        //if (_propertyName.equals("OneTargetTypedPrivilege")) {
                          //  System.out.println("in ACSFactoryUtilityImpl."+
                            //        "getL_MapKeysForSpecializedRights(), l_result: ");
                            //for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                              //      System.out.println("-------------- "+ite.next());
                            //} // for i
                        //} // if
      
      return (String[]) l_result.toArray(new String[0]);
  } // getL_MapKeysForSpecializedRights()

    /**
   * Gets the specialized right domain when an AcsAddon defines a property right
   * in its proper extension classes, if there are specific constraints on the allowed values
   * through specialized rights. The alternative is to set the constraints on rights
   * in the standard rights map of the ACS. Otherwise, the ACS vocabulary map contains
   * the key ACSADDON_SPECIALIZED_RIGHT_DOMAINS to get
   * the specialized domain among the values in L_SPECIALIZED_RIGHT_DOMAINS.
   * Called by getL_MapKeysForSpecializedRights().
   * @see #ACSADDON_SPECIALIZED_RIGHT_DOMAINS
   * @param _acs ACS which manages these specialized rights. Never null.
   * @param _propertyName the name of the property. Never null.
   * Has to contain 'Rights' since it is a right property.
   * @return the specialized right domain. Null if there is no matching value.
   */
  public static String getAcsAddonExtensionSpecializedRightDomain(
              ImmutableACS _acs, String _propertyName) {

      String[] l_names = (String[]) _acs.getM_AcsVocabulary().get(
                                        ACSADDON_SPECIALIZED_RIGHT_DOMAINS);
      boolean  error = false;
      int      index = -1;
      String   result = null;

      if (l_names == null || l_names.length == 0) {

                      // for testing and tracking
                      //System.out.println("--- "+ //ACSFactoryUtilityImpl."+
                            //"getAcsAddonExtensionSpecializedRightDomain(() "+
                          //    " do not found the specialized right domain");

              return null;
      } // (classes == null

                      // for testing and tracking
                      //System.out.println("ACSFactoryUtilityImpl."+
                        //    "getAcsAddonExtensionSpecializedRightDomain(() "+
                          //    " for property:"+_propertyName);

      for (int i = 0; i < l_names.length; i++) {
          if (l_names[i].indexOf(".") == -1) {
              error = true;
              index = i;
              break;
          }

          if (l_names[i].substring(0, l_names[i].indexOf(".")).equals(_propertyName)) {
              result = l_names[i].substring(l_names[i].indexOf(".") + 1);
              break;
          }
      } // for i

      if (error)
          throw new InternalError("in ACSFactoryUtilityImpl."+
                            "getAcsAddonExtensionSpecializedRightDomain(() "+
                            "wrong format for "+l_names[index]+" in the ACS "+
                            _acs.getName());

      l_names = null;;
      return result;
  } // getAcsAddonExtensionSpecializedRightDomain()



  /**
   * This method returns the triplet (object.domain.key) of allowed values 
   * for the first three Strings in the quartet of Strings which is used as
   * map key for specialized rights. 
   * Examples of quartets (object.domain.key.subkey) are: <P>
   *     - 'Target.AclRightsSet.Interface.Directory', for ACL rights 
   *           when the target is a Directory instance <P>
   *     - 'Target.AccRightsSet.Type.script', for Account rights 
   *           when the target has the type 'script' <P>
   *     - 'Source.AclRightsSet.Type.group', for ACL rights 
   *           when the source has the type 'group' <P>
   * The object is Target or Source to indicate that these rights are for the target
   * (or source) which has the correct interface or type. <P>
   * The domain is one right category:
   * 'AccRightsSet' for Target only, 'GpORightsSet' for Target only, 
   * 'AclRightsSet' for Target only, 'BdgRightsSet', 'RooRightsSet' for root. <P>
   * The key is 'Interface' or 'Type' to indicate that the subkey belongs to one
   * of these sets. The subkey is not defines in this method.<P>
   * Note: specialized rights for privileges in ACSImpl have the format
   * 'PFType/PFLink.PrvRightsSet.Type.second type', which is derived from
   * the key 'PFType/Link.SecondType.second type' returned
   * by getL_MapKeysForPrivilegeRights().
   * Called by ACSFactoryImpl.addSpecializedRights().
   * @return an array of three Lists containing the allowed String values for respectivly,
   *   objects, domains and keys in a map key
   */
  public static List<String>[] getL_MapKeysStartsForSpecializedRights() {
    
    List<String>[] l_result = new List[3];
    
    l_result[0] = new ArrayList<String>(10);
    l_result[0].add("Target");
    l_result[0].add("Source");

    l_result[1] = new ArrayList<String>(10);
    l_result[1].add("AccRightsSet");
    l_result[1].add("GpORightsSet");
    l_result[1].add("AclRightsSet");
    l_result[1].add("BdgRightsSet");
    l_result[1].add("RooRightsSet");

    l_result[2] = new ArrayList<String>(10);
    l_result[2].add("Interface");
    l_result[2].add("Type");

    return l_result;
  } // getL_MapKeysStartsForSpecializedRights()

  /**
   * This method returns the allowed keys for the standard and for the specialized
   * privilege rights: <P>
   * - 'MainType.main type' for each of the 6 PrivilegeAbst main types, which sets
   * the 7 keys 'PrivilegeRights.MainType.main type' for the standard rights in ACSImpl, <P>
   * - 'PFType/PFLink.SecondType.second type' for the specialized rights, and
   * there is a returned value for each typed privilege second type
   * and each linked privilege second type. <P>
   * Note: for the specialized rights, the format is close to the format delivered
   * by getL_MapKeysForSpecializedRights().<P>
   * Called by the wrapper ACSFactoryImpl.getL_MapKeysForPrivilegeRights() and
   * addPrivilegeRights().
   * @param _factory the ACSFactoryImpl, for which the 2 sets of privilege types are known
   * @return list of the keys for the privilege rights
   */
  public static List<String> getL_MapKeysForPrivilegeRights(ACSFactoryImpl _factory) {

    List<String> l_result = new ArrayList<String>(10);

    for (String st : PrivilegeAbst.getL_PrivilegeTypes()) {
        l_result.add("MainType."+st);
    } // for st

    String[] l_types = _factory.getL_PrivilegeForTypeTypes();
    for (int i = 0; i < l_types.length; i++) {
        l_result.add("PFType.SecondType."+l_types[i]);
    }
    l_types = _factory.getL_PrivilegeForLinksTypes();
    for (int i = 0; i < l_types.length; i++) {
        l_result.add("PFLink.SecondType."+l_types[i]);
    }

                    // for testing and tracking
                    //System.out.println("ACSFactoryUtilityImpl.getL_MapKeysForPrivilegeRights() result: ");
                    //for (String st : l_result) {
                      //  System.out.println("----- "+st);
                    //}

    l_types = null;
    return l_result;
  } // getL_MapKeysStartsForPrivilegeRights()

  /**
   * From a key from the standard rights map or from the specialized rights map of an ACS,
   * this method returns the associated key to use to configure the privilege rights in a new ACS,
   * through the call to ACSFactoryimpl.addPrivilegeRights().
   * @param _privilegeKey key from a standard rights map or a specialized rights map
   * @return the key to use as argument for ACSFactoryimpl.addPrivilegeRights(). May be null.
   */
  public static String convertKeyOfPrivilegeRightsForNewACS(String _privilegeKey) {
        // from ACSImpl.getM_SpecializedRights()
      if (_privilegeKey.startsWith("PFType.PrvRightsSet.Type."))
          return "PFType.SecondType.".concat(_privilegeKey.substring(25));
      else if (_privilegeKey.startsWith("PFLink.PrvRightsSet.Type."))
          return "PFLink.SecondType.".concat(_privilegeKey.substring(25));
        // from ACSImpl.getM_StandardRights()
      else if (_privilegeKey.startsWith("PrivilegeRights.MainType."))
          return "MainType.".concat(_privilegeKey.substring(25));      
      else
          return null;

  } // convertKeyOfPrivilegeRightsForNewACS()

  /**
   * Gets the specialized types which are allowed for a key in the
   * types-and-explorer map, from the current types in the type lists of the ACSFactory.
   * The returned map depends on the current ACSFactory structure only for UserIDs,
   * GroupIDs, GroupID trees, Resources, AclEntries, Directories, Actors, VirtualFolders,
   * PrivilegeForTypes and PrivilegeForLinks. The default value is always
   * all the possible values, when there is no key to match.
   * Caution: the 'GroupID' and 'GroupIDMember' keys return the values from the
   * current list of the EligibleParty types, where some values may be forbidden
   * by the ACS type policy.
   * The 35 possible keys and their associated values are: <P>
   *     - 'Explorer.NoAcsNode', where the values have the format 'ACSTree.XXX',
   *            to forbid the display in the explorer, of some standard nodes:
   *            AclEntries, ACS rights, UserIDs, GroupIDs, Resource Virtual Folders,
   *            Eligible Party Virtual Folders, Privileges <P>
   *     - 'Explorer.SubAcsNode' sets some subACS nodes in the explorer from any
   *            current type starting with '&lt;byAcsCreator&gt; XXX' for Resources or EligibleParties,
   *            where the values have the format 'ResourceType.&lt;byAcsCreator&gt; XXX'
   *            or 'GroupType.&lt;byAcsCreator&gt; XXX', to specify the display
   *            under a dedicated node '&lt;XXX_subacs&gt;', of all the resources or groups
   *            having this relevant type. This type starting
   *            is reserved to these children in the explorer.  <P>
   *     - 'Explorer.AcsGroupTree' sets some group root nodes in the explorer from any
   *            current type starting with '&lt;byAcsCreator&gt; YYY' for EligibleParties,
   *            where the values have the format 'GroupType.&lt;byAcsCreator&gt; YYY',
   *            to specify the display directly under a dedicated node '&lt;YYY_tree&gt;'
   *            of all the groupIDs having the relevant type. This type starting
   *            is reserved to these children in the explorer.  <P>
   *     - 'CreationByBeamer.NoType' where the values are the types,
   *            to forbid the creation through the GUI, of ACSObjects having these types
   *            when they are Resource, EligibleParty, Directory, Actor, UserID, GroupID,
   *            or VirtualFolder <P>
   *     - 'GroupIDMember.NoMemberOf', where the values are the types of the GroupIDmembers for
   *            which the role of member of another group is forbidden <P>
   *     - 'GroupIDMember.IsNotConditionalAclSource', where the values are the types of
   *            the GroupIDmembers for which the conditional ACL is forbidden as source <P>
   *     - 'GroupID.NoPrimaryGroup', where the values are the types of the groups for
   *            which the role of primary group for an account is forbidden <P>
   *     - 'GroupID.NoMainGroup' where the values are the types of the groups for
   *            which the role of main group of a resource is forbidden <P> 
   *     - 'GroupID.NoSecondaryGroup' where the values are the types of the groups for
   *            which the role of secondary group of an actor is forbidden <P>
   *     - 'GroupID.MemberFromAdmin' where the values are the types of the groups
   *            where every member has to be an administrator <P>
   *     - 'GroupID.NoUserIDAsMember' where the values are the types of the groups
   *            where every member has to not be an UserID <P>
   *     - 'GroupID.NoGroupIDAsMember' where the values are the types of the groups
   *            where every member has to not be a GroupID <P>
   *     - 'GroupID.NoMoreThanOneMember' where the values are the types of the groups
   *            where only one member is allowed <P>
   *     - 'GroupID.ConditionalAclGroup' where the values are the types of the groups
   *            which may be condition groups in some flexible or right-defined conditional ACLs <P>
   *     - 'Resource.OneRightACL' where the values are the types of the resources
   *            which handled only AclEntries with no or one right <P>
   *     - 'Resource.NoAccount' where the values are the types of the resources
   *            which do not handle an account as owner (a group is possible from parent) <P>
   *     - 'Resource.NoConditionalACL' where the values are the types of the resources
   *            which never handle flexible or right-defined conditional AclEntries <P>
   *     - 'Resource.NoNonConditionalACL' where the values are the types of the resources
   *            for which an AclEntry without condition group(s) is never operational <P>
   *     - 'Resource.OneConditionGroupInACL' where the values are the types of the resources
   *            which handled only conditional AclEntries with no or one condition group <P>
   *     - 'Actor.NoCurrentAccount' where the values are the types of the actors
   *            for which a current account is forbidden <P>
   *     - 'Actor.NoCurrentGroup' where the values are the types of the actors
   *            for which a current group is forbidden <P>
   *     - 'Actor.NoNullCurrentAccount' where the values are the types of the actors
   *            for which a current account is mandatory <P> 
   *     - 'Actor.IsNotBridgeTarget' where the values are the types of the actors
   *            which cannot be bridge targets <P>
   *     - 'Actor.IsNotBridgeSource' where the values are the types of the actors
   *            which cannot be bridge sources <P>
   *     - 'Actor.IsNotConditionalAclSource', where the values are the types of
   *            the Actors for which the conditional ACL is forbidden as source <P>
   * <P>
   * The following keys are always ended by an allowed type, after the last point:
   *     - 'GroupID.TypesOfMemberFor.allowed_GroupID_type' where the values are 
   *            the allowed types of the members for a group having this type <P>
   *     - 'GroupIDMember.NoMoreThanOneGroup.allowed_GroupIDMember_type' where the values 
   *            are the GroupID types for which the groupIDmember may be member of 
   *            one group of this type at most <P>
   *     - 'GroupIDMember.RecommandedGroupsAsMember.allowed_GroupIDMember_type' 
   *            where the values are the GroupID types for which a message
   *            recommands to the user to put the groupIDmember as member;
   *            there, it is not possible to describe some alternatives
   *            like 'group_type_one<OR>group_type_two', although it may done
   *            by program in an ACSFactoryImpl initialization <P>
   *     - 'Directory.TypesOfChildFor.allowed_Directory_type' where the values are
   *            the allowed types of the children for a Directory having this type,
   *            or if the type is 'NULL', for a Resource without parent <P>
   *     - 'VirtualFolder.TypesOfMemberFor.allowed_VirtualFolder_type' where the values are
   *            the allowed types of the members for a VirtualFolder having this type <P>
   *     - 'VirtualFolder.TypesOfChildFor.allowed_VirtualFolder_type' where the values are
   *            the allowed types of the children for a VirtualFolder having this type,
   *            or if the type is 'NULL', for a VirtualFolder without parent <P>
   *     - 'Resource.SecondTypesOfPrivilegeFor.allowed_Resource_type' where the values are
   *            the allowed second types of the privileges for a Resource having this type <P>
   *     - 'EligibleParty.SecondTypesOfPrivilegeFor.allowed_EligibleParty_type' where the values are
   *            the allowed second types of the privileges for an EligibleParty having this type <P>
   *     - 'VirtualFolder.SecondTypesOfPrivilegeFor.allowed_VirtualFolder_type' where the values are
   *            the allowed second types of the privileges for a VirtualFolder having this type <P>
   *     - 'Resource.TypesOfSourceForConditionalACL.allowed_Resource_type' where the values are
   *            the allowed EligibleParty types as source of a conditional ACL
   *            in a Resource having this type <P>
   * Called by ACSFactoryImpl.getL_LimitedValuesForTypesAndExplorer() at the request
   * to the StringMapPropertyEditor from the value returned by ACSFactoryImplBeanInfo.
   * @param _acs the acs factory for which this method is called
   * @param _key for the constrained values
   * @return list of the allowed values. Never empty nor null. 
   */
  public static List<String> getL_LimitedValuesForTypesAndExplorer(
          ACSFactoryImpl _acs, String _key) {

      List<String>  l_result = new ArrayList(INITIAL_CAPACITY*3); 
      String[]      l_r = null;
      String[]      l_r2 = null;
      Map<String, String[]> m_subPolicy = _acs.getM_SubAndGroupPolicy();

      if (_key.equals("Explorer.NoAcsNode")) {
          l_result.add("ACSTree.ACLENTRIES");
          l_result.add("ACSTree.ACSRIGHTS");
          l_result.add("ACSTree.USERS");
          l_result.add("ACSTree.GROUPS");
          l_result.add("ACSTree.RESVFOLDERS");
          l_result.add("ACSTree.EPVFOLDERS");
          l_result.add("ACSTree.PRIVILEGES");
      }
      else if (_key.equals("Explorer.SubAcsNode")) {
          if (m_subPolicy.isEmpty()) return l_result;

          for (Map.Entry entry : m_subPolicy.entrySet()) {
              if (((Object[])entry.getValue())[0].equals("<ResourceSubACS>")) {
                  l_r = _acs.getL_ResourceTypes();
                  for (int i = 0; i < l_r.length; i++) {
                      if (l_r[i].startsWith("<byAcsCreator> "+entry.getKey()))
                          l_result.add("ResourceType."+l_r[i]);
                  }
              } // if
              if (((Object[])entry.getValue())[0].equals("<GroupSubACS>")) {
                  l_r = _acs.getL_EPTypes();
                  for (int i = 0; i < l_r.length; i++) {
                      if (l_r[i].startsWith("<byAcsCreator> "+entry.getKey()))
                          l_result.add("GroupType."+l_r[i]);
                  }
              } // if
          } // for entry
      }  
      else if (_key.equals("Explorer.AcsGroupTree")) {
          if (m_subPolicy.isEmpty()) return l_result;

          for (Map.Entry entry : m_subPolicy.entrySet()) {
              if (((Object[])entry.getValue())[0].equals("<GroupTree>")) {
                  l_r = _acs.getL_EPTypes();
                  for (int i = 0; i < l_r.length; i++) {
                      if (l_r[i].startsWith("<byAcsCreator> "+entry.getKey()))
                          l_result.add("GroupType."+l_r[i]);
                  }
              } // if
          } // for entry
      }
      else if (_key.equals("CreationByBeamer.NoType")) {
          if (_acs.getManageResources()) {
              l_r = _acs.getL_ResourceTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add("ResourceType."+l_r[i]);
              }
              if (_acs.getManageGroupTrees()) {
                  for (int i = 0; i < l_r.length; i++) {
                      l_result.add("DirectoryType."+l_r[i]);
                  }
              } // if
              if (_acs.getManageActors()) {
                  for (int i = 0; i < l_r.length; i++) {
                      l_result.add("ActorType."+l_r[i]);
                  }
              } // if
          } // if

          if (_acs.getManageAccounts() || _acs.getManageGroups()) {
              l_r = _acs.getL_EPTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add("EPType."+l_r[i]);
              }
              if (_acs.getManageAccounts()) {
                  for (int i = 0; i < l_r.length; i++) {
                      l_result.add("UserIDType."+l_r[i]);
                  }
              } // if
              if (_acs.getManageGroups()) {
                  for (int i = 0; i < l_r.length; i++) {
                      l_result.add("GroupIDType."+l_r[i]);
                  }
              } // if
          } // if

          if (_acs.getManageResourceVirtualFolders()) {
              l_r = _acs.getL_ResourceVirtualFolderTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add("VirtualFolderType."+l_r[i]);
              }
          } // if

          if (_acs.getManageEPVirtualFolders()) {
              l_r = _acs.getL_EPVirtualFolderTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add("VirtualFolderType."+l_r[i]);
              }
          } // if

          if (_acs.getManagePrivilegeRights()) {
              l_r = _acs.getL_PrivilegeForTypeTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add("PrivilegeSecondType."+l_r[i]);
              }
              l_r = _acs.getL_PrivilegeForLinksTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add("PrivilegeSecondType."+l_r[i]);
              }
          } // if
      } 
      else if (_key.equals("GroupID.NoPrimaryGroup")) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }             
      else if (_key.equals("GroupIDMember.NoMemberOf") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("GroupIDMember.IsNotConditionalAclSource") &&
              _acs.getManageGroups() && _acs.getManageAccounts() &&
              (_acs.getManageRightDefinedConditionalAcls() ||
              _acs.getManageFlexibleConditionalAcls())) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("GroupID.NoMainGroup") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }       
      else if (_key.equals("GroupID.NoSecondaryGroup") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("GroupID.MemberFromAdmin") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("GroupID.NoUserIDAsMember") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("GroupID.NoGroupIDAsMember") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }      
      else if (_key.equals("GroupID.NoMoreThanOneMember") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } 
      else if (_key.equals("GroupID.ConditionalAclGroup") &&
              _acs.getManageGroups() &&
              (_acs.getManageFlexibleConditionalAcls() || _acs.getManageRightDefinedConditionalAcls())) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Resource.OneRightACL") &&
              _acs.getManageResources()) {
          l_r = _acs.getL_ResourceTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } 
      else if (_key.equals("Resource.NoAccount") &&
              _acs.getManageResources()) {
          l_r = _acs.getL_ResourceTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Resource.NoConditionalACL") &&
              _acs.getManageResources() &&
              (_acs.getManageFlexibleConditionalAcls() || _acs.getManageRightDefinedConditionalAcls())) {
          l_r = _acs.getL_ResourceTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Resource.NoNonConditionalACL") &&
              _acs.getManageResources() &&
              (_acs.getManageFlexibleConditionalAcls() || _acs.getManageRightDefinedConditionalAcls())) {
          l_r = _acs.getL_ResourceTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Resource.OneConditionGroupInACL") &&
              _acs.getManageResources() &&
              (_acs.getManageFlexibleConditionalAcls() || _acs.getManageRightDefinedConditionalAcls())) {
          l_r = _acs.getL_ResourceTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Actor.NoCurrentAccount") &&
              _acs.getManageActors() &&
              _acs.getManageAccounts() ) {
          l_r = _acs.getL_ResourceTypes();
          l_r2 = _acs.getL_EPTypes();

          for (int i = 0; i < l_r.length; i++) {
              if (Arrays.asList(l_r2).contains(l_r[i]))
                  l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Actor.NoCurrentGroup") &&
              _acs.getManageActors() &&
              _acs.getManageGroups() ) {
          l_r = _acs.getL_ResourceTypes();
          l_r2 = _acs.getL_EPTypes();
          
          for (int i = 0; i < l_r.length; i++) {
              if (Arrays.asList(l_r2).contains(l_r[i]))
                  l_result.add(l_r[i]);
          }
      } //
      else if (_key.equals("Actor.NoNullCurrentAccount") &&
              _acs.getManageActors() &&
              _acs.getManageAccounts() ) {
          l_r = _acs.getL_ResourceTypes();
          l_r2 = _acs.getL_EPTypes();

          for (int i = 0; i < l_r.length; i++) {
              if (Arrays.asList(l_r2).contains(l_r[i]))
                  l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Actor.IsNotBridgeTarget") &&
              _acs.getManageActors() &&
              _acs.getManageBridges() ) {
          l_r = _acs.getL_ResourceTypes();
          l_r2 = _acs.getL_EPTypes();

          for (int i = 0; i < l_r.length; i++) {
              if (Arrays.asList(l_r2).contains(l_r[i]))
                  l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Actor.IsNotBridgeSource") &&
              _acs.getManageActors() &&
              _acs.getManageBridges() ) {
          l_r = _acs.getL_ResourceTypes();
          l_r2 = _acs.getL_EPTypes();

          for (int i = 0; i < l_r.length; i++) {
              if (Arrays.asList(l_r2).contains(l_r[i]))
                  l_result.add(l_r[i]);
          }
      }
      else if (_key.equals("Actor.IsNotConditionalAclSource") &&
              _acs.getManageActors() &&
              (_acs.getManageRightDefinedConditionalAcls() ||
              _acs.getManageFlexibleConditionalAcls())) {
          l_r = _acs.getL_ResourceTypes();
          l_r2 = _acs.getL_EPTypes();

          for (int i = 0; i < l_r.length; i++) {
              if (Arrays.asList(l_r2).contains(l_r[i]))
                  l_result.add(l_r[i]);
          }
      }

      // -----------------------------------------------------
      // test of type-dependant keys
      else if (_key.startsWith("GroupID.TypesOfMemberFor.") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }       
      else if (_key.startsWith("GroupIDMember.NoMoreThanOneGroup.")&&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } 
      else if (_key.startsWith("GroupIDMember.RecommandedGroupsAsMember.") &&
              _acs.getManageGroups()) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } 
      else if (_key.startsWith("Directory.TypesOfChildFor.") &&
              _acs.getManageResourceTrees()) {
          l_r = _acs.getL_ResourceTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } 
      else if (_key.startsWith("VirtualFolder.TypesOfMemberFor.") &&
          _acs.getManageVirtualFolders()) {
          if (_acs.getManageResourceVirtualFolders()) {
              l_r = _acs.getL_ResourceTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add(l_r[i]);
              }
          } // if
          if (_acs.getManageEPVirtualFolders()) {
              l_r = _acs.getL_EPTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add(l_r[i]);
              }
          } // if
      } 
      else if (_key.startsWith("VirtualFolder.TypesOfChildFor.") &&
          _acs.getManageVirtualFolderTrees()) {
          if (_acs.getManageResourceVirtualFolders()) {
              l_r = _acs.getL_ResourceVirtualFolderTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add(l_r[i]);
              }
          } // if
          if (_acs.getManageEPVirtualFolders()) {
              l_r = _acs.getL_EPVirtualFolderTypes();
              for (int i = 0; i < l_r.length; i++) {
                  l_result.add(l_r[i]);
              }
          } // if
      }
      else if (_key.startsWith("Resource.SecondTypesOfPrivilegeFor.") &&
              _acs.getManageResources() && _acs.getManagePrivilegeRights()) {
          l_r = _acs.getL_PrivilegeForTypeTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
          l_r = _acs.getL_PrivilegeForLinksTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } 
      else if (_key.startsWith("EligibleParty.SecondTypesOfPrivilegeFor.") &&
               _acs.getManagePrivilegeRights()) {
          l_r = _acs.getL_PrivilegeForTypeTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
          l_r = _acs.getL_PrivilegeForLinksTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } 
      else if (_key.startsWith("VirtualFolder.SecondTypesOfPrivilegeFor.") &&
              _acs.getManageVirtualFolders() && _acs.getManagePrivilegeRights()) {
          l_r = _acs.getL_PrivilegeForTypeTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
          l_r = _acs.getL_PrivilegeForLinksTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      } //
      else if (_key.startsWith("Resource.TypesOfSourceForConditionalACL.") &&
              _acs.getManageResources() &&
              (_acs.getManageFlexibleConditionalAcls() || _acs.getManageRightDefinedConditionalAcls())) {
          l_r = _acs.getL_EPTypes();
          for (int i = 0; i < l_r.length; i++) {
              l_result.add(l_r[i]);
          }
      }
      
      l_r = null;
      l_r2 = null;
      m_subPolicy = null;
      return l_result;
  } // getL_LimitedValuesForTypesAndExplorer()

  /**
   * This method defines the component nodes in the IS nametree:
   * 'physical' and 'logical'. The return of 'physical' depends on the key and the acs parent.
   * The two values are returned if the arguments are null.
   * Called by StringMapPropertyEditor from the value returned by ACSFactoryImplBeanInfo,
   * and by ActionNewACSsyst.
   * @param _acs the acs factory which calls this method. May be null.
   * @param _key for the constrained values. May be null.
   * @return list of the allowed values. Never empty nor null.
   */
  public static List<String> getL_LimitedValuesForComponentNodes(
          ACSFactoryImpl _acs, String _key) {

      List<String>  l_result = new ArrayList(INITIAL_CAPACITY);

      if (_acs == null && _key == null)
          l_result.add("physical");
      else if (_acs != null && _key != null &&
              _key.startsWith("IsChildOfNewAcsParent.") ) {
          String parentType = null;
          try {
              parentType = ((ImmutableNameISTree)BaseManagerImpl.getInstance().
                    getISTree(_acs.getISName())).getType(_acs.getParentName());
          } catch (BaseError b) {
              throw new InternalError("ACSFactoryUtilityimpl.getL_LimitedValuesForComponentNodes() "+
                      "throws "+b.toString());
          }
          if (ISFactoryUtilityImpl.getChildTypesForISnode(parentType).contains("physical"))
            l_result.add("physical");
          parentType = null;
      } // else if

      l_result.add("logical");

      return l_result;
  } // getL_LimitedValuesForComponentNodes()

  /**
   * This method defines the 3 types of composite nodes in the IS nametree:
   * 'acs', 'acs.authorization_server' and 'subacs'.
   * Called by StringMapPropertyEditor from the value returned by ACSFactoryImplBeanInfo.
   * Caution: any change in this list should update 
   * ACSFactoryImpl.getL_LimitedValuesForBelongsToComposite() and addAlternate()
   * because they test the value 'subacs'.
   * @return list of the allowed values. Never empty nor null.
   */
  public static List<String> getL_LimitedValuesForCompositeNodes() {
      List<String>  l_result = new ArrayList(INITIAL_CAPACITY);
      l_result.add("acs");
      l_result.add("acs.authorization_server");
      l_result.add("subacs");

      return l_result;
  } // getL_LimitedValuesForCompositeNodes()

  /**
   * This method defines the subacs and the group root nodes in the GUI explorer.
   * Called by ACSFactoryImpl for StringMapPropertyEditor.
   * '&lt;ResourceSubACS&gt;' means that a node for this subACS has to be displayed
   * directly under the Resources node, in the explorer.
   * '&lt;GroupSubACS&gt;' for a subACS means that a node has to be displayed
   * directly under the Groups node, in the explorer.
   * '&lt;GroupTree&gt;' for a groups tree means that a group root node has to be displayed
   * directly under the Groups node, in the explorer.
   * @return list of the allowed values. Never empty nor null.
   */
  public static List<String> getL_LimitedValuesForSubAndGroupPolicy() {
      List<String>  l_result = new ArrayList(INITIAL_CAPACITY);
      l_result.add("<ResourceSubACS>");
      l_result.add("<GroupSubACS>");
      l_result.add("<GroupTree>");

      return l_result;
  } // getL_LimitedValuesForSubAndGroupPolicy()

  /**
   * This method defines the alternate types, the right types for alternates and 
   * the options. Call ISFactoryUtilityImpl.getAlternateTypes(), getRightTypesForAlternate()
   * and getOptionsForAlternate(). Only one alternate type
   * should be chosen by the user, and there is no other costraint.
   * Called by ACSFactoryImpl for StringTwoKeysMapPropertyEditor,
   * from the value returned by ACSFactoryImplBeanInfo.
   * @return list of the allowed values. Never empty nor null.
   */
  public static List<String> getL_LimitedValuesForAlternates() {
      List<String>  l_result = new ArrayList(INITIAL_CAPACITY);
      l_result.addAll(ISFactoryUtilityImpl.getAlternateTypes());
      l_result.addAll(ISFactoryUtilityImpl.getRightTypesForAlternate());
      l_result.addAll(ISFactoryUtilityImpl.getOptionsForAlternate());
      return l_result;
  } // getL_LimitedValuesForAlternates()


    /**
   * Controls the alternate values policy. Does not check the argument length,
   * since this method may be used to control a partial array. 
   * The rules are: <P>
   * - the first value is among the types in ISFactoryUtilityImpl.getAlternateTypes(), <P>
   * - then, one or several right types among the types in ISFactoryUtilityImpl.getRightTypesForAlternate(), <P>
   * - then, as options, one or several options in ISFactoryUtilityImpl.getOptionsForAlternate(). <P>
   * The test of the SWITCHING_ALTERNATE unicity for each right type is done in
   * ACSFactoryImpl.addAlternate(), not here.
   * Used by ACSFactoryImpl.addAlternate().
   * @param _l_values for alternates. Never null and may be empty.
   * @return null if the integrity rules are followed or if _l_values is empty,
   *   or return a message which explains the first integrity error.
   */
  public static String controlAlternateValues(String[] _l_values) {
    int index = -1;
    boolean error = false;

    if (_l_values == null) return "null argument";
    if (_l_values.length == 0) return null;

    for (int i = 0; i < _l_values.length; i++) {
        if (!getL_LimitedValuesForAlternates().contains(_l_values[i])) {
            index = i;
            break;
        }
    } // for i

    if (index != -1)
        return "The value "+_l_values[index]+" is not authorized.";
    
    if (!ISFactoryUtilityImpl.getAlternateTypes().contains(_l_values[0]))
        return "The first value is not one of the 3 alternate Switching, Limiting or Subtracting types.";

    if (_l_values.length > 1 &&
            !ISFactoryUtilityImpl.getRightTypesForAlternate().contains(_l_values[1]))
        return "The second value is not one of the 3 AGO/ACL/Bridge rights types.";

    if (_l_values.length < 3) return null;
    
    for (int i = 2; i < _l_values.length; i++) {
        if (ISFactoryUtilityImpl.getAlternateTypes().contains(_l_values[i])) {
            error = true;
            break;
        }
    } // for i

    if (error)
        return "Only the first value may be an alternate Switching, Limiting or Subtracting type.";

    for (int i = 2; i < _l_values.length; i++) {
        if (ISFactoryUtilityImpl.getOptionsForAlternate().contains(_l_values[i])) {
            index = i;
            break;
        }
    } // for i

    if (index == -1 || index == _l_values.length - 1) return null;

    // index is now the start of options, and there is at most a value at index+1
    for (int i = index+1; i < _l_values.length; i++) {
        if (ISFactoryUtilityImpl.getRightTypesForAlternate().contains(_l_values[i])) {
            error = true;
            index = i;
            break;
        }
    } // for i

    if (error)
        return "The type '"+_l_values[index]+"' should not be after an option.";
    return null;
  } // controlAlternateValues()

  /**
   * Returns the subclasses of a given super class, that are declared in the ACS
   * vocabulary as extension classes in an AcsAddon gBase package.
   * The super class must be one of the constants in L_CLASSIC_CLASS_KEYS.
   * The super class may be a key in the vocabulary map to define an AcsAddon classic class, or
   * it may be a generic gBase class. <P>
   * For instance, in the AcsAddon Ubuntu package, 'Executable' is returned by this
   * method as a subclass of 'Actor'.
   * Called by ActionNewResource, ActionNewEligibleParty.
   * @see #ACSADDON_EXTENSION_CLASSES
   * @param _m_vocabulary from an ACSRun.getM_AcsVocabulary()
   * @param _superClassName is a class that may have subclasses as extension classes
   * in an AcsAddon gBase package.
   * @return the reduced names of the extension classes that are subclasses of _className.
   * May be empty but never null.
   * @exception InternalError if an argument is null or if _superClassName is not in
   * L_CLASSIC_CLASS_KEYS
   */
  public static List<String> getL_AcsAddonExtensionClassNames(
          Map _m_vocabulary, String _superClassName) {
      List<String>  l_result = new ArrayList();
      String[]      classes = null;
      String[]      l_segments = null;

      if (_m_vocabulary == null || _superClassName == null)
          throw new InternalError("From ACSFActoryUtilityImpl.getL_AcsAddonExtensionClassNames(), "+
                  "an argument is null");
      if (!L_CLASSIC_CLASS_KEYS.contains(_superClassName))
          throw new InternalError("From ACSFActoryUtilityImpl.getL_AcsAddonExtensionClassNames(), "+
                  "the super class name is unknown in L_CLASSIC_CLASS_KEYS");

                      // for testing and tracking
                      //System.out.println("ACSFactoryUtilityImpl.getL_AcsAddonExtensionClassNames(), "+
                        //      "start of the method for _superClassName: "+_superClassName);

      classes = (String[]) _m_vocabulary.get(ACSADDON_EXTENSION_CLASSES);

      if (classes == null || classes.length == 0) {

                      // for testing and tracking
                      //System.out.println("ACSFactoryUtilityImpl.getL_AcsAddonExtensionClassNames() "+
                        //      "with an empty result since the vocabulary has no ACSADDON_EXTENSION_CLASSES key");

              return l_result;
      } // (classes == null

      for (int i = 0; i < classes.length; i++) {

                      // for testing and tracking
                      //System.out.println("ACSFactoryUtilityImpl.getL_AcsAddonExtensionClassNames(), "+
                        //      "classes[i]: "+classes[i]);

          l_segments = classes[i].split("/is/");
          
          if (l_segments[1].equals(_superClassName))
              l_result.add(l_segments[0]);
      } // for i

                      // for testing and tracking
                      //System.out.println("ACSFactoryUtilityImpl.getL_AcsAddonExtensionClassNames(), "+
                        //      "result: ");
                      //for (Object obj : l_result) {
                        //  System.out.println("--- "+obj);
                      //} // for obj

      classes = null;
      return l_result;
  } // getL_AcsAddonExtensionClassNames()

} // ACSFactoryUtilityImpl
