//Title:        Access Road
//Version:      0.7.1
//Copyright:    Copyright 2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;

import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Map;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

 /**
 * This important class is responsible for modeling an access control entry which manages
 * some access rights or some access restrictions from an eligible party to a resource.
 * An AclEntry is external if its eligible party or its resource is from an ACS which
 * is not the ACS of the AclEntry. It is possible to have a third ACS for the other end.
 * It is not possible to have the same ACS for the two ends and another ACS for the ACL.
 * For an internal ACL, the ACL ACS is the ACS of the two ends.
 * An AclEntry is conditional when it uses some condition group(s). There are two types of
 * conditional AclEntry, the simple-condition (or flexible) ACL and the right-condition ACL.
 * A conditional ACL is for a GroupIDMember or an Actor (new in 0.7.1) as source.
 * An external ACL may be conditional. Then, it cannot have more than one condition group,
 * and this group has to belong to the ACS of one of the two ends (it may not be the ACL ACS).
 * An ACS cannot manage both the external AclEntries and the external linked Privileges.
 * A resource in an ACS addon may inherit the ACL of its (in)direct parent(s),
 * following the inheritance rules of the ACS addon. <P>
 * The condition group(s) belong to the ACL source ACS or to the ACL target ACS.
 * In most of the cases, the rights activation may be handled internally by the conditional AclEntry,
 * except for an Actor as source of a conditional ACL, because the Actor AG context
 * drives the rights activation, and it depends then on the access path
 * into which the Actor is currently considered.  <P>
 * An AclEntry is created and managed by its resource, and it is deleted
 * when the eligible party or the resource is deleted, or the condition group if it is mandatory.
 * A non-conditional AclEntry cannot become a conditional one, and
 * a conditional AclEntry can become a non-conditional one only if it is a right-condition ACL.
 * A simple-condition (or flexible) AclEntry is uniquely defined by the set
 * (its ACS, its Resource, its EligibleParty, first condition GroupID, sens of the rights).
 * The first condition GroupID is mandatory at the creation only if it is a simple
 * conditional AclEntry. The other types of AclEntry are uniquely defined by the set
 * (its ACS, its Resource, its EligibleParty, sens of the rights).
 * This is why a right-condition AclEntry may change its condition groups if the ACS policy
 * authorizes the operation. The sens of the rights is immutable after the creation
 * of the AclEntry, and any right changing has to comply with this sens. <P>
 * In a conditional ACL with a GroupIDMember source, the effective rights
 * the ACL delivers through getL_Rights() are not empty only if the ACL source
 * is a direct or indirect member of each condition group. In a conditional ACL
 * with an Actor source (new in 0.7.1), the rights the ACL delivers through getL_Rights()
 * are the internal rights, not the effective rights, since the ACL cannot know
 * the AG context of its source. In that second case, the effective rights are set
 * by the access paths search, outside this gBase package. For such a conditional ACL,
 * the value of the property 'Enabled rights' is always 'true'. <P>
 * An external AclEntry has only one condition group, while all the types of internal
 * AclEntry may have several condition groups if the ACS policy permits it.
 * An AclEntry is simple-condition (or flexible) simply when
 * the user selects one first condition group at the creation of the AclEntry.
 * An AclEntry is right-defined conditional (or right-condition) when it uses a conditional
 * right at the creation. Such a right includes the reference to one or several groups
 * like in the Linux right 'authorize&lt;IF&gt;&lt;Console&gt;&lt;IF&gt;&lt;ActiveSession&gt;'. <P>
 * The ACL source (an EligibleParty) is implied in the ACL-rights policy
 * only if it is internal, that is, if it belongs to the ACL ACS.
 * Otherwise, only the resource type is taken in account for getting the allowed rights
 * by the ACL ACS, not by the resource ACS. As these rights depend on the resource type
 * (and sometimes the EP type), and since these objects may be external to the ACL ACS,
 * the ACL-rights policy in the ACL ACS has to be defined in accordance to these external types.
 * If the ACL ACS right policy define no rights, the AclEntry cannot have any right.
 * For a conditional ACL, it is also true if a condition group cannot, through the policy types,
 * have the ACL source as member, or an item of the ACL Actor AG context. <P>
 * Performance limitation: if the AclEntry GroupIDMember source is not a member
 * of one condition group, it is necessary to listen the changes in the membership relations
 * to update the effective rights, if the source becomes an indirect member
 * of the condition group. This listening along a chain of groups is applied to
 * all the groups for which the source is indirect member.
 * However, the distance of these listened groups to the source, that is the number
 * of intermediate groups, has to be inferior to 5. Otherwise, a new indirect membership
 * is not detected immediately by Access Road, and the view cannot be updated.
 * The indirect membership detection is always correct at the opening of a view,
 * whatever the distance source/condition group. This performance limitation
 * is not applicable when the source is an Actor, because the AG context
 * of the acccess path is then updated at each relevant property change. <P>
 * The right-defined conditional AclEntry defines its condition group(s)
 * through the single StringRight of the AclEntry. This right has to contain
 * the word 'IF' before the name(s) of the condition group(s). Such a right is
 * not a metaright. It is an ACS right which is created at the ACS construction.
 * In this case, the condition groups may be changed by the user after the ACS creation,
 * if the ACS has other conditional rights.
 * The flexible conditional (simple-condition) AclEntry has its first condition group selected
 * by the user at the creation of the AclEntry. Other condition group(s) may be added to
 * a flexible AclEntry if the AclEntry is not external. <P>
 * To process the inherited AclEntry, it is mandatory to code in an AcsAddon how
 * the Directory creates and removes the inherited AclEntry in its direct children.
 * This may include the choice of the ACL and the child, and the definition
 * of an order to select them in each target. The default behavior, in this generic class,
 * is to inherit all the applicable AclEntries, and to sum up all the rights.
 * After the inherited object creation to code in an AcsAddon, the right updating
 * of the inherited object is handled by the generic DirectoryImpl class, at every change
 * on the AclEntry rights or on the condition groups. <P>
 * An AclEntry has no type, so it is not possible to set it immutable for the user,
 * contrary of a Privilege. But the AclEntry instance and its rights are immutable
 * if its Resource or its EligibleParty is immutable. The condition groups may be
 * edited, even with an immutable group, but not the first condition group in
 * a flexible conditional AclEntry. The right-condition AclEntry is used in the RBAC application.
 * The flexible conditional AclEntry is managed for example in the MySQL ACS,
 * for handling the database and the host MySQL rights.
 * It is an alternative to the LINKED_IF_ALL_SOURCES privilege. The main differences are
 * (1) the ACL rights may be changed by the user, (2) the conditional AclEntry
 * is focused to deliver rights to one source which has to be member of
 * all the condition groups, and through its AG context if it is an Actor.  <P>
 * The use of conditional AclEntries may be reserved to some types of Resource
 * by the ACS type policy. For the GroupIDMember or Actor, the use of AclEntries
 * may be constrained only through the specialized rights, not by the type policy.
 * About the ACL targets, a conditional ACL may be forbidden or set mandatory
 * for some resource types. A given type of resource may have constraints
 * on the allowed sources of its conditional AclEntries.
 * If the ACS type policy enforces it, the resource may have only conditional ACLs.
 * A StringRight may be controled by the ACS specialized rights on AclEntries. <P>
 * All these rules on the ACSObject type policy are handled through keys as followings: <P>
 *     - 'GroupID.ConditionalAclGroup' where the values are the types of the groups
 *            which may be condition groups in some flexible or right-defined conditional ACLs <P>
 *     - 'Resource.OneRightACL' where the values are the types of the resources
 *            which handled only AclEntries with no or one right <P>
 *     - 'Resource.NoConditionalACL' where the values are the types of the resources
 *            which never handle flexible or right-defined conditional AclEntries <P>
 *     - 'Resource.NoNonConditionalACL' where the values are the types of the resources
 *            for which an AclEntry without condition group(s) is never operational <P>
 *     - 'Resource.OneConditionGroupInACL' where the values are the types of the resources
 *            which handled only conditional AclEntries with no or one condition group <P>
 *     - 'GroupIDMember.IsNotConditionalAclSource', where the values are the types of
 *            the GroupIDmembers for which the conditional ACL is forbidden as source <P>
 *     - 'Actor.IsNotConditionalAclSource', where the values are the types of
 *            the Actors for which the conditional ACL is forbidden as source <P>
 *     - 'Resource.TypesOfSourceForConditionalACL.allowed_Resource_type' where the values
 *            are the allowed EligibleParty types as source of a conditional ACL
 *            in a Resource having this type <P>
 * If the ACS type policy 'Resource.NoNonConditionalACL' enforces the resource
 * to have only conditional ACLs and if there is no conditional right and no condition group
 * at the creation of the AclEntry, the constructor assumes it is a right-defined conditional ACL,
 * or, if it is not allowed, throws an exception. For a right-defined conditional ACL,
 * the constructor sets then an empty array of rights and no exception is thrown.
 * After the creation of the AclEntry, the user may set directly
 * the conditional right to produce the derived condition group(s). <P>
 * This class is a javabean with the following bound properties:
 * 'Comment', 'AclRights', 'ConditionSourceGroups'. <P>
 * The listeners are proper to this instance.
 * Excepting for EPRViewInBaseImpl, they are outside gBase,
 * they are transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user. <P>
 * This class has several subclasses in the AcsAddon packages.
 * Most of the variables are protected.
 * @see ARoad0.gBaseInterface.Resource#addAclEntry
 * @version 0.7.0
 */
public class AclEntryImpl implements AclEntry, Serializable, Comparable {

  // private static and protected attributes
  private static final long       serialVersionUID__ = 0;
  private static final int        PRIME = 1000003;
  protected static final int      INITIAL_CAPACITY = 10; // the default value: 101 for map

  // private attributes
  protected String                  detailledName_;
  protected ImmutableACS            aCS_;                           // external object reference
  protected ImmutableACS            resACS_;                        // external object reference
  protected ImmutableName           resName_ = NameImpl.EMPTY_INSTANCE;
  protected String                  resType_ = null;
  protected ImmutableEligibleParty  ep_;                            // external object reference
  protected ImmutableName           acsName_;
  protected ImmutableName           epName_;
  protected ImmutableName           nameOfResAcs_;
  protected ImmutableName           nameOfEpAcs_;
  protected StringRight[]           l_right_ = new StringRight[0];
  protected boolean                 positiveRight_ = true;
  protected String                  comment_;
  
  /**
   * To activate the ACL rights, the EP has to be member of the GroupIDs in this list.
   * May be null.
   */
  protected ArrayList<ImmutableGroupID>   l_ConditionSourceGroups_ = null;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final AclEntryImpl EMPTY_INSTANCE = new AclEntryImpl();

   /**
   * manage all the property change listeners
   */
  private transient PropertyChangeSupport changeSupport_ = null;

  /** variable for the JDK 2 serialization
   * "aCS_",ACSImpl.class,
   *  "resACS_",ACSImpl.class,
   *  "resName_",NameImpl.class,
   *  "resType_",String.class,
   *  "acsName_",NameImpl.class,
   *  "nameOfResAcs_",NameImpl.class,
   *  "epName_",NameImpl.class,
   *  "nameOfEpAcs_",NameImpl.class,
   *  "ep_",EligiblePartyImpl.class,
   *  "l_right_",String[].class,
   *  "positiveRight_",Boolean.TYPE,
   *  "comment_",String.class,
   *  "l_ConditionSourceGroups_",ArrayList.class
   *  "detailledName_",String.class
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("aCS_",ACSImpl.class),
      new ObjectStreamField("resACS_",ACSImpl.class),
      new ObjectStreamField("resName_",NameImpl.class),
      new ObjectStreamField("resType_",String.class),
      new ObjectStreamField("acsName_",NameImpl.class),
      new ObjectStreamField("nameOfResAcs_",NameImpl.class),
      new ObjectStreamField("epName_",NameImpl.class),
      new ObjectStreamField("nameOfEpAcs_",NameImpl.class),
      new ObjectStreamField("ep_",EligiblePartyImpl.class),
      new ObjectStreamField("l_right_",String[].class),
      new ObjectStreamField("positiveRight_",Boolean.TYPE),
      new ObjectStreamField("comment_",String.class),
      new ObjectStreamField("l_ConditionSourceGroups_",ArrayList.class),
      // new in 070
      new ObjectStreamField("detailledName_",String.class)
  };

 /*
  * -------------------------------- Constructors -----------------------------
  */

    /**
     * Constructor to use only as a transient value for initialization, or
     * used from ARoad0.gBase.AclEntryImplBeanInfo.getPropertyDescriptors()
     * in ARoad0.Gui1.CollectionPropertyEditor.getValue().
     */
    public AclEntryImpl() {
    }

  /**
   * This is the standard way to create an instance, only from a resource in gBase.
   * Registers the new acl in the condition group(s).
   * The acl registration in the ACS and in the EligibleParty is done by the resource. <P>
   * An AclEntry is right-defined conditional when it uses a conditional right.
   * An AclEntry is simple (or flexible) conditional when the user set an immutable
   * condition group as argument of the AclEntry constructor.
   * A simple-condition AclEntry is uniquely defined by the set
   * (its ACS, its Resource, its EligibleParty, first condition GroupID, sens of the rights).
   * The first condition GroupID is mandatory at the creation only if it is a simple
   * conditional AclEntry. The other types of AclEntry are uniquely defined by the set
   * (its ACS, its Resource, its EligibleParty, sens of the rights).<P>
   * The constructor calls the method setL_Rights() to control and setup the rights.
   * If the ACL source is not external, it is used for getting the rights policy from the ACL ACS.
   * In all cases, the ACL target is used to get them. In a right-defined conditional ACL,
   * the method setL_Rights() sets the condition groups.
   * The error messages for the user are as following: <P>
   * - An external conditional AclEntry must have its Eligible Party ACS or
   * its Resource ACS being its condition group ACS. <P>
   * - This ACS does not manage the simple-condition AclEntries, into which the rights
   * are activated only if the right user is member of all the condition group(s); <P>
   * - A condition group has to be from the source ACS or from the target ACS. <P>
   * - With the type policy 'Resource.NoConditionalACL', the ACS forbids this Resource type
   * as target of a simple-condition AclEntry. <P>
   * - With the type policy 'GroupIDMember.IsNotConditionalAclSource', the ACS forbids
   * this UserID or GroupID type as source of a simple-condition AclEntry. <P>
   * - No conditional right in a simple-condition AclEntry, since the condition groups
   * are directly set. <P>
   * Called only by ResourceImpl.
   * @param _aCS open controller which owns this AclEntry. May be the main ACS
   *    of _res, or not. If not, is recorded as ACS controller of this main ACS.
   * @param _res External Object Reference to the resource.
   * @param _ep External Object Reference to the EligibleParty, which
   *    may belong to another ACS. Has to be a GroupIDMember if it is a conditional AclEntry.
   * @param _l_right is an array of rights for _ep on _res. Create an empty array
   *  with a length of 1 if the argument is null or if its first element is null
   *  (even if there is a second non-null element). Controls if the rights are
   *  allowed by the ACS and if not, the rights are not set. For a right-defined
   *  conditional AclEntry, only one right is allowed.
   * @param _positiveRight is true if the rights are oriented to grant,
   *  and false if they are oriented to deny. _l_right is checked.
   * @param _conditionGroup first condition group. May be null. If null, no condition
   * group may be added directly after the AclEntry creation. If not null, _l_right
   * cannot contain a conditional right.
   * @exception CreateError if (_res or _ep).getName() or (_res or _ep).getEorACS() is null,
   *  if _ep is not an EligiblePartyImpl nor an ActorImpl, if _right has not the sens
   *   of _positiveRight, or if they are conditional and more than one, or more than
   *   one right when the acs forbids it, or external.
   */
   protected AclEntryImpl( ImmutableACS _aCS, ImmutableResource _res,
      ImmutableEligibleParty _ep, StringRight[] _l_right, boolean _positiveRight,
      ImmutableGroupID _conditionGroup) throws CreateError {

        if (_res.getName()==null || _res.getEorACS()==null ||
            _ep.getName()==null || _ep.getEorACS().getName()==null) {
              throw new InternalError("new AclEntryImpl with a null argument");
        }
        resName_ = (Name)_res.getName(); // getName() returns a clone
        resType_ = _res.getType();
        epName_ = _ep.getName();
        resACS_ = _res.getEorACS();
        ep_ = _ep;
        aCS_ = _aCS;        
        positiveRight_ = _positiveRight;
        acsName_ = _aCS.getName();
        nameOfResAcs_ = resACS_.getName();
        nameOfEpAcs_ = _ep.getAcsName();
        comment_ = "";

        String name = null;

                                  // for testing and tracking
                                 //System.out.println("AclEntryImpl constructor: starting");

        // no use of equals() to detect bugs on external ACL
        // from ACSImpl.checkClosedExternalAclEntries()
        if (isExternalAclEntry() && _conditionGroup != null &&
                (_ep.getEorACS() != _conditionGroup.getEorACS() &&
                _res.getEorACS() != _conditionGroup.getEorACS())) {


                                  // for testing and tracking
                                 //System.out.println("AclEntryImpl constructor: false condition group:");
                                 //System.out.println("--- _conditionGroup.getEorACS(): "+
                                   //      _conditionGroup.getEorACS());
                                 //System.out.println("--- _ep.getEorACS(): "+
                                   //      _ep.getEorACS());
                                 //System.out.println("--- _res.getEorACS(): "+
                                   //      _res.getEorACS());

            resName_ = null;
            epName_ = null;
            acsName_ = null;
            throw new CreateError(" An external conditional AclEntry must have"+
                    " its Eligible Party ACS or its Resource ACS being its condition group ACS.");
        }
        
        // The detailled name is build up first for an internal AclEntry
        if (_conditionGroup != null)
            detailledName_ = buildUpDetailledName(
                ((Name)epName_.clone()).subname(
                    nameOfEpAcs_.noOfComponents(), epName_.noOfComponents()),
                ((Name)resName_.clone()).subname(
                    nameOfResAcs_.noOfComponents(), resName_.noOfComponents()),
                positiveRight_, _conditionGroup.getName().lastComponent());
        else
            detailledName_ = buildUpDetailledName(
                ((Name)epName_.clone()).subname(
                    nameOfEpAcs_.noOfComponents(), epName_.noOfComponents()),
                ((Name)resName_.clone()).subname(
                    nameOfResAcs_.noOfComponents(), resName_.noOfComponents()),
                positiveRight_, null);

        // For an external ACL, the form of the detailled name is: <P>
        // - ACS: ACS_name | (full or incomplete) Eligible Party name >>
        // (full or incomplete) Resource name || grant OR deny ||
        // <Source ACS group> OR <Target ACS group> ||
        // first condition group last component.
        if (isExternalAclEntry()) {
            StringBuilder sb = new StringBuilder();

                                      // for testing and tracking
                                      //System.out.println("AclEntryImpl constructor, "+
                                        //      "detailledName_ before processing for external: "+
                                          //    detailledName_);

            sb.append("ACS:");
            sb.append(acsName_.toString());
            sb.append("|");

            if (!nameOfEpAcs_.equals(acsName_)) {
                name = ep_.getAcsName().toString();
                // remove the two last spaces
                sb.append(name.substring(0, name.length() - 2));
            }
            // add 'incomplete Eligible Party name >>' with the first space
            sb.append(detailledName_.substring(
                    0,
                    detailledName_.indexOf(">>") + 2));

            if (!nameOfResAcs_.equals(acsName_)) {
                name = nameOfResAcs_.toString();
                // remove the twp last spaces
                sb.append(name.substring(0, name.length() - 2));
            }
            // add 'incomplete Resource name ||' with the first space
            sb.append(detailledName_.substring(
                    detailledName_.indexOf(">>") + 2,
                    detailledName_.indexOf("||") + 2));

            if (detailledName_.indexOf("||") == detailledName_.lastIndexOf("||")) {
                // add 'grant OR deny'
                sb.append(detailledName_.substring(
                        detailledName_.indexOf("||") + 2,
                        detailledName_.length()));
            }
            else {
                // add 'grant OR deny ||'
                sb.append(detailledName_.substring(
                        detailledName_.indexOf("||") + 2,
                        detailledName_.lastIndexOf("||") + 2));

                // insert ' <Source ACS group> OR <Target ACS group> ||'
                if (nameOfEpAcs_.equals(_conditionGroup.getAcsName())) {
                    sb.append(" <Source ACS group> ||");

                                      // for testing and tracking
                                      //System.out.println("AclEntryImpl constructor, "+
                                        //      "add '<Source ACS group>' from nameOfEpAcs_: "+
                                          //    nameOfEpAcs_);

                }
                else {
                    sb.append(" <Target ACS group> ||");
                }

                // add 'first condition group last component'
                sb.append(detailledName_.substring(
                        detailledName_.lastIndexOf("||") + 2,
                        detailledName_.length()));

            } // else

            detailledName_ = sb.toString();
            sb = null;
            name = null;
        } // if (isExternalAclEntry())

                                      // for testing and tracking
                                      //System.out.println("AclEntryImpl constructor, "+
                                        //      "detailledName_: "+detailledName_);

                                      // for testing and tracking
                                     //System.out.println("AclEntryImpl constructor: "+
                                       //      "after buildUpDetailledName()");

        try {
            if (_conditionGroup != null) {
                if (!aCS_.getManageFlexibleConditionalAcls())
                  throw new CreateError(" This ACS does not manage the simple-condition AclEntries, "+
                          "into which the rights are activated only if the right user is member"+
                          " of all the condition group(s).");

                if (_conditionGroup.getEorACS() != ep_.getEorACS() &&
                        !_conditionGroup.getAcsName().equals(nameOfResAcs_) )
                    throw new CreateError(" A condition group has to be from the source ACS "+
                            "or from the target ACS.");

                Map<String, String[]> m_policy = aCS_.getM_TypeOrientedPolicy();

                if (m_policy.get("Resource.NoConditionalACL") != null &&
                        Arrays.asList(m_policy.get("Resource.NoConditionalACL")).
                            contains(UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        getEorResource().getResourceType()))
                        )
                    throw new CreateError(" With the type policy 'Resource.NoConditionalACL',"+
                            " the ACS forbids this Resource type "+
                            "as target of a simple-condition AclEntry.");

                if (_ep instanceof GroupIDMember &&
                        m_policy.get("GroupIDMember.IsNotConditionalAclSource") != null &&
                        Arrays.asList(m_policy.get("GroupIDMember.IsNotConditionalAclSource")).
                            contains(UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(ep_.getEPType()))
                        )
                    throw new CreateError(" With the type policy 'GroupIDMember.IsNotConditionalAclSource',"+
                            " the ACS forbids this UserID or GroupID type "+
                            "as source of a simple-condition AclEntry.");

                if (_ep instanceof Actor &&
                        m_policy.get("Actor.IsNotConditionalAclSource") != null &&
                        Arrays.asList(m_policy.get("Actor.IsNotConditionalAclSource")).
                            contains(UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(ep_.getEPType()))
                        )
                    throw new CreateError(" With the type policy 'Actor.IsNotConditionalAclSource',"+
                            " the ACS forbids this Actor type "+
                            "as source of a simple-condition AclEntry.");

                // adds at 0 in the groups list
                addConditionSourceGroup(_conditionGroup);
                m_policy = null;
            } // if (_conditionGroup != null)

                                      // for testing and tracking
                                     //System.out.println("AclEntryImpl constructor: "+
                                       //      "after addConditionSourceGroup()"+
                                         //    " for detailled name: "+detailledName_);

            setL_Rights(_l_right); // uses aCS_, resACS_, resName_, ep_, acsName_

            if (_conditionGroup != null && isRightDefinedConditionalAclEntry())
                throw new CreateError(" No conditional right in a simple-condition AclEntry,"+
                        " since the condition groups are directly set.");
        }
        catch (CreateError c) {

                                      // for testing and tracking
                                     //System.out.println("AclEntryImpl constructor: "+
                                       //      "before finalizeClosedAcl() "+
                                         //    "for detailled name: "+detailledName_);

            finalizeClosedAcl();
            throw c;
        }

                                      // for testing and tracking
                                     //System.out.println("AclEntryImpl constructor: ending "+
                                       //      "for detailled name: "+detailledName_);

  } // AclEntryImpl()

   /**
    * This method creates the detailled name of an internal AclEntry, that is unique in its ACS,
    * and it provides the preprocessing of the detailled name for an external AclEntry.
    * The detailled name may include its ACS name and the ACS names of the Resource/EP
    * only if the ACL is external. Being external means that at least one of
    * the two EP/Resource ACS is not the ACL ACS.
    * It is then possible to have a third ACS for the other end.
    * It is not possible to have the same ACS for the two ends and another ACS for the ACL.
    * For an internal ACL, the ACL ACS is the ACS of the two ends.
    * Otherwise, for an internal ACL, the ACL ACS is the ACS of the two ends.
    * Called by the constructors and getDetailledNameFromName(),
    * and by some AcsAddon classes like DirectoryMySQLImpl.
    * The format of the detailled name for an internal ACL is as following: <P>
    * - incomplete Eligible Party name >> incomplete Resource name ||
    * grant OR deny || first condition group last component. <P>
    * For an external ACL, the format of the detailled name is as following: <P>
    * - ACS: ACS_name | (full or incomplete) Eligible Party name >>
    * (full or incomplete) Resource name || grant OR deny ||
    * &lt;Source ACS group&gt; OR &lt;Target ACS group&gt; ||
    * first condition group last component. <P>
    * An Eligible Party or Resource name is incomplete if it belongs to the AclEntry ACS.
    * The condition group is put in the name or the detailled name only for
    * a flexible conditional AclEntry. '&lt;XXXX ACS group&gt;' is used
    * only for an external ACL, when the condition group belongs
    * to the Source or Target (XXXX) ACS.
    * @param _epName name of the EligibleParty, without its ACS name. Never null nor empty.
    * @param _resName name of the Resource, without its ACS name. Never null nor empty.
    * @param _positiveRight is true for a granting right, false for a denying one
    * @param _conditionGroupName first condition group name last component. May be null.
    * @return the detailled name of an internal AclEntry, even if all the parameters are null,
    * or the first form of the detailled name for an external AclEntry.
    * @exception CreateError if a name is null or empty
    */
    public static final String buildUpDetailledName(ImmutableName _epName,
      ImmutableName _resName, boolean _positiveRight, String _conditionGroupName)
      throws CreateError {

        if (_epName == null || _epName.isEmpty())
           throw new CreateError(
              " Null or empty EligibleParty name");
        if (_resName == null || _resName.isEmpty())
           throw new CreateError(
              " Null or empty Resource name");

        StringBuilder sb = new StringBuilder();

        sb.append(_epName.toString() );
        sb.append(">>");
        sb.append(_resName.toString() );

        sb.append(" || ");

        if (_positiveRight) sb.append("grant");
        else sb.append("deny");

        if (_conditionGroupName != null) {
            sb.append(" || ");
            sb.append(_conditionGroupName);
        }

        return sb.toString();
    } // buildUpDetailledName()

 //
 //  --------------------------- interface methods ---------------------------
 //

  /**
  * This method returns the detailled name for the explorer,
  * which gives the key components of the object, without the acs name.
  * The form of the detailled name is: <P>
  * (full or incomplete) Eligible Party name >> (full or incomplete) Resource name ||
  * grant/deny || first condition group last component. <P>
  * The condition group is put in the name or the detailled name only for
  * a flexible conditional AclEntry. The EP and Resource names may be incomplete
  * because the detailled name includes the ACS names of the Resource/EP
  * only if the ACL is external, and it means that at least one of
  * the two EP/Resource ACS is not the ACL ACS.
  * Works on cloned aclEntries.
  * @see #getKeyPropertiesFromDetailledName
  * @return the detailled name, or 'null AclEntry'
  */
  public String getDetailledName() {
    if (nameOfEpAcs_ == null) return "null AclEntry";
    else return detailledName_;
  } // getDetailledName()

  /**
  * Gets the resource through a call to the ACS map, so quite slow.
  * getResName() and getResType() are much more faster.
  * Returns null if it is a clone or a finalized instance.
  * Caution: Use resourceACS.getEorM_Resources(), so do NOT work after
  * the ACS closing.
  * @return the resource that applies this AclEntry
  */
  public ImmutableResource getEorResource() {
      if (resACS_ != null) {
          return (ImmutableResource)resACS_.getEorM_Resources().get(resName_);
      } else return null;
  } // getEorResource()

  /**
  * Gets the access source displayed in the beamer.
  * The form is:
  * 'ep name as string'+'(ACL)'.
  * @return the source name. May be null.
  */
  public String getSource() {
      if (epName_ != null) {
         return epName_.toString() + "(ACL)";
      } else return null;
  } // getSource()

  /**
  * Gets the access target displayed in the beamer.
  * The form is:
  * 'resource name as string'+'(by ACL)'.
  * @return the target name. May be null.
  */
  public String getTarget() {
      if (resName_ != null) {
         return resName_.toString() + "(by ACL)";
      } else return null;
  } // getTarget()

    /**
    * Returns null for a clone or a finalized instance.
    * @return ImmutableEligibleParty, as an External Object Reference
    */
     public final ImmutableEligibleParty getEorEP() {
        return ep_;
     }

     /**
    * Returns null for a clone or a finalized instance.
    * @return ImmutableACS of this AclEntry, as an External Object Reference
    */
    public final ImmutableACS getEorACS() {
        return aCS_;
    }

    /**
    * @return ImmutableACS of the AclEntry resource, as an External Object Reference
    */
    public final ImmutableACS getEorResACS() {
        return resACS_;
    }

  /**
   * Sets the rights of the AclEntry, from which the effective rights
   * given by getRight() are derived.
   * Called by the GUI. 
   * Calls setL_Rights() and resetRightsAsInheritedAclEntry().
   * Fires a PropertyChangeEvent "AclRights" to the listeners.
   * @param _st array of non-null, non-empty rights
   *    for the couple (resource, EligibleParty) in the AclEntry ACS.
   *    The authorized rights depend on the AclEntry ACS. May be null.
   * @exception CreateError if the resource or the eligible party is immutable,
   * or from setL_Rights()
   */
   public synchronized void setL_RightsForUser(StringRight[] _st) throws CreateError {

       if (getEorResource().getResourceType().startsWith("<immutable> ") ||
               getEorResource().getResourceType().endsWith("<immutable> "))
          throw new CreateError(
            " The type of "+getEorResource().getName().lastComponent()+
            " sets it as immutable for its ACLs.");
       if (getEorEP().getEPType().startsWith("<immutable> ") ||
               getEorEP().getEPType().endsWith("<immutable> "))
          throw new CreateError(
            " The type of "+getEorEP().getName().lastComponent()+
            " sets it as immutable for its ACLs.");

       StringRight[]       oldRight = new StringRight[l_right_.length];
       System.arraycopy(l_right_, 0, oldRight, 0, l_right_.length);

       setL_Rights(_st);
       resetRightsAsInheritedAclEntry();

       try {
            // no firing if the rights are the same
            if (!Arrays.asList(oldRight).equals(Arrays.asList(l_right_)))
                firePropertyChange("AclRights", oldRight, getNoConditionalRight());

       } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From AclEntryImpl.setL_RightsForUser(), firePropertChange() throws:");
            e.printStackTrace();
       }

       oldRight = null;
   } // setL_RightsForUser()

  /**
   * Gets the effective rights for the eligible party upon the resource.
   * If there is no condition group, the effective rights are the internal rights
   * as returned by getNoConditionalRight().
   * If there are conditional groups and if the eligible party is a group member,
   * the eligible party has to be a direct or indirect member of each conditional group,
   * to return the internal rights. Otherwise, the condition rights are not workable,
   * and an empty array is returned. If the eligible party is not a group member,
   * the condition group(s) are forbbiden.
   * Calls getNoConditionalRight(), GroupIDImpl.containsAsMember()
   * and containsAsIndirectMember().
   * @return effective rights for the couple (Resource, EligibleParty)
   * Never null. May be an empty array.
   */
    public final StringRight[] getL_Rights() {
        
        if (!getEnabledRights()) {
            return new StringRight[0];
        } // if
        else {
            return getNoConditionalRight();
        } // else
    } // getRight()

  /**
   * Gets the list of the internal conditional or no-conditional rights
   * for the EligibleParty upon the Resource, whatever the compliance to
   * the conditions if there are some conditional groups.
   * Called by setL_Rights() and getL_Rights().
   * @return rights as READ, CONNECT, DELETE... for the couple (resource, eligible party).
   *  Never null. May be an empty array.
   */
    public final StringRight[] getNoConditionalRight() {

                                  // for testing and tracking
                                 //System.out.println("call to AclEntryImpl.getNoConditionalRight()");

        StringRight[] l_right = new StringRight[l_right_.length];
        
        if (l_right_.length != 0)
            System.arraycopy(l_right_, 0, l_right, 0, l_right_.length);
        return l_right;
    } // getNoConditionalRight()

    /**
     * Gets true for a right-defined conditional AclEntry.
     * A simple-condition AclEntry is not a right-condition AclEntry.
     * Called by setConditionSourceGroupsFromRights().
     * @return true if the first right is a conditional right
     */
    public final boolean isRightDefinedConditionalAclEntry() {
        if (l_right_ != null && l_right_.length != 0 &&
                l_right_[0].getRight().indexOf("<IF>") != -1)
            return true;
        else
            return false;
    } // isRightDefinedConditionalAclEntry()

    /**
     * Gets true for a right-defined or a flexible conditional AclEntry.
     * @return true if the condition groups list is not null
     */
    public final boolean isConditionalAclEntry() {
        if (l_ConditionSourceGroups_ != null) return true;
        else
            return false;
    } // isConditionalAclEntry() getEnabledRights

    /**
     * Gets false only for a conditional AclEntry where the GroupIDMember source is not
     * (in)direct member of a condition group. A conditional ACL having an Actor as
     * source has always enabled rights (new in 0.7.1).
     * @return true if the rights are workable
     */
    public final boolean getEnabledRights() {
        boolean result = true;
        ImmutableGroupID grp = null;

        if (l_ConditionSourceGroups_ != null) {
            if (!(ep_ instanceof GroupIDMember)) return true;
            else {
                for (Iterator ite = l_ConditionSourceGroups_.iterator(); ite.hasNext(); ) {
                    grp = (ImmutableGroupID)ite.next();
                    // containsAsIndirectMember() returns false if grp is finalized
                    if (grp == null ||
                            (!grp.containsAsMember((GroupIDMember)ep_) &&
                            !grp.containsAsIndirectMember((GroupIDMember)ep_))) {
                        result = false;
                        break;
                    } // if
                } // for ite
            } // else
        } // if

                        //if (!result) {
                                // for testing and tracking
                            //if (grp != null)
                              //  System.out.println("AclEntryImpl.getEnabledRights(): conditional right with"+
                                //    " a false condition for the group "+grp.getNickName());
                            //else
                              //  System.out.println("AclEntryImpl.getEnabledRights(): null conditional right"+
                               //     " for "+this);
                        //} // if

        grp = null;
        return result;
    } // getEnabledRights()

    /**
     * @return _b true if the AclEntry rights ('r' for example)
     *  grant access to the eligible party, and false if the rights
     *  deny an access type.
     */
    public final boolean getPositiveRight() {
        return positiveRight_ ;
    }

   /**
    * Caution: does not return a clone.
    * @return ImmutableName name of the AclEntry resource ACS. 
    * No cloned from the internal property. May be null.
    */
    public ImmutableName getNameOfResACS() {
        return nameOfResAcs_;
    }

   /**
    * Caution: does not return a clone.
    * @return ImmutableName name of the AclEntry resource. 
    * No cloned from the internal property.
    *   May be NameImpl.EMPTY_INSTANCE.
    */
    public ImmutableName getResName() {
        return resName_;
    }

   /**
    * To use when it is possible to replace getEorResource().
    * @return the AclEntry resource type.
    */
    public String getResType() {
        return resType_;
    }

   /**
    * Caution: does not return a clone.
    * @return ImmutableName name of the AclEntry ACS. 
    * No cloned from the internal property. May be null.
    */
    public ImmutableName getAcsName() {
        return acsName_;
    }

   /**
    * Caution: does not return a clone.
    * @return ImmutableName name of the AclEntry EligibleParty ACS.  
    * No cloned from the internal property. May be null.
    */
    public ImmutableName getNameOfEpACS() {
        return nameOfEpAcs_;
    }

   /**
    * Caution: does not return a clone.
    * @return ImmutableName name of the AclEntry EligibleParty.  
    * No cloned from the internal property. May be null.
    */
    public ImmutableName getEpName() {
        return epName_;
    }

  /**
  * Set any comment related to this AclEntry. It is a bound property.
  * @param _s comment which is associated to the AclEntry
  */
  public synchronized void setComment(String _s)	{
      String oldComment = comment_;
      comment_ = _s;

      try {
          firePropertyChange("Comment", oldComment, comment_);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From AclEntryImpl.setComment(), firePropertChange() throws:");
        e.printStackTrace();
      }
      oldComment = null;
  }

  /**
   * @return comment which is associated to the AclEntry
   */
  public String getComment()	{
          return comment_;
  }   
  
  /**
   * To activate the ACL rights in a conditional AclEntry, the EP
   * is a  GroupIDMember, and it has to be a member of each condition GroupID.
   * If the EP is not a GroupIDMember, the effective rights are null.
   * @return condition groups. May be empty, never null.
   */
  public List<ImmutableGroupID> getL_ConditionSourceGroups() {
      if (l_ConditionSourceGroups_ == null) return new ArrayList(INITIAL_CAPACITY);
      else return (List<ImmutableGroupID>)l_ConditionSourceGroups_.clone();
  }  
  
  /**
  * Add a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the 'Comment', 'AclRights', 'ConditionSourceGroups'
  * properties.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null) 
          changeSupport_ = new PropertyChangeSupport(this);
 
                                  // for testing and tracking
                                  //if (ep_.getNickName().equals("jerry")) {
                                    // System.out.println(
                                      //   "call to AclEntryImpl.addPropertyChangeListener()"+
                                        // " on 'jerry' for the ACL "+this);
                                     //System.out.println(" ---- and for the property "+
                                       //      _propertyName+
                                         //   " by the listener "+_l);
                                  //} // if


      synchronized (changeSupport_) 
      {changeSupport_.addPropertyChangeListener(_propertyName, _l);} 
  }

  /**
  * Add a listener to the bean.
  * @param _l PropertyChangeListener to add
  */
  public synchronized void addPropertyChangeListener(PropertyChangeListener _l){

      if (changeSupport_ == null) 
          changeSupport_ = new PropertyChangeSupport(this);
 
                                  // for testing and tracking
                                  //if (ep_.getNickName().equals("jerry"))
                                    // System.out.println(
                                      //   "call to AclEntryImpl.addPropertyChangeListener()"+
                                        // " for the target 'jerry' and the source "+
                                         //ep_.getNickName());
      
      synchronized (changeSupport_) 
      {changeSupport_.addPropertyChangeListener(_l);} 
  }

  /**
  * remove a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the 'Comment', 'AclRights', 'ConditionSourceGroups'
  *  properties.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){


      if (changeSupport_ == null) 
          changeSupport_ = new PropertyChangeSupport(this);
      
      synchronized (changeSupport_) 
      {changeSupport_.removePropertyChangeListener(_propertyName, _l);} 
  } // removePropertyChangeListener()

  /**
  * remove a listener to the bean.
  * @param _l PropertyChangeListener to remove
  */
  public synchronized void removePropertyChangeListener(PropertyChangeListener _l){

      if (changeSupport_ == null) 
          changeSupport_ = new PropertyChangeSupport(this);
      
      synchronized (changeSupport_) 
      {changeSupport_.removePropertyChangeListener(_l);} 
  }
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("AclEntryImpl.getPropertyChangeListeners(),
                    // call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      return changeSupport_.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()

  /**
   * Based on the AclEntry name comparisons on the EP first, and then on the Resource,
   * if the EP names are equals. On Resources, the nodes are always smaller than their leaves.
   * The UserIDs are smaller than the GroupIDs.
   * The condition groups are compared at the end, through the comparison
   * of the detailled names where only the first group name last component appears, if any,
   * and the ACS name is it is an external AclEntry.
   * The rights are NOT compared.
   * Called by ACSImpl.newEorAclEntry() (through Arrays.sort())
   * ACSTreeUtilities.addDetailledNameUnderNode() and by
   * Gui1.CollectionPropertyEditor.propertyChange() to display the AclEntries.
   * @param o to compare must be an ImmutableAclEntry
   * @exception ClassCastException - if the specified object's type
   *  prevents it from being compared to this current Object
   * @return 0 if the ACL are almost equals, a negative integer if this object
   * is lesser than the argument, and a positive integer if this object
   * is greater than the argument.
   */
  public int compareTo(Object o) throws ClassCastException {
    if (!(o instanceof ImmutableAclEntry))
      throw new ClassCastException("In AclEntry.compareTo(), the argument "+
        "is not an instance of ImmutableAclEntry");
    // if (equals(o) ) return 0;
    if (epName_ == null) throw new ClassCastException("In AclEntry.compareTo(),"+
      " the object name is null");
    int result = epName_.compareTo( ((ImmutableAclEntry)o).getEpName() );
    if (result != 0) return result;
    result = resName_.compareTo( ((ImmutableAclEntry)o).getResName() );
    if (result != 0) return result;
    result = detailledName_.compareTo( ((ImmutableAclEntry)o).getDetailledName() );

                // for testing and tracking
                //System.out.println("AclEntryImpl.compareTo() result: "+result);
                //System.out.println("----- on the AclEntry: "+getName());
                //System.out.println("------- with argument: "+((ImmutableAclEntry)o).getName());

    return result;
  } // compareTo()
  
  /**
   * Called by removeEorAclEntry() in the events firing.
   * The ACS, the EP and the resource ACS are null in the returned AclEntry.
   * The rights array is copied without cloning of the rights (shallow copy).
   * Thus the rights networks of metarights and acsrights are linked to the ACL clone.
   * The method getEorResource() returns null while getResName() returns a clone
   * of the resource name. The condition groups list is cloned in a shallow copy.
   * The methods getRight() and getDetailledName() work well, as toString().
   * The properties are protected, so may be accessed by the package.
   * Caution: the clone is not registered in its ACS, its resource, its EP or
   * its confition groups, but equals() returns true when it compares a clone and its source.
   * @return Object without external references
   * (ACS, resource ACS and EligibleParty) and copy of the other properties.
   */
   public Object clone() {
      try {
          AclEntryImpl clon = (AclEntryImpl) super.clone();
         
          clon.resName_ = (ImmutableName) resName_.clone();
          clon.aCS_ = null; // to cut external links
          clon.resACS_ = null; // to cut external links
          clon.ep_ = null; // to cut external links
          clon.positiveRight_ = positiveRight_;
          clon.detailledName_ = detailledName_;
          
          clon.l_right_ = new StringRight[l_right_.length];
          System.arraycopy(l_right_, 0, clon.l_right_, 0, l_right_.length);          
          
          if (acsName_ != null)
              clon.acsName_ = (ImmutableName) acsName_.clone();
          if (nameOfResAcs_ != null)
              clon.nameOfResAcs_ = (ImmutableName) nameOfResAcs_.clone();
          if (epName_ != null)
              clon.epName_ = (ImmutableName) epName_.clone();
          if (nameOfEpAcs_ != null)
              clon.nameOfEpAcs_ = (ImmutableName) nameOfEpAcs_.clone();
          
          if (l_ConditionSourceGroups_ != null)
              clon.l_ConditionSourceGroups_ = (ArrayList<ImmutableGroupID>)
                      l_ConditionSourceGroups_.clone();

                                // for testing and tracking
                            //System.out.println("AclEntryImpl.clone() for"+this+
                              //      " and the result is "+clon);


          return clon;

      } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
  	    if (aCS_ != null) {
  	            aCS_.getEorBaseError().addException(e.toString());
  	    }
        throw new InternalError(e.toString());
      } // catch
   } // clone()

   /**
    * This method returns true for a clone, even if it has no direct references
    * to its ACS, resource or EP. This method casts _obj to AclEntryImpl, to have
    * a direct access to the protected properties. The result is true if the two
    * ACL have the same rights in a different order.
   * @param _obj is the object to compare.
   * @return boolean true if _obj is equal with the same class,
   *  not only an implementation of AclEntry.
   */
  public synchronized boolean equals(Object _obj) {
      // initialization
      Name nullName = new NameImpl();

      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      AclEntryImpl obj = (AclEntryImpl) _obj;      

      // resACS_  null if obj is empty, created with "new AclEntryImpl()"
      if (nameOfResAcs_ == null || obj.nameOfResAcs_ == null) {
          if (nameOfResAcs_ != null || obj.nameOfResAcs_ != null)
            return false;  // if
      } // if
      else {
        if (!nameOfResAcs_.equals( obj.nameOfResAcs_ )) return false; }

      // true if resName_ == nullName && obj.getEorResource() == null
      if (obj.getEorResource() != null ) {
        if (!resName_.equals( obj.getEorResource().getName() ))
            return false;
      }
      if ( obj.getEorResource() == null && !resName_.equals(nullName) ) return false;

      if (acsName_ == null || obj.acsName_ == null) {
          if (acsName_ != null || obj.acsName_ != null)
            return false;  // if
      } // if
      else {
        if (!acsName_.equals( obj.acsName_ )) return false; }


      if (epName_ == null || obj.epName_ == null) {
          if (epName_ != null || obj.epName_ != null)
            return false;  // if
      } // if
      else {
        if (!epName_.equals( obj.epName_ )) return false; }

      if (l_right_.length != obj.l_right_.length )
          return false;
      for (int i = 0; i < l_right_.length; i++) {
        if (obj.l_right_[i] == null) {
          continue; // i++
        } // if
        else {
          if (!Arrays.asList(l_right_).contains( obj.l_right_[i] ))
            return false;
        } // else
      } // for

      if (positiveRight_ != obj.getPositiveRight() ) return false;
      
      if (l_ConditionSourceGroups_ == null || obj.l_ConditionSourceGroups_ == null) {
          if (l_ConditionSourceGroups_ != null || obj.l_ConditionSourceGroups_ != null)
            return false;  // if
      } // if
      else {
        if (!l_ConditionSourceGroups_.equals( obj.l_ConditionSourceGroups_ )) return false; }      

      return true;
  } // equals(Object)

    /**
    * @return the hashcode value from the detailled name only
    */
    public int hashCode() {
        int result = 0;

        result = PRIME*result + detailledName_.hashCode();

        //if (l_ConditionSourceGroups_ != null) {
          //  result = PRIME*result + l_ConditionSourceGroups_.hashCode();
        //}

        //result = PRIME*result + (positiveRight_ ? 1:0);

        return result;
    } // hashcode()

    /**
     * Return long value for serialization
     */
    public long getSerialVersionUID()	{
            return serialVersionUID__;
    }
 
   /**
    * An external ACL has the source or the target from another ACS than the ACL ACS.
    * It may be a conditional ACL. Called by the constructor.
   * @return true if the source or the target is from another ACS
   */
  public boolean isExternalAclEntry() {
      // a finalized instance has a non-null acsName_ and a null nameOfEpAcs_
      return (!acsName_.equals(nameOfEpAcs_) || !acsName_.equals(nameOfResAcs_));
  }

  /**
  * @return boolean true if equal to new AclEntryImpl()
  */
  public boolean isEmpty() {
    return equals(new AclEntryImpl() );
  }

  //
  // ---------------- protected methods out of the implemented interfaces -----
  //
  
  /**
   * Copy the AclEntry for a resource of the same ACS.
   * The AclEntry contains the same properties than this instance, and
   * it is registered in its ACS, resource and EP.
   * @param _newRes is a resource belonging to the same ACS
   * @return an AclEntry with a new resource and a copy of the other
   *   properties.
   * @exception InternalError if _newRes is null.
   * @exception CreateError if _ep is already recorded in this resource
   *  with the same ACS and sens, or thrown by the ACS.newEorAclEntry().
   */
  public AclEntry copy(Resource _newRes) throws CreateError {
      if (_newRes == null || _newRes == getEorResource())
          throw new InternalError(" In AclEntryImpl.copy(), the argument is null "+
                  "or is this ACL resource.");
      else if (_newRes.getEorACS() != aCS_)
          throw new CreateError(" In AclEntryImpl.copy(), the argument is not "+
                  "from the same ACS.");
      
      AclEntry result = null;
      try {
          if (l_ConditionSourceGroups_ == null || isRightDefinedConditionalAclEntry())
              result = _newRes.addAclEntry(aCS_, ep_, l_right_,
                            positiveRight_, null);
          else
              result = _newRes.addAclEntry(aCS_, ep_, l_right_,
                            positiveRight_, getL_ConditionSourceGroups().iterator().next());
      }
      catch (CreateError u) {
          throw new InternalError("AclEntryImpl.copy(): "+u);
      }
      return result;
  } // copy()
  
  /**
  * Fire an event to every registered listener, in any order.
  * Called by setComment(), setRights().
  * For the properties 'Comment', 'AclRights', 'ConditionSourceGroups'.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

      // no action if the transiant variables are null
      if (changeSupport_ == null) return;
    
      synchronized (changeSupport_) 
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);} 
  } // firePropertyChange()

  /**
   * Currently unused method, that is from the DetailledName interface.
   * @exception no UpDateError throwing
   */
  public synchronized void finalizeForBase() throws UpDateError {
      
  }

  /**
   * Currently unused method, that is from the DetailledName interface.
   * @exception no UpDateError throwing
   */
  public synchronized void finalizeForUser() throws UpDateError {
      
  }

  /**
   * Replacing finalizeForBase() and finalizeForUser(), this method removes
   * the AclEntry from the aclConnectedResources map
   * of its EligibleParty, and from the list of its ACS.
   * The removing in the ACS fires a change property event to the GUI, and
   * may remove the AclEntry Resource or EP in the ACS lists if they are external.
   * Most of the variables are set to null, so the instance is equals to a new AclEntryImpl().
   * It is called only by its resource, which deletes the AclEntry
   * and its own references to it.
   * The StringRights are not finalized, but the links are cut.
   * Only ActorImpls and EligiblePartyImpls are managed as EligibleParties.
   * @param _epCount is the number of references to ep_ in the AclEntries list
   *  of its Resource. If it is equal to '1', the aclConnectedResources map
   *    of ep_ is updated.
   * @exception UpDateError if the Resource is unknown of the main ACS,
   *    or if this AclEntry is recorded more than once.
   */
  protected synchronized void finalizeForUser(short _epCount) throws UpDateError {
    if (ep_ != null && _epCount == 1) {
    // removes acl in the eligible party, with test
    // if it is an ActorImpl or an EligiblePartyImpl,
    // since removeAclEntryToEP() is not in the EligibleParty interface
    // to be 'protected'
    // count this eligible party in _acl.getEorResource().getL_aclEntry()

               //for testing and tracking
              //System.out.println("Call of AclEntryImpl.finalizeForUser() for: "+getName());

      try {
        Class ep_class = Class.forName("ARoad0.gBase.EligiblePartyImpl");
        Class act_class = Class.forName("ARoad0.gBase.ActorImpl");

        if (ep_class.isAssignableFrom(ep_.getClass() ) )
          ((EligiblePartyImpl)ep_).removeAclEntryToEP(this);
        else if (act_class.isAssignableFrom(ep_.getClass() ) )
          ((ActorImpl)ep_).removeAclEntryToEP(this);

        else throw new InternalError(" In AclEntryImpl.finalizeForUser(),"+
          " ep_ is not an EligiblePartyImpl nor an ActorImpl.");
      } catch (ClassNotFoundException f) {
        throw new InternalError(" In AclEntryImpl.finalizeForUser(): " +
          f.toString() ); //  that should NEVER happen.
      }
    } // ep_ != null

    // removes the AclEntry in the acl ACS
    if (aCS_ != null) {
        // this method uses resACS_, and also resName_ in the call of this.getEorResource().
        // it uses nameOfEpAcs_, nameOfResAcs_ when fire an event.
        // The event fired send a clone of this, without external references
        // to ep_, aCS_ and resName_.
        ((ACSImpl)aCS_).removeEorAclEntry(this);
    }

    ep_ = null;
    aCS_ = null;
    //acsName_ = null; // not possible to clear, since used in ViewRightsNetworkObserver...
    resName_ = NameImpl.EMPTY_INSTANCE;
    resType_ = null;
    epName_ = null;
    resACS_ = null;
    nameOfResAcs_ = null;
    nameOfEpAcs_ = null;
    positiveRight_ = true;
    l_right_ = new StringRight[0];
    l_ConditionSourceGroups_ = null;
    detailledName_ = null;
  } // finalizeForUser()

  /**
   * This method allows the detection by the garbage collector of unused
   * closed ACLs, or copied ACLs through copyForExternal(), after the closing of their ACS,
   * or after the deleting of their resource or eligible party.
   * This method sets the properties to their default values, without updating
   * their resource or eligible party as finalizeForUser() does it.
   * It sets to null almost all variables, thus the instance becames empty.
   * It is called by the constructor, by BaseManagerImpl.saveAndCloseOpenACS()
   * and closeOpenACS(), by ResourceImpl.finalizeFromDeserialization
   * and by ACSImpl.checkClosedAclEntries(),
   * to cleanup the AclEntry instances in ACSImpl.l_ClosedAclEntries_ or in ResourceImpl
   */
   protected synchronized void finalizeClosedAcl() {
        if (ep_ == null) return;

        ep_ = null;
        aCS_ = null;
        acsName_ = null;
        resName_ = NameImpl.EMPTY_INSTANCE;
        resType_ = null;
        epName_ = null;
        resACS_ = null;
        nameOfResAcs_ = null;
        nameOfEpAcs_ = null;
        positiveRight_ = true;
        l_right_ = new StringRight[0];
        l_ConditionSourceGroups_ = null;
        detailledName_ = null;
    } // finalizeClosedAcl()

  //
  // ---------------- public methods out of the implemented interfaces -----
  //

  /**
   * The full name is unique for the Access Road program.
   * There, its is equal to the name with all the rights and all the condition groups.
   * @return a copy of the name. Null for an empty AclEntry.
   */
    public ImmutableName getFullName() {
       Name result = (Name) getName();
       if (result == null) return null;
       result.extended(" ACLRIGHTS ");
       for (int i = 0; i < l_right_.length; i++) {
         if (l_right_[i] != null)
             result.extended(l_right_[i].getRight());
       } // for i
       if (l_ConditionSourceGroups_ != null) {
           result.extended(" CONDITIONS ");
           for (ImmutableGroupID obj : l_ConditionSourceGroups_) {
             result.extended(obj.getNickName());
           } // for obj
       } // if
       return result;
    } // getFullName()
    
  /**
   * This short name is NOT unique for the Access Road program nor the ACS if
   * it is an ACS object. The format is: <P>
   * EligibleParty Name last component >> first right nick name >> Resource Name last component
   * @return the AclEntry nick name. Never null.
   */
    public String getNickName() {
       StringBuilder result =  new StringBuilder(epName_.lastComponent());
       result.append(" >> ");
       result.append(getNoConditionalRight()[0].getNickName());
       result.append(" >> ");
       result.append(resName_.lastComponent());
       return result.toString();
    }  
   
    /**
     * Gets the BaseObject name of the AclEntry from getNameFromDetailledName().
     * The AclEntry name is composed of: <P>
     * ACS:: acsName::  ELIGIBLEPARTY:: epName::  >> RESOURCE:: resName || 
     * grant/deny || CONDITIONS:: first condition group name <P>
     * (here, ':: ' is added by gBase.NameImpl.toString()),
     * where epName is the eligible Party full name, and resName is the Resource full name. <P>
     * The condition group is put in the name or the detailled name only for
     * a flexible conditional AclEntry.
     * If _acsName, _epName or _resName is null, like in an empty instance,
     * they are not put in the name. This method works for a clone.
     * This name is used only outside gBase, following BaseObject interface.
     * The IS name is in the first components of the ACS name.
     * @return ImmutableName of the AclEntry, or null if it is a finalized AclEntry
     */
    public ImmutableName getName() {
      try {
          return getNameFromDetailledName(aCS_, detailledName_);
      } catch (BaseError b) {
          return null;
      }
    } // getName()

  /**
   * This method is to use only for the flexible conditional ACL, and if there
   * is alrady a non-null first condition group set by the constructor.
   * If the condition group is immutable, the operation is not aborted.
   * Controls the ACS type policy on 'Resource.NoConditionalACL'.
   * An external AclEntry handles only one condition group.
   * To activate the ACL rights, the source has to be member of the condition GroupIDs,
   * or, if the source is an Actor, has to have an AG context which is member of the condition GroupIDs.
   * If the AclEntry uses right-defined condition group(s),
   * it is not possible to update them directly. If a condition group is finalized,
   * this AclEntry does not listen the event but the method getRight() returns empty.
   * Controls of the ACL policy for 'Resource.NoConditionalACL', 'GroupID.ConditionalAclGroup',
   * 'Resource.OneConditionGroupInACL', 'Resource.TypesOfSourceForConditionalACL.resource_type'.
   * Calls addConditionSourceGroup() and resetRightsAsInheritedAclEntry().
   * @param _grp condition group from the AclEntry ACS. Never null.
   * @exception UpDateError if the AclEntry ACS does not manage the flexible
   * conditional group, if the first condition group is null, already known, 
   * if the argument is null or not from the AclEntry ACS, if it is an external AclEntry
   */
  public void addForUserConditionSourceGroup(ImmutableGroupID _grp) throws UpDateError {
        String[] l_remove = null;
        boolean error = false;

        if (!aCS_.getManageFlexibleConditionalAcls()) {
          throw new UpDateError(" This ACS does not manage the simple conditional AclEntries.");
        }

        if (l_ConditionSourceGroups_ == null) {
          throw new UpDateError(" This first condition group is null, thus no group may be added.");
        }

        if (_grp == null) {
          throw new UpDateError(" The AclEntry requires a non-null condition group.");
        }

        if (isExternalAclEntry()) {
            throw new UpDateError(" Only one condition group is allowed to "+
                    "an external AclEntry.");
        } // if

        if (_grp.getEorACS() != ep_.getEorACS() && 
                !_grp.getAcsName().equals(nameOfResAcs_) ) {
            throw new UpDateError(" A condition group has to be from the source ACS "+
                    "or from the target ACS.");
        } // if

        if (l_ConditionSourceGroups_.contains(_grp)) {
          throw new UpDateError(" This condition group is already known.");
        }

        if (aCS_.getM_TypeOrientedPolicy() != null &&
            getEorResource().getResourceType() != null) {
          // no adding of the ACL if conditional ACL is not allowed
            l_remove = (String[]) aCS_.getM_TypeOrientedPolicy().
                      get("Resource.NoConditionalACL");
            if (l_remove != null) {
                for (int i = 0; i < l_remove.length; i++) {
                    if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    getEorResource().getResourceType()))) {
                        error = true;
                        break;
                    }
                } // for i
            } // if (l_remove != null)

            if (error)
                throw new UpDateError(" With the type policy 'Resource.NoConditionalACL',"+
                        " the Resource type forbids this ACL "+
                        "to be a conditional AclEntry.");
        } // if (aCS_.getM_TypeOrientedPolicy() != null)
        l_remove = null;

        if (aCS_.getM_TypeOrientedPolicy() != null &&
            getEorResource().getResourceType() != null) {
          // no adding if the group cannot be a condition group
            l_remove = (String[]) aCS_.getM_TypeOrientedPolicy().
                      get("GroupID.ConditionalAclGroup");
            error = true;
            if (l_remove != null) {
                for (int i = 0; i < l_remove.length; i++) {
                    if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    _grp.getType()))) {
                        error = false;
                        break;
                    }
                } // for i
            } // if (l_remove != null)

            if (error)
                throw new UpDateError(" With the type policy 'GroupID.ConditionalAclGroup',"+
                        " the ACS forbids this group to be "+
                        "a condition group for ACL.");
        } // if (aCS_.getM_TypeOrientedPolicy() != null)
        l_remove = null;

        if (aCS_.getM_TypeOrientedPolicy() != null &&
            getEorResource().getResourceType() != null) {
          // no adding if this conditional ACL cannot have more than one condition group
            l_remove = (String[]) aCS_.getM_TypeOrientedPolicy().
                      get("Resource.OneConditionGroupInACL");
            if (l_remove != null && l_ConditionSourceGroups_ != null &&
                    !l_ConditionSourceGroups_.isEmpty()) {
                for (int i = 0; i < l_remove.length; i++) {
                    if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    getEorResource().getResourceType()))) {
                        error = true;
                        break;
                    }
                } // for i
            } // if (l_remove != null) 

            if (error)
                throw new UpDateError(" With the type policy 'Resource.OneConditionGroupInACL',"+
                        " the ACS forbids this AclEntry "+
                        "to have more than one condition group.");
        } // if (aCS_.getM_TypeOrientedPolicy() != null)
        l_remove = null;

        if (aCS_.getM_TypeOrientedPolicy() != null &&
            getEorResource().getResourceType() != null) {
          // no adding if this conditional ACL cannot have this source
            l_remove = (String[]) aCS_.getM_TypeOrientedPolicy().
                      get("Resource.TypesOfSourceForConditionalACL."+
                        UtilityImpl.getFirstInstance().  
                            getObjectTypeWithoutImmutable(
                                    getEorResource().getResourceType()));
            if (l_remove != null) {
                error = true;
                for (int i = 0; i < l_remove.length; i++) {
                    if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    getEorEP().getEPType()))) {
                        error = false;
                        break;
                    }
                } // for i
            } // if (l_remove != null) 

            if (error)
                throw new UpDateError(" With the type policy 'Resource.TypesOfSourceForConditionalACL.xxx',"+
                        " the ACS forbids this conditional ACL to have"+
                        " this type 'xxx' of source.");
        } // if (aCS_.getM_TypeOrientedPolicy() != null)
        l_remove = null;
        
        if (_grp != null) {
          addConditionSourceGroup(_grp);
          resetRightsAsInheritedAclEntry();
        } // if
  } // addForUserConditionSourceGroup()

  /**
   * To activate the ACL rights, the EP has to be member of the condition GroupIDs,
   * if it is a GroupIDMember. The first condition group cannot be removed.
   * For a right-defined AclEntry, it is not possible to update them directly.
   * If a condition group is finalized,
   * this AclEntry does not listen the event but the method getRight() returns empty.
   * Calls removeConditionSourceGroup() and resetRightsAsInheritedAclEntry().
   * @param _grp conditionnal group to remove
   * @exception UpDateError if the AclEntry uses static condition group, or the argument
   * is not from the AclEntry ACS or it is the first condition group
   */
  public void removeForUserConditionSourceGroup(ImmutableGroupID _grp) throws UpDateError {

                  // for testing and tracking
                  //System.out.println("AclEntryImpl.removeForUserConditionSourceGroup() starting ");

      if (!aCS_.getManageFlexibleConditionalAcls()) {
          throw new UpDateError(" This ACL may change its condition groups only "+
                  "through an editing of the rights,"+
                  " if it is allowed by the ACS policy.");
      }

      try {
          removeConditionSourceGroup(_grp);
      }
      catch (CreateError c) {
          throw new UpDateError(c.toString());
      }

      resetRightsAsInheritedAclEntry();
  } // removeForUserConditionSourceGroup()

  /**
  * Extracts the BaseObject name from the detailled name of an AclEntry.
  * The form of the detailled name for an internal ACL is: <P>
  * - incomplete Eligible Party name >> incomplete Resource name ||
  * grant OR deny || first condition group last component. <P>
  * For an external ACL, the form of the detailled name is: <P>
  * - ACS: ACS_name | (full or incomplete) Eligible Party name >>
  * (full or incomplete) Resource name || grant OR deny ||
  * &lt;Source ACS group&gt; OR &lt;Target ACS group&gt; ||
  * first condition group last component. <P>
  * An Eligible Party or Resource name is incomplete if it belongs to the AclEntry ACS.
  * The condition group is put in the name or the detailled name only for
  * a flexible conditional AclEntry. '&lt;XXXX ACS group&gt;' is used
  * only for an external ACL, when the condition group belongs to the XXXX ACS. <P>
  * The AclEntry name is composed of: <P>
  * ACS:: acsName::  ELIGIBLEPARTY:: epName::  >> RESOURCE:: resName ||
  * grant/deny || CONDITIONS:: first condition group name <P>
  * (here, ':: ' is added by gBase.NameImpl.toString()),
  * where epName is the Eligible Party full name, and resName is
  * the Resource full name.<P>
  * The condition group is put in the name or the detailled name only for
  * a flexible conditional AclEntry.
  * This is a pseudo-static method, since it does not depend on the instance.
  * Calls BaseUtilityImpl.getEorBaseObject().
  * Called by getName(), TreeManager.analyzeSelectedNodeInExplorer() and
  * SimplePropertyEditor.viewOneElement().
  * @param _detailledName is the viewable name of the object in the explorer
  * @return name of the AclEntry
  * @exception BaseError if an argument is null, or if _detailledName is not well formed
  */
  public ImmutableName getNameFromDetailledName(ImmutableACS _acs, String _detailledName)
    throws BaseError {
      if (_acs == null) {
          InternalError i = new InternalError(
                  "AclEntryImpl.getNameFromDetailledName() with a null acs");
          i.printStackTrace();
      } // if
      if (_detailledName == null) {
          InternalError i = new InternalError(
                  "AclEntryImpl.getNameFromDetailledName() with a null detailled name");
          i.printStackTrace();
      } // if

                  // for testing and tracking
                  //System.out.println("AclEntryImpl.getNameFromDetailledName() starts with "+
                    //"_acs: "+_acs.getNickName()+" and _detailledName: "+_detailledName);

      ImmutableName[]   l_Names = getKeyPropertiesFromDetailledName(_detailledName);
      boolean           external = false;
      boolean           externalFromSource = false;
      ImmutableName     acs2name = null;

      if (!_acs.containsOwnEligiblePartyName(
              ((Name)_acs.getName()).extended(l_Names[1])) ) {
          acs2name = ((ACSObject)BaseUtilityImpl.getEorBaseObject(l_Names[1])).getAcsName();
          external = true;
          externalFromSource = true;
      }
      if (!external && !_acs.containsOwnResourceName(
              ((Name)_acs.getName()).extended(l_Names[0])) ) {
          acs2name = ((ACSObject)BaseUtilityImpl.getEorBaseObject(l_Names[0])).getAcsName();
          external = true;
      }

                  // for testing and tracking
                  //System.out.println("AclEntryImpl.getNameFromDetailledName() external: "+
                    //external+" externalFromSource: "+externalFromSource+
                    //" acs2name: "+acs2name);

      Name name = new NameImpl("ACS");
      name.extended(_acs.getName());
      name.extended("ELIGIBLEPARTY" );
      if (external && externalFromSource) {
          name.extended(l_Names[1] );
      }
      else {
          name.extended(((Name)_acs.getName()).extended(l_Names[1]) );
      }
      name.extended(" >> RESOURCE" );
      if (external && !externalFromSource) {
          name.extended(l_Names[0] );
      }
      else {
          name.extended(((Name)_acs.getName()).extended(l_Names[0]) );
      }
      name.extended(" || " );
      name.extended(l_Names[2].lastComponent());
      if (l_Names.length == 4) {
          name.extended(" || CONDITIONS" );
          if (!external) {
              name.extended(((Name)_acs.getName())).extended("<G>").extended(l_Names[3].
                  lastComponent());
          }
          else {
            if (_detailledName.indexOf("<Source ACS group>") != -1) {
                // use _acs or acs2name
              if (externalFromSource)
                  name.extended(acs2name).extended("<G>").
                          extended(l_Names[3].lastComponent());
              else
                  name.extended(_acs.getName()).extended("<G>").
                          extended(l_Names[3].lastComponent());
            }
            else {
                // inverse use _acs or acs2name
              if (externalFromSource)
                  name.extended(_acs.getName()).extended("<G>").
                          extended(l_Names[3].lastComponent());
              else
                  name.extended(acs2name).extended("<G>").
                          extended(l_Names[3].lastComponent());
            } // else
          } // else
      } // if

                  // for testing and tracking
                  //System.out.println("AclEntryImpl.getNameFromDetailledName() result: "+
                    //name);

      l_Names = null;
      acs2name = null;
      return name;
  } // getNameFromDetailledName()

  /**
  * Extracts the resource name first, and the EligibleParty name
  * in the detailled name of an AclEntry, then the right sens and, if not null,
  * the first condition group name. The ACS name is never returned.
  * The form of the detailled name for an internal ACL is as following: <P>
  * - incomplete Eligible Party name >> incomplete Resource name ||
  * grant OR deny || first condition group last component. <P>
  * For an external ACL, the form of the detailled name is as following: <P>
  * - ACS: ACS_name | (full or incomplete) Eligible Party name >>
  * (full or incomplete) Resource name || grant OR deny ||
  * &lt;Source ACS group&gt; OR &lt;Target ACS group&gt; ||
  * first condition group last component. <P>
  * An Eligible Party or Resource name is incomplete if it belongs to the AclEntry ACS.
  * The condition group is put in the name or the detailled name only for
  * a flexible conditional AclEntry. '&lt;XXXX ACS group&gt;' is used
  * only for an external ACL, when the condition group belongs to the XXXX ACS. <P>
  * This is a pseudo-static method, since it does not depend on the instance.
  * Used by the method getNameFromDetailledName.
  * @see #getDetailledName
  * @param _displayedName is the short viewable name of an AclEntry
  * @return array of 3 or 4 names, with
  *   - at the index 0, the resource name, without the ACS name if it belongs to the ACL ACS
  *   - at the index 1, the eligible party name, without the ACS name if it belongs to the ACL ACS
  *   - at the index 2, 'grant' or 'deny' in a one-component name
  *   - at the index 3, the first condition group name, without '&lt;G&gt;' nor its ACS name (as option)
  * @exception BaseError if the argument is not an AclEntry name string,
  *  or if a main component string is not provided by NameImpl.toString().
  */
  public ImmutableName[] getKeyPropertiesFromDetailledName(
    String _displayedName) throws BaseError {
      
      if (_displayedName.indexOf(">>") == -1 ||
            _displayedName.indexOf(" || ") == -1)
            throw new BaseError(" In AclEntryImpl.getKeyPropertiesFromDetailledName()"+
              " the argument is not a true AclEntry name.");
      
      // variables
      ImmutableName[]   l_result = null;
      
      // first case: internal ACL
      if (!_displayedName.startsWith("ACS: ") && _displayedName.indexOf(" | ") == -1) {
          if (_displayedName.indexOf(" || ") != _displayedName.lastIndexOf(" || "))
              l_result = new ImmutableName[4];
          else
              l_result = new ImmutableName[3];

          // extract the Resource name
          l_result[0] = NameImpl.getNameFromString(
            _displayedName.substring(_displayedName.indexOf(">>") + 2,
            _displayedName.indexOf(" || ")) );
          // extract the EligibleParty name
          l_result[1] = NameImpl.getNameFromString(
            _displayedName.substring(0, _displayedName.indexOf(">>") ) );
          // extract the right
          if (l_result.length == 3)
              l_result[2] = new NameImpl(
                _displayedName.substring(_displayedName.indexOf(" || ") + 4,
                _displayedName.length() ) );
          else
              l_result[2] = new NameImpl(
                _displayedName.substring(_displayedName.indexOf(" || ") + 4,
                _displayedName.lastIndexOf(" || ") ) );
          // extract the first condition group, if needed
          if (l_result.length == 4)
              l_result[3] = new NameImpl(
                _displayedName.substring(_displayedName.lastIndexOf(" || ") + 4,
                _displayedName.length() ) );

          return l_result;
      }
      else {
          // process an external ACL
          ImmutableACS      acs = null;
          //ImmutableName     name = null;
          String            noACSdisplayedName = _displayedName.substring(
                                            _displayedName.indexOf(" | ") + 2);

          acs = BaseManagerImpl.getInstance().getEorM_OpenACS().get(
                  NameImpl.getNameFromString(
                  _displayedName.substring(4, _displayedName.indexOf("| "))));


                  // for testing and tracking
                  //System.out.println("AclEntryImpl.getKeyPropertiesFromDetailledName() with "+
                    //      "noACSdisplayedName: "+noACSdisplayedName);

          if (noACSdisplayedName.indexOf("<Source ACS group>") != -1 ||
                 noACSdisplayedName.indexOf("<Target ACS group>") != -1)
              l_result = new ImmutableName[4];
          else
              l_result = new ImmutableName[3];

          // extract the (full or incomplete) Resource name
          l_result[0] = NameImpl.getNameFromString(
            noACSdisplayedName.substring(noACSdisplayedName.indexOf(">>") + 2,
            noACSdisplayedName.indexOf(" || ")) );
          /*
          // this assumes that acs.getName() extended by 'name'
          // does not exist if 'name' is a full name which starts with an ACS name
          if (acs.containsOwnResourceName(((NameImpl)acs.getName()).extended(name)))
              l_result[0] = ((NameImpl)acs.getName()).extended(name);
          else {
              try {
                  BaseUtilityImpl.getEorBaseObject(name);
              }
              catch (BaseError b) {
                  throw new BaseError(" In AclEntryImpl.getKeyPropertiesFromDetailledName()"+
                    " the argument contains a wrong AclEntry Resource name.");
              }
              l_result[0] = name;
          } // else
           */

          // extract the (full or incomplete) EligibleParty name
          l_result[1] = NameImpl.getNameFromString(
            noACSdisplayedName.substring(0, noACSdisplayedName.indexOf(">>") ) );

          // extract the right
          if (l_result.length == 3) {
              l_result[2] = new NameImpl(
                noACSdisplayedName.substring(noACSdisplayedName.indexOf(" || ") + 4,
                noACSdisplayedName.length() ) );
          }
          else {
              if (noACSdisplayedName.substring(noACSdisplayedName.indexOf(" || ") + 4,
                noACSdisplayedName.indexOf(" || ") + 9 ).equals("grant") )
                    l_result[2] = new NameImpl("grant");
              else
                    l_result[2] = new NameImpl("deny");
          }
          // extract the first condition group, if needed
          if (l_result.length == 4) {
              l_result[3] = new NameImpl(
                    noACSdisplayedName.substring(noACSdisplayedName.lastIndexOf(" || ") + 4,
                    noACSdisplayedName.length() ) );
          } // if

                      // for testing and tracking
                      //if (l_result.length == 4) {
                        //  System.out.println("AclEntryImpl.getKeyPropertiesFromDetailledName() result: ");
                          //  System.out.println("--- Resource name: "+l_result[0]);
                            //System.out.println("--- EP name: "+l_result[1]);
                            //System.out.println("--- sens: "+l_result[2]);
                            //System.out.println("--- condition group: "+l_result[3]);
                      //} // if
                      //else {
                        //  System.out.println("AclEntryImpl.getKeyPropertiesFromDetailledName() result: ");
                          //  System.out.println("--- Resource name: "+l_result[0]);
                            //System.out.println("--- EP name: "+l_result[1]);
                            //System.out.println("--- sens: "+l_result[2]);
                      //}


          noACSdisplayedName = null;
          acs = null;
          //name = null;
          return l_result;
      } // else
  } // getKeyPropertiesFromDetailledName()

  /**
  * Gets the detailled name from the AclEntry name, like a static method since
  * the AclEntry properties are not used.
  * The form of the detailled name for an internal ACL is as following: <P>
  * - incomplete Eligible Party name >> incomplete Resource name ||
  * grant OR deny || first condition group last component. <P>
  * For an external ACL, the form of the detailled name is as following: <P>
  * - ACS: ACS_name | (full or incomplete) Eligible Party name >>
  * (full or incomplete) Resource name || grant OR deny ||
  * &lt;Source ACS group&gt; OR &lt;Target ACS group&gt; ||
  * first condition group last component. <P>
  * The AclEntry name is composed of: <P>
  * ACS:: acsName::  ELIGIBLEPARTY:: epName::  >> RESOURCE:: resName ||
  * grant/deny || CONDITIONS:: first condition group name <P>
  * Note: the detailled name is not unique in the base.
  * Calls BaseManagerImpl.getEorM_OpenACS() and knowsACS() to find the ACS name.
  * Calls getKeyPropertiesFromName() and buildUpDetailledName().
  * @param _name name of an AclEntry
  * @return the detailled name of the AclEntry
  * @exception BaseError if the argument has a wrong format
  */
  public String getDetailledNameFromName(ImmutableName _name) throws BaseError {
      
    ImmutableName[] l_names = getKeyPropertiesFromName(_name);

    if (l_names[0] == null)
        throw new BaseError(" In AclEntryImpl.getDetailledNameFromName(), the argument "+
                _name+" does not contain an ACS name.");

    boolean         positiveRight = false;
    String          result = null;
    boolean         isExternal = true;
    ACS             acs = (ACS) BaseManagerImpl.getInstance().getEorM_OpenACS().get(l_names[0]);
    String          condition = null;
    String          name = null;
    ImmutableName   conditionGroupACSname = null;
    StringBuilder   sb = null;
    ImmutableName   epAcsName = null;
    ImmutableName   resAcsName = null;
    BaseManagerImpl base = BaseManagerImpl.getInstance();

    if (acs == null)
        throw new BaseError(" In AclEntryImpl.getDetailledNameFromName(), the argument "+
                _name+" does not contain the ACS name of an open ACS.");

    if (l_names[3].lastComponent().equals("grant")) positiveRight = true;

    // if the EP and the Resource are from the ACL ACS, the ACL is not external
    // note: l_names[1].context(l_names[0].noOfComponents() - 1).equals(l_names[0])
    // is not used since the Resource ACS name may start with the ACL ACS name.
    if (acs.containsOwnEligiblePartyName(l_names[1]) &&  acs.containsOwnResourceName(l_names[2])) {
                isExternal = false;
    } // if


                      // for testing and tracking
                      //System.out.println("AclEntryImpl.getDetailledNameFromName() for an internal ACL: "+isExternal);
                        //System.out.println("ACS name: "+l_names[0]);
                      //System.out.println(" with Resource name: "+
                          //  l_names[1].subname(l_names[0].noOfComponents(), l_names[1].noOfComponents()));
                        //System.out.println("EP name: "+
                          //  l_names[2].subname(l_names[0].noOfComponents(), l_names[2].noOfComponents()));

    if (l_names.length == 5) {
            condition = l_names[4].lastComponent();
            // remove the two last components of the GroupID name
            conditionGroupACSname = l_names[4].subname(0, l_names[4].noOfComponents() - 2);
    }

    try {

                        //System.out.println("Resource name: "+
                          //  l_names[1].subname(l_names[0].noOfComponents(), l_names[1].noOfComponents()));
                        //System.out.println("EP name: "+
                          //  l_names[2].subname(l_names[0].noOfComponents(), l_names[2].noOfComponents()));

        // remove the ACS names, but for an external ACL, this ACS is not always the ACL ACS.
        // condition may be null.
        if (!isExternal) {
            
                      // for testing and tracking
                      //System.out.println("AclEntryImpl.getDetailledNameFromName() "+
                        //      "for an internal ACL");
                        //System.out.println("ACS name: "+l_names[0]);

            result = buildUpDetailledName(
                l_names[1].subname(l_names[0].noOfComponents(), l_names[1].noOfComponents()),
                l_names[2].subname(l_names[0].noOfComponents(), l_names[2].noOfComponents()),
                positiveRight, condition);

                      // for testing and tracking
                      //System.out.println("AclEntryImpl.getDetailledNameFromName() "+
                        //      "for an internal ACL");
                        //System.out.println("--- result: "+result);

        }
        else {
            epAcsName = (ImmutableName) l_names[1].clone();
            while (!base.knowsACS(epAcsName)) {

                      // for testing and tracking
                      //System.out.println("AclEntryImpl.getDetailledNameFromName() "+
                        //      "for an external ACL. Search of the EP ACS:");
                        //System.out.println("--- epAcsName: "+epAcsName);

                epAcsName = epAcsName.context();
            }
            resAcsName = (ImmutableName) l_names[2].clone();
            while (!base.knowsACS(resAcsName)) {

                      // for testing and tracking
                      //System.out.println("AclEntryImpl.getDetailledNameFromName() "+
                        //      "for an external ACL. Search of the resource ACS:");
                        //System.out.println("--- resAcsName: "+resAcsName);

                resAcsName = resAcsName.context();
            }

            result = buildUpDetailledName(
                l_names[1].subname(epAcsName.noOfComponents(), l_names[1].noOfComponents()),
                l_names[2].subname(resAcsName.noOfComponents(), l_names[2].noOfComponents()),
                positiveRight, condition);
        } // else

        // For an external ACL, the form of the detailled name is: <P>
        // - ACS: ACS_name | (full or incomplete) Eligible Party name >>
        // (full or incomplete) Resource name || grant OR deny ||
        // <Source ACS group> OR <Target ACS group> ||
        // first condition group last component.
        if (isExternal) {
            sb = new StringBuilder();

            sb.append("ACS:");
            sb.append(l_names[0].toString());
            sb.append("|");

            if (!epAcsName.equals(l_names[0])) {
                name = epAcsName.toString();
                // remove the two last spaces
                sb.append(name.substring(0, name.length() - 2));
            }
            // add 'incomplete Eligible Party name >>' with the first space
            sb.append(result.substring(
                    0,
                    result.indexOf(">>") + 2));

            if (!resAcsName.equals(l_names[0])) {
                name = resAcsName.toString();
                // remove the twp last spaces
                sb.append(name.substring(0, name.length() - 2));
            }
            // add 'incomplete Resource name ||' with the first space
            sb.append(result.substring(
                    result.indexOf(">>") + 2,
                    result.indexOf("||") + 2));

            if (result.indexOf("||") == result.lastIndexOf("||")) {
                // add 'grant OR deny'
                sb.append(result.substring(
                        result.indexOf("||") + 2,
                        result.length()));
            }
            else {
                // add 'grant OR deny ||'
                sb.append(result.substring(
                        result.indexOf("||") + 2,
                        result.lastIndexOf("||") + 2));

                // insert '<Source ACS group> OR <Target ACS group> ||'
                if (epAcsName.equals(conditionGroupACSname)) {
                    sb.append(" <Source ACS group> ||");
                }
                else {
                    sb.append(" <Target ACS group> ||");
                }

                // add 'first condition group last component'
                sb.append(result.substring(
                        result.lastIndexOf("||") + 2,
                        result.length()));

            } // else

            result = sb.toString();
                      
        } // if (isExternalAclEntry())

    }
    catch (CreateError c) {
        throw new BaseError(c.toString());
    }
    l_names = null;
    acs = null;
    condition = null;
    conditionGroupACSname = null;
    sb = null;
    name = null;
    epAcsName = null;
    resAcsName = null;
    base = null;

                      // for testing and tracking
                      //System.out.println("in AclEntryImpl.getDetailledNameFromName(), "+
                        //      "result: "+result);

    return result;
  } // getDetailledNameFromName()

   /**
    * The form is 'Eligible party: KKKK+ , ACL ACS: GGG + ; Resource: DDDD+
    * Right(s): bbbb| vvvv| nnnn + positive right: true/false"+ Condition groups:
    * UUUU| OOOO| .
    * Called by BaseManagerImpl.importACS() to display the closed AclEntries,
    * and in the ResourceImpl.addAclEntry() error messages.
   * @return string form of the AclEntry with most of the properties
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();

    sb.append("Eligible party: ");
    sb.append(epName_);

    sb.append("+ , ACL ACS: ");
  	if (acsName_ != null) {
        sb.append(acsName_);
    } else sb.append("<none>");

    sb.append(" + ; Resource: ");
    sb.append(resName_);

    sb.append("+ Right(s): ");
    if (l_right_.length != 0) {
      for (int i = 0; i < l_right_.length; i++) {
        if (l_right_[i] != null)
            sb.append(l_right_[i].toString());
        sb.append("| ");
      } //for
    } // if
    else sb.append("<none>");

    sb.append("+ positive right: ");
    sb.append(positiveRight_);
    
    if (l_ConditionSourceGroups_ != null) {
      sb.append("+ Condition groups: ");
      for (Iterator ite = l_ConditionSourceGroups_.iterator(); ite.hasNext(); ) {
        sb.append(((ImmutableGroupID)ite.next()).getName());
        sb.append("| ");
      } //for
    } // if
    
    return " "+sb;
  } // toString

  // -------------------- public static methods --------------------------------

  /**
   * Gets the component names from the AclEntry name.
   * The AclEntry name is composed of: <P>
   * ACS:: acsName::  ELIGIBLEPARTY:: epName::  >> RESOURCE:: resName ||
   * grant/deny || CONDITIONS:: first condition group name <P>
   * (here, ':: ' is added by gBase.NameImpl.toString()),
   * where epName is the EligibleParty full name, and resName is the Resource full name.
   * The condition group is put in the name or the detailled name only for
   * a flexible conditional AclEntry.
   * Called by the methods getKeyReferencesFromName, getDetailledNameFromName,
   * getPropertyNamesFromName.
   * @param _aclName name of the AclEntry.
   * @return an array of 4 or 5 names:<P>
   *  - the ACS full name at index 0,<P>
   *  - the EligibleParty full name at index 1,<P>
   *  - the Resource full name at index 2, <P>
   *  - the 'grant' or 'deny' string in a one-component name, <P>
   *  - the first condition group name at index 4, or no index 4.
   * @exception BaseError the argument is not an AclEntry name with the key words.
   */
    public static ImmutableName[] getKeyPropertiesFromName(ImmutableName _aclName)
      throws BaseError {
        if ( _aclName.noOfComponents() < 10
          || !_aclName.contains(new NameImpl("ACS"))
          || !_aclName.contains(new NameImpl("ELIGIBLEPARTY"))
          || !_aclName.contains(new NameImpl(" >> RESOURCE")))
            throw new BaseError("In AclEntryImpl.getKeyPropertiesFromName()"+
              " the argument is not an AclEntry name.");

                      // for testing and tracking
                      //System.out.println("AclEntryImpl.getKeyPropertiesFromName() starting on "+
                        //      _aclName);

        // variables
        ImmutableName[] l_result = null;
        int             epIndex = 0;
        int             resourceIndex = 0;
        int             conditionIndex = 0;

        for (int i = 0 ; i < _aclName.noOfComponents() ; i++) {
          if (_aclName.component(i).equals("ELIGIBLEPARTY")) {
            epIndex = i;
            continue;
          }
          if (_aclName.component(i).equals(" >> RESOURCE")) {
            resourceIndex = i;
            continue;
          }
          if (_aclName.component(i).equals(" || CONDITIONS")) {
            conditionIndex = i;
            break;
          }
        } // for i

        if (conditionIndex != 0)
            l_result = new NameImpl[5];
        else
            l_result = new NameImpl[4];

        l_result[0] = _aclName.subname(1, epIndex);
        l_result[1] = _aclName.subname(
                epIndex + 1, resourceIndex);

        if (conditionIndex != 0) {
            l_result[2] = _aclName.subname(
                    resourceIndex + 1, conditionIndex - 2);
            l_result[3] = _aclName.subname(
                    conditionIndex - 1, conditionIndex );
            l_result[4] = _aclName.subname(
                conditionIndex + 1, _aclName.noOfComponents() );
        }
        else {
            l_result[2] = _aclName.subname(
                    resourceIndex + 1, _aclName.noOfComponents() - 2);
            l_result[3] = _aclName.subname(
                    _aclName.noOfComponents() - 1, _aclName.noOfComponents() );
        } // else

                      // for testing and tracking
                      //System.out.println("AclEntryImpl.getKeyPropertiesFromName() result: ");
                        //System.out.println("--- ACS name: "+l_result[0]);
                        //System.out.println("--- EP name: "+l_result[1]);
                        //System.out.println("--- resource name: "+l_result[2]);
                        //System.out.println("--- right: "+l_result[3]);

        return l_result;
    } // getKeyPropertiesFromName()

  /**
   * Gets the 4 or 5 AclEntry main components from the AclEntry name.
   * These components fully define the AclEntry instance.
   * The AclEntry name is composed of: <P>
   * ACS:: acsName::  ELIGIBLEPARTY:: epName::  >> RESOURCE:: resName ||
   * grant/deny || CONDITIONS:: first condition group name <P>
   * where epName is the EligibleParty full name, and resName is
   * the Resource full name.
   * The condition group is put in the name or the detailled name only for
   * a flexible conditional AclEntry.
   * To have names rather than instances, use getKeyPropertiesFromName().
   * This method is used by Gui1.PropertyEditor to remove
   * an AclEntry from a list displaying the AclEntry name in the beamer.
   * This method is returned by the PropertyDescriptor value 'removeArgumentsFromName'
   * in ResourceImplBeanInfo, for CollectionPropertyEditor.
   * @param _aclName the AclEntry name, get through getName()
   * @return array of the 4 or 5 AclEntry main components:<P>
   *  - index 0: ACS which manages this AclEntry<P>
   *  - index 1: AclEntry EligibleParty<P>
   *  - index 2: AclEntry Resource<P>
   *  - index 3: Boolean for the right sens
   *  - index 4: the first condition group
   * @exception BaseError if the argument is not an AclEntry name string,
   *  or if a main component string is not provided by NameImpl.toString()
   */
    public static Object[] getKeyReferencesFromName(ImmutableName _aclName)
      throws BaseError {

      // variables
      ImmutableName[]   l_resultNames = getKeyPropertiesFromName(_aclName);
      Object[]          l_result = new Object[l_resultNames.length];

      l_result[0] = BaseUtilityImpl.getEorBaseObject(l_resultNames[0]);
      l_result[1] = BaseUtilityImpl.getEorBaseObject(l_resultNames[1]);
      l_result[2] = BaseUtilityImpl.getEorBaseObject(l_resultNames[2]);

      if (l_resultNames[3].lastComponent().equals("grant"))
        l_result[3] = Boolean.TRUE;
      else l_result[3] = Boolean.FALSE;

      if (l_resultNames.length == 5)
        l_result[4] = BaseUtilityImpl.getEorBaseObject(l_resultNames[4]);

      l_resultNames = null;
      return l_result;
    } // getKeyReferencesFromName()

  /**
   * Gets the main components names from the aclEntry name as a string,
   * returned by toString(). These key components names define completly
   * the AclEntry instance, following the KeyPropertiesInName pattern.
   * The AclEntry name is composed of: <P>
   * ACS:: acsName::  ELIGIBLEPARTY:: epName::  >> RESOURCE:: resName ||
   * grant/deny || CONDITIONS:: first condition group name <P>
   * The condition group is put in the name or the detailled name only for
   * a flexible conditional AclEntry.
   * @param _aclName the AclEntry name, get through getName().toString()
   * @return array of the 4 or 5 AclEntry main components names:<P>
   *  - index 0: name of the ACS which manages this AclEntry<P>
   *  - index 1: name of the EligibleParty<P>
   *  - index 2: name of the Resource<P>
   *  - index 3: the 'grant' or 'deny' string in a one-component name
   *  - index 4: the first condition group
   * @exception BaseError if the argument is not an AclEntry name string,
   *  or if a main component string is not provided by NameImpl.toString().
   */
    public static ImmutableName[] getPropertyNamesFromName(String _aclName)
      throws BaseError {
        return getKeyPropertiesFromName(NameImpl.getNameFromString(_aclName));
    } // getPropertyNamesFromName()

  //
  // --------- protected  methods --------------------------------
  //

  /**
   * Sets the rights of the AclEntry, from which the effective rights
   * given by getRight() are derived. Only header rights are used.
   * The ACS constraints on 'AclRights' are applied even if they are null,
   * and they depend on both the EligibleParty and the Resource. This is not
   * like for AGO rights. To explain the result, an exception is thrown as an user
   * message when the ACS constraints are no fullfilled.
   * Controls the ACL rights constraints in the ACS from the resource,
   * and also from the EP only if the AclEntry is not external (that is, the ACS
   * is the same for the AclEntry and the EP. <P>
   * Only non-null, non-empty rights are set as new rights.
   * Creates a 0-size array if the argument is null.
   * Controls if each right has the correct sens.
   * Controls the ACS type policy on 'Resource.OneRightACL', 'Resource.NoConditionalACL',
   * 'Resource.NoNonConditionalACL', 'GroupIDMember.IsNotConditionalAclSource' and
   * 'Actor.IsNotConditionalAclSource'. <P>
   * A right-defined conditional AclEntry must have a single non-null right.
   * It may be be external.
   * For a conditional AclEntry, _ep is a GroupIDMember, and the right is workable
   * only if _ep is a member of each of these condition groups (or the _ep AG context).
   * Otherwise, the resulting effective right from getRight() is a 0-size array.
   * Calls BaseUtilityImpl.getL_AcsRestrictedRights() and UtilityImpl.selectHeaders(). <P>
   * Note: 'AclRights' is also the bound property name of the ACLs in ResourceImpl,
   * to allow the work of the Gui1 package. For Gui2.GraphicViewBaseListenerImpl
   * that listens the 'AclRights' events both from the Resource and the AclEntry,
   * the source and the value types in the event allows to process the event properly. <P>
   * No action if this instance is a clone or has been finalized.
   * Called by the constructor and setRightsForUser().
   * @see BaseUtilityImpl#getL_AcsRestrictedRights
   * @param _st array of non-null, non-empty rights
   *    for the couple (resource, EligibleParty) in the AclEntry ACS.
   *    The authorized rights depend on the AclEntry ACS. May be null.
   * @exception CreateError if a right is null, empty,
   *    if a right is not allowed, has not the sens of getPositiveRight(),
   *    or does not fulfil with the ACL constraints.
   */
   protected void setL_Rights(StringRight[] _st) throws CreateError {
                 // for testing and tracking
                 //System.out.println("In AclEntryImpl.setL_Rights(), the arguments are:");
                 //if (_st != null) {
                   //for (int i = 0; i < _st.length ; i++) {
                     //   System.out.println("--------- "+_st[i]);
                       // if (!(_st[i] instanceof StringMetaRightImpl)) continue;
                       // if (((StringMetaRightImpl)_st[i]).m_ACSRights_ == null) continue;

                        //Map.Entry entry = null;
                        //System.out.println("and its m_ACSRights_ if not null:");
                        //for (Iterator ite = ((StringMetaRightImpl)_st[i]).m_ACSRights_.entrySet().iterator(); ite.hasNext(); ) {
                          //  entry = (Map.Entry)ite.next();
                            //if (entry.getValue() instanceof StringRight)
                              //  System.out.println("-- for "+entry.getKey()+" value: "+entry.getValue());
                            //else if (entry.getValue() instanceof Set) {
                              //  System.out.println("-- for "+entry.getKey()+", the set is: ");
                                //for (Iterator ita = ((Set)entry.getValue()).iterator(); ita.hasNext(); ) {
                                  //  System.out.println("----- value: "+(StringRight)ita.next());
                                //} // for ita
                            //} // else if
                        //} // for ite
                    //} // for
                 //}
                 //else
                   //  System.out.println("--------- null array ");
                 
       
       if (aCS_ == null) {
                 // for testing and tracking
                 //System.out.println("In AclEntryImpl.setL_Rights(), noop since aCS_ is null");
           
           return; // it is a clone or it has been finalized
       }
       
        boolean errorOne = false;
        boolean errorTwo = false;
        boolean errorThree = false;
        boolean errorFour = false;
        boolean errorFive = false;
        boolean errorSix = false;
        boolean errorSeven = false;
        boolean errorEight = false;
        boolean errorNine = false;
        boolean errorTen = false;
        int index = 0;
        if (_st != null && _st.length != 0 ) {
            for (int i = 0; i < _st.length; i++) {
              if (_st[i] == null || _st[i].isEmpty() ) {
                errorOne = true;
                index = i;
                break;
              } // if
              if ((_st[i].isGrantingRight() && !positiveRight_) ||
                      (!_st[i].isGrantingRight() && positiveRight_)) {
                errorTwo = true;
                index = i;
                break;
              } // if
              if (_st[i].getRight().indexOf("<IF>") != -1) {

                if (_st[i].getRight().lastIndexOf("<IF>") + 4 == _st[i].getRight().length()) {
                    errorThree = true;
                    break;
                } // if
                if (_st.length > 1) {
                    errorFour = true;
                    break;
                } // if
                if (!nameOfEpAcs_.equals(acsName_)) {
                    errorFive = true;
                    break;
                } // if

              } // if (_st[i].getRight().indexOf("<IF>") != -1)

            } // for i
        } // if

        if (!isRightDefinedConditionalAclEntry() &&
                (_st != null && _st.length != 0 &&
                _st[0].getRight().indexOf("<IF>") != -1) &&
                !getL_ConditionSourceGroups().isEmpty()) {
            errorEight = true;
        } // if

        String[] l_remove = null;
        if (aCS_.getM_TypeOrientedPolicy() != null) {

            l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                      get("Resource.OneRightACL");
            if (l_remove != null) {
                  for (int i = 0; i < l_remove.length; i++) {
                      
                                 // for testing and tracking
                                 //System.out.println("In AclEntryImpl.setL_Rights(), for 'Resource.OneRightACL' "+
                                   //      "l_remove[i]: ");                        
                      
                      if (l_remove[i].equals(
                                UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    getEorResource().getResourceType())) &&
                              _st.length > 1) {
                          errorFour = true;
                          break;
                      }
                  } // for i
            } // if (l_remove !=

            if (aCS_.getManageRightDefinedConditionalAcls() &&
                getEorResource().getResourceType() != null) {
              // no adding of the ACL if it is not a right-defined conditional ACL
                l_remove = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Resource.NoConditionalACL");
                if (l_remove != null) {
                    for (int i = 0; i < l_remove.length; i++) {
                        if (l_remove[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        getEorResource().getResourceType()))) {
                            errorSeven = true;
                            break;
                        }
                    } // for i
                } // if (l_types != null)

                if (errorSeven) {
                    errorSeven = false;
                    if (_st != null && _st.length != 0 && _st[0].getRight().indexOf("<IF>") != -1)
                        errorSeven = true;
                } // if (errorSix)
            } // if (aCS_.getManageRightDefinedConditionalAcls()

            l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                      get("GroupIDMember.IsNotConditionalAclSource");
            if (ep_ instanceof GroupIDMember && l_remove != null) {
                  for (int i = 0; i < l_remove.length; i++) {

                                 // for testing and tracking
                                 //System.out.println("In AclEntryImpl.setL_Rights(), for 'Resource.OneRightACL' "+
                                   //      "l_remove[i]: ");

                      if (l_remove[i].equals(
                                UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    ep_.getEPType()))) {
                          errorNine = true;
                          break;
                      }
                  } // for i
            } // if (l_remove !=

            l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                      get("Actor.IsNotConditionalAclSource");
            if (ep_ instanceof Actor && l_remove != null) {
                  for (int i = 0; i < l_remove.length; i++) {

                                 // for testing and tracking
                                 //System.out.println("In AclEntryImpl.setL_Rights(), for 'Resource.OneRightACL' "+
                                   //      "l_remove[i]: ");

                      if (l_remove[i].equals(
                                UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    ep_.getEPType()))) {
                          errorTen = true;
                          break;
                      }
                  } // for i
            } // if (l_remove !=
            l_remove = null;
        } // if (getEorACS().getM_TypeOrientedPolicy()
        
        if (_st != null && _st.length != 0) {
            if (errorOne) throw new CreateError(
                      " The right at "+index+" is null or empty.");
            else if (errorTwo) throw new CreateError(
                      " The right "+_st[index]+" has not the correct sens in the acs "+acsName_);
        } // if
        if (errorThree) throw new CreateError(
                  " The right is a conditional one with a false format.");
        else if (errorFour) throw new CreateError(
                  " Only one right is allowed, "+
                  "because it is a right-condition ACL, or from the ACS policy.");
        else if (errorFive) throw new CreateError(
                  " The AclEntry ACS must be the EP ACS (conditional right).");
        else if (errorSix) throw new CreateError(
                  " The AclEntry must be a right-condition AclEntry.");
        else if (errorSeven) throw new CreateError(
                " With the type policy 'Resource.NoConditionalACL',"+
                  " the ACS forbids this ACL to be "+
                  "a right-condition AclEntry.");
        else if (errorEight) throw new CreateError(
                " A simple-condition AclEntry cannot become "+
                "a right-condition AclEntry.");
        else if (errorNine) throw new CreateError(
                " With the type policy 'GroupIDMember.IsNotConditionalAclSource',"+
                            " the ACS forbids this UserID or GroupID type "+
                            "as source of a simple-condition AclEntry");
        else if (errorTen) throw new CreateError(
                " With the type policy 'Actor.IsNotConditionalAclSource',"+
                            " the ACS forbids this Actor type "+
                            "as source of a simple-condition AclEntry.");

        StringRight[]       oldRight = new StringRight[l_right_.length];
        System.arraycopy(l_right_, 0, oldRight, 0, l_right_.length);
        
        List                l_limitedRights = null;
        Set                 l_newRights = new HashSet(INITIAL_CAPACITY / 2);
        StringRight[]       l_restrictedRights = null;
        StringRight         right = null;
        boolean             error = false;
        
        try {
            // controls the acl rights restriction in the ACS from the resource,
            // and the ep only if it is not external
            if (ep_.getEorACS() == aCS_)
                l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                       null, getEorResource(), "AclRights", aCS_, ep_.getEPType());
            else
                l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                       null, getEorResource(), "AclRights", aCS_, null);
        } catch (Exception c) {
            System.out.println("From AclEntryImpl.setL_Rights(), "+
                    "BaseUtilityImpl.getL_AcsRestrictedRights() throws:");
            c.printStackTrace();
            throw new CreateError(" The search on the ACS rights policy constraints "
                    +"throws an exception.");
        } // catch

        if (l_restrictedRights != null) 
            l_limitedRights = Arrays.asList( l_restrictedRights );
        else if (_st != null) {
            oldRight = null;
            l_newRights = null;
            throw new CreateError(
                  " No AclEntry right is allowed by the AclEntry ACS for this pair (ep, resource).");
        } // else
        
                         // for testing and tracking
                        //System.out.println("AclEntryImpl.setL_Rights(), the restricted rights are:");
                        //for (Iterator ite = l_limitedRights.iterator(); ite.hasNext(); ) {
                          //System.out.println("------- "+ite.next());
                        //} // for  
        
        if (_st != null) {
          for (int i = 0; i < _st.length; i++) {
            if (_st[i] != null) {
                if (l_limitedRights.contains(_st[i])) {
        
                         // for testing and tracking
                        //System.out.println("AclEntryImpl.setL_Rights(), add in l_newRights of "+_st[i]);
                        
                    l_newRights.add(_st[i]);
                } else {
                    right = _st[i];
                    errorOne = true;
                    break;
                } 
            } // if
          } // for i
          
          if (errorOne) {
            oldRight = null;
            l_newRights = null;
            l_restrictedRights = null;
            l_limitedRights = null;
            System.out.println("AclEntryImpl.setRight(), "+right.getRight()+" is not allowed.");
            throw new CreateError(
                " The ACL right "+right.getRight()+" is not allowed by the AclEntry ACS "+
                "for this pair (ep, resource).");
          } // if (errorOne
        } // if (_st != null)           
        
                         // for testing and tracking
                        //System.out.println("AclEntryImpl.setL_Rights(), the new rights are:");
                        //for (Iterator ite = l_newRights.iterator(); ite.hasNext(); ) {
                          //System.out.println("------- "+ite.next());
                        //} // for

        // updates the acl rights
        if (!l_newRights.isEmpty()) {
          try {
            UtilityImpl.getFirstInstance().selectHeaders(l_newRights);
          }
          catch (BaseError b) {
              throw new CreateError(" The new AclEntry rights cannot be set. "+b.toString());
          }
          l_right_ = (StringRight[])
                  l_newRights.toArray(new StringRight[l_newRights.size()]);
        } else {
          l_right_ = new StringRight[0];
        }

        // updates the conditional groups, if any
        int re = setConditionSourceGroupsFromRights();

        if (re < 0) throw new CreateError(" For this conditional right, "+re+
                " groups are unknown by the AclEntry ACS. The rights have been updated. They "+
                "should be deleted, or the missing groups should be created.");

        // set empty rights if 'Resource.NoNonConditionalACL' is not applied
        if (l_ConditionSourceGroups_ == null || l_ConditionSourceGroups_.isEmpty()) {
            String[] l_types = null;
            if (aCS_.getM_TypeOrientedPolicy() != null &&
                getEorResource().getResourceType() != null) {
              // no rights if the resource enforces to have a conditional ACL
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Resource.NoNonConditionalACL");
                if (l_types != null) {
                    for (int i = 0; i < l_types.length; i++) {
                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        getEorResource().getResourceType()))) {
                            error = true;
                            break;
                        }
                    } // for i
                } // if (l_remove != null)

            } // if (aCS_.getM_TypeOrientedPolicy() != null)
            l_types = null; //

            if (error) {
                // reverse operation
                l_right_ = oldRight;

                setConditionSourceGroupsFromRights();

                throw new CreateError(" No operation. The ACS policy on "+
                        "'Resource.NoNonConditionalACL' "+
                        "enforces conditional AclEntries for the resource type: "+
                        getEorResource().getResourceType());
            } // if (error)
        } // if (l_ConditionSourceGroups_ == null) {

                        // for testing and tracking
                     //System.out.println("In AclEntryImpl.setL_Rights(), values at the end:");
                     //if (l_right_.length != 0) {
                       // for (int i = 0; i < l_right_.length ; i++) {
                         // System.out.println("---- "+i+": "+l_right_[i]);
                        //} // for
                     //}
                     //else
                       // System.out.println("---- null array");

        oldRight = null;
        l_newRights = null;
        l_restrictedRights = null;
        l_limitedRights = null;
        right = null;
    } // setL_Rights()
  
  /**
  * Main method to update the right-defined condition group at every rights updating.
  * Noop if the ACS does not manage the right-defined conditional ACL.
  * Only one external condition group is allowed for an external AclEntry.
  * Controls the ACS policy for 'GroupID.ConditionalAclGroup', 'Resource.OneConditionGroupInACL',
  * 'Resource.TypesOfSourceForConditionalACL.'.
  * Called by setL_Rights() and ACSImpl.checkClosedAclEntriesTo().
  * Calls addConditionSourceGroup(), but not removeConditionSourceGroup().
  * Fires the event 'ConditionSourceGroups' for the removing of all the current groups,
  * then one 'ConditionSourceGroups' event for each new group.
  * @return if negative, it is the number of new condition groups which 
  *     are unknown by the acs, else return 0.
  * @exception CreateError if the ACS does not manage right-condition ACL while
  * the first right contains 'IF', or the ACS of the condition group is not the AclEntry ACS.
  */
  protected int setConditionSourceGroupsFromRights() throws CreateError {
    // variables
    int                 result = 0;
    String              text = null;
    String              groupNameLastComponent = null;
    ImmutableGroupID    grp = null;
    String[]            l_types = null;
    boolean             error = false;
    HashSet<ImmutableGroupID> l_old = null;
            
                            // for testing and tracking purpose
                            //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                              //      " starting with aCS_.getManageRightDefinedConditionalAcls(): "+
                                //    aCS_.getManageRightDefinedConditionalAcls());

    // updates the condition groups, if any
    if (aCS_.getManageRightDefinedConditionalAcls() && isRightDefinedConditionalAclEntry()) {

        // with only only one condition group
        if (l_right_[0].getRight().indexOf("<IF>") == l_right_[0].getRight().lastIndexOf("<IF>")) {
            
                            // for testing and tracking purpose
                            //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                              //      "for one group ");

            if (!l_right_[0].isTargetClassTyped()) {
                groupNameLastComponent = l_right_[0].getRight().substring(
                                l_right_[0].getRight().indexOf("<IF>") + 4);
            } else {
                groupNameLastComponent = l_right_[0].getRight().substring(
                                l_right_[0].getRight().indexOf("<IF>") + 4,
                                l_right_[0].getRight().lastIndexOf("_"+
                                    l_right_[0].getTargetClassType()));
            }

            try {
                grp = (ImmutableGroupID) BaseUtilityImpl.getEorBaseObject(
                            ((Name)acsName_.clone()).extended("<G>").extended(
                            groupNameLastComponent ));
            } catch (BaseError b) {

                            // for testing and tracking purpose
                            //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                              //      " does not find "+groupNameLastComponent);

                result--;
            }

                            // for testing and tracking purpose
                            //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                              //      " finds the group type: "+
                                //    UtilityImpl.getFirstInstance().
                                  //  getObjectTypeWithoutImmutable(
                                    //    grp.getType()));

            if (!error && aCS_.getM_TypeOrientedPolicy() != null &&
                getEorResource().getResourceType() != null) {
              // no adding if the group cannot be a condition group
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("GroupID.ConditionalAclGroup");
                
                if (l_types != null) {
                    error = true;
                    for (int i = 0; i < l_types.length; i++) {

                                    // for testing and tracking purpose
                                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                                      //      " for 'GroupID.ConditionalAclGroup', test of "+l_types[i]);

                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        grp.getType()))) {
                            error = false;

                                    // for testing and tracking purpose
                                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                                      //      " no error on 'GroupID.ConditionalAclGroup'.");

                            break;
                        }
                    } // for i
                } // if (l_types != null)

            } // if (aCS_.getM_TypeOrientedPolicy() != null)
            l_types = null;

            if (!error && aCS_.getM_TypeOrientedPolicy() != null &&
                getEorResource().getResourceType() != null) {
              // no adding if this conditional ACL cannot have more than one condition group
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Resource.OneConditionGroupInACL");
                if (l_types != null && l_ConditionSourceGroups_ != null &&
                        !l_ConditionSourceGroups_.isEmpty()) {
                    for (int i = 0; i < l_types.length; i++) {
                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        getEorResource().getResourceType()))) {
                            error = true;

                                    // for testing and tracking purpose
                                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                                      //      " error on 'Resource.OneConditionGroupInACL.'.");

                            break;
                        }
                    } // for i
                } // if (l_remove != null)

            } // if (aCS_.getM_TypeOrientedPolicy() != null)
            l_types = null;

            if (!error && aCS_.getM_TypeOrientedPolicy() != null &&
                getEorResource().getResourceType() != null) {
              // no adding if this conditional ACL cannot have this source
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Resource.TypesOfSourceForConditionalACL."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                        getEorResource().getResourceType()));

                                    // for testing and tracking purpose
                                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                                      //      " for 'Resource.TypesOfSourceForConditionalACL.', "+
                                        //    "compare to "+
                                          //  UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                            //    getEorEP().getEPType()));

                if (l_types != null) {
                    error = true;
                    for (int i = 0; i < l_types.length; i++) {

                                    // for testing and tracking purpose
                                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                                      //      " for 'Resource.TypesOfSourceForConditionalACL.', test of "+l_types[i]);

                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        getEorEP().getEPType()))) {
                            error = false;

                                    // for testing and tracking purpose
                                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                                      //      " no error on 'Resource.TypesOfSourceForConditionalACL.'.");

                            break;
                        }
                    } // for i
                } // if (l_types != null)

            } // if (aCS_.getM_TypeOrientedPolicy() != null)
            l_types = null;

            if (!error && grp != null) {
                if (grp.getEorACS() != ep_.getEorACS() &&
                        !grp.getAcsName().equals(nameOfResAcs_) )
                    throw new CreateError(" A condition group has to be from the source ACS "+
                            "or from the target ACS.");
            } // if

            // main operation
            if (!error) {

              if (l_ConditionSourceGroups_ != null) {
                  l_old = (HashSet<ImmutableGroupID>)l_ConditionSourceGroups_.clone();
                  l_ConditionSourceGroups_.clear();

                  try {
                      firePropertyChange("ConditionSourceGroups", l_old, null);
                  } catch (Exception e) {
                    ARoad0.Gui1.DialogBox.showAlarmMessage(
                      "There is an exception after the base updating, from outside the database."+
                      "You should save the base and stop the program, then run it once again.");
                    System.out.println("From AclEntryImpl.setConditionSourceGroupsFromRights(), "+
                            "firePropertChange() throws:");
                    e.printStackTrace();
                  }
              } // if

                            // for testing and tracking purpose
                            //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                              //      " single grp: "+grp);

              if (grp != null) {
                  addConditionSourceGroup(grp);
              } // if
            } // if
            else {

                            // for testing and tracking purpose
                            //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights()"+
                              //      " error with grp: "+grp);
            }
        } // if
        else { // several <IF>, then several groups to add

                    // for testing and tracking purpose
                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights(): "+
                      //      "several <IF>, then several groups to add with the right "+
                        //    l_right_[0].getRight());

            if (isExternalAclEntry()) {
                throw new CreateError(" Only one condition group is allowed "+
                        "for an external AclEntry.");
            } // if

            // process the last group
            if (!l_right_[0].isTargetClassTyped()) {
                groupNameLastComponent = l_right_[0].getRight().substring(
                                l_right_[0].getRight().lastIndexOf("<IF>") + 4);
            } else {
                groupNameLastComponent = l_right_[0].getRight().substring(
                                l_right_[0].getRight().lastIndexOf("<IF>") + 4,
                                l_right_[0].getRight().lastIndexOf("_"+
                                    l_right_[0].getTargetClassType()));
            }

                    // for testing and tracking purpose
                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights(): "+
                      //      "several <IF>, then groupNameLastComponent: "+
                        //    groupNameLastComponent);

            try {
                grp = (ImmutableGroupID) BaseUtilityImpl.getEorBaseObject(
                            ((Name)acsName_.clone()).extended("<G>").extended(
                            groupNameLastComponent ));
            } catch (BaseError b) {

                    // for testing and tracking purpose
                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights(): "+
                      //      "several <IF>, last group with BaseError: "+b.toString());

                result--;
            }

            if (aCS_.getM_TypeOrientedPolicy() != null &&
                getEorResource().getResourceType() != null) {
              // no adding of the group cannot be a condition group
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("GroupID.ConditionalAclGroup");
                
                if (l_types != null) {
                    error = true;
                    for (int i = 0; i < l_types.length; i++) {
                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        grp.getType()))) {
                            error = false;
                            break;
                        }
                    } // for i
                } // if (l_types != null)

                if (error) {

                                // for testing and tracking purpose
                                //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights(): "+
                                  //      "several <IF>, 'GroupID.ConditionalAclGroup' matches for: "+
                                    //    grp.getNickName());

                    result--;
                }
            } // if (aCS_.getM_TypeOrientedPolicy() != null)
            l_types = null;

            if (!error && aCS_.getM_TypeOrientedPolicy() != null &&
                getEorResource().getResourceType() != null) {
              // no adding if this conditional ACL cannot have more than one condition group
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Resource.OneConditionGroupInACL");
                if (l_types != null && l_ConditionSourceGroups_ != null &&
                        !l_ConditionSourceGroups_.isEmpty()) {
                    for (int i = 0; i < l_types.length; i++) {
                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        getEorResource().getResourceType()))) {
                            error = true;
                            break;
                        }
                    } // for i
                } // if (l_remove != null)

            } // if (aCS_.getM_TypeOrientedPolicy() != null)
            l_types = null; 

            if (!error && aCS_.getM_TypeOrientedPolicy() != null &&
                getEorResource().getResourceType() != null) {
              // no adding if this conditional ACL cannot have this source
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Resource.TypesOfSourceForConditionalACL."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                        getEorResource().getResourceType()));
                if (l_types != null) {
                    error = true;
                    for (int i = 0; i < l_types.length; i++) {
                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(
                                        getEorEP().getEPType()))) {
                            error = false;
                            break;
                        }
                    } // for i
                } // if (l_types != null)

            } // if (aCS_.getM_TypeOrientedPolicy() != null)
            l_types = null;
            
            if (!error) {

                    // for testing and tracking purpose
                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights(): "+
                      //      "several <IF>, add the first condition group: "+
                        //    grp.getNickName());

              if (l_ConditionSourceGroups_ != null) {
                  l_old = (HashSet<ImmutableGroupID>)l_ConditionSourceGroups_.clone();
                  l_ConditionSourceGroups_.clear();

                  try {
                      firePropertyChange("ConditionSourceGroups", l_old, null);
                  } catch (Exception e) {
                    ARoad0.Gui1.DialogBox.showAlarmMessage(
                      "There is an exception after the base updating, from outside the database."+
                      "You should save the base and stop the program, then run it once again.");
                    System.out.println("From AclEntryImpl.setConditionSourceGroupsFromRights(), "+
                            "firePropertChange() throws:");
                    e.printStackTrace();
                  }
              } // if

              if (grp != null) {
                if (grp.getEorACS() != ep_.getEorACS() &&
                        !grp.getAcsName().equals(nameOfResAcs_) )
                    throw new CreateError(" A condition group has to be from the source ACS "+
                            "or from the target ACS.");
                addConditionSourceGroup(grp);
              } // if

            } // if (!error)
            
            // process the other groups in the inverse order of l_right_[0].getRight()
            text = l_right_[0].getRight().substring(0,
                                l_right_[0].getRight().lastIndexOf("<IF>"));

            while (text.indexOf("<IF>") != -1 && text.length() > text.lastIndexOf("<IF>") + 4) {
                groupNameLastComponent = text.substring(
                                text.lastIndexOf("<IF>") + 4);

                try {
                    grp = (ImmutableGroupID) BaseUtilityImpl.getEorBaseObject(
                            ((Name)acsName_.clone()).extended("<G>").extended(
                            groupNameLastComponent ));
                } catch (BaseError b) {
                    result--;
                }

                if (aCS_.getM_TypeOrientedPolicy() != null &&
                    getEorResource().getResourceType() != null) {
                  // no adding of the group cannot be a condition group
                    l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                              get("GroupID.ConditionalAclGroup");
                    error = true;
                    if (l_types != null) {
                        for (int i = 0; i < l_types.length; i++) {
                            if (l_types[i].equals(
                                    UtilityImpl.getFirstInstance().
                                        getObjectTypeWithoutImmutable(
                                            grp.getType()))) {
                                error = false;
                                break;
                            }
                        } // for i
                    } // if (l_types != null)

                    if (error) {

                                    // for testing and tracking purpose
                                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights(): "+
                                      //      "several <IF>, 'GroupID.ConditionalAclGroup' matches for: "+
                                        //    grp.getNickName());

                        result--;
                    }
                } // if (aCS_.getM_TypeOrientedPolicy() != null)
                l_types = null;

                if (!error) {

                            // for testing and tracking purpose
                            //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights(): "+
                              //      "several <IF>, add another condition group: "+
                                //    grp.getNickName());

                    if (grp != null) {
                        if (grp.getEorACS() != ep_.getEorACS() &&
                                !grp.getAcsName().equals(nameOfResAcs_) )
                            throw new CreateError(" A condition group has to be from the source ACS "+
                                    "or from the target ACS.");

                        addConditionSourceGroup(grp);
                    } // if
                } // if

                text = text.substring(0, text.lastIndexOf("<IF>"));
            } // while

        } // else

    } // if (l_right_.length != 0 && l_right_[0].getRight().indexOf("<IF>") != -1)          
    else if (!aCS_.getManageRightDefinedConditionalAcls() && isRightDefinedConditionalAclEntry()) {
        throw new CreateError(" The ACS of this AclEntry does not manage the right-condition ACLs");
    }
    else if (aCS_.getManageRightDefinedConditionalAcls()) {

        if (l_right_ != null &&
                (l_right_.length == 0 || l_right_[0].getRight().indexOf("<IF>") == -1)) {

                        // for testing and tracking purpose
                        //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights() "+
                          //      "without condition groups");

              if (l_ConditionSourceGroups_ != null) {
                  l_old = (HashSet<ImmutableGroupID>)l_ConditionSourceGroups_.clone();
                  l_ConditionSourceGroups_.clear();

                  try {
                      firePropertyChange("ConditionSourceGroups", l_old, null);
                  } catch (Exception e) {
                    ARoad0.Gui1.DialogBox.showAlarmMessage(
                      "There is an exception after the base updating, from outside the database."+
                      "You should save the base and stop the program, then run it once again.");
                    System.out.println("From AclEntryImpl.setConditionSourceGroupsFromRights(), "+
                            "firePropertChange() throws:");
                    e.printStackTrace();
                  } // catch
              } // if

        } // if
          
    }
    else {

                    // for testing and tracking purpose
                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights() NOOP");
    }

                    // for testing and tracking purpose
                    //System.out.println("AclEntryImpl.setConditionSourceGroupsFromRights() result: "+
                      //      result);
    
    text = null;
    groupNameLastComponent = null;
    grp = null;
    l_types = null;
    l_old = null;
    return result;
  } // setConditionSourceGroupsFromRights()  

  /**
   * Adds a condition group without controls. The ACS of the group may not be
   * the ACL ACS.
   * To activate the ACL rights, the EP has to be member of the condition GroupIDs.
   * If the EP is not a GroupIDMember while there is at least one condition group,
   * the rights are empty. The argument is not updated.
   * GroupIDImpl.finalizeForBase() delete a conditional AclEntry having a finalized group
   * as condition group.
   * Called by the constructor, setConditionSourceGroupsFromRights() and
   * addForUserConditionSourceGroup().
   * Fires a PropertyChangeEvent "ConditionSourceGroups" to the listeners.
   * @param _grp conditionnal group to add. Never null.
   */
  protected void addConditionSourceGroup(ImmutableGroupID _grp) {
      ArrayList<ImmutableGroupID> l_old = null;
      if (l_ConditionSourceGroups_ == null)
          l_ConditionSourceGroups_ = new ArrayList<ImmutableGroupID>(INITIAL_CAPACITY);
      else
          l_old = (ArrayList<ImmutableGroupID>)l_ConditionSourceGroups_.clone();

      l_ConditionSourceGroups_.add(_grp);

      try {
          firePropertyChange("ConditionSourceGroups", l_old,
              l_ConditionSourceGroups_.clone());
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From AclEntryImpl.addConditionSourceGroup(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
      }
  } // addConditionSourceGroup()

  /**
   * Removes a condition group. The result is never null nor empty, since the first
   * group cannot be removed.
   * To activate the ACL rights, the EP has to be member of the condition GroupIDs,
   * if it is a GroupIDMember. The argument is not updated.
   * The first condition group cannot be removed.
   * Called by removeForUserConditionSourceGroup(), but not by
   * setConditionSourceGroupsFromRights().
   * If the target is a Directory in an ACS that manages the ACL inheritance,
   * calls the method resetInheritedAclEntryRights() on the Directory.
   * Fires a PropertyChangeEvent "ConditionSourceGroups" to the listeners.
   * @param _grp conditionnal group to remove
   * @exception CreateError if _grp is not registered or it is the first condition group
   */
  protected void removeConditionSourceGroup(ImmutableGroupID _grp) throws CreateError {
      ArrayList<ImmutableGroupID> l_old = null;
      if (l_ConditionSourceGroups_ == null )
          throw new CreateError("This AclEntry is not a conditional AclEntry.");
      else
          l_old = (ArrayList<ImmutableGroupID>)l_ConditionSourceGroups_.clone();
      if (!l_ConditionSourceGroups_.contains(_grp))
          throw new CreateError("This ACL does not know this condition group");
      if (l_ConditionSourceGroups_.get(0) == _grp)
          throw new CreateError("The first condition group cannot be removed");

      l_ConditionSourceGroups_.remove(_grp);

      //if (l_ConditionSourceGroups_.isEmpty())
        //  l_ConditionSourceGroups_ = null;

      if (aCS_.getManageACLRightsInheritance()) {
          Resource res = (Resource) getEorResource();
          if (res instanceof Directory) {
              res.resetInheritedAclEntryRights(this);
          }
          res = null;
      } // if

      try {
          //if (l_ConditionSourceGroups_ != null)
              firePropertyChange("ConditionSourceGroups", l_old,
                l_ConditionSourceGroups_.clone());
          //else
            //  firePropertyChange("ConditionSourceGroups", l_old, null);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From AclEntryImpl.removeConditionSourceGroup(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
      }
  } // removeConditionSourceGroup()

  /**
   * If the target is a Directory in an ACS that manages the ACL inheritance,
   * calls Directory.resetInheritedAclEntryRightsOfChildren(), to reset
   * the effective rights in the inherited AclEntry map of each Directory child.
   * An AcsAddon may use also this call to resetInheritedAclEntryRightsOfChildren()
   * for other AclEntry-related operations, like in MySQL.
   * Called by the methods setL_RightsForUser() and add/removeForUserConditionSourceGroup()
   * after the updating of the effective rights, and by GroupIDImpl.add/removeEorMember()
   * after the updating of the (in)direct members of a condition group.
   * Calls Directory.resetInheritedAclEntryRightsOfChildren().
   * Note: the creation of the inherited AclEntry in a child is not handled by
   * the generic gBase package, contrary of the rights updating of the inherited AclEntry.
   * Fires no PropertyChangeEvent.
   */
  protected void resetRightsAsInheritedAclEntry() {
      if (aCS_.getManageACLRightsInheritance()) {
          Resource res = (Resource) getEorResource();
          if (res instanceof Directory) {
              ((Directory)res).resetInheritedAclEntryRightsOfChildren(this);
          }
          res = null;
      } // if
  } // resetRightsAsInheritedAclEntry()
  
  /**
   * Called by ACSImpl.closeAclEntriesTo() only for the external ACLs,
   * when the links to other ACS have to be cut and cloned.
   * The ACS, the EP and the resource ACS are null in the returned AclEntry.
   * The rights array is copied without cloning of the rights (shallow copy).
   * Thus the rights networks of metarights and acsrights are linked to the ACL clone.
   * The method getEorResource() returns null while getResName() returns a clone
   * of the resource name. An external ACL has conditional group(s) from the source ACS
   * or the target ACS.
   * The methods getRight() and getDetailledName() work well, as toString().
   * The properties are protected, so may be accessed by the package.
   * The returned instance may be finalized with finalizeClosedAcl().
   * Caution: the clone is not registered in its ACS, its resource, its EP or
   * its confition groups, but equals() returns true when it compares a clone and its source.
   * @return Object without external references
   * (ACS, resource ACS and EligibleParty) and copy of the other properties,
   * excluding the condition groups.
   */
   protected AclEntry copyForExternal() {
      if (!isExternalAclEntry())
            throw new InternalError("AclEntryImpl.copyForExternal() called "+
                    "on a non-external ACL");

      try {
          AclEntryImpl clon = (AclEntryImpl) super.clone();
         
          clon.resName_ = (ImmutableName) resName_.clone();
          if (epName_ != null)
              clon.epName_ = (ImmutableName) epName_.clone();
          clon.aCS_ = null; // to cut external links
          clon.resACS_ = null; // to cut external links
          clon.ep_ = null; // to cut external links
          clon.positiveRight_ = positiveRight_;
          clon.detailledName_ = detailledName_;

          // keep the external links to acsrights and metarights
          clon.l_right_ = new StringRight[l_right_.length];
          System.arraycopy(l_right_, 0, clon.l_right_, 0, l_right_.length);          
          
          if (acsName_ != null)
              clon.acsName_ = (ImmutableName) acsName_.clone();
          if (nameOfResAcs_ != null)
              clon.nameOfResAcs_ = (ImmutableName) nameOfResAcs_.clone();
          if (nameOfEpAcs_ != null)
              clon.nameOfEpAcs_ = (ImmutableName) nameOfEpAcs_.clone();

          if (l_ConditionSourceGroups_ != null) {
              // the single group is from the source ACS or the target ACS
              clon.l_ConditionSourceGroups_ = (ArrayList)l_ConditionSourceGroups_.clone();
          } // if (l_ConditionSourceGroups_

                                // for testing and tracking
                            //System.out.println("AclEntryImpl.clone() for"+this+
                              //      " and the result is "+clon);


          return clon;

      } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
  	    if (aCS_ != null) {
  	            aCS_.getEorBaseError().addException(e.toString());
  	    }
        throw new InternalError(e.toString());
      } // catch
   } // copyForExternal()

} // AclEntryImpl