//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.BaseError;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Collections;
import java.util.Iterator;

import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;

/**
 * This class is responsible for modeling an acting entity which may have access
 * to controlled resources. It may also play some special roles,
 * as a bridge, between an another actor and some resources.
 * Exemples of such actors are servers, printers, LDAP objects or executables.
 * This class also models the operating system process which has a current
 * account and group, and which belongs to several groups. By default,
 * the current account and group are the account and group of the actor as resource.
 * The current account and group, and the secondary groups, cannot be updated
 * if the resource type does not allow it. But the resource type does not control
 * the bridge target. The actor alias must be another-acs actor.
 * It is a javabean with the following bound properties:
 * 'SecondaryGroups', 'l_bridgeSources',
 * 'CurrentGroupID', 'CurrentUserID', 'ACSBridgeTarget', 'EPType', 'IsRightProxy',
 * 'AclConnectedResources', 'OneSourceTypedPrivilege',
 * 'OneSourceLinkedPrivilege'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * This class has subclasses in this package.
 * All variables are private to secure the white box inheritances from this class,
 * excepting for epType_, for which the values are associated
 * to the class and have to be set by the subclass constructors. <P>
 * This class has several subclasses in this package and in the AcsAddon packages.
 * Most of the variables are protected.
 * @version 0.7.1
 */
public class ActorImpl extends ResourceImpl implements Actor, Serializable {
  /* private variables */
  private static final long     serialVersionUID__ = 0;

  /* protected variables */
  protected transient ImmutableActor[]    l_bridgeSources_; // External Object Reference
  protected ImmutableGroupID              currentGroupID_;// External Object Reference
  protected transient ImmutableActor      bridgeTarget_;  // External Object Reference
  /**
   * with this initial value, currentUserID_ is returned as null in the get method.
   * currentUserID_ is set to null only in finalizeForBase(), which detects a second
   * call through the test of this variable.
   */
  protected ImmutableUserID     currentUserID_ = UserIDImpl.EMPTY_INSTANCE; // External Object Reference
  protected ImmutableGroupID[]  l_SecondaryGroup_; // External Object Reference
  /**
   * Map<ImmutableName, Resource>
   */
  protected SortedMap<ImmutableName, ImmutableResource> m_AclEntriesResources_;
  protected boolean             isRightProxy_ = true;
   
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ActorImpl EMPTY_INSTANCE = new ActorImpl();

  /** variable for the JDK 2 serialization
   * @serial ((
   * ("currentGroupID_",GroupIDImpl.class), ("currentUserID_",UserIDImpl.class),
   * ("l_SecondaryGroup_",GroupID[].class),
   * ("epType_",String.class), ("m_AclEntriesResources_",TreeMap.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      //new ObjectStreamField("l_bridgeSources_",Actor[].class),
      new ObjectStreamField("currentGroupID_",GroupIDImpl.class),
      new ObjectStreamField("currentUserID_",UserIDImpl.class),
      new ObjectStreamField("l_SecondaryGroup_",GroupID[].class),
      new ObjectStreamField("m_AclEntriesResources_",TreeMap.class),
      //new ObjectStreamField("bridgeTarget_",ActorImpl.class),
      new ObjectStreamField("isRightProxy_",Boolean.TYPE)
  };

 /**
  * Default constructor.
  */
  public ActorImpl(){
    actorActivated_ = true;
  }
  
  /**
  * Registers the new instance in the eligible parties map in the main ACS.
  * Used for the creation of an instance through the GUI:
  * used by the beamer editors for the ACS resources list and for the
  * ACS EP list, from 'New'.
  * The actor name _and_ the resource name are structured.
  * _bridgeRef must belong to another ACS than this instance ACS.
  * Note: this is added to the main ACS as a resource, in super().
  * @param _name of the actor as a resource.
  * @param _acs main ACS of the actor.
  * @param _parent of the actor.
  * @param _owner owns the resource; may be null.
  * @param _resourceType is also the EP type, and may be null if the ACS allows it.
  * @param _bridgeTarget another-ACS Actor for which this instance is a bridge source;
  *   may be null. 
  * @exception CreateError: this(_name, _acs, _parent, _actor_name, _bridgeRef, false)
  */
  public ActorImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType, ImmutableActor _bridgeTarget)
    throws CreateError {
      
      this(_name, _acs, _parent, _owner, _resourceType, _bridgeTarget, true);
    
  } // constructor
  

  /**
  * Used for the creation of an actor.
  * May register the new instance in the EligibleParties map in the main ACS.
  * Sets the owner as the current UserID.
  * The ACS type-oriented policy is used for the key 'Actor.NoNullCurrentAccount'
  * to verify that this type of actor can have no owner.
  * Sets the current group if the parent UserID has a primary group.
  * Note: this is added to the main ACS as a resource, in the super class constructor.
  * The actor name _and_ the resource name are structured.
  * _bridgeTarget must belong to another ACS than this instance ACS.
  * Used by copy().
  * @param _name of the actor as a resource.
  * @param _acs main ACS of the actor.
  * @param _parent of the actor.
  * @param _owner owns the resource and is the current UserID; may be null.
  * @param _resourceType is also the EP type, and may be null if the ACS allows it. 
  * @param _bridgeTarget another-ACS Actor for which this instance is a bridge source;
  *  may be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in its ACS.
  * @exception CreateError: see super(_name, _acs, _parent) exception,
  *   if _name is already known in ACSCatalog as an EligibleParty, or if the
  *   bridge target belongs to another ACS, if an owner is mandatory.
  */
  public ActorImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType, ImmutableActor _bridgeTarget, boolean _toRegisterInManager)
    throws CreateError {
      
      super(_name, _acs, _parent, _owner, _resourceType, false);
      
      if (_bridgeTarget != null ) {

          String[] l_types = aCS_.getM_TypeOrientedPolicy().get("Actor.IsNotBridgeSource");

          try {
              if (l_types != null && Arrays.asList(l_types).contains(
                      UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(_resourceType)) ) {
                          // works because _toRegisterInManager is true from the GUI creation
                         finalizeForBase();
                         throw new CreateError("From the ACS policy 'Actor.IsNotBridgeSource',"+
                                 "the type of this actor forbids"+
                                 " it to be a bridge source.");
              }

              if (_bridgeTarget.getEorACS() != _acs) {
                 try {
                     setEorACSBridgeTarget((Actor)_bridgeTarget, this);
                 } catch (CreateError c) {
                     // works because _toRegisterInManager == true from the GUI creation
                    finalizeForBase();                 
                    System.out.println("ActorImpl constructor: the bridge target "+
                   "registration throws:");
                    c.printStackTrace();
                    
                    throw new CreateError(" The bridge relation"+
                       " is not correct. "+c.getMessage());
                 }
             } else {  
                  // works because _toRegisterInManager is true from the GUI creation
                 finalizeForBase();
                 throw new CreateError(" The bridge target "+
                   "must belong to another ACS.");
             }
         } catch (UpDateError u) {
                    // message 'the EligibleParty  xxx is unknown in the ACS '
                    //r.printStackTrace();
         }  // catch            
      } // if (_bridgeTarget != null
      
      actorActivated_ = true; // used by ACS.getEorM_Actors()
      
      try {
          setEorCurrentUserID(_owner);
      } catch (UpDateError u) {

            try {
                finalizeForBase();
            } catch (UpDateError r) {
                    // message 'the EligibleParty  xxx is unknown in the ACS '
                    //r.printStackTrace();
            }

            CreateError c = new CreateError(" From the Actor constructor, "+
                    "for the current userID: "+u.toString());
            c.printStackTrace();
            throw c;
      } // catch

        // checkup 'Actor.NoNullCurrentAccount'
      if (getEorCurrentUserID() == null) {
        String reducedType = null;
        String[] l_policy = null;

                        // for testing and tracking
                        //System.out.println("ActorImpl constructor with a null current UserID for "+
                          //      name_);


        reducedType = UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(type_);

        if (aCS_.getM_TypeOrientedPolicy() != null) {
            l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get("Actor.NoNullCurrentAccount");

            if (l_policy != null && Arrays.asList(l_policy).contains(reducedType)) {
                reducedType = null;
                l_policy = null;

                        // for testing and tracking
                        //System.out.println("ActorImpl constructor with exception on "+
                          //      "'Actor.NoNullCurrentAccount' for the type "+type_+" on "+
                           //     name_);

                try {
                    finalizeForBase();
                } catch (UpDateError r) {
                    // message 'the EligibleParty  xxx is unknown in the ACS '
                    //r.printStackTrace();
                }

                throw new CreateError(
                    " The ACS type-oriented policy for 'Actor.NoNullCurrentAccount'"+
                    " forbids this type of actor to have no account.");
            } // if (l_policy != null &&

            reducedType = null;
            l_policy = null;
        } // if
      } // if (currentUserID_== null)
       
      // adds this to the map of resources in the main ACS
      // and sends an event "OneResource" to the GUI listeners.
      // may throw a CreateError
      try {
          if (_toRegisterInManager) ((ACSImpl)aCS_).newEorResource(this);
      } catch (CreateError c) {
              try {
                  finalizeForBase();
              } catch (UpDateError r) {
                r.printStackTrace();
              }
              throw c;
      }

      // add this to the map of EligibleParties in the main ACS
      // may throw a CreateError
      if (_toRegisterInManager) {
          try {
            ((ACSImpl)getEorACS()).newEorEligibleParty(this);
          } catch (CreateError c) {
                  try {
                      finalizeForBase();
                  } catch (UpDateError r) {
                    throw new CreateError(" "+r.toString() +" is thrown in "+c.toString());
                  }
                  throw c;
          } // catch
      } // if
  } // constructor

  //
  // -------------  methods of the implemented interfaces -----------------
  //
  
  /**
  * This method provides a copy of the instance
  * with another name, following Copyable.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Each aclEntry is copied, but not the list of ViewInBase
  * that contains this instance. The bridge target is copied if
  * this instance is a bridge source.
  * Caution: do not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Calls copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually created as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
     May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ActorImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null || 
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError(
          "The name or the acs is null,"+
          " or the parent ACS is not the instance ACS.");
      }
      
      ActorImpl result = new ActorImpl(
            _name, _acs, _parent, null, getResourceType(), null,
            _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);
      
      return result;
  } // copy(,,,)

  
  /**
  * This method provides a copy of the instance
  * as copy() but without copying internal aclEntries.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Each aclEntry is copied, but not the list of ViewInBase
  * that contains this instance. The bridge target is copied if
  * this instance is a bridge source.
  * Caution: does not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Caution: does not copy the bridge sources, if they exists.
  * Calls copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually created as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
     May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ActorImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null || 
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError("in ActorImpl.copy(),"+
          " _name or _acs is null,"+
          " or _parent ACS is not _acs.");
      }
      
      ActorImpl result;
      result = new ActorImpl(
            _name, _acs, _parent, null, getResourceType(), bridgeTarget_,
            _toRegisterInManager.booleanValue());
      
      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclEntry copy
      
      //if (result.getEorL_AclConnectedResources().isEmpty()) return result;
      //else return EligiblePartyImpl.deleteAclEntries(result);
      return result;
  } // copyWithoutAcl(,,,)
  
  /**
  * The current userID is not null when, typically, the actor is an executable
  * launched under an userID. Called by the GUI. Calls setEorCurrentUserID().
  * @param _u current userID of the actor. May be null.
  * @exception UpDateError if the resource type starts with 'byAcsCreator' or 'immutable',
  * or from setEorCurrentUserID().
  */
  public void setEorCurrentUserIDForUser(ImmutableUserID _u) throws UpDateError {
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable>") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            " The resource type of "+getName().lastComponent()+
            " sets it as immutable for its current userID.");
    
    setEorCurrentUserID(_u);
  } // setEorCurrentUserIDForUser()
 
  /**
  * The current userID is not null when, typically, the actor is an executable
  * launched under an userID. This class does not override setEorUserID()
  * since Resource.userID is an access target property, and 
  * Actor.currentUserID is a rights user property.
  * Noop if the resource type starts with 'byAcsCreator' or 'immutable',
  * typically when it is launched by the system at its starting.
  * The ACS type-oriented policy is used for the key 'Actor.NoCurrentAccount'
  * to verify that this type of actor can have a current account.
  * Called by the constructor.
  * If it is applicable, calls setEorCurrentGroupID() to set the _u primary group
  * as the current group, and setEorL_SecondaryGroup() to add as secondary groups
  * all the groups having the argument as member.
  * Calls UserID.add/removeCurrentActor() and getAuthorizeCurrentGroupID() that uses
  * the Resource type, setEorCurrentGroupID(), setEorL_SecondaryGroup().
  * Fires an 'CurrentUserID' event.
  * @param _u current userID of the actor. May be null.
  * @exception UpDateError if the ACS type policy
  * forbids a current account for this type of actor.
  */
  protected synchronized void setEorCurrentUserID(ImmutableUserID _u) throws UpDateError {
    try {
      if (currentUserID_ != null && currentUserID_ != UserIDImpl.EMPTY_INSTANCE)
        ((UserIDImpl)currentUserID_).removeCurrentActor(this);
    } catch (UpDateError u) {
      getEorACS().getEorBaseError().addException(
        "in ActorImpl.setEorCurrentUserID(), the actor is unknown in the userID");
    }

    // checkup 'Actor.NoCurrentAccount'
    if (type_ != null) {
        String reducedType = null;
        String[] l_policy = null;

        reducedType = UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(type_);

        if (aCS_.getM_TypeOrientedPolicy() != null) {
            l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get("Actor.NoCurrentAccount");

            if (l_policy != null && Arrays.asList(l_policy).contains(reducedType)) {
                reducedType = null;
                l_policy = null;
                
                throw new UpDateError(
                    " The ACS type-oriented policy for 'Actor.NoCurrentAccount'"+
                    " forbids this type of actor to have an account.");
            } // if (l_policy != null &&

            reducedType = null;
            l_policy = null;
        } // if
    } // if (type_ != null)

    ImmutableUserID oldCurrentUserID = currentUserID_;

    if (_u == null || _u == UserIDImpl.EMPTY_INSTANCE)
      currentUserID_ = UserIDImpl.EMPTY_INSTANCE;
    else {
      currentUserID_ = _u;
      ((UserIDImpl)currentUserID_).addCurrentActor(this);

                        // for testing and tracking
                        //System.out.println("ActorImpl.setEorCurrentUserID() for "+
                          //      name_.lastComponent()+", before setEorCurrentGroupID()"+
                            //    " with the new current user "+
                              //  currentUserID_);
      
      // set the current groupID, if applicable
      if (getAuthorizeCurrentGroupID())
          setEorCurrentGroupID(_u.getEorPrimGroup());
      // set the secondary groups from _u
      try {
          setEorL_SecondaryGroup(_u.getEorL_GroupID());
      } catch (UpDateError u) {
          throw new InternalError("ActorImpl.setEorCurrentUserID() "+u.toString());
      }
    } // else   
    
    try {
        firePropertyChange("CurrentUserID", oldCurrentUserID, currentUserID_);
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.setEorCurrentUserID(), firePropertChange() throws:");
            e.printStackTrace();
    } 

                        // for testing and tracking
                        //System.out.println("end of ActorImpl.setEorCurrentUserID() for "+
                          //      name_.lastComponent()+" with the current user "+currentUserID_);
    
    oldCurrentUserID = null;
  } // setEorCurrentUserID()

 /**
 * If this property has not been set, it is null.
 * When an executable is launched from a userID, its current
 * userID often becames this userID. 
 * @return the current userID. May be null.
  */
  public ImmutableUserID getEorCurrentUserID() {      
    if (currentUserID_== null ||
      currentUserID_.equals(UserIDImpl.EMPTY_INSTANCE) || 
      currentUserID_.toString() == null) {
          return null;
    }
    else {
        return currentUserID_;
    }
  } // getEorCurrentUserID()

  /**
  * Set the group for which this actor is member as right user.
  * The current groupID is not null when, typically, the actor is a program which is
  * launched under an another groupID. This class does not overread setEorGroupID()
  * since the Resource groupID may be different, for instance when the ACL rights
  * authorizes the launching. Resource.groupID is an access target property, 
  * and Actor.currentGroupID is a rights user property.
  * Calls GroupID.add/removeCurrentActor(this).
  * Reads the ACS 'Actor.NoCurrentGroup' values to check if the argument
  * may be a current group.
  * The ACS type-oriented policy is used for the key 'Actor.NoCurrentGroup'
  * to verify that this type of actor can have a current group.
  * Overridden by the AcsAddon Ubuntu Actor and Executable classes.
  * Called by the GUI. Calls setEorCurrentGroupID().
  * @param _g the current groupID of the actor. May be null.
  * @exception UpDateError if the resource type starts with 'byAcsCreator' or 'immutable',
  * or from setEorCurrentGroupID()
  */
  public synchronized void setEorCurrentGroupIDForUser(ImmutableGroupID _g) throws UpDateError {
     if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable>") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            " The resource type of "+getName().lastComponent()+
            " sets it as immutable for its current group.");
    
      setEorCurrentGroupID(_g);
  } // setEorCurrentGroupIDForUser()

  /**
  * Set the group for which this actor is member as right user.
  * The current groupID is not null when, typically, the actor is a program which is
  * launched under an another groupID. This class does not overread setEorGroupID()
  * since the Resource groupID may be different, for instance when the ACL rights
  * authorizes the launching. Resource.groupID is an access target property, 
  * and Actor.currentGroupID is a rights user property.
  * Calls GroupID.add/removeCurrentActor(this).
  * Reads the ACS 'Actor.NoCurrentGroup' values to check if the argument
  * may be a current group.
  * The ACS type-oriented policy is used for the key 'Actor.NoCurrentGroup'
  * to verify that this type of actor can have a current group.
  * Called by setEorCurrentUserID() and setEorCurrentGroupIDForUser().
  * Fires an 'CurrentGroupID' event.
  * @param _g the current groupID of the actor. May be null.
  * @exception UpDateError if the current groupID is immutable, or the ACS type policy
  * forbids a current account for this type of actor.
  */
  protected synchronized void setEorCurrentGroupID(ImmutableGroupID _g) throws UpDateError {

    if (!getAuthorizeCurrentGroupID())
        throw new UpDateError(
                    " The ACS type-oriented policy for 'Actor.NoCurrentGroup'"+
                    " forbids all the current groups for this type of actor.");

     String[] l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
              get("Actor.NoCurrentGroup");
     boolean remove = false;
     if (l_remove != null) {
          for (int i = 0; i < l_remove.length; i++) {
              if (_g != null && l_remove[i].equals(
                      UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(_g.getEPType()))) {
                  remove = true;
                  break;
              }
          } // for i
     } // if (l_remove !=
     l_remove = null;
     if (remove)
          throw new UpDateError(" The ACS type-oriented policy for 'Actor.NoCurrentGroup' "+
                  "forbids this group to be a current group.");
    
    try {
      if (currentGroupID_ != null )
      ((GroupIDImpl)currentGroupID_).removeCurrentActor(this);
    } catch (UpDateError u) { 
        throw new InternalError(" ActorImpl.setEorCurrentGroupID() - "+
                "the actor is unknown in the GroupID");
    } // catch
    
    ImmutableGroupID oldCurrentGroupID = currentGroupID_;
    currentGroupID_ = _g;
    if (currentGroupID_ != null) ((GroupIDImpl)currentGroupID_).addCurrentActor(this); 
    
    try {
        firePropertyChange("CurrentGroupID", oldCurrentGroupID, currentGroupID_);
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.setEorCurrentGroupID(), firePropertChange() throws:");
            e.printStackTrace();
    } 

                        // for testing and tracking
                        //System.out.println("end of ActorImpl.setEorCurrentGroupID() for "+
                          //      name_.lastComponent()+
                            //    " with the new current group "+
                              //  currentGroupID_);
    
    oldCurrentGroupID = null;
  } // setEorCurrentGroupID()

  /**
  * Get the group for which this actor is member as right user.
  * @return ImmutableGroupID used currently by the actor, that may be changed
  * by itself, and it is different to the actor groupID attached to any resource.
  * For instance, when an executable is launched from a groupID, its current
  * groupID often becames the launcher groupID. May be null.
  */
  public ImmutableGroupID getEorCurrentGroupID() {
    return currentGroupID_;
  }
  
  /**
  * A secondary group is used by an Actor as rights user, to get complementary
  * accesses to targets only when a current userID is set.
  * Noop if the resource type starts with 'byAcsCreator' or 'immutable'.
  * Calls setEorL_SecondaryGroup().
  * @param _L array of the secondary groupID attached to this actor. May be null.
  * @exception UpDateError if the secondary groupIDs are immutable.
  */
  public void setEorL_SecondaryGroupForUser(ImmutableGroupID[] _L) 
          throws UpDateError {
      
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable>") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            " The resource type of "+getName().lastComponent()+" sets it "+
            "as immutable for its secondary groups."); 
    
     setEorL_SecondaryGroup(_L);
  } // setEorL_SecondaryGroupForUser()
  
  /**
  * A secondary group is used by an Actor as rights user, to get complementary
  * accesses to targets only when a current userID is set. When the actor
  * is accessed as any target, the current userID of the rights user is activated
  * with its secondary groups, and the secondary groups returned by this method
  * are not activated.
  * Unregisters the previous groupIDs and set the new ones.
  * Removes in the argument, the current group or GroupIDImpl.EMPTY_INSTANCE.
  * Reads the ACS 'GroupID.NoSecondaryGroup' values to check if each argument
  * may be a secondary group.
  * Called by setEorL_SecondaryGroupForUser() and setEorCurrentUserID().
  * Calls GroupID.add/removeOwnActor().
  * Fire an event 'SecondaryGroups'.
  * @param _L array of the secondary groupID attached to this actor. May be null.
  * @exception UpDateError if the secondary groupIDs are immutable.
  */
  protected synchronized void setEorL_SecondaryGroup(ImmutableGroupID[] _L)
          throws UpDateError {
    
    if (!aCS_.getManageSecGroupsForActor()) return;
      
    ImmutableGroupID[] l_OldSecondaryGroup = getEorL_SecondaryGroup();
    // removes all the previous secondary groupID
    if (l_SecondaryGroup_ != null) {
      try {
        for (int i = 0; i < l_SecondaryGroup_.length; i++) {
          if (l_SecondaryGroup_[i] != null && 
                  !l_SecondaryGroup_[i].equals (GroupIDImpl.EMPTY_INSTANCE ) )
            ((GroupIDImpl)l_SecondaryGroup_[i]).removeOwnActorAsSecondaryGroup(this);
        } // for
      } catch (UpDateError u) { getEorACS().getEorBaseError().addException
      ("in ActorImpl.setEorL_SecondaryGroup(), the actor "+getName().toString()
        +"  was not registered in a previous groupID");
      }
    } // if

    // sets a new l_SecondaryGroup_
    if (_L == null) {
      l_SecondaryGroup_ = null;
      try {
          firePropertyChange("SecondaryGroups", l_OldSecondaryGroup, null);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.setEorL_SecondaryGroup(), firePropertChange() throws:");
            e.printStackTrace();
      } 
          
      l_OldSecondaryGroup = null;
      return;
    } else { // _L != null
      String[] l_remove = (String[]) getEorACS().getM_TypesAndGUIPolicy().
              get("GroupID.NoSecondaryGroup");
      boolean remove = false;
      int index = -1;
         
      for (int  j = 0 ; j < _L.length ; j++) {
          if (_L[j] == null) continue;
          if (l_remove != null) {
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(_L[j].getEPType())) {
                      remove = true;
                      index = j;
                      break;
                  }
              } // for i
          } // if (l_remove !=
          if (remove) break;  
      } // for j
      l_remove = null;
      if (remove)
          throw new UpDateError("The "+_L[index]+" type is in the ACS values"+
                  "for 'GroupID.NoSecondaryGroup': "+
                  "it cannot be the secondary group of an actor.");        
    } // else
    
    // register this actor in all the new secondary groupID
    short shiftingBack = 0;
    ImmutableGroupID[] newArray = new GroupIDImpl[_L.length];
    ImmutableGroupID[] result = new GroupIDImpl[1];

    for (int  i = 0 ; i < _L.length ; i++) {
      if (_L[i] != null) {
        if (_L[i].equals (new GroupIDImpl() ) ||
                _L[i].equals( currentGroupID_ )) {
              // incremente shiftingBack for not copying _L[i]
          shiftingBack++;
        }
        else {
          newArray[i - shiftingBack] = _L[i];
          ((GroupIDImpl)_L[i]).addOwnActorAsSecondaryGroup(this); // registers this actor
        }
      }
      else shiftingBack++; // for not copying _L[i]
    } // for

    if (_L.length != shiftingBack) {
      result = (ImmutableGroupID[]) new GroupIDImpl[_L.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
    }
    else result = null;
    l_SecondaryGroup_ = result;

    try {
        firePropertyChange("SecondaryGroups", l_OldSecondaryGroup, getEorL_SecondaryGroup());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ActorImpl.setEorL_SecondaryGroup(), firePropertChange() throws:");
        e.printStackTrace();
    } 

    newArray = null;
    result = null;
    return;
  } // setEorL_SecondaryGroup()

  /**
  * A secondary GroupID adds new access to the actor as rights user.
  * A secondary GroupID is independant to the GroupID of the actor
  * as resource, but this primary group may be also a seocndary group.
  * @return secondary GroupIDs under which this actor runs a rights user.
  *      May be empty, but not null.
  */
  public ImmutableGroupID[] getEorL_SecondaryGroup() {
    if (l_SecondaryGroup_ == null) return new ImmutableGroupID[0];
    ImmutableGroupID[] result = new GroupIDImpl[l_SecondaryGroup_.length];
    System.arraycopy(l_SecondaryGroup_, 0, result, 0, l_SecondaryGroup_.length);
    return result;
  } // getEorL_SecondaryGroup()

  /**
  * @return true if _grp is a secondary groupID which have this actor
  * as membership.
  */
  public boolean isMemberToSecondaryGroup(ImmutableGroupID _grp) {
    if (l_SecondaryGroup_ == null) return false;
    for (int i = 0 ; i < l_SecondaryGroup_.length ; i++) {
      if (l_SecondaryGroup_[i] == _grp) return true;
    }
    return false;
  }

/**
 * Gets the resources connected to this eligible party.
 * The resources may belong to differents ACS,
 * as the aclEntries. One resource may have differents AclEntries
 * (of different ACS) which connect it to this eligible party.
 * Then, the returned list contains only one reference to this resource.
 * @return the unmodifiable list of references on resources. 
 * May be Collections.EMPTY_LIST, but not null.
 */
  public final List<ImmutableResource> getEorL_AclConnectedResources()	{
    if (m_AclEntriesResources_ != null) {
        List<ImmutableResource> L = new ArrayList( m_AclEntriesResources_.values() );
        return Collections.unmodifiableList(L);
    } else return Collections.EMPTY_LIST;
  }

  /**
   * If _st is allowed by the ACS, set the type of this EligibleParty which is also a resource.
   * This method is used to switch from an immutable type
   * which forbids the updates of a protected property, and to return to this protecting type.
   * Then, it is possible to switch for example from 'actor' to
   * 'immutable actor', but not to 'immutable group'.
   * From the default value 'undefined', any first update may be done.
   * Calls super.setResourceType().
   * Fires the events 'ResourceType' and 'EPType'.
   * @param _st type of the resource.
   * @exception UpDateError if the type is null, equal to &lt;byAcsCreator&gt;,
   *    or not known by the ACS, or not associated to the previous type.
   */
  public synchronized void setEPType(String _st) throws UpDateError {
      String old = type_;
      // calls BasicImpl.setType()
      setResourceType(_st);

      try {
          // 'ResourceType' is fired by BasicImpl.setType()
          firePropertyChange("EPType", old, type_);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ResourceImpl.setResourceType(), firePropertChange() throws:");
            e.printStackTrace();
      } // catch

      old = null;
  } // setEPType()

  /**
   * As an EligibleParty, but it is simply the Basic type.
   * @return type of the eligible party. May be null.
   */
  public String getEPType()	{
          return type_;
  }

  /**
   * A bridge is an unidirectionnal connexion between two ACS through 
   * the mean of two actors, one in each ACS.
   * @return ImmutableActor[] external actors for which this actor is a bridge target. 
   *    May be empty, but not null.
   */
  public final ImmutableActor[] getEorL_bridgeSources() {
    if (l_bridgeSources_ == null) return new ImmutableActor[0];
    ImmutableActor[] result = new ActorImpl[l_bridgeSources_.length];
    System.arraycopy(l_bridgeSources_, 0, result, 0, l_bridgeSources_.length);
    return result;
  }

  /**
   * @return true if this actor is a bridge source,
   *  created through a specific constructor, and updated when the
   *  external actor is closed or deleted.
   */
  public boolean getIsACSBridgeSource() {
    if (bridgeTarget_ != null) return true;
    else return false;
  }

  /**
   * Get the another-ACS actor for which this resource is a bridge source.
   * @return a reference on the actor. Is null if
   *   this instance is not a bridge, if the target ACS is not opened.
   */
  public final ImmutableActor getEorACSBridgeTarget() {
    return bridgeTarget_;
  }

  /**
   * The default value is true.
   * @return true if this Actor may be a proxy in the access paths, that is,
   *    as rights user, it gets some rights to the other
   *    rights users which have an 'execute' right on it
   */
  public final boolean getIsRightProxy() {
    return isRightProxy_;
  }

  /**
   * Noop if the resource type contains 'byAcsCreator' or 'immutable'.
   * Calls ACSImpl.removeNoRightProxyActor(). Overridden by the AcsAddon Ubuntu,
   * that calls it. Fires an event 'isRightProxy'.
   * @param _b true if this Actor may be a proxy in the access paths, that is,
   *    as rights user, it gets some rights to the other
   *    rights users which have an 'execute' right on it
  * @exception UpDateError if the right proxy flag is immutable.
   */
  public void setIsRightProxy(boolean _b) throws UpDateError {
      
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable>") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The resource type of "+getName().lastComponent()+" sets it "+
            "as immutable for its right proxy flag."); 
      
      boolean old = isRightProxy_; 
      isRightProxy_ = _b;
      
        // update the ACS list
      if (old && !isRightProxy_ ) {
                 ((ACSImpl)aCS_).addNoRightProxyActor(this);
      } // if
      else if (!old && isRightProxy_) {
                 ((ACSImpl)aCS_).removeNoRightProxyActor(this);
      } // if
      
      try {
        firePropertyChange("IsRightProxy", old, isRightProxy_);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.setIsRightProxy(), firePropertChange() throws:");
            e.printStackTrace();
      } 
  } // setIsRightProxy()

 //
 // -------- Public methods from EligibleParty -------------------
 //

   /**
   * Sets a privilege as applicable to this access source, and this is called only by the program.
   * The gBase classes are the only caller of this method
   * at any creation/deletion of a typed privilege or an access source.
   * Fires 'OneSourceTypedPrivilege' events.
   * @param _priv is a Privilege to register in the source
   * @exception BaseError if the ACS does not manage the privileges,
   *   or if this instance is not a privilege source.
   */
  public void addEorPrivilegeForTypeToSource(PrivilegeForType _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if

    privHandler_.addEorPrivilegeToSource(_priv);
  }
  
   /**
   * Sets a privilege as applicable to this access source.
   * Called by PrivilegeForLinksImpl.addEorPrivilegeSourceLink().
   * If the source is also a source container,
   * the privilege is used by the children of the target if the target
   * is a GroupIDMember (as any right), or, if it is a DirectoryEP,
   * it may be used by the children if its ACS policy specifies it.
   * Fires 'OneSourceLinkedPrivilege' events.
   * @param _priv is a Privilege to register in the source
   * @exception BaseError if the ACS does not manage the privileges,
   *   if the type sets the instance as immutable,
   *   this instance is not a privilege source, or: <P>
   *   for all linked privileges: <P>
   *   - if the number of sources will be superior to 4, or <P>
   *   - the source is null, empty, or its name is null, or <P>
   *   - the source name is already known by this instance.
   *   for LINKED_FOR_ONE_TO_ONE privileges, if: <P>
   *   - there is already a source, or <P>
   *   - the parameter is not a GroupID <P>
   *   and, for all PrivilegeForLinks excepted for LINKED_FOR_ONE_TO_ONE privileges:
   *   - the source belongs to another acs than the privilege acs
   */
  protected void addEorPrivilegeForLinksToSource(PrivilegeForLinks _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new BaseError(
            "The type of "+getName().lastComponent()+" sets it as immutable for its privileges.");

    privHandler_.addEorPrivilegeToSource(_priv);
  } // addEorPrivilegeToSource()

   /**
   * Unregisters the privilege, and this instance in the linked privilege
   * when it is no more used by this privilege source.
   * Fires 'OneSourcePrivilege' event.
   * @param _priv to unregister
   * @exception BaseError if the privilege and this source do not known each other,
   *   if the type sets the instance as immutable.
   *    User-displayable message.
   */
  public void removeEorPrivilegeToSource(Privilege _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (_priv instanceof PrivilegeForLinks && type_ != null &&
            (type_.startsWith("<byAcsCreator>") ||
            type_.startsWith("<immutable> ") ||
            type_.endsWith("<immutable> ")))
          throw new BaseError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its privileges.");

    privHandler_.removeEorPrivilegeToSource(_priv);
  } // removeEorPrivilegeToSource()

   /**
   * Gets all the privileges linked to this source.
   * There is two types of privilege links (allowed or not by the ACS):
   * - PrivilegeForLinks: linked privilege established for a couple
   *      (ACSObject, privilege)
   * - PrivilegeForType: typed privilege, for which the ACSObject and the privilege
   *      has a common type (and a common class if it is a bindedClasses
   *      privilege)
   * @return List of Privileges. May be empty, but never null.
   */
  public List getEorL_SourcePrivileges() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_SourcePrivileges();
  } // getEorL_SourcePrivileges()
  
   /**
   * Gets all the PrivilegeForTypes linked to this source.
   * @return List of PrivilegeForType. May be empty, but never null.
   */
  public List getEorL_SourcePrivilegeForTypes() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_SourcePrivilegeForTypes();
  }
  
   /**
   * Gets all the privilegeForLinks linked to this source.
   * Note: for LINKED_IF_ALL linked privileges, the presence of the privilege
   * does not mean that the master has the privilege rights.
   * Returns an immutable empty list if getIsSource() is false.
   * @return List of PrivilegeForLinks. May be empty, but never null.
   */
  public List getEorL_SourcePrivilegeForLinks() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_SourcePrivilegeForLinks();
  } // getEorL_SourcePrivilegeForLinks()

 //
 // ------- Public methods from Resource -------------------------
 //
 
  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'SecondaryGroups', 'l_bridgeSources',
  * 'CurrentGroupID', 'CurrentUserID', 'ACSBridgeTarget', 'EPType', 'IsRightProxy',
  * 'ActorAlias', 'AclConnectedResources', 'OneSourceTypedPrivilege',
  * 'OneSourceLinkedPrivilege'.
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (_propertyName.equals("OneSourceTypedPrivilege") ||
                _propertyName.equals("OneSourceLinkedPrivilege")) {
                    if (privHandler_ == null ) {
                        try {
                          privHandler_ = new CPrivilegeUserImpl(this);
                        } catch (CreateError c) {
                            new InternalError("From "+name_+"- "+c.toString());
                        }
                    } // if

                    privHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else super.addPropertyChangeListener(_propertyName, _l);

  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'SecondaryGroups', 'l_bridgeSources',
  * 'CurrentGroupID', 'CurrentUserID', 'ACSBridgeTarget', 'EPType', 'IsRightProxy',
  * 'ActorAlias', 'AclConnectedResources', 'OneSourceTypedPrivilege',
  * 'OneSourceLinkedPrivilege'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      if (_propertyName.equals("OneSourceTypedPrivilege") ||
                _propertyName.equals("OneSourceLinkedPrivilege") ) {
                    if (privHandler_ == null ) {
                        try {
                          privHandler_ = new CPrivilegeUserImpl(this);
                        } catch (CreateError c) {
                            new InternalError("From "+name_+"- "+c.toString());
                        }
                    } // if

                    privHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else super.removePropertyChangeListener(_propertyName, _l);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {

      if (_propertyName.equals("OneSourceTypedPrivilege") ||
                _propertyName.equals("OneSourceLinkedPrivilege")) {
          
            if (privHandler_ != null )
                return privHandler_.getPropertyChangeListeners(_propertyName);
            else
                return null;
      }
      else return super.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()
  
   /**
   * This method removes the actor in the eligible party ACS list,
   * and in the resources connected through an aclEntry.
   * An unused object has to be deleted to inform the others objects, and this method
   * HAS TO BE USED before deleting the actor.
   * Note: when this method is called by the ResourceImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the eligible party is unknown by the main ACS
   * (if it is a clone, typically), if the ACS of an aclEntry is closed or unknown,
   *  or if the bridge relation, if any, cannot be broken, or if the type of
   *  an associated object fires an exception.
   */
   public synchronized void finalizeForBase() throws UpDateError {

                              // for testing purpose
                              //System.out.println("call of ActorImpl.finalizeForBase() for "+
                                //      getName()+" with a current userID: "+currentUserID_);

    // noop if it is the second call to this method
    if (folderMember_ == null) {
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");
      return;
    }
    // if called by the Resource constructor, it is an instance initialized only as Resource
    if (!actorActivated_) {
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");
      super.finalizeForBase();
      return;
    }
    
    // intialization
    Resource res = null;
    
    // removes the recorded aclEntries in the relevant resources, in m_AclEntriesResources_.
    // if it is the same ACS, the work has been already done through super.finalizeForBase()
    if (m_AclEntriesResources_ != null) {
        List L = new ArrayList( m_AclEntriesResources_.values() );
        Iterator ite = L.iterator();
        for (Iterator iter = ite ; iter.hasNext() ;) {
            res = (Resource)iter.next();
            AclEntry[] aclArray = res.getL_aclEntry();
            if (aclArray != null) {
              for (int i = 0 ; i < aclArray.length ; i++) {

                  if (aclArray[i].getEorEP() == this) {
                      if (!aclArray[i].getL_ConditionSourceGroups().isEmpty())
                          res.removeAclEntry(aclArray[i].getEorACS(), this,
                            aclArray[i].getPositiveRight(),
                            aclArray[i].getL_ConditionSourceGroups().iterator().next());
                      else
                          res.removeAclEntry(aclArray[i].getEorACS(), this,
                            aclArray[i].getPositiveRight(),
                            null);
                  } // if
                  
              } // for int i
            } // for if (aclArray != null)
        } //for iter
        L = null;
    } // if

    // removes this in its current userID
    if (currentUserID_ != null && currentUserID_ != UserIDImpl.EMPTY_INSTANCE)
      ((UserIDImpl)currentUserID_).removeCurrentActor(this);

    // removes this in its current groupID
    if (currentGroupID_ != null)
      ((GroupIDImpl)currentGroupID_).removeCurrentActor(this);

    currentGroupID_ = null;
    currentUserID_ = null;

    // removes all the previous secondary groupID
    if (l_SecondaryGroup_ != null) {
      try {
        for (int i = 0; i < l_SecondaryGroup_.length; i++) {
          if (l_SecondaryGroup_[i] != null && !l_SecondaryGroup_[i].
            equals ( GroupIDImpl.EMPTY_INSTANCE ) ) {
              ((GroupIDImpl)l_SecondaryGroup_[i]).removeOwnActorAsSecondaryGroup(this);

                        //System.out.println("In ActorImpl.finalizeForBase(),"+
                          //" removing of: "+l_SecondaryGroup_[i]);
            }
        } // for
      } catch (UpDateError u) {
        getEorACS().getEorBaseError().addException
          ("in ActorImpl.finalizeForBase(), the actor "+getName().toString()
          +"  has not been registered in a previous groupID");
        getEorACS().getEorBaseError().addException(u);
      }
    } // if

    // removes all the previous bridge sources
    if (l_bridgeSources_ != null) {
      try {
          // updates l_bridgeSources_ through a call to removeBridgeSource()
        ActorImpl[] l_copy = new ActorImpl[l_bridgeSources_.length];
        System.arraycopy(l_bridgeSources_, 0, l_copy, 0, l_bridgeSources_.length);

        for (int i = 0; i < l_copy.length; i++) {
          if (l_copy[i] != null  ) {
              ((ActorImpl)l_copy[i]).setEorACSBridgeTarget(null, (BaseObject)this);

                                //System.out.println("In ActorImpl.finalizeForBase(),"+
                                  //" after the removing of the target in the ACS "+
                                  //l_copy[i].getEorACS().getNickName());
           } // if
        } // for
        l_copy = null;
      } catch (CreateError u) {
        getEorACS().getEorBaseError().addException
          ("in ActorImpl.finalizeForBase(), the actor "+getName().toString()
          +"  has not been unregistered in a bridge source.");
        getEorACS().getEorBaseError().addException(u);
      } // catch
    } // if

                                        // for testing and tracking
                                    //System.out.println("in ActorImpl.finalizeForBase() "+
                                      //"call to removeEorEligibleParty() "+
                                      //"for the thread "+Thread.currentThread().getName()+
                                      //" with the priority: "+Thread.currentThread().getPriority());
    
    // removes (this) to the map of eligible parties in the main ACS.
    // if it is a clone, it may doesn't exist for the ACS and UpDateError is thrown.
    // if it is a bridge source with a closed target, updates the closed bridges map.
    ((ACSImpl)getEorACS()).removeEorEligibleParty(this);

    l_bridgeSources_ = null;
    //l_isAGateFor_ = null;
    l_SecondaryGroup_ = null;
    m_AclEntriesResources_ = null;

    try {
         if (bridgeTarget_ != null) {
             // removes the previous value in the two ACS
             ((ACSImpl)bridgeTarget_.getEorACS()).
                  removeOwnActorBridgeTargetFromSource(this, this);
             // calls back getEorACSBridgeTarget() for reading bridgeTarget_
             ((ACSImpl)getEorACS()).removeOwnActorBridgeSource(this);

             ((ActorImpl)bridgeTarget_).removeBridgeSource(this);
             bridgeTarget_ = null;
         } // if (bridgeTarget_ != null)
    } catch (CreateError c) {
        throw new UpDateError("in ActorImpl.finalizeForBase(), the bridge relation"+
          " is wrong.");
    }   
    
    super.finalizeForBase(); // remove (this) to the map of resources in the main ACS.
  } // finalizeForBase()

  /**
   * @return the static serialVersionUID__ of this Actor class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * Returns a deep copy without registering of the new object in the ACSs.
   * Caution: if the instance is a bridge target, this relation
   * is not cloned since there is no information on the source.
   * But if it is a bridge source, its target is cloned.
   * @exception CloneNotSupportedException
   * @return Object is not registred in the ACSs. The names, the aclEntries,
   * the Linux/Unix rights, the ACS controller names, the type,
   * the bridge target are cloned.
   * The external object references are not cloned (see "Eor" in get method).
   */
  public synchronized Object clone() {
    ActorImpl clon = (ActorImpl) super.clone();

    if (l_bridgeSources_ != null) {
      clon.l_bridgeSources_ = new ActorImpl[l_bridgeSources_.length];
      System.arraycopy(l_bridgeSources_, 0, clon.l_bridgeSources_, 0, l_bridgeSources_.length);
    }

    //if (l_isAGateFor_ != null) {
      //clon.l_isAGateFor_ = new ActorImpl[l_isAGateFor_.length];
      //System.arraycopy(l_isAGateFor_, 0, clon.l_isAGateFor_, 0, l_isAGateFor_.length);
    //}

    if (l_SecondaryGroup_ != null) {
      clon.l_SecondaryGroup_ = new GroupIDImpl[l_SecondaryGroup_.length];
      System.arraycopy(l_SecondaryGroup_, 0, clon.l_SecondaryGroup_, 0, l_SecondaryGroup_.length);
    } else clon.l_SecondaryGroup_ = null;

    clon.currentUserID_ = currentUserID_;
    clon.currentGroupID_ = currentGroupID_;

    if (m_AclEntriesResources_ != null)
        clon.m_AclEntriesResources_ = (SortedMap) ((TreeMap)m_AclEntriesResources_).clone();
    
    clon.bridgeTarget_ = bridgeTarget_;
    
    return clon;
  } // clone()

  /**
   * Tests all the internal Actor properties, and all the superclasses properties.
   * @param _obj
   * @return true if _obj is an instance of ActorImpl (or an inherited class)
   * equals to this instance
   */
  public synchronized boolean equals (Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      if (!super.equals(_obj))
          return false;

      ActorImpl obj = (ActorImpl) _obj;

      if (l_bridgeSources_ != null ) {
        if (l_bridgeSources_.length != obj.getEorL_bridgeSources().length)
          return false;
        for (int i = 0; i < l_bridgeSources_.length; i++) {
          if (l_bridgeSources_[i] == null | obj.getEorL_bridgeSources()[i] == null) {
            if (l_bridgeSources_[i] != null | obj.getEorL_bridgeSources()[i] != null)
              return false;
            continue; // i++
          } // if
          else {
            if (!l_bridgeSources_[i].equals( obj.getEorL_bridgeSources()[i] ))
              return false;
          } // else
        } // for
      } else if(obj.getEorL_bridgeSources() != null) return false;

      if (l_SecondaryGroup_ == null || obj.getEorL_SecondaryGroup() == null) {
          if (l_SecondaryGroup_ != null || obj.getEorL_SecondaryGroup() != null)
            return false;
      } // if
      else {
        if (l_SecondaryGroup_.length != obj.getEorL_SecondaryGroup().length)
          return false;
        for (int i = 0; i < l_SecondaryGroup_.length; i++) {
          if (l_SecondaryGroup_[i] == null | obj.getEorL_SecondaryGroup()[i] == null) {
            if (l_SecondaryGroup_[i] != null | obj.getEorL_SecondaryGroup()[i] != null)
              return false;
            continue; // i++
          } // if
          else {
            if (!l_SecondaryGroup_[i].equals( obj.getEorL_SecondaryGroup()[i] ))
              return false;
          } // else
        } // for
      } // else

      if (currentUserID_ == null | obj.getEorCurrentUserID() == null) {
          if (currentUserID_ != null | obj.getEorCurrentUserID() != null)
            return false;
      } // if
      else {
        if (!currentUserID_.equals( obj.getEorCurrentUserID() ))
          return false;
      } //else

      if (currentGroupID_ == null | obj.getEorCurrentGroupID() == null) {
          if (currentGroupID_ != null | obj.getEorCurrentGroupID() != null)
            return false;
      } // if
      else {
          if (!currentGroupID_.equals( obj.getEorCurrentGroupID() ))
            return false;
      } //else

      if (getEorL_AclConnectedResources() == null || obj.getEorL_AclConnectedResources() == null) {
          if (getEorL_AclConnectedResources() != null || obj.getEorL_AclConnectedResources() != null)
            return false;  // if
      } // if
      else {
        if (!getEorL_AclConnectedResources().equals( obj.getEorL_AclConnectedResources() )) return false;
      }

      if (bridgeTarget_ == null || obj.getEorACSBridgeTarget() == null) {
          if (bridgeTarget_ != null || obj.getEorACSBridgeTarget() != null)
            return false;
      } // if
      else {
          if (!bridgeTarget_.getName().equals( obj.getEorACSBridgeTarget().getName() ))
            return false;
      } // else
      
      return true;
  } // equals

    /**
   * @return the hashcode value for this object
   */
  public int hashCode() {
    int result = super.hashCode();

/*
    if (l_bridgeSources_ != null) {
      for (int i = 0; i < l_bridgeSources_.length; i++) {
        if (l_bridgeSources_[i] != null) {
          result = PRIME*result + l_bridgeSources_[i].hashCode();
        }
      } // for
    }

    if (currentGroupID_ != null) {
      if (currentGroupID_.getName() != null) {
        result = PRIME*result + currentGroupID_.getName().hashCode();
      }
    }
    if (currentUserID_ != null) {
      if (currentUserID_.getName() != null) {
        result = PRIME*result + currentUserID_.getName().hashCode();
      }
    }
 */

//    if(actor_alias_ != null)
//      result = PRIME*result + actor_alias_.hashCode();

        // epType_ not used

/*
    if (l_SecondaryGroup_ != null) {
      for (int i = 0; i < l_SecondaryGroup_.length; i++) {
        if (l_SecondaryGroup_[i] != null) {
          if (l_SecondaryGroup_[i].getName() != null) {
            result = PRIME*result + l_SecondaryGroup_[i].getName().hashCode();
          }
        }
      } // for
    } // if
*/

    if (bridgeTarget_ != null ) {
        result = PRIME*result + bridgeTarget_.getName().hashCode();
    }
    
    return result;
  } // hashCode()
  

  /**
   * Sets the another-ACS actor for which this instance is a bridge source.
   * Fires a 'ACSBridgeTarget' event.
   * @param _target the bridge target. May be null.
   * @exception CreateError if the target belongs to the same ACS, or if the
   *    target changing is not validated by the connected objects.
   */
  public void setEorACSBridgeTarget(Actor _target) 
  throws CreateError {
      setEorACSBridgeTarget(_target, (BaseObject)null);
  } //setEorACSBridgeTarget()

  //
  // ------------- PROTECTED METHODS ----------------------------------------
  //

  /**
   * Sets the another-ACS actor for which this resource is a bridge source.
   * Updates the bridge target and the two ACS.
   * Fires a 'ACSBridgeTarget' event after the ACS relevant events.
   * Does not check the resource type to allow the updating.
   * Called by the constructor, by ACSImpl.closeBridges(), checkClosedBridges(),
   * removeOwnActorBridgeTargetFromSource() and setEorACSBridgeTarget(ImmutableActor),
   * and by finalizeForBase().
   * Calls add/removeBridgeSource().
   * @param _target is the bridge target. May be null.
   * @param _caller is the object which calls this method. Used to avoid
   *   a loop with ACSImpl.removeOwnActorBridgeTargetFromSource(). Null
   *   when called from setEorACSBridgeTarget(ImmutableActor).
   * @exception CreateError if the target belongs to the same ACS, or if the
   *    target changing is not validated by the connected objects.
   */
  protected void setEorACSBridgeTarget(Actor _target, BaseObject _caller)
    throws CreateError {
      
     if (bridgeTarget_ == _target || !getEorACS().getManageBridges())
         return;
     if (_target != null && _target.getEorACS() == getEorACS())
         throw new CreateError("The new bridge target does not belong to another ACS.");
     ImmutableActor oldTarget; 
     String[]       l_types = null;

     if (_target != null) {
         l_types = _target.getEorACS().getM_TypeOrientedPolicy().get(
             "Actor.IsNotBridgeTarget");

         if (l_types != null && Arrays.asList(l_types).contains(
                          UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                          _target.getResourceType())) )
             throw new CreateError("From the ACS policy 'Actor.IsNotBridgeSource',"+
                                     "the type of the bridge target forbids"+
                                     " it to be such a target.");
     } // if

     try {
         // removes the previous value in the target ACS
         if (bridgeTarget_ != null && _caller != bridgeTarget_.getEorACS())
              ((ACSImpl)bridgeTarget_.getEorACS()).
                  removeOwnActorBridgeTargetFromSource(this, this);
         if (bridgeTarget_ != null) {
             // calls back getEorACSBridgeTarget() for reading bridgeTarget_
             ((ACSImpl)getEorACS()).removeOwnActorBridgeSource(this);

             ((ActorImpl)bridgeTarget_).removeBridgeSource(this);
         } // if

         // sets the new value
         oldTarget = bridgeTarget_;
         bridgeTarget_ = _target;        

         // sets the two ACS
         if (_target != null) {

                                            // for testing and tracking
                                            //System.out.println("In ActorImpl.setEorACSBridgeTarget(),"+
                                              //" set the new target "+_target.getName());

             ((ACSImpl)getEorACS()).addOwnActorBridgeSource(this);

             ((ACSImpl)_target.getEorACS()).
               addOwnActorBridgeTarget(this, _target);
             // may fires InternalError in an ACS addon
             ((ActorImpl)_target).addBridgeSource(this);
             
         } // if
         else {

                                            // for testing and tracking
                                            //System.out.println("In ActorImpl.setEorACSBridgeTarget(),"+
                                              //" set a null target ");
         } // else
     } catch (CreateError c) {
         throw new CreateError("ActorImpl.setEorACSBridgeTarget() - the bridge relation"+
           " is wrong: "+c);
     } catch (InternalError c) {
         throw new CreateError("ActorImpl.setEorACSBridgeTarget() - the bridge relation"+
           " is wrong: "+c);
     } // catch

     try {
         firePropertyChange("ACSBridgeTarget", oldTarget, _target);
     } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.setEorACSBridgeTarget(), "+
                    "on the event 'ACSBridgeTarget', firePropertChange() throws:");
            e.printStackTrace();
     }
     oldTarget = null;
     l_types = null;
  } // setEorACSBridgeTarget()
 
  /**
  * Adds an actor as source to this instance which is then a bridge target for this source.
  * Invoked by Actor.setEorACSBridgeTarget().
  * Fires an event on "l_bridgeSources".
  * @param _act is a bridge source for this actor as target.
  */
  protected void addBridgeSource(Actor _act) {
    // variables
    Actor[]     newArray = new ActorImpl[1];
    boolean     recorded = false;

    if (l_bridgeSources_ == null) l_bridgeSources_ = (Actor[]) new ActorImpl[1];

    Actor[]     oldValue = new ActorImpl[l_bridgeSources_.length];
    System.arraycopy(l_bridgeSources_, 0, oldValue, 0, l_bridgeSources_.length);

    // add in an empty place if possible
    for (int i = 0 ; i < l_bridgeSources_.length ; i++) {
        if (l_bridgeSources_[i] == null || l_bridgeSources_[i].equals(ActorImpl.EMPTY_INSTANCE) ) {
            l_bridgeSources_[i] = _act;
            recorded = true;
            break;
        }
    }
    // else, add at the end of l_bridgeSources_
    if (!recorded) {
        newArray = new ActorImpl[l_bridgeSources_.length + 1];
        System.arraycopy(l_bridgeSources_, 0, newArray, 0, l_bridgeSources_.length);
        newArray[l_bridgeSources_.length] = _act;
        l_bridgeSources_ = new ActorImpl[newArray.length];
        l_bridgeSources_ = newArray;
    }
    else {
        newArray = new ActorImpl[l_bridgeSources_.length];
        System.arraycopy(l_bridgeSources_, 0, newArray, 0, l_bridgeSources_.length);
    }

                                            // for testing and tracking
                                            //System.out.println("In ActorImpl.addBridgeSource(),"+
                                              //" l_bridgeSources_: ");
                                            //for (int i = 0; i < l_bridgeSources_.length; i++) {
                                              //  System.out.println("---- "+
                                                //        l_bridgeSources_[i].getName());
                                            //} // for i


    try {
        firePropertyChange("l_bridgeSources", oldValue, newArray);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ActorImpl.addBridgeSource(), firePropertChange() throws:");
        e.printStackTrace();
    }
    oldValue = null;
    newArray = null;
  } // addBridgeSource()
 
  /**
  * Remove an actor as bridge source for this instance.
  * Invoked by Actor.setEorACSBridgeTarget().
  * Fire an event on "l_bridgeSources".
  * @param _act is a bridge source for this actor as target.
  */
  protected void removeBridgeSource(Actor _act) {
    // noop if finalizeForBase() has been called
    if (l_bridgeSources_ == null) return;

    // exceptions
    if (_act == null) {
      getEorACS().getEorBaseError().addException(
        "In ActorImpl.removeBridgeSource(_act), _act == null ");
    }

    // variables and initializations
    short        count = 0;
    short        shiftingBack = 0;
    // l_bridgeSources_ can't be null in this method, so:
    ImmutableActor[]      newArray = new ActorImpl[l_bridgeSources_.length];
    ImmutableActor[]      result = new ActorImpl[0];
    ImmutableActor[]      oldValue = new ActorImpl[l_bridgeSources_.length];
    System.arraycopy(l_bridgeSources_, 0, oldValue, 0, l_bridgeSources_.length);

    // count the occurences of _act:
    for (int i = 0 ; i < l_bridgeSources_.length ; i++) {
      if (l_bridgeSources_[i] != null ) {
        if (l_bridgeSources_[i].equals(_act) ) count++;
      }
    } // for i

    // exceptions
    if (count == 0 && l_bridgeSources_.length != 0) {
        getEorACS().getEorBaseError().addException(
          "In ActorImpl.removeBridgeSource(_act), _act is unknown ");
    }

    // update l_bridgeSources_
    for (int  i = 0 ; i < l_bridgeSources_.length ; i++) {
      if (l_bridgeSources_[i] != null) {
        // incremente shiftingBack delete the reference l_bridgeSources_[i]:
        if (l_bridgeSources_[i].equals(_act) ) shiftingBack++;
        else newArray[i - shiftingBack] = l_bridgeSources_[i];
      }
      else shiftingBack++; // delete null l_bridgeSources_[i]
    } // for

    if (l_bridgeSources_.length != shiftingBack) {
      result = new ActorImpl[l_bridgeSources_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, l_bridgeSources_.length - shiftingBack);
      l_bridgeSources_ = new ActorImpl[result.length];
      System.arraycopy(result, 0, l_bridgeSources_, 0, result.length);
    } // if
    else {
      l_bridgeSources_ = result;
    } // else

    try {
        firePropertyChange("l_bridgeSources", oldValue, getEorL_bridgeSources());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ActorImpl.removeBridgeSource(), firePropertChange() throws:");
        e.printStackTrace();
    }
    oldValue = null;
    newArray = null;
    result = null;
  } // removeBridgeSource()
 
  /**
   * A bridge is an unidirectionnal connexion between two ACS through 
   * the mean of two actors, one in each ACS.
   * Called by copyOwnVariables().
   * Fire an event 'l_bridgeSources'.
   * @param _L ImmutableActor[] external actors for which this actor is a bridge target. 
   *    May be empty or null.
   */
  protected void setEorL_bridgeSources(ImmutableActor[] _L) {
    ImmutableActor[] oldL_bridgeSources = l_bridgeSources_;
    if (_L != null) {
      l_bridgeSources_ = new ActorImpl[_L.length];
      System.arraycopy(_L, 0, l_bridgeSources_, 0, _L.length);
    } else l_bridgeSources_ = null;

    try {
        firePropertyChange("l_bridgeSources", oldL_bridgeSources, _L);
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.setEorL_bridgeSources(), firePropertChange() throws:");
            e.printStackTrace();
    } 
    
    oldL_bridgeSources = null;
  } // setEorL_bridgeSources()
  
   /**
  * This method finalizes all the attributes, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null.
  * Called by EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

              // for testing purpose
              //System.out.println("call of ActorImpl.finalizeFromDeserialization() for "+getName());

    if (folderMember_ == null) return;

    l_bridgeSources_ = null;
    //l_isAGateFor_ = null;
    currentGroupID_ = null;
    currentUserID_ = null;
    l_SecondaryGroup_ = null;
    m_AclEntriesResources_ = null;
    bridgeTarget_ = null;
    
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()
  
  /**
  * This method provides a copy of the instance variables.
  * Copy the aclEntries.
  * Caution: copies the current user and group, and the secondary groups,
  * if it is the same ACS,
  * or else in every UserID and GroupID copy which is found.
  * In this second case, the user or group name must be made
  * of the acs name ended by the user or group string name.
  * Caution: does not copy the bridge sources, if they exist, nor the actor alias.
  * Uses the super overridden method with one argument.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError
  **/
  protected ActorImpl copyOwnVariables(ActorImpl _result, boolean _withoutInternalAcl)
    throws CreateError {

              // for testing and tracking
              //System.out.println("Call of ActorImpl.copyOwnVariables() with _result: "+
                //_result+"; and _result ACS: "+_result.getEorACS());
        
      if (_result == null) {
          throw new CreateError("in ActorImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      _result = (ActorImpl)super.copyOwnVariables(_result, _withoutInternalAcl);
      
      _result.setEorL_bridgeSources(l_bridgeSources_);
      //_result.setEorL_isAGateFor(l_isAGateFor_);
      try {
          if (_result.getEorACS() == getEorACS()) {
              _result.setEorCurrentUserID(getEorCurrentUserID()); // does not copy directly
              _result.setEorCurrentGroupID(currentGroupID_);
              _result.setEorL_SecondaryGroup(l_SecondaryGroup_);
          } else {
            // if the copy is for another ACS, try to find
            // the users and groups
            // with the same name last components.
            // there, the user and group name must be made
            // of the acs name ended by the group string name
            ImmutableName name;
            if (getEorCurrentUserID() != null) {
                name = ((Name)_result.getEorACS().getName()).extended( 
                currentUserID_.getName().lastComponent() );
                try {
                  _result.setEorCurrentUserID(
                    (UserID)BaseUtilityImpl.getEorBaseObject(name));
                } catch (BaseError b) {
                  // no operation if name is not an user name
                } // catch
            }
            if (getEorCurrentGroupID() != null) {
                name = ((Name)_result.getEorACS().getName()).extended( 
                currentGroupID_.getName().lastComponent() );
                try {
                  _result.setEorCurrentGroupID(
                    (GroupID)BaseUtilityImpl.getEorBaseObject(name));
                } catch (BaseError b) {
                  // no operation if name is not a group name
                } // catch
            }
            name = null;

            if (l_SecondaryGroup_ != null) {
                short shiftingBack = 0;
                ImmutableGroupID[] newArray = new GroupIDImpl[l_SecondaryGroup_.length];
                ImmutableGroupID[] l_secGroups = new GroupIDImpl[1];
                ImmutableName grp_name;

                for (int  i = 0 ; i < l_SecondaryGroup_.length ; i++) {
                    if (l_SecondaryGroup_[i] == null) {
                        shiftingBack++;
                        continue;
                    }
                    if (l_SecondaryGroup_[i].equals (new GroupIDImpl() )) {
                          // incremente shiftingBack for not copying _L[i]
                      shiftingBack++;
                    }
                    else {
                        grp_name = ((Name)_result.getEorACS().
                            getName()).extended( 
                            l_SecondaryGroup_[i].getName().lastComponent() );
                        try {
                            newArray[i - shiftingBack] =
                              (GroupID)BaseUtilityImpl.getEorBaseObject(grp_name);
                        } catch (BaseError b) {
                              // no operation if grp_name is not a group name
                        } // catch
                    } // else
                } // for

                if (l_SecondaryGroup_.length != shiftingBack) {
                  l_secGroups = (ImmutableGroupID[]) new GroupIDImpl[
                    l_SecondaryGroup_.length - shiftingBack];
                  System.arraycopy(newArray, 0, l_secGroups, 0,
                    newArray.length - shiftingBack);
                }
                else l_secGroups = null;
                if (l_secGroups != null) _result.setEorL_SecondaryGroup(l_secGroups);

                grp_name = null;
                newArray = null;
                l_secGroups = null;
            } // if (l_SecondaryGroup_ != null) 

          } // else
      } catch (UpDateError u) {}
      
      // to copy all the external aclEntries and the internal ones
      // if _withoutInternalAcl is false
      if (m_AclEntriesResources_ != null) {
            // attributes
            AclEntry[]            l_aclCopy = null;
            Resource              res = null;

            for (Iterator ite = m_AclEntriesResources_.values().iterator(); ite.hasNext(); ) {
                res = (Resource)ite.next();
                l_aclCopy = res.getL_aclEntry();
                // iterate on all aclEntries of the resource
                for (int i = 0 ; i < l_aclCopy.length ; i++) {
                    if (l_aclCopy[i] == null) continue;
                    if ( l_aclCopy[i].isEmpty() ) continue;
                    // copy the aclEntry with _result as EP
                    if ( l_aclCopy[i].getEorEP() == this ) {
                        if (!_withoutInternalAcl) {
                            res.addAclEntry(l_aclCopy[i].getEorACS(), _result,
                              l_aclCopy[i].getL_Rights(), l_aclCopy[i].getPositiveRight(),
                              l_aclCopy[i].getL_ConditionSourceGroups().iterator().next() );
                        } // if
                        // always copies the acl if it is an external one
                        else if (res.getEorACS() != getEorACS()) {
                            res.addAclEntry(l_aclCopy[i].getEorACS(), _result,
                              l_aclCopy[i].getL_Rights(), l_aclCopy[i].getPositiveRight(),
                              l_aclCopy[i].getL_ConditionSourceGroups().iterator().next());
                        } // else if
                    } // if
                } // for i
            } // for ite

            l_aclCopy = null;
            res = null;
      } // if
      
      return _result;
  } // copyOwnVariables(,,,)

   /**
   * Adds an aclEntry to this instance as an EligibleParty, not as a Resource.
   * Called by the ResourceImpl.addAclEntry().
   * There, the AclEntry ACS knowns the AclEntry.
   * CAUTION: any update of this method must be also done in EligiblePartyImpl.
   * @param _acl to this EP. 
   *  The resource connected to _acl must have a name following the pattern
   *  IS_name / ACS_name / resource_name, to allow the unicity
   *  of the name. The ACS of the aclEntry may be different to the ACS
   *  of the eligible party.
   */
  protected final void addAclEntryToEP(AclEntry _acl) {

    List l_OldAclEntriesResources = getEorL_AclConnectedResources(); // may be null

    if (_acl.getEorResource() == null) throw new InternalError("add AclEntry"+
      " with AclEntry.getEorResource() == null");

    ImmutableName name = _acl.getEorResource().getName();
    if (m_AclEntriesResources_ == null) m_AclEntriesResources_ = new TreeMap();

    synchronized (m_AclEntriesResources_) {
      if (!m_AclEntriesResources_.containsKey(name)) {

        m_AclEntriesResources_.put( (ImmutableName)name.clone() , _acl.getEorResource());
        
        try {
            firePropertyChange("AclConnectedResources", l_OldAclEntriesResources,
              getEorL_AclConnectedResources());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.addAclEntryToEP(), firePropertChange() throws:");
            e.printStackTrace();
        }    
 
      } // if
    } // synchro

    name = null;
    l_OldAclEntriesResources = null;
  } // addAclEntryToEP()
  
   /**
   * Removes the related resource of an aclEntry,
   * as a kind of an EligibleParty, not as a resource, if the argument is the single
   * AclEntry for the resource. This condition is handled by the calling method.
   * Called from AclEntryImpl.finalizeForUser().
   * The related resource is unrecorded in L_AclConnectedResources
   * since _acl is the last one connected to this resource.
   * CAUTION: any update of this method must be also done
   *  in EligiblePartyImpl, and any other EligibleParty direct implementation.
   * @param _acl to remove
   */
  protected final void removeAclEntryToEP(AclEntry _acl) {
    List l_OldAclEntriesResources = getEorL_AclConnectedResources(); // may be empty

              // for testing and tracking
              //System.out.println("Call of ActorImpl.removeAclEntryToEP() in: "+getName());


    synchronized (m_AclEntriesResources_) {
      Object obj = m_AclEntriesResources_.remove( ((AclEntryImpl)_acl).getResName() );

      if (obj == null) throw new InternalError("in EligiblePartyImpl."+
        "removeAclEntryToEP(), null resource or null resource name");

      if (m_AclEntriesResources_.isEmpty()) m_AclEntriesResources_ = null;

      try {
          firePropertyChange("AclConnectedResources", l_OldAclEntriesResources,
             getEorL_AclConnectedResources());
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.removeAclEntryToEP(), firePropertChange() throws:");
            e.printStackTrace();
      } 

      obj = null;
    } // synchro

    l_OldAclEntriesResources = null;
  } // removeAclEntryToEP()

  /**
   * Remove the current userID of the resource.
   * To use only when the userID is finalizing.
   * In the other cases, use setEorCurrentUserID().
   * The main differences with setEorCurrentUserID() are that removeCurrentActor(this)
   * is not invoked in this method, and there is no checking of the resource type.
   * Called by UserID.finalizeForBase().
   */
  protected void removeEorCurrentUserID() {
    ImmutableUserID oldCurrentUserID = currentUserID_;
    currentUserID_ = UserIDImpl.EMPTY_INSTANCE;

    if (oldCurrentUserID != null && oldCurrentUserID != UserIDImpl.EMPTY_INSTANCE) {
      try {
        firePropertyChange("CurrentUserID", oldCurrentUserID, null);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.removeEorCurrentUserID(), firePropertChange() throws:");
            e.printStackTrace();
      } 
        
      oldCurrentUserID = null;
    }
  } // removeEorCurrentUserID()


  /**
 * Remove the current groupID of the resource.
 * To use only when the groupID is finalizing.
 * In the other cases, use setEorCurrentGroupID().
 * The main differences with setEorCurrentGroupID() are that removeCurrentActor(this)
 * is not invoked in this method, and there is no checking of the resource type.
 * Called by UserID.removePrimGroup().
  */
  protected void removeEorCurrentGroupID() {
    ImmutableGroupID oldCurrentGroupID = currentGroupID_;
    currentGroupID_ = null;

    if (oldCurrentGroupID != null) {
      try {
        firePropertyChange("CurrentGroupID", oldCurrentGroupID, null);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ActorImpl.removeEorCurrentGroupID(), firePropertChange() throws:");
            e.printStackTrace();
      } 
        
      oldCurrentGroupID = null;
    }
  }
  
  /**
  * A secondary group is used by an Actor as rights user, to get complementary
  * accesses to targets.
  * Call GroupID.add/removeOwnActor().
  * Fire an event 'SecondaryGroups'.
  * Called by GroupID.finalizeForBase(), which is not called back.
  * @param _grp secondary groupID attached to this actor. May be null.
  */
  protected synchronized void removeSecondaryGroup(ImmutableGroupID _grp) {
      
    ImmutableGroupID[] l_OldSecondaryGroup = getEorL_SecondaryGroup();
    short shiftingBack = 0;
    ImmutableGroupID[] newArray = new GroupIDImpl[l_SecondaryGroup_.length];
    ImmutableGroupID[] result = new GroupIDImpl[1];

    for (int  i = 0 ; i < l_SecondaryGroup_.length ; i++) {
      if (l_SecondaryGroup_[i] != null) {
        if (l_SecondaryGroup_[i].equals (_grp )) {
              // incremente shiftingBack for not copying l_SecondaryGroup_[i]
          shiftingBack++;
        }
        else {
          newArray[i - shiftingBack] = l_SecondaryGroup_[i];
        }
      }
      else shiftingBack++; // for not copying l_SecondaryGroup_[i]
    } // for

    if (l_SecondaryGroup_.length != shiftingBack) {
      result = (ImmutableGroupID[]) new GroupIDImpl[l_SecondaryGroup_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
    }
    else result = null;
    l_SecondaryGroup_ = result;

    try {
        firePropertyChange("SecondaryGroups", l_OldSecondaryGroup, getEorL_SecondaryGroup());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ActorImpl.removeSecondaryGroup(), firePropertChange() throws:");
        e.printStackTrace();
    } 

    newArray = null;
    result = null;
    l_OldSecondaryGroup = null;
  } // removeSecondaryGroup()

  /**
   * Reads the key 'Actor.NoCurrentGroup' in the type-oriented ACS policy,
   * and returns false only if this instance cannot have a current group.
   * Called by setEorCurrentGroupID() and setEorCurrentUserID().
  * @return true if the type-oriented ACS policy authorizes this instance to have
   * a current GroupID, or if the type is null.
  */
  protected boolean getAuthorizeCurrentGroupID() {

    // checkup 'Actor.NoCurrentGroup'
    if (type_ != null) {
        String reducedType = null;
        String[] l_policy = null;

        reducedType = UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(type_);

        if (aCS_.getM_TypeOrientedPolicy() != null) {
            l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get("Actor.NoCurrentGroup");

            if (l_policy != null && Arrays.asList(l_policy).contains(reducedType)) {
                reducedType = null;
                l_policy = null;
                return false;
            } // if (l_policy != null &&

            reducedType = null;
            l_policy = null;
        } // if
    } // if (type_ != null)
    return true;
  } // getAuthorizeCurrentGroupID()

} // ActorImpl

