//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.ErrB.BaseError;
import ARoad0.Pattern.Singleton;
import ARoad0.Pattern.BeanInfoPattern;
import ARoad0.gDMak.ActionNewEligibleParty;

import java.beans.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.Void;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Collections;
import java.util.Arrays;
import java.lang.reflect.Method;

/**
* This class is a utility class for gBase to manage the association scripts.
* Associations to do are calls to constructors and methods in gBase,
* with all the required arguments. <P>
* Associations may be to process when a new object of a given class and type is created in an ACS,
* or when a child ACS specifies constraints on its ACSObject roots in its parent ACS.
* The AcsAddon classes are managed. <P>
* The first type of association is in the same ACS, at the creation or at the deletion
* of a new object which is called the master object.
* The second type of associations to handle is
* for a child ACS which has an associate, called root, in its parent ACSs. <P>
* For instance, it allows, after each creation of an account,
* to create a group in the same ACS having the same last component in its name, 
* then to register the new account in this new group. <P>
* A second example is when a child ACS opening sets the creation
* of a root like an Actor in a direct or indirect parent ACS. <P>
* A third example is when a child ACS sets the integrity of a root in a parent ACS,
* as a criterion to ensure its proper working. <P>
* The main limitations are: <P>
* - it is not possible to add associations to a new AclEntry or a new Privilege.
* It is possible to create an AclEntry or a Privilege after the creation of
* an ACSObject, and then to edit the properties of this AclEntry or Privilege. <P>
* - it is possible to use an ACSObject as argument in a constructor or a method
* if its name may be get simply from the new object name or the ACS name. It is not
* possible to compute a name for instance from the parent name of the new object,
* or to get it through a given method called on this new object.  <P>
* These limitations about the associations may be overcomed by the way of
* a dedicated AcsAddon. It is easy to add to a gBase class some methods
* for the handling of the object constructions. The AcsAddon designer has the choice
* to use an association for calling these specific methods, or to directly code
* their calling in the AcsAddon gBase classes. With an association, the script
* is visible in the GUI for the advanced users, on the contrary of a hard-coded function. <P>
* The allowed String map keys are described in the methods 
* getL_MapKeyStartsForAssociationsOnNewObject and getL_MapKeyStartsForAllRoots. <P>
* The map values define one or several actions to process or to check. 
* The keys that are applicable are searched for every creation/deletion of an ACSObject, 
* and at every opening/closing of an ACS. <P>
* The controlAssociations method gives details on the syntax of the scripts.
* @see #controlAssociations
* @version 0.7.0
*/
public class AssociateUtilityImpl implements Singleton {
      /* the only instance of the class, following singleton pattern */
  private static AssociateUtilityImpl  instance__ = new AssociateUtilityImpl();

  private static final int        INITIAL_CAPACITY = 10; // the default value: 101

    /* Constructors
     */

  /** with this private constructor, the compiler won't generate
  * a default public constructor.
  */
  private AssociateUtilityImpl() {
  }

    /**
    * Only method to obtain the unique instance of AssociateUtilityImpl.
    * @return AssociateUtilityImpl
    */
  public static AssociateUtilityImpl getInstance() {
    return instance__;
  } // getInstance()

  /**
   * This method returns the allowed startings for the values
   * which are used as map keys for associations on new objects.
   * Does not allow the associations on new AclEntries or Privileges.
   * Not used for the child ACS roots. <P>
   * The allowed keys are starting by a range of values including: <P>
   * - 'Class'.'Property'.&lt;Value&gt;.XXX: a new object property has the value XXX; <P>
   * - 'Class'.'Property'.&lt;ValueContains&gt;.XXX: a new object property value contains the value XXX;
   *    when the property is a Name, this is the name last component which contains the value XXX. <P>
   * - 'Class'.'Property'.&lt;Class&gt;.XXX: a new object property has a class with
   *        exactly the simple name XXX; <P>
   * - 'Class'.'Property'.&lt;NameStartsWith&gt;.XXX: the new object property name
   *        starts with the ACS name plus the subname XXX; <P>
   * - 'Class'.&lt;NameStartsWith&gt;.XXX: the new object name of the class
   *        starts with the ACS name plus the subname XXX; <P>
   * - &lt;Class&gt;.&lt;Value&gt;.XXX: the new object class is exactly gBase.XXX,
   * both for a generic class and for an AcsAddon class; <P>
   * The first word 'Class' may have the value 'Resource', 'Directory', 'Actor', 'GroupID',
   * 'UserID' or 'VirtualFolder'.
   * The key starting 'ACS.Name.&lt;Value&gt;' is forbidden there, since it is reserved
   * to associations for the roots.
   * Any implied object property has to be an argument in the object constructor.
   * If a returned value is ending by a point, it is a true starting value.
   * Otherwise, it is a completed value which forbids any adding.
   * @see #addAssociatesAfterCreation
   * @see #getL_MapKeyStartsForAllRoots
   * @return List containing the allowed String values
   */
  public static final List<String> getL_MapKeyStartsForAssociationsOnNewObject() {

    List<String> l_result = new ArrayList<String>(10);
    l_result.add("Resource.Name.<ValueContains>.");
    l_result.add("Resource.ParentNode.<Value>.");
    l_result.add("Resource.ParentNode.<NameStartsWith>.");
    l_result.add("Resource.UserID.<Value>.");
    l_result.add("Resource.ResourceType.<Value>.");
    l_result.add("Resource.ResourceType.<ValueContains>.");
    l_result.add("Resource.UserID.<Class>.");

    l_result.add("Directory.Name.<ValueContains>.");
    l_result.add("Directory.ParentNode.<Value>.");
    l_result.add("Directory.ParentNode.<NameStartsWith>.");
    l_result.add("Directory.UserID.<Value>.");
    l_result.add("Directory.ResourceType.<Value>.");
    l_result.add("Directory.ResourceType.<ValueContains>.");
    l_result.add("Directory.UserID.<Class>.");

    l_result.add("Actor.Name.<ValueContains>.");
    l_result.add("Actor.ParentNode.<Value>.");
    l_result.add("Actor.ParentNode.<NameStartsWith>.");
    l_result.add("Actor.UserID.<Value>.");
    l_result.add("Actor.ResourceType.<Value>.");
    l_result.add("Actor.ResourceType.<ValueContains>.");
    l_result.add("Actor.UserID.<Class>.");
    l_result.add("Actor.EPType.<Value>.");
    l_result.add("Actor.EPType.<ValueContains>.");

    l_result.add("GroupID.Name.<ValueContains>.");
    l_result.add("GroupID.<NameStartsWith>.");
    l_result.add("GroupID.EPType.<Value>.");
    l_result.add("GroupID.EPType.<ValueContains>.");

    l_result.add("UserID.Name.<ValueContains>.");
    l_result.add("UserID.<NameStartsWith>.");
    l_result.add("UserID.EPType.<Value>.");
    l_result.add("UserID.EPType.<ValueContains>.");

    l_result.add("VirtualFolder.Name.<ValueContains>.");
    l_result.add("VirtualFolder.ParentNode.<Value>.");
    l_result.add("VirtualFolder.ParentNode.<NameStartsWith>.");
    l_result.add("VirtualFolder.Type.<Value>.");
    l_result.add("VirtualFolder.Type.<ValueContains>.");
    l_result.add("VirtualFolder.ForResourcesOnly.<Value>.");

    l_result.add("<Class>.<Value>.Resource");
    l_result.add("<Class>.<Value>.Directory");
    l_result.add("<Class>.<Value>.Actor");
    l_result.add("<Class>.<Value>.GroupID");
    l_result.add("<Class>.<Value>.UserID");
    l_result.add("<Class>.<Value>.VirtualFolder");

    return l_result;
  } // getL_MapKeyStartsForAssociationsOnNewObject()

  /**
   * Defines the possible cases for associations, and it depends on the argument type.
   * The possible types are 'Resource', 'Directory', 'Actor', 'UserID', 'GroupID'
   * and 'VirtualFolder'. Only one type is selected, if any.
   * Not for child ACS roots.
   * Called by add/removeAssociatesAfterCreation() and
   * ACSFactoryImpl.addAssociationsToNewObject().
   * @see #getL_MapKeyStartsForAssociationsOnNewObject
   * @param _obj just-created by the GUI, having or not some associates
   * @return List containing the keys which match to the new object. May be null
   *    or empty.
   */
  public static final Set<String> getL_MapKeyMatchingForAssociationsOnNewObject(ACSObject _obj) {

    Map             m_asso = null;
    if (_obj instanceof Resource)
        m_asso = ((Resource)_obj).getEorACS().getM_AssociationsOnNewObject();
    else if (_obj instanceof EligibleParty)
        m_asso = ((EligibleParty)_obj).getEorACS().getM_AssociationsOnNewObject();
    else if (_obj instanceof VirtualFolder)
        m_asso = ((VirtualFolder)_obj).getEorACS().getM_AssociationsOnNewObject();

    if (m_asso == null || m_asso.isEmpty()) {
        m_asso = null;
        return null;
    }

    Set<String>     l_result = new HashSet<String>(20);
    String          test = null;
    String          type = null;
    String          key = null;
    List<String>    l_ValueContains = new ArrayList<String>(15);
    List<String>    l_NameStarts = new ArrayList<String>(15);

    // initializes the type
    if (_obj instanceof Resource) {
            if (!(_obj instanceof Directory) &&
                !(_obj instanceof Actor)) type = "Resource";
            else if (_obj instanceof Directory) type = "Directory";
            else if (_obj instanceof Actor) type = "Actor";
    }
    else if (_obj instanceof EligibleParty) {
            if (_obj instanceof UserID) type = "UserID";
            else if (_obj instanceof GroupID) type = "GroupID";
    }
    else if (_obj instanceof VirtualFolder) {
            type = "VirtualFolder";
    }

    // initializes l_ValueContains for 'Name.<ValueContains>' and
    // any 'String property.<ValueContains>', for 'ResourceType', 'EPType' or 'Type',
    // for which the value is get through a method having a name following
    // the pattern get'property name', like getResourceType().
    for (Iterator ite = m_asso.keySet().iterator(); ite.hasNext(); ) {
        test = (String)ite.next();
        if (test.indexOf(".Name.<ValueContains>.") != -1)
            l_ValueContains.add(test);
        else if (test.indexOf(".Name.<ValueContains>.") == -1 &&
                test.indexOf(".<ValueContains>.") != -1)
            l_ValueContains.add(test);
    } // for ite

    // searching for in l_ValueContains
    for (Iterator ite = l_ValueContains.iterator(); ite.hasNext(); ) {
        key = (String)ite.next();
        if (key.startsWith(type+".Name.<ValueContains>.") &&
                _obj.getName().lastComponent().indexOf(
                    key.substring(type.length()+22) ) != -1)
                        l_result.add(key);
        else if (key.startsWith(type+".ResourceType.<ValueContains>.") &&
                ((Resource)_obj).getResourceType().indexOf(
                    key.substring(type.length()+30)) != -1)
                        l_result.add(key);
        else if (key.startsWith(type+".EPType.<ValueContains>.") &&
                ((EligibleParty)_obj).getEPType().indexOf(
                    key.substring(type.length()+24)) != -1)
                        l_result.add(key);
        else if (key.startsWith(type+".Type.<ValueContains>.") &&
                ((VirtualFolder)_obj).getType().indexOf(
                    key.substring(type.length()+22)) != -1)
                        l_result.add(key);
    } // for ite

    if (_obj instanceof Resource ) {

        if (((Resource)_obj).getEorUserID() != null) {
            test = type+".UserID.<Value>."+
                        ((Resource)_obj).getEorUserID().getName().toString();
            if (m_asso.containsKey(test)) l_result.add(test);
        } // if

        test = type+".ResourceType.<Value>."+
                        ((Resource)_obj).getResourceType();
        if (m_asso.containsKey(test)) l_result.add(test);

        if (((Resource)_obj).getEorUserID() != null) {
            test = type+".UserID.<Class>."+
                        ((Resource)_obj).getEorUserID().getClass().getSimpleName();
            if (m_asso.containsKey(test)) l_result.add(test);
        } // if


        if (type.equals("Resource") && m_asso.containsKey("<Class>.<Value>.Resource") )
                l_result.add("<Class>.<Value>.Resource");
        else if (type.equals("Directory") && m_asso.containsKey("<Class>.<Value>.Directory") )
                l_result.add("<Class>.<Value>.Directory");
        else if (type.equals("Actor") && m_asso.containsKey("<Class>.<Value>.Actor") )
                l_result.add("<Class>.<Value>.Actor");

        if (((Resource)_obj).getEorParent() != null) {
            test = type+".ParentNode.<Value>."+
                        ((Resource)_obj).getEorParent().getName().toString();
            if (m_asso.containsKey(test)) l_result.add(test);

            // initializes l_NameStarts
            for (Iterator ite = m_asso.keySet().iterator(); ite.hasNext(); ) {
                test = (String)ite.next();
                if (test.indexOf(".ParentNode.<NameStartsWith>.") != -1)
                    l_NameStarts.add(test);
            } // for ite

            // searching for in l_NameStarts for 'type.ParentNode.<NameStartsWith>.'
            for (Iterator ite = l_NameStarts.iterator(); ite.hasNext(); ) {
                key = (String)ite.next();
                if (key.startsWith(type+".ParentNode.<NameStartsWith>.") &&
                        ((Resource)_obj).getEorParent().getName().toString().startsWith(
                            key.substring(type.length()+29) ) )
                                l_result.add(key);
            } // for ite
        } // if (((Resource)_obj).getEorParent() != null) {

    } // if (_obj instanceof Resource

    if (_obj instanceof EligibleParty ) {

        test = type+".EPType.<Value>."+
                        ((EligibleParty)_obj).getEPType();
        if (m_asso.containsKey(test)) l_result.add(test);

        if (type.equals("UserID") && m_asso.containsKey("<Class>.<Value>.UserID") )
                l_result.add("<Class>.<Value>.UserID");
        else if (type.equals("GroupID") && m_asso.containsKey("<Class>.<Value>.GroupID") )
                l_result.add("<Class>.<Value>.GroupID");

        // initializes l_NameStarts
        l_NameStarts.clear();
        for (Iterator ite = m_asso.keySet().iterator(); ite.hasNext(); ) {
            test = (String)ite.next();
            if (test.indexOf(".<NameStartsWith>.") != -1 &&
                    test.indexOf(".ParentNode.") == -1) l_NameStarts.add(test);
        } // for ite

        // searching for in l_NameStarts for 'UserID.<NameStartsWith>.' and
        // 'GroupID.<NameStartsWith>.'
        for (Iterator ite = l_NameStarts.iterator(); ite.hasNext(); ) {
            key = (String)ite.next();
            if (key.startsWith(type+".<NameStartsWith>.") &&
                    _obj.getName().lastComponent().startsWith(
                        key.substring(type.length()+18) ) )
                            l_result.add(key);
        } // for ite

    } // if (_obj instanceof EligibleParty

    if (_obj instanceof VirtualFolder ) {

        test = type+".SecondaryType.<Value>."+
                        ((VirtualFolder)_obj).getType();
        if (m_asso.containsKey(test)) l_result.add(test);
        test = type+".ForResourcesOnly.<Value>.true";
        if (m_asso.containsKey(test) &&
                ((VirtualFolder)_obj).getType().equals(VirtualFolderImpl.RES_VFOLDER) )
            l_result.add(test);
        test = type+".ForResourcesOnly.<Value>.false";
        if (m_asso.containsKey(test) &&
                ((VirtualFolder)_obj).getType().equals(VirtualFolderImpl.EP_VFOLDER) )
            l_result.add(test);
        
        if (m_asso.containsKey("<Class>.<Value>.VirtualFolder") )
                l_result.add("<Class>.<Value>.VirtualFolder");

        if (((VirtualFolder)_obj).getEorParent() != null) {
            test = type+".ParentNode.<Value>."+
                        ((VirtualFolder)_obj).getEorParent().getName().toString();
            if (m_asso.containsKey(test)) l_result.add(test);

            // initializes l_NameStarts which is empty
            for (Iterator ite = m_asso.keySet().iterator(); ite.hasNext(); ) {
                test = (String)ite.next();
                if (test.indexOf(".ParentNode.<NameStartsWith>.") != -1)
                    l_NameStarts.add(test);
            } // for ite

            // searching for in l_NameStarts for 'type.ParentNode.<NameStartsWith>.'
            for (Iterator ite = l_NameStarts.iterator(); ite.hasNext(); ) {
                key = (String)ite.next();
                if (key.startsWith(type+".ParentNode.<NameStartsWith>.") &&
                        ((VirtualFolder)_obj).getEorParent().getName().toString().startsWith(
                            key.substring(type.length()+29) ) )
                                l_result.add(key);
            } // for ite
        } // if (((VirtualFolder)_obj).getEorParent() != null) {

    } // if (_obj instanceof VirtualFolder

                            // for testing and tracking 
                    //System.out.println("in AssociateUtilityImpl.getL_MapKeyMatchingForAssociationsOnNewObject(), "+
                      //      "for "+_obj.getName()+
                        //    ", the type is "+type+", l_result is: ");
                    //for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                      //      System.out.println("-------------- "+ite.next());
                    //} // for i

    m_asso = null;
    test = null;
    type = null;
    key = null;
    l_ValueContains = null;
    l_NameStarts = null;

    return l_result;
  } // getL_MapKeyMatchingForAssociationsOnNewObject()

  /**
   * This method returns the allowed startings for the values
   * which are used as map keys for the child ACS roots. Not used for
   * associations on new objects. <P>
   * The criteria is a map key with one allowed form for the starting, which
   * is 'ACS.Name.&lt;Value&gt;.'.
   * Any implied object property has to be an argument in the object constructor.
   * If a returned value is ending by a point, it is a true starting value.
   * Otherwise, it is a complete value which forbids any adding.
   * @see ACSFactoryImpl#addOpeningNoClosingRoots
   * @see #getL_MapKeyStartsForAssociationsOnNewObject
   * @return List containing the allowed String values
   */
  public static final List<String> getL_MapKeyStartsForAllRoots() {

    List<String> l_result = new ArrayList<String>(10);
    l_result.add("ACS.Name.<Value>.");

    return l_result;
  } // getL_MapKeyStartsForAllRoots()
  
  /**
   * Gets the associations to handle by the mean of actions to do by the program.
   * Actions have numbers from 1 to 999.
   * Called by controlAssociations(), ACSFactoryImpl.addAssociations(),
   * getM_AssociationsOnNewObject(), getM_OpeningAndClosingRoots(), getM_OpeningNoClosingRoots(),
   * getM_IntegrityRoots() and createACS().
   * @param _m_Associations actions to do. May be null or empty.
   * @param _filtered true if, since the internal map may contain values where 
   *   the first characters are not 'N.', 'NN.' nor 'NNN.', the first no-number value
   *   and all the followings are not returned by this method
   * @return a copied map where the key is a String, and the value is an array
   *   of Strings, started (if _filtered false) or not by a number between 1 to 999
   *   followed by a point. May be empty, but not null.
   */
  public static final HashMap<String,String[]> getM_FilteredAssociations(
          HashMap<String, String[]> _m_Associations, boolean _filtered) {
      
      HashMap<String,String[]>      m_result = 
                                        new HashMap<String,String[]>(INITIAL_CAPACITY); 
      if (_m_Associations == null || _m_Associations.isEmpty())
          return m_result;
      
      String[]                      l_values = null;
      Map.Entry<String,String[]>     entry = null;
      String[]                      result = null;
      int                           jump = -1;

              // for testing and tracking
              //System.out.println("in AssociateUtilityImpl.getM_FilteredAssociations(): ");
      
      for (Iterator ite = _m_Associations.entrySet().iterator(); ite.hasNext(); ) {
         entry = (Map.Entry<String,String[]>)ite.next();

         l_values = entry.getValue();
         
         if (l_values == null ) continue;
         if (_filtered) {
             for (int i = 0; i < l_values.length; i++) {
                 if (l_values[i] == null ||
                         (!l_values[i].substring(1, 2).equals(".") &&
                         !l_values[i].substring(2, 3).equals(".") &&
                         !l_values[i].substring(3, 4).equals("."))) {
                            jump = i;
                            break;
                 } // if
             } // for i
             if (jump == 0) {
                 jump = -1;
                 continue;
             }
         } // _filtered
         
         if (jump == -1) {
             result = new String[l_values.length];
             System.arraycopy(l_values, 0, result, 0, l_values.length);
         }  
         else {
             result = new String[jump];
             System.arraycopy(l_values, 0, result, 0, jump);
             jump = -1;
         }
           
         m_result.put(entry.getKey(), result);
         
              // for testing and tracking
              //System.out.println("AssociateUtilityImpl.getM_FilteredAssociations() in _filtered: "+_filtered);
              //System.out.println("--- for the key "+entry.getKey()); 
              //for (int i = 0; i < result.length; i++) {
                //    System.out.println("--------- at "+i+": "+result[i]);                  
              //} // for i
         
      } // for ite

      
      l_values = null;
      entry = null;
      result = null;     
      return m_result;
  } // getM_FilteredAssociations()

  /**
   * Checks up the two types of associations to handle for a new ACS.
   * The first type of association is in the same ACS, at the creation or at the deletion
   * of a new object which is called the master object.
   * The second type of associations to handle is
   * for a child ACS which has an associate, called root, in its parent ACSs. <P>
   * It is not possible to add associations to a new AclEntry or a new Privilege.
   * The allowed String map keys are described in the methods
   * getL_MapKeyStartsForAssociationsOnNewObject and getL_MapKeyStartsForAllRoots. <P>
   * The map values define one or several actions to process or to check.
   * The keys that are applicable are searched for every creation/deletion of an ACSObject,
   * and at every opening/closing of an ACS. <P>
   * In the allowed String map values, the first value has a specific type
   * among 5 possible forms. <P>
   * After the first value, the allowed String map values have another specific type
   * among 10 possible forms. For the masters, the values after the first one are
   * all starting by one of the 9 'PropertyName' forms, since the association specifies
   * then the arguments to use. For the child ACS roots, it is more complex since
   * there are 2 options after the first value: <P>
   * 1 - the association specifies a root in the parent ACS which should be there or created
   * at the child ACS opening, or should be deleted at the child ACS closing, and
   * then the allowed map values, after the first one, are arguments to create it,
   * to check its presence or to delete it, and these arguments are like the master arguments,
   * using both &lt;AssociateClass&gt; and some &lt;CalledMethod&gt; first values, <P>
   * 2 - the association specifies a root integrity as a criterion for
   * the working of the child ACS, and this means that any change on the inner properties
   * of the root may alterate the working of the child ACS, including all its not-simulated 
   * root inner properties (for instance a child ACS parameter, if the root is 
   * a configuration text file); then, the first value is starting by the key word
   * &lt;AssociateClass&gt;, the second value is a 'PropertyName' to give the name of this root,
   * and the third and last value is the key word '&lt;GeneralCriterion&gt;.&lt;RootIntegrity&gt;'. <P> <P>
   * Action specification: <P>
   * There are 5 allowed forms for the first value of the action N for the masters,
   * but only 3 authorized forms for the child ACS roots: <P>
   * - &lt;AssociateClass&gt;.XXX which specifies the class name XXX in the generic gBase
   *    or in the AcsAddon gBase, without the 'Impl' end,
   *    of an associate object to build up or to finalize when the master is build up
   *    or finalized, or the child ACS root to open, to close or to check up <P>
   * - &lt;CalledMethod&gt;.&lt;Master&gt;.YYY which specifies the method YYY to call on the master
   *    (forbidden for child ACS roots), <P>
   * - &lt;CalledMethod>.&lt;ACS&gt;.YYY which specifies the method YYY to call on the master or the parent ACS;
   *    for the masters, the only authorized values are 'handleAssociationsForInnerObject',
   *    for the parent ACS, there is no method restriction, <P>
   * - &lt;CalledMethod&gt;.&lt;Action&gt;.C.ZZZ which specifies the method ZZZ
   *    to call on the associate previously created by the action number C <P>
   * - &lt;CalledMethod&gt;.&lt;Object&gt;.XXX+&lt;Class&gt;+YYY+&lt;Value&gt;+ZZZ which specifies the method ZZZ
   *    to call on the future ACSObject having the name XXX and the class ZZZ,
   *     where the parameters must be the master or an associate
   *    (to set the reverse operation when it is finalized) (forbidden for child ACS roots). <P> <P>
   * With '&lt;AssociateClass&gt;', the associate instance is not created if it is returned
   * by BaseUtilityImpl.getBaseObject() from the name in the first argument.
   * As usual ,the associate is then put in the list of accessible objects for
   * any '&lt;CalledMethod&gt;.&lt;Action&gt;' association. <P>
   * Argument specification: <P>
   * After each first value, there is the optional list of the specifications for the arguments.
   * Arguments may be for an object constructor, for the root integrity marker (case 1)
   * or for the method named in the action (cases 2 to 5).
   * There is one specification per argument, so one command per argument. <P>
   * Each argument specification must fulfill one of the 10 allowed forms. They are
   * presented herein without the number 'N.', sometimes at the start for the action N: <P>
   * (1) 'PropertyName'.&lt;Null&gt;: null value for the argument having the name 'PropertyName'. <P>
   * For any first value. For the master creating/deleting, and for the child ACS root
   * opening/deleting. <P>
   * (2) 'PropertyName'.&lt;Master&gt;: the master is the argument <P>
   * For any first value. For the master creating/deleting only,
   * and for the child ACS root opening/deleting where 'Master' is the parent ACS. <P>
   * (3) 'PropertyName'.&lt;MasterCreation&gt;: boolean true if the master is creating, <P>
   * false if it is finalizing. For any first value. For the master creating/deleting,
   * and for the child ACS root creating/deleting. <P>
   * (4) 'PropertyName'.&lt;ArgumentValue&gt;.C: the associate number C object. <P>
   * For any first value. For the master creating/deleting, and for the child
   * ACS root opening/deleting. <P>
   * (5) 'PropertyName'.&lt;Value&gt;.XXX: the argument is XXX as String, ImmutableName <P>
   * if it is the property type, true/false if the type is boolean,
   * or any BaseObject get from its name. For any first value. For the master creating/deleting,
   * and for the child ACS root opening/deleting. <P>
   * (6) 'PropertyName'.&lt;Value&gt;.XXX+&lt;MasterValue&gt;+YYY: the master value <P>
   * for the same property in its BeanInfo (the read method has to not return an array),
   * with some optional pre and post String values if the property type is String.
   * For any first value. For the master creating/deleting,
   * and for the child ACS root opening/deleting where 'Master' is the parent ACS. <P>>
   * (7) 'PropertyName'.&lt;Value&gt;.XXX+&lt;MasterNameLastComponent&gt;+YYY: <P>
   * the last component in the master name, with some optional pre and post Name values as Strings.
   * For any first value. For the master creating/deleting, and for the child
   * ACS root opening/deleting where 'Master' is the parent ACS. <P>
   * (8) 'PropertyName'.&lt;Value&gt;.&lt;MasterNameContext&gt;+YYY <P>
   * the context in the master name, with a post Name value as String
   * For any first value. For the master creating/deleting, and for the child ACS
   * root opening/deleting where 'Master' is the parent ACS. <P>
   * (9) 'PropertyName'.&lt;Value&gt;.&lt;ACSName&gt;+YYY: the ACS name, <P>
   * with a post Name value as String. For any first value. For the master creating/deleting,
   * it is the master ACS name. For the child ACS root opening/deleting and
   * the integrity criteria, it is the parent ACS name. <P>
   * (10) 'PropertyName'.&lt;Value&gt;.XXX+&lt;ChildNameLastComponent&gt;+YYY: <P>
   * the last component in the child ACS name, with some optional pre and post
   * Name values as Strings. For any first value. Only for the child ACS
   * root opening/deleting and the integrity criteria. <P> <P>
   * The root integrity marker is not an argument specification. It is the 11th allowed form
   * after the first value, and only one occurence is authorized in the array of values.
   * If it is there, '&lt;GeneralCriterion&gt;.' must always be in the last value: <P>
   * - &lt;GeneralCriterion&gt;.&lt;RootIntegrity&gt;: any change on the inner properties <P>
   * of the root may alterate the working of the child ACS, <P>
   * including all the not-simulated root inner properties.  <P>
   * Only after a &lt;AssociateClass&gt; first value, and <P>
   * a 'Name.&lt;Value&gt;.&lt;ACSName&gt;+' second value <P>
   * with the parent ACS name, and only for a child ACS integrity criteria. <P> <P>
   * Controls in this method are on the presence of the class and the constructor,
   * or the presence of the method, the number of arguments,
   * but no control on arguments for the syntax, the types or the values.
   * Meanwhile, it checks up for &lt;AssociateClass&gt; the argument names (not the values)
   * from the BeanInfo descriptor of these names,
   * returned by the value 'constrArgumentPropertyNames'.
   * The returned message 'Error: lack of arguments' is the only error message
   * used by AssociationMapPropertyEditor.
   * Caution: any change in this method has to be also done in
   * Gui1.AssociationMapPropertyEditor.addRestrictedValueElements().
   * Gui1.ActionMapPropertyEditor.addValueElements() uses the control message.
   * Calls getL_MethodsForKeyOfAssociations().
   * Called by ACSFactoryImpl.checkACSbeforeCreation(), add/removeAssociations().
   * @see #getL_MapKeyStartsForAssociationsOnNewObject
   * @see #getL_MapKeyStartsForAllRoots
   * @param _factory caller of this method
   * @param _m_Associations actions to do. May be null or empty.
   * @param _masterKey the key in the associations map. It must start with
   * 'ACS.Name.&lt;Value&gt;.' for the child ACS roots.
   * @param _l_values is a String array which contains the values for the criteria.
   *   Never null but the lenght may be 0.
   * @param _withNumber true if all the items in _l_values have to start with 'N.',
   *   and false otherwise.
   * @return null if the checking rules are fulfilled, or
   *   a message which explains the first integrity error, and starts with 'Error: '
   *   or  'Error to delete:'.
   */
  public static final String controlAssociations(ACSFactoryImpl _factory,
          HashMap<String, String[]> _m_Associations,
          String _masterKey, String[] _l_values, boolean _withNumber) {
      // variables
      boolean           found = false;
      int               index = -1;
      int               actionIndex = -1;
      int               numberShift = -1;
      int               methIndex = -1;
      Method[]          l_meth = null;
      Map               m_asso = null;
      String[]          l_val = null;
      String            className = null;
      BeanInfoPattern   newBaseObjectBeanInfo = null;
      String            beanName =  null;
      Class             oneclass = null;
      BeanDescriptor    beanDescriptor = null;
      Class[]           l_class = null;
      Name              name = null;
      String            stringName =  null;
      BaseObject        obj = null;
      boolean           forChildACSroots = false;
      boolean           generalCriterion = false;
      ACS               acsRoots = null;


      if (_masterKey == null || _l_values == null ||
              _l_values.length == 0 || _l_values[0] == null || _l_values[0].length() < 20)
          return "Error: null key/master acs/values/first value, or too short first value";

      if (_withNumber && _l_values[0].substring(3, 4).equals(".")) numberShift = 4;
      else if (_withNumber && _l_values[0].substring(2, 3).equals(".")) numberShift = 3;
      else if (_withNumber) numberShift = 2;

      if (_masterKey.startsWith("ACS.Name.<Value>.")) {
          try {
              NameImpl.getNameFromString(" "+_masterKey.substring(17)+"  ");
          }
          catch (BaseError b) {
              return "Error: the key is not ended by a parent ACS name - "+b;
          }
          forChildACSroots = true;
      }

      if (forChildACSroots) {
          try {
              name = NameImpl.getNameFromString(" "+_masterKey.substring(17)+"  ");
          }
          catch (BaseError b) {
              return "Error to delete: the key does not contain an ACS name";
          }
          acsRoots = (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(name);
      } // if (forChildACSroots)

                        // for testing and tracking
                        //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          //      "numberShift = "+numberShift+", and _l_values:");
                        //for (int i = 0; i < _l_values.length; i++) {
                          //  System.out.println("-------- "+_l_values[i]);
                        //} // for i

      // checks the first value
      if (!_withNumber && !_l_values[0].startsWith("<AssociateClass>.") &&
              !_l_values[0].startsWith("<CalledMethod>.<Master>.") &&
              !_l_values[0].startsWith("<CalledMethod>.<ACS>.") &&
              !_l_values[0].startsWith("<CalledMethod>.<Action>.") &&
              !_l_values[0].startsWith("<CalledMethod>.<Object>."))
          return "Error to delete: forbidden first value "+_l_values[0];
      if (_withNumber && !_l_values[0].substring(numberShift).startsWith("<AssociateClass>.") &&
              !_l_values[0].substring(numberShift).startsWith("<CalledMethod>.<Master>.") &&
              !_l_values[0].substring(numberShift).startsWith("<CalledMethod>.<ACS>.") &&
              !_l_values[0].substring(numberShift).startsWith("<CalledMethod>.<Action>.") &&
              !_l_values[0].substring(numberShift).startsWith("<CalledMethod>.<Object>."))
          return "Error to delete: forbidden first value "+_l_values[0];

      /*
      // checks the first value <AssociateClass>
      if (!_withNumber && _l_values[0].startsWith("<AssociateClass>.") &&
              !_l_values[0].substring(17).startsWith("Resource") &&
              !_l_values[0].substring(17).startsWith("Directory") &&
              !_l_values[0].substring(17).startsWith("Actor") &&
              !_l_values[0].substring(17).startsWith("EligibleParty") &&
              !_l_values[0].substring(17).startsWith("UserID") &&
              !_l_values[0].substring(17).startsWith("GroupID") &&
              !_l_values[0].substring(17).startsWith("AclEntry") &&
              !_l_values[0].substring(17).startsWith("VirtualFolder") &&
              !_l_values[0].substring(17).startsWith("PrivilegeForType") &&
              !_l_values[0].substring(17).startsWith("PrivilegeForLinks"))
          return "Error to delete: forbidden associate class at 0";
      if (_withNumber && _l_values[0].substring(numberShift).startsWith("<AssociateClass>.") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("Resource") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("Directory") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("Actor") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("EligibleParty") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("UserID") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("GroupID") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("AclEntry") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("VirtualFolder") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("PrivilegeForType") &&
              !_l_values[0].substring(numberShift).substring(17).startsWith("PrivilegeForLinks"))
          return "Error to delete: forbidden associate class at 0";
       */

                              // for testing purpose
                        //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          //"end of checkings for <AssociateClass>, and _l_values[0]: "+_l_values[0]);

      // checks the first value <CalledMethod>.<Master>.
      if ((!_withNumber && _l_values[0].startsWith("<CalledMethod>.<Master>.")) ||
              (_withNumber && _l_values[0].substring(numberShift).startsWith(
                "<CalledMethod>.<Master>."))) {

          if (forChildACSroots) return "Error to delete: Master is forbidden to child ACS roots";

          l_meth = getL_MethodsForKeyOfAssociations(_factory, _masterKey);
          if (!_withNumber) index = 24;
          else index = 24 + numberShift;

          for (int i = 0; i < l_meth.length; i++) {
              if (l_meth[i].getName().contains(_l_values[0].substring(index))) {
                  found = true;
                  methIndex = i;
                  break;
              }
          } // for i
          if (!found) {
              System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          "Error to delete: master method not found - l_meth:");
              for (int i = 0; i < l_meth.length; i++) {
                  System.out.println("----- i: "+l_meth[i].getName());
              } // for i
              System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          "Error to delete: master method not found - _l_values[0].substring(index): "+
                          _l_values[0].substring(index));

              return "Error to delete: master method not found";
          } // if (!found)
      } // if

      // checks the first value <CalledMethod>.<ACS>., only for
      // the creation of new objects, to allow the calling of
      // 'handleAssociationsForInnerObject'
      if ((!_withNumber && _l_values[0].startsWith("<CalledMethod>.<ACS>.")) ||
              (_withNumber && _l_values[0].substring(numberShift).startsWith("<CalledMethod>.<ACS>."))) {

          if (!_withNumber && !_l_values[0].substring(21).equals("handleAssociationsForInnerObject") &&
                  !forChildACSroots)
            return "Error to delete: forbidden ACS method in the master ACS";
          if (_withNumber && !_l_values[0].substring(21 + numberShift).equals("handleAssociationsForInnerObject") &&
                  !forChildACSroots )
            return "Error to delete: forbidden ACS method in the master ACS";

          try {
              if (!forChildACSroots)
                  l_meth = Class.forName(
                                UtilityImpl.getFirstInstance().getEffectiveClassName(
                                    _factory.m_AcsVocabulary_,
                                    ACSFactoryUtilityImpl.ACS_CLASS)
                            ).getMethods();
              else
                  l_meth = acsRoots.getClass().getMethods();
          }
          catch (ClassNotFoundException c) {
                throw new InternalError("From AssociateUtilityImpl.controlAssociations(): "+
                        c.toString());
          }

          if (!_withNumber) index = 21;
          else index = 21 + numberShift;
          for (int i = 0; i < l_meth.length; i++) {
              if (l_meth[i].getName().contains(_l_values[0].substring(index))) {
                  found = true;
                  methIndex = i;
                  break;
              }
          } // for i
          if (!found) {
              System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                      "Error to delete: this ACS method not found - _l_values[0].substring(index): "+
                      _l_values[0].substring(index));

              return "Error to delete: ACS method not found";
          } // if (!found)

                              // for testing purpose
                        //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          //"end of checkings for <CalledMethod>.<ACS>., and _l_values[0]: "+_l_values[0]);
      } // if

      // checks the first value <CalledMethod>.<Action>.C.ZZZ
      // where the action C contains '<AssociateClass>'
      if ((!_withNumber && _l_values[0].startsWith("<CalledMethod>.<Action>.")) ||
        (_withNumber && _l_values[0].substring(numberShift).startsWith("<CalledMethod>.<Action>.")) ) {

          m_asso = getM_FilteredAssociations(_m_Associations, true);
          if (m_asso.isEmpty()) {
              m_asso = null;
              return "Error to delete: no called method since empty map";
          }
          l_val = (String[])m_asso.get(_masterKey);
          if (l_val == null || l_val.length == 0 )
              return "Error to delete: no method to call, no previous action for this key";
          found = false;
          // gets the first character of the action number
          actionIndex = _l_values[0].indexOf("<CalledMethod>.<Action>.")+24;
          if (_l_values[0].length() <= actionIndex+2 ||
                      _l_values[0].substring(actionIndex).indexOf(".") <= 0 )
              return "Error to delete: the called method is not well defined";

          for (int i = 0; i < l_val.length; i++) {
              if (l_val[i].indexOf("<AssociateClass>.") != -1 &&
                      l_val[i].startsWith( _l_values[0].substring(
                        actionIndex,
                        actionIndex + _l_values[0].substring(actionIndex).indexOf(".") )) ) {
                  found = true;
                  index = i; // i is on the first value of the matching set
                  break;
              } // if
          } // for i
          if (!found) {
              m_asso = null;
              l_val = null;
              return "Error to delete: no called method since false number at "+
                      _l_values[0].substring(
                      actionIndex,
                      actionIndex + _l_values[0].substring(actionIndex).indexOf("."));
          }
          
                                  // for testing purpose
                            //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                              //      "for <CalledMethod>.<Action>.C.ZZZ, _l_values[0]: "+
                                //    _l_values[0]+" and found index C: "+index+" gives l_val[index]: "+
                                  //  l_val[index]);

          // no test of _withNumber since it is always a numbered action
          if (!forChildACSroots)
              l_meth = getL_MethodsForKeyOfAssociations(_factory, l_val[index].substring(
                  l_val[index].indexOf("<AssociateClass>.")+17 ));
          else
              l_meth = getL_MethodsForKeyOfAssociations(acsRoots, l_val[index].substring(
                  l_val[index].indexOf("<AssociateClass>.")+17 ));

          if (l_meth != null) { // null if it is not a '<AssociateClass>' case
              found = false;
              for (int i = 0; i < l_meth.length; i++) {
                  if (l_meth[i].getName().contains(_l_values[0].substring(
                          actionIndex + _l_values[0].substring(actionIndex).indexOf(".") + 1 ))) {
                      found = true;
                      methIndex = i;
                      break;
                  }
              } // for i
              if (!found) {// '.substring(24)' to display 'C.'

                          // for testing
                          //System.out.println("Error to delete: no called method in "+
                            //      l_val[index].substring(l_val[index].indexOf("<AssociateClass>.")+17 ));
                          //System.out.println("--- do not found "+
                            //      _l_values[0].substring(actionIndex + _l_values[0].substring(actionIndex).indexOf(".")+1)+
                              //    " in the methods:");
                          //for (int i = 0; i < l_meth.length; i++) {
                            //  System.out.println("------- "+l_meth[i].getName());
                          //}

                return "Error to delete: no called method "+_l_values[0]+
                      " in "+l_val[index].substring(l_val[index].indexOf("<AssociateClass>.")+17 );
              } // if
          } // if (l_meth != null
          else {
                                   // for testing purpose
                            System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                                    "for <CalledMethod>.<Action>.C.ZZZ, _l_values[0]: "+
                                    _l_values[0]+" and found index C: "+index+" gives l_val[index]: "+
                                    l_val[index]+"; but with l_meth null.");
          }
               // else
      } // if (_l_values[0].startsWith("<CalledMethod>.<Action>."))

      // checks the first value <CalledMethod>.<Object>.XXX+<Class>+YYY+<Value>+ZZZ
      if ((!_withNumber && _l_values[0].startsWith("<CalledMethod>.<Object>.")) ||
        (_withNumber && _l_values[0].substring(numberShift).startsWith("<CalledMethod>.<Object>.")) ) {

          if (forChildACSroots) return "Error to delete: Object is forbidden to child ACS roots";

         if (_l_values[0].indexOf("+<Class>+") == -1 ||
                 _l_values[0].indexOf("+<Value>+") == -1 )
             return "Error to delete: no Class or no Value in the action";

         name = null;
         stringName = _l_values[0].substring(
                            _l_values[0].indexOf(".<Object>.") + 10,
                            _l_values[0].indexOf("+<Class>+"));

         try {
             if (stringName != null && stringName.endsWith("::  ") &&
                     stringName.startsWith(" "))
                 name = NameImpl.getNameFromString( stringName );
             else return "Error to delete: no object name or <Value> in "+stringName;
         } catch (BaseError b) {
             return "Error to delete: no object name in "+stringName;
         }

         try {
             // find the AcsAddon class name, if necessary
             oneclass = Class.forName(
                     UtilityImpl.getFirstInstance().getEffectiveClassName(
                        _factory.m_AcsVocabulary_,
                        _l_values[0].substring(
                            _l_values[0].indexOf("+<Class>+") + 9,
                            _l_values[0].indexOf("+<Value>+"))));
             if (oneclass == null)
                 return "Error to delete: object class not found in "+
                         _l_values[0].substring(
                     _l_values[0].indexOf("+<Class>+") + 9,
                     _l_values[0].indexOf("+<Value>+"));
         } catch (ClassNotFoundException b) {
             return "Error to delete: object class not found in "+
                     _l_values[0].substring(
                     _l_values[0].indexOf("+<Class>+") + 9,
                     _l_values[0].indexOf("+<Value>+"));
         }

         l_meth = oneclass.getMethods();

         if (_l_values[0].indexOf("+<Value>+") + 9 > _l_values[0].length())
             return "Error to delete: method not found in "+
                     _l_values[0].substring(_l_values[0].indexOf("+<Value>+"));

          found = false;
          index = _l_values[0].indexOf("+<Value>+") + 9;
          for (int i = 0; i < l_meth.length; i++) {
              if (l_meth[i].getName().equals(_l_values[0].substring(index))) {
                  found = true;
                  methIndex = i;
                  break;
              }
          } // for i
          if (!found) {

                      // for testing
                      //System.out.println("Error to delete: no called method - methods:");
                      //for (int i = 0; i < l_meth.length; i++) {
                        //  System.out.println("----- "+l_meth[i]);
                      //}

              return "Error to delete: method not found in "+_l_values[0];
          } // if

      } // if (_l_values[0].startsWith("<CalledMethod>.<Object>."))

      //
      // end of checkings for the first value
      //

                              // for testing purpose
                        //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          //"end of checkings for the first value, and _l_values[0]: "+_l_values[0]);

       // <GeneralCriterion>.<RootIntegrity>:
        // Only after a <AssociateClass> first value 
        // and a 'Name.<Value>.<ACSName>+' second value,
        // with the parent ACS name, and only for a child ACS integrity criteria.
      found = true;
      if ((!_withNumber && _l_values[0].startsWith("<AssociateClass>.")) ||
              (_withNumber && _l_values[0].substring(numberShift).startsWith("<AssociateClass>."))) {
          for (int i = 1; i < _l_values.length; i++) {
              
              if (_l_values[i].indexOf("<GeneralCriterion>") != -1) {
                  generalCriterion = true;
                  
                  if (!forChildACSroots || _l_values.length != 3 || i != 2 ||
                      !_l_values[1].startsWith("Name.<Value>.<ACSName>+")) {
                              found = false;
                              break;
                  } // if
              } // if
              if (generalCriterion && !found) return "Error to delete: GeneralCriterion "+
                      "for a master, or with a wrong form";
          } // for i
      } // if
      if ((!_withNumber && _l_values[0].startsWith("<CalledMethod>.")) ||
        (_withNumber && _l_values[0].substring(numberShift).startsWith("<CalledMethod>.")) ) {
          for (int i = 1; i < _l_values.length; i++) {
              if (_l_values[i].indexOf("<GeneralCriterion>") != -1) {
                  found = false;
                  break;
              }
              if (!found) return "Error: GeneralCriterion is reserved to <AssociateClass>";
          } // for i
      } // if

      // checks the arguments names, but not the values, for <AssociateClass>
      if (!generalCriterion &&
             ((!_withNumber && _l_values[0].startsWith("<AssociateClass>.")) ||
              (_withNumber && _l_values[0].substring(numberShift).startsWith("<AssociateClass>.")))) {

        // initialization from the property BeanInfo
        if (!_withNumber) {
           if (!forChildACSroots)
                className = UtilityImpl.getFirstInstance().getEffectiveClassName(
                        _factory.m_AcsVocabulary_,
                        _l_values[0].substring(17));
           else
                className = UtilityImpl.getFirstInstance().getEffectiveClassName(
                        acsRoots.getM_AcsVocabulary(),
                        _l_values[0].substring(17));
        }
        else {
           if (!forChildACSroots)
                className = UtilityImpl.getFirstInstance().getEffectiveClassName(
                        _factory.m_AcsVocabulary_,
                        _l_values[0].substring(17+numberShift));
           else 
                className = UtilityImpl.getFirstInstance().getEffectiveClassName(
                        acsRoots.getM_AcsVocabulary(),
                        _l_values[0].substring(17+numberShift));
        } // else

        if (className == null) {
              l_meth = null;
              m_asso = null;
              l_val = null;
              className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;
              if (!_withNumber)
                  return "Error to delete: <AssociateClass> not found in "+
                          _l_values[0].substring(17);
              else
                  return "Error to delete: <AssociateClass> not found in "+
                          _l_values[0].substring(17+numberShift);
        } // if

        newBaseObjectBeanInfo = null;
        try {
                         //if (forChildACSroots) {
                           //  System.out.println("AssociateUtilityImpl.controlAssociations() for the "+
                             //        "child ACS roots: "+acsRoots.getNickName());
                             //for (Object ob : acsRoots.getM_AcsVocabulary().entrySet()) {
                               //  Map.Entry entry = (Map.Entry)ob;
                                // System.out.println("---------- key: "+entry.getKey()+
                                  //       "-- value at 0: "+ ((String[])entry.getValue())[0]);
                             //} // for ob
                         //} // if


                                  // for testing purpose
                          //if (!_withNumber)
                            //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                             // "for <AssociateClass>, className:|"+className+"| for"+
                              //" _l_values[0].substring(17): "+_l_values[0].substring(17));
                          //else
                            //  System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                             // "for <AssociateClass>, className:|"+className+"| for"+
                              //" _l_values[0].substring(17+numberShift): "+
                                //      _l_values[0].substring(17+numberShift));

            beanName =  UtilityImpl.getFirstInstance().
                            getBeanInfoNameFromClass(className);

                                  // for testing purpose
                            //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                              //"for <AssociateClass>, beanName:|"+beanName+"|");

            oneclass = Class.forName(beanName);
        } catch (ExceptionInInitializerError e) {
            throw new InternalError("AssociateUtilityImpl.controlAssociations(): "
              +e.toString()
              );
        } catch (LinkageError l) {
            throw new InternalError("AssociateUtilityImpl.controlAssociations(): "
              +l.toString()
              );
        } catch (ClassNotFoundException c) {
            throw new InternalError("AssociateUtilityImpl.controlAssociations(): "
              +c.toString()
              );
        } // initialization of the property BeanInfo

        try {
            newBaseObjectBeanInfo = (BeanInfoPattern) oneclass.newInstance();
            beanName = null;
            oneclass = null;
        } catch (Exception c) {
            throw new InternalError("AssociateUtilityImpl.controlAssociations() "+
              "on oneclass.newInstance(): "+c.toString());
        } // initialization of the property BeanInfo

          // initializes from the BeanInfo
        beanDescriptor = newBaseObjectBeanInfo.getBeanDescriptor();
        l_val = (String[]) beanDescriptor.getValue("constrArgumentPropertyNames");

                              // for testing purpose
                        //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          //"check the arguments for <AssociateClass>, l_val: ");
                        //for (int i = 0; i < l_val.length; i++) {
                          //  System.out.println("------ at "+i+": "+l_val[i]);
                        //} // for i

        if (l_val == null) {
              l_meth = null;
              m_asso = null;
              l_val = null;
              //className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;
              return "Error to delete: no description in the BeanInfo for "+
                      className;
        }
        if (_l_values.length < l_val.length + 1) {
              l_meth = null;
              m_asso = null;
              //className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;
              l_val = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;
              return "Error: lack of arguments in the BeanInfo for "+
                      className;
        } // if
        else if (_l_values.length > l_val.length + 1) {
              l_meth = null;
              m_asso = null;
              l_val = null;
              //className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;
              return "Error to delete: too much arguments in the BeanInfo for "+
                      className;
        } // if

        found = true;
        for (int i = 0; i < l_val.length; i++) {
              if ((!_withNumber && !_l_values[i+1].startsWith( l_val[i])) ||
                      (_withNumber && !_l_values[i+1].substring(numberShift).startsWith( l_val[i])) ) {
                  found = false;
                  index = i;
                  break;
              } // if
        } // for i
        if (!found) {
              l_meth = null;
              m_asso = null;
              //className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;

                              // for testing purpose
                        System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          "check the arguments for <AssociateClass>, "+
                          "the arguments array has a false argument name at index = "+index+
                          ", and _withNumber = "+_withNumber);
                        System.out.println("   where the argument "+_l_values[index+1]+" does not start with"+
                          " the BeanInfo element at index: ");
                        for (int i = 0; i < l_val.length; i++) {
                            System.out.println("------ at "+i+": "+l_val[i]);
                        } // for

              l_val = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;
              return "Error to delete: false argument name in the BeanInfo for "+
                      className+" at the index "+index+1+": "+_l_values[index+1];
        } // if (!found)
      } // if (_l_values[0].startsWith("<AssociateClass>.")

      // checks the arguments number only, for all the <CalledMethod> actions;
      // then, l_meth and methIndex have been initialized
      else if (_l_values[0].indexOf("<CalledMethod>.") != -1) {

          l_class = l_meth[methIndex].getParameterTypes();

          if (l_class == null && _l_values.length != 1) {
              l_meth = null;
              m_asso = null;
              l_val = null;
              className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;
              return "Error to delete: too much arguments";
          }
          else if (_l_values.length < l_class.length + 1) {
              l_meth = null;
              m_asso = null;
              l_val = null;
              className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;

                              // for testing purpose
                        System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          "check the arguments for <CalledMethod>.<Master> or <Action>, "+
                          "where the error is '_l_values.length < l_class.length + 1'");
                        System.out.println("--- with l_values: ");
                        for (int i = 0; i < _l_values.length; i++) {
                            System.out.println("------ at "+i+": "+_l_values[i]);
                        } // for
                        System.out.println("--- with l_class: ");
                        for (int i = 0; i < l_class.length; i++) {
                            System.out.println("------ at "+i+": "+l_class[i]);
                        } // for

              return "Error: lack of arguments";
          } // if
          else if (_l_values.length > l_class.length + 1) {
              l_meth = null;
              m_asso = null;
              l_val = null;
              className = null;
              newBaseObjectBeanInfo = null;
              beanName =  null;
              oneclass = null;
              beanDescriptor = null;
              name = null;
              stringName =  null;
              obj = null;
              acsRoots = null;
              return "Error to delete: too much arguments";
          } // else if

      } // else

      // checking for <CalledMethod>.<Object>.XXX+<Value>+ZZZ, where
      // the argument types must contain <Master> or <AssociateValue>
      if ((!_withNumber && _l_values[0].startsWith("<CalledMethod>.<Object>.")) ||
        (_withNumber && _l_values[0].substring(numberShift).startsWith("<CalledMethod>.<Object>.")) ) {

          if ((!_withNumber && _l_values[0].startsWith("<CalledMethod>.<Object>.")) ||
                (_withNumber && _l_values[0].substring(numberShift).startsWith(
                    "<CalledMethod>.<Object>.")) ) {
              found = false;
              if (_l_values.length != 1) {
                  for (int i = 1; i < _l_values.length; i++) {
                    if (_l_values[i].contains("<Master>") || _l_values[i].contains(
                                "<AssociateValue>")) {
                        found = true;
                        break;
                    }
                  } // for i
              } // if
              if (!found) return "Error: no Master nor AssociateValue argument";
          } // if
      } // if


                              // for testing purpose
                        //System.out.println("AssociateUtilityImpl.controlAssociations(), "+
                          //"end of method with _l_values[0]: "+_l_values[0]);

      l_meth = null;
      m_asso = null;
      l_val = null;
      l_class = null;
      className = null;
      newBaseObjectBeanInfo = null;
      beanName =  null;
      oneclass = null;
      beanDescriptor = null;
      name = null;
      stringName =  null;
      obj = null;
      acsRoots = null;
      return null;
  } // controlAssociations()

  /**
   * Gets the methods which are applicable for a given criteria in the map of
   * associations for objects. Used also for the child ACS roots key.
   * The master key is a map key among the followings: <P>
   * - 'Class'.'Property'.&lt;Value&gt;.XXX: a new object property has the value XXX; <P>
   * - 'Class'.'Property'.&lt;ValueContains&gt;.XXX: a new object property value contains the value XXX; <P>
   * - 'Class'.'Property'.&lt;Class&gt;.XXX: a new object property has exactly the class XXX; <P>
   * - 'Class'.'Property'.&lt;NameStartsWith&gt;.XXX: the new object property name
   *        starts with the ACS name plus the subname XXX; <P>
   * - 'Class'.&lt;NameStartsWith&gt;.XXX: the new object name of the class
   *        starts with the ACS name plus the subname XXX; <P>
   * - &lt;Class&gt;.&lt;Value&gt;.XXX: the new object class is exactly ARoad0.gBase.XXX; <P>
   * - &lt;Class&gt;.&lt;ValueImplements&gt;.XXX: the new object class is a subclass of ARoad0.gBase.XXX; <P>
   * - for a child ACS roots key, it is a classic class or an extension class from
   * an AcsAddon. <P>
   * Called by controlAssociations().
   * @see #getL_MapKeyStartsForAssociationsOnNewObject
   * @see ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ACSFactoryUtilityImpl#ACSADDON_EXTENSION_CLASSES
   * @param _factory for which associations are set, to find the AcsAddon class if necessary;
   * it is the root ACS for the root associations
   * @param _masterKey the key in the associations map, the first value of a numbered action,
   * a classic class or an extension class from an ACSAddon.
   * @return array of the applicable methods for the master, or null.
   */
  public static final Method[] getL_MethodsForKeyOfAssociations(ACSRun _factory,
          String _masterKey) {
      Method[] l_result = null;

                              // for testing purpose
                        //System.out.println("AssociateUtilityImpl.getL_MethodsForKeyOfAssociations(), "+
                          //"for the argument: "+_masterKey);

      try {
          if (_masterKey.startsWith("Resource") ||
                  (_masterKey.startsWith("<Class>") && _masterKey.endsWith("Resource")))
              l_result = Class.forName(
                            UtilityImpl.getFirstInstance().getEffectiveClassName(
                                _factory.getM_AcsVocabulary(),
                                ACSFactoryUtilityImpl.RESOURCE_CLASS)
                        ).getMethods();

          else if (_masterKey.startsWith("Actor") ||
                  (_masterKey.startsWith("<Class>") && _masterKey.endsWith("Actor")))
              l_result = Class.forName(
                             UtilityImpl.getFirstInstance().getEffectiveClassName(
                                _factory.getM_AcsVocabulary(),
                                ACSFactoryUtilityImpl.ACTOR_CLASS)
                        ).getMethods();

          else if (_masterKey.startsWith("EligibleParty") ||
                  (_masterKey.startsWith("<Class>") && _masterKey.endsWith("EligibleParty")))
              l_result = Class.forName(
                             UtilityImpl.getFirstInstance().getEffectiveClassName(
                                _factory.getM_AcsVocabulary(),
                                ACSFactoryUtilityImpl.ELIGIBLEPARTY_CLASS)
                        ).getMethods();

          else if (_masterKey.startsWith("Directory") ||
                  (_masterKey.startsWith("<Class>") && _masterKey.endsWith("Directory")))
              l_result = Class.forName(
                             UtilityImpl.getFirstInstance().getEffectiveClassName(
                                _factory.getM_AcsVocabulary(),
                                ACSFactoryUtilityImpl.DIRECTORY_CLASS)
                        ).getMethods();

          else if (_masterKey.startsWith("GroupID") ||
                  (_masterKey.startsWith("<Class>") && _masterKey.endsWith("GroupID")))
              l_result = Class.forName(
                             UtilityImpl.getFirstInstance().getEffectiveClassName(
                                _factory.getM_AcsVocabulary(),
                                ACSFactoryUtilityImpl.GROUP_CLASS)
                        ).getMethods();

          else if (_masterKey.startsWith("UserID") ||
                  (_masterKey.startsWith("<Class>") && _masterKey.endsWith("UserID")))
              l_result = Class.forName(
                             UtilityImpl.getFirstInstance().getEffectiveClassName(
                                _factory.getM_AcsVocabulary(),
                                ACSFactoryUtilityImpl.USER_CLASS)
                        ).getMethods();

          else if (_masterKey.startsWith("VirtualFolder") ||
                  (_masterKey.startsWith("<Class>") && _masterKey.endsWith("VirtualFolder")))
              l_result = Class.forName(
                             UtilityImpl.getFirstInstance().getEffectiveClassName(
                                _factory.getM_AcsVocabulary(),
                                ACSFactoryUtilityImpl.VIRTUALFOLDER_CLASS)
                        ).getMethods();
          else {
              String[] l_extension_classes = (String[])_factory.getM_AcsVocabulary().get(
                  ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES);

              if (l_extension_classes != null) {
                  for (int i = 0; i < l_extension_classes.length; i++) {
                      
                      if (l_extension_classes[i].indexOf(_masterKey) != -1) {
                          l_result = Class.forName(
                                        UtilityImpl.getFirstInstance().getEffectiveClassName(
                                            _factory.getM_AcsVocabulary(),
                                            _masterKey)
                                     ).getMethods();
                          break;
                      } // if
                  } // for i
              } // if

              l_extension_classes = null;
          } // else

      } catch (ClassNotFoundException c) {
          throw new InternalError("AssociateUtilityImpl.getL_MethodsForKeyForAssociations()"+
                  " checks the first value: "+c.toString());
      } // catch

      return l_result;
  } // getL_MethodsForKeyOfAssociations()

  /**
   * Splits the list of numbered actions to return arrays where are put all the actions
   * having the same number. Then there is one array per number.
   * Called by ACSFactoryImpl.checkACSbeforeCreation(), add/removeAssociateActions(),
   * getL_AssociatesAndCalledMethods().
   * @param _l_values values in an association map for an acs and a given key. Not null.
   * @return each set of actions having the same number, is in an array in the list.
   *        Never null but may be empty.
   */
  public static final List<String[]> splitActionSetsForAssociations(String[] _l_values) {
     List<String[]> l_result = new ArrayList(20);
     String[]       l_oneAction = null;
     int            count = 0;
     int            endOfLastAction = -1;
     int            numberSize = -1;

                // for testing and tracking
                //System.out.println("AssociateUtilityImpl.splitActionSetsForAssociations(), _l_values:");
                //for (int i = 0; i < _l_values.length; i++) {
                  //  System.out.println("-------- "+_l_values[i]);
                //} // for i

     for (int i = 0; i < _l_values.length; i++) {
         // continue if it is in the set of the last controlled action
         if (i <= endOfLastAction) continue;
         if (!_l_values[i].substring(1, 2).equals(".") &&
                 !_l_values[i].substring(2, 3).equals(".") &&
                 !_l_values[i].substring(3, 4).equals(".")) {
             throw new InternalError("no-number command for "+_l_values[i]);
         } // if
         try {
            if ( _l_values[i].substring(1, 2).equals(".") ) {
                new Integer(_l_values[i].substring(0, 1));
                numberSize = 1;
            } else if ( _l_values[i].substring(2, 3).equals(".") ) {
                new Integer(_l_values[i].substring(0, 2));
                numberSize = 2;
            } else if ( _l_values[i].substring(3, 4).equals(".") ) {
                new Integer(_l_values[i].substring(0, 3));
                numberSize = 3;
            }
         } catch (NumberFormatException n) {
             throw new InternalError("no-number command for "+_l_values[i]);
         } // catch
         count = 1;
         while (i+count < _l_values.length &&
                 _l_values[i].substring(0, numberSize).equals(
                    _l_values[i+count].substring(0, numberSize) )) {
             count++;
         } // while
         l_oneAction = new String[count];
         System.arraycopy(_l_values, i, l_oneAction, 0, count);
         l_result.add(l_oneAction);
         endOfLastAction = i+count-1;
     } // for i

     l_oneAction = null;
     return l_result;
  } // splitActionSetsForAssociations()

  /**
   * Gets the arguments to use for a constructor in a method or an action.
   * Each useful item in the argument _l_values must belong to a set of 10 allowed forms,
   * presented herein without the number 'N.' at the start, for the action N: <P>
   * - 'PropertyName'.&lt;Null&gt;: null value for the argument having the name 'PropertyName'. <P>
   *        For any first value. For the master creating/deleting, <P>
   *        and for the child ACS root opening/deleting. <P>
   * - 'PropertyName'.&lt;Master&gt;: the master is the argument <P>
   *        For any first value. For the master creating/deleting only, <P>
   *        and for the child ACS root opening/deleting where 'Master' is the parent ACS. <P>>
   * - 'PropertyName'.&lt;MasterCreation&gt;: boolean true if the master is creating, <P>
   *        false if it is finalizing <P>
   *        For any first value. For the master creating/deleting, <P>
   *        and for the child ACS root creating/deleting. <P>
   * - 'PropertyName'.&lt;ArgumentValue&gt;.C: the associate number C object <P>
   *        For any first value. For the master creating/deleting, <P>
   *        and for the child ACS root opening/deleting. <P>
   * - 'PropertyName'.&lt;Value&gt;.XXX: the argument is XXX as String, ImmutableName <P>
   *        if it is the property type, true/false if the type is boolean or Boolean, <P>
   *        or any BaseObject get from its name <P>
   *        For any first value. For the master creating/deleting, <P>
   *        and for the child ACS root opening/deleting. <P>
   * - 'PropertyName'.&lt;Value&gt;.XXX+&lt;MasterValue&gt;+YYY: the master value <P>
   *        for the same property in its BeanInfo (the read method has to not return an array), <P>
   *        with some optional pre and post String values <P>
   *        if the property type is String <P>
   *        For any first value. For the master creating/deleting, <P>
   *        and for the child ACS root opening/deleting where 'Master' is the parent ACS. <P>>
   * - 'PropertyName'.&lt;Value&gt;.XXX+&lt;MasterNameLastComponent&gt;+YYY: <P>
   *        the last component in the master name, with some optional <P>
   *        pre and post Name values as Strings <P>
   *        For any first value. For the master creating/deleting, <P>
   *        and for the child ACS root opening/deleting where 'Master' is the parent ACS. <P>
   * - 'PropertyName'.&lt;Value&gt;.&lt;MasterNameContext&gt;+YYY <P>
   *        the context in the master name, with a post Name value as String <P>
   *        For any first value. For the master creating/deleting, <P>
   *        and for the child ACS root opening/deleting where 'Master' is the parent ACS. <P>
   * - 'PropertyName'.&lt;Value&gt;.&lt;ACSName&gt;+YYY: the ACS name, <P>
   *        with a post Name value as String <P>
   *        For any first value. For the master creating/deleting, it is <P>
   *        the master ACS name. For the child ACS root opening/deleting and
   *        the integrity criteria, it is the parent ACS name. <P>
   * - 'PropertyName'.&lt;Value&gt;.XXX+&lt;ChildNameLastComponent&gt;+YYY: <P>
   *        the last component in the child ACS name, with some optional <P>
   *        pre and post Name values as Strings <P>
   *        For any first value. Only for the child ACS root opening/deleting and
   *        the integrity criteria. <P>
   * The last item in the argument _l_values may start with '&lt;GeneralCriterion>.',
   * and it is then not returned by this method.
   * Called by add/removeAssociateActions(), getL_AssociatesAndCallesMethods().
   * @param _object just-created or just-finalized ACSObject, or parent ACS to update
   * @param _m_arguments the previous associates or intermediate arguments as values
   *        in a map, where the key is the action number of the associate or argument.
   *        Never null but may be empty.
   * @param _creation true if the master is creating, false if it is finalizing
   * @param _l_values set of actions having the same number, where the first item
   *        defines the constructor or method to call (do not processed there),
   *        and the other items give the arguments or the returned value.
   *        Never null nor empty.
   * @param _l_argumentsClass classes of the arguments for the constructor or
   *        the method to use. Never null.
   * @param _acsChild Child ACS. Not null only when _object is a parent ACS
   *        to update.
   * @return the arguments to use. An argument may be an array, for instance through
   *    '&lt;ArgumentValue&gt;'.
   * @exception BaseError if _l_values is not correct
   */
  public static final Object[] getL_ArgumentsForAssociations(BaseObject _object,
          Map _m_arguments, Boolean _creation, String[] _l_values,
          Class[] _l_argumentsClass, ACS _acsChild) throws BaseError {

      Object[]      l_result = null;
      if (_l_values[_l_values.length - 1].indexOf("<GeneralCriterion>.") == -1)
          l_result = new Object[_l_values.length - 1];
      else
          l_result = new Object[_l_values.length - 2];

      String[]      l_argumentsNames = null;
      Class         beanInfoClass = null;
      BeanInfoPattern      bean = null;
      BeanDescriptor beanDescriptor = null;
      String        val = null;
      PropertyDescriptor[] l_descriptors = null;
      Method        meth = null;
      Object        param = null;
      ImmutableName name = null;

      try {
          // from i = 1
         for (int i = 1; i < _l_values.length; i++) {

             if (_l_values[i].endsWith(".<Null>")) l_result[i-1] = null;
             else if (_l_values[i].endsWith(".<Master>")) l_result[i-1] = _object;
             else if (_l_values[i].endsWith(".<MasterCreation>"))
                 l_result[i-1] = _creation;
             else if (_l_values[i].indexOf(".<ArgumentValue>.") != -1)
                 // may be an array
                 l_result[i-1] = _m_arguments.get(
                         _l_values[i].substring(_l_values[i].indexOf(".<ArgumentValue>.")+17) );
             else if (_l_values[i].indexOf(".<Value>.") == -1)
                 throw new InternalError("AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                     " incorrect value "+_l_values[i]+" for the key "+_object.getName());
             // 'PropertyName'.<Value>.XXX+<MasterValue>+YYY
             else if (_l_values[i].indexOf("<MasterValue>") != -1 ) {

                 beanInfoClass = Class.forName(
                      UtilityImpl.getFirstInstance().getBeanInfoNameFromClass(
                        _object.getClass().getName()));
                 bean = (BeanInfoPattern)beanInfoClass.newInstance();

                 beanDescriptor = bean.getBeanDescriptor();

                 l_argumentsNames = (String[]) beanDescriptor.getValue(
                    "constrArgumentPropertyNames");
                 val = null;
                 for (int j = 0; j < l_argumentsNames.length; j++) {
                     if (_l_values[i].substring(2).startsWith( l_argumentsNames[j] )) {
                         val = l_argumentsNames[j];
                         break;
                     }
                 } // for j
                 if (val == null) {
                     throw new InternalError("AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                     " incorrect argument name in the value "+_l_values[i]
                             +" from the BeanInfo class "+beanInfoClass+
                             ", and for the master "+_object.getName());
                 } // if (val == null)

                 l_descriptors = bean.getPropertyDescriptors();
                 for (int j = 0; j < l_descriptors.length; j++) {
                     if (l_descriptors[j].getName().equals(val)) {
                         meth = l_descriptors[j].getReadMethod();
                         break;
                     }
                 } // for j
                 try {
                    param = (Object)meth.invoke(_object, (Object[])null);
                 } catch (Exception e) {
                     throw new InternalError(
                             "AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                             " incorrect call of the method "+meth
                             +" on the master "+_object.getName()+" with the exception "+e.toString());
                 }

                 if (!(param instanceof String) &&
                         (_l_values[i].indexOf("<MasterValue>") != _l_values[i].indexOf(".<Value>.") + 9 ||
                          _l_values[i].indexOf("<MasterValue>") + 13 != _l_values[i].length() ))
                     throw new InternalError("AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                            " incorrect String addings in the value "+_l_values[i]
                             +" for the master "+_object.getName());

                 if (!(param instanceof String) || 
                         (_l_values[i].indexOf(".<Value>.") + 9 >= _l_values[i].indexOf("<MasterValue>") - 1 &&
                         _l_values[i].indexOf("<MasterValue>") + 13 == _l_values[i].length()))
                     l_result[i-1] = param;

                 else if (_l_values[i].indexOf(".<Value>.") + 9 < _l_values[i].indexOf("<MasterValue>") - 1 &&
                         _l_values[i].indexOf("<MasterValue>") + 13 < _l_values[i].length()) {
                     l_result[i-1] = _l_values[i].substring(_l_values[i].indexOf(".<Value>.") + 9,
                             _l_values[i].indexOf("<MasterValue>")) +
                             (String)param +
                             _l_values[i].substring(_l_values[i].indexOf("<MasterValue>") + 13);

                 } else if (_l_values[i].indexOf(".<Value>.") + 9 < _l_values[i].indexOf("<MasterValue>") &&
                         _l_values[i].indexOf("<MasterValue>") + 13 == _l_values[i].length()) {
                     l_result[i-1] = _l_values[i].substring(_l_values[i].indexOf(".<Value>.") + 9,
                             _l_values[i].indexOf("<MasterValue>") - 1) +
                             (String)param;

                 } else if (_l_values[i].indexOf(".<Value>.") + 9 == _l_values[i].indexOf("<MasterValue>") &&
                         _l_values[i].indexOf("<MasterValue>") + 13 < _l_values[i].length()) {
                     l_result[i-1] = (String)param +
                             _l_values[i].substring(_l_values[i].indexOf("<MasterValue>") + 14);
                 } // else

             // 'PropertyName'.<Value>.XXX+<MasterNameLastComponent>+YYY
             } else if (_l_values[i].indexOf("<MasterNameLastComponent>") != -1) {

                 if ((_l_values[i].indexOf(".<Value>.") + 9 >= _l_values[i].indexOf("<MasterNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<MasterNameLastComponent>") + 25 == _l_values[i].length() ))
                     l_result[i-1] = _object.getName().lastComponent();

                 else if (_l_values[i].indexOf(".<Value>.") + 9 < _l_values[i].indexOf("<MasterNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<MasterNameLastComponent>") + 25 < _l_values[i].length() )
                     l_result[i-1] = _l_values[i].substring(_l_values[i].indexOf(".<Value>.") + 9 ,
                             _l_values[i].indexOf("<MasterNameLastComponent>") - 1) +
                             _object.getName().lastComponent() +
                             _l_values[i].substring(_l_values[i].indexOf("<MasterNameLastComponent>") + 26);

                 else if (_l_values[i].indexOf(".<Value>.") + 9 < _l_values[i].indexOf("<MasterNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<MasterNameLastComponent>") + 25 >= _l_values[i].length() )
                     l_result[i-1] = _l_values[i].substring(_l_values[i].indexOf(".<Value>.") + 9 ,
                             _l_values[i].indexOf("<MasterNameLastComponent>") - 1) +
                             _object.getName().lastComponent();

                 else if (_l_values[i].indexOf(".<Value>.") + 9 >= _l_values[i].indexOf("<MasterNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<MasterNameLastComponent>") + 25 < _l_values[i].length() )
                     l_result[i-1] =  _object.getName().lastComponent() +
                             _l_values[i].substring(_l_values[i].indexOf("<MasterNameLastComponent>") + 26);

             // 'PropertyName'.<Value>.XXX+<ChildNameLastComponent>+YYY 
             } else if (_l_values[i].indexOf("<ChildNameLastComponent>") != -1) {

                 if ((_l_values[i].indexOf(".<Value>.") + 9 >= _l_values[i].indexOf("<ChildNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<ChildNameLastComponent>") + 24 == _l_values[i].length() ))
                     l_result[i-1] = _acsChild.getName().lastComponent();

                 else if (_l_values[i].indexOf(".<Value>.") + 9 < _l_values[i].indexOf("<ChildNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<ChildNameLastComponent>") + 24 < _l_values[i].length() )
                     l_result[i-1] = _l_values[i].substring(_l_values[i].indexOf(".<Value>.") + 9 ,
                             _l_values[i].indexOf("<ChildNameLastComponent>") - 1) +
                             _acsChild.getName().lastComponent() +
                             _l_values[i].substring(_l_values[i].indexOf("<ChildNameLastComponent>") + 24);

                 else if (_l_values[i].indexOf(".<Value>.") + 9 < _l_values[i].indexOf("<ChildNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<ChildNameLastComponent>") + 24 >= _l_values[i].length() )
                     l_result[i-1] = _l_values[i].substring(_l_values[i].indexOf(".<Value>.") + 9 ,
                             _l_values[i].indexOf("<ChildNameLastComponent>") - 1) +
                             _acsChild.getName().lastComponent();

                 else if (_l_values[i].indexOf(".<Value>.") + 9 >= _l_values[i].indexOf("<ChildNameLastComponent>") - 1 &&
                          _l_values[i].indexOf("<ChildNameLastComponent>") + 24 < _l_values[i].length() )
                     l_result[i-1] =  _acsChild.getName().lastComponent() +
                             _l_values[i].substring(_l_values[i].indexOf("<ChildNameLastComponent>") + 24);

             // 'PropertyName'.<Value>.<MasterNameContext>+YYY
             } else if (_l_values[i].indexOf(".<MasterNameContext>") != -1) {

                 if (_l_values[i].indexOf("<MasterNameContext>") + 19 >= _l_values[i].length() )
                     l_result[i-1] = _object.getName().context().toString();

                 else l_result[i-1] =  _object.getName().context().toString() +
                             _l_values[i].substring(_l_values[i].indexOf("<MasterNameContext>") + 19);

             // 'PropertyName'.<Value>.<ACSName>+YYY where the optional YYY is
                 // the String form of an ImmutableName (then starting by ' ')
                 // or any BaseObject get from its name
             } else if (_l_values[i].indexOf(".<ACSName>+") != -1) {

                 if (_l_values[i].indexOf("<ACSName>+") + 10 >= _l_values[i].length() ) {
                     if (_object instanceof ACSObject) name = ((ACSObject)_object).getAcsName();
                     else if (_object instanceof ACS) name = ((ACS)_object).getName();

                 } else {
                     if (_object instanceof ACSObject)
                         name = ((NameImpl)((ACSObject)_object).getAcsName()).
                            extended(NameImpl.getNameFromString(
                                _l_values[i].substring(_l_values[i].indexOf("<ACSName>+") + 10)));
                     else if (_object instanceof ACS)
                         name = ((NameImpl) ((ACS)_object).getName()).
                            extended(NameImpl.getNameFromString(
                                _l_values[i].substring(_l_values[i].indexOf("<ACSName>+") + 10)));
                 } // else

                 if (name != null &&
                         ImmutableName.class.isAssignableFrom( _l_argumentsClass[i - 1]))
                     l_result[i-1] = name;
                 else if (name != null &&
                         BaseObject.class.isAssignableFrom( _l_argumentsClass[i - 1]))
                     l_result[i-1] = BaseUtilityImpl.getEorBaseObject(name);

                 if (l_result[i-1] == null)
                     throw new InternalError(
                         "AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                         " incorrect parsing for the parameter "+_l_values[i]
                         +" on the master "+_object.getName());

             // 'PropertyName'.<Value>.XXX where the property is a String (never null),
                 // an ImmutableName, true/false if the type is boolean or Boolean,
                 // a number if the type is int or Integer,
                 // or any BaseObject get from its name (never null)
             } else {


                                      // for testing purpose
                                //System.out.println("AssociateUtilityImpl."+
                                  //      "getL_ActionArgumentsForAssociations(), for"+
                                    //    "'PropertyName'.<Value>.XXX with the argument class :"+
                                      //  _l_argumentsClass[i - 1]+" for the action "+
                                        //_l_values[0]+" with the native argument value "+
                                        //_l_values[i]);

                 if (String.class.isAssignableFrom( _l_argumentsClass[i - 1])) {
                     l_result[i-1] = _l_values[i].substring(
                         _l_values[i].indexOf(".<Value>.") + 9 );
                     // null if empty
                     //if (((String)l_result[i-1]).length() == 0)
                       // l_result[i-1] = null;
                 } else if (Integer.class.isAssignableFrom( _l_argumentsClass[i - 1])) {
                     l_result[i-1] = new Integer( _l_values[i].substring(
                         _l_values[i].indexOf(".<Value>.") + 9 ));
                 } else if (Integer.TYPE.isAssignableFrom( _l_argumentsClass[i - 1])) {
                     l_result[i-1] = new Integer( _l_values[i].substring(
                         _l_values[i].indexOf(".<Value>.") + 9 ));
                 } else if (Boolean.TYPE.isAssignableFrom( _l_argumentsClass[i - 1])) {
                     if (_l_values[i].substring(_l_values[i].
                             indexOf(".<Value>.") + 9 ).equals("true"))
                         l_result[i-1] = true;
                     else if (_l_values[i].substring(_l_values[i].
                             indexOf(".<Value>.") + 9 ).equals("false"))
                         l_result[i-1] = false;
                     else throw new InternalError(
                             "AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                             " incorrect boolean value for the parameter "+_l_values[i]
                             +" on the master "+_object.getName());
                 } else if (Boolean.class.isAssignableFrom( _l_argumentsClass[i - 1])) {
                     if (_l_values[i].substring(_l_values[i].
                             indexOf(".<Value>.") + 9 ).equals("true"))
                         l_result[i-1] = Boolean.TRUE;
                     else if (_l_values[i].substring(_l_values[i].
                             indexOf(".<Value>.") + 9 ).equals("false"))
                         l_result[i-1] = Boolean.FALSE;
                     else throw new InternalError(
                             "AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                             " incorrect Boolean value for the parameter "+_l_values[i]
                             +" on the master "+_object.getName());
                 } else {

                                      // for testing purpose
                                //System.out.println("AssociateUtilityImpl."+
                                  //      "getL_ActionArgumentsForAssociations(), "+
                                  //"for the argument i = "+i+", it complies to the <Value> form"+
                                  //" with the name: ["+
                                  //_l_values[i].substring( _l_values[i].indexOf(".<Value>.") + 9 )+
                                  //"]");


                     name = null;
                     // when the last component of a name is entered, the argument class
                     // is Name, and the entered last component may be a Name
                     // or a simple String
                     if (_l_values[i].endsWith("::  "))
                         name = NameImpl.getNameFromString( _l_values[i].substring(
                             _l_values[i].indexOf(".<Value>.") + 9 ));
                     else
                         l_result[i-1] = _l_values[i].substring(
                             _l_values[i].indexOf(".<Value>.") + 9 );

                     if (name != null &&
                             ImmutableName.class.isAssignableFrom( _l_argumentsClass[i - 1]))
                         l_result[i-1] = name;
                     else if (name != null &&
                             BaseObject.class.isAssignableFrom( _l_argumentsClass[i - 1]))
                         l_result[i-1] = BaseUtilityImpl.getEorBaseObject(name);

                     if (l_result[i-1] == null)
                         throw new InternalError(
                             "AssociateUtilityImpl.getL_ArgumentsForAssociations()"+
                             " incorrect parsing for the parameter "+_l_values[i]
                             +" on the master "+_object.getName());
                 } // else
             } // else

         } // for i
     } catch (ClassNotFoundException b) {
          throw new BaseError("NoSuchMethodException in the "+
            "creation from "+_l_values[0]+" for the master "+_object+
            ", the exception is "+b.toString());
     }
     catch (InstantiationException b) {
          throw new BaseError("InstantiationException in the "+
            "creation from "+_l_values[0]+" for the master "+_object+
            ", the exception is "+b.toString());
     }
     catch (IllegalAccessException b) {
      throw new BaseError("IllegalAccessException in the "+
        "creation from "+_l_values[0]+" for the master "+_object+
        ", the exception is: "+b.toString());
     }

     l_argumentsNames = null;
     beanInfoClass = null;
     bean = null;
     beanDescriptor = null;
     val = null;
     l_descriptors = null;
     meth = null;
     param = null;


              // for testing and tracking
              //if (_l_values[0].equals("23.<CalledMethod>.<Action>.22.setResourceType")) {
                  //System.out.println(
                    //"AssociateUtilityImpl.getL_ArgumentsForAssociations() result:");
                  //for (int i = 0; i < l_result.length; i++) {
                    //if (l_result[i] != null)
                      //  System.out.println("--------- "+i+", class: "+l_result[i].getClass()+
                        //    ", value: "+l_result[i]);
                    //else System.out.println("--------- "+i+
                      //      ", value: "+l_result[i]);
                  //} // for
              //} // if

     return l_result;
  } // getL_ArgumentsForAssociations()

  /**
   * Sets the absolute arguments to use in the call of the constructor,
   * through an analyze of the relative names to concatenate.
   * The relative name is a String or an ImmutableName, and the argument
   * to put as context is the name of a BaseObject. <P>
   * This method is inspired by BaseObjectHandler.setFinalArguments(), since
   * it reads the values 'constrArgumentConcatenedNameSource', 'constrArgumentClasses',
   * and 'constrArgumentPropertyNames' in the BeanInfo class.
   * This method follows also the naming policy for UserIDs and GroupIDs,
   * from the EligiblePartyImpl constructor. <P>
   * Called by addAssociateActions().
   * @param _l_argumentsValues relative values set by getL_ActionArgumentsForAssociations(),
   *            which is updated by this method if it is a concatenated value.
   * @param _beanDescriptor descriptor of the ACSObject to create, from its BeanInfo
   * @return the updated _l_argumentsValues
   */
  public static final Object[] setFinalArgumentsForAssociateCreation(
          Object[] _l_argumentsValues, BeanDescriptor _beanDescriptor) {

    Integer[] l_argumentsConcatening = (Integer[]) _beanDescriptor.getValue(
                    "constrArgumentConcatenedNameSource");
    Class[]   l_argumentsClass = (Class[]) _beanDescriptor.getValue(
                    "constrArgumentClasses");
    String[]  l_argumentsNames = (String[]) _beanDescriptor.getValue(
                    "constrArgumentPropertyNames");
    int startIndex = -1;

    // it may exist some relative names that imply to use an other argument
    // to set the first components of their name
    for (int i = 0; i < _l_argumentsValues.length; i++) {
      // if it is a relative name to concatenate, process it,
      // test if the argument to concatenate is a BaseObject and is not null
      if (l_argumentsConcatening[i] == null) continue;

        if (_l_argumentsValues[l_argumentsConcatening[i].intValue()] != null &&
          _l_argumentsValues[l_argumentsConcatening[i].intValue()] instanceof BaseObject) {


                           // for testing and tracking purpose
                           //System.out.println("call to BaseObjectHandler.setFinalArguments(); ");

                           //System.out.println("        the class is "+l_argumentsClass_[i]);
                           //System.out.println("        the root name to add is "+
                              //(Name)((BaseObject)
                                    //l_argumentsValues_[l_argumentsConcatening_[i].intValue()]).
                                    //getName());


            if (!l_argumentsClass[i].isArray()){

                    // l_argumentsValues_[i] is a String or an ImmutableName,
                    // and the context is given by the name of BaseObject
              if (_l_argumentsValues[i] instanceof ImmutableName)
                  _l_argumentsValues[i] = ((Name)((BaseObject)
                    _l_argumentsValues[l_argumentsConcatening[i].intValue()]).
                    getName()).extended( (ImmutableName)_l_argumentsValues[i] );
              else if (_l_argumentsValues[i] instanceof String)
                  _l_argumentsValues[i] = ((Name)((BaseObject)
                    _l_argumentsValues[l_argumentsConcatening[i].intValue()]).
                    getName()).extended( (String)_l_argumentsValues[i] );

            } // if (!l_argumentsClass_[i].isArray())
            else {
                    // l_argumentsValues_[i] is a String or an ImmutableName
                    // and the context is given by the name of BaseObject
              if (_l_argumentsValues[i] instanceof ImmutableName)
                _l_argumentsValues[i] = new Name[] {((Name)((BaseObject)
                  _l_argumentsValues[l_argumentsConcatening[i].intValue()]).
                  getName()).extended( (ImmutableName)_l_argumentsValues[i] )};
              else if (_l_argumentsValues[i] instanceof String)
                _l_argumentsValues[i] = new Name[] {((Name)((BaseObject)
                  _l_argumentsValues[l_argumentsConcatening[i].intValue()]).
                  getName()).extended( (String)_l_argumentsValues[i] )};

            } // else
        }
        else { // the start of the name is null, so add the ACS name at the start

          // searching for the ACS name in the arguments
          for (int j = 0; j < l_argumentsNames.length; j++) {
            if (l_argumentsNames[j].equals("ACS") ) {
                startIndex = j;
                break;
            } // if
          } // for

          // the ACS name is non null
          if (!l_argumentsClass[i].isArray()){

                    // l_argumentsValues_[i] is a String or an ImmutableName
              if (_l_argumentsValues[i] instanceof ImmutableName)
                  _l_argumentsValues[i] = ((Name)((BaseObject)
                    _l_argumentsValues[startIndex]).
                    getName()).extended( (ImmutableName)_l_argumentsValues[i] );
              else if (_l_argumentsValues[i] instanceof String)
                  _l_argumentsValues[i] = ((Name)((BaseObject)
                    _l_argumentsValues[startIndex]).
                    getName()).extended( (String)_l_argumentsValues[i] );

          } // if (!l_argumentsClass_[i].isArray())
          else {
                    // l_argumentsValues_[i] is a String or an ImmutableName
              if (_l_argumentsValues[i] instanceof ImmutableName)
                _l_argumentsValues[i] = new Name[] {((Name)((BaseObject)
                  _l_argumentsValues[startIndex]).
                  getName()).extended( (ImmutableName)_l_argumentsValues[i] )};
              else if (_l_argumentsValues[i] instanceof String)
                _l_argumentsValues[i] = new Name[] {((Name)((BaseObject)
                  _l_argumentsValues[startIndex]).
                  getName()).extended( (String)_l_argumentsValues[i] )};

          } // else
        } // else

    } // for

    l_argumentsConcatening = null;
    l_argumentsClass =  null;
    l_argumentsNames = null;
    return _l_argumentsValues;
  } // setFinalArgumentsForAssociateCreation

  /**
   * Executes the ACS associations after the creation of an ACSObject.
   * Called by ActionNewResource, ActionNewEligibleParty.
   * Calls addAssociateActions() that does the main job,
   * and ActionNewEligibleParty.handleNewGroup().
   * @param _master just-created object, having or not some associates and actions
   *        to process. Not null.
   * @exception BaseError if the processing is not correct
   */
  public static final void addAssociatesAfterCreation(ACSObject _master) throws BaseError {
          // variables
      Set<String>   l_matchedKeys = getL_MapKeyMatchingForAssociationsOnNewObject(_master);

      if (l_matchedKeys == null) return;
      if (l_matchedKeys.isEmpty()) {
          l_matchedKeys = null;
          return;
      }

      String[]      l_actions = null;
      Map           m_asso = null;
      String        key = null;

      if (_master instanceof Resource)
        m_asso = ((Resource)_master).getEorACS().getM_AssociationsOnNewObject();
      else if (_master instanceof EligibleParty)
        m_asso = ((EligibleParty)_master).getEorACS().getM_AssociationsOnNewObject();
      else if (_master instanceof VirtualFolder)
        m_asso = ((VirtualFolder)_master).getEorACS().getM_AssociationsOnNewObject();


      for (Iterator ite = l_matchedKeys.iterator(); ite.hasNext(); ) {
          key = (String)ite.next();
          l_actions = (String[])m_asso.get(key);
          if (l_actions == null) continue;

          addAssociateActions(_master, l_actions, null);
      } // for ite

      l_matchedKeys = null;
      l_actions = null;
      m_asso = null;
  } // addAssociatesAfterCreation()

  /**
   * Removes the ACS associations after the deletion of an ACSObject.
   * Called by Basic.finalizeForUser().
   * @param _deletedObject just-deleted by the GUI, having or not some associates
   * @param _acsChild Child ACS. Not null only when _object is a parent ACS
   *        to update.
   * @exception BaseError if the processing is correct
   */
  public static final void removeAssociatesAfterDeletion(ACSObject _deletedObject,
          ACS _acsChild) throws BaseError {
          // variables
      Set<String>   l_matchedKeys = getL_MapKeyMatchingForAssociationsOnNewObject(_deletedObject);

      if (l_matchedKeys == null) return;
      if (l_matchedKeys.isEmpty()) {
          l_matchedKeys = null;
          return;
      }

      String[]      l_actions = null;
      Map           m_asso = null;

      if (_deletedObject instanceof Resource)
        m_asso = ((Resource)_deletedObject).getEorACS().getM_AssociationsOnNewObject();
      else if (_deletedObject instanceof EligibleParty)
        m_asso = ((EligibleParty)_deletedObject).getEorACS().getM_AssociationsOnNewObject();
      else if (_deletedObject instanceof VirtualFolder)
        m_asso = ((VirtualFolder)_deletedObject).getEorACS().getM_AssociationsOnNewObject();

      for (Iterator ite = l_matchedKeys.iterator(); ite.hasNext(); ) {
          l_actions = (String[])m_asso.get(ite.next());
          if (l_actions == null) continue;

          removeAssociateActions(_deletedObject, l_actions, _acsChild);

                  // for testing and tracking
                  //System.out.println("AssociateUtilityImpl.removeAssociatesAfterCreation()"+
                    //" the final state of the through-constructor object is: "+result.toString());

      } // for ite

      l_matchedKeys = null;
      l_actions = null;
      m_asso = null;
  } // removeAssociatesAfterDeletion()

  /**
   * Applicates the creation commands for the ACS roots after the creation or
   * the opening of a child ACS, in the case where these roots do not exist yet.
   * The creation commands in the argument _m_asso are get through a filtering of
   * _child.getM_RootsForOpeningAndClosing() and/or
   * _child.getM_RootCommandsForOpeningNoClosing(), for only one map key: <P>
   *  'ACS.Name.&lt;Value&gt;._parentACS_name'. <P>
   * Called by ACSImpl.setAcsRelationsInIS().
   * Does not update the parent ACS map for child ACSs, nor the relevant root property.
   * Calls addAssociateActions().
   * @param _child is an open ACS having or not some associates and actions
   * to process on its parent ACSs. Not null.
   * @param _m_asso the applicable root commands for the ACS creation or opening.
   * Never null.
   * @exception BaseError if the processing or a key is not correct
   * @return the associates created in the parent ACS, and the
   * Objects that some called methods returns. May be empty but not null.
   */
  public static final List<Object> addRootsInParentACSs(ACS _child,
         Map _m_asso) throws BaseError {
          // variables
      String[]          l_actions = null;
      List<Object>      l_result = null;
      Name              name = null;
      boolean           error = false;

      for (Object key : _m_asso.keySet()) {

          l_actions = (String[])_m_asso.get(key);
          if (l_actions == null) continue;

          if (!((String)key).startsWith("ACS.Name.<Value>.")) {
              error = true;
              break;
          }
          
          name = NameImpl.getNameFromString(" "+((String)key).substring(17)+"  ");
          if (l_result == null)
              l_result = addAssociateActions(
                  // gets the parent ACS
                  (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(name),
                  l_actions, _child);
          else
              l_result.addAll(addAssociateActions(
                  // gets the parent ACS
                  (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(name),
                  l_actions, _child)
                      );
      } // for key

      if (error)
          throw new BaseError("AssociateUtilityImpl.addRootsInParentACSs(): "+
                  "a key is not well formed.");

      l_actions = null;
      name = null;
      return l_result;
  } // addRootsInParentACSs()

  /**
   * Gets the roots that a child ACS needs to have in its parent ACS, and which are not
   * to delete at the closing of the child ACS. NO USE.
   * The data is get from _child.getM_RootsForOpeningNoClosing(),
   * for only one map key: <P>
   *  'ACS.Name.&lt;Value&gt;._parentACS_name'. <P>
   * Call getL_AssociatesAndCalledMethods().
   * @param _child is an open ACS having or not some associates and actions
   *        to process on its parent ACSs. Not null.
   * @exception BaseError if the processing or a key is not correct
   * @return root names. May be empty or null.
   */
  public static final Set<ImmutableName> getL_RootsWithoutDeletingInParentACSs(
          ACS _child) throws BaseError {
          // variables
      String[]          l_actions = null;
      Map               m_asso = null;
      Set<ImmutableName>   l_result = null;
      Name              name = null;
      boolean           error = false;

      m_asso = _child.getM_RootCommandsForOpeningNoClosing();

      if (m_asso == null) return null;

      for (Object key : m_asso.keySet()) {

          l_actions = (String[])m_asso.get(key);
          if (l_actions == null) continue;

          if (!((String)key).startsWith("ACS.Name.<Value>.")) {
              error = true;
              break;
          }

          name = NameImpl.getNameFromString(" "+((String)key).substring(17)+"  ");

          if (l_result == null)
              l_result = getL_AssociatesAndCalledMethods(
                  // get the parent ACS
                  (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(name),
                  l_actions, _child).keySet();
          else
              l_result.addAll(getL_AssociatesAndCalledMethods(
                  // get the parent ACS
                  (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(name),
                  l_actions, _child).keySet()
                      );
      } // for key

      if (error)
          throw new BaseError("AssociateUtilityImpl.getL_RootsWithoutDeletingInParentACSs() "+
                  "a key in _child.getM_RootCommandsForOpeningNoClosing() is not well formed.");

      l_actions = null;
      m_asso = null;
      name = null;
      return l_result;
  } // getL_RootsWithoutDeletingInParentACSs()

  /**
   * Executes the deleting commands for the ACS roots at the closing of a child ACS,
   * in the case where this deleting is required.
   * The deleting commands in the argument _m_asso are get through the child ACS
   * getM_RootsForOpeningAndClosing() method, for only one type of map key: <P>
   *  'ACS.Name.&lt;Value&gt;._parentACS_name'. <P>
   * Called by ACSImpl.finalizeForBase().
   * Calls removeAssociateActions().
   * @param _m_asso the applicable commands for the root deletions. Never null.
   * @param _child is an open ACS having or not some associates and actions
   *        to process on its parent ACSs. Not null.
   * @exception BaseError if the processing is not correct
   */
  public static final void removeRootsInParentACSs(Map _m_asso, ACS _child) throws BaseError {
          // variables
      Name name = null;

      for (Object entry : _m_asso.entrySet()) {

          name = NameImpl.getNameFromString(
                  ((String)((Map.Entry)entry).getKey()).substring(17));
          
          removeAssociateActions(
                  (ImmutableACS)BaseManagerImpl.m_OpenAcs__.get(name),
                  (String[])((Map.Entry)entry).getValue(),
                  _child);
      } // for entry

                  // for testing and tracking
                  //System.out.println("AssociateUtilityImpl.removeRootsInParentACSs()"+
                    //" the final state of the through-constructor object is: "+result.toString());

      name = null;
  } // removeRootsInParentACSs()

  /**
   * This method defines the first value for an association. For a new object,
   * this method returns 5 values, and it returns 3 values for a child ACS.
   * Called by ACSFactoryImpl for AssociationMapPropertyEditor,
   * with an argument get from the BeanInfo.
   * @param _key for the constrained values. The single workable value is 'ForChildACS'.
   * @return list of the allowed values. Never empty nor null.
   */
  public static final List<String> getL_LimitedFirstValuesForAssociation(String _key) {
      List<String>  l_result = null;

      if (_key != null && _key.equals("ForChildACS"))
          l_result = Arrays.asList(
                        new String[] {"<AssociateClass>.",
                        "<CalledMethod>.<ACS>.", "<CalledMethod>.<Action>."});
      else
          l_result = Arrays.asList(
              new String[] {"<AssociateClass>.", "<CalledMethod>.<Master>.",
                        "<CalledMethod>.<ACS>.", "<CalledMethod>.<Action>.",
                        "<CalledMethod>.<Object>."});
      return l_result;
  } // getL_LimitedFirstValueForAssociate()

  /**
   * Gets the relations a root parent ACSObject may have
   * with a constrained and opened child ACS.
   * There are 3 types of relation: <P>
   * - '&lt;ForChildAcsOpeningAndClosing&gt;': the root parent ACSObject, or one
   *    of its property value, has to be there when the child ACS is opened,
   *    to ensure its ability to exist,
   *    and has to be closed or updated (for a property value) at its closing, <P>
   * - '&lt;ForChildAcsOpeningNoClosing&gt;': the root parent ACSObject, or one
   *    of its property value, has to be there when the child ACS is opened,
   *    to ensure its ability to exist,
   *    but there is no constraint at its closing, <P>
   * - '&lt;ForChildAcsIntegrity&gt;': the root parent ACSObject, or one
   *    of its property value, has to be there when the child ACS is opened,
   *    to ensure its ability to work well, in other words, following its specifications. <P>
   * Each relation is associated to a key with the same name.
   * Called by SimplePropertyEditor.getRestrictedValues() to display the Resource
   * property 'M_ConstrainedChildACS', and by ResourceImpl.add/removeConstrainedChildACS().
   * @return array of the 3 allowed keys.
   */
  public static String[] getKeysForConstrainedChildACSMap() {
      return new String[] {
          "<ForChildAcsOpeningAndClosing>",
          "<ForChildAcsOpeningNoClosing>",
          "<ForChildAcsIntegrity>"};
  }
 
  /**
   * Called by ActionNewACSyst.fillDerivedACSproperties() to fill the new ACS with
   * all the association scripts for new objects to be copied, or for the roots
   * in the ACS parent.
   * Replaces the _refAcs name by the _factory name in the scripts.
   * Replaces in the map key, and for roots only, the _refAcs parent by the _factory
   * parent.
   * Removes the starting numbers in each command line.
   * Derived from addAssociateActions().
   * @param _factory of the new ACS
   * @param _refAcs is the reference acs from which the new acs is build up when
   *    it is a derived acs
   * @param _associationMap the associations to convert, from _refAcs
   * @return new Map of associations for _factory, where the key is the association
   *    key, and the value is the list of String arrays, with one array for each
   *    elementary command
   * */
   public static final Map fillDerivedAssociations(ACSFactoryImpl _factory,
           ACS _refAcs, Map _associationMap) {
    
      Map           m_result = new HashMap(20);
      String[]      l_actions = null;
      String[]      l_oneActionsSet = null;
      ArrayList<String[]>  l_commandsForOneKey = new ArrayList(100);
      Name          name = null;
      String        stringName =  null;
      int           index = -1;
      int           numberShift = -1;
      Map.Entry      entry = null;
      ImmutableNameISTree    tree = (ImmutableNameISTree) BaseManagerImpl.getInstance().
                                        getISTree(_factory.getISName());
      
      try {
          for (Iterator ite = _associationMap.entrySet().iterator(); ite.hasNext(); ) {
              entry = (Map.Entry)ite.next();
              l_actions = (String[])entry.getValue();
              if (l_actions == null) continue;
              l_commandsForOneKey.clear();

                           // for testing and tracking purpose
                           //System.out.println("call to AssociateUtilityImpl.fillDerivedAssociations() "+
                             //      "for the processing of "+entry.getKey());
              
              for (Iterator ita = AssociateUtilityImpl.splitActionSetsForAssociations(l_actions).iterator();
                ita.hasNext(); ) {
                  l_oneActionsSet = (String[])ita.next();

                  if (l_oneActionsSet[0].substring(3, 4).equals(".")) numberShift = 4;
                  else if (l_oneActionsSet[0].substring(2, 3).equals(".")) numberShift = 3;
                  else numberShift = 2;
                  
                  // N.<CalledMethod>.<Object>.XXX+<Class>+YYY+<Value>+ZZZ
                  if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Object>.") != -1) {
                         name = null;
                         stringName = l_oneActionsSet[0].substring(24+numberShift, 
                                     l_oneActionsSet[0].indexOf("+<Class>+"));

                         try {
                             if (stringName != null && stringName.endsWith("::  ") &&
                                     stringName.startsWith(" "))
                                 name = NameImpl.getNameFromString( stringName );
                         } catch (BaseError b) {
                             throw new BaseError(
                                     "AssociateUtilityImpl.fillDerivedAssociationsForNewObject()"+
                                     " incorrect search of the object name in "+l_oneActionsSet[0]
                                     +", with the exception "+b.toString());
                         }
                         
                         // replaces the _refAcs name by the _factory name in l_oneActionsSet[0]
                         if (name.subname(0, _refAcs.getName().noOfComponents()).equals(_refAcs.getName())) {

                                              // for testing and tracking
                                              //System.out.println("AssociateUtilityImpl.fillDerivedAssociations() for the key "+
                                                //      entry.getKey()+" and 'N.<CalledMethod>.<Object>.XXX+<Class>+YYY+<Value>+ZZZ'");
                                              //System.out.println("previous l_oneActionsSet[0]:"+l_oneActionsSet[0]);
                                              
                             l_oneActionsSet[0] = l_oneActionsSet[0].substring(0, 24+numberShift) +
                                     ((Name)_factory.getName()).extended(
                                        name.subname(_refAcs.getName().noOfComponents(), name.noOfComponents())).toString() +
                                     l_oneActionsSet[0].substring(l_oneActionsSet[0].indexOf("+<Class>+"));

                                                // for testing and tracking
                                              //System.out.println(
                                                //      "AssociateUtilityImpl.fillDerivedAssociationsForNewObject() with "+
                                                  //    "new l_oneActionsSet[0]: "+l_oneActionsSet[0]);

                         } // if (name.subname(0
                  } // if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Object>.") != -1)

                  if (l_oneActionsSet.length > 1) {
                      for (int i = 1; i < l_oneActionsSet.length; i++) {
                          // process the case 'PropertyName'.<Value>.XXX: the argument is XXX as String, ImmutableName  
                           //        if it is the property type, true/false if the type is boolean,
                           //        or any BaseObject get from its name <P>
                          if (!l_oneActionsSet[i].endsWith(".<Null>") &&
                                !l_oneActionsSet[i].endsWith(".<Master>") &&
                                !l_oneActionsSet[i].endsWith(".<MasterCreation>") && 
                                l_oneActionsSet[i].indexOf(".<ArgumentValue>.") == -1 &&
                                l_oneActionsSet[i].indexOf("<MasterNameLastComponent>") == -1 &&
                                l_oneActionsSet[i].indexOf("<MasterValue>") == -1 &&
                                l_oneActionsSet[i].indexOf("<MasterNameLastComponent>") == -1 &&
                                l_oneActionsSet[i].indexOf(".<MasterNameContext>") == -1 &&
                                l_oneActionsSet[i].indexOf(".<ACSName>+") == -1) {
                                    index = l_oneActionsSet[i].indexOf(_refAcs.getName().toString());
                                    if (index != -1) {

                                              // for testing and tracking
                                              //System.out.println("AssociateUtilityImpl.fillDerivedAssociations() for the key "+
                                                //      entry.getKey()+" and l_oneActionsSet[i] = "+l_oneActionsSet[i]);
                                        
                                        if (l_oneActionsSet[i].length() > index + 
                                                    _refAcs.getName().toString().length())
                                            l_oneActionsSet[i] = l_oneActionsSet[i].substring(0, index) +
                                                ((Name)_factory.getName()).extended(
                                                l_oneActionsSet[i].substring(index + 
                                                    _refAcs.getName().toString().length())).toString();
                                        else
                                             l_oneActionsSet[i] = l_oneActionsSet[i].substring(0, index) +
                                                _factory.getName().toString();

                                    } // if
                          } // if 
                      } // for i
                  } // if (l_oneActionsSet.length > 1)

                              // for testing and tracking
                              //System.out.println("AssociateUtilityImpl.fillDerivedAssociations() for the key "+
                                //      entry.getKey()+", result:");
                  
                  // removes the starting numbers, which are added in addAssociateActions()
                  for (int i = 0; i < l_oneActionsSet.length; i++) {
                      l_oneActionsSet[i] = l_oneActionsSet[i].substring(numberShift);

                                                // for testing and tracking
                                              //System.out.println(
                                                //      "AssociateUtilityImpl.fillDerivedAssociationsForNewObject() with "+
                                                //System.out.println("----- "+l_oneActionsSet[i]);

                  } // for i
                  // adds to the previous commands for the key
                  l_commandsForOneKey.add(l_oneActionsSet);
                  
              } // for (Iterator ita

              // does the main job, with a switch of the parent name for roots
              if (!((String)entry.getKey()).startsWith("ACS.Name.<Value>"))
                  m_result.put((String)entry.getKey(), l_commandsForOneKey.clone());
              else {
                  name = NameImpl.getNameFromString(((String)entry.getKey()).substring(17));
                  if (!_refAcs.getParentName().equals(name))
                      throw new InternalError("The roots command "+entry.getKey()+" does not use "+
                              "the parent name of "+_refAcs+" in the key. "+
                              "The associations cannot be derived.");
                  m_result.put(
                          "ACS.Name.<Value>."+_factory.getParentName().toString(),
                          l_commandsForOneKey.clone());
              } // else
          } // for ite 

          l_actions = null;
          l_oneActionsSet = null;
          name = null;
          stringName =  null; 
          entry = null;
          l_commandsForOneKey = null;
      } // try
      catch (Exception e) {

          e.printStackTrace();     
          throw new InternalError(
            "... From AssociateUtilityImpl.fillDerivedAssociations(). ");

      } // catch

      return m_result;
   } // fillDerivedAssociations()

// ---------------------- protected methods --------------------------------------

  /**
   * Executes the creating and updating commands. A '&lt;GeneralCriterion&gt;.' command
   * is not processed, since it is about control, and not for object creation.
   * With '&lt;AssociateClass&gt;', an instance is not created if it is returned by
   * BaseUtilityImpl.getBaseObject() from the name in the first argument, and
   * the instance is then put in the list of accessible objects for
   * any '&lt;CalledMethod&gt;.&lt;Action&gt;' association.
   * A called method may returns an object that has to be used by a next command as an argument.
   * For the creation of a GroupID, calls ActionNewEligibleParty.handleNewGroup()
   * to listen the new members group in the GUI.
   * Caution: any returned ACSObject will be put in the root names lists of an ACS
   * if it does not belong to this ACS. Then, when a root command calls a method
   * to get an argument for a next root command, it should not be such an ACSObject,
   * to avoid to be seen as a root.
   * Note: fillDerivedAssociations() is directly written from this method.
   * Called by addAssociatesAfterCreation() and addRootsInParentACSs().
   * Calls ActionNewEligibleParty.handleNewGroup() for any new GroupID.
   * @see #controlAssociations
   * @see ARoad0.gDMak.ActionNewEligibleParty#handleNewGroup(ARoad0.gBaseInterface.ImmutableGroupID) 
   * @param _object just-created ACSObject, having or not some associates and actions
   *        to process, or having a parent ACS to update. Not null.
   * @param _l_actions is a set of commands to execute
   * @param _acsChild Child ACS. Not null only when _object is a parent ACS
   *        to update.
   * @exception BaseError if the processing is not correct
   * @return the associates created in the own ACS or in the parent ACS, and the
   * Objects that some called methods returns. May be empty but not null.
   */
  protected static final List<Object> addAssociateActions(BaseObject _object,
          String[] _l_actions, ACS _acsChild) throws BaseError {
          // variables
      List<Object>          l_result = new ArrayList(INITIAL_CAPACITY);
      Class[]               l_argumentsClass = null;
      Object[]              l_argumentsValues = null;
      Class                 newBaseObjectClass = null;
      Class                 beanInfoClass = null;
      BeanInfoPattern       bean = null;
      BeanDescriptor        beanDescriptor = null;
      Object                argument = null;
      Map                   m_arguments = new HashMap(INITIAL_CAPACITY);

      String[]              l_oneActionsSet = null;
      Method                meth = null;
      Method[]              l_meth = null;
      ImmutableACS          acs = null;
      String                key = null;
      Name                  name = null;
      String                stringName =  null;
      BaseObject            obj = null;
      int                   index = -1;
      int                   numberShift = -1;

      try {
          for (Iterator ita = splitActionSetsForAssociations(_l_actions).iterator();
            ita.hasNext(); ) {
              l_oneActionsSet = (String[])ita.next();

                       // for testing and tracking purpose
                       //System.out.println("call to AssociateUtilityImpl.addAssociateActions(), "+
                         //      "on _object "+_object.getName()+" for  "+
                           //    "l_oneActionsSet[0]: "+l_oneActionsSet[0]);

              // does not process the '<GeneralCriterion>.' action sets,
              // which are for the integrity controls
              if (l_oneActionsSet[l_oneActionsSet.length - 1].indexOf("<GeneralCriterion>.") != -1)
                    continue;
              
              if (l_oneActionsSet[0].substring(3, 4).equals(".")) numberShift = 4;
              else if (l_oneActionsSet[0].substring(2, 3).equals(".")) numberShift = 3;
              else numberShift = 2;

              // N.<AssociateClass>.XXX
              // the class is not always from the package of _object, for an AcsAddon package
              if (l_oneActionsSet[0].indexOf(".<AssociateClass>.") != -1) {
                    if (_object instanceof ImmutableACS)
                        acs = (ImmutableACS) _object;
                    else if (_object instanceof ACSObject)
                        acs = ((ACSObject)_object).getEorACS();

                    newBaseObjectClass = Class.forName(
                          UtilityImpl.getFirstInstance().getEffectiveClassName(
                                acs.getM_AcsVocabulary(),
                                l_oneActionsSet[0].substring(17+numberShift)));
                    beanInfoClass = Class.forName(
                            UtilityImpl.getFirstInstance().getBeanInfoNameFromClass(
                                UtilityImpl.getFirstInstance().getEffectiveClassName(
                                    acs.getM_AcsVocabulary(),
                                    l_oneActionsSet[0].substring(17+numberShift))));
                    bean = (BeanInfoPattern)beanInfoClass.newInstance();

                    beanDescriptor = bean.getBeanDescriptor();

                    l_argumentsClass = (Class[]) beanDescriptor.getValue(
                        "constrArgumentClasses");

                    l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                            _object, m_arguments, true, l_oneActionsSet, 
                            l_argumentsClass, _acsChild);
                    l_argumentsValues = setFinalArgumentsForAssociateCreation(
                            l_argumentsValues, beanDescriptor);

                    // creates the associate, if necessary
                    if (ImmutableName.class.isAssignableFrom(l_argumentsClass[0])) {
                        // try to get the argument
                        argument = UtilityImpl.getFirstInstance().getBaseObject(
                                BaseUtilityImpl.getBaseObjectTrueName(
                                    (ImmutableName)l_argumentsValues[0],
                                    newBaseObjectClass));
                        if (argument == null) {

                                       // for testing and tracking purpose
                                       //System.out.println("AssociateUtilityImpl.addAssociateActions(), "+
                                         //      "argument is not found with the name: "+l_argumentsValues[0]);

                            // this is the standard way; the associate is created
                            argument = (ACSObject) newBaseObjectClass.
                                    getConstructor(l_argumentsClass).newInstance(l_argumentsValues);
                        } // if
                    }
                    else
                        argument = (ACSObject) newBaseObjectClass.
                            getConstructor(l_argumentsClass).newInstance(l_argumentsValues);

                    m_arguments.put(l_oneActionsSet[0].substring(0, numberShift - 1), argument);
                    l_result.add(argument);

                    // for a GroupID, calls a method to listen its members in the GUI
                    if (argument instanceof ImmutableGroupID)
                        ActionNewEligibleParty.handleNewGroup((ImmutableGroupID)argument);
              } // if (l_actions[0].indexOf(".<AssociateClass>.")

              // N.<CalledMethod>.<Master>.YYY
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Master>.") != -1) {
                     l_meth = _object.getClass().getMethods();
                     for (int j = 0; j < l_meth.length; j++) {
                         if (l_meth[j].getName().equals(
                                 l_oneActionsSet[0].substring(24+numberShift))) {
                             meth = l_meth[j];
                             break;
                         }
                     } // for j
                     l_argumentsClass = meth.getParameterTypes();

                    l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                            _object, m_arguments, true, l_oneActionsSet, 
                            l_argumentsClass, _acsChild);

                     try {
                        if (!meth.getReturnType().equals(Void.class))
                            l_result.add(meth.invoke(_object, l_argumentsValues));
                        else
                            meth.invoke(_object, l_argumentsValues);
                     } catch (InvocationTargetException i) {
                         System.out.println(i.getCause());
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the master "+_object.getName()
                                 +" with the exception "+i.getCause()+" in "+i);
                     } catch (Exception e) {
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the master "+_object.getName()
                                 +" with the exception "+e.toString());
                     }  // catch
              }

              // N.<CalledMethod>.<ACS>.YYY
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<ACS>.") != -1) {
                    if (_object instanceof ACSObject)
                         acs = ((ACSObject)_object).getEorACS();
                    else if (_object instanceof ACS)
                         acs = (ACS)_object;

                    l_meth = acs.getClass().getMethods();

                    for (int j = 0; j < l_meth.length; j++) {
                         if (l_meth[j].getName().equals(
                                 l_oneActionsSet[0].substring(21+numberShift))) {
                             meth = l_meth[j];
                             break;
                         }
                    } // for j
                    l_argumentsClass = meth.getParameterTypes();

                    l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                            _object, m_arguments, true, l_oneActionsSet, 
                            l_argumentsClass, _acsChild);

                     try {
                        if (!meth.getReturnType().equals(Void.class))
                            l_result.add(meth.invoke(acs, l_argumentsValues));
                        else
                            meth.invoke(acs, l_argumentsValues);
                     } catch (InvocationTargetException i) {
                         System.out.println(i.getCause());
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the ACS of "+_object.getName()
                                 +" with the exception "+i.getCause()+" in "+i);
                     } catch (Exception e) {
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the ACS of "+_object.getName()
                                 +" with the exception "+e.toString());
                     }  // catch
              }

              // N.<CalledMethod>.<Action>.C.ZZZ
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Action>.") != -1) {
                     key = l_oneActionsSet[0].substring(24 + numberShift);
                     argument = (Object)m_arguments.get(
                             key.substring(0, key.indexOf(".")));
                     if (!(argument instanceof ACSObject))
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call on the associate "+argument
                                 +" which is not an ACSObject.");
                     
                     l_meth = argument.getClass().getMethods();
                     for (int j = 0; j < l_meth.length; j++) {
                         if (l_meth[j].getName().equals(
                                 l_oneActionsSet[0].substring(
                                    numberShift + 25 +
                                    // 'C' may take one to three characters
                                    key.substring(0, key.indexOf(".")).length() ))) {
                             meth = l_meth[j];
                             break;
                         }
                     } // for j

                             // for testing and tracking
                             //System.out.println("AssociateUtilityImpl.addAssociateActions() "+
                               //      "on "+argument+", apply N.<CalledMethod>.<Action>.C.ZZZ where "+
                                 //    "the method ZZZ is "+meth+" with "+
                                   //  "l_oneActionsSet[0].substring(26 + numberShift) = "+
                                     //l_oneActionsSet[0].substring(26 + numberShift));

                     l_argumentsClass = meth.getParameterTypes();

                     l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                            _object, m_arguments, true, l_oneActionsSet, 
                            l_argumentsClass, _acsChild);

                     try {
                        if (!meth.getReturnType().equals(Void.class))
                            l_result.add(meth.invoke(argument, l_argumentsValues));
                        else
                            meth.invoke(argument, l_argumentsValues);
                     } catch (InvocationTargetException i) {
                         // for testing and tracking
                         if (l_argumentsValues != null) {
                             System.out.println("exception from AssociateUtilityImpl."+
                                     "addAssociateActions() "+
                                     "where l_argumentsValues classes are: ");
                             for (int k = 0; k < l_argumentsValues.length; k++) {
                                if (l_argumentsValues[k] != null)
                                    System.out.println("-------- "+l_argumentsValues[k].getClass());
                                else
                                    System.out.println("--------  null for "+k);
                             }
                         } // if
                         System.out.println("l_oneActionsSet[0]: "+l_oneActionsSet[0]);
                         System.out.println(i.getCause());

                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the associate "+((ACSObject)argument).getName()
                                 +" with the exception "+i.getCause()+" in "+i);
                     } catch (Exception e) {
                         // for testing and tracking
                         System.out.println("exception from AssociateUtilityImpl."+
                                 "addAssociateActions()"+
                                 " incorrect call of the method "+meth
                                 +" on the associate "+((ACSObject)argument).getName()
                                 +" with the exception "+e.toString());
                         System.out.println("l_oneActionsSet[0]: "+l_oneActionsSet[0]);
                         System.out.println(e.getCause());
                         
                         if (l_argumentsValues != null) {
                             System.out.println("AssociateUtilityImpl.addAssociateActions() "+
                                     "where l_argumentsValues classes are: ");
                             for (int k = 0; k < l_argumentsValues.length; k++) {
                                if (l_argumentsValues[k] != null)
                                    System.out.println("-------- "+l_argumentsValues[k].getClass());
                                else
                                    System.out.println("--------  null for "+k);
                             }
                         } // if

                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the associate "+((ACSObject)argument).getName()
                                 +" with the exception "+e.toString());
                     }  // catch
              }

              // N.<CalledMethod>.<Object>.XXX+<Class>+YYY+<Value>+ZZZ
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Object>.") != -1) {
                     name = null;
                     stringName = l_oneActionsSet[0].substring(24+numberShift,
                                 l_oneActionsSet[0].indexOf("+<Class>+"));

                     try {
                         if (stringName != null && stringName.endsWith("::  ") &&
                                 stringName.startsWith(" "))
                             name = NameImpl.getNameFromString( stringName );
                     } catch (BaseError b) {
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect search of the object name in "+l_oneActionsSet[0]
                                 +", with the exception "+b.toString());
                     }

                     try {
                         obj = BaseUtilityImpl.getEorBaseObject(name);
                     } catch (BaseError b) {
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect search of the object in "+l_oneActionsSet[0]
                                 +", with the exception "+b.toString());
                     }

                     l_meth = obj.getClass().getMethods();

                     index = l_oneActionsSet[0].indexOf("+<Value>+") + 9;
                     for (int i = 0; i < l_meth.length; i++) {
                          if (l_meth[i].getName().equals(l_oneActionsSet[0].substring(index))) {
                              meth = l_meth[i];
                              break;
                          }
                     } // for i
                     l_argumentsClass = meth.getParameterTypes();

                     l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                            _object, m_arguments, true, l_oneActionsSet, 
                            l_argumentsClass, _acsChild);


                                 // for testing and tracking
                                 //System.out.println("AssociateUtilityImpl.addAssociateActions() "+
                                   //      ", apply N.<CalledMethod>.<Object>. where "+
                                     //    "the method is "+meth+" on the object "+obj.getNickName());
                                 //System.out.println("-- AssociateUtilityImpl.addAssociateActions() "+
                                   //      "l_argumentsValues elements:");
                                 //for (int i = 0; i < l_argumentsValues.length; i++) {
                                   //  System.out.println("----- "+l_argumentsValues[i]);
                                 //} // for i


                     try {
                        if (!meth.getReturnType().equals(Void.class))
                            l_result.add(meth.invoke(obj, l_argumentsValues));
                        else
                            meth.invoke(obj, l_argumentsValues);
                     } catch (InvocationTargetException i) {
                         i.printStackTrace();
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the object "+obj.getName()
                                 +" with the exception "+i.getCause()+" in "+i);
                     } catch (Exception e) {
                         e.printStackTrace();
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the object "+obj.getName()
                                 +" with the exception "+e.toString());
                     }  // catch
              }
          } // for (Iterator ita

                      // for testing and tracking
                      //System.out.println("AssociateUtilityImpl.addAssociateActions()"+
                        //" the final state of the through-constructor object is: "+result.toString());

    } // try
    catch (NoSuchMethodException b) {
      throw new BaseError("NoSuchMethodException in the "+
        "creation from "+l_oneActionsSet[0]+" for the master or parent "+_object+
        ", the exception is "+b.toString());
    }
    catch (InstantiationException b) {
      throw new BaseError("InstantiationException in the "+
        "creation from "+l_oneActionsSet[0]+" for the master or parent "+_object+
        ", the exception is "+b.toString());
    }
    catch (InvocationTargetException t) {
      t.printStackTrace();
      throw new BaseError("InvocationTargetException in the "+
        "creation from "+l_oneActionsSet[0]+" for the master or the parent "+_object);
    }
    catch (IllegalAccessException b) {
      throw new BaseError("IllegalAccessException in the "+
        "creation from "+l_oneActionsSet[0]+" for the master or the parent "+_object+
        ", the exception is: "+b.toString());
    }
    catch (ClassNotFoundException b) {
      throw new BaseError("ClassNotFoundException in the "+
        "creation from "+l_oneActionsSet[0]+" for the master or the parent "+_object+
        ", the exception is: "+b.toString());
    }

    l_argumentsClass = null;
    l_argumentsValues = null;
    newBaseObjectClass = null;
    beanInfoClass = null;
    bean = null;
    beanDescriptor = null;
    argument = null;
    m_arguments = null;
    l_oneActionsSet = null;
    meth = null;
    l_meth = null;
    acs = null;
    name = null;
    stringName =  null;
    obj = null;
    name = null;
    return l_result;
  } // addAssociateActions()

  /**
   * Executes the deleting and updating operations in the ACS association commands.
   * A '&lt;GeneralCriterion&gt;.' command is not processed. It is about control, and not removing.
   * The commands N.&lt;CalledMethod&gt;.&lt;Action&gt;.C.ZZZ
   * and N.&lt;CalledMethod&gt;.&lt;Object&gt;.XXX+&lt;Class&gt;+YYY+&lt;Value&gt;+ZZZ are no more processed,
   * since the deleting of the master and associates should remove these actions.
   * Called by removeAssociatesAfterCreation() and removeRootsInParentACSs().
   * @param _object ACSObject having or not some associates and actions
   *        to process, or parent ACS to update. Not null.
   * @param _l_actions is an association script to execute. Never null.
   * @param _acsChild Child ACS. Not null only when _object is a parent ACS
   *        to update.
   * @exception BaseError if the processing is not correct
   */
  protected static final void removeAssociateActions(BaseObject _object, 
          String[] _l_actions, ACS _acsChild) throws BaseError {
          // variables
      ImmutableName name = null;
      Class[]       l_argumentsClass = null;
      Object[]      l_argumentsValues = null;
      Class         newBaseObjectClass = null;
      Class         beanInfoClass = null;
      BeanInfoPattern       bean = null;
      BeanDescriptor        beanDescriptor = null;
      Map           m_associates = new HashMap(INITIAL_CAPACITY);
      String[]      l_oneActionsSet = null;
      Method        meth = null;
      Method[]      l_meth = null;
      ImmutableACS  acs = null;
      int           numberShift = -1;

      if (_object instanceof Resource)
        acs = ((Resource)_object).getEorACS();
      else if (_object instanceof EligibleParty)
        acs = ((EligibleParty)_object).getEorACS();
      else if (_object instanceof VirtualFolder)
        acs = ((VirtualFolder)_object).getEorACS();
      else if (_object instanceof ImmutableACS)
        acs = (ImmutableACS) _object;
      else return;

      try {

          for (Iterator ita = splitActionSetsForAssociations(_l_actions).iterator();
            ita.hasNext(); ) {
              l_oneActionsSet = (String[])ita.next();

              // does not process the '<GeneralCriterion>.' action sets,
              // which are for the integrity controls
              if (l_oneActionsSet[l_oneActionsSet.length - 1].indexOf("<GeneralCriterion>.") != -1)
                    continue;

              if (l_oneActionsSet[0].substring(3, 4).equals(".")) numberShift = 4;
              else if (l_oneActionsSet[0].substring(2, 3).equals(".")) numberShift = 3;
              else numberShift = 2;

              // N.<AssociateClass>.XXX
              // the class is always from the package of _object, even for an AcsAddon package
              if (l_oneActionsSet[0].indexOf(".<AssociateClass>.") != -1) {
                    if (_object instanceof ImmutableACS)
                        acs = (ImmutableACS) _object;
                    else if (_object instanceof ACSObject)
                        acs = ((ACSObject)_object).getEorACS();

                    newBaseObjectClass = Class.forName(
                          UtilityImpl.getFirstInstance().getEffectiveClassName(
                                acs.getM_AcsVocabulary(),
                                l_oneActionsSet[0].substring(17+numberShift)));
                    beanInfoClass = Class.forName(
                            UtilityImpl.getFirstInstance().getBeanInfoNameFromClass(
                                UtilityImpl.getFirstInstance().getEffectiveClassName(
                                    acs.getM_AcsVocabulary(),
                                    l_oneActionsSet[0].substring(17+numberShift))));


                    bean = (BeanInfoPattern)beanInfoClass.newInstance();

                    beanDescriptor = bean.getBeanDescriptor();

                    l_argumentsClass = (Class[]) beanDescriptor.getValue(
                        "constrArgumentClasses");
                    l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                            _object, m_associates, true, l_oneActionsSet, 
                            l_argumentsClass, _acsChild);
                    l_argumentsValues = setFinalArgumentsForAssociateCreation(
                            l_argumentsValues, beanDescriptor);

                    // copy from the UserID and GroupID constructors
                    if (UserID.class.isAssignableFrom(newBaseObjectClass)) {
                      // add '<U>' between the ACS name and the end of _name
                      name = ((Name)acs.getName()).extended("<U>").extended(
                                 ((Name)l_argumentsValues[0]).subname(
                                   acs.getName().noOfComponents(),
                                   ((Name)l_argumentsValues[0]).noOfComponents()));
                    } else if (GroupID.class.isAssignableFrom(newBaseObjectClass)) {
                      // add '<G>' between the ACS name and the end of _name
                      name = ((Name)acs.getName()).extended("<G>").extended(
                                 ((Name)l_argumentsValues[0]).subname(
                                   acs.getName().noOfComponents(),
                                   ((Name)l_argumentsValues[0]).noOfComponents()));
                    } else name = (ImmutableName)l_argumentsValues[0];

                    // deletes the associate
                    if (Resource.class.isAssignableFrom(newBaseObjectClass))
                        ((Resource)BaseUtilityImpl.getEorBaseObject(name)).finalizeForBase();
                    else if (EligibleParty.class.isAssignableFrom(newBaseObjectClass))
                        ((EligibleParty)BaseUtilityImpl.getEorBaseObject(name)).finalizeForBase();
                    else if (VirtualFolder.class.isAssignableFrom(newBaseObjectClass))
                        ((VirtualFolder)BaseUtilityImpl.getEorBaseObject(name)).finalizeForBase();
              } // if (l_actions[0].indexOf(".<AssociateClass>.")

              // N.<CalledMethod>.<Master>.YYY
              // the deleting of the master and associates should remove this action

              // N.<CalledMethod>.<ACS>.YYY
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<ACS>.") != -1) {
                    if (_object instanceof ACSObject)
                         acs = ((ACSObject)_object).getEorACS();
                    else if (_object instanceof ACS)
                         acs = (ACS)_object;

                    l_meth = Class.forName(
                            UtilityImpl.getFirstInstance().getEffectiveClassName(
                                acs.getM_AcsVocabulary(),
                                "ACS")).getMethods();

                    for (int j = 0; j < l_meth.length; j++) {
                         if (l_meth[j].getName().equals(
                                 l_oneActionsSet[0].substring(21+numberShift))) {
                             meth = l_meth[j];
                             break;
                         }
                    } // for j
                    l_argumentsClass = meth.getParameterTypes();

                    l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                            _object, m_associates, false, l_oneActionsSet, 
                            l_argumentsClass, _acsChild);

                     try {
                        meth.invoke(acs, l_argumentsValues);
                     } catch (InvocationTargetException i) {
                         System.out.println(i.getCause());
                         throw new BaseError(
                                 "AssociateUtilityImpl.removeAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the ACS of "+_object.getName()
                                 +" with the exception "+i.getCause()+" in "+i);
                     } catch (Exception e) {
                         throw new BaseError(
                                 "AssociateUtilityImpl.removeAssociatesAfterCreation()"+
                                 " incorrect call of the method "+meth
                                 +" on the ACS of "+_object.getName()
                                 +" with the exception "+e.toString());
                     }  // catch
              }

              // N.<CalledMethod>.<Action>.C.ZZZ
              // N.<CalledMethod>.<Object>.XXX+<Class>+YYY+<Value>+ZZZ
              // the deleting of the master and associates should remove these actions
          } // for (Iterator ita

                      // for testing and tracking
                      //System.out.println("AssociateUtilityImpl.removeAssociateAction()"+
                        //" the final state of the through-constructor object is: "+result.toString());

    } // try
    catch (InstantiationException b) {
      throw new BaseError("InstantiationException in the "+
        "deletion from "+l_oneActionsSet[0]+" for the master "+_object+
        ", the exception is "+b.toString());
    }
    catch (IllegalAccessException b) {
      throw new BaseError("IllegalAccessException in the "+
        "deletion from "+l_oneActionsSet[0]+" for the master "+_object+
        ", the exception is: "+b.toString());
    }
    catch (BaseError p) {
      throw new BaseError(
        "deletion from "+l_oneActionsSet[0]+" for the master "+_object+
        ", the exception is: "+p.toString());
    } // catch
    catch (Exception e) {

              // for testing and tracking
              System.out.println(
                "deletion from "+l_oneActionsSet[0]+" for the master "+_object+
                ", the exception is: "+e.toString());
              for (int i = 0; i < l_oneActionsSet.length; i++) {
                System.out.println(i+": "+l_oneActionsSet[i]);
              } // for

      e.printStackTrace();
      throw new BaseError(
        "deletion from "+l_oneActionsSet[0]+" for the master "+_object+
        ", the exception is: "+e.toString());

    } // catch

      l_argumentsClass = null;
      l_argumentsValues = null;
      newBaseObjectClass = null;
      beanInfoClass = null;
      bean = null;
      beanDescriptor = null;
      m_associates = null;
      l_oneActionsSet = null;
      meth = null;
      l_meth = null;
      acs = null;
      name = null;
  } // removeAssociateActions()

  /**
   * Parses the ACS association commands to get the associates in a large sens,
   * and the methods which are implied. <P>
   * The status of a classical associate is described through the value '.&lt;AssociateClass&gt;.',
   * where the key is the associate name. <P>
   * The status of a root with an integrity constraint for a child ACS is described
   * through the '&lt;GeneralCriterion&gt;.&lt;RootIntegrity&gt;', where the key is the root name. <P>
   * The argument _object or its ACS may be a returned associate. <P>
   * This method is directly written from the addAssociateActions() code,
   * but it does not operate to create or update the associates.
   * Called by ACSFactoryImpl.createACS() and ACSImpl.initializeInnerAcsRelationsInIS().
   * @param _object just-created ACSObject, having or not some associates and actions
   *        to process, or parent ACS to update. Not null.
   * @param _l_actions is an association script to execute
   * @param _acsChild Child ACS. Not null only when _object is a parent ACS
   *        to update.
   * @exception BaseError if the processing is not correct
   * @return the map where the key is the name of the associate, and the value
   * is a list of the method names of the associate or, for an associate to create or to check,
   * the value list contains respectively the String '.<AssociateClass>.' or
   * '<GeneralCriterion&gt;.<RootIntegrity&gt;'.
   * May be empty but not null.
   */
  protected static final Map<ImmutableName, List<String>> getL_AssociatesAndCalledMethods(
          BaseObject _object, String[] _l_actions, ACS _acsChild) throws BaseError {
          // variables
      Map<ImmutableName, List<String>>   l_result = new HashMap(INITIAL_CAPACITY);
      Class[]               l_argumentsClass = null;
      Object[]              l_argumentsValues = null;
      Class                 newBaseObjectClass = null;
      Class                 beanInfoClass = null;
      BeanInfoPattern       bean = null;
      BeanDescriptor        beanDescriptor = null;
      Map                   m_associateClasses = new HashMap(INITIAL_CAPACITY);
      Map                   m_associateNames = new HashMap(INITIAL_CAPACITY);

      String[]              l_oneActionsSet = null;
      Method                meth = null;
      Method[]              l_meth = null;
      ImmutableACS          acs = null;
      String                key = null;
      Name                  name = null;
      String                stringName =  null;
      BaseObject            obj = null;
      int                   index = -1;
      int                   numberShift = -1;
      List<String>          l_methods = null;

      try {
          for (Iterator ita = splitActionSetsForAssociations(_l_actions).iterator();
            ita.hasNext(); ) {
              l_oneActionsSet = (String[])ita.next();

                       // for testing and tracking purpose
                       //System.out.println("call to AssociateUtilityImpl."+
                         //      "getL_AssociatesAndCalledMethods(), "+
                           //    "l_oneActionsSet[0]: "+l_oneActionsSet[0]);

              if (l_oneActionsSet[0].substring(3, 4).equals(".")) numberShift = 4;
              else if (l_oneActionsSet[0].substring(2, 3).equals(".")) numberShift = 3;
              else numberShift = 2;

              // N.<AssociateClass>.XXX
              if (l_oneActionsSet[0].indexOf(".<AssociateClass>.") != -1) {

                    if (l_oneActionsSet[l_oneActionsSet.length - 1].indexOf(
                            "<GeneralCriterion>") != -1) {
                        
                        // The name is of the associate is put in a one-item array.
                        l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                                _object, Collections.EMPTY_MAP, true, l_oneActionsSet,
                                new Class[] {ImmutableName.class}, _acsChild);

                        // m_associateClasses and m_associateNames are not initialized
                        // since they are never used by the other action sets

                        l_methods = l_result.get((Name)l_argumentsValues[0]);
                        if (l_methods == null) {
                             l_methods = new ArrayList(INITIAL_CAPACITY);
                        }
                        l_methods.add("<GeneralCriterion>.<RootIntegrity>");
                        l_result.put((ImmutableName)l_argumentsValues[0], l_methods);
                    }
                    else {
                        // classical case with <AssociateClass>
                        if (_object instanceof ImmutableACS)
                            acs = (ImmutableACS) _object;
                        else if (_object instanceof ACSObject)
                            acs = ((ACSObject)_object).getEorACS();

                        newBaseObjectClass = Class.forName(
                              UtilityImpl.getFirstInstance().getEffectiveClassName(
                                    acs.getM_AcsVocabulary(),
                                    l_oneActionsSet[0].substring(17+numberShift)));
                        beanInfoClass = Class.forName(
                                UtilityImpl.getFirstInstance().getBeanInfoNameFromClass(
                                    UtilityImpl.getFirstInstance().getEffectiveClassName(
                                        acs.getM_AcsVocabulary(),
                                        l_oneActionsSet[0].substring(17+numberShift))));

                        bean = (BeanInfoPattern)beanInfoClass.newInstance();

                        beanDescriptor = bean.getBeanDescriptor();

                        l_argumentsClass = (Class[]) beanDescriptor.getValue(
                            "constrArgumentClasses");
                        // The associations map may be empty. The name last component is the unique
                        // useful data there with the ACS, not the other constructor arguments,
                        // since the name last component is the first argument,
                        // with the acs name as context in another argument.
                        l_argumentsValues = (Object[]) getL_ArgumentsForAssociations(
                                _object, Collections.EMPTY_MAP, true, l_oneActionsSet, 
                                l_argumentsClass, _acsChild);
                        l_argumentsValues = setFinalArgumentsForAssociateCreation(
                                l_argumentsValues, beanDescriptor);

                        m_associateClasses.put(l_oneActionsSet[0].substring(0, numberShift - 1),
                                newBaseObjectClass);
                        m_associateNames.put(l_oneActionsSet[0].substring(0, numberShift - 1),
                                l_argumentsValues[0]);

                        l_methods = new ArrayList(INITIAL_CAPACITY);
                        l_methods.add(".<AssociateClass>.");
                        l_result.put((Name)l_argumentsValues[0], l_methods);
                    } // else
              } // if (l_actions[0].indexOf(".<AssociateClass>.")

              // N.<CalledMethod>.<Master>.YYY
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Master>.") != -1) {
                     l_meth = _object.getClass().getMethods();
                     for (int j = 0; j < l_meth.length; j++) {
                         if (l_meth[j].getName().equals(
                                 l_oneActionsSet[0].substring(24+numberShift))) {
                             meth = l_meth[j];
                             break;
                         }
                     } // for j
                     
                     l_methods = l_result.get(_object.getName());
                     if (l_methods == null) {
                         l_methods = new ArrayList(INITIAL_CAPACITY);
                     }
                     l_methods.add(meth.getName());
                     l_result.put(_object.getName(), l_methods);
              }

              // N.<CalledMethod>.<ACS>.YYY
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<ACS>.") != -1) {
                    if (_object instanceof ACSObject)
                         acs = ((ACSObject)_object).getEorACS();
                    else if (_object instanceof ACS)
                         acs = (ACS)_object;

                    l_meth = Class.forName(
                            UtilityImpl.getFirstInstance().getEffectiveClassName(
                                acs.getM_AcsVocabulary(),
                                "ACS")).getMethods();

                    for (int j = 0; j < l_meth.length; j++) {
                         if (l_meth[j].getName().equals(
                                 l_oneActionsSet[0].substring(21+numberShift))) {
                             meth = l_meth[j];
                             break;
                         }
                    } // for j

                     l_methods = l_result.get(acs.getName());
                     if (l_methods == null) {
                         l_methods = new ArrayList(INITIAL_CAPACITY);
                     }
                     l_methods.add(meth.getName());
                     l_result.put(acs.getName(), l_methods);
              }

              // N.<CalledMethod>.<Action>.C.ZZZ
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Action>.") != -1) {
                     key = l_oneActionsSet[0].substring(24 + numberShift);
                     newBaseObjectClass = (Class)m_associateClasses.get(
                             key.substring(0, key.indexOf(".")));
                     l_meth = newBaseObjectClass.getMethods();
                     for (int j = 0; j < l_meth.length; j++) {
                         if (l_meth[j].getName().equals(
                                 l_oneActionsSet[0].substring(26 + numberShift))) {
                             meth = l_meth[j];
                             break;
                         }
                     } // for j

                     name = (Name)m_associateNames.get(key.substring(0, key.indexOf(".")));

                     l_methods = l_result.get(name);
                     l_methods.add(meth.getName());
                     l_result.put(name, l_methods);
              }

              // N.<CalledMethod>.<Object>.XXX+<Class>+YYY+<Value>+ZZZ
              else if (l_oneActionsSet[0].indexOf(".<CalledMethod>.<Object>.") != -1) {
                     name = null;
                     stringName = l_oneActionsSet[0].substring(24+numberShift,
                                 l_oneActionsSet[0].indexOf("+<Class>+"));

                     try {
                         if (stringName != null && stringName.endsWith("::  ") &&
                                 stringName.startsWith(" "))
                             name = NameImpl.getNameFromString( stringName );
                     } catch (BaseError b) {
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect search of the object name in "+l_oneActionsSet[0]
                                 +", with the exception "+b.toString());
                     }

                     try {
                         obj = BaseUtilityImpl.getEorBaseObject(name);
                     } catch (BaseError b) {
                         throw new BaseError(
                                 "AssociateUtilityImpl.addAssociatesAfterCreation()"+
                                 " incorrect search of the object in "+l_oneActionsSet[0]
                                 +", with the exception "+b.toString());
                     }

                     l_meth = obj.getClass().getMethods();

                     index = l_oneActionsSet[0].indexOf("+<Value>+") + 9;
                     for (int i = 0; i < l_meth.length; i++) {
                          if (l_meth[i].getName().equals(l_oneActionsSet[0].substring(index))) {
                              meth = l_meth[i];
                              break;
                          }
                     } // for i

                     l_methods = l_result.get(obj.getName());
                     if (l_methods == null) {
                         l_methods = new ArrayList(INITIAL_CAPACITY);
                     }
                     l_methods.add(meth.getName());
                     l_result.put(obj.getName(), l_methods);
              }
          } // for (Iterator ita

                      // for testing and tracking
                      //System.out.println("AssociateUtilityImpl.getL_AssociatesAndCalledMethods()"+
                        //" the final keys in the result are: ");
                      //for (ImmutableName nm : l_result.keySet()) {
                        //  System.out.println("       "+nm);
                      //}

    } // try
    catch (InstantiationException b) {
      throw new BaseError("InstantiationException in the "+
        "creation from "+l_oneActionsSet[0]+" for the master "+_object+
        ", the exception is "+b.toString());
    }
    catch (IllegalAccessException b) {
      throw new BaseError("IllegalAccessException in the "+
        "creation from "+l_oneActionsSet[0]+" for the master "+_object+
        ", the exception is: "+b.toString());
    }
    catch (Exception e) {

              // for testing and tracking
              //System.out.println(
                //"creation from "+l_oneActionsSet[0]+" for the master "+_master+
                //", the exception is: "+e.toString());
              //for (int i = 0; i < l_oneActionsSet.length; i++) {
                //System.out.println(i+": "+l_oneActionsSet[i]);
              //} // for

      e.printStackTrace();
      throw new BaseError(
        "creation from "+l_oneActionsSet[0]+" for the master "+_object);

    } // catch

    l_argumentsClass = null;
    l_argumentsValues = null;
    newBaseObjectClass = null;
    beanInfoClass = null;
    bean = null;
    beanDescriptor = null;
    m_associateClasses = null;
    m_associateNames = null;
    l_oneActionsSet = null;
    meth = null;
    l_meth = null;
    acs = null;
    name = null;
    stringName =  null;
    obj = null;
    name = null;
    return l_result;
  } // getL_AssociatesAndCalledMethods()

} // AssociateUtilityImpl

