//Title:        Access Road
//Version:      0.7.2
//Copyright:    Copyright 2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.Gui1.Desktop;

import ARoad0.ErrB.*;
import ARoad0.ErrW.DialogError;
import ARoad0.Pattern.Singleton;

import java.util.*;

import java.io.ObjectStreamField;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.File;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;


/**
* This large class is the single manager for the Access Road object-oriented database.
* It allows to register, save, close, open and remove the ACS and the views.
* It allows to create, update, save and remove the information systems,
* their nodes and components. It allows to register and unregister an AcsAddon.
* After ACSFactoryImpl and ACSImpl, this is the most complex class of the gBase package.
* <P> 
* This class manages a map of nametrees which describes the tree structures of IS,
* components and ACS. The key is the IS name, and the value is a NameISTree which
* describes the structure. An information system may contain some ACS, subACS, 
* composite ACS, composite subACS, physical components and logical components.
* A physical component is a hardware in the information system.
* A logical component is simply a software which have no ACS functions.
* An information system cannot contain another one. <P> 
* The IS structure is a tree of named nodes, each having a type and some associated properties.
* The ACS nodes are opened and closed in the program, and all the other nodes 
* are created/opened/closed at the ACS creation/opening/closing. This is true 
* for the subACS of the ACS, but also for any composite ACS which is opened only 
* when all the ACS of its elements are opened, and for any physical or logicial component, 
* since it is opened in the IS at the opening of one ACS which is its child. <P>
* This structure has a specific format which is defined in the class ISNodeImpl.
* There is a default IS for all the ACS without an explicit IS,
* and all these ACS are in a flat list under the default IS root.
* Deleting is done into the Access Road base, not the files into the Access Road
* operating system work directory.
* <P>
* A view is a user selection of several ACS objects. <P>
* This class offers lists of the current open or closed ACS and views.
* It does not create the ACS and the views directly, but registers and saves them
* when there are created through ACSFactory or their view constructors.
* On the other hand, the information system nametrees are created directly by this class.<P>
* Caution: a new ACS may have the name of an ACSObject in a closed ACS,
* and this is not checked. The ACSObjects in the open ACS are all checked.
* When the relevant closed ACS is opened, there is no more checking
* of the name doublons. During the search of objects from their names, the ACS
* is selected first.
* <P>
* There is only one instance of this class.
* It is a javabean with the following bound properties: <P>
* 'OpenACS', 'OpenViews', 'OneInformationSystem', 'OneInformationSystemNode',
* 'OneInformationSystemNodeOpening'. Since the AcsAddons are (un)registered
* at the starting of the program, there is no bound property for them.
* <P>
* Other changes in nametree nodes are managed by ACSFactory.
* The listeners are proper to this instance, outside gBase, transient.
* They are called in any order.<P>
* Nevertheless, an order exists between these events and the events fired
* by the ACS and the views, when there are some operations on them:<P>
* - any change on 'OpenACS' and 'OpenViews' is fired - after - the creation or opening,
*       but, it is fired - before - the closing or removing, and
*       the events from the ACS or the view.<P>
* The listeners receive only a copy of the new value, to protect the property.
* All the exceptions from the listeners are catched,
* and a dialog box is displayed to inform the user.
* <P>
* All the Access Road data is saved as serialized Java objects under the Access Road
* working directory, with the file extension 'acr'. This includes the AcsAddon files.
* There are 3 main types of objects to serialize independently:<P>
*   - ACS: for each one, the file name starts with 'ACS0_';
*     its serialization maintains references among
*     the objects of almost all classes in gBase, excepted for ViewInBase classes
*     for which there are no references,<P>
*   - ViewInBase: for each one, the file name starts with 'VIEW0_';
*     its serialization does not maintain references among
*     the objects in all the relevant ACS, since only names are saved. 
*     There are no references to other ViewInBases into it.<P>
*   - Information System NameTree: the single file name is 'm_nametrees_v0.acr'; 
*     it is serialized without external references into it.<P>
* The other small objects to save are the general descriptors of the Access Road
* context. That includes:<P>
* - 'l_acs_v0.acr' for the ACS names,<P>
* - 'l_views_v0.acr' for the view names,<P>
* - 'l_metarights_v0.acr' for the metarights,<P>
* - 'l_acsaddon_v0.acr' for the ACS addon names,<P>
* - 'GUI_config.acr' for the locations of the objects in the GUI.
*/
public class BaseManagerImpl implements ACSManager, ViewManager, Singleton {
  private static final int        INITIAL_CAPACITY = 10; // the default value: 101
 
  /**
   * ACS names
   */
  private static List<ImmutableName> l_Acs__ = new ArrayList(INITIAL_CAPACITY*2);
  //  protected List                  l_Streams_ = new ArrayList();
  
  /**
   * m_NameTrees__ is a map of NameISTrees which describes the structure of each IS.
   * The key is the IS name, and the value is a NameISTree where the root is the IS name.
   * The NameISTree has immutable names as nodes, which are the names of the ACS,
   * physical component, logical components, subACS and composite ACS of the information system.
   * The NameISTree defines the properties of each node and its links with its
   * environment in the information system. The only node property which is often
   * updated is the boolean to tell the node is opened or closed. The other node
   * properties in the NameISTree are updated only when a node is created or removed from
   * the base, at the creation/removing of an ACS.
   * 
   */
  private static Map<ImmutableName, ImmutableNameISTree> m_NameTrees__ = null;
  /**
   * The form of a value is 'Editor: full_editor_name - AcsAddon: addon_name',
   * like in ACSImpl.getEditorAndAddonNames().
   */
  private static List<String> l_AcsAddonDescriptors__ = new ArrayList(INITIAL_CAPACITY);
  /**
   * View names
   */
  private static List<ImmutableName> l_Views__ = new ArrayList(INITIAL_CAPACITY);
        // NO PRIVATE FOR TESTING and for BaseUtilityImpl
  /**
   * ACS map; ACS name is the key
   */
  protected static Map<ImmutableName, ImmutableACS> m_OpenAcs__  = new HashMap(INITIAL_CAPACITY);
  /**
   * View map; view name is the key
   */
  protected static Map<ImmutableName, ViewInBase> m_OpenViews__ = new HashMap(INITIAL_CAPACITY);
  //  protected Map                   m_OpenStreams_ = new HashMap();
  //  protected Map                   m_OpenIS_ = new HashMap();
  
  /**
   * List of metarights.
   */
  protected static List<StringMetaRight> l_MetaRights__ = null;
  
  /**
   * Map of metarights where the getRight() value is the key; 
   * the value in the map is the metaright. No saving.
   */
  protected static Map<String, StringMetaRight> m_MetaRightsFromRight__ = new HashMap(INITIAL_CAPACITY*4);
  
  /**
   * Map of metarights where the getPower() value is the key; 
   * the value in the map is an array of all the metarights having the same power.
   * No saving.
   */
  protected static Map<Integer, StringMetaRight[]> m_MetaRightsArrayFromPower__ = new HashMap(INITIAL_CAPACITY*4);
  
  
  private static final long       serialVersionUID = 0;

  // Transient variables
   /**
   * manage all the property change listeners
   */
  private transient PropertyChangeSupport changeSupport_ = null;
  
      /* the only instance of the class, following singleton pattern */
  protected static final BaseManagerImpl  INSTANCE = new BaseManagerImpl();

  
      /** variable for the JDK 2 serialization. UNUSED.
      * @serial l_Acs__, l_Views__, m_OpenAcs__, m_OpenViews__,
      *   serialVersionUID, l_MetaRights__, m_NameTrees__
      */
  public final static ObjectStreamField[] serialPersistentFields =
  {
//      new ObjectStreamField("l_Streams_",ArrayList.class),
      new ObjectStreamField("l_Acs__",ArrayList.class),
      new ObjectStreamField("l_Views__",ArrayList.class),
//      new ObjectStreamField("m_OpenStreams_",HashMap.class),
      new ObjectStreamField("serialVersionUID",Long.TYPE),
      new ObjectStreamField("l_MetaRights__",ArrayList.class),
      new ObjectStreamField("m_NameTrees__",HashMap.class)
  };


  // --------------------- static initialization -----------------------------
  //static {
      
  //}

    //
    // ----------- CONSTRUCTORS ---------------------
    //

  /** 
   * Reads the files "l_metarights_v0.acr", "l_is_v0.acr", "m_nametrees_v0.acr",
   * l_acs_v0.acr" and "l_views_v0.acr" to initialize the rights, IS, components,
   * ACS and view maps. Writes the file 'l_acsaddon_v0.acr'.
   * Uses Desktop.AR_URI which has to be intialized. GetGUI() is the first method
   * called by Gui1.Desktop after the construction.
  */
  private BaseManagerImpl() {
                    
                        // for testing and tracking        
                    //System.out.println("Call to the BaseManagerImpl constructor");
   
   //      l_Streams_ = new ArrayList(MAP_INITIAL_CAPACITY);

      // variables
    FileInputStream     fis = null;
    ObjectInputStream   in = null;

    //
    // open the known metarights file
    //
    try {
        fis = new FileInputStream("l_metarights_v0.acr");
        in = new ObjectInputStream(fis);
//      catch (SecurityException s) {
//        throw new InternalError("BaseManagerImpl.getInstance() failed: "+s.toString());
//
    } catch (IOException i) {

        // no l_metarights_v0.acr in the default directory, so create a list
        // of basic metarights
        l_MetaRights__ = new ArrayList(ACSFactoryUtilityImpl.initializeBasicMetaRights() );
        List l_copy = new ArrayList(l_MetaRights__);
        l_MetaRights__.clear();  
        // update l_MetaRights__, sort from the greater to the lesser
        try {
            for (Iterator ite = l_copy.iterator(); ite.hasNext(); ) {
                //the internal metarights maps are updated, without metaright saving
                newMetaRight( (StringMetaRight)ite.next(), false );
            } // for ite
        } catch (CreateError c) {
            throw new InternalError ("from BaseManagerImpl constructor on l_MetaRights__: "+
                    c.toString());
        }  

        // FOR StringRight TESTING only: replace all the lines above by the line below
            //l_MetaRights__ = new ArrayList(30 );        
    } // catch

    // if no IOException, read from the file
    if (l_MetaRights__ == null && in != null) {
        try {
            l_MetaRights__ = (List<StringMetaRight>) in.readObject();
            in.close();
        } catch (ClassNotFoundException f) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              " ClassNotFoundException to open the metarights list: "+f.toString());
        } catch (IOException i) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              " IOException to open the metarights list: "+i.toString());
        }
        // the internal metarights maps are created
        createMetaRightMaps();        
    } // if
    else if (!l_MetaRights__.isEmpty()) { 
        // save new basic metarights created after IOException
        try {
            saveMetaRights();
        } catch (BaseError b) {
            throw new InternalError("from BaseManagerImpl constructor: "+b.toString());
        }
    } // else
    
    //
    // open the known nameTree file
    //
    try {
        fis = new FileInputStream("m_nametrees_v0.acr");
        in = new ObjectInputStream(fis);
//      catch (SecurityException s) {
//        throw new InternalError("BaseManagerImpl.getInstance() failed: "+s.toString());
//
    } catch (IOException i) {
        // if no m_nametrees_v0.acr in the default directory, create an empty list
        m_NameTrees__ = new HashMap(INITIAL_CAPACITY);
    } // catch

    // if no IOException
    if (m_NameTrees__ == null && in != null) {
        try {
            m_NameTrees__ = (Map) in.readObject();
            in.close();
        } catch (ClassNotFoundException f) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              " ClassNotFoundException to open the nameTree map: "+f.toString());
        } catch (IOException i) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              " IOException to open the nameTree map: "+i.toString());
        }
    } // if

    // creates and saves the IPV4_ADDRESSES_SPACE and the IPV6_ADDRESSES_SPACE IS
    if (m_NameTrees__.get(new NameImpl(ISFactoryUtilityImpl.IPV4_ADDRESSES_SPACE)) == null) {
        try {
            addIS(new NameImpl(ISFactoryUtilityImpl.IPV4_ADDRESSES_SPACE), true);
            addIS(new NameImpl(ISFactoryUtilityImpl.IPV6_ADDRESSES_SPACE), true);
        }
        catch (BaseError b) {
            throw new InternalError("in the BaseManagerImpl.getInstance() call,"+
              " exception to create the IP IS: "+b.toString());
        }
    } // if

    // updates and saves the list of ACS addons
    try {
        updateAndSaveAcsAddonDescriptors();
    }
    catch (BaseError b) {
            throw new InternalError("in the BaseManagerImpl.getInstance() call,"+
              " exception to create the list of ACS addons: "+b.toString());
    }

                    // for testing and tracking
                    //System.out.println("From BaseManagerImpl constructor, m_NameTrees__ keys:");
                    //for (ImmutableName name : m_NameTrees__.keySet()) {
                      //  System.out.println("---------- "+name);
                    //}

    
    //
    // open the known ACS file
    //
    try {
        fis = new FileInputStream("l_acs_v0.acr");
        in = new ObjectInputStream(fis);
//      catch (SecurityException s) {
//        throw new InternalError("BaseManagerImpl.getInstance() failed: "+s.toString());
//
    } catch (IOException i) {
        // no l_acs_v0.acr in the default directory,
        // so l_Acs__ = new ArrayList(INITIAL_CAPACITY);
        fis = null;
        in = null;
        //return INSTANCE;
            
                    // for testing and tracking        
                    //System.out.println("From BaseManagerImpl constructor,"+
                      //" IOException in ACS reading:");
                    //i.printStackTrace();
    }

                        /*                
                    // l_Acs__ is null there
                         // for testing and tracking 
                    for(Iterator ite = l_Acs__.iterator() ;ite.hasNext() ; ) {
                      System.out.println("From BaseManagerImpl constructor,"+
                                      " a registered ACS:"+(Name)ite.next());
                    }
                         */
    if (in != null) {
        try {
            l_Acs__ = (List) in.readObject();
            in.close();
        } catch (ClassNotFoundException f) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              " ClassNotFoundException to open the ACS list: "+f.toString());
        } catch (IOException i) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              " IOException to open the ACS list: "+i.toString());
        }
    } // if

                    // for testing and tracking 
                    //for(Iterator ite = l_Acs__.iterator() ;ite.hasNext() ; ) {
                      //System.out.println("From BaseManagerImpl constructor,"+
                        //              " a registered ACS:"+(Name)ite.next());
                    //}
    
    //
    // open the known VIEWS file
    //
    try {
        fis = new FileInputStream(
        "l_views_v0.acr");
        in = new ObjectInputStream(fis);
//      catch (SecurityException s) {
//        throw new InternalError(s.toString()+ "BaseManagerImpl.openEorACS() failed");
//
    } catch (IOException i) {
        // no l_views_v0.acr in the default directory,
        // l_Views__ = new ArrayList(INITIAL_CAPACITY);
        fis = null;
        in = null;
        //return INSTANCE;
            
                    // for testing and tracking        
                    //System.out.println("From BaseManagerImpl constructot, "+
                     //   "IOException in views reading:");
                    //i.printStackTrace();
                    
    } // catch

    if (in != null) {
        try {
            l_Views__ = (List) in.readObject();
            in.close();
        } catch (ClassNotFoundException f) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              "ClassNotFoundException to open the views list: "+f.toString());
        } catch (IOException i) {
            throw new InternalError("in BaseManagerImpl.getInstance() call,"+
              "IOException to open the views list: "+i.toString());
        }
    } // if

    fis = null;
    in = null;
  
  } // constructor

    /**
    * Only method to obtain the unique instance of BaseManagerImpl.
    * Test the file l_acs_v0.acr in the default directory and
    * read it if it exists. Reads also l_is_v0.acr, l_components_v0.acr,
    * l_views_v0.acr
    * @return BaseManagerImpl
    */
  public static final BaseManagerImpl getInstance() {

                            // for testing and tracking
                        //System.out.println("BaseManagerImpl.getInstance()"+//view_.getName()+
                          //      "; run in EDT: "+SwingUtilities.isEventDispatchThread());

    return INSTANCE;
  } // getInstance()

    /**
    * Create the generic metaright maps which are internal to this instance.
    * Called by the constructor.
    */
  private static final void createMetaRightMaps() {
     
    StringMetaRight right;
    StringMetaRight[] l_rights_one;
    StringMetaRight[] l_rights_two;
    Integer power;

    for (Iterator ite = l_MetaRights__.iterator(); ite.hasNext();) {
        right = (StringMetaRight) ite.next();
        m_MetaRightsFromRight__.put(right.getRight(), right);
        
        power = new Integer(right.getPower());
        if (!m_MetaRightsArrayFromPower__.containsKey( power )) {
             m_MetaRightsArrayFromPower__.put(power,  new StringMetaRight[]{right} );
        } else {
            l_rights_one = (StringMetaRight[]) m_MetaRightsArrayFromPower__.get(power);
            l_rights_two = new StringMetaRight[l_rights_one.length + 1];
            System.arraycopy(l_rights_one, 0, l_rights_two, 0, l_rights_one.length);
            l_rights_two[l_rights_two.length - 1] = right;
            m_MetaRightsArrayFromPower__.put(power,  l_rights_two );
        } // else
    } // for ite

    right = null;
    l_rights_one = null;
    l_rights_two = null;
    power = null;
  } // createMetaRights()
  
    //
    // -------- the ImmutableNameTree interface METHODS ----------
    //

  /**
  * @param _name ImmutableName of an IS
  * @return boolean true if _name is known as an IS name.
  */
  public boolean knowsIS(ImmutableName _name){
    return m_NameTrees__.keySet().contains(_name);
  }  

  /**
  * Gets the IS name if the name is known as a node name or an area name.
  * The beginning of the argument should be the IS name, and this method checks
  * if the name is really known by the IS.
  * The name may be an IS, an ACS, a subACS, a composite, a physical component or
  * a logical component. To use for testing the presence of a node or an area.
  * Note: it is not allowed to have two components with the same name
  * in two different IS.
  * @param _name of the node or area. Never null.
  * @return IS name, or null.
  */
  public ImmutableName getISofName(ImmutableName _name) {
      boolean known = false;
      ImmutableName result = null;

      for (ImmutableNameTree nameTree : m_NameTrees__.values()) {
          if (nameTree.isNode(_name)) {
              known = true;
              result = nameTree.getRoot();
              break;
          } // if
          else if (nameTree.isArea(_name)) {
              known = true;
              result = nameTree.getRoot();
              break;
          } // if
      } // for ite

      if (!known) return null;
      else return result;
  } // getISofName()
  
  /**
  * Returns the names of the registered IS, which are also the IS nametree names.
  * The order of the IS is the order of their creations. This order is used
  * to display both the IS trees and the ACS in the explorer.
  * @return set of IS names. Never null.
  */
  public Set<ImmutableName> getL_IS() {

                            // for testing and tracking
                        //System.out.println("BaseManagerImpl.getL_IS(), return:");
                        //for (ImmutableName name : m_NameTrees__.keySet()) {
                          //  System.out.println("------ "+name);
                        //}

      Set<ImmutableName> result = new HashSet(INITIAL_CAPACITY);
      ImmutableName ISname;

      for (Map.Entry entry : m_NameTrees__.entrySet() ) {
          ISname = (ImmutableName) entry.getKey();
          result.add( (ImmutableName)ISname.clone());
      } // for ite

      ISname = null;
      return result;
  }  
  
  /**
  * Returns the map of all the registered IS nametrees.
  * @return map of immutable and no-cloned NameISTrees. Never null.
  */
  public Map<ImmutableName, ImmutableNameISTree> getM_ISTrees() {
      Map<ImmutableName, ImmutableNameISTree> result = new HashMap(INITIAL_CAPACITY);
      ImmutableName ISname;

      for (Map.Entry entry : m_NameTrees__.entrySet() ) {
          ISname = (ImmutableName) entry.getKey();
          result.put(
                  (ImmutableName)ISname.clone(), 
                  (ImmutableNameISTree)entry.getValue() );
      } // for ite

      ISname = null;
      return result;
  } // getM_ISTrees()

  /**
  * Returns an IS nametree.
  * @param _name name of the IS; not null
  * @return the nameTree. May be null or an empty nameTree.
  */
  public ImmutableNameISTree getISTree(ImmutableName _name) {
      if (!m_NameTrees__.containsKey(_name)) return null;
      return (ImmutableNameISTree) m_NameTrees__.get(_name);
  } // getISTree()
  
  /**
  * Called by newEorACS() and ActionNewACSsyst to add an information system,
  * to save it in the NameTrees map through a new NameISTree with a passive root
  * which is the IS name. Fires the property change 'OneInformationSystem'.
  * @param _name name of a new IS. Not null.
  * @param _is_hub is true if the IS is a node hub, which means that for this node,
  *    any direct child may have exchanges with any other direct child, if the rights allow them.
  * @exception BaseError if the IS name is already known as ACS, IS or component,
  *   or FileNotFoundException and IOException.
  */
  public final void addIS(ImmutableName _name, boolean _is_hub) throws BaseError {
    if (_name == null || getISofName(_name) != null )
       throw new BaseError("in BaseManagerImpl, the IS name is null or already known.");

    try {
        m_NameTrees__.put((ImmutableName)_name.clone(), new NameISTreeImpl(_name, _is_hub) );
    } catch (CreateError c) {
        throw new BaseError(c.toString());
    }
    
    saveNameTrees();
    
    // fire an event
    try {
        firePropertyChange("OneInformationSystem", null, _name );
     } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.addIS(), firePropertChange() throws:");
        e.printStackTrace();
    }     
  } // addIS()
  
  /**
  * Called by removeACS() to remove an information system, and to save the new
  * NameTrees map. Removes all the ACS of the IS.
  * This is a public method only for testing.
  * Fires the property changes 'OneInformationSystem', and 'OpenACS' for each closed ACS.
  * @param _name of an IS. Never null.
  * @exception BaseError if the IS name is not known, or
  *  FileNotFoundException and IOException.
  */
  public final void removeIS(ImmutableName _name) throws BaseError {
    if (_name == null || m_NameTrees__.get(_name) == null)
       throw new BaseError("in BaseManagerImpl, the IS name is null or unknown.");
    
    NameISTreeImpl iSTree = (NameISTreeImpl) m_NameTrees__.get(_name);
    List<ImmutableName> l_children = iSTree.getL_AllChildren(_name);
    
    // remove the ACS in any order - removing of composite ACS is unuseful
    for (ImmutableName nodename : l_children) {
        if (iSTree.getType(nodename).equals(ISFactoryUtilityImpl.ACS_NODE)) {
            removeACS(nodename);
        } // if
    } // for
    
    m_NameTrees__.remove(_name);
    
    saveNameTrees();
    
    try {
        firePropertyChange("OneInformationSystem", _name, null );
     } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.removeIS(), firePropertChange() throws:");
        e.printStackTrace();
    }

    iSTree.finalizeForUser();
    iSTree = null;
    l_children = null;
  } // removeIS()
   
  /**
  * Called by ACSFactory to add an ACS or subACS to the IS nametrees, after the call
  * to newEorACS(), and by importACS().
  * Do not handle the Tied-Child, Composite, Alternate relations of the new ACS or subACS
  * (see addSecondaryPropertiesToAcsOrSubNode()).
  * Conditional firing of a property change 'OneInformationSystemNode'
  * with the old and the new nameISTree.
  * @param _ISname the IS name
  * @param _parent is a node in the nametree. May be null.
  * @param _name is the new ACS or subACS name, which is put at the end
  *   of the children array of _parent
  * @param _node_type is the _name type; it is an allowed type for the _parent 
  *     children, following the NameTreeImpl policy.
  * @param _is_hub is true if the child is a node hub, which means that for this node,
  *    any direct child may have exchanges with any another direct child.
  * @param _visible_space is the visible space for the ACS. Never null.
  * @param _server true if the new ACS is an authorization server.
  *         Not used for a subACS, for which this method sets it to false.
  * @param _fireChange true to fire a property change 'OneInformationSystemNode'.
  * To set to false if Tied-Child, Composite or Alternate relations have still to be set.
  * @exception BaseError if _parent is not a node of this nametree,
  *     if _child is null, is already a nametree node, if 
  *     the parent does not allow this type for its children, 
  *     wrong node type, wrong visible space,
  *     or if its length is greater than 12,
  *     or FileNotFoundException and IOException.
  */
  protected final void addAcsOrSubAcsNodeToIS(ImmutableName _ISname,  
      Name _parent, ImmutableName _name, String _node_type, boolean _is_hub,
      String _visible_space, boolean _server, boolean _fireChange)
      throws BaseError {
      
        if (_ISname == null || getISofName(_ISname) == null ||
                (_parent != null &&  getISofName(_parent) == null) ||
                _name == null)
           throw new BaseError("in BaseManagerImpl.addAcsNodeToIS(), "+
                   "a name is null or unknown.");

        // variables
        //NameISTreeImpl  old = (NameISTreeImpl)((NameISTreeImpl) m_NameTrees__.get(_ISname)).clone();
        NameISTreeImpl  iSTree = (NameISTreeImpl) m_NameTrees__.get(_ISname);
        String[]        l_links = null;
        
        synchronized(m_NameTrees__) {        
            if (_node_type.equals(ISFactoryUtilityImpl.ACS_NODE) && _server) {
                l_links = new String[] {ISFactoryUtilityImpl.IS_AUTHORIZATION_SERVER};
            } else
                l_links = new String[0];
            // checks _node_type
            iSTree.addChild(_parent, _name, _node_type, _is_hub, l_links);
            iSTree.setExtendedVisibleSpace(_name, _visible_space);

            // saves before the opening of the node
            saveNameTrees();
            iSTree.setOpen(_name, true);
        } // synchronized(m_NameTrees__)

                            // for testing and tracking
                        //System.out.println("BaseManagerImpl.addAcsOrSubAcsNodeToIS(), for "+
                          //      _name);
                        //System.out.println("BaseManagerImpl.addAcsOrSubAcsNodeToIS(), IS names:");
                        //for (ImmutableName name : m_NameTrees__.keySet()) {
                          //  System.out.println("------ "+name);
                        //}
                        //System.out.println("BaseManagerImpl.addAcsOrSubAcsNodeToIS(), "+
                          //      "new node name: "+_name);
                        //System.out.println("BaseManagerImpl.addAcsOrSubAcsNodeToIS(), "+
                          //      "getISofName(_name): "+getISofName(_parent));

        try {
            if (_fireChange)
                firePropertyChange("OneInformationSystemNode", null, _name);
         } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.addAcsNodeToIS(), firePropertChange() throws:");
            e.printStackTrace();
        }

        //old = null;
        iSTree = null;
        l_links = null;
  } // addAcsOrSubAcsNodeToIS()
    
  /**
  * Called by ACSFactory to add some Tied-Child, Composite, Alternate relations 
  * to an ACS or subACS in the IS nametree, after the call to addAcsOrSubAcsNodeToIS().
  * This method is independant to newEorACS() which is called by the ACS constructor
  * to register the ACS in the ACS list. It may be  called several times, with varied arguments. <P>
  * The Composite relation is defined in the mandatory paths of _name, and
  * _name is added as subcomposite to the ImmutableNameISTree.Area which defines the Composite. <P>
  * The Alternate relation is defined through the two last method arguments, which must
  * be coherent. The items in _l_alternates are ACS/SubACS/Composite ACS/Composite SubACS nodes
  * in the IS nametree. Almost all the couples are allowed in the Alternate relation,
  * excepted when a ground ACS has an alternate SubACS or Composite SubACS. <P>
  * The _m_alternateProperties keys must be strictly the _l_alternates items. 
  * The _m_alternateProperties value for a key is a String array 
  * with the ISFactoryUtilityImpl constants: ALTERNATE_TYPE at the index 0 (mandatory),
  * then, without any specific order in the array, some right types (mandatory),
  * OPTIONAL_ALTERNATE (optional), OPTIONAL_GROUND (optional).
  * The argument _m_alternateProperties is stored in specific links in the IS nametree.
  * A double name derived from _name and the alternate name is used as key for the link.
  * These specific keys are not handled by NameISTreeImpl, so they are managed by this method
  * and the class ISNodeImpl. <P>
  * It is possible to change, through this method, some of the 3 relations
  * for a given ACS or subACS. There are some dependencies among the relations,
  * and there are checked by this method which throws an exception if a constraint
  * is not fulfilled. The dependencies between relations are listed hereafter: <P>
  * - a subcomposite cannot be a ground for alternates. <P>
  * - a subcomposite ACS may be a Tied-Child of its parent, but it is not workable
  * until all the subcomposites of the Composite are opened. <P>
  * The ACS node is always opened by this method. <P>
  * May fire a property change 'OneInformationSystemNode' with the new node name.
  * @param _ISname the IS name
  * @param _name is the ACS or subACS node name in the current IS nametree 
  * @param _is_tied_child true if the new ACS is a tied child to its parent.
  * Not used for a subACS, for which this method sets it to false.
  * @param _composite name of the Composite ACS or Composite SubACS for which
  * this ACS or subACS is a subcomposite. This subcomposite is added
  * to the Composite list. May be null.
  * @param _l_alternates are the alternate names in the same IS for this ACS or subACS
  * as a ground. No more than 4 items. May be null or empty. Each argument is updated.
  * @param _m_alternateProperties where each name in _l_alternates is a map key,
  * and the map value is an array of the properties for the alternate. Must be
  * not null if _l_alternates is not null nor empty, with the good format.
  * @param _fireChange true to fire a property change 'OneInformationSystemNode'.
  * To set to false if it is not the last call to this method in the same sequence.
  * @exception BaseError if a constraint is not fulfilled,
  * or FileNotFoundException and IOException.
  */
  protected final void addSecondaryPropertiesToAcsOrSubNode(ImmutableName _ISname,
      ImmutableName _name, 
      boolean _is_tied_child,  ImmutableName _composite, ImmutableName[] _l_alternates,
      Map<ImmutableName, String[]> _m_alternateProperties, boolean _fireChange)
      throws BaseError {
                            // for testing and tracking
                            //System.out.println("BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(),"+
                              //      " for "+_name);

        if (_ISname == null || getISofName(_ISname) == null || _name == null ||
                getISofName(_name) == null || !getISofName(_name).equals(_ISname))
           throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                   "a name is null or unknown in the IS.");
        if (_composite != null && (getISofName(_composite) == null ||
                !getISofName(_composite).equals(_ISname) ))
           throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                   "the composite name is unknown in the IS.");
        if (_composite != null && _l_alternates != null && _l_alternates.length != 0)
           throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                   "a subcomposite cannot be a ground for alternates.");
        
        // variables
        //NameISTreeImpl  old = (NameISTreeImpl)((NameISTreeImpl) m_NameTrees__.get(_ISname)).clone();
        NameISTreeImpl  iSTree = (NameISTreeImpl) m_NameTrees__.get(_ISname);
        boolean         error = false;
        String[]        l_links = null;
        ImmutableName   doubleName = null;
        int             size = 0;
        
        if (!iSTree.getType(_name).equals(ISFactoryUtilityImpl.ACS_NODE) &&
                !iSTree.getType(_name).equals(ISFactoryUtilityImpl.SUB_ACS_NODE) )
           throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                   "_name is not an ACS or subACS node.");
        
        if (_composite != null) {
            if (iSTree.getType(_name).equals(ISFactoryUtilityImpl.ACS_NODE) &&
                    !iSTree.getType(_composite).equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) )
               throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                       "the composite has not the right type.");
            if (iSTree.getType(_name).equals(ISFactoryUtilityImpl.SUB_ACS_NODE) &&
                    !iSTree.getType(_composite).equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE) )
               throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                       "the composite has not the right type.");
            if (iSTree.getMandatoryPaths(_name) != null)
               throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                       "the composite of _name is already set.");
            if (iSTree.getAlternates(_name) != null) {
               for (ImmutableName name : iSTree.getAlternates(_name)) {
                   if (name.component(0).equals(ISFactoryUtilityImpl.IS_ALTERNATE_OF)) {
                       error = true;
                       break;
                   } // if
               } // for i
               if (error)
                   throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                       "_name cannot be a subcomposite since it is already an alternate.");
            } // if  (iSTree.getAlternates(_name) != null)
            // do not test if there are paths of more than one item in the returned arrays
            if (iSTree.getMandatoryPaths(_name) != null)
                   throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                       "_name is already a subcomposite.");
        } // if  (_composite != null)
        
        if (_l_alternates != null && _l_alternates.length != 0) {
            for (int i = 0; i < _l_alternates.length; i++) {
                if (getISofName(_l_alternates[i]) == null ||
                        !getISofName(_l_alternates[i]).equals(_ISname) ||
                        ( iSTree.getType(_name).equals(ISFactoryUtilityImpl.ACS_NODE) &&
                            (iSTree.getType(_l_alternates[i]).equals(ISFactoryUtilityImpl.SUB_ACS_NODE) ||
                            iSTree.getType(_l_alternates[i]).equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE)))
                    ) {
                    error = true;
                    break;
                } // if
            } // for i
            if (_m_alternateProperties == null) error = true;
            else {
                if (_m_alternateProperties.keySet().size() != _l_alternates.length)
                    error = true;
                // _m_alternateProperties keys are _l_alternates items
                for (int i = 0; i < _l_alternates.length; i++) {
                    if (!_m_alternateProperties.keySet().contains(_l_alternates[i])) {
                        error = true;
                        break;
                    }
                } // for i
            } // else
        } // if
        if (error)
            throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                    "an alternate is unknown, not from the same IS,"+
                    " has a (Composite) SubACS type while _name has the ACS type, "+
                    "or it is not a key in the properties map.");
        
        // the _m_alternateProperties value is String[]: 
        // <ALTERNATE_TYPE>, 
        // 1, 2 or 3 RightsTypes,
        // (option) OPTIONAL_ALTERNATE
        // (option) OPTIONAL_GROUND
        for (Map.Entry entry : _m_alternateProperties.entrySet()) {
            l_links = (String[])entry.getValue();
            if (l_links.length < 2 || l_links.length > 6) {
                error = true;
                break;
            }
            if (!ISFactoryUtilityImpl.getAlternateTypes().contains(l_links[0])) {
                error = true;
                break;
            }
        } // for entry
        if (error)
            throw new BaseError("in BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), "+
                    "the length of the alternate properties or the first value is not correct.");
        
        
        // end of controls
        
        synchronized(m_NameTrees__) {        
            
            if (iSTree.getType(_name).equals(ISFactoryUtilityImpl.ACS_NODE) && _is_tied_child) {
                l_links = iSTree.getLinks(_name);
                if (l_links.length != 0) {
                    l_links = new String[l_links.length + 1];
                    System.arraycopy(iSTree.getLinks(_name), 0, l_links, 0, l_links.length - 1);
                    l_links[l_links.length - 1] = ISFactoryUtilityImpl.IS_TIED_CHILD;
                } else
                    l_links = new String[] {ISFactoryUtilityImpl.IS_TIED_CHILD};
                iSTree.setLinks(_name, l_links);
            } // if
             
            if (_composite != null) {
                iSTree.addMandatoryPath(_name, new ImmutableName[] {_composite});
                iSTree.addNodeToArea(_composite, _name);
            } // if
            
            if (_l_alternates != null && _l_alternates.length != 0) {
                
                for (int i = 0; i < _l_alternates.length; i++) {
                    iSTree.addAlternate(_name, _l_alternates[i]);
                    
                    doubleName = ((Name)((NameImpl)_name).clone()).extended(_l_alternates[i]);
                    l_links = iSTree.getLinks(doubleName);
                    size = l_links.length;
                    
                    if (size != 0) {
                        l_links = new String[size + 
                                ((String[])_m_alternateProperties.get(_l_alternates[i])).length];
                        System.arraycopy(iSTree.getLinks(doubleName), 0, l_links, 0, l_links.length - size);
                    } else
                        l_links = new String[((String[])_m_alternateProperties.get(_l_alternates[i])).length];
                    
                    System.arraycopy((String[])_m_alternateProperties.get(_l_alternates[i]), 0, 
                            l_links, size, l_links.length - size);
                    iSTree.setLinks(doubleName, l_links);                    
                    
                } // for i

            } // if  (_l_alternates != null)
            
            // save before the opening of the node
            saveNameTrees();
            iSTree.setOpen(_name, true);
        } // synchronized(m_NameTrees__)

        try {
            if (_fireChange)
                firePropertyChange("OneInformationSystemNode", null, _name);
         } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.addSecondaryPropertiesToAcsOrSubNode(), firePropertChange() throws:");
            e.printStackTrace();
        }

        //old = null;
        iSTree = null;
        l_links = null;
        doubleName = null;
  } // addSecondaryPropertiesToAcsOrSubNode()
  
  /**
  * Called by ACSFactory to add a Composite ACS or a Composite subACS to the IS nametrees.
  * This method has to be called before the creation of the subcomposites in the IS nametree. <P>
  * It implements a Composite through the ImmutableNameISTree.Area property.
  * The property is_hub_for_children is always true for a composite.
  * The extended visible space of a composite is always NameISTreeImpl.ALL_THE_IS_SPACE.
  * A composite is never a tied child of its parent. Its working depends on
  * the opening of all its subcomposites, and this is managed by this class at each
  * opening and closing of an ACS.
  * The standard information systems IPV4/6_ADDRESSES_SPACE do not allow a composite node.
  * The Composite node is not opened by this method. <P>
  * Firing of a property change 'OneInformationSystemNode' with the new node name.
  * @param _ISname the IS name
  * @param _name the last component of the new Composite ACS or subACS name
  * @param _node_type is the _name type; it is an allowed type for the _parent 
  *     children, following the ISImpl policy.
  * @param _is_authorization_server true if the new Composite ACS is an authorization server.
  *         Not used for a Composite subACS, for which this method sets it to false.
  * @exception BaseError if _parent is not a node of this nametree,
  *     if _child is null, is already a nametree node, if 
  *     the parent does not allow this type for its children, wrong node type, 
  *     or FileNotFoundException and IOException.
  */
  protected final void addCompositeNodeToIS(ImmutableName _ISname,
      String _name, String _node_type, boolean _is_authorization_server)
      throws BaseError {
        // does not control the composite in the nametree
        if (_ISname == null || getISofName(_ISname) == null || _name == null)
           throw new BaseError("in BaseManagerImpl.addCompositeNodeToIS(), "+
                   "a name is null or unknown.");
        if (_ISname.equals(new NameImpl(ISFactoryUtilityImpl.IPV4_ADDRESSES_SPACE)) ||
                _ISname.equals(new NameImpl(ISFactoryUtilityImpl.IPV6_ADDRESSES_SPACE)))
           throw new BaseError("in BaseManagerImpl.addCompositeNodeToIS(), "+
                   "the two standard information systems IPv4/6 do not allow a composite node.");
        if (_node_type == null || (!_node_type.equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) &&
                !_node_type.equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE)))
           throw new BaseError("in BaseManagerImpl.addCompositeNodeToIS(), "+
                   "it is not a composite type.");

        //NameISTreeImpl  old;
        NameISTreeImpl  iSTree;
        ImmutableName   name = ((NameImpl)_ISname.clone()).extended(_name);
        String[]        l_links = null;
        
        synchronized(m_NameTrees__) {        
            //old = (NameISTreeImpl)((NameISTreeImpl) m_NameTrees__.get(_ISname)).clone();
            iSTree = (NameISTreeImpl) m_NameTrees__.get(_ISname);
            
            if (_node_type.equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) &&
                    _is_authorization_server) {
                l_links = new String[] {ISFactoryUtilityImpl.IS_AUTHORIZATION_SERVER};
            } else
                l_links = new String[0];
            // check _node_type
            iSTree.addArea( name, _node_type, l_links);

            saveNameTrees();
            //iSTree.setOpenNode(_name, true);
            
        } // synchronized(m_NameTrees__)

        try {
            firePropertyChange("OneInformationSystemNode", null, name);
         } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.addCompositeNodeToIS(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }

        name = null;
        iSTree = null;
        l_links = null;
  } // addCompositeNodeToIS()  

  /**
  * Called by ACSFactoryImpl and importACS() to add a physical or logical component
  * to the IS nametrees, and by ActionNewACSyst.handleSelectionDialog().
  * This type of node has no extended visible space. Iit is always a hub.
  * The component parent name is the context of _name.
  * The two standard information systems IPv4/6 allow only a logical node for the two first levels.
  * The two first levels of an IPv4 address has to be a number lesser than 256.
  * Firing of a property change 'OneInformationSystemNode' with the new node name.
  * @param _ISname the IS name
  * @param _name is the new component name, which is put at the end
  *   of the children array of its parent
  * @param _node_type is the _name type; it is an allowed type for the _parent
  *     children, following the ISImpl policy.
  * @exception BaseError if the parent is not a node of this nametree,
  *     if _child is null, is already a nametree node, if
  *     the parent does not allow this type for its children,
  *     wrong node type, or FileNotFoundException and IOException.
  */
  public final void addComponentNodeToIS(ImmutableName _ISname,
      ImmutableName _name, String _node_type) throws BaseError {
        // do not control the component in the nametree
        if (_ISname == null || getISofName(_ISname) == null || _name == null ||
                getISofName(_name.context()) == null )
           throw new BaseError(" "+
                   "A name is null or unknown for the new component in the IS "+_ISname);
        if (_node_type == null ||
                (!_node_type.equals(ISFactoryUtilityImpl.PHYSICAL_COMPONENT_NODE) &&
                    !_node_type.equals(ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE)))
           throw new BaseError(" "+
                   "A new component in an IS has to be a logical node or a physical node.");
        // check the IPv4/6 information systems
        if ((_ISname.equals(new NameImpl(ISFactoryUtilityImpl.IPV4_ADDRESSES_SPACE)) ||
               _ISname.equals(new NameImpl(ISFactoryUtilityImpl.IPV6_ADDRESSES_SPACE))) &&
                !_node_type.equals(ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE) &&
                _name.noOfComponents() < 4 )
           throw new BaseError(" "+
                   "The two standard information systems IPv4/6 allow only a logical node "+
                   "at the two first levels.");

        if (_ISname.equals(new NameImpl(ISFactoryUtilityImpl.IPV4_ADDRESSES_SPACE))) {
               if (_name.noOfComponents() < 4 || _name.lastComponent().length() > 3)
                    throw new BaseError(" "+
                        "The node name for the two first levels of an IPv4 address is too large.");

               Integer in = null;
               try {
                   in = new Integer(_name.lastComponent());
               }
               catch (NumberFormatException n) {
                   throw new BaseError(" "+
                        "The node name for the two first levels of an IPv4 address is not a number.");
               }
               
               if (in.intValue() > 255) {
                   throw new BaseError(" "+
                        "The node name for the two first levels of an IPv4 address is a number greater than 255.");
               } // if
               in = null;
        } // if (_ISname.equals(new NameImpl(NameISTreeImpl.IPV4_ADDRESSES_SPACE)))

        NameISTreeImpl  iSTree;

                                    // for testing and tracking
                                //System.out.println("BaseManagerImpl.addComponentNodeToIS(), "+
                                  //      "for the new component: "+_name+", the current IS are:");
                                //for (ImmutableName name : m_NameTrees__.keySet()) {
                                  //  System.out.println("------ "+name);
                                //}


        synchronized(m_NameTrees__) {
            //old = (NameISTreeImpl)((NameISTreeImpl) m_NameTrees__.get(_ISname)).clone();
            iSTree = (NameISTreeImpl) m_NameTrees__.get(_ISname);

            // checks _node_type
            iSTree.addChild(_name.context(), _name, _node_type, true, new String[0]);
            // no ExtendedVisibleSpace(_name, NameISTreeImpl.ALL_THE_IS_SPACE);

            saveNameTrees();
            // the node is not open

        } // synchronized(m_NameTrees__)

        try {
            firePropertyChange("OneInformationSystemNode", null, _name);
         } catch (Exception e) {
            //ARoad0.Gui1.DialogBox.showAlarmMessage(
              //"There is an exception after the base updating, from outside the database."+
              //"You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.addComponentNodeToIS(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }

        iSTree = null;
  } // addComponentNodeToIS()

  /**
  * Called to remove an ACS to the IS nametrees and save it, and to remove also
  * all its direct and indirect children, including its subACS and all the ACS and subACS
  * which are its children. <P>
  * This method is called only by removeACS().
  * Handles the Tied-Child, Composite, Alternate relations of the ACS and its subACS. <P>
  * Some of the Alternate properties are stored in specific links in the IS nametree.
  * A double name derived from _name and the alternate name is used as key for the link.
  * These specific keys are not handled by NameISTreeImpl, so they are managed by this method. <P>
  * Firing of a property change 'OneInformationSystemNode' with the old node name.
  * @param _ISname the IS name
  * @param _name is the ACS name
  * @exception BaseError if a name is null or unknown, if it is not an ACS node,
  *     or FileNotFoundException and IOException.
  */
  private final void removeAcsNodeToIS(ImmutableName _ISname, ImmutableName _name)
          throws BaseError {

                            // for testing and tracking
                        //System.out.println("BaseManagerImpl.removeAcsNodeToIS(), for the ACS:"+_name);
                        //for (ImmutableName name : m_NameTrees__.keySet()) {
                          //  System.out.println("------ "+name);
                        //}

        if (_ISname == null || getISofName(_ISname) == null || _name == null)
           throw new BaseError("in BaseManagerImpl.addAcsNodeToIS(), "+
                   "a name is null or unknown.");

        // variables
        //NameISTreeImpl          old = (NameISTreeImpl)((NameISTreeImpl)
          //                              m_NameTrees__.get(_ISname)).clone();
        NameISTreeImpl          iSTree = (NameISTreeImpl) m_NameTrees__.get(_ISname);
        List<ImmutableName>     l_alternates = iSTree.getAlternates(_name);
        ImmutableName           doubleName = null;
        List<ImmutableName>     l_grounds = iSTree.getGroundsOfAlternate(_name);
        List<ImmutableName>     l_children = iSTree.getL_AllChildren(_name);
        List<ImmutableName[]>   l_paths = iSTree.getMandatoryPaths(_name);

        if (!iSTree.getType(_name).equals(ISFactoryUtilityImpl.ACS_NODE) )
           throw new BaseError("in BaseManagerImpl.removeAcsNodeToIS(), "+
                   "_name is not an ACS node.");

        synchronized(m_NameTrees__) {
            // removes first the Alternate and Composite secondary properties
            if (l_alternates != null) {
                for (ImmutableName name : l_alternates) {
                    doubleName = ((Name)((NameImpl)_name).clone()).extended(name);
                    iSTree.setLinks(doubleName, null);
                } // for i
            } // if  (_l_alternates != null)
            if (l_grounds != null) {
                for (ImmutableName name : l_grounds) {
                    doubleName = ((Name)((NameImpl)name).clone()).extended(_name);
                    iSTree.setLinks(doubleName, null);
                } // for i
            } // if  (_l_alternates != null)
            if (l_paths != null) {
                for (ImmutableName[] array : l_paths) {
                    for (int i = 0; i < array.length; i++) {
                        if (iSTree.getType(array[i]).equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) ||
                            iSTree.getType(array[i]).equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE) )
                                    iSTree.removeNodeToArea(array[i], _name);
                    } // for i
                } // for array
            } // if (l_paths

            // same work for the indirect and direct children which are ACS and subACS
            for (ImmutableName childName : l_children) {
                if (!iSTree.getType(childName).equals(ISFactoryUtilityImpl.ACS_NODE) &&
                        !iSTree.getType(childName).equals(ISFactoryUtilityImpl.SUB_ACS_NODE))
                    continue;
                l_alternates = iSTree.getAlternates(childName);
                l_grounds = iSTree.getGroundsOfAlternate(childName);
                l_paths = iSTree.getMandatoryPaths(_name);
                if (l_alternates != null) {
                    for (ImmutableName name : l_alternates) {
                        doubleName = ((Name)((NameImpl)childName).clone()).extended(name);
                        iSTree.setLinks(doubleName, null);
                    } // for i
                } // if  (_l_alternates != null)
                if (l_grounds != null) {
                    for (ImmutableName name : l_grounds) {
                        doubleName = ((Name)((NameImpl)name).clone()).extended(childName);
                        iSTree.setLinks(doubleName, null);
                    } // for i
                } // if  (_l_alternates != null)
                if (l_paths != null) {
                    for (ImmutableName[] array : l_paths) {
                        for (int i = 0; i < array.length; i++) {
                            if (iSTree.getType(array[i]).equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) ||
                                iSTree.getType(array[i]).equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE) )
                                        iSTree.removeNodeToArea(array[i], _name);
                        } // for i
                    } // for array
                } // if (l_paths
            } // for childName

            // updates iSTree.l_alternates through removeLeaf()
            iSTree.removeNodeAndSubTree(_name);

            saveNameTrees();
        } // synchronized(m_NameTrees__)

        try {
            firePropertyChange("OneInformationSystemNode", _name, null);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.removeAcsNodeToIS(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }

        //old = null;
        iSTree = null;
        l_alternates = null;
        l_grounds = null;
        l_children = null;
        doubleName = null;
  } // removeAcsNodeToIS()

  /**
  * Called to remove a Composite ACS or a Composite subACS to the IS nametrees.
  * This method handles the ImmutableNameISTree.Area property, and
  * it removes the subcomposite property of the subcomposites.
  * For a Composite ACS, the subcomposites are also removed from the IS nametree.
  * For a Composite subACS, the subcomposites remain in the IS nametree since their ACS
  * stay there, but the subcomposites will be unworkable.
  * Firing of a property change 'OneInformationSystemNode' with the old node name.
  * @param _ISname the IS name
  * @param _name is the Composite ACS or Composite subACS name
  * @exception BaseError if _parent is not a node of this nametree,
  *     if _name is null, or FileNotFoundException and IOException.
  */
  public final void removeCompositeNodeToIS(ImmutableName _ISname,
      ImmutableName _name) throws BaseError {

        if (_ISname == null || getISofName(_ISname) == null || _name == null)
           throw new BaseError("in BaseManagerImpl.removeCompositeNodeToIS(), "+
                   "a name is null or unknown.");

        //NameISTreeImpl  old = (NameISTreeImpl)((NameISTreeImpl) m_NameTrees__.get(_ISname)).clone();
        NameISTreeImpl  iSTree = (NameISTreeImpl) m_NameTrees__.get(_ISname);
        List<ImmutableName> l_subs = iSTree.getL_NodesOfArea(_name);

        // Each subcomposite subACS keeps its mandatory path to _name,
        // to tell this subACS is not workable.
        // The subcomposite ACS is removed, and this operation updates the composite.
        if ( iSTree.getType(_name).equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) ) {
           for (ImmutableName name : l_subs) {
                removeAcsNodeToIS(_ISname, name);
           } // for name
        } // if

        List<ImmutableName> l_s = null;

        synchronized(m_NameTrees__) {
            l_s = iSTree.getGroundsOfAlternate(_name);
            if (l_s != null) {
                for (ImmutableName ground : l_s) {
                    iSTree.removeAlternate(ground, _name);
                } // for
            } // if
            l_s = iSTree.getAlternates(_name);
            if (l_s != null) {
                for (ImmutableName ground : l_s) {
                    iSTree.removeAlternate(_name, ground);
                } // for
            } // if
            
            iSTree.removeArea(_name);

            saveNameTrees();

        } // synchronized(m_NameTrees__)

        try {
            firePropertyChange("OneInformationSystemNode", _name, null);
         } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.removeCompositeNodeToIS(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }

        //old = null;
        iSTree = null;
        l_s = null;
  } // removeCompositeNodeToIS()

  /**
  * Called to remove a physical or logical component to the IS nametrees, and all its
  * direct and indirect children, including its subACS and all the ACS and subACS
  * which are its children. <P>
  * Firing of a property change 'OneInformationSystemNode' with the old node name.
  * @param _ISname the IS name
  * @param _name is the new ACS or subACS name
  * @exception BaseError if a name is null or unknown, if it is not an ACS node,
  *     or FileNotFoundException and IOException.
  */
  public final void removeComponentNodeToIS(ImmutableName _ISname,  ImmutableName _name)
          throws BaseError {

        if (_ISname == null || getISofName(_ISname) == null || _name == null)
           throw new BaseError("in BaseManagerImpl.addAcsNodeToIS(), "+
                   "a name is null or unknown.");

        // variables
        //NameISTreeImpl  old = (NameISTreeImpl)((NameISTreeImpl) m_NameTrees__.get(_ISname)).clone();
        NameISTreeImpl  iSTree = (NameISTreeImpl) m_NameTrees__.get(_ISname);
        List<ImmutableName> l_children = iSTree.getL_AllChildren(_name);

        if (!iSTree.getType(_name).equals(ISFactoryUtilityImpl.PHYSICAL_COMPONENT_NODE) &&
                !iSTree.getType(_name).equals(ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE))
           throw new BaseError("in BaseManagerImpl.setPropertiesToACS(), "+
                   "_name is not a physical or logical component node.");

        synchronized(m_NameTrees__) {

            // works for the indirect and direct children which are ACS and subACS
            for (ImmutableName childName : l_children) {
                if (iSTree.getType(childName).equals(ISFactoryUtilityImpl.ACS_NODE) ) {
                                removeACS(childName);
                                //removeAcsNodeToIS(_ISname, childName);
                } //else if (iSTree.getType(childName).equals(ISFactoryUtilityImpl.SUB_ACS_NODE))
                    //            removeAcsNodeToIS(_ISname, childName);
            } // for childName

            // update iSTree.l_alternates, and all the children including the subcomponents
            iSTree.removeNodeAndSubTree(_name);

            saveNameTrees();

        } // synchronized(m_NameTrees__)

        try {
            firePropertyChange("OneInformationSystemNode", _name, null);
         } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.removeComponentNodeToIS(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }

        //old = null;
        iSTree = null;
        l_children = null;
  } // removeComponentNodeToIS()

  /**
  * Get the Composite ACS and a Composite subACS of an IS nametrees.
  * @param _ISname the IS name
  * @return the composite node names. Never null. Empty if _ISname is unknown.
  */
  public final Set<ImmutableName> getL_CompositeNodesOfIS(ImmutableName _ISname) {

        if (_ISname == null || getISofName(_ISname) == null )
           return Collections.EMPTY_SET;

        return ((NameISTreeImpl) m_NameTrees__.get(_ISname)).getL_AreaNames();
  } // getL_CompositeNodesOfIS()

    /**
     * Save the NameTrees map in the file 'l_metarights_v0.acr'.
     * Called by all the update methods for the nametrees.
     * @exception BaseError if FileNotFoundException and IOException.
     */
  private final void saveNameTrees() throws BaseError {
      try {
          FileOutputStream fos = new FileOutputStream(
            "m_nametrees_v0.acr");
          ObjectOutputStream out = new ObjectOutputStream(fos);
          out.writeObject( m_NameTrees__ );
          out.close();
          fos = null;
          out = null;
      } catch (FileNotFoundException f) {
          throw new BaseError("in saveNameTrees(), FileNotFoundException: "+f.toString());
      }
      catch (SecurityException s) {
          throw new InternalError(
            "BaseManagerImpl.saveNameTrees() fails for security: "+s.toString());
      }
      catch (IOException i) {
          throw new BaseError("in saveNameTrees() call, IOException: "+i.toString());
      }

  } // saveNameTrees()

    /**
     * Fills up the list l_AcsAddonDescriptors__ with the ACS addon editors and names.
     * The form of a value is 'Editor: editor_name - AcsAddon: addon_name'.
     * Saves the ACS addon editors and names in the file 'l_acsaddon_v0.acr',
     * that is never read by the program.
     * Scans the packages under the directory 'ARoad0.AcsAddon'
     * to update and to save the list of ACS addons.
     * The 4 standard directories (gBase, gBaseBeanInfo, gBaseInterface, gWork)
     * of an AcsAddon have to be there, even if one may be empty.
     * The standard file AcsFactoryXXX has to be in the directory 'gWork',
     * where 'XXX' is the AcsAddon package name.
     * Caution: two ACS from the same AcsAddon may have different ACS policy checksums
     * if one has been derived from another one, with changes in the ACS policy.
     * Uses Desktop.AR_URI which has to be intialized.
     * Called by the constructor.
     * @exception BaseError if FileNotFoundException and IOException, or if an AcsAddon
     * have not the 4 standard directories (gBase, gBaseBeanInfo, gBaseInterface, gWork).
     */
  private final void updateAndSaveAcsAddonDescriptors() throws BaseError {

      File      AcsAddon = new File(Desktop.AR_URI.resolve("AcsAddon"));
      File[]    l_children = null;
      File[]    l_children2 = null;
      File[]    l_children3 = null;
      String[]  l_child = null;
      boolean   error = false;
      boolean   error2 = false;
      boolean   error3 = false;
      boolean   error4 = false;
      boolean   find = false;
      String    message = null;
      List<String>  l_packages = Arrays.asList(
                                    new String[] {"gBase", "gBaseBeanInfo",
                                                    "gBaseInterface", "gWork"});

      if (AcsAddon.exists()) {
          l_children = AcsAddon.listFiles();

                                        // for testing and tracking
                                     //System.out.println(
                                       //      "BaseManagerImpl.updateAndSaveAcsAddonDescriptors(),"+
                                         //       " process the directory "+AcsAddon.getName()+
                                           //     " with children:");
                                     //for (int i = 0; i < l_children.length; i++) {
                                       //  System.out.println("---"+l_children[i].getName());
                                     //} // i

          for (int i = 0; i < l_children.length; i++) {
              if (!l_children[i].isDirectory() )
                  continue;

              l_children2 = l_children[i].listFiles();

                                        // for testing and tracking
                                     //System.out.println(
                                       //      "BaseManagerImpl.updateAndSaveAcsAddonDescriptors(),"+
                                         //       " process the directory "+l_children[i].getName()+
                                           //     " with children:");
                                     //for (int j = 0; j < l_children2.length; j++) {
                                       //  System.out.println("-------"+l_children2[j].getName());
                                     //} // j


              for (int j = 0; j < l_children2.length; j++) {
                  if (!l_children2[j].isDirectory() )
                      continue;

                                        // for testing and tracking
                                     //System.out.println("BaseManagerImpl.updateAndSaveAcsAddonDescriptors(),"+
                                     //" process the directory "+l_children2[j].getName());

                  l_AcsAddonDescriptors__.add(
                          "Editor: "+l_children[i].getName()+
                          " - AcsAddon: "+l_children2[j].getName());

                  l_children3 = l_children2[j].listFiles();

                  if (l_children3.length != 4) {
                      error = true;
                      message = "Editor: "+l_children[i].getName()+
                          " - AcsAddon: "+l_children2[j].getName();
                      break;
                  } // if

                  for (int k = 0; k < l_children3.length; k++) {
                      if (!l_children3[k].isDirectory() ||
                              !l_packages.contains(l_children3[k].getName())) {
                          error2 = true;
                          message = "Editor: "+l_children[i].getName()+
                          " - AcsAddon: "+l_children2[j].getName()+
                                  ", forbidden package: "+l_children3[k].getName();
                          break;
                      } // if

                      // the class AcsFactoryXXX is mandatory
                      if (l_children3[k].getName().equals("gWork")) {
                          find = false;
                          l_child = l_children3[k].list();

                          for (int m = 0; m < l_child.length; m++) {
                              if (l_child[m].equals(
                                      "AcsFactory"+l_children2[j].getName()+".class")) {
                                  find = true;
                                  break;
                              } // if
                          } // for m

                          if (!find) {
                              error4 = true;
                              message = "Editor: "+l_children[i].getName()+
                                      " - AcsAddon: "+l_children2[j].getName()+
                                      ", do not find the class 'AcsFactory"+
                                      l_children2[j].getName()+
                                      ".class' in "+l_children3[k].getName();
                              break;
                          } // if
                      } // if (l_children3[k].getName().equals("gWork"))

                  } // for k

                  if (error2 || error4) break;
              } // for j

              if (error || error2 || error4) break;

          } // for i
      } // if (ARoad0.exists())
      else {

                            // for testing and tracking
                         //System.out.println("BaseManagerImpl.updateAndSaveAcsAddonDescriptors(),"+
                         //" the directory 'ARoad0/AcsAddon' is not found. No AcsAddons. "+
                           //      "Desktop.AR_URI: "+Desktop.AR_URI.toString());
      } // else

      if (error) {
          throw new BaseError("in updateAndSaveAcsAddonDescriptors() call, IOException: "+
                  "lack of the 4 standard packages for the ACSAddon "+message);
      }
      if (error2 || error4) {
          throw new BaseError("in updateAndSaveAcsAddonDescriptors() call, IOException: "+
                  message);
      }

      Collections.sort(l_AcsAddonDescriptors__);

      try {
          FileOutputStream fos = new FileOutputStream(
            "l_acsaddon_v0.acr");
          ObjectOutputStream out = new ObjectOutputStream(fos);
          out.writeObject( l_AcsAddonDescriptors__ );
          out.close();
          fos = null;
          out = null;
      } catch (FileNotFoundException f) {
          throw new BaseError("in updateAndSaveAcsAddonDescriptors(), FileNotFoundException: "+
                  f.toString());
      }
      catch (SecurityException s) {
          throw new InternalError(
            "BaseManagerImpl.updateAndSaveAcsAddonDescriptors() fails for security: "+
            s.toString());
      }
      catch (IOException i) {
          throw new BaseError("in updateAndSaveAcsAddonDescriptors() call, IOException: "+
                  i.toString());
      }

                            // for testing and tracking
                         //System.out.println("BaseManagerImpl.updateAndSaveAcsAddonDescriptors(),"+
                         //" the final list is:");
                        //for (String va : l_AcsAddonDescriptors__) {
                          //  System.out.println("------ "+va);
                        //}

      AcsAddon = null;
      l_children = null;
      l_children2 = null;
      l_children3 = null;
      l_packages = null;
      l_child = null;
  } // updateAndSaveAcsAddonDescriptors()
  
    //
    // -------- the ACSManager interface METHODS --------------
    //

    /**
    * Opens a registered and closed ACS from a file in the current directory.
    * Opens all the child nodes in the NameISTree which are not ACS.
    * Analyzes the need to open the closed Bridges, AclEntries and Privileges
    * in every open ACS and in this ACS, and the need to create the closed bridge links.
    * Calls setOpenNode() on the IS nametree, to open the direct and indirect children
    * if they are not ACS, or a child of a child ACS.
    * Creates the roots in the parent ACS which have been deleted at the ACS closing.
    * Fires the events 'OpenACS' and several 'OneInformationSystemNodeOpening' events.
    * @param _name ImmutableName of the ACS to open. Its file name is built
    *   with the two last components of the name, or with the single component
    * @exception BaseError if the ACS is not known, the ACS is open, the open
    *   ACS has not the same name, or ClassNotFoundException, IOException with
    *   invalidClassException if ACSImpl has changed.
    * @return the opened ACS after a setting of all the external references
     * to open ACS, and deleting all the references to closed ACS.
    */
  public ACSImpl openACS(ImmutableName _name) throws BaseError {
    // variables
    String              str = "ACS0";
    FileInputStream     fis = null;
    ObjectInputStream   in = null;
    ACSImpl             acs = null;
    NameISTreeImpl      iSTree = null;
    List<ImmutableName> l_childACSandCo = null;
    List<ImmutableName> l_childToOpen = null;

                    // for testing and tracking
                //System.out.println("BaseManagerImpl.openACS() for name: "+
                  //  _name);


    // tests
    if (_name == null || _name.lastComponent() == null) {
      throw new BaseError("in BaseManagerImpl.openACS(_name),"+
        " _name or _name.lastComponent() == null");
    }
    if (!l_Acs__.contains(_name) ) {
        throw new BaseError("in ACSManager.openACS(name) call,"+
          " the name is unknown");
    }
    if (m_OpenAcs__.containsKey(_name) ) {
        throw new BaseError("in ACSManager.openACS(name) call,"+
          " the acs is already open");
    }

    int index = 0;
    while (index < _name.noOfComponents()) {
        str = str.concat("_");
        str = str.concat(_name.component(index));
        index++;
    }

    // opens the file
    try {
        fis = new FileInputStream( str + ".acr");
        in = new ObjectInputStream(fis);
        acs = (ACSImpl) in.readObject();
        in.close();
    } catch (ClassNotFoundException f) {
        fis = null;
        in = null;
        acs = null;
        
        throw new BaseError("in BaseManagerImpl.openACS() call,"+
          " ClassNotFoundException");
    }
//      catch (SecurityException s) {
//        throw new InternalError(s.toString()+ "BaseManagerImpl.openEorACS() failed");
//      }
    catch (IOException i) {
        fis = null;
        in = null;
        acs = null;
        
        throw new BaseError("in BaseManagerImpl.openACS() call,"+
          " IOException" + i.toString());
    }

    if (!_name.equals(acs.getName()) ) {
        fis = null;
        in = null;
        acs = null;
        
        throw new BaseError("in BaseManagerImpl.openACS(_name) call,"+
          " _name not equals to acs.getName()");
    }

    // registers acs in the ACSManager before the call to the restoring methods
    Map oldOpenACS = new HashMap(m_OpenAcs__);
    synchronized(m_OpenAcs__){m_OpenAcs__.put( (ImmutableName)_name.clone(), acs);}

    // opens the nodes in the IS nametree
    //old = (NameISTreeImpl)((NameISTreeImpl) m_NameTrees__.get(acs.getISName())).clone();
    iSTree = (NameISTreeImpl) m_NameTrees__.get(acs.getISName());
    // noop on children
    iSTree.setOpen(_name, true);
    // intializes l_childACSandCo
    for (ImmutableName name : iSTree.getL_DirectChildren(_name)) {
        if (iSTree.getType(name).equals(ISFactoryUtilityImpl.ACS_NODE)) {
            if (l_childACSandCo == null) {
                l_childACSandCo = iSTree.getL_AllChildren(name);
                l_childACSandCo.add(name);
            }
            else {
                l_childACSandCo.addAll(iSTree.getL_AllChildren(name));
                l_childACSandCo.add(name);
            }
        } // if
    } // for name
    // The direct and indirect children are opened,
    // including the subACSs and the logical components,
    // if they are not a child ACS, or a child of a child ACS
    l_childToOpen = iSTree.getL_AllChildren(_name);

                    // for testing and tracking
                //System.out.println("in BaseManagerImpl.openACS(), l_childToOpen: "+
                  //  l_childToOpen);

    if (l_childACSandCo != null)
        l_childToOpen.removeAll(l_childACSandCo);
    try {
        for (ImmutableName name : l_childToOpen) {
            iSTree.setOnlyThisOpen(name, true);
            firePropertyChange("OneInformationSystemNodeOpening", null, name);
        } // for name
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.openACS(), firePropertChange() throws:");
        e.printStackTrace();
    }

    // handles parent ACS relations
    acs.setAcsRelationsInIS(false, false);
    // and other ACS relations
    restoreACSExternalLinks(acs);
    restoreACSInternalLinks(acs);

    // fires the events
    try {
        firePropertyChange("OpenACS",
            oldOpenACS, new HashMap( getEorM_OpenACS() ));
        firePropertyChange("OneInformationSystemNodeOpening", null, _name);

            // unuseful since the global structure tree is repainted at the first event
        //for (ImmutableName name : l_childToOpen) {
          //  firePropertyChange("OneInformationSystemNodeOpening", null, name);
        //} // for name
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.openACS(), firePropertChange() throws:");
        e.printStackTrace();
    }
        
    oldOpenACS = null;
    fis = null;
    in = null;
    iSTree = null;
    l_childACSandCo = null;
    l_childToOpen = null;

    return acs;
  } // openACS()

/**
* Imports an unknown ACS from a file in the current directory, creates it as a node
* in the right IS, opens and saves it. The saving updates the initial file of the imported ACS.
* If the IS is unknown, adds it to the list as a hub node.
* Compares the ACS parent name to the current nodes of the ACS IS.
* If the parent name does exist in the IS, registers the ACS under this node if
* there is no child having the same name.
* If there is no chain of nodes in the IS to match to the ACS parent name,
* a chain of nodes derived from the ACS name, is created with logical nodes.
* During the importation, two following IS nametree properties are missed. They are
* the Composite and Alternate relations to an ACS or subACS in the IS nametree.
* An imported ACS has no composite and no alternate, and it is not possible to edit
* them after the importation. The other properties of the ACS node in the source IS nametree
* are copied into the target IS nametree.
* <P>
* This method restores all the inner properties of the imported ACS, including
* its structure, behavior and ACSobjects. This includes the ACS roots, but they are
* applied only if the parent is an ACS. Otherwise, the roots are deleted during
* the import. If the parent is an ACS, it has to comply to the structure and behavior
* of the parent ACS source, so that the roots action may be applied without error.
* However, errors in the creation of roots do not stop the import operation.
* <P>
* This method never opens the external ACSObjects of the imported ACS,
* and its closed Bridges, AclEntries and linked Privileges.
* It rather returns a map containing the description of these closed objects.
* The user may choose, in the class ActionImportACS,
* to create some of these closed objects. The returned map described each Bridge,
* AclEntry or Privilege, including the rights. Even if they are ACS rights
* from the imported ACS, they are always fully defined. The import operation
* resets the links between the ACS rights and the current metarights of the running
* Access Road instance.
* Called by ACS_Facade.getImportedACS().
* Fires an event 'OpenACS' after the ACS opening.
* @see ARoad0.gDMak.ActionImportACS
* @param _name ImmutableName of the ACS to import. Its file name is built
*   with the two last components of the name, or with the single component
* @exception BaseError if the ACS name is known or null argument, 
*   or SecurityException, or the IS name of the imported acs or
*   its parent in the IS ImmutableNameTree are unknown, or wrong saving
* @exception DialogError if the open ACS has not the same name,
*   or ClassNotFoundException, IOException with
*   invalidClassException if ACSImpl has changed.
* @return the map of the closed external ACSObjects which have a Bridge,
* AclEntries or linked Privileges in this imported ACS. May be empty but not null.
*/
  public Map<String, List<String>> importACS(ImmutableName _name)
          throws BaseError, DialogError {
    // variables
    String              str = "ACS0";
    FileInputStream     fis = null;
    ObjectInputStream   in = null;
    FileOutputStream    fos = null;
    ObjectOutputStream  out = null;
    ACSImpl             acs = null;
    Object              obj = null;
    ImmutableName       iSname = null;
    NameISTreeImpl      iS = null;
    Name                currentName = null;
    Map                 oldOpenACS = null;
    Map<String, List<String>> m_result = new HashMap(INITIAL_CAPACITY*10);
    ArrayList<String>   l_values = new ArrayList(INITIAL_CAPACITY);

    // tests
    if (_name == null || _name.lastComponent() == null) {
      throw new BaseError("in BaseManagerImpl.importACS(_name),"+
        " _name or _name.lastComponent() == null");
    }
    if (l_Acs__.contains(_name) ) {
        throw new BaseError("in ACSManager.importACS(name) call,"+
          " the name is already known");
    }

    // try to open the file

    int index = 0;
    while (index < _name.noOfComponents()) {
        str = str.concat("_");
        str = str.concat(_name.component(index));
        index++;
    }

    try {
        fis = new FileInputStream(str + ".acr");
        in = new ObjectInputStream(fis);
        obj = in.readObject();
        in.close();
    } catch (ClassNotFoundException f) {
        fis = null;
        in = null;
        acs = null;
        // this detects an uncompatible version
        throw new DialogError(" The ACS is not imported. "+
                "A ClassNotFoundException has been thrown. This is probably due to "+
                "a missing AcsAddon in your Access Road directory. Access Road updates "+
                "its list of AcsAddons at each starting. Read the imported ACS "+
                "documentation to install the AcsAddon classes this ACS requires,"+
                " with the right version.");
    }
    catch (SecurityException s) {
        throw new BaseError(s.toString()+ " in BaseManagerImpl.importACS() failed");
    }
    catch (IOException i) {
        fis = null;
        in = null;
        acs = null;
        
        throw new DialogError(" The file 'ACS0_" + str +
                ".acr' is not found under the working directory."+
                " The ACS is not imported.");
    }
    if (!(obj instanceof ACSImpl)) {
        fis = null;
        in = null;
        acs = null;
        
        throw new DialogError(" "+_name+" is not an ACS. The ACS is not imported.");
    }
    else acs = (ACSImpl) obj;
    
    if (!_name.equals(acs.getName()) ) {
        fis = null;
        in = null;
        acs = null;
        
        throw new DialogError(" "+_name+" is not equals to the true name of the imported acs."+
                "The ACS is not imported.");
    }

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 1: adds "+_name+" in the ACS list");

    // record acs in the ACSManager before the call to the restoring methods
    // updates and saves the new l_Acs__
    synchronized(l_Acs__){
        l_Acs__.add( _name );
    }
    try {
      fos = new FileOutputStream("l_acs_v0.acr");
      out = new ObjectOutputStream(fos);
      out.writeObject(l_Acs__ );
      out.close();
    } catch (FileNotFoundException f) {
      throw new BaseError("in BaseManagerImpl.importACS() call, FileNotFoundException");
    }
    catch (SecurityException s) {
      throw new InternalError("in BaseManagerImpl.importACS() failed" + s.toString());
    }
    catch (IOException i) {
      throw new BaseError("in BaseManagerImpl.importACS() call, IOException");
    }


    // register the IS in the IS nametree, if unknown
    iSname = acs.getISName();
    if (iSname != null && !iSname.isEmpty() ) {
        iS = (NameISTreeImpl)getISTree(iSname);

        if (iS.isNode(acs.getName())) {
            fis = null;
            in = null;
            acs = null;
            l_Acs__.remove(_name);
            throw new DialogError(" "+_name+" is a node in the IS. "+
                    "The acs is not imported.");
        } // if (iS.isNode(acs.getName()))
        if (iS.isArea(acs.getName())) {
            fis = null;
            in = null;
            acs = null;
            l_Acs__.remove(_name);
            throw new DialogError(" "+_name+" is an area name in the IS. "+
                    "The acs is not imported.");
        } // if (iS.isNode(acs.getName()))

        if (!knowsIS(iSname)) {
            addIS(iSname, true);
        } // if
    }
    else {
        fis = null;
        in = null;
        acs = null;
        l_Acs__.remove(_name);
        throw new DialogError(" "+_name+" is an ACS whithout IS. "+
                "The acs is not imported.");
    } // else

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 2: adds "+_name+" in the IS nametree, with"+
                                    //" the missing parent compoments");

    // If the parent name is in the IS, register the ACS under this node.
    if (iS.isNode(acs.getParentName())) {
        // The following properties are missed: Composite,
        // Alternate relations to an ACS or subACS in the IS nametree.
        addAcsOrSubAcsNodeToIS(iSname, (Name)acs.getParentName(), acs.getName(),
                ISFactoryUtilityImpl.ACS_NODE, acs.getHubProperty(),
                acs.getVisibleSpace(),
                acs.getIsAuthorizationServer(), true);

        // registers the subACSs in the IS nametree
        // (copied from ACSFactoryImpl.createACSforGenericBase())
        if (acs.m_SubAndGroupPolicy_ != null) {

              for (String subname : acs.m_SubAndGroupPolicy_.keySet()) {

                  // does not process the group tree roots
                  if (acs.m_SubAndGroupPolicy_.get(subname)[0].equals("<GroupTree>") )
                      continue;
                  // processes '<ResourceSubACS>' and '<GroupSubACS>'

                  // is_hub_ and visibleSpace_ are from the ACS.
                  // a subACS is never an authorization server.
                  BaseManagerImpl.getInstance().addAcsOrSubAcsNodeToIS(
                        iSname, (Name) acs.getName(),
                        ((Name)acs.getName()).extended(subname),
                        ISFactoryUtilityImpl.SUB_ACS_NODE, false,
                        NameISTreeImpl.ALL_THE_IS_SPACE, false, true);

              } // for subname

          } // if (acs.m_SubAndGroupPolicy_ != null)

    } // if (iS.isNode(acs.getParentName()))
    // Otherwise, if there is no chain of nodes in the IS to match to the ACS names,
    // the chain of nodes is created with logical nodes derived from the ACS name.
    else {
        boolean firstNode = false;
        currentName = (Name) iSname.clone();
        try {
            for (String compo : acs.getParentName().components()) {
                // do not process the IS node
                if (!firstNode) {
                    firstNode = true;
                    continue;
                }
                addComponentNodeToIS(iSname, (Name)((Name)currentName.extended(compo)).clone(),
                        ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE);


                iS.setOpen(currentName, true);

                firePropertyChange("OneInformationSystemNode", null, _name);
            } // on compo
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BaseManagerImpl.importACS(), for the event "+
                    "'OneInformationSystemNode', firePropertChange() throws:");
            e.printStackTrace();
        }

        // roots are deleted, if any
        acs.m_RootCommandsForAcsOpeningAndClosing_ = null;
        acs.m_RootCommandsForAcsOpeningNoClosing_ = null;
        acs.m_RootCommandsForAcsIntegrity_ = null;
        acs.l_RootNamesForAcsOpeningNoClosing_ = null;
    } // else

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 3: restores the external bridges ");

      // sets the external ACS links in the returned map, without restoring,
      // and clears the relevant ACS lists and maps
    if (acs.m_closedSourcesOfActorBridgeTargets_ != null) {
        for (Map.Entry<ImmutableName, ArrayList<ImmutableName>> entry :
            acs.m_closedSourcesOfActorBridgeTargets_.entrySet()) {

                l_values.clear();
                for (ImmutableName name : entry.getValue()) {

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 3: process the external bridge source "+
                                    //name.toString());
                                
                    l_values.add(name.toString());
                } // for name
                m_result.put("<External closed Bridge Source(s) of the Target: >"+
                                entry.getKey().toString(),
                            (List<String>) l_values.clone());

        } // for entry
        //acs.m_closedSourcesOfActorBridgeTargets_ = null;
    } // if (acs.m_closedSourcesOfActorBridgeTargets_ != null)
    
    if (acs.m_closedTargetsOfActorBridgeSources_ != null) {
        for (Map.Entry<ImmutableName, ImmutableName> entry :
            acs.m_closedTargetsOfActorBridgeSources_.entrySet()) {

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 3: process the external bridge target "+
                                    //entry.getValue().toString());

                l_values.clear();
                l_values.add(entry.getValue().toString());
                m_result.put("<External closed Bridge Target of the Source: >"+
                                entry.getKey().toString(),
                            (List<String>) l_values.clone());

        } // for entry
        //acs.m_closedTargetsOfActorBridgeSources_ = null;
    } // if (acs.m_closedSourcesOfActorBridgeTargets_ != null)

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 4: restores the external ACLs ");

    // note: one ACL target may have several external sources, but they are
    // displayed in each ACL. These external ACLs are not grouped in a list
    // with the ACL target as key.
    if (acs.l_ClosedAclEntries_ != null) {
        for (ImmutableAclEntry acl : acs.l_ClosedAclEntries_) {
            if (!acl.getNameOfEpACS().equals(acs.getName())) {

                l_values.clear();
                l_values.add(acl.toString());
                m_result.put("<External closed Source in the AclEntry: >"+
                                acl.getDetailledName(),
                            (List<String>) l_values.clone());

            } // if
            if (!acl.getNameOfResACS().equals(acs.getName())) {

                l_values.clear();
                l_values.add(acl.toString());
                m_result.put("<External closed Target in the AclEntry: >"+
                                acl.getDetailledName(),
                            (List<String>) l_values.clone());

            } // if
        } // for acl
        //acs.l_ClosedAclEntries_ = null;
    } // if (acs.l_ClosedAclEntries_ != null)

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 5: restores the external privileges ");

    if (acs.l_ClosedPrivileges_ != null) {
        for (PrivilegeForLinks priv : acs.l_ClosedPrivileges_) {
            if (!priv.getNameOfSingleSourceACS().equals(acs.getName())) {

                l_values.clear();
                l_values.add(priv.toString());
                m_result.put("<External closed Source in the Privilege: >"+
                                priv.getDetailledName(),
                            (List<String>) l_values.clone());

            } // if
            if (!priv.getNameOfSingleTargetACS().equals(acs.getName())) {

                l_values.clear();
                l_values.add(priv.toString());
                m_result.put("<External closed Target in the Privilege: >"+
                                priv.getDetailledName(),
                            (List<String>) l_values.clone());

            } // if
        } // for priv
        //acs.l_ClosedPrivileges_ = null;
    } // if (acs.l_ClosedPrivileges_ != null)

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 6: adds "+_name+" in the open ACS map.");

    oldOpenACS = new HashMap(m_OpenAcs__);
    synchronized(m_OpenAcs__){
        m_OpenAcs__.put( (ImmutableName)_name.clone(), acs);
    }

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 7: sets the ACS relations in the IS ");

    // handles parent ACS relations
    acs.setAcsRelationsInIS(true, true);

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 8: restores the ACS rights and the internal links ");

    acs.checkClosedAcsRights();

    restoreACSInternalLinks(acs);

                                    // for testing and tracking
                                //System.out.println("in BaseManagerImpl.importACS(), "+
                                  //  "step 9: saves "+_name);

    saveOpenACS(acs);

    // fire an event
    try {
        firePropertyChange("OpenACS", oldOpenACS, new HashMap( getEorM_OpenACS() ));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.importACS(), firePropertChange() throws:");
        e.printStackTrace();
    }
        
    oldOpenACS = null;
    fis = null;
    in = null;
    obj = null;
    iSname = null;
    iS = null;
    currentName = null;
    l_values = null;

                                    // for testing and tracking
                                //System.out.println("BaseManagerImpl.importACS() "+
                                  //  "end ");
    
    return m_result;
  } // importACS()

    /**
    * Saves the base with the ACS and view lists, even if there are empty.
    * Updates the file 'l_acs_v0.acr'.
    * First, all the open views are saved in some files (one per open view) 
    * in the working directory.
    * With saveOpenACS(), each open ACS is saved in some files (one per open ACS) 
    * in the working directory.
    * The metarights are not saved since it is done on line, but their links
    * to the acsrights are updated.
    * Fires no event.
    * @exception BaseError if an acs or a view is not open, ACSname.lastComponent() is null,
    *   FileNotFoundException or IOException.
    */
    public synchronized void saveDefaultBase() throws BaseError {
      // variables
      Map     copyM_OpenAcs = new HashMap(m_OpenAcs__);
      ACSImpl acs;

      // save all views
      saveAllViews();
      
      // save l_acs_, even if it is empty
      try {
        FileOutputStream fos = new FileOutputStream("l_acs_v0.acr");
        ObjectOutputStream out = new ObjectOutputStream(fos);
        out.writeObject(l_Acs__ );
        out.close();
      } catch (FileNotFoundException f) {
        throw new BaseError("in BaseManagerImpl.saveDefaultBase() call, FileNotFoundException");
      }
      catch (SecurityException s) {
        throw new InternalError("in BaseManagerImpl.saveDefaultBase() failed" + s.toString());
      }
      catch (IOException i) {
        throw new BaseError("in BaseManagerImpl.saveDefaultBase() call, IOException");
      }

      // breaks the external references between open ACS
      for (Iterator ite = copyM_OpenAcs.values().iterator() ; ite.hasNext() ; ) {
        acs = ( (ACSImpl)ite.next() );
        saveOpenACS(acs);
      } // ite

      copyM_OpenAcs = null;
      acs = null;
    } // saveDefaultBase()

    /**
    * Closes the base without savings, with all the open ACS and all the open views first.
    * Called by ACS_Facade.closeDefaultBase() for ActionExit.
    * @exception BaseError if an acs is not open.
    */
    public void closeDefaultBase() throws BaseError {
      // variables
      Map copyM_OpenAcs = new HashMap(m_OpenAcs__);
      Map copyM_OpenViews = new HashMap(m_OpenViews__);
      ACSImpl acs = null;
      NameISTreeImpl      iSTree = null;

      // closes all open views
      for (Iterator ite = copyM_OpenViews.values().iterator() ; ite.hasNext() ; ) {
        closeOpenView((ViewInBase) ite.next() );
      }

                    // for testing and tracking
                //System.out.println("in BaseManagerImpl.closeDefaultBase(), "+
                //    "all the views are closed. ");

      // closes all open ACS, in breaking the references between us
      for (Iterator ite = copyM_OpenAcs.values().iterator() ; ite.hasNext() ; ) {
          acs = (ACSImpl)ite.next();
          iSTree = (NameISTreeImpl)m_NameTrees__.get(acs.getISName());
          // closes all the ACS children of this ACS, from the bottom of the tree
          // to let the deleting of the roots in the upper-node ACS
          if (!iSTree.isChildOfAcsNode(acs.getName()))
            closeOpenACS(acs, false);
      } // for ite

                    // for testing and tracking
                //System.out.println("in BaseManagerImpl.closeDefaultBase(), "+
                //    "all the ACS are closed. ");

      copyM_OpenAcs = null;
      copyM_OpenViews = null;
      acs = null;
    } // closeDefaultBase()

    /**
    * Save and close the base with the ACS and the view lists, even if it is empty.
    * All open ACS and views, but not the closed ones,
    * are saved in some files (one per open ACS and view) in the working directory.
    * Call closeOpenACS().
    * @exception BaseError if an acs or a view is not open, ACSname.lastComponent() is null,
    * FileNotFoundException or IOException.
    */
    public void saveAndCloseDefaultBase() throws BaseError {
      // variables
      Map       copyM_OpenAcs = new HashMap(m_OpenAcs__);
      Map       copyM_OpenViews = new HashMap(m_OpenViews__);
      ACSImpl   acs = null;
      NameISTreeImpl      iSTree = null;

      // save all views
      saveAllViews();
      // closes all open views
      for (Iterator ite = copyM_OpenViews.values().iterator() ; ite.hasNext() ; ) {
        closeOpenView((ViewInBase) ite.next() );
      }

      // save l_acs_, even if it is empty
      try {
        FileOutputStream fos = new FileOutputStream("l_acs_v0.acr");
        ObjectOutputStream out = new ObjectOutputStream(fos);
        out.writeObject(l_Acs__ );
        out.close();
      } catch (FileNotFoundException f) {
        throw new BaseError("in BaseManagerImpl.saveAndCloseDefaultBase() call, FileNotFoundException");
      }
      catch (SecurityException s) {
        throw new InternalError("in BaseManagerImpl.saveAndCloseDefaultBase() failed" + s.toString());
      }
      catch (IOException i) {
        throw new BaseError("in BaseManagerImpl.saveAndCloseDefaultBase() call, IOException");
      }

      // breaks the external references between open ACS
      for (Iterator ite = copyM_OpenAcs.values().iterator() ; ite.hasNext() ; ) {
          acs = ( (ACSImpl)ite.next() );

          iSTree = (NameISTreeImpl)m_NameTrees__.get(acs.getISName());
          // closes and saves also all the ACS children of this ACS
          if (!iSTree.isChildOfAcsNode(acs.getName()))
            closeOpenACS(acs, true); // 'true' to save it
      } // ite

      copyM_OpenAcs = null;
      copyM_OpenViews = null;
      acs = null;
      iSTree = null;
    } // saveAndCloseDefaultBase()

    /**
    * Saves an open ACS without closing it, with a synchronization on it.
    * The ACS instance is not modified.<P>
    * - If one of its Resources or one of its EligibleParty
    * is controlled by the AclEntry or the linked  Privilege of an another ACS,
    * it is NOT unrecorded in this another ACS.<P>
    * - If _acs is a controller, its external AclEntries and linked Privileges
    * are unrecorded in the controlled Resources and relevant EligibleParties,
    * and in their main ACS, then _acs is saved and its external AclEntries and
    * linked Privileges are recorded again with _acs.checkClosedAclEntries(),
    * but they are new instances THAT MUST BE TAKEN IN ACCOUNT BY THE APPLICATION.<P>
    * - If _acs has bridge links, they are also cuts then restored.<P>
    * Called by saveDefaultBase(). Calls saveACS(), and
    * the code is synchronized on the argument.
    * @param _acs to be saved
    * @exception BaseError if _acs is not open, ACSname.lastComponent() is null,
    *  or FileNotFoundException and IOException.
    */
    public synchronized void saveOpenACS(ACSImpl _acs) throws BaseError {
        // variables

        // exception
        if (!m_OpenAcs__.containsKey(_acs.getName()) ) {
            throw new BaseError("in BaseManagerImpl.saveOpenACS(), "+
              _acs.getName()+" is not open.");
        }
        
        synchronized(_acs) {
            // cuts references to _acs objects in other open ACS
            cutACSExternalLinks(_acs);
            // fills the VirtualFolder internal properties
            cutACSInternalLinks(_acs);

            // saves _acs from m_openAcs_
            saveACS(_acs);

            // restores all the external references between open ACS
            restoreACSExternalLinks(_acs);
            // restoreACSInternalLinks() processes only the VirtualFolders
            // which remains usable after cutACSInternalLinks()
        } // synchronized
    } // saveOpenACS()

    /**
    * Closes an open ACS with saving or not, and closes all its children ACS.
    * Uses this method after a saving, or when an ACS instance is corrupted,
    * before the opening of the old saving.
    * Updates the IS nametree node states, including all the children.
    * The ACS is still known in ACSManager.
    * Calls closeACS() for each ACS under _acs in the IS nametree, and for _acs.
    * After the call of this method, each processed ACS should be set to null.
    * closeACS() fires 'OpenACS' and 'OneInformationSystemNodeOpening' events.
    * @param _acs ACS to be closed and finalized.
    * @param _withSaving true to save the ACS
    * @exception BaseError if _acs is not open.                 
    */
    public void closeOpenACS(ACSImpl _acs, boolean _withSaving) throws BaseError {
        // exception
        if (!m_OpenAcs__.containsKey(_acs.getName()) ) {
            throw new BaseError("in ACSManager.closeOpenACS(), "
              +_acs.getName()+" is not opened.");
        } // if

        // variables
        NameISTreeImpl      iSTree = (NameISTreeImpl)m_NameTrees__.get(_acs.getISName());
        List<ImmutableName> l_children = iSTree.getL_AllChildren(_acs.getName());

        // updates the IS nametree, including all the children
        iSTree.setOpen(_acs.getName(), false);

        saveNameTrees();

        // finalizes and removes from the open acs map.
        // iterates on the list with the garanty from iSTree.getL_AllChildren() that
        // each processed node is a leaf.
        for (ImmutableName name : l_children) {
            if (iSTree.getType(name).equals(ISFactoryUtilityImpl.ACS_NODE) &&
                    isOpenACS(name) ) {

                closeACS((ACSImpl)m_OpenAcs__.get(name), _withSaving);
            } // if
        } // for name

        closeACS(_acs, _withSaving);

        iSTree = null;
        l_children = null;
    } // closeOpenACS()

  /**
  * The ACS name is removed from the ACS list and from the IS nametree, like
  * all its children in the IS nametree if they exist. Its roots in the parent ACS
  * are deleted if there are roots only for this ACS.
  * If it is opened, the ACS is closed and saved before.
  * So, it will be possible to restore the ACS in the futur.
  * To handle a corrupted ACS instance and to recover a correct ACS from file,
  * it is recommended to use closeOpenACS().
  * Call removeOneACS() for the ACS and its children.
  * No direct firing of event changes.
  * @param _name ACS to remove from the base.
  * @exception BaseError if the ACS is unknown, or it is not a leaf in the IS nametree,
  *  or if ACSname.lastComponent() is null, FileNotFoundException and IOException.
  */
  public void removeACS(ImmutableName _name) throws BaseError {

    if (!l_Acs__.contains(_name)) throw new BaseError(
        "in BaseManagerImpl.removeACS() call, _name is not a known ACS.");

    ImmutableName   iSname = getISofName(_name);

                            // for testing and tracking
                        //System.out.println("in BaseManagerImpl.removeACS(), "+
                          //  "iSname: "+iSname);

    NameISTreeImpl  iSTree = (NameISTreeImpl) m_NameTrees__.get(iSname);

                            // for testing and tracking
                        //System.out.println("in BaseManagerImpl.removeACS(), "+
                          //  "iSTree: "+iSTree);

    List<ImmutableName> l_children = iSTree.getL_AllChildren(_name);

    // finalizes and removes the ACS children
    for (ImmutableName name : l_children) {
        if (iSTree.getType(name).equals(ISFactoryUtilityImpl.ACS_NODE) ) {

            removeOneACS(name);
        } // if
    } // for name

    removeOneACS(_name);

    iSname = null;
    iSTree = null;
    l_children = null;
  } //removeACS()

  /**
  * Return the names of the registered ACS, open ones and closed ones.
  * @return cloned list of ACS names
  */
  public List<ImmutableName> getL_ACSNames(){
      return (List)((ArrayList)l_Acs__).clone();
  }

  /**
   * Called in the generic access paths searching, then may be called
   * from a view worker thread.
  * @return unmodifiable map of non-null instances of the open ACS, 
  *   where the ACS name is the key. Never null.
  */
  public synchronized Map<ImmutableName, ImmutableACS> getEorM_OpenACS() {
    return Collections.unmodifiableMap(m_OpenAcs__);
  }

  /**
  * @param _name ImmutableName of an ACS
  * @return boolean true if _name is known as an ACS name.
  * The ACS may be open or closed.
  */
  public boolean knowsACS(ImmutableName _name) {
    return l_Acs__.contains(_name);
  }

  /**
  * @param _name ImmutableName of an ACS
  * @return boolean true if _name is known as an ACS name and the ACS is open.
  */
  public boolean isOpenACS(ImmutableName _name) {
    return m_OpenAcs__.containsKey(_name);
  }
 
    /**
    * Restores all the internal properties of an ACS.
    * Fills in the ACS the l_NoRightProxyActors_ list and the m_RightsOfInternalKey_ map,
    * and restores the internal properties of VirtualFolders.
    * It is not necessary to call this method after cutACSInternalLinks() during
    * the simple saving of the ACS, since the workable properties of the VirtualFolders
    * remains usable after ACSImpl.closeVirtualFolders(), and the other ACS
    * properties this method updates are not changed by cutACSInternalLinks().
    * The code is synchronized on the argument.
    * Called by openACS(), importACS(), AcsFactory.configure() and the same class and
    * method in the ACS addons.
    * Caution: has to be called by any AcsFactoryForAcsAddon.configure() method which
    * creates some Actors.
    * Calls ACSImpl.fillNoRightProxyActors(), fillVirtualFolders(), fillRightsOfInternalKey().
    * @param _acs to be restored.   
    */
  final public void restoreACSInternalLinks(ACSImpl _acs) {

    _acs.fillNoRightProxyActors();
    _acs.fillVirtualFolders();
    _acs.fillRightsOfInternalKey();

  } // restoreACSInternalLinks()
  
  //
  // general base manager methods
  //
  
  /**
   * @return long
   */
  public long getSerialVersionUID() {
    return serialVersionUID;
  }

  /**
  * @return boolean true if the base supports XML exports
  * for their Meta rights, ACS and IS.
  */
  public boolean getSupportsXML()	{
    return (false);
  }

  /**
  * Add a listener to the bean.
  * Currently unused.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public synchronized void addPropertyChangeListener(PropertyChangeListener _l){
      
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      synchronized (changeSupport_)
      {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Add a listener to the bean.
  * For the properties 'OpenViews', 'OneInformationSystem', 'OneInformationSystemNode',
  * 'OneInformationSystemNodeOpening', 'OpenACS'.
  * Not used for the opening and the closing, which is managed
  * directly by the gDMak package.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

                // for testing and tracking
                //if (_propertyName.equals("OpenACS"))
                  //  System.out.println("in BaseManagerImpl.addPropertyChangeListener(),"+
                    //        " add a listener "+_l.getClass()+" for OpenACS");

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}
  } // addPropertyChangeListener()

  /**
  * remove a listener to the bean.
  * No use.
  * @param _l PropertyChangeListener to remove
  */
  public synchronized void removePropertyChangeListener(PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_l);}
  }

  /**
  * remove a listener to the bean.
  * For the properties 'OpenViews', 'OneInformationSystem', 'OneInformationSystemNode',
  * 'OneInformationSystemNodeOpening', 'OpenACS'.
  * Used by gDMak.XXXTreeBaseListenerImpl.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()

    //
    // ------------------ the ACS non-public METHODS -------------------------
    //

  /**
  * Called by the ACSFactoryImpl.createACS().
  * Calls addIS() for a new IS, then fires 'OpenACS' in all cases.
  * Does not save the new ACS and does not register it in the IS nametree.
  * Registers the ACS in the ACS list and the ACS IS in the IS nametree.
  * The user error messages are the followings: <P>
  * - The ACS name is already known as an ACS name. <P>
  * - The ACS name is already known as the name of an object into another open ACS. <P>
  * The caller may invoke later the methods addAcsOrSubAcsNodeToIS() and
  * addSecondaryPropertiesToAcsOrSubNode(). <P>
  * Caution: the new ACS may have the name of an ACSObject in a closed ACS,
  * and this is not checked. When the closed ACS is opened, there is no more checking
  * of the name doublons. In the search of objects from their names, the ACS
  * is selected first.
  * @see #addAcsOrSubAcsNodeToIS(ARoad0.gBaseInterface.ImmutableName, ARoad0.gBaseInterface.Name, ARoad0.gBaseInterface.ImmutableName, java.lang.String, boolean, java.lang.String, boolean, boolean)
  * @param _acs is the new ACS to be recorded in the base manager.
  *  _acs must not be null (throws an InternalError)
  * @param _ISroot_is_hub is true for a new IS for which the root in the nametree
  * is a hub. There is a new IS if _acs.getISName() returns an unknown IS name, and
  * otherwise this argument is not used.
  * @exception BaseError if the ACS name is already known as ACS, IS or component,
  *   if the IS name is not known, if _parentName is not a node of the IS nametree
  */
  protected final void newEorACS(ACS _acs, boolean _ISroot_is_hub) throws BaseError {
    // variables
    FileOutputStream    fos = null;
    ObjectOutputStream  out = null;
    Map                 oldOpenACS = null;
    boolean             unkown = false;

          // for testing and tracking
          //System.out.println("Call to BaseManagerImpl.newEorACS()");

    if (l_Acs__.contains(_acs.getName() ) || knowsACS(_acs.getName()) )
        throw new BaseError(" The ACS name is already known as an ACS name.");

    try {
        BaseUtilityImpl.getEorBaseObject(_acs.getName());
    }
    catch (BaseError b) {
        unkown = true;
    }
    if (!unkown) {
        throw new BaseError(" The ACS name is already known as the name "+
                "of an object into another open ACS.");
    }

    // updates and saves the new l_Acs__
    synchronized(l_Acs__){l_Acs__.add(_acs.getName() );}
    try {
      fos = new FileOutputStream("l_acs_v0.acr");
      out = new ObjectOutputStream(fos);
      out.writeObject(l_Acs__ );
      out.close();
    } catch (FileNotFoundException f) {
      throw new BaseError("in BaseManagerImpl.newEorACS() call, FileNotFoundException");
    }
    catch (SecurityException s) {
      throw new InternalError("in BaseManagerImpl.newEorACS() failed" + s.toString());
    }
    catch (IOException i) {
      throw new BaseError("in BaseManagerImpl.newEorACS() call, IOException");
    }
    
    // if the ACS has an IS, registers the IS in the IS nametree, saves it and fires 'ISNodes'
    ImmutableName iSname = _acs.getISName();
    if (iSname != null && !iSname.isEmpty() ) {
        if (!knowsIS(iSname)) {
            addIS(iSname, _ISroot_is_hub);
        } // if (!knowsIS
        //iSTree = null;
    }

    // updates m_OpenACS and fires an event (but it is an unused event)
    oldOpenACS = new HashMap(m_OpenAcs__);
    synchronized(m_OpenAcs__){m_OpenAcs__.put( (ImmutableName)_acs.getName(), _acs);}
    
    try {
      firePropertyChange("OpenACS", oldOpenACS, new HashMap( getEorM_OpenACS() ));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updatinge, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.newEorACS(), firePropertChange() throws:");
        e.printStackTrace();
    }

    fos = null;
    out = null;
    oldOpenACS = null;
  } // newEorACS

    /**
     * Saves an open ACS without closing for it.
     * If its objects are still referenced by
     * other open ACS, a crash may corrupt the base integrity.
     * Called by closeACS(), saveOpenACS(), and the calling code
     * has to be synchronized on the argument.
     * @param _acs ACS to be saved.
     * @exception BaseError if ACSname.lastComponent() is null,
     *  or FileNotFoundException and IOException.
     */
  final private void saveACS(ACSImpl _acs) throws BaseError {
    // variables
    Name name = (Name) _acs.getName();
    String str = "ACS0";
    FileOutputStream fos = null;
    ObjectOutputStream out = null;
    
    if (name.lastComponent() == null) {
        throw new BaseError("in BaseManagerImpl.saveACS(),"+
          " the ACS name last component is null");
    }

    int index = 0;
    while (index < name.noOfComponents()) {
        str = str.concat("_");
        str = str.concat(name.component(index));
        index++;
    }

    try {
        fos = new FileOutputStream( str + ".acr");
        out = new ObjectOutputStream(fos);
        out.writeObject( _acs );
        out.close();
    } catch (FileNotFoundException f) {
        throw new BaseError("in ACSManager.saveACS(),"+
          " FileNotFoundException");
    }
    catch (SecurityException s) {
        throw new InternalError(s.toString()+
          "BaseManagerImpl.saveACS() failed");
    }
    catch (IOException i) {
        throw new BaseError("in ACSManager.saveACS(), IOException "+i.getMessage());
    }

    name = null;
    str = null;
    fos = null;
    out = null;
  } // saveACS()

  /**
  * The ACS name is removed from the ACS list and from the IS nametree, like
  * all its children in the IS nametree if they exist. Its roots in the parent ACS
  * are deleted if there are roots only for this ACS.
  * If it is opened, the ACS is closed and saved before.
  * So, it will be possible to restore the ACS in the futur.
  * To handle a corrupted ACS instance and to recover a correct ACS from file,
  * it is recommended to use closeOpenACS().
  * The ACS saving file is not deleted in the files system. Of course,
  * it should be manually removed by the user. ACSManager.openACS(_name) wont work.
  * Calls removeIS() if this ACS is the last ACS of its IS, and this is the single way
  * to remove an IS.
  * Call closeOpenACS() if the acs is open. Call removeAcsNodeToIS().
  * No direct firing of event changes.
  * Called by removeACS().
  * @param _name of the ACS to remove from the base.
  * @exception BaseError if the ACS is unknown, or it is not a leaf in the IS nametree,
  *  or if ACSname.lastComponent() is null, FileNotFoundException and IOException.
  */
  private void removeOneACS(ImmutableName _name) throws BaseError {

    if (!l_Acs__.contains(_name)) throw new BaseError(
        "in BaseManagerImpl.removeOneACS() call, _name is not a known ACS.");

    ImmutableName   iSname = getISofName(_name);
    NameISTreeImpl  iSTree = (NameISTreeImpl) m_NameTrees__.get(iSname);

    if (isOpenACS(_name) ) closeOpenACS( (ACSImpl)m_OpenAcs__.get(_name), true);

                  // for testing and tracking
                  //System.out.println("in BaseManagerImpl.removeOneACS(), "+
                    //"call to removeAcsNodeToIS on IS: "+iSname);

    // saves the new IS nametree
    removeAcsNodeToIS(iSname, _name);

    // if _name is the last ACS, removes the IS
    if (!iSTree.isParentOfAcsNode(iSTree.getRoot())) {

                  // for testing and tracking
                  //System.out.println("in BaseManagerImpl.removeoneACS(), call to removeIS");

        removeIS(iSname);
    } // if

    // updates and saves l_Acs__
    synchronized(l_Acs__){l_Acs__.remove(_name );}

          // for testing and tracking
          //System.out.println("in BaseManagerImpl.removeOneACS(), l_Acs__: "+l_Acs__);

    try {
      FileOutputStream fos = new FileOutputStream("l_acs_v0.acr");
      ObjectOutputStream out = new ObjectOutputStream(fos);
      out.writeObject(l_Acs__ );
      out.close();
    } catch (FileNotFoundException f) {
      throw new BaseError("in BaseManagerImpl.removeACS() call,"+
        " FileNotFoundException");
    }
    catch (SecurityException s) {
      throw new InternalError("in BaseManagerImpl.removeACS() failed" + s.toString());
    }
    catch (IOException i) {
      throw new BaseError("in BaseManagerImpl.removeACS() call, IOException");
    }

    iSname = null;
    iSTree = null;
  } //removeOneACS()

    /**
    * Cuts all the external links to another open ACS and to the metarights. <P>
    * If one of the ACS resource or eligible party
    * is controlled by the AclEntry of an another ACS,
    * to avoid failed external references, it is unregistered
    * in this another ACS with ACS.ExternalAclEntries(_acs). <P>
    * If _acs is a controller, its external AclEntries are unregistered in the
    * controlled resources and relevant eligible parties, and in their main ACS,
    * with _acs.ExternalAclEntries(all ACS). <P>
    * If one of the ACS resource or eligible party
    * is controlled through the linked privilege of another ACS, or controls through a linked
    * privilege of another ACS, its external privileges are unregistered with
    * ACS.closeExternalLinkedPrivileges().
    * If _acs has bridge links, their are closed with closeBridges()
    * in the two sens. <P>
    * The code is synchronized on the argument and also on each other open ACS to process.
    * Called by saveOpenACS(), closeACS().
    * @param _acs to isolate from the other open ACS
    */
    final private void cutACSExternalLinks(ACSImpl _acs) {
        ACSImpl acs;
        
                      // for testing and tracking
                      //System.out.println("BaseManagerImpl.cutACSExternalLinks() on the ACS"+
                        //      _acs.getName());
                      //System.out.println("BaseManagerImpl.cutACSExternalLinks(), right metarights for "+_acs.getName());
                      //StringRight[] l_Rights_ = _acs.getL_Rights();
                      //for (int i = 0; i < l_Rights_.length; i++) {
                          // for testing and tracking
                        //  System.out.println("------------ at "+i+" right metaright is "+l_Rights_[i].getEorMetaRight());
                      //} // for ite

         synchronized (_acs) {

            for (Iterator ite = m_OpenAcs__.values().iterator() ; ite.hasNext() ; ) {
                  acs = (ACSImpl)ite.next();
                  if (acs == null || acs == _acs) continue;

                  synchronized (acs) {
                      _acs.closeExternalAclEntries(acs);
                      _acs.closeBridges(acs);
                      _acs.closeExternalLinkedPrivileges(acs);

                      acs.closeExternalAclEntries(_acs);
                      acs.closeBridges(_acs);
                      acs.closeExternalLinkedPrivileges(_acs);
                  } // synchro (acs)
            } // for

            _acs.closeAcsRights();
         } // synchro (_acs)
        
        acs = null;
    } // cutACSExternalLinks()
 
    /**
    * Cuts all the internal links.
    * Fills the VirtualFolder internal properties.
    * The code is synchronized on the argument.
    * Called by saveOpenACS(), closeACS().
    * @param _acs to be saved.
    */
    final private void cutACSInternalLinks(ACSImpl _acs) {
        synchronized (_acs) {
            _acs.closeVirtualFolders();
        }
    } // cutACSInternalLinks()

    /**
    * Restores all the external links to another open ACS and in the metarights.
    * Process the metarights, the AclEntries and the bridges which has to be opened.
    * Called by openACS(), saveDefaultBase(), saveOpenACS(), importACS().
    * The code is synchronized on the argument and also on each other open ACS to process.
    * @param _acs to restore
    */
    final private void restoreACSExternalLinks(ACSImpl _acs) {

                        // for testing and tracking
                        //System.out.println("BaseManagerImpl.restoreACSExternalLinks() "+
                          //"called for " + _acs.getName() );

        ACSImpl acs2;

        synchronized (_acs) {
            _acs.checkClosedAcsRights();

            for (Iterator ite = m_OpenAcs__.values().iterator() ; ite.hasNext() ; ) {
                acs2 = (ACSImpl)ite.next();
                if (acs2 == null || acs2 == _acs) continue;

                synchronized (acs2) {
                    // acs2 processes _acs
                    if (acs2.getL_ClosedAclEntriesACSNames() != null &&
                      acs2.getL_ClosedAclEntriesACSNames().contains(_acs.getName()) ) {

                            // for testing and tracking
                            //System.out.println("BaseManagerImpl.restoreACSExternalLinks(), "+
                              //"call to checkClosedExternalAclEntries() for the ACS " + acs2.getName()+
                                // "and the argument " + _acs.getName());

                        acs2.checkClosedExternalAclEntries(_acs);
                    } // if

                    if (acs2.getL_ClosedACSBridgeSources() != null)
                        acs2.checkClosedBridges(_acs);

                    if (acs2.getL_ClosedPrivilegesACSNames() != null &&
                      acs2.getL_ClosedPrivilegesACSNames().contains(_acs.getName()) ) {

                            // for testing and tracking
                            //System.out.println("BaseManagerImpl.restoreACSExternalLinks(), "+
                              //"call to checkClosedExternalLinkedPrivileges() for the ACS "+acs2.getName()+
                                // "and the argument " + _acs.getName());

                        acs2.checkClosedExternalLinkedPrivileges(_acs);
                    } // if

                    // _acs processes acs2
                    if (_acs.getL_ClosedAclEntriesACSNames() != null &&
                      _acs.getL_ClosedAclEntriesACSNames().contains(acs2.getName()) ) {

                                // for testing and tracking
                                //System.out.println("BaseManagerImpl.restoreACSExternalLinks(), "+
                                  //"call to checkClosedExternalAclEntries() for the ACS " + _acs.getName()++
                                    //" the argument " + acs2.getName() );

                        _acs.checkClosedExternalAclEntries(acs2);
                    } // if

                    if (_acs.getL_ClosedACSBridgeSources() != null)
                        _acs.checkClosedBridges(acs2);

                    if (_acs.getL_ClosedPrivilegesACSNames() != null &&
                      _acs.getL_ClosedPrivilegesACSNames().contains(acs2.getName()) ) {

                            // for testing and tracking
                            //System.out.println("BaseManagerImpl.restoreACSExternalLinks(), "+
                              //"call to checkClosedExternalLinkedPrivileges() for the ACS " + _acs.getName()+
                                // "and the argument " + acs2.getName());

                        _acs.checkClosedExternalLinkedPrivileges(acs2);
                    } // if
                } // synchro (acs2)

            } // for ite
        } // synchronized (_acs)
    
        acs2 = null;
    } // restoreACSExternalLinks()

    /**
    * Closes an open ACS with or without saving it.
    * Uses this method after a saving, or when an ACS instance is corrupted,
    * before the opening of the old saving.
    * The ACS is still known in ACSManager.
    * In this method, THE OBJECT _ACS IS SET TO NULL. <P>
    *   - If one of its resources or one of its eligible party
    *   is controlled by the aclEntry of an another ACS,
    *   to avoid failed external references, it is unrecorded
    *   in this another ACS with ACS.closeAclEntriesTo(_acs). <P>
    *   - If _acs is a controller, its aclEntries are unrecorded in the
    *   controlled resources and relevant eligible parties, and in their main ACS,
    *   with closeAclEntriesTo(_acs). <P>
    *   - If there are bridge links, closes them. <P>
    *   - cuts the links with metarights. <P>
    *   - then finalizes the ACS. <P>
    * After the call of this method, set the _acs to null, or
    * restore it with _acs = openACS(cloned _acs name).
    * Fires an 'OpenACS' event to the ACSTreeBaseListenerImpls and ViewInBases
    * before the the finalization of the ACS and its true removing in m_OpenAcs__.
    * Then, the open views that contains some objects from this ACS are closed
    * without user message.
    * Fires an 'OneInformationSystemNodeOpening' event to the BaseListeners after
    * the finalization of the ACS and its removing.
    * Called by closeOpenACS(). Calls saveACS(), and
    * the code is synchronized on the argument.
    * @param _acs to be closed and finalized
    * @param _withSaving true to save the ACS
    * @exception BaseError if _acs is not open.
    */
    private void closeACS(ACSImpl _acs, boolean _withSaving) throws BaseError {
        synchronized(_acs) {

                            // for testing and tracking
                        //System.out.println("call to BaseManagerImpl.closeACS(), "+
                          //  "for "+_acs.getName());

            HashMap<ImmutableName, ImmutableACS> m_next = new HashMap(m_OpenAcs__);

            // cuts references to _acs objects in other open ACS
            cutACSExternalLinks((ACSImpl)_acs);

            // does not use saveOpenACS() which restores the links with the external objects
            if (_withSaving) {

                    // fills the VirtualFolder internal properties
                    cutACSInternalLinks((ACSImpl)_acs);
                    saveACS(_acs);

            } // if

            // fires an event to ACSTreeBaseListenerImpl and ViewInBase
            m_next.remove(_acs.getName());
            try {
                firePropertyChange("OpenACS", new HashMap(m_OpenAcs__), m_next);
            } catch (Exception e) {
                ARoad0.Gui1.DialogBox.showAlarmMessage(
                  "There is an exception after the base updating, from outside the database."+
                  "You should save the base and stop the program, then run it once again.");
                System.out.println("From BaseManagerImpl.closeACS(), firePropertChange() throws:");
                e.printStackTrace();
            }

            // finalizes, but _acs.getName() still works
            ((ACSImpl)_acs).finalizeForBase();
            m_OpenAcs__.remove(_acs.getName());

            // fires an event to BaseListeners after the removing
            try {
                firePropertyChange("OneInformationSystemNodeOpening", _acs.getName(), null);
            } catch (Exception e) {
                ARoad0.Gui1.DialogBox.showAlarmMessage(
                  "There is an exception after the base updating, from outside the database."+
                  "You should save the base and stop the program, then run it once again.");
                System.out.println("From BaseManagerImpl.closeACS(), firePropertChange() throws:");
                e.printStackTrace();
            }

            m_next = null;
        } // synchronized
    } // closeACS()
    
    //
    // ------------ the ViewManager interface METHODS ------------------
    //


  /**
  * Open all known views, when there are all closed.
  * Fire an event 'OpenViews' for each open view.
  * @exception BaseError if a view is not known or ClassNotFoundException,
  *   a view object belongs to a closed ACS (with the relevant closed
  *   ACS names that may be get through BaseError.getL_Exception()),
  *   or IOException, or if a view is open at the start of the method (if so,
  *   close it before a new call to this method).
  * @exception DialogError if a view contains closed-ACS objects.
  */
  public void openAllViews() throws BaseError, DialogError {
    // variables
    FileInputStream     fis = null;
    ObjectInputStream   in = null;

    // exception
    if (!m_OpenViews__.isEmpty() ) throw new BaseError("in BaseManagerImpl."+
      "openAllViews(), a view is already open when the method is called.");

    // open the known objects file:
    try {
        fis = new FileInputStream(
            "l_views_v0.acr");
        in = new ObjectInputStream(fis);
//    } catch (ClassNotFoundException f) {
//        throw new BaseError("in BaseManagerImpl.openAllViews() call, ClassNotFoundException");
//    }
//      catch (SecurityException s) {
//        throw new InternalError(s.toString()+ "BaseManagerImpl.openAllViews() failed");
//
    } catch (IOException i) {
        // no l_views_v0.acr in the default directory
        l_Views__ = new ArrayList(INITIAL_CAPACITY );
        m_OpenViews__ = new HashMap( INITIAL_CAPACITY );
        return;
    }
    try {
        l_Views__ = (List) in.readObject();
        in.close();
    } catch (ClassNotFoundException f) {
        throw new InternalError("in BaseManagerImpl.openAllViews() call,"+
          " ClassNotFoundException");
    } catch (IOException i) {
        throw new InternalError("in BaseManagerImpl.openAllViews() call,"+
          " IOException");
    }

    // open all the views and update m_OpenViews__, throw BaseError
    for (Iterator ite = l_Views__.iterator() ; ite.hasNext() ; ) {
      openView((ImmutableName) ite.next() ) ;
    }

    fis = null;
    in = null;
  } // openAllViews()


  /**
  *  Save all the open views in some files (one per view)
  *  in the default data directory.
  * @exception BaseError if viewName.lastComponent() is null,
  *   FileNotFoundException or IOException.
  */
  public synchronized void saveAllViews() throws BaseError {

    // save l_views_, even if it is empty
    try {
      FileOutputStream fos = new FileOutputStream(
          "l_views_v0.acr");
      ObjectOutputStream out = new ObjectOutputStream(fos);
      out.writeObject(l_Views__ );
      out.close();
    } catch (FileNotFoundException f) {
      throw new BaseError("in BaseManagerImpl.saveAllViews() call,"+
        " FileNotFoundException");
    }
    catch (SecurityException s) {
      throw new InternalError("in BaseManagerImpl.saveAllViews() failed" +
        s.toString());
    }
    catch (IOException i) {
      throw new BaseError("in BaseManagerImpl.saveAllViews() call,"+
        " IOException");
    }

    // save each open view
    for (Iterator ite = m_OpenViews__.values().iterator() ; ite.hasNext() ; ) {
      saveView( (ViewInBaseImpl)ite.next() );
    } // ite

  } // saveAllViews()

  /**
  *  Close an open view and save it before.
  *  Fire an event 'OpenViews'. Work if an implied acs is closed.
  *  The view is still known in ViewManager.
  *  In this method, THE OBJECT _VIEW IS SET TO NULL.
  *  After the call of this method, SET the object _view TO NULL.
  * @param _view ViewInBase to be saved, closed and finalized.
  * @exception BaseError if ViewName.lastComponent() is null,
  *  or FileNotFoundException and IOException.
  */
  public synchronized void saveAndCloseOpenView(ViewInBase _view) throws BaseError {
      saveView(_view);
      closeOpenView(_view);
  } // saveAndCloseOpenView()

  /**
  *  Saves an open view without closing it.
  *  The view instance is not modified.
  * @param _view ViewInBase to be saved.
  * @exception BaseError if _view is not open, ViewName.lastComponent() is null,
  *  or FileNotFoundException and IOException.
  */
  public synchronized void saveOpenView(ViewInBase _view) throws BaseError {
      // exception
      if (!m_OpenViews__.containsKey(_view.getName()) ) {
          throw new BaseError("in ViewManager.saveOpenView(), "+
              _view.getName()+"+ is not open.");
      }
      // saves _view from m_OpenViews__
      saveView(_view);
  } // saveOpenView()

  /**
  * Gets the names of the open or closed views.
  * @return unmodifiable list of view names
  */
  public List<ImmutableName> getL_ViewNames(){
    return Collections.unmodifiableList(l_Views__);
  }


  /**
  * @return unmodifiable map of non-null instances of the open views, 
  *   where the view name is the key Never null.
  */
  public  Map<ImmutableName, ViewInBase> getEorM_OpenViews() {
    
            // for testing and tracking  
            //for (Iterator ite = m_OpenViews__.values().iterator() ; ite.hasNext() ; ) {
                //System.out.println("in BaseManagerImpl.getEorM_OpenViews(), "+
                    //"the OpenViews map contains an object with the class: "+ite.next().getClass());
            //}
      
      return Collections.unmodifiableMap(m_OpenViews__);
  }

  /**
  * @param _name ImmutableName of a view
  * @return boolean true if _name is known as a view name.
  * The view may be open or closed.
  */
  public boolean knowsView(ImmutableName _name){
    return l_Views__.contains(_name);
  }

  /**
  * @param _name ImmutableName of a view
  * @return boolean true if _name is known as a view name
  * and the view is open.
  */
  public boolean isOpenView(ImmutableName _name){
    return m_OpenViews__.containsKey(_name);
  }   

  /**
  * Opens a registered and closed view from a file in the current data directory.
  * Calls ViewInBase.prepareAfterRestoring() and ARoad0.Gui1.DialogBox.showAlarmMessage()
  * to inform about the deleted objects which has been removed from the view.
  * If ViewInBase.prepareAfterRestoring() throws an exception, the view is not opened
  * and the DialogError is thrown by this method to inform the user.
  * Else, if prepareAfterRestoring() returns a non-null message, it is displayed to the user,
  * typically to inform that a deleted object cannot be kept in the opening view.
  * Fires an event 'OpenViews'.
  * remove '<>' at the start and the end of a 'NMNM' indicator, for a NoMoreNoLess view
  * (new in 0.7.2). 0.7.0 and 0.7.1 backward compatibility:
  * opens a NoMore-NoLess view having a file name with '&lt;NMNL&gt;'.
  * @param _name ImmutableName of an EPRViewInBaseImpl to open. Its file name is built
  * with the three last components of the name, at most.
  * @exception BaseError if the view is not known, the view is open,
  *   the open view has not the same name,
  *   or ClassNotFoundException, IOException.
  * @exception DialogError if the view contains an object from a closed ACS, 
  *    and the exception message should be shown to the user
  * @return the open view.
  */
  public synchronized ViewInBase openView(ImmutableName _name)
    throws BaseError, DialogError {
    // variables
    String              str = null;
    FileInputStream     fis = null;
    ObjectInputStream   in = null;
    ViewInBaseImpl      view = null;
    String              preparing = null;

                // for testing and tracking
                //System.out.println("in BaseManagerImpl.openView(), m_OpenViews__: ") ;
                //for (Iterator ite = m_OpenViews__.keySet().iterator() ; ite.hasNext() ; ) {
                  //  System.out.println("--- "+ite.next() ) ;
                //}
                //System.out.println("in BaseManagerImpl.openView(), l_Views__: ") ;
                //for (Iterator ite = l_Views__.iterator() ; ite.hasNext() ; ) {
                  //  System.out.println("--- "+ite.next() ) ;
                //}
                
    // tests
    if (_name == null || _name.lastComponent() == null) {
      throw new BaseError("in BaseManagerImpl.openView(_name),"+
        " _name or _name.lastComponent() == null");
    }
    if (!l_Views__.contains(_name) ) {
        throw new BaseError("in ViewManager.openView(name) call,"+
          " the name is unknown");
    }
    if (m_OpenViews__.containsKey(_name) ) {
        throw new BaseError("in ViewManager.openView(name) call,"+
          " the view is already open");
    }

    if (_name.noOfComponents() < 2 ) {
      str = _name.lastComponent();
    }
    else if (_name.noOfComponents() == 2) {
      str = _name.context().lastComponent() + "_" + _name.lastComponent();
    }
    else if (_name.noOfComponents() == 3) {
                  // remove '<>' at the start and the end
      str = _name.component(1).substring(1, _name.component(1).length() - 1);
      str = _name.component(0) + "_" + str + "_" + _name.lastComponent();
    }


                // for testing and tracking
                //System.out.println("in BaseManagerImpl.openView(), str to open: "+str) ;
    
    // opens the file
    try {
        fis = new FileInputStream("VIEW0_" +str +".acr");
        in = new ObjectInputStream(fis);
        view = (ViewInBaseImpl) in.readObject();
        in.close();
    } catch (ClassNotFoundException f) {
        throw new BaseError("in BaseManagerImpl.openView() call,"+
          " ClassNotFoundException");
    }
//      catch (SecurityException s) {
//        throw new InternalError(s.toString()+ "BaseManagerImpl.openEorACS() failed");
//      }
    catch (IOException i) {
        // for the NMNL views, 0.7.0 and 0.7.1 backward compatibility:
        // opens a view having a file name with '<NMNL>'.
        if (_name.noOfComponents() == 3) {
            str = _name.component(0) + "_" + _name.component(1) + "_" + _name.lastComponent();
            try {
                fis = new FileInputStream("VIEW0_" +str +".acr");
                in = new ObjectInputStream(fis);
                view = (ViewInBaseImpl) in.readObject();
                in.close();
            } catch (Exception f) {
                throw new BaseError("in BaseManagerImpl.openView() for a NoMoreNoLess view: "+f);
            }
        }
        else
            throw new BaseError("in BaseManagerImpl.openView() call, IOException "+
                    i.getMessage());
    } // catch

    if (!_name.equals(view.getName()) ) {
        throw new BaseError("in BaseManagerImpl.openView(_name) call,"+
          " _name is not equals to view.getName()");
    }
    
    try {
        preparing = view.prepareAfterRestoring();
    } catch (DialogError d) {
        view.finalizeFromDeserialization();

                // for testing and tracking
                //System.out.println("in BaseManagerImpl.openView(), DialogError thrown: "+
                    //d.toString()) ;
        
        fis = null;
        in = null;
        throw d;
    } // catch
    
    // registers view in the ViewManager
    Map oldOpenViews = new HashMap(m_OpenViews__);
    synchronized(m_OpenViews__){m_OpenViews__.put( (ImmutableName)_name.clone(), view);}

    // fire an event
    try {
         firePropertyChange("OpenViews",
                oldOpenViews, new HashMap(getEorM_OpenViews()));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.openView(), firePropertChange() throws: "+e);
    }
    
    if (preparing != null) 
        ARoad0.Gui1.DialogBox.showOkMessage("In "+view.getName()+": "+preparing );
    
    fis = null;
    oldOpenViews = null;
    in = null;
    preparing = null;

    return view;
  } // openView()


  /**
  * Gets the elements of a closed or opened ViewInBase.
  * For the closed views, calls FolderAbst.prepareDataAfterRestoring().
  * Fires no event. A closed view is not opened by this method.
  * @param _name of an opened or closed ViewInBase.
  * @param _acsNames false to get the view elements names, true to get the names of
  *     the ACS which own the view elements
  * @exception BaseError if the view is not known, 
  *   the open view has not the same name,
  *   or ClassNotFoundException, IOException.
  * @return the element names or the ACS names; if for ACS, it may contains a null item.
   *    Never null.
  */
  public synchronized Set<ImmutableName> getElementsOfOneView(
          ImmutableName _name, boolean _acsNames) throws BaseError {
    // variables
    String              str = null;
    FileInputStream     fis = null;
    ObjectInputStream   in = null;
    ViewInBaseImpl      view = null;
    Set<ImmutableName>  l_result = null;

                // for testing and tracking
                //System.out.println("in BaseManagerImpl.getElementsOfOneView(), m_OpenViews__: ") ;
                //for (Iterator ite = m_OpenViews__.keySet().iterator() ; ite.hasNext() ; ) {
                  //  System.out.println("--- "+ite.next() ) ;
                //}
                //System.out.println("in BaseManagerImpl.getElementsOfOneView(), l_Views__: ") ;
                //for (Iterator ite = l_Views__.iterator() ; ite.hasNext() ; ) {
                  //  System.out.println("--- "+ite.next() ) ;
                //}

    // tests
    if (_name == null || _name.lastComponent() == null) {
      throw new BaseError("in BaseManagerImpl.getElementsOfOneView(_name),"+
        " _name or _name.lastComponent() == null");
    }
    if (!l_Views__.contains(_name) ) {
        throw new BaseError("in ViewManager.getElementsOfOneView(name) call,"+
          " the name is unknown");
    }
    if (m_OpenViews__.containsKey(_name) ) {
        if (!_acsNames)
            return new HashSet(((ViewInBase)m_OpenViews__.get(_name)).getEorM_BaseObjects().keySet());
        else
            return new HashSet( ((ViewInBase)m_OpenViews__.get(_name)).getEorL_ConnectedACSNames() );
    }

    // it is a closed view, so open the file.
    if (_name.noOfComponents() < 2 ) {
      str = _name.lastComponent();
    }
    else if (_name.noOfComponents() == 2) {
      str = _name.context().lastComponent() + "_" + _name.lastComponent();
    }
    else if (_name.noOfComponents() == 3) {
      str = _name.component(0) + "_" + _name.component(1) + "_" + _name.lastComponent();
    }

    try {
        fis = new FileInputStream("VIEW0_" +str +".acr");
        in = new ObjectInputStream(fis);
        view = (ViewInBaseImpl) in.readObject();
        in.close();
    } catch (ClassNotFoundException f) {
        throw new BaseError("in BaseManagerImpl.getElementsOfOneView() call,"+
          " ClassNotFoundException");
    }
    catch (IOException i) {
        throw new BaseError("in BaseManagerImpl.getElementsOfOneView() call, IOException");
    }

    if (!_name.equals(view.getName()) ) {
        throw new BaseError("in BaseManagerImpl.getElementsOfOneView(_name) call,"+
          " _name is not equals to view.getName()");
    }

    // with or without a DialogError on setDataAfterRestoring(),
    // it is possible to call getEorM_BaseObjects() and getEorL_ConnectedACSNames()
    // and to finalize the instance with finalizeFromDeserialization().
    try {
        view.prepareDataAfterRestoring();
    } catch (DialogError d) {

                                //System.out.println("in BaseManagerImpl.getElementsOfOneView(), "+
                                        //"DialogError from view.prepareAfterRestoring(): "+
                                        //d.toString());
    } // catch

    if (!view.isEmpty()) {
        if (!_acsNames)
            l_result = new HashSet(view.getEorM_BaseObjects().keySet());
        else
            l_result = new HashSet(view.getEorL_ConnectedACSNames());
    }
    else
        l_result = Collections.EMPTY_SET;


                                //System.out.println("end of BaseManagerImpl.getElementsOfOneView(),"+
                                        //" for the view "+_name+" and _acsNames: "+_acsNames+
                                        //", l_result: ") ;
                                //for (Iterator ite = l_result.iterator() ; ite.hasNext() ; ) {
                                    //System.out.println("-------- "+ite.next() ) ;
                                //}

    view.finalizeFromDeserialization();
    view = null;
    fis = null;
    in = null;
    return l_result;
  } // getElementsOfOneView()

  /**
  * Closes an open view without saving it.
  * Fires an event 'OpenViews' to the Gui2.ViewTreeBaseListenerImpl before
  * the closing, to done it before the view proper events. Finalizes the view.
  * The file is not deleted.
  * @param _view to be closed and finalized. Is null when a view
  *   is closed after the closing of one of its ACS, because gDMak.ActionClose
  *   cannot get the view then.
  */
  public void closeOpenView(ViewInBase _view) {

                // for testing and tracking
                //for (Iterator ite = m_OpenViews__.keySet().iterator() ; ite.hasNext() ; ) {
                    //System.out.println("in BaseManagerImpl.closeOpenView()"+
                    //  ", m_OpenViews__: "+ite.next() ) ;
                //}

      // exception
      if (_view == null || !m_OpenViews__.containsKey(_view.getName()) ) {
          
                    //System.out.println("in BaseManagerImpl.closeOpenView()"+
                      //", unknown view: "+_view ) ;
          
          return;

          //throw new BaseError("in ViewManager.closeOpenACS(), "
          //  +_view.getName()+" is not open.");
      }

      // fires an event to BaseListeners
      synchronized (m_OpenViews__) {
            Map newOpenViews = new HashMap(m_OpenViews__);


                    // for testing and tracking
                    //System.out.println("in BaseManagerImpl.closeOpenView(), newOpenViews keys "+
                      //      "before the removing of "+_view.getName());
                    //for (Iterator ite = newOpenViews.keySet().iterator(); ite.hasNext(); ) {
                      // System.out.println("-------- "+ite.next()); 
                    //}
            
            newOpenViews.remove(_view.getName());

                    // for testing and tracking
                    //System.out.println("in BaseManagerImpl.closeOpenView(), newOpenViews keys "+
                      //      "after the removing of "+_view.getName()+" and before the 'OpenViews' firing");
                    //for (Iterator ite = newOpenViews.keySet().iterator(); ite.hasNext(); ) {
                      // System.out.println("-------- "+ite.next()); 
                    //}
            
            // fire an event
            try {
                 firePropertyChange("OpenViews",
                    new HashMap(getEorM_OpenViews()), newOpenViews);
            } catch (Exception e) {
                ARoad0.Gui1.DialogBox.showAlarmMessage(
                  "There is an exception after the base updating, from outside the database."+
                  "You should save the base and stop the program, then run it once again.");
                System.out.println("From BaseManagerImpl.closeOpenview(), firePropertChange() throws:");
                e.printStackTrace();
            }

                    // for testing and tracking
                    //System.out.println("in BaseManagerImpl.closeOpenView(),"+
                      //" call to _view.finalizeForBase(), when the fired new map of open views is:");
                    //for (Iterator ite = newOpenViews.keySet().iterator(); ite.hasNext(); ) {
                      // System.out.println("-------- "+ite.next()); 
                    //}

            // removes in m_OpenViews__
            m_OpenViews__.remove(_view.getName() );

            // removes as a listener and clears the internal maps
            try {
                ((ViewInBaseImpl)_view).finalizeForBase();
            }
            catch (UpDateError u) {
                throw new InternalError("from BaseManagerImpl.closeOpenView(): "+u);
            }

            newOpenViews = null;
      } // synchronized

                // for testing and tracking
                //System.out.println("in BaseManagerImpl.closeOpenView(),"+
                //  " l_OpenViewsChangeSupport__ does NOT know listeners:"+
                //  l_OpenViewsChangeSupport__.isEmpty());

  } // closeOpenView()

  /**
  *  Close the sketch view without saving it.
  *  No 'OpenViews' firing. Finalize the view.
  *  No associated file.
  * @param _view to be closed and finalized.
  */
  public void closeSketchView(EPRViewInBaseImpl _view) {
      // exception
      if (_view == null ) return;

      // this method is reserved to the sketch view
      _view.finalizeForGUI();

  } // closeSketchView()  
  
  /**
  * The view name is removed from the view list in ViewManager.
  * If it is open, the view is closed and saved before.
  * The view saving file is not deleted, but it could be (or removed) by the user.
  * Then, ViewManager.openView(this view) can't work.
  * The view objects in the JVM memory are not finalized.
  * @param _name ImmutableName of the view to remove.
  * @exception BaseError if the view is unknown,
  *   viewName.lastComponent() is null,
  *  or FileNotFoundException and IOException.
  */
  public void removeView(ImmutableName _name) throws BaseError {

    if (!l_Views__.contains(_name)) throw new BaseError(
      "in BaseManagerImpl.removeView() call, _name is not a known view.");

    if (isOpenView(_name) )
      saveAndCloseOpenView( (ViewInBase)m_OpenViews__.get(_name) );

    // update and save the new l_Views__
    l_Views__.remove(_name );

          // for testing and tracking
          //System.out.println("in BaseManagerImpl.removeView(), l_Views__: "+l_Views__);

    try {
      FileOutputStream fos = new FileOutputStream(
      "l_views_v0.acr");
      ObjectOutputStream out = new ObjectOutputStream(fos);
      out.writeObject(l_Views__ );
      out.close();
    } catch (FileNotFoundException f) {
      throw new BaseError("in BaseManagerImpl.removeView(),"+
        " FileNotFoundException");
    }
    catch (SecurityException s) {
      throw new InternalError("in BaseManagerImpl.removeView() failed"
        + s.toString());
    }
    catch (IOException i) {
      throw new BaseError("in BaseManagerImpl.removeView(), IOException");
    }

  } //removeView()

  //
  // -------------------- GUI METHODS --------------------------------------
  //
  
    /**
     * Saves the GUI configuration in the file 'GUI_config.acr'.
     * Called by ActionSaveAll, ActionSave.
     * @param _config is the map of the configuration objects, where the object 
     *   name is the key
     * @exception BaseError if FileNotFoundException and IOException.
     */
  final public void saveGUI(Map _config) throws BaseError {
    // variables
    FileOutputStream fos = null;
    ObjectOutputStream out = null;

          // for testing and tracking
          //System.out.println("in BaseManagerImpl.saveGUI(), _config keys are: ");
          //for (Iterator ite = _config.keySet().iterator() ; ite.hasNext() ; ) {
            //    System.out.println("--------- "+ite.next());
          //} // for   
    
    try {
        fos = new FileOutputStream("GUI_config.acr");
        out = new ObjectOutputStream(fos);
        out.writeObject( _config );
        out.close();
    } catch (FileNotFoundException f) {
        throw new BaseError("in ACSManager.saveGUI(),"+
          " FileNotFoundException");
    }
    catch (SecurityException s) {
        throw new InternalError(s.toString()+
          "BaseManagerImpl.saveGUI() failed");
    }
    catch (IOException i) {
        throw new BaseError("in ACSManager.saveGUI(), IOException");
    }

    fos = null;
    out = null;
  } // saveGUI()  
  
/**
 * Gets the GUI configuration from the 'GUI_config.acr' file.
 * Called by ActionSave, ActionSaveAll, ActionExit, Desktop.
 * @exception BaseError if ClassNotFoundException and IOException.
 * @return the map of the configuration objects in the GUI, where the object
 *   name is the key. Null if there is no 'GUI_config.acr' file.
 */
  final public Map getGUI() throws BaseError {
    // variables
    FileInputStream fis = null;
    ObjectInputStream in = null;
    Map m_result = null;

    try {
        fis = new FileInputStream("GUI_config.acr");
        in = new ObjectInputStream(fis);
        m_result = (Map) in.readObject();
        in.close();        
    } catch (FileNotFoundException f) {
        fis = null;
        in = null;
        return null;
    } catch (ClassNotFoundException f) {
        throw new BaseError("in BaseManagerImpl.getGUI() call,"+
          " ClassNotFoundException");
    }
    catch (IOException i) {
        throw new BaseError("in ACSManager.getGUI(), IOException");
    }

    fis = null;
    in = null;
    return m_result;
  } // getGUI() 
  
  //
  // --------------------- META-RIGHTS METHODS ------------------------------
  //

  /**
  * Gets all the metarights in an sorted set.
  * @return the metarights. Never null nor empty.
  */
  public List<StringMetaRight> getL_MetaRights() {
    return Collections.unmodifiableList(l_MetaRights__);
  } //getL_MetaRights() 

  /**
  * Gets all the descriptors of the ACS addons that are loaded in the current
  * working directory of Access Road.
  * The form of a value is 'Editor: editor_name - AcsAddon: addon_name'.
  * The public descriptor of the ACS Addon is the String property PUBLIC_ACS_ADDON_NAME
  * in the class gWork.AcsFactoryXXX, where XXX is the name
  * of the Addon package as it is get from ACS.getM_AcsVocabulary(),
  * for the key ACSFactoryUtilityImpl.ACSADDON_PACKAGE_EDITOR.
  * Called by ISNodeImpl.getL_AcsAddonDescriptors(), gDMak.ActionAcsAddonGuides
  * and ActionNewACSyst.
  * @return the descriptors. Never null and may be empty.
  */
  public List<String> getL_AcsAddonDescriptors() {
    return Collections.unmodifiableList(l_AcsAddonDescriptors__);
  } //getL_AcsAddonDescriptors()

  /**
  * Gets all the non-administrative metarights in an ordered list, which is a subset
  * of the metarights for GUI displaying and user selecting. A non-administrative
  * metaright has its right which does not ended by '|'.
  * @return the metarights list. Never null nor empty.
  */
  public List<StringMetaRight> getL_NonAdministrativeMetaRights() {
     List<StringMetaRight> l_result = new ArrayList(l_MetaRights__);
     for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
         if ( ((StringMetaRight)ite.next()).getRight().endsWith("|") )
             ite.remove();
     } // ite
     return l_result;
  } //getL_NonAdministrativeMetaRights() 
  
  /**
  * Gets the metaright with the correct right, following the creations in
  * BaseUtilityImpl.initializeBasicMetaRights(). It may be desactivated.
  * Called in the generic access paths searching, then may be called
  * from a view worker thread.
  * @param _right is from StringRight.getRight()
  * @return the metaright. May be null.
  */
  public synchronized StringMetaRight getMetaRight(String _right) {

                    // for testing and tracking
                    //if (_right.indexOf("do_action_four") != -1 && m_MetaRightsFromRight__.get(_right) != null) {
                      //System.out.println("in BaseManagerImpl.getMetaRight() for do_action_four,");
                        //Map.Entry entry = null;
                        //StringMetaRightImpl right = (StringMetaRightImpl)m_MetaRightsFromRight__.get(_right);
                        //System.out.println("its m_ACSRights_ if not null:");
                        //if (right.m_ACSRights_ != null) {
                          //  for (Iterator ite = right.m_ACSRights_.entrySet().iterator(); ite.hasNext(); ) {
                            //    entry = (Map.Entry)ite.next();
                              //  if (entry.getValue() instanceof StringRight)
                                //    System.out.println("-- for "+entry.getKey()+" value: "+entry.getValue());
                                //else if (entry.getValue() instanceof Set) {
                                  //  System.out.println("-- for "+entry.getKey()+", the set is: ");
                                  //  for (Iterator ita = ((Set)entry.getValue()).iterator(); ita.hasNext(); ) {
                                    //    System.out.println("----- value: "+(StringRight)ita.next());
                                    //} // for ita
                                //} // else if
                            //} // for ite
                        //} // if (right.m_ACSRights_
                    //} // if
      
      
    return (StringMetaRight) m_MetaRightsFromRight__.get(_right);
  } //getMetaRight()  
  
  /**
  * Gets the metaright with the correct name through getName().
  * @param _name
  * @return the metaright. May be null.
  */
  public StringMetaRight getMetaRightFromName(ImmutableName _name) {
    if (_name == null) return null;
    StringMetaRight result = null;
    StringMetaRight right = null;
    
    for (Iterator ite = l_MetaRights__.iterator(); ite.hasNext(); ) {
        right = (StringMetaRight)ite.next();
        if (right.getName().equals(_name)) {
            result = right;
            break;
        }
    } // for
    
    right = null;
    return result;
  } //getMetaRightFromName() 
  
  /**
  * Gets all the metarights with the same power.
  * @param _power is the power of the searched metarights
  * @return the metaright array. May be null.
  */
  public StringMetaRight[] getL_MetaRightsFromPower(int _power) {
    Integer power = new Integer(_power);
    return (StringMetaRight[]) m_MetaRightsArrayFromPower__.get(power);
  } //getL_MetaRightsFromPower()  

  //
  // --------------------- META-RIGHTS non-public METHODS ------------------------------
  //

  /**
  * Register a new metaright. Used only by the StringMetaRightImpl constructor
  * for testing.
  * @param _right the new metaright
  * @param _toSave true if the saving must be provided
  * @exception CreateError if the metaright is already known
  */
  public static void newMetaRight(StringMetaRight _right, boolean _toSave)
          throws CreateError {
 
    // FOR StringMetaRight TESTING only, remove this test
    if (m_MetaRightsFromRight__.get(_right.getRight()) != null) {
        
                // for testing and tracking
               //System.out.println("In BaseManagerImpl.newMetaRight(), _right "+_right+" is known");
               
        throw new CreateError("In BaseManagerImpl.newMetaRight(), "+
                "the new metaright has a right which is already known.");       
    } // if
      
    l_MetaRights__ = BaseUtilityImpl.addObjectInDecreasingSort(l_MetaRights__, _right);
  

              // for testing and tracking
              //System.out.println("in BaseManagerImpl.newMetaRight(), add to "+
                //_right+" is successfull: "+ l_MetaRights__.contains(_right)+
                //", and the resulting list size is "+l_MetaRights__.size()+": ");
              //for (Iterator ite = l_MetaRights__.iterator(); ite.hasNext(); ) {
                //  System.out.println("------------- "+ite.next());
              //}
    
    try {
        if (_toSave) saveMetaRights();
    } catch (BaseError b) {
        throw new CreateError("from newMetaRight(): "+b.toString());
    }
    
    m_MetaRightsFromRight__.put(_right.getRight(), _right);
    
    StringMetaRight[] l_rights_one;
    StringMetaRight[] l_rights_two;
    Integer power; 
    power = new Integer(_right.getPower());
    
    if (!m_MetaRightsArrayFromPower__.containsKey( power )) {
         m_MetaRightsArrayFromPower__.put(power,  new StringMetaRight[]{_right} );
    } else {
        l_rights_one = (StringMetaRight[]) m_MetaRightsArrayFromPower__.get(power);
        l_rights_two = new StringMetaRight[l_rights_one.length + 1];
        System.arraycopy(l_rights_one, 0, l_rights_two, 0, l_rights_one.length);
        l_rights_two[l_rights_two.length - 1] = _right;
        m_MetaRightsArrayFromPower__.put(power,  l_rights_two );
    } // else
    l_rights_one = null;
    l_rights_two = null;
    power = null;
        
                // for testing and tracking
               //System.out.println("In BaseManagerImpl, end of newMetaRight()"); 

  } //newMetaRight() 
  
  /**
  * UNUSED METHOD.
  * Unregisters a metaright. Called by the metaright finalizeForUser() method.
  * @exception CreateError if the metaright is not known
  */
  protected synchronized void removeMetaRight(StringMetaRight _right) throws CreateError {
    if (m_MetaRightsFromRight__.get(_right.getRight()) == null) {
        throw new CreateError("In BaseManagerImpl.removeMetaRight(), the metaright "+
                _right+" is not registered.");
    } // if
    
    l_MetaRights__.remove(_right);
    try {
        saveMetaRights();
    } catch (BaseError b) {
        throw new CreateError("from removeMetaRight(): "+b.toString());
    }
    
    m_MetaRightsFromRight__.remove(_right.getRight() );
    
    StringMetaRight[]   l_rights_one;
    StringMetaRight[]   l_rights_two;
    Integer             power;
    int                 index = -1;
    power = new Integer(_right.getPower());

    if ( ((StringMetaRight[]) m_MetaRightsArrayFromPower__.get(power)).length == 1   ) {
         m_MetaRightsArrayFromPower__.remove(power);
    } else {
        l_rights_one = (StringMetaRight[]) m_MetaRightsArrayFromPower__.get(power);
        l_rights_two = new StringMetaRight[l_rights_one.length - 1];
        for (int i = 0; i < l_rights_one.length; i++) {
            if (l_rights_one[i] == _right) {
                index = i;
                break;
            } // if
        } // for
        
        if (index == -1) {
              throw new InternalError("in BaseManagerImpl.removeMetaRight(), "+
                "the right is not known in m_MetaRightsArrayFromPower__ for power: "+power+
                "; and the right: "+_right.getRight());      
        } // if (index)
        
        // copy the rights before index
        System.arraycopy(l_rights_one, 0, l_rights_two, 0, index);
        // copy the rights after index
        if (index != l_rights_one.length - 1)
            System.arraycopy(l_rights_one, index + 1, l_rights_two, index,
                l_rights_one.length - index - 1);
        m_MetaRightsArrayFromPower__.put(power,  l_rights_two );
    } // else
    l_rights_one = null;
    l_rights_two = null;
    power = null;

  } //removeMetaRight() 
  
    /**
     * Save the metarights list.
     * @exception BaseError if FileNotFoundException or IOException
     */
  final static private void saveMetaRights() throws BaseError {
    try {
                // for testing and tracking
               //System.out.println("In BaseManagerImpl.saveMetaRights(), beginning of method");
        
        FileOutputStream fos = new FileOutputStream("l_metarights_v0.acr");
        ObjectOutputStream out = new ObjectOutputStream(fos);
        out.writeObject(l_MetaRights__ );
        out.close();
        fos = null;
        out = null;
        
                // for testing and tracking
               //System.out.println("In BaseManagerImpl.saveMetaRights(), end of method");
                  //"call to "+listener+".propertyChange()"
                  //+" for the property: "+_propertyName);        
        
    } catch (FileNotFoundException f) {
        throw new BaseError("in BaseManagerImpl saveMetaRights(), FileNotFoundException");
    }
    catch (SecurityException s) {
        throw new InternalError("in BaseManagerImpl,  saveMetaRights() failed" + s.toString());
    }
    catch (IOException i) {
        throw new BaseError("in BaseManagerImpl saveMetaRights(), IOException");
    } // catch

  } // saveMetaRights()
  
  //
  // --------------------- VIEW non-public METHODS --------------------------
  //

  /**
  * Fire an event to every registered listener, in any order.
  * For the properties 'OpenViews', 'OneInformationSystem', 'OneInformationSystemNode',
  * 'OpenACS'.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and if the two values are not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;


            // for testing and tracking
            //if (_oldValue != null && _propertyName.equals("OneBaseObject")) {
              //  System.out.println( "BaseManagerImpl.firePropertyChange()
                // with a non-null old value: "+
                //        _oldValue+", for a 'OneBaseObject' event ");
                //System.out.println("PropertyChangeListeners for 'OneBaseObject' : ");
                //PropertyChangeListener[] l_prop = changeSupport_.getPropertyChangeListeners("OneBaseObject");
                //for (int i = 0; i < l_prop.length; i++) {
                  //  System.out.println(" -------------- "+l_prop[i].toString());
                //}
            //} // if

      synchronized (changeSupport_)
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
  * Called by the ViewInBaseImpl constructor for registering the instance.
  * @param _view is the new view to be recorded.
  *  _view must not be null (throws InternalError).
  * @exception CreateError if the view name is already known
  */
  protected synchronized final void newEorView(ViewInBase _view) throws CreateError {
    if (l_Views__.contains(_view.getName() ))
        throw new CreateError("in ViewManager.newEorView(),"+
          " the view name "+_view.getName()+"is already known - "+
            "if you are testing, delete first the l_views_v0.acr file");

    // update and save the new l_Views__
    l_Views__.add(_view.getName() );
    try {
      FileOutputStream fos = new FileOutputStream(
      "l_views_v0.acr");
      ObjectOutputStream out = new ObjectOutputStream(fos);
      out.writeObject(l_Views__ );
      out.close();
    } catch (FileNotFoundException f) {
      throw new CreateError("in BaseManagerImpl.newEorView(),"+
        " FileNotFoundException");
    }
    catch (SecurityException s) {
      throw new InternalError("in BaseManagerImpl.newEorView() failed"
          + s.toString());
    }
    catch (IOException i) {
      throw new CreateError("in BaseManagerImpl.newEorView(), IOException");
    }

    // update m_OpenViews and fire an event
    Map<ImmutableName, ViewInBase> oldOpenViews = new HashMap(m_OpenViews__);
    synchronized(m_OpenViews__){
      m_OpenViews__.put( (ImmutableName)_view.getName(), _view);
    }

    try {
        firePropertyChange("OpenViews", oldOpenViews, new HashMap(m_OpenViews__));
     } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BaseManagerImpl.newEorView(), firePropertChange() throws:");
        e.printStackTrace();
    }

    oldOpenViews = null;

  } // newEorView()

  /**
  * Called by the ViewInBaseImpl copyAndFinalize() method for registering
  * the instance and to save it, while it is not open.
  * @param _view is the new view to be recorded.
  *  _view must not be null (throws InternalError).
  * @exception BaseError if the view name is already known
  */
  protected synchronized final void newEorCopiedViewToSave(ViewInBase _view)
   throws BaseError {
    if (l_Views__.contains(_view.getName() ))
        throw new BaseError("in BaseManagerImpl.newEorCopiedViewToSave(),"+
          " the view name is already known");

    // update and save the new l_Views__
    l_Views__.add(_view.getName() );
    try {
      FileOutputStream fos = new FileOutputStream(
      "l_views_v0.acr");
      ObjectOutputStream out = new ObjectOutputStream(fos);
      out.writeObject(l_Views__ );
      out.close();
    } catch (FileNotFoundException f) {
      throw new BaseError("in BaseManagerImpl.newEorCopiedViewToSave(),"+
        " FileNotFoundException");
    }
    catch (SecurityException s) {
      throw new InternalError("in BaseManagerImpl.newEorCopiedViewToSave() failed"
          + s.toString());
    }
    catch (IOException i) {
      throw new BaseError("in BaseManagerImpl.newEorCopiedViewToSave(), IOException");
    }

    // save the copy
    try {
      saveView(_view);
    } catch (BaseError b) {
      b.printStackTrace();
    }

  } // newEorCopiedViewToSave()

    /**
     * Saves a view without closing it. Prepares it if it is a ViewInBaseImpl.
     * Does not update the view names file.
     * For a NoMoreNoLess view, remove '<>' at the start and the end of 'NMNL' (new in 0.7.2).
     * 0.7.0 and 0.7.1 backward compatibility: deletes the old file with '&lt;NMNL&gt;', if any.
     * Called by saveAndCloseEorOpenView(), saveOpenView(), newEorViewCopy(),
     * saveAllViews().
     * @param _view to be saved.
     * @exception BaseError if ViewName.lastComponent() is null,
     *  or FileNotFoundException and IOException.
     */
  final private void saveView(ViewInBase _view) throws BaseError {
      // variables
      Name name = (Name) _view.getName();
      
      // prepare view - works also for ViewInBaseImpl subclasses,
      // or if an implied acs is closed
      ((ViewInBaseImpl)_view).prepareBeforeSaving();

                // for testing and tracking
                //System.out.println("In BaseManagerImpl.saveView(), the view "+
                  //      _view.getName()+" contains: ");
                //for (Iterator ite = _view.getEorM_BaseObjects().keySet().iterator(); ite.hasNext() ; ) {
                  //  System.out.println("--------- "+(Name)ite.next());
                //} // for ite

      
      // string build up with the two last components (if there is only one, it is alone)
      String str = null;
      if (name.lastComponent() == null) {
          throw new BaseError("in BaseManagerImpl.saveView(),"+
            " the view name last component is null.");
      }

      if (name.noOfComponents() < 2 ) {
          str = name.lastComponent();
      }
      else if (name.noOfComponents() == 2) {
          str = name.context().lastComponent() + "_" + name.lastComponent();
      }
      else if (name.noOfComponents() == 3) {
        // remove '<>' at the start and the end
        str = name.component(1).substring(1, name.component(1).length() - 1);
        str = name.component(0) + "_" + str + "_" + name.lastComponent();

        // 0.7.0 and 0.7.1 backward compatibility: deletes the old file
        // with '&lt;NMNL&gt;', if any.
        File f = new File("VIEW0_" + name.component(0) + "_" + name.component(1) +
             "_" + name.lastComponent() + ".acr");

        if (f.exists()) {
         if (!f.canWrite())
           throw new BaseError(
                    "Unprevious write protected on "+
                "a NoMore-NoLess view file from an Access Road instance before 0.7.1-1 :"+
                "VIEW0_" + name.component(0) + "_" + name.component(1) +
                "_" + name.lastComponent() + ".acr");

         if (!f.delete())
           throw new BaseError(
                       "Deletion failed for the NoMore-NoLess view file: "+
                       "VIEW0_" + name.component(0) + "_" + name.component(1) +
                       "_" + name.lastComponent() + ".acr");
        } // if (f.canWrite())

        f = null;
      } // else if (name.noOfComponents() == 3) {

      try {
          FileOutputStream fos = new FileOutputStream(
          "VIEW0_" + str + ".acr");
          ObjectOutputStream out = new ObjectOutputStream(fos);
          out.writeObject( _view );
          out.close();
      } catch (FileNotFoundException f) {
          throw new BaseError("in saveView(), FileNotFoundException "+f.getCause()+
                  "; "+f.getMessage());
      }
      catch (SecurityException s) {
          throw new InternalError(s.toString()+ "BaseManagerImpl.saveView() failed "+s.getCause());
      }
      catch (IOException i) {
          throw new BaseError("in saveView() call, IOException "+i.getCause());
      }

      if (name.noOfComponents() == 3) {
        // for the NMNL views,
        // 0.7.0 and 0.7.1 backward compatibility: deletes the old file
        // with '<NMNL>', if any.
        File f = new File("VIEW0_" + name.component(0) + "_" + name.component(1) +
             "_" + name.lastComponent() + ".acr");

        if (f.exists()) {
                // not really useful...
             if (!f.canWrite())
               throw new BaseError(
                        "Unprevious write protected on "+
                    "a NoMore-NoLess view file from an Access Road instance before 0.7.1-1 :"+
                    "VIEW0_" + name.component(0) + "_" + name.component(1) +
                    "_" + name.lastComponent() + ".acr");

             if (!f.delete())
               throw new BaseError(
                           "Deletion failed for the NoMore-NoLess view file: "+
                           "VIEW0_" + name.component(0) + "_" + name.component(1) +
                           "_" + name.lastComponent() + ".acr");

                                // for testing and tracking
                                System.out.println("BaseManagerImpl.saveView() message: the view "+
                                        _view.getName()+" has an old file with '<NMNL>', "+
                                        "and it has been deleted. ");
        } // if (f.exists())

        f = null;
      } // if (name.noOfComponents() == 3) {

      name = null;
  } // saveView()


} // class BaseManagerImpl


