//Title:        Access Road
//Version:      0.7.0
//Copyright:    Copyright 2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.Pattern.Singleton;
import ARoad0.Pattern.BeanInfoPattern;

import java.beans.*;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Arrays;

/**
* This class is an utility class for all the ACS, with only static methods.
* It provides synchronized methods to copy, get a BaseObject or a BeanInfo.
* This class is not used by the view core algorithms, since its methods are dedicated
* to the updating and the displaying of the gBase class properties.
* A lot of methods are final, to provide generic operations.
* The subclasses of this class are in the AcsAddon packages.
* Caution: this class is not defined in ACSFactoryUtilityImpl.M_GENERIC_ACS_VOCABULARY,
* then it is not possible to get its subclasses in the AcsAddons, through
* ARoad0.gBaseInterface.ACSRun.getM_AcsVocabulary(). Then, a BaseUtilityImpl subclass
* is always reserved to its AcsAddon packages and its own subclasses.
* @version 0.7.0
*/
public class BaseUtilityImpl implements Singleton {
      /* the only instance of the class, following singleton pattern */
  protected static BaseUtilityImpl  instance__ = new BaseUtilityImpl();

  protected static final int        INITIAL_CAPACITY = 10; // the default value: 101
  
    // Constructors

  /**
   * Protected constructor for the subclasses.
  */
  protected BaseUtilityImpl() {
  }

    /**
    * Method to obtain the unique instance of BaseUtilityImpl.
    * @return BaseUtilityImpl
    */
  public static BaseUtilityImpl getInstance() {
    return instance__;
  } // getInstance()

/**
* Gets the BaseObject from the open entities only. It must be an ACS, a view, a metaright,
* or an ACSObject.
* If a view or an acs has the same name than an ACSObject, the view or the acs
* is returned without warning.
* Called by BaseManagerImpl.newEorACS(), and it is there normal to not find the BaseObject.
* Calls the first instance of UtilityImpl.getBaseObject().
* @param _name name of the object
* @return the non-null object
* @exception BaseError if there is no open object with this name
*/
  public static synchronized final BaseObject getEorBaseObject(ImmutableName _name)
          throws BaseError {

    BaseObject result = UtilityImpl.getFirstInstance().getBaseObject(_name);

    if (result != null) {
      return result;
    }
    else {
        BaseError e = new BaseError(
          "BaseUtilityImpl.getEorBaseObject - "+_name+
          " is not the name of an open base object.");
        //e.printStackTrace();
        throw e;
    }
  } // getEorBaseObject(name)

  /**
   * For UserIDs and GroupIDs, adds '&lt;U&gt;' or '&lt;G&gt;' before the name last component.
   * Called by AssociateUtilityImpl.addAssociateActions().
   * @param _initialName to analyze. Not null.
   * @param _class of the BaseObject. Never null.
   * @return the true name of the BaseObject. It is _initialName in most of the cases.
   */
  public static final ImmutableName getBaseObjectTrueName(ImmutableName _initialName, Class _class) {

                   // for testing and tracking purpose
                   //System.out.println("call to BaseUtilityImpl.getBaseObjectTrueName() for "+
                     //      "_initialName: "+_initialName+", and _class: "+_class);


      if (UserID.class.isAssignableFrom(_class)) {
          return _initialName.context().extended("<U>").extended(
                  _initialName.lastComponent());
      }
      else if (GroupID.class.isAssignableFrom(_class)) {
          return _initialName.context().extended("<G>").extended(
                  _initialName.lastComponent());
      }
      else return _initialName;

  } // getBaseObjectTrueName()

/**
 * A Basic name starts with its ACS name, and this is used to detect if a name
 * is from a known ACS in the full structure. Does not work for an AclEntry or a Privilege.
 * The returned value is the first ACS name which is found in removing the last components
 * in the argument.
* Called by FolderAbst.prepareAfterRestoring().
* @param _name the name of the ACSObject, which may be unknown by the ACS
* @return the name of the known ACS which owns the ACSObject, and this ACS
*   may be closed or not, or null if it is not a known ACS
*/
  public static synchronized final ImmutableName getFromKownACS(ImmutableName _name) {
      if (_name == null || _name.isEmpty()) return null;
      NameImpl result = (NameImpl)_name;

      while (!BaseManagerImpl.getInstance().knowsACS(result) && result.noOfComponents() > 1) {
          result = (NameImpl)result.subname(0, result.noOfComponents() - 1);
      }
      if (result.noOfComponents() == 1) {
          result = null;
      }

      return result;
  } // getFromKownACS()
  
  /**
   * Gets the restricted String or StringRight values for the argument of a new ACSObject.
   * When a value is restricted, the allowed values are get:<P>
   * - first, if _propertyName contains 'Rights', from the ACS.getM_SpecializedRights() map
   * if there is a matched key, and then, the values are in a StringRight array,<P>
   * - second, if _propertyName contains 'Rights', from the ACS.getM_StandardRights() map
   * if there is a matched key, and then, the values are in a StringRight array,<P>
   * - third, from the ACS.getM_SpecializedTypesAndExplorer() map if there is a matched key,
   * and then, the values are in a String array,<P>
   * - fourth, from the property descriptor in the BeanInfo if 'limitedSetForValues' is not null,
   * and then, the values are in a String array.<P>
   * For the privilege 'MainType' property in the constructor, returns only the
   * simple typed and the simple linked privilege types.
   * Called:<P>
   * - by BaseObjectHandler.getArgumentEditor() with a null or non-null ACSRun argument,<P>
   * - by CollectioPropertyEditor.actionPerformed() for the Select button. <P>
   * Calls getL_AcsRestrictedRights() and getL_AcsRestrictedStrings().
   * @param _baseObjectBeanInfo is the BeanInfo instance where the properties are described
   * @param _propertyName is the name of the restricted property in the BeanInfo,
   *     with spaces between words or not.
   * @param _acs is the new object ACS or ACSFactory. May be null.
   * @param _source is generally the object having _propertyName as property. May be null
   *     only if _propertyName does not contain 'Rights', since it is used for
   *     searching specialized rights. If _source is instance of AclEntry, the
   *     caller has to replace it by the AclEntry Resource to test the type.
   *     If it is for Privilege rights, _source has to be a Privilege.
   * @param _type not null only when _propertyName is 'AclRights' or 'PrivilegeRights',
   *    to give the type of the AclEntry EligibleParty or the Privilege type. May be null.
   * @return the restricted values as a String or a StringRight array, 
   *     or null if they are not found for these arguments.
   */
  public static synchronized final Object[] getRestrictedValues(
          BeanInfoPattern _baseObjectBeanInfo, String _propertyName, ACSRun _acs,
          BaseObject _source, String _type) {
      // initialization
    Object[]                l_result = null;
    PropertyDescriptor[]    l_propertyDescriptor =
          _baseObjectBeanInfo.getPropertyDescriptors();

                // for testing and tracking
                //System.out.println("in BaseUtilityImpl.getRestrictedValues(),"+
                  //" before the search of the restricted values in the ACS:");
                //System.out.println("----- _propertyName: "+_propertyName);
                //System.out.println("----- _baseObjectBeanInfo: "+_baseObjectBeanInfo);
                //System.out.println("----- effective name is: "+removeSpacesInString(_propertyName)+
                  //", acs: "+_acs+
                  //", source: "+_source+
                  //", type: "+_type);
                
    // first searching for the restricted values in the ACS
    if (_acs != null && _acs instanceof ImmutableACS) {   
        // if it is about rights
        if (_propertyName.indexOf("Rights") != -1 && _source != null) {
            l_result = getL_AcsRestrictedRights(_baseObjectBeanInfo,
                               _source, _propertyName, (ImmutableACS)_acs, _type);

                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.getRestrictedValues(),"+
                      //" result from getL_AcsRestrictedRights(): "+l_result);
        } // if
        // for Privilege only, since their constructor arguments about 'type'
        // has to be constrained by the current types in the ACS, but the values
        // cannot be provided by getL_AcsRestrictedTypes() since it does not return
        // the forbidden values to create a new Resource or an EP, although these
        // forbidden values may be chosen at the construction of a privilege.
        else if (_baseObjectBeanInfo.getBeanDescriptor().getDisplayName().
                indexOf("Privilege") != -1 && _propertyName.indexOf("Type") != -1 &&
                !_propertyName.equals("SecondType")) {

            if (_propertyName.equals("MainType") &&
                    _baseObjectBeanInfo.getBeanDescriptor().getDisplayName().
                    indexOf("Typed") != -1)
                l_result = new String[] {PrivilegeAbst.TYPED, PrivilegeAbst.TYPED_CLASSED,
                                PrivilegeAbst.TYPED_CLASSED_SPECIFIC, PrivilegeAbst.TYPED_FOR_SEED};

            if (_propertyName.equals("MainType") &&
                    _baseObjectBeanInfo.getBeanDescriptor().getDisplayName().
                    indexOf("Linked") != -1)
                l_result = new String[] {PrivilegeAbst.LINKED, PrivilegeAbst.LINKED_FOR_ONE_TO_ONE,
                                PrivilegeAbst.LINKED_IF_ALL_SOURCES};

            else if (_propertyName.equals("SourceType"))
                l_result = (String[])((ImmutableACS)_acs).getM_TypesAndGUIPolicy().
                        get("EligibleParty.EPType");

            else if (_propertyName.equals("TargetType")) {
                Set list = null;
                String[] l_v = (String[])((ImmutableACS)_acs).getM_TypesAndGUIPolicy().
                        get("VirtualFolder.ResourceVFType");

                if (l_v != null)
                    list = new HashSet(Arrays.asList(l_v));
                l_v = (String[])((ImmutableACS)_acs).getM_TypesAndGUIPolicy().
                        get("VirtualFolder.EPVFType");

                if (l_v != null) {
                    if (list == null)
                        list = new HashSet(Arrays.asList(l_v));
                    else
                        list.addAll(Arrays.asList(l_v));
                } // if

                l_v = (String[])((ImmutableACS)_acs).getM_TypesAndGUIPolicy().
                        get("Resource.ResourceType");
                if (l_v != null) {
                    if (list == null)
                        list = new HashSet(Arrays.asList(l_v));
                    else
                        list.addAll(Arrays.asList(l_v));
                } // if

                if (list != null)
                    l_result = list.toArray(new String[0]);
                l_v = null;
                list = null;
            } // else if
 
                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.getRestrictedValues(),"+
                      //" l_restrictedValues is set by the ACS for "+_propertyName+"."+
                      //" The restricted values are from the ACS, for a privilege: ");
                    //if (l_result != null) {
                      // for (int t = 0; t < l_result.length; t++) {
                        //   System.out.println("-------------- "+l_result[t]);
                       //} // for t
                    //}

        } // else if (_baseObjectBeanInfo.getBeanDescriptor().getDisplayName()
        else l_result = getL_AcsRestrictedTypes(
                   _baseObjectBeanInfo, _propertyName, (ImmutableACS)_acs);
    } // if (_acs != null)
    
    if (l_result != null) {

                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.getRestrictedValues(),"+
                      //" l_restrictedValues is set by the ACS for "+_propertyName+"."+
                      //" The restricted values are from the ACS: ");
                    //if (l_result != null) {
                      // for (int t = 0; t < l_result.length; t++) {
                        //   System.out.println("-------------- "+l_result[t]);
                       //} // for t
                    //}
        
        l_propertyDescriptor = null;
        return l_result;
    }
    
    // second searching for the restricted strings in the property descriptors
    for (int j = 0; j < l_propertyDescriptor.length; j++) {

                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.getRestrictedValues(),"+
                      //" for "+j+", the property name is "+l_propertyDescriptor[j].getName());        
        
        if (l_propertyDescriptor[j].getName().equals(removeSpacesInString(_propertyName)) ) {

            l_result =
              (String[]) l_propertyDescriptor[j].getValue("limitedSetForValues");

                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.getRestrictedValues(),"+
                      //" l_restrictedValues is found in the BeanInfo for "+_propertyName);

            break;

         } // if
    } // for j

    l_propertyDescriptor = null;

                // for testing and tracking
                //System.out.println("in BaseUtilityImpl.getRestrictedValues(),"+
                  //" l_restrictedValues is set by BeanInfo for "+_propertyName+"."+
                  //" The restricted values are from the property descriptors: ");
                //if (l_result != null) {
                  // for (int t = 0; t < l_result.length; t++) {
                    //   System.out.println(l_result[t]);
                   //} // for t
                //}

    return l_result;
  } // getRestrictedValues()

   /**
   * Gets the restricted rights from the ACS for the property of an ACSObject.<P>
   * - first, search of the specialized right constraints,<P>
   * - second, search of the general right constraints.<P>
   * The class in the ACS must be one of the interfaces of the ACSObject class.
   * Returns the first matching result which is found. In each of the two searches,
   * the search order is the order of interfaces get from getClassicSuperInterfacesFromBeanInfo().
   * When _propertyName is 'AclRights', it is recommanded to have a non-null _type
   * only when the eligible party belongs to _acs, like in AclEntryImpl.setRights().
   * When _propertyName is 'PrivilegeRights', there is no search of
   * the specialized right constraints.
   * Called by getRestrictedValues(), getNoDisplayThisRightProperty(),
   * ResourceImpl, AclEntryImpl, PrivilegeAbst.
   * Calls getClassicSuperInterfacesFromBeanInfo(), removeSpacesInString(),
   * getL_AcsSpecializedRights(), getClassNameFromArrayOrInterface() and
   * getACSVocabularyFromArrayOrInterface().
   * BaseUtilityImpl.getClassNameFromArrayOrInterface() and getBeanInfoNameFromClass().
   * @param _baseObjectBeanInfo is the BeanInfo instance where the properties are described.
   *    Used only when _source is a Privilege.
   * @param _source is the object for which there is the property and a BeanInfo. 
   *    For AclEntry, it is the AclEntry Resource or EligibleParty. 
   *    For Privilege, it is an empty PrivilegeForTypeImpl, even for a linked privilege.
   *    For other objects, it is the source of the property.
   * @param _propertyName is the property name, with spaces in the name or not
   * @param _acs is the ACSObject ACS
   * @param _type to use only when _propertyName is 'AclRights' for an ACL source,
   * to give the type of the eligible party or the privilege type. May be null,
   * even with the property 'AclRights'.
   * @return the StringRight array of the restricted values for the property.
    *   Null if there is no applicable restricted values in _acs for these arguments.
   */
  public static synchronized final StringRight[] getL_AcsRestrictedRights(
            BeanInfoPattern _baseObjectBeanInfo, BaseObject _source,
            String _propertyName, ImmutableACS _acs, String _type) {
        String resourceType = null;
        String epType = null;
        
        if (_source instanceof EligibleParty && _source instanceof Resource) {
            resourceType = ((Resource)_source).getResourceType();
            epType = ((EligibleParty)_source).getEPType();               
        } else if (_source instanceof EligibleParty && !(_source instanceof Resource))
            epType = ((EligibleParty)_source).getEPType();
        else if (!(_source instanceof EligibleParty) && (_source instanceof Resource)) {
            resourceType = ((Resource)_source).getResourceType();
        }
        // resourceType and epType remain null for 'PrivilegeRights' in a Privilege instance
        // in Resource, EligibleParty and VirtualFolder
        if (_propertyName.equals("AclRights") ) epType = _type;
      
        // init
        BeanInfoPattern baseObjectBeanInfo = null;
        String gName = null;
        String beanName = null;
        Class beanInfoClass = null;
        // gets the source BeanInfo
        gName = UtilityImpl.getFirstInstance().getClassNameFromArrayOrInterface(
                UtilityImpl.getFirstInstance().
                    getACSVocabularyFromArrayOrInterface(_source.getClass().getName()),
                _source.getClass().getName());
        beanName = UtilityImpl.getFirstInstance().getBeanInfoNameFromClass(gName);

                        // for testing and tracking
                        //System.out.println("in BaseUtilityImpl.getRestrictedRights(),"+
                          //" for propertyName "+_propertyName+
                          //", source class is "+_source.getClass().getName()+
                          //", gName: "+gName+
                          //", beanName: "+beanName);


        try {
            beanInfoClass = Class.forName(beanName);

                        // for testing and tracking
                        //System.out.println("in BaseUtilityImpl.getRestrictedRights(),"+
                          //" for propertyName "+_propertyName+", beanInfoClass: "+beanInfoClass);

            baseObjectBeanInfo = (BeanInfoPattern) beanInfoClass.newInstance(); 
        } catch (Exception e) {
            throw new InternalError("in BaseUtilityImpl.getL_AcsRestrictedRights(), exception "+e);
        }

        List             l_interfaces = getClassicSuperInterfacesFromBeanInfo(baseObjectBeanInfo);
        StringRight[]   l_result = null;

        if (!_propertyName.equals("PrivilegeRights")) {
            List      l_interfaces2 = l_interfaces;

            if (_propertyName.equals("AclRights")) {
                l_interfaces2 = new ArrayList(INITIAL_CAPACITY);
                l_interfaces2.add("AclEntry");
            }

            // search from the specialized rights before the standard right constraints
            l_result = getL_AcsSpecializedRights(
                                            _propertyName, _acs, resourceType, epType,
                                            l_interfaces2);

            resourceType = null;
            epType = null;
            baseObjectBeanInfo = null;
            //baseObjectBeanInfo2 = null;
            gName = null;
            beanName = null;
            beanInfoClass = null;
            l_interfaces2 = null;

            // gets from specialized rights
            if (l_result != null) {

                        // for testing and tracking
                        //System.out.println("in BaseUtilityImpl.getRestrictedRights(),"+
                          //" for propertyName "+_propertyName+", some specialized rights are found.");

                l_interfaces = null;
                return l_result;
            } // if
        } // if (!_propertyName.equals("PrivilegeRights")) {
        
        
                // for testing and tracking
                //System.out.println("in BaseUtilityImpl.getL_AcsRestrictedRights(),"+
                  //" l_class length: "+l_interfaces.length);
        
        String   AcsSetForValuesKey = null;
        String   propertyName = removeSpacesInString(_propertyName);
        String   keyName = null;
        Map      m_rights = _acs.getM_StandardRights();
 
                    // for testing and tracking
                    //if (_source.getName().lastComponent().equals("res_one"))
                        //System.out.println("in BaseUtilityImpl.getRestrictedRights(),"+
                          //  "test of standard rights for the source: "+
                            //_source);
        
        // search through the standard rights
        if (!propertyName.equals("PrivilegeRights")) {
            for (Iterator ite = l_interfaces.iterator(); ite.hasNext(); ) {
                keyName = (String)ite.next();

                // for the 'PrivilegeRights.MainType.'+PrivilegeAbst.TYPED/LINKED formats
                AcsSetForValuesKey = keyName.concat(".").concat( propertyName );

                                // for testing and tracking
                            //System.out.println("--- Test of the key for standard rights: "+
                            //AcsSetForValuesKey);

                // stops the search at the first matching of a StringRight array
                l_result = (StringRight[])m_rights.get(AcsSetForValuesKey);
                if (l_result != null) {
                    break;
                }
            } // for ite
        } // if (propertyName.equals("PrivilegeRights"))
        else {
            if (PrivilegeForType.class.isAssignableFrom(
                    (Class)_baseObjectBeanInfo.getBeanDescriptor().getValue("commonBeanClass")))
                AcsSetForValuesKey = propertyName.concat(".MainType.").concat( PrivilegeAbst.TYPED );
            else if (PrivilegeForLinks.class.isAssignableFrom(
                    (Class)_baseObjectBeanInfo.getBeanDescriptor().getValue("commonBeanClass")))
                AcsSetForValuesKey = propertyName.concat(".MainType.").concat( PrivilegeAbst.LINKED );

                                 // for testing and tracking
                            //System.out.println("_source: "+_source);
                            //System.out.println("beanInfoClass: "+beanInfoClass);
                            //System.out.println("Test of the key for standard rights: "+ AcsSetForValuesKey);

             l_result = (StringRight[])m_rights.get(AcsSetForValuesKey);
        } // else

        l_interfaces = null;
        AcsSetForValuesKey = null;
        propertyName = null;
        keyName = null;
        m_rights = null; // no clear(): unsupported operation
        beanInfoClass = null;
        
        return l_result;
  } // getL_AcsRestrictedRights()
  
   /**
   * Gets the restricted ACSObject types for the property of an ACSObject
   * to set in the GUI. The values to remove are defined in the ACS by
   * the key 'CreationByBeamer.NoType' in getM_TypesAndGUIPolicy().
   * The class in the ACS must be one of the interfaces of the ACSObject interface.
   * The integrity rules on types are so not enforced by the gBase package, but by
   * GUI packages, typically with the PropertyEditor classes and the ActionNewResource class. <P>
   * The search order of the possible values in getM_TypesAndGUIPolicy()
   * is the order of interfaces get from the method getClassicSuperInterfacesFromBeanInfo(),
   * where the searched keys have the format 'interface name.property name'.
   * The possible values are get from the first key to match. <P>
   * The user can create some ACSObjects having only the types the ACS authorizes.
   * Typically, the types 'ResourceType.&lt;byAcsCreator&gt;' and 'EPType.&lt;byAcsCreator&gt;'
   * are not allowed. After the first selection of allowed types, it may have a removing
   * of some values from the getM_TypesAndGUIPolicy() key 'CreationByBeamer.NoType'.
   * The forbidden types in the array have one of the following forms :
   * 'ResourceType.forbidden type', 'EPType.forbidden type', 'DirectoryType.forbidden type',
   * 'ActorType.forbidden type', 'UserIDType.forbidden type', 'GroupIDType.forbidden type',
   * 'VirtualFolderType.forbidden type', 'PrivilegeSecondType.forbidden type'.
   * A forbidden type for EligibleParty is not forbidden to all the EligibleParty subclasses.
   * This is also true for a forbidden type for Resource. Like the order of interfaces,
   * the 'ActorType.forbidden type' list is the single list to applied to an Actor,
   * just like 'DirectoryType.forbidden type' for the Directories,
   * and 'UserIDType.forbidden type' for the UserID. <P>
   * Caution: this method does provide all the types of the objects the program has created
   * at the ACS construction, although these types may be useful in varied operations
   * like the construction of a new typed privilege. <P>
   * 'VirtualFolderType' is converted to ResourceVFType or to EPVFType.
   * ResourceVFType is the first choice, if _acs manages the VirtualFolders of Targets.
   * Then if the 2 VirtualFolder types are used by the ACS,
   * the key 'VirtualFolder.EPVFType' is unused,
   * and the selection of a type for a VirtualFolder of EP Targets
   * is let to the user among the extended list. If its type choice is wrong,
   * the VirtualFolder constructor fires an exception. <P>
   * Called by getRestrictedValues() and SimplePropertyEditor.getRestrictedValues().
   * @see ARoad0.gBaseInterface.ACSCatalog#getM_TypesAndGUIPolicy() 
   * @param _baseObjectBeanInfo is the BeanInfo of a BaseObject class
   * @param _propertyName is the property name, with spaces in the name or not
   * @param _acs is the ACSObject ACS
   * @return the array of the restricted values for the property.
    *   Null if there is no applicable restricted types in _acs for these arguments.
   */
  public static synchronized final String[] getL_AcsRestrictedTypes(
    BeanInfoPattern _baseObjectBeanInfo, String _propertyName, ImmutableACS _acs) {
    
        String[] l_result = null;
        // the order is Actor, then Directory then Resource in l_interfaces
        // and UserID then GroupID then EligibleParty. Actually used only to have Actor
        // before Resource.
        List     l_interfaces = getClassicSuperInterfacesFromBeanInfo(_baseObjectBeanInfo);
        String   AcsSetForValuesKey = null;
        String   propertyName = null;
        String   keyName = null;
        
                                        // for testing and tracking
                                        //System.out.println("BaseUtilityImpl.getL_AcsRestrictedTypes()"+
                                          //" for the BeanInfo "+_baseObjectBeanInfo.getClass()+
                                          //" and the property "+_propertyName);

        // gets away the spaces in the property name
        propertyName = removeSpacesInString(_propertyName);

        // 'VirtualFolderType' is converted to ResourceVFType or to EPVFType.
        // ResourceVFType is the first choice, if _acs manages the VirtualFolders of Targets.
        // Then if the 2 VirtualFolder types are used by the ACS,
        // the key 'VirtualFolder.EPVFType' is unused,
        // and the selection of a type for a VirtualFolder of EP Targets
        // is let to the user among the extended list. If its type choice is wrong,
        // the VirtualFolder constructor fires an exception.
        if (propertyName.equals("VirtualFolderType")) {
            if (_acs.getManageResourceVirtualFolders())
                propertyName = "ResourceVFType";
            else
                propertyName = "EPVFType";
        }

        // search from _acs.getM_TypesAndGUIPolicy()
        for (Iterator ite = l_interfaces.iterator(); ite.hasNext(); ) {  
            keyName = (String)ite.next();
            
            AcsSetForValuesKey = keyName.concat(".").concat( propertyName );

                                    // for testing and tracking
                                    //System.out.println("in BaseUtilityImpl.getL_AcsRestrictedTypes(),"+
                                      //" l_restrictedValues is set by the ACS. "+
                                      //" test of the values key: "+AcsSetForValuesKey);

            // stops the search at the first matching of a String array
            l_result = (String[])_acs.getM_TypesAndGUIPolicy().get(AcsSetForValuesKey);
            if (l_result != null) {

                            // for testing and tracking
                            //System.out.println("in BaseUtilityImpl.getL_AcsRestrictedTypes(), "+
                              //      "result for "+AcsSetForValuesKey+" before any removing: ");
                            //for (int i = 0; i < l_result.length; i++) {
                              //  System.out.println(" ---------- "+l_result[i]);
                            //} // for i
                
                break;
            }
        } // for ite
        
        propertyName = null;
        keyName = null;
        
        if (l_result == null) return l_result;
        
        String[]    l_remove = null;
        boolean     remove = false;
        if (l_interfaces.contains("Resource") || 
                l_interfaces.contains("EligibleParty") ||
                l_interfaces.contains("VirtualFolder") ||
                l_interfaces.contains("Privilege") )
                        l_remove = (String[])_acs.getM_TypesAndGUIPolicy().
                                get("CreationByBeamer.NoType");

                            // for testing and tracking
                            //if (l_remove != null) {
                              //  System.out.println("    l_remove: ");
                                //for (int i = 0; i < l_remove.length; i++) {
                                  //  System.out.println(" ---------- "+l_remove[i]);
                                //} // for i
                            //} // if

        if (l_remove == null) return l_result;
        
        // removes the ACSObject types which are defined in 'CreationByBeamer.NoType'
        String[]     l_newArray = new String[l_result.length];
        String[]     l_res = new String[0];
        int          shiftingBack = 0;
        String       resourceApplicableInterface = null;
        String       epApplicableInterface = null;
        String       removing = null;

        // set the applicable interface for a Resource or a EligibleParty,
        // in the right order
        if (l_interfaces.contains("Actor")) {
            resourceApplicableInterface = "Actor";
        }
        else if (l_interfaces.contains("Directory")) {
            resourceApplicableInterface = "Directory";
        }
        else if (l_interfaces.contains("Resource")) {
            resourceApplicableInterface = "Resource";
        }

        if (l_interfaces.contains("UserID")) {
            epApplicableInterface = "UserID";
        }
        else if (l_interfaces.contains("GroupID")) {
            epApplicableInterface = "GroupID";
        }
        else if (l_interfaces.contains("EligibleParty")) {
            epApplicableInterface = "EligibleParty";
        }
        
                                    // for testing and tracking
                                    //System.out.println("BaseUtilityImpl.getL_AcsRestrictedTypes()"+
                                      //" resourceApplicableInterface = "+resourceApplicableInterface+
                                      //" - epApplicableInterface = "+epApplicableInterface);


        // initializes l_newArray from l_result with a filtering from l_remove
        for (int  i = 0 ; i < l_result.length ; i++) {
          if (l_result[i] != null) {
              if (l_remove == null) {
                  l_newArray[i] = l_result[i];
                  continue;
              }
              remove = false;
              for (int j = 0; j < l_remove.length; j++) {

                if (resourceApplicableInterface != null &&
                        (resourceApplicableInterface+"Type." + l_result[i]).equals(l_remove[j]) ) {
                    remove = true;

                    removing = l_remove[j];

                    break;
                }
                else if (epApplicableInterface != null &&
                        (epApplicableInterface+"Type." + l_result[i]).equals(l_remove[j]) ) {
                    remove = true;
                    break;
                }
                else if (l_interfaces.contains("VirtualFolder") && 
                        ("VirtualFolderType." + l_result[i]).equals(l_remove[j]) ) {
                    remove = true;
                    break;
                }
                else if (l_interfaces.contains("Privilege") &&
                        ("PrivilegeSecondType." + l_result[i]).equals(l_remove[j]) ) {
                    remove = true;
                    break;
                }
              } // for j on l_remove

              if (remove) {
                  // incrementing shiftingBack deletes each reference
                  shiftingBack++;

                                    // for testing and tracking
                                    //System.out.println("BaseUtilityImpl.getL_AcsRestrictedTypes() "+
                                      //      "removing for Resource of the value: "+removing);

              }
              else {
                  l_newArray[i - shiftingBack] = l_result[i];
              }
          } // if (l_result[i] != null)
          else shiftingBack++; 
        } // for i on l_result

        if (l_result.length != shiftingBack) {
          l_res = new String[l_result.length - shiftingBack];
          System.arraycopy(l_newArray, 0, l_res, 0, l_newArray.length - shiftingBack);
        }
        // else l_result is already equal to new String[0]
                                    
        l_result = l_res;        

                                        // for testing and tracking
                                        //System.out.println(
                                          //    "BaseUtilityImpl.getL_AcsRestrictedTypes(), result: ");
                                        //for (int i = 0; i < l_result.length; i++) {
                                          //  System.out.println(" ---------- "+l_result[i]);
                                        //} // for i

        AcsSetForValuesKey = null;
        l_interfaces = null;        
        l_newArray = null;
        l_res = null;
        l_remove = null;
        resourceApplicableInterface = null;
        epApplicableInterface = null;
        return l_result;
  } // getL_AcsRestrictedTypes()
  
    /**   
   * Gets the allowed specialized rights from the ACS for the property of an ACSObject.
   * If there are several matched keys, the returned array contains the rights
   * which are in all the matched arrays. So, to forbid a source type for instance,
   * a null array should be associated to it in the ACS properties.
   * Called by getL_AcsRestrictedRights(). 
   * Calls getClassicSuperInterfacesFromBeanInfo() and getL_MapKeysForSpecializedRights().
   * @param _propertyName is the property name, with spaces in the name or not.
   *    Has to contain 'Rights' since it is a right property.
   * @param _acs is the base object ACS
   * @param _resourceType is the type of the new object as Resource. May be null.
   * @param _EPType is the type of the new object as EligibleParty. May be null.
   * @param _l_interfaces interfaces to consider
   * @return the array of the restricted values for the property. Null if there is
   *   no restricted values in _acs for these arguments.
   */
  public static synchronized final StringRight[] getL_AcsSpecializedRights(
    String _propertyName, ImmutableACS _acs, String _resourceType, String _EPType,
    Collection _l_interfaces) {

                // for testing and tracking
                //System.out.println("in BaseUtilityImpl.getL_AcsSpecializedRights(),"+
                  //" before the search for the specialized rights in the ACS:");
                //System.out.println("------- property name argument is: "+_propertyName+
                  //", resource type is: "+_resourceType);
                //System.out.println("------- EPType is: "+_EPType);
                //System.out.println("----- and _l_interfaces are: ");
                //for (Iterator ite = _l_interfaces.iterator(); ite.hasNext(); ) {
                  //  System.out.println("------------ "+ite.next());
                //}
      
      
        Set<StringRight> l_result = new HashSet(INITIAL_CAPACITY);
        StringRight[]   l_rights = null;
        // build up all the applicable keys which should match 
        // in the specialized rights map from the acs
        String[]        l_s = ACSFactoryUtilityImpl.
                            getL_MapKeysForSpecializedRights(
                                _acs, _l_interfaces,
                                _propertyName, _resourceType, _EPType);
        Map             l_sRights = _acs.getM_SpecializedRights();
        Map             m_matchingArray = new HashMap(INITIAL_CAPACITY);
        StringRight[]   l_one = null;
        boolean         present = false;

                    // for testing and tracking
                    //System.out.println("BaseUtilityImpl.getL_AcsSpecializedRights(),"+
                      //" propertyName: "+_propertyName);                
                    //System.out.println("BaseUtilityImpl.getL_AcsSpecializedRights(), interfaces:");
                    //for (Iterator ite =  _l_interfaces.iterator(); ite.hasNext(); ) {
                      //  System.out.println("---------- "+ite.next());
                    //}
                    //System.out.println("in BaseUtilityImpl.getL_AcsSpecializedRights(), "+
                      //      "map keys in l_s:");
                    //for (int i = 0; i < l_s.length; i++ ) {
                      //  System.out.println("---------- "+l_s[i]);
                    //}
                    //if (_propertyName.equals("AclRights")) {
                      //  Map.Entry<String,StringRight[]> entry = null;
                      //  System.out.println("BaseUtilityImpl.getL_AcsSpecializedRights(), l_sRights:");
                        //for (Iterator ite = l_sRights.entrySet().iterator(); ite.hasNext(); ) {
                          //  entry = (Map.Entry)ite.next();
                          //  System.out.println("---------"+"key: "+entry.getKey()+"--- value: "+entry.getValue());

                        //} // for ite
                    //} // if
        
        for (int i = 0; i < l_s.length; i++ ) {  
            
                                    // for testing and tracking
                                    //if (_propertyName.equals("AclRights"))
                                      //  System.out.println("in BaseUtilityImpl.getL_AcsSpecializedRights(),"+
                                        //      "Test of the key: "+l_s[i]+" and the value: "+l_sRights.get(l_s[i]));

            l_rights = (StringRight[]) l_sRights.get(l_s[i]);
            if (l_rights != null) {
                
                                    // for testing and tracking
                                    //if (_propertyName.equals("AclRights"))
                                        //System.out.println(
                                          //    "-------- put in m_matchingArray the key: "+l_s[i]);
                                    
                m_matchingArray.put(l_s[i], l_rights);
            }
        } // for i
        
        l_s = null;
        l_sRights.clear();
        l_sRights = null;
        
        if (m_matchingArray.isEmpty()) {
            l_rights = null;
            l_result = null;
            m_matchingArray = null;
            l_sRights = null;

                                        // for testing and tracking
                                        //System.out.println(
                                          //      "in BaseUtilityImpl.getL_AcsSpecializedRights(), "+
                                            //    "result is null");
            
            return null;
        }
        else if (m_matchingArray.size() == 1) {
            l_rights = (StringRight[])m_matchingArray.get(
                        m_matchingArray.keySet().iterator().next());
            l_result = null;
            m_matchingArray = null;
            l_sRights = null;

                                        // for testing and tracking
                                        //System.out.println("in BaseUtilityImpl.getL_AcsSpecializedRights(), "+
                                          //      "result with only one matched key:");
                                        //for (int i = 0; i < l_rights.length; i++) {
                                          //  System.out.println("---------- "+l_rights[i]);
                                        //}
            
            
            return l_rights;
        }
        else {
            // puts in l_result the items which are in each array,
            // with a search from the first array.
            // Works if m_matchingArray contains only one array, but unused.
            // If an array has a length of 0, no rights are allowed.
            l_one = (StringRight[])m_matchingArray.get(m_matchingArray.keySet().iterator().next());
            for (int i = 0; i < l_one.length; i++) {
                for (Iterator ite = m_matchingArray.keySet().iterator(); ite.hasNext(); ) {
                        present = false;
                        l_rights = (StringRight[])m_matchingArray.get(ite.next());
                        for (int j = 0; j < l_rights.length; j++) {
                            if (l_rights[j].equals(l_one[i])) {
                                present = true;
                                break;
                            }
                        } // for j
                        if (!present) {

                                    // for testing and tracking
                                    //System.out.println("BaseUtilityImpl.getL_AcsSpecializedRights(), right"+
                                      //      " to remove: "+l_one[i]);
                    
                            break;
                        } // if
                } // for ite
                if (present) l_result.add(l_one[i]);
             } // for i            
        } // else
        
        l_rights = null;
        m_matchingArray = null;
        l_sRights = null;
        l_one = null;
        l_s = null;

                                    // for testing and tracking
                                    //System.out.println("in BaseUtilityImpl.getL_AcsSpecializedRights(), result"+
                                      //      " with several matched keys:");
                                    //for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                                      //  System.out.println("---------- "+ite.next());
                                    //}
        
        return l_result.toArray(new StringRight[0]);
  } // getL_AcsSpecializedRights()
  
    /**   
   * Gets the set of all the classical interfaces and super-interfaces of the BaseObject
   * from its BeanInfo.getBeanDescriptor().getBeanClass().
   * The hierarchy of package names ('ARoad0.gBaseInterface.') is not returned.
   * The order of the items is get initially from the Class.getInterfaces() method, but
   * the final sorting is Actor, then Directory then Resource,
   * and UserID then GroupID then EligibleParty.
   * For the AcsAddons, the returned list does not include the interfaces from
   * the AcsAdon, like 'ActorZZZ' from the interface 'ARoad0.XXX.YYY.gBaseInterface.ActorZZZ'.
   * Called by getL_AcsRestrictedValues() and getL_AcsSpecializedRights().
   * @param _baseObjectBeanInfo is the bean info of a BaseObject class. Never null.
   * @return classic interfaces without the package names. Never null.
   */
  public static synchronized final List<String> getClassicSuperInterfacesFromBeanInfo(
     BeanInfoPattern _baseObjectBeanInfo) {
        Class   curClass = null;
        List<String>     l_result = new ArrayList(INITIAL_CAPACITY);
        Set<Class>     l_classes = new HashSet(INITIAL_CAPACITY);
        Set<Class>     l_currentInterfaces = new HashSet(INITIAL_CAPACITY);
        Set<Class>     l_nextInterfaces = new HashSet(INITIAL_CAPACITY);
        String  keyName = null;        
                
        // defines the set of all the interfaces and super interfaces
        Class[] l_class = _baseObjectBeanInfo.getBeanDescriptor().getBeanClass()
            .getInterfaces();
        for (int i = 0; i < l_class.length; i++ ) {
            l_currentInterfaces.add(l_class[i]);
            l_classes.add(l_class[i]);
        } // i

        while (!l_currentInterfaces.isEmpty()) {
            for (Iterator ite = l_currentInterfaces.iterator(); ite.hasNext(); ) {  
                curClass = (Class)ite.next();
                l_class = curClass.getInterfaces();
                for (int i = 0; i < l_class.length; i++ ) {
                    l_nextInterfaces.add(l_class[i]);
                } // i
            } // for ite
            l_currentInterfaces = (Set)((HashSet)l_nextInterfaces).clone();
            l_classes.addAll(l_nextInterfaces);
            l_nextInterfaces.clear();
        } // while
        
        curClass = null;
        l_currentInterfaces = null;
        l_nextInterfaces = null;

        // switch the AcsAddon values 'ARoad0.XXX.YYY.gBaseZZZ.'
        for (Iterator ite = l_classes.iterator(); ite.hasNext(); ) {  
            keyName = ((Class)ite.next()).getName();

                        // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.getClassicSuperInterfacesFromBeanInfo(), "+
                      //  "keyName: "+keyName);
            
            if (keyName.indexOf("ARoad0.gBaseInterface") != -1) {
                keyName = keyName.substring(22);

                        // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.getClassicSuperInterfacesFromBeanInfo(), "+
                      //  "add keyName "+keyName);

                l_result.add(keyName);
            } else if (keyName.indexOf("ARoad0.gBase") != -1) {
                keyName = keyName.substring(13, keyName.length() - 4);

                        // for testing and tracking
                    //System.out.println("BaseUtilityImpl.getClassicSuperInterfacesFromBeanInfo(), "+
                      //  "add keyName "+keyName);

                l_result.add(keyName);
            } // else if
        } // for ite

        // the bean class is not in l_classes
        keyName = _baseObjectBeanInfo.getBeanDescriptor().getBeanClass().getName();
        if (keyName.indexOf("ARoad0.gBaseInterface") != -1) {
            keyName = keyName.substring(22);

                    // for testing and tracking
                //System.out.println("in BaseUtilityImpl.getClassicSuperInterfacesFromBeanInfo(), "+
                  //  "add keyName "+keyName);

            l_result.add(keyName);
        } else if (keyName.indexOf("ARoad0.gBase") != -1) {
            keyName = keyName.substring(13, keyName.length() - 4);

                    // for testing and tracking
                //System.out.println("BaseUtilityImpl.getClassicSuperInterfacesFromBeanInfo(), "+
                  //  "add keyName "+keyName);

            l_result.add(keyName);
        } // else if
        
        l_classes = null;
        keyName = null;

        int indexOne = l_result.indexOf("Directory");
        int indexTwo = l_result.indexOf("Actor");

        // Actor and Directory have to be before Resource
        if (indexOne != -1) {
            l_result.remove("Directory");
            l_result.add(0, "Directory");
        }
        if (indexTwo != -1) {
            l_result.remove("Actor");
            l_result.add(0, "Actor");
        }
 
        indexOne = l_result.indexOf("GroupID");
        indexTwo = l_result.indexOf("UserID");
        
        // UserID and GroupID have to be before EligibleParty
        if (indexOne != -1) {
            l_result.remove("GroupID");
            l_result.add(0, "GroupID");
        }
        if (indexTwo != -1) {
            l_result.remove("UserID");
            l_result.add(0, "UserID");
        }

                                // for testing and tracking
                            //System.out.println("in BaseUtilityImpl.getClassicSuperInterfacesFromBeanInfo(), "+
                              //  "last adding in result: "+
                                //keyName.substring(keyName.indexOf("ARoad0.gBaseInterface") + 22));

                            // for testing and tracking
                        //System.out.println("in BaseUtilityImpl.getClassicSuperInterfacesFromBeanInfo(), "+
                          //  "for _baseObjectBeanInfo "+_baseObjectBeanInfo+", result: ");
                        //for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                          //  System.out.println("----------- "+ite.next());
                        //} // for i
                        //System.out.println("----- for _baseObjectBeanInfo bean class: "+
                          //      _baseObjectBeanInfo.getBeanDescriptor().getClass().toString()+
                            //    ", with getBeanClass(): "+
                              //  _baseObjectBeanInfo.getBeanDescriptor().getBeanClass());
        
        return l_result;
  } // getClassicSuperInterfacesFromBeanInfo
  
  /**
  * Detects the need to display a right property in the beamer, following the ACS policy
  * about its standard and specialized rights.
  * The check of the specialized rights is based on the right category, but
  * it does not include the verification of the target or the source, and 
  * the interface/type of the object to display.
  * Does not display, for a Resource, the panes 'AG' and 'Inheritance' if the userID
  * is null and the type contains '&lt;byAcsCreator&gt;'.
  * Calls getL_AcsRestrictedRights() for testing the ACL and Privilege rights in Resources.
  * Called by BeamerManager.getNoDisplayThisProperty().
  * getL_AcsSpecializedRights() should allow to get a more precise view of what is
  * authorized, but it is not fast, and this method is called for each property
  * to display in the beamer. So, the CollectionPropertyEditor 'Select' button is
  * processed to explain to the user when there is no authorized values for a
  * property of a given ACSObject.
  * @param _object to display. May be null.
  * @param _propertyDescriptor is a property descriptor from a BeanInfo
  * @param _acs contains _object
  * @return false if the property has to NOT be displayed.
  *   Always true if _object is null.
  */
  public static synchronized final boolean getNoDisplayThisRightProperty(
    ACSObject _object, PropertyDescriptor _propertyDescriptor, ACS _acs) {
      // initialization
      boolean       result = true;
      String        paneName = (String)_propertyDescriptor.getValue("tabbedPaneName");
      StringRight[] l_rights = null;
      
      // ACS right categories are 'AccRightsSet',
      // 'GpORightsSet', 'AclRightsSet', 'PrvRightsSet', 'BdgRightsSet', 'RooRightsSet'.
      //Set l_cat = ((ACSImpl)_acs).getCategoriesInSpecializedRights();
     
     if (
             // 'ACL' pane in resources and eligible parties
         (paneName.equals("  ACL  ") 
           && !_acs.getManageInternalAcls() && !_acs.getManageExternalAcls() ) ||

             // 'Privileges' pane in resources, virtual folders and eligible parties 
           // TODO - lire getManagePrivilegeRightsForResources()
         (paneName.equals("  Privileges  ") 
           && !_acs.getManagePrivilegeRights() ) ||

             // 'AG' pane in groupIDmembers
         (paneName.equals("  AG  ")
           && _object instanceof GroupIDMember
           && !_acs.getManageActors() )  ||

             // 'AG' pane in resources
         (paneName.equals("  AG  ")
           && _object instanceof Resource
           && !_acs.getManageAGORights() )  ||

             // 'AG' pane in resources
         (paneName.equals("  AG  ")
           && _object instanceof Resource
           && ((Resource)_object).getEorUserID() == null
           && ((Resource)_object).getResourceType().indexOf("<byAcsCreator>") != -1 )  ||

             // 'Inheritance' pane in resources and virtual folders
         (paneName.indexOf("Inheritance") != -1
           && !_acs.getManageResourceTrees()
           && !_acs.getManageVirtualFolderTrees() )  ||

             // ' AG Inheritance' pane in resources
         (paneName.equals("  AG Inheritance  ")
           && !_acs.getManageAGORightsInheritance()  )  ||

             // ' ACL Inheritance' pane in resources
         (paneName.equals("  ACL Inheritance  ")
           && !_acs.getManageACLRightsInheritance())  ||

             // ' PRI Inheritance' pane in resources and virtual folders
         (paneName.equals("  PRI Inheritance  ")
           && !_acs.getManageLPRIRightsInheritance())  ||

             // 'AG Inheritance' pane in resources <byAcsCreator>
         (paneName.equals("  AG Inheritance  ")
           && _object instanceof Resource
           && ((Resource)_object).getEorUserID() == null
           && ((Resource)_object).getResourceType().indexOf("<byAcsCreator>") != -1 )  ||

             // memberships in resources and eligible parties
         (paneName.equals("  Memberships  ")
           && !_acs.getManageAccounts() && !_acs.getManageGroups()
           && !_acs.getManageAGORights() )  ||

             // 'Resources' pane in groupIDs and userIDs
         (paneName.equals("  Resources  ")
           && _object instanceof GroupIDMember
           && (!_acs.getManageResources() || !_acs.getManageAGORights()))  ||

             // 'Alias' pane in VirtualFolders and Basics
         (paneName.equals("  Alias  ")
           && !_acs.getManageSoftAlias() && !_acs.getManageInternalHardAlias() )  ||

             // 'Typed Privileges' in Privileges
           // TODO - when they are not simple
         (_propertyDescriptor.getName().equals("OneSourceTypedPrivilege")
           && !_acs.getManageSimpleTypedPrivileges() ) ||

             // 'Linked Privileges' in Privileges
           // TODO - when they are not simple nor one-to-one
         (_propertyDescriptor.getName().equals("OneSourceLinkedPrivilege")
           && !_acs.getManageSimpleLinkedPrivileges()
           && !_acs.getManageInternalForOneToOneLinkedPrivileges()
           && !_acs.getManageExternalForOneToOneLinkedPrivileges() ) ||

             // 'Typed Privileges' in Privileges
           // TODO - when they are not simple
         (_propertyDescriptor.getName().equals("OneTargetTypedPrivilege")
           && !_acs.getManageSimpleTypedPrivileges() ) ||

             // 'Linked Privileges' in Privileges
           // TODO - when they are not simple nor one-to-one
         (_propertyDescriptor.getName().equals("OneTargetLinkedPrivilege")
           && !_acs.getManageSimpleLinkedPrivileges()
           && !_acs.getManageInternalForOneToOneLinkedPrivileges()
           && !_acs.getManageExternalForOneToOneLinkedPrivileges()) ||
           
             // 'AcsControllers' in Resources
         (_propertyDescriptor.getName().equals("AcsControllers")
           && !_acs.getManageExternalAcls()
           && !_acs.getManageExternalForOneToOneLinkedPrivileges()) ||
           
             // Owner rights in resource
         (_propertyDescriptor.getName().equals("OwnerRights")
           && !_acs.getManageAGORights()  )  ||

             // Group rights in resource
         (_propertyDescriptor.getName().equals("GroupRights")
           && !_acs.getManageGroups() )  ||

             // Group rights in resource
         (_propertyDescriptor.getName().equals("GroupRights")
           && !_acs.getManageAGORights()  )  ||
           
            // EnforceChildGID in directory 
         (_propertyDescriptor.getName().equals("EnforceChildGID")
           && !_acs.getManageAGORights() ) ||
           
            // EnforceChildGID in directory 
         (_propertyDescriptor.getName().equals("EnforceChildGID")
           && ((Resource)_object).getEorUserID() == null
           && ((Resource)_object).getResourceType().indexOf("<byAcsCreator>") != -1 ) 
           
         ) {             


            // for testing purpose
            //System.out.println("in BaseUtilityImpl.getNoDisplayThisRightProperty(),"+
              //" no display for "+_propertyDescriptor.getName()+" in "+
              //acs.getName());             

             result = false;
     } // if

      // to be faster, calls getL_AcsRestrictedRights() only if the result is still true
     if (result && _object instanceof Resource && paneName.equals("  ACL  ")) {
                // _baseObjectBeanInfo is null since _object is not a Privilege
                // _epType is null since there is an array of ACL, not only one
         l_rights = getL_AcsRestrictedRights(null, _object, "AclRights", _acs, null);

             // 'ACL' pane in resources
         if ((l_rights == null || l_rights.length == 0) && !_acs.getManageACLRightsInheritance() ) {
                result = false;

                            // for testing purpose
                            //System.out.println("in BaseUtilityImpl.getNoDisplayThisRightProperty(),"+
                              //" no display for the pane 'ACL'");


         } // if
     } // if (result
     if (result && _object instanceof Resource &&
              _object.getEorACS().getM_TypeOrientedPolicy() != null &&
              paneName.equals("  Privileges  ")) {
                // 
         if (!(_object instanceof Actor) &&
                _object.getEorACS().getM_TypeOrientedPolicy().get(
                "Resource.SecondTypesOfPrivilegeFor."+((Resource)_object).getType()) == null)
                    result = false;
         else if (_object instanceof Actor &&
                 _object.getEorACS().getM_TypeOrientedPolicy().get(
                "Resource.SecondTypesOfPrivilegeFor."+((Resource)_object).getType()) == null &&
                 _object.getEorACS().getM_TypeOrientedPolicy().get(
                "EligibleParty.SecondTypesOfPrivilegeFor."+((Resource)_object).getType()) == null)
                    result = false;

                            // for testing purpose
                            //System.out.println("in BaseUtilityImpl.getNoDisplayThisRightProperty(),"+
                              //" no restricted rights, then no display for "+
                              //_propertyDescriptor.getName()+" in "+_acs.getName());
             
             
     } // if (result
      
     return result;
  } // getNoDisplayThisRightProperty()
  
    /**
    * Removes the whitespaces. Example: "Acl Rights" becomes "AclRights".
    * Called by getRestrictedValues() and getL_AcsRestrictedValues().
    * By the way, this is used by BaseObjectHandler and CollectionPropertyEditor, but
    * it is not usefull, since the property names in getBeanDescriptor() have no more spaces.
    * @param _propertyName is a set of words separated or not by one or several spaces,
     *   may be with whitespaces at the end and the beginning
    * @return a compacted set of characters, in one word
    */
  public static synchronized final String removeSpacesInString(String _propertyName) {
        String result = ""; 
        String property = _propertyName;
        property.trim();
        int spaceIndex = property.indexOf(" ");
        int lastIndex = 0;
        
                        // for testing and tracking
                //System.out.println("in BaseUtilityImpl.removeSpacesInString(), "+
                  //        "the string to compact is: "+_propertyName+
                    //      " and first index for space is: "+spaceIndex);
        
        if (spaceIndex == -1) result = property;
        else {
                    
            while (spaceIndex != -1) {
                result = result.concat(
                  property.substring(lastIndex,spaceIndex) );

                lastIndex = spaceIndex+1;
                if (property.substring(lastIndex).indexOf(" ") != -1)
                    spaceIndex = property.substring(lastIndex).indexOf(" ")+lastIndex;
                else break;
                
        
                        // for testing and tracking
                //System.out.println("in BaseUtilityImpl.removeSpacesInString(), "+
                  //        "the index for the next space is: "+spaceIndex);                
                
            }  // while
            result = result.concat( property.substring(lastIndex) );
        } // else
        property = null;
        return result;
  } // removeSpacesInString()
  
  /**
  * Copies a DirectoryImpl, and copies all the children and children of children
  * without any specific order. 
  * Called from ActionCopy.handleNewBaseObject().
  * Calls copyDirectoriesAndActorsWithoutInternalAcl(),
  *  ResourceImpl.copyOwnAcl() and copyDirectoriesAndResources().
  * @param _dirToCopy is the ResourceImpl to copy
  * @param _dir is the instance which has been just created,
  *  and is updated with _dirToCopy variables copies
  * @return Directory _dir
  */
  public static synchronized Directory copyDirectoryChildren(Directory _dirToCopy,
    Directory _dir) {
        try {
            copyDirectoriesAndActorsWithoutInternalAcl(_dirToCopy, _dir);

            ((ResourceImpl)_dirToCopy).copyOwnAcl((ResourceImpl)_dir, false, true);

            copyDirectoriesAndResources(_dirToCopy, _dir);
        } catch (CreateError c) {
           throw new InternalError("in BaseUtilityImpl.copyDirectoryChildren():"+c.toString());
        }

            // for testing and tracking
            //System.out.println("End of BaseUtilityImpl.copyDirectoryChildren().");

        return _dir;
  } // copyDirectoryChildren()

  /**
  * Copies all the ACS directories, and copies all the children and children of children
  * with the standard copy() method of the Copyable interface.
  * Has the copyDirectoryChildren() structure.
  * Called from ACSImpl.copy().
  * Calls copyDirectoriesAndActorsWithoutInternalAcl(),
  * copyDirectoriesAndResources(), ResourceImpl.copyOwnAcl() and
  * Resource.copyWithoutInternalAcl().
  * No action if _acsToCopy is _acs.
  * @param _acsToCopy is the source
  * @param _acs is the copied instance
  * @return _acs
  */
  public static synchronized ACS copyDirectoryChildrenForACS(ACS _acsToCopy, ACS _acs) {

    if (_acsToCopy == _acs) return _acs;
    if (_acsToCopy.getEorM_Resources().isEmpty()) return _acs;
      
    Resource res = null;
    Resource resCopy = null;
    Map m_resourcesToCopy = new HashMap(_acsToCopy.getEorM_Resources());
    
    try {
      for (Iterator ite = m_resourcesToCopy.values().iterator(); ite.hasNext();) {
          res = (Resource)ite.next();
          if (res.getEorACS() != _acsToCopy || res.getEorParent() != null) continue;

          // copies the first-level resource and registers the copy in _acs
          resCopy = (Resource)res.copyWithoutInternalAcl(
            ((Name)_acs.getName()).extended( res.getName().lastComponent()),
            _acs, null, Boolean.TRUE);

                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.copyDirectoryChildrenForACS(), "+
                      //"copy of a first-level resource: "+resCopy);

          // copies all the children of children in _acs
          if (res instanceof Directory) {

                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.copyDirectoryChildrenForACS(), "+
                      //"copy of the directories and actors for the directory: "+res);

              copyDirectoriesAndActorsWithoutInternalAcl((Directory)res, (Directory)resCopy);
          } // if

      } // for

      // at the start of this second iteration, all copied directories and actors
      // have been created with their external acls and without their internal acls
      m_resourcesToCopy = new HashMap(_acsToCopy.getEorM_Resources());
      for (Iterator ite = m_resourcesToCopy.values().iterator(); ite.hasNext();) {
          res = (ResourceImpl)ite.next();
          if (res.getEorACS() != _acsToCopy || res.getEorParent() != null) continue;
          
          resCopy = (Resource)_acs.getEorM_Resources().get(
            ((Name)_acs.getName()).extended( res.getName().lastComponent()) );
          
          // creates the internal acls for the first-level resources
          ((ResourceImpl)res).copyOwnAcl((ResourceImpl)resCopy, false, true);

          if (res instanceof Directory) {
                    // for testing and tracking
                    //System.out.println("in BaseUtilityImpl.copyDirectoryChildrenForACS(), "+
                      //"copy of acl and copy of the resources directory for: "+res);

              copyDirectoriesAndResources((Directory)res, (Directory)resCopy);
          } // if

      } // for
    } catch (CreateError c) {
       throw new InternalError("in BaseUtilityImpl.copyDirectoryChildrenForACS():"+c.toString());
    }

                        // for testing and tracking
                        //System.out.println("End of BaseUtilityImpl.copyDirectoryChildrenForACS().");
    
    res = null;
    resCopy = null;
    m_resourcesToCopy = null;
    return _acs;
  } // copyDirectoryChildrenForACS()

  /**
  * Copies all the actors and all the subdirectories without internal acl copy.
  * Called by copyDirectoryChildrenForACS() and copyDirectoryChildren().
  * @param _dirToCopy is the source
  * @param _dir is the copied instance
  * @return Directory _dir
  */
  public static synchronized Directory copyDirectoriesAndActorsWithoutInternalAcl(
   Directory _dirToCopy, Directory _dir) {
      
    if (_dirToCopy.getEorL_Children() == null ||
      _dirToCopy.getEorL_Children().length == 0) return _dir;
       
    // initialization
    Set l_copiedParent_one = new HashSet(INITIAL_CAPACITY);
    Set l_copiedParent_two = new HashSet(INITIAL_CAPACITY);
    Set l_toCopyChildren_one = new HashSet(INITIAL_CAPACITY);
    Set l_toCopyChildren_two;
    Map m_toCopyChildren_one = new HashMap(INITIAL_CAPACITY);
    Map m_toCopyChildren_two = new HashMap(INITIAL_CAPACITY);
    Resource child;
    Resource childCopy = null;
    Directory parent;
    int iter = 0;
    
    // init with _dir as copied parent
    l_copiedParent_one.add(_dir);
    ImmutableLeaf[] l_toCopyChildren = _dirToCopy.getEorL_Children();
    for (int i = 0; i < l_toCopyChildren.length; i++ ) {
        if (l_toCopyChildren[i] instanceof EligibleParty
          || l_toCopyChildren[i] instanceof Directory)
            l_toCopyChildren_one.add(l_toCopyChildren[i]);
    }
    m_toCopyChildren_one.put(_dir.getName(), l_toCopyChildren_one);

    // copies all the directories and actors of a stage n, then all ones
    // of the stage n+1, until there is no directory with new children to copy
    while (!l_copiedParent_one.isEmpty()) {
        try {
            for (Iterator ite = l_copiedParent_one.iterator(); ite.hasNext(); ) {  
              parent = (Directory)ite.next();
              l_toCopyChildren_one = (Set)m_toCopyChildren_one.get(parent.getName());

              for (Iterator itu = l_toCopyChildren_one.iterator(); itu.hasNext(); ) {
                  child = (Resource)itu.next();
                  if (child == null ||
                    child.getName().lastComponent() == null) continue;

                  if (child instanceof EligibleParty) {
                        childCopy = (Resource)((EligibleParty)child).copyWithoutInternalAcl( 
                          ((Name)parent.getName()).extended(child.getName().lastComponent()),
                          (ACS)parent.getEorACS(), parent, Boolean.TRUE);
                  }
                  else if (child instanceof Directory) {

                        childCopy = (Resource)((DirectoryImpl)child).copyWithoutInternalAcl(
                          ((Name)parent.getName()).extended(child.getName().lastComponent()),
                          (ACS)parent.getEorACS(), parent, Boolean.TRUE);

                        // adds to m_toCopyChildren_two and l_copiedParent_two
                        // with child.getEorL_Children() and childCopy
                        l_toCopyChildren = ((Directory)child).getEorL_Children();
                        if (l_toCopyChildren == null || l_toCopyChildren.length == 0) continue;
                        l_toCopyChildren_two = new HashSet(INITIAL_CAPACITY);
                        for (int i = 0; i < l_toCopyChildren.length; i++ ) {
                          if (l_toCopyChildren[i] instanceof EligibleParty
                            || l_toCopyChildren[i] instanceof Directory)
                                l_toCopyChildren_two.add(l_toCopyChildren[i]);
                        }

                                // for testing and tracking
                                //System.out.println("in BaseUtilityImpl.copyDirectoriesAndActorsWithoutAcl(),"+
                                      //" in loop "+iter+"; Directory children to copy: ");
                                //for (Iterator itetest = l_toCopyChildren_two.iterator(); itetest.hasNext(); ) {
                                    //System.out.println(" ------------- "+(ImmutableLeaf)itetest.next());
                                //}

                        m_toCopyChildren_two.put(childCopy.getName(), l_toCopyChildren_two);
                        l_copiedParent_two.add(childCopy);

                  } // else if (child instanceof Directory)
                  else continue;

                                // for testing and tracking
                                //System.out.println("in BaseUtilityImpl.copyDirectoriesAndActorsWithoutInternalAcl(),"+
                                      //" in loop "+iter+", Resource copy: "+childCopy.getName() );
                                //System.out.println(" ------ with source ACS: "+_dirToCopy.getEorACS().getName()+
                                      //"; and copy ACS: "+parent.getEorACS().getName());


              } // for itu

                                // for testing and tracking
                                //System.out.println("in BaseUtilityImpl.copyDirectoriesAndActorsWithoutInternalAcl(),"+
                                      //" in loop "+iter+"; after the parent: "+parent.getName()+" the children to copy are");
                                //for (Iterator itetest = m_toCopyChildren_two.keySet().iterator(); itetest.hasNext(); ) {
                                    //System.out.println(" ------------ "+(ImmutableName)itetest.next());
                                //}

            } // for ite
        } catch (CreateError c) {
          throw new InternalError("in BaseUtilityImpl.copyDirectoriesAndActorsWithoutInternalAcl():"+c.toString());
        }
          
        if (!l_copiedParent_two.isEmpty()) {
            l_copiedParent_one = (HashSet)((HashSet)l_copiedParent_two).clone();
            m_toCopyChildren_one = (HashMap)((HashMap)m_toCopyChildren_two).clone();
        } else {
            l_copiedParent_one.clear();
        } // else
        l_copiedParent_two.clear();
        m_toCopyChildren_two.clear();

                // for testing and tracking
                //System.out.println("in BaseUtilityImpl.copyDirectoriesAndActorsWithoutInternalAcl(),"+
                      //" in loop "+iter+" l_copiedParent_one at the end of the while loop: ");
                //for (Iterator itetest = l_copiedParent_one.iterator(); itetest.hasNext(); ) {
                    //System.out.println((BaseObject)itetest.next());
                //}
                
        iter++;
        
    } // while (!l_copiedParent_one.isEmpty())

    
    l_copiedParent_one = null;
    l_copiedParent_two = null;
    l_toCopyChildren_one = null;
    l_toCopyChildren_two = null;
    m_toCopyChildren_one = null;
    m_toCopyChildren_two = null;
    child = null;
    childCopy = null;
    parent = null;
    //acl = null;
    //dir = null;
    
    return _dir;
  } // copyDirectoriesAndActorsWithoutInternalAcl()
  
  /**
  * Applies ((ResourceImpl)child).copyOwnAcl() to each element under _dir.
  * Called from copyDirectoryChildrenForACS(), and copyDirectoryChildren().
  * Must be called AFTER copyDirectoriesAndActorsWithoutInternalAcl().
  * @param _dirToCopy is the source
  * @param _dir is the copied instance
  * @return Directory _dir
  */
  public static synchronized Directory copyDirectoriesAndResources(
   Directory _dirToCopy, Directory _dir) {
      
    if (_dirToCopy.getEorL_Children() == null ||
      _dirToCopy.getEorL_Children().length == 0) return _dir;
    
      // initialization
    Set l_copiedParent_one = new HashSet(INITIAL_CAPACITY);
    Set l_copiedParent_two = new HashSet(INITIAL_CAPACITY);
    Set l_toCopyChildren_one = new HashSet(INITIAL_CAPACITY);
    Set l_toCopyChildren_two = new HashSet(INITIAL_CAPACITY);
    Map m_toCopyChildren_one = new HashMap(INITIAL_CAPACITY);
    Map m_toCopyChildren_two = new HashMap(INITIAL_CAPACITY);
    Resource child;
    Resource childCopy;
    Directory parent;
    
    ImmutableLeaf[] l_toCopyChildren = _dirToCopy.getEorL_Children();
    l_copiedParent_one.add(_dir);
    for (int i = 0; i < l_toCopyChildren.length; i++ ) {
        l_toCopyChildren_one.add(l_toCopyChildren[i]);
    }
    m_toCopyChildren_one.put(_dir, l_toCopyChildren_one);

    // copies all the resources of a stage n, then all ones of the stage n+1,
    // until there is no directory with children to copy
    while (!l_copiedParent_one.isEmpty()) {
        
        for (Iterator ite = l_copiedParent_one.iterator(); ite.hasNext(); ) {  
          parent = (Directory)ite.next();
          l_toCopyChildren_one = (Set)m_toCopyChildren_one.get(parent);
          
          for (Iterator itu = l_toCopyChildren_one.iterator(); itu.hasNext(); ) {
              child = (Resource)itu.next();
              if (child == null ||
                child.getName().lastComponent() == null) continue;

              try {             
                  if (!(child instanceof Directory) && !(child instanceof EligibleParty)) {
                    
                      childCopy = (Resource)child.copy( 
                        ((Name)parent.getName()).extended(child.getName().lastComponent()),
                        (ACS)parent.getEorACS(), parent, Boolean.TRUE);


                            // for testing and tracking
                            //System.out.println("in BaseUtilityImpl.copyResourcesInDirectories(),"+
                                  //" Resource copy with name: "+childCopy.getName());
                  
                  } else {
                      
                      childCopy = (Resource)parent.getEorACS().getEorM_Resources().get(
                        ((Name)parent.getName()).extended(child.getName().lastComponent()));
                      
                      if (childCopy == null) throw new InternalError("in BaseUtilityImpl."+
                        "copyDirectoriesAndResources(), the copy is missing for the name: "+
                        ((Name)parent.getName()).extended(child.getName().lastComponent()));
                      
                      ((ResourceImpl)child).copyOwnAcl((ResourceImpl)childCopy, false, true);
                  } // else
                  
                   
                  if (child instanceof Directory) {
                        l_toCopyChildren = ((Directory)child).getEorL_Children();
                        if (l_toCopyChildren == null) continue;
                        l_toCopyChildren_two = new HashSet(INITIAL_CAPACITY);
                        for (int i = 0; i < l_toCopyChildren.length; i++ ) {
                            l_toCopyChildren_two.add(l_toCopyChildren[i]);
                        }

                                // for testing and tracking
                                //for (Iterator itetest = l_toCopyChildren_two.iterator(); itetest.hasNext(); ) {
                                    //System.out.println("in BaseUtilityImpl.copyResourcesInDirectories(), "+
                                      //"a Directory child to explore: "+(ImmutableLeaf)itetest.next());
                                //}

                        m_toCopyChildren_two.put(childCopy, l_toCopyChildren_two);
                        l_copiedParent_two.add(childCopy);
                  } // if (child
                  
              } catch (CreateError c) {
                  throw new InternalError("in BaseUtilityImpl.copyResourcesInDirectories():"+c.toString());
              }

          } // for itu
        } // for ite
        
        if (!l_copiedParent_two.isEmpty()) {
            l_copiedParent_one = (HashSet)((HashSet)l_copiedParent_two).clone();
            m_toCopyChildren_one = (HashMap)((HashMap)m_toCopyChildren_two).clone();
        } else {
            l_copiedParent_one.clear();
        } // else
        l_copiedParent_two.clear();
        m_toCopyChildren_two.clear();

    } // while    
    
    l_copiedParent_one = null;
    l_copiedParent_two = null;
    l_toCopyChildren_one = null;
    l_toCopyChildren_two = null;
    m_toCopyChildren_one = null;
    m_toCopyChildren_two = null;
    child = null;
    childCopy = null;
    parent = null;

    return _dir;
  } // copyDirectoriesAndResources()
  
  /**
  * Copies or adds members to a GroupID from another one members.
  * - if the ACS are the same, it adds the members without copy, 
  * - if the ACS are not the same, it searches the relevant members
  *   in the _grp ACS, and it adds them to _grp. If it does
  *   not exist, it creates the copy, it adds the copy as member,
  *   then it adds the copy to the returned array.
  * Generally, _grp is a copy of _groupToCopy but it is not mandatory.
  * Called from ActionCopy.handleNewBaseObject().
  * Caution: if the ACS is already displayed in the explorer
  *   before the calling of this method,
  *   each calling method must register the Gui2.ACSTreeBaseListeners 
  *   to listen the 'l_Members' events from _grp, and after the calling,
  *   must use the returned array to register the new returned groupIDs. 
  * @param _groupToCopy is the GroupID to copy, which has non-null members. 
  * @param _grp is the instance which has been created
  * @return array of the new goupIDs which may be processed then for
  *  Gui2.ACSTreeBaseListeners 'l_Members' registering. 
  *  The array structure is a list of couples (groupToCopy, copiedGroup),
  *  so the lenght is always a pair number.
  *  For a copy within an ACS, groupToCopy is equal to copiedGroup.
  *  May be empty but not null, and does not contain null element.
  */
  public static synchronized GroupID[] copyOrAddGroupMembers(
          GroupID _groupToCopy, GroupID _grp) {

    GroupID[] l_result = new GroupID[0]; 
    if (_groupToCopy == null || _grp == null ||
      _groupToCopy.getEorL_Members().length == 0) return l_result;
    
    ImmutableGroupIDMember[] l_toCopyMembers = _groupToCopy.getEorL_Members();
    GroupIDMemberImpl grpMemberCopy = null;
    GroupIDMemberImpl grpMember = null;
    GroupID[] result = null;
    Name newName = null;
    try { 
        if (_groupToCopy.getEorACS() == _grp.getEorACS()) {
             
            for (int i = 0; i < l_toCopyMembers.length; i++ ) {
                
                if (!_grp.containsAsMemberWithoutAliasing((GroupIDMemberImpl)l_toCopyMembers[i]))
                  _grp.addEorMember((GroupIDMemberImpl)l_toCopyMembers[i]);
                if (l_toCopyMembers[i] instanceof GroupID) {
                    // adds l_toCopyMembers[i] to l_result
                    result = new GroupID[l_result.length+2];
                    System.arraycopy(l_result, 0, result, 0, l_result.length);
                    result[l_result.length] = (GroupID)l_toCopyMembers[i];
                    result[l_result.length + 1] = (GroupID)l_toCopyMembers[i];
                    l_result = result;
                } // if
              
            } // for
            
        } else {
             for (int i = 0; i < l_toCopyMembers.length; i++ ) {
                if (l_toCopyMembers[i].getName().lastComponent() == null) continue;
                
                newName = ((Name)_grp.getName()).context().extended(
                       l_toCopyMembers[i].getName().lastComponent());
                
                // gets the right member in _grp.getEorACS()
                // and does not put the member in l_result, since it has already registered
                // the ACSTreeBaseListeners
                if (_grp.getEorACS().containsOwnEligiblePartyName(newName) ) {
                    grpMember = (GroupIDMemberImpl)_grp.getEorACS().
                      getEorM_EligibleParties().get(newName);
                    if (!_grp.containsAsMemberWithoutAliasing(grpMember))
                      _grp.addEorMember(grpMember);
                } else {
                  // else, creates the source member with copy(), and
                  // registers the new group as member of all _groupToCopy groups
                  // if they exist in the target ACS.
                  grpMemberCopy = (GroupIDMemberImpl)l_toCopyMembers[i].copy(
                    newName, (ACS)_grp.getEorACS(), null, Boolean.TRUE);
                  
                  // if grpMember belongs to _groupToCopy and if the last name component
                  // of _groupToCopy (it is an user choice) is not the same as _grp,
                  // grpMemberCopy has not be registered by copy() as member of _groupToCopy.
                  // So, a special test is needed
                  if (!_grp.containsAsMemberWithoutAliasing(grpMemberCopy))
                      _grp.addEorMember((GroupIDMemberImpl)grpMemberCopy);
                  
                  if (grpMemberCopy instanceof GroupID) {
                      // adds l_toCopyMembers[i] and grpMemberCopy to l_result
                      result = new GroupID[l_result.length+2];
                      System.arraycopy(l_result, 0, result, 0, l_result.length);
                      result[l_result.length] = (GroupID)l_toCopyMembers[i];
                      result[l_result.length + 1] = (GroupID)grpMemberCopy;
                      l_result = result;
                      
                            // for testing and tracking
                            //System.out.println("in BaseUtilityImpl.copyGroupMember(), groupID copy with name: "
                                  //+newName+"; and source groupID: "+l_toCopyMembers[i].getName());
                      
                  } // if
                } // else
             } // for             
         } // else
    } catch (CreateError c) {
        throw new InternalError("in BaseUtilityImpl.copyGroupMembers():"+c.toString());
    }
    l_toCopyMembers = null;
    grpMemberCopy = null;
    grpMember = null;
    result = null;
    newName = null;

    return l_result;
  } // copyGroupMembers()  
 

  /**
  * UNUSED - copies a GroupID tree with all members,
  * to register in _grp (not copied) all the _groupToCopy members.
  * - if the ACS are the same, it adds the members without copy, 
  * - if the ACS are not the same, it searches the relevant members
  *   in the _grp ACS, and it adds them to _grp. If it does
  *   not exist, it creates the copy, it adds the copy as member,
  *   then it add it to the returned array.
  * Caution: if a _groupToCopy member is also member of another groupID
  * which is not in the groupID tree, then the member link is not
  * created.
  * Called from ActionCopy.handleNewBaseObject().
  * If the ACS is already displayed in the explorer
  *   before the calling of this method,
  *   each calling method must register the Gui2.ACSTreeBaseListeners 
  *   to listen the 'l_Members' events from _grp, and after the calling,
  *   must use the returned array to register the new groupIDs for the same reason. 
  * @param _groupToCopy is the GroupID to copy, which has non-null members. 
  * @param _grp is the instance which has been created
  * @return array of the new goupIDs which may be processed then for
  *  Gui2.ACSTreeBaseListeners 'l_Members' registering, and for
  *  copying their members through the call to this method.
  *  May be empty but not null, and do not contain null element.
  */
  public static synchronized GroupID[] copyGroupsTree(
          GroupID _groupToCopy, GroupID _grp) {
    // initialization
    GroupID[] l_result = new GroupID[0];
    if (_groupToCopy == null || _grp == null ||
      _groupToCopy.getEorL_Members().length == 0) return l_result;
    
    ImmutableGroupIDMember[] l_toCopyMembers = _groupToCopy.getEorL_Members();
    GroupIDMember grpMember = null;
    GroupIDMember grpMemberCopy = null;
    GroupID[] result = null;
    GroupID parent = null;
    //GroupID group = null;
    GroupID[] l_groups = null;
    ImmutableName grp_name;
    
    Set l_copiedParent_one = new HashSet(INITIAL_CAPACITY);
    Set l_copiedParent_two = new HashSet(INITIAL_CAPACITY);
    Set l_toCopyMember_one = new HashSet(INITIAL_CAPACITY);
    Set l_toCopyMember_two = new HashSet(INITIAL_CAPACITY);
    Map m_toCopyMember_one = new HashMap(INITIAL_CAPACITY);
    Map m_toCopyMember_two = new HashMap(INITIAL_CAPACITY);

    l_copiedParent_one.add(_grp);
    for (int i = 0; i < l_toCopyMembers.length; i++ ) {
        l_toCopyMember_one.add(l_toCopyMembers[i]);
    }
    m_toCopyMember_one.put(_grp, l_toCopyMember_one);

    // copy all the members of a stage n, then all ones of the stage n+1,
    // until there is no groupID with members to copy
    while (!l_copiedParent_one.isEmpty()) {
        
        for (Iterator ite = l_copiedParent_one.iterator(); ite.hasNext(); ) {  
          parent = (GroupID)ite.next();
          l_toCopyMember_one = (Set)m_toCopyMember_one.get(parent);
          
          for (Iterator itu = l_toCopyMember_one.iterator(); itu.hasNext(); ) {
              grpMember = (GroupIDMember)itu.next();
              if (grpMember == null ||
                grpMember.getName().lastComponent() == null) continue;

              try {

                            // for testing and tracking
                            //System.out.println("in BaseUtilityImpl.copyGroupsTree(), member copy with name: "
                                  //+((Name)parent.getName()).extended(grpMember.getName().lastComponent())+
                                  //"; and ACS: "+_groupToCopy.getEorACS().getName()+
                                  //"; and groupID: "+parent+"; and parent ACS: "+parent.getEorACS().getName());
                  
                  // does not test if all the groups owning grpMember have been already copied
                  // and if not, does not create each copy, because only groups under _groupToCopy
                  // have to be copied.
                            /*
                  l_groups = (GroupID[])grpMember.getEorL_GroupID();
                  if (l_groups.length != 0) {
                      for (int i = 0; i < l_groups.length; i++) {
                          grp_name = ((Name)_grp.getEorACS().getName()).extended( 
                            l_groups[i].getName().lastComponent() );
                          if (_grp.getEorACS().containsOwnEligiblePartyName(grp_name))
                              continue;

                          grpMemberCopy = (GroupIDMember)grpMember.copy( grp_name,
                             (ACS)_grp.getEorACS(), null, null, Boolean.TRUE);

                          // add grpMemberCopy to l_result
                          result = new GroupID[l_result.length + 1];
                          System.arraycopy(l_result, 0, result, 0, l_result.length);
                          result[l_result.length + 1] = (GroupID)grpMemberCopy;
                          l_result = result;
                      } // for i 
                  } // if (l_groups.length != 0)
                             */
                  
                  grp_name = ((Name)_grp.getEorACS().getName()).extended(grpMember.getName().lastComponent());          
                  if (!_grp.getEorACS().containsOwnEligiblePartyName(grp_name)) {
                      
                      grpMemberCopy = (GroupIDMember)grpMember.copy( 
                        grp_name,
                        (ACS)parent.getEorACS(), null, Boolean.TRUE);
                      // if grpMember owns to _groupToCopy and if the last name component
                      // of _groupToCopy (it is an user choice) is not the same as _grp,
                      // grpMemberCopy has not be registered y copy() as member of _groupToCopy.
                      // So, a special test is needed
                      if (_groupToCopy.containsAsMemberWithoutAliasing(grpMember) &&
                        !_grp.containsAsMemberWithoutAliasing(grpMemberCopy))
                          _grp.addEorMember((GroupIDMemberImpl)grpMemberCopy);
                          
                              
                      if (grpMember instanceof GroupID) {
                          // add grpMemberCopy to l_result
                          result = new GroupID[l_result.length + 1];
                          System.arraycopy(l_result, 0, result, 0, l_result.length);
                          result[l_result.length] = (GroupID)grpMemberCopy;
                          l_result = result;
                      } // if (grpMember
                  
                  } // if (!_grp.getEorACS()
                  else {
                      grpMemberCopy = (GroupIDMember)
                        _grp.getEorACS().getEorM_EligibleParties().get(grp_name);
                      if (!parent.containsAsMember(grpMemberCopy)) {
                          parent.addEorMember((GroupIDMemberImpl)grpMemberCopy);
                      } // if
                  } // else

              } catch (CreateError c) {
                  throw new InternalError("in BaseUtilityImpl.copyGroupsTree():"+c.toString());
              }
              if (grpMember instanceof GroupID) {
                  
                  l_toCopyMembers = ((GroupID)grpMember).getEorL_Members();
                    if (l_toCopyMembers.length == 0) continue;
                    l_toCopyMember_two = new HashSet(INITIAL_CAPACITY);
                    for (int i = 0; i < l_toCopyMembers.length; i++ ) {
                        if (l_toCopyMembers[i] != null) l_toCopyMember_two.add(l_toCopyMembers[i]);
                    }

                            // for testing and tracking
                            //for (Iterator itetest = l_toCopyMember_two.iterator(); itetest.hasNext(); ) {
                                //System.out.println("in BaseUtilityImpl.copyGroupsTree(), "+
                                  //"a member is a GroupID which has the member: "+(GroupIDMember)itetest.next());
                            //}

                    m_toCopyMember_two.put(grpMemberCopy, l_toCopyMember_two);
                    l_copiedParent_two.add(grpMemberCopy);
              } // if (grpMember

          } // for itu
        } // for ite
        
        if (!l_copiedParent_two.isEmpty()) {
            l_copiedParent_one = (HashSet)((HashSet)l_copiedParent_two).clone();
            m_toCopyMember_one = (HashMap)((HashMap)m_toCopyMember_two).clone();
        } else {
            l_copiedParent_one.clear();
        } // else
        l_copiedParent_two.clear();
        m_toCopyMember_two.clear();
            
    } // while
    
    l_copiedParent_one = null;
    l_copiedParent_two = null;
    l_toCopyMember_one = null;
    l_toCopyMember_two = null;
    m_toCopyMember_one = null;
    m_toCopyMember_two = null;
    grpMember = null;
    grpMemberCopy = null;
    parent = null;
    l_groups = null;
    grp_name = null;
    
    return l_result;
  } // copyGroupsTree()

  /**
   * Gets the name of the BeanInfo class from the ISNodeImpl type.
   * For instance, with the ISFactoryUtilityImpl.INFORMATION_SYSTEM_NODE argument,
   * it returns the value 'ARoad0.gBaseBeanInfo.ISComponentBeanInfo'.
   * Called by ActionOpenGlobalStructure.getISnodePropertyDescriptors().
   * @param _text a node type in ISFactoryUtilityImpl
   * @return the right BeanInfo class name
   */
   public static synchronized final String getBeanInfoNameForISnode(String _text) {
      if (ISFactoryUtilityImpl.STRUCTURE_ROOT_STRING.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISStructureRootBeanInfo";

      if (!ISFactoryUtilityImpl.getNodeTypesForIS().contains(_text))
          throw new InternalError("BaseUtilityImpl.getBeanInfoNameForISnode()"+
                  " with a wrong argument.");

      if (ISFactoryUtilityImpl.INFORMATION_SYSTEM_NODE.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISComponentBeanInfo";
      else if (ISFactoryUtilityImpl.PHYSICAL_COMPONENT_NODE.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISComponentBeanInfo";
      else if (ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISComponentBeanInfo";
      else if (ISFactoryUtilityImpl.ACS_NODE.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISAcsBeanInfo";
      else if (ISFactoryUtilityImpl.SUB_ACS_NODE.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISSubACSBeanInfo";
      else if (ISFactoryUtilityImpl.COMPOSITE_ACS_NODE.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISCompositeBeanInfo";
      else if (ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE.equals(_text))
          return "ARoad0.gBaseBeanInfo.ISCompositeBeanInfo";
      else return null;

   } // getBeanInfoNameForISnode()

  /**
   * Adds a right, and sort the resulting set with StringRightImpl.compareTo()
   * to put the right in a decreasing order. If the current array is not ordered,
   * the right is added before the first occurence which is lesser than the right.
   * Called by ACS_Facade, RightsFactory_Facade and ACSFactoryImpl.
   * @param _l_currentRights sorted set. May be null or empty.
   * @param _newRight to add. May be null.
   * @return _l_currentRights containing _newRight if not null nor
   *          empty
   */
   public static synchronized final StringRight[] addRightInDecreasingSort(
            StringRight[] _l_currentRights, StringRight _newRight) {
        
        if (_newRight == null || _newRight.isEmpty()) return _l_currentRights;
        if (_l_currentRights == null || _l_currentRights.length == 0)
            return new StringRight[] {_newRight};
        
        StringRight[]   l_result = new StringRight[_l_currentRights.length + 1];
        boolean         inserted = false;
        
        for (int i = 0; i < _l_currentRights.length; i++) {
            if ( _l_currentRights[i].compareTo(_newRight) < 0 && !inserted) {
                l_result[i] = _newRight;
                l_result[i + 1] = _l_currentRights[i];
                inserted = true;
            }
            else if (!inserted) {
                l_result[i] = _l_currentRights[i];
            }
            else {
                l_result[i + 1] = _l_currentRights[i];
            }
        } // for i
        if (!inserted) l_result[l_result.length - 1] = _newRight;
 
        return l_result;
   } // addRightInDecreasingSort()
    
  /**
   * Adds a comparable object, and sort the resulting set with compareTo()
   * to put the object in a decreasing order. If the current array is not ordered,
   * the object is added before the first occurence which is lesser than the object.
   * @param _l_current sorted list of java.lang.Comparable to update. Never null.
   * @param _newObject to add. Never null.
   * @return updated _l_current with _newObject at the right place
   */
   public static synchronized List addObjectInDecreasingSort(
            List _l_current, Comparable _newObject) {
        
        Comparable          obj = null;
        int                 count = 0;
        
        
        if (!_l_current.isEmpty()) {
            // sorts from the greater to the lesser
            for (Iterator ite = _l_current.iterator(); ite.hasNext(); ) {
                obj = (Comparable)ite.next();
                if ( obj.compareTo(_newObject) < 0 ) {
                    break;
                }
                count++;
            } // for ite
        } // else

        _l_current.add(count, _newObject);


                              // for testing purpose
                        //System.out.println("in BaseUtilityImpl.addObjectInDecreasingSort(), "+
                          //"the added object "+_newObject+" is put at: "+count+" in the list of "+
                          //_l_current.size()+" items");
        
        
        obj = null;
        return _l_current;
   } // addObjectInDecreasingSort()

  /**
  * Analyzes the directory rights to say if the AG inherited rights which
  * are derived from the directory rights allow some access to the child.
  * Return true if the Directory AG rights contain an equivalent to the metarights
  * 'gothrough_for_directory', 'read_for_directory' or 'full_control'.
  * On the contrary of ACL and PRI inherited rights, the AG rights inheritance
  * requires that each parent node allows the go accross it, to have an access.
  * Called by the Resource constructor, Directory.setL_ownerRights() and setL_groupRights().
  * @param _l_rights rights of a Directory
  * @return true if the rights allow an access to the Directory children
  **/
  public static synchronized boolean withAcrossToDirectoryAGRights(StringRight[] _l_rights) {
      if (_l_rights == null || _l_rights.length == 0) return false;
      boolean result = false;
      StringRight right;
      StringRight got = BaseManagerImpl.getInstance().getMetaRight("gothrough_for_directory");
      StringRight read = BaseManagerImpl.getInstance().getMetaRight("read_for_directory");
      StringRight ctrl = BaseManagerImpl.getInstance().getMetaRight("full_control");      
       
      // test rights
      for (int k = 0; k < _l_rights.length; k++) {
        if (_l_rights[k] == null) continue;
        right = (StringRight)_l_rights[k].getEorMetaRight();
        if (right == null) continue;
        if (right.equals(got) || right.equals(read) || right.equals(ctrl)) {
                result = true;
                break;
        }
      } // for ite

              // for testing and tracking       
              //if (_l_rights != null && _l_rights[0].getRight().equals("xdir_Node"))
                //  System.out.println("in BaseUtilityImpl.withAcrossToDirectoryAGRights(), result for 'xdir_Node' at 0: ");
                //  System.out.println("----------------------- "+result);
      
      right = null; 
      got = null;
      read = null;
      ctrl = null;
      return result;
  } // withAcrossToDirectoryAGRights()

  /**
  * Gets all the direct and all the indirect members of a group.
  * No use in the core classes.
  * @see #getAllIndirectGroupMembersAndRunningUnderActors(ARoad0.gBaseInterface.GroupID) getAllIndirectGroupMembersAndRunningUnderActors
  * @param _grp group to analyse
  * @return set of the direct and indirect GroupIDmembers.
  *  May be empty but not null, and does not contain null element.
  */
  public static Set<ImmutableGroupIDMember> getAllIndirectGroupMembers(GroupID _grp) {

    Set<ImmutableGroupIDMember> l_result = new HashSet(30);
    if (_grp == null || _grp.getEorL_Members().length == 0) return l_result;

    // variables
    Set                         l_initialGroups = null;
    ImmutableGroupIDMember[]    l_members = _grp.getEorL_Members();

    for (int i = 0; i < l_members.length; i++) {
        l_result.add(l_members[i]);
        if (l_members[i] instanceof GroupID) {
            if (l_initialGroups == null)
                l_initialGroups = new HashSet(INITIAL_CAPACITY);
            l_initialGroups.add(l_members[i]);
        } // if
    } // for i

    if (l_initialGroups == null || !_grp.getEorACS().getManageGroupTrees()) {
        l_members = null;
        l_initialGroups = null;
        return l_result;
    } // if

    // variables
    Set                         l_groups;
    Set                         l_nextGroupsToTest = (Set) ((HashSet)l_initialGroups).clone();
    Set                         l_testedGroups = new HashSet(30);
    ImmutableGroupIDMember[]    l_groupmembers;
    ImmutableGroupID            grp;
    int                         iterationsCount = 0;

    // Loop to get all the indirect groupIDmembers.
    // Stops when there is no new group to test.
    // Each iteration tries to add GroupIDMembers
    // to the result, through GroupID.getEorL_Members().
    while (!l_nextGroupsToTest.isEmpty() && iterationsCount < 40) {

      iterationsCount++;

                        // for testing and tracking
                        //System.out.println("in BaseUtilityImpl."+
                        //  "getAllIndirectGroupMembers(), THE CURRENT STATE IS:");
                        //System.out.println("---_grp: "+_grp.getName());
                        //System.out.println("--- iterationsCount: "+iterationsCount);
                        //System.out.println("------- l_nextGroupsToTest: ");
                        //for (Iterator itetest = l_nextGroupsToTest.iterator(); itetest.hasNext();) {
                        //  System.out.println("--------------- : "+itetest.next().toString());
                        //}

      l_groups = (Set) ((HashSet)l_nextGroupsToTest).clone();
      l_nextGroupsToTest.clear();

      for (Iterator ito = l_groups.iterator(); ito.hasNext();) {

            grp = (ImmutableGroupID) ito.next();
            l_testedGroups.add(grp);
            l_groupmembers = grp.getEorL_Members();

            for (int j = 0; j < l_groupmembers.length; j++) {
              if (l_groupmembers[j] == null) continue;

                                // for testing and tracking
                                //System.out.println("in BaseUtilityImpl."+
                                //  "getAllIndirectGroupMembers(), start of "+
                                //  "the loop for l_groupmembers[j] "+
                                //  l_groupmembers[j].getName().lastComponent());

              if (l_groupmembers[j] instanceof GroupID && !l_testedGroups.contains(l_groupmembers[j]))
                  l_nextGroupsToTest.add(l_groupmembers[j]);

              l_result.add(l_groupmembers[j]);
            } // for j

      } // for ito

    } // while

    if (iterationsCount == 40) throw new InternalError("in BaseUtilityImpl."+
                          "getAllIndirectGroupMembers(), "+
                          "the number of iterations is greater than 40.");

    grp = null;
    l_groups = null;
    l_members = null;
    l_nextGroupsToTest = null;
    l_testedGroups = null;
    l_initialGroups = null;

    return l_result;
  } // getAllIndirectGroupMembers()
  
  /**
  * Gets all the direct and indirect members of a group, and all the running-under
  * actors for which this group is a primary or a secondary group.
  * Called by NoThanViewInBaseImpl.setUpContextLists().
  * @see #getAllIndirectGroupMembers(ARoad0.gBaseInterface.GroupID) 
  * @param _grp is the group to analyze
  * @return set of groupIDmembers or Actors.
  *  May be empty but not null, and does not contain null element.
  */
  public static synchronized Set<ImmutableEligibleParty>
          getAllIndirectGroupMembersAndRunningUnderActors(GroupID _grp) {
      
    Set<ImmutableEligibleParty> l_result = new HashSet(30); 
    if (_grp == null || _grp.getEorL_Members().length == 0) return l_result;
    
    // variables
    Set                         l_initialGroups = new HashSet(30);
    ImmutableGroupIDMember[]    l_members = _grp.getEorL_Members();
    Actor[]                     l_act = null;
    
    for (int i = 0; i < l_members.length; i++) {
        l_result.add(l_members[i]);
        if (l_members[i] instanceof GroupID) {
            l_initialGroups.add(l_members[i]);
        }

        // test of running-under actors of _grp
        l_act = _grp.getEorL_GroupIDCurrentActors();
        for (int j = 0; j < l_act.length; j++) {
          l_result.add(l_act[j]);
        } // for j  
        l_act = _grp.getEorL_GroupIDCurrentActorsAsSecondaryGroup();
        for (int j = 0; j < l_act.length; j++) {
          l_result.add(l_act[j]);
        } // for j 
    } // for i
    
    if (l_initialGroups.isEmpty() || !_grp.getEorACS().getManageGroupTrees()) {
        l_members = null;
        l_initialGroups = null;
        l_act = null;
        return l_result;
    }

    // variables
    Set                         l_groups;
    Set                         l_nextGroupsToTest = (Set) ((HashSet)l_initialGroups).clone();
    Set                         l_testedGroups = new HashSet(30);
    ImmutableGroupIDMember[]    l_groupmembers;
    ImmutableGroupID            grp;
    int                         iterationsCount = 0;


    // Loop to get all the indirect groupIDmembers. 
    // Stops when there is no new group to test.
    // Each iteration tries to add GroupIDMembers
    // to the result, through GroupID.getEorL_Members().
    while (!l_nextGroupsToTest.isEmpty() && iterationsCount < 40) {

      iterationsCount++;

                        // for testing and tracking
                        //System.out.println("in ACS_Facade."+
                        //  "getAllIndirectGroupMembers(), THE CURRENT STATE IS:");
                        //System.out.println("---_grp: "+_grp.getName());
                        //System.out.println("--- iterationsCount: "+iterationsCount);
                        //System.out.println("------- l_nextGroupsToTest: ");
                        //for (Iterator itetest = l_nextGroupsToTest.iterator(); itetest.hasNext();) {
                        //  System.out.println("--------------- : "+itetest.next().toString());
                        //}

      l_groups = (Set) ((HashSet)l_nextGroupsToTest).clone();
      l_nextGroupsToTest.clear();

      for (Iterator ito = l_groups.iterator(); ito.hasNext();) {

            grp = (ImmutableGroupID) ito.next();
            l_testedGroups.add(grp);
            l_groupmembers = grp.getEorL_Members();

            for (int j = 0; j < l_groupmembers.length; j++) {
              if (l_groupmembers[j] == null) continue;

                                // for testing and tracking
                                //System.out.println("in ACS_Facade."+
                                //  "getAllIndirectGroupMembers(), start of "+
                                //  "the loop for l_groupmembers[j] "+
                                //  l_groupmembers[j].getName().lastComponent());

              if (l_groupmembers[j] instanceof GroupID &&
                      !l_testedGroups.contains(l_groupmembers[j]))
                  l_nextGroupsToTest.add(l_groupmembers[j]);

              l_result.add(l_groupmembers[j]);
            } // for j
        
            // test of running-under actors of grp
            l_act = grp.getEorL_GroupIDCurrentActors();
            for (int i = 0; i < l_act.length; i++) {
              l_result.add(l_act[i]);
            } // for i  
            l_act = grp.getEorL_GroupIDCurrentActorsAsSecondaryGroup();
            for (int i = 0; i < l_act.length; i++) {
              l_result.add(l_act[i]);
            } // for i 
        
        
      } // for ito

    } // while

    if (iterationsCount == 40) throw new InternalError("in BaseUtilityImpl."+
                          "getAllIndirectGroupMembersAndRunningUnderActors(), "+
                          "the number of iterations is greater than 40.");

    grp = null;
    l_groups = null;
    l_members = null;
    l_nextGroupsToTest = null;
    l_testedGroups = null;
    l_initialGroups = null;
    l_act = null;

    return l_result;
  } // getAllIndirectGroupMembersAndRunningUnderActors()    
   
  /**
   * Gets the child inherited rights from the AG directory rights, for the children.
   * This is a wrapper for getChildRightsSetFromParentAGRights().
   * It is protected to be reserved to the gBase classes, from the generic gBase or
   * the AcsAddon gBases.
   * Called by ResourceImpl.getL_inheritedOwnerRights() and getL_inheritedGroupRights(),
   * and the subclasses.
   * @param _l_parentRights acsrights or metarights of a directory. Never null.
   * @param _acs of the Resource calling this method. Never null.
   * @return the child rights, without the rights in _l_parentRights which are not
   *     directory rights. May be a null-length array but not null.
   */
   protected static synchronized StringRight[] getChildRightsFromParentAGRights(
            StringRight[] _l_parentRights, ImmutableACS _acs) {
        
      Set<StringRight>      l_result = getChildRightsSetFromParentAGRights(
                                    _l_parentRights, _acs);

      if (!l_result.isEmpty()) return l_result.toArray(new StringRight[0]);
      else {
          l_result = null;
          return new StringRight[0];
      }
    } // getChildRightsFromParentAGRights()
   
  /**
   * Gets the child inherited rights from the AG directory rights, for the children.
   * It is protected to be reserved to the gBase classes, from the generic gBase or
   * the AcsAddon gBases.
   * Uses the pattern 'directory/child' in the metaright names.
   * For instance, the metaright 'write_for_nxdirectory' has the image
   * 'write_for_nxchild' which has the lower rights 'createchild' and 'deletechild',
   * and these rights as arguments produce, through this method, the children rights
   * 'create' and 'delete'.
   * There is no way to use 'xxxchild' metarights as effective rights. 
   * They are just the intermediate values for the processing, and they are desactivated.
   * The metaright 'full_control' has a dedicated processing since it has two images
   * 'full_controlchild' and 'full_controlnxchild', which works as the sum 
   * of the images 'read_for_child' and 'write_for_child' for the first image,
   * and 'read_for_child', 'write_for_nxchild' and 'execute_for_nxchild' for the second image.
   * The argument _acs is used then for choosing the relevant rights.
   * It is the same way for 'deny_all' and its two images
   * 'deny_allchild' and 'deny_allnxchild'.
   * Called by getChildRightsFromParentAGRights() and the subclasses.
   * @param _l_parentRights acsrights or metarights of a directory. Never null.
   * @param _acs of the Resource calling this method. Never null.
   * @return the child rights, without the rights in _l_parentRights which are not
   *     directory rights. May be an empty set but not null.
   */
   protected static synchronized Set<StringRight> getChildRightsSetFromParentAGRights(
            StringRight[] _l_parentRights, ImmutableACS _acs) {
        
      Set<StringRight>      l_result = new HashSet(INITIAL_CAPACITY);
      StringRight[]         l_lowerRights = null;
      String                value = null;
      StringRight           right = null;
      int                   size = 0;
      boolean               error = false;
      int                   index = -1;

      if (!_acs.getManageAGORightsInheritance()) return l_result;

                            //System.out.println("BaseUtilityImpl.getChildRightsSetFromParentAGRights()");

      for (int i = 0; i < _l_parentRights.length; i++) {

                                  // for testing purpose
                            //System.out.println("BaseUtilityImpl.getChildRightsSetFromParentAGRights(), "+
                              //"for "+i+", the value _l_parentRights[i] is "+_l_parentRights[i]+
                              //" and the metaright is "+_l_parentRights[i].getEorMetaRight());

          value = _l_parentRights[i].getEorMetaRight().getRight();          
          
          if (value.equals("full_control")) {
              if (_acs.getACSType().indexOf(ACSImpl.LINUX) != -1
            ) {
                  value = "full_controlnxdirectory";
              } else {
                  value = "full_controldirectory";
              }
          }
          else if (value.equals("deny_full_control")) {
              if (_acs.getACSType().indexOf(ACSImpl.LINUX) != -1 ) {
                  value = "deny_allnxdirectory";
              } else {
                  value = "deny_alldirectory";
              }
          } // else
          
          // rights which are directory rights are  NOT added (as for instance 'cp'
          // from a subtraction of 'dl' and 'rn' from 'xdir' in a Linux Ubuntu)
          if (!value.endsWith("directory")) {
              //l_result.add( _l_parentRights[i] );
              continue;
          }
          
          // selects the 'child' image of _l_parentRights
          size = value.length();
          right = BaseManagerImpl.getInstance().getMetaRight(
                  value.substring(0, size - 9)+"child" );
          if (right == null) continue;
          
          // right may contain 'child', like in the metaright 'deleteRenameCopychild'
          l_lowerRights = right.getLowerRights();
          // adds the lower rights having 'child' in name, in removing 'child',
          // searches an acsright first or adds the metaright
          for (int j = 0; j < l_lowerRights.length; j++) {
              
              if (l_lowerRights[j].getRight().endsWith("child")) {
                  size = l_lowerRights[j].getRight().length();
                  value = l_lowerRights[j].getRight().substring(0, size - 5);

                                  // for testing purpose
                            //System.out.println("BasetUtilityImpl.getChildRightsSetFromParentAGRights(), "+
                              //"for "+l_lowerRights[j]+" the value is "+value);
                  
                  right = _acs.getFirstRightFromMeta(value); // the first matching right
                  if (right != null) l_result.add( right );
                  else {
                      right = BaseManagerImpl.getInstance().getMetaRight(value);
                      l_result.add( right );
                  }
                  if (right == null) {
                      error = true;
                      index = i;
                      break;
                  }
              } // if endsWith("child")

          } // for j
      } // for i 
      
      if (error) {
            throw new InternalError("BaseUtilityImpl.getChildRightsSetFromParentAGRights() - "+
              "the first error in the process is for "+_l_parentRights[index]);          
      }

      l_lowerRights = null;
      value = null;
      right = null;
      return l_result;
    } // getChildRightsSetFromParentAGRights()
   
  /**
   * Get the right from the string returned by getName().toString(), both
   * for StringRightImpl and StringMetaRightImpl instances.
   * Do not handle names from StringRightImpl.getFullName(), but it operates
   * on StringMetaRightImpl.getFullName() since the full name is the name in metarights.
   * If _acs is null for an acsright, the ACS is searching through the open ACS. 
   * So, if it is an acsright during the creation of the ACS, the ACSFactory
   * has to be get to found the right.
   * Called by ActionBeamer, SimplePropertyEditor, BaseObjectHandler.
   * @param _text is the name, as get through getName().toString()
   * @param _acs is the ACSFactory or the ACS which may own the right if it is
   *   an ascright. May be null.
   * @return the right. May be null.
   * @exception BaseError if the string argument is not from a name, or
   *   if it is not a known right in an open ACS, or from NameImpl.getNameFromString(_text).
   */
    public static synchronized final StringRight getRightFromNameAsString(
            String _text, ACSRun _acs) throws BaseError {
      // variables
      StringRight   result = null;
      Name          buffer = null;
      Name          acsName = null;
      ACSRun        acs = null;

      // exception and simple cases
      if (_text == null || _text.length() == 0) 
          throw new BaseError(
                "in StringRightUtilityImpl.getRightFromNameAsString(), "+
                "the string argument is null or empty.");
      
      buffer = NameImpl.getNameFromString(_text);
      
                    // for testing
                    //System.out.println("in UtilityImpl.getRightFromNameAsString(), buffer: "+buffer);
                    
      
      if (!buffer.component(buffer.noOfComponents() - 2).equals("<VALUE>") || 
              !buffer.component(0).startsWith("<FOR") )
          throw new BaseError(
                "in StringRightUtilityImpl.getRightFromNameAsString(), "+
                "the argument is not from StringRight.getName().");
      
      // if it is a StringRightImpl in an ACS or in an ACSFactory
      if (buffer.component(0).equals("<FOR_ACS>")) {
          if (  buffer.component(1).equals("no_acs"))
              throw new BaseError(
                    "in StringRightUtilityImpl.getRightFromNameAsString(), "+
                    "the right is embbeded in a metaright and not accessible.");  
          else acsName = buffer.subname(1, buffer.noOfComponents() - 2);
          acs = _acs;
          if ( acs == null)
              acs = (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(acsName);
          if ( acs == null)
              throw new BaseError(
                    "in StringRightUtilityImpl.getRightFromNameAsString(), "+
                    "the ACS of this acsright is not open, and its ACS has not been get."); 
      
                    // for testing
                    //System.out.println("in UtilityImpl.getRightFromNameAsString(), right: |"+
                      //      buffer.component(buffer.noOfComponents() - 1)+"|");
          
          
          result = acs.getOneRight(buffer.component(buffer.noOfComponents() - 1));
      } else { // it is a metaright
          result = BaseManagerImpl.getInstance().getMetaRightFromName(buffer);
      } // else

      buffer = null;
      acsName = null;
      acs = null;      
      return result;
    } // getRightFromNameAsString()
    
  /**
   * Get the ACS name from the acsright name. Not applicable to metarights.
   * Called by ActionOpenBeamer and BeamerManager.
   * @param _rightName name of an acsright
   * @return the ACS name. Never null.
   * @exception BaseError if it is not an acsright name.
   */
    public static synchronized final ImmutableName getAcsNameFromName(ImmutableName _rightName)
            throws BaseError {

      // exception and simple cases
      if (_rightName.isEmpty() || !_rightName.component(0).equals("<FOR_ACS>") ||
              !_rightName.component(_rightName.noOfComponents() - 2).equals("<VALUE>")) 
          throw new BaseError(
                "in StringRightUtilityImpl.getAcsNameFromName(), "+
                "the argument is not an acsright name or is empty.");
   
      return _rightName.subname(1, _rightName.noOfComponents() - 2);
    } // getRightFromNameAsString()
    
  /**
   * Called by BeamerManager. Do not check the presence of the acsright in its ACS.
   * @param _rightName name
   * @return true if it is an acsright name
   */
    public static synchronized final boolean isAcsRightName(ImmutableName _rightName) {

      // exception and simple cases
      if (_rightName.isEmpty() || !_rightName.component(0).equals("<FOR_ACS>") ||
              !_rightName.component(_rightName.noOfComponents() - 2).equals("<VALUE>")) 
          return false;

      else return true;
    } // isAcsRightName()
    
  /**
   * Called by ACSImpl and PrivilegeForLinksImpl.finalizeForBase(),
   * to check the presence of the ACSObject named _objName
   * as a source of an external AclEntry or an external Linked Privilege, if _forSource
   * is true, or as a target, also for an external AclEntry or an external linked Privilege.
   * For an Actor, this method may be called twice to get a complete result.
   * Caution: this method does not check if _objName is an _acs object, while it is forbidden.
   * @param _acs ACS to analyze
   * @param _objName name of the ACSObject to search
   * @param _forSource true to search an EligibleParty, and false to search an ImmutableTarget
   * @return true if _objName is the name of an external ACSObject in _acs
   */
    public static final boolean isExternalACSobject(ACSImpl _acs, ImmutableName _objName,
            boolean _forSource) {

        boolean result = false;

        if (_forSource && _acs.l_AclEntry_ != null) {
            for (int i = 0 ; i < _acs.l_AclEntry_.length ; i++) {
                if ( _objName.equals( _acs.l_AclEntry_[i].getEpName()) ) {
                  result = true;
                  break;
                }
            } // for i
        } // if (_forSource &&

        if (!result && _forSource && _acs.l_Privilege_ != null) {
            for (int i = 0 ; i < _acs.l_Privilege_.length ; i++) {
                if (_acs.l_Privilege_[i].getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE) &&
                        _objName.equals( ((PrivilegeForLinks)_acs.l_Privilege_[i]).
                        getSingleSourceName()) ) {
                      result = true;
                      break;
                }
            } // for i
        } // if (!result &&

        if (!_forSource && _acs.l_AclEntry_ != null) {
            for (int i = 0 ; i < _acs.l_AclEntry_.length ; i++) {
                if ( _objName.equals( _acs.l_AclEntry_[i].getResName()) ) {
                  result = true;
                  break;
                }
            } // for i
        } // if (!_forSource &&

        if (!result && !_forSource && _acs.l_Privilege_ != null) {
            for (int i = 0 ; i < _acs.l_Privilege_.length ; i++) {
                if (_acs.l_Privilege_[i].getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE) &&
                        _objName.equals( ((PrivilegeForLinks)_acs.l_Privilege_[i]).
                        getSingleTargetName()) ) {
                      result = true;
                      break;
                }
            } // for i
        } // if (!result &&

        return result;
    } // isAcsRightName()

} // BaseUtilityImpl