//Title:        Access Road
//Version:      0.7.0
//Copyright:    2009-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.BaseError;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * This class is responsible for modeling the member of a folder (VirtualFolder
 * or ViewInBase) in an ACS, and it ensures the handling of Privileges.
 * It is the super class of EligibleParty and Resource.
 * To let the PrivilegeSource capacity to EligibleParty, this class does not
 * implement PrivilegeSource. Nonetheless, it handles fully privilege rights
 * so its subclasses have not to do it.
 * The PrivilegeForTypes are managed directly by this class.
 * The PrivilegeForLinks are also managed directly, or with cascading
 * in  the subclass Resource.<P>
 * This class implements the common alias methods for its subclasses which
 * are ImmutableSource or ImmutableTarget.
 * There are two types of alias relation this class may handle: <P>
 * - the internal hard alias produces a direct alias-to-reference relation
 * in the same ACS, and the two objects know this relation, <P>
 * - the soft alias produces an indirect reference-to-alias relation
 * where the alias is external, and it does not register this relation.
 * It is the responsibility of the ACS associations or the AcsAddon code
 * to set an internal alias for one new ACSObject, as an alias or as an alias reference.
 * On the other hand, the soft alias relation may be only set by an AcsAddon.
 * <P>
 * This class is a javabean with the following bound properties:
 * 'Comment', 'ImpliedOpenViews', 'VirtualFolders', 'ResourceType' or 'EPType',
 * 'M_ConstrainedChildACS', 'HardAliasReference', 'HardAlias', 'SoftAlias',
 * 'AcsOfSoftReferences' and 'hasSoftAlias'. <P>
 * The listeners are proper to this instance.
 * Excepting for EPRViewInBaseImpl, they are outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched.
 * <P>
 * This class is a composite class which uses the following embedded classes
 * as private attributes:
 * - CFolderMemberImpl for handling VirtualFolders and ViewInBases <P>
 * - optional CPrivilegeUserImpl for handling PrivilegeForTypes and PrivilegeForLinks <P>
 * - optional CAliasImpl for handling the hard and soft aliasing relations <P>
 * This class has several subclasses in this package.
 * All variables are private to secure the white box inheritances from this class,
 * excepting for final variables and actorActivated_, for which the value is associated
 * to the class and has to be set by the Actor constructors. <P>
 * This class has several subclasses in this package and in the AcsAddon packages.
 * Most of the variables are protected.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see CFolderMemberImpl
 * @see CPrivilegeUserImpl
 * @see CAliasImpl
 * @version 0.7.0
 */
public class BasicImpl implements Basic, Serializable {

  private static final long   serialVersionUID = 0;
  //  private Name[]          l_Access_;
  protected ImmutableName     name_;
  protected String            comment_;
  //  private Icon              icon_;
  protected ImmutableACS      aCS_;  // main ACS; External Object Reference
  protected ImmutableName     iS_Name_;
  protected ImmutableName     aCS_Name_;
  protected static final int  PRIME = 1000003;
  protected static final int  INITIAL_CAPACITY = 10; // the default value: 101 for map
  protected static final int  INITIAL_CAPACITY_2 = 4 ;  // the default value is 10
  protected static final int  INCREMENT_CAPACITY = 2;
  /**
   * This is the Resource type or the EligibleParty type,
   * to set by one of the relevant constructors.
   * The type is used by PrivilegeForType instances.
   */
  protected String            type_= "<undefined>";
  
  // Protected variables -------------------------------------

  /**
   * A constrained (opened or closed) child ACS has one or several of
   * the following relations to this object:
   * &lt;ForChildAcsOpeningAndClosing&gt;, &lt;ForChildAcsOpeningNoClosing&gt;
   * or &lt;ForChildAcsIntegrity&gt;.
   * Each relation is associated to a key with the same name.
   */
  protected HashMap<String, ImmutableName[]> m_ConstrainedChildACS_ = null;

  /**
   * folders handler of this instance, both for view links and optionally,
   * for virtual folder links.
   */
  protected CFolderMemberImpl   folderMember_;
  /**
   * This optional privileges handler is used only
   * by the subclasses, EligiblePartyImpl and ResourceImpl.
   */
  protected CPrivilegeUserImpl  privHandler_ = null;
  /**
   * Optional alias handler of this instance.
   */
  protected CAliasImpl          aliasHandler_ = null;

  // -------- transient properties -------------------------------

   /**
   * Manages all the property change listeners of this class and its subclasses.
   * GetPropertyChangelisteners() allows the AcsAddon to have the listeners for
   * a given property.
   */
  protected transient PropertyChangeSupport changeSupport_ = null;

  // ----------- public static properties --------------------------------

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final BasicImpl EMPTY_INSTANCE = new BasicImpl();

  /** 
   * variable for the JDK 2 serialization
   * @serial serialVersionUID, name_, aCS_, iS_Name_, aCS_Name_,
   * comment_, privHandler_, folderMember_, type_, m_ConstrainedChildACS_,
   * aliashandler_
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("serialVersionUID", Long.TYPE),
      new ObjectStreamField("name_", NameImpl.class),
      new ObjectStreamField("aCS_", ACSImpl.class),
      new ObjectStreamField("iS_Name_", NameImpl.class),
      new ObjectStreamField("aCS_Name_", NameImpl.class),
//      new ObjectStreamField("l_Access_", NameImpl[].class),
      new ObjectStreamField("comment_", String.class),
//      new ObjectStreamField("icon_", IconImpl.class),
      new ObjectStreamField("privHandler_", CPrivilegeUserImpl.class),
      new ObjectStreamField("folderMember_", CFolderMemberImpl.class),
      new ObjectStreamField("type_", String.class),
      new ObjectStreamField("m_ConstrainedChildACS_", Map.class),
      new ObjectStreamField("aliasHandler_", CAliasImpl.class)
  };

   /**
   * Default constructor. No registration in the ACS.
   */
  public BasicImpl() {

                          // for testing purpose
                     //System.out.println("Call to BasicImpl.BasicImpl()");

      name_ = NameImpl.EMPTY_INSTANCE;
      iS_Name_ = NameImpl.EMPTY_INSTANCE;
      aCS_Name_= NameImpl.EMPTY_INSTANCE;
        //      l_Access_ = new NameImpl[1];
      comment_ = "";
        //      icon_ = new Icon();
      folderMember_ = null;
  }

   /**
   * Used for the creation of a Basic by the subclass constructors.
   * For an UserID or a GroupID, adds '&lt;U&gt;' or '&lt;G&gt;' between the ACS name
   * and the end of _name. Controls if the name is not a virtual folder name or an ACS name.
   * Used by copy().
   * The properties privHandler_ and aliasHandler_ remains null.
   * The user error messages are the followings: <P>
   * - The name of the new instance is null. <P>
   * - The name of the new instance contains a character that is not a letter. <P>
   * - The found IS name is null. <P>
   * - This is the name of a node or an area in the Information System, or the name
   * of a virtual folder in the same ACS.
   * Caution: this constructor does not register the new instance
   * in the ACS; it is done by the subclasses.
   * @param _name of the eligible party to create, before the adding of the component
   *    '&lt;G&gt;' if it is a GroupID or '&lt;U&gt;' if it is an UserID. Never null.
   * @param _acs is the main ACS of the Basic to create. Never null.
   * @exception CreateError if the instance name is null,
   * if it contains a character that is not a letter, or contains '::' or '&gt;&gt;',
   * if the IS name is unknown in the ACS name,
   * if it is the name of a node or area in the information system, or the name of a virtual folder.
   */
  protected BasicImpl(ImmutableName _name, ACS _acs) throws CreateError {

      this();

      if (_name == null) {
        iS_Name_ = null;
        aCS_Name_ = null;
        throw new CreateError(" The name of the new instance is null.");
      } //if
      char[] l_chars;
      boolean containsFaultyValue = false;
      for (int j = 0; j < _name.noOfComponents(); j++) {
          l_chars = _name.component(j).toCharArray();
          for (int i = 0; i < l_chars.length; i++) {
            if (Character.isIdentifierIgnorable(l_chars[i]) ||
             Character.isISOControl(l_chars[i])) {
                  containsFaultyValue = true;
                  System.out.println("in the BasicImpl constructor, a character "+
                        "is ignorable or is an ISO control: "+l_chars[i]);
                  break;
            }
          } // for i
          if (_name.component(j).indexOf("::") != -1 ||
                  _name.component(j).indexOf(">>") != -1)
                        containsFaultyValue = true;
          if (containsFaultyValue) break;
      } // for j
      if (containsFaultyValue) {
        iS_Name_ = null;
        aCS_Name_ = null;
        l_chars = null;
        throw new CreateError(" The name contains a character "+
          "that is not a letter, or contains '::' or '>>'.");
      } //if
      l_chars = null;

      aCS_ =  (ImmutableACS) _acs;
      aCS_Name_ = aCS_.getName();
      
      if (this instanceof UserID) {
          // adds '<U>' between the ACS name and the end of _name
          name_ = ((Name)aCS_Name_.clone()).extended("<U>").extended(
                     _name.subname(
                       aCS_Name_.noOfComponents(), _name.noOfComponents()));
      } else if (this instanceof GroupID) {
          // adds '<G>' between the ACS name and the end of _name
          name_ = ((Name)aCS_Name_.clone()).extended("<G>").extended(
                     _name.subname(
                       aCS_Name_.noOfComponents(), _name.noOfComponents()));

              // for testing and tracking
              //System.out.println("Call of BasicImpl constructor "+
                //"for: "+name_+", with the initial argument name: "+_name);
      }
      else name_ = (ImmutableName) _name.clone();

      iS_Name_ = BaseManagerImpl.getInstance().getISofName(aCS_Name_);

      if (iS_Name_ == null) {
        aCS_ = null;
        name_ = null; 
        iS_Name_= null;
        aCS_Name_= null;
        comment_ = null;
        throw new CreateError(" The found IS name is null.");
      }
      if (BaseManagerImpl.getInstance().getISTree(iS_Name_).isNode(name_) ||
              BaseManagerImpl.getInstance().getISTree(iS_Name_).isArea(name_) ||
              aCS_.getOneVirtualFolder(name_) != null) {
        aCS_ = null;
        name_ = null; 
        iS_Name_= null;
        aCS_Name_= null;
        comment_ = null;
        throw new CreateError(" This is the name of a node or an area"+
                " in the Information System, or the name of a virtual folder in the same ACS.");
      }
      
      // privHandler_ is created only if needed, in the subclasses
      // aliasHandler_ is created only if needed
      folderMember_ = new CFolderMemberImpl(this);
      
                          // for testing purpose
                     //System.out.println("Call to BasicImpl constructor with"+
                        // " (ImmutableName _name, ACS _acs)");
      
  } // BasicImpl constructor
  
 //
 // -----------  Methods of the implemented interfaces --------------------
 //

  
  /**
  * This method provides a copy of the instance
  * with another name. Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Most of the properties are copied, but not the list of ViewInBase
  * that contains this instance. Most of the privileges are copied.
  * Caution: does not copy the privilege if _result
  * does not belong to _acs and if there are no privilege
  * in the _result ACS which has a similar name - that is that
  * have the same short privilege name.
  * Calls copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name is a copy name. Usually creates as an extension of _parent name.
  * @param _acs is a main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used.
  * @param _toRegisterInManager is not used.
  * @return BasicImpl
  * @exception CreateError of the constructor with the same parameters
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError(
            "The name or the acs is null.");
      }
      
      BasicImpl result = new BasicImpl(_name, _acs);

      result = copyOwnVariables(result); 
      
      return result;
  } // copy(,,,)

  /**
   * The Basic owns to an access control system (ACS) which owns
   * to an information system (IS). IS name is usually the first component
   * of the ACS name. Not described in the BeanInfo.
   * @return ImmutableName of the IS
   */
  public ImmutableName getIS_Name()	{
          return (ImmutableName) iS_Name_.clone();
  }

  /**
   * @return ImmutableACS is an External Object Reference to the ACS
   * that owns this Basic (main ACS)
   */
  public ImmutableACS getEorACS()	{
          return aCS_;
  }

  /**
   * @return ImmutableName of the ACS that owns this object
   */
  public ImmutableName getAcsName() {
    return (ImmutableName)aCS_Name_.clone();
  }

/**
 * @return
 */
/*	public Icon getIcon()
    {
            return (Icon) icon_.clone();
    }
*/

    /**
    * Set any comment related to this Basic. It is a bound property.
    * @param _s is a comment which is associated to the Basic.
    */
    public synchronized void setComment(String _s)	{
        String oldComment = comment_;
        comment_ = _s;

        try {
              firePropertyChange("Comment", oldComment, comment_);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BasicImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
        }
        oldComment = null;
        return;
    }

    /**
     * @return comment which is associated to the Basic. May be null.
     */
    public String getComment()	{
       return comment_;
    }

    /**
    * The name is unique for the Access Road program.
    * @return a copy of the Basic name. May be empty, but not null.
    */
    public ImmutableName getName()	{
       return (ImmutableName) name_.clone();
    }

  /**
   * The full name is unique for the Access Road program.
   * It is there equal to the name returned by getName().
   * @return a copy of the Basic full name. May be empty, but not null.
   */
    public ImmutableName getFullName() {
       return (ImmutableName) name_.clone();
    }
    
  /**
   * This short name is NOT unique for the Access Road program nor the ACS if
   * it is an ACS object.
   * It is often the last component name.
   * @return the Basic nick name. May be null.
   */
    public String getNickName() {
       return name_.lastComponent();
    }

  /**
    * The type is defined at the creation of the object, then it may be updated by the user.
    * It has several uses: <P>
    * - to define the type updating policy from the ACS, <P>
    * - for defining the authorized rights for each kind of rights (ACL, AGO...), <P>
    * - to manage the capacity to delete the instance, <P>
    * - and by the PrivilegeForType instances to apply their rights. <P>
    * It is the resource type if the object is a Resource, the EP type
    * if it is an EP. So, this method from Immutable Source and ImmutableTarget
    * (not from ImmutableBasic) is for the Basic subclasses.
    * @return String of type, as for exemple "executable" or "file".
    *  Never null.
   */
  public String getType() {
    return type_;
  }

  /**
   * Get the unique ACSObject, if any, of this ACS as reference for which this object
   * is an internal hard alias.
   * This class implements the common alias methods for its subclasses which
   * are ImmutableSource or ImmutableTarget.
   * @return the hard reference. May be null.
   */
  public final ACSObject getEorReferenceAsInternalHardAlias() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getEorReferenceAsInternalHardAlias();
  } // getEorReferenceAsInternalHardAlias()

  /**
   * Get the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard reference.
   * This class implements the common alias methods for its subclasses which
   * are ImmutableSource or ImmutableTarget.
   * @return the hard alias. May be null.
   */
  public final ACSObject getEorAliasAsInternalHardReference() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getEorAliasAsInternalHardReference();
  } // getEorAliasAsInternalHardReference()

  /**
  * Gets a sorted array of names of the internal or external soft alias.
  * This class implements the common alias methods for its subclasses which
  * are ImmutableSource or ImmutableTarget.
  * @return ImmutableName[] of the soft alias. May be null but not empty.
  */
  public final ImmutableName[] getL_SoftAlias() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getL_SoftAlias();
  } // getL_SoftAlias()

  /**
  * Gets a sorted array of names of the ACS of the soft references.
  * @return ImmutableName[] of the ACS. May be null but not empty.
  */
  public final ImmutableName[] getL_AcsOfSoftReferences() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getL_AcsOfSoftReferences();
  } // getL_SoftAlias()

  /**
   * This property is not displayed in the GUI.
   * This class implements the common alias methods for its subclasses which
   * are ImmutableSource or ImmutableTarget.
  * @return true if there is at least one external or internal soft alias.
  */
  public final boolean getHasSoftAlias() {
    if (aliasHandler_ == null ) {
        return false;
    } // if
    return aliasHandler_.getHasSoftAlias();
  } // getHasSoftAlias()

   /**
   * Get the open views containing this BaseObject.
   * @return List of ViewInBase. May be empty,
   *   but never null.
   */
  public List<ViewInBase> getImpliedViews() {
      return folderMember_.getImpliedViews();
  } // getImpliedViews()

  /**
   * A constrained (opened or closed) child ACS has one or several of
   * the following relations to this object: <P>
   * &lt;ForChildAcsOpeningAndClosing&gt;, &lt;ForChildAcsOpeningNoClosing&gt;
   * or &lt;ForChildAcsIntegrity&gt;.<P>
   * Each relation is associated to a key with the same name.
   * @return a map of 3 ACS name lists for the 3 keys. May be empty
   *    but never null. A value may be an empty array but is never null.
   */
  public Map<String, ImmutableName[]> getM_ConstrainedChildACS() {
      if (m_ConstrainedChildACS_ == null) return new HashMap(INITIAL_CAPACITY);
      else return (Map) m_ConstrainedChildACS_.clone();      
  }

  /**
  * Adds a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      synchronized (changeSupport_)
      {changeSupport_.addPropertyChangeListener(_l);}

      folderMember_.addPropertyChangeListener(_l);

      if (privHandler_ != null)
          privHandler_.addPropertyChangeListener(_l);
      else {
            try {
              privHandler_ = new CPrivilegeUserImpl(this);
            } catch (CreateError c) {
                new InternalError("From "+name_+"- "+c.toString());
            }

            privHandler_.addPropertyChangeListener(_l);
      } // else
      
      if (aliasHandler_ != null) aliasHandler_.addPropertyChangeListener(_l);
      else {
            try {
              aliasHandler_ = new CAliasImpl(this);
            } catch (CreateError c) {
                new InternalError("From "+name_+"- "+c.toString());
            }

            aliasHandler_.addPropertyChangeListener(_l);
      } // else

                // for testing and tracking
                //System.out.println("BasicImpl.addPropertyChangeListener(), call from "+
                  //  _l+" to "+name_+" for all the properties ");

  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'Comment', 'ImpliedOpenViews', 'VirtualFolders', 'ResourceType' or 'EPType',
  * 'M_ConstrainedChildACS'.
  * The following properties are managed by aliasHandler_:
  * 'HardAliasReference', 'HardAlias', 'SoftAlias', 'AcsOfSoftReferences'
  * and 'hasSoftAlias'.
  * Called for all the subclasses, then there is no check up of the property name
  * if it is not one of the previously named properties.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      // privHandler_ is managed by the subclasses
      if (_propertyName.equals("ImpliedOpenViews") || _propertyName.equals("VirtualFolders"))
          folderMember_.addPropertyChangeListener(_propertyName, _l);

      else if (_propertyName.equals("HardAliasReference") ||
                        _propertyName.equals("HardAlias") ||
                        _propertyName.equals("SoftAlias") ||
                        _propertyName.equals("AcsOfSoftReferences") ||
                        _propertyName.equals("hasSoftAlias")) {
            if (aliasHandler_ == null ) {
                try {
                  aliasHandler_ = new CAliasImpl(this);
                } catch (CreateError c) {
                    new InternalError("From "+name_+"- "+c.toString());
                }
            } // if

            aliasHandler_.addPropertyChangeListener(_propertyName, _l);
      }

      else synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("BasicImpl.addPropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the adding of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //System.out.println("listeners of all properties after the adding of "+_l+": ");
                    //l_list = changeSupport_.getPropertyChangeListeners();
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //l_list = null;
                //} // if


  } // addPropertyChangeListener(String _propertyName,

  /**
  * Removes a listener to the bean, for all the bound properties.
  * Used by ViewInBaseImpl.
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(PropertyChangeListener _l){

      // no action if the transiant variable is null
      // It is null when a PropertyEditor calls back a finalized instance.
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_l);}

      folderMember_.removePropertyChangeListener(_l);

      if (privHandler_ != null)
          privHandler_.removePropertyChangeListener(_l);

      if (aliasHandler_ != null)
          aliasHandler_.removePropertyChangeListener(_l);

  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'Comment', 'ImpliedOpenViews', 'VirtualFolders', 'ResourceType' or 'EPType',
  * 'M_ConstrainedChildACS'.
  * The following properties are managed by aliasHandler_:
  * 'HardAliasReference', 'HardAlias', 'SoftAlias', 'AcsOfSoftReferences'
  * and 'hasSoftAlias'.
  * Called for all the subclasses, then there is no check up of the property name
  * if it is not one of the previously named properties.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("BasicImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      // privHandler_ is managed by the subclasses
      if (_propertyName.equals("ImpliedOpenViews") || _propertyName.equals("VirtualFolders"))
          folderMember_.removePropertyChangeListener(_propertyName, _l);

      else if (_propertyName.equals("HardAliasReference") ||
                        _propertyName.equals("HardAlias") ||
                        _propertyName.equals("SoftAlias") ||
                        _propertyName.equals("AcsOfSoftReferences") ||
                        _propertyName.equals("hasSoftAlias")) {
            if (aliasHandler_ != null )
                aliasHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else synchronized (changeSupport_){
          changeSupport_.removePropertyChangeListener(_propertyName, _l);
      }

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("BasicImpl.getPropertyChangeListeners(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      // privHandler_ is managed by the subclasses
      if (_propertyName.equals("ImpliedOpenViews") || _propertyName.equals("VirtualFolders"))
          return folderMember_.getPropertyChangeListeners(_propertyName);

      else if (_propertyName.equals("HardAliasReference") ||
                        _propertyName.equals("HardAlias") ||
                        _propertyName.equals("SoftAlias") ||
                        _propertyName.equals("AcsOfSoftReferences") ||
                        _propertyName.equals("hasSoftAlias")) {
            if (aliasHandler_ != null )
                return aliasHandler_.getPropertyChangeListeners(_propertyName);
            else
                return null;
      }
      else synchronized (changeSupport_){
          return changeSupport_.getPropertyChangeListeners(_propertyName);
      }

  } // getPropertyChangeListeners()

 /**
  * Returns a copy without registering the new object in the ACSs,
  * and without links to the privileges of this instance.
  * Caution: since the clone() interface is broken, copy constructors should be
  * better.
   * Caution: don't use any updating method on this clone,
   *  because that may destroy the base integrity.
   * @exception CloneNotSupportedException
   * @return Object - The names are cloned.
   * The external object references are not cloned (see "Eor" in get method).
   * The links between the EOR and the new instance are not set, but the lists
   * of OER are cloned (viewInBase and privileges) so it is possible to
   * get these active objects.
   */
  public synchronized Object clone() {
    try {
      BasicImpl clon = (BasicImpl) super.clone();
      clon.name_ = (ImmutableName) name_.clone();
      clon.aCS_Name_ = (ImmutableName) aCS_Name_.clone();
      clon.iS_Name_ = (ImmutableName) iS_Name_.clone();
  //    clon.setL_Access(l_Access_);
      clon.setComment(comment_);
      try {
          clon.setType(type_);
      }
      catch (UpDateError u) {
          throw new InternalError(u.toString());
      }
      //    clon.setIcon(icon_);
      clon.aCS_ = aCS_;

      // these embeded classes are cloned with the result as master
      if (privHandler_ != null) clon.privHandler_ = privHandler_.cloneC(clon);
      if (aliasHandler_ != null) clon.aliasHandler_ = aliasHandler_.cloneC(clon);
      clon.folderMember_ = folderMember_.cloneC(clon);

      return clon;

    } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
      if (aCS_ == null) throw new InternalError("In BasicImpl, the main ACS is null");
      else {
        aCS_.getEorBaseError().addException(e.toString() + "BasicImpl.clone() fails");
        throw new InternalError(e.toString());
      }
    } // catch
  } // clone()

  /**
  * Does not compare the comment. Works even if the instance is user-finalized.
  * @param _obj
  * @return boolean true if _obj is an instance of BasicImpl (or a derived class)
  *   equals to this instance (with the same class)
  */
  public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      BasicImpl obj = (BasicImpl) _obj;
      
      /*
      if (comment_ == null | obj.getComment() == null) {
          if (comment_ != null | obj.getComment() != null)
            return false;
      } // if
      else {
          if (!comment_.equals( obj.getComment() ))
            return false;
      } // else
      */

//      if (!icon_.equals( obj.getIcon() )
//          return false;
      if(aCS_ != null) {
        if (aCS_ != obj.getEorACS() ) return false;
      } else {
        if(obj.getEorACS() != null) return false; }

      if (name_ == null | obj.getName() == null) {
          if (name_ != null | obj.getName() != null)
            return false;
      } // if
      else {
        if (!name_.equals( obj.getName() ))
          return false;
      } //else

      if (iS_Name_ == null || obj.getIS_Name() == null) {
          if (iS_Name_ != null || obj.getIS_Name() != null)
            return false;
      } // if
      else {
          if (!iS_Name_.equals( obj.getIS_Name() ))
            return false;
      } // else

      if (privHandler_ == null || ((BasicImpl)obj).privHandler_ == null) {
          if (privHandler_ != null || ((BasicImpl)obj).privHandler_ != null)
            return false;
      } // if
      else if (!privHandler_.equals( ((BasicImpl)obj).privHandler_ ))
        return false;

      if (aliasHandler_ == null || ((BasicImpl)obj).aliasHandler_ == null) {
          if (aliasHandler_ != null || ((BasicImpl)obj).aliasHandler_ != null)
            return false;
      } // if
      else if (!aliasHandler_.equals( ((BasicImpl)obj).aliasHandler_ ))
        return false;

      if (folderMember_ == null || ((BasicImpl)obj).folderMember_ == null) {
          if (folderMember_ != null || ((BasicImpl)obj).folderMember_ != null)
            return false;
      } // if
      else if (!folderMember_.equals( ((BasicImpl)obj).folderMember_ ))
        return false;

      if (type_ == null | obj.getType() == null) {
          if (type_ != null | obj.getType() != null)
            return false;
      } // if
      else {
          if (!type_.equals( obj.getType() ))
            return false;
      } // else

      
      return true;
  } // equals

  /**
   * @return the hashcode derived from the type and folder meber hashcodes
   */
  public int hashCode() {
    int result = 0;

    if (type_ != null) {
      result = type_.hashCode();
    }
    if (folderMember_ != null) {
      result = PRIME*result + folderMember_.hashCode();
    }

    return result;
  } // hashcode()

  /**
   * For the java serialization.
   * @return long
   */
  public long getSerialVersionUID()	{
          return serialVersionUID;
  }

  /**
   * The comparison is based on the name compareTo() method of each
   * leaf, except when there is a node.
   * This method is based on name_, so Basic.compareTo(same instance)
   * does not return 0, but a positive or a negative value.
   * The nodes are always smaller than the leaves.
   * @param _o to compare must be an ImmutableLeaf
   * @exception ClassCastException - if the specified object's type
   *  prevents it from being compared to this current Object
   * @return int 0 if the leaves are equals, a negative integer if this leaf
   * is lesser than the argument, and a positive integer if this leaf
   * is greater than the argument.
   */
  public int compareTo(Object _o) throws ClassCastException {
    if (!(_o instanceof ImmutableBasic))
      throw new ClassCastException("in Basic.compareTo(), the argument "+
        "is not an instance of ImmutableBasic");
    // if (equals(o) ) return 0;
    if (name_ == null || _o == null || ((ImmutableBasic)_o).getName() == null)
        throw new InternalError("in Basic.compareTo(),"+
           " the object name, the argument or the argument name is null");
    if (this instanceof ImmutableNode && !(_o instanceof ImmutableNode)) return -1;
    else if (!(this instanceof ImmutableNode) && (_o instanceof ImmutableNode)) return +1;
    else return name_.compareTo( ((ImmutableBasic)_o).getName() );
  }

   /**
   * This method controls the immutability of the object, removes the associates,
   * then calls finalizeForBase().
   * An unused object has to be deleted to inform the other objects, and this method
   * HAS TO BE CALLED before deleting this object.
   * Called by the GUI. Calls finalizeForBase() and
   * ACSFactoryUtilityImpl.removeAssociatesAfterDeletion() to remove the associates.
   * This method does not raise PropertyChangeEvents since the BeanInfo for this class
   * specifies that its ACS informs the listeners when this instance is finalized.
   * To test if an instance has been finalized, uses (folderMember_ == null).
   * @exception UpDateError if the object type contains 'byAcsCreator', starts with
   * or ends with 'immutable', or from finalizeFromBase() or
   * AssociateUtilityImpl.removeAssociatesAfterDeletion()
   */
  public synchronized void finalizeForUser() throws UpDateError {
        if (folderMember_ == null) return;

        if (type_.indexOf("<byAcsCreator>") != -1 || type_.startsWith("<immutable> ") ||
                type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of '"+getName().lastComponent()+"' sets it as without deleting.");

        // removes the associates, if any
        try {
            AssociateUtilityImpl.removeAssociatesAfterDeletion(this, (ACS)aCS_);
        } catch (BaseError b) {
            // finalize the instance in all cases
            finalizeForBase();
            throw new UpDateError("from AssociateUtilityImpl.removeAssociatesAfterDeletion(), "+
                    b.toString());
        }

                    // for testing and tracking
                    //System.out.println("call of BasicImpl.finalizeForUser(), for " +
                      //      name_);

        finalizeForBase();
  } // finalizeForUser()

   /**
   * This method cuts the linked privileges links, if any, and
   * removes the instance in the virtual folders, if any.
   * It is public to let the gWork methods to call it, but is is not described
   * in the BeanInfo since the beamer must use only finalizeForUser() to delete an object.
   * Called by finalizeForUser(), which controls the immutability of the object
   * and removes the associates before the call.
   * This method is called by finalize() and the overridding methods.
   * Caution: the overridding methods in ResourceImpl and EligiblePartyImpl
   * unregister the instance in its ACS.
   * Calls setType() with a null argument, to unregister the instance in the
   * associated privileges. An unused object has to be deleted to inform the others objects,
   * and this method HAS TO BE CALLED before deleting the Basic.
   * This method does not raise PropertyChangeEvents since the BeanInfo for this class
   * specifies that its ACS informs the listeners when this instance is finalized.
   * To test if an instance has been finalized, uses (folderMember_ == null).
   * @exception UpDateError if the CPrivilegeUserImpl or CFolderMemberImpl throws it.
   */
  public synchronized void finalizeForBase() throws UpDateError {

                        //System.out.println("BasicImpl.finalizeForBase() for "+name_ );

    if (folderMember_ == null) return;

    setType(null);
    
    if (privHandler_ != null) privHandler_.finalizeForBase();
    if (aliasHandler_ != null) aliasHandler_.finalizeForBase();
    folderMember_.finalizeForBase();

    aCS_ = null;
    //name_ = null; //not cleared; used in maps
    iS_Name_= null;
    //aCS_Name_= null; // not cleared; used in ACSTreeBaseListener, ViewInBaseImpl...
    comment_ = null;
    privHandler_ = null;
    aliasHandler_ = null;
    folderMember_ = null;
    m_ConstrainedChildACS_ = null;
    changeSupport_ = null;

    try {
        super.finalize();
    } catch (Throwable t) {
        throw new InternalError("in BasicImpl.finalizeForBase(), "+
          "super.finalize throws: "+t.toString());
    }
  } // finalizeForBase()

  /**
   * Gets the name last component of the instance.
   * Used in the explorer to display the element.
   * Never null.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append(name_.lastComponent() );
/*    sb.append("BasicIMPL :");
    sb.append("; name_ = ");
    sb.append(name_);

// for actor class tests and others inherited classes, add here : /*
    sb.append("; aCS_ = ");
    if (aCS_ != null) sb.append(aCS_.getName());
    else sb.append("null");
/*    sb.append("; actorActivated_ = ");
    sb.append(actorActivated_);
    sb.append("; isACSBridge_ = ");
    sb.append(isACSBridge_);

    sb.append("; parent_ = ");
    if (parent_ != null && parent_.getName() != null)
      sb.append(parent_.getName() );
    if (bridgeTargetEligibleParty_ != null && bridgeTargetEligibleParty_.getName() != null) {
      sb.append("; bridgeTargetEligibleParty_ = ");
      sb.append(bridgeTargetEligibleParty_.getName() );
    }
    sb.append("l_Access_ = ");
    for (int i = 0; i < l_Access_.length; i++) {
        sb.append(l_Access_[i]);
        sb.append(", ");
    } //for

    sb.append("; comment_ = ");
    sb.append(comment_);
    sb.append("; l_aclEntry_ =");
    for (int i = 0; i < l_aclEntry_.length; i++) {
        sb.append(l_aclEntry_[i]);
        sb.append(", ");
    } //for

    return " [[[" + sb + "]]]";
//*/
    return " " + sb;
  } // toString()

  //
  // ---------------- protected methods ------------------------------
  //

  /**
   * Set the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard alias.
   * The generic gBase package does not call this method, nor the user from the GUI.
   * It is the responsibility of the ACS associations or the AcsAddon code
   * to set an internal alias for one new ACSObject, as an alias or as an alias reference.
   * If the instance is already a reference, it cannot become an alias.
   * Fires the property change 'HardAliasReference'.
   * @param _ref the reference. May be null.
   * @exception UpDateError _ref does not belong to the same ACS or the type is
   * immutable
   */
  protected void setEorReferenceAsInternalHardAlias(ACSObject _ref) throws UpDateError {
    if (folderMember_ == null) return;
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new UpDateError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.setEorReferenceAsInternalHardAlias(_ref);
  } // setEorReferenceAsInternalHardAlias()

  /**
   * Set the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard reference.
   * The generic gBase package does not call this method, nor the user from the GUI.
   * It is the responsibility of the ACS associations or the AcsAddon code
   * to set an internal alias for one new ACSObject, as an alias or as an alias reference.
   * If the instance is already an alias, it cannot become a reference.
   * Fires the property change 'HardAlias'.
   * @param _ali the alias. May be null.
   * @exception UpDateError _basic does not belong to the same ACS or
   * the master is already an alias
   */
  protected void setEorAliasAsInternalHardReference(ACSObject _ali) throws UpDateError {
    if (folderMember_ == null) return;
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new UpDateError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its alias.");

    aliasHandler_.setEorAliasAsInternalHardReference(_ali);      
  } // setEorAliasAsInternalHardReference()

  /**
   * Adds an internal or external soft alias.
   * The alias is not updated. Used only by the ACS.
   * Has to be used by the AcsAddon from a gBase class.
   * Fires the events 'SoftAlias' and 'hasSoftAlias'.
   * @param _alias ACSObject. No action if null, or already known.
   * @exception UpDateError the type is immutable
   */
  protected void addEorSoftAlias(ImmutableName _alias) throws UpDateError {
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.addEorSoftAlias(_alias);
  } // addEorSoftAlias()

  /**
   * Removes an external or internal alias.
   * The external alias is not updated. Used only by the ACS, in an AcsAddon
   * and from a gBase class.
   * Detects the alias through a name comparison.
   * Fires the 'SoftAlias' and the 'hasSoftAlias' property change events.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the list is null.
   * @param _alias is a registered alias. No action if null.
   * @exception UpDateError if _alias is not known.
   */
  protected void removeEorSoftAlias(ImmutableName _alias) throws UpDateError {
    if (folderMember_ == null) return;
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.removeEorSoftAlias(_alias);
  } // removeEorSoftAlias()

  /**
   * Adds the ACS name of an external or internal soft reference to the master as soft alias.
   * The reference is not updated there but by the ACS calling method.
   * Used only by the ACS, which may add controls on the parameter class.
   * Has to be used by the AcsAddon.
   * Fires the event 'AcsOfSoftReferences'.
   * @param _acs ACS name. No action if null, or already known.
   * @exception UpDateError if the type is immutable
   */
  protected void addAcsOfSoftReferences(ImmutableName _acs)  throws UpDateError {
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.addEorSoftAlias(_acs);      
  } // addAcsOfSoftReferences()

  /**
   * Removes an external or internal alias from the master as soft reference.
   * The alias is not updated there but by the ACS calling method.
   * Used only by the ACS, which may add controls on the parameter class.
   * Detects the alias through a name comparison.
   * Fires 'AcsOfSoftReferences' property change events.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the list is null.
   * @param _acs ACS name. No action if null.
   * @exception UpDateError if _acs is not known.
   */
  protected void removeAcsOfSoftReferences(ImmutableName _acs) throws UpDateError {
    if (folderMember_ == null) return;
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.removeAcsOfSoftReferences(_acs);
  } // removeAcsOfSoftReferences()

   /**
   * Registers the view as containing this BaseObject.
   * Called by EPRViewInBaseImpl.addEorResource/EligibleParty().
   * Fires 'ImpliedOpenViews' event.
   * @param _view to add
   */
  protected void addInViewList(ViewInBase _view) {
      folderMember_.addInViewList(_view);
  } // addInViewList()

   /**
   * Unregisters the view when it does not contain this BaseObject.
   * Called by EPRViewInBaseImpl.removeEorResource/EligibleParty().
   * Fires 'ImpliedOpenViews' event.
   * @param _view to remove
   * @return false if the view is unknown
   */
  protected boolean removeInViewList(ViewInBase _view) {
      return folderMember_.removeInViewList(_view);
  } // removeInViewList()

   /**
   * A constrained (opened or closed) child ACS has one or several of
   * the following relations to this object: <P>
   * &lt;ForChildAcsOpeningAndClosing&gt;, &lt;ForChildAcsOpeningNoClosing&gt;
   * or &lt;ForChildAcsIntegrity&gt;.<P>
   * Each relation is associated to a key with the same name, given by
   * ACSFactoryUtilityImpl.getKeysForConstrainedChildACSMap().
   * Fires a 'M_ConstrainedChildACS' event.
   * Called by the child ACS setAcsRelationsInIS() method.
   * @param _key must be among the allowed keys. Otherwise, throws an InternalError.
   * @param _l_acs array of constrained and opened child ACSs to add.
   *           Noop if null or empty.
   */
  protected void addConstrainedChildACS(String _key, ImmutableName[] _l_acs) {
      if (m_ConstrainedChildACS_ == null)
          m_ConstrainedChildACS_ = new HashMap(INITIAL_CAPACITY);

                    // for testing and tracking
                    //System.out.println("From BasicImpl.addConstrainedChildACS(), "+
                      //      "values to add for "+_key);

      if (_key == null ||
              !Arrays.asList(AssociateUtilityImpl.
                getKeysForConstrainedChildACSMap()).contains(_key))
                      throw new InternalError("BasicImpl.addConstrainedChildACS() : "+
                      "false key as argument.");

      if (_l_acs == null || _l_acs.length == 0) return;

      Map               m_old = (Map)m_ConstrainedChildACS_.clone();
      ImmutableName[]    array = null;
      ImmutableName[]    l_result = null;
      Set<ImmutableName>  l_new = null;

      synchronized (m_ConstrainedChildACS_) {
            array = (ImmutableName[])m_ConstrainedChildACS_.get(_key);
            if (array == null) {
                l_result = new ImmutableName[_l_acs.length];
                System.arraycopy(_l_acs, 0, l_result, 0, _l_acs.length);
                m_ConstrainedChildACS_.put(_key, l_result);
            }
            else if (Arrays.asList(array).containsAll(Arrays.asList(_l_acs))) {
                m_old = null;
                array = null;
                return;
            }
            else {
                l_new = new HashSet(INITIAL_CAPACITY);
                for (int i = 0; i < array.length; i++) {
                    l_new.add(array[i]);
                } // for i
                l_new.addAll(Arrays.asList(_l_acs) );
                m_ConstrainedChildACS_.put(
                        _key, (ImmutableName[])l_new.toArray(new ImmutableACS[0]));
            }
      } // synchro

      firePropertyChange("M_ConstrainedChildACS", m_old,
              (Map)m_ConstrainedChildACS_.clone());

      array = null;
      m_old = null;
      l_result = null;
      l_new = null;
  } // addConstrainedChildACS()

   /**
   * A constrained (opened or closed) child ACS has one or several of
   * the following relations to this object: <P>
   * &lt;ForChildAcsOpeningAndClosing&gt;, &lt;ForChildAcsOpeningNoClosing&gt;
   * or &lt;ForChildAcsIntegrity&gt;.<P>
   * Each relation is associated to a key with the same name.
   * Fires a 'M_ConstrainedChildACS' event.
   * Calls ACSFactoryUtilityImpl.getKeysForConstrainedChildACSMap().
   * Called by the child ACS.
   * @param _key must be among the allowed keys.
   * @param _l_acs constrained and opened child ACSs to remove. To set to null
    *   for deleting _key in the map. May be empty but without null element.
   * @exception UpDateError if:<P>
   *    - the key is null or unknown, <P>
   *    - one value in _l_acs is null, <P>
   *    - a value is not registered.
   */
  protected void removeConstrainedChildACS(String _key, ImmutableName[] _l_acs)
    throws UpDateError {

      if (_key == null ||
                !Arrays.asList(AssociateUtilityImpl.
                    getKeysForConstrainedChildACSMap()).contains(_key))
                          throw new UpDateError("BasicImpl.removeConstrainedChildACS() : "+
                          "wrong key as argument.");

      if (m_ConstrainedChildACS_ == null) {
        throw new UpDateError("The property is not well initialized.");
      }
      if (m_ConstrainedChildACS_.get(_key) == null) {
        throw new UpDateError("The property does not know this key.");
      }

                    // for testing and tracking
                    //System.out.println("From BasicImpl.removeConstrainedChildACS(), "+
                      //      "values to remove for "+_key);

    HashMap m_old = (HashMap) m_ConstrainedChildACS_.clone();

    if ( _l_acs == null) {
        m_ConstrainedChildACS_.remove(_key);

        try {
              firePropertyChange("M_ConstrainedChildACS", m_old,
                      (Map)m_ConstrainedChildACS_.clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From BasicImpl.removeConstrainedChildACS(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        } // catch
        m_old = null;
        return;
    } // if ( _l_values == null)

    boolean error = false;

    for (int i = 0; i < _l_acs.length; i++) {
      if (_l_acs[i] == null)  {
          error = true;
          break;
      }
      else {
                    // for testing and tracking
                    //System.out.println("----------"+_l_values[i]);
      }
    } // for i
    if (error) {
        throw new UpDateError("A value in the argument is null.");
    }

    ImmutableName[] l_current = m_ConstrainedChildACS_.get(_key);
    List list = new ArrayList( Arrays.asList(l_current) );

    short count = 0;
    boolean found = false;

    for (int  i = 0 ; i < _l_acs.length ; i++) {
        found = list.remove(_l_acs[i]);

        if (!found) {
            error = true;
            break;
        } // if
        else count++;

    } // for i

    if (error) {
        m_old = null;
        l_current = null;
        list = null;
        throw new UpDateError("A value to remove is not found.");
    }

    if (l_current.length != count) {
        m_ConstrainedChildACS_.put(
                _key,
                (ImmutableName[]) list.toArray(new ImmutableName[list.size()]));
    } else { // if _l_acs contains all the current values
        m_ConstrainedChildACS_.put(
                _key, new ImmutableName[0]);
    }

    try {
          firePropertyChange("M_ConstrainedChildACS", m_old,
                  (Map)m_ConstrainedChildACS_.clone());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BasicImpl.removeConstrainedChildACS(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    } // catch

    m_old = null;
    l_current = null;
    list = null;
  } // removeConstrainedChildACS()

   /**
   * Registers the folder as containing this instance.
   * Called by VirtualFolderImpl.addEorBaseObject().
   * Fires 'VirtualFolders' event.
   * @param _folder to add
   * @exception CreateError if the ACS does not manage the virtual folders.
   */  
  protected void addInVFolderList(VirtualFolder _folder) throws CreateError {
      folderMember_.addInVFolderList(_folder);
  }

   /**
   * Gets the virtual folders containing the master.
   * @return List of VirtualFolders. May be empty,
   *   but never null.
   */
  public List<VirtualFolder> getVirtualFolders() {
      return folderMember_.getVirtualFolders();
  }

   /**
   * Unregisters the folder when it does not contain this instance.
   * Called by VirtualFolderImpl.removeEorBaseObject().
   * Fires 'VirtualFolders' event.
   * @exception CreateError if the ACS does not manage the virtual folders.
   * @return boolean false if the folder is unknown
   */
  protected boolean removeInVFolderList(VirtualFolder _folder) throws CreateError {
      return folderMember_.removeInVFolderList(_folder);
  }
  
  /**
   * If the argument is allowed by the ACS, sets the type of this instance.
   * If the object is a Resource, the value is equal to the resource type.
   * If the object is an EligibleParty, the value is equal to the EP type.
   * This method is used by the subclasses constructors, and to switch from/to
   * an '&lt;immutable&gt;' type. <P>
   * An '&lt;immutable&gt;' type forbids most of the updatings of the instance.
   * Indeed, it is possible to switch for example from 'actor' to
   * '&lt;immutable&gt; actor', but not to 'immutable group'.
   * The rule is that the 'immutable ' key word, if it is used, must be at the start
   * or at the end of the type. <P>
   * From the default value 'undefined', any updating may be done.
   * The user cannot change the type if it starts with '&lt;byAcsCreator&gt;',
   * but the program can do it. For doing it, '12345678901234567890' is
   * a program-reserved value. Even if it is not allowed by the ACS,
   * '12345678901234567890' is the only new value
   * it is not possible to change from a type starting with '&lt;byAcsCreator&gt;'.
   * If the current type does not start with '&lt;byAcsCreator&gt;' but contains it,
   * the type is editable while the instance cannot be finalized by the user.
   * From '12345678901234567890', any new value may be set from the allowed ones,
   * and this starting value is the only one which allows to set a value containing
   * '&lt;byAcsCreator&gt;'.  <P>
   * Since the typed privileges use depends on the type, this method registers
   * the typed privileges having the same type (modulo the key word '&lt;immutable&gt;').
   * However, this is not true for the for_seed typed privileges,
   * nor for the specific ones, because a specific privilege cannot be created
   * if the source or the target is not there.
   * A switching from/to an '&lt;immutable&gt;' type to the other format never
   * changes the associated PrivilegeForTypes. <P>
   * Every change in this method should be done in VirtualFolderImpl.setType().
   * Fires a property change on 'ResourceType' or 'EPType'.
   * This method is indirectly called from the GUI. Called by finalizeForBase(),
   * ResourceImpl.setResourceType() and EligiblePartyImpl.setEPType(),
   * at the request of their subclass constructor, or for the GUI.
   * Calls ACSImpl.getM_TypesAndGUIPolicy() and getM_TypeOrientedPolicy(),
   * BaseUtilityImpl.getObjectTypeWithoutImmutable().
   * @param _type as for instance "Cayro_file" or "tablespace".
   *  May be null, or equals to the current type.
   * @exception UpDateError if the new value and the old one are not compatible, or
   * if the type handling for privileges throws it.
   */
  protected void setType(String _type) throws UpDateError {
      if (type_ != null && type_.equals(_type))
          return;
      else if (type_ == null && _type == null)
          return;

      // variables
      String[]            l_allowed = null;
      String              oldType = type_;
      List                l_priv;
      PrivilegeForType    priv;
      String[]            l_policy = null;
      String              reducedOldType = null;
      String              reducedType = null;
      
      if (this instanceof Resource)
          l_allowed = (String[])aCS_.
                    getM_TypesAndGUIPolicy().get("Resource.ResourceType");
      else if (this instanceof EligibleParty)
          l_allowed = (String[])aCS_.
                    getM_TypesAndGUIPolicy().get("EligibleParty.EPType");

                          // for testing purpose
                          //System.out.println("for "+getName()+", BasicImpl.setType() "+
                            //      "for the type argument "+_type);
                          //System.out.println("   allowed values: ");
                          //for (int i = 0; i < l_allowed.length; i++) {
                            //  System.out.println(" ------- "+l_allowed[i]);
                          //} // for i


      if (_type != null && !_type.equals("<12345678901234567890>") &&
              l_allowed != null && !Arrays.asList(l_allowed).contains(_type) ) {

                          // for testing purpose
                          //System.out.println("Forbidden type argument: "+_type);
                          //System.out.println("   allowed values: ");
                          //for (int i = 0; i < l_allowed.length; i++) {
                            //  System.out.println(" ------- "+l_allowed[i]);
                          //} // for i

          throw new UpDateError(" The new type '"+_type+"' of "+name_.lastComponent()+
                  " is forbidden by the ACS.");
      } // if
      
      if (type_.indexOf("<byAcsCreator>") != -1 &&
              (_type != null && !_type.equals("<12345678901234567890>")))
          throw new UpDateError(" "+name_.lastComponent()+" has a type which"+
                  " forbids a type editing.");
          
      else if (!type_.equals("<undefined>")) {

          if (_type != null && _type.startsWith("<byAcsCreator>") &&
                  !type_.equals("<12345678901234567890>"))
                throw new UpDateError(getName()+" new type cannot contain '<byAcsCreator>'.");
          if ( type_.indexOf("<byAcsCreator>") == -1 &&
                  !type_.equals("<12345678901234567890>")) {

              if (!type_.contains("<immutable>") && (_type != null &&
                      (!_type.contains("<immutable>") || !_type.contains(type_))))
                    throw new UpDateError(" The new type is not near to the previous type "+type_+
                            ", like 'actor' and '<immutable> actor' for instance.");
              if (type_.contains("<immutable>") && _type != null && !type_.contains(_type))
                    throw new UpDateError(" The "+getName().lastComponent()+" new type "+
                            _type+" is not derived from the previous one "+type_+
                            ", like '<immutable> actor' and 'actor' for instance.");

          } // if ( type_.indexOf("<byAcsCreator>") == -1
      } // if (!type_.equals("<undefined>")

    type_ = _type;


                // for testing and tracking
                //System.out.println("In BasicImpl, call of " + name_ + ".setType() with: ");
                //System.out.println("---- new type: " + type_ + ", old type: "+oldType+
                  //      " reduced old type: "+BaseUtilityImpl.getObjectTypeWithoutImmutable(oldType));


    
    if (aCS_.getManagePrivilegeRights() && aCS_.getManageSimpleTypedPrivileges()
            // && aCS_.getManageClassedTypedPrivileges()
            // && aCS_.getManageForSeedTypedPrivileges()
            ) {
        try {
            // from/to an 'immutable' type does not change the associated typed privileges
            if (oldType == null || type_ == null ||
                    !UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(type_).
                            equals(UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(oldType)) ) {

                // unlistenings from 'oldType'
                if (oldType != null) {
                    reducedOldType = UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(oldType);

                    if (aCS_.getM_TypeOrientedPolicy() != null)
                        l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                                "EligibleParty.SecondTypesOfPrivilegeFor."+reducedOldType);

                    if (this instanceof PrivilegeSource) {

                        // the reduced type is unuseful there
                        l_priv = aCS_.getEorL_PrivilegeForTypesForSource(oldType);

                        for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                            priv = (PrivilegeForType)ite.next();
                            if (l_policy != null &&
                                    !Arrays.asList(l_policy).contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(
                                                priv.getSecondType()) ))
                                                    continue; // avoid to throw BaseError

                            if (priv.getMainType().equals(PrivilegeAbst.TYPED))
                              ((PrivilegeSource)this).removeEorPrivilegeToSource(priv);
                            else if (priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED) &&
                              Class.forName("ARoad0.gBaseInterface."+
                                ((PrivilegeForType)priv).getSourceClass())
                                .isAssignableFrom(this.getClass()) )
                                        ((PrivilegeSource)this).removeEorPrivilegeToSource(priv);
                        } // for ite

                    } // if

                    if (aCS_.getM_TypeOrientedPolicy() != null)
                        l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                            "Resource.SecondTypesOfPrivilegeFor."+reducedOldType);

                    if (this instanceof PrivilegeTarget) {

                        // the reduced type is unuseful there
                        l_priv = aCS_.getEorL_PrivilegeForTypesForTarget(oldType);

                        for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                            priv = (PrivilegeForType)ite.next();
                            if (l_policy != null &&
                                    !Arrays.asList(l_policy).contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(
                                                priv.getSecondType()) ))
                                                    continue; // avoid to throw BaseError

                            if (priv.getMainType().equals(PrivilegeAbst.TYPED))
                              ((PrivilegeTarget)this).removeEorPrivilegeToTarget(priv);
                            else if (priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED) &&
                              Class.forName("ARoad0.gBaseInterface."+
                              ((PrivilegeForType)priv).getSourceClass())
                              .isAssignableFrom(this.getClass()) )
                                    ((PrivilegeTarget)this).removeEorPrivilegeToTarget(priv);
                        } // for ite

                    } // if
                } // if (oldType != null) {

                // new listenings from 'type_'
                if (type_ != null) {
                    reducedType = UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(type_);

                    if (aCS_.getM_TypeOrientedPolicy() != null)
                        l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                                "EligibleParty.SecondTypesOfPrivilegeFor."+reducedType);

                    if (this instanceof PrivilegeSource) {

                        // the reduced type is unuseful there
                        l_priv = aCS_.getEorL_PrivilegeForTypesForSource(type_);

                        for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                            priv = (PrivilegeForType)ite.next();
                            if (l_policy != null &&
                                    !Arrays.asList(l_policy).contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(
                                                priv.getSecondType()) ))
                                                    continue; // avoid to throw BaseError

                            // noop on TYPED_FOR_SEED and TYPED_CLASSED_SPECIFIC
                            if (priv.getMainType().equals(PrivilegeAbst.TYPED))
                              ((PrivilegeSource)this).addEorPrivilegeForTypeToSource(priv);
                            else if (priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED) &&
                              Class.forName("ARoad0.gBaseInterface."+
                                ((PrivilegeForType)priv).getSourceClass())
                                .isAssignableFrom(this.getClass()) )
                                        ((PrivilegeSource)this).addEorPrivilegeForTypeToSource(priv);
                        } // for ite

                    } // if

                    if (aCS_.getM_TypeOrientedPolicy() != null)
                        l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                            "Resource.SecondTypesOfPrivilegeFor."+reducedType);

                    if (this instanceof PrivilegeTarget) {

                        // the reduced type is unuseful there
                        l_priv = aCS_.getEorL_PrivilegeForTypesForTarget(type_);
                        
                        for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                            priv = (PrivilegeForType)ite.next();
                            if (l_policy != null &&
                                    !Arrays.asList(l_policy).contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(priv.getSecondType()) ))
                                                    continue; // avoid to throw BaseError

                            // noop on TYPED_FOR_SEED and TYPED_CLASSED_SPECIFIC
                            if (priv.getMainType().equals(PrivilegeAbst.TYPED))
                              ((PrivilegeTarget)this).addEorPrivilegeForTypeToTarget(priv);
                            else if (priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED) &&
                              Class.forName("ARoad0.gBaseInterface."+
                              ((PrivilegeForType)priv).getSourceClass()).
                                isAssignableFrom(this.getClass()) )
                                    ((PrivilegeTarget)this).addEorPrivilegeForTypeToTarget(priv);
                        } // for ite

                    } // if
                } // if (type_ != null) {
            } // if ((oldType == null

        } catch (BaseError b) {
            type_ = null;
            // caller of this method, but no loop since (type_ == null && _type == null)
            finalizeForBase();
            throw new InternalError("From BasicImpl.setType(), "+b.toString());
        } // catch
        catch (ClassNotFoundException c) {
            type_ = null;
            finalizeForBase();
            throw new InternalError("From BasicImpl.setType(), "+c.toString());
        } // catch
    } // if (aCS_.getManageSimpleTypedPrivileges()) {
    
    try {
          if (this instanceof Resource)
              firePropertyChange("ResourceType", oldType, type_);
          else if (this instanceof EligibleParty)
              firePropertyChange("EPType", oldType, type_);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BasicImpl.setType(), firePropertChange() throws:");
        e.printStackTrace();
    }

    l_allowed = null;
    oldType = null;
    l_priv = null;
    priv = null;
    l_policy = null;
    reducedOldType = null;
    reducedType = null;
  }  // setType()
  
  /**
   * UNUSED
  * This method provides a copy of the instance variables.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy).
  * Copy the privileges only if it is possible.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @return the updated _result
  * @exception CreateError from CFolderMemberImpl and CPrivilegeUserImpl
  **/
  protected BasicImpl copyOwnVariables(BasicImpl _result) throws CreateError {

      if (_result == null) {
          throw new CreateError("in BasicImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      //
      // TODO: control the values from the ACS
      //
      _result.iS_Name_ = getIS_Name();
      
      if (privHandler_ != null) 
          _result.privHandler_ = privHandler_.copy(_result);
      if (aliasHandler_ != null) 
          _result.aliasHandler_ = aliasHandler_.copy(_result);
      if (folderMember_ != null)
          _result.folderMember_ = folderMember_.copy(_result);

      return _result;
  } // copyOwnVariables(,,,) 

  /**
  * Fires an event to every registered listener, in any order.
  * For the bound properties 'Comment'.
  * These following properties are fired by folderMember_, not by this method:
  * 'ImpliedOpenViews', 'VirtualFolders'.
  * The following properties are fired by aliasHandler_, not by this method:
  * 'HardAliasReference', 'HardAlias', 'SoftAlias' and 'hasSoftAlias'.
  * Called by all the updating methods from all the subclasses, then there is
  * no check up of the property name, including calls from AcsAddon classes
  * like from BasesSetMySQLImpl.resetInheritedAclEntryRightsOfChildren().
  * @param _propertyName of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  public void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

                // for testing and tracking
                //System.out.println("in BasicImpl.firePropertyChange(), call to "+
                //  "propertyChange() of "+name_+" for the property "+_propertyName);

      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //System.out.println("BasicImpl.firePropertyChange(), call to "+
                  //"propertyChange() of "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
   * This method removes the Basic from the ACS lists, but there is no
   * guarantee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {
        u.printStackTrace();
    }
  } // finalize()

  /**
   * NO USE.
  * This method finalizes all the attributes, but not the ACS.
  * There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Called by EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In BasicImpl, call of " +
                //name_ + ".finalizeFromDeserialization()");
    if (folderMember_ == null) return;

    aCS_ = null;
    name_ = null;
    iS_Name_= null;
    aCS_Name_= null;
    comment_ = null;
    if (privHandler_ != null) privHandler_.finalizeFromDeserialization();
    if (aliasHandler_ != null) aliasHandler_.finalizeFromDeserialization();
    if (folderMember_ != null) folderMember_.finalizeFromDeserialization();
    privHandler_ = null;
    aliasHandler_ = null;
    folderMember_ = null;
    type_ = null;

    if (m_ConstrainedChildACS_ != null)
        m_ConstrainedChildACS_.clear();
    m_ConstrainedChildACS_ = null;

    try {
        super.finalize();
    } catch (Throwable t) {
        throw new InternalError("in BasicImpl.finalizeFromDeserialization(), "+
          "super.finalize throws: "+t.toString());
    }
  } // finalizeFromDeserialization()
  
} // BasicImpl

