// Title:        Access Road
// Version:      0.7.0
// Copyright:    2011 EURL ACCBEE
// Author:       Patrick Thazard
// Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road CAlias).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import java.util.Arrays;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * This class is an embedded class responsible for modeling an ACSObject
 * as reference to an alias, and/or as alias of a reference. The alias is the source
 * of an ARoad0.CNot.AccessControlLink.IS_ALIAS relation to its reference as target.
 * The hard alias setting is done through a call to the master of the CAliasImpl instance.
 * The hard reference of an alias is get through the methods of this class.
 * The soft reference of an alias is get through a call to getReferenceOfSoftAlias()
 * on the reference ACS. This is the correct way gWork.compoundRightsFactoryImpl
 * uses in the methods selectDirectHiddenReferenceForAlias() and
 * selectDirectReferencesInViewForAlias(), both for the hard and the soft aliasing.
 * The CAliasImpl master may be an alias or a reference.
 * It may be a BasicImpl or a VirtualFolderImpl. There is no constraints
 * on the reference class for a given alias class. Both for hard and soft relations,
 * the alias relation is not multi-level: a reference cannot be the alias of another object.
 * There are two types of alias relation this class may handle: <P>
 * (1) the internal hard alias produces a direct reference-to-alias relation
 * in the same ACS, and the two objects know this one-to-one relation, <P>
 * (2) the soft alias produces an indirect reference-to-alias relation
 * where the alias is internal or external, where the reference knows
 * all its soft aliases, and the alias simply knows the ACS of its unique soft reference. <P>
 * The soft alias results in a specific processing in each AcsAddon ACS,
 * since the reference is returned by the method ACS.getReferenceOfSoftAlias()
 * for a given external ACSObject as argument. This ACS method registers
 * the soft alias relation in the CAliasImpl instances of the reference and the alias. <P>
 * It is the responsibility of the ACS associations or the AcsAddon code
 * to set an internal alias for one new ACSObject, as an alias or as an alias reference.
 * On the other hand, the soft Alias relation may be only set by an AcsAddon. <P>
 * Generally speaking, in the standard way of use, the alias transfers
 * directly its rights as target to its reference, including, for an internal alias,
 * its Account/Groups context. If the elements of the AG context are not yet
 * in the path, this may define new links, from the reference to each account or group
 * of the context. This is the propagation of the indirect AG context.
 * The hard alias relation is alone to handle this tranfert in Access Road.
 * There is no such a propagation in an AclEntry or a Privilege relation,
 * even if an AclEntry transfers the direct AG context.
 * This case may occur when: <P>
 * (1) a GroupIDMember is a hard alias in a view,
 * and the hard reference receives the AG rights of its alias. <P>
 * (2) an Actor is a hard alias running as actor under a given AG context,
 * and the hard reference receives the AG rights of its alias. <P>
 * These cases are simple when the hard alias is in the view, since this implies
 * the reference is also in the view, and the paths from the alias are displayed
 * so that the user may see them directly.
 * Otherwise, the cases occur into an access path when the hard alias is an intermediate
 * node, and if the node before the hard alias do not transmit its proper AG context. <P>
 * An exemple of indirect AG context propagation is for a 3-nodes path having the form
 * 'ep/privilege/actor started from xid/alias/virtual folder',
 * where the actor runs under the account U and the group G. The generic rule
 * is that the rights of 'ep' on the virtual folder depend on the actor AG context,
 * if any, since the privileges do not propagate the Account/Groups context of 'ep'.
 * While U and G are not intermediate nodes in this path, if the actor is an intermediate node,
 * the path is extended by the links 'run under/U' and 'run under/G'.
 * If the actor does not run under U, there is no AG context to
 * propagate through the alias link. It is important to note that, while there is
 * an alias relation, the 'ep' rights on the virtual folder are always,
 * defined by the privileges. Simply, the AG context of the actor provides the rights
 * to the virtual folder. <P>
 * This exemple shows how it is possible to extend the properties of a given
 * class of the generic model, since VirtualFolder becomes there able to run
 * under a GroupIDMember. Such an extension of the model may be provided to
 * any ACSObject class, each time the instance is the reference of an alias.
 * This feature is independant to the VirtualFolder ACS property for managing
 * the rights propagation to the members. <P>
 * A second exemple of indirect AG context propagation is in the MySQL AcsAddon.
 * A MySQL view is a VirtualFolder having an Actor alias. The view provides access
 * to its members through the account or the group of the VirtualFolder AG context.
 * This may be the AG context of the VirtualFolder alias - if this alias is such
 * a self-AG-context Actor (MySQL DEFINER mode) - or it may be the direct AG context
 * from the nodes before the haard alias in the path (MySQL INVOKER mode). <P>
 * One non-standard way of use for this embedded class, in an AcsAddon, is to set
 * some of its properties without complying with the previous recommandations.
 * For exemple, the MySQL AcsAddon needs to model the PROXY relation between
 * several ProxyMySQL and one UserIDMySQL. The hard aliasing is not adapted, but
 * this AcsAddon uses the methods of the soft aliasing to set the PROXY link.
 * However, the MySQL ACS does not manage the soft aliasing, so there is no overload
 * in the access paths searches. Detecting a PROXY target through the method
 * ACS.getReferenceOfSoftAlias() would be more heavy when there are several open ACS.
 * This is why the MySQL AcsAddon chooses to complete the access paths searching,
 * in NodeRightsMySQLImpl, by a direct reading of a ProxyMySQL property returning
 * the current PROXY target. Such a non-standard way allows to display
 * the soft aliasing properties in the beamer, because the MySQL ACS manages
 * the hard aliasing for another need (that is the view delegate). <P>
 * All variables are private to secure the white box inheritances from this class.
 * It is a javabean with the following bound properties: <P>
 * - 'HardAliasReference', 'HardAlias', 'hasSoftAlias', 'SoftAlias', 'AcsOfSoftReferences'. <P>
 * The listeners are proper to this instance. Excepting for EPRViewInBaseImpl,
 * they are outside gBase, transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * This class is public to be overriden in the AcsAddons, and in this case,
 * the subclass is only called by some classes in the same package.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBaseInterface.ImmutableBasic#getEorReferenceAsInternalHardAlias
 * @see ARoad0.gBaseInterface.ImmutableACS#getReferenceOfSoftAlias(ARoad0.gBaseInterface.ACSObject) 
 * @see ARoad0.gWork.CompoundRightsFactoryImpl#selectDirectHiddenReferenceForAlias(ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, ARoad0.gBaseInterface.ACSObject) 
 */
public class CAliasImpl implements Serializable {
  /**
  * Protected variables
  */
  protected static final int  INITIAL_CAPACITY = 10; // the default value: 101 for map
  protected static final int  INITIAL_CAPACITY_2 = 4 ;  // the default value is 10
  protected static final int  INCREMENT_CAPACITY = 2;
  /* private variables
  */
  private static final long       serialVersionUID = 0;

  protected ACSObject             master_;
  /*
   * The soft aliases of the master, which by default is a 0-size array.
   */
  protected ImmutableName[]       l_SoftAlias_ = new ImmutableName[0];
  /*
   * The ACS of the soft references of the master. The master ACS has to add an ACS name
   * to this array when the method ACS.getReferenceOfSoftAlias() is called.
   * By default, it is a 0-size array.
   */
  protected ImmutableName[]       l_AcsOfSoftReferences_ = new ImmutableName[0];
  /*
   * The hard alias or the hard reference of the master. It is not possible
   * for the master to be both an alias and an alias reference.
   */
  protected transient ACSObject   hardReferenceOrAlias_ = null;
  protected ImmutableName         hardReferenceOrAliasName_ = null;
  /*
   * False if the master is a hard alias, while hardReferenceOrAlias_ is not null,
   * to detect is this object is the alias or the reference of the master.
   */
  protected boolean               isHardReference_ = false;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  static final CAliasImpl EMPTY_INSTANCE = new CAliasImpl();

  // -------- transient properties -------------------------------

   /**
   * manages all the property change listeners of this class and its subclasses
   */
  protected transient PropertyChangeSupport changeSupport_ = null;

  /** variable for the JDK 2 serialization
  * @serial
  * ("hardReferenceOrAliasName_", ImmutableName.class),
  * ("isHardReference_", boolean.class),
  * ("l_SoftAlias_", ImmutableName[].class),
  * ("l_AcsOfSoftReferences_", ImmutableName[].class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("hardReferenceOrAliasName_", ImmutableName.class),
      new ObjectStreamField("isHardReference_", boolean.class),
      new ObjectStreamField("l_SoftAlias_", ImmutableName[].class),
      new ObjectStreamField("l_AcsOfSoftReferences_", ImmutableName[].class)
  };

 /**
  * The basic constructor.
  */
  protected CAliasImpl() {


                          // for testing purpose
                     //System.out.println("Call to CAliasImpl.CAliasImpl()");
  }

  /**
  * Used by the master constructor. No checking of the management of hard and soft
  * alias by the master ACS.
  * @param _master is the master  of this embedded class
  * @exception CreateError the master is null
  */
  protected CAliasImpl(ACSObject _master) throws CreateError {
        if (_master == null )
             throw new CreateError("At the CAlias construction, "+
                "the master is null.");
        master_ = _master;
  } // constructor


  /**
  * This method provides a copy of the instance
  * with another name, without children copying
  * because the copy order will not be always correct.
  * @param _masterCopy is the master of the new copy
  * @return copy as CAliasImpl instance
  * @exception CreateError of the constructor with the same parameters.
  **/
  protected CAliasImpl copy(ACSObject _masterCopy) throws CreateError {

      CAliasImpl result = new CAliasImpl(_masterCopy);

      // TODO: copy the properties

      return result;
  } // copy()

  //
  // -------------- Methods of the implemented interfaces ---------------
  //

 /**
 * @return a value derived from the parent nick name hashcode
 */
 public int hashCode() {
   return "CAliasImpl".hashCode() + master_.getNickName().hashCode();
 } // hashCode()

  /**
  * As a soft reference, gets a sorted array of names of the soft alias.
  * @return ImmutableName[] of the alias. May be null but not empty.
  */
  protected final ImmutableName[] getL_SoftAlias() {

                // for testing and tracking
                //System.out.println("call to CAliasImpl.getEorL_SoftAlias()");

      // l_SoftAlias_ == null only if the CAlias has been finalized
    if (l_SoftAlias_ == null || l_SoftAlias_.length == 0)
        return null;

    ImmutableName[] result = new ImmutableName[l_SoftAlias_.length];
    System.arraycopy(l_SoftAlias_, 0, result, 0, l_SoftAlias_.length);

                // for testing and tracking
                //System.out.println("CAliasImpl.getEorL_SoftAlias() non-null result: ");
                //for (int i = 0; i < result.length; i++) {
                  //  System.out.println("----- "+result[i].lastComponent());
                //} // for i

    return result;
  } // getL_SoftAlias()

  /**
  * As a soft alias, sets a sorted array of names of the ACS of the soft references.
  * @return ImmutableName[] of the acs. May be null but not empty.
  */
  protected final ImmutableName[] getL_AcsOfSoftReferences() {
    // l_AcsOfSoftReferences_ == null only if the CAlias has been finalized
    if (l_AcsOfSoftReferences_ == null || l_AcsOfSoftReferences_.length == 0)
        return null;

    ImmutableName[] result = new ImmutableName[l_AcsOfSoftReferences_.length];
    System.arraycopy(l_AcsOfSoftReferences_, 0, result, 0, l_AcsOfSoftReferences_.length);

                // for testing and tracking
                //System.out.println("CAliasImpl.getEorL_AcsOfSoftReferences()are: ");
                //for (int i = 0; i < result.length; i++) {
                    //System.out.println("----- "+new Boolean(result[i] instanceof ImmutableResource));
                //} // for i

    return result;
  } // getL_AcsOfSoftReferences()

  /**
  * @return true if, as a soft reference, there is at least one soft alias.
  */
  protected final boolean getHasSoftAlias() {
    // l_SoftAlias_ == null only if the CAlias has been finalized
    if (l_SoftAlias_ == null || l_SoftAlias_.length == 0)
        return false;

    return true;
  } // getHasSoftAlias()

  /**
   * Get the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard alias. Calls UtilityImpl.getBaseObject().
   * @return the hard reference. May be null.
   */
  protected ACSObject getEorReferenceAsInternalHardAlias() {
    if (hardReferenceOrAliasName_ == null) return null;
    if (isHardReference_) return null;

    // if the name is not null and the reference is,
    // try to get it, after an ACS opening for instance
    if (hardReferenceOrAlias_ == null)
        hardReferenceOrAlias_ = (ACSObject) UtilityImpl.getFirstInstance().getBaseObject(
                hardReferenceOrAliasName_);

    if (hardReferenceOrAlias_ == null) {
        hardReferenceOrAliasName_ = null;
        return null;
    }
    return hardReferenceOrAlias_;
  } // getEorReferenceAsInternalHardAlias()

  /**
   * Get the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard reference. Calls UtilityImpl.getBaseObject().
   * @return the hard alias. May be null.
   */
  protected ACSObject getEorAliasAsInternalHardReference() {
    if (hardReferenceOrAliasName_ == null) return null;
    if (!isHardReference_) return null;

    // if the name is not null and the reference is,
    // try to get it, after an ACS opening for instance
    if (hardReferenceOrAlias_ == null)
        hardReferenceOrAlias_ = (ACSObject) UtilityImpl.getFirstInstance().getBaseObject(
                hardReferenceOrAliasName_);

    if (hardReferenceOrAlias_ == null) {
        hardReferenceOrAliasName_ = null;
        return null;
    }
    return hardReferenceOrAlias_;
  } // getEorAliasAsInternalHardReference()

  //
  // --------------------- other protected CAlias methods ----------------------
  //

  /**
   * Sets the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard alias.
   * The generic gBase package does not call this method, nor the user from the GUI.
   * It is the responsibility of the ACS associations or the AcsAddon code
   * to set an internal alias for one new ACSObject, as an alias or as an alias reference.
   * If the master is already a reference, it cannot become an alias.
   * Fires the property change 'HardAliasReference'.
   * @param _ref the reference. May be null.
   * @exception UpDateError _basic does not belong to the same ACS or the master
   * is already an alias reference
   */
  protected void setEorReferenceAsInternalHardAlias(ACSObject _ref) throws UpDateError {
      // noop if the instance is finalized
    if (l_SoftAlias_ == null) return;
    if (_ref != null && _ref.getEorACS() != master_.getEorACS())
        throw new UpDateError("This object does not belongs to the same ACS");
    if (isHardReference_)
        throw new UpDateError("This object is a hard reference; it cannot become an alias.");

    ImmutableName old = hardReferenceOrAliasName_;
    // reset by getEorReferenceAsInternalHardAlias(), and it is not a bound property
    hardReferenceOrAlias_ = null;
    if (_ref != null)
        hardReferenceOrAliasName_ = _ref.getName();
    else
        hardReferenceOrAliasName_ = null;

    try {
        firePropertyChange("HardAliasReference", old, hardReferenceOrAliasName_);
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From CAliasImpl.setEorReferenceAsInternalHardAlias(), firePropertChange() throws:");
            e.printStackTrace();
    }
  } // setEorReferenceAsInternalHardAlias()

  /**
   * Sets the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard reference. Calls setEorReferenceAsInternalHardAlias()
   * on the alias.
   * The generic gBase package does not call this method, nor the user from the GUI.
   * It is the responsibility of the ACS associations or the AcsAddon code
   * to set an internal alias for one new ACSObject, as an alias or as an alias reference.
   * If the master is already an alias, it cannot become a reference.
   * Calls getEorReferenceAsInternalHardAlias(). Called by finalizeForBase()
   * from the alias.
   * Fires the property change 'HardAlias'.
   * @param _ali the alias. May be null.
   * @exception UpDateError _basic does not belong to the same ACS or
   * the master is already an alias
   */
  protected void setEorAliasAsInternalHardReference(ACSObject _ali) throws UpDateError {
      // noop if the instance is finalized
    if (l_SoftAlias_ == null) return;
    if (_ali != null && _ali.getEorACS() != master_.getEorACS())
        throw new UpDateError("This object does not belongs to the same ACS");
    if (getEorReferenceAsInternalHardAlias() != null)
        throw new UpDateError("This object is a hard alias; it cannot become a reference.");

    ImmutableName old = hardReferenceOrAliasName_;
    // reset by getEorReferenceAsInternalHardAlias(), and it is not a bound property
    hardReferenceOrAlias_ = null;
    if (_ali != null)
        hardReferenceOrAliasName_ = _ali.getName();
    else
        hardReferenceOrAliasName_ = null;
    
    isHardReference_ = true;

    if (_ali instanceof VirtualFolderImpl)
        ((VirtualFolderImpl)_ali).setEorReferenceAsInternalHardAlias(master_);
    else if (_ali instanceof BasicImpl)
        ((BasicImpl)_ali).setEorReferenceAsInternalHardAlias(master_);
    else if (_ali != null)
        throw new InternalError("CAliasImpl.setEorAliasAsInternalHardReference()"+
                " with a wrong class for the master: "+_ali.getClass());
    
                        // for testing and tracking
                        //System.out.println("CAliasImpl.setEorAliasAsInternalHardReference() fires "+
                          //      " the property change 'HardAlias' for "+master_.getName());


    try {
        firePropertyChange("HardAlias", old, hardReferenceOrAliasName_);
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From CAliasImpl.setEorAliasAsInternalHardReference(), firePropertChange() throws:");
            e.printStackTrace();
    }
  } // setEorAliasAsInternalHardReference()

  /**
   * Adds an external or internal soft alias to the master as soft reference.
   * The alias is normally updated by the ACS calling method.
   * In the standard way of use, called only by the ACS through the master,
   * which may add controls on the parameter class.
   * Has to be used by the AcsAddon.
   * Fires the events 'SoftAlias' and 'hasSoftAlias'.
   * @param _alias ACSObject. No action if null, or already known.
   * @exception UpDateError _alias belongs to the same ACS or the type is
   * immutable
   */
  protected void addEorSoftAlias(ImmutableName _alias) throws UpDateError {
      // noop if the instance is finalized
    if (l_SoftAlias_ == null) return;

    // variables
    boolean recorded = false;
    ImmutableName[] l_oldSoftAlias;
    ImmutableName[] newArray;
    
                            // for testing and tracking
                            //System.out.println("Call to CNodeImpl.addEorChild() on "+
                              //      master_.getNickName());

    if (_alias == null ) return;

    if (l_SoftAlias_ == null)
        throw new InternalError("In CAliasImpl.addEorSoftAlias(), the list is null.");

    //if (master_.getEorACS().getEorM_Resources().get(_alias) != null ||
      //      master_.getEorACS().getEorM_EligibleParties().get(_alias) != null ||
        //    master_.getEorACS().getOneVirtualFolder(_alias) != null )
        //throw new UpDateError("In CAliasImpl.addEorSoftAlias(), "+
          //      "the alias is not external, from another ACS.");
    
    // tests if _alias is known
    for (int i = 0 ; i < l_SoftAlias_.length ; i++) {
        if (_alias.equals(l_SoftAlias_[i]) ) {
          recorded = true;
          break;
        }
    } // for
    if (recorded) return;

    l_oldSoftAlias = new ImmutableName[l_SoftAlias_.length];
    System.arraycopy(l_SoftAlias_, 0, l_oldSoftAlias, 0, l_SoftAlias_.length);

    // unknown, so update the array in an empty place if possible
    for (int i = 0 ; i < l_SoftAlias_.length ; i++) {
        if (l_SoftAlias_[i] == null) {
            l_SoftAlias_[i] = _alias;
            recorded = true;
            break;
        }
    } // for
        // else, adds at the end of the array
    if (!recorded) {
        newArray = new ImmutableName[l_SoftAlias_.length + 1];
        System.arraycopy(l_SoftAlias_, 0, newArray, 0, l_SoftAlias_.length);
        newArray[l_SoftAlias_.length] = _alias;
        l_SoftAlias_ = newArray;
    }

    try {
      Arrays.sort(l_SoftAlias_);
    } catch (ClassCastException c) {
      throw new InternalError("In CAliasImpl.addEorSoftAlias(), exception"+
        " on sort operation: "+c.toString());
    } 
    
    try {
      firePropertyChange("SoftAlias", l_oldSoftAlias, getL_SoftAlias());
      if (l_oldSoftAlias.length == 0)
          firePropertyChange("hasSoftAlias", false, true);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CAliasImpl.addEorSoftAlias(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    l_oldSoftAlias = null;
    newArray = null;
  } // addEorSoftAlias()

  /**
   * Removes an external or internal alias from the master as soft reference.
   * The alias is normally updated by the ACS calling method.
   * In the standard way of use, called only by the ACS through the master,
   * which may add controls on the parameter class.
   * Detects the alias through a name comparison.
   * Fires the 'SoftAlias' and the 'hasSoftAlias' property change events.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the list is null.
   * @param _alias alias of the master. No action if null.
   * @exception UpDateError if _alias is not known.
   */
  protected void removeEorSoftAlias(ImmutableName _alias) throws UpDateError {
      // noop if the instance is finalized
    if (l_SoftAlias_ == null) return;

    boolean         recorded = false;
    ImmutableName[] l_oldSoftAlias;
    short           shiftingBack = 0;
    
        // for testing and tracking
        //System.out.println("Call to CAliasImpl.removeEorChild() on "+getName());

    if (_alias == null ) return;
    if (l_SoftAlias_ == null || l_SoftAlias_.length == 0)
        throw new UpDateError("In CAliasImpl.removeEorSoftAlias(), _alias is unknown.");

    l_oldSoftAlias = new ImmutableName[l_SoftAlias_.length];
    System.arraycopy(l_SoftAlias_, 0, l_oldSoftAlias, 0, l_SoftAlias_.length);

    ImmutableName[]     newArray = new ImmutableName[l_SoftAlias_.length];
    ImmutableName[]     result = new ImmutableName[0];
    
    for (int  i = 0 ; i < l_SoftAlias_.length ; i++) {
      if (l_SoftAlias_[i] != null) {
        if (l_SoftAlias_[i].equals(_alias) ) {
          // increments shiftingBack delete each reference
          shiftingBack++;
          recorded = true;
        }
        else newArray[i - shiftingBack] = l_SoftAlias_[i];
      }
      else shiftingBack++; // deletes null value
    } // for
    
    if (!recorded)
        throw new UpDateError("In CAliasImpl.removeEorSoftAlias(), _alias is unknown.");

    if (l_SoftAlias_.length != shiftingBack) {
      result = new ImmutableName[l_SoftAlias_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
    }
              // else result is a 0-size array
    l_SoftAlias_ = result;

    try {
      Arrays.sort(l_SoftAlias_);
    } catch (ClassCastException c) {
      throw new InternalError("In CAliasImpl.removeEorSoftAlias(), exception"+
        " on sort operation: "+c.toString());
    }    

    try {
      firePropertyChange("SoftAlias", l_oldSoftAlias, getL_SoftAlias());
      if (l_SoftAlias_.length == 0)
          firePropertyChange("hasSoftAlias", true, false);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CAliasImpl.removeEorSoftAlias(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    l_oldSoftAlias = null;
    result = null;
    newArray = null;
  } // removeEorSoftAlias()

  /**
   * Adds the ACS name of an external or internal soft reference to the master as soft alias.
   * The reference is updated by the ACS calling method.
   * Used only by the ACS through the master, which may add controls on the parameter class.
   * Has to be used by the AcsAddon.
   * Fires the event 'AcsOfSoftReferences'.
   * @param _acs ACS name. No action if null, or already known.
   */
  protected void addAcsOfSoftReferences(ImmutableName _acs) throws UpDateError {
      // noop if the instance is finalized
    if (l_AcsOfSoftReferences_ == null) return;

    // variables
    boolean recorded = false;
    ImmutableName[] l_old;
    ImmutableName[] newArray;
    
        // for testing and tracking
        //System.out.println("Call to CAliasImpl.addAcsOfSoftReferences() on "+getName());

    if (_acs == null ) return;

    if (l_AcsOfSoftReferences_ == null)
        throw new InternalError("In CAliasImpl.addAcsOfSoftReferences(), the list is null.");
    // tests if _alias is known
    for (int i = 0 ; i < l_AcsOfSoftReferences_.length ; i++) {
        if (_acs.equals(l_AcsOfSoftReferences_[i]) ) {
          recorded = true;
          break;
        }
    } // for
    if (recorded) return;

    l_old = new ImmutableName[l_AcsOfSoftReferences_.length];
    System.arraycopy(l_AcsOfSoftReferences_, 0, l_old, 0, l_AcsOfSoftReferences_.length);

    // unknown, so update the array in an empty place if possible
    for (int i = 0 ; i < l_AcsOfSoftReferences_.length ; i++) {
        if (l_AcsOfSoftReferences_[i] == null) {
            l_AcsOfSoftReferences_[i] = _acs;
            recorded = true;
            break;
        }
    } // for
        // else, adds at the end of the array
    if (!recorded) {
        newArray = new ImmutableName[l_AcsOfSoftReferences_.length + 1];
        System.arraycopy(l_AcsOfSoftReferences_, 0, newArray, 0, l_AcsOfSoftReferences_.length);
        newArray[l_AcsOfSoftReferences_.length] = _acs;
        l_AcsOfSoftReferences_ = newArray;
    }

    try {
      Arrays.sort(l_AcsOfSoftReferences_);
    } catch (ClassCastException c) {
      throw new InternalError("In CAliasImpl.addAcsOfSoftReferences(), exception"+
        " on sort operation: "+c.toString());
    } 
    
    try {
      firePropertyChange("AcsOfSoftReferencess", l_old, getL_AcsOfSoftReferences());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CAliasImpl.addAcsOfSoftReferences(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    l_old = null;
    newArray = null;
  } // addEorSoftAlias()

  /**
   * Removes an external or internal alias from the master as soft reference.
   * The alias is updated by the ACS calling method.
   * Used only by the ACS through the master, which may add controls on the parameter class.
   * Detects the alias through a name comparison.
   * Fires 'AcsOfSoftReferences' property change events.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the list is null.
   * @param _acs ACS name. No action if null.
   * @exception UpDateError if _alias is not known.
   */
  protected void removeAcsOfSoftReferences(ImmutableName _acs) throws UpDateError {
      // noop if the instance is finalized
    if (l_AcsOfSoftReferences_ == null) return;

    boolean         recorded = false;
    ImmutableName[] l_old;
    short           shiftingBack = 0;
    
        // for testing and tracking
        //System.out.println("Call to CAliasImpl.removeEorChild() on "+getName());

    if (_acs == null ) return;
    if (l_AcsOfSoftReferences_ == null || l_AcsOfSoftReferences_.length == 0)
        throw new UpDateError("In CAliasImpl.removeAcsOfSoftReferences(), _alias is unknown.");

    l_old = new ImmutableName[l_AcsOfSoftReferences_.length];
    System.arraycopy(l_AcsOfSoftReferences_, 0, l_old, 0, l_AcsOfSoftReferences_.length);

    ImmutableName[]     newArray = new ImmutableName[l_AcsOfSoftReferences_.length];
    ImmutableName[]     result = new ImmutableName[0];
    
    for (int  i = 0 ; i < l_AcsOfSoftReferences_.length ; i++) {
      if (l_AcsOfSoftReferences_[i] != null) {
        if (l_AcsOfSoftReferences_[i].equals(_acs) ) {
          // increments shiftingBack delete each reference
          shiftingBack++;
          recorded = true;
        }
        else newArray[i - shiftingBack] = l_AcsOfSoftReferences_[i];
      }
      else shiftingBack++; // deletes null value
    } // for
    
    if (!recorded)
        throw new UpDateError("In CAliasImpl.removeAcsOfSoftReferences(), _acs is unknown.");

    if (l_AcsOfSoftReferences_.length != shiftingBack) {
      result = new ImmutableName[l_AcsOfSoftReferences_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
    }
              // else result is a 0-size array
    l_AcsOfSoftReferences_ = result;

    try {
      Arrays.sort(l_AcsOfSoftReferences_);
    } catch (ClassCastException c) {
      throw new InternalError("In CAliasImpl.removeAcsOfSoftReferences(), exception"+
        " on sort operation: "+c.toString());
    }    

    try {
      firePropertyChange("AcsOfSoftReferences", l_old, getL_AcsOfSoftReferences());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CAliasImpl.removeAcsOfSoftReferences(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    l_old = null;
    result = null;
    newArray = null;
  } // removeAcsOfSoftReferences()

  // ------------------ package methods --------------------------

  /**
  * Adds a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to add
  */
  protected void addPropertyChangeListener(PropertyChangeListener _l){

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for all the bound properties:
  * 'SoftAlias', 'hasSoftAlias', 'AcsOfSoftReferences', 'HardAliasReference', 'HardAlias'.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to add
  */
  protected void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}

              // for testing and tracking
              //System.out.println("in CAliasImpl.addPropertyChangeListener(), "+
              //  "call from "+_l.getClass()+" to "+getName()+" for the property: "
              //  +_propertyName);
  } // addPropertyChangeListener()

  /**
  * Removes a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_l);}
  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for all the bound properties:
  * 'SoftAlias', 'hasSoftAlias', 'AcsOfSoftReferences', 'HardAliasReference', 'HardAlias'.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

      return changeSupport_.getPropertyChangeListeners(_propertyName);
  } // getPropertyChangeListeners()

  /**
  * Finalizes the CAlias and the hard alias, for a master hard reference. Does not finalize
  * the hard reference of a master alias, but unregister the master in the reference.
  * Calls getEorReferenceAsInternalHardAlias() and getEorAliasAsInternalHardReference().
  * Calls setEorAliasAsInternalHardReference() on the reference.
  * HAVE TO BE USED before the deleting of the CAlias.
  * @exception UpDateError if the CAlias is unknown of the main ACS (if
  *  it is a clone, typically)
  */
  protected synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call of CAliasImpl.finalizeForUser() for "+getName());

    // noop if it is the second call to this method
    if (l_SoftAlias_ == null) {
      return;
    }

    // finalize the master hard alias, if any
    if (getEorAliasAsInternalHardReference() != null) {
        ACSObject obj = getEorAliasAsInternalHardReference();
        if (obj instanceof BasicImpl)
            ((BasicImpl)obj).finalizeForBase();
        else if (obj instanceof VirtualFolderImpl)
            ((VirtualFolderImpl)obj).finalizeForBase();
    }

    // unregisters in the master hard reference, if any
    if (getEorReferenceAsInternalHardAlias() != null) {
        ACSObject obj = getEorReferenceAsInternalHardAlias();
        if (obj instanceof BasicImpl)
            ((BasicImpl)obj).setEorAliasAsInternalHardReference(null);
        else if (obj instanceof VirtualFolderImpl)
            ((VirtualFolderImpl)obj).setEorAliasAsInternalHardReference(null);
    }

    hardReferenceOrAliasName_ = null;
    hardReferenceOrAlias_ = null;
    l_SoftAlias_ = null;
    l_AcsOfSoftReferences_ = null;

    master_ = null;
  } // finalizeForBase()

  /**
  * This method finalizes the instance, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called for EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

              // for testing purpose
              //System.out.println("call of CAliasImpl.finalizeForUser() for "+getName());

    // noop if it is the second call to this method
    if (l_SoftAlias_ == null) {
      return;
    }

    hardReferenceOrAliasName_ = null;
    hardReferenceOrAlias_ = null;
    l_SoftAlias_ = null;
    l_AcsOfSoftReferences_ = null;

    master_ = null;
  } // finalizeFromDeserialization()

   /**
   * Clone for embeddd class without registering the new object in the ACSs.
   * Caution: don't use any modifying method that may altere the integrity
   * of the base.
   * The soft alias name list is cloned, and not the hard alias reference.
   * The external object references are not cloned (see "Eor" in get method).
   * @param _master the clone master
   * @exception CloneNotSupportedException
   * @return cloned object with _master as the master
   */
  protected synchronized CAliasImpl cloneC(ACSObject _master) {
    CAliasImpl clon;
    try {
        clon = (CAliasImpl) super.clone();

        clon.l_SoftAlias_ = new ImmutableName[l_SoftAlias_.length];
        System.arraycopy(l_SoftAlias_, 0, clon.l_SoftAlias_, 0, l_SoftAlias_.length);

        clon.l_AcsOfSoftReferences_ = new ImmutableName[l_AcsOfSoftReferences_.length];
        System.arraycopy(l_AcsOfSoftReferences_, 0, clon.l_AcsOfSoftReferences_, 0, l_AcsOfSoftReferences_.length);

        clon.master_ = _master;
    } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
      if (master_ == null) throw new InternalError("In CAliasImpl, the master is null");
      else {
        ((ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(master_.getAcsName())).
                getEorBaseError().addException(e.toString() + "CAliasImpl.clone() failed");
        throw new InternalError(e.toString());
      }
    } // catch

    return clon;
  } // cloneC()

  /**
 * @param _obj Object
 * @return true if _obj is a CAliasImpl instance (or an instance of an
 *  extended class) equals to the current one
 */
public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      CAliasImpl obj = (CAliasImpl) _obj;

      if (master_ != obj.master_ )
          return false;

      if (l_SoftAlias_ == null || obj.getL_SoftAlias() == null) {
          if (l_SoftAlias_ != null || obj.getL_SoftAlias() != null)
            return false;
      } // if
      else if (l_SoftAlias_.length != obj.getL_SoftAlias().length) // no test: length == 0
        return false;
      else {
          for (int i = 0; i < l_SoftAlias_.length; i++) {
            if (l_SoftAlias_[i] == null || obj.getL_SoftAlias()[i] == null) {
                if (l_SoftAlias_[i] != null || obj.getL_SoftAlias()[i] != null)
                  return false;
                continue; // i++
            } // if
            else {
                        // test of equals only on children names
                if (!l_SoftAlias_[i].equals( obj.getL_SoftAlias()[i]))
                  return false;
            } // else
          } // for
      } // else

      if (l_AcsOfSoftReferences_ == null || obj.getL_AcsOfSoftReferences() == null) {
          if (l_AcsOfSoftReferences_ != null || obj.getL_AcsOfSoftReferences() != null)
            return false;
      } // if
      else if (l_AcsOfSoftReferences_.length != obj.getL_AcsOfSoftReferences().length) // no test: length == 0
        return false;
      else {
          for (int i = 0; i < l_AcsOfSoftReferences_.length; i++) {
            if (l_AcsOfSoftReferences_[i] == null || obj.getL_AcsOfSoftReferences()[i] == null) {
                if (l_AcsOfSoftReferences_[i] != null || obj.getL_AcsOfSoftReferences()[i] != null)
                  return false;
                continue; // i++
            } // if
            else {
                        // test of equals only on children names
                if (!l_AcsOfSoftReferences_[i].equals( obj.getL_AcsOfSoftReferences()[i]))
                  return false;
            } // else
          } // for
      } // else

      return true;
  } // equals

    /**
     * @return long: serial version UID for serialization of the instance
     */
    public long getSerialVersionUID()	{
            return serialVersionUID;
    }

  /**
  * Fires an event to every registered listener, in any order.
  * For the properties 'SoftAlias', 'hasSoftAlias', 'AcsOfSoftReferences',
  * 'HardAliasReference', 'HardAlias'.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //System.out.println("CAliasImpl.firePropertyChange(), call to "+
                  //"propertyChange() of "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
   * This method removes the CAlias from the ACS lists, but there is no
   * guarantee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {}

    super.finalize();
  }


  //
  // ------------------------- private methods -----------------------------
  //


} // CAliasImpl


