//Title:        Access Road
//Version:      0.7.0
//Copyright:    2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road CFolderMember).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Collections;

import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * This class is an embedded class responsible for modeling a folder member 
 * which is contained in a EPRViewInBase or in a VirtualFolder.
 * An instance is used as a private attribute of a composite class instance
 * (called the master instance) which delegates the processing of its membership
 * to an instance of this class.
 * The master class is a Basic or a VirtualFolder. When it is a VirtualFolder,
 * it is only to handle the EPRViewInBases, and never the VirtualFolders. <P>
 * This class implements some methods of its master,
 * but this is not directly controlled by Java since this class does not
 * implement the interface to avoid to be a public class.
 * All variables are private to secure the white box inheritances
 * from this class.
 * <P>
 * It is a javabean with the following bound properties: <P>
 * - 'ImpliedOpenViews', 'VirtualFolders'. <P>
 * The listeners are proper to this instance.
 * Excepting for ViewInBase, they are outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. The exceptions from the listeners are not catched.
 * This class is public to be overriden in the AcsAddons, and in this case,
 * the subclass is only called by some classes in the same package.
 * @version 0.7.0
 * @see ARoad0.gBaseInterface.ViewInBase
 * @see ARoad0.gBaseInterface.VirtualFolder
*/
public class CFolderMemberImpl implements Serializable {
  /**
  * Protected variables
  */
  protected static final int  INITIAL_CAPACITY = 10; // the default value: 101 for map
  protected static final int  INITIAL_CAPACITY_2 = 4 ;  // the default value is 10
  protected static final int  INCREMENT_CAPACITY = 2;
  protected static final int  PRIME = 1000003;
  /**
  * private variables
  */
  private static final long       serialVersionUID = 0;

  /*
   * The master is a BasicImpl or a VirtualFolderImpl
   */
  protected ACSObject             master_;
  /**
   * list of open ViewInBase that contains this instance
   * Created only when it is needed.
   */
  protected transient ArrayList<ViewInBase>  l_OpenViews_;
  /**
   * list of VirtualFolder that contains this instance.
   * Created only when it is needed. Always null if the master is a VirtualFolderImpl.
   */
  protected ArrayList<VirtualFolder>         l_VirtualFolders_ = null;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final CFolderMemberImpl EMPTY_INSTANCE = new CFolderMemberImpl();

  /* Out of the extended interfaces */

   /**
   * manages all the property change listeners
   */
  protected transient PropertyChangeSupport changeSupport_ = null;
  
  /** variable for the JDK 2 serialization
  * @serial
  * ("l_VirtualFolders_", ArrayList.class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
        new ObjectStreamField("l_VirtualFolders_", ArrayList.class)
  };

 /**
  * The basic constructor.
  */
  public CFolderMemberImpl() {


                          // for testing purpose
                     //System.out.println("Call to CFolderMemberImpl.CFolderMemberImpl()");
  }
 
  /**
  * Used by the master constructor.
  * @param _master master of this embedded class
  * @exception CreateError the master is null
  */
  public CFolderMemberImpl(ACSObject _master) throws CreateError {
        if (_master == null ||
                (!(_master instanceof BasicImpl) && !(_master instanceof VirtualFolderImpl)))
             throw new CreateError("At the CFolderMember construction, "+
                "the master is null, or is not a BasicImpl nor a VirtualFolderImpl.");
        master_ = _master;
  } // constructor
  
  
  /**
  * This method provides a copy of the instance
  * with another name, without children copying
  * because the copy order will not be always correct.
  * @param _masterCopy is the master of the new copy
  * @return copy as CFolderMemberImpl instance
  * @exception CreateError of the constructor with the same parameters.
  **/
  protected CFolderMemberImpl copy(ACSObject _masterCopy) throws CreateError {
      
      CFolderMemberImpl result = new CFolderMemberImpl(_masterCopy);

      // TODO: copy the properties
      
      return result;
  } // copy()

   /**
   * Gets the open views containing the master.
   * @return List of ViewInBase. May be empty,
   *   but never null.
   */
  protected List<ViewInBase> getImpliedViews() {
      if (l_OpenViews_ == null) return Collections.EMPTY_LIST;
      else return (List)((ArrayList)l_OpenViews_).clone();
  } // getImpliedViews()  

   /**
   * Gets the virtual folders containing the master.
   * @return List of VirtualFolders. May be empty,
   *   but never null.
   */
  protected List<VirtualFolder> getVirtualFolders() {
      if (l_VirtualFolders_ == null) return Collections.EMPTY_LIST;
      else return (List)l_VirtualFolders_.clone();
  } // getVirtualFolders() 

   /**
   * Registers the view as containing the master.
   * Fires 'ImpliedOpenViews' event.
   */
  protected void addInViewList(ViewInBase _view) {
      if (l_OpenViews_ == null)
          l_OpenViews_ = new ArrayList(INITIAL_CAPACITY_2);
      List l_oldList = (List)l_OpenViews_.clone();
      synchronized (l_OpenViews_) {
	l_OpenViews_.add(_view);
      }
      firePropertyChange("ImpliedOpenViews", l_oldList, getImpliedViews());
      l_oldList = null;
  } // addInViewList()

   /**
   * Unregisters the view when it does not contain the master.
   * Fires 'ImpliedOpenViews' event.
   * @return false if the view is unknown or the instance is finalized
   */
  protected boolean removeInViewList(ViewInBase _view) {
      if (l_OpenViews_ == null) return false;
      boolean result = false;
      List l_oldList = (List)l_OpenViews_.clone();
      synchronized (l_OpenViews_) {      
        result = l_OpenViews_.remove(_view);
      }
      if (result) firePropertyChange("ImpliedOpenViews", l_oldList, getImpliedViews());
      l_oldList = null;
      return result;      
  } // removeInViewList()  
  
   /**
   * Registers the folder as containing the master.
   * Fires 'VirtualFolders' event.
   * @exception CreateError if the ACS or the master does not manage the virtual folders.
   */
  protected void addInVFolderList(VirtualFolder _folder) throws CreateError {
      if (!(master_ instanceof BasicImpl))
          throw new CreateError("in CFolderMemberImpl.addInVFolderList(), "+
             "the master does not manage the virtual folders as members.");
      if (!((BasicImpl)master_).getEorACS().getManageVirtualFolders())
          throw new CreateError("in CFolderMemberImpl.addInVFolderList(), "+
             "the ACS does not manage the virtual folders.");
      if (l_VirtualFolders_ == null)
          l_VirtualFolders_ = new ArrayList(INITIAL_CAPACITY_2);
      List l_oldList = (List)l_VirtualFolders_.clone();
      synchronized (l_VirtualFolders_) {
	l_VirtualFolders_.add(_folder);
      }
      firePropertyChange("VirtualFolders", l_oldList, getVirtualFolders());
      l_oldList = null;
  } // addInVFoldersList()

   /**
   * Unregisters the folder when it does not contain the master.
   * Called by VirtualFolderImpl.removeEorBaseObject(), sometimes from the method
   * finalizeForBase() of this instance.
   * Fires the 'VirtualFolders' event.
   * @param _folder to remove. May be finalized, e. g. with a null baseError_.
   * @exception CreateError if the ACS or the master does not manage the virtual folders.
   * @return false if the folder is unknown, if the instance or the argument is finalized
   */
  protected boolean removeInVFolderList(VirtualFolder _folder) throws CreateError {
      if (l_VirtualFolders_ == null) {

                      // for testing and tracking
                      //System.out.println("CFolderMemberImpl.removeInVFolderList(), "+
                        //"null list in the removing for "+_folder.getName());

          return false;
      }
      if (!(master_ instanceof BasicImpl))
          throw new CreateError("in CFolderMemberImpl.removeInVFolderList(), "+
             "the master does not manage the virtual folders as members.");
      if (!((BasicImpl)master_).getEorACS().getManageVirtualFolders())
          throw new CreateError("in CFolderMemberImpl.removeInVFolderList(), "+
             "the ACS does not manage the virtual folders.");

      boolean result = false;
      List l_oldList = (List)l_VirtualFolders_.clone();

                      // for testing and tracking
                      //System.out.println("CFolderMemberImpl.removeInVFolderList(), "+
                        //"start of removing for "+_folder.getName());

      synchronized (l_VirtualFolders_) {      
        result = l_VirtualFolders_.remove(_folder);
      }
      // ((VirtualFolderImpl)_folder).baseError_ null if _folder is finalized
      if (((VirtualFolderImpl)_folder).baseError_ != null && !result)
          throw new CreateError("in CFolderMemberImpl.removeInVFolderList(), "+
             "no operation for "+_folder.getName());

      if (result)
          firePropertyChange("VirtualFolders", l_oldList, getVirtualFolders());
              
      l_oldList = null;
      return result;      
  } // removeInVFoldersList()   
  
  /**
  * Adds a listener for all the bound properties.
  * @param _l PropertyChangeListener to add
  */
  protected void addPropertyChangeListener(PropertyChangeListener _l){

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used for all the bound properties:
  *   'ImpliedOpenViews', 'VirtualFolders'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add
  */
  protected void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}

              // for testing and tracking
              //System.out.println("in CFolderMemberImpl.addPropertyChangeListener(), "+
              //  "call from "+_l.getClass()+" to "+getName()+" for the property: "
              //  +_propertyName);

  } // addPropertyChangeListener()

  /**
  * Removes a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("CFolderMemberImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_l);}
  } // removePropertyChangeListener()

  /**
  * Removes a listener.
  * Used for all the bound properties:
  *   'ImpliedOpenViews', 'VirtualFolders'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("CFolderMemberImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

      return changeSupport_.getPropertyChangeListeners(_propertyName);
  } // getPropertyChangeListeners()

  /**
  * Finalizes the CFolderMember. Removes the master in the virtual folders.
  * HAVE TO BE USED before deleting the CFolderMember instance.
  * @exception UpDateError if VirtualFolder.removeEorBaseObject() throws an exception
  */
  protected synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call of CFolderMemberImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (l_OpenViews_ == null) {
      return;
    }

    if (l_VirtualFolders_ != null) {
        List l_oldList = (List)l_VirtualFolders_.clone();
        try {
            for (Iterator ite = l_oldList.iterator(); ite.hasNext();) {
                ((VirtualFolder)ite.next()).removeEorBaseObject(master_);
            }
        } catch (CreateError b) {
            throw new UpDateError("CFolderMemberImpl.finalizeForBase(): "+b.toString());
        }
        l_oldList = null;
    } // if
    
    //ViewInBases are informed through the master ACS
    l_OpenViews_ = null;
    l_VirtualFolders_ = null;
    master_ = null;
  } // finalizeForBase()

  /**
  * This method finalizes all the virtual folders, since it is used after
  * a deserialization. There, they have been deserialized but
  * are not 'true' objects of the base. No action on the open views, always empty.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called for EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

              // for testing purpose
              //System.out.println("call of CFolderMemberImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (l_OpenViews_ == null) {
      return;
    }

    if (l_VirtualFolders_ != null) {
        List l_oldList = (List)l_VirtualFolders_.clone();
        for (Iterator ite = l_oldList.iterator(); ite.hasNext();) {
            ((VirtualFolderImpl)ite.next()).finalizeFromDeserialization();
        }
        l_oldList = null;
    } // if
    
    l_OpenViews_ = null;
    l_VirtualFolders_ = null;
    master_ = null;
  } // finalizeFromDeserialization()


   /**
   * Clone for embedded class without registering the new object in the ACSs.
   * Caution: don't use any modifying method that may altere the integrity
   * of the base.
   * The list are cloned without deep copy.
   * The external object references are not cloned (see "Eor" in get method).
   * The links between the EOR and the new instance are not set, but the lists
   * of OER are cloned (viewInBase, virtualFolder) so it is possible to
   * get these active objects.
   * @exception CloneNotSupportedException
   * @return cloned object with _master as the master
   */
  protected synchronized CFolderMemberImpl cloneC(Basic _master) {
    CFolderMemberImpl clon;
    try {
        clon = (CFolderMemberImpl) super.clone();

        clon.l_OpenViews_ = (ArrayList) l_OpenViews_.clone();
        if (l_VirtualFolders_ != null) 
            clon.l_VirtualFolders_ = (ArrayList)l_VirtualFolders_.clone();

        clon.master_ = _master;
    } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
      if (master_ == null) throw new InternalError("In CFolderMemberImpl, the master is null");
      else {
        throw new InternalError(e.toString());
      }
    } // catch
        
    return clon;
  }

  /**
   * @return the hashcode derived from the master nick name and the open views
   */
  public int hashCode() {
    int result = 0;

    if (master_ != null) {
      result = "CFolderMemberImpl".hashCode() + master_.getNickName().hashCode();
    }

    //if (l_VirtualFolders_ != null) {
      //result = PRIME*result + l_VirtualFolders_.hashCode();
    //}

    //if (l_OpenViews_ != null) {
            // FolderAbst.hashcode() return null when name_ is null
      //result = PRIME*result + l_OpenViews_.hashCode();
    //}

    return result;
  } // hashcode()
  
  /**
 * @param _obj Object
 * @return true if _obj is a CFolderMemberImpl instance (or an instance of an
 *  extended class) equals to the current one
 */
 public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      CFolderMemberImpl obj = (CFolderMemberImpl) _obj;

      if (master_ != obj.master_ )
          return false;
      /*
      if (l_OpenViews_ == null || obj.getImpliedViews().equals(Collections.EMPTY_LIST)) {
          if (l_OpenViews_ != null || !obj.getImpliedViews().equals(Collections.EMPTY_LIST))
            return false;
      } // if
      else {
          if (!l_OpenViews_.equals( obj.getImpliedViews() ))
            return false;
      } // else
      
      if (l_VirtualFolders_ == null || obj.getVirtualFolders().equals(Collections.EMPTY_LIST)) {
          if (l_VirtualFolders_ != null || !obj.getVirtualFolders().equals(Collections.EMPTY_LIST))
            return false;
      } // if
      else {
          if (!l_VirtualFolders_.equals( obj.getVirtualFolders() ))
            return false;
      } // else
       */

      return true;
  } // equals

    /**
     * @return long: serial version UID for serialization of the instance
     */
    public long getSerialVersionUID()	{
            return serialVersionUID;
    }

  /**
  * Fires an event to every registered listener, in any order.
  * For the properties: 'ImpliedOpenViews', 'VirtualFolders'.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

      if (_oldValue != null && _newValue != null && _oldValue.equals(_newValue)) {
          return;
      }

      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //System.out.println("CFolderMemberImpl.firePropertyChange(), call to "+
                  //"propertyChange() of "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
   * This method removes the CFolderMember from the ACS lists, but there is no
   * guaranee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {}

    super.finalize();
  }

} // CFolderMemberImpl

