// Title:        Access Road
// Version:      0.7.0
// Copyright:    2006, 2010 EURL ACCBEE
// Author:       Patrick Thazard
// Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road CNode).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import java.util.Arrays;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * This class is an embedded class responsible for modeling a node 
 * which contains children (leaves and nodes)
 * as a private attribute of a composite class instance, called the master instance.
 * The master delegates the processing of its children to an instance of this class.
 * The master class may be a Directory or a VirtualFolder, and
 * it transfers the operations on the children to this class.
 * The children are Resources or VirtualFolders. They all deleted when the node is deleted,
 * on the contrary of the members of a virtual folder, for instance.
 * This class implements the Node interface for its master,
 * but this is not directly controlled by Java since this class does not
 * implement the interface to avoid to be a public class.
 * The Leaf interface is not processed, so the master has to managed it
 * since any node is a leaf. <P>
* All variables are private to secure the white box inheritances
* from this class.
* It is a javabean with the following bound properties: <P>
* - 'Children'. <P>
* The listeners are proper to this instance.
* Excepting for EPRViewInBaseImpl, they are outside gBase,
* transient (not serialized in this instance backup).
* They are called in any order. They receive only a copy of the new value,
* to protect the property. All the exceptions from the listeners are catched,
* and a dialog box is displayed to inform the user.
* This class is public to be overriden in the AcsAddons, and in this case,
* the subclass is only called by some classes in the same package.
* See the <a href="{@docRoot}/copyright.html">Copyright</a>.
* @see ARoad0.gBaseInterface.Node
* @see ARoad0.gBaseInterface.Resource
* @see ARoad0.gBaseInterface.VirtualFolder
 */
public class CNodeImpl implements Serializable {
  /**
  * Protected variables
  */
  protected static final int  INITIAL_CAPACITY = 10; // the default value: 101 for map
  protected static final int  INITIAL_CAPACITY_2 = 4 ;  // the default value is 10
  protected static final int  INCREMENT_CAPACITY = 2;    
  /* private variables
  */
  private static final long       serialVersionUID = 0;

  protected Node                  master_;
  /*
   * The children of the CNode, which by default is a 0-size array.
   */
  protected ImmutableLeaf[]       l_Children_ = new ImmutableLeaf[0];  // External Object Reference
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  static final CNodeImpl EMPTY_INSTANCE = new CNodeImpl();

  // -------- transient properties -------------------------------

   /**
   * manages all the property change listeners of this class and its subclasses
   */
  protected transient PropertyChangeSupport changeSupport_ = null;
  
  /** variable for the JDK 2 serialization
  * @serial
  * ("l_Children_",ImmutableLeaf[].class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
        new ObjectStreamField("l_Children_",ImmutableLeaf[].class)
  };

 /**
  * The basic constructor.
  */
  protected CNodeImpl() {


                          // for testing purpose
                     //System.out.println("Call to CNodeImpl.CNodeImpl()");
  }
 
  /**
  * Used by the master constructor.
  * @param _master is the master  of this embedded class
  * @exception CreateError the master is null
  */
  protected CNodeImpl(Node _master) throws CreateError {
        if (_master == null )
             throw new CreateError("At the CNode construction, "+
                "the master is null.");
        master_ = _master;
  } // constructor
  
  
  /**
  * This method provides a copy of the instance
  * with another name, without children copying
  * because the copy order will not be always correct.
  * @param _masterCopy is the master of the new copy
  * @return copy as CNodeImpl instance
  * @exception CreateError of the constructor with the same parameters.
  **/
  protected CNodeImpl copy(Node _masterCopy) throws CreateError {
      
      CNodeImpl result = new CNodeImpl(_masterCopy);

      // TODO: copy the properties
      
      return result;
  } // copy()

  //
  // -------------- Methods of the implemented interfaces ---------------
  //
  
 /**
 * @return a value derived from the parent nick name hashcode
 */
 public int hashCode() {
   return "CNodeImpl".hashCode() + master_.getNickName().hashCode();
 } // hashCode()

  /**
  * Gets a sorted array of external object references on the children
  *  of this node, which may be leaves or nodes.
  * @return ImmutableLeaf[]. May be null but not empty.
  */
  protected final ImmutableLeaf[] getEorL_Children() {
    // l_Children_ == null only if the CNode has been finalized
    if (l_Children_ == null || l_Children_.length == 0)
        return null;

    ImmutableLeaf[] result = new ImmutableLeaf[l_Children_.length];
    System.arraycopy(l_Children_, 0, result, 0, l_Children_.length);

                // for testing and tracking
                //System.out.println("CNodeImpl.getEorL_Children() member are ImmutableResource: ");
                //for (int i = 0; i < result.length; i++) {
                    //System.out.println("----- "+new Boolean(result[i] instanceof ImmutableResource));
                //} // for i

    return result;
  } // getEorL_Children()
  
  //
  // --------------------- protected CNode methods ----------------------
  //
  
  /**
   * Adds a child to the CNode, without control on the master type and the ACS type policy.
   * Used only by the master, which may add controls on the parameter class.
   * Has to be used by the leaf constructor, so does not update
   * the parent internal variable in the child.
   * Controls the child ACS name. No control on the ACS type policy, since it is done
   * in the ResourceImpl and VirtualFolderImpl constructors which calls the master
   * for adding a new child.
   * Detects the child presence through a name comparison.
   * Fires the event 'Children'.
   * @param _child is a leaf to put in this CNode. No action if null, or already known.
   * @exception InternalError if the leaves list is null. If
   *  firePropertyChange() throws an exception, displays an user message.
   */
  protected void addEorChild(ImmutableLeaf _child) {
    // method variable
    boolean recorded = false;
    ImmutableLeaf[] l_oldChildren;
    ImmutableLeaf[] newArray;

        // for testing and tracking
        //System.out.println("Call to CNodeImpl.addEorChild() on "+getName());

    if (_child == null ) return;
    if (l_Children_ == null)
        throw new InternalError("In CNodeImpl.addEorChild(), the children list is null.");
    if (!_child.getAcsName().equals(master_.getAcsName()) )
        throw new InternalError("In CNodeImpl.addEorChild(), the child has not the correct ACS.");
    
    // tests if _Child is not known
    for (int i = 0 ; i < l_Children_.length ; i++) {
        if (_child.getName().equals(l_Children_[i].getName()) ) {
          recorded = true;
          break;
        }
    } // for
    if (recorded) return;

    l_oldChildren = new ImmutableLeaf[l_Children_.length];
    System.arraycopy(l_Children_, 0, l_oldChildren, 0, l_Children_.length);

    // unknown, so update the array in an empty place if possible
    for (int i = 0 ; i < l_Children_.length ; i++) {
        if (l_Children_[i] == null) {
            l_Children_[i] = _child;
            recorded = true;
            break;
        }
    } // for
        // else, adds at the end of the array
    if (!recorded) {
        newArray = new ImmutableLeaf[l_Children_.length + 1];
        System.arraycopy(l_Children_, 0, newArray, 0, l_Children_.length);
        newArray[l_Children_.length] = _child;
        l_Children_ = newArray;
    }

    try {
      Arrays.sort(l_Children_);
    } catch (ClassCastException c) {
      throw new InternalError("In CNodeImpl.addEorChil(), exception"+
        " on sort operation: "+c.toString());
    } 
    
    try {
      firePropertyChange("Children", l_oldChildren, getEorL_Children());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CNodeImpl.addEorChild(), firePropertChange() throws:");
        e.printStackTrace();
    }

    l_oldChildren = null;
    newArray = null;
  } // addEorChild()

  /**
   * Removes a child from the CNode.
   * Used by the children finalizing methods.
   * Does not update the parent internal variable in the child.
   * Detects the child presence through a name comparison.
   * Fires 'Children' property.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the children list is null, to speed up the
   * finalizing of this instance, when children are called to be finalizing.
   * @param _child is a leaf which is in this CNode. No action if null.
   * @exception UpDateError if _child is not known.
   */
  protected void removeEorChild(ImmutableLeaf _child) throws UpDateError {
    // method variable
    boolean recorded = false;
    ImmutableLeaf[] l_oldChildren;
    short shiftingBack = 0;
    
        // for testing and tracking
        //System.out.println("Call to CNodeImpl.removeEorChild() on "+getName());

    if (_child == null ) return;
    if (l_Children_ == null || l_Children_.length == 0)
        throw new UpDateError("In CNodeImpl.removeEorChild(), _child is unknown.");

    l_oldChildren = new ImmutableLeaf[l_Children_.length];
    System.arraycopy(l_Children_, 0, l_oldChildren, 0, l_Children_.length);

    ImmutableLeaf[]     newArray = new ImmutableLeaf[l_Children_.length];
    ImmutableLeaf[]     result = new ImmutableLeaf[0];
    
    for (int  i = 0 ; i < l_Children_.length ; i++) {
      if (l_Children_[i] != null) {
        if (l_Children_[i].getName().equals(_child.getName()) ) {
          // increments shiftingBack delete each reference
          shiftingBack++;
          recorded = true;
        }
        else newArray[i - shiftingBack] = l_Children_[i];
      }
      else shiftingBack++; // deletes null value
    } // for
    
    if (!recorded)
        throw new UpDateError("In CNodeImpl.removeEorChild(), _child is unknown.");

    if (l_Children_.length != shiftingBack) {
      result = new ImmutableLeaf[l_Children_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
    }
              // else result is a 0-size array
    l_Children_ = result;

    try {
      Arrays.sort(l_Children_);
    } catch (ClassCastException c) {
      throw new InternalError("In CNodeImpl.removeEorChil(), exception"+
        " on sort operation: "+c.toString());
    }    

    try {
      firePropertyChange("Children", l_oldChildren, getEorL_Children());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From CNodeImpl.removeEorChild(), firePropertChange() throws:");
        e.printStackTrace();
    }

    l_oldChildren = null;
    result = null;
    newArray = null;
  } // removeEorChild()

  // ------------------ package methods --------------------------
  
  /**
  * Adds a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to add
  */
  protected void addPropertyChangeListener(PropertyChangeListener _l){

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for all the bound properties:
  *   'Children'.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to add
  */
  protected void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}

              // for testing and tracking
              //System.out.println("in CNodeImpl.addPropertyChangeListener(), "+
              //  "call from "+_l.getClass()+" to "+getName()+" for the property: "
              //  +_propertyName);
  } // addPropertyChangeListener()

  /**
  * Removes a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_l);}
  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for all the bound properties:
  *   'Children'.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

      return changeSupport_.getPropertyChangeListeners(_propertyName);
  } // getPropertyChangeListeners()

  /**
  * Finalizes the CNode and ALL ITS CHILDREN.
  *  HAVE TO BE USED before the deleting of the CNode.
  * @exception UpDateError if the CNode is unknown of the main ACS (if
  *  it is a clone without setName(), typically)
  */
  protected synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call of CNodeImpl.finalizeForUser() for "+getName());

    // noop if it is the second call to this method
    if (l_Children_ == null) {
      return;
    }

    // sets parent to null in all the children and finalizes the children.
    // the  exception is not catched
    ImmutableLeaf[] l_oldChildren = new ImmutableLeaf[l_Children_.length];
    System.arraycopy(l_Children_, 0, l_oldChildren, 0, l_Children_.length);

    // to speed up the call to removeEorChild() from children
    l_Children_ = null;
    // finalizes all children
    for (int  i = 0 ; i < l_oldChildren.length ; i++) {
      if (l_oldChildren[i] != null) {
            l_oldChildren[i].finalizeForBase();
      } // if
    } // for
    l_oldChildren = null;
    
    master_ = null;
  } // finalizeForBase()

  /**
  * This method finalizes all the children, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called for EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

              // for testing purpose
              //System.out.println("call of CNodeImpl.finalizeForUser() for "+getName());

    // noop if it is the second call to this method
    if (l_Children_ == null) {
      return;
    }

    // sets parent to null in all the children and finalizes the children.
    // the  exception is not catched
    ImmutableLeaf[] l_oldChildren = new ImmutableLeaf[l_Children_.length];
    System.arraycopy(l_Children_, 0, l_oldChildren, 0, l_Children_.length);

    // to speed up the call to removeEorChild() from children
    l_Children_ = null;
    // finalizes all children
    for (int  i = 0 ; i < l_oldChildren.length ; i++) {
      if (l_oldChildren[i] != null) {
            if (l_oldChildren[i] instanceof ResourceImpl)
                ((ResourceImpl)l_oldChildren[i]).finalizeFromDeserialization();
            else if (l_oldChildren[i] instanceof VirtualFolderImpl)
                ((VirtualFolderImpl)l_oldChildren[i]).finalizeFromDeserialization();
      } // if
    } // for
    l_oldChildren = null;

    master_ = null;
  } // finalizeFromDeserialization()

   /**
   * Clone for embeddd class without registering the new object in the ACSs.
   * Caution: don't use any modifying method that may altere the integrity
   * of the base.
   * The children list is cloned without deep copy.
   * The external object references are not cloned (see "Eor" in get method).
   * The links between the EOR and the new instance are not set, but the lists
   * of OER are cloned (children) so it is possible to
   * get these active objects.
   * @param _master the clone master
   * @exception CloneNotSupportedException
   * @return cloned object with _master as the master
   */
  protected synchronized CNodeImpl cloneC(Node _master) {
    CNodeImpl clon;
    try {
        clon = (CNodeImpl) super.clone();

        clon.l_Children_ = new ImmutableLeaf[l_Children_.length];
        System.arraycopy(l_Children_, 0, clon.l_Children_, 0, l_Children_.length);

        clon.master_ = _master;
    } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
      if (master_ == null) throw new InternalError("In CNodeImpl, the master is null");
      else {
        ((ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(master_.getAcsName())).
                getEorBaseError().addException(e.toString() + "CNodeImpl.clone() failed");
        throw new InternalError(e.toString());
      }
    } // catch
        
    return clon;
  } // cloneC()
  
  /**
 * @param _obj Object
 * @return true if _obj is a CNodeImpl instance (or an instance of an
 *  extended class) equals to the current one
 */
public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      CNodeImpl obj = (CNodeImpl) _obj;

      if (master_ != obj.master_ )
          return false;
      
      if (l_Children_ == null || obj.getEorL_Children() == null) {
          if (l_Children_ != null || obj.getEorL_Children() != null)
            return false;
      } // if
      else if (l_Children_.length != obj.getEorL_Children().length) // no test: length == 0
        return false;
      else {
          for (int i = 0; i < l_Children_.length; i++) {
            if (l_Children_[i] == null || obj.getEorL_Children()[i] == null) {
                if (l_Children_[i] != null || obj.getEorL_Children()[i] != null)
                  return false;
                continue; // i++
            } // if
            else {
                        // test of equals only on children names
                if (!l_Children_[i].getName().equals( obj.getEorL_Children()[i].getName() ))
                  return false;
            } // else
          } // for
      } // else

      return true;
  } // equals

    /**
     * @return long: serial version UID for serialization of the instance
     */
    public long getSerialVersionUID()	{
            return serialVersionUID;
    }
  
  //
  // -------------- protected methods -------------------------
  //

  /**
  * Fires an event to every registered listener, in any order.
  * For the properties 'Children'.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //System.out.println("CNodeImpl.firePropertyChange(), call to "+
                  //"propertyChange() of "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
   * This method removes the CNode from the ACS lists, but there is no
   * guarantee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {}

    super.finalize();
  }

  
  //
  // ------------------------- private methods -----------------------------
  //


} // CNodeImpl

