//Title:        Access Road
//Version:      0.7.1
//Copyright:    2009-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.BaseError;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Collections;
import java.util.Arrays;
import java.util.TreeMap;
import java.util.Set;
import java.util.HashSet;

import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * This class is an embedded one, responsible for modeling a privilege user.
 * It is used as a private attribute of a composite class instance, called the master,
 * which delegates the processing of its privileges to this class.
 * This class responsability covers both the source and the target privileges.
 * The target inherited privileges may be managed by an AcsAddon, with the support
 * of some methods in this class.
 * The master class may be an EligibleParty, a Resource or a VirtualFolder, and
 * it transfers directly all the actions on privileges to this class.
 * <P>
 * The master may be both a PrivilegeSource and a PrivilegeTarget.
 * If the master is only an EligibleParty, it is only a PrivilegeSource and
 * all the target operations are desactivated. If it is only a PrivilegeTarget
 * (VirtualFolder or Resource), all the source operations are desactivated.
 * EligiblePartyImpl and ResourceImpl are both subclasses of BasicImpl,
 * and this explains why the CPrivilegeUserImpl instance is defined in BasicImpl
 * without use. It is handled in its 2 direct subclasses. It is not possible
 * to create a CPrivilegeUserImpl instance from a BasicImpl instance, because
 * the master must be a PrivilegeSource or a PrivilegeTarget.
 * <P>
 * This class implements the PrivilegeSource and PrivilegeTarget interfaces for the master.
 * This implementation is not declared to avoid to be a public class.
 * The PrivilegeForTypes are managed directly by this class.
 * The PrivilegeForLinks are managed directly or with a target cascading.
 * There is no source cascading. For the target cascading, the inherited rights
 * are managed independantly to the direct rights of the target.
 * The master parent updates dynamically the inherited privileges applied to
 * its children, which are the master of the CPrivilegeUser instances.
 * The methods setEorL_TargetInheritedLinkedPrivileges(), getEorL_TargetInheritedLinkedPrivileges()
 * and resetInheritedPrivilegeRights() are reserved to the master parent.
 * This privilege inheritance depends mainly on each AcsAddon policy, since
 * it is not used in this generic gBase package.
 * The inherited PrivilegeForLinks have not the master as registered target,
 * and they have only one target. The LINKED_IF_ALL_SOURCES privileges are forbidden
 * in the privilege inheritance.
 * <P>
 * It is a javabean with the following bound properties:
 * 'OneSourceTypedPrivilege', 'OneSourceLinkedPrivilege', 'OneTargetTypedPrivilege',
 * 'OneTargetLinkedPrivilege', 'InheritedTargetPrivileges'.
 * The listeners are proper to the master instance.
 * Excepting for EPRViewInBaseImpl and this class, they are outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. The exceptions from the listeners are not catched.
 * This class is public to be overriden in the AcsAddons, and in this case,
 * the subclass is only called by some classes in the same package.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @version 0.7.1
 */
public class CPrivilegeUserImpl implements Serializable {

  // Protected variables
  protected static final int  INITIAL_CAPACITY = 10; // the default value: 101 for map
  protected static final int  INITIAL_CAPACITY_2 = 4 ;  // the default value is 10
  protected static final int  INCREMENT_CAPACITY = 2;
  protected static final int  PRIME = 1000003;

   // Private variables
  private static final long   serialVersionUID = 0;
  /**
   * The master is the instance for which this class operates as an embedded one.
   * It may be a PrivilegeSource, a PrivilegeTarget, or both.
   * To use cascading, it has to be a Resource.
   */
  protected PrivilegeUser       master_;

  /**
   * list of typed privileges for the master as target. isTarget_ has to be true.
   */
  protected ArrayList<PrivilegeForType>     l_T_TypedPrivileges_ = null;

  /**
   * list of linked privileges for the master as target. isTarget_ has to be true.
   */
  protected ArrayList<PrivilegeForLinks>    l_T_LinkedPrivileges_ = null;

  /**
   * TreeMap of inherited linked privilege rights for the master as target.
   * isTarget_ has to be true.
   * The key is the name of the EligibleParty having the rights, with,
   * as last component, the order of the parent which has given the rights.
   * For instance, the direct parent has the string '1' and its direct parent
   * has the string '2' as last name component.
   * A subclass should use the constructor TreeMap(Comparator<? super K> comparator)
   * to set an ordering other than the natural ordering of the keys.
   */
  protected Map<String, StringRight[]>   m_inheritedT_LinkedPRIRights_ = new TreeMap();

  /**
   * List of the current inherited linked privileges for the master as target.
   * isTarget_ has to be true. This is a data source of m_inheritedT_LinkedPRIRights_.
   */
  protected ArrayList<PrivilegeForLinks>        l_inheritedT_LinkedPrivileges_ = null;

  /**
   * list of typed privileges of the master as source. isSource_ has to be true.
   */
  protected ArrayList          l_S_TypedPrivileges_ = null;

  /**
   * list of linked privileges of the master as source. isSource_ has to be true.
   */
  protected ArrayList          l_S_LinkedPrivileges_ = null;

  /**
    * True if the master is a PrivilegeSource.
    */
  protected  boolean             isSource_ = false;
  /**
    * True if the master is a PrivilegeTarget.
    */
  protected boolean             isTarget_ = false;
  
  /* Out of the extended interfaces */
  
   /**
   * manages all the property change listeners and also for its subclasses
   */
  protected transient PropertyChangeSupport changeSupport_ = null;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  static final CPrivilegeUserImpl EMPTY_INSTANCE = new CPrivilegeUserImpl();

  /** 
   * variable for the JDK 2 serialization
   * @serial ("serialVersionUID",Long.TYPE),
   * ("master_",PrivilegeUser.class),
   * ("isSource_",Boolean.TYPE),
   * ("isTarget_",Boolean.TYPE),
   * ("l_T_TypedPrivileges_",ArrayList.class),
   * ("l_T_LinkedPrivileges_",ArrayList.class),
   * "m_inheritedT_LinkedPRIRights_",TreeMap.class),
   * "l_inheritedT_LinkedPrivileges_",ArrayList.class),
   * ("l_SPrivileges_",ArrayList.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("serialVersionUID",Long.TYPE),
      new ObjectStreamField("master_",PrivilegeUser.class),
      new ObjectStreamField("isSource_",Boolean.TYPE),
      new ObjectStreamField("isTarget_",Boolean.TYPE),
      //new ObjectStreamField("aCS_",ACSImpl.class),
      //new ObjectStreamField("parent_",DirectoryImpl.class),
//      new ObjectStreamField("l_Access_",NameImpl[].class),
      new ObjectStreamField("l_T_TypedPrivileges_",ArrayList.class),
      new ObjectStreamField("l_T_LinkedPrivileges_",ArrayList.class),
      new ObjectStreamField("m_inheritedT_LinkedPRIRights_",TreeMap.class),
      new ObjectStreamField("l_inheritedT_LinkedPrivileges_",ArrayList.class),
      new ObjectStreamField("l_SPrivileges_",ArrayList.class)
  };

   /**
   * Default constructor. No registration in the master.
   */
  protected CPrivilegeUserImpl(){

                          // for testing purpose
                     //System.out.println("Call to CPrivilegeUserImpl.CPrivilegeUserImpl()");


      //aCS_Name_= NameImpl.EMPTY_INSTANCE;
  }

   /**
   * Called by the methods addEorPrivilegeForType/LinksToSource/Target
   * in the PrivilegeSources and the PrivilegeTargets.
   * Does not set the linked privilege cascading on the targets.
   * @param _master owns this instance. If not a PrivilegeSource and nor
   * a PrivilegeTarget, fires an InternalError.
   * @exception CreateError if the argument is null.
   * Does not control the ACS privilege management.
   */
  protected CPrivilegeUserImpl(PrivilegeUser _master) throws CreateError {
        if (_master == null)
             throw new CreateError(" This PrivilegeUser is not "+
                "a PrivilegeSource nor a PrivilegeTarget.");
        if (!(_master instanceof PrivilegeSource) && !(_master instanceof PrivilegeTarget))
             throw new InternalError(" This PrivilegeUser is not "+
                "a PrivilegeSource nor a PrivilegeTarget.");

        master_ = _master;
        if (master_ instanceof PrivilegeSource)
            isSource_ = true;
        if (master_ instanceof PrivilegeTarget)
            isTarget_ = true;
        
                          // for testing purpose
                     //System.out.println("Call to CPrivilegeUserImpl.CPrivilegeUserImpl("+
                       // "ImmutableName _name, ACS _acs, Node _parent, String _CPrivilegeUserType)");

  } // CPrivilegeUserImpl constructor

 //
 // -----------  Methods of the implemented interfaces --------------------
 //

  /**
  * This method provides a copy of the instance with another master.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * When _result does not belong to aCS_, does not copy a privilege link
  * if there is no privilege in the result ACS with the same key
  * (the rights may be different).
  * @param _masterCopy is the master of the new copy
  * @return a new instance of CPrivilegeUserImpl
  * @exception CreateError of the constructor with the same parameters.
  * @version 0.5.3
  **/
  protected CPrivilegeUserImpl copy(PrivilegeUser _masterCopy) throws CreateError {
      
      CPrivilegeUserImpl result = new CPrivilegeUserImpl(
        _masterCopy);

      // copies all the properties
      // TODO 
      
      return result;
  } // copy(,,,)
  
 /**
 * The master may be both a PrivilegeSource and a PrivilegeTarget.
 * @return true if the master_ is a privilege source,
 *  and all the relevant methods are activated in this instance
 */
  protected final boolean getIsSource() {
      return isSource_;
  }
  
 /**
 * The master may be both a PrivilegeSource and a PrivilegeTarget.
 * @return true if the master_ is a privilege target
 *  and all the relevant methods are activated in this instance
 */
  protected final boolean getIsTarget() {
      return isTarget_;
  }

   /**
   * Sets a privilege as applicable to this access target.
   * Uses ACS.getM_TypeOrientedPolicy().'Resource OR VirtualFolder.
   * SecondTypesOfPrivilegeFor.'
   * to control if the privilege type is authorized by the ACS.
   * No operation if the privilege is already known.
   * For PrivilegeForTypes, the TYPED and TYPED_CLASSED privilege are allowed as argument,
   * while a TYPED_FOR_SEED or a TYPED_CLASSED_SPECIFIC privilege fires an InternalError.
   * Caution: for a PrivilegeForType, the master constructor calls this method
   * while the privilege is not still registered in the ACS.
   * Fires 'OneTargetTypedPrivilege' or 'OneTargetLinkedPrivilege' events, with
   * names that are similar but different to the Privileges event names.
   * Calls PrivilegeForTypeImpl.fireOneTypedPrivilegeTargetEvent() when
   * the argument is a PrivilegeForType.
   * @param _priv to register in the target
   * @exception BaseError if the master is not a privilege target, or,
   *   if _priv is a typed privilege: <P>
   *   - the privilege second type is not authorized for this target.
   * @return always false if it is not a Directory or not a PrivilegeForLinks,
   *     and if they are, true if children.getParentReadyForRightCascading()
   *     has changed, even if a child may not apply it for changing its rights.
    */
  protected boolean addEorPrivilegeToTarget(Privilege _priv) throws BaseError {

                      // for testing purpose
                 //System.out.println(
                   //      "Call to CPrivilegeUserImpl.addEorPrivilegeToTarget(): "+_priv);

      // controls
      if (!isTarget_) throw new BaseError(
              " "+master_.getNickName()+" is not a privilege target.");
      if (_priv.getMainType().equals(PrivilegeAbst.TYPED_FOR_SEED) ||
              _priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC))
          throw new InternalError(_priv.getNickName()+" has not to be registered in the target.");

      String type = null;
      String inter = null;

      // for PrivilegeForLinks, the type is controlled by PrivilegeForLinks
      if (_priv instanceof PrivilegeForType) {
          type = ((ImmutableTarget)master_).getType();

          if (master_ instanceof ImmutableResource)
              inter = "Resource";
          else if (master_ instanceof ImmutableVirtualFolder)
              inter = "VirtualFolder";

          if (master_.getEorACS().getM_TypeOrientedPolicy() != null &&
                master_.getEorACS().getM_TypeOrientedPolicy().get(
                inter+".SecondTypesOfPrivilegeFor."+type) != null) {

                    if (!Arrays.asList((String[])master_.getEorACS().getM_TypeOrientedPolicy().get(
                            inter+".SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(type))).
                                    contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(
                                                _priv.getSecondType()) )) {

                                  // for testing purpose
                             //System.out.println("Call to CPrivilegeUserImpl.addEorPrivilegeToTarget() "+
                               // "with a reject of the privilege type '"+_priv.getSecondType()+
                                //"' since the allowed values are:");
                             //String[] l_val = (String[]) master_.getEorACS().
                               //      getM_TypeOrientedPolicy().get(
                                 //   inter+".SecondTypesOfPrivilegeFor."+type);
                             //for (int i = 0; i < l_val.length; i++) {
                               //  System.out.println("------'" +l_val[i]+"'");
                             //}

                            throw new BaseError(
                                    " The privilege type "+_priv.getSecondType()+
                                    " is not authorized by the ACS for this object, "+
                                    "through the key "+inter+".SecondTypesOfPrivilegeFor."+type);
                    } // if (!Arrays
          } // if (master_
      } // if

      // action
      if (_priv instanceof PrivilegeForType && l_T_TypedPrivileges_ == null)
          l_T_TypedPrivileges_ = new ArrayList(INITIAL_CAPACITY_2);
      else if (_priv instanceof PrivilegeForLinks && l_T_LinkedPrivileges_ == null)
          l_T_LinkedPrivileges_ = new ArrayList(INITIAL_CAPACITY_2);

      if (_priv instanceof PrivilegeForType) {
          // for the typed privileges, there is no registration in the privilege,
          // but the privilege listeners are informed by a 'OneTypedPrivilegeTarget' event
          ((PrivilegeForTypeImpl)_priv).fireOneTypedPrivilegeTargetEvent
                  (null, (PrivilegeTarget)master_);

          synchronized (l_T_TypedPrivileges_) {
              if (l_T_TypedPrivileges_.contains(_priv)) {

                              // for testing purpose
                         //System.out.println(
                           //      "CPrivilegeUserImpl.addEorPrivilegeToTarget() knows "+_priv);

                  type = null;
                  inter = null;
                  return false;
              }
                              // for testing purpose
                         //System.out.println(
                           //      "CPrivilegeUserImpl.addEorPrivilegeToTarget() adds "+_priv);

              l_T_TypedPrivileges_.add((PrivilegeForType)_priv);
          } // synchro
      }
      else {
          synchronized (l_T_LinkedPrivileges_) {
              if (l_T_LinkedPrivileges_.contains(_priv)) {

                              // for testing purpose
                         //System.out.println(
                           //      "CPrivilegeUserImpl.addEorPrivilegeToTarget() knows "+_priv);

                  type = null;
                  inter = null;
                  return false;
              }
                              // for testing purpose
                         //System.out.println(
                           //      "CPrivilegeUserImpl.addEorPrivilegeToTarget() adds "+_priv);
                         
              l_T_LinkedPrivileges_.add((PrivilegeForLinks)_priv);
          } // synchro
      } // else
      
      if (_priv instanceof PrivilegeForType)
          firePropertyChange("OneTargetTypedPrivilege", null, _priv);
      else
          firePropertyChange("OneTargetLinkedPrivilege", null, _priv);

                      // for testing purpose
                 //System.out.println("End of CPrivilegeUserImpl.addEorPrivilegeToTarget(): "+_priv);

      inter = null;
      type = null;
      return false;      
  } // addEorPrivilegeToTarget()

   /**
   * Unregisters the privilege, and this instance in the linked privilege
   * when it is no more used by this privilege target.
   * Called by the master and by finalizeForBase().
   * Fires 'OneTargetTypedPrivilege' or 'OneTargetLinkedPrivilege' events, with
   * names that are similar but different to the Privileges event names.
   * @return always false if it is not a Directory or not a PrivilegeForLinks,
   *     and if they are, true if child.getParentReadyForRightCascading()
   *     has changed (even if the child does not apply it
   *     for changing its rights).
   * @exception BaseError if the privilege is unknown.
   */
  protected boolean removeEorPrivilegeToTarget(Privilege _priv)
    throws BaseError {
      boolean exception = false;

                            // for testing and tracking
                          //if (_priv instanceof PrivilegeForLinksImpl)
                            //  System.out.println("CPrivilegeUserImpl.removeEorPrivilegeToTarget() call "+
                              // "for "+_priv+
                               //", and l_T_LinkedPrivileges_ contains it: "+
                               //l_T_LinkedPrivileges_.contains(_priv));
      
      if (_priv instanceof PrivilegeForType) {
          synchronized (l_T_TypedPrivileges_) {
            exception = l_T_TypedPrivileges_.remove(_priv);
          } // synchro
      }
      else {
          synchronized (l_T_LinkedPrivileges_) {
            exception = l_T_LinkedPrivileges_.remove(_priv);
          } // synchro          
      } // else
      
      if (!exception) {
          if (_priv instanceof PrivilegeForLinksImpl) {
              System.out.println("CPrivilegeUserImpl.removeEorPrivilegeToTarget(),"+
                      " l_T_LinkedPrivileges_:");
              for (Object pri : l_T_LinkedPrivileges_) {
                System.out.println("--- String form: "+pri);
              }
          } // if

          throw new BaseError(" Database error! "+
            "The privilege "+_priv+" is unknown in the target. "+
            "Please restart the base from the last saving.");
      } // if
      
      if (_priv instanceof PrivilegeForLinksImpl &&
        ((PrivilegeForLinks)_priv).getEorLinkedTargets().contains(master_)) {
            // throws 'OneLinkedTarget' event
            // test l_T_LinkedPrivileges_ and do not call back this method
            ((PrivilegeForLinksImpl)_priv).removeEorPrivilegeTargetLink(
                (PrivilegeTarget)master_);
      } // if    

      if (_priv instanceof PrivilegeForType) {
          firePropertyChange("OneTargetTypedPrivilege", _priv, null);
          ((PrivilegeForTypeImpl)_priv).fireOneTypedPrivilegeTargetEvent
                  ((PrivilegeTarget)master_, null);
      } 
      else {
          firePropertyChange("OneTargetLinkedPrivilege", _priv, null);
      }
      return false;        
  } // removeEorPrivilegeToTarget()

   /**
   * Gets all the privileges linked to this target.
   * There are two types of privilege links (activated or not by the ACS):
   * - PrivilegeForLinks: linked privilege established for a couple
   *      (ACSObject, privilege)
   * - PrivilegeForType: typed privilege, where the ACSObject and the privilege
   *      has a common type (and a common class if it is a bindedClasses
   *      privilege)
   * Returns an immutable empty list if getIsTarget() is false.
   * @return List of Privileges. May be Collections.EMPTY_LIST, but never null.
   */
  protected List<Privilege> getEorL_TargetPrivileges() {
      if (l_T_TypedPrivileges_ == null && l_T_LinkedPrivileges_ == null)
          return Collections.EMPTY_LIST;
      else if (l_T_TypedPrivileges_ != null && l_T_LinkedPrivileges_ == null)
          return (List)l_T_TypedPrivileges_.clone();
      else if (l_T_TypedPrivileges_ == null && l_T_LinkedPrivileges_ != null)
          return (List)l_T_LinkedPrivileges_.clone();
      else {
          List result = (List)l_T_TypedPrivileges_.clone();
          result.addAll(l_T_LinkedPrivileges_);
          return result;
      }
  } // getEorL_TargetPrivileges()

   /**
   * Gets all the privilegeForLinks linked to this target, not the
   * privilegeForTypes. Return both direct and cascading privileges,
   * if the cascading is activated.
   * Returns am empty list if getIsTarget() is false.
   * @return List of PrivilegeForLinks. May be empty, but never null.
   */
  protected List<PrivilegeForLinks> getEorL_TargetPrivilegeForLinks() {
      if (l_T_LinkedPrivileges_ == null) return Collections.EMPTY_LIST;
      else return (List)l_T_LinkedPrivileges_.clone();
  } // getEorPrivilegeForLinks()

   /**
   * Gets all the privilegeForTypes linked to this target, not the
   * privilegeForLinks. May include the privileges from the parents,
   * if cascading is activated.
   * Return an immutable empty list if getIsTarget() is false.
   * @return List of PrivilegeForTypes. May be Collections.EMPTY_LIST, but never null.
   */
  protected List<PrivilegeForType> getEorL_TargetPrivilegeForTypes() {
      if (l_T_TypedPrivileges_ == null) return Collections.EMPTY_LIST;
      else return (List)l_T_TypedPrivileges_.clone();
  } // getEorPrivilegeForTypes()

   /**
   * Sets a privilege as applicable to this access source.
   * Uses ACS.getM_TypeOrientedPolicy().'EligibleParty.SecondTypesOfPrivilegeFor.'
   * to control if the privilege type is authorized by the ACS.
   * No operation if the privilege is already known.
   * For PrivilegeForTypes, the TYPED and TYPED_CLASSED privilege are allowed as argument,
   * while a TYPED_FOR_SEED or a TYPED_CLASSED_SPECIFIC privilege fires an InternalError.
   * Caution: for PrivilegeForTypes, the constructor calls this method,
   * and the privilege is not still registered in the ACS.
   * Fires 'OneSourceTypedPrivilege' and 'OneSourceLinkedPrivilege' events.
   * @param _priv to register in the source
   * @exception BaseError if the master is not a privilege source, or,
   *   if _priv is a typed privilege: <P>
   *   - the privilege second type is not authorized for this source.
   */
  protected void addEorPrivilegeToSource(Privilege _priv) throws BaseError {

                      // for testing purpose
                 //System.out.println("Call to CPrivilegeUserImpl.addEorPrivilegeToSource() ");

      // controls
      if (!isSource_)
          throw new BaseError(" "+master_.getNickName()+" is not a privilege source.");
      if (_priv.getMainType().equals(PrivilegeAbst.TYPED_FOR_SEED) ||
              _priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC))
          throw new InternalError(_priv.getNickName()+" has not to be registered in the source.");

      // the master has rejected the call to this method if it is immutable
      String type = ((ImmutableSource)master_).getType();

      // for PrivilegeForLinks, the type is controlled by PrivilegeForLinks
      if (_priv instanceof PrivilegeForType) {
          if (master_.getEorACS().getM_TypeOrientedPolicy() != null &&
              master_.getEorACS().getM_TypeOrientedPolicy().get(
              "EligibleParty.SecondTypesOfPrivilegeFor."+type) != null) {

                    if (!Arrays.asList((String[])master_.getEorACS().getM_TypeOrientedPolicy().get(
                            "EligibleParty.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(type))).
                                    contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(_priv.getSecondType()) )) {

                                              // for testing purpose
                                         //System.out.println("Call to CPrivilegeUserImpl.addEorPrivilegeToSource() "+
                                           // "with a reject of the privilege type '"+_priv.getSecondType()+
                                           // "' since the allowed values are:");
                                         //String[] l_val = (String[]) master_.getEorACS().
                                           //      getM_TypeOrientedPolicy().get(
                                             //   "EligibleParty.SecondTypesOfPrivilegeFor."+type);
                                         //for (int i = 0; i < l_val.length; i++) {
                                           //  System.out.println("------'" +l_val[i]+"'");
                                         //} // for i

                            throw new BaseError(
                                    " The privilege type "+_priv.getSecondType()+
                                    " is not authorized by the ACS for this object, "+
                                    "through the key EligibleParty.SecondTypesOfPrivilegeFor."+type);
                    } // if (!Arrays.asList((String[])master_.getEorACS().getM_TypeOrientedPolicy()
          } // if (master_.getEorACS().getM_TypeOrientedPolicy() != null
      } // if (_priv instanceof PrivilegeForType)

      // action
      if (_priv instanceof PrivilegeForType && l_S_TypedPrivileges_ == null)
          l_S_TypedPrivileges_ = new ArrayList(INITIAL_CAPACITY_2);
      else if (_priv instanceof PrivilegeForLinks && l_S_LinkedPrivileges_ == null)
          l_S_LinkedPrivileges_ = new ArrayList(INITIAL_CAPACITY_2);

      // there is no registration in the privilege
      if (_priv instanceof PrivilegeForType) {
          // synchro on a non-final field to limit the memory print of the CPrivilegeUserImpl instance
          synchronized (l_S_TypedPrivileges_) {
              if (l_S_TypedPrivileges_.contains(_priv)) {
                  type = null;
                  return;
              }
              l_S_TypedPrivileges_.add(_priv);
          } // synchro

                      // for testing purpose
                 //System.out.println("CPrivilegeUserImpl.addEorPrivilegeToSource() adds "+
                   //      _priv+" for "+master_.getNickName());

      }
      else {
          synchronized (l_S_LinkedPrivileges_) {
              if (l_S_LinkedPrivileges_.contains(_priv)) {
                  type = null;
                  return;
              }
            l_S_LinkedPrivileges_.add(_priv);
          } // synchro
      } // else

      if (_priv instanceof PrivilegeForType)
          firePropertyChange("OneSourceTypedPrivilege", null, _priv);
      else
          firePropertyChange("OneSourceLinkedPrivilege", null, _priv);

      type = null;
  } // addEorPrivilegeToSource()

   /**
   * Unregisters the privilege, and the master in the linked privilege
   * when it is no more used by this privilege source.
   * Called by finalizeForBase().
   * Fires 'OneSourceTypedPrivilege' and 'OneSourceLinkedPrivilege' events.
   * @exception BaseError if the privilege is unknown.
   */
  protected void removeEorPrivilegeToSource(Privilege _priv) throws BaseError {
      boolean exception = false;

                            // for testing and tracking
                          //if (_priv instanceof PrivilegeForLinksImpl)
                            //  System.out.println("CPrivilegeUserImpl.removeEorPrivilegeToSource() call "+
                              // "for "+_priv+
                               //", and l_S_LinkedPrivileges_ contains it: "+
                               //l_S_LinkedPrivileges_.contains(_priv));


      if (_priv instanceof PrivilegeForType) {
          synchronized (l_S_TypedPrivileges_) {
            exception = l_S_TypedPrivileges_.remove(_priv);
          } // synchro
      }
      else {
          synchronized (l_S_LinkedPrivileges_) {
            exception = l_S_LinkedPrivileges_.remove(_priv);
          } // synchro
      } // else
      
      if (!exception) {
          if (_priv instanceof PrivilegeForLinksImpl) {
              System.out.println("CPrivilegeUserImpl.removeEorPrivilegeToSource(),"+
                      " l_S_LinkedPrivileges_:");
              for (Object pri : l_S_LinkedPrivileges_) {
                System.out.println("--- String form: "+pri);
              }
          } // if

          throw new BaseError(" Database error! "+
            "Unknown privilege "+_priv+" by the source "+master_.getName()+
            ". Restart the base from the last saving!");
      } // if
      
      if (_priv instanceof PrivilegeForLinksImpl &&
        ((PrivilegeForLinks)_priv).getEorLinkedSources().contains(master_) ) {


                            // for testing and tracking
                          //System.out.println("CPrivilegeUserImpl.removeEorPrivilegeToSource() call "+
                            //   "PrivilegeForLinksImpl.removeEorPrivilegeSourceLink() ");

          // throws 'OneLinkedSource' event
          // test l_S_LinkedPrivileges_ and do not call back this method
          ((PrivilegeForLinksImpl)_priv).removeEorPrivilegeSourceLink(
            (PrivilegeSource)master_);
      } // if     

      if (_priv instanceof PrivilegeForType)
          firePropertyChange("OneSourceTypedPrivilege", _priv, null);
      else
          firePropertyChange("OneSourceLinkedPrivilege", _priv, null);
  } // removeEorPrivilegeToSource()

   /**
   * Gets all the privileges linked to this source.
   * There is two types of privilege links (allowed or not by the ACS):
   * - PrivilegeForLinks: linked privilege established for a couple
   *      (ACSObject, privilege)
   * - PrivilegeForType: typed privilege, where the ACSObject and the privilege
   *      has a common type (and a common class if it is a bindedClasses
   *      privilege)
   * Returns an immutable empty list if getIsSource() is false.
   * @return List of Privileges. May be Collections.EMPTY_LIST, but never null.
   */
  protected List getEorL_SourcePrivileges() {
      if (l_S_TypedPrivileges_ == null && l_S_LinkedPrivileges_ == null)
          return Collections.EMPTY_LIST;
      else if (l_S_TypedPrivileges_ != null && l_S_LinkedPrivileges_ == null)
          return (List)l_S_TypedPrivileges_.clone();
      else if (l_S_TypedPrivileges_ == null && l_S_LinkedPrivileges_ != null)
          return (List)l_S_LinkedPrivileges_.clone();
      else {
          List result = (List)l_S_TypedPrivileges_.clone();
          result.addAll(l_S_LinkedPrivileges_);
          return result;
      }
  } // getEorL_SourcePrivileges()
  
   /**
   * Gest all the PrivilegeForTypes linked to this source.
   * Returns an empty list if getIsSource() is false.
   * @return List of PrivilegeForType. May be Collections.EMPTY_LIST, but never null.
   */
  protected List getEorL_SourcePrivilegeForTypes() {
      if (l_S_TypedPrivileges_ == null) return Collections.EMPTY_LIST;
      else {
          PrivilegeForType tpri = null;
          for (Iterator ite = l_S_TypedPrivileges_.iterator(); ite.hasNext(); ) {
              tpri = (PrivilegeForType)ite.next();
              if (tpri.getMainType() == null) {
                  ite.remove();

                                  // for testing and tracking
                                  System.out.println("CPrivilegeUserImpl."+
                                          "getEorL_SourcePrivilegeForTypes() on "+
                                          master_.getNickName()+" removes a null TPRI: "+
                                          tpri);
              }
          } // for ite


          return (List)l_S_TypedPrivileges_.clone();
      }
  }
  
   /**
   * Gets all the PrivilegeForLinks linked to this source.
   * Note: for LINKED_IF_ALL linked privileges, the presence of the privilege
   * does not mean that the master has the privilege rights.
   * Returns an immutable empty list if getIsSource() is false.
   * @return List of PrivilegeForLinks. May be Collections.EMPTY_LIST, but never null.
   */
  protected List getEorL_SourcePrivilegeForLinks() {
      if (l_S_LinkedPrivileges_ == null) return Collections.EMPTY_LIST;
      else return (List)l_S_LinkedPrivileges_.clone();
  }

  /**
   * Gets the Privilege rights on the target master.
   * The linked privileges may be inherited from the parent. 
   * @return Map of access rights in the main ACS,
   *  where the key is the getSouce() String for a privilege,
   *  and the value is a StringRight array.
   *  May be empty, never null.
   */
  //Map<String, StringRight[]> getM_PRIRightsAsTarget() {
    //  Map<String, StringRight[]> result = new TreeMap();
    //  if (l_TPrivileges_ == null || !isTarget_) return result;
    //  Privilege priv = null;
      
    //  for (Iterator ite = l_TPrivileges_.iterator(); ite.hasNext();) {
      //    priv = (Privilege) ite.next();

        //  result.put(priv.getSource(), priv.getL_rights());
      //} // for
      //priv = null;
      //return result;
  //} // getM_PRIRightsAsTarget()

  /**
   * Gets the Privilege rights on the source master.
   * @return Map of access rights in the main ACS and other ACS,
   *  where the map key is an item in getTargets() for a privilege,
   *  and the map value is a StringRight array.
   *  May be empty, never null.
   */
  //Map<String, StringRight[]> getM_PRIRightsAsSource() {
    //  Map result = new TreeMap();
    //  if (l_SPrivileges_ == null || !isSource_) return result;
    //  Privilege priv = null;
    //  String[] l_targets = null;
      
    //  for (Iterator ite = l_SPrivileges_.iterator(); ite.hasNext();) {
      //    priv = (Privilege) ite.next();
      //    l_targets = priv.getTargets();
      //    if (l_targets == null) continue;

      //    for (int i = 0; i < l_targets.length; i++) {
        //    result.put(l_targets[i], priv.getL_rights());
         // } // for i
     // } // for ite
     // priv = null;
     // l_targets = null;
     // return result;
  //} // getM_PRIRightsAsSource()

   /**
   * Gets all the immutable list of the inherited linked privileges of the target.
   * The LINKED_IF_ALL_SOURCE type is forbidden, and any privilege having more than one target.
   * To use by the consumers in the gWork packages, which deny any access if
   * the returned value is null when the privilege inheritance is managed.
   * It is recommanded to avoid getEorL_TargetInheritedLinkedPrivileges() for these consumers.
   * Null is the default and immutable returned value in the generic gBase class,
   * and this has to be changed in an AcsAddon implementing the inheritance.
   * The privilege may be a clone, and the target may not be the master.
   * Caution: a clone is not registered in its ACS, nor in its sources and targets.
   * However, it is possible in the beamer to navigate from a cloned privilege
   * to its targets or sources.
   * @return immutable list of Linked Privileges. May be empty but never null.
   */
  protected List<PrivilegeForLinks> getEorImmutableL_TargetInheritedLinkedPrivileges() {
      if (l_inheritedT_LinkedPrivileges_ == null ||
              l_inheritedT_LinkedPrivileges_.isEmpty()) return Collections.EMPTY_LIST;
      else return Collections.unmodifiableList(l_inheritedT_LinkedPrivileges_);
  } // getEorL_TargetInheritedLinkedPrivileges()

   /**
   * Gets all the inherited linked privileges of this target.
   * The LINKED_IF_ALL_SOURCE type is forbidden, and any privilege having more than one target.
   * To use by the parent class in the gBase package, to prepare the updating of the list
   * with a call to setEorL_TargetInheritedLinkedPrivileges().
   * No use in this generic gBase package, so for the AcsAddon packages.
   * @return mutable list of Linked Privileges. May be a new empty list or an old
   * empty list, but never null.
   */
  protected List<PrivilegeForLinks> getEorL_TargetInheritedLinkedPrivileges() {
      if (l_inheritedT_LinkedPrivileges_ == null) return new ArrayList(INITIAL_CAPACITY_2);
      else return l_inheritedT_LinkedPrivileges_;
  } // getEorL_TargetInheritedLinkedPrivileges()

   /**
   * Sets all the inherited linked privileges for this target where the
   * LINKED_IF_ALL_SOURCE type is forbidden, and any  privilege having more than one target.
   * Called only by the master parents.
   * Do not register the master in the inherited linked privileges.
   * Calls setM_TargetInheritedLinkedPRIRights() to set the map (source name, StringRight[]).
   * No use in this generic gBase package, so for the AcsAddon packages.
   * Fires 'InheritedTargetPrivileges' events.
   * @param _l_priv list of one-target Linked Privileges. May be empty, but never null.
   * @throws InternalError if a privilege has the LINKED_IF_ALL_SOURCE type, or
   * if it has not exactly one target.
   */
  protected void setEorL_TargetInheritedLinkedPrivileges(ArrayList<PrivilegeForLinks> _l_priv) {
      if (_l_priv == null) throw new InternalError("CPrivilegeUserImpl."+
              "setEorL_TargetInheritedLinkedPrivileges() with a null argument.");
      if (!isTarget_) throw new InternalError("CPrivilegeUserImpl."+
              "setEorL_TargetInheritedLinkedPrivileges() for a non-target master.");
      
      ArrayList l_oldList = null;
      boolean   error = false;
      boolean   error2 = false;

      if (l_inheritedT_LinkedPrivileges_ == null) {
          l_oldList = new ArrayList(INITIAL_CAPACITY_2);
      } else {
          l_oldList = (ArrayList) l_inheritedT_LinkedPrivileges_.clone();
      } // else

      for (Object obj : _l_priv) {
          if (((PrivilegeForLinks)obj).getMainType().equals(PrivilegeAbst.LINKED_IF_ALL_SOURCES)) {
              error = true;
              break;
          } // if
          if (((PrivilegeForLinks)obj).getEorLinkedTargets().size() != 1) {
              error2 = true;
              break;
          }
      } // for obj

      if (error)
          throw new InternalError("CPrivilegeUserImpl."+
              "setEorL_TargetInheritedLinkedPrivileges() with a LINKED_IF_ALL_SOURCES privilege.");

      if (error2)
          throw new InternalError("CPrivilegeUserImpl."+
              "setEorL_TargetInheritedLinkedPrivileges() with a privilege having"+
              " not exactly one target.");

      l_inheritedT_LinkedPrivileges_ = _l_priv;

      setM_TargetInheritedLinkedPRIRights(_l_priv);
      
      firePropertyChange(
              "InheritedTargetPrivileges", l_oldList, l_inheritedT_LinkedPrivileges_.clone());
      
      l_oldList = null;
  } // setEorL_TargetInheritedLinkedPrivileges()

    /**
    * Resets the inherited Privilege rights after an user edition.
    * Called by the master parent when there is a change in the effective rights of
    * an inherited Privilege. This is only when there is a change in the sources.
    * Contrary to AclEntry, the rights are immutable and there is no conditional group.
    * Reads l_inheritedT_LinkedPrivileges_ and updates m_inheritedT_LinkedPRIRights_
    * only for the relevant keys, that is one per source.
    * Note: the creation of the inherited Privilege in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited Privilege
    * is handled.
    * @param _lpri inherited Privilege with updated rights. Never null.
    */
  protected synchronized void resetInheritedPrivilegeRights(PrivilegeForLinks _lpri) {

                                  // for testing and tracking
                                  //System.out.println("CPrivilegeUserImpl."+
                                    //      "resetInheritedPrivilegeRights() "+
                                      //    "on "+master_.getName().lastComponent()+
                                        //  " for the ACL "+_lpri);

      if (l_inheritedT_LinkedPrivileges_ == null) {
          InternalError i = new InternalError("CPrivilegeUserImpl."+
              "resetInheritedPrivilegeRights(), the target "+
              master_.getName().lastComponent()+
              " has no Privielges.");
          i.printStackTrace();
          throw i;
      } // if

      PrivilegeForLinks pri = null;
      boolean           found = false;
      ImmutableName     targetName = _lpri.getEorLinkedTargets().get(0).getName();
      StringRight[]     l_rights = null;
      Set<StringRight>  l_newRights = null;
      String            order = UtilityImpl.getFirstInstance().getParentOrder(
                                  (PrivilegeTarget) master_,
                                  (ImmutableNode) UtilityImpl.getFirstInstance().getBaseObject(
                                        targetName));
      String            key = null;

      for (PrivilegeSource source : _lpri.getEorLinkedSources()) {
          key = ((Name)source.getName()).extended(order).toString();

                                    // for testing and tracking
                                    //if (master_.getName().lastComponent().equals("Products table") &&
                                      //  m_inheritedT_LinkedPRIRights_.get(key) != null) {

                                        //    StringRight[] rights = m_inheritedT_LinkedPRIRights_.get(key);
                                          //  for (int i = 0; i < rights.length; i++) {
                                            //    System.out.println("--- old right "+rights[i]);
                                           // } // for ite

                                    //} // if (name_
                                    //else if (master_.getName().lastComponent().equals("Customers table"))
                                      //  System.out.println("CPrivilegeUserImpl."+
                                        //      "resetInheritedPrivilegeRights() "+
                                          //    "for the resource "+
                                            //  master_.getName().lastComponent()+
                                              //" without old rights for the key "+key);


          // clears the map to update
          StringRight[]     l_oldRights = m_inheritedT_LinkedPRIRights_.remove(key);

                                      // for testing and tracking
                                      //System.out.println("CPrivilegeUserImpl."+
                                        //      "resetInheritedPrivilegeRights() "+
                                          //    "for the keyName "+key);


          // reset the couple (source name + order, StringRight[])
          for (Iterator ite = l_inheritedT_LinkedPrivileges_.iterator(); ite.hasNext();) {
            pri = (PrivilegeForLinks) ite.next();


                              // for testing and tracking
                              //System.out.println("--- iteration on inherited LPRI "+pri);

            // review all the privileges with the _lpri keyName
            if (!pri.getEorLinkedTargets().get(0).equals(_lpri.getEorLinkedTargets().get(0)) ||
                    !source.equals(_lpri.getEorLinkedSources().get(0))) {
                continue;
            }

            if (pri == _lpri) found = true;

            // pri is _lpri in one of the iterations
            l_rights = pri.getL_Rights();

            if (l_rights.length == 0) {

                              // for testing and tracking
                              //System.out.println("--- empty rights for "+pri);

                continue;
            }

                              // for testing and tracking
                              //System.out.println("--- non-empty rights to add for "+pri);

            if (m_inheritedT_LinkedPRIRights_.get(key) == null) {

                m_inheritedT_LinkedPRIRights_.put(key, l_rights);

            } // if
            else {
                l_newRights = new HashSet(
                        Arrays.asList(m_inheritedT_LinkedPRIRights_.get(key)));
                l_newRights.addAll(Arrays.asList(l_rights));

                m_inheritedT_LinkedPRIRights_.put(
                        key,
                        l_newRights.toArray(new StringRight[0]));
            } // else
          } // for ite

          if (!found) {
              InternalError i = new InternalError("CPrivilegeUserImpl."+
                  "resetInheritedAclEntryRights(), the target "+
                  master_.getName().lastComponent()+
                  " does not know the inherited privilege "+_lpri);
              i.printStackTrace();
              throw i;
          } // if (!found)

      } // for source

                                // for testing and tracking
                                //if (master_.getName().lastComponent().equals("Products table") &&
                                  //  m_inheritedT_LinkedPRIRights_.get(key) != null) {

                                    //    System.out.println("-- New rights for "+key);
                                      //  StringRight[] rights = m_inheritedT_LinkedPRIRights_.get(key);
                                        //for (int i = 0; i < rights.length; i++) {
                                          //  System.out.println("----- new right "+rights[i]);
                                        //} // for ite

                                //} // if (name_

                                  // for testing and tracking
                                  //System.out.println("end of CPrivilegeUserImpl."+
                                    //      "resetInheritedPrivilegeRights() ");

    pri = null;
    targetName = null;
    l_rights = null;
    l_newRights = null;
    order = null;
    key = null;
  } // resetInheritedPrivilegeRights()

  /**
   * Sets the internal map of the inherited linked privilege rights, where
   * the key is the name of the EligibleParty having the rights, extended with,
   * as last component, the parent order of the master parent which has given the rights.
   * For instance, the direct parent has the string '1', and its direct parent
   * has the string '2' as parent order in the last name component, and so on.
   * On the privilege, calls the method getL_rights(), getEorLinkedSources() and
   * getEorLinkedTargets(). The privilege may be a clone.
   * Caution: a clone is not registered in its ACS, nor in its sources and targets.
   * Called by setEorL_TargetInheritedLinkedPrivileges() only, and do not fire any event.
   * No use in this generic gBase package, so for the AcsAddon packages.
   * @param _l_priv argument of setEorL_TargetInheritedLinkedPrivileges().
   * May be empty, but never null.
   */
  private void setM_TargetInheritedLinkedPRIRights(List<PrivilegeForLinks> _l_priv) {
      PrivilegeForLinks priv = null;
      StringRight[]     l_rights = null;
      Set<StringRight>  l_newRights = null;
      String            order = null;
      boolean           error = false;
      String            key = null;
      
      if (m_inheritedT_LinkedPRIRights_ == null)
          m_inheritedT_LinkedPRIRights_ = new TreeMap();
      else
          m_inheritedT_LinkedPRIRights_.clear();
      
      // add the couple (source name + order, StringRight[])
      if (!_l_priv.isEmpty()) {
          try {
              for (Iterator ite = _l_priv.iterator(); ite.hasNext();) {
                  priv = (PrivilegeForLinks) ite.next();
                  l_rights = priv.getL_Rights();
                  //if (l_rights.length == 0)
                    //  continue;

                  order = UtilityImpl.getFirstInstance().getParentOrder(
                          (ImmutablePrivilegeTarget)master_,
                          (ImmutableNode)priv.getEorLinkedTargets().get(0));
                  if (Integer.valueOf(order) < 0) {
                      error = true;
                      break;
                  }

                  for (Object obj : priv.getEorLinkedSources()) {
                    key = ((Name)((ACSObject)obj).getName()).extended(order).toString();

                    if (m_inheritedT_LinkedPRIRights_.get(key) == null) {
                        
                        m_inheritedT_LinkedPRIRights_.put(key, l_rights);

                    } // if
                    else if (l_rights.length != 0) {
                        l_newRights = new HashSet(
                                Arrays.asList(m_inheritedT_LinkedPRIRights_.get(key)));
                        l_newRights.addAll(Arrays.asList(l_rights));

                        m_inheritedT_LinkedPRIRights_.put(
                                key,
                                l_newRights.toArray(new StringRight[0]));
                    } // else
                  } // for obj

              } // for ite
          } catch (NumberFormatException n) {
              throw new InternalError("in CPrivilegeUserImpl."+
                      "setM_TargetInheritedLinkedPRIRights(), "+
                        "the exception for "+order+" is: "+n.toString());
          }
      } // if

      if (error)
          throw new InternalError("CPrivilegeUserImpl."+
              "setM_TargetInheritedLinkedPRIRights(), the target of a privilege "+
              "is not the parent of the master.");
      
      priv = null;
      l_rights = null;
      l_newRights = null;
      order = null;
      key = null;
  } // setM_TargetInheritedLinkedPRIRights()
   
  /**
   * Gets the internal map of the inherited linked privilege rights, where
   * the key is the name of the EligibleParty having the rights on the master, extended with,
   * as last component, the parent order of the master parent which has given the rights.
   * For instance, the direct parent has the string '1', and its direct parent
   * has the string '2' as parent order in the last name component, and so on.
   * The map value is a non-null StringRight array of inherited rights, and it may be empty.
   * @return unmodifiable map of inherited rights from the linked privileges.
   * May be empty but never null.
   */
  protected Map<String, StringRight[]> getM_TargetInheritedLinkedPRIRights() {
      if (m_inheritedT_LinkedPRIRights_ == null) return Collections.EMPTY_MAP;
      else return Collections.unmodifiableMap(m_inheritedT_LinkedPRIRights_);
  }

  /**
  * Adds a listener to the bean, for all the bound properties.
  * Used by the master only.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  protected void addPropertyChangeListener(PropertyChangeListener _l){
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by the master for the properties:
  * 'OneSourceTypedPrivilege', 'OneSourceLinkedPrivilege',
  * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege',
  * 'InheritedTargetPrivileges'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  protected void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(master_);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}
  } // addPropertyChangeListener(String _propertyName,

  /**
  * Removes a listener to the bean, for all the bound properties.
  * Used by the master.
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("CPrivilegeUserImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_l);}
  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by the master for the properties:
  * 'OneSourceTypedPrivilege', 'OneSourceLinkedPrivilege',
  * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege',
  * 'InheritedTargetPrivileges'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  protected void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("CPrivilegeUserImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

      return changeSupport_.getPropertyChangeListeners(_propertyName);
  } // getPropertyChangeListeners()

 /**
  * Clone for embedded class, without registering the new object in the ACSs,
  * and without links to the privileges of this instance.
  * The privilege list is cloned without deep copy.
   * The external object references are not cloned (see "Eor" in get method).
   * The links between the EOR and the new instance are not set, but the lists
   * of OER are cloned (privileges) so it is possible to get
   * these active objects. Used by the master clone() method.
   * @param _master must be a clone of this instance master
   * @return instance with _master as master.
   */
  protected synchronized CPrivilegeUserImpl cloneC(PrivilegeUser _master) {
    CPrivilegeUserImpl clon;
    try {
      clon = (CPrivilegeUserImpl) super.clone();

      clon.isSource_ = isSource_;
      clon.isTarget_ = isTarget_;
      
      clon.master_ = _master;

      if (l_T_TypedPrivileges_ != null)
          clon.l_T_TypedPrivileges_ = (ArrayList)l_T_TypedPrivileges_.clone();
      else clon.l_T_TypedPrivileges_ = null;

      if (l_T_LinkedPrivileges_ != null)
          clon.l_T_LinkedPrivileges_ = (ArrayList)l_T_LinkedPrivileges_.clone();
      else clon.l_T_LinkedPrivileges_ = null;

      if (m_inheritedT_LinkedPRIRights_ != null)
          clon.m_inheritedT_LinkedPRIRights_ = (Map)((TreeMap)m_inheritedT_LinkedPRIRights_).clone();
      else clon.m_inheritedT_LinkedPRIRights_ = null;

      if (l_inheritedT_LinkedPrivileges_ != null)
          clon.l_inheritedT_LinkedPrivileges_ = (ArrayList)l_inheritedT_LinkedPrivileges_.clone();
      else clon.l_inheritedT_LinkedPrivileges_ = null;

      if (l_S_TypedPrivileges_ != null)
          clon.l_S_TypedPrivileges_ = (ArrayList)l_S_TypedPrivileges_.clone();
      else clon.l_S_TypedPrivileges_ = null;

      if (l_S_LinkedPrivileges_ != null)
          clon.l_S_LinkedPrivileges_ = (ArrayList)l_S_LinkedPrivileges_.clone();
      else clon.l_S_LinkedPrivileges_ = null;
      
    } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
      if (master_ == null) throw new InternalError("In CPrivilegeUserImpl, the master is null");
      else {
        master_.getEorACS().getEorBaseError().addException(e.toString() + 
          "CPrivilegeUserImpl.clone() failed");
        throw new InternalError(e.toString());
      }
    } // catch
    
    return clon;
  } // cloneC()

  /**
   * @return the hashcode derived from the master nick name, the first typed and
   * the first linked privileges
   */
  public int hashCode() {
    int result = 0;

    if (master_ != null) {
      result = "CPrivilegeUserImpl".hashCode() + master_.getNickName().hashCode();
    }

    if (l_T_TypedPrivileges_ != null && !l_T_TypedPrivileges_.isEmpty()) {
      result = PRIME*result + l_T_TypedPrivileges_.get(0).hashCode();
    }

    if (l_S_LinkedPrivileges_ != null && !l_S_LinkedPrivileges_.isEmpty()) {
      result = PRIME*result + l_S_LinkedPrivileges_.get(0).hashCode();
    }
    
    return result;
  } // hashcode()

  /**
  * Works even if the instance is user-finalized.
  * @param _obj
  * @return true if _obj is an instance of CPrivilegeUserImpl (or a derived class)
  *   equals to this instance (with the same class)
  */
  public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      CPrivilegeUserImpl obj = (CPrivilegeUserImpl) _obj;

      if (master_ != obj.master_ )
          return false;
      /*
      if (isSource_ != obj.getIsSource() )
          return false;
      
      if (isTarget_ != obj.getIsTarget() )
          return false;     

      if (l_T_TypedPrivileges_ == null || obj.getEorL_TargetPrivilegeForTypes().equals(Collections.EMPTY_LIST)) {
          if (l_T_TypedPrivileges_ != null || !obj.getEorL_TargetPrivilegeForTypes().equals(Collections.EMPTY_LIST))
            return false;
      } // if
      else {
          if (!l_T_TypedPrivileges_.equals( obj.getEorL_TargetPrivilegeForTypes() ))
            return false;
      } // else

      if (l_T_LinkedPrivileges_ == null || obj.getEorL_TargetPrivilegeForLinks().equals(Collections.EMPTY_LIST)) {
          if (l_T_LinkedPrivileges_ != null || !obj.getEorL_TargetPrivilegeForLinks().equals(Collections.EMPTY_LIST))
            return false;
      } // if
      else {
          if (!l_T_LinkedPrivileges_.equals( obj.getEorL_TargetPrivilegeForLinks() ))
            return false;
      } // else

      if (m_inheritedT_LinkedPRIRights_ == null || obj.getEorL_TargetInheritedLinkedPrivileges().equals(Collections.EMPTY_LIST)) {
          if (m_inheritedT_LinkedPRIRights_ != null || !obj.getEorL_TargetInheritedLinkedPrivileges().equals(Collections.EMPTY_LIST))
            return false;
      } // if
      else {
          if (!m_inheritedT_LinkedPRIRights_.equals( obj.getEorL_TargetInheritedLinkedPrivileges() ))
            return false;
      } // else
      
      if (l_S_TypedPrivileges_ == null || obj.getEorL_SourcePrivilegeForTypes().equals(Collections.EMPTY_LIST)) {
          if (l_S_TypedPrivileges_ != null || !obj.getEorL_SourcePrivilegeForTypes().equals(Collections.EMPTY_LIST))
            return false;
      } // if
      else {
          if (!l_S_TypedPrivileges_.equals( obj.getEorL_SourcePrivilegeForTypes() ))
            return false;
      } // else

      if (l_S_LinkedPrivileges_ == null || obj.getEorL_SourcePrivilegeForLinks().equals(Collections.EMPTY_LIST)) {
          if (l_S_LinkedPrivileges_ != null || !obj.getEorL_SourcePrivilegeForLinks().equals(Collections.EMPTY_LIST))
            return false;
      } // if
      else {
          if (!l_S_LinkedPrivileges_.equals( obj.getEorL_SourcePrivilegeForLinks() ))
            return false;
      } // else
       */
      
      return true;
  } // equals

  /**
   * @return long
   */
  public long getSerialVersionUID()	{
          return serialVersionUID;
  }

  /**
   * Gets the name last component of each privilege. Never null.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("PRIVILEGE_USER: ");

    if (l_T_TypedPrivileges_ != null) {
        for (Iterator iter = l_T_TypedPrivileges_.iterator(); iter.hasNext(); ) {
          sb.append(iter.next().toString());
          sb.append(", ");
        } //for
    } // if

    if (l_T_LinkedPrivileges_ != null) {
        for (Iterator iter = l_T_LinkedPrivileges_.iterator(); iter.hasNext(); ) {
          sb.append(iter.next().toString());
          sb.append(", ");
        } //for
    } // if

    if (l_inheritedT_LinkedPrivileges_ != null) {
        for (Iterator iter = l_inheritedT_LinkedPrivileges_.iterator(); iter.hasNext(); ) {
          sb.append(iter.next().toString());
          sb.append(", ");
        } //for
    } // if

    if (l_S_TypedPrivileges_ != null) {
        for (Iterator iter = l_S_TypedPrivileges_.iterator(); iter.hasNext(); ) {
          sb.append(iter.next().toString());
          sb.append(", ");
        } //for
    } // if

    if (l_S_LinkedPrivileges_ != null) {
        for (Iterator iter = l_S_LinkedPrivileges_.iterator(); iter.hasNext(); ) {
          sb.append(iter.next().toString());
          sb.append(", ");
        } //for
    } // if

    return " " + sb;
  } // toString()

  //
  // ---------------- methods out of the implemented interfaces --------------
  //

   /**
   * This method cuts the links to the linked privileges, and informs the typed privileges.
   * If the master is a PrivilegeTarget, the linked privileges are deleted if they
   * are LINKED_one-to-one privileges. Otherwise, the linked privileges are kept
   * without the links to the finalizing master.
   * This method is called by finalize(), but finalize() may be not called.
   * An unused object has to be deleted to inform the others objects, and this method
   * HAS TO BE CALLED before deleting the CPrivilegeUser.
   * This method does not raise PropertyChangeEvents since the BeanInfo for this class
   * specifies that its ACS informs the listeners when this instance is finalized.
   * To test if an instance has been finalized, let's use (master_ == null).
   * @exception UpDateError if the main ACS do not known the CPrivilegeUser (if
   * it is a clone without setName(), typically), if a privilege is not correct.
   */
  protected synchronized void finalizeForBase() throws UpDateError {

                //System.out.println("In CPrivilegeUserImpl, call of " + name_ +
                    //".finalizeForBase()");

    if (master_ == null) return;

    // removes the master in linked privileges,
    // since typed privileges do not register the sources and targets
    try {
        Privilege priv;
        if (l_T_LinkedPrivileges_ != null) {
            for(Iterator ite = ((ArrayList)l_T_LinkedPrivileges_.clone()).iterator(); ite.hasNext();) {
                priv = (Privilege)ite.next();
                if (((PrivilegeForLinksImpl)priv).getOneRoleOneTarget()) {
                    priv.finalizeForUser(); // call removeEorPrivilegeToTarget(priv)
                } else removeEorPrivilegeToTarget(priv);
            } // for
        } // if

        if (l_T_TypedPrivileges_ != null) {
            for(Iterator ite = ((ArrayList)l_T_TypedPrivileges_.clone()).iterator(); ite.hasNext();) {
                priv = (Privilege)ite.next();
                removeEorPrivilegeToTarget(priv);
            } // for
        } // if

        if (l_S_LinkedPrivileges_ != null) {
            for(Iterator ite = ((ArrayList)l_S_LinkedPrivileges_.clone()).iterator(); ite.hasNext();) {
                priv = (Privilege)ite.next();

                                // for testing and tracking
                                //if (master_.getNickName().equals("acc_two"))
                                  //  System.out.println("CPrivilegeUserImpl.finalizeForBase(), call "+
                                    //    " for the source 'acc_two' to  remove the privilege"+priv);


                removeEorPrivilegeToSource(priv);
            } // for
        } // if

        if (l_S_TypedPrivileges_ != null) {
            for(Iterator ite = ((ArrayList)l_S_TypedPrivileges_.clone()).iterator(); ite.hasNext();) {
                priv = (Privilege)ite.next();
                removeEorPrivilegeToSource(priv);
            } // for
        } // if

        priv = null;
    } catch (BaseError b) {
        throw new UpDateError(b.toString() );     
    }
        
    master_ = null;
    l_T_TypedPrivileges_ = null;
    l_T_LinkedPrivileges_ = null;
    m_inheritedT_LinkedPRIRights_ = null;
    l_inheritedT_LinkedPrivileges_ = null;
    l_S_TypedPrivileges_ = null;
    l_S_LinkedPrivileges_ = null;
    isSource_ = false;
    isTarget_ = false;
  } // finalizeForBase()

  /**
  * This method finalizes all the privileges, since it is used after
  * a deserialization of the master. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called for EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {


                //System.out.println("In CPrivilegeUserImpl, call of " + name_ +
                    //".finalizeFromDeserialization()");

    if (master_ == null) return;

    // finalizes all the privileges,
    PrivilegeForLinksImpl lpriv;
    PrivilegeForTypeImpl tpriv;
    if (l_T_LinkedPrivileges_ != null) {
        for(Iterator ite = ((ArrayList)l_T_LinkedPrivileges_.clone()).
                iterator(); ite.hasNext();) {
            lpriv = (PrivilegeForLinksImpl)ite.next();
            lpriv.finalizeFromDeserialization();
        } // for
    } // if

    if (l_S_LinkedPrivileges_ != null) {
        for(Iterator ite = ((ArrayList)l_S_LinkedPrivileges_.clone()).
                iterator(); ite.hasNext();) {
            lpriv = (PrivilegeForLinksImpl)ite.next();
            lpriv.finalizeFromDeserialization();
        } // for
    } // if

    if (l_T_TypedPrivileges_ != null) {
        for(Iterator ite = ((ArrayList)l_T_TypedPrivileges_.clone()).
                iterator(); ite.hasNext();) {
            tpriv = (PrivilegeForTypeImpl)ite.next();
            tpriv.finalizeFromDeserialization();
        } // for
    } // if

    if (l_S_TypedPrivileges_ != null) {
        for(Iterator ite = ((ArrayList)l_S_TypedPrivileges_.clone()).
                iterator(); ite.hasNext();) {
            tpriv = (PrivilegeForTypeImpl)ite.next();
            tpriv.finalizeFromDeserialization();
        } // for
    } // if
    lpriv = null;
    tpriv = null;
        
    master_ = null;
    l_T_TypedPrivileges_ = null;
    l_T_LinkedPrivileges_ = null;
    m_inheritedT_LinkedPRIRights_ = null;
    l_inheritedT_LinkedPrivileges_ = null;
    l_S_TypedPrivileges_ = null;
    l_S_LinkedPrivileges_ = null;
  } // finalizeFromDeserialization()

  /**
  * Fires an event to every registered listener, in any order.
  * For the bound properties 'OneSourceTypedPrivilege', 'OneSourceLinkedPrivilege',
  * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege', 'InheritedTargetPrivileges'.
  * Called by all the updating methods.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

                // for testing and tracking
                //System.out.println("in CPrivilegeUserImpl.firePropertyChange(), call to "+
                  //"propertyChange() of "+master_.getNickName()+" for the property "+
                  //_propertyName+" with the old value "+_oldValue);

      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
        {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
   * This method removes the CPrivilegeUser from the ACS lists, but there is no
   * guaranee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {u.printStackTrace();}

    super.finalize();
  } // finalize()
  
} // CPrivilegeUserImpl


