//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;


import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.util.Arrays;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
* This class is responsible for modeling a directory which contains children,
* as resources and other directories.
* It manages fully the AG inherited rights of its children.
* It manages only partially the ACL and PRI inherited rights of its children,
* and this may be completed by the AcsAddons.
* There are three flows of data to manage the inherited AclEntries: <P>
* - one at the ACL creation/deletion, and this is not done in the generic gBase, <P>
* - one at their inner rights editing, and this is the responsability of this class, <P>
* - and the last flow, only for the conditional ACL, at the editing of the membership
* relation between the source and a condition group, and this is the responsability
* of this class.  <P>
* In this later case, this is the calling sequence to reset the conditional AclEntry: <P>
* - GroupIDImpl.addEorMember(), <P>
* - if GroupIDImpl.isConditionalACLtoResetAfterMembershipChangeInThisGroup(), <P>
* - AclEntry.resetRightsAsInheritedAclEntry(). <P>
* There are two flows of data to manage the inherited Privileges: <P>
* - one at the privilege creation/deletion, and this is not done in the generic gBase, <P>
* - one at their sources editing, and this is the responsability of this class. <P>
* A CNodeImpl embedded instance is used to handle the children.
* All variables are protected.
* It is a javabean with the following bound properties:
* 'Children', 'EnforceChildGID'.
* The listeners are proper to this instance, outside gBase,
* transient (not serialized in this instance backup).
* They are called in any order. They receive only a copy of the new value,
* to protect the property. All the exceptions from the listeners are catched,
* and a dialog box is displayed to inform the user. <P>
* This class has several subclasses in this package and in the AcsAddon packages.
* Most of the variables are protected.
* @see ARoad0.gBaseInterface.Directory
* @see ARoad0.gBaseInterface.Resource
*/
public class DirectoryImpl extends ResourceImpl
	implements Directory, Serializable   // , DirectoryIterator to implement ?
{
  /* private variables
  */
  private static final long       serialVersionUID__ = 0;

  /**
   * children handler of this instance
   */
  protected CNodeImpl             nodeHandler_;
  protected boolean               EnforceChildGID_ = false;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final DirectoryImpl EMPTY_INSTANCE = new DirectoryImpl();

  /* Out of the extended interfaces */

  /** variable for the JDK 2 serialization
  * @serial ("EnforceChildGID_",Boolean.TYPE),
  * ("nodeHandler_",CNodeImpl.class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
        new ObjectStreamField("EnforceChildGID_",Boolean.TYPE),
        new ObjectStreamField("nodeHandler_",CNodeImpl.class)
  };

 /**
  * The basic constructor
  */
  public DirectoryImpl() {
      nodeHandler_ = null;

                          // for testing purpose
                     //System.out.println("Call to DirectoryImpl constructor");
  }

   /**
   * This constructor sets this instance as the ACS own resources root if
   * _parent is null. Used by the beamer editor for the ACS resources list,
   * from the 'New' button, or from the gWork.AcsFactory.
   * @param _name of the directory
   * @param _acs is the main ACS of the directory
   * @param _parent of the directory; may be null.
   * @param _resourceType is the type, and may be null if the ACS allows it.
   * @param _owner owns the resource; may be null.
   * @exception CreateError if ACSImpl.newEorResource() throws it -
   *  the instance must then be deleted !
   */
  public DirectoryImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType) throws CreateError {
      
      this(_name, _acs, _parent, _owner, _resourceType, true);

  } // constructor
 
  /**
  * This constructor sets this instance as own resources root if the ACS has
  * not one yet. Used for the creation of a copy.
  * @param _name ImmutableName of the directory.
  * @param _acs main ACS of the directory.
  * @param _parent Node which owns the directory. 
  * @param _owner owns the resource; may be null.
  * @param _resourceType is the type, and may be null if the ACS allows it.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @exception CreateError: see super(_name, _acs, _parent, _resourceType,
  *   _toRegisterInManager) exception
  */
  public DirectoryImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType, boolean _toRegisterInManager) throws CreateError {
      
      super(_name, _acs, _parent, _owner, _resourceType, false);
      
      // update the ACS root name if this is the first or the second node as root
      try {
          if (_parent == null) {
            if (((ACSImpl)_acs).ownResourcesRootName_ == null)
                        ((ACSImpl)_acs).setOwnResourcesRootName(_name);
            else ((ACSImpl)_acs).setOwnResourcesRootName(null); // this is the second node
          } // if
      } catch (UpDateError u) {
        throw new CreateError("in DirectoryImpl constructor: "+u.toString());
      } 

      // update the enforceChildGID_ value
      if (_parent != null && _parent instanceof Directory && 
              ((Directory)_parent).getEnforceChildGID())
                    EnforceChildGID_ = true;

      nodeHandler_ = new CNodeImpl(this);


      // adds this to the map of resources in the main ACS
      // and sends an event "OneResource" to the GUI listeners.
      // may throw a CreateError
      try {
          if (_toRegisterInManager) ((ACSImpl)aCS_).newEorResource(this);
      } catch (CreateError c) {
              try {
                  finalizeForBase();
              } catch (UpDateError r) {}
              throw c;
      }
  } // constructor
  
  /**
  * This method provides a copy of the instance
  * with another name, without children copying
  * because the copy order will not be always correct.
  * Caution:  the user and the group will not be copied if
  * the copy ACS is different and without similar user or group.
  * This method is overridden by the subclasses.
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
  *   May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return copy as DirectoryImpl instance
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null || 
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError("in ResourceImpl.copy(),"+
          " _name or _acs is null"+
          " or _parent ACS is not _acs.");
      }
      
      DirectoryImpl result = new DirectoryImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());
      
      result = copyOwnVariables(result, false);
      
      return result;
  } // copy()
  
  /**
  * This method provides a copy of the instance as Copyable.copy(),
  * but without copying the internal aclEntries.
  * It is used for ACSImpl.copy(), because all the EP must be copied
  * first before the copy of all the resources with their aclEntries.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy),
  * but without copy for the internal aclEntry nor the list of ViewInBase
  * that contains this instance.
  * Called from BaseUtilitiesImpl.copyDirectoriesAndActors().
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of this class
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs,
    Directory _parent, Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in DirectoryImpl.copyWithoutInternalAcl(),"+
          " _name or _acs is null.");
      }

      DirectoryImpl result = new DirectoryImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclentries copy
      
      //if (result.m_AclEntriesResources_ == null) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copyWithoutInternalAcl(,,,)

  //
  // -------------- Methods of the implemented interfaces ---------------
  //

  /**
   * Calls the other setEorL_Children() with the _caller argument equal to
   * this instance to have no action.
   * This method is created for the BeanInfo; it is used by Gui1 beamer
   * which listen the 'Children' changes.
   * Caution: never call this method from a new or a deprecated child
   * of this directory; use then the other method.
   * @deprecated add/removeEorChild methods replaces it
   * @param _l_Child array of leaves which are in this directory. May be
   *  equal to ResourceImpl.EMPTY_INSTANCE, or null.
   * @exception BaseError if a leaf is not owned by the main ACS of
   *  this directory (or InternalError for several rare conditions)
   */
  public final void setEorL_Children(ImmutableLeaf[] _l_Child)
    throws BaseError {
      //setEorL_Children(_l_Child, this);
      return;
  } // setEorL_Children(ImmutableLeaf[] _l_Child)

  /**
  * Gets a sorted array of external object references on the children
  * of this node, which may be leaves (resources) or nodes.
  * The returned type is not immutableLeaf[] to allow BaseUtilityImpl to found
  * the true returned type from the method signature, for BaseObjectHandler
  * when it has to create a new child from the property editor.
  * @return a copy of the array. May be null but not empty.
  */
  public ImmutableResource[] getEorL_Children() {
      if (nodeHandler_ == null) return null;
      ImmutableLeaf[] l_child = nodeHandler_.getEorL_Children();
      if (l_child == null) return null;

      // adaptor from ImmutableLeaf[] to ImmutableResource[]
      ImmutableResource[] l_result = new ImmutableResource[l_child.length];
      System.arraycopy(l_child, 0, l_result, 0, l_result.length);

      l_child = null;
      return l_result;
  }
  
  /**
   * Noop if 'resource 1: new child/node inherits the current G, if activated, 
   * and G is not the account primary group' is not in the ACS rules.
   * @param _a true if the groupID is changed for the newly-created children
   * to this-directory groupID, and the newly-created sub-directories have
   * this property set to true.
   */
  public synchronized void setEnforceChildGID(boolean _a) {
    if (!aCS_.getL_ImmutableRules().contains(
            "resource 1: new child/node inherits the current G, if activated, "+
            "and G is not the account primary group")) return;
    boolean oldValue = EnforceChildGID_;
    EnforceChildGID_ = _a;

                // for testing purpose
                //System.out.println("call to DirectoryImpl.setEnforceChildGID() "+
                //  ", oldValue: "+oldValue+", EnforceChildGID_: "+EnforceChildGID_);    
    
    try {
        firePropertyChange("EnforceChildGID", new Boolean(oldValue), new Boolean(_a));
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From DirectoryImpl.setEnforceChildGID(), firePropertChange() throws:");
        e.printStackTrace();
    }
  } // setEnforceChildGID()

  /**
   * @return boolean true if the groupID is changed for the newly-created files
   * to this directory groupID, and the newly-created sub-directories have
   * this property set to true.
   */
  public boolean getEnforceChildGID() {
          return EnforceChildGID_;
  }

 /**
 * At each rights changing into a direct ACL, if the ACL inheritance is applicable,
 * reset the inherited AclEntry rights in the children.
 * There are three flows of data to manage the inherited AclEntries: <P>
 * - one at the ACL creation/deletion, and this is not done in the generic gBase package, <P>
 * - one at their inner rights editing, <P>
 * - and the last flow, only for the condition ACL, at the editing of the membership
 * relation between the source and a condition group. <P>
 * This method is called for handling the second and the third cases,
 * to update the children of the Directory the AclEntry which is the target of the AclEntry.
 * Not described in the BeanInfo since the user does not call it.
 * Overridden in the AcsAddons.
 * Called by AclEntryImpl.resetRightsAsInheritedAclEntry().
 * @param _acl own AclEntry with updated rights
 */
  public synchronized void resetInheritedAclEntryRightsOfChildren(AclEntry _acl) {
    if (!aCS_.getManageACLRightsInheritance()) {
          InternalError i = new InternalError("DirectoryImpl."+
                    "resetInheritedPrivilegeRightsOfChildren() "+
                    "is called in an ACS that does not manage the ACL inheritance. ");
          i.printStackTrace();
          throw i;
    } // if

    if (!_acl.getResName().equals(name_)) {
          InternalError i = new InternalError("DirectoryImpl."+
                    "resetInheritedAclEntryRightsOfChildren() "+
                    "is called by an AclEntry with a wrong target, on "+
                    name_.lastComponent());
          i.printStackTrace();
          throw i;
    } // if

    // reset the ACL as inherited ACL on the children
    ImmutableResource[] l_child = getEorL_Children();
    AclEntry acl = null;

    if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            acl = selectInheritedAclEntry(_acl, l_child[i].getResourceType());
            if (acl == null) continue;

            ((Resource)l_child[i]).resetInheritedAclEntryRights(acl);

        } // for i
    } // if (l_child != null)

    l_child = null;
    acl = null;
  } // resetInheritedAclEntryRightsOfChildren()

    /**
    * Resets the inherited AclEntry rights after an user edition.
    * Called by the parent when there is a change in the effective rights
    * of an inherited AclEntry.
    * May be overread only in the AcsAddons handling the ACL inheritance.
    * Not described in the BeanInfo since the user does not call it.
    * Note: the creation of the inherited AclEntry in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited AclEntry is handled.
    * Called by resetInheritedAclEntryRightsOfChildren() and
    * this method from the parent.
    * Overrides the super method and calls it.
    * @param _acl inherited AclEntry with updated rights
    */
  public synchronized void resetInheritedAclEntryRights(AclEntry _acl) {


                                  // for testing and tracking
                                  System.out.println(" ");
                                  System.out.println("DirectoryImpl."+
                                          "resetInheritedAclEntryRights() "+
                                          "on "+name_.lastComponent()+
                                          " for the ACL "+_acl);

    super.resetInheritedAclEntryRights(_acl);
      
    // reset the ACL as inherited ACL on the children
    ImmutableResource[] l_child = getEorL_Children();
    AclEntry acl = null;

    if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            acl = selectInheritedAclEntry(_acl, l_child[i].getResourceType());
            if (acl == null) continue;

            ((Resource)l_child[i]).resetInheritedAclEntryRights(acl);

        } // for i
    } // if (l_child != null)

                                  // for testing and tracking
                                  System.out.println("end of DirectoryImpl."+
                                          "resetInheritedAclEntryRights() "+
                                          "on "+name_.lastComponent()+
                                          " for the ACL "+_acl);


    l_child = null;
    acl = null;
  } // resetInheritedAclEntryRights()

    /**
    * Resets the inherited Privilege rights after an user edition.
    * Called by the parent when there is a change in the effective rights
    * of an inherited Privilege. This is only when it is about a change in the sources.
    * Contrary to AclEntry, the rights are immutable and there is no condition.
    * The subclassing of CPrivilegeUserImpl is possible in an AcsAddon, but it is
    * also possible to overread the same method in a subclass of this class.
    * Not described in the BeanInfo since the user does not call it.
    * Note: the creation of the inherited Privilege in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited Privilege
    * is handled.
    * Called by Directory.resetInheritedPrivilegeRightsOfChildren().
    * Overridden in DirectoryImpl and called from it.
    * @param _lpri inherited Privilege with updated rights. Never null.
    */
  public synchronized void resetInheritedPrivilegeRights(PrivilegeForLinks _lpri) {
                                  // for testing and tracking
                                  System.out.println(" ");
                                  System.out.println("DirectoryImpl."+
                                          "resetInheritedAclEntryRights() "+
                                          "on "+name_.lastComponent()+
                                          " for the privilege "+_lpri);

    super.resetInheritedPrivilegeRights(_lpri);
      
    // reset the ACL as inherited ACL on the children
    ImmutableResource[] l_child = getEorL_Children();
    PrivilegeForLinks pri = null;

    if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            pri = selectInheritedPrivilege(_lpri, l_child[i].getType());
            if (pri == null) continue;

            ((Resource)l_child[i]).resetInheritedPrivilegeRights(pri);

        } // for i
    } // if (l_child != null)

                                  // for testing and tracking
                                  System.out.println("end of DirectoryImpl."+
                                          "resetInheritedPrivilegeRights() "+
                                          "on "+name_.lastComponent()+
                                          " for the privilege "+_lpri);


    l_child = null;
    pri = null;
  } // resetInheritedPrivilegeRights()

 /**
 * Resets the inherited Privilege rights in the children, if the Privilege
 * inheritance is applicable.
 * There are two flows of data to manage the inherited Privileges: <P>
 * - one at the privilege creation/deletion, and this is not done in the generic gBase package, <P>
 * - one at their sources editing, <P>
 * This method is called for handling the second case,
 * to update the children of the Directory which is the target of the Privilege.
 * Not described in the BeanInfo since the user does not call it.
 * Overridden by the AcsAddons.
 * Called by PrivilegeForLinksImpl.resetRightsAsInheritedPrivilege().
 * @param _lpri inherited Privilege with updated rights
 */
  public synchronized void resetInheritedPrivilegeRightsOfChildren(PrivilegeForLinks _lpri) {
    if (!aCS_.getManageLPRIRightsInheritance()) {
          InternalError i = new InternalError("DirectoryImpl."+
                    "resetInheritedPrivilegeRightsOfChildren() "+
                    "is called in an ACS that does not manage the PRI inheritance. ");
          i.printStackTrace();
          throw i;
    } // if

    if (_lpri.getEorLinkedTargets().get(0) != this) {
          InternalError i = new InternalError("DirectoryImpl."+
                    "resetInheritedPrivilegeRightsOfChildren() "+
                    "is called by a Privilege with a wrong target, on "+
                    name_.lastComponent());
          i.printStackTrace();
          throw i;
    } // if

    // reset the Privilege as inherited Privilege on the children
    ImmutableResource[] l_child = getEorL_Children();
    PrivilegeForLinks pri = null;

    if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            pri = selectInheritedPrivilege(_lpri, l_child[i].getType());
            if (pri == null) continue;

            ((Resource)l_child[i]).resetInheritedPrivilegeRights(pri);

        } // for i
    } // if (l_child != null)

    l_child = null;
    pri = null;
  } // resetInheritedPrivilegeRightsOfChildren()

  //
  // --------------------- protected Directory methods ----------------------
  //
  
  /**
  * This method provides a copy of the instance variables.
  * Does not copy the children.
  * Called by copy(), and by the copyOwnVariables() methods
  * of the subclasses.
  * Calls super.copyOwnVariables().
  * Caution:  the user and the group will not be copied if
  * the copy ACS is different and without similar user or group.
  * @param _result to update with this instance variables copies.
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception if super throws it
  **/
  protected DirectoryImpl copyOwnVariables(DirectoryImpl _result,
    boolean _withoutInternalAcl) throws CreateError {
        
      _result = (DirectoryImpl)super.copyOwnVariables(
        _result, _withoutInternalAcl);
      
      _result.EnforceChildGID_ = getEnforceChildGID();
      
      return _result;
  } // copyOwnVariables(,,,)

  /**
   * Adds a child to this Node. A child is not a member.
   * Used by the children constructor.
   * Does not update the parent internal variable in the child.
   * No control on the ACS type policy, since it is done
   * in the ResourceImpl constructor which calls this class
   * for adding a new child. A parent type which contains 'immutable' does not forbid
   * to add children to the parent.
   * Detects the child through a name comparison.
   * Fires the event 'Children'.
   * @param _child is a Resource (not a simple leaf) to put in this Node.
   *   No action if null, or already known.
   * @exception InternalError if the leaves list is null, or if _child
   *  is not a VirtualFolder. If firePropertyChange() throws an exception,
   *  displays an user message.
   */
  protected synchronized void addEorChild(ImmutableLeaf _child) {
      if (!(_child instanceof Resource))
          throw new InternalError("In DirectoryImpl.addEorChild(), "+
            "the child is not a resource.");
      nodeHandler_.addEorChild(_child);
  }

  /**
   * Removes a child from this Node.
   * Used by the children finalizing methods.
   * Does not update the parent internal variable in the child.
   * Detects the child presence through a name comparison.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the children list is null, to speed up the
   * finalizing of this instance, when children are called to be finalizing.
   * @param _child is a virtual folder which is in this CNode. No action if
   *  null.
   * @exception UpDateError if _child is not known.
   */
  protected synchronized void removeEorChild(ImmutableLeaf _child) throws UpDateError {
      nodeHandler_.removeEorChild(_child);
  }
  
  //
  // ------- Overreading methods from BasicImpl or ResourceImpl ---------
  //
 
  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'Children', 'EnforceChildGID'.
  * Overrides the super method and calls it.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (_propertyName.equals("Children") ) {
          nodeHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else super.addPropertyChangeListener(_propertyName, _l);

  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'Children', 'EnforceChildGID'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      if (_propertyName.equals("Children") ) {
          if (nodeHandler_ != null)
              nodeHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else super.removePropertyChangeListener(_propertyName, _l);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

      if (_propertyName.equals("Children")) {
          
            if (nodeHandler_ != null )
                return nodeHandler_.getPropertyChangeListeners(_propertyName);
            else
                return null;
      }
      else return super.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()

  /**
   * Overrides the super method to change the inherited owner rights on the children.
   * Simply calls the super method if the ACS does not manage the AG inherited rights.
   * Called by the ResourceImpl constructor, so nodeHandler_ is null then.
   * Calls BaseUtilityImpl.withAcrossToDirectoryAGRights(),
   * ResourceImpl.setAGAccessFromTheParentAccount() and fireL_inheritedOwnerRights(),
   * and the super method.
   * Fires a PropertyChange for 'OwnerRights'.
   * @param _s array of acsrights and/or metarights. May be null. 
   * @exception UpDateError if the owner rights are immutable.
   */
  public synchronized void setL_ownerRights(StringRight[] _s) throws UpDateError {

                // for testing and tracking
                //System.out.println("DirectoryImpl.setL_ownerRights() in "+getName()+
                  //      ", entered rights at the start: ");
                //for (int i = 0; i < _s.length; i++) {
                  //  System.out.println("------------ "+_s[i]);
                //} // for i 
      
    boolean         old = BaseUtilityImpl.withAcrossToDirectoryAGRights(getL_ownerRights());
    boolean         allowToGoAcross = false;
    StringRight[]   l_old = null;
    StringRight[]   l_new = null;
    Resource[]      l_Children = null;
    
    if (nodeHandler_ != null) {
        l_Children = (Resource[])nodeHandler_.getEorL_Children();
        if (l_Children != null) {
            for (int  i=0 ; i < l_Children.length ; i++) {
              if (l_Children[i] == null ) continue;
              // l_old is the same for all the children
              l_old = l_Children[i].getL_inheritedOwnerRights();
              break;
            } // for
        } // if
    } // if (nodeHandler_
    
    super.setL_ownerRights(_s);

    if (!aCS_.getManageAGORightsInheritance() || l_Children == null) {
        l_Children = null;
        l_old = null;
        return;
    } // if
    
    allowToGoAcross = BaseUtilityImpl.withAcrossToDirectoryAGRights(getL_ownerRights());
    
    if ((old && !allowToGoAcross) || (!old && allowToGoAcross)) {
        for (int  i=0 ; i < l_Children.length ; i++) {
          if (l_Children[i] == null ) continue;
          ((ResourceImpl)l_Children[i]).
                  setAGOAccessFromTheParentAccount( allowToGoAcross );
        } // for 
    } // if
    
    for (int  i=0 ; i < l_Children.length ; i++) {
      if (l_Children[i] == null ) continue;
      // l_new is the same for all the children
      l_new = l_Children[i].getL_inheritedOwnerRights();
      break;
    } // for 
    
    // calls fireL_inheritedXXX... even if setAGOAccessFromTheParentXXX has been called,
    // to update the property editor in the GUI
    if ( l_Children.length != 0 &&
            !Arrays.asList(l_old).equals(Arrays.asList(l_new)) ) {

                // for testing purpose
                //System.out.println("DirectoryImpl.setL_groupRights(): "+
                  //"call to l_Children_.fireL_inheritedGroupRights()");
        
        for (int  i=0 ; i < l_Children.length ; i++) {
          if (l_Children[i] == null ) continue;
          ((ResourceImpl)l_Children[i]).fireL_inheritedOwnerRights(l_old, l_new);
        } // for 
    } // if      

    l_old = null;
    l_new = null;
    l_Children = null;
  } // setL_ownerRights()
 
  /**
   * Overrides the super method to call getL_inheritedOwnerRights(),
   * BaseUtilityImpl.withAcrossToDirectoryAGRights(), fireL_inheritedGroupRights() and
   * setAGOAccessFromTheParentGroup() on the children.
   * Simply calls the super method if the ACS does not manage the AG rights inheritance.
   * Called by the ResourceImpl constructor, so nodeHandler_
   * is null then. Calls the super method.
   * Fires a PropertyChange for 'GroupRights'.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the group rights are immutable.
   */
  public synchronized void setL_groupRights(StringRight[] _s) throws UpDateError {
    boolean         old = BaseUtilityImpl.withAcrossToDirectoryAGRights(getL_groupRights());
    boolean         allowToGoAcross = false;
    StringRight[]   l_old = null;
    StringRight[]   l_new = null;
    ImmutableLeaf[] l_Children = null;

                        // for testing purpose
                        //System.out.println("call to DirectoryImpl.setL_groupRights() for"+
                          //      name_);

    if (nodeHandler_ != null) {
        l_Children = nodeHandler_.getEorL_Children();
        if (l_Children != null) {
            for (int  i=0 ; i < l_Children.length ; i++) {
              if (l_Children[i] == null ) continue;
              // l_old is the same for all the children
              l_old = ((Resource)l_Children[i]).getL_inheritedGroupRights();
              break;
            } // for
        } // if
    } // if (nodeHandler_
    
    super.setL_groupRights(_s);

    if (!aCS_.getManageAGORightsInheritance() || l_Children == null) {

                        // for testing purpose
                        //System.out.println("DirectoryImpl.setL_groupRights() returns, with: "+
                          //"'right access 1: child/node access only if the current parent tree rights"+
                            //" allow to go across i': "+aCS_.getL_ImmutableRules().contains(
                            //"right access 1: child/node access only if the current parent tree rights"+
                            //" allow to go across it")+
                          //", and l_Children: "+l_Children);

        l_Children = null;
        l_old = null;
        return;
    } // if (!aCS_.getL_ImmutableRules().contains(
    
    allowToGoAcross = BaseUtilityImpl.withAcrossToDirectoryAGRights(getL_groupRights());

                        // for testing purpose
                        //System.out.println("DirectoryImpl.setL_groupRights(): "+
                          //"old allowToGoAcross: "+old+
                          //", new allowToGoAcross: "+allowToGoAcross);

    if ((old && !allowToGoAcross) || (!old && allowToGoAcross)) {
        for (int  i=0 ; i < l_Children.length ; i++) {
          if (l_Children[i] == null ) continue;


                        // for testing purpose
                        //System.out.println("DirectoryImpl.setL_groupRights(): "+
                          //"call to l_Children_.setAGOAccessFromTheParentGroup()");

          ((ResourceImpl)l_Children[i]).
                  setAGOAccessFromTheParentGroup( allowToGoAcross );
        } // for 
    } // if
    
    for (int  i=0 ; i < l_Children.length ; i++) {
      if (l_Children[i] == null ) continue;
      // l_new is the same for all the children
      l_new = ((ResourceImpl)l_Children[i]).getL_inheritedGroupRights();
      break;
    } // for 
    
    if (l_Children.length != 0 &&
             !Arrays.asList(l_old).equals(Arrays.asList(l_new)) ) {

                        // for testing purpose
                        //System.out.println("DirectoryImpl.setL_groupRights(): "+
                          //"call to l_Children_.fireL_inheritedGroupRights()");
        
        for (int  i=0 ; i < l_Children.length ; i++) {
          if (l_Children[i] == null ) continue;
          ((ResourceImpl)l_Children[i]).fireL_inheritedGroupRights(l_old, l_new);
        } // for 
    } // if

    l_Children = null;
    l_old = null;
    l_new = null;
  } // setL_groupRights()

  /**
  * Finalizes the directory and ALL ITS CHILDREN.
  * If this instance is the current resources root, set it to null.
  * Note: when this method is called by the ResourceImpl constructor, the proper
  * variables of this class are not still initialized.
  * @exception UpDateError if the directory is unknown of the main ACS (if
  *  it is a clone without setName(), typically)
  */
  public synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call of DirectoryImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (folderMember_ == null) {
      return;
    }

    // finalizes all the children
    if (nodeHandler_ != null)
        nodeHandler_.finalizeForBase();

    if (getEorACS().getOwnResourcesRootName().equals(getName()))
        ((ACSImpl)getEorACS()).setOwnResourcesRootName(NameImpl.EMPTY_INSTANCE);
    
    nodeHandler_ = null;
    super.finalizeForBase();
  } // finalizeForBase()

   /**
   * Clone without register the new object in the ACSs.
   * Caution: don't use any modifying method that may altere the integrity
   * of the base.
   * @exception CloneNotSupportedException
   * @return Object; is not registred in the ACSs. The names, the aclEntries,
   * the Linux/Unix rights, the ACS controller names are cloned.
   * The external object references are not cloned (see "Eor" in get method).
   * So, the children are not cloned.
   */
  public synchronized Object clone() {
    DirectoryImpl clon = (DirectoryImpl) super.clone();

    //clon.nodeHandler_ = nodeHandler_.cloneC(this);

    clon.setEnforceChildGID(EnforceChildGID_);

    return clon;
  }

/**
 * @param _obj Object
 * @return true if _obj is a DirectoryImpl instance (or an instance of an
 *  extended class) equals to the current one (but test of equals only on
 *  names, ACS names of children, and parent name)
 */
public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      DirectoryImpl obj = (DirectoryImpl) _obj;

      // getEorL_Children() is never null, but to have the same pattern, we do it
      if (nodeHandler_ == null || obj.nodeHandler_ == null) {
          if (nodeHandler_ != null || obj.nodeHandler_ != null)
            return false;
      } // if
      else if (!nodeHandler_.equals(obj.nodeHandler_)) 
        return false;

      if (getEorParent() == null || obj.getEorParent() == null) {
          if (getEorParent() != null || obj.getEorParent() != null)
            return false;
      } // if
      else {
          if (!getEorParent().getName().equals( obj.getEorParent().getName() ))
            return false;
      } // else

      if (EnforceChildGID_ != obj.getEnforceChildGID() )
          return false;

      return true;
  } // equals

 /**
  * Overrides the super method.
 * @return the hashcode value for this object
 */
public final int hashCode() {
    int result = super.hashCode();


     // l_Children_ can't be implied, because l_Children_[i].hashCode() uses
     // parent_.hashCode() and it makes a circular relationship

    result = PRIME*result + (EnforceChildGID_ ? 1:0);
    if (parent_ != null)
      result = PRIME*result + ((BaseObject)parent_).getName().hashCode();

    return result;
  } // hashCode()

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /* Out of the extended interfaces */

   /**
   * @return only the last component of the name.
   *  Never null.
   */
  public String toString() {
    StringBuilder sb = new StringBuilder(super.toString());
    //sb.append(getName().lastComponent() );
/*    sb.append("DIRECTORYIMPL : ");
    sb.append(super.toString());
    sb.append("END of RESOURCEIMPL |||");

    sb.append("; l_Children_.length = ");
    sb.append(l_Children_.length);
    sb.append("; l_Children_[0].getName() = ");
    if(l_Children_[0] != null && l_Children_[0] != ResourceImpl.EMPTY_INSTANCE)
      sb.append(l_Children_[0].getName().toString());

    sb.append(", l_Children_[l_Children_.length - 1].getName() = ");
    if(l_Children_[l_Children_.length - 1] != null  && l_Children_[0] != ResourceImpl.EMPTY_INSTANCE)
      sb.append(l_Children_[l_Children_.length - 1].getName().toString());
/*
    sb.append("; EnforceChildGID_ = ");
    sb.append(EnforceChildGID_);

    sb.append("; stickyBit_ = ");
    sb.append(stickyBit_);

    return "[[[" + sb + " END OF DIRECTORYIMPL]]]";
*/
    return " " + sb;
  } // toString()

  //
  // ------------- protected methods ----------------------------------
  //

  /**
  * This method finalizes all the children, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called by EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In DirectoryImpl, call of " +
                //name_ + ".finalizeFromDeserialization()");

    if (folderMember_ == null) return;

    // set parent to null in all the children
    // the  exception is not catched
    nodeHandler_.finalizeFromDeserialization();

    nodeHandler_ = null;

    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

}

