//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrW.DialogError;

import java.util.TreeMap;
import java.util.SortedMap;
import java.util.Map;
import java.util.Iterator;
import java.util.Collections;
import java.io.Serializable;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeSupport;

/**
 * This class is responsible for modeling in the base, a view
 * about access control issues between eligible parties (EP of EPRViewInBase)
 * or virtual folders as access sources, and resources (R of EPRViewInBase)
 * or virtual folders as access targets.
 * It is used by the full views in the GUI to display
 * all the rights between any couple of view objects, since an access source
 * may also be an access target in some precise cases.
 * It registers its sources and its targets from one or several open ACS.
 * It is closed when any of its objects belongs to a closing ACS.
 * It is not closed when one of its objects is referenced in the AclEntry
 * of a closing ACS, or when the object is deleted from its ACS.
 * The EPRViewInBase adds automatically in the view the reference of an alias
 * and the alias of a reference. This is not done in its two subclasses.
 * It is a javabean with the following bound properties (see also the super class):
 * 'OnePropertyInView'.
 * Caution: no equals() nor clone() methods.
 * @see ARoad0.Gui2.GraphicEPRView
 * @version 0.7.0
 */
public class EPRViewInBaseImpl extends ViewInBaseImpl
  implements EPRViewInBase, Serializable {
  /* protected attributes */
  protected transient SortedMap<ImmutableName, ImmutableSource>
          m_Sources_ = new TreeMap();
  protected transient SortedMap<ImmutableName, ImmutableTarget>
          m_Targets_ = new TreeMap();
  protected transient SortedMap<ImmutableName, ImmutableTarget>
          m_NoSourceTargets_ = new TreeMap();
  protected transient PropertyChangeEvent lastEvent_ = null;


  /* public attributes */
  public static final EPRViewInBase EMPTY_INSTANCE = new EPRViewInBaseImpl();

 /**
  * Default constructor
  */
  public EPRViewInBaseImpl() {
    super();
    type_ = "EPR";
  }

  /**
   * @param _name of the view. it is copied.
   * @exception CreateError if the view name is already known in BaseManagerImpl.
   */
  public EPRViewInBaseImpl(ImmutableName _name) throws CreateError {
      super(_name);
      type_ = "EPR";
  }

  /**
   * This constructor is used by BaseManagerImpl.openView() to create
   * every instance. So, all the objects to listen (ACS...) receive a registration
   * from this instance. It would not be the case in a simple deserialization of
   * an EPRViewInBase.
   * Called by RightsFactory_Facade.detectNoThanRights() with _toRegisterInManager
   * false, in a NoThan view worker thread.
   * Calls addEorHardAliasOrReference() and may complements the list of BaseObjects.
   * Calls the super constructor with the same arguments.
   * @param _name of the view. It is copied.
   * @param _m_BaseObjects is the map (name, object) of the view objects.
   *  It is copied. May be null.
   * @param _toRegisterInManager false if this constructor must not register
   *  the instance in BaseManagerImpl and in its view objects.
   * @exception BaseError if the constructor of ViewInBaseImpl fires it, or
   * if _m_BaseObjects contains an object that is not an eligible party,
   * nor a resource.
   */
  public EPRViewInBaseImpl(ImmutableName _name, Map _m_BaseObjects,
    boolean _toRegisterInManager) throws CreateError {

      super(_name, _m_BaseObjects, false);
      type_ = "EPR";
      
      // updates m_Targets_, m_NoSourceTargets_ and m_Sources_
      if (_m_BaseObjects != null) {
          BaseObject obj;
          for (Iterator ite = _m_BaseObjects.values().iterator() ; ite.hasNext() ; ) {
            obj = (BaseObject)ite.next();

            if (!(obj instanceof ImmutableSource) && !(obj instanceof ImmutableTarget))
              throw new CreateError("in EPRViewInBaseImpl constructor, "+obj+
                " is not an access source nor an access target.");

            if (obj instanceof ImmutableTarget) {
                m_Targets_.put( obj.getName(), (ImmutableTarget)obj);

                if (!(obj instanceof ImmutableSource)) {
                m_NoSourceTargets_.put( obj.getName(), (ImmutableTarget)obj);
                }

                // add the hard alias or reference, not the soft alias or reference
                if (((ImmutableTarget)obj).getEorAliasAsInternalHardReference() != null ||
                    ((ImmutableTarget)obj).getEorReferenceAsInternalHardAlias() != null)
                        addEorHardAliasOrReference((ACSObject)obj);
            } // if
            if (obj instanceof ImmutableSource) {
                m_Sources_.put( obj.getName(), (ImmutableSource)obj);

                // add the hard alias or reference, not the soft alias or reference
                if (!(obj instanceof ImmutableTarget) &&
                        (((ImmutableSource)obj).getEorAliasAsInternalHardReference() != null ||
                    ((ImmutableSource)obj).getEorReferenceAsInternalHardAlias() != null))
                        addEorHardAliasOrReference((ACSObject)obj);
            } // if

          } // for
      } // if (_m_BaseObjects != null

      // adds to the map of Views in BaseManagerImpl after the variables initialization.
      // throws BaseError if it already exists.
      if (_toRegisterInManager) {
        try {
            BaseManagerImpl.getInstance().newEorView(this);
        } catch (CreateError c) {
            this.finalizeForCopy();  // resources and EP are not unlistened
            throw new CreateError(c.toString());
        }
        // registers for 'OpenACS' events
        addOwnListening(true);
      } // if
      
  } // constructor

  //
  // ----------- Methods of the implemented interfaces ----------------
  //
  
  /**
  * Creates, registers in the base and saves a view copy with another name.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy): baseError_,
  * m_BaseObjects_, m_NoSourceTargets_, m_Sources_.
  * Overridden by NoThanViewInBaseImpl.
  * @param _name of the copy. It is copied without test.
  * @exception BaseError if _name is null or already known in BaseManagerImpl,
  * or if _view is null or not known.
  * @return the copied view, which may be a NoThan view
  */
  public EPRViewInBase copyAndSave(ImmutableName _name) throws BaseError {
      
      ViewInBase result = copy(_name, this, true);

      // add result to the map of Views in BaseManagerImpl.
      // throws BaseError if already exists.
        // for testing the view without ViewManager: TAKE OUT THIS LIGNE
      //BaseManagerImpl.getInstance().newEorCopiedViewToSave(result);

      BaseManagerImpl.getInstance().saveOpenView(result);
      
      return (EPRViewInBase)result;
  } // copyAndSave
  
  /**
   * @param _targ to test
   * @return true if the view contains it
   */
  public boolean containsTarget(ImmutableTarget _targ) {
    if (_targ == null) return false;
    else return m_Targets_.containsKey(_targ.getName() );
  }

  /**
   * @param _sour to test
   * @return true if the view contains it
   */
  public boolean containsSource(ImmutableSource _sour) {
    if (_sour == null) return false;
    else return m_Sources_.containsKey(_sour.getName() );
  }

  /**
   * Gets all the access targets in this view, and this covers simple resources,
   * actors, directories, executables, accounts, groups and virtual folders.
   * @return unmodifiableSortedMap of the view targets
   */
  public SortedMap<ImmutableName, ImmutableTarget> getEorM_Targets() {
    return Collections.unmodifiableSortedMap(m_Targets_);
  }

  /**
   * Gets the targets in this view which are not eligible parties (actors, accounts, groups)
   * nor virtual folders.
   * @return unmodifiableSortedMap of the view targets
   */
  public SortedMap<ImmutableName, ImmutableTarget> getEorM_NoSourceTargets() {
    return Collections.unmodifiableSortedMap(m_NoSourceTargets_);
  }

  /**
   * Gets all the sources  in this view, and this covers the actors, accounts,
   * groups and virtual folders.
   * @return unmodifiableSortedMap of the view sources
   */
  public SortedMap<ImmutableName, ImmutableSource> getEorM_Sources() {

          // for testing and tracking
          //System.out.println("in EPRViewInBaseImpl.getEorM_EligibleParties(), "+
          //  "the ep are: ");
          //for (Iterator ite = m_Sources_.values().iterator() ; ite.hasNext() ; ) {
          //  System.out.println( "---------- "+ite.next().toString() );
          //}

    return Collections.unmodifiableSortedMap(m_Sources_);
  }

  /**
   * Adds a baseObject to the base.
   * If it is both a source and an access target, adds it in all the relevant maps.
   * Calls addEorHardAliasOrReference(), addEorTarget() and addEorSource().
   * Overrides the super method but calls it indirectly through addEorTarget() and
   * addEorSource().
   * Caution: ImmutableGroupIDMember extends ImmutableTarget.
   * @param _obj non-null BaseObject to add in the view
   * @exception CreateError if the object is already known by the view,
   * or is not a source nor an target.
   */
  public synchronized void addEorBaseObject(BaseObject _obj) throws CreateError {
      // for finalized instances that are still in the memory stack
    if (m_BaseObjects_ == null) return;
    // calls the correct method
    if (_obj instanceof ImmutableTarget) {
        addEorTarget( (ImmutableTarget)_obj );

        if ((((ImmutableTarget)_obj).getEorAliasAsInternalHardReference() != null ||
                ((ImmutableTarget)_obj).getEorReferenceAsInternalHardAlias() != null))
            addEorHardAliasOrReference((ACSObject)_obj);
    }
    else if (_obj instanceof ImmutableSource) {
        addEorSource( (ImmutableSource)_obj );

        if ((((ImmutableSource)_obj).getEorAliasAsInternalHardReference() != null ||
                ((ImmutableSource)_obj).getEorReferenceAsInternalHardAlias() != null))
            addEorHardAliasOrReference((ACSObject)_obj);
    }
    else throw new CreateError("in EPRViewInBaseImpl.addEorBaseObject(_obj)"+
            " for "+name_+", _obj is not a source nor an access target: "+_obj);

  } // addEorBaseObject()

  /**
   * Overrides ViewInBaseImpl.removeEorBaseObject() to update
   * sources and targets maps.
   * If it is both a source and a target, removes in all the relevant maps.
   * Calls removeEorAlias(), removeEorTarget() and removeEorSource().
   * Overrides the super method but calls it indirectly through removeEorTarget() and
   * removeEorSource()
   * @param _obj non-null BaseObject to remove in the view
   * @exception CreateError if the object is null or not known by the view,
   * or if the object is not a source nor a target, or
   * if the object name is empty.
   */
  public synchronized void removeEorBaseObject(BaseObject _obj) throws CreateError {
    if (!(_obj instanceof ImmutableTarget ) && !(_obj instanceof ImmutableSource))
        throw new CreateError("in EPRViewInBaseImpl.removeEorBaseObjects(_obj) for "+
            getName()+", _obj is not a source nor a target: "+_obj);

    if (_obj instanceof ImmutableTarget ) {
        removeEorTarget( (ImmutableTarget)_obj );

        if ((((ImmutableTarget)_obj).getEorAliasAsInternalHardReference() != null ||
                ((ImmutableTarget)_obj).getEorReferenceAsInternalHardAlias() != null))
            removeEorHardAliasOrReference((ACSObject)_obj);
    }
    else if (_obj instanceof ImmutableSource){
        removeEorSource( (ImmutableSource)_obj );

        if ((((ImmutableSource)_obj).getEorAliasAsInternalHardReference() != null ||
                ((ImmutableSource)_obj).getEorReferenceAsInternalHardAlias() != null))
            removeEorHardAliasOrReference((ACSObject)_obj);
    }

  } // removeEorBaseObject()

  /**
   * Updates all the internal maps after a test of the parameter type.
   * So, an Actor is also added to the Sources internal map.
   * Invokes super.addEorBaseObject(_res). 
   * No operation if _res is null or already known.
   * Registers this instance as a listener of the argument and, for all its
   * indirect parents, a listener of the AGO inherited rights. But the 'AclRights'
   * events are not processed by propertyChange(), since this is the responsability
   * of the Gui2.GraphicViewBaseListenerImpl of the view.
   * Calls BasicImpl.addInViewList() or VirtualFolderImpl.addInViewList().
   * Overridden and called by the NoThanViewInBaseImpl subclass.
   * Called by addEorBaseObject() and calls FolderAbst.addEorBaseObject()
   * which fires an event.
   * @param _targ to add to the view
   * @exception CreateError if the target is already known by the view,
   * if the name is null, or the object name is null or empty.
   */
  public void addEorTarget(ImmutableTarget _targ) throws CreateError {
      // for finalized instances that are still in the memory stack
    if (m_BaseObjects_ == null) return;
    // exception
    if (_targ == null || m_Targets_.containsKey(_targ.getName())) {
      return;
    }

    m_Targets_.put( _targ.getName(), _targ);

    if (!(_targ instanceof ImmutableSource)) {
      m_NoSourceTargets_.put( _targ.getName(), _targ);
    }
    else {
      m_Sources_.put( _targ.getName(), (ImmutableSource)_targ);
    }

    _targ.addPropertyChangeListener(this);

    if (_targ instanceof BasicImpl)
        ((BasicImpl)_targ).addInViewList(this);
    else if (_targ instanceof VirtualFolderImpl)
        ((VirtualFolderImpl)_targ).addInViewList(this);

    // to follow the access accross the nodes tree of each resource.
    if (_targ instanceof Resource) {

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the AcsAddon
        if (_targ.getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().
                  addAddonParentPropertyChangeListener(this, _targ);

        ImmutableNode res = ((Resource)_targ).getEorParent();

        while (res != null) {
          res.addPropertyChangeListener("FromTheParentAccount", this);
          res.addPropertyChangeListener("FromTheParentGroup", this);
          res.addPropertyChangeListener("InheritedOwnerRights", this);
          res.addPropertyChangeListener("InheritedGroupRights", this);
          res.addPropertyChangeListener("InheritedTargetPrivileges", this);
          res.addPropertyChangeListener("InheritedTargetAclEntries", this);

          res = res.getEorParent();
        } // while
        
        res = null;
    } // if (_targ
    else if (_targ instanceof VirtualFolder) {

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the AcsAddon
        if (((ACSObject)_targ).getEorACS() instanceof AcsAddon) {
            // unsuseful since all the property changes are listened
            // vf has to be an ACSObjectAddon, following the AcsAddon framework
            //UtilityImpl.getFirstInstance().
              //    addAddonInnerPropertyChangeListener(this, (ACSObjectAddon)_targ);

           UtilityImpl.getFirstInstance().
                  addAddonParentPropertyChangeListener(this, (ACSObject)_targ);
        } // if

        VirtualFolder vf = (VirtualFolder)((VirtualFolder)_targ).getEorParent();

        while (vf != null) {
          vf.addPropertyChangeListener("InheritedTargetPrivileges", this);

          vf = (VirtualFolder)vf.getEorParent();
        } // while

        vf = null;
    } // else if (_targ instanceof VirtualFolder)

    // updates and fires an event
    super.addEorBaseObject(_targ); // throws BaseError
  } // addEorTarget()

  /**
   * Updates all the internal maps after a test of the parameter type.
   * Unregisters this instance as a listener of the argument and, for all its
   * indirect parents, a listener of the AGO inherited rights.
   * Calls BasicImpl.removeInViewList()
   * Invokes super.removeEorBaseObject(_targ).
   * Overridden and called by the NoThanViewInBaseImpl subclass.
   * Called by removeEorBaseObject() and calls FolderAbst.removeEorBaseObject().
   * @param _targ to remove in the view
   * @exception CreateError if the target is not known by the view, or
   * if the object name is empty.
   */
  public void removeEorTarget(ImmutableTarget _targ) throws CreateError {

                    // for testing and tracking
                    //System.out.println("In EPRViewInBaseImpl.removeEorResource() starts,"+
                      //" for the target "+_targ+" the targets are:");
                    //for (Iterator ite = getEorM_Targets().values().iterator() ; ite.hasNext() ; ) {
                      //System.out.println( ((BaseObject)ite.next()).toString() );
                    //}

    if(_targ == null) {
      getEorBaseError().addException("In EPRViewInBaseImpl.removeEorResource(Resource),"+
        " the target is null in the ACS");
      throw new CreateError("In EPRViewInBaseImpl.removeEorResource(Resource),"+
        " the target is null in the ACS");
    }
    if (((ImmutableTarget)_targ).getEorAliasAsInternalHardReference() == null &&
                ((ImmutableTarget)_targ).getEorReferenceAsInternalHardAlias() == null &&
                            !m_Targets_.containsKey(_targ.getName() )) {
      getEorBaseError().addException("In EPRViewInBaseImpl.removeEorResource(Resource),"+
        " the target is unknown in the ACS");
      throw new CreateError("In EPRViewInBaseImpl.removeEorResource(Resource),"+
        " the target is unknown in the ACS");
    }
    else if (!m_Targets_.containsKey(_targ.getName() ))
        return;

    Resource res = null;
    ACSObject obj = null;

    // update super at the end, because it fires an event for the GUI
    m_Targets_.remove( _targ.getName());

    if (!(_targ instanceof ImmutableSource)) {
      m_NoSourceTargets_.remove( _targ.getName());
    }
    else {
      m_Sources_.remove( _targ.getName());
    }

    _targ.removePropertyChangeListener(this);
    
    if (_targ instanceof BasicImpl)
        ((BasicImpl)_targ).removeInViewList(this);
    else if (_targ instanceof VirtualFolderImpl)
        ((VirtualFolderImpl)_targ).removeInViewList(this);   

    // to follow the access accross the nodes tree of each resource
    if (_targ instanceof Resource) {

        // Removes the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the ACS Addon
        if (_targ.getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().
                  removeAddonParentPropertyChangeListener(this, _targ);

        res = (Resource)((Resource)_targ).getEorParent();

        while (res != null) {
          res.removePropertyChangeListener("FromTheParentAccount", this);
          res.removePropertyChangeListener("FromTheParentGroup", this);
          res.removePropertyChangeListener("InheritedOwnerRights", this);
          res.removePropertyChangeListener("InheritedGroupRights", this);
          res.removePropertyChangeListener("InheritedTargetPrivileges", this);
          res.removePropertyChangeListener("InheritedTargetAclEntries", this);

          res = (Resource)res.getEorParent();
        } // while

        // but these removings may have affected other resources having
        // some common nodes with res, so reinitializes all the nodes from m_NoSourceTargets_
        for (Iterator ite = m_NoSourceTargets_.values().iterator() ; ite.hasNext() ; ) {
            obj = (ACSObject)ite.next();
            if (obj instanceof Resource) res = (Resource)obj;
            else continue;

            while (res != null && res.getEorParent() != null) {
              res.removePropertyChangeListener("FromTheParentAccount", this);
              res.removePropertyChangeListener("FromTheParentGroup", this);
              res.removePropertyChangeListener("InheritedOwnerRights", this);
              res.removePropertyChangeListener("InheritedGroupRights", this);
              res.removePropertyChangeListener("InheritedTargetPrivileges", this);
              res.removePropertyChangeListener("InheritedTargetAclEntries", this);

              res = (Resource)res.getEorParent();
            } // while

            // to follow the access accross the nodes tree of each resource
            while (res != null && res.getEorParent() != null) {
              res.addPropertyChangeListener("FromTheParentAccount", this);
              res.addPropertyChangeListener("FromTheParentGroup", this);
              res.addPropertyChangeListener("InheritedOwnerRights", this);
              res.addPropertyChangeListener("InheritedGroupRights", this);
              res.addPropertyChangeListener("InheritedTargetPrivileges", this);
              res.addPropertyChangeListener("InheritedTargetAclEntries", this);

              res = (Resource)res.getEorParent();
            } // while
        } // for

        res = null;
    } // if (_targ instanceof Resource)
    else if (_targ instanceof VirtualFolder) {

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the ACS Addon
        if (((ACSObject)_targ).getEorACS() instanceof AcsAddon) {
            // unsuseful since all the property changes are listened
            // vf has to be an ACSObjectAddon, following the AcsAddon framework
            //UtilityImpl.getFirstInstance().
              //    addAddonInnerPropertyChangeListener(this, (ACSObjectAddon)vf);

           UtilityImpl.getFirstInstance().
                  removeAddonParentPropertyChangeListener(this, (ACSObject)_targ);
        } // if

        VirtualFolder vf = (VirtualFolder)((VirtualFolder)_targ).getEorParent();

        while (vf != null) {
          vf.removePropertyChangeListener("InheritedTargetPrivileges", this);

          vf = (VirtualFolder)vf.getEorParent();
        } // while

        vf = null;
    } // if (_targ
    
    res = null;    
    
    super.removeEorBaseObject(_targ);

                    // for testing and tracking
                    //System.out.println("In EPRViewInBaseImpl.removeEorResource(),"+
                    //  " the resources after the removing are:");
                    //for (Iterator ite = getEorM_Targets().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ((BaseObject)ite.next()).toString() );
                    //}
    
  } // removeEorTarget()

  /**
   * Updates all the internal maps after a test of the parameter type.
   * So, an Actor is also added to the Targets internal map.
   * Registers this instance as a listener of the argument.
   * Called by addEorBaseObject().
   * Calls BasicImpl.addInViewList() and super.addEorBaseObject() which fires an event.
   * Overridden and called by the NoThanViewInBaseImpl subclass.
   * @param _sour to add to the view. No operation if null or already known.
   * @exception CreateError if the source is already known by the view,
   * if the name is null, or the object name is null or empty.
   */
  public void addEorSource(ImmutableSource _sour) throws CreateError {

                    // for testing and tracking
                    //System.out.println("call to EPRViewInBaseImpl.addEorSource() for "+_sour);

      // for finalized instances that are still in the memory stack
    if (m_BaseObjects_ == null) return;
    if (_sour == null || m_Sources_.containsKey(_sour.getName())) {

                        // for testing and tracking
                    //System.out.println("in EPRViewInBaseImpl.addEorSource(), "+
                    //"_sour is already known: "+_sour);
        
        return;
    }

    m_Sources_.put( _sour.getName(), _sour);

    if (_sour instanceof ImmutableTarget && !m_Targets_.containsKey(_sour.getName())) {
      m_Targets_.put( _sour.getName(), (ImmutableTarget)_sour);
    }

    _sour.addPropertyChangeListener(this);
    if (_sour instanceof BasicImpl)
        ((BasicImpl)_sour).addInViewList(this);
    else if (_sour instanceof VirtualFolderImpl)
        ((VirtualFolderImpl)_sour).addInViewList(this);

    // updates and fires an event
    super.addEorBaseObject(_sour); // throws CreateError
  } // addEorSource()

  /**
   * Updates all the internal maps after a test of the parameter type.
   * Unregisters this instance as a listener of the argument.
   * Called by removeEorBaseObject().
   * Calls BasicImpl.removeInViewList() and super.removeEorBaseObject() which fires an event.
   * Overridden and called by the NoThanViewInBaseImpl subclass.
   * @param _sour to remove in the view
   * @exception CreateError if the source is null or not known by the view, or
   * if the object name is empty.
   */
  public void removeEorSource(ImmutableSource _sour) throws CreateError {

                    // for testing and tracking
                    //System.out.println("call to EPRViewInBaseImpl.removeEorSource() for "+_sour);

    if (_sour == null ) {
      getEorBaseError().addException("In EPRViewInBaseImpl.removeEorSource()"+
        ", the source is null. ");
      CreateError c = new CreateError("In EPRViewInBaseImpl.removeEorSource()"+
        ", the source is null");
      c.printStackTrace();
      throw c;
    }
    if (!m_Sources_.containsKey(_sour.getName() )) {
      getEorBaseError().addException("In EPRViewInBaseImpl.removeEorSource()"+
        ", the source is unknown in the ACS: "+_sour);
      CreateError c = new CreateError("In EPRViewInBaseImpl.removeEorSource()"+
        ", the source is unknown in the ACS: "+_sour);
      c.printStackTrace();
      throw c;
    }

    // update super at the end, because it fires an event for the GUI
    m_Sources_.remove( _sour.getName());

    if (_sour instanceof ImmutableTarget && m_Targets_.containsKey(_sour.getName())) {
      m_Targets_.remove( _sour.getName());
    }

    _sour.removePropertyChangeListener(this);
    if (_sour instanceof BasicImpl)
        ((BasicImpl)_sour).removeInViewList(this);
    else if (_sour instanceof VirtualFolderImpl)
        ((VirtualFolderImpl)_sour).removeInViewList(this);
    
    super.removeEorBaseObject(_sour);

                    // for testing and tracking
                    //System.out.println("In EPRViewInBaseImpl.removeEorSource(),"+
                    //  " the ep after the removing are:");
                    //for (Iterator ite = getEorM_Sources().values().iterator() ; ite.hasNext() ; ) {
                    //  System.out.println( ((BaseObject)ite.next()).toString() );
                    //}

  } // removeEorSource()

  /**
   * Observer of the changes on the objects set in addOwnListening().
   * Overrides the super method, and calls it at the first line.
   * Does not listen 'AclRights' from Resource (this event from AclEntry is not listened),
   * nor 'AclConnectedResources' from EligibleParty, which are handled by
   * Gui2.GraphicViewBaseListenerImpl.
   * Does not listen 'ResourceType' from Resource, 'EPType' from EligibleParty,
   * 'VirtualFolderType' from VirtualFolder.
   * Does not fire a 'InheritedOwnerRights' event after a 'FromTheParentAccount' event
   * from the same source, and so on for the group events.
   * Does not fire a 'InheritedTargetPrivileges' or a 'InheritedTargetAclEntries' event
   * if the last event is from the parent.
   * Fires all the base objects changes to the 'OnePropertyInView' listeners,
   * and the events from AcsAddon objects, without any processing.
   * @param _evt PropertyChangeEvent
   */
  public void propertyChange(PropertyChangeEvent _evt) {
    // for an empty view
    if (m_BaseObjects_ == null || m_BaseObjects_.isEmpty()) return;
      
    if (_evt.getPropertyName().equals("AclRights")) return;                
    else if (_evt.getPropertyName().equals("AclConnectedResources")) return;
    else if (_evt.getPropertyName().equals("ResourceType")) return;
    else if (_evt.getPropertyName().equals("EPType")) return;
    else if (_evt.getPropertyName().equals("VirtualFolderType")) return;
    else if (_evt.getPropertyName().equals("Comment")) return;
    else if (_evt.getPropertyName().equals("ImpliedOpenViews")) return; 
    
    if (lastEvent_ != null && lastEvent_.getSource() == _evt.getSource() ) {
        if (lastEvent_.getPropertyName().equals("FromTheParentAccount") &&
                _evt.getPropertyName().equals("InheritedOwnerRights")) return; 
        else if (lastEvent_.getPropertyName().equals("FromTheParentGroup") &&
                _evt.getPropertyName().equals("InheritedGroupRights")) return; 
    } // if
    
    if (lastEvent_ != null && _evt.getSource() instanceof PrivilegeTarget &&
            lastEvent_.getSource() == ((PrivilegeTarget)_evt.getSource()).getEorParent() ) {
        if (lastEvent_.getPropertyName().equals("InheritedTargetPrivileges") &&
                _evt.getPropertyName().equals("InheritedTargetPrivileges")) return;
    } // if
    if (lastEvent_ != null && _evt.getSource() instanceof Resource &&
            lastEvent_.getSource() == ((Resource)_evt.getSource()).getEorParent() ) {
        if (lastEvent_.getPropertyName().equals("InheritedTargetAclEntries") &&
                _evt.getPropertyName().equals("InheritedTargetAclEntries")) return;
    } // if

                    // for testing and tracking
                    //System.out.println("EPRViewInBaseImpl.propertyChange() processed for "+
                      //      _evt.getPropertyName()+" from "+_evt.getSource());
                    
    lastEvent_ = _evt;
    
      // super.propertyChange() may close this instance or add/remove an object
    super.propertyChange(_evt);

    // do not process the ACS events that are not AcsAddon events
    if (_evt.getSource() instanceof ACS) {
        if (!(_evt.getSource() instanceof AcsAddon)) {

                    // for testing and tracking
                    //System.out.println("EPRViewInBaseImpl.propertyChange() after "+
                       //     "super.propertyChange(); no process for "+
                         //   _evt.getPropertyName()+" from "+_evt.getSource());

            return;
        }
        // 'OneBaseObject' is listened only by FolderAbst
        else if ( !_evt.getPropertyName().equals("OneBaseObject"))
            firePropertyChange(_evt);
        else
            return;
    } // if


    if (!(_evt.getSource() instanceof ACSObject) ) return;

    // fires all the view changes on Resources, VirtualFolders and EligibleParties
    // for the view objects and also their indirect parents

    // from ResourceImpl (and VirtualFolderImpl for Privilege events)
    if ( _evt.getPropertyName().equals("UserID") ||
      _evt.getPropertyName().equals("GroupID") ||
      _evt.getPropertyName().equals("RootRights") ||
      _evt.getPropertyName().equals("GroupRights") ||
      _evt.getPropertyName().equals("OwnerRights") ||
      _evt.getPropertyName().equals("FromTheParentAccount") ||
      _evt.getPropertyName().equals("InheritedOwnerRights") ||
      _evt.getPropertyName().equals("InheritedGroupRights") ||
      _evt.getPropertyName().equals("OneTargetTypedPrivilege") ||
      _evt.getPropertyName().equals("OneTargetLinkedPrivilege") ||
      _evt.getPropertyName().equals("InheritedTargetPrivileges") ||
      _evt.getPropertyName().equals("InheritedTargetAclEntries"))
        firePropertyChange(_evt);
    // from DirectoryImpl InheritedOwnerRights
    else if (_evt.getPropertyName().equals("Children") ||
      _evt.getPropertyName().equals("EnforceChildGID") ||
      _evt.getPropertyName().equals("StickyBit") )
            firePropertyChange(_evt);
    // from ActorImpl
    else if (_evt.getPropertyName().equals("SecondaryGroups") ||
      _evt.getPropertyName().equals("l_isABridgeFor") ||
      _evt.getPropertyName().equals("l_isAGateFor") ||
      _evt.getPropertyName().equals("CurrentGroupID") ||
      _evt.getPropertyName().equals("CurrentUserID") ||
      _evt.getPropertyName().equals("OneSourceTypedPrivilege") ||
      _evt.getPropertyName().equals("OneSourceLinkedPrivilege"))
            firePropertyChange(_evt);
    // from UserIDImpl
    else if (_evt.getPropertyName().equals("PrimGroup") ||
      _evt.getPropertyName().equals("UMask") ||
      _evt.getPropertyName().equals("Order") )
            firePropertyChange(_evt);
    // from EligiblePartyImpl
    else if (_evt.getPropertyName().equals("OneSourceTypedPrivilege") ||
      _evt.getPropertyName().equals("OneSourceLinkedPrivilege") )
            firePropertyChange(_evt);
    // from GroupIDMemberImpl
    else if (_evt.getPropertyName().equals("AssociatedGroups"))
        firePropertyChange(_evt);
    // from GroupIDImpl
    else if (_evt.getPropertyName().equals("l_Members") ||
      _evt.getPropertyName().equals("l_GroupIDCurrentActors") ||
      _evt.getPropertyName().equals("l_GroupIDOwnResources") ||
      _evt.getPropertyName().equals("l_GroupIDOwnActorsAsSecondaryGroup") )
            firePropertyChange(_evt);
    // from VirtualFolderImpl
    else if (_evt.getPropertyName().equals("OneBaseObjectInFolder"))
        firePropertyChange(_evt);

    // all the AcsAddon events from the AcsAddon objects
    else if (_evt.getSource() instanceof ACSObject &&
            ((ACSObject)_evt.getSource()).getEorACS() instanceof AcsAddon)
        firePropertyChange(_evt);

  } // propertyChange()

  /**
   * no equals: each instance is equals only to itself.
   * @param _obj: object to compare
   * @return true if the two objects are identical instances
   * of identical classes that implemente ViewInBase
   */
   /*public synchronized boolean equals(Object _obj) {
    if (this == _obj)
        return true;
    if (_obj == null)
        return false;
    if (_obj.getClass() != getClass())
        return false;

    if (!super.equals(_obj))
        return false;

    return true;
  } //equals()
    */

  /**
   * Tne name is the only property which is used to compute the hashcode,
   * because after finalizeForBase(), hashcode() is used in BaseManagerImpl Maps.
   * @return the hashcode value for the view, based on its name.
   */
  /*public int hashCode() {
    int result = super.hashCode();
    //if (name_ != null) {
    //  result = PRIME*result + name_.hashCode();
    //}
    //if (m_BaseObjects_ != null) {
    //  result = PRIME*result + m_BaseObjects_.keySet().hashCode();
    //}
    return result;
  } // hashcode()
   */

  /**
  * Adds a listener to the bean for the property 'OnePropertyInView',
  * or for any superclass property.
  * @param _propertyName is the name of the property
  * @param _l to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      // for finalized instances that are still in the memory stack
      if (m_BaseObjects_ == null) return;
      if (changeSupport_ == null) {
        changeSupport_= new PropertyChangeSupport(this);
      }

      if (_propertyName.equals("OnePropertyInView")) {
          synchronized (changeSupport_) {
            changeSupport_.addPropertyChangeListener( _propertyName, _l);}
      }
      else super.addPropertyChangeListener(_propertyName, _l);
  }

  /**
  * Removes a listener to the bean for the property 'OnePropertyInView',
  * or for any superclass property.
  * @param _propertyName String is the name of the property
  * @param _l to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      if (changeSupport_ == null) return;

      if (_propertyName.equals("OnePropertyInView")) {
          synchronized (changeSupport_) {
            changeSupport_.removePropertyChangeListener( _propertyName, _l);}
      }
      else super.removePropertyChangeListener(_propertyName, _l);

          // for testing and tracking
          //System.out.println("For the view "+name_+", invoke removePropertyChangeListener(),"+
          //  "for the property 'OnePropertyInView': ");
          //for (Iterator ite = l_OnePropertyInViewChangeSupport_.iterator() ; ite.hasNext() ; ){
          //  System.out.println((PropertyChangeListener)ite.next() );
          //}
          //}

  } // removePropertyChangeListener()

        //
        // ------------- methods out of the extended interfaces --------------------
        //

  /**
  * Creates a copy with another name, and may register it.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy): baseError_,
  * m_BaseObjects_, m_NoSourceTargets_, m_Sources_.
  * Called by copyAndSave(). Overridden by NoThanViewInBaseImpl.
  * @param _name of the view. It is copied.
  * @param _view to copy.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in BaseManagerImpl.
  * @exception BaseError if _name is null or already known in BaseManagerImpl,
  * or if _view is null or not known.
  * @return a copy
  */
  protected static ViewInBase copy(ImmutableName _name,
    EPRViewInBase _view, boolean _toRegisterInManager) throws BaseError {

      if (_view == null || _name == null ||
        !BaseManagerImpl.getInstance().knowsView(_view.getName() )) {
          throw new BaseError("in EPRViewInBaseImpl.copy(_name, _view, boolean),"+
          " _name or _view is null, or _view is not known by the base.");
      }

      // does not call result.addOwnListening()
      EPRViewInBaseImpl result = null;
      try {
          result = new EPRViewInBaseImpl(
            _name, (Map)((TreeMap)((EPRViewInBaseImpl)_view).m_BaseObjects_).clone(), false);
      } catch (CreateError c) {
          throw new BaseError(c.toString());
      }

      // copy _view
      result.baseError_ = ((EPRViewInBaseImpl)_view).getEorBaseError();

        // adds result to the map of Views in BaseManagerImpl.
        // throws BaseError if already exists.
      if (_toRegisterInManager) {
        try {
            BaseManagerImpl.getInstance().newEorView(result);
        } catch (CreateError c) {
            result.finalizeForCopy();
            throw new BaseError(c.toString());
        }
        // registers for 'OpenACS' and 'OneBaseObject' events
        result.addOwnListening(true);
      } 
      
      return result;
  } // copy
  
  /**
  * Registers the instance as a property event listener
  * on the EligibleParties, and on the no-Actor Resources and their parents,
  * to follow the AG/PRI/ACL inherited rights.
  * Called by the constructor, copy() and prepareAfterRestoring() methods. 
  * It does not override the super method to avoid to call it 
  * from super.prepareAfterRestoring(), which has to be called by
  * prepareAfterRestoring() before the updating of the transiant maps and the
  * call to this method. Adds the listening of
  * the specific events from any AcsAddon ACS or any of its objects,
  * and BaseUtilityImpl.addAddonParentPropertyChangeListener() is called for that.
  * Calls super.addOwnListening().
  * @param _callSuper false if this instance must not listen
  *     the open ACS events in BaseManagerImpl and the objects events in ACSImpl; 
  *     if true, calls super.addOwnListening().
  */
  protected void addOwnListening(boolean _callSuper) {
      // for finalized instances that are still in the memory stack
      if (m_BaseObjects_ == null) return;
      
      Resource res;
      ImmutableSource ep;
      VirtualFolder vf;
      ACSObject obj;

      // registering for resources, virtual folders and eligible parties events.
      // Delete operations are handled by the superclass
      for (Iterator ite = m_NoSourceTargets_.values().iterator() ; ite.hasNext() ; ) {
        obj = (ACSObject)ite.next();
        if (obj instanceof Resource) {

            // Gets the specific events to listen from an ACS Addon ACSObject,
            // because these events should change the rights of an ACSObject in the ACS Addon
            if (((ACSObject)obj).getEorACS() instanceof AcsAddon) {
                // unsuseful since all the property changes are listened
                // obj has to be an ACSObjectAddon, following the AcsAddon framework
                //UtilityImpl.getFirstInstance().
                  //    addAddonInnerPropertyChangeListener(this, (ACSObjectAddon)obj);

                UtilityImpl.getFirstInstance().
                      addAddonParentPropertyChangeListener(this, (ACSObject)obj);
            } // if
            
            ((BasicImpl)obj).addInViewList(this);
            ((Resource)obj).addPropertyChangeListener(this);

            res = (Resource)((Resource)obj).getEorParent();

            // to follow the access accross the nodes tree of each resource
            while (res != null) {
              res.addPropertyChangeListener("FromTheParentAccount", this);
              res.addPropertyChangeListener("FromTheParentGroup", this);
              res.addPropertyChangeListener("InheritedOwnerRights", this);
              res.addPropertyChangeListener("InheritedGroupRights", this);
              res.addPropertyChangeListener("InheritedTargetPrivileges", this);
              res.addPropertyChangeListener("InheritedTargetAclEntries", this);

              res = (Resource)res.getEorParent();
            } // while

        } // if (obj
        else if (obj instanceof VirtualFolder) {

            // Gets the specific events to listen from an ACS Addon ACSObject,
            // because these events should change the rights of an ACSObject in the ACS Addon
            if (((ACSObject)obj).getEorACS() instanceof AcsAddon) {
                // unsuseful since all the property changes are listened
                // vf has to be an ACSObjectAddon, following the AcsAddon framework
                //UtilityImpl.getFirstInstance().
                  //    addAddonInnerPropertyChangeListener(this, (ACSObjectAddon)vf);

               UtilityImpl.getFirstInstance().
                      addAddonParentPropertyChangeListener(this, (ACSObject)obj);
            } // if

            ((VirtualFolderImpl)obj).addInViewList(this);
            ((VirtualFolder)obj).addPropertyChangeListener(this);

            vf = (VirtualFolder)((VirtualFolder)obj).getEorParent();

            // to follow the access accross the nodes tree of each virtual folder
            while (vf != null) {
              vf.addPropertyChangeListener("InheritedTargetPrivileges", this);

              vf = (VirtualFolder)vf.getEorParent();
            } // while

        } // if (obj
      } // for

      for (Iterator ite = m_Sources_.values().iterator() ; ite.hasNext() ; ) {
        ep = (ImmutableSource)ite.next();
        if (ep instanceof BasicImpl)
            ((BasicImpl)ep).addInViewList(this);
        else if (ep instanceof EligiblePartyImpl)
            ((EligiblePartyImpl)ep).addInViewList(this);

        ep.addPropertyChangeListener(this);
      } // for ite

      // ACL are listen by GraphicView.getRightsRelatedBaseObjects()
      
      if (_callSuper) addOwnListening();
      
      res = null;
      ep = null;
      obj = null;
      vf = null;
  } // addOwnListening 
  
  /**
  * Unregisters the instance as a property event listener on BaseManagerImpl.
  * Called by finalizeForBase/Copy/GUI(), and the removeOwnListening() methods
  * of the subclass. Removes the listening of
  * the specific events from any AcsAddon ACS or any of its objects,
  * and BaseUtilityImpl.removeAddonParentPropertyChangeListener()
  * is called for that.
  * Calls super.removeOwnListening().
  * @param _callSuper false if this method must not unregister
  *     the open ACS events in BaseManagerImpl and the objects events in ACSImpl; 
  *     if true, call super.removeOwnListening().
  */
  protected void removeOwnListening(boolean _callSuper) {
      Resource res;
      ImmutableSource ep;
      VirtualFolder vf;
      ACSObject obj;

      // unregisters for events.
      for (Iterator ite = m_NoSourceTargets_.values().iterator() ; ite.hasNext() ; ) {
        obj = (ACSObject)ite.next();
        if (obj instanceof Resource) {

            // removes the specific events to listen from an ACS Addon ACSObject,
            // because these events should change the rights of an ACSObject in the ACS Addon
            if (((ACSObject)obj).getEorACS() instanceof AcsAddon)
              UtilityImpl.getFirstInstance().
                      removeAddonParentPropertyChangeListener(
                          this, (ACSObject)obj);
            
            ((BasicImpl)obj).removeInViewList(this);
            ((BasicImpl)obj).removePropertyChangeListener(this);

            res = (Resource)((Resource)obj).getEorParent();

            while (res != null) {
              res.removePropertyChangeListener("FromTheParentAccount", this);
              res.removePropertyChangeListener("FromTheParentGroup", this);
              res.removePropertyChangeListener("InheritedOwnerRights", this);
              res.removePropertyChangeListener("InheritedGroupRights", this);
              res.removePropertyChangeListener("InheritedTargetPrivileges", this);
              res.removePropertyChangeListener("InheritedTargetAclEntries", this);

              res = (Resource)res.getEorParent();
            } // while

        } // if (obj
        else if (obj instanceof VirtualFolder) {

            // removes the specific events to listen from an ACS Addon ACSObject,
            // because these events should change the rights of an ACSObject in the ACS Addon
            if (((ACSObject)obj).getEorACS() instanceof AcsAddon)
              UtilityImpl.getFirstInstance().
                      removeAddonParentPropertyChangeListener(
                          this, (ACSObject)obj);

            ((VirtualFolderImpl)obj).removeInViewList(this);
            ((VirtualFolder)obj).removePropertyChangeListener(this);

            vf = (VirtualFolder)((VirtualFolder)obj).getEorParent();

            // to not follow the access accross the nodes tree of each virtual folder
            while (vf != null) {
              vf.removePropertyChangeListener("InheritedTargetPrivileges", this);

              vf = (VirtualFolder)vf.getEorParent();
            } // while

        } // if (obj
      } // for

      for (Iterator ite = m_Sources_.values().iterator() ; ite.hasNext() ; ) {
        ep =(ImmutableSource)ite.next();
        if (ep instanceof BasicImpl)
            ((BasicImpl)ep).removeInViewList(this);
        else if (ep instanceof EligiblePartyImpl)
            ((EligiblePartyImpl)ep).removeInViewList(this);
        ep.removePropertyChangeListener(this);
      } // for ite
             
      res = null;
      ep = null;
      vf = null;
      
      if (_callSuper) removeOwnListening();      
  } // removeOwnListening   
  
  /**
  * Fires an event to every registered listener, in any order,
  * for a view object bound property.
  * Called by propertyChange() of this class and subclass, which send all 
  * the base objects events to the listeners of the property 'OnePropertyInView',
  * to update the GraphicView. 
  * No link with the ViewInBaseImpl bound properties and firePropertyChange()
  * method.
  * @param _evt is fired by a view object
  */
  protected void firePropertyChange(PropertyChangeEvent _evt) {

      if (changeSupport_ == null) {
            // for testing and tracking
            //System.out.println("In EPRViewInBaseImpl, call to firePropertyChange() from "+name_+
            //  " with null variables.");
        return;
      }

      PropertyChangeListener[] l_listeners = null;
      
      // controls
      if(!changeSupport_.hasListeners("OnePropertyInView")) return;
      if (_evt.getOldValue() != null && _evt.getNewValue() != null &&
        _evt.getOldValue().equals(_evt.getNewValue())) {
          return;
      }

      l_listeners = changeSupport_.getPropertyChangeListeners("OnePropertyInView");

                    // for testing and tracking
                  //if (_evt.getPropertyName().equals("FromTheParentGroup")) {
                    //System.out.println("call of the view "+name_+".firePropertyChange()."+
                      //      " For the event 'FromTheParentGroup'."+
                        //  " Value of l_listeners_ before the calling to listeners: ");
                    //for (int i = 0; i < l_listeners.length; i++) {
                      //System.out.println((PropertyChangeListener)l_listeners[i] );
                    //}
                  //} // if

      for (int i = 0; i < l_listeners.length; i++) {
          l_listeners[i].propertyChange(_evt);
      } // for i

      l_listeners = null;
  } // firePropertyChange()

  /**
   * If an added ACSObject in the view is an alias or an alias reference,
   * adds to the view its associated object. No adding of the alias or reference
   * of the associated object, to avoid to add a chain of aliasings.
   * The test of an alias relation has to be done in the caller method.
   * Called by the constructor and addEorBaseObject().
   * Calls addEorTarget() and addEorSource().
   * Overridden by the NoThanViewInBaseImpl subclass to provide an empty method.
   * @param _obj with an alias relation to process
   * @exception CreateError if the name is null, or the object name is null or empty.
   */
  protected void addEorHardAliasOrReference(ACSObject _obj) throws CreateError {
      
    ACSObject ob = null;
    if (_obj instanceof ImmutableSource) {
        ob = ((ImmutableSource)_obj).getEorAliasAsInternalHardReference();
        if (ob == null)
            ob = ((ImmutableSource)_obj).getEorReferenceAsInternalHardAlias();
    }
    else if (_obj instanceof ImmutableTarget) {
        ob = ((ImmutableTarget)_obj).getEorAliasAsInternalHardReference();
        if (ob == null)
            ob = ((ImmutableTarget)_obj).getEorReferenceAsInternalHardAlias();
    }

    if (ob == null) throw new InternalError("EPRViewInBaseImpl.addEorAlias(): "+
            "no hard alias nor hard alias reference for "+_obj.getName());
    
    if (m_BaseObjects_.containsKey(ob.getName())) return;

    if (ob instanceof ImmutableTarget) {
      addEorTarget( (ImmutableTarget)ob );
    }
    else if (ob instanceof ImmutableSource) {
      addEorSource( (ImmutableSource)ob );
    }

    ob = null;
  } // addEorHardAliasOrReference()

  /**
   * If a removed ACSObject in the view is an alias or an alias reference,
   * removes from the view its associated object.
   * The test of an alias relation has to be done in the caller method.
   * Called by removeEorBaseObject().
   * Calls removeEorTarget() and removeEorSource().
   * Overridden by the NoThanViewInBaseImpl subclass to provide an empty method.
   * @param _obj with an alias relation to process
   * @exception CreateError if the name is null, or the object name is null or empty.
   */
  protected void removeEorHardAliasOrReference(ACSObject _obj) throws CreateError {

    ACSObject ob = null;
    if (_obj instanceof ImmutableSource) {
        ob = ((ImmutableSource)_obj).getEorAliasAsInternalHardReference();
        if (ob == null)
            ob = ((ImmutableSource)_obj).getEorReferenceAsInternalHardAlias();
    }
    else if (_obj instanceof ImmutableTarget) {
        ob = ((ImmutableTarget)_obj).getEorAliasAsInternalHardReference();
        if (ob == null)
            ob = ((ImmutableTarget)_obj).getEorReferenceAsInternalHardAlias();
    }

    if (ob == null) throw new InternalError("EPRViewInBaseImpl.removeEorAlias(): "+
            "no hard alias nor hard alias reference for "+_obj.getName());

    if (!m_BaseObjects_.containsKey(ob.getName())) return;

    if (ob instanceof ImmutableTarget) {
      removeEorTarget( (ImmutableTarget)ob );
    }
    else if (ob instanceof ImmutableSource) {
      removeEorSource( (ImmutableSource)ob );
    }

    ob = null;
  } // removeEorHardAliasOrReference()

  /**
  * The view is removed as a listener and the internal maps are set to null.
  * The view is NOT still closed in BaseManager.
  * Called by BaseManagerImpl.closeOpenView() and, indirectly, by propertyChange().
  * @exception UpDateError from the super class
  */
  public synchronized void finalizeForBase() throws UpDateError {
    // noop if it is the second call to this method
    if (m_Targets_ == null) {
        
              // for testing purpose
              //System.out.println("in EPRViewInBaseImpl.finalizeForBase(), for "+
                //      getName()+", second call to finalization.");

      return;
    }
    
    removeOwnListening(false);

    m_Targets_.clear();
    m_NoSourceTargets_.clear();
    m_Sources_.clear();
    m_Targets_ = null;
    m_NoSourceTargets_ = null;
    m_Sources_ = null;
    lastEvent_ = null;
    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * This method does not clear all the view objects maps, since they are transient.
  * It does not finalizes the view objects, since there are objects from open ACS
  * (see FolderAbst.prepareAfterRestoring()).
  * The view is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null. Only calls the super method.
  * Called by BaseManager.openView() when prepareAfterRestoring() throws an exception.
  */
  protected synchronized void finalizeFromDeserialization() {
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  /**
  * The internal maps are set to null.
  * The view is NOT open in BaseManager.
  * Called by copy(), RightsFactory_Facade.detectNoThanRights() and
  * BaseManagerImpl.getElementsOfOneView().
  */
  public synchronized void finalizeForCopy() {
    // noop if it is the second call to this method
    if (m_Targets_ == null) {
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");

      return;
    }
     
    m_Targets_.clear();
    m_NoSourceTargets_.clear();
    m_Sources_.clear();
    m_Targets_ = null;
    m_NoSourceTargets_ = null;
    m_Sources_ = null;
    super.finalizeForCopy();
    return;
  } // finalizeForCopy()

  /**
  * This method is reserved to the sketch view. This view is removed as
  * a listener and the internal maps are set to null.
  * The sketch view is never registered in BaseManager.
  * Called by BaseManagerImpl.closeSketchView(). Calls removeOwnListening() and
  * super.finalizeForCopy().
  */
  protected synchronized void finalizeForGUI() {
    // noop if it is the second call to this method
    if (m_Targets_ == null) {
        
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");

      return;
    }
    
    removeOwnListening(false);
    
    SortedMap     m_ep = (SortedMap)((TreeMap)m_Sources_).clone();
    SortedMap     m_nar = (SortedMap)((TreeMap)m_NoSourceTargets_).clone();
    try {
        // removes as a listener (never as a listened instance)
        for (Iterator ite = m_nar.values().iterator() ; ite.hasNext() ; ) {
          removeEorTarget( (ImmutableTarget)ite.next() );
        }
        for (Iterator ite = m_ep.values().iterator() ; ite.hasNext() ; ) {
          removeEorSource( (ImmutableSource)ite.next() );
        }
    } catch (CreateError b) {
        throw new InternalError("In EPRViewInBaseImpl.finalizeForGUI(), error: "+
            b.toString());
    }

    m_Targets_.clear();
    m_NoSourceTargets_.clear();
    m_Sources_.clear();
    m_Targets_ = null;
    m_NoSourceTargets_ = null;
    m_Sources_ = null;
    m_ep = null;
    m_nar = null;
    super.finalizeForCopy();
  } // finalizeForGUI()
  
  /**
   * This method must be called by BaseManagerImpl after the opening of the view,
   * to update the transient maps, and to check the open ACS
   * and their objects which are in the restored view.
   * Noop if the folder objects has been restored, having a non-empty BaseObjects map.
   * Calls the super method at the first line.
   * Overrides the super method.
   * @exception DialogError by the super class  if the view contains an object 
   *    from a closed ACS, before the registering of the view as listener; 
   *    the exception message should be shown to the user
   * @exception InternalError if an object is not a Resource nor an EligibleParty.
   * @return a non-null message if some objects has been removed from the view,
   *    or null.
   */
  protected String prepareAfterRestoring() throws DialogError {
    // noop if the folder objects has been restored
    if (m_BaseObjects_ != null && !m_BaseObjects_.isEmpty())
        return null;
    String result = super.prepareAfterRestoring();
    
      // init  
    if (m_Targets_ != null) m_Targets_.clear();
    else m_Targets_ = new TreeMap();
    if (m_NoSourceTargets_ != null) m_NoSourceTargets_.clear();
    else m_NoSourceTargets_ = new TreeMap();
    if (m_Sources_ != null) m_Sources_.clear();
    else m_Sources_ = new TreeMap();
    BaseObject      obj = null;

    // updates m_NoSourceTargets_, m_NoSourceTargets_ and m_Sources_
    for (Iterator ite = m_BaseObjects_.values().iterator() ; ite.hasNext() ; ) {
        obj = (BaseObject)ite.next();

        if (!(obj instanceof EligibleParty) && !(obj instanceof ImmutableTarget)){
            throw new InternalError("EPRViewInBaseImpl.prepareAfterRestoring() where "+
                    obj+" has not a right type.");
        } // if

        if (obj instanceof ImmutableTarget) {
          m_Targets_.put( obj.getName(), (ImmutableTarget)obj);

          if (!(obj instanceof ImmutableSource)) {
            m_NoSourceTargets_.put( obj.getName(), (ImmutableTarget)obj);
          }
        }
        if (obj instanceof ImmutableSource){
          m_Sources_.put( obj.getName(), (ImmutableSource)obj);
        }

    } // for  
    
     // registers as listener on events, to complement super.addOwnListening()
     // called by super.prepareAfterRestoring().
    addOwnListening(false);
    
    obj = null;   
    return result;    
  }  // prepareAfterRestoring()

} // EPRViewInBaseImpl