//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.BaseError;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Collections;
import java.io.ObjectStreamField;

import java.beans.PropertyChangeListener;

/**
* This class is responsible for modeling the simplest rights user
* for which an ACS grants access or denies access to
* some targets through AclEntries or Privileges.
* This kind of entity may be an account (UserID), a group (GroupID) or an actor (Actor).
* This class manages the resources connected to the EligibleParty
* through AclEntries, and it manages its Privileges.
* The constructor adds '&lt;U&gt;' or '&lt;G&gt;' after the ACS name, in the name components,
* if the new instance is an UserID or a GroupID, which are subclasses of this class.
* It is a javabean with the following bound properties:
* 'AclConnectedResources', 'OneSourceTypedPrivilege', 'OneSourceLinkedPrivilege'.
* The listeners are proper to this instance, outside gBase,
* transient (not serialized in this instance backup).
* They include ViewInBaseImpl, gWork.RightsMediatorImpl.
* They are called in any order. They receive only a copy of the new value,
* to protect the property. All the exceptions from the listeners are catched,
* and a dialog box is displayed to inform the user.
* finalizeForUser() uses the ARoad0.Gui1.DialogBox class. <P>
* This class has several subclasses in this package and in the AcsAddon packages.
* Most of the variables are protected.
* @version 0.7.2
*/
public class EligiblePartyImpl extends BasicImpl implements EligibleParty {

    /* private variables*/
  private static final long   serialVersionUID__ = 0;
  //  private Icon              icon_;
  //private boolean             isAlias_;

  /**
   * map(resource name, resource). Is set to null when it is empty.
   */
  protected SortedMap<ImmutableName, ImmutableResource>         m_AclEntriesResources_ = null;

  // ------------------ transient properties ----------------------------------
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final EligiblePartyImpl EMPTY_INSTANCE = new EligiblePartyImpl();

  /** 
  * Properties for the serialization.
  * @serial m_AclEntriesResources_,
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
        new ObjectStreamField("m_AclEntriesResources_",TreeMap.class),
//        new ObjectStreamField("isAlias_",Boolean.TYPE),
//        new ObjectStreamField("trueEP_",EligiblePartyImpl.class),
  };

 /**
  * Default constructor
  */
  public EligiblePartyImpl() {
      
    //      m_AclEntriesResources_ = new TreeMap(); to keep small memory use
  }

   /**
  * Constructor used in the beamer by the editor for the ACS EP list, from 'New'.
  * Adds '&lt;U&gt;' or '&lt;G&gt;' between the ACS name and the end of _name if it is an
  * UserID or a GroupID. This '&lt;G&gt;' adding is used in AclEntryImpl.setRight().
  * @see AssociateUtilityImpl#setFinalArgumentsForAssociateCreation must apply
  *     the same UserID and GroupID naming policy
  * @param _name of the eligible party
  * @param _acs main ACS of the eligible party
  * @param _epType the eligible party type which drives the rights and the immutability
   * of the instance. If null, the value is set to '&lt;undefined&gt;'.
   * @exception CreateError if ACSImpl.newEorEligibleParty() throws it - the instance
   *  must then be deleted!
  */
  public EligiblePartyImpl(ImmutableName _name, ACS _acs, String _epType) throws CreateError {
      this(_name, _acs, _epType, true);
  } // constructor

   /**
   * Used for the creation of an eligible party alias without base registration.
   * Adds a prefix in the name, following the effective class of the created object:
   * UserID or GroupID. Calls setType().
   * @param _name of the eligible party to create, before the adding of the component
   *    'G' if it is a GroupID or 'U' if it is an UserID
   * @param _acs is the ACS of the eligible party to create
   * @param _epType is the eligible party type which drives the rights and the immutability
   * of the instance. If null, the value is set to '&lt;undefined&gt;'.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if ACSImpl.newEorEligibleParty() or setType() throws it -
    * the instance must then be deleted!
   */
  public EligiblePartyImpl(ImmutableName _name, ACS _acs, String _epType,
    boolean _toRegisterInManager) throws CreateError {
      super(_name, _acs);

      try {
          if (_epType != null) {
              setEPType(_epType);
          } // if (_epType
      }
      catch (UpDateError u) {
          throw new CreateError(u.toString());
      }

      //if (_aliasRef != null) {
        //isAlias_ = true;
        //trueEP_ = _aliasRef;
      //}
      
      // adds this to the map of eligible parties in its ACS
      if (_toRegisterInManager) {
          try {
              ((ACSImpl)aCS_).newEorEligibleParty(this);
          } catch (CreateError c) {
              try {
                  finalizeForBase();
              } catch (UpDateError u) {
                throw new InternalError("From the EligibleParty constructor: "+u);
              }

              throw c;
          } // catch
      } // if
  } // constructor

  //
  // ------------ implemented interfaces ------------------------------------
  //
  
  /**
  * This method provides a copy of the instance
  * with another name. Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Comments are not copied, nor the aclEntry resources.
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ResourceImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      EligiblePartyImpl result = new EligiblePartyImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);
      
      return result;
  } // copy(,,,)

  
  /**
  * This method provides a copy of the instance as Copyable.copy(),
  * but without copying the internal aclEntries.
  * It is used for ACSImpl.copy(), because all the EP must be copied
  * first before the copy of all the resources with their aclEntries.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy),
  * but without copy for the internal aclEntry nor the list of ViewInBase
  * that contains this instance.
  * Called from BaseUtilitiesImpl.copyDirectoriesAndActors().
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of EligiblePartyImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs,
    Directory _parent, Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in EligiblePartyImpl.copyWithoutAcl(),"+
          " _name or _acs is null.");
      }

      EligiblePartyImpl result = new EligiblePartyImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclentries copy
      
      //if (result.m_AclEntriesResources_ == null) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copyWithoutInternalAcl(,,,)

  /**
   * Sets the type which drives the immutablility and the allowed rights of the object.
   * If _st is allowed by the ACS, set the type of this eligible party.
   * This method is used to switch from an '&lt;immutable&gt;' type
   * which forbids the updates of a protected property, and to return to this protecting type.
   * Then , it is possible to switch for example from 'actor' to
   * '&lt;immutable&gt; actor', but not to '&lt;immutable&gt; group'.
   * From the default value '&lt;undefined&gt;', any first update may be done.
   * '&lt;12345678901234567890&gt;' is a program-reserved value. Even if not
   * allowed by the ACS, '&lt;12345678901234567890&gt;' is the only new value
   * for which a type starting with '&lt;byAcsCreator' may be changed.
   * From '&lt;12345678901234567890&gt;', any new value may be set from the allowed ones,
   * and this starting value is the only one which allow to set a value containing
   * '&lt;byAcsCreator&gt;'. Calls ACSImpl.getM_SpecializedTypesAndExplorer().
   * Called by the constructor.
   * This is just a wrapper of super.setType() and it fires the event 'EPType'.
   * @param _st type of the resource.
   * @exception UpDateError if the type is null, or does not comply to the criteria.
   */
  public synchronized void setEPType(String _st) throws UpDateError {


      setType(_st);
                    // for testing and tracking
                    //if (this instanceof GroupID)
                      //System.out.println("end of EligiblePartyIDImpl.setEPType() "+
                        //      "with type_: "+type_+" for "+name_);
      
  } // setEPType()

  /**
   * @return String type of the eligible party. May be null.
   */
  public String getEPType()	{
    return type_;
  }

  /**
   * Gets the resources connected to this eligible party.
   *  The resources may belong to differents ACS,
   *  as the aclEntries. One resource may have differents aclEntries
   *  (of different ACS) which connect it to this eligible party;
   *  then, the returned list contains only one reference to this resource.
   * @return the unmodifiable list of references on resources. May be empty,
   *  but not null.
   */
  public final List<ImmutableResource> getEorL_AclConnectedResources()	{
    if (m_AclEntriesResources_ != null) {
        List L = new ArrayList( m_AclEntriesResources_.values() );
        return Collections.unmodifiableList(L);
    } else return new ArrayList(INITIAL_CAPACITY);
  }

   /**
   * Sets a privilege as applicable to this access source.
   * This method is called only by PrivilegeForTypeImpl.
   * For PrivilegeForTypes, only the TYPED and TYPED_CLASSED privilege may be argument.
   * This is false for TYPED_CLASSED_SPECIFIC and TYPED_FOR_SEED privileges,
   * for which an InternalError is thrown.
   * The gBase classes are the only caller of this method
   * at any creation/deletion of a typed privilege or an access source.
   * Fires 'OneSourceTypedPrivilege' events.
   * @param _priv is a Privilege to register in the source
   * @exception BaseError if this instance is not a privilege source,
   * if the privilege second type is not authorized for this source.
   */
  public void addEorPrivilegeForTypeToSource(PrivilegeForType _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if

    privHandler_.addEorPrivilegeToSource(_priv);
  }

   /**
   * Sets a privilege as applicable to this access source.
   * Called by PrivilegeForLinksImpl.addEorPrivilegeSourceLink().
   * Caution: a role-to-one privilege is added even if this instance is immutable,
   * ahile the other linked privileges are not added in this case.
   * Fires 'OneSourceLinkedPrivilege' events.
   * @param _priv to register in the source
   * @exception BaseError if this instance is not a privilege source or is immutable,
   * if the privilege second type is not authorized for this source.
   */
  protected void addEorPrivilegeForLinksToSource(PrivilegeForLinks _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (!_priv.getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE) &&
            (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
            type_.endsWith("<immutable> ")))
          throw new BaseError(
            "The type of "+getName().lastComponent()+" sets it as immutable for its privileges.");

    privHandler_.addEorPrivilegeToSource(_priv);
  } // addEorPrivilegeToSource()

   /**
   * Unregisters the privilege, and this instance in the linked privilege
   * when it is no more used by this privilege source.
   * Called by PrivilegeForTypeImpl.finalizeForBase() and PrivilegeForLinksImpl.
   * Fires 'OneSourceTypedPrivilege' and 'OneSourceLinkedPrivilege' events.
   * @param _priv to unregister
   * @exception BaseError if the privilege is unknown.
   */
  public void removeEorPrivilegeToSource(Privilege _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if

    privHandler_.removeEorPrivilegeToSource(_priv);
  } // removeEorPrivilegeToSource()

   /**
   * Gets all the privileges linked to this source.
   * There is two types of privilege links (allowed or not by the ACS): <P>
   * - PrivilegeForLinks: linked privilege established for a couple
   * (ACSObject, privilege), <P>
   * - PrivilegeForType: typed privilege, where the ACSObject and the privilege
   * has a common type, but without the TYPED_FOR_SEED and
   * TYPED_CLASSED_SPECIFIC privileges. <P>
   * @return List of Privileges. May be empty, but never null.
   */
  public List<Privilege> getEorL_SourcePrivileges() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_SourcePrivileges();
  } // getEorL_SourcePrivileges()
  
   /**
   * Gets all the PrivilegeForTypes of this source.
   * Does not include the TYPED_FOR_SEED and TYPED_CLASSED_SPECIFIC privileges.
   * @return List of PrivilegeForType. May be empty, but never null.
   */
  public List<PrivilegeForType> getEorL_SourcePrivilegeForTypes() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_SourcePrivilegeForTypes();
  }
  
   /**
   * Gets all the PrivilegeForLinks linked to this source.
   * Note: for LINKED_IF_ALL linked privileges, the presence of the privilege
   * does not mean that the master has the privilege rights.
   * Returns an immutable empty list if getIsSource() is false.
   * @return List of PrivilegeForLinks. May be empty, but never null.
   */
  public List<PrivilegeForLinks> getEorL_SourcePrivilegeForLinks() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_SourcePrivilegeForLinks();
  }
 
  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'AclConnectedResources', 'OneSourceTypedPrivilege', 'OneSourceLinkedPrivilege'.
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      //if (changeSupport_ == null)
        //  changeSupport_ = new PropertyChangeSupport(this);

      //if (_propertyName.equals("AclConnectedResources"))
        //  changeSupport_.addPropertyChangeListener(_propertyName, _l);
      if (_propertyName.equals("OneSourceTypedPrivilege") ||
              _propertyName.equals("OneSourceLinkedPrivilege")) {
                    if (privHandler_ == null ) {
                        try {
                          privHandler_ = new CPrivilegeUserImpl(this);
                        } catch (CreateError c) {
                            new InternalError("From "+name_+"- "+c.toString());
                        }
                    } // if

                    privHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else super.addPropertyChangeListener(_propertyName, _l);

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("EligiblePartyImpl.addPropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the adding of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //System.out.println("listeners of all properties after the adding of "+_l+": ");
                    //l_list = changeSupport_.getPropertyChangeListeners();
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //l_list = null;
                //} // if


  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'AclConnectedResources', 'OneSourcePrivilege'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      //if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("EligiblePartyImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      //if (_propertyName.equals("AclConnectedResources"))
        //  changeSupport_.removePropertyChangeListener(_propertyName, _l);
      if (_propertyName.equals("OneSourceTypedPrivilege") ||
              _propertyName.equals("OneSourceLinkedPrivilege")) {
                    if (privHandler_ != null )
                        privHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else super.removePropertyChangeListener(_propertyName, _l);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("EligiblePartyImpl.getPropertyChangeListeners(),
                    // call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("AclConnectedResources") )
            return changeSupport_.getPropertyChangeListeners(_propertyName);

      else if (_propertyName.equals("OneSourceTypedPrivilege") ||
              _propertyName.equals("OneSourceLinkedPrivilege")) {
            if (privHandler_ != null )
                return privHandler_.getPropertyChangeListeners(_propertyName);
            else
                return null;
      }
      else return super.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()

   /**
   * This method removes the EligibleParty in the ACS list,
   * and in the Resources, deletes any AclEntry having this instance as source.
   * Then, it removes this instance in every group.
   * UserID.finalizeForBase() removes in resources through
   * a variable userID_ and currentUserID_ (for Actors).
   * GroupID.finalizeForBase() removes in resources through
   * the variable groupID_ and, for Actors, currentGroupID_ and l_SecondaryGroup_.
   * GroupID.finalizeForBase() removes also in userIDs the primary group if it is this. 
   * An unused object has to be deleted to inform the others objects,
   * and this method HAS TO BE USED before deleting the eligible party.
   * Calls ACSImpl.removeEorEligibleParty() only if the object is known by the ACS.
   * Called by the ACSImpl.finalizeForBase() and setAcsRelationsInIS(),
   * and by finalizeForUser().
   * This method does not raise PropertyChangeEvents since the BeanInfo for this class
   * specifies that the ACS informs the listeners when this instance is finalized.
   * Caution: The listeners may use the groups list since it is not still deleted.
   * @exception UpDateError if the eligible party is unknown
   * by the main ACS (if it is a clone, typically) or if the ACS of an aclEntry
   *  is closed or unknown, or if a linked immutable object fires an exception,
   *  or from ACSFactoryUtilityImpl.removeAssociatesAfterDeletion().
   */
  public synchronized void finalizeForBase() throws UpDateError {

                //System.out.println("Call of EligiblePartyImpl.finalizeForBase()");

    // noop if it is the second call to this method
    if (folderMember_ == null) return;

    // initialization
    Resource    res = null;
    List        list = new ArrayList( INITIAL_CAPACITY );

                
    // remove from its views TODO
                // l_impliedOpenViews_ = null;
    
    // initialization avoid new ArrayList(m_AclEntriesResources_.values())
            // to have two separate iterator on 'list' and m_AclEntriesResources_
    if (m_AclEntriesResources_ != null) {
      for (Iterator ite = m_AclEntriesResources_.values().iterator() ; ite.hasNext() ;) {
        list.add(ite.next() );
      }
    }

    // removes the recorded aclEntries in the relevant resources,
    // in m_AclEntriesResources_
    if (m_AclEntriesResources_ != null) {
        for (Iterator iter = list.iterator() ; iter.hasNext() ;) {
            res = (Resource)iter.next();
            AclEntry[] aclArray = res.getL_aclEntry();
            if (aclArray != null) {
              for (int i = 0 ; i < aclArray.length ; i++) {
                  if (aclArray[i].getEorEP() == this) {

                      // for testing and tracking
                      //System.out.println("In EligiblePartyImpl.finalizeForBase(): "+
                      //  ", for the ep: "+name_+" , remove the aclEntry for the resource: "+
                      //  res.getName());

                      // indirect call to this.removeAclEntryToEP():
                      res.removeAclEntry(aclArray[i].getEorACS(), this,
                        aclArray[i].getPositiveRight(),
                        aclArray[i].getL_ConditionSourceGroups().iterator().next());
                  }
              } // for i
            } // for if (aclArray != null)
        } //for iter
    }

    // removes (this) to the map of eligible parties in the main ACS.
    // if it is a clone, it doesn't exist for the ACS and UpDateError is thrown.
    // must be done before the call of GroupID.removeEorMember(this), because
    // the list group is used in ACSTreeBaseListener to analyze the ep
    if (aCS_ != null && aCS_.getEorM_EligibleParties().get(name_) != null)
        ((ACSImpl)aCS_).removeEorEligibleParty(this);

    m_AclEntriesResources_ = null;
    //trueEP_ = null;

    super.finalizeForBase();
    
    res = null;
    list = null;
    
                // for testing and tracking        
                //if (type_ == null) throw new InternalError(
                //"in EligiblePartyImpl.finalizeForBase(), ep is finalized");
  } // finalizeForBase()

  /**
   * Clone without register the new object in the ACSs.
   * Caution: don't use any modifying method that may altere the integrity
   * of the base. A finalized instance returns a clone with getEorACS() which returns null,
   * but getName() and getAcsName() operates.
   * @exception CloneNotSupportedException
   * @return not registred in the ACSs. The names, the type, the comment are cloned.
   * The external object references are not cloned (see "Eor" in get method).
   */
  public Object clone() {
      EligiblePartyImpl clon = (EligiblePartyImpl) super.clone();
      //clon.isAlias_ = isAlias_;
      //clon.trueEP_ = trueEP_;

      if (m_AclEntriesResources_ != null)
        clon.m_AclEntriesResources_ = (SortedMap)
          ((TreeMap)m_AclEntriesResources_).clone();

      return clon;
  } // clone()

    /**
     * Does not compare the comment.
     * @param _obj object to compare
     * @return true if _obj is an instance of EligiblePartyImpl
     * (or a derived class) equals to this instance (with the same class)
     */
    public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      EligibleParty obj = (EligibleParty) _obj;

      //if (trueEP_ == null | obj.getEorTrueEP() == null) {
        //  if (trueEP_ != null | obj.getEorTrueEP() != null)
          //  return false;
      //} // if
      //else {
        //  if (!trueEP_.equals( obj.getEorTrueEP() ))
          //  return false;
      //} // else

      if (getEorL_AclConnectedResources() == null || obj.getEorL_AclConnectedResources() == null) {
          if (getEorL_AclConnectedResources() != null || obj.getEorL_AclConnectedResources() != null)
            return false;  // if
      } // if
      else {
        if (!getEorL_AclConnectedResources().equals( obj.getEorL_AclConnectedResources() )) return false;
      }

      return true;
  } // equals()

  /**
   * @since 0.7.0
   * @return the hashcode derived from the names of the ACL-connected resources, 
   * the type and folder member hashcodes
   */
  public int hashCode() {
    int result = super.hashCode();

    if (m_AclEntriesResources_ != null) {
      result = PRIME*result + m_AclEntriesResources_.keySet().hashCode();
    }

    return result;
  } // hashcode()

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }
  
  //
  // ---------------- methods out of the implemented interfaces --------------
  //
  
  /**
  * This method provides a copy of the instance variables.
  * Copy the aclEntries which have different resources ACS
  * than this instance, and copy the others following _withoutIntResAcl.
  * Comments are not copied.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError if _result is null or if Resource.addAclEntry()
  *   throws it. 
  **/
  protected EligiblePartyImpl copyOwnVariables(EligiblePartyImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

        if (_result == null) {
          throw new CreateError("in EligiblePartyImpl.copyOwnVariables(),"+
          " _result is null.");
        }

        super.copyOwnVariables(_result);
      
        if (m_AclEntriesResources_ != null) {
            // attributes
            AclEntry[]            l_aclCopy = null;
            Resource              res = null;

            for (Iterator ite = m_AclEntriesResources_.values().iterator(); ite.hasNext(); ) {
                res = (Resource)ite.next();
                l_aclCopy = res.getL_aclEntry();
                // iterate on all aclEntries of the resource
                for (int i = 0 ; i < l_aclCopy.length ; i++) {
                    if (l_aclCopy[i] == null) continue;
                    if ( l_aclCopy[i].isEmpty() ) continue;

                    // copy the aclEntry with _result as EP
                    if ( l_aclCopy[i].getEorEP() == this ) {
                        // always copies the acl if it is an external one
                        if (res.getEorACS() != getEorACS()) {
                            res.addAclEntry(l_aclCopy[i].getEorACS(), _result,
                              l_aclCopy[i].getL_Rights(), l_aclCopy[i].getPositiveRight(),
                              l_aclCopy[i].getL_ConditionSourceGroups().iterator().next());
                        } // if
                        else if (!_withoutInternalAcl) {
                            res.addAclEntry(l_aclCopy[i].getEorACS(), _result,
                              l_aclCopy[i].getL_Rights(), l_aclCopy[i].getPositiveRight(),
                              l_aclCopy[i].getL_ConditionSourceGroups().iterator().next());
                        } // else if
                    } // if

                } // for i
            } // for ite

            l_aclCopy = null;
            res = null;
        } // if

        return _result;
  } // copyOwnVariables(,,,)
  
   /**
   * CAUTION: any change in this method code MUST also be done in ActorImpl.
   * Called by the ResourceImpl.addAclEntry().
   * There, the AclEntry ACS knowns the AclEntry.
   * @param _acl appplied to this EP. 
   *  The resource connected to _acl must have a name following the pattern
   *  IS_name / ACS_name / resource_name, to allow the unicity
   *  of the name. The aclEntry ACS may be different to
   *  the eligible party ACS.
   */
  protected synchronized void addAclEntryToEP(AclEntry _acl) {
    List l_OldAclEntriesResources = getEorL_AclConnectedResources(); // may be null

    if (_acl.getEorResource() == null) throw new InternalError("add aclEntry"+
      " with aclEntry.getEorResource() == null");

    ImmutableName name = _acl.getEorResource().getName();
    if (m_AclEntriesResources_ == null) m_AclEntriesResources_ = new TreeMap();

    synchronized (m_AclEntriesResources_) {
      if (!m_AclEntriesResources_.containsKey(name)) {

        m_AclEntriesResources_.put( (ImmutableName)name.clone() , _acl.getEorResource());
        
        try {
            firePropertyChange("AclConnectedResources", l_OldAclEntriesResources,
               getEorL_AclConnectedResources());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From EligiblePartyImpl.addAclEntryToEP(), firePropertChange() throws:");
            e.printStackTrace();
        }

      } // if
    } // synchro

    name = null;
    l_OldAclEntriesResources = null;
  } // addAclEntryToEP()

   /**
   * Removes the related resource of an aclEntry, if the argument is the single
   * AclEntry for the resource. This condition is handled by the calling method.
   * Called from AclEntryImpl.finalizeForUser().
   * The related resource is unrecorded in L_AclConnectedResources
   * since _acl is the last one connected to its Resource in this EligibleParty.
   * CAUTION: any update of this method must be also done
   * in ActorImpl and any other EligibleParty direct implementation.
   * @param _acl to remove
   */
  protected void removeAclEntryToEP(AclEntry _acl) {
    List l_OldAclEntriesResources = getEorL_AclConnectedResources(); // may be null

              // for testing and tracking
              //System.out.println("Call of EligiblePartyImpl.removeAclEntryToEP() "+
              //  "in: "+name_+", for the aclEntry: "+_acl.getName());

    synchronized (m_AclEntriesResources_) {
      Object obj = m_AclEntriesResources_.remove( ((AclEntryImpl)_acl).getResName() );

      if (obj == null) throw new InternalError("in EligiblePartyImpl."+
        "removeAclEntryToEP(), null resource or null resource name");

      if (m_AclEntriesResources_.isEmpty()) m_AclEntriesResources_ = null;

      try {
          firePropertyChange("AclConnectedResources", l_OldAclEntriesResources,
            getEorL_AclConnectedResources());
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From EligiblePartyImpl.removeAclEntryToEP(), firePropertChange() throws:");
            e.printStackTrace();
       }

      obj = null;
    } // synchro

    l_OldAclEntriesResources = null;
  } // removeAclEntryToEP()

   /**
   * Removes all the aclEntries.
   * Called from copyWithoutAcl() from this class
   * and each of its subclasses, and from ActorImpl and each of its subclasses.
   * @param _ep is an EligibleParty
   * @return _ep.
   */
  protected final static EligibleParty deleteAclEntries(EligibleParty _ep) {

        // attributes
        AclEntry[]            l_aclCopy = null;
        Resource              res = null;
        List                  l_aclResources =  _ep.getEorL_AclConnectedResources();

        try {
            for (Iterator ite = l_aclResources.iterator(); ite.hasNext(); ) {
                res = (Resource)ite.next();
                l_aclCopy = res.getL_aclEntry(); 
                // iterate on all AclEntries of the resource
                for (int i = 0 ; i < l_aclCopy.length ; i++) {
                    if (l_aclCopy[i] == null) continue;
                    if ( l_aclCopy[i].isEmpty() ) continue;

                    // deletes the AclEntry with result as EP
                    if ( l_aclCopy[i].getEorEP() == _ep ) {
                       res.removeAclEntry(l_aclCopy[i].getEorACS(), _ep,
                              l_aclCopy[i].getPositiveRight(),
                              l_aclCopy[i].getL_ConditionSourceGroups().iterator().next() );

                    } // if

                } // for i
            } // for ite
        } catch (UpDateError u) {
            throw new InternalError("In EligiblePartyImpl.deleteAclEntries(): "+u.toString());
        }
        l_aclCopy = null;
        res = null;
        l_aclResources = null;

        return _ep;
  } // deleteAclEntries()


  /**
   * NO USE.
  * This method finalizes all the attributes, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null.
  * Does not call finalizeForBase(). Called only by EPRViewInBaseImpl.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In EligiblePartyImpl, call of " + name_ +
                    //".finalizeFromDeserialization()");
    if (folderMember_ == null) return;

    if (m_AclEntriesResources_ != null)
        m_AclEntriesResources_.clear();
    m_AclEntriesResources_ = null;
    
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

} // EligiblePartyImpl

