//Title:        Access Road
//Version:      0.7.2
//Copyright:    2007, 2010, 2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrW.DialogError;

import java.util.Collections;
import java.util.List;
import java.util.ArrayList;
import java.util.SortedMap;
import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeSupport;

/**
 * This abstract class is responsible for managing a folder of BaseObjects
 * (even if most of the code is adapted to ACSObjects folder), where the members
 * are added and removed from the container _without_ object deleting.
 * This is the main differene with a Node which deletes its children
 * when they are removed from it.
 * It is a javabean with the following bound properties:
 * 'OneBaseObjectInFolder' (when one base object is added or removed),
 * 'Comment'.
 * The subclasses have different managers on which they are registered:
 * - BaseManagerImpl for ViewInBaseImpl,
 * - ACSImpl for VirtualFolderImpl. 
 * The listeners are proper to the instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order.
 * Excepted for 'OneBaseObjectInFolder', the listeners receive only a copy 
 * of the new value to protect the property. All the exceptions
 * from the listeners are catched, and a dialog box is displayed to inform the user.
 * Caution: no equals() nor clone() methods.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 */
public abstract class FolderAbst implements Folder, Serializable {
  /* protected attributes */
    /*
     * The default value is '<undefined>'. To be defined by the subclass.
     */
  protected String            type_ = "<undefined>";
  protected static final int  PRIME = 1000003;
  protected static final int  INITIAL_CAPACITY = 20; // default value for map: 101
  protected static final int  INITIAL_CAPACITY_5 = 5;
  protected ImmutableName     name_;
  protected BaseError         baseError_;
  protected transient SortedMap<ImmutableName, BaseObject>  m_BaseObjects_;// open objects only

  /* private attributes*/
  private static final long   serialVersionUID = 0;
  protected String              comment_;
  /*
   * True if the objects in the folder are from several ACS.
   * Always false for a VirtualFolder.
   */
  protected boolean             connectedToSeveralACS_ = false;
  /**
   * The connected ACS are the ACS of the BaseObjects, and it is empty
   * if it is a view of non-ACSObjects like 'Architecture'.
   * Transient list wich is filled, by the constructor and at the opening of the folder
   * from the items in m_ObjectsNames_, by the method prepareAfterRestoring().
   **/
  protected transient List<ImmutableName>      l_connectedACSNames_;
  /**
  * Used in the folder saving, to avoid some links to the ACSObjects, and then
  * to not include the connected ACSs in the folder saving. Used only to save and restore,
  * and by BaseManagerImpl.getElementsOfOneView() directly without updating.
  * No initializing at the instance creation, so the default value is null.
  * Is correct only after the call of prepareBeforeSaving().
  * The key is the BaseObject name, and the value is the ACS name if it is an ACSObject.
   * */
  protected SortedMap<ImmutableName, ImmutableName>           m_ObjectsNames_ = null;

   /**
   * manages all the property change listeners of this classes and its subclasses
   */
  protected transient PropertyChangeSupport changeSupport_ = null;

  /**
   * variable for the JDK 2 serialization.
   * m_BaseObjects_ and l_connectedACSNames_ are not persistent, 
   * and they are updated by prepareAfterRestoring().
   * @serial ("serialVersionUID",Long.TYPE),
   * ("baseError_",BaseError.class),
   * ("name_",NameImpl.class),
   * ("type_",String.class),
   * ("comment_",String.class),
   * ("connectedToSeveralACS_",Boolean.TYPE),
   * ("m_ObjectsNames_",SortedMap.class)
   */
    public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("serialVersionUID",Long.TYPE),
      new ObjectStreamField("baseError_",BaseError.class),
      new ObjectStreamField("name_",NameImpl.class),
      new ObjectStreamField("type_",String.class),
      new ObjectStreamField("comment_",String.class),
      new ObjectStreamField("connectedToSeveralACS_",Boolean.TYPE),
      new ObjectStreamField("m_ObjectsNames_",SortedMap.class)
  };

 /**
  * Default constructor
  */
  protected FolderAbst() {
      name_ = (ImmutableName) new NameImpl();
      baseError_ = new BaseError();
      m_BaseObjects_ = new TreeMap();
      comment_ = "";
      l_connectedACSNames_ = new ArrayList(INITIAL_CAPACITY_5);
      changeSupport_ = new PropertyChangeSupport(this);
  }

  /**
   * @param _name is copied.
   * @exception CreateError if the folder name is null,
   * or if it contains a character that is not a letter
   */
  protected FolderAbst(ImmutableName _name) throws CreateError {
      if (_name == null) {
        throw new CreateError("The name of the new folder is null.");
      } //if
      char[] l_chars;
      boolean containsFaultyValue = false;
      for (int j = 0; j < _name.noOfComponents(); j++) {
          l_chars = _name.component(j).toCharArray();
          for (int i = 0; i < l_chars.length; i++) {
            if (Character.isIdentifierIgnorable(l_chars[i]) ||
             Character.isISOControl(l_chars[i])) {
                  containsFaultyValue = true;
                  System.out.println("in FolderAbst constructor, a character "+
                        "is ignorable or is an ISO control: "+l_chars[i]+" end.");
                  break;
            }
          } // for i
          if (containsFaultyValue) break;
      } // for j
      if (containsFaultyValue) {
        l_chars = null;
        throw new CreateError("The name of the new folder contains a character "+
          "that is not a letter.");
      } //if
      l_chars = null;

      name_ = (ImmutableName) _name.clone();
      baseError_ = new BaseError();
      m_BaseObjects_ = new TreeMap();
      comment_ = "";
      l_connectedACSNames_ = new ArrayList(INITIAL_CAPACITY_5);
      changeSupport_ = new PropertyChangeSupport(this);
  } // constructor

  /**
  * This constructor provides a FolderAbst instance from a map of base objects.
  * Called by ViewInBaseImpl and VirtualFolderImpl.
  * l_connectedACSNames_ is filled only by ACSObject ACS names,
  * so it does not work for Architecture views with only ACSs as elements.
  * @param _name ImmutableName of the folder. It is copied.
  * @param _m_BaseObjects is the map (name, object) of the folder objects.
  *     It is copied. May be null.
  * @exception CreateError if _m_BaseObjects contains a null,
  *     no BaseObject, or empty-name object.
  */
  protected FolderAbst(ImmutableName _name, Map _m_BaseObjects) throws CreateError {

      this(_name);

      // updates m_BaseObjects_
      Object obj = null;
      boolean nok = false;
      boolean nok2 = false;
      
      if (_m_BaseObjects != null) {
          for (Iterator ite = _m_BaseObjects.values().iterator() ; ite.hasNext() ; ) {
            obj = ite.next();
            if (!(obj instanceof BaseObject) ) {
                nok = true;
                break;
            } // if
            if (((BaseObject)obj).getName().isEmpty() ) {
                nok2 = true;
                break;
            } // if

            // caution: these commands must also be in prepareDataAfterRestoring()
            m_BaseObjects_.put( ((BaseObject)obj).getName(), (BaseObject)obj);

            // updates l_connectedACSNames_ and connectedToSeveralACS_ if necessary
            if (obj instanceof ACSObject) {
              if (!l_connectedACSNames_.contains( ((ACSObject)obj).getAcsName() ) )
                l_connectedACSNames_.add( ((ACSObject)obj).getAcsName() );
            } // if (_obj instanceof ACSObject)
            // ... and noop for Architecture views with ACS as elements
          } // for ite
          
          if (l_connectedACSNames_.size() > 1) connectedToSeveralACS_ = true;

                                        // for testing and tracking
                                        //System.out.println("FolderAbs constructor of "+
                                          //name_+" calls to addOwnListening()");
          
          // registers for ACSObject deletions in the ACS.
          // faster than adddEorBaseObject().
          addOwnListening();
      } // if (_m_BaseObjects != null)
      
      if (nok) {
          name_ = null;
          m_BaseObjects_.clear();
          m_BaseObjects_ = null;
          throw new CreateError("in the FolderAbst constructor, "+obj+
            " is not a BaseObject.");
      } // if
      if (nok2) {
          name_ = null;
          m_BaseObjects_.clear();
          m_BaseObjects_ = null;
          throw new CreateError("in the FolderAbst constructor, the object "+obj+
            " is empty.");
      } // if
  } // constructor with _m_BaseObjects

  // --------- methods of the implemented interfaces ------------------

  /**
   * @return long the serialVersionUID used by the serialization process.
   */
  public synchronized long getSerialVersionUID() {
    return serialVersionUID;
  }

  /**
   * @return ImmutableName of this folder. Not null after the finalization.
   */
  public final ImmutableName getName() {
    if (name_ == null) return null;
    else return (ImmutableName)name_.clone();
  }

  /**
   * The full name is unique for the Access Road program.
   * Its is equal to the ACS name.
   * @return a copy of the name. Should be null after the finalization.
   */
    public ImmutableName getFullName() {
       return getName();
    }
    
  /**
   * This short name is NOT unique for the Access Road program nor the ACS if
   * it is an ACS object.
   * It is often the last component name. Overridden in ViewInBaseImpl.
   * @return the nick name. May be null.
   */
    public String getNickName() {
       if (name_ == null) return null;
       else return name_.lastComponent();
    }  
  
  /**
   * @return an External Object Reference to the baseError instance
   */
  public BaseError getEorBaseError() {
    return baseError_;
  }

  /**
   * The type is defined at the creation of the object, by a subclass.
   * For VirtualFolders only, it is used like a Basic type.
   * setType() is defined in VirtualFolderImpl and fires a 'VirtualFolderType' event.
   * For views, there is one immutable type per class.
   * @return type of the folder
   */
  public String getType(){
    return type_;
  }

  /**
   * Always false when the elements are not ACSObjects, like
  * in an Architecture view, or for a VirtualFolder.
  * @return boolean true if the folder is connected to ACSObjects
  * belonging to different ACS.
  */
  public boolean getConnectedToSeveralACS() {
    return connectedToSeveralACS_;
  }

  /**
  * @return List of names of the ACS having ACS Objects displayed in the folder.
  * Never null.
  */
  public List<ImmutableName> getEorL_ConnectedACSNames() {
      return (List)((ArrayList)l_connectedACSNames_).clone();
  }

  /**
   * Caution: an empty folder is not equals to a new FolderAbst() instance,
   * since it may have an non-null name and other non-null properties.
  * @return boolean true if equal to a folder wich does not contain objects.
  */
  public boolean isEmpty() {
    if (m_BaseObjects_ == null || m_BaseObjects_.isEmpty())
        return true;
    else
        return false;
  }
  
 /**
 * @return int derived from the name hashcode.
 */
 public final int hashCode() {
   return "FOLDERABST".hashCode() + name_.hashCode();
 } // hashCode()

  /**
   * Tests if a BaseObject is within the folder.
   * @param _obj BaseObject to test
   * @return boolean true if the folder references this instance
   */
  public boolean containsBaseObject(BaseObject _obj){
    if (_obj == null) return false;
    if (m_BaseObjects_ == null) return false;
        //System.out.println("in FolderAbst.containsBaseObject(), "+
          //"m_BaseObjects_ is null");
    //else if (_obj != null) System.out.println(
      //"in FolderAbst.containsBaseObject() _obj is:"+_obj);
    return m_BaseObjects_.containsKey(_obj.getName() );
  } // containsBaseObject()

  /**
   * Gets all the base objects displayed in this folder.
   * @return unmodifiableSortedMap of the folder BaseObjects. Never null.
   */
  public SortedMap<ImmutableName, BaseObject> getEorM_BaseObjects() {

                                // for testing and tracking
                              //System.out.println("FolderAbst.getEorM_BaseObjects() for "+
                                //            getName()+" with the BaseObjects: ");
                              //if (m_BaseObjects_ != null) {
                                //  for (Iterator ite = m_BaseObjects_.entrySet().iterator(); ite.hasNext();) {
                                  //          System.out.println("------- "+
                                    //                ((Map.Entry)ite.next()).getKey() );
                                  //} // for ite
                              //}
                              //else {
                                //  System.out.println("----- null map ");
                              //}

    return Collections.unmodifiableSortedMap(m_BaseObjects_);
  } // getEorL_BaseObjects()

  /**
   * Adds a base object to the folder.
   * Fires the event 'OneBaseObjectInFolder'. If the argument is from a new ACS,
   * registers this instance as listener of 'OneBaseObject' from this ACS, and fires
   * the change property 'ConnectedACS'.
   * Caution: this instance is not registered in _obj, and this is done by the subclasses.
   * Overridden method.
   * @param _obj non-null BaseObject to add in the folder
   * @exception CreateError if the name is null, or the object name is null or empty,
   * or if the name is already known by the folder.
   */
  public synchronized void addEorBaseObject(BaseObject _obj) throws CreateError {
    // exception
    if (_obj == null || _obj.getName() == null || _obj.getName().isEmpty())
        throw new CreateError("in FolderAbst.addEorBaseObject()"+
          " the name is null, or the object name is null or empty.");
    if (containsBaseObject(_obj) )
      throw new CreateError("in FolderAbst.addEorBaseObject()"+
        " the object name is already known.");

                            // for testing and tracking
                            //System.out.println("call to FolderAbst.addEorBaseObject(), "+
                              //"in "+name_+" for adding "+_obj);

    // updates m_BaseObjects_
    m_BaseObjects_.put(_obj.getName(),_obj);

    // updates l_connectedACSNames_ and connectedToSeveralACS_ if necessary,
    // and registers this instance to the ACS if necessary
    List l_oldValue = null;
    ACS acs = null;
    if (_obj instanceof ACSObject) {
      if (!l_connectedACSNames_.contains( ((ACSObject)_obj).getAcsName() )) {

        l_oldValue = (List) ((ArrayList)l_connectedACSNames_).clone();
        // adds a name
        l_connectedACSNames_.add( ((ACSObject)_obj).getAcsName() );

        if (l_connectedACSNames_.size() > 1) connectedToSeveralACS_ = true;

        // registers in the ACS
        acs = (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().
            get(((ACSObject)_obj).getAcsName());
        if (!acs.isPropertyChangeListener("OneBaseObject", this))
            acs.addPropertyChangeListener("OneBaseObject", this);

        try {
            firePropertyChange("ConnectedACS", l_oldValue,
               ((ArrayList)l_connectedACSNames_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From FolderAbst.addEorBaseObject(), firePropertChange() throws:");
            e.printStackTrace();
        }

      } // if
    } // if (_obj instanceof ACSObject) {

    try {
        firePropertyChange("OneBaseObjectInFolder", null, _obj);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From FolderAbst.addEorBaseObject(), firePropertChange() throws:");
        e.printStackTrace();
    }
    l_oldValue = null;
    acs = null;
  } // addEorBaseObject()

  /**
   * Removes a base object to the folder.
   * Fires the events 'OneBaseObjectInFolder'.
   * Overridden method.
   * @param _obj non-null BaseObject to remove in the folder
   * @exception CreateError if the object name is empty, or
   * the object is not known by the folder.
   */
  public synchronized void removeEorBaseObject(BaseObject _obj) throws CreateError {
    if (baseError_ == null) {
              // for testing purpose
              //System.out.println("In FolderAbst.removeEorBaseObject(), for "+
                //      getName()+", second call to finalization.");

      return;
    }
    boolean toKeep = false;
    Object obj = null;

                    // for testing and tracking
                    //if (m_BaseObjects_ != null) {
                      //  System.out.println("In FolderAbst.removeEorBaseObject(),"+
                        //  " the base objects before the removing are not null:");
                        //for (Iterator ite = m_BaseObjects_.values().iterator() ; ite.hasNext() ; ) {
                          //System.out.println("--- "+ ((BaseObject)ite.next()).toString() );
                        //}
                    //} // if

    // exception
    if (_obj == null || _obj.getName() == null ||
      _obj.getName().isEmpty() )
        throw new CreateError("in FolderAbst.removeEorBaseObject()"+
          "the object is null, or its name is null or empty.");
    if (!containsBaseObject(_obj) )
      throw new CreateError("in FolderAbst.removeEorBaseObject()"+
        "the object name is unknown: "+_obj.getName());


                    // for testing and tracking
                    //System.out.println("In FolderAbst.removeEorBaseObject(),"+
                      //" the base objects before the removing are:");
                    //for (Iterator ite = getEorM_BaseObjects().values().iterator() ; ite.hasNext() ; ) {
                      //System.out.println("--- "+ ((BaseObject)ite.next()).toString() );
                    //}

    // updates m_BaseObjects_
    m_BaseObjects_.remove(_obj.getName() );

    // updates l_connectedACSNames_ and connectedToSeveralACS_ if necessary
    List l_oldValue;
    if (_obj instanceof ACSObject) {
      for (Iterator ite = m_BaseObjects_.values().iterator() ; ite.hasNext() ; ) {
        obj = ite.next();
        if (obj instanceof ACSObject &&
          ((ACSObject)_obj).getAcsName().equals(((ACSObject)obj).getAcsName()) ) {
            toKeep = true;
            break;
        } // if
      } // for ite

      if (!toKeep) {
        l_oldValue = (List) ((ArrayList)l_connectedACSNames_).clone();
        l_connectedACSNames_.remove( ((ACSObject)_obj).getAcsName() );
        if (l_connectedACSNames_.size() < 2) connectedToSeveralACS_ = false;
        // unregister in the ACS
        if (BaseManagerImpl.getInstance().getEorM_OpenACS().
                get(((ACSObject)_obj).getAcsName()) != null)
            ((ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().
                get(((ACSObject)_obj).getAcsName())).
                removePropertyChangeListener("OneBaseObject", this);

        try {
            firePropertyChange("ConnectedACS", l_oldValue,
              ((ArrayList)l_connectedACSNames_).clone());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("FolderAbst.removeEorBaseObject(), firePropertChange() throws:");
            e.printStackTrace();
        }

        l_oldValue = null;
      } // if (!notLastOne)
    } // if (_obj instanceof ACSObject)
    
                    // for testing and tracking
                    //System.out.println("In FolderAbst.removeEorBaseObject(),"+
                      //" fires 'OneBaseObjectInFolder' for "+_obj);
                            //" the base objects after the removing are:");
                    //for (Iterator ite = getEorM_BaseObjects().values().iterator() ; ite.hasNext() ; ) {
                      //System.out.println("--- "+ ((BaseObject)ite.next()).toString() );
                    //}

    try {
        firePropertyChange("OneBaseObjectInFolder", _obj, null);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From FolderAbst.removeEorBaseObject(), firePropertChange() throws:");
        e.printStackTrace();
    }
    obj = null;
  } //removeEorBaseObject()

  /**
   * Observes the changes in the base objects list from the ACS.
   * This is for the 'OneBaseObject' bound property, to remove a folder object.
   * Overridden method that is called by all the overridding methods.
   * @param _evt PropertyChangeEvent
   */
  public void propertyChange(PropertyChangeEvent _evt){

    if (_evt.getSource() instanceof ACS &&
      _evt.getPropertyName().equals("OneBaseObject")) {

          // does not remove if it is an external controlled object,
          // for example through an acl which is deleted
          if (containsBaseObject( (BaseObject)_evt.getOldValue()) &&
            ((ACSObject)_evt.getOldValue()).getAcsName().equals(
              ((ACS)_evt.getSource()).getName() ) 
          ) {
            try {

                    // for testing and tracking
                    //System.out.println("In FolderAbst.propertyChange(),"
                     //+" remove and object.");

              // this method fires some events (including 'OneBaseObjectInFolder')
                removeEorBaseObject( (BaseObject)_evt.getOldValue() );

            } catch (CreateError b) {
              b.printStackTrace();
            }
          } // if

    } // if (_evt.getSource() instanceof ACS
  } // propertyChange()

  /**
  * Sets any comment related to this folder. It is a bound property.
  * @param _s comment which is associated to the folder.
  */
  public synchronized void setComment(String _s) {
        String oldComment = comment_;
        comment_ = _s;

                    // for testing and tracking
                    //System.out.println("In FolderAbst.setComment(),"
                      //+" new comment: "+comment_);
        
        try {
              firePropertyChange("Comment",oldComment,comment_);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From FolderAbst.setComment(), firePropertChange() throws:");
            e.printStackTrace();
        }
        oldComment = null;
  }

  /**
  * @return String: comment which is associated to the folder. May be null.
  */
  public String getComment() {
        return comment_;
  }

  /**
  * Adds a listener to the bean for the porperties 'OneBaseObjectInFolder',
  * 'ConnectedACS', 'Comment'.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public synchronized void addPropertyChangeListener(PropertyChangeListener _l){

    if (changeSupport_ == null) {
      changeSupport_= new PropertyChangeSupport(this);
    }

    synchronized (changeSupport_) {changeSupport_.addPropertyChangeListener(_l);}
  }

  /**
  * Add a listener to the bean for the properties 'OneBaseObjectInFolder',
  * 'ConnectedACS', 'Comment'.
  * @param _propertyName is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

    if (changeSupport_ == null) {
      changeSupport_= new PropertyChangeSupport(this);
    }

    synchronized (changeSupport_) {
        changeSupport_.addPropertyChangeListener(_propertyName, _l);
    }
  }

  /**
  * remove a listener to the bean for the properties 'OneBaseObjectInFolder',
  * 'ConnectedACS', 'Comment'.
  * @param _l PropertyChangeListener to remove
  */
  public synchronized void removePropertyChangeListener(PropertyChangeListener _l){

    if (changeSupport_ == null) return;

    synchronized (changeSupport_) {changeSupport_.removePropertyChangeListener(_l);}
  }

  /**
  * remove a listener to the bean for the properties 'OneBaseObjectInFolder',
  * 'ConnectedACS', 'Comment'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

    if (changeSupport_ == null) return;

    synchronized (changeSupport_) {
        changeSupport_.removePropertyChangeListener(_propertyName, _l);
    }

  } // removePropertyChangeListener()

  /**
   * no equals: each instance is equals only to itself.
   * @param _obj: object to compare
   * @return true if the two objects are identical instances
   * with the same name, not the same BaseObjects list
   */
   /*public synchronized boolean equals(Object _obj) {
    if (this == _obj)
        return true;
    if (_obj == null)
        return false;
    if (_obj.getClass() != getClass())
        return false;

    FolderAbst obj = (FolderAbst) _obj;

    if (name_ == null || obj.getName() == null) {
        if (name_ != null || obj.getName() != null)
          return false;  // if
    } // if
    else {
      if (!name_.equals( obj.getName() )) return false; }

    //if (!baseError_.equals(obj.getEorBaseError()) ) return false;
    
    if (m_BaseObjects_ == null || obj.getEorM_BaseObjects() == null) {
      if (m_BaseObjects_ != null || obj.getEorM_BaseObjects() != null)
        return false;  // if
    } // if
    else {
      if (!m_BaseObjects_.keySet().equals( obj.getEorM_BaseObjects().keySet() ))
        return false; }
    
    
    return true;
  } //equals()
  */
  
  /**
   * The name is the only property which is used to compute the hashcode,
   * because after finalizeForBase(), hashcode() is used in BaseManagerImpl Maps.
   * @return the hashcode value for the folder, based on its name.
   */
  /*public int hashCode() {
    int result = 0;
    if (name_ != null) {
      result = name_.hashCode();
    }
    //if (m_BaseObjects_ != null) {
    //  result = PRIME*result + m_BaseObjects_.keySet().hashCode();
    //}
    return result;
  } // hashcode()
  */
   
  /**
  * @return derived from the nick name
  */
  public String toString() {

      StringBuilder sb = new StringBuilder();
      sb.append(getNickName() );
      /*
      sb.append("  FOLDERABST.TOSTRING():");

      sb.append(" name_ = ");
      sb.append(name_);

      sb.append("; size of the objects map = ");
      sb.append(m_BaseObjects_.size() );

      return " [" + sb + "] ";
      */
      return " " + sb;


/*
      StringBuilder sb = new StringBuilder();

      sb.append("  FOLDERABST.TOSTRING():");

      sb.append(" name_ = ");
      sb.append(name_);

      sb.append("; size of the objects map = ");
      sb.append(m_BaseObjects_.size() );

      sb.append("------- start of objects map");
      for (Iterator ite = m_BaseObjects_.values().iterator() ; ite.hasNext() ; ) {
        sb.append( ((BaseObject)ite.next()).toString() );
      }
      sb.append("------- end of connected ACS list");

      sb.append("------- start of objects map");
      for (Iterator ite = l_connectedACSNames_.iterator() ; ite.hasNext() ; ) {
        sb.append( ((Name)ite.next()).toString() );
      }
      sb.append("------- end of connected ACS list");
 
      sb.append("; comment_ = ");
      sb.append(comment_);

      return " [" + sb + "] ";
*/
  } // toString()

  /**
  * The folder is removed as an ACS listener.
  * "OneBaseObjectInFolder" events are not fired to listeners.
  * All properties are set to null, but the name is not cleared.
  * Called by the subclass methods.
  * @exception UpDateError is not thrown
  */
  public synchronized void finalizeForBase() throws UpDateError {

      // noop if it is the second call to this method
    if (baseError_ == null) {
              // for testing purpose
              //System.out.println("In ViewInBaseImpl.finalizeForBase(), for "+
                //      getName()+", second call to finalization.");

      return;
    }

                            // for testing and tracking
                            //System.out.println("call to FolderAbst.finalizeForBase()");

    // removes as a listener (never as a listened instance)
    removeOwnListening();

    //name_ = null;
    m_BaseObjects_.clear();
    l_connectedACSNames_.clear();
    m_BaseObjects_ = null;
    l_connectedACSNames_ = null;
    baseError_ = null;
    connectedToSeveralACS_ = false;
    comment_ = null;
    m_ObjectsNames_ = null;
  } // finalizeForBase()

  // ------------ Methods out of the extended interfaces --------------------


  /**
  * Fire an event to every registered listener, in any order,
  * for a bound property.
  * Called by add/removeEorBaseObject() methods and the subclasses.
  * For the properties 'OneBaseObjectInFolder', 'ConnectedACS',
  * 'Comment' in this class, and 'OneMoreCriteria', 'OneLessCriteria'
  * in the subclass NoThanViewInBase.
  * There is a method with another arguments in the subclass EPRViewInBase.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and if the two values are not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {
      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //System.out.println("FolderAbs.firePropertyChange(), call to "+
                  //"propertyChange() of "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

  /**
  * The folder is NOT removed as an ACS listener.
  * "OneBaseObjectInFolder" change is NOT fired to listeners.
  * m_BaseObjects_ is cleared whitout finalizing the objects, since they are
  * from open ACS (see prepareAfterRestoring()).
  * All properties are set to null, INCLUDING the name.
  * Overridden by EPRViewInBaseImpl and VirtualFolderImpl.
  * Called by BaseManagerImpl.openView() when prepareAfterRestoring() throws an exception,
  * and by getElementsOfOneView() in all cases.
  */
  protected synchronized void finalizeFromDeserialization() {
    type_ = null;
    name_ = null;
    baseError_ = null;
    m_BaseObjects_.clear();
    m_BaseObjects_ = null;
    comment_ = null;
  } // finalizeFromDeserialization()

  /**
  * The folder is removed as an ACS listener.
  * "OneBaseObjectInFolder" change is NOT fired to listeners.
  * All properties are set to null, including the name.
  * Overridden in ViewInBase subclass, not in VirtualFolder.
  */
  protected synchronized void finalizeForCopy() {
      type_ = null;
      name_ = null;
      baseError_ = null;
      
      // registers for ACSObject deletions in the ACS
      Object obj = null;
      ACS acs = null;
      for (Iterator ite = m_BaseObjects_.values().iterator() ; ite.hasNext() ; ) {
        obj = ite.next();
        if (obj instanceof ACSObject) {
            acs = (ACS) BaseManagerImpl.getInstance().getEorM_OpenACS().get(
             ((ACSObject)obj).getAcsName() );
            acs.removePropertyChangeListener("OneBaseObject", this);
        } // if
      } // for
      acs = null;
      obj = null;
      
      m_BaseObjects_.clear();
      m_BaseObjects_ = null;
      comment_ = null;
  } // finalizeForCopy()

  /**
   * This method removes the object from the application space,
   * but there is no guarantee that the JVM calls this method before
   * the end of the application.
   * @exception  Throwable
   */
  protected void finalize() throws Throwable {
    finalizeForBase();
    super.finalize();
  }

  /**
  * This method registers the instance as a property event listener of
  * 'OneBaseObject' on the ACS to detect the deleting of a folder object.
  * Also done in addEorBaseObject().
  * This method is necessary because the subclasses add some complementary
  * processings.
  * Called by the constructors, copy(), prepareAfterRestoring() methods,
  * and by the identical subclass methods.
  */
  protected void addOwnListening() {
      if (l_connectedACSNames_ != null && !l_connectedACSNames_.isEmpty()) {
          // register for base object deletions in the ACS
          ACS acs;
          for (Iterator ite = l_connectedACSNames_.iterator() ; ite.hasNext() ; ) {
            acs = (ACS) BaseManagerImpl.getInstance().getEorM_OpenACS().get(
                    (ImmutableName)ite.next());

                                        // for testing and tracking
                                        //System.out.println("FolderAbs.addOwnListening() of "+
                                          //name_+" calls to "+acs.getNickName()+
                                          //".addPropertyChangeListener() for 'OneBaseObject' ");

            if (!acs.isPropertyChangeListener("OneBaseObject", this))
                acs.addPropertyChangeListener("OneBaseObject", this);
          } // for ite
          acs = null;
      } // if (_m_BaseObjects != null

  } // addOwnListening

  /**
  * This method unregister the instance as a property event listener of
  * 'OneBaseObject' on the relevant ACS.
  * Called by finalizeForBase() and the subclasses.
  */
  protected void removeOwnListening() {

      if (l_connectedACSNames_ != null && !l_connectedACSNames_.isEmpty()) {
          // register for base object deletions in the ACS
          ACS acs;
          for (Iterator ite = l_connectedACSNames_.iterator() ; ite.hasNext() ; ) {
            acs = (ACS) BaseManagerImpl.getInstance().getEorM_OpenACS().get(ite.next());
            if (acs != null)
                acs.removePropertyChangeListener("OneBaseObject", this);
          }
          acs = null;
      } // if (_m_BaseObjects != null

  } // removeOwnListening

  /**
   * This method must be called by BaseManagerImpl before the saving of the folder,
   * to update m_ObjectsNames_ from m_BaseObjects_ which is not saved nor updated.
   * It is not necessary to clear m_BaseObjects_ since it is transient. According
   * to the same way, the subclasses maps are not cleared,
   * since they are derived from m_BaseObjects_ and also transient.
   * Overridden by NoThanViewInBase which calls this method.
   */
  protected void prepareBeforeSaving() {
    ImmutableName   name = null;
    Map.Entry       entry = null;
    if (m_ObjectsNames_ != null) m_ObjectsNames_.clear();
    else m_ObjectsNames_ = new TreeMap();

                                        // for testing and tracking
                                        //System.out.println("FolderAbs.prepareBeforeSaving() of "+
                                          //name_+" for:");

    for (Iterator ite = m_BaseObjects_.entrySet().iterator(); ite.hasNext();) {
      entry = (Map.Entry)ite.next();
      name = (ImmutableName)((Name)entry.getKey()).clone();

                                        // for testing and tracking
                                        //System.out.println("------- "+name);

      // does not check if obj is a view or an acs
      if (entry.getValue() instanceof ACSObject)
          m_ObjectsNames_.put(name, ((ACSObject)entry.getValue()).getAcsName());
      else
          m_ObjectsNames_.put(name, null);
    } // for ite

    name = null;
    entry = null;
  }  // prepareBeforeSaving()

  /**
   * This method is called by BaseManagerImpl after the opening of the folder,
   * to update m_BaseObjects_ and l_connectedACSNames_ from m_ObjectsNames_,
   * and to check the open ACS and their objects which are in the restored view.
   * The folder has to be closed if it contains an object from a closed or removed ACS.
   * It is not necessary to call this method when ACSImpl.closeVirtualFolders() has been called
   * to save the ACS, since the workable properties of the VirtualFolders remains
   * usable after closeVirtualFolders().
   * Noop if the folder objects has been restored or has never been saved,
   * having a non-empty BaseObjects map.
   * Overridden method, but called by the subclasses at the first line.
   * @exception DialogError if the folder contains an object from a closed or removed ACS,
   *    before the registering of the folder as listener;
   *    the exception message should be shown to the user, and this instance has to be
   *    finalized through finalizeFromDeserialization().
   * @return a non-null message if some objects has been removed from the folder,
   *    or null.
   */
  protected String prepareAfterRestoring() throws DialogError {
    // noop if there is no saving, or if the folder objects has been restored
    if (m_ObjectsNames_ == null) {

                                        // for testing and tracking
                                        //System.out.println("FolderAbs.prepareAfterRestoring() of "+
                                          //name_+"- noop: null m_ObjectsNames_");

        return null;
    }
    if (m_ObjectsNames_.isEmpty()) {
        
                                        // for testing and tracking
                                        //System.out.println("FolderAbs.prepareAfterRestoring() of "+
                                          //name_+"- noop: empty m_ObjectsNames_");

        m_BaseObjects_ = new TreeMap();
        return null;
    }
    if (m_BaseObjects_ != null && !m_BaseObjects_.isEmpty()) {

                                        // for testing and tracking
                                        //System.out.println("FolderAbs.prepareAfterRestoring() of "+
                                          //name_+"- noop: non-empty m_BaseObjects_");

        return null;
    }




      // updates m_BaseObjects_ and l_connectedACSNames_
    String result = prepareDataAfterRestoring();

                                        // for testing and tracking
                                        //System.out.println("FolderAbs.prepareAfterRestoring() of "+
                                          //name_+"- calls to addOwnListening()");

    // registers as listener on BaseManagerImpl events.
    addOwnListening();

    return result;
  }  // prepareAfterRestoring()

  /**
   * This method updates m_BaseObjects_ and l_connectedACSNames_ from m_ObjectsNames_,
   * and checks up the open ACS and their objects which are in the restored view.
   * The folder has to be closed if it contains an object from a closed or removed ACS. <P>
   * If this method is called by BaseManagerImpl.getElementsOfOneView() for a closed view,
   * it is used for updating m_BaseObjects_ and l_connectedACSNames_,
   * then to finalize the instance with finalizeFromDeserialization().
   * Clears m_BaseObjects_ if it is not null.
   * Called by prepareAfterRestoring() and BaseManagerImpl.getElementsOfOneView().
   * @exception DialogError if the folder contains an object from a closed or removed ACS;
   *    the exception message should be shown to the user, and this instance has to be
   *    finalized through finalizeFromDeserialization().
   * @return a non-null message if some objects has been removed from the folder, or null.
   */
  protected String prepareDataAfterRestoring() throws DialogError {
    // init
    if (m_BaseObjects_ != null) m_BaseObjects_.clear();
    else m_BaseObjects_ = new TreeMap();
    l_connectedACSNames_ = new ArrayList(INITIAL_CAPACITY_5);

                              // for testing and tracking
                              //System.out.println("call to FolderAbst.prepareDataAfterRestoring(),"+
                                //      " for "+getName());

    BaseObject      obj = null;
    ImmutableName   name = null;
    Map.Entry       entry = null;
    StringBuilder    result = new StringBuilder();
    ImmutableName   acsname = null;
    ImmutableName   error_acsname = null;
    ImmutableName   error_name = null;

    for (Iterator ite = m_ObjectsNames_.entrySet().iterator(); ite.hasNext();) {
      entry = (Map.Entry)ite.next();
      name = (ImmutableName)entry.getKey();
      obj = null;
      try {
          // obj is from the open ACS, not from a deserialized version of an ACS
          obj = BaseUtilityImpl.getEorBaseObject(name);
      } catch (BaseError b) {
          if (entry.getValue() != null) {
              if (BaseManagerImpl.getInstance().isOpenACS((ImmutableName)entry.getValue()) ) {
                  // if it is a deleted object from an open ACS, the view opening is running,
                  // but a message is returned
                  result.append("The view object '"+name+"' has been deleted in its ACS.");
                  // never added to the view
                  continue;
              } // if
              else {
                    // DialogError to throw, but added to the view: put in m_BaseObjects_
                    // and l_connectedACSNames_

                              // for testing and tracking
                              //System.out.println("In FolderAbst.prepareDataAfterRestoring(), for "+
                                //      getName()+", the acsname "+entry.getValue()+
                                  //    " is not from an opened acs for: "+name);

                  acsname = BaseUtilityImpl.getFromKownACS(name);
                  if (acsname == null) {
                      // it is from a deleted ACS
                      error_name = name;
                      error_acsname = null;
                  } // if
                  else {
                      // it is from a closed ACS
                      error_name = name;
                      error_acsname = acsname;
                  }
                  // m_BaseObjects_ does not contain this value of name,
                  // and BaseManagerImpl will call finalizeFromDeserialization()
                  // for clearing the current m_BaseObjects_
                  break;
              } // else
          } // if
          // for the Architecture view
          else {
              if (BaseManagerImpl.getInstance().knowsACS(name)) {
                  result.append(name+" is a closed ACS.");
              } // if
              else {
                  result.append(name+" is a removed ACS; it is removed from the view.");
              } // else
              // never added to the view
              continue;
          } // else
      } // catch

                              // for testing and tracking
                              //System.out.println("FolderAbst.prepareDataAfterRestoring() adds "+
                                //      name+" in m_BaseObjects_ ");

      // does not check if obj is an acs
      // caution: if DialogError to throw just before, obj is null
      m_BaseObjects_.put((ImmutableName)((NameImpl)name).clone(), obj);

        // updates l_connectedACSNames_ and connectedToSeveralACS_ if necessary
        // for common views...
      if (obj != null) {
        if (obj instanceof ACSObject) {
          if (!l_connectedACSNames_.contains( (ImmutableName)entry.getValue()) )
            l_connectedACSNames_.add( (ImmutableName)entry.getValue() );
          if (l_connectedACSNames_.size() > 1) connectedToSeveralACS_ = true;
        } // if (_obj instanceof ACSObject)
        // ... and noop for Architecture views with ACSs
      }
      // if DialogError to throw just before, obj is null
      else {
        if ((ImmutableName)entry.getValue() != null) {
          if (!l_connectedACSNames_.contains( (ImmutableName)entry.getValue()) )
            l_connectedACSNames_.add( (ImmutableName)entry.getValue() );
          if (l_connectedACSNames_.size() > 1) connectedToSeveralACS_ = true;
        } // if ((ImmutableName)entry.getValue()
      } // else
    } // for ite

    // throw the exception - never null for the subclass VirtualFolderImpl.
    if (error_name != null) {

                      // for testing purpose
                      //System.out.println("In ViewInBaseImpl.prepareAfterRestoring(), for "+
                              //getName()+", DialogError with error_acsname: "+error_acsname);

        obj = null;
        name = null;
        entry = null;
        result = null;

        if (error_acsname == null) {
          throw new DialogError("The folder object '"+error_name.lastComponent() +
                  "' is from a removed ACS"+
                  ". The folder cannot be opened, and should be removed.");
        } // if
        else
          throw new DialogError("The folder object '"+error_name.lastComponent() +
                  "' is from the closed ACS "+error_acsname+
                  ". To open the folder, open this ACS first.");
    } // if

    obj = null;
    name = null;
    entry = null;
    error_acsname = null;
    error_name = null;

    if (result.length() == 0)
        return null;
    else
        return result.toString();
  }  // prepareDataAfterRestoring()

} // FolderAbst