//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import java.util.List;
import java.util.Map;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Arrays;
import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;

/**
 * This class is responsible for modeling a GroupIDMember that owns a set of members,
 * and which has Account/Group rights on Resources.
 * Typically, it handles a set of user accounts and groupIDs.
 * The membership links between the groupID and each member are managed by the GroupID.
 * Resources and actors may also be registered as owned by a groupID,
 * in two different lists. An actor may also have a groupID as a current group
 * under which it runs, and such an actor is registered in a third list. <P>
 * It is a javabean with the following bound properties:
 * 'l_Members', 'l_GroupIDCurrentActors', 'l_GroupIDOwnResources',
 * 'l_GroupIDCurrentActorsAsSecondaryGroup'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user. <P>
 * This class has several subclasses in this package and in the AcsAddon packages.
 * Most of the variables are protected.
 * @version 0.7.0
 */
public class GroupIDImpl extends GroupIDMemberImpl implements GroupID, Serializable {
  private static final long   serialVersionUID__ = 0;

  // protected variables
  protected GroupIDMember[]     l_Members_ = new GroupIDMemberImpl[0];// External Object Reference
  protected int                 order_ = 0;
  protected Resource[]          l_ownResources_ ;
  protected Actor[]             l_currentActors_ ;
  protected Actor[]             l_secGroupActors_ ;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final GroupIDImpl EMPTY_INSTANCE = new GroupIDImpl();


  /** variable for the JDK 2 serialization
  * @serial ("l_Members_",GroupIDMemberImpl[].class),
  *   ("order_",Integer.TYPE),
  *   ("l_ownResources_",ResourceImpl[].class),
  *   ("l_currentActors_",ActorImpl[].class),
  *   ("l_secGroupActors_",ActorImpl[].class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("l_Members_",GroupIDMemberImpl[].class),
      new ObjectStreamField("order_",Integer.TYPE),
      new ObjectStreamField("l_ownResources_",ResourceImpl[].class),
      new ObjectStreamField("l_currentActors_",ActorImpl[].class),
      new ObjectStreamField("l_secGroupActors_",ActorImpl[].class)
  };

 /**
  * Default constructor
  */
  public GroupIDImpl() {
  }

   /**
   * Constructor used in the beamer by the editor for the ACS EP list, from 'New'.
   * The last component of _name cannot end with '_tree&gt;' to avoid any
   * confusion with the second nodes under GROUPS in the explorer. 
   * @param _name ImmutableName of the groupID
   * @param _acs ACS of the groupID
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users
   * @exception CreateError if _name ends with '_tree&gt;', or if
   *    ACSImpl.newEorEligibleParty() throws it - the instance must then be deleted !
   */
  public GroupIDImpl(ImmutableName _name, ACS _acs, String _epType, Boolean _isAdmin)
   throws CreateError {
          super(_name, _acs, _epType, _isAdmin);
          if (_name.lastComponent().endsWith("_tree>")) {
              try {
                  finalizeForBase();
              } catch (UpDateError u) {
                    //u.printStackTrace();
              }
              throw new CreateError(" The name cannot end with '_tree>'");
          } // if

        // for testing and tracking
        //System.out.println("call to 2-arguments GroupIDImpl constructor with l_Members_: "+l_Members_);      
      
  } // GroupIDImpl()
  
   /**
   * Used for the creation of a groupID.
   * This method is used by copy(). 
   * @param _name ImmutableName of the eligible party to create
   * @param _acs Main ACS of the eligible party to create
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if ACSImpl.newEorEligibleParty() throws it
   * - the instance must then be deleted !
   */
  public GroupIDImpl(ImmutableName _name, ACS _acs, String _epType, boolean _isAdmin,
    boolean _toRegisterInManager) throws CreateError {
        super(_name, _acs, _epType, _isAdmin, _toRegisterInManager);
  }
  
  /**
  * This method provides a copy of the instance
  * with another name.
  * The copy order is 3 times the source order.
  * Does not copy the GroupID members, nor the own resources,
  * the current actors, the own actors.
  * To copy all these links, you have to copy first the GroupID,
  * then to copy the GroupIDMembers, the Resources and the Actors
  * to copy the links. 
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of GroupImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager)
    throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in GroupIDImpl.copy(),"+
          " _name or _acs is null.");
      }
      
      GroupIDImpl result = new GroupIDImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);

             // for testing and tracking
             //System.out.println("in GroupIDImpl.copy(), the source: "+this+" and the result: "+result );
      
      
      return result;
  } // copy(,,,)

  
  /**
  * This method provides a copy of the instance
  * as copy(), but without copying the EligibleParty internal aclEntries.
  * The copy order is 3 times the source order.
  * Does not copy the GroupID members, nor the own resources,
  * the current actors, the own actors.
  * To copy all these links, you have to copy first the GroupID,
  * then to copy the GroupIDMembers, the Resources and the Actors
  * to copy the links. 
  * Calls copyOwnVariables().
  * Called by ACSImpl.copy().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of GroupImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in GroupIDImpl.copy(),"+
          " _name or _acs is null.");
      }
      
      GroupIDImpl result = new GroupIDImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclEntry copy
      
      // if (result.getEorL_AclConnectedResources().isEmpty()) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copy(,,,)
  
  /**
  * Defines the groupID members, and
  * updates the members through the add/removeGroupToGM methods.
  * This method is synchronized on l_Members_.
  * Fire an 'l_Members' event. Not used by the GUI which used add/removeEorMember().
  * Does not control the type. Called by finalizeForBase(null) and gWork.AcsFactory.
  * Control the presence of a loop between groupIDs.
  * Note: members are independant to the own actors and current actors of the groupID.
  * @param _L is the members of the GroupID. May be null.
  *  The new children array is sorted with ImmutableEligibleParty.compareTo();
  *  null and empty instances are removed.
  * @exception InternalError if one element is not known by the
  *   main ACS, if the main ACS has no registered eligible parties.
  * @exception CreateError if a _L element does not belong to the ACS,
  *   or if it is a GroupID which has this instance has
  *   direct or indirect member
  */
  public void setEorL_Members(GroupIDMember[] _L) throws CreateError {

    GroupIDMember[] newArray = new GroupIDMemberImpl[1];
    GroupIDMember[] oldArray;
    GroupIDMember[] L;
    List newList;
    short shiftingBack = 0;

    if (getEorACS() == null) // aCS_ from EligibleParty
      throw new InternalError("aCS_ of GroupID " + getName().toString() + " is null");

    if (getEorACS().getEorM_OwnEligibleParties() != null && _L != null) {
      Map eligiblePartiesMap = (Map) getEorACS().getEorM_OwnEligibleParties();
      
      for (int i = 0; i <_L.length; i++) {
        if (_L[i] == null) continue;
        if (!eligiblePartiesMap.containsKey(_L[i].getName()) ) {
          getEorACS().getEorBaseError().addException("In GroupIDImpl."+
            "setEorL_Members(GroupIDMember[]), the new member GroupIDMember[" + i +
            "] is not owned by the main ACS "+getEorACS().getName());
          throw new InternalError("In GroupIDImpl."+
            "setEorL_Members(GroupIDMember[]), a new member is not owned "+
            "by the main ACS "+getEorACS().getName());
        } // if
        if (_L[i] instanceof GroupID && (((GroupID)_L[i]).containsAsIndirectMember(this) ||
          ((GroupID)_L[i]).containsAsMember(this)))
            throw new CreateError("No creation of the member "+
              _L[i].getName()+" - "+getName()+
              " is already a direct or an indirect member of it!");
        if (containsAsIndirectMember(_L[i]))
            throw new CreateError("No creation of the member "+_L[i].getName()+" which is already "+
            "an indirect member of "+getName());
        if (containsAsMember(_L[i]))
            throw new CreateError("No creation of the member "+_L[i].getName()+" which is already "+
            "a direct member of "+getName());       
      } // for i
      
    } //if
    else {
      if (getEorACS().getEorM_OwnEligibleParties() == null) {
        getEorACS().getEorBaseError().addException("For GroupIDImpl."+
          "setEorL_Members(GroupIDMember[]), aCS_.getEorM_OwnEligibleParties() is null");
        throw new InternalError("For GroupIDImpl.setEorL_Members(GroupIDMember[]),"+
          " aCS_.getEorM_OwnEligibleParties() is null");
      }
    } //else

    synchronized (l_Members_) {

      // unregistered this groupID in the old members if they are not created by Access Road
      for (int  i=0 ; i < l_Members_.length ; i++) {
          if (l_Members_[i] != null )
                ((GroupIDMemberImpl)l_Members_[i]).removeGroupToGM(this);
      } // for


      // remove null and empty members, and sort the others members
      // in a ImmutableEligibleParty.compareTo() order
      if (_L != null) {
          
        newArray = new GroupIDMember[_L.length];
            // remove null and empty members
        for (int  i=0 ; i < _L.length ; i++) {
          if (_L[i] == null || _L[i].equals(GroupIDMemberImpl.EMPTY_INSTANCE))
            shiftingBack++;
          else newArray[i - shiftingBack] = _L[i];
        } // for
        
      } // if
      if (_L != null && _L.length != shiftingBack) {
          
        L = new GroupIDMemberImpl[newArray.length - shiftingBack];
        System.arraycopy(newArray, 0, L, 0, newArray.length - shiftingBack);
          // sort the array
        try {
          Arrays.sort(L);
        } catch (ClassCastException c) {
            c.printStackTrace();
            throw new InternalError("in GroupIDImpl.setEorL_Members(), exception on array sort.");
        }

      } // if
      else L = new GroupIDMemberImpl[0]; // value if _L is null

      // saves the current l_Members_ in oldArray
      oldArray = new GroupIDMemberImpl[l_Members_.length];
      System.arraycopy(l_Members_, 0, oldArray, 0, l_Members_.length);

      // updates l_Members_
      l_Members_ = new GroupIDMemberImpl[L.length];
      System.arraycopy(L, 0, l_Members_, 0, L.length);
    } // synchronized

    // update the new EP
    for (int  i = 0 ; i < l_Members_.length ; i++) {
        ((GroupIDMemberImpl)l_Members_[i]).addGroupToGM(this);
    } // for    
    
    try {
        firePropertyChange("l_Members", oldArray, L);
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.setEorL_Members(), firePropertChange() throws:");
            e.printStackTrace();
    }
    
    newArray = null;
    oldArray = null;
    L = null;
    newList = null;
  } // setEorL_Members(l_Members_)

  /**
  * Adds a GroupIDMember as a member of this groupID, and
  * update the member through the addGroupToGM method.
  * Controls the presence of a loop through nested groupIDs, including aliasings.
  * Controls the ACS constraints for 'GroupIDMember.NoMemberOf',
  * 'GroupID.MemberFromAdmin', 'GroupID.NoUserIDAsMember', 'GroupID.NoGroupIDAsMember',
  * 'GroupID.NoMoreThanOneMember', 'GroupID.TypesOfMemberFor',
  * 'GroupIDMember.NoMoreThanOneGroup.'
  * to checkup the operation is allowed by the ACS.
  * Calls isConditionalACLtoResetAfterMembershipChangeInThisGroup() and
  * AclEntry.resetRightsAsInheritedAclEntry().
  * This method is synchronized on l_Members_.
  * Fire an 'l_Members' event.
  * @param _gm new member of the GroupID. No action if null, empty.
  * @exception CreateError if _gm is a GroupID which has this instance has
  *   direct or indirect member, if _gm is already a member, or if the _gm type
  *   is not authorized by the ACS, or if the type does not allow the operation,
  *   if this group is immutable, or if it is mutable and not the new member.
  */
  public void addEorMember(GroupIDMember _gm) throws CreateError {
    if (type_.startsWith("<immutable> ") || type_.endsWith("<immutable> "))
          throw new CreateError(
            " The type of "+getName().lastComponent()+" sets it as immutable for its members.");  
    
    if (_gm == null || _gm.equals(GroupIDMemberImpl.EMPTY_INSTANCE) ) return;

                    // for testing
                    //System.out.println("GroupIDImpl.addEorMember() with member: "+
                      //      _gm.getName());
    
    if ( _gm.getEPType().indexOf("<immutable>") != -1) 
          throw new CreateError(
            " "+getName().lastComponent()+" is mutable, but not the new member.");
    
    if (_gm.getEorACS() != getEorACS())
        throw new CreateError(" No creation of the member - it does not belong "+
        "to the same ACS.");
    if (!getEorACS().getManageGroupTrees() && _gm instanceof GroupID)
        throw new CreateError(" Groups are not authorized to have other groups "+
          "as members, in this ACS.");

    
    if (_gm instanceof GroupID && (((GroupID)_gm).containsAsIndirectMember(this) ||
      ((GroupID)_gm).containsAsMember(this)))
        throw new CreateError(" No operation - "
          +getName().lastComponent()+" is member of "+_gm.getName().lastComponent());

    if (containsAsIndirectMember(_gm))
        throw new CreateError(" "+_gm.getName().lastComponent()+" is already "+
        "an indirect member of "+getName().lastComponent());
    if (containsAsMember(_gm))
        throw new CreateError(" "+_gm.getName().lastComponent()+" is already "+
        "a direct member of "+getName().lastComponent());

    // variables
    GroupIDMember[] newArray;
    GroupIDMember[] l_oldMembers;    
    String[]        l_remove = null;
    boolean         remove = false;
    boolean         admin = false;
    ImmutableGroupID[]  l_groups = null;
    ImmutableAclEntry[] l_acl = null;
    
    if (getEorACS().getM_TypeOrientedPolicy() != null) {
        l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                  get("GroupIDMember.NoMemberOf");
        remove = false;
        if (l_remove != null) {
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(
                          UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(_gm.getEPType()))) {
                      remove = true;
                      break;
                  }
              } // for i
        } // if (l_remove !=
        l_remove = null;
        if (remove)
              throw new CreateError(" Through the ACS type policy 'GroupIDMember.NoMemberOf', "+
                      "the argument "+_gm.getNickName()+
                      " has a type which forbids it to be the member of a group.");

        l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                  get("GroupID.MemberFromAdmin");
        remove = false;
        l_groups = _gm.getEorL_GroupID();
        if (l_remove != null) {
              for (int i = 0; i < l_groups.length; i++) {
                  if (l_groups[i].isAdmin()) {
                      admin = true;
                      break;
                  }
              } // for i
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(getEPType())) &&
                        !admin) {
                      remove = true;
                      break;
                  }
              } // for i
        } // if (l_remove !=
        l_remove = null;
        l_groups = null;
        if (remove)
              throw new CreateError(" Through the ACS type policy 'GroupID.MemberFromAdmin', "+
                      "the group forbids members which are not administrators.");

        l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                  get("GroupID.NoUserIDAsMember");
        remove = false;
        if (l_remove != null) {
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(getEPType())) &&
                        _gm instanceof UserID) {
                      remove = true;
                      break;
                  }
              } // for i
        } // if (l_remove !=
        l_remove = null;
        if (remove)
              throw new CreateError(" Through the ACS type policy 'GroupID.NoUserIDAsMember', "+
                      "the group forbids accounts as members.");

        l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                  get("GroupID.NoGroupIDAsMember");
        remove = false;
        if (l_remove != null) {
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(getEPType())) &&
                        _gm instanceof GroupID) {
                      remove = true;
                      break;
                  }
              } // for i
        } // if (l_remove !=
        l_remove = null;
        if (remove)
              throw new CreateError(" Through the ACS type policy 'GroupID.NoGroupIDAsMember', "+
                      "the group forbids groups as members.");

        l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                  get("GroupID.NoMoreThanOneMember");
        remove = false;
        if (l_remove != null) {
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(getEPType())) &&
                        l_Members_ != null && l_Members_.length == 1) {
                      remove = true;
                      break;
                  }
              } // for i
        } // if (l_remove !=
        l_remove = null;
        if (remove)
              throw new CreateError(" Through the ACS type policy 'GroupID.NoMoreThanOneMember', "+
                      "the group does not accept more than one member.");

        l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                  get("GroupID.TypesOfMemberFor."+getEPType());
        remove = false;
        if (l_remove != null) {
              StringBuilder message = new StringBuilder();
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(
                          UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(_gm.getEPType()))) {
                      remove = true;
                  }
                  message.append(l_remove[i]);
                  message.append(", ");
              } // for i
              
              l_remove = null;
              if (!remove) // caution: 'remove' means there 'keep'!
                    throw new CreateError(" Through the ACS type policy 'GroupID.TypesOfMemberFor', "+
                      "the group '"+getNickName()+"' forbids the type of "+_gm.getNickName()+
                      " - The allowed type(s) are "+message);
              message = null;
        
        } // if (l_remove != 
 

        l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
                  get("GroupIDMember.NoMoreThanOneGroup."+_gm.getEPType());
        remove = false;
        if (l_remove != null) {
              for (int i = 0; i < l_remove.length; i++) {
                  if (l_remove[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(getEPType()))) {
                      remove = true;
                      break;
                  }
              } // for i
              if (remove) { // test if _gm is already in a group of the same type
                  remove = false;
                  ImmutableGroupID[] l_grps = _gm.getEorL_GroupID();
                  for (int i = 0; i < l_grps.length; i++) {
                      if (l_grps[i].getEPType().equals(getEPType())) {
                          remove = true;
                          break;
                      }                      
                  } // for i
                  l_grps = null;
              } // if
        } // if (l_remove !=  
        l_remove = null;
        if (remove) 
              throw new CreateError(" Through the ACS type policy "+
                      "'GroupIDMember.NoMoreThanOneGroup', "+
                      "the member does not accept to be in a second group of this type."); 
    } // if (getEorACS().getM_TypeOrientedPolicy() != null)

    synchronized (l_Members_) {
      // save the old Members list, since l_Members_ is a bound property
      l_oldMembers = new GroupIDMemberImpl[l_Members_.length];
      System.arraycopy(l_Members_, 0, l_oldMembers, 0, l_Members_.length);
      
      // add in an empty place if possible
      boolean recorded = false;
      for (int i = 0 ; i < l_Members_.length ; i++) {
        if (l_Members_[i] == null || l_Members_[i].equals(GroupIDMemberImpl.EMPTY_INSTANCE) ) {
          l_Members_[i] = _gm;
          recorded = true;
          break;
        }
      }
      // else, add at the end of l_Members_
      if (recorded == false) {
          newArray = new GroupIDMemberImpl[l_Members_.length + 1];
          System.arraycopy(l_Members_, 0, newArray, 0, l_Members_.length);
          newArray[l_Members_.length] = _gm;
          l_Members_ = newArray;
      } else newArray = l_Members_;

                     // for testing and tracking
                     //System.out.println("in GroupIDImpl.addEorMember(_gm), "+
                       //      "members before _gm.addGroupToGM(this) are:" );
                     //for (int i = 0; i < l_Members_.length; i++) {
                       //  System.out.println("------"+l_Members_[i].getName());
                     //} // for i
      
      // updates the group list in _gm.
      // if _gm is a generic UserIDImpl, calls setEorPrimGroup() which calls back
      // containsAsMemberWithoutAliasing().
      _gm.addGroupToGM(this);

      try {
          Arrays.sort(l_Members_);
      } catch (ClassCastException c) {
          c.printStackTrace();
          throw new InternalError("in GroupIDImpl.addEorMember(), exception on the array sorting.");
      }
    } // synchronized

    try {
          firePropertyChange("l_Members", l_oldMembers, getEorL_Members());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From GroupIDImpl.addEorMember(), firePropertChange() throws:");
        e.printStackTrace();
    } 
                    // for testing
                    //System.out.println("GroupIDImpl.addEorMember() before "+
                      //      "the test on conditional ACL.");

    // reset the condition ACLs, and throw some property changes
    if (aCS_.getManageFlexibleConditionalAcls() ||
                aCS_.getManageRightDefinedConditionalAcls()) {
        
        l_acl = aCS_.getEorL_AclEntries();
        for (int i = 0; i < l_acl.length; i++ ) {

            if (isConditionalACLtoResetAfterMembershipChangeInThisGroup(l_acl[i], _gm))
                ((AclEntryImpl)l_acl[i]).resetRightsAsInheritedAclEntry();

        } // for i
        
    } // if (aCS_.getManageFlexibleConditionalAcls()

                    // for testing
                    //System.out.println("end of GroupIDImpl.addEorMember() with member: "+
                      //      _gm.getName());
    
    
    l_oldMembers = null;
    newArray = null;
    l_acl = null;
  } // addEorMember()

  /**
  * Removes a groupID member, and update the member through the removeGroupToGM method,
  * if the group is not immutable.
  * Calls isConditionalACLtoResetAfterMembershipChangeInThisGroup() and
  * AclEntry.resetRightsAsInheritedAclEntry().
  * This method is synchronized on l_Members_.
  * @param _gm member of the GroupID. No action if unknown, null, empty.
  * @exception CreateError if the type does not allow the operation
  */
  public void removeEorMember(GroupIDMember _gm) throws CreateError {
    if (type_.startsWith("<immutable> ") || type_.endsWith("<immutable> "))
          throw new CreateError(
            " The type of "+getName().lastComponent()+
            " sets it as immutable for its members."); 
    
    if (_gm == null || _gm.equals(GroupIDMemberImpl.EMPTY_INSTANCE)
      || l_Members_ == null || l_Members_.length == 0) return;

    // variables
    GroupIDMember[] newArray;
    GroupIDMember[] l_oldMembers;
    GroupIDMember[] result;
    short               shiftingBack = 0;
    ImmutableAclEntry[] l_acl = null;

    synchronized (l_Members_) {
      // save the old Members list, since l_Members_ is a bound property
      l_oldMembers = new GroupIDMemberImpl[l_Members_.length];
      System.arraycopy(l_Members_, 0, l_oldMembers, 0, l_Members_.length);

      // update l_Members_
          // this code deletes all relevant members, even if count > 1,
      newArray = new GroupIDMemberImpl[l_Members_.length];
      result = new GroupIDMemberImpl[0];

      for (int  i = 0 ; i < l_Members_.length ; i++) {
        if (l_Members_[i] != null) {
          if (l_Members_[i].equals(_gm)) {
                // incremente shiftingBack delete the reference l_Members_[i]
            shiftingBack++;
          }
          else newArray[i - shiftingBack] = l_Members_[i];
        }
        else shiftingBack++; // delete null l_Members_[i]
      } // for

      if (l_Members_.length != shiftingBack) {
        result = new GroupIDMemberImpl[l_Members_.length - shiftingBack];
        System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
      }
                // else result (already egal to) new GroupIDMemberImpl[0]
      l_Members_ = result;

             // for testing and tracking
             //System.out.println("in GroupIDImpl.removeEorMember(), the new members are:" );
             //for (int i = 0; i < l_Members_.length; i++) {
               //  System.out.println("------"+l_Members_[i].getName());
             //} // for i
      
      /*
      try {
          Arrays.sort(l_Members_);
      } catch (ClassCastException c) {
          c.printStackTrace();
          throw new InternalError("in GroupIDImpl.removeEorMember(), exception on array sort.");
      }
       */
             
    } // synchronized

    // update the group list in _gm
    _gm.removeGroupToGM(this);
    
    try {
        firePropertyChange("l_Members", l_oldMembers, getEorL_Members());
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.removeEorMember(), firePropertChange() throws:");
            e.printStackTrace();
    } // catch

    // reset the condition ACLs, and throw some property changes
    if (aCS_.getManageFlexibleConditionalAcls() ||
                aCS_.getManageRightDefinedConditionalAcls()) {
        
        l_acl = aCS_.getEorL_AclEntries();
        for (int i = 0; i < l_acl.length; i++ ) {

            if (isConditionalACLtoResetAfterMembershipChangeInThisGroup(l_acl[i], _gm))
                ((AclEntryImpl)l_acl[i]).resetRightsAsInheritedAclEntry();
            
        } // for i
        
    } // if (aCS_.getManageFlexibleConditionalAcls()

    newArray = null;
    l_oldMembers = null;
    result = null;
    l_acl = null;
  } // removeEorMember()

  /**
  * Get the ordered array of this GroupID members.
  * If empty, returns GroupIDMemberImpl[0].
  * @return a copy of GroupID members. Never null. This is an ordered list
  * without null member, and without empty member after the first update.
  */
  public ImmutableGroupIDMember[] getEorL_Members() {
    // l_Members_ is null when the beamer tries to show a new groupID
    if (l_Members_ == null || l_Members_.length == 0) {

            // for testing and tracking
            //System.out.println("in GroupIDImpl.getEorL_Members(), "
            //+" no member for "+getName());

      return new GroupIDMemberImpl[0];
    }
    ImmutableGroupIDMember[] result = new GroupIDMemberImpl[l_Members_.length];
    System.arraycopy(l_Members_, 0, result, 0, l_Members_.length);

            // for testing and tracking
            //System.out.println("in GroupIDImpl.getEorL_Members(), for "+getName());
            //System.out.println("the members are:");
            //for (int i = 0; i < l_Members_.length; i++) {
            //  System.out.println("------"+l_Members_[i].getName());
            //} // for i

    return result;
  } // getEorL_Members()

  /**
  * Tests if an eligible party is a direct member of this GroupID, or if it has
  * an alias which is a direct member. It is an identity test, not with equals().
  * Returns for the moment the same result as containsAsMemberWithoutAliasing().
  * Synchronized on l_Members_.
  * @param _ep to test
  * @return true if _ep is a member of the GroupID.
  */
  public boolean containsAsMember(ImmutableGroupIDMember _ep) {
      boolean result = false;
      if (_ep == null || l_Members_ == null || l_Members_.length == 0) return false;
      synchronized (l_Members_) {
        for (int i = 0 ; i < l_Members_.length ; i++ ) {
          if (l_Members_[i] == _ep) {
            result = true;
            break;
          }
          //if (l_Members_[i].getIsAlias() &&
            //_ep == l_Members_[i].getEorTrueEP()) {
              //result = true;
              //break;
          //}
        } // for
      } // synchro


            // for testing and tracking
            //System.out.println("in GroupIDImpl.containsAsMember("+_ep+"), the"
            //+" result for "+getName()+" is: "+ result);

      return result;
  } // containsAsMember()

  /**
  * Test if an eligible party is a direct member of this GroupID,
  * and do not test if it has an alias which is a direct member.
  * Return actually the same result that containsAsMember().
  * Called by UserID.setEorPrimGroup().
  * Synchronized on l_Members_.
  * @deprecated _ep to test
  * @return true if _ep is a member of the GroupID.
  */
  public boolean containsAsMemberWithoutAliasing(ImmutableGroupIDMember _ep) {

            // for testing and tracking
            //if (_ep != null)
              //  System.out.println("call to GroupIDImpl.containsAsMemberWithoutAliasing(), "+
                //    "on :"+getName()+" for "+_ep.getName());

            // for testing and tracking
            //System.out.println("call to GroupIDImpl.containsAsMemberWithoutAliasing(), "+
              //      "l_Members_.length: "+l_Members_.length);
            
      boolean result = false;
      if (_ep == null || l_Members_ == null || l_Members_.length == 0) return false;
      
      synchronized (l_Members_) {
        for (int i = 0 ; i < l_Members_.length ; i++ ) {
          if (l_Members_[i] == _ep) {
            result = true;
            break;
          }
        } // for
      } // synchro


            // for testing and tracking
            //System.out.println("in GroupIDImpl.containsAsMemberWithoutAliasing("+_ep+"), the"
                //+" result for "+getName()+" is: "+ result);

      return result;
  } // containsAsMemberWithoutAliasing()

  /**
  * Test if an eligible party is an indirect member of this GroupID,
  * or if it is an hard alias having a reference which is an indirect member.
  * Caution: if the argument is the alias of a GroupIDMember, the test is applied
  * only to the reference as potential member, and not to the argument.
  * Synchronized on l_Members_.
  * @param _ep to test
  * @return true if _ep is an indirect member of the GroupID, or its alias reference.
  */
  public boolean containsAsIndirectMember(ImmutableGroupIDMember _ep) {
      boolean result = false;
      if (_ep == null || l_Members_ == null || l_Members_.length == 0) return false;
      ImmutableGroupIDMember mem = null;

      synchronized (l_Members_) {
        for (int i = 0 ; i < l_Members_.length ; i++ ) {
          if (l_Members_[i] instanceof GroupID) {

            if (aCS_.getManageInternalHardAlias() &&
                    _ep.getEorReferenceAsInternalHardAlias() != null &&
                    _ep.getEorReferenceAsInternalHardAlias() instanceof GroupIDMember) {
                mem = (GroupIDMember)_ep.getEorReferenceAsInternalHardAlias();
            }
            else {
                mem = _ep;
            }

            if ( ((GroupID)l_Members_[i]).containsAsMember(mem) ||
                    ((GroupID)l_Members_[i]).containsAsIndirectMember(mem) ) {
                result = true;
                break;
            } // if

          } // if (l_Members_[i] instanceof GroupID)
        } // for
      } // synchro

            // for testing and tracking
            //System.out.println("in GroupIDImpl.containsAsIndirectMember("+_ep+"), the"
            //+" result for "+getName()+" is: "+ result);

      mem = null;
      return result;
  } // containsAsIndirectMember()

  /**
  * Gets the first member which contains the argument as member or indirect member.
  * Calls containsAsMember() and containsAsIndirectMember().
  * Returns null if _ep is a member.
  * Synchronized on l_Members_.
  * @param _ep to test. May be null.
  * @return the first member. May be null.
  */
  public GroupID getFirstIntermediateForIndirectMember(ImmutableGroupIDMember _ep) {
      if (_ep == null || l_Members_ == null || l_Members_.length == 0) return null;
      ImmutableGroupIDMember mem = null;
      GroupID result = null;

      synchronized (l_Members_) {
        for (int i = 0 ; i < l_Members_.length ; i++ ) {
          if (l_Members_[i] instanceof GroupID) {

            if (aCS_.getManageInternalHardAlias() &&
                    _ep.getEorReferenceAsInternalHardAlias() != null &&
                    _ep.getEorReferenceAsInternalHardAlias() instanceof GroupIDMember) {
                mem = (ImmutableGroupIDMember)_ep.getEorReferenceAsInternalHardAlias();
            }
            else {
                mem = _ep;
            }

            if ( ((GroupID)l_Members_[i]).containsAsMember(mem) ||
                    ((GroupID)l_Members_[i]).containsAsIndirectMember(mem) ) {
                result = (GroupID)l_Members_[i];
                break;
            } // if

          } // if (l_Members_[i] instanceof GroupID)
        } // for
      } // synchro

            // for testing and tracking
            //System.out.println("in GroupIDImpl.getFirstIntermediateForIndirectMember("+_ep+"), the"
            //+" result for "+getName()+" is: "+ result);

      mem = null;
      return result;
  } // getFirstIntermediateForIndirectMember()


  /**
  * Test if an eligible party is an indirect member of this GroupID,
  * and do not test if it is an alias having a reference which is an indirect member.
  * Synchronized on l_Members_.
  * @return true if _ep is a member of the GroupID.
  */
  public boolean containsAsIndirectMemberWithoutAliasing(ImmutableGroupIDMember _ep) {
      boolean result = false;
      if (_ep == null || l_Members_ == null || l_Members_.length == 0) return false;
      synchronized (l_Members_) {
        for (int i = 0 ; i < l_Members_.length ; i++ ) {
          if (l_Members_[i] instanceof GroupID) {
            if ( ((GroupID)l_Members_[i]).containsAsMember(_ep)
              || ((GroupID)l_Members_[i]).containsAsIndirectMember(_ep) ) {
                  result = true;
                  break;
            }
          } // if (l_Members_[i] instanceof GroupID)
        } // for
      } // synchro


            // for testing and tracking
            //System.out.println("in GroupIDImpl.containsAsIndirectMember("+_ep+"), the"
            //+" result for "+getName()+" is: "+ result);

      return result;
  } // containsAsIndirectMemberWithoutAliasing()
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("GroupImpl.getPropertyChangeListeners(),
                    // call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      return super.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()

  /**
  * Set the order of the GroupID which determines its priority
  * among in access control decisions - on Unix : 16-bit group identifier (GID).
  * Not used by the GUI
  * @param _i is the order of the groupID - on Unix : 16-bit group identifier (GID)
  */
  public synchronized void setOrder(int _i)	{
    order_ = _i;
  }

  /**
  * @return the order of the groupID - on Unix : 16-bit user identifier (UID)
  */
  public int getOrder()	{
	return order_;
  }

  /**
  * Invoked by Actor.setEorL_SecondaryGroup().
  * Fires an event 'l_GroupIDCurrentActorsAsSecondaryGroup'.
  * @param _act Actor for which this groupID is a secondary groupID.
  *  May be null.
  */
  protected void addOwnActorAsSecondaryGroup(Actor _act) {
    // variables
    Actor[]     newArray = new ActorImpl[1];
    boolean     recorded = false;
    Actor       nullAct = new ActorImpl();

    // unuseful, since Actor.setEorL_SecondaryGroup() controls:
    // if (_act == null || nullAct.equals(_act) ) return;

    if (l_secGroupActors_ == null) l_secGroupActors_ = (Actor[]) new ActorImpl[1];

    Actor[]     oldValue = new ActorImpl[l_secGroupActors_.length];
    System.arraycopy(l_secGroupActors_, 0, oldValue, 0, l_secGroupActors_.length);

    // add in an empty place if possible
    for (int i = 0 ; i < l_secGroupActors_.length ; i++) {
        if (l_secGroupActors_[i] == null || l_secGroupActors_[i].equals(nullAct) ) {
            l_secGroupActors_[i] = _act;
            recorded = true;
            break;
        }
    }
    // else, add at the end of l_secGroupActors_
    if (!recorded) {
        newArray = new ActorImpl[l_secGroupActors_.length + 1];
        System.arraycopy(l_secGroupActors_, 0, newArray, 0, l_secGroupActors_.length);
        newArray[l_secGroupActors_.length] = _act;
        l_secGroupActors_ = new ActorImpl[newArray.length];
        l_secGroupActors_ = newArray;
    }
    else {
        newArray = new ActorImpl[l_secGroupActors_.length];
        System.arraycopy(l_secGroupActors_, 0, newArray, 0, l_secGroupActors_.length);
    }

    try {
        firePropertyChange("l_GroupIDCurrentActorsAsSecondaryGroup", oldValue,
          getEorL_GroupIDCurrentActorsAsSecondaryGroup());
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.addOwnActorAsSecondaryGroup(), "+
              "firePropertChange() throws:");
            e.printStackTrace();
    } 
    
    newArray = null;
    nullAct = null;
  } // addOwnActorAsSecondaryGroup()

  /**
  *  A currentGroupID or a secondary groupID is a property of an actor, which is
  *  a kind of resource. But these actors are NOT returned by this method.
  * @return array of the resources in the ACS for which this group is their primary groupID.
   *        May be empty, but never null.
  */
  public Resource[] getEorL_GroupIDOwnResources() {
        if (l_ownResources_ == null) return new ResourceImpl[0];
        Resource[] result = new ResourceImpl[l_ownResources_.length];
        System.arraycopy(l_ownResources_, 0, result, 0, l_ownResources_.length);

            // for testing and tracking
            //System.out.println("in GroupIDImpl.getEorL_GroupIDOwnResources(), the first"
            //+" own resource of "+getName()+" is: "+ result[0]);


        return result;
  }

  /**
  * @return true if it owns the resource.
  */
  public boolean ownsResource(Resource _res) {
      if (_res == null) return false;
      boolean result = false;
      for (int i = 0; i < l_ownResources_.length; i++) {
        if (l_ownResources_[i] == _res) {
          result = true;
          break;
        }
      } // for
      return result;
  }

   /**
    * A current actor runs under this groupID.
    * A current actor may be an own resource of this groupID or not.
   * @return array of the current actors attached to this groupID.
   *  May be empty but not null.
   */
   public Actor[] getEorL_GroupIDCurrentActors() {
        if (l_currentActors_ == null) return new ActorImpl[0];
        Actor[] result = new ActorImpl[l_currentActors_.length];
        System.arraycopy(l_currentActors_, 0, result, 0, l_currentActors_.length);
        return result;
  }

   /**
   *  Get the actors in the ACS that have, in the role of a running processes,
   *   this group as a secondary groupID.
   *  If this groupID is the main groupID  or the current groupID of an actor,
   *  it is NOT returned by this method.
   * @return array of the actors in the ACS that have this group as
   *  a secondary groupID. May be empty but not null.
   */
   public Actor[] getEorL_GroupIDCurrentActorsAsSecondaryGroup() {
        if (l_secGroupActors_ == null) return new ActorImpl[0];
        Actor[] result = new ActorImpl[l_secGroupActors_.length];
        System.arraycopy(l_secGroupActors_, 0, result, 0, l_secGroupActors_.length);
        return result;
   }
   
   /**
   * This method removes the members, and the resources through the variable groupID_ 
   * and, for Actors, through currentGroupID_ and l_SecondaryGroup_. It deletes
   * the conditional AclEntries having this group as condition.
   * super.finalizeForBase() removes the groupID in the ACS lists (main and controllers),
   * and removes in the resources connected through an aclEntry.
   * This method removes in userIDs the primary group if it is this.
   * Overrides the super method and calls it at the end.
   * An unused object has to be deleted to inform the others objects, and this method
   * HAS TO BE USED before deleting the groupID.
   * Note: when this method is called by the EligiblePartyImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the userID is unknown of the main ACS
   *  (if it is a clone, typically), if this instance is registered has the
   *  primary group of another-ACS userID, or if it is a conditional group in an
    * AclEntry which is not well finalized
   */
   public synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call of GroupIDImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (folderMember_ == null ) {
          //super.finalizeForBase();
          return;
    }
    
    // if this group is a conditional group in an internal ACL, finalize the ACL
    ImmutableAclEntry[] l_acl = getEorACS().getEorL_AclEntries();
    if (l_acl != null) {
        for (int i = 0; i < l_acl.length; i++) {
           if (l_acl[i] != null && !l_acl[i].isEmpty() && 
                   l_acl[i].getL_ConditionSourceGroups().contains(this)) 
               // remove the ACL
               ((ResourceImpl)l_acl[i].getEorResource()).removeACL(
                       l_acl[i].getEorACS(),
                       l_acl[i].getEorEP(), 
                       l_acl[i].getPositiveRight(),
                       l_acl[i].getL_ConditionSourceGroups().iterator().next());
        } // for
    } // if

    try {
      if (l_Members_ != null) setEorL_Members(null);
    } catch (CreateError c) {
      c.printStackTrace();
    }

    // set groupID_ and, for Actors, currentGroupID_ and secondary groups, to null
    // in all the resources that are connected to this groupID
    if (l_ownResources_ != null ) {
      for (int i = 0 ; i < l_ownResources_.length ; i++) {
        if (l_ownResources_[i] != null ) ((ResourceImpl)l_ownResources_[i]).removeEorGroupID();
      } // for i
    }
    if (l_currentActors_ != null ) {
      for (int i = 0 ; i < l_currentActors_.length ; i++) {
        if (l_currentActors_[i] != null ) ((ActorImpl)l_currentActors_[i]).removeEorCurrentGroupID();
      } // for i
    }
    
    // set primGroup_  to null in all the relevant userIDs
    Iterator ite = getEorACS().getEorM_OwnEligibleParties().values().iterator();
    for (Iterator iter = ite ; iter.hasNext() ; ) {
      EligibleParty ep = (EligibleParty)iter.next();
      if (ep != null && ep instanceof UserID) {
        if ( ((UserID)ep).getEorPrimGroup() == this) ((UserIDImpl)ep).removePrimGroup();
      } // if (ep != null)
    } // for

    // remove this in Actor.setEorL_SecondaryGroup()
    if (l_secGroupActors_ != null ) {
      for (int i = 0 ; i < l_secGroupActors_.length ; i++) {
        if (l_secGroupActors_[i] != null )
          ((ActorImpl)l_secGroupActors_[i]).removeSecondaryGroup(this);
      } // for i
    }

    l_ownResources_ = null;
    l_currentActors_ = null;
    l_secGroupActors_ = null;
    l_Members_ = null;
    
    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * This method finalizes each group members, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null.
  * Does not call finalizeForBase(). Called only by EPRViewInBaseImpl.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In GroupIDImpl, call of " + name_ +
                    //".finalizeFromDeserialization()");

    // noop if it is the second call to this method
    if (folderMember_ == null ) {
          return;
    }

    // if this group is a conditional group in an internal ACL, finalizes the resource
    ImmutableAclEntry[] l_acl = aCS_.getEorL_AclEntries();
    if (l_acl != null) {
        for (int i = 0; i < l_acl.length; i++) {
           if (l_acl[i] != null && !l_acl[i].isEmpty() &&
                   l_acl[i].getL_ConditionSourceGroups() != null &&
                   l_acl[i].getL_ConditionSourceGroups().contains(this))
               // removes the ACL and its resource
               ((ResourceImpl)l_acl[i].getEorResource()).finalizeFromDeserialization();
        } // for
    } // if

    // finalizes the members
    for (int  i=0 ; i < l_Members_.length ; i++) {
      if (l_Members_[i] != null )
            ((GroupIDMemberImpl)l_Members_[i]).finalizeFromDeserialization();
    } // for

    // finalizes all the resources and current actors that are connected to this groupID
    if (l_ownResources_ != null ) {
      for (int i = 0 ; i < l_ownResources_.length ; i++) {
        if (l_ownResources_[i] != null ) 
            ((ResourceImpl)l_ownResources_[i]).finalizeFromDeserialization();
      } // for i
    }
    if (l_currentActors_ != null ) {
      for (int i = 0 ; i < l_currentActors_.length ; i++) {
        if (l_currentActors_[i] != null ) 
            ((ActorImpl)l_currentActors_[i]).finalizeFromDeserialization();
      } // for i
    }

    // finalizes the secondary group actors
    if (l_secGroupActors_ != null ) {
      for (int i = 0 ; i < l_secGroupActors_.length ; i++) {
        if (l_secGroupActors_[i] != null )
          ((ActorImpl)l_secGroupActors_[i]).finalizeFromDeserialization();
      } // for i
    }

    l_ownResources_ = null;
    l_currentActors_ = null;
    l_secGroupActors_ = null;
    l_Members_ = null;

    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  /**
  * Clone without register the new object in the ACSs.
  * Caution: don't use any modifying method that may altere the integrity
  * of the base.
  * @exception CloneNotSupportedException
  * @return Object; is not registred in the ACSs. The names, the type, the comment,
  * the alias, the order are cloned.
  * The external object references are not cloned (see "Eor" in get method).
  */
  public Object clone() {
      GroupIDImpl clon = (GroupIDImpl) super.clone();

            // for testing and tracking
            //System.out.println("call of GroupIDImpl.clone() for "+getName());


      if (l_Members_ != null) {
        clon.l_Members_ = new GroupIDMemberImpl[l_Members_.length];
        System.arraycopy(l_Members_, 0, clon.l_Members_, 0, l_Members_.length);
      } else l_Members_ = new GroupIDMemberImpl[0]; // for a finalized instance

      clon.setOrder(order_);

      if (l_ownResources_ != null) {
        clon.l_ownResources_ = new ResourceImpl[l_ownResources_.length];
        System.arraycopy(l_ownResources_, 0, clon.l_ownResources_, 0, l_ownResources_.length);
      } else clon.l_ownResources_ = null;

      if (l_currentActors_ != null) {
        clon.l_currentActors_ = new ActorImpl[l_currentActors_.length];
        System.arraycopy(l_currentActors_, 0, clon.l_currentActors_, 0, l_currentActors_.length);
      } else clon.l_currentActors_ = null;

      if (l_secGroupActors_ != null) {
        clon.l_secGroupActors_ = new ActorImpl[l_secGroupActors_.length];
        System.arraycopy(l_secGroupActors_, 0, clon.l_secGroupActors_, 0, l_secGroupActors_.length);
      } else clon.l_secGroupActors_ = null;

      return clon;
  } // clone()

  /**
  * @param _obj object to compare
  * @return true if _obj is an instance of GroupIDImpl (or a derived class)
  *  equals to this instance (with the same class)
  */
  public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      GroupID obj = (GroupID) _obj;


      if (order_ == 0 || obj.getOrder() == 0) {
          if (order_ != 0 || obj.getOrder() != 0)
            return false;
      } // if
      else {
          if (order_ != obj.getOrder() )
            return false;
      } // else

      if (l_Members_ == null || obj.getEorL_Members() == null) {
        if (l_Members_ != null || obj.getEorL_Members() != null)
          return false;
      } // if
      else if (l_Members_.length != obj.getEorL_Members().length)
        return false;
      else {
        for (int i = 0; i < l_Members_.length; i++) {
          if (l_Members_[i] == null || obj.getEorL_Members()[i] == null) {
              if (l_Members_[i] != null || obj.getEorL_Members()[i] != null)
                return false;
              continue; // i++
          } // if
          else {
                      // test of names
              if (!l_Members_[i].getName().equals( obj.getEorL_Members()[i].getName() ))
                return false;
                      // test of ACS
              if (!l_Members_[i].getEorACS().getName().equals( obj.getEorL_Members()[i].getEorACS().getName() ))
                return false;
          } // else
        } // for
      } // else


      if (l_ownResources_ == null || obj.getEorL_GroupIDOwnResources() == null) {
        if (l_ownResources_ != null || obj.getEorL_GroupIDOwnResources() != null)
          return false;
      } // if
      else {
        if (l_ownResources_ != null && obj.getEorL_GroupIDOwnResources() != null) {
            if (l_ownResources_.length != obj.getEorL_GroupIDOwnResources().length)
                return false;
            for (int i = 0; i < l_ownResources_.length; i++) {
                if (l_ownResources_[i] == null || obj.getEorL_GroupIDOwnResources()[i] == null) {
                    if (l_ownResources_[i] != null || obj.getEorL_GroupIDOwnResources()[i] != null)
                        return false;
                    continue; // i++
                } // if
                else {
                    if (!l_ownResources_[i].equals( obj.getEorL_GroupIDOwnResources()[i] ))
                        return false;
                } // else
            } // for
        } // if
      } // else


      if (l_currentActors_ == null || obj.getEorL_GroupIDCurrentActors() == null) {
        if (l_currentActors_ != null || obj.getEorL_GroupIDCurrentActors() != null)
          return false;
      } // if
      else {
        if (l_currentActors_ != null && obj.getEorL_GroupIDCurrentActors() != null) {
            if (l_currentActors_.length != obj.getEorL_GroupIDCurrentActors().length)
                return false;
            for (int i = 0; i < l_currentActors_.length; i++) {
                if (l_currentActors_[i] == null || obj.getEorL_GroupIDCurrentActors()[i] == null) {
                    if (l_currentActors_[i] != null || obj.getEorL_GroupIDCurrentActors()[i] != null)
                        return false;
                    continue; // i++
                } // if
                else {
                    if (!l_currentActors_[i].equals( obj.getEorL_GroupIDCurrentActors()[i] ))
                        return false;
                } // else
            } // for
        } // if
      } // else

      if (l_secGroupActors_ == null || obj.getEorL_GroupIDCurrentActorsAsSecondaryGroup() == null) {
        if (l_secGroupActors_ != null || obj.getEorL_GroupIDCurrentActorsAsSecondaryGroup() != null)
          return false;
      } // if
      else {
        if (l_secGroupActors_ != null && obj.getEorL_GroupIDCurrentActorsAsSecondaryGroup() != null) {
            if (l_secGroupActors_.length != obj.getEorL_GroupIDCurrentActorsAsSecondaryGroup().length)
                return false;
            for (int i = 0; i < l_secGroupActors_.length; i++) {
                if (l_secGroupActors_[i] == null || obj.getEorL_GroupIDCurrentActorsAsSecondaryGroup()[i] == null) {
                    if (l_secGroupActors_[i] != null || obj.getEorL_GroupIDCurrentActorsAsSecondaryGroup()[i] != null)
                        return false;
                    continue; // i++
                } // if
                else {
                    if (!l_secGroupActors_[i].equals( obj.getEorL_GroupIDCurrentActorsAsSecondaryGroup()[i] ))
                        return false;
                } // else
            } // for
        } // if
      } // else

      return true;
  } // equals()

   /**
   * @return the hashcode derived from the order, name of the first groupID and the first member,
   * the type and folder member hashcodes
   */
   public int hashCode() {
    int result = super.hashCode();

    result = PRIME*result + order_;

    if (l_Members_ != null && l_Members_.length != 0 && l_Members_[0] != null)
      result = PRIME*result + l_Members_[0].getName().hashCode();

    return result;
  }

  /**
  * @return String: the variables of the instance : name (last component).
  *  May be null.
  */
  public String toString() {
/*    StringBuilder sb = new StringBuilder();

        // else, 'null' is added to sb
    if (getName().lastComponent() != null) sb.append("name: "+name_.lastComponent() );

    sb.append(" | groupID order : " );
    sb.append(order_ );

    sb.append("GROUPIDIMPL :");
    sb.append(super.toString());
    sb.append("END of ELIGIBLEPARTYIMPL");

    sb.append("; l_Members_[0].getName() = ");
    if(l_Members_.length != 0 && l_Members_[0] != null)
      sb.append(l_Members_[0].getName().toString());

    sb.append(", l_Members_[l_Members_.length - 1].getName() = ");
    if(l_Members_.length != 0 && l_Members_[l_Members_.length - 1] != null)
      sb.append(l_Members_[l_Members_.length - 1].getName().toString());


    sb.append("; l_ownResources_ =");
    for (int i = 0; i < l_ownResources_.length; i++) {
        sb.append(l_ownResources_[i]);
        sb.append(", ");
    } //for

    sb.append("; l_currentActors_ =");
    for (int i = 0; i < l_currentActors_.length; i++) {
        sb.append(l_currentActors_[i]);
        sb.append(", ");
    } //for

    sb.append("; l_secGroupActors_ =");
    for (int i = 0; i < l_secGroupActors_.length; i++) {
        sb.append(l_secGroupActors_[i]);
        sb.append(", ");
    } //for

    return " [[[" + sb + " end of GROUPIDIMPL ]]]";
    return sb;
*/
    return getName().lastComponent() ;
  } // toString()

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  //
  // -------------------- STATIC PUBLIC METHODS ---------------------------------
  //


  /**
   * Gets, for a conditional AclEntry, all the effective or potential intermediate GroupIDs,
   * in the chain from the source to each condition group.
   * If the AclEntry source is not a member of one condition group,
   * returns the near groups for which the source is an indirect member. 
   * The distance of these groups to the ACL source (number of intermediate groups)
   * has to be inferior to 5.
   * Calls AclEntry.getL_ConditionSourceGroups(), containsAsMember(),
   * containsAsIndirectMember(), getFirstIntermediateForIndirectMember().
   * Called by getL_ConditionalACLsOfOneConditionGroupIndirectMember() to detect
   * the need to call the method AclEntryImpl.resetRightsAsInheritedAclEntry(),
   * and by Gui2.GraphicViewBaseListenerImpl and gWork.RighsMediatorImpl
   * to listen the member changes in the condition groups.
   * @param _acl AclEntry to analyze
   * @return array of intermediate GroupIDs. Null if it is not a conditional AclEntry,
   * the source is not a GroupIDMember, and empty if the source is a direct member
   * of each condition group.
   */
    public static ImmutableGroupID[] getL_IntermediateGroupsOfAclConditionalGroups(
            ImmutableAclEntry _acl) {

        if (!_acl.isConditionalAclEntry() || !(_acl.getEorEP() instanceof GroupIDMember))
            return null;

        HashSet<ImmutableGroupID> l_result = new HashSet(INITIAL_CAPACITY);
        GroupID             grp = null;
        GroupID             grp2 = null;
        ImmutableGroupID[]  l_grps = null;
        ImmutableGroupID[]  l_grps2 = null;
        ImmutableGroupID[]  l_grps3 = null;
        ImmutableGroupID[]  l_grps4 = null;
        ImmutableGroupID[]  l_grps5 = null;
        boolean             bottomUpIsDone = false;

        for (Iterator ite = _acl.getL_ConditionSourceGroups().iterator(); ite.hasNext(); ) {
            grp = (GroupIDImpl)ite.next();

            if (grp == null || grp.containsAsMember((GroupIDMember)_acl.getEorEP()) )
                continue;

            if (!grp.containsAsIndirectMember((GroupIDMember)_acl.getEorEP()) && !bottomUpIsDone) {

                bottomUpIsDone = true;

                l_grps = ((GroupIDMember)_acl.getEorEP()).getEorL_GroupID();
                // returns only the five first levels of groups from ep_
                for (int j = 0; j < l_grps.length; j++) {
                    l_result.add(l_grps[j]);
                    l_grps2 = l_grps[j].getEorL_GroupID();

                    for (int i = 0 ; i < l_grps2.length ; i++ ) {
                        l_result.add(l_grps2[i]);
                        l_grps3 = l_grps2[i].getEorL_GroupID();

                        for (int k = 0 ; k < l_grps3.length ; k++ ) {
                            l_result.add(l_grps3[k]);
                            l_grps4 = l_grps3[k].getEorL_GroupID();

                            for (int u = 0 ; u < l_grps4.length ; u++ ) {
                                l_result.add(l_grps4[u]);
                                l_grps5 = l_grps4[u].getEorL_GroupID();

                                for (int v = 0 ; v < l_grps5.length ; v++ ) {
                                    l_result.add(l_grps5[v]);
                                } // for v
                            } // for u
                        } // for k
                    } // for i
                } // for j

            }
            else {
                while (grp != null) {
                    // limit: if there are several member paths, the first one is detected
                    grp2 = grp.getFirstIntermediateForIndirectMember(
                            (GroupIDMember)_acl.getEorEP());
                    if (grp2 != null)
                        l_result.add(grp2);
                    grp = grp2;
                } // while
            } // else

        } // for ite

        grp = null;
        grp2 = null;
        l_grps = null;
        l_grps2 = null;
        l_grps3 = null;
        l_grps4 = null;
        l_grps5 = null;

        return l_result.toArray(new ImmutableGroupID[0]);
    } // getL_IntermediateGroupsOfAclConditionalGroups()

  /**
   * NO USE. Gets, for a group, all the conditional AclEntries of the same ACS
   * for which the group is in the membership chain from itself to at least one condition group.
   * The aim is to return all the active AclEntries for which the rights
   * have to be analyzed because a relevant membership relation may have changed.
   * If the ACL inner rights are null (possible only for a flexible conditional ACL)
   * the ACL is not returned, since it is not an active ACL.
   * Calls ACS.getEorL_AclEntries() and getL_IntermediateGroupsOfAclConditionalGroups().
   * @param _grp to analyze. Never null.
   * @return conditional AclEntries for which _grp is in the membership chain
   * from itself to at least one condition group. May be null.
   */
    public static AclEntry[] getL_ConditionalACLsOfOneConditionGroupIndirectMember(
            ImmutableGroupID _grp) {

        if (!_grp.getEorACS().getManageFlexibleConditionalAcls() &&
                !_grp.getEorACS().getManageRightDefinedConditionalAcls())
            return null;

        HashSet<ImmutableAclEntry>  l_result = null;
        ImmutableAclEntry[]         l_acl = _grp.getEorACS().getEorL_AclEntries();

        for (int i = 0; i < l_acl.length; i++ ) {
            if (!l_acl[i].isConditionalAclEntry() || l_acl[i].getL_Rights().length == 0)
                continue;

            if (!Arrays.asList(getL_IntermediateGroupsOfAclConditionalGroups(
                    l_acl[i])).contains(_grp))
                continue;

            if (l_result == null)
                l_result = new HashSet(INITIAL_CAPACITY);

            l_result.add(l_acl[i]);
        } // for i

        l_acl = null;

        if (l_result == null)
            return null;
        else
            return l_result.toArray(new AclEntryImpl[0]);
    } // getL_ConditionalACLsOfOneConditionGroupIndirectMember()

  /**
   * Detects if it is a conditional AclEntry from this group ACS,
   * for which a change in the groupidmember membership relation to this group is able to
   * change the membership chain from the AclEntry source to at least one condition group
   * of the AclEntry. This group ACS has to manage the conditional Aclentries,
   * and this is checked by this method.
   * If the method returns true, the method AclEntryImpl.resetRightsAsInheritedAclEntry()
   * has to be called because the AclEntry effective rights may have changed, after
   * the membership change in this group.
   * Returns false if there is no link from _acl.getEorEP() to _member.
   * Returns false if there is no link from one _acl condition group to this group.
   * Whatever the sens of the groupidmember membership change, that is an
   * adding or a removing of _member in this group, this method says the AclEntry rights
   * are changing. The simplest case is processed, where _member is the Acl source
   * and this group is an _acl condition group.
   * This method is faster than a suite of tests having the form
   * conditionGroup.containsAsIndirectMember(aclSource)
   * before and after the groupidmember membership change.
   * Calls containsAsMember(), containsAsIndirectMember(), and not
   * containsAsIndirectMemberWithoutAliasing().
   * Called by add/removeEorMember().
   * @param _acl from the _member ACS. Never null.
   * @param _member new or previous direct member of this group. Never null.
   * @return true if _acl has to be reset through AclEntryImpl.resetRightsAsInheritedAclEntry().
   * Return false in all other cases, including an ACS which does not manage the conditional ACL,
   * or a non-conditional _acl, or no identity or membership relation from the _acl source
   * to _member, or no membership relation from this group to any _acl condition group.
   */
    public boolean isConditionalACLtoResetAfterMembershipChangeInThisGroup(
            ImmutableAclEntry _acl, ImmutableGroupIDMember _member) {

        if (!_acl.isConditionalAclEntry())
            return false;
        if (!(_acl.getEorEP() instanceof ImmutableGroupIDMember))
            return false;

        GroupIDMember   aclSource = (GroupIDMember)_acl.getEorEP();
        GroupID         grp = null;
        boolean         result = false;

        // false if there is no link from _acl.getEorEP() to _member
        if (aclSource != _member) {
            if (!(_member instanceof GroupID))
                return false;
            else if (!((GroupID)_member).containsAsMember(aclSource) &&
                    !((GroupID)_member).containsAsIndirectMember(aclSource))
                return false;
        } // if (aclSource != _member)

        // false if there is no link from one _acl condition group to this group
        for (Iterator ite = _acl.getL_ConditionSourceGroups().iterator(); ite.hasNext(); ) {
            grp = (GroupIDImpl)ite.next();

            if (grp != null &&
                    (grp == this || grp.containsAsMember(this) ||
                        grp.containsAsIndirectMember(this))) {
                result = true;
                break;
            } // if
        } // for ite

        aclSource = null;
        grp = null;
        return result;
    } // isConditionalACLtoResetAfterMembershipChangeInThisGroup()

   //
   // ---------------------- PROTECTED METHODS --------------------------------
   //

  /**
  * This method provides a copy of the instance variables.
  * Copies the own actors and the result as groupIDMember
  * (done by GroupIDMemberImpl.copyOwnVariables()).
  * Does not copy the groupIDMembers (done by the copy() caller).
  * Does not copy the own resources, nor
  * the current actors,
  * but the resource and actor copies do the job.
  * The copy order is 3 times the source order.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError 
  **/
  protected GroupIDImpl copyOwnVariables(GroupIDImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in GroupIDImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      _result = (GroupIDImpl)super.copyOwnVariables(_result, _withoutInternalAcl);
           
      _result.order_ = getOrder()*3;
      
      // add _result as a secondary group
      try {
          if (getEorL_GroupIDCurrentActorsAsSecondaryGroup().length != 0) {
            Actor[] l_actors = getEorL_GroupIDCurrentActorsAsSecondaryGroup();
            ImmutableGroupID[] l_groups = null;
            ImmutableGroupID[] l_newGroups = null;
            for (int i = 0; i < l_actors.length; i++) {
                l_groups = l_actors[i].getEorL_SecondaryGroup();
                l_newGroups = new ImmutableGroupID[l_groups.length + 1];
                System.arraycopy(l_groups, 0, l_newGroups, 0, l_groups.length);
                l_newGroups[l_groups.length] = _result;
                ((ActorImpl)l_actors[i]).setEorL_SecondaryGroup(l_newGroups);
            } // for i
          } // if
      } catch (UpDateError u) {
        throw new CreateError(u.toString());
      }

      return _result;
  } // copyOwnVariables(,,,)
  
    /**
    * Used by Resource.setEorGroupID().
    * Fire an 'l_GroupIDOwnResources' event.
    * @param _res Resource owned by this groupID. May be null.
    */
    protected void addOwnResource(Resource _res) {
      // variables
      Resource[]  newArray = new ResourceImpl[1];
      boolean     recorded = false;

      if (l_ownResources_ == null) l_ownResources_ = (Resource[]) new ResourceImpl[1];

      Resource[]  oldValue = new ResourceImpl[l_ownResources_.length];
      System.arraycopy(l_ownResources_, 0, oldValue, 0, l_ownResources_.length);

      // add in an empty place if possible
      for (int i = 0 ; i < l_ownResources_.length ; i++) {
          if (l_ownResources_[i] == null ||
            l_ownResources_[i].equals(ResourceImpl.EMPTY_INSTANCE) ) {
              l_ownResources_[i] = _res;
              recorded = true;
              break;
          }
      }
      // else, add at the end of l_ownResources_
      if (!recorded) {
          newArray = new ResourceImpl[l_ownResources_.length + 1];
          System.arraycopy(l_ownResources_, 0, newArray, 0, l_ownResources_.length);
          newArray[l_ownResources_.length] = _res;
          l_ownResources_ = new ResourceImpl[newArray.length];
          l_ownResources_ = newArray;
      }
      else { // copy to newArray
          newArray = new ResourceImpl[l_ownResources_.length];
          System.arraycopy(l_ownResources_, 0, newArray, 0, l_ownResources_.length);
      }

      try {
          firePropertyChange("l_GroupIDOwnResources", oldValue, getEorL_GroupIDOwnResources());
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.addOwnResource(), firePropertChange() throws:");
            e.printStackTrace();
      }
      
      newArray = null;
    } // addOwnResource()

    /**
    * Used by Actor.setEorCurrentGroupID.
    * Fire an event "l_GroupIDCurrentActors".
    * @param _act Actor currently attached to this groupID.  May be null.
    */
    protected void addCurrentActor(Actor _act) {
      // variables
      Actor[]     newArray = new ActorImpl[1];
      boolean     recorded = false;

      if (_act == null) return;
      if (l_currentActors_ == null) l_currentActors_ = (Actor[]) new ActorImpl[1];

      Actor[]     oldValue = new ActorImpl[l_currentActors_.length];
      System.arraycopy(l_currentActors_, 0, oldValue, 0, l_currentActors_.length);

      // add in an empty place if possible
      for (int i = 0 ; i < l_currentActors_.length ; i++) {
          if (l_currentActors_[i] == null || l_currentActors_[i].equals(ActorImpl.EMPTY_INSTANCE) ) {
              l_currentActors_[i] = _act;
              recorded = true;
              break;
          }
      }
      // else, add at the end of l_currentActors_
      if (!recorded) {
          newArray = new ActorImpl[l_currentActors_.length + 1];
          System.arraycopy(l_currentActors_, 0, newArray, 0, l_currentActors_.length);
          newArray[l_currentActors_.length] = _act;
          l_currentActors_ = new ActorImpl[newArray.length];
          l_currentActors_ = newArray;
      }
      else {
          newArray = new ActorImpl[l_currentActors_.length];
          System.arraycopy(l_currentActors_, 0, newArray, 0, l_currentActors_.length);
      }

      try {
          firePropertyChange("l_GroupIDCurrentActors", oldValue, newArray);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.addcurrentActor(), firePropertChange() throws:");
            e.printStackTrace();
      } 
      
      newArray = null;
    } // addCurrentActor()

   /**
   * Used by Resource.setEorGroupID() and finalizeForBase().
   * Fire an event on "l_GroupIDOwnResources".
   * @param _res Resource to remove in this groupID
   * @exception UpDateError if _res is null or unknown
   */
   protected void removeOwnResource(Resource _res) throws UpDateError {
    // noop if finalizeForBase() has been called
    if (l_ownResources_ == null) return;

    Resource[]  oldValue = new ResourceImpl[l_ownResources_.length];
    System.arraycopy(l_ownResources_, 0, oldValue, 0, l_ownResources_.length);

    // variables and initializations
    short           count = 0;
    short           shiftingBack = 0;
    // l_ownResources_ can't be null when this method is invoked, so:
    Resource[]      newArray = new ResourceImpl[l_ownResources_.length];
    Resource[]      result = new ResourceImpl[1];

    // count the occurences of _res:
    for (int i = 0 ; i < l_ownResources_.length ; i++) {
      if (l_ownResources_[i] != null ) {
        if (l_ownResources_[i].equals(_res) ) count++;

                // for testing and tracking
                //System.out.println("In GroupIDImpl.removeOwnResource(), "+
                //  " an initial resource is "+l_ownResources_[i]+", and _res is "+_res);

      } // if
    } // for i

    // exceptions
    if (_res == null) {
      getEorACS().getEorBaseError().addException("In GroupIDImpl."+
       "removeOwnResource(_res), _res == null ");
      throw new UpDateError("In GroupIDImpl.removeOwnResource(_res), _res == null ");
    }
    if (count == 0 && l_ownResources_.length != 0) {
        getEorACS().getEorBaseError().addException("In GroupIDImpl."+
          "removeOwnResource(_res), _res is unknown ");
        throw new UpDateError("In GroupIDImpl.removeOwnResource(_res), _res is unknown ");
    }

    // updates l_ownResources_
    for (int  i = 0 ; i < l_ownResources_.length ; i++) {
      if (l_ownResources_[i] != null) {
        // incrementing shiftingBack deletes the reference l_ownResources_[i]:
        if (l_ownResources_[i].equals(_res) ) shiftingBack++;
        else newArray[i - shiftingBack] = l_ownResources_[i];
      }
      else shiftingBack++; // delete null l_ownResources_[i]
    } // for

    if (l_ownResources_.length != shiftingBack) {
      result = new ResourceImpl[l_ownResources_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
      l_ownResources_ = new ResourceImpl[result.length];
      System.arraycopy(result, 0, l_ownResources_, 0, result.length);
    } // if
    else {
      l_ownResources_ = null;
      result = null;
    } // else

    try {
        firePropertyChange("l_GroupIDOwnResources", oldValue, getEorL_GroupIDOwnResources());
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.removeOwnResource(), firePropertChange() throws:");
            e.printStackTrace();
    }
    
    oldValue = null;
    newArray = null;
    result = null;
  } // removeOwnResource()

   /**
   * Used by Actor.finalizeForBase(), Actor.setEorCurrentGroupID().
   * Fire an event.
   * @param _act Actor to remove in this groupID
   * @exception UpDateError if _act is null or unknown
   */
   protected void removeCurrentActor(Actor _act) throws UpDateError {
    // noop if finalizeForBase() has been called
    if (l_currentActors_ == null) return;

    Actor[]     oldValue = new ActorImpl[l_currentActors_.length];
    System.arraycopy(l_currentActors_, 0, oldValue, 0, l_currentActors_.length);

    // variables and initializations
    short        count = 0;
    short        shiftingBack = 0;
    // l_currentActors_ can't be null when this method is invoked, so:
    Actor[]      newArray = new ActorImpl[l_currentActors_.length];
    Actor[]      result = new ActorImpl[1];

    // count the occurences of _act:
    for (int i = 0 ; i < l_currentActors_.length ; i++) {
      if (l_currentActors_[i] != null ) {
        if (l_currentActors_[i].equals(_act) ) count++;
      }
    } // for i

    // exceptions
    if (_act == null) {
      getEorACS().getEorBaseError().addException("In UserIDImpl.removeCurrentActor(_act), _act == null ");
      throw new UpDateError("In UserIDImpl.removeCurrentActor(_act), _act == null ");
    }
    if (count == 0 && l_currentActors_.length != 0) {
        getEorACS().getEorBaseError().addException("In UserIDImpl.removeCurrentActor(_act), _act is unknown ");
        throw new UpDateError("In UserIDImpl.removeCurrentActor(_act), _act is unknown ");
    }

    // update l_currentActors_
    for (int  i = 0 ; i < l_currentActors_.length ; i++) {
      if (l_currentActors_[i] != null) {
        // incremente shiftingBack delete the reference l_currentActors_[i]:
        if (l_currentActors_[i].equals(_act) ) shiftingBack++;
        else newArray[i - shiftingBack] = l_currentActors_[i];
      }
      else shiftingBack++; // delete null l_currentActors_[i]
    } // for

    if (l_currentActors_.length != shiftingBack) {
      result = new ActorImpl[l_currentActors_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
      l_currentActors_ = new ActorImpl[result.length];
      System.arraycopy(result, 0, l_currentActors_, 0, result.length);
    } // if
    else {
      l_currentActors_ = null;
      result = null;
    } // else

    try {
        firePropertyChange("l_GroupIDCurrentActors", oldValue, getEorL_GroupIDCurrentActors());
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.removeCurrentActor(), firePropertChange() throws:");
            e.printStackTrace();
    }
    
    oldValue = null;
    newArray = null;
    result = null;
  } // removeCurrentActor()

   /**
   * Invoked by Actor.setEorL_SecondaryGroup() and Actor.finalizeForBase().
   * Fires an event 'l_GroupIDCurrentActorsAsSecondaryGroup'.
   * @param _act Actor to remove in this groupID
   * @exception UpDateError if _act is null or unknown, but not if the actor
   *  list is null
   */
   protected void removeOwnActorAsSecondaryGroup(Actor _act) throws UpDateError {

    if (l_secGroupActors_ == null) return; // possible with finalizeForBase()

    // for the event 'l_GroupIDOwnActors'
    Actor[]     oldValue = new ActorImpl[l_secGroupActors_.length];
    System.arraycopy(l_secGroupActors_, 0, oldValue, 0, l_secGroupActors_.length);

    // variables and initializations
    short        count = 0;
    short        shiftingBack = 0;
    Actor[]      newArray = new ActorImpl[l_secGroupActors_.length];
    // for the event 'l_GroupIDOwnActors'
    Actor[]      newValue = new ActorImpl[1];

    // count the occurences of _act
    for (int i = 0 ; i < l_secGroupActors_.length ; i++) {
      if (l_secGroupActors_[i] != null ) {
        if (l_secGroupActors_[i].equals(_act) ) count++;
      }
    } // for i

    // exceptions
    if (_act == null) {
      getEorACS().getEorBaseError().addException(
        "In UserIDImpl.removeCurrentActor(_act), _act == null ");
      throw new UpDateError(
        " In UserIDImpl.removeCurrentActor(_act), _act == null ");
    }
    if (count == 0 && l_secGroupActors_.length != 0) {
        getEorACS().getEorBaseError().addException(
          "In UserIDImpl.removeCurrentActor(_act), _act is unknown ");
        throw new UpDateError(
          " In UserIDImpl.removeCurrentActor(_act), _act is unknown ");
    }

    // updates l_secGroupActors_
    for (int  i = 0 ; i < l_secGroupActors_.length ; i++) {
      if (l_secGroupActors_[i] != null) {
        // incremente shiftingBack delete the reference l_secGroupActors_[i]:
        if (l_secGroupActors_[i].equals(_act) ) {
          shiftingBack++;

                // for testing and tracking
                //System.out.println("In GroupIDImpl.removeOwnActor(), "+
                  //" the actor "+_act+" is removed.");
        }
        else newArray[i - shiftingBack] = l_secGroupActors_[i];
      }
      else shiftingBack++; // delete null l_secGroupActors_[i]
    } // for

    if (l_secGroupActors_.length != shiftingBack) {
      newValue = new ActorImpl[newArray.length - shiftingBack];
      System.arraycopy(newArray, 0, newValue, 0, newArray.length - shiftingBack);
      l_secGroupActors_ = new ActorImpl[newValue.length];
      System.arraycopy(newValue, 0, l_secGroupActors_, 0, newValue.length);
    } // if
    else {
      l_secGroupActors_ = null;
      newValue = null;
    } // else

    try {
        firePropertyChange("l_GroupIDCurrentActorsAsSecondaryGroup",
          oldValue, getEorL_GroupIDCurrentActorsAsSecondaryGroup());
    } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From GroupIDImpl.removeOwnActor(), firePropertChange() throws:");
            e.printStackTrace();
    }
    
    oldValue = null;
    newArray = null;
    newValue = null;
  } // removeOwnActor()


  /**
  * This method removes the groupID from the ACS lists, but there is no
  * guaranee that the JVM calls this method before the end of the application.
  * @exception java.lang.Throwable
  * @exception  Throwable
  */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {
      super.finalize();
      throw new InternalError("in GroupImpl.finalize(): "+u.toString());
    }

    super.finalize();
  }

} // GroupIDImpl

