//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.GroupIDMember;
import ARoad0.gBaseInterface.Name;
import ARoad0.gBaseInterface.GroupID;
import ARoad0.gBaseInterface.EligibleParty;
import ARoad0.gBaseInterface.Directory;
import ARoad0.gBaseInterface.ACS;
import ARoad0.gBaseInterface.ImmutableGroupID;
import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;

import java.io.Serializable;
import java.io.ObjectStreamField;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
* This class is responsible for modeling the entity which is a GroupID member.
* This class offers a method to get the groups the instance is member of.
* It is a javabean with the following bound properties:
* 'AssociatedGroups'.
* The listeners are proper to this instance, outside gBase,
* transient (not serialized in this instance backup).
* They are called in any order. They receive only a copy of the new value,
* to protect the property. All the exceptions from the listeners are catched,
* and a dialog box is displayed to inform the user.<P>
* This class has several subclasses in this package and in the AcsAddon packages.
* Most of the variables are protected.
* @version 0.7.0
*/
public class GroupIDMemberImpl extends EligiblePartyImpl implements
  Serializable, GroupIDMember {
  private static final long   serialVersionUID__ = 0;

  /**
   * true only if the instance has an administrative role in the ACS, for
   * the technical administration or the users administration.
   */
  protected boolean     isAdmin_ = false;
    
  // private variables

  /**
   * array of groupIDs that own directly this GroupIDMember.
   * Initialized to new GroupIDImpl[0]
   */
  protected GroupIDImpl[]   l_Groups_ = new GroupIDImpl[0];

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final GroupIDMemberImpl EMPTY_INSTANCE = new GroupIDMemberImpl();

  /** variable for the JDK 2 serialization
  * ("l_Groups_",GroupIDImpl[].class),
  * ("isAdmin_", Boolean.TYPE)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("l_Groups_",GroupIDImpl[].class),
      new ObjectStreamField("isAdmin_", Boolean.TYPE)
  };

 /**
  * Simple constructor that calls EligiblePartyImpl().
  */
  public GroupIDMemberImpl() {
  }

   /**
   * Constructor used in the beamer by the editor for the ACS EP list, from 'New'.
   * @param _name ImmutableName of the GroupIDMember as a kind of eligible party
   * @param _acs ACS owning the GroupIDMember
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users administration.
   * @exception CreateError if ACSImpl.newEorEligibleParty() throws it - the instance
   * must then be deleted !
   */
  public GroupIDMemberImpl(ImmutableName _name, ACS _acs, String _epType, Boolean _isAdmin)
   throws CreateError  {
      this(_name, _acs, _epType, _isAdmin.booleanValue(), true);

        // for testing and tracking
        //System.out.println("in GroupIDMemberImpl constructor, l_Groups_ length: "+
          //l_Groups_.length);
  }

   /**
   * Used for the creation of a GroupIDMember alias, without base registration.
   * @param _name ImmutableName of the eligible party to create
   * @param _acs Main ACS of the eligible party to create
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users administration.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if ACSImpl.newEorEligibleParty() throws it -
   *  the instance must then be deleted !
   */
  public GroupIDMemberImpl(ImmutableName _name, ACS _acs, String _epType,
    boolean _isAdmin, boolean _toRegisterInManager) throws CreateError {
      super(_name, _acs, _epType, _toRegisterInManager);
      isAdmin_ = _isAdmin;
  }
  
  /**
  * This method provides a copy of the instance
  * with another name.
  * Register the copy in every GroupID as member, if it is the same ACS,
  * or else in every GroupID copy which is found.
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _aliasRef this argument is not used there.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of GroupIDMemberImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Object _aliasRef, Boolean _toRegisterInManager)
    throws CreateError {

      if (_acs == null || _name == null || 
        (_aliasRef != null && !(_aliasRef instanceof EligibleParty)) ) {
          throw new CreateError("in GroupIDMemberImpl.copy(),"+
          " _name or _acs is null, or _aliasRef is not an EligibleParty.");
      }
      
      GroupIDMemberImpl result = new GroupIDMemberImpl(
          _name, _acs, getEPType(), isAdmin_,
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);
      
      return result;
  } // copy(,,,)
  
  /**
  * This method provides a copy of the instance
  * as copy() but without copying the aclEntries.
  * Register the copy in every GroupID as member, if it is the same ACS,
  * or else in every GroupID copy which is found.
  * _aliasRef may belong to another ACS than this instance ACS.
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _aliasRef this argument is not used there.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of GroupIDMemberImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Object _aliasRef, Boolean _toRegisterInManager)
    throws CreateError {

      if (_acs == null || _name == null || 
        (_aliasRef != null && !(_aliasRef instanceof EligibleParty)) ) {
          throw new CreateError("in GroupIDMemberImpl.copy(),"+
          " _name or _acs is null, or _aliasRef is not an EligibleParty.");
      }
      
      GroupIDMemberImpl result = new GroupIDMemberImpl(
          _name, _acs, getEPType(), isAdmin_,
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, true);
      
      return result;
  } // copy(,,,)
  
  /**
   * Gets the groups for which this GroupIDMember is a member.
   * @return a copy of the groups from this eligible party ACS.
   *    Never null. It may be an empty array.
   */
  public ImmutableGroupID[] getEorL_GroupID() {
    ImmutableGroupID[] result;
    if (l_Groups_ != null) {
        result = new GroupIDImpl[l_Groups_.length];
        System.arraycopy(l_Groups_, 0, result, 0, l_Groups_.length);
    }
    else
        result = new GroupIDImpl[0];
    return result;
  } // getEorL_GroupID()
  
  /**
   * True only if the group member has an administrative role in the ACS, for
   * the technical administration or the users administration.
   */
  public boolean isAdmin() {
      return isAdmin_;
  }

    /**
    * This method removes the GroupIDMember from the ACS lists, but there is no
    * guaranee that the JVM calls this method before the end of the application.
    * @exception java.lang.Throwable
    */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {
      super.finalize();
      throw new InternalError("in GroupIDMemberImpl.finalize(): "+u.toString());
    }

    super.finalize();
  }
  
   /**
   * super.finalizeForBase() removes the GroupIDMember in the ACS lists (main and controllers),
   * and removes it in the resources connected through an AclEntry.
   * This method removes the instance in the groups it is member, and before,
   * the groups array is set to null.
   * This method HAS TO BE USED before deleting the GroupIDMember.
   * Note: when this method is called by the EligiblePartyImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the GroupIDMember is unknown of the main ACS
   * (if it is a clone, typically), or if a linked immutable object fires an exception
   */
   public synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call 1 of GroupIDMemberImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (folderMember_ == null) {
      return;
    }

              // for testing purpose
              //System.out.println("call 2 of GroupIDMemberImpl.finalizeForBase() for "+getName());


    // removes (this) in every group, and use an array copy
    // to update l_Groups_ through the call to each group.removeEorMember(this).
    // no CreateError normally, since the type is not immutable.
    if (l_Groups_ != null) {
        try {
            if (l_Groups_.length != 0) {
              GroupIDImpl[] l_copy = new GroupIDImpl[l_Groups_.length];
              System.arraycopy(l_Groups_, 0, l_copy, 0, l_Groups_.length);

              l_Groups_ = null; // may be used by removeEorMember()
              for (int i = 0 ; i < l_copy.length ; i++) {
                  if (l_copy[i] != null && l_copy[i] != GroupIDImpl.EMPTY_INSTANCE) {
                      l_copy[i].removeEorMember(this);
                  }
              } // for
              l_copy = null;
            } else l_Groups_ = null;
        } catch (CreateError c) {
            throw new InternalError("GroupIDMemberImpl.finalizeForBase(): "+c);
        } // catch
    } // if (l_Groups_ != null)

    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * This method finalizes each group in the groups set, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null.
  * Does not call finalizeForBase(). Called only by EPRViewInBaseImpl.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In GroupIDMemberImpl, call of " + name_ +
                    //".finalizeFromDeserialization()");

    // noop if it is the second call to this method
    if (folderMember_ == null) {
      return;
    }
    
    // finalizes every group and calls back this instance
    if (l_Groups_.length != 0) {
      GroupIDImpl[] l_copy = new GroupIDImpl[l_Groups_.length];
      System.arraycopy(l_Groups_, 0, l_copy, 0, l_Groups_.length);

      l_Groups_ = null; // to speed up removeGroupToGM() which is invoked by the groups
      for (int i = 0 ; i < l_copy.length ; i++) {
          if (l_copy[i] != null && l_copy[i] != GroupIDImpl.EMPTY_INSTANCE) {
              l_copy[i].finalizeFromDeserialization();
          }
      } // for
      l_copy = null;
    } else l_Groups_ = null;
    
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  /**
  * Clone without register the new object in the ACSs.
  * Caution: don't use any modifying method that may altere the integrity
  * of the base.
  * @exception CloneNotSupportedException
  * @return Object; is not registred in the ACSs. The names, the type, the comment,
  * the alias, the umask, the order are cloned.
  * The external object references are not cloned (see "Eor" in get method).
  */
  public Object clone() {
      GroupIDMemberImpl clon = (GroupIDMemberImpl) super.clone();

      if (l_Groups_ != null) {
            clon.l_Groups_ = new GroupIDImpl[l_Groups_.length];
            System.arraycopy(l_Groups_, 0, clon.l_Groups_, 0, l_Groups_.length);
      }
      else
          clon.l_Groups_ = new GroupIDImpl[0];

      clon.isAdmin_ = isAdmin_;
      return clon;
  } // clone()

  /**
  * @param _obj Object to compare
  * @return true if _obj is an instance of GroupIDMemberImpl (or a derived class)
  * equals to this instance (with the same class)
  */
  public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      GroupIDMember obj = (GroupIDMember) _obj;

      // with removeGroupToEP(), l_Groups_ can't be null
      // but they should be null whatever
      if (l_Groups_ == null || obj.getEorL_GroupID() == null) {
          if (l_Groups_ != null || obj.getEorL_GroupID() != null)
            return false;
      } // if
      else {
        if (l_Groups_.length != obj.getEorL_GroupID().length) return false;
        for (int i = 0; i < l_Groups_.length; i++) {
            if (l_Groups_[i] == null | obj.getEorL_GroupID()[i] == null) {
              if (l_Groups_[i] != null | obj.getEorL_GroupID()[i] != null)
                return false;
              continue; // i++
            } // if
            else if (!l_Groups_[i].equals( obj.getEorL_GroupID()[i] ))
                    return false;
        } // for
      } // else
      
      if (isAdmin_ && !obj.isAdmin())
          return false;
      if (!isAdmin_ && obj.isAdmin())
          return false;
      
      return true;
  } // equals()

   /**
   * @return the hashcode derived from the name of the first groupID,
   * the type and folder member hashcodes
   */
  public int hashCode() {
    int result = super.hashCode();

    if (l_Groups_ != null && l_Groups_.length != 0 && l_Groups_[0] != null)
      result = PRIME*result + l_Groups_[0].getName().hashCode();

    return result;
  }

   //
   // --------------- Overridden  public methods ------------------------
   //

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'AssociatedGroups'.
  * Calls the super method.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      super.addPropertyChangeListener(_propertyName, _l);

  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'AssociatedGroups'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Calls the super method.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      super.removePropertyChangeListener(_propertyName, _l);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      return super.getPropertyChangeListeners(_propertyName);
  } // getPropertyChangeListeners()

  /**
   * @return a string from the name last component
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append(getName().lastComponent() );
    /*    sb.append("  | " );

    sb.append("; l_Groups_ :");
    for (int i = 0; i < l_Groups_.length; i++) {
        sb.append(l_Groups_[i]);
        sb.append(", ");
    } //for


    return " [[[" + sb + " end of GroupIDMemberImpl ]]]";
*/
    return sb.toString() ;
  } // toString()

  /**
   * @return the static serialVersionUID__ of this class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  //
  // ----------- PROTECTED METHODS -------------------------------------
  //
  
  /**
  * This method provides a copy of the instance variables.
  * Registers the copy in every GroupID as member, if it is the same ACS,
  * or else in every GroupID which has a sililar name - that is ending
  * to the same component.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError of l_Groups_[i].addEorMember()
  **/
  protected GroupIDMemberImpl copyOwnVariables(GroupIDMemberImpl _result,
     boolean _withoutInternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in GroupIDMemberImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      _result = (GroupIDMemberImpl)super.copyOwnVariables(_result, _withoutInternalAcl);
      
      if (_result.getEorACS() == getEorACS()) {
          for (int i = 0; i < l_Groups_.length; i++) {
              l_Groups_[i].addEorMember(_result);
          }
      } else {
          // if the copy is for another ACS, try to find the groups
          // with the same name last components.
          // there, the group name must be made of the acs name
          // ended by the group string name
          ImmutableName grp_name;
          for (int i = 0; i < l_Groups_.length; i++) {
              grp_name = ((Name)_result.getEorACS().getName()).extended( 
                l_Groups_[i].getName().lastComponent() );
              if (!_result.getEorACS().containsOwnEligiblePartyName(grp_name)) {

                            // for testing and tracking
                            //System.out.println("in GroupIDMemberImpl.copyOwnVariables(), "+
                              //"for "+getName()+", the group name "+
                              //grp_name+" is not found in "+_result.getEorACS());
                  
                  continue;
              }
              try {
                  ((GroupID)_result.getEorACS().getEorM_EligibleParties().get(grp_name)).
                    addEorMember(_result);
              } catch (CreateError c) {
                  // no operation if grp_name is not a group name
                  throw new InternalError("from GroupIDMemberImpl.copyOwnVariables(), "+
                    "no adding as member to a groupID.");
              } // catch
          } // for i 
          grp_name = null;
      } // else
      
      _result.isAdmin_ = isAdmin_;
      
      return _result;
  } // copyOwnVariables(,,,)
  
   /**
   * Adds a group for which this GroupIDMember is a member.
   * Fires changes on 'AssociatedGroups'.
   * Used only by GroupIDImpl. Overridden and called by UserIDImpl.
   * @param _grp is the groupID to add. No action if it is null or known.
   */
  public synchronized void addGroupToGM(GroupID _grp) {
    // variables
    boolean recorded = false;
    GroupIDImpl[] newArray;
    GroupIDImpl[] l_oldGroups;

    if (_grp == null) return;
    // test if _grp is not known
    for (int i = 0 ; i < l_Groups_.length ; i++) {
        if (_grp == l_Groups_[i] ) {
          recorded = true; //_grp is known
          break;
        }
    } // for i
    if (recorded) return;

    // saves the old Group list, since l_Groups_ is a bound property
    l_oldGroups = new GroupIDImpl[l_Groups_.length];
    System.arraycopy(l_Groups_, 0, l_oldGroups, 0, l_Groups_.length);

    // adds at the end of l_Groups_
    newArray = new GroupIDImpl[l_Groups_.length + 1];
    System.arraycopy(l_Groups_, 0, newArray, 0, l_Groups_.length);
    newArray[l_Groups_.length] = (GroupIDImpl)_grp;
    l_Groups_ = newArray;

    try {
        firePropertyChange("AssociatedGroups", l_oldGroups, getEorL_GroupID());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From GroupIDMemberImpl.addGroupToGM(), firePropertChange() throws:");
        e.printStackTrace();
    }  
    
    newArray = null;
    l_oldGroups = null;
  } // addGroupToGM()

   /**
   * Removes a group for which this GroupIDMember is a member.
   * Fires changes on 'AssociatedGroups'.
   * No exception if _grp is not registered.
   * Used only by GroupIDImpl. Overridden by UserIDImpl which calls it.
   * @param _grp group to remove
   * @exception InternalError if _grp is not found
   */
  public void removeGroupToGM(ImmutableGroupID _grp) {
    // variables
    int index = -1;
    GroupIDImpl[] newArray;
    GroupIDImpl[] l_oldGroups;

                            // for testing and tracking
                            //System.out.println("in GroupIDMemberImpl.removeGroupToGM(), for "+getName()+
                              //      ", start for the argument: "+_grp);

    if (_grp == null) return;
    // is true only if finalizeForUser() has been invoked
    if (l_Groups_ == null) return;       

    // saves the old Group list, since l_Groups_ is a bound property
    l_oldGroups = new GroupIDImpl[l_Groups_.length];
    System.arraycopy(l_Groups_, 0, l_oldGroups, 0, l_Groups_.length);

                            // for testing and tracking
                            //System.out.println("in GroupIDMemberImpl.removeGroupToGM(), "+
                              //      "l_Groups_ before the detection of the group to remove:");
                              //for (int i = 0 ; i < l_Groups_.length; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("at "+i+": "+l_Groups_[i].getName());
                              //} // for 
    
    for (int i = 0 ; i < l_Groups_.length ; i++) {
        if (l_Groups_[i] == _grp ) {
            index = i;
            break;
        } // if
    } // for

    // detects error
    if (l_Groups_.length == 0)
      throw new InternalError("in GroupIDMemberImpl.removeGroupToGM()"+
        ", the current groupID array is emmpty.");
    if (index == -1)
      throw new InternalError("in GroupIDMemberImpl.removeGroupToGM()"+
        ", the groupID to remove is not found.");

                            // for testing and tracking
                            //System.out.println("in GroupIDMemberImpl.removeGroupToGM(), for "+getName()+
                              //      ", l_Groups_ before the removing:");
                              //for (int i = 0 ; i < l_Groups_.length ; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("-- at "+i+": "+l_Groups_[i].getName());
                                  //else
                                    //  System.out.println("-- at "+i+": NULL");
                              //} // for
    
    // remove the null element if it is not the last one
    if (index < l_Groups_.length - 1) {
      for (int i = index ; i < l_Groups_.length - 1 ; i++) {
          l_Groups_[i] = l_Groups_[i + 1];
      } // for
    } // if

                            // for testing and tracking
                            //System.out.println("in GroupIDMemberImpl.removeGroupToGM(), for "+getName()+
                              //      ", l_Groups_ after the removing, before the removing of the last element:");
                              //for (int i = 0 ; i < l_Groups_.length ; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("-- at "+i+": "+l_Groups_[i].getName());
                                  //else
                                    //  System.out.println("-- at "+i+": NULL");
                              //} // for
    
    
    if (l_Groups_.length > 1) { // remove the last element
      newArray = new GroupIDImpl[l_Groups_.length - 1];
      System.arraycopy(l_Groups_, 0, newArray, 0, l_Groups_.length - 1);
      l_Groups_ = newArray;
    } else l_Groups_ = new GroupIDImpl[0];

                            // for testing and tracking
                            //System.out.println("in GroupIDMemberImpl.removeGroupToGM(), for "+getName()+
                              //      ", l_Groups_ at the end:");
                            //if (l_Groups_.length != 0) {
                              //for (int i = 0 ; i < l_Groups_.length ; i++) {
                                //  if (l_Groups_[i] != null)
                                  //    System.out.println("-- at "+i+": "+l_Groups_[i].getName());
                                  //else
                                    //  System.out.println("-- at "+i+": NULL");
                              //} // for
                            //}
                            //else {
                              //  System.out.println("-- null array");
                            //}

    try {
        firePropertyChange("AssociatedGroups", l_oldGroups, getEorL_GroupID());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From GroupIDMemberImpl.removeGroupToGM(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    newArray = null;
    l_oldGroups = null;
  } // removeGroupToGM()


} // GroupIDMemberImpl