//Title:        Access Road
//Version:      0.7.2
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
 
package ARoad0.gBase;

import ARoad0.Pattern.Singleton;

import java.util.List;
import java.util.Arrays;
import java.util.Collections;

/**
* This class is responsible to define the IS constraints and controls to use 
* for an ImmutableNameISTree by BaseManager and other classes.
* It defines the node types which are allowed for each parent type. <P>
* The IS NameLinkTree structure has the following format, where the node types are constants. 
* The nametree root is a node with a specific type (INFORMATION_SYSTEM_NODE), 
* and there are 5 possible types for its direct children: <P>
* - a physical component (PHYSICAL_COMPONENT_NODE type), <P> 
* - a logical component (LOGICAL_COMPONENT_NODE type), <P>
* - an ACS (ACS_NODE type), <P>
* - a composite ACS (COMPOSITE_ACS_NODE type). <P>
* - a composite subACS (COMPOSITE_SUB_ACS_NODE type). <P>
* For a physical component, there are 3 possible types for its direct children: <P>
* - a physical component, <P>
* - a logical component, <P>
* - an ACS. <P>
* For a logical component, there are 2 possible types for its direct children: <P>
* - a logical component, <P>
* - an ACS. <P>
* For an ACS, there are 2 possible types for its direct children: <P>
* - a logical component, <P>
* - a subACS (SUB_ACS_NODE type). <P>
* For a composite ACS, there are 1 possible type for its direct children: <P>
* - an ACS. <P>
* For a composite subACS, there are 1 possible type for its direct children: <P>
* - a subACS. <P>
* For a subACS, there are no direct children.
* The methods in this class return these values to apply these constraints on types.<P>
* The complementary static variables are: <P>
* - IS_AUTHORIZATION_SERVER for an ACS only, to set in NameISTree.Links. <P>
* - IS_TIED_CHILD for an ACS only, to set in NameISTree.Links. <P>
* - the AGO, ACL, Bridge and Linked Privielge Right types, to set in NameISTree.Links
* for the Alternate relation between two ACS or two subACS,
* to state that the alternate handles these rights for the ground. <P>
* The alternate replaces or changes the ground handling, depending on the alternate relation type.
* @see NameISTreeImpl
* @see BaseManagerImpl#addSecondaryPropertiesToAcsOrSubNode(ARoad0.gBaseInterface.ImmutableName, ARoad0.gBaseInterface.ImmutableName, boolean, ARoad0.gBaseInterface.ImmutableName, ARoad0.gBaseInterface.ImmutableName[], java.util.Map, boolean) 
*/
public class ISFactoryUtilityImpl implements Singleton {
    /**
     * This string signal the root of the tree in the 'IS structure' window,
     * by TreeManager, and also the nodes searched
     * by an ISTreeBaseListener for adding or removing an ACS object.
     */
    public static final String  STRUCTURE_ROOT_STRING = "IS Structure";
    /**
     * Type of a NameLinkTree node which represents an information system.
     */
    public static final String DEFAULT_IS_NAME = "Default_IS";
    /**
     * Type of a NameLinkTree node which represents an information system.
     */
    public static final String INFORMATION_SYSTEM_NODE = "<Information System>";

    /**
     * Type of a NameLinkTree node which represents a physical component.
     * Named 'physical' in NameISTrees.
     */
    public static final String PHYSICAL_COMPONENT_NODE = "<Hardware>";
    
    /**
     * Type of a NameLinkTree node which represents a logical component, in other words
     * a software which is not an ACS, subACS, composite ACS or composite subACS.
     * Named 'logical' in NameISTrees.
     */
    public static final String LOGICAL_COMPONENT_NODE = "<Non-Acs Software>";

    /**
     * Type of a NameLinkTree node which represents an Access Control System.
     */
    public static final String ACS_NODE = "<ACS>";

    /**
     * Type of a NameLinkTree node which represents a Sub Access Control System.
     */
    public static final String SUB_ACS_NODE = "<subACS>";

    /**
     * Type of a NameLinkTree node which represents a Composite Access Control System.
     */
    public static final String COMPOSITE_ACS_NODE = "<Composite ACS>";

    /**
     * Type of a NameLinkTree node which represents a Composite Sub Access Control System.
     */
    public static final String COMPOSITE_SUB_ACS_NODE = "<Composite subACS>";
    /**
     * Optional property for an ACS only, to set in NameISTree.Links.
     */
    public static final String IS_AUTHORIZATION_SERVER = "Is an Authorization Store";
    /**
     * Optional property for an ACS only, to set in NameISTree.Links.
     */
    public static final String IS_TIED_CHILD = "Is a Tied Child to its parent";    
    /**
     * Optional property for an ACS or a subACS, to use in the GUI. 
     */
    public static final String IS_SUBCOMPOSITE_OF = "<Subcomposite for: >"; 
    /**
     * Keyword for the optional property of a ground ACS or a subACS, 
     * to set in NameISTree.Alternates,
     * to say that this node is a ground for an alternate in a name having the format
     * 'IS_ALTERNATE_OF+alternate_name'.
     */
    public static final String IS_ALTERNATE_OF = "<Alternate for: >"; 
    /**
     * Alternate relation type, to set in NameISTree.Links,
     * to say that the alternate replaces the groung handling for the relevant right types. 
     * For instance, LDAP is a SWITCHING_ALTERNATE ACS of the Linux Network SubACS. 
     */
    public static final String SWITCHING_ALTERNATE = "Switching Alternate";     
    /**
     * Alternate relation type, to set in NameISTree.Links,
     * to say that the alternate sets positive rights which limits the ground positive rights
     * for the relevant right types. 
     * For instance, SELinux is a LIMITING_ALTERNATE SubACS of the Linux Files SubACS.
     */
    public static final String LIMITING_ALTERNATE = "Limiting Alternate"; 
    /**
     * Alternate relation type, to set in NameISTree.Links,
     * to say that the alternate sets negative rights which are subtracted from
     * the ground positive rights, for the relevant right types. 
     */
    public static final String SUBTRACTING_ALTERNATE = "Subtracting Alternate"; 
    /**
     * Account/Group/Other Rights type, to set in NameISTree.Links for the Alternate relation,
     * to say that the alternate handles these rights for the ground. The alternate replaces
     * or changes the ground handling, depending on the alternate relation type. 
     */
    public static final String AGO_RIGHTS = "For AGO rights";
    /**
     * Access Control List Rights type, to set in NameISTree.Links for the Alternate relation,
     * to say that the alternate handles these rights for the ground. The alternate replaces
     * or changes the ground handling, depending on the alternate relation type. 
     */
    public static final String ACL_RIGHTS = "For ACL rights";
    /**
     * Bridge Rights type, to set in NameISTree.Links for the Alternate relation,
     * to say that the alternate handles these rights for the ground. The alternate replaces
     * or changes the ground handling, depending on the alternate relation type. 
     */
    public static final String BDG_RIGHTS = "For Bridge rights";
    /**
     * Linked Privilege Rights type, to set in NameISTree.Links for the Alternate relation,
     * to say that the alternate handles these rights for the ground. The alternate replaces
     * or changes the ground handling, depending on the alternate relation type. 
     */
    public static final String LPRI_RIGHTS = "For linked Privilege rights";
    /**
     * Alternate relation when the alternate may be closed when the ground ACS or subACS is opened,
     * To set in NameISTree.Links. 
     */
    public static final String OPTIONAL_ALTERNATE = "The alternate may be closed"; 
    /**
     * Alternate relation when the ground may be closed when the alternate ACS or subACS is opened,
     * To set in NameISTree.Links. 
     */
    public static final String OPTIONAL_GROUND = "The ground may be closed";
  /**
  * Name of the IPv4 network addresses space, as a standard information system.
  */
    public final static String IPV4_ADDRESSES_SPACE = "IPv4";
  /**
  * Name of the IPv6 network addresses space, as a standard information system.
  */
    public final static String IPV6_ADDRESSES_SPACE = "IPv6";
    
      /* the only instance of the class, following singleton pattern */
    protected static final ISFactoryUtilityImpl  INSTANCE = new ISFactoryUtilityImpl();   
    
    /**
     * Does not return the type for the IS structure root.
     * @return immutable list of the allowed types for an IS NameLinkTree node.
     */
    public static final List<String> getNodeTypesForIS() {
        return Arrays.asList(INFORMATION_SYSTEM_NODE,
                PHYSICAL_COMPONENT_NODE,
                LOGICAL_COMPONENT_NODE,
                ACS_NODE,
                SUB_ACS_NODE,
                COMPOSITE_ACS_NODE,
                COMPOSITE_SUB_ACS_NODE);
    }
    
    /**
     * Get the allowed types for the children of an IS NameTree node, 
     * following this class policy which constraints the NameTree structure.
     * @param _parentType is an ISImpl allowed node type of a node in the IS NameTree. 
     *          Returns null otherwise.
     * @return immutable list of the allowed types for the direct children of a _parentType node. 
     * May be empty (it is then Collections.EMPTY_LIST) for instance
     * if the argument is SUB_ACS_NODE. Never null.
     */
    public static final List<String> getChildTypesForISnode(String _parentType) {
        if (_parentType == null) return Collections.EMPTY_LIST;
        else if (INFORMATION_SYSTEM_NODE.equals(_parentType))
            return Arrays.asList(
                PHYSICAL_COMPONENT_NODE,
                LOGICAL_COMPONENT_NODE,
                ACS_NODE,
                COMPOSITE_ACS_NODE,
                COMPOSITE_SUB_ACS_NODE);
        else if (PHYSICAL_COMPONENT_NODE.equals(_parentType))
            return Arrays.asList(
                PHYSICAL_COMPONENT_NODE,
                LOGICAL_COMPONENT_NODE,
                ACS_NODE);
        else if (LOGICAL_COMPONENT_NODE.equals(_parentType))
            return Arrays.asList(
                LOGICAL_COMPONENT_NODE,
                ACS_NODE);
        else if (ACS_NODE.equals(_parentType))
            return Arrays.asList(
                LOGICAL_COMPONENT_NODE,
                ACS_NODE,
                SUB_ACS_NODE);
        else if (COMPOSITE_ACS_NODE.equals(_parentType))
            return Arrays.asList(
                ACS_NODE);
        else if (COMPOSITE_SUB_ACS_NODE.equals(_parentType))
            return Arrays.asList(
                SUB_ACS_NODE);  
        else return Collections.EMPTY_LIST;
    } // getChildTypesForISnode()
    
    /**
     * Gets the 3 types for an alternate relation between an ACS or a subACS and
     * another ACS or subACS.
     * SWITCHING_ALTERNATE says that the alternate replaces the groung handling
     * for the relevant right types.
     * For instance, LDAP is a SWITCHING_ALTERNATE ACS of the Linux Network SubACS.
     * LIMITING_ALTERNATE says that the alternate sets positive rights which limits
     * the ground positive rights for the relevant right types.
     * For instance, SELinux is a LIMITING_ALTERNATE SubACS of the Linux Files SubACS.
     * SUBTRACTING_ALTERNATE says that the alternate sets negative rights
     * which are subtracted from the ground positive rights, for the relevant right types.
     * @return immutable list of the allowed types for an Alternate relation.
     */
    public static final List<String> getAlternateTypes() {
        return Arrays.asList(SWITCHING_ALTERNATE,
                LIMITING_ALTERNATE,
                SUBTRACTING_ALTERNATE);
    } // getAlternateTypes()
    
    /**
     * Get the 4 right types, that are AGO, ACL, BDG and LPRI.
     * @return immutable list of the rights types.
     */
    public static final List<String> getRightTypesForAlternate() {
        return Arrays.asList(AGO_RIGHTS, ACL_RIGHTS, BDG_RIGHTS, LPRI_RIGHTS);
    } // getRightTypesForAlternate()
    
    /** 
     * Get the 2 options.
     * @return immutable list of optional_ground, optional_alternate.
     */
    public static final List<String> getOptionsForAlternate() {
        return Arrays.asList(OPTIONAL_ALTERNATE,
                OPTIONAL_GROUND);
    } // getOptionsForAlternate()
    
    /**
     * Get the 3 types of ACS and subACS, without COMPOSITE_SUB_ACS_NODE.
     * @return immutable list of the ACS types.
     */
    public static final List<String> getMainAcsTypes() {
        return Arrays.asList(ACS_NODE,
                SUB_ACS_NODE,
                COMPOSITE_ACS_NODE);
    } // getMainAcsTypes()
    
    
} // IS