//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Arrays;
import java.util.Collections;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

 /**
  * This class is responsible for describing the IS nametree nodes
  * and composites for displaying them in the GUI. It is an adaptor to read
  * the properties of a node from the relevant NameISTreeImpl in the BaseManager.
  * This same class is used while the properties are quite different when the node type changes.
  * This works because the GUI reads the properties through the data (e.g. the method names,...)
  * a BeanInfo provides. Selecting the right BeanInfo is the responsability of
  * the method getBeanInfoNameForISnode() in BaseUtilityImpl.
  * Called by ActionOpenGlobalStructure.valueChanged() and openFrame().
  * @see ARoad0.gDMak.ActionOpenGlobalStructure#valueChanged(javax.swing.event.TreeSelectionEvent) 
  * @version 0.7.0
  */
public class ISNodeImpl implements PropertyChangeListener, BoundBean, BaseObject {
  /**
   * Remains null only for the IS structure root.
   */
  private ImmutableNameISTree   iSTree_ = null;
  private Name                  nodeName_ = null;

   /**
   * manages all the property change listeners
   */
  private transient PropertyChangeSupport changeSupport_ = null;

  protected static final int INITIAL_CAPACITY = 10;

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ISNodeImpl EMPTY_INSTANCE = new ISNodeImpl();

  /**
  * The default constructor.
  */
  public ISNodeImpl() {
  }

  /**
   * Registers this instance has a listener of BaseManagerImpl
   * for the events 'OneInformationSystem', 'InformationSystemChange',
   * if the node is not the IS structure root.
   * This constructor has to be fast since it is used at each new selection
   * in the IS structure tree, in the GUI.
   * Called by ActionOpenGlobalStructure.openFrame() and valueChanged().
   * @param _iSTree a known NameISTree. Null only for the IS structure root.
   * @param _nodeName is the name of this nameISTree node. Not null.
   * @exception CreateError if _iSname is unknown, if the node does not belong to 
   * the nametree and is not the IS structure root
   */
  public ISNodeImpl(ImmutableNameISTree _iSTree, ImmutableName _nodeName)
          throws CreateError {

    if (_nodeName == null || _nodeName.isEmpty())
        throw new CreateError("ISNodeImpl constructor: the node name is null or empty");
    if (_iSTree == null &&
            !_nodeName.equals(new NameImpl(ISFactoryUtilityImpl.STRUCTURE_ROOT_STRING)))
        throw new CreateError("ISNodeImpl constructor: the NameISTree is null while it is not"+
                " for the IS structure root.");

    if (_iSTree != null) {
        iSTree_ = _iSTree;

        if (!iSTree_.isNode(_nodeName) && !iSTree_.isArea(_nodeName))
            throw new CreateError("ISNodeImpl constructor:" +
              " the node does not belong to the nametree.");

        nodeName_ = (Name)_nodeName.clone();

        BaseManagerImpl.getInstance().addPropertyChangeListener(
                "OneInformationSystem", this);
        BaseManagerImpl.getInstance().addPropertyChangeListener(
                "InformationSystemChange", this);
    }
    else {
        // the IS structure root has immutable properties, and
        // iSTree_ has to remian null for the working of getL_AcsAddonDescriptors()
        nodeName_ = (Name)_nodeName.clone();
    }
  } // constructor

  /**
   * Receives the events fired by BaseManagerImpl,
   * to send a property change to all the PropertyEditors in the GUI.
   * For 'OneInformationSystem', 'InformationSystemChange' events.
   * As the event is about a node which is not this instance node but is indeed
   * a node which is referenced by this node, it is more simple to fire a change event
   * to all the property editors of this instance properties, at every change in the IS.
   * @param _evt from BaseManagerImpl
   */
  public final void propertyChange(PropertyChangeEvent _evt) {
      if (_evt.getPropertyName().equals("OneInformationSystem")) {
          if (_evt.getOldValue() != null && _evt.getOldValue().equals(iSTree_.getRoot())) {
              firePropertyChange("FinalizedNode", null, null);
              return;
          }
          else return;
      } // if
      else if (_evt.getPropertyName().equals("InformationSystemChange")) {
          // nameTree is never null
          ImmutableNameISTree nameTree = (ImmutableNameISTree)_evt.getNewValue();

          if (!nameTree.getRoot().equals(iSTree_.getRoot())) {
              return;
          }

          if (!nameTree.isNode(nodeName_) && !nameTree.isArea(nodeName_)) {
              firePropertyChange("FinalizedNode", null, null);
              return;
          } // if

          firePropertyChange("OpenNode", null, null);
          firePropertyChange("NodeType", null, null);
          firePropertyChange("HubNode", null, null);
          firePropertyChange("VisibleSpace", null, null);
          firePropertyChange("AuthorizationServer", null, null);
          // for CollectionPropertyEditor which uses the new value
          firePropertyChange("NodesOfComposite", null, getL_NodesOfComposite());
          firePropertyChange("Composite", null, null);
          firePropertyChange("TiedChild", null, null);
          firePropertyChange("Alternates", null, null);
          // for CollectionPropertyEditor which uses the new value
          firePropertyChange("GroundsOfAlternate", null, getL_GroundsOfAlternate());
      } // if
  } // propertyChange()

  /**
  * Add a listener to the bean.
  * @param _l the listener
  */
  public final void addPropertyChangeListener(PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      synchronized (changeSupport_)
      {changeSupport_.addPropertyChangeListener(_l);}
  }

  /**
  * Adds a listener to the bean.
  * For the bound properties 'OpenNode', 'NodeType', 'HubNode', 'VisibleSpace',
  * 'AuthorizationServer', 'NodesOfComposite', 'Composite', 'TiedChild',
  * 'Alternates', 'GroundsOfAlternate', and 'FinalizedNode' when the node or its IS
  * is deleted.
  * @param _propertyName the property to listen
  * @param _l the listener
  */
  public final void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}
  }

  /**
  * Removes a listener to the bean.
  * @param _l the listener
  */
  public final void removePropertyChangeListener(PropertyChangeListener _l) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_l);}
  }

  /**
  * Removes a listener to the bean.
  * For the bound properties 'OpenNode', 'NodeType', 'HubNode', 'VisibleSpace',
  * 'AuthorizationServer', 'NodesOfComposite', 'Composite', 'TiedChild',
  * 'Alternates', 'GroundsOfAlternate', and 'FinalizedNode' when the node or its IS
  * is deleted.
  * @param _propertyName the property to listen
  * @param _l the listener
  */
  public final void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

      synchronized (changeSupport_)
      {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  }

  /**
   * @return name of the base object. May be empty, but not null.
   */
  public final ImmutableName getName() {
      return (ImmutableName)nodeName_.clone();
  }

  /**
   * @return like getName()
   */
  public final ImmutableName getFullName() {
      return (ImmutableName)nodeName_.clone();
  }

  /**
   * This short name is NOT unique for the Access Road program nor the ACS if
   * it is an ACS object.
   * It is often the last component name.
   * @return the nick name. May be null.
   */
  public final String getNickName() {
        return nodeName_.lastComponent();
  }

  /**
  * Gets the state open/closed of a node or area.
  * The IS structure root is always open.
  * @return true if the node is open; always false if it is null or unknown.
  */
  public final boolean isOpen() {
      // Remains null only for the IS structure root
      if (iSTree_ == null)
          return true;
      else
          return iSTree_.isOpen(nodeName_);
  }

  /**
   * Gets the type of the node or area, or ISFactoryUtilityImpl.STRUCTURE_ROOT_STRING
   * for the IS structure root.
   * @return the node type. May be null.
   */
   public final String getType() {
      if (iSTree_ == null)
          return ISFactoryUtilityImpl.STRUCTURE_ROOT_STRING;

      try {
          return iSTree_.getType(nodeName_);
      }
      catch (BaseError b) {
          new InternalError("ISNodeImpl.getType(): "+b.toString());
      }
      return null; // unuseful
   } // getType()
 
  /**
   * Tests if a node of the nametree is set as a 'hub' for its children.
   * A node may be a hub, which means that for this node,
   * any direct child may have exchanges with any other direct child. 
   * For a non-hub node, exchanges between its children are not allowed.
   * The visible space (see NameISTree) of a node having a hub parent always 
   * includes all the children of its parent.
   * An area is always a hub node.
   * @return true if the node is a hub.
   */
   public final boolean isHubNode() {
      try {
          return iSTree_.isHubNode(nodeName_);
      }
      catch (BaseError b) {
          new InternalError("ISNodeImpl.isHubNode(): "+b.toString());
      }
      return false; // logically unuseful
   } // isHubNode()

  /**
  * Gets the nodes which are visible for the node.
  * This extends the native ImmutableNameTree visible space, which is the parent children
  * (if ImmutableNameTree.isHubNode is true) and the ImmutableNameTree areas of the node, if any.
  * The value is one of the constants returned by getExtendedVisibleSpaceConstants().
  * An area has always ALL_THE_IS_SPACE as visible space.
  * @return the extended visible space from the node. May be null.
  */
  public final String getExtendedVisibleSpace() {
      return iSTree_.getExtendedVisibleSpace(nodeName_);
  }

  /**
   * True if the node is an ACS or a composite ACS which is an authorization server.
   */
  public final boolean isAuthorizationServer() {
      return Arrays.asList(iSTree_.getLinks(nodeName_)).contains(
              ISFactoryUtilityImpl.IS_AUTHORIZATION_SERVER);
  }
  
  /**
  * Gets the nodes belonging to a composite in the nametree.
  * @return list of the subcomposite names. May be empty. Never null.
  */
  public final List<ImmutableName> getL_NodesOfComposite() {
      return iSTree_.getL_NodesOfArea(nodeName_);
  } // getAreasOfComposite()

  /** 
   * True if the node is a tied node.
   */
  public final boolean isTiedChild() {
      return Arrays.asList(iSTree_.getLinks(nodeName_)).contains(
              ISFactoryUtilityImpl.IS_TIED_CHILD);
  }

  /**
  * Gets the node property for the Alternate relation. The key is the alternate name.
  * The value for a key is a String array
  * with the ISImpl constants: ALTERNATE_TYPE at the index 0 (mandatory),
  * then, without any specific order in the array, one or several RightsTypes (mandatory),
  * then OPTIONAL_ALTERNATE (option) and OPTIONAL_GROUND (option).
  * @return the array of alternates for the node. May be empty but not null.
  */
  public final Map<String, String[]> getM_Alternates() {
      if (iSTree_.getAlternates(nodeName_) == null)
          return Collections.emptyMap();

      String[]              l_alter = null;
      Map<String, String[]> m_result = new HashMap(INITIAL_CAPACITY);

      for (ImmutableName name : iSTree_.getAlternates(nodeName_)) {
          l_alter = (String[]) iSTree_.getLinks(
                  ((Name)((NameImpl)nodeName_).clone()).extended(name));
          m_result.put(name.toString(), l_alter);
      } // for name

      l_alter = null;
      return m_result;
  } // getM_Alternates()

  /**
  * Gets the composite if the node is a subcomposite.
  * @return the composite name or null
  */
  public final ImmutableName getComposite() {
      if (iSTree_.getMandatoryPaths(nodeName_) == null ) return null;
      else return (ImmutableName)iSTree_.getMandatoryPaths(nodeName_).get(0)[0];
  } // getComposite()

  /**
  * Gets the Grounds for which the node is an Alternate.
  * @return the list of grounds for the node. May be empty but not null.
  */   
  public final List<ImmutableName> getL_GroundsOfAlternate() {
      List<ImmutableName> l_result = iSTree_.getGroundsOfAlternate(nodeName_);
      
      if (l_result != null) return l_result;
      else return new ArrayList(INITIAL_CAPACITY);
  } // getL_GroundsOfAlternate()

  /**
  * Gets all the descriptors of the ACS addons that are loaded in the current
  * working directory of Access Road. This method is reserved to the IS structure root.
  * It may be called on an empty instance.
  * @return the list of the ACS addons. May be empty but not null.
  */
  public final List<String> getL_AcsAddonDescriptors() {
      if (iSTree_ != null)
          throw new InternalError("ISNodeImpl.getL_AcsAddonDescriptors() is reserved"+
                  " to the IS structure root.");
      return BaseManagerImpl.getInstance().getL_AcsAddonDescriptors();
  }

  /**
  * @return a cloned and no-backed instance, with clone of all names.
  */
  public final Object clone() {
      ISNodeImpl clon = null;

      try {
          clon = (ISNodeImpl) super.clone();
      }
      catch (CloneNotSupportedException c) {
          throw new InternalError("ISNodeImpl.clone(): "+c.toString());
      }

      clon.iSTree_ = iSTree_;
      clon.nodeName_ = (NameImpl)((NameImpl)nodeName_).clone();
      return clon;
  }   // clone()

  /**
   * Compares the values in the internal maps of this class.
  * @param _obj is the object to compare
  * @return true if the lists of values are equal
  */
  public final boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      if (!super.equals(_obj))
          return false;

      ISNodeImpl obj = (ISNodeImpl) _obj;

      if (obj.iSTree_ != iSTree_)
            return false;  

      if (obj.nodeName_ == null && nodeName_ != null)
            return false;
      if (obj.nodeName_ != null && !obj.nodeName_.equals(nodeName_))
            return false;  

      return true;
  } // equals()

  /**
  * @return the hashcode from the node name.
  */
  public final int hashCode() {
       return nodeName_.hashCode();
  }

  //
  // ------------- protected methods -------------------
  //

  /**
   * This method nullifies all the inner properties and unregisters it as property listener.
   */
  protected final synchronized void finalizeForUser() {
    iSTree_ = null;
    nodeName_ = null;

    BaseManagerImpl.getInstance().removePropertyChangeListener(
            "OneInformationSystem", this);
    BaseManagerImpl.getInstance().removePropertyChangeListener(
            "InformationSystemChange", this);
  } // finalizeForUser()

  /**
   * This method calls finalizeForUser().
   * @exception java.lang.Throwable
   */
  protected final void finalize() throws Throwable {
    finalizeForUser();
    super.finalize();
  } // finalize()

  /**
  * Fire an event to every registered listener, in any order.
  * For the bound properties 'OpenNode', 'NodeType', 'HubNode', 'VisibleSpace',
  * 'AuthorizationServer', 'NodesOfComposite', 'Composite', 'TiedChild',
  * 'Alternates', 'GroundsOfAlternate', and 'FinalizedNode' when the node or its IS
  * is deleted.
  * Called by propertyChange() for all the listeners of all the inner properties.
  * @param _propertyName of the updated property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected final void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //System.out.println("ISNodeImpl.firePropertyChange(), call to "+
                  //"propertyChange() of "+nodeName_+" for the property "+_propertyName);

      synchronized (changeSupport_)
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()

} // ISNodeImpl class
