//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Arrays;
import java.util.HashSet;

import java.io.ObjectStreamField;

 /**
  * This class is responsible to manage for IS nametrees,
  * a tree of ImmutableNames with specialized links to the other nodes.
  * Most of the properties are directly managed by the calling class.
  * The ISFactoryUtilityImpl class provides the constraints on the nodes inheritance
  * in the nametree.<P>
  * A node in an ImmutableNameISTree may have to apply some inner properties or
  * some constraints in its links to other nodes of the whole nametree:<P>
  * - 'Open' is a mandatory node property which has the value 'true' if the node
  * is active, like an open ACS, subACS or Composite ACS/subACS, 
  * which is currently opened in the program, or if the node is a passive one 
  * as a the direct or indirect parent of an opened active node. <P>        
  * - 'ExtendedVisibleSpace' is a mandatory node property which extends the parent children
  * (if ImmutableNameTree.isHubNode is true) and the ImmutableNameTree areas of the node, if any.
  * This property is an ImmutableName/String Map:<P>
  *      mandatory key: node_name; <P>
  *      the possible String values are ImmutableNameISTree constants: <P>
  *          ALL_THE_IS_SPACE (the default value), <P>
  *          ONE_STEP_NODES_SPACE, TWO_STEPS_NODES_SPACE, THREE_STEPS_NODES_SPACE, <P>
  *          DIRECT_CHILDREN_SPACE, INDIRECT_CHILDREN_SPACE,  <P>
  *          PARENT_INDIRECT_CHILDREN_SPACE, PARENT_SPACE, <P>
  *          NO_VISIBLE_SPACE. <P>
  * - 'Alternates' is an optional node property, used only for the Alternate relation,
  * with an ImmutableName/ImmutableName[] Map:<P>
  *      mandatory key: node_name for a ground ACS or subACS; <P>
  *      the values are the names of the alternates for the key (limited to 4). <P>
  * - 'MandatoryPaths' is an optional node property, used only for the Composites
  * which are implemented as ImmutableNameTree.Areas, with an ImmutableName/ImmutableName[] Map:<P>
  *      optional key: node_name is the beginning of the path; <P>
  *      value: the optional array of node names is the structurally-mandatory path 
  *      to go accross, where the structure is a one-item List, or a List<P>
  *      (first_node, second_node, beginning_node, third_node, fourth_node, beginning_node), <P>
  *      which indicates that there are two paths: from beginning_node to second_node, <P>
  *      and from beginning_node to third_node and fourth_node, <P>
  *      and (feature to be implemented) where all the nodes are in the visible space of the beginning node. <P>
  *      The use of this property in BaseManagerimpl is to set the unique composite of a subcomposite. <P>
  * - 'Links' is an optional node property, for IS nametrees,
  *      with an ImmutableName/String[] map. <P>
  *      Each property may have any format, and is defined by the user of this class.
  *      The key is usually a node name. <P>
  * <P> 
  * The Links map is used for IS nametrees, and the ACS factory initializes
  * the ImmutableNameISTree links following some specific and complex uses.<P>
  * All the properties of a ImmutableNameISTree are independant to the others.
  * A node which is opened or closed may have any visible space, any mandatory paths
  * and any links. Before the use of the data about a mandatory path or
  * a link, it is recommanded to see if the relevant node is opened or not.
  * See also the properties 'getNodeType' and 'isHubNode' in the super class.
  * @see BaseManagerImpl#addSecondaryPropertiesToAcsOrSubNode(ARoad0.gBaseInterface.ImmutableName, ARoad0.gBaseInterface.ImmutableName, boolean, ARoad0.gBaseInterface.ImmutableName, ARoad0.gBaseInterface.ImmutableName[], java.util.Map, boolean)
  * @version 0.7.0
  */
public class NameISTreeImpl extends NameTreeImpl implements ImmutableNameISTree {
  
  /*
  * private variables
  */
  private static final long serialVersionUID = 0;

  /**
   * ImmutableName/Boolean map of the state open/closed of each node.<P>
   */
  private Map<ImmutableName, Boolean>  m_OpenNodes_ = null;
  
  /**
   * ImmutableName/String map of visible spaces for a node, which extends the parent children
   * (if ImmutableNameTree.isHubNode is true) and the ImmutableNameTree areas of the node, if any.<P>
   * It is a mandatory node property:<P>
   *      mandatory key: node_name; <P>
   *      the possible String values are 'xxx_SPACE' ImmutableNameISTree constants. <P>
   */
  private Map<ImmutableName, String>  m_ExtendedVisibleSpaces_ = null;

  /**
  * Optional node property for the Alternate relation: <P>
  *      mandatory key: node_name for a ground ACS or subACS; <P>
  *      the values are the names of the alternates for the key. <P> 
   */
  private Map<ImmutableName, ImmutableName[]>   m_Alternates_ = null;
  
  /** 
   * ImmutableName/ImmutableName[] map of the optional mandatory paths <P>
   * It is an optional node property:<P>
   *      optional key: node_name is the beginning of the first path; <P>
   *      value: the optional array of node names is the mandatory paths to go accross <P>
   *      (if rights allow it), <P>
   *      where the structure is a one-item List, or a List<P>
  *      (first_node, second_node, beginning_node, third_node, fourth_node, beginning_node), <P>
   *      which indicates that there are two paths: from beginning_node to second_node, <P>
   *      and from beginning_node to third_node and fourth_node, <P>
   *      and where all the nodes are in the visible space of the beginning node <P>
   */
  private Map<ImmutableName, ImmutableName[]>  m_MandatoryPaths_ = null;

  /**
   * ImmutableName/String[] map of optional links which are managed by the user of this class<P>
   * It is an optional node property with an ImmutableName/String[] map: <P>
   *      the optional key is the node or area name, or the sum of two node names, <P>
   *      the value is an optional String array, where all the properties are also optional
   *        and without constraints on the format.<P>
   */
  private Map<ImmutableName, String[]>  m_Links_ = null;

  /*
  * public variables
  */  
  /**
  * to see all the space
  */
  public final static String ALL_THE_IS_SPACE = "all the space";
  /**
  * to see the parent and the direct children of the node.
  */
  public final static String ONE_STEP_NODES_SPACE = "one-step nodes";
  /**
  * to see the parent and the parent of this parent, the
  * direct children and the direct children of these direct children.
  */
  public final static String TWO_STEPS_NODES_SPACE = "one and two-steps nodes";
  /**
  * to see the parent, the first two-levels parents of this parent, the
  * direct children and the first two-levels children of these direct children.
  */
  public final static String THREE_STEPS_NODES_SPACE = "one, two and three-steps nodes";
  /**
  * to see the direct children.
  */
  public final static String DIRECT_CHILDREN_SPACE = "direct children";
  /**
  * to see the direct and all the indirect children, whatever their levels.
  */
  public final static String INDIRECT_CHILDREN_SPACE = "direct and all indirect children";
  /**
  * to see the parent, the direct and all the indirect children, whatever theirs levels.
  */
  public final static String PARENT_INDIRECT_CHILDREN_SPACE =
          "parent, direct and all indirect children";
  /**
  * to see the parent.
  */
  public final static String PARENT_SPACE = "parent";  
  /**
  * to see nothing; the closed visible space.
  */
  public final static String NO_VISIBLE_SPACE = "no visible space";
  
  /**
  * The list of the allowed values for the visible space.
  */
  private final static ArrayList<String> L_VISIBLE_SPACES = new ArrayList(INITIAL_CAPACITY);
  
  // --------------------- static initialization -----------------------------
  static {
      L_VISIBLE_SPACES.add( ALL_THE_IS_SPACE );
      L_VISIBLE_SPACES.add( ONE_STEP_NODES_SPACE );
      L_VISIBLE_SPACES.add( TWO_STEPS_NODES_SPACE );
      L_VISIBLE_SPACES.add( THREE_STEPS_NODES_SPACE );
      L_VISIBLE_SPACES.add( DIRECT_CHILDREN_SPACE );
      L_VISIBLE_SPACES.add( INDIRECT_CHILDREN_SPACE );
      L_VISIBLE_SPACES.add( PARENT_INDIRECT_CHILDREN_SPACE );
      L_VISIBLE_SPACES.add( PARENT_SPACE );
      L_VISIBLE_SPACES.add( NO_VISIBLE_SPACE );
  }  // static

  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final NameISTreeImpl EMPTY_INSTANCE = new NameISTreeImpl();

  /** standardized variables for the serialization
  * @serial ("m_OpenNodes_", HashMap.class), 
  * ("m_ExtendedVisibleSpaces_", HashMap.class),
  * ("m_MandatoryPaths_", HashMap.class), 
  * ("m_Links_", HashMap.class),
  * ("m_Alternates_", HashMap.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("m_OpenNodes_", HashMap.class),
      new ObjectStreamField("m_ExtendedVisibleSpaces_", HashMap.class),
      new ObjectStreamField("m_MandatoryPaths_", HashMap.class),
      new ObjectStreamField("m_Links_", HashMap.class),
      new ObjectStreamField("m_Alternates_", HashMap.class)
  };

  /**
  * The default constructor.
  */
  public NameISTreeImpl() {
  }

  /**
   * Constructor.
   * The returned instance is empty if _root_name is null or empty.
   * The root visible space is 'ALL_THE_IS_SPACE'.
   * The root type is ISImpl.INFORMATION_SYSTEM_NODE.
   * @param _root_name is the name of the nametree root. Not null.
   * @param _is_hub is true if the root is a node hub
   * @exception CreateError if _root_name is null or empty
   */
  public NameISTreeImpl(ImmutableName _root_name, boolean _is_hub) throws CreateError {
        
        super(_root_name, ISFactoryUtilityImpl.INFORMATION_SYSTEM_NODE, _is_hub);

        m_OpenNodes_ = new HashMap(INITIAL_CAPACITY);
        m_ExtendedVisibleSpaces_ = new HashMap(INITIAL_CAPACITY);
        m_ExtendedVisibleSpaces_.put(_root_name, ALL_THE_IS_SPACE);

        m_MandatoryPaths_ = new HashMap(INITIAL_CAPACITY*5);
        m_Links_ = new HashMap(INITIAL_CAPACITY*5);
        m_Alternates_ = new HashMap(INITIAL_CAPACITY);
  } // constructor    

  /**
  * The method called from ACSFactoryImplBeanInfo for calling getExtendedVisibleSpaceConstants()
  * to set the ACS visible space.
  * @return an empty instance of this class
  */
  public static NameISTreeImpl getInstance() {
      return new NameISTreeImpl();
  }
  
  /**
  * Gets the state open/closed of a node or area.
  * @param _node is a node in the ImmutableNameISTree
  * @return true if the node is opened; always false if it is null or unknown.
  */   
  public boolean isOpen(ImmutableName _node) {
      if (_node == null || m_OpenNodes_.get(_node) == null) return false;
      else return (Boolean)m_OpenNodes_.get(_node);
  }
  
  /**
  * Sets the state open/closed of a node or area. Sets the same state for all its parents
  * if _open is true. Otherwise, sets the same state for the parents which have
  * only this node as still open direct or indirect child. Set the same state
  * for all the children if _open is false, and noop on children otherwise.
  * Calls getTreePath(), getL_DirectChildren() and getL_AllChildren().
  * @see #setOnlyThisOpen
  * @param _node is a non-null node or area in the ImmutableNameISTree
  * @param _open true if the node is opened.
  * @exception BaseError if _node is null, empty or unknown, or thrown by a called method.
  */   
  protected synchronized void setOpen(ImmutableName _node, boolean _open)
          throws BaseError {
      if (_node == null || _node.isEmpty() || (!isNode(_node) && !isArea(_node)) )
          throw new BaseError("in NameISTreeImpl.setOpenNode()," +
            " the node or area is null, empty, or unknown.");

      if (isArea(_node)) {
          if (_open) {
              m_OpenNodes_.put((ImmutableName)_node.clone(), new Boolean(true));
          }
          else {
              m_OpenNodes_.put((ImmutableName)_node.clone(), new Boolean(false));
          }
          return;
      } // if isArea()

      Object[]   l_path = getTreePath(getRoot(), _node).getPath();
      boolean    toProcess = false;
      boolean    closedChildren = false;

      if (l_path.length < 3) {
          l_path = null;
          return;
      }

      if (_open) {
          for (int i = 0; i < l_path.length; i++) {
              m_OpenNodes_.put((ImmutableName)l_path[i], new Boolean(true));
          } // for i
      }
      else {
          m_OpenNodes_.put((ImmutableName)_node.clone(), new Boolean(false));

          // closes the parents if necessary, from _node to the root
          for (int i = l_path.length - 2; i > 0; i--) {
              if (getL_DirectChildren((ImmutableName)l_path[i]).size() == 1) {
                  if (closedChildren) {
                      m_OpenNodes_.put((ImmutableName)l_path[i], new Boolean(false));
                  }
              } else {
                  toProcess = true;

                  for (ImmutableName name : getL_DirectChildren((ImmutableName)l_path[i])) {
                      if (name.equals(l_path[i + 1])) continue;

                      if (isOpen(name)) toProcess = false;

                      // closes l_path[i] if l_path[i + 1] is the unique child to be opened
                      for (ImmutableName name2 : getL_AllChildren(name)) {
                          if (isOpen(name2)) {
                              toProcess = false;
                              break;
                          }
                      } // for name2
                      if (!toProcess) break;
                  } // for name

                                    // for testing and tracking
                                //System.out.println("NameISTreeImpl.setOpen(), "+
                                  //      "for "+l_path[i]+", closedChildren: "+closedChildren+
                                    //    ", toProcess: "+toProcess);

                  // no closing if the children in the path have not been closed
                  if (toProcess && closedChildren)
                      m_OpenNodes_.put((ImmutableName)l_path[i], new Boolean(false));
                  else if (toProcess && i == l_path.length - 2) {
                      m_OpenNodes_.put((ImmutableName)l_path[i], new Boolean(false));
                      closedChildren = true;
                  }
                  else closedChildren = false;

              } // else
          } // for i

          // closes all the children
          for (ImmutableName name : getL_AllChildren(_node)) {
              m_OpenNodes_.put(name, new Boolean(false));
          } // for name
      } // else
      l_path = null;
  } // setOpen()
  
  /**
  * Sets the state open/closed of a node or area, and does not process its parents or children.
  * @see #setOpen
  * @param _node is a non-null node or area in the ImmutableNameISTree
  * @param _open true if the node is opened.
  * @exception BaseError if _node is null, empty or unknown.
  */   
  protected synchronized void setOnlyThisOpen(ImmutableName _node, boolean _open)
          throws BaseError {
      if (_node == null || _node.isEmpty() || (!isNode(_node) && !isArea(_node)) )
          throw new BaseError("in NameISTreeImpl.setOpenNode()," +
            " the node is null, empty, or unknown.");
      
      m_OpenNodes_.put((ImmutableName)_node.clone(), new Boolean(_open));
  } // setOnlyThisOpen()

  /**
  * Gets if a node is the direct or indirect parent of at least one node
  * having the type ISImpl.ACS_NODE. The argument may be an area.
  * @param _node is a node or an area in the ImmutableNameISTree
  * @return true if the argument is a parent
  */   
  public boolean isParentOfAcsNode(ImmutableName _node) {
      if (_node == null) return false;
      else {
          List<ImmutableName> l_children = getL_AllChildren(_node);
          boolean result = false;
          try {
              if (l_children == null) return false;
              for (ImmutableName node : l_children) {
                  if (getType(node).equals(ISFactoryUtilityImpl.ACS_NODE)) {
                      result = true;
                      break;
                  }
              } // for node
          } catch (BaseError b) {
              b.printStackTrace();
              throw new InternalError("from NameISTreeImpl.isParentOfAcsNode().");
          }
          return result;
      } // else
  } // isParentOfAcsNode()

  /**
  * Gets if a node is the direct or indirect child of at least one node
  * having the type ISImpl.ACS_NODE. The argument may be an area.
  * @param _node is a node or an area in the ImmutableNameISTree
  * @return true if the argument is a child
  */   
  public boolean isChildOfAcsNode(ImmutableName _node) {
      if (_node == null) return false;
      else {
          Object[] l_ancestors = null;
          boolean result = false;
          try {
              l_ancestors = getTreePath(getRoot(), _node).getPath();
              if (l_ancestors == null) return false;
              for (int i = 0; i < l_ancestors.length; i++) {
                  if (getType((ImmutableName)l_ancestors[i]).equals(ISFactoryUtilityImpl.ACS_NODE)) {
                      result = true;
                      break;
                  }
              } // for node
          } catch (BaseError b) {
              b.printStackTrace();
              throw new InternalError("from NameISTreeImpl.isChildOfAcsNode().");
          }
          return result;
      } // else
  } // isChildOfAcsNode()
  
  /**
  * Gets the authorized visible spaces which extends the ImmutableNameTree.isHubNode.
  * Called by SimplePropertyEditor.getRestrictedValues()
  * for ACSFactoryImplBeanInfo.
  * @return a copy of the authorized visible spaces for the NameISTrees.
  */   
  public String[] getExtendedVisibleSpaceConstants() {
      return (String[]) ((List) L_VISIBLE_SPACES.clone()).toArray(new String[0]);
  }
  
  /**
  * Gets the nodes which are visible for the argument.
  * This extends the native ImmutableNameTree visible space, which is the parent children 
  * (if ImmutableNameTree.isHubNode is true) and the ImmutableNameTree areas of the node, if any. 
  * The value is one of the constants returned by getExtendedVisibleSpaceConstants().
  * An area has always ALL_THE_IS_SPACE as visible space.
  * @param _node is a node or an area in the ImmutablelNameTree
  * @return the extended visible space from the node. May be null.
  */   
  public String getExtendedVisibleSpace(ImmutableName _node) {
      if (isArea(_node)) return ALL_THE_IS_SPACE;
      else return (String)m_ExtendedVisibleSpaces_.get(_node);
  }
  
  /**
  * Sets the nodes which are visible for the node argument.
  * This extends the native ImmutableNameTree visible space, which is the parent children 
  * (if ImmutableNameTree.isHubNode is true) and the ImmutableNameTree areas of the node, if any. 
  * The value is one of the constants returned by getExtendedVisibleSpaceConstants().
  * @param _node is a non-null node in the ImmutableNameISTree. Never an area.
  * @param _space is the visible space from the node. Never null.
  * @exception BaseError if _node is null, empty, an area or unknown, if _space is not
  *   an authorized value.
  */   
  protected void setExtendedVisibleSpace(ImmutableName _node, String _space) throws BaseError {
      if (_node == null || _node.isEmpty() || !isNode(_node) )
          throw new BaseError("in NameISTreeImpl.setExtendedVisibleSpace()," +
            " the node is null, empty, or unknown.");
      if (!L_VISIBLE_SPACES.contains(_space) )
          throw new BaseError("in NameISTreeImpl.setExtendedVisibleSpace()," +
            " the visible space is not an authorized value.");
      m_ExtendedVisibleSpaces_.put((ImmutableName)_node.clone(), _space);
  } // setExtendedVisibleSpace()

  /**
  * Adds a node property only for the Alternate relation: <P>
  *      mandatory key: node_name; <P>
  *      the values are the names of the alternates for the key which is a ground,
  *         and (ISFactoryUtilityImpl.IS_ALTERNATE_OF+ground name) for the key which is an alternate.<P>
  * If _node is an ACS, _alter should be also an ACS or an ACS composite ACS.
  * If _node is a subACS, _alter should be also a subACS or a subACS composite.
  * @param _node is a non-null node of an ACS or a subACS
  * @param _alter is an alternate name to add. Never null.
  * @exception BaseError if an argument is null or is not a nametree node,
  *    if the two nodes have not the same type (ACS or subACS), or if
  *    the number of alternates for _node is greater than 4.
  */   
  protected synchronized void addAlternate(ImmutableName _node, 
          ImmutableName _alter) throws BaseError {

      if (_node == null || _alter == null || !isNode(_node) || !isNode(_alter) ||
              _alter.equals(_node))
          throw new BaseError("in NameISTreeImpl.addAlternate()," +
            " an argument is null or not a nametree node,"+
            " or the alternate is equal to the ground.");
      if (!getType(_node).equals(ISFactoryUtilityImpl.ACS_NODE) &&
              !getType(_node).equals(ISFactoryUtilityImpl.SUB_ACS_NODE))
          throw new BaseError("in NameISTreeImpl.addAlternate()," +
            " the ground is not an ACS nor a subACS.");
      if (getType(_node).equals(ISFactoryUtilityImpl.ACS_NODE) &&
              !getType(_alter).equals(ISFactoryUtilityImpl.ACS_NODE) &&
              !getType(_alter).equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE))
          throw new BaseError("in NameISTreeImpl.addAlternate()," +
            " the alternate is not an ACS while it is the type of the grounde.");
      if (getType(_node).equals(ISFactoryUtilityImpl.SUB_ACS_NODE) &&
              !getType(_alter).equals(ISFactoryUtilityImpl.SUB_ACS_NODE) &&
              !getType(_alter).equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE))
          throw new BaseError("in NameISTreeImpl.addAlternate()," +
            " the alternate is not a subACS while it is the type of the ground.");
      
      // add _alter at the end of the current array
      ImmutableName[] l_old = (ImmutableName[]) m_Alternates_.get(_node);
      
      if (l_old == null) 
          setAlternates(_node, new ImmutableName[] {_alter});
      else {
          ImmutableName[] l_new = new ImmutableName[l_old.length + 1];
          System.arraycopy(l_old, 0, l_new, 0, l_old.length);
          l_new[l_new.length - 1] = _alter;
          setAlternates(_node, l_new);
          l_new = null;
      } // else

      // adds IS_ALTERNATE_OF+_node at the end of the current array for _alter
      l_old = (ImmutableName[]) m_Alternates_.get(_alter);
      
      if (l_old == null) 
          setAlternates(_alter, new ImmutableName[] {
            (new NameImpl(ISFactoryUtilityImpl.IS_ALTERNATE_OF)).extended(_node) });
      else {
          ImmutableName[] l_new = new ImmutableName[l_old.length + 1];
          System.arraycopy(l_old, 0, l_new, 0, l_old.length);
          l_new[l_new.length - 1] = (new NameImpl(ISFactoryUtilityImpl.
                  IS_ALTERNATE_OF)).extended(_node);
          setAlternates(_alter, l_new);
          l_new = null;
      } // else

      l_old = null;   
  } // addAlternate()
  
  /**
  * Removes a node property for the Alternate relation: <P>
  *      mandatory key: node_name; <P>
  *      the values are the names of the alternates for the key which is a ground,
  *         and (ISFactoryUtilityImpl.IS_ALTERNATE_OF+ground name) for the key which is an alternate. <P>
  * @param _node is a non-null node of a ground ACS or subACS
  * @param _alter is an alternate name. Never null.
  * @exception BaseError if an argument is null or is not a nametree node, or
  *    if _alter is not known as an alternate of _node.
  */   
  protected synchronized void removeAlternate(ImmutableName _node, ImmutableName _alter)
    throws BaseError {
      if (_node == null || _alter == null || !isNode(_node) || !isNode(_alter) ||
              m_Alternates_.get(_node) == null )
          throw new BaseError("in NameISTreeImpl.removeAlternate()," +
            " an argument is null or not a nametree node,"+
            " or the node has no alternates.");
      if (getGroundsOfAlternate(_alter) == null ||
              !getGroundsOfAlternate(_alter).contains(_node))
          throw new BaseError("in NameISTreeImpl.removeAlternate(), " +
                  "_alter is not an alternate of _node");

      List<ImmutableName>   l_new;
      ImmutableName[]       l_old = (ImmutableName[]) m_Alternates_.get(_node);
      
      if (l_old.length == 1) {
            // there is no more alternate for _node
            synchronized (m_Alternates_) {
                setAlternates(_node, null);
            }          
            l_old = null;
      } // if (l_old.length ==
      
      // there are other alternates or grounds to keep in l_old
      else {
            // remove _alter
            l_new = new ArrayList(5);
            for (int i = 0; i < l_old.length; i++) {
                if (l_old[i].equals(_alter))
                    continue;
                else
                    l_new.add(l_old[i]);
            }
            synchronized (m_Alternates_) {
                setAlternates(_node, l_new.toArray(new ImmutableName[0]));
            }          
            l_old = null;
            l_new = null;
      } // if (match_count == 0

      l_old = (ImmutableName[]) m_Alternates_.get(_alter);

      if (l_old.length == 1) {
            // there is no more alternate or ground for _alter
            synchronized (m_Alternates_) {
                setAlternates(_node, null);
            }
            l_old = null;
            return;
      } // if (l_old.length ==

      // there are other alternates or grounds to keep in l_old
      else {
            // remove _alter
            l_new = new ArrayList(5);
            for (int i = 0; i < l_old.length; i++) {
                if (l_old[i].equals((new NameImpl(ISFactoryUtilityImpl.IS_ALTERNATE_OF)).
                        extended(_node)))
                    continue;
                else
                    l_new.add(l_old[i]);
            }
            synchronized (m_Alternates_) {
                setAlternates(_node, l_new.toArray(new ImmutableName[0]));
            }
            l_old = null;
            l_new = null;
            return;
      } // if (match_count == 0

  } // removeAlternate()
  
  /**
  * Gets the grounds for which the argument is an alternate.
  * @param _alter is the node name of an Alternate.
  * @return the list of grounds for _alter. May be null.
  */   
  public List<ImmutableName> getGroundsOfAlternate(ImmutableName _alter) {
      if (m_Alternates_.isEmpty() ) return null;
      if (_alter == null) return null;

      List<ImmutableName> l_result = new ArrayList(INITIAL_CAPACITY);
      
      for (ImmutableName name : m_Alternates_.get(_alter)) {
          if (name.component(0).equals(ISFactoryUtilityImpl.IS_ALTERNATE_OF))
              l_result.add( (ImmutableName) name.subname(1, name.noOfComponents() - 1));
      } // for name
 
      if (l_result.isEmpty()) {
          l_result = null;
          return null;
      } else return l_result;
  } // getGroundsOfAlternate()
  
  /**
  * Gets the alternates of a node. <P>
  * @param _node is a non-null node name of a ground
  * @return the array of alternates for _node. May be null but not empty.
  */   
  public List<ImmutableName> getAlternates(ImmutableName _node) {
      if (m_Alternates_.isEmpty() ) return null;
      if (_node == null) return null;

      List<ImmutableName> l_result = new ArrayList(INITIAL_CAPACITY);

      for (ImmutableName name : m_Alternates_.get(_node)) {
          if (!name.component(0).equals(ISFactoryUtilityImpl.IS_ALTERNATE_OF))
              l_result.add( (ImmutableName) name.clone());
      } // for name

      if (l_result.isEmpty()) {
          l_result = null;
          return null;
      } else return l_result;
  } // getAlternates()
  
  /**
  * Sets the node property for the alternate relation: <P>
  *      mandatory key: node_name for a ground ACS or subACS; <P>
  *      the values are the names for the key. <P> 
  * @param _node is a node with the type ACS or subACS
  * @param _l_alters is the array of alternates to copy. Null to remove _node.
  * @exception BaseError if _l_alters length is greater than 4
  */   
  private void setAlternates(ImmutableName _node, ImmutableName[] _l_alters)
    throws BaseError {
      if (_l_alters.length > 4 )
          throw new BaseError("in NameISTreeImpl.setAlternates()," +
            " _l_alters length is greater than 4.");
      
      if (_l_alters == null) {
          m_Alternates_.remove(_node);
          return;
      }

      ImmutableName[] l_result = new ImmutableName[_l_alters.length];
      for (int i = 0; i < _l_alters.length; i++) {
          l_result[i] = (ImmutableName)_l_alters[i].clone();
      } // for i
      m_Alternates_.put((ImmutableName)_node.clone(), l_result);
      l_result = null;     
  } // setAlternates()
  
  /**
  * Gets the mandatory paths from a node or an area to access to other nodes, if rights allow it.
  * The returned value is a list of node name arrays, which are 
  * the mandatory paths to go accross. 
  * Exemples of the array structure is an one-item List, or a List.<P>
  * For instance, if the internal mandatory paths array for _node is: <P>
  *     - (first_node, second_node, _node, third_node, fourth_node, _node), <P>
  * then the first path in the returned list is from _node to second_node, <P>
  * through first_node, where all the nodes are in the visible space of _node <P>
  * The use of this property in BaseManagerImpl is to set the unique composite of a subcomposite. <P>
  * Caution: the returned list may contain a name which is not an area in the nametree.
  * This is because, for a composite subACS relation, if the composite subACS has been removed,
  * the path remains in each subACS to tell 'this is an unworkable subACS'.
  * @param _node is a node in the ImmutableNameISTree
  * @return the list of arrays for mandatory path copies. May be null nut not empty.
  */   
  public List<ImmutableName[]> getMandatoryPaths(ImmutableName _node) {
      ImmutableName[] l_p = (ImmutableName[])m_MandatoryPaths_.get(_node);      
      if (l_p == null) return null;
      ImmutableName[] l_copy = new ImmutableName[l_p.length];
      ImmutableName[] l_new;
      List l_result = new ArrayList(10);
      int pathSize = 0;
      
      for (int i = 0; i < l_p.length; i++) {
          if (!l_p[i].equals(_node)) {
              l_copy[i] = (ImmutableName)l_p[i].clone();
              pathSize++;
          } else {
              l_new = new ImmutableName[pathSize];
              System.arraycopy(l_copy, i - pathSize, l_new, 0, pathSize);
              l_result.add(l_new);
              pathSize = 0;
          } // else
      } // for i
      
      l_p = null;
      l_copy = null;
      l_new = null;
      return l_result;
  } // getMandatoryPaths()
    
  /**
  * Tests a path as mandatory path from a node to access to other nodes.
  * The path is described as an array of node names which are the mandatory paths
  * to go accross. An exemple of the array structure is: <P>
  *      (first_node||second_node||beginning_node||third_node||fourth_node||beginning_node), <P>
  *      which indicates that there are two paths: from beginning_node to second_node, <P>
  *      and from beginning_node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * Call getMandatoryPathIndex().
  * @param _node is a non-null node in the ImmutableNameISTree
  * @param _l_path is the array of the path to test, without _node. Never null.
  * @return true if _l_path is a mandatory path of _node
  */   
  public boolean isMandatoryPath(ImmutableName _node, ImmutableName[] _l_path) {
      if (getMandatoryPathIndex(_node, _l_path) == -1 ) return false;
      else return true;
  } // isMandatoryPath()
   
  /**
  * Adds a mandatory path from a node to access to other nodes.
  * The path is described as an array of node names which are the mandatory paths
  * to go accross. An exemple of the array structure is: <P>
  *      (first_node||second_node||_node||third_node||fourth_node||_node), <P>
  *      which indicates that there are two mandatory paths: from _node to second_node, <P>
  *      and from _node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * This property is used in BaseManagerImpl to set the unique composite of a subcomposite. <P>
  * @param _node is a non-null node or area in the ImmutableNameISTree
  * @param _l_path is the array of one mandatory path, without _node. Never null.
  * @exception BaseError if _node is null, empty or unknown, if _l_path has not
  *     not the right ImmutableNames structure, if _l_path is null or contains _node,
  *     or if its length is greater than 4, if the names are not in the _node visible space.
  */   
  protected synchronized void addMandatoryPath(ImmutableName _node, ImmutableName[] _l_path)
    throws BaseError {

      if (_l_path == null || _l_path.length > 4 )
          throw new BaseError("in NameISTreeImpl.addMandatoryPath()," +
            " _l_path is null or its length is greater than 4.");
      int index_error = -1;
      for (int i = 0; i < _l_path.length; i++) {
          if (_l_path[i] == null || _l_path[i].equals(_node)) {
              index_error = i;
              break;
          }
      } // for i
      if (index_error != -1)
          throw new BaseError("in NameISTreeImpl.addMandatoryPath()," +
                 " the _l_path value at "+index_error+" is null or is _node.");
      
      // add _node at the end of _l_path
      ImmutableName[] l_p = new ImmutableName[_l_path.length+1];
      System.arraycopy(_l_path, 0, l_p, 0, _l_path.length);
      l_p[l_p.length - 1] = (ImmutableName) _node.clone();
      
      ImmutableName[] l_old = (ImmutableName[]) m_MandatoryPaths_.get(_node);
      ImmutableName[] l_new;
      
      if (l_old == null) setMandatoryPaths(_node, l_p);
      else {
        l_new = new ImmutableName[l_old.length + l_p.length];
        System.arraycopy(l_old, 0, l_new, 0, l_old.length);
        System.arraycopy(l_p, 0, l_new, l_old.length, l_p.length);
        synchronized (m_MandatoryPaths_) {
            setMandatoryPaths(_node, l_new);
        }
      } // else
      l_p = null;
      l_old = null;
      l_new = null;     
  } // addMandatoryPath()
   
  /**
  * Removes a mandatory path from a node.
  * An exemple of the array structure is: <P>
  *      (first_node||second_node||_node||third_node||fourth_node||_node), <P>
  *      which indicates that there are two paths: from _node to second_node, <P>
  *      and from _node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * @param _node is a non-null node or area in the ImmutableNameISTree
  * @param _l_path is the array of one mandatory path, without _node which 
  *     is the beginning node of the mandatory path. Never null.
  * @exception BaseError if _node is null, empty or unknown, if _l_path is null,
  *   empty, or not a _node mandatory path.
  */   
  protected synchronized void removeMandatoryPath(ImmutableName _node, ImmutableName[] _l_path)
    throws BaseError {
      int beginning_search_index = getMandatoryPathIndex(_node, _l_path);

      if (beginning_search_index == -1)
          throw new BaseError("in NameISTreeImpl.removeMandatoryPath()," +
            " _node is null, empty or unknown, if _l_path is null,"+
            " empty, or not a _node mandatory path.");
      
      ImmutableName[] l_old = (ImmutableName[]) m_MandatoryPaths_.get(_node);
      ImmutableName[] l_new;
      
      if (l_old.length == _l_path.length) {
            // there is no more path for _node
            synchronized (m_MandatoryPaths_) {
                m_MandatoryPaths_.remove(_node);
            }          
            l_old = null;
            return;
      } // if (l_old.length ==
      
      // there are other paths to keep in l_old
      else {
            // remove _l_path and the beginning node after it
            l_new = new ImmutableName[l_old.length - _l_path.length - 1];
            if (beginning_search_index != 0)
                // copy the paths before _l_path
                System.arraycopy(l_old, 0, l_new, 0, beginning_search_index);
            // copy the paths after _l_path, excluding the beginning node after it
            System.arraycopy(l_old, beginning_search_index + _l_path.length + 1,
                  l_new, beginning_search_index, 
                  l_old.length - _l_path.length - 1);
            synchronized (m_MandatoryPaths_) {
                m_MandatoryPaths_.put((ImmutableName)_node.clone(), l_new);
            }          
            l_old = null;
            l_new = null;
            return;
      } // if (match_count == 0
    
  } // removeMandatoryPath()
  
  /**
  * Gets the optional links for a name, which depends on the ImmutableNameISTree use.
  * To test if a node is a tied child, uses the instruction
  * Arrays.asList(ISTree.getLinks(nodeName)).contains(ISFactoryUtilityImpl.IS_TIED_CHILD).
  * @param _name is a node/area name or the association of two node/area names
  * in the ImmutableNameISTree
  * @return the links. May be a 0-length array. Never null.
  */   
  public String[] getLinks(ImmutableName _name) {
      String[] l_s = (String[]) m_Links_.get(_name);
      if (l_s == null) return new String[0];
      String[] l_result = new String[ l_s.length ];
      System.arraycopy(l_s, 0, l_result, 0, l_s.length);
      l_s = null;
      return l_result;
  } // getLinks()
   
  /**
  * Tests the presence of a link for a name.
  * @param _name is a node/area name or the association of two node/area names
  * in the ImmutableNameISTree
  * @param _link is a link to test for the name.
  * @return true if _link is a link of _name
  */   
  public boolean isLinked(ImmutableName _name, String _link) {
      if (_name == null || _name.isEmpty() || !isNode(_name) )
          return false;
      if (_link == null) return false;
      boolean result = false;
      String[] l_s = (String[]) m_Links_.get(_name);
      if (l_s == null) return false;

      for (int i = 0; i < l_s.length; i++) {
          if (l_s[i].equals(_link)) {
              result = true;
              break;
          }
      } // for i
      l_s = null;
      return result;        
  } // isLinked()
  
  /**
  * Sets the special links of the name, which depends on the ImmutableNameISTree use.
  * It is an optional node property. There is no control on the _l_links values.
  * Sees this class comments about the use of this method by BaseManager.
  * @param _name is a non-null node/are or the association of two node/area names
  *     in the ImmutableNameISTree
  * @param _l_links defines the nature of the links for the name. Null to remove the _name links.
  * @exception BaseError if _name is null, empty or unknown, _l_links null,
  *    or if its length is greater than 12.

  */   
  protected synchronized void setLinks(ImmutableName _name, String[] _l_links)
          throws BaseError {
      if (_name == null || _name.isEmpty() ||
              (!isNode(_name) && ! isArea(_name)))
          throw new BaseError("in NameISTreeImpl.setLinks()," +
            " the node is null, empty, or unknown.");
      if (_l_links.length > 12 )
          throw new BaseError("in NameISTreeImpl.setMandatoryPaths()," +
            " _l_links is null or its length is greater than 12.");
      
      if (_l_links == null) m_Links_.remove(_name);
      else m_Links_.put((ImmutableName)_name.clone(), _l_links);
  } // setLinks()
  
  /**
  * Adds a child as a new node in the nametree. Calls the super method.
  * The default value of the visible space is ALL_THE_IS_SPACE. The resulting node
  * is closed.
  * @param _parent is a node in the nametree. May be null, and then the child
  *     is added to the root. Never an area.
  * @param _child is the unique name of the new child of _parent, which is put at the end
  *   of the children array
  * @param _node_type is an allowed type for the _parent children, following the
   *        ISImpl policy.
  * @param _is_hub is true if the child is a node hub, which means that for this node,
  *    any direct child may have exchanges with any other direct child, if the rights allow them.
  * @param _l_links defines the nature of the links for the node. Never null.
  * @exception BaseError if _parent is not a node of this nametree,
  *     if _child is null, is already a nametree node, if 
  *     the parent does not allow this type for its children, _l_links null,
  *     or if its length is greater than 12.
  */
  protected synchronized void addChild(ImmutableName _parent, ImmutableName _child,
    String _node_type, boolean _is_hub, String[] _l_links) throws BaseError {

                                    // for testing and tracking
                                //System.out.println("NameISTreeImpl.addChild() for "+
                                  //  _child);

        if (_parent != null &&
                !ISFactoryUtilityImpl.getChildTypesForISnode(getType(_parent)).
                    contains(_node_type))
            throw new BaseError("in NameTreeImpl.addChild()" +
                    ", the parent type "+getType(_parent)+" does not allow this type "+
                    _node_type+" for its children.");
        else if (_parent == null &&
                !ISFactoryUtilityImpl.getChildTypesForISnode(getType(getRoot())).
                    contains(_node_type)) 
            throw new BaseError("in NameTreeImpl.addChild()" +
            ", the parent does not allow this type for its children.");

        if (_parent != null)
            super.addChild(_parent, _child, _node_type, _is_hub);
        else
            super.addChild(getRoot(), _child, _node_type, _is_hub);

                                    // for testing and tracking
                                //System.out.println("NameISTreeImpl.addChild(), "+
                                  //      "call to setLinks() for "+_child);
        
        setLinks(_child, _l_links);
  } // addChild()
  
  /**
  * Adds an empty area in the nametree, which is defined as a set of nodes in the nametree.
  * These implied nodes may be in any locations. <P>
  * Areas are used to define virtual structures like a composite ACS which contains several
  * pieces of software running on different servers, and defined in the nametree
  * as nodes. <P>
  * The area name cannot be a node name, and areas are not explicitly in the nametree.
  * Nevertheless, they are often put graphically under the nametreee node. 
  * Areas may be included in areas, but the nametree does not manage this relation.
  * The extended visible space of an area node is always ALL_THE_IS_SPACE.
  * Overrides the NameTreeImpl method.
  * @param _area is the area name.
  * @param _node_type is a value of the type. May be null.
  * @param _l_links defines the nature of the links for the node. Never null.
  * @exception BaseError if _area is already an area name or a node name.
  */
  protected synchronized void addArea(ImmutableName _area, String _node_type,
         String[] _l_links) throws BaseError {
        super.addArea(_area, _node_type);

                    // for testing and tracking
                //System.out.println("in NameISTreeImpl.addArea(), call to setLinks() for "+
                  //  _area);

        setLinks(_area, _l_links);
        m_ExtendedVisibleSpaces_.put(_area, ALL_THE_IS_SPACE);
  } // addArea()

  /**
  * Removes an area in the nametree. Overrides the super method.
  * @param _area is an area name
  * @exception BaseError if _area is not an area name
  */
  protected synchronized void removeArea(ImmutableName _area) throws BaseError {
        super.removeArea(_area);
        setLinks(_area, null);
        m_ExtendedVisibleSpaces_.remove(_area);
        m_OpenNodes_.remove(_area);
  } // removeArea()

  /**
   * Removes the internal data array for a leaf in the nametree,
   * with an update of the parent, the nametree internal data,
   * but no removing of the node children.
   * Caution: any special link where the key contains _node has to be removed
   * by the calling class.
   * The root cannot be removed.
   * Overrides the NameTreeImpl method.
   * @exception BaseError if _node is the root, or does not belong to the nametree,
   *   or is not a leaf, or if the base is not consistent.
   */
   protected synchronized void removeLeaf(ImmutableName _node) throws BaseError {
        super.removeLeaf(_node);
        m_OpenNodes_.remove(_node);
        m_MandatoryPaths_.remove(_node);

        ImmutableName[] l_names = null;
        // remove _node as value in mandatory paths
        for (Object name : new HashSet(m_MandatoryPaths_.keySet())) {
            for (Object array : getMandatoryPaths((ImmutableName) name)) {
                l_names = (ImmutableName[]) array;
                if (Arrays.asList(l_names).contains(_node)) {
                    removeMandatoryPath((ImmutableName) name, l_names);
                }
            } // for list
        } // for name

        m_ExtendedVisibleSpaces_.remove(_node);
        
        List<ImmutableName> l_s = getGroundsOfAlternate(_node);
        if (l_s != null) {
            for (ImmutableName ground : l_s) {
                removeAlternate(ground, _node);
            } // for
        } // if
        l_s = getAlternates(_node);
        if (l_s != null) {
            for (ImmutableName ground : l_s) {
                removeAlternate(_node, ground);
            } // for
        } // if

        m_Links_.remove(_node);

        l_names = null;
        l_s = null;
  } // removeLeaf()  
   
  /**
  * @return a cloned and no-backed instance of NameISTree, with clone of all names.
  */
  public Object clone() {
      NameISTreeImpl clon = (NameISTreeImpl) super.clone();

      // variables
      Map.Entry         ent;
      ImmutableName[]   array;
      ImmutableName[]   array_two;
      String            value;
      String[]          array_three;
      String[]          array_four;
      
      clon.m_OpenNodes_ = new HashMap(INITIAL_CAPACITY);
      for (Iterator ite = m_OpenNodes_.entrySet().iterator(); ite.hasNext(); ) {
          ent = (Map.Entry) ite.next();
          clon.m_OpenNodes_.put(
                  (ImmutableName)((ImmutableName)ent.getKey()).clone(), (Boolean)ent.getValue());
      } // for ent
      
      clon.m_ExtendedVisibleSpaces_ = new HashMap(INITIAL_CAPACITY);
      for (Iterator ite = m_ExtendedVisibleSpaces_.entrySet().iterator(); ite.hasNext(); ) {
          ent = (Map.Entry) ite.next();
          value = (String) ent.getValue();
          clon.m_ExtendedVisibleSpaces_.put(
                  (ImmutableName)((ImmutableName)ent.getKey()).clone(), value);
      } // for ite
      
      clon.m_Alternates_ = new HashMap(INITIAL_CAPACITY);
      for (Iterator ite = m_Alternates_.entrySet().iterator(); ite.hasNext(); ) {
          ent = (Map.Entry) ite.next();
          array = (ImmutableName[]) ent.getValue();
          array_two = new ImmutableName[array.length];
          for (int i = 0; i < array.length; i++) {
              array_two[i] = (ImmutableName) ((ImmutableName)array[i]).clone();
          }
          clon.m_Alternates_.put((ImmutableName)((ImmutableName)ent.getKey()).clone(), array_two);
      } // for ite      
      
      clon.m_MandatoryPaths_ = new HashMap(INITIAL_CAPACITY);
      for (Iterator ite = m_MandatoryPaths_.entrySet().iterator(); ite.hasNext(); ) {
          ent = (Map.Entry) ite.next();
          array = (ImmutableName[]) ent.getValue();
          array_two = new ImmutableName[array.length];
          for (int i = 0; i < array.length; i++) {
              array_two[i] = (ImmutableName) ((ImmutableName)array[i]).clone();
          }
          clon.m_MandatoryPaths_.put(
                  (ImmutableName)((ImmutableName)ent.getKey()).clone(), array_two);
      } // for ite

      clon.m_Links_ = new HashMap(INITIAL_CAPACITY);
      for (Iterator ite = m_Links_.entrySet().iterator(); ite.hasNext(); ) {
          ent = (Map.Entry) ite.next();
          array_three = (String[]) ent.getValue();
          array_four = new String[array_three.length];
          System.arraycopy(array_three, 0, array_four, 0, array_four.length);
          clon.m_Links_.put(
                  (ImmutableName)((ImmutableName)ent.getKey()).clone(), array_four);
      } // for ent

      array = null;
      array_two = null;
      array_three = null;
      array_four = null;
      ent = null;
      value = null;
      return clon;
  }   // clone()
  
  /**
   * Compares the values in the internal maps of this class.
  * @param _obj is the object to compare
  * @return true if the lists of values are equal
  */
  public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      if (!super.equals(_obj))
          return false;
      
      NameISTreeImpl obj = (NameISTreeImpl) _obj;

      if (m_OpenNodes_ == null || obj.m_OpenNodes_ == null) {
          if (m_OpenNodes_ != null || obj.m_OpenNodes_ != null)
            return false;  // if
      } // if
      else {
        if (!m_OpenNodes_.values().equals( obj.m_OpenNodes_.values() ))
          return false; }  
      
      if (m_ExtendedVisibleSpaces_ == null || obj.m_ExtendedVisibleSpaces_ == null) {
          if (m_ExtendedVisibleSpaces_ != null || obj.m_ExtendedVisibleSpaces_ != null)
            return false;  // if
      } // if
      else {
        if (!m_ExtendedVisibleSpaces_.values().equals( obj.m_ExtendedVisibleSpaces_.values() ))
          return false; }     
      
      if (m_MandatoryPaths_ == null || obj.m_MandatoryPaths_ == null) {
          if (m_MandatoryPaths_ != null || obj.m_MandatoryPaths_ != null)
            return false;  // if
      } // if
      else {
        if (!m_MandatoryPaths_.values().equals( obj.m_MandatoryPaths_.values() ))
          return false; } 
      
      if (m_Links_ == null || obj.m_Links_ == null) {
          if (m_Links_ != null || obj.m_Links_ != null)
            return false;  // if
      } // if
      else {
        if (!m_Links_.values().equals( obj.m_Links_.values() ))
          return false; }       

      if (m_Alternates_ == null || obj.m_Alternates_ == null) {
          if (m_Alternates_ != null || obj.m_Alternates_ != null)
            return false;  // if
      } // if
      else {
        if (!m_Alternates_.values().equals( obj.m_Alternates_.values() ))
          return false; } 
      
      return true;
  } // equals()
  
  /**
  * @return the hashcode from the super class.
  */
  public int hashCode() {
       return super.hashCode();
  }   

  /**
  * @return long serial version UID for serialization of the class
  */
  public long getSerialVersionUID() {
       return serialVersionUID;
  }    
  
  //
  // ------------- protected methods -------------------
  //

  /**
   * This method nullifies all the ImmutableNameISTree inner properties.
   */
  protected synchronized void finalizeForUser() {
    m_OpenNodes_ = null;
    m_ExtendedVisibleSpaces_ = null;
    m_MandatoryPaths_ = null;
    m_Links_ = null;
    m_Alternates_ = null;
    super.finalizeForUser();
  } // finalizeForUser()

  /**
   * This method calls finalizeForUser().
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    finalizeForUser();
    super.finalize();
  } // finalize()
  
  //
  // ------------ private methods ----------------------
  //
  
  /**
  * Sets the mandatory paths from a node to access to other nodes.
  * These paths may be not active if the proper inter-node rights are not there.
  * The paths are in an array of node names which are the structurally-mandatory paths
  * to go accross (if rights allow it). Exemples of the array structure is: <P>
  *      (first_node, second_node, _node, third_node, fourth_node, _node), <P>
  *      which defines two mandatory paths: from _node to second_node, <P>
  *      and from _node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * The true use of this property is to set the subcomposites of a composite,
  * which is an area, and the unique composite of a subcomposite. 
  * So, there is only single-node paths for the moment.
  * For a composite subACS, the subcomposite names are the ACS name
  * ended, as last component, by the subACS name. <P>
  * @param _node is a non-null node in the ImmutableNameISTree
  * @param _l_paths is the array of mandatory paths to copy. Never null.
  * @exception BaseError if _node is null, empty or unknown, if _l_paths has not
  *     not the right ImmutableNames structure, if _l_paths is null or is not ending with _node,
  *     or if its length is greater than 20, if the names are not in the _node visible space.
  */   
  private void setMandatoryPaths(ImmutableName _node, ImmutableName[] _l_paths)
    throws BaseError {
      if (_node == null || _node.isEmpty() || !isNode(_node) )
          throw new BaseError("in NameISTreeImpl.setMandatoryPaths()," +
            " the node is null, empty, or unknown.");
      if (_l_paths == null || _l_paths.length > 20 )
          throw new BaseError("in NameISTreeImpl.setMandatoryPaths()," +
            " _l_paths is null or its length is greater than 20.");
      int index_error = -1;
      for (int i = 0; i < _l_paths.length; i++) {
          if (!isNode(_l_paths[i]) && !isArea(_l_paths[i])) { // for now, it should be an area
              index_error = i;
              break;
          }
      } // for i
      if (index_error != -1)
          throw new BaseError("in NameISTreeImpl.setMandatoryPaths()," +
                 " the _l_paths value at "+index_error+" is not a nametree area or node.");
      if (!_l_paths[_l_paths.length - 1].equals(_node))
          throw new BaseError("in NameISTreeImpl.setMandatoryPaths()," +
                 " the last _l_paths value is not _node.");

      ImmutableName[] l_result = new ImmutableName[_l_paths.length];
      for (int i = 0; i < _l_paths.length; i++) {
          l_result[i] = (ImmutableName)_l_paths[i].clone();
      } // for i
      m_MandatoryPaths_.put((ImmutableName)_node.clone(), l_result);
      l_result = null;     
  } // setMandatoryPaths()
  
  /**
  * Tests a path as mandatory path from a node to access to other nodes.
  * The path is described as an array of node names which are the mandatory paths
  * to go accross. An exemple of the array structure is: <P>
  *      (first_node||second_node||beginning_node||third_node||fourth_node||beginning_node), <P>
  *      which indicates that there are two paths: from beginning_node to second_node, <P>
  *      and from beginning_node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * Called by removeManadatoryPath() and isMandatoryPath().
  * @param _node is a non-null node in the ImmutableNameISTree
  * @param _l_path is the array of the path to test, without _node. Never null.
  * @return -1 if _l_path is not a mandatory path of _node, or the index of the beginning
  *    of _l_path in the _node paths
  */   
  private int getMandatoryPathIndex(ImmutableName _node, ImmutableName[] _l_path) {

      if (_l_path == null || _l_path.length == 0)
          return -1;
      int index_error = -1;
      for (int i = 0; i < _l_path.length; i++) {
          if (_l_path[i] == null || _l_path[i].equals(_node)) {
              index_error = i;
              break;
          } // if
      } // for i
      if (index_error != -1)
          return -1;
      
      ImmutableName[] l_old = (ImmutableName[]) m_MandatoryPaths_.get(_node);
      if (l_old == null)
          return -1;
      if (l_old.length < _l_path.length) {
          l_old = null;
          return -1;
      }
      
      boolean match = false;
      int beginning_search_index = -1;
      
      for (int search = 0; search < l_old.length - _l_path.length + 1; 
        search++) {
          match = true;
          // detect if _l_path is a part of l_old
          for (int i = 0; i < _l_path.length; i++) {
              if (!l_old[i + search].equals(_l_path[i])) {
                  match = false;
                  break;
              } // if
          } // for i

          // _l_path is a part of l_old
          if (match) {
              beginning_search_index = search;
              break;
          } 
      } //for (int search = 0

      l_old = null;
      return beginning_search_index;      
  } // getMandatoryPathIndex()  
  
} // NameISTreeImpl class