//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;


import ARoad0.gBaseInterface.Name;
import ARoad0.gBaseInterface.ImmutableName;

import ARoad0.ErrB.*;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.Iterator;
import java.util.Arrays;
import java.util.Collections;
import java.io.Serializable;
import java.io.ObjectStreamField;

 /**
  * This class is responsible for managing the names for all the database objects.
  * Examples of real name are "http://accbee.com", "c:\program files\access road",
  * "net/users/john". The separator character depends on the ACS. All the real names
  * are converted in Access Road names implemented by this class. The single separator
  * in all the Access Road names is ':: '.
  * Name follows the OMG CORBA Object Services Specification (OMG-COSS)
  * for a naming service which names an object for later retrieval.
  * Methods added to the OMG specification are extended(ImmutableName),
  * extended(String), clone(), equals(), hashCode(), getSerialVersion(),
  * getNameFromString(). It does NOT follow the JNDI name interface.
  * <P>
  * The name is an ordered list of components in a java.util.ArrayList.
  * Each component is a string that may be null. Case must be respected.
  * ':: ' is a forbidden component. It is controlled in every creating or updating method.
  * The usual struture of an entity name is (IS name).(ACS name).(entity name)
  * Normally, the usual name of an instance of Name is equals
  * to name.lastComponent().
  * To remove any component, use without(). To add a component at the end,
  * use an extended() method. To modify the last component, use without() then extended().
  * For more complex modifying operations, use the list returned by components().
  * <P>
  * With threads, the calls to the Name methods should be done in a block synchronized
  * on the Name instance. If the name is used as a key in a map, it is much safer
  * to avoid a mutable name. Use then the ImmutableName interface to construct the key.
  * But the update will remain possible by casting the ImmutableName in Name.
  * The hashcode() and equals() methods are coherent over serialization,
  * then a map with a name key handles the same values.
  * @version 0.7.0
  * @since 0.4.0
  */
public final class NameImpl implements Name, Serializable {

  /*
  * private variables
  */
  private static final long serialVersionUID__ = 0;
  private static final int INITIAL_CAPACITY = 8 ;  // the default value is 10
  private static final int INCREMENT_CAPACITY = 2;
  private static final int PRIME = 1000003;

  /**
   * fCompo_ is an ArrayList that stores the name components.
   * With ArrayList, hashcode() and equals() are coherent over serialization,
   * then a map with a name key handles the same values after its deserialization.
   */
  private ArrayList fCompo_ = null;

	  /** standardized variable for the JDK 2 serialization
	  * @serial ("fCompo_", ArrayList.class)
	  */

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final NameImpl EMPTY_INSTANCE = new NameImpl();

  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("fCompo_", ArrayList.class)
  };

  /*
  * contructors.
  * The arguments when they exist are copied. So, the initial object
  * referenced by an argument constructor is never updated.
  * /

  /**
  * @return instance
  */
  public NameImpl() {
  }

  /**
   * The returned instance is empty if a component contains ':: '.
   * @param _components contains only strings, without ':: '. _components is copied.
   */
  public NameImpl(ArrayList _components) {

    boolean withTwoPoints = false;
    fCompo_ = new ArrayList(INITIAL_CAPACITY);

    for (int i = 0 ; i < _components.size() ; i++) {
      if (((String)_components.get(i)).indexOf(":: ") != -1 ) {
          withTwoPoints = true;
          break;
      }
    } // for

    if (!withTwoPoints) fCompo_ = (ArrayList)_components.clone();
  }

   /**
   * Returns a new empty name if the parameter is null or contains ':: '.
   * @param _st is the string name.
   */
  public NameImpl(String _st) {
    fCompo_ = new ArrayList(INITIAL_CAPACITY);
    if (_st != null && _st.indexOf(":: ") == -1) {
      fCompo_.add(_st);
    }
  } // constructor


  // ------------------------ Methods of the Name interface ------------------


  /**
   * @return true if the name has no component, or if all components
   * are null or equal to empty Strings.
   */
   public final boolean isEmpty() {
        if (fCompo_ == null) return true;
        boolean result = true;
        for (int i = 0 ; i < fCompo_.size() ; i++) {
          if (fCompo_.get(i) != null &&
            !((String)fCompo_.get(i)).equals(new String()) ) {
              result = false;
              break;
          }
        } // for
        return result;
    }

  /**
  * @return a copy of the current instance without the last component.
  * May be null.
  */
  public final Name context() {
    if (fCompo_ != null) {
        int p = fCompo_.size();
        ArrayList vcompo = new ArrayList(fCompo_);  // fCompo_ is copied
        vcompo.remove(p-1);
        return new NameImpl(vcompo);
    } else return null;
  }

  /**
   * Caution: this method changes the instance.
   * @param _name is removed from the current name. _name may be at the start,
   *    the end or the middle of the current name.
   *    If there are several sections of the current name that are strictly
   *    identical to _name, without() removes the last matched part,
   *    at the end of the current name.
   * @return this instance
   * With threads, this non-synchronized method (for performance)
   * must be used in a synchronized block. May be null.
   */
   public final Name without(Name _name) {

    if (fCompo_ == null) return null;
    if (fCompo_.containsAll(_name.components() )) {

      int sizeName = _name.noOfComponents();
      int sizeCompo = fCompo_.size();

      // use arrays to be fast, and the List interface to be easy to modify
      List lC = Arrays.asList(fCompo_.toArray(new String[sizeCompo]));
      List lN = Arrays.asList(_name.components().toArray(new String[sizeName]));

      int lastIndex = 100;   // arbitrary initial value
      int count = 0;   // count the number of identical elements (strings)
                       // in _name and fCompo_

      for (int i = sizeCompo; i > 0 ; i--) {
          // compare the last component of _name with every fCompo_ component
          count = 0;  // set to 0 before every test of _name last component

              while (((String)lC.get(i-1)).equals((String)lN.get(sizeName-1)) ) {
                  ++count;

                  if (sizeName == 1) break ;  // next test is usefulness

                  // test if all the _name components before the current last one
                  // are also matched, and increment count for each match :
                  if (i > 1 && sizeName > 1) {
                       // if i == 1 && sizeName > 1, _name is not extracted
                    for (int m = i-1, n = sizeName-1; m > 0 && n > 0 ; m--, n--)
                    {
                          if (((String)lC.get(m-1)).equals((String)lN.get(n-1)) )
                              ++count;
                    } // for m, n

                  } // if (i && sizeName)

                  break;  // the while loop is used only once

              } // while equals

          if (count != sizeName) continue;

          // if count == sizeName : the number of connections with fCompo_
          // is equals to the _name size
          lastIndex = i-1; // the last index to remove at the end of fCompo_
          break;

      }  // for i

      if (lastIndex == 0) fCompo_.remove(0);

      if (lastIndex != 100 && lastIndex != 0) {

          for (int k = lastIndex; k > (lastIndex - sizeName); k--) {
              // (lastIndex - sizeName) can't be negative : see loop on i
              fCompo_.remove(k);
          } // for k

      } // if lastIndex != 100 && 0

    } // if (fCompo_.containsAll(_name))

    return this;
  } // without(Name)

    /**
    * For extending the current name. Caution: this method updates the instance.
    * Noop if the argument is null or empty, or if ':: ' is in the argument,
    * because it is used in toString() and getNameFromString().
    * With threads, this method must be used in a synchronized block.
    * @param _nc adds a new component to the current instance
    * @return this instance
    */
  public final Name extended(String _nc) {
      if (_nc == null || _nc.indexOf(":: ") != -1 || _nc.equals("")) return this;
      if (fCompo_ == null) fCompo_ = new ArrayList(INITIAL_CAPACITY);
      fCompo_.add(_nc);    // does a copy of _nc as a new string at the end of fCompo_
      return this;
  } // extended(String)

    /**
    * For extending the current name.
    * The name is unchanged if _name is null or an empty instance,
    * or has a component which contains ":: " (see getNameFromString()).
    * @param _name adds a new name at the end of the current instance
    * @return this instance
    * @since 0.5.1
    */
  public final Name extended(ImmutableName _name) {
      if (_name == null || _name.equals(NameImpl.EMPTY_INSTANCE) ) return this;
      if (fCompo_ == null) fCompo_ = new ArrayList(INITIAL_CAPACITY);

      // test if all the name components does not contain ":: "
      boolean withTwoPoints = false;
      for (Iterator ite = _name.immutableComponents().iterator() ; ite.hasNext() ; ) {
        if (((String)ite.next()).lastIndexOf(":: ") != -1) {
          withTwoPoints = true;
          break;
        } // if
      } // for

      if (!withTwoPoints) {
        for (int i = 0 ; i < _name.noOfComponents() ; i++) {
          fCompo_.add(_name.component(i));
        } // for
      } // if
      return this;
  } // extended(ImmutableName)

  /**
   * Gets a copy of the current name less some last components.
   * @param _end is the _inclusive_ index of the last component to be returned
   * @return the context (from 0 to _end) of the current name in a new Name,
   *    this name if _end = (size of the name) - 1,
   *    this name.context() if _end = (size of the name) - 2,
   *    the first name component if _end = 0. The context is not backed
   *    by the first components of the current name. Return is null if
   *    the name is empty, or if _end is equal or superior to the name size.
   */
  public final Name context(int _end) {
    // fCompo_.subList would return a List and this is not
    // allowed by the NameImpl constructors. So create a new vector.
    if (fCompo_ == null) return null;
    if (fCompo_.size() <= _end) return null;
    int p = fCompo_.size();
    ArrayList vcompo = null;

    if (_end != p-1) {
        vcompo = new ArrayList(fCompo_.subList(0, _end + 1));
        return new NameImpl(vcompo);
    } // if
    else return new NameImpl( fCompo_ );

  } // context(int)

  /**
   * Returns a new Name instance. For instance, the name 'A' with 3 components
   * '1', '2' and '3' returns a new name with 1 component '2' through
   * the command A.subname(1, 2).
   * @param _start inclusive fromIndex
   * @param _end exclusive toIndex
   * @return an extract of the current name that is copied in a new Name. May be null.
   */
  public final Name subname(int _start, int _end) {
    // fCompo_.subList returns a List that is not allowed
    // by the NameImpl constructors. So, create a new vector.
    if (fCompo_ == null) return null;
    ArrayList vcompo = new ArrayList(fCompo_.subList(_start, _end));
    return new NameImpl(vcompo);
  } // subname(int, int)

   /**
    * Faster than immutableComponents().
   * @return List of the components, as strings, that is backed by the list
   *  of the current name.
   *  Then, the returned name is modifiable but
   *  the only proprietary of the returned list is the current name.
   *  So, to avoid any bug, USE THE RETURNED LIST AS A TRANSIENT OBJECT,
   *  for a limited time. Never delete it directly. May be null.
   */
  public final List<String> components() {
      return fCompo_;
  }

  /**
   * @return List of the components (strings) in the name. is immutable. may be null.
   */
  public final List immutableComponents() {
      if (fCompo_ == null) return null;
      return Collections.unmodifiableList(fCompo_);
  }

   /**
   * @return int number of components in the name.
   */
  public final int noOfComponents() {
    if (fCompo_ != null) return fCompo_.size();
    else return 0;
  }

   /**
    * Gets the component at a given position.
   * @param _i index of the component. No limit on the value.
   * @return the component at _i in the name. May be null.
   */
  public final String component(int _i) {
    if (fCompo_ == null || _i < 0 || _i > fCompo_.size() - 1) return null;
    String result = (String) fCompo_.get(_i); // do not return a direct reference

                // for testing and tracking
            //System.out.println("in NameImpl.component(), result: "+result);
    
    return result;
  }

   /**
    * Gets the last component of the name.
   * @return the last component (may be null)
   */
  public final String lastComponent() {
    if (fCompo_ == null) return null;
    int n = fCompo_.size() - 1;
    String s = new String();
    if (n == -1) return s;
    if (n != 0) {
      s = (String) fCompo_.get(n); // do not return pass a direct reference
      return s;
    }
    else {
      s = (String) fCompo_.get(0);
      return s;
    }
  }

   /**
   * @return true if _name components are all there in the instance,
   *  but not always in the same order
   * @param _name never null.
   */
  public final boolean contains(ImmutableName _name) {
    if (fCompo_ == null) return false;
    return fCompo_.containsAll(_name.components());
  }

  /**
   * Get a clone of name.
   * @exception InternalError if CloneNotSupportedException
   * @return a cloned and no-backed instance of mutable name
   * that may be cast as a Name or an ImmutableName. May be null.
   */
  public final Object clone() {
    NameImpl clon = null;
    try {
      clon = (NameImpl) super.clone();
      if (fCompo_ != null) clon.fCompo_ = (ArrayList) fCompo_.clone();
      else clon.fCompo_ = null;
    } catch (CloneNotSupportedException e) {
      // this shouldn't happen, since we are Cloneable
      throw new InternalError(e.toString() + "NameImpl.clone() failed");
    } // catch
    return clon;
  } // clone()

    /**
     * Compares the two names.
     * @param _obj the object to compare
     * @return true if the list of strings is equal - not only identical -
     *  even if the two objects are instances of the same class
     *  that implements Name (and not only ImmutableName)
     */
    public final synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (components() == null || ((Name)_obj).components() == null) {
        if (components() != null || ((Name)_obj).components() != null )
          return false;
        else return true; // if the two are null
      } // if
      else {
        if (!components().equals(((Name)_obj).components()) )
          return false;
      } // else

      return true;
  } // equals(Object)

    /**
     * @return the hashcode value for this instance of Name. May be null.
     */
    public final int hashCode() {
    int result = 0;
    if (fCompo_ != null) {
      result = fCompo_.hashCode();
    } // if
    return result;
  } // hashcode()

    /**
     * @param o Object to compare must be an ImmutableName (true for Name)
     * @exception ClassCastException - if the specified object's type
     *  prevents it from being compared to this current Object
     * @return 0 if the names are equals, a negative integer if this Name object
     *  is lesser than the argument, and a positive integer if this Name object
     *  is greater than the argument.
     *  The comparison is based on the Unicode value of each character
     *  in each Name component (starting from the first one)
     *  and the number of components (negative return if all strings of this Name
     *  are equals but with less components, or if the argument is null).
     */
    public final synchronized int compareTo(Object o) throws ClassCastException {
     int result = 0;
     if (!(o instanceof ImmutableName)) {
        ClassCastException c = new ClassCastException(" For the name "+lastComponent()+
        ", the argument of Name.compareTo() is not an ImmutableName instance "+o.toString());
        c.printStackTrace();
        throw c;
     } // if
     
     if (equals(o)) return result; // equals is true if the two have null fCompo_
     if (fCompo_ == null) return -1;

     ImmutableName obj = (ImmutableName)o;
     ListIterator iteratorF = fCompo_.listIterator(0);
     ListIterator iteratorO = obj.immutableComponents().listIterator(0);
     String stF;
     String stO;

     while ( (result == 0) && iteratorF.hasNext() && iteratorO.hasNext()) {
        stF = (String)iteratorF.next();
        stO = (String)iteratorO.next();
        if (stF != null && stO != null) result = stF.compareTo(stO);
        if (stF == null && stO != null) result = -1;
        if (stF != null && stO == null) result = 1;
        if (stF == null && stO == null) result = 0;
     }
     // if all compared strings are equals, sizes may be differents
     if (result == 0) {
        if (noOfComponents() < obj.noOfComponents() ) result = -1;
        if (noOfComponents() > obj.noOfComponents() ) result = 1;
     }
     return result;
  } // compareTo(Object)

  /**
   * @return long serial version UID for serialization of the class
   */
  public final long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * methods out of the extended interfaces:
   */

   /**
   * Gets a String representation with one leading whitespace and '::  ' at the end,
   * or gets '  null  '.
   * @return String
   */
  public final String toString() {
    StringBuilder sb = new StringBuilder();
    if (fCompo_ == null) {
        sb.append(" null ");
    }
    else {
        int p = fCompo_.size();

        for (int i = 0; i < p; i++) {
            sb.append(fCompo_.get(i));
            //if (i != p-1)
            sb.append(":: "); // adds each component of the name in sb
        } //for
    }
    return " " + sb + " ";
  } // toString()

  /**
   * Gets the name from the string returned by toString().
   * The toString() method returns a text with one leading and one trailing whitespaces.
   * To handle the user entering of a text, the argument may have more than one whitespace
   * before, and more than two whitespaces after the effective characters, since
   * this method gets off all these whitespaces before the processing.
   * @param _text name, as get through toString()
   * @return the name. Empty if _text equals "  null  ".
   * @exception BaseError if the string argument is null or empty,
   *  or if the string does not start with ' ' or is not ended by '::  '.
   */
    public final static Name getNameFromString(String _text) throws BaseError {
      // variables
      Name   result = new NameImpl();
      String text = null;

                              // for testing and tracking
                  //System.out.println("call to NameImpl.getNameFromString(),
                    //for the argument "+_text);

      // exception and simple cases
      if (_text == null || _text.length() < 6) throw new BaseError(
        " This argument "+_text+" cannot produce a name - it is null or empty.");
      
      if (_text.equals("  null  ")) return result;

      if (!_text.endsWith("  ")) {

                              // for testing and tracking
                  //System.out.println("in NameImpl.getNameFromString(), "+
                    //"error on _text: "+_text );
                  
        throw new BaseError(
          " This argument "+_text+" cannot produce a name - it is not ended by two whitespaces. ");
      }
      if (!_text.startsWith(" ")) {

                              // for testing and tracking
                  //System.out.println("in NameImpl.getNameFromString(), "+
                    //"error on _text: "+_text );
                  
        throw new BaseError(
          " This argument "+_text+" cannot produce a name - it does not start with one whitespace.");
      }

      // so, it works even if _text has more than 1 leading or 2 trailing whitespaces
      // new in 0.7.0
      text = _text.trim();

      if (!text.endsWith("::")) {

                              // for testing and tracking
                  //System.out.println("in NameImpl.getNameFromString(), "+
                    //"error on _text: "+_text );
                  
        throw new BaseError(
          " This argument "+_text+" cannot produce a name - it is not ended by 2 ':' and a whitespace.");
      }
      
      text = text + " "; // text is there a structured name to process
      int indexComponentStart = 0;
      int indexComponentEnd = text.indexOf(":: ");

      while(true) {

                // for testing and tracking
            //System.out.println("in NameImpl.getNameFromString()"+
            //  "; the argument is "+_text+
            //  "; indexComponentStart = "+indexComponentStart+
            //  "; indexComponentEnd = "+indexComponentEnd);

        // catch the component and put it in the result
        result.extended(text.substring(indexComponentStart, indexComponentEnd));

        // break if indexComponentEnd points on the last ":: " which ends the name
        if ( (indexComponentEnd + 3) >= text.length()) break;

        // move the two index to the next component in the text
        indexComponentStart = indexComponentEnd + 3;
        indexComponentEnd = indexComponentStart +
          text.substring(indexComponentStart,text.length()).indexOf(":: ");

      } // while

      text = null;
      return result;
    } // getNameFromString()

} // NameImpl