//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import javax.swing.tree.TreePath;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Collections;
        
import java.io.ObjectStreamField;
import java.io.Serializable;

 /**
  * This class is responsible for managing a fast-methods tree of ImmutableNames.
  * They are not designed to be displayed by the Gui1.CommonPropertyEditor subclasses,
  * but by trees in dedicated Gui1 internal frames.
  * The nodes are ImmutableNames, without any constraint between a node name 
  * and its parent node name. The node name may include the parent name or not.
  * However, the current use of this class is to enforce the parent name including.
  * The links between nodes are parent-child relations. <P>
  * This class reuses the names of the relevant methods in javax.swing.tree.DefaultTreeModel.
  * There is always a non-null name as root.
  * There is no sorting in the children list of a node.
  * A node name cannot be two times in the nametree. <P>
  * This class is used by the IS trees.
  * It is possible to get new subtrees, that is an independant nametree with a copy
  * of all the indirect children. The NameTree manages mainly nodes and areas.
  * An area is defined as a group of nametree-connected or
  * nametree-unconnected nodes in the nametree, with a specific area name.
  * Areas are used to define virtual structures like a composite ACS which contains several
  * pieces of software running on different servers, and defined in the nametree
  * as classical nodes. 
  * The area name cannot be a node name, and areas are put by convention under the nametreee root.
  * Areas may be included in areas, but the nametree does not manage this relation.
  * The visible space (see NameISTree) of an area node always includes this area.
  * Most of the methods are applicable both to nodes and areas, and the exceptions
  * are documented.  <P>
  * A node may have an immutable type. A node may be a hub, which means that for this node,
  * any direct child may have exchanges with any other direct child. 
  * For a non-hub node, exchanges between its children are not allowed.
  * The visible space (see NameISTree) of a node having a hub parent always 
  * includes all the children of its parent. <P>
  * The data structure is optimized for searchings from the root, with no more than
  * 20 nodes in any level.
  * This class does not fire property changes.
  * @version 0.7.0
  */
public class NameTreeImpl implements ImmutableNameTree, Serializable {

  /**
   * arraylists array, one array per level
   *   for the root:
   *     root_name, node_type, is_hub, nb_of_children, null
   *   for each level, arraylist of Object[5]
   *   one Object[5] for each node:
   *     0: node name (it is the key), 1: node type, 2: is_hub,
   *     3: nb_of_children (at the lower level), 4: parent name (at the upper level)
   * hashmap of nodes from levels:
   *     key: node_name
   *     value: node level
   *   index of a child: get through a search in the right level vector
   * hashmap of areas:
   *     key: area_name
   *     value: area nodes list
   */  

  /*
  * private variables
  */
  protected static final long serialVersionUID = 37L;
  protected static final int INITIAL_CAPACITY = 200 ;
  protected static final int INCREMENT_CAPACITY = 3;
  protected static final int PRIME = 1000003;
  
  /**
   * array of lists where the index is the level.
   * in each list, the value is an an array Object[5] for each node description:
   *    0: node name (it is the key), 1: node type, 2: is_hub,
   *    3: nb_of_children, 4: parent name
   * the nametree root is at the index 0, as a single item in the list
   */
  protected ArrayList[] l_NodeDescriptors_ = null;
  
  /**
   * hashmap of node levels:
   *    key: node_name
   *    value: node level in l_NodeDescriptors_
   * so, the key set gives all the node names.
   */
  protected HashMap<ImmutableName, Integer> m_Levels_ = null;
  
  /**
   * hashmap of areas:
   *    key: area_name, which may be a node name
   *    value: String node type at the index 0, and then node names of items in the area
   */
  protected HashMap<ImmutableName, List<String>> m_Areas_ = null;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final NameTreeImpl EMPTY_INSTANCE = new NameTreeImpl();

  /** standardized variable for the JDK 2 serialization
  * @serial ("l_NodeDescriptors_", ArrayList.class),
  * ("m_Levels_", HashMap.class), ("m_Areas_", HashMap.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("l_NodeDescriptors_", ArrayList[].class),
      new ObjectStreamField("m_Levels_", HashMap.class),
      new ObjectStreamField("m_Areas_", HashMap.class)
  };

  /**
  * The default constructor.
  */
  public NameTreeImpl() {
  }

  /**
   * Constructor.
   * The returned instance is empty if _root_name is null or empty.
   * @param _root_name is the name of the nametree root. Not null.
   * @param _node_type is a type for the root node. May be null.
   * @param _is_hub is true if the root is a node hub, which means that
   *    any direct child of this node may have exchanges with any other direct child,
   *    if the rights allow them.
   * @exception CreateError if _root_name is null or empty.
   */
  public NameTreeImpl(ImmutableName _root_name, String _node_type, boolean _is_hub) 
   throws CreateError {
    if (_root_name == null || _root_name.isEmpty() )
        throw new CreateError("in NameTreeImpl constructor, the root name is null or empty");
      
    Object[] l_root = new Object[5];
    l_root[0] = _root_name.clone();
    l_root[1] = _node_type;
    l_root[2] = new Boolean( _is_hub );
    l_root[3] = new Integer(0); // nb of children
    // l_root[4] is the parent
    
    l_NodeDescriptors_ = new ArrayList[1];
    l_NodeDescriptors_[0] = new ArrayList(INITIAL_CAPACITY);
    l_NodeDescriptors_[0].add(l_root);
    
    m_Levels_ = new HashMap(INITIAL_CAPACITY);
    m_Levels_.put((ImmutableName)_root_name.clone(), new Integer(0));
    
    m_Areas_ = new HashMap(INITIAL_CAPACITY);
    
    l_root = null;
  } // constructor

  // ------------------------ Methods of the ImmutableNameTree interface ------------------
    
    
  /**
   * @return true if the nametree has null properties, or has only a root as node.
   */
   public boolean isEmpty() {
       if (l_NodeDescriptors_ == null || l_NodeDescriptors_.length < 2) return true;
       else return false;
   } 
   
  /**
  * Gets the root of the nametree
  * @return the nametree root or null if finalized
  */
  public final ImmutableName getRoot() {

                // for testing and tracking
            //if (l_NodeDescriptors_ != null) {
              //  Object[] l_root = (Object[]) l_NodeDescriptors_[0].get(0);
              //  System.out.println("NameTreeImpl.getRoot(), l_NodeDescriptors_[0].get(0): ");
              //  for (int i = 0; i < l_root.length; i++) {
              //      System.out.println("--------- "+l_root[i]);
              //  }
            //}


       if (l_NodeDescriptors_ == null)
           return null;
       else
           return (ImmutableName) ((ImmutableName) (
                   (Object[]) l_NodeDescriptors_[0].get(0))[0]).clone();
  } 
  
  /**
   * Gets an independant subtree from the nametree, with all the indirect children.
   * This method is a good test of the nametree code bugs, and throws
   * an error when unconsistent data is detected.<P>
   * Changes on the returned nametree do not change this nametree.
   * Caution: the subclass does not override this method, so the returned object
   * is never an ImmutableNameISTree.
   * @param _root node of the nametree which will be the returned subtree root.
   *            Never the nametree root nor an area.
   * @exception BaseError if the root is the nametree root, null, empty, is an area name,
   *   does not belong to the nametree, or if non-consistent data is detected.
   * @return a deep copy of the subtree where _root is the new tree root.
   */
  public final ImmutableNameTree getSubTree(ImmutableName _root) throws BaseError {
      if (_root == null || _root.isEmpty() || !isNode(_root) ||
        getRoot().equals(_root) || getL_AreaNames().contains(_root))
          throw new BaseError(
            "in NameTreeImpl.getSubTree(), the root is the nametree root, null, "+
              "empty, is an area name, or does not belong to the nametree.");
      
      // init
      Integer   root_level = (Integer) m_Levels_.get(_root);
      int       root_index;
      root_index = getIndexInParent(_root);

      if (root_index == -1) throw new BaseError("in NameTreeImpl.getSubTree(), "+
        "getIndexInParent() does not found the index.");
      
      NameTreeImpl result;
      try {
          Object[] l_root = (Object[]) l_NodeDescriptors_[root_level.intValue()].
            get(root_index);
          result = new NameTreeImpl(_root,
            (String) l_root[1],
            ((Boolean) l_root[2]).booleanValue() );
          l_root = null;
      } catch (CreateError c) {
          // not possible
          throw new InternalError(c.toString());
      }
      
      if (getChildCount(_root) == 0) return result;
      
      // else, built up the _root subtree   
      Integer   nb_children = (Integer)
        ((Object[]) l_NodeDescriptors_[root_level.intValue()].get(root_index))[3];
      int       current_level = root_level.intValue()+ 1;
      int       child_count = 0;
      boolean   error = false;
      Object[]  l_values = null;
      int       nbOfLevels = nbOfLevels();
      
        // downing iteration to add the children to the parent
      while (current_level < nbOfLevels) {
          for (Iterator ite = l_NodeDescriptors_[current_level].iterator(); ite.hasNext(); ) {
                l_values = (Object[]) ite.next();
                // if result contains the parent, add the current node
                // to result as a child
                if ( result.isNode((ImmutableName) l_values[4]) ) {
                    result.addChild((ImmutableName) ((ImmutableName)l_values[4]).clone(),
                        (ImmutableName) ((ImmutableName)l_values[0]).clone(),
                        (String) l_values[1],
                        ((Boolean) l_values[2]).booleanValue());                    
                    child_count++;
                } // if
          } // for ite
          // test of child count only for _root
          if ((current_level == root_level.intValue()+ 1) &&
            child_count != nb_children.intValue()) {
                error = true;
                break;
          } // if
          current_level++;
      } // while current_level
      if (error) throw new BaseError(
              "in NameTreeImpl.getSubTree(), the first level of children" +
              " has not the right number of children.");
      
      l_values = null;
      return result;
  } // getSubTree()
  
  /**
   * Gets all the areas of this nametree, with their node names.
   * @return a Map of node Names. The key is the area name. 
   *        The value is a clone of the nodes list. May be empty.
   */
  public Map getM_AreaNodes() {
      Map result = new HashMap(INITIAL_CAPACITY);
      
      ImmutableName name;

      for (Iterator ite = m_Areas_.keySet().iterator(); ite.hasNext(); ) {
          name = (ImmutableName)ite.next();
          result.put(name.clone(), getL_NodesOfArea(name));
      } // for ite

      name = null;
      return result;
  } // getM_AreaNodes()
  
  /**
  * Gets the set of all the nametree nodes, including the root and excluding the areas.
  * @return ImmutableNames of nodes. Never null.
  */
  public Set<ImmutableName> nodes() {
      Set result = new HashSet(INITIAL_CAPACITY);
      for (Iterator ite = m_Levels_.keySet().iterator(); ite.hasNext(); ) {
          result.add( ((ImmutableName)ite.next()).clone() );
      } // for ite
      return result;
  } // nodes()
  
  /**
  * Gets the set of all the nametree area names, if any.
  * @return ImmutableNames of area names. Never null.
  */
  public final Set<ImmutableName> getL_AreaNames() {
      Set result = new HashSet(INITIAL_CAPACITY);
      for (Iterator ite = m_Areas_.keySet().iterator(); ite.hasNext(); ) {
          result.add( ((ImmutableName)ite.next()).clone() );
      } // for ite
      return result;
  } // areas()
  
  /**
  * @return the number of nodes in the nametree, including the root and the areas.
  */
  public final int nbOfNodesAndAreas() {
       return m_Levels_.keySet().size() + m_Areas_.keySet().size();
  }

  /**
   * The returned value is at least '1'.
  * @return the number of levels in the nametree, including the root and excluding the areas.
  */
  public final int nbOfLevels() {
      return l_NodeDescriptors_.length;
  }
  
  /**
  * Tests a name as node of the nametree, including the root.
  * @param _name to test
  * @return true if _name is a nametree node
  */
  public final boolean isNode(ImmutableName _name) {
      getRoot();
      return m_Levels_.keySet().contains(_name);
  }
  
  /**
  * Tests a name as area of the nametree.
  * @param _name to test
  * @return true if _name is a nametree area
  */
  public boolean isArea(ImmutableName _name) {
       return m_Areas_.keySet().contains(_name);
  }
  
  /**
  * Gets the nodes belonging to an area of the nametree.
  * Faster than getAreasOfNode().
  * The area name cannot be a node name, and areas are put in the nametree,
  * be convention under the root.
  * Areas may be included in areas, but the nametree does not manage this relation.
  * The visible space (see NameISTree) of an area node always includes this area.
  * @param _name of an area in the nametree
  * @return list of the area node names. May be empty. Never null.
  */
  public final List<ImmutableName> getL_NodesOfArea(ImmutableName _name) {
      List l_result = new ArrayList(INITIAL_CAPACITY);
      List l_nodes = (List) m_Areas_.get(_name);
      Object node = null;
      if (l_nodes == null) return l_result;
      for (Iterator ite = l_nodes.iterator(); ite.hasNext();) {
          node = ite.next();
          if (node instanceof String) continue;
          l_result.add( ((ImmutableName)node).clone() );
      } // ite
      l_nodes = null;
      return l_result;
  } // getAreasOfNode()
  
  /**
  * Gets the areas of the nametree the node belongs to.
  * @param _name of a node of the nametree
  * @return list of the area node names. May be empty.
  */
  public final List<ImmutableName> getL_AreasOfNode(ImmutableName _name) {
      List l_result = new ArrayList(INITIAL_CAPACITY);
      Map.Entry entry;
      for (Iterator ite = m_Areas_.entrySet().iterator(); ite.hasNext();) {
          entry = (Map.Entry) ite.next();
          
          if (((List)entry.getValue()).contains(_name)) {
              l_result.add( ((ImmutableName)entry.getKey()).clone() );
          } // if
      } // ite
      entry = null;
      return l_result;
  } // getAreasOfNode()
  
  /**
  * Gets the level of a node in the nametree. The root is at the level 0.
  * @param _node has to be a nametree node or area
  * @return the level of _node in the nametree, or -1 if not in the tree
  */
  public final int getLevel(ImmutableName _node) {
      if (isArea(_node)) return 1;
      else if (!isNode(_node)) return -1;
      else return ((Integer)m_Levels_.get(_node)).intValue();
  }
  
  /**
  * Gets the parent of a nametree node or area.
  * @param _child has to be a nametree node name, or an area name
  * @return the copy of parent in the nametree, or null if
  *  _child is null, empty, an area name, the root or it is not a nametree node
  */
  public final ImmutableName getParent(ImmutableName _child) {
      if (_child == null || _child.isEmpty() || _child.equals(getRoot())
         || (!isNode(_child) && !isArea(_child))) return null;

      else if (getL_AreaNames().contains(_child)) return getRoot();

      ImmutableName result;
      try {
            result = (ImmutableName) ((ImmutableName) getNodeArray(_child)[4]).clone();
      } catch (BaseError b) {
          // not possible
          throw new InternalError("in NameTreeImpl.getParent(): "+b.toString());
      } // catch
      return result;
  } // getParent()
  
  /**
  * Gets one of the children in the nametree
  * @param _parent is a node or area in the nametree
  * @param _index in the parent's child array
  * @return the copy of child at _index, or null if
  *   _parent is not a nametree node, not a parent, or if _index is not
  *  in the right values.
  */
  public final ImmutableName getChild(ImmutableName _parent, int _index) {
      try {
          if (getChildCount(_parent) - 1 < _index || _index < 0)
              return null;
      } catch (BaseError b) {
          return null;
      }

      if (isArea(_parent))
          return (ImmutableName) ((ImmutableName)
                  getL_NodesOfArea(_parent).get(_index)).clone();


      ImmutableName result = null;
      Object[] l_values;
      int search = 0;
      // searching of the child at its level, where some nodes may have
      // another parent(s), so _index is not over all the vector
      for (Iterator ite = l_NodeDescriptors_[((Integer)m_Levels_.get(_parent)).intValue() + 1].
        iterator(); ite.hasNext(); ) {
            l_values = (Object[]) ite.next();
            // search among _parent children only
            if (!_parent.equals( l_values[4] )) continue;
            if (search == _index) {
                result = (ImmutableName) ((ImmutableName)l_values[0]).clone();
                break;
            } // if
            else search++;
      } // for ite      
      l_values = null;
      return result;     
  } // getChild()
  
  /**
  * Gets the number of children for a node in the nametree,
  * in reading its proper children count.
  * Generally speaking, this method is faster than getChildren() which
  * searches children in the level above the parent level.
  * Call getNodeArray().
  * @param _parent is a node or an area in the nametree
  * @exception BaseError if the parent is null or does not belong to the nametree
  * @return the number of children for the parent
  */
  public final int getChildCount(ImmutableName _parent) throws BaseError {
      if (_parent == null || (!isNode(_parent) && !isArea(_parent)) ) throw new BaseError(
            "in NameTreeImpl.getChildCount()" +
            ", the parent is null or does not belong to the nametree.");
      if (isNode(_parent))
          return ((Integer) getNodeArray(_parent)[3]).intValue();
      else if (isArea(_parent))
          return getL_NodesOfArea(_parent).size();
      else return 0;
  } // getChildCount()

  /**
  * Gets all the direct children of a node or area in the nametree.
  * The order of the returned childen is the tree index order.
  * Called by getL_AllChildren().
  * @see #getL_AllChildren
  * @param _parent is a node or area or the root in the nametree
  * @return clones of ImmutableNames. Null if _parent is not a node
  *    of the nametree. Empty if _parent is a leaf.
  */
  public final List<ImmutableName> getL_DirectChildren(ImmutableName _parent) {
      if (isArea(_parent))
          return getL_NodesOfArea(_parent);

      if (!isNode(_parent)) return null;
      List<ImmutableName> l_result = new ArrayList(INITIAL_CAPACITY);
      
      // empty result if _parent is at the last level
      if (l_NodeDescriptors_.length == ((Integer)m_Levels_.get(_parent)).intValue() + 1 )
          return l_result;
      
      Object[] l_values;
      
      // search the children, and empty result if there is no child
      for (Iterator ite = l_NodeDescriptors_[((Integer)m_Levels_.get(_parent)).intValue() + 1].
        iterator(); ite.hasNext(); ) {
            l_values = (Object[]) ite.next();
            if (_parent.equals(l_values[4])) {
                l_result.add((ImmutableName) ((ImmutableName)l_values[0]).clone() );
            } // if
      } // for ite

      if (getRoot().equals(_parent)) l_result.addAll(getL_AreaNames());

                            // for testing and tracking
                        //System.out.println("NameTreeImpl.getL_DirectChildren() result"+
                          //      " for the parent "+_parent);
                        //for (ImmutableName name : l_result) {
                          //  System.out.println("--- "+name);
                        //}
      
      l_values = null;
      return l_result;       
  } // getL_DirectChildren()

  /**
  * Gets all the direct and indirect children of a node or area in the nametree.
  * The sorting of the returned childen is the decreasing order in levels, so
  * the direct children are the last items in the returned list. This order allows
  * to iterate on the list for removing the nodes, with the garantee that
  * each processed node is a leaf.
  * Calls getL_DirectChildren(). Called by NameISTreeImpl.isParentOfAcsNode(),
  * ISTree.createISTree().
  * @see #getL_DirectChildren
  * @param _parent is a node or area in the nametree
  * @return clones of ImmutableNames. Null if _parent is not a node
  *    of the nametree. Empty if _parent is a leaf.
  */
  public final List<ImmutableName> getL_AllChildren(ImmutableName _parent) {
      if (isArea(_parent))
          return getL_NodesOfArea(_parent);

      List<ImmutableName> l_result = getL_DirectChildren(_parent);
      if (l_result == null || l_result.isEmpty()) return l_result;
      
      // return if _parent is at the level before the last level
      if (l_NodeDescriptors_.length == ((Integer)m_Levels_.get(_parent)).intValue() + 2 )
          return l_result;

      // tp keep the order of getL_DirectChildren() in the final result
      Collections.reverse(l_result);

      List      l_children;
      List      l_current = new ArrayList(l_result);
      List      l_nextCurrent = new ArrayList(INITIAL_CAPACITY);

      // for the root, getL_DirectChildren(_parent) has also returned the areas
      // for which the children are nodes somewhere in the tree, so
      // no complementary search on the area children
      if (getRoot().equals(_parent)) l_current.removeAll(getL_AreaNames());
      
      // search the indirect children, if any
      while (!l_current.isEmpty()) {
          for (Iterator ite = l_current.iterator(); ite.hasNext(); ) {
                l_children = getL_DirectChildren((ImmutableName)ite.next());
                if (l_children == null || l_children.isEmpty()) continue;
                Collections.reverse(l_children);
                l_result.addAll(l_children);
                l_nextCurrent.addAll(l_children);
          } // for ite 
          
          l_current.clear();
          l_current.addAll(l_nextCurrent);
          l_nextCurrent.clear();
      } // while
      
      l_children = null;
      l_current = null;
      l_nextCurrent = null;
      
      Collections.reverse(l_result); 
      return l_result;
  } // getL_AllChildren()  
  
  /**
  * Gets the current index of a child in its parent array
  * in the nametree. Return -1 if _child is the root,
  * if either the child is null or without parent, or not in the nametree.
  * Caution: the convention is to put the areas as first children under the root,
  * before the nodes.
  * @param _child is the node or an area we are interrested in
  * @return the current index of the child in its level, or -1 if there is no parent
  */   
  public final int getIndexInParent(ImmutableName _child) {
      ImmutableName parent = getParent(_child);
      if (parent == null) return -1;

      if (isArea(_child))
          return getL_NodesOfArea(_child).indexOf(_child);

      Object[] l_values;
      int result = 0;
      int search = 0;
      
      // search the child
      for (Iterator ite = l_NodeDescriptors_[((Integer)m_Levels_.get(parent)).intValue() + 1].
        iterator(); ite.hasNext(); ) {
            l_values = (Object[]) ite.next();
             // search among _parent children only
            if (!parent.equals( l_values[4] )) continue;           
            if (_child.equals( l_values[0] )) {
                result = search;
                break;
            } // if
            else search++;
      } // for ite 
      
      l_values = null;
      parent = null;
      
      return result;      
  }  // getIndexInParent()
  
  /**
  * Gets the current index of a child in its nametree, as the child
  * appears in the IS structure o the GUI, following a depth-first order.
  * For the level n, the index contains the value (getIndexInParent() + 1)
  * for the ancestor of this level n. The number of digits into
  * the index is always the number of levels in the nametree.
  * '0' means there is no ancestor at this level.
  * The nametree root has the index '10...0'.
  * For instance, a typical suite of indexes this method should return,
  * in a sequence of calls for all the nodes of a nametree having 4 levels,
  * is: '1000', '1100', '1110', '1120', '1121', '1122', '1200', '1210', '1211'.
  * This is the suite for a nametree having under the root '1000', two nodes with
  * the indexes '1100' and '1200', 4 children and indirect children under '1100',
  * and 2 children and indirect children under '1200'.
  * Calls getTreePath() and getIndexInParent().
  * @param _child is the node or an area of the nametree
  * @return the current index of the child in the nametree, 
  * for traversing the tree in a depth-first order.
  * @exception BaseError if the arguments does not belong to the nametree,
  *   or if the nametree data is not consistent.
  */   
  public final int getDepthFirstIndexInNameTree(ImmutableName _child) throws BaseError {

                            // for testing and tracking
                        //System.out.println("call to NameTreeImpl.getDepthFirstIndexInNameTree()"+
                          //      " for "+_child);

      Object[]      l_ancestors = getTreePath(getRoot(), _child).getPath();
      ImmutableName parent = getParent(_child);
      StringBuffer  result = new StringBuffer("1");
      int           lev = nbOfLevels();
      int           nb = -1;

      if (parent == null) {
          while (lev > 1) {
              result.append("0");
              lev--;
          }
          l_ancestors = null;
          return Integer.valueOf(result.toString());
      } // if
      
      // add the path nodes
      for (int i = 1; i < l_ancestors.length; i++ ) {

                            // for testing and tracking
                        //System.out.println("NameTreeImpl.getDepthFirstIndexInNameTree() ancestor:");
                        //System.out.println("---- "+l_ancestors[i]);

          result.append(getIndexInParent((ImmutableName)l_ancestors[i]) + 1);
      } // for i

      if (lev != result.toString().length()) {
          nb = lev - result.toString().length();
          while (nb > 0) {
              result.append("0");
              nb--;
          }
      } // if
      
      l_ancestors = null;
      parent = null;
      return Integer.valueOf(result.toString());
  }  // getDepthFirstIndexInNameTree()
  
  /**
   * Gets the type of the node or the area.
   * @param _name of the node or area
   * @exception BaseError if _name is null or an empty name,
   *   or does not belong to the nametree.
   * @return the type for the node or area. May be null.
   */
   public String getType(ImmutableName _name) throws BaseError {
      if (_name == null || _name.isEmpty() ||
              (!isNode(_name) && !isArea(_name)))
          throw new BaseError("in NameTreeImpl.getType()" +
              ", the node is null, empty, or does not belong to the nametree."); 
      if (isNode(_name))
          return (String) getNodeArray(_name)[1];
      else
          return (String) ((List)m_Areas_.get(_name)).get(0);

   } // getType()
 
  /**
   * Tests if a node of the nametree is set as a 'hub' for its children.
   * A node may be a hub, which means that for this node,
   * any direct child may have exchanges with any other direct child. 
   * For a non-hub node, exchanges between its children are not allowed.
   * The visible space (see NameISTree) of a node having a hub parent always 
   * includes all the children of its parent.
   * An area is always a hub node.
   * @exception BaseError if _node is null or an empty name,
   *   or does not belong to the nametree.
   * @return true if the node is a hub.
   */
   public boolean isHubNode(ImmutableName _node) throws BaseError {
       if (isArea(_node)) return true;
       if (_node == null || _node.isEmpty() || !isNode(_node))
          throw new BaseError("in NameTreeImpl.isHubNode()" +
              ", the node is null, empty, or does not belong to the nametree."); 
       return ((Boolean) getNodeArray(_node)[2]).booleanValue();
   } // isHubNode() 
   
  /**
  * Adds a child as a new node in the nametree. Not to use for an area.
  * Overridden by NameISTreeImpl.
  * @param _parent is the parent node of the child in the nametree. Never null.
  * @param _child is the unique name of the new child
  * @param _node_type is a value of the type. May be null.
  * @param _is_hub is true if the child is a node hub, which means that for this node,
  *    any direct child may have exchanges with any other direct child, if the rights allow them.
  * @exception BaseError if _parent is not a node of this nametree,
  *  if _child is null, an area name, is already a nametree node, or
  *  contains a character that is not a letter
  */
  protected final synchronized void addChild(ImmutableName _parent, ImmutableName _child,
    String _node_type, boolean _is_hub) throws BaseError {
        if (!isNode(_parent))
            throw new BaseError("in NameTreeImpl.addChild()" +
            ", the parent does not belong to the nametree.");
        
        if (_child == null || isNode(_child) || isArea(_child)) 
            throw new BaseError(
                "in NameTreeImpl.addChild()" +
                ", the child is null, an area name, or belongs to the nametree.");

        // character control in _child
        char[] l_chars;
        boolean containsFaultyValue = false;
        for (int j = 0; j < _child.noOfComponents(); j++) {
          l_chars = _child.component(j).toCharArray();
          for (int i = 0; i < l_chars.length; i++) {
            if (Character.isIdentifierIgnorable(l_chars[i]) ||
             Character.isISOControl(l_chars[i])) {
                  containsFaultyValue = true;
                  System.out.println("in NameTreeImpl.addChild(), a character "+
                        "is ignorable or is an ISO control:"+l_chars[i]+"end.");
                  break;
            }
          } // for i
          if (containsFaultyValue) break;
        } // for j
        if (containsFaultyValue) {
        l_chars = null;
        throw new BaseError("The name of the new child contains a character "+
          "that is not a letter.");
        } //if
        l_chars = null;

        Object[] l_values = new Object[5];
        l_values[0] = (ImmutableName)_child.clone();
        l_values[1] = _node_type;
        l_values[2] = new Boolean(_is_hub);
        l_values[3] = new Integer(0); // nb of children
        l_values[4] = (ImmutableName)_parent.clone();
   
        // incrementes the _parent children count
        Object[]    parentArray = getNodeArray(_parent);
        Integer     nbOfChildren = (Integer) parentArray[3];

        parentArray[3] = new Integer(nbOfChildren.intValue() + 1);
      
        // adds a level if necessary to l_NodeDescriptors_
        if (nbOfLevels() == getLevel(_parent) + 1) {
            ArrayList[] l_v = new ArrayList[l_NodeDescriptors_.length + 1];
            System.arraycopy(l_NodeDescriptors_, 0, l_v, 0, l_NodeDescriptors_.length);
            l_v[l_NodeDescriptors_.length] = new ArrayList(INITIAL_CAPACITY);
            l_NodeDescriptors_ = l_v;
            
            l_NodeDescriptors_[l_NodeDescriptors_.length - 1].add(l_values);
            m_Levels_.put(
                    (ImmutableName)_child.clone(), new Integer(l_NodeDescriptors_.length - 1));
            
            l_v = null;
        } // if
        // or adds l_values directly to l_vectors
        else {
            int level = getLevel(_parent) + 1;
            l_NodeDescriptors_[level].add(l_values);
            m_Levels_.put(
                    (ImmutableName)_child.clone(), new Integer(level));
        } // else

        parentArray = null;
        nbOfChildren = null;
        l_values = null;
  } // addChild()
  
  /**
  * Removes a node from the nametree (excepting the root),
  * and removes all the subtree which may be associated to the node.
  * Does not remove an area. Removes also the node in an ImmutableNameISTree, but
  * does not remove the NameISTree.links where the keys are not strictly
  * the node name. Then, these specific links should be removed independantly.
  * @param _node is a node of the nametree
  * @exception BaseError if _node is null, or not in the nametree, or is the root
  * @return the number of removed nodes, including _node
  */
  protected final synchronized int removeNodeAndSubTree(ImmutableName _node)
          throws BaseError {
        if (_node == null || !isNode(_node) || getRoot().equals(_node))
          throw new BaseError(
            "in NameTreeImpl.removeNodeAndSubTree()" +
            ", the node is null, the root, or does not belong to the nametree.");
        
        int result = 1;
        
        if (getChildCount(_node) == 0 ) {
            removeLeaf(_node);
            return result;
        } // if
        
        ImmutableName node;
        List l_children;
        result = 0;
        
        // sets the list of the current _node indirect children, then
        // iteration to remove the leaves and to update the children count in the parents,
        // until _node is a leaf
        while (getChildCount(_node) != 0) {
            l_children = getL_AllChildren(_node);
            for (Iterator ite = l_children.iterator(); ite.hasNext(); ) {
                node = (ImmutableName)ite.next();
                if (getChildCount(node) == 0) {
                    removeLeaf(node);
                    result++;
                }
            } // for ite
        } // while
        
        removeLeaf(_node);
        node = null;
        l_children = null;
        return result++;
  } // removeNodeAndSubTree()
  
  /**
  * Adds an empty area in the nametree, which is defined as a set of nodes in the nametree.
  * These implied nodes may be in any locations. <P>
  * Areas are used to define virtual structures like a composite ACS which contains several
  * pieces of software running on different servers, and defined in the nametree
  * as nodes. By convention, an area is put under the root of the nametree. <P>
  * The area name cannot be a node name, and areas are not explicitly in the nametree.
  * Nevertheless, they are often put graphically under the nametreee root.
  * Areas may be included in areas, but the nametree does not manage this relation.
  * The extended visible space (see NameISTree) of an area node is always ALL_THE_IS_SPACE.
  * @param _area the area name.
  * @param _node_type a value of the type. May be null.
  * @exception BaseError if _area is already an area name or a node name, or if
   * it contains a character that is not a letter
  */
  protected final synchronized void addArea(ImmutableName _area, String _node_type)
    throws BaseError {
        if (_area == null)
            throw new BaseError("in NameTreeImpl.addArea()" +
               ", the area name is null.");

        // character control in _area
        char[] l_chars;
        boolean containsFaultyValue = false;
        for (int j = 0; j < _area.noOfComponents(); j++) {
          l_chars = _area.component(j).toCharArray();
          for (int i = 0; i < l_chars.length; i++) {
            if (Character.isIdentifierIgnorable(l_chars[i]) ||
             Character.isISOControl(l_chars[i])) {
                  containsFaultyValue = true;
                  System.out.println("in NameTreeImpl.addArea(), a character "+
                        "is ignorable or is an ISO control:"+l_chars[i]+"end.");
                  break;
            }
          } // for i
          if (containsFaultyValue) break;
        } // for j
        if (containsFaultyValue) {
        l_chars = null;
        throw new BaseError("The name of the new area contains a character "+
          "that is not a letter.");
        } //if
        l_chars = null;

        if (m_Areas_.containsKey(_area) )
            throw new BaseError("in NameTreeImpl.addArea()" +
               ", the area name is known.");
        if (m_Levels_.containsKey(_area))
            throw new BaseError("in NameTreeImpl.addArea()" +
               ", the area name is known as node.");
             
        List<String> l_area = new ArrayList(INITIAL_CAPACITY);

        l_area.add(_node_type);
        m_Areas_.put((ImmutableName)_area.clone(), l_area);

        l_area = null;
  } // addArea() 
  
  /**
  * Adds a node to an area in the nametree, only if the new node is a nametree
  * node which is a direct child of an area external node.
  * @param _area is the area name
  * @param _node of the nametree to add to the area
  * @exception BaseError if _node is not a node name of this nametree,
  *   if _area is not an area name, if _node
  *   is not a direct child of an area external node.
  */
  protected final synchronized void addNodeToArea(ImmutableName _area, ImmutableName _node)
    throws BaseError {       
        if (!m_Areas_.containsKey(_area) || !m_Levels_.containsKey(_node))
            throw new BaseError("in NameTreeImpl.addNodeToArea()" +
               ", the area or the node is unknown.");
        
        ((List)m_Areas_.get(_area)).add((ImmutableName)_node.clone());
  } // addNodeToArea() 
  
  /**
  * Removes a node in an area in the nametree.
  * @param _area is an area name
  * @param _node of the nametree to remove to the area
  * @exception BaseError if _areaName is not an area name
  */
  protected final synchronized void removeNodeToArea(ImmutableName _area, ImmutableName _node)
          throws BaseError {
        boolean error = false;        
        if (!m_Areas_.containsKey(_area) )
            throw new BaseError("in NameTreeImpl.removeNodeToArea()" +
               ", the area name is unknown.");          

        if (!((List)m_Areas_.get(_area)).contains(_node))
            throw new BaseError("in NameTreeImpl.removeNodeToArea()" +
               ", the node name is unknown.");
        ((List)m_Areas_.get(_area)).remove(_node);
  } // removeNodeToArea()

  /**
  * Removes an area in the nametree.
  * @param _area is an area name
  * @exception BaseError if _area is not an area name
  */
  protected synchronized void removeArea(ImmutableName _area) throws BaseError {
        boolean error = false;        
        if (!m_Areas_.containsKey(_area) )
            throw new BaseError("in NameTreeImpl.removeArea()" +
               ", the area name is unknown.");          

        ((List)m_Areas_.get(_area)).clear();
        m_Areas_.remove(_area);       
  } // removeArea() 
  
  /**
  * Gets the path of nodes between two nodes in the nametree,
  * if one of the two nodes is an indirect child of the other one.
  * This method is faster than getSubTree(), and calls getNodeArray().
  * @param _startingNode is a nametree node. Never an area.
  * @param _endingNode is a nametree node which is at a lower or a greater level
  *   than _startingNode. Never an area.
  * @exception BaseError if one of the two arguments does not belong to the nametree,
  *   or if the nametree data is not consistent.
  * @return the TreePath of node names, including the two arguments, and the first one
  *   (index 0) is the node in the lower nametree level. Null if there is no tree path.
  *   Contains only _startingNode if the two nodes are equals.
  */  
  public final TreePath getTreePath(ImmutableName _startingNode, ImmutableName _endingNode)
    throws BaseError {
        if (!isNode(_startingNode) || !isNode(_endingNode))
            throw new BaseError("in NameTreeImpl.getTreePath()" +
               ", one of the arguments does not belong to the nametree.");
        if ( _startingNode.equals(_endingNode) )
            return new TreePath(new ImmutableName[]{(ImmutableName)_startingNode.clone()});
        
        int startingLevel = getLevel(_startingNode);
        int endingLevel = getLevel(_endingNode);
        if ( startingLevel == endingLevel ) return null;

                // for testing and tracking
            //System.out.println("NameTreeImpl.getTreePath(), startingLevel = "+
              //      startingLevel+"; endingLevel = "+endingLevel);

        boolean         error = false;
        ImmutableName[] l_nodes;
        ImmutableName   currentNode;
        ImmutableName   lowerNode;
        TreePath        l_result = null;
        // iterate from the last level in the treepath
        int             currentLevel;
        // lower level in the resulting treepath
        int             lowerLevel;
        // index in l_nodes
        int             index;
        
        if ( startingLevel < endingLevel ) {
            l_nodes = new ImmutableName[endingLevel - startingLevel + 1];
            currentNode = _endingNode;
            lowerNode = _startingNode;
            currentLevel = endingLevel;
            lowerLevel = startingLevel;
            index = endingLevel - startingLevel;
        } else {
            l_nodes = new ImmutableName[startingLevel - endingLevel + 1];
            currentNode = _startingNode;
            lowerNode = _endingNode;
            currentLevel = startingLevel;
            lowerLevel = endingLevel;
            index = startingLevel - endingLevel;
        } // else
        
        // l_nodes is initialized from the upper index
        // since getParent() is the faster way to go accross the treepath
        while (currentLevel != lowerLevel - 1) {
            l_nodes[index] = (ImmutableName) currentNode.clone();
            
            // iteration up to the parent
            // compares to the root
            if (currentNode.equals( ((Object[]) l_NodeDescriptors_[0].get(0))[0] ))
                currentNode = null;
            // else, gets the parent
            else currentNode = (ImmutableName) getNodeArray(currentNode)[4];
            currentLevel--;
            index--;
        } // while
        
        if (index != -1) {

                // for testing and tracking
            //System.out.println("NameTreeImpl.getTreePath(), with index != -1 - l_nodes:");
            //for (int i = 0; i < l_nodes.length; i++) {
              //  System.out.println("--------- "+l_nodes[i]);
            //} // for i

            throw new InternalError("in NameTreeImpl.getTreePath()" +
               ", the search is not correct.");
        }
        
        // null result if lowerNode is not an indirect parent of the second argument
        if (!l_nodes[0].equals(lowerNode) ) error = true;
        
        if (!error) l_result = new TreePath(l_nodes);
        else {

                // for testing and tracking
            //System.out.println("NameTreeImpl.getTreePath(), with error true - l_nodes:");
            //for (int i = 0; i < l_nodes.length; i++) {
              //  System.out.println("--------- "+l_nodes[i]);
            //} // for i

        } // else


                // for testing and tracking
            //System.out.println("NameTreeImpl.getTreePath(), with error false - l_nodes:");
            //for (int i = 0; i < l_nodes.length; i++) {
              //  System.out.println("--------- "+l_nodes[i]);
            //} // for i

        l_nodes = null;
        currentNode = null;  
        lowerNode = null;
        return l_result;
  } // getTreePath()
  
  /**
  * Tests if a node is a leaf of the nametree, in reading its proper children count.
  * Generally speaking, this method is faster than getChildren() which
  * searches children in the level above the parent level.
  * Call getChildCount().
  * @param _node is a node or an area to test
  * @exception BaseError if the node is null or does not belong to the nametree
  * @return true if _node is a leaf, without any child
  */  
  public final boolean isLeaf(ImmutableName _node) throws BaseError {
      int children_count = getChildCount(_node);
      if (children_count == 0) return true;
      else return false;
  } // isLeaf()
  
  /**
  * @return a cloned and no-backed instance of NameTree, with clone of all names.
  */
  public Object clone() {
      try {
          NameTreeImpl clon = (NameTreeImpl) super.clone();
          
          ImmutableName key;
          Object[] l_root;
          Object[] l_obj;
          List  list;
          
          // full clone of all the names in the arrays
          if (l_NodeDescriptors_ != null) {
              
              /**
               * array of lists where the index is the level.
               * in each list, the value is an an array Object[5] for each node description:
               *    0: node name (it is the key), 1: node type, 2: is_hub,
               *    3: nb_of_children, 4: parent name
               * the nametree root is at the index 0, as a single item in the list
               */
              l_root = getNodeArray(getRoot());

              l_root[0] = ((ImmutableName)l_root[0]).clone();
              //l_root[4] = ((ImmutableName)l_root[4]).clone();

              clon.l_NodeDescriptors_ = new ArrayList[1];
              clon.l_NodeDescriptors_[0] = new ArrayList(INITIAL_CAPACITY);
              clon.l_NodeDescriptors_[0].add(l_root); 
              
              // The root level is already processed, and the nodes are added
              // after their parent
              for (int i = 1; i < l_NodeDescriptors_.length; i++) {  
                for (Iterator ite = l_NodeDescriptors_[i].iterator(); ite.hasNext(); ) {
                    l_obj = (Object[])ite.next();
                    clon.addChild(
                            (ImmutableName)((ImmutableName)l_obj[4]).clone(), 
                            (ImmutableName)((ImmutableName)l_obj[0]).clone(), 
                            (String)l_obj[1], 
                            ((Boolean)l_obj[2]).booleanValue());

                } // for ite
              } // for i
          } // if (l_NodeDescriptors_ != null)
          
          if (m_Levels_ != null) {
            clon.m_Levels_ = new HashMap(INITIAL_CAPACITY);

            for (Iterator ite = m_Levels_.keySet().iterator(); ite.hasNext(); ) {
                key = (ImmutableName)ite.next();
                clon.m_Levels_.put(
                  (ImmutableName)key.clone(),
                  new Integer( ((Integer)m_Levels_.get(key)).intValue() )
                );
            } // for ite
          } // if
          
          clon.m_Areas_ = new HashMap(INITIAL_CAPACITY);
          for (Iterator ite = m_Areas_.keySet().iterator(); ite.hasNext(); ) {
              key = (ImmutableName)ite.next();
              list = new ArrayList(INITIAL_CAPACITY);
              for (Object obj : ((List) m_Areas_.get(key))) {
                  if (obj instanceof String)
                      list.add(obj);
                  else list.add(((ImmutableName)obj).clone());
              } // for obj
              clon.m_Areas_.put((ImmutableName)key.clone(), list);
          } // for ite

          key = null;
          l_root = null;
          l_obj = null;
          key = null;
          
          return clon;

      } catch (CloneNotSupportedException e) {
            throw new InternalError(e.toString());      
      } catch (BaseError b) {
            throw new InternalError(b.toString());
      } // catch 
  } // clone()  
  
  /**
   * This method compares the node and area lists, not the parent-child relations
   * nor the area elements.
  * @param _obj is the object to compare
  * @return true if the lists of node names and area names are equal
  */
  public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      NameTreeImpl obj = (NameTreeImpl) _obj;
      
      if (m_Levels_ == null || obj.m_Levels_ == null) {
          if (m_Levels_ != null || obj.m_Levels_ != null)
            return false;  // if
      } // if
      else if (!m_Levels_.keySet().equals( obj.m_Levels_.keySet() ))
          return false;

      if (m_Areas_ == null || obj.m_Areas_ == null) {
          if (m_Areas_ != null || obj.m_Areas_ != null)
            return false;  // if
      } // if
      else if (!m_Areas_.keySet().equals( obj.m_Areas_.keySet() ))
          return false;
      
      return true;
  } // equals
  
  /**
  * @return the hashcode value from the root and the number of nodes.
  */
  public int hashCode() {
      if (l_NodeDescriptors_ == null)
          return 100001;
      else
          return getRoot().hashCode() + 77711*nbOfNodesAndAreas();
  }   

  /**
  * @return long serial version UID for serialization of the class
  */
  public long getSerialVersionUID() {
       return serialVersionUID;
  }    

  /**
   * This method nullify all the nameTree inner properties.
   */
  protected synchronized void finalizeForUser() {
    l_NodeDescriptors_ = null;
    m_Levels_.clear();
    m_Levels_ = null;
    m_Areas_.clear();
    m_Areas_ = null;
  } // finalizeForUser()

  /**
   * This method calls finalizeForUser().
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    finalizeForUser();
    super.finalize();
  } // finalize()
  
   //
   // ----------- protected methods ----------------------------
   //
   
  /**
   * Remove the internal data for a leaf in the nametree,
   * with an update of the parent, the nametree internal data, the areas,
   * but no removing of the node children.
   * The root cannot be removed.
   * Overridden method in NameISTreeImpl.
   * @param _node is a node of the nametree, not an area.
   * @exception BaseError if _node is the root, or does not belong to the nametree,
   *   or is not a leaf, or if the base is not consistent.
   */
   protected synchronized void removeLeaf(ImmutableName _node) throws BaseError {
      if (getChildCount(_node) != 0 || _node.equals(getRoot()) ) throw new BaseError(
        "in NameTreeImpl.removeLeaf()" +
        ", the node is not a leaf or is the nametree root.");
       
      // decrementes the _parent children count
      Object[] parentArray = getNodeArray( (ImmutableName) getNodeArray(_node)[4] );
      Integer nbOfChildren = (Integer) parentArray[3];
      parentArray[3] = new Integer(nbOfChildren.intValue() - 1);
      parentArray = null;
      nbOfChildren = null;
      
      boolean error = true;
      Object[] l_values;
      
      // searching for the node to remove, and nullify the relevant values
      for (Iterator ite = l_NodeDescriptors_[((Integer)m_Levels_.get(_node)).intValue() ].
        iterator(); ite.hasNext(); ) {
            l_values = (Object[]) ite.next();
            if (_node.equals(l_values[0])) {
                for (int i = 0; i < l_values.length; i++) {
                    l_values[i] = null;
                } // for i
                ite.remove();
                error = false;
                break;
            } // if
      } // for ite      
      l_values = null;
      if (error) throw new BaseError("in NameTreeImpl.removeLeaf()" +
        ", the node does not belong to the vector array.");
      
      // if _node is the last node of its level, it is then the upper level,
      // and the level is removing from l_vectors
      if (l_NodeDescriptors_[((Integer)m_Levels_.get(_node)).intValue() ].isEmpty()) {
          
          if (l_NodeDescriptors_.length != ((Integer)m_Levels_.get(_node)).intValue() + 1)
              throw new InternalError("in NameTreeImpl.removeLeaf()," +
                " the levels in l_NodeDescriptors_ are not consistent.");
          
          ArrayList[] l_v = new ArrayList[l_NodeDescriptors_.length - 1];
          System.arraycopy(l_NodeDescriptors_, 0, l_v, 0, l_v.length);
          l_NodeDescriptors_ = l_v; 
          l_v = null;
      } // if    

      m_Levels_.remove(_node);
      
      // searching for the node in the areas
      List l_areas = getL_AreasOfNode(_node);
      for (Iterator ite = l_areas.iterator(); ite.hasNext(); ) {
          ((List)m_Areas_.get(ite.next() )).remove(_node);
      } // for
      l_areas = null;
      
      return;        
   } // removeLeaf()  

   //
   // ------------ private methods ----------------------------
   //
  
  /**
   * Gets the internal description of a node in the nametree.
   * Caution: no control on _node, which must be a nametree node, and not an area.
   * @param _node is the node we are interrested in. Never an area.
   * @exception BaseError if the node does not belong to this nametree.
   * @return a link to the internal data array where <P>
   *    0: node name (it is the key), 1: node type (may be null), 2: is_hub,<P>
   *    3: nb_of_children, 4: parent name
   */
   private Object[] getNodeArray(ImmutableName _node) throws BaseError {
      
      Object[] result = null;
      boolean error = true;
      Object[] l_values;
      // searching for the node
      for (Iterator ite = l_NodeDescriptors_[((Integer)m_Levels_.get(_node)).intValue() ].
        iterator(); ite.hasNext(); ) {
            l_values = (Object[]) ite.next();
            if (_node.equals(l_values[0])) {
                result = l_values;
                error = false;
                break;
            } // if
      } // for ite      
      l_values = null;
      if (error) throw new BaseError("in NameTreeImpl.getNodeArray()" +
        ", the node does not belong to the vector array.");
      
      return result;        
   } // getNodeArray()
  
   
} // NameTreeImpl class