//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;


import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeSupport;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrW.DialogError;

/**
* This class is responsible for managing the No-More-Than and No-Less-Than views.
* The aim is to check if some Eligible Parties (or Sources) has rights on the access target
* above a given 'no-more' right and below a given 'no-less' right.
* As any view, the result depends on all the opened ACSs. <P>
* The 'no-less' right criterion is generally used to check the compliance to the access needs.
* The 'no-more' right criterion is generally used to check the compliance to the access security policy.
* This view handles one target, a set of basic rights users as standard EPRViewInBase objects,
* another set of excluded Sources which belong or run under one of the
* basic right user, a StringRight 'no-more' criterion and a StringRight 'no-less' criterion. <P>
* The checking perimeter includes the view sources and the listening perimeter.
* The associates of a source are defined as their direct/indirect members
* and their direct/indirect running-under Actors.
* This listening perimeter is get by the subtraction of excluded sources and their associates
* from the view Sources list (also called 'basic rights users') and their associates.
* For instance, if the group A is a view Source, and has only group B and group C
* has direct members, and if the group C is a view excluded Source,
* the listening perimeter contains only the group B and its associates. <P>
* The number of basic right users is limited to 9, like the reduced users,
* and the number of sources to check in the checking perimeter is limited to 50.
* Caution: every Source is indeed a Target, since a GroupDMember for example
* is the target of the RUN_UNDER relation.
* The two right criteria have to be positive. It is not possible to have
* a less criterion greater than the more criterion, but they may be equal or equivallent.
* The comparison between a criterion and the effective rights of the checking perimeter
* uses the comparison of the respective metarights, even if they are ACS rights.
* It is possible to have as criterion an administrative right and a non-administrative right,
* or criteria mixing a metaright and an acsright, or two ACS rights from different ACS. <P>
* This is a javabean with the following bound properties:<P>
* - 'OneExcludedEP' when an excluded Source is added or removed, <P>
* - 'FirstAccessTarget' when an access target is added or removed <P> 
* - 'OneLessCriterion' when a less criterion is added or removed,<P>
* - 'OneMoreCriterion', <P>
* - 'CheckingPerimeter', <P>
* - 'ReducedEPsInViewEPs', <P>
* - 'ObjectChangesInListeningPerimeter', <P>
* - 'OnePropertyInAListeningPerimeterObject'. <P>
* The updating of the Target resets the two criteria, since they have a strong relation
* with the Target.
* The adding and the removing of the Target and the excluded sources is processed
* by Gui2.GraphicViewBaseListenerImpl almost like the adding or removing of
* an EPRViewInBase object, to refresh the draw of the view nodes
* and to analyze the new rights. The only difference is the alias relation handling,
* since the EPRViewInBase adds automatically in the view the reference of an alias
* and the alias of a reference, while this is not done in a NoThanViewInBase.
* The bound properties of the target are listened
* and processed like EPRViewInBase does for its standard objects. 
* The bound properties of the excluded Sources have not to be listened.
* When a source is a GroupID containing GroupIDs, the bound property 'l_Members'
* of each group in the members tree has to be listened, and also the properties
* for the actors 'l_GroupIDCurrentActors' and 'l_GroupIDCurrentActorsAsSecondaryGroup'
* in each of these indirect members, if these objects are not excluded sources.
* @version 0.7.0
* @see ARoad0.Gui2.GraphicNoThanView
*/
public class NoThanViewInBaseImpl extends EPRViewInBaseImpl implements NoThanViewInBase {
  /* protected attributes */
    /*
     * Map<String,StringRight[]> for 'more single' and 'more team', 
     * with the key Name.toString() for each target. For this class version,
     * there is only one key and one right as value for this key.
     * Used by getMoreCriterion() and getEorFirstAccessTarget(). May be empty.
     */
  protected Map<String,StringRight[]>     m_MoreCriteria_ = new HashMap(INITIAL_CAPACITY_5);
    /*
     * Map<String,StringRight[]> for 'less single', 
     * with the key Name.toString() for each target. For this class version,
     * there is only one key and one right as value for this key. 
     * Used by getLessCriterion() and getEorFirstAccessTarget(). May be empty.
     */
  protected Map<String,StringRight[]>     m_LessCriteria_ = new HashMap(INITIAL_CAPACITY_5);
  /*
   * List of excluded Source (currently, Eligible Party) names which are connected
   * to some view sources, but have not to be checked since they are explicitly
   * put away by the user. Used to fill l_excludedSources_ by prepareAfterRestoring().
   */
  protected List<ImmutableName>           l_excludedSourceNames_ =
          new ArrayList(INITIAL_CAPACITY_5);
  /*
   * Transient list of excluded Sources which are connected to some view sources, but
   * have not to be checked since they are explicitly put away by the user. Filled
   * by prepareAfterRestoring().
   */
  protected transient List<ImmutableSource> l_excludedSources_ = new ArrayList(INITIAL_CAPACITY_5);
  /*
   * List of Sources (currently Eligible Parties) which are in the perimeter of checking.
   * Gets by the sustraction of l_excludedSources_ from the view EPs list and all
   * their direct or indirect members, and their direct or indirect running-under actors.
   * Filled by setUpContextLists().
   */
  protected transient List<ImmutableSource> l_checkingPerimeter_ = new ArrayList(INITIAL_CAPACITY);
  /*
   * List of Sources (currently Eligible Parties) from the view sources list
   * which have at most a connected member or a running-under actor
   * being an excluded source. Filled by setUpContextLists().
   */
  protected transient List<ImmutableSource> l_reducedSourcesInViewSources_ =
          new ArrayList(INITIAL_CAPACITY_5);
  
  /* public attributes */
  public static final NoThanViewInBase    EMPTY_INSTANCE = new NoThanViewInBaseImpl();

  /*
   * The super class has no serialPersistentFields property, but ViewInBaseImpl does.
   * variable for the JDK 2 serialization
   * @serial ("m_MoreCriteria_",HashMap.class), ("m_LessCriteria_",HashMap.class)
   * ("l_excludedSourceNames_",ArrayList.class),
   */ 
    public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("m_MoreCriteria_",HashMap.class),
      new ObjectStreamField("m_LessCriteria_",HashMap.class),
      new ObjectStreamField("l_excludedSourceNames_",ArrayList.class)
  };

 /**
  * Default constructor
  */
  public NoThanViewInBaseImpl() {
    super();
    type_ = "No-More-Than/No-Less-Than";
  }

  /**
   * @param _name of the view. it is copied.
   * @exception CreateError if the view name is already known in BaseManagerImpl.
   */
  public NoThanViewInBaseImpl(ImmutableName _name) throws CreateError {
      super(_name);
      type_ = "No-More-Than/No-Less-Than";
  }

  /**
   * This constructor is used by BaseManagerImpl.openView() to create
   * every instance. So, all the objects to listen (ACS...) do receive a registration
   * from this instance, and that would not be the case in a simple deserialization.
   * Calls the super constructor with the same arguments, and calls addOwnListening(true, true).
   * @param _name of the view. It is copied.
   * @param _m_BaseObjects is the map (name, object) of the view objects, without
   *  no-Actor Resource(s). It is copied. The maximal size is 9. May be null.
   * @param _toRegisterInManager false if this constructor must not register
   *  the instance in BaseManagerImpl and in its view objects.
   * @exception CreateError if the constructor of EPRViewInBaseImpl fires it, or
   * if _m_BaseObjects contains an object that is not a source.
   */
  public NoThanViewInBaseImpl(ImmutableName _name, Map _m_BaseObjects,
    boolean _toRegisterInManager) throws CreateError {

      // 'super' may complements the list of BaseObjects
      super(_name, _m_BaseObjects, false);
      type_ = "No-More-Than/No-Less-Than";
      
      // controls _m_BaseObjects has a no-Source Target
      if (_m_BaseObjects != null && !m_NoSourceTargets_.isEmpty()) {
          this.finalizeForCopy();  // targets and EP are not unlistened
          throw new CreateError("The view objects map contains an access target "+
                  "which is not an access source");
      } // if (_m_BaseObjects != null
      // controls if _m_BaseObjects has more than 9 elements
      if (_m_BaseObjects != null && m_BaseObjects_.size() > 9) {
          this.finalizeForCopy();  // targets and EP are not unlistened
          throw new CreateError("The view objects map contains more than 9 elements");
      } // if (_m_BaseObjects != null
      
      // adds to the map of Views in BaseManagerImpl after the variables init.
      // throws BaseError if already exists.
      if (_toRegisterInManager) {
        try {
            BaseManagerImpl.getInstance().newEorView(this);
        } catch (CreateError c) {
            this.finalizeForCopy();  // resources and EP are not unlistened
            throw c;
        }
        setUpContextLists();
        // registering by the two super classes, but the access target is still null
        addOwnListening(true, true);
      } // if
      
  } // constructor

  //
  // ----------- Methods of the implemented interfaces ----------------
  //  
    
  /**
   * This short name is NOT unique for the Access Road program nor the ACS if
   * it is an ACS object.
   * It is often the last component name, but not for a NoThanViewInBase where it is
   * the first and the last components, with ':: ' as separator. Overrides ViewInBaseImpl.
   * @return the nick name. May be null.
   */
    public String getNickName() {
       if (name_ == null) 
           return null;
       else return name_.component(name_.noOfComponents() - 3) + ":: " +
                   name_.component(name_.noOfComponents() - 1);
    } // getNickName()

  /**
   * This view handles exactly one access target.
   * @return the target. May be null only before the first updating
   */
  public ImmutableTarget getEorFirstAccessTarget() {
      if (m_MoreCriteria_.isEmpty() ) return null;
      Name name = null;
      ImmutableTarget result = null;
      try {
          name = NameImpl.getNameFromString((String)m_MoreCriteria_.keySet().iterator().next());
      } catch (BaseError b) {
          throw new InternalError("NoThanViewInBaseImpl.getEorFirstAccessTarget(): "+b.toString());
      }
      try {
          result = (ImmutableTarget)BaseUtilityImpl.getEorBaseObject(name);
      } catch (BaseError b) {
          // the target has been deleted, or the target acs
          result = null;
      }
      name = null;
      return result;
  } // getEorFirstAccessTarget()
  
  /**
   * This view handles exactly one access target. The More and Less criteria are set to null.
   * Calls addOwnListeningOnTarget(). Fires the 'FirstAccessTarget' property change.
   * @param _targ the target to add. Never null, so a current target cannot be removed.
   * @exception BaseError if the target is also a registered source in this view,
   * if the number of targets is too large,
   * if the argument is null, or the object name is null or empty.
   */
  public void setEorFirstAccessTarget(ImmutableTarget _targ) throws BaseError {
      if (_targ == null || _targ.getName() == null || _targ.getName().isEmpty())
          throw new BaseError("NoThanViewInBaseImpl.setEorFirstAccessTarget(): "+
                  "the argument or its name is null or empty");
      if (containsBaseObject(_targ))
          throw new BaseError(" The argument is already registered as souce in this view");
      
      ImmutableTarget old = getEorFirstAccessTarget();
      if (old != null) {
          removeOwnListeningOnTarget();
          setMoreCriterion(null);
          setLessCriterion(null);
      } // if
      
      m_MoreCriteria_.put(_targ.getName().toString(), null);
      m_LessCriteria_.put(_targ.getName().toString(), null);
      // false since the super classes have already been called through the coonstructor
      addOwnListeningOnTarget();

    try {
        firePropertyChange("FirstAccessTarget", old, _targ);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("NoThanViewInBaseImpl.ImmutableTarget(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }  
  } // setEorFirstAccessTarget()
  
  /**
   * NO USE SO FAR.
   * This method handles exactly one target. If there is already a target,
   * the More and Less criteria are set to null.
   * Overrides the superclass method, and do not call BasicImpl.addInViewList()
   * or VirtualFolderImpl.addInViewList(), nor FolderAbst.addEorBaseObject(),
   * like the superclass does it.
   * Calls addOwnListeningOnTarget(). Fires the 'FirstAccessTarget' property change.
   * @param _targ the target to add
   * @exception BaseError if the resource is also a registered source in this view,
   * if the number of targets is too large,
   * if the argument is null, or the object name is null or empty.
   */
public void addEorNoThanAccessTarget(ImmutableTarget _targ) throws BaseError {
      // for finalized instances that are still in the memory stack
    if (m_BaseObjects_ == null) return;
    if (!m_MoreCriteria_.isEmpty())
          throw new BaseError(" There is already a target");
      setEorFirstAccessTarget(_targ);
  } // addEorNoThanAccessTarget()

  /**
   * This method handles exactly one target. The More and Less criteria are set to null.
   * Called by propertyChange().
   * Fires the 'FirstAccessTarget' property change.
   * @param _targ the target to remove
   * @exception BaseError if the argument is not the current target,
   * if the argument is null, or the object name is null or empty.
   */
  public void removeEorNoThanAccessTarget(ImmutableTarget _targ) throws BaseError {
      if (m_MoreCriteria_.isEmpty() )
          throw new BaseError("NoThanViewInBase.removeEorAccessTarget(): there is "+
                  "no target");   
      if (!m_MoreCriteria_.keySet().contains(_targ.getName()))
          throw new InternalError("NoThanViewInBase.removeEorAccessTarget(): "+
                  "the argument is not the current target");
      
      removeOwnListeningOnTarget();
      setMoreCriterion(null);
      setLessCriterion(null);

      try {
          firePropertyChange("FirstAccessTarget", _targ, null);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("NoThanViewInBaseImpl.removeEorAccessTarget(), firePropertChange() throws:");
        e.printStackTrace();
      } 
  } // removeEorNoThanAccessTarget(
  
  /**
   * The aim is to check if an Eligible Party has rights on the target
   * which are greater than the 'more' criteria.
   * @return the 'more' criteria. May be null.
   */
  public StringRight getMoreCriterion() {
      if (m_MoreCriteria_.isEmpty()) {

              // for testing purpose
              //System.out.println("");
              //System.out.println("in NoThanViewInBaseImpl.getMoreCriterion(), the "+
                //      "map is empty, so null result.");

          return null;
      }
      String more = m_MoreCriteria_.keySet().iterator().next();
      if (m_MoreCriteria_.get(more) == null) {

              // for testing purpose
              //System.out.println("");
              //System.out.println("in NoThanViewInBaseImpl.getMoreCriterion(), the "+
                //      "map does not contain value for the first key, so null result.");

          return null;
      }
      else {

              // for testing purpose
              //System.out.println("");
              //System.out.println("in NoThanViewInBaseImpl.getMoreCriterion(), the "+
                //      "result is "+((StringRight[])m_MoreCriteria_.get(more))[0]);

          return ((StringRight[])m_MoreCriteria_.get(more))[0];
      }
  } // getMoreCriterion()
    
  /**
   * The aim is to check if an Eligible Party has rights on the target
   * which are greater than the 'more' criteria. Fires the 'OneMoreCriterion' property change
   * with the old and the new values.
   * @param _right is the 'more' criteria. May be null or have the same power than the less right.
   * @exception BaseError if the argument is empty, or the target is null, or
   *    the argument is inferior to the less right.
   */
  public void setMoreCriterion(StringRight _right) throws BaseError {
    if (_right != null && (_right.isEmpty() || !_right.isGrantingRight()))
      throw new BaseError(" The argument is empty or negative");
    if (getEorFirstAccessTarget() == null)
      throw new BaseError(" There is no target");
    // check if the less right is not superior
    if (_right != null && getLessCriterion() != null && 
            _right.compareTo(getLessCriterion()) < 0)
      throw new BaseError(" The new more right "+
              "is inferior to the less right");         
    
    StringRight old = getMoreCriterion();
    String more = m_MoreCriteria_.keySet().iterator().next();
    if (_right != null && !_right.isEmpty())
        m_MoreCriteria_.put(more, new StringRight[]{_right}); 
    else
        m_MoreCriteria_.put(more, null);

    try {
        firePropertyChange("OneMoreCriterion", old, _right);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("NoThanViewInBaseImpl.setMoreCriterion(), firePropertChange() throws:");
        e.printStackTrace();
    } 
    old = null;
    more = null;
  } // setMoreCriterion()

  /**
   * The aim is to check if an Eligible Party has rights on the target,
   * which are lesser than the 'less' criteria.
   * @return the 'less' criteria. May be null.
   */
  public StringRight getLessCriterion() {
      if (m_LessCriteria_.isEmpty()) return null;
      String less = m_LessCriteria_.keySet().iterator().next();
      if (m_LessCriteria_.get(less) == null) return null;
      else return ((StringRight[])m_LessCriteria_.get(less))[0];      
  } // getLessCriterion()
  
  /**
   * The aim is to check if an Eligible Party has rights on the target,
   * which are lesser than the 'less' criteria. Fire the 'OneLessCriterion' property change
   * with the old and the new values.
   * @param _right is the 'less' criteria. May be null.
   * @exception BaseError if the argument is empty, the target is null, or
   *    the argument is superior to the more right.
   */
  public void setLessCriterion(StringRight _right) throws BaseError {
    if (_right != null && (_right.isEmpty() || !_right.isGrantingRight()))
      throw new BaseError(" The argument is empty or negative");
    if (getEorFirstAccessTarget() == null)
      throw new BaseError(" There is no target");
    // check if the more right is not inferior
    if (_right != null && getMoreCriterion() != null &&
            _right.compareTo(getMoreCriterion()) > 0)
      throw new BaseError(" The new less right "+
              "is superior to the more right"); 
    
    StringRight old = getLessCriterion();
    
    String less = m_LessCriteria_.keySet().iterator().next();
    if (_right != null && !_right.isEmpty())
        m_LessCriteria_.put(less, new StringRight[]{_right}); 
    else
        m_LessCriteria_.put(less, null);
    
    try {
        firePropertyChange("OneLessCriterion", old, _right);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("NoThanViewInBaseImpl.setLessCriterion(), firePropertChange() throws:");
        e.printStackTrace();
    } 
    old = null;
    less = null;
  } // setLessCriterion()
  
  /**
   * An excluded Source is connected to a view basic Source, but it
   * has not to be checked since it is explicitly put away by the user.
   * So its compliance to the more and to the less criteria
   * will stay unknown.
   * @return list of the excluded sources. Never null.
   */
  public List<ImmutableSource> getExcludedSources() {
      return (List<ImmutableSource>)((ArrayList)l_excludedSources_).clone();
  } // getExcludedSources()
  
  /**
   * An excluded Source is connected to a view basic Source, but it
   * has not to be checked since it is explicitly put away by the user.
   * So its compliance to the more and to the less criteria will stay unknown.
   * @param _sour to test
   * @return true if the view references this Source as excluded
   */
  public boolean containsExcludedSource(ImmutableSource _sour) {
      return l_excludedSources_.contains(_sour);
  } // containsExcludedSource()

  /**
   * Adds an excluded Source, which is is in the perimeter of checking, but
   * has not to be checked. So its compliance to the more and to the less criteria
   * will stay unknown. The perimeter of checking is defined as the set of the Sources
   * which are registered in the view (but such Source may not be excluded),
   * and the Source which are  direct or indirect member of, or run under,
   * a registered Source of the view.
   * Fires 'OneExcludedEP' event.
   * @param _sour source to exclude
   * @exception BaseError if the argument is a registered Source,
   * is not in the perimeter of checking,
   * if the argument is null, or the object name is null or empty.
   */
  public void addEorExcludedSource(ImmutableSource _sour) throws BaseError {
      // for finalized instances that are still in the memory stack
      if (m_BaseObjects_ == null) return;
      if (_sour == null || _sour.getName() == null || _sour.getName().isEmpty())
          throw new BaseError(" The argument or its name is null or empty");
      if (containsBaseObject(_sour))
          throw new BaseError(" The argument is a basic rights user");
      if (l_excludedSources_.contains(_sour))
          throw new BaseError(" The argument is already an excluded Source");
      if (l_excludedSources_.size() > 8)
             throw new BaseError(" The excluded sources list cannot contain more than 9 elements.");
      
      boolean found = false;
      boolean found2 = false;
      EligibleParty ep = null;
      Actor[] l_act = null;
      
      for (Iterator ite = m_Sources_.values().iterator(); ite.hasNext(); ) {
          ep = (EligibleParty)ite.next();
          if (ep instanceof Actor) continue;
          if (ep instanceof UserID) {
              if (_sour instanceof Actor) {
                l_act = ((UserID)ep).getEorL_UserIDCurrentActors();
                for (int i = 0; i < l_act.length; i++) {
                    if (_sour.equals(l_act[i])) {
                        found = true;
                        break;
                    }
                } // for i
              } // if
          } // if
          else { // ep instanceof GroupID
              if (_sour instanceof Actor) {
                l_act = ((GroupID)ep).getEorL_GroupIDCurrentActors();
                for (int i = 0; i < l_act.length; i++) {
                    if (_sour.equals(l_act[i])) {
                        found2 = true;
                        break;
                    }
                } // for i
              }
              else { // _ep instanceof UserID or GroupID
                if (((GroupID)ep).containsAsIndirectMember((GroupIDMember)_sour) ||
                        ((GroupID)ep).containsAsMember((GroupIDMember)_sour) ) {
                    found2 = true;
                    break;
                }
              }            
          } // else
          if (found) break;
          if (found2) break;
      } // for ite

      if (!found && !found2)
          throw new BaseError("NoThanViewInBase.addEorExcludedEligibleParty(): "+
                  "the argument is not an Actor running under a view userID, "+
                  "nor a GroupIDMember running under a view groupID.");

      ep = null;
      l_act = null;
      
      removeOwnListeningOnPerimeterEPs();
      l_excludedSources_.add(_sour);
      setUpContextLists(); // fires 'CheckingPerimeter', 'ReducedEPsInViewEPs', 'ObjectChangesInListeningPerimeter'
      addOwnListeningOnPerimeterEPs();

                        // for testing purpose
                        //System.out.println("in NoThanViewInBaseImpl."+
                          //"addEorExcludedEligibleParty() fires 'OneExcludedEP' ");
      
      try {
          firePropertyChange("OneExcludedEP", null, _sour);
      } catch (Exception e) { 
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("NoThanViewInBaseImpl.addEorExcludedEligibleParty(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
      } 

  } // addEorExcludedSource()

  /**
   * Removes an excluded Source, which is in the perimeter of checking, but
   * has not to be checked. So its compliance to the more and to the less criteria
   * will stay unknown. The perimeter of checking is defined as the set of the Source
   * which are registered in the view (but such Source may not be excluded),
   * and the Source which are direct or indirect member of, or run under,
   * a registered Source of the view.
   * Fires 'OneExcludedEP' event.
   * @param _sour non-null Source to remove as excluded Source
   * @exception BaseError if the Source is unknown as excluded, or
   * if the argument is null, if the object name is empty.
   */
  public void removeEorExcludedSource(ImmutableSource _sour) throws BaseError {
      if (_sour == null || _sour.getName() == null || _sour.getName().isEmpty())
          throw new BaseError(" The argument or its name is null or empty");
      if (!l_excludedSources_.contains(_sour))
             throw new BaseError(" The argument is not an excluded eligible party.");
      
      removeOwnListeningOnPerimeterEPs();
      l_excludedSources_.remove(_sour);
      // fire 'CheckingPerimeter', 'ReducedEPsInViewEPs', 'ObjectChangesInListeningPerimeter'
      setUpContextLists(); 
      addOwnListeningOnPerimeterEPs();
      
      try {
          firePropertyChange("OneExcludedEP", _sour, null);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("NoThanViewInBaseImpl.addEorExcludedEligibleParty(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
      } // catch
  } // removeEorExcludedSource()
  
  /**
   * List of Sources which are in the perimeter of checking, and that
   * includes the view BaseObjects and the listening perimeter.
   * Gets the checking perimeter which includes the view BaseObjects and the listening perimeter.
   * The associates of an Source are defined as their direct/indirect members
   * and their direct/indirect running-under actors.
   * This listening perimeter is get by the subtraction of the excluded Sources
   * and their associates from the view Sources list (also called 'basic rights users')
   * and their associates. <P>
   * For instance, if the group A is a view Source, and has only group B and
   * group C has direct members, and if the group C is a view excluded Source,
   * the listening perimeter contains only the group B and its associates. <P>
   * The Sources in the listening perimeter do not registered this NoThanViewInBase
   * has an implied view. The size is limited to 50.
   * @return the Sources in the checking perimeter. Never null.
   */
  public Set<ImmutableSource> getCheckingPerimeter() {
      return new HashSet(l_checkingPerimeter_);
  } // getCheckingPerimeter()
  
  /**
   * List of Sources from the view Sources list which have at most a connected
   * member or a running-under actor being an excluded Source. Do not contain
   * the reduced Sources in the checking perimeter which are not view Sources, that is
   * when they are members or running-under Actors of a view Source.
   * @return the Sources. Never null.
   */
  public Set<ImmutableSource> getReducedSourcesInViewSources() {
      return new HashSet(l_reducedSourcesInViewSources_);
  } // getReducedEPsInViewEPs()
  
  /**
   * Checks if there is not more that 8 elements, and forbids it otherwise.
   * Overrides the super method and calls it.
   * Called by the constructor through the super constructor and addEorHardAliasOrReference().
   * @param _targ new right user to analyze
   * @exception CreateError if the argument is not an Actor, if the limit of 9 view objects
   * is get, if the target is already known by the view,
   * if the argument is null, or the object name is null or empty,
   * or if the name is already known by the view.
   */
  public void addEorTarget(ImmutableTarget _targ) throws CreateError {
      // for finalized instances that are still in the memory stack
      if (m_BaseObjects_ == null) return;
      if (m_BaseObjects_.size() > 8)
             throw new CreateError(" The Right Users map cannot contain more than 9 elements.");
      // null if called by the constructor through the super constructor
      if (l_checkingPerimeter_ != null)
          removeOwnListeningOnPerimeterEPs();

      super.addEorTarget(_targ); // fires 'OneBaseObjectInFolder' and 'connectedACS'

      if (l_checkingPerimeter_ != null) {
        setUpContextLists(); // fires 'CheckingPerimeter', 'ReducedEPsInViewEPs', 'ObjectChangesInListeningPerimeter'
        addOwnListeningOnPerimeterEPs();
      } // if
  } // addEorTarget()

  /**
   * Adds a view Source which is not a Target, and this does not exist.
   * Checks if there is not more that 8 elements, and forbid it otherwise.
   * Overrides the super method and calls it.
   * Called by the constructor through the super constructor and addEorHardAliasOrReference().
   * @param _sour new right user to analyze
   * @exception CreateError if _sour name is null, the limit of 9 view objects
   * or 50 objects in the checking perimeter is get.
   */
   public synchronized void addEorSource(ImmutableSource _sour) throws CreateError {
      // for finalized instances that are still in the memory stack
      if (m_BaseObjects_ == null) return;
      if (m_BaseObjects_.size() > 8)
             throw new CreateError(" The Right Users map cannot contain more than 9 elements.");

            // null if called by the constructor through the super constructor
      if (l_checkingPerimeter_ != null)
          removeOwnListeningOnPerimeterEPs();

      super.addEorSource(_sour); // fires 'OneBaseObjectInFolder' and 'connectedACS'
      
      if (l_checkingPerimeter_ != null) {
          setUpContextLists(); // fires 'CheckingPerimeter', 'ReducedEPsInViewEPs', 'ObjectChangesInListeningPerimeter'

          if (l_checkingPerimeter_.size() > 50) {
              removeEorSource(_sour);
              throw new CreateError(" The checking perimeter list "+
                      "cannot contain more than 50 elements.");
          }

          addOwnListeningOnPerimeterEPs();
      } // if
  } // addEorSource()
  
  /**
   * Removes a view node.
   * Overrides the super method and calls it.
   * @param _targ right user to remove
   * @exception CreateError if the resource is not known by the view, or
   * if the object name is empty.
   */
  public void removeEorTarget(ImmutableTarget _targ) throws CreateError {
      removeOwnListeningOnPerimeterEPs();
      super.removeEorTarget(_targ); // fire 'OneBaseObjectInView' and 'connectedACS'
      setUpContextLists(); // fire 'CheckingPerimeter', 'ReducedEPsInViewEPs', 'ObjectChangesInListeningPerimeter'
      addOwnListeningOnPerimeterEPs();
  } // removeEorTarget()

  /**
   * Removes a view Source which is not a Target, and this does not exist.
   * Overrides the super method and calls it.
   * @param _sour right user to remove
   * @exception CreateError if the Source is unknown as excluded, or
   * if the argument is null, if the object name is empty.
   */
  public void removeEorSource(ImmutableSource _sour) throws CreateError {
      removeOwnListeningOnPerimeterEPs();
      super.removeEorSource(_sour); // fire 'OneBaseObjectInView' and 'connectedACS'
      setUpContextLists(); // fire 'CheckingPerimeter', 'ReducedEPsInViewEPs', 'ObjectChangesInListeningPerimeter'
      addOwnListeningOnPerimeterEPs();
  } // removeEorSource()
  
  /**
   * Observer of the property changes and adding/removing of the accesss target,
   * the adding/removing of excluded EPs, the adding/removing of the listener perimeter,
   * the property changes of every EP in the checking perimeter, 
   * the right changes on any implied ACL.
   * Overrides the super method, and calls it at the last line if necessary.
   * Listens 'OneBaseObject' on the ACS of the excluded EP, to remove them when
   * they are deleting.
   * Does not listen 'ResourceType', 'EPType', 'VirtualFolderType', 'Comment', 'ImpliedOpenViews'
   * from the listened objects.
   * Does not fire a 'InheritedOwnerRights' event after a 'FromTheParentAccount' event
   * from the target or an actor of the perimeter, and so on for the group similar events.
   * Does not fire a 'InheritedTargetPrivileges' or a 'InheritedTargetAclEntries' event
   * if the last event is from the parent.
   * @see #addOwnListeningOnTarget()
   * @see #addOwnListeningOnPerimeterEPs()
   * @param _evt PropertyChangeEvent
   */
  public void propertyChange(PropertyChangeEvent _evt) {

    // the event 'OnePropertyInACheckingPerimeterObject' is fired when any
                // EP of the checking perimeter has a property change.
    // when the checking perimeter changes, it produces always a 'OneBaseObjectInView'
                // or a 'OneExcludedEP' event, 
                // but there is also a complementary event 'OnePropertyInACheckingPerimeterObject'
                // with AclEntry[] as values, to follow the ACL changes.
                // 2 cases:
                // - actor adding or removing in the perimeter, then listen each of its ACL
                // - ACL adding or removing on an actor of the perimeter, then listen the ACL
      
      // for finalized or empty instance
    if (m_BaseObjects_ == null || m_BaseObjects_.isEmpty()) return;

    if (_evt.getPropertyName().equals("ResourceType")) return;
    else if (_evt.getPropertyName().equals("EPType")) return;
    else if (_evt.getPropertyName().equals("VirtualFolderType")) return;
    else if (_evt.getPropertyName().equals("Comment")) return;
    else if (_evt.getPropertyName().equals("ImpliedOpenViews")) return;    
    
    // copied from the super class
    if (lastEvent_ != null && lastEvent_.getSource() == _evt.getSource() ) {
        if (lastEvent_.getPropertyName().equals("FromTheParentAccount") &&
                _evt.getPropertyName().equals("InheritedOwnerRights")) return; 
        else if (lastEvent_.getPropertyName().equals("FromTheParentGroup") &&
                _evt.getPropertyName().equals("InheritedGroupRights")) return; 
    }

    if (lastEvent_ != null && _evt.getSource() instanceof PrivilegeTarget &&
            lastEvent_.getSource() == ((PrivilegeTarget)_evt.getSource()).getEorParent() ) {
        if (lastEvent_.getPropertyName().equals("InheritedTargetPrivileges") &&
                _evt.getPropertyName().equals("InheritedTargetPrivileges")) return;
    } // if
    if (lastEvent_ != null && _evt.getSource() instanceof Resource &&
            lastEvent_.getSource() == ((Resource)_evt.getSource()).getEorParent() ) {
        if (lastEvent_.getPropertyName().equals("InheritedTargetAclEntries") &&
                _evt.getPropertyName().equals("InheritedTargetAclEntries")) return;
    } // if

                    // for testing and tracking
                    //System.out.println("NoThanViewInBaseImpl.propertyChange() processed for "+
                      //      _evt.getPropertyName()+" from "+_evt.getSource());
                    
    lastEvent_ = _evt;
    
    // Fire all the view changes on the target, but not the target deleting (see hereafter)
    // nor the ACLs (done by GraphicViewBaseListenerImpl)
    if (_evt.getSource() == getEorFirstAccessTarget() ) {
        if ( _evt.getPropertyName().equals("UserID") ||
          _evt.getPropertyName().equals("GroupID") ||
          _evt.getPropertyName().equals("RootRights") ||
          _evt.getPropertyName().equals("GroupRights") ||
          _evt.getPropertyName().equals("OwnerRights") ||
          _evt.getPropertyName().equals("FromTheParentAccount") ||
          _evt.getPropertyName().equals("FromTheParentGroup") ||
          _evt.getPropertyName().equals("InheritedOwnerRights") ||
          _evt.getPropertyName().equals("InheritedGroupRights") ||
          _evt.getPropertyName().equals("OneTargetTypedPrivilege") ||
          _evt.getPropertyName().equals("OneTargetLinkedPrivilege") ||
          _evt.getPropertyName().equals("InheritedTargetPrivileges") ||
          _evt.getPropertyName().equals("InheritedTargetAclEntries"))
            firePropertyChange(_evt);
        // from DirectoryImpl
        else if (_evt.getPropertyName().equals("Children") ||
          _evt.getPropertyName().equals("EnforceChildGID") ||
          _evt.getPropertyName().equals("StickyBit") )
                firePropertyChange(_evt);
        // from ActorImpl
        else if (_evt.getPropertyName().equals("SecondaryGroups") ||
          _evt.getPropertyName().equals("l_isABridgeFor") ||
          _evt.getPropertyName().equals("l_isAGateFor") ||
          _evt.getPropertyName().equals("CurrentGroupID") ||
          _evt.getPropertyName().equals("CurrentUserID") ||
          _evt.getPropertyName().equals("OneSourceTypedPrivilege") ||
          _evt.getPropertyName().equals("OneSourceLinkedPrivilege"))
                firePropertyChange(_evt);
        // from ExecutableImpl
        else if (_evt.getPropertyName().equals("EnforceUID") ||
          _evt.getPropertyName().equals("EnforceGID") )
                firePropertyChange(_evt);
        // from EligiblePartyImpl
        else if (_evt.getPropertyName().equals("OneSourceTypedPrivilege") ||
          _evt.getPropertyName().equals("OneSourceLinkedPrivilege") )
                firePropertyChange(_evt);
        // from VirtualFolderImpl
        else if (_evt.getPropertyName().equals("OneBaseObject")  )
                firePropertyChange(_evt); 
        // all the AcsAddon events from the AcsAddon objects
        else if (((ACSObject)_evt.getSource()).getEorACS() instanceof AcsAddon)
            firePropertyChange(_evt);
    }

    // for the listening perimeter EP and their ACLs by GraphicViewBaseListenerImpl
    else if (_evt.getSource() instanceof ACSObject &&
            l_checkingPerimeter_.contains(_evt.getSource())) {
        try {    
            firePropertyChange("OnePropertyInAListeningPerimeterObject",
                    _evt.getOldValue(), _evt.getNewValue()); // unused
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From NoThanViewInBaseImpl.propertyChange(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
    }
    else if (_evt.getSource() instanceof AclEntry &&
            l_checkingPerimeter_.contains(((AclEntry)_evt.getSource()).getEorEP()) ) {
        try {    
                    firePropertyChange("OnePropertyInAListeningPerimeterObject", 
                            _evt.getOldValue(), _evt.getNewValue()); // unused
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From NoThanViewInBaseImpl.propertyChange(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
    } // else if
    
    // removes a deleted excluded source, a deleted target,
    // a deleted source in the listening perimeter
    else if (_evt.getSource() instanceof ACS && _evt.getNewValue() == null &&
      _evt.getPropertyName().equals("OneBaseObject") ) {

                            // for testing and tracking
                            //System.out.println("In NoThanViewInBaseImpl.propertyChange(),"
                              //+" the view "+name_+" has received an 'OneBaseObject' event ");
        
          if (_evt.getOldValue() instanceof ImmutableSource &&
                  containsExcludedSource( (ImmutableSource)_evt.getOldValue()) ) {
            try {

                        // for testing and tracking
                        //System.out.println("In NoThanViewInBaseImpl.propertyChange(),"
                          //+" remove the object "+( (BaseObject)_evt.getOldValue() ).getName());

                // this method fires the 'OneExcludedEP' event
                removeEorExcludedSource( (ImmutableSource)_evt.getOldValue() );

            } catch (BaseError b) {
              b.printStackTrace();
            }

          } // if
          else if (_evt.getOldValue() instanceof ImmutableTarget &&
                  _evt.getOldValue() == getEorFirstAccessTarget() ) {
            try {

                        // for testing and tracking
                        //System.out.println("In NoThanViewInBaseImpl.propertyChange(),"
                          //+" remove the object "+( (BaseObject)_evt.getOldValue() ).getName());

                // this method fires the 'FirstAccessTarget' event
                removeEorNoThanAccessTarget( (ImmutableTarget)_evt.getOldValue() );

            } catch (BaseError b) {
              b.printStackTrace();
            }

          } // if
          // process only the listening perimeter in the checking perimeter
          else if (_evt.getOldValue() instanceof ImmutableSource &&
                  !containsBaseObject((BaseObject)_evt.getOldValue()) &&
                  getCheckingPerimeter().contains( _evt.getOldValue() ) ) {

                        // for testing and tracking
                        //System.out.println("In NoThanViewInBaseImpl.propertyChange(),"
                          //+" remove the object "+( (BaseObject)_evt.getOldValue() ).getName());

                removeOwnListeningOnPerimeterEPs();
                // fires 'CheckingPerimeter', 'ReducedEPsInViewEPs',
                // 'ObjectChangesInListeningPerimeter'
                setUpContextLists();
                addOwnListeningOnPerimeterEPs();

          } // if          
    } // if (_evt.getSource() instanceof ACS for 'OneBaseObject'

    // FolderAbst.propertyChange() may remove a folder object in m_BaseObjects_,
    // ViewInBaseImpl.propertyChange() may close this instance, and
    // EPRViewInBaseImpl.propertyChange() fires all the view changes on
     // Resources, VirtualFolders and EligibleParties
    super.propertyChange(_evt);

  } // propertyChange()

  /**
   * no equals: each instance is equals only to itself.
   * @param _obj: object to compare
   * @return true if the two objects are identical instances
   * of identical classes that implemente ViewInBase
   */
   /*public synchronized boolean equals(Object _obj) {
    if (this == _obj)
        return true;
    if (_obj == null)
        return false;
    if (_obj.getClass() != getClass())
        return false;

    if (!super.equals(_obj))
        return false;

    return true;
  } //equals()
    */

  /**
   * Tne name is the only property which is used to compute the hashcode,
   * because after finalizeForBase(), hashcode() is used in BaseManagerImpl Maps.
   * @return the hashcode value for the view, based on its name.
   */
  /*public int hashCode() {
    int result = super.hashCode();
    //if (name_ != null) {
    //  result = PRIME*result + name_.hashCode();
    //}
    //if (m_BaseObjects_ != null) {
    //  result = PRIME*result + m_BaseObjects_.keySet().hashCode();
    //}
    return result;
  } // hashcode()
   */

  /**
  * Adds a listener to the bean for the properties 'FirstAccessTarget', 'OneExcludedEP',
  * 'OneLessCriterion', 'OneMoreCriterion', 'CheckingPerimeter', 'ReducedEPsInViewEPs',
  * 'ObjectChangesInListeningPerimeter', 'OnePropertyInAListeningPerimeterObject',
  * or for any superclass property. Calls the super method.
  * @param _propertyName is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      // for finalized instances that are still in the memory stack
      if (m_BaseObjects_ == null) return;
      if (changeSupport_ == null) {
        changeSupport_= new PropertyChangeSupport(this);
      }

      if (_propertyName.equals("OneExcludedEP") ||
              _propertyName.equals("FirstAccessTarget") ||
              _propertyName.equals("OneLessCriterion") ||
              _propertyName.equals("OneMoreCriterion") ||
              _propertyName.equals("l_CheckingPerimeter") ||
              _propertyName.equals("l_ReducedEPsInViewEPs") ||
              _propertyName.equals("l_ObjectChangesInListeningPerimeter") ||
              _propertyName.equals("OnePropertyInAListeningPerimeterObject")) {
          synchronized (changeSupport_) {
            changeSupport_.addPropertyChangeListener( _propertyName, _l);}
      }
      else super.addPropertyChangeListener(_propertyName, _l);
  } // addPropertyChangeListener()

  /**
  * Removes a listener to the bean for the properties 'FirstAccessTarget',
  * 'OneExcludedEP', 'OneLessCriterion', 'OneMoreCriterion',
  * 'CheckingPerimeter', 'ReducedEPsInViewEPs',
  * 'ObjectChangesInListeningPerimeter', 'OnePropertyInAListeningPerimeterObject',
  * or for any superclass property. Call the super method.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      if (changeSupport_ == null) return;

      if (_propertyName.equals("OneExcludedEP") ||
              _propertyName.equals("FirstAccessTarget") ||
              _propertyName.equals("OneLessCriterion") ||
              _propertyName.equals("OneMoreCriterion") ||
              _propertyName.equals("l_CheckingPerimeter") ||
              _propertyName.equals("l_ReducedEPsInViewEPs") ||
              _propertyName.equals("l_ObjectChangesInListeningPerimeter") ||
              _propertyName.equals("OnePropertyInAListeningPerimeterObject")) {
          synchronized (changeSupport_) {
            changeSupport_.removePropertyChangeListener( _propertyName, _l);}
      }
      else super.removePropertyChangeListener(_propertyName, _l);

          // for testing and tracking
          //System.out.println("For the view "+name_+", invoke removePropertyChangeListener(),"+
          //  "for the property 'OnePropertyInView': ");
          //for (Iterator ite = l_OnePropertyInViewChangeSupport_.iterator() ; ite.hasNext() ; ){
          //  System.out.println((PropertyChangeListener)ite.next() );
          //}
          //}
  } // removePropertyChangeListener()
  
  /**
  * Creates, registers in the base and saves a view copy with another name.
  * Calls copy().
  * @param _name of the copy. It is copied without test.
  * @exception BaseError if _name is null or already known in BaseManagerImpl,
  * or if _view is null or not known.
  * @return the copied view
  */
  public NoThanViewInBase copyAndSave(ImmutableName _name) throws BaseError {
      
      ViewInBase result = copy(_name, this, true);

      // add result to the map of Views in BaseManagerImpl.
      // throws BaseError if already exists.
        // for testing the view without ViewManager: TAKE OUT THIS LIGNE
      //BaseManagerImpl.getInstance().newEorCopiedViewToSave(result);

      BaseManagerImpl.getInstance().saveOpenView(result);
      
      return (NoThanViewInBase)result;
  } // copyAndSave
  
        //
        // ------------- methods out of the extended interfaces --------------------
        //

  /**
  * Creates a copy with another name, and may register it.
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy): baseError_,
  * m_BaseObjects_, m_NoActorResources_, m_Sources_.
  * Called by copyAndSave().
  * @param _name of the view. It is copied.
  * @param _view to copy.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in BaseManagerImpl.
  * @exception BaseError if _name is null or already known in BaseManagerImpl,
  * or if _view is null or not known.
  * @return a copy
  */
  protected static ViewInBase copy(ImmutableName _name,
    NoThanViewInBase _view, boolean _toRegisterInManager) throws BaseError {

              // for testing and tracking
              System.out.println("For the NoThan view, copy()");
                //"for the property 'OnePropertyInView': ");
              //for (Iterator ite = l_OnePropertyInViewChangeSupport_.iterator() ; ite.hasNext() ; ){
              //  System.out.println((PropertyChangeListener)ite.next() );
              //}
              //}
      
      if (_view == null || _name == null ||
        !BaseManagerImpl.getInstance().knowsView(_view.getName() )) {
          throw new BaseError("in NoThanViewInBaseImpl.copy(_name, _view, boolean),"+
          " _name or _view is null, or _view is not known by the base.");
      }

      NoThanViewInBaseImpl result = null;

      try {
          result = new NoThanViewInBaseImpl(
                _name,
                (Map)((TreeMap)((NoThanViewInBaseImpl)_view).m_BaseObjects_).clone(),
                false);
      } catch (CreateError c) {
            throw new BaseError(c.toString());
      }

      // copy _view
      result.baseError_ = ((NoThanViewInBaseImpl)_view).getEorBaseError();
            // clone the names in l_excludedSources_
      List<ImmutableSource> list = _view.getExcludedSources();
      result.l_excludedSources_ = new ArrayList(list.size());
      for (Iterator ite = list.iterator(); ite.hasNext(); ) {
          result.l_excludedSources_.add((EligibleParty)ite.next() );
      } // for ite
      result.m_MoreCriteria_ = (Map<String,StringRight[]>)((HashMap)
              ((NoThanViewInBaseImpl)_view).m_MoreCriteria_).clone();
      result.m_LessCriteria_ = (Map<String,StringRight[]>)((HashMap)
              ((NoThanViewInBaseImpl)_view).m_LessCriteria_).clone();
      result.l_checkingPerimeter_ = (List<ImmutableSource>)((ArrayList)
              ((NoThanViewInBaseImpl)_view).l_checkingPerimeter_).clone();
      result.l_reducedSourcesInViewSources_ = (List<ImmutableSource>)((ArrayList)
              ((NoThanViewInBaseImpl)_view).l_reducedSourcesInViewSources_).clone();
      
        // add result to the map of Views in BaseManagerImpl.
        // throws BaseError if already exists.
      if (_toRegisterInManager) {
        try {
            BaseManagerImpl.getInstance().newEorView(result);
        } catch (CreateError c) {
            result.finalizeForCopy();
            throw new BaseError(c.toString());
        }
        // registering by the two super classes
        result.addOwnListening(true, true);
      } 
      
      return result;
  } // copy
  
  /**
  * Registers the instance as a property event listener
  * for any change on the access target or the checking perimeter,
  * since they are not handled as a view object in the super classes.
  * Called by the constructor, copy(), addEorAccessTarget() and the
  * prepareAfterRestoring() methods. 
  * Calls addOwnListeningOnTarget(), addOwnListeningOnPerimeterEPs()
  * and super.addOwnListening(_listenACSs).
  * @param _callSuper false if this instance must not call the method super class
  *     EPRViewInBaseImpl.addOwnListening(),
  *     to listen the open ACS events in BaseManagerImpl and the objects events in ACSImpl,
  *     nor the changes in the no-actor resources rights
  * @param _listenACSs false if this instance must not listen
  *     the open ACS events in BaseManagerImpl and the objects events in ACSImpl,
  *     through the indirect call to ViewInBaseImpl.addOwnListening(). If true,
  *     _callSuper has to be true. Else, an InternalError is thrown.
  */
  protected void addOwnListening(boolean _callSuper, boolean _listenACSs) {      
    if (_listenACSs && !_callSuper)
        throw new InternalError("NoThanViewInBaseImpl.addOwnListening() with "+
                "_listenACSs true and _callSuper false");
      
    addOwnListeningOnTarget();
    addOwnListeningOnPerimeterEPs();
    
    if (_callSuper) super.addOwnListening(_listenACSs);
 
  } // addOwnListening 
  
  /**
  * This method unregister the instance as a property event listener
  * for any change on the access target or the checking perimeter,
  * since it is not handled as a view object in the super classes.
  * Called by finalizeForBase().
  * Calls removeOwnListeningOnTarget() and super.removeOwnListening(_listenACSs).
  * @param _callSuper false if this instance must not be unregistered for the listening
  *     of the open ACS events in BaseManagerImpl and the objects events in ACSImpl,
  *     nor the changes in the no-actor resources rights
  * @param _listenACSs false if this instance must not listen
  *     the open ACS events in BaseManagerImpl and the objects events in ACSImpl,
  *     through the indirect call to ViewInBaseImpl.addOwnListening(). If true,
  *     _callSuper has to be true. Else, an InternalError is thrown.
   */
  protected void removeOwnListening(boolean _callSuper, boolean _listenACSs) {
      
    if (_listenACSs && !_callSuper)
        throw new InternalError("NoThanViewInBaseImpl.removeOwnListening() with "+
                "_listenACSs true and _callSuper false");
      
    removeOwnListeningOnTarget();
    removeOwnListeningOnPerimeterEPs();
    
    if (_callSuper) super.removeOwnListening(_listenACSs);
     
  } // removeOwnListening   
  
  /**
  * This method registers the instance as a property event listener
  * for any change on the access target, since it is not handled as a view object
  * in the super classes. Adds the listening of
  * the specific events from any AcsAddon ACS or any of its objects,
  * and BaseUtilityImpl.addAddonParentPropertyChangeListener() is called for that.
  * Called by addOwnListening() and setEorFirstAccessTarget().
  */
  protected void addOwnListeningOnTarget() {
    ImmutableTarget targ = getEorFirstAccessTarget();
    if (targ == null) return;
    
    // listen the deletings.
    // done for the view objects in ViewInBaseImpl.addOwnListening()
    if (targ instanceof Resource) {
        ((BasicImpl)targ).getEorACS().addPropertyChangeListener("OneBaseObject", this);

        // process the target like the no-actor resources in EPRViewInBase,
        // but completed by the listening of the AclEntries (done for the view
        // BaseObjects, in EPRViewInBase, by the ViewRightsNetworkObserver
        // which does not cover the NoThanViewInBase target)
        ((BasicImpl)targ).addInViewList(this);

        AclEntry[] l_acl = ((ResourceImpl)targ).getL_aclEntry();
        for (int i = 0; i < l_acl.length; i++) {
          if (l_acl[i] != null) {
              l_acl[i].addPropertyChangeListener("AclRights", this);
              l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);
          } // if
        } // for i
        l_acl = null;

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the ACS Addon
        if (((ACSObject)targ).getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().addAddonParentPropertyChangeListener(
                      this, (ACSObject)targ);

        // to follow the access accross the nodes tree of the Resource
        while (((Resource)targ).getEorParent() != null) {
          ((Resource)targ).addPropertyChangeListener("FromTheParentAccount", this);
          ((Resource)targ).addPropertyChangeListener("FromTheParentGroup", this);
          ((Resource)targ).addPropertyChangeListener("InheritedOwnerRights", this);
          ((Resource)targ).addPropertyChangeListener("InheritedGroupRights", this);
          ((Resource)targ).addPropertyChangeListener("InheritedTargetPrivileges", this);
          ((Resource)targ).addPropertyChangeListener("InheritedTargetAclEntries", this);

          targ = (ImmutableTarget)((ResourceImpl)targ).getEorParent();
        } // while
    }
    else if (targ instanceof VirtualFolder) {

        VirtualFolder vf = (VirtualFolder)targ;
        vf.getEorACS().addPropertyChangeListener("OneBaseObject", this);

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the ACS Addon
        if (vf.getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().addAddonParentPropertyChangeListener(
                      this, vf);

        while (vf.getEorParent() != null) {
          vf.addPropertyChangeListener("InheritedTargetPrivileges", this);

          vf = (VirtualFolder)vf.getEorParent();
        } // while

        // process the target like the no-actor resources in EPRViewInBase,
        // but completed by the listening of the AclEntries (done for the view
        // BaseObjects, in EPRViewInBase, by the ViewRightsNetworkObserver
        // which does not cover the NoThanViewInBase target)
        ((VirtualFolderImpl)targ).addInViewList(this);
    } // else

    targ.addPropertyChangeListener(this);

    targ = null;
  } // addOwnListeningOnTarget 
  
  /**
  * This method unregisters the instance as a property event listener
  * for any change on the access target, since it is not handled as a view object
  * in the super classes. Removes the listening of
  * the specific events from any AcsAddon ACS or any of its objects,
  * and BaseUtilityImpl.removeAddonParentPropertyChangeListener()
  * is called for that.
  * Called by removeOwnListening(), removeEorNoThanAccessTarget() and setEorFirstAccessTarget().
   */
  protected void removeOwnListeningOnTarget() { 
    ImmutableTarget targ = getEorFirstAccessTarget();
    if (targ == null) return;
    
    if (targ instanceof ResourceImpl) {
        // listen the deletings.
        // done for the view objects in ViewInBaseImpl.addOwnListening()
        ((BasicImpl)targ).getEorACS().removePropertyChangeListener("OneBaseObject", this);

        // process the target like the no-actor resources in EPRViewInBase
        ((BasicImpl)targ).removeInViewList(this);

        AclEntry[] l_acl = ((ResourceImpl)targ).getL_aclEntry();
        for (int i = 0; i < l_acl.length; i++) {
          if (l_acl[i] != null) {
              l_acl[i].removePropertyChangeListener("AclRights", this);
              l_acl[i].removePropertyChangeListener("ConditionSourceGroups", this);
          } // if
        } // for i
        l_acl = null;

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events may change the rights of an ACSObject in the ACS Addon
        if (((ACSObject)targ).getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().removeAddonParentPropertyChangeListener(
                      this, (ACSObject)targ);

        // to follow the access accross the nodes tree of the target
        while (((Resource)targ).getEorParent() != null) {
          ((Resource)targ).removePropertyChangeListener("FromTheParentAccount", this);
          ((Resource)targ).removePropertyChangeListener("FromTheParentGroup", this);
          ((Resource)targ).removePropertyChangeListener("InheritedOwnerRights", this);
          ((Resource)targ).removePropertyChangeListener("InheritedGroupRights", this);
          ((Resource)targ).removePropertyChangeListener("InheritedTargetPrivileges", this);
          ((Resource)targ).removePropertyChangeListener("InheritedTargetAclEntries", this);

          targ = (ImmutableTarget)((ResourceImpl)targ).getEorParent();
        } // while

    } // if (targ instanceof ResourceImpl) {
    else if (targ instanceof VirtualFolder) {

        VirtualFolder vf = (VirtualFolder)targ;
        vf.getEorACS().removePropertyChangeListener("OneBaseObject", this);

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events may change the rights of an ACSObject in the ACS Addon
        if (vf.getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().removeAddonParentPropertyChangeListener(
                      this, vf);

        while (vf.getEorParent() != null) {
          vf.removePropertyChangeListener("InheritedTargetPrivileges", this);

          vf = (VirtualFolder)vf.getEorParent();
        } // while

        // process the target like the no-actor resources in EPRViewInBase,
        // but completed by the listening of the AclEntries (done for the view
        // BaseObjects, in EPRViewInBase, by the ViewRightsNetworkObserver
        // which does not cover the NoThanViewInBase target)
        ((VirtualFolderImpl)targ).removeInViewList(this);
    } // else

    targ.removePropertyChangeListener(this);

    targ = null;
  } // removeOwnListeningOnTarget
  
  /**
  * This method registers the instance as a property event listener
  * for any change on the listening perimeter.
  * Caution: to call only after a l_checkingPerimeter_ updating by setUpContextLists().
  * Called by addOwnListening(), add/removeEorExcludedEligibleParty(),
  * add/removeEorRemove/EligibleParty(). 
  * Listens the property changes of every EP in the listening perimeter,
  * and the right changes on any implied ACL where the EP is the rights user.
  * If the EP is an actor, does not listen the AGO inherited rights since
  * only the actor rights on the target are searched.
  */
  protected void addOwnListeningOnPerimeterEPs() {
    ImmutableSource       sour = null;
    AclEntry[]            l_acl = null;
    Resource              res = null;
    
    for (Iterator ite = l_checkingPerimeter_.iterator(); ite.hasNext(); ) {
      sour = (ImmutableSource)ite.next();
      
      if (m_BaseObjects_.values().contains(sour)) continue;
          
      sour.addPropertyChangeListener(this);
      
      for (Iterator ita = ((ImmutableEligibleParty)sour).getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
          l_acl = ((Resource)ita.next()).getL_aclEntry();
          if (l_acl == null || l_acl.length == 0) continue;
          
          for (int i = 0; i < l_acl.length; i++) {
            if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                  l_acl[i].addPropertyChangeListener("AclRights", this);
                  l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);
            } // if
          } // for          
      } // for ita

      if (sour instanceof ImmutableActor) {
        res = (Resource)sour;

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the ACS Addon
        if (res.getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().addAddonParentPropertyChangeListener(
                      this, res);

        // to follow the access accross the nodes tree of the Resource
        while (res.getEorParent() != null) {
          res.addPropertyChangeListener("FromTheParentAccount", this);
          res.addPropertyChangeListener("FromTheParentGroup", this);
          res.addPropertyChangeListener("InheritedOwnerRights", this);
          res.addPropertyChangeListener("InheritedGroupRights", this);
          res.addPropertyChangeListener("InheritedTargetPrivileges", this);
          res.addPropertyChangeListener("InheritedTargetAclEntries", this);

          res = (Resource)res.getEorParent();
        } // while
      } // if (sour immutableActor
      
    } // for ite
    sour = null;
    l_acl = null;
  } // addOwnListeningOnPerimeterEPs 
  
  /**
  * This method unregisters the instance as a property event listener
  * for any change on the listening perimeter.
  * Caution: to call after a l_checkingPerimeter_ updating by setUpContextLists().
  * Called by removeOwnListening(), add/removeEorExcludedEligibleParty(),
  * add/removeEorRemove/EligibleParty(), add/removeEorTarget().
   */
  protected void removeOwnListeningOnPerimeterEPs() {

    ImmutableSource       sour = null;
    AclEntry[]            l_acl = null;
    Resource              res = null;
    
    for (Iterator ite = l_checkingPerimeter_.iterator(); ite.hasNext(); ) {
      sour = (ImmutableSource)ite.next();
      
      if (m_BaseObjects_.values().contains(sour)) continue;
          
      sour.removePropertyChangeListener(this);
      
      for (Iterator ita = ((ImmutableEligibleParty)sour).getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
          l_acl = ((Resource)ita.next()).getL_aclEntry();
          if (l_acl == null || l_acl.length == 0) continue;
          
          for (int i = 0; i < l_acl.length; i++) {
            if (l_acl[i] != null && l_acl[i].getEorEP() == sour ) {
                  l_acl[i].removePropertyChangeListener("AclRights", this);
                  l_acl[i].removePropertyChangeListener("ConditionSourceGroups", this);
            } // if
          } // for          
      } // for ita
      
      if (sour instanceof ImmutableActor) {
        res = (Resource)sour;

        // Gets the specific events to listen from an ACS Addon ACSObject,
        // because these events should change the rights of an ACSObject in the ACS Addon
        if (res.getEorACS() instanceof AcsAddon)
          UtilityImpl.getFirstInstance().removeAddonParentPropertyChangeListener(
                      this, res);
        
        // to follow the access accross the nodes tree of the Resource
        while (res.getEorParent() != null) {
          res.removePropertyChangeListener("FromTheParentAccount", this);
          res.removePropertyChangeListener("FromTheParentGroup", this);
          res.removePropertyChangeListener("InheritedOwnerRights", this);
          res.removePropertyChangeListener("InheritedGroupRights", this);
          res.removePropertyChangeListener("InheritedTargetPrivileges", this);
          res.removePropertyChangeListener("InheritedTargetAclEntries", this);

          res = (Resource)res.getEorParent();
        } // while        
      } // if (sour
      
    } // for ite
    sour = null;
    l_acl = null;
  } // removeOwnListeningOnPerimeterEPs

  /**
   * Empty method since the NoThanView does not handle the alias relation as the EPRView.
   * Overriddes the superclass method.
   * @param _obj with an alias relation to process
   * @exception CreateError if the name is null, or the object name is null or empty.
   */
  protected void addEorHardAlias(ACSObject _obj) throws CreateError {

  } // addEorHardAlias()

  /**
   * Empty method since the NoThanView does not handle the alias relation as the EPRView.
   * Overriddes the superclass method.
   * @param _obj with an alias relation to process
   * @exception CreateError if the name is null, or the object name is null or empty.
   */
  protected void removeEorHardAlias(ACSObject _obj) throws CreateError {

  } // removeEorHardAlias()
  
  /**
  * The view is removed as a listener and the internal maps are set to null.
  * The view is NOT still closed in BaseManager.
  * Called by BaseManagerImpl.closeOpenView() and, indirectly, by propertyChange().
  * @exception UpDateError
  */
  public synchronized void finalizeForBase() throws UpDateError {
    // noop if it is the second call to this method
    if (m_Sources_ == null) {
        
              // for testing purpose
              //System.out.println("in NoThanViewInBaseImpl.finalizeForBase(), for "+
                //      getName()+", second call to finalization.");

      return;
    }
    
    removeOwnListening(true, true);

    l_excludedSources_.clear();
    l_excludedSourceNames_.clear();
    m_MoreCriteria_.clear();
    m_LessCriteria_.clear();
    l_excludedSources_ = null;
    l_excludedSourceNames_ = null;
    m_MoreCriteria_ = null;
    m_LessCriteria_ = null;
    l_checkingPerimeter_ = null;
    l_reducedSourcesInViewSources_ = null;

    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * This method clears l_excludedSourceNames_, m_MoreCriteria_ and m_LessCriteria_.
  * The view is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null. Calls the super method.
  * Called by BaseManager.openView() when prepareAfterRestoring() throws an exception.
   */
  protected synchronized void finalizeFromDeserialization() {
    l_excludedSourceNames_.clear();
    m_MoreCriteria_.clear();
    m_LessCriteria_.clear();

    l_excludedSourceNames_ = null;
    m_MoreCriteria_ = null;
    m_LessCriteria_ = null;
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  /**
  * The internal maps are set to null.
  * The view is NOT open in BaseManager.
  * Called by copy() and BaseManagerImpl.getElementsOfOneView().
   */
  public synchronized void finalizeForCopy() {
    // noop if it is the second call to this method
    if (m_Sources_ == null) {
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");

      return;
    }
     
    l_excludedSources_.clear();
    l_excludedSourceNames_.clear();
    m_MoreCriteria_.clear();
    m_LessCriteria_.clear();
    l_excludedSources_ = null;
    l_excludedSourceNames_ = null;
    m_MoreCriteria_ = null;
    m_LessCriteria_ = null;
    l_checkingPerimeter_ = null;
    l_reducedSourcesInViewSources_ = null;
    super.finalizeForCopy();
  } // finalizeForCopy()
  
  /**
   * This method must be called by BaseManagerImpl before the saving of the view,
   * to update l_excludedSourceNames_ from l_excludedSources_ which is not saved.
   * It is not necessary to clear l_excludedSources_ since it is transient. According
   * to the same principle, it is not necessary to clear the lists
   * which are derived from l_excludedSources_ since they are transient.
   * Override the ViewInBase method.
   */
  protected void prepareBeforeSaving() {
    if (l_excludedSourceNames_ != null) l_excludedSourceNames_.clear();
    else l_excludedSourceNames_ = new ArrayList(INITIAL_CAPACITY_5);
    
    for (Iterator ite = l_excludedSources_.iterator(); ite.hasNext();) {
      l_excludedSourceNames_.add(((ImmutableSource)ite.next()).getName());
    } // for ite
    
    super.prepareBeforeSaving();
  }  // prepareBeforeSaving()
  
  /**
   * This method must be called by BaseManagerImpl after the opening of the view,
   * to update the transient properties, and to check up the open ACS
   * and their objects which are in the restored view.
   * The deleting of ACS rights as criteria is not handled.
   * Noop if the folder objects has been restored, having a non-empty BaseObjects map.
   * Calls the super method at the first line.
   * @exception DialogError if the view contains an object from a closed ACS, before
   *    the registering of the view as listener; 
   *    the exception message should be shown to the user, and this instance finalized.
   * @exception InternalError if an object is not a Resource nor an EligibleParty.
   * @return a non-null message if some objects has been removed from the view,
   *    or null.
   */
  protected String prepareAfterRestoring() throws DialogError {
    // noop if the folder objects has been restored
    if (m_BaseObjects_ != null && !m_BaseObjects_.isEmpty())
        return null;
    
    ImmutableName   acsname = null;
    ImmutableName   name = null;  
    BaseObject      obj = null;
    
    String result = super.prepareAfterRestoring();
    
    // prepares the target
    ImmutableTarget target = getEorFirstAccessTarget();
    if (target == null && !m_MoreCriteria_.isEmpty()) {
      try {
          acsname = BaseUtilityImpl.getFromKownACS(
                    NameImpl.getNameFromString(
                        (String)m_MoreCriteria_.keySet().iterator().next()));
      } catch (BaseError b) {
          throw new InternalError("NoThanViewInBase.prepareAfterRestoring(): "+b.toString());
      }
      
      if (BaseManagerImpl.getInstance().isOpenACS(acsname) ) {
          // if it is a deleted object from an open ACS, the view opening is processed,
          // but a message is returned
          if (result == null)
              result = "The view target has been deleted in its ACS; it is removed from the view.";
          else
              result = result + " The view target has been deleted in its ACS; it is removed from the view.";
            // remove the target
          m_MoreCriteria_.clear();
          m_LessCriteria_.clear();
      } // if
      else {
            if (acsname == null) {
              throw new DialogError("The view target is from a removed ACS "+
                      ". The view is not opened, and should be removed from the base.");                    
            } // if
            else {
              throw new DialogError("The view target belongs to the closed ACS "+
                      acsname+". The view is not opened."); 
            } // else
      } // else
    } // if (target
    
    // prepare thes criteria which may have a transient metaright to restore
    if (getMoreCriterion() != null) {
        StringRight right = getMoreCriterion();
        if (right instanceof StringRightImpl) {
            ((StringRightImpl)right).meta_ = BaseManagerImpl.getInstance().
                    getMetaRight(((StringRightImpl)right).metaRight_);
        } // if
        right = null;
    } // if
    if (getLessCriterion() != null) {
        StringRight right = getLessCriterion();
        if (right instanceof StringRightImpl) {
            ((StringRightImpl)right).meta_ = BaseManagerImpl.getInstance().
                    getMetaRight(((StringRightImpl)right).metaRight_);
        } // if
        right = null;
    } // if
    
    // prepares the excluded sources from l_excludedSourceNames_
    l_excludedSources_ = new ArrayList(INITIAL_CAPACITY_5);
    for (Iterator ite = l_excludedSourceNames_.iterator(); ite.hasNext(); ) {
        name = (ImmutableName)ite.next();
        try {
            obj = BaseUtilityImpl.getEorBaseObject(name);
            if (!(obj instanceof EligibleParty)) {
              if (result == null)
                  result = "The excluded source "+name+" has been deleted in its ACS;"+
                          " it is removed from the view.";
              else
                  result = result + "The excluded source "+name+" has been deleted in its ACS;"+
                          " it is removed from the view.";
              continue;
            }
            l_excludedSources_.add((EligibleParty)obj);
        } catch (BaseError b) {
          acsname = BaseUtilityImpl.getFromKownACS(name);
          
          if (BaseManagerImpl.getInstance().isOpenACS(acsname) ) {
              // if it is a deleted object from an open ACS, the view opening is processed,
              // but a message is returned
              if (result == null)
                  result = "The excluded source "+name+" has been deleted in its ACS;"+
                          " it is removed from the view.";
              else
                  result = result + "The excluded source "+name+" has been deleted in its ACS;"+
                          " it is removed from the view.";
                // remove the excluded ep
              ite.remove();
          } // if
          else {
                if (acsname == null) {
                  throw new DialogError("An excluded source is from a removed ACS."+
                          " The view is not opened, and should be removed from the base.");                    
                } // if
                else {
                  throw new DialogError("An excluded source is from the closed ACS "+
                          acsname+". This ACS should be opened first.");
                } // else
          } // else
        } // catch
        
    } // for ite 
    
    // sets l_checkingPerimeter_ and l_reducedSourcesInViewSources_
    l_checkingPerimeter_ = new ArrayList(INITIAL_CAPACITY_5);
    l_reducedSourcesInViewSources_ = new ArrayList(INITIAL_CAPACITY_5);
    setUpContextLists(); // fire 'CheckingPerimeter', 'ReducedEPsInViewEPs', 'ObjectChangesInListeningPerimeter'
    
     // registers as listener on events, to complement super.addOwnListening(false)
     // called by super.prepareAfterRestoring().
    addOwnListening(false, false);
    
    obj = null;   
    return result;    
  }  // prepareAfterRestoring()  

  /**
   * This method updates l_checkingPerimeter_ and l_reducedSourcesInViewSources_,
   * which are transient lists.
   * The checking perimeter includes the view EPs and the listening perimeter. 
   * The associates of an Eligible Party are defined as their direct/indirect members 
   * and their direct/indirect running-under actors.
   * This listening perimeter is get by the subtraction of l_excludedSources_ and their associates
   * from the view EPs list (also called 'basic rights users') and their associates.
   * For instance, if the group A is a view EP, and has only group B and group C has direct members,
   * and if the group C is a view excluded EP, the listening perimeter contains only
   * the group B and its associates, but not an associate which is also a C associate. <P>
   * Called by prepareAfterRestoring(), add/removeEorExcludedEligibleParty(),
   * add/removeEorRemove/EligibleParty().
   * Call BaseUtilityImpl.getAllIndirectGroupMembersAndRunningUnderActors().
   * Fire the events 'CheckingPerimeter' and 'ReducedEPsInViewEPs' for the beamer,
   * and 'ObjectChangesInListeningPerimeter' which sends only the deprecated objects
   * and the new objects to listen, and this last event is listened 
   * by Gui2.GraphicViewBaseListenerImpl.
   */
  protected void setUpContextLists() {
    // variables
    Set                   set = null;
    ImmutableSource       ep = null;
    GroupID               grp = null;
    UserID                usr = null;
    Actor[]               l_act = null;
    boolean               removed = false;
    List<ImmutableSource>   l_oldPer = new ArrayList(l_checkingPerimeter_);
    List<ImmutableSource>   l_oldRedEP = new ArrayList(l_reducedSourcesInViewSources_);
    List<ImmutableSource>   l_newObjects = new ArrayList(INITIAL_CAPACITY);
    List<ImmutableSource>   l_oldObjects = new ArrayList(INITIAL_CAPACITY);
    Set<ImmutableSource>    l_allExcludedEPsAndAssociates = new HashSet(INITIAL_CAPACITY);

    l_checkingPerimeter_.clear();
    l_reducedSourcesInViewSources_.clear();

    // search all the associates of the excluded EPs to fill l_allExcludedEPsAndAssociates
    l_allExcludedEPsAndAssociates.addAll(l_excludedSources_);
    for (Iterator ite = l_excludedSources_.iterator(); ite.hasNext(); ) {
      ep = (ImmutableSource)ite.next();

      if (ep instanceof UserID) {
          usr = (UserID) ep;
          l_act = usr.getEorL_UserIDCurrentActors();
          for (int i = 0; i < l_act.length; i++) {
              l_allExcludedEPsAndAssociates.add(l_act[i]);
          } // for i
      }
      else if (ep instanceof GroupID) {
          grp = (GroupID) ep;
          l_allExcludedEPsAndAssociates.addAll(
                  BaseUtilityImpl.getAllIndirectGroupMembersAndRunningUnderActors(grp));
      } // else if GroupID    
    } // for ite
    
    // fills l_checkingPerimeter_ with the EPs which are direct/indirect members
    //  or direct/indirect runnning-under actors, if they are not excluded EPs
    // or associates
    for (Iterator ite = m_BaseObjects_.values().iterator(); ite.hasNext(); ) {
      ep = (ImmutableSource)ite.next();
      removed = false;

      l_checkingPerimeter_.add(ep);

      if (ep instanceof UserID) {
          usr = (UserID) ep;
          l_act = usr.getEorL_UserIDCurrentActors();
          for (int i = 0; i < l_act.length; i++) {
              if (!l_allExcludedEPsAndAssociates.contains(l_act[i]))
                  l_checkingPerimeter_.add(l_act[i]);
              else if (!l_reducedSourcesInViewSources_.contains(ep))
                  l_reducedSourcesInViewSources_.add(ep);
          } // for i
      }

      else if (ep instanceof GroupID) {
          grp = (GroupID) ep;
          l_act = grp.getEorL_GroupIDCurrentActors();
          for (int i = 0; i < l_act.length; i++) {
              if (!l_allExcludedEPsAndAssociates.contains(l_act[i]))
                  l_checkingPerimeter_.add(l_act[i]);
              else if (!l_reducedSourcesInViewSources_.contains(ep))
                  l_reducedSourcesInViewSources_.add(ep);
          } // for i  
          l_act = grp.getEorL_GroupIDCurrentActorsAsSecondaryGroup();
          for (int i = 0; i < l_act.length; i++) {
              if (!l_allExcludedEPsAndAssociates.contains(l_act[i]))
                  l_checkingPerimeter_.add(l_act[i]);
              else if (!l_reducedSourcesInViewSources_.contains(ep))
                  l_reducedSourcesInViewSources_.add(ep);
          } // for i 
          set = BaseUtilityImpl.getAllIndirectGroupMembersAndRunningUnderActors(grp);
          removed = set.removeAll(l_allExcludedEPsAndAssociates);
          l_checkingPerimeter_.addAll(set);
          if (removed && !l_reducedSourcesInViewSources_.contains(ep))
                  l_reducedSourcesInViewSources_.add(ep);
      } // else if GroupID
    } // for ite

    // fill l_newObjects and l_oldObjects
    for (Iterator ite = l_checkingPerimeter_.iterator(); ite.hasNext(); ) {
        ep = (ImmutableSource)ite.next();
        if (!m_BaseObjects_.values().contains(ep) && !l_oldPer.contains(ep)) {
            l_newObjects.add(ep);
        } // if
    } // for ite

    for (Iterator ite = l_oldPer.iterator(); ite.hasNext(); ) {
        ep = (ImmutableSource)ite.next();
        if (!m_BaseObjects_.values().contains(ep) && !l_checkingPerimeter_.contains(ep)) {
            l_oldObjects.add(ep);
        } // if
    } // for ite

    try {
        firePropertyChange("l_CheckingPerimeter", l_oldPer,
           ((ArrayList)l_checkingPerimeter_).clone());
        firePropertyChange("l_ReducedEPsInViewEPs", l_oldRedEP,
           ((ArrayList)l_reducedSourcesInViewSources_).clone());

        if (!l_oldObjects.isEmpty() || !l_newObjects.isEmpty())
                // for GraphicViewBaseListenerImpl
                firePropertyChange("l_ObjectChangesInListeningPerimeter", 
                        l_oldObjects, l_newObjects);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From NoThanViewInBaseImpl.setUpContextLists(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    set = null;
    ep = null;
    grp = null;
    usr = null;
    l_act = null;
    l_oldPer = null;
    l_oldRedEP = null;
    l_newObjects = null;
    l_oldObjects = null;
    l_allExcludedEPsAndAssociates = null;
  } // setUpContextLists()
  
} // NoThanViewInBaseImpl

