//Title:        Access Road
//Version:      0.7.0
//Copyright:    2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;
import ARoad0.Pattern.BeanInfoPattern;

import java.util.Set;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;

import java.io.Serializable;
import java.io.ObjectStreamField;
import java.lang.CloneNotSupportedException;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

 /**
 * This abstract class is used by its subclasses for creating the Privileges.
 * A privilege instance is controlled by an ACS, and delivers rights
 * applicable to couples (source or sources set or source containers set, 
 * target or targets set). The rights of a privilege are immutable
 * after its creation, while the AclEntry has immutable source and target,
 * but with editable rights after the creation. The objects in the couple may be external
 * sometimes. In other words, they do not belong to the privilege ACS.
 * If the ACS constraints define no rights, the privilege cannot have any effective right.
 * These ACS constraints depend on the privilege main type.
 * <P>
 * There are 7 main types of privilege defined in this class,
 * and two interfaces which are implemented by two classes:
 * - PrivilegeForType interface, where the source and target types are the main criteria
 *     to apply the privilege rights to sources and targets which are not known by the privilege,
 *     with the four types TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC,
 *     and TYPED_FOR_SEED.
 *     TYPED_CLASSED privilege is a TYPED one with contraints on source and target classes. 
 *     TYPED_CLASSED_SPECIFIC is a TYPED_CLASSED privilege for designated source or target.
 *     TYPED_FOR_SEED is a TYPED privilege reserved for Authorization (or Rights) Servers.
 * - PrivilegeForLinks interface, where the dynamic links to the source and the target
 *     are the single criteria to apply the privilege, which then knows to which
 *     entities it is applied, with the three types LINKED, LINKED_IF_ALL_SOURCES and
 *     LINKED_FOR_ONE_TO_ONE. The access sources are EligibleParties, and the access targets
 *     are Resources or VirtualFolders.
 *     The LINKED type allows to define in one privilege, a immutable set of rights,
 *     and each EligibleParty which is source have all the rights on all the targets.
 *     With LINKED_IF_ALL_SOURCES, the true access sources are not the linked sources,
 *     since they must be EligibleParties belonging to ALL the set of source containers
 *    (GroupIDMembers or DirectoryEPs) which are the authorized linked sources.
 *     The LINKED_FOR_ONE_TO_ONE type is very like an AclEntry, but with static rights,
 *     for GroupIDs only as sources, but for Resources, EP and Resource VirtualFolders
 *     as targets (VirtualFolders are not used by AclEntry).<P>
 * <P> 
 * Directory and Resource may manage container-for-target privilege inheritance
 * from its parent and to its children, but this is provided only for PrivilegeForLinks
 * having the types LINKED and LINKED_FOR_ONE_TO_ONE, and only one target is then possible.
 * At the source side, the container-for-source principle is an inner principle
 * in GroupID to deliver rights to its members. The container-for-source principle
 * is also managed in DirectoryEP (future version) to deliver rights to its children.
 * <P>
 * A Privilege instance is always dedicated to one ACS.
 * There is a 'See why' description to complement the comment.
 * For PrivilegeForTypes, sources and target may be any ACSObject, to allow to
 * give a privilege to a Directory, and to allow certificate seeds in Authorization Servers.
 * For most of the PrivilegeForLinks, the source is any EligibleParty, but only
 * GroupIDMembers are allowed for the LINKED_IF_ALL_SOURCES type, and only
 * GroupIDs are allowed for the LINKED_FOR_ONE_TO_ONE type.
 * For PrivilegeForLinks, access targets are only Resources or VirtualFolders.
 * <P>
 * The privileges may be applicable to external objects (that is, other ACS than the Privilege ACS)
 * for LINKED_FOR_ONE_TO_ONE PrivilegeForLinks, and they are always applicable for
 * the TYPED_FOR_SEED PrivilegeForTypes. In the other cases,
 * privileges are limited to their own ACS objects.
 * The copy of PrivilegeForType is for the same ACS, and the copy
 * of PrivilegeForLinks may be for another ACS.
 * The TYPED_FOR_SEED PrivilegeForTypes are the only privileges which
 * are not registered in their sources and targets.
 * The external LINKED_FOR_ONE_TO_ONE PrivilegeForLinks change their detailled names
 * when the state external/internal changes. An ACS saving produces the deleting
 * of the external LINKED_FOR_ONE_TO_ONE PrivilegeForLinks to be replaced, at the
 * next ACS restoring, by some true copies.
 * An ACS cannot manage both the external AclEntries and the external linked Privileges.
 * <P>
 * All the typed privileges have a detailled name which starts with 'TYPED'.
 * All the linked privileges have a detailled name which starts with 'LINKED',
 * but after the ACS name if it is an external linked privilege.
 * Starting with 'TYPED' is detected by Gui2.ExplorerTreeCellRenderer.getTreeCellRendererComponent().
 * Each main type may be used to set some constraints in the ACS on the
 * allowed rights. For instance, 'Privilege.PrivilegeRights.TYPED' is
 * a possible key for a standard rights policy in the ACS. It is also possible to
 * use specialized rights and to select them from the type, exactly like a Resource
 * or an EligibleParty.
 * <P>
 * It is a javabean with the following bound properties: <P>
 * - 'Comment' <P>
 * - 'SeeWhy' is an optional description, usually longer than the comment <P>
 * - 'SecondType' <P>
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched,
 * and a dialog box is displayed to inform the user.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBaseInterface.Privilege
 */
public abstract class PrivilegeAbst implements Serializable, Privilege {

  // private static and protected attributes
  private static final long       serialVersionUID = 0;
  protected static final int      PRIME = 1000003;
  protected static final int      INITIAL_CAPACITY = 10; // the default value: 101 for map
  protected static final int      INCREMENT_CAPACITY = 5; // for ArrayList
  
  /*
   * Used in the privilege key, following the implementation class and its attributes.
   **/
  public static final String      TYPED = "TYPED";
  /*
   * Used in the privilege key, following the implementation class and its attributes.
   **/
  public static final String      TYPED_CLASSED = "TYPED_classed";
  /*
   * Used in the privilege key, following the implementation class and its attributes.
   **/
  public static final String      TYPED_CLASSED_SPECIFIC = "TYPED_specific";
  /*
   * Used in the privilege key, following the implementation class and its attributes.
   **/
  public static final String      TYPED_FOR_SEED = "TYPED_certificate";
    /*
   * Used in the privilege key, following the implementation class and its attributes.
   **/
  public static final String      LINKED = "LINKED";
  /*
   * Used in the privilege key, following the implementation class and its attributes.
   **/
  public static final String      LINKED_IF_ALL_SOURCES = "LINKED_sourced";
    /*
   * Used in the privilege key, following the implementation class and its attributes.
   **/
  public static final String      LINKED_FOR_ONE_TO_ONE = "LINKED_one";
  
  /*
   * List of the privilege types. Used in buildUpDetailledName().
   **/
  protected static final List   L_TYPES = new ArrayList(INITIAL_CAPACITY);
  
  static {
        L_TYPES.add(TYPED);
        L_TYPES.add(TYPED_CLASSED);
        L_TYPES.add(TYPED_CLASSED_SPECIFIC);
        L_TYPES.add(TYPED_FOR_SEED);
        L_TYPES.add(LINKED);
        L_TYPES.add(LINKED_IF_ALL_SOURCES);
        L_TYPES.add(LINKED_FOR_ONE_TO_ONE);
  }

  // protected attributes
   /**
    * The ACS which manages the privilege.
    */
    protected ImmutableACS aCS_;  // external object reference
    /**
     * The acs name, wich is used in clone()
     */
    protected ImmutableName acsName_;
   /**
    * The privilege key is built up with the format:
    * privilegeType >> nickName || grant/deny.
    */
    protected String detailledName_ = null;
    /**
    * Short privilege description displayed to the user in the beamer
    */
    protected String nickName_ = null;
   /**
    * Long description to understand the privilege
    */
    protected String seeWhy_ = null;
   /**
    * Privilege comment updated by the user
    */
    protected String comment_ = null;
    /**
    * Positive or negative rights set for the privilege
    */
    protected StringRight[] l_rights_ = null;  
   /**
    * The possible values of the main type are in L_TYPES.
    */
    protected String mainType_ = null;
   /*
    * True if the privilege grants access, and false if it denies it
    */
    protected boolean positiveRight_ = true;
   /**
    * The possible values of the second type are defined by the ACS.
    * The default value is '<undefined>'.
    */
    protected String secondType_ = "<undefined>";
    
    /**
     * The empty instance is an empty PrivilegeForType
     */
    public static final Privilege EMPTY_INSTANCE = new PrivilegeForTypeImpl();

  // -------- transient properties -------------------------------

   /**
   * manages all the property change listeners of this classes and its subclasses
   */
  protected transient PropertyChangeSupport changeSupport_ = null;
  
  /** variable for the JDK 2 serialization
  * @serial ("aCS_",ACS.class),
  * ("acsName_",ImmutableName.class),
  * ("detailledName_",String.class),
  * ("nickName_",String.class),
  * ("seeWhy_",String.class),
  * ("l_rights_",StringRight[].class),
  * ("comment_",String.class),
  * ("mainType_",String.class),
  * ("positiveRight_",Boolean.TYPE),
  * ("secondType_",String.class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("aCS_",ACS.class),
      new ObjectStreamField("acsName_",ImmutableName.class),
      new ObjectStreamField("detailledName_",String.class),
      new ObjectStreamField("nickName_",String.class),
      new ObjectStreamField("seeWhy_",String.class),
      new ObjectStreamField("l_rights_",StringRight[].class),
      new ObjectStreamField("comment_",String.class),
      new ObjectStreamField("mainType_",String.class),
      new ObjectStreamField("positiveRight_",Boolean.TYPE),
      new ObjectStreamField("secondType_",String.class)
  };

 //
 // -------------------------------- Constructors -----------------------------
 //

    /**
     * Constructor to use only as a transient value for initialization, or
     * used from ARoad0.gBase.PrivilegeImplBeanInfo.getPropertyDescriptors()
     * in ARoad0.Gui1.CollectionPropertyEditor.getValue().
     */
    public PrivilegeAbst() {
    }

  /**
   * This is the only way to create an instance.
   * Used by the subclasses, which have to check the unicity of the detailled name.
   * @param _aCS is the ACS which owns this Privilege.
   *   Non-null External Object Reference.
   *   This instance is not registered in the ACS; it is done
   *   by subclasses.
   * @param _nickName is the short privilege name. Cannot be null.
   * It may be the StringRight.getRight() value of the main right in the privilege.
   * @param _mainType must be one of the PrivilegeAbst constant
   * @param _l_rights is a StringRight array of privilege rights.
   *  May be null, but a null value is not allowed at the first position.
   *  Creates an empty array with a length of 0 if the argument is null.
   *  All rights have to be all positive or all negative.
   * @param _secondType second type which drives the rights and the immutability,
   * just like a Basic type. If null, the value is set to "&lt;undefined&gt;".
   * @exception CreateError
   *   if the rights or the privilege type is not correct,
   *   or if the nick name contains '::' or '&gt;&gt;', contains a character that is not a letter,
   *   if a null right or a right which is not authorized.
   * @exception InternalError if the ACS does not manage the privilege rights,
   *   or the required privilege type.
   */
   PrivilegeAbst(ACS _aCS, String _nickName, String _mainType,
           StringRight[] _l_rights, String _secondType) throws CreateError {

        if (_aCS == null || _nickName == null || _mainType == null) {
              throw new CreateError(
                " Null argument which is not authorized.");
        }
        if (!L_TYPES.contains(_mainType))
           throw new CreateError(
              " The privilege type is not allowed.");
        if (
          !_aCS.getManagePrivilegeRights()
        ) {
          throw new InternalError(" This ACS does not manage the privilege rights.");
        }

        if (
          LINKED.equals(_mainType) && !_aCS.getManageSimpleLinkedPrivileges()
        ) {
          throw new InternalError(" This ACS does not manage the simple linked privileges.");
        }

        if (
          LINKED_IF_ALL_SOURCES.equals(_mainType) && !_aCS.getManageIfAllSourceLinkedPrivileges()
        ) {
          throw new InternalError(" This ACS does not manage the If-All-sources linked privileges.");
        }

        if (
          LINKED_FOR_ONE_TO_ONE.equals(_mainType) && !_aCS.getManageInternalForOneToOneLinkedPrivileges() &&
          !_aCS.getManageExternalForOneToOneLinkedPrivileges()
        ) {
          throw new InternalError(" This ACS does not manage the One-To-One linked privileges.");
        }

        if (
          TYPED.equals(_mainType) && !_aCS.getManageSimpleTypedPrivileges()
        ) {
          throw new InternalError(" This ACS does not manage the typed privileges.");
        }

        if (
          TYPED_CLASSED.equals(_mainType) && !_aCS.getManageClassedTypedPrivileges()
        ) {
          throw new InternalError(" This ACS does not manage the classed typed privileges.");
        }

        if (
          TYPED_CLASSED_SPECIFIC.equals(_mainType) && !_aCS.getManageClassedSpecificTypedPrivileges()
        ) {
          throw new InternalError(" This ACS does not manage the classed "+
                  "specific typed privileges.");
        }

        if (
          TYPED_FOR_SEED.equals(_mainType) && !_aCS.getManageAuthorizationServerTypedPrivileges()
        ) {
          throw new InternalError(" This ACS does not manage the authorization server"+
                  " typed privileges.");
        }

        char[] l_chars = _nickName.toCharArray();
        boolean containsFaultyValue = false;
        for (int i = 0; i < l_chars.length; i++) {
            if (Character.isIdentifierIgnorable(l_chars[i]) ||
             Character.isISOControl(l_chars[i])) {
                  containsFaultyValue = true;
                  System.out.println("in the PrivilegeAbst constructor, a character "+
                        "is ignorable or is an ISO control: "+l_chars[i]+"end.");
                  break;
            } // if
        } // for i
        l_chars = null;
        if (_nickName.indexOf("::") != -1 || _nickName.indexOf(">>") != -1)
          containsFaultyValue = true;
        if (containsFaultyValue) {
            throw new CreateError(" The nick name contains a character "+
              "that is not a letter, or contains '::' or '>>'.");
        } // if

        aCS_ = _aCS;
        acsName_ = _aCS.getName();
        nickName_ = _nickName;
        comment_ = "";
        mainType_ = _mainType;

        try {
            if (_secondType != null)
                setSecondType(_secondType);
            else
                setSecondType("<undefined>");
        }
        catch (UpDateError u) {
            aCS_ = null;
            nickName_ = null;
            detailledName_ = null;
            l_rights_ = null;
            secondType_ = null;
            throw new CreateError(u.toString());
        }
        
        try {
            // also sets positiveRight_
            setRights(_l_rights);
        } catch (CreateError i) {
            aCS_ = null;
            nickName_ = null;
            detailledName_ = null;
            l_rights_ = null;
            secondType_ = null;
            throw new CreateError(" Null right, a right is positive and another one is negative," +
                " or a right is not authorized. "+i.toString());
        } // catch

        detailledName_ = buildUpDetailledName(mainType_, nickName_, l_rights_[0].getRight(),
                null, null);
        
  } // PrivilegeAbst()

 //
 // -------------------------- static methods --------------------------------
 //

   /**
    * This method allows to create the detailled name, unique in the ACS.
    * Used in the Privilege implementation constructor.
    * The format of the returned value is as following: <P>
    * privilegeType || nickName || first right <P>
    * Called by the constructors, PrivilegeForType.getDetailledNameFromNameForType()
    * and PrivilegeForLinks.getDetailledNameFromNameForLinks().
    * Overridden by PrivilegeForTypeImpl and PrivilegeForLinksImpl.
    * @param _mainType main type of the privilege. Never null.
    * @param _nickName nick name of the privilege. Never null.
    * @param _firstRight first right of the privilege. Never null.
    * @param _sourceType source type of the privilege. Used only in PrivilegeForTypeImpl
    * which overrides this method, and unused otherwise.
    * @param _targetType target type of the privilege. Used only in PrivilegeForTypeImpl
    * which overrides this method, and unused otherwise.
    * @return the privilege detailled name, even if all the parameters are null.
    * @exception CreateError if '||', '>>' or '::' is in _nickName
    */
    protected static String buildUpDetailledName(String _mainType, String _nickName,
            String _firstRight, String _sourceType, String _targetType) throws CreateError {

        if (_nickName.indexOf("||") != -1 || _nickName.indexOf(">>") != -1 ||
                _nickName.indexOf("::") != -1)
           throw new CreateError(" The nick name description contains "+
                   "a forbidden pattern ('||', '>>' or '::')");

        StringBuilder key = new StringBuilder();

        key.append(_mainType);
        key.append(" || ");
        key.append(_nickName);
        key.append(" || ");
        key.append(_firstRight);

        return key.toString();
    } // buildUpDetailledName()

  /**
   * Gets the immutable main types, not the second types which depends on the ACS.
   * Called by BaseUtilityImpl.getL_AcsRestrictedRights().
  * @return the 7 privilege main types.
  */
  public static final List<String> getL_PrivilegeTypes() {
    return L_TYPES;
  }

 //
 //  --------------------------- interface methods ---------------------------
 //
  
  /**
  * Gets the ACS from the privilege name.
  * @param _name a privilege name
  * @return the opened ACS.
  * @exception BaseError if the argument is not a privilege name, if
  * it is a closed ACS or if it is unknown
  */
  public ImmutableACS getAcsFromName(ImmutableName _name) throws BaseError {
    ImmutableACS    acs = null;
    ImmutableName   acsName = null;
    int             typeIndex = 0;
    Object[]        l_commponents = _name.immutableComponents().toArray();

    for (int i = 0 ; i < l_commponents.length ; i++) {
      if ( l_commponents[i].equals("|| PRIV_TYPE") ) {
          typeIndex = i ;
          break;
      }
    } // for
    if (typeIndex == 0)
        throw new BaseError(" The name does not include the component '|| PRIV_TYPE'.");

    acsName = _name.subname(1, typeIndex);

    // controls the ACS
    acs = (ACS) BaseManagerImpl.getInstance().getEorM_OpenACS().get(acsName);
    if (acs == null)
        throw new BaseError(" The privilege name "+
            " is a closed ACS or it is unknown.");
    
    acsName = null;
    l_commponents = null;
    return acs;
  } // getAcsFromName(


  /**
  * Gets the detailled name from the privilege name.
  * The format of the returned value depends on the privilege, since a typed
  * privilege and a linked privilege have not the same format for their detailled names.
  * Calls PrivilegeForTypeImpl.getDetailledNameFromNameForType() and
  * PrivilegeForLinksImpl.getDetailledNameFromNameForLinks().
  * Called by ACS_Facade.getPrivilege().
  * @see ARoad0.gBase.PrivilegeAbst#buildUpDetailledName
  * @param _name name of a typed or a linked privilege
  * @return the detailled name of the privilege
  * @exception BaseError if the argument has a wrong format
  */
  public String getDetailledNameFromName(ImmutableName _name) throws BaseError {
    if (_name == null || _name.noOfComponents() < 9)
        throw new BaseError(
            " The name is null, or it is too short to have the right format'.");

      // initialization
    int             typeIndex = -1;
    Object[]        l_commponents = _name.immutableComponents().toArray();
    String          result = null;

    for (int i = 0 ; i < l_commponents.length ; i++) {
      if ( l_commponents[i].equals("|| PRIV_TYPE") ) {
          typeIndex = i ;
          break;
      }
    } // for
    if (typeIndex == -1)
        throw new BaseError(
            " The name does not include the component '|| PRIV_TYPE'.");

    if (((String)l_commponents[typeIndex + 1]).indexOf("TYPED") != -1)
            result = PrivilegeForTypeImpl.getDetailledNameFromNameForType(
                    _name, typeIndex);
    else
        result = PrivilegeForLinksImpl.getDetailledNameFromNameForLinks(
                _name, typeIndex);

    l_commponents = null;
    return result;
  } // getDetailledNameFromName()

  /**
  * @return the ACS name for the acs that owns this object, even if the acs
  *    is closed.
  */
  public ImmutableName getAcsName() {
    return acsName_;
  }

  /**
   * This short name is NOT unique for the Access Road program nor the ACS if
   * it is an ACS object.
   * @return the nick name. Never null.
   */
    public String getNickName() {
       return nickName_;
    }

  /**
  * This method returns the detailled name for the explorer,
  * which gives the key components of the object.
  * The formats of the returned value are: <P>
  * privilegeType || nickName || grant/deny <P>
  * or, for a PrivilegeForType: <P>
  * privilegeType || nickName || first right || source type >> target type <P>
  * Furthermore, an external privilege starts with 'ACS: acs_name | '.
  * @see #getKeyPropertiesFromDetailledName
  * @return the detailled name. May be null.
  */
  public String getDetailledName() {
      return detailledName_;
  } // getDetailledName()

  /**
  * This method returns the previous detailled name for the explorer,
  * which gives the key components of the object.
  * An external object has a detailled name starting with 'ACS: acs_name | '.
  * When the state of external changes, this method allows to get the previous detailled name,
  * with or without the ACS name. This is used only by One-to-One PrivilegeForLinks
  * when the single source or the single target are updated with an external object.
  * Otherwise, the previous detailled name is the detailled name.
  * @return the detailled name if it has not changed, or the other version of the detailled name
  */
  public String getPreviousDetailledName() {
      return detailledName_;
  }

  /**
    * The main type has several uses: <P>
    * - to define the general behavior of the privilege, <P>
    * - to define the authorized privilege standard rights. <P>
   * @return the privilege type. As a PrivilegeAbst constant, the
   *   value may be TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC, TYPED_FOR_SEED,
   *   LINKED, LINKED_IF_ALL_SOURCE, LINKED_FOR_ONE_TO_ONE
   */
  public final String getMainType() {
          return mainType_;
  }
  
    /**
    * The second type is defined at the creation of the object, then it may be
    * updated by the user. It has several uses: <P>
    * - to define the type updating policy from the ACS, <P>
    * - to define the authorized privilege specialized rights, <P>
    * - to manage the capacity to delete the instance. <P>
    * This is not the main type, which is get by getMainType().
    * @return second type, as for exemple "read" or "role one". Not null.
    */
    public String getSecondType() {
        return secondType_;
    }
 
  /**
   * If the new value is allowed by the ACS, sets the second type of this privilege.
   * This method is used to switch from an 'immutable' type which forbids
   * the updates of a protected property, and to return to this protecting type.
   * Then, it is possible to switch for instance from 'read' to
   * ''immutable' read', but not to ''immutable' write'.
   * From the default value 'undefined', any first update may be done. 
   * '12345678901234567890' is a program-reserved value. Even if not
   * allowed by the ACS, '12345678901234567890' is the only new value 
   * for which a current type containing 'byAcsCreator' may be changed.
   * From '12345678901234567890', any new value may be set from the allowed ones,
   * and this starting value is the only one which allows to set a value containing
   * 'byAcsCreator'. Calls ACS.getM_TypesAndGUIPolicy() to have the allowed values
   * for the key 'PrivilegeForType.SecondType' if the instance is a PrivilegeForType,
   * or 'PrivilegeForLinks.SecondType' if the instance is a PrivilegeForLinks.
   * Called by the constructor and the GUI. Fires the event 'SecondType'.
   * @param _st second type of the privilege. It is not the main type.
   * @exception UpDateError if the type is null, starts with '&lt;byAcsCreator&gt;',
   * or not known by the ACS, or not associated to the previous type.
   */
  public synchronized void setSecondType(String _st) throws UpDateError {
      String[] l_allowed = null;

      if (this instanceof PrivilegeForType)
          l_allowed = (String[])aCS_.
                    getM_TypesAndGUIPolicy().get("PrivilegeForType.SecondType");
      else
          l_allowed = (String[])aCS_.
                    getM_TypesAndGUIPolicy().get("PrivilegeForLinks.SecondType");

                          // for testing and tracking purpose
                     //System.out.println("in PrivilegeAbst.setSecondType(), new SecondType "+
                        //"for Privileges (they are typed OR linked): "+_st);
                     //System.out.println("SecondType for Privileges (they are typed OR linked) from the ACS for "+getName());
                     //for (String st : l_allowed) {
                       //  System.out.println("------------------ "+st);
                     //} // for ite
                     //System.out.println("new Privilege class: "+this.getClass());
      
      if (_st == null ) 
          throw new UpDateError(getName()+" new type is null.");
      else if (!_st.equals("<12345678901234567890>") && l_allowed == null )
          throw new UpDateError(getName()+
                  " new type is forbidden by the ACS which does not allow any value.");
      else if (!_st.equals("<12345678901234567890>") &&
              l_allowed != null && !Arrays.asList(l_allowed).contains(_st) ) 
          throw new UpDateError(getName()+
                  " new type is forbidden by the ACS.");
      l_allowed = null;
      
      if (secondType_.startsWith("<byAcsCreator>") &&
              !_st.equals("<12345678901234567890>")) 
          throw new UpDateError(getName()+" has been created to be immutable;"+
                  " its type cannot be changed.");
          
      else if (!secondType_.equals("<undefined>")) {

          if (_st.startsWith("<byAcsCreator>") &&
                  !secondType_.equals("<12345678901234567890>"))
                throw new UpDateError(getName()+" new type cannot contain '<byAcsCreator>'.");
          if ( secondType_.indexOf("<byAcsCreator>") == -1 &&
                  !secondType_.equals("<12345678901234567890>")) {

              if (!secondType_.contains("<immutable>") &&
                      (!_st.contains("<immutable>") || !_st.contains(secondType_)))
                    throw new UpDateError("The new type is not linked to the previous type,"+
                            " as 'actor' and '<immutable> actor' for instance");
              if (secondType_.contains("<immutable>") && !secondType_.contains(_st))
                    throw new UpDateError(getName()+
                            " new type "+_st+" is not a reduction of the previous one "+secondType_+", "+
                            " as '<immutable> actor' and 'actor'.");

          } // if ( type_.indexOf("<byAcsCreator>") == -1
      } // if (!type_.equals("<undefined>")

      
      String old = secondType_;
      secondType_ = _st;

      try {
          firePropertyChange("SecondType", old, secondType_);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From PrivilegeAbst.setSecondType(), firePropertChange() throws:");
            e.printStackTrace();
      } // catch

                          // for testing and tracking purpose
                     //System.out.println("PrivilegeForType.SecondType from the ACS for "+getName());
                     //for (Iterator ite = Arrays.asList( (String[])aCS_.getM_SpecializedTypesAndExplorer().
                       //     get("PrivilegeForType.SecondType") ).iterator(); ite.hasNext(); ) {
                       //  System.out.println("------------------ "+ite.next());
                     //} // for ite
                     
  } // setSecondType()

    /**
     * @return true if the privilege rights
     *  grant access to the source, and false if the rights
     *  deny an access type.
     */
    public final boolean getPositiveRight() {
        return positiveRight_;
    }
    
   /**
    * Currently unused.
    * @return a long description which is displayed to the user. May be null.
    */
    public final String getSeeWhy() {
        return seeWhy_;
    }

   /**
    * Gets the effective privilege rights.
    * Unworkable for a PrivilegeForType with a wrong initialization.
    * If a PrivilegeForType is classed but with a null source class or target class,
    * returns an empty array.
    * @return the rights. May be an empty array.
    */
    public final StringRight[] getL_Rights() {
        if (this instanceof PrivilegeForType && ((PrivilegeForType)this).getClassed()
            && (((PrivilegeForType)this).getSourceClass() == null ||
                    ((PrivilegeForType)this).getTargetClass() == null))
            return new StringRight[0];
        
        StringRight[] l_rights = new StringRight[l_rights_.length];
        System.arraycopy(l_rights_, 0, l_rights, 0, l_rights_.length);
        return l_rights;
    } // getL_Rights()

  /**
   * Sets the rights array mainly at the instance construction. ACSImpl, when a linked
   * privilege is restored after an ACS opening, calls this method if there are rights
   * which has been deleted by the user before the opening. <P>
   * Controls the rights constraints in the ACS from the privilege main type.
   * If there is a null array of allowed rights, no right is set and there is an exception. <P>
   * Creates a 0-size array if the argument is null.
   * Controls if the rights are allowed by the ACS, and it may depend on the privilege type
   * if they are some matched specialized rights in the ACS. Only non-null, non-empty rights
   * are set as new rights, if they have the sens of the property positiveRight_.
   * If they have another sens, they are simply forgotten without exception.
   * It is not a bound property; no event "PrivilegeRights" firing to the listeners.
   * Sets the positiveRight_ property.
   * Calls BaseUtilityImpl.getL_AcsRestrictedRights() for having the range 
   * of the authorized values. Calls UtilityImpl.selectHeaders().
   * Called only by the constructors.
   * @param _st is a StringRight array of rights
   *    for the couple (resource, eligibleparty) in the Privilege ACS.
   *    The actual list of authorized rights depends on the ACS.
   *    _st cannot be null.
   * @exception CreateError if a right is not allowed by the ACS, or null rights,
   * or if all rights have not the same sens.
   */
    protected synchronized void setRights(StringRight[] _st) throws CreateError {
        
        // controls the privilege rights restriction in the ACS
        StringRight[] l_restrictedRights = null;
        Set           l_newRights = new HashSet(INITIAL_CAPACITY / 2);
        boolean       positive = false;

        try {
            // gets the rights restriction in the ACS.
            // EMPTY_INSTANCE is used to find in its
            // BeanInfo the allowed interfaces, including 'Privilege' which is necessary there
            if (this instanceof PrivilegeForType)
                l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                       (BeanInfoPattern) Class.forName(
                            "ARoad0.gBaseBeanInfo.PrivilegeForTypeImplBeanInfo").newInstance(),
                        PrivilegeForTypeImpl.EMPTY_INSTANCE,
                       "PrivilegeRights", aCS_, mainType_);
            else if (this instanceof PrivilegeForLinks)
                l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                       (BeanInfoPattern) Class.forName(
                            "ARoad0.gBaseBeanInfo.PrivilegeForLinksImplBeanInfo").newInstance(),
                        PrivilegeForLinksImpl.EMPTY_INSTANCE,
                       "PrivilegeRights", aCS_, mainType_);
        } catch (Exception c) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From PrivilegaAbst.setRights(), "+
                    "BaseUtilityImpl.getL_AcsRestrictedRights() throws:");
            c.printStackTrace();
            throw new CreateError(" Error to get the authorized rights by the ACS.");
        } // catch

        if (l_restrictedRights != null) {
            if (_st != null) {
              if (_st[0] == null)
                throw new CreateError("The first right is null.");
              else if (_st[0].isGrantingRight())
                  positive = true;

              for (int i = 0; i < _st.length; i++) {
                if (_st[i] != null && !_st[i].isEmpty()) {
                    if (_st[i].isGrantingRight() != positive)
                        throw new CreateError(
                                " The right at "+i+" has not the sens of the first right");
                    l_newRights.add(_st[i]);
                }
              } // for i
            } // if
            Set l_limitedRights = new HashSet(INITIAL_CAPACITY / 2);
            for (int i = 0; i < l_restrictedRights.length; i++) {
              l_limitedRights.add(l_restrictedRights[i]);
            } // for i

            if (_st != null && !l_newRights.isEmpty() && 
              !l_limitedRights.containsAll(l_newRights)) {
                l_newRights = null;
                l_limitedRights = null;
                l_restrictedRights = null;
                throw new CreateError(
                        " Some of the rights are not authorized by the ACS.");
            } else { // all is ok
                l_limitedRights = null;
                l_restrictedRights = null;
            }
        } // if (l_restrictedRights

        try {
            if (!l_newRights.isEmpty())
                UtilityImpl.getFirstInstance().selectHeaders(l_newRights);
        }
        catch (BaseError b) {
            b.printStackTrace();
            throw new InternalError("PrivilegeAbst.setRight(): "+b.toString());
        }

        // updates the rights
        if (!l_newRights.isEmpty()) {
          l_rights_ = (StringRight[])
                  l_newRights.toArray(new StringRight[l_newRights.size()]);

          positiveRight_ = positive;
        }
        else {
          throw new CreateError(" The rights are null.");
        }

        l_restrictedRights = null;
        l_newRights = null;
    } // setRights()

   /**
    * Currently unused.
    * Sets the description which is displayed to the user.
    * @param _see may be null.
    */
    public final synchronized void setSeeWhy(String _see) {
      String oldSee = seeWhy_;
      seeWhy_ = _see;

      try {
          firePropertyChange("SeeWhy", oldSee, seeWhy_);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From PrivilegeAbst.setSeeWhy(), firePropertChange() throws:");
        e.printStackTrace();
      }
      oldSee = null;
    }

  /**
  * Sets any comment related to this privilege. It is a bound property.
  * @param _s comment which is associated to the privilege.
  */
  public final synchronized void setComment(String _s)	{
      String oldComment = comment_;
      comment_ = _s;

      try {
          firePropertyChange("Comment",oldComment,comment_);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From PrivilegeAbst.setComment(), firePropertChange() throws:");
        e.printStackTrace();
      }
      oldComment = null;
  }

  /**
   * @return the comment which is associated to the privilege.
   */
  public final String getComment()	{
          return comment_;
  }

  /**
   * @return the ACS of this PrivilegeImpl, as an External Object Reference
   */
  public final ImmutableACS getEorACS() {
       return aCS_;
  }
  
  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the 'Comment'
  * and 'SeeWhy' properties. Overidden method.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

                        // for testing and tracking
                        //System.out.println(
                          //      "PrivilegeAbst.addPropertyChangeListener(_propertyName, _l),"+
                            //    "call to "+nickName_+" for the property "+_propertyName+
                              //  " and the listener "+_l);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}
  }

  /**
  * Adds a listener to the bean. Overidden method.
  * @param _l PropertyChangeListener to add
  */
  public void addPropertyChangeListener(PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

                        // for testing and tracking
                        //System.out.println("PrivilegeAbst.addPropertyChangeListener(_l), call "+
                          //"to "+nickName_+" for the listener "+_l);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}
  }

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the 'Comment'
  * and 'SeeWhy' properties. Overidden method.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("PrivilegeImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()

  /**
  * removes a listener to the bean. Overidden method.
  * @param _l PropertyChangeListener to remove
  */
  public synchronized void removePropertyChangeListener(PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("PrivilegeImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_l);}
  }
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("PrivilegeAbst.getPropertyChangeListeners(),
                    // call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      return changeSupport_.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()

  /**
   * Used to clone a Privilege when an acs is closing, or to clone an ACS.
   * Clones the acs name. Shallow copy of the rights array, thus with the rights
   * network of the metarights and the rights network of the acsrights.
   * @return Object without any change in the ACS reference
   *  but copy of this reference.
   */
    public Object clone() {
      try {
          PrivilegeAbst clon = (PrivilegeAbst) super.clone();
          clon.aCS_ = aCS_;
          clon.acsName_ = (ImmutableName)acsName_.clone();
          clon.detailledName_ = detailledName_;
          clon.nickName_ = nickName_;
          clon.seeWhy_ = seeWhy_;
          
          clon.l_rights_ = new StringRight[l_rights_.length];
          System.arraycopy(l_rights_, 0, clon.l_rights_, 0, l_rights_.length);
          
          clon.comment_ = comment_;
          clon.mainType_ = mainType_;
          clon.secondType_ = secondType_;
          clon.positiveRight_ = positiveRight_;
          return clon;

      } catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
  	    if (aCS_ != null) {
  	            aCS_.getEorBaseError().addException(e.toString());
  	    }
        throw new InternalError(e.toString());
      } // catch
    } // clone()

  /**
   * Note: cannot use detailledName_ because it may be changed in PrivilegeForLinksImpl.
   * @return the hashcode value for this object
   */
  public int hashCode() {
    int result = 0;

    if (nickName_ != null) {
      result = PRIME*result + nickName_.hashCode();
    }

    if (acsName_ != null) {
      result = PRIME*result + acsName_.hashCode();
    }
    
    return result;
  } // hashcode()

   /**
   * @param _obj to compare.
   * @return true if _obj is equal with the same class,
   *  not only an implementation of Privilege.
   */
  public synchronized boolean equals(Object _obj) {

      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      PrivilegeAbst obj = (PrivilegeAbst) _obj;

      // aCS_  null if obj is empty, created with "new PrivilegeAbst()"
      if (aCS_ == null || obj.getEorACS() == null) {
          if (aCS_ != null || obj.getEorACS() != null)
            return false;  // if
      } // if
      else {
        if (aCS_ != obj.getEorACS() ) return false; }

      //if (detailledName_ == null || obj.detailledName_ == null) {
        //  if (detailledName_ != null || obj.detailledName_ != null)
          //  return false;  // if
      //} // if
      //else {
        //if (!detailledName_.equals( obj.detailledName_ )) return false; }

      if (nickName_ == null || obj.getNickName() == null) {
          if (nickName_ != null || obj.getNickName() != null)
            return false;  // if
      } // if
      else {
        if (!nickName_.equals( obj.getNickName() )) return false; }

      //if (seeWhy_ == null || obj.getSeeWhy() == null) {
        //  if (seeWhy_ != null || obj.getSeeWhy() != null)
          //  return false;  // if
      //} // if
      //else {
        //if (!seeWhy_.equals( obj.getSeeWhy() )) return false; }
          
      if (l_rights_.length != obj.getL_Rights().length )
          return false;
      for (int i = 0; i < l_rights_.length; i++) {
        if (l_rights_[i] == null || obj.getL_Rights()[i] == null) {
          if (l_rights_[i] != null || obj.getL_Rights()[i] != null) {

                    //System.out.println("PrivilegeImpl.equals() on l_rights_[i], with i :" + i );

            return false;}
          continue; // i++
        } // if
        else {
          if (!l_rights_[i].equals( obj.getL_Rights()[i] ))
            return false;
        } // else
      } // for
      
      if (mainType_ == null || obj.getMainType() == null) {
          if (mainType_ != null || obj.getMainType() != null)
            return false;  // if
      } // if
      else {
        if (!mainType_.equals( obj.getMainType() )) return false; }
      
      if (secondType_ == null || obj.getSecondType() == null) {
          if (secondType_ != null || obj.getSecondType() != null)
            return false;  // if
      } // if
      else {
        if (!secondType_.equals( obj.getSecondType() )) return false; }
      
      //if (positiveRight_ != obj.getPositiveRight()) return false;
      
      return true;
  } // equals(Object) 

  /**
   * Based on the main type first, and then on the nick name,
   * if the main types are equals, then
   * on the detailled names at the end.
   * The rights are NOT compared.
   * Called by ACSImpl.newEorPrivilege() (through Arrays.sort()),
   * ACSTreeUtilities.addDetailledNameUnderNode() and by
   * Gui1.CollectionPropertyEditor.propertyChange() to display the Privileges.
   * @param o to compare must be an ImmutableAclEntry
   * @exception ClassCastException - if the specified object's type
   *  prevents it from being compared to this current Object
   * @return 0 if the ACL are almost equals, a negative integer if this object
   * is lesser than the argument, and a positive integer if this object
   * is greater than the argument.
   */
  public int compareTo(Object o) throws ClassCastException {
    if (!(o instanceof ImmutablePrivilege))
      throw new ClassCastException("In PrivilegeAbst.compareTo(), the argument "+
        "is not an instance of ImmutablePrivilege");
    // if (equals(o) ) return 0;
    if (nickName_ == null) throw new ClassCastException("In PrivilegeAbst.compareTo(),"+
      " the object nick name is null");
    int result = mainType_.compareTo( ((ImmutablePrivilege)o).getMainType() );
    if (result != 0) return result;
    result = nickName_.compareTo( ((ImmutablePrivilege)o).getNickName() );
    if (result != 0) return result;
    result = detailledName_.compareTo( ((ImmutablePrivilege)o).getDetailledName() );

                // for testing and tracking
                //System.out.println("PrivilegeAbst.compareTo() result: "+result);
                //System.out.println("----- on the Privilege: "+getName());
                //System.out.println("------- with argument: "+((ImmutablePrivilege)o).getName());

    return result;
  } // compareTo()

  /**
   * Return long value for serialization
   */
  public long getSerialVersionUID()	{
       return serialVersionUID;
  }

  /**
  * Extracts the 3 key properties from the key of an object.
  * From the DetailledName interface.
  * @param _displayedName is the detailled name of a privilege
  * @return an array of 3 names: <P>
  *   - at the index 0, the main type in a one-component name <P>
  *   - at the index 1, the nick name in a one-component name <P>
  *   - at the index 2, the first right name in a one-component name <P>
  * @exception BaseError if the argument is not a valid key,
  *  or if a main component string is not provided by NameImpl.toString().
  */
  public final ImmutableName[] getKeyPropertiesFromDetailledName(
    String _displayedName) throws BaseError {
        
      if (_displayedName.indexOf(" || ") == -1 ||
              _displayedName.lastIndexOf(" || ") == _displayedName.indexOf(" || "))
               throw new BaseError(" The argument is not a valid key.");
        
      // variables
      ImmutableName[]   l_result = new ImmutableName[3];
      int               index = 0;
      
      // extracts the main type
      l_result[0] = NameImpl.getNameFromString(
        _displayedName.substring(0, _displayedName.indexOf(" || ") ) );
      // extracts the nick name
      index = _displayedName.indexOf(" || ");
      l_result[1] = NameImpl.getNameFromString(
        _displayedName.substring(index + 4,
            _displayedName.indexOf(" || ", index + 4) ) );
      // extracts the right
      index = _displayedName.indexOf(" || ", index + 4);
      l_result[2] = new NameImpl(
        _displayedName.substring(index + 4,
            _displayedName.indexOf(" || ", index + 4) ) );

      return l_result;
  } // getKeyPropertiesFromDetailledName()

  /**
  * Extracts the BaseObject name from the detailled name of a Privilege.
  * This is a pseudo-static method, since it does not depend on the instance.
  * Calls ACSCatalog.getOnePrivilegeFromDetailledName().
  * Called by TreeManager.analyzeSelectedNodeInExplorer() and
  * SimplePropertyEditor.viewOneElement().
  * @param _detailledName the viewable name of the object in the explorer
  * @return name of the Privilege, or empty
  */
  public ImmutableName getNameFromDetailledName(ImmutableACS _acs, String _detailledName)
    throws BaseError {
      if (_acs == null || _detailledName == null)
          return NameImpl.EMPTY_INSTANCE;

      Privilege priv = _acs.getOnePrivilegeFromDetailledName(_detailledName);

      if (priv == null) return NameImpl.EMPTY_INSTANCE;
      else return priv.getName();
  } // getNameFromDetailledName()

  /**
   * It sets to null almost all variables, so the instance is equals
   * to a new PrivilegeForTypeImpl().
   * This method removes the associates, then calls finalizeForBase().
   * Called by the GUI. Calls finalizeForBase() and
   * AssociateUtilityImpl.removeAssociatesAfterDeletion() to remove the associates.
   * It is public to let the gWork methods to call it, but is is not described
   * in the BeanInfo since the beamer must use only finalizeForUser() to delete an object.
   * This method does not raise PropertyChangeEvents since the BeanInfo for this class
   * specifies that its ACS informs the listeners when this instance is finalized.
   * To test if an instance has been finalized, uses (aCS_ == null).
   * @exception UpDateError from finalizeFromBase()
   */
  public synchronized void finalizeForUser() throws UpDateError {
    if (aCS_ == null ) return;

    // removes the associates, if any
    try {
        AssociateUtilityImpl.removeAssociatesAfterDeletion(this, (ACS)aCS_);
    } catch (BaseError b) {
        throw new UpDateError("from AssociateUtilityImpl.removeAssociatesAfterDeletion(), "+
                b.toString());
    }
              // for testing and tracking
              //System.out.println("Call of PrivilegeForTypeImpl.finalizeForUser() in: "+getName());

    finalizeForBase();
  } // finalizeForUser()

  /**
   * It sets to null almost all variables, so the instance is equals
   * to a new PrivilegeAbst().
   * The StringRights are not finalized, but the links are cut.
   * The privilege is not unregistered in its ACS.
   * Overridden by all the subclasses.
   * Called by the subclasses methods, including finalizeFromDeserialization().
   * @exception no UpDateError throwing
   */
  public synchronized void finalizeForBase() throws UpDateError {
    if (aCS_ == null ) return;
              // for testing and tracking
              //System.out.println("Call of PrivilegeImpl.finalizeForUser() in: "+getName());

    aCS_ = null;
    acsName_ = null;
    detailledName_ = null;
    nickName_ = null;
    seeWhy_ = null;
    comment_ = "";
    l_rights_ = null;
    mainType_ = null;
    secondType_ = null;
    positiveRight_= true;
  } // finalizeForBase()

  //
  // ---------------- protected methods out of the implemented interfaces -----
  //

  /**
  * Fires an event to every registered listener, in any order.
  * Called by setComment() for the bound properties: <P>
  * - 'Comment', <P>
  * - 'SeeWhy', <P>
  * - 'SecondType', <P>
  * and for all the subclasses property changes.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

      // no action if the transiant variables are null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //System.out.println("PrivilegeAbst.firePropertyChange(), call to "+
                  //nickName_+" for the property "+_propertyName);

      synchronized (changeSupport_)
      {changeSupport_.firePropertyChange(_propertyName, _oldValue, _newValue);}
  } // firePropertyChange()
  
} // PrivilegeAbst