//Title:        Access Road
//Version:      0.7.1
//Copyright:    2006, 2010, 2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls.


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.Arrays;

import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

 /**
 * This class is responsible for modeling the direct links which manages some access rights
 * from a EligibleParty list (or only one GroupID) to a privilege targets list
 * (or only one target). Targets are Resources or VirtualFolders.
 * There are three types of PrivilegeForLinks: LINKED,
 * LINKED_IF_ALL_SOURCES and LINKED_FOR_ONE_TO_ONE.
 * This privilege is usable by any couple (all sources set, one of the target)
 * from the privilege ACS, or from another ACS for an external LINKED_FOR_ONE_TO_ONE privilege.
 * A privilege is external if its source or its target is from an ACS which
 * is not the ACS of the privilege. It is not possible to have a third ACS for the other end.
 * A linked privilege may be inherited from the (in)direct parents of the target,
 * except the LINKED_IF_ALL_SOURCES privilege. Only one target is authorized
 * if the first target is from an ACS that manages the PRI inheritance.
 * <P>
 * Eligible Party and Resource types and classes are never used,
 * on the contrary of PrivilegeForTypes, and the linked sources and targets
 * register and unregister themself dynamically in the Privilege.
 * So, it is possible to modify a Privilege on the fly following
 * a program dynamic rule. The rights of a Privilege are immutable
 * after its creation, while the AclEntry has immutable source and target,
 * but with editable rights after the creation.
 * <P>
 * This is a javabean with the following bound properties: <P>
 * - 'OneLinkedPrivilegeSource' <P>
 * - 'OneLinkedPrivilegeTarget' <P>
 * - 'ExternalPrivilege' <P>
 * There are three types of PrivilegeForLinks: LINKED,
 * LINKED_IF_ALL_SOURCES and LINKED_FOR_ONE_TO_ONE. 
 * For the last one, when it is finalized, "TPrivileges" and "SPrivileges"
 * are thrown by the source and the target after 'OneLinkedPrivilegeSource' and
 * 'OneLinkedPrivilegeTarget'.
 * CAUTION: the gWork package manages only the LINKED PrivilegeForLinks, in this version.
 * <P>
 * The LINKED type allows to define in one privilege, a generic set of rights,
 * and each EligibleParty which is source have all the rights on all the targets.
 * The LINKED privilege is defined by its rights, then in a large ACS, such a privilege
 * may have numerous linked sources or targets.
 * With LINKED_IF_ALL_SOURCES, the rights on the targets are used by any EligibleParty
 * which belongs to ALL the source containers, which are GroupIDMembers or DirectoryEPs.
 * For instance, if there are a GroupID and an UserID as privilege sources,
 * any Actor which is a member of the GroupID AND belongs to the UserID, 
 * has the rights of the privilege.
 * If an actor belongs to the GroupID but is not under the UserID, it has not
 * the privilege. To have an OR relation between the GroupID and the UserID,
 * rather than AND, let's use 2 LINKED_FOR_ONE_TO_ONE privileges or two AclEntries.
 * The conditional AclEntry is an alternative to the LINKED_IF_ALL_SOURCES privilege,
 * but it is focused to deliver rights to one source which has to be member of the
 * condition groups, and the couple (source, target) is immutable for an AclEntry.
 * The number of privilege sources is limited to 4.
 * <P>
 * With LINKED_FOR_ONE_TO_ONE, there is at most one source, and it is an EligibleParty,
 * and at most one PrivilegeTarget. A LINKED_FOR_ONE_TO_ONE privilege is like
 * an ACL from an EP to a resource or a virtual folder. This is the only linked privilege
 * able to be external. A LINKED_FOR_ONE_TO_ONE PrivilegeForLinks is an alternative
 * to an AclEntry. Both may have one external access source or target,
 * where the target is immutable while only the privilege can change its source,
 * it is deleted when its source or its target is deleted but an AclEntry may be conditional,
 * it can change dynamically its rights, and there are other differences as following:<P>
 * - on sources, the AclEntry is as powerful as a LINKED_FOR_ONE_TO_ONE PrivilegeForLinks.<P>
 * - on targets, the AclEntry is less powerful since it cannot be applied to VirtualFolders.
 * In a one-to-one privilege, the target cannot be removed after the first adding,
 * while it is possible for the source.
 * For an external LINKED_FOR_ONE_TO_ONE privilege, one end (the source or the target)
 * must belong to the privilege ACS, while the other end does not. The rights belongs
 * always to the ACS privilege, if they are ACS rights. <P>
 * The format of the detailled name for an internal linked privilege is as following: <P>
 * privilegeType || nickName || first right <P>
 * For an external linked privilege, the format of the detailled name is as following: <P>
 * - ACS: ACS_name | privilegeType || nickName || first right <P>
 * This means the detailled name changes when an empty privilege becomes an external one
 * during the setting of an end by the user. This change is visible in the explorer for instance.
 * The method getPreviousDetailledName returns the old detailled name if it is different.
 * An Eligible Party or PrivilegeTarget name is incomplete if it belongs to
 * the Privilege ACS. Since the source and the target are unknown at the creation
 * of the PrivilegeForLinks, the privilege may become external only after the setting
 * of the first external end, the source or the target. Then, its detailled name
 * changes in the GUI.
 * <P>
 * A specific use of PrivilegeForLinks is to extend the structural relations among
 * the ACS objects, without any change in the types model of AccessControlLink.
 * For example, for a given ACS, it is possible to
 * create privileges linking some pattern-name Directories to each
 * item which is member of them whithout being a child of these Directories.
 * This pattern-name Directory is not a VirtualFolder and allows AclEntries as target.
 * The main feature is then to avoid to code a new link in the ACS addon.
 * <P>
 * To process the inherited PrivilegeForLinks, it is mandatory to code in an AcsAddon
 * how the Directory or VirtualFolder creates and removes the inherited PrivilegeForLinks
 * in its direct children. This may include the choice of the objects that becomes
 * inherited ones, and the definition of an order to select them in each target.
 * The default behavior, in this generic class, is to inherit all the applicable
 * PrivilegeForLinks, and to sum up all the rights.
 * After the inherited object creation to code in an AcsAddon, the right updating
 * of the inherited object is handled by this generic class, at every change on the sources.
 * The target ACS cannot manage the PRI inheritance in a LINKED_IF_ALL_SOURCES privilege.
 * For the other types, only one target is authorized if the first target is from
 * an ACS that manages the PRI inheritance.
 * <P>
 * The privilege key is used by the ACS to find a privilege.
 * The key is the couple (privilege type, nick name).
 * The main privilege value is an unmodifiable list of rights.
 * The privilege has positive or negative rights to grant or deny the access.
 * The rights and the key CANNOT be updated after the privilege creation,
 * nor the rights sens (grant or deny).
 * There is a 'See Why' description to complement the comment.<P>
 * This is a typical sequence of calls, in the inverse order,
 * when the ACS of an external linked privilege is saved: <P>
 * ARoad0.Gui1.CollectionPropertyEditor.propertyChange() <P>
 * at java.beans.PropertyChangeSupport.firePropertyChange(PropertyChangeSupport.java:339) <P>
 * at java.beans.PropertyChangeSupport.firePropertyChange(PropertyChangeSupport.java:347) <P>
 * at java.beans.PropertyChangeSupport.firePropertyChange(PropertyChangeSupport.java:276) <P>
 * at ARoad0.gBase.ACSImpl.firePropertyChange(ACSImpl.java:5454) <P>
 * at ARoad0.gBase.ACSImpl.removeAcsController(ACSImpl.java:7981) <P>
 * at ARoad0.gBase.ACSImpl.removeEorResource(ACSImpl.java:4014) <P>
 * at ARoad0.gBase.PrivilegeForLinksImpl.removeEorPrivilegeTargetLink(PrivilegeForLinksImpl.java:1447) <P>
 * at ARoad0.gBase.CPrivilegeUserImpl.removeEorPrivilegeToTarget(CPrivilegeUserImpl.java:481) <P>
 * at ARoad0.gBase.ResourceImpl.removeEorPrivilegeToTarget(ResourceImpl.java:1999) <P>
 * at ARoad0.gBase.PrivilegeForLinksImpl.finalizeForBase(PrivilegeForLinksImpl.java:1829) <P>
 * at ARoad0.gBase.PrivilegeAbst.finalizeForUser(PrivilegeAbst.java:1281) <P>
 * at ARoad0.gBase.ACSImpl.closeExternalLinkedPrivileges(ACSImpl.java:6327) <P>
 * at ARoad0.gBase.BaseManagerImpl.cutACSExternalLinks(BaseManagerImpl.java:2915) <P>
 * at ARoad0.gBase.BaseManagerImpl.saveOpenACS(BaseManagerImpl.java:2393) <P>
 * at ARoad0.gWork.ACS_Facade.saveACS(ACS_Facade.java:188) <P>
 * at ARoad0.gDMak.ActionSave.save(ActionSave.java:210) <P>
 * at ARoad0.gDMak.ActionSaveAll.saveAll(ActionSaveAll.java:302) <P>
 * at ARoad0.gDMak.ActionSaveAll.actionPerformed(ActionSaveAll.java:87) <P>
 * This class has subclasses in the AcsAddon packages.
 * Most of the variables are protected.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBase.PrivilegeAbst
 * @see ARoad0.gBase.PrivilegeForTypeImpl
 * @version 0.7.1
 */
public class PrivilegeForLinksImpl extends PrivilegeAbst implements PrivilegeForLinks,
    PreviousDetailledName {
   
   /**
    * if true, the instance has the LINKED_IF_ALL_SOURCES type,
    * with only one internal or external GroupID as source,
    * and only one internal or external target.
    */
    protected boolean ifAllSources_;
   /**
    * if true, the instance has the LINKED_FOR_ONE_TO_ONE type,
    * with only one internal or external EligibleParty as source,
    * and only one internal or external target.
    */
    protected boolean oneRoleOneTarget_;
    /**
    * Access sources which are connected to this privilege.
    * Restricted to one item, if oneRoleOneTarget_ is true.
    */
    protected ArrayList<PrivilegeSource> l_sources_;
   /**
    * Access targets are Resources and VirtualFolders connected to this privilege.
    * Restricted to one item, if oneRoleOneTarget_ is true.
    * Updated by the targets.
    */
    protected ArrayList<PrivilegeTarget> l_targets_;
    /**
     * If LINKED_FOR_ONE_TO_ONE, the single source full name, if any.
     */
    protected ImmutableName sourceName_ = null;
    /**
     * If LINKED_FOR_ONE_TO_ONE, the single target full name, if any. 
     */
    protected ImmutableName targetName_ = null;
    /**
     * If LINKED_FOR_ONE_TO_ONE, the single target ACS name, if any. 
     */
    protected ImmutableName nameOfSourceAcs_ = null;
    /**
     * If LINKED_FOR_ONE_TO_ONE, the single target ACS name, if any.
     */
    protected ImmutableName nameOfTargetAcs_ = null;
    /**
     * The source types that the ACS authorizes for this privilege
     */
    protected List<String>  l_internalAllowedSourceTypes_ = new ArrayList(INITIAL_CAPACITY);
    /**
     * The target types that the ACS authorizes for this privilege
     */
    protected List<String>  l_internalAllowedTargetTypes_ = new ArrayList(INITIAL_CAPACITY);
    /**
     * True to return the previous detailled name, after a change of the external
     * source or target.
     */
    protected transient boolean changedDetailledName_ = false;
    
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of the constructor for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final PrivilegeForLinksImpl EMPTY_INSTANCE = new PrivilegeForLinksImpl();
  
  /** variable for the JDK 2 serialization
  * @serial ("ifAllSources_",Boolean.TYPE),
   * ("oneRoleOneTarget_",Boolean.TYPE),
   * ("l_sources_",ArrayList.class),
   * ("l_targets_",ArrayList.class),
   * ("sourceName_",ImmutableName.class),
   * ("targetName_",ImmutableName.class),
   * ("nameOfSourceAcs_",ImmutableName.class),
   * ("nameOfTargetAcs_",ImmutableName.class),
   * ("l_internalAllowedSourceTypes_",ArrayList.class),
   * ("l_internalAllowedTargetTypes_",ArrayList.class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("ifAllSources_",Boolean.TYPE),
      new ObjectStreamField("oneRoleOneTarget_",Boolean.TYPE),
      new ObjectStreamField("l_sources_",ArrayList.class),
      new ObjectStreamField("l_targets_",ArrayList.class),
      new ObjectStreamField("sourceName_",ImmutableName.class),
      new ObjectStreamField("targetName_",ImmutableName.class),
      new ObjectStreamField("nameOfSourceAcs_",ImmutableName.class),
      new ObjectStreamField("nameOfTargetAcs_",ImmutableName.class),
      new ObjectStreamField("l_internalAllowedSourceTypes_",ArrayList.class),
      new ObjectStreamField("l_internalAllowedTargetTypes_",ArrayList.class)
  };

 /*
  * -------------------------------- Constructors -----------------------------
  */

    /**
     * Constructor to use only as a transient value for initialization, or
     * used from ARoad0.gBase.PrivilegeImplBeanInfo.getPropertyDescriptors()
     * in ARoad0.Gui1.CollectionPropertyEditor.getValue().
     */
    public PrivilegeForLinksImpl() {
    }

  /**
   * This is the only way to create a linked privilege instance from the GUI.
   * Calls the second constructor.
   * @param _aCS is the ACS which owns this Privilege.
   *   Non-null external object reference.
   * @param _nickName is the short privilege name. Cannot be null.
   * @param _mainType is equal to PrivilegeAbst.LINKED, LINKED_FOR_ONE_TO_ONE,
   *   or to LINKED_IF_ALL_SOURCES
   * @param _l_rights is a StringRight array of privilege rights.
   *  May be null, but a null value is not allowed at the first position.
   *  Creates an empty array with a length of 0 if the argument is null.
   *  All rights have to be all positive or all negative.
   * @param _secondType second type which drives the rights and the immutability,
   * just like a Basic type. If null, the internal value is set to '&lt;undefined&gt;'.
   * @exception CreateError if this privilege already exists
   *   (_aCS.newEorPrivilege() throws it), if the rights or the
   *   privilege type is not correct, if _nickName contains '::' or '>>',
   *   or if a parameter is null.
   */
   public PrivilegeForLinksImpl( 
      ACS _aCS, String _nickName, String _mainType, StringRight[] _l_rights,
      String _secondType) throws CreateError {

        this(_aCS, _nickName, _mainType, _l_rights, _secondType, true);
        
  } // PrivilegeForLinksImpl()

  /**
   * This is the only way to create a linked privilege instance.
   * @param _aCS is the ACS which owns this Privilege.
   *   Non-null external object reference.
   * @param _nickName is the short privilege name. Cannot be null.
   * @param _mainType is equal to PrivilegeAbst.LINKED, LINKED_FOR_ONE_TO_ONE,
   *   or to LINKED_IF_ALL_SOURCES
   * @param _l_rights is a StringRight array of privilege rights.
   *  May be null, but a null value is not allowed at the first position.
   *  Creates an empty array with a length of 0 if the argument is null.
   *  All rights have to be all positive or all negative.
   * @param _secondType second type which drives the rights and the immutability,
   * just like a Basic type. If null, the internal value is set to '&lt;undefined&gt;'.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if this privilege already exists,
   *   if the rights or the privilege type is not correct,
   *   or if the nick name contains '::' or '&gt;&gt;', contains a character that is not a letter,
   *   if a null right or a right which is not authorized.
   *   of if a parameter is null.
   */
   public PrivilegeForLinksImpl( 
      ACS _aCS, String _nickName, String _mainType,
      StringRight[] _l_rights, String _secondType,
      boolean _toRegisterInManager) throws CreateError {

        super(_aCS, _nickName, _mainType, _l_rights, _secondType);

        if (aCS_.getOnePrivilegeFromDetailledName(detailledName_) != null) {
            aCS_ = null;
            nickName_ = null;
            detailledName_ = null;
            l_rights_ = null;
            mainType_ = null;
            secondType_ = null;
            throw new CreateError(" In the privilege ACS, this new privilege already exists.");
        } // if
        
        if (_mainType.equals(PrivilegeAbst.LINKED)) {
            ifAllSources_ = false;
            oneRoleOneTarget_ = false;
        }
        else if (_mainType.equals(PrivilegeAbst.LINKED_IF_ALL_SOURCES)) {
            ifAllSources_ = true;
            oneRoleOneTarget_ = false;
        }    
        else if (_mainType.equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE)) {
            ifAllSources_ = false;
            oneRoleOneTarget_ = true; 
        } 
        else {
            aCS_ = null;
            nickName_ = null;
            detailledName_ = null;
            l_rights_ = null;
            secondType_ = null;
            throw new CreateError(" The privilege type is wrong.");
        } // else
        
        l_sources_ = new ArrayList(INITIAL_CAPACITY);
        l_targets_ = new ArrayList(INITIAL_CAPACITY);

        String[] l_epTypes = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                                "EligibleParty.EPType");

        if (l_epTypes != null) {
            for (int i = 0; i < l_epTypes.length; i++) {

                if (aCS_.getM_TypeOrientedPolicy().get(
                        "EligibleParty.SecondTypesOfPrivilegeFor."+
                        UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(l_epTypes[i])) == null)
                    continue;

                if (Arrays.asList((String[])aCS_.getM_TypeOrientedPolicy().get(
                        "EligibleParty.SecondTypesOfPrivilegeFor."+
                        UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(l_epTypes[i]))).
                                contains(
                                    UtilityImpl.getFirstInstance().
                                        getObjectTypeWithoutImmutable(getSecondType()) )) {
                    if (!l_internalAllowedSourceTypes_.contains(l_epTypes[i]))
                            l_internalAllowedSourceTypes_.add(l_epTypes[i]);
                } // if
            } // for i

                            // for testing and tracking
                            //System.out.println("PrivilegeForLinksImpl constructor, "+
                              //      "l_internalAllowedSourceTypes_:");
                            //for (Object obj : l_internalAllowedSourceTypes_) {
                              //  System.out.println("--- "+obj);
                            //}

            l_epTypes = null;
        } // if (l_epTypes
        
        String[]  l_resTypes = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                                "Resource.ResourceType");
        String[]  l_actTypes = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                                "Actor.ResourceType");
        String[]  l_resFoldTypes = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                                "VirtualFolder.ResourceVFType");
        String[]  l_epFoldTypes = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                                "VirtualFolder.EPVFType");

        if (l_resTypes != null || l_actTypes != null ||
              l_resFoldTypes != null || l_epFoldTypes != null) {

            if (l_resTypes != null && aCS_.getManageResources()) {
              for (int i = 0; i < l_resTypes.length; i++) {

                    if (aCS_.getM_TypeOrientedPolicy().get(
                            "Resource.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_resTypes[i])) == null)
                        continue;

                    if (Arrays.asList((String[])aCS_.getM_TypeOrientedPolicy().get(
                            "Resource.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_resTypes[i]))).
                                    contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(getSecondType()) )) {
                        if (!l_internalAllowedTargetTypes_.contains(l_resTypes[i]))
                            l_internalAllowedTargetTypes_.add(l_resTypes[i]);
                    } // if
              } // for i
            } // if (l_res != null
            if (l_actTypes != null && aCS_.getManageActors()) {
              for (int i = 0; i < l_actTypes.length; i++) {

                    if (aCS_.getM_TypeOrientedPolicy().get(
                            "Resource.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_actTypes[i])) == null)
                        continue;

                    if (Arrays.asList((String[])aCS_.getM_TypeOrientedPolicy().get(
                            "Resource.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_actTypes[i]))).
                                    contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(getSecondType()) )) {
                        if (!l_internalAllowedTargetTypes_.contains(l_actTypes[i]))
                            l_internalAllowedTargetTypes_.add(l_actTypes[i]);
                    } // if
              } // for i
            } // if (l_res != null
            if (l_resFoldTypes != null && aCS_.getManageResourceVirtualFolders()) {
              for (int i = 0; i < l_resFoldTypes.length; i++) {

                    if (aCS_.getM_TypeOrientedPolicy().get(
                            "VirtualFolder.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_resFoldTypes[i])) == null)
                        continue;

                    if (Arrays.asList((String[])aCS_.getM_TypeOrientedPolicy().get(
                            "VirtualFolder.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_resFoldTypes[i]))).
                                    contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(getSecondType()) )) {
                        if (!l_internalAllowedTargetTypes_.contains(l_resFoldTypes[i]))
                            l_internalAllowedTargetTypes_.add(l_resFoldTypes[i]);
                    } // if
              } // for i
            } // if (l_res != null
            if (l_epFoldTypes != null && aCS_.getManageEPVirtualFolders()) {
              for (int i = 0; i < l_epFoldTypes.length; i++) {

                    if (aCS_.getM_TypeOrientedPolicy().get(
                            "VirtualFolder.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_epFoldTypes[i])) == null)
                        continue;

                    if (Arrays.asList((String[])aCS_.getM_TypeOrientedPolicy().get(
                            "VirtualFolder.SecondTypesOfPrivilegeFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(l_epFoldTypes[i]))).
                                    contains(
                                        UtilityImpl.getFirstInstance().
                                            getObjectTypeWithoutImmutable(getSecondType()) )) {
                        if (!l_internalAllowedTargetTypes_.contains(l_epFoldTypes[i]))
                            l_internalAllowedTargetTypes_.add(l_epFoldTypes[i]);
                    } // if
              } // for i
            } // if (l_res != null

                            // for testing and tracking
                            //System.out.println("PrivilegeForLinksImpl constructor, "+
                              //      "l_internalAllowedTargetTypes_:");
                            //for (Object obj : l_internalAllowedTargetTypes_) {
                              //  System.out.println("--- "+obj);
                            //}

            l_resTypes = null;
            l_actTypes = null;
            l_resFoldTypes = null;
            l_epFoldTypes = null;
        } // if (l_resTypes != null

        // adds this to the map of privileges in the ACS
        // and sends an event "OneLinkedPrivilege" to the GUI listeners.
        // May throw a CreateError if the privilege already exists.
        if (_toRegisterInManager) ((ACSImpl)aCS_).newEorPrivilege(this);
        
  } // PrivilegeForLinksImpl()

  /**
  * This method returns the detailled name or the previous detailled name for the GUI,
  * which gives the key components of the object.
  * An external privilege has a detailled name starting with 'ACS: acs_name | '.
  * When the state of external changes, this method allows to get the right detailled name,
  * with or without the ACS name.
  * If the privilege ACS does not manage the external One-to-One linked privileges,
  * of if this privilege is not an One-to-One linked privilege,
  * the detailled name is returned. This is also true if the property changedDetailledName_ is false.
  * Otherwise, the specific previous detailled name is returned, that is with 'ACS'
  * if the current detailled name does not contain it, or without 'ACS' otherwise.
  * The 4 methods add/removeEorPrivilegeTarget/SourceLink() update the property
  * changedDetailledName_ to change the value this method returns to the GUI.
  * @return the detailled name if it has not changed, or the other version of the detailled name
  */
  public String getPreviousDetailledName() {
      if (!oneRoleOneTarget_ || !aCS_.getManageExternalForOneToOneLinkedPrivileges() ||
              !changedDetailledName_)
          return detailledName_;
      else if (!detailledName_.startsWith("ACS:"))
          return "ACS:"+ acsName_.toString()+ "| " + detailledName_;
      else
          return detailledName_.substring(detailledName_.indexOf(" | ") + 3);
  }

//
// ---------------------------- static methods ------------------------------
//

  /**
  * Gets the registered detailled name from the privilege name.
  * The format of the returned value is: <P>
  * - privilegeType || nickName || first right, for an internal privilege, <P>
  * - ACS: ACS_name | privilegeType || nickName || first right, for an external privilege. <P>
  * Calls ACS.getOnePrivilegeFromDetailledName() twice to test if there is
  * an ACS or not in the detailled name privilege the ACS knows.
  * Calls PrivilegeAbst.buildUpDetailledName() and
  * BaseManagerImpl.getInstance().getEorM_OpenACS().
  * Called by PrivilegeAbst.getDetailledNameFromName() and clone().
  * @see ARoad0.gBase.PrivilegeAbst#buildUpDetailledName
  * @param _name linked privilege name
  * @param _typeIndex index of the component '|| PRIV_TYPE' in _name
  * @return the detailled name of the privilege
  * @exception BaseError if there is a wrong format for _name
  */
  protected static String getDetailledNameFromNameForLinks(
          ImmutableName _name, int _typeIndex) throws BaseError {
    if (_name == null || _name.noOfComponents() < 9)
        throw new BaseError(
            " The privilege name is null, or it is too short.");

      // initialization
    String          result = null;

    try {
        result = buildUpDetailledName(
            _name.component(_typeIndex + 1),
            _name.component(_typeIndex + 3),
            _name.component(_typeIndex + 5),
            null, null);
    } catch (CreateError c) {
        throw new BaseError(c.toString());
    } // catch

    Name    acs_name = _name.subname(1, _typeIndex);
    ACS     acs = (ACS)BaseManagerImpl.getInstance().getEorM_OpenACS().get(acs_name);
    
    if (acs == null) {
        throw new BaseError(
            " The ACS name is malformed: "+acs_name);
    }
    else if (acs.getOnePrivilegeFromDetailledName(result) != null) {
        acs_name = null;
        acs = null;
        return result;
    }
    else if (acs.getOnePrivilegeFromDetailledName(
            "ACS:"+ acs_name.toString()+ "| " + result) != null) {
        result = "ACS:"+ acs_name.toString()+ "| " + result;
        acs_name = null;
        acs = null;
        return result;
    }
    else
        throw new BaseError(
            " The privilege name is unknown in the ACS "+acs_name.lastComponent()+
            ", from the name argument "+_name);
  } // getDetailledNameFromNameForLinks()
   
 //
 //  --------------------------- interface methods --------------------------
 //

  /**
   * The full name is unique for the Access Road program.
   * There, it is equal to the privilege name.
   * @return a copy of the name. Null for an empty privilege.
   */
    public ImmutableName getFullName() {
       return getName();
    } 
   
    /**
     * Gets the BaseObject name of the privilege.
     * The linked privilege name is composed of: <P>
     * ACS:: ACS name <P>
     * ::  || PRIV_TYPE::  main type <P>
     * ::  || NICK_NAME::  privilege name <P>
     * ::  || FIRST_RIGHT::  first right <P>
     * here, ':: ' is added by gBase.NameImpl.toString().
     * The reverse function of getName() is getKeyReferencesFromName().
     * This name is used only outside gBase, following the BaseObject interface.
     * @return ImmutableName of the privilege, or null if it is an empty privilege
     */
    public ImmutableName getName() {
      if (aCS_ == null ) return null;
        
      Name name = new NameImpl("ACS");
      name.extended(aCS_.getName());
      name.extended("|| PRIV_TYPE" );
      name.extended(mainType_ );
      name.extended("|| NICK_NAME" );
      name.extended(nickName_ );
      name.extended("|| FIRST_RIGHT" );
      name.extended(l_rights_[0].getRight() );
      
      return name;
    } // getName() 

   /**
   * @return true if the type of the instance is LINKED_IF_ALL.
   */
  public boolean getIfAllSources() {
      return ifAllSources_;
  } 
    
   /**
   * @return true if the type of the instance is LINKED_FOR_ONE_TO_ONE.
   */
  public boolean getOneRoleOneTarget() {
      return oneRoleOneTarget_;
  } 

   /**
   * Only the LINKED_FOR_ONE_TO_ONE privileges may be external ones, among the linked privileges.
   * An external privilege has exactly one source or one target that does not belong
   * to the privilege ACS. If the source or the target is null, returns always false.
   * An external privilege has a different detailled name format.
   * @return true if the source or the target is from another ACS
   */
  public boolean isExternalPrivilege() {

                            // for testing and tracking
                            //System.out.println("call to PrivilegeForLinksImpl. "+
                              //      "isExternalPrivilege()");

      if (!oneRoleOneTarget_) return false;
      else {
          if (nameOfSourceAcs_ != null && !nameOfSourceAcs_.equals(acsName_))
              return true;
          else if (nameOfTargetAcs_ != null && !nameOfTargetAcs_.equals(acsName_))
              return true;
          else return false;
      } // else
  } // isExternalPrivilege()
  
   /**
   * Gets the sources set which have this privilege.
   * If LINKED_IF_ALL_SOURCES, the privilege is applied only
   * to the objects belonging to every linked source.
   * If LINKED_FOR_ONE_TO_ONE, there is at most one EligibleParty
   * as linked source.
   * @return List of PrivilegeSources. May be empty, but never null.
   */
  public List<PrivilegeSource> getEorLinkedSources() {
      if (l_sources_ == null) return Collections.EMPTY_LIST;
      else return (List)l_sources_.clone();
  } 
 
   /**
    * @return the single source name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getSingleSourceName() {
        return sourceName_;
    }
  
   /**
    * @return the single target name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getSingleTargetName() {
        return targetName_;
    }

   /**
    * @return the single source ACS name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getNameOfSingleSourceACS() {
        return nameOfSourceAcs_;
    }

   /**
    * @return the single target ACS name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getNameOfSingleTargetACS() {
        return nameOfTargetAcs_;
    }
    
  /**
  * Gets the access sources set displayed in the beamer.
  * The format is 'full role name + (LPRI)' if LINKED_FOR_ONE_TO_ONE,
  * or, if there are 3 sources for instance, 
  * 'source last-component name 1 &lt;+&gt; source last-component name 2 &lt;+&gt;
  * source last-component name 3 &lt;+&gt; (LPRI)'
  * The number of sources is from 0 to 4. 
  * @return the sources last-component or full names. Never null.
  */
  public String getSource() {
      if (oneRoleOneTarget_)
          if (l_sources_.size() == 0) return "no_source (LPRI)";
          else return ((ACSObject)l_sources_.get(0)).getName().toString() + "(LPRI)";
      else {
          if (l_sources_.size() == 0) return "no_sources (LPRI)";
          StringBuilder result = new StringBuilder();
          for (Iterator ite = l_sources_.iterator(); ite.hasNext();) {
              result.append( ((ACSObject)ite.next()).getName().lastComponent() );
              if (ite.hasNext()) result.append(" <+> ");
          }
          result.append(" (LPRI)");
          return result.toString();
      } //else
  } // getSource()
  
  /**
   * Links to a source which uses this PrivilegeForLinks.
   * The format of the detailled name for an internal linked privilege is as following: <P>
   * privilegeType || nickName || first right <P>
   * For an external linked privilege, the format of the detailled name is as following: <P>
   * - ACS: ACS_name | privilegeType || nickName || first right <P>
   * An Eligible Party or PrivilegeTarget name is incomplete if it belongs to
   * the Privilege ACS. Since the source and the target are unknown at the creation
   * of the PrivilegeForLinks, the privilege may become external only after the setting
   * of the first source or target. Then, its detailled name changes. The ACS
   * listeners and the privilege target reset the detailled name.
   * Because the detailled name changes when the state of external is updated, 
   * this method calls removeEorPrivilegeToTarget() then addEorPrivilegeToTarget() 
   * on the target, if any, to handle the new detailled name. <P>
   * Calls resetRightsAsInheritedPrivilege() and PrivilegeSource.addEorPrivilegeToSource().
   * Caution: contains a summary of addEorPrivilegeTargetLink(), and the two methods
   * have to remain consistent.
   * Fires the events 'OneLinkedPrivilegeSource', 'OneLinkedPrivilegeTarget',
   * 'OneLinkedPrivilege' and 'ExternalPrivilege'.
   * @param _ep is a non-null source to connect
   * @exception BaseError if the source type is not authorized, or
   * when the instance is a LINKED_FOR_ONE_TO_ONE privilege, if: <P>
   *   - there is already a source <P>
   *   - the new detailled name of the external privilege is not unique in its ACS <P>
   *   or, when the instance is a LINKED_IF_ALL_SOURCES privilege, if: <P>
   *   - _ep is not a GroupIDMember <P>
   *   for all the privilege types: <P>
   *   - if the second type sets the privilege as immutable, <P>
   *   - if the number of sources will be superior to 4, or <P>
   *   - _ep is null or is immutable, or its name is null, <P>
   *   - _ep is  empty, <P>
   *   - _ep name is already known by this instance. <P>
   *   for all the types excepted for LINKED_FOR_ONE_TO_ONE privilege: <P>
   *   - _ep belongs to another acs
   */
  public void addEorPrivilegeSourceLink(PrivilegeSource _ep) throws BaseError {

                    // for testing and tracking
                    //System.out.println("PrivilegeForLinksImpl.addEorPrivilegeSourceLink() start");

    // exception
    if (secondType_.startsWith("<immutable> ") || secondType_.endsWith("<immutable> "))
          throw new BaseError(
            " The second type of this privilege ensures it is immutable.");

    if (_ep.getType().startsWith("<byAcsCreator>") ||
              _ep.getType().startsWith("<immutable> ") ||
              _ep.getType().endsWith("<immutable> "))
          throw new BaseError(
            "The type of "+_ep.getName().lastComponent()+
            " sets it as immutable for its privileges.");

    if (oneRoleOneTarget_) {
        if (l_sources_.size() > 0)
          throw new BaseError(
            " This is an one-to-one privilege, and there is already a source.");
        if (!l_targets_.isEmpty() && ((Resource)l_targets_.get(0)).getEorACS() != aCS_ &&
                _ep.getEorACS() != aCS_)
          throw new BaseError(
            " An external one-to-one privilege cannot have both an external source"+
            " and an external target.");
    } // if
    else if (ifAllSources_) {
        if (!(_ep instanceof GroupIDMember))
          throw new BaseError(
            " This is a sourced privilege, and the parameter is not a GroupIDMember.");
    } // else if

    if (_ep.getEorACS() != aCS_) {
        if (!oneRoleOneTarget_)
          throw new BaseError(
            " The source is not from the same ACS, and this is not an one-to-one privilege.");
        if (!aCS_.getManageExternalForOneToOneLinkedPrivileges())
            throw new BaseError(
                " This privilege has an ACS that forbids it to be an external privilege.");
        // to allow an own component in the privilege of another ACS.
        if (!_ep.getEorACS().getManageExternalForOneToOneLinkedPrivileges())
            throw new BaseError(
                " This source has an ACS that forbids it to be into the privilege of another ACS.");
        if (!l_targets_.isEmpty() && ((ACSObject)l_targets_.get(0)).getEorACS() == aCS_) {
            if (!aCS_.getManagePrivilegesInResources() &&  l_targets_.get(0) instanceof Resource)
                throw new BaseError(
                " This privilege has an ACS that forbids it to have an external privilege on an own resource.");
            else if (!aCS_.getManagePrivilegesInVirtualFolders() && l_targets_.get(0) instanceof VirtualFolder)
                throw new BaseError(
                " This privilege has an ACS that forbids it to have an external privilege on an own virtual folder.");
        } // if
    } // if (_ep.getEorACS() != aCS_)
    else if (oneRoleOneTarget_) {
        if (!aCS_.getManageInternalForOneToOneLinkedPrivileges() &&
                !l_targets_.isEmpty() && ((Resource)l_targets_.get(0)).getEorACS() == aCS_)
            throw new BaseError(
                " This privilege has an ACS that forbids it to be an internal privilege.");
    } // else
    
    if (_ep == null || _ep.getName() == null ||
      _ep.getName().equals(NameImpl.EMPTY_INSTANCE) )
        throw new BaseError(
          " The source is null, or the source name is null or empty.");
    if (l_sources_.contains(_ep) )
      throw new BaseError(
        " The source is already known.");
    if (l_sources_.size() == 4 )
      throw new BaseError(" The number of sources is limited to 4.");

    String epType = _ep.getType();

    if (_ep.getEorACS().getM_TypeOrientedPolicy() != null &&
          _ep.getEorACS().getM_TypeOrientedPolicy().get(
          "EligibleParty.SecondTypesOfPrivilegeFor."+epType) != null) {

            if (!Arrays.asList((String[])_ep.getEorACS().getM_TypeOrientedPolicy().get(
                    "EligibleParty.SecondTypesOfPrivilegeFor."+
                    UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(epType))).
                            contains(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(getSecondType()) )) {

                                      // for testing purpose
                                 //System.out.println("Call to CPrivilegeUserImpl.addEorPrivilegeToSource() "+
                                   // "with a reject of the privilege type '"+getSecondType()+
                                    //"' since the allowed values are:");
                                 //String[] l_val = (String[]) _ep.getEorACS().
                                   //      getM_TypeOrientedPolicy().get(
                                     //   "EligibleParty.SecondTypesOfPrivilegeFor."+epType);
                                 //for (int i = 0; i < l_val.length; i++) {
                                   //  System.out.println("------'" +l_val[i]+"'");
                                 //} // for i

                    throw new BaseError(
                            " The privilege type "+getSecondType()+
                            " is not authorized by the ACS for this source, "+
                            "through the key EligibleParty.SecondTypesOfPrivilegeFor."+epType);
            } // if (!Arrays.asList((String[])master_.getEorACS().getM_TypeOrientedPolicy()
    } // if (_ep.getEorACS().getM_TypeOrientedPolicy() != null

    boolean previousExternal = false;
    if (oneRoleOneTarget_) previousExternal = isExternalPrivilege();

    // updates l_sources_
    l_sources_.add(_ep);
    if (oneRoleOneTarget_) {
        sourceName_ = _ep.getName();
        nameOfSourceAcs_ = _ep.getEorACS().getName();

        if (_ep.getEorACS() != aCS_ && !detailledName_.startsWith("ACS: ")) {

            if (aCS_.getOnePrivilegeFromDetailledName(
                    "ACS:"+ acsName_.toString()+ "| " + detailledName_) != null) {
                l_sources_.remove(_ep);
                sourceName_ = null;
                nameOfSourceAcs_ = null;
                throw new BaseError(" In the privilege ACS, the new detailled name"+
                        " starting with the ACS name already exists.");
            } // if

            // change the detailled name,
            // and this enforces the removing into the ACS listeners and into the target.
            // The ACS privileges list is unchanged, and ACSTreeBaseListenerImpl is a listener.
            changedDetailledName_ = false;
            ((ACSImpl)aCS_).firePropertyChange("OneLinkedPrivilege", this, null);
            PrivilegeTarget target = null;
            if (!l_targets_.isEmpty()) {
                target = (PrivilegeTarget)l_targets_.get(0);
                target.removeEorPrivilegeToTarget(this);
            }

            detailledName_ = "ACS:"+ acsName_.toString()+ "| " + detailledName_;
            changedDetailledName_ = true;

            if (target != null) {
                // summary of addEorPrivilegeTargetLink()
                l_targets_.add(target);
                targetName_ = target.getName();
                nameOfTargetAcs_ = target.getEorACS().getName();

                firePropertyChange("OneLinkedPrivilegeTarget", null, target);

                if (target instanceof ResourceImpl) {
                    ((ResourceImpl)target).addEorPrivilegeForLinksToTarget(this);
                }
                else if (target instanceof VirtualFolderImpl) {
                    ((VirtualFolderImpl)target).addEorPrivilegeForLinksToTarget(this);
                }
            } // if
            ((ACSImpl)aCS_).firePropertyChange("OneLinkedPrivilege", null, this);
            target = null;

            // register _ep as external EP in the privilege ACS
            try {
                if (!aCS_.containsEligiblePartyName(_ep.getName()))
                    ((ACSImpl)aCS_).addEligibleParty((ImmutableEligibleParty)_ep);
            }
            catch (CreateError c) {
                throw new BaseError(
                    " This external source is rejected by the privilege ACS.");
            }

        } // if (_ep.getEorACS() != aCS_
        else if (_ep.getEorACS() != aCS_) { // l_targets_.get(0) is external
            throw new InternalError(
               " This external privilege cannot have two external ends.");
        }
    } // if

    try {
        firePropertyChange("OneLinkedPrivilegeSource", null, _ep);
        if (oneRoleOneTarget_ && previousExternal != isExternalPrivilege())
            firePropertyChange("ExternalPrivilege", previousExternal, isExternalPrivilege());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From PrivilegeForLinksImpl.addEorSource(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }
    
    if (_ep instanceof GroupIDMember)
        ((EligiblePartyImpl)_ep).addEorPrivilegeForLinksToSource(this);
    else if (_ep instanceof Actor)
        ((ActorImpl)_ep).addEorPrivilegeForLinksToSource(this);

    if (!getMainType().equals(LINKED_IF_ALL_SOURCES))
        resetRightsAsInheritedPrivilege();

                    // for testing and tracking
                    //System.out.println("PrivilegeForLinksImpl.addEorPrivilegeSourceLink() end");
    
  } // addEorPrivilegeSourceLink()

  /**
   * Adds a connected target which has this PrivilegeForLinks.
   * Calls ACSImpl.addAcsController() and PrivilegeTarget.addEorPrivilegeToTarget().
   * Because the detailled name changes when the state of external is updated,
   * this method calls removeEorPrivilegeToSource() then addEorPrivilegeToSource()
   * on the target, if any, to handle the new detailled name.
   * Caution: contains a summary of addEorPrivilegeSourceLink(), and the two methods
   * have to remain consistent.
   * Fires the events 'OneLinkedPrivilegeSource', 'OneLinkedPrivilegeTarget',
   * 'OneLinkedPrivilege' and 'ExternalPrivilege'.
   * @param _res is a non-null target to connect
   * @exception BaseError when the instance is a LINKED_FOR_ONE_TO_ONE privilege, if: <P>
   *   - there is already a target <P>
   *   - the new detailled name of the external privilege is not unique in its ACS <P>
   *   for all privilege types: <P>
   *   - if the privilege type is not authorized to the target, <P>
   *   - if the second type sets the privilege as immutable, <P>
   *   - if _res is null or immutable, or its name is null, <P>
   *   - _res is  empty, <P>
   *   - _res name is already known by this instance. <P>
   *   for all types excepted for LINKED_FOR_ONE_TO_ONE privilege: <P>
   *   - _res belongs to another acs
   *   and, for the LINKED_IF_ALL_SOURCES privilege: <P>
   *   - the target ACS cannot manage the PRI inheritance in a LINKED_IF_ALL_SOURCES privilege.
   *   and, for all types excepted for LINKED_IF_ALL_SOURCES privilege: <P>
   *   - the first target ACS manages the PRI inheritance, then no target may be added
   */
  public void addEorPrivilegeTargetLink(PrivilegeTarget _res) throws BaseError {

                    // for testing and tracking
                    //System.out.println("PrivilegeForLinksImpl.addEorPrivilegeTargetLink() start");

    // exception
    if (secondType_.startsWith("<immutable>") || secondType_.endsWith("<immutable> "))
          throw new BaseError(
            " The second type of "+detailledName_+" sets it as immutable.");

    if (_res.getType().startsWith("<byAcsCreator>") || 
              _res.getType().startsWith("<immutable> ") ||
              _res.getType().endsWith("<immutable> "))
          throw new BaseError(
            "The type of "+_res.getName().lastComponent()+
            " sets it as immutable for its privileges.");

    if (oneRoleOneTarget_) {
        if (l_targets_.size() > 0)
          throw new BaseError(
            " This is an one-to-one privilege and there is already a target.");
        if (!l_sources_.isEmpty() && ((EligibleParty)l_sources_.get(0)).getEorACS() != aCS_ &&
                _res.getEorACS() != aCS_)
          throw new BaseError(
            " An external one-to-one privilege cannot have both an external source"+
            " and an external target.");
    } // if

    if (_res.getEorACS() != aCS_) {
        if (!oneRoleOneTarget_)
          throw new BaseError(
            " The target is not from the same ACS, and this is not an one-to-one privilege.");
        if (!aCS_.getManageExternalForOneToOneLinkedPrivileges())
            throw new BaseError(
                " This privilege has an ACS that forbids it to be an external privilege.");  
        // to allow an own component in the privilege of another ACS.
        if (!_res.getEorACS().getManageExternalForOneToOneLinkedPrivileges())
            throw new BaseError(
                " This target has an ACS that forbids it to be into the privilege of another ACS.");
    } // if (_res.getEorACS() != aCS_)
    else if (oneRoleOneTarget_) {
        if (!l_sources_.isEmpty() && ((EligibleParty)l_sources_.get(0)).getEorACS() != aCS_) {
            if (!aCS_.getManagePrivilegesInResources() && _res instanceof Resource)
                throw new BaseError(
                    " This privilege has an ACS that forbids it to have an external privilege"+
                        "on an own resource.");
            else if (!aCS_.getManagePrivilegesInVirtualFolders() && _res instanceof VirtualFolder)
                throw new BaseError(
                    " This privilege has an ACS that forbids it to have an external privilege"+
                        "on an own virtual folder.");
        } // if
        if (!aCS_.getManageInternalForOneToOneLinkedPrivileges() &&
                !l_sources_.isEmpty() && ((EligibleParty)l_sources_.get(0)).getEorACS() == aCS_)
            throw new BaseError(
                " This privilege has an ACS that forbids it to be an internal privilege.");
    } // else
        
    if (_res == null || _res.getName() == null ||
      _res.getName().equals(NameImpl.EMPTY_INSTANCE) )
        throw new BaseError(
          " The target or the target name is null, or the object name is null or empty.");
    if (l_targets_.contains(_res) )
      throw new BaseError(
        " The target is already known.");
    if (getMainType().equals(LINKED_IF_ALL_SOURCES) &&
            _res.getEorACS().getManageLPRIRightsInheritance())
        throw new BaseError(
          " The target ACS cannot manage the privilege inheritance in a sourced privilege.");
    if (!getMainType().equals(LINKED_IF_ALL_SOURCES) && !l_targets_.isEmpty() &&
            ((PrivilegeTarget)l_targets_.get(0)).getEorACS().getManageLPRIRightsInheritance())
        throw new BaseError(
          " The first target ACS manages the PRI inheritance, then no target may be added.");
    if (_res instanceof Resource && !aCS_.getManagePrivilegesInResources()) {
        throw new BaseError(
            " This privilege has an ACS that forbids it to have a Resource as target.");
    }
    if (_res instanceof VirtualFolder && !aCS_.getManagePrivilegesInVirtualFolders()) {
        throw new BaseError(
            " This privilege has an ACS that forbids it to have a VirtualFolder as target.");
    }

      String inter = null;
      String resType = _res.getType();

      if (_res instanceof ImmutableResource)
          inter = "Resource";
      else if (_res instanceof ImmutableVirtualFolder)
          inter = "VirtualFolder";

      if (_res.getEorACS().getM_TypeOrientedPolicy() != null &&
            _res.getEorACS().getM_TypeOrientedPolicy().get(
            inter+".SecondTypesOfPrivilegeFor."+resType) != null) {

                if (!Arrays.asList((String[])_res.getEorACS().getM_TypeOrientedPolicy().get(
                        inter+".SecondTypesOfPrivilegeFor."+
                        UtilityImpl.getFirstInstance().
                            getObjectTypeWithoutImmutable(resType))).
                                contains(
                                    UtilityImpl.getFirstInstance().
                                        getObjectTypeWithoutImmutable(getSecondType()) )) {

                              // for testing purpose
                         //System.out.println("PrivilegeForLinksImpl.addEorPrivilegeTargetLink() "+
                           // "with a reject of the privilege type '"+getSecondType()+
                            //"' since the allowed values are:");
                         //String[] l_val = (String[]) _res.getEorACS().
                           //      getM_TypeOrientedPolicy().get(
                             //   inter+".SecondTypesOfPrivilegeFor."+resType);
                         //for (int i = 0; i < l_val.length; i++) {
                           //  System.out.println("------'" +l_val[i]+"'");
                         //}

                        throw new BaseError(
                                "The privilege target type "+resType+
                                " is not authorized by the ACS, "+
                                "through the key "+inter+".SecondTypesOfPrivilegeFor."+resType);
                } // if (!Arrays
    } // if (_res


    boolean previousExternal = false;
    if (oneRoleOneTarget_) previousExternal = isExternalPrivilege();

    // update l_targets_
    l_targets_.add(_res);
    if (oneRoleOneTarget_) {
        targetName_ = _res.getName();
        nameOfTargetAcs_ = _res.getEorACS().getName();

        if (_res.getEorACS() != aCS_) {

            if (!detailledName_.startsWith("ACS: ")) {

                if (aCS_.getOnePrivilegeFromDetailledName(
                        "ACS:"+ acsName_.toString()+ "| " + detailledName_) != null) {
                    l_targets_.remove(_res);
                    targetName_ = null;
                    nameOfTargetAcs_ = null;
                throw new BaseError(" In the privilege ACS, the new detailled name"+
                        " starting with the ACS name already exists.");
                } // if

                // Change the detailled name,
                // and this enforces the removing into the ACS listeners and into the source.
                // The ACS privileges list is unchanged, and ACSTreeBaseListenerImpl is a listener.
                changedDetailledName_ = false;
                ((ACSImpl)aCS_).firePropertyChange("OneLinkedPrivilege", this, null);
                EligibleParty source = null;
                if (!l_sources_.isEmpty()) {
                    source = (EligibleParty)l_sources_.get(0);
                    ((PrivilegeSource)l_sources_.get(0)).removeEorPrivilegeToSource(this);
                }

                detailledName_ = "ACS:"+ acsName_.toString()+ "| " + detailledName_;
                changedDetailledName_ = true;

                if (source != null) {
                    // summary of addEorPrivilegeSourceLink()
                    l_sources_.add(source);
                    sourceName_ = source.getName();
                    nameOfSourceAcs_ = source.getEorACS().getName();

                    firePropertyChange("OneLinkedPrivilegeSource", null, source);

                    if (source instanceof GroupIDMember)
                        ((EligiblePartyImpl)source).addEorPrivilegeForLinksToSource(this);
                    else if (source instanceof Actor)
                        ((ActorImpl)source).addEorPrivilegeForLinksToSource(this);

                } // if
                
                try {
                    // like in ACSImpl.newEorAclEntry(), for the target and not for the source
                    ((ACSImpl)_res.getEorACS()).addAcsController(acsName_);
                    ((ACSImpl)aCS_).firePropertyChange("OneLinkedPrivilege", null, this);
                    source = null;

                    // register _res as external target in the privilege ACS
                    if (_res instanceof ImmutableResource &&
                            !aCS_.containsResourceName(_res.getName()))
                        ((ACSImpl)aCS_).addResource((ImmutableResource)_res);
                    else if (_res instanceof VirtualFolder &&
                            !aCS_.containsVirtualFolderName(_res.getName()))
                        ((ACSImpl)aCS_).addVirtualFolder((VirtualFolder)_res);
                }
                catch (CreateError c) {
                    // never occurs
                    throw new BaseError(
                        " This external target is rejected by the privilege ACS.");
                }

            } // if (externalPrivilege
            else { // l_sources_.get(0) is external
                throw new InternalError(
                   "This external privilege cannot have two external ends: "+this);
            }
        } // if (_res.getEorACS() != aCS_)
        //else {
            // Managing external privileges always implies that all the own sources of this ACS
            // may have an inner external privilege. To allow an own resource or an own virtual folder
            // of this ACS to get an inner external privilege, the privileges in resources or
            // in virtual folders have to be managed. This allows to limit
            // the inner external privileges to the own sources of the ACS, excluding the own targets,
            // if there are no internal privileges to manage. On the other hand, having
            // internal privileges means that the inner external privileges always accept
            // as target an own resource or own virtual folder.
        //}
    } // if (oneRoleOneTarget)

    try {
        firePropertyChange("OneLinkedPrivilegeTarget", null, _res);
        if (oneRoleOneTarget_ && previousExternal != isExternalPrivilege())
            firePropertyChange("ExternalPrivilege", previousExternal, isExternalPrivilege());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From PrivilegeForLinksImpl.addEorTarget(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    // no exception
    if (_res instanceof Resource)
        ((ResourceImpl)_res).addEorPrivilegeForLinksToTarget(this);
    else if (_res instanceof VirtualFolder)
        ((VirtualFolderImpl)_res).addEorPrivilegeForLinksToTarget(this);

                    // for testing and tracking
                    //System.out.println("PrivilegeForLinksImpl.addEorPrivilegeTargetLink() end");
                    
  } // addEorPrivilegeTargetLink()

  /**
   * Removes a connected source using this PrivilegeForLinks.
   * Called by CPrivilegeUserImpl, for instance when this privilege is finalizing,
   * and by the GUI.
   * Calls resetRightsAsInheritedPrivilege() and PrivilegeSource.removeEorPrivilegeToTarget().
   * Because the detailled name changes when the state of external is updated,
   * this method calls removeEorPrivilegeToTarget() then addEorPrivilegeToTarget()
   * on the target, if any, to handle the new detailled name.
   * Caution: contains a summary of addEorPrivilegeTargetLink(), and the two methods
   * have to remain consistent.
   * Fires the events 'OneLinkedPrivilegeSource', 'OneLinkedPrivilegeTarget',
   * 'OneLinkedPrivilege' and 'ExternalPrivilege'.
   * @param _ep is a source to unset
   * @exception BaseError if the name is null, or the source is null or immutable,
   *  if the second type sets the privilege as immutable,
   *  or if the source is not known.
   */
  public void removeEorPrivilegeSourceLink(PrivilegeSource _ep) throws BaseError {

                    // for testing and tracking
                    //System.out.println("PrivilegeForLinksImpl.removeEorPrivilegeSourceLink() start");

    // exception
    if (secondType_.startsWith("<immutable>") || secondType_.endsWith("<immutable> "))
          throw new BaseError(
            " The second type of "+detailledName_+" sets it as immutable.");
    if (_ep == null || _ep.getName() == null ||
      _ep.getName().equals(NameImpl.EMPTY_INSTANCE) )
        throw new BaseError(" The name is null, or the object name is null or empty.");

    // _ep.getType() null when called by _ep.finalizeForBase()
    // through CPrivilegeUserImpl.removeEorPrivilegeToSource()
    if (_ep.getType() != null &&
              (_ep.getType().startsWith("<byAcsCreator>") ||
              _ep.getType().startsWith("<immutable> ") ||
              _ep.getType().endsWith("<immutable> ")))
          throw new BaseError(
            "The type of "+_ep.getName().lastComponent()+
            " sets it as immutable for its privileges.");

    if (!l_sources_.contains(_ep) )
      throw new BaseError(
        " The object name is unknown in this privilege.");

    boolean previousExternal = false;
    if (oneRoleOneTarget_) previousExternal = isExternalPrivilege();

    // update l_sources_
    l_sources_.remove(_ep);
    if (oneRoleOneTarget_) {
        sourceName_ = null;
        nameOfSourceAcs_ = null;

        // Change the detailled name,
        // and this enforces the removing into the ACS listeners and into the target.
        // The ACS privileges list is unchanged, and ACSTreeBaseListenerImpl is a listener.
        if (_ep.getEorACS() != aCS_ && detailledName_.startsWith("ACS: ")) {

            if (aCS_.getOnePrivilegeFromDetailledName(
                    detailledName_.substring(detailledName_.indexOf(" | ") + 3)) != null) {
                l_sources_.add(_ep);
                sourceName_ = _ep.getName();
                nameOfSourceAcs_ = _ep.getEorACS().getName();
                throw new BaseError(" In the privilege ACS, the new detailled name"+
                        " without ACS name already exists. The source cannot be removed.");
            } // if

            changedDetailledName_ = false;
            ((ACSImpl)aCS_).firePropertyChange("OneLinkedPrivilege", this, null);
            PrivilegeTarget target = null;
            if (!l_targets_.isEmpty()) {
                target = (PrivilegeTarget)l_targets_.get(0);
                target.removeEorPrivilegeToTarget(this);
            }

            detailledName_ = detailledName_.substring(detailledName_.indexOf(" | ") + 3);

            // if changedDetailledName_ = false;
            if (target != null) {
                // summary of addEorPrivilegeTargetLink()
                l_targets_.add(target);
                targetName_ = target.getName();
                nameOfTargetAcs_ = target.getEorACS().getName();

                firePropertyChange("OneLinkedPrivilegeTarget", null, target);

                if (target instanceof Resource)
                    ((ResourceImpl)target).addEorPrivilegeForLinksToTarget(this);
                else if (target instanceof VirtualFolder)
                    ((VirtualFolderImpl)target).addEorPrivilegeForLinksToTarget(this);
            } // if
            ((ACSImpl)aCS_).firePropertyChange("OneLinkedPrivilege", null, this);

            // finalizeForBase() when it calls CPrivilegeForUserImpl.removeEorPrivilegeToSource()
            // which fires 'OneSourceLinkedPrivilege' after this method.
            changedDetailledName_ = true;

            // unregister _ep as external EP in the privilege ACS.
            // An ACS cannot manage both the external AclEntries and the external linked Privileges.
            try {
                boolean toRemain = false;

                for (Privilege pri : aCS_.getEorL_PrivilegeForLinks()) {
                    if (pri == this) continue;
                    else if (((PrivilegeForLinks)pri).getEorLinkedSources().contains(_ep)) {
                        toRemain = true;
                        break;
                    }
                } // for pri

                if (!toRemain)
                    ((ACSImpl)aCS_).removeEorEligibleParty((ImmutableEligibleParty)_ep);
            }
            catch (UpDateError u) {
                throw new BaseError(
                    "This external source cannot be unregistered by the privilege ACS.");
            }

        } // if (!externalPrivilege
    } // if

    try {
        firePropertyChange("OneLinkedPrivilegeSource", _ep, null);
        if (oneRoleOneTarget_ && previousExternal != isExternalPrivilege())
            firePropertyChange("ExternalPrivilege", previousExternal, isExternalPrivilege());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From PrivilegeForLinksImpl.removeEorSource(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
    }

    // false if this method has been called by _ep
    if (_ep.getEorL_SourcePrivilegeForLinks().contains(this))
        _ep.removeEorPrivilegeToSource(this);

    if (!getMainType().equals(LINKED_IF_ALL_SOURCES))
        resetRightsAsInheritedPrivilege();

                    // for testing and tracking
                    //System.out.println("PrivilegeForLinksImpl.removeEorPrivilegeSourceLink() end");
                    
  } // removeEorPrivilegeSourceLink()

  /**
   * Removes a connected target using this PrivilegeForLinks.
   * Called by the CPrivilegeUserImpl of the target when this privilege is finalizing,
   * and by the GUI.
   * Calls PrivilegeTarget.removeEorPrivilegeForLinksToTarget().
   * In a one-to-one privilege, the target can be removed only when the privilege
   * is finalizing.
   * Fires the events 'OneLinkedPrivilegeTarget' and 'ExternalPrivilege'.
   * @param _res target to remove
   * @exception BaseError if the name is null, or the target is null or immutable,
   *  if the second type sets the privilege as immutable,
   *  or if the target is not known.
   */
  public void removeEorPrivilegeTargetLink(PrivilegeTarget _res) throws BaseError {
    // exception
    if (secondType_.startsWith("<immutable>") || secondType_.endsWith("<immutable> "))
          throw new BaseError(
            " The second type of "+detailledName_+" sets it as immutable.");
    if (_res == null || _res.getName() == null ||
      _res.getName().equals(NameImpl.EMPTY_INSTANCE) )
        throw new BaseError(" The target to remove is null, or the target name is null or empty.");

    // _ep.getType() null when called by _ep.finalizeForBase()
    // through CPrivilegeUserImpl.removeEorPrivilegeToTarget()
    if (_res.getType() != null &&
            (_res.getType().startsWith("<byAcsCreator>") ||
              _res.getType().startsWith("<immutable> ") ||
              _res.getType().endsWith("<immutable> ")))
          throw new BaseError(
            "The type of "+_res.getName().lastComponent()+
            " sets it as immutable for its privileges.");

    if (!l_targets_.contains(_res) )
      throw new BaseError(" The target to remove is unknown by the privilege.");
    // _res returns false if this method has been called by _res for this.finalizeForBase()
    if (oneRoleOneTarget_ && _res.getEorL_TargetPrivilegeForLinks().contains(this)) {
      throw new BaseError(" In an one-to-one privilege, the target cannot be removed.");
    }
    
    boolean previousExternal = false;
    if (oneRoleOneTarget_) previousExternal = isExternalPrivilege();
    
    // updates l_targets_
    l_targets_.remove(_res);

    try {
        firePropertyChange("OneLinkedPrivilegeTarget", _res, null);
        if (oneRoleOneTarget_) {
            if (_res.getEorACS() != aCS_) {
                // unregister _res as an external object in the privilege ACS.
                // An ACS cannot manage both the external AclEntries and the external Linked Privileges.
                try {
                    boolean toRemain = false;

                    // for CollectionPropertyEditor.propertyChange() which
                    // listens the property change 'OneTargetLinkedPrivilege'
                    // from the CPrivilegeUserImpl.removeEorPrivilegeToTarget()
                    // of this privilege target, after the call to this method.
                    detailledName_ = detailledName_.substring(detailledName_.indexOf(" | ") + 3);
                    // finalizeForBase() when it calls removeEorPrivilege()
                    // which fires 'OneLinkedPrivilege' and other events.
                    //changedDetailledName_ = true;

                    for (Privilege pri : aCS_.getEorL_PrivilegeForLinks()) {
                        if (pri == this) continue;
                        else if (((PrivilegeForLinks)pri).getEorLinkedTargets().contains(_res)) {
                            toRemain = true;
                            break;
                        }
                    } // for pri

                    if (!toRemain) {

                                    // for testing and tracking
                                    //System.out.println(
                                      //  "PrivilegeForLinksImpl.removeEorPrivilegeTargetLink() calls"+
                                        //" remove target on the ACS");

                        if (_res instanceof ImmutableResource)
                            ((ACSImpl)aCS_).removeEorResource((ImmutableResource)_res);
                        else if (_res instanceof VirtualFolder)
                            ((ACSImpl)aCS_).removeEorVirtualFolder((VirtualFolder)_res);
                    } // if
                }
                catch (UpDateError u) {
                    throw new BaseError(
                        "This external target cannot be unregistered by the privilege ACS.");
                }
            } // if (_res.getEorACS() != aCS_

            if (previousExternal != isExternalPrivilege())
                firePropertyChange("ExternalPrivilege", previousExternal, isExternalPrivilege());
        } // if (oneRoleOneTarget_
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From PrivilegeForLinksImpl.removeEorTarget(), firePropertChange() throws:");
        e.printStackTrace();
    }

    // false if this method has been called by _res for this.finalizeForBase()
    if (_res.getEorL_TargetPrivilegeForLinks().contains(this))
        _res.removeEorPrivilegeToTarget(this);
    
  } // removeEorPrivilegeTargetLink()
  
   /**
   * Gets the targets which use this privilege.
   * If LINKED_FOR_ONE_TO_ONE, there is at most one target.
   * If the targets are external ones, all targets belong to the same ACS.
   * @return List of PrivilegeTargets. May be empty,
   *   but never null.
   */
  public List<PrivilegeTarget> getEorLinkedTargets() {
      if (l_targets_ == null) return Collections.EMPTY_LIST;
      else return (List)l_targets_.clone();
  } 
    
   /**
    * Caution: the list is from the privilege ACS, and an external privilege
    * may accept types from another ACS.
    * @return unmodifiable list of the source types that the ACS authorizes for this privilege.
    *  May be empty but never null.
    */
  public List<String> getL_InternalAllowedSourceTypes() {
      return Collections.unmodifiableList(l_internalAllowedSourceTypes_);
  } // getL_InternalAllowedSourceTypes()
    
   /**
    * Caution: the list is from the privilege ACS, and an external privilege
    * may accept types from another ACS.
    * A target may be a resource or a VirtualFolder.
    * @return unmodifiable list of the target types that the ACS authorizes for this privilege.
    *  May be empty but never null.
    */
  public List<String> getL_InternalAllowedTargetTypes() {
      return Collections.unmodifiableList(l_internalAllowedTargetTypes_);
  } // getL_InternalAllowedTargetTypes()

  /**
   * If the target is in an ACS that manages the PRI inheritance,
   * calls resetInheritedPrivilegeyRightsOfChildren() to reset
   * the effective rights in the inherited Privilege map of each child.
   * Called by add/removeEorPrivilegeSourceLink().
   * Calls Directory.resetInheritedPrivilegeRightsOfChildren() and
   * VirtualFolder.resetInheritedPrivilegeRightsOfChildren().
   * Note: the creation of the inherited Privilege in a child is not handled by
   * the generic gBase package, while the rights updating of the inherited Privilege
   * is handled.
   * Fires no PropertyChangeEvent.
   */
  protected void resetRightsAsInheritedPrivilege() {
      if (!aCS_.getManageLPRIRightsInheritance() || l_targets_.isEmpty()) {
          return;
      }

      if (l_targets_.get(0) instanceof Resource) {
          Resource res = (Resource) l_targets_.get(0);
          if (res instanceof Directory) {
              ((Directory)res).resetInheritedPrivilegeRightsOfChildren(this);
          }
          res = null;
      }
      else if (l_targets_.get(0) instanceof VirtualFolder) {
          VirtualFolder vf = (VirtualFolder) l_targets_.get(0);
          ((VirtualFolder)vf).resetInheritedPrivilegeRightsOfChildren(this);
          vf = null;
      }
  } // resetRightsAsInheritedPrivilege()

  /**
  * Adds a listener to the bean for the events 'OneLinkedPrivilegeSource',
 * 'OneLinkedPrivilegeTarget', 'ExternalPrivilege'. <P>
  * Used by Gui1.CommonPropertyEditor for the 'Comment'
  * and 'SeeWhy' properties. Overiddes the super method.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      if (_propertyName.equals("OneLinkedPrivilegeSource") ||
              _propertyName.equals("OneLinkedPrivilegeTarget") ||
              _propertyName.equals("ExternalPrivilege") )
          changeSupport_.addPropertyChangeListener(_propertyName, _l);
      else super.addPropertyChangeListener(_propertyName, _l);
  }

  /**
  * Removes a listener to the bean for the events 'OneLinkedPrivilegeSource',
 * 'OneLinkedPrivilegeTarget', 'ExternalPrivilege'. <P>
  * Overiddes the super method.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("PrivilegeImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("OneLinkedPrivilegeSource") ||
              _propertyName.equals("OneLinkedPrivilegeTarget") ||
              _propertyName.equals("ExternalPrivilege") )
          changeSupport_.removePropertyChangeListener(_propertyName, _l);
      else super.removePropertyChangeListener(_propertyName, _l);
  } // removePropertyChangeListener()
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("PrivilegeForLinksImpl.getPropertyChangeListeners(),
                    // call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("OneLinkedPrivilegeSource") ||
              _propertyName.equals("OneLinkedPrivilegeTarget") ||
              _propertyName.equals("ExternalPrivilege") )
            return changeSupport_.getPropertyChangeListeners(_propertyName);

      else return super.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()
  
  /**
   * Used to clone a privilege.
   * Provides a shallow copy of sources and target lists.
   * Caution: the clone is not registered in its ACS, nor in its sources and targets.
   * @return Object without change in the external references
   *  but copy of these references.
   */
    public Object clone() {
      //try {
          PrivilegeForLinksImpl clon = (PrivilegeForLinksImpl) super.clone();

          clon.ifAllSources_ = ifAllSources_;          
          clon.oneRoleOneTarget_ = oneRoleOneTarget_;
          
          clon.l_sources_ = (ArrayList) l_sources_.clone();
          clon.l_targets_ = (ArrayList) l_targets_.clone();
          
          if (sourceName_ != null)
              clon.sourceName_ = (ImmutableName)sourceName_.clone();
          else clon.sourceName_ = null;
          if (targetName_ != null)
              clon.targetName_ = (ImmutableName)targetName_.clone();
          else clon.targetName_ = null;
          if (nameOfSourceAcs_ != null)
              clon.nameOfSourceAcs_ = (ImmutableName)nameOfSourceAcs_.clone();
          else clon.nameOfSourceAcs_ = null;
          if (nameOfTargetAcs_ != null)
              clon.nameOfTargetAcs_ = (ImmutableName)nameOfTargetAcs_.clone();
          else clon.nameOfTargetAcs_ = null;
          
          return clon;

      //} catch (CloneNotSupportedException e) {
	//    // this shouldn't happen, since we are Cloneable
  	  //  if (aCS_ != null) {
  	    //        aCS_.getEorBaseError().addException(e.toString());
  	  //  }
        //throw new InternalError(e.toString());
      //} // catch
    } // clone()

   /**
    * In this method, the current sources and targets are not compared.
   * @param _obj to compare.
   * @return boolean true if _obj is equal with the same class
   *  and the same detailled name.
   */
  public synchronized boolean equals(Object _obj) {

      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      if (!super.equals(_obj))
          return false;
      
      PrivilegeForLinksImpl obj = (PrivilegeForLinksImpl) _obj;

      if (ifAllSources_ != obj.getIfAllSources() ) return false;      
      if (oneRoleOneTarget_ != obj.getOneRoleOneTarget() ) return false;

      if (detailledName_ == null || obj.detailledName_ == null) {
          if (detailledName_ != null || obj.detailledName_ != null)
            return false;  // if
      } // if
      else {
          if (!detailledName_.equals( obj.detailledName_ )) return false;
      }
      
      return true;
  } // equals(Object)

  /**
  * @return boolean true if equal to new PrivilegeForLinksImpl()
  */
  public boolean isEmpty() {
    return equals(EMPTY_INSTANCE );
  }
  
  /**
   * Gets the main components from the linked privilege name as a string,
   * returned by getName(). These key components define fully
   * the privilege instance.
   * @see PrivilegeAbst#getAcsFromName(ARoad0.gBaseInterface.ImmutableName)
   * @param _privName has been returned by getName()
   * @return null if the name format is wrong, or an array of the 6 Privilege components: <P>
   *  - index 0: name of the ACS which manages this privilege (ImmutableName) <P>
   *  - index 1: privilege type (String) <P>
   *  - index 2: nick name (String) <P>
   *  - index 3: first right (String) <P>
   *  - index 4: null <P>
   *  - index 5: null
   * @exception CreateError if the argument or a component is null or not correct,
   *  not a privilege name string, or if a main component string
   *  is not provided by NameImpl.toString()
   */
    public Object[] getKeyReferencesFromName(ImmutableName _privName)
      throws CreateError {
        if (_privName == null || aCS_ == null)
            throw new CreateError("In PrivilegeForLinksImpl."+
                "getKeyReferencesFromName(), the argument or the ACS is null.");
        int index = aCS_.getName().noOfComponents();
        if (!_privName.component(0).equals("ACS")
          || !_privName.component(index+1).equals("|| PRIV_TYPE")
          || !_privName.component(index+3).equals("|| NICK_NAME")
          || !_privName.component(index+5).equals("|| FIRST_RIGHT" ))
            throw new CreateError("In PrivilegeForLinksImpl.getKeyReferencesFromName()"+
              " the argument is not a linked privilege name.");

        // variables
        Object[] l_result = new NameImpl[6];
        index++; // for "|| PRIV_TYPE"

        l_result[0] = _privName.subname(1, index); // ACS name
        l_result[1] = _privName.subname(index + 1, index + 2).component(0); // privilege type
        l_result[2] = _privName.subname(index + 3, index + 4).component(0); // nick name
        l_result[3] = _privName.subname(index + 5, index + 6).component(0); // first right
        
        if (l_result[0] == null || ((Name)l_result[0]).isEmpty()
          || (!l_result[1].equals(PrivilegeAbst.LINKED_IF_ALL_SOURCES) && 
                   !l_result[1].equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE) && 
                   !l_result[1].equals(PrivilegeAbst.LINKED))
          || l_result[2] == null
          || l_result[3] == null)
            throw new CreateError(
              " A component of the key references is not from a linked privilege name.");
        
              // for testing and tracking
              //System.out.println("in PrivilegeForLinksImpl.getKeyReferencesFromName(): "+
              //  "ACS name: "+l_result[0]+
              //  "privilege type: "+l_result[1]+
              //  "nick name: "+l_result[2]+
              //  "gant/deny: "+l_result[3]);

        return l_result;
    } // getKeyReferencesFromName()  

   /**
    * The form is ' Detailled name: KKKK + Source(s): FFFF + Target(s): NNNNNNNN +
    * Right(s): jjjj| kkkk| mmmm| ', where the detailed name has the form:
    * 'privilegeType || nickName || first right'.
    * Called by BaseManagerImpl.importACS() to display the closed Privileges.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("Detailled name: ");
    sb.append(detailledName_);
    
    sb.append(" + Source(s): ");
    if (oneRoleOneTarget_ && sourceName_ != null)
        sb.append(sourceName_);
    else if (l_sources_ != null) {
            for (PrivilegeSource so : l_sources_) {
                sb.append(so.getName()+"| ");
            } // for so
    }
    else
        sb.append("<none>");
    
    sb.append(" + Target(s): ");
    if (oneRoleOneTarget_ && targetName_ != null)
        sb.append(targetName_);
    else if (l_targets_ != null) {
            for (PrivilegeTarget tg : l_targets_) {
                sb.append(tg.getName()+"| ");
            } // for so
    }
    else
        sb.append("<none>");

    sb.append(" + Right(s): ");
    if (l_rights_.length != 0) {
      for (int i = 0; i < l_rights_.length; i++) {
        if (l_rights_[i] != null)
            sb.append(l_rights_[i].toString());
        sb.append("| ");
      } //for
    } // if
    else sb.append("<none>");
    
    return " "+sb;
  } // toString()

  /**
   * It sets to null almost all variables, so the instance is equals
   * to a new empty PrivilegeForLinksImpl.
   * The relevant sources and targets register/unregister the linked privilege when the source,
   * the target or the privilege is created/deleted.
   * An unused object has to be deleted to inform the others objects, and this method
   * HAS TO BE CALLED before deleting this object.
   * This method is called by finalize(), finalizeForUser() and
   * ACSImpl.closeExternalLinkedPrivileges().
   * @exception UpDateError if the privilege is unknwon from the ACS, the sources
   *   or the targets.
   */
  public synchronized void finalizeForBase() throws UpDateError {
    if (aCS_ == null ) return;

                  // for testing and tracking
                  //System.out.println("PrivilegeForLinksImpl.finalizeForUser() for: "+getName());

    PrivilegeSource source = null;
    PrivilegeTarget target = null;
    boolean         reset = false;

    if (oneRoleOneTarget_ && !l_sources_.isEmpty() &&
            ((EligibleParty)l_sources_.get(0)).getEorACS() != aCS_)
        reset = true;

    // removes the sources and targets, calls back the remove methods of this class,  
    try {
        for(Iterator ite = ((ArrayList)l_sources_.clone()).iterator(); ite.hasNext();) {
            source = (PrivilegeSource)ite.next();
            source.removeEorPrivilegeToSource(this);
        } // for ite
        for(Iterator ite = ((ArrayList)l_targets_.clone()).iterator(); ite.hasNext();) {
            target = (PrivilegeTarget)ite.next();
            target.removeEorPrivilegeToTarget(this);
        } // for ite
    } catch (BaseError b) {
        throw new UpDateError(b.toString());
    }
    
    sourceName_ = null;
    targetName_ = null;
    nameOfSourceAcs_ = null;
    nameOfTargetAcs_ = null;
    l_sources_ = null;
    l_targets_ = null;
    l_internalAllowedSourceTypes_ = null;
    l_internalAllowedTargetTypes_ = null;

    // ACSTreeBaseListener.changeOnePrivilege() gets the previous detailled name.
    // for an external one-to-one privilege, source.removeEorPrivilegeToSource(this)
    // has called removeEorPrivilegeSourceLink() which set a non-external name
    // in the ACSTree, with changedDetailledName_ true. Changing changedDetailledName_
    // to false allows ACSTreeBaseListener.changeOnePrivilege() finds this non-external name.
    if (reset) changedDetailledName_ = false;

    // removes the privilege in the ACS. Throws UpDateError.
    // getPreviousDetailledName() still operates.
    ((ACSImpl)aCS_).removeEorPrivilege(this);

    source = null;
    target = null;
    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * This method finalizes all the sources and all the targets, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called for EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {
    if (aCS_ == null ) return;
          // for testing and tracking
          //System.out.println("Call of PrivilegeImpl.finalizeForUser() in: "+getName());

    // finalizes the sources and targets
    for (Iterator ite = ((ArrayList)l_sources_.clone()).iterator(); ite.hasNext();) {
        ((EligiblePartyImpl)ite.next()).finalizeFromDeserialization();
    }
    for (Object obj : (ArrayList)l_targets_.clone()) {
        if (obj instanceof ResourceImpl)
            ((ResourceImpl)obj).finalizeFromDeserialization();
        if (obj instanceof VirtualFolderImpl)
            ((VirtualFolderImpl)obj).finalizeFromDeserialization();
    } // for obj


    sourceName_ = null;
    targetName_ = null;
    nameOfSourceAcs_ = null;
    nameOfTargetAcs_ = null;
    l_sources_ = null;
    l_targets_ = null;
    l_internalAllowedSourceTypes_ = null;
    l_internalAllowedTargetTypes_ = null;

    try {
        super.finalizeForBase();
    }
    catch (UpDateError u) {
        throw new InternalError(u.toString());
    }
  } // finalizeFromDeserialization()

  /**
   * This method is to use only for closed LINKED_FOR_ONE_TO_ONE privileges
   * in the relevant ACS list.
   * It cuts all the external references,
   * which can be from and to another ACS, without updating
   * these objects as finalizeForUser() does.
   * It set to null almost all variables, so the instance becames empty.
   * It is called by ACSImpl.checkClosedExternalLinkedPrivileges(),
   * to delete the instance in ACSImpl.l_ClosedPrivileges_.
   */
   protected synchronized void finalizeClosedPrivilege() {
        if (aCS_ == null ) return;
        if (!mainType_.equals(LINKED_FOR_ONE_TO_ONE)) return;

              // for testing and tracking
              //System.out.println("Call of PrivilegeForLinksImpl.finalizeForUser() in: "+
                //getName());

        aCS_ = null;
        acsName_ = null;
        detailledName_ = null;
        nickName_ = null;
        seeWhy_ = null;
        comment_ = "";
        l_rights_ = null;
        mainType_ = null;
        positiveRight_= true;
        l_sources_.clear();
        l_sources_ = null;
        l_targets_.clear();
        l_targets_ = null;
        sourceName_ = null;
        targetName_ = null;
        nameOfSourceAcs_ = null;
        nameOfTargetAcs_ = null;
        l_internalAllowedSourceTypes_ = null;
        l_internalAllowedTargetTypes_ = null;
    } // finalizeClosedPrivilege()  
  
  /** 
   * TO REVIEW
  * This method provides a copy of the instance,
  * in the same ACS or another one, on the contrary of
  * PrivilegeForTypeImpl.copy().
  * If the copy is provided inside the privilege ACS, all the source
  * and target links are copied, with an update of the original
  * sources and targets on which is so added one new privilege link.
  * If the copy is provided outside the privilege ACS, the copy
  * is linked to sources and targets in its ACS if instances
  * with the same names may be found. If not, the link is not
  * copied.
  * Comment is not copied.
  * @param _name is the copy name must be a one-component name,
  *   and the component is the nick name
  *   of the new privilege. May be equals to the nick name
  *   of this privilege, but only for another ACS.
  * @param _acs is the ACS of the copy to create. May be different
  *   to the ACS of this instance.
  * @param _parent must be null here.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in its ACS.
  * @return Object is a copy of the object, in the same class.
  * @exception CreateError if _name or _acs is null, or _parent is not null,
  *   or _name is not a one-component name.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Node _parent,
    Boolean _toRegisterInManager) throws CreateError {
  
        if (_name == null || _acs == null || _parent != null ||
          _name.noOfComponents() != 1)
            throw new CreateError(
                " in PrivilegeForLinksImpl.copy(), _name or _acs is null,"+
                " or _parent is not null,"+
                " or _name is not a one-component name.");
        
        String nickName = (String) _name.component(0);
        
        PrivilegeForLinksImpl result;
        PrivilegeTarget       priTarget;
        
        if (_acs == aCS_) {
            if (oneRoleOneTarget_) {
                result = new PrivilegeForLinksImpl(
                     _acs, 
                     PrivilegeAbst.LINKED_FOR_ONE_TO_ONE,
                     nickName, getL_Rights(),
                     secondType_,
                     _toRegisterInManager.booleanValue());
            } else if (ifAllSources_) {
                result = new PrivilegeForLinksImpl(
                     _acs,
                     PrivilegeAbst.LINKED_IF_ALL_SOURCES,
                     nickName, getL_Rights(),
                     secondType_,
                     _toRegisterInManager.booleanValue());
            } else {
                result = new PrivilegeForLinksImpl(
                     _acs,
                     PrivilegeAbst.LINKED,
                     nickName, getL_Rights(),
                     secondType_,
                     _toRegisterInManager.booleanValue());
            } // else (oneRole
                
            // registers the new privilege in the same sources
            try {
                if (this instanceof PrivilegeForType) {
                    for(Iterator ite = l_sources_.iterator(); ite.hasNext();) {
                        ((PrivilegeSource)ite.next()).addEorPrivilegeForTypeToSource(
                                (PrivilegeForType)result);
                    }
                }
                else {
                    for(Iterator ite = l_sources_.iterator(); ite.hasNext();) {
                        ((EligiblePartyImpl)ite.next()).addEorPrivilegeForLinksToSource(
                                (PrivilegeForLinks)result);
                    }
                }
            } catch (BaseError b) {
                throw new CreateError(b.toString());
            }
            
            result.setSeeWhy(seeWhy_);
            // registers the new privilege in the same targets
            try {
                if (this instanceof PrivilegeForType) {
                    for(Iterator ite = l_targets_.iterator(); ite.hasNext();) {
                        ((PrivilegeTarget)ite.next()).addEorPrivilegeForTypeToTarget(
                                (PrivilegeForType)result);
                    }
                }
                else {
                    for(Iterator ite = l_targets_.iterator(); ite.hasNext();) {
                        priTarget = (PrivilegeTarget)ite.next();

                        if (priTarget instanceof Resource)
                            ((ResourceImpl)priTarget).addEorPrivilegeForLinksToTarget(
                                    (PrivilegeForLinks)result);
                        else if (priTarget instanceof VirtualFolder)
                            ((VirtualFolderImpl)priTarget).addEorPrivilegeForLinksToTarget(
                                    (PrivilegeForLinks)result);

                    } // for
                } // else
            } catch (BaseError b) {
                throw new CreateError(b.toString());
            }
            priTarget = null;
        } // if (_acs ==
        else {
            ImmutableName ep_name = null;

            if (oneRoleOneTarget_) {
                EligibleParty copyEP = null;
                EligibleParty thisEP = null;
                PrivilegeTarget thisTarget = null;
                PrivilegeTarget copyTarget = null;
                
                if (!l_sources_.isEmpty()) {
                      thisEP = (EligibleParty) l_sources_.get(0);
                      // if the copy is for another ACS, try to find copySource
                      // with the same name last components
                      ep_name = ((Name)_acs.getName()).extended( 
                          thisEP.getName().lastComponent() );
                      if (_acs.containsOwnEligiblePartyName(ep_name))
                          copyEP = (EligibleParty)_acs.getEorM_OwnEligibleParties().
                              get(ep_name);
                      else System.out.println("in PrivilegeForLinksImpl.copy(), "+
                          "the source has not been found in the target ACS for the copy of "+
                            detailledName_);
                } // if (!l_sources_.isEmpty()
                if (!l_targets_.isEmpty()) {
                      thisTarget = (PrivilegeTarget) l_targets_.get(0);
                      // if the copy is for another ACS, try to find copyTarget
                      // with the same name last components
                      ep_name = ((Name)_acs.getName()).extended( 
                          thisTarget.getName().lastComponent() );
                      
                      // TO ADD:
                      //if (_acs.containsOwnVirtualFolderName(ep_name))
                          //copyTarget = (PrivilegeForLinksTarget)_acs.getEorM_OwnVirtualFolders().
                            //  get(ep_name);
                      // else
                      if (_acs.containsOwnResourceName(ep_name))
                          copyTarget = (PrivilegeTarget)_acs.getEorM_OwnResources().
                              get(ep_name);
                      else System.out.println("in PrivilegeForLinksImpl.copy(), "+
                          "the copy target has not been found in the target ACS for the copy of "+
                            detailledName_);
                } // if (!l_sources_.isEmpty()
                
                result = new PrivilegeForLinksImpl(
                     _acs, PrivilegeAbst.LINKED_FOR_ONE_TO_ONE,
                     nickName, getL_Rights(),
                     secondType_,
                     _toRegisterInManager.booleanValue());
                copyEP = null;
                thisEP = null;
            } else if (ifAllSources_) {
                result = new PrivilegeForLinksImpl(
                     _acs, PrivilegeAbst.LINKED_IF_ALL_SOURCES,
                     nickName, getL_Rights(),
                     secondType_,
                     _toRegisterInManager.booleanValue());
            } else {
                result = new PrivilegeForLinksImpl(
                     _acs, PrivilegeAbst.LINKED,
                     nickName, getL_Rights(),
                     secondType_,
                     _toRegisterInManager.booleanValue());
            } // else (oneRole  
            
            PrivilegeSource copySource = null;
            PrivilegeSource thisSource = null;
            // registers the new privilege in the ACS sources
            try {
                for(Iterator ite = l_sources_.iterator(); ite.hasNext();) {

                   thisSource = (PrivilegeSource)ite.next();
                  // if the copy is for another ACS, try to find copySource
                  // with the same name last components
                  ep_name = ((Name)_acs.getName()).extended( 
                      thisSource.getName().lastComponent() );
                  copySource = (PrivilegeSource)
                     _acs.getEorM_EligibleParties().get(ep_name);
                  if (copySource != null) {
                        if (result instanceof PrivilegeForType)
                            copySource.addEorPrivilegeForTypeToSource(
                                    (PrivilegeForType)result);
                        else
                            ((EligiblePartyImpl)copySource).addEorPrivilegeForLinksToSource(
                                    (PrivilegeForLinks)result);
                  } else System.out.println("in PrivilegeForLinksImpl.copy(), "+
                      "the copy EP has not been found in the target ACS for the copy of:"+
                         ep_name);

                } // for(Iterator ite 
            } catch (BaseError b) {
                throw new CreateError(b.toString());
            }
            copySource = null;
            thisSource = null;

            result.setSeeWhy(seeWhy_);
            // whatever the type, the targets are to be copied
            PrivilegeTarget copyTarget = null;
            PrivilegeTarget thisTarget = null;
            // registers the new privilege in the ACS targets
            try {
                for(Iterator ite = l_targets_.iterator(); ite.hasNext();) {

                   thisTarget = (PrivilegeTarget)ite.next();
                  // if the copy is for another ACS, try to find copyTarget
                  // with the same name last components
                  ep_name = ((Name)_acs.getName()).extended( 
                      thisTarget.getName().lastComponent() );
                  // try to find in folders, then in resources
                  copyTarget = (PrivilegeTarget)
                     _acs.getOneVirtualFolder(ep_name);
                  if (copyTarget == null)
                      copyTarget = (PrivilegeTarget)
                        _acs.getEorM_Resources().get(ep_name);
                  if (copyTarget != null) {
                        if (result instanceof PrivilegeForType)
                            copyTarget.addEorPrivilegeForTypeToTarget((PrivilegeForType)result);
                        else {
                            if (copyTarget instanceof Resource)
                                ((ResourceImpl)copyTarget).addEorPrivilegeForLinksToTarget(
                                        (PrivilegeForLinks)result);
                            else if (copyTarget instanceof VirtualFolder)
                                ((VirtualFolderImpl)copyTarget).addEorPrivilegeForLinksToTarget(
                                        (PrivilegeForLinks)result);
                        } // else
                  } else System.out.println("in PrivilegeForLinksImpl.copy(), "+
                      "the copy EP has not been found in the target ACS for the copy of "+
                         ep_name);

                } // for(Iterator ite 
            } catch (BaseError b) {
                throw new CreateError(b.toString());
            }
            copyTarget = null;
            thisTarget = null;
            ep_name = null;
        } // else (_acs ==
        
        // todo :     sourceName_ 
        //targetName_ nameOfSourceAcs_ nameOfTargetAcs_
        
        return result;
  } // copy
  
  /*
   * This provides a new empty instance of a typed privilege, with a null type.
   * @return Privilege
   */
  public Privilege getEmptyInstance() {
      return EMPTY_INSTANCE;
  }
  
  //
  // ---------------- protected methods -----------------------------------
  //   
  
  /**
   * Called by ACSImpl.closeExternalLinkedPrivileges() only for the external privileges,
   * thus for the LINKED_FOR_ONE_TO_ONE privileges,
   * when the links to other ACS have to be cut and cloned.
   * No copy of the sources and target lists, nor the detailled name.
   * Caution: the copy is not registered in its ACS, nor in its sources and targets.
   * @return Object without change in the external references
   *  but copy of these references.
   */
    protected PrivilegeForLinksImpl copyForExternal() {
        if (!isExternalPrivilege())
            throw new InternalError("PrivilegeForLinks.copyForExternal() called "+
                    "on a non-external privilege");

      //try {
          PrivilegeForLinksImpl clon = (PrivilegeForLinksImpl) super.clone();

          clon.aCS_ = null; // from PrivilegeAbst
          
          clon.ifAllSources_ = false;
          clon.oneRoleOneTarget_ = true;
          
          //clon.l_sources_ = (ArrayList) l_sources_.clone();
          //clon.l_targets_ = (ArrayList) l_targets_.clone();
          
          if (sourceName_ != null)
              clon.sourceName_ = (ImmutableName)sourceName_.clone();
          else clon.sourceName_ = null;
          if (targetName_ != null)
              clon.targetName_ = (ImmutableName)targetName_.clone();
          else clon.targetName_ = null;
          if (nameOfSourceAcs_ != null)
              clon.nameOfSourceAcs_ = (ImmutableName)nameOfSourceAcs_.clone();
          else clon.nameOfSourceAcs_ = null;
          if (nameOfTargetAcs_ != null)
              clon.nameOfTargetAcs_ = (ImmutableName)nameOfTargetAcs_.clone();
          else clon.nameOfTargetAcs_ = null;
          
          return clon;

      //} catch (CloneNotSupportedException e) {
	//    // this shouldn't happen, since we are Cloneable
  	  //  if (aCS_ != null) {
  	    //        aCS_.getEorBaseError().addException(e.toString());
  	  //  }
        //throw new InternalError(e.toString());
      //} // catch
    } // copyForExternal()


} // PrivilegeForLinksImpl