//Title:        Access Road
//Version:      0.7.1
//Copyright:    2006, 2010, 2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;
import ARoad0.Gui1.DialogBox;

import java.util.Iterator;
import java.util.Map;
import java.util.Arrays;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

 /**
 * This class is responsible for modeling a set of rights from a source with
 * a given type, to an access target with another given type. 
 * A source has rights upon a target through a typed privilege if the privilege source type 
 * is the type of the given source, and if the privilege target type is the type 
 * of the given target. One typed privilege is designed to produce rights 
 * for numerous sources and numerous targets, and without ability to change them
 * except through a straight deletion of the typed privilege.
 * The rights of a privilege are immutable after its creation,
 * while the AclEntry has immutable source and target, but with editable rights
 * after the creation. <P>
 * A typed privilege is managed by an ACS, and it is registered by the program
 * in the relevant sources and targets. They may be external ones (that is, their ACS is
 * not the Privilege ACS) only for TYPED_FOR_SEED privileges, and then the privilege
 * is not registered in its sources and targets.
 * Access sources are usually EligibleParties, but not only. This allows to
 * give a PrivilegeForType to a Directory for example. <P>
 * The ACSObject type is the main criteria to apply the privilege to sources and 
 * targets. However, the prefix '&lt;immutable&gt;' in a type does not change the
 * PrivilegeforTypes of the ACSObject.
 * There are the four privilege subtypes: TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC,
 * and TYPED_FOR_SEED.
 * TYPED_CLASSED privilege is a TYPED privilege with contraints on source and target classes. 
 * TYPED_CLASSED_SPECIFIC is a TYPED_CLASSED privilege for named source and target.
 * TYPED_FOR_SEED is a TYPED privilege reserved for the authorization servers and
 * for external sources and targets.
 * CAUTION: the gWork package manages only the TYPED PrivilegeForType, in this version.
 * <P>
 * Following the ACS policy, sources and targets may have static or dynamic types.
 * 'dynamic' means that the user may change the type after the object creation.
 * If the source type (or the target type) is set '&lt;undefined&gt;' in the privilege,
 * then there is no restriction for the relevant type.
 * It may have some limitations about the allowed source 
 * and target types, and this is controlled by the ACS.
 * <P>
 * The relevant sources and targets register/unregister the typed privilege when the source,
 * the target, the type or the privilege is created/deleted. Then, sources and targets know
 * their typed privileges, but typed privileges do not know the sources and
 * targets which use them. To detect the applicable typed privileges between
 * a source and a target, the two privilege lists are compared through '==', not equals().
 * But the relevant sources and targets do NOT register/unregister TYPED_FOR_SEED privileges,
 * which have external sources and targets having no link to these privileges.
 * <P>
 * When a target is created or deleted, it calls the method
 * fireOneTypedPrivilegeTargetEvent() in its typed privileges, which fires
 * an event 'OneTypedPrivilegeTarget' to its listeners.
 * This is used by gWork.RightsMediatorImpl to detect when a view may have to change its rights.
 * <P>
 * This is a javabean with the following bound properties: <P>
 * - 'PrivilegeSourceClass' <P>
 * - 'PrivilegeTargetClass' <P>
 * - 'OneTypedPrivilegeTarget' <P>
 * The privilege key is used by the ACS to find a privilege.
 * The key is the couple (main type, nick name).
 * The main privilege value is an immutable list of rights.
 * The privilege gets positive or negative rights: it grants or denies accesses.
 * The rights and the key CANNOT be updated after the privilege creation,
 * nor the rights sens (grant or deny).
 * There is a 'See Why' description to complement the comment.
 * <P>
 * Sources and target may be any ACSObject, to allow to
 * give a privilege to a Directory, and to allow certificate seeds in Authorization Servers.
 * A Privilege instance is always dedicated to one ACS.
 * The privileges may be applicable to external (from other ACS) objects only for
 * TYPED_FOR_SEED privileges. In the other cases, privileges are limited to their own ACS. <P>
 * With a TYPED, a TYPED_CLASSED or a TYPED_CLASSED_SPECIFIC privilege,
 * the standard rights are never classed nor typed (not possible if they are metarights).
 * With a TYPED_CLASSED privilege, the source and target classes are constrained,
 * in addition to their type restrictions. For instance, the access source may have
 * to be an ImmutableEligibleParty, like many other right types, and the target
 * may have to be a Resource.
 * A TYPED_CLASSED_SPECIFIC privilege is reserved to a named source and a named target.
 * Their two names are put in the privilege source or target type, while
 * the source an the target do not register the privilege.
 * At the construction of a classed privilege, it is mandatory for the user to
 * call setSourceClassOneTime() and setTargetClassOneTime() to define the source class
 * and the target class. Otherwise, PrivilegeAbst.getL_Rights() returns an empty array. <P>
 * An access control seed is the ACS data which allows an ACS,
 * as an Authorization Server, to provide a certificat at the request of external actors.
 * The true access control right is a certificate or an equivalent.
 * It is then built up by the gWork package from the seed given by the Authorization Server.
 * A TYPED_FOR_SEED privilege is then used by the Authorization Server,
 * where the criterion to deliver the certificate is put in the privilege type.
 * This criteria is applicable to the source and to the target. It has generally
 * the following structure: Basic_FullName.toString().
 * The Authorization Server has to not use any constraint on the target types in privileges.
 * The right is translated and stored by gWork for any access source
 * which requests to the Authorization Server and is authorized to receive it.
 * A TYPED_FOR_SEED privilege is always external. <P>
 * This class has subclasses in the AcsAddon packages.
 * Most of the variables are protected.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBase.PrivilegeAbst
 * @see ARoad0.gBase.PrivilegeForLinksImpl
 * @version 0.7.1
 */
public class PrivilegeForTypeImpl extends PrivilegeAbst implements PrivilegeForType {
   
   /**
    * if true, the instance has the TYPED_CLASSED or a TYPED_CLASSED_SPECIFIC type
    */
    protected boolean classed_;
   /**
    * if true, the instance has the TYPED_CLASSED_SPECIFIC type
    */
    protected boolean specific_;
   /**
    * if true, the instance has the TYPED_FOR_SEED type
    */
    protected boolean seed_;
    /**
    * The source type. May be null.
    */
    protected String sourceType_;
   /**
    * The target type. May be null.
    */
    protected String targetType_;
   /**
    * Class of the privilege users, if TYPED_CLASSED or TYPED_CLASSED_SPECIFIC. 
    * May be null, even if classed.
    * This class may not be an EligibleParty.
    */
    protected Class sourceClass_ = null;
   /**
    * Class on which the privilege is applied, if TYPED_CLASSED or TYPED_CLASSED_SPECIFIC.
    * May be null, even if classed.
    * This class may not be a Resource.
    */
    protected Class targetClass_ = null;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final PrivilegeForTypeImpl EMPTY_INSTANCE = new PrivilegeForTypeImpl();
    
  /** variable for the JDK 2 serialization
  * @serial "classed_",Boolean.TYPE),
  * ("specific_",Boolean.TYPE,
  * ("seed_",Boolean.TYPE)),
  * ("sourceType_",String.class),
  * ("targetType_",String.class),
  * ("targetClass_",Class.class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("classed_",Boolean.TYPE),
      new ObjectStreamField("specific_",Boolean.TYPE),
      new ObjectStreamField("seed_",Boolean.TYPE),
      new ObjectStreamField("sourceType_",String.class),           
      new ObjectStreamField("targetType_",String.class),           
      new ObjectStreamField("sourceClass_",Class.class),
      new ObjectStreamField("targetClass_",Class.class)
  };

 /*
  * -------------------------------- Constructors -----------------------------
  */

    /**
     * Constructor to use only as a transient value for initialization, or
     * used from ARoad0.gBase.PrivilegeImplBeanInfo.getPropertyDescriptors()
     * in ARoad0.Gui1.CollectionPropertyEditor.getValue().
     */
    public PrivilegeForTypeImpl() {
    }

  /**
   * This is the only way to request the creation of a typed privilege instance from the GUI.
   * Calls the second constructor.
   * @param _aCS owns this Privilege.
   *   Non-null External Object Reference.
   * @param _nickName short privilege name. Cannot be null.
   * @param _sourceType type of the object which have the access rights.
   * Cannot be null, but '&lt;undefined&gt;' is a neutral value.
   * For a TYPED_FOR_SEED privilege or a TYPED_CLASSED_SPECIFIC privilege,
   * it has to follow the format Name.toString().
   * @param _targetType type of the object on which access rights are applied.
   * Cannot be null, but '&lt;undefined&gt;' is a neutral value.
   * For a TYPED_FOR_SEED privilege or a TYPED_CLASSED_SPECIFIC privilege,
   * it has to follow the format Name.toString().
   * @param _mainType must be one of the PrivilegeAbst constant
   * @param _l_rights StringRight array of privilege rights.
   *  May be null, but a null value is not allowed at the first position.
   *  Creates an empty array with a length of 0 if the argument is null.
   *  All rights have to be all positive or all negative. If the type is
   *  PrivilegeAbs.TYPED, rights have to be non-classed, non-typed rights.
   * @param _secondType the second type which drives the rights and the immutability,
   * just like a Basic type. If null, the internal value is set to "&lt;undefined&gt;".
   * @exception CreateError if this privilege already exists
   *   (_aCS.newEorPrivilege() throws it), if the rights or the
   *   privilege type is not correct, or if _nickName contains '::'
   *   or '>>', or if a parameter is null, or if a relevant source or a target
   *   does not register this privilege, or if typed privilege with classed or typed rights.
   */
   public PrivilegeForTypeImpl(
      ACS _aCS, String _nickName, String _sourceType, String _targetType,
      String _mainType, StringRight[] _l_rights,
      String _secondType) throws CreateError {
        this(_aCS, _sourceType, _targetType,
            _nickName, _mainType, _l_rights, _secondType, true);
   }

  /**
   * This is the only way to create a typed privilege, with one of the four possible types.
   * A TYPED privilege is the most simple one.
   * A TYPED_CLASSED privilege is a TYPED privilege with contraints on source and target classes.
   * A TYPED_FOR_SEED privilege is used as a certificate seed,
   * where the criteria to deliver the certificate is put in the two types.
   * This criteria is applicable to the source and to the target. It has
   * the following structure: Basic_FullName.toString(). <P>
   * A TYPED_CLASSED_SPECIFIC privilege is reserved to a named source and a named target.
   * Their two names are put in the privilege source or target type, while
   * the source an the target do not register the privilege.
   * This criteria is applicable to the source and to the target. It has
   * the following structure: Basic_FullName.toString(). <P>
   * The relevant sources and targets register/unregister the typed privilege when the source,
   * the target or the privilege is created/deleted, so sources and targets know
   * their typed privileges, but typed privileges do not know the sources and
   * targets which use them. This is also true for TYPED_FOR_SEED privileges,
   * which have external sources and targets.
   * If the instance is classed, displays a message to the user to recall him
   * that the source class and the target class have to be set. For this purpose,
   * calls ARoad0.Gui1.DialogBox.showOkMessage().
   * @param _aCS owns this Privilege.
   *   Non-null External Object Reference.
   * @param _sourceType type of the object which have the access rights.
   *   Cannot be null, but '&lt;undefined&gt;' is a neutral value.
   *   If it is a TYPED_FOR_SEED or a TYPED_CLASSED_SPECIFIC privilege,
   *   the argument has to use the format Name.toString().
   *   Otherwise, if the argument contains 'immutable ', this word is removed in the type.
   * @param _targetType type of the object on which access rights are applied.
   *   Cannot be null, but '&lt;undefined&gt;' is a neutral value.
   *   If it is a TYPED_FOR_SEED or a TYPED_CLASSED_SPECIFIC privilege,
   *   the argument has to use the format Name.toString().
   *   Otherwise, if the argument contains '&lt;undefined&gt;', this word is removed in the type.
   * @param _nickName short privilege name. Cannot be null.
   * @param _mainType must be one of the PrivilegeAbst constant
   * @param _l_rights StringRight array of privilege rights.
   *  May be null, but a null value is not allowed at the first position.
   *  Creates an empty array with a length of 0 if the argument is null.
   *  All rights have to be all positive or all negative. If the type is
   *  PrivilegeAbs.TYPED, rights have to be non-classed, non-typed rights.
   * @param _secondType the second type drives the rights and the immutability,
   * just like a Basic type. If null, the internal value is set to '&lt;undefined&gt;'.
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if this privilege already exists,
   * if the rights or the privilege type is not correct,
   * or if the nick name contains '::' or '&gt;&gt;', contains a character that is not a letter,
   * if a null right or a right which is not authorized.
   * if a relevant source or a target does not register this privilege,
   * or if it is a typed privilege with classed or typed rights.
   */
   public PrivilegeForTypeImpl( 
      ACS _aCS, String _sourceType, String _targetType,
      String _nickName, String _mainType, StringRight[] _l_rights, String _secondType,
      boolean _toRegisterInManager) throws CreateError {

        super(_aCS, _nickName, _mainType, _l_rights, _secondType);

        if (_sourceType == null || _targetType == null ||
          _sourceType.length() == 0 || _targetType.length() == 0 ) {
             try {
                 super.finalizeForBase();
             }
             catch (UpDateError u) {
                 throw new CreateError(" Null source type or null target type.");
             }
             throw new CreateError(" Null source type or null target type.");
        } // if

        String sourceType = null;
        String targetType = null;
        int    index = -1;

        // remove '<immutable> ' in the types, if necessary
        if (_mainType.equals(PrivilegeAbst.TYPED) || 
             _mainType.equals(PrivilegeAbst.TYPED_CLASSED)) {

                sourceType = UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(_sourceType);

                targetType = UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(_targetType);
        } // if (_mainType
        else {
            sourceType = _sourceType;
            targetType = _targetType;
        }

        detailledName_ = buildUpDetailledName(mainType_, nickName_, l_rights_[0].getRight(),
                sourceType, targetType);

        if (aCS_.getOnePrivilegeFromDetailledName(detailledName_) != null) {
            try {
                super.finalizeForBase();
            }
            catch (UpDateError u) {
                throw new CreateError(" In the ACS, this new privilege already exists.");
            }
            throw new CreateError(" In the ACS, this new privilege already exists.");
        } // if

        // checks up the compatibility of the type, the source type and the target type
        // for the ACS policy
        String[] l_typePolicy = null;
        String[] l_typePolicy2 = null;
        String[] l_privilegePolicy = null;
        boolean  exception = true;

        if (aCS_.getM_TypeOrientedPolicy() != null) {
            
            // checks up sourceType
            l_privilegePolicy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                "EligibleParty.SecondTypesOfPrivilegeFor."+sourceType);

            if (l_privilegePolicy == null ||
                    Arrays.asList(l_privilegePolicy).contains(
                        UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(secondType_))) {
                exception = false;
            } // if
            
            if (exception) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(" The ACS sets that the second type "+_secondType+
                        " cannot be used with the source type "+sourceType);
                 }
                throw new CreateError(" The ACS sets that the second type "+_secondType+
                        " cannot be used with the source type "+sourceType);
            } // if (exception)
            
            // checks up targetType_
            l_typePolicy = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                    "Resource.ResourceType");

            if (aCS_.getManageResources() && l_typePolicy != null &&
                  Arrays.asList(l_typePolicy).contains(targetType)) {

                    l_privilegePolicy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                        "Resource.SecondTypesOfPrivilegeFor."+targetType);

                    if (l_privilegePolicy == null ||
                            Arrays.asList(l_privilegePolicy).contains(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(secondType_))) {
                        exception = false;
                    } // if
            } // if (aCS_.getManageResources()

            l_typePolicy = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                    "VirtualFolder.ResourceVFType");
            l_typePolicy2 = (String[]) aCS_.getM_TypesAndGUIPolicy().get(
                    "VirtualFolder.EPVFType");
            if (exception && aCS_.getManageVirtualFolders() && 
                 ((l_typePolicy != null && Arrays.asList(l_typePolicy).contains(targetType) ) ||
                 (l_typePolicy2 != null && Arrays.asList(l_typePolicy2).contains(targetType) )) ) {

                    l_privilegePolicy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                        "VirtualFolder.SecondTypesOfPrivilegeFor."+targetType);

                    if (l_privilegePolicy == null ||
                            Arrays.asList(l_privilegePolicy).contains(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(secondType_))) {
                        exception = false;
                    } // if
            } // if (exception &&

            if (exception) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(" The ACS sets that the second type "+_secondType+
                        " cannot be used with the target type "+targetType);
                 }
                throw new CreateError(" The ACS sets that the second type "+_secondType+
                        " cannot be used with the target type "+targetType);
            } // if (exception)        
        } // if (aCS_.getM_TypeOrientedPolicy() != null)

        // if not a TYPED_FOR_SEED,
        // exception if a right is typed or classed
        exception = false;
        if (!_mainType.equals(PrivilegeAbst.TYPED_FOR_SEED) && l_rights_ != null) {
            for (int i = 0; i < _l_rights.length; i++) {
                if (_l_rights[i].isTargetClassed() ) {
                      exception = true;
                      break;
                }
            } // for i
            if (exception) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(" In the typed privilege constructor, "+
                   "a right is typed or classed while the privilege is not a TYPED_FOR_SEED.");
                 }
                 throw new CreateError(" In the typed privilege constructor, "+
                   "a right is typed or classed while the privilege is not a TYPED_FOR_SEED.");
            } // if
        } // if (!_type.equals(PrivilegeAbst.TYPED_FOR_SEED))

        // ------------------ End of first controls -----------------------------------
        
        classed_ = false;
        sourceType_ = sourceType;
        targetType_ = targetType;
        sourceClass_ = null;
        targetClass_ = null;

        // if TYPED_FOR_SEED, source and target types have to be names
        if (_mainType.equals(PrivilegeAbst.TYPED_FOR_SEED)) {
            seed_ = true;
            specific_ = false;
            try {
               NameImpl.getNameFromString(sourceType_);
            } // try
            catch (BaseError b) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(" Call to the PrivilegeforTypeImpl constructor with an "+
                   "malformed source type for the TYPED_FOR_SEED privilege.");
                 }
                 sourceType_ = null;
                 targetType_ = null;
                 throw new CreateError(" Call to the PrivilegeforTypeImpl constructor with an "+
                   "malformed source type for the TYPED_FOR_SEED privilege.");
                 
            } // catch
            try {
               NameImpl.getNameFromString(targetType_);
            } // try
            catch (BaseError b) {
                
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(" Call to the PrivilegeforTypeImpl constructor with an "+
                   "malformed target type for the TYPED_FOR_SEED privilege.");
                 }
                 sourceType_ = null;
                 targetType_ = null;
                 throw new CreateError(" Call to the PrivilegeforTypeImpl constructor with an "+
                   "malformed target type for the TYPED_FOR_SEED privilege.");
                 
            } // catch
        } // if _type.equals(PrivilegeAbst.TYPED_FOR_SEED)
        else if (_mainType.equals(PrivilegeAbst.TYPED) ||
                _mainType.equals(PrivilegeAbst.TYPED_CLASSED)) {
            seed_ = false;
            specific_ = false;
            try {
                // registers this instance in the sources and targets
                // having the right type - 'immutable' types are processed.
                if (!sourceType_.equals("<undefined>")) {
                    Map m_sources = aCS_.getEorM_OwnEligibleParties();
                    PrivilegeSource source;
                    for (Iterator ite = m_sources.values().iterator(); ite.hasNext(); ) {
                        source = (PrivilegeSource)ite.next();

                                  // for testing and tracking
                                  //System.out.println("PrivilegeForTypeImpl constructor with "+
                                          //"the type "+sourceType_+", and it "+
                                          //"may call source.addEorPrivilegeForTypeToSource() on "+
                                          //source.getNickName()+" which has the type "+
                                          //source.getType());

                        if (sourceType.equals(
                                UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                    source.getType())) )
                            source.addEorPrivilegeForTypeToSource(this);
                    } // for ite
                    m_sources = null;
                    source = null;
                } // if
                if (!targetType_.equals("<undefined>")) {

                        Map m_targets = aCS_.getEorM_OwnResources();
                        PrivilegeTarget target;
                        for (Iterator ite = m_targets.values().iterator(); ite.hasNext(); ) {
                            target = (PrivilegeTarget)ite.next();
                            if (targetType.equals(
                                    UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                        target.getType())))
                                target.addEorPrivilegeForTypeToTarget(this);
                        } // for ite
                        m_targets = null;
                        target = null;

                        VirtualFolder[] l_targets = aCS_.getEorL_VirtualFolders();
                        for (int i = 0; i < l_targets.length; i++ ) {
                            if (targetType.equals(
                                    UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                        l_targets[i].getType())) )
                                l_targets[i].addEorPrivilegeForTypeToTarget(this);
                        } // for ite
                        l_targets = null;  

                } // if
            } catch (BaseError b) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(b.toString());
                 }
                 sourceType_ = null;
                 targetType_ = null;
                 throw new CreateError(b.toString());
            } // catch
        } // else
        else { // for PrivilegeAbst.TYPED_CLASSED_SPECIFIC, source and target types have to be names
            classed_ = true;
            seed_ = false;
            specific_ = true;

            if (!sourceType_.equals(UtilityImpl.getFirstInstance().
                    getObjectTypeWithoutImmutable(sourceType_)))
                throw new CreateError(" The source type contains the word 'immutable'.");
            
            if (!targetType_.equals(UtilityImpl.getFirstInstance().
                    getObjectTypeWithoutImmutable(targetType_)))
                throw new CreateError(" The target type contains the word 'immutable'.");

            // type must have the structure Name.toString()
            try {
                NameImpl.getNameFromString(sourceType_);
            } catch (BaseError b) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(" The source type has not the string form of a name.");
                 }
                 sourceType_ = null;
                 targetType_ = null;
                 sourceClass_ = null;
                 targetClass_ = null;
                 throw new CreateError(
                      " The source type has not the string form of a name.");
            } // if
            try {
                NameImpl.getNameFromString(targetType_);
            } catch (BaseError b) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(" The target type has not the string form of a name.");
                 }
                 sourceType_ = null;
                 targetType = null;
                 sourceClass_ = null;
                 targetClass_ = null;
                 throw new CreateError(
                      " The target type has not the string form of a name.");
            } // if
        } // if _type.equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)

        if (_mainType.equals(PrivilegeAbst.TYPED_CLASSED)) {
            classed_ = true;
        } // else

        // adds this instance to the map of privileges in the ACS
        // and sends an event "OneTypedPrivilege" to the GUI listeners.
        // May throw a CreateError if the privilege already exists.
        try {
            if (_toRegisterInManager) ((ACSImpl)aCS_).newEorPrivilege(this);
        } catch (CreateError c) {
                 try {
                     super.finalizeForBase();
                 }
                 catch (UpDateError u) {
                     throw new CreateError(c.toString());
                 }
                 sourceType_ = null;
                 targetType_ = null;
                 throw new CreateError(c.toString());         
        } // catch

        if (classed_)
            DialogBox.showOkMessage("The typed privilege has been created. "+
                    "Caution: it is a classed privilege. It must have "+
                    "a non-null source class and a non-null target class. "+
                    "It is recommanded to set them now.");

        l_typePolicy = null;
        l_typePolicy2 = null;
        l_privilegePolicy = null;
        sourceType = null;
        targetType = null;
        index = -1;
  } // PrivilegeForTypeImpl()

//
//   ------------------------ protected methods ----------------------------
//

   /**
    * This method allows to create the detailled name, unique in the ACS.
    * Used in the Privilege implementation constructor.
    * The format of the returned value is: <P>
    * privilegeType || nickName || first right || source type >> target type
    * @param _mainType one of the allowed privilege types. Never null.
    * @param _nickName privilege nick name. Never null.
    * @param _firstRight first right of the privilege. Never null.
    * @param _sourceType source type of the privilege. Never null.
    * @param _targetType target type of the privilege. Never null.
    * @return the privilege detailled name, even if all the parameters are null.
    * @exception CreateError if '||', '>>' or '::' is in _nickName
    */
    protected static String buildUpDetailledName(String _mainType, String _nickName,
            String _firstRight, String _sourceType, String _targetType) throws CreateError {

        if (_nickName.indexOf("||") != -1 || _nickName.indexOf(">>") != -1 ||
                _nickName.indexOf("::") != -1)
           throw new CreateError(" The nick name description contains "+
                   "a forbidden pattern ('||', '>>' or '::')");

        StringBuilder key = new StringBuilder();

        key.append(_mainType);
        key.append(" || ");
        key.append(_nickName);
        key.append(" || ");
        key.append(_firstRight);
        key.append(" || ");
        key.append(_sourceType);
        key.append(" >> ");
        key.append(_targetType);

        return key.toString();
    } // buildUpDetailledName()

 //
 //  --------------------------- interface methods --------------------------
 //

  /**
  * Gets the detailled name from the privilege name.
  * The format of the returned value is: <P>
  * privilegeType || nickName || first right || source type >> target type
  * Calls buildUpDetailledName().
  * Called only by PrivilegeAbst.getDetailledNameFromName().
  * @see ARoad0.gBase.PrivilegeAbst#buildUpDetailledName
  * @param _name typed privilege name
  * @param _typeIndex index of the component '|| PRIV_TYPE' in _name
  * @return the detailled name of the privilege
  * @exception BaseError if there is a wrong format for _name
  */
  protected static String getDetailledNameFromNameForType(ImmutableName _name,
          int _typeIndex) throws BaseError {
    if (_name == null || _name.noOfComponents() < 9)
        throw new BaseError(
            " The name is null or it is too short to have the right format'.");

      // initialization
    String          result = null;
    try {
        result = buildUpDetailledName(
            _name.component(_typeIndex + 1),
            _name.component(_typeIndex + 3),
            _name.component(_typeIndex + 5),
            _name.component(_typeIndex + 7),
            _name.component(_typeIndex + 9));
    } catch (CreateError c) {
        throw new BaseError(c.toString());
    } // catch

    return result;
  } // getDetailledNameFromNameForType()

  /**
   * Sets the interface which may be assignable from the effective source class.
   * It is a criteria which complements the type to select the privilege source.
   * The source class is reserved to the 2 main types TYPED_CLASSED and
   * TYPED_CLASSED_SPECIFIC of privileges. This method may be called only one time,
   * and it is mandatory to do it just after the creation of the privilege.
   * Fires the event 'PrivilegeSourceClass'.
   * @param _class end of a gBaseInterface class name, like 'Resource' for
   * the class ARoad0.gBaseInterface.Resource
   * @exception CreateError if the privilege has not the main type TYPED_CLASSED or
   * TYPED_CLASSED_SPECIFIC, if _class does not define a gBaseInterface class or
   * if it is not the first call to this method
   */
    public void setSourceClassOneTime(String _class) throws CreateError {
        if (!mainType_.equals(TYPED_CLASSED) &&
                !mainType_.equals(TYPED_CLASSED_SPECIFIC))
            throw new CreateError(" This typed privilege must have a null source class");

        if (sourceClass_ != null)
            throw new CreateError(" The source class is set one time, and it cannot be changed.");

        try {
            sourceClass_ = Class.forName("ARoad0.gBaseInterface."+_class);
        }
        catch (ClassNotFoundException c) {
            throw new CreateError(" The source class "+_class+
                    " is not found in ARoad0.gBaseInterface");
        }

        if (mainType_.equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)) {
            Name name_one;
            try {
                try {
                   name_one = NameImpl.getNameFromString(sourceType_);
                } // try
                catch (BaseError b) {
                     finalizeForBase();
                     throw new CreateError(
                       " Malformed source type for the TYPED_CLASSED_SPECIFIC privilege."+
                       " The privilege is deleted.");
                } // catch

                PrivilegeSource source = (PrivilegeSource)
                  aCS_.getEorM_OwnEligibleParties().get(name_one);

                if (source == null) {
                     finalizeForBase();
                     throw new CreateError(
                       " Malformed source type for the TYPED_CLASSED_SPECIFIC privilege."+
                       " The privilege is deleted.");
                } // if (source == null

                try {
                    source.addEorPrivilegeForTypeToSource(this);
                } catch (BaseError b) {
                     finalizeForBase();
                     throw new CreateError(
                       " Exception from the source for the TYPED_CLASSED_SPECIFIC privilege."+
                       " The privilege is deleted.");
                } // catch

                source = null;
                name_one = null;
            }
            catch (UpDateError u) {
                 u.printStackTrace();
                 throw new InternalError("PrivilegeForTypeImpl.setSourceClass() with an "+
                   "exception for the TYPED_CLASSED_SPECIFIC privilege."+
                   " The privilege should be deleted.");
            }

        } // if _type.equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)
        else {
            try {
                try {
                    if (!sourceType_.equals("<undefined>")) {
                        Map m_sources = aCS_.getEorM_OwnEligibleParties();
                        PrivilegeSource source;
                        for (Iterator ite = m_sources.values().iterator(); ite.hasNext(); ) {
                            source = (PrivilegeSource)ite.next();
                            if (sourceType_.equals(source.getType()) &&
                              sourceClass_.isAssignableFrom(source.getClass()) )
                                source.addEorPrivilegeForTypeToSource(this);
                        } // for ite
                        m_sources = null;
                        source = null;
                    } // if (!sourceType_
                } catch (BaseError b) {
                         finalizeForBase();
                         throw new CreateError(
                           " Exception from the source for the TYPED_CLASSED privilege."+
                           " The privilege is deleted.");
                } // catch
            }
            catch (UpDateError u) {
                 u.printStackTrace();
                 throw new InternalError(
                   " Exception for the TYPED_CLASSED privilege."+
                   " The privilege should be deleted.");
            }

        } // else

        try {
            firePropertyChange("PrivilegeSourceClass", null, getSourceClass());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From PrivilegeForTypeImpl.setSourceClassOneTime(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }

    } // setSourceClassOneTime()

  /**
   * Sets the interface which may be assignable from the effective targets class.
   * It is a criteria which complements the type to select the privilege targets.
   * The target class is reserved to the 2 main types TYPED_CLASSED and
   * TYPED_CLASSED_SPECIFIC of privileges. This method may be called only one time,
   * and it is mandatory to do it just after the creation of the privilege.
   * Fires the event 'PrivilegeTargetClass'.
   * @param _class end of a gBaseInterface class name, like 'Resource' for
   * the class ARoad0.gBaseInterface.Resource
   * @exception CreateError if the privilege has not the type TYPED_CLASSED or
   * TYPED_CLASSED_SPECIFIC, if _class does not define a gBaseInterface class or
   * if it is not the first call to this method
   */
    public void setTargetClassOneTime(String _class) throws CreateError {
        if (!mainType_.equals(TYPED_CLASSED) &&
                !mainType_.equals(TYPED_CLASSED_SPECIFIC))
            throw new CreateError(" This typed privilege must have a null target class");

        if (targetClass_ != null)
            throw new CreateError(" The target class is set one time, and it cannot be changed.");

        try {
            targetClass_ = Class.forName("ARoad0.gBaseInterface."+_class);
        }
        catch (ClassNotFoundException c) {
            throw new CreateError(" The target class "+_class+
                    " is not found in ARoad0.gBaseInterface");
        }

        if (mainType_.equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)) {
            try {
                Name name_two = null;
                try {
                   name_two = NameImpl.getNameFromString(targetType_);
                } // try
                catch (BaseError b) {
                         finalizeForBase();
                         throw new CreateError(
                           " Exception from the source for the TYPED_CLASSED_SPECIFIC privilege."+
                           " The privilege is deleted.");
                } // catch

                PrivilegeTarget target = (PrivilegeTarget) aCS_.getEorM_OwnResources().
                  get(name_two);
                if (target == null)
                    target = (PrivilegeTarget) aCS_.getOneVirtualFolder(name_two);

                if (target == null) {
                         finalizeForBase();
                         throw new CreateError(
                           " Exception from the source for the TYPED_CLASSED_SPECIFIC privilege."+
                           " The privilege is deleted.");
                } // if (source == null

                try {
                    target.addEorPrivilegeForTypeToTarget(this);
                } catch (BaseError b) {
                     finalizeForBase();
                     throw new CreateError(
                       " Exception from the source for the TYPED_CLASSED_SPECIFIC privilege."+
                       " The privilege is deleted.");
                } // catch

                target = null;
                name_two = null;
            }
            catch (UpDateError u) {
                 u.printStackTrace();
                 throw new InternalError("PrivilegeForTypeImpl.setTargetClass() with an "+
                   "exception for the TYPED_CLASSED_SPECIFIC privilege."+
                   " The privilege should be deleted.");
            }
        } // if _type.equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)
        else {
            try {
                try {
                    if (!targetType_.equals("<undefined>")) {

                        if (ImmutableResource.class.isAssignableFrom(targetClass_) ) {
                            Map m_targets = aCS_.getEorM_OwnResources();
                            PrivilegeTarget target;
                            for (Iterator ite = m_targets.values().iterator(); ite.hasNext(); ) {
                                target = (PrivilegeTarget)ite.next();
                                if (targetType_.equals(target.getType()) &&
                                  targetClass_.isAssignableFrom(target.getClass()) )
                                    target.addEorPrivilegeForTypeToTarget(this);
                            } // for ite
                            m_targets = null;
                            target = null;
                        } // if  ( Class.forName("ARoad0

                        else if (ImmutableVirtualFolder.class.isAssignableFrom(targetClass_) ) {
                            VirtualFolder[] l_targets = aCS_.getEorL_VirtualFolders();
                            for (int i = 0; i < l_targets.length; i++ ) {
                                if (targetType_.equals(l_targets[i].getType()) &&
                                  targetClass_.isAssignableFrom(l_targets[i].getClass()) )
                                    l_targets[i].addEorPrivilegeForTypeToTarget(this);
                            } // for ite
                            l_targets = null;
                        } // if  ( Class.forName("ARoad0

                    } // if ( targetClass_
                } catch (BaseError b) {
                         finalizeForBase();
                         throw new CreateError(
                           " Exception from the source for the TYPED_CLASSED privilege."+
                           " The privilege is deleted.");
                } // catch
            }
            catch (UpDateError u) {
                 u.printStackTrace();
                 throw new InternalError(
                   " Exception for the TYPED_CLASSED privilege."+
                   " The privilege should be deleted.");
            }
        } // else

        try {
            firePropertyChange("PrivilegeTargetClass", null, getTargetClass());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From PrivilegeForTypeImpl.setTargetClassOneTime(), "+
                    "firePropertChange() throws:");
            e.printStackTrace();
        }
        
    } // setTargetClassOneTime()

  /**
   * The full name is unique for the Access Road program.
   * Its is equal to the privilege name.
   * @return a copy of the name. Null for an empty privilege.
   */
    public ImmutableName getFullName() {
       return getName();
    }
   
    /**
     * Gets the BaseObject name of the privilege.
     * The typed Privilege name is composed of:
     * ACS:: aCS name
     * ::  || PRIV_TYPE::  main type
     * ::  || NICK_NAME::  privilege name
     * ::  || FIRST_RIGHT::  first right
     * ::  || SOUR_TYPE::  sourceType
     * ::  >> TARG_TYPE::  targetType
     * here, ':: ' is added by gBase.NameImpl.toString().
     * The reverse functions of getName() are getKeyReferencesFromName() and
     * getDetailledNameFromName().
     * This name is used only outside gBase, following BaseObject interface.
     * @return ImmutableName of the privilege, or null if it is an empty privilege
     */
    public ImmutableName getName() {

      Name name = new NameImpl("ACS");
      name.extended(aCS_.getName());
      name.extended("|| PRIV_TYPE" );
      name.extended(mainType_ );
      name.extended("|| NICK_NAME" );
      name.extended(nickName_ );
      name.extended("|| FIRST_RIGHT" );
      name.extended(l_rights_[0].getRight() );
      name.extended("|| SOUR_TYPE");
      name.extended(sourceType_);
      name.extended(">> TARG_TYPE" );
      name.extended(targetType_ );

      return name;
    } // getName() 
    
   /**
    * @return the source type. May be null.
    */
    public final String getSourceType() {
        return sourceType_;
    }

   /**
    * @return the target type. May be null.
    */
    public final String getTargetType() {
        return targetType_;
    }

    /**
     * @return boolean true if the Privilege type is
     *    PrivilegeAbst.TYPED_CLASSED or TYPED_CLASSED_SPECIFIC.
     */
    public final boolean getClassed() {
        return classed_;
    }

    /**
     * @return boolean true if the Privilege type is
     *    PrivilegeAbst.TYPED_CLASSED_SPECIFIC.
     */
    public final boolean getSpecific() {
        return specific_;
    }

    /**
     * @return boolean true if the Privilege type is
     *    PrivilegeAbst.TYPED_FOR_SEED.
     */
    public final boolean getForSeed() {
        return seed_;
    }
    
  /**
   * Gets the interface which may be assignable from the effective source class.
   * It is a criteria which complements the type to select the privilege source.
   * This property is never null when it is a TYPED_CLASSED or a
   * TYPED_CLASSED_SPECIFIC privilege, and always null otherwise.
   * 'EligibleParty' is an example of value.
   * @return the class name in gBaseInterface. May be null.
   */
  public final String getSourceClass() {
       if (sourceClass_ != null)
           return sourceClass_.getName().substring(
               sourceClass_.getName().indexOf("gBaseInterface.") + 15);
       else return null;
  }

  /**
   * Gets the interface which may be assignable from the effective target class.
   * It is a criteria which complements the type to select the privilege targets.
   * This property is never null when it is a TYPED_CLASSED or a
   * TYPED_CLASSED_SPECIFIC privilege, and always null otherwise.
   * 'Resource' is an example of value.
   * @return the class name in gBaseInterface. May be null.
   */
  public final String getTargetClass() {
       if (targetClass_ != null)
           return targetClass_.getName().substring(
               targetClass_.getName().indexOf("gBaseInterface.") + 15);
       else return null;
  }

  /**
  * Adds a listener to the bean for the events 'PrivilegeSourceClass',
  * 'PrivilegeTargetClass', 'OneTypedPrivilegeTarget'.
  * Used by Gui1.CommonPropertyEditor for the 'Comment'
  * and 'SeeWhy' properties. Overiddes the super method.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

      if (_propertyName.equals("PrivilegeSourceClass") ||
              _propertyName.equals("PrivilegeTargetClass") ||
              _propertyName.equals("OneTypedPrivilegeTarget") )
          changeSupport_.addPropertyChangeListener(_propertyName, _l);
      else super.addPropertyChangeListener(_propertyName, _l);
  }

  /**
  * Removes a listener to the bean for the events 'PrivilegeSourceClass',
  * 'PrivilegeTargetClass', 'OneTypedPrivilegeTarget'.
  * Overiddes the super method.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("OneTypedPrivilegeTarget"))
                  //  System.out.println("PrivilegeForTypeImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("PrivilegeSourceClass") ||
              _propertyName.equals("PrivilegeTargetClass") ||
              _propertyName.equals("OneTypedPrivilegeTarget") )
          changeSupport_.removePropertyChangeListener(_propertyName, _l);
      else super.removePropertyChangeListener(_propertyName, _l);
  } // removePropertyChangeListener()
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("PrivilegeForTypeImpl.getPropertyChangeListeners(),
                    // call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("PrivilegeSourceClass") ||
              _propertyName.equals("PrivilegeTargetClass") ||
              _propertyName.equals("OneTypedPrivilegeTarget") )
            return changeSupport_.getPropertyChangeListeners(_propertyName);

      else return super.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()
  
  /**
   * Used to clone an ACS.
   * Caution: the clone is not registered in its ACS, nor in its source and targets.
   * @return Object without change in the external references
   *  but copy of these references.
   */
    public Object clone() {
      //try {
          PrivilegeForTypeImpl clon = (PrivilegeForTypeImpl) super.clone();
          
          clon.classed_ = classed_;
          clon.specific_ = specific_;
          clon.seed_ = seed_;
          clon.sourceType_ = sourceType_;
          clon.targetType_ = targetType_;
          
          clon.sourceClass_ = sourceClass_;
          clon.targetClass_ = targetClass_;
          
          
          return clon;

      //} catch (CloneNotSupportedException e) {
	//    // this shouldn't happen, since we are Cloneable
  	  //  if (aCS_ != null) {
  	    //        aCS_.getEorBaseError().addException(e.toString());
  	  //  }
        //throw new InternalError(e.toString());
      //} // catch
    } // clone()

   /**
   * Do NOT use this method to compare the typed privileges of a source and a target.
   * Rather use the '==' fonction.
   * @param _obj to compare.
   * @return boolean true if _obj is equal with the same class,
   *  as implementation of PrivilegeForTypeImpl.
   */
  public synchronized boolean equals(Object _obj) {

      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass() )
          return false;

      if (!super.equals(_obj))
          return false;
      
      PrivilegeForTypeImpl obj = (PrivilegeForTypeImpl) _obj;

      if (getClassed() != obj.getClassed() ) return false;
      if (getSpecific() != obj.getSpecific() ) return false;
      if (getForSeed() != obj.getForSeed() ) return false;
      
      if (sourceType_ == null || obj.getSourceType() == null) {
          if (sourceType_ != null || obj.getSourceType() != null)
            return false;  // if
      } // if
      else {
        if (!sourceType_.equals( obj.getSourceType() )) return false; }

      if (targetType_ == null || obj.getTargetType() == null) {
          if (targetType_ != null || obj.getTargetType() != null)
            return false;  // if
      } // if
      else {
        if (!targetType_.equals( obj.getTargetType() )) return false; }
      
      if (sourceClass_ == null || obj.getSourceClass() == null) {
          if (sourceClass_ != null || obj.getSourceClass() != null)
            return false;  // if
      } // if
      else {
        if (!sourceClass_.equals( obj.getSourceClass() )) return false; }
      
      if (targetClass_ == null || obj.getTargetClass() == null) {
          if (targetClass_ != null || obj.getTargetClass() != null)
            return false;  // if
      } // if
      else {
        if (!targetClass_.equals( obj.getTargetClass() )) return false; }
      
      return true;
  } // equals(Object)

  /**
  * @return boolean true if equal to new PrivilegeForTypeImpl()
  */
  public boolean isEmpty() {
    return equals(EMPTY_INSTANCE );
  }
  
  /*
   * This provides a new empty instance of a linked privilege
   * @return Privilege
   */
  public Privilege getEmptyInstance() {
      return EMPTY_INSTANCE;
  }

   /**
   * Only LINKED_FOR_SEED privileges are external ones, among the typed privileges,
   * and this is true for them in all cases.
   * @return true if the source or the target is from another ACS
   */
  public boolean isExternalPrivilege() {
      if (seed_) return true;
      else return false;
  }

   /**
    * The format is ' Detailled name: KKKK; source type: FFFF; target type: NNNNNNNN'.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("Detailled name: ");
    sb.append(detailledName_);
    
    sb.append("; source type: ");
    sb.append(sourceType_);

    sb.append("; target type: ");
    sb.append(targetType_);

    return " "+sb;
  } // toString()

  /**
   * It sets to null almost all variables, so the instance is equals
   * to a new PrivilegeForTypeImpl().
   * The relevant sources and targets register/unregister the typed privilege when the source,
   * the target or the privilege is created/deleted, so sources and targets know
   * their typed privileges, but typed privileges do not know the sources and
   * targets which use them. This is NOT true for TYPED_FOR_SEED privileges,
   * which have external sources and targets.
   * An unused object has to be deleted to inform the others objects, and this method
   * HAS TO BE CALLED before deleting this object.
   * This method is called by finalize() and finalizeForUser().
   * @exception UpDateError if the privilege is unknwon from the ACS, the sources
   *   or the targets.
   */
  public synchronized void finalizeForBase() throws UpDateError {
    if (aCS_ == null ) return;
          // for testing and tracking
          //System.out.println("Call of PrivilegeImpl.finalizeForUser() in: "+getName());
    
    // if not TYPED_FOR_SEED, unregisters this instance in the sources and targets
    // having the right type, class, or specific
    if (!mainType_.equals(PrivilegeAbst.TYPED_FOR_SEED))  {

            if (mainType_.equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC))  {
                Name name_one;
                Name name_two;
                try {
                   name_one = NameImpl.getNameFromString(sourceType_);
                   name_two = NameImpl.getNameFromString(targetType_);
                } // try
                catch (BaseError b) {
                     throw new UpDateError("in PrivilegeforTypeImpl.finalizeForUser(), "+
                       "malformed source or target type for a TYPED_CLASSED_SPECIFIC privilege.");
                } // catch                
                PrivilegeSource source = (PrivilegeSource)
                  aCS_.getEorM_OwnEligibleParties().get(name_one);
                PrivilegeTarget target = (PrivilegeTarget) aCS_.getEorM_OwnResources().
                  get(name_two);
                if (target == null)
                    target = (PrivilegeTarget) aCS_.getOneVirtualFolder(name_two);

                if (source == null || target == null) {
                     throw new UpDateError("in PrivilegeforTypeImpl.finalizeForUser(), "+
                       "a source or a target, for a TYPED_CLASSED_SPECIFIC privilege, "+
                       "is not from the ACS.");
                } // if (source == null

                try {
                    source.removeEorPrivilegeToSource(this);
                    target.removeEorPrivilegeToTarget(this);
                } catch (BaseError b) {
                     throw new UpDateError(b.toString());
                } // catch  
                
                name_one = null;
                name_two = null;
                source = null;
                target = null;
            } // if (mainType_.equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC))
            else {
                try {            
                    if (!sourceType_.equals("<undefined>")) {

                        Map m_sources = aCS_.getEorM_OwnEligibleParties();
                        PrivilegeSource source;
                        for (Iterator ite = m_sources.values().iterator(); ite.hasNext(); ) {
                            source = (PrivilegeSource)ite.next();
                            if (sourceClass_ == null && sourceType_.equals(
                                    UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                        source.getType())))
                                source.removeEorPrivilegeToSource(this);
                            else if (sourceClass_ != null && sourceType_.equals(
                                    UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                        source.getType())) &&
                                    sourceClass_.isAssignableFrom(source.getClass()))
                                source.removeEorPrivilegeToSource(this);
                        } // for ite
                        m_sources = null;
                        source = null;

                    } // if
                    
                    if (!mainType_.equals(PrivilegeAbst.TYPED_CLASSED)) {
                        
                            Map m_targets = aCS_.getEorM_OwnResources();
                            PrivilegeTarget target;
                            for (Iterator ite = m_targets.values().iterator(); ite.hasNext(); ) {
                                target = (PrivilegeTarget)ite.next();
                                if (targetType_.equals(
                                        UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                            target.getType())) )
                                    target.removeEorPrivilegeToTarget(this);
                            } // for ite
                            m_targets = null;
                            target = null;

                            VirtualFolder[] l_targets = aCS_.getEorL_VirtualFolders();
                            for (int i = 0; i < l_targets.length; i++ ) {
                                if (targetType_.equals(
                                        UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                            l_targets[i].getType()) ))
                                    l_targets[i].removeEorPrivilegeToTarget(this);
                            } // for ite
                            l_targets = null;
                        
                    } else { // for TYPED_CLASSED privilege

                        if ( ImmutableResource.class.isAssignableFrom(targetClass_) ) {
                            Map m_targets = aCS_.getEorM_OwnResources();
                            PrivilegeTarget target;
                            for (Iterator ite = m_targets.values().iterator(); ite.hasNext(); ) {
                                target = (PrivilegeTarget)ite.next();
                                if (targetType_.equals(
                                        UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                            target.getType())) && targetClass_ != null &&
                                  targetClass_.isAssignableFrom(target.getClass()) )
                                    target.removeEorPrivilegeToTarget(this);
                            } // for ite
                            m_targets = null;
                            target = null;
                        } // if  ( Class.forName("ARoad0

                        if ( ImmutableVirtualFolder.class.isAssignableFrom(targetClass_) ) {
                            VirtualFolder[] l_targets = aCS_.getEorL_VirtualFolders();
                            for (int i = 0; i < l_targets.length; i++ ) {
                                if (targetType_.equals(
                                        UtilityImpl.getFirstInstance().getObjectTypeWithoutImmutable(
                                            l_targets[i].getType())) &&
                                  targetClass_ != null &&
                                  targetClass_.isAssignableFrom(l_targets[i].getClass()) )
                                    l_targets[i].removeEorPrivilegeToTarget(this);
                            } // for ite
                            l_targets = null;
                        } // if  ( Class.forName("ARoad0     
                     
                    } //else, for TYPED_CLASSED privilege
                    
                } catch (BaseError b) {
                    throw new UpDateError(b.toString());
                } // catch
            } // else
    } // if

    // removes the privilege in the ACS. Throw UpDateError.
    // uses the privilege properties since the privilege is an ACS
    // bound property, without privilege copy in the change property firing
    ((ACSImpl)aCS_).removeEorPrivilege(this);

    sourceType_ = null;           
    targetType_ = null;           
    sourceClass_ = null;
    targetClass_ = null;
    super.finalizeForBase();      
  } // finalizeForBase()
  
  //
  // ---------------- public methods out of the implemented interfaces -----
  //   

  /**
  * This method provides a copy of the instance
  * with another name, only for another ACS.
  * On the contrary of PrivilegeForLinksImpl.copy(),
  * the PrivilegeForType copy in the same ACS has no interest
  * without changing the rights or the types.
  * Each internal variable that is an external
  * object reference have a shallow copy or a deep copy.
  * Comment is not copied.
  * @param _name is NOT the copy name; must be a one-component name,
  *   and the component is the nick name of the copy.
  * @param _acs is the main ACS of the copy to create. Have to be different
  *   to the ACS of this instance.
  * @param _parent must be null here.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in its ACS.
  * @return Object is a copy of the object, in the same class.
  * @exception CreateError if _name or _acs is null, or _parent is not null,
  *   or _acs is this privilege ACS, or _name is not a one-component name.
  * @version 0.7.0
   **/
  public Object copy(ImmutableName _name, ACS _acs, Node _parent,
    Boolean _toRegisterInManager) throws CreateError {

        if (_name == null || _acs == null || _parent != null ||
          _acs == aCS_ || _name.noOfComponents() != 1)
            throw new CreateError(
                "in PrivilegeForTypeImpl.copy(), _name or _acs is null,"+
                " or _parent is not null, or _acs is this privilege ACS,"+
                " or _name is not a one-component name.");

        String nickName = (String) _name.component(0);

        PrivilegeForTypeImpl result;

        result = new PrivilegeForTypeImpl(
                 _acs, sourceType_, targetType_,
                 nickName,
                 mainType_, getL_Rights(), secondType_,
                 _toRegisterInManager.booleanValue());

        if (classed_) {
            result.setSourceClassOneTime(getSourceClass());
            result.setTargetClassOneTime(getTargetClass());
        }

        result.setSeeWhy(seeWhy_);
        //result.setComment(comment_);

        return result;
  } // copy

  /**
   * Gets the main components from the typed privilege name as a string,
   * returned by getName(). These key components define fully
   * the privilege instance.
   * This method is used by Gui1.PropertyEditor to remove
   * a privilege from a list displaying the privilege name in the beamer.
   * Called by ACS_Facade.getPrivilege().
   * @see PrivilegeAbst#getAcsFromName(ARoad0.gBaseInterface.ImmutableName)
   * @param _privName is the privilege name, returned by getName()
   * @return null if the name format is wrong (for PrivilegeForType or PrivilegeForLinks),
   * or an array of the 6 Privilege components:<P>
   *  - index 0: name of the ACS which manages this privilege (ImmutableName)
   *  - index 1: privilege type (String)
   *  - index 2: nick name (String)
   *  - index 3: grant or deny (String)
   *  - index 4: source type (String)
   *  - index 5: target type (String)
   * @exception CreateError if the argument or a component is null or not correct,
   *  not a privilege name string, or if a main component string
   *  is not provided by NameImpl.toString()
   */
    public Object[] getKeyReferencesFromName(ImmutableName _privName)
      throws CreateError {
        if (_privName == null || aCS_ == null) 
            throw new CreateError(" In PrivilegeForTypeImpl."+
                "getKeyReferencesFromName(), the argument or the ACS is null.");
        int index = aCS_.getName().noOfComponents();
        if (!_privName.component(0).equals("ACS")
          || !_privName.component(index+1).equals("|| PRIV_TYPE")
          || !_privName.component(index+3).equals("|| NICK_NAME")
          || !_privName.component(index+5).equals("|| SENS")
          || !_privName.component(index+7).equals("|| SOUR_TYPE")
          || !_privName.component(index+9).equals(">> TARG_TYPE"))
            throw new CreateError(" In PrivilegeForTypeImpl.getKeyReferencesFromName()"+
              " the argument is not a typed privilege name.");

        // variables
        Object[] l_result = new NameImpl[6];
        index++; // for "|| PRIV_TYPE"

        l_result[0] = _privName.subname(1, index); // ACS name
        l_result[1] = _privName.subname(index + 1, index + 2).component(0); // privilege type
        l_result[2] = _privName.subname(index + 3, index + 4).component(0); // nick name
        l_result[3] = _privName.subname(index + 5, index + 6).component(0); // grant/deny
        l_result[4] = _privName.subname(index + 7, index + 8).component(0); // source type
        l_result[5] = _privName.subname(index + 9, index + 10).component(0); // target type
        
        if (l_result[0] == null || ((Name)l_result[0]).isEmpty()
          || (!l_result[1].equals(PrivilegeAbst.TYPED) && 
                   !l_result[1].equals(PrivilegeAbst.TYPED_CLASSED) )
          || l_result[2] == null
          || l_result[3] == null
          || l_result[4] == null
          || l_result[5] == null)
            throw new CreateError(" In PrivilegeForTypeImpl.getKeyReferencesFromName()"+
              " a component of the result is not from a typed privilege name.");
        
              // for testing and tracking
              //System.out.println("in PrivilegeImpl.getKeyReferencesFromName(): "+
              //  "ACS name: "+l_result[0]+
              //  "privilege type: "+l_result[1]+
              //  "nick name: "+l_result[2]+
              //  "gant/deny: "+l_result[3]);

        return l_result;
    } // getKeyReferencesFromName() 

  /**
  * Activates the firing of a 'OneTypedPrivilegeTarget' event to all its listeners.
  * Called by each new and finalized PrivilegeTarget that uses this PrivilegeForType.
  * Noop if the two arguments are null, if the two ones are not null, if the two
  * arguments are identical.
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  public void fireOneTypedPrivilegeTargetEvent(ImmutablePrivilegeTarget _oldValue,
          ImmutablePrivilegeTarget _newValue) {

      if (_oldValue == null && _newValue == null)
          return;
      else if (_oldValue != null && _newValue != null)
          return;
      else if (_oldValue == _newValue)
          return;      
      
      firePropertyChange("OneTypedPrivilegeTarget", _oldValue, _newValue);
  } // fireOneTypedPrivilegeTargetEvent()

  //
  // ---------------- protected methods ------------------------------
  //

  /**
   * This method calls finalizeForBase(), but there is no
   * guarantee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {u.printStackTrace();}

    super.finalize();
  } // finalize()

  //
  // ---------------- private methods --------------------------------
  //

  /**
  * This method does not finalize the sources and the targets, even for a
  * TYPED_CLASSED_SPECIFIC.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called for EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {
    if (aCS_ == null ) return;
          // for testing and tracking
          //System.out.println("Call of PrivilegeForTypeImpl.finalizeFromDeserialization() in: "+
            //getName());

    sourceType_ = null;
    targetType_ = null;
    sourceClass_ = null;
    targetClass_ = null;
    try {
        super.finalizeForBase();
    }
    catch (UpDateError u) {
        throw new InternalError(u.toString());
    }
  } // finalizeFromDeserialization()

} // PrivilegeForTypeImpl