//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.BaseError;

import java.util.Map;
import java.util.Set;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Collections;
import java.util.TreeMap;

import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;

/**
 * This class is responsible for modeling a resource, that is an access target,
 * in an access control system on which it is registered.
 * The resource may have a parent in the resource tree of this ACS.
 * The resource is accessed from EligibleParties (userID, groupID, actor)
 * through AclEntries belonging to this ACS or to another one.
 * The resource has an userID (account) and a groupID (group) in this ACS.
 * It has AG rights for them and for the super-administrator of the ACS.
 * The resource is a PrivilegeTarget for PrivilegeForTypes and PrivilegeForLinks. <P>
 * Directory may manage container-for-target cascading rights to its Resource children, 
 * both for the Account/Group rights, the AclEntry rights and the Privilege rights.
 * But only one of these three inheritances is possible in a given ACS.
 * The generic packages handle fully only the AG inherited rights.
 * For the AcsAddons handling of the AclEntry or the Privilege inherited rights,
 * the generic gBase package provides the basic methods on Resource
 * to set and get the inherited rights. The generic gBase package provides
 * also the handling of the listeners for the two property changes
 * 'InheritedTargetPrivileges' and 'InheritedTargetAclEntries', but without any firing
 * of these changes. The generic gWork package handles
 * these three inherited rights in the access path search.
 * An AcsAddon package, to fully process these two inherited rights, has then only
 * to code how the Directory subclass updates the inherited rights in its children,
 * and to code the LinkRightsImpl subclass for overriding some of
 * the 'ThroughNodesTree' methods which implemente the algorithms for the rights inheritance. <P>
 * All the rights and the type are constrained by the ACS.
 * If 'immutable' is in the Resource type, the class rejects any updating on the userID,
 * groupID, parent and all the rights. Nonetheless, an AclEntry may be added wihtout
 * any removing.
 * If 'byAcsCreator' is in the type, the same behavior is set, and in addition,
 * the user cannot finalize this instance.
 * <P>
 * It is a javabean with the following bound properties:
 * 'AclRights', 'UserID', 'GroupID',
 * 'RootRights', 'GroupRights', 'OwnerRights',
 * 'EPsThroughAclEntry', 'AcsControllers', 
 * 'FromTheParentAccount', 'FromTheParentGroup',
 * 'InheritedOwnerRights', 'InheritedGroupRights',
 * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege',
 * and from the subclasses in the AcsAddon,
 * 'InheritedTargetPrivileges', 'InheritedTargetAclEntries'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched
 * in the 'set' method, and a dialog box is displayed to inform the user.
 * <P>
 * This class has several subclasses in this package and in the AcsAddon packages.
 * Most of the variables are protected.
 * @version 0.7.1
 */
public class ResourceImpl extends BasicImpl implements Resource {
  /**
  * Protected variables
  */
  protected boolean           actorActivated_ = false; // always false in this class

  protected StringRight[]     l_rootRights_; // rights of root (userID_ with getOrder() == 0)
  protected StringRight[]     l_groupRights_; // rights of groupID_
  protected StringRight[]     l_ownerRights_; // rights of userID_
  protected UserID            userID_;     // resource owner; External Object Reference
  protected GroupID           groupID_;   // resource main group; External Object Reference
  protected Directory         parent_;  // External Object Reference

  /**
   * 2 booleans for Account/Group inheritance rights, in this order.
   * Updated by the direct and indirect parents in DirectoryImpl.setAllowToGoAccross().
   * The first boolean is true if, for its userID, 
   * this resource may be accessed through the parent tree.
   * Used in get/setL_xxxRights(). The default value is true for all the booleans.
   * In the AcsAddon Ubuntu, this property is completed by a third boolean
   * to handle the 'other' property.
   */
  protected boolean[]         l_parentTreeAGRightsAllowToGoAcross_ = new boolean[] {true, true};
  
  /**
   * Private variables
   */
  private static final long   serialVersionUID__ = 0;
  //  private Name[]          l_Access_;
//  private Icon              icon_;

  protected AclEntry[]        l_aclEntry_;

  /**
   * TreeMap of inherited AclEntry rights. by default, it is not displayed to the user in the
   * generic BeanInfo class.
   * The default key is the name of the EligibleParty having the rights, extended with,
   * as last component, the String order of the parent which has given the rights.
   * For instance, the direct parent has the string '1' and its direct parent
   * has the string '2' as last name component. A subclass may change the key form,
   * and it may choose to not use this property.
   * A subclass may use the constructor TreeMap(Comparator comparator)
   * to set an ordering which is different to the natural ordering of the keys.
   * An AcsAddon may choose to not use this property while it handles the inherited ACL.
   */
  protected Map<String, StringRight[]>   m_inheritedT_ACLRights_ = new TreeMap();

  /**
   * List of the current inherited AclEntry. The default value is null.
   * This is the data source of m_inheritedT_ACLRights_.
   */
  protected ArrayList<AclEntry> l_inheritedT_AclEntries_ = null;

  /**
   * Array of ACS that control this resource through external AclEntries
   * or external LINKED_FOR_ONE_TO_ONE privileges. May be null.
   */
  protected ImmutableName[]     l_OpenAcsControllerNames_ = null;

  // --------- public static properties -------------------------------

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final ResourceImpl EMPTY_INSTANCE = new ResourceImpl();

  /**
   * Properties for the serialization
   * @serial 
   *  ("parent_",DirectoryImpl.class),
   *  ("actorActivated_",Boolean.TYPE),
   *  ("l_aclEntry_",AclEntryImpl[].class),
   *  ("l_rootRights_",StringRight[].class),
   *  ("l_groupRights_",StringRight[].class),
   *  ("l_ownerRights_",StringRight[].class),
   *  ("userID_",UserIDImpl.class),
   *  ("groupID_",GroupIDImpl.class),
   *  ("l_OpenAcsControllerNames_",NameImpl[].class),
   *  ("l_parentTreeAGRightsAllowToGoAcross_", boolean[].class),
   *  ("m_inheritedT_ACLRights_",TreeMap.class,
   *  ("l_inheritedT_AclEntries_",ArrayList.class
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("parent_",DirectoryImpl.class),
//      new ObjectStreamField("l_Access_",NameImpl[].class),
      new ObjectStreamField("actorActivated_",Boolean.TYPE),
      new ObjectStreamField("l_aclEntry_",AclEntryImpl[].class),
      new ObjectStreamField("l_rootRights_",StringRight[].class),
      new ObjectStreamField("l_groupRights_",StringRight[].class),
      new ObjectStreamField("l_ownerRights_",StringRight[].class),
      new ObjectStreamField("userID_",UserIDImpl.class),
      new ObjectStreamField("groupID_",GroupIDImpl.class),
      new ObjectStreamField("l_OpenAcsControllerNames_",NameImpl[].class),
      new ObjectStreamField("l_parentTreeAGRightsAllowToGoAcross_", boolean[].class),
      new ObjectStreamField("m_inheritedT_ACLRights_",TreeMap.class),
      new ObjectStreamField("l_inheritedT_AclEntries_",ArrayList.class)
  };

   /**
   * Default constructor. No registration in the ACS.
   * Defines "full control" default rights for the owner,
   * and no right for the group.
   */
  public ResourceImpl() {

                          // for testing purpose
                     //System.out.println("Call to ResourceImpl.ResourceImpl()");

      super();
      l_aclEntry_ = new AclEntryImpl[0];
      l_rootRights_ = new StringRight[] {BaseManagerImpl.getInstance().
              getMetaRight("|is_superadmin|")};
      l_groupRights_ = new StringRight[0];
      l_ownerRights_ = new StringRight[0];
      //l_OpenAcsControllerNames_ = new NameImpl[0];
      parent_ = null;     
  }

   /**
   * Used by the beamer editor for the ACS resources list, from the 'New' button,
   * and by the gWork.AcsFactory. Calls the other constructor.
   * @param _name ImmutableName of the resource
   * @param _acs main ACS of the resource
   * @param _parent of the resource; may be null.
   * @param _owner owns the resource; may be null.
   * @param _resourceType is the resource type which drives the rights and the immutability
   * of the instance. If null, the value is set to "&lt;undefined&gt;".
   * @exception CreateError if ACSImpl.newEorResource() throws it;
   * the instance has then to be deleted !
   */
  public ResourceImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
    String _resourceType) throws CreateError {

      this(_name, _acs, (Directory)_parent, _owner, _resourceType, true);
        
                          // for testing purpose
                     //System.out.println("Call to ResourceImpl.ResourceImpl("+
                       // "ImmutableName _name, ACS _acs, Node _parent, String _resourceType)"+
                       // " for "+_name);

  } // ResourceImpl constructor

   /**
   * Used for the creation of a resource with its type.
   * If _owner is null, the new resource account is the parent account.
   * If _owner is not null, the resource account is _owner, and its group
   * is the account primary group if it is not null and if the parent group
   * does not enforce 'ChildGID'. Else, the new resource group is the parent group.
   * If _resourceType has the format '&lt;byAcsCreator&gt; fffff', the parent
   * must have '&lt;fffff_subacs&gt;' as a component in its name.
   * If this format is the parent type format, no child adding is allowed.
   * If _resourceType has not the format '&lt;byAcsCreator&gt; fffff'
   * nor '&lt;byAcsCreator&gt;', with the ACS having 'ffffff' as a subACS type,
   * the parent must not have '&lt;fffff_subacs&gt;' as a component in its name.
   * Controls the ACS constraints for 'Directory.TypesOfChildFor.' on the parent type,
   * and 'Resource.NoAccount'.
   * Defines "full_control" default rights for the owner and "|is_superadmin|" for root.
   * Used by copy() and the subclass constructors.
   * @param _name non-null ImmutableName of the resource to create
   * @param _acs main ACS of the resource to create
   * @param _parent has the resource as child; may be null or immutable.
   * @param _owner owns the resource; if null, the parent account is used if any.
   * @param _resourceType is the resource type which drives the rights and the immutability
   * of the instance. If null, the value is set to "&lt;undefined&gt;".
   * @param _toRegisterInManager false if this method must not register
   *   the instance in its ACS.
   * @exception CreateError if ACSImpl.newEorResource() throws it, if
   *   _parent is not a DirectoryImpl or does not allow an adding of child,
   *   if the parent type is '&lt;byAcsCreator&gt; fffff',
   *   if the resource has a strong type but is not under the right parent - 
   *   the instance has then to be deleted! - or if _bridgeRef does not belong to another ACS
   *  - the new instance is then finalized before the exception is thrown.
   */
  public ResourceImpl(ImmutableName _name, ACS _acs, Directory _parent, UserID _owner,
     String _resourceType, boolean _toRegisterInManager) throws CreateError {

      super(_name, _acs);

                              // for testing purpose
                         //System.out.println("ResourceImpl constructor"+
                           // " for "+_name+", with the parent argument: "+
                            //_parent);

      l_aclEntry_ = new AclEntryImpl[0];
      l_rootRights_ = new StringRight[] {BaseManagerImpl.getInstance().
              getMetaRight("|is_superadmin|")};
      l_groupRights_ = new StringRight[0];
      l_ownerRights_ = new StringRight[0];
      //l_OpenAcsControllerNames_ = new NameImpl[0];
      parent_ = null;
          
      // variables
      String[]        l_types = null;
      boolean         keep = false;

      if (_owner != null && !_owner.getAcsName().equals(_acs.getName()))
          throw new CreateError(" The account is not from the same ACS.");
      
      if (_resourceType != null) {
          String crea = "<byAcsCreator> ";
          String tot = null;
          if (_parent != null) {
              if (_resourceType.startsWith(crea) && _resourceType.length() > crea.length()) {
                    // If _resourceType has the format '<byAcsCreator> fffff',
                    // the parent must have '<fffff_subacs>' as a component in its name.
                      tot = _resourceType.substring(crea.length());
                      if (!_parent.getName().components().contains("<"+tot+"_subacs>")) {
                          try {
                              finalizeForBase();
                          } catch (UpDateError r) {
                                // message 'the Resource  xxx is unknown in the ACS '
                                //r.printStackTrace();
                          }
                          throw new CreateError(" "+name_.lastComponent()+
                                  " has a controlled-display type but its parent name "+
                                  "does not contain "+"<"+tot+"_subacs>");
                      } // if
              } // if (_resourceType.
              else {
                    // If _resourceType has not the format '<byAcsCreator> fffff'
                    // nor '<12345678901234567890>',
                    // the parent must not have '<fffff_subacs>' as a component in its name.
                  for (Iterator ite = _parent.getName().components().iterator(); ite.hasNext(); ) {
                      tot = (String)ite.next();
                      
                                    // for testing and tracking
                                    /*
                                    if (tot.length() > 9) {
                                        System.out.println("ResourceImpl constructor, '<fffff_subacs>': "+
                                            crea + tot.substring(1).substring(0, tot.length() - 9));                      

                                        System.out.println("ResourceImpl constructor, acs returns "+
                                                "a string array for Explorer.SubAcsNode: "+
                                            (_acs.getM_SpecializedTypesAndExplorer().
                                                get("Explorer.SubAcsNode") instanceof String[]) );
                                        
                                        System.out.println("ResourceImpl constructor, acs allows the type: "+
                                            Arrays.asList(_acs.getM_SpecializedTypesAndExplorer().
                                                get("Explorer.SubAcsNode")).contains(
                                                "ResourceType."+ crea + tot.substring(1).substring(0, tot.length() - 9))); 
                                    } // if
                                     * */
 
                      if (!tot.endsWith("_subacs>") || tot.length() < 10)
                          continue;
                      l_types = (String[]) _acs.getM_TypesAndGUIPolicy().
                            get("Explorer.SubAcsNode");
                      if (l_types != null) {
                          boolean present = false;
                          for (int i = 0; i < l_types.length; i++) {
                            if (l_types[i].equals(
                                    "ResourceType."+ crea + 
                                    tot.substring(1).substring(0, tot.length() - 9))) {
                                            present = true;
                                            break;
                            } // if
                          } // for i                      

                          if (present && !_resourceType.equals("<byAcsCreator>") ) {
                                  throw new CreateError(" "+name_.lastComponent()+
                                          " has not the type '<byAcsCreator>' "+
                                          "while it has "+tot+" has (in)direct parent. "+
                                          "If '<byAcsCreator>' is not proposed as type, "+
                                          "then no child may be added to this node.");
                          } // if
                      } // if l_types
                      l_types = null;
                  } // for ite
              } // else
          } // if (_parent != null
          else if (_resourceType.startsWith(crea) && _resourceType.length() > crea.length()) {
                // If _resourceType has the format '<byAcsCreator> fffff',
                // it must have a parent with '<fffff_subacs>' as a component in its name,
                // so the parent cannot be null.
              try {
                  finalizeForBase();
              } catch (UpDateError r) {
                  throw new CreateError(" "+name_.lastComponent()+
                      " has a subACS type but is not under "+
                      "a node having the format <fffff_subacs>"+
                          " To finalize: "+r.toString());
              }
              throw new CreateError(" "+name_.lastComponent()+
                      " has a subACS type but is not under "+
                      "a node having the format <fffff_subacs>");
          } // else
          crea = null;
          tot = null;

          try {
              // calls back addEorPrivilegeForTypeToTarget(),
              // through BasicImpl.setType().
              setResourceType(_resourceType);
          }
          catch (UpDateError u) {
              throw new CreateError(u.toString());
          }
      } // if (_resourceType != null)
      else {
          try {
              // calls back addEorPrivilegeForTypeToTarget(),
              // through BasicImpl.setType().
              setResourceType("<undefined>");
          }
          catch (UpDateError u) {
              throw new CreateError(u.toString());
          }
      } // else
      
      if (_parent != null) {
          if (!(_parent instanceof DirectoryImpl))
              throw new CreateError(" The parent is not a DirectoryImpl.");
          if (!_parent.getAcsName().equals(_acs.getName()))
              throw new CreateError(" The parent is not from the same ACS.");

          // no adding in _parent if its type is '<byAcsCreator> fffff'
          // but 'immutable' and the type 'fffff <byAcsCreator>' are allowed in the parent type
          String crea = "<byAcsCreator>";
          if (_parent.getResourceType().startsWith(crea) &&
                  _parent.getResourceType().length() > crea.length()) {
                      try {
                          finalizeForBase();
                      } catch (UpDateError r) {
                        throw new CreateError(" The type of the parent "+
                            _parent.getName().lastComponent()+" does not allow any new child."+
                            "To finalize: "+r.toString());
                      }
                      throw new CreateError(" The type of the parent "+
                            _parent.getName().lastComponent()+" does not allow any new child.");
          } // if
          crea = null;

          if (aCS_.getM_TypeOrientedPolicy() != null) {
              // no adding in _parent if its type forbids the resource type
              // in the key 'Directory.TypesOfChildFor', modulo '<immutable> '
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Directory.TypesOfChildFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    _parent.getResourceType()));
                keep = false;
                if (l_types != null) {
                    StringBuilder types = new StringBuilder();

                                          // for testing purpose
                                     //System.out.println("Call to ResourceImpl constructor"+
                                       // " for "+_name+", and the 'Directory.TypesOfChildFor' policy "+
                                        //"for the parent type "+UtilityImpl.getFirstInstance().
                                        //getObjectTypeWithoutImmutable(_parent.getResourceType()));

                    for (int i = 0; i < l_types.length; i++) {

                                          // for testing purpose
                                     //System.out.println(" ------- "+l_types[i]);


                          if (l_types[i].equals(
                                  UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(type_))) {
                              keep = true;
                          }
                          types.append(l_types[i]);
                          types.append(", ");
                    } // for i
                    if (!keep) {
                          System.out.println("ResourceImpl constructor, wrong type for "+
                                  _name);
                          try {
                              finalizeForBase();
                          } catch (UpDateError r) {
                            //throw new CreateError(" Through the ACS constraint 'Directory.TypesOfChildFor', "+
                              //    "the parent accepts only children having the types "+types+
                                //  " To finalize: "+r.toString());
                          }
                          throw new CreateError(" Through the ACS constraint "+
                                  "'Directory.TypesOfChildFor', "+
                                  "the parent accepts only children having the types "+
                                  types);
                    } // if (!keep)
                    types = null;
                } // if (l_types !=
                l_types = null;
          } // if (aCS_.getM_TypeOrientedPolicy()

                              // for testing purpose
                         //System.out.println("ResourceImpl constructor"+
                           // " for "+_name+", set the parent.  ");
          
          // no action if 'this' is already known  
          ((DirectoryImpl)_parent).addEorChild(this); // send the event "Children" to the GUI
          parent_ = _parent;         
      } // if (_parent != null)
      else {

          if (aCS_.getM_TypeOrientedPolicy() != null) {
              // no adding in _parent if its type forbids the resource type
              // in the key 'Directory.TypesOfChildFor', modulo '<immutable> '
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Directory.TypesOfChildFor."+"NULL");
                keep = false;
                if (l_types != null) {
                    StringBuilder types = new StringBuilder();

                                          // for testing purpose
                                     //System.out.println("Call to ResourceImpl constructor"+
                                       // " for "+_name+", and the 'Directory.TypesOfChildFor' policy "+
                                        //"for the parent type NULL");

                    for (int i = 0; i < l_types.length; i++) {

                                          // for testing purpose
                                     //System.out.println(" ------- "+l_types[i]);


                          if (l_types[i].equals(
                                  UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(type_))) {
                              keep = true;
                          }
                          types.append(l_types[i]);
                          types.append(", ");
                    } // for i
                    if (!keep) {
                          System.out.println("ResourceImpl constructor, error on the type for "+
                                  _name);
                          try {
                              finalizeForBase();
                          } catch (UpDateError r) {
                            //throw new CreateError(" Through the ACS constraint 'Directory.TypesOfChildFor', "+
                              //    "the parent accepts only children having the types "+types+
                                //  " To finalize: "+r.toString());
                          }
                          throw new CreateError(" Through the ACS constraint "+
                                  "'Directory.TypesOfChildFor.NULL', "+
                                  "the no-parent resource may have only the types "+
                                  types);
                    } // if (!keep)
                    types = null;
                } // if (l_types != null
                l_types = null;
          } // if (aCS_.getM_TypeOrientedPolicy()
      } // else if (_parent != null)

      if (aCS_.getManageAccounts() ) { 
          if (aCS_.getM_TypeOrientedPolicy() != null && type_ != null) {
              // no adding of _owner if the resource type forbids it
              // in the key 'Resource.NoAccount'
                l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("Resource.NoAccount");
                keep = true;
                if (l_types != null) {
                    for (int i = 0; i < l_types.length; i++) {
                        if (l_types[i].equals(
                                UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(type_))) {
                            keep = false;
                            break;
                        }
                    } // for i
                }
          } // if (aCS_.getM_TypeOrientedPolicy()
          
          if (_owner != null && keep) userID_ = _owner;
          else if (_parent != null && keep) userID_ = (UserID) _parent.getEorUserID();

                              // for testing purpose
                         //System.out.println("Call to ResourceImpl constructor"+
                           // " for "+_name+", keep = "+keep+", and the account is: "+
                           // userID_.getNickName());

          if (userID_ != null) {
              ((UserIDImpl)userID_).addOwnResource(this);
              l_ownerRights_ = new StringRight[] {BaseManagerImpl.getInstance().
                                        getMetaRight("full_control")};
          }
      } // if (aCS_.getManageAccounts() )
      if (aCS_.getManageGroups() ) {
          if ((_parent == null || !_parent.getEnforceChildGID()) && userID_ != null) {

                              // for testing purpose
                         //System.out.println("Call to ResourceImpl.ResourceImpl()"+
                           // " for "+_name+", to set the main group: "+
                           // userID_.getEorPrimGroup());
              
              groupID_ = userID_.getEorPrimGroup();
          } else if (_parent != null)
              groupID_ = (GroupID) _parent.getEorGroupID();

          if (groupID_ != null) ((GroupIDImpl)groupID_).addOwnResource(this);
      }

      // updates the AG inherited rights
      if (userID_ != null) {
          if (parent_ != null) {
              l_parentTreeAGRightsAllowToGoAcross_[0] =
                  BaseUtilityImpl.withAcrossToDirectoryAGRights(parent_.getL_ownerRights());
              l_parentTreeAGRightsAllowToGoAcross_[1] =
                  BaseUtilityImpl.withAcrossToDirectoryAGRights(parent_.getL_groupRights());
          } // if parent_ != null
      } // if (userID_ != null
      
      // adds this to the map of resources in the main ACS
      // and sends an event "OneResource" to the GUI listeners.
      // may throw a CreateError
      try {
          if (_toRegisterInManager) ((ACSImpl)aCS_).newEorResource(this);
      } catch (CreateError c) {
              try {
                  finalizeForBase();
              } catch (UpDateError r) {
                throw new InternalError("From the Resource constructor: "+r);
              }
              throw c;
      }
      
  } // ResourceImpl constructor

 //
 // -----------  Methods of the implemented interfaces --------------------
 //

  
  /**
  * Provides a copy of the instance
  * with another name. Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Most of the aclEntries are copied, but not the list of ViewInBase
  * that contains this instance.
  * Caution: does not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same last component in this instance user or group names.
  * <P>
  * For aclEntries, if _result ACS is not this resource ACS, and if
  * the aclEntry ACS is this resource ACS, sets _result ACS as
  * the aclEntry ACS of the copied acl.
  * Caution: does not copy an aclEntry if: <P>
  *  - _acs is not this resource ACS, <P>
  *  - the acl eligible party ACS is this resource ACS, <P>
  *  - and there is no matched EligibleParty (with the same last components)
  *      in _acs <P>
  * Calls copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
     May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of ResourceImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public synchronized Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null || 
        (_parent != null && _parent.getEorACS() != _acs)) {
          throw new CreateError(
          "The name or the acs is null, or the bridge target is not a Resource,"+
          " or the parent ACS is not the instance ACS.");
      }
      
      ResourceImpl result = new ResourceImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());

      // copies all the aclEntries
      result = copyOwnVariables(result, false); 
      
      return result;
  } // copy(,,,)

  /**
  * Provides a copy of the instance as Copyable.copy(),
  * but without copying the internal aclEntries.
  * It is used for ACSImpl.copy(), because all the EP must be copied
  * first before the copy of all the resources with their aclEntries.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy),
  * but without copy for the internal aclEntry nor the list of ViewInBase
  * that contains this instance.
  * Called from BaseUtilitiesImpl.copyDirectoriesAndActors().
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of this class
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public synchronized Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs,
    Directory _parent, Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in ResourceImpl.copyWithoutInternalAcl(),"+
          " _name or _acs is null.");
      }

      ResourceImpl result = new ResourceImpl(
          _name, _acs, _parent, null, getResourceType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclentries copy
      
      //if (result.m_AclEntriesResources_ == null) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copyWithoutInternalAcl(,,,)

    /**
    * Defines the parent node and updates it.
    * Noop if the type contains 'byAcsCreator' or 'immutable'.
    *  The previous parent is updated, except if it is the caller
    *  of this method, or if this instance is a new clone (see clone()).
    * @deprecated finalizeForBase() uses parent_.removeEorChild()
    * @param _par is the new parent of this leaf, as an external
    *  object reference. The new parent is updated. May be null.
    * @param _caller of this method; if it is the parent or this instance, the
    *   parent is not updated
    * @exception BaseError if the parent is not own by the same acs, or
    *   InternalError if parent.setEorL_Children() throws a BaseError, or
    *   if it is immutable for its parent.
    */
  public final synchronized void setEorParent(Node _par, Object _caller)
    throws BaseError {
  } // setEorParent()

  /**
   * Gets the current parent node of the resource
   * @return a reference on the parent node. May be null.
   */
  public final ImmutableNode getEorParent() {
          return parent_;
  }

  /**
   * Gets the open ACS that manages an external AclEntry or Linked Privilege
   * for this resource. The resource ACS forbids to have both external ACL and external privilege.
   * For compatibility purpose about the version 0.7.0, this method removes 
   * the null values in the array.
   * @return the open ACS controller names. May be null, never empty nor with null items.
   */
  public synchronized ImmutableName[] getL_AcsControllers() {
       if (l_OpenAcsControllerNames_ == null ||
               l_OpenAcsControllerNames_.length == 0) return null;
       
        int                 shiftingBack = 0;
        ImmutableName[]     newArray = new ImmutableName[l_OpenAcsControllerNames_.length];
        ImmutableName[]     result = new ImmutableName[0];
                // this code works even if shiftingBack > 1
        for (int  i = 0 ; i < l_OpenAcsControllerNames_.length ; i++) {
             // deletes the empty l_OpenAcsControllerNames_[i]
            if (l_OpenAcsControllerNames_[i] == null)
                shiftingBack++;
            else newArray[i - shiftingBack] = (ImmutableName)l_OpenAcsControllerNames_[i];
        } // for i

        if (l_OpenAcsControllerNames_.length != shiftingBack) {
            result = new ImmutableName[l_OpenAcsControllerNames_.length - shiftingBack];
            System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
        }

        l_OpenAcsControllerNames_ = new NameImpl[result.length];
        System.arraycopy( result, 0, l_OpenAcsControllerNames_, 0, result.length);
        newArray = null;
        return result;
  } // getL_AcsControllers()

  /**
   * Sets the userID of the resource, and set the current userID in the own actors
   * if it was the previous userID. It is a bound property.
   * Changes the group, which becomes the primary group of the new account if it is not null.
   * Noop and error mesage to the user if the resource type starts with '&lt;byAcsCreator&gt;'
   * or '&lt;immutable&gt;', or if 'Resource.NoAccount' in the ACS container policy forbids
   * the account for this resource type.
   * Overridden in the AcsAddon Ubuntu.
   * @param _u owns this resource, in the main ACS. May be null.
   * @exception UpDateError if _u does not belong to this resource main ACS, or
   *               if it is immutable for its userID.
   */
  public synchronized void setEorUserID(ImmutableUserID _u) throws UpDateError {
      if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+" sets it as immutable for its account.");
      if (_u != null) {
        if (aCS_ == null) throw new UpDateError(
          "In ResourceImpl.setEorUserID(), the main ACS is null");
        if (_u.getEorACS() != aCS_) {
          aCS_.getEorBaseError().addException("The selected account "+
            "does not belong to the ACS of this object.");

            // for testing and tracking
            //System.out.println("Exception fromResourceImpl.setEorUserID(): "+
            //  aCS_.getEorBaseError().toString());

          // error message to be displayed to the user
          throw new UpDateError(
            "The selected account does not belong to the ACS of this object.");
        }
        boolean keep = true;
        if (aCS_.getM_TypeOrientedPolicy() != null && type_ != null) {
          // no adding of _u if the resource type forbids it
          // in the key 'Resource.NoAccount'
           String[] l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                      get("Resource.NoAccount");

           if (l_types != null) {
                for (int i = 0; i < l_types.length; i++) {
                    if (l_types[i].equals(
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(type_))) {
                        keep = false;
                        break;
                    }
                } // for i
           } // if
        } // if (aCS_.getM_TypeOrientedPolicy()
        if (!keep)
          // error message to be displayed to the user
          throw new UpDateError(
            "The ACS forbids all accounts for the resources of this type.");
      } // if
      
      UserID oldUserID = userID_;
      if (userID_ != null) ((UserIDImpl)userID_).removeOwnResource(this);
      userID_ = (UserID) _u;
      if (_u != null) {
          ((UserIDImpl)_u).addOwnResource(this);
          if (userID_.getEorPrimGroup() != null && aCS_.getManageGroups())
              setEorGroupID(userID_.getEorPrimGroup());
      } // if
  
      try {
          firePropertyChange("UserID", oldUserID, userID_);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ResourceImpl.setEorUserID(), firePropertChange() throws:");
            e.printStackTrace();
      }

      oldUserID = null;
  } // setEorUserID()

  /**
   * @return ImmutableUserID which owns this resource, in the main ACS. May be null.
   */
  public ImmutableUserID getEorUserID()	{
    return userID_;
  }

  /**
   * Sets the rights for the account which owns this resource.
   * The rights may be acsrights and metarights, positive and negative rights.
   * The ACS constraints on 'OwnerRights' are applied only if they are not null,
   * otherwise all the arguments are validated. If a new right
   * is not allowed by the ACS, it is removed without comment to the user,
   * and the other rights are kept.
   * The negative rights are subtracted from the positive rights.
   * No action if the instance has been finalized. 
   * Noop if the resource type contains 'byAcsCreator' or 'immutable'.
   * No adding of null or negative elements. The 'full_control' equivallent is 
   * the default right if there is no umask.
   * Called by the constructor to copy the parent rights.
   * Calls BaseUtilityImpl.subtractNegativeFromPositiveInOneACS() and
   * UtilityImpl.selectHeaders().
   * Overridden and called by DirectoryImpl to update l_parentTreeAGRightsAllowToGoAcross_.
   * Fires a PropertyChange for 'OwnerRights'.
   * @param _s array of acsrights and/or metarights. May be null. 
   * @exception UpDateError if the owner rights are immutable,
   * if there are more than 20 increasing chains to analyze from one lesser right.
   */
  public synchronized void setL_ownerRights(StringRight[] _s) throws UpDateError {

                // for testing and tracking
                //System.out.println("ResourceImpl.setL_ownerRights() in "+getName()+
                  //      ", entered rights at the start: ");
                //for (int i = 0; i < _s.length; i++) {
                  //  System.out.println("------------ "+_s[i]);
                //} // for i 
      
    if (l_ownerRights_ == null) return;
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
            type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its owner rights."); 
    
    // variables
    StringRight[]       l_oldRights = new StringRight[l_ownerRights_.length];
    Set<StringRight>    l_newRights = new HashSet(INITIAL_CAPACITY / 2);
    StringRight[]       l_restrictedRights = null; 
    List<StringRight>   l_limitedRights = null;
    Set<StringRight>    l_posRights = new HashSet(INITIAL_CAPACITY / 2);
    Set<StringRight>    l_negRights = new HashSet(INITIAL_CAPACITY / 2);
    StringRight         right = null;
    boolean             toThrow = false;
 
    System.arraycopy(l_ownerRights_, 0, l_oldRights, 0, l_ownerRights_.length);    
    
    try {       
        // gets the rights restriction in the ACS
        l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                   null, this, "OwnerRights", aCS_, null);
    } catch (Exception c) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setL_groupRights(), "+
                "BaseUtilityImpl.getL_AcsRestrictedRights() throws:");
        c.printStackTrace();
    } // catch
    
    if (l_restrictedRights != null) 
        l_limitedRights = new ArrayList(Arrays.asList( l_restrictedRights ));
    else
        l_limitedRights = new ArrayList(INITIAL_CAPACITY);
    l_limitedRights.add(aCS_.getFullControlRight()); // for owner only
    
    if (_s != null) {
      for (int i = 0; i < _s.length; i++) {
        if (_s[i] != null) {
            if (l_limitedRights.contains(_s[i]))
                l_newRights.add(_s[i]);
        } // if
      } // for i
    }
 
                // for testing and tracking
                //System.out.println("ResourceImpl.setL_ownerRights()in "+getName()+
                  //      ",  rights after filtering from restricted rights: ");
                //for (Iterator ite= l_newRights.iterator(); ite.hasNext(); ) {
                  //  System.out.println("------------ "+ite.next());
                //} // for i
    
    // filters the positive and negative rigths
    if (!l_newRights.isEmpty()) {
        for (Iterator ite = l_newRights.iterator(); ite.hasNext(); ) {
            right = (StringRight)ite.next();
            if (right.isGrantingRight()) l_posRights.add(right);
            else l_negRights.add(right);
        }
    } // if

                // for testing and tracking
                //System.out.println("ResourceImpl.setL_ownerRights()in "+getName()+
                  //      ", new rights: ");
                //for (int i = 0; i < l_ownerRights_.length; i++) {
                  //  System.out.println("------------ "+l_ownerRights_[i]);
                //} // for i
    
    // subtracts the negative rights to the positive ones
    if (!l_negRights.isEmpty()) 
        l_newRights = UtilityImpl.getFirstInstance().
                    subtractNegativeNotRemovingInOneACS(
           l_posRights, l_negRights, aCS_, this instanceof ImmutableDirectory);
    else l_newRights = l_posRights;

    try {    
        UtilityImpl.getFirstInstance().selectHeaders(l_newRights);
 
                // for testing and tracking
                //System.out.println("ResourceImpl.setL_ownerRights()in "+getName()+
                  //      ",  result: ");
                //for (Iterator ite= l_newRights.iterator(); ite.hasNext(); ) {
                  //  System.out.println("------------ "+ite.next());
                //} // for i
        
        // updates the owner rights
        if (!l_newRights.isEmpty()) {
          l_ownerRights_ = (StringRight[])
                  l_newRights.toArray(new StringRight[l_newRights.size()]);
        } else {
          l_ownerRights_ = new StringRight[0];
        }
     
        if (l_oldRights.length != l_ownerRights_.length )
            toThrow = true;
        else if (l_oldRights.length == l_ownerRights_.length &&
                l_ownerRights_.length == 0)
            toThrow = true;
        else {
            for (int i = 0; i < l_oldRights.length; i++) {
                if (!l_newRights.contains( l_oldRights[i] ) ) {
                    toThrow = true;
                    break;
                }
            } // for i
        } // else
        
        if (toThrow)
            firePropertyChange("OwnerRights", l_oldRights, 
                getL_ownerRights());

    } catch (BaseError b) {
        throw new InternalError(b.toString());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setL_groupRights(), firePropertChange() throws:");
        e.printStackTrace();
    }              
                
    l_oldRights = null;
    l_newRights = null;
    l_limitedRights = null;
    l_restrictedRights = null; 
    l_posRights = null;
    l_negRights = null;
    right = null;
  } // setL_ownerRights()

  /**
   * Gets the rights of the resource owner. Caution: do not take these rights in account
   * if the parent tree does not allow any access, while the AG rights inheritance is managed.
   * @return an array of positive access rights for the UserID which owns this resource,
   *   in the main ACS. No null value. May be empty, or null only if the instance is finalized.
   */
  public final synchronized StringRight[] getL_ownerRights() {
    if (l_ownerRights_ == null) return null;

                // for testing and tracking
                //System.out.println("ResourceImpl.getL_ownerRights() in "+getName()+
                  //      ", rights: ");
                //for (int i = 0; i < l_ownerRights_.length; i++) {
                  //  System.out.println("------------ "+l_ownerRights_[i]);
                //} // for i
    
    StringRight[] result = new StringRight[l_ownerRights_.length];
    System.arraycopy(l_ownerRights_, 0, result, 0, l_ownerRights_.length);
    return result;
  }

  /**
   * Defines the main group owning this resource as an access target. 
   * Called by setEorUserID() to set as new group, the primary group of the new account.
   * Sets the group rights to null.
   * Noop and error mesage to the user if the resource type contains '&lt;byAcsCreator&gt;'
   * or '&lt;immutable&gt;'.
   * Reads the ACS 'GroupID.NoMainGroup' values to check if the argument may be a main group.
   * It is a bound property. 
   * @param _g GroupID which owns this resource, in the main ACS. May be null.
   * @exception UpDateError if _g does not belong to the main ACS
   *  of this resource, or if it is immutable for its groupID.
   */
  public synchronized void setEorGroupID(ImmutableGroupID _g) throws UpDateError	{
      if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+" sets it as immutable for its groupID.");
      if (_g != null) {
        if (aCS_ == null) throw new UpDateError(
          "In ResourceImpl, the main ACS is null");
        if (aCS_ != _g.getEorACS() ) {
          aCS_.getEorBaseError().addException("User error: the selected GroupID "+
            "has to belong to the ACS of this resource.");

            // for testing and tracking
            //System.out.println("Exception fromResourceImpl.setEorGroupID(): "+
            //  aCS_.getEorBaseError().toString());

          // error message to the user
          throw new UpDateError(
            "The selected group does not belong to the ACS of this object.");
        }
      } // if

      String[] l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
              get("GroupID.NoMainGroup");
      boolean remove = false;
      if (l_remove != null) {
          for (int i = 0; i < l_remove.length; i++) {
              if (_g != null && l_remove[i].equals(
                      UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(_g.getEPType()))) {
                  remove = true;
                  break;
              }
          } // for i
      } // if (l_remove !=
      l_remove = null;
      if (remove)
          throw new UpDateError("This group type is in the ACS values for 'GroupID.NoMainGroup': "+
                  "it cannot be the main group of an access target.");      
      
      GroupID oldGroupID = groupID_;
      if (groupID_ != null) ((GroupIDImpl)groupID_).removeOwnResource(this);
      groupID_ = (GroupID) _g;
      if (groupID_ != null) ((GroupIDImpl)groupID_).addOwnResource(this);

      try {
          firePropertyChange("GroupID", oldGroupID, groupID_);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setEorGroupID(), firePropertChange() throws:");
        e.printStackTrace();
      }

      setL_groupRights(null);

      oldGroupID = null;
      return;
  } // setEorGroupID()

  /**
   * @return ImmutableGroupID primary group which owns this resource,
   * in the main ACS. May be null.
   */
  public ImmutableGroupID getEorGroupID()	{

                  // for testing and tracking
          //if (name_ != null)
          //  System.out.println("for the ResourceImpl "+name_+
          //    "the groupID is: "+groupID_);

  return groupID_;
  }

  /**
   * Sets the rights for the members of the groupID which owns this resource.
   * The rights may be acsrights and metarights, positive and negative rights.
   * The ACS constraints on 'GroupRights' are applied only if they are not null,
   * otherwise all the arguments are validated.
   * The negative rights are subtracted from the positive rights.
   * No action if the instance has been finalized. If a new right
   * is not allowed by the ACS, it is removed and the other rights are kept.
   * No adding of null or negative elements.
   * Noop if the resource type contains '&lt;byAcsCreator&gt;' or '&lt;immutable&gt;'.
   * Calls BaseUtilityImpl.subtractNegativeFromPositiveInOneACS() and
   * UtilityImpl.selectHeaders().
   * Called by the constructor and setEorGroupID().
   * Overridden, and also called, by DirectoryImpl to update l_parentTreeAGRightsAllowToGoAcross_.
   * Overridden by the AcsAddon Resource subclasses.
   * Fires a PropertyChange for 'GroupRights'.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the group rights are immutable,
   * or if there are more than 20 increasing chains to analyze from one lesser right.
   */
  public synchronized void setL_groupRights(StringRight[] _s) throws UpDateError {
    if (l_groupRights_ == null) return;
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
            type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its group rights.");; 

                          // for testing and tracking purpose
                        //System.out.println("call to ResourceImpl.setL_groupRights() for "+name_);
    
    // variables
    StringRight[]       l_oldRights = new StringRight[l_groupRights_.length];
    Set                 l_newRights = new HashSet(INITIAL_CAPACITY / 2);
    StringRight[]       l_restrictedRights = null;
    List<StringRight>   l_limitedRights = null;
    Set<StringRight>    l_posRights = new HashSet(INITIAL_CAPACITY / 2);
    Set<StringRight>    l_negRights = new HashSet(INITIAL_CAPACITY / 2);
    StringRight         right = null;
    boolean             toThrow = false;
    
    System.arraycopy(l_groupRights_, 0, l_oldRights, 0, l_groupRights_.length);
    
    try {       
        // gets the rights restriction in the ACS
        l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                   null, this, "GroupRights", aCS_, null);
    } catch (Exception c) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setL_groupRights(), "+
                "BaseUtilityImpl.getL_AcsRestrictedRights() throws:");
        c.printStackTrace();
    } // catch
    
    if (l_restrictedRights != null) 
        l_limitedRights = Arrays.asList( l_restrictedRights );
    
    if (_s != null) {
      for (int i = 0; i < _s.length; i++) {
        if (_s[i] != null) {
            if (l_limitedRights == null ||
                    (l_limitedRights != null && l_limitedRights.contains(_s[i]))) {
                l_newRights.add(_s[i]);

                          // for testing and tracking purpose
                     //System.out.println("in ResourceImpl.setL_groupRights(), "+
                      //"before umask, in l_newRights addition of: "+_s[i]); 
                     
            } // if
        } // if
      } // for i
    }
    
    // filters the positive and negative rights
    if (!l_newRights.isEmpty()) {
        for (Iterator ite = l_newRights.iterator(); ite.hasNext(); ) {
            right = (StringRight)ite.next();
            if (right.isGrantingRight()) l_posRights.add(right);
            else l_negRights.add(right);
        }
    } // if
    
    // subtracts the negative rights to the positive ones, without removing
    // the common lower rights
    if (!l_negRights.isEmpty()) 
        l_newRights = UtilityImpl.getFirstInstance().
                    subtractNegativeNotRemovingInOneACS(
           l_posRights, l_negRights, aCS_, this instanceof ImmutableDirectory);
    else l_newRights = l_posRights;

    try { 

                          // for testing and tracking purpose
                     //System.out.println("in ResourceImpl.setL_groupRights(), "+
                      //"l_newRights before selectHeaders()");
                     //for (Iterator ite = l_newRights.iterator(); ite.hasNext(); ) {
                       //  System.out.println("-------- "+((StringRight)ite.next()).getRight());
                     //} // for ite
        
        UtilityImpl.getFirstInstance().selectHeaders(l_newRights);

                          // for testing and tracking purpose
                     //System.out.println("in ResourceImpl.setL_groupRights(), "+
                      //"l_newRights after selectHeaders()");
                     //for (Iterator ite = l_newRights.iterator(); ite.hasNext(); ) {
                       //  System.out.println("-------- "+((StringRight)ite.next()).getRight());
                     //} // for ite
        
        // updates the group rights
        if (!l_newRights.isEmpty()) {
          l_groupRights_ = (StringRight[])
                  l_newRights.toArray(new StringRight[l_newRights.size()]);
        } else {
          l_groupRights_ = new StringRight[0];
        }

        if (l_oldRights.length != l_groupRights_.length )
            toThrow = true;
        else if (l_oldRights.length == l_groupRights_.length &&
                l_groupRights_.length == 0)
            toThrow = true;
        else {
            for (int i = 0; i < l_oldRights.length; i++) {
                if (!l_newRights.contains( l_oldRights[i] ) ) {
                    toThrow = true;
                    break;
                }
            } // for i
        } // else
        
        if (toThrow)
            firePropertyChange("GroupRights", l_oldRights,
                getL_groupRights());
    } catch (BaseError b) {
        throw new InternalError(b.toString());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setL_groupRights(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldRights = null;
    l_newRights = null;
    l_limitedRights = null;
    l_restrictedRights = null;
    l_posRights = null;
    l_negRights = null;
    right = null;
    return;
  } // setL_groupRights()

  /**
   * caution: does not take these rights in account if the parent tree
   * does not allow any access while AG rights inheritance is managed.
   * @return an array of positive access rights for the GroupID which owns this resource,
   *   in the main ACS. No null value. May be empty, or null only if the instance is finalized.
   */
  public final synchronized StringRight[] getL_groupRights() {
    if (l_groupRights_ == null) return null;
    StringRight[] result = new StringRight[l_groupRights_.length];
    System.arraycopy(l_groupRights_, 0, result, 0, l_groupRights_.length);
    return result;
  } // getL_groupRights()

  /**
   * If there is no parent, return true to avoid useless processing.
   * Return also true if the 4 compared objects are null.
   * Note: unused in gWork since even when the parent has the same A and G,
   * this parent may have different rights than this resource rights.
  * @return true if this instance has the same userID and main group than its direct parent.
  */
  public final boolean hasTheAccountAndGroupOfParent() {
      if (parent_ == null) return true;
      if (userID_ == parent_.getEorUserID() && groupID_ == parent_.getEorGroupID()) 
          return true;
      else return false;
  } // hasTheAccountAndGroupOfParent()

  /**
   * Note: the access from the parent userID depends only to the userID rights.
  * @return true if the parent authorizes its userID to have access to this instance.
  */
  public boolean getAGOAccessFromTheParentAccount() {
      if (l_parentTreeAGRightsAllowToGoAcross_ != null)
          return l_parentTreeAGRightsAllowToGoAcross_[0];
      else return false;
  }

  /**
  * @return true if the parent authorizes its groupID to have access to this instance.
  */
  public boolean getAGOAccessFromTheParentGroup() {
      if (l_parentTreeAGRightsAllowToGoAcross_ != null)
          return l_parentTreeAGRightsAllowToGoAcross_[1];
      else return false;
  }
 
  /**
   * Gets the inherited rights from the parent of the resource owner.
   * Calls BaseUtilityImpl.getChildRightsFromParentAGRights().
   * @return an array of access rights for the UserID which owns this resource,
   *   in the main ACS. May be empty but not null.
   */
  public synchronized StringRight[] getL_inheritedOwnerRights() {
    if (parent_ == null) {
                // for testing and tracking
                //if (name_.lastComponent().equals("f_jerry"))
                  //  System.out.println(
                    //        "for f_jerry, inherited owner rights is empty since the parent is null");
        
        return new StringRight[0];
    }
    else {
        StringRight[] l_right = BaseUtilityImpl.getChildRightsFromParentAGRights(
            ((Directory)parent_).getL_ownerRights(), aCS_);

                // for testing and tracking
                //if (name_.lastComponent().equals("f_jerry"))
                  //  System.out.println(
                    //        "for f_jerry, inherited owner rights size is: "+l_res.length);        
        
                // for testing and tracking
                //if (name_.lastComponent().equals("f_jerry")) {
                  //  for (int i = 0; i < l_res.length; i++) {
                    //    System.out.println("for f_jerry, inherited owner rights: "+l_res[i]);
                  //  }
                //} // if
        
        return l_right;
    }
  } // getL_inheritedOwnerRights()
  
  /**
   * Gets the inherited rights from the parent of the resource group. 
   * Calls BaseUtilityImpl.getChildRightsSetFromParentAGRights().
   * @return an array of access rights for the GroupID which contains this resource,
   *   in the main ACS. May be empty but not null.
   */
  public synchronized StringRight[] getL_inheritedGroupRights() {
    if (parent_ == null) return new StringRight[0];
        
                    // for testing and tracking
                    //System.out.println("ResourceImpl.getL_inheritedGroupRights(): "+
                     //       "call to UtilityImpl.getChildRightsSetFromParentAGRights() "+
                      //      "from the parent "+parent_.getName());


    else {
        StringRight[] l_right = BaseUtilityImpl.getChildRightsFromParentAGRights(
            ((Directory)parent_).getL_groupRights(), aCS_);

                // for testing and tracking
                //if (name_.lastComponent().equals("f_jerry"))
                  //  System.out.println(
                    //        "for f_jerry, inherited group rights size is: "+l_res.length);

                // for testing and tracking
                //if (name_.lastComponent().equals("f_jerry")) {
                  //  for (int i = 0; i < l_res.length; i++) {
                    //    System.out.println("for f_jerry, inherited group rights: "+l_res[i]);
                  //  }
                //} // if

        return l_right;
    }
  } // getL_inheritedGroupRights()
  
  /**
   * Sets the rights for the userID 'root'.
   * 'full_control' equivallent is a mandatory right, even if it is not
   * in the constrained rights list.
   * No action if the instance has been finalized or if a right is not allowed.
   * Noop if the resource type contains '&lt;byAcsCreator&gt;' or '&lt;immutable&gt;'.
   * Fires a PropertyChange for 'RootRights'.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the root rights are immutable.
   */
  public synchronized void setL_rootRights(StringRight[] _s) throws UpDateError {
    if (l_rootRights_ == null) return;
    if (type_.startsWith("<byAcsCreator>") ||type_.startsWith("<immutable> ") ||
            type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+" sets it as immutable for its root rights.");         
            
    StringRight[]       l_oldRights = new StringRight[l_rootRights_.length];
    System.arraycopy(l_rootRights_, 0, l_oldRights, 0, l_rootRights_.length);

    Set                 l_newRights = new HashSet(INITIAL_CAPACITY / 2);
    StringRight[]       l_restrictedRights = null; 
    List<StringRight>   l_limitedRights = new ArrayList(INITIAL_CAPACITY);
    boolean             toThrow = false;
    
    try {       
        // gets the rights restriction in the ACS
        l_restrictedRights = BaseUtilityImpl.getL_AcsRestrictedRights(
                   null, this, "RootRights", aCS_, null);
    } catch (Exception c) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setL_groupRights(), "+
                "BaseUtilityImpl.getL_AcsRestrictedRights() throws:");
        c.printStackTrace();
    } // catch
    
    if (l_restrictedRights != null) // null there
        l_limitedRights = Arrays.asList( l_restrictedRights ); 
    
    if (_s != null) {
      for (int i = 0; i < _s.length; i++) {
        if (_s[i] != null) {
            if (l_limitedRights.isEmpty())
                l_newRights.add(_s[i]);
            else if (l_limitedRights.contains(_s[i]))
                l_newRights.add(_s[i]);
        } // if
      } // for i
    }
    l_newRights.add(aCS_.getFullControlRight()); // for root only, never null  
    
    if (!l_newRights.isEmpty()) {
      l_rootRights_ = (StringRight[])
              l_newRights.toArray(new StringRight[l_newRights.size()]);
    } else { // not possible
      l_rootRights_ = new StringRight[0];
    }
         
    if (l_oldRights.length != l_rootRights_.length )
        toThrow = true;
    else if (l_oldRights.length == l_rootRights_.length &&
            l_rootRights_.length == 0)
        toThrow = true;
    else {
            for (int i = 0; i < l_oldRights.length; i++) {
                if (!l_newRights.contains( l_oldRights[i] ) ) {
                    toThrow = true;
                    break;
                }
            } // for i
    } // else
    
    try {
        if (toThrow)
            firePropertyChange("RootRights", l_oldRights,
                getL_rootRights());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setL_groupRights(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    l_oldRights = null;
    l_newRights = null;
    l_limitedRights = null;
    l_restrictedRights = null;  
    return;
  } // setL_rootRights()

  /**
   * @return an array of access rights for the root UserID,
   *  in the main ACS, following the Unix/Linux scheme. No null value.
   *  May be empty, or null only if the instance is finalized.
   */
  public final synchronized StringRight[] getL_rootRights()	{
    if (l_rootRights_ == null) return null;
    StringRight[] result = new StringRight[l_rootRights_.length];
    System.arraycopy(l_rootRights_, 0, result, 0, l_rootRights_.length);
    return result;
  }

  /**
   * Sets the type which drives the immutablility and the allowed rights of the object.
   * Called by the constructor and ActorImpl.setEPType() since the EPType
   * is then the ResourceType.
   * This is just a wrapper of super.setType() and it fires the event 'ResourceType'.
   * @param _st type of the resource.
   * @exception UpDateError if the type is null, starts with <byAcsCreator>,
   *    or not known by the ACS, or not associated to the previous type.
   */
  public synchronized void setResourceType(String _st) throws UpDateError {
      setType(_st);

                          // for testing and tracking purpose
                     //System.out.println("ResourceImpl.ResourceType of "+getName());
                     //for (Iterator ite = Arrays.asList( (String[])aCS_.getM_SpecializedTypesAndExplorer().
                       //     get("Resource.ResourceType") ).iterator(); ite.hasNext(); ) {
                       //  System.out.println("------------------ "+ite.next());
                     //} // for ite
                     

  } // setResourceType()

  /**
   * The resource types are generally defined at the ACS level.
   * @return String of resource type, as for exemple "executable" or "file".
   *  May be null.
   */
  public String getResourceType()	{
    return type_;
  }

  /**
   * Creates a new AclEntry which delivers rights to an eligible party
   * to access or to deny access to this resource.
   * The resource name must have a name following the pattern
   * IS_name / ACS_name / resource_name, to allow the identification
   * of the ACS in the EligibleParty _ep.
   * The new AclEntry is added at the end of the current list.
   * This method rejects a new non-conditional ACL having the set
   * (ACS, source, target, sens) of a known non-conditional ACL,
   * while a conditional ACL may have the same set.
   * This method rejects a new right-condition ACL having the set
   * (ACS, source, target, sens) of a known conditional ACL,
   * while a non-conditional ACL may have the same set.
   * This method rejects a new flexible conditional ACL having the first condition group
   * of a known conditional ACL and the same set (ACS, source, target, sens),
   * but it may be the second condition group of this conditional ACL. <P>
   * The control of the ACS policy is done in the AclEntry constructor and the ACS.
   * For the creation of the ACL by the user, if there is no rights to select,
   * BaseObjectHandler.setRestrictedSelectedArguments() sets an array
   * with a null object as first value, and this method produces an exception.
   * On the other hand, an association may use a null value for the rights array,
   * and this is allowed to create an ACL without rights.
   * There is an error mesage to the user if the resource or the EP is immutable.
   * The creation of the inherited AclEntry in a child is not handled by
   * the generic gBase package, while the rights updating of the inherited AclEntry
   * is handled by the generic classes AclEntryImpl, ResourceImpl and DirectoryImpl.
   * The sequence of calls from this method is the following: <P>
   * - new AclEntryImpl, <P>
   * - ACS.newAclEntry(), <P>
   * - EligibleParty.addEorAclEntryToEP(), <P>
   * - updating of l_aclEntry_, <P>
   * - addAcsController(), <P>
   * - firePropertyChange(). <P>
   * 'AclRights' and 'EPsThroughAclEntry' are bound properties.
   * The 'AclRights' event sends some copies of the old and the new array of AclEntry.
   * 'AclRights' must be also the bound property name of the rights in AclEntryImpl,
   * to allow the work of GUI1 classes, and the sended values are then a StringRight
   * array. 'AclRights' is listen by every gWork.RightsMediatorImpl where the view
   * depends on the AclEntries of this resource.
   * @param _aclAcs ImmutableACS which owns the new AclEntry, and to be recorded
   *  as an ACS controller of this resource if it is not its main ACS.
   * @param _ep EligibleParty connected to the resource through a new aclEntry.
   * It has to not be immutable, but may be created by the program.
   * @param _l_right is an array of positive or negative rights of _ep on this resource
   *  (that is, if READ is a right, _ep can read on the resource). May be null.
   * @param _positiveRight is true if the rights are set to authorize (grant),
   *  and false if they are set to forbid (deny).
   * @param _conditionGroup first condition group. May be null.
   * @exception InternalError if this resource has a null name or a null ACS,
   *  or if _ep is not an EligiblePartyImpl nor an ActorImpl.
   * @exception CreateError if _ep or _aclAcs is null, or if _ep is already recorded
   * in this resource with the same ACS, or thrown by the AclEntry constructor or
   * ACS.newEorAclEntry().
   * @return the new AclEntry.
   */
  public synchronized AclEntry addAclEntry(ImmutableACS _aclAcs,
    ImmutableEligibleParty _ep, StringRight[] _l_right, Boolean _positiveRight,
    ImmutableGroupID _conditionGroup) throws CreateError {
    
                                // for tracking and testing
                                // if (_aclAcs.getName().equals(new NameImpl("acs1") ) )
                                //System.out.println("ResourceImpl.addAclEntry() start");

    boolean     isExternal = false;
    boolean     match = false;
    String[]    l_types = null;
      
    // exceptions
    if (type_.startsWith("<immutable>") || type_.endsWith("<immutable> "))
          throw new CreateError(
            " The type of "+getName().lastComponent()+" sets it as immutable for its AclEntries.");
    if (_ep.getEPType().startsWith("<immutable>") || _ep.getEPType().endsWith("<immutable> "))
          throw new CreateError(
            " The type of "+_ep.getName().lastComponent()+" sets it as immutable for its ACLs.");
    
    if (aCS_ == null) throw new CreateError(" In ResourceImpl "+
      name_+", the AclEntry can't be created with a null main ACS.");
    if (_ep == this) throw new CreateError(" In ResourceImpl "+
      name_+", the AclEntry can't be created with"+
      " this resource as ImmutableEligibleParty argument.");
    if (_ep == null) {
      aCS_.getEorBaseError().addException("the AclEntry can't be created; _ep is null ");
      throw new CreateError(" The AclEntry can't be created - the EligibleParty is null.");
    }
    if (_aclAcs == null) {
      aCS_.getEorBaseError().addException("the AclEntry can't be created; _aclAcs is null ");
      throw new CreateError(" The AclEntry can't be created - the ACS is null.");
    }
    
    // tests getManageExternalAcls(), where the 3 ACS have to manage them.
    // the constructor of AclEntryImpl manages the constraint on the condition group ACS.
    if (aCS_ != _aclAcs || _ep.getEorACS() != _aclAcs) isExternal = true;
    if (isExternal) {
        if (!_aclAcs.getManageExternalAcls() ) {
          aCS_.getEorBaseError().addException("the AclEntry is external and its ACS does not manage it.");
          throw new CreateError(" The AclEntry is external, and the AclEntry ACS does not manage it.");
        }
        if (!aCS_.getManageExternalAcls() ) {
          aCS_.getEorBaseError().addException("the AclEntry is external and this resource ACS does not manage it.");
          throw new CreateError(" The AclEntry is external and the resource ACS does not manage it.");
        }
        if (!_ep.getEorACS().getManageExternalAcls() ) {
          aCS_.getEorBaseError().addException("the AclEntry is external and the EP ACS does not manage it.");
          throw new CreateError(" The AclEntry is external and the EP ACS does not manage it.");
        }
    } // if (isExternal

    // test getManageInternalAcls()
    else if (!_aclAcs.getManageInternalAcls() ) {
      aCS_.getEorBaseError().addException("the AclEntry is internal, and its ACS does not manage it.");
      throw new CreateError(" The AclEntry is internal, and the AclEntry ACS does not manage it.");
    }

    // if no rights to select, BaseObjectHandler.setRestrictedSelectedArguments()
    // sets an array with a null object as first value.
    // on the other hand, an association may use a null value for the rights array,
    // and this is allowed to create an ACL without rights.
    if (_l_right != null && _l_right.length != 0 && _l_right[0] == null) {
        aCS_.getEorBaseError().addException("An AclEntry cannot be created for this couple"+
                " (source, target).");
        throw new CreateError(" An AclEntry cannot be created for this couple"+
                " (source, target).");
    } // if

    for (int i = 0 ; i < l_aclEntry_.length ; i++) {
      if (l_aclEntry_[i] != null) {
        if (l_aclEntry_[i].getEorEP() == _ep) {
              // this control forbids a new right-condition ACL having
              // the set (ACS, source, target, sens) of a conditional ACL
           if (_l_right != null && _l_right.length != 0 &&
                _l_right[0].getRight().indexOf("<IF>") != -1 &&
                l_aclEntry_[i].isConditionalAclEntry()) {
               // it is a right-condition ACL
              if (l_aclEntry_[i].getEorACS() == _aclAcs &&
                l_aclEntry_[i].getPositiveRight() == _positiveRight) {
                  aCS_.getEorBaseError().addException("the AclEntry can't be created -"+
                    " it is already registered. ");
                  throw new CreateError(" This AclEntry cannot be created -"+
                    " it is already registered as "+l_aclEntry_[i]);
              }

           }
           else { // not a right-condition ACL
               
              if (l_aclEntry_[i].getEorACS() == _aclAcs &&
                l_aclEntry_[i].getPositiveRight() == _positiveRight &&
                l_aclEntry_[i].getL_ConditionSourceGroups().isEmpty() &&
                _conditionGroup == null ) {
                  aCS_.getEorBaseError().addException("the AclEntry can't be created:"+
                    " it is already registered. ");
                  throw new CreateError(" This AclEntry cannot be created -"+
                    " it is already registered as "+l_aclEntry_[i]);
              }
              // this control forbids a new flexible conditional ACL having
              // the first condition group of a known condition ACL,
              // but it may be the second condition group.
              else if (l_aclEntry_[i].getEorACS() == _aclAcs &&
                l_aclEntry_[i].getPositiveRight() == _positiveRight &&
                !l_aclEntry_[i].getL_ConditionSourceGroups().isEmpty() &&
                l_aclEntry_[i].getL_ConditionSourceGroups().iterator().next() == _conditionGroup ) {
                  aCS_.getEorBaseError().addException("the AclEntry can't be created -"+
                    " it is already registered. ");
                  throw new CreateError(" This AclEntry cannot be created -"+
                    " it is already registered as "+l_aclEntry_[i]);
              }
               
           } // else

        } // if (l_aclEntry_[i].getEorEP() == _ep)
      }
    } // for i
    
    // no control on the acl rights restriction in _aclAcs;
    // done at the acl creation: Resource.NoNonConditionalACL

    // initialization
    AclEntry        nullAcl = new AclEntryImpl();
    AclEntry[]      newArray;
    AclEntry[]      l_oldAclEntry;
    Set             l_oldEPthroughAclEntry = getEorL_EPthroughAclEntry();
    boolean         knowsEP = false;

    // saves the old AclEntry list, since AclEntries is a bound property
    l_oldAclEntry = new AclEntryImpl[l_aclEntry_.length];
    System.arraycopy(l_aclEntry_, 0, l_oldAclEntry, 0, l_aclEntry_.length);

    // creates the AclEntry
    
                        // for tracking and testing
                        // if (_aclAcs.getName().equals(new NameImpl("acs1") ) )
                        //System.out.println("ResourceImpl.addAclEntry(), creates acl from "+name_+
                          //      " with ep: "+_ep.getName() );

    AclEntry acl = new AclEntryImpl( _aclAcs, this, _ep, _l_right, 
                        _positiveRight, _conditionGroup); // throw CreateError()

                            // for tracking and testing
                          //System.out.println("ResourceImpl.addAclEntry(), after the acl "+
                            //      "creation from "+name_+" with ep: "+_ep.getName() );
    
    // registers ACL in _aclAcs.
    // registers the resource and the ep if they are external ones.
    // display the ACL in the beamer, if it is a listener of 'AclEntry' from the ACS,
         // before its adding in l_aclEntry_
    ((ACSImpl)_aclAcs).newEorAclEntry(acl);

    // new acl in the eligible party, with test
    // if it is an ActorImpl or an EligiblePartyImpl,
    // since addAclEntryToEP() is not in the EligibleParty interface.
    // this is done after _aclACS.newEorAclEntry() to allow
    // AclEntryImpl.getAclEntry() to found the acl from its acs.
    if (_ep instanceof EligiblePartyImpl )
      ((EligiblePartyImpl)_ep).addAclEntryToEP(acl);
    else if (_ep instanceof ActorImpl)
      ((ActorImpl)_ep).addAclEntryToEP(acl);
    else throw new CreateError(" In new AclEntryImpl(), "+
      "ep is not an EligiblePartyImpl nor an ActorImpl.");

                // for testing and tracking
                //System.out.println("In ResourceImpl.addAclEntry() after addAclEntryToEP(). ");
                        //"created in "+name_+
                  //" with the acs: "+_aclAcs.getName()+" and with ep:"+_ep.getName() );

    // fires a change on "EPsThroughAclEntry" if it was the first link to the acl EP
    for (int  i = 0 ; i < l_aclEntry_.length ; i++) {
      if (l_aclEntry_[i] != null &&  l_aclEntry_[i].getEorEP() == _ep) {
          knowsEP = true;
          break;
      } // if (l_aclEntry_[i] != null
    } // for

    // add in an empty place if possible (it never occurs, indeed)
    boolean recorded = false;
    for (int i = 0 ; i < l_aclEntry_.length ; i++) {
      if (l_aclEntry_[i] == null || l_aclEntry_[i].equals(nullAcl) ) {
        l_aclEntry_[i] = acl;
        recorded = true;
        break;
      }
    }
    // else, add at the end of l_aclEntry_
    if (recorded == false) {
      newArray = new AclEntryImpl[l_aclEntry_.length + 1];
      System.arraycopy(l_aclEntry_, 0, newArray, 0, l_aclEntry_.length);
      newArray[l_aclEntry_.length] = acl;
      l_aclEntry_ = new AclEntryImpl[newArray.length];
      l_aclEntry_ = newArray;
    }

    // caution: this is the ACS controllers list of this resource,
    // not the list of the resource ACS
    if (!(_aclAcs.getName()).equals(aCS_.getName()) )
        addAcsController((Name)_aclAcs.getName() ); //throw CreateError()

                // for testing and tracking
                //System.out.println("In ResourceImpl.addAclEntry() after addAcsController(). ");

    try {
        firePropertyChange("AclRights", l_oldAclEntry, getL_aclEntry());
        if (!knowsEP)
          firePropertyChange("EPsThroughAclEntry", l_oldEPthroughAclEntry,
                  getEorL_EPthroughAclEntry());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.addAclEntry(), firePropertChange() throws:");
        e.printStackTrace();
    }

    l_types = null;
    l_oldAclEntry = null;
    newArray = null;
    l_oldEPthroughAclEntry = null;
    return acl;
  } // addAclEntry()

    /**
    * Deletes an aclEntry which delivers rights to an eligible party
    * to access or to deny access to this resource.
    * 'EPsThroughAclEntry' and 'AclRights' are bound properties,
    * and 'AclRights' is the last firing.
    * This method unregisters the relevant AclEntry in the EligibleParty, in this Resource,
    * and in the relevant ACS. The ACS of the AclEntry may be unregistered in this
    * resource if there are no other AclEntry owned by this ACS in this resource.
    * Noop and error mesage to the user if the resource or the EP is immutable.
    * Calls removeACL().
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, if _aclAcs is closed or unknown,
    *  or if the type does not allow this operation.
    */
    public synchronized void removeAclEntry(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError	{

              // for testing and tracking
              //System.out.println("Call of ResourceImpl.removeAclEntry() in: "+name_);
        
    if (type_.startsWith("<immutable> ") || type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+" sets it as immutable for its ACLs."); 
    if (_ep.getEPType().startsWith("<immutable>") || _ep.getEPType().endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+_ep.getName().lastComponent()+" sets it as immutable for its ACLs.");
    
    removeACL(_aclAcs, _ep, _positiveRight, _conditionGroup);
  } // removeAclEntry()


    /**
    * Gets the AclEntries of this resource.
    * The AclEntries are not ordering from the AcsControllers names.
    * @return a new array with direct references to the AclEntries.
    * May be empty. If the instance is finalized, return null.
    */
    public final synchronized AclEntry[] getL_aclEntry()	{
        if (l_aclEntry_ == null) return null;
        AclEntry[] result = new AclEntryImpl[l_aclEntry_.length];
        System.arraycopy(l_aclEntry_, 0, result, 0, l_aclEntry_.length);
        return result;
    }

  /**
  * Tests if an EligibleParty is associated to the resource through an
  * AclEntry.
  * @param _ep EligibleParty to test. May be null.
  * @return true if _ep is referenced in at least
  * one aclEntry in this resource.
  */
  public synchronized boolean containsAclEntryToEP(EligibleParty _ep) {

    if (_ep == null || l_aclEntry_ != null) return false;
    for (int i = 0; i < l_aclEntry_.length; i++) {
      if (l_aclEntry_[i] != null && l_aclEntry_[i].getEorEP() == _ep)
        return true;
    } // for

    return false;
  } // containsAclEntryToEP()

  /**
  * Gets all the EligibleParties associated to the Resource through at least
  * one AclEntry.
  * @return Set of EligibleParties, through external object references.
  * May be empty, but not null.
  */
  public synchronized Set<ImmutableEligibleParty> getEorL_EPthroughAclEntry() {
    Set l_result = new HashSet(INITIAL_CAPACITY/2);
    if (l_aclEntry_ == null) return l_result;
    synchronized (l_aclEntry_) {
      for (int i = 0; i < l_aclEntry_.length; i++) {
        if (l_aclEntry_[i] != null && !l_aclEntry_[i].isEmpty()) {

            // for testing and tracking
            //System.out.println("in ResourceImpl.getEorL_EPthroughAclEntry(), for "+name_+
            //  ", l_aclEntry_["+i+"].getEorEP(): "+l_aclEntry_[i].getEorEP().getName());

          l_result.add(l_aclEntry_[i].getEorEP() ) ;
        } // if
      } // for
    } // synchro
    return l_result;
  } // getEorL_EPthroughAclEntry()

   /**
   * Gets all the privilegeForLinks linked to this target, not the
   * privilegeForTypes. Return only the direct privileges,
   * even if the privilege cascading is activated.
   * @return List of PrivilegeForLinks. May be empty, but never null.
   */
  public List<PrivilegeForLinks> getEorL_TargetPrivilegeForLinks() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_TargetPrivilegeForLinks();
  }

   /**
   * Gets all the PrivilegeForTypes of this target.
   * Does not include the TYPED_FOR_SEED and TYPED_CLASSED_SPECIFIC privileges.
   * @return List of PrivilegeForTypes. May be empty, but never null.
   */
  public List<PrivilegeForType> getEorL_TargetPrivilegeForTypes() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_TargetPrivilegeForTypes();
  }

   /**
   * Gets all the direct privileges linked to this target, not the inherited ones.
   * There is two types of privilege links (allowed or not by the ACS): <P>
   * - PrivilegeForLinks: linked privilege established for a couple
   * (ACSObject, privilege), <P>
   * - PrivilegeForType: typed privilege, where the ACSObject and the privilege
   * has a common type, but without the TYPED_FOR_SEED and
   * TYPED_CLASSED_SPECIFIC privileges. <P>
   * @return List of Privileges. May be empty, but never null.
   */
  public List<Privilege> getEorL_TargetPrivileges() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_TargetPrivileges();
  }

   /**
   * Sets a privilege as applicable to this access target.
   * This method is called only by PrivilegeForTypeImpl 
   * at any creation/deletion of a typed privilege or an access target.
   * For the PrivilegeForTypes, only the TYPED and TYPED_CLASSED privilege may be argument.
   * This is false for the TYPED_CLASSED_SPECIFIC and TYPED_FOR_SEED privileges,
   * and this produces the throwning of an InternalError.
   * Called by BasicImpl.setType(), to the indirect request of the instance constructor.
   * Fires 'OneTargetTypedPrivilege' events.
   * @param _priv to register in the target
   * @exception BaseError if this instance is not a privilege target,
   * if the privilege second type is not authorized for this target.
   */
  public void addEorPrivilegeForTypeToTarget(PrivilegeForType _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    
      //if (type_ != null && (type_.startsWith("<immutable> ") ||  type_.endsWith("<immutable> ")))
        //  throw new BaseError(
          //  "The type of "+getNickName()+" sets it as immutable for its privileges.");

    privHandler_.addEorPrivilegeToTarget(_priv);      
  }
  
   /**
   * Unregisters the privilege when it is no more used by this privilege target.
   * Called by PrivilegeForLinksImpl and PrivilegeForTypeImpl, NOT by the GUI.
   * PrivilegeForLinksimpl.removeEorPrivilegeTargetLink() checks the type of the target
   * to forbid a change on an immutable target.
   * Calls removeAcsController() for an external one-to-one linked privilege.
   * Fires 'OneTargetTypedPrivilege' or 'OneTargetLinkedPrivilege' event,
   * and, for a typed privilege, calls it fireOneTypedPrivilegeTargetEvent() method.
   * @param _priv to remove
   * @exception BaseError if the privilege or this source is unknown, or from removeAcsController().
   */
  public synchronized void removeEorPrivilegeToTarget(Privilege _priv) throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    
    privHandler_.removeEorPrivilegeToTarget(_priv);

    if (_priv.getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE)) {
      boolean maintainAcs = false;
      // removes the LPRI ACS from the controller list if it was its last LPRI.
      // a Resource cannot handles both external ACL and external LPRI.
      if (!_priv.getAcsName().equals(aCS_.getName()) ) {
        for (PrivilegeForLinks pri : privHandler_.getEorL_TargetPrivilegeForLinks()) {

            if ( _priv.getAcsName().equals(pri.getAcsName()) ) {
              maintainAcs = true;
              break;
            } // if

        } // for
        try {
            if (!maintainAcs )
                removeAcsController(_priv.getAcsName() ); // fires UpDateError
        }
        catch (UpDateError u) {
            throw new BaseError("ResourceImpl.removeEorPrivilegeToTarget() from "+
                    "removeAcsController() - "+u.toString());
        } // catch
      } // if (!_priv.getAcsName().equals(aCS_.getName()) )
    } // if (_priv.getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE

  } // removeEorPrivilegeToTarget()

   /**
   * Gets all the immutable list of the inherited linked privileges of the target.
   * The LINKED_IF_ALL_SOURCE type is forbidden, and any privilege having more than one target.
   * The generic gWork package denies any access if the returned value is null
   * and if the Privilege inheritance is managed.
   * Null is the default and immutable returned value in this generic gBase class,
   * and this has to be changed in an AcsAddon implementing the inheritance.
   * The privilege is not a clone, and the target is never this instance.
   * In the BeanInfo class of the Resource subclass that extends this class in an
   * AcsAddon package, this property has to be put in the pane '  PRI Inheritance  '.
   * Called by gWork.LinkRights.withAccessThroughNodesTreeFromEP() and
   * getL_accessRightsThroughNodesTree().
   * @return immutable list of Linked Privileges. May be empty but never null.
   */
  public List<PrivilegeForLinks> getEorImmutableL_TargetInheritedLinkedPrivileges() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorImmutableL_TargetInheritedLinkedPrivileges();
  }
   
  /**
   * Gets the internal map of the inherited linked privilege rights, where
   * the default key is the name of the EligibleParty having the rights, extended with,
   * as last component, the parent order of the direct or indirect parent 
   * which has given the rights.
   * For instance, the direct parent has the string '1', and its direct parent
   * has the string '2' as parent order in the last name component, and so on.
   * This property is not displayed to the user. A subclass may change the key form.
   * The map value is always a non-null StringRight array of inherited rights,
   * and it may be empty.
   * Called by gWork.LinkRights.withAccessThroughNodesTreeFromEP() and
   * getL_accessRightsThroughNodesTree(), but only in the AcsAddons implementing
   * the linked privilege inheritance. Such an AcsAddon may also choose to not use this property.
   * @return unmodifiable map of inherited rights from the linked privileges.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedLinkedPRIRights() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getM_TargetInheritedLinkedPRIRights();      
  }

   /**
   * Gets the immutable list of all the inherited AclEntries of the target.
   * To use by the consumers in the GUI and gWork packages.
   * The generic gWork package denies any access if the returned value is null
   * and if the AclEntry inheritance is managed.
   * Null is the default and immutable returned value in this generic gBase class,
   * and this has to be changed in an AcsAddon implementing the inheritance.
   * The AclEntry is not a clone, and the target is never this instance.
   * In the BeanInfo class of the Resource subclass that extends this class in an
   * AcsAddon package, this property has to be put in the pane '  ACL Inheritance  '.
   * Called by gWork.NodeRights.withAccessThroughNodesTreeFromEP() and
   * getL_accessRightsThroughNodesTree().
   * @see #getEorL_TargetInheritedAclEntries()
   * @return immutable list of AclEntries. May be empty or null.
   */
  public List<AclEntry> getEorImmutableL_TargetInheritedAclEntries() {
      if (l_inheritedT_AclEntries_ == null) return null;
      else if (l_inheritedT_AclEntries_.isEmpty()) return Collections.EMPTY_LIST;
      else {
                            // for testing and tracking
                            //if (name_.lastComponent().equals("function_one")) {
                              //  System.out.println(
                                //        "in ResourceImpl.getEorImmutableL_TargetInheritedAclEntries(), for "+
                                  //      name_+", l_inheritedT_AclEntries_:");

                                //for (Object obj : l_inheritedT_AclEntries_) {
                                  //  System.out.println("------ "+obj);
                                //} // for
                            //} // if

          return Collections.unmodifiableList(l_inheritedT_AclEntries_);
      }
  } // getEorImmutableL_TargetInheritedAclEntries()
   
  /**
   * Gets the internal map of the inherited AclEntry rights, where
   * the default key is the name of the EligibleParty having the rights on the target,
   * extended with, as last component, the parent order of the target parent
   * which has given the rights.
   * For instance, the direct parent has the string '1', and its direct parent
   * has the string '2' as parent order in the last name component, and so on.
   * This property is not displayed to the user. A subclass may change the key form.
   * A subclass may use the constructor TreeMap(Comparator comparator)
   * to set an ordering which is different to the natural ordering of the keys.
   * The map value is always a non-null StringRight array of inherited rights,
   * and it may be empty.
   * Called for the gWork consumers, but only in the AcsAddons implementing 
   * the ACL inheritance. Such an AcsAddon may also choose to not use this property.
   * @return unmodifiable map of inherited rights from the AclEntries.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedACLRights() {
      if (m_inheritedT_ACLRights_ == null) return Collections.EMPTY_MAP;
      else {

                            // for testing and tracking
                            //if (name_.lastComponent().equals("product_name")) {
                              //  System.out.println(
                                //        "in ResourceImpl.getM_TargetInheritedACLRights(), for "+
                                  //      name_+", m_inheritedT_ACLRights_ keys and values:");

                                //StringRight[] rights = null;

                                //for (String name : m_inheritedT_ACLRights_.keySet()) {
                                  //  System.out.println("--- key "+name);
                                    //rights = m_inheritedT_ACLRights_.get(name);
                                    //if (rights != null) {
                                      //  for (int i = 0; i < rights.length; i++) {
                                        //    System.out.println("------- right "+rights[i]);
                                        //} // fori
                                    //} // if
                                //} // for
                            //} // if

          return Collections.unmodifiableMap(m_inheritedT_ACLRights_);
      } // else
  } // getM_TargetInheritedACLRights()

  /**
   * Used by ACS.getEorM_OwnActors(), and by ACSTreeBaseListenerImpl
   * for detecting the case of a wrong actor creation.
   * @return true if this resource is an actor.
   */
  public boolean getActorActivated()	{
          return actorActivated_;
  }

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'AclRights', 'UserID', 'GroupID',
  * 'RootRights', 'GroupRights', 'OwnerRights',
  * 'EPsThroughAclEntry', 'AclAcsControllers', 'PriAcsController',
  * 'FromTheParentAccount', 'FromTheParentGroup',
  * 'InheritedOwnerRights', 'InheritedGroupRights',
  * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege',
  * and from the subclasses in the AcsAddon,
  * 'InheritedTargetPrivileges', 'InheritedTargetAclEntries'.
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      //if (changeSupport_ == null)
        //  changeSupport_ = new PropertyChangeSupport(this);

      //if (_propertyName.equals("AclRights") ||
        //      _propertyName.equals("UserID") ||
          //    _propertyName.equals("GroupID") ||
            //  _propertyName.equals("RootRights") ||
              //_propertyName.equals("GroupRights") ||
              //_propertyName.equals("OwnerRights") ||
              //_propertyName.equals("EPsThroughAclEntry") ||
              //_propertyName.equals("AclAcsControllers") ||
              //_propertyName.equals("PriAcsControllers") ||
              //_propertyName.equals("FromTheParentAccount") ||
              //_propertyName.equals("FromTheParentGroup") ||
              //_propertyName.equals("InheritedOwnerRights") ||
              //_propertyName.equals("InheritedGroupRights") ||
              //_propertyName.equals("InheritedTargetAclEntries"))
                //    changeSupport_.addPropertyChangeListener(_propertyName, _l);
      if (_propertyName.equals("OneTargetTypedPrivilege") ||
              _propertyName.equals("OneTargetLinkedPrivilege") ||
              _propertyName.equals("InheritedTargetPrivileges") ) {
                    if (privHandler_ == null ) {
                        try {
                          privHandler_ = new CPrivilegeUserImpl(this);
                        } catch (CreateError c) {
                            new InternalError("From "+name_+"- "+c.toString());
                        }
                    } // if

                            // for testing and tracking
                            //if (_propertyName.equals("OneTargetLinkedPrivilege"))
                              //System.out.println("ResourceImpl.addPropertyChangeListener(), "+
                                //      "call to privHandler_ for 'OneTargetLinkedPrivilege'");

                    privHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else super.addPropertyChangeListener(_propertyName, _l);

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("ResourceImpl.addPropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the adding of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //System.out.println("listeners of all properties after the adding of "+_l+": ");
                    //l_list = changeSupport_.getPropertyChangeListeners();
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}

                    //l_list = null;
                //} // if


  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'AclRights', 'UserID', 'GroupID',
  * 'RootRights', 'GroupRights', 'OwnerRights',
  * 'EPsThroughAclEntry', 'AclAcsControllers', 'PriAcsController',
  * 'FromTheParentAccount', 'FromTheParentGroup',
  * 'InheritedOwnerRights', 'InheritedGroupRights',
  * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege',
  * and from the subclasses in the AcsAddon,
  * 'InheritedTargetPrivileges', 'InheritedTargetAclEntries'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      //if (changeSupport_ == null) return;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("BasicImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      //if (_propertyName.equals("AclRights") ||
        //      _propertyName.equals("UserID") ||
          //    _propertyName.equals("GroupID") ||
            //  _propertyName.equals("RootRights") ||
              //_propertyName.equals("GroupRights") ||
              //_propertyName.equals("OwnerRights") ||
              //_propertyName.equals("EPsThroughAclEntry") ||
              //_propertyName.equals("AclAcsControllers") || // and PriAcsController
              //_propertyName.equals("FromTheParentAccount") ||
              //_propertyName.equals("FromTheParentGroup") ||
              //_propertyName.equals("InheritedOwnerRights") ||
              //_propertyName.equals("InheritedGroupRights") ||
              //_propertyName.equals("InheritedTargetAclEntries") )
                //    changeSupport_.removePropertyChangeListener(_propertyName, _l);
      if (_propertyName.equals("OneTargetTypedPrivilege") ||
              _propertyName.equals("OneTargetLinkedPrivilege") ||
              _propertyName.equals("InheritedTargetPrivileges")) {
                    if (privHandler_ != null )
                        privHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else super.removePropertyChangeListener(_propertyName, _l);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (changeSupport_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("ResourceImpl.getPropertyChangeListeners(),
                    // call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      //if (_propertyName.equals("AclRights") ||
        //      _propertyName.equals("UserID") ||
          //    _propertyName.equals("GroupID") ||
            //  _propertyName.equals("RootRights") ||
              //_propertyName.equals("GroupRights") ||
              //_propertyName.equals("OwnerRights") ||
              //_propertyName.equals("EPsThroughAclEntry") ||
              //_propertyName.equals("AclAcsControllers") || // and 'PriAcsController'
              //_propertyName.equals("FromTheParentAccount") ||
              //_propertyName.equals("FromTheParentGroup") ||
              //_propertyName.equals("InheritedOwnerRights") ||
              //_propertyName.equals("InheritedGroupRights") ||
              //_propertyName.equals("InheritedTargetAclEntries")  )
            //return changeSupport_.getPropertyChangeListeners(_propertyName);

      if (_propertyName.equals("OneTargetTypedPrivilege") ||
              _propertyName.equals("OneTargetLinkedPrivilege") ||
              _propertyName.equals("InheritedTargetPrivileges")) {
          
            if (privHandler_ != null )
                return privHandler_.getPropertyChangeListeners(_propertyName);
            else
                return null;
      }
      else return super.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()

 /**
  * Clones without registering of the new object in the ACSs.
   * Caution: don't use any updating method on this clone,
   * because that may destroy the base integrity. A finalized instance returns
   * a clone with getEorACS() which returns null,
   * but getName() and getAcsName() operates.
   * @exception CloneNotSupportedException
   * @return Object; is not registred in the ACSs. The names, the aclEntries,
   * the AG rights, the ACS controller names are cloned.
   * The AclEntries are weakly cloned, so they have no direct link to this resource.
   * The external object references are not cloned
   */
  public synchronized Object clone() {
    //try {
      ResourceImpl clon = (ResourceImpl) super.clone();
  //    clon.setL_Access(l_Access_);
      clon.parent_ = parent_;

      if (l_aclEntry_ != null) {
          clon.l_aclEntry_ = new AclEntryImpl[l_aclEntry_.length];
          for (int i = 0 ; i < l_aclEntry_.length ; i++) {
            if (l_aclEntry_[i] != null) {
              clon.l_aclEntry_[i] = (AclEntry) ((AclEntryImpl)l_aclEntry_[i]).clone();
            }
          }
      } else l_aclEntry_ = new AclEntryImpl[0];

      try {
          clon.setL_rootRights(l_rootRights_);
          clon.setL_ownerRights(l_ownerRights_);
          clon.setL_groupRights(l_groupRights_);
      } catch (UpDateError u) {}
      clon.userID_ = userID_;
      clon.groupID_ = groupID_;

      if (l_OpenAcsControllerNames_ != null) {
          clon.l_OpenAcsControllerNames_ = new NameImpl[l_OpenAcsControllerNames_.length];
          for (int i = 0 ; i < l_OpenAcsControllerNames_.length ; i++) {
            if (l_OpenAcsControllerNames_[i] != null) {
              clon.l_OpenAcsControllerNames_[i] = (Name) l_OpenAcsControllerNames_[i].clone();
            }
          }
      } else l_OpenAcsControllerNames_ = new NameImpl[0];

      if (l_inheritedT_AclEntries_ != null) {
          clon.l_inheritedT_AclEntries_ = (ArrayList)l_inheritedT_AclEntries_.clone();
      }

      if (m_inheritedT_ACLRights_ != null) {
          clon.m_inheritedT_ACLRights_ = new TreeMap();
          for (Map.Entry entry : m_inheritedT_ACLRights_.entrySet()) {
             clon.m_inheritedT_ACLRights_.put(
                     (String)entry.getKey(), (StringRight[])entry.getValue());
          }
      } // if

      return clon;

    //} catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
      //if (aCS_ != null) {
        //aCS_.getEorBaseError().addException(e.toString() + "ResourceImpl.clone() failed");
      //}
      //throw new InternalError(e.toString());
    //}
  } // clone()

  /**
  * Does not compare the comment. Works even if the instance is user-finalized.
  * @param _obj
  * @return true if _obj is an instance of ResourceImpl (or a derived class)
  *   equals to this instance (with the same class)
  */
  public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      Resource obj = (Resource) _obj;

      if (parent_ == null | obj.getEorParent() == null) {
          if (parent_ != null | obj.getEorParent() != null)
            return false;
      } // if
      else {
          if (!parent_.equals( obj.getEorParent() ))
            return false;
      } // else

      /*if (l_Access_.length != obj.getL_Access().length) // no test: length == 0
        return false;
      for (int i = 0; i < l_Access_.length; i++) {
        if (l_Access_[i] == null | obj.getL_Access()[i] == null) {
          if (l_Access_[i] != null | obj.getL_Access()[i] != null)
            return false;
          continue; // i++
        } // if
        else {
          if (!l_Access_[i].equals( obj.getL_Access()[i] ))
            return false;
        } // else
      } // for
      */
      if (actorActivated_ != obj.getActorActivated() )
          return false;


      /*
      if (comment_ == null | obj.getComment() == null) {
          if (comment_ != null | obj.getComment() != null)
            return false;
      } // if
      else {
          if (!comment_.equals( obj.getComment() ))
            return false;
      } // else
      */

        //      if (!icon_.equals( obj.getIcon() )
        //          return false;

      // l_aclEntry_ may be null if the instance is finalized
      if (l_aclEntry_ == null || obj.getL_aclEntry() == null) {
          if (l_aclEntry_ != null || obj.getL_aclEntry() != null)
            return false;
      } // if
      else {
          if (l_aclEntry_.length != obj.getL_aclEntry().length)//no test for lenght == 0
            return false;
          for (int i = 0; i < l_aclEntry_.length; i++) {
            if (l_aclEntry_[i] == null || obj.getL_aclEntry()[i] == null) {
              if (l_aclEntry_[i] != null || obj.getL_aclEntry()[i] != null)
                return false;
              continue; // i++
            } // if
            else {
              if (!l_aclEntry_[i].equals( obj.getL_aclEntry()[i] ))
                return false;
            } // else
          } // for
      } // else

      // l_rootRights_ may be null if the instance is finalized
      if (l_rootRights_ == null || obj.getL_rootRights() == null) {
          if (l_rootRights_ != null || obj.getL_rootRights() != null)
            return false;
      } // if
      else {
          if (l_rootRights_.length != obj.getL_rootRights().length)  // no test for lenght == 0
            return false;
          for (int i = 0; i < l_rootRights_.length; i++) {
            if (l_rootRights_[i] == null || obj.getL_rootRights()[i] == null) {
              if (l_rootRights_[i] != null || obj.getL_rootRights()[i] != null)
                return false;
              continue; // i++
            } // if
            else {
              if (!l_rootRights_[i].equals( obj.getL_rootRights()[i] ))
              return false;
            } // else
          } // for
      } // else

      // l_ownerRights_ may be null if the instance is finalized
      if (l_ownerRights_ == null || obj.getL_ownerRights() == null) {
          if (l_ownerRights_ != null || obj.getL_ownerRights() != null)
            return false;
      } // if
      else {
          if (l_ownerRights_.length != obj.getL_ownerRights().length)  // no test for lenght == 0
            return false;
          for (int i = 0; i < l_ownerRights_.length; i++) {
            if (l_ownerRights_[i] == null || obj.getL_ownerRights()[i] == null) {
              if (l_ownerRights_[i] != null || obj.getL_ownerRights()[i] != null)
                return false;
              continue; // i++
            } // if
            else {
              if (!l_ownerRights_[i].equals( obj.getL_ownerRights()[i] ))
                return false;
            } // else
          } // for
      } // else

      // l_groupRights_ may be null if the instance is finalized
      if (l_groupRights_ == null || obj.getL_groupRights() == null) {
          if (l_groupRights_ != null || obj.getL_groupRights() != null)
            return false;
      } // if
      else {
          if (l_groupRights_.length != obj.getL_groupRights().length)  // no test for lenght == 0
            return false;
          for (int i = 0; i < l_groupRights_.length; i++) {
            if (l_groupRights_[i] == null || obj.getL_groupRights()[i] == null) {
              if (l_groupRights_[i] != null || obj.getL_groupRights()[i] != null)
                return false;
              continue; // i++
            } // if
            else {
              if (!l_groupRights_[i].equals( obj.getL_groupRights()[i] ))
                return false;
            } // else
          } // for
      } // else

      if (userID_ == null | obj.getEorUserID() == null) {
          if (userID_ != null | obj.getEorUserID() != null)
            return false;
      } // if
      else {
      if (!userID_.equals( obj.getEorUserID() ))
          return false;
      } //else

      if (groupID_ == null | obj.getEorGroupID() == null) {
          if (groupID_ != null | obj.getEorGroupID() != null)
            return false;
      } // if
      else {
      if (!groupID_.equals( obj.getEorGroupID() ))
          return false;
      } //else

      // l_OpenAcsControllerNames_ may be null if the instance is finalized
      if (l_OpenAcsControllerNames_ == null || obj.getL_AcsControllers() == null) {
          if (l_OpenAcsControllerNames_ != null || obj.getL_AcsControllers() != null)
            return false;
      } // if
      else {
          if (l_OpenAcsControllerNames_.length != obj.getL_AcsControllers().length) // no test for lenght == 0
            return false;
          for (int i = 0; i < l_OpenAcsControllerNames_.length; i++) {
            if (l_OpenAcsControllerNames_[i] == null || obj.getL_AcsControllers()[i] == null) {
              if (l_OpenAcsControllerNames_[i] != null || obj.getL_AcsControllers()[i] != null)
                return false;
              continue; // i++
            } // if
            else {
              if (!l_OpenAcsControllerNames_[i].equals( obj.getL_AcsControllers()[i] ))
                return false;
            } // else
          } // for
      } // else

      if (l_inheritedT_AclEntries_ != null) {
          if (!getEorImmutableL_TargetInheritedAclEntries().equals(
                  obj.getEorImmutableL_TargetInheritedAclEntries()))
              return false;
      } // if

      // m_inheritedT_ACLRights_ is derived from l_inheritedT_AclEntries_

      return true;
  } // equals

  /**
   * @return the hashcode derived from the userID, the type and folder member hashcodes
   */
  public int hashCode() {
    int result = super.hashCode();

    if (userID_ != null) {
      result = PRIME*result + userID_.hashCode();
    }

    return result;
  } // hashcode()

  /**
   * @return the static serialVersionUID__ of this Resource class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  /**
   * The comparison is based on the name compareTo() method of each
   * resource when they are not equals, except when there is a node.
   * The nodes are always smaller than the leaves.
   * @param o Object to compare must be an ImmutableResource
   * @exception ClassCastException - if the specified object's type
   *  prevents it from being compared to this current Object
   * @return 0 if the resources are equals, a negative integer if this Resource object
   * is lesser than the argument, and a positive integer if this Resource object
   * is greater than the argument.
   */
  public int compareTo(Object o) throws ClassCastException {
    if (!(o instanceof ImmutableResource))
      throw new ClassCastException("in Resource.compareTo(), the argument "+
        "is not an instance of ImmutableResource");
    // if (equals(o) ) return 0;
    if (name_ == null) throw new ClassCastException("in Resource.compareTo(),"+
      " the object name is null");
    if (this instanceof ImmutableNode && !(o instanceof ImmutableNode)) return -1;
    else if (!(this instanceof ImmutableNode) && (o instanceof ImmutableNode)) return +1;
    else return name_.compareTo( ((ImmutableResource)o).getName() );
  } // compareTo()
  
   /**
   * This method removes the resource from the ACS lists (main and controllers),
   * from its parent, its userID and its groupID. The view list is not updated,
   * and views are informed by the ACS through removeEorResource().
   * This method is called by finalize(), but finalize() may be not called.
   * An unused object has to be deleted to inform the other objects, and this method
   * HAS TO BE CALLED before deleting the resource.
   * Called by ACSImpl.finalizeForBase() and setAcsRelationsInIS(),
   * parent Directory.finalizeForBase() and by finalizeForUser().
   * Calls ACSImpl.removeEorResource() only if the object is known by the ACS.
   * Calls parent.setEorL_Children() rather than parent.removeEorChild().
   * This method does not raise PropertyChangeEvents since the BeanInfo for this class
   * specifies that its ACS informs the listeners when this instance is finalized,
   * but each called method like UserIDImpl.removeOwnResource() throws a ChangeEvent
   * to the Gui1/Gui2/gWork classes. So each call products a new searching of the true
   * rights if this resource belongs to an open view.
   * To test if an instance has been finalized, uses (l_aclEntry_ == null).
   * @exception UpDateError if the main ACS do not known the resource (if
   * it is a clone without setName(), typically)
   */
  public synchronized void finalizeForBase() throws UpDateError {

                    // for testing and tracking
                //System.out.println("In ResourceImpl, call of " + name_ + ".finalizeForBase()");

    if (folderMember_ == null) return;
    
    // takes out this resource to its parent, in l_Children_, if the parent
    // is not already unrecorded in the main ACS by its own finalize() method.
    // the InternalError exception is not catched
    if (parent_ != null && parent_.getEorL_Children()  != null && 
            aCS_ != null && aCS_.containsOwnResourceName(parent_.getName() )) {
      short shiftingBack = 0;
      ImmutableResource[] leafArray = (ImmutableResource[])parent_.getEorL_Children();
      ImmutableResource[] newArray = new ResourceImpl[leafArray.length];

      if (leafArray != null) {
        for (int  i=0 ; i < leafArray.length ; i++) {
            if (leafArray[i] == this) shiftingBack++;
            else newArray[i - shiftingBack] = leafArray[i];
        } // for
      }
      else newArray = null;

      try {
                          //System.out.println("use of parent_.setEorL_Children() in "
                          //  + name_ + ".finalizeForBase()");

        ((DirectoryImpl)parent_).removeEorChild(this);
      } catch (InternalError i) { // not relevant
        i.printStackTrace();
      }
      leafArray = null;
      newArray = null;
    } // if parent_ != null

    // removes the aclEntries, Acs controllers in this and in its main ACS
    // (not very fast)
    AclEntry[] aclArray = new AclEntry[1];
    if (l_aclEntry_ != null) {
      aclArray = new AclEntry[l_aclEntry_.length];
      System.arraycopy(l_aclEntry_, 0, aclArray, 0, l_aclEntry_.length);
      for (int  i = 0 ; i < aclArray.length ; i++) {
        if (aclArray[i] != null && !aclArray[i].isEmpty()) {
          if (!aclArray[i].getL_ConditionSourceGroups().isEmpty())
              removeAclEntry(aclArray[i].getEorACS(), aclArray[i].getEorEP(),
                aclArray[i].getPositiveRight(),
                aclArray[i].getL_ConditionSourceGroups().iterator().next() );
          else
              removeAclEntry(aclArray[i].getEorACS(), aclArray[i].getEorEP(),
                aclArray[i].getPositiveRight(),
                null );
        } // if
      } // for
    }

    // removes this to its userID
    if (userID_ != null) ((UserIDImpl)userID_).removeOwnResource(this);

    // removes this to its groupID
    if (groupID_ != null) ((GroupIDImpl)groupID_).removeOwnResource(this);

    // removes (this) to the map of resources in the main ACS.
    // as a clone or with the call of ResourceSubclass.finalize(),
    // it may doesn't exist for the ACS and UpDateError is thrown.
    if (aCS_ != null && aCS_.getEorM_Resources().get(name_) != null)
        ((ACSImpl)aCS_).removeEorResource(this);

    aclArray = null;
    
    l_aclEntry_ = null; // may be tested to detect a finalized object
    parent_ = null;
    l_rootRights_ = null;
    l_groupRights_ = null;
    l_ownerRights_ = null;
    l_OpenAcsControllerNames_ = null;
    l_parentTreeAGRightsAllowToGoAcross_ = new boolean[] {true, true};
    l_inheritedT_AclEntries_ = null;
    m_inheritedT_ACLRights_ = null;
    
    super.finalizeForBase();
  } // finalizeForBase()

  /**
   * Gets the name last component of the instance.
   * Used in the explorer to display the element.
   * Never null.
   * @return String
   */
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append(name_.lastComponent() );
/*    sb.append("RESOURCEIMPL :");
    sb.append("; name_ = ");
    sb.append(name_);

// for actor class tests and others inherited classes, add here : /*
    sb.append("; aCS_ = ");
    if (aCS_ != null) sb.append(aCS_.getName());
    else sb.append("null");
/*    sb.append("; actorActivated_ = ");
    sb.append(actorActivated_);
    sb.append("; isACSBridge_ = ");
    sb.append(isACSBridge_);

    sb.append("; parent_ = ");
    if (parent_ != null && parent_.getName() != null)
      sb.append(parent_.getName() );
    if (bridgeTargetEligibleParty_ != null && bridgeTargetEligibleParty_.getName() != null) {
      sb.append("; bridgeTargetEligibleParty_ = ");
      sb.append(bridgeTargetEligibleParty_.getName() );
    }
    sb.append("l_Access_ = ");
    for (int i = 0; i < l_Access_.length; i++) {
        sb.append(l_Access_[i]);
        sb.append(", ");
    } //for

    sb.append("; comment_ = ");
    sb.append(comment_);
    sb.append("; l_aclEntry_ =");
    for (int i = 0; i < l_aclEntry_.length; i++) {
        sb.append(l_aclEntry_[i]);
        sb.append(", ");
    } //for
    sb.append("; type_ = ");
    sb.append(type_);
    sb.append("; rootRights_ = " + rootRights_.length);
    if (rootRights_.length != 0) {
      for (int i = 0; i < rootRights_.length; i++) {
          sb.append(rootRights_[i]);
          sb.append(", ");
      } //for
    } // if
    else sb.append("null ");
*/ /*   sb.append("; l_otherRights_ = ");
    if (l_otherRights_.length != 0) {
      for (int i = 0; i < l_otherRights_.length; i++) {
        sb.append(l_otherRights_[i]);
        sb.append(", ");
      } //for
    } // if
    else sb.append("null ");
    sb.append("; groupRights_ = ");
    if (l_groupRights_.length != 0) {
      for (int i = 0; i < l_groupRights_.length; i++) {
        sb.append(l_groupRights_[i]);
        sb.append(", ");
      } //for
    } // if
    else sb.append("null ");
    sb.append("; ownerRights_ = ");
    if (l_ownerRights_.length != 0) {
      for (int i = 0; i < l_ownerRights_.length; i++) {
        sb.append(l_ownerRights_[i]);
        sb.append(", ");
      } //for
    } // if
    else sb.append("null ");

    sb.append("; userID_ = ");
    if (userID_ != null) sb.append(userID_);
    else sb.append("null");
    sb.append("; groupID_ = ");
    if (groupID_ != null) sb.append(groupID_);
    else sb.append("null");
    sb.append("; l_OpenAcsControllerNames_ = ");
    if (l_OpenAcsControllerNames_.length != 0) {
      for (int i = 0; i < l_OpenAcsControllerNames_.length; i++) {
        if (l_OpenAcsControllerNames_[i] != null) {
          sb.append(l_OpenAcsControllerNames_[i]);
          sb.append(", ");
        } // if
        else sb.append("null, ");
      } //for
    } // if
    else sb.append("null ");
    sb.append("; iS_Name_ = ");
    sb.append(iS_Name_);

    return " [[[" + sb + "]]]";
//*/
    return " " + sb;
  } // toString()

    /**
    * Resets the inherited AclEntry rights after an user edition.
    * Called by the parent when there is a change in the effective rights of
    * an inherited AclEntry.
    * Reads l_inheritedT_AclEntries_ and updates m_inheritedT_ACLRights_
    * only for one relevant key. May be overridden only in the AcsAddons
    * handling the ACL inheritance.
    * Not described in the BeanInfo since the user does not call it.
    * Note: the creation of the inherited AclEntry in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited AclEntry is handled.
    * Called by Directory.resetInheritedAclEntryRightsOfChildren().
    * Overridden in DirectoryImpl and called from it.
    * @param _acl inherited AclEntry with updated rights. Never null.
    */
  public synchronized void resetInheritedAclEntryRights(AclEntry _acl) {

                                  // for testing and tracking
                                  //System.out.println("ResourceImpl."+
                                    //      "resetInheritedAclEntryRights() "+
                                      //    "on "+name_.lastComponent()+
                                        //  " for the ACL "+_acl);

      if (l_inheritedT_AclEntries_ == null) {
          InternalError i = new InternalError("ResourceImpl."+
              "resetInheritedAclEntryRights(), the target "+name_.lastComponent()+
              " has no AclEntries.");
          i.printStackTrace();
          throw i;
      } // if

      AclEntry          acl = null;
      boolean           found = false;
      StringRight[]     l_rights = null;
      Set<StringRight>  l_newRights = null;
      String            order = UtilityImpl.getFirstInstance().getParentOrder(
                                  this,
                                  (ImmutableNode) UtilityImpl.getFirstInstance().getBaseObject(
                                        _acl.getResName()));
      String            key = ((Name)_acl.getEpName().clone()).extended(order).toString();


                                // for testing and tracking
                                //if (name_.lastComponent().equals("Products table") &&
                                  //  m_inheritedT_ACLRights_.get(keyName) != null) {

                                    //    StringRight[] rights = m_inheritedT_ACLRights_.get(keyName);
                                      //  for (int i = 0; i < rights.length; i++) {
                                        //    System.out.println("--- old right "+rights[i]);
                                        //} // for ite

                                //} // if (name_
                                //else if (name_.lastComponent().equals("Customers table"))
                                  //  System.out.println("ResourceImpl."+
                                    //      "resetInheritedAclEntryRights() "+
                                      //    "for the resource "+name_.lastComponent()+
                                        //  " without old rights for the key "+keyName);


      // clears the map to update
      StringRight[]     l_oldRights = m_inheritedT_ACLRights_.remove(key);

                                  // for testing and tracking
                                  //System.out.println("ResourceImpl."+
                                    //      "resetInheritedAclEntryRights() "+
                                      //    "for the keyName "+keyName);


      // reset the couple (source name + order, StringRight[])
      for (Iterator ite = l_inheritedT_AclEntries_.iterator(); ite.hasNext();) {
        acl = (AclEntry) ite.next();


                          // for testing and tracking
                          //System.out.println("--- iteration on inherited ACL "+acl);

        // review all the AclEntry with the _acl keyName
        if (!acl.getResName().equals(_acl.getResName()) ||
                !acl.getEpName().equals(_acl.getEpName())) {
            continue;
        }

        if (acl == _acl) found = true;

        // acl is _acl in one of the iterations
        l_rights = acl.getL_Rights();

        if (l_rights.length == 0) {

                          // for testing and tracking
                          //System.out.println("--- empty rights for "+acl);

            continue;
        }

                          // for testing and tracking
                          //System.out.println("--- non-empty rights to add for "+acl);

        if (m_inheritedT_ACLRights_.get(key) == null) {

            m_inheritedT_ACLRights_.put(key, l_rights);

        } // if
        else {
            l_newRights = new HashSet(
                    Arrays.asList(m_inheritedT_ACLRights_.get(key)));
            l_newRights.addAll(Arrays.asList(l_rights));

            m_inheritedT_ACLRights_.put(
                    key,
                    l_newRights.toArray(new StringRight[0]));
        } // else

      } // for ite

      if (!found) {
          InternalError i = new InternalError("ResourceImpl."+
              "resetInheritedAclEntryRights(), the target "+name_.lastComponent()+
              " does not know the inherited ACL "+_acl);
          i.printStackTrace();
          throw i;
      } // if (!found)


                                // for testing and tracking
                                //if (name_.lastComponent().equals("Products table") &&
                                  //  m_inheritedT_ACLRights_.get(keyName) != null) {

                                    //    System.out.println("-- New rights for "+keyName);
                                      //  StringRight[] rights = m_inheritedT_ACLRights_.get(keyName);
                                        //for (int i = 0; i < rights.length; i++) {
                                          //  System.out.println("----- new right "+rights[i]);
                                        //} // for ite

                                //} // if (name_

                                  // for testing and tracking
                                  //System.out.println("end of ResourceImpl."+
                                          //"resetInheritedAclEntryRights() ");

    acl = null;
    l_rights = null;
    l_newRights = null;
    order = null;
    key = null;
    l_oldRights = null;
  } // resetInheritedAclEntryRights()

    /**
    * Returns the AclEntry to use as inherited AclEntry in a child.
    * Empty method which returns the argument. Should be never called, and throws
    * an InternalError. Has to be overridden by any AcsAddon handling the ACL inheritance.
    * @param _acl the initial AclEntry which may be given to the child
    * @param _type type of a child, as Resource
    * @return the inherited AclEntry if the child inherits this initial AclEntry, or null
    */
  public AclEntry selectInheritedAclEntry(AclEntry _acl, String _type) {

                            // always thrown
                            //if (_acl instanceof AclEntry) {
                              //  InternalError i = new InternalError("In ResourceImpl, call of " +
                                // "selectInheritedAclEntry()");
                                //i.printStackTrace();
                                //throw i;
                            //}

      return _acl;
  } // selectInheritedAclEntry()

    /**
    * Resets the inherited Privilege rights after an user edition.
    * Called by the master parent when there is a change in the effective rights of
    * an inherited Privilege. This is only when it is about a change in the sources.
    * Contrary to AclEntry, the rights are immutable and there is no condition.
    * The subclassing of CPrivilegeUserImpl is possible in an AcsAddon, but it is
    * also possible to overread the same method in a subclass of this class.
    * Not described in the BeanInfo since the user does not call it.
    * Note: the creation of the inherited Privilege in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited Privilege
    * is handled.
    * Called by Directory.resetInheritedPrivilegeRightsOfChildren().
    * Overridden in DirectoryImpl and called from it.
    * @param _lpri inherited Privilege with updated rights. Never null.
    */
  public synchronized void resetInheritedPrivilegeRights(PrivilegeForLinks _lpri) {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    privHandler_.resetInheritedPrivilegeRights(_lpri);
  } // resetInheritedPrivilegeRights()

    /**
    * Returns the Privilege to use as inherited Privilege in a child.
    * Empty method which returns the argument. Should be never called, and throws
    * an InternalError. Has to be overridden into an AcsAddon handling
    * the privilege inheritance.
    * @param _lpri the initial Privilege which may be given to the child
    * @param _type type of a child, as Resource
    * @return the inherited Privilege if the child inherits this initial Privilege, or null
    */
  public PrivilegeForLinks selectInheritedPrivilege(PrivilegeForLinks _lpri, String _type) {

                            // always thrown
                            if (_lpri instanceof PrivilegeForLinks) {
                                InternalError i = new InternalError("In ResourceImpl, call of " +
                                 "selectInheritedPrivilege()");
                                i.printStackTrace();
                                throw i;
                            }

      return _lpri;
  } // selectInheritedPrivilege()

   /**
   * Gets a reference on all the inherited AclEntries of this target.
   * This method is not in the Resource interface.
   * Called by the parent class in the gBase package, to prepare the updating of the list
   * with a call to setEorL_TargetInheritedAclEntries(). To use by the AcsAddons.
   * @see #getEorImmutableL_TargetInheritedAclEntries()
   * @return the internal list of inherited AclEntries, or an empty list.
   * May be a new empty list or an old empty list, but never null. It is not a copy.
   */
  public ArrayList<AclEntry> getEorL_TargetInheritedAclEntries() {
      if (l_inheritedT_AclEntries_ == null) return new ArrayList(INITIAL_CAPACITY_2);
      else return l_inheritedT_AclEntries_;
  } // getEorL_TargetInheritedAclEntries()

   /**
   * Gets all the inherited linked privileges of this target.
   * This method is not in the Resource interface.
   * The LINKED_IF_ALL_SOURCE type is forbidden, and any privilege having more than one target.
   * To use by the parent in the gBase package, to prepare the updating of the list
   * with a call to setEorL_TargetInheritedLinkedPrivileges().
   * The privilege may be a clone, and the target may not be this instance.
   * Caution: a clone is not registered in its ACS, nor in its sources and targets.
   * However, it is possible in the beamer to navigate from a cloned privilege
   * to its targets or sources.
   * @see #getEorL_TargetInheritedLinkedPrivileges()
   * @return reference on the internal list of inherited Linked Privileges.
   * May be a new empty list or an old empty list, but never null.
   */
  public List<PrivilegeForLinks> getEorL_TargetInheritedLinkedPrivileges() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_TargetInheritedLinkedPrivileges();
  }
  
  //
  // ---------------- protected methods ----------------------------------
  //

  /**
   * Called by DirectoryImpl.setL_ownerRights().
   * Do not control if the ACS manages the AG rights inheritance.
   * Note: the access from the parent userID depends only to the userID rights.
  * @param _b true if the parent authorizes its userID to have access to this instance.
  */
  protected void setAGOAccessFromTheParentAccount(boolean _b) {
      boolean old = l_parentTreeAGRightsAllowToGoAcross_[0];
      l_parentTreeAGRightsAllowToGoAcross_[0] = _b;
      try {
        firePropertyChange("FromTheParentAccount", old, _b);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setAGOAccessFromTheParentAccount(),"+
                "firePropertChange() throws:");
        e.printStackTrace();
      }
  } // setAGOAccessFromTheParentAccount()

  /**
   * Called by DirectoryImpl.setL_groupRights().
   * Do not control if the ACS manages the AG rights inheritance.
  * @param _b true if the parent authorizes its groupID to have access to this instance.
  */
  protected void setAGOAccessFromTheParentGroup(boolean _b) {
      boolean old = l_parentTreeAGRightsAllowToGoAcross_[1];
      l_parentTreeAGRightsAllowToGoAcross_[1] = _b;
      try {
        firePropertyChange("FromTheParentGroup", old, _b);
      } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.setAGOAccessFromTheParentGroup(), "+
                "firePropertChange() throws:");
        e.printStackTrace();
      }
  } // setAGOAccessFromTheParentGroup()

  /**
   * Fires the event 'InheritedOwnerRights', from the parent of the resource owner.
   * Called by the parent Directory.setL_ownerRights() method.
   * @param _l_old an array of the old access rights for the UserID which owns this resource,
   *   in the main ACS. May be empty.
   * @param _l_new an array of the new access rights for the UserID which owns this resource,
   *   in the main ACS. May be empty.
   */
  protected final void fireL_inheritedOwnerRights(StringRight[] _l_old, StringRight[] _l_new) {
    if (parent_ != null)
        firePropertyChange("InheritedOwnerRights", _l_old, _l_new);
  } // fireL_inheritedOwnerRights()

  /**
   * Fires the event 'InheritedGroupRights', from the parent of the resource group.
   * Called by the parent Directory.setL_groupRights() method.
   * @param _l_old an array of the old access rights for the UserID which owns this resource,
   *   in the main ACS. May be empty.
   * @param _l_new an array of the new access rights for the UserID which owns this resource,
   *   in the main ACS. May be empty.
   */
  protected final void fireL_inheritedGroupRights(StringRight[] _l_old, StringRight[] _l_new) {
    if (parent_ != null)
        firePropertyChange("InheritedGroupRights", _l_old, _l_new);
  } // fireL_inheritedGroupRights()
  
  /**
  * This method provides a copy of the instance variables.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy).
  * Most of the aclEntries are copied, but not the list of ViewInBase
  * that contains this instance, nor comments.
  * In most of the cases, the aclEntry arguments are copied,
  * but see the exceptions.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * Calls copyOwnAcl().
  * Caution: does not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same last component in this instance user or group names.<P>
  * For AclEntries, if _result ACS is not this resource ACS, and if
  * the AclEntry ACS is this resource ACS, sets _result ACS as
  * the AclEntry ACS of the copied ACL.
  * Caution: does not copy an aclEntry if: 
  *  - _result ACS is not this resource ACS,
  *  - the eligible party ACS is this resource ACS,
  *  - and there is no matched EligibleParty (with the same last components)
  *      in the _result ACS
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return the updated _result
  * @exception CreateError of addAclEntry()
  **/
  protected ResourceImpl copyOwnVariables(ResourceImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in ResourceImpl.copyOwnVariables(),"+
          " _result is null.");
      }

      _result = copyOwnAcl(_result, _withoutInternalAcl, false);
      
      _result.type_ = getResourceType();
      _result.l_rootRights_ = getL_rootRights();
      _result.l_groupRights_ = getL_groupRights();
      _result.l_ownerRights_ = getL_ownerRights();
      try {
          if (_result.aCS_ == aCS_) {
                  _result.setEorUserID( getEorUserID() );
                  _result.setEorGroupID( getEorGroupID() );
          } else {
              if (getEorUserID() != null) {
                // if the copy is for another ACS, try to find the user
                // with the same name last components
                ImmutableName usr_name = ((Name)_result.aCS_.getName()).extended( 
                  getEorUserID().getName().lastComponent() );
                if (_result.aCS_.containsOwnEligiblePartyName(usr_name)
                ) {
                    _result.setEorUserID( (ImmutableUserID)_result.aCS_.
                      getEorM_EligibleParties().get(usr_name) );
                }
                else System.out.println("in ResourceImpl.copyOwnVariables(), "+
                  "the copy user has not been found in the target ACS.");
                usr_name = null;
              } // if
              if (getEorGroupID() != null) {
                // if the copy is for another ACS, try to find the group
                // with the same name last components
                ImmutableName grp_name = ((Name)_result.aCS_.getName()).extended( 
                  getEorGroupID().getName().lastComponent() );
                if (_result.aCS_.containsOwnEligiblePartyName(grp_name)
                ) {
                    _result.setEorGroupID( (ImmutableGroupID)_result.aCS_.
                      getEorM_EligibleParties().get(grp_name)  );
                }
                else System.out.println("in ResourceImpl.copyOwnVariables(), "+
                  "the copy group has not been found in the target ACS.");
                grp_name = null;
              } // if
          } // else
      } catch (UpDateError u) { 
            throw new InternalError("in ResourceImpl.copyOwnVariables(): "+u);
      }
      _result.iS_Name_ = getIS_Name();

      return _result;
  } // copyOwnVariables(,,,) 
  
  /**
  * This method provides a copy of the AclEntries,
  * all the AclEntries (see restrictions) or only the external ones.
  * Most of the aclEntries are copied, but not the list of ViewInBase
  * that contains this instance, nor comments.
  * If _result ACS is not this resource ACS, and if
  * the AclEntry ACS is this resource ACS, sets _result ACS as
  * the AclEntry ACS of the copied acl.
  * Caution: does not copy an internal AclEntry if:
  *  - _result ACS is not this resource ACS,
  *  - the eligible party ACS is this resource ACS,
  *  - and there is no matched EligibleParty (with the same last components)
  *     in the _result ACS
  * Called by copyOwnVariables() and by
  * BaseUtilitiesImpl.copyDirectoryChildrenForACS().
  * @param _result to update with this instance AclEntries copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @param _withoutExternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
   * @return updated _result
  * @exception CreateError of addAclEntry()
  **/
  protected ResourceImpl copyOwnAcl(ResourceImpl _result,
    boolean _withoutInternalAcl, boolean _withoutExternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in ResourceImpl.copyOwnAcl(),"+
          " _result is null.");
      }
      AclEntry[] l_acl = getL_aclEntry();
      if (l_acl == null || l_acl.length == 0) return _result;

                            // for testing and tracking      
                            //System.out.println("call to ResourceImpl.copyOwnAcl() on "+name_+
                              //" for the copy "+_result.getName()+" and _withoutInternalAcl:"+
                              //_withoutInternalAcl);
      
      try {
        for (int i = 0 ; i < l_acl.length ; i++) {
            if (l_acl[i] == null || l_acl[i].isEmpty()) continue;

                            // for testing and tracking        
                            //System.out.println("in ResourceImpl.copyOwnAcl(), "+
                                 //     "copy the acl: "+l_acl[i]);

            if (_result.aCS_ == aCS_) {

                            // for testing and tracking      
                            //System.out.println("in ResourceImpl.copyOwnAcl() for an internal copy.");
                
              if (l_acl[i].getEorEP().getEorACS() != aCS_) {
                  if (!_withoutExternalAcl)
                      _result.addAclEntry(l_acl[i].getEorACS(),
                         l_acl[i].getEorEP(), l_acl[i].getL_Rights(),
                         l_acl[i].getPositiveRight(),
                         l_acl[i].getL_ConditionSourceGroups().iterator().next());
              } else if (!_withoutInternalAcl)
                  _result.addAclEntry(l_acl[i].getEorACS(),
                     l_acl[i].getEorEP(), l_acl[i].getL_Rights(),
                     l_acl[i].getPositiveRight(),
                     l_acl[i].getL_ConditionSourceGroups().iterator().next());
            } else {
                boolean isInnerAcl = false;
                if (l_acl[i].getEorACS() == aCS_) isInnerAcl = true;

                if (l_acl[i].getEorEP().getEorACS() != aCS_) {

                            // for testing and tracking      
                            //System.out.println("in ResourceImpl.copyOwnAcl() for an external copy, "+
                                      //"add an external acl if _withoutExternalAcl is false: "+
                                      //_withoutExternalAcl);
                            
                    if (_withoutExternalAcl) continue;                    
                    // if it is an inner acl, provides a translation with
                    // _result ACS as acl ACS
                    if (isInnerAcl)
                      _result.addAclEntry(_result.aCS_,
                        l_acl[i].getEorEP(), l_acl[i].getL_Rights(),
                        l_acl[i].getPositiveRight(),
                        l_acl[i].getL_ConditionSourceGroups().iterator().next());
                    else _result.addAclEntry(l_acl[i].getEorACS(),
                        l_acl[i].getEorEP(), l_acl[i].getL_Rights(),
                        l_acl[i].getPositiveRight(),
                        l_acl[i].getL_ConditionSourceGroups().iterator().next());

                } else if (!_withoutInternalAcl) {
                    // tries to find out the EP in the _result ACS
                    // with the same name last components
                    ImmutableName ep_name = ((Name)_result.aCS_.getName()).extended( 
                      ((Name)l_acl[i].getEorEP().getName()).without((Name)aCS_.getName()) );

                            // for testing and tracking      
                            //System.out.println("in ResourceImpl.copyOwnAcl() for an external copy, "+
                                      //"add an internal acl to the external ep: "+ep_name);

                    if (_result.aCS_.containsOwnEligiblePartyName(ep_name)
                    ) {
                            // for testing and tracking      
                            //System.out.println("in ResourceImpl.copyOwnAcl() for an external copy, "+
                                      //"the external EP acl is found.");

                        // makes a translation in the _result ACS
                        // if l_acl[i] is an inner aclEntry which belongs to aCS_
                      if (isInnerAcl) _result.addAclEntry(_result.aCS_,
                        (ImmutableEligibleParty)_result.aCS_.getEorM_EligibleParties().get(ep_name),
                        l_acl[i].getL_Rights(),
                        l_acl[i].getPositiveRight(),
                        l_acl[i].getL_ConditionSourceGroups().iterator().next());
                      else _result.addAclEntry(l_acl[i].getEorACS(),
                        (ImmutableEligibleParty)_result.aCS_.getEorM_EligibleParties().get(ep_name),
                        l_acl[i].getL_Rights(),
                        l_acl[i].getPositiveRight(),
                        l_acl[i].getL_ConditionSourceGroups().iterator().next());
                    } // if  (_result.aCS_.containsOwnEligiblePartyName(ep_name)
                    // else the right ep is not found, and the aclEntry is not copied

                } // else if (!_withoutInternalAcl)
            } // else of 'if (_result.aCS_ == aCS_)'

        } // for
      } catch (CreateError c) {
        throw new CreateError("in ResourceImpl.copyOwnAcl(), exception: "+c);
      }

      l_acl = null;
      return _result;
  } // copyOwnAcl(,,,)
  
  /**
   * Removes the userID of the resource. To use ONLY when the UserID is finalizing.
   * In the other cases, use setEorUserID(). It is a bound property.
   * The main difference with setEorUserID() is that _u.removeOwnResource(this)
   * is not invoked in this method.
   */
  protected void removeEorUserID() throws UpDateError {
    UserID oldUserID = userID_;
    userID_ = null;

    firePropertyChange("UserID", oldUserID, userID_);
    oldUserID = null;
  } // removeEorUserID()

  /**
   * Removes the groupID of the resource.
   * To use only when the groupID is finalizing.
   * In the other cases, use setEorGroupID(). It is a bound property.
   * The main difference with setEorGroupID() is that removeOwnResource(this)
   * is not invoked in this method.
   */
 protected void removeEorGroupID() throws UpDateError {
    
      GroupID oldGroupID = groupID_;
      groupID_ = null;

      firePropertyChange("GroupID", oldGroupID, groupID_);
      oldGroupID = null;
  } // removeEorGroupID()

    /**
    * Deletes an inner AclEntry which delivers rights to an EligibleParty
    * to access or to deny access to this Resource.
    * 'EPsThroughAclEntry' and 'AclRights' are bound properties,
    * and 'AclRights' is the last firing.
    * The resulting array may be a 0-size array. It does not contain null or empty AclEntries.
    * After a removing, the order of the remaining AclEntries is not changed.
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, or if _aclAcs is closed or unknown.
    */
  protected final synchronized void removeACL(ImmutableACS _aclAcs,
      ImmutableEligibleParty _ep, boolean _positiveRight, ImmutableGroupID _conditionGroup)
      throws UpDateError {

                              // for testing and tracking
                              //System.out.println("Call of ResourceImpl.removeACL() in: "+name_);
                              //System.out.println("----- for ACS: "+_aclAcs.getName());
                              //System.out.println("----- for EligibleParty: "+_ep.getName());
                              //System.out.println("----- for PositiveRight: "+_positiveRight);
                              //if (_conditionGroup != null)
                                //  System.out.println("----- for ConditionGroup: "+_conditionGroup.getName());
                              //else
                                //  System.out.println("----- for null ConditionGroup");

      // initialization
      short           epCount = 0; // number of references for this resource in _ep
      boolean         maintainAcs = false;
      boolean         maintainEP = false;
      AclEntryImpl    acl = null;
      AclEntry[]      l_oldAclEntry;
      Set             l_oldEPthroughAclEntry = getEorL_EPthroughAclEntry();

      // exceptions
      if (aCS_ == null) throw new UpDateError("In ResourceImpl, the main ACS is null");
      if (_ep == null) {
        aCS_.getEorBaseError().addException("In ResourceImpl.removeACL(_ep),"+
          " _ep == null for the resource: "+name_);
        throw new UpDateError("In ResourceImpl.removeACL(_ep), _ep == null "+
          "for the resource: "+name_);
      }
      // count the occurences of _ep
      for (int i = 0 ; i < l_aclEntry_.length ; i++) {
        if (l_aclEntry_[i] != null) {
          if (l_aclEntry_[i].getEorEP() == _ep) {
            epCount++;
          }
        }
      } // for i
      // exception
      if (epCount == 0) {
          aCS_.getEorBaseError().addException("In ResourceImpl.removeACL(),"
            +" _ep is unknown ");
          throw new UpDateError("In the AclEntry Resource, the EligibleParty of the AclEntry "+
                  "to remove is unknown. ");
      } // if

      // get the old aclEntry list
      l_oldAclEntry = new AclEntryImpl[l_aclEntry_.length];
      System.arraycopy(l_aclEntry_, 0, l_oldAclEntry, 0, l_aclEntry_.length);

      // update l_aclEntry_
          // this code deletes all the relevant AclEntries even if epCount > 1,
          // but only one AclEntry is finalized
      short         shiftingBack = 0;
      AclEntry[]    newArray = new AclEntryImpl[l_aclEntry_.length];
      AclEntry[]    result = new AclEntryImpl[0];

      for (int  i = 0 ; i < l_aclEntry_.length ; i++) {

                              // for testing and tracking
                              //System.out.println("ResourceImpl.removeACL() in: "+name_+
                                //", scanning of the ACL: "+l_aclEntry_[i]);

        if (l_aclEntry_[i] != null && !l_aclEntry_[i].isEmpty()) {
          if (l_aclEntry_[i].getEorEP() == _ep &&
            l_aclEntry_[i].getEorACS() == _aclAcs &&
            l_aclEntry_[i].getPositiveRight() == _positiveRight &&
            _conditionGroup == null) {
                  // incrementing shiftingBack deletes the reference l_aclEntry_[i]
              shiftingBack++;

                              // for testing and tracking
                              //System.out.println("ResourceImpl.removeACL() in: "+name_+
                                //" removes at "+i+" the ACL: "+l_aclEntry_[i]);

                  // set acl to unrecord the aclEntry in the eligible party, and in its ACS
                  // (plus, if necessary, unrecord this controlled resource in the aclEntry ACS):
              acl = (AclEntryImpl) l_aclEntry_[i];
          }
          else if (l_aclEntry_[i].getEorEP() == _ep &&
            l_aclEntry_[i].getEorACS() == _aclAcs &&
            l_aclEntry_[i].getPositiveRight() == _positiveRight &&
            !l_aclEntry_[i].isRightDefinedConditionalAclEntry() &&
            !l_aclEntry_[i].getL_ConditionSourceGroups().isEmpty() &&
                  // operator '==' does not work there n all cases (see DirectoryMySQLImpl)
            ((GroupID)l_aclEntry_[i].getL_ConditionSourceGroups().toArray()[0]).
                       getNickName().equals(_conditionGroup.getNickName()) ) {
                  // incrementing shiftingBack deletes the reference l_aclEntry_[i]
              shiftingBack++;

                              // for testing and tracking
                              //System.out.println("ResourceImpl.removeACL() in: "+name_+
                                //" removes at "+i+" the ACL: "+l_aclEntry_[i]);

                  // set acl to unrecord the aclEntry in the eligible party, and in its ACS
                  // (plus, if necessary, unrecord this controlled resource in the aclEntry ACS):
              acl = (AclEntryImpl) l_aclEntry_[i];
          }
          else newArray[i - shiftingBack] = l_aclEntry_[i];
        }
        else shiftingBack++; // delete null and empty l_aclEntry_[i]
      } // for

      if (l_aclEntry_.length != shiftingBack) {
        result = new AclEntryImpl[l_aclEntry_.length - shiftingBack];
        System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
      } // else result (already egal to) new AclEntryImpl[0]
      
      l_aclEntry_ = result;

                              // for testing and tracking
                            //System.out.println("ResourceImpl.removeACL() new list of ACLs: ");
                            //for (int  i = 0 ; i < l_aclEntry_.length ; i++) {
                              //  System.out.println("--- "+l_aclEntry_[i]);
                            //} // for i


                                      // for testing and tracking
                                      //System.out.println("ResourceImpl.removeACL() in: "+name_+
                                        //" before the AclEntry finalization of "+acl);

      // finalize the aclEntry with l_aclEntry_ already updated.
      // this method updates the ACS and, only if epCount == 1, the EP of the acl
      acl.finalizeForUser(epCount);

                      // for testing and tracking
                      //System.out.println("ResourceImpl.removeACL() in: "+name_+
                        //" after the AclEntry finalization.");

      // removes _aclAcs from the controller list if it was its last AclEntry.
      // a resource cannot handles both external ACL and external LPRI.
      if (!_aclAcs.getName().equals(aCS_.getName()) ) {
        for (int  i = 0 ; i < l_aclEntry_.length ; i++) {
          if (l_aclEntry_[i] != null) {
            if ( l_aclEntry_[i].getEorACS() == _aclAcs ) {
              maintainAcs = true;
              break;
            } // if
          } // if (l_aclEntry_[i] != null
        } // for
        if (!maintainAcs ) removeAcsController(_aclAcs.getName() ); // fires UpDateError
      } // if (!_aclAcs.getName().equals(aCS_.getName()) )

      // fires a change on "EPsThroughAclEntry" if it was the last link to the acl EP
      for (int  i = 0 ; i < l_aclEntry_.length ; i++) {
        if (l_aclEntry_[i] != null) {
          if ( l_aclEntry_[i].getEorEP() == _ep) {
            maintainEP = true;
            break;
          } // if
        } // if (l_aclEntry_[i] != null
      } // for

                          // for testing and tracking
                  //System.out.println("at the end of ResourceImpl.removeACL(), "+
                    //"the aclEntry list is:");
                  //for (int i = 0; i < l_aclEntry_.length; i++) {
                    //  System.out.println(name_+".l_aclEntry_["+ i +"] :" + l_aclEntry_[i] );
                  //}

      // fires the bound property
      try {
          firePropertyChange("AclRights", l_oldAclEntry, getL_aclEntry());
          if (!maintainEP) {
            firePropertyChange("EPsThroughAclEntry", l_oldEPthroughAclEntry,
              getEorL_EPthroughAclEntry());
          }
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ResourceImpl.removeACL(), firePropertChange() throws:");
            e.printStackTrace();
      }

      l_oldAclEntry = null;
      newArray = null;
      l_oldEPthroughAclEntry = null;
      result = null;
      acl = null;
  } // removeACL()

  /**
   * NO USE.
  * This method finalizes the AclEntries and the other attibutes, since it is used
  * after a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called by EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In ResourceImpl, call of " + name_ +
                //".finalizeFromDeserialization()");

    if (folderMember_ == null) return;

    // removes the aclEntries, Acs controllers in this and in its main ACS
    // (not very fast)
    AclEntry[] aclArray = new AclEntry[1];
    if (l_aclEntry_ != null) {
      aclArray = new AclEntry[l_aclEntry_.length];
      System.arraycopy(l_aclEntry_, 0, aclArray, 0, l_aclEntry_.length);
      for (int  i = 0 ; i < aclArray.length ; i++) {
        if (aclArray[i] != null && !aclArray[i].isEmpty()) {
          ((AclEntryImpl)aclArray[i]).finalizeClosedAcl();
        }
      } // for
    } // if

    l_aclEntry_ = null;
    parent_ = null;
    l_rootRights_ = null;
    //l_otherRights_ = null;
    l_groupRights_ = null;
    l_ownerRights_ = null;
    l_OpenAcsControllerNames_ = null;
    l_inheritedT_AclEntries_ = null;
    m_inheritedT_ACLRights_ = null;
    
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  /**
   * Called by addAclEntry() and addEorPrivilegeForLinksToTarget().
   * Fires the PropertyChangeEvent 'AclAcsController' or 'PriAcsController'.
   * @param _ACSname open ACS which controls this resource. No exception if it is
   * already known.
   * @exception CreateError if the Name is null, the ACS is closed, or if
   *  it is the main ACS of this resource.
   */
  protected void addAcsController(Name _ACSname) throws CreateError {

                        // for testing and tracking
                    //System.out.println("ResourceImpl.addAcsController() starts");
      
    if (aCS_ == null) return;

    // method variable:
    Map             m_openACS = BaseManagerImpl.getInstance().getEorM_OpenACS();
    boolean         known = false;
    ImmutableName[] newArray = null;
    ImmutableName[] l_oldAcsControllers = null;

    // test of UpDateError exceptions
    if (_ACSname == null) {
      aCS_.getEorBaseError().addException("Resource.addAcsController(Name): Name==null ");
      throw new CreateError(" The controller name of the resource is null.");
    }
    if (m_openACS.get(_ACSname) == aCS_) {
      aCS_.getEorBaseError().addException("Resource.addAcsController(Name): Name is main ACS name");
      throw new CreateError(" The controller name of the resource is the main ACS name.");
    }
    if (!m_openACS.containsKey(_ACSname) ) {
        aCS_.getEorBaseError().addException("Resource.addAcsController(Name): Name of a closed ACS");
        throw new CreateError(" The controller name of the resource is the name of a closed ACS.");
    }

    if (l_OpenAcsControllerNames_ != null && l_OpenAcsControllerNames_.length != 0) {
        // test if _ACSname is not known:
        for (int i = 0 ; i < l_OpenAcsControllerNames_.length ; i++) {
            if (_ACSname.equals(l_OpenAcsControllerNames_[i]) ) {
              known = true;
              break;
            }
        } // for
    } // if (l_OpenAcsControllerNames_

    if (known) {
        m_openACS = null;
        return;
    }

    l_oldAcsControllers = getL_AcsControllers();

    // _ACSname unknown, so update l_OpenAcsControllerNames_ in an empty place if possible
    if (l_OpenAcsControllerNames_ != null && l_OpenAcsControllerNames_.length != 0) {
        for (int i = 0 ; i < l_OpenAcsControllerNames_.length ; i++) {
            if (l_OpenAcsControllerNames_[i] == null) {
                l_OpenAcsControllerNames_[i] = (Name)_ACSname.clone();
                known = true;
                break;
            }
        } // for
    } // if
        // else, add at the end of l_OpenAcsControllerNames_
    if (!known) {

        if (l_OpenAcsControllerNames_ != null && l_OpenAcsControllerNames_.length != 0) {
            newArray = new NameImpl[l_OpenAcsControllerNames_.length + 1];
            System.arraycopy(l_OpenAcsControllerNames_, 0, newArray, 0, l_OpenAcsControllerNames_.length);
            newArray[l_OpenAcsControllerNames_.length] = (Name)_ACSname.clone();
        }
        else {
            newArray = new NameImpl[1];
            newArray[0] = (Name)_ACSname.clone();
        } // else
        l_OpenAcsControllerNames_ = newArray;
        
    }

    try {
         if (aCS_.getManageExternalAcls())
             firePropertyChange("AclAcsControllers", l_oldAcsControllers, getL_AcsControllers());
         else if (aCS_.getManageExternalForOneToOneLinkedPrivileges())
             firePropertyChange("PriAcsControllers", l_oldAcsControllers, getL_AcsControllers());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.addAcsController(), firePropertChange() throws:");
        e.printStackTrace();
    }

                        // for testing and tracking
                    //System.out.println("ResourceImpl.addAcsController() ends");

    l_oldAcsControllers = null;
    m_openACS = null;
    newArray = null;
  } // addAcsController()

  /**
   * Called by removeAclEntry() and removeEorPrivilegeToTarget().
   * Fires the PropertyChangeEvent 'AclAcsController' or 'PriAcsController'.
   * @param _ACSname open ACS that controls this resource.
   * @exception UpDateError if _ACSname is null or unknown.
   */
  protected void removeAcsController(ImmutableName _ACSname) throws UpDateError {

                        // for testing and tracking
                    //System.out.println("ResourceImpl.removeAcsController() starts");
      
    if (aCS_ == null) return;

    ImmutableName[]     l_oldAcsControllers = getL_AcsControllers();

    // test of UpDateError exceptions
    if (_ACSname == null || l_oldAcsControllers == null) {
      aCS_.getEorBaseError().addException("ResourceImpl.removeAcsController(Name): null argument or null list ");
      throw new UpDateError("ResourceImpl.removeAcsController() with a null argument or a null list.");
    }

    // update l_OpenAcsControllerNames_
        // this code deletes all the relevant names even if count > 1,
    short               shiftingBack = 0;
    ImmutableName[]     newArray = new NameImpl[l_oldAcsControllers.length];
    ImmutableName[]     result = null;

    for (int  i = 0 ; i < l_oldAcsControllers.length ; i++) {
        if (l_oldAcsControllers[i].equals(_ACSname) ) {
              // incrementing shiftingBack deletes the reference
          shiftingBack++;
        }
        else newArray[i - shiftingBack] = l_oldAcsControllers[i];

    } // for

    if (shiftingBack == 0) {
        l_oldAcsControllers = null;
        newArray = null;
        throw new UpDateError(
                "ResourceImpl.removeAcsController() with an unknown argument "+
                _ACSname);
    } // if

    if (l_oldAcsControllers.length != shiftingBack) {
      result = new NameImpl[l_oldAcsControllers.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
    }
              // else result is null
    l_OpenAcsControllerNames_ = result;

    try {
         if (aCS_.getManageExternalAcls())
             firePropertyChange("AclAcsControllers", l_oldAcsControllers, getL_AcsControllers());
         else if (aCS_.getManageExternalForOneToOneLinkedPrivileges())
             firePropertyChange("PriAcsControllers", l_oldAcsControllers, getL_AcsControllers());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From ResourceImpl.removeAcsController(), firePropertChange() throws:");
        e.printStackTrace();
    }
    
    newArray = null;
    result = null;
    l_oldAcsControllers = null;
  } // removeAcsController()

   /**
   * Sets a privilege as applicable to this access target, without right inheritance.
   * Called by PrivilegeForLinksImpl.addEorPrivilegeTargetLink() which checks
   * the type of the target to forbid a change on an immutable target.
   * This instance is registered in the privilege by this method.
   * Calls addAcsController() for an external one-to-one linked privilege.
   * Fires 'OneTargetLinkedPrivilege' events.
   * @param _priv to register in the target
   * @exception BaseError if this instance is not a privilege target,
   * if the privilege second type is not authorized for this target, or
   * from addAcsController().
   */
  protected void addEorPrivilegeForLinksToTarget(PrivilegeForLinks _priv)
          throws BaseError {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if

    privHandler_.addEorPrivilegeToTarget(_priv);

    if (_priv.getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE)) {
      boolean addAcs = true;
      // adds the LPRI ACS from the controller list if it was its last LPRI.
      // a Resource cannot handles both external ACL and external LPRI.
      if (!_priv.getAcsName().equals(aCS_.getName()) ) {
        for (PrivilegeForLinks pri : privHandler_.getEorL_TargetPrivilegeForLinks()) {

            if (_priv == pri) continue;
            if ( _priv.getAcsName().equals(pri.getAcsName()) ) {
              addAcs = false;
              break;
            } // if

        } // for
        try {
            if (addAcs )
                addAcsController((Name)_priv.getAcsName() ); // fires CreateError
        }
        catch (CreateError u) {
            throw new BaseError("ResourceImpl.addEorPrivilegeForLinksToTarget() from "+
                    "addAcsController() - "+u.toString());
        } // catch
      } // if (!_priv.getAcsName().equals(aCS_.getName()) )
    } // if (_priv.getMainType().equals(PrivilegeAbst.LINKED_FOR_ONE_TO_ONE

  } // addEorPrivilegeForLinksToTarget()
  
   /**
    * Note: To use in a subclass for an AcsAddon implementing the ACL inheritance.
    *
   * Sets all the inherited AclEntries for this target. The AclEntry may be a clone.
   * To use by the parent class in the gBase package.
   * No use in this generic gBase package, so it is reserved to the AcsAddon packages.
   * Calls setM_TargetInheritedACLRights() to set the map (source name + order, StringRight[]).
   * Fires 'InheritedTargetAclEntries' events.
   * @param _l_acl list of AclEntries. May be empty, but never null.
   */
  //protected void setEorL_TargetInheritedAclEntries(ArrayList<AclEntry> _l_acl) {
    //  if (_l_acl== null) throw new InternalError("ResourceImpl."+
      //        "setEorL_TargetInheritedAclEntries() with a null argument.");

      //ArrayList l_oldList = null;

      //if (l_inheritedT_AclEntries_ == null) {
        //  l_oldList = new ArrayList(INITIAL_CAPACITY_2);
      //} else {
        //  l_oldList = (ArrayList) l_inheritedT_AclEntries_.clone();
      //} // else

      //l_inheritedT_AclEntries_ = _l_acl;

      //setM_TargetInheritedACLRights(_l_acl);

      //firePropertyChange(
        //      "InheritedTargetAclEntries", l_oldList, l_inheritedT_AclEntries_.clone());

      //l_oldList = null;
  //} // setEorL_TargetInheritedAclEntries()

   /**
    * Note: To use in a subclass for an AcsAddon implementing the PRI inheritance.
    *
   * Sets all the inherited linked privileges for this target.
   * The LINKED_IF_ALL_SOURCE type is forbidden, and any privilege having more than one target.
   * It is recommanded that only the instance parents call this method.
   * Do not register this instance in the inherited linked privileges.
   * No use in this generic gBase package, so for the AcsAddon packages.
   * Fires 'InheritedTargetPrivileges' events.
   * @param _l_priv list of one-target Linked Privileges. May be empty, but never null.
   * @throws InternalError if a privilege has the LINKED_IF_ALL_SOURCE type, or
   * if it has not exactly one target.
   */
  //protected void setEorL_TargetInheritedLinkedPrivileges(ArrayList<PrivilegeForLinks> _l_priv) {
    //if (privHandler_ == null ) {
      //  try {
        //  privHandler_ = new CPrivilegeUserImpl(this);
        //} catch (CreateError c) {
          //  new InternalError("From "+name_+"- "+c.toString());
        //}
    //} // if
    //privHandler_.setEorL_TargetInheritedLinkedPrivileges(_l_priv);
  //}


  //
  // ----------------- package and private methods ---------------------------
  //

  
} // ResourceImpl

