//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;

import java.util.List;
import java.util.TreeMap;
import java.util.SortedMap;
import java.util.Map;
import java.util.Iterator;
import java.io.Serializable;


/**
 * This class is responsible for modeling in the base, a sketch view
 * about access control issues between access sources and an ACSObject, called the center,
 * and between this center and access targets (Resources or VirtualFolders).
 * It is used by the GraphicSketchViews in the GUI, so there is only one running instance,
 * and it is never saved on the base. m_NoSourceTargets_ is used indirectly
 * by GraphicSketchView.setGraphicNodes(), and not the other internal maps.
 * Caution: no equals() nor clone() methods.
 * @see ARoad0.Gui2.GraphicEPRView
 * @version 0.7.0
 */
public class SketchViewInBaseImpl extends EPRViewInBaseImpl
  implements SketchViewInBase, Serializable {

  /* public attributes */
  public static final SketchViewInBase EMPTY_INSTANCE = new SketchViewInBaseImpl();

 /**
  * Default constructor
  */
  public SketchViewInBaseImpl() {
    super();
    type_ = "SketchView";
  }

  /**
   * @param _name of the view. it is copied.
   * @exception CreateError if the view name is already known in BaseManagerImpl.
   */
  public SketchViewInBaseImpl(ImmutableName _name) throws CreateError {
      super(_name);
      type_ = "SketchView";
  }

  /**
   * This constructor is used by ViewInBase_Facade.createSketchView().
   * A sketch view is not register in the base.
   * Calls the super constructor with the same two arguments, and the third argument
   * at 'false'.
   * @param _name of the view. It is copied.
   * @param _m_BaseObjects is the map (name, object) of the view objects.
   *  It is copied. May be null.
   * @exception BaseError if the constructor of EPRViewInBaseImpl fires it
   */
  public SketchViewInBaseImpl(ImmutableName _name, Map _m_BaseObjects) throws CreateError {

      super(_name, _m_BaseObjects, false);
      type_ = "SketchView";
  } // constructor

  /**
   * Sets the BaseObjects displayed in the sketch view, including the center.
   * This method does not listen the changes in the view objects,
   * on the contrary of the add/remove methods.
   * Called by gDMak.ActionOpenSketcher.openSelection() and
   * ViewInBase_Facade.populateSketchView().
   * m_NoSourceTargets_ is used indirectly by GraphicSketchView.setGraphicNodes(),
   * and not the other maps.
   * @param _list of BaseObjects to display in the view. Not null. Includes the view center.
   * @exception BaseError if _list is empty
   */
  public synchronized void setEorL_BaseObjectsForSketcher(List<BaseObject> _list)
          throws BaseError {
    if (_list == null ) throw new BaseError("in SketchViewInBaseImpl."+
      "setEorL_BaseObjectsForSketcher(), the list is null.");

    // variables
    BaseObject bo;

    // removes old values
    for (Iterator ite = ((SortedMap)((TreeMap)m_BaseObjects_).clone()).
            values().iterator() ; ite.hasNext() ; ) {
        bo = (BaseObject)ite.next();
        //derived from removeEorXXX() and super.removeEorBaseObject()
        if (bo instanceof ImmutableSource) {
            m_Sources_.remove( bo.getName());
        }
        if (bo instanceof ImmutableTarget) {
            m_Targets_.remove( bo.getName());
            if (!(bo instanceof ImmutableSource)) {
              m_NoSourceTargets_.remove( bo.getName());
            }
        }
        m_BaseObjects_.remove(bo.getName() );
    } // for ite

    // adds new values
    for (Iterator ite = _list.iterator() ; ite.hasNext() ; ) {
        bo = (BaseObject)ite.next();
        if (bo == null) continue;
        //derived from addEorXXX() and super.addEorBaseObject()
        if (bo instanceof ImmutableSource) {
            m_Sources_.put(bo.getName(), (ImmutableSource)bo );
        }
        if (bo instanceof ImmutableTarget) {
            m_Targets_.put(bo.getName(), (ImmutableTarget)bo );
            if (!(bo instanceof ImmutableSource)) {
              m_NoSourceTargets_.put( bo.getName(), (ImmutableTarget)bo);
            }
        }
        m_BaseObjects_.put(bo.getName(), bo );
    } // for ite

                    // for testing and tracking
                    //System.out.println("In SketchViewInBaseImpl.setEorL_BaseObjectsForSketcher(),"+
                      //" at the end view objects are:");
                    //for (Iterator ite = m_BaseObjects_.keySet().iterator() ; ite.hasNext() ; ) {
                      //System.out.println( ((ImmutableName)ite.next()).toString() );
                    //}


    bo = null;
  } // setEorL_BaseObjectsForSketcher()


} // SketchViewInBaseImpl
