//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;


import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.*;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
//import java.util.LinkedList;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

 /**
  * This class is responsible for managing the canonical (or generic) rights
  * for all the base objects. A metaright is a StringRight, 
  * so it provides rights in a rights network, but without link to an ACS. 
  * It handles the links to the StringRights in every rights network managed by an ACS.
  * A metaright has a position in a matrix (power,field). So each metaright has: <P>
  * - a power, which is bigger as the power of the right is powerfull <P>
  * - a field, which is the name of a column where the closer metarights are grouped <P>
  * A metaright may be in a floating subnetwork of one or several ACS rights which are
  * not associated to the main stream of metarights, without any capacity to compare
  * the powers. <P>
  * A metaright may be used as a right in an ACS, to avoid StringRight creation
  * and to have a pool of rights for several ACS. A metaright is never typed and may be
  * classed.
  * Most of denying rights have 'deny_' as first caracters in their right, and
  * 'none' is the single exception.
  * Note: the implementation class is not a subclass of StringRightImpl, to be faster,
  * and it uses an embedded StringRightImpl instance, following the 'composition' pattern.
 * It is a javabean with the following bound properties:
 * 'Comment', 'Parent', 'Slave', 'Desactivated', 'UpperRights', 'LowerRights',
 * 'ACSRights'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched
 * in the updating methods, and a dialog box is displayed to inform the user.
  * @see UtilityImpl#getFirstIncreasingRightsChain
  * @version 0.7.0
  */
public final class StringMetaRightImpl implements StringMetaRight {

  /*
  * private variables
  */
  private static final long serialVersionUID = 0;
  private static final int INITIAL_CAPACITY = 5 ;  // the default value is 10
  private static final int INCREMENT_CAPACITY = 3;
  
  /**
   * The right, as an embedded instance.
   */
  protected StringRightImpl sRight_;
  
  /**
   * This is the power, which is the line value in the matrix (power, field) 
   * which helps to locate the metaright.
   */
  private int power_ = 0;
    
  /**
   * This is the field, which is the colomn value in the matrix (power, field) 
   * which helps to locate the metaright.
   */
  private int field_ = 0;
  
  /**
  * This is the ACS StringRights which are strictly equivallent to this StringMetaRight 
  * in each open ACS. This map is transient to avoid problems at the serialization.
  * Caution: getM_ACSRights() has not these keys and values.
  * There, the key is an ACS name, and the value is a StringRightImpl or a Set of them.
  */
  protected transient HashMap m_ACSRights_ = new HashMap(INITIAL_CAPACITY, INCREMENT_CAPACITY);
 
  /**
   * boolean true if the metaright is in a floating subnetwork for an ACS,
   *    without any association outside the subnetwork.
   */
  private boolean inFloatingSubNetwork_ = false;
 
  /**
   * boolean true if the metaright is in a granting right, not a denying one.
   */
  private boolean grant_ = true;

  /**
   * list of PropertyChangeListeners for the property 'ACSRights'.
   * No use of SwingChangePropertySupport class (bug in hasListeners() method).
   * The other bound properties are managed by sRight_.
   */
  private transient Set l_ACSRightsChangeSupport_=
       new HashSet(INITIAL_CAPACITY);   
  
  /**
   * This is an array of metarights which have the same power and are in
   *  the same field that this instance. 
   */
  //private StringMetaRight[] l_NeighbourMetaRights_;
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final StringMetaRightImpl EMPTY_INSTANCE = new StringMetaRightImpl();

  /** standardized variable for the JDK 2 serialization
  * @serial 
  *  ("sRight_", StringRightImpl.class),
   * ("power_", Integer.class),
   * ("field_", Integer.class),
   * ("inFloatingSubNetwork_", Boolean.class)
   * ("grant_", Boolean.class)
   */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("sRight_", StringRightImpl.class),
      new ObjectStreamField("power_", Integer.class),
      new ObjectStreamField("field_", Integer.class),
      //new ObjectStreamField("m_ACSRights_", HashMap.class),
      //new ObjectStreamField("l_NeighbourMetaRights_", StringMetaRight[].class),
      new ObjectStreamField("inFloatingSubNetwork_", Boolean.class),
      new ObjectStreamField("grant_", Boolean.class)
  };
  
  /**
   * Constructor used by EMPTY_INSTANCE.
  */
  public StringMetaRightImpl() {
       sRight_ = new StringRightImpl();
  }

  /**
   * Effective constructor. Does not call BaseManagerImpl.newMetaRight(this),
   * so it has to be done after the construction for any new non-basic metaright.
   * Called by ACSFactoryUtilityImpl.initializeBasicMetaRights() and ACSImpl.closeRights().
   * @param _kernelRight is the right if the StringRight is not typed, or otherwise
   *    it is the beginning of the right. Never null.
   * @param _targetClass is the target class, non-null only when this StringRight is typed.
   *    Do not use the 'Immutable' version of the interface. A typical value
   *    is "ARoad0.gBaseInterface.Directory".
   * @param _internalKey is the internal key which allows rights comparisons.
   *    From 1 to 99999.
   * @param _power is the line value in the matrix (power, field). From 1 to 99. Not null.
   * @param _field is the colomn value in the matrix (power, field). From 1 to 99. Not null.
   * @param _grant is true if it is a granting metaright. False for a denying right.
   * @exception CreateError if the StringRightImpl constructor
   *   - the instance must then be deleted !
   */
  public StringMetaRightImpl(String _kernelRight, String _targetClass,
    Integer _internalKey, Integer _power, Integer _field, boolean _grant)
    throws CreateError {

        if (_power == null || _power.intValue() < 1 || _power.intValue() > 99)
            throw new CreateError("in StringMetaRightImpl constructor, the power "+
              "is not correct.");      
        if (_field == null || _field.intValue() < 1 || _field.intValue() > 99)
            throw new CreateError("in StringMetaRightImpl constructor, the field "+
              "is not correct."); 
        
        sRight_ = new StringRightImpl(_kernelRight, _internalKey, _targetClass); // throws CreateError
        
        power_ = _power.intValue();
        field_ = _field.intValue(); 
        grant_ = _grant;

        // TO USE ONLY FOR TESTING since metarights are created 
        // by BaseManager.getInstance() at present.
        // (a metaright cannot be finalized by the user)
               //BaseManagerImpl.newMetaRight(this, false);
        
                // for testing and tracking
               //System.out.println("End of the StringMetaRightImpl constructor for "+_kernelRight);        

  } // StringRightImpl()


  // ------------------------ Methods of the interface ------------------
   
  /**
   * @return an empty name.
   */
  public final ImmutableName getAcsName() {
      return NameImpl.EMPTY_INSTANCE;
  }

 /**
 * Get the full name of the right, which is unique for the program.
 * The name is composed of:
 * &lt;FOR&gt;::  target class wrapped with spaces, or '   all   '
 * &lt;VALUE&gt;::  right <P>
 * here, ':: ' is added by gBase.NameImpl.toString().
 * This name is used only outside gBase, following BaseObject interface.
 * All the denying rights have 'deny_' as first caracters in their right.
 * @return ImmutableName of the right, which is empty if it is an empty right
 */
  public final ImmutableName getFullName() {
      if (isEmpty()) return NameImpl.EMPTY_INSTANCE;
      
      String clas = null;
      
      Name name = new NameImpl("<FOR>");

      if (sRight_.getTargetClass() != null) {
          clas = sRight_.getTargetClass().substring(22); // '22' to remove 'ARoad0.gBaseInterface.'
          if (clas.length() == 3 )
              clas = "    "+clas+"    ";
          else if (clas.length() == 4 || clas.length() == 5)
              clas = "   "+clas+"   ";
          else if (clas.length() == 6)
              clas = "  "+clas+"  ";
          else if (clas.length() == 7)
              clas = " "+clas+" ";
          else if (clas.length() == 8)
              clas = clas+" ";
          name.extended( clas ); 
      } else name.extended("     all     ");
      
      name.extended("<VALUE>" );
      name.extended(sRight_.getRight());      
      return name;
  } // getFullName()
  
  /**
 * Get the BaseObject name of the metaright, that is its full name.
 * @return getFullName()
 * @see #getFullName
 * @see BaseUtilityImpl#getRightFromNameAsString
   */
  public final ImmutableName getName() {
      return getFullName();
  } // getName()
    
  /**
   * Gets a short name limited to 5 characters if positive, 16 first characters if starts with
   * '|' and positive, and 7 if negative, for displaying this right in the views. <P>
   * The exception is for the administrative rights where the 15 first characters
   * are put in the nick name: <P>
   * If right_ is '|modify_security_descriptor|', the nick name is '|modify_security|'. <P>
   * If right_ is 'read', the nick name is the same.
   * If right_ is 'read&execute', the nick name is 'read&'.
   * If right_ contains '_', the nick name 2 last characters are '_%' where '%'
   * is here the character at the right of the last '_' in right_. For example,
   * 'write_for_directory' gives 'wri_d'.
   * The nick name for a negative right starts with 'd_', if the right
   * starts by 'deny_'. The other parts of the nick name are like for a positive right.
   * If right is 'deny_read', the nick name is 'd_read'.
   * If the right is 'deny_write_for_directory', the nick name is 'd_wri_d'.
   * If the negative right does not start with 'deny_', it is processed like
   * a positive right and the nick name size is limited to 5.
   * This short name is NOT unique for the Access Road program, nor for the ACS if
   * it is an ACS object.
   * @return the nick name. Never null.
   */
    public final String getNickName() {
       return sRight_.getNickName();
    } 
    
  /**
   * @return the internal key
   */
    public final int getInternalKey() {
       return sRight_.internalKey_;
    } 
    
  /**
   * @return true if the right is always associated to a class for a target,
   *   as for exemple 'delete_Actor' for the class Actor.
   */
   public final boolean isTargetClassed() {
       return sRight_.isTargetClassed();
   }
    
  /**
   * @return boolean true if the right is always associated to a type for a target
   *   class, as for exemple in LCA Lotus Notes 'read_javascript' for the class Actor.
   */
   public final boolean isTargetClassTyped() {
       return sRight_.isTargetClassTyped();
   }
    
  /**
   * @return true if the metaright is a granting right - the opposite one is
   *   a denying right.
   */
   public boolean isGrantingRight() {
       return grant_;
   }
   
  /**
   * @return true if the embedded StringRightImpl is equals to an empty instance
   */
  public boolean isEmpty() {
      return sRight_.isEmpty();
  }
   
  /**
   * If the metaright is non-classed, the result contains the kernel right only.
  * If the metaright is classed,
  * the result contains the kernel right and the class target.
  * @return the right to apply
 */
  public final String getRight() {
      return sRight_.getRight();
  }

    /**
    * @return the complete name of the class from which the type is provided.
    *   Null if the right is not typed. 
    */
  public final String getTargetClass() {
      return sRight_.getTargetClass();
  }

    /**
    * @return null since a metaright has no type.
    */
  public final String getTargetClassType() {
      return null;
  }

    /**
    * Set any comment related to this right. It is a bound property.
    * @param _s is an user comment which is associated to the right.
    */
    public synchronized void setComment(String _s)	{
        sRight_.setComment(_s);
        return;
    }

    /**
     * @return the comment which is associated to the right. May be null.
     */
    public final String getComment()	{
        return sRight_.getComment();
    }  
  
  /**
   * Add an upper metaright which are strictly greater than this metaright
   * without any intermediate right, following the network of metarights.
   * The two rights may have not the same sens, as for '|modify_security_descriptor|'
   * and 'deny_all'.
   * An upper right cannot be an administrative right (that is, having a strenght
   * superior to 90000) if this instance is not an administrative right.
   * It is a bound property.
   * @param _right is directly included in this right. Never null.
   * @exception BaseError if the right is null, if this instance is a slave,
   *    or if the upper right is not superior to this instance,
   *    or an upper right is an administration right while the new right is not,
   *    or the number of upper rights is already 5.
   */
  public final void addUpperRight(StringMetaRight _right) throws BaseError {

      // test if the upper right is comparable and superior to this right 
      if (_right.compareTo(this) <= 0)
        throw new BaseError("in StringRightImpl.addUpperRight(), the upper right "+_right+
              " is not superior to this metaright "+this);
      if (getInternalKey() <= 90000  && _right.getInternalKey() > 90000)
        throw new BaseError("in StringMetaRightImpl.addUpperRight(), the upper right "+
              "is an administration right "+_right+" while this right is not: "+this); 
      
      sRight_.addUpperMetaRight((StringMetaRightImpl)_right); 
      ((StringMetaRightImpl)_right).addLowerRight(this);      
      return;
  } // addUpperRight()
  
  /**
   * Remove a upper metaright which are strictly greater than this metaright
   * without any intermediate right, following the network of metarights.
   * May be called by the greater metaright finalizeForUser() method.
   * It is a bound property.
   * @param _right is directly included in this right. Never null.
   * @param _caller is the object which  calls this method; used to
   *   identify the upper right finalizing operation with _caller == _right.
   */
  public final void removeUpperRight(StringMetaRight _right, Object _caller) {
      sRight_.removeUpperRight(_right, _caller);
      if (_right != _caller) ((StringMetaRightImpl)_right).removeLowerRight(this);      
      return;     
  } // removeUpperRight
  
  /**
   * Get the set of upper metarights which are strictly greater than this metaright
   * without any intermediate right, following the network of metarights.
   * @return an array of connected StringMetaRightImpl[] which are strictly greater than
   *    this right. Never null but may be empty.
   */
  public final StringRight[] getUpperRights() {
      return sRight_.getUpperRights();
  }

  /**
   * Get the set of lower metarights which are strictly lesser than this metaright
   * without any intermediate right, following the network of metarights.
   * @return an array of connected StringMetaRightImpl[] which are strictly lesser than 
   *    this right. Never null but may be empty.
   */
  public final StringRight[] getLowerRights() {
      return sRight_.getLowerRights();      
  }

  /**
   * Get the global Access Road right which is a generic description of this ACS right.
   *    Metarights are used in the rights comparision and analysis to
   *    compare rights from different ACS.
   * @return itself.
   */
  public final StringMetaRight getEorMetaRight() {
      return this;
  }
    
  /**
   * @return true if the right is a parent of lower rights - the slaves -
   *   in a multi-level right.
   */
   public final boolean isParent() {
       return sRight_.isParent();
   }
    
  /**
   * A slave right has a strong link with a parent right.
   * @return true if the right has an unique upper right - the parent -
   *   in a multi-level right.
   */
   public final boolean isSlave() {
       return sRight_.isSlave();
   }
  
  /**
   * UNUSED, the value is always false.
   * Set the right as parent of some lower rights - the slaves -
   *   in a multi-level right. There is no inverse method.
   * This method has to be used before the registration of the lower rights.
   * Caution: some lower rights may be not slaves, so this instance may be not parent
   * for all its lower rights.
   * @param _b is true to set the right as a parent of slave rights
   * @exception BaseError if there are lower rights.
   */
   public final void setParent(boolean _b) throws BaseError {
       sRight_.setParent(_b);
       return;
   } // setParent()
    
  /**
   * UNUSED, the value is always false.
   * Set that the right has an unique upper right - the parent,
   * already registered - in a multi-level right. 
   * @param _b is true to set the right as a slave of a parent right
   * @exception BaseError if there is more than one upper right,
   *    or if it is not a parent.
   */
   public final void setSlave(boolean _b) throws BaseError {
       sRight_.setSlave(_b);
       return;       
   } // setSlave()
    
  /**
   * Set that the right is desactivated. Used to create a list of flat rights
   * with one single desactivated root. All the associated open ACS rights 
   * are desactivated.
   * @param _b is true to desactivate the right
   */
   public final void setDesactivated(boolean _b) {
       sRight_.setDesactivated(_b);
      // null after a deserialization
      if (m_ACSRights_ == null)
          m_ACSRights_ = new HashMap(INITIAL_CAPACITY, INCREMENT_CAPACITY);
       
       Map.Entry ent = null;
      
      for (Iterator ite = m_ACSRights_.entrySet().iterator(); ite.hasNext(); ) {
          ent = (Map.Entry)ite.next();
          if (ent.getValue() instanceof StringRight)
              ((StringRight)ent.getValue()).setDesactivated(_b);
          else if (ent.getValue() instanceof Set) {
              for (Iterator ita = ((Set)ent.getValue()).iterator(); ita.hasNext(); ) {
                  ((StringRight)ita.next()).setDesactivated(_b);
              } // for ite
          } // else if (ent.getValue() instanceof Set)
      } // for ite
       return;       
   } // setDesactivated()
    
  /**
   * Get if the right is desactivated.
   */
   public final boolean isDesactivated() {
       return sRight_.isDesactivated();
   } // isDesactivated()

  /**
  * Add a listener to the bean, for all the bound properties.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public final void addPropertyChangeListener(PropertyChangeListener _l){
    sRight_.addPropertyChangeListener(_l);
    // initialize the transiant variables after a deserialization
    if (l_ACSRightsChangeSupport_ == null) {
      l_ACSRightsChangeSupport_= new HashSet(INITIAL_CAPACITY);      
    }

    synchronized (l_ACSRightsChangeSupport_)
      {l_ACSRightsChangeSupport_.add(_l);}

              // for testing and tracking
              //System.out.println("in StringRightImpl.addPropertyChangeListener(_l), "+
              //  "call from "+_l.getClass()+" to "+right_+" for the property 'Comment'.");  
    return;
  } // addPropertyChangeListener()

  /**
  * Add a listener to the bean. For the properties:
  * 'Comment', 'Parent', 'Slave', 'Desactivated', 'LowerRights', 'UpperRights',
  * 'ACSRights'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public final void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      sRight_.addPropertyChangeListener(_propertyName, _l);
      
      // initialize the transiant variables after a deserialization
      if (l_ACSRightsChangeSupport_ == null) {
        l_ACSRightsChangeSupport_= new HashSet(INITIAL_CAPACITY);
      }

      if (_propertyName.equals("ACSRights")) {
          synchronized (l_ACSRightsChangeSupport_) {
            l_ACSRightsChangeSupport_.add(_l);}
      } // if
      
      return;
  } // addPropertyChangeListener(String _propertyName,

  /**
  * Remove a listener to the bean, for all the bound properties.
  * Used by ViewInBaseImpl.
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(PropertyChangeListener _l){

    // no action if the transiant variables are null, after a deserialization
    if (l_ACSRightsChangeSupport_ == null) return;

    sRight_.removePropertyChangeListener(_l);
    synchronized (l_ACSRightsChangeSupport_)
      {l_ACSRightsChangeSupport_.remove(_l);}
    
    return;
  } // removePropertyChangeListener()

  /**
  * Remove a listener to the bean. For the properties:
  * 'Comment', 'Parent', 'Slave', 'Desactivated', 'LowerRights', 'UpperRights',
  * 'ACSRights'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variables are null, after a deserialization
      if (l_ACSRightsChangeSupport_ == null) return;

      if (_propertyName.equals("ACSRights")) {
          synchronized (l_ACSRightsChangeSupport_) {
            l_ACSRightsChangeSupport_.remove(_l);}
      } // if
      
      sRight_.removePropertyChangeListener(_propertyName, _l);
      
      return;
  } // removePropertyChangeListener()   
  
  /**
   * Get a clone of the instance. No use since a right does not change.
   * Caution: the isDesactivated_ property is always set to false in the clone.
   * @exception InternalError if CloneNotSupportedException
   * @return is a cloned and no-backed instance of StringRight.
   *   May be null.
   */
  /*public Object clone() {
    StringMetaRightImpl clon = null;
    try {
      clon = (StringMetaRightImpl) super.clone();

      clon.sRight_ = sRight_;

      int incr = 1;
      clon.power_ = power_ + incr;
      clon.field_ = field_;
      if (m_ACSRights_ != null) 
          clon.m_ACSRights_ = (HashMap) ((HashMap)m_ACSRights_).clone();

      
      //if (l_NeighbourMetaRights_ != null) {
        //  clon.l_NeighbourMetaRights_ = new StringMetaRight[
        //    l_NeighbourMetaRights_.length];
        //  System.arraycopy(l_NeighbourMetaRights_, 0, clon.l_NeighbourMetaRights_,
        //    0, l_NeighbourMetaRights_.length);
      //} // if
      
      
      clon.inFloatingSubNetwork_= inFloatingSubNetwork_;
      
    } catch (CloneNotSupportedException e) {
      // this shouldn't happen, since we are Cloneable
      throw new InternalError(e.toString() + "StringMetaRightImpl.clone() failed");
    } // catch
    return clon;
  } // clone()
   **/
  
    /**
     * Compares the two StringMetaRights without comparison of the property
     * m_ACSRights_, since a deserialized metaright in an ACS rights array, may have
     * a null m_ACSRights_ althought it has be to considered as equal to the
     * reference metaright from BaseManagerImpl.
     * @param _obj to compare
     * @return boolean true only if the variables are the same
     */
    public final synchronized boolean equals(Object _obj) {

                // for testing and tracking
                //System.out.println("StringMetaRightImpl.equals() for "+this.getRight());
        
      if (this == _obj) {
                // for testing and tracking
                //System.out.println("StringMetaRightImpl.equals(): objects are identical ");
          return true;
      }         
      if (_obj == null) {
                // for testing and tracking
                //System.out.println("StringMetaRightImpl.equals(): false since _obj is null ");
          return false;
      }
      if (_obj.getClass() != getClass())
          return false;

      StringMetaRightImpl obj = (StringMetaRightImpl)_obj;

      if (!sRight_.equals( obj.sRight_ )) {
                // for testing and tracking
                //System.out.println("StringMetaRightImpl.equals(): false since the embedded rights are not equal ");
          return false;
      }
      
      if (power_ != obj.power_ ) return false;
      
      if (field_ != obj.field_ ) return false;
      
      if ((grant_ && !obj.grant_) || (!grant_ && obj.grant_) ) {
                // for testing and tracking
                //System.out.println("StringMetaRightImpl.equals(): sens are different ");          
          return false;
      }
      
      // no control of m_ACSRights_ since the deserialized metarights in the ACS
      // constraints, have no updated m_ACSRights_ as in BaseManager
      /*
      if (m_ACSRights_ != null) {
          if ( !m_ACSRights_.equals(obj.m_ACSRights_)   ) {
              
                // for testing and tracking
                System.out.println("StringMetaRightImpl.equals(): m_ACSRights_ are different. m_ACSRights_: "); 
                Map.Entry entry = null;
                for (Iterator ite = m_ACSRights_.entrySet().iterator(); ite.hasNext(); ) {
                    entry = (Map.Entry)ite.next();
                    if (entry.getValue() instanceof StringRight)
                        System.out.println("-- for "+entry.getKey()+" value: "+entry.getValue());
                    else if (entry.getValue() instanceof Set) {
                        System.out.println("-- for "+entry.getKey()+", the set is: ");
                        for (Iterator ita = ((Set)entry.getValue()).iterator(); ita.hasNext(); ) {
                            System.out.println("----- value: "+(StringRight)ita.next());
                        } // for ita
                    } // else if
                } // for ite
                System.out.println("and obj.m_ACSRights_: "); 
                for (Iterator ite = obj.m_ACSRights_.entrySet().iterator(); ite.hasNext(); ) {
                    entry = (Map.Entry)ite.next();
                    if (entry.getValue() instanceof StringRight)
                        System.out.println("-- for "+entry.getKey()+" value: "+entry.getValue());
                    else if (entry.getValue() instanceof Set) {
                        System.out.println("-- for "+entry.getKey()+", the set is: ");
                        for (Iterator ita = ((Set)entry.getValue()).iterator(); ita.hasNext(); ) {
                            System.out.println("----- value: "+(StringRight)ita.next());
                        } // for ita
                    } // else if
                } // for ite                
                
              return false;
          }
      }
       * */

                // for testing and tracking
                //System.out.println("StringMetaRightImpl.equals(): true for "+sRight_.getRight());       
      
      return true;
  } // equals(Object)

  /**
  * @return hashcode value for this instance. 
  */
  public final int hashCode() {
    int result = 73;
    if (sRight_ != null) {
      result = result*1000003 + sRight_.hashCode();
    } // if
    return result;
  } // hashcode()

    /**
     * Compare two StringRight instances through the internal key
     * and, if not null, acsname comparisons, whatever the right, power and field.
     * All metarights are superior to any acsright.
     * @param _o to compare
     * @exception ClassCastException - if the specified object's type
     *  prevents it from being compared to this current Object
     * @return 0 if the rights have the same internal key and acsname,
     *  a negative integer if this object is lesser than the argument, 
     *  and a positive integer if this object is greater than the argument.
     */
    public final synchronized int compareTo(Object _o) throws ClassCastException {
     if (!(_o instanceof StringRight))
        throw new ClassCastException(
          "the argument of StringMetaRightImpl.compareTo() is not a StringRight instance");         
     if (_o instanceof StringMetaRightImpl)
         return sRight_.compareTo( ((StringMetaRightImpl)_o).sRight_);
     else return sRight_.compareTo(_o);
  } // compareTo(Object)

  /**
   * @return long serial version UID for serialization of the class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID;
  }

   /**
   * @return the full name as String.
   */
  public final String toString() {
    return getFullName().toString();
  } // toString()
  
  /**
  *  // FROM StringRightImpl, TODO
  * This method provides a copy of the instance for the same or another acs. 
  * It is designed to facilitate the copy of a whole rights network from
  * the most powerfull rights. <P>
  * Each unmodified internal variable that is
  * an external object reference - metaright, lower and upper rights - 
  * have a shallow copy (no deep copy) if the ACS is the same.
  * Otherwise, there is still a shallow copy of the metaright, but the
  * the upper rights are searched in the copy ACS and the lower rights
  * are NOT copied. <P>
  * If the instance is a typed right, the type class and type value are
  * copied.
  * @param _name is the copy name; must be a one-component name,
  *   and the component is the kernelRight_ parameter for the copy.
  * @param _acs is the ACS of the copy to create. May be different
  *   to the ACS of this instance.
  * @param _parent must be null here.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in its ACS, if it exists.
  * @return Object is a copy of the object, in the same class.
  * @exception CreateError if the acs, the kernel right is null.
  * @version 0.5.3
  **/ /*
  public Object copy(ImmutableName _name, ACS _acs, Node _parent,
    Boolean _toRegisterInManager) throws CreateError {

      //if (_kernelRight == null || 
        //_meta != null || _internalKey < 1 || _internalKey > 999) {
          //throw new CreateError(
            //"The right or the metaright is null,"+
            //" or the internal key is not correct.");
      //}
      
      StringRightImpl result = null;
      //if (_acs == null) 

      //else if (_acs == acs_) {

      //else {
          // TODO
      //}
      
      // TODO: set the other arguments

      
      return result;
  } // copy()
       **/

  /**
   * Get the power, which is the line value in the matrix (power, field) 
   * which helps to locate the metaright.
   * @return a bigger value as the power of the right is higher
   */
   public final int getPower() {
       return power_;
   }
    
  /**
   * Get the field, which is the colomn value in the matrix (power, field) 
   * which helps to locate the metaright.
   * @return the number of a column where the near metarights are grouped,
   * all from a limited set of ACS.
   */
   public final int getField() {
       return field_;
   }
  
  /**
  * Get the ACS StringRights which are equivallent to 
  * this MetaStringRight in each ACS. Include the closed-ACS rights.
  * @return Map where the key is an ACS name as String, and the value 
  *   is an array of StringRightImpls. May be empty.
  */
   public final Map<String, StringRight[]> getM_ACSRights() {
      // null after a deserialization
      if (m_ACSRights_ == null)
          m_ACSRights_ = new HashMap(INITIAL_CAPACITY, INCREMENT_CAPACITY);

      if (m_ACSRights_.isEmpty())
          return new HashMap(INITIAL_CAPACITY, INCREMENT_CAPACITY);      
      
      Map<String, StringRight[]> result = new HashMap(INITIAL_CAPACITY, INCREMENT_CAPACITY);
      Map.Entry ent = null;
      
      for (Iterator ite = m_ACSRights_.entrySet().iterator(); ite.hasNext(); ) {
          ent = (Map.Entry)ite.next();
          if (ent.getValue() instanceof StringRight)
              result.put(
                  ((Name)ent.getKey()).toString() ,
                  new StringRight[]{(StringRight)ent.getValue()});
          else if (ent.getValue() instanceof Set)
              result.put(
                  ((Name)ent.getKey()).toString() ,
                  (StringRight[])((Set)ent.getValue()).toArray(new StringRight[0])  );
      } // for ite
      
      ent = null;
      return result;      
   } // getM_ACSRights()
  
  /**
  * Get the StringRights which are equivallent to 
  * this MetaStringRight in one ACS. Applicable to the rights of a closed ACS.
  * @param _aCSname an ACS name
  * @return an array of StringRightImpls. May be empty. Never null.
  */
   public final StringRight[] getACSRightsForOneACS(ImmutableName _aCSname) {
      if (m_ACSRights_ == null && _aCSname == null) return new StringRight[0];
      Object result = m_ACSRights_.get(_aCSname);
      
      if (result instanceof StringRight)
          return new StringRight[] {(StringRight)result};
      else if (result instanceof Set)
          return (StringRight[])((Set)result).toArray(new StringRight[0]);
      
      else return new StringRight[0];      
   } // getACSRightsForOneACS()
   
  /**
   * @return StringMetaRight[] is an array of rights which have the same power and are in
   *  the same field that this instance. May be null.
   */
  //public StringMetaRight[] getL_NeighbourMetaRights() {
    //  if (l_NeighbourMetaRights_ == null) return null;
    //  StringMetaRight[] l_rights = new StringMetaRight[l_NeighbourMetaRights_.length];
    //  System.arraycopy(l_NeighbourMetaRights_, 0, l_rights, 0, l_NeighbourMetaRights_.length);
    //  return l_rights;       
  //}
 
  /**
   * @return boolean true if the metaright is in a floating subnetwork for a specific ACS,
   *    without any association to most of the metarights.
   */
  public final boolean getInFloatingSubNetwork() {
      return inFloatingSubNetwork_;
  }
  
  /**
   * It is not a bound property, since it is set only at the creation.
   * @param _b is true if the metaright is in a floating subnetwork for an ACS,
   *    without any association to the main metarights in the network.
   */
  public void setInFloatingSubNetwork(boolean _b) {
      inFloatingSubNetwork_ = _b;
  }
  
  /**
   * When some relevant ACS are closed, the link to them remain in this metaright.
   * Caution: a metaright may be used in some ACS directly, wihtout any ACS rights.
   * @return boolean true if the metaright has no ACS rights associated to it.
   */
  public final boolean isUnused() {
      if (m_ACSRights_ == null || m_ACSRights_.size() == 0)
          return true;
      else return false;
  }
  
  /**
   * TODO: coherence with compareTo() and equals()???
   * Get a clone of the instance.
   * Caution: the properties grant_ and the embedded StringRightImpl are cloned,
   * and all the other are not cloned nor copied. For instance, getM_ACSRights(),
   * getPower() and getACSRightsForOneACS() are not workable on a clone.
   * Used by ACS.closeAclEntriesTo() when the links to other ACS have to 
   * be cut and cloned.
   * @exception InternalError if CloneNotSupportedException
   * @return is a cloned instance of StringRightImpl.
   */
  /*
  public Object clone() {
    StringMetaRightImpl clon = null;
    try {
      clon = (StringMetaRightImpl) super.clone();

      clon.grant_ = grant_;
      clon.sRight_ = (StringRightImpl) sRight_.clone();
      
    } catch (CloneNotSupportedException e) {
      // this shouldn't happen, since we are Cloneable
      throw new InternalError(e.toString() + "StringRightImpl.clone() failed");
    } // catch
    return clon;
  } // clone()
   * */
  
   /**
    * UNUSED METHOD.
   * Calls finalizeForBase().
   * @exception UpDateError if the base does not known the right (if
   *   it is a clone, typically).
   */
  public final synchronized void finalizeForUser() throws UpDateError {
      finalizeForBase();
  }
  
   /**
    * UNUSED METHOD.
   * This method removes the StringMetaRight from the base list, 
   * finalizes the associated rights,
   * and removes from the lower and upper metarights.
   * This method is called by finalize(), but finalize() may be not called.
   * An unused object has to be deleted to inform the other objects, and this method
   * HAS TO BE CALLED before deleting the instance.
   * Sees finalizeClosedRight() for closing a deprecated right without updating the
   *    external references.
   * @exception UpDateError if the base does not known the right (if
   *   it is a clone, typically).
   */
  public final synchronized void finalizeForBase() throws UpDateError {

                //System.out.println("In StringMetaRightImpl, call of " + name_ + ".finalizeForUser()");

    if (sRight_ == null) return;
    
    StringRight[] l_rights = sRight_.getLowerRights();
    if (l_rights != null) {
      
      if (!sRight_.isParent()) {
          // remove this instance in the lower rights, without call back
          for (int i = 0; i < l_rights.length; i++) {
              ((StringMetaRightImpl)l_rights[i]).removeUpperRight(this, this); 
          } // for
      } else {
          for (int i = 0; i < l_rights.length; i++) {
              // if parent, delete only the slaves, not all the lower rights
              if (l_rights[i].isSlave()) l_rights[i].finalizeForUser();
              // remove this instance in the lower rights
              else {
                  ((StringMetaRightImpl)l_rights[i]).removeUpperRight(this, this);                  
              } // else
          } // for          
      } // else
      
    } // if (l_rights
    l_rights = sRight_.getUpperRights();
    if (l_rights != null) {
      
      // remove this instance in the upper rights, without call back
      for (int i = 0; i < l_rights.length; i++) {
          ((StringMetaRightImpl)l_rights[i]).removeLowerRight(this); 
      } // for
      
    } // if (l_rights    
    l_rights = null;
    
    try {
       BaseManagerImpl.getInstance().removeMetaRight(this);
    } catch (CreateError c) {
        throw new UpDateError(c.toString());
    }
    
    sRight_.finalizeForUser();
    sRight_ = null;
    
    if (m_ACSRights_ == null) {
        power_ = 0;
        field_ = 0;
        return;
    } // if
        
    // finalize all the open ACS rights 
    HashMap map = (HashMap) m_ACSRights_.clone();
    ACSImpl acs = null;
    Map.Entry ent = null;
    Name acsname = null;

    for (Iterator ite = map.entrySet().iterator(); ite.hasNext();) {
        ent = (Map.Entry)ite.next();
        acsname = (Name) ent.getKey();
        acs = (ACSImpl) BaseManagerImpl.getInstance().getEorM_OpenACS().get(acsname);
        if (acs == null) continue;

        if (ent.getValue() instanceof String)
            ((StringRightImpl) acs.getOneRight((String) ent.getValue()) ).finalizeForUser();
        else if (ent.getValue() instanceof Set) {
          Set copy = new HashSet((Set) ent.getValue() );
          for (Iterator ita = copy.iterator(); ita.hasNext();) {
            ((StringRightImpl) acs.getOneRight((String) ent.getValue()) ).finalizeForUser();
          } // for
          copy = null;
        } // else
    } // for ite

    m_ACSRights_.clear();
    m_ACSRights_ = null;   
    
    power_ = 0;
    field_ = 0;
    //l_NeighbourMetaRights_ = null;
    map = null;
    acs = null;
    ent = null;
    acsname = null;    
    return;
  } // finalizeForBase()
   
   //
   // -------------- protected methods -------------------------------
   //
   
  /**
   * Set the power, which is the line value in the matrix (power, field) 
   * which helps to locate the metaright.
   */
   protected void setPower(int _p) {
       power_ = _p;
       return;
   }
    
  /**
   * Set the field, which is the colomn value in the matrix (power, field) 
   * which helps to locate the metaright.
   */
   protected void setField(int _f) {
       field_ = _f;
       return;
   }
   
  /**
   * Add an ACS right which is associated to this metaright in an ACS.
   * Called by the ACS right constructor and at the ACS opening by ACSImpl.checkClosedRights().
   * This metaright may have already some
   * right(s) associated to the same ACS. It is a bound property.
   * @param _right which associates this metaright to it. No action if null.
   */
  protected final void addACSRight(StringRightImpl _right) {
        if (_right == null) return;
        // null after a deserialization
        if (m_ACSRights_ == null)
          m_ACSRights_ = new HashMap(INITIAL_CAPACITY, INCREMENT_CAPACITY);        
        Map old = getM_ACSRights();
        
        if (m_ACSRights_.get(_right.getAcsName()) == null)
          m_ACSRights_.put(_right.getAcsName(),  _right);
               
        else if (m_ACSRights_.get(_right.getAcsName()) instanceof StringRight) {
          StringRight previous = (StringRight) m_ACSRights_.get(_right.getAcsName());
          Set l_rights = new HashSet(INITIAL_CAPACITY);
          l_rights.add(previous);
          l_rights.add(_right);
          m_ACSRights_.put(_right.getAcsName(),  l_rights ); 
          previous = null;
          l_rights = null;
        } 
        
        else {
          Set l_previous = (Set) m_ACSRights_.get(_right.getAcsName());
          l_previous.add(_right);
          l_previous = null;
        } // else
      
                        // for testing and tracking
                        //if (sRight_.getRight().indexOf("do_action_four") != -1) {
                          //System.out.println("in StringMetaRightImpl.addACSRight(), for "+sRight_.getRight()+", m_ACSRights_ is..."  );  
                          //if (m_ACSRights_ != null && _right != null && m_ACSRights_.get(_right.getAcsName()) instanceof Set) {
                            //  System.out.println("a Set for "+_right.getAcsName()+" which contains:"  );
                              //int index = 0;
                              //for (Iterator ite = ((Set)m_ACSRights_.get(_right.getAcsName())).iterator(); ite.hasNext();) {
                                //  System.out.println("------ at "+index+": "+ite.next()  );
                                  //index++;
                              //}
                          //} // if
                          //else if (m_ACSRights_ != null && _right != null && m_ACSRights_.get(_right.getAcsName()) instanceof StringRight)
                            //  System.out.println("a single right for "+_right.getAcsName()+", which is:"+_right  );
                        //} // if

        try {
              firePropertyChange("ACSRights", old, getM_ACSRights());
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
        }       
        old = null;     
      
        return;
  } // addACSRight()
  
  /**
   * Remove an ACS right which is associated to this metaright.
   * Called by the ACS right finalizeForUser() method and ACSImpl.closeAcsRights().
   * It is a bound property.
   * @param _right directly associates this right to it. No action if null or
   *     unknown.
   */
  protected void removeACSRight(StringRightImpl _right) {
      
                // for testing
                //System.out.println("call of StringMetaRightImpl.removeACSRight for the right:  "+
                  //      _right.getRight()+" for the acsrights map with the keys: ");
                //for (Iterator ite = m_ACSRights_.keySet().iterator(); ite.hasNext(); ) {
                 //   System.out.println("----------- "+ite.next());
                //} // for ite
                //if(_right != null)
                  //  throw new InternalError("from StringMetaRightImpl.removeACSRight");
      
    if (_right == null) return;
    Map old = getM_ACSRights(); // then, m_ACSRights_ is never null

    if (m_ACSRights_.get(_right.getAcsName()) instanceof StringRight) {
            // for testing
            //System.out.println("in StringMetaRightImpl.removeACSRight, one single removing of "+
              //      _right.getAcsName());

      m_ACSRights_.remove(_right.getAcsName());
    } else { // the value in m_ACSRights_ is a Set
      Set l_previous = (Set) m_ACSRights_.get(_right.getAcsName());
      if (l_previous == null) {
          m_ACSRights_.remove(_right.getAcsName());
          return;
      }
      l_previous.remove(_right);
      if (l_previous.size() == 1) {
          m_ACSRights_.put(_right.getAcsName(), (StringRight)l_previous.iterator().next());
      }
      l_previous = null;
    } // else    

    try {
          firePropertyChange("ACSRights", old, getM_ACSRights());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
        e.printStackTrace();
    }       
    old = null; 

    return;      
  } // removeACSRight
  
  /**
   * Add a lower metaright which are strictly lesser than this metaright
   * without any intermediate right, following the network of metarights.
   * Called by the lower metaright addUpperRight() method. It is a bound property.
   * @param _right is directly included in this right. Never null.
   */
  protected final void addLowerRight(StringMetaRight _right) {
      sRight_.addLowerRight(_right);
      return;
  } // addLowerRight()
  
  /**
   * Remove a lower metaright which are strictly lesser than this metaright
   * without any intermediate right, following the network of metarights.
   * Called by the lower metaright removeUpperRight() method and by 
   * the lower metaright finalizeForUser() method. It is a bound property.
   * @param _right is directly included in this right. Never null.
   */
  protected void removeLowerRight(StringMetaRight _right) {
      sRight_.removeLowerRight(_right);
      return;      
  } // removeLowerRight

  /**
  * Fire an event to every registered listener, in any order.
  * For the properties: 'ACSRights'.
  * Called by addACSRight() and removeACSRight().
  * Used as the same method in java.beans.PropertyChangeSupport, which is not
  * used because hasListeners() has a bug in JDK 1.3.0-C of SUN.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected final void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

      if (_oldValue != null && _newValue != null && _oldValue.equals(_newValue)) {
          return;
      }

      // no action if the transiant variables are null, after a deserialization
      if (l_ACSRightsChangeSupport_ == null) {

            // for testing and tracking
            //System.out.println("in DirectoryImpl.firePropertyChange(), "+
            //  "no action since l_ChildrenChangeSupport_ is null.");

        return;
      }

        // for testing and tracking
        //System.out.println("in StringMetaRightImpl.firePropertyChange(), call "+
        //  "in "+getName()+" for the property "+_propertyName);

      PropertyChangeEvent     evt = new PropertyChangeEvent(this,
                                          _propertyName, _oldValue, _newValue);

      PropertyChangeListener  listener;
      Set                     toFire = null;

      if (_propertyName.equals("ACSRights"))
        toFire = (Set) ((HashSet)l_ACSRightsChangeSupport_).clone(); // no synchro problem

      for (Iterator ite = toFire.iterator() ; ite.hasNext() ; ) {
        listener = (PropertyChangeListener)ite.next();
        listener.propertyChange(evt);

        // for testing and tracking
        //System.out.println("in DirectoryImpl.firePropertyChange(), call to "+
        //  listener+".propertyChange() from "+getName()+"for the property"
        //  +_propertyName);

      } // for

      evt = null;
      listener = null;
      toFire = null;
      return;
  } // firePropertyChange() 
  
   /**
   * This method is to use only for closed rights in the relevant ACS list, to
   * set an empty instance.
   * It cuts all the Object External References, without updating
   * these objects as finalizeForUser() does, and finalize the embedded StringRightImpl. 
   * It set to null all variables, so the instance becames empty.
   * It is called by ACSImpl.checkClosedAclEntriesTo().
   * Call StringRightImpl.finalizeClosedRight().
   */
  protected final synchronized void finalizeClosedRight()  {

                //System.out.println("In StringMetaRightImpl, call of " + name_ + ".finalizeClosedRight()");

    if (sRight_ == null) return;
    
    sRight_.finalizeClosedRight();

    if (m_ACSRights_ != null) m_ACSRights_.clear();  
    
    power_ = 0;
    field_ = 0;
    return;
  } // finalizeClosedRight()  
  
} // StringMetaRightImpl