//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;


import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.*;

import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.util.Arrays;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeSupport;

 /**
  * This class is responsible for managing a right. A right is defined 
  * without connection to the user right and the access target. This
  * connection is the role of the right categories as AGO rights or ACL rights for instance.
  * There are two types of rights, the canonical metarights, outside any ACS,
  * and the ACS rights:<P>
  * - this class manages the ACS rights, registered in their ACSRun, <P>
  * - this class manages also the embedded right including in a generic or metaright. <P>
  * Each ACS right has a metaright as equivallent which is a StringMetaRightImpl,
  * subclass of StringRightImpl.
  * The generic right is used to compare rights from different ACS, and
  * to explain precisely the ACS right semantic. <P>
  * The two types of right may be classed, that is associated to a target class,
  * or ACS right only may be typed, that is associated to a type in a target class. 
  * If an ACS right is classed, it is also the generic right class. 
  * If it is typed, it is always classed.
  * For a generic right, the class isnot added at the end of the kernel right.
  * For an ACS right, the class and the type are added at the end of the kernel right.<P>
  * For instance, in LCA Lotus Notes, 'read_javascript' is a typed right
  * for the target class Actor, where 'javascript' is a type of Actor.  
  * When a right is classed or typed, the class and the type may be superior
  * to the right container (AGO, ACL, PRI, BDG) specification, or may be
  * inferior, and this is specified by the ACS policy. For instance,
  * if an ACL has an actor as target, and if it uses a classed right where
  * the target class is Directory, then the ACL will be effective only if
  * its ACS sets that the ACL specification is superior to the right class (Directory, there).
  * The ACS may also set that both specifications are applicable if they are
  * compatible, and if not, they are forgotten. <P>
  * There are several levels of right handling in the program: <P>
  * - the generic or ACS right contains the semantic of a granting or a denying, <P>
  * - the ACS defines the set of authorized generic/acs rights for each right category,
  *     plus some subsets managed in the specialized rights, <P>
  * - the ACS right category like AGO rights, ACL rights,... handle a form of connection
  *     between a right user, a generic/acs right, and an access target, <P>
  * - the right realization is the expression of a right category through the
  *     filter of ACS rules and inter-ACS applicable adaptors, <P>
  * - the AccessControlLink is the set of all the right realizations between
  *     two nodes of an access path in an information system, <P>
  * - the access path is a collection of AccessControlLinks from an user right 
  *     to an access target, and it is handles by a DisplayableLinkImpl, <P>
  * - the effective rights of an access path are the expression of the right categories
  *     through the filter of ACS rules and inter-ACS applicable adaptors. <P>
  * The right handles the links to the upper and lower rights 
  * in a rights network, allowing to compare them. Each ACS has a rights network,
  * somtimes a very simple one. There is also one metarights network.
  * The next upper and lower rights are managed, including
  * the synchronization from the lower rights when the link is updated.
  * So, the upper rights are updated when they are unregistered
  * in the lower rights. There are at most 5 upper rights.<P>
  * A two-levels right (in Windows for instance) is managed as one parent right which
  * has a detailled description in a set of lower rights.
  * The parent is the unique upper right of these slave rights.
  * Multi-level rights is a generalization where a right is both
  * a parent and a slave. Some rights set may have these tree structure
  * (it is not a network) without be marked as multi-level rights. <P>
  * A right may be desactivated, and then it is not used by the program in the search
  * of access paths. <P>
  * It is possible to use a quite empty right  when only the getRight() method 
  * is needed: use then the constructor with a null ACS, 
  * the right and an empty metaright, and use finalizeForUser() at the end.
 * A right may be conditional, and it is then used in an AclEntry to activate
 * it when the condition criterion is fulfilled. The criterion is to be membership to
 * one or several groups. These groups are defined in the kernel right with
 * the separator &lt;IF&gt; before each last component of the group name, for
 * instance in the kernel right 'read&lt;IF&gt;admin' where 'admin' is the name of a group.
 * Caution: the meta_ property is transient and should be restored after each
 * deserialization of an instance.
 * It is a javabean with the following bound properties:
 * 'Comment', 'Parent', 'Slave', 'Desactivated', 'UpperRights', 'LowerRights'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup).
 * They are called in any order. They receive only a copy of the new value,
 * to protect the property. All the exceptions from the listeners are catched
 * in the 'set' method, and a dialog box is displayed to inform the user.
  * @version 0.7.0
  */
public final class StringRightImpl implements StringRight {

  /*
  * private variables
  */
  private static final long serialVersionUID = 0;
  protected static final int  INITIAL_CAPACITY = 5; // the default value: 101 for map  

  /**
   * The name of the ACSRun which manages this right. Never null for an ACS right.
   * For an embedded right in a StringMetaRightImpl, it is always null.
   */
  private ImmutableName acsName_;

  /**
   * The ACSFactory which creates this right when it is an acsright. Used
   * only in the constructor, clone() and finalizeForUser().
   */
  private transient ACSRun acsRun_ = null;
  
  /**
   * The effective right, for typed and untyped rights. The default value is '<EMPTY_RIGHT>'.
   */
  private String right_ = "<EMPTY_RIGHT>";
  
  /**
   * The right key, to compare it in the network rights. 
   * Two StringRights may have the same key.
   * Key values are from 1 to 999 for StringRightImpl,
   * and from 1 to 99999 for StringMetaRightImpl.
   * Metarights with the same power have keys in a space of 100.
   * This key is used in gWork and for the NoThanViewInBases, to check the fullfilment
   * to the criteria.
   */
  protected int internalKey_;
  
  /**
   * If it is a classed right, this is the target class for the target of the right.
   * Exemple: in the right 'execute_script', 'script' is a
   * targetClassType_ for the targetClass_ Actor in the relevant ACS.
   * If non-classed right, targetClass_ is null. Otherwise, it is non-null.
   */
  private Class targetClass_ = null;
  
  /**
   * If it is a typed right, this is the target type for which this right is applied,
   * only for the target class defined by targetClass_. This value is at the end of
   * the effective right. Exemple: in the right 'execute script', 'script' is a
   * targetClassType_ for the targetClass_ Actor in the relevant ACS.
   * If non-typed right, targetClassType_ is null. Otherwise, it is non-null.
   */
  private String targetClassType_ = null;
  
  /**
   * The user comment about this right.
   */
  private String comment_ = "";
  
  /**
   * This is the list of the ACS rights or metarights (if embedded) which are directly more powerfull,
   * than this instance in the rights network. May be null.
   */
  protected StringRight[] l_upperRights_ = null;

  /**
   * This is the list of the ACS rights or metarights (if embedded) which are directly less powerfull,
   * than this instance in the rights network. May be null.
   */
  protected StringRight[] l_lowerRights_;
  
  /**
   * For ACS rights, this is the equivalent to this instance in the metarights space.
   * It is used to compare rights from different ACS. At the ACS saving, this property
   * is transient to avoid a direct link to the metarights network, and the
   * property metaRight_ is used to restore this instance meta_ value when
   * the ACS is opened or when the NoThanViewInBase is opened.
   * It is not copied in clone().
   * For an embedded right in a StringMetaRightImpl, it is always null.
   */
  transient StringMetaRight meta_ = null;
  
  /**
   * For ACS rights, this is the value returned by meta_.getRight().
   */
  String metaRight_ = null;
  
  /**
   * True when the right is set as a parent.
   */
  private boolean isParent_ = false;
  
  /**
   * True when the right is set as a slave, with only one upper right (the parent).
   */
  private boolean isSlave_ = false;  
  
  /**
   * True when the right is desactivated. It is then used only as a node
   * in a right network. It is no more used to deliver rights, but it remains
   * usefull in the search of child rights in StringRightUtilityImpl.
   */
  private boolean isDesactivated_ = false; 


   /**
   * manages all the property change listeners of this classes and its subclasses
   */
  protected transient PropertyChangeSupport changeSupport_ = null;
  /**
   * list of PropertyChangeListeners for the property 'Comment'.
   * No use of SwingChangePropertySupport class (bug in hasListeners() method).
   */
  private transient Set l_CommentChangeSupport_=
       new HashSet(INITIAL_CAPACITY);  

  /**
   * list of PropertyChangeListeners for the property 'Parent'.
   * No use of SwingChangePropertySupport class (bug in hasListeners() method).
   */
  private transient Set l_ParentChangeSupport_=
       new HashSet(INITIAL_CAPACITY); 

  /**
   * list of PropertyChangeListeners for the property 'Slave'.
   * No use of SwingChangePropertySupport class (bug in hasListeners() method).
   */
  private transient Set l_SlaveChangeSupport_=
       new HashSet(INITIAL_CAPACITY); 

  /**
   * list of PropertyChangeListeners for the property 'Desactivated'. 'UpperRights', 'LowerRights'
   * No use of SwingChangePropertySupport class (bug in hasListeners() method).
   */
  private transient Set l_DesactivatedChangeSupport_=
       new HashSet(INITIAL_CAPACITY); 

  /**
   * list of PropertyChangeListeners for the property 'UpperRights'.
   * No use of SwingChangePropertySupport class (bug in hasListeners() method).
   */
  private transient Set l_UpperRightsChangeSupport_=
       new HashSet(INITIAL_CAPACITY); 

  /**
   * list of PropertyChangeListeners for the property 'LowerRights'.
   * No use of SwingChangePropertySupport class (bug in hasListeners() method).
   */
  private transient Set l_LowerRightsChangeSupport_=
       new HashSet(INITIAL_CAPACITY); 
  
  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for temporary values.
   * Caution: never change the state of the returned instance.
   */
  public static final StringRightImpl EMPTY_INSTANCE = new StringRightImpl();

  /** standardized variable for the JDK 2 serialization
  * @serial ("acsName_", NameImpl.class)
  *  ("right_", String.class),
   * ("internalKey_", Integer.class),
   * ("targetClass_", Class.class),
   * ("targetClassType_", String.class),
   * ("comment_",String.class),
   * ("l_upperRights_", StringRight[].class),
   * ("l_lowerRights_", StringRight[].class),
   * ("metaRight_", String.class),
   * ("isParent_", Boolean.class),
   * ("isSlave_", Boolean.class),
   * ("isDesactivated_", Boolean.class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("acsName_", NameImpl.class),
      new ObjectStreamField("right_", String.class),
      new ObjectStreamField("internalKey_", Integer.class),
      new ObjectStreamField("targetClass_", Class.class),
      new ObjectStreamField("targetClassType_", String.class),
      new ObjectStreamField("comment_",String.class),      
      new ObjectStreamField("l_upperRights_", StringRight[].class),
      new ObjectStreamField("l_lowerRights_", StringRight[].class),
      new ObjectStreamField("metaRight_", String.class),
      new ObjectStreamField("isParent_", Boolean.class),
      new ObjectStreamField("isSlave_", Boolean.class),
      new ObjectStreamField("isDesactivated_", Boolean.class)
  };
  
  /**
   * Constructor used by EMPTY_INSTANCE.
  */
  public StringRightImpl() {
  }

  /**
   * Constructor used for the creation of an embedded right 
   * in a StringMetaRightImpl instance. <P>
   * There is no links between the most upper metarights dedicated to the rights administration,
   * like '|is_superadmin|', and the lower positive metarights headed by 'full_control'
   * and which have opposite negative rights (like 'deny_all' for 'full_control').
   * So a subtraction of any lower right to one administration right
   * does not change the administration right. These administration metarights have
   * a strenght superior to 90000, which is the 'full_control' strenght.<P>
   * Used by BaseUtilityImpl.initializeBasicMetaRights().
   * @param _kernelRight is the right if the StringRight. Never null.
   * @param _internalKey is the internal key which allows ACS right comparisons.
   *    From 1000 to 99999. '1000' is the lesser value. Never null.
   * @param _targetClass is the target interface (in gBaseInterface) 
   *    for the ACS or embedded right; if not null, this StringRight is classed.
   *    Do not use the 'Immutable' version of the interface as _targetClass.
   * @exception CreateError if the target class is not a gBase interface or if
   *   the kernel right or the internal key is not correct,
   *   the instance has then to be deleted!
   */
  protected StringRightImpl(String _kernelRight, Integer _internalKey,
    String _targetClass) throws CreateError {
    
        // common tests for embedded and ACS rights
    if (_kernelRight == null || _kernelRight.trim().length() == 0 ||
            _kernelRight.indexOf("-") != -1 ||
            _kernelRight.indexOf("<") != -1 ||
            _kernelRight.indexOf(">") != -1)
        throw new CreateError(//"in StringRightImpl constructor, "+
          " The right is not null or contains one of the characters '-', '<', '>'.");
    if (_targetClass != null && _targetClass.indexOf("ARoad0.gBaseInterface.") == -1) {
        throw new CreateError(" The target class "+
          "is not an interface of the Access Road database.");
    } // if ARoad0.gBaseInterface.
    if (_internalKey.intValue() > 99999)
            throw new CreateError(" The internal key "+
              "is not correct for a metaright for "+_kernelRight);
    
    // for embedded and ACS rights
    right_ = _kernelRight.trim();
    if (_targetClass != null) {
        try {
            targetClass_ = Class.forName(_targetClass); 
        } catch (ClassNotFoundException c) {
            throw new CreateError("In the StringRight constructor, the target class "+
              "has not a correct name "+c);        
        }
    } // if
    
    internalKey_ = _internalKey.intValue();          
  } // StringRightImpl()
  
  /**
   * Constructor used for the creation of an ACS right.
   * It is designed to facilitate the creation of a whole network rights, down from
   * the most powerfull rights. <P>
   * Registers this instance in its ACS and its metaright.
   * An upper right cannot be an administrative right (that is, its metaright
   * has a strenght superior to 90000) if this instance is not an administrative right.
   * The argument _targetClassType is not null only when the metaright is classed,
   * and the type should be ampong the class of the metaright. But this is
   * not controlled by the constructor, and if it is not true, the right will
   * simply not be used in the right analysis.
   * The _acsRun argument delivers both the name of the ACSFactoryImpl which
   * creates the new ACS, and the name of the created ACS.
   * Used by the GUI to create the ACS rights for an ACSFactoryImpl instance. 
   * @param _kernelRight is the right if the StringRight if not typed,
   *     or otherwise it is the start of the right. Never null and do not contain
   *     the character '-'. Contain one or several '&lt;IF&gt;' to have
   *     a conditional right.
   * @param _meta is the StringMetaRight which is equivallent to this StringRight.
   *    Never null, empty or desactivated.
   * @param _internalKey is the internal key which allows ACS right comparisons.
   *    From 1 to 999. '1' is the lesser value. Never null.
   * @param _l_upperRights is the set of upper rights which are strictly greater than this right
   *    without any intermediate rights, following the network of rights in the same ACS. 
   *    The number of upper rights is from 0 to 5 (including).  May be null.
   * @param _targetClassType is the type which is the end of the right.
   *    To use only when the _meta target class is not null. May be null.
   *    Is added after '_' at the end of the kernel right, for having the displayed right.
   * @param _acsRun is the ACSFactory owning this right.
   *    Never null. 
   * @exception CreateError if ACSFactoryImpl.addRight() throws it, or if
   *   the kernel right, the metaright or the internal key is not correct,
   *   an upper right is not superior to this instance;
   *   an upper right has a metaright which is inferior to _meta;
   *   an upper right is an administration right while the new right is not;
   *   the instance has then to be deleted!
   */
  public StringRightImpl(String _kernelRight,
    StringMetaRight _meta, Integer _internalKey, StringRightImpl[] _l_upperRights,
    String _targetClassType, ACSRun _acsRun) throws CreateError {
    
    if (_kernelRight == null || _kernelRight.trim().length() == 0 ||
            _kernelRight.indexOf("-") != -1)
        throw new CreateError(//"in StringRightImpl constructor, "+
          " The right "+_kernelRight+" is not null or contains the character "+
          "'-'.");
    if (_acsRun == null )
        throw new CreateError(" The ACSFactory cannot be null.");    
    if (_meta == null || _meta.isDesactivated() || _meta.isEmpty())
        throw new CreateError(" The metaright "+
          "cannot be null, desactivated or empty for an ACS right: "+_kernelRight);
 
    if (_internalKey.intValue() < 1 || _internalKey.intValue() > 999)
        throw new CreateError(" The internal key "+_internalKey+
          " is not correct for an ACS right.");
    if (_l_upperRights != null && _l_upperRights.length > 5)
        throw new CreateError(" The number "+_l_upperRights.length+
          " of upper rights is superior to 5.");
    
    meta_ = _meta; // needed for using isGrantingRight() hereafter
    metaRight_ = _meta.getRight();
    
    if (_meta.getTargetClass() != null) {
        try {
            targetClass_ = Class.forName(_meta.getTargetClass()); 
        } catch (ClassNotFoundException c) {
            throw new CreateError(" In the StringRight constructor, the target class "+
              "has not a correct name "+c);        
        }
    } // if
   
    
                    // for testing and tracking
                    //System.out.println("StringRightImpl constructor for "+_kernelRight);
    
                    
    //   Test if an upper right is inferior to this instance, from the same ACS,
    //   and if an upper right has a metaright which is inferior to _meta; 
    int error = -1;
    int error2 = -1;
    int error3 = -1;    
    int error4 = -1;
    //int error5 = -1;

    if ( _l_upperRights != null ) {
        if (_acsRun != null ) { 
          for (int i = 0; i < _l_upperRights.length; i++) {
              if (_l_upperRights[i].internalKey_ <= _internalKey.intValue() ||
                      _l_upperRights[i].isGrantingRight() != isGrantingRight()) {
                  error = i;
                  break;
              }
              if (!_acsRun.getName().equals(_l_upperRights[i].getAcsName())) {
                  error2 = i;
                  break;
              }
              if (_l_upperRights[i].getEorMetaRight().compareTo(_meta) < 0) {
                  error3 = i;
                  break;
              }
              if (_meta.getInternalKey() <= 90000  && 
                      _l_upperRights[i].getEorMetaRight().getInternalKey() > 90000) {
                  error4 = i;
                  break;
              }
               /* 
              if (targetClass_ != null && _l_upperRights[i].isTargetClassed()) {
                  // _l_upperRights[i] may be not classed, but its non-null class
                  // must be a superclass
                  if (!_l_upperRights[i].targetClass_.isAssignableFrom(targetClass_) ) {
                                error5 = i;
                                break;
                  }
              } // if (_targetClass != null )   
               */           

          } // for i
        } // if (_acsName   
    } // if ( _l_upperRights

    if (error != -1)
        throw new CreateError(" The upper right at "+
          error + " is not superior to the new right or has not the same sens.");
    if (error2 != -1)
        throw new CreateError(" The upper right at "+
          error2 + " is not from the same ACS - "+_l_upperRights[error2].getAcsName()+
          " and "+_acsRun.getName());
    if (error3 != -1)
        throw new CreateError(" The upper right at "+
          error3 + " has a metaright which is inferior to the new right metaright.");  
    if (error4 != -1)
        throw new CreateError(" For '"+_kernelRight+"', the upper right at "+
          error4 + " is an administration right while the new right is not.");    
    /*
    if (error5 != -1)
        throw new CreateError("The upper right at "+
          error5 + " has a target class which is not a superclass of the new right target class."); 
    */   
    
    right_ = _kernelRight.trim();
    if (targetClass_ != null) {
        if (_targetClassType != null && _targetClassType.trim().length() != 0)
            targetClassType_ = _targetClassType.trim();
        
        // the kernel right is completed with the type but never the class,
        // right_ may be not unique among the ACS rights
        if (targetClassType_ != null)
            right_ = _kernelRight.trim() + "_" + targetClassType_;
        //else right_ = _kernelRight.trim()+ "_" + targetClass_.toString().substring(32); // remove 'interface ARoad0.gBaseInterface.'                
    } // if
    
    internalKey_ = _internalKey.intValue();  
    
    acsRun_ = _acsRun;
   
                        // for testing
                        //if (_kernelRight.equals("x"))
                          //  System.out.println("in StringRightImpl constructor of 'x', right_: "+right_);        
    
    acsName_ = (ImmutableName)_acsRun.getName();

    if (_l_upperRights != null && _l_upperRights.length != 0) {
      l_upperRights_ = new StringRight[_l_upperRights.length];
      System.arraycopy(_l_upperRights, 0, l_upperRights_, 0, _l_upperRights.length);  
      // declares this instance in the upper rights
      for (int i = 0; i < _l_upperRights.length; i++) {
          _l_upperRights[i].addLowerRight(this); // use equals()
      } // for
    } // if               

    // ACSpreFactoryImplBeanInfo indicates to BaseObjectHandler how to register
    // this instance in acsRun_ (fire 'OneRight'to the GUI). Then,
    // acsRun_ will register this instance in the new ACS.
    acsRun_.addRight(this);

    // registers this instance in the metaright
    ((StringMetaRightImpl)_meta).addACSRight(this);  
  } // StringRightImpl()


  // ------------------------ Methods of the interface ------------------
   
  /**
   * Called from the constructor through StringMetaRightImpl.addACSRight()
   * It is both the name of the ACS and then name of its ACSFactory.
   * This method implements the ACSObject without declaring this class as
   * implementing ACSObject, since the subclass StringMetaRight is never an ACSObject.
   * @return ImmutableName of the ACSRun that owns this object. 
   *    Empty if it is an embedded right.
   */
  public final ImmutableName getAcsName() {
      if (acsName_ != null) return (ImmutableName)acsName_.clone();
      else return NameImpl.EMPTY_INSTANCE;
  }

  /**
   * After the ACS creation, this method calls BaseManager to get the open ACS.
   * It also works during the creation of ACS rights by an ACSFactory.
   * So, BeamerManager can be informed when an ACS rights has been
   * deleted, since it calls this method to be registered as listener in the
   * acsright ACS or ACSRun.
   * @return ACS or ACSFactory creating the ACS, if this right is an acsright. 
   *     Null otherwise.
   */
  public final ACSRun getEorACS() {
      if (acsName_ == null) return null;
      ACSRun result = (ImmutableACS) 
              BaseManagerImpl.getInstance().getEorM_OpenACS().get(acsName_);
      if (result == null) result = acsRun_;// acsRun_ remains non-null after the ACS creation
      return result;
  }
  
  /**
 * Gets the full name of the right, which is unique for the program.
 * The name is composed of:
 * :: &lt;FOR_ACS&gt;:: ACS name or no_acs
 * plus, if there is a target class,
 * ::  &lt;AND&gt;::  target class, or Target class:: class type
 * &lt;VALUE&gt;::  right <P>
 * here, ':: ' is added by gBase.NameImpl.toString().
 * This name is used only outside gBase, following BaseObject interface.
 * @return the full name of the right, which is empty if it is an empty right
   */
  public final ImmutableName getFullName() {
      if (isEmpty()) return NameImpl.EMPTY_INSTANCE;
      
      Name name = new NameImpl("<FOR_ACS>");
      if (acsName_ != null) name.extended(acsName_);
      else name.extended("no_acs"); 
      if (targetClass_ != null) {
          name.extended("<AND>" );
          name.extended(targetClass_.toString().substring(32) ); // '32' to remove 'interface ARoad0.gBaseInterface.'
          if (targetClassType_ != null) name.extended(targetClassType_);
      }
      name.extended("<VALUE>");
      name.extended(right_ );

      return name;
  } // getFullName()
  
  /**
 * Gets the BaseObject name of the right, as a current name with less information
 * than the full name. The name is composed of: 
 * :: &lt;FOR_ACS&gt;:: ACS name or no_acs
 * &lt;VALUE&gt;::  right <P>
 * This name is used only outside gBase, following BaseObject interface.
 * @return ImmutableName of the right, which is empty if it is an empty right
 * @see #getFullName
 * @see BaseUtilityImpl#getRightFromNameAsString
   */
  public final ImmutableName getName() {
      if (isEmpty()) return NameImpl.EMPTY_INSTANCE;
      
      Name name = new NameImpl("<FOR_ACS>");
      if (acsName_ != null) name.extended(acsName_);
      else name.extended("no_acs");

      name.extended( "<VALUE>" );
      name.extended(right_ ); 
      
      return name;
  } // getName()
    
  /**
   * Gets a short name limited to 5 characters if positive, 14 first characters if
   * it starts with '|' and positive, and 7 if negative, for displaying this right
   * in the views. <P>
   * The exception is for the administrative rights where the 17 last characters
   * are used from the nick name. <P>
   * If right_ is '|modify_security_descriptor|', the nick name is '|modify_security|'. <P>
   * If right_ is 'read', the nick name is the same.
   * If right_ is 'read&execute', the nick name is 'read&'.
   * If right_ contains '_', the nick name 2 last characters are '_%' where '%'
   * is here the character at the right of the last '_' in right_. For example,
   * 'write_for_directory' gives 'wri_d'.
   * The nick name for a negative right starts with 'd_', if the right
   * starts by 'deny_'. The other parts of the nick name are like for a positive right.
   * If right is 'deny_read', the nick name is 'd_read'.
   * If the right is 'deny_write_for_directory', the nick name is 'd_wri_d'.
   * If the negative right does not start with 'deny_', it is processed like
   * a positive right and the nick name size is limited to 5.
   * This short name is NOT unique for the Access Road program, nor for the ACS if
   * it is an ACS object.
   * @return the nick name. Never null.
   */
    public final String getNickName() {
        if (isGrantingRight() || !right_.startsWith("deny_")) {
           if (right_.length() < 6 && !right_.startsWith("|")) {
               return right_;
           }
           else if (right_.startsWith("|")) {
               if (right_.length() < 18) return right_;
               //else if (right_.length() > 25)
                 //      return "|"+right_.substring(right_.length() - 17);
               else return "|"+right_.substring(right_.length() - 17);
           }
           else if (right_.lastIndexOf("_") > 3 && 
                   right_.lastIndexOf("_")+1 < right_.length()) {
               return right_.substring(0, 3) + right_.substring(
                       right_.lastIndexOf("_"), right_.lastIndexOf("_")+2);
           }
           else {
               return right_.substring(0, 5);
           } // else
        } // if (isGrantingRight()
        else {
           if (right_.length() < 10) {
               return "d_"+right_.substring(5); 
           }
           else if (right_.lastIndexOf("_") > 5 && right_.lastIndexOf("_")+1 < right_.length()) {
               if (8 < right_.lastIndexOf("_")) {
                   return "d_"+right_.substring(5, 8) + 
                       right_.substring(
                          right_.lastIndexOf("_"), right_.lastIndexOf("_")+2);
               }
               else {
                   return "d_"+right_.substring(5, right_.lastIndexOf("_")) +
                       right_.substring(
                          right_.lastIndexOf("_"), right_.lastIndexOf("_")+2);
               } // else
           } 
           else {
               return "d_"+right_.substring(5, 10);
           }
        } // else
    }  // getNickName()
    
  /**
   * @return the internal key
   */
    public final int getInternalKey() {
       return internalKey_;
    }
    
  /**
   * @return true if the right is always associated to a class for a target,
   *   as for exemple 'delete_Actor' for the class Actor.
   */
   public final boolean isTargetClassed() {
       if (targetClass_ != null) return true;
       else return false;
   }
    
  /**
   * @return true if the right is always associated to a type for a target
   *   class, as for exemple in LCA Lotus Notes 'read_javascript' for the class Actor.
   */
   public final boolean isTargetClassTyped() {
       if (targetClassType_ != null) return true;
       else return false;
   }

  /**
  * If the right is not typed, the result contains the kernel right only.
  * If the right is typed (for ACS rights only), the result contains the kernel right,
  * then '_' and the target class type.  
  * @return the right to apply
 */
  public final String getRight() {
      return right_;
  }

    /**
    * @return the target class for which the right is provided.
    *   Null if the right is not classed. 
    */
  public final String getTargetClass() {
      if (targetClass_ == null) return null;
      else {
          //System.out.println("in StringRightImpl.getTargetClass(), targetClass_.toString(): "+
            //      targetClass_.toString());
          
          return targetClass_.toString().substring(10); // substring to remove 'interface '
      } // else
  }

    /**
    * @return the target type. Null if the right is not typed.
    */
  public final String getTargetClassType() {
      return targetClassType_;
  }

    /**
    * Sets any comment related to this right. It is a bound property.
    * @param _s is an user comment which is associated to the right.
    */
   public synchronized void setComment(String _s)	{
        String oldComment = comment_;
        comment_ = _s;

        try {
              firePropertyChange("Comment",oldComment,comment_);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
        }
        oldComment = null;
        return;
   } // setComment

    /**
     * @return the comment which is associated to the right. May be null.
     */
   public final String getComment()	{
        return comment_;
   }  
  
  /**
   * Called from the constructor for ACS rights, by getNickName() and compareTo().
   * @return true if the right is a granting rights - the opposite one is
   *   a denying right.
   */
   public boolean isGrantingRight() {
       if (meta_ != null) return meta_.isGrantingRight();
       else return false; // meta_ is null only for embedded rights, so used only in this class
   }
  
  /**
   * @return true if the instance is equals to an empty instance
   */
  public final boolean isEmpty() {
      if (EMPTY_INSTANCE.equals(this)) return true;
      else return false;
  }
  
  /**
   * For two-levels rights, this right manages tasks, and its slaves manage operations.
   * @return true if the right is a parent of lower rights - the slaves -
   *   in a multi-level right.
   */
   public final boolean isParent() {
       return isParent_;
   }
    
  /**
   * A slave right has a strong link with a parent right.
   * For two-levels rights, this right manages operations, and its parent manages tasks.
   * @return true if the right has an unique upper right - the parent -
   *   in a multi-level right.
   */
   public final boolean isSlave() {
       return isSlave_;
   }
    
  /**
   * @return true if it is an acsright managed by an open ACS after its creation
   */
    public boolean isAcsRightUnderOpenACS() {

      if (acsName_ != null && 
              BaseManagerImpl.getInstance().getEorM_OpenACS().get(acsName_) != null) 
          return true;

      else return false;
    } // isAcsRightName()
   
  /**
   * Sets the right as parent of lower rights - the slaves -
   * in a multi-level right.
   * For two-levels rights, this right manages tasks, and its slaves manage operations.
   * This method has to be used before the registration of the lower rights.
   * Caution: some lower rights may be not slaves, so this instance may be not parent
   * for all its lower rights.
   * It is a bound property.
   * @param _b is true to set the right as parent
   * @exception BaseError when _b true if there are some lower rights or if it is after the
   *    ACS creation process.
   */
   public final void setParent(boolean _b) throws BaseError {       
      if (isAcsRightUnderOpenACS())
          throw new BaseError(" The parent flag cannot be set after the ACS creation.");
       if (_b && (l_lowerRights_ != null && l_lowerRights_.length != 0)) 
            throw new BaseError(" The slaves are already defined.");
      
       boolean old = isParent_;
       isParent_ = _b;
       
        try {
              firePropertyChange("Parent", old, isParent_);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
        } 
       return;
   } // setParent()
    
  /**
   * Sets that the right has an unique upper right - the parent,
   * already registered - in a multi-level right. 
   * For two-levels rights, this right manages operations, and its parent manages tasks.
   * It is a bound property.
   * @param _b is true to set the right as slave
   * @exception BaseError when _b true if there is not exactly one upper right,
   *    or if the upper right is not a parent, or if it is after the
   *    ACS creation process.
   */
   public final void setSlave(boolean _b) throws BaseError {
       
      if (isAcsRightUnderOpenACS())
          throw new BaseError(" The slave flag cannot be set after the ACS creation.");
       
      if (_b && (l_upperRights_ == null || l_upperRights_.length != 1 ||
         !l_upperRights_[0].isParent())) 
            throw new BaseError("in StringRightImpl.setSlave(), "+
              "the first upper right is not a parent.");
      
      boolean old = isSlave_;
      isSlave_ = _b;
       
      try {
              firePropertyChange("Slave", old, isSlave_);
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
      }        
      return;       
   } // setSlave()
    
  /**
   * Sets that the right is desactivated. Used to create a list of flat rights
   * with one single desactivated root. This method is called when the metaright
   * equivallent method is called.
   * It is a bound property.
   * @param _b is true to desactivate the right
   */
   public final void setDesactivated(boolean _b){
       boolean old = isDesactivated_;
       isDesactivated_ = _b;
       
        try {
              firePropertyChange("Desactivated", old, isDesactivated_);
        } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
            e.printStackTrace();
        }        
       return;       
   } // setDesactivated()
    
  /**
   * Get if the right is desactivated.
   */
   public final boolean isDesactivated() {
       return isDesactivated_;
   } // isDesactivated()
  
  /**
   * Gets the canonical metaright which is a generic description of this ACS right.
   * Metarights are used in the rights comparison and analysis to
   * compare rights from different ACS and in the same ACS.
   * This property is set to null by ACSImpl.closeAcsRights().
   * @return as a generic description of the right. Is null
   * only when the instance is embedded in a StringMetaRightImpl, or after
   * the closing of its ACS.
   */
  public final StringMetaRight getEorMetaRight() {
      return meta_;
  }
   
  /**
   * Gets the set of lower rights which are strictly lesser than this right
   * without any intermediate right, following the network of rights in the ACS.
   * For an embedded right, return an array of StringMetaRightImpl.
   * @return is an array of rights which are strictly lesser than 
   *    right. Never null but may be empty.
   */
  public final StringRight[] getLowerRights() {
      if (l_lowerRights_ == null) return new StringRight[0];
      StringRight[] l_rights = new StringRight[l_lowerRights_.length];
      System.arraycopy(l_lowerRights_, 0, l_rights, 0, l_lowerRights_.length);
      return l_rights;      
  }
  
  /**
   * Gets the set of upper rights which are strictly greater than this right
   * without any intermediate right, following the network of rights in the ACS.
   * For an embedded right, return an array of StringMetaRightImpl.
   * @return is an array of connected rights which are strictly greater than
   *    this right. Never null but may be empty.
   */
  public final StringRight[] getUpperRights() {
      if (l_upperRights_ == null) return new StringRight[0];
      StringRight[] l_rights = new StringRight[l_upperRights_.length];
      System.arraycopy(l_upperRights_, 0, l_rights, 0, l_upperRights_.length);
      return l_rights;
  }  
  
  /**
   * TODO: consistent with compareTo() and equals()
   * Get a clone of the instance.
   * Caution: the metaright meta_ and the comment are not copied, but metaRight_ 
   * is copied and the upper/lower rights references are copied in the arrays.
   * Used by ACS.closeAclEntriesTo() when the links to other ACS have to 
   * be cut and cloned.
   * @exception InternalError if CloneNotSupportedException
   * @return is a cloned instance of StringRightImpl.
   */
  /*
  public Object clone() {
    StringRightImpl clon = null;
    try {
      clon = (StringRightImpl) super.clone();

      clon.acsName_ = (ImmutableName)acsName_.clone();
      clon.right_ = right_;
      clon.internalKey_ = internalKey_;
      clon.targetClass_ = targetClass_;
      clon.targetClassType_ = targetClassType_;
      clon.isParent_ = isParent_;
      clon.isSlave_ = isSlave_;
      clon.isDesactivated_ = isDesactivated_; 
      if (l_upperRights_ != null) {
        clon.l_upperRights_ = new StringRight[l_upperRights_.length];
        System.arraycopy(l_upperRights_, 0, clon.l_upperRights_, 0, l_upperRights_.length);
      }
      if (l_lowerRights_ != null) {
        clon.l_lowerRights_ = new StringRight[l_lowerRights_.length];
        System.arraycopy(l_lowerRights_, 0, clon.l_lowerRights_, 0, l_lowerRights_.length);
      }
      clon.metaRight_ = metaRight_;
      
    } catch (CloneNotSupportedException e) {
      // this shouldn't happen, since we are Cloneable
      throw new InternalError(e.toString() + "StringRightImpl.clone() failed");
    } // catch
    return clon;
  } // clone()
  */
  
    /**
     * Compares the two StringRights, without comparison of the properties
     * l_upperRights_ and l_lowerRights_.
     * @param _obj to compare
     * @return true if the properties acsName_, right_ are equals and 
     *      internalKey_ and meta_ are the same
     */
    public final synchronized boolean equals(Object _obj) {

                // for testing and tracking
                //System.out.println("StringRightImpl.equals() on "+right_);
        
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      StringRightImpl obj = (StringRightImpl)_obj;

      if (acsName_ == null || obj.acsName_ == null) {
        if (acsName_ != null || obj.acsName_ != null ) {
                // for testing and tracking
                //System.out.println("StringRightImpl.equals(): acsName_ are not all null ");  
            return false;
        } // if
      } // if      
      else if (!acsName_.equals(obj.acsName_) ) {
                // for testing and tracking
                //System.out.println("StringRightImpl.equals(): acsName_ are not the same ");            
          return false;
      }
 
      if (right_ == null || obj.right_ == null) {
        if (right_ != null || obj.right_ != null )
          return false;
      } // if      
      else if (!right_.equals( obj.right_ ) ) {
                // for testing and tracking
                //System.out.println("StringRightImpl.equals(): right_ are not the same; ");            
                //System.out.println("StringRightImpl.equals(): right_; |"+right_+"|");
                //System.out.println("StringRightImpl.equals(): obj.right_; |"+obj.right_+"|");
          return false;
      }
      
      if (internalKey_ != obj.internalKey_ ) {
                // for testing and tracking
                //System.out.println("StringRightImpl.equals(): internalKey_ are not the same ");  
          return false;
      }

              // for testing and tracking
              //System.out.println("In StringRightImpl.equals() after internalKey_");

      if (targetClass_ == null || obj.targetClass_ == null) {
        if (targetClass_ != null || obj.targetClass_ != null )
          return false;
      } // if      
      else if (!targetClass_.equals( obj.targetClass_ ) ) return false;

      if (targetClassType_ == null || obj.targetClassType_ == null) {
        if (targetClassType_ != null || obj.targetClassType_ != null )
          return false;
      } // if      
      else if (!targetClassType_.equals( obj.targetClassType_ ) ) return false;

      if (isParent_ != obj.isParent_ ) return false;
      
      if (isSlave_ != obj.isSlave_ ) return false;
      
      if (isDesactivated_ != obj.isDesactivated_ ) return false;

      // for testing and tracking
      //System.out.println("In StringRightImpl.equals() after isDesactivated_");
      /*
      boolean found_1;
      boolean found_2;
      if (l_upperRights_ == null || obj.l_upperRights_ == null) {
        if (l_upperRights_ != null || obj.l_upperRights_ != null )
          return false;
      } // if
      else if (l_upperRights_ != null) {
          if (l_upperRights_.length != obj.l_upperRights_.length)  // no test for lenght == 0
            return false;
          found_1 = false;
          found_2 = true;
          for (int i = 0; i < l_upperRights_.length; i++) {
              for (int j = 0; j < obj.l_upperRights_.length; j++) {
                  if (l_upperRights_[i].getAcsName().equals( obj.l_upperRights_[j].getAcsName() ) &&
                          l_upperRights_[i].getName().equals( obj.l_upperRights_[j].getName() ) &&
                          l_upperRights_[i].compareTo( obj.l_upperRights_[j] ) == 0 ) {
                    found_1 = true;
                    break;
                  } // if
              } // for j
              if (!found_1) {
                  found_2 = false;
                  break;
              }
              else found_1 = false;
          } // for i
          if (!found_2) return false;
      } // else

              // for testing and tracking
              //System.out.println("In StringRightImpl.equals() after l_upperRights_; ");
              //System.out.println("l_lowerRights_.length: "+l_lowerRights_.length);
              //System.out.println("obj.l_lowerRights_.length: "+obj.l_lowerRights_.length);

      if (l_lowerRights_ == null || obj.l_lowerRights_ == null) {
        if (l_lowerRights_ != null || obj.l_lowerRights_ != null )
          return false;
      } // if
      else if (l_lowerRights_ != null) {
          if (l_lowerRights_.length != obj.l_lowerRights_.length)  // no test for lenght == 0
            return false;
          found_1 = false;
          found_2 = true;
          for (int i = 0; i < l_lowerRights_.length; i++) {
              for (int j = 0; j < obj.l_lowerRights_.length; j++) {
                  if (l_lowerRights_[i].getAcsName().equals( obj.l_lowerRights_[j].getAcsName() ) &&
                          l_lowerRights_[i].getName().equals( obj.l_lowerRights_[j].getName() ) &&
                          l_lowerRights_[i].compareTo( obj.l_lowerRights_[j] ) == 0) {
                    found_1 = true;
                    break;
                  } // if
              } // for j
              if (!found_1) {
                  found_2 = false;
                  break;
              }
              else found_1 = false;
          } // for i
          if (!found_2) return false;
      } // else
       * */

      // for testing and tracking
      //System.out.println("In StringRightImpl.equals()");
      
      if (meta_ == null || obj.meta_ == null) {
            if (meta_ != null || obj.meta_ != null ) {
                    // for testing and tracking
                    //System.out.println("StringRightImpl.equals(): only one meta_ is null ");
                    //System.out.println("StringRightImpl.equals(): right_; |"+right_+"|");
                    //System.out.println("StringRightImpl.equals(): obj.right_; |"+obj.right_+"|");
                    //System.out.println("StringRightImpl.equals(): meta_; |"+meta_+"|");
                    //System.out.println("StringRightImpl.equals(): obj.meta_; |"+obj.meta_+"|");

                return false;
            } // if
      } // if      
      else if (!meta_.equals( obj.meta_ ) ) {
                // for testing and tracking
                //System.out.println("StringRightImpl.equals(): meta_ are not equal ");           
            return false;
      }
      
                  // for testing and tracking
                  //System.out.println("StringRightImpl.equals(): true for "+right_);  
                  
      return true;
  } // equals(Object)

    /**
     * @return hashcode value for this instance.
     */
  public final int hashCode() {
    int result = 7;
    if (acsName_ != null) {
      result = result*1000003 + acsName_.hashCode();
    } // if
    if (right_ != null) {
      result = result*1000003 + right_.hashCode();
    } // if
    return result;
  } // hashcode()

    /**
     * Compares two StringRight instances through the internal key, the sens and the right,
     * in that order, and, if not null, through an acs names comparison, 
     * whatever the powers and the fields.
     * All metarights are superior to any acsright. The comparison 
     * 'xx'.compareTo('deny_xx') returns +1 since a positive right is superior to its
     * opposite.
     * Note: it is recommanded to always compare the metarights of two acsrights,
     * rather than the acsrights directly.
     * @param _o to compare
     * @exception ClassCastException - if the specified object's type
     *  prevents it from being compared to this current Object
     * @return 0 if the rights have the same internal key, sens, right and acs name,
     *  a negative integer if this object is lesser than the argument, 
     *  and a positive integer if this object is greater than the argument.
     */
  public final synchronized int compareTo(Object _o) throws ClassCastException {
     if (!(_o instanceof StringRight))
        throw new ClassCastException(
        "the argument of StringRightImpl.compareTo() is not a StringRight instance.");
     
     if (right_ == null) throw new ClassCastException("in StringRightImpl.compareTo(),"+
      " the object right is null");
     
     // init
     int result = 1; // for (internalKey_ > o.internalKey_)
     StringRightImpl o;
     if ((_o instanceof StringRightImpl))
         o = (StringRightImpl)_o;
     else o = ((StringMetaRightImpl)_o).sRight_;
     
     if (internalKey_ == o.internalKey_) {
         if (isGrantingRight() && !o.isGrantingRight()) result = +1;
         else if (!isGrantingRight() && o.isGrantingRight()) result = -1;
         else {
            if (right_.compareTo(o.right_) < 0) result = -1;
            else if (right_.compareTo(o.right_) == 0) {
                if (acsName_ == null && o.acsName_ == null) result = 0;
                 // metarights greater than acsrights
                else if (acsName_ == null && o.acsName_ != null) result = +1;
                else if (acsName_ != null && o.acsName_ == null) result = -1;
                else if (acsName_.compareTo(o.acsName_) < 0) result = -1;
                else if (acsName_.compareTo(o.acsName_) == 0) result = 0;
            }
         } // else
     }
     else if (internalKey_ < o.internalKey_) result = -1;

     return result;
  } // compareTo(Object)

  /**
   * @return long serial version UID for serialization of the class
   */
  public long getSerialVersionUID()	{
     return serialVersionUID;
  }

   /**
   * @return getRight()
   */
  public final String toString() {
    return getRight();
  } // toString()
    
   /**
   * This method removes the StringRight from the ACS list, from the metaright,
   * from the lower and upper rights. If the instance is a parent, all slaves
   * are also finalized. Calls finalizeForBase().
   * @exception UpDateError if the ACS does not known the right (if
   *   it is a clone, typically), if the metaright is wrong,
   *   or if a lower or upper right is not correct.
   */
  public void finalizeForUser() throws UpDateError {
      finalizeForBase();
  }
  
   /**
   * This method removes the StringRight from the ACS list, from the metaright,
   * from the lower and upper rights. If the instance is a parent, all slaves
   * are also finalized. This method may be called after an exception from
   * the constructor, to reinitialize the links to other objects.
   * This method is called by finalize(), but finalize() may be not called.
   * An unused object has to be deleted to inform the other objects, and this method
   * HAS TO BE CALLED before deleting the StringRight.
   * See finalizeClosedRight() for closing a deprecated right without updating the
   * external references.
   * @exception UpDateError if the ACS does not known the right (if
   *   it is a clone, typically), if the metaright is wrong,
   *   or if a lower or upper right is not correct.
   */
  public synchronized void finalizeForBase() throws UpDateError {

                //System.out.println("In StringRightImpl, call of " + name_ + ".finalizeForUser()");

    if (right_ == null) return;

    // if it is an ACS right.
    // otherwise, finalizeForUser() is called by StringMetaRightImpl.finalizeForUser()
    if (acsName_ != null) {
        ((StringMetaRightImpl)meta_).removeACSRight(this); // where sRight_ == null if the caller is meta_
        
                    // for testing
            //System.out.println("in StringRightImpl.finalizeForUser() for "+right_+
              //      ", after call of meta_.removeACSRight(this)");
        
        meta_ = null;
        metaRight_ = null;
        ACSImpl acs = (ACSImpl) BaseManagerImpl.getInstance().
          getEorM_OpenACS().get(acsName_);
        // acs is null for an acsright before the ACS creation by ACSFactory
        if (acs != null) acs.removeRight(this); // fire 'OneRight' event
        else acsRun_.removeRight(this); // fire 'OneRight' event
        acs = null;
        acsRun_ = null;
    }  // if
    
    StringRight[] l_rights;
    if (l_upperRights_ != null) {
      l_rights = new StringRight[l_upperRights_.length];
      System.arraycopy(l_upperRights_, 0, l_rights, 0, l_upperRights_.length);  
      // remove this instance in the upper rights
      for (int i = 0; i < l_rights.length; i++) {
          if (acsName_ != null)
              ((StringRightImpl)l_rights[i]).removeLowerRight(this);
          // else: processed in StringMetaRightImpl
      } // for
    } // if
    
    if (l_lowerRights_ != null) {
        
      l_rights = new StringRight[l_lowerRights_.length];
      System.arraycopy(l_lowerRights_, 0, l_rights, 0, l_lowerRights_.length); 
      if (!isParent_) {
          // remove this instance in the lower rights, without call back
          for (int i = 0; i < l_rights.length; i++) {
              if (acsName_ != null)
                  ((StringRightImpl)l_rights[i]).removeUpperRight(this, this);
              // else: processed in StringMetaRightImpl
          } // for
      } else {
          for (int i = 0; i < l_rights.length; i++) {
              // if parent, delete only the slaves, not all the lower rights
              if (l_rights[i].isSlave()) l_rights[i].finalizeForUser();
              // remove this instance in the lower rights
              else {
                  if (acsName_ != null)
                      ((StringRightImpl)l_rights[i]).removeUpperRight(this, this);
                  // else: processed in StringMetaRightImpl                  
              } // else
          } // for          
      } // else
      
    } // if (l_lowerRights_
      
    targetClassType_ = null;
    targetClass_ = null;
    right_ = null;
    internalKey_ = 0;
    l_rights = null;
    comment_ = null;
    l_lowerRights_ = null;
    l_upperRights_ = null;
    isParent_ = false;
    isSlave_ = false;
    acsName_ = null; 
  } // finalizeForBase()

  /**
  * Adds a listener to the bean, for all the bound properties.
  * Used by ViewInBaseImpl to detect the changes in the view rights.
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(PropertyChangeListener _l){
      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

                        // for testing and tracking
                        //System.out.println(
                          //      "StringRightImpl.addPropertyChangeListener(_l),"+
                            //    "call to "+right_+" for the property "+_propertyName+
                              //  " and the listener "+_l);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_l);}
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'Comment', 'Parent', 'Slave', 'Desactivated', 'LowerRights', 'UpperRights'.
  * @param _propertyName String is the name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (changeSupport_ == null)
          changeSupport_ = new PropertyChangeSupport(this);

                        // for testing and tracking
                        //System.out.println(
                          //      "StringRightImpl.addPropertyChangeListener(_propertyName, _l),"+
                            //    "call to "+right_+" for the property "+_propertyName+
                              //  " and the listener "+_l);

      synchronized (changeSupport_)
        {changeSupport_.addPropertyChangeListener(_propertyName, _l);}
  } // addPropertyChangeListener(String _propertyName,

  /**
  * Removes a listener to the bean, for all the bound properties.
  * @param _l is the PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                    // for testing and tracking
                  //  System.out.println("StringRightImpl.removePropertyChangeListener(_l), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_l);}
  } // removePropertyChangeListener()

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'Comment', 'Parent', 'Slave', 'Desactivated', 'LowerRights', 'UpperRights'.
  * @param _propertyName String is the name of the property
  * @param _l is the PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      // no action if the transiant variable is null
      if (changeSupport_ == null) return;

                    // for testing and tracking
                  //  System.out.println("StringRightImpl.removePropertyChangeListener(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      synchronized (changeSupport_)
        {changeSupport_.removePropertyChangeListener(_propertyName, _l);}
  } // removePropertyChangeListener()  
  
  /**
  *  // TODO
  * This method provides a copy of the instance for the same or another acs. 
  * It is designed to facilitate the copy of a whole rights network from
  * the most powerfull rights. <P>
  * Each unmodified internal variable that is
  * an external object reference - metaright, lower and upper rights - 
  * have a shallow copy (no deep copy) if the ACS is the same.
  * Otherwise, there is still a shallow copy of the metaright, but the
  * the upper rights are searched in the copy ACS and the lower rights
  * are NOT copied. <P>
  * If the instance is a typed right, the type class and type value are
  * copied.
  * @param _name is the copy name; must be a one-component name,
  *   and the component is the kernelRight_ parameter for the copy.
  * @param _acs is the ACS of the copy to create. May be different
  *   to the ACS of this instance.
  * @param _parent must be null here.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in its ACS, if it exists.
  * @return Object is a copy of the object, in the same class.
  * @exception CreateError if the acs, the kernel right is null.
  * @version 0.5.3
  **/ /*
  public Object copy(ImmutableName _name, ACS _acs, Node _parent,
    Boolean _toRegisterInManager) throws CreateError {

      //if (_kernelRight == null || 
        //_meta != null || _internalKey < 1 || _internalKey > 999) {
          //throw new CreateError(
            //"The right or the metaright is null,"+
            //" or the internal key is not correct.");
      //}
      
      StringRightImpl result = null;
      //if (_acs == null) 

      //else if (_acs == acsName_) {

      //else {
          // TODO
      //}
      
      // TODO: set the other arguments
      result.isParent_ = isParent_;
      result.isSlave_ = isSlave_;
      
      return result;
  } // copy()
       **/

  //
  // -----------------  protected methods  ------------------------
  //

  /**
   * Adds, for an embedded right only, an upper metaright which are strictly
   * greater than this embedded right owner, without any intermediate metaright.
   * Register the owner in the upper metaright (its embedded right, in fact).
   * Called only by the owner metaright. The upper right is not updated.
   * No comparison between the upper right target class and this right target class,
   * but BaseUtilityImpl.initializeBasicMetaRights() is written to avoid
   * an upper right with a more restricted target class.
   * Noop if the new right is already known. It is a bound property.
   * @param _right which directly includes this right. Never null.
   * @exception BaseError if the right is a slave, or the parameter is null,
   *    or the number of upper rights is already 5,
   *    or an upper right is an administration right while the new right is not,
   *    or it is not an embedded right.
   */
  protected void addUpperMetaRight(StringMetaRightImpl _right) throws BaseError {
      if (_right == null || acsName_ != null)
          throw new BaseError("in StringRightImpl.addUpperRight(), "+
            "the parameter is null or it is not an embedded right.");
      if (isSlave_)
          throw new BaseError("in StringRightImpl.addUpperRight(), "+
            "the right is a slave; it cannot accept another upper right.");
      
      if (l_upperRights_ != null && l_upperRights_.length > 4)
        throw new BaseError("in StringRightImpl.addUpperRight(), the upper rights "+
              "set has the maximal size of 5.");  
      
      StringRight[] old = l_upperRights_;
    
      if (l_upperRights_ == null || l_upperRights_.length == 0) {
          l_upperRights_ = new StringRight[1];
          l_upperRights_[0] = _right;
      } else if ( Arrays.asList(l_upperRights_).contains(_right) ) {
          old = null;
          return;      
      } else {
          StringRight[] l_rights = new StringRight[l_upperRights_.length + 1];
          System.arraycopy(l_upperRights_, 0, l_rights, 0, l_upperRights_.length);
          l_rights[l_rights.length - 1] = _right;
          l_upperRights_ = l_rights;
          l_rights = null;
      } // else

    try {
          firePropertyChange("UpperRights", old, getUpperRights());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
        e.printStackTrace();
    }       
    old = null;
    return;
  } // addUpperMetaRight()
  
  /**
   * Removes a upper right which are strictly greater than this right
   * without any intermediate right, following the network of metarights.
   * Called by the upper right finalizeForUser() and,
   * for embedded right, by the owner metaright. It is a bound property.
   * @param _right directly includes this right. Never null.
   * @param _caller is the object which  calls this method; used to
   *   identify the upper right finalizing operation with _caller == _right
   *   for an ACS right only.
   */
  protected void removeUpperRight(StringRight _right, Object _caller) {
    if (_right == null || l_upperRights_ == null )
      throw new InternalError("in StringRightImpl.removeUpperRight(), "+
        "the parameter or the array of rights is null or empty.");

    int index = 10000;
    for (int i = 0; i < l_upperRights_.length; i++) {
      if (_right == l_upperRights_[i]) {
          index = i;
          break;
      }
    } // for

    if (index == 10000) throw new InternalError("in StringRightImpl.removeUpperRight(), "+
    "the parameter is not known.");

    StringRight[] old = l_upperRights_;
    
    if (l_upperRights_.length == 1) l_upperRights_ = null;
    else {
      StringRight[] l_rights = new StringRight[l_upperRights_.length - 1];
      System.arraycopy(l_upperRights_, 0, l_rights, 0, index);
      if (index != l_upperRights_.length - 1)
          System.arraycopy(l_upperRights_, index + 1, 
                  l_rights, index, l_upperRights_.length - index - 1);
      l_upperRights_ = l_rights;
      l_rights = null;
    } // else

    // if slave, there was exactly one upper right, so it is then no more a slave.
    if (isSlave_) isSlave_ = false;

    if (_caller != _right) { // it is not a finalizing of the upper right, for an ACS right only
      if (acsName_ != null)  ((StringRightImpl)_right).removeLowerRight(this);
    } // if (_caller

    try {
          firePropertyChange("UpperRights", old, getUpperRights());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
        e.printStackTrace();
    }       
    old = null;
    return;      
  } // removeUpperRight
  
  /**
   * Adds a lower right which are strictly lesser than this right
   * without any intermediate right, following the network of rights in the ACS.
   * Called by the lower StringRightImpl constructor and by StringMetaRightImpl.addUpperRight()
   * through addLowerRight(), after a check if it is a lower right.
   * Noop if the new right is already known. It is a bound property.
   * @param _right is directly included in this right. Never null.
   */
  protected void addLowerRight(StringRight _right) {
    if (_right == null)
      throw new InternalError("in StringRightImpl.addLowerRight(), "+
        "the parameter is null.");
    StringRight[] old = l_lowerRights_;

    if (l_lowerRights_ == null || l_lowerRights_.length == 0) {
      l_lowerRights_ = new StringRight[1];
      l_lowerRights_[0] = _right;
    } else if ( Arrays.asList(l_lowerRights_).contains(_right) ) {
      old = null;
      return;
    } else {
      StringRight[] l_rights = new StringRight[l_lowerRights_.length + 1];
      System.arraycopy(l_lowerRights_, 0, l_rights, 0, l_lowerRights_.length);
      l_rights[l_rights.length - 1] = _right;
      l_lowerRights_ = l_rights;
      l_rights = null;
    } // else

    try {
          firePropertyChange("LowerRights", old, getLowerRights());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
        e.printStackTrace();
    }       
    old = null;
    return;
  } // addLowerRight()
  
  /**
   * Removes a lower rights which are strictly lesser than this right
   * (or the metaright, if this instance is embedded)
   * without any intermediate right, following the network of rights in the ACS.
   * Called by the lower right finalizeForUser() method, and by StringMetaRightImpl
   * when it is an embedded right.
   * It is a bound property.
   * @param _right is directly included in this right. Never null.
   */
  protected void removeLowerRight(StringRight _right) {
    if (_right == null || l_lowerRights_ == null ) {

      //System.out.println("In StringRightImpl.removeLowerRight(), _right:"+_right+
        //      " and l_lowerRights_:"+l_lowerRights_+" for the right "+right_);

      throw new InternalError("in StringRightImpl.removeLowerRight(), "+
        "the parameter or the array of rights is null or empty.");
    }

    int index = -1;
    for (int i = 0; i < l_lowerRights_.length; i++) {
      if (_right == l_lowerRights_[i]) {
          index = i;
          break;
      }
    } // for

    if (index == -1) throw new InternalError("in StringRightImpl.removeLowerRight(), "+
       "the parameter is not known.");
    StringRight[] old = l_lowerRights_;    
    
    if (l_lowerRights_.length == 1) {
      l_lowerRights_ = null;
      //isParent_ = false;
    } else {
      StringRight[] l_rights = new StringRight[l_lowerRights_.length - 1];
      System.arraycopy(l_lowerRights_, 0, l_rights, 0, index);
      if (index != l_lowerRights_.length - 1)
          System.arraycopy(l_lowerRights_, index + 1, 
                  l_rights, index, l_lowerRights_.length - index - 1);
      l_lowerRights_ = l_rights;
      l_rights = null;
    } // else

    try {
      firePropertyChange("LowerRights", old, getLowerRights());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From StringRightImpl.setComment(), firePropertChange() throws:");
        e.printStackTrace();
    }       
    old = null;      
    return;      
  } // removeLowerRight

  /**
  * Fires an event to every registered listener, in any order.
  * For the bound properties 'Comment', 'Parent', 'Slave', 'Desactivated',
  * 'LowerRights', 'UpperRights'.
  * Called by all the updating methods.
  * Used as the same method in java.beans.PropertyChangeSupport, which is not
  * used because hasListeners() has a bug in JDK 1.3.0-C of SUN.
  * @param _propertyName name of the changing property
  * @param _oldValue of the property
  * @param _newValue of the property. No operation if equal to _oldValue
  *   and not null.
  */
  protected void firePropertyChange(String _propertyName,
    Object _oldValue, Object _newValue) {

                // for testing and tracking
                //System.out.println("in StringRightImpl.firePropertyChange(), call to "+
                //  "propertyChange() of "+name_+" for the property "+_propertyName);

      if (_oldValue != null && _newValue != null && _oldValue.equals(_newValue)) {
          return;
      }

      // no action if the transiant variables are null, after a deserialization
      if (l_CommentChangeSupport_ == null) return;


      PropertyChangeEvent evt = new PropertyChangeEvent(this,
                                          _propertyName, _oldValue, _newValue);

      PropertyChangeListener listener;
      Set toFire = null;

      if (_propertyName.equals("Comment"))
        toFire = (Set) ((HashSet)l_CommentChangeSupport_).clone(); // no synchro problem
      else if (_propertyName.equals("Parent"))
        toFire = (Set) ((HashSet)l_ParentChangeSupport_).clone();
      else if (_propertyName.equals("Slave"))
        toFire = (Set) ((HashSet)l_SlaveChangeSupport_).clone();
      else if (_propertyName.equals("Desactivated"))
        toFire = (Set) ((HashSet)l_DesactivatedChangeSupport_).clone();      
      else if (_propertyName.equals("LowerRights"))
        toFire = (Set) ((HashSet)l_LowerRightsChangeSupport_).clone();
      else if (_propertyName.equals("UpperRights"))
        toFire = (Set) ((HashSet)l_UpperRightsChangeSupport_).clone();
      
              // for testing and tracking
              //System.out.println("in StringRightImpl.firePropertyChange(), call to "+toFire.size()+
              //  " listeners from "+name_+"for the property "+_propertyName);

      for (Iterator ite = toFire.iterator() ; ite.hasNext() ; ) {
        listener = (PropertyChangeListener)ite.next();

              // for testing and tracking
              //System.out.println("in StringRightImpl.firePropertyChange(), call to "+
                //listener+".propertyChange() from "+name_+"for the property "+_propertyName);

        listener.propertyChange(evt);
      } // for
      evt = null;
      listener = null;
      toFire = null;
      return;
  } // firePropertyChange()  
  
   /**
   * This method is to use only for closed rights in the relevant ACS list,
   * to set an empty instance.
   * It cuts all the Object External References, without updating
   * these objects as finalizeForUser() does.
   * It sets to null all variables, so the instance becames empty.
   * Called by StringMetaRightImpl.finalizeClosedRight().
   */
  protected synchronized void finalizeClosedRight()  {

                //System.out.println("In StringRightImpl, call of " + name_ + ".finalizeClosedRight()");

    if (right_ == null) return;

    meta_ = null;
    metaRight_ = null;
    targetClassType_ = null;
    targetClass_ = null;
    right_ = null;
    internalKey_ = 0;  
    l_lowerRights_ = null;
    l_upperRights_ = null;
    isParent_ = false;
    isSlave_ = false;
    acsName_ = null; 
    return;
  } // finalizeClosedRight()
  
  /**
   * This method removes the right from the ACS list, but there is no
   * guarantee that the JVM calls this method before the end of the application.
   * @exception java.lang.Throwable
   */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {u.printStackTrace();}

    super.finalize();
  } // finalize()    
  
} // StringRightImpl