//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.beans.PropertyChangeListener;

import java.io.Serializable;
import java.io.ObjectStreamField;


/**
* This class is responsible for modeling an account in an ACS,
* as a kind of GroupIDMember that owns some resources and a primary groupID.
* It has some current actors, that are programs starting their execution under it.
* It is a javabean with the following bound properties:
* 'l_UserIDCurrentActors', 'l_UserIDOwnResources', 'PrimGroup', 'Order'.
* The listeners are proper to this instance, outside gBase,
* transient (not serialized in this instance backup).
* They are called in any order. Excepted for 'PrimGroup', they receive only a copy
* of the new value, to protect the property. All the exceptions from the listeners
* are catched, and a dialog box is displayed to inform the user.<P>
* All variables are private to secure the white box inheritances from this class. <P>
* This class has several subclasses in this package and in the AcsAddon packages.
* Most of the variables are protected.
* @version 0.7.0
*/
public class UserIDImpl extends GroupIDMemberImpl
	implements Serializable, UserID {
  // protected variables
  private static final long   serialVersionUID__ = 0;

  protected GroupID     primGroup_ = GroupIDImpl.EMPTY_INSTANCE;  // External Object Reference
  protected int         order_ = -1;
  protected Resource[]  l_ownResources_ = (Resource[]) new ResourceImpl[0];
  protected Actor[]     l_currentActors_ = (Actor[]) new ActorImpl[0];

  /**
   * A reusable empty instance for initialization,
   * to avoid the use of 'new' for default values.
   * Caution: never change the state of the returned instance.
   */
  public static final UserIDImpl EMPTY_INSTANCE = new UserIDImpl();

  /** variable for the JDK 2 serialization
  * @serial ("primGroup_",GroupIDImpl.class),
  * ("order_",Integer.TYPE),
  * ("l_ownResources_",ResourceImpl[].class),
  * ("l_currentActors_",ActorImpl[].class)
  */
  public final static ObjectStreamField[] serialPersistentFields =
  {
      new ObjectStreamField("primGroup_",GroupIDImpl.class),
      new ObjectStreamField("order_",Integer.TYPE),
      new ObjectStreamField("l_ownResources_",ResourceImpl[].class),
      new ObjectStreamField("l_currentActors_",ActorImpl[].class),
  };

 /**
  * simple constructor
  */
  public UserIDImpl() {
  }

   /**
   * Constructor used in the beamer by the editor for the ACS EP list, from the 'New' button.
   * @param _name ImmutableName of the userID as a kind of eligible party
   * @param _acs ACS owning the userID
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users
   * @exception CreateError if ACS.newEorEligibleParty() throws it - the instance
   * must then be deleted !
   */
  public UserIDImpl(ImmutableName _name, ACS _acs, String _epType, Boolean _isAdmin)
          throws CreateError  {
      this(_name, _acs, _epType, _isAdmin, true);

          // for testing purpose
          //System.out.println("call to UserIDImpl constructor with 2 arguments, for "+_name);
  }

   /**
   * Used for the creation of an UserID.
   * @param _name ImmutableName of the eligible party to create
   * @param _acs Main ACS of the eligible party to create
   * @param _epType is the eligible party type, and may be null if the ACS allows it.
   * @param _isAdmin true only if the instance has an administrative role in the ACS,
   *    for the technical administration or the users
   * @param _toRegisterInManager false if this method must not register
   *  the instance in its ACS.
   * @exception CreateError if ACS.newEorEligibleParty() throws it -
   *  the instance must then be deleted !
   */
  public UserIDImpl(ImmutableName _name, ACS _acs, String _epType, boolean _isAdmin,
    boolean _toRegisterInManager) throws CreateError {
      super(_name, _acs, _epType, _isAdmin, _toRegisterInManager);

          // for testing purpose
          //System.out.println("call to UserIDImpl constructor with 3 arguments, for "+_name);

  }
  
  /**
  * This method provides a copy of the instance with another name.
  * Do not copy the own resources or the current actors.
  * The copy order is 3 times the source order.
  * Caution: do not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Call copyOwnVariables().
  * This method is overridden by the subclasses.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of UserIDImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager)
    throws CreateError {

      if (_acs == null || _name == null  ) {
          throw new CreateError("in UserIDImpl.copy(),"+
          " _name or _acs is null.");
      }
      
      UserIDImpl result = new UserIDImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      result = copyOwnVariables(result, false);
      
      return result;
  } // copy(,,,)
  
  /**
  * This method provides a copy of the instance
  * as copy() but without copying the internal AclEntries.
  * Does not copy the own resources or the current actors.
  * The copy order is 3 times the source order.
  * Caution: does not copy the user and the group if _result
  * does not belong to aCS_ and if there are no user or group
  * in the _result ACS which have similar names - that is that
  * are ended by the same Strings in this instance user or group names.
  * Calls copyOwnVariables().
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is not used, wathever the value.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a new instance of UserIDImpl
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null ) {
          throw new CreateError("in UserIDImpl.copy(),"+
          " _name or _acs is null.");
      }
      
      UserIDImpl result = new UserIDImpl(
          _name, _acs, getEPType(),
          _toRegisterInManager.booleanValue());

      // copies the external aclEntries if they exist
      result = copyOwnVariables(result, true); // no internal aclEntry copy
      
      //if (result.getEorL_AclConnectedResources().isEmpty()) return result;
      //else return deleteAclEntries(result);
      return result;
  } // copy(,,,)
  
  /**
   * Sets a primary group, and it must have this instance as member.
   * The primary group is used by an Actor that has a current UserID, to set
   * the primary group as current GroupID of the Actor. If the resource parent
   * does not enforce the child GID, the primary group is the group of any new resource
   * owned by this UserID.
   * Updates the own actors which have the previous primary group as current group.
   * Removes the current primary group in the current actors which have it as
   * their current group. Fire the PropertyChange 'PrimGroup'.
   * Noop and error mesage to the user if the instance type starts with 'byAcsCreator'
   * or 'immutable' after the creation.
   * To do it in that case, remove this instance from all its groups,
   * and then, the first registered group will be the primary group.
   * Reads the ACS 'GroupID.NoPrimaryGroup' values to check if the argument
   * may be a primary group.
   * @param _g GroupID that is the primary group of the userID. May be null.
   * @exception UpDateError if _g does not belong to the main ACS
   *  of this resource or cannot be a primary group, or does not contain this instance,
   *  or if the type is immutable after the first creation.
   */
  public synchronized void setEorPrimGroup(GroupID _g) throws UpDateError {
    if ((type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
            type_.endsWith("<immutable> ")) &&
                primGroup_ != null && !primGroup_.equals(GroupIDImpl.EMPTY_INSTANCE)) 
          throw new UpDateError(
            "This object has been created immutable, or the primary group value"+
            " is immutable.");
    
                        // for testing and tracking
                        //System.out.println("call to UserIDImpl.setEorPrimGroup() 1 for "+_g);
    
    if (_g != null) {
      if (getEorACS() == null) {
                      // for testing and tracking
                      //System.out.println("call to UserIDImpl.setEorPrimGroup() with getEorACS() null");
          throw new UpDateError("In UserIDImpl, the main ACS is null");
      }
      if (getEorACS() != _g.getEorACS() ) {
                      // for testing and tracking
                    //System.out.println("call to UserIDImpl.setEorPrimGroup() getEorACS():"+getEorACS());
                    //System.out.println("call to UserIDImpl.setEorPrimGroup() getEorACS().getEorBaseError():"+
                      //      getEorACS().getEorBaseError());
        getEorACS().getEorBaseError().addException("The selected GroupID "+
          "has to belong to the ACS of this resource.");

                      // for testing and tracking
                      //System.out.println("Exception from UserIDImpl.setEorPrimGroup(): "+
                      //  getEorACS().getEorBaseError().toString());
    
                        // for testing and tracking
                        //System.out.println("call to UserIDImpl.setEorPrimGroup() 2");
        
        throw new UpDateError("In UserIDImpl.setEorPrimGroup(), "+
          "the argument does not belong to this resource main ACS");
      } // if (getEorACS() != _g.getEorACS()
      
      if (!_g.containsAsMemberWithoutAliasing(this))
        throw new UpDateError("The primary group has to have this userID as member.");
    
                          // for testing and tracking
                          //System.out.println("call to UserIDImpl.setEorPrimGroup() 3");
      
      String[] l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
              get("GroupID.NoPrimaryGroup");
      boolean remove = false;
      if (l_remove != null) {
          for (int i = 0; i < l_remove.length; i++) {
              if (l_remove[i].equals(
                      UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(_g.getEPType()))) {
                  remove = true;
                  break;
              }
          } // for i
      } // if (l_remove !=
      l_remove = null;
    
                          // for testing and tracking
                          //System.out.println("call to UserIDImpl.setEorPrimGroup() 4");      
      
      if (remove) {
          throw new UpDateError("This group type is in the ACS values "+
                  "for 'GroupID.NoPrimaryGroup'; "+
                  "it cannot be a primary group.");
          //l_remove = null;  
          //return;
      }
    } // if (_g != null) {
    
    GroupID oldValue = primGroup_;
    
    for (int i = 0; i < l_currentActors_.length; i++) {
        if (l_currentActors_[i].getEorCurrentGroupID() == primGroup_)
            l_currentActors_[i].setEorCurrentGroupIDForUser(_g);
    } // for i
    
                        // for testing and tracking
                        //System.out.println("call to UserIDImpl.setEorPrimGroup() 5");
    
    if (_g != null) primGroup_ = _g;
    else primGroup_ = GroupIDImpl.EMPTY_INSTANCE;

    try {
        firePropertyChange("PrimGroup", oldValue, _g);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDImpl.setEorPrimGroup(), firePropertChange() throws:");
        e.printStackTrace();
    }
  } // setEorPrimGroup()

  /**
   * The primary group is used by an Actor that has a current UserID, to set
   * the primary group as current GroupID of the Actor. If the resource parent
   * does not enforce the child GID, the primary group is the group of any new resource
   * owned by this UserID.
   * @return the primary group of the userID. May be null.
   */
  public GroupID getEorPrimGroup() {
    if (primGroup_ == null ||
      primGroup_.equals(GroupIDImpl.EMPTY_INSTANCE) ||
      primGroup_.toString() == null) {
          return null;
    }
    else return primGroup_;
  } // getEorPrimGroup()

  /**
 * Sets the strengh of the UserID which determines its priority order
 * in access control decisions.
 * Noop without error message if the ep type starts with 'byAcsCreator' or 'immutable',
 * with a current order that is different to the initial value '-1'.
 * @param _i order of the userID
 * @exception UpDateError if immutable
 */
  public synchronized void setOrder(int _i) throws UpDateError {
    if ((type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
            type_.endsWith("<immutable> "))
            && order_ != -1)
                  throw new UpDateError(
                        "This object has been created immutable, or the order value"+
                        " after the first creation is immutable.");
    
    int oldValue = order_;
    order_ = _i;
    
    try {
        firePropertyChange("Order", new Integer(oldValue), new Integer(_i));
     } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDImpl.setOrder(), firePropertChange() throws:");
        e.printStackTrace();
    }
  }

  /**
   * @return the UID of the userID
   */
  public int getOrder()	{
    return order_;
  }

  //
  // -------- Overridden public methods ------------------
  //

  /**
  * Adds a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'l_UserIDCurrentActors', 'l_UserIDOwnResources', 'PrimGroup', 'Order'.
  * Calls the super method.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      super.addPropertyChangeListener(_propertyName, _l);

  } // addPropertyChangeListener(String _propertyName, _l)

  /**
  * Removes a listener to the bean.
  * Used by Gui1.CommonPropertyEditor for the properties:
  * 'l_UserIDCurrentActors', 'l_UserIDOwnResources', 'PrimGroup', 'Order'.
  * Note: using this method does not remove a listener which has been
  * registered through addPropertyChangeListener(PropertyChangeListener _l).
  * Overrides the super method and calls it.
  * @param _propertyName name of the property
  * @param _l PropertyChangeListener to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){

      super.removePropertyChangeListener(_propertyName, _l);

                //if (_propertyName.equals("AclRights")) {
                  //  System.out.println("AclRights listeners after the removing of "+_l+": ");
                    //PropertyChangeListener[] l_list = changeSupport_.getPropertyChangeListeners("AclRights");
                    //for (int i = 0; i < l_list.length; i++) {
                      //  System.out.println("-------"+l_list[i]);
                    //}
                    //l_list = null;
                //} // if

  } // removePropertyChangeListener()
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      return super.getPropertyChangeListeners(_propertyName);
  } // getPropertyChangeListeners()

  /**
   * @return the static serialVersionUID__ of this UserID class
   */
  public long getSerialVersionUID()	{
          return serialVersionUID__;
  }

  //
  // ------- protected methods ---------------------
  //
  
  /**
  * Invoked by Resource.setEorUserID().
  * Fires an event on 'l_UserIDOwnResources'.
  * @param _res owned by this userID.
  */
  protected void addOwnResource(Resource _res) {
    // variables
    Resource[]  newArray = new ResourceImpl[1];
    boolean     recorded = false;

    //if (l_ownResources_ == null) l_ownResources_ = (Resource[]) new ResourceImpl[1];

    Resource[]  oldValue = new ResourceImpl[l_ownResources_.length];
    System.arraycopy(l_ownResources_, 0, oldValue, 0, l_ownResources_.length);

    // add in an empty place if possible
    for (int i = 0 ; i < l_ownResources_.length ; i++) {
        if (l_ownResources_[i] == null || l_ownResources_[i].equals(ResourceImpl.EMPTY_INSTANCE) ) {
            l_ownResources_[i] = _res;
            recorded = true;
            break;
        }
    }
    // else, add at the end of l_ownResources_
    if (recorded == false) {
        newArray = new ResourceImpl[l_ownResources_.length + 1];
        System.arraycopy(l_ownResources_, 0, newArray, 0, l_ownResources_.length);
        newArray[l_ownResources_.length] = _res;
        l_ownResources_ = new ResourceImpl[newArray.length];
        l_ownResources_ = newArray;
    }
    else {
        newArray = new ResourceImpl[l_ownResources_.length];
        System.arraycopy(l_ownResources_, 0, newArray, 0, l_ownResources_.length);
    }

    try {
        firePropertyChange("l_UserIDOwnResources", oldValue, newArray);
     } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDImpl.addOwnResource(), firePropertChange() throws:");
        e.printStackTrace();
    }
    oldValue = null;
    newArray = null;
  } // addOwnResource(Resource _res)

  /**
   * A current actor is running under this userID whatever its caller.
  * Invoked by Actor.setEorCurrentUserID().
  * Fire an event on "l_UserIDCurrentActors".
  * @param _act currently attached to this userID.
  */
  protected void addCurrentActor(Actor _act) {
    // variables
    Actor[]     newArray = new ActorImpl[1];
    boolean     recorded = false;

    //if (l_currentActors_ == null) l_currentActors_ = (Actor[]) new ActorImpl[1];

    Actor[]     oldValue = new ActorImpl[l_currentActors_.length];
    System.arraycopy(l_currentActors_, 0, oldValue, 0, l_currentActors_.length);

    // add in an empty place if possible
    for (int i = 0 ; i < l_currentActors_.length ; i++) {
        if (l_currentActors_[i] == null || l_currentActors_[i].equals(ActorImpl.EMPTY_INSTANCE) ) {
            l_currentActors_[i] = _act;
            recorded = true;
            break;
        }
    }
    // else, add at the end of l_currentActors_
    if (!recorded) {
        newArray = new ActorImpl[l_currentActors_.length + 1];
        System.arraycopy(l_currentActors_, 0, newArray, 0, l_currentActors_.length);
        newArray[l_currentActors_.length] = _act;
        l_currentActors_ = new ActorImpl[newArray.length];
        l_currentActors_ = newArray;
    }
    else {
        newArray = new ActorImpl[l_currentActors_.length];
        System.arraycopy(l_currentActors_, 0, newArray, 0, l_currentActors_.length);
    }

    try {
        firePropertyChange("l_UserIDCurrentActors", oldValue, newArray);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDImpl.addCurrentActor(), firePropertChange() throws:");
        e.printStackTrace();
    }
    oldValue = null;
    newArray = null;
  } // addCurrentActor()
  
   /**
   * Overrides the super class method. Calls setEorPrimGroup(_grp) if no primary group.
   * @param _grp to add. No action if it is null, or with null name.
   */
  public synchronized void addGroupToGM(GroupID _grp) {

                    // for testing
                    //System.out.println("UserIDImpl.addGroupToGM() for "+getName()+
                      //      " for a fgroup "+first);
      
      // updates l_Groups_
      super.addGroupToGM(_grp);
 
      if (getEorPrimGroup() != null) return;
      
      String    temp = null;
      String[]  l_remove = null;
      boolean   remove = false;
      
      if (type_.startsWith("<byAcsCreator>")) {
          temp = type_;
          type_ = "12345678901234567890";
      }
      if (type_.startsWith("<immutable> ") || type_.endsWith("<immutable> ")) {
          temp = type_;
          type_ = "01234567891234567890";
      }

      l_remove = (String[]) getEorACS().getM_TypeOrientedPolicy().
              get("GroupID.NoPrimaryGroup");
      
      if (l_remove != null) {
          for (int i = 0; i < l_remove.length; i++) {
              if (l_remove[i].equals(
                      UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(_grp.getEPType()))) {
                  remove = true;

                // for testing
                //System.out.println("UserIDImpl.addGroupToGM() for "+getName()+
                  //      ", 'remove' set to true with "+l_remove[i]); 

                  break;
              }
          } // for i
      } // if (l_remove !=

                // for testing
                //System.out.println("UserIDImpl.addGroupToGM() for "+getName()+
                  //      ", do not call setEorPrimGroup(): "+remove);           

      // set _grp as primary group if it is the first group,
      // and if it is not forbidden by the acs
      try {
          if (!remove) setEorPrimGroup(_grp);
      } catch (UpDateError u) {
                System.out.println("from UserIDImpl.addGroupToGM(), "+u.toString());           
      }

      if (type_.equals("12345678901234567890") ||
          type_.equals("01234567891234567890")) {
            type_ = temp;
      }
      temp = null;
      l_remove = null;

                    // for testing
                    //System.out.println("UserIDImpl.addGroupToGM(): end with type "+type_);
      
  } // addGroupToGM(
  
  /**
   * Removes the current primary group in the current actors which have it as
   * their current group.
   * Called by GroupID.finalizeForBase() and removeGroupToGM().
   * Calls ActorImpl.removeEorCurrentGroupID().
   * Fire the Propertychange 'PrimGroup'.
   */
  protected void removePrimGroup() {    
    GroupID oldValue = primGroup_;

    for (int i = 0; i < l_currentActors_.length; i++) {
        if (l_currentActors_[i].getEorCurrentGroupID() == primGroup_)
            ((ActorImpl)l_currentActors_[i]).removeEorCurrentGroupID();
    } // for i

    
    primGroup_ = GroupIDImpl.EMPTY_INSTANCE;

    try {
        firePropertyChange("PrimGroup", oldValue, null);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDImpl.setEorPrimGroup(), firePropertChange() throws:");
        e.printStackTrace();
    }
    return;
  } // removePrimGroup()  
  
   /**
   * A current actor is running under this userID whatever its caller.
   * It may be an own resource of this userID or not. 
   * @return array of the current actors attached to this userID.
   *  May be empty, but never null.
   */
   public Actor[] getEorL_UserIDCurrentActors() {
        if (l_currentActors_ == null) return new ActorImpl[0];
        Actor[] result = new ActorImpl[l_currentActors_.length];
        System.arraycopy(l_currentActors_, 0, result, 0, l_currentActors_.length);
        return result;
  }

  /**
  * An own resource is a same-ACS resource that have this as userID.
  * If it is an Actor, it may be a current actor of this userID or not.
  * @return array of the own resources. May be empty, but never null.
  */
  public Resource[] getEorL_UserIDOwnResources() {
      if (l_ownResources_ == null) return new ResourceImpl[0];
      Resource[] result = new ResourceImpl[l_ownResources_.length];
      System.arraycopy(l_ownResources_, 0, result, 0, l_ownResources_.length);
      return result;
  }

  /**
  * @return true if it owns the resource.
  */
  public boolean ownsResource(Resource _res) {
      if (_res == null || l_ownResources_ == null) return false;
      boolean result = false;
      for (int i = 0; i < l_ownResources_.length; i++) {
        if (l_ownResources_[i] == _res) {
          result = true;
          break;
        }
      } // for
      return result;
  }

    /**
    * This method removes the userID from the ACS lists, but there is no
    * guaranee that the JVM calls this method before the end of the application.
    * @exception java.lang.Throwable
    * @exception  Throwable
    */
  protected void finalize() throws Throwable {
    try {
      finalizeForBase();
    } catch (UpDateError u) {
      super.finalize();
      throw new InternalError("in UserIDImpl.finalize(): "+u.toString());
    }

    super.finalize();
  }

   /**
   * super.finalizeForBase() removes the userID in the ACS lists (main and controllers),
   * and removes it in the resources connected through an aclEntry.
   * This method removes (this) in own resources and in current actors,
   * so they have no more UserID connected to them.
   * Overrides the super method and calls it.
   * This method HAS TO BE USED before deleting the userID.
   * Note: when this method is called by the EligiblePartyImpl constructor, the proper
   * variables of this class are not still initialized.
   * @exception UpDateError if the userID is unknown of the main ACS
   * (if it is a clone, typically), or if a linked immutable object fires an exception
   */
   public synchronized void finalizeForBase() throws UpDateError {

              // for testing purpose
              //System.out.println("call 1 to UserIDImpl.finalizeForBase() for "+getName());

    // noop if it is the second call to this method
    if (folderMember_ == null) {
      return;
    }

              // for testing purpose
              //System.out.println("call to UserIDImpl.finalizeForBase() for "+getName()+
                //      " with folderMember_: "+folderMember_);

    // set userID_ and currentUserID_ (for Actors) to null in all the resources
    // that are connected to this userID

            // for testing purpose
            //System.out.println("in UserIDImpl.finalizeForBase(), l_ownResources_.length: "+
            //  l_ownResources_.length);

    if (l_ownResources_ != null) { // false if called by EligiblePartyImpl constructor
        for (int i = 0 ; i < l_ownResources_.length ; i++) {
          if (l_ownResources_[i] != null ) {
            // l_ownResources_[i] can't be equals to nullRes, so:
            ((ResourceImpl)l_ownResources_[i]).removeEorUserID();
          }
        } // for i
    } // if (l_ownResources_

    if (l_currentActors_ != null) {
        for (int i = 0 ; i < l_currentActors_.length ; i++) {
          if (l_currentActors_[i] != null ) {
            // l_currentActors_[i] can't be equals to nullAct, so:
            ((ActorImpl)l_currentActors_[i]).removeEorCurrentUserID();
          }
        } // for i
    } // if (l_currentActors_

    primGroup_ = null;
    l_ownResources_ = null;
    l_currentActors_ = null;
    super.finalizeForBase();
  } // finalizeForBase()

  /**
  * This method finalizes each group in the groups set, since it is used after
  * a deserialization. There, the attributes have been deserialized but
  * are not 'true' objects of the base.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null.
  * Does not call finalizeForBase(). Called only by EPRViewInBaseImpl.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In UserIDImpl, call of " + name_ +
                    //".finalizeFromDeserialization()");


    // noop if it is the second call to this method
    if (folderMember_ == null) {
      return;
    }

    // finalizes all the resources and the current actors
    // that are connected to this userID
    for (int i = 0 ; i < l_ownResources_.length ; i++) {
      if (l_ownResources_[i] != null ) {
        // l_ownResources_[i] can't be equals to nullRes, so:
        ((ResourceImpl)l_ownResources_[i]).finalizeFromDeserialization();
      }
    } // for i

    for (int i = 0 ; i < l_currentActors_.length ; i++) {
      if (l_currentActors_[i] != null ) {
        ((ActorImpl)l_currentActors_[i]).finalizeFromDeserialization();
      }
    } // for i

    primGroup_ = null;
    l_ownResources_ = null;
    l_currentActors_ = null;
    
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  /**
  * Clone without register the new object in the ACSs.
  * Caution: don't use any modifying method that may altere the integrity
  * of the base.
  * @return Object is not registred in the ACSs. The names, the type, the comment,
  * the alias, the order are cloned.
  * The external object references are not cloned (see "Eor" in get method).
  */
  public Object clone() {
      UserIDImpl clon = (UserIDImpl) super.clone();
      clon.primGroup_ = primGroup_;
      clon.order_ = order_;

      if (l_ownResources_ != null) {
        clon.l_ownResources_ = new ResourceImpl[l_ownResources_.length];
        for (int i = 0 ; i < l_ownResources_.length ; i++) {
          if (l_ownResources_[i] != null) {
            //clon.l_ownResources_[i] = (Resource) l_ownResources_[i].clone();
          }
        }
      } else clon.l_ownResources_ = null;

      if (l_currentActors_ != null) {
        clon.l_currentActors_ = new ActorImpl[l_currentActors_.length];
        for (int i = 0 ; i < l_currentActors_.length ; i++) {
          if (l_currentActors_[i] != null) {
            //clon.l_currentActors_[i] = (Actor) l_currentActors_[i].clone();
          }
        }
      } else clon.l_currentActors_ = null;

      return clon;
  }

  /**
  * @param _obj Object to compare
  * @return true if _obj is an instance of UserIDImpl (or a derived class)
  * equals to this instance (with the same class)
  */
  public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      if (!super.equals(_obj))
          return false;

      UserID obj = (UserID) _obj;

      if (primGroup_ == null || obj.getEorPrimGroup() == null) {
        if (primGroup_ != null || obj.getEorPrimGroup() != null)
          return false;
      } // if
      else {
        if (!primGroup_.equals( obj.getEorPrimGroup()) ) return false;
      } // else

      if (order_ != obj.getOrder() ) return false;

      if (l_ownResources_ == null || obj.getEorL_UserIDOwnResources() == null) {
        if (l_ownResources_ != null || obj.getEorL_UserIDOwnResources() != null)
          return false;
      } // if
      else {
        if (l_ownResources_ != null && obj.getEorL_UserIDOwnResources() != null) {
            if (l_ownResources_.length != obj.getEorL_UserIDOwnResources().length)
                return false;
            for (int i = 0; i < l_ownResources_.length; i++) {
                if (l_ownResources_[i] == null || obj.getEorL_UserIDOwnResources()[i] == null) {
                    if (l_ownResources_[i] != null || obj.getEorL_UserIDOwnResources()[i] != null)
                        return false;
                    continue; // i++
                } // if
                else {
                    if (!l_ownResources_[i].equals( obj.getEorL_UserIDOwnResources()[i] ))
                        return false;
                } // else
            } // for
        } // if
      } // else


      if (l_currentActors_ == null || obj.getEorL_UserIDCurrentActors() == null) {
        if (l_currentActors_ != null || obj.getEorL_UserIDCurrentActors() != null)
          return false;
      } // if
      else {
        if (l_currentActors_ != null && obj.getEorL_UserIDCurrentActors() != null) {
            if (l_currentActors_.length != obj.getEorL_UserIDCurrentActors().length)
                return false;
            for (int i = 0; i < l_currentActors_.length; i++) {
                if (l_currentActors_[i] == null || obj.getEorL_UserIDCurrentActors()[i] == null) {
                    if (l_currentActors_[i] != null || obj.getEorL_UserIDCurrentActors()[i] != null)
                        return false;
                    continue; // i++
                } // if
                else {
                    if (!l_currentActors_[i].equals( obj.getEorL_UserIDCurrentActors()[i] ))
                        return false;
                } // else
            } // for
        } // if
      } // else

      return true;
  } // equals()

   /**
   * @return the hashcode derived from the name of the primary groupID, the first groupID,
   * the type and folder member hashcodes
   */
  public int hashCode() {
    int result = super.hashCode();

    if (primGroup_ != null) {
      result = PRIME*result + primGroup_.getName().hashCode();
    }

    return result;
  }

   //
   // -------------- Out of the extended interfaces -------------------------
   //


  /**
  * @return String: the variables of the instance : name (last component).
  *  May be null.
   */
  public String toString() {
    /*
      StringBuilder sb = new StringBuilder();
    // else, 'null' is added to sb
    if (getName().lastComponent() != null) sb.append(getName().lastComponent() );

        sb.append("  | " );
    sb.append("userID order: " );
    sb.append(order_ );

    sb.append("USERIDIMPL :");
    sb.append(super.toString());
    sb.append("END of ELIGIBLEPARTYIMPL");

    sb.append("; primGroup_ = ");
    sb.append(primGroup_);

    sb.append("; l_ownResources_ =");
    for (int i = 0; i < l_ownResources_.length; i++) {
        sb.append(l_ownResources_[i]);
        sb.append(", ");
    } //for

    sb.append("; l_currentActors_ =");
    for (int i = 0; i < l_currentActors_.length; i++) {
        sb.append(l_currentActors_[i]);
        sb.append(", ");
    } //for

    return " [[[" + sb + " end of USERIDIMPL ]]]";
*/
    return getName().lastComponent();
  } // toString()

  //
  // ----------- PROTECTED METHODS -------------------------------------
  //
  
  /**
  * This method provides a copy of the instance variables.
  * Copies the result as groupIDMember
  * (done by GroupIDMemberImpl.copyOwnVariables()).
  * Does not copy the own resources and the current actors,
  * but the resource and actor copies do the job.
  * The copy order is 3 times the source order.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @param _withoutInternalAcl true if the internal aclEntries 
  *   - that is with a resource having the same ACS - must not be copied
  * @return updated _result
  * @exception CreateError 
  **/
  protected UserIDImpl copyOwnVariables(UserIDImpl _result,
    boolean _withoutInternalAcl) throws CreateError {

      if (_result == null) {
          throw new CreateError("in UserIDImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      _result = (UserIDImpl)super.copyOwnVariables(_result, _withoutInternalAcl);

      _result.order_ = getOrder()*3;
      
      return _result;
  } // copyOwnVariables(,,,)
  
   /**
   * Invoked by Resource.setEorUserID(), Resource.finalizeForBase()
   * Fires an event on "l_UserIDOwnResources".
   * @param _res Resource to remove in this userID
   * @exception UpDateError if _res is null or unknown
   */
   protected void removeOwnResource(Resource _res) throws UpDateError {
    // noop if finalizeForBase() has been called
    if (l_ownResources_ == null) return;

    // variables and initializations
    short           count = 0;
    short           shiftingBack = 0;
    // l_ownResources_ can't be null in this method, so:
    Resource[]      newArray = new ResourceImpl[l_ownResources_.length];
    Resource[]      result = new ResourceImpl[0];
    Resource[]      oldValue = new ResourceImpl[l_ownResources_.length];
    System.arraycopy(l_ownResources_, 0, oldValue, 0, l_ownResources_.length);

    // count the occurences of _res:
    for (int i = 0 ; i < l_ownResources_.length ; i++) {
      if (l_ownResources_[i] != null ) {
        if (l_ownResources_[i].equals(_res) ) count++;
      }
    } // for i

    // exceptions
    if (_res == null) {
      getEorACS().getEorBaseError().addException(
              "In UserIDImpl.removeOwnResource(_res), _res == null ");
      throw new UpDateError("In UserIDImpl.removeOwnResource(_res), _res == null ");
    }
    if (count == 0 && l_ownResources_.length != 0) {
        getEorACS().getEorBaseError().addException(
                "In UserIDImpl.removeOwnResource(_res), _res is unknown ");
        throw new UpDateError("In UserIDImpl.removeOwnResource(_res), _res is unknown ");
    }

    // update l_ownResources_
    for (int  i = 0 ; i < l_ownResources_.length ; i++) {
      if (l_ownResources_[i] != null) {
        // incremente shiftingBack delete the reference l_ownResources_[i]:
        if (l_ownResources_[i].equals(_res) ) shiftingBack++;
        else newArray[i - shiftingBack] = l_ownResources_[i];
      }
      else shiftingBack++; // delete null l_ownResources_[i]
    } // for

    if (l_ownResources_.length != shiftingBack) {
      result = new ResourceImpl[l_ownResources_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
      l_ownResources_ = new ResourceImpl[result.length];
      System.arraycopy(result, 0, l_ownResources_, 0, result.length);
    } // if
    else {
                // else result (already egal to) new ResourceImpl[0]
      l_ownResources_ = result;
    } // else

    try {
        firePropertyChange("l_UserIDOwnResources", oldValue, getEorL_UserIDOwnResources());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDImpl.removeOwnResource(), firePropertChange() throws:");
        e.printStackTrace();
    }
    oldValue = null;
    newArray = null;
    result = null;
  } // removeOwnResource()

   /**
   * Invoked by Actor.finalizeForBase(), Actor.setEorCurrentUserID()
   * or by gWork. Fires an event on "l_UserIDCurrentActors".
   * @param _act to remove in this userID
   * @exception UpDateError if _act is null or unknown
   */
   protected void removeCurrentActor(Actor _act) throws UpDateError {
    // noop if finalizeForBase() has been called
    if (l_currentActors_ == null) return;

    // variables and initializations
    short        count = 0;
    short        shiftingBack = 0;
    // l_currentActors_ can't be null in this method, so:
    Actor[]      newArray = new ActorImpl[l_currentActors_.length];
    Actor[]      result = new ActorImpl[0];
    Actor[]      oldValue = new ActorImpl[l_currentActors_.length];
    System.arraycopy(l_currentActors_, 0, oldValue, 0, l_currentActors_.length);

    // count the occurences of _act:
    for (int i = 0 ; i < l_currentActors_.length ; i++) {
      if (l_currentActors_[i] != null ) {
        if (l_currentActors_[i].equals(_act) ) count++;
      }
    } // for i

    // exceptions
    if (_act == null) {
      getEorACS().getEorBaseError().addException(
        "In UserIDImpl.removeCurrentActor(_act), _act == null ");
      throw new UpDateError(
        "In UserIDImpl.removeCurrentActor(_act), _act == null ");
    }
    if (count == 0 && l_currentActors_.length != 0) {
        getEorACS().getEorBaseError().addException(
          "In UserIDImpl.removeCurrentActor(_act), _act is unknown ");
        throw new UpDateError("In the UserIDImpl "+getName()+
          ".removeCurrentActor(_act), "+_act+" is unknown ");
    }

    // update l_currentActors_
    for (int  i = 0 ; i < l_currentActors_.length ; i++) {
      if (l_currentActors_[i] != null) {
        // incremente shiftingBack delete the reference l_currentActors_[i]:
        if (l_currentActors_[i].equals(_act) ) shiftingBack++;
        else newArray[i - shiftingBack] = l_currentActors_[i];
      }
      else shiftingBack++; // delete null l_currentActors_[i]
    } // for

    if (l_currentActors_.length != shiftingBack) {
      result = new ActorImpl[l_currentActors_.length - shiftingBack];
      System.arraycopy(newArray, 0, result, 0, newArray.length - shiftingBack);
      l_currentActors_ = new ActorImpl[result.length];
      System.arraycopy(result, 0, l_currentActors_, 0, result.length);
    } // if
    else {
      l_currentActors_ = result;
    } // else

    try {
        firePropertyChange("l_UserIDCurrentActors", oldValue, getEorL_UserIDCurrentActors());
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From UserIDImpl.removeCurrentActor(), firePropertChange() throws:");
        e.printStackTrace();
    }
    oldValue = null;
    newArray = null;
    result = null;
  } // removeCurrentActor()

   /**
   * Remove a group for which this GroupIDMember is a member, and remove it as primary group.
   * Fire changes on 'AssociatedGroups'.
   * No exception if _grp is not registered.
   * Used only by GroupIDImpl. Call removePrimGroup().
   * Overriddes the GroupIDMemberImpl method.
   * @param _grp to remove
   * @exception InternalError if _grp is not found.
   */
  public final void removeGroupToGM(ImmutableGroupID _grp) {

    if (_grp == null) return;
    
    if (_grp == primGroup_) removePrimGroup();
    
    super.removeGroupToGM(_grp);
  } // removeGroupToGM()
  
} // UserIDImpl

