//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBase;


import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.*;

import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Map;
import java.util.Collections;
import java.util.HashMap;
import java.util.Collection;
//import java.util.SortedMap;

import java.beans.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import ARoad0.Pattern.BeanInfoPattern;
import ARoad0.gWorkInterface.CoreAlgorithm;

 /**
  * This class is an utility class for BaseObjects, StringRights, the BeanInfo and the AcsAddon
  * management. It cannot be subclassed in an AcsAddon. As a CoreAlgorithm,
  * each worker thread of a GraphicView uses a dedicated instance of this class.
  * It applies the ACS rules: <P>
  * - 'AGO right: no denying rights' <P>
  * - 'bridge right: no denying rights' <P>
  * - 'sens metarule: ACS denyings are superior to ACS grantings
  * (there are + and - operations), for each right type' <P>
  * - 'right type order: default order: root, A in AGO, all other right types' <P>
  * It is used by gBase and the AcsAddon gBase packages.
  * It also implements some core algorithms that gWork uses in the access paths search. <P>
  * ARoad0.gWork.RightsMediatorImpl uses this class to create view instances
  * in the view worker thread, while the first instance is used by the gBase consumers
  * in the main thread. <P>
  * The direct calls to ACS in the core algorithms are for
  * the methods getL_Rights(), getFirstRightFromMeta(), getEorM_EligibleParties(),
  * getEorL_VirtualFolders(), getEorL_Resources().
  * The direct calls to BaseManagerImpl in the core algorithms are for the method
  * getMetaRight() and in getBaseObject(), isOpenACS(), isOpenView(), getMetaRightFromName()
  * getEorM_OpenACS(). <P>
  * The methods have to be fast in the access paths search. Some private sets
  * are created at the instance construction to avoid the creation of sets
  * at each method call. Nonetheless, this class is stateless, since each method
  * clears all its sets at its starting. <P>
  * For the event management of the ACS Addons, the 4 ACSObjectAddon methods
  * add/removeAddonInnerPropertyChangeListener() and
  * add/removeAddonParentPropertyChangeListener() allow the views to listen
  * the specific events from their nodes, when they belong to an AcsAddon.
  */
public final class UtilityImpl implements CoreAlgorithm {
      /**
       * the first instance of the class, used by gBase and the GUI,
       * not by the GraphicView worker threads
       */
  private static final UtilityImpl  instance__ = new UtilityImpl();

  private static final int          INITIAL_CAPACITY = 10; // the default value: 101
  
  /**
   * used to speed up subtractNotRemovingCommonLowerNodes() and
   * subtractInOneSetInOneACS()
   */
  private final Set<StringRight>  l_test_ = new HashSet(INITIAL_CAPACITY*3);
  /**
   * used to speed up subtractNotRemovingCommonLowerNodes()
   */
  private final Set<StringRight>  l_to_next_test_ = new HashSet(INITIAL_CAPACITY*3);
  /**
   * used to speed up subtractNegativeHeadersNotRemovingInOneACS()
   */
  private final Set<StringRight>  l_to_retain_ = new HashSet(INITIAL_CAPACITY*3);
  /**
   * used to speed up selectHeaders() and subtractNotRemovingCommonLowerNodes()
   */
  private final Set<StringRight>  l_removed_ = new HashSet(INITIAL_CAPACITY*3);
  /**
   * used to speed up selectHeaders()
   */
  private final Set<StringRight>  l_headers_ = new HashSet(INITIAL_CAPACITY*3);
  /**
   * used to speed up getFirstIncreasingRightsChain()
   */
  private final LinkedList        l_methodResult_ = new LinkedList();
  private final Set               l_rightsUpFromCurrent_ = new HashSet(INITIAL_CAPACITY*3);
  private final Set               l_PathsUpFromLesser_ = new HashSet(INITIAL_CAPACITY*3);
  private final Set               l_toAddTol_PathsUpFromLesser_ = new HashSet(INITIAL_CAPACITY*3);
  
  /** 
  * Called by gWork.RightsMediatorImpl to be used in the GraphicView worker thread.
  */
  public UtilityImpl() {
  }
  
    /**
    * Method to obtain the first instance of UtilityImpl, for the gBase
    * consumers and generally the main thread and the event dispatch thread.
    * The GraphicView worker threads use dedicated instances set by gWork.RighsMediatorImpl.
    * @return ACSFactoryUtilityImpl
    */
  public static UtilityImpl getFirstInstance() {
    return instance__;
  } // getInstance()
  
 /**
  * Gets a StringRights increasing chain where the two arguments are at the two ends,
  * if the chain exists. This method has to be fast because it is used in
  * subtractNotRemovingCommonLowerNodes(), withExecuteRight(), selectHeaders(),
  * withDenyExecuteRight() and withDirectExecuteRight().
  * If there is a positive right and a negative one, or if compareTo() returns 0,
  * the result is null.
  * The lesser right is the first item in the returned LinkedList. 
  * It is an increasing chain, so rights are more powerfull when the index increments.
  * Return the first increasing chain which is found, but it may exist several chains
  * in the rights network between the two right arguments.
  * Return null if the comparison (compareTo method) returns 0, 
  * if the arguments have not the same sens, or if there is no result.
  * This method does not use a double up-down search, because rights have more
  * lower rights than upper rights, and so, a single down-up search is generally faster.
  * Acs rights have generally a right network much more simple than metarights,
  * so the result is faster for acs rights.
  * Metarights and acsrights have never any chain linking them, and the result
  * null is quite fast to found since acsrights are inferior and the search begins
  * from them without exploring the metarights network. It the two arguments are 
  * ACS rights from more than one ACS, the method returns null.
  * A non-null _upperToAvoid is used by subtractNotRemovingCommonLowerNodes().
  * Note: to ensure performance, if there are more than 20 increasing chains
  * to analyze from the lesser right, an exception is thrown. This should never be,
  * and it indicates that the design of rights is very poor. 
  * @param _rightOne is one end of the chain of rights to get. Not null.
  * @param _rightTwo is the second end of the chain of rights to get. Not null.
  * @param _upperToAvoid a direct upper right of the lesser right among the two arguments.
  *          No action if it is not a direct upper right of the lesser right, or
  *          if it is the greater right among the two arguments.
  *          The result will never include this right. May be null. 
  * @exception BaseError if an argument is null,
  *    if there are more than 20 increasing chains to analyze from the lesser right
  * @return a chain of linked increasing rights, from the index 0. 
  *   May be null, or its minimal lenght is 2.
  */
  private final LinkedList getFirstIncreasingRightsChain(
          StringRight _rightOne, StringRight _rightTwo, StringRight _upperToAvoid) 
          throws BaseError {

      if (_rightOne == null || _rightTwo == null)
          throw new BaseError("in UtilityImpl.getFirstIncreasingRightsChain(),"+
                  " one of the two arguments is null.");      

      if (_rightOne.isGrantingRight() && !_rightTwo.isGrantingRight()) return null;
      if (!_rightOne.isGrantingRight() && _rightTwo.isGrantingRight()) return null;
      if (_rightOne.compareTo(_rightTwo) == 0) return null;
      if (_rightOne.isDesactivated() || _rightTwo.isDesactivated()) return null; 
      if (!_rightOne.getAcsName().isEmpty() && !_rightTwo.getAcsName().isEmpty() &&
              !_rightOne.getAcsName().equals( _rightTwo.getAcsName() ) ) return null;
      
      // initialization
      final int         PATHS_SIZE_LIMIT = 20;
      StringRight       lesser = null;
      StringRight       greater = null;
      l_methodResult_.clear();
      l_rightsUpFromCurrent_.clear();
       
      if (_rightOne.compareTo(_rightTwo) < 0) {
          l_methodResult_.add(0, _rightOne);
          lesser = _rightOne;
          l_rightsUpFromCurrent_.addAll( Arrays.asList(_rightOne.getUpperRights()) );
          greater = _rightTwo;
      } else {
          l_methodResult_.add(0, _rightTwo);
          lesser = _rightTwo;
          l_rightsUpFromCurrent_.addAll( Arrays.asList(_rightTwo.getUpperRights()) );
          greater = _rightOne;
      } // else
      
      // to be faster if 'greater' is a direct 'lesser' upper right
      if (l_rightsUpFromCurrent_.contains(greater)) {
          l_rightsUpFromCurrent_.clear();
          lesser = null;
          l_methodResult_.addLast(greater);
          greater = null;          
          return l_methodResult_;
      } // if
      
      // avoid to go through _upperToAvoid
      l_rightsUpFromCurrent_.remove(_upperToAvoid);
      
      // initialization
        // last item in l_currentList
      StringRight   current = null;
      boolean       detect = false;
      boolean       noNewPath = false;
        // sets of LinkedList, with a list for each path of raising rights
      l_PathsUpFromLesser_.clear();
      l_PathsUpFromLesser_.add(l_methodResult_);
        // contains the new paths built up from the new detected upper rigths
      l_toAddTol_PathsUpFromLesser_.clear();
      LinkedList    mainList = null;
      LinkedList    list = null;
      StringRight   right = null;
        // to control the number of search paths
      int           pathCount = 0;
      boolean       tooLargeNumberOfPaths = false;
      
      while ( !detect && !noNewPath && !tooLargeNumberOfPaths ) {

          l_PathsUpFromLesser_.addAll( l_toAddTol_PathsUpFromLesser_ );
          l_toAddTol_PathsUpFromLesser_.clear();
          
          for (Iterator itu = l_PathsUpFromLesser_.iterator(); itu.hasNext(); ) {

              mainList = (LinkedList) itu.next();
              current = (StringRight) mainList.getLast(); // never empty?
              l_rightsUpFromCurrent_.clear();
              l_rightsUpFromCurrent_.addAll( Arrays.asList(current.getUpperRights()) );
              
              if (l_rightsUpFromCurrent_.contains(greater)) {
                  detect = true;
                  break;
              } // if
              else if (l_rightsUpFromCurrent_.isEmpty()) {
                  // l_currentList is ending
                  itu.remove();
                  pathCount--;
                  continue;
              } // else
              else {
                  
                  for (Iterator ite = l_rightsUpFromCurrent_.iterator(); ite.hasNext(); ) {
                     right = (StringRight) ite.next();
                     // removes an upper right if stronger than greater or desactivated
                     if (right.compareTo(greater) >= 0 || right.isDesactivated()) 
                         ite.remove();
                  } // for ite
                  
                  if (l_rightsUpFromCurrent_.isEmpty()) {
                     // l_currentList last right has only stronger upper rights
                     itu.remove();
                     pathCount--;
                     continue;
                  } // if
                  
                  // Adds the remaining paths with the correct upper rights
                  // in l_PathsUpFromLesser
                  for (Iterator ite = l_rightsUpFromCurrent_.iterator(); ite.hasNext(); ) {
                     right = (StringRight) ite.next();
                     list = (LinkedList)mainList.clone();
                     list.addLast(right);
                     l_toAddTol_PathsUpFromLesser_.add( list );
                     pathCount++;
                  } // for ite
                  
                  // l_currentList has to be remove from l_PathsUpFromLesser
                  itu.remove();
                  pathCount--;
              
              } // else

          } // for itu
          
          if (l_PathsUpFromLesser_.isEmpty() && l_toAddTol_PathsUpFromLesser_.isEmpty())
              noNewPath = true;
          
          if (pathCount > PATHS_SIZE_LIMIT)
              tooLargeNumberOfPaths = true;

      } // while !detect

      current = null;
      list = null;
      right = null;      

      if (tooLargeNumberOfPaths) {
          l_methodResult_.clear();
          greater = null;
          throw new BaseError(
            "in UtilityImpl.getFirstIncreasingRightsChain(), "+
            "there are too many chains to analyze from the lesser right "+lesser);
      }
      else if (detect) {
          l_methodResult_.clear();
          l_methodResult_.addAll(mainList);
          l_methodResult_.addLast(greater);
          greater = null;
          lesser = null;
          mainList = null;          
          return l_methodResult_;
      }
      else {
          l_methodResult_.clear();
          greater = null;
          lesser = null;
          return null;
      }
  } // getFirstIncreasingRightsChain()
  
  /**
  * Wrapper and caller of subtractNegativeNotRemovingInOneACS().
  * Caution: this method updates its argument by the removing of negative rights,
  * so the recommanded use is 'set = substractNegativeFromPositiveRights(set)' to
  * avoid any error on the 'set' variable.
  * Called by RightsFactory_Facade.getEffectiveRightsForPair() and
  * getFirstShapeEffectiveRights(), 
  * DisplayablelinkUtilities.getDLinksFromNegativeACLPaths(),
  * AclRightsFactoryImpl.getL_aclEntryRights().
  * @param _l_upd_rights to analyze. Not null. Updated by the method.
  * @return an all-positive or all-negative rights set.
  *     May be empty but not null.
  **/
  public final Set<StringRight> substractNegativeFromPositiveRights(
          Set<StringRight> _l_upd_rights) {
    // variables
    Set<StringRight>            l_result = null;
    Set<StringRight>            l_posRights  = null;
    Set<StringRight>            l_negRights = new HashSet(INITIAL_CAPACITY);    
    StringRight                 right = null;

                        // for testing and tracking    
                        //System.out.println("");
                        //System.out.println(
                                //"UtilityImpl.substractNegativeFromPositiveRights() start.");

    for (Iterator ito = _l_upd_rights.iterator(); ito.hasNext(); ) {
        right = (StringRight)ito.next();
        if (!right.isGrantingRight()) {
            l_negRights.add(right);
            ito.remove();
        } // if
    } // for ito
        
                      // for testing and tracking
                      //System.out.println("-- all the positive rights before subtract(): ");
                      //for (Iterator ite = _l_upd_rights.iterator(); ite.hasNext(); ) {
                          //System.out.println("----- "+ite.next());
                      //}
                      //System.out.println("-- all the negative rights before subtract(): ");
                      //for (Iterator ite = l_negRights.iterator(); ite.hasNext(); ) {
                          //System.out.println("----- "+ite.next());
                      //}

    // if no positive right, the negative rights are returned
    if (_l_upd_rights.isEmpty() && !l_negRights.isEmpty()) {
        right = null;
        return l_negRights;
    } // if    
    
    // no acs, so generic rights may be returned
    // no node, since the inheritance rights on node are not relevant
    l_posRights = subtractNegativeNotRemovingInOneACS(
            _l_upd_rights, l_negRights, null, false);

    // if l_posRights is empty, the negative rights are stronger
    if (l_posRights.isEmpty() && !l_negRights.isEmpty()) {
        l_result = l_negRights;
    } // if
    else
        l_result = l_posRights;

                      // for testing and tracking
                      //System.out.println("----- resulting rights: ");
                      //for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                          //System.out.println("----- "+ite.next());
                      //}
   
    l_posRights  = null;
    l_negRights = null; 
    right = null;
    
    return l_result;
  } // substractNegativeFromPositiveRights()
  
 /**
  * Substracts the second right to the first right. This operation
  * returns the lower rights which remain applicable when _rightOne is reduced
  * by the opposite of _rightTwo in the same rights set. <P>
  * _rightTwo and all the upper rights of _rightTwo are removed in _rightOne.
  * For the lower rights, if a _rightOne (direct or indirect) lower right, say R,  
  * is also a _rightTwo (direct or indirect) lower right, the common node R 
  * is put in the result.<P>
  * Returns, if any, all the lower rights of _rightOne if it is equal to _rightTwo.
  * If _rightOne is not equal to _rightTwo, return _rightOne if 
  * getFirstIncreasingRightsChain(_rightOne, _rightTwo) is null or starts with _rightOne.
  * If _rightOne and _rightTwo are acs rights from different ACS, returns _rightOne.
  * Called by subtractNegativeHeadersNotRemovingInOneACS(), so
  * this method has to be fast.
  * Calls getFirstIncreasingRightsChain(). Calls withAssociatedDirectoryRight()
  * to detect the 'directory' rights, that have 'read', 'write' or 'execute' as
  * metaright.
  * Does not call selectHeaders(), so l_removed_ may be used by the two methods
  * independantly.
  * Uses the instance properties l_test_, l_to_next_test_ and l_removed_.
  * @see #subtractNegativeHeadersNotRemovingInOneACS
  * @param _rightOne first right. Not null.
  * @param _rightTwo second right to subtract. Must have the same class and the same sens 
  *        that _rightOne. Not null.
  * @param _acs uses all the rights in the two first arguments. May be null.
  * @param _forNode true if these rights are for a Node (Directory or VirtualFolder);
  *         the 'directory' rights are not removed, and their equivallent like 'read' is removed
  * @exception BaseError if an argument is null,
  *    if there is a positive right and a negative one, or an acsright and a metaright,
  *    or if there are more than 20 increasing chains to analyze from the lesser right
  * @return a subset of the _rightOne direct and indirect lower rights, or _rightOne
  *     if _rightTwo is totally independant. May be empty but never null. All the items
  *     in the result are acsrights of _acs or metarights used by _acs.
  */
  public final Set<StringRight> subtractNotRemovingCommonLowerNodes(
          StringRight _rightOne, StringRight _rightTwo, ImmutableACS _acs,
          boolean _forNode) throws BaseError {
 
      if (_rightOne == null || _rightTwo == null)
          throw new BaseError("in UtilityImpl.subtractNotRemovingCommonLowerNodes(),"+
                  " one of the two arguments is null.");      
      
      if ((_rightOne.isGrantingRight() && !_rightTwo.isGrantingRight()) ||
              (!_rightOne.isGrantingRight() && _rightTwo.isGrantingRight()) ||
              (_rightOne instanceof StringMetaRight && !(_rightTwo instanceof StringMetaRight)) ||
              (!(_rightOne instanceof StringMetaRight) && _rightTwo instanceof StringMetaRight)) {
                  throw new BaseError("in UtilityImpl.subtractNotRemovingCommonLowerNodes(),"+
                          " the two arguments have not the same sens or the same class."); 
      } // if

      // variables
      Set<StringRight> l_result = new HashSet(INITIAL_CAPACITY*3);
      StringRight[]    l_lower = null;
      
      if (_rightOne.equals(_rightTwo)) {
          l_lower = _rightOne.getLowerRights();
          
          if (l_lower.length != 0)
              l_result.addAll( Arrays.asList( l_lower ));

 
                  // for testing and tracking
                  //System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes(), "+
                    //      "the result is all the _rightOne lower rights, if not empty: ");
                  //for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                    //  System.out.println("--- "+(StringRight)ite.next());
                  //} // for i       
          
          l_lower = null;
          return l_result;
      } // if (_rightOne.equals(_rightTwo))
      
      // caution: chain is cleared at the next call to this method
      LinkedList chain = getFirstIncreasingRightsChain(_rightOne, _rightTwo, null);
      if (chain == null) {
          l_result.add(_rightOne);
          l_lower = null;
          return l_result;           
      }
      if (chain.getFirst() == _rightOne) {
 
              // for testing and tracking
              //System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes(), the result is empty "+
                //          "since _rightOne "+_rightOne+" is lesser than "+_rightTwo);  
          
          l_result.add(_rightOne);
          l_lower = null;
          chain = null;
          return l_result; 
      }
      if (_rightOne.isDesactivated()) {
          l_lower = null;
          chain = null;
          return l_result;
      }
      if (_rightTwo.isDesactivated()) {
          l_result.add(_rightOne);
          l_lower = null;
          chain = null;
          return l_result;
      } // if
      
      l_test_.clear();
      l_to_next_test_.clear();
      l_removed_.clear();
      StringRight       testedRight = null;
      boolean           testedRightToRemove = false;
      
      
      l_test_.add(_rightOne);
      
      while (!l_test_.isEmpty()) {
          // goes down in all the lower rights network to select the rights without
          // any lower link to __rightTwo
          for (Iterator ite = l_test_.iterator(); ite.hasNext();) {
            testedRight = (StringRight)ite.next();
            if (l_removed_.contains(testedRight)) continue;
 
                                  // for testing and tracking
                                  //if (_rightOne.getRight().equals("ctrl") && _rightTwo.getRight().equals("w"))
                                    //    System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes(), search of "+
                                      //    "lower rights of "+testedRight);
            
            l_lower = testedRight.getLowerRights();
            
            for (int i = 0; i < l_lower.length; i++) {
              if (l_lower[i].equals(_rightTwo) || l_lower[i].isDesactivated()) continue;
              if (!_forNode && l_lower[i].getEorMetaRight().getRight().endsWith("directory")) 
                  continue;
              if (_forNode && withAssociatedDirectoryRight(l_lower[i])) 
                  continue;
              // caution: chain is cleared only at the next call to this method
              chain = getFirstIncreasingRightsChain(l_lower[i], _rightTwo, testedRight);
              if (chain == null ) {
                  // l_lower[i] has no link to _rightTwo
                  l_result.add(l_lower[i]);
                  l_to_next_test_.add(l_lower[i]);
 
                      // for testing and tracking
                      //System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes(), add in "+
                        //      "l_result before optional removing at the next iteration: "+
                          //    l_lower[i]);
              
              } // if
              else if (chain.getFirst() == _rightTwo) {
                  // l_lower[i] is greater than _rightTwo
                  l_to_next_test_.add(l_lower[i]);
                  testedRightToRemove = true;
 
                      // for testing and tracking
                      //System.out.println("StringRightImpl.subtractNotRemovingCommonLowerNodes(), add in "+
                        //      "l_toNextTest of "+l_lower[i]+" and removing for: "+
                          //    testedRight);
              } // if
              // the lesser right is l_lower[i], then noop
              //else {
 
                              // for testing and tracking
                              //if (_rightOne.getRight().equals("ctrl") && _rightTwo.getRight().equals("w"))
                                //System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes(), "+
                                  //    "removing for: "+testedRight);
                      
                //  testedRightToRemove = true;
              //}  
            }// for i on l_lower
            
            // to keep in l_result only rights which have all their applicable 
            // lower rights in l_result - 'directory' rights are applicable only to Nodes.
            if (testedRightToRemove && l_result.contains(testedRight)) {
                l_removed_.add(testedRight);
                l_result.remove(testedRight);

                              // for testing and tracking
                              //if (_rightOne.getRight().equals("ctrl") && _rightTwo.getRight().equals("w"))
                                //System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes(), "+
                                  //    "removing of "+testedRight);
                
            } // if (testedRightToRemove)
            
            testedRightToRemove = false;
            
          } // for ite on l_test
          
          l_test_.clear();
          l_test_.addAll(l_to_next_test_);
          l_to_next_test_.clear();
          
      } // while
 
                          // for testing and tracking
                          //if (_rightOne.getRight().equals("ctrl") && _rightTwo.getRight().equals("w")) {
                            //  System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes() at the end, BEFORE "+
                              //        "selection from ACS:");
                              //System.out.println("--------- rights are:");
                              //for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                                //      System.out.println(" ------------------------ "+ite.next());
                              //} // for
                          //} // if
       
      if (_acs != null) {
          List l_inAcs = Arrays.asList( _acs.getL_Rights() );
          for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
              if (!l_inAcs.contains(ite.next())) ite.remove();
          } // for ite
          l_inAcs = null;
      } // if
  
              // for testing and tracking
              //System.out.println("UtilityImpl.subtractNotRemovingCommonLowerNodes() at the end,"+
                //      "subtract "+_rightTwo+" from "+_rightOne);
              //System.out.println("--------- rights are:");
              //for (Iterator ita = l_result.iterator(); ita.hasNext(); ) {
                      // for testing and tracking
                //     System.out.println(" ---------- "+ita.next());
              //} // for
      
      chain = null;      
      testedRight = null;
      l_lower = null;
      testedRightToRemove = false; 
      
      // does not remove the lesser rights through a call to selectHeaders()
      return l_result;
  } // subtractNotRemovingCommonLowerNodes()
  
 /**
  * Selects the stronguest rights, in removing all the rights in the argument
  * which have a direct or indirect upper rights in the argument.
  * Acsrights and metarights, granting and denying rights, acsrights from different acs 
  * may be mixed, but each of these four categories will generate at most one header 
  * in the result.
  * Called by ResourceImpl, PrivilegeAbst, AclEntryImpl,
  * RightsFactory_Facade.getEffectiveRightsForPair(),
  * DisplayableLinkUtilities.getEffectiveRightsAndCommentOnRights(),
  * subtractNegativeNotRemovingInOneACS(), subtractInOneSetInOneACS(),
  * subtractNegativeHeadersNotRemovingInOneACS().
  * Calls getFirstIncreasingRightsChain().
  * @param _l_updatedRights metarights and acsrights to filter, which may be granting
  *    or denying ones. Not null and no null item. Is updated by this method.
  * @exception BaseError if an item is null in the argument,
  *    if there are more than 20 increasing chains to analyze from the lesser right
  */
  public final void selectHeaders(
          Set<StringRight> _l_updatedRights) throws BaseError {

        l_headers_.clear();
        l_removed_.clear();
        if (_l_updatedRights.isEmpty()) return;
        
        StringRight         right = null;
        StringRight         right2 = null; 
        LinkedList          l_chain = null;
        boolean             presentUnderHead = false;
        boolean             removeHead = false;
   
                                  // for testing and tracking
                                  //System.out.println(" ");
                                  //System.out.println("Start of UtilityImpl.selectHeaders()");
                                  //System.out.println("--------- rights are:");
                                  //for (Iterator ita = _l_updatedRights.iterator(); ita.hasNext(); ) {
                                           //for testing and tracking
                                    //      System.out.println(" ------------------------ "+ita.next());
                                  //} //for
        
      
        for (Iterator ita = _l_updatedRights.iterator(); ita.hasNext(); ) {
            right2 = (StringRight)ita.next();
            if (right2.isDesactivated()) continue;

                              // for testing and tracking
                              //System.out.println("UtilityImpl.selectHeaders(), "+
                                //      "analyze ---------- "+right2);
            

                
            for (Iterator ite = l_headers_.iterator(); ite.hasNext(); ) {
                // cannot be right2, so direct call to getFirstIncreasingRightsChain()
                right = (StringRight)ite.next();
                // caution: l_chain is cleared at the next call to this method
                l_chain = getFirstIncreasingRightsChain(right2, right, null);
                if (l_chain != null) {
                    
                              // for testing and tracking
                              //System.out.println("UtilityImpl.selectHeaders(), "+
                                //      "l_chain: "+l_chain);
                    
                    if (l_chain.getLast() == right) {
                        presentUnderHead = true;

                              // for testing and tracking
                              //if (right2.getEorMetaRight().getRight().equals("deleteRenameCopy"))
                              //System.out.println("UtilityImpl.selectHeaders(), "+
                                //  "no-added lesser right: "+right2+", due to greater right "+right);

                        break;
                    } else {
                        removeHead = true;
                        l_removed_.add(right);
                    }
                } // if (l_chain != null)
            } // ite

            if (!presentUnderHead) { 
                // right2 is a new 'head' of the rights network
                l_headers_.add(right2);

                          // for testing and tracking
                          //System.out.println("UtilityImpl.selectHeaders(), "+
                            //      "add to result: "+right2);

                if (removeHead) {

                          // for testing and tracking                         
                          //System.out.println("UtilityImpl.selectHeaders(), "+
                            //          "remove from result: "+l_removed_+", due to greater right "+right2);

                    l_headers_.removeAll(l_removed_);
                    l_removed_.clear();
                } // if            
            } // if (!present  
            removeHead = false;
            presentUnderHead = false;

        } // for ita 

                                      // for testing and tracking
                                      //System.out.println("UtilityImpl.selectHeaders() ");
                                      //System.out.println("--------- l_result contains:");
                                      //for (Iterator ita = l_headers_.iterator(); ita.hasNext(); ) {
                                               //for testing and tracking
                                        //      System.out.println(" ------------------------ "+ita.next());
                                      //} //for
                                      //System.out.println(" ");
              
        right = null;
        right2 = null; 
        l_chain = null;
        _l_updatedRights.clear();
        _l_updatedRights.addAll(l_headers_);
  } // selectHeaders()
  
  /**
   * Tests if the metaright of the argument is 'goto', 'read', 'write' or 'execute'.
  * Called by subtractNotRemovingCommonLowerNodes().
  * @param _right acsright or metaright to analyze
  * @return true only if _right metaright is 'read', 'write' or 'execute'.
  * */
  public final boolean withAssociatedDirectoryRight(StringRight _right) {
    if (_right.getEorMetaRight().getRight().equals("goto")) return true;
    if (_right.getEorMetaRight().getRight().equals("read")) return true;
    if (_right.getEorMetaRight().getRight().equals("write")) return true;
    if (_right.getEorMetaRight().getRight().equals("execute")) return true;    
    else return false;
  } // withAssociatedDirectoryRight()

  /**
  * Detects the rights which are directly or indirectly linked to the metaright 'execute',
  * as 'read & execute' for instance.
  * If the argument contains some negative rights, this method subtracts
  * them to the positive rights.
  * Called by ACSImpl, ActorImpl, AclRightsFactoryImpl and CompoundRightsFactoryImpl.
  * Calls subtractInOneSetInOneACS() and getFirstIncreasingRightsChain().
  * @param _l_rights to analyze. May be null or empty.
  * @param _forNode true if these rights are for a node, and so the 'directory'
  *         rights are not removed
  * @return true only if the rights contain an 'execute' right
  **/
  public final boolean withExecuteRight(
          StringRight[] _l_rights, boolean _forNode) {

                      //for testing and tracking
                      //System.out.println("call to UtilityImpl.withExecuteRight()");
                        //"processing of right: "+right);
      
      if (_l_rights == null || _l_rights.length == 0) {

                      //for testing and tracking
                      //System.out.println("UtilityImpl.withExecuteRight(), "+
                        //"with null/empty rights arguments");

          return false;
      }
      
      boolean           result = false;
      StringRight       right;
      StringRight       execute = BaseManagerImpl.getInstance().getMetaRight("execute");
      LinkedList        list = null;
      // tests executing right after a subtraction of the negative rights
      Set<StringRight>  set = subtractInOneSetInOneACS(_l_rights, null, _forNode);
      
      // test first the easy cases
      for (Iterator ite = set.iterator(); ite.hasNext(); ) {

            right = ((StringRight)ite.next()).getEorMetaRight();
            if (right == null) continue;
            if (right.equals( execute )) {
                    result = true;
                    break;
            } // if

      } // for ite

      if (result) {
          right = null;
          execute = null;
          set = null;

          return result;
      } // if

      try {
          // test the chain of rights
          // is not limited by an acsrights set
          for (Iterator ite = set.iterator(); ite.hasNext(); ) {

                right = ((StringRight)ite.next()).getEorMetaRight();

                      //for testing and tracking
                      //System.out.println("in UtilityImpl.withExecuteRight(), "+
                        //"processing of right: "+right);

                // caution: list is cleared at the next call to getFirstIncreasingRightsChain()
                list = getFirstIncreasingRightsChain(right, execute, null);

                      //for testing and tracking
                      //if (list != null) {
                        //  System.out.println("--- list from getFirstIncreasingRightsChain(): ");
                          //for (Object obj : list) {
                            //  System.out.println("----- "+obj);
                          //} // for obj
                      //} // if (list


                if (list != null && list.getFirst().equals(execute)) {
                    result = true;
                    break;
                } // if
          } // for ite
        } catch (BaseError b) {
            throw new InternalError("UtilityImpl.withExecuteRight(): "+b.toString());
        }
      
      right = null;
      execute = null;
      list = null;
      set = null;
      
      return result;
  } // withExecuteRight()

  /**
  * Detects the rights which are directly or indirectly linked to the metaright 'deny_execute'.
  * Calls getFirstIncreasingRightsChain() and BaseManagerImpl.getInstance().getMetaRight().
  * @param _l_rights to analyze
  * @return true only if the rights contain one 'deny_execute' right, or contain
  *   a right to modify the current rights for having a deny_execution right.
  * */
  public final boolean withDenyExecuteRight(StringRight[] _l_rights) {
      if (_l_rights == null) return false;
      boolean result = false;
      StringRight right;
      StringRight execute = BaseManagerImpl.getInstance().getMetaRight("deny_execute");
      LinkedList list = null;
      
      for (int k = 0; k < _l_rights.length; k++) {
        if (_l_rights[k] == null) continue;
        right = (StringRight)_l_rights[k].getEorMetaRight();
        
        if (right.equals( execute )) {
                result = true;
                break;
        }
        try {
            // caution: list is cleared at the next call to getFirstIncreasingRightsChain()
            list = getFirstIncreasingRightsChain(right, execute, null);
        } catch (BaseError b) {
            throw new InternalError("from UtilityImpl.withDenyExecuteRight(): "+b);
        }
        if (list != null && list.getFirst().equals(execute)) { 
            result = true;
            break;
        } // if 
        
        list = null;
      } // for k
      
      right = null;
      execute = null;
      list = null;
      return result;
  } // withDenyExecuteRight()
  
  /**
  * Detects the rights which are directly linked to the metaright 'execute',
  * as 'read & execute' for instance. False if it contains '|modify_security_descriptor|',
  * since it is an indirect access to the 'execute' right.
  * Calls getFirstIncreasingRightsChain() and BaseManagerImpl.getInstance().getMetaRight().
  * @param _l_rights to analyze
  * @return true only if the argument contains an 'execute' right.
  * */
  public final boolean withDirectExecuteRight(StringRight[] _l_rights) {
      if (_l_rights == null) {

                      //for testing and tracking
                      //System.out.println("UtilityImpl.withDirectExecuteRight(), "+
                        //"with null rights arguments");

          return false;
      }
      boolean result = false;
      StringRight right;
      StringRight execute = BaseManagerImpl.getInstance().getMetaRight("execute");
      StringRight modify_security_descriptor = 
              BaseManagerImpl.getInstance().getMetaRight("|modify_security_descriptor|");
      LinkedList list = null;

                      //for testing and tracking
                      //System.out.println("UtilityImpl.withDirectExecuteRight()"+
                        //" number of processed rights: "+_l_rights.length);
      
      for (int k = 0; k < _l_rights.length; k++) {

                      //for testing and tracking
                      //System.out.println("UtilityImpl.withDirectExecuteRight()"+
                        //" processed right: "+_l_rights[k]);

        if (_l_rights[k] == null) continue;
        right = (StringRight)_l_rights[k].getEorMetaRight();
        
        if (right.equals( execute )) {
                result = true;
                break;
        }
        try {
            // caution: list is cleared at the next call to getFirstIncreasingRightsChain()
            list = getFirstIncreasingRightsChain(right, execute, null);
        } catch (BaseError b) {
            throw new InternalError("from UtilityImpl.withDirectExecuteRight(): "+b);
        }
        // inferior to modify_security_descriptor to avoid indirect access to the 'execute' right
        if (list != null && list.getFirst().equals(execute) &&
                right.compareTo(modify_security_descriptor) < 0) { 
            result = true;
            break;
        } // if 
        
        list = null;
      } // for k
 
                      //for testing and tracking
                      //System.out.println("UtilityImpl.withDirectExecuteRight(), "+
                        //"result: "+result);

      right = null;
      execute = null;
      modify_security_descriptor = null;
      list = null;
      return result;
  } // withDirectExecuteRight()
  
  /**
  * Detect the inverse right, if any, that it is defined, if _right is positive,
  * as the negative right which as the same semantic and strenght. Use the pattern
  * of name 'xxx' and 'deny_xxx'. 'goto' and 'deny_all'
  * are not in this pattern, but they are handled too. But do not process
  * 'full_controlchild' and 'full_controlnxchild', which are desactivated. For an acsright,
  * the return value is the first right to be associated to the correct metaright.
  * Called by subtractNegativeFromPositiveInOneACS().
  * @param _right acsright or metaright to analyze
  * @return the inverse right, or null.
  * */
  public final StringRight withInverseRight(StringRight _right) {
    if (_right == null) return null;
    ACS acs = null;
    StringRight result = null;

    if (!_right.isGrantingRight() && _right.getEorMetaRight().getRight().startsWith("deny_") ) {
        if (_right instanceof StringMetaRight) {
            if (_right.getRight().equals("deny_all"))
                result = BaseManagerImpl.getInstance().getMetaRight( "goto" );
            else result = BaseManagerImpl.getInstance().getMetaRight(
                _right.getEorMetaRight().getRight().substring(5) );
        } else {
            acs = (ACS)((StringRightImpl)_right).getEorACS();
            if (_right.getEorMetaRight().getRight().equals("deny_all"))
                result = acs.getFirstRightFromMeta( "goto" );
            else result = acs.getFirstRightFromMeta(
                    BaseManagerImpl.getInstance().getMetaRight(
                _right.getEorMetaRight().getRight().substring(5) ).getRight() );            
        }
    }
    else {
        if (_right instanceof StringMetaRight)
            if (_right.getRight().equals("goto"))
                result = BaseManagerImpl.getInstance().getMetaRight( "deny_all" );
            else result = BaseManagerImpl.getInstance().getMetaRight("deny_"+
                _right.getEorMetaRight().getRight() );
        else {
            acs = (ACS)((StringRightImpl)_right).getEorACS();
            if (_right.getEorMetaRight().getRight().equals("goto"))
                result = acs.getFirstRightFromMeta( "deny_all" );
            else result = acs.getFirstRightFromMeta(
                    BaseManagerImpl.getInstance().getMetaRight("deny_"+
                _right.getEorMetaRight().getRight() ).getRight() );
        }
    } // else

                      //for testing and tracking
                      //System.out.println("in RightsFactoryUtilities.withInverseRight(), "+
                      //  "result: "+result);

    return result;
  } // withInverseRight() 
  
  /**
  * Operates the reduction of each positive right by all the negative rights,
  * afer a filtering of the headers for each sens.
  * Each negative right is subtracted to each positive right, and the positive results,
  * if any for each positive right, are put in the returned set.
  * Each negative right and all its upper rights are removed in each positive right.
  * For the lower rights, if a positive right (direct or indirect) lower right, say R,
  * is also a negative right (direct or indirect) lower right,
  * this common node R is put in the result.
  * If a negative right is stronger than all positive rights, the returned set is empty,
  * and the calling method could use _l_headNegRights as the resulting right set if necessary.
  * If _acs is null, the result is not limited by the _acs rights set.
  * Called by ResourceImpl/DirectoryImpl.setL_xxxRights(),
  * their constructors to process the umasks,
  * AclRightsFactoryImpl.getL_aclEntryRights(),
  * DisplayableLinkUtilities.getDLinksFromNegativeACLPaths() and
  * getEffectiveRightsAndComment(), substractNegativeFromPositiveRights().
  * Wrapper and caller of subtractNegativeHeadersNotRemovingInOneACS().
  * @see UtilityImpl#selectHeaders
  * @param _l_posRights positive rights in an HashSet. Not null and no null item.
  *     No update.
  * @param _l_negRights negative rights in an HashSet. Not null and no null item.
  *     No update.
  * @param _acs uses all the rights in the two right sets. May be null.
  * @param _forNode true if these rights are for a Node (Directory or VirtualFolder):
  *         the 'directory' rights are not removed if applicable
  * @return the set of positive rights are all headers, and are used by _acs
  *         if not null. May be empty, not null.
  **/
  public final Set<StringRight> subtractNegativeNotRemovingInOneACS(
           Set<StringRight> _l_posRights, Set<StringRight> _l_negRights, 
           ImmutableACS _acs, boolean _forNode) {
       
    Set l_headPosRights = null;
    Set l_headNegRights = null;
    
    try {          
        // select the headers in each set
            // clone() since the argument is updated
        l_headPosRights = (Set<StringRight>) ((HashSet)_l_posRights).clone();
        
                              // for testing and tracking
                              //System.out.println("UtilityImpl.subtractNegativeNotRemovingInOneACS() "+
                                //      "l_headPosRights after the clone:");
                              //for (Iterator ite = l_headPosRights.iterator(); ite.hasNext(); ) {
                                      // for testing and tracking
                                //      System.out.println(" ----- "+ite.next());
                              //} // for
        
        if (l_headPosRights.size() > 1)
            selectHeaders(l_headPosRights); 
        
                              // for testing and tracking
                              //System.out.println("UtilityImpl.subtractNegativeNotRemovingInOneACS() "+
                                //      "_l_negRights before selectHeaders():");
                              //for (Iterator ite = _l_negRights.iterator(); ite.hasNext(); ) {
                                      // for testing and tracking
                                //      System.out.println(" ------------------------ "+ite.next());
                              //} // for        
        
        l_headNegRights = (Set<StringRight>) ((HashSet)_l_negRights).clone();
        if (l_headNegRights.size() > 1)
            selectHeaders(l_headNegRights);

        
                              // for testing and tracking
                              //System.out.println("-- l_headNegRights after selectHeaders():");
                              //for (Iterator ite = l_headNegRights.iterator(); ite.hasNext(); ) {
                                      // for testing and tracking
                                //      System.out.println(" ----- "+ite.next());
                              //} // for
    }
    catch (BaseError b) {
        throw new InternalError("in UtilityImpl.subtractNegativeNotRemovingInOneACS(), "+
                "from StringRightImpl "+b);
    }
       
    return subtractNegativeHeadersNotRemovingInOneACS(
            l_headPosRights, l_headNegRights, _acs, _forNode);
  } // subtractNegativeNotRemovingInOneACS()  
  
  /**
  * Operates the reduction of each positive right by all the negative rights.
  * Each negative right negRight is subtracted to each positive right, and the positive results,
  * if any for each positive right posRight, are put in the returned set.
  * All the upper rights of negRight are removed in each posRight. For the lower rights,
  * this method processes in a different way than subtractNegativeFromPositiveInOneACS():
  * If a posRight (direct or indirect) lower right, say R, is also a negRight 
  * (direct or indirect) lower right, the common node R IS applicable, and then
  * posRight too. 
  * If a negative right is stronger than all positive rights, the returned set is empty,
  * and the calling method could use _l_headNegRights as the resulting right set if necessary.
  * If _acs is null, the result is not limited by the _acs rights set.
  * Called by withExecuteRight().
  * Calls subtractNegativeHeadersNotRemovingInOneACS().
  * @see UtilityImpl#selectHeaders
  * @param _l_right rights array not null nor empty.
  * @param _acs uses all the rights in the two right sets. May be null.
  * @param _forNode true if these rights are for a node, and so the 'directory'
  *         rights are not removed
  * @return the set of positive rights are all headers, and are used by _acs
  *         if not null. May be empty, not null.
  **/
   public final Set<StringRight> subtractInOneSetInOneACS(
           StringRight[] _l_right, ImmutableACS _acs, boolean _forNode) {
       
    if (_l_right == null) return null;
    Set<StringRight> l_posRights = new HashSet(INITIAL_CAPACITY); // capacity == 10
    l_test_.clear();

    for (int i = 0; i < _l_right.length; i++) {
        if (_l_right[i] == null) continue;
        if (_l_right[i].isGrantingRight()) l_posRights.add(_l_right[i]);
        else l_test_.add(_l_right[i]);
    }
    
    try {          
        // select the headers in each set
        selectHeaders(l_posRights); 
        selectHeaders(l_test_);
    }
    catch (BaseError b) {
        throw new InternalError("in UtilityImpl.subtractNegativeFromPositiveInOneACS(), "+
                "from StringRightImpl "+b);
    }
       
    return subtractNegativeHeadersNotRemovingInOneACS(
            l_posRights, l_test_, _acs, _forNode);
  } // subtractInOneSetInOneACS() 
   
  /**
  * Operates the reduction of each positive right by all the negative rights.
  * Each negative right is subtracted to each positive right, and the positive results,
  * if any for each positive right, are put in the returned set.
  * Each negative right and all the upper rights are removed in each positive right.
  * If a negative right is stronger than all positive rights, the returned set is empty,
  * and the calling method could use _l_headNegRights as the resulting right set if necessary.
  * For the lower rights, if a positive right (direct or indirect) lower right, say R,
  * is also a negative right (direct or indirect) lower right,
  * this common node R is put in the result.
  * If _acs is null, the result is not limited by the _acs rights set. <P>
  * Applies the ACS rules: <P>
  * - 'AGO right: no denying rights' <P>
  * - 'bridge right: no denying rights' <P>
  * - 'sens metarule: ACS denyings are superior to ACS grantings
  * (there are + and - operations), for each right type' <P>
  * - 'right type order: default order: root, A in AGO, all other right types' <P>
  * Uses the instance property l_to_retain_.
  * Called by subtractNegativeNotRemovingInOneACS() and subtractInOneSetInOneACS().
  * Calls StringRightImpl.selectHeaders() and subtractNotRemovingCommonLowerNodes(),
  * withInverseRight().
  * @see UtilityImpl#selectHeaders
  * @param _l_headPosRights positive rights with only headers. Not null and no null item.
  *     No update in this method, but the method may return it.
  * @param _l_headNegRights negative rights with only headers. Not null and no null item.
  *     No update.
  * @param _acs uses all the rights in the two right sets. May be null.
  * @param _forNode true if these rights are for a node, and so the 'directory'
  *         rights are not removed
  * @return the set of all-positive rights are all headers, and are used by _acs
  *         if not null. May be empty, not null.
  **/
  public final Set<StringRight> subtractNegativeHeadersNotRemovingInOneACS(
           Set _l_headPosRights, Set _l_headNegRights, ImmutableACS _acs, boolean _forNode) {
    Set             l_result = null;
    StringRight     right = null;
    StringRight     right2 = null;
    StringRight     right3 = null;
    Set<StringRight>    l_rights = null;
    l_to_retain_.clear();

    try {          
        
        Set l_subtractedPosRights = new HashSet(INITIAL_CAPACITY*3);
        
                              // for testing and tracking
                              //System.out.println(" ");
                              //System.out.println("UtilityImpl.subtractNegativeHeadersNotRemovingInOneACS() "+
                                //      "at the start:");
                              //System.out.println("--- positive rights are:");
                              //for (Iterator ite = _l_headPosRights.iterator(); ite.hasNext(); ) {
                                      // for testing and tracking
                                //      System.out.println(" ------------------------ "+ite.next());
                              //} // for
                              //System.out.println("--- negative rights are:");
                              //for (Iterator ite = _l_headNegRights.iterator(); ite.hasNext(); ) {
                                      // for testing and tracking
                                //      System.out.println(" ------------------------ "+ite.next());
                              //} // for

        // subtracts the negative rights from the positive rights.
        // this search may be not very fast if the right network is complex...
        if (!_l_headPosRights.isEmpty() && !_l_headNegRights.isEmpty()) {

            for (Iterator ite = _l_headPosRights.iterator(); ite.hasNext(); ) {
                right = (StringRight)ite.next();

                              // for testing and tracking
                              //System.out.println("UtilityImpl.subtractNegativeHeadersNotRemovingInOneACS() subtracting "+
                                //      "the positive right is "+right);

                // each negative right is used to remove the positive right ite.next()
                // or to reduce it if the two subnetworks have one common subnode or more
                for (Iterator ita = _l_headNegRights.iterator(); ita.hasNext(); ) {
                    right3 = (StringRight)ita.next();
                    right2 = withInverseRight(right3);

                              // for testing and tracking
                              //System.out.println("UtilityImpl.subtractNegativeHeadersNotRemovingInOneACS() subtracting "+
                                //      "iteration on the negative right "+right3);

                    if (right2 == null) {

                              // for testing and tracking
                              //System.out.println(
                                //      "--- the negative rights: inverse right not found.");
                              
                        continue;
                    } else {

                              // for testing and tracking
                              //System.out.println(
                                //      "--- the negative rights: the inverse right is found: "+right2);
                        
                    }
                    
                    l_rights = subtractNotRemovingCommonLowerNodes(
                                right, right2, _acs, _forNode);
                
                                  // for testing and tracking
                                  //System.out.println("UtilityImpl.subtractNegativeHeadersNotRemovingInOneACS() "+
                                    //      "for "+right2+" subtracting "+right+", the result are:");
                                  //for (Iterator iti = l_rights.iterator(); iti.hasNext(); ) {
                                    //      System.out.println(" ----- "+iti.next());
                                  //} // for
                    
                    if (l_to_retain_.isEmpty())
                        l_to_retain_.addAll( l_rights);
                    else {
                                  // for testing and tracking
                                  //System.out.println("-- l_to_next_test_ BEFORE retainAll():");
                                  //for (Iterator iti = l_to_retain_.iterator(); iti.hasNext(); ) {
                                    //      System.out.println(" ----- "+iti.next());
                                  //} // for
                                      
                        l_to_retain_.retainAll( l_rights );

                                  // for testing and tracking
                                  //System.out.println("-- l_to_next_test_ AFTER retainAll():");
                                  //for (Iterator iti = l_to_retain_.iterator(); iti.hasNext(); ) {
                                    //      System.out.println(" ----- "+iti.next());
                                  //} // for
                    } // else
                } // for ita
                
                // the subtracted rights of one positive right are not reduced
                // by the subtracted rights of another positive right
                l_subtractedPosRights.addAll(l_to_retain_);
                l_to_retain_.clear();
            } // for ite

            if (!l_subtractedPosRights.isEmpty()) {
                
                                  // for testing and tracking
                                  //System.out.println("UtilityImpl.subtractNegativeHeadersNotRemovingInOneACS() "+
                                    //      "AFTER subtracting and BEFORE selecting headers, the rights are:");
                                  //for (Iterator ite = l_subtractedPosRights.iterator(); ite.hasNext(); ) {
                                    //      System.out.println(" ----- "+ite.next());
                                  //} // for
                
                l_result = l_subtractedPosRights;
                selectHeaders(l_result); 
            }
            //if (!_l_headNegRights.isEmpty())
              //  l_result = _l_headNegRights;        
        
        } // if
        else if (!_l_headPosRights.isEmpty()) {
            l_result = _l_headPosRights;
        }
 
                                  // for testing and tracking
                                  //System.out.println("UtilityImpl.subtractNegativeHeadersNotRemovingInOneACS() AT THE END:");
                                  //System.out.println("----- final rights are:");
                                  //if (l_result != null) {
                                    //  for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                                              // for testing and tracking
                                      //        System.out.println(" --------- "+ite.next());
                                      //} // for
                                  //} else System.out.println("--------- no rights");

        l_subtractedPosRights = null;
        right = null;
        right2 = null;
        right3 = null;
    }
    catch (BaseError b) {
        throw new InternalError("in UtilityImpl.subtractNegativeHeadersNotRemovingInOneACS(), "+
                "from StringRightImpl "+b);
    }
       
    if (l_result == null)
        return Collections.EMPTY_SET;
    else return l_result;
  } // subtractNegativeHeadersNotRemovingInOneACS()     
  
  /**
   * UNUSED.
  * Detect the positive and negative rights, and put them in two separate sets.
  * @param _l_right array of acsrights or metarights to analyze
  * @return two Set of StringRights: the positive ones at 0 and 
  *     the negative ones at the index 1 of the returned array. Each set may be empty.
  * */
  public final Set<StringRight>[] separateRightsFromSens(
          StringRight[] _l_right) {
    
    if (_l_right == null) return null;
    Set<StringRight> l_result = new HashSet(INITIAL_CAPACITY); // capacity == 10
    Set<StringRight> l_negResult = new HashSet(INITIAL_CAPACITY); // capacity == 10
    Set<StringRight>[] result = new Set[] {l_result, l_negResult};

    for (int i = 0; i < _l_right.length; i++) {
        if (_l_right[i] == null) continue;
        if (_l_right[i].isGrantingRight()) l_result.add(_l_right[i]);
        else l_negResult.add(_l_right[i]);
    }

                      //for testing and tracking
                      //System.out.println("in RightsFactoryUtilities.separateRights(), "+
                      //  "result: "+result);

    return result;
  } // withNegativeRightFromSens()
   
  /**
  * UNUSED.
  * Subtract the 4 'delete' lower rights from a set.
  * Call subtractNegativeFromPositiveInOneACS(). Do not handle acs rights.
  * @param _l_posRights positive metarights. Not null and no null item.
  * @param _acs uses all the rights in the two right sets. May be null.
  * @param _forNode true if these rights are for a node, and so the 'directory'
  *         rights are not removed
  * @return the set of positive rights are all headers, and are used by _acs
  *         if not null. May be empty, not null. 
  **/
   public final Set<StringRight> subtractDeleteRightsInOneACS(
           Set<StringRight> _l_posRights, ImmutableACS _acs, boolean _forNode) {

       Set<StringRight> l_negRights = new HashSet(INITIAL_CAPACITY); // capacity == 15
       l_negRights.add(BaseManagerImpl.getInstance().getMetaRight("delete"));
       l_negRights.add(BaseManagerImpl.getInstance().getMetaRight("delete_child"));
       l_negRights.add(BaseManagerImpl.getInstance().getMetaRight("delete_item"));
       l_negRights.add(BaseManagerImpl.getInstance().getMetaRight("delete_tree"));       
       return subtractNegativeNotRemovingInOneACS(_l_posRights, l_negRights, _acs, _forNode);
   } // subtractDeleteRightsInOneACS()
 
   /**
   * Returns the argument without the key word 'immutable ' at the start or
   * the end of the string.
   * Called by gWork.FactoryUtilities.isTargetForPrivilegeForType(),
   * BasicImpl.setType(), ACSImpl.getEorL_PrivilegeForTypesForSource/Target(),
   * ACSFactoryImpl.getL_MapKeysForTypePolicies(), most of the ACSObject classes
   * and CPrivilegeUserImpl.
   * @param _type of a BaseObject
   * @return the type without 'immutable '; may be _type.
   */
  public String getObjectTypeWithoutImmutable(String _type) {
      String  result = null;

      if (_type == null)
          return null;

      if (_type.startsWith("<immutable> ")) {
          result = _type.substring(12);
      }
      else if (_type.endsWith("<immutable> ")) {
          result = _type.substring(0, _type.length() - 12);
      }
      else {
          result = _type;
      }

      return result;
  } // getObjectTypeWithoutImmutable()


  //
  // ------------- BeanInfo management methods ------------------------------
  //

  /**
  * Analyzes the BaseObject to select the right BeanInfo in gBase.
  * Works for the generic gBase and the AcsAddon gBase packages.
  * Called by the static initializer of the delete button in Gui1.BeamerManager,
  * Gui1.BeamerManager.updateBeamerPane(), SimplePropertyEditor, CollectionPropertyEditor,
  * ActionNewACSyst, ACSTree.createACSExplorer(), gWork.RightsMediatorImpl.
  * Does not process the IS nodes in the nameISTree for their displaying in the GUI,
  * since they are not BaseObjects.
  * If the BaseObject is an interface, select the BeanInfo of the relevant
  * implementation class if it exists.
  * If the BaseObject is an array, selects the BeanInfo of the implementation class.
  * Follows the names pattern in gBase to select the right BeanInfo.
  * Throws an InternalError if the BeanInfo can't be found.
  * Called in the generic access paths search by
  * UtilityImpl.add/removeAddonInner/ParentPropertyChangeListener(),
  * then may be called from a view worker thread. The other calls are from the
  * main thread or the event dispatch thread.
  * @param _objectInBeamer is the object to display in the beamer
  * @return the BeanInfo of the object
  */
  public BeanInfoPattern selectBeanInfo(BaseObject _objectInBeamer) {

    // initialization
    BeanInfoPattern   result = null; // no-sens init
    Class             beanInfoClass;
    String            objectClassName = _objectInBeamer.getClass().getName();
    String            rightClassName = objectClassName;

    if(_objectInBeamer.getClass().isInterface() || _objectInBeamer.getClass().isArray()) {
        // if 'Immutable' is at the beginning of the interface name,
        // without look at the package names, removes 'Immutable'.
        // adds 'Impl' to the interface name to find the class name
      rightClassName = getClassNameFromArrayOrInterface(
              getACSVocabularyFromArrayOrInterface(objectClassName),
              objectClassName);
    } // if

                        // for testing and tracking purpose
                     //System.out.println("in UtilityImpl.selectBeanInfo(), "+
                       //         "for the object "+_objectInBeamer.getNickName()+
                         //       " and the "+_objectInBeamer.getClass()+
                           //     ", the rightClassName is: "+rightClassName);

    try {

        beanInfoClass = Class.forName(getBeanInfoNameFromClass(rightClassName));
      
        result = (BeanInfoPattern) beanInfoClass.newInstance();

    } catch (Exception l) {
      new BaseError("from UtilityImpl.selectBeanInfo: "+l.toString());
      throw new InternalError("from UtilityImpl.selectBeanInfo(): "+l.toString());
    } catch (Error e) {
      new BaseError("from UtilityImpl.selectBeanInfo: "+e.toString());
      throw new InternalError("from UtilityImpl.selectBeanInfo(): "+e.toString());
    }

    beanInfoClass = null;
    objectClassName = null;
    rightClassName = null;

                        // for testing and tracking purpose
                         //System.out.println("in UtilityImpl.selectBeanInfo(), "+
                          //"the result is: "+result);

    return result;
  } // selectBeanInfo()

  /**
   * Get the name of the BeanInfo class from the operational class.
   * Simply add 'BeanInfo' at the end of the 'gBase' package and at the end of the argument.
   * For instance, with the 'ARoad0.gBase.ResourceImpl' argument, it returns
   * the value 'ARoad0.gBaseBeanInfo.ResourceImplBeanInfo'.
   * Works for the generic gBase and the ACS Addon gBase packages.
   * @param _text a gBase class name
   * @return the BeanInfo class name
   */
   public String getBeanInfoNameFromClass(String _text) {
      int index;
      String result;
      
      index = _text.indexOf("gBase");
      // add 'BeanInfo' at the end of the 'gBase' package
      result = _text.substring(0,
            index+5)+"BeanInfo"+
            _text.substring(
            index+5,
            _text.length());        

                          // for testing purpose
                    //System.out.println("UtilityImpl.getBeanInfoNameFromClass(), "+
                      //"the result for "+_text+" is:|"+result+"BeanInfo|");
      
      return result+"BeanInfo";
   } // getBeanInfoNameFromClass()

  /**
   * Gets the name of the operational class from the interface name, with 'Immutable' or not,
   * as an array or not, following the pattern of packages and class names in Access Road. <P>
   * Works for the generic gBase and for the AcsAddon gBase packages.
   * For instance, with the '[LARoad0.gBaseInterface.ImmutableResource' argument,
   * this method returns the value 'ARoad0.gBase.Resource'.
   * Returns the argument if it is the name of a class.
   * Follows the gBase pattern for classes and interface names, and the AcsAddon pattern.
   * Searches in gBase from interfaces in gBaseInterface.
   * Called by BeamerManager.getElementaryPropertyType(), BaseObjectHandler,
   * DialogBox and most of the PropertyEditors.
   * Calls ACSFactoryUtilityImpl.getEffectiveClassName().
   * @param _m_vocabulary from getACSVocabularyFromArrayOrInterface(). Never null.
   * For a class or an interface which is not managed in the AcsAddon framework,
   * like for StringRight, this property has to be empty.
   * @param _text an interface or a class name in a gBase or a gBaseInterface package.
   * Never null.
   * @return the class name, or null if _text is a generic interface like BaseObject or Node,
   * if the _text cental word is not found in ACSFactoryUtilityImpl.L_ACS_GENERIC_CLASSES,
   * or if the relevant AcsAddon class name does not start with 'ARoad0.AcsAddon.'
   */
   public final String getClassNameFromArrayOrInterface(
            Map<String, String[]> _m_vocabulary, String _text) {
      int       index;
      String    result;
      String    addonName = null;

                          // for testing purpose
                    //System.out.println("in UtilityImpl.getClassNameFromArrayOrInterface(), "+
                      //"the parameter text is: "+_text);

      if (_text == null || _text.indexOf("gBase") == -1) {

                              // for testing purpose
                        //System.out.println("in UtilityImpl.getClassNameFromArrayOrInterface(), "+
                          //"the result is null and the parameter is: "+_text);

          return null;
      }

      // if it is an array, so we have to avoid '[L' and ';'
      // at the beginning and at the end of the class name
      if (_text.indexOf("[L") == 0) result = _text.substring(2,_text.length() - 1);
      else result = _text;

      index = result.indexOf("Immutable");
      if (index != -1) {
        // result contains 'Immutable', so remove it
        result = result.substring(0,
            index)+
            result.substring(
            index+9,
            result.length());
      } // if

      index = result.indexOf("gBaseInterface");
      if (index != -1) {
        // result contains 'gBaseInterface', so removes Interface and keeps gBase
        result = result.substring(0,
            index+5)+
            result.substring(
            index+14,
            result.length());
      } // index

      if (result.endsWith("Impl"))
          result = result.substring(0, result.length() - 4);

      index = result.indexOf("ARoad0.AcsAddon.");
      if (index != -1) {
          index = result.indexOf("gBase.");
          // the addon name is just before 'gBase'
          addonName = result.substring(
                  result.substring(0, index - 1).lastIndexOf(".") + 1,
                  index - 1);
      } // if

      // remove the AcsAddon name at the end of the class name
      if (addonName != null && result.endsWith(addonName)) {
          result = result.substring(0, result.length() - addonName.length());
      }

                              // for testing purpose
                        //System.out.println("in UtilityImpl.getClassNameFromArrayOrInterface(), "+
                          //"the second argument of getEffectiveClassName() is: "+
                          //result.substring(result.indexOf("gBase.") + 6));

      addonName = null;

      // get the generic class or the AcsAddon class
      return getEffectiveClassName(_m_vocabulary,
              result.substring(result.indexOf("gBase.") + 6));

   } // getClassNameFromArrayOrInterface()

  //
  // -------------- AcsAddon management method ----------------------------
  //

  /**
   * Gets the map of the ACS vocabulary from the right ACSFactory,
   * following the pattern of packages and class names in Access Road. <P>
   * Works for the generic gBase and for the AcsAddon gBase packages.
   * For instance, with the '[LARoad0.gBaseInterface.ImmutableResource' argument,
   * this method returns the vocabulary from the constructor of ARoad0.gBase.ACSFactoryImpl.
   * With 'ARoad0.AcsAddon.Accbee.Ubuntu.gBaseInterface.ResourceUbuntu' as parameter,
   * this method returns the vocabulary from the constructor of
   * ARoad0.AcsAddon.Accbee.Ubuntu.gBase.ACSFactoryUbuntuImpl.
   * Follows the gBase pattern for class and interface names, and the AcsAddon pattern.
   * Called by selectBeanInfo(), BaseUtilityImpl.getL_AcsRestrictedRights(),
   * BaseObjectHandler, DialogBox and most of the PropertyEditors, often to be used with
   * getClassNameFromArrayOrInterface().<P>
   * Gets the ACSFactoryImpl empty constructor or one of its subclass constructor, to call
   * getM_AcsVocabulary(). For a generic gBaseInterface that is not managed
   * in the AcsAddon framework, returns always an empty map. These generic
   * gBaseInterface interfaces are ACSObject, BaseObject, Leaf, Node, NameTree, NameISTree,
   * Source, Target, DetailledName, StringRight, StringMetaRight. Only the two last interfaces
   * have an implementation in the generic gBase package, and StringMetaRightImpl is not
   * a subclass of StringRightImpl.
   * @param _text an interface or a class name in a gBase or a gBaseInterface package,
   * from the generic packages or from an AcsAddon package. Never null.
   * @return the ACS vocabulary, or an empty map if it is not managed in the AcsAddon framework,
   * or null if _text does not contain the word 'gBase', if the _text cental word
   * is not found in ACSFactoryUtilityImpl.L_ACS_GENERIC_CLASSES,
   * or if the relevant AcsAddon class name does not start with 'ARoad0.AcsAddon.'
   */
    public final Map<String, String[]> getACSVocabularyFromArrayOrInterface(
            String _text) {
      int                   index;
      String                className;
      Map<String, String[]>   m_result = new HashMap(25);
      Object                factory = null;
      String                addonName = null;

      if (_text == null || _text.indexOf("gBase") == -1)
          return null;

      // test the generic gBaseInterfaces that are not managed in the AcsAddon framework
      if (_text.indexOf("ACSObject") != -1 || _text.indexOf("BaseObject") != -1 ||
              _text.indexOf("Leaf") != -1 || _text.indexOf("Node") != -1 ||
              _text.indexOf("NameTree") != -1 || _text.indexOf("NameISTree") != -1 ||
              _text.indexOf("Source") != -1 || _text.indexOf("Target") != -1 ||
              _text.indexOf("StringRight") != -1 || _text.indexOf("StringMetaRight") != -1 ||
              _text.indexOf("DetailledName") != -1)
          return Collections.EMPTY_MAP;

                              // for testing purpose
                        //System.out.println("in UtilityImpl.getACSVocabularyFromArrayOrInterface(), "+
                          //"the parameter is: "+_text);

      // if it is an array, so we have to avoid '[L' and ';'
      // at the beginning and at the end of the class name
      if (_text.indexOf("[L") == 0)
          className = _text.substring(2,_text.length() - 1);
      else className = _text;

      index = className.indexOf("Immutable");
      if (index != -1) {
        // className contains 'Immutable', so remove it
        className = className.substring(0,
            index)+
            className.substring(
            index+9,
            className.length());
      } // if

      index = className.indexOf("gBaseInterface");
      if (index != -1) {
        // className contains 'gBaseInterface', so removes Interface and keeps gBase
        className = className.substring(0,
            index+5)+
            className.substring(
            index+14,
            className.length());
      } // index
      
      index = className.indexOf("ARoad0.AcsAddon.");
      if (index != -1) {
          index = className.indexOf("gBase.");
          // the AcsAddon name is just before 'gBase'
          addonName = className.substring(
                  className.substring(0, index - 1).lastIndexOf(".") + 1,
                  index - 1);

                              // for testing purpose
                        //System.out.println(
                          //      "in UtilityImpl.getACSVocabularyFromArrayOrInterface(), "+
                            //    "for _text: "+_text+", addonName: "+addonName);


          try {
              factory = Class.forName(className.substring(0, index + 6)+
                      "ACSFactory"+addonName+"Impl").
                      getConstructor(new Class[0]).newInstance(new Object[0]);
          }
          catch (Exception e) {
              throw new InternalError("UtilityImpl.getACSVocabularyFromArrayOrInterface() "+
                      "for className "+className+", throws "+e.toString());
          }
      }
      else {

                              // for testing purpose
                        //System.out.println(
                          //      "in UtilityImpl.getACSVocabularyFromArrayOrInterface(), "+
                            //    "for _text: "+_text+", NO addonName. ");

          factory = new ARoad0.gBase.ACSFactoryImpl();
      } // else

      m_result = ((ACSFactoryImpl)factory).getM_AcsVocabulary();

                                // for testing purpose
                                //System.out.println(
                                  //      "in UtilityImpl.getACSVocabularyFromArrayOrInterface(), for "+
                                    //    _text+", the result is from "+factory.getClass());

      className = null;
      factory = null;
      addonName = null;
      return m_result;
    } // getACSVocabularyFromArrayOrInterface()

  /**
   * Returns the name of the effective class to use.
   * It may be a generic class from ARoad0.gBase, and the returned value
   * is 'ARoad0.gBase._classNameImpl'.
   * Or it may be a class from an AcsAddon, and the returned value is
   * 'ARoad0.AcsAddon.xxxx.yyyy.gBase._classNameyyyyImpl' where 'xxxx' is the editor name
   * and 'yyyy' is the AcsAddon package name. <P>
   * In this second case, it may be an AcsAddon classic class that subclass and replace
   * a generic gBase class, or it may be an extension class which is defined
   * at the key ACSADDON_EXTENSION_CLASSES in the vocabulary map. <P>
   * It is possible to replace a generic class, in a package
   * named 'yyyy', by a classic class from a package 'zzzz'. On the other hand,
   * this is not authorized for the extension classes, since they must
   * be in the same AcsAddon gBase package. Furthermore, an extension class 
   * must be the subclass of a generic gBase class or of an AcsAddon classic class. <P>
   * Called by getClassNameFromArrayOrInterface(), AssociateUtilityImpl,
   * ACSFactoryImpl, ActionNewResource, ActionNewEligibleParty.
   * @param _m_vocabulary from ACSRun.getM_AcsVocabulary() or
   * getACSVocabularyFromArrayOrInterface().
   * For a class or an interface which is not managed in the AcsAddon framework,
   * like for StringRight, this property has to be empty.
   * @param _className one of the constants in L_CLASSIC_CLASS_KEYS, or
   * an AcsAddon extension class which is described with the key ACSADDON_EXTENSION_CLASSES
   * in _m_vocabulary, with the format _className/is/one_classic_class_name
   * @return the effective class name. Null if it is a generic interface like BaseObject or Node,
   * or if the AcsAddon class name does not start with 'ARoad0.AcsAddon.'
   */
  public final String getEffectiveClassName(Map _m_vocabulary, String _className) {

                            // for testing and tracking
                            //System.out.println("UtilityImpl.getEffectiveClassName() arguments: ");
                            //if (_m_vocabulary != null &&
                              //      _m_vocabulary.get(ACSFactoryUtilityImpl.ACS_CLASS) != null)
                                //System.out.println("--- _m_vocabulary from: "+
                                  //  ((String[])_m_vocabulary.get(ACSFactoryUtilityImpl.ACS_CLASS))[0]);
                            //System.out.println("--- _className: "+_className);

      if (_m_vocabulary == null || _className == null) return null;
      String    className = null;
      String    result = null;
      boolean   isExtensionClass = false;

      // For a generic gBaseInterface that is not managed in the AcsAddon framework,
      // returns null if it is not a StringRight
      if (_m_vocabulary.isEmpty()) {
          
          if (_className.equals("StringRight") || _className.equals("StringMetaRight"))
              result = "ARoad0.gBase."+_className+"Impl";

                              // for testing and tracking
                          //System.out.println("in UtilityImpl.getEffectiveClassName() "+
                            //      "empty vocabulary map  with result: "+result);

          return result;
      } // if

      if (!ACSFactoryUtilityImpl.L_CLASSIC_CLASS_KEYS.contains(_className)) {

          String[] classes = (String[]) _m_vocabulary.get(
                  ACSFactoryUtilityImpl.ACSADDON_EXTENSION_CLASSES);

          if (classes == null || classes.length == 0) {
              
                          // for testing and tracking
                          //System.out.println("--- "+ //UtilityImpl.getEffectiveClassName() "+
                            //      "with null result: L_CLASSIC_CLASS_KEYS and ACSADDON_EXTENSION_CLASSES"+
                              //    " do not contain _className");

                  return null;
          } // (classes == null

          for (int i = 0; i < classes.length; i++) {
              if (classes[i].split("/is/")[0].equals(_className)) {
                  className = _className;
                  isExtensionClass = true;
                  break;
              }
          } // for i

          if (className == null) {

                          // for testing and tracking
                          //System.out.println("--- "+ //UtilityImpl.getEffectiveClassName() "+
                            //      "with null result: L_CLASSIC_CLASS_KEYS and ACSADDON_EXTENSION_CLASSES"+
                              //    "do not contain _className");
                          
                  return null;
          } // (classes == null

          classes = null;
      } // if (!L_CLASSIC_CLASS_KEYS.
      else {
          className = _className;
      }

                                  // for testing and tracking
                                  //System.out.println("--- UtilityImpl.getEffectiveClassName() "+
                                    //      "className: "+className+
                                      //    " and _m_vocabulary.get(className) :"+
                                        //  _m_vocabulary.get(className));

      if (_m_vocabulary.containsKey(className)) {

          result = ((String[])_m_vocabulary.get(className))[0];
          if (result.startsWith("ARoad0.AcsAddon.")) {

                          // for testing and tracking
                          //System.out.println("--- UtilityImpl.getEffectiveClassName() "+
                            //      "for _className "+_className+"; result with 'ARoad0.AcsAddon' ");

          }
          //else {

                          // for testing and tracking
            //              System.out.println("--- UtilityImpl.getEffectiveClassName() "+
              //                    "with null result, with _className "+_className+"since "+
                //                  result+" does not start with ARoad0.AcsAddon.");

              //return null;
          //} // else
      }
      else if (ACSFactoryUtilityImpl.L_CLASSIC_CLASS_KEYS.contains(className)) {

                          // for testing and tracking
                          //System.out.println("--- UtilityImpl.getEffectiveClassName() "+
                            //      "for _className "+_className+";result 2: "+
                              //    "ARoad0.gBase."+className+"Impl");

          result = "ARoad0.gBase."+className+"Impl";
      }
      // this is an AcsAddon extension class
      else if (isExtensionClass) {

                            // for testing and tracking
                          //System.out.println("--- UtilityImpl.getEffectiveClassName() "+
                            //      "for _className "+_className+";result 3: "+
                              //    "ARoad0.AcsAddon."+
                                // ((String[])_m_vocabulary.get(ACSFactoryUtilityImpl.ACSADDON_EDITOR))[0]+"."+
                                 //((String[])_m_vocabulary.get(ACSFactoryUtilityImpl.ACSADDON_NAME))[0]+".gBase."+
                                  //className+
                                  //((String[])_m_vocabulary.get(ACSFactoryUtilityImpl.ACSADDON_NAME))[0]+"Impl");

          result = "ARoad0.AcsAddon."+
                  ((String[])_m_vocabulary.get(ACSFactoryUtilityImpl.ACSADDON_PACKAGE_EDITOR))[0]+"."+
                  ((String[])_m_vocabulary.get(ACSFactoryUtilityImpl.ACSADDON_NAME))[0]+".gBase."+
                  className+
                  ((String[])_m_vocabulary.get(ACSFactoryUtilityImpl.ACSADDON_NAME))[0]+"Impl";
      } // else
      else
          result = null;

                          // for testing and tracking
                          //System.out.println("in UtilityImpl.getEffectiveClassName() "+
                            //      "for _className "+_className+"; result: "+result);

      className = null;
      return result;
  } // getEffectiveClassName()


   /**
   * Registers for the specific events to listen,
   * because these events should change the rights of an ACSObject in an ACS Addon.
   * Then, these events are listened to update the views containing these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see addAddonParentPropertyChangeListener()).
   * Called by ARoad0.gBase.EPRViewInBaseImpl, NoThanViewInBaseImpl,
   * gWork.RightsMediatorImpl.
   * Calls changeAddonInnerPropertyChangeListener().
   * @param _listener to register
   * @param _obj the basic object for which some ACS events or ACSObject events have
   * to be listened, to catch the changings in its rights.
   * Noop if _obj is null, or if _obj is not from an AcsAddon ACS.
   */
  public void addAddonInnerPropertyChangeListener(
          PropertyChangeListener _listener, ACSObjectAddon _obj) {

      changeAddonInnerPropertyChangeListener(_listener,_obj, true);
  }

   /**
   * Registers the specific events to listen on a Resource or a VirtualFolder,
   * and its direct and indirect parents,
   * because these events should change the rights of the Resource or VirtualFolder
   * in an ACS Addon.
   * Then, these events are listened to update the views which contain these ACSObjects.
   * This method does not cover the inner properties of an ACSObject
   * (see addAddonInnerPropertyChangeListener()).
   * Called by ARoad0.gBase.EPRViewInBaseImpl, NoThanViewInBaseImpl,
   * gWork.RightsMediatorImpl.
   * Calls changeAddonParentPropertyChangeListener().
   * @param _listener to register
   * @param _obj the basic object of an AcsAddon ACS, for which some ACS events or
   * ACSObject events have to be listened, to catch the changings in its rights.
   * Noop if _obj is null, not a Resource or a VirtualFolder, or if
   * _obj is not from an AcsAddon ACS.
   */
  public void addAddonParentPropertyChangeListener(PropertyChangeListener _listener,
          ACSObject _obj)  {

      changeAddonParentPropertyChangeListener(_listener, _obj, true);
  }

   /**
   * Unregisters for the specific events to listen,
   * because these events should change the rights of an ACSObject in an ACS Addon.
   * Then, these events are listened to update the views which contain these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see removeAddonParentPropertyChangeListener()).
   * Called by ARoad0.gBase.EPRViewInBaseImpl, NoThanViewInBaseImpl,
   * gWork.RightsMediatorImpl.
   * Calls changeAddonInnerPropertyChangeListener().
   * @param _listener to register
   * @param _obj the basic object for which some ACS events or ACSObject events have
   * to be listened, to catch the changings in its rights.
   * Noop if _obj is null, or if _obj is not from an AcsAddon ACS.
   */
  public void removeAddonInnerPropertyChangeListener(PropertyChangeListener _listener,
          ACSObjectAddon _obj) {

      changeAddonInnerPropertyChangeListener(_listener,_obj, false);
  } // removeAddonInnerPropertyChangeListener()

   /**
   * Unregisters the specific events to listen on a Resource or a VirtualFolder,
   * and its direct and indirect parents,
   * because these events should change the rights of the Resource or VirtualFolder
   * in an ACS Addon.
   * Then, these events are listened to update the views which contain these ACSObjects.
   * This method does not cover the inner properties of an ACSObject
   * (see removeAddonInnerPropertyChangeListener()).
   * Called by ARoad0.gBase.EPRViewInBaseImpl, NoThanViewInBaseImpl,
   * gWork.RightsMediatorImpl.
   * Calls changeAddonParentPropertyChangeListener().
   * @param _listener to register
   * @param _obj the basic object for which some ACS events or ACSObject events have
   * to be listened, to catch the changings in its rights.
   * Noop if _obj is null, is not a Resource or a VirtualFolder,
   * or if _obj is not from an AcsAddon ACS.
   */
  public void removeAddonParentPropertyChangeListener(PropertyChangeListener _listener,
          ACSObject _obj) {

        changeAddonParentPropertyChangeListener(_listener, _obj, false);
  } // removeAddonParentPropertyChangeListener()

   /**
   * Registers or unregisters for the specific events to listen,
   * because these events should change the rights of an ACSObject in an ACS Addon.
   * Then, these events are listened to update the views which contain these ACSObjects.
   * This method does not cover the parent properties of a Resource or a VirtualFolder
   * (see addAddonParentPropertyChangeListener()).
   * Called by add/removeAddonParentPropertyChangeListener().
   * Calls ACSObjectAddon.getAddonInnerPropertiesToListen() and selectBeanInfo().
   * @see ACSObjectAddon#getAddonInnerPropertiesToListen()
   * @param _listener to register
   * @param _obj the basic object for which some ACS events or ACSObject events have
   * to be listened, to catch the changings in its rights.
   * Noop if _obj is null, or if _obj is not from an AcsAddon ACS.
   * @param _toAdd true for adding the listeners, false to remove them
   */
  private final void changeAddonInnerPropertyChangeListener(
          PropertyChangeListener _listener, ACSObjectAddon _obj, boolean _toAdd) {
      Map.Entry             entry = null;
      String                methodName = null;
      Method                method = null;
      List                  returnList = null;
      String                eventName = null;
      BeanInfoPattern       beanInfo = null;
      PropertyDescriptor[]  l_propertyDescriptor = null;
      boolean               toTest = false;
      BaseBeanManager       obj = null;

      if (_obj == null || !(_obj.getEorACS() instanceof AcsAddon))
              return;

      Map<String, String> m_properties = _obj.getAddonInnerPropertiesToListen();

      try {
          if (m_properties != null) {
              for (Object ob : m_properties.entrySet()) {
                  entry = (Map.Entry)ob;

                  // methodName, method for the key, to initialize returnList.
                  // 'obj' is the argument for the ACS method to call
                  if (((String)entry.getKey()).startsWith("<ACS>.")) {
                      methodName = ((String)entry.getKey()).substring(6);
                      method = _obj.getEorACS().getClass().getDeclaredMethod(
                                    methodName, (Class<?>[]) null);
                      returnList = (List) method.invoke(
                              _obj.getEorACS(), new Object[] {_obj});
                  } // if
                  else if (((String)entry.getKey()).equals("<THIS>")) {
                      returnList = new ArrayList(5);
                      returnList.add(_obj);
                  } // else

                  // methodName, method for the value, to initialize method and eventName.
                  // remove '<THIS>.' at the start.
                  if (((String)entry.getValue()).indexOf("<IF>") != -1) {
                      methodName = ((String)entry.getValue()).substring(7,
                          ((String)entry.getValue()).indexOf("<IF>"));
                      toTest = true;
                  } else {
                      methodName = ((String)entry.getValue()).substring(7);
                  }

                  for (Object occ : returnList ) {
                      obj = (BaseBeanManager)occ;

                      // gets on the first element, the res method to test
                      if (beanInfo == null) {
                          beanInfo = selectBeanInfo((BaseObject)obj);
                          l_propertyDescriptor = beanInfo.getPropertyDescriptors();
                          for (int i = 0; i < l_propertyDescriptor.length; i++) {
                              if (methodName.equals(
                                      l_propertyDescriptor[i].getReadMethod().getName())) {
                                  method = l_propertyDescriptor[i].getReadMethod();
                                  eventName = l_propertyDescriptor[i].getName();
                                  break;
                              } // if (methodName.equals(
                          } // for i
                      } // if (beanInfo == null

                      if (toTest && ((String)entry.getValue()).endsWith("<FALSE>")) {

                          if (!(Boolean) method.invoke(obj, (Object[])null)) {
                              if (_toAdd)
                                  obj.addPropertyChangeListener(eventName, _listener);
                              else
                                  obj.removePropertyChangeListener(eventName, _listener);
                          } // if

                      } // if (toTest
                      else if (toTest && ((String)entry.getValue()).endsWith("<TRUE>")) {

                          if ((Boolean) method.invoke(obj, (Object[])null)) {
                              if (_toAdd)
                                  obj.addPropertyChangeListener(eventName, _listener);
                              else
                                  obj.removePropertyChangeListener(eventName, _listener);
                          } // if

                      } // if (toTes
                      else {

                              if (_toAdd)
                                  obj.addPropertyChangeListener(eventName, _listener);
                              else
                                  obj.removePropertyChangeListener(eventName, _listener);

                      } // if (toTes

                  } // if ita

              } // for ob
          } // if (m_properties != null
      } catch (NoSuchMethodException n) {
          throw new InternalError("From UtilityImpl.changeAddonInnerPropertyChangeListener(): "+
                  n.toString());
      } catch (IllegalAccessException n) {
          throw new InternalError("From UtilityImpl.changeAddonInnerPropertyChangeListener(): "+
                  n.toString());
      } catch (InvocationTargetException n) {
          throw new InternalError("From UtilityImpl.changeAddonInnerPropertyChangeListener(): "+
                  n.toString());
      } catch (SecurityException n) {
          throw new InternalError("From UtilityImpl.changeAddonInnerPropertyChangeListener(): "+
                  n.toString());
      }

      entry = null;
      methodName = null;
      method = null;
      returnList = null;
      eventName = null;
      beanInfo = null;
      l_propertyDescriptor = null;
      toTest = false;
      m_properties = null;
      obj = null;
  } // changeAddonInnerPropertyChangeListener()

   /**
   * Registers or unregisters the specific events to listen on a Resource or a VirtualFolder,
   * and its direct and indirect parents,
   * because these events should change the rights of the Resource or VirtualFolder
   * in an ACS Addon.
   * Then, these events are listened to update the views which contain these ACSObjects.
   * This method does not cover the inner properties of an ACSObject
   * (see addAddonInnerPropertyChangeListener()).
   * Called by add/removeAddonParentPropertyChangeListener().
   * Calls AcsAddon.getAddonParentPropertiesToListen().
   * @see AcsAddon#getAddonParentPropertiesToListen() 
   * @param _listener to register
   * @param _obj the basic object of an AcsAddon ACS, for which some ACS events or
   * ACSObject events have to be listened, to catch the changings in its rights.
   * Noop if _obj is null, not a Resource or a VirtualFolder, or if
   * _obj is not from an AcsAddon ACS.
   * @param _toAdd true for adding the listeners, false to remove them
   */
  private final void changeAddonParentPropertyChangeListener(PropertyChangeListener _listener,
          ACSObject _obj, boolean _toAdd) {
      Map.Entry             entry = null;
      String                methodName = null;
      Method                method = null;
      List                  returnList = null;
      String                eventName = null;
      BeanInfoPattern       beanInfo = null;
      PropertyDescriptor[]  l_propertyDescriptor = null;
      boolean               toTest = false;
      ImmutableLeaf         leaf = null;
      Map<String, String>   m_properties = null;

      if (_obj == null || !(_obj.getEorACS() instanceof AcsAddon))
              return;

      try {
          if (_obj instanceof Resource || _obj instanceof VirtualFolder ) {

              m_properties = ((AcsAddon)_obj.getEorACS()).getAddonParentPropertiesToListen();

              if (m_properties != null) {

                     // ACSUbuntuImpl.getAddonParentPropertiesToListen() returns:
                    //m_result.put(
                     //       "<THIS>",
                       //     "<PARENT>.getAGOAccessFromTheParentOther<IF><FALSE>");
                    //m_result.put(
                      //      "<THIS>",
                        //    "<PARENT>.getL_inheritedOtherRights");
                    //m_result.put(
                      //      "<ACS>.getEorL_FromOtherInCurrentContextActors",
                        //    "<PARENT>.getAGOAccessFromTheParentOther<IF><FALSE>");
                   //

                  for (Object ob : m_properties.entrySet()) {
                      entry = (Map.Entry)ob;

                      // methodName, method for the key, to initialize returnList
                      if (((String)entry.getKey()).startsWith("<ACS>.")) {
                          methodName = ((String)entry.getKey()).substring(6);
                          method = _obj.getEorACS().getClass().getDeclaredMethod(
                                        methodName, (Class<?>[]) null);
                          returnList = (List) method.invoke(
                                  ((ACSObject)_obj).getEorACS(), (Object[])null);
                      } // if
                      else if (((String)entry.getKey()).equals("<THIS>")) {
                          returnList = new ArrayList(5);
                          returnList.add(_obj);
                      } // else

                      // methodName, method for the value, to initialize method and eventName.
                      // remove '<PARENT>.' at the start.
                      if (((String)entry.getValue()).indexOf("<IF>") != -1) {
                          methodName = ((String)entry.getValue()).substring(9,
                              ((String)entry.getValue()).indexOf("<IF>"));
                          toTest = true;
                      } else {
                          methodName = ((String)entry.getValue()).substring(9);
                          toTest = false;
                      }

                                              // for testing and tracking
                                              //System.out.println("in UtilityImpl."+
                                                // "changeAddonParentPropertyChangeListener()"+
                                                 //" for entry "+entry+"; method name: "+methodName+
                                                 //" for toTest: "+toTest);

                      for (Object occ : returnList ) {
                          leaf = (ImmutableLeaf)occ;

                                              // for testing and tracking
                                              //System.out.println("in UtilityImpl."+
                                                // "changeAddonParentPropertyChangeListener()"+
                                                // " for the leaf "+leaf);

                          if (toTest) {

                              while (leaf.getEorParent() != null) {

                                  // the method to test has to be get from the effective
                                  // BeanInfo of the current leaf
                                  beanInfo = selectBeanInfo(leaf);

                                  l_propertyDescriptor = beanInfo.getPropertyDescriptors();

                                              // for testing and tracking
                                              //System.out.println("in UtilityImpl."+
                                                // "changeAddonParentPropertyChangeListener()"+
                                                 //" for the BeanInfo "+beanInfo.getClass());
                                              //for (int i = 0; i < l_propertyDescriptor.length; i++) {
                                                //  System.out.println("---- i = "+i+" : "+l_propertyDescriptor[i]);
                                              //} // for i

                                  for (int i = 0; i < l_propertyDescriptor.length; i++) {

                                      if (methodName.equals(
                                              l_propertyDescriptor[i].getReadMethod().getName())) {
                                          method = l_propertyDescriptor[i].getReadMethod();
                                          eventName = l_propertyDescriptor[i].getName();
                                          break;
                                      } // if (methodName.equals(
                                  } // for i

                                              // for testing and tracking
                                              //System.out.println("in UtilityImpl."+
                                                // "changeAddonParentPropertyChangeListener()"+
                                                 //" for leaf "+leaf.getName()+"; method: "+method);

                                  if (((String)entry.getValue()).endsWith("<FALSE>")) {
                                      if (!(Boolean) method.invoke(leaf, (Object[])null)) {
                                          if (_toAdd)
                                              leaf.addPropertyChangeListener(eventName, _listener);
                                          else
                                              leaf.removePropertyChangeListener(eventName, _listener);

                                          leaf = leaf.getEorParent();
                                      }
                                      else break;
                                  } // if (((String)entry.getValue()).endsWith("<FALSE>")
                                  else if (((String)entry.getValue()).endsWith("<TRUE>")) {
                                      if ((Boolean) method.invoke(leaf, (Object[])null)) {
                                          if (_toAdd)
                                              leaf.addPropertyChangeListener(eventName, _listener);
                                          else
                                              leaf.removePropertyChangeListener(eventName, _listener);

                                          leaf = leaf.getEorParent();;
                                      }
                                      else break;
                                  } // else

                              } // while (leaf.getEorParent() != null)

                          } // if (toTest
                          else {
                              while (leaf.getEorParent() != null) {

                                  // the method to test has to be get from the effective
                                  // BeanInfo of the current leaf
                                  beanInfo = selectBeanInfo(leaf);

                                  l_propertyDescriptor = beanInfo.getPropertyDescriptors();
                                  for (int i = 0; i < l_propertyDescriptor.length; i++) {
                                      if (methodName.equals(
                                              l_propertyDescriptor[i].getReadMethod().getName())) {
                                          eventName = l_propertyDescriptor[i].getName();
                                          break;
                                      } // if (methodName.equals(
                                  } // for i

                                  if (_toAdd)
                                      leaf.addPropertyChangeListener(eventName, _listener);
                                  else
                                      leaf.removePropertyChangeListener(eventName, _listener);

                                  leaf = leaf.getEorParent();
                              } // while
                          } // else

                      } // if ita

                  } // for ob
              } // if (m_properties != null
          } // if (obj instanceof Resource || obj instanceof VirtualFolder )
      } catch (NoSuchMethodException n) {
          throw new InternalError("From UtilityImpl.changeAddonParentPropertyChangeListener(): "+
                  n.toString());
      } catch (IllegalAccessException n) {
          throw new InternalError("From UtilityImpl.changeAddonParentrPropertyChangeListener(): "+
                  n.toString());
      } catch (InvocationTargetException n) {
          throw new InternalError("From UtilityImpl.changeAddonParentPropertyChangeListener(): "+
                  n.toString());
      } catch (SecurityException n) {
          throw new InternalError("From UtilityImpl.changeAddonParentPropertyChangeListener(): "+
                  n.toString());
      //} catch (ClassNotFoundException n) {
        //  throw new InternalError("From UtilityImpl.changeAddonParentPropertyChangeListener(): "+
          //        n.toString());
      //} catch (InstantiationException n) {
        //  throw new InternalError("From UtilityImpl.changeAddonParentPropertyChangeListener(): "+
          //        n.toString());
      }

                              // this previous code implemented the function without reflection:
                                        // ACS.getAddonInheritedPropertiesToListen() returns
                                        // key: "<ACS>.getEorL_FromOtherInCurrentContextActors",
                                        // value: "<PARENT>.getAGOAccessFromTheParentOther<IF><FALSE>")
                              // to follow the OtherInCurrentContextActors which have no AGO access,
                              // to compute the new rights when the AGO access changes (those
                              // which have an AGO access are followed outside)
                          //List  l_otherActors = ((Resource)obj).
                            //      getEorACS().getEorL_FromOtherInCurrentContextActors();
                          //for (Iterator ita = l_otherActors.iterator(); ita.hasNext(); ) {
                              //res = (Resource)ita.next();
                              //while (res.getEorParent() != null) {
                                  //if (!res.getAGOAccessFromTheParentOther())
                                    //  res.addPropertyChangeListener("FromTheParentOther", this);

                                  //if (res.getAGOAccessFromTheParentOther() )
                                    //    res = (Resource)res.getEorParent();
                                  //else break;
                              //} // while
                          //} // if ita
                          //l_otherActors = null;

      entry = null;
      methodName = null;
      method = null;
      returnList = null;
      eventName = null;
      beanInfo = null;
      l_propertyDescriptor = null;
      toTest = false;
      m_properties = null;
      leaf = null;
  } // changeAddonParentPropertyChangeListener()
    
  /**
   * Gets the order of a node from a leaf.
   * For instance, the direct parent of a leaf has the order '1', and its direct parent
   * has the order '2', and so on.
   * Called by CPrivilegeUserImpl.setM_TargetInheritedLinkedPRIRights(),
   * resetInheritedPrivilegeRights() and ResourceImpl.resetInheritedAclEntryRights().
   * @param _leaf is never null
   * @param _node is never null. Cannot be _leaf.
   * @return the parent order as String of a number.
   * Returns '0' if _leaf has no parent, and '-1' if _node is not a direct or 
   * indirect parent of _leaf.
   */
   public synchronized final String getParentOrder(
           ImmutableLeaf _leaf, ImmutableNode _node) {
       int              order = 0;
       ImmutableNode    node = _leaf.getEorParent();
       boolean          found = false;

                          // for testing and tracking
                          //System.out.println("UtilityImpl.getParentOrder() "+
                            //      "for leaf "+_leaf.getNickName()+", for node "+_node.getName());

       while (node != null) {

                          // for testing and tracking
                          //System.out.println("--- test of parent "+node.getName());


           order++;
           if (_node == node) {
               found = true;
               break;
           }
           else
               node = node.getEorParent();
       } // while

       if (!found) order = -1;
       node = null;
       return new String(Integer.toString(order));
   } // getParentOrder()

/**
* Gets the BaseObject from the names of an open entity. It must be a metaright,
* an ACS right (which is not an ACSObject), an ACSObject or a ViewInBase.
* Open ACS and open views are looked for first, since an ACSObject should have the name
* of an ACS or a view. Thus, an ACSObject having the name of an ACS or view is undetectable.
* If a view or an ACS has the same name than an ACSObject, the view or the acs
* is returned without warning message.
* After the open ACS and the open view names, the method searches for among the metarights,
* then among the ACSObjects of each open ACS.
* Looking for the ACSObject over all the open ACS, but not in the closed ACS.
* Caution: a new ACS may have the name of an ACSObject in a closed ACS,
* and this is not checked. The ACSObjects in the open ACS are all checked.
* When the relevant closed ACS is opened, there is no more checking
* of the name doublons. During the search of objects from their names, the ACS
* is selected first.
* @param _name object name
* @return the object or null
*/
public synchronized final BaseObject getBaseObject(ImmutableName _name) {

    if (_name == null || _name.isEmpty() ) {
        return null;
    }

    if (BaseManagerImpl.getInstance().isOpenACS(_name)) {
      return (BaseObject) BaseManagerImpl.m_OpenAcs__.get(_name);
    }

    if (BaseManagerImpl.getInstance().isOpenView(_name)) {
      return (BaseObject) BaseManagerImpl.m_OpenViews__.get(_name);
    }

    BaseObject result = BaseManagerImpl.getInstance().getMetaRightFromName(_name);
    if (result != null) {
      return result;
    }

    ACS         acs = null;
    BaseObject  o1 = null;
    Collection  l_openACS = BaseManagerImpl.getInstance().getEorM_OpenACS().values();
    //SortedMap   m_ep = null;

    for (Iterator ite = l_openACS.iterator() ; ite.hasNext() ; ) {
      acs = (ACSImpl)ite.next() ;

                // for testing and tracking
                //System.out.println("UtilityImpl.getEorBaseObjects "+
                  //      "acs.getEorM_Resources().get(_name): "+
                    //    acs.getEorM_Resources().get(_name));

      if (!_name.contains(acs.getFullName())) continue;

      o1 = (BaseObject) acs.getEorM_EligibleParties().get(_name);
      if (o1 != null &&
              ((EligibleParty)o1).getEorACS() == acs ) {
            result = o1;
            break;
      } // if

      o1 = (BaseObject) acs.getEorM_Resources().get(_name);
      if (o1 != null &&
              ((Resource)o1).getEorACS() == acs) {
            result = o1;
            break;
      } // if

                                // for testing and tracking
                                //System.out.println("UtilityImpl.getBaseObject() search for "+
                                  //      _name);
                                //System.out.println("-- on the ACS "+acs.getNickName()+
                                  //      " - internal ACLs: "+acs.getManageInternalAcls()+
                                    //    " - external ACLs: "+acs.getManageExternalAcls());

      if (_name.component(0).equals("ACS") &&
              (acs.getManageInternalAcls() || acs.getManageExternalAcls())) {
            ImmutableAclEntry[] l_acl = acs.getEorL_AclEntries();

            for (int i = 0 ; i < l_acl.length ; i++) {

                                // for testing and tracking
                                //System.out.println("---- UtilityImpl.getBaseObject() on ACL "+
                                  //      l_acl[i].getName());

                if (l_acl[i].getName().equals(_name)) {
                    result = l_acl[i];
                    break;
                } // if
            } // for

            l_acl = null;
            if (result != null) break;
      } // if

      if (acs.getManageVirtualFolders()) {
            VirtualFolder[] l_vf = acs.getEorL_VirtualFolders();

            for (int i = 0 ; i < l_vf.length ; i++) {
                if (l_vf[i].getName().equals(_name)) {
                    result = l_vf[i];
                    break;
                } // if
            } // for

            l_vf = null;
            if (result != null) break;
      }

      if (acs.getManagePrivilegeRights()) {
            Privilege[] l_priv = acs.getEorL_Privileges();

            for (int i = 0 ; i < l_priv.length ; i++) {
                if (l_priv[i].getName().equals(_name)) {
                    result = l_priv[i];
                    break;
                } // if
            } // for

            l_priv = null;
            if (result != null) break;
      }

      if (acs.getManageAcsRights() ) {
            result = acs.getOneRight(_name);
            if (result != null) break;
      } // else

                // for testing and tracking
                //System.out.println("UtilityImpl.getBaseObject, the ACS manages acsrights: "+
                  //  acs.getManageAcsRights()+" and the rights are ");
                //StringRight[] l_rights = acs.getL_Rights();
                //for (int j= 0; j < l_rights.length; j++) {
                  //  System.out.println("----------- "+ l_rights[j]);
                //} // for j

    } // for ite
    
    acs = null;
    o1 = null;
    l_openACS = null;

    return result;
  } // getBaseObject(name)

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess(), but not for the first instance.
   */
  public void finalizeForProcess() {
    l_test_.clear();
    l_to_next_test_.clear();
    l_removed_.clear();
    l_headers_.clear();
    l_methodResult_.clear();
    l_rightsUpFromCurrent_.clear();
    l_PathsUpFromLesser_.clear();
    l_toAddTol_PathsUpFromLesser_.clear();
  } // finalizeForProcess()

} // UtilityImpl
