//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;

import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;

import java.beans.PropertyChangeEvent;

/**
 * This class is responsible for modeling a view which is the backing data
 * of a GraphicView in the GUI.
 * It is both a base listener and a bean with bounding properties.
 * It is a direct listener of the ACS openings and ACS closings,
 * since a view is opened only if all its ACS are opened.
 * It is NOT a listener of the base object opening and closing which is handled
 * by the super class.
 * It is NOT a listener of the properties of base objects which belong to this view,
 * since this feature is provided by the subclasses.
 * This class is used directly only for the 'Architecture' view, where the view objects
 * are ACS. Otherwise, the subclasses handle view objects as ACSObjects.
 * At the opening of a view, the deserialized instance does not contain
 * the view objects, and their value are put in the view from the open ACSs.
 * This is why the super method prepareAfterRestoring() has to be called
 * by BaseManagerImpl after the opening of a view.
 * Caution: no equals() nor clone() methods.
 * @version 0.7.0
 */
public class ViewInBaseImpl extends FolderAbst implements ViewInBase, BoundBean {

  /* private attributes*/
  private static final long   serialVersionUID__ = 0;
  
  public static final String  ARCHI_VIEW_STRING = "Architecture of ACS";
  public static final Name    ARCHI_VIEW_NAME = new NameImpl(ARCHI_VIEW_STRING);

  /* public attributes */
  public static final ViewInBase EMPTY_INSTANCE = new ViewInBaseImpl();

 /**
  * Default constructor
  */
  public ViewInBaseImpl() {
      type_ = "<basic_view>";
  }

  /**
   * This constructor is used by the GUI. Calls the last constructor.
   * @param _name ImmutableName is copied.
   * @exception CreateError if the view name is null, already known in BaseManagerImpl,
   *  or if it contains an ISO control value.
   */
  public ViewInBaseImpl(ImmutableName _name) throws CreateError {
      
      this(_name, true);

  } // constructor

  /**
   * This constructor allows the subclasses to manage directly the registration
   * in BaseManagerImpl, to initialize their properties before
   * the firing of the 'OpenViews' event by BaseManagerImpl.
   * Called by the previous constructor.
   * @param _name ImmutableName is copied.
   * @param _toRegisterInManager false if this constructor must not register
   *  the instance in BaseManagerImpl. If true, this constructor is identical to
   *  ViewInBaseImpl(_name)
   * @exception CreateError if the view name is null, already known in BaseManagerImpl,
   * or if it contains an ISO control value.
   */
  public ViewInBaseImpl(ImmutableName _name, boolean _toRegisterInManager)
          throws CreateError {
      
      super(_name);
      type_ = "<basic_view>";

      // adds this to the map of Views in BaseManagerImpl.
      // throws BaseError if already exists.
      // for testing the view without ViewManager: TAKE OUT THIS LIGNE
      if (_toRegisterInManager) {
        try {
            BaseManagerImpl.getInstance().newEorView(this);
        } catch (CreateError c) {
            this.finalizeForCopy();  // resources and EP are not unlistened
            throw new CreateError(c.toString());
        }
        // registers for 'OpenACS' events
        addOwnListening();
      } // if (_toRegisterInManager)
  } // constructor

  /**
  * This constructor provides a ViewInBaseImpl instance from a map of base objects.
  * Invoked by EPRViewInBaseImpl(ImmutableName _name, Map _m_BaseObjects) and
  * ViewInBase_Facade.createView().
  * This constructor is used by BaseManagerImpl.openView() to create
  * every instance. So, all the objects to listen (ACS...)
  * do receive a registration from this instance through the call to
  * the addOwnListening() method.
  * @param _name of the view. It is copied.
  * @param _m_BaseObjects is the map (name, object) of the view objects.
  *     It is copied. May be null.
  * @param _toRegisterInManager false if this constructor must not register
  *     the instance in BaseManagerImpl.
  * @exception CreateError if _name is already known in BaseManagerImpl,
  *     or if _m_BaseObjects contains a null, no BaseObject, or empty-name object.
  */
  public ViewInBaseImpl(ImmutableName _name, Map _m_BaseObjects,
    boolean _toRegisterInManager) throws CreateError {

      super(_name, _m_BaseObjects);
      type_ = "<basic_view>";
      
        // adds this to the map of Views in BaseManagerImpl.
        // throws BaseError if already exists.
        // TAKE OUT THIS LIGNE for testing the base without ViewManager
      if (_toRegisterInManager) {
        try {
            BaseManagerImpl.getInstance().newEorView(this);
        } catch (CreateError c) {
            this.finalizeForCopy();  // resources and EP are not unlistened
            throw new CreateError(c.toString());
        }
        // registers for 'OpenACS' events, and
        // ACSObject deletions in the ACS
        addOwnListening();
      } // if

  } // constructor with _m_BaseObjects
  
  
  // --------- methods of the implemented interfaces ------------------

  /**
   * @return long the serialVersionUID used by the serialization process.
   */
  public synchronized long getSerialVersionUID() {
    return serialVersionUID__;
  }
    
  /**
   * This short name is NOT unique for the Access Road program, nor its ACS if
   * it is an ACS object.
   * It is often the last component name, but not for a ViewInBase where it is
   * the two last components with ':: ' as separator. Overridden by NoThanViewInBaseImpl.
   * @return the nick name. May be null.
   */
    public String getNickName() {
       if (name_ == null) 
           return null;
       else if (name_.noOfComponents() == 1) 
           return name_.lastComponent();
       else
           return name_.component(name_.noOfComponents() - 2) + ":: " +
                   name_.component(name_.noOfComponents() - 1);
    } // getNickName()

  /**
   * False if at least one ACS structure modeling is incomplete for an used ACS.
   * The modeling of the ACS structure is incomplete when some ACSObjects miss
   * while they participate to the decision for some important access controls,
   * or simply when these missed ACSObjects are important for the overall ACS security,
   * or when their important relations with some other ACSObjects are not set.
   * An ACS structure is incomplete if one of the two methods ACS.getIncompleteStructureModeling()
   * and getIncompleteBehaviorModeling() return false.
   * On the other hand, the modeling remains complete if some secondary ACSObjects
   * are not modelled, like for instance image files in an operating system file system.
   * When the modeling of an ACS structure is incomplete, all the views which
   * use this ACS are set incomplete in their results.
   * @return false if the result is incomplet since one ACS structure is incomplete.
   */
  public boolean getCompleteResults() {
      boolean result = true;
      ACSObject b = null;
      for (Object obj : m_BaseObjects_.values()) {
          if (!(obj instanceof ACSObject)) continue;
          else {
              b = (ACSObject) obj;
              if (b.getEorACS().getIncompleteStructureModeling() ||
                      b.getEorACS().getIncompleteBehaviorModeling()) {
                  result = false;
                  break;
              }
          }
      } // for obj
      return result;
  } // getCompleteResults()

  /**
   * Observes the changes 'OpenACS' in the open ACS,
   * and closes the instance if a connected ACS has been closed.
   * Note: the ACS is then already finalized and unregistered in BaseManagerImpl.
   * Fires property change on 'ConnectedACS' if one connected ACS is closing (and
   * not for the opening ACS) only if the view contains ACSObjects (unlike the
   * 'Architecture' view).
   * For the 'Architecture' view, fire property change on 'OneBaseObjectInView' 
   * when an acs is opening or closing.
   * For a 'OneBaseObject' event, may call removeEorBaseObject().
   * Fires property change 'CompleteResults' if one connected ACS is opening or closing,
   * or if the view elements has changed.
   * Call BaseManagerImpl.getInstance().saveAndCloseOpenView().
   * Calls the super method for the 'OneBaseObject' events. Overridden method.
   * May invoke BaseManagerImpl.closeOpenView(this).
   * @param _evt
   */
  public void propertyChange(PropertyChangeEvent _evt){
    boolean toClose = false;
    ImmutableName name = null;
    List l_ACSnames = getEorL_ConnectedACSNames();
    
    // for empty view
    if (m_BaseObjects_ == null || m_BaseObjects_.isEmpty()) return;

    if (_evt.getSource() == BaseManagerImpl.getInstance() &&
      _evt.getPropertyName().equals("OpenACS")) {

                            // for testing and tracking
                            //System.out.println("In ViewInBaseImpl.propertyChange(),"
                              //+" the view "+name_+" has received an 'OpenACS' event "+
                              //"with l_connectedACSNames_ size: "+l_connectedACSNames_.size());

      Map m_openACS = (Map)_evt.getNewValue();

      try {
          // not for the 'Architecture' view
          if (!l_ACSnames.isEmpty()) {
              for (Iterator ite = l_ACSnames.iterator() ; ite.hasNext() ; ) {
                if (!m_openACS.containsKey(ite.next())) {

                            // for testing and tracking
                            //System.out.println("In ViewInBaseImpl.propertyChange(),"
                              //+" fire 'ConnectedACS' ");
                    
                   toClose = true;
                  // toClose activates the view closing,
                  // but without sending the fact that it is due to an ACS closing.
                  // So, fire another event to inform ViewTreeBaseListenerImpl
                  // before the 'OpenViews' event from BaseManagerImpl
                  firePropertyChange("ConnectedACS", l_ACSnames, null);
                  break;
                } // if
              } // for

              // updates the boolean 'CompleteResults', where the old and new values
              // do not matter in the property change event
              firePropertyChange("CompleteResults", null, getCompleteResults());
          }
          else {
              if (!name_.equals(ARCHI_VIEW_NAME)) {
                  l_ACSnames = null;
                  m_openACS = null;
                  return;
              }
              // for the 'Architecture' view only, fires a change about the closing or opening acs
              // note: works for any BaseObject
              BaseObject  obj = null;
              if (m_openACS.size() > ((Map)_evt.getOldValue()).size()) {
                  for (Iterator ite = m_openACS.keySet().iterator(); ite.hasNext(); ) {
                      name = (ImmutableName)ite.next();
                      if ( !((Map)_evt.getOldValue()).keySet().contains(name) ) {
                          obj = (BaseObject)m_openACS.get(name);
                          break;
                      } // if
                  } // for ite
                  if (obj != null) addEorBaseObject(obj);  
              }
              else {
                  for (Iterator ite = ((Map)_evt.getOldValue()).keySet().iterator(); ite.hasNext(); ) {
                      name = (ImmutableName)ite.next();
                      if ( !m_openACS.keySet().contains(name) ) {
                          obj = (BaseObject)((Map)_evt.getOldValue()).get(name);
                          break;
                      } // if
                  } // for ite
                  if (obj != null) removeEorBaseObject(obj);
              } // else if (m_openACS.size() > ((Map)_evt.getOldValue()).size())
              obj = null;
          } // else if (!l_ACSnames.isEmpty())
      } catch (Exception e) {
            ARoad0.Gui1.DialogBox.showAlarmMessage(
              "There is an exception after the base updating, from outside the database."+
              "You should save the base and stop the program, then run it once again.");
            System.out.println("From ViewInBaseImpl.propertyChange(), firePropertChange() throws:");
            e.printStackTrace();
      }  
    } // if (_evt.getSource() == BaseManagerImpl.getInstance()

    else if (_evt.getSource() instanceof ACS &&
      _evt.getPropertyName().equals("OneBaseObject")) {

            if (containsBaseObject((BaseObject)_evt.getOldValue()) ||
                    containsBaseObject((BaseObject)_evt.getNewValue())) {

                            // for testing and tracking
                            //System.out.println("In ViewInBaseImpl.propertyChange(),"
                              //+" the view "+name_+" has received an applicable"+
                                //" 'OneBaseObject' event ");

                super.propertyChange(_evt);

                // updates the boolean 'CompleteResults', where the old and new values
                // do not matter in the property change event
                firePropertyChange("CompleteResults", null, getCompleteResults());
            } // if
                
    } // if (_evt.getSource() instanceof ACS

    // saves and closes the view
    if (toClose) {

                    // for testing and tracking
                    //System.out.println("In ViewInBaseImpl.propertyChange(),"
                    //  +" close the view "+name_);
      try {
          // calls back finalizeForBase()
        BaseManagerImpl.getInstance().saveAndCloseOpenView(this);
      } catch (BaseError b) {
        b.printStackTrace();
      } // catch

    } // toClose

    name = null;
    l_ACSnames = null;
  } // propertyChange()


  // ------------ Methods out of the extended interfaces --------------------

  /**
  * This method creates a copy with another name.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy): baseError_,
  * m_BaseObjects_. Overridden by the subclass.
  * <P>
  * @param _name ImmutableName of the view. It is copied.
  * @param _view ViewInBaseImpl to copy.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in BaseManagerImpl.
  * @exception BaseError if _name is null or already known in BaseManagerImpl,
  * or if _view is null or not known.
  * @return a copy
  */
  protected static ViewInBase copy(ImmutableName _name,
    ViewInBase _view, boolean _toRegisterInManager) throws BaseError {

      if (_view == null || _name == null ||
        !BaseManagerImpl.getInstance().knowsView(_view.getName() )) {
          throw new BaseError("in ViewInBaseImpl.copy(_name, _view, boolean),"+
          " _name or _view is null, or _view is not known by the base.");
      }

      ViewInBaseImpl result = null;

      try {
          // no call to result.addOwnListening()
          result = new ViewInBaseImpl(
            _name, (Map)((TreeMap)((ViewInBaseImpl)_view).m_BaseObjects_).clone(), false);

          // copy _view but not m_ObjectsNames_
          result.baseError_ = ((ViewInBaseImpl)_view).getEorBaseError();

            // add result to the map of Views in BaseManagerImpl.
            // throws BaseError if already exists.
            // TAKE OUT THIS LIGNE for testing the base without ViewManager
          if (_toRegisterInManager) {

            BaseManagerImpl.getInstance().newEorView(result);
            // register for 'OpenACS' events
            result.addOwnListening();
          }
      } catch (CreateError c) {
          if (result != null) result.finalizeForCopy();  // resources and EP are not unlistened
          throw new BaseError(c.toString());
      }
      
      return result;
  } // copy

  /**
  * This method registers the instance as a property event listener of
  * 'OpenACS' on BaseManagerImpl to detect the closing of a relevant acs,
  * and 'OneBaseObject' on ACSImpl to detect the deleting of a view object.
  * Does not register for 'OpenViews', which is done by Gui2.ViewTreeBaseListenerImpl.
  * Called by the constructors, copy(), prepareAfterRestoring() methods,
  * and by the addOwnListening(boolean) subclass method.
  */
  protected void addOwnListening() {
      BaseManagerImpl.getInstance().addPropertyChangeListener("OpenACS", this);

      super.addOwnListening();
  } // addOwnListening 
  
  /**
  * This method unregisters the instance as a property event listener of
  * 'OpenACS' on BaseManagerImpl, and 'OneBaseObject' on ACSImpl if it is not closed
  * (the acs closing produces a call to finalizeForBase()).
  * Called by finalizeForBase(). Overrides and calls the super method.
  */
  protected void removeOwnListening() {

      BaseManagerImpl.getInstance().removePropertyChangeListener("OpenACS", this);

      super.removeOwnListening();
      
  } // removeOwnListening 
          
} // ViewInBaseImpl