//Title:        Access Road
//Version:      0.7.1
//Copyright:    2010-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBase;

import ARoad0.gBaseInterface.*;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

import java.util.List;
import java.util.TreeMap;
import java.util.SortedMap;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Collections;
import java.util.Map;
import java.util.ArrayList;

import java.io.Serializable;
import java.io.ObjectStreamField;
import java.beans.PropertyChangeListener;

/**
 * This class is a container of Resources belonging to one ACS,
 * responsible for being a PrivilegeTarget, an alias or an alias reference.
 * A VirtualFolder is also an ImmutableSource. It manages HAS_VIRTUAL_MEMBER relations
 * to its members. The relation VirtualFolder/member is weak. The deleting
 * of the container does not delete the members, on the contrary of Directory.
 * There are two main types of virtual folders, where the members are always Resources: <P>
 * - for Resource members which are not EligibleParties ('RES_VFOLDER' type) <P>
 * - for Resource members which may be EligibleParties ('EP_VFOLDER' type) <P>
 * Thus, an Virtual Folder is an access target, and it is an access source only as reference
 * of an alias or as container of its members.<P>
 * The default behavior in the rights propagation is the RIGHTS_PASSING_OFF_TO_MEMBERS mode.
 * This means that the VirtualFolder transfers none of its rights as target to its members.
 * In the other mode RIGHTS_PASSING_ON_TO_MEMBERS, the ACS has set this mode for its
 * virtual folders, and all the virtual folder rights are transfered to its members.<P>
 * Trees of virtual folders are possible. There is no rights inheritance in the tree
 * for the generic gBase classes, but an AcsAddon may override the hook methods.
 * A virtual folder may be included in: <P>
 * - a virtual tree of Resource Folders, which is not the Resources tree in the ACS, <P>
 * - or a virtual tree of Eligible Party Folders. <P>
 * The virtual folder manages: <P>
 * - its members, as any folder, and they are Basics of one ACS; they are not
 * deleted when this folder is deleted, on the contrary of Directory children, <P>
 * - its children, as any node, and they are virtual folders (no Basic) which are deleted
 * when this node is deleted, <P>
 * - its privileges, as any privilege target, and this may include external
 * LINKED_FOR_ONE_TO_ONE privileges. <P>
 * For each virtual folder type, there is in an ACS only one folders tree and then,
 * one folder root. <P>
 * There are two types of alias relation this class may handle: <P>
 * - the internal hard alias produces a direct alias-to-reference relation
 * in the same ACS, and the two objects know this relation, <P>
 * - the soft alias produces an indirect reference-to-alias relation
 * where the alias is external or internal, and it does not register this relation. <P>
 * As a FolderAbst, the members are not serialized when a VirtualFolderImpl instance
 * is saved. It is the responsability of BaseManagerImpl to call
 * ACSImpl.closeVirtualFolders() to prepare the virtual folders before their saving.
 * This class is a composite class which uses the following embedded classes
 * as private attributes: <P>
 * - CFolderMemberImpl for handling only EPRViewInBases (no VirtualFolders!) <P>
 * - CPrivilegeUserImpl for handling PrivilegeForTypes and PrivilegeForLinks as target <P>
 * - optional CNodeImpl for handling the VirtualFolder children <P>
 * - optional CAliasImpl for handling the hard and soft alias relations <P>
 * If there is a folder tree, it is possible to force each member to be member at most
 * one time in all the folder tree (EP or Resource tree).
 * There is no link between the members of a virtual folder and the members of children
 * of this folder. There is no link between the privileges of a virtual folder
 * and the privileges of its children.
 * As any node, the deleting of a virtual folder implies the deleting of its children.
 * The typed privileges may be apply to the two current types of virtual folder.
 * They may be used to defined default virtual folder rights.
 * There is no property 'ACS controllers' in a virtual folder having external
 * LINKED_FOR_ONE_TO_ONE privileges, contrary to the Resource. <P>
 * It is a javabean with the following bound properties (see also the superclass):
 * 'ImpliedOpenViews', 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege',
 * 'Children', 'VirtualFolderType','HardAliasReference', 'HardAlias', 'SoftAlias'
 * 'AcsOfSoftReferences' and 'hasSoftAlias'. <P>
 * The superclass FolderAbst needs to be prepared before a saving and after a restoring.
 * This is the responsability of the ACS methods 'close/fillVirtualFolders', called
 * by the BaseManager when an ACS is saved or opened. If the restoring is already done,
 * there is no action. <P>
 * This class has several subclasses in the AcsAddon packages.
 * Most of the variables are protected.
 * Note: no clone() method.
 * @see CFolderMemberImpl
 * @see CPrivilegeUserImpl
 * @see CNodeImpl
 * @see CAliasImpl
 * @see ARoad0.gWork.CompoundRightsFactoryImpl#selectDirectHiddenMembersForVirtualFolder(ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, ARoad0.gBaseInterface.ImmutableVirtualFolder)
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 */
public class VirtualFolderImpl extends FolderAbst implements VirtualFolder, Serializable {
  /**
   * The VFType when the folder is for Eligible Parties which are also Resources,
   * excluding the GroupIDMembers.
   **/
  public static final String        EP_VFOLDER = "Virtual Folder of Right User Resources";
  /**
   * The VFType when the folder is for simple Resources only, which are not Eligible Parties.
   **/
  public static final String        RES_VFOLDER = "Virtual Folder of simple Resources";
    
  // protected attributes
  protected ACS                     aCS_;
  protected ImmutableVirtualFolder  parent_;  // External Object Reference
  /**
   * The default value of VFType_ is never used
   **/
  protected String                  VFType_ = "<undefined>";

  /**
   * folders handler of this instance, only for ViewInBase links. Then it is
   * a transient property.
   */
  protected transient CFolderMemberImpl   folderMember_;
  /**
   * privileges handler of this instance
   */
  protected CPrivilegeUserImpl      privHandler_;
  /**
   * children handler of this instance
   */
  protected CNodeImpl               nodeHandler_;
  /**
   * Optional alias handler of this instance.
   */
  protected CAliasImpl              aliasHandler_ = null;
  /**
   * True if an ACSObject cannot be in a folders tree more than one time,
   * as member of one of the virtual folders.
   * If false, several folders in an ACS may have the same object as member.
   * This is set by the ACS.
   */
  protected boolean                 memberOneTimeInTree_ = false;

  // ------------- public attributes -----------------------------------------
  /*
   * Used to create a new instance in CollectionPropertyEditor from the BeanInfo description
   * for ACSs and VirtualFolders
   */
  public static final VirtualFolderImpl EMPTY_INSTANCE = new VirtualFolderImpl();

  /**
   * variable for the JDK 2 serialization
   * @serial ("aCS_",ACSImpl.class),
   * ("parent_",ImmutableVirtualFolder.class),
   * ("memberOneTimeInTree_",Boolean.TYPE),
   * ("privHandler_",CPrivilegeUserImpl.class),
   * ("nodeHandler_",CNodeImpl.class),
   * ("VFType_",String.class),
   * "aliasHandler_", CAliasImpl.class)
   */
    public final static ObjectStreamField[] serialPersistentFields =
        {
          new ObjectStreamField("aCS_",ACSImpl.class),
          new ObjectStreamField("parent_",ImmutableVirtualFolder.class),
          new ObjectStreamField("memberOneTimeInTree_",Boolean.TYPE),
          new ObjectStreamField("privHandler_",CPrivilegeUserImpl.class),
          new ObjectStreamField("nodeHandler_",CNodeImpl.class),
          new ObjectStreamField("VFType_",String.class),
          new ObjectStreamField("aliasHandler_", CAliasImpl.class)
        };

 /**
  * Default constructor
  */
  public VirtualFolderImpl() {
    aCS_ = null;
    parent_ = null;
    folderMember_ = null;
    privHandler_ = null;
    nodeHandler_ = null;
  }

   /**
   * Used by the beamer editor for the ACS VirtualFolders list, from the 'New' button,
   * and by the gWork.AcsFactory. Calls the other constructor.
   * @param _name name of the folder name. Its last component is added to the _parent name
   * if _parent is not null, or to the _acs name, to get the folder name. Never null.
   * @param _acs is the main ACS of the folder to create. Never null.
   * @param _parent is the folder which contains this folder, and with the same type;
   *   may be null or immutable.
   * @param _type is the folder type. Used like the Resource type. If null, the internal value
   *  is "&lt;undefined&gt;". Not the main type, get through getMainVFType.
   * @param _forResourcesOnly defines the basic folder type; true if the folder
   *   manages only Resources as children (RES_VFOLDER type), and false when
   *   Resource Eligible Parties may also be children (EP_VFOLDER type)
  * @exception CreateError if _m_BaseObjects contains a null,
  *     no ACSObject, empty-name or not _acs object, or with double link in
  *     the folders tree for a ACS, if _parent has not the same folder type,
  *     if the name is the name of a known IS node, VirtualFolder, Resource or EligibleParty,
  *     or if the object type is not managed by the ACS.
   */
  public VirtualFolderImpl(ImmutableName _name, ACS _acs, VirtualFolder _parent, 
          String _type, Boolean _forResourcesOnly) 
          throws CreateError {

      this(_name, _acs, _parent, _type, _forResourcesOnly, true);
        
                          // for testing purpose
                     //System.out.println("Call to ResourceImpl.ResourceImpl("+
                       // "ImmutableName _name, ACS _acs, Node _parent, String _resourceType)"+
                       // " for "+_name);

  } // ResourceImpl constructor

  /**
   * This is the main constructor.
   * If the parent is null and if there is no root in the folders tree,
   * registers this instance as root in the ACS.
   * The relevant sources and targets register/unregister the typed privilege when the source,
   * the target or the privilege is created/deleted, so sources and targets know
   * their typed privileges, but typed privileges do not know the sources and
   * targets which use them. This is NOT true for the TYPED_FOR_SEED privileges,
   * which have external sources and targets having no link to these privileges.
   * The property aliasHandler_ remains null in this constructor.
   * @param _name name of the folder name. Its last component is added to the _parent name
   * if _parent is not null, or to the _acs name, to get the folder name. Never null.
   * @param _acs is the main ACS of the folder to create. Never null.
   * @param _parent is the folder which contains this folder, and with the same type;
   *   may be null or immutable.
   * @param _type is the folder type. Used like the Resource type. If null, the internal value
   *  is "&lt;undefined&gt;". Not the main type, get through getMainVFType.
   * @param _forResourcesOnly defines the basic folder type; true if the folder
   *   manages only Resources as children (RES_VFOLDER type), and false when
   *   Resource Eligible Parties may also be children (EP_VFOLDER type)
   * @param _toRegisterInManager false if this method must not register
   *   the instance in its ACS.
   * @exception CreateError if _m_BaseObjects contains a null,
   *     no ACSObject, empty-name or not _acs object, or with double link in
   *     the folders tree for a ACS, if _parent has not the same folder type,
   *     if the name is the name of a known IS node, VirtualFolder, Resource or EligibleParty,
   *     or if the object type is not managed by the ACS.
   */
  public VirtualFolderImpl(ImmutableName _name, ACS _acs, VirtualFolder _parent, 
          String _type, Boolean _forResourcesOnly, boolean _toRegisterInManager)
          throws CreateError {

      super(((Name)_acs.getName()).extended(_name.lastComponent()), null);


                          // for testing and tracking
                          //System.out.println("VirtualFolderImpl constructor with arguments: ");
                          //System.out.println("-------- _name: "+_name);
                          //System.out.println("-------- _acs: "+_acs.getName());
                          //System.out.println("-------- _type: "+_type);
                          //System.out.println("-------- _forResourcesOnly: "+_forResourcesOnly);


      // updates name_
      // works well because super(_name, null) uses _name only to update name_
      if (_parent != null)
          name_ = ((Name)_parent.getName()).extended(_name.lastComponent());
      
      if (_forResourcesOnly) VFType_ = RES_VFOLDER;
      else VFType_ = EP_VFOLDER;

      aCS_ = _acs;

      if (BaseManagerImpl.getInstance().getISTree(_acs.getISName()).isNode(name_) ||
              BaseManagerImpl.getInstance().getISTree(_acs.getISName()).isArea(name_)) {
        aCS_ = null;
        name_ = null;
        comment_ = null;
        throw new CreateError("VirtualFolderImpl constructor with the name of a node or an area"+
                " in the Information System.");
      }

      if (_acs.getEorM_EligibleParties().get(name_) != null ||
              _acs.getEorM_Resources().get(name_) != null) {
        aCS_ = null;
        name_ = null;
        comment_ = null;
        throw new CreateError("VirtualFolderImpl constructor with the name of a Resource or"+
                " an EligibleParty in the same ACS.");
      }

      if (_acs.getOneVirtualFolder(name_) != null) {
          try {
              finalizeForBase();
          } catch (UpDateError u) {
                    //u.printStackTrace();
          }
          throw new CreateError(" The folder name already exists.");
      }
      if (!_acs.getManageVirtualFolders() ) {
          try {
              finalizeForBase();
          } catch (UpDateError u) {
                    //u.printStackTrace();
          }
          throw new CreateError(
              " The virtual folders are not managed in the ACS ");
      }
      if (_parent != null && !_acs.getManageVirtualFolderTrees() ) {
          try {
              finalizeForBase();
          } catch (UpDateError u) {
                    //u.printStackTrace();
          }
          throw new CreateError(
              " The virtual folders are not managed in the ACS ");
      }
      if (_acs.containsOwnResourceName(name_) || _acs.containsEligiblePartyName(name_) ) {
          try {
              finalizeForBase();
          } catch (UpDateError u) {
                    //u.printStackTrace();
          }
          throw new CreateError(
                      " The folder name is the name of a known Resource or EligibleParty. ");
      }
      
      if (_type != null) {
          try {
              setType(_type);
          } catch (UpDateError u) {
             throw new CreateError(" In the VirtualFolder constructor: "+u.toString());
          }
      } // if

      // controls _parent
      if (_parent != null) {
          if (!_parent.getAcsName().equals(_acs.getName())) {
              try {
                  finalizeForBase();
              } catch (UpDateError r) {
                    //u.printStackTrace();
              }
              throw new CreateError(" In the VirtualFolder constructor,"+
                " the parent has not the same ACS");
          } // if (!_parent.getAcsName()
          if (!_parent.getMainVFType().equals(VFType_)) {
              try {
                  finalizeForBase();
              } catch (UpDateError r) {
                    //u.printStackTrace();
              }
              throw new CreateError(" In the VirtualFolder constructor,"+
                " the parent has not the same virtual folder main type");
          } // if (!_parent.getMainVFType()

          // no adding in _parent if its type is '<byAcsCreator> fffff'
          // but 'immutable' is allowed in the parent type
          String crea = "<byAcsCreator>";
          if (_parent.getType().indexOf(crea) != -1 &&
                  _parent.getType().length() > crea.length()) {
                      try {
                          finalizeForBase();
                      } catch (UpDateError r) {
                            //u.printStackTrace();
                      }
                      throw new CreateError(" The type of the parent "+
                            _parent.getName().lastComponent()+" does not allow any new child.");
          } // if
          crea = null;


          if (aCS_.getM_TypeOrientedPolicy() != null) {
              // no adding in _parent if its type forbids the resource type
              // in the key 'VirtualFolder.TypesOfChildFor', modulo '<immutable> '
                String[] l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("VirtualFolder.TypesOfChildFor."+
                            UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(
                                    _parent.getType()));
                boolean keep = false;
                if (l_types != null) {
                    StringBuilder types = new StringBuilder();
                    for (int i = 0; i < l_types.length; i++) {
                          if (l_types[i].equals(
                                  UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(_type))) {
                              keep = true;
                          }
                          types.append(l_types[i]);
                          types.append(", ");
                    } // for i
                    if (!keep) {
                          System.out.println("VirtualFolderImpl constructor, error "+
                                  "on the type for "+_name);
                          try {
                              finalizeForBase();
                          } catch (UpDateError r) {
                                                //u.printStackTrace();
                          }
                          if (types.length() != 0)
                              throw new CreateError(" Through the ACS policy key "+
                                        "'VirtualFolder.TypesOfChildFor', "+
                                        "the parent accepts only children having the types "+types);
                          else
                              throw new CreateError(" Through the ACS policy key "+
                                        "'VirtualFolder.TypesOfChildFor', "+
                                        "the parent accepts no child at all. ");
                    } // if (!keep)
                    types = null;
                } // if (l_types !=
                l_types = null;
          } // if (aCS_.getM_TypeOrientedPolicy()

      } // if (_parent != null)
      else {

          if (aCS_.getM_TypeOrientedPolicy() != null) {
              // no adding in _parent if its type forbids the resource type
              // in the key 'VirtualFolder.TypesOfChildFor', modulo '<immutable> '
                String[] l_types = (String[]) aCS_.getM_TypeOrientedPolicy().
                          get("VirtualFolder.TypesOfChildFor."+"NULL");
                boolean keep = false;
                if (l_types != null) {
                    StringBuilder types = new StringBuilder();
                    for (int i = 0; i < l_types.length; i++) {
                          if (l_types[i].equals(
                                  UtilityImpl.getFirstInstance().
                                    getObjectTypeWithoutImmutable(_type))) {
                              keep = true;
                          }
                          types.append(l_types[i]);
                          types.append(", ");
                    } // for i
                    if (!keep) {
                          System.out.println("VirtualFolderImpl constructor, error "+
                                  "on the type for "+_name);
                          try {
                              finalizeForBase();
                          } catch (UpDateError r) {
                                                //u.printStackTrace();
                          }
                          if (types.length() != 0)
                              throw new CreateError(" Through the ACS policy key "+
                                        "'VirtualFolder.TypesOfChildFor.NULL', "+
                                        "the no-parent folder may have only the types "+
                                        types);
                          else
                              throw new CreateError(" Through the ACS policy key "+
                                        "'VirtualFolder.TypesOfChildFor.NULL', "+
                                        "the no-parent folder is forbidden. ");
                    } // if (!keep)
                    types = null;
                } // if (l_types !=
                l_types = null;
          } // if (aCS_.getM_TypeOrientedPolicy()
      } // else if (_parent != null)
      
      // --------- instanciates ------------------------

      if (_acs.getManageOneTimeInVirtualFolderTreeForMember()) {
          memberOneTimeInTree_ = true;
      }

      if (_parent != null) {
          ((VirtualFolderImpl)_parent).addEorChild(this); // sends the "Children" event
          parent_ = _parent;
      } // if _parent

      folderMember_ = null;
      privHandler_ = new CPrivilegeUserImpl(this);
      nodeHandler_ = new CNodeImpl(this);
      // aliasHandler_ is created only if needed
      
      boolean nok = false;
      try {
          if (_toRegisterInManager) {

                          // for testing and tracking
                          //System.out.println("VirtualFolderImpl constructor calls "+
                            //      "the method ACS.newEorVirtualFolder()");

              // adds to the map of BaseObjects in ACS after the variables init.
              // throws CreateError if already exists or InternalError.
              // for testing the folder without ACS: TAKE OUT THIS LIGNE
              ((ACSImpl)aCS_).newEorVirtualFolder(this);

              // updates the ACS if this is the first folder
              // throws UpDateError
              if (_parent == null) {
                if (_forResourcesOnly) {
                    if (((ACSImpl)_acs).getOwnResourceFoldersRootName().equals(
                       NameImpl.EMPTY_INSTANCE))
                          ((ACSImpl)_acs).setOwnResourceFoldersRootName(name_);
                }
                else {
                    if (((ACSImpl)_acs).getOwnEPFoldersRootName().equals(
                       NameImpl.EMPTY_INSTANCE))
                          ((ACSImpl)_acs).setOwnEPFoldersRootName(name_);
                } // else
              } // if (_parent
          } // if
      } catch (CreateError c) {
            nok = true;
      } catch (UpDateError u) {
            nok = true;
      }  
      
      if (nok) {
          try {
              finalizeForBase();
          } catch (UpDateError u) {
              //u.printStackTrace();
          }
          throw new CreateError(" Error in VirtualFolderImpl constructor "+
                "in ACS registering or folder root name registering");
      } // if

                          // for testing purpose
                          //System.out.println("end of the VirtualFolderImpl.constructor for "+
                            //                getName());

  } // constructor

  /**
   * No use, since CollectionPropertyEditor calls BaseObjectHandler.
   * Calls the main constructor.
   * Caution: it has to not be used as a Resource name in the same ACS,
   * because linked privileges have resources and folders as targets.
   * @param _name of the view. It is copied. 
   * @param _acs is the main ACS of the folder to create. Never null.
   * @param _parent is the folder which contains this folder, and with the same type;
   *   may be null.
   * @param _type is the folder type. If null, the internal value is "<undefined>".
   * Used for some typed privileges. The main type is get through getMainVFType.
   * @param _forResourcesOnly defines the basic folder type; true if the folder
   *   manages only Resources as children (RES_VFOLDER type), and false when
   *   Resource Eligible Parties may also be children (EP_VFOLDER type)
  * @exception CreateError if _acs is null, or from the constructor.
   */
  public static VirtualFolder valueOf(ImmutableName _name, ACS _acs, VirtualFolder _parent,
          String _type, Boolean _forResourcesOnly) throws CreateError {

                          // for testing purpose
                          System.out.println("call to VirtualFolderImpl.valueOf() ");

      if (_acs == null) throw new CreateError("error in VirtualFolderImpl constructor "+
              "with a null ACS");

      return new VirtualFolderImpl(_name, _acs, _parent, _type, _forResourcesOnly);

  } // valueOf()
  
  /**
  * This method provides a copy of the instance
  * with another name, without children copying
  * because the copy order will not be always correct.
  * @param _name is the copy name. Caution: it has to not be used
  *   as a Resource name in the same ACS, because linked privileges have
  *    resources and folders as targets.
  * @param _acs is the main ACS of the copy to create. May not be the ACS of this instance.
  * @param _parent is the VirtualFolder (not Node!) which belongs to _acs
  *   and will have the copy. May be null.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return Object is a copy as VirtualFolderImpl instance
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent is not a virtual folder, or _parent ACS is not _acs.
  **/
  public Object copy(ImmutableName _name, ACS _acs, Node _parent,
    Boolean _toRegisterInManager) throws CreateError {

      if (_acs == null || _name == null || 
        (_parent != null && ((VirtualFolder)_parent).getEorACS() != _acs) ||
        !(_parent instanceof VirtualFolder)) {
          throw new CreateError("in VirtualFolderImpl.copy(),"+
          " _name or _acs is null,"+
          " or _parent is not a virtual folder, or _parent ACS is not _acs.");
      }
      
      boolean forRes;
      if (VFType_.equals(RES_VFOLDER) ) forRes = true;
      else forRes = false;
      
      VirtualFolderImpl result = new VirtualFolderImpl(
          _name, _acs, (VirtualFolder)_parent, type_,
          forRes);

      // Todo: adds the base objects

      if (privHandler_ != null)
          result.privHandler_ = privHandler_.copy(result);
      if (aliasHandler_ != null)
          result.aliasHandler_ = aliasHandler_.copy(result);
      
      return result;
  } // copy()
  
  /**
   * UNUSED
  * This method provides a copy of the instance variables.
  * Each internal variable that is an external
  * object reference has a shallow copy (no deep copy).
  * Copy the privileges only if it is possible.
  * Called by copy() and by the copyOwnVariables() methods
  * of the subclasses.
  * @param _result to update with this instance variables copies
  * @return the updated _result
  * @exception CreateError from
  **/
  protected VirtualFolderImpl copyOwnVariables(VirtualFolderImpl _result) throws CreateError {

      if (_result == null) {
          throw new CreateError("in VirtualFolderImpl.copyOwnVariables(),"+
          " _result is null.");
      }
      
      //
      // TODO: control the values from the ACS
      //

      return _result;
  } // copyOwnVariables(,,,)

  /**
   * This method is used when memberOneTimeInTree_ is true,
   * to detect the member in all the folders of the tree, 
   * for given virtual folder type.
   * No exception if the member is there more than one time.
   * @param _acs is the ACS of the folders tree
   * @param _folderType is RES_VFOLDER or EP_VFOLDER
   * @param _obj is a new member which could be already
   *    in the folder tree
   * @return boolean true if _obj has been detected one time
   *   or more in the tree. Always false if one parameter is null,
   *   or if _folderType is not a folder type.
   */
  protected static final boolean detectMemberInFolderTree(
    ACS _acs, String _folderType, ACSObject _obj) {
        boolean result = false;
        
        if (_acs == null || _folderType == null || _obj == null) return result;
      
        if (_folderType.equals(RES_VFOLDER)) {
            for (Iterator ite = _acs.getEorL_ResourceVirtualFolders().iterator(); ite.hasNext();) {
                
                if (((VirtualFolder)ite.next()).containsBaseObject(_obj)) {
                    result = true;
                    break;
                }
                
            } // for
        } else {
            for (Iterator ite = _acs.getEorL_EPVirtualFolders().iterator(); ite.hasNext();) {
                
                if (((VirtualFolder)ite.next()).containsBaseObject(_obj)) {
                    result = true;
                    break;
                }
                
            } // for             
        } // else

        return result;
  } // detectMemberInFolderTree(obj)
  
  //
  // ----------- Methods of the implemented interfaces ----------------
  //

  /**
   * @return ImmutableACS is an External Object Reference to the ACS
   *   that owns this folder (main ACS)
   */
  public ImmutableACS getEorACS() {
          return aCS_;
  }
  
  /**
   * The virtual folder main type is defined by the constructor.
   * @return String of main VF type, which is "RES_VFOLDER" or "EP_VFOLDER".
   *    Not null.
   */
  public String getMainVFType() {
    return VFType_;
  }
  
  /**
   * If the argument is allowed by the ACS, sets the secondary type of this instance.
   * The main type is EP_VFOLDER or RES_VFOLDER.
   * This method is used by the constructor, and to switch from/to an immutable type. <P>
   * An immutable type forbids most of the updatings of the instance.
   * Indeed, it is possible to switch for example from 'actor' to
   * ''immutable actor', but not to 'immutable group'.
   * The rule is that the ''immutable ' key word, if it is used, must be at the start
   * or at the end of the type. <P>
   * From the default value '&lt;undefined&gt;', any updating may be done.
   * The user cannot change the type if it starts with 'byAcsCreator',
   * but the program can do it. To do that, '&lt;12345678901234567890&gt;' is
   * a program-reserved value. Even if it is not allowed by the ACS,
   * '&lt;12345678901234567890&gt;' is the only new value
   * it is possible to change from a type with '&lt;byAcsCreator&gt;'.
   * If the current type does not start with '&lt;byAcsCreator&gt;' but contains it,
   * the type is editable while the instance cannot be finalized by the user.
   * From '&lt;1234567890123456789&gt;', any new value may be set from the allowed ones,
   * and this starting value is the only one which allows to set a value containing
   * 'byAcsCreator'.  <P>
   * Since the typed privileges use depends on the type, this method registers
   * the typed privileges having the same type (modulo the key word 'immutable').
   * However, this is not true for the for_seed typed privileges,
   * nor for the specific ones, because a specific privilege cannot be created
   * if the source or the target is not there.
   * A switching from/to an immutable type to the other format never changes
   * the associated PrivilegeForTypes. <P>
   * Every change in this method should be done in BasicImpl.setType().
   * This method is called from the GUI, by finalizeForBase()
   * and by the constructor.
   * Calls ACS.getM_TypesAndGUIPolicy() to have the allowed values,
   * if any, for the keys 'VirtualFolder.ResourceVFType' and 'VirtualFolder.EPVFType'.
   * If there is no allowed values, all the values are possible. <P>
   * Calls also ACSImpl.getM_TypeOrientedPolicy(), BaseUtilityImpl.getObjectTypeWithoutImmutable().
   * Fires the event 'VirtualFolderType'.
   * @param _type as for instance "Cayro_file" or "tablespace".
   *  May be null, or equals to the current type.
   * @exception UpDateError if the new value and the old one are not compatible, or
   * if the type handling for privileges throws it.
   */
  public void setType(String _type) throws UpDateError {
      if (type_ != null && type_.equals(_type))
          return;
      else if (type_ == null && _type == null)
          return;

      // variables
      String[]            l_allowed = null;
      String              oldType = type_;
      List                l_priv;
      PrivilegeForType    priv;
      String[]            l_policy = null;
      String              reducedOldType = null;
      String              reducedType = null;
      
      if (VFType_.equals(RES_VFOLDER))
          l_allowed = (String[])aCS_.
                    getM_TypesAndGUIPolicy().get("VirtualFolder.ResourceVFType");
      else
          l_allowed = (String[])aCS_.
                    getM_TypesAndGUIPolicy().get("VirtualFolder.EPVFType");

      if (_type != null && !_type.equals("<12345678901234567890>") &&
              l_allowed != null && !Arrays.asList(l_allowed).contains(_type) ) {


                          // for testing purpose
                          //System.out.println("for "+getName()+", VirtualFolderImpl.setType() "+
                            //      "with exception for the type argument "+_type+
                              //    " for the Resources main VF type: "+VFType_.equals(RES_VFOLDER));
                          //System.out.println("   allowed values: ");
                          //for (int i = 0; i < l_allowed.length; i++) {
                            //  System.out.println(" ------- "+l_allowed[i]);
                          //} // for i


          throw new UpDateError(" The new type '"+_type+"' of "+name_.lastComponent()+
                  " is forbidden by the ACS. It is a Resources main VF type: "+
                  VFType_.equals(RES_VFOLDER));
      } // if
      
      if (type_.indexOf("<byAcsCreator>") != -1 &&
              (_type != null && !_type.equals("<12345678901234567890>")))
          throw new UpDateError(" "+name_.lastComponent()+" has a type which"+
                  " forbids a type editing.");
          
      else if (!type_.equals("<undefined>")) {

          if (_type != null && _type.startsWith("<byAcsCreator>") &&
                  !type_.equals("<12345678901234567890>"))
                throw new UpDateError(getName()+" new type cannot contain '<byAcsCreator>'.");
          if ( type_.indexOf("<byAcsCreator>") == -1 &&
                  !type_.equals("<12345678901234567890>")) {

              if (!type_.contains("<immutable> ") && (_type != null &&
                      (!_type.contains("<immutable> ") || !_type.contains(type_))))
                    throw new UpDateError(" The new type is not linked to the previous type,"+
                            " as 'actor' and '<immutable> actor' for instance");
              if (type_.contains("<immutable> ") && _type != null && !type_.contains(_type))
                    throw new UpDateError(" The "+getName().lastComponent()+" new type "+
                            _type+" is not derived from the previous one "+type_+", "+
                            " as '<immutable> actor' and 'actor'.");

          } // if ( type_.indexOf("<byAcsCreator>") == -1
      } // if (!type_.equals("<undefined>")

    type_ = _type;
    
    if (aCS_.getManagePrivilegeRights() && aCS_.getManageSimpleTypedPrivileges()
            // && aCS_.getManageClassedTypedPrivileges()
            // && aCS_.getManageForSeedTypedPrivileges()
            ) {
        try {
            // from/to an 'immutable' type does not change the associated typed privileges
            if (oldType == null || type_ == null ||
                    !type_.equals(UtilityImpl.getFirstInstance().
                        getObjectTypeWithoutImmutable(oldType)) ) {

                // unlistenings from 'oldType'
                if (oldType != null) {
                    reducedOldType = UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(oldType);

                    if (aCS_.getM_TypeOrientedPolicy() != null)
                        l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                                "VirtualFolder.SecondTypesOfPrivilegeFor."+reducedOldType);

                    // the reduced type is unuseful there
                    l_priv = aCS_.getEorL_PrivilegeForTypesForTarget(oldType);

                    for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                        priv = (PrivilegeForType)ite.next();
                        if (l_policy != null &&
                                !Arrays.asList(l_policy).contains(
                                    UtilityImpl.getFirstInstance().
                                        getObjectTypeWithoutImmutable(
                                            priv.getSecondType()) ))
                                                continue; // avoid to throw BaseError

                        if (priv.getMainType().equals(PrivilegeAbst.TYPED))
                          ((PrivilegeTarget)this).removeEorPrivilegeToTarget(priv);
                        else if (priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED) &&
                          Class.forName("ARoad0.gBaseInterface."+
                          ((PrivilegeForType)priv).getSourceClass())
                          .isAssignableFrom(this.getClass()) )
                                ((PrivilegeTarget)this).removeEorPrivilegeToTarget(priv);
                    } // for ite

                } // if (oldType != null) {

                // new listenings from 'type_'
                if (type_ != null) {
                    reducedType = UtilityImpl.getFirstInstance().
                                getObjectTypeWithoutImmutable(type_);

                    if (aCS_.getM_TypeOrientedPolicy() != null)
                        l_policy = (String[]) aCS_.getM_TypeOrientedPolicy().get(
                            "VirtualFolder.SecondTypesOfPrivilegeFor."+reducedType);

                    // the reduced type is unuseful there
                    l_priv = aCS_.getEorL_PrivilegeForTypesForTarget(type_);

                    for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                        priv = (PrivilegeForType)ite.next();
                        if (l_policy != null &&
                                !Arrays.asList(l_policy).contains(
                                    UtilityImpl.getFirstInstance().
                                        getObjectTypeWithoutImmutable(
                                            priv.getSecondType()) ))
                                                continue; // avoid to throw BaseError

                        // noop on TYPED_FOR_SEED and TYPED_CLASSED_SPECIFIC
                        if (priv.getMainType().equals(PrivilegeAbst.TYPED))
                          ((PrivilegeTarget)this).addEorPrivilegeForTypeToTarget(priv);
                        else if (priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED) &&
                          Class.forName("ARoad0.gBaseInterface."+
                          ((PrivilegeForType)priv).getSourceClass())
                          .isAssignableFrom(this.getClass()) )
                                ((PrivilegeTarget)this).addEorPrivilegeForTypeToTarget(priv);
                    } // for ite

                } // if (type_ != null) {
            } // if ((oldType == null

        } catch (BaseError b) {
          try {
              type_ = null;
              // caller of this method, but no loop since (type_ == null && _type == null)
              finalizeForBase();
          } catch (UpDateError r) {
            throw new InternalError("From VirtualFolderImpl.setType(): "+
                  r.toString());
          }
          throw new InternalError("From VirtualFolderImpl.setType(): "+
                  b.toString());
        } // catch
        catch (ClassNotFoundException c) {
          try {
              type_ = null;
              // caller of this method, but no loop since (type_ == null && _type == null)
              finalizeForBase();
          } catch (UpDateError r) {
            throw new InternalError("From VirtualFolderImpl.setType(): "+
                  r.toString());
          }
          throw new InternalError("From VirtualFolderImpl.setType(): "+
                  c.toString());
        } // catch
    } // if (aCS_.getManagePrivilegeRights()) {
    
    try {
          firePropertyChange("VirtualFolderType", oldType, type_);
    } catch (Exception e) {
        ARoad0.Gui1.DialogBox.showAlarmMessage(
          "There is an exception after the base updating, from outside the database."+
          "You should save the base and stop the program, then run it once again.");
        System.out.println("From BasicImpl.setType(), firePropertChange() throws:");
        e.printStackTrace();
    }

    l_allowed = null;
    oldType = null;
    l_priv = null;
    priv = null;
    l_policy = null;
    reducedOldType = null;
    reducedType = null;
  }  // setType()
  
  /**
   * Adds a member to this virtual folder.
   * For a RES_VFOLDER, it has to be a Resource and not an EligibleParty.
   * For a EP_VFOLDER, the member has to be a Resource and it may be an EligibleParty.
   * A member is not a child in the folders tree.
   * If each member has to be limited to one instance in the folders tree,
   * and this is controlled.
   * Uses ACSImpl.getM_TypeOrientedPolicy().'VirtualFolder.TypesOfMemberFor.'
   * to control the type of the new member.
   * Calls BasicImpl.addInVFolderList().
   * Overrides FolderAbst.addEorBaseObject() to control the new member and
   * to register this instance in the argument, then calls the super method.
   * @param _obj non-null ACSObject to add in the folder (Resource or EligibleParty)
   * @exception CreateError if the object is already known by the view,
   *   if it is known in the tree when it is restricted to one instance,
   *   or it is not a resource or an actor (following the main type),
   *   or if its ACS is not the folder ACS, or the object type is not authorized.
   */
  public synchronized void addEorBaseObject(BaseObject _obj)
    throws CreateError {
        
    if (!(_obj instanceof BasicImpl)) {
        getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects()"+
          "the argument is not a BasicImpl.");
        throw new CreateError("The new element is not a Basic.");
    }
    if (VFType_.equals(RES_VFOLDER) && 
            (!(_obj instanceof Resource) || (_obj instanceof EligibleParty))) {
        getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects()"+
          "the argument is not a Resource, or it is also an EligibleParty.");
        throw new CreateError("The new element is not a Resource, or it is also an EligibleParty.");
    } else if (VFType_.equals(EP_VFOLDER) && 
            (!(_obj instanceof Resource))
            ) {
        getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects()"+
          "the argument is not a Resource.");
        throw new CreateError("The new element is not a Resource.");
    } // setManageActorInEPVirtualFolders(true)

    if (VFType_.equals(RES_VFOLDER) && !aCS_.getManageSimpleResourceVirtualFolders() &&
            !(_obj instanceof Directory)) {
        getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects()"+
          "the argument is a simple Resource, and it is not managed by the ACS.");
        throw new CreateError("The new element is a simple Resource and this is forbidden.");
    }
    if (VFType_.equals(RES_VFOLDER) && !aCS_.getManageDirectoryInResourceVirtualFolders() &&
            _obj instanceof Directory) {
        getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects()"+
          "the argument is a Directory, and it is not managed by the ACS.");
        throw new CreateError("The new element is a Directory and this is forbidden.");
    } //
    if (VFType_.equals(EP_VFOLDER) && !aCS_.getManageActorInEPVirtualFolders() &&
            _obj instanceof Actor
            ) {
        getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects()"+
          "the argument is an Actor and it is not managed by the ACS.");
        throw new CreateError("The new element is an Actor and this is forbidden.");
    } //

    if (!((ACSObject)_obj).getAcsName().equals(aCS_.getName())) {
        getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects()"+
          "the argument has not the right ACS.");
        throw new CreateError("The new element has not the right ACS.");
    }

    if (type_.startsWith("<immutable> ") || type_.endsWith("<immutable> "))
          throw new CreateError(
            " The type of "+getName().lastComponent()+" sets it as immutable for its members.");

    if (aCS_.getM_TypeOrientedPolicy() != null &&
            aCS_.getM_TypeOrientedPolicy().get(
                "VirtualFolder.TypesOfMemberFor."+type_) != null) {

                          // for testing purpose
                          //System.out.println("for "+getName()+
                            //      ", VirtualFolderImpl.addEorBaseObject() "+
                              //    "with the object type "+((BasicImpl)_obj).getType());

        if (!Arrays.asList((String[])aCS_.getM_TypeOrientedPolicy().get(
            "VirtualFolder.TypesOfMemberFor."+type_)).contains(
                 UtilityImpl.getFirstInstance().
                     getObjectTypeWithoutImmutable(((BasicImpl)_obj).getType()) )) {

                          // for testing purpose
                          System.out.println("for "+getName()+
                                  ", VirtualFolderImpl.addEorBaseObject() "+
                                  "negative test of the key "+
                                  "VirtualFolder.TypesOfMemberFor."+type_);

                throw new CreateError(
                    "The type of the new element is not authorized for this virtual folder.");
        } // if
    } // if

    if (memberOneTimeInTree_ && detectMemberInFolderTree(aCS_, VFType_, (ACSObject)_obj)) {
          getEorBaseError().addException("in VirtualFolderImpl.addEorBaseObjects(_obj)"+
             "_obj is already in the folders tree and cannot be added.");
          throw new CreateError(
             "The new element is already in the folders tree of this folder.");
    }

    ((BasicImpl)_obj).addInVFolderList(this);
    
    super.addEorBaseObject(_obj);
  } // addEorBaseObject()

  /**
   * Removes a member to the folder and unregisters it in the object.
   * Overrides the superclass method.
   * @param _obj non-null object to remove in the folder
   * @exception CreateError if the object name is empty, or
   *    the object is not known by the folder.
   */
  public synchronized void removeEorBaseObject(BaseObject _obj)
    throws CreateError {
        super.removeEorBaseObject(_obj);

        ((BasicImpl)_obj).removeInVFolderList(this);
  } //removeEorBaseObject()

   /**
   * Get the open ViewInBases containing this BaseObject.
   * @return List of ViewInBase. May be empty,
   *   but never null.
   */
  public List getImpliedViews() {
      if (folderMember_ == null) return Collections.emptyList();
      else return folderMember_.getImpliedViews();
  } // getImpliedViews()

  /**
   * Get the unique ACSObject, if any, of this ACS as reference for which this object
   * is an internal hard alias.
   * @return the reference. May be null.
   */
  public ACSObject getEorReferenceAsInternalHardAlias() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getEorReferenceAsInternalHardAlias();
  } // getEorReferenceAsInternalHardAlias()

  /**
   * Get the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard reference.
   * @return the hard alias. May be null.
   */
  public ACSObject getEorAliasAsInternalHardReference() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getEorAliasAsInternalHardReference();      
  } // getEorAliasAsInternalHardReference()

  /**
   * This property is not displayed in the GUI.
  * @return true if there is at least one external soft alias.
  */
  public final boolean getHasSoftAlias() {
    if (aliasHandler_ == null ) {
        return false;
    } // if
    return aliasHandler_.getHasSoftAlias();
  } // getHasSoftAlias()

  /**
  * Gets a sorted array of names for the external soft aliases.
  * @return ImmutableName[]. May be null but not empty.
  */
  public final ImmutableName[] getL_SoftAlias() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getL_SoftAlias();
  } // getL_SoftAlias()

  /**
  * Gets a sorted array of names of the ACS of the soft references.
  * @return ImmutableName[] of the ACS. May be null but not empty.
  */
  public final ImmutableName[] getL_AcsOfSoftReferences() {
    if (aliasHandler_ == null ) {
        return null;
    } // if
    return aliasHandler_.getL_AcsOfSoftReferences();
  } // getL_SoftAlias()
  
/**
 * Tests of equals() on name, children and parent.
 * @param _obj to compare to this instance
 * @return true if _obj is a VirtualFolderImpl instance (or an instance of an
 *  extended class) equals to the current one
 */
public synchronized boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;

      //if (!super.equals(_obj))
        //  return false;

      VirtualFolderImpl obj = (VirtualFolderImpl) _obj;

      if (name_ == null || obj.name_== null) {
          if (name_ != null || obj.name_ != null)
            return false;
      } // if
      else if (!name_.equals(obj.name_))
        return false;

      // getEorL_Children() is never null, but to have the same pattern, we do it
      if (nodeHandler_ == null || obj.nodeHandler_ == null) {
          if (nodeHandler_ != null || obj.nodeHandler_ != null)
            return false;
      } // if
      else if (!nodeHandler_.equals(obj.nodeHandler_))
        return false;

      if (getEorParent() == null || obj.getEorParent() == null) {
          if (getEorParent() != null || obj.getEorParent() != null)
            return false;
      } // if
      else {
          if (!getEorParent().getName().equals( obj.getEorParent().getName() ))
            return false;
      } // else

      return true;
  } // equals
  
  /**
  * Adds a listener to the bean.
  * Used for the superclass properties, and for: 'ImpliedOpenViews',
  * 'OneTargetPrivilege', 'TaPrivReadyFromParent',
  * 'AllowTaPrivFromParent', 'Children'.
  * The following properties are managed by aliasHandler_:
  * 'HardAliasReference', 'HardAlias', 'SoftAlias', 'AcsOfSoftReferences'
  * and 'hasSoftAlias'.
  * @param _l to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public synchronized void addPropertyChangeListener(PropertyChangeListener _l){
    super.addPropertyChangeListener(_l);

    try {
        if (folderMember_ == null)
            folderMember_ = new CFolderMemberImpl(this);
    } catch (CreateError c) {
        throw new InternalError(c.toString());
    }
    folderMember_.addPropertyChangeListener(_l);

    privHandler_.addPropertyChangeListener(_l);
    nodeHandler_.addPropertyChangeListener(_l);

    if (aliasHandler_ != null) aliasHandler_.addPropertyChangeListener(_l);
    else {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }

        aliasHandler_.addPropertyChangeListener(_l);
    } // else
  } // addPropertyChangeListener()

  /**
  * Adds a listener to the bean.
  * Used for the superclass properties, and for: 'ImpliedOpenViews',
  * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege', 'TaPrivReadyFromParent',
  * 'AllowTaPrivFromParent', 'Children'.
  * The following properties are managed by aliasHandler_:
  * 'HardAliasReference', 'HardAlias', 'SoftAlias', 'AcsOfSoftReferences'
  * and 'hasSoftAlias'.
  * @param _propertyName is the name of the property
  * @param _l to add. Caution: the listener must NOT
  *   have specific hashCode() and equals() methods, to work with HashSet
  *   in the removing operations.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l) {

      if (_propertyName.equals("OneTargetTypedPrivilege") ||
              _propertyName.equals("OneTargetLinkedPrivilege") ||
        _propertyName.equals("TaPrivReadyFromParent") ||
        _propertyName.equals("AllowTaPrivFromParent")) {
            privHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else if (_propertyName.equals("ImpliedOpenViews") ) {
            try {
                if (folderMember_ == null)
                    folderMember_ = new CFolderMemberImpl(this);
            } catch (CreateError c) {
                throw new InternalError(c.toString());
            }

            folderMember_.addPropertyChangeListener(_propertyName, _l);
      }
      else if (_propertyName.equals("Children") ) {
            nodeHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else if (_propertyName.equals("HardAliasReference") ||
                        _propertyName.equals("HardAlias") ||
                        _propertyName.equals("SoftAlias") ||
                        _propertyName.equals("AcsOfSoftReferences") ||
                        _propertyName.equals("hasSoftAlias")) {
            if (aliasHandler_ == null ) {
                try {
                  aliasHandler_ = new CAliasImpl(this);
                } catch (CreateError c) {
                    new InternalError("From "+name_+"- "+c.toString());
                }
            } // if

            aliasHandler_.addPropertyChangeListener(_propertyName, _l);
      }
      else super.addPropertyChangeListener(_propertyName, _l);;
  }

  /**
  * removes a listener to the bean.
  * @param _l to remove
  */
  public synchronized void removePropertyChangeListener(PropertyChangeListener _l){
    super.removePropertyChangeListener(_l);

    if (nodeHandler_ == null) return;

    if (folderMember_ != null)
        folderMember_.removePropertyChangeListener(_l);

    privHandler_.removePropertyChangeListener(_l);
    nodeHandler_.removePropertyChangeListener(_l);

    if (aliasHandler_ != null)
      aliasHandler_.removePropertyChangeListener(_l);
  }

  /**
  * removes a listener to the bean for the properties:
  * Used for the superclass properties, and for: 'ImpliedOpenViews',
  * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege', 'TaPrivReadyFromParent',
  * 'AllowTaPrivFromParent', 'Children'.
  * The following properties are managed by aliasHandler_:
  * 'HardAliasReference', 'HardAlias', 'SoftAlias', 'AcsOfSoftReferences'
  * and 'hasSoftAlias'.
  * @param _propertyName is the name of the property
  * @param _l to remove
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l){
      // no action if the transiant variable is null
      // It is null when a PropertyEditor calls back a finalized instance.
      if (nodeHandler_ == null) return;
      
      if (_propertyName.equals("OneTargetTypedPrivilege") ||
              _propertyName.equals("OneTargetLinkedPrivilege") ||
        _propertyName.equals("TaPrivReadyFromParent") ||
        _propertyName.equals("AllowTaPrivFromParent")) {
            privHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else if (_propertyName.equals("ImpliedOpenViews") ) {
            if (folderMember_ != null)
                    folderMember_.removePropertyChangeListener(_propertyName, _l);
      }
      else if (_propertyName.equals("Children") ) {
            nodeHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else if (_propertyName.equals("HardAliasReference") ||
                        _propertyName.equals("HardAlias") ||
                        _propertyName.equals("SoftAlias") ||
                        _propertyName.equals("AcsOfSoftReferences") ||
                        _propertyName.equals("hasSoftAlias")) {
            if (aliasHandler_ != null )
                aliasHandler_.removePropertyChangeListener(_propertyName, _l);
      }
      else super.removePropertyChangeListener(_propertyName, _l);
  } // removePropertyChangeListener()

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName) {
      // no action if the transiant variable is null
      if (nodeHandler_ == null) return null;

                // for testing and tracking
                //if (_propertyName.equals("AclRights"))
                  //  System.out.println("VirtualFolderImpl.getPropertyChangeListeners(), call from "+
                    //_l.getClass()+" to "+name_+" for the property "+_propertyName);

      if (_propertyName.equals("OneTargetTypedPrivilege") ||
              _propertyName.equals("OneTargetLinkedPrivilege") ||
        _propertyName.equals("TaPrivReadyFromParent") ||
        _propertyName.equals("AllowTaPrivFromParent")) {
            return privHandler_.getPropertyChangeListeners(_propertyName);
      }
      else if (_propertyName.equals("ImpliedOpenViews") ) {
            if (folderMember_ != null)
               return folderMember_.getPropertyChangeListeners(_propertyName);
            else
                return null;
      }
      else if (_propertyName.equals("Children") ) {
            return nodeHandler_.getPropertyChangeListeners(_propertyName);
      }
      else if (_propertyName.equals("HardAliasReference") ||
                        _propertyName.equals("HardAlias") ||
                        _propertyName.equals("SoftAlias") ||
                        _propertyName.equals("AcsOfSoftReferences") ||
                        _propertyName.equals("hasSoftAlias")) {
            if (aliasHandler_ != null )
                return aliasHandler_.getPropertyChangeListeners(_propertyName);
            else
                return null;
      }
      else return changeSupport_.getPropertyChangeListeners(_propertyName);

  } // getPropertyChangeListeners()

        //
        // ------------- methods out of the extended interfaces --------------------
        //
  
  /**
   * Set the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard alias.
   * The generic gBase package does not call this method, nor the user from the GUI.
   * It is the responsibility of the ACS associations to set an internal alias
   * for one new ACSObject, as an alias or as an alias reference.
   * Fires the property change 'HardAliasReference'.
   * @param _ref the reference. May be null.
   * @exception UpDateError _ref does not belong to the same ACS or the type is
   * immutable
   */
  protected void setEorReferenceAsInternalHardAlias(ACSObject _ref) throws UpDateError {
    // noop if finalized
    if (nodeHandler_ == null) {
      return;
    }
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new UpDateError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.setEorReferenceAsInternalHardAlias(_ref);
  } // setEorReferenceAsInternalHardAlias()

  /**
   * Set the unique ACSObject, if any, of this ACS for which this object
   * is an internal hard reference.
   * The generic gBase package does not call this method, nor the user from the GUI.
   * It is the responsibility of the ACS associations or an ACS addon to set
   * an internal alias for one new ACSObject, as an alias or as an alias reference.
   * If the instance is already an alias, it cannot become an alias reference.
   * Fires the property change 'HardAlias'.
   * @param _ali the alias. May be null.
   * @exception UpDateError _basic does not belong to the same ACS or
   * the master is already an alias
   */
  protected void setEorAliasAsInternalHardReference(ACSObject _ali) throws UpDateError {
    // noop if finalized
    if (nodeHandler_ == null) {
      return;
    }
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new UpDateError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.setEorAliasAsInternalHardReference(_ali);      
  } // setEorAliasAsInternalHardReference()

  /**
   * Adds an internal or external soft alias.
   * The alias is not updated. Used only by the ACS.
   * Has to be used by an AcsAddon from a gBase class.
   * Fires the events 'SoftAlias' and 'hasSoftAlias'.
   * @param _alias internal or external ACSObject. No action if null, or already known.
   * @exception UpDateError _alias belongs to the same ACS or the type is
   * immutable
   * @exception InternalError if the list is null. If
   *  firePropertyChange() throws an exception, displays an user message.
   */
  protected void addEorSoftAlias(ImmutableName _alias) throws UpDateError {
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.addEorSoftAlias(_alias);;
  } // addEorSoftAlias()

  /**
   * Removes a soft alias.
   * The alias is not updated. Used only by the ACS, in an AcsAddon and
   * from a gBase class.
   * Detects the alias through a name comparison.
   * Fires the 'SoftAlias' and the 'hasSoftAlias' property change events.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the list is null.
   * @param _alias is an external alias. No action if null.
   * @exception UpDateError if _alias is not known.
   */
  protected void removeEorSoftAlias(ImmutableName _alias) throws UpDateError {
    // noop if finalized
    if (nodeHandler_ == null) {
      return;
    }
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.removeEorSoftAlias(_alias);
  } // removeEorSoftAlias()

  /**
   * Adds the ACS name of an external or internal soft reference to the master as soft alias.
   * The reference is not updated there but by the ACS calling method.
   * Used only by the ACS, which may add controls on the parameter class.
   * Has to be used by the AcsAddon.
   * Fires the event 'AcsOfSoftReferences'.
   * @param _acs ACS name. No action if null, or already known.
   * @exception UpDateError if the type is immutable
   */
  protected void addAcsOfSoftReferences(ImmutableName _acs)  throws UpDateError {
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.addEorSoftAlias(_acs);      
  } // addAcsOfSoftReferences()

  /**
   * Removes an external or internal alias from the master as soft reference.
   * The alias is not updated there but by the ACS calling method.
   * Used only by the ACS, which may add controls on the parameter class.
   * Detects the alias through a name comparison.
   * Fires 'AcsOfSoftReferences' property change events.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the list is null.
   * @param _acs ACS name. No action if null.
   * @exception UpDateError if _acs is not known.
   */
  protected void removeAcsOfSoftReferences(ImmutableName _acs) throws UpDateError {
    if (folderMember_ == null) return;
    if (aliasHandler_ == null ) {
        try {
          aliasHandler_ = new CAliasImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    if (type_.startsWith("<byAcsCreator>") || type_.startsWith("<immutable> ") ||
              type_.endsWith("<immutable> "))
          throw new UpDateError(
            "The type of "+getName().lastComponent()+
            " sets it as immutable for its reference.");

    aliasHandler_.removeAcsOfSoftReferences(_acs);
  } // removeAcsOfSoftReferences()

  /**
   * This method controls the immutability of the object, removes the associates,
   * then calls finalizeForBase().
   * @exception UpDateError if the object type contains 'byAcsCreator', starts with
   * or ends with 'immutable', from finalizeFromBase() or
  * AssociateUtilityImpl.removeAssociatesAfterDeletion()
  */
  public synchronized void finalizeForUser() throws UpDateError {
    // noop if it is the second call to this method
    if (nodeHandler_ == null) {
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");

      return;
    }
    
    if (type_.indexOf("<byAcsCreator>") != -1 || type_.startsWith("<immutable>") ||
            type_.endsWith("<immutable>"))
      throw new UpDateError(
        "The type of '"+getName().lastComponent()+"' sets it as immutable.");

    // removes the associates, if any
    try {
        AssociateUtilityImpl.removeAssociatesAfterDeletion(this, (ACS)aCS_);
    } catch (BaseError b) {
        throw new UpDateError("from AssociateUtilityImpl.removeAssociatesAfterDeletion(), "+
                b.toString());
    }
    
    finalizeForBase();
  } // finalizeForUser()

  /**
  * All the internal maps are set to null and the folder is unregistered in its ACS.
  * Called by finalizeForUser(). Calls superclass finalizeForBase().
  * The relevant sources and targets register/unregister the typed privilege when the source,
  * the target or the privilege is created/deleted, so sources and targets know
  * their typed privileges, but typed privileges do not know the sources and
  * targets which use them. This is NOT true for TYPED_FOR_SEED privileges,
  * which have external sources and targets having no link to these privileges.
  * Calls setType() to unregister the typed privielges.
  * @exception UpDateError from Basic.removeInVFolderList()
  */
  public synchronized void finalizeForBase() throws UpDateError {
    // noop if it is the second call to this method
    if (nodeHandler_ == null) {
              // for testing purpose
              //System.out.println("for "+getName()+", second call to finalization.");

      return;
    }

    // takes out this leaf to its parent, if the parent
    // is not already unrecorded in the main ACS by its own finalize() method.
    if (parent_ != null && aCS_ != null && aCS_.containsVirtualFolderName(parent_.getName() )) {
        ((VirtualFolderImpl)parent_).removeEorChild(this);
    } // if parent_ != null

    SortedMap     m_obj = (SortedMap)((TreeMap)m_BaseObjects_).clone();
    // removes as a listener (not as a listened instance)
    try {
        for (Iterator ite = m_obj.values().iterator() ; ite.hasNext() ; ) {
          ((BasicImpl)ite.next()).removeInVFolderList(this);
        }
    } catch (CreateError b) {
        throw new UpDateError(b.toString());        
    }

    // unregisters the typed privileges
    setType(null);
    if (folderMember_ != null) folderMember_.finalizeForBase();
    // finalizes all the children
    nodeHandler_.finalizeForBase();
    // finalizes only OTO LPRI
    privHandler_.finalizeForBase();
    folderMember_ = null;
    nodeHandler_ = null;
    privHandler_ = null;
    if (aliasHandler_ != null) aliasHandler_.finalizeForBase();
    aliasHandler_ = null;
    
    ((ACSImpl)aCS_).removeEorVirtualFolder(this); // uses name_ and VFType_
    aCS_ = null;
    parent_ = null;
    VFType_ = null;
    
    super.finalizeForBase();
  } // finalizeForBase()

  /**
   * The comparison is based on the name compareTo() method of each
   * leaf, except when there is a node.
   * The nodes are always smaller than the leaves.
   * @param _o to compare must be an ImmutableLeaf
   * @exception ClassCastException - if the specified object's type
   *  prevents it from being compared to this current Object
   * @return int 0 if the leaves are equals, a negative integer if this leaf
   * is lesser than the argument, and a positive integer if this leaf
   * is greater than the argument.
   */
  public int compareTo(Object _o) throws ClassCastException {
    if (!(_o instanceof ImmutableLeaf))
      throw new ClassCastException("in VirtualFolder.compareTo(), the argument "+
        "is not an instance of ImmutableLeaf");
    // if (equals(o) ) return 0;
    if (getName() == null || _o == null || ((ImmutableLeaf)_o).getName() == null)
        throw new InternalError("in VirtualFolder.compareTo(),"+
           " the object name, the argument or the argument name is null");
    if (this instanceof ImmutableNode && !(_o instanceof ImmutableNode)) return -1;
    else if (!(this instanceof ImmutableNode) && (_o instanceof ImmutableNode)) return +1;
    else return getName().compareTo( ((ImmutableLeaf)_o).getName() );
  }

  /**
   * @return ImmutableName of the ACS that owns this object
   */  
  public ImmutableName getAcsName() {
      return aCS_.getName();
  }

  /**
  * An array of external object references on the children of this node.
  * Children are not members.
  * The returned type is set to allow BaseUtilityImpl to found
  * the true returned type from the method signature, for BaseObjectHandler
  * when it has to create a new child from the property editor.
  * @return the children. May be null but not empty.
  */
  public ImmutableVirtualFolder[] getEorL_Children() {
      ImmutableLeaf[] l_children = nodeHandler_.getEorL_Children();
      if (l_children == null) return null;
      ImmutableVirtualFolder[] result = new ImmutableVirtualFolder[l_children.length];
      System.arraycopy(l_children, 0, result, 0, l_children.length);
      l_children = null;
      return result;
  }

  /**
   * Gets the current parent node of the leaf
   * @return ImmutableNode is a reference on the current parent node. May be null.
   */
  public ImmutableNode getEorParent() {
      return parent_;
  }

   /**
   * Gets all the PrivilegeForLinks linked to this target, not the
   * privilegeForTypes. Returns only the direct privileges,
   * since the cascading is not activated in VirtualFolders.
   * @return List of PrivilegeForLinks. May be empty, but never null.
   */
  public List getEorL_TargetPrivilegeForLinks() {
      return privHandler_.getEorL_TargetPrivilegeForLinks();
  }

   /**
   * Gets all the PrivilegeForTypes of this target.
   * Does not include the TYPED_FOR_SEED and TYPED_CLASSED_SPECIFIC privileges.
   * @return List of PrivilegeForTypes. May be empty, but never null.
   */
  public List getEorL_TargetPrivilegeForTypes() {
      return privHandler_.getEorL_TargetPrivilegeForTypes();
  }

   /**
   * Gets all the privileges linked to this target.
   * There is two types of privilege links (allowed or not by the ACS): <P>
   * - PrivilegeForLinks: linked privilege established for a couple
   * (ACSObject, privilege), <P>
   * - PrivilegeForType: typed privilege, where the ACSObject and the privilege
   * has a common type, but without the TYPED_FOR_SEED and
   * TYPED_CLASSED_SPECIFIC privileges. <P>
   * @return List of Privileges. May be empty, but never null.
   */
  public List getEorL_TargetPrivileges() {
      return privHandler_.getEorL_TargetPrivileges();
  }

   /**
   * Sets a privilege as applicable to this access target, and called only by the program.
   * This method is called only by PrivilegeForTypeImpl 
   * at any creation/deletion of a typed privilege or an access target.
   * For PrivilegeForTypes, only the TYPED and TYPED_CLASSED privilege may be argument.
   * This is false for the TYPED_CLASSED_SPECIFIC and TYPED_FOR_SEED privileges,
   * and this produces the throwning of an InternalError.
   * Called by setType() to the indirect request of the VirtualFolderImpl constructor.
   * Fires 'OneTargetTypedPrivilege' events.
   * @param _priv to register in the target
   * @exception BaseError if this instance is not a privilege target,
   * if the privilege second type is not authorized for this target.
   */
  public void addEorPrivilegeForTypeToTarget(PrivilegeForType _priv) throws BaseError {

      //if (type_ != null && (type_.startsWith("<immutable> ") ||  type_.endsWith("<immutable> ")))
        //  throw new BaseError(
          //  "The type of "+getNickName()+" sets it as immutable for its privileges.");

      privHandler_.addEorPrivilegeToTarget(_priv);
  }

   /**
   * Sets a privilege as applicable to this access target.n.
   * Called by PrivilegeForLinksImpl.addEorPrivilegeTargetLink() which checks
   * the type of the target to forbid a change on an immutable target.
   * Fires 'OneTargetLinkedPrivilege' events.
   * @param _priv to register in the target
   * @exception BaseError if this instance is not a privilege target,
   * if the privilege second type is not authorized for this target.
   */
  protected void addEorPrivilegeForLinksToTarget(PrivilegeForLinks _priv) throws BaseError {
      privHandler_.addEorPrivilegeToTarget(_priv);
  }

   /**
   * Unregisters the privilege when it is no more used by this privilege target.
   * Called by PrivilegeForLinksImpl and PrivilegeForTypeImpl, NOT by the GUI.
   * PrivilegeForLinksImpl.removeEorPrivilegeTargetLink() checks the type of the target
   * to forbid a change on an immutable target.
   * Fires 'OneTargetTypedPrivilege' or 'OneTargetLinkedPrivilege' events.
   * @param _priv to remove
   * @exception BaseError if the privilege and this source is unknown.
   */
  public synchronized void removeEorPrivilegeToTarget(Privilege _priv) throws BaseError {
      privHandler_.removeEorPrivilegeToTarget(_priv);
  }

   /**
   * Gets all the immutable list of the inherited linked privileges of the target.
   * The LINKED_IF_ALL_SOURCE type is forbidden, and any privilege having more than one target.
   * To use by the consumers in the gWork packages, which deny any access if
   * the returned value is null when the privilege inheritance is managed.
   * It is recommanded to avoid getEorL_TargetInheritedLinkedPrivileges() for these consumers.
   * Null is the default and immutable returned value in this generic gBase class,
   * and this has to be changed in an AcsAddon implementing the inheritance
   * Called by gWork.LinkRights.withAccessThroughNodesTreeFromEP() and
   * getL_accessRightsThroughNodesTree().
   * Caution: a clone is not registered in its ACS, nor in its sources and targets.
   * However, it is possible in the beamer to navigate from a cloned privilege
   * to its targets or sources.
   * @return immutable list of Linked Privileges. May be null but never empty.
   */
  public List<PrivilegeForLinks> getEorImmutableL_TargetInheritedLinkedPrivileges() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorImmutableL_TargetInheritedLinkedPrivileges();
  }

   /**
   * Gets all the inherited linked privileges of this target.
   * The LINKED_IF_ALL_SOURCE type is forbidden, and any privilege having more than one target.
   * To use by the parent in the gBase package, to prepare the updating of the list
   * with a call to setEorL_TargetInheritedLinkedPrivileges().
   * No use in this generic gBase package, so for the AcsAddon packages.
   * In the BeanInfo class of the VirtualFolder subclass that extends this class in an
   * AcsAddon package, this property has to be put in the pane '  PRI Inheritance  '.
   * The privilege may be a clone, and the target may not be this instance.
   * Caution: a clone is not registered in its ACS, nor in its sources and targets.
   * However, it is possible in the beamer to navigate from a cloned privilege
   * to its targets or sources.
   * @return mutable list of Linked Privileges. May be a new empty list or an old
   * empty list, but never null.
   */
  protected List<PrivilegeForLinks> getEorL_TargetInheritedLinkedPrivileges() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getEorL_TargetInheritedLinkedPrivileges();
  }

  /**
   * Gets the internal map of the inherited linked privilege rights, where
   * the key is the name of the EligibleParty having the rights, extended with,
   * as last component, the parent order of the direct or indirect parent
   * which has given the rights.
   * For instance, the direct parent has the string '1', and its direct parent
   * has the string '2' as parent order in the last name component, and so on.
   * The map value is a non-null StringRight array of inherited rights, and it may be empty.
   * Called by gWork.LinkRights.withAccessThroughNodesTreeFromEP() and
   * getL_accessRightsThroughNodesTree(), but only in the AcsAddons implementing
   * the linked privilege inheritance. Such an AcsAddon may also choose to not use this property.
   * @return unmodifiable map of inherited rights from the linked privileges.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedLinkedPRIRights() {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    return privHandler_.getM_TargetInheritedLinkedPRIRights();
  }

   /**
   * Sets all the inherited linked privileges for this target.
   * Called only by the parents.
   * Do not register this instance in the inherited linked privileges.
   * Calls setM_TargetInheritedLinkedPRIRights() to set the map (source name+order, StringRight[]).
   * No use in this generic gBase package, so for the AcsAddon packages.
   * Fires 'InheritedTargetPrivileges' events.
   * @param _l_priv list of one-target Linked Privileges. May be empty, but never null.
   * @throws InternalError if a privilege has the LINKED_IF_ALL_SOURCE type, or
   * if it has not exactly one target.
   */
  protected void setEorL_TargetInheritedLinkedPrivileges(ArrayList<PrivilegeForLinks> _l_priv) {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    privHandler_.setEorL_TargetInheritedLinkedPrivileges(_l_priv);
  }
 
    /**
    * Resets the inherited Privilege rights after an user edition.
    * Called by the parent when there is a change in the effective rights
    * of an inherited Privilege. This is only when there is a change in the sources.
    * Contrary to AclEntry, the rights are immutable and there is no conditional group.
    * The subclassing of CPrivilegeUserImpl is possible in an AcsAddon, but it is
    * also possible to overread the same method in a subclass of this class.
    * Not described in the BeanInfo since the user does not call it.
    * Note: the creation of the inherited Privilege in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited Privilege
    * is handled.
    * Called by resetInheritedPrivilegeRightsOfChildren().
    * @param _lpri inherited Privilege with updated rights. Never null.
    */
  public synchronized void resetInheritedPrivilegeRights(PrivilegeForLinks _lpri) {
    if (privHandler_ == null ) {
        try {
          privHandler_ = new CPrivilegeUserImpl(this);
        } catch (CreateError c) {
            new InternalError("From "+name_+"- "+c.toString());
        }
    } // if
    privHandler_.resetInheritedPrivilegeRights(_lpri);
  } // resetInheritedPrivilegeRights()

    /**
    * Returns the Privilege to use as inherited Privilege in a child.
    * Empty method which returns the argument. Should be never called, and throws
    * an InternalError. Has to be overridden into an AcsAddon handling 
    * the privilege inheritance.
    * @param _lpri the initial Privilege which may be given to the child
    * @param _type type of a child, as VirtualFolder
    * @return the inherited Privilege if the child inherits this initial Privilege, or null
    */
  public PrivilegeForLinks selectInheritedPrivilege(PrivilegeForLinks _lpri, String _type) {

                            // always thrown
                            if (_lpri instanceof PrivilegeForLinks) {
                                InternalError i = new InternalError("In VirtualFolderImpl, call of " +
                                 "selectInheritedPrivilege()");
                                i.printStackTrace();
                                throw i;
                            }

      return _lpri;
  } // selectInheritedPrivilege()

    /**
    * Resets the inherited Privilege rights in the children, if the Privilege
    * inheritance is applicable.
    * There are two flows of data to manage the inherited Privileges: <P>
    * - one at the privilege creation/deletion, and this is not done in the generic gBase package, <P>
    * - one at their sources editing, <P>
    * This method is called for handling the second case,
    * to update the children of the VirtualFolder which is the target of the Privilege.
    * Not described in the BeanInfo since the user does not call it.
    * Called by PrivilegeForLinksImpl.resetRightsAsInheritedPrivilege().
    * @param _lpri inherited privilege with updated sources
    */
  public synchronized void resetInheritedPrivilegeRightsOfChildren(PrivilegeForLinks _lpri) {
    if (!aCS_.getManageLPRIRightsInheritance()) {
          InternalError i = new InternalError("VirtualFolderImpl."+
                    "resetInheritedPrivilegeRightsOfChildren() "+
                    "is called in an ACS that does not manage the PRI inheritance. ");
          i.printStackTrace();
          throw i;
    } // if

    if (_lpri.getEorLinkedTargets().get(0) != this) {
          InternalError i = new InternalError("VirtualFolderImpl."+
                    "resetInheritedPrivilegeRightsOfChildren() "+
                    "is called by a Privilege with a wrong target, on "+
                    name_.lastComponent());
          i.printStackTrace();
          throw i;
    } // if

    // reset the Privilege as inherited Privilege on the children
    ImmutableVirtualFolder[] l_child = getEorL_Children();
    PrivilegeForLinks pri = null;

    if (l_child != null) {
        for (int i = 0; i < l_child.length; i++) {

            pri = selectInheritedPrivilege(_lpri, l_child[i].getType());
            if (pri == null) continue;

            ((VirtualFolder)l_child[i]).resetInheritedPrivilegeRights(pri);

        } // for i
    } // if (l_child != null)

    l_child = null;
    pri = null;
  } // resetInheritedPrivilegeRightsOfChildren()

/**
 * @return boolean true if the ACS does not allow an ACSObject to
 *   be more than one time in the folders tree of this folder.
 */
  public boolean getMemberOneTimeInTree() {
      return memberOneTimeInTree_;
  }
  
  //
  // ------------   protected and package methods ------------------
  //

   /**
   * Registers the view as containing this BaseObject.
   * Called by EPRViewInBaseImpl.addEorResource/EligibleParty().
   * Fires 'ImpliedOpenViews' event.
   * @param _view to add
   */
  protected void addInViewList(ViewInBase _view) {
      try {
          if (folderMember_ == null)
            folderMember_ = new CFolderMemberImpl(this);
      } catch (CreateError c) {
            throw new InternalError(c.toString());
      }

      folderMember_.addInViewList(_view);
  } // addInViewList()

   /**
   * Unregisters the view when it does not contain this BaseObject.
   * Called by EPRViewInBaseImpl.removeEorResource/EligibleParty().
   * Fires 'ImpliedOpenViews' event.
   * @param _view to remove
   * @return false if the view is unknown
   */
  protected boolean removeInViewList(ViewInBase _view) {
      try {
          if (folderMember_ == null)
            folderMember_ = new CFolderMemberImpl(this);
      } catch (CreateError c) {
            throw new InternalError(c.toString());
      }

      return folderMember_.removeInViewList(_view);
  } // removeInViewList()
  
  /**
   * Adds a child to this Node. A child is not a member.
   * Used by the children constructor.
   * Does not update the parent internal variable in the child.
   * No control on the ACS type policy, since it is done
   * in the VirtualFolderImpl constructor which calls this class
   * for adding a new child. A parent type which contains 'immutable' does not forbid
   * to add children to the parent.
   * Detects the child through a name comparison.
   * Fires the event 'Children'.
   * @param _child is a VirtualFolder (not a simple leaf) to put in this Node.
   *   No action if null, or already known.
   * @exception InternalError if the leaves list is null, or if _child
   *  is not a VirtualFolder. If firePropertyChange() throws an exception,
   *  displays an user message.
   */
  protected synchronized void addEorChild(ImmutableLeaf _child) {
      if (!(_child instanceof VirtualFolder))
          throw new InternalError("In VirtualFolderImpl.addEorChild(), "+
            "the child is not a virtual folder.");
      nodeHandler_.addEorChild(_child);
  }
  
  /**
   * Removes a child from this Node.
   * Used by the children finalizing methods.
   * Does not update the parent internal variable in the child.
   * Detects the child presence through a name comparison.
   * If firePropertyChange() throws an exception, displays an user message.
   * No action if the children list is null, to speed up the
   * finalizing of this instance, when children are called to be finalizing.
   * @param _child is a virtual folder which is in this CNode. No action if
   *  null.
   * @exception UpDateError if _child is not known.
   */
  protected synchronized void removeEorChild(ImmutableLeaf _child) throws UpDateError {
      nodeHandler_.removeEorChild(_child);
  }

  /**
  * This method does not finalize the attributes, since there are only weak links.
  * It it is not necessary to cut them after a deserialization.
  * The instance is NOT removed as a listener, since there is no listener.
  * The internal maps are set to null, but not the private bound properties sets.
  * Does not call finalizeForBase(). Called by EPRViewInBaseImpl only.
   */
  protected synchronized void finalizeFromDeserialization() {

                //System.out.println("In ResourceImpl, call of " +
                //name_ + ".finalizeFromDeserialization()");

    if (aCS_ == null) return;

    aCS_ = null;
    parent_ = null;
    if (nodeHandler_ != null) nodeHandler_.finalizeFromDeserialization();
    if (privHandler_ != null) privHandler_.finalizeFromDeserialization();
    if (folderMember_ != null) folderMember_.finalizeFromDeserialization();
    if (aliasHandler_ != null) aliasHandler_.finalizeFromDeserialization();
    nodeHandler_ = null;
    privHandler_ = null;
    folderMember_ = null;
    aliasHandler_ = null;
    
    super.finalizeFromDeserialization();
  } // finalizeFromDeserialization()

  /**
  * @return the fastHashCode value for this object, faster than hashCode()
  *  because Resource.hashCode() is not called, the hashcodes of
  *  the children are not implied, only name_ and parent_ are used.
   * @deprecated
  */
  public int fastHashCode() {
      return 0;
  }
  
} // VirtualFolderImpl