//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseBeanInfo;

import ARoad0.gBaseInterface.*;

import ARoad0.gBase.*;

import java.beans.BeanDescriptor;
import java.beans.PropertyDescriptor;
import java.beans.IntrospectionException;
import java.lang.reflect.Method;

import ARoad0.Pattern.*;
import ARoad0.ErrW.ProcessError;

/**
 * This class is responsible for describing the ResourceImpl class.
 * It is used for displaying and editing the properties of ResourceImpl,
 * to create it and to copy it.
 * This is the reference class for BeanInfo descriptions.
 * @version 0.7.1
 */
public class ResourceImplBeanInfo extends BasicImplBeanInfo
  implements BeanInfoPattern {

  public ResourceImplBeanInfo() {}

  /**
   * Gets the description of the bean.
   * The properties to handle in the GUI are set in the descriptor.
   * They are get through BeanDescriptor.getValue().
   * @return the bean descriptor
   */
  public BeanDescriptor getBeanDescriptor() {
      // used by BaseUtilityImpl.getSuperInterfacesFromBeanInfo()
    BeanDescriptor result = new BeanDescriptor(Resource.class );
    result.setDisplayName("Resource");
      
    return getCommonBeanDescriptor(result);
  } // getBeanDescriptor()

  /**
   * Get the common description of the bean. The bean descriptor stores some values :<P>
   * - 'withTabbedPanes' is true when several tabbed panes are necessary
   * to display all the panels (property, value), <P>
   * - 'withDeleteMethod' is true to display the 'delete' button in the beamer, <P>
   * - 'commonBeanClass' is the class which has the bean info defining getCommonBeanDescriptor(),
   * and then which is the class or the superclass of the bean, <P>
   * - 'constrArgumentClasses' is the list of argument classes
   * for the class constructor, <P>
   *  - 'constrArgumentPropertyNames' is the list of property names
   * for the class constructor, <P>
   * - 'constrArgumentConcatenedNameSource' indicates, when an argument
   * is entered by the user as a relative name
   * which has then to be concatenate to another constructor argument,
   * what is this another argument which provides the first part of the name
   * if it is not null. The relative name is an ImmutableName or a String, <P>
   * - 'constrNullArgumentPossible' indicates if an argument may be null in the constructor, <P>
   * - 'constrOptionalArgument' indicates if an argument must not be used in the
   * call to the constructor when it is null, <P>
   * - 'constrRestrictedArgument' indicates if an argument is restricted
   * to some specified values, which are strings or BaseObject names. Is optional.
   * If it is true, the array of restricted values for an argument is searched
   * through BaseUtilityImpl.getRestrictedValues():
   * 1/ in 3 ACS maps which define the constraints from the ACS,
   * and then "constrArgumentPropertyNames" must contain the property descriptor name;
   * 2/ in the value "staticSetOfItems" of the property in this BeanInfo, <P>
   * - 'constrNonVisibleArgument' indicates if an argument is masqued
   * in the constructor window and sets to null in the constructor;
   * true when the ACS does not manage the property, and then,
   * this value gets the name of the ACS method which returns a boolean 'false'
   * when the argument has to be null and to be not displayed; if
   * the returned value is 'true', the argument is displayed as usual;
   * for each non null value,
   * "constrNullArgumentPossible" has to be Boolean.TRUE,
   * "constrOptionalArgument" has to be Boolean.FALSE. <P>
   * There is two property descriptors for the property 'AcsControllers'.
   * The first one is for the ACL tabbed pane, and the second one is for the
   * Privileges tabbed pane. In most of the cases, a Resource has not both AclEntries
   * and Linked Privileges, and the BeamerManager does not display the two descriptors.
   * In the rare cases where there are both managed, the Resource ACS forbids to have
   * both external AclEntries and external Linked Privileges. The property 'AcsControllers'
   * is then filled up by one of the two types of external objects. <P>
   * Used by the subclasses.
   * @param _result is the BeanDescriptor to configure
   * @return the bean descriptor
   */
  protected BeanDescriptor getCommonBeanDescriptor(BeanDescriptor _result) {

    // 'withTabbedPanes' is true since several tabbed panes are necessary
    // to display all the panels (property, value) of ResourceImpl
    _result.setValue("withTabbedPanes", Boolean.TRUE );

    // 'withDeleteMethod' is true to display the 'delete' button in the beamer
    _result.setValue("withDeleteMethod", Boolean.TRUE );

    // 'commonBeanClass' is the class which has the bean info defining getCommonBeanDescriptor(),
    // and then which is the class or the superclass of the bean.
    // caution: it is not possible in BeanDescriptor to change the value of a given key
    _result.setValue("commonBeanClass", ResourceImpl.class );
    
    // ----------------------- constructor parameters ---------------------------

    // 'constrArgumentClasses' is the list of argument classes
    // for the class constructor
    _result.setValue("constrArgumentClasses", new Class[]{
      ImmutableName.class, ACS.class, Directory.class, UserID.class, String.class} );
    
    // 'constrArgumentPropertyNames' is the list of property names
    // for the class constructor
    _result.setValue("constrArgumentPropertyNames", new String[]{
      "Name", "ACS", "Parent Node", "UserID", "Resource Type"} );

    // 'constrArgumentFromSelect' is the list of cases
    // where the argument is selected in the explorer,
    // in the order of "constrArgumentPropertyNames", or, as here,
    // in the order of the constructor method in ResourceImpl
    _result.setValue("constrArgumentFromSelect", new Boolean[]{
      Boolean.FALSE, Boolean.TRUE, Boolean.TRUE, Boolean.TRUE, Boolean.FALSE} );

    // 'constrArgumentConcatenedNameSource' indicates, when an argument
    // is entered by the user as a relative name
    // which has then to be concatenate to another constructor argument,
    // what is this another argument which provides the first part of the name
    // if it is not null. The relative name is an ImmutableName or a String.
    _result.setValue("constrArgumentConcatenedNameSource", new Integer[]{
      new Integer(2), null, null, null, null} );

    // 'constrNullArgumentPossible' indicates if an argument may be null in the constructor
    _result.setValue("constrNullArgumentPossible", new Boolean[]{
      Boolean.FALSE, Boolean.FALSE, Boolean.TRUE, Boolean.TRUE, Boolean.FALSE} );

    // 'constrOptionalArgument' indicates if an argument must not be used in the
    // call to the constructor when it is null.
    _result.setValue("constrOptionalArgument", new Boolean[]{
      Boolean.FALSE, Boolean.FALSE, Boolean.FALSE, Boolean.FALSE, Boolean.FALSE} );

    // 'constrRestrictedArgument' indicates if an argument is restricted
    // to some specified values, which are strings or BaseObject names. Is optional.
    // If it is true, the array of restricted values for an argument is searched
    // through BaseUtilityImpl.getRestrictedValues():
    // - first, in 3 ACS maps which define the constraints from the ACS,
    // and then "constrArgumentPropertyNames" must contain the property descriptor name.
    // The name words may be separated by spaces in "constrArgumentPropertyNames",
    // although there is no separation in the ACS map.
    // - second, in the value "staticSetOfItems" of the property in this BeanInfo.
    _result.setValue("constrRestrictedArgument", new Boolean[]{
      Boolean.FALSE, Boolean.FALSE, Boolean.FALSE, Boolean.FALSE, Boolean.TRUE} );
    
    // 'constrNonVisibleArgument' indicates if an argument is masqued
    // in the constructor window and sets to null in the constructor.
    // True when the ACS does not manage the property, and then, 
    // this value gets the name of the ACS method which returns a boolean 'false'
    // when the argument has to be null and to be not displayed. If
    // the returned value is 'true', the argument is displayed as usual.
    // For each non null value:
    // "constrNullArgumentPossible" has to be Boolean.TRUE,
    // "constrOptionalArgument" has to be Boolean.FALSE.
    _result.setValue("constrNonVisibleArgument", new String[]{
      null, null, "getManageResourceTrees", "getManageAccounts", null} ); 

    // ----------------------- copy parameters ---------------------------

    // 'copyArgumentClasses' is the list of argument classes
    // for the copy method, excepted for _toRegister at the end,
    //  which is another argument set to true.
    _result.setValue("copyArgumentClasses", new Class[]{
      ImmutableName.class, ACS.class, Directory.class} );
      
    // 'copyArgumentPropertyNames' is the list of property names
    // for the copy method, excepted for _toRegister which is set to true
    // following the copy() pattern.
    _result.setValue("copyArgumentPropertyNames", new String[]{
      "Name", "ACS", "Parent Node"} );

    // 'copyArgumentFromExplorer' is the list of cases
    // where the argument is selected in the explorer,
    // in the order of "copyArgumentPropertyNames"
    _result.setValue("copyArgumentFromExplorer", new Boolean[]{
      Boolean.FALSE, Boolean.TRUE, Boolean.TRUE} );

    // 'copyArgumentConcatenedNameSource' indicates, when an argument
    // is entered by the user as a relative name
    // which has then to be concatenate to another argument,
    // what is this another argument which provides the first part of the name
    // if it is not null. The relative name is an ImmutableName or a String.
    _result.setValue("copyArgumentConcatenedNameSource", new Integer[]{
      new Integer(2), null, null} );

    // 'copyNullArgumentPossible' indicates if an argument may be null
    //  in the method
    _result.setValue("copyNullArgumentPossible", new Boolean[]{
      Boolean.FALSE, Boolean.FALSE, Boolean.TRUE} );

    // 'copyOptionalArgument' indicates if an argument must not be used
    //  in the call to the method when it is null.
    _result.setValue("copyOptionalArgument", new Boolean[]{
      Boolean.FALSE, Boolean.FALSE, Boolean.FALSE} );

    // 'copyRestrictedArgument' indicates if an argument is restricted
    // to some specified values, which are strings or BaseObject names.
    //  Is optional.
    // If it is true, the array of restricted values for an argument is searched:
    // - first, in the value "staticSetOfItems" of the property in this BeanInfo,
    // - second, in the ACS.getM_staticSetOfItems() map
    // and then "constrArgumentPropertyNames" must contain the property descriptor name
    // (but the name words may be separated by spaces in "constrArgumentPropertyNames").
    _result.setValue("copyRestrictedArgument", new Boolean[]{
      Boolean.FALSE, Boolean.FALSE, Boolean.FALSE} );
      
    // ----------------------- other parameters ---------------------------

    // 'hidden' is true if the bean is hidden to the users, without
    // graphical user interface
    _result.setHidden(false);
    _result.setExpert(true);

    return _result;
  } // getCommonBeanDescriptor()

  /**
   * Each property descriptor has specific values that are get
   * with PropertyDescriptor.getValue().
   * To read a property, use the read method in the prooperty descriptor. <P>
   * To update a property, there are 3 means, in this order: <P>
   * - if it is not a single value, use the "choiceMethodForNew" value to add one value <P>
   * - or, if it is not a single value, use the "addMethod" value to add one value
   *  and "removeMethod", for BaseObjects only, <P>
   * - or, in any cases, use the write method in the property descriptor
   * @return array of property descriptors
   */
  public PropertyDescriptor[] getPropertyDescriptors() {
    PropertyDescriptor[] l_pdFromParent;
    PropertyDescriptor[] l_result;

    try {
      // ------------------- property "ParentNode" ------------------------
      PropertyDescriptor pd_three = new PropertyDescriptor(
        "ParentNode", ResourceImpl.class, "getEorParent", null);
      pd_three.setDisplayName("Parent");
      pd_three.setShortDescription("The node which is the parent in "+
        "the tree structure");
      //pd_three.setBound(false);
      //pd_three.setConstrained(false);
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_three.setValue("tabbedPaneName", "  AG Inheritance  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
        //pd_three.setValue("panelGroupName", "  Main attributes  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the ResourceImpl instance
      // does it, or when is is not applicable.
      //pd_three.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the PropertyEditor
      // has to display an 'See' button if the property is viewable
      // in the beamer.
      pd_three.setValue("withViewButton", Boolean.TRUE );

      // ------------------- property "GroupID" --------------------
      PropertyDescriptor pd_five = new PropertyDescriptor(
        "GroupID", ResourceImpl.class, "getEorGroupID", "setEorGroupID");
      pd_five.setDisplayName("main Group");
      pd_five.setShortDescription("The main group which contains this resource");
      pd_five.setBound(true);
      //pd_five.setConstrained(false);
      //pd_five.setPropertyEditorClass(ARoad0.Gui1.SimplePropertyEditor.class);
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_five.setValue("tabbedPaneName", "  AG  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_five.setValue("panelGroupName", "  Account and Group  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the ResourceImpl instance
      // does it, or when is is not applicable.
      //pd_five.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the PropertyEditor
      // has to display an 'See' button if the property is viewable
      // in the beamer.
      pd_five.setValue("withViewButton", Boolean.TRUE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_five.setValue("withSelectButton", Boolean.TRUE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select' if 'Clear' is not used,
      // when the property is an external object reference.
        //pd_five.setValue("withRemoveButton", Boolean.TRUE );
      // 'withClearButton' is the boolean which set if the PropertyEditor
      // has to display a Clear button
      pd_five.setValue("withClearButton", Boolean.TRUE );
      // 'withKnownArgumentsForNew' allows to define the constructor/add/copy arguments
      // which has to be fixed - with 'choiceMethodForNew' or not, add method or not -
      // when a new instance is created through the 'New' button.
      // First case for the fixed argument: the string 'this' is used
      // to point to the object in the beamer;
      // but this object is NOT get as argument if the 'this' index in the array
      // is after the last specified argument for the constructor or the add method.
      // Second case: search the right property descriptor in the beamer,
      // and call the read method to initialize a l_fixedValue from this object.
      // Third case for the fixed argument: the string 'thisname' is used
      // to point to the value returned from getName() by the object in the beamer.
      // Fourth case, the string 'true' or 'false' is used
      // to define a boolean value.
      // Fifth case, if none of the previous cases matches: takes directly
      // the fixed argument if it is a String.
      // Used by SimplePropertyEditor.getFixedValuesForNewElement(), which
      // may use a subset of the returned arguments if the class of the object to create
      // waits a smaller set of arguments.
      // Caution: if 'choiceMethodForNew' exists, then the known arguments
      // are applied to the constructor even if the type of the new instance
      // is not the type of the property (exemple: children of Directory).
      // Caution: the known arguments have to be Objects.
      //pd_five.setValue("withKnownArgumentsForNew", new Object[] {null, "ACS", null} );

      // ------------------- property "UserID" --------------------
      PropertyDescriptor pd_six = new PropertyDescriptor(
        "UserID", ResourceImpl.class, "getEorUserID", "setEorUserID");
      pd_six.setDisplayName("Account");
      pd_six.setShortDescription("The owner of this resource (may be proper to an user or not)");
      pd_six.setBound(true);
      //pd_six.setConstrained(false);
      //pd_six.setPropertyEditorClass(ARoad0.Gui1.SimplePropertyEditor.class);
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_six.setValue("tabbedPaneName", "  AG  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_six.setValue("panelGroupName", "  Account and Group  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the ResourceImpl instance
      // does it, or when is is not applicable.
      //pd_six.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the PropertyEditor
      // has to display an 'See' button if the property is viewable
      // in the beamer.
      pd_six.setValue("withViewButton", Boolean.TRUE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_six.setValue("withSelectButton", Boolean.TRUE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select' if 'Clear' is not used,
      // when the property is an external object reference.
        //pd_six.setValue("withRemoveButton", Boolean.TRUE );
      // 'withClearButton' is the boolean which set if the PropertyEditor
      // has to display a Clear button
      pd_six.setValue("withClearButton", Boolean.TRUE );
      // 'withKnownArgumentsForNew' allows to define the constructor/add/copy arguments
      // which has to be fixed - with 'choiceMethodForNew' or not, add method or not -
      // when a new instance is created through the 'New' button.
      // First case for the fixed argument: the string 'this' is used
      // to point to the object in the beamer;
      // but this object is NOT get as argument if the 'this' index in the array
      // is after the last specified argument for the constructor or the add method.
      // Second case: search the right property descriptor in the beamer,
      // and call the read method to initialize a l_fixedValue from this object.
      // Third case for the fixed argument: the string 'thisname' is used
      // to point to the value returned from getName() by the object in the beamer.
      // Fourth case, the string 'true' or 'false' is used
      // to define a boolean value.
      // Fifth case, if none of the previous cases matches: takes directly
      // the fixed argument if it is a String.
      // Used by SimplePropertyEditor.getFixedValuesForNewElement(), which
      // may use a subset of the returned arguments if the class of the object to create
      // waits a smaller set of arguments.
      // Caution: if 'choiceMethodForNew' exists, then the known arguments
      // are applied to the constructor even if the type of the new instance
      // is not the type of the property (exemple: children of Directory).
      // Caution: the known arguments have to be Objects.
      //pd_six.setValue("withKnownArgumentsForNew", new Object[] {null, "ACS", null} );

      // ------------------- property "AclRights" --------------------
      // note: 'AclRights' must be defined in AclEntryImplBeanInfo.getBeanDescriptor()
      // for the value 'externalFinalizeProperty', to allow the beamer to finalize an ACL
      PropertyDescriptor pd_seven = new PropertyDescriptor(
        "AclRights", ResourceImpl.class, "getL_aclEntry", null);
      // If the display name is "", the property name is not displayed.
      pd_seven.setDisplayName(""); // ACLs as target
      pd_seven.setShortDescription("AclEntries for which this object is the access target. "+
              "May be disabled by the ACL inheritance, if any.");
      pd_seven.setBound(true);
      //pd_seven.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_seven.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is a collection or a map.
      // Then, the value is the type of the values in the collection or in the map.
      // as DetailledName, is used only in a CollectionPropertyEditor
      // to display the detailled name rather than the name.
      pd_seven.setValue("valueTypeInGroup", new Object[] {AclEntry.class} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_seven.setValue("tabbedPaneName", "  ACL  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
          //pd_seven.setValue("panelGroupName", "  List  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when is is not applicable.
          //pd_seven.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_seven.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the PropertyEditor
      // has to display a New button if the property is viewable
      // in the beamer.
      pd_seven.setValue("withNewButton", Boolean.TRUE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable or if it is a String list,
      // or if the only write method is a removing method.
      // It is possible to have 'New' and 'Select'.
      pd_seven.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_seven.setValue("withDeleteButton", Boolean.TRUE );
      // 'withoutListSorting' is the boolean which forbids, if true, the sorting
      // of the displayed array by the CollectionPropertyEditor, using Object.compareTo().
      pd_seven.setValue("withoutListSorting", Boolean.FALSE );
      // 'withKnownArgumentsForNew' allows to define the constructor/add/copy arguments
      // which has to be known - with 'choiceMethodForNew' or not, add method or not -
      // when a new instance is created through the 'New' button.
      // First case for the knwon argument: the string 'this' is used
      // to point to the object in the beamer;
      // but this object is NOT get as argument if the 'this' index in the array
      // is after the last specified argument for the constructor or the add method.
      // Second case: search the right property descriptor in the beamer,
      // and call the read method to initialize a known value from this object.
      // Third case for the knwon argument: the string 'thisname' is used
      // to point to the value returned from getName() by the object in the beamer.
      // Fourth case, the string 'true' or 'false' is used
      // to define a boolean value.
      // Fifth case, if none of the previous cases matches: takes directly
      // the known argument if it is a String.
      // Used by SimplePropertyEditor.getFixedValuesForNewElement(), which
      // may use a subset of the returned arguments if the class of the object to create
      // waits a smaller set of arguments.
      // Caution: if 'choiceMethodForNew' exists, then the known arguments
      // are applied to the constructor even if the type of the new instance
      // is not the type of the property (exemple: children of Directory).
      // Caution: the known arguments have to be Objects.
      pd_seven.setValue("withKnownArgumentsForNew", new Object[] {
        null, null, null, null, null, "this"} );
      // 'addMethod' defines the 'add' method to invoke on the object for the property,
      // when there is no write method.  The property has to be a BaseObject.
      // The returned array defines:
      // - the name of the add method, at the index 0, which allows, if it is alone,
      //      to have the right BeanInfo and its constructor description from the class 
      //      of the return value of this add method,
                // - optionnaly, the name of the BeanInfo class where the add method arguments
                //      are defined, at the index 1
                // - optionnaly, the property names of the add method arguments, in the BeanInfo class,
                //      in the correct order
      // These elements allow the PropertyEditor to call BaseObjectHandler which 
      // displays a dialog to enter the arguments and to call the add method.
      pd_seven.setValue("addMethod", "addAclEntry");
      // 'removeMethod' defines the 'remove' method for an element of the property,
      // even if there is a write method. The property has to be a BaseObject.
      // finalizeForUser() in the 'valueTypeInGroup' class may be used
      // if there is "withDeleteButton", and the returned value has to be "finalizeForUser".
      // Else, the returned array defines the name of the remove method.
      // These elements allow the PropertyEditor to call the remove method
      // of the property.
      pd_seven.setValue("removeMethod", "removeAclEntry");
            //"ARoad0.gBase.AclEntryImplBeanInfo", "ACS", "EligibleParty"} );
      // Following the KeyPropertiesInName pattern for CollectionPropertyEditor,
      // 'removeArgumentsFromName' defines how to find the right arguments for
      // the 'remove' method for the property, when the name of the object
      // to remove is known, and when finalizeForUser() is not used.
      // If "removeArgumentsFromName" is null, the remove method is invoked
      // with the object to remove as single argument.
      // The returned array defines:
      // - the name of the class to call, at the index 0
      // - the name of the static method to call, at the index 1, with
      //    the name of the object to remove as argument, and with
      //    an Object[] array as returned value
      // - in the Object[] array returned as value by the static method,
      //    the sequence of index where are the n arguments to use
      //    in the remove method, in the correct order, if they do exist.
      // These elements allow the PropertyEditor to call the remove method
      // of the property.
      pd_seven.setValue("removeArgumentsFromName", new Object[] {"AclEntryImpl",
        "getKeyReferencesFromName", new Integer(0), new Integer(1), new Integer(3),
        new Integer(4)} );

      // ------------------- property "RootRights" -------------------- unused
      PropertyDescriptor pd_eight = new PropertyDescriptor(
        "RootRights", ResourceImpl.class, "getL_rootRights", null);
       // "" to do not display the name
      pd_eight.setDisplayName("");
      pd_eight.setShortDescription("The rights of the ACS super-administrateur"+
        " on this object, and they may be inferior to the maximum.");
      pd_eight.setBound(true);
      //pd_eight.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_eight.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
        //pd_eight.setValue("valueTypeInGroup", new Object[] {...} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_eight.setValue("tabbedPaneName", "  General  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_eight.setValue("panelGroupName", "  Super-Administrator Current Rights  ");
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the instance
      // does it, or when is is not applicable.
        //pd_eight.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_eight.setValue("withViewButton", Boolean.FALSE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_eight.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_eight.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_eight.setValue("withDeleteButton", Boolean.FALSE );
      // 'isRestrictedArgument' is true if it is a constrained property that is not
      // in the constructor (else, see 'constrRestrictedArgument' in the BeanDescriptor).
      // The restricted values are get from 'staticSetOfItems', for Strings only,
      // from 'choiceMethodForNew', or through a search in the ACS by the PropertyEditor
      // like here.
      pd_eight.setValue("isRestrictedArgument", Boolean.TRUE );
      
      // ------------------- property "GroupRights" --------------------
      PropertyDescriptor pd_ten = new PropertyDescriptor(
        "GroupRights", ResourceImpl.class, "getL_groupRights", "setL_groupRights");
      pd_ten.setDisplayName("main Group rights");
      pd_ten.setShortDescription("The rights on this object of"+
        " the main group. May be disabled by the AG inheritance, if any.");
      pd_ten.setBound(true);
        //pd_ten.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_ten.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
      //pd_ten.setValue("valueTypeInGroup", new Object[] {...} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_ten.setValue("tabbedPaneName", "  AG  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_ten.setValue("panelGroupName", "  Basic AG rights  ");
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when is is not applicable.
        //pd_ten.setValue("BaseBeanManager", );
      // 'withPreviousButton' is set to true to display a 'Previous' button
      // to copy all the String values of the previous property for this current property,
      // following the order of the property descriptors returned by getPropertyDescriptors().
      pd_ten.setValue("withPreviousButton", Boolean.TRUE );
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_ten.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_ten.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_ten.setValue("withSelectButton", Boolean.TRUE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_ten.setValue("withRemoveButton", Boolean.TRUE );
      // 'limitedSizeForPresentation' is the maximal size of the visible list.
      // The casual values are 5, 8, 10 (standard height), 16, 32.
      pd_ten.setValue("limitedSizeForPresentation", new Integer(8));
       // 'isRestrictedArgument' is true if it is a constrained property that is not
      // in the constructor (else, see 'constrRestrictedArgument' in the BeanDescriptor).
      // The restricted values are get from 'staticSetOfItems', for Strings only,
      // secondly from 'choiceMethodForNew', or through a search in the ACS by the PropertyEditor
      // like here.
      pd_ten.setValue("isRestrictedArgument", Boolean.TRUE );
      // 'staticSetOfItems' is the array of possible String values for the property.
      // The type of the array has to be the type of the returned elements
      // in the 'get' property method.
            //pd_ten.setValue("staticSetOfItems",  );
      // 'choiceMethodForNew' defines a specific instance/method to invoke 
      // to let the user defines the new restricted key, in a map, or the new instance, 
      // for instance to choose a sub-class
      // of the property class or to select among the restricted values.
      // With 'withKnownArgumentsForNew', the specific method is used even if 
      // there is an add method or a write method. With 'isRestrictedArgument', 
      // and only if 'staticSetOfItems' is null,
      // the specific method is used to get the restricted String values.
      // The value is an array of two Objects; the instance to invoke, or null if it is the source,
      // and the name of the method (for instance, 
      // ActionNewResource.createResource() or NameLinkTreeImpl.getVisibleSpaceConstants()).
      // The method has no argument.
      // The array is equal to 'new Object[1]' when the value 'withKnownArgumentsForNew' is null.
            //pd_ten.setValue("choiceMethodForNew", new Object[] {
                //NameLinkTreeImpl.getInstance(), "getVisibleSpaceConstants"} );
      
      // ------------------- property "OwnerRights" -------------------- 
      PropertyDescriptor pd_eleven = new PropertyDescriptor(
        "OwnerRights", ResourceImpl.class, "getL_ownerRights", "setL_ownerRights");
      pd_eleven.setDisplayName("Account rights");
      pd_eleven.setShortDescription("The rights on this object of the account which owns it. "+
              "May be disabled by the AG inheritance, if any.");
      pd_eleven.setBound(true);
        //pd_eleven.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_eleven.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
        //pd_eleven.setValue("valueTypeInGroup", new Object[] {...} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_eleven.setValue("tabbedPaneName", "  AG  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_eleven.setValue("panelGroupName", "  Basic AG rights  ");
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when is is not applicable.
        //pd_eleven.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_eleven.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_eleven.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_eleven.setValue("withSelectButton", Boolean.TRUE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_eleven.setValue("withRemoveButton", Boolean.TRUE );
      // 'limitedSizeForPresentation' is the maximal size of the visible list.
      // The casual values are 5, 8, 10 (standard height), 16, 32.
      pd_eleven.setValue("limitedSizeForPresentation", new Integer(8));
       // 'isRestrictedArgument' is true if it is a constrained property that is not
      // in the constructor (else, see 'constrRestrictedArgument' in the BeanDescriptor).
      // The restricted values are get from 'staticSetOfItems', for Strings only,
      // secondly from 'choiceMethodForNew', or through a search in the ACS by the PropertyEditor
      // like here.
      pd_eleven.setValue("isRestrictedArgument", Boolean.TRUE );
      // 'staticSetOfItems' is the array of possible String values for the property.
      // The type of the array has to be the type of the returned elements
      // in the 'get' property method.
            //pd_eleven.setValue("staticSetOfItems",  );
      // 'choiceMethodForNew' defines a specific instance/method to invoke 
      // to let the user defines the new restricted key, in a map, or the new instance, 
      // for instance to choose a sub-class
      // of the property class or to select among the restricted values.
      // With 'withKnownArgumentsForNew', the specific method is used even if 
      // there is an add method or a write method. With 'isRestrictedArgument', 
      // and only if 'staticSetOfItems' is null,
      // the specific method is used to get the restricted String values.
      // The value is an array of two Objects; the instance to invoke, or null if it is the source,
      // and the name of the method (for instance, 
      // ActionNewResource.createResource() or NameLinkTreeImpl.getVisibleSpaceConstants()).
      // The method has no argument.
      // The array is equal to 'new Object[1]' when the value 'withKnownArgumentsForNew' is null.
            //pd_eleven.setValue("choiceMethodForNew", new Object[] {
                //NameLinkTreeImpl.getInstance(), "getVisibleSpaceConstants"} );
      
      // ------------------- property "AclAcsControllers" for '  ACL  ' --------------------
      PropertyDescriptor pd_twelve = new PropertyDescriptor(
        "AclAcsControllers", ResourceImpl.class, "getL_AcsControllers", null);
      pd_twelve.setDisplayName("ACS controllers");
      pd_twelve.setShortDescription("These open ACS control this resource through "+
              "an external AclEntry.");
      pd_twelve.setBound(true);
        //pd_twelve.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_twelve.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
        //pd_twelve.setValue("valueTypeInGroup", new Object[] {...} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twelve.setValue("tabbedPaneName", "  ACL  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_twelve.setValue("panelGroupName", "  In the ACL, as target  ");
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when is is not applicable.
        //pd_twelve.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_twelve.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_twelve.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_twelve.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_twelve.setValue("withDeleteButton", Boolean.FALSE );
      
      // ------------------- second property "PriAcsControllers" for ' Privileges ' --------------------
      PropertyDescriptor pd_twentyeight = new PropertyDescriptor(
        "PriAcsControllers", ResourceImpl.class, "getL_AcsControllers", null);
      pd_twentyeight.setDisplayName("ACS controllers");
      pd_twentyeight.setShortDescription("These open ACS control this resource through "+
              "an external Linked Privilege.");
      pd_twentyeight.setBound(true);
        //pd_twelve.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_twentyeight.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
        //pd_twentyeight.setValue("valueTypeInGroup", new Object[] {...} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twentyeight.setValue("tabbedPaneName", "  Privileges  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_twentyeight.setValue("panelGroupName", "  A - Privileges as Target  ");
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when is is not applicable.
        //pd_twentyeight.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_twentyeight.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_twentyeight.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_twentyeight.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_twentyeight.setValue("withDeleteButton", Boolean.FALSE );

      // ------------------- property "EPsThroughAclEntry" --------------------
      PropertyDescriptor pd_thirteen = new PropertyDescriptor(
        "EPsThroughAclEntry", ResourceImpl.class, "getEorL_EPthroughAclEntry", null);
      pd_thirteen.setDisplayName("Right users in ACL");
      pd_thirteen.setShortDescription("The internal and external "+
        "eligible parties in its ACLs.");
      pd_thirteen.setBound(true);
        //pd_thirteen.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_thirteen.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
      pd_thirteen.setValue("valueTypeInGroup", new Object[] {ImmutableEligibleParty.class} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_thirteen.setValue("tabbedPaneName", "  ACL  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_thirteen.setValue("panelGroupName", "  In the ACL, as target  ");
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_thirteen.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_thirteen.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_thirteen.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_thirteen.setValue("withDeleteButton", Boolean.FALSE );

      // ------------------- property "ResourceType" ------------------------
      PropertyDescriptor pd_seventeen = new PropertyDescriptor(
        "ResourceType", ResourceImpl.class, "getResourceType", "setResourceType");
      pd_seventeen.setDisplayName("Access Target type");
      pd_seventeen.setShortDescription("If the type is not immutable, "+
              "the new type must be associated to the old one.");
      pd_seventeen.setValue("tabbedPaneName", "  General  " );
      pd_seventeen.setBound(true);
       // 'isRestrictedArgument' is true if it is a constrained property that is not
      // in the constructor (else, see 'constrRestrictedArgument' in the BeanDescriptor).
      // The restricted values are get from 'staticSetOfItems', for Strings only,
      // secondly from 'choiceMethodForNew', or through a search in the ACS
      // by the PropertyEditor like here.
      pd_seventeen.setValue("isRestrictedArgument", Boolean.TRUE );
      
      // ------------------ property "AccountAndGroupOfParent" ---------------
      PropertyDescriptor pd_eighteen = new PropertyDescriptor(
        "AccountAndGroupOfParent", ResourceImpl.class, "hasTheAccountAndGroupOfParent", null);
      pd_eighteen.setDisplayName("has parent A and G");
      pd_eighteen.setShortDescription("true if this target has the account and main group of its parent.");
      pd_eighteen.setBound(false);
      //pd_eighteen.setConstrained(false);
      //pd_eighteen.setPropertyEditorClass(ARoad0.Gui1.SimplePropertyEditor.class);
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_eighteen.setValue("tabbedPaneName", "  AG Inheritance  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_eighteen.setValue("panelGroupName", "  AG access through the parent  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. Is null when the ACSImpl instance
      // does it, or when is is not applicable.
      //pd_eighteen.setValue("BaseBeanManager", null);
  
      // ------------------ property "FromTheParentAccount" ---------------
      PropertyDescriptor pd_nineteen = new PropertyDescriptor(
        "FromTheParentAccount", ResourceImpl.class, "getAGOAccessFromTheParentAccount", null);
      pd_nineteen.setDisplayName("Parent A direct access");
      pd_nineteen.setShortDescription("true if this resource may be accessed "+
              "through its parent account AG rights.");
      pd_nineteen.setBound(true);
      //pd_nineteen.setConstrained(false);
      //pd_nineteen.setPropertyEditorClass(ARoad0.Gui1.SimplePropertyEditor.class);
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_nineteen.setValue("tabbedPaneName", "  AG Inheritance  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_nineteen.setValue("panelGroupName", "  AG access through the parent  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. Is null when the ACSImpl instance
      // does it, or when is is not applicable.
      //pd_nineteen.setValue("BaseBeanManager", null);
 
      // ------------------ property "FromTheParentGroup" ---------------
      PropertyDescriptor pd_twenty = new PropertyDescriptor(
        "FromTheParentGroup", ResourceImpl.class, "getAGOAccessFromTheParentGroup", null);
      pd_twenty.setDisplayName("Parent G direct access");
      pd_twenty.setShortDescription("true if this resource may be accessed "+
              "through its parent group AG rights.");
      pd_twenty.setBound(true);
      //pd_twenty.setConstrained(false);
      //pd_twenty.setPropertyEditorClass(ARoad0.Gui1.SimplePropertyEditor.class);
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twenty.setValue("tabbedPaneName", "  AG Inheritance  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_twenty.setValue("panelGroupName", "  AG access through the parent  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. Is null when the ACSImpl instance
      // does it, or when is is not applicable.
      //pd_twenty.setValue("BaseBeanManager", null);
      
      // ------------------- property "InheritedOwnerRights" --------------------  
      PropertyDescriptor pd_twentytwo = new PropertyDescriptor(
        "InheritedOwnerRights", ResourceImpl.class, "getL_inheritedOwnerRights", null);
      pd_twentytwo.setDisplayName("Rights of parent A");
      pd_twentytwo.setShortDescription("The inherited 'Account' rights from the parent "+
        "- not workable if no account matching");
      pd_twentytwo.setBound(true);
        //pd_twentytwo.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_twentytwo.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
        //pd_twentytwo.setValue("valueTypeInGroup", new Object[] {...} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twentytwo.setValue("tabbedPaneName", "  AG Inheritance  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_twentytwo.setValue("panelGroupName", "  AG from the parent A&G point of view -"+
              " Caution if 'has parent A and G' is false  ");
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when is is not applicable.
        //pd_twentytwo.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_twentytwo.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_twentytwo.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_twentytwo.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_twentytwo.setValue("withRemoveButton", Boolean.FALSE );
      // 'limitedSizeForPresentation' is the maximal size of the visible list.
      // The casual values are 5, 8, 10 (standard height), 16, 32.
      pd_twentytwo.setValue("limitedSizeForPresentation", new Integer(5));
      
      // ------------------- property "InheritedGroupRights" -------------------- 
      PropertyDescriptor pd_twentythree = new PropertyDescriptor(
        "InheritedGroupRights", ResourceImpl.class, "getL_inheritedGroupRights", null);
      pd_twentythree.setDisplayName("Rights of parent G");
      pd_twentythree.setShortDescription("The inherited 'Group' rights of the parent "+
        "- not workable if account matching");
      pd_twentythree.setBound(true);
        //pd_twentythree.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and the default value is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_twentythree.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
        //pd_twentythree.setValue("valueTypeInGroup", new Object[] {...} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twentythree.setValue("tabbedPaneName", "  AG Inheritance  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      // caution: "panelGroupName" is superior to "tabbedPaneName" in the BeamerManager analysis.
      pd_twentythree.setValue("panelGroupName", "  AG from the parent A&G point of view -"+
              " Caution if 'has parent A and G' is false  ");
      // 'withViewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'View' button if the property is viewable in the beamer.
      pd_twentythree.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_twentythree.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable, if it is a String list,
      // or if the only write method is a removing method.
      pd_twentythree.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes generally with 'Select'
      // when the property is an external object reference.
      pd_twentythree.setValue("withRemoveButton", Boolean.FALSE );
      // 'limitedSizeForPresentation' is the maximal size of the visible list.
      // The casual values are 5, 8, 10 (standard height), 16, 32.
      pd_twentythree.setValue("limitedSizeForPresentation", new Integer(5));      

      // ------------------- property "M_ConstrainedChildACS" --------------------
      PropertyDescriptor pd_twentyfive = new PropertyDescriptor(
        "M_ConstrainedChildACS", ResourceImpl.class,
        "getM_ConstrainedChildACS", null);

      pd_twentyfive.setDisplayName("Supported child ACS");
      pd_twentyfive.setShortDescription("This object supports "+
              "the execution or the integrity of these child ACSs.");
      pd_twentyfive.setBound(true);
      //pd_twentyfive.setConstrained(false);
      // set the type of property editor to use in the Gui1 package.
      pd_twentyfive.setValue("isBaseObjectMap", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle. 
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
      pd_twentyfive.setValue("valueTypeInGroup", new Object[] {ImmutableACS.class} );

      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twentyfive.setValue("tabbedPaneName", "  General  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_twentyfive.setValue("panelGroupName", "    Supports the execution "+
              "or the integrity of (opened or closed) child ACS  " );
      // 'withViewButton' is the boolean which is set if the collectionPropertyEditor
      // has to display a 'View' button if the property is viewable
      // in the beamer. 
      pd_twentyfive.setValue("withViewButton", Boolean.TRUE );
      // 'KeysUpdateInMap' is true if the key list may be updated. If it is false,
      // and if there is some updating methods, only the value list in the map
      // may be updated for any key.
        //pd_twentyfive.setValue("KeysUpdateInMap", Boolean.TRUE);
      // 'isRestrictedArgument' is true if it is a constrained property that is not
      // in the constructor (else, see 'constrRestrictedArgument' in the BeanDescriptor).
      // The restricted values are get from 'staticSetOfItems', for Strings only,
      // secondly from 'choiceMethodForNew' like here,
      // or through a search in the ACS by the PropertyEditor.
        //pd_twentyfive.setValue("isRestrictedArgument", Boolean.TRUE );
      // 'choiceMethodForNew' defines a specific instance/method to invoke 
      // to let the user defines the new restricted key, in a map, or the new instance, 
      // for instance to choose a sub-class
      // of the property class or to select among the restricted values.
      // With 'withKnownArgumentsForNew', the specific method is used even if 
      // there is an add method or a write method. With 'isRestrictedArgument', 
      // and only if 'staticSetOfItems' is null,
      // the specific method is used to get the restricted String values.
      // The value is an array of two Objects; the instance to invoke, or null if it is the source,
      // and the name of the method (for instance, 
      // ActionNewResource.createResource() or NameLinkTreeImpl.getVisibleSpaceConstants()).
      // The method has no argument.
        //pd_twentyfive.setValue("choiceMethodForNew", new Object[] {
            //ACSFactoryUtilityImpl.getInstance(), "getKeysForConstrainedChildACSMap"} );
      // 'addMethod' defines the 'add' method to invoke on the object for the property,
      // when there is no write method.  The property has to be a BaseObject.
      // The returned array defines:
      // - the name of the add method, at the index 0, which allows, if it is alone,
      //      to have the right BeanInfo and its constructor description from the class 
      //      of the return value of this add method,
                // - option 1: the name of the BeanInfo class where the add method arguments
                //      are defined, at the index 1
                // - option 2: the property names of the add method arguments, 
                //      in the BeanInfo class, in the correct order
      // These elements allow the PropertyEditor to call BaseObjectHandler which 
      // displays a dialog to enter the arguments and call the add method.
        //pd_twentyfive.setValue("addMethod", "addAclEntry");
      // 'removeMethod' defines the 'remove' method for an element of the property,
      // even if there is a write method. The property has to be a BaseObject.
      // If finalizeForUser() exists, is public, and if there is "withDeleteButton",
      // the returned value is "finalizeForUser".
      // Else, the returned array defines the name of the remove method.
      // These elements allow the PropertyEditor to call the remove method
      // of the property. To remove a key, use this method with a null argument
      // for the value.
        //pd_twentyfive.setValue("removeMethod", "removeAclEntry");

      // ------------------- property "OneTargetLinkedPrivilege" --------------------
      PropertyDescriptor pd_twentysix = new PropertyDescriptor(
        "OneTargetLinkedPrivilege", ResourceImpl.class,
        "getEorL_TargetPrivilegeForLinks", null);
      pd_twentysix.setDisplayName("Linked Privileges");
      pd_twentysix.setShortDescription("The linked privileges of this access target. "+
              "May be disabled by the PRI inheritance, if any.");
      pd_twentysix.setBound(true);
      //pd_twentysix.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and is FALSE when it is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_twentysix.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, the value is the type of the values in the collection or in the map.
      // as DetailledName, is used only in a CollectionPropertyEditor
      // to display the detailled name rather than the name.
      pd_twentysix.setValue("valueTypeInGroup", new Object[] {PrivilegeForLinks.class} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twentysix.setValue("tabbedPaneName", "  Privileges  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_twentysix.setValue("panelGroupName", "  A - Privileges as Target  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when it is not applicable.
          //pd_twentysix.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which is set if the collectionPropertyEditor
      // has to display a 'View' button if the property is viewable in the beamer.
      pd_twentysix.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_twentysix.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable or if it is a String list,
      // or if the only write method is a removing method.
      // It is possible to have 'New' and 'Select'. 'Select' implies 'Remove'.
        //pd_twentysix.setValue("withSelectButton", Boolean.TRUE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes automatically with 'Select'
      // when the property is an external object reference.
        //pd_twentysix.setValue("withRemoveButton", Boolean.TRUE );
      // 'withKnownArgumentsForNew' allows to define the constructor arguments
      // which has to be preset - with 'choiceMethodForNew' or not, add method or not -
      // when a new instance is created through the 'New' button.
      // First case for the known argument: the string 'this' is used
      // to point to the object in the beamer;
      // but this object is NOT get as argument if the 'this' index in the array
      // is after the last specified argument for the constructor or the add method.
      // Second case: search the right property descriptor of the object
      // in the beamer, and call the read method
      // to initialize a l_fixedValue from this object.
      // Caution: if 'choiceMethodForNew' exists, then the known arguments
      // are applied to the constructor even if the type of the new instance
      // is not the type of the property (exemple: children of Directory).
      // Caution: the known arguments have to be Objects.
        //pd_twentysix.setValue("withKnownArgumentsForNew", new Object[] {
          //null, null, null, null, "this"} );
      // 'addMethod' defines the 'add' method to invoke on the object for the property,
      // when there is no write method.  The property has to be a BaseObject.
      // The returned array defines:
      // - the name of the add method, at the index 0
                // - the name of the BeanInfo class where the add method arguments
                //      are defined, at the index 1
                // - the property names of the add method arguments, in the BeanInfo class,
                //      in the correct order
      // These elements allow the PropertyEditor to display a window to create
      // all the required arguments and to call the add method of the property.
        //pd_twentysix.setValue("addMethod", "addEorPrivilegeForLinksToTarget");
          //"ARoad0.gBase.AclEntryImplBeanInfo", "ACS", "EligibleParty", "Rights"} );
      // 'removeMethod' defines the 'remove' method for an element of the property,
      // even if there is a write method. The property has to be a BaseObject.
      // If finalizeForUser() exists, is public, and if there is "withDeleteButton",
      // the returned value is "finalizeForUser".
      // Else, the returned array defines the name of the remove method.
      // These elements allow the PropertyEditor to call the remove method
      // of the property.
        //pd_twentysix.setValue("removeMethod", "removeEorPrivilegeToTarget");
            //"ARoad0.gBase.AclEntryImplBeanInfo", "ACS", "EligibleParty"} );
      // Following the KeyPropertiesInName pattern,
      // 'removeArgumentsFromName' defines how to find the right arguments for
      // the 'remove' method for the property, when the name of the object
      // to remove is known, and when finalizeForUser() is not used.
      // If "removeArgumentsFromName" is null, the remove method is invoked
      // with the object to remove as single argument.
      // The returned array defines:
      // - the name of the class to call, at the index 0
      // - the name of the static method to call, at the index 1, with
      //    the name of the object to remove as argument, and with
      //    an Object[] array as returned value
      // - in the Object[] array returned as value by the static method,
      //    the sequence of index where are the n arguments to use
      //    in the remove method, in the correct order
      // These elements allow the PropertyEditor to call the remove method
      // of the property.
        // pd_twentysix.setValue("removeArgumentsFromName", ... );
      // 'limitedSizeForPresentation' is the maximal size of the visible map and value lists.
      // Not ridden if less than 4.
      // The casual values are 5, 8, 10 (standard height), 16, 32.
      pd_twentysix.setValue("limitedSizeForPresentation", new Integer(5));

      // ------------------- property "OneTargetTypedPrivilege" --------------------
      PropertyDescriptor pd_twentyseven = new PropertyDescriptor(
        "OneTargetTypedPrivilege", ResourceImpl.class, "getEorL_TargetPrivilegeForTypes", null);
      pd_twentyseven.setDisplayName("Typed Privileges");
      pd_twentyseven.setShortDescription("The typed privileges of this access target.");
      pd_twentyseven.setBound(true);
      //pd_twentyseven.setConstrained(false);
      // 'isCollection' is TRUE when the right property editor is CollectionPropertyEditor,
      // and is FALSE when it is SimplePropertyEditor.
      // If the display name is "", it is not displayed.
      pd_twentyseven.setValue("isCollection", Boolean.TRUE );
      // 'valueTypeInGroup' is not null only when the property is an array, a collection or a map.
      // Then, it defines the type of the elementary object to handle.
      // If the map value is a collection or an array, the given type
      // is the type of the element in this collection or array.
      // as DetailledName is used only in a CollectionPropertyEditor
      // to display the detailled name rather than the name.
      pd_twentyseven.setValue("valueTypeInGroup", new Object[] {PrivilegeForType.class} );
      // 'tabbedPaneName' is the name (String value) of the pane where this property
      // has to be displayed in the beamer.
      pd_twentyseven.setValue("tabbedPaneName", "  Privileges  " );
      // 'panelGroupName' is the name (String value) of the group where this property
      // has to be displayed with other related properties. May be empty.
      pd_twentyseven.setValue("panelGroupName", "  A - Privileges as Target  " );
      // 'BaseBeanManager' is the manager (Object value) which fires an event
      // when the property is closed or deleted. No value when the AclEntryImpl instance
      // does it, or when it is not applicable.
          //pd_twentyseven.setValue("BaseBeanManager", );
      // 'withViewButton' is the boolean which is set if the collectionPropertyEditor
      // has to display a 'View' button if the property is viewable in the beamer.
      pd_twentyseven.setValue("withViewButton", Boolean.TRUE );
      // 'withNewButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'New Value(s)' button if the property is viewable
      // in the beamer. False if it is not editable trough a constructor.
      pd_twentyseven.setValue("withNewButton", Boolean.FALSE );
      // 'withSelectButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Add Selection(s)' button if the property is viewable
      // in the beamer. False if it is not editable or if it is a String list,
      // or if the only write method is a removing method.
      // It is possible to have 'New' and 'Select'. 'Select' implies 'Remove'.
      pd_twentyseven.setValue("withSelectButton", Boolean.FALSE );
      // 'withDeleteButton' is the boolean which is set if the collectionPropertyEditor
      // has to display an 'Delete Selection(s)' button if the property is viewable
      // in the beamer. If false, there is a 'Remove Selection(s)' button.
      // 'Delete' goes generally with 'New'.
      // On the contrary, 'Remove' goes automatically with 'Select'
      // when the property is an external object reference.
      pd_twentyseven.setValue("withRemoveButton", Boolean.FALSE );
      // 'withKnownArgumentsForNew' allows to define the constructor arguments
      // which has to be preset - with 'choiceMethodForNew' or not, add method or not -
      // when a new instance is created through the 'New' button.
      // First case for the known argument: the string 'this' is used
      // to point to the object in the beamer;
      // but this object is NOT get as argument if the 'this' index in the array
      // is after the last specified argument for the constructor or the add method.
      // Second case: search the right property descriptor of the object
      // in the beamer, and call the read method
      // to initialize a l_fixedValue from this object.
      // Caution: if 'choiceMethodForNew' exists, then the known arguments
      // are applied to the constructor even if the type of the new instance
      // is not the type of the property (exemple: children of Directory).
      // Caution: the known arguments have to be Objects.
        //pd_twentyseven.setValue("withKnownArgumentsForNew", new Object[] {
          //null, null, null, null, "this"} );
      // 'addMethod' defines the 'add' method to invoke on the object for the property,
      // when there is no write method.  The property has to be a BaseObject.
      // The returned array defines:
      // - the name of the add method, at the index 0
                // - the name of the BeanInfo class where the add method arguments
                //      are defined, at the index 1
                // - the property names of the add method arguments, in the BeanInfo class,
                //      in the correct order
      // These elements allow the PropertyEditor to display a window to create
      // all the required arguments and to call the add method of the property.
        //pd_twentyseven.setValue("addMethod", "addEorPrivilegeForLinksToTarget");
          //"ARoad0.gBase.AclEntryImplBeanInfo", "ACS", "EligibleParty", "Rights"} );
      // 'removeMethod' defines the 'remove' method for an element of the property,
      // even if there is a write method. The property has to be a BaseObject.
      // If finalizeForUser() exists, is public, and if there is "withDeleteButton",
      // the returned value is "finalizeForUser".
      // Else, the returned array defines the name of the remove method.
      // These elements allow the PropertyEditor to call the remove method
      // of the property.
        //pd_twentyseven.setValue("removeMethod", "removeEorPrivilegeToTarget");
            //"ARoad0.gBase.AclEntryImplBeanInfo", "ACS", "EligibleParty"} );
      // Following the KeyPropertiesInName pattern,
      // 'removeArgumentsFromName' defines how to find the right arguments for
      // the 'remove' method for the property, when the name of the object
      // to remove is known, and when finalizeForUser() is not used.
      // If "removeArgumentsFromName" is null, the remove method is invoked
      // with the object to remove as single argument.
      // The returned array defines:
      // - the name of the class to call, at the index 0
      // - the name of the static method to call, at the index 1, with
      //    the name of the object to remove as argument, and with
      //    an Object[] array as returned value
      // - in the Object[] array returned as value by the static method,
      //    the sequence of index where are the n arguments to use
      //    in the remove method, in the correct order
      // These elements allow the PropertyEditor to call the remove method
      // of the property.
        // pd_twentyseven.setValue("removeArgumentsFromName", ... );
      // 'limitedSizeForPresentation' is the maximal size of the visible map and value lists.
      // Not ridden if less than 4.
      // The casual values are 5, 8, 10 (standard height), 16, 32.
      pd_twentyseven.setValue("limitedSizeForPresentation", new Integer(5));
      
      // ------- End of properties description ---------------------------------

      // start from the parent class bean info
      l_pdFromParent = super.getPropertyDescriptors(); // 9 items
      l_result = new PropertyDescriptor[l_pdFromParent.length + 18];
      System.arraycopy(l_pdFromParent, 0, l_result, 0, 4);
      l_result[4] = pd_seventeen; // Resource Type
      l_result[5] = l_pdFromParent[4]; // Virtual Folders
      l_result[6] = l_pdFromParent[5]; // Open Views
      l_result[7] = l_pdFromParent[6]; //
      l_result[8] = l_pdFromParent[7]; //
      l_result[9] = l_pdFromParent[8]; //

      // The order in l_pd defines the display order, plus the values in
      // "tabbedPaneName" and "panelGroupName" for each property, and also,
      // it is used in the sequence (pd_eleven, pd_ten) for the Previous button

      // pd_eight on 'RootRights' is unused, so they are 19 properties, and only 18 are returned
      PropertyDescriptor[] l_pd = new PropertyDescriptor[]
        {pd_six, pd_five,  
          pd_eleven, pd_ten, // rights
          pd_twentyfive,
          // Inheritance
          pd_eighteen, pd_nineteen, pd_twenty,
          pd_three,
          pd_twentytwo, pd_twentythree,
          // ACL
          pd_seven, pd_twelve, pd_thirteen,
          // Privileges
          pd_twentyseven, pd_twentysix, pd_twentyeight // almost copy of pd_twelve
        };
      System.arraycopy(l_pd, 0, l_result, 10, l_pd.length);

      
                    //System.out.println("ResourceImplBeanInfo.getPropertyDescriptors() "+
                      //      "with 'VirtualFolders' at 7:");
                    //for (int i = 0; i < l_result.length; i++) {
                      //  System.out.println("------ "+l_result[i].getName());
                    //} // for i


      l_pdFromParent = null;
      l_pd = null;
      return l_result;
    } catch (IntrospectionException e) {
      new ProcessError("In ResourceImplBeanInfo.getPropertyDescriptors(), "+
        "IntrospectionException: "+e);
      throw new InternalError("In ResourceImplBeanInfo.getPropertyDescriptors(), "+
        "IntrospectionException: "+e);
    } // catch

  } // getPropertyDescriptors()

    /**
     * Get the instance, if it exists, which fires an event when
     * the bean is closed or deleted.
     * @return null.
     */
    public BaseBeanManager getBaseBeanManagerInstance() {
      return null;
    }

    /**
     * Get the bean method, without parameters, that allows to get
     * the BaseBeanManager instance, if it exists,
     * which fires an event when the bean instance is closed or deleted.
     * @return the get method. Is null when there is
     *  an active BeanInfoPattern.getBaseBeanManagerInstance(),
     *  or when it is not applicable.
     */
    public Method getBaseBeanManagerGetMethod() {
      Object result = ResourceImpl.EMPTY_INSTANCE; // non-sens init
      try {
        result = ResourceImpl.class.getMethod("getEorACS", (Class<?>[])null);
      } catch (NoSuchMethodException n) {
        new ProcessError("in ResourceImplBeanInfo."+
          "getBaseBeanManagerGetMethod(): "+n.toString());
      } catch (SecurityException s) {
      }
      return (Method) result;
    }

    /**
     * Get the property name for which a change fires an PropertyChangeEvent
     * in the BaseBeanManager, when the bean instance is closed or deleted.
     * This is used to register the BeamerManager as a PropertyChangeListener.
     * @return the property name.
     */
    public String getBaseBeanManagerBoundProperty() {
      return "m_OwnResources";
    }

} // ResourceImplBeanInfo