//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.SortedMap;
import java.util.Map;
import java.util.List;

/**
* This interface is responsible for delivering a catalog
* of the access control system (ACS) structure and its ACS objects.
* @version 0.7.1
*/
public interface ACSCatalog {  
    
  /**
  * @return unmodifiable SortedMap of recorded open resources in the ACS base.
  * The resources are external object references.
  */
  public SortedMap<ImmutableName, ImmutableResource> getEorM_Resources();
  
  /**
  * @return unmodifiable SortedMap of recorded own resources copies,
  *  for which it is the main ACS. The resources are external object references.
  */
  public SortedMap<ImmutableName, ImmutableResource> getEorM_OwnResources();
  
  /**
   * @return unmodifiable SortedMap of registred own resources copies,
   *  for which there is no parent. Return a subset of the getEorM_Resources
   *  method, then is much slower.
   */
  public SortedMap<ImmutableName, ImmutableResource> getEorM_OwnNoParentResources();
  
  /**
   * An external resource is owned by another ACS.
   * @return unmodifiable SortedMap of registred external resources copies,
   * for which it is NOT the main ACS. 
   */
  public SortedMap<ImmutableName, ImmutableResource> getEorM_ExternalResources();
  
  /**
   * @return unmodifiable sorted map of registered open eligible parties
   */
  public SortedMap<ImmutableName, ImmutableEligibleParty> getEorM_EligibleParties();
  
  /**
    * @return unmodifiable SortedMap of recorded eligible parties
    * in the ACS base. The eligible parties are external object references.
    */
  public SortedMap<ImmutableName, ImmutableEligibleParty> getEorM_OwnEligibleParties();
  
  /**
   * Get the external eligible parties, which are defined as owned by another ACS.
   * @return unmodifiable SortedMap of registred external eligible party copies,
   *  for which it is NOT the main ACS.
   */
  public SortedMap<ImmutableName, ImmutableEligibleParty> getEorM_ExternalEligibleParties();
  
  /**
   * Get the aclEntries managed by this ACS. They may connect a resource
   * and an eligible party from this ACS or from others ones.
   * Caution: They can't be updated directly. Use their resources for that.
   * @return ImmutableAclEntry[]: copy of the aclEntries of this ACS. May be
   * with empty aclEntries, or equal to 'new AclEntryImpl[] {new AclEntryImpl()}'
   */
  public ImmutableAclEntry[] getEorL_AclEntries();
  
  /**
  * Get an AclEntry managed by this ACS from its detailled name.
  * @param _detailledName detailled name
  * @return AclEntry - May be null.
  */
  public AclEntry getOneAclEntryFromDetailledName(String _detailledName);
  
  /**
  * Get an AclEntry managed by this ACS from its name.
  * @param _name name
  * @return AclEntry - May be null.
  */
  //public AclEntry getOneAclEntry(String _name);
  
/**
 * Get the number of own resources in the ACS,
 * excluding the external objects.
 * @return number of resources managed by the ACS
 */
public int getNumberOfResources();
  
/**
 * Get the number of own eligible parties in the ACS,
 * excluding the external objects.
 * @return number of eligible parties managed by the ACS
 */
public int getNumberOfEPs();

/**
 * Test if the resource is opened, and owned or controlled by the ACS.
 * @param _name ImmutableName of the resource to test in the ACS
 * @return boolean  true if the ACS knows this resource as opened
 */
public boolean containsResourceName(ImmutableName _name);

/**
 * @param _name ImmutableName of the eligible party to test
 * @return boolean true if the ACS knows this EP as its own or as an external one
 */
public boolean containsEligiblePartyName(ImmutableName _name);
        
/**
* @param _name ImmutableName of the own resource to test in the ACS
* @return boolean true if the ACS owns this resource
*/
public boolean containsOwnResourceName(ImmutableName _name);
        
/**
 * @param _name ImmutableName of the own eligible party to test in the ACS
 * @return boolean true if the ACS owns this EP or this actor
 */
public boolean containsOwnEligiblePartyName(ImmutableName _name);
        
/**
 * Test if an AclEntry is known. Use '==' and not the equals() method.
 * @param _acl AclEntry to test
 * @return boolean true if the ACS has recorded this aclEntry instance
 */
public boolean containsAclEntry(ImmutableAclEntry _acl);

/**
* The actors here are the controlled ones, they are internal or external.
* They can't be external eligible actors, referenced in some aclEntries.
* @return unmodifiable SortedMap of registred actors.
*  Return a subset of getM_Resources(), is slower than getM_Resources().
*  The actors are external object references.
*/
public SortedMap<ImmutableName, ImmutableActor> getEorM_Actors();
    
/**
 * @return unmodifiable SortedMap of registred own actors,
 *  for which it is the main ACS. Return a subset of getM_OwnResources(),
 *  is slower than getM_OwnResources() and getEorM_Actors().
 *  The actors are external object references.
 */
public SortedMap<ImmutableName, ImmutableActor> getEorM_OwnActors();

/**
 * Gets, in Linux/Unix ACS only, the actors with an 'execute' right in 'other' rights,
 * and having a non-null current account/group. If Executable, this current value
 * may be set by an enforced account/group.
 * These Actors are not accessible if their direct and indirect directories 
 * have not the right 'r_nxdir', but this is not checked during the filling of this list.
 * @return unmodifiable List of actors
 */
//public List getEorL_FromOtherInCurrentContextActors();

  /**
   * Gets all the own actors which are bridge sources for external actors.
   * @return set of ImmutableActors. May be null.
   */
  public List<ImmutableActor> getEorL_OwnActorBridgeSources();

  /**
   * Gets all the own actors which are bridge targets for external actors.
   * @return set of ImmutableActors. May be null.
   */
  public List<ImmutableActor> getEorL_OwnActorBridgeTargets();
  
  /**
  * Get the ACS Grasp, in the search of it in the Grasp node.
  * @param _name the Grasp name
  * @return the Grasp; null if it does not exist.
  */
  //public Grasp getOneGrasp(Name _name); 

 /**
  * Get all the open ACS that have an external AclEntry or an external Linked Privilege
  * for a resource or a virtual folder from this ACS.
  * @return the open ACS controller names. May be empty but not null.
  * No null value in the array.
 */
public ImmutableName[] getL_AcsControllers();

/**
 * @param
 * @exception
 * @return
 */
//	public SortedMap getEorM_Accesses();


  /**
   * The modeling of the structure is incomplete when some ACSObjects miss
   * while they participate to the decision for some important access controls,
   * or simply when these missed ACSObjects are important for the overall ACS security,
   * or when their important relations with some other ACSObjects are not set.
   * On the other hand, the modeling remains complete if some secondary ACSObjects
   * are not modeled, like for instance image files in an operating system file system.
   * When the modeling of an ACS structure is incomplete, all the views which
   * use this ACS are set incomplete in their results.
   * @return true if the modeling of the ACS structure is incomplete.
   */
  public boolean getIncompleteStructureModeling();

  /**
   * The modeling of the behavior is incomplete when important constraints on
   * ACSObjects are not modeled, for instance to set their allowed rights, or
   * their effective default rights.
   * On the other hand, the modeling remains complete if some secondary relations
   * or constraints are not modelled by the program, like for instance the presence
   * of image files in an operating system file system for a given application software.
   * The user may complements manually the properties of an ACSObject for which
   * the behavior has not been totally defined in the ACS model.
   * When the modeling of an ACS behavior is incomplete, the views which
   * use this ACS remain complete in their results, but the creation or deletion
   * of a new ACSObject should be reviewed by the user to ensure that all the properties
   * are still well defined in the ACS.
   * @return true if the modeling of the ACS behavior is incomplete.
   */
  public boolean getIncompleteBehaviorModeling();
  /**
   * @return true if the resources are managed.
   */
  public boolean getManageResources();

  /**
   * @return true if the resource trees are managed.
   */
  public boolean getManageResourceTrees();

  /**
   * @return true if the accounts (or users) are managed.
   */
  public boolean getManageAccounts();
  
  /**
   * @return true if the GroupIDs are managed.
   */
  public boolean getManageGroups();

  /**
   * Caution: the name of this method is confusing, because a collection of
   * nested groups may be more than one single group tree.
   * @return true if the nested groups are managed.
   */
  public boolean getManageGroupTrees();

  /**
   * @return true if the internal acl entries are managed.
   */
  public boolean getManageInternalAcls();

  /**
   * @return true if the external acl entries are managed, as
   *   controler of other-acs objects, and/or as acs which is controled by other acs.
   */
  public boolean getManageExternalAcls();
  
  /** 
   * @return true if the flexible conditional AclEntries are managed
   */
  public boolean getManageFlexibleConditionalAcls();
  
  /** 
   * @return true if the right-defined conditional AclEntries are managed.
   */
  public boolean getManageRightDefinedConditionalAcls();
  
  /**
   * @return true if the aclentry rights inheritance from the parent are managed
   * (not necessary all these rights).
   */
  public boolean getManageACLRightsInheritance();

  /**
   * @return true if the actors are managed.
   */
  public boolean getManageActors();

  /**
   * @return true if the ACS bridges are managed.
   */
  public boolean getManageBridges();

  /**
   * @return true if the secondary groups are managed for actors (no
   *   answer on directoryEPs).
   */
  public boolean getManageSecGroupsForActor();
  
  /**
   * @return true if the directoryEP are managed.
   */
  public boolean getManageDirectoryEPs();

  /**
   * @return true if some of the Account-Group-Other rights are managed (not
   * necessary all these rights).
   */
  public boolean getManageAGORights();
  
  /**
   * @return true if Account-Group rights inheritance from the parent are managed (not 
   * necessary all these rights).
   */
  public boolean getManageAGORightsInheritance();

  /**
   * @return true if the privileges are managed. 
   */
  public boolean getManagePrivilegeRights();

  /**
   * @return true if the virtual folders 
   *   are managed for the ACS. 
   */
  public boolean getManageVirtualFolders();
  
  /**
   * @return true if the Grant rights are managed. 
   */
  public boolean getManageGrantRights();
  
  /**
   * @return true if the Deny rights are managed. 
   */
  public boolean getManageDenyRights();      
   
  /**
   * @deprecated
   * @return true if the acs is running on the parent acs.
   */
  public boolean getRunningOnParentACS();
  
  /**
   * @deprecated
   * @return true if the acs is embedded in the parent acs.
   */
  public boolean getEmbeddedInParentACS();

  /**
   * Caution: if getManageExternalForOneToOneLinkedPrivileges() returns true,
   * the privilege of another ACS may always have as target a Resource of this ACS.
   * @return true if the privilege rights in Resources are managed.
   */
  public boolean getManagePrivilegesInResources();

  /**
   * Caution: if getManageExternalForOneToOneLinkedPrivileges() returns true,
   * the privilege of another ACS may always have as target a VirtualFolder of this ACS.
   * @return true if the privilege rights in VirtualFolders are managed.
   */
  public boolean getManagePrivilegesInVirtualFolders();
  
  /**
   * @return true if the simple linked privilege rights are managed.
   */
  public boolean getManageSimpleLinkedPrivileges();
  
  /**
   * @return true if the simple typed privilege rights are managed. 
   */
  public boolean getManageSimpleTypedPrivileges();
  
  /**
   * @return true if the if-all-sources linked privilege rights are managed.
   */
  public boolean getManageIfAllSourceLinkedPrivileges();
  
  /**
   * @return true if the internal for-one-to-one linked privilege rights are managed. 
   */
  public boolean getManageInternalForOneToOneLinkedPrivileges();

  /**
   * May be true if the privileges are managed but all the other privilege features
   * are not managed, to allow an own component in the privilege of another ACS.
   * Managing external privileges always implies that all the own sources of this ACS
   * may have an inner external privilege. To allow an own resource or an own virtual folder
   * of this ACS to get an inner external privilege, the privileges in resources or
   * in virtual folders have to be managed. This allows to limit the
   * inner external privileges to the own sources of the ACS, excluding the own targets,
   * if there are no internal privileges to manage. On the other hand, having
   * internal privileges means that the inner external privileges always accept as target
   * an own resource or virtual folder.
   * @return true if the external for-one-to-one linked privilege rights are managed
   * as inner privileges or as privileges of another ACS.
   */
  public boolean getManageExternalForOneToOneLinkedPrivileges();

  /**
   * @return true if the classed typed privilege rights are managed.
   */
  public boolean getManageClassedTypedPrivileges();

  /**
   * @return true if the classed specific typed privilege rights are managed,
   * for designated source and target.
   */
  public boolean getManageClassedSpecificTypedPrivileges();

  /**
   * @return true if the typed privilege rights for authorization servers are managed.
   */
  public boolean getManageAuthorizationServerTypedPrivileges();

  /**
   * @return true if the granting rights in privileges are managed.
   */
  public boolean getManageGrantingRightPrivileges();

  /**
   * @return true if the denying rights in privileges are managed.
   */
  public boolean getManageDenyingRightPrivileges();
  
  /**
   * @return true if the linked privilege rights inheritance from the parent are managed
   * (not necessary all these rights).
   */
  public boolean getManageLPRIRightsInheritance();
  
  /**
   * @return true if the virtual folders of resources are managed for the ACS,
   * and where actors are forbidden.
   */
  public boolean getManageResourceVirtualFolders();

  /**
   * @return true if the virtual folders of actors are managed for the ACS,
   * and where other resources are forbidden.
   */
  public boolean getManageEPVirtualFolders();

  /**
   * @return true if the resource virtual folders manage simple resources for the ACS.
   */
  public boolean getManageSimpleResourceVirtualFolders();

  /**
   * @return true if the resource virtual folders manage the directories for the ACS.
   */
  public boolean getManageDirectoryInResourceVirtualFolders();

  /**
   * @return true if the virtual folders of eligible parties manage actors for the ACS.
   */
  public boolean getManageActorInEPVirtualFolders();

  /**
   * @return true if the EP virtual folders manage the directoryEPs for the ACS.
   */
  public boolean getManageDirectoryEPinEPVirtualFolders();

  /**
   * @return true if the virtual folder trees are managed for the ACS.
   */
  public boolean getManageVirtualFolderTrees();

  /**
   * @return true if a member is no more than one time in each virtual folder tree.
   */
  public boolean getManageOneTimeInVirtualFolderTreeForMember();

  /**
   * @return true if a virtual folder propagates its target rights to all its members.
   */
  public boolean getManageVirtualFolderRightsPropagationToMembers();
  
  /**
   * @return true if the acs is an authorization server.
   */
  public boolean getIsAuthorizationServer();
  
  /**
   * @return true if the acs manages ACS rights (it main manages
   *    metarights too).
   */
  public boolean getManageAcsRights();
  
  /**
   * @return true if the acs manages metarights (it main manages
   *    ACS rights too).
   */
  public boolean getManageMetaRights();

  /**
   * @return true if the acs manages the internal or external soft aliases
   */
  public boolean getManageSoftAlias();

  /**
   * @return true if the acs manages the internal hard aliases
   */
  public boolean getManageInternalHardAlias();
   
  /**
   * Get the String values which define the allowed ACSObject types.
   * The default value is always all the possible values, when there is no key
   * to match. The types are used to manage the display of nodes in the explorer,
   * and to forbid the creation of instances having some specified types for a given class.
   * The 10 possible keys and values are: <P>
   *     - 'Resource.ResourceType' defines the allowed types for the Resources <P>
   *     - 'Actor.ResourceType' defines the allowed types for the Actors <P>
   *     - 'EligibleParty.EPType' defines the allowed types for the Eligible Parties <P>
   *     - 'VirtualFolder.VirtualFolderType' defines the allowed String types for
   *        the VirtualFolders of the ACS <P>
   *     - 'PrivilegeForType.SecondType' defines the allowed String types for
   *        the PrivilegeForTypes of the ACS <P>
   *     - 'PrivilegeForLinks.SecondType' defines the allowed String types for
   *        the PrivilegeForLinks of the ACS <P>
   *     - 'Explorer.NoAcsNode', where the values have the format 'ACSTree.XXX',
   *            to forbid the display in the explorer, of some standard nodes like
   *            'AclEntries...' or 'GroupIDs', <P>
   *     - 'Explorer.SubAcsNode' sets some subacs nodes in the explorer,
   *            where the values have the format 'ResourceType.<byAcsCreator> XXX'
   *            or 'GroupType.<byAcsCreator> XXX', to specify the display
   *            under a dedicated node '<XXX_subacs>', of all the resources or groups
   *            having a type starting with  '<byAcsCreator> XXX'. This type starting
   *            is reserved to these children.  <P>
   *     - 'Explorer.AcsGroupTree' sets some group trees in the explorer,
   *            where the values have the format 'GroupType.<byAcsCreator> XXX',
   *            to specify the display under a dedicated node '<XXX_tree>'
   *            as its direct children, all the groupIDs having a type which starts
   *            with '<byAcsCreator> YYY'. This type starting
   *            is reserved to these children.  <P>
   *     - 'CreationByBeamer.NoType' where the values are some forbidden types,
   *            to forbid the creation by the user and through the GUI, 
   *            of ACSObjects having these types for a given class. <P>
   * With the key 'CreationByBeamer.NoType', the user can create some ACSObjects
   * only with the types the ACS authorizes. Typically, the types
   * 'ResourceType.<byAcsCreator>' and 'EPType.<byAcsCreator>' are not allowed.
   * After the first selection of allowed types, it may have a removing of some values
   * from the getM_TypesAndGUIPolicy() key 'CreationByBeamer.NoType'.
   * The forbidden types have one of the following forms :
   * 'ResourceType.forbidden type', 'EPType.forbidden type', 'DirectoryType.forbidden type',
   * 'ActorType.forbidden type', 'UserIDType.forbidden type', 'GroupIDType.forbidden type',
   * 'VirtualFolderType.forbidden type', 'PrivilegeSecondType.forbidden type'.
   * A forbidden type for EligibleParty is not forbidden to all the EligibleParty subclasses.
   * This is also true for a forbidden type for Resource. Like the order of interfaces,
   * the 'ActorType.forbidden type' list is the single list to applied to an Actor,
   * just like 'DirectoryType.forbidden type' for the Directories,
   * and 'UserIDType.forbidden type' for the UserID.
   * @return an unmodifiable HashMap where
  *   the key is a String of ACSObject class interface (like Resource) ended by
  *   a bound property name (like ResourceType), with '.' as separator,
  *   and the value is a String array of the values for this key.
  *   May be null but never empty.
  */ 
  public Map<String, String[]> getM_TypesAndGUIPolicy();
  
  /**
   * Gets the ACS constraints on the allowed or forbidden ACSObject types
   * to set the GroupID, Node, Resource, AclEntry and Privilege policies.
   * The values are types of ACSObject, that is
   * the type of a Basic or a VirtualFolder, or the second type of a Privilege.
   * The default value is always all the possible values, when there is no key
   * to match.
   * The keys are composed of a structured word which defines the application domain,
   * like 'GroupID.NoPrimaryGroup', and there is sometimes a type at the end. <P>
   * There are 29 forms for the key, presented hereinafter: <P>
   *     - 'GroupIDMember.NoMemberOf', where the values are the types of the groupmembers for
   *            which the role of member of another group is forbidden <P>
   *     - 'GroupIDMember.IsNotConditionalAclSource', where the values are the types of
   *            the GroupIDmembers for which the conditional ACL is forbidden as source <P>
   *     - 'GroupID.NoPrimaryGroup', where the values are the types of the groups for
   *            which the role of primary group for an account is forbidden <P>
   *     - 'GroupID.NoMainGroup' where the values are the types of the groups for
   *            which the role of main group of a resource is forbidden <P>
   *     - 'GroupID.NoSecondaryGroup' where the values are the types of the groups for
   *            which the role of secondary group of an actor is forbidden <P>
   *     - 'GroupID.MemberFromAdmin' where the values are the types of the groups
   *            where every member has to be an administrator <P>
   *     - 'GroupID.NoUserIDMember' where the values are the types of the groups
   *            where every member has to not be an UserID <P>
   *     - 'GroupID.NoGroupIDMember' where the values are the types of the groups
   *            where every member has to not be a GroupID <P>
   *     - 'GroupID.NoMoreThanOneMember' where the values are the types of the groups
   *            where only one member is allowed <P>
   *     - 'GroupID.ConditionalAclGroup' where the values are the types of the groups
   *            which may be condition groups in some conditional ACLs <P>
   *     - 'Resource.OneRightACL' where the values are the types of the resources
   *            which handled only AclEntries with no or one right <P>
   *     - 'Resource.NoAccount' where the values are the types of the resources
   *            which do not handle an account as owner (a group is possible from parent) <P>
   *     - 'Resource.NoConditionalACL' where the values are the types of the resources
   *            which never handle flexible or right-defined conditional AclEntries <P>
   *     - 'Resource.NoNonConditionalACL' where the values are the types of the resources
   *            for which an AclEntry without condition group(s) is never operational <P>
   *     - 'Resource.OneConditionGroupInACL' where the values are the types of the resources
   *            which handled only conditional AclEntries with no or one condition group <P>
   *     - 'Actor.NoCurrentAccount' where the values are the types of the actors
   *            for which a current account is forbidden <P>
   *     - 'Actor.NoNullCurrentAccount' where the values are the types of the actors
   *            for which a current account is mandatory <P>
   *     - 'Actor.NoCurrentGroup' where the values are the types of the actors
   *            for which a current group is forbidden <P> 
   *     - 'Actor.IsNotBridgeTarget' where the values are the types of the actors
   *            which cannot be bridge targets <P>
   *     - 'Actor.IsNotBridgeSource' where the values are the types of the actors
   *            which cannot be bridge sources <P>
   *     - 'Actor.IsNotConditionalAclSource', where the values are the types of
   *            the Actors for which the conditional ACL is forbidden as source <P>
   * <P>
   * The following keys include an ACSObject type, after the last point:
   *     - 'GroupID.TypesOfMemberFor.allowed_GroupID_type' where the values are 
   *            the allowed types of the members for a group having this type <P>
   *     - 'GroupIDMember.NoMoreThanOneGroup.allowed_GroupIDMember_type' where the values 
   *            are the GroupID types for which the groupIDmember may be member of 
   *            one group of this type at most <P>
   *     - 'GroupIDMember.RecommandedGroupsAsMember.allowed_GroupIDMember_type' 
   *            where the values are the GroupID types for which a message
   *            recommands to the user to put the groupIDMember as member;
   *            there, it is possible to describe some alternatives
   *            like 'group_type_one<OR>group_type_two' <P>
   *     - 'Directory.TypesOfChildFor.allowed_Directory_type' where the values are
   *            the allowed types of the children for a Directory having this type,
   *            or if the type is 'NULL', for a Resource without parent <P>
   *     - 'VirtualFolder.TypesOfMemberFor.allowed_VirtualFolder_type' where the values are 
   *            the allowed types of the members for a VirtualFolder having this type <P>
   *     - 'VirtualFolder.TypesOfChildFor.allowed_VirtualFolder_type' where the values are
   *            the allowed types of the children for a VirtualFolder having this type,
   *            or if the type is 'NULL', for a VirtualFolder without parent <P>
   *     - 'Resource.SecondTypesOfPrivilegeFor.allowed_Resource_type' where the values are
   *            the allowed second types of the privileges for a Resource having this type <P>
   *     - 'EligibleParty.SecondTypesOfPrivilegeFor.allowed_EligibleParty_type' where the values are 
   *            the allowed second types of the privileges for an EligibleParty having this type <P>
   *     - 'VirtualFolder.SecondTypesOfPrivilegeFor.allowed_VirtualFolder_type' where the values are 
   *            the allowed second types of the privileges for a VirtualFolder having this type <P>
   *     - 'Resource.TypesOfSourceForConditionalACL.allowed_Resource_type' where the values
   *            are the allowed EligibleParty types as source of a conditional ACL
   *            in a Resource having this type <P>
   * @return an unmodifiable HashMap<String, String[]> where
  *   the key is an ACSObject class interface (like Resource) ended by
  *   a bound property name (like ResourceType), with '.' as separator,
  *   and the value is a String array of the authorized types.
  *   May be null but never empty.
  */ 
  public Map<String, String[]> getM_TypeOrientedPolicy();
  
  /**
  * This map is initialized at the ACS creation to set the standard values
  * for the ACSObject Right properties (for example, the Linux rights), but only  
  * when there is no matching specialized rights. The possible keys are: <P>
  * - 'Resource.OwnerRights',<P>
  * - 'Resource.GroupRights',<P>
  * - 'Resource.OtherRights',<P>
  * - 'Resource.AclRights',<P>
  * - 'Actor.BridgeRights'. <P>
  * An AcsAddon ACS may have complementary keys, like for instance 'Resource.OtherRights'
  * in the ACSUbuntuImpl class. In all cases, these AcsAddon keys have to start
  * with the name of an interface, to be detected by
  * ACSFactoryUtilityImpl.getL_MapKeysForSpecializedRights(),
  * and they have to end with a property change name containing 'Rights',
  * to be detected by BaseUtilityImpl.getRestrictedValues().
  * @see #getM_SpecializedRights
  * @return unmodifiable HashMap where
  *   the key is a String of ACSObject class interface (like Resource) ended by
  *   a bound property name including 'Rights'(like groupRights), with '.' as separator,
  *   and the value is a StringRight array of the authorized standard values for 
  *   this bound property. May be null but never empty.
  */ 
  public Map<String, StringRight[]> getM_StandardRights();
  
  /**
   * Gets the specialized rights extending or constraining the allowed values
   * which are defined for each right type set in getM_StandardRights().
   * The specialized rights are then superior to the standard ones, which
   * are unworkable if some specialized rights are applicable.
   * If several sets of specialized rights are applicable, it is their intersection
   * which is workable. <P>
   * As map keys, examples of quartets (object.domain.key.subkey) are: <P>
   *     - 'Target.AclRightsSet.Interface.Directory', for ACL rights
   *           when the target is a Directory instance <P>
   *     - 'Target.AccRightsSet.Type.script', for Account rights
   *           when the target has the type 'script' <P>
   *     - 'Source.AclRightsSet.Type.group', for ACL rights
   *           when the source has the type 'group' <P>
   *     - 'PFType.PrvRightsSet.Type.role one', for PrivilegeForTypes
   *            having the second type 'role one' <P>
   * The object is Target or Source to indicate that these rights are for the target
   * (or source) which has the correct interface or type. <P>
   * For Privileges only, PFType indicates rights for PrivilegeForTypes, and
   * PFLink indicates that it is for PrivilegeForLinks. <P>
   * The domain may be one of the right categories:
   * 'AccRightsSet' for Target only, 'GpORightsSet' for Target only,
   * 'AclRightsSet' for Target only, 'BdgRightsSet', 'RooRightsSet' for root.
   * 'PrvRightsSet' for Privileges only. <P>
   * The key is 'Interface' or 'Type' to indicate that the subkey belongs to one
   * of these categories. <P>
   * The subkey is an interface name or a Basic type (or second type for a Privilege),
   * for which the specialized rights have to be used. <P>
   * The value in the returned map is an array of StringRight which may have
   * a lenght of 0.
   * @return an unmodifiable Map where the key is a quartet of Strings, and the value is an array
   * of StringRights. May be null but never empty.
   */
  public Map<String, StringRight[]> getM_SpecializedRights();
  
  /**
  * @return the immutable rules which are derived from the ACS structure
  */
  public List<String> getL_ImmutableRules();
  
  /**
  * Gets the ACS or metarights having one of the arguments as nick name.
  * @param _r1 nick name of an ACS right. May be null.
  * @param _r2 nick name of an ACS right. May be null.
  * @param _r3 nick name of an ACS right. May be null.
  * @param _r4 nick name of an ACS right. May be null.
  * @param _r5 nick name of an ACS right. May be null.
  * @return from one to five required ACS rights
  */
  public StringRight[] getL_SelectedRights(String _r1, String _r2, String _r3,
          String _r4, String _r5);

  
  // new in v0.7.0: Privileges and VirtualFolders
  
  /**
   * The current root is the first non-finalized virtual folder for resources
   * without parent which has been created.
   * Returns name of the root for the recorded virtual folders in the ACS base.
   * @return ImmutableName. May be equal to VectorNameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnResourceFoldersRootName();

  /**
   * The current root is the first non-finalized virtual folder for
   * eligible parties, without parent which has been created.
   * Returns name of the root for the recorded virtual folders in the ACS base.
   * @return ImmutableName. May be equal to VectorNameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnEPFoldersRootName();
  
  /**
  * Gets a privilege that is managed by this ACS.
  * @param _detailledName privilege detailled name
  * @return Privilege - May be null.
  */
  public Privilege getOnePrivilegeFromDetailledName(String _detailledName);
  
  /**
  * Gets a privilege that is managed by this ACS.
  * @param _name privilege name
  * @return Privilege - May be null.
  */
  //public Privilege getOnePrivilege(String _name);
  
  /**
  * Gets the privileges managed by this ACS. They may connect several resources
  * and several eligible parties from this ACS.
  * @return Privilege[] is a copy of the privileges of this ACS. May be null,
  * with empty privileges, or equal to 'new Privilege[] {new Privilege()}'.
  */
  public Privilege[] getEorL_Privileges();
  
   /**
   * Gets all the privilegeForLinks of the ACS, not the privilegeForTypes.
   * @return List of PrivilegeForLinks. May be empty, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForLinks();

   /**
   * Gets all the privilegeForTypes of the ACS, not the PrivilegeForLinks. 
   * @return List of PrivilegeForTypes. May be empty, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForTypes();

   /**
   * Gets all the PrivilegeForTypes which have the right source type. The comparison
   * returns true if a value starts with the key word '<immutable> ',
   * and not the other value. 
   * @param _type is the searched source type. No action if null.
   * @return List of PrivilegeForTypes. May be empty and then immutable, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForTypesForSource(String _type);
  
   /**
   * Gets all the PrivilegeForTypes which have the right target type. The comparison
   * returns true if a value starts with the key word '<immutable> ',
   * and not the other value.
   * @param _type is the searched target type. No action if null.
   * @return List of PrivilegeForTypes. May be empty and then immutable, but never null.
   */
  public List<Privilege> getEorL_PrivilegeForTypesForTarget(String _type);
  
  /**
  * Gets the virtual folders managed by this ACS.
  * @return VirtualFolder[]: copy of the virtual folders list. May be null or empty.
  */
  public VirtualFolder[] getEorL_VirtualFolders();

  /**
  * Gets a virtual folder managed by this ACS.
  * @param _name is the folder name
  * @return VirtualFolder. May be null.
  */
  public VirtualFolder getOneVirtualFolder(ImmutableName _name);
  
  /**
  * Gets the virtual folders managed by this ACS and which contain only
  * resources.
  * @return full copy of the relevant virtual folders. May be empty but not null.
  */
  public List<VirtualFolder> getEorL_ResourceVirtualFolders();
  
  /**
  * Gets the virtual folders managed by this ACS and which contain only
  * eligible parties.
  * @return full copy of the virtual folders list. May be empty but not null.
  */
  public List<VirtualFolder> getEorL_EPVirtualFolders();

  /**
   * Copy of the external virtual folders in a new array. 
   * They are defined as owned by another ACS but having rights through 
   * an external Privilege that is managed by this ACS.
   * @return copy of the virtual folders list. May be null.
  */
  public VirtualFolder[] getEorL_ExternalVirtualFolders();

  /**
   * Tests if a privilege is known.
   * @param  _name is an ImmutableName of the privilege to test
   * @return boolean true if the ACS has recorded this privilege instance
   */
  public boolean containsPrivilegeName(ImmutableName _name);

  /**
   * Tests if a virtual folder is known.
   * @param  _name is an ImmutableName of the virtual folder to test
   * @return true if the ACS has recorded this virtual folder instance
   */
  public boolean containsVirtualFolderName(ImmutableName _name);

  /**
   * Tests if the ACS owns the virtual folder.
   * @param  _name of the virtual folder to test
   * @return boolean true if this virtual folder is from this ACS
   */
  public boolean containsOwnVirtualFolderName(ImmutableName _name);

  /**
   * Gets the structure of this ACS. For the basic properties, there is
   * a dedicated method to test them, like for instance the method getManageActors().
   * For the non-basic properties,
   * they are included as keys and it is not necessary to have a Boolean.TRUE as value
   * to use the property, since the presence of the key is simply tested.
   * @return the immutable ACS structure. Never empty nor null.
   */
  public SortedMap<String, Boolean> getM_Structure();

  /**
   * Gets the ACS rights and the metarights this ACS handles and for a given internal key value
   * of the associated metaright. The map key is the metaright internal key.
   * @return the unmodifiable rights map. Never empty nor null.
   */
  public SortedMap<String, StringRight[]> getM_RightsOfInternalKey();

  /**
   * The format is 'Editor: full_editor_name - AcsAddon: addon_name'.
   * The AcsAddon name is 'NONE' if the ACS is not from an AcsAddon.
   * @return the editor name and the AcsAddon name for the ACS. Never null.
   */
  public String getEditorAndAddonNames();

} //ACSCatalog

