//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.gBase.*;
import java.util.Map;
import java.util.List;

import ARoad0.ErrB.BaseError;

/**
* This interface is responsible for managing
* the ACSs in the generic base through specific methods.
* @see BaseManagerImpl
* @version 0.7.0
*/
public interface ACSManager extends BaseBeanManager, BoundBean
{

  /**
  *  Opens the base with all known ACS that are open.
  * @exception BaseError if an acs is not known or ClassNotFoundException,
  *  or IOException, or....
  */
  //public void openDefaultBase() throws BaseError;

    /**
    * Opens a registered and closed ACS from a file in the current directory.
    *  Analyze the need to open the closed aclEntries in every open ACS
    *  and in this ACS, with checkClosedAclEntries().
    *  No need for closing open aclEntries, since all external AclEntries
    *  have been closed during the last save of this ACS.
    * @param _name ImmutableName of the ACS to open. Its file name is built with the two last
    * components of the name, or with the only component if it is alone.
    * @exception BaseError if the ACS is not known, the ACS is open, the open
    *   ACS has not the same name, or ClassNotFoundException, IOException.
    * @return ACS: direct reference to the opened ACS after setting
    *  all the external references to open ACS, and deleting all the references
    *  to closed ACS.
    */
  public ACS openACS(ImmutableName _name) throws BaseError;

  /**
  * @param _name ImmutableName of an ACS
  * @return boolean true if _name is known as an ACS name. The ACS may be
  * open or closed.
  */
  public boolean knowsACS(ImmutableName _name);

  /**
  * @param _name ImmutableName of an ACS
  * @return boolean true if _name is known as an ACS name and the ACS is open.
  */
  public boolean isOpenACS(ImmutableName _name);

  /**
  * Returns the names of open or closed ACS.
  * @return unmodifiable list of ACS names
  */
  public List getL_ACSNames();

  /**
  * @return unmodifiable map of non-null instances of ACS
  */
  public Map getEorM_OpenACS();

    /**
    *  Saves the base with the ACS and view lists, even if there are empty.
    *  All the open ACS, but not the closed ones,
    *  are saved in some files (one per open ACS) in the default directory.
    *  All the open views, but not the closed ones,
    *  are saved in some files (one per open view) in the default directory.
    *  Fires a 'OpenACS' event before each ACS closing, to done it before the ACS
    *  proper events.
    *  Fires a 'OpenViews' event before each view closing, to done it before the view
    *  proper events.
    * @exception BaseError if an acs or a view is not open, ACSname.lastComponent() is null,
    * FileNotFoundException or IOException.
    */
    public void saveDefaultBase() throws BaseError;

    /**
    *  Saves and closes the base with the ACS and the view lists, even if it is empty.
    *  All open ACS and views, but not the closed ones,
    *  are saved in some files (one per open ACS and view) in the default directory.
    * @exception BaseError if an acs or a view is not open, ACSname.lastComponent() is null,
    * FileNotFoundException or IOException.
    */
    public void saveAndCloseDefaultBase() throws BaseError;

    /**
    *  Saves an opens ACS with out close it.
    *  The ACS instance is not modified.
    *   - If one of its resources or one of its eligible party
    *   is controlled by the aclEntry of an another ACS, it is NOT unrecorded
    *   in this another ACS.
    *   - If _acs is a controller, its external aclEntries are unrecorded in the
    *   controlled resources and relevant eligible parties, and in their main ACS,
    *   with _acs.closeAclWithExternalReferences(), then _acs is saved and
    *   its external aclEntries are recorded again with _acs.checkClosedAclEntries(),
    *   but they are new instances THAT MUST BE TAKEN IN ACCOUNT BY THE APPLICATION.
    *   ACSManager then fires a swithExternalAclEntriesEvent
    *   to the registered listeners in the application (  TO DO  ).
    * @param _acs ACSImpl to be saved.
    * @exception BaseError if _acs is not open, ACSname.lastComponent() is null,
    *  or FileNotFoundException and IOException.
    */
    public void saveOpenACS(ACSImpl _acs) throws BaseError;

    /**
    *  Closes an open ACS without saving it.
    *  Uses this method after a saving, or when an ACS instance is corrupted,
    *  before the opening of the old saving.
    *  The ACS is still known in ACSManager.
    *  Call closeACS() for each ACS under _acs in the IS nametree, and for _acs.
    *  After the call of this method, set the _acs to null, or
    *  restore it with _acs = openACS(cloned _acs name).
    * @param _acs ACS to be closed and finalized.
    * @param _withSaving true to save the ACS
    * @exception BaseError if _acs is not open.
    */
    public void closeOpenACS(ACSImpl _acs, boolean _withSaving) throws BaseError;

    /**
    * Closes the base without savings, with all the ACS and all the views.
    * @exception BaseError if an acs is not open.
    */
    public void closeDefaultBase() throws BaseError;

  /**
  * The ACS name is removed from the ACS list and from the IS nametree, like
  * its children in the IS nametree if they exist.
  * If it is open, the ACS is closed and saved before.
  * So, it will be possible to restore the ACS in the futur.
  * To handle a corrupted ACS instance and recover a correct ACS from file,
  * you must rather use closeOpenACS().
  * The ACS saving file is not deleted, but it could be by the user, or be moved for
  * a future use. ACSManager.openACS(_name) wont work..
  * @param _name ImmutableName of the ACS to remove from the base.
  * @exception BaseError if the ACS is unknown, or it is not a leaf in the IS nametree,
  *  or if ACSname.lastComponent() is null, FileNotFoundException and IOException.
  */
  public void removeACS(ImmutableName _name) throws BaseError;

  /**
  * @return long static final
  */
  public long getSerialVersionUID();


} // ACSManager class

