//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.beans.PropertyChangeListener;
import java.util.Map;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

/**
* This interface is responsible for returning the first immutable ACS properties
* and the first rights. It is used to create the acsrights before the selection of
* the allowed rights for each right type (Group, ACL,...). It contains
* the associations to handle at the creation or the deletion of some ACSObjects,
* by the mean of actions to do by the program. It allows to get
* the ACS vocabulary that defines the AcsAddon classes.
* This interface is then immplemented by both ACSFactoryImpl and ACSImpl. 
* 'OneRight' is the single bound property.
* @version 0.7.0
*/
public interface ACSRun extends BaseObject, BaseBeanManager {
  
  /**
   * Get the name of the information system which owns this acs.
   * @return a copy of the IS name. May be empty, but not null.
   */
  public ImmutableName getISName();
    
    /**
 * Returns the ACS type, as ACSImpl.GENERIC, LINUX,...
 * @return the acs type
 */
public String getACSType();
   
  /**
   * Get the mandatory parent name for the ACS in the IS nametree.
   * @return the parent name. May be null or empty.
   */
  public ImmutableName getParentName(); 

 /**
   * Get all the rights which are allowed in this ACSRun.
   * May be StringMetaRights which then do not belong to this ACSRun.
   * @return all the acs and meta rights.
   */
  public StringRight[] getL_Rights();
  
  /**
   * Get the first acsright which has 'full_control' as metaright, or this metaright
   * if no acsright complies this condition.
   * 'full_control' metaright has 'deny_all' metaright for opposite,
   * and its single user upper right is 'modify_security_descriptor'.
   * @return the right. Never null. 
   */
  public StringRight getFullControlRight();
  
  /**
   * Get a right which is registered by this ACSRun. It may be a StringMetaRight.
   * @param _nickName is the nick name of the right, provided by StringRight.getNickName().
   * @return the right; null if it does not exist. 
   */
  public StringRight getOneRight(String _nickName); 
  
  /**
   * Get a right which is registered by this ACSRun. It may be a StringMetaRight.
   * @param _name is the name of the right, provided by StringRight.getName().
   * @return the right; null if it does not exist.
   */
  public StringRight getOneRight(ImmutableName _name); 
  
  /**
   * Get a right which is registered by this ACS. It may be a StringMetaRight.
   * @param _metaRight is the right of the metaright of the searched right, 
   *             provided by StringRight.getRight().
   * @return the right; null if it does not exist.
   */
  public StringRight getFirstRightFromMeta(String _metaRight);
  
  /**
   * Register a new acsright in this ACS. 
   * 'OneRight' is a bound property.
   * @param _right is the new right to register. Not null.
   * @exception CreateError if the right is null, a metaright or if
   *   the right is already used.
   */
   public void addRight(StringRight _right ) throws CreateError;

  /**
   * Unregisters a right in this ACSRun.
   * Used by the StringRight.finalizeForUser() method.
   * 'OneRight' is a bound property.
   * @param _right the right to unregister.
   * @exception UpDateError if the right is null or not known.
   */
   public void removeRight(StringRight _right ) throws UpDateError;  
  /**
  * Add a listener to the bean.
  * Now only for the property 'OneRight'.
  */
  public void addPropertyChangeListener(String _propertyName,
    PropertyChangeListener _l);

  /**
  * Add a listener to the bean.
  * Now only for the property 'OneRight'.
  * @param _propertyName name of the property
  * @param _l listener
  */
  public void removePropertyChangeListener(String _propertyName,
    PropertyChangeListener _l);
  
  /**
   * Get the associations to handle at the creation or the deletion of some ACSObjects,
   * by the mean of actions to do by the program. Actions have numbers from 1 to 9.
   * The criteria is a map key among the followings: <P>
   * - 'Class'.'Property'.<Value>.XXX: a new object property has the value XXX; <P>
   * - 'Class'.'Property'.<ValueContains>.XXX: a new object property value contains the value XXX; <P>
   * - 'Class'.'Property'.<Class>.XXX: a new object property has exactly the class XXX; <P>
   * - 'Class'.'Property'.<NameStartsWith>.XXX: the new object property name 
   *        starts with the ACS name plus the subname XXX; <P>
   * - 'Class'.<NameStartsWith>.XXX: the new object name of the class
   *        starts with the ACS name plus the subname XXX; <P>
   * - <Class>.<Value>.XXX: the new object class is exactly ARoad0.gBase.XXX; <P>
   * - <Class>.<ValueImplements>.XXX: the new object class is a subclass of ARoad0.gBase.XXX. <P>
   * <P>
   * The allowed String map values depends lightly on the map key.
   * The values define one or several actions to do if the key matches.
   * There are 4 possible formats for the first value of the action N: <P>
   * - N.<AssociateClass>.XXX which specifies the class name XXX in gBase, without the Impl String,<P>
   *    of an associate object to build up or to finalize when the master is build up <P>
   *    or finalized <P>
   * - N.<CalledMethod>.<Master>.YYY which specifies the method YYY to call on the master, <P>
   * - N.<CalledMethod>.<ACS>.YYY which specifies the method YYY to call on the master ACS, <P>
   * - N.<CalledMethod>.<Action>.C.ZZZ which specifies the method ZZZ <P>
   *    to call on the associate previously created by the action number C. <P>
   * After each first value, there is the optional list of the constraints on the arguments
   * of the constructor (case 1) or the method (cases 2 to 4),
   * and there is one constraint specification per argument.
   * Each argument constraint specification belongs to a set of 13 allowed formats,
   * presented herein for the action N: <P>
   * - N.'PropertyName'.<Null>: null value for the argument having the name 'PropertyName' <P>
   * - N.'PropertyName'.<Master>: the master is the argument <P>
   * - N.'PropertyName'.<MasterValue>: the master value for the same property <P>
   * - N.'PropertyName'.<MasterNameLastComponent>: the last component in the master name <P>
   * - N.'PropertyName'.<MasterState>: boolean true if the master is creating, false if it is finalizing <P>
   * - N.'PropertyName'.<Value>.XXX: the argument is XXX <P>
   * - N.'PropertyName'.<Value>.XXX+<MasterValue>+YYY: the master value <P>
   *        for the same property, with some optional pre and post String values <P>
   * - N.'PropertyName'.<Value>.XXX+<MasterNameLastComponent>+YYY: <P>
   *        the last component in the master name, with some optional  <P>
   *        pre and post Name values as Strings <P>
   * - N.'PropertyName'.<Value>.<MasterNameContext>+YYY <P>
   *        the context in the master name, with a post Name value as String <P>
   * - N.'PropertyName'.<Value>.<ACSName>+YYY: the master ACS name, <P>
   *        with a post Name value as String <P>
   * - N.'PropertyName'.<Value>.XXX+<AssociateValue>.C.+YYY: the associate number C <P>
   *        value for the same property, with some optional pre and post String values  <P>
   * - N.'PropertyName'.<Value>.XXX+<AssociateNameLastComponent>.C.+YYY: <P>
   *        the last component in the associate number C name, with a post String value  <P>
   * - N.'PropertyName'.<Value>.<AssociateNameContext>.C.+YYY: <P>
   *        the context in the associate number C name, with a post String value  <P>
   * The internal map may contain values where the first 2 characters are not 'N.', but
   * the first no-number value and its followings are returned by this method which is used by
   * StringMapPropertyEditor to edit the internal map.
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings started by a number between 1 to 9 followed by a point.
   *   May be empty, but not null.
   */
  public Map<String,String[]> getM_AssociationsOnNewObject();
  
  /**
   * Gets comments to explain how this ACS simulates a real system/software.
   * The criteria is an ACS property displaying name. The value is a String array
   * of free user comments on the settings of the related ACS property. <P>
   * @return a copied Map where the key is a String, and the value is an array
   *   of Strings. May be empty, but not null. A value may be empty but is never null.
   */
  public Map getM_CommentsOnProperties();

  /**
   * Gets the ACS vocabulary. This may cover several independant domains: <P>
   * - the AcsAddon editor, the AcsAddon package editor and the AcsAddon name,
   * in three keys, <P>
   * - how to name the L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS constants for this ACS,
   * as names of the ACS tree nodes in the proper ACS internal frame. An example
   * is to replace the tree node 'Actors (right user target)' by 'Scripts'. <P>
   * - what are the gBase classes to use, from the generic package or from
   * the AcsAddon gBase package (see L_CLASSIC_CLASS_KEYS), like when
   * the Addon Ubuntu subclasses ResourceImpl with ResourceUbuntuImpl. The
   * full Java name of the subclass has to be put in the value array. <P>
   * - If the ACS is an AcsAddon, what are the specific new classes in
   * the AcsAddon gBase package (see ACSADDON_NEW_CLASSES). In this case,
   * the class names are separated by one space, like in
   * 'ExecutableUbuntu XXXXUbuntu YYYYUbuntu', without the full name of the classes. <P>
   * - If the ACS is an AcsAddon, what are the non-generic rights in the
   * ACSObjects on which the ACS constraints of the specialized rights
   * are applicable, and in that case, what domain is applicable. An example
   * is the domain 'GpORightsSet' for the property 'OtherRights', in Ubuntu,
   * that is declared by the value 'OtherRights.GpORightsSet'. <P>
   * A new ACSFactory in an AcsAddon package sets its vocabulary in its constructor,
   * so it is possible to call this method just after the construction of the factory.
   * @see ARoad0.gBase.ACSFactoryUtilityImpl#L_ACS_TREE_UPPER_NODES_FOR_ONE_ACS
   * @see ARoad0.gBase.ACSFactoryUtilityImpl#L_CLASSIC_CLASS_KEYS
   * @see ARoad0.gBase.ACSFactoryUtilityImpl#L_ACS_TREE_GENERIC_UPPER_NODES
   * @return an unmodifiable Map where the key is a String, and the value is a String
   * array. May be null. A map value is never empty but may be null.
   */
  public Map<String,String[]> getM_AcsVocabulary();
  
} // ACSRun

