//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

/**
 * This interface is responsible for modeling the access control list entry which manages
 * some access grantings or some access denyings from an EligibleParty to a Resource.
 * An AclEntry is conditional when it uses condition group(s) for which the source
 * has to be member for activating the AclEntry rights. There are two types of conditional AclEntry,
 * the simple-condition and the right-condition AclEntry.
 * The AclEntry, the EligibleParty and the Resource may belong to
 * three different access control systems, or, in the simplest case, to only one ACS.
 * An AclEntry is created by the resource, that specifies an eligible party 
 * as argument for the constructor. <P>
 * An AclEntry may be conditioned by the groups of the source, and then, the returned rights
 * are null if the source is not member of all the condition group(s).
 * The conditional AclEntry is an alternative to the LINKED_IF_ALL_SOURCES privilege,
 * but it is focused to deliver rights to one source which has to be member of
 * condition groups. A conditional AclEntry has the same ACS for managing
 * both the AclEntry, the condition group and the eligible party. Thus, a conditional AclEntry
 * has never an external eligible party. <P>
 * An AclEntry is created and managed by its resource, and it is deleted
 * when the resource is deleted. An AclEntry is uniquely defined by the set
 * (ACS, Resource, EligibleParty, first condition GroupID, sens of the right)
 * where the first condition GroupID is null if it is not a conditional AclEntry.
 * The AclEntry identity changes for a conditional AclEntry
 * when its first condition GroupID changes, and this is why the first condition group
 * is immutable after the creation of the AclEntry. <P>
 * This interface manages the following ACS constraints on the ACSObject type policy: <P>
 *     - 'GroupID.ConditionalAclGroup' where the values are the types of the groups
 *            which may be condition groups in some flexible or right-defined conditional ACLs <P>
 *     - 'Resource.OneRightACL' where the values are the types of the resources
 *            which handled only AclEntries with no or one right <P>
 *     - 'Resource.NoConditionalACL' where the values are the types of the resources
 *            which never handle flexible or right-defined conditional AclEntries <P>
 *     - 'Resource.NoNonConditionalACL' where the values are the types of the resources
 *            for which an AclEntry without condition group(s) is never operational <P>
 *     - 'Resource.OneConditionGroupInACL' where the values are the types of the resources
 *            which handled only conditional AclEntries with no or one condition group <P>
 *     - 'Resource.TypesOfSourceForConditionalACL.allowed_Resource_type' where the values are
 *            the allowed EligibleParty types as source of a conditional ACL
 *            in a Resource having this type <P>
 * It is a javabean with the following bound properties:
 * 'Comment', 'AclRights', 'ConditionSourceGroups'.
* @see ARoad0.gBaseInterface.EligibleParty
* @see ARoad0.gBaseInterface.Resource
*/
public interface AclEntry extends ImmutableAclEntry, BoundBean {

  /**
   * @param _ep the Eligible Party implied in the access control entry
   * @return void
   */
//	public  void setEorEP(ImmutableEligibleParty _ep);

  /**
   * Sets the rights of the AclEntry, from which the effective rights
   * given by getRight() are derived. Only header rights are used.
   * The ACS constraints on 'AclRights' are applied even if they are null,
   * and they depend on both the EligibleParty and the Resource. This is not
   * like for AGO rights. To explain the result, an exception is thrown as an user
   * message when the ACS constraints are no fullfilled.
   * Controls the ACL rights constraints in the ACS from the resource,
   * and also from the EP only if the AclEntry is not external (that is, the ACS
   * is the same for the AclEntry and the EP). <P>
   * Only non-null, non-empty rights are set as new rights.
   * Creates a 0-size array if the argument is null.
   * Controls if each right has the correct sens.
   * Controls the ACS type policy on 'Resource.OneRightACL', 'Resource.NoConditionalACL',
   * 'Resource.NoNonConditionalACL'. <P>
   * A right-defined conditional AclEntry must have a single non-null right.
   * For a conditinal AclEntry, _ep is a GroupIDMember, and the right is workable
   * only if _ep is a member of each of these condition groups. Otherwise,
   * the resulting effective right from getRight() is a 0-size array.
   * Calls BaseUtilityImpl.getL_AcsRestrictedRights() and UtilityImpl.selectHeaders(). <P>
   * Fires a PropertyChangeEvent "AclRights" to the listeners.
   * Note: 'AclRights' is also the bound property name of the ACLs in ResourceImpl,
   * to allow the work of the Gui1 package. For Gui2.GraphicViewBaseListenerImpl
   * that listens the 'AclRights' events both from the Resource and the AclEntry,
   * the source and the value types in the event allows to process the event properly.
   * @param _st is an array of non-null, non-empty rights
   *    for the couple (resource, EligibleParty) in the AclEntry ACS.
   *    The authorized rights depend on the AclEntry ACS. May be null.
   * @exception CreateError if a right is null, empty,
   *    if the resource or the eligible party is immutable,
   *    if a right is not allowed, has not the sens of getPositiveRight(),
   *    or does not fulfil with the ACL constraints.
   */
  public  void setL_RightsForUser(StringRight[] _st) throws CreateError;


  /**
  * Sets any comment related to this resource. It is a bound property.
  * @param _s comment which is associated to the resource.
  */
  public void setComment(String _s);

  /**
   * To activate the ACL rights, the EP has to be member of the conditionnal GroupIDs.
   * If the EP is not a GroupIDMember while there is at least one conditional group,
   * the rights are empty. If the AclEntry uses static conditional group(s),
   * it is not possible to update them directly. If a conditional group is finalized, 
   * this AclEntry does not listen the event but the method getRight() returns empty.
   * @param _grp conditionnal group to add
   * @exception UpDateError if the AclEntry ACS does not manage the flexible
   * conditional group, or the argument is null or not from the AclEntry ACS
   */
  public void addForUserConditionSourceGroup(ImmutableGroupID _grp) throws UpDateError;

  /**
   * To activate the ACL rights, the EP has to be member of the conditionnal GroupIDs,
   * if it is a GroupIDMember. If the AclEntry uses static conditional group(s),
   * it is not possible to update them directly. If a conditional group is finalized, 
   * this AclEntry does not listen the event but the method getRight() returns empty.
   * @param _grp conditionnal group to remove. Never null.
   * @exception UpDateError if the AclEntry uses static conditional group, or the argument
   * is null or not from the AclEntry ACS
   */
  public void removeForUserConditionSourceGroup(ImmutableGroupID _grp) throws UpDateError;

  /**
   * To activate the ACL rights, the EP has to be member of the conditionnal GroupIDs
   * if it is a  group member.
   * Fires a PropertyChangeEvent "ConditionSourceGroups" to the listeners.
   * @param _grp conditionnal group to add
   */
  //public void addConditionSourceGroup(ImmutableGroupID _grp);

  /**
   * To activate the ACL rights, the EP has to be member of the conditionnal GroupIDs,
   * if it is a  group member.
   * Fires a PropertyChangeEvent "ConditionSourceGroups" to the listeners.
   * @param _grp conditionnal group to remove
   * @exception CreateError if _grp is not registered
   */
  //public void removeConditionSourceGroup(ImmutableGroupID _grp) throws CreateError;  
  
  /**
   * @param _b boolean true if the right allows access to the resource, false
   *  if the right forbids access
   * @return void
   */
//	public  void setPositiveRight(boolean _b);

} // AclEntry