//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.Map;

/**
 * This interface is responsible for modeling an access control system (ACS)
 * that is used in an ACS Addon package.
 * It defines some methods to inform the PropertyChange listeners of the ViewInBase,
 * what are the specific events to listen from in the direct and indirect parents of an ACSObject,
 * and at what conditions.
 * The listeners are gBase.ViewInBaseImpl and gWork.RightsMediatorImpl.
 * If the ACSObject is a Resource or a VirtualFolder, the events to listen are from the
 * indirect parents for the changings in some inherited rights. For instance, there is
 * the event 'FromTheParentOther' in the AcsAddon Ubuntu for the class ResourceUbuntuImpl.
 * @version 0.7.0
 */
public interface AcsAddon {
    
  /**
   * Gets the specific events to listen on a Resource or a VirtualFolder,
   * and its direct and indirect parents, because these events may change
   * the rights of the Resource or VirtualFolder in an ACS Addon.
   * Thhese events are listened to update the views containing these ACSObjects.
   * This method covers only the parent properties of a Resource or a VirtualFolder
   * (see ARoad0.gBaseInterface.ACSObjectAddon.getAddonInnerPropertiesToListen()
   * for the other cases).
   * To define the events, the map key may have the following values: <P>
   * - 'THIS' to say that the single basic object is the Resource or VirtualFolder, <P>
   * - 'ACS.method_name' to say that the list of Resources (or VirtualFolders) is returned
   * by the method 'method_name' in the Addon ACS class. <P>
   * For each map key, the values have the following format: <P>
   * 'AAA.method_name IF BBB' <P>
   * where: <P>
   * - AAA: PARENT (mandatory), <P>
   * - method_name: 'get' method name of each basic objet parent
   * that is associated, in the relevant BeanInfo class, to the event to listen (mandatory) <P>
   * - IF BBB: is optional, where BBB is TRUE, or FALSE, or NULL, or NOT-NULL, to say that
   * the event must be listened if and only if the returned value of the method_name is
   * respectively a boolean TRUE, or FALSE, or an object NULL, or NOT-NULL. <P>
   * When the map key starts with 'ACS', the ACS method to call has no argument.
   * When the map value starts with 'PARENT', the parent method to call has no argument.
   * Called by ARoad0.gBase.BaseUtilityImpl.add/removeAddonInner/ParentPropertyChangeListener().
   * For instance, 'PARENT.getAGOAccessFromTheParentOther IF FALSE' is the value
   * the ACS Addon Ubuntu ACS returns for the key ACS.getEorL_FromOtherInCurrentContextActors,
   * to define the listening of the 'FromTheParentOther' event on all the indirect parents
   * of a ResourceUbuntu in a view.
   * @return map to define the specific ACS Addon events to listen. May be null.
   */
  public Map<String, String> getAddonParentPropertiesToListen();


} // AcsAddon
