//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.gBase.*;
import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.UpDateError;

/**
 * This class is responsible for modeling an acting entity which may access
 * to resources controlled by the ACS, and which may play
 * some special roles, as a bridge, between an another actor and
 * some resources.
 * Exemples of such actors are servers or programs.
 * This class also models the operating system process which has a current
 * account and group, and belongs to several groups.
 * The current account and group, and the secondary groups, cannot be updated
 * if the resource type does not allow it. But the resource type does not control
 * the bridge target.
 * The setActorAlias() method in 0.6 version has been deleted, and it is
 * replaced by ImmutableBasic.getEorReferenceAsInternalHardAlias() for the AcsAddons
 * only.
 * @version 0.7.0
 */
public interface Actor extends Resource, EligibleParty, ImmutableActor
{
    
  /**
  * The current userID is not null when, typically, the actor is an executable
  * launched under an another userID. This class does not overiddes Resource.setEorUserID()
  * since the userID may be different, for instance when the group rights
  * authorizes the launching to the launcher. Resource.userID is
  * an access target property, and Actor.currentUserID is a rights user property.
  * Fires an event.
  * @param _u the current userID of the actor. May be null.
  * @exception UpDateError if the current userID is immutable.
   */
  public void setEorCurrentUserIDForUser(ImmutableUserID _u) throws UpDateError;

  /**
  * The current groupID is not null when, typically, the actor is an executable
  * launched under an another groupID. This class does not overiddes Resource.setEorGroupID()
  * since the groupID may be different, for instance when the ACL rights
  * authorizes the launching to the launcher. Resource.groupID is
  * an access target property, and Actor.currentGroupID is a rights user property.
  * @param _g the current groupID, that for an actor may be
  * changed by itself and different to the resource groupID
  * (an actor is a kind of resource).
  * @exception UpDateError if the current groupID is immutable.
  */
  public  void setEorCurrentGroupIDForUser(ImmutableGroupID _g) throws UpDateError;

  /**
   * A secondary group is used by an Actor as rights user, to get complementary
   * accesses to targets only when a current userID is set. So, when the actor
   * is accessed as any target, the current userID of the rights user is activated
   * with its secondary groups, and the secondary groups returned by this method
   * are not activated.
   * Unregister the previous groupIDs and set the new ones.
   * Remove in the argument, the current group or GroupIDImpl.EMPTY_INSTANCE.
   * Noop if the resource type contains <byAcsCreator> or <immutable>.
   * Fire an event 'SecondaryGroups'.
   * @param _L array of the secondary groupID attached to this actor. May be null.
   * @exception UpDateError if the secondary groupIDs are immutable.
   */
  public void setEorL_SecondaryGroupForUser(ImmutableGroupID[] _L) throws UpDateError;

  /**
   * @param _L ImmutableActor[] array of actors for which this actor
   *  is a gate. Used as an External Object Reference. May be null.
   */
  //public  void setEorL_isAGateFor(ImmutableActor[] _L);

  /**
   * Sets the another-ACS actor for which this actor is a bridge source.
   * @param _target a reference on the bridge target. May be null.
   * @exception CreateError if the target belongs to the same ACS, or if the
   *    target changing is not validated by the connected objects.
   */
  public void setEorACSBridgeTarget(Actor _target) 
    throws CreateError;

  /**
   * Noop if the resource type is <byAcsCreator> or contains <immutable>.
   * @param _b true if this Actor may be a proxy in the access paths, that is,
   *    as rights user, it gets some rights to the other
   *    rights users which have an 'execute' right on it
   * @exception UpDateError if the right proxy flag is immutable.
   */
  public void setIsRightProxy(boolean _b) throws UpDateError;
  
} // Actor

