//Title:        Access Road
//Version:      0.7.1
//Copyright:    2002, 2004, 2007, 2008, 2009, 2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.BaseError;

/**
* This is an interface to inform that it is an object with a detailled name
* which gives the key components of the object. Allows a clearer presentation
* of an object in the explorer and into the beamer. The beamer tests if an object
* is a DetailledName, and if yes, displays the detailled name rather than the name.
* Used for the AclEntries and the Privileges.
* To allow to get the BaseObject without errors, a detailled name needs to be completed
* by the ACS name. However, an external entity needs to include its ACS into its detailled name.
* The classes AclEntryImpl and PrivilegeAbstract implements the method getNameFromDetailledName(). <P>
* To know if the detailled name is about an AclEntry or a Privilege, the following rules
* are applied: <P>
* - a detailled name contains at least one sequence '||' and at most one sequence '>>',
* - an AclEntry detailled name contains the one sequence '>>', and '>>' is before
* the first '||' sequence.
* - a Privilege detailled name may not contain the sequence '>>',
* and otherwise, '>>' is after the last '||' sequence. <P>
* For a type privilege, the possible forms are: <P>
* - privilegeType || nickName || first right || source type >> target type
* For a linked privilege, the possible forms are: <P>
* - privilegeType || nickName || first right, for a internal privilege, <P>
* - ACS: ACS_name | privilegeType || nickName || first right, for a external privilege. <P>
* The form of the detailled name for an internal ACL is: <P>
* - incomplete Eligible Party name >> incomplete Resource name ||
* grant OR deny || first condition group last component. <P>
* The last part '|| first condition group last component' is there only if there
* is a condition group.
* The form of the detailled name for an external ACL is: <P>
* - ACS: ACS_name | (full or incomplete) Eligible Party name >>
* (full or incomplete) Resource name || grant OR deny ||
* &lt;Source ACS group&gt; OR &lt;Target ACS group&gt; ||
* first condition group last component. <P>
* An Eligible Party or Resource name is incomplete if it belongs to the AclEntry ACS.
* The condition group is put in the name or the detailled name only for
* a flexible conditional AclEntry. '&lt;Source ACS group&gt;' is used
* when the condition group belongs to the source (Eligible Party) ACS.
* Due to the method setValue() of CollectionPropertyEditor, there are 3 constraints: <P>
* - an array or a collection of DetailledName cannot be updated with a global 'set' method,
* nor a 'choiceMethodForNew' parameter.
* Some 'add/remove' methods are mandatory if there is no constructor/finalizing method. <P>
* - a DetailledName instance can be deleted through a 'finalizeForUser' method
* like a Privilege, or through any other method the BeanInfo property descriptor specifies,
* like an AclEntry where a Resource method is unused. <P>
* - it it not possible to use the beamer 'Prev.' to get the values of the previous
* property the beamer displays.
*/
public interface DetailledName extends FinalizedObjectForUser {

  /**
  * This method returns the detailled name for the explorer,
  * which gives the key components of the object.
  * On the contrary of getName(), it may do not contain the ACS name,
  * and returns a string which is more easily readable by the Access Road user.
  * An external object has a detailled name starting with 'ACS: acs_name | '.
  */
  public String getDetailledName();

  /**
  * Gets the detailled name from the BaseObject name.
  * The format of the returned value depends on the class of the BaseObject.
  * @see ARoad0.gBase.PrivilegeAbst#buildUpDetailledName(java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String) 
  * @see ARoad0.gBase.AclEntryImpl#buildUpDetailledName(ARoad0.gBaseInterface.ImmutableName, ARoad0.gBaseInterface.ImmutableName, boolean, java.lang.String)
  * @param _name name of the BaseObject
  * @return the detailled name
  */
  public String getDetailledNameFromName(ImmutableName _name) throws BaseError;

  /**
  * Extracts the key property names from the detailled name
  * of an object.
  * @param _detailledName is the viewable name of the object
  * @return array of properties which describes the object, and the items
  * are implementation dependant.
  */
  public ImmutableName[] getKeyPropertiesFromDetailledName(
    String _detailledName) throws BaseError;

  /**
  * Extracts the BaseObject name from the detailled name of an object.
  * This is a pseudo-static method, since it does not depend on the instance.
  * @param _acs ACS of the BaseObject
  * @param _detailledName the viewable name of the object
  * @return name of the object
  */
  public ImmutableName getNameFromDetailledName(ImmutableACS _acs,
    String _detailledName) throws BaseError;
}