//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.SortedMap;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.BaseError;

/**
 * This class is responsible for modeling in the base, a view
 * about access control issues between eligible parties (EP of EPRViewInBase)
 * as access sources, and resources (R of EPRViewInBase) or virtual folders
 * (containing Resources or EPs) as access targets.
 * It is used by the full views and the sketch views in the GUI to display
 * all the rights between any couple of view objects, since an access source
 * may also be an access target sometimes.
 * @see ARoad0.Gui2.GraphicEPRView
 * @see ARoad0.Gui2.GraphicSketchView
 */
public interface EPRViewInBase extends ViewInBase {

  /**
   * @param _targ to test
   * @return true if the view contains it
   */
  public boolean containsTarget(ImmutableTarget _targ);

  /**
   * @param _sour to test
   * @return true if the view contains it
   */
  public boolean containsSource(ImmutableSource _sour);

  /**
   * Updates all the internal maps after a test of the parameter type.
   * So, an Actor is also added to the Sources internal map.
   * Invoke super.addEorBaseObject(_targ). Overridden by the subclass.
   * @param _targ to add to the view
   * @exception CreateError if the target is already known by the view,
   * if the name is null, or the object name is null or empty,
   * or if the name is already known by the view.
   */
  public void addEorTarget(ImmutableTarget _targ) throws CreateError;

  /**
   * Updates all the internal maps after a test of the parameter type.
   * So, an Actor is also added to the Targets internal map.
   * Invoke super.addEorBaseObject(_sour). Overridden by the subclass.
   * @param _sour to add to the view
   * @exception CreateError if the source is already known by the view,
   * if the name is null, or the object name is null or empty,
   * or if the name is already known by the view.
   */
  public void addEorSource(ImmutableSource _sour) throws CreateError;

  /**
   * Updates all the internal maps after a test of the parameter type.
   * Invoke super.removeEorBaseObject(_targ). Overridden by the subclass.
   * @param _targ to remove in the view
   * @exception CreateError if the target is not known by the view, or
   * if the object name is empty.
   */
  public void removeEorTarget(ImmutableTarget _targ) throws CreateError;

  /**
   * Updates all the internal maps after a test of the parameter type.
   * Invoke super.removeEorBaseObject(_sour). Overridden by the subclass.
   * @param _sour to remove in the view
   * @exception CreateError if the source is null or not known by the view, or
   * if the object name is empty.
   */
  public void removeEorSource(ImmutableSource _sour) throws CreateError;

  /**
   * Gets all the access targets in this view, and this covers simple resources,
   * actors, directories, executables, accounts, groups and virtual folders.
   * @return unmodifiableSortedMap of the view targets
   */
  public SortedMap<ImmutableName, ImmutableTarget> getEorM_Targets();

  /**
   * Gets the targets in this view which are not eligible parties (actors, accounts, groups)
   * nor virtual folders.
   * @return unmodifiableSortedMap of the view targets
   */
  public SortedMap<ImmutableName, ImmutableTarget> getEorM_NoSourceTargets();

  /**
   * Gets all the sources  in this view, and this covers the actors, accounts,
   * groups and virtual folders.
   * @return unmodifiableSortedMap of the view sources
   */
  public SortedMap<ImmutableName, ImmutableSource> getEorM_Sources();

  /**
  * This method creates, registers and saves a view copy with another name. 
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy).
  * Calls copy(), saves the new instance and calls finalizeForCopy().
  * Overridden by the subclass.
  * @param _name of the copy. It is copied without test.
  * @exception BaseError if _name is null or already known in BaseManagerImpl,
  * or if _view is null or not known.
  * @return the copied view
  */
  public EPRViewInBase copyAndSave(ImmutableName _name) throws BaseError;
  
} // EPRViewInBase