//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.List;
import java.util.Map;
import javax.swing.Icon;
import java.beans.PropertyChangeListener;

import ARoad0.gBase.*;

/**
 * This interface is responsible for modeling an access control system (ACS).
 * It registers the controlled (internal and external) resources,
 * the eligible parties (accounts, groups, actors), and the aclEntries
 * of the ACS.
 * It registers the other ACS which control the ACS.
 * It cuts or restores in its objects, the external references to other ACS
 * when this ACS is closed or open.
 * @version 0.7.0
 */
public interface ImmutableACS extends ACSCatalog, ACSRun, ImmutableCommonBase, Cloneable {
   
  /**
   * Get the mandatory parent name for the ACS in the IS nametree.
   * @return the parent name. May be null or empty.
   */
  public ImmutableName getParentName(); 

   /**
      * True if an other open ACS controls this ACS,
      * that is has an external AclEntry or Linked Privilege
      * for a resource or a virtual folder from this ACS.
      * If an another ACS has an external AclEntry or Privilege
      * which is connected to an eligible party from this ACS,
      * this ACS is NOT controlled. The concept of control is for targets only.
      * @return boolean true if the ACS is controlled by one or several ACS
      */
    public boolean getIsControlled();

    /**
     * @return true if, through its open external AclEntries or Linked Privileges,
     * the ACS is a controller of the target of another ACS, or if
     * it delivers some rights to an EligibleParty in another ACS.
     * Closed AclEntries and Privileges are not taken in account.
     */
    public boolean getIsExternalOriented();

    /**
     * Gets the presence of actors which are bridge targets.
     * @return boolean true if, through bridge relations,
     * the Acs has bridge targets.
     */
    public boolean getContainsBridgeTarget();

    /**
     * Gets the presence of actors which are bridge sources.
     * @return boolean true if, through bridge relations,
     * the Acs has bridge sources.
     */
    public boolean getContainsBridgeSource();
    
  /**
   * @return ImmutableName: root of the recorded own resources in the ACS base.
   */
    public ImmutableName getOwnResourcesRootName();

  /**
   * The current root is the first non-finalized group
   * without parent which has been created.
   * return name of the root for the recorded own groups in the ACS base.
   * @return ImmutableName - May be equal to VectorNameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnGroupsRootName();
  
  /**
   * The current root is the first non-finalized virtual folder for resources
   * without parent which has been created.
   * return name of the root for the recorded virtual folders in the ACS base.
   * @return ImmutableName - May be equal to VectorNameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnResourceFoldersRootName();

  /**
   * The current root is the first non-finalized virtual folder for
   * eligible parties, without parent which has been created.
   * return name of the root for the recorded virtual folders in the ACS base.
   * @return ImmutableName - May be equal to VectorNameImpl.EMPTY_INSTANCE.
   */
  public ImmutableName getOwnEPFoldersRootName();  
  
  /**
   * @return true if some specialized rights have been defined.
   */
  public boolean getHaveSpecializedRights();
  
  /**
   * Gets the hub property for the ACS node in the IS nametree.
   * @since 0.7.0
   * @return the is_hub property
   */
  public boolean getHubProperty(); 
  
  /**
   * Gets the visible space for the ACS in the IS nametree.
   * @since 0.7.0
   * @return the visible space property
   */
  public String getVisibleSpace();

  /**
   * Gets the IS nametree property.
   * @since 0.7.0
   * @return true if the acs is tied to its parent acs.
   */
  public boolean getTiedToParentACS();
  
  /**
   * A tied child ACS has to be opened when this instance is opened.
   * Like any child ACS, it has to be closed when this instance is closed, and 
   * it may be closed alone.
   * @return the names of the child ACS which are tied up to this instance
   */
  public List<ImmutableName> getL_TiedChildACS();

  /**
   * Gets the subACS names and their mode (for Resources or for GroupIDs) and the
   * root names for groups trees, as nodes to display in the explorer.
   * A subACS is always open with its ACS.
   * In the returned value, the key is the subACS name or a group root name, in one word,
   * and the value is an array which may contain one or two items among the following values: <P>
   * <ResourceSubACS>, <GroupSubACS> and <GroupTree>. <P>
   * For instance, the value 'yyy' for the key <GroupTree>,
   * means that a group root node '<yyy_tree>' has to be displayed
   * directly under the GroupIDs node, in the explorer. For the two other keys,
   * a value 'xxxx' means that a node '<xxxx_subacs>' has to be displayed directly
   * under the Resources and the AclEntries nodes, or under the GroupIDs node.
   * @return the map of the subACSs and the groups tree roots. May be empty but never null.
   */
  public Map<String, String[]> getM_SubAndGroupPolicy();

  /**
   * Gets the map which says if the ACS or a subACS belongs to a Composite ACS
   * in the IS Name Tree. <P>
   * In the returned value, the key is the subACS name, in one word, 
   * or the last component of the ACS name, and the value is the Composite name.
   * @return the map of the composite relations. Never null and may be empty.
   */
  public Map<String, ImmutableName> getM_BelongsToComposite();

  /**
   * For a parent ACS, gets the list of the inner ACSObjects having with an opened child ACS,
   * a relation among <ForChildAcsOpeningAndClosing>, <ForChildAcsOpeningNoClosing>
   * and <ForChildAcsIntegrity>.
   * These relations set that the deleting or the updating of the ACSObject will
   * stop the child ACS or will modify its ability to run correctly.
   * @see #getL_RootsForOpening
   * @see #getL_RootsForClosing
   * @see #getL_RootsForIntegrity
   * @return the map of the inner ACSObjects which support at least one child ACS.
   * The child ACS names are the keys, and the values are arrays of root names
   * from this ACS. Never null and may be empty.
   */
  public Map<String, ImmutableName[]> getM_InnerRootsForChildACS();

  /**
   * Gets for a child ACS, the list of associations to roots in the parent ACSs. <P>
   * This relation set that the root parent ACSObject, or one of its property value,
   * has to be there when the child ACS is opened, to ensure its ability to exist,
   * and has to be closed or updated (for a property value) at its closing.
   * @see AssociateUtilityImpl
   * @return The map key has the format: <P>
   * - ACS.Name.<Value>.XXX: a parent ACS name having the value XXX. <P>
   * The map value is a sequence of commands which are executed by AssociateUtilityImpl.
   */
  public Map<String, String[]> getM_RootCommandsForOpeningAndClosing();

  /**
   * Gets for a child ACS, the list of associations to roots in the parent ACS which
   * have a root relation with this child ACS. <P>
   * This relation set that the root parent ACSObject, or one of its property value,
   * has to be there when the child ACS is opened, to ensure its ability to exist,
   * but there is NO constraint to free the associations at its closing.
   * @see AssociateUtilityImpl
   * @return The map key has the format: <P>
   * - ACS.Name.<Value>.XXX: a parent ACS name having the value XXX. <P>
   * The map value is a sequence of commands which are executed by AssociateUtilityImpl.
   */
  public Map<String, String[]> getM_RootCommandsForOpeningNoClosing();
  
  /**
   * Get for a child ACS, the list of associations to roots in the parent ACS which
   * have a root relation with this child ACS. <P>
   * This relation set that the root parent ACSObject, or one of its property value, 
   * has to be there when the child ACS is opened, to ensure its ability to work well, 
   * in other words, following its specifications. 
   * @see AssociateUtilityImpl
   * @return The map key has the format: <P>
   * - ACS.Name.<Value>.XXX: a parent ACS name having the value XXX. <P>
   * The map value is a sequence of commands which are executed by AssociateUtilityImpl.
   */
  public Map<String, String[]> getM_RootCommandsForIntegrity();


  /**
   * Gets the list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root has to be in the parent ACS at the opening of this ACS to ensure its
   * working.
   * There is no common roots with the values in getL_RootsForIntegrity(), but
   * some common values may exist with getL_RootsForClosing().
   * This list is generated by AssociateUtilityImpl from the commands in
   * getM_RootCommandsForOpeningAndClosing() and getM_RootCommandsForOpeningNoClosing().
   * @return the root names.
   */
  public List<ImmutableName> getL_RootsForOpening();

  /**
   * Gets the list of the ACSObject roots of this ACS in an opened parent ACS, where
   * each root has to be deleted in the parent ACS at the closing of this ACS.
   * There is no common roots with the values in getL_RootsForIntegrity(), but
   * all the values should exist in getL_RootsForOpening().
   * This list is generated by AssociateUtilityImpl from the commands in
   * getM_RootCommandsForOpeningAndClosing().
   * @return the root names.
   */
  public List<ImmutableName> getL_RootsForClosing();

  /**
   * Gets the list of the ACSObject roots of this ACS in an opened parent ACS, 
   * where each root in the parent ACS has some property updatings which may weaken
   * the integrity of this ACS. There is no common roots with the values in
   * getL_RootsForOpening() and getL_RootsForClosing().
   * This list is generated by AssociateUtilityImpl from the commands in
   * getM_RootCommandsForIntegrity().
   * @return the root names. May be empty but never null.
   */
  public List<ImmutableName> getL_RootsForIntegrity();

  /**
   * @return true if the ACS has some roots in its parent ACS
   */
  public boolean getContainsRootActions();

  /**
   * Gets the reference in this ACS for an external or internal soft alias,
   * by the way of a specific algorithm. The internal reference may have several aliases.
   * This method returns null in this ACSImpl class.
   * It has to be overridden in the AcsAddons
   * where an external or internal soft alias is managed, and it is then
   * responsible to update the soft alias and the soft reference properties.
   * This means that a call to this method is the only mean to update these bound
   * properties. If the soft aliasing links may vary, it is the responsivility
   * of the AcsAddon to call this method at the right moment. This is true
   * typically when an object is deleted, or when there is a change into
   * the parameters defining the soft aliasing rules.
   * This method is not declared in the ACS
   * BeanInfo since it is not based on a property.
   * @param _alias ACSObject from another ACS
   * @return the internal ACSObject which is the reference of the argument.
   * May be null.
   */
  public ACSObject getReferenceOfSoftAlias(ACSObject _alias);
  
  /**
   * @return the icon associated to the ACS. May be null.
   */
  public Icon getEorIcon();

  /**
   * @return the name of the icon associated to the ACS. May be null.
   */
  public String getIconName();

  /**
   * To detect an AcsAddon, this method tests the chain 'ARoad0.AcsAddon.' in the ACS class name.
   */
  public boolean getIsFromAcsAddon();

  /**
   * Tests a listener and returns true if it is registered for the listening of
   * a given bound property.
   * @param _listener to test
   * @param _propertyName name of the bound property to listen
   * @return true if _listener is registered
   */
  public boolean isPropertyChangeListener(
         String _propertyName, PropertyChangeListener _listener);

}
