//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.List;
import ARoad0.ErrB.CreateError;

import java.beans.PropertyChangeListener;

  /**
 * This important interface is responsible for modeling an access control entry which manages
 * some access rights or some access restrictions from an eligible party to a resource.
 * An AclEntry is external if its eligible party or its resource is from an ACS which
 * is not the ACS of the AclEntry. It is possible to have a third ACS for the other end.
 * It is not possible to have the same ACS for the two ends and another ACS for the ACL.
 * For an internal ACL, the ACL ACS is the ACS of the two ends.
 * An AclEntry is conditional when it uses some condition group(s). There are two types of
 * conditional AclEntry, the simple-condition (or flexible) ACL and the right-condition ACL.
 * A conditional ACL is for a GroupIDMember or an Actor (new in 0.7.1) as source.
 * An external ACL may be conditional. Then, it cannot have more than one condition group,
 * and this group has to belong to the ACS of one of the two ends (that is not always the ACL ACS).
 * A resource in an ACS addon may inherit the ACL of its (in)direct parent(s),
 * following the inheritance rules of the ACS addon. <P>
 * The AclEntry is an alternative to the Privilege, but it allows the change of its rights
 * while the Privilege cannot do it.
 * The conditional AclEntry is an alternative to the LINKED_IF_ALL_SOURCES privilege,
 * but it is focused to deliver rights to one source which has to be member of the
 * condition groups. <P>
 * The form of the detailled name for an internal ACL is: <P>
 * - incomplete Eligible Party name >> incomplete Resource name ||
 * grant OR deny || first condition group last component. <P>
 * For an external ACL, the form of the detailled name is: <P>
 * - ACS: ACS_name | (full or incomplete) Eligible Party name >>
 * (full or incomplete) Resource name || grant OR deny ||
 * &lt;Source ACS group&gt; OR &lt;Target ACS group&gt; ||
 * first condition group last component. <P>
 * An Eligible Party or Resource name is incomplete if it belongs to the AclEntry ACS.
 * The condition group is put in the name or the detailled name only for
 * a flexible conditional AclEntry. '&lt;XXXX ACS group&gt;' is used
 * only for an external ACL, when the condition group belongs to the XXXX ACS.
 * <P>
 * It is a javabean with the following bound properties:
 * 'Comment', 'AclRights', 'ConditionSourceGroups'.
 * @see ARoad0.gBaseInterface.EligibleParty
 * @see ARoad0.gBaseInterface.Resource
 * @version 0.7.0
 */
public interface ImmutableAclEntry extends Cloneable, ACSObject, DetailledName, Comparable {

  /**
  * Gets the access source displayed in the beamer.
  * The format is:
  * 'full ep name as string'+'(ACL)'.
  * @return String is the source name. May be null.
  */
  public String getSource();

  /**
  * Gets the access target displayed in the beamer.
  * The format is:
  * 'full resource name as string'+'(ACL)'.
  * @return String is the target name. May be null.
  */
  public String getTarget();
    
  /**
  * @return boolean true if equal to new AclEntryImpl()
  */
  public boolean isEmpty();    
 
   /**
   * @return true if the source or the target is from another ACS
   */
  //public boolean isExternalAclEntry();

  /**
  * @return an external object reference to the resource that applies this AclEntry,
  *  without backup copy. Quite low, since it uses resourceACS.getEorM_Resources().
  */
  public ImmutableResource getEorResource();

  /*
   * @return ImmutableEligibleParty as external object reference
   * of the Eligible Party
   */
  public ImmutableEligibleParty getEorEP();

  /**
   * Gets the effective rights for the eligible party upon the resource.
   * If there is no condition group, the effective rights are the internal rights
   * as returned by getNoConditionalRight().
   * If there are conditional groups and if the eligible party is a group member,
   * the eligible party has to be a direct or indirect member of each conditional group,
   * to return the internal rights. Otherwise, the condition rights are not workable,
   * and an empty array is returned. If the eligible party is not a group member,
   * the condition group(s) are forbbiden.
   * @return String[] of rights as READ, CONNECT, DELETE...
   *  for the couple (resource, eligible party). 
   *  Never null. May be an empty array.
   */
   public  StringRight[] getL_Rights();

  /**
   * Gets the list of the internal conditional or no-conditional rights
   * for the eligible party upon the resource, whatever the compliance to
   * the conditions if there are some conditional groups.
   * @return String[] of rights as READ, CONNECT, DELETE...
   *  for the couple (resource, eligible party). 
   *  Never null. May be an empty array .
   */
    public StringRight[] getNoConditionalRight();
   
   /**
   * @return ImmutableACS of this AclEntry, as an External Object Reference
   */
    public ImmutableACS getEorACS();

   /**
   * @return ImmutableACS of the AclEntry resource, as an External Object Reference.
   */
    public ImmutableACS getEorResACS();

  /**
   * @return true if the right allows access to the resource, false if it forbids access.
   */
  public  boolean getPositiveRight();

   /**
    * Caution: does not return a clone.
    * @return ImmutableName of the AclEntry EligibleParty. May be null.
    */
    public ImmutableName getEpName();

   /**
    * Caution: does not return a clone.
    * @return ImmutableName of the AclEntry resource.
    *   May be NameImpl.EMPTY_INSTANCE.
    */
    public ImmutableName getResName();

   /**
    * To use when it is possible to replace getEorResource().
    * @return the AclEntry resource type.
    */
    public String getResType();

   /**
    * Caution: does not return a clone.
    * @return ImmutableName of the AclEntry EligibleParty ACS. May be null.
    */
    public ImmutableName getNameOfEpACS();

   /**
    * Caution: does not return a clone.
    * @return ImmutableName of the AclEntry resource ACS. May be null.
    */
    public ImmutableName getNameOfResACS();
  
  /**
   * To activate the ACL rights in a conditional AclEntry, the EP,
   * if it is a  groupmember, has to be member of each conditional GroupID.
   * If the EP is not a groupmember, the effective rights are null.
   * @return list of the conditional groups. May be empty, not null.
   */
  public List<ImmutableGroupID> getL_ConditionSourceGroups();
    
    /**
     * Gets true for a right-defined conditional AclEntry.
     * @return true if the first right is a conditional right
     */
  public boolean isRightDefinedConditionalAclEntry();

    /**
     * Gets true for a right-defined or a flexible conditional AclEntry.
     * @return true if the condition groups list is not null
     */
  public boolean isConditionalAclEntry();
 
   /**
    * An external ACL has the source or the target from another ACS than the ACL ACS.
    * It may be a conditional ACL.
   * @return true if the source or the target is from another ACS
   */
  public boolean isExternalAclEntry();

    /**
     * Gets false only for a conditional AclEntry where the GroupIDMember source is not
     * (in)direct member of a condition group. A conditional ACL having an Actor as
     * source has always enabled rights (new in 0.7.1).
     * @return true if the rights are workable
     */
  public boolean getEnabledRights();

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName);

  /**
   * Used by ACS.closeAclEntriesTo() when the links to other ACS have to 
   * be cut and cloned.
   * The ACS, the EP and the resource ACS are null in the returned AclEntry,
   * and getEorResource() returns null. The properties are protected, so
   * may be accessed by the package.
   * Caution: the clone is not registered in its ACS, its resource or its EP,
   * but equals() returns true when it compares a clone and its source.
   * @return Object without external references
   * (ACS, resource ACS and EligibleParty) and copy of the other properties,
   * including the condition groups.
   */
    public Object clone();
  
  /**
   * Copy the AclEntry for a resource of the same ACS.
   * The AclEntry contains the same properties than this instance, and
   * it is registered in its ACS, resource and EP.
   * @param _newRes is a resource belonging to the same ACS
   * @return an AclEntry with a new resource and a copy of the other
   *   properties.
   * @exception InternalError if _newRes is null.
   * @exception CreateError if _ep is already recorded in this resource
   *  with the same ACS and sens, or thrown by the ACS.newEorAclEntry().
   */
  public AclEntry copy(Resource _newRes) throws CreateError;

  /**
  * @param _obj is an Object to compare
  * @return true if the variables are equals - not only identicals -
  * and are from the _same_ class implementing AclEntry interface.
  */
    public boolean equals(Object _obj);

  /**
  * @return the hashcode value for this object
  */
   public int hashCode();

/**
 * @return long is the uid used for serialization of the instance.
 */
   public long getSerialVersionUID();

/**
* The String presentation of the AclEntry
* @return String
*/
  public String toString();

} // ImmutableAclEntry
