//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

/**
 * This immutable interface is responsible for modeling
 * an acting entity which may access to resources controlled by
 * the ACS, and which may play some special roles, as a bridge, between
 * an another actor and some resources.
 * Exemples of such actors are servers or executables.
 * The getActorAlias() method in 0.6 version has been deleted, and it is
 * replaced by ImmutableBasic.getEorReferenceAsInternalHardAlias() for the AcsAddons
 * only.
 * This interface also models the operating system process which has a current
 * account and group, and which belongs to several groups.
 */
public interface ImmutableActor extends ImmutableResource, ImmutableEligibleParty {

 /**
 * By default, is the value from the resource.
 * @return ImmutableUserID the current userID, that may be changed
 * by itself, and different to the actor userID attached to any resource.
 * For instance, when an executable is launched from a userID, its current
 * userID often becames the launcher userID. May be null.
 */
  public ImmutableUserID getEorCurrentUserID();

  /**
  * By default, is the value from the resource.
  * @return ImmutableGroupID used currently by the actor, that may be changed
  * by itself, and different to the actor groupID attached to any resource.
  * For instance, when an executable is launched from a groupID, its current
  * groupID often becames the launcher groupID. May be null.
 */
  public ImmutableGroupID getEorCurrentGroupID();

  /**
  * A secondary groupID adds new accesses to the actor as rights user.
  * A secondary groupID is independant to the primary groupID of the actor
  * as resource, but this primary group may be also a seocndary group.
  * @return secondary groupIDs under which this actor runs a rights user.
  *      May be empty, but not null.
  */
  public ImmutableGroupID[] getEorL_SecondaryGroup();

  /**
  * @return true if _grp is a secondary groupID which have this actor
  * as membership.
  */
  public boolean isMemberToSecondaryGroup(ImmutableGroupID _grp);

  /**
   * @return ImmutableActor[] as an External Object Reference.
   *  May be empty, but not null.
   */
  //public ImmutableActor[] getEorL_isAGateFor();

  /**
   * A bridge is an unidirectionnal connexion between two ACS through 
   * the mean of two actors, one in each ACS.
   * @return ImmutableActor[] external actors for which this actor is a bridge target. 
   *    May be empty, but not null.
   */
  public ImmutableActor[] getEorL_bridgeSources();

  /**
   * @return true if this actor is a bridge source,
   *  created through a specific constructor, and updated when the
   *  external actor is closed or deleted.
   */
  public boolean getIsACSBridgeSource();

  /**
   * Get the another-ACS actor for which this resource is a bridge.
   * @return a reference on the actor. Is null if
   *   this instance is not a bridge, if the target ACS is not opened.
   */
  public ImmutableActor getEorACSBridgeTarget();

  /**
   * @return true if this Actor may be a proxy in the access paths, that is,
   *    as rights user, it gets some rights to the other
   *    rights users which have an 'execute' right on it
   */
  public boolean getIsRightProxy();
  
} // ImmutableActor

