//Title:        Access Road
//Version:      0.6.0
//Copyright:    2002, 2004, 2007, 2008, 2009 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

/**
 * This immutable interface is responsible for modeling a GroupIDMember that owns a set of members,
 * and which has Account/Group rights on Resources.
 * Typically, it handles a set of userIDs and GroupIDs.
 * The membership links between the groupID and each member are managed by
 * the GroupID instance.
 * Resources may also be registered as owned by a GroupID. Aa rights user,
 * an actor may also have a groupID as a current group, and it is then registered
 * in a third list. An actor may have several secondary groups as rights user,
 * in a fourth list.
 * These relations are managed by this interface.
 * @version 0.6.0
 */
public interface ImmutableGroupID extends ImmutableGroupIDMember {

  /**
  * Get the members of the GroupID.
  * @return members of the groupID
  *  as an External Object Reference. This is an ordered list
  *  without null member, and without empty member after the first update.
  */
  public ImmutableGroupIDMember[] getEorL_Members();

  /**
  * Test if an eligible party is a direct member of this GroupID,
  * or if it has an alias which is a direct member
  * Synchronized on l_Members_.
  * @return true if _ep is a member of the GroupID.
  */
  public boolean containsAsMember(ImmutableGroupIDMember _ep);

  /**
  * Test if an eligible party is a direct member of this GroupID,
  * and do not test if it has an alias which is a direct member.
  * Synchronized on l_Members_.
  * @return true if _ep is a member of the GroupID.
  */
  public boolean containsAsMemberWithoutAliasing(ImmutableGroupIDMember _ep);

  /**
  * Gets the first member which contains the argument as member or indirect member.
  * Synchronized on l_Members_.
  * @param _ep to test. May be null.
  * @return the first member. May be null.
  */
  public GroupID getFirstIntermediateForIndirectMember(ImmutableGroupIDMember _ep);

  /**
  * Test if an eligible party is an indirect member of this GroupID,
  * or if it is an hard alias having a reference which is an indirect member.
  * Caution: if the argument is the alias of a GroupIDMember, the test is applied
  * only to the reference as potential member, and not to the argument.
  * Synchronized on l_Members_.
  * @return true if _ep is an indirect member of the GroupID.
  */
  public boolean containsAsIndirectMember(ImmutableGroupIDMember _ep);

  /**
  * Test if an eligible party is a indirect member of this GroupID,
  * and do not test if it has an alias which is an indirect member.
  * Synchronized on l_Members_.
  * @return true if _ep is an indirect member of the GroupID.
  */
  public boolean containsAsIndirectMemberWithoutAliasing(ImmutableGroupIDMember _ep);

  /**
  * @return the order of the groupID - on Unix/Linux : 16-bit group identifier (GID)
  */
  public  int getOrder();

  /**
  *  A currentGroupID or a secondary groupID is a property of an actor, which is
  *  a kind of resource. But these actors are NOT returned by this method.
  * @return array of the resources in the ACS for which this group is their primary groupID.
  */
  public Resource[] getEorL_GroupIDOwnResources();

   /**
    * A current actor runs under this groupID.
    * A current actor may be an own resource of this groupID or not.
   * @return array of the current actors attached to this groupID.
   *  May be empty but not null.
   */
   public Actor[] getEorL_GroupIDCurrentActors();

   /**
   *  Get the actors in the ACS that have, in the role of a running processes,
   *   this group as a secondary groupID.
   *  If this groupID is the main groupID  or the current groupID of an actor,
   *  it is NOT returned by this method.
   * @return array of the actors in the ACS that have this group as
   *  a secondary groupID. May be empty but not null.
   */
   public Actor[] getEorL_GroupIDCurrentActorsAsSecondaryGroup();

   /**
   * @return true if it owns the resource.
   */
   public boolean ownsResource(Resource _res);
}
