//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009 2010 ACCBEE EURL
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.List;

 /**
  * This interface is responsible for managing the names for all the database objects.
  * Examples of real name are "http://accbee.com", "c:\program files\access road",
  * "net/users/john". The separator character depends on the ACS. All the real names
  * are converted in Access Road names implemented by this class. The single separator
  * in all the Access Road names is ':: '.
  * Every ImmutableName instance must be a Name instance.
  * The name is an list of components in a java.util.ArrayList.
  * Each component is a string that may be null. Case must be respected.
  * component(int) and lastComponent() return only copies of the component, as
  * the others methods which return new objects, no linked to the current name.
  * compareTo() allows an immutable name to be a key in a sorted map.
  */
public interface ImmutableName extends Cloneable, Comparable {

  /**
   * @return true if the name has no component, or if all components
   * are null or equal to empty Strings.
   */
   public boolean isEmpty();

  /**
   * Get a copy of the current name less the last component.
   * @return a copy of the current name less than the last component.
   *    The context returned is not backed by the current name,
   *    so modify the context doesn't modify the current name. May be null.
   */
  public  Name context();

  /**
   * Get a copy of the current name less some last components.
   * @param _end is the _inclusive_ index of the last component to be returned
   * @return the context (from 0 to _end) of the current name in a new Name,
   *    this name if _end = (size of the name) - 1,
   *    this name.context() if _end = (size of the name) - 2,
   *    the first name component if _end = 0. The context is not backed
   *    by the first components of the current name. Return is null if
   *    the name is empty, or if _end is equal or superior to the name size.
   */
  public  Name context(int _end);
  /**
     * @param _start inclusive fromIndex
     * @param _end exclusive toIndex
     * @return an extract of the current name that is copied in a new Name. May be null.
     */
  public  Name subname(int _start, int _end);
   /**
     * @return List of the components (strings) that is backed by the list
     *  of the current name.
     *  Then, the returned name is modifiable but
     *  the only proprietary of the returned list is the current name.
     *  So, to avoid any bug, use the returned list as a transient object,
     *  for a limited time. Never delete it directly. May be null.
     */
  public List<String> components();
   /**
     * @return List of the components (strings) in the name. is immutable. may be null.
     */
  public  List immutableComponents();
   /**
     * @return int number of components in the name.
     */
  public  int noOfComponents();
   /**
    * No exception thrown.
   * @param _i index of the component. No limit on the value.
   * @return the component at _i in the name. May be null.
     */
  public  String component(int _i);
   /**
     * @return String the last component in the name (may be null)
     */
  public  String lastComponent();
   /**
     * @param _name Never null.
     * @return true if _name components are all there in the instance,
     *  but not always in the same order
     */
  public  boolean contains(ImmutableName _name);
    /**
     * @return a cloned and no-backed instance of mutable name
     * that may be cast as a Name or an ImmutableName. May be null.
     */
  public Object clone();
    /**
     * @param _obj the object to compare
     * @return true if the list of strings is equal - not only identical -
     *  even if the two objects are instances of the same class
     *  that implementes Name (and not only ImmutableName)
     */
  public boolean equals(Object _obj);
    /**
     * @return the hashcode value for this instance of Name. May be null.
     */
  public int hashCode();
    /**
     * @param o Object to compare must be an ImmutableName (true for Name)
     * @exception ClassCastException - if the specified object's type
     *  prevents it from being compared to this current Object
     * @return 0 is the names are equals, a negative integer if this Name object
     *  is lesser than the argument, and a positive integer if this Name object
     *  is greater than the argument.
     *  The comparison is based on the Unicode value of each character
     *  in each Name component (starting from the first one)
     *  and the number of components (negative return if all strings of this Name
     *  are equals but with less components, or if the argument is null).
     */
  public int compareTo(Object o) throws ClassCastException;

    /**
     * @return long serial version UID for serialization of the class
     */
  public long getSerialVersionUID();
}
