//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.gBase.*;

import java.util.List;

 /**
  * This class is responsible for managing specifically for IS nametrees,
  * a tree of ImmutableNames with specific links to the other nodes.
  * Most of the properties are directly managed by the calling class.<P>
  * A node in a ImmutableNameLinkTree may have to apply some inner properties or 
  * some constraints in its links to other nodes of the whole nametree:<P>
  * - 'Open' is a mandatory node property which has the value 'true' if the node
  * is active, like an open ACS, subACS or Area, which is currently opened in the program,
  * or if the node is a passive one which is the direct or indirect parent of
  * an opened active node. <P>        
  * - 'ExtendedExtendedVisibleSpace' is a mandatory node property which extends the parent children
  * (if ImmutableNameTree.isHubNode is true) and the ImmutableNameTree areas of the node, if any. This
  * property is an ImmutableName/String Map:<P>
  *      the possible String values are ImmutableNameLinkTree constants: <P>
  *          ALL_THE_NAMETREE_SPACE, <P>
  *          ONE_STEP_NODES_SPACE, TWO_STEPS_NODES_SPACE, THREE_STEPS_NODES_SPACE, <P>
  *          DIRECT_CHILDREN_SPACE, INDIRECT_CHILDREN_SPACE,  <P>
  *          PARENT_INDIRECT_CHILDREN_SPACE, PARENT_SPACE, <P>
  *          NO_VISIBLE_SPACE,  <P>
  *          the first value is the default one. <P>
  * - 'Alternates' is an optional node property, used only for the Alternate relation,
  * with an ImmutableName/ImmutableName[] Map:<P>
  *      mandatory key: node_name for a ground ACS or subACS; <P>
  *      the values are the names of the alternates for the key (limited to 4). <P>
  * - 'MandatoryPaths' is an optional node property, used for IS nametrees only,
  * with an ImmutableName/ImmutableName[] Map:<P>
  *      optional key: node_name is the beginning of the path; <P>
  *      value: the optional array of node names is the structurally-mandatory path 
  *      to go accross, where the structure is a one-item List, or a List<P>
  *      (first_node, second_node, beginning_node, third_node, fourth_node, beginning_node), <P>
  *      which indicates that there are two paths: from beginning_node to second_node, <P>
  *      and from beginning_node to third_node and fourth_node, <P>
  *      and (feature to be implemented) where all the nodes are in the visible space of the beginning node. <P>
  * - 'Links' is an optional node property, for IS nametrees,
  *      with an ImmutableName/String[] map. <P>
  *      Each property may have any format, and is defined by the user of this class.<P> 
  * This Links map is used for IS nametrees, and the ACS factory initializes
  * the ImmutableNameLinkTree links following some specific and complex uses.<P>
  * All the properties of a ImmutableNameLinkTree are mutually independant. In other words,
  * a node which is opened or closed may have any visible space, any mahdatory paths
  * and any links. Then, before the use of the data about a mandatory path or
  * a link, it is recommanded to see if the relevant node is opened or not.
  * See also 'isActiveNode' and 'isHubNode' properties in the super class ImmutableNameTree.
  * @version 0.7.0
  */
public interface ImmutableNameISTree extends ImmutableNameTree {
  
  /**
  * Get the state open/closed of a node.
  * @param _node is a node in the ImmutableNameLinkTree
  * @return true if the node is opened; always false if it is null or unknown.
  */   
  public boolean isOpen(ImmutableName _node);
  
  /**
  * Get the grounds for which the argument is an Alternate.
  * @param _alter is a node name or derived from a node name.
  * @return the list of grounds for _node as alternate. May be null.
  */   
  public List<ImmutableName> getGroundsOfAlternate(ImmutableName _alter);
  
  /**
  * Get the node property for the Alternate relation: <P>
  *      mandatory key: node_name; <P>
  *      the values are the names of the alternates for the key which is a ground,
  *         and (ISFactoryUtilityImpl.IS_ALTERNATE_OF+ground name) for the key which is an alternate. <P>
  * @param _node is a non-null node name of a ground
  * @return the array of alternates for _node. May be null but not empty.
  */
  public List<ImmutableName> getAlternates(ImmutableName _node);
  
  /**
  * Get the authorized visible spaces.
  * @return a copy of the authorized visible spaces for the NameLinkTrees.
  */   
  public String[] getExtendedVisibleSpaceConstants();  
    
  /**
  * Returns the nodes which are visible. The value is one of these constants:
  * ALL_THE_NAMETREE_SPACE/PARENT_INDIRECT_CHILDREN_SPACE/
  * TWO_STEPS_NODES_SPACE/ONLY_PARENT_SPACE_DIRECT_CHILDREN_SPACE/ONLY_DIRECT_CHILDREN_SPACE/
  * ONLY_PARENT_SPACE/NO_VISIBLE_SPACE<P>.
  * @param _node is a non-null node in the ImmutableNameLinkTree
  * @return the visible space from the node. May be null.
  */   
  public String getExtendedVisibleSpace(ImmutableName _node);
  
  /**
  * Set the nodes which are visible from a node. The value is one of these constants:
  * ALL_THE_NAMETREE_SPACE/PARENT_INDIRECT_CHILDREN_SPACE/
  * TWO_STEPS_NODES_SPACE/ONLY_PARENT_SPACE_DIRECT_CHILDREN_SPACE/ONLY_DIRECT_CHILDREN_SPACE/
  * ONLY_PARENT_SPACE/NO_VISIBLE_SPACE<P>
  * @param _node is a non-null node in the ImmutableNameLinkTree
  * @param _space is the visible space from the node. Never null.
  * @exception BaseError if _node is null, empty or unknown, if _space is not
  *   an authorized value.
  * @return void
  */   
  //public void setExtendedVisibleSpace(ImmutableName _node, String _space) throws BaseError;
  
  /**
  * Returns the mandatory paths from a node to access to other nodes, if rights allow it.
  * The returned value is a list of node name arrays, which are 
  * the mandatory paths to go accross. Exemples of the array structure is an one-item List, or a List <P>
  *      (first_node, second_node, beginning_node, third_node, fourth_node, beginning_node), <P>
  *      which indicates that there are two paths: from beginning_node to second_node, <P>
  *      and from beginning_node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * @param _node is a node in the ImmutableNameLinkTree
  * @return the list of arrays for mandatory paths copies. May be null.
  */   
  public List<ImmutableName[]> getMandatoryPaths(ImmutableName _node);
  
  /**
  * Test a path as mandatory path from a node to access to other nodes. 
  * The path is described as an array of node names which are the mandatory paths
  * to go accross. An exemple of the array structure is: <P>
  *      (first_node||second_node||beginning_node||third_node||fourth_node||beginning_node), <P>
  *      which indicates that there are two paths: from beginning_node to second_node, <P>
  *      and from beginning_node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * @param _node is a non-null node in the ImmutableNameLinkTree
  * @param _l_path is the array of the path to test, without _node. Never null.
  * @return true if _l_path is a mandatory path of _node
  */   
  public boolean isMandatoryPath(ImmutableName _node, ImmutableName[] _l_path);   
  
  /**
  * Set the mandatory paths from a node to access to other nodes. 
  * These paths may be not active if the proper inter-node rights are not there.
  * The paths are in an array of node names which are the mandatory paths to go accross. <P>
  * An exemple of the array structure is: <P>
  *      (first_node||second_node||beginning_node||third_node||fourth_node), <P>
  *      which indicates that there are two paths: from beginning_node to second_node, <P>
  *      and from beginning_node to third_node and fourth_node, <P>
  *      and where all the nodes are in the visible space of the beginning node <P>
  * @param _node is a non-null node in the ImmutableNameLinkTree
  * @param _l_paths is the array of mandatory paths. Never null.
  * @exception BaseError if _node is null, empty or unknown, if _l_paths is null or 
  *     is not ending with _node, or if its length is greater than 20,
  *     if the names are not in the _node visible space.
  * @return void
  */   
  //public void setMandatoryPaths(ImmutableName _node, ImmutableName[] _l_paths)
    //throws BaseError;
   
  /**
  * Test a special link for a node.
  * @param _node is a non-null node in the ImmutableNameLinkTree
  * @param _link is a link to test for the node.
  * @return true if _link is a link of _node
  */   
  public boolean isLinked(ImmutableName _node, String _link);
  
  /**
  * Gets the optional links for a name, which depends on the ImmutableNameISTree use.
  * @param _name is a node/area name or the association of two node/area names
  * in the ImmutableNameISTree
  * @return the links. May be a 0-length array. Never null.
  */   
  public String[] getLinks(ImmutableName _name);
   
  /**
  * Set the special links of the node, which depends on the ImmutableNameLinkTree use.
  * It is an optional node property. There is no control on the _l_links values.
  * See this class comments about the use of this method by ACSFactory.
  * @param _node is a non-null node in the ImmutableNameLinkTree
  * @param _l_links defines the nature of the links for the node. Never null.
  * @exception BaseError if _node is null, empty or unknown, or _l_links null.
  * @return void
  */   
  //public void setLinks(ImmutableName _node, String[] _l_links) throws BaseError;

  /**
  * Get if a node is the direct or indirect parent of at least one node
  * having the type ISImpl.ACS_NODE.
  * @param _node is a node in the ImmutableNameISTree
  * @return true if the node is a ACS_NODE parent.
  */   
  public boolean isParentOfAcsNode(ImmutableName _node);
  
  /**
  * Gets if a node is the direct or indirect child of at least one node
  * having the type ISImpl.ACS_NODE. The argument may be an area.
  * @param _node is a node or an area in the ImmutableNameISTree
  * @return true if the argument is a child
  */
  public boolean isChildOfAcsNode(ImmutableName _node);
  
} // ImmutableNameLinkTree class