//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.BaseError;
import java.util.Set;
import java.util.List;
import java.util.Map;
import javax.swing.tree.TreePath;

 /**
  * This class is responsible for managing a tree of ImmutableNames
  * for all the base objects. The nodes are ImmutableNames, but there is no
  * constraint between a node name and its parent node name (that is,
  * the node name may integer the parent name or not).
  * The links between nodes are parent-child relations. <P>
  * This interface uses the relevant methods of javax.swing.tree.DefaultTreeModel.
  * There is always a non-null name as root.
  * There is no order in the children list of a node.
  * An name cannot be a node twice in the same nametree. <P>
  * This class is used by IS, component and ACS trees.
  * It is possible to get subtrees, that is a new nametree with all the
  * indirect children.
  * It is also possible to get areas. Area is defined as a group of
  * connected nodes in the nametree, with a single root,
  * which is used for mapping components to a virtual structure like an IS domain. <P>
  * @version 0.7.0
  */
public interface ImmutableNameTree extends Cloneable {

  /**
   * @return true if the nametree has null properties, or has only a root as node.
   */
   public boolean isEmpty();
   
  /**
  * Get the root of the nametree
  * @return the nametree root
  */
  public ImmutableName getRoot(); 
  
  /**
   * Get a subtree from the nametree, with all the indirect children.
   * This method is a good test of the nametree code bugs, and throws
   * an Error when unconsistent data is detected.
   * Caution: change on the returned nametree does not change this nametree.
   * @param _root this non-null node of the nametree is the subtree root,
   *            which is not the nametree root.
   * @exception BaseError if the root is the nametree root, null, empty,
   *   does not belong to the nametree, or if non-consistent data is detected.
   * @return a deep copy of the subtree where _root is the new tree root.
   */
  public ImmutableNameTree getSubTree(ImmutableName _root) throws BaseError;
  
  /**
  * Get the path of nodes between two nodes in the nametree,
  * if one of the two nodes is an indirect child of the other one.
  * @param _startingNode is a nametree node
  * @param _endingNode is a nametree node which is at a lower or a greater level
  *   than _startingNode
  * @exception BaseError if one of the two arguments does not belong to the nametree,
  *   or if the nametree data is not consistent.
  * @return the TreePath of node names, including the two arguments, and the first one
  *   (index 0) is the node in the lower nametree level. Null if there is no tree path.
  *   Contains only _startingNode if the two nodes are equals.
  */  
  public TreePath getTreePath(ImmutableName _startingNode, ImmutableName _endingNode)
    throws BaseError;
  
  /**
  * Get the set of all the nametree nodes, including the root and excluding the areas.
  * @return node names. Never null.
  */
  public Set<ImmutableName> nodes();
  
  /**
  * Get the set of all the nametree areas, if any.
  * @return area names. Never null.
  */
  public Set<ImmutableName> getL_AreaNames();
  
  /**
   * Get all the areas of this nametree, with their node names.
   * @return a Map of node Names. The key is the area name. May be empty.
   */
  public Map getM_AreaNodes();
  
  /**
  * @return the number of nodes in the nametree, including the root and excluding the areas.
  */
  public int nbOfNodesAndAreas();
    
  /**
  * @return the number of levels in the nametree, including the root.
  */
  public int nbOfLevels();
  
  /**
  * Test a name as node of the nametree, including the root.
  * @param _name to test
  * @return true if _name is a nametree node
  */
  public boolean isNode(ImmutableName _name);

  /**
  * Test a name as area of the nametree.
  * @param _name to test
  * @return true if _name is a nametree area
  */
  public boolean isArea(ImmutableName _name);
  
  /**
  * Get the nodes belonging to an area of the nametree.
  * @param _name of an area of the nametree
  * @return list for the area node names. May be empty.
  */
  public List<ImmutableName> getL_NodesOfArea(ImmutableName _name);
  
  /**
  * Get the areas of the nametree, the node belongs to.
  * @param _name of a node of the nametree
  * @return a deep copy of the list for the node areas. May be empty.
  */
  public List<ImmutableName> getL_AreasOfNode(ImmutableName _name); 
  
  /**
  * Get the level of a node in the nametree. The root is at the level 0.
  * @param _node has to be a nametree node
  * @return the level of _node in the nametree, or -1 if not in the tree
  */
  public int getLevel(ImmutableName _node);
  
  /**
  * Get the parent of a nametree node.
  * @param _child has to be a nametree node name, not an area name
  * @return the copy of parent in the nametree, or null if
  *  _child is null, empty, an area name, the root or it is not a nametree node
  */
  public ImmutableName getParent(ImmutableName _child);
  
  /**
  * Get one of the children in the nametree
  * @param _parent is a parent in the nametree
  * @param _index in the parent's child array
  * @return the child of _parent at _index, or null if
  *   _parent is not a nametree node, not a parent, or if _index is negative
  */
  public ImmutableName getChild(ImmutableName _parent, int _index);

  /**
  * Get all the direct children of a node in the nametree.
  * The order of the returned childen is the tree index order.
  * @see #getL_AllChildren
  * @param _parent is a node in the nametree
  * @return clones of ImmutableNames. Null if _parent is not a node
  *    of the nametree. Empty if _parent is a leaf.
  */
  public List<ImmutableName> getL_DirectChildren(ImmutableName _parent);

  /**
  * Get all the direct and indirect children of a node in the nametree.
  * The order of the returned childen is the increasing order in levels, so
  * the direct children are the first items in the returned list.
  * @see #getL_DirectChildren
  * @param _parent is a node in the nametree
  * @return clones of ImmutableNames. Null if _parent is not a node
  *    of the nametree. Empty if _parent is a leaf.
  */
  public List<ImmutableName> getL_AllChildren(ImmutableName _parent);
  
  /**
   * Get the type of the node.
   * @exception BaseError if _node is null or an empty name,
   *   or does not belong to the nametree.
   * @return the node type.
   */
   public String getType(ImmutableName _node) throws BaseError;
 
  /**
   * Test if a node of the nametree is set as a 'hub' for its children.
   * A node may be a hub, which means that for this node,
   * any direct child may have exchanges with any other direct child. 
   * For a non-hub node, exchanges between its children are not allowed.
   * This concept is extended in NameLinkTree by the visible space property. 
   * @exception BaseError if _node is null or an empty name,
   *   or does not belong to the nametree.
   * @return true if the node is a hub.
   */
   public boolean isHubNode(ImmutableName _node) throws BaseError; 
   
  /**
  * Add a child to a node in the nametree, with the default false values
  * for is_passive and is_hub.
  * @param _parent is a node in the nametree
  * @param _child is the new child, which is put at the end
  *   of the children array
  * @exception BaseError if _parent is not a node of this nametree,
  *  if _child is null, or is already a nametree node
  * @return void
  */
  //public void addChild(ImmutableName _parent, ImmutableName _child)
    //throws BaseError;  
  
  /**
  * Add a child to a node in the nametree
  * @param _parent is a node in the nametree
  * @param _child is the new child, which is put at the end
  *   of the children array
  * @param _is_passive is true if the root is empty
  * @param _is_hub is true if the root is a node hub
  * @exception BaseError if _parent is not a node,
  *   if_child is null, or is already a nametree node
  * @return void
  */
  //public void addChild(ImmutableName _parent, ImmutableName _child,
    //boolean _is_passive, boolean _is_hub) throws BaseError;
  
  /**
  * Remove a node from the nametree, and the subtree which is
  * associated to the node.
  * @param _node is a node of the tree
  * @exception BaseError if _node is null
  * @return the number of removed nodes, including _node
  */
  //public int removeNodeAndSubTree(ImmutableName _node) throws BaseError;
  
  /**
  * Get the number of children for a node in the nametree,
  * in reading its proper children count.
  * Generally speaking, this method is faster than getChildren() which
  * searches children in the level above the parent level.
  * @param _parent is a node in the nametree
  * @exception BaseError if the parent is null or does not belong to the nametree
  * @return the number of children for the parent
  */
  public int getChildCount(ImmutableName _parent) throws BaseError;  

  /**
  * Test if a node is a leaf of the nametree, in reading its proper children count.
  * Generally speaking, this method is faster than getChildren() which
  * searches children in the level above the parent level.
  * Call getChildCount().
  * @param _node is a node to test
  * @exception BaseError if the node is null or does not belong to the nametree
  * @return true if _node is a leaf, without any child
  */  
  public boolean isLeaf(ImmutableName _node) throws BaseError;
  
  /**
  * Returns the current index of a child in its parent array
  * in the nametree. Returns -1 if _child is the root,
  * if either the child is null or without parent, or not in the nametree.
  * @param _child is the node we are interested in
  * @return the current index of the child in its level
  */   
  public int getIndexInParent(ImmutableName _child);
  
  /**
  * Gets the current index of a child in its nametree, as the child
  * appears in the IS structure o the GUI, following a depth-first order.
  * For the level n, the index contains the value (getIndexInParent() + 1)
  * for the ancestor of this level n. The number of digits into
  * the index is always the number of levels in the nametree.
  * '0' means there is no ancestor at this level.
  * The nametree root has the index '10...0'.
  * For instance, a typical suite of indexes this method should return,
  * in a sequence of calls for all the nodes of a nametree having 4 levels,
  * is: '1000', '1100', '1110', '1120', '1121', '1122', '1200', '1210', '1211'.
  * This is the suite for a nametree having under the root '1000', two nodes with
  * the indexes '1100' and '1200', 4 children and indirect children under '1100',
  * and 2 children and indirect children under '1200'.
  * @param _child is the node or an area of the nametree
  * @return the current index of the child in the nametree, 
  * for traversing the tree in a depth-first order.
  * @exception BaseError if the arguments does not belong to the nametree,
  *   or if the nametree data is not consistent.
  */   
  public int getDepthFirstIndexInNameTree(ImmutableName _child) throws BaseError;
  
  /**
  * @return a cloned and no-backed instance of mutable nametree, without the areas.
  *   May be null.
  */
  //public Object clone();
  
  /**
  * This method compare the node lists, not the parent-child relations.
  * @param _obj is the object to compare
  * @return true if the list of nodes is equal
  */
  public boolean equals(Object _obj);
  
  /**
  * @return the hashcode value for the root.
  */
  public int hashCode();

  /**
  * @return long serial version UID for serialization of the class
  */
  public long getSerialVersionUID();  
  
} // ImmutableNameTree class