//Title:        Access Road
//Version:      0.7.0
//Copyright:    2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.CreateError;
import ARoad0.ErrB.BaseError;

import java.beans.PropertyChangeListener;

 /**
 * This interface models the Privilege, which, like ACL or AGO rights,
 * is a type of granted or denied access right delivered to an access source
 * for accessing to a target.
 * A privilege instance is controlled by an ACS, and delivers immutable rights
 * applicable to couples (source or sources set or source containers set,
 * target or targets set). The objects in the couple may be external
 * in some precise cases, in other words, they may do not belong to the privilege ACS.
 * The rights of a privilege are immutable after its creation, while the AclEntry
 * has immutable source and target, but with editable rights after the creation.
 * <P>
 * There are 7 types of privilege defined in this interface,
 * and two interfaces which are implemented by two classes:
 * - PrivilegeForType interface, where the static type is the main criteria 
 *     to apply the privilege to sources and targets which are not known by the privilege,
 *     with the four subtypes TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC,
 *     and TYPED_FOR_SEED.
 *     TYPED_CLASSED privilege is a TYPED one with contraints on source and target classes. 
 *     TYPED_CLASSED_SPECIFIC is a TYPED_CLASSED privilege for a designated source or target.
 *     TYPED_FOR_SEED is a TYPED privilege reserved for Authorization Servers.
 * - PrivilegeForLink interface, where the dynamic links to the source and the target
 *     are the single criteria to apply the privilege, which then knows to which
 *     entities it is applied, with the three subtypes LINKED, LINKED_IF_ALL_SOURCES and
 *     LINKED_FOR_ONE_TO_ONE, and with EligibleParty as access sources, Resource and
 *     VirtualFolder as access targets.
 *     The LINKED type allows to define in one privilege, a generic set of rights,
 *     and each EligibleParty which is source have all the rights on all the targets.
 *     With LINKED_IF_ALL_SOURCES, the true access sources are not the linked sources,
 *     since they must belong to ALL the set of source containers (GroupIDMembers
 *     or DirectoryEPs) which are the authorized linked sources.
 *     The LINKED_FOR_ONE_TO_ONE type is very like an AclEntry, but with static rights
 *     and for EP and Resource VirtualFolders as targets.<P>
 * <P>
 * Directory may manage container-for-target cascading privileges
 * from its parent and to its children, but it is not true for VirtualFolders.
 * At the source side, the container-for-source
 * principle is an inner principle in GroupIDMember, and it is managed explicitly
 * in DirectoryEP.
 * <P>
 * A Privilege instance is always dedicated to one ACS.
 * There is a long privilege description associated to the short one.
 * For PrivilegeForTypes, sources and target may be any ACSObject, to allow to
 * give a privilege to a Directory, and to allow certificate seeds in Authorization Servers.
 * For PrivilegeForLinks, sources are always EligibleParties, and only
 * GroupIDMembers if it has the LINKED_IF_ALL_SOURCES type.
 * For PrivilegeForLinks, access targets are Resources or VirtualFolders.
 * <P>
 * The privileges are applicable to external (that is, other ACS than the Privilege ACS)
 * objects optionally for LINKED_FOR_ONE_TO_ONE PrivilegeForLinks,
 * and always for TYPED_FOR_SEED PrivilegeForTypes. In the other cases,
 * privileges are limited to their own ACS.
 * The copy of a PrivilegeForType is for the same ACS, and the copy
 * of a PrivilegeForLink may be for another ACS.
 * The TYPED_FOR_SEED PrivilegeForTypes are the only privileges which
 * are not registered in their sources and targets.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBase.PrivilegeAbst
 */
public interface ImmutablePrivilege extends ACSObject, Cloneable, BoundBean,
  FinalizedObjectForUser, DetailledName, Comparable {

    /**
     * Gets the BaseObject name of the privilege.
     * The Privilege name is composed of: <P>
     * ACS:: ACS name <P>
     * ::  || PRIV_TYPE::  type <P>
     * ::  || NICK_NAME::  privilege name <P>
     * ::  || SENS::  'grant' or 'deny' <P>
     * ::  || XXXX::  xxxx value <P>
     * ::  >> YYYY::  yyyy value <P>
     * where: <P>
     *   - ':: ' is added by gBase.NameImpl.toString(), <P>
     *   - the type is one of the types defined in the PrivilegeAbst class, <P>
     *   - XXXX and YYYY are implementation-dependant values (see the relevant classes). <P>
     * The reverse function of getName() is getKeyReferencesFromName().
     * This name is used only outside gBase, following BaseObject interface.
     * @return the privilege name, or null if it is an empty privilege
     */
    public ImmutableName getName();

  /**
    * The main type has several uses: <P>
    * - to define the general behavior of the privilege, <P>
    * - for defining the authorized privilege standard rights. <P>
   * @return the privilege type. As a PrivilegeAbst constant, the
   *   value may be TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC, TYPED_FOR_SEED,
   *   LINKED, LINKED_IF_ALL_SOURCE, LINKED_FOR_ONE_TO_ONE
   */
  public String getMainType();
  
    /**
    * The second type is defined at the creation of the object, then it may be updated by the user.
    * It has several uses: <P>
    * - to define the type updating policy from the ACS, <P>
    * - for defining the authorized privilege specialized rights, <P>
    * - to manage the capacity to delete the instance. <P>
    * This is not the main type, which is get by getMainType().
    * @return second type, as for exemple "read" or "role one".
    *  Not null.
    */
    public String getSecondType();

    /**
     * @return true if the privilege rights
     *  grant access to the source, and false if the rights
     *  deny an access type.
     */
    public boolean getPositiveRight();
    
   /**
    * @return a long description which is displayed to the user. May be null.
    */
    public String getSeeWhy();
  
   /**
    * Gets the effective privilege rights.
    * Unworkable for a PrivilegeForType with a wrong initialization.
    * If a PrivilegeForType is classed but with a null source class or target class,
    * returns a 0-size array.
    * @return the rights. May be a 0-size array.
    */
    public StringRight[] getL_Rights();

  /**
   * @return privilege comment which is associated to the resource.
   */
  public String getComment();

  /**
   * @return ACS of this Privilege, as an External Object Reference
   */
  public ImmutableACS getEorACS();

  /**
  * @return true if equal to new Privilege instance
  */
  public boolean isEmpty();
 
  /*
   * This provides a new empty instance of a privilege
   * @return a Privilege
   */
  public Privilege getEmptyInstance();     
 
   /**
   * Only the privileges TYPED_FOR_SEED and LINKED_FOR_ONE_TO_ONE may be external.
   * An external privilege has exactly one source or one target that does not belong
   * to the privilege ACS. For linked privileges, if the source or the target is null,
   * returns always false.
   * @return for linked privilege, true if the source or the target is from another ACS,
   * and for a typed privilege, true if the source or the target may be from another ACS
   */
  public boolean isExternalPrivilege();
  
  /**
   * Gets the main components from the privilege name as a string,
   * returned by getName(). These key components define fully
   * the privilege instance.
   * This method is implemented by each PrivilegeAbst subclass,
   * and the result depends on the subclass: PrivilegeForType or PrivilegeForLinks. <P>
   * The Privilege name is composed of
   * ACS:: _aCS name::  >> PRIV_TYPE::  type::  >> PRIV_NAME::  _shortPrivilege
   * ::  >> XXXX::  _firstObject::  >> YYYY::  _secondObject <P>
   * where: <P>
   *   - ':: ' is added by gBase.NameImpl.toString(),<P>
   *   - the type is one of the constants defined in the PrivilegeAbst class<P>
   *   - XXXX and YYYY are implementation-dependant values<P>
   *   - YYYY is optional.<P>
   * This method is used by Gui1.PropertyEditor to remove
   * a privilege from a list displaying the privilege name in the beamer.
   * @param _privName is the privilege name, get through getName()
   * @return null if the name format is wrong (for PrivilegeForType or PrivilegeForLinks),
   * or an array of the 6 Privilege components:<P>
   *  - index 0: name of the ACS which manages this privilege (ImmutableName)<P>
   *  - index 1: a privilege type among the constant types defined in
   *      the PrivilegeAbst class (String)<P>
   *  - index 2: short privilege description (String)<P>
   *  - index 3: XXXX is an implementation-dependant value.<P>
   *  - index 4: YYYY is an implementation-dependant value, if it exists.
   * @exception CreateError if a main component string is not provided by NameImpl.toString()
   */
    public Object[] getKeyReferencesFromName(ImmutableName _privName)
      throws CreateError;
  
  /**
  * Gets the ACS from the privilege name.
  * @param _name a privilege name
  * @return the opened ACS.
  * @exception BaseError if the argument is not a privilege name, if
  * it is a closed ACS or if it is unknown
  */
  public ImmutableACS getAcsFromName(ImmutableName _name) throws BaseError;
  
  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName);

} // ImmutablePrivilege