//Title:        Access Road
//Version:      0.7.0
//Copyright:    2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBaseInterface;

import java.util.List;
import java.util.Map;

 /**
 * This interface is responsible for modeling
 * the target which is connected to Privileges.
 * A privilege instance is contolled by the ACS, and it is
 * then appliable to any couple (source, target) which
 * belong to the ACS.
 * There are different types of privilege (defined in PrivilegeAbst
 * class) and interfaces.
 * Fires the event 'InheritedTargetPrivileges' when the parent changes them.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBase.PrivilegeAbst to understand the different types of privilege
 */
public interface ImmutablePrivilegeTarget extends PrivilegeUser, ImmutableTarget,
    ImmutableLeaf {

  /**
   * The type is generally defined by the user.
   * It is used by PrivilegeForType instances.
   * If the user does not change it, it is equal to the
   * resource type if the object is a Resource, or to the EP type
   * if it is an EP.
   * @return String of type, as for exemple "executable" or "file".
   *  Not null.
   */
  public String getType();
  
   /**
   * Gets all the direct privileges linked to this target, not the inherited ones.
   * There is two types of privilege links (allowed or not by the ACS): <P>
   * - PrivilegeForLinks: linked privilege established for a couple
   * (ACSObject, privilege), <P>
   * - PrivilegeForType: typed privilege, where the ACSObject and the privilege
   * has a common type, but without the TYPED_FOR_SEED and
   * TYPED_CLASSED_SPECIFIC privileges. <P>
   * @return List of Privileges. May be empty, but never null.
   */
  public List getEorL_TargetPrivileges();
  
   /**
   * Gets all the privilegeForTypes of this target.
   * Does not include the TYPED_FOR_SEED and the TYPED_CLASSED_SPECIFIC privileges.
   * @return List of PrivilegeForTypes. May be empty, but never null.
   */
  public List getEorL_TargetPrivilegeForTypes();
  
   /**
   * Gets all the privilegeForLinks linked to this instance; not the
   * privilegeForTypes. Return only the direct privileges,
   * even if the privilege cascading is activated.
   * @return List of PrivilegeForLinks. May be empty, but never null.
   */
  public List getEorL_TargetPrivilegeForLinks();

   /**
   * Gets all the immutable list of the inherited linked privileges of the target.
   * To use by the consumers in the gWork packages, which deny any access if
   * the returned value is null.
   * @return immutable list of Linked Privileges. May be empty but never null.
   */
  public List<PrivilegeForLinks> getEorImmutableL_TargetInheritedLinkedPrivileges();

  /**
   * Gets the internal map of the inherited linked privilege rights, where
   * the default key is the name of the EligibleParty having the rights, extended with,
   * as last component, the parent order of the direct or indirect parent
   * which has given the rights.
   * For instance, the direct parent has the string '1', and its direct parent
   * has the string '2' as parent order in the last name component, and so on.
   * This property is not displayed to the user. A subclass may change the key form.
   * The map value is always a non-empty StringRight array of inherited rights.
   * Called for the gWork consumers, but only in the AcsAddons implementing
   * the privilege inheritance. An AcsAddon may choose to not use this property.
   * @return unmodifiable map of inherited rights from the linked privileges.
   * May be null but never empty.
   */
  public Map<String, StringRight[]> getM_TargetInheritedLinkedPRIRights();

    /**
    * Returns the Privilege to use as inherited Privilege in a child.
    * Empty method which has to be overridden into an AcsAddon handling
    * the privilege inheritance.
    * @param _lpri the initial Privilege which may be given to the child
    * @param _type type of a child, as Resource
    * @return the inherited Privilege if the child inherits this initial Privilege, or null
    */
  public PrivilegeForLinks selectInheritedPrivilege(PrivilegeForLinks _lpri, String _type);
  
} // ImmutablePrivilegeTarget