//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.Set;
import java.util.Map;
import java.util.List;

/**
 * This interface is the immutable version of Resource.
 * The resource may have a parent in a resource tree in this ACS,
 * As a general resource, it is accessed
 * from EligibleParties (userID, groupID, actor) through
 * AclEntries belonging to this ACS or to another one.
 * As a resource with AGO rights, it has an userID (account)
 * and a groupID (group) in this ACS, and rights for them in this ACS.
 * As a PrivilegeTarget, it handles the Privileges.
 * This interface has several subinterfaces in this package.
 * Directories, files, executable progams are examples of resources.
 * @version 0.7.0
 */
public interface ImmutableResource extends ImmutableBasic, ImmutablePrivilegeTarget {

  /**
   * Gets the open ACS that manages an external AclEntry or Linked Privilege
   * for this resource. The resource ACS forbids to have both external ACL and external privilege.
   * @return the open ACS controller names. May be null, never empty nor with null items.
   */
  public ImmutableName[] getL_AcsControllers();

  /**
   * Get the userID, following the AGO rights.
   * @return owns this resource, in the main ACS. May be null.
   */
  public ImmutableUserID getEorUserID();

  /**
   * Get the rights of the resource owner.
   * @return array of positive access rights for the UserID which owns this resource,
   *   in the main ACS. May be empty, but not null.
   */
  public  StringRight[] getL_ownerRights();

  /**
   * @return primary group which owns this resource,
   * in the main ACS. May be null.
   */
  public ImmutableGroupID getEorGroupID();

  /**
   * Get the rights of the resource group.
   * @return array of positive access rights for the GroupID which owns this resource,
   *   in the main ACS. May be empty, but not null.
   */
  public  StringRight[] getL_groupRights();

  /**
   * Get the rights of the root, following AGO rights concept.
   * @return array of access rights for the root UserID,
   *  in the main ACS, following the Unix/Linux scheme. May be empty, but not null.
   */
  public  StringRight[] getL_rootRights();

  /**
   * Get resource type, as for exemple "executable" or "file".
   * @return String of resource type.
   */
  public String getResourceType();

  /**
   * Get the list where each AclEntry delivers rights to an EligibleParty
   * to access or to deny access to this Resource.
   * @return the AclEntry list. May be null.
   */
  public AclEntry[] getL_aclEntry();

  /**
  * True if an EligibleParty is associated to the resource through an AclEntry.
  * @param _ep to test. May be null.
  * @return true if _ep is referenced in at least one aclEntry in this resource.
  */
  public boolean containsAclEntryToEP(EligibleParty _ep);

  /**
  * If there is no parent, return true to avoid useless processing.
  * @return true if this instance has the same userID and main group than its direct parent.
  */
  public boolean hasTheAccountAndGroupOfParent();

  /**
  * @return true if the parent authorizes its userID to have access to this instance.
  */
  public boolean getAGOAccessFromTheParentAccount();

  /**
  * @return true if the parent authorizes its groupID to have access to this instance.
  */
  public boolean getAGOAccessFromTheParentGroup();
  
  /**
  * Get all the EligibleParty associated to the resource through at least one
  * AclEntry.
  * @return EligibleParties, through external object references.
  * May be empty.
  */
  public Set getEorL_EPthroughAclEntry();

  /**
   * @return boolean true if this resource is an actor.
   */
  public boolean getActorActivated();

  /**
   * Get the inherited rights from the parent of the resource owner. 
   * @return an array of access rights for the UserID which owns this resource,
   *   in the main ACS. May be empty but not null.
   */
  public StringRight[] getL_inheritedOwnerRights();
  
  /**
   * Get the inherited rights from the parent of the resource group. 
   * @return an array of access rights for the GroupID which contains this resource,
   *   in the main ACS. May be empty but not null.
   */
  public StringRight[] getL_inheritedGroupRights(); 

   /**
   * Gets all the immutable list of the inherited AclEntries of the target.
   * To use by the consumers in the gWork packages, which deny any access if
   * the returned value is null.
   * @return immutable list of AclEntries. May be empty or null.
   */
  public List<AclEntry> getEorImmutableL_TargetInheritedAclEntries();

  /**
   * Gets the internal map of the inherited AclEntry rights, where
   * the default key is the name of the EligibleParty having the rights on the target,
   * extended with, as last component, the parent order of the target parent
   * which has given the rights.
   * For instance, the direct parent has the string '1', and its direct parent
   * has the string '2' as parent order in the last name component, and so on.
   * This property is not displayed to the user. A subclass may change the key form.
   * A subclass may use the constructor TreeMap(Comparator<? super K> comparator)
   * to set an ordering other than the natural ordering of the keys.
   * The map value is always a non-null StringRight array of inherited rights,
   * and it may be empty.
   * Called for the gWork consumers, but only in the AcsAddons implementing
   * the ACL inheritance. An AcsAddon may choose to not use this property.
   * @return unmodifiable map of inherited rights from the AclEntries.
   * May be empty but never null.
   */
  public Map<String, StringRight[]> getM_TargetInheritedACLRights();

    /**
    * Returns the AclEntry to use as inherited AclEntry in a child.
    * Empty method which has to be overridden by any AcsAddon handling the ACL inheritance.
    * @param _acl the initial AclEntry which may be given to the child
    * @param _type type of a child, as Resource
    * @return the inherited AclEntry if the child inherits this initial AclEntry, or null
    */
  public AclEntry selectInheritedAclEntry(AclEntry _acl, String _type);

  /**
   * A constrained (opened or closed) child ACS has one or several
   * of the following relations to this Resource: <P>
   * <ForChildAcsOpeningAndClosing>, <ForChildAcsOpeningNoClosing> 
   * or <ForChildAcsIntegrity>.<P>
   * Each relation is associated to a key with the same name.
   * @return a map of 3 ACS name lists for the 3 keys <ForChildAcsOpeningAndClosing>, 
   *    <ForChildAcsOpeningNoClosing> and <ForChildAcsIntegrity>. May be empty but
   *    never null.
   */
  public Map<String, ImmutableName[]> getM_ConstrainedChildACS();
  
} // ImmutableResource
