//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBaseInterface;

import java.beans.PropertyChangeListener;

/**
 * This interface models a container of Resources belonging to one ACS,
 * responsible for being a PrivilegeTarget, an alias or an alias reference.
 * A VirtualFolder is also an ImmutableSource. It manages HAS_VIRTUAL_MEMBER relations
 * to its members. The relation VirtualFolder/member is weak. The deleting
 * of the container does not delete the members, on the contrary of Directory.
 * There are two main types of virtual folders, where the members are always Resources: <P>
 * - for Resource members which are not EligibleParties ('RES_VFOLDER' type) <P>
 * - for Resource members which may be EligibleParties ('EP_VFOLDER' type) <P>
 * Thus, an Virtual Folder is an access target, and it is an access source only as reference
 * of an alias or as container of its members.<P>
 * The default behavior in the rights propagation is the RIGHTS_PASSING_OFF_TO_MEMBERS mode.
 * This means that the VirtualFolder transfers none of its rights as target to its members.
 * In the other mode RIGHTS_PASSING_ON_TO_MEMBERS, the ACS has set this mode for its
 * virtual folders, and all the virtual folder rights are transfered to its members.<P>
 * Trees of virtual folders are possible. There is no rights inheritance in the tree
 * for the generic gBase classes, but an AcsAddon may override the hook methods. <P>
 * The virtual folder manages: <P>
 * - its members, as any folder, and they are Basics of one ACS; they are not
 * deleted when this folder is deleted, <P>
 * - its children, as any node, and they are virtual folders (no Basic) which are deleted
 * when this node is deleted, <P>
 * - its privileges, as any privilege target. <P>
 * There are two main types of virtual folders: <P>
 * - for Resource members which are not EligibleParties ('RES_VFOLDER' type) <P>
 * - for Resource members which are EligibleParties ('EP_VFOLDER' type) <P>
 * Each type allows only one folders tree and then, one folder root in an ACS. <P>
 * It is possible to force each member to be member at most
 * one time in all the folder tree (EP or Resource tree). <P>
 * There is no link between the members of a virtual folder
 * and the members of its children.
 * There is no link between the privileges of a virtual folder
 * and the privileges of its children.
 * But, as any node, the deleting of a virtual folder implies the deleting of its
 * children.
 * The typed privileges are applies to the virtual folder types.
 * Then, they may be used to defined default virtual folder rights.
 * A VirtualFolder is an ImmutableSource since it manages HAS_VIRTUAL_MEMBER relations
 * to its members.
 * Caution: no equals() nor clone() methods.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @version 0.7.0
 */
public interface ImmutableVirtualFolder extends ImmutableFolder, ImmutableNode,
  ImmutablePrivilegeTarget, ImmutableSource {
  
    /**
     * True if the ACS does not allow an ACSObject to
     * be more than one time in the folders tree of
     * relevant type (EP tree or Resource tree).
     */
    public boolean getMemberOneTimeInTree();

  /**
   * The virtual folder main types are defined by the program.
   * @return String of VF type, as "RES_VFOLDER" and "EP_VFOLDER".
   *    Not null.
   */
  public String getMainVFType();

  /**
  * An array of external object references on the children
  *  of this node, which are virtual folders.
  * @return ImmutableVirtualFolder[]. May be null but not empty.
  */
  public ImmutableVirtualFolder[] getEorL_Children();

/**
 * Tests of equals() on name, children and parent.
 * @param _obj to compare to this instance
 * @return true if _obj is a VirtualFolderImpl instance (or an instance of an
 *  extended class) equals to the current one
 */
public boolean equals(Object _obj);

  /**
   * Returns the change listeners for a property. Used by the AcsAddons.
   * @param _propertyName name of the listened property.
   * @return the change listeners. May be null or empty.
   */
  public PropertyChangeListener[] getPropertyChangeListeners(String _propertyName);

} // ImmutableVirtualFolder