//Title:        Access Road
//Version:      0.7.0
//Copyright:    2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import java.util.List;
import java.util.Set;

import ARoad0.ErrB.BaseError;

/**
* This class is responsible for managing the No-More-Than and No-Less-Than views.
* The aim is to check if some Eligible Parties (or Sources) has rights on the access target
* above a given 'no-more' right and below a given 'no-less' right.
* The 'less' right criteria is generally used to check the compliance to the access needs.
* The 'more' right criteria is generally used to check the compliance to the access security policy.
* This view handles one Target, a set of Sources,
* another set of excluded Sources which belong or run under one of the
* first-set Source, a StringRight 'more' criteria and a StringRight 'less' criteria.
* Caution: indeed, every Source is a Target, since a GroupDMember for example
* is the target of the RUN_UNDER relation.
* @version 0.7.0
* @see ARoad0.Gui2.GraphicNoThanView
*/
public interface NoThanViewInBase extends EPRViewInBase {

  /**
   * This view handles exactly one target.
   * @return the target. May be null only before the first updating.
   */
  public ImmutableTarget getEorFirstAccessTarget();
  
  /**
   * This view handles exactly one access target. The More and Less criteria are set to null.
   * @param _targ the target to add. Never null, so a current target cannot be removed.
   * @exception BaseError if the target is also a registered source in this view,
   * if the number of targets is too large,
   * if the argument is null, or the object name is null or empty.
   */
  public void setEorFirstAccessTarget(ImmutableTarget _targ) throws BaseError;
  
  /**
   * This view handles exactly one target, in this class version.
   * @param _targ the target to add
   * @exception BaseError if the target is also a registered source in this view,
   * if the number of targets is too large,
   * if the argument is null, or the object name is null or empty.
   */
  public void addEorNoThanAccessTarget(ImmutableTarget _targ) throws BaseError;

  /**
   * This view handles exactly one target.
   * @param _targ the target to remove
   * @exception BaseError if the argument is not the current target,
   * if the argument is null, or the object name is null or empty.
   */
  public void removeEorNoThanAccessTarget(ImmutableTarget _targ) throws BaseError;
    
  /**
   * The aim is to check if an Eligible Party has rights on the target,
   * which are below the 'more' criteria.
   * @return the 'more' criteria. May be null.
   */
  public StringRight getMoreCriterion();
    
  /**
   * The aim is to check if an Eligible Party has rights on the target,
   * which are above the 'more' criteria.
   * @param _right is the 'more' criteria
   * @exception BaseError if the argument is null.
   */
  public void setMoreCriterion(StringRight _right) throws BaseError;

  /**
   * The aim is to check if an Eligible Party has rights  on the target,
   * which are below the 'less' criteria.
   * @return the 'less' criteria. May be null.
   */
  public StringRight getLessCriterion();
  
  /**
   * The aim is to check if an Eligible Party has rights  on the target,
   * which are below the 'less' criteria.
   * @param _right is the 'less' criteria
   * @exception BaseError if the argument is null.
   */
  public void setLessCriterion(StringRight _right) throws BaseError;
  
  /**
   * An excluded Source is connected to a view basic Source, but it
   * has not to be checked since it is explicitly put away by the user. 
   * So its compliance to the more and to the less criteria
   * will stay unknown.
   * @return list of the excluded sources. Never null.
   */
  public List<ImmutableSource> getExcludedSources();
  
  /**
   * An excluded Source is connected to a view basic Source, but it
   * has not to be checked since it is explicitly put away by the user. 
   * So its compliance to the more and to the less criteria will stay unknown.
   * @param _sour to test
   * @return true if the view references this Source as excluded
   */
  public boolean containsExcludedSource(ImmutableSource _sour);
  
  /**
   * List of Sources which are in the perimeter of checking, and that
   * includes the view BaseObjects and the listening perimeter. 
   * Gets the checking perimeter which includes the view BaseObjects and the listening perimeter.
   * The associates of an Source are defined as their direct/indirect members
   * and their direct/indirect running-under actors.
   * This listening perimeter is get by the subtraction of the excluded Sources
   * and their associates from the view Sources list (also called 'basic rights users')
   * and their associates. <P>
   * For instance, if the group A is a view Source, and has only group B and
   * group C has direct members, and if the group C is a view excluded Source,
   * the listening perimeter contains only the group B and its associates. <P>
   * The Sources in the listening perimeter do not registered this NoThanViewInBase
   * has an implied view. The size is limited to 50.
   * @return the Sources in the checking perimeter. Never null.
   */
  public Set<ImmutableSource> getCheckingPerimeter();
  
  /**
   * List of Sources from the view Sources list which have at most a connected
   * member or a running-under actor being an excluded Source. Do not contain
   * the reduced Sources in the checking perimeter which are not view Sources, that is
   * when they are members or running-under Actors of a view Source.
   * @return the Sources. Never null.
   */
  public Set<ImmutableSource> getReducedSourcesInViewSources();
  
  /**
   * An excluded Source is in the perimeter of checking, but
   * has not to be checked. So its compliance to the more and to the less criteria
   * will stay unknown. The perimeter of checking is defined as the set of the Sources
   * which are registered in the view (but such Source may not be excluded),
   * and the Source which are  direct or indirect member of, or run under,
   * a registered Source of the view.
   * @param _sour to exclude
   * @exception BaseError if the argument is a registered Source,
   * is not in the perimeter of checking,
   * if the argument is null, or the object name is null or empty.
   */
  public void addEorExcludedSource(ImmutableSource _sour) throws BaseError;

  /**
   * An excluded Source is in the perimeter of checking, but
   * has not to be checked. So its compliance to the more and to the less criteria
   * will stay unknown. The perimeter of checking is defined as the set of the Source
   * which are registered in the view (but such Source may not be excluded),
   * and the Source which are direct or indirect member of, or run under,
   * a registered Source of the view.
   * @param _sour non-null Source to remove as excluded Source
   * @exception BaseError if the Source is unknown as excluded, or
   * if the argument is null, if the object name is empty.
   */
  public void removeEorExcludedSource(ImmutableSource _sour) throws BaseError;
  
  /**
  * Creates, registers in the base and saves a view copy with another name.
  * Calls copy(). Overrides the super method.
  * @param _name of the copy. It is copied without test.
  * @exception BaseError if _name is null or already known in BaseManagerImpl,
  * or if _view is null or not known.
  * @return the copied view
  */
  public NoThanViewInBase copyAndSave(ImmutableName _name) throws BaseError;
  
} // NoThanViewInBase
