//Title:        Access Road
//Version:      0.7.0
//Copyright:    2006, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBaseInterface;

import java.util.List;

 /**
 * This interface is responsible for modeling the privileges which manages some access rights
 * from a privilege sources list (or only one GroupID) to a privilege targets list
 * (or only one target).
 * There are three types of PrivilegeForLinks: LINKED,
 * LINKED_IF_ALL_SOURCES and LINKED_FOR_ONE_TO_ONE. 
 * This privilege is usable by any couple (all sources set, one of the target)
 * from the privilege ACS, or from two external ACS for LINKED_FOR_ONE_TO_ONE privileges.
 * A linked privilege may be inherited from the (in)direct parents of the target,
 * except the LINKED_IF_ALL_SOURCES privilege.
 * <P>
 * Eligible Party and Resource classes are never used,
 * on the contrary of PrivilegeForTypes, and the linked sources and targets
 * types are used to register and unregister them in the privilege.
 * So, it is possible to modify the privilege links to sources and targets on the fly.
 * <P>
 * This is a javabean with the following bound properties:
 * - 'OneLinkedSource'
 * - 'OneLinkedTarget'
 * <P>
 * The LINKED type allows to define in one privilege, a generic set of rights,
 * and each EligibleParty which is source has all the rights on all the targets.
 * With LINKED_IF_ALL_SOURCES, the rights on the targets are used by any EligibleParty
 * which belongs to ALL the source containers, which are GroupIDMembers or DirectoryEPs.
 * For instance, if there are a GroupID and an UserID as privilege sources,
 * any Actor which is a member of the GroupID AND belongs to the UserID, 
 * has the rights of the privilege.
 * If an actor belongs to the GroupID but is not under the UserID, it has not
 * the privilege. To have an OR relation between the GroupID and the UserID,
 * rather than AND, let's use 2 LINKED_FOR_ONE_TO_ONE privileges or two AclEntries.
 * The number of privilege sources is limited to 4.
 * In its principle, a LINKED_IF_ALL_SOURCES privilege is close to a conditional AclEntry,
 * where the effective rights are conditioned by the groups of the source.
 * <P>
 * With LINKED_FOR_ONE_TO_ONE, there is at most one source, and it is an EligibleParty,
 * and at most one target, which is a PrivilegeForLinksTarget.
 * A LINKED_FOR_ONE_TO_ONE privilege is like an ACL from an EP to
 * a resource or a virtual folder.
 * A LINKED_FOR_ONE_TO_ONE PrivilegeForLink is an alternative to an AclEntry
 * because both have one (external) access source and one (external) access target,
 * but the differences are that AclEntry cannot change its source and target,
 * it is deleted when its source or its target is deleted,
 * it can change dynamically its rights, and:<P>
 * - on sources, it is as powerfull as a LINKED_FOR_ONE_TO_ONE PrivilegeForLinks.<P>
 * - on targets, it is less powerfull since it cannot be applied to Resource or
 *   EligibleParty VirtualFolders,
 *   whereas a LINKED_FOR_ONE_TO_ONE PrivilegeForLink can do it. 
 * In a one-to-one privilege, the target may be added once but it cannot be removed.
 * <P>
 * The rights are applicable to external (other ACS than the privilege ACS)
 * objects only for LINKED_FOR_ONE_TO_ONE privileges, and isExternal() is true
 * when the source or the target is from another ACS.
 * The target ACS cannot manage the PRI inheritance in a LINKED_IF_ALL_SOURCES privilege.
 * For the other types, only one target is authorized if the first target is from
 * an ACS that manages the PRI inheritance.
 * <P>
 * The privilege key is used by the ACS to find a privilege.
 * The key is the couple (privilege type, short privilege description).
 * The main privilege value is an unmodifiable list of rights.
 * The privilege gets positive or negative rights: it grants or denies accesses.
 * The rights and the key CANNOT be updated after the privilege creation,
 * nor the rights sens (grant or deny).
 * There is a long privilege description associated to the short one.
 * <P>
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBaseInterface.Privilege
 * @see ARoad0.gBaseInterface.PrivilegeSource
 * @see ARoad0.gBaseInterface.PrivilegeTarget
 */
public interface PrivilegeForLinks extends Privilege {

   /**
   * @return true if the type of the instance is LINKED_FOR_ONE_TO_ONE.
   */
  public boolean getOneRoleOneTarget();
    
   /**
   * Get the sources set which have this privilege.
   * If LINKED_IF_ALL, the privilege is applied only
   * to the objects belonging to every linked source.
   * If LINKED_FOR_ONE_TO_ONE, there is at most one GroupID
   * as linked source.
   * @return List of PrivileForLinksSources. May be empty,
   *   but never null.
   */
  public List<PrivilegeSource> getEorLinkedSources();

   /**
   * Get the targets which use this privilege.
   * If LINKED_FOR_ONE_TO_ONE, there is at most one target.
   * If the targets are external ones, all targets belong to the same ACS.
   * @return List of PrivileForLinksTargets. May be empty,
   *   but never null.
   */
  public List<PrivilegeTarget> getEorLinkedTargets();
 
   /**
    * @return the single source name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getSingleSourceName();
  
   /**
    * @return the single target name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getSingleTargetName();

   /**
    * @return the single source ACS name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getNameOfSingleSourceACS();

   /**
    * @return the single target ACS name, for LINKED_FOR_ONE_TO_ONE privilege only,
    *   or null.
    */
    public ImmutableName getNameOfSingleTargetACS();

   /**
    * Caution: the list is from the privilege ACS, and an external privilege
    * may accept types from another ACS.
    * @return the source types that the ACS authorizes for this privilege.
    *  May be empty but never null.
    */
    public List<String> getL_InternalAllowedSourceTypes();

   /**
    * Caution: the list is from the privilege ACS, and an external privilege
    * may accept types from another ACS.
    * @return the target types that the ACS authorizes for this privilege.
    *  May be empty but never null.
    */
    public List<String> getL_InternalAllowedTargetTypes();

} // PrivilegeForLinks