//Title:        Access Road
//Version:      0.7.0
//Copyright:    2006, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.BaseError;

 /**
 * This interface is responsible for modeling
 * the source which is connected to Privileges.
 * EligibleParties are PrivilegeSources.
 * A privilege instance is controlled by the ACS, and it is
 * then appliable to any couple (source, target) which
 * belong or not to the privilege ACS.
 * There are different types of privilege (defined in PrivilegeAbst class).
 * The sources and the targets are defined through
 * dynamic or static links to the privilege.
 * It is a javabean with the following bound properties:
 * 'OneSourceTypedPrivilege' and 'OneSourceLinkedPrivilege'.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBase.PrivilegeAbst to understand the different types of privilege
 * @see ARoad0.gBaseInterface.PrivilegeTarget
 */
public interface PrivilegeSource extends ImmutablePrivilegeSource {

   /**
   * Sets a privilege as applicable to this access source, and this is called only by the program.
   * The gBase classes are the only caller of this method
   * at any creation/deletion of a typed privilege or an access source.
   * Fires 'OneSourceTypedPrivilege' events.
   * @param _priv is a Privilege to register in the source
   * @exception BaseError if the ACS does not manage the privileges, or if
   *   this instance is not a privilege source.
   */
  public void addEorPrivilegeForTypeToSource(PrivilegeForType _priv) throws BaseError;

   /**
   * Sets a privilege as applicable to this access source, and called by the program
   * or by the user.
   * This instance is registered in the privilege by this method.
   * If the source is also a source container,
   * the privilege is used by the children of the target if the target 
   * is a GroupIDMember (as any right), or, if it is a DirectoryEP,
   * it may be used by the children if its ACS policy specifies it.
   * Fires 'OneSourceLinkedPrivilege' events.
   * @param _priv is a Privilege to register in the source
   * @exception BaseError if the ACS does not manage the privileges,
   *   this instance is not a privilege source, or: <P>
   *   for all linked privileges: <P>
   *   - if the number of sources will be superior to 4, or <P>
   *   - the source is null, empty, or its name is null, or <P>
   *   - the source name is already known by this instance.
   *   for LINKED_FOR_ONE_TO_ONE privileges, if: <P>
   *   - there is already a source, or <P>
   *   - the parameter is not a GroupID <P>
   *   and, for all PrivilegeForLinks excepted for LINKED_FOR_ONE_TO_ONE privileges:
   *   - the source belongs to another acs than the privilege acs
   */
  //public void addEorPrivilegeForLinksToSource(PrivilegeForLinks _priv) throws BaseError;

   /**
   * Unregisters on the privilege when it is no more used by
   * this privilege source.
   * Fires 'OneSourcePrivilege' events.
   * @param _priv is a Privilege to register in the source
   * @exception BaseError if the privilege does not known this source.
   *    User-displayable message.
   */
  public void removeEorPrivilegeToSource(Privilege _priv) 
    throws BaseError; 

} // PrivilegeForLinksSource