//Title:        Access Road
//Version:      0.7.0
//Copyright:    2006, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones. The relevant court is at Nanterre (92).
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.BaseError;

 /**
 * This interface is responsible for modeling
 * the target which is connected to Privileges.
 * Resources and VirtualFolders are PrivilegeTargets.
 * A privilege instance is contolled by the ACS, and it is
 * then appliable to any couple (source, target) which
 * belongs or not to the privilege ACS.
 * It is a javabean with the following bound properties:
 * 'OneTargetTypedPrivilege', 'OneTargetLinkedPrivilege'.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @see ARoad0.gBase.PrivilegeAbst to understand different types of privilege
 * @see ARoad0.gBaseInterface.PrivilegeSource
 */
public interface PrivilegeTarget extends ImmutablePrivilegeTarget {

   /**
   * Sets a privilege as applicable to this access target, and called only by the program.
   * The program, not the user, is the only caller of this method
   * at any creation/deletion of a typed privilege or an access target.
   * Fires 'OneTargetTypedPrivilege' events.
   * @param _priv to register in the target
   * @exception BaseError if the ACS does not manage the privileges, or if
   *  this instance is not a privilege target
   */
  public void addEorPrivilegeForTypeToTarget(PrivilegeForType _priv) throws BaseError ;

   /**
   * Sets a privilege as applicable to this access target, and called by the program
   * or by the user.
   * This instance is registered in the privilege by this method.
   * The cascading policy sets, when the target
   * is a Directory, if the children are also targets for the same privilege.
   * Fires 'OneTargetLinkedPrivilege' events.
   * @param _priv to register in the target
   * @exception BaseError if the ACS does not manage the privileges,
   *  this instance is not a privilege target, or: <P>
   *   if it is a LINKED_FOR_ONE_TO_ONE privilege, if: <P>
   *   - there is already a target <P>
   *   for all PrivilegeForLinks: <P>
   *   - if privileges are inherited from the parent, <P>
   *   - if the target is null, empty or its name is null, or <P>
   *   - the target name is already known by the privilege. <P>
   *   and, for all PrivilegeForLinks excepted for LINKED_FOR_ONE_TO_ONE privileges: <P>
   *   - the target belongs to another acs than the privilege acs
   */
  //public void addEorPrivilegeForLinksToTarget(PrivilegeForLinks _priv) throws BaseError ;

   /**
   * Unregisters on the privilege when it is no more applicable to
   * this access target.
   * Fires 'OneTargetTypedPrivilege' and 'OneTargetLinkedPrivilege' events.
   * @param _priv to remove from the target
   * @exception BaseError for PrivilegeForLinks only, if the privilege
   * does not known this source. User-displayable message.
   */
  public void removeEorPrivilegeToTarget(Privilege _priv) 
    throws BaseError; 

} // PrivilegeForLinksTarget