//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.UpDateError;
import ARoad0.ErrB.CreateError;

/**
 * This interface is responsible for modeling a resource in an ACS.
 * The resource may have a parent in a resource tree in this ACS,
 * and may be the alias of an EligibleParty of another ACS.
 * <P>
 * As a general resource it is accessed
 * from EligibleParties (userID, groupID, actor) through
 * AclEntries belonging to this ACS or to another one.
 * As a resource with AGO rights, it has an userID (account)
 * and a groupID (group) in this ACS, and rights for them.
 * The Resource is a PrivilegeTarget.
 * This interface has several subinterfaces in this package.
 * Directories, files, executable progams are examples of resources.
 * The resource is automatically registered in at most one ACS
 * (the main one) plus others (the controllers). So, finalizeForUser()
 * HAVE TO BE CALLED BEFORE DELETING the resource.
 * <P>
 * It is a javabean with the following bound properties:
 * 'Comment', 'AclEntries', 'UserID', 'GroupID',
 * 'RootRights', 'OtherRights', 'GroupRights', 'OwnerRights',
 * 'EPsThroughAclEntry', 'AcsControllers',
 * and from the subclasses in the AcsAddon,
 * 'InheritedTargetPrivileges', 'InheritedTargetAclEntries'.
 * The listeners are proper to this instance, outside gBase,
 * transient (not serialized in this instance backup), and they are called in any order.
 * As a BaseBeanManager, fire closing/deleting events on AclEntries.
 * @version 0.7.0
 * @since 0.4.0
 */
public interface Resource extends ImmutableResource, Leaf, Basic, PrivilegeTarget {

  /**
  * Defines the group owning this resource. Some resources
  *  as actors have secondary groups in the same ACS.
  * @param _g ImmutableGroupID which owns this resource, in the main ACS.
  * @exception UpDateError if _g does not belong to the main ACS of this resource.
  */
  public void setEorGroupID(ImmutableGroupID _g) throws UpDateError;

/**
   * Set the userID of the resource, and set the current userID in the own actors
   * if it was the previous userID. It is a bound property.
   * Change the group, which becomes the primary group of the new account if it is not null.
   * @param _u owns this resource, in the main ACS. May be null.
   * @exception UpDateError if _u does not belong to this resource main ACS.
 */
public void setEorUserID(ImmutableUserID _u) throws UpDateError;

/**
   * Sets the rights for the account which owns this resource.
   * The rights may be acsrights and metarights, positive and negative rights.
   * The ACS constraints are always applied.
   * No action if the instance has been finalized. If a new right
   * is not allowed by the ACS, it is removed and the other rights are kept.
   * No adding of null elements. 'full_control' equivallent is the default right
   * if there is no umask. 
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the owner rights are immutable.
 */
public void setL_ownerRights (StringRight[] _s) throws UpDateError;


/**
   * Set the rights for the members of the groupID which owns this resource.
   * The rights may be acsrights and metarights, positive and negative rights.
   * The ACS constraints are always applied.
   * No action if the instance has been finalized. If a new right
   * is not allowed by the ACS, it is removed and the other rights are kept.
   * No adding of null elements.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the group rights are immutable.
 */
public void setL_groupRights(StringRight[] _s) throws UpDateError;

/**
   * Sets the rights for the userID 'root'.
   * "full_control" equivallent is a mandatory right, even if it is not 
   * in the constrained rights list.
   * No action if the instance has been finalized or if a right is not allowed.
   * Fires a PropertyChange for 'RootRights'.
   * @param _s array of acsrights and/or metarights. May be null.
   * @exception UpDateError if the root rights are immutable.
 */
public  void setL_rootRights(StringRight[] _s) throws UpDateError;

  /**
   * If _st is allowed by the ACS, set the type of this resource.
   * This method is used to switch from an <immutable> type
   * which forbids the updates of a protected property, and to return to this protecting type.
   * Then , it is possible to switch for example from 'actor' to
   * '<immutable> actor', but not to '<immutable> directory'.
   * From the default value '<undefined>', any first update may be done.
   * Fire a PropertyChange for 'ResourceType'.
   * @param _st type of the resource.
   * @exception UpDateError if the type is null, equal to <byAcsCreator>,
   *    or not known by the ACS, or not associated to the previous type.
   */
  public void setResourceType(String _st) throws UpDateError;

  /**
   * Creates a new aclEntry which delivers rights to an eligible party
   * to access or to deny access to this resource.
   * The resource name must have a name following the pattern
   * IS_name / ACS_name / resource_name, to allow the identification
   * of the ACS in _ep.
   * 'AclEntries' and 'EPsThroughAclEntry'  are bound properties.
   * @param _aclAcs ImmutableACS which owns the new AclEntry, and to be recorded
   *  as an ACS controller of this resource if it is not its main ACS.
   * @param _ep eligible party connected to the resource through a new aclEntry.
   * @param _right is an string array of positive or negative rights of _ep on this resource
   *  (that is, if READ is a right, _ep can read on the resource). May be null.
   * @param _positiveRight is true if the rights are set to authorize (grant),
   *  and false if they are set to forbid (deny).
    * @param _conditionGroup first condition group. May be null.
   * @exception InternalError if this resource has a null name or a null ACS,
   *  or if _ep is not an EligiblePartyImpl nor an ActorImpl.
   * @exception CreateError if _ep or _aclAcs is null, or if _ep is already recorded
   *  in this resource with the same ACS, or thrown by the ACS.newEorAclEntry().
   * @return new AclEntry.
  */
  public AclEntry addAclEntry(ImmutableACS _aclAcs,ImmutableEligibleParty _ep,
    StringRight[] _right, Boolean _positiveRight, ImmutableGroupID _conditionGroup)
    throws CreateError;


  /**
    * Deletes an aclEntry which delivers rights to an eligible party
    * to access or to deny access to this resource.
    * 'AclEntries' and 'EPsThroughAclEntry'  are bound properties.
    * @param _aclAcs ImmutableACS which owns the AclEntry, and is recorded
    *  as an ACS controller of this resource.
    * @param _ep eligible party connected to the resource through an AclEntry.
    *  This method unrecords the relevant aclEntry in _ep, in this resource,
    *  and in the relevant ACS. The ACS of the aclEntry may be unrecorded in this
    *  resource if there are no other aclEntry owned by this ACS in this resource.
    * @param _positiveRight is true if the rights are oriented to authorize,
    *  and false if they are oriented to forbid.
    * @param _conditionGroup first condition group. May be null.
    * @exception UpDateError if _ep is null, unknown, or if _aclAcs is closed or unknown.
    */
   public void removeAclEntry(ImmutableACS _aclAcs, ImmutableEligibleParty _ep,
    boolean _positiveRight, ImmutableGroupID _conditionGroup) throws UpDateError;

    /**
    * Resets the inherited AclEntry rights after an user edition.
    * Only to be overridden in an AcsAddon, so it is an empty method in the generic class.
    * Called by an AclEntry when it fires 'AclRights' or a 'ConditionSourceGroups'
    * property change, since this implies a change in the effective rights of
    * itself as inherited AclEntry in the children of the Directory it has as target.
    * This method has to be overread only in the Resource implementations of 
    * the AcsAddons handling the ACL inheritance.
    * Not described in the BeanInfo since the user does not call it.
    * @param _acl inherited AclEntry with updated rights
    */
  public void resetInheritedAclEntryRights(AclEntry _acl);

    /**
    * Resets the inherited Privilege rights after an user edition.
    * Called by the master parent when there is a change in the effective rights of
    * an inherited Privilege. This is only when it is about a change in the sources.
    * Contrary to AclEntry, the rights are immutable and there is no condition.
    * Not described in the BeanInfo since the user does not call it.
    * Note: the creation of the inherited Privilege in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited Privilege
    * is handled.
    * @param _lpri inherited Privilege with updated rights. Never null.
    */
  public void resetInheritedPrivilegeRights(PrivilegeForLinks _lpri);

    /**
    * Returns the Privilege to use as inherited Privilege in a child.
    * Empty method which returns the argument. Should be never called, and throws
    * an InternalError. Has to be overridden into an AcsAddon handling
    * the privilege inheritance.
    * @param _lpri the initial Privilege which may be given to the child
    * @param _type type of a child, as Resource
    * @return the inherited Privilege if the child inherits this initial Privilege, or null
    */
  public PrivilegeForLinks selectInheritedPrivilege(PrivilegeForLinks _lpri, String _type);

  /**
  * This method provides a copy of the instance as Copyable.copy(),
  * but without copying the internal aclEntries.
  * It is used in ACSImpl.copy().
  * Each internal variable that is an external
  * object reference have a shallow copy (no deep copy),
  * but without copy for the internal aclEntry nor the list of ViewInBase
  * that contains this instance.
  * @param _name copy name. Usually creates as an extension of _parent name.
  * @param _acs main ACS of the copy to create. May not be the ACS
  *   of this instance.
  * @param _parent DirectoryImpl which belongs to _acs and will have the copy.
  *   May be null. Must be null for some implementations.
  * @param _toRegisterInManager false if this method must not register
  *  the instance in the ACS.
  * @return a copy of the object, in the same class.
  * @exception CreateError of the constructor with the same parameters,
  *   or if _parent ACS is not _acs, or addAclEntry() exceptions.
  **/
  public Object copyWithoutInternalAcl(ImmutableName _name, ACS _acs, Directory _parent,
    Boolean _toRegisterInManager) throws CreateError;
  
} // Resource

