//Title:        Access Road
//Version:      0.6.0
//Copyright:    2002, 2004, 2007, 2008, 2009 EURL ACCBEE
//Author:       Patrick Thazard ( )
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBaseInterface;


import ARoad0.ErrB.*;
import java.util.Map;

 /**
  * This interface is responsible for managing the canonical rights
  * for all the base objects, called metarights. As any StringRight, it provides rights
  * in a rights network, but metarights have no links to ACS. <P>
  * For each ACS, the metaright may be associated to one StringRightImpl in the rights network 
  * managed by every ACS.<P>
  * A metaright has a location in the matrix (power,field) of metarights. 
  * So, each metaright has:<P>
  * - a power, which is bigger as the power of the right is higher <P>
  * - a field, which is the number of a column where the near metarights are grouped<P>
  * A metaright may be in a floating subnetwork of one or several ACS rights which are
  * not associated to the main stream of metarights, without any capacity to compare
  * the powers. <P>
  * Note: the implementation class is not a subclass of StringRightImpl, to be faster.
  * @version 0.6.0
  */
public interface StringMetaRight extends StringRight {
  /**
   * Get the power, which is the line value in the matrix (power, field) 
   * which helps to locate the metaright.
   * @return a bigger value as the power of the right is higher
   */
   public int getPower();
    
  /**
   * Get the field, which is the colomn value in the matrix (power, field) 
   * which helps to locate the metaright.
   * @return the number of a column where the near metarights are grouped,
   *    from a limited set of ACS.
   */
   public int getField();
  
  /**
  * Get the ACS StringRights which are strictly equivallent to 
  * this MetaStrinRight in each ACS.
  * @return Map where the key is an ACS name as String, and the value 
  *   is an array of StringRightImpls. May be empty.
  */
   public Map getM_ACSRights();
  
  /**
  * Get the StringRights which are equivallent to 
  * this MetaStringRight in one ACS. Applicable to the rights of a closed ACS.
  * @param _aCSname an ACS name
  * @return an array of StringRightImpls. May be empty. Never null.
  */
   public StringRight[] getACSRightsForOneACS(ImmutableName _aCSname);
   
  /**
   * @return true if the metaright is in a floating subnetwork for an ACS,
   *    without any association to the main metarights in the network.
   */
  public boolean getInFloatingSubNetwork();
 
  /**
   * @param _b is true if the metaright is in a floating subnetwork for an ACS,
   *    without any association to the main metarights in the network.
   */
  public void setInFloatingSubNetwork(boolean _b);
  
  /**
   * When some relevant ACS are closed, the link to them remain in this metaright.
   * Caution: a metaright may be used in some ACS directly, wihtout any ACS rights.
   * @return boolean true if the metaright has no ACS rights associated to it.
   */
  public boolean isUnused();  

  /**
   * Add an upper metaright which are strictly greater than this metaright
   * without any intermediate right, following the network of metarights.
   * The two rights may have not the same sens, as for 'modify_security_descriptor'
   * and 'deny_full_control'.
   * It is a bound property.
   * @param _right is directly included in this right. Never null.
   * @exception BaseError if the right is a slave, or the upper right
   *    is not superior to this instance, or the parameter is null.
   */
  public void addUpperRight(StringMetaRight _right) throws BaseError;
  
  /**
   * Remove a upper metaright which are strictly greater than this metaright
   * without any intermediate right, following the network of metarights.
   * Called by the greater metaright finalizeForUser() method.
   * @param _right is directly included in this right. Never null.
   * @param _caller is the object which  calls this method; used to
   *   identify the finalize operation.
   */
  public void removeUpperRight(StringMetaRight _right, Object _caller);
  
} // StringMetaRight