//Title:        Access Road
//Version:      0.6.0
//Copyright:    2002, 2004, 2007, 2008, 2009 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


package ARoad0.gBaseInterface;

import ARoad0.ErrB.*;

import java.io.Serializable;
//import java.util.LinkedList;

 /**
  * This interface is responsible for managing the ACS rights,
  * of all the base objects in an ACS. It handles the links to the
  * upper and lower rights in the rights network, allowing
  * comparison, and the next upper and lower rights. <P>
  * The ACS right may be classed, that is associated to a target class,
  * or typed, that is associated to a type in a target class. If it is typed,
  * it is classed. <P>
  * For instance, in LCA Lotus Notes, 'read_javascript' is a typed right
  * for the target class Actor, where 'javascript' is a type of Actor. 
  * When a right is classed or typed, the class and the type may be superior
  * to the right container (AGO, ACL, PRI, BDG) specification, or may be
  * inferior, and this is specified by the ACS policy. For instance,
  * if an ACL has an actor as target, and if it uses a classed right where
  * the target class is Directory, then the ACL will be effective only if
  * its ACS sets that the ACL specification is superior to the right specification.
  * The ACS may also set that both specifications are applicable.<P>
  * Two-levels right is managed as one parent right which
  * has a detailled description in a set of lower rights.
  * The parent is the unique upper right of these slave rights.
  * Multi-level rights is a generalization where a right is both
  * a parent and a slave. Some rights set may have these tree structure
  * without be marked and displayed as multi-level rights.<P>
  * The right has a generic equivallent called a StringMetaRight,
  * which is used to compare rights from different ACS.
  * @version 0.6.0
  */
public interface StringRight extends BaseObject, Serializable, Comparable, 
        FinalizedObjectForUser, BoundBean {
  /**
   * @return ImmutableName of the ACS that owns this object. 
   *    Empty if it is an embedded right or a metaright.
   */
  public ImmutableName getAcsName();

  /**
   * The full name is unique for the Access Road program.
   * It contains generally more information than getName().
   * @return a copy of the name. Empty for an empty object.
   */
    public ImmutableName getFullName();
    
  /**
   * Get a short name limited to 5 characters if positive, 6 characters if starts with
   * '|' and positive, and 7 if negative, for displaying this right in the views.
   * If right_ is 'read', the nick name is the same.
   * If right_ is 'read&execute', the nick name is 'read&'.
   * If right_ is '|run_under|', the nick name is '|run_u'.
   * If right_ contains '_', the nick name 2 last characters are '_%' where '%'
   * is here the character at the right of the last '_' in right_. For example,
   * 'write_for_directory' gives 'wri_d'.
   * The nick name for a negative right starts with 'd_', if the right
   * starts by 'deny_'. The other parts of the nick name are like for a positive right.
   * If right is 'deny_read', the nick name is 'd_read'.
   * If the right is 'deny_write_for_directory', the nick name is 'd_wri_d'.
   * If the negative right does not start with 'deny_', it is processed like
   * a positive right and the nick name size is limited to 5.
   * This short name is NOT unique for the Access Road program, nor for the ACS if
   * it is an ACS object.
   * @return the nick name. Never null.
   */
    public String getNickName();   
  
  /**
   * @return true if the right is always associated to a class for a target,
   *   as for exemple 'delete_Actor' for the class Actor.
   */
   public boolean isTargetClassed();
   
  /**
   * @return true if the right is always associated to a type for a target
   *   class, as for exemple in LCA Lotus Notes 'read_javascript' for the class Actor.
   */
   public boolean isTargetClassTyped();
    
  /**
   * @return true if the right is a granting right - the opposite one is
   *   a denying right.
   */
   public boolean isGrantingRight();
   
  /**
   * @return true if the right is a parent of lower rights - the slaves -
   *   in a multi-level right.
   */
   public boolean isParent();
    
  /**
   * A slave right has a strong link with a parent right.
   * @return true if the right has an unique upper right - the parent -
   *   in a multi-level right.
   */
   public boolean isSlave();   
    
  /**
   * Set the right as parent of lower rights - the slaves -
   *   in a multi-level right. There is no inverse method.
   * This method has to be used before the set of the lower rights.
   * Caution: some lower rights may be not slaves, so this instance may be not parent
   * for all its lower rights.
   * @param _b is true to set the right as parent
   * @exception BaseError if there are lower rights.
   */
   public void setParent(boolean _b) throws BaseError;
    
  /**
   * Set the right an unique upper right - the parent -
   *   in a multi-level right. There is no inverse method.
   * @param _b is true to set the right as slave
   * @exception BaseError if there is not exactly one upper right,
   *    or if it is not a parent.
   */
   public void setSlave(boolean _b) throws BaseError;  
   
  /**
   * Set that the right is desactivated. Used to create a list of flat rights
   * with one single desactivated root. 
   * @param _b is true to desactivate the right
   */
   public void setDesactivated(boolean _b);
    
  /**
   * Get if the right is desactivated.
   */
   public boolean isDesactivated();
   
  /**
  * If the right is not typed, the result contains the kernel right only.
  * If the right is typed (for ACS rights only), the result contains the kernel right
  * and the target class type.
  * @return the right to apply
 */
  public String getRight();
    
  /**
   * @return the internal key
   */
  public int getInternalKey();
  
    /**
    * @return the class from which the type is provided. Null if the right is not classed. 
    */
  public String getTargetClass();

    /**
    * @return the target typed right. Null if the right is not typed or if it is
    *   a metaright.
    */
  public String getTargetClassType();

  /**
  * @param is the kernel right to apply.
  * @return void
 */
  //public void setKernelRight(String _st);

    /**
    * @param Class from which the type is provided. Null if the right is not typed. 
    * @return void
    */
  //public void setTypingClass(Class _c);

    /**
    * @param String which is the target class typed right, or null if 
    *    the right is not typed.
    * @return void
    */
  //public void setTargetClassType(String _t);
  
  /**
   * Define the set of upper rights which are strictly greater than this right
   * without any intermediate rights, following the network of rights in the ACS.
   * @param StringRight[] is an array of connected rights which include
   *    this instance right. May be null.
   * @return void
   */
  //public void setUpperRights(StringRight[] _l_rights) throws BaseError;

  /**
   * Define the set of upper rights which are strictly lesser than this right
   * without any intermediate right, following the network of rights in the ACS.
   * @return StringRight[] is an array of rights which are directly included 
   *    in this right. May be null.
   */
  //public void setLowerRights(StringRight[] _l_rights) throws BaseError;
  
  /**
   * Get the set of upper rights which are strictly greater than this right
   * without any intermediate right, following the network of rights.
   * @return is an array of connected rights which are strictly greater than
   *    this right. Never null but may be empty.
   */
  public StringRight[] getUpperRights();

  /**
   * Get the set of lower rights which are strictly lesser than this right
   * without any intermediate right, following the network of rights.
   * @return is an array of rights which are strictly lesser than 
   *    right. Never null but may be empty.
   */
  public StringRight[] getLowerRights();

  /**
   * Get the global Access Road right which is a generic description of this ACS right.
   *    Metarights are used in the rights comparision and analysis to
   *    compare rights from different ACS.
   * @return as a generic description of the right. Is null
   *   only when the instance is embedded in a StringMetaRightImpl.
   */
  public StringMetaRight getEorMetaRight();

 /**
  * Get the StringRights chain where this right is one end and _right is
  * the another end, if the chain exists.
  * @param _right is the other end of the chain of rights to get.
  * @return is a chain of linked rights. May be null.
  */
  //public StringRight[] getRightsChain(StringRight _obj);
  
  /**
   * @return true if the instance is equals to an empty instance
   */
  public boolean isEmpty();

    /**
     * @return a comment which is associated to the right. May be null.
     */
    public String getComment();
  
    /**
     * @param _s is a comment which is associated to the right.
     */
    public void setComment (String _s);  
 
 /**
  * Get a StringRights increasing chain where the two arguments are at the two ends,
  * if the chain exists.
  * The lesser right is the first item in the returned LinkedList. 
  * It is an increasing chain, so rights are greater when the index raises.
  * Return the first increasing chain which is found, but it may exist several chains
  * in the rights network.
  * Return null if the comparison (compareTo method) returns 0, or if there is no result.
  * This method does not use a double up-down search, because rights have more
  * lower rights than upper rights, and so, a single raising search is generally faster.
  * Caution: to ensure performance, if there are more than 20 increasing chains
  * to analyze from the lesser right, an exception is thrown. This should never be,
  * and it indicates that the design of rights is very poor. 
  * @param _rightOne is one end of the chain of rights to get. Not null.
  * @param _rightTwo is the second end of the chain of rights to get. Not null.
  * @exception BaseError if an argument is null, if they have not the same class,
  *    if there are more than 20 increasing chains to analyze from the lesser right
  * @return a chain of linked increasing rights, from the index 0.
  *   May be null, or its minimal lenght is 2.
  */
  //public LinkedList getFirstIncreasingRightsChain(
    //      StringRight _rightOne, StringRight _rightTwo) throws BaseError;    
    
    /**
     * Compares the two StringRights.
     * @param _obj to compare
     * @return true if the properties are the same
     */
    public boolean equals(Object _obj);

    /**
     * @return is the hashcode value for this instance. May be null.
     */
    public int hashCode();

    /**
     * Compare two StringRight instances through the internal key
     * and, if not null, acsname comparisons, whatever the right, sens, power and field.
     * All metarights are superior to any acsright.  The comparison 
     * of the rights xx and deny_xx returns 0.
     * @param _o to compare
     * @exception ClassCastException - if the specified object's type
     *  prevents it from being compared to this current Object
     * @return 0 if the rights have the same internal key and acsname,
     *  a negative integer if this object is lesser than the argument, 
     *  and a positive integer if this object is greater than the argument.
     */
    public int compareTo(Object _o) throws ClassCastException;
  
  /**
   * Get a clone of the instance.
   * Caution: the metaright meta_ and the comment are not copied, but metaRight_ 
   * is copied and the upper/lower rights references are copied in the arrays.
   * Used by ACS.closeAclEntriesTo() when the links to other ACS have to 
   * be cut and cloned.
   * @exception InternalError if CloneNotSupportedException
   * @return is a cloned instance of StringRightImpl.
   */
  //public Object clone();
    
   /**
   * This method removes the StringRight from the ACS list, from the metaright,
   * from the lower and upper rights. If the instance is a parent, all slaves
   * are also finalized. 
   * @exception UpDateError if the ACS does not known the right (if
   *   it is a clone, typically), if the metaright is wrong,
   *   or if a lower or upper right is not correct.
   */
  public void finalizeForUser() throws UpDateError;    
    
   /**
   * @return getRight()
   */
  public String toString();
  
} // StringRight