//Title:        Access Road
//Version:      0.6.0
//Copyright:    2002, 2004, 2007, 2008, 2009 EURL ACCBEE
//Author:       Patrick Thazard ( )
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gBaseInterface;

import ARoad0.ErrB.BaseError;
import ARoad0.ErrW.DialogError;

import java.util.List;
import java.util.Map;

/**
* This interface is responsible for managing
* the views in the generic base through specific methods.
* @version 0.5.1
* @see ARoad0.gBase.BaseManagerImpl
*/
public interface ViewManager extends BaseBeanManager {

    /**
    * Open all known views.
    * @exception BaseError if a view is not known or ClassNotFoundException,
    *   a view object belongs to a closed ACS (with the relevant closed
    *   ACS names that may be get through BaseError.getL_Exception()),
    *   or IOException, or if a view is open at the start of the method (if so,
    *   close it before a new call to this method).
    * @exception DialogError if a view contains closed-ACS objects.
    */
    public void openAllViews() throws BaseError, DialogError;

    /**
    * Open a registered and closed view from a file in the current data directory.
    * @param _name ImmutableName of the view to open. Its file name is built
    * with the two last components of the name, or with the sole component.
    * @exception BaseError if the view is not known, the view is open,
    *   the open view has not the same name,
    *   or ClassNotFoundException, IOException.
    * @exception DialogError if the view contains closed-ACS objects,
    * and closed-ACS names are then put in the DialogError Exceptions List,
    * or if the view contains an object which is no more known by its ACS,
    * and removed object name is then put in the DialogError Exceptions List.
    * @return ViewInBase: the open view.
    */
  public ViewInBase openView(ImmutableName _name)
    throws BaseError, DialogError;


    /**
    *  Save all the open views in some files (one per view)
    *  in the default data directory.
    * @exception BaseError if viewName.lastComponent() is null,
    *   FileNotFoundException or IOException.
    */
  public void saveAllViews() throws BaseError;

    /**
    *  Save and close all the views, even if it is empty.
    *  All open views, but not the closed ones,
    *  are saved in some files (one per open view) in the default data directory.
    * @exception BaseError if viewName.lastComponent() is null,
    * FileNotFoundException or IOException.
    */
  //public void saveAndCloseAllViews() throws BaseError;

    /**
    *  Close an open view and save it before.
    *  The view is still known in ViewManager.
    *  In this method, THE OBJECT _VIEW IS SET TO NULL.
    *  After the call of this method, SET the object _view TO NULL.
    * @param _view ViewInBase to be saved, closed and finalized.
    * @exception BaseError if _view is not open,
    *   ViewName.lastComponent() is null,
    *  or FileNotFoundException and IOException.
    */
  public void saveAndCloseOpenView(ViewInBase _view) throws BaseError;

    /**
    *  Save an open view without close it.
    *  The view instance is not modified.
    * @param _view ViewInBase to be saved.
    * @exception BaseError if _view is not open, ViewName.lastComponent() is null,
    *  or FileNotFoundException and IOException.
    */
  public void saveOpenView(ViewInBase _view) throws BaseError;

  /**
  * Return the names of the open or closed views.
  * @return unmodifiable list of view names
  */
  public List getL_ViewNames();

  /**
  * @return unmodifiable map of non-null instances of the open views
  */
  public  Map getEorM_OpenViews();

  /**
  * @return long
  */
  public long getSerialVersionUID();

  /**
  * @param _name ImmutableName of a view
  * @return boolean true if _name is known as a view name.
  * The view may be open or closed.
  */
  public boolean knowsView(ImmutableName _name);

  /**
  * @param _name ImmutableName of a view
  * @return boolean true if _name is known as a view name
  * and the view is open.
  */
  public boolean isOpenView(ImmutableName _name);

  /**
  * The view name is removed from the view list in ViewManager.
  * If it is open, the view is closed and saved before.
  * The view saving file is not deleted, but it could be (or removed) by the user.
  * Then, ViewManager.openView(this view) can't work.
  * @param _name ImmutableName of the view to remove.
  * @exception BaseError if the view is unknown,
  *   viewName.lastComponent() is null,
  *  or FileNotFoundException and IOException.
  */
  public void removeView(ImmutableName _name) throws BaseError;

    /**
    *  Close an open view without save it.
    *  The file is not deleted.
    * @param _view ViewInBase to be closed and finalized.
    * @exception BaseError if _view is not open.
    */
  public void closeOpenView(ViewInBase _view) throws BaseError;


} // ViewManager