//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls


/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation, purely the version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details (gpl.html at the root of Access Road directory).
This license is applicable for 20 years since the publication of the program.
The applicable laws are the French ones, and the relevant court is at Nanterre
(92), France.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation, Inc.,
675 Mass Ave, Cambridge, MA 02139, USA.
*/

package ARoad0.gBaseInterface;

/**
 * This interface is a Resources folder belonging to one ACS,
 * responsible for being a privilege target, an alias or an alias reference. <P>
 * The default behavior in the rights propagaton is the RIGHTS_PASSING_OFF_TO_MEMBERS mode.
 * This means that the VirtualFolder transfers none of its rights as target to its member.
 * In the other mode RIGHTS_PASSING_ON_TO_MEMBERS, the ACS as set this mode for its
 * virtual folders, and all the virtual folder rights are transfered to the members.<P>
 * There are two main types of virtual folders: <P>
 * - for Resource members which are not EligibleParties ('RES_VFOLDER' type) <P>
 * - for Resource members which are EligibleParties ('EP_VFOLDER' type) <P>
 * Trees of virtual folders are possible. They have no rights inheritance in the
 * generic gBase, while an AcsAddon may override the hook methods. <P>
 * A virtual folder may be included in: <P>
 * - a virtual tree of Resource Folders, which is not the Resources tree in the ACS, <P>
 * - or a virtual tree of Eligible Party Folders. <P>
 * The virtual folder manages:<P>
 * - its members, as any folder, and they are Basics of one ACS; they are not
 * deleted when this folder is deleted, <P>
 * - its children, as any node, and they are virtual folders (no Basic) which are deleted
 * when this node is deleted, <P>
 * - privileges, as any target.<P>
 * Each type allows only one folders tree and then, one folder root in an ACS.
 * It is possible to force each member to be member at most
 * one time in all the folder tree (EP or Resource tree).<P>
 * There is no link between the members of a virtual folder
 * and the members of the folder children.
 * There is no link between the privileges of a virtual folder
 * and the privileges of its children.
 * But, as any node, the deleting of the virtual folder implies the deleting of its
 * children.<P>
 * The typed privileges are applied to the virtual folder types.
 * Then, they may be used to defined default virtual folder rights.<P>
 * There are two flows of data to manage the inherited Privileges: <P>
 * - one at the privilege creation/deletion, and this is not done in the generic gBase package, <P>
 * - one at their sources editing, <P>
 * This interface handles the second case,
 * to update the children of the VirtualFolder which is the target of the Privilege.
 * Caution: no equals() nor clone() methods.
 * See the <a href="{@docRoot}/copyright.html">Copyright</a>.
 * @version 0.7.0
 */
public interface VirtualFolder extends ImmutableVirtualFolder, Folder, Node,
  PrivilegeTarget {

    /**
    * Resets the inherited Privilege rights after an user edition.
    * Called by the master parent when there is a change in the effective rights
    * of an inherited Privilege. This is only when there is a change in the sources.
    * Contrary to AclEntry, the rights are immutable and there is no conditional group.
    * Not described in the BeanInfo since the user does not call it.
    * Note: the creation of the inherited Privilege in a child is not handled by
    * the generic gBase package, while the rights updating of the inherited Privilege
    * is handled.
    * @param _lpri inherited Privilege with updated rights. Never null.
    */
  public void resetInheritedPrivilegeRights(PrivilegeForLinks _lpri);

    /**
    * Returns the Privilege to use as inherited Privilege in a child.
    * Empty method which returns the argument. Has to be implmented into an AcsAddon
    * handling the privilege inheritance.
    * @param _lpri the initial Privilege which may be given to the child
    * @param _type type of a child, as VirtualFolder
    * @return the inherited Privilege if the child inherits this initial Privilege, or null
    */
  public PrivilegeForLinks selectInheritedPrivilege(PrivilegeForLinks _lpri, String _type);

    /**
    * Resets the inherited Privilege rights in the children, if the Privilege
    * inheritance is applicable.
    * There are two flows of data to manage the inherited Privileges: <P>
    * - one at the privilege creation/deletion, and this is not done in the generic gBase package, <P>
    * - one at their sources editing, <P>
    * This method is called for handling the second case,
    * to update the children of the VirtualFolder which is the target of the Privilege.
    * Not described in the BeanInfo since the user does not call it.
    * @param _lpri inherited Privilege with updated sources
    */
  public void resetInheritedPrivilegeRightsOfChildren(PrivilegeForLinks _lpri);
    
} // VirtualFolder