//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gDMak;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.*;
import java.util.Map;
import java.util.Iterator;
import java.util.Arrays;

import ARoad0.Gui1.*;
import ARoad0.Gui2.*;
import ARoad0.gWork.*;
import ARoad0.ErrW.*;
import ARoad0.ErrB.BaseError;

import ARoad0.gBase.NameImpl;
import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.ISFactoryUtilityImpl;
import ARoad0.gBase.ACSImpl;

import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.ACS;
import ARoad0.gBaseInterface.ViewInBase;
import ARoad0.gBaseInterface.ImmutableNameISTree;

/**
 * An action that closes an ACS or a view in the base,
 * closes its main frame if open, all its open frames and its graphical objects
 * in other open frames, and removes its node in the explorer.
 * When an ACS is closed, all the views which contain an element from this ACS
 * are closed, and all the ACS which are direct or indirect children of this ACS
 * are closed. <P>
 * Manages closing commands from 5 sources: <P>
 * - direct closing of the main frame of an acs or a view, from the frame
 *    listener <P>
 * - direct closing of the tab of a view in a views tabbed pane, from
 *    the closing button <P>
 * - from 'Close' in the main menu when the main frame is activated <P>
 * - from 'Close' in the main menu when in the explorer, the ACS or view node
 *    is selected <P>
 * - from ActionRemove, ActionOpenAllFullViews, ActionOpenAllNoThanViews and
 * ViewTreeBaseListenerImpl, which call directly the close methods. <P>
 * For views in the 'All views' or 'NoThan views' tabbed pane, there is 3 types of closing: <P>
 * - close the ViewInBase in the program and then in the explorer <P>
 * - close the ViewInBase tab in the tabbed pane without ViewInBase closing <P>
 * - close the GraphicView in the tab to constuct a new one, without tab closing.
 * @see ActionUtilities
 * @version 0.7.0
 */
public class ActionClose extends  AbstractAction implements TabClosingHandler {

  /* the only instance of the class, following singleton pattern */
  private static ActionClose instance__ = new ActionClose();

  // variables for the gui
  private static ViewInBase_Facade  ViewBase__ = ViewInBase_Facade.getInstance();

  /**
  * with this private constructor, the compiler won't generate
  * a default public constructor
  */
  private ActionClose() {
    super("Close");
  }
  /**
   * constructor()
   * @return ActionClose
   */
  public static synchronized ActionClose getInstance(){
    return instance__;
  } // getInstance()


  /**
   * Manages closing commands from three sources:
   * - direct closing of the main frame of an ACS or a view, from the 'All views' tab,
   *    at the frame listener
   * - from 'Close' in the main menu when the main frame of an ACS or a view is activated
   * - from 'Close' in the main menu when, in the explorer, an ACS or a view node
   *    is selected
   * @param _ev ActionEvent
   */ 
   public void actionPerformed(ActionEvent _ev) {

    // no action if the explorer is currently used for an object selection
    // or if there is an user dialog currently
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
    
    // initialization
    Object[]               l_analyzedEvent = new Object[3];
    // userChoice[0] is true if closing is a user request
    // userChoice[1] is true if closing and saving are user requests
    boolean[]              userChoice = new boolean[2];


    l_analyzedEvent = ActionUtilities.getInstance().analyzeActionEvent(_ev);
    
    if (l_analyzedEvent[0] == null ) {
            // display a warning message
            DialogBox.showErrorMessage(
              "No selection or unworkable selection.");
            l_analyzedEvent = null;
            return;
    } // if
    
                      // for testing and tracking
                //System.out.println("in ActionClose.actionPerformed()"+
                  //"l_analyzedEvent: ");
                //for (int i = 0 ; i < l_analyzedEvent.length ; i++) {
                  //System.out.println(l_analyzedEvent[i].toString());
                //}

    if (l_analyzedEvent[0] == null ||
      (!((String)l_analyzedEvent[0]).equals("acs") &&
      !((String)l_analyzedEvent[0]).equals("view") )
      ) {
        // display a warning message
        DialogBox.showErrorMessage(
          "Only selected ACS or view can be closed.");
        return;
    }

    // whatever the event source, request to the user
    // if he wants a closing and a saving of the base object, knowning that
    // the main frame of the base object is or has to be closed,
    // but not the view tab if open.
    // userChoice is a boolean[2]: first (0) is true if closing is
    // a user request, second is true if closing and saving are user requests.
    if (((String)l_analyzedEvent[0]).equals("acs"))
      userChoice = DialogBox.showClosingAndSavingDialog(
        (ImmutableName)l_analyzedEvent[1]);

    else if (((String)l_analyzedEvent[0]).equals("view"))
      userChoice[0] = DialogBox.showConfirmDialog("Do you want to close "+
        ((ImmutableName)l_analyzedEvent[1]).toString()+" in the GUI?", 
        "Say NO to close only the main frame");

    if (!userChoice[0] ) {
      // dont close the ACS or the view in the explorer, close the main frame
      // but not the view tab (which may be closed by the tab button)  
      if (l_analyzedEvent[2] instanceof JInternalFrame)
          closeMainFrameInGUI((ImmutableName)l_analyzedEvent[1],
          (JInternalFrame)l_analyzedEvent[2], true);
      // else, l_analyzedEvent[2] is the JPanel of a view in the 'All views' frame
      // so noop...
      //else closeMainFrameInGUI((ImmutableName)l_analyzedEvent[1], null, true);
      l_analyzedEvent = null;
      return;
    }
    // closes the ACS and save it if userChoice[1] is true,
    // closes the view tab and the view, without saving of the GUI configuration:
    // if it is a view, userChoice[1] is false
    else close((ImmutableName)l_analyzedEvent[1], userChoice[1]);

    l_analyzedEvent = null;
    userChoice = null;
  } // actionPerformed(ActionEvent) 

  /**
   * Manages a tab closing in the 'All Views' frame.
   * Called by the BeamerManager.initializeViewTabComponent() and
   * closeAllViewsFrame().
   * @param _tab is the name of the BaseObject displayed in the tab
   */
  public void tabClosingPerformed(ImmutableName _tab) {
      closeOneAllViewsTab(
              _tab, 
              (JPanel) InternalFrameModel.getJComponentsOfBaseObjects().get(_tab));
  } // tabClosingPerformed()
   
  /**
  * Closes an open ACS or an open view, and saves it if necessary.
  * Called by actionPerformed(), ActionRemove, ActionExit and 
  * ViewTreeBaseListenerImpl.propertyChange().
  * Calls ACS_Facade.saveAndCloseOpenACS/View(),
  * ACS_Facade.closeOpenACS/View() and closeInGUI().
  * If the 'Architecture' view is opened during an ACS closing, updates this view.
  * @param _name of the entity to close
  * @param _withSaving boolean true if the user requests a saving with the closing
  */ 
  public synchronized void close(ImmutableName _name, boolean _withSaving) {
    // variables
    ACSImpl                 acs = (ACSImpl)ACS_Facade.getOpenACS().get(_name);
    ViewInBase              view = (ViewInBase)ViewBase__.getAllStandardOpenViews().get(_name);
    if (_name.equals(ActionOpenArchitecture.ARCHI_VIEW_NAME))
        view = ViewBase__.getArchitectureView();
    JInternalFrame          frame;
    JPanel                  pane = null;
    //ImmutableNameISTree     nametree = null;
    //List<ImmutableName>     l_children = null;
    
                      // for testing and tracking
                    //System.out.println("ActionClose.close() for "+_name);
    
    frame = (JInternalFrame)InternalFrameModel
                                  .getMainFramesOfBaseObjects().get(_name);
    // if a view is not in a frame, it may be in a tab of the 'All views' frame
    if (frame == null) pane = (JPanel)InternalFrameModel
                                        .getJComponentsOfBaseObjects().get(_name);

    // a tied child ACS cannot be closed alone
    if (acs != null) {
        ImmutableNameISTree nametree = BaseManagerImpl.getInstance().getISTree(
                        BaseManagerImpl.getInstance().getISofName(acs.getName()));

        if (Arrays.asList(nametree.getLinks(acs.getName())).contains(
              ISFactoryUtilityImpl.IS_TIED_CHILD)) {
                   frame = null;
                   pane = null;
                   acs = null;
                   nametree = null;
                   DialogBox.showErrorMessage("Sorry, this ACS is tied to its parent. To "+
                           "close it, you must close its parent.");
                   return;
        } // if
        nametree = null;
    } // if

    // closes in the GUI
    if (frame != null) closeInGUI(_name, frame);
    else if (pane != null) closeInGUI(_name, pane);
    else closeInGUI(_name, null); // if it is open only in the explorer

    // closes the ACS or view in the base, even if frame and pane are null
    try {
        
        if (_withSaving) {
          if (acs != null) {
              ACS_Facade.saveAndCloseOpenACS(acs);
            //} // else

          } // if (acs != null)
          else {
              ViewBase__.saveAndCloseOpenView(view);
          }
        // without saving; exception handling with an user message
        } else {
          if (acs != null) {
              // close also the open views with some objects from this ACS,
              // and there is no user message for this.
              ACS_Facade.closeOpenACS(acs);
          }
          else {
              ViewBase__.closeOpenView(view);
          }
        } // else
 
    } catch (BaseError b){
        DialogBox.showOkMessage("The entity closing has just been done"+
          " by another thread, perhaps with saving. An exception has been thrown: "+
          b.toString() );
    } // catch
    
    acs = null;
    view = null;
    frame = null;
    pane = null;
    //l_children = null;
    //nametree = null;
  } // close()

  /**
  * Closes in the GUI an open or closed ACS or an open view, but no closing in the base.
  * The closing is done in the explorer and the main frame for an acs or a view,
  * and for a view only, it is done in the tab and the 'See why' frame.
  * For an ACS closing, the views using some of this-ACS objects listen the ACS closing,
  * and for their proper closing, they fire an event to ViewTreeBaseListenerImpl
  * that calls this method.
  * Removes all the TreeBaseListeners associated to the ACS or view.
  * Called by close(), ACSTreeBaseListernerImpl.changeOpenACS()
  * and ViewTreeBaseListenerImpl.propertyChange().
  * The sketch view does not call this method. Its closing uses InternalFrame,
  * since it is still registered in ViewModel when the GUI is closed through
  * the ViewTreeBaseListenerImpl.
  * Uses ACSTreeUtilities, TreeManager and ViewModel.
  * Calls closeMainFrameInGUI() if _compo is a JInternalFrame, and
  * closeOneAllViewsTab() if _compo is a Panel.
  * @param _name of the entity to close
  * @param _compo is the main component that displays the entity.
  *   Null when only the explorer as to be updated, or it is 
  *   a JInternalFrame or a JPanel (for view JTabbedPanes).
  */ 
  public synchronized void closeInGUI(ImmutableName _name, JComponent _compo) {
    // variables
    ACS acs = (ACS)ACS_Facade.getOpenACS().get(_name);

    // if it is not an open ACS nor a view name, return
    //if (acs == null && !ViewBase__.knowsView(_name) ) {
      //          System.out.println("in ActionClose.closeInGUI(),"+
        //          " it is not an open ACS nor an open view.");
      //return;
    //}
    
    if (_compo instanceof JInternalFrame) {
        if (acs == null)
            closeMainFrameInGUI(_name, (JInternalFrame)_compo, true);
        else
            closeMainFrameInGUI(_name, (JInternalFrame)_compo, false);
    } else if (_compo instanceof JPanel)
        closeOneAllViewsTab(_name, (JPanel)_compo);

    // removes the entity node in the explorer
    ACSTreeUtilities.removeACSorViewNode(_name);
    // resets the beamer
    if (InternalFrameModel.withBeamer())
        BeamerManager.getInstance().propertyChange(
                new PropertyChangeEvent(this, "objectClosing", null, null));

    // removes all the TreeBaseListeners associated to the ACS or view,
    // except for the main frame, which has been already processed
    try {
      TreeManager.removeAllTreeBaseListenerFor(_name);
    } catch (SmallError s) {
        DialogBox.showOkMessage("On-going entity closing,"+
          " but the treeBaseListeners are not all removed."+
          " An exception has been thrown: "+
          s.toString() );
    }

    acs = null;
  } // closeInGUI()
 
  /**
   * Manages the closing of all the tabs in the 'All Full views' tabbed pane, then
   * the closing of the JInternalFrame if it has not been done.
   * Called by InternalFrameUtilities.addAllFullViewsFrameListener().
   */
  public void closeAllFullViewsFrame() {
      ImmutableName name = null;
    
                      // for testing and tracking
                    //System.out.println("start of ActionClose.closeAllFullViewsFrame()");

      for (Iterator ite = ViewBase__.getOpenFullViews().keySet().iterator(); ite.hasNext(); ) {
          name = (ImmutableName)ite.next();
          // no tab if the view is closed in the FullViews frame
          if (InternalFrameModel.getJComponentsOfBaseObjects().get(name) != null)
                tabClosingPerformed(name);
      }
      ActionOpenAllFullViews.getInstance().finalizePropertyEditors();

      // new in 0.7.0
      ViewModel.getAllFullViewsJTabbedPane().removeAll();
      
      name = new NameImpl("<AllFullViews>");
      closeMainFrameInGUI(name, InternalFrameModel.getAllFullViewsFrame(), false);

      if (DialogBox.showConfirmDialog("Do you want to close all the full views, without saving?",
              "Otherwise, they will remain open in the explorer.")) {

          for (Object obj : ViewInBase_Facade.getOpenFullViews().keySet()) {
            close((ImmutableName)obj, false);
          }

      } // if

      name = null;
  } // closeAllFullViewsFrame()
 
  /**
   * Manages the closing of all the tabs in the 'All NoThan views' tabbed pane, then
   * the closing of the JInternalFrame.
   * Called by InternalFrameUtilities.addAllNoThanViewsFrameListener().
   */
  public void closeAllNoThanViewsFrame() {
      ImmutableName name = null;

      for (Iterator ite = ViewBase__.getOpenNoThanViews().keySet().iterator(); ite.hasNext(); ) {
          name = (ImmutableName)ite.next();
          // no tab if the view is closed in the NoThan frame
          if (InternalFrameModel.getJComponentsOfBaseObjects().get(name) != null)
            tabClosingPerformed(name);
      }
      ActionOpenAllNoThanViews.getInstance().finalizePropertyEditors();
      
      // new in 0.7.0
      ViewModel.getAllNoThanViewsJTabbedPane().removeAll();

      name = new NameImpl("<AllNoThanViews>");
      closeMainFrameInGUI(name, InternalFrameModel.getAllNoThanViewsFrame(), false);

      if (DialogBox.showConfirmDialog("Do you want to close all the NoMore-NoLess views, without saving?",
              "Otherwise, they will remain open in the explorer.")) {

          for (Object obj : ViewInBase_Facade.getOpenNoThanViews().keySet()) {
            close((ImmutableName)obj, false);
          }

      } // if

      name = null;
  } // closeAllNoThanViewsFrame()
 
  /**
   * Manages the closing of the 'IS structure' pane, then
   * the closing of the JInternalFrame.
   * Called by InternalFrameUtilities.addStructureFrameListener().
   * Calls ActionOpenGlobalStructure.finalizePropertyEditors() and
   * closeMainFrameInGUI(), TreeManager.removeAllTreeBaseListenerFor().
   * @see ARoad0.Gui1.InternalFrameUtilities#addStructureFrameListener(javax.swing.JInternalFrame) 
   */
  public void closeStructureFrame() {
      ImmutableName name = null;

      ActionOpenGlobalStructure.getInstance().finalizePropertyEditors();

      name = new NameImpl("<Structure>");
      closeMainFrameInGUI(name, InternalFrameModel.getStructureFrame(), false);

        // removes the ISTreeBaseListener
      name = new NameImpl("<BaseManager>");
      try {
          TreeManager.removeAllTreeBaseListenerFor(name);
      } catch (SmallError s) {
            DialogBox.showOkMessage("On-going closing of the IS structure,"+
              " but the treeBaseListeners are not all removed."+
              " An exception has been thrown: "+
              s.toString() );
      }

      name = null;
  } // closeStructureFrame()

  // --------------- Private methods --------------------------------------
  
  /**
  * Closes the tab pane of an open EPRView or NoThanView in the relevant
  * TabbedPane of the GUI, with the graphic view and all the associated objects.
  * Uses a call to the TabButton on the view tab to close it.
  * Removes it in the InternalFrameModel,
  * but no removing of the view node in the explorer,
  * and no removing of the view in the base.
  * Called by closeInGUI(), tabClosingPerformed().
  * Calls InternalFrameModel.getJComponentsOfBaseObjects().
  * @param _name of the entity to close
  * @param _panel that displays the entity objects in a JTabbedPane.
  */
  private synchronized void closeOneAllViewsTab(ImmutableName _name,
    JPanel _panel) {
       // variables
       Map          baseObjectsInFrame = InternalFrameModel.getJComponentsOfBaseObjects();
       JTabbedPane  tPane = null;
       if (ViewBase__.isNoThanViewName(_name))
           tPane = ViewModel.getAllNoThanViewsJTabbedPane();
       else
           tPane = ViewModel.getAllFullViewsJTabbedPane();
       int          index = tPane.indexOfComponent(_panel);
       GraphicView  gview = ViewModel.getGraphicViewOfView(_name);

      // if the pane is closed or the panel is not found, noop
      if (!baseObjectsInFrame.containsKey(_name) || index == -1) {
    
                      // for testing and tracking
                    System.out.println("ActionClose.closeOneAllViewsTab() for "+_name+
                            " but no tab to close with index = "+index+
                            " and baseObjectsInFrame.containsKey(_name): "+
                            baseObjectsInFrame.containsKey(_name));
          
        baseObjectsInFrame = null;
        tPane = null;
        gview = null;
        return;
      } // if
      
      // inform that the base object is implied in an on-going
      // close operation by an ActionClose class. So, the pane closing listener
      // does not invoke ActionClose.
      ViewModel.addOnGoingClosingOnView(_name);

      // close the tab and finalize the GraphicViewPropertyEditor
      // with a call to the TabButton, subclass of ActionListener,
      // which is an inner method class initialized in 
      // BeamerManager.initializeViewTabComponent()
      ((ActionListener)ViewModel.getClosingTabButtonInTabbedPane(_name))
              .actionPerformed(new ActionEvent(this, 0, ""));
      
      // unregister the JPanel for the ViewInBase
      InternalFrameModel.removeBaseObjectJComponent(_name);
 
      // remove the ViewRightsNetworkObserverImpl and the GraphicViewBaseListener
      closeGraphicView(_name);
      
      baseObjectsInFrame = null;
      tPane = null;
      gview = null;      
  } // closeOneAllViewsTab()
  
   //
   // ------------- STATIC METHODS ---------------------------------------
   //
  
  /**
  * Close in the GUI the main frame of an open ACS, an open view, the 'All full views'
  * or 'All NoThan views' frames.
  * Remove it in the InternalFrameModel,
  * but no removing of the view node in the explorer,
  * and no removing of the view in the base,
  * nor, for a view, in its view tab in the 'All views' pane.
  * For a ViewInBase, close the 'See why' frame if it is open, and
  * _closeGraphicView sets if the graphic view is finalized.
  * Called by actionPerformed(), closeInGUI(), closeAllFullViewsFrame(), closeAllNoThanViewsFrame(),
  * ActionOpenAllFullViews, ActionOpenAllNoThanViews and GraphicViewBaseListenerImpl.changeFrameView().
  * Calls InternalFrameModel.getBaseObjectsInMainFrames() and
  * InternalFrameModel.addOnGoingClosingOnBaseObject().
  * @param _name of the entity to close
  * @param _frame that displays all the entity objects. May be null.
  * @param _closeGraphicView true when the entity is a ViewInBase with a GraphicView
  *   to finalize. ActionOpenAll/NoThanViews call with an argument at 'false'.
  */ 
  public static synchronized void closeMainFrameInGUI(ImmutableName _name,
    JInternalFrame _frame, boolean _closeGraphicView) {
      // variables
      Map  baseObjectsInFrame =
                            InternalFrameModel.getBaseObjectsInMainFrames();

      // if the main frame is closed and it is not one of the general frame, noop
      if (!baseObjectsInFrame.containsKey(_name) && 
              _frame != InternalFrameModel.getAllFullViewsFrame() &&
              _frame != InternalFrameModel.getAllNoThanViewsFrame() &&
              _frame != InternalFrameModel.getStructureFrame()) {
          
          // for testing and tracking        
          System.out.println("in ActionClose.closeMainFrameInGUI(),"+
                  " for the name: "+_name+", the frame is unknown by InternalFrameModel");
          
                        baseObjectsInFrame = null;
                        return;
      }
      
      // inform InternalFrameUtilities that the base object is implied in an on-going
      // closing operation by an ActionClose class. So, the frame closing listener
      // does not invoke ActionClose.
      InternalFrameModel.addOnGoingClosingFromActionClass(_name);

      // close this frame
      // the closing calls the inner class Gui1.InternalFrameUtilities.InternalFrameAdapter()
      try {
        _frame.setClosed(true);
      } catch (PropertyVetoException p) {
        System.out.println("in ActionClose.closeMainFrameInGUI(),"+
          " the frame closing fires a PropertyVetoException: "+
          p.toString());
      }
      
      if (_closeGraphicView) closeGraphicView(_name);

      baseObjectsInFrame = null;
  } // closeMainFrameInGUI()
  
  /**
  * Finalize the ViewRightsNetworkObserverImpl and the GraphicViewBaseListener
  * for a ViewInBase, even if it is unknown by the base.
  * Not for the sketch view.
  * Called by closeMainFrameInGUI() and closeOneAllViewsTab().
  * @param _name of the ViewInBase
  */ 
  public static void closeGraphicView(ImmutableName _name) {
      
    GraphicView gview = ViewModel.getGraphicViewOfView(_name);

    try {
        if (_name.equals(ActionOpenArchitecture.ARCHI_VIEW_NAME)) {
            ViewModel.removeViewRightsMediator( (ViewInBase)
                ViewBase__.getArchitectureView());
            ViewModel.removeGraphicViewBaseListener( (ViewInBase)
                ViewBase__.getArchitectureView());
        }
        else {
            ViewModel.removeViewRightsMediator( (ViewInBase)
                ViewBase__.getAllStandardOpenViews().get(_name));
            ViewModel.removeGraphicViewBaseListener( (ViewInBase)
                ViewBase__.getAllStandardOpenViews().get(_name));
        }
    } catch (SmallError s) {
        s.printStackTrace();
    }

    gview.finalizeForGui();
    gview = null;
  } // closeGraphicView()
  
  
  /**
  * Close the 'See why' main frame of an open view in the GUI, but not in a JTabbedPane.
  * Remove it in the InternalFrameModel,
  * but no removing of the view main frame, and view node in the explorer,
  * and no removing of the view in the base, nor the 'See why' panel in one of the two
  * 'All views' frames.
  * No action if the name is not a view name or no 'See why' frame.
  * Called by closeInGUI() and GraphicViewBaseListenerImpl.changeFrameView().
  * Use Desktop.getCurrentFrames().
  * @param _name of the view
  * @return true if the 'See why' frame has been found
  */ 
  public static synchronized boolean closeSeeWhyFrameInGUI(ImmutableName _name) {
    // variables
    JInternalFrame[]        l_frames = ARoad0.Gui1.Desktop.getCurrentFrames();
    boolean                 result = false;

    // if it is not a registered view name, return
    if ( _name == null || !ViewBase__.knowsView(_name) ) {
                System.out.println("in ActionClose.closeSeeWhyFrameInGUII(),"+
                  " it is not a view name or an open view.");
      return result;
    }

    // search and close the 'See why' view frame
    for (int i = 0; i < l_frames.length; i++) {
        
        if (l_frames[i].getTitle().equals(_name.toString()+ " 'See why' text") ) {
              try {
                l_frames[i].setClosed(true);
              } catch (PropertyVetoException p) {
                System.out.println("in ActionClose.closeSeeWhyFrameInGUI(),"+
                  " a 'See why' view frame closing fires "+
                  "a PropertyVetoException: "+p.toString());
              }
              result = true;
              break;
        } // if

    } // for

      // for testing and tracking
    //System.out.println("in ActionClose.closeSeeWhyFrameInGUI(), result: "+result);
    
    l_frames = null;
    return result;
  } // closeSeeWhyFrameInGUI()  
 
  
} // class ActionClose