//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gDMak;

import javax.swing.*;
import java.awt.event.*;
import java.util.*;

import ARoad0.Pattern.*;
import ARoad0.Gui1.*;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrW.*;
import ARoad0.gWork.ACS_Facade;
import ARoad0.gWork.ViewInBase_Facade;

import ARoad0.gBaseInterface.Name;
import ARoad0.gBaseInterface.ACS;

import ARoad0.gBase.NameImpl;

/**
 * An action that imports in the base an unknown access control system from a file.
 * After the import operation, a saving updates the initial file of the imported ACS.
 * The imported ACS must have the same ACSImpl version than the current ones.
 * This class manages all the dialog boxes required for the user,
 * controls the inputs, imports the ACS in the base from its file,
 * helps the user to create the external links to the current open ACS,
 * then displays the imported ACS in the explorer and the beamer.
 * <P>
 * An ACS has to belong to an information system (IS), named below 'IS-name'.
 * This name has only one String component.
 * The imported ACS name has the String form ' IS-name:: aaaa:: bbbb:: cc::  ddddddddddd:: eeee:: ',
 * where xxx is both a component of the name and a component in the IS nametree.
 * The ACS file name depends on all the components of the ACS name.
 * This may forbids the importation of an ACS if there is a current ACS having 
 * the same file name, or if there is an IS node or area in the same case.
 * It is not possible to change the name of the imported ACS or the name of a current ACS.
 * <P>
 * To import an ACS, it has to be closed before, into its source Access Road instance.
 * It is not necessary to remove it from this source instance, and the ACS has not
 * to be prepared for the export.
 * The imported ACS file is simply copied from its source Access Road working directory
 * to the working directory of the targeted Access Road instance. For importing 
 * in its Access Road instance, the user has to enter the full name of the ACS.
 * An error is displayed if the associated file, derived from the ACS name, is not find
 * into the working directory. To import two ACS, one being the child of the second one,
 * the user has to import the parent first.
 * <P>
 * If the information system is unknown, is is added into the IS list as a hub node.
 * The ACS parent name is compared to the current nodes of the ACS IS.
 * If the parent name does exist in the IS, the ACS is registered under this node if
 * there is no child having the same name.
 * If there is no chain of nodes in the IS to match to the ACS parent name,
 * a chain of nodes derived from the ACS name, is created with logical nodes.
 * During the importation, two following IS nametree properties are missed. They are
 * the Composite and Alternate relations to an ACS or subACS in the IS nametree.
 * An imported ACS has no composite and no alternate, and it is not possible to edit
 * them after the import. The other properties of the ACS node in the source IS nametree
 * are copied into the target IS nametree.
 * <P>
 * In the imported ACS, all the inner properties are restored, including
 * its structure, its behavior and its set of ACSobjects. This includes the ACS roots
 * in the parent ACS, if the parent is an ACS. Otherwise, the roots are deleted during
 * the import. If the parent is an ACS, it has to comply with the structure and behavior
 * of the parent ACS source, so that the roots action may be applied without error.
 * However, errors in the creation of roots do not stop the import operation.
 * This is why the set of roots may be partially restored, or not at all.
 * <P>
 * This method never opens the external ACSObjects of the imported ACS,
 * and its closed Bridges, AclEntries and linked Privileges.
 * It rather returns a map containing the description of these closed objects.
 * The user may chooses, in the class ActionImportACS,
 * to create some of these closed objects. The returned map described each Bridge,
 * AclEntry or Privilege, including the names of the rights. Even if they are ACS rights
 * from the imported ACS, they are always fully defined. The import operation
 * resets the links between the ACS rights and the current metarights of the running
 * Access Road instance.
 * <P>
 * As the inter-ACS links are detected through the objects names,
 * some 'ghosts' are possible in specific configurations. For instance,
 * if the ACS 'rrr' has been closed and removed while it had a bridge with the open ACS
 * 'aaa', then 'aaa' keeps the memory of this inter-ACS link, through the name
 * of a 'rrr' Actor. If the imported ACS is also named 'rrr' with a same-named Actor,
 * the ACS 'aaa' will try to restore a 'ghost' bridge with the new ACS.
 * This may also occur on AclEntries or Privileges external links.
 * This feature is applicable in the two sens, both for bridges, AclEntries and Privileges.
 * For the bridges, it is suffisant to have the bridge description in one of the
 * relevant ACS to recreate the bridge. For the AclEntries and the Privileges,
 * the full restoration data is only into the owner of the AclEntry or Privilege.
 * The user should review the external links of the imported ACS, to delete
 * these ghost links after their creation by the import procedure, or to keep them
 * after their creation, or to recreate them if they have not been created by
 * the import procedure.
 * For detecting them just after the import, it is possible to open all the current ACS
 * in the base, before the import operation. If there are too many ACS in the base,
 * the work may be done on the per-ACS basis.
 * The ghost links are then useful to replace an ACS to remove by an imported
 * ACS, without changing some of its inter-ACS links of the previous ACS.
 * This is of course a choice to explain in the documentation of the import operation.
 * @version 0.7.0
 */
public class ActionImportACS extends AbstractAction implements Singleton {

  /**
   *  the only instance of the class, following singleton pattern
   */
  private static ActionImportACS    instance__ = new ActionImportACS();

  // variables for the dialog
  private static String             lineSepar__ = System.getProperty("line.separator");  


  /**
  * with this private constructor, the compiler won't generate
  * a default public constructor
  */
  private ActionImportACS() {
    super("Import ACS");

    //initialization

  }

  /**
   * only public method to get the unique instance
   * @return ActionImportACS, the unique instance
   */
  public static synchronized ActionImportACS getInstance() {
    return instance__;
  } // getInstance()

  /**
  * Manages the actions fired from the main menu, and displays a dialog box
  * to enter the imported ACS name.
  * @param ev ActionEvent
  */
  public void actionPerformed(ActionEvent ev) {

    // no action if the explorer is currently used for an object selection
    // or if there is an user dialog currently
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;

    String enter = null;
    Name name = null;

    while (enter == null) {
        enter = DialogBox.showSingleEditor("Full name of the ACS to import"+
                " (not its file name...)");
        try {
            if (enter != null)
                name = NameImpl.getNameFromString(enter);
            else {
                DialogBox.showErrorMessage("The ACS import is aborted");
                return;
            } // else
        }
        catch (BaseError b) {
            DialogBox.showErrorMessage(b.toString());
            enter = null;
        } // catch
    } // while

    if (!name.isEmpty())
        prepareACS(name);
    else {
        DialogBox.showErrorMessage("The name is empty."+
          lineSepar__+"The ACS import is aborted.");
    }

    enter = null;
    name = null;
  } // actionPerformed(ActionEvent)
  
  /**
   * Invoked by actionPerformed() to import the ACS from its name.
   * If the ACS-name field is empty, an error message is displayed.
   * If the complete ACS name is already known by the base, an error message is displayed.
   * There is no control of space character at the end of any name
   * entered by the user, nor control of the name size.
   * Call ACS_Facade.getImportedACS() and display the error message if any.
   * @param _name of the imported ACS
   * */
   private void prepareACS(Name _name) {

    Map<String, List<String>> m_closedObjects = null;

    // validateName(_name) displays a warning message
    if (!ACS_Facade.validateFrameName(_name) ) {
        DialogBox.showErrorMessage(
          "A name must do not contain ':: ', or '>>', or '<>' at the start and the end.");
        return;
    }

    // displays a warning message if the name is already known
    if (ACS_Facade.knowsACS(_name) || ViewInBase_Facade.knowsView(_name) ) {
        DialogBox.showErrorMessage(
          "This is the name of a known ACS or a known view."+
          lineSepar__+"To import an ACS having this name, remove this object in the base.");
        return;
    } // if

    InternalFrame.openExplorer();

    try {
        m_closedObjects = ACS_Facade.getImportedACS(_name);
    }
    catch (BaseError b) {
        System.out.println("ActionImportACS.prepareACS() throws for "+_name);
        b.printStackTrace();
        return;
    }
    catch (DialogError d) {
        System.out.println("ActionImportACS.prepareACS() throws for "+_name+
                " the dialog error: "+d.getMessage());
        DialogBox.showErrorMessage(d.getMessage());
        return;
    }

    // displays the ACS in the GUI
    ActionOpenACSyst.getInstance().displayOneACS(
            (ACS)ACS_Facade.getOpenACS().get(_name) );

    DialogBox.showOkMessage("The imported ACS is open. It may have been updated "+
            "for the external inter-ACS links. "+
            "It has been saved in the importation file.");

    if (!m_closedObjects.isEmpty())
        DialogBox.showImmutableStringMap(
            "The created inter-ACS links. "+
            "If not created, do it at hand, or open the missing ACS.",
            m_closedObjects);

    m_closedObjects = null;
  } // prepareACS()

   
} // class ActionImportACS
