//Title:        Access Road
//Version:      0.6.0
//Copyright:    2002, 2004, 2007, 2008, 2009 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gDMak;

import javax.swing.*;
import javax.swing.tree.*;
import java.awt.event.*;
import java.util.List;
import java.util.Arrays;
import java.util.Set;
import java.util.ArrayList;
import java.util.Map;
import java.util.Enumeration;
import java.util.Collections;

import ARoad0.Pattern.*;
import ARoad0.Gui1.*;
import ARoad0.Gui2.*;
import ARoad0.ErrB.BaseError;
import ARoad0.ErrW.*;
import ARoad0.gWork.ACS_Facade;

import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.ACS;
import ARoad0.gBaseInterface.ImmutableNameISTree;

import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.ISFactoryUtilityImpl;

/**
 * An action that opens one or several Access Control System
 * selected in the main menu.
 * @version 0.5.3
 * @since 0.5.0
 */
public class ActionOpenACSyst extends AbstractAction implements Singleton {

  // the only instance of the class, following singleton pattern
  private static ActionOpenACSyst   instance__ = new ActionOpenACSyst();

  // environment
  private static String             currentDir__ = System.getProperty("user.dir");
  private static String             userDir__ = System.getProperty("user.home");
  private static String             fileSepar__ = System.getProperty("file.separator");
  private static String             lineSepar__ = System.getProperty("line.separator");

  // variables for the user data
  private JDialog                   dialog_ = new JDialog();
  private JList                     jlist_ = new JList();

  /**
 * with this private constructor, the compiler won't generate
 * a default public constructor.
 */
  private ActionOpenACSyst() {
    super("Open Access Control System", new ImageIcon("Gui1/images/copy.gif"));
  }

  /**
  * To get an instance.
  * @return ActionOpenACSyst. This is the unique mean to get
  * an instance.
  */
  public static synchronized ActionOpenACSyst getInstance(){
    return instance__;
  } // getInstance()

  /**
  * Handles the click on the menu item "Open ACS".
  * Builds a new frame for the ACS in the desktop (one or more).
  * Refuses ACS main frame creation if there is more than 20 open frames,
  * but the ACS is created.
  * Refuses frame creation if it is a second call.
  * @param _ev ActionEvent produced by the click for "Open ACS".
  */
  public void actionPerformed(ActionEvent _ev) {

    // no action if the explorer is currently used for an object selection
    // or if there is an user dialog currently
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;
      
      // variables of the dialog box
    JOptionPane     pane = new JOptionPane();
    List            list = null;

    // creates a dialog box from the list of closed ACS
    // valueChanged() in jlist is NOT overidden because user must click
    // on a button in the dialog box to valid its input
    list = ACS_Facade.getClosedACSnames();
    jlist_ = new JList(list.toArray() );

    // no work, because selection is null when the user clicks on OK
    if (list.isEmpty() ) {
      // display a message
      DialogBox.showErrorMessage("No ACS to open");
      return;
    }

    // shows the dialog box
    dialog_ = DialogBox.showListDialog(
        "Select all the ACS to open", jlist_);

    // here, dialog_ waits an enter from the user,
    // then the next instructions are processed.

    pane = (JOptionPane)dialog_.getContentPane().getComponent(0);
    // no action if CLOSED, and close the dialog box
    if (pane.getValue() == null) return;
    // else, it is an Integer
    int value = ((Integer)pane.getValue()).intValue();

    // no action if CANCEL, and close the dialog box
    if (value == JOptionPane.CANCEL_OPTION) {
      //Object nullObject = new Object();
      //nullObject = null;
      jlist_.removeAll();
      return;
    }
    // if OK, action and close the dialog box
    if (value == JOptionPane.OK_OPTION) 
        openACS( jlist_.getSelectedValues());

    pane = null;
    return;
  }

  /**
   * Invoked by actionPerformed(), Desktop.restoreLastGUI(), ActionImportACS and 
   * ActionNewACSyst. Opens first the ancestor nodes in the IS, including the ACS.
   * Calls openOneACS(). Opens the explorer first. Opens the tied ACS children.
   * @param _l_names of acs to open
   * */
   public void openACS(Object[] _l_names) {

        if (_l_names == null) {
            // displays a warning mesage if the selection is empty
            DialogBox.showErrorMessage(
              "Please select at least one name to open an ACS.");
            return;
        }
        
        ImmutableNameISTree nametree = null;
        ImmutableName       name = null;
        Object[]            l_ancestors = null;

        InternalFrame.openExplorer();

        try {
            // for each selected ACS name
            for (int i = 0; i < _l_names.length ; i++) {
                name = (ImmutableName)_l_names[i];
                nametree = BaseManagerImpl.getInstance().getISTree(
                        BaseManagerImpl.getInstance().getISofName(name));

                if (!ACS_Facade.isOpenACS(name) && !nametree.isChildOfAcsNode(name))
                    openOneACS(name);
                else {
                    
                    l_ancestors = nametree.getTreePath(nametree.getRoot(), name).getPath();
                    
                    for (int j = 0; j < l_ancestors.length; j++) {
                      if (!ACS_Facade.isOpenACS((ImmutableName)l_ancestors[j]) &&
                              nametree.getType((ImmutableName)l_ancestors[j]).
                                equals(ISFactoryUtilityImpl.ACS_NODE)) {

                                  openOneACS(
                                    (ImmutableName)l_ancestors[j]);

                      } // if
                    } // for node

                } // else
                // opens the tied ACS children of name
                for (ImmutableName na : nametree.getL_DirectChildren(name)) {

                    if (!ACS_Facade.isOpenACS(na) &&
                            ACS_Facade.knowsACS(na) &&
                            Arrays.asList(nametree.getLinks(na)).contains(
                                ISFactoryUtilityImpl.IS_TIED_CHILD)) {
                        
                                        openOneACS(na);
                    } // if

                } // for na

            } // for i
        } catch (BaseError b) {
              b.printStackTrace();
              throw new InternalError("from ActionOpenACSyst.openACS().");
        }

        nametree = null;
        name = null;
        l_ancestors = null;
   } // openACS

  /**
   * Opens the selected ACS, to display their resources in trees,
   * to update the explorer, the beamer and the 'Architecture' view.
   * ACSTreeBaseListenerImpl is not still created, then no fired event to the GUI.
   * Called by openACS(). Calls ACS_Facade.openOrGetACS().
   * @param _name of the ACS to open
   * */
   private void openOneACS(ImmutableName _name) {
      try {
          displayOneACS(ACS_Facade.openOrGetACS(_name));
      }
      catch (BaseError b) {
        DialogBox.showErrorMessage(
          "There is an error in the ACS opening process."+
            lineSepar__+"See the program messages." );
        System.out.println("There is an error in the ACS opening process."+
            " The exception is:"+lineSepar__+b.toString());
      } // catch
   } // openOneACS()

  /**
   * Displays an open ACS, with all its objects.
   * Sets the ACS node at the right place into the explorer, following the order
   * of the IS in the base and the order of the ACS into the relevant NameISTree.
   * Updates the beamer and the 'Architecture' view.
   * The explorer has to be open.
   * ACSTreeBaseListenerImpl is not still created, then no fired event to the GUI.
   * Called by openOneACS() and ActionNewACSyst.endACScreation().
   * Calls BaseManagerImpl..getM_ISTrees(), NameTree.getDepthFirstIndexInNameTree().
   * @param _acs opened ACS to display
   * */
   protected void displayOneACS(ACS _acs) {
       // variables
      JTree                   eTree = InternalFrameModel.getTreeInExplorer();
      DefaultTreeModel        eTreeModel = (DefaultTreeModel) eTree.getModel();
      DefaultMutableTreeNode  acsNode = null;
      int                     index = 0;
      int                     numberOfAcs = 0;
      ImmutableName           acsName = null;
      ImmutableName           iSName2 = null;
      ImmutableName           iSNameOfNewACS = BaseManagerImpl.getInstance().
                                getISofName(_acs.getName());
      Map<ImmutableName, ImmutableNameISTree> m_isTrees =
                                    BaseManagerImpl.getInstance().getM_ISTrees();
      ImmutableNameISTree     nametree = m_isTrees.get(iSNameOfNewACS);
      List<ImmutableName>     l_IS = new ArrayList(BaseManagerImpl.getInstance().getL_IS());
      int                     depthIndexOfNewACS = -1;
      int                     depthIndex = -1;

                          // for testing and tracking
                          //System.out.println("call to ActionOpenACSyst.displayOneACS() for "+_acs);

      try {
          depthIndexOfNewACS = nametree.getDepthFirstIndexInNameTree(_acs.getName());

                                      // for testing and tracking
                                      //System.out.println("ActionOpenACSyst.displayOneACS() "+
                                        //      " depthIndexOfNewACS = "+depthIndexOfNewACS);

          // updates the explorer
              // number of displayed ACS in the explorer
          numberOfAcs = eTreeModel.getChildCount(((DefaultMutableTreeNode)eTreeModel.
              getRoot()).getChildAt(0) );

          acsNode = new DefaultMutableTreeNode(_acs );

          if (numberOfAcs != 0) {
              for (ImmutableName iSName : l_IS) {

                                      // for testing and tracking
                                      //System.out.println("-- ActionOpenACSyst.displayOneACS() "+
                                        //      " for iSName: "+iSName+
                                          //    ", l_IS.indexOf(iSName) = "+l_IS.indexOf(iSName));

                  // the IS of _acs is not after
                  if (l_IS.indexOf(_acs.getISName()) < l_IS.indexOf(iSName)) {
                    break;
                  }

                    // search, among the ACS nodes, the right position of the new TreeNode
                  for (Enumeration enu = ((DefaultMutableTreeNode)eTreeModel.getRoot()).
                          getChildAt(0).children(); enu.hasMoreElements(); ) {

                      // the ACS name is the name associated to the selected node
                      acsName = (ImmutableName)
                        ((ACS)((DefaultMutableTreeNode)enu.nextElement()).getUserObject()).getName();

                      if (_acs.getName().equals(acsName))
                          throw new BaseError("---- ActionOpenACSyst.displayOneACS() finds "+
                                  "that the ACS is already displayed by the explorer.");

                      iSName2 = BaseManagerImpl.getInstance().getISofName(acsName);

                      // each IS has to be counted once
                      if (!iSName.equals(iSName2)) {
                          continue;
                      }

                              // for testing and tracking
                              //System.out.println("---- ActionOpenACSyst.displayOneACS() search from "+
                                //      acsName+" with index = "+index);

                      if (!iSNameOfNewACS.equals(iSName2)) {

                                      // for testing and tracking
                                      //System.out.println("---- ActionOpenACSyst.displayOneACS() "+
                                        //      "not on the searched IS ");

                              index++;
                              continue;
                      } // if (!iSNameOfNewACS.equals(iSName2))

                      depthIndex = nametree.getDepthFirstIndexInNameTree(acsName);

                                      // for testing and tracking
                                      //System.out.println("---- ActionOpenACSyst.displayOneACS() "+
                                        //      "depthIndex = "+depthIndex);

                      if (depthIndexOfNewACS > depthIndex) {

                                      // for testing and tracking
                                      //System.out.println("---- ActionOpenACSyst.displayOneACS() "+
                                        //      "depthIndex is lesser: "+
                                          //    depthIndex+
                                            //  " than "+depthIndexOfNewACS);

                          index++;
                          continue;
                      }
                      else {
                          break;
                      }
                  } // for
              } // for iSNAme
          } // if (numberOfAcs != 0)

                          // for testing and tracking
                          //System.out.println("ActionOpenACSyst.displayOneACS() - index = "+
                            //      index);

          eTreeModel.insertNodeInto(
              acsNode,
              (DefaultMutableTreeNode)(((DefaultMutableTreeNode)eTreeModel.
                getRoot()).getChildAt(0)),
              index);

                          // for testing and tracking
                          //System.out.println("in ActionOpenACSyst.displayOneACS(),"+
                            //" add an ACS: "+_acs.getName()+", at the index: "+index+
                            //", under the node: "+((DefaultMutableTreeNode)eTreeModel.
                            //getRoot()).getChildAt(0)+", under the tree root: "+eTreeModel.
                            //getRoot());

          try {
                ACSTree.createACSExplorer(_acs, eTree, acsNode);
          } catch (ProcessError p) {
              System.out.println(
                      "exception in ActionOpenACSSyst.openACS() on ACSTree.createACSExplorer():"+
                      p.toString());
              throw new BaseError(p.toString());
          }

          // changes the selection in the explorer, and in the beamer
          eTree.setSelectionPath(new TreePath( acsNode.getPath() ));
          // if the beamer is closed, open it with this acs
          if ( !InternalFrameModel.withBeamer() ) {
            try {
              InternalFrame.openBeamer(
                BeamerManager.getInstance().updateBeamerPane(_acs));
            } catch (ProcessError p) {
              DialogBox.showAlarmMessage("ActionOpenACS.openACS() , the beamer can't detect"+
                " the acs to display.");
            }
          } // if

      } catch (BaseError b) {
        DialogBox.showErrorMessage(
          "There is an error in the ACS opening process."+
            lineSepar__+"See the program messages." );
        System.out.println("There is an error in the ACS opening process."+
            " The exception is:"+lineSepar__+b.toString());
      } // catch
                          
      eTree = null;
      eTreeModel = null;
      acsNode = null;
      nametree = null;
      acsName = null;
      iSName2 = null;
      iSNameOfNewACS = null;
      m_isTrees = null;
      l_IS = null;
   } // displayOneACS()

} // class ActionOpenACSyst


