//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gDMak;

import javax.swing.*;
import java.awt.event.*;
import javax.swing.tree.*;
import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Arrays;

import ARoad0.Pattern.*;
import ARoad0.Gui1.*;
import ARoad0.Gui2.TreeManager;
import ARoad0.gWork.*;
import ARoad0.ErrB.BaseError;

import ARoad0.gBase.NameImpl;
import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.ISFactoryUtilityImpl;

import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.ViewInBase;
import ARoad0.gBaseInterface.ImmutableNameISTree;

/**
 * An action that removes an open or closed ACS or view,
 * a closed composite, a physical or logical component,
 * from the Access Road base, without backup.
 * The IS is always removed when its last ACS is removed.
 * The removed entity may be open or closed. See ActionUtilities.
 * @version 0.7.0
 */
public class ActionRemove extends AbstractAction implements Singleton {

  /* the only instance of the class, following singleton pattern */
  private static ActionRemove instance__ = new ActionRemove();

  private static ViewInBase_Facade  viewBase__ = ViewInBase_Facade.getInstance();

  /**
   * with this private constructor, the compiler won't generate
   * a default public constructor
   */
  private ActionRemove() {
    super("Remove");
  }

  /**
   * constructor()
   * @return ActionRemove
   */
  public static synchronized ActionRemove getInstance(){
    return instance__;
  } // getInstance()

  /**
   * Manages removing commands from one source:
   * - from 'Remove' in the main menu when the explorer is open, or
   * - from the main frame of an ACS or a view is activated.
   * Every removing is confirmed by the user, and the entity is closed
   * before if it is open.
   * First, if there is at least one closed ACS, a dialog requests to the user if
   * he wants to remove one of them, and the user may select one
   * and removes it. If there is no closed ACS, if the user answer 'no',
   * a dialog requests if the user wants to close an open tied-child ACS.
   * Otherwise, if there is at least one closed view, a dialog requests to the user
   * if he wants to remove one of them, and the user may select one
   * and remove a closed view.
   * Finally, if the user rejects the two first options, if
   * the 'IS Structure' tree is open and if there is a closed composite,
   * a physical or logical component that is selected, this object is removed.
   * @param _ev ActionEvent
   */
   public void actionPerformed(ActionEvent _ev) {

    // no action if the explorer is currently used for an object selection
    // or if there is an user dialog currently
    if (DialogBox.onGoingObjectSelectionOrUserDialog()) return;

                                      // for testing and tracking
                                  //System.out.println("call to ActionRemove.actionPerformed()");
    
    // initialization
    Object[]               l_analyzedEvent = new Object[3];
    //Name                   name = new NameImpl();
    JList                  jlist = new JList();
    JDialog                dialog = new JDialog();
    boolean                removeClosedEntityToSelect = false;
    boolean                removeClosedEntity = false;
    Map                    m_GUIconfig = null;     
    Map                    m_LastConfig = null;
    Map                    m_PreviousConfig = null;
    ImmutableName          name = null;

    // request for removing an ACS if there are closed ACS
    if (!ACS_Facade.getClosedACSnames().isEmpty() &&
      DialogBox.showConfirmDialog("Remove a closed ACS?", null) ) {

        jlist = new JList(ACS_Facade.getClosedACSnames().toArray() );
        l_analyzedEvent[0] = "acs";
        removeClosedEntityToSelect = true;

    } // if (!ACS_Facade.getInstance().getClosedACS().isEmpty()

    // request for removing an open ACS that is a tied child
    else if (DialogBox.showConfirmDialog("Remove an open tied-child ACS?", null) ) {

        ImmutableNameISTree nametree = null;
        Set                 set = new HashSet(10);

        for (Object obj : ACS_Facade.getOpenACS().keySet()) {
            nametree = BaseManagerImpl.getInstance().getISTree(
                        BaseManagerImpl.getInstance().getISofName((ImmutableName)obj));
            if ((Arrays.asList(nametree.getLinks((ImmutableName)obj))).contains(
                    ISFactoryUtilityImpl.IS_TIED_CHILD))
                          set.add((ImmutableName)obj);
        } // for obj

        if (!set.isEmpty()) {
            removeClosedEntityToSelect = true;
            jlist = new JList(set.toArray() );
            l_analyzedEvent[0] = "acs";
        } // if (!set
        nametree = null;
        set = null;
    } // if (if (DialogBox.showConfirmDialog("Remove a open tied-child ACS?"

    // request for removing a view, if ACS removing is not selected
    else if (!removeClosedEntityToSelect && !ViewInBase_Facade.getClosedViews().isEmpty() &&
      DialogBox.showConfirmDialog("Remove a closed view?", null) ) {

        jlist = new JList(ViewInBase_Facade.getClosedViews().toArray() );
        l_analyzedEvent[0] = "view";
        removeClosedEntityToSelect = true;

    } // if (informationOnACS

    // request for removing a closed composite, a closed physical or logical component
    else if (!removeClosedEntityToSelect && InternalFrameModel.withStructure() &&
      TreeManager.analyzeSelectedNodeInStructure() != null ) {

        DefaultMutableTreeNode treeNode = TreeManager.analyzeSelectedNodeInStructure();

        name = TreeManager.getTrueUserObjectNameInStructureTree(treeNode);

        ImmutableNameISTree nameTree = BaseManagerImpl.getInstance().getISTree(
                            BaseManagerImpl.getInstance().getISofName(name));

        try {
            if (!nameTree.isOpen(name)) {
                l_analyzedEvent[0] = nameTree.getType(name);
                if (l_analyzedEvent[0].equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) ||
                        l_analyzedEvent[0].equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE) ||
                        l_analyzedEvent[0].equals(ISFactoryUtilityImpl.PHYSICAL_COMPONENT_NODE) ||
                        l_analyzedEvent[0].equals(ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE)) {
                            l_analyzedEvent[1] = name;
                            removeClosedEntity = true;

                                      // for testing and tracking
                                  //System.out.println("in ActionRemove.actionPerformed(), "+
                                    //"the selected entity is a node in the structure. ");
                } // if
            } // if
        }
        catch (BaseError b) {
            throw new InternalError("ActionRemove.actionPerformed() for a structure node: "+
                    b.toString());
        }
        treeNode = null;
        nameTree = null;
    } // if (!ACS_Facade.getInstance().getClosedACS().isEmpty()

    if (!removeClosedEntityToSelect && !removeClosedEntity) {
        l_analyzedEvent = null;
        jlist = null;
        dialog = null;
        DialogBox.showOkMessage("There is no removing.");
        return;
    }


    if (removeClosedEntityToSelect) {
      // works on the selected closed ACS or view
      jlist.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      dialog = DialogBox.showListDialog(
        "Select the entity to remove", jlist);

      // here, dialog_ waits an enter from the user,
      // then the next instructions are processed.

      JOptionPane pane = (JOptionPane)dialog.getContentPane().getComponent(0);
      // no action if CLOSED, and close the dialog box
      if (pane.getValue() == null) {
        l_analyzedEvent = null;
        jlist = null;
        dialog = null;
        pane = null;
        return;
      }
      // else, it is an Integer
      int value = ((Integer)pane.getValue()).intValue();

      // if OK, closes the dialog box and initialize l_analyzedEvent
      // for processing it
      if (value == JOptionPane.OK_OPTION) {
        l_analyzedEvent[1] = (ImmutableName)jlist.getSelectedValue();
        // l_analyzedEvent[2] == null;
        pane = null;
        removeClosedEntity = true;

                  // for testing and tracking
              //System.out.println("in ActionRemove.actionPerformed(), "+
                //"the selected entity is: "+(ImmutableName)jlist.getSelectedValue());
      } // if (value == JOptionPane.OK_OPTION)
      else {
            l_analyzedEvent = null;
            jlist = null;
            dialog = null;
            return;
      }      

    } // if (removeClosedEntity)

          // for testing and tracking
    //System.out.println("in ActionRemove.actionPerformed()"+
    //  "l_analyzedEvent: ");
    //for (int i = 0 ; i < l_analyzedEvent.length ; i++) {
    //  System.out.println(l_analyzedEvent[i].toString());
    //}

    if (!removeClosedEntity) {
            l_analyzedEvent = null;
            jlist = null;
            dialog = null;
            return;
    } // if

    // the user reads the ACS dependancies on other ACS and views,
    // before confirming the removing of an ACS
    if (removeClosedEntityToSelect && ((String)l_analyzedEvent[0]).equals("acs"))
        informationOnACSremoving((ImmutableName)l_analyzedEvent[1]);

    if (!DialogBox.showConfirmDialog(
                ((ImmutableName) l_analyzedEvent[1]).toString()+"removing?",
                "The file '.acr' "+
                "would NOT be deleted by the program.")) {
                    l_analyzedEvent = null;
                    jlist = null;
                    dialog = null;
                    return;
    } // if

    if (((String)l_analyzedEvent[0]).equals("acs")) {

        // closes the ACS without saving
        //if (ACS_Facade.isOpenACS((ImmutableName)l_analyzedEvent[1]))
          //ActionClose.getInstance().close((ImmutableName)l_analyzedEvent[1],
            //false);

        try {
            // removes the ACS
            ACS_Facade.removeACS((ImmutableName)l_analyzedEvent[1]);
        } catch (BaseError b){
          DialogBox.showOkMessage("The ACS removing is not correct."+
            "An exception has been thrown: "+b.toString() );
        } // catch
    } // if (((String)l_analyzedEvent[0]).equals("acs"))

    else if (((String)l_analyzedEvent[0]).equals("view")) {
        // closes the view without saving
        //if (viewBase__.isOpenView((ImmutableName)l_analyzedEvent[1]))
          //ActionClose.getInstance().close((ImmutableName)l_analyzedEvent[1],
            //false);
        // removes the view
        try {
          viewBase__.removeView((ImmutableName)l_analyzedEvent[1]);
        } catch (BaseError b){
          DialogBox.showOkMessage("The view removing is not correct."+
            " An exception has been thrown: "+b.toString() );
        } // catch
    } // if (((String)l_analyzedEvent[0]).equals("view"))

    else if (l_analyzedEvent[0].equals(ISFactoryUtilityImpl.COMPOSITE_ACS_NODE) ||
            l_analyzedEvent[0].equals(ISFactoryUtilityImpl.COMPOSITE_SUB_ACS_NODE)) {
        // removes the composite
        try {
          BaseManagerImpl.getInstance().removeCompositeNodeToIS(
                    BaseManagerImpl.getInstance().getISofName((ImmutableName) l_analyzedEvent[1]),
                    (ImmutableName) l_analyzedEvent[1]);
        } catch (BaseError b){
          DialogBox.showOkMessage("The node removing is not correct."+
            " An exception has been thrown: "+b.toString() );
        } // catch
    } // if (((String)l_analyzedEvent[0]).equals("COMPOSITE"))

    else if (l_analyzedEvent[0].equals(ISFactoryUtilityImpl.PHYSICAL_COMPONENT_NODE) ||
            l_analyzedEvent[0].equals(ISFactoryUtilityImpl.LOGICAL_COMPONENT_NODE)) {

                                      // for testing and tracking
                                  //System.out.println("in ActionRemove.actionPerformed(), "+
                                    //"removes a component node in the structure. ");

        if (BaseManagerImpl.getInstance().getISTree((ImmutableName) l_analyzedEvent[1]).
                                          isParentOfAcsNode((ImmutableName) l_analyzedEvent[1]) ) {
            if (!DialogBox.showConfirmDialog("This component is the ancestor of some ACS.",
                    "Do you confirm its removing with all its children?") ) {
                l_analyzedEvent = null;
                m_GUIconfig = null;
                jlist = null;
                dialog = null;
                return;
            } // if
        } // if

        // removes the composite
        try {
          BaseManagerImpl.getInstance().removeComponentNodeToIS(
                    BaseManagerImpl.getInstance().getISofName((ImmutableName) l_analyzedEvent[1]),
                    (ImmutableName) l_analyzedEvent[1]);
        } catch (BaseError b){
          DialogBox.showOkMessage("The node removing is not correct."+
            " An exception has been thrown: "+b.toString() );
        } // catch
    } // if (((String)l_analyzedEvent[0]).equals("COMPOSITE"))

      // removes the frame and the graphic view disposition from the GUI saving
    try {
        if (removeClosedEntityToSelect) {
            m_GUIconfig = BaseManagerImpl.getInstance().getGUI();
            if (m_GUIconfig != null) {
                m_LastConfig = (Map)m_GUIconfig.get( new NameImpl("<Last GUI>") );
                m_PreviousConfig = (Map)m_GUIconfig.get( new NameImpl("<Previous GUI>") );
            }

            if (((String)l_analyzedEvent[0]).equals("acs")) {

              if (m_PreviousConfig != null ) {
                  m_PreviousConfig.remove( ((ImmutableName)l_analyzedEvent[1]).toString() );
                  m_PreviousConfig.remove((ImmutableName)l_analyzedEvent[1] );
              }
              if (m_LastConfig != null ) {
                  m_LastConfig.remove( ((ImmutableName)l_analyzedEvent[1]).toString() );
                  m_LastConfig.remove( (ImmutableName)l_analyzedEvent[1] );
              }

            } // if (((String)l_analyzedEvent[0]).equals("acs"))
            else if (((String)l_analyzedEvent[0]).equals("view")) {

              if (m_PreviousConfig != null ) {
                  m_PreviousConfig.remove( ((ImmutableName)l_analyzedEvent[1]).toString() );
                  m_PreviousConfig.remove((ImmutableName)l_analyzedEvent[1] );
                  m_PreviousConfig.remove(
                          (new NameImpl("<GraphicView for>")).extended((ImmutableName)l_analyzedEvent[1]) );
                  m_PreviousConfig.remove(
                          (new NameImpl("<See why for>")).extended((ImmutableName)l_analyzedEvent[1]) );
              } // if
              if (m_LastConfig != null ) {
                  m_LastConfig.remove( ((ImmutableName)l_analyzedEvent[1]).toString() );
                  m_LastConfig.remove((ImmutableName)l_analyzedEvent[1] );
                  m_LastConfig.remove(
                          (new NameImpl("<GraphicView for>")).extended((ImmutableName)l_analyzedEvent[1]) );
                  m_LastConfig.remove(
                          (new NameImpl("<See why for>")).extended((ImmutableName)l_analyzedEvent[1]) );
              } // if

            } // else 'view'

            BaseManagerImpl.getInstance().saveGUI(m_GUIconfig);
            
            m_GUIconfig = null;
            m_PreviousConfig = null;
            m_LastConfig = null;
        } // if (removeClosedEntityToSelect)
    } catch (BaseError b) {
            l_analyzedEvent = null;
            m_GUIconfig = null; 
            jlist = null;
            dialog = null;            
            throw new InternalError("in ActionRemove.actionPerformed(), error: "+b);
    } // catch  
        
    m_GUIconfig = null;
    l_analyzedEvent = null;
    jlist = null;
    dialog = null;
    name = null;
  } // actionPerformed(ActionEvent)

  /**
  * The user has to read the ACS dependencies on children ACS and views,
  * before confirming the removing of an ACS.
  * Displays the list of these associated ACS and views.
  * @param _name of the ACS to remove
  */
  private void informationOnACSremoving(ImmutableName _name) {
    if (!ACS_Facade.knowsACS(_name)) return;

    // variables
    StringBuilder           buffer = null;
    ImmutableName           name = null;
    ImmutableName           iSname = BaseManagerImpl.getInstance().getISofName(_name);
    Set<ImmutableName>      l_associates = new HashSet();
    Set<ImmutableName>      l_views = new HashSet();
    ImmutableNameISTree     nametree = BaseManagerImpl.getInstance().getISTree(iSname);
    List<ImmutableName>     l_children = nametree.getL_AllChildren(_name);

    // get the children ACS
    try {
        for (ImmutableName na : l_children) {
            if (nametree.getType(na).equals(ISFactoryUtilityImpl.ACS_NODE) )
                l_associates.add(na) ;
        } // for name
    }
    catch (BaseError b) {
        throw new InternalError("ActionRemove.informationOnACSremoving(): "+
             b.toString());
    }

    if (!l_associates.isEmpty()) {
        buffer = new StringBuilder();

                            // for testing and tracking
                            //System.out.println("in ActionRemove.informationOnACSremoving(), "+
                            // "views in l_associates: ");
                            //for (ImmutableName na :  l_associates) {
                                //System.out.println("----------"+ na );
                            //}


        for (Iterator ite = l_associates.iterator(); ite.hasNext(); ) {
            buffer.append( (ImmutableName)ite.next() );
            if (ite.hasNext()) buffer.append(",");
        }

        if (l_associates.size() > 1)
            DialogBox.showOkMessage("The ACS to remove is the parent of these ACS. "+
                "They would be removed too: "+buffer.toString());
        else
            DialogBox.showOkMessage("The ACS to remove is the parent of this ACS. "+
                "It would be removed too: "+buffer.toString());
    } else
        DialogBox.showOkMessage("The ACS to remove has no child ACS.");


    // analyze the views for all the implied ACS
    l_associates.add(_name);

    try {
      for (ImmutableName name2 : l_associates) {
        // gets the open implied views
        for (Object obj: ViewInBase_Facade.getAllStandardOpenViews().keySet() ) {
            name = (ImmutableName) obj;
            if (ViewInBase_Facade.isNoThanViewName(name) &&
                ((ViewInBase)ViewInBase_Facade.getOpenNoThanViews().get(name)).
                        getEorL_ConnectedACSNames().contains(name2)) {
                                l_views.add(name);
            } // if
            else if ( ViewInBase_Facade.getOpenFullViews().get(name) != null &&
                    ((ViewInBase)ViewInBase_Facade.getOpenFullViews().get(name)).
                        getEorL_ConnectedACSNames().contains(name2)) {
                                l_views.add(name);
            } // else if
        } // for obj


                            // for testing and tracking
                            //System.out.println("in ActionRemove.informationOnACSremoving(), "+
                            // "views in l_associates: ");
                            //for (ImmutableName na :  l_associates) {
                                //System.out.println("----------"+ na );
                            //}

            // adds the closed implied views, if any
            for (Iterator ite = ViewInBase_Facade.getClosedViews().iterator(); ite.hasNext(); ) {
                name = (ImmutableName)ite.next();

                                // for testing and tracking
                                //System.out.println("in ActionRemove.informationOnACS(), "+
                                  //      "closed view: "+name);

                if (ViewInBase_Facade.getElementsOfOneView(name, true).contains(name2))
                            l_views.add(name);
            } // for ite

        } // for name2
    } catch (BaseError b) {
        throw new InternalError("from ActionRemove.informationOnACSremoving(): "+b.toString());
    }

    if (!l_views.isEmpty()) {
        buffer = new StringBuilder();

                            // for testing and tracking
                            //System.out.println("in ActionRemove.informationOnACSremoving(), l_views: ");
                            //for (Iterator ite = l_views.iterator(); ite.hasNext(); ) {
                                //System.out.println("----------"+ (ImmutableName)ite.next() );
                            //}


        for (Iterator ite = l_associates.iterator(); ite.hasNext(); ) {
            buffer.append( (ImmutableName)ite.next() );
            if (ite.hasNext()) buffer.append(",");
        }
        if (l_views.size() > 1)
            DialogBox.showOkMessage("An ACS to remove is used by these views. "+
                "They contain element(s) from this ACS, but these views would not be removed: "+
                buffer.toString());
        else
            DialogBox.showOkMessage("An ACS to remove is used by this view. "+
                "It contains element(s) from this ACS, but the view would not be removed: "+
                buffer.toString());
    } else
        DialogBox.showOkMessage("The ACS to remove is not used by any view.");

    l_associates = null;
    buffer = null;
    name = null;
    iSname = null;
    l_associates = null;
    l_views = null;
    nametree = null;
    l_children = null;
  } // informationOnACSremoving()

} // class ActionRemove