//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import ARoad0.gWorkInterface.*;
import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;

import ARoad0.CNot.*;
import ARoad0.ErrW.*;

import ARoad0.gBaseInterface.*;

import ARoad0.gBase.UtilityImpl;
import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.StringRightImpl;

/**
 * This class is responsible for all the core algorithms about AclEntry rights analysis,
 * and most of the Bridge rights analysis.
 * These rights are direct or inherited rights for an EligibleParty (the access user)
 * on a Resource (the access target).
 * This class is stateless, and RightsMediatorImpl at its creation
 * setups an instance for its ViewInBase.
 * No direct call to the ACS methods.
 * Upper-level methods begin by detect...(),
 * middle-level methods begin by select...(),
 * lower-level methods begin by add...().
 * @see ARoad0.gBase.AclEntryImpl
 * @version 0.7.0
 */
public class AclRightsFactoryImpl implements CoreAlgorithm {
  private AlgorithmInterpreter      interpreter_ = null;
  private UtilityImpl               utility_ = null;
  private DisplayableLinkUtilities  linkUtil_ = null;

  /**
  * only one protected constructor
  * @param _interpreter algorithm interpreter
  */
  protected AclRightsFactoryImpl(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      utility_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
  }

  /**
  * Defines the effective applicable rights of an EligibleParty for a Resource,
  * when there are direct AclEntries between them.
  * For a sketch view, only AclEntries from/to _center are processed.
  * Called by RightsFactory_Facade.detectEPRRights(), detectNoThanRights()
  * and detectSketchRights(). <P>
  * Synchronized to _viewInBase from the GraphicView call.
  * @param _viewInBase the EPRViewInBase of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks lists,
  *   with ACL links associated to every pair (Source, Target).
  * */
  protected final Map detectACLRights( 
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks, BaseObject _center) {
      // variables
      ImmutableSource         so;
      EligibleParty           ep;
      Resource                res;
      AccessControlLinkImpl   acLink;
      Object                  obj;
      LinkRights              linkRights = null;

      for (Iterator ite = _viewInBase.getEorM_Sources().values().iterator(); ite.hasNext() ; ) {
        so = (ImmutableSource)ite.next();
        if (so == null || !(so instanceof EligibleParty)) continue;
        ep = (EligibleParty) so;

        for (Iterator iti = _viewInBase.getEorM_Targets().values().iterator(); iti.hasNext() ; ) {
            obj = iti.next();
            if (!(obj instanceof Resource)) continue;
            res = (Resource) obj;

            // for a sketch view
            if (_center != null && _center != ep && _center != res) continue;

            if (ep == res) continue;
                    
                                        // for testing and tracking
                                    //System.out.println("AclRightsFactoryImpl.detectACLRights(), "+
                                      //      "for the source "+ep.getNickName()+
                                        //    " and for the target "+res.getNickName()+
                                          //  ", call to addAclInAccessControlLink()");


            acLink = addAclInAccessControlLink(res, ep, null);

            if (acLink != null) {
                linkRights = interpreter_.getLinkRightsInstances().get(res.getName());
                linkRights.addDisplayableLink(
                                new DisplayableLinkImpl(ep, res, acLink),
                                _m_l_DisplayableLinks);
            } // if
            
        } // for (Iterator iti
      }  // for (Iterator ite    
      
      res = null;
      ep = null;
      acLink = null;
      so = null;
      obj = null;
      
                                        // for testing and tracking
                                    //System.out.println("end of AclRightsFactoryImpl.detectACLRights() ");
      
      return _m_l_DisplayableLinks;
  } // detectACLRights()

  /**
  * Detects all the right-proxy Actors that are not in _viewInBase, and are directly
  * accessible from _ep through an AclEntry or a Bridge,
  * with some executing rights or not. The rights are set.
  * These actors will define new links that should do not transfer the AG context
  * of the argument _ep. With an internal ACL, the AG context is transferred, but
  * it is not possible to say if the ACL target, as actor, do not provide another
  * links outside the current AG context. This is why all the ACL are detected.
  * Note: this method does not take account of the inherited AclEntries in an AcsAddon,
  * and this may be done in NodeRights.selectAddonNewHiddenNodeForGroupIDMember()
  * and selectAddonNewHiddenNodeForActor(), if these methods are overidden by the AcsAddon.
  * Called twice by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),
  * with true for _withFinalExecute to find the access paths, and false to find
  * the actors outside any path but to be listened.
  * Calls addAclBridgeRightsInSimpleDisplayableLink().
  * @param _viewInBase view on which _gview is based
  * @param _ep is in viewInBase or not
  * @param _isProxyWithFinalExecute true for detecting only if there is an execute
  * right on a right-proxy actor. If false, any actor without executing right
  * and/or not being a right-proxy is returned.
  * @exception ProcessError if there is a loop in the rights search
  * @return Set of simple DisplayableLinkImpl for which the first end is _ep,
  *   the second end is an actor not in _viewInBase.
  *   May be null, but not empty.
  * */
  public final Set selectDirectHiddenAclBridgeLinksToActorForEP(
    EPRViewInBase _viewInBase, ImmutableEligibleParty _ep,
    boolean _isProxyWithFinalExecute) throws ProcessError {

    ImmutableResource     res;
    Set                   l_result = null;
    DisplayableLinkImpl   dLink = null;
    NodeRights            nodeRights = null;
    ImmutableActor        act = null;

    for (Iterator iti = _ep.getEorL_AclConnectedResources().iterator(); iti.hasNext() ; ) {
          res = (ImmutableResource) iti.next();

          // take the internal and the external ACLs
          if (res instanceof ImmutableActor &&
                (!_isProxyWithFinalExecute ||
                (_isProxyWithFinalExecute && ((ImmutableActor)res).getIsRightProxy())) &&
            !_viewInBase.containsSource((ImmutableActor)res)) {

                            // for testing and tracking
                            //System.out.println("in AclRightsFactoryImpl."+
                              //"selectDirectHiddenExternalAclBridgeLinksToActorForEP(), add the simple link for ACL"+
                              //" with _withFinalExecute: "+_withFinalExecute);

                    nodeRights = interpreter_.getNodeRightsInstances().get(res.getName());
                    if (nodeRights == null) {
                        interpreter_.addInstancesForNode(res);
                    } // if

                    dLink = addAclBridgeRightsInSimpleDisplayableLink(
                            _ep, res, _isProxyWithFinalExecute);

                    if (dLink != null) {
                      if (l_result == null)
                          l_result = new HashSet(INITIAL_CAPACITY / 3);
                      l_result.add(dLink);
                    } // if

                            // for testing and tracking
                            //System.out.println("in AclRightsFactoryImpl."+
                              //"selectDirectHiddenExternalAclBridgeLinksToActorForEP(), "+
                              //"add the simple link for ACL:");
                            //System.out.println("----------- "+dLink);

          } // if (res
    } // for iti

    if (_ep instanceof ImmutableActor &&
            ((ImmutableActor)_ep).getIsACSBridgeSource()) {

            act = ((ImmutableActor)_ep).getEorACSBridgeTarget();
            if (!_viewInBase.containsSource( act )) {

                    nodeRights = interpreter_.getNodeRightsInstances().get(
                            act.getName());
                    if (nodeRights == null) {
                        interpreter_.addInstancesForNode(act);
                    } // if

                    dLink = addAclBridgeRightsInSimpleDisplayableLink(
                          _ep, act, _isProxyWithFinalExecute);

                    if (dLink != null) {
                      if (l_result == null)
                          l_result = new HashSet(INITIAL_CAPACITY / 3);
                      l_result.add(dLink);
                    } // if
            } // if (!_viewInBase.containsSource( act ))
    } // (((Actor)ep).getIsACSBridgeSource()
    
    res = null;
    dLink = null;
    nodeRights = null;
    act = null;
    return l_result;
  } // selectDirectHiddenAclBridgeLinksToActorForEP()

  /**
  * Selects the right aclEntry managed by the resource, from the ACS list because
  * the resource list is updated after the ACS has fired its change property events
  * for the AclEntry creation.
  * Called by CompoundRightsFactoryImpl.detectAcsLinks().
  * @param _res with some AclEntries to filter
  * @param _ep is the EligibleParty associated to the AclEntry
  * @param _aclAcs ImmutableACS which owns the new AclEntry
  * @param _positiveRight is true is the searched AclEntry is psitive
  * @return the ImmutableAclEntry, or null.
  * */
  public static final ImmutableAclEntry getAclEntryFromACS(ImmutableResource _res,
    ImmutableEligibleParty _ep, ImmutableACS _aclAcs, boolean _positiveRight) {

      ImmutableAclEntry[] l_acl = _aclAcs.getEorL_AclEntries();
      if (l_acl == null || l_acl.length == 0) return null;
      ImmutableAclEntry result = null; 
      
      for (int i = 0; i < l_acl.length; i++) {
        if (l_acl[i] != null && l_acl[i].getEorEP() == _ep &&
          l_acl[i].getEorResource() == _res &&
          l_acl[i].getPositiveRight() == _positiveRight) {
              result = l_acl[i];
              break;
        }
      } // for
      
      l_acl = null;
      return result;
  } // getAclEntryFromACS()

  //
  // --------- sensible methods ----------------------------------------
  //

  /**
  * Gets or creates the simple DisplayableLinkImpl associated to a view object pair,
  * and updates the AccessControlLink with ACL or bridge rights.
  * The main function is to handle the need to have executing rights
  * when the last argument requires it, and this is necessary for the intermediate
  * actors in the DisplayableLinks.
  * Each right is provided by the bridge or the AclEntry ACS,
  * and it may have several ACS which are implied in one DisplayableLinkImpl.
  * So, puts in comments the acl ACS name for each ACL right.
  * Note : this method is sensible since it is used for both
  * the two-link method and a multiple-links method.
  * Called by selectDirectHiddenExternalAclBridgeLinksToActorForEP(),
  * ThreeNodesRightsFactoryImpl.addPathsFromGroupIDMemberAclActor().
  * Calls addAclInAccessControlLink().
  * Note: when the argument _withExecute is true, the method is much more complex.
  * AccessControlLinkImpl.getRedondantRights() is called to have all the rights
  * from all the implied ACSs, with one positive or negative right per ACS, and
  * BaseUtilityImpl.withExecuteRight() gets the effective global rights to evaluate
  * if there is an executing right. This means that there is no priorities
  * among the implied ACS and the implied right types (ACL and Bridge, here).
  * @param _ep is a rights user
  * @param _res is an access target. Has to be known in the AlgorithmInterpreter.
  * @param _withExecute true for returning null if there is no effective executing
  *   right in the link.
  * @return a new DisplayableLinkImpl. Null if there is no acl nor bridge to add.
  * */
  public final DisplayableLinkImpl addAclBridgeRightsInSimpleDisplayableLink(
     ImmutableEligibleParty _ep, ImmutableResource _res, boolean _withExecute) {

        // variables
        DisplayableLinkImpl     result = null;
        AccessControlLinkImpl   acLink = null;
        boolean                 withBridge = false;

        if (_ep instanceof Actor) {
            if (((Actor)_ep).getIsACSBridgeSource() && _res == ((Actor)_ep).getEorACSBridgeTarget()) {
                withBridge = true;
                acLink = new AccessControlLinkImpl();
                result = new DisplayableLinkImpl(_ep, _res, acLink);
                acLink.addRight(
                    AccessControlLink.BRIDGE,
                    BaseManagerImpl.getInstance().getMetaRight("execute") );
            } // if
        } // if (_ep

        // updates GLOBAL if there are rights from several ACS
        acLink = addAclInAccessControlLink(_res, _ep, acLink);

        if (acLink == null) {
          return result;
        } // if

        // if withBridge true, there is an executing right,
        // so _withExecute is tested only if withBridge is false
        if (_withExecute && !withBridge) {

            // withExecuteRight is called only if there are not rights from several ACS in acLink
            if (acLink.getLinkTypeComments(AccessControlLink.GLOBAL).
                    indexOf(" ACLs from different ACS: ") == -1) {
                        // apply the _withExecute condition for one ACS
                        if (!utility_.withExecuteRight(
                            (StringRight[])acLink.getRedondantRights().toArray(new StringRight[0]),
                            _res instanceof Node) ) {

                                    // for testing and tracking
                                    //System.out.println("in AclRightsFactoryImpl."+
                                    //"addAclBridgeRightsInSimpleDisplayableLink(), "+
                                      //"the executing condition is not fulfilled,"+
                                      //"with ep: "+_ep.getName()+", res: "+_res.getName());

                                result = null;
                        } // if (!BaseUtilityImpl
                        else {
                                result = new DisplayableLinkImpl(_ep, _res, acLink);
                        } // else if (!BaseUtilityImpl
            } // if (acLink.getLinkTypeComments(
            else {
                // there are rights from several ACS in acLink.
                result = new DisplayableLinkImpl(_ep, _res, acLink);
            } // else if (acLink.getLinkTypeComments(
        } // if (_withExecute
        else if (!withBridge) // result is initialized if withBridge false
            result = new DisplayableLinkImpl(_ep, _res, acLink);

        acLink = null;
        return result;
  } // addAclBridgeRightsInSimpleDisplayableLink()

  /**
  * Adds the AclEntry rights and comments to an AccessControlLinkImpl. 
  * It calls AccessControlLinkImpl.getRight() to filter the rights if there are negative rights.
  * It reads the positive and negative inherited rights if the _res ACS manages
  * the ACL rights inheritance. In this case, if the right user and the target are from the same ACS,
  * it returns the direct rights if the method NodeRights.getL_accessRightsThroughNodesTree()
  * returns an empty, but not null AccessControlLink, and also if
  * NodeRights.withAccessThroughNodesTreeFromEP() returns true.
  * With the same conditions but nodeRights.withAccessThroughNodesTreeFromEP() false,
  * it returns null. With the same conditions but NodeRights.withAccessThroughNodesTreeFromEP()
  * returning null, it returns null.
  * The non-null inherited rights may delete all rights or be addded to the direct ones,
  * through the method NodeRights.mergeInheritedAclPriRightsAndComments(). <P>
  * This method puts the rights as comment. It also puts in the GLOBAL comment all
  * the ACL ACS names if they are more than one. It may adds 'conditional ACL'.
  * The method NodeRights.getMergedInheritedAclPriRightsAndComments() is responsible
  * to add the specific AcsAddon comments on the direct AclEntries, and to set
  * the comment AccessControlLink.C_INHERITED_ACL about the inherited AclEntries.
  * The method NodeRights.withAccessThroughNodesTreeFromEP() may add some additional comments
  * for the inherited AclEntries. <P>
  * Note: this method is sensible since it is used for both the one-link method,
  * the two-links method and a multiple-links method.
  * Called by detectACLRights(), addAclBridgeRightsInSimpleDisplayableLink(),
  * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),
  * ThreeNodesRightsFactoryImpl.selectOneHiddenNodePathsForVirtualFolder(),
  * endsPathsFromActorWithAclBridgePrivilege() and selectOneHiddenNodePathsForFirstLinkPrivileges(),
  * endsPathsFromGroupIDMemberWithAclPrivilege().
  * The argument _upd_acLink is not null only for addAclBridgeRightsInSimpleDisplayableLink().
  * Calls the _res NodeRights.detectL_aclEntryRights(), mergeInheritedAclPriRightsAndComments(),
  * getL_accessRightsThroughNodesTree() and withAccessThroughNodesTreeFromEP().
  * @param _res with some AclEntries to filter. Has to be known in the AlgorithmInterpreter.
  * @param _ep may be associated to some _res AclEntries
  * @param _upd_acLink is the AccessControlLink to update. May be null.
  * @return null if no right has been added, otherwise returns the updated _upd_acLink,
  *  or a new AccessControlLinkImpl if _upd_acLink is null.
  * */
  public final AccessControlLinkImpl addAclInAccessControlLink(
          ImmutableResource _res, ImmutableEligibleParty _ep,
          AccessControlLinkImpl _upd_acLink) {

                                    // for testing and tracking
                                    //if (_res.getNickName().equals("BASE One"))
                                      //  System.out.println("AclRightsFactoryImpl."+
                                        //   "addAclInAccessControlLink() start"+
                                          // " for _res "+_res.getNickName()+" and _ep "+
                                           //_ep.getNickName());

        StringRight[]           l_inheritedRights = null;
        AccessControlLinkImpl   inhLink = null;

        NodeRights        nodeRights = interpreter_.getNodeRightsInstances().get(_res.getName());
        if (nodeRights == null) {
            interpreter_.addInstancesForNode(_res);
            nodeRights = interpreter_.getNodeRightsInstances().get(_res.getName());
        } // if

        // sets l_inheritedRights and inhLink
        if (_res.getEorACS().getManageACLRightsInheritance()) {
          nodeRights = interpreter_.getNodeRightsInstances().get(_res.getName());

                                        // for testing and tracking
                                        //if (nodeRights == null
                                            // && _ep.getNickName().equals("anonymous@localhost")
                                            //)
                                                //System.out.println("AclRightsFactoryImpl."+
                                                  // "addAclInAccessControlLink()"+
                                                   //" with a nodeRights for res: "+
                                                   //nodeRights.getClass());
                                                   //" in the view "+
                                                   //interpreter_.getSource().getNickName()+
                                                   //" for res "+_res.getNickName()+
                                                   //" and for ep "+_ep.getNickName());

          if (_res.getEorACS() == _ep.getEorACS() &&
                  nodeRights.withAccessThroughNodesTreeFromEP(_ep, _res, null)) {

              if (!(_ep instanceof ImmutableGroupID)) {
                  inhLink = nodeRights.getL_accessRightsThroughNodesTree(
                                    _res, _ep, null);
                  if (inhLink != null)
                      l_inheritedRights = inhLink.getRights(AccessControlLink.ACL);
              }
              else {
                  inhLink = nodeRights.getL_accessRightsThroughNodesTree(
                                    _res, null, (ImmutableGroupID)_ep);
                  if (inhLink != null)
                      l_inheritedRights = inhLink.getRights(AccessControlLink.ACL);
              } // else

          } // if

                                        // for testing and tracking
                                        //if (inhLink != null)
                                          //if (_res.getNickName().equals("BASE One"))
                                          //   System.out.println("AclRightsFactoryImpl."+
                                            //       "addAclInAccessControlLink()"+
                                              //     " for _res "+_res.getNickName()+
                                                //   " and _ep "+_ep.getNickName()+
                                                  // " with inhLink: "+inhLink);


          // if inhLink is empty, l_inheritedRights is null but the direct rights are workable
          if (_res.getEorACS() == _ep.getEorACS() && inhLink == null) {
              nodeRights = null;
                                        // for testing and tracking
                                        //System.out.println("AclRightsFactoryImpl."+
                                          //         "addAclInAccessControlLink()"+
                                            //       ", for _ep "+_ep.getNickName()+
                                              //     ", for _res "+_res.getNickName()+
                                                //   ", return null with a null inhLink.");

              return null;
          } // if

        } // if (_res.getEorACS().getManageACLRightsInheritance())

        StringRight           right = null;
        boolean               severalACS = false;
        boolean               conditionalGroups = false;
        char                  open = '[';
        char                  close = ']';
        char                  space = ' ';
        AccessControlLinkImpl result = null;

        // both positive and negative rights only if the AclEntries are from different ACS
        List<StringRight> l_rights = nodeRights.detectL_aclEntryRights(_res, _ep, false);

        if (l_rights == null &&
              (l_inheritedRights == null || l_inheritedRights.length == 0)) {

                                        // for testing and tracking
                                        //if (_res.getNickName().equals("BASE One"))
                                            //System.out.println(
                                              //  "AclRightsFactoryImpl.addAclInAccessControlLink()"+
                                                //" for _res "+_res.getNickName()+" and _ep "+
                                                //_ep.getNickName()+
                                                //" returns null with null direct and null inherited ACLs ");

          return null;
        } // if (l_rights == null)

        if (_upd_acLink != null) result = _upd_acLink;
        else result = new AccessControlLinkImpl();

        if (l_rights != null) {
            for (Iterator ite = l_rights.iterator(); ite.hasNext(); ) {
                right = (StringRight)ite.next();
                if (right == null || right.equals(StringRightImpl.EMPTY_INSTANCE)) continue;

                                //for testing and tracking
                                //System.out.println("AclRightsFactoryImpl.addAclInAccessControlLink()"+
                                  //" add a right: "+right);

                if (right.isGrantingRight()) {
                  result.addRight(AccessControlLink.ACL, right );
                  // put in comments the right - new in v065
                  result.addComment(AccessControlLink.ACL,
                          right.getRight().replace(open, space).replace(close, space) );
                } else {
                  result.addRight(AccessControlLink.NEG_ACL, right);
                  // put in comments the right - new in v065
                  result.addComment(AccessControlLink.NEG_ACL,
                          right.getRight().replace(open, space).replace(close, space) );
                }

            } // for ite
        } // if (l_rights != null)

                                        //for testing and tracking
                                        //if (_ep.getNickName().equals(""))
                                            //System.out.println("AclRightsFactoryImpl.addAclInAccessControlLink()"+
                                              //" intermediate direct result, before the call to "+
                                              //"nodeRights.mergeInheritedAclPriRightsThroughNodesTree(): "+
                                              //result);

        // merge the inherited rights and the direct rights.
        // add some proper comments and some ones from NodeRights.getL_accessRightsThroughNodesTree()
        result = nodeRights.getMergedInheritedAclPriRightsAndComments(
              _ep, _res, result, l_inheritedRights);
        if (l_inheritedRights != null && result != null) {
              result.mergeCommentsWithoutFiltering(inhLink);
        } // if (l_inheritedRights != null)

        // 'result' is null if the non-null inherited rights merging deletes all rights.
        if (result == null) {

                                    // for testing and tracking
                                    //if (_res.getNickName().equals("BASE One"))
                                        //System.out.println(
                                          //"end of AclRightsFactoryImpl.addAclInAccessControlLink()"+
                                          //" for _res "+_res.getNickName()+" and _ep "+
                                          //_ep.getNickName()+"- null result after "+
                                          //"nodeRights.getMergedInheritedAclPriRightsAndComments()");

            l_rights = null;
            right = null;
            l_inheritedRights = null;
            inhLink = null;
            return result;
        } // if

        // l_acl is used only for the comments in the result
        AclEntry[]        l_acl = _res.getL_aclEntry();

        if (l_acl == null || l_acl.length == 0) {

                                    // for testing and tracking
                                    //if (_res.getNickName().equals("BASE One"))
                                        //System.out.println(
                                          //"end of AclRightsFactoryImpl.addAclInAccessControlLink()"+
                                          //" for _res "+_res.getNickName()+" and _ep "+
                                          //_ep.getNickName()+"- result with null direct ACL: "+result);

            l_rights = null;
            right = null;
            l_inheritedRights = null;
            inhLink = null;
            return result;
        } // if

        StringBuilder     comment = null;

        // put the comment 'conditional ACL' from the direct ACLs, with the group names
        // note: the same comment from the inherited ACLs in an AcsAddon,
        // should be done in a subclass of NodeRightsImpl, in getL_accessRightsThroughNodesTree().
        for (int i = 0; i < l_acl.length; i++) {
            if (l_acl[i] == null ||
                    l_acl[i].getEorEP() != _ep ||
                    l_acl[i].getL_Rights().length == 0 || // includes the ACL desabling
                    !l_acl[i].isConditionalAclEntry())
                continue;

            // add for conditional ACLs, the name of the first conditional group
            if (!conditionalGroups) {
                conditionalGroups = true;
                if (comment == null)
                    comment = new StringBuilder(AccessControlLink.C_CONDITIONAL_ACL);

                  // grp name as comment
                comment.append(
                        " through "+
                        ((GroupID)l_acl[i].getL_ConditionSourceGroups().get(0)).getNickName());

            } // if (!conditionalGroups)
            else {
                comment.append(
                        ", "+
                        ((GroupID)l_acl[i].getL_ConditionSourceGroups().get(0)).getNickName());
            } // else

                                        //for testing and tracking
                                    //System.out.println(
                                      //"NodeRightsImpl.getMergedInheritedAclPriRightsThroughNodesTree()"+
                                      //" for _res "+_res.getNickName()+" and _ep "+
                                      //_ep.getNickName()+" adding of the comment "+
                                      //"'conditional ACL' for the group "+
                                      //((GroupID)l_acl[i].getL_ConditionSourceGroups().get(0)).getNickName());

        } // for i

        if (conditionalGroups) {
            result.setLinkTypeToTrue(AccessControlLink.ACL);
            result.addComment(AccessControlLink.ACL, comment.toString());
        } // if (conditionalGroups)


        // puts in comments the acl ACS names, if they are several ones
        for (int i = 1; i < l_acl.length; i++) { // from 1
            if (l_acl[i] == null ||
                    l_acl[i].getEorEP() != _ep ||
                    l_acl[i].getL_Rights().length == 0)
                continue;

            // detects the first case where there are 2 ACS
            if (l_acl[i].getEorACS() != l_acl[i - 1].getEorACS() && !severalACS) {
                severalACS = true;
                comment = new StringBuilder("ACLs from different ACS: ");
                comment.append(l_acl[i - 1].getEorACS().getNickName() + ", ");

                                        // for testing and tracking
                                    //System.out.println(
                                      //"AclRightsFactoryImpl.addAclInAccessControlLink()"+
                                      //" for _res "+_res.getNickName()+" and _ep "+
                                      //_ep.getNickName()+" adding for the comment "+
                                      //"ACLs from different ACS");

            } // if (l_acl[i].getEorACS() !=

            if (severalACS && comment.indexOf(l_acl[i].getEorACS().getNickName()) == -1) {
                comment.append(l_acl[i].getEorACS().getNickName());
                if (i != l_acl.length - 1) comment.append(", ");
            } // if (severalACS
        } // for i

        if (severalACS) {
            result.setLinkTypeToTrue(AccessControlLink.GLOBAL);
            result.addComment(AccessControlLink.GLOBAL, comment.toString());
        } // if (severalACS)

                                    // for testing and tracking
                                    //if (_res.getNickName().equals("BASE One"))
                                        //System.out.println(
                                          //"end of AclRightsFactoryImpl.addAclInAccessControlLink()"+
                                          //" for _res "+_res.getNickName()+" and _ep "+
                                          //_ep.getNickName()+"- result: "+result);

        l_acl = null;
        comment = null;
        l_rights = null;
        right = null;
        l_inheritedRights = null;
        inhLink = null;
        return result;
  } // addAclInAccessControlLink()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
    interpreter_ = null;
    utility_ = null;
    linkUtil_ = null;
  } // finalizeForProcess()
  
} // AclRightsFactoryImpl
