//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;
import java.util.Collection;
import java.util.Arrays;

import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.CNot.AccessControlLink;
import ARoad0.CNot.AccessControlLinkImpl;

import ARoad0.Gui2.GraphicView;

import ARoad0.gBaseInterface.*;
import ARoad0.gWorkInterface.*;

import ARoad0.gBase.UserIDImpl; 
import ARoad0.gBase.GroupIDImpl; 
import ARoad0.gBase.BaseManagerImpl;

/**
 * This class is responsible for all the core algorithms about Account/Group and
 * Bridge rights.
 * These rights are based mainly on UserID rights, GroupID rights
 * in operating systems.
 * This class is stateless, and RightsMediatorImpl at its creation
 * setups an instance for its ViewInBase.
 * The unique direct call to ACS is for ACS.getManageSecGroupsForActor().
 * The upper-level methods begin by detect...(),
 * The middle-level methods begin by select...(),
 * The lower-level methods begin by add...().
 * Reads the ACS structure for the property: <P>
 * 'AGO right inheritance 2: at each access, if the real account
 * is not the parent account, it uses first its primary group inherited rights,
 * if it is the parent group'. <P>
 * @version 0.7.0
 */
public final class AgoRightsFactoryImpl implements CoreAlgorithm {
  private AlgorithmInterpreter      interpreter_ = null;
  private DisplayableLinkUtilities  linkUtil_ = null;

  /**
  * only one protected constructor
  * @param _interpreter algorithm interpreter
  */
  protected AgoRightsFactoryImpl(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
  }

  /**
  * Defines the effective applicable rights of an UserID for a Resource
  * which it owns (between an actor and its current userID)
  * when they are in the view, defines the effective rights between an actor 
  * and a resource when their common userID and they are in the view,
  * and defines the effective rights between an actor and its current userID.
  * Called by RightsFactory_Facade.detectSketchRights().
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @param _viewInBase EPRViewInBase
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  * with Owner links associated to every pair (EP, Resource).
  * */
  protected final Map detectUserRights(EPRViewInBase _viewInBase,
    Map _m_l_DisplayableLinks, BaseObject _center) {
      
      return detectOwnerContainRights(
              _viewInBase, _m_l_DisplayableLinks, true, _center);
      
  } // detectUserRights()

  /**
  * Defines all types of effective applicable direct group rights.
  * Defines the effective applicable rights of
  * a GroupID for a Resource which is its member.
  * Defines the effective rights between an actor and a resource
  * when their common groupID is on the view.
  * Defines the effective rights for an actor through its secondary groups.
  * Defines also the direct member relations between each GroupIDMember
  * and the GroupIDs in the view. Apply the rule 'AGO right: A rights overlay G rights'.
  * Called by RightsFactory_Facade.detectSketchRights().
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @param _viewInBase EPRViewInBase
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  * with Contain links associated to every pair (EP,Resource).
  * */
  protected final Map detectGroupRights(EPRViewInBase _viewInBase,
    Map _m_l_DisplayableLinks, BaseObject _center) {

      ImmutableSource       so;
      ImmutableSource       so2;
      EligibleParty         ep;
      EligibleParty         grp;
      Collection            l_ep = _viewInBase.getEorM_Sources().values();
      DisplayableLinkImpl   dLink;
      LinkRights            linkRights = null;
      
      // detects the 'contain' relations from an owner to a resource in the view
      Map m_result = detectOwnerContainRights(_viewInBase, _m_l_DisplayableLinks, false, _center);

                     // for testing and tracking
                    //System.out.println("In LinuxRightsFactory.detectGroupRights(), "+
                      //"the DisplayableLinkImpls after detectOwnerContainRights() are:");
                    //ImmutableName key;
                    //for (Iterator ite = m_result.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--for key: "+key);
                      // for (Iterator ita = ((List)m_result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
      
      // for an actor to its secondary GroupID in the view,
      // if its current userID is not null
      m_result = detectSecondaryGroupRights(_viewInBase, m_result, _center);

                     // for testing and tracking
                    //System.out.println("In LinuxRightsFactory.detectGroupRights(), "+
                      //"the DisplayableLinkImpls after detectSecondaryGroupRights() are:");
                    //ImmutableName key;
                    //for (Iterator ite = m_result.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--for key: "+key);
                      // for (Iterator ita = ((List)m_result.get(key)).iterator() ; ita.hasNext() ; ) {
                       // System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
      
      // detects all the direct member relations
      // between a GroupIDMember and a GroupID in the view
      for (Iterator ita = l_ep.iterator() ; ita.hasNext() ; ) {
          so = (ImmutableSource) ita.next();
          if (so == null || !(so instanceof GroupIDMember)) continue;
          ep = (EligibleParty) so;

          for (Iterator ite = l_ep.iterator() ; ite.hasNext() ; ) {
              so2 = (ImmutableSource) ite.next();
              if (so2 == ep || !(so2 instanceof ImmutableGroupID)) continue;
              grp = (EligibleParty) so2;
              if (!(((ImmutableGroupID)grp).containsAsMember((GroupIDMember)ep))) continue;
                // for a sketch view
              else if (_center != null && _center != ep && _center != grp) continue;
              else {
                  // updates m_result
                  dLink = linkUtil_.getDisplayableLink(ep, grp, null, m_result);
                  if (dLink == null) {
                      dLink = new DisplayableLinkImpl(ep, grp, (BaseObject[])null);
                      linkRights = interpreter_.getLinkRightsInstances().get(grp.getName());
                      dLink = linkRights.addDisplayableLink(dLink, m_result);
                  } // else
                  if (dLink != null)
                      dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_MEMBER);
                  
                    // for testing and tracking
                    //System.out.println("In AgoRightsFactoryImpl.detectGroupRights(), "+
                            //"the link is set to true for the IS_MEMBER type: "+dLink);
                  
              } // else

          } // for

      } // for

      so  = null;
      so2 = null;
      ep = null;
      grp = null;
      l_ep = null;
      dLink = null;
      return m_result;
  } // detectGroupRights()

  /**
  * Defines the effective applicable rights of an EligibleParty to a Resource (not
  * to a GroupIDMember or a VirtualFolder) in _viewInBase, when there is
  * exactly one intermediate node, and only AGO rights.
  * When the intermediate node is in the view, it is processed by detectUserRights()
  * and detectGroupRights().<P>
  * An Actor is an EligibleParty, but its processing is different. An actor
  * has no proper AGO rights whether it does not run under a current UserID.
  * Then, the current UserID, the current GroupID and
  * the Actor secondary groups are processed.
  * The path (actor started from xid/current account-group/actor started from other xid)
  * is detected in CompoundRightsFactoryImpl.
  * When there is a chain of hidden userID/groupID/bridges/ACLs
  * which defines the rights, it is processed by detectHiddenChainedGroupsRights().
  * _m_l_DisplayableLinks is updated only with n DisplayableLinks, where n
  * is the number of GroupID paths through which a _viewInBase EligibleParty
  * has an indirect access right on a _viewInBase Resource.
  * All the links are stored in _m_l_DisplayableLinks.
  * Applies the rule 'AGO right: A rights overlay G rights'.
  * A link is build up for the UserID that owns the target, even whithout rights.
  * Called by RightsFactory_Facade.detectEPRRights() and detectNoThanRights().
  * Caution: called by detectNoThanRights(), _viewInBase is not the view of _gview,
  * since this method uses transient EPRViewInBases with for each of them,
  * a pair of nodes from the initial NoThanViewInBase.
  * Calls addOwnerContainRightsToDisplayableLink(),
  * NodeRights.selectHiddenDirectOwnerContainGlobalForActorAsEP(),
  * selectHiddenDirectOwnerContainForActorAsResource(),
  * selectHiddenDirectOwnerContainForTarget().
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @return Map of DisplayableLinks, with hidden UserID/GroupID/Actor links 
  *          associated to every pair (EP,Resource).
  * @exception InterruptedException
  * */
  public final Map detectOneHiddenNodeWithCommonAGORights(GraphicView _gview,
    EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks) throws InterruptedException {


                    // for testing and tracking
                    //System.out.println("In AgoRightsFactoryImpl."+
                      //"detectOneHiddenNodeWithCommonAGORights(), "+
                      //"the DisplayableLinkImpls at the start are:");
                    //ImmutableName key;
                    //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      // variables
      ImmutableSource         so;
      EligibleParty           ep;
      GroupIDMember           ep_2 = null;
      Resource                res;
      ImmutableTarget         targ;
      EligibleParty           hiddenDirectEP;
      Set                     l_hiddenDirectEP = null;
      Set                     l_hiddenDirectOwner;
      DisplayableLinkImpl     dLink = null;
      HashSet                 l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();
      NodeRights              nodeRights = null;
      NodeRights              nodeRights2 = null;
      LinkRights              linkRights = null;

      // iterate on the view eligible parties 
      for (Iterator ite = _viewInBase.getEorM_Sources().values().
        iterator() ; ite.hasNext() ; ) {

        so = (ImmutableSource) ite.next();
        if (so == null || so instanceof VirtualFolder) continue;
        ep = (EligibleParty) so;
        nodeRights = interpreter_.getNodeRightsInstances().get(ep.getName());

        if (!(ep instanceof Actor))
          l_hiddenDirectEP = nodeRights.selectHiddenDirectGroupIDForGroupIDMember(
            _viewInBase, (GroupIDMember)ep);
        else 
          l_hiddenDirectEP = nodeRights.selectHiddenDirectOwnerContainGlobalForActorAsEP(
            _viewInBase, (Actor)ep);

        // iterates on Resource in the view
        for (Iterator iti = _viewInBase.getEorM_Targets().values().iterator() ;
          iti.hasNext() ; ) {
            targ = (ImmutableTarget) iti.next();
            if (!(targ instanceof Resource) || targ == ep) continue;
            res = (Resource) targ;
            nodeRights2 = interpreter_.getNodeRightsInstances().get(res.getName());

                        // for testing and tracking
                        //System.out.println("in AgoRightsFactoryImpl.detectOneHiddenNodeWithCommonAGORights()"+
                          //", interpreter_.getNodeRightsInstances(): ");
                        //for (Object obj : interpreter_.getNodeRightsInstances().keySet()) {
                          //  System.out.println("--- :"+obj);
                        //}

            if (res instanceof Actor)
              l_hiddenDirectOwner = nodeRights2.selectHiddenDirectOwnerContainForActorAsResource(
                _viewInBase, (Actor)res);
            else l_hiddenDirectOwner = nodeRights2.selectHiddenDirectOwnerContainForTarget(
              _viewInBase, res);

                        // for testing and tracking
                        //System.out.println("in AgoRightsFactoryImpl."+
                          //"detectOneHiddenNodeWithCommonAGORights()"+
                          //", loop on the resource: "+res.getName());
            
            // ACS rule: 'AGO right: A rights overlay G rights'
            if ((ep instanceof Actor) && ((Actor)ep).getEorCurrentUserID() == res.getEorUserID()) {
                if (!_viewInBase.containsSource(res.getEorUserID())) {
                    if (l_hiddenDirectEP != null)
                        l_hiddenDirectEP.clear();
                    else
                        l_hiddenDirectEP = new HashSet(5);
                    l_hiddenDirectEP.add(res.getEorUserID());
                } 
                // res.getEorUserID() is in the view
                else continue; 
            } // if
            
            // selects the set of Not-In-View direct UserID and GroupID
            // which contains/owns ep (or for an actor, under which it runs), 
            // and setup each last node before ep as ep_2
            if (l_hiddenDirectEP != null && !l_hiddenDirectEP.isEmpty()) {
                for (Iterator ita = l_hiddenDirectEP.iterator() ; ita.hasNext() ; ) {

                    hiddenDirectEP = (EligibleParty) ita.next();

                    // compares the hiddenDirectEP
                    // to the Not-In-View Direct UserID and GroupID of res,
                    // and adds to _m_l_DisplayableLinks a new DisplayableLinkImpl
                    // that is created from dLink, for the pair (ep, res)
                    if (l_hiddenDirectOwner != null) {
                        for (Iterator ito = l_hiddenDirectOwner.iterator(); ito.hasNext() ; ) {

                            ep_2 = (GroupIDMember) ito.next();

                            if (hiddenDirectEP != ep_2) continue;
                            linkRights = interpreter_.getLinkRightsInstances().get(res.getName());

                            // dLink is build up for an UserID, even whithout inherited rights
                            if ((ep_2 instanceof ImmutableGroupID) &&
                                    !nodeRights2.withAccessThroughNodesTreeFromEP(
                                        ep, res, ep_2))
                                continue;

                            dLink = new DisplayableLinkImpl(ep, res, new EligibleParty[] {ep_2});

                            dLink = addOwnerContainRightsToDisplayableLink(dLink);

                                                // for testing and tracking
                                                //System.out.println("in AgoRightsFactoryImpl."+
                                                  //"detectOneHiddenNodeWithCommonAGORights(), new dLink: "+dLink);

                            linkRights = interpreter_.getLinkRightsInstances().get(res.getName());
                            dLink = linkRights.addDisplayableLink(
                                    dLink, _m_l_DisplayableLinks);

                            if (dLink != null)
                                l_ObjectToRegister.add(ep_2);

                        } // for (Iterator ito
                    } // if (l_hiddenDirectOwner != null

                } // for (Iterator ita
            } // if l_hiddenDirectEP != null

        } // for iti - number 1 

        if (Thread.interrupted()) {
            throw new InterruptedException();
        }

      }  // for (Iterator ite
      
      _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);
 
                    // for testing and tracking
                    //System.out.println("In AgoRightsFactoryImpl."+
                      //"detectOneHiddenNodeWithCommonAGORights(), "+
                      //"the DisplayableLinkImpls at the end are:");
                    //ImmutableName key;
                    //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
      
      
    res = null;
    ep = null;
    ep_2 = null;
    hiddenDirectEP = null;
    l_hiddenDirectEP = null;
    l_hiddenDirectOwner = null;
    dLink = null;
    so = null;
    targ = null;
    nodeRights = null;
    nodeRights2 = null;
    linkRights = null;

    return _m_l_DisplayableLinks;
  } // detectOneHiddenNodeWithCommonAGORights()

  /**
  * Defines the effective applicable rights of an EligibleParty for a Resource
  * in _viewInBase, when there are several UserID or GroupID external to the view,
  * which may define specific access control links based on relations 
  * (EligibleParty to UserID/GroupID), (Actor to BridgeTarget to UserID/GroupID),
  * (GroupIDMember to GroupID), (UserID/GroupID to Resource) and
  * (EligibleParty to Actor in another access context). <P>
  * _upd_m_l_DisplayableLinks is updated only with n DisplayableLinks, where n
  * is the number of GroupID paths through which a _viewInBase EligibleParty
  * has an indirect access right on a _viewInBase Resource.
  * All the contain, own and bridge relations are stored in _upd_m_l_DisplayableLinks.
  * Called by RightsFactory_Facade.detectRights() and detectNoThanRights().
  * Caution: called by detectNoThanRights(), _viewInBase is not the view of _gview,
  * since this method uses transient EPRViewInBases with for each of them,
  * a pair of nodes from the initial NoThanViewInBase. <P>
  * Calls NodeRightsImpl.selectHiddenButNoDirectGroupLinks() and
  * selectHiddenDirectOwnerContainForTarget(),
  * LinkRightsImpl.withAccessThroughNodesTreeFromEP() and
  * detectAccountPriorityInLastLinkAGORights(),
  * RightsFactoryUtilities.addOwnerContainRightsToDisplayableLink().
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze
  * @param _upd_m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @exception InterruptedException
  * @return _upd_m_l_DisplayableLinks,
  *    with hidden UserID/GroupID links associated to every pair (EP, Resource).
  * */
  public final Map detectHiddenChainedGroupsRights(GraphicView _gview,
    EPRViewInBase _viewInBase, Map _upd_m_l_DisplayableLinks) throws InterruptedException {
      // variables
      ImmutableSource         so;
      EligibleParty           ep;
      GroupIDMember           ep_2;
      Resource                res;
      ImmutableTarget         targ;
      ImmutableActor          act;
      DisplayableLinkImpl     dLink;
      DisplayableLinkImpl     dLink_2;
      DisplayableLinkImpl     dLink_3;
      ImmutableGroupIDMember  epOwner;
      Set                     l_processedPairKeys = new HashSet(INITIAL_CAPACITY);
      Name                    pairKey_1;
      AccessControlLinkImpl   lastLink;
      AccessControlLinkImpl   acLink;
      Set                     l_hiddenButNoDirectGroupLinks;
      Set                     l_hiddenDirectOwnerForTarget;
      HashSet                 l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();
      NodeRights              nodeRights = null;
      NodeRights              nodeRights2 = null;
      LinkRights              linkRights = null;

      // iterate on EP in the view
      for (Iterator ite = _viewInBase.getEorM_Sources().values().
        iterator() ; ite.hasNext() ; ) {
        so = (ImmutableSource) ite.next();
        if (so == null || !(so instanceof EligibleParty)) continue;
        ep = (EligibleParty) so;
        nodeRights = interpreter_.getNodeRightsInstances().get(ep.getName());

        l_hiddenButNoDirectGroupLinks = nodeRights.selectHiddenButNoDirectGroupLinks(
              _viewInBase, ep);

        // adds the bridge target current AG, if they exist
        if (ep instanceof Actor && ((Actor)ep).getIsACSBridgeSource() ) {

            act = ((Actor)ep).getEorACSBridgeTarget();
            if (!_viewInBase.containsSource(act) && act.getIsRightProxy()) {
                if ( act.getEorCurrentUserID() != null &&
                  !_viewInBase.containsSource( act.getEorCurrentUserID() )) {
                    acLink = new AccessControlLinkImpl();
                    acLink.addRight(
                                AccessControlLink.BRIDGE,
                                BaseManagerImpl.getInstance().getMetaRight("execute") );

                    dLink = new DisplayableLinkImpl(ep, act, acLink);
                    lastLink = new AccessControlLinkImpl();
                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, 
                      act.getEorCurrentUserID(), lastLink);
                    dLink_2 = addOwnerContainRightsToDisplayableLink(dLink_2);

                    if (l_hiddenButNoDirectGroupLinks == null)
                        l_hiddenButNoDirectGroupLinks = new HashSet(5);

                    l_hiddenButNoDirectGroupLinks.add(dLink_2);
                }
                if (act.getEorCurrentGroupID() != null &&
                  !_viewInBase.containsSource( act.getEorCurrentGroupID() )) {
                    acLink = new AccessControlLinkImpl();
                    acLink.addRight(
                                AccessControlLink.BRIDGE,
                                BaseManagerImpl.getInstance().getMetaRight("execute") );

                    dLink = new DisplayableLinkImpl(ep, act, acLink);
                    lastLink = new AccessControlLinkImpl();
                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, 
                      act.getEorCurrentGroupID(), lastLink);
                    dLink_2 = addOwnerContainRightsToDisplayableLink(dLink_2);

                    if (l_hiddenButNoDirectGroupLinks == null)
                        l_hiddenButNoDirectGroupLinks = new HashSet(5);

                    l_hiddenButNoDirectGroupLinks.add(dLink_2);
                } // if
            } // if
        } // if
        
        // iterates on Resource in the view
        for (Iterator iti = _viewInBase.getEorM_Targets().values().iterator() ;
          iti.hasNext() ; ) {
            targ = (ImmutableTarget) iti.next();
            if (!(targ instanceof Resource) || targ == ep) continue;
            res = (Resource) targ;

            // construct the keys with the names of ep and res
            pairKey_1 = ((Name)ep.getName()).extended(res.getName());
            if (l_processedPairKeys.contains(pairKey_1)) continue;
            else l_processedPairKeys.add(pairKey_1);

            nodeRights2 = interpreter_.getNodeRightsInstances().get(res.getName());
            linkRights = interpreter_.getLinkRightsInstances().get(res.getName());

            l_hiddenDirectOwnerForTarget = nodeRights2.selectHiddenDirectOwnerContainForTarget(
                                                _viewInBase, res);

            if (l_hiddenDirectOwnerForTarget == null) continue;

            // selects the set of Not-In-View Applicable UserID and GroupID
            // which contains/owns ep, and set each last node before ep
            // as epUserIDandGroupIDs
            if (l_hiddenButNoDirectGroupLinks != null) {
                for (Iterator ita = l_hiddenButNoDirectGroupLinks.iterator() ; ita.hasNext() ; ) {

                    dLink = (DisplayableLinkImpl)ita.next();

                    epOwner = (ImmutableGroupIDMember) dLink.getEorSecondEnd();

                    l_ObjectToRegister.addAll(Arrays.asList(dLink.getEorAllNodes()));

                    if (!l_hiddenDirectOwnerForTarget.contains(epOwner))
                        continue;

                                  // for testing and tracking
                                  //System.out.println(
                                    //"in AgoRightsFactoryImpl.detectHiddenChainedGroupsRights()"+
                                    //" for the EP "+ep.getName()+
                                    //", one hidden not-direct owner is "+
                                    //epOwner.getName() );

                    // detects all direct 'root' relations
                    if ((epOwner instanceof UserID) && ((UserID)epOwner).getOrder() == 0)
                        continue;

                    if (!nodeRights2.withAccessThroughNodesTreeFromEP(ep, res, epOwner))
                        continue;

                    // each possible path produces a new DisplayableLink
                    // where res is the second end
                    lastLink = new AccessControlLinkImpl();
                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, res, lastLink);
                    if (dLink_2 == null) continue;

                    // ACS rule is:
                    // 'AGO right: A rights overlay G rights'
                    // no dLink_2 to add if it is a group and if the userID rights
                    // are already applicable
                    pairKey_1 = ((Name)((BaseObject)dLink_2.getEorFirstEnd()).
                          getName()).extended(res.getName());
                    if (_upd_m_l_DisplayableLinks.get(pairKey_1) != null) {
                        if (linkRights.detectAccountPriorityInLastLinkAGORights(
                                dLink_2, (List)_upd_m_l_DisplayableLinks.get(pairKey_1)))
                            continue;
                    } // if

                    dLink_2 = addOwnerContainRightsToDisplayableLink(dLink_2);

                    if (dLink_2 != null) {
                        // updates _upd_m_l_DisplayableLinks without checking on AGO priorities
                        linkRights = interpreter_.getLinkRightsInstances().get(res.getName());
                        dLink = linkRights.addDisplayableLink(
                            dLink_2, _upd_m_l_DisplayableLinks);
                    } // if (dLink_2
                    // else, rights are null in dLink_2

                } // for ita
            } // if (l_hiddenButNoDirectGroupLinks != null)

        } // for (Iterator iti
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }

      }  // for (Iterator ite
      
                // for testing and tracking
                //System.out.println("in AgoRightsFactoryImpl.detectHiddenChainedGroupsRights(), "+
                  //      "at the end, l_RightsRelatedBaseObjects_ contains:");
                //for (Iterator ite = l_RightsRelatedBaseObjects_.iterator(); ite.hasNext(); ) {
                  //  System.out.println(((BaseObject)ite.next()).getName());
                //} // for ite
      
      _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);
      
      res = null;
      targ = null;
      ep = null;
      ep_2 = null;
      act = null;
      dLink = null;
      epOwner = null;
      l_processedPairKeys = null;
      pairKey_1 = null;
      dLink_2 = null;
      acLink = null;
      lastLink = null;
      l_hiddenButNoDirectGroupLinks = null;
      l_hiddenDirectOwnerForTarget = null;
      dLink_3 = null;
      so = null;
      nodeRights = null;
      nodeRights2 = null;
      linkRights = null;

      return _upd_m_l_DisplayableLinks;
  } // detectHiddenChainedGroupsRights()

  /**
  * Updates a DisplayableLinkImpl with group and user rights, including for 'root'.
  * The rights are updated following the type of the last node (GroupID, UserID)
  * connected to the resource which is the second end. If this end
  * is a GroupID or an UserID, the link types and the rights are also set.
  * An empty owner relation is set for forbidding the other DisplayableLinks
  * based on the group. <P>
  * When the first node is an actor and the next node is its current user or group,
  * a RUN_UNDER relation is set. When the GLOBAL relation is get through
  * the secondary group of the first node,
  * it is put in a comment AccessControlLink.C_SECONDARY_GROUP in the link
  * from the actor to its secondary group. This is also done when the actor is
  * before the last node. <P>
  * For all the intermediates nodes, the single operation is for the MEMBER
  * relation from a GroupIDMember to a GroupID. In a chain of nodes,
  * IS_INDIRECT_MEMBER replaces IS_MEMBER in the last link.
  * This is to reply to the detectHiddenChainedGroupsRights() needs.
  * Updates also root rights. Processes _dLink with a bridge, but does not add
  * the BRIDGE type or any comment in the link. <P>
  * Called by detectOneHiddenNodeWithCommonAGORights(),
  * detectHiddenChainedGroupsRights(),
  * ThreeNodesRightsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeAlias().
  * Calls addOwnerContainRightsToLastLinkInDisplayableLink().
  * @param _dLink is a simple or not-simple link:
  *  - with a Resource as second end, and a GroupID or UserID as previous node,
  *  - or with a GroupID or UserID as second end, and a member or an actor as previous node,
   * - and/or an Actor as first end with its current user/group as next node.
  * @return the updated _dLink, or null if _dLink is null, or if
  *   addOwnerContainRightsToLastLinkInDisplayableLink() returns null.
  * */
  public final DisplayableLinkImpl addOwnerContainRightsToDisplayableLink(
     DisplayableLinkImpl _dLink) {

        if (_dLink == null) return null;

        DisplayableLinkImpl dLink = addOwnerContainRightsToLastLinkInDisplayableLink(_dLink);

        if (dLink == null) dLink = _dLink;

        BaseObject[]            l_nodes = _dLink.getEorAllNodes();
        AccessControlLinkImpl[] l_links = _dLink.getEorAllACLinks();
        Actor                   act;
        GroupIDMember           member;
        boolean                 indirectMember = true;

                            // for testing and tracking
                            //System.out.println("in AgoRightsFactoryImpl."+
                              //"addOwnerContainRightsToDisplayableLink(), "+
                              //"l_links is: ");
                            //for (int i = 0; i < l_links.length; i++) {
                              //System.out.println("------ "+l_links[i].toString());
                            //}

        // set the type on each other acLink of dLink
        // if the intermediate nodes are GroupIDMembers having the IS_INDIRECT_MEMBER relation,
        // and no operation if it is for example an actor
        if (l_links.length > 1) {
          for (int i = 0; i < l_links.length - 1; i++) {
            // l_nodes[i+1] is the node after l_links[i] in dLink
            if (l_nodes[i+2] instanceof GroupID &&
                    l_nodes[i+1] instanceof GroupID &&
                    l_nodes[i] instanceof GroupIDMember) {

                  l_links[i+1].setLinkTypeToFalse(AccessControlLink.IS_MEMBER);
                  l_links[i+1].setLinkTypeToTrue(AccessControlLink.IS_INDIRECT_MEMBER);

                  // detect the indirect member relation
                  if (!indirectMember) {
                      indirectMember = true;
                  }

                            // for testing and tracking
                            //System.out.println("in AgoRightsFactoryImpl."+
                              //"addOwnerContainRightsToDisplayableLink(), "+
                              //"adds indirect member in: "+_dLink);

            } // if (l_nodes[i+2]
          } // for i
        } // if (l_links.length > 1) {

        // for an Actor as first end, set RUN_UNDER
        if (dLink.getEorFirstEnd() instanceof Actor) { 
            if (dLink.getEorAllNodes()[1] instanceof GroupIDMember) {
                // set the GLOBAL type if the first end is an actor and the next node,
                // its current user/group or secondary group            
                  act = (Actor)dLink.getEorFirstEnd();
                  member = (GroupIDMember)dLink.getEorAllNodes()[1];
                  if (member == ((Actor)act).getEorCurrentGroupID() ||
                    member == ((Actor)act).getEorCurrentUserID() ) {
                         dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                  } // if
                  else if (member instanceof GroupID &&
                     act.isMemberToSecondaryGroup((GroupID)member)) {
                        dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                        dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink.getEorFirstLink().addComment(AccessControlLink.GLOBAL,
                                AccessControlLink.C_SECONDARY_GROUP);
                  } // else
            }
        } // if (dLink.getEorFirstEnd() instanceof Actor        

                          // for testing and tracking
                          //System.out.println("in AgoRightsFactoryImpl."+
                            //"addOwnerContainRightsToDisplayableLink(), "+
                            //"set in the last link the rights "+rightType+
                            //" for the link: "+dLink);
                          //for (int i = 0; i < l_rights.length; i++) {
                            //System.out.println("----- for "+i+": "+l_rights[i]);
                          //}

        l_nodes = null;
        l_links = null;
        act = null;
        member = null;

        return dLink;
  } // addOwnerContainRightsToDisplayableLink()

  /**
  * Updates a DisplayableLinkImpl with group and user rights, including for 'root'.
  * The rights are updated following the type of the last node (GroupID, UserID)
  * connected to the resource which is the second end. If this end
  * is a GroupID or an UserID, the link types and the rights are also set.
  * An empty owner or group relation is set for forbidding the other DisplayableLinks,
  * if the owner (or group) rights are null. <P>
  * When the last node is the current user or group of the previous node,
  * a RUN_UNDER relation is set. When it is the secondary group,
  * a GLOBAL relation is set with a comment AccessControlLink.C_SECONDARY_GROUP. <P>
  * Updates also root rights. Processes _dLink with a bridge, but does not add
  * the BRIDGE type or any comment in the link. <P>
  * If the method returns null, the argument is not updated.
  * Called by addOwnerContainRightsToDisplayableLink(),
  * NodeRightsImpl.selectHiddenButNoDirectGroupLinks(),
  * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),
  * ThreeNodesrightsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeMemberOwnContain().
  * Caution: does not call NodeRights.withAccessThroughNodesTreeFromEP(),
  * and this is why it has to be called before, by the caller of this method, to be sure
  * that there is a DisplayableLinkImpl to set.
  * Calls LinkRights.getL_accessRightsThroughNodesTree() if the last link is from
  * a GroupIDMember to a Resource, and throws an InternalError if these AGO inherited rights
  * are null or empty.
  * @param _dLink is a simple or not-simple link:
  *  - with a Resource as second end, and a GroupID or UserID as previous node,
  *  - or with a GroupID or UserID as second end, and a member or an actor as previous node.
  * @return the updated _dLink, or null if _dLink is null, or if all
  *   the rights are not applicable nor link type settings.
  * */
  public final DisplayableLinkImpl addOwnerContainRightsToLastLinkInDisplayableLink(
     DisplayableLinkImpl _dLink) {

        if (_dLink == null) return null;

        BaseObject[]            l_nodes = _dLink.getEorAllNodes();
        ImmutableSource         nodeBeforeSecondNode = (ImmutableSource)
                                  l_nodes[l_nodes.length - 2];
        AccessControlLinkImpl   link = _dLink.getEorLastLink();
        StringRight[]           l_rights = new StringRight[0];
        String                  rightType = null;
        boolean                 added = false;
        boolean                 added_one = false;
        boolean                 added_2 = false;
        boolean                 added_3 = false;
        Resource                res = null;
        NodeRights              nodeRights = null;
        AccessControlLinkImpl   inhLinkRights = null;

                            // for testing and tracking
                            //System.out.println("in AgoRightsFactoryImpl."+
                              //"addOwnerContainRightsToDisplayableLink(), "+
                              //"the argument is: "+_dLink);
        
        if (nodeBeforeSecondNode instanceof GroupID &&
                _dLink.getEorSecondEnd() instanceof Resource) {
            
          res = (Resource)_dLink.getEorSecondEnd(); 
          if (res.getEorGroupID() == nodeBeforeSecondNode ) {
              rightType = AccessControlLink.CONTAIN;

              l_rights = res.getL_groupRights();
              if (l_rights.length == 0) {
                // an empty contain dLink is set
                link.setLinkTypeToTrue(rightType);
                added_one = true;
              }
          } // if

        } // if (owner instanceof GroupID)
        else if (nodeBeforeSecondNode instanceof UserID &&
                _dLink.getEorSecondEnd() instanceof Resource) {
          res = (Resource)_dLink.getEorSecondEnd();
          
          if (((UserID)nodeBeforeSecondNode).getOrder() != 0 ) {
            l_rights = res.getL_ownerRights();
          } else {
            l_rights = res.getL_rootRights();             
          } 

          if (res.getEorUserID() == nodeBeforeSecondNode) {
              rightType = AccessControlLink.OWNER;
              if (l_rights.length == 0) {
                // an empty owner dLink is set
                link.setLinkTypeToTrue(rightType);
                added_one = true;
              }                
          } // if (res.getEorUserID()          
          
        } // else if
        else if (_dLink.getEorSecondEnd() instanceof GroupID) {
          added_2 = true;
          if (nodeBeforeSecondNode instanceof GroupIDMember) {
              _dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.IS_MEMBER);
          } else if (nodeBeforeSecondNode instanceof Actor &&
              _dLink.getEorSecondEnd() == ((Actor)nodeBeforeSecondNode).getEorCurrentGroupID()) {
                 _dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
          }
          else if (nodeBeforeSecondNode instanceof Actor &&
             ((Actor)nodeBeforeSecondNode).isMemberToSecondaryGroup((GroupID)_dLink.getEorSecondEnd())) {
                _dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                _dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                _dLink.getEorLastLink().addComment(AccessControlLink.GLOBAL,
                        AccessControlLink.C_SECONDARY_GROUP);
          } // else
          else
              added_2 = false;
        }
        else if (_dLink.getEorSecondEnd() instanceof UserID) {
          if (nodeBeforeSecondNode instanceof Actor &&
              _dLink.getEorSecondEnd() == ((Actor)nodeBeforeSecondNode).getEorCurrentUserID()) {
                 _dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                 added_2 = true;
          }
        }  // else if
        
        // sets the rights if the second end is a resource,
        // with an OWNER or CONTAIN relation
        if (l_rights.length != 0 && rightType != null) {
          for (int i = 0; i < l_rights.length; i++) {
            added = link.addRight(rightType, l_rights[i]);
            if (added && !added_3) added_3 = true;
          } // for

                          // for testing and tracking
                          //System.out.println("in AgoRightsFactoryImpl."+
                            //"addOwnerContainRightsToLastLinkInDisplayableLink(), "+
                            //"set in the last link the rights "+rightType+
                            //" for the link: "+_dLink);
                          //System.out.println("-- add the basic rights:");
                          //for (int i = 0; i < l_rights.length; i++) {
                            //System.out.println("----- for "+i+": "+l_rights[i]);
                          //}

        } // if

        if (_dLink.getEorSecondEnd() instanceof Resource && rightType != null) { // res is set
            if (nodeBeforeSecondNode instanceof UserID) {
                nodeRights = interpreter_.getNodeRightsInstances().get(res.getName());
                inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(
                                    res,
                                    (ImmutableUserID)nodeBeforeSecondNode,
                                    null);
                if (inhLinkRights != null)
                      l_rights = inhLinkRights.getRights(AccessControlLink.GLOBAL);
            }
            else if (nodeBeforeSecondNode instanceof GroupID) {
                nodeRights = interpreter_.getNodeRightsInstances().get(res.getName());
                inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(
                                    res,
                                    null,
                                    (GroupID)nodeBeforeSecondNode);
                if (inhLinkRights != null)
                      l_rights = inhLinkRights.getRights(AccessControlLink.GLOBAL);
            } // else
            
            if (l_rights != null && l_rights.length != 0) {
                for (int j = 0; j < l_rights.length; j++) {
                    link.addRight(rightType, l_rights[j]);
                } // for j
                
                if (inhLinkRights != null) {;
                    link.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                    link.addComment(AccessControlLink.GLOBAL,
                        AccessControlLink.C_INHERITED_AG);
                    link.addComment(AccessControlLink.GLOBAL,
                        inhLinkRights.getLinkTypeComments(AccessControlLink.GLOBAL));
                } // if (inhLinkRights != null)
            } //if
            // getL_accessRightsThroughNodesTree() has been called,
            // but if an empty owner/contain dLink is set, AGO inherited rights are optional
            else if (!added_one)
                throw new InternalError("in AgoRightsFactoryImpl."+
                        "addOwnerContainRightsToLastLinkInDisplayableLink(), "+
                        "the GroupIDMember does not own or contain the Resource,"+
                        "and the AGO inherited rights cannot be null.");
        } // if

                          // for testing and tracking
                          //System.out.println("in AgoRightsFactoryImpl."+
                            //"addOwnerContainRightsToLastLinkInDisplayableLink(), "+
                            //"set the rights for the acLink: "+link);
                          //System.out.println("-- add the inherited rights:");
                          //if (l_rights != null) {
                            //  for (int i = 0; i < l_rights.length; i++) {
                              //  System.out.println("----- for "+i+": "+l_rights[i]);
                              //}
                          //} // if

        nodeBeforeSecondNode = null;
        l_nodes = null;
        link = null;
        l_rights = null;
        rightType = null;
        res = null;
        nodeRights = null;

        if (added_one || added_2 || added_3) return _dLink;
        else return null;
  } // addOwnerContainRightsToLastLinkInDisplayableLink()
  
  /**
  * Defines the effective applicable rights of an UserID or a GroupID
  * for a Resource which is its member, when the two objects are in the view.
  * Defines the effective rights between an actor and a resource
  * when their common UserID or GroupID is on the view with them.
  * Defines also the relation when a view actor is owned by an UserID
  * or a GroupID which is also in the view, even if the relevant rights of the actor are null,
  * because this relation may allow access from the actor as rights user.
  * Comments the links to view actors having currentUID/GID. <P>
  * For a sketch view (non-null _center), the own resources of an UserID or a GroupID
  * are not detected to keep short the view size.
  * Does NOT define the root rights for the resources when the UserID 'root',
  * with the order equal to 0, is in the view.
  * Does NOT define the IS_MEMBER relations.
  * Does NOT define the group rights when user rights are applicable.
  * Does NOT define the 'other' rights for Linux/Unix operating systems, since the
  * user or group may be activated but hidden, outside the view.
  * Does NOT define the secondary group rights, if the secondary groups are in the view.
  * Process only the view objects. <P>
  * Called by detectUserRights() and detectGroupRights(), only for
  * RightsFactory_Facade.detectSketchRights().
  * Calls LinkRights.withAccessThroughNodesTreeFromEP() and
  * getL_accessRightsThroughNodesTree().
  * Synchronized to _viewInBase in the calling GraphicView.
  * @param _viewInBase EPRViewInBase
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _forUser true if the user rights are to be detected, and false if
  *   the group rights are to be detected
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  *   with Owner/Contain type links associated to every pair (EP, Resource).
  * */
  public final Map detectOwnerContainRights(EPRViewInBase _viewInBase,
    Map _m_l_DisplayableLinks, boolean _forUser, BaseObject _center) {
      // variables
      Resource                res;
      Actor                   act;
      ImmutableTarget         targ;
      AccessControlLinkImpl   acLink;
      DisplayableLinkImpl     dLink;
      ImmutableGroupID        resGroup = GroupIDImpl.EMPTY_INSTANCE;
      ImmutableUserID         resUser = UserIDImpl.EMPTY_INSTANCE;
      ImmutableGroupIDMember  groupOrUser;
      StringRight[]           l_rights;
      Map                     m_l_DisplayableLinks = _m_l_DisplayableLinks;
      NodeRights              nodeRights = null;
      LinkRights              linkRights = null;
      AccessControlLinkImpl   inhLinkRights = null;

      // process link AG/resource
      for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ;
        ite.hasNext() ; ) {
        targ = (ImmutableTarget) ite.next();
        if (!(targ instanceof Resource))
            continue;
        res = (Resource) targ;
        
        if (_forUser) {
          resUser = res.getEorUserID();
          if (resUser == null || resUser.getOrder() == 0)
              continue;
          l_rights = res.getL_ownerRights();
          groupOrUser = resUser;
        } else {
          resGroup = res.getEorGroupID();
          if (resGroup == null)
              continue;
          l_rights = res.getL_groupRights();
          groupOrUser = resGroup;
        } // else
        // for a sketch view, no right is also processing
        if (_center == null && (l_rights.length == 0 ||
                (l_rights.length == 1 && l_rights[0] == null) ||
                (l_rights.length == 1 && l_rights[0].isEmpty()) ))
            continue;

                    // for testing and tracking
                    //System.out.println("in AgoRightsFactoryImpl.detectOwnerContainRights(), for "+
                      //"the resource "+res.getName()+", the user or group is "+groupOrUser+
                      //" and the rights are:");
                    //for (int j = 0; j < l_rights.length; j++) {
                      //  System.out.println("------- "+l_rights[j]);
                    //} // for j
        
        // process the pair (resource, its groupOrUser) when they are in the view
        if (!_viewInBase.containsBaseObject( groupOrUser ))
            continue;
                // for a sketch view, noop if _center == groupOrUser
                // to avoid the display of the numerous own resources of _center
        else if (_center != null && _center != res )
            continue;

        nodeRights = interpreter_.getNodeRightsInstances().get(res.getName());
        linkRights = interpreter_.getLinkRightsInstances().get(res.getName());
        
        if (!nodeRights.withAccessThroughNodesTreeFromEP(groupOrUser, res, null))
            continue;
        
                // for testing and tracking
                //System.out.println("in AgoRightsFactoryImpl.detectOwnerContainRights(), "+
                  //"step 1: detect one pair of OWNER/CONTAIN relation from "+groupOrUser.getName()+
                  //" to "+res.getName());

        // updates m_l_DisplayableLinks,
        // without taking account of the effective displayable link
        // with GroupID, from ep to res (done in detectHidenGroupIDRights())
        dLink = linkUtil_.getDisplayableLink(
              groupOrUser, res, null, m_l_DisplayableLinks);
        if (dLink == null) {
          dLink = new DisplayableLinkImpl(
                  groupOrUser, res, (BaseObject[])null);
          dLink = linkRights.addDisplayableLink(dLink, m_l_DisplayableLinks);
        } // else

        if (dLink == null) continue;

        acLink = dLink.getEorFirstLink();

        if (_forUser) {
            for (int j = 0; j < l_rights.length; j++) {
                acLink.addRight(AccessControlLink.OWNER, l_rights[j]);
            } // for j

            inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(res,
                                (ImmutableUserID)groupOrUser, null);
            if (inhLinkRights != null)
                      l_rights = inhLinkRights.getRights(AccessControlLink.C_INHERITED_AG);
            // l_rights never null since nodeRights.withAccessThroughNodesTreeFromEP() true
            for (int j = 0; j < l_rights.length; j++) {
                acLink.addRight(AccessControlLink.OWNER, l_rights[j]);
            } // for j

            if (inhLinkRights != null) {
                acLink.addComment(AccessControlLink.OWNER,
                        AccessControlLink.C_INHERITED_AG);
                acLink.addComment(AccessControlLink.GLOBAL,
                        inhLinkRights.getLinkTypeComments(AccessControlLink.GLOBAL));
            } // if
        }
        else {

                    // for testing and tracking
                    //System.out.println("in AgoRightsFactoryImpl.detectOwnerContainRights(), for "+
                      //"the resource "+res.getName()+", the group rights are :");
                    //for (int j = 0; j < l_rights.length; j++) {
                         // for testing and tracking
                        //System.out.println("------- "+l_rights[j]);
                    //} // for j


            acLink.addComment(AccessControlLink.CONTAIN, AccessControlLink.C_WEAK_LINK);
            for (int j = 0; j < l_rights.length; j++) {
                acLink.addRight(AccessControlLink.CONTAIN, l_rights[j]);
            } // for j            
            
            inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(res,
                    null, (ImmutableGroupID)groupOrUser);
            if (inhLinkRights != null)
                      l_rights = inhLinkRights.getRights(AccessControlLink.C_INHERITED_AG);
            // l_rights never null since nodeRights.withAccessThroughNodesTreeFromEP() true
            for (int j = 0; j < l_rights.length; j++) {
                acLink.addRight(AccessControlLink.CONTAIN, l_rights[j]);
            } // for j

            if (inhLinkRights != null) {
                 acLink.addComment(AccessControlLink.CONTAIN,
                    AccessControlLink.C_INHERITED_AG);
                acLink.addComment(AccessControlLink.GLOBAL,
                    inhLinkRights.getLinkTypeComments(AccessControlLink.GLOBAL));
            } // if
        } // else        

      } // for (Iterator ite

      // process link actor/currentAG
      for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ;
        ite.hasNext() ; ) {
        targ = (ImmutableTarget) ite.next();
        if (!(targ instanceof Actor)) continue;
        act = (Actor) targ;
        
        if (_forUser) {
          resUser = act.getEorCurrentUserID();
          if (resUser == null || resUser.getOrder() == 0) continue;
          //l_rights = res.getL_ownerRights();
          groupOrUser = resUser;
        } else {
          resGroup = act.getEorCurrentGroupID();
          if (resGroup == null) continue;
          //l_rights = res.getL_groupRights();
          groupOrUser = resGroup;
        } // else

                    // for testing and tracking
                    //System.out.println("in AgoRightsFactoryImpl.detectOwnerContainRights(), for "+
                      //"the resource "+res.getName()+", the user or group rights are :");
                    //for (int j = 0; j < l_rights.length; j++) {
                         // for testing and tracking
                      //  System.out.println("------- "+l_rights[j]);
                    //} // for j    
        
        // process the pair (resource, its groupOrUser) when they are in the view
        if (!_viewInBase.containsBaseObject( groupOrUser )) continue;
                // for a sketch view, if _center == groupOrUser,
                // display the current actors of _center
        else if (_center != null && _center != act && _center != groupOrUser) continue;

                // for testing and tracking
                //System.out.println("in AgoRightsFactoryImpl.detectOwnerContainRights(), "+
                  //"step 2: detect one pair of OWNER/CONTAIN relation from "+groupOrUser.getName()+
                  //" to "+res.getName());

        dLink = linkUtil_.getDisplayableLink(
              act, groupOrUser, null, m_l_DisplayableLinks);
        if (dLink == null) {
          dLink = new DisplayableLinkImpl(
                  act, groupOrUser, (BaseObject[])null);
          linkRights = interpreter_.getLinkRightsInstances().get(groupOrUser.getName());
          if (linkRights == null) {
                interpreter_.addInstancesForNode(groupOrUser);
                linkRights = interpreter_.getLinkRightsInstances().get(
                    groupOrUser.getName());
          } // if
          dLink = linkRights.addDisplayableLink(dLink, m_l_DisplayableLinks);
        } // if

        if (dLink == null) continue;

        acLink = dLink.getEorFirstLink();
        acLink.setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
      } // for (Iterator ite
      
      res = null;
      targ = null;
      act = null;
      acLink = null;
      dLink = null;
      resGroup = null;
      l_rights = null;
      nodeRights = null;
      linkRights = null;
      inhLinkRights = null;

                    // for testing and tracking
                    //System.out.println("\n In AgoRightsFactoryImpl.detectOwnerContainRights(), "+
                      //"the DisplayableLinkImpls in the result are:");
                    //for (Iterator ite = m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {
                      // for (Iterator ita = ((List)ite.next()).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite


      return m_l_DisplayableLinks;
  } // detectOwnerContainRights()
  
  /**
  * Defines the effective applicable rights of the administrative account
  * with an order at 0, and for every actor running under such an account,
  * for any resource which is in the view. 
  * Called by RightsFactory_Facade.detectEPRRights() and detectSketchRights().
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @param _viewInBase EPRViewInBase
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks, with 'root' links associated 
  * to every pair (EP,Resource).
  * */
  protected final Map detectRootRights(EPRViewInBase _viewInBase,
    Map _m_l_DisplayableLinks, BaseObject _center) {
      ImmutableSource       so;
      EligibleParty         ep;
      ACSObject             obj;
      StringRight[]         l_rights;
      DisplayableLinkImpl   dLink;
      LinkRights            linkRights = null;

                          // for testing and tracking
                          //System.out.println("In AgoRightsFactoryImpl."+
                            //"detectRootRights(), start for the view "
                            //+_viewInBase.getName());
      
      Map m_result = _m_l_DisplayableLinks;

      for (Iterator ita = _viewInBase.getEorM_Sources().values()
        .iterator() ; ita.hasNext() ; ) {
          so = (ImmutableSource) ita.next();
          if (so == null || !(so instanceof EligibleParty)) continue;
          ep = (EligibleParty)so;

          // detect all direct 'root' relations
          if (
            ((ep instanceof UserID) && ((UserID)ep).getOrder() == 0) ||
            ((ep instanceof Actor) &&
            ((Actor)ep).getEorCurrentUserID() != null &&
            ((Actor)ep).getEorCurrentUserID().getOrder() == 0) ||
            ((ep instanceof Actor) &&
            ((Actor)ep).getEorUserID() != null &&
            ((Actor)ep).getEorUserID().getOrder() == 0)
          ) {

                          // for testing and tracking
                          //System.out.println("In AgoRightsFactoryImpl."+
                            //"detectRootRights(), "+
                            //"process an ep: "+ep.getName());

            for (Iterator ite = _viewInBase.getEorM_BaseObjects().values()
              .iterator() ; ite.hasNext() ; ) {
                obj = (ACSObject) ite.next();
                
                if (!ep.getAcsName().equals( obj.getAcsName()) ) continue;
                if (ep == obj) continue;
                
                // for a sketch view
                if (_center != null && _center != obj) continue;
                
                if (obj instanceof Resource) l_rights = ((Resource)obj).getL_rootRights();
                else l_rights = new StringRight[] {
                    BaseManagerImpl.getInstance().getMetaRight("|is_superadmin|")};

                // updates m_l_DisplayableLinks if res and ep have the same ACS
                dLink = linkUtil_.getDisplayableLink(
                      ep, obj, null, m_result);
                if (dLink == null) {
                  dLink = new DisplayableLinkImpl(
                          ep, obj, (BaseObject[])null);
                  linkRights = interpreter_.getLinkRightsInstances().get(obj.getName());
                  dLink = linkRights.addDisplayableLink(dLink, m_result);
                } // dLink == null

                if (dLink == null) continue;

                dLink.getEorFirstLink().setRights(
                        AccessControlLink.OWNER, l_rights);

                          // for testing and tracking
                          //System.out.println("In AgoRightsFactoryImpl."+
                            //"detectRootRights(), "+
                            //"add a DisplayableLinkImpl: "+dLink);

            } // for
          } // if ((ep instanceof UserID)
          
      } // for
      
      ep = null;
      obj = null;
      dLink = null;
      l_rights = null;
      so = null;
      linkRights = null;
      return m_result;
  } // detectRootRights()

/**
  * Defines the effective applicable RUN_UNDER right through a secondary GroupID
  * for an Actor as a rights user. There is no right for a secondary GroupID upon an Actor.
  * Defines the rights when the two objects are in the view,
  * and define the rights between this actor and any view resource
  * belonging to a secondary GroupID.
  * Does not define the root rights when the UserID 'root', with the order equal to 0,
  * is in the view.
  * Called by detectGroupRights(). Synchronized to _viewInBase.
  * Calls ACS.getManageSecGroupsForActor().
  * @param _viewInBase EPRViewInBase
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks, with Contain links associated to every pair.
  **/
  public final Map detectSecondaryGroupRights(EPRViewInBase _viewInBase,
    Map _m_l_DisplayableLinks, BaseObject _center) {
      // variables
      ImmutableSource         so;
      Actor                   act;
      DisplayableLinkImpl     dLink;
      Map                     m_l_DisplayableLinks = _m_l_DisplayableLinks;
      Collection              l_ep = _viewInBase.getEorM_Sources().values();
      ImmutableGroupID[]      l_secGroups;
      LinkRights              linkRights = null;
      
      // detects the direct relations for an actor through the secondary groups
      // which are in the view
      for (Iterator ita = l_ep.iterator() ; ita.hasNext() ; ) {
            so = (ImmutableSource) ita.next();
            if (so == null || !(so instanceof Actor)) continue;
            act = (Actor) so;
            if (!act.getEorACS().getManageSecGroupsForActor()) continue;           
            l_secGroups = act.getEorL_SecondaryGroup();
          
            for (int i = 0; i < l_secGroups.length; i++ ) {
                if (!_viewInBase.containsBaseObject(l_secGroups[i])) continue;
                // for a sketch view
                else if (_center != null && _center != act && _center != l_secGroups[i])
                    continue;                 
                
                // updates _m_l_DisplayableLinks
                dLink = linkUtil_.getDisplayableLink(
                      act, l_secGroups[i], null, m_l_DisplayableLinks);
                if (dLink == null) {
                  dLink = new DisplayableLinkImpl(
                          act, l_secGroups[i], (BaseObject[])null);
                  linkRights = interpreter_.getLinkRightsInstances().get(
                          l_secGroups[i].getName());
                  if (linkRights == null) {
                        interpreter_.addInstancesForNode(l_secGroups[i]);
                        linkRights = interpreter_.getLinkRightsInstances().get(
                            l_secGroups[i].getName());
                  } // if
                  dLink = linkRights.addDisplayableLink(dLink, m_l_DisplayableLinks);
                } // dLink == null

                if (dLink == null) continue;

                // set the GLOBAL link
                dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);


                    // for testing and tracking
                    //System.out.println("In AgoRightsFactoryImpl.detectSecondaryGroupRights(), ");
                
            } // for i
          
      } // for ita


      so = null;
      act = null;
      dLink = null;
      l_ep = null;
      l_secGroups = null;
      linkRights = null;

                    // for testing and tracking
                    //System.out.println("In AgoRightsFactoryImpl.detectSecondaryGroupRights(), "+
                      //"the DisplayableLinkImpls in the result are:");
                    //for (Iterator ite = m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {
                       //for (Iterator ita = ((List)ite.next()).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      return m_l_DisplayableLinks;
  } // detectSecondaryGroupRights()

  /**
   * Gets all the Account and Group links to Resource and Actors, but
   * only when all the objects are in the view. Detect the links hereinafter:<P>
   * - 2 links AG/own or contain/resource,<P>
   * - 3 links actor/run under/current A, current G or secondary group, <P>
   * - link groupidmember/member/groupID. <P>
  * Does not apply the rules 'AGO right: A rights overlay G rights' since the group/resource
  * direct link has to be displayed if the group is in the view. But this
  * group/resource direct link is weak, and it is put in the link comment.
  * Studying of this method is a good start to understand the AGO rights processing.<P>
  * Called by RightsFactory_Facade.detectEPRRights() and detectNoThanRights().
  * Calls selectDirectOwnerContainGlobalForActorAsEP(),
  * NodeRightsImpl.selectDirectOwnerContainGlobalForActorAsEP(),
  * withAccessThroughNodesTreeFromEP() and getL_accessRightsThroughNodesTree().
  * @param _viewInBase EPRViewInBase
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @return Map of DisplayableLinks
  * @exception InterruptedException
  * */
  protected final Map detectAGRights(EPRViewInBase _viewInBase,
          Map _m_l_DisplayableLinks) throws InterruptedException {


                    // for testing and tracking
                    //System.out.println("start of AgoRightsFactoryImpl."+
                      //"detectAGRights(), for the view "+_viewInBase.getName());//+", the targets are:");
                     //"the DisplayableLinkImpls at the start are:");
                    //ImmutableName key;
                    //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
      
      // variables
      ImmutableSource         so;
      ImmutableSource         so2;
      Resource                res;
      ImmutableTarget         targ;
      AccessControlLinkImpl   acLink;
      DisplayableLinkImpl     dLink;
      ImmutableUserID         resUser = null;
      ImmutableEligibleParty  groupOrUser = null;
      ImmutableEligibleParty  groupOrUser2 = null;
      StringRight[]           l_rights = null;
      EligibleParty           ep;
      ImmutableGroupID        grp;
      Collection              l_ep = _viewInBase.getEorM_Sources().values();
      Set                     l_groupIdMembers;
      NodeRights              nodeRights = null;
      LinkRights              linkRights = null;
      AccessControlLinkImpl   inhLinkRights = null;

                            // for testing and tracking
                          //for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ; ite.hasNext() ; ) {
                            //targ = (ImmutableTarget) ite.next();

                                        // for testing and tracking
                              //          System.out.println("--- target "+targ.getName());
                          //} // for ite


      // process link AG/resource
      for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ; ite.hasNext() ; ) {
        targ = (ImmutableTarget) ite.next();
        if (!(targ instanceof Resource)) continue;
        else res = (Resource) targ;


                    // for testing and tracking
                    //System.out.println("--- target "+res.getName());


        resUser = res.getEorUserID();
         // no root account, but process the default value '-1'
        if (resUser != null && resUser.getOrder() != 0) {
            // process the pair (groupOrUser, resource) when they are in the view
            if (_viewInBase.containsBaseObject( resUser )) {
              l_rights = res.getL_ownerRights();
              groupOrUser = resUser;
            } 
        } // if

        nodeRights = interpreter_.getNodeRightsInstances().get(res.getName());
        linkRights = interpreter_.getLinkRightsInstances().get(res.getName());

                    // for testing and tracking
                    //System.out.println("in AgoRightsFactoryImpl.detectAGRights(), for "+
                      //"the resource "+res.getName()+", the user or group rights are :");
                    //for (int j = 0; j < l_rights.length; j++) {
                         // for testing and tracking
                      //  System.out.println("------- "+l_rights[j]);
                    //} // for j           

                // for testing and tracking
                //System.out.println("in AgoRightsFactoryImpl.detectAGRights(), "+
                  //"detect one pair of OWNER/CONTAIN relation from "+groupOrUser.getName()+
                  //" to "+res.getName());

        // updates m_l_DisplayableLinks for the res account
        if (groupOrUser != null &&
                nodeRights.withAccessThroughNodesTreeFromEP(
                    groupOrUser, res, null )) {

            dLink = linkUtil_.getDisplayableLink(
                  groupOrUser, res, null, _m_l_DisplayableLinks);
            if (dLink == null) {
              dLink = new DisplayableLinkImpl(
                      groupOrUser, res, (BaseObject[])null);
            } // dLink == null

                            // for testing and tracking
                            //System.out.println("in AgoRightsFactoryImpl.detectAGRights(), "+
                              //"dLink "+dLink+" with  nodeRights "+nodeRights.getClass());

            acLink = dLink.getEorFirstLink();

            // add the direct rights
            for (int j = 0; j < l_rights.length; j++) {
                acLink.addRight(AccessControlLink.OWNER, l_rights[j]);
            } // for j

            // add the right from the parent tree
            l_rights = null;
            inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(res,
                                    groupOrUser, null);
            if (inhLinkRights != null)
                  l_rights = inhLinkRights.getRights(AccessControlLink.GLOBAL);

            // l_rights may be null with nodeRights.withAccessThroughNodesTreeFromEP() true
            if (l_rights != null) {
                for (int j = 0; j < l_rights.length; j++) {
                    acLink.addRight(AccessControlLink.OWNER, l_rights[j]);
                } // for j

                acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                acLink.addComment(AccessControlLink.GLOBAL,
                    AccessControlLink.C_INHERITED_AG);
                // acLink.setLinkTypeToTrue(AccessControlLink.OWNER);
                acLink.addComment(AccessControlLink.OWNER,
                    inhLinkRights.getLinkTypeComments(AccessControlLink.GLOBAL));
            } // if

                            // for testing and tracking
                            //System.out.println("in AgoRightsFactoryImpl.detectAGRights(), "+
                              //"add or update acLink with inhLinkRights: "+acLink);


            dLink = linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);

        } // if (groupOrUser != null

        if (res.getEorGroupID() != null && 
                _viewInBase.containsBaseObject( res.getEorGroupID() )) {
          groupOrUser2 = res.getEorGroupID();
          l_rights = res.getL_groupRights();

                    // for testing and tracking
                    //System.out.println("in AgoRightsFactoryImpl.detectAGRights(), for "+
                      //"the resource "+res.getName()+", for the group "+groupOrUser2+
                      //" the rights are:");
                    //for (int j = 0; j < l_rights.length; j++) {
                      //  System.out.println("------- "+l_rights[j]);
                    //} // for j

        } // if
        
        // updates m_l_DisplayableLinks
        if (groupOrUser2 != null &&
                nodeRights.withAccessThroughNodesTreeFromEP(
                    groupOrUser2, res, null)) {

            dLink = linkUtil_.getDisplayableLink(
                  groupOrUser2, res, null, _m_l_DisplayableLinks);
            if (dLink == null) {
              dLink = new DisplayableLinkImpl(
                      groupOrUser2, res, (BaseObject[])null);
              dLink = linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);
            } // dLink == null

                    // for testing and tracking
                    //System.out.println("----- the DisplayableLink: "+dLink);

            if (dLink != null) {
                acLink = dLink.getEorFirstLink();
                acLink.addComment(AccessControlLink.CONTAIN, AccessControlLink.C_WEAK_LINK );

                for (int j = 0; j < l_rights.length; j++) {
                    acLink.addRight(AccessControlLink.CONTAIN, l_rights[j]);
                } // for j

                        // for testing and tracking
                        //System.out.println("------ the first AccessControlLink: "+acLink);

                l_rights = null;
                inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(res,
                                            null, (ImmutableGroupID)groupOrUser2);
                if (inhLinkRights != null)
                      l_rights = inhLinkRights.getRights(AccessControlLink.GLOBAL);

                // l_rights may be null with linkRights.withAccessThroughNodesTreeFromEP() true
                if (l_rights != null) {
                    for (int j = 0; j < l_rights.length; j++) {
                        acLink.addRight(AccessControlLink.CONTAIN, l_rights[j]);
                    } // for j

                    acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                    acLink.addComment(AccessControlLink.GLOBAL,
                        AccessControlLink.C_INHERITED_AG);
                    // acLink.setLinkTypeToTrue(AccessControlLink.CONTAIN);
                    acLink.addComment(AccessControlLink.CONTAIN,
                        inhLinkRights.getLinkTypeComments(AccessControlLink.GLOBAL));
                } // if
            } // if (dLink != null)

        } // if (groupOrUser2 != null
        
        //forUser = false;
        groupOrUser = null;
        groupOrUser2 = null;

        if (!(res instanceof Actor)) continue;

                    // for testing and tracking
                    //if (nodeRights == null) {
                      //  System.out.println("in AgoRightsFactoryImpl.detectAGRights(), for "+
                        //    "the view "+_viewInBase.getName()+" and its resource "+res.getName()+
                          //  ", there is no NodeRights.");
                        //System.out.println(" The current keys are:");
                        //for (Object obj : interpreter_.getNodeRightsInstances().keySet()) {
                          //  System.out.println("--- "+obj);
                        //}
                    //} // if


        // process the links actor/currentAG and actor/secondary group
        l_groupIdMembers = nodeRights.selectDirectOwnerContainGlobalForActorAsEP(
                                _viewInBase, (Actor) res);
        if (l_groupIdMembers != null) {
            for (Object obj : l_groupIdMembers ) {

              acLink = new AccessControlLinkImpl();
              acLink.setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

              if (obj instanceof GroupID &&
                    ((Actor)res).isMemberToSecondaryGroup((GroupID)obj)) {
                  acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                  acLink.addComment(AccessControlLink.GLOBAL,
                          AccessControlLink.C_SECONDARY_GROUP);
              } // if

              linkRights.addDisplayableLink(
                  new DisplayableLinkImpl(res, (GroupIDMember)obj, acLink),
                  _m_l_DisplayableLinks);
            } // for obj
        } // if (l_groupIdMembers
      
      } // for (Iterator ite
      
                     // for testing and tracking
                    //System.out.println("In AgoRightsFactory.detectAGRights(), "+
                      //"the DisplayableLinkImpls after the RUN_UNDER detection are:");
                    //ImmutableName key;
                    //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--for key: "+key);
                       //for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
      
      // process link (GroupIDMember, GroupID) in the view
      for (Iterator ita = l_ep.iterator() ; ita.hasNext() ; ) {
          so = (ImmutableSource) ita.next();
          if (so == null || !(so instanceof EligibleParty)) continue;
          ep = (EligibleParty)so;

          if (ep instanceof GroupIDMember) {
              for (Iterator ite = l_ep.iterator() ; ite.hasNext() ; ) {
                  so2 = (ImmutableSource) ite.next();
                  if (so2 == ep || !(so2 instanceof ImmutableGroupID)) continue;
                  grp = (ImmutableGroupID)so2;
                  if (!(grp.containsAsMember((GroupIDMember)ep))) continue;
                    // for a sketch view
                  else {
                        // updates m_l_DisplayableLinks
                        dLink = linkUtil_.getDisplayableLink(
                              ep, grp, null, _m_l_DisplayableLinks);
                        if (dLink == null) {
                          dLink = new DisplayableLinkImpl(
                                  ep, grp, (BaseObject[])null);
                          linkRights = interpreter_.getLinkRightsInstances().get(grp.getName());
                          dLink = linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);
                        } // dLink == null

                        if (dLink != null)
                            dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_MEMBER);

                                        // for testing and tracking
                                        //System.out.println("In AgoRightsFactoryImpl.detectAGRights(), "+
                                                //"the link is set to true for the IS_MEMBER type: "+dLink);

                  } // else
              } // for ite
          } // if

          if (Thread.interrupted()) {
            throw new InterruptedException();
          }
      } // for    ita   
      
      res = null;
      targ = null;
      resUser = null;
      acLink = null;
      dLink = null;
      l_rights = null;
      ep = null;
      grp = null;
      l_ep = null;
      so = null;
      so2 = null;
      l_groupIdMembers = null;
      nodeRights = null;
      linkRights = null;
      inhLinkRights = null;

      return _m_l_DisplayableLinks;
  } // detectAGRights()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
    interpreter_ = null;
    linkUtil_ = null;
  } // finalizeForProcess()
  
} // AgoRightsFactoryImpl
