//Title:        Access Road
//Version:      0.7.2
//Copyright:    2002, 2004, 2007-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import java.util.Arrays;
import java.util.Collection;

import ARoad0.ErrW.*;
import ARoad0.Gui2.GraphicView;
import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.CNot.AccessControlLink;
import ARoad0.CNot.AccessControlLinkImpl;

import ARoad0.gBaseInterface.*;
import ARoad0.gWorkInterface.*;

import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.VirtualFolderImpl;
import ARoad0.gBase.UtilityImpl;

/**
 * This class is responsible for all the core algorithms about the long access paths,
 * mixing the searches among all types of rights. This class contains the most complex
 * algorithms of Access Road, in the method detectHiddenCompoundEpRights() responsible to
 * find all the access paths having more than 3 nodes. It provides
 * also some small methods for handling the ACS links, the Bridges, the Alias
 * and the virtual members.
 * This class is stateless, and RightsMediatorImpl at its creation
 * setups an instance for its ViewInBase.
 * The calls to ACS are only in the method detectAcsLinks().
 * @version 0.7.2
 */
public final class CompoundRightsFactoryImpl implements CoreAlgorithm {

  private Set<ImmutableSource>          l_hiddenSourceToTest_ = new HashSet(INITIAL_CAPACITY*10);
  private Set<DisplayableLinkImpl>      l_hiddenSourceLinks_ = new HashSet(INITIAL_CAPACITY*40);
  private Set<ImmutableGroupIDMember>   l_ownersForOwnAndContainToTest_ = new HashSet(INITIAL_CAPACITY*2);
  private Set<ImmutableActor>           l_actorsToTest_ = new HashSet(INITIAL_CAPACITY*2);
  private Set<ImmutableVirtualFolder>   l_virtualFoldersToTest_ = new HashSet(INITIAL_CAPACITY*20);
  private Set<ImmutableGroupIDMember>   l_ownersForXIDToTest_ = new HashSet(INITIAL_CAPACITY*2);
  private Set<DisplayableLinkImpl>      l_currentOwnersLinks_ = new HashSet(INITIAL_CAPACITY);
  private Set<DisplayableLinkImpl>      l_currentActorsLinks_ = new HashSet(INITIAL_CAPACITY);
  private Set<DisplayableLinkImpl>      l_currentVirtualFoldersLinks_ = new HashSet(INITIAL_CAPACITY);
  private Set<DisplayableLinkImpl>      l_hiddenOwnersLinks_ = new HashSet(INITIAL_CAPACITY*20);
  private Set<DisplayableLinkImpl>      l_hiddenActorsLinks_ = new HashSet(INITIAL_CAPACITY*20);
  private Set<DisplayableLinkImpl>      l_hiddenVirtualFoldersLinks_ = new HashSet(INITIAL_CAPACITY*20);
  private Set<DisplayableLinkImpl>      l_allHiddenOwnersLinks_ = new HashSet(INITIAL_CAPACITY*20);
  private Set<DisplayableLinkImpl>      l_allHiddenActorsLinks_ = new HashSet(INITIAL_CAPACITY*20);
  private Set<DisplayableLinkImpl>      l_allHiddenVirtualFoldersLinks_ = new HashSet(INITIAL_CAPACITY*20);
  private Set<DisplayableLinkImpl>      l_endOfLinks_ = new HashSet(INITIAL_CAPACITY);
  private Set<DisplayableLinkImpl>      l_endOfLinks2_ = new HashSet(INITIAL_CAPACITY);
  //private Set<DisplayableLinkImpl>           l_currentSearchingLinks_ = new HashSet(INITIAL_CAPACITY);

  private AlgorithmInterpreter          interpreter_ = null;
  private DisplayableLinkUtilities      linkUtil_ = null;
  private AclRightsFactoryImpl          acl_ = null;
  private AgoRightsFactoryImpl          ago_ = null;
  private PrivilegeRightsFactoryImpl    pri_ = null;
  private UtilityImpl                   utility_ = null;

  /**
  * only one protected constructor
  * @param _interpreter algorithm interpreter
  */
  protected CompoundRightsFactoryImpl(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
      acl_ = (AclRightsFactoryImpl)interpreter_.getCoreAlgorithm("Acl");
      ago_ = (AgoRightsFactoryImpl)interpreter_.getCoreAlgorithm("Ago");
      pri_ = (PrivilegeRightsFactoryImpl)interpreter_.getCoreAlgorithm("Pri");
      utility_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
  }

  /**
  * This is the most complex Access Road method, which defines 
  * the effective applicable rights through all the access paths containing
  * more than 3 nodes. This method contains more than 1000 logical lines of code. <P>
  * All the work may be viewed as building up new CNot.DisplayableLinkImpls from the list
  * of the current ones, as the result of the analysis of rights and structure.
  * This is why the main parameter is the map of all the current DisplayableLinkImpls,
  * named _upd_m_l_DisplayableLinks.
  * This map is filled with new DisplayableLinkImpls by each method like this one.
  * A good start to understand this method is to read the code of
  * the ThreeNodesRightsFactoryImpl class. <P>
  * The search deep is up to 40 iterations, and all the access paths are found.
  * To be fast and maintenable, this method is based on very structured loops
  * that could be seen as elementay blocks. 
  * There are 3 main steps in this method: <P>
  * - the initializing search defines all the simple DisplayableLinkImpls from
  * the view nodes that are sources; this implies 9 cases, including 3 hooks for
  * the AcsAddons, <P>
  * - the central search extends these simple DisplayableLinkImpls in a continuous
  * loop that ends only when there is no new path to found; this implies the test of 19 cases,
  * including 7 tests for Actors, 8 tests for GroupIDMembers, 4 tests for Virtualfolders,
  * with 3 tests on hooks for the AcsAddons on Actors, GroupIDMembers and VirtualFolders, <P>
  * - the final search defines the last AccessControlLinkImpls from the
  * current DisplayableLinkImpls to the view nodes that are targets; this implies 
  * the test of 14 cases, with 4 hooks for the AcsAddons, including 3 final tests
  * for filtering about the rules to reject some DisplayableLinkImpls. <P>
  * This method fills the resulting map with new DisplayableLinkImpls.
  * There are two cases where a link has to be removed from the map after its insertion. <P>
  * The first case is about the filtering of access paths to apply some priority rules
  * among the AGO, PRI, ACL, BDG rights. Some of these priority rules
  * may be from an AcsAddon, through the call to a method LinkRights.detectAddonPriorityInXXX().
  * These methods have the property to detect a prioritary link
  * which enforces to remove one or several longer links. The inverse
  * is not true, since a long link cannot enforce, by its existence,
  * the removing of a shorter link. Caution: due to the sequence of calls to the ACS addons,
  * the longer link may be added in the resulting map before the shorter link
  * which produces the removing of this longer link. <P>
  * The second case is for applying a rule on the virtual folders.
  * This rule is 'virtual folder: an indirect access to a Resource that is not
  * a virtual member, is always through a virtual member'. To detect this second case,
  * the method DisplayableLinkUtilitites.getContainVirtualFolderWithoutMember()
  * is called. <P>
  * Called by RightsFactory_Facade.detectRights().
  * Calls selectDirectHiddenMembersForVirtualFolder(),
  * AclRightsFactoryImpl.selectDirectHiddenExternalAclBridgeLinksToActorForEP(),
  * addAclBridgeRightsInDisplayableLink(),
  * AgoRightsFactoryImpl.selectHiddenDirectGroupIDForGroupIDMember(),
  * selectHiddenDirectOwnerContainForActor(),
  * selectHiddenDirectOwnerContainForTarget(),
  * addOwnerContainRightsToDisplayableLink(),
  * PrivilegeRightsFactoryImpl.selectDirectHiddenPrivilegeLinksForEP(),
  * addPrivilegeInAccessControlLink(),
  * and FactoryUtilities.withExecuteRight(),
  * FactoryUtilities.detectPriorityInAllLinksAllRights(),
  * DisplayableLinkImpl.getAllEndsInDisplayableLinks(), mergoTwoLinksCollection(),
  * selectDisplayableLinksForSecondEnd().
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase
  * @param _upd_m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @exception InterruptedException, ProcessError if there is a loop in the rights search
  * @return updated _upd_m_l_DisplayableLinks, with hidden coumpond rights links associated to every pair.
  * */
  public final Map detectHiddenCompoundEpRights(GraphicView _gview,
    EPRViewInBase _viewInBase, Map _upd_m_l_DisplayableLinks) 
    throws ProcessError, InterruptedException {

                          //System.out.println(
                            //"Start of CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()");

      l_hiddenSourceToTest_.clear();
      l_hiddenSourceLinks_.clear();
      l_ownersForOwnAndContainToTest_.clear();
      l_actorsToTest_.clear();
      //l_testedActors_.clear();
      //l_actorsForXIDToTest_.clear();
      l_virtualFoldersToTest_.clear();
      l_ownersForXIDToTest_.clear();
      //l_currentOwnersLinks_.clear();
      //l_currentActorsLinks_.clear();
      //l_currentOtherContextActorLinks_.clear();
                  //l_whileLoopOtherContextActorLinks_.clear();
                  //l_otherContextActorLinks_.clear();
            //l_currentVirtualFoldersLinks_.clear();
      l_hiddenOwnersLinks_.clear();
      l_hiddenActorsLinks_.clear();
      l_hiddenVirtualFoldersLinks_.clear();
      l_allHiddenOwnersLinks_.clear();
      l_allHiddenActorsLinks_.clear();
      l_allHiddenVirtualFoldersLinks_.clear();
      //l_currentSearchingLinks_.clear();

      // variables
      HashSet                  l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();
      Set<ImmutableGroupIDMember>  l_hiddenDirectEP;
      Set<DisplayableLinkImpl> l_startOfLinks;
      Set<DisplayableLinkImpl> l_temporaryLinks = null;
      Set<DisplayableLinkImpl> l_temporaryLinks2 = null;
      Set<DisplayableLinkImpl> l_temporaryLinks3 = null;
      List<DisplayableLinkImpl> l_temporaryLinks4 = null;
      List<ACSObject>           l_temporaryLinks5 = new ArrayList(20);
      Set<ImmutableGroupID>    l_groups;
      ImmutableEligibleParty   ep;
      ImmutableResource        viewRes;
      ImmutableEligibleParty   viewEp;
      ImmutableGroupIDMember   member;
      ImmutableActor           act;
      ImmutableTarget          viewTarg;
      DisplayableLinkImpl      dLink;
      DisplayableLinkImpl      dLink_2;
      int                      iterationsCount = 0;
      AccessControlLinkImpl    acLink;
      AccessControlLinkImpl    lastLink;
      Resource[]               l_resources;
      boolean                  ok_1 = true; // for the first while loop iteration
      boolean                  ok_2 = false;
      boolean                  ok_3 = false;
      ImmutableSource          so;
      ImmutableVirtualFolder   vf;
      Name                     pairKey = null;
      boolean                  priority = false;
      String                   rightType = null;
      Set<ImmutableGroupIDMember> l_members = null;
      NodeRights               nodeRights = null;
      NodeRights               nodeRights2 = null;
      LinkRights               linkRights = null;
      LinkRights               linkRights2 = null;
      //BaseObject[]             l_nodes = null;
      int                      nextIndex = -1;

      // INITIALIZING SEARCH, for GROUPIDMEMBERS and ACTORS, including 8 tests.
      // iterates on each source in the view
      for (Iterator iti = _viewInBase.getEorM_Sources().values().iterator() ; iti.hasNext() ; ) {

          so = (ImmutableSource) iti.next();
          if (so == null) continue;

          l_currentOwnersLinks_.clear();
          l_currentActorsLinks_.clear();
          l_currentVirtualFoldersLinks_.clear();

          nodeRights = interpreter_.getNodeRightsInstances().get(so.getName());

          l_temporaryLinks = selectDirectHiddenReferenceForAlias(_gview, _viewInBase, so);
          if (l_temporaryLinks != null) {
              for (Object obj : l_temporaryLinks) {
                     dLink = (DisplayableLinkImpl)obj;
                     if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember)
                         l_currentOwnersLinks_.add(dLink);
                     else if (dLink.getEorSecondEnd() instanceof ImmutableActor)
                         l_currentActorsLinks_.add(dLink);
                     else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder)
                         l_currentVirtualFoldersLinks_.add(dLink);
                     // noop if the reference is a resource for instance
              } // for
          } // if (l_temporaryLinks != null)

          // so is a VirtualFolder
          if (so instanceof VirtualFolder) {

              // only EP VirtualFolders are processed
              if ( !((VirtualFolder)so).getMainVFType().equals(VirtualFolderImpl.EP_VFOLDER))
                  continue;

              l_temporaryLinks = selectDirectHiddenMembersForVirtualFolder(
                    _gview, _viewInBase, (VirtualFolder)so);
              if (l_temporaryLinks != null) {
                  // virtual members may be actors, since it is an EP_FOLDER VirtualFolder
                  for (Object obj : l_temporaryLinks) {
                     dLink = (DisplayableLinkImpl)obj;
                     if (dLink.getEorSecondEnd() instanceof ImmutableActor)
                         l_currentActorsLinks_.add(dLink);
                  } // for
              } // if

                // if node.getSelectAddonNewHiddenNodeForActor()
                // detects the VirtualFolders which have complementary links
                // from their AcsAddon ACS, updates l_currentActorsLinks_,
                // and puts the new ends in the view l_ObjectToRegister
              if (so instanceof VirtualFolder &&
                      nodeRights.getSelectAddonNewHiddenNodeForVirtualFolder()) {

                    l_temporaryLinks5.clear();
                    l_temporaryLinks4 = nodeRights.selectAddonNewHiddenNodeForVirtualFolder(
                                _viewInBase, (VirtualFolder)so, null, _upd_m_l_DisplayableLinks,
                                l_temporaryLinks5);
                    l_ObjectToRegister.addAll(l_temporaryLinks5);

                    if (l_temporaryLinks4 != null) {
                        for (Iterator itu = l_temporaryLinks4.iterator();
                                itu.hasNext(); ) {

                              dLink = (DisplayableLinkImpl)itu.next();

                              if (dLink.getEorSecondEnd() instanceof Actor)
                                  l_currentActorsLinks_.add( dLink );
                        } // for itu
                    } // if (l_temporaryLinks4 != null)

              } // if node.getSelectAddonNewHiddenNodeForVirtualFolder()
          } // if VirtualFolder
          else {
              ep = (EligibleParty)so;

              if (ep instanceof GroupIDMember) {
                  // adds all the is_member relations on groups
                l_hiddenDirectEP = nodeRights.
                  selectHiddenDirectGroupIDForGroupIDMember(
                    _viewInBase, (GroupIDMember)ep);

                    // adds all the owner/contain relations on actors
                    // that are external-acl, bridge or privilege sources, or alias
                l_temporaryLinks = nodeRights.
                        selectForGroupIDMemberItsHiddenActorsWithNextLinks(
                        _viewInBase, (GroupIDMember)ep);
                if (l_temporaryLinks != null)
                    l_currentActorsLinks_.addAll( l_temporaryLinks );
                
              } else // ep may not be a right proxy
                  // adds all the run_under relations on groupidmembers
                l_hiddenDirectEP = nodeRights.
                    selectHiddenDirectOwnerContainGlobalForActorAsEP(
                        _viewInBase, (Actor)ep);

              if (l_hiddenDirectEP != null) {
                  for (Iterator itu = l_hiddenDirectEP.iterator() ; itu.hasNext() ; ) {
                    member = (ImmutableGroupIDMember)itu.next();
                    acLink = new AccessControlLinkImpl();

                    if (ep instanceof GroupIDMember) {
                        // 'member' is a GroupID
                        acLink.setLinkTypeToTrue(AccessControlLink.IS_MEMBER);
                        if (ep instanceof GroupID)
                                // the first acLink is marked
                            acLink.addComment(AccessControlLink.CONTAIN,
                                AccessControlLink.C_WEAK_LINK);
                    } else { // ep is Actor
                        if (member instanceof GroupID &&
                                ((Actor)ep).isMemberToSecondaryGroup((GroupID)member)) {
                            acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                            acLink.addComment(
                                AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                        }
                        // 'member' is an UserID or a GroupID
                        acLink.setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                    } // else
                    
                    l_currentOwnersLinks_.add(
                      new DisplayableLinkImpl(ep, member, acLink));
                  } // for itu
              } // if (l_hiddenDirectEP != null

                        //select the actors that are acl or bridge targets for ep as source
              l_temporaryLinks = acl_.
                selectDirectHiddenAclBridgeLinksToActorForEP(_viewInBase, ep, true);
              if (l_temporaryLinks != null)
                  l_currentActorsLinks_.addAll( l_temporaryLinks  );

                        //select privilege links for Actors and VirtualFolders
              l_temporaryLinks = pri_.
                selectDirectHiddenPrivilegeLinksForEP(_gview, _viewInBase, ep, true);
              if (l_temporaryLinks != null) {
                  for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                      dLink = (DisplayableLinkImpl)itu.next();

                      if (dLink.getEorSecondEnd() instanceof Actor)
                          l_currentActorsLinks_.add( dLink );
                      else if (dLink.getEorSecondEnd() instanceof VirtualFolder)
                          l_currentVirtualFoldersLinks_.add( dLink );
                  } // if
              } // if (l_temporaryLinks

                // detects the GroupIDMembers which have complementary links
                // from their AcsAddon ACS, updates l_currentActorsLinks_,
                // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
                // and puts the new ends in l_ObjectToRegister
              if (ep instanceof GroupIDMember &&
                      nodeRights.getSelectAddonNewHiddenNodeForGroupIDMember()) {

                    l_temporaryLinks5.clear();
                    l_temporaryLinks4 = nodeRights.selectAddonNewHiddenNodeForGroupIDMember(
                            _viewInBase, (GroupIDMember)ep, null, _upd_m_l_DisplayableLinks,
                            l_temporaryLinks5);
                    l_ObjectToRegister.addAll(l_temporaryLinks5);

                    if (l_temporaryLinks4 != null) {
                        for (Iterator itu = l_temporaryLinks4.iterator();
                                itu.hasNext(); ) {

                              dLink = (DisplayableLinkImpl)itu.next();

                              if (dLink.getEorSecondEnd() instanceof Actor)
                                  l_currentActorsLinks_.add( dLink );
                              else if (dLink.getEorSecondEnd() instanceof GroupIDMember)
                                  l_currentOwnersLinks_.add( dLink );
                              else if (dLink.getEorSecondEnd() instanceof VirtualFolder)
                                  l_currentVirtualFoldersLinks_.add( dLink );
                        } // for itu
                    } // if (l_temporaryLinks4 != null)
              } // if (ep instanceof ImmutableGroupIDMember

                // detects the Actors which have complementary links
                // from their AcsAddon ACS, updates l_currentActorsLinks_,
                // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
                // and puts the new ends in l_ObjectToRegister
              if (ep instanceof Actor && 
                      nodeRights.getSelectAddonNewHiddenNodeForActor()) {

                    l_temporaryLinks5.clear();
                    l_temporaryLinks4 = nodeRights.selectAddonNewHiddenNodeForActor(
                                _viewInBase, (Actor)ep, null, _upd_m_l_DisplayableLinks,
                                l_temporaryLinks5);
                    l_ObjectToRegister.addAll(l_temporaryLinks5);

                    if (l_temporaryLinks4 != null) {
                        for (Iterator itu = l_temporaryLinks4.iterator();
                                itu.hasNext(); ) {

                              dLink = (DisplayableLinkImpl)itu.next();

                              if (dLink.getEorSecondEnd() instanceof Actor)
                                  l_currentActorsLinks_.add( dLink );
                              else if (dLink.getEorSecondEnd() instanceof GroupIDMember)
                                  l_currentOwnersLinks_.add( dLink );
                              else if (dLink.getEorSecondEnd() instanceof VirtualFolder)
                                  l_currentVirtualFoldersLinks_.add( dLink );
                        } // for itu
                    } // if (l_temporaryLinks4 != null)

              } // if (ep instanceof Actor

          } // so instanceof EligibeParty
          
          // SET THE USABLE LINKS
          for (Iterator itu = l_currentOwnersLinks_.iterator(); itu.hasNext(); ) {
                ago_.addOwnerContainRightsToDisplayableLink(
                        (DisplayableLinkImpl)itu.next());
          }

          l_hiddenOwnersLinks_.addAll(l_currentOwnersLinks_);
          l_hiddenActorsLinks_.addAll(l_currentActorsLinks_);
          l_hiddenVirtualFoldersLinks_.addAll(l_currentVirtualFoldersLinks_);
           
          // SET THE RIGHTS-RELATED OBJECTS
          // sets the graphic view with the intermediates, to listen for their impact on rights
          for (Iterator itu = l_currentOwnersLinks_.iterator(); itu.hasNext(); ) {
              dLink = (DisplayableLinkImpl)itu.next();
              l_ObjectToRegister.add( dLink.getEorSecondEnd() );
          }
          for (Iterator itu = l_currentActorsLinks_.iterator(); itu.hasNext(); ) {
              dLink = (DisplayableLinkImpl)itu.next();
              l_ObjectToRegister.add( dLink.getEorSecondEnd() );
          }
          for (Iterator itu = l_currentVirtualFoldersLinks_.iterator(); itu.hasNext(); ) {
              dLink = (DisplayableLinkImpl)itu.next();
              l_ObjectToRegister.add( dLink.getEorSecondEnd() );
          }

          // SET THE OBJECTS TO TEST
          // selects the set of Not-In-View, not direct, applicable UserIDs
          // and GroupIDs which contain the ep
          l_ownersForOwnAndContainToTest_.addAll(linkUtil_.
            getAllEndsInDisplayableLinks(l_currentOwnersLinks_, false));

          // selects the set of Not-In-View Actors with 'execute' right or bridge
          l_actorsToTest_.addAll(linkUtil_.
            getAllEndsInDisplayableLinks(l_currentActorsLinks_, false));

          // selects the set of Not-In-View VirtualFolders
          l_virtualFoldersToTest_.addAll(linkUtil_.
            getAllEndsInDisplayableLinks(l_currentVirtualFoldersLinks_, false));

      } // iterator iti

      // used in the final loops
      l_hiddenSourceToTest_.addAll((Set)((HashSet)l_ownersForOwnAndContainToTest_).clone());
      l_hiddenSourceToTest_.addAll((Set)((HashSet)l_actorsToTest_).clone());
      l_hiddenSourceToTest_.addAll((Set)((HashSet)l_virtualFoldersToTest_).clone());

      if (Thread.interrupted()) {
        throw new InterruptedException();
      }
                        // for testing and tracking
                        //System.out.println("\n in CompoundRightsFactoryImpl."+
                          //"detectHiddenCompoundEpRights(), for the view "+_viewInBase.getName()+
                          //", THE FIRST STATE IS:");
                        //System.out.println("----- l_hiddenSourceToTest_: ");
                        //for (Iterator itetest = l_hiddenSourceToTest_.iterator(); itetest.hasNext();) {
                          //System.out.println("--------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("----- l_hiddenOwnersLinks_: ");
                         //for (Iterator itetest = l_hiddenOwnersLinks_.iterator() ; itetest.hasNext() ; ) {
                          //System.out.println("------------- :"+
                            //((DisplayableLinkImpl)itetest.next()).toString() );
                         //}
                        //System.out.println("----- l_hiddenActorsLinks_ (are sources of ACL/bridge/privilege/alias links): ");
                         //for (Iterator itetest = l_hiddenActorsLinks_.iterator() ; itetest.hasNext() ; ) {
                          //System.out.println("------------- :"+
                            //((DisplayableLinkImpl)itetest.next()).toString() );
                         //}
                        //System.out.println("----- l_hiddenVirtualFoldersLinks_ (contains privilege links): ");
                         //for (Iterator itetest = l_hiddenVirtualFoldersLinks_.iterator() ; itetest.hasNext() ; ) {
                          //System.out.println("------------- :"+
                            //((DisplayableLinkImpl)itetest.next()).toString() );
                         //}

      //
      // --------- START OF THE CENTRAL WHILE LOOP --------------------------------------
      //
      while ( iterationsCount < 40 && (ok_1 || ok_2 || ok_3) ) {

          iterationsCount++;
          
                          // for testing and tracking
                          //System.out.println(
                            //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),"+
                            //" start in the central loop of the iteration number "+iterationsCount);
          
          l_currentActorsLinks_.clear();
          l_currentOwnersLinks_.clear();
          l_currentVirtualFoldersLinks_.clear();
          
          // FIRST CENTRAL ITERATION, on GROUPIDMEMBERS, including 8 tests.
          // 1 - analyzes the GroupIDMembers which have complementary links
          // from their AcsAddon ACS, may update l_startOfLinks for the next tests,
          // l_currentActorsLinks_, l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
          // and puts the new ends in l_ObjectToRegister
          // 2 - detects the actors connected to the analyzed GroupIDMember
          // through an executing acl,
          // - as links that may do not transfer the AG context -
          // and updates l_currentActorsLinks_.
          // 3 - detects the direct groups this groupidmember belongs to
          // and updates l_currentOwnersLinks_.
                  // to adapt for data_exchange: detects the objects having XID actor connexions
                  // to update l_currentOwnersLinks_.
          // 4 - detects the bridge sources, external-acl or privilege sources
          // owned by an analyzed groupIDMember and executable by it,
          // and updates l_currentActorsLinks_.
          // 5 - analyzes the GroupIDMembers with privilege to Actors and VirtualFolders
          // and updates l_currentActorsLinks_ or l_currentVirtualFoldersLinks_.
          // 6 - analyzes the GroupIDMembers having actors
          // and puts them in view.l_ObjectToRegister if not here,
          // since they may become bridge, acl or privilege sources,
          // to allow RightsMediatorImpl to listen the properties of these actors.
          // 7 - analyzes the GroupIDMembers which are aliases,
          // updates l_currentActorsLinks_,
          // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
          // and puts the new ends in l_ObjectToRegister
          // 8 - analyzes the alias references with an Actor alias
          // having a proper and applicable AG context,
          // and puts each element of this context as a new end if it is not in the view,
          // updates l_currentOwnersLinks_
          for (Iterator ite = l_ownersForOwnAndContainToTest_.iterator() ; ite.hasNext() ; ) {
            ep = (ImmutableGroupIDMember)ite.next();

                        // for testing and tracking
                        //System.out.println("-- in l_ownersForOwnAndContainToTest_"+
                          //" iteration on ep: "+ep);

            l_startOfLinks = linkUtil_.
                selectDisplayableLinksForSecondEnd(ep, l_hiddenOwnersLinks_);

            if (l_startOfLinks == null) continue;

                        // for testing and tracking
                        //System.out.println("--- in l_ownersForOwnAndContainToTest_"+
                          //" iteration number "+iterationsCount+
                          //", l_startOfLinks:");
                        //for (Iterator itetest = l_startOfLinks.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}

            nodeRights = interpreter_.getNodeRightsInstances().get(ep.getName());
            if (nodeRights == null) {
                interpreter_.addInstancesForNode(ep);
                nodeRights = interpreter_.getNodeRightsInstances().get(ep.getName());
            } // if

            // ---- first test in the first central iteration:
            // detects the GroupIDMembers which have complementary links
            // from their AcsAddon ACS, updates l_currentActorsLinks_,
            // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
            // and puts the new ends in l_ObjectToRegister
            if (nodeRights.getSelectAddonNewHiddenNodeForGroupIDMember()) {

                l_temporaryLinks5.clear();
                l_temporaryLinks4 = nodeRights.selectAddonNewHiddenNodeForGroupIDMember(
                        _viewInBase, (GroupIDMember)ep, l_startOfLinks, 
                        _upd_m_l_DisplayableLinks, l_temporaryLinks5);
                // new in 0.7.2, to listen the sources able to become intermediate nodes
                l_ObjectToRegister.addAll(l_temporaryLinks5);

                              // for testing and tracking
                              //System.out.println(
                                //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),"+
                                //" first test in the first central iteration, objects to register: ");
                                //for (ACSObject ob : l_temporaryLinks5) {
                                  //  System.out.println("--- "+ob.getName());
                                //}

                if (l_temporaryLinks4 != null) {
                    for (Iterator itu = l_temporaryLinks4.iterator();
                            itu.hasNext(); ) {

                          dLink = (DisplayableLinkImpl)itu.next();
                          l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                          if (dLink.getEorSecondEnd() instanceof ImmutableActor &&
                                  !l_allHiddenActorsLinks_.contains(dLink))
                              l_currentActorsLinks_.add( dLink );
                          else if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember &&
                                  !l_allHiddenOwnersLinks_.contains(dLink))
                              l_currentOwnersLinks_.add( dLink );
                          else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder &&
                                  !l_allHiddenVirtualFoldersLinks_.contains(dLink))
                              l_currentVirtualFoldersLinks_.add( dLink );
                    } // for itu
                } // if (l_temporaryLinks4 != null)

            } // if node.getSelectAddonNewHiddenNodeForGroupIDMember()

            // ---- second test in the first central iteration
            // detects the actors connected to the analyzed groupidmember
            // through an executing acl,
            // - typically as links that may do not transfer the AG context -
            // and updates l_currentActorsLinks_.
            l_endOfLinks_ = acl_.
              selectDirectHiddenAclBridgeLinksToActorForEP(_viewInBase, ep, true);
            // new in 0.7.2, to listen the actors without executing right
            l_temporaryLinks = acl_.
              selectDirectHiddenAclBridgeLinksToActorForEP(_viewInBase, ep, false);

            if (l_temporaryLinks != null) {
                if (l_endOfLinks_ != null) {
                    l_temporaryLinks.removeAll(l_endOfLinks_);
                } // if
                for (Iterator ito = l_temporaryLinks.iterator(); ito.hasNext();) {
                    act = (Actor)((DisplayableLinkImpl)ito.next()).getEorSecondEnd();

                                    // for testing and tracking
                                      //System.out.println("CompounRightsFactoryImpl."+
                                        //      "detectHiddenCompoundEpRights(), "+
                                          //    "second test in the first central iteration, "+
                                            //  "register for listening the actor "+
                                              //act.getName());

                    l_ObjectToRegister.add(act);
                } //for ito
            } // if l_temporaryLinks

            if (l_endOfLinks_ != null) {

                        // for testing and tracking
                        //if (iterationsCount == 8) {
                          //  System.out.println("--- in CompounRightsFactoryImpl, "+
                            //  "l_ownersForOwnAndContainToTest_"+
                            //  " for owners, first iteration with selectDirectHiddenExternalAclBridgeLinksToActorForEP(),"+
                            //  " l_endOfLinks_ is not empty:");
                            //for (Iterator itetest = l_endOfLinks_.iterator(); itetest.hasNext();) {
                              //System.out.println("------------- : "+itetest.next().toString());
                            //}
                        //}

                l_temporaryLinks = linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_);
              
                for (Iterator ito = l_endOfLinks_.iterator(); ito.hasNext();) {
                    act = (Actor)((DisplayableLinkImpl)ito.next()).getEorSecondEnd();

                                    // for testing and tracking
                                      //System.out.println("CompounRightsFactoryImpl."+
                                        //      "detectHiddenCompoundEpRights(), "+
                                          //    "second test in the first central iteration, "+
                                            //  "register for listening the actor "+
                                              //act.getName());

                    l_ObjectToRegister.add(act);
                } //for ito

                // may filter the ACL last link
                for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                    dLink = (DisplayableLinkImpl)itu.next();
                    linkRights = interpreter_.getLinkRightsInstances().get(
                            dLink.getEorSecondEnd().getName());
                    if (linkRights == null) {
                        interpreter_.addInstancesForNode(dLink.getEorSecondEnd());
                        linkRights = interpreter_.getLinkRightsInstances().get(
                            dLink.getEorSecondEnd().getName());
                    } // if

                    // the other values of iterationsCount have a control
                    // in the final search
                    if (linkRights.getDetectAddonPriorityInLastLinksNonAGORights() &&
                            (iterationsCount == 3 ||
                            iterationsCount == 6 ||
                            iterationsCount == 9 ||
                            iterationsCount == 13 ||
                            iterationsCount == 18 ||
                            iterationsCount == 23 ||
                            iterationsCount == 28 ||
                            iterationsCount == 33)) {
                            priority = linkRights.detectAddonPriorityInLastLinkNonAGORights(
                                                    dLink, l_temporaryLinks);
                            if (priority) {
                                itu.remove();
                                continue;
                            }
                    } // if
                } // for obj

                if (!l_allHiddenActorsLinks_.containsAll(l_temporaryLinks))
                    l_currentActorsLinks_.addAll(l_temporaryLinks);

            } // if (l_endOfLinks_ != null) {

            // ---- third test in first central iteration:
            // detects the direct groups this groupidmember belongs to
            // and updates l_currentOwnersLinks_.
            if (l_endOfLinks_ != null)
                l_endOfLinks_.clear();
            else
                l_endOfLinks_ = new HashSet(INITIAL_CAPACITY);
            l_groups = nodeRights.selectHiddenDirectGroupIDForGroupIDMember(
                _viewInBase, (ImmutableGroupIDMember)ep);
            if (l_groups != null) {
                for (Iterator ito = l_groups.iterator(); ito.hasNext();) {
                    member = (GroupIDMember)ito.next();

                    dLink = new DisplayableLinkImpl(
                        ep,
                        member,
                        new AccessControlLinkImpl());

                    dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.IS_MEMBER);

                    l_endOfLinks_.add(dLink);
                    l_ObjectToRegister.add(member);
                } // for ito
            } // if (l_groups != null
            l_groups = null;
            
            if (!l_endOfLinks_.isEmpty()) {

                        // for testing and tracking
                        //if (iterationsCount == 8) {
                          //  System.out.println("--- in l_ownersForOwnAndContainToTest_"+
                            //  " second loop in first iteration: direct and indirect groups,"+
                            //  " l_endOfLinks_:");
                            //for (Iterator itetest = l_endOfLinks_.iterator(); itetest.hasNext();) {
                              //System.out.println("------------- : "+itetest.next().toString());
                            //}
                        //}
                
              l_temporaryLinks = linkUtil_.
                mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_);

              if (!l_allHiddenOwnersLinks_.containsAll(l_temporaryLinks))
                  l_currentOwnersLinks_.addAll(l_temporaryLinks);

                        // for testing and tracking
                        //System.out.println("--- second loop in first iteration:"+
                          //      "l_temporaryLinks:");
                        //for (Iterator itetest = l_temporaryLinks.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
              
            } // if (!l_endOfLinks_.isEmpty()) {
            
            if (ep instanceof ImmutableUserID || ep instanceof ImmutableGroupID) {

                // ---- fourth test in the first central iteration
                // detects the bridge sources, external-acl or privilege sources
                // owned by an analyzed groupIDMember and executable by it,
                // and updates l_currentActorsLinks_.
              l_endOfLinks_ = nodeRights.selectForGroupIDMemberItsHiddenActorsWithNextLinks(
                            _viewInBase, (ImmutableGroupIDMember)ep);

                        // for testing and tracking
                        //System.out.println("--- in CompounRightsFactoryImpl"+
                          //" in the first ite, start of the third loop for ep:"+ep.getName());

              if (l_endOfLinks_ != null) {
                  if (l_temporaryLinks == null)
                        l_temporaryLinks = new HashSet(INITIAL_CAPACITY / 3);
                  else l_temporaryLinks.clear();

                  l_temporaryLinks.addAll(linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_));

                  for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                      dLink = (DisplayableLinkImpl)itu.next();
                      if (dLink.noOfNodes() < 3) continue;
                      l_temporaryLinks2 = linkUtil_.
                              selectDisplayableLinksWithOneNodeAndTwoEnds(
                                  // this is the common source before ep
                                  dLink.getEorAllNodes()[dLink.noOfNodes() - 3],
                                  dLink.getEorFirstEnd(), dLink.getEorSecondEnd(),
                                  l_currentActorsLinks_);
                      l_temporaryLinks3 = (Set<DisplayableLinkImpl>)
                              ((HashSet)l_temporaryLinks2).clone();

                      linkRights = interpreter_.getLinkRightsInstances().get(
                                dLink.getEorSecondEnd().getName());
                      if (linkRights == null) {
                            interpreter_.addInstancesForNode(dLink.getEorSecondEnd());
                            linkRights = interpreter_.getLinkRightsInstances().get(
                                dLink.getEorSecondEnd().getName());
                      } // if

                        // ACS rule: 'AGO right: A rights overlay G rights'
                        // continue if userID rights are already applicable,
                        // and may update _upd_m_l_DisplayableLinks if ep is an UserID
                      priority = linkRights.detectAccountPriorityInLastLinkAGORights(
                                                dLink, l_temporaryLinks2);
                      if (priority) {
                          itu.remove();

                                    // for testing and tracking
                                    //System.out.println("--- in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                      //      "remove a dLink after linkRights.detectAccountPriorityInLastLinkAGORights()");

                          continue;
                      }
                      else if (linkRights.getDetectAddonPriorityInLastLinksAGORights()
                              //&&
                            //(iterationsCount == 3 ||
                            //iterationsCount == 6 ||
                            //iterationsCount == 9 ||
                            //iterationsCount == 13 ||
                            //iterationsCount == 18 ||
                            //iterationsCount == 23 ||
                            //iterationsCount == 28 ||
                            //iterationsCount == 33)
                            ){
                            priority = linkRights.detectAddonPriorityInLastLinkAGORights(
                                                    dLink, l_temporaryLinks2);
                            if (priority) {
                                itu.remove();

                                    // for testing and tracking
                                    //System.out.println("--- in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                      //      "remove a dLink after linkRights.detectAddonPriorityInLastLinkAGORights()");

                                continue;
                            }
                      } // else if

                      l_ObjectToRegister.add( dLink.getEorSecondEnd());

                       // a removing in l_temporaryLinks2 has to be also
                       // done in l_currentActorsLinks_
                      if (ep instanceof ImmutableUserID) {
                           dLink_2 = null;
                           for (Object obj : l_temporaryLinks3) {
                                dLink = (DisplayableLinkImpl)obj;
                                if (!l_temporaryLinks2.contains(dLink)) {
                                    dLink_2 = dLink;
                                    break;
                                }
                           } // for obj
                           if (dLink_2 != null)
                               l_currentActorsLinks_.remove(dLink_2);
                      } // if (ep instanceof ImmutableUserID)
                  } // for itu on l_temporaryLinks

                  if (!l_allHiddenActorsLinks_.containsAll(l_temporaryLinks))
                      l_currentActorsLinks_.addAll(l_temporaryLinks);
              } // if (!l_endOfLinks_.isEmpty())

                // ---- fifth test in first central iteration:
                // analyzes the GroupIDMembers having privilege to Actors and VirtualFolders
                // and updates l_currentActorsLinks_ or l_currentVirtualFoldersLinks_.
              if (l_endOfLinks_ != null)
                    l_endOfLinks_.clear();
              else
                    l_endOfLinks_ = new HashSet(INITIAL_CAPACITY);
              l_endOfLinks2_.clear();
              l_temporaryLinks = pri_.
                  selectDirectHiddenPrivilegeLinksForEP(_gview, _viewInBase, ep, true);
              if (l_temporaryLinks != null) {
                  for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                          dLink = (DisplayableLinkImpl)itu.next();
                          l_ObjectToRegister.add( dLink.getEorSecondEnd() );
                          if (dLink.getEorSecondEnd() instanceof ImmutableActor)
                              l_endOfLinks_.add( dLink );
                          else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder)
                              l_endOfLinks2_.add( dLink );
                  } // for itu
              } // if (l_temporaryLinks

              if (!l_endOfLinks_.isEmpty()) {
                    l_temporaryLinks = linkUtil_.
                        mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_);

                    // may filter the Privilege last link
                    for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                        dLink = (DisplayableLinkImpl)itu.next();

                        linkRights = interpreter_.getLinkRightsInstances().get(
                                    dLink.getEorSecondEnd().getName());
                        if (linkRights == null) {
                                interpreter_.addInstancesForNode(dLink.getEorSecondEnd());
                                linkRights = interpreter_.getLinkRightsInstances().get(
                                    dLink.getEorSecondEnd().getName());
                        } // if
                        // the other iterationCount values have a control in
                        // the final search
                        if (linkRights.getDetectAddonPriorityInLastLinksNonAGORights() &&
                            (iterationsCount == 3 ||
                            iterationsCount == 6 ||
                            iterationsCount == 9 ||
                            iterationsCount == 13 ||
                            iterationsCount == 18 ||
                            iterationsCount == 23 ||
                            iterationsCount == 28 ||
                            iterationsCount == 33)){
                                priority = linkRights.detectAddonPriorityInLastLinkNonAGORights(
                                                        dLink, l_temporaryLinks);
                                if (priority) {
                                    itu.remove();

                                    // for testing and tracking
                                    //System.out.println("--- in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                      //      "remove a dLink after linkRights.detectAddonPriorityInLastLinkNonAGORights()");


                                    continue;
                                }
                        } // if
                    } // for obj

                    if (!l_allHiddenActorsLinks_.containsAll(l_temporaryLinks))
                        l_currentActorsLinks_.addAll(l_temporaryLinks);
              }

              if (!l_endOfLinks2_.isEmpty()) {
                  l_temporaryLinks = linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks2_);

                  if (!l_allHiddenVirtualFoldersLinks_.containsAll(l_temporaryLinks))
                      l_currentVirtualFoldersLinks_.addAll(l_temporaryLinks);
              } // if

            } // if (ep instanceof UserID ||

            // ---- sixth test in first central iteration:
          // analyzes the GroupIDMembers having actors
          // and puts them in view.l_ObjectToRegister if not here,
          // since they may become bridge, acl or privilege sources,
          // to allow RightsMediatorImpl to listen the properties of these actors.
            if (ep instanceof ImmutableGroupID) {
                l_resources = ((ImmutableGroupID)ep).getEorL_GroupIDOwnResources();
                for (int i = 0; i < l_resources.length; i++) {
                    if (l_resources[i] instanceof ImmutableActor)
                        l_ObjectToRegister.add(l_resources[i]);
                }
            } // if
            else if (ep instanceof ImmutableUserID) {
                l_resources = ((ImmutableUserID)ep).getEorL_UserIDOwnResources();
                for (int i = 0; i < l_resources.length; i++) {
                    if (l_resources[i] instanceof ImmutableActor)
                        l_ObjectToRegister.add(l_resources[i]);
                }
            } // if

            // ---- seventh test in the first central iteration:
            // detects the GroupIDMembers which are aliases,
            // updates l_currentActorsLinks_,
            // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
            // and puts the new ends in l_ObjectToRegister
            l_endOfLinks_ = selectDirectHiddenReferenceForAlias(_gview, _viewInBase, ep);

            if (l_endOfLinks_ != null) {
                  if (l_temporaryLinks == null)
                        l_temporaryLinks = new HashSet(INITIAL_CAPACITY / 3);
                  else l_temporaryLinks.clear();

                  l_temporaryLinks.addAll(linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_));

                  for (Object obj : l_temporaryLinks) {
                         dLink = (DisplayableLinkImpl)obj;

                         if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember &&
                                 !l_allHiddenOwnersLinks_.contains(dLink))
                             l_currentOwnersLinks_.add(dLink);
                         else if (dLink.getEorSecondEnd() instanceof ImmutableActor &&
                                 !l_allHiddenActorsLinks_.contains(dLink))
                             l_currentActorsLinks_.add(dLink);
                         else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder &&
                                 !l_allHiddenVirtualFoldersLinks_.contains(dLink))
                             l_currentVirtualFoldersLinks_.add(dLink);
                         // noop if the reference is a resource for instance
                  } // for
            } // if (l_endOfLinks_  != null)

              // ---- eighth test in the first central iteration:
              // analyzes the alias references with an Actor alias
              // having a proper and applicable AG context,
              // and puts each element of this context as a new end if it is not in the view,
              // updates l_currentOwnersLinks_
            for (Object obj : l_startOfLinks) {
                dLink = (DisplayableLinkImpl)obj;
                if (!dLink.getEorLastLink().getLinkTypeValue(AccessControlLink.IS_ALIAS))
                    continue;
                l_members = linkUtil_.getSecondEndAGcontext(dLink);

                if (l_members == null || dLink.getProperties() == null ||
                            dLink.getProperties().get(
                                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null)
                                    continue;

                nextIndex = ((Integer)dLink.getProperties().get(
                            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                            intValue();
                // nextIndex has to be the index of the actor just before the second end
                if (nextIndex == dLink.noOfNodes() - 2) // != dLink.noOfNodes() - 2)
                    throw new InternalError("eighth test in the first central iteration");
                    //continue;

                for (Object memb : l_members) {

                    if (dLink.getEorFirstEnd() == memb ||
                            _viewInBase.containsBaseObject((BaseObject)memb))
                        continue;

                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, (GroupIDMember)memb,
                            new AccessControlLinkImpl());
                    if (dLink_2 == null) continue;

                    dLink_2.getEorConnectedLinks((GroupIDMember)dLink.getEorSecondEnd())[0].
                            setLinkTypeToTrue(AccessControlLink.IS_ALIAS);
                    if (memb instanceof GroupID) {
                        dLink_2.getEorConnectedLinks((GroupIDMember)dLink.getEorSecondEnd())[0].
                              addComment(
                                AccessControlLink.IS_ALIAS,
                                AccessControlLink.C_SET_GROUP + ": "+((GroupIDMember)memb).getNickName());
                    }
                    else if (memb instanceof UserID) {
                      dLink_2.getEorConnectedLinks((GroupIDMember)dLink.getEorSecondEnd())[0].
                              addComment(
                                AccessControlLink.IS_ALIAS,
                                AccessControlLink.C_SET_USER + ": "+((GroupIDMember)memb).getNickName());
                    } // else

                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

                    // set the GLOBAL type if the first end is an actor, and the next node
                    // is a secondary group
                    if (memb instanceof GroupID && dLink.getEorSecondEnd() instanceof Actor &&
                      ((Actor)dLink.getEorSecondEnd()).isMemberToSecondaryGroup((GroupID)memb) ) {
                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink_2.getEorLastLink().addComment(
                                  AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                    } // if

                    if (!l_allHiddenOwnersLinks_.contains(dLink_2))
                        l_currentOwnersLinks_.add(dLink_2);
                } // for memb

            } // for obj
          } // for ite, end of the first iteration
          
          // SECOND CENTRAL ITERATION, on ACTORS, including 7 tests.
          // 1 - analyzes the Actors which are in a link, after a VirtualFolder having
          // no member as next node in the link, and removes their links.
          // may removing links in l_hiddenActorsLinks_ (not in l_currentActorsLinks_)
          // 2 - analyzes the Actors which have complementary links
          // from their AcsAddon ACS, may update l_startOfLinks for the next tests,
          // l_currentActorsLinks_, l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
          // and puts the new ends in l_ObjectToRegister
          // 3 - analyzes the Actors having an executing acl on another Actor
          // or a bridge relation, as sources
          // - as links that may do not transfer the AG context -
          // and updates l_currentActorsLinks_
          // 4 - analyzes the Actors having privilege to Actors and VirtualFolders
          // and updates l_currentActorsLinks_ or l_currentVirtualFoldersLinks_
          // 5 - analyzes the Actors that have a proper AG context,
          // that are bridge targets, virtual members,
          // or that are ACL targets or privilege targets...
          // detects if their current Account/Group context replaces partially or totally
          // the AG context of the previous node, and updates l_currentOwnersLinks_
          // 6 - analyzes the Actors which are aliases,
          // updates l_currentActorsLinks_,
          // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
          // and puts the new ends in l_ObjectToRegister
          // 7 - analyzes the alias references with an Actor alias
          // having a proper and applicable AG context,
          // and puts each element of this context as a new end if it is not in the view,
          // updates l_currentOwnersLinks_
                        // removing tests
          for (Iterator ite = l_actorsToTest_.iterator() ; ite.hasNext() ; ) {
            ep = (ImmutableActor)ite.next();

            l_startOfLinks = linkUtil_.
                selectDisplayableLinksForSecondEnd(ep, l_hiddenActorsLinks_);

            if (l_startOfLinks == null) continue;

            // ---- first removing test in the second central iteration:
            // analyzes the Actors which are in a link, after a VirtualFolder having
            // no member as next node in the link, and removes their links.
            // may removing links in l_hiddenActorsLinks_ (not in l_currentActorsLinks_)
            for (Iterator iti = l_startOfLinks.iterator(); iti.hasNext(); ) {
                dLink = (DisplayableLinkImpl)iti.next();

                if (linkUtil_.getContainVirtualFolderWithoutMember(dLink)) {
                    l_hiddenActorsLinks_.remove(dLink);
                    l_allHiddenActorsLinks_.remove(dLink);
                } // if
            } // for link
          } // for ite
                        // adding tests
          for (Iterator ite = l_actorsToTest_.iterator() ; ite.hasNext() ; ) {
            ep = (ImmutableActor)ite.next();

            l_startOfLinks = linkUtil_.
                selectDisplayableLinksForSecondEnd(ep, l_hiddenActorsLinks_);

            if (l_startOfLinks == null) continue;

            nodeRights = interpreter_.getNodeRightsInstances().get(ep.getName());
            if (nodeRights == null) {
                interpreter_.addInstancesForNode(ep);
                nodeRights = interpreter_.getNodeRightsInstances().get(ep.getName());
            } // if

            // ---- second test in the second central iteration:
            // analyzes the Actors which have complementary links
            // from their AcsAddon ACS, may update l_startOfLinks for the next tests,
            // l_currentActorsLinks_, l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
            // and puts the new ends in l_ObjectToRegister
            if (nodeRights.getSelectAddonNewHiddenNodeForActor()) {

                l_temporaryLinks5.clear();
                l_temporaryLinks4 = nodeRights.selectAddonNewHiddenNodeForActor(
                            _viewInBase, (Actor)ep, l_startOfLinks, 
                            _upd_m_l_DisplayableLinks, l_temporaryLinks5);
                // new in 0.7.2
                l_ObjectToRegister.addAll(l_temporaryLinks5);

                if (l_temporaryLinks4 != null) {
                    for (Iterator itu = l_temporaryLinks4.iterator();
                            itu.hasNext(); ) {

                          dLink = (DisplayableLinkImpl)itu.next();
                          l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                          if (dLink.getEorSecondEnd() instanceof Actor &&
                                  !l_allHiddenActorsLinks_.contains(dLink))
                              l_currentActorsLinks_.add( dLink );
                          else if (dLink.getEorSecondEnd() instanceof GroupIDMember &&
                                  !l_allHiddenOwnersLinks_.contains(dLink))
                              l_currentOwnersLinks_.add( dLink );
                          else if (dLink.getEorSecondEnd() instanceof VirtualFolder &&
                                  !l_allHiddenVirtualFoldersLinks_.contains(dLink))
                              l_currentVirtualFoldersLinks_.add( dLink );
                    } // for itu
                } // if (l_temporaryLinks4 != null)

            } // if node.getSelectAddonNewHiddenNodeForActor()

          // ---- third test in the second central iteration
          // analyzes the Actors having an executing acl on another Actor
          // or a bridge relation, as sources
          // - typically as links that may do not transfer the AG context
            l_endOfLinks_ = acl_.
              selectDirectHiddenAclBridgeLinksToActorForEP(_viewInBase, ep, true);
            if (l_endOfLinks_ != null) {

                        // for testing and tracking
                        //if (iterationsCount == 8) {
                            //System.out.println("--- in CompounRightsFactoryImpl, "+
                              //"l_ownersForOwnAndContainToTest_"+
                              //" for actors, second ite, first step with selectDirectHiddenExternalAclBridgeLinksToActorForEP(),"+
                              //" l_endOfLinks_ is not empty to detect an executing acl on another Actor"+
                              //" or a bridge relation, as sources:");
                            //for (Iterator itetest = l_endOfLinks_.iterator(); itetest.hasNext();) {
                              //System.out.println("------------- : "+itetest.next().toString());
                            //}
                        //}

                l_temporaryLinks = linkUtil_.mergeTwoLinksCollection(
                    l_startOfLinks, l_endOfLinks_);

                for (Iterator itu = l_endOfLinks_.iterator(); itu.hasNext(); ) {
                  l_ObjectToRegister.add( ((DisplayableLinkImpl)itu.next()).getEorSecondEnd() );
                }

                // may filter the ACL or Bridge last link
                for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                    dLink = (DisplayableLinkImpl)itu.next();

                    linkRights = interpreter_.getLinkRightsInstances().get(
                                dLink.getEorSecondEnd().getName());
                    if (linkRights == null) {
                            interpreter_.addInstancesForNode(dLink.getEorSecondEnd());
                            linkRights = interpreter_.getLinkRightsInstances().get(
                                dLink.getEorSecondEnd().getName());
                    } // if
                    // the other iterationCount values have a control
                    // in the final search
                    if (linkRights.getDetectAddonPriorityInLastLinksNonAGORights()
                            //&&
                            //(iterationsCount == 3 ||
                            //iterationsCount == 6 ||
                            //iterationsCount == 9 ||
                            //iterationsCount == 13 ||
                            //iterationsCount == 18 ||
                            //iterationsCount == 23 ||
                            //iterationsCount == 28 ||
                            //iterationsCount == 33)
                            ){
                            priority = linkRights.detectAddonPriorityInLastLinkNonAGORights(
                                                    dLink, l_temporaryLinks);
                            if (priority) {
                                itu.remove();

                                    //System.out.println("--- in CompounRightsFactoryImpl, "+
                                      //"l_ownersForOwnAndContainToTest_"+
                                      //" for actors, second ite, first step with selectDirectHiddenExternalAclBridgeLinksToActorForEP(),"+
                                      //" removed link: "+dLink);

                                continue;
                            } // if
                    } // if
                } // for obj

                if (!l_allHiddenActorsLinks_.containsAll(l_temporaryLinks))
                    l_currentActorsLinks_.addAll(l_temporaryLinks);
            } // if (l_endOfLinks_ != null)


            // ---- fourth test in second central iteration
          // analyzes the Actors having privilege to Actors and VirtualFolders
          // and updates l_currentActorsLinks_ or l_currentVirtualFoldersLink
            if (l_endOfLinks_ != null)
                l_endOfLinks_.clear();
            else
                l_endOfLinks_ = new HashSet(INITIAL_CAPACITY);
            l_endOfLinks2_.clear();
            l_temporaryLinks = pri_.
                  selectDirectHiddenPrivilegeLinksForEP(_gview, _viewInBase, ep, true);
            if (l_temporaryLinks != null) {
                for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                          dLink = (DisplayableLinkImpl)itu.next();
                          l_ObjectToRegister.add( dLink.getEorSecondEnd() );
                          if (dLink.getEorSecondEnd() instanceof Actor)
                              l_endOfLinks_.add( dLink );
                          else if (dLink.getEorSecondEnd() instanceof VirtualFolder)
                              l_endOfLinks2_.add( dLink );
                } // for itu
            } // if (l_temporaryLinks

            if (!l_endOfLinks_.isEmpty()) {
                l_temporaryLinks = linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_);
            
                // may filter the Privilege last link, but for Actors only
                for (Iterator itu = l_temporaryLinks.iterator(); itu.hasNext(); ) {
                    dLink = (DisplayableLinkImpl)itu.next();

                    linkRights = interpreter_.getLinkRightsInstances().get(
                                dLink.getEorSecondEnd().getName());
                    if (linkRights == null) {
                            interpreter_.addInstancesForNode(dLink.getEorSecondEnd());
                            linkRights = interpreter_.getLinkRightsInstances().get(
                                dLink.getEorSecondEnd().getName());
                    } // if
                    // the other iterationCount values have a control
                    // in the final search
                    if (linkRights.getDetectAddonPriorityInLastLinksNonAGORights()
                            //&&
                            //(iterationsCount == 3 ||
                            //iterationsCount == 6 ||
                            //iterationsCount == 9 ||
                            //iterationsCount == 13 ||
                            //iterationsCount == 18 ||
                            //iterationsCount == 23 ||
                            //iterationsCount == 28 ||
                            //iterationsCount == 33)
                            ){
                            priority = linkRights.detectAddonPriorityInLastLinkNonAGORights(
                                                    dLink, l_temporaryLinks);
                            if (priority) {
                                itu.remove();

                                    // for testing and tracking
                                    //System.out.println("--- in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                      //      "remove a dLink after linkRights.detectAddonPriorityInLastLinkNonAGORights()");

                                continue;
                            }
                    } // if
                } // for obj

                if (!l_allHiddenActorsLinks_.containsAll(l_temporaryLinks))
                    l_currentActorsLinks_.addAll(l_temporaryLinks);
            }
            if (!l_endOfLinks2_.isEmpty()) {
                l_temporaryLinks = linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks2_);

                if (!l_allHiddenVirtualFoldersLinks_.containsAll(l_temporaryLinks))
                    l_currentVirtualFoldersLinks_.addAll(l_temporaryLinks);
            } // if

                        // for testing and tracking
                        //System.out.println("--- the second test in second iteration "+
                          //" detects the Actors having privilege to Actors and VirtualFolders:");
                        //System.out.println("--- in l_currentActorsLinks_:");
                        //for (Iterator itetest = l_currentActorsLinks_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("--- in l_currentVirtualFoldersLinks_:");
                        //for (Iterator itetest = l_currentVirtualFoldersLinks_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
            
            // ---- fifth test in the second central iteration
          // analyzes the actors that have a proper AG context,
          // that are bridge targets, virtual members, reference of an alias,
          // or that are ACL targets or privilege targets...
          // detects if their current Account/Group context replaces partially or totally
          // the proper AG context of the previous node, and updates l_currentOwnersLinks_
                        
                                // for testing and tracking
                                //System.out.println("in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(), "+
                                  //"in the central loop, start of the second ite, fifth step");

            if (l_temporaryLinks == null)
                l_temporaryLinks = new HashSet(INITIAL_CAPACITY / 3);
            else l_temporaryLinks.clear();         

            for (Iterator itu = l_startOfLinks.iterator(); itu.hasNext() ;) {
                dLink = (DisplayableLinkImpl)itu.next();

                // defines the Account/Groups context of the dLink second end
                l_members = linkUtil_.getSecondEndAGcontext(dLink);
                        
                                // for testing and tracking
                                //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(), "+
                                  //"in the central loop, the second ite, fifth step, after the call to getSecondEndAGcontext() "+
                                  //"--- processes the link "+dLink);

                // changed in 0.7.1
                if (l_members == null ||
                        dLink.getProperties() == null ||
                        dLink.getProperties().get(
                                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null) {

                                // for testing and tracking
                                //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(), "+
                                  //"in the central loop, the second ite, fifth step, after the call to getSecondEndAGcontext() "+
                                  //"--- no processing of the link "+dLink);

                                        continue;
                } // if

                nextIndex = ((Integer)dLink.getProperties().get(
                            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                            intValue();
                // nextIndex has to be the index of the actor as second end
                if (nextIndex != dLink.noOfNodes() - 1)
                    continue;

                for (Object memb : l_members) {
                    if (dLink.getEorFirstEnd() == memb ||
                                _viewInBase.containsBaseObject((BaseObject)memb)) continue;

                    rightType = null;
                    // set the XID comment if the first end is an actor and the next node,
                    // an actor which runs under its current user/group or secondary group
                    rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                    if (rightType != null && memb instanceof GroupID)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_GROUP);
                    else if (rightType != null && memb instanceof UserID)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_USER);

                    dLink_2 = DisplayableLinkImpl.valueOf(
                            dLink, (BaseObject)memb,
                            new AccessControlLinkImpl());

                    if (dLink_2 == null) continue;

                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

                    if (memb instanceof GroupID &&
                            ((ImmutableActor)ep).isMemberToSecondaryGroup((GroupID)memb)) {
                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink_2.getEorLastLink().addComment(
                            AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                    } // if

                                // for testing and tracking
                                //System.out.println(" --- add the link "+dLink_2);

                    l_temporaryLinks.add(dLink_2);
                    l_ObjectToRegister.add( memb ); // new in v065
                } // for memb

                            // for testing and tracking
                            //if (iterationsCount == 8) {
                                //System.out.println("in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(), "+
                                  //"dLink: "+dLink+", and the node before the last one: "+
                                  //dLink.getEorAllNodes()[dLink.noOfNodes() - 2]);
                            //}

            } // for itu

                            // for testing and tracking
                            //System.out.println("CompoundRightsFactoryImpl.detectHiddenCompoundEpRights():");
                            //System.out.println(" --- in the central loop at iteration "+iterationsCount);
                            //System.out.println(" --- in the second ite, the fourth step on Actors, the current result is: ");
                            //System.out.println("as owners links:");
                            //for (Iterator itetest = l_temporaryLinks.iterator(); itetest.hasNext();) {
                              //System.out.println("------------- : "+itetest.next().toString());
                            //}

            if (!l_allHiddenOwnersLinks_.containsAll(l_temporaryLinks))
                l_currentOwnersLinks_.addAll(l_temporaryLinks);

            // ---- sixth test in the second central iteration:
            // detects the Actors which are aliases,
            // updates l_currentActorsLinks_,
            // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
            // and puts the new ends in l_ObjectToRegister
            l_endOfLinks_ = selectDirectHiddenReferenceForAlias(_gview, _viewInBase, ep);

            if (l_endOfLinks_ != null) {
                  if (l_temporaryLinks == null)
                        l_temporaryLinks = new HashSet(INITIAL_CAPACITY / 3);
                  else l_temporaryLinks.clear();

                  l_temporaryLinks.addAll(linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_));

                  for (Object obj : l_temporaryLinks) {
                         dLink = (DisplayableLinkImpl)obj;

                                    // for testing and tracking
                                    //System.out.println("in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights():");
                                    //System.out.println(" --- sixth test in the second central iteration "+iterationsCount);
                                    //System.out.println(" --- add the link "+dLink);


                         if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember &&
                                 !l_allHiddenOwnersLinks_.contains(dLink))
                             l_currentOwnersLinks_.add(dLink);
                         else if (dLink.getEorSecondEnd() instanceof ImmutableActor &&
                                 !l_allHiddenActorsLinks_.contains(dLink))
                             l_currentActorsLinks_.add(dLink);
                         else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder &&
                                 !l_allHiddenVirtualFoldersLinks_.contains(dLink))
                             l_currentVirtualFoldersLinks_.add(dLink);
                         // noop if the reference is a resource for instance
                  } // for obj
            } // if (l_endOfLinks_!= null)

              // -----seventh test in the second central iteration:
              // analyzes an Actor as reference of an alias, and
              // having a proper and applicable AG context,
              // and puts each element of this context as a new end if it is not in the view,
              // updates l_currentOwnersLinks_
            for (Object obj : l_startOfLinks) {
                dLink = (DisplayableLinkImpl)obj;
                if (!dLink.getEorLastLink().getLinkTypeValue(AccessControlLink.IS_ALIAS))
                    continue;

                                    // for testing and tracking
                                    //System.out.println("in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights():");
                                    //System.out.println(" --- seventh test in the second central iteration "+iterationsCount);
                                    //System.out.println(" --- analyze the link "+dLink);

                l_members = linkUtil_.getSecondEndAGcontext(dLink);
                if (l_members == null || dLink.getProperties() == null ||
                            dLink.getProperties().get(
                                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null)
                                        continue;

                nextIndex = ((Integer)dLink.getProperties().get(
                            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                            intValue();
                // nextIndex has to be the index of the actor just before the second end
                if (nextIndex == dLink.noOfNodes() - 2) // != dLink.noOfNodes() - 2)
                    throw new InternalError("seventh test in the second central iteration");
                    //continue;

                for (Object memb : l_members) {
                    if (dLink.getEorFirstEnd() == memb ||
                                _viewInBase.containsBaseObject((BaseObject)memb)) continue;

                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, (GroupIDMember)memb,
                            new AccessControlLinkImpl());
                    if (dLink_2 == null) continue;

                    dLink_2.getEorConnectedLinks((GroupIDMember)dLink.getEorSecondEnd())[0].
                            setLinkTypeToTrue(AccessControlLink.IS_ALIAS);
                    if (memb instanceof GroupID) {
                        dLink_2.getEorConnectedLinks((GroupIDMember)dLink.getEorSecondEnd())[0].
                              addComment(
                                AccessControlLink.IS_ALIAS,
                                AccessControlLink.C_SET_GROUP + ": "+((GroupIDMember)memb).getNickName());
                    }
                    else if (memb instanceof UserID) {
                      dLink_2.getEorConnectedLinks((GroupIDMember)dLink.getEorSecondEnd())[0].
                              addComment(
                                AccessControlLink.IS_ALIAS,
                                AccessControlLink.C_SET_USER + ": "+((GroupIDMember)memb).getNickName());
                    } // else

                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

                    // set the GLOBAL type if the first end is an actor, and the next node
                    // is a secondary group
                    if (memb instanceof GroupID && dLink.getEorSecondEnd() instanceof Actor &&
                      ((Actor)dLink.getEorSecondEnd()).isMemberToSecondaryGroup((GroupID)memb) ) {
                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink_2.getEorLastLink().addComment(
                                  AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                    } // if

                                    //System.out.println(" --- add the link "+dLink_2);

                    if (!l_allHiddenOwnersLinks_.contains(dLink_2))
                        l_currentOwnersLinks_.add(dLink_2);
                } // for memb

            } // for obj
          } // for ite

                            // for testing and tracking
                            //System.out.println("in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights():");
                            //System.out.println(" --- in the central loop at iteration "+iterationsCount);
                            //System.out.println(" --- start of the third ite on EP-VF");

          // THIRD CENTRAL ITERATION, on the EP VIRTUAL FOLDERS, including 4 tests.
          // 1 - analyzes the VirtualFolders which have complementary links
          // from their AcsAddon ACS, may update l_startOfLinks for the next tests,
          // l_currentActorsLinks_, l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
          // and puts the new ends in l_ObjectToRegister
          // 2 - detects the hidden actors through a virtual member relation
          // and updates l_currentActorsLinks_
          // where rights on the actor are copied from the rights on the VirtualFolder
          // 3 - detects the VirtualFolders which are aliases,
          // updates l_currentActorsLinks_,
          // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
          // and puts the new ends in l_ObjectToRegister
          // 4 - detects the VirtualFolders which are references of an Actor alias
          // having a proper and applicable AG context,
          // and puts each element of this context as a new end if it is not in the view,
          // updates l_currentOwnersLinks_
          for (Iterator ite = l_virtualFoldersToTest_.iterator() ; ite.hasNext() ; ) {
            vf = (ImmutableVirtualFolder)ite.next();
            
            if (!vf.getMainVFType().equals(VirtualFolderImpl.EP_VFOLDER))
                continue;
            
            l_startOfLinks = linkUtil_.
                selectDisplayableLinksForSecondEnd(vf, l_hiddenVirtualFoldersLinks_);
            
            if (l_startOfLinks == null) continue;

            nodeRights = interpreter_.getNodeRightsInstances().get(vf.getName());
            if (nodeRights == null) {
                interpreter_.addInstancesForNode(vf);
                nodeRights = interpreter_.getNodeRightsInstances().get(vf.getName());
            } // if

            // ---- first test in the third central iteration:
            // analyzes the VirtualFolders which have complementary links
            // from their AcsAddon ACS, may update l_startOfLinks for the next tests,
            // l_currentActorsLinks_, l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
            // and puts the new ends in l_ObjectToRegister
            if (nodeRights.getSelectAddonNewHiddenNodeForVirtualFolder()) {

                l_temporaryLinks5.clear();
                l_temporaryLinks4 = nodeRights.selectAddonNewHiddenNodeForVirtualFolder(
                            _viewInBase, vf, l_startOfLinks, _upd_m_l_DisplayableLinks,
                            l_temporaryLinks5);
                // new in 0.7.2
                l_ObjectToRegister.addAll(l_temporaryLinks5);

                if (l_temporaryLinks4 != null) {
                    for (Iterator itu = l_temporaryLinks4.iterator();
                            itu.hasNext(); ) {

                          dLink = (DisplayableLinkImpl)itu.next();
                          l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                          if (dLink.getEorSecondEnd() instanceof Actor &&
                                  !l_allHiddenActorsLinks_.contains(dLink))
                              l_currentActorsLinks_.add( dLink );
                    } // for itu
                } // if (l_temporaryLinks4 != null)

            } // if node.getSelectAddonNewHiddenNodeForVirtualFolder()

            // ---- second test in the third central iteration
            // Selects the actors as virtual members of vf
            if (l_endOfLinks_ != null)
                l_endOfLinks_.clear();
            else
                l_endOfLinks_ = new HashSet(INITIAL_CAPACITY);
            l_temporaryLinks = selectDirectHiddenMembersForVirtualFolder(
                                    _gview, _viewInBase, vf);
            if (l_temporaryLinks != null) {
                  // virtual members may be actors, since it is an EP_FOLDER VirtualFolder
                  for (Object obj : l_temporaryLinks) {
                     dLink = (DisplayableLinkImpl)obj; 
                     if (dLink.getEorSecondEnd() instanceof ImmutableActor)
                         l_endOfLinks_.add(dLink);
                  } // for
            } // if

            if (!l_endOfLinks_.isEmpty()) {

                        // for testing and tracking
                        //if (iterationsCount == 8) {
                            //System.out.println("--- the first test in the third iteration "+
                              //" detects the hidden actors through a virtual member relation.");
                            //for (Iterator itetest = l_endOfLinks_.iterator(); itetest.hasNext();) {
                              //System.out.println("------------- : "+itetest.next().toString());
                            //}
                        //}

              l_temporaryLinks = linkUtil_.mergeTwoLinksCollection(
                l_startOfLinks, l_endOfLinks_);

              // the rights on vf are copied as rights on the actor which is a virtual member.
              // the properties of the link are not changed.
              for (Iterator ito = l_temporaryLinks.iterator(); ito.hasNext(); ) {
                  dLink = (DisplayableLinkImpl)ito.next();
                  dLink.getEorLastLink().
                          mergeAccessControlLinkWithoutFiltering(
                            dLink.getEorAllACLinks()[dLink.noOfNodes() - 2]);

                            //System.out.println("--- the first test in the third iteration "+
                              //" merge in the last ACLink, the ACLink:");
                            //System.out.println("--- "+dLink.getEorAllACLinks()[dLink.noOfNodes() - 2]);

              } // for ito

              if (!l_allHiddenActorsLinks_.containsAll(l_temporaryLinks))
                  l_currentActorsLinks_.addAll(l_temporaryLinks);
            } // if (!l_endOfLinks_.isEmpty()) {

            // ---- third test in the third central iteration:
            // detects the VirtualFolders which are aliases,
            // updates l_currentActorsLinks_,
            // l_currentOwnersLinks_ or l_currentVirtualFoldersLinks_,
            // and puts the new ends in l_ObjectToRegister
            l_endOfLinks_ = selectDirectHiddenReferenceForAlias(_gview, _viewInBase, vf);

            if (l_endOfLinks_ != null) {
                  if (l_temporaryLinks == null)
                        l_temporaryLinks = new HashSet(INITIAL_CAPACITY / 3);
                  else l_temporaryLinks.clear();

                  l_temporaryLinks.addAll(linkUtil_.
                    mergeTwoLinksCollection(l_startOfLinks, l_endOfLinks_));

                  for (Object obj : l_temporaryLinks) {
                         dLink = (DisplayableLinkImpl)obj;

                         if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember &&
                                 !l_allHiddenOwnersLinks_.contains(dLink)) {
                             l_currentOwnersLinks_.add(dLink);
                             l_ObjectToRegister.add(dLink.getEorSecondEnd());
                         }
                         else if (dLink.getEorSecondEnd() instanceof ImmutableActor &&
                                 !l_allHiddenActorsLinks_.contains(dLink)) {
                             l_currentActorsLinks_.add(dLink);
                             l_ObjectToRegister.add(dLink.getEorSecondEnd());
                         }
                         else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder &&
                                 !l_allHiddenVirtualFoldersLinks_.contains(dLink)) {
                             l_currentVirtualFoldersLinks_.add(dLink);
                             l_ObjectToRegister.add(dLink.getEorSecondEnd());
                         }
                         // noop if the reference is a resource for instance
                  } // for obj
            } // if (l_endOfLinks != null)

              // ---- fourth test in the third central iteration:
              // detects the VirtualFolders which are references of an Actor alias
              // having a proper and applicable AG context,
              // and puts each element of this context as a new end if it is not in the view,
              // updates l_currentOwnersLinks_
            for (Object obj : l_startOfLinks) {
                dLink = (DisplayableLinkImpl)obj;
                if (!(dLink.getEorNodeJustBeforeTheSecondEnd() instanceof Actor) ||
                        !dLink.getEorLastLink().getLinkTypeValue(AccessControlLink.IS_ALIAS))
                    continue;

                                    // for testing and tracking
                                    //if (dLink.getEorSecondEnd().getNickName().equals("Amounts view")) {
                                      //  System.out.println("CompoundRightsFactoryImpl.detectHiddenCompoundEpRights():");
                                      //  System.out.println(" --- fourth test in the third central iteration "+iterationsCount);
                                      //  System.out.println(" --- detects the VirtualFolders which are references of an Actor alias ");
                                      //  System.out.println(" --- analyze the link "+dLink);
                                    //}

                // updates dLink
                l_members = linkUtil_.getSecondEndAGcontext(dLink);
                
                if (l_members == null) {

                                    // for testing and tracking
                                    //if (dLink.getEorSecondEnd().getNickName().equals("Amounts view"))
                                      //  System.out.println(" --- no AG context members for "+dLink);

                    continue;
                } // if
                if (dLink.getProperties() == null ||
                            dLink.getProperties().get(
                                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null) {

                                     // for testing and tracking
                                     //if (dLink.getEorSecondEnd().getNickName().equals("Amounts view"))
                                       // System.out.println(" --- no applicable AG context for "+dLink);

                    continue;
                } // if

                nextIndex = ((Integer)dLink.getProperties().get(
                            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                            intValue();
                // nextIndex has to be the index of the actor just before the second end
                if (nextIndex == dLink.noOfNodes() - 2) // != dLink.noOfNodes() - 2)
                    throw new InternalError("fourth test in the third central iteration");
                    //continue;

                for (Object memb : l_members) {
                    if (dLink.getEorFirstEnd() == memb ||
                                _viewInBase.containsBaseObject((BaseObject)memb)) continue;

                                    // for testing and tracking
                                    //if (dLink.getEorSecondEnd().getNickName().equals("Amounts view"))
                                      //  System.out.println(" --- try to add memb: "+memb+
                                        //    ", on the link "+dLink);
                    
                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, (GroupIDMember)memb,
                            new AccessControlLinkImpl());
                    if (dLink_2 == null) {

                                    // for testing and tracking
                                    //if (dLink.getEorSecondEnd().getNickName().equals("Amounts view"))
                                      //  System.out.println(" --- do not add memb "+memb);

                        continue;
                    }

                                    // for testing and tracking
                                    //if (dLink.getEorSecondEnd().getNickName().equals("Amounts view"))
                                      //  System.out.println(" --- adding of memb "+memb);
                    
                    dLink_2.getEorConnectedLinks((VirtualFolder)dLink.getEorSecondEnd())[0].
                            setLinkTypeToTrue(AccessControlLink.IS_ALIAS);
                    if (memb instanceof GroupID) {
                        dLink_2.getEorConnectedLinks((VirtualFolder)dLink.getEorSecondEnd())[0].
                              addComment(
                                AccessControlLink.IS_ALIAS,
                                AccessControlLink.C_SET_GROUP + ": "+((GroupIDMember)memb).getNickName());
                    }
                    else if (memb instanceof UserID) {
                      dLink_2.getEorConnectedLinks((VirtualFolder)dLink.getEorSecondEnd())[0].
                              addComment(
                                AccessControlLink.IS_ALIAS,
                                AccessControlLink.C_SET_USER + ": "+((GroupIDMember)memb).getNickName());
                    } // else

                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

                                    //System.out.println(" --- add the link "+dLink_2);

                    if (!l_allHiddenOwnersLinks_.contains(dLink_2))
                        l_currentOwnersLinks_.add(dLink_2);

                } // for memb

            } // for obj
          } // for ite

                            // for testing and tracking
                        //System.out.println("\n CompoundRightsFactoryImpl.detectHiddenCompoundEpRights():");
                        //System.out.println("-- end of the central loop at iteration "+iterationsCount);

                          //if (//iterationsCount ==  1 || iterationsCount == 2 ||
                            //      iterationsCount == 3 || iterationsCount == 4 //||
                                  //iterationsCount == 5 || iterationsCount == 6 || iterationsCount == 7
                                  //) {
                                        // for testing and tracking
                                    //    System.out.println("--- l_currentOwnersLinks_ at iteration: "+iterationsCount);
                                      //  for (Iterator itetest = l_currentOwnersLinks_.iterator(); itetest.hasNext();) {
                                        //    dLink = (DisplayableLinkImpl) itetest.next();
                                          //  if (dLink.noOfNodes() == iterationsCount + 2)
                                            //    System.out.println("------------- : "+dLink.toString());
                                            //else
                                              //  System.out.println("----no right nb of nodes : "+dLink.toString());
                                        //}
                                        //System.out.println("--- l_currentActorsLinks_ at iteration: "+iterationsCount);
                                        //for (Iterator itetest = l_currentActorsLinks_.iterator(); itetest.hasNext();) {
                                          //  dLink = (DisplayableLinkImpl) itetest.next();
                                            //if (dLink.noOfNodes() == iterationsCount + 2)
                                              //  System.out.println("------------- : "+dLink.toString());
                                            //else
                                              //  System.out.println("----no right nb of nodes : "+dLink.toString());
                                        //}
                                        //System.out.println("--- l_currentVirtualFoldersLinks_ at iteration: "+iterationsCount);
                                        //for (Iterator itetest = l_currentVirtualFoldersLinks_.iterator(); itetest.hasNext();) {
                                          //  dLink = (DisplayableLinkImpl) itetest.next();
                                            //if (dLink.noOfNodes() == iterationsCount + 2)
                                              //  System.out.println("------------- : "+dLink.toString());
                                            //else
                                              //  System.out.println("----no right nb of nodes : "+dLink.toString());
                                        //}
                          //} // if
                        //System.out.println("--- at the end of iteration "+iterationsCount+
                          //", l_whileLoopOtherContextActorLinks_ for the next iteration:");
                        //for (Iterator itetest = l_currentOtherContextActorLinks_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}

          // END OF THE THREE LOOPS IN THE CENTRAL LOOP
          // initializing of the next while iteration...
          ok_1 = false;
          ok_2 = false;
          ok_3 = false;

          // from l_currentActorsLinks_
          if (!l_currentActorsLinks_.isEmpty()) {
              l_actorsToTest_ = linkUtil_.
                getAllEndsInDisplayableLinks(l_currentActorsLinks_, false);
              l_hiddenSourceToTest_.addAll((Set)((HashSet)l_actorsToTest_).clone());
              ok_1 = true;
              l_allHiddenActorsLinks_.addAll(l_currentActorsLinks_);
              //l_currentSearchingLinks_.addAll(l_currentActorsLinks_);
              l_hiddenActorsLinks_ = new HashSet(l_currentActorsLinks_);
          } // if (!l_currentActorsLinks_.isEmpty())
          
          // from l_currentOwnersLinks_
          if (!l_currentOwnersLinks_.isEmpty()) {
              l_ownersForOwnAndContainToTest_ = linkUtil_.
                getAllEndsInDisplayableLinks(l_currentOwnersLinks_, false);
              l_hiddenSourceToTest_.addAll((Set)((HashSet)l_ownersForOwnAndContainToTest_).clone());
              ok_2 = true;
              l_allHiddenOwnersLinks_.addAll(l_currentOwnersLinks_);
              //l_currentSearchingLinks_.addAll(l_currentOwnersLinks_);
              l_hiddenOwnersLinks_ = new HashSet(l_currentOwnersLinks_);
          } // if (!l_currentOwnersLinks_
          
          // from l_currentVirtualFoldersLinks_
          if (!l_currentVirtualFoldersLinks_.isEmpty()) {
              l_virtualFoldersToTest_ = linkUtil_.
                getAllEndsInDisplayableLinks(l_currentVirtualFoldersLinks_, false);
              l_hiddenSourceToTest_.addAll((Set)((HashSet)l_virtualFoldersToTest_).clone());
              ok_3 = true;
              l_allHiddenVirtualFoldersLinks_.addAll(l_currentVirtualFoldersLinks_);
              //l_currentSearchingLinks_.addAll(l_currentVirtualFoldersLinks_);
              l_hiddenVirtualFoldersLinks_ = new HashSet(l_currentVirtualFoldersLinks_);
          } // if (!l_currentVirtualFoldersLinks_

          // from l_whileLoopOtherContextActorLinks_
          //l_whileLoopOtherContextActorLinks_ = (Set)((HashSet)l_currentOtherContextActorLinks_).clone();
          //ok_3 = l_otherContextActorLinks_.addAll(l_currentOtherContextActorLinks_);
          
          // ok_1/2/3: break the while loop if the 3 current DLinks set are 
          // not changing over the last iteration

          if (Thread.interrupted()) {
            throw new InterruptedException();
          }

          //if (iterationsCount == 20) {

                        // for testing and tracking
                        //System.out.println("in CompoundRightsFactoryImpl."+
                          //"detectHiddenCompoundEpRights(), 20 ITERATIONS IN THE CENTRAL SEARCHING LOOP :");
                        //System.out.println("--- at the end of iteration "+iterationsCount+
                          //", l_currentOwnersLinks_:");
                        //for (Iterator itetest = l_currentOwnersLinks_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("--- at the end of iteration "+iterationsCount+
                          //", l_currentActorsLinks_:");
                        //for (Iterator itetest = l_currentActorsLinks_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("--- at the end of iteration "+iterationsCount+
                          //", l_currentOtherContextActorLinks_ for the next iteration:");
                        //for (Iterator itetest = l_currentOtherContextActorLinks_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("--- at the end of iteration "+iterationsCount+
                          //", l_actorsToTest_:");
                        //for (Iterator itetest = l_actorsToTest_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("--- at the end of iteration "+iterationsCount+
                          //", l_ownersForOwnAndContainToTest_:");
                        //for (Iterator itetest = l_ownersForOwnAndContainToTest_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("--- at the end of iteration "+iterationsCount+
                          //", l_virtualFoldersToTest_ for the next iteration:");
                        //for (Iterator itetest = l_virtualFoldersToTest_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}
                        
          //} // if (iterationsCount
          
      } // while ( (!l_actorsToTest_.isEmpty() ||
        // !l_ownersForOwnAndContainToTest_.isEmpty() ||
        // !l_virtualFoldersToTest_.isEmpty() ) &&
        // iterationsCount < 40)

      //
      // ----------------- END OF THE CENTRAL WHILE LOOP -----------------------------
      // -------------- THAT SEARCHES THE INTERMEDIATE NODES -------------------------
      //
      
      if (iterationsCount == 40) throw new InternalError(
        "The rights search is aborted, because "+
        "in an access path, there are 40 intermediate nodes. "+
        "This may be due to a network of open ACS which is too complex.");
      
      for (Iterator iti = l_allHiddenOwnersLinks_.iterator(); iti.hasNext(); ) {
          dLink = (DisplayableLinkImpl) iti.next();
          // to avoid the cases handled by the other methods
          if (dLink.noOfNodes() < 3) iti.remove();
      } // for iti
      for (Iterator iti = l_allHiddenActorsLinks_.iterator(); iti.hasNext(); ) {
          dLink = (DisplayableLinkImpl) iti.next();
          // to avoid the cases handled by the other methods
          if (dLink.noOfNodes() < 3) iti.remove();
      } // for iti
      for (Iterator iti = l_allHiddenVirtualFoldersLinks_.iterator(); iti.hasNext(); ) {
          dLink = (DisplayableLinkImpl) iti.next();
          // to avoid the cases handled by the other methods
          if (dLink.noOfNodes() < 3) iti.remove();
      } // for iti

                        // for testing and tracking
                        //System.out.println("in CompoundRightsFactoryImpl."+
                          //"detectHiddenCompoundEpRights(), THE STATE AFTER THE CENTRAL SEARCH IS:");
                        //System.out.println("----- l_allHiddenOwnersLinks_: ");
                        //for (Iterator itetest = l_allHiddenOwnersLinks_.iterator() ; itetest.hasNext() ; ) {
                          //System.out.println("------------- :"+
                            //((DisplayableLinkImpl)itetest.next()).toString() );
                        //}
                        //System.out.println("----- l_allHiddenActorsLinks_: ");
                        //for (Iterator itetest = l_allHiddenActorsLinks_.iterator() ; itetest.hasNext() ; ) {
                          //System.out.println("------------- :"+
                            //((DisplayableLinkImpl)itetest.next()).toString() );
                        //}
                        //System.out.println("------- l_allHiddenVirtualFoldersLinks_: ");
                        //for (Iterator itetest = l_allHiddenVirtualFoldersLinks_.iterator(); itetest.hasNext();) {
                          //System.out.println("------------- : "+itetest.next().toString());
                        //}

      l_hiddenSourceLinks_.addAll(l_allHiddenOwnersLinks_);
      l_hiddenSourceLinks_.addAll(l_allHiddenActorsLinks_);
      l_hiddenSourceLinks_.addAll(l_allHiddenVirtualFoldersLinks_);

      //
      // ------------- START OF THE FINAL 'ITI, ITE, ITO, ITA' LOOPS ------------------------
      // ---------------- THEN ANOTHER 'ITI, ITO' LOOPS -------------------------------------
      // ------------- WHERE DISPLAYABLE LINKS ARE UPDATED IN _upd_m_l_DisplayableLinks -----
      // ------------- THROUGH THE ADDING OF THE LAST ACLINK CONNECTED TO A VIEW OBJECT -----
      //
 
                                // for testing and tracking
                                //System.out.println("");
                                //System.out.println("In CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                  //      "before the first final loop, _upd_m_l_DisplayableLinks: ");
                                //ImmutableName key;
                                //for (Iterator ite = _upd_m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                                  // key = (ImmutableName) ite.next();
                                   //System.out.println("--- for key: "+key);
                                   //for (Iterator ita = ((List)_upd_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                                    //System.out.println("------------"+
                                     // ((DisplayableLinkImpl)ita.next()).toString() );
                                  //} // for ita
                                //} // for ite

      // FIRST FINAL LOOP ON EACH SELECTED SOURCE AND THE VIEW TARGETS, including 11 tests:
      // first test in the iti/ite on groupIDMember:
        // detects the GroupIDMembers which have complementary links
        // from their AcsAddon ACS
      // second test in the iti/ite on actor:
        // detects the Actors which have complementary links
        // from their AcsAddon ACS
      // third test in the iti/ite on virtual folder only:
        // detects the VirtualFolders which have complementary links
        // from their AcsAddon ACS
      // fourth test in iti/ite:
        // if so is a virtual folder with virtual members
      // fifth test in iti/ite:
        // if so is a virtual folder running under a groupidmember
      // sixth test in iti/ite:
        // if ep is a privilege source to a view target
      // seventh test in iti/ite:
        // if ep is a view target owner or group
      // eighth test in iti/ite:
        // if there is an acl link for the last link on the path
      // nineth test in iti/ite on actor only:
        // if ep is the bridge source of a view actor
      // tenth test in iti/ite:
        // if ep is an alias for a reference in the view
      // eleventh test in iti/ite
        // if ep is an alias reference having an AG context from the previous actor,
        // and if an element of the context is in the view
      for (Iterator iti = l_hiddenSourceToTest_.iterator() ; iti.hasNext() ; ) {
        so = (ImmutableSource) iti.next();
        if (so == null) continue;
        if (_viewInBase.getEorM_Sources().containsKey(so.getName())) {

                          // for testing and tracking
                          //System.out.println(
                            //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                            //" there is a abnormal FINAL LOOP on the source, "+so.getName()+
                            //" that is a view node in "+_viewInBase.getName()+
                            //" and it is also in l_hiddenSourceToTest_.");
            
            continue;
        } // if

        nodeRights = interpreter_.getNodeRightsInstances().get(so.getName());
        if (nodeRights == null) {
            interpreter_.addInstancesForNode(so);
            nodeRights = interpreter_.getNodeRightsInstances().get(so.getName());
        } // if

                          // for testing and tracking
                          //System.out.println(
                            //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                            //" FINAL LOOP on the hidden source "+so.getName());

        for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ;
          ite.hasNext() ; ) {
            viewTarg = (ImmutableTarget) ite.next();
            // first test in the first final loop on groupIDMember:
            // detects the GroupIDMembers which have complementary links
            // from their AcsAddon ACS
            if (so instanceof ImmutableGroupIDMember) {
                if (nodeRights.getSelectAddonLastViewNodeAfterGroupIDMember()) {

                     for (Iterator ito = l_hiddenSourceLinks_.iterator(); ito.hasNext(); ) {

                        dLink = (DisplayableLinkImpl) ito.next();
                        if (dLink.getEorSecondEnd() != so ) continue;

                        dLink_2 = nodeRights.selectAddonLastViewNodeAfterGroupIDMember(
                            _viewInBase, viewTarg, dLink, _upd_m_l_DisplayableLinks);

                        // rule 'virtual folder: an indirect access to a Resource that is not
                        // a virtual member, is always through a virtual member'
                        if (dLink_2 == null ||
                                linkUtil_.getContainVirtualFolderWithoutMember(dLink_2)) continue;

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the first test in the first final loop produces: "+dLink_2);

                        linkRights = interpreter_.getLinkRightsInstances().get(viewTarg.getName());
                        linkRights.addDisplayableLink(
                                        dLink_2, _upd_m_l_DisplayableLinks);

                     } // for (Iterator ito

                } // if node.getSelectAddonNewHiddenNodeForGroupIDMember()
            } // if (ep instanceof ImmutableGroupIDMember)

            // second test in the first final loop on actor:
            // detects the Actors which have complementary links
            // from their AcsAddon ACS
            if (so instanceof ImmutableActor) {

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the second test on actor, in the first final loop, starts. ");

                if (nodeRights.getSelectAddonLastViewNodeAfterActor()) {

                     for (Iterator ito = l_hiddenActorsLinks_.iterator(); ito.hasNext(); ) {

                        dLink = (DisplayableLinkImpl) ito.next();
                        if (dLink.getEorSecondEnd() != so ) continue;

                        dLink_2 = nodeRights.selectAddonLastViewNodeAfterActor(
                            _viewInBase, viewTarg, dLink, _upd_m_l_DisplayableLinks);

                        if (dLink_2 == null)
                            continue;

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the second test in the first final loop produces: "+dLink_2);

                        linkRights = interpreter_.getLinkRightsInstances().get(
                                viewTarg.getName());
                        linkRights.addDisplayableLink(
                                    dLink_2, _upd_m_l_DisplayableLinks);

                     } // for (Iterator ito

                } // if node.getSelectAddonNewHiddenNodeForActor()
            } // if (ep instanceof ImmutableActor)

            // third test in the first final loop on virtual folder:
            // detects the VirtualFolders which have complementary links
            // from their AcsAddon ACS
            if (so instanceof ImmutableVirtualFolder && viewTarg instanceof ImmutableResource) {
                if (nodeRights.getSelectAddonLastViewNodeAfterVirtualFolder()) {

                     for (Iterator ito = l_allHiddenVirtualFoldersLinks_.iterator(); ito.hasNext(); ) {

                        dLink = (DisplayableLinkImpl) ito.next();
                        if (dLink.getEorSecondEnd() != so ) continue;

                        dLink_2 = nodeRights.selectAddonLastViewNodeAfterVirtualFolder(
                            _viewInBase, (ImmutableResource)viewTarg, dLink,
                            _upd_m_l_DisplayableLinks);

                        if (dLink_2 == null)
                            continue;

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the third test in the first final loop produces: "+dLink_2);

                        linkRights = interpreter_.getLinkRightsInstances().get(
                                viewTarg.getName());
                        linkRights.addDisplayableLink(
                                    dLink_2, _upd_m_l_DisplayableLinks);

                     } // for (Iterator ito

                } // if node.getSelectAddonNewHiddenNodeForGroupIDMember()
            } // if (ep instanceof ImmutableGroupIDMember)

        } // ite on _viewInBase.getEorM_Targets().values() for viewTarg

        if (so instanceof VirtualFolder) {

            for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ;
              ite.hasNext() ; ) {
                viewTarg = (ImmutableTarget) ite.next();

                // fourth test in the first final loop:
                // if so is a virtual folder with virtual members.
                // iterates over each Resource
                if (viewTarg instanceof ImmutableResource) {
                    viewRes = (ImmutableResource)viewTarg;

                    if (((VirtualFolder)so).containsBaseObject(viewRes)) {

                        for (Iterator ito = l_allHiddenVirtualFoldersLinks_.iterator() ;
                          ito.hasNext() ; ) {

                            dLink = (DisplayableLinkImpl) ito.next();
                            if (dLink.getEorSecondEnd() != so ) continue;
                            if (dLink.getEorFirstEnd() == viewTarg ) continue;

                                      // for testing and tracking
                                      //System.out.println(
                                        //"------ in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" extension for a virtual member with the initial link 
                                        // in l_hiddenVirtualFoldersLinks_: "+dLink);

                            dLink_2 = DisplayableLinkImpl.valueOf(
                                    dLink, viewRes, new AccessControlLinkImpl());

                            dLink_2.getEorLastLink().
                                    setLinkTypeToTrue(AccessControlLink.HAS_VIRTUAL_MEMBER);

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the fourth test in the first final loop produces: "+dLink_2);


                            linkRights = interpreter_.getLinkRightsInstances().get(
                                    viewRes.getName());
                            linkRights.addDisplayableLink(
                                dLink_2, _upd_m_l_DisplayableLinks);

                        } // for ito

                    } // if
                } // if (!(viewTarg instanceof ImmutableResource)

               // fifth test in the first final loop:
                // if so is a virtual folder with runs under a GroupIDMember.
                // iterates over each GroupIDMember in the view
                if (viewTarg instanceof ImmutableGroupIDMember) {
                    member = (ImmutableGroupIDMember)viewTarg;

                    for (Iterator ito = l_allHiddenVirtualFoldersLinks_.iterator() ;
                      ito.hasNext() ; ) {

                        dLink = (DisplayableLinkImpl) ito.next();
                        if (dLink.getEorSecondEnd() != so ) continue;
                        if (dLink.getEorFirstEnd() == member ) continue;

                        l_members = linkUtil_.getSecondEndAGcontext(dLink);
                        if (l_members == null || 
                                dLink.getProperties() == null ||
                                dLink.getProperties().get(
                                    DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null ||
                                !l_members.contains(member))
                                        continue;

                        nextIndex = ((Integer)dLink.getProperties().get(
                                    DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                                    intValue();
                        // nextIndex has to be the index of the virtual folder as second end
                        if (nextIndex != dLink.noOfNodes() - 1)
                            continue;

                        dLink_2 = DisplayableLinkImpl.valueOf(
                                dLink, member, new AccessControlLinkImpl());

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the fifth test in the first final loop produces: "+dLink_2);


                        dLink_2.getEorLastLink().
                                setLinkTypeToTrue(AccessControlLink.GLOBAL);

                        if (member instanceof GroupID)
                          dLink_2.getEorConnectedLinks((VirtualFolder)dLink.getEorSecondEnd())[0].
                                  addComment(AccessControlLink.GLOBAL, AccessControlLink.C_SET_GROUP);
                        else if (member instanceof UserID)
                          dLink_2.getEorConnectedLinks((VirtualFolder)dLink.getEorSecondEnd())[0].
                                  addComment(AccessControlLink.GLOBAL, AccessControlLink.C_SET_USER);

                        dLink_2.getEorLastLink().
                                setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

                        linkRights = interpreter_.getLinkRightsInstances().get(
                                member.getName());
                        linkRights.addDisplayableLink(
                            dLink_2, _upd_m_l_DisplayableLinks);

                    } // for ito
                } // if (viewTarg instanceof ImmutableGroupIDMember)
            } // for ite on _viewInBase.getEorM_Targets().values() for viewTarg
            
            // stop the iti loop
            continue;
        } // if so instanceof VirtualFolder

        ep = (EligibleParty)so;

        // iterates over each target in the view
        for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ; ite.hasNext() ; ) {
            viewTarg = (ImmutableTarget) ite.next();
            if (ep == viewTarg) continue;
            if (!(viewTarg instanceof ImmutablePrivilegeTarget)) {
                continue;
            }

            nodeRights2 = interpreter_.getNodeRightsInstances().get(viewTarg.getName());
            linkRights = interpreter_.getLinkRightsInstances().get(viewTarg.getName());

            // sixth test in the first final loop:
            // if there is a privilege to a view target (Resource or VirtualFolder)
            acLink = pri_.addPrivilegeInAccessControlLink(
                        (ImmutablePrivilegeTarget)viewTarg, ep, null);

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the sixth test, acLink: "+acLink);

            if (acLink != null) {
                for (Iterator ito = l_hiddenSourceLinks_.iterator() ;
                          ito.hasNext() ; ) {
                    dLink = (DisplayableLinkImpl) ito.next();
                    if (dLink.getEorSecondEnd() != ep ) continue;
                    if (dLink.getEorFirstEnd() == viewTarg ) continue;

                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, viewTarg, acLink);

                        // rule 'virtual folder: an indirect access to a Resource that is not
                        // a virtual member, is always through a virtual member'
                    if (linkUtil_.getContainVirtualFolderWithoutMember(dLink_2)) continue;

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the sixth test in the first final loop produces: "+dLink_2);

                    linkRights.addDisplayableLink(
                                dLink_2,
                                _upd_m_l_DisplayableLinks);
                } // for ito
            } // if (acLink != null)

            if (!(viewTarg instanceof ImmutableResource)) {
                continue;
            }
            viewRes = (ImmutableResource)viewTarg;

                          // for testing and tracking
                          //System.out.println(
                            //"--- in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                            //" 3 loops on the target "+viewTarg.getName())

            // seventh test in the first final loop:
            // if ep is a view target owner or group
            if (viewRes instanceof Actor)
                l_groups = nodeRights2.selectHiddenDirectOwnerContainForActorAsResource(
                                _viewInBase, (Actor)viewRes);
            else
                l_groups = nodeRights2.selectHiddenDirectOwnerContainForTarget(
                                _viewInBase, viewRes);

            if (l_groups != null) {
                for (Iterator ito = l_groups.iterator(); ito.hasNext() ; ) {
                    // ep is a GroupID or an UserID
                    if (ep != ito.next() ) continue;

                    // selects all the dLinks with ep as second end
                    for (Iterator ita = l_allHiddenOwnersLinks_.iterator() ; ita.hasNext() ; ) {

                        dLink = (DisplayableLinkImpl) ita.next();
                        if (dLink.getEorSecondEnd() != ep) continue;
                        if (dLink.getEorFirstEnd() == viewRes ) continue;

                        if (!nodeRights2.withAccessThroughNodesTreeFromEP(
                                        ep, viewRes, null)) continue;

                        // each possible path produces a new DisplayableLink
                        // where res is the second end
                        dLink_2 = DisplayableLinkImpl.valueOf(
                                    dLink, viewRes, new AccessControlLinkImpl());

                        // ACS rule: 'AGO right: A rights overlay G rights'
                        // continue if userID rights are already applicable,
                        // and may update _upd_m_l_DisplayableLinks if ep is an UserID
                        pairKey = ((Name)((BaseObject)dLink.getEorFirstEnd()).
                              getName()).extended(viewRes.getName());
                        if (_upd_m_l_DisplayableLinks.get(pairKey) != null) {
                            priority = linkRights.detectAccountPriorityInLastLinkAGORights(
                                                    dLink_2,
                                                    (List)_upd_m_l_DisplayableLinks.get(pairKey));
                            if (priority) {

                                    // for testing and tracking
                                    //System.out.println("--- in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                      //      "remove after linkRights.detectAccountPriorityInLastLinkAGORights() the dLink: "+
                                        //    dLink_2);

                                continue;
                            }
                            else if (linkRights.getDetectAddonPriorityInLastLinksAGORights()){
                                priority = linkRights.detectAddonPriorityInLastLinkAGORights(
                                                        dLink_2,
                                                        (List)_upd_m_l_DisplayableLinks.get(pairKey));
                                if (priority) {

                                    // for testing and tracking
                                    //System.out.println("--- in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                      //      "remove after linkRights.detectAddonPriorityInLastLinkAGORights() the dLink: "+
                                        //    dLink_2);

                                    continue;
                                }
                            } // else
                        } // if

                        dLink_2 = ago_.
                                addOwnerContainRightsToDisplayableLink(dLink_2);

                                  // for testing and tracking
                                  //System.out.println(
                                    //"------ in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),"+
                                    //" final ep iteration, extension of EP "+ep+" with the link: "+
                                    //dLink_2);

                        if (dLink_2 == null)
                            continue;

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the seventh test in the first final loop produces: "+dLink_2);

                         linkRights.addDisplayableLink(dLink_2, _upd_m_l_DisplayableLinks);
                    } // for ita

                } // for ito
            } // if (l_groups != null)

            // eighth test in the first final loop:
            // if there is an acl link for the last link on the path.
            // update _gview with the ACLs to listen their right changes.
            for (Iterator ito = l_hiddenSourceLinks_.iterator() ;
              ito.hasNext() ; ) {

                dLink = (DisplayableLinkImpl) ito.next();
                if (dLink.getEorSecondEnd() != ep ) continue;
                if (dLink.getEorFirstEnd() == viewRes ) continue;

                                  // for testing and tracking
                                  //System.out.println(
                                    //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                    //" eighth test in the first final loop: try extension to dLink: "+
                                    //dLink);

                // each possible path produces a new DisplayableLink
                // where viewRes is the second end
                acLink = acl_.addAclInAccessControlLink(viewRes, ep, null);
                if (acLink != null) {
                    dLink_2 = DisplayableLinkImpl.valueOf(
                            dLink, viewRes, acLink);

                        // rule 'virtual folder: an indirect access to a Resource that is not
                        // a virtual member, is always through a virtual member'
                    if (dLink_2 == null ||
                            linkUtil_.getContainVirtualFolderWithoutMember(dLink_2)) {

                                  // for testing and tracking
                                  //System.out.println(
                                    //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                    //" eighth test in the first final loop, "+
                                    //"linkUtil_.getContainVirtualFolderWithoutMember() false for "+
                                    //dLink_2);
                        
                        continue;
                    } // if

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the eighth test in the first final loop produces: "+dLink_2);

                    linkRights.addDisplayableLink(
                        dLink_2, _upd_m_l_DisplayableLinks);

                } // if (acLink != null)

            } // for ito
            
            if (ep instanceof Actor) {
                  
                // nineth test in the first final loop on actor:
                // if ep is the bridge source of a view resource
                for (Iterator ito = l_allHiddenActorsLinks_.iterator(); ito.hasNext() ; ) {

                    dLink = (DisplayableLinkImpl) ito.next();
                    if (dLink.getEorSecondEnd() != ep ) continue;
                    if (dLink.getEorFirstEnd() == viewRes ) continue;

                    if (((Actor)ep).getIsACSBridgeSource() &&
                      ((Actor)ep).getEorACSBridgeTarget() == viewRes) {

                        lastLink = new AccessControlLinkImpl();

                        lastLink.addRight(
                            AccessControlLink.BRIDGE,
                            BaseManagerImpl.getInstance().getMetaRight("execute") );

                        // each possible path produces a new DisplayableLink
                        // where viewRes is the second end
                        dLink_2 = DisplayableLinkImpl.valueOf(
                                    dLink, viewRes, lastLink);


                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the nineth test in the first final loop produces: "+dLink_2);

                        linkRights.addDisplayableLink(
                                dLink_2, _upd_m_l_DisplayableLinks);

                    } // if (ep.getIsACSBridgeSource()

                } // for ito

            } // if (ep instanceof Actor)

            // tenth test in the first final loop:
            // if ep is an alias for a reference in the view
            l_temporaryLinks2 = selectDirectReferencesInViewForAlias(
                    _viewInBase, (ACSObject)ep);
            if (l_temporaryLinks2 != null) {

                for (Object obj : l_temporaryLinks2) {
                    for (Iterator ito = l_hiddenSourceLinks_.iterator() ; ito.hasNext() ; ) {

                        dLink = (DisplayableLinkImpl) ito.next();
                        if (dLink.getEorSecondEnd() != ep ) continue;
                        if (dLink.getEorFirstEnd() == ((DisplayableLinkImpl)obj).getEorSecondEnd() )
                            continue;

                        dLink_2 = linkUtil_.mergeLinks(dLink, (DisplayableLinkImpl)obj);

                        // rule 'virtual folder: an indirect access to a Resource that is not
                        // a virtual member, is always through a virtual member'
                        if (linkUtil_.getContainVirtualFolderWithoutMember(dLink_2))
                            continue;

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the tenth test in the first final loop produces: "+dLink_2);


                        linkRights2 = interpreter_.getLinkRightsInstances().get(
                                ((DisplayableLinkImpl)obj).getEorSecondEnd().getName());
                        linkRights2.addDisplayableLink(
                                dLink_2, _upd_m_l_DisplayableLinks);

                    } // for ito        
                } // for obj
            } // if (l_temporaryLinks2 != null)

            // eleventh test in the first final loop:
            // if ep as EligibleParty is an alias reference having an AG context,
            // and an element of this context is in the view, so there is
            // a strong RUN_UNDER relation to add to the current link
            for (Object obj : l_hiddenSourceLinks_) {
                dLink = (DisplayableLinkImpl)obj;
                if (dLink.getEorSecondEnd() != ep ||
                        !dLink.getEorLastLink().getLinkTypeValue(AccessControlLink.IS_ALIAS))
                    continue;

                l_members = linkUtil_.getSecondEndAGcontext(dLink);

                if (l_members == null ||
                        dLink.getProperties() == null ||
                            dLink.getProperties().get(
                                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null)
                                        continue;

                nextIndex = ((Integer)dLink.getProperties().get(
                            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                            intValue();
                // nextIndex has to be the index of the EP as second end
                if (nextIndex != dLink.noOfNodes() - 1)
                    continue;

                for (Object memb : l_members) {
                    if (!_viewInBase.containsBaseObject((GroupIDMember)memb)) continue;
                    if (dLink.getEorFirstEnd() == memb) continue;

                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, (GroupIDMember)memb,
                            new AccessControlLinkImpl());

                    if (memb instanceof GroupID)
                      dLink_2.getEorConnectedLinks((EligibleParty)dLink.getEorSecondEnd())[0].
                              addComment(AccessControlLink.IS_ALIAS, AccessControlLink.C_SET_GROUP);
                    else if (memb instanceof UserID)
                      dLink_2.getEorConnectedLinks((EligibleParty)dLink.getEorSecondEnd())[0].
                              addComment(AccessControlLink.IS_ALIAS, AccessControlLink.C_SET_USER);

                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

                    // set the GLOBAL type if the first end is an actor, and the next node
                    // is a secondary group
                    if (memb instanceof GroupID && dLink.getEorSecondEnd() instanceof Actor &&
                      ((Actor)dLink.getEorSecondEnd()).isMemberToSecondaryGroup((GroupID)memb) ) {
                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink_2.getEorLastLink().addComment(
                                  AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                    } // if

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the eleventh test in the first final loop produces: "+dLink_2);

                    linkRights2 = interpreter_.getLinkRightsInstances().get(
                        ((GroupIDMember)memb).getName());
                    linkRights2.addDisplayableLink(dLink_2, _upd_m_l_DisplayableLinks);
                } // for memb

            } // for obj

        } // for ite on _viewInBase.getEorM_Targets().values() for viewTarg

        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
      }  // for iti on l_hiddenSourceToTest_ for so then ep

                                // for testing and tracking
                                //System.out.println("");
                                //System.out.println("In CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                  //      "before the second final loop on MEMBER and RUN_UNDER, _upd_m_l_DisplayableLinks: ");
                                //ImmutableName key;
                                //for (Iterator ite = _upd_m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                                  // key = (ImmutableName) ite.next();
                                   //System.out.println("--- for key: "+key);
                                   //for (Iterator ita = ((List)_upd_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                                    //System.out.println("------------"+
                                      //((DisplayableLinkImpl)ita.next()).toString() );
                                  //} // for ita
                                //} // for ite

      // SECOND FINAL LOOP ON EACH SELECTED SECOND END AND THE VIEW EPs, including 2 tests:
      // first test:
          // the second end is a GroupIDMember that is member of a view group
      // second test:
          // analyzes the actors that have a proper AG context,
          // that are bridge targets, virtual members,
          // or the actors that are ACL targets or privilege targets...
          // detects if their current Account/Group context replaces
          // the AG context of the previous node, and if it contains a view node
      for (Iterator iti = l_hiddenSourceLinks_.iterator(); iti.hasNext() ; ) {
        dLink = (DisplayableLinkImpl) iti.next();


                          // for testing and tracking
                          //System.out.println(
                            //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                            //" FINAL LOOP on the link in l_hiddenSourceLinks_: "+dLink);

        for (Iterator ite = _viewInBase.getEorM_Sources().values().iterator() ;
          ite.hasNext() ; ) {
            so = (ImmutableSource) ite.next();
            if (so == null || !(so instanceof ImmutableGroupIDMember)) continue;
            viewEp = (EligibleParty)so;;

            if (dLink.getEorFirstEnd() == viewEp) { // viewEP as first node
                continue;
            }

            // first test in the second final loop:
            // the second end is a GroupIDMember that is member of a view group
            if (dLink.getEorSecondEnd() instanceof GroupIDMember &&
                    viewEp instanceof ImmutableGroupID) {
                member = (GroupIDMember)dLink.getEorSecondEnd();

                if (Arrays.asList( ((ImmutableGroupID)viewEp).getEorL_Members()).contains(member)) {
                    dLink_2 = DisplayableLinkImpl.valueOf(
                                dLink,
                                viewEp,
                                new AccessControlLinkImpl());

                    // detects the IS_MEMBER relation
                    dLink = ago_.
                      addOwnerContainRightsToDisplayableLink(dLink_2);

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the first test in the second final loop produces: "+dLink);

                    linkRights2 = interpreter_.getLinkRightsInstances().get(
                                ((ImmutableGroupIDMember)viewEp).getName());
                    linkRights2.addDisplayableLink(
                            dLink, _upd_m_l_DisplayableLinks); 
                } // if (Arrays.asList(
            } // if (dLink.getEorSecondEnd() instanceof GroupIDMember
            
            // second test in the second final loop:
          // analyzes the actors that have a proper AG context,
          // that are bridge targets, virtual members,
          // or the actors that are ACL targets or privilege targets...
          // detects if their current Account/Group context replaces
          // the AG context of the previous node, and if it contains a view node
            if (!(dLink.getEorSecondEnd() instanceof Actor) ) {

                        // for testing and tracking
                        //System.out.println(
                          //"------ in CompoundRightsFactoryImpl."+
                          //"detectHiddenCompoundEpRights(), dLink does not ended by an actor"+
                          //dLink.getEorFirstLink().getLinkTypeComments(
                            //AccessControlLink.GLOBAL));

                    continue;
            } // if not actor

            // defines the Account/Groups context
            l_members = linkUtil_.getSecondEndAGcontext(dLink);

            // change in 0.7.1
            if (l_members == null ||
                    dLink.getProperties() == null ||
                    dLink.getProperties().get(
                            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null ||
                    !l_members.contains(viewEp))
                                continue;
            
            nextIndex = ((Integer)dLink.getProperties().get(
                        DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                        intValue();
            // nextIndex has to be the index of the EP as second end
            if (nextIndex != dLink.noOfNodes() - 1)
                continue;
            
              // detects when the second end is an actor starting from a view GroupIDMmember
            if (viewEp instanceof ImmutableGroupID) {
                    // set the XID comment if the first end is an actor and the next node,
                    // an actor which runs under its current user/group or secondary group
                    rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                    if (rightType != null)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_GROUP);
            } else if (viewEp instanceof ImmutableUserID) {
                    // set the XID comment if the first end is an actor and the next node,
                    // an actor which runs under its current user
                    rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                    if (rightType != null)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_USER);
            } // else if
             
            dLink_2 = DisplayableLinkImpl.valueOf(
                        dLink,
                        viewEp,
                        new AccessControlLinkImpl());

            if (dLink_2 == null)
                continue;

            dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
            // set the GLOBAL type if the first end is an actor, and the next node
            // is a secpndary group
            if (viewEp instanceof GroupID &&
                ((ImmutableActor) dLink.getEorSecondEnd()).isMemberToSecondaryGroup((GroupID)viewEp) ) {
                dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                dLink_2.getEorLastLink().addComment(
                          AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
            } // if

                                      // for testing and tracking
                                      //System.out.println(
                                        //"in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()"+
                                        //" the second test in the second final loop produces: "+dLink_2);

            linkRights2 = interpreter_.getLinkRightsInstances().get(
                            ((ImmutableGroupIDMember)viewEp).getName());
            linkRights2.addDisplayableLink(dLink_2, _upd_m_l_DisplayableLinks);

                        // for testing and tracking
                        //System.out.println(
                          //"------ in CompoundRightsFactoryImpl."+
                          //"detectHiddenCompoundEpRights(), in the last while loop "+
                          //"AgoRightsFactoryImpl.addOwnerContainRightsToDisplayableLink() "+
                          //"returns dLink "+dLink);

        } // for ite on _viewInBase.getEorM_Sources().values() for so then viewEp

        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
      } // for iti on l_hiddenSourceLinks_

                                // for testing and tracking
                                //System.out.println("");
                                //System.out.println("In CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() "+
                                  //      "before the third final loop on priority rules, _upd_m_l_DisplayableLinks: ");
                                //ImmutableName key;
                                //for (Iterator ite = _upd_m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                                  // key = (ImmutableName) ite.next();
                                   //System.out.println("--- for key: "+key);
                                   //for (Iterator ita = ((List)_upd_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                                    //System.out.println("------------"+
                                      //((DisplayableLinkImpl)ita.next()).toString() );
                                  //} // for ita
                                //} // for ite


      // THIRD FINAL LOOP on all DisplayableLinks, with 3 tests for filtering
      // about the priority rules. No removing in l_ObjectToRegister.
      for (Object obj : _upd_m_l_DisplayableLinks.values()) {
          
        for (Object ob : interpreter_.getAllLinkRights((Collection)obj)) {
            linkRights = (LinkRightsImpl)ob;
            
              // applies the rule 'AGO right: A rights overlay G rights'
            linkRights.detectAccountPriorityInAllLinksAGORights((List)obj);

            if (linkRights.getDetectAddonPriorityInAllLinksAGORights())
                linkRights.detectAddonPriorityInAllLinksAGORights((List)obj);
            // applies the ACL and Privilege priority rules, if any
            if (linkRights.getDetectAddonPriorityInAllLinksNonAGORights())
                linkRights.detectAddonPriorityInAllLinksNonAGORights((List)obj);
        } // for ob

        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
      } // for obj

                                // for testing and tracking
                                //System.out.println("In CompoundRightsFactoryImpl."+
                                  //      "detectHiddenCompoundEpRights(), "+
                                    //    "call to GraphicView.setRightsRelatedBaseObjects()");

      _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);

      l_hiddenDirectEP = null;
      l_startOfLinks = null;
      l_temporaryLinks = null;
      l_temporaryLinks2 = null;
      l_temporaryLinks3 = null;
      l_temporaryLinks4 = null;
      l_temporaryLinks5 = null;
      l_groups = null;
      ep = null;
      viewRes = null;
      viewEp = null;
      member = null;
      act = null;
      viewTarg = null;
      dLink = null;
      dLink_2 = null;
      acLink = null;
      lastLink = null;
      l_resources = null;
      so = null;
      vf = null;
      pairKey = null;
      rightType = null;
      l_members = null;
      nodeRights = null;
      nodeRights2 = null;
      linkRights = null;
      linkRights2 = null;
      l_ObjectToRegister = null;
      
      return _upd_m_l_DisplayableLinks;
  } // detectHiddenCompoundEpRights()
  
  /**
  * Defines the direct BRIDGE relations between the view objects.
  * This is a simple method which may be studied to understand the gWork logic.
  * For a sketch view, only the links from/to the _center are processed.
  * Adds the metaright 'execute' to each AccessControlLink with the bridge type.
  * Called by RightsFactory_Facade.detectRights().
  * Synchronized to _viewInBase.
  * @param _viewInBase EPRViewInBase
  * @param _upd_m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks lists,
  *    with BRIDGE links associated to every pair (Actor, Actor).
  * */
  public final Map detectBridges(EPRViewInBase _viewInBase,
    Map _upd_m_l_DisplayableLinks, BaseObject _center) {
      // variables
      Actor                   act;
      Resource                res;
      ImmutableTarget         targ;
      DisplayableLinkImpl     dLink;
      AccessControlLinkImpl   acLink;
      LinkRights              linkRights = null;

      if (!_viewInBase.getConnectedToSeveralACS())
          return _upd_m_l_DisplayableLinks;
      
      for (Iterator ite = _viewInBase.getEorM_Targets().values().iterator() ;
        ite.hasNext() ; ) {
            targ = (ImmutableTarget) ite.next();
            if (targ == null) continue;
            if (!(targ instanceof Actor)) continue;
            act = (Actor)targ;
            if (!act.getIsACSBridgeSource() ) continue;
            if (!_viewInBase.containsSource(act.getEorACSBridgeTarget()))
                continue; // there is no direct bridge relation
            
            if (_center != null && _center != act && _center != act.getEorACSBridgeTarget())
                continue;

            acLink = new AccessControlLinkImpl();

            acLink.addRight(
                AccessControlLink.BRIDGE,
                BaseManagerImpl.getInstance().getMetaRight("execute") );

            dLink = new DisplayableLinkImpl(
                    act, act.getEorACSBridgeTarget(), acLink);

            // updates _upd_m_l_DisplayableLinks
            linkRights = interpreter_.getLinkRightsInstances().get(
                                act.getEorACSBridgeTarget().getName());

            linkRights.addDisplayableLink(dLink, _upd_m_l_DisplayableLinks);

                                // for testing and tracking
                                //System.out.println(
                                  //"------ in CompoundRightsFactoryImpl."+
                                  //"detectBridges(), detection of a new dLink:"+
                                  //dLink);
        
      }  // for (Iterator ite

      res = null;
      act = null;
      dLink = null;
      acLink = null;
      linkRights = null;
      
      return _upd_m_l_DisplayableLinks;
  } // detectBridges()
  
  /**
  * Defines the BRIDGE and ACL relations between all the ACS which are in the view,
  * that are Bridge links between actors, or AclEntries between eligible parties
  * and resources.
  * Called by RightsFactory_Facade.detectRights().
  * Synchronized to _viewInBase.
  * @param _gview is the graphic view associated to the view
  * @param _viewInBase is the view
  * @param _upd_m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @return Map of DisplayableLinks lists,
  *    with BRIDGE and ACL links associated to some pair (ACS, ACS).
  * */
  protected static final Map detectAcsLinks(GraphicView _gview, ViewInBase _viewInBase,
    Map _upd_m_l_DisplayableLinks) {
      // variables
      Map                     m_acs = _viewInBase.getEorM_BaseObjects();
      Actor                   act;
      BaseObject              obj;
      ImmutableEligibleParty  ep;
      Resource                res;
      ACS                     acs;
      ImmutableAclEntry       acl;
      ImmutableAclEntry[]     l_acl;
      DisplayableLinkImpl     dLink;
      AccessControlLinkImpl   acLink;
      List<DisplayableLinkImpl>  l_link;
      Name                    pairKey;
      HashSet                 l_RightsRelatedBaseObjects = new HashSet(INITIAL_CAPACITY);
      
      for (Iterator ite = m_acs.values().iterator() ; ite.hasNext() ; ) {
            obj = (BaseObject) ite.next();
            if (obj == null) continue;
            if (!(obj instanceof ACS) || 
                (!((ACS)obj).getContainsBridgeSource() && !((ACS)obj).getIsExternalOriented() )) {

                                // for testing and tracking
                                //System.out.println(
                                  //"------ in CompoundRightsFactoryImpl."+
                                  //"detectAcsLinks(), detection of an object without external links:"+
                                  //obj.getName());
                    
                      continue;                                
            } // if
            acs = (ACS)obj;

                                // for testing and tracking
                                //System.out.println(
                                  //"------ in CompoundRightsFactoryImpl."+
                                  //"detectAcsLinks(), analysis of the ACS: "+
                                  //acs.getName());
            
            // Bridge analysis
            for (Iterator ito = acs.getEorL_OwnActorBridgeSources().iterator();
              ito.hasNext(); ) {
                
                act = (Actor)ito.next();
                if (!m_acs.containsKey(act.getEorACSBridgeTarget().getAcsName())) continue;
                acLink = new AccessControlLinkImpl();
                // sets default right 'execute'
                acLink.setLinkTypeToTrue(AccessControlLink.BRIDGE);

                pairKey = ((Name)acs.getName()).
                  extended(act.getEorACSBridgeTarget().getAcsName());

                dLink = new DisplayableLinkImpl(acs,
                  act.getEorACSBridgeTarget().getEorACS(), acLink);

                // adds dLink to _upd_m_l_DisplayableLinks, which is a map
                // of DisplayableLinkImpl lists, where the map key
                // is pairKey: the first end name extended by the second end name.
                if (!_upd_m_l_DisplayableLinks.containsKey(pairKey) ) {
                  l_link = new ArrayList(INITIAL_CAPACITY/3);
                  l_link.add(dLink);
                  _upd_m_l_DisplayableLinks.put(pairKey, l_link);
                }
                else {
                  // adds dLink in the current list
                  ((List)_upd_m_l_DisplayableLinks.get(pairKey)).add(dLink);
                } 

                                // for testing and tracking
                                //System.out.println(
                                  //"------ in CompoundRightsFactoryImpl."+
                                  //"detectAcsLinks(), detection of a new dLink from a bridge:"+
                                  //dLink);
                
            } // for ito
            
            // AclEntry analysis from external EP
            for (Iterator ito = acs.getEorM_ExternalEligibleParties().values().iterator();
              ito.hasNext(); ) {
                
                ep = (EligibleParty)ito.next();
                if (!m_acs.containsKey(ep.getAcsName())) continue;
                
                // search if there are positive inter-ACS ACL which is managed by 'acs'
                for (Iterator ita = ep.getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                    res = (Resource)ita.next();
                    // process if the ACL links 2 objects of the same ACS
                    //if (res.getEorACS() == ep.getEorACS()) continue;
                    
                    // noop if the resource ACS is not in the view
                    if (!m_acs.containsKey(res.getAcsName()) ) continue;
                    
                    // call 'FromACS' because the resource has not still be updated
                    acl = AclRightsFactoryImpl.getAclEntryFromACS(res, ep, acs, true);
                    // noop if the ACL is negative
                    if (acl == null) {
                                // for testing and tracking
                                //System.out.println(
                                  //"------ in CompoundRightsFactoryImpl."+
                                  //"detectAcsLinks(), detection of a negative ACL: "+
                                  //acl);                        
                        continue;
                    }

                    acLink = new AccessControlLinkImpl();
                    acLink.setLinkTypeToTrue(AccessControlLink.ACL);
                    
                    if (res.getEorACS() != ep.getEorACS()) {                       
                        pairKey = ((Name)ep.getAcsName()).extended(res.getAcsName());

                        dLink = new DisplayableLinkImpl(ep.getEorACS(),
                          res.getEorACS(), acLink);
                    } else {
                        acLink.addComment(AccessControlLink.ACL,
                                        AccessControlLink.C_CONTROL_OTHER_ACS);
                        pairKey = ((Name)acs.getName()).extended(ep.getAcsName());

                        dLink = new DisplayableLinkImpl(acs,
                          ep.getEorACS(), acLink);                        
                    } // else

                    // adds dLink to _upd_m_l_DisplayableLinks, which is a map
                    // of DisplayableLinkImpl lists, where the map key
                    // is pairKey: the first end name extended by the second end name.
                    if (!_upd_m_l_DisplayableLinks.containsKey(pairKey) ) {
                      l_link = new ArrayList(INITIAL_CAPACITY/3);
                      l_link.add(dLink);
                      _upd_m_l_DisplayableLinks.put(pairKey, l_link);
                    }
                    else {
                      // adds dLink in the current list
                      if( !((List)_upd_m_l_DisplayableLinks.get(pairKey)).contains(dLink) )
                              ((List)_upd_m_l_DisplayableLinks.get(pairKey)).add(dLink);
                    } // else
                    
                                // for testing and tracking
                                //System.out.println(
                                  //"------ in CompoundRightsFactoryImpl."+
                                  //"detectAcsLinks(), detection of a new dLink from an external EP: "+
                                  //dLink);                    
                } // for ita                
            } // for ito
            
            // AclEntry analysis from external resources not detected before,
            // then connected to an 'acs' EP
            for (Iterator ito = acs.getEorM_ExternalResources().values().iterator();
              ito.hasNext(); ) {
                
                res = (Resource)ito.next();
                if (!m_acs.containsKey(res.getAcsName())) continue;
                l_acl = res.getL_aclEntry();
                
                // search if there are positive ACL with an EP from 'acs'
                for (int i = 0; i < l_acl.length; i++) {
                    ep = l_acl[i].getEorEP();
                    // ep is owned by 'acs'; otherwise, the previous ito loop
                    // has detected the ACL
                    if (ep.getEorACS() != acs) continue;
                    if (!l_acl[i].getPositiveRight()) continue;

                    acLink = new AccessControlLinkImpl();
                    acLink.setLinkTypeToTrue(AccessControlLink.ACL);

                    if (res.getEorACS() != ep.getEorACS()) {                       
                        pairKey = ((Name)ep.getAcsName()).extended(res.getAcsName());

                        dLink = new DisplayableLinkImpl(ep.getEorACS(),
                          res.getEorACS(), acLink);
                    } else {
                        acLink.addComment(AccessControlLink.ACL,
                                        AccessControlLink.C_CONTROL_OTHER_ACS);
                        pairKey = ((Name)acs.getName()).extended(ep.getAcsName());

                        dLink = new DisplayableLinkImpl(acs,
                          ep.getEorACS(), acLink);                        
                    } // else

                    // adds dLink to _upd_m_l_DisplayableLinks, which is a map
                    // of DisplayableLinkImpl lists, where the map key
                    // is pairKey: the first end name extended by the second end name.
                    if (!_upd_m_l_DisplayableLinks.containsKey(pairKey) ) {
                      l_link = new ArrayList(INITIAL_CAPACITY/3);
                      l_link.add(dLink);
                      _upd_m_l_DisplayableLinks.put(pairKey, l_link);
                    }
                    else {
                      // adds dLink in the current list
                      if( !((List)_upd_m_l_DisplayableLinks.get(pairKey)).contains(dLink) )
                              ((List)_upd_m_l_DisplayableLinks.get(pairKey)).add(dLink);
                    } // else
                    
                                // for testing and tracking
                                //System.out.println(
                                  //"------ in CompoundRightsFactoryImpl."+
                                  //"detectAcsLinks(), detection of a new dLink from an external resource:"+
                                  //dLink);                    
                } // for ita                
            } // for ito 
                
      }  // for (Iterator ite

      // sets the graphic view with the BaseObjects to listen for their impact on rights.
      // listen all the ACS of the view
      l_RightsRelatedBaseObjects.addAll(m_acs.values());
      _gview.setRightsRelatedBaseObjects(l_RightsRelatedBaseObjects);
      l_RightsRelatedBaseObjects.clear();      

      m_acs = null;
      obj = null;
      act = null;
      dLink = null;
      acLink = null;
      l_link = null;
      pairKey = null;
      l_acl = null;
      res = null;
      ep = null;
      l_RightsRelatedBaseObjects = null;
      return _upd_m_l_DisplayableLinks;
  } // detectAcsLinks()

  /**
  * Creates or sets some new simple negative-rights links (or updates the current simple links),
  * which may be first-rate ones, in various cases. The aim is to prepare a set
  * of DisplayableLinksImpls for the method RightsFactory_Facade.getEffectiveRightsForPair(),
  * where the negative-right types are processed with the positive rights
  * only if the negative-right types link is simple.
  * The followed ACS generic rule is 'this-ACS denyings are superior to this-ACS grantings
  * (there are + and – operations), for each right type'.
  * The algorithm is described hereinafter, as a generic access path search: <P>
  * - if there is a non-simple DisplayableLinkImpl (A to B) having negative right(s)
  * in its last AccessControlLink, a new simple link (A to B) is created with
  * a copy of these negative rights, or, if there is already a simple link,
  * these negative rights are copied into it. <P>
  * - if there is a simple negative-right DisplayableLinkImpl (A to B) while
  * there are an indirect positive-rights or structural paths from A to B
  * through one or several nodes of the view,
  * the link (A to B) is set as 'first-rate' to say that the negative right is a prioritary one,
  * and that A cannot access to B through the other view nodes. <P>
  * If the negative rights of the non-simple DisplayableLinkImpl are from several
  * types, all the negative rights are added. Thus, the resulting simple link has
  * one of the applicable types: AccessControlLink.NEG_TPRI is the prioritary one,
  * or AccessControlLink.NEG_LPRI, and AccessControlLink.NEG_ACL is the last one.
  * This method uses the indicator 'AccessControlLink.C_COPIED_FROM_LINK' when a current
  * simple DisplayableLinkimpl is updated with the negative rights of another DisplayableLinkImpl.
  * This method uses a metaright '|transfer_all_rights|' to indicate a structural link. <P>
  * The results of this method may be changed by an AcsAddon, through the call
  * to the heavy method NodeRights.detectAddonFinalAllNodes() after this method. <P>
  * Called by RightsFactory_Facade.detectEPRRights().
  * Calls RightsFactory_Facade.getFirstShapeEffectiveRights() and
  * linkUtil_.containsIndirectPaths().
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *     associated to the view, and to update.
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
  * */
  public final Map addPathFromNegativeRightsInOneACS(GraphicView _gview,
    EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks) throws InterruptedException {

                                // for testing and tracking
                                //System.out.println("CompoundRightsFactoryImpl."+
                                  //      "addPathFromNegativeRightsInOneACS() start");
                        
    // variables
    Map.Entry                   entry = null;
    List<Name>                  l_negRightsKeys = new ArrayList(INITIAL_CAPACITY);
    List<DisplayableLinkImpl>   l_dLinks = new ArrayList(INITIAL_CAPACITY);
    List<DisplayableLinkImpl>   l_dLinks2 = new ArrayList(INITIAL_CAPACITY);
    Name                        name = null;
    LinkRights                  linkRights = null;
    DisplayableLinkImpl         dLink = null;
    DisplayableLinkImpl         simpleLink = null;
    DisplayableLinkImpl         firstLink = null;
    String                      negType = null;
    HashSet<StringRight>        l_rights = new HashSet(INITIAL_CAPACITY);

    if (Thread.interrupted()) {
        throw new InterruptedException();
    }
    
    // initialize the keys where there is a non-simple negative-right link
    for (Iterator ito = _m_l_DisplayableLinks.entrySet().iterator() ; ito.hasNext() ; ) {

      entry = (Map.Entry)ito.next();
      l_dLinks = (List)entry.getValue();

      // initialize l_negRightsKeys
      for (Iterator ite = l_dLinks.iterator() ; ite.hasNext() ; ) {
            dLink = (DisplayableLinkImpl)ite.next();
            if (!dLink.isSimple() && dLink.getEorLastLink().isTrueForOneNegativeLinkType())
                l_negRightsKeys.add( ((Name)dLink.getEorFirstEnd().getName()).
                        extended(dLink.getEorSecondEnd().getName()) );
      } // for ite
    } // for ito
    
    if (l_negRightsKeys.isEmpty()) {
        l_negRightsKeys = null;
        l_dLinks = null;
        l_dLinks2 = null;
        entry = null;
        l_rights = null;

                                // for testing and tracking
                                //System.out.println("CompoundRightsFactoryImpl."+
                                  //      "addPathFromNegativeRightsInOneACS() "+
                                    //    "no processing of negative rights ");

        return _m_l_DisplayableLinks;
    } // if

                                // for testing and tracking
                                //System.out.println("CompoundRightsFactoryImpl."+
                                  //      "addPathFromNegativeRightsInOneACS() "+
                                    //    "processes negative rights ");

    for (Iterator ita = l_negRightsKeys.iterator(); ita.hasNext(); ) {
        name = (Name)ita.next();
        l_dLinks = (List)_m_l_DisplayableLinks.get(name);
        firstLink = l_dLinks.iterator().next();

        simpleLink = null;
        negType = null;
        l_rights.clear();

                                // for testing and tracking
                                //System.out.println("------ for "+name);

        for (DisplayableLinkImpl link : l_dLinks) {

            if (link.isSimple()) {
                simpleLink = link;
            }

            if (link.getEorLastLink().isTrueForOneNegativeLinkType()) {
                if (link.getEorLastLink().getLinkTypeValue(AccessControlLink.NEG_ACL)) {
                    negType = AccessControlLink.NEG_ACL;
                    l_rights.addAll(
                            Arrays.asList((StringRight[])link.getEorLastLink().
                                getRights(AccessControlLink.NEG_ACL)));
                }
                if (link.getEorLastLink().getLinkTypeValue(AccessControlLink.NEG_LPRI)) {
                    negType = AccessControlLink.NEG_LPRI;
                    l_rights.addAll(
                            Arrays.asList((StringRight[])link.getEorLastLink().
                                getRights(AccessControlLink.NEG_LPRI)));
                }
                if (link.getEorLastLink().getLinkTypeValue(AccessControlLink.NEG_TPRI)) {
                    negType = AccessControlLink.NEG_TPRI;
                    l_rights.addAll(
                            Arrays.asList((StringRight[])link.getEorLastLink().
                                getRights(AccessControlLink.NEG_TPRI)));
                }
            } // if

            if (simpleLink != null && negType != null)
                break;
        } // for link

        // 'negType' is not null
        if (simpleLink != null) {
            simpleLink.getEorLastLink().setRights(
                        negType,
                        ((Set<StringRight>)l_rights.clone()).toArray(new StringRight[0]));
            simpleLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
            simpleLink.getEorLastLink().addComment(
                    AccessControlLink.GLOBAL, AccessControlLink.C_COPIED_FROM_LINK);
        }
        else {
            simpleLink = new DisplayableLinkImpl(
                    firstLink.getEorFirstEnd(), firstLink.getEorSecondEnd(),
                    new AccessControlLinkImpl());
            simpleLink.getEorLastLink().setRights(
                    negType,
                    ((Set<StringRight>)l_rights.clone()).toArray(new StringRight[0]));
        } // else if (simpleLink != null)


        linkRights = interpreter_.getLinkRightsInstances().get(
                    simpleLink.getEorSecondEnd().getName());
        simpleLink = linkRights.addDisplayableLink(
                simpleLink, _m_l_DisplayableLinks);

        l_dLinks2.add(simpleLink);


                                // for testing and tracking
                                //System.out.println("CompoundRightsFactoryImpl."+
                                  //      "addPathFromNegativeRightsInOneACS() "+
                                    //    "adds the negative-rights simple link: "+simpleLink);
    } // for ita

    if (l_dLinks2.isEmpty()) {

        l_negRightsKeys = null;
        l_dLinks = null;
        l_dLinks2 = null;
        entry = null;
        name = null;
        linkRights = null;
        simpleLink = null;
        firstLink = null;
        negType = null;
        l_rights = null;

        return _m_l_DisplayableLinks;
    } // if

    // m_rights after the adding of the negative-rights simple links
    Map<ImmutableName, Set<StringRight>> m_rights =
            ((RightsFactory_Facade)interpreter_.getCoreAlgorithm("Facade")).
                getFirstShapeEffectiveRights(_m_l_DisplayableLinks);

    // add the structural links where getFirstShapeEffectiveRights()
    // has returned an empty list, for linkUtil_getCurrentEndKeys()
    // called by containsIndirectPaths().
    for (Set<StringRight> m_r : m_rights.values()) {
        if (m_r.isEmpty()) {
            m_r.add(BaseManagerImpl.getInstance().getMetaRight("|transfer_all_rights|"));
        }
    } // for m_r

    // when some indirect paths have positive rights which modify
    // the direct negative-rights simple paths, these latter links
    // are set as 'first-rate' links
    l_dLinks = linkUtil_.containsIndirectPaths(l_dLinks2, m_rights, true);

    if (l_dLinks == null) {
        m_rights = null;
        l_negRightsKeys = null;
        l_dLinks2 = null;
        entry = null;
        name = null;
        linkRights = null;
        simpleLink = null;
        firstLink = null;
        negType = null;
        l_rights = null;

        return _m_l_DisplayableLinks;
    } // if

    for (DisplayableLinkImpl link : l_dLinks) {
        link.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
        link.getEorLastLink().addComment(
                 AccessControlLink.GLOBAL, AccessControlLink.C_FIRST_RATE);

                                // for testing and tracking
                                //System.out.println("CompoundRightsFactoryImpl."+
                                  //      "addPathFromNegativeRightsInOneACS() "+
                                    //    "sets link as 'first-rate' link: "+link);
    } // for link
      
    m_rights = null;
    l_negRightsKeys = null;
    l_dLinks = null;
    l_dLinks2 = null;
    entry = null;
    name = null;
    linkRights = null;
    simpleLink = null;
    firstLink = null;
    negType = null;
    l_rights = null;

    return _m_l_DisplayableLinks;
  } // addPathFromNegativeRightsInOneACS()
  
  /**
  * Defines the direct links from a VirtualFolder to its members,
  * when all the objects are in the view, even if the member ACS does not manage
  * the VirtualFolder rights propagation to members.
  * Works also for a sketch view, where the argument _center is not null.
  * Called by RightsFactory_Facade.detectEPRRights() and detectSketchRights().
  * Synchronized to _viewInBase from the GraphicView call.
  * @param _viewInBase the EPRViewInBase to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks lists,
  *   with HAS_VIRTUAL_MEMBER links associated to every pair (VirtualFolder, member in the view).
  * */
  protected final Map detectVirtualMembers(
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks, BaseObject _center) {
      // variables
      BaseObject              bo;
      ImmutableBasic          bas;
      Collection              viewObjects = _viewInBase.getEorM_BaseObjects().values();
      DisplayableLinkImpl     dLink;
      LinkRights              linkRights = null;

                    // for testing and tracking
                    //System.out.println(
                      //      "start of CompoundRightsFactoryImpl.detectVirtualMembers()");

      // process the link virtual folder/member
      for (Iterator ite = viewObjects.iterator(); ite.hasNext() ; ) {
        bo = (BaseObject) ite.next();

        if (!(bo instanceof ImmutableBasic)) continue;
        bas = (ImmutableBasic) bo;
        //if (!bas.getEorACS().getManageVirtualFolderRightsPropagationToMembers()) {

                    // for testing and tracking
                    //System.out.println(
                      //      "CompoundRightsFactoryImpl.detectVirtualMembers() with "+
                        //    "getManageVirtualFolderRightsPropagationToMembers() false for "+
                          //  bas.getEorACS().getName());
            
            //continue;
        //} // if

        for (Object obj : bas.getVirtualFolders()) {

            if (!viewObjects.contains(obj)) continue;
            // for a sketch view
            if (_center != null && _center != obj && _center != bas) continue;

                    // for testing and tracking
                    //System.out.println("CompoundRightsFactoryImpl.detectVirtualMembers(), "+
                      //"call to linkUtil_.getOrCreateDisplayableLink() for "+
                      //bas.getNickName()+" and "+((ACSObject)obj).getNickName());

            dLink = linkUtil_.getDisplayableLink(
                  (BaseObject)obj, bas, null, _m_l_DisplayableLinks);

            if (dLink == null) {
              dLink = new DisplayableLinkImpl(
                      (BaseObject)obj, bas, (BaseObject[])null);
              dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.HAS_VIRTUAL_MEMBER);

              linkRights = interpreter_.getLinkRightsInstances().get(
                                bas.getName());
              dLink = linkRights.addDisplayableLink(dLink, _m_l_DisplayableLinks);
            } // dLink == null
            else {
                dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.HAS_VIRTUAL_MEMBER);
            }
              
        } // for obj
      } // for (Iterator ite
      
      bo = null;
      bas = null;
      viewObjects = null;
      dLink = null;
      linkRights = null;

                    // for testing and tracking
                    //System.out.println("\n At the end of CompoundRightsFactoryImpl.detectVirtualMembers(), "+
                      //"for the view "+_viewInBase.getNickName()+
                      //", the DisplayableLinkImpls in the result are:");
                    //for (Iterator ite = _m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {
                      // for (Iterator ita = ((List)ite.next()).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      return _m_l_DisplayableLinks;
  } // detectVirtualMembers()
  
  /**
  * Defines the simple links from an ACSObject as alias to its reference,
  * when all the objects are in the view. The alias is the source, and it may be
  * an external soft alias or an internal hard alias.
  * Unused for a sketch view, where the argument _center is not null, since
  * a sketch view does not display the soft alias relations.
  * _center is analyzed as alias, not as reference.
  * Called by RightsFactory_Facade.detectEPRRights() and detectNoThanRights().
  * Synchronized to _viewInBase from the GraphicView call.
  * Calls selectDirectReferencesInViewForAlias().
  * @param _viewInBase the EPRViewInBase to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks lists,
  *   with IS_ALIAS links associated to every pair (ACSObject alias, reference in the view).
  * */
  protected final Map detectReferencesForAliases(
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks, BaseObject _center) {
      // variables
      BaseObject              bo;
      Collection              viewObjects = _viewInBase.getEorM_BaseObjects().values();
      Set<DisplayableLinkImpl> l_dLinks = null;
      LinkRights              linkRights = null;

      for (Iterator ite = viewObjects.iterator(); ite.hasNext() ; ) {
        bo = (BaseObject) ite.next();

                    // for testing and tracking
                    //System.out.println("\n In CompoundRightsFactoryImpl.detectReferencesForAlias(), "+
                      //"process the BaseObject "+bo.getName());

        if (!(bo instanceof ACSObject) || _center != null && _center != bo)
            continue;

        l_dLinks = selectDirectReferencesInViewForAlias(_viewInBase, (ACSObject)bo);
        if (l_dLinks == null) continue;

        for (Object link : l_dLinks) {
            linkRights = interpreter_.getLinkRightsInstances().get(
                                ((DisplayableLinkImpl)link).getEorSecondEnd().getName());
            linkRights.addDisplayableLink((DisplayableLinkImpl)link, _m_l_DisplayableLinks);
        } // for link
      } // for ite
      
      bo = null;
      viewObjects = null;
      linkRights = null;

                    // for testing and tracking
                    //System.out.println("In CompoundRightsFactoryImpl.detectReferencesForAlias(), "+
                      //"for the view "+_viewInBase.getNickName()+
                      //", the DisplayableLinkImpls in the result are:");
                    //for (Iterator ite = _m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {
                      // for (Iterator ita = ((List)ite.next()).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      return _m_l_DisplayableLinks;
  } // detectReferencesForAliases()
  
  /**
  * Defines the links from an ACSObject as alias to its hard reference,
  * when the objects are in the view. The alias is the source, and it is not
  * considered as an external soft alias.
  * Works only for a sketch view, where the argument _center is not null.
  * _center is analyzed as alias, not as reference.
  * Called by RightsFactory_Facade.detectSketchRights().
  * Synchronized to _viewInBase from the GraphicView call.
  * Calls selectDirectHardReferenceInViewForAlias().
  * @param _viewInBase the EPRViewInBase to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks lists,
  *   with IS_ALIAS links associated to every pair (ACSObject alias, reference in the view).
  * */
  protected final Map detectHardReferencesForAliases(
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks, BaseObject _center) {
      // variables
      BaseObject              bo;
      Collection              viewObjects = _viewInBase.getEorM_BaseObjects().values();
      DisplayableLinkImpl     link = null;
      LinkRights              linkRights = null;

      for (Iterator ite = viewObjects.iterator(); ite.hasNext() ; ) {
        bo = (BaseObject) ite.next();

        if (!(bo instanceof ACSObject) || _center != null && _center != bo)
            continue;

        link = selectDirectHardReferenceInViewForAlias(_viewInBase, (ACSObject)bo);
        if (link != null) {
            linkRights = interpreter_.getLinkRightsInstances().get(
                                link.getEorSecondEnd().getName());
            linkRights.addDisplayableLink(link, _m_l_DisplayableLinks);
        } // if

      } // for ite
      
      bo = null;
      viewObjects = null;
      link = null;
      linkRights = null;

                    // for testing and tracking
                    //System.out.println("\n In CompoundRightsFactoryImpl.detectHardReferencesForAlias(), "+
                      //"for the view "+_viewInBase.getNickName()+
                      //", the DisplayableLinkImpls in the result are:");
                    //for (Iterator ite = _m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {
                      // for (Iterator ita = ((List)ite.next()).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      return _m_l_DisplayableLinks;
  } // detectHardReferencesForAliases()
  
  /**
  * Defines the links from an ACSObject as reference for its hard alias,
  * when the objects are in the view. The alias is the source, and it is not
  * considered as an external soft alias.
  * Used only for a sketch view, where the argument _center is not null.
  * _center is analyzed as reference, not as alias.
  * Called by RightsFactory_Facade.detectSketchRights().
  * Synchronized to _viewInBase from the GraphicView call.
  * Calls selectDirectHardReferenceInViewForAlias().
  * @param _viewInBase the EPRViewInBase to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks lists,
  *   with IS_ALIAS links associated to every pair (ACSObject alias, reference in the view).
  * */
  protected final Map detectHardAliasesForReferences(
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks, BaseObject _center) {
      // variables
      BaseObject              bo;
      Collection              viewObjects = _viewInBase.getEorM_BaseObjects().values();
      DisplayableLinkImpl     link = null;
      LinkRights              linkRights = null;

      for (Iterator ite = viewObjects.iterator(); ite.hasNext() ; ) {
        bo = (BaseObject) ite.next();

        if (!(bo instanceof ACSObject) || _center != null && _center != bo)
            continue;

        link = selectDirectHardAliasInViewForReference(_viewInBase, (ACSObject)bo);
        if (link != null) {
            linkRights = interpreter_.getLinkRightsInstances().get(
                                link.getEorSecondEnd().getName());
            linkRights.addDisplayableLink(link, _m_l_DisplayableLinks);
        } // if

      } // for ite
      
      bo = null;
      viewObjects = null;
      link = null;
      linkRights = null;

                    // for testing and tracking
                    //System.out.println("\n In CompoundRightsFactoryImpl.detectHardReferencesForAlias(), "+
                      //"for the view "+_viewInBase.getNickName()+
                      //", the DisplayableLinkImpls in the result are:");
                    //for (Iterator ite = _m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {
                      // for (Iterator ita = ((List)ite.next()).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      return _m_l_DisplayableLinks;
  } // detectHardAliasesForReferences()
  
  /**
  * Defines the links from an ACSObject as alias to its references
  * when the references are in the view. The alias is the source, and it may be
  * an external or internal soft alias, or an internal hard alias.
  * This method adds to the type comment AccessControlLink.ALIAS
  * the text AccessControlLink.C_RECEIVE_ALIAS_RIGHTS. <P>
  * Calls ACS.getReferenceOfSoftAlias() which may be a heavy processing load
  * in some ACS.
  * Called by detectHiddenCompoundEpRights(), detectReferencesForAliases() and
  * ThreeNodesRightsFactoryImpl.endsPathsFromActorWithAclBridgePrivilege(),
  * endsPathsFromVirtualFolderWithMemberAlias(),
  * selectOneHiddenNodePathsForVirtualFolder(),
  * selectOneHiddenNodePathsForFirstLinkPrivileges().
  * Synchronized to _viewInBase from the GraphicView call.
  * @param _viewInBase the EPRViewInBase
  * @param _alias may be an alias or not. Never null.
  * @return Set of simple DisplayableLinkImpl for which the first end is _alias,
  *   the second end is an _alias reference in _viewInBase.
  *   May be null but not empty.
  * */
  protected final Set<DisplayableLinkImpl> selectDirectReferencesInViewForAlias(
          EPRViewInBase _viewInBase, ACSObject _alias) {
    // variables
    Set<DisplayableLinkImpl>  l_result = null;
    DisplayableLinkImpl     dLink;
    ACSObject               ob2;
    ImmutableACS            acs;
    ImmutableName[]         l_acsNames = null;

    if (_alias instanceof ImmutableSource) {
        l_acsNames = ((ImmutableSource)_alias).getL_AcsOfSoftReferences();
    }
    else if (_alias instanceof ImmutableTarget) {
        l_acsNames = ((ImmutableTarget)_alias).getL_AcsOfSoftReferences();
    }

    if (l_acsNames != null) {
        for (Object o : _viewInBase.getEorL_ConnectedACSNames()) {
            acs = (ImmutableACS)utility_.getBaseObject((ImmutableName)o);
            if (!acs.getManageSoftAlias())
                continue;
            if (!Arrays.asList(l_acsNames).contains(acs.getName()))
                continue;

            // process the link external soft alias
            ob2 = acs.getReferenceOfSoftAlias(_alias);

            if (ob2 == null || !_viewInBase.containsBaseObject(ob2))
                continue;

            dLink = new DisplayableLinkImpl(_alias, ob2, new AccessControlLinkImpl());
            dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_ALIAS);
            dLink.getEorFirstLink().addComment(
                    AccessControlLink.IS_ALIAS, AccessControlLink.C_RECEIVE_ALIAS_RIGHTS);

            if (l_result == null)
                l_result = new HashSet(INITIAL_CAPACITY / 3);
            l_result.add(dLink);
        } // for o
    } // if (l_acsNames != null)

    dLink = selectDirectHardReferenceInViewForAlias(_viewInBase, _alias);
    if (dLink != null) {
        if (l_result == null)
                l_result = new HashSet(INITIAL_CAPACITY / 3);
        l_result.add(dLink);
    } // if
      
    dLink = null;
    ob2 = null;
    acs = null;
    l_acsNames = null;

                    // for testing and tracking
                    //System.out.println("\n In CompoundRightsFactoryImpl."+
                      //      "selectDirectReferencesInViewForAlias(), "+
                      //"for the view "+_viewInBase.getNickName()+
                      //", the DisplayableLinkImpls in the result are:");
                    //if (l_result != null) {
                      //   for (Iterator ita = l_result.iterator() ; ita.hasNext() ; ) {
                        //    System.out.println("------------"+
                          //    ((DisplayableLinkImpl)ita.next()).toString() );
                         //} // for ita
                    //} // if
                    //else System.out.println("----- null");

    return l_result;
  } // selectDirectReferencesInViewForAlias()
  
  /**
  * Defines the link from an ACSObject as hard alias to its reference
  * when the reference is in the view. The alias is the source, and it is not
  * analyzed as a soft alias.
  * Adds the comment AccessControlLink.C_RECEIVE_ALIAS_RIGHTS or, if the argument
  * is an Actor running under a proper AG context, adds the user name of this context. <P>
  * Called by selectDirectReferencesInViewForAlias() and detectHardReferencesForAliases().
  * Synchronized to _viewInBase from the GraphicView call.
  * @param _viewInBase the EPRViewInBase
  * @param _alias may be an alias or not. Never null.
  * @return simple DisplayableLinkImpl for which the first end is _alias,
  *   the second end is the _alias hard reference in _viewInBase.
  *   May be null.
  * */
  protected final DisplayableLinkImpl selectDirectHardReferenceInViewForAlias(
          EPRViewInBase _viewInBase, ACSObject _alias) {
    // variables
    DisplayableLinkImpl     result = null;
    ACSObject               ob2 = null;

    if ((_alias instanceof ImmutableBasic || _alias instanceof ImmutableVirtualFolder) &&
            _alias.getEorACS().getManageInternalHardAlias()) {

            // process the link internal hard alias
        if (_alias instanceof ImmutableSource)
            ob2 = ((ImmutableSource)_alias).getEorReferenceAsInternalHardAlias();
        else if (_alias instanceof ImmutableTarget)
            ob2 = ((ImmutableTarget)_alias).getEorReferenceAsInternalHardAlias();

        if (ob2 != null && _viewInBase.containsBaseObject(ob2)) {

            result = new DisplayableLinkImpl(_alias, ob2, new AccessControlLinkImpl());
            result.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_ALIAS);

            if (!(_alias instanceof Actor) || ((Actor)_alias).getEorCurrentUserID() == null) {
                result.getEorFirstLink().addComment(
                    AccessControlLink.IS_ALIAS, AccessControlLink.C_RECEIVE_ALIAS_RIGHTS);
            }
            else {
                if (((Actor)_alias).getEorCurrentGroupID() != null)
                    result.getEorFirstLink().addComment(
                        AccessControlLink.IS_ALIAS, AccessControlLink.C_MAY_SET_USER_GROUP +
                            ((Actor)_alias).getEorCurrentUserID().getNickName() +
                            "/" + ((Actor)_alias).getEorCurrentGroupID().getNickName());
                else
                    result.getEorFirstLink().addComment(
                        AccessControlLink.IS_ALIAS, AccessControlLink.C_MAY_SET_USER_GROUP +
                            ((Actor)_alias).getEorCurrentUserID().getNickName() +
                            "/NULL");
            } // else

        } // if
    } // if
      
    ob2 = null;

                                // for testing and tracking
                                //System.out.println("\n In CompoundRightsFactoryImpl."+
                                  //      "selectDirectHardReferenceInViewForAlias(), "+
                                  //"for the alias "+_alias.getNickName()+":");
                                //System.out.println("--- the result is: "+result);

    return result;
  } // selectDirectHardReferenceInViewForAlias()
  
  /**
  * Defines the link from an ACSObject as hard reference to its alias
  * when the alias is in the view. The alias is the source, and it is not
  * analyzed as an external soft alias.<P>
  * Called by detectHardAliasesForReferences().
  * Synchronized to _viewInBase from the GraphicView call.
  * @param _viewInBase the EPRViewInBase
  * @param _ref may be an alias reference or not. Never null.
  * @return simple DisplayableLinkImpl for which the first end is the alias,
  *   the second end is the _ref in _viewInBase.
  *   May be null.
  * */
  protected final DisplayableLinkImpl selectDirectHardAliasInViewForReference(
          EPRViewInBase _viewInBase, ACSObject _ref) {
    // variables
    DisplayableLinkImpl     result = null;
    ACSObject               ob2 = null;

    if ((_ref instanceof ImmutableBasic || _ref instanceof ImmutableVirtualFolder) &&
            _ref.getEorACS().getManageInternalHardAlias()) {

            // process the link internal hard alias
        if (_ref instanceof ImmutableSource)
            ob2 = ((ImmutableSource)_ref).getEorAliasAsInternalHardReference();
        else if (_ref instanceof ImmutableTarget)
            ob2 = ((ImmutableTarget)_ref).getEorAliasAsInternalHardReference();

        if (ob2 != null && _viewInBase.containsBaseObject(ob2)) {

            result = new DisplayableLinkImpl(ob2, _ref, new AccessControlLinkImpl());
            result.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_ALIAS);

        } // if
    } // if
      
    ob2 = null;

                    // for testing and tracking
                    //System.out.println("\n In CompoundRightsFactoryImpl."+
                      //      "selectDirectHardaliasInViewForReference(), "+
                      //"for the view "+_viewInBase.getNickName()+
                      //", the DisplayableLinkImpl in the result is: "+result);

    return result;
  } // selectDirectHardAliasInViewForReference()

  /**
  * Detects the paths from a VirtualFolder to its members, considering the paths
  * only if the member ACS manages the VirtualFolder rights propagation to the members.
  * In that case, the VirtualFolder transfers directly its rights as target to its members.
  * Nonetheless, this is done in the caller of this method, while this method
  * simply returns the good links. The link types are set.
  * The members which are not EligibleParties may be returned.
  * If only one member in the VirtualFolder has an ACS that manages the VirtualFolder
  * rights propagation, then the result contains one link to this member. <P>
  * Trees of virtual folders are possible. There is no privilege inheritance in the tree
  * for the generic gBase classes, while an AcsAddon may override the hook methods. <P>
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() and
  * ThreeNodesRightsFacoryImpl.selectOneHiddenNodePathsForVirtualFolder().
  * @param _gview is the graphic view to update
  * @param _viewInBase view on which _gview is based
  * @param _vf is in _viewInBase or not
  * @return Set of DisplayableLinkImpl for which the node before the second end is _vf,
  *   the second end is a _vf member or a GroupIDMember outside _viewInBase.
  *   May be null but not empty.
  * */
  public final Set<DisplayableLinkImpl> selectDirectHiddenMembersForVirtualFolder(
        GraphicView _gview, EPRViewInBase _viewInBase, ImmutableVirtualFolder _vf) {

    ImmutableResource     res;
    Set<DisplayableLinkImpl>  l_result = null;
    DisplayableLinkImpl   dLink = null;
    HashSet               l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();
    boolean               registering = false;

    for (Iterator iti = _vf.getEorM_BaseObjects().values().iterator();
      iti.hasNext() ; ) {
          res = (ImmutableResource) iti.next();

          if (_viewInBase.containsTarget(res)) continue;
          if (!res.getEorACS().getManageVirtualFolderRightsPropagationToMembers())
              continue;

          dLink = new DisplayableLinkImpl(_vf, res, new AccessControlLinkImpl());
          dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.HAS_VIRTUAL_MEMBER);
          if (l_result == null)
              l_result = new HashSet(INITIAL_CAPACITY / 3);
          l_result.add(dLink);
          l_ObjectToRegister.add(res);
          registering = true;

                            // for testing and tracking
                            //System.out.println("in CompoundRightsFactoryImpl."+
                              //"selectVirtualMembersForVirtualFolder(), add the simple link for virtual members:");
                            //System.out.println("-----------"+dLink);
    } // for iti

    if (registering)
        _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);

    res = null;
    dLink = null;
    l_ObjectToRegister = null;
    return l_result;
  } // selectDirectHiddenMembersForVirtualFolder()

  /**
  * Detects the references of an alias that are not in _viewInBase.
  * The link types are set. Several references are possible, since
  * there may be external or internal soft aliases, and one internal hard alias.
  * This method searches the soft references of _ali by calling getReferenceOfSoftAlias()
  * on each open ACS managing the soft aliasing for _ali, so it should be time-consuming. <P>
  * Generally speaking, the hard alias transfers directly its rights as target
  * to its reference, including its Account/Groups context. If the elements
  * of the AG context are not yet in the path, this may define new links,
  * from the reference to each account or group of the context.
  * This is called the propagation of the indirect AG context, and the alias relation
  * is alone to handle that in Access Road. There is no such a propagation in
  * an AclEntry relation, even if it transfers the direct AG context.
  * The case occurs when an Actor is an alias running as actor under an account, if,
  * in the path, the nodes before this actor do not transmit their proper AG context.
  * The responsability of this method is not to set the good AG context on the reference.
  * It does not know the current paths having _ali as second end. But it is the responsibility
  * of the caller of this method to handle properly the reference AG context
  * (see the ARoad0.gBase.CAliasImpl documentation). <P>
  * In this version, the visible space of the _ali ACS is not considered
  * in the search for the external soft aliases. All the open ACS are considered.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() and
  * ThreeNodesRightsFacoryImpl.selectOneHiddenNodePathsForVirtualFolder(),
  * addPathsFromActorStartsUnderGroupIDMember(),
  * addPathsFromActorAliasReference() and
  * addPathsFromGroupIDMemberAliasReference().
  * @see ARoad0.gBase.CAliasImpl
  * @param _gview is the graphic view to update
  * @param _viewInBase view on which _gview is based
  * @param _ali is in _viewInBase or not
  * @return Set of simple DisplayableLinkImpl for which the first end is _ali,
  *   the second end is an _ali reference outside _viewInBase.
  *   May be null but not empty.
  * */
  public final Set<DisplayableLinkImpl> selectDirectHiddenReferenceForAlias(
        GraphicView _gview, EPRViewInBase _viewInBase, ACSObject _ali) {

    Set<DisplayableLinkImpl>    l_result = null;
    DisplayableLinkImpl         dLink = null;
    HashSet                     l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();
    ImmutableACS                acs;
    ACSObject                   ob2 = null;
    boolean                     registering = false;
    ImmutableName[]             l_acsNames = null;

    if (_ali instanceof ImmutableSource) {
        l_acsNames = ((ImmutableSource)_ali).getL_AcsOfSoftReferences();
    }
    else if (_ali instanceof ImmutableTarget) {
        l_acsNames = ((ImmutableTarget)_ali).getL_AcsOfSoftReferences();
    }

    if (l_acsNames != null) {
        for (Object o : BaseManagerImpl.getInstance().getEorM_OpenACS().values()) {
            acs = (ImmutableACS)o;
            if (!acs.getManageSoftAlias())
                continue;
            if (!Arrays.asList(l_acsNames).contains(acs.getName()))
                continue;

            // process the link soft alias
            ob2 = acs.getReferenceOfSoftAlias(_ali);

            if (ob2 == null || _viewInBase.containsBaseObject(ob2))
                continue;

            dLink = new DisplayableLinkImpl(_ali, ob2, new AccessControlLinkImpl());
            dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_ALIAS);
            dLink.getEorFirstLink().addComment(
                        AccessControlLink.IS_ALIAS, AccessControlLink.C_RECEIVE_ALIAS_RIGHTS);

            if (l_result == null)
                l_result = new HashSet(INITIAL_CAPACITY / 3);
            l_result.add(dLink);
            l_ObjectToRegister.add(ob2);
            registering = true;
        } // for o
    } // if (l_acsNames != null)

    if (_ali.getEorACS().getManageInternalHardAlias()) {

        ob2 = null;
            // process the link internal hard alias
        if (_ali instanceof ImmutableSource)
            ob2 = ((ImmutableSource)_ali).getEorReferenceAsInternalHardAlias();
        else if (_ali instanceof ImmutableTarget)
            ob2 = ((ImmutableTarget)_ali).getEorReferenceAsInternalHardAlias();

        if (ob2 != null && !_viewInBase.containsBaseObject(ob2)) {
            dLink = new DisplayableLinkImpl(_ali, ob2, new AccessControlLinkImpl());
            dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_ALIAS);
            dLink.getEorFirstLink().addComment(
                    AccessControlLink.IS_ALIAS, AccessControlLink.C_RECEIVE_ALIAS_RIGHTS);

            if (l_result == null)
                l_result = new HashSet(INITIAL_CAPACITY / 3);
            l_result.add(dLink);
            l_ObjectToRegister.add(ob2);
            registering = true;
        } // if
    } // if (_al.getEorACS().getManageInternalHardAlias())

    if (registering)
        _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);

    ob2 = null;
    acs = null;
    dLink = null;
    l_ObjectToRegister = null;
    return l_result;
  } // selectDirectHiddenReferenceForAlias()

  /**
  * Adds a SOFT_RUN_UNDER relation between two view objects of the same ACS.
  * If the second end of a link is a process (indeed, an Actor or a VirtualFolder)
  * having an AG context so that its runs under a view GroupIDMember, this method
  * adds a simple SOFT_RUN_UNDER link in the map of DisplayableLinks.
  * Typically, the second end may be an internal hard alias.
  * If its reference runs under a view account, there is a simple SOFT_RUN_UNDER link
  * to create from the reference to the account.
  * If the link delivers an executing right to the first end GroupIDMember
  * on the second-end process, this method creates a new link from the process
  * to the GroupIDMember.<P>
  * This method has to be called at the end of the access paths search.
  * If the two view objects are not from the same ACS, or if a simple link
  * for the same ends does already exist, this method does not operate.
  * Called only by RightsFactory_Facade.detectEPRRights().
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase
  * @param _upd_m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair).
  * Never null. May be updated by new simple links.
  * @return updated _upd_m_l_DisplayableLinks
  **/
  public final Map detectLastSimpleSoftRunUnderDisplayableLinks(GraphicView _gview,
    EPRViewInBase _viewInBase, Map _upd_m_l_DisplayableLinks) {
      // variables
      List<DisplayableLinkImpl> l_dLinks;
      DisplayableLinkImpl       dLink;
      Set<ImmutableGroupIDMember>   l_members = null;
      List<DisplayableLinkImpl> l_new_dLinks = null;
      LinkRights                linkRights = null;
      int                       nextIndex = -1;

      for (Iterator ite = _upd_m_l_DisplayableLinks.values().iterator() ; ite.hasNext() ; ) {
          l_dLinks = (List<DisplayableLinkImpl>) ite.next();

          for (DisplayableLinkImpl link : l_dLinks) {

            if (!(link.getEorSecondEnd() instanceof Actor) &&
                    !(link.getEorSecondEnd() instanceof VirtualFolder))
                                        continue;

            // no test of (link.getProperties() == null), because a 2-nodes link
            // may have no AG context

            // l_members is the second end AG context
            l_members = linkUtil_.getSecondEndAGcontext(link);

            if (l_members == null ||
                    link.getProperties() == null ||
                    link.getProperties().get(
                        DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null) {

                                    // for testing and tracking
                                    //if (((ACSObject)link.getEorFirstEnd()).
                                      //      getNickName().equals("sys"))
                                        //System.out.println("\n CompoundRightsFactoryImpl."+
                                          //"detectLastSimpleSoftRunUnderDisplayableLinks(), "+
                                          //"no member for the link: "+link);

                continue;
            } // if

            // new in 0.7.1
            nextIndex = ((Integer)link.getProperties().get(
                        DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                        intValue();
            // nextIndex has to be the index of the 'link' second end
            if (nextIndex != link.noOfNodes() - 1)
                continue;

                                    // for testing and tracking
                                    //if (((ACSObject)link.getEorFirstEnd()).
                                      //      getNickName().equals("sys"))
                                        //System.out.println("\n CompoundRightsFactoryImpl."+
                                          //"detectLastSimpleSoftRunUnderDisplayableLinks(), "+
                                          //"members to analyze for the link: "+link);

            for (Object memb : l_members) {
                if (!_viewInBase.containsBaseObject((GroupIDMember)memb) ||
                        link.getEorSecondEnd() == memb ||
                        ((ACSObject)link.getEorSecondEnd()).getEorACS() !=
                            ((GroupIDMember)memb).getEorACS())
                                continue;

                                    // for testing and tracking
                                    //if (((ACSObject)link.getEorFirstEnd()).
                                      //      getNickName().equals("sys"))
                                        //System.out.println("CompoundRightsFactoryImpl."+
                                      //"detectLastSimpleSoftRunUnderDisplayableLinks(), "+
                                      //"process for the member "+memb+", the link: "+link);


                // no creation if there is already a RUN_UNDER type in the simple link
                dLink = linkUtil_.getDisplayableLink(
                            link.getEorSecondEnd(), (GroupIDMember)memb,
                            null, _upd_m_l_DisplayableLinks);

                if (dLink != null &&
                        dLink.getEorLastLink().getLinkTypeValue(AccessControlLink.RUN_UNDER))
                            continue;

                dLink = new DisplayableLinkImpl(
                        // this is a view object
                        link.getEorSecondEnd(),
                        (GroupIDMember)memb,
                        new AccessControlLinkImpl());

                dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.SOFT_RUN_UNDER);

                if (l_new_dLinks == null)
                    l_new_dLinks = new  ArrayList(INITIAL_CAPACITY*10);
                l_new_dLinks.add(dLink);
            } // for memb

          } // for link
      } // for (Iterator ite

      if (l_new_dLinks != null) {
          for (DisplayableLinkImpl link : l_new_dLinks) {

                                    // for testing and tracking
                                    //System.out.println("CompoundRightsFactoryImpl."+
                                      //"detectLastSimpleSoftRunUnderDisplayableLinks(), "+
                                      //"add the simple link for SOFT_RUN_UNDER: "+link);


              linkRights = interpreter_.getLinkRightsInstances().get(
                                link.getEorSecondEnd().getName());
              linkRights.addDisplayableLink(link, _upd_m_l_DisplayableLinks);
          } // for link
      } // if

      l_dLinks = null;
      dLink = null;
      l_members = null;
      l_new_dLinks = null;
      linkRights = null;
      
      return _upd_m_l_DisplayableLinks;
  } // detectLastSimpleSoftRunUnderDisplayableLinks()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
    interpreter_ = null;
    linkUtil_ = null;
    acl_ = null;
    ago_ = null;
    pri_ = null;
    utility_ = null;

    l_hiddenSourceToTest_.clear();
    l_hiddenSourceLinks_.clear();
    l_ownersForOwnAndContainToTest_.clear();
    l_actorsToTest_.clear();
    l_virtualFoldersToTest_.clear();
    l_ownersForXIDToTest_.clear();
    l_currentOwnersLinks_.clear();
    l_currentActorsLinks_.clear();
    l_currentVirtualFoldersLinks_.clear();
    l_hiddenOwnersLinks_.clear();
    l_hiddenActorsLinks_.clear();
    l_hiddenVirtualFoldersLinks_.clear();
    l_allHiddenOwnersLinks_.clear();
    l_allHiddenActorsLinks_.clear();
    l_allHiddenVirtualFoldersLinks_.clear();
    if (l_endOfLinks_ != null)
        l_endOfLinks_.clear();
    l_endOfLinks2_.clear();
  } // finalizeForProcess()

} // CompoundRightsFactoryImpl
