//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import ARoad0.gWorkInterface.CoreAlgorithm;
import ARoad0.gWorkInterface.AlgorithmInterpreter;
import java.awt.Font;
import java.awt.Color;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;
import java.util.Collection;
import java.util.ArrayList;

import javax.swing.JTextPane;
import javax.swing.text.Document;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.StyleConstants;

import ARoad0.CNot.*;
import ARoad0.Gui2.GraphicView;
import ARoad0.Gui2.GraphicEPRView;
import ARoad0.Gui2.GraphicNoThanView;

import ARoad0.gBase.BaseManagerImpl;
import ARoad0.gBase.UtilityImpl;
import ARoad0.gBaseInterface.*;

/**
 * This class provides some core algorithms about the rights and the 'See why' texts.
 * It consists exclusively of final methods, and most of them are static.
 * This class is stateless, and RightsMediatorImpl at its creation
 * setups an instance for its ViewInBase.
 * No direct call to the ACS methods.
 * It is used by gWork and Gui2 classes.
 * @see ARoad0.CNot.AccessControlLinkImpl
 * @version 0.7.0
 */
public class FactoryUtilities implements CoreAlgorithm {

  // environment
  private static String             lineSepar__ = System.getProperty("line.separator");

  // table of the acronyms for the standardized rights
  private static Map                M_RIGHTS_ACRONYMS = new HashMap(INITIAL_CAPACITY);

  private AlgorithmInterpreter      interpreter_ = null;
  private UtilityImpl               baseUtil_ = null;
  private DisplayableLinkUtilities  linkUtil_ = null;
  private Set<StringRight>          l_rights_ = new HashSet(INITIAL_CAPACITY);

  // --------------------- static initialization -----------------------------
  static {

    M_RIGHTS_ACRONYMS.put("|unknown_rights|", "<unknown>");
    M_RIGHTS_ACRONYMS.put("|modify_from_parent_acs|", "<modify from root>");
    M_RIGHTS_ACRONYMS.put("|modify_owner|", "<modify owner>");
    M_RIGHTS_ACRONYMS.put("|modify_security_descriptor|", "<modify right>");
    M_RIGHTS_ACRONYMS.put("|is_superadmin|", "<superadmin>");

    M_RIGHTS_ACRONYMS.put("|read_from_parent_acs|", "<read from root>");
    M_RIGHTS_ACRONYMS.put("|get_target_rights|", "<has target right>");
    M_RIGHTS_ACRONYMS.put("|unverified_rights|", "<unverified right>");
    M_RIGHTS_ACRONYMS.put("|transfer_all_rights_to_third_party|", 
            "<may grant all to third>");
    M_RIGHTS_ACRONYMS.put("|transfer_limited_rights_to_third_party|", 
            "<may grant to third>");
    M_RIGHTS_ACRONYMS.put("|transfer_all_rights|",
            "<grant all to second>");
    M_RIGHTS_ACRONYMS.put("|transfer_limited_rights|",
            "<grant to second>");
  } // static

  /**
  * only one protected constructor
  * @param _interpreter algorithm interpreter
  */
  protected FactoryUtilities(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      baseUtil_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
  }

  /**
  * Returns true if each node pair in the DisplayableLink is activated, that is
  * each AccessControlLink contains one pseudo-structural link or non-null rights.
  * This method is used to verify that the negative rights
  * in an AccessControlLink have not deleted the
  * positive rights which exist in the same AccessControlLink.
  * Following the AGO right priorities, it is also used to remove an 'owner' path
  * without active rights, which is there to show the need to remove all
  * the similar 'group' and 'other' paths.
  * An active path may provide no rights at all, when there is a pseudo-structural link.
  * All the _dLink nodes have to be ACSObject to get an optimized result using
  * the selection of the ACS rights, but otherwise, the result is still provided.
  * Called by RightsFactory_Facade.getEffectiveRightsForPair().
  * Calls AccessControlLinkImpl.getRights(), isTrueForOnePseudoStructuralLinkType()
  * and subtractNegativeNotRemovingInOneACS().
  * @param _dLink to analyze
  * @return false if _dLink is not active.
  **/
  public final boolean isActivePath(DisplayableLinkImpl _dLink) {
    // variables
    AccessControlLinkImpl[] l_acLinks = _dLink.getEorAllACLinks();
    Set<StringRight>[]      l_l_rights = null;
    ImmutableACS            acs = null;
    boolean                 notActive = false;
    boolean                 forNode = false;
    boolean                 forVirtualFolder = false;
    int                     immutableRange = 1; // only 'owner' rights are immutable
    l_rights_.clear();
    
                      // for testing and tracking
                      //System.out.println("FactoryUtilities.isActivePath(), start "+
                        //"for dLink "+_dLink);
    
    // do not test the last link
    for (int i = 0; i < l_acLinks.length - 1; i++ ) {
      if ( l_acLinks[i].isTrueForOnePseudoStructuralLinkType() ) {

                      // for testing and tracking
                      //System.out.println(//"FactoryUtilities.isActivePath(), "+
                        //"acLink.isTrueForOnePseudoStructuralLinkType() is true for i: "+i);
                        //" in dLink "+_dLink);

             // it may be an active path
          continue;
      } // if
      
      // ---------- TODO ------------------------
      // if there is an ACL in the link, the ACS is the ACL ACS only if
      // there is not other activated link types. The general expression of rights is:
      // - for each of the N ACL ACS, pos rights - neg rights through getRights()
      // - for other types, pos rights - neg rights through getRights()
      // - sum of the N+1 results
      if (_dLink.getEorConnectedNodes(l_acLinks[i])[1] instanceof ACSObject) {
          acs = (ImmutableACS) BaseManagerImpl.getInstance().getEorM_OpenACS().get(
              ((ACSObject)_dLink.getEorConnectedNodes(l_acLinks[i])[1]).getAcsName());
      } // if
      
      if (_dLink.getEorConnectedNodes(l_acLinks[i])[1] instanceof Node)
          forNode = true;
      if (_dLink.getEorConnectedNodes(l_acLinks[i])[1] instanceof VirtualFolder)
          forVirtualFolder = true;

            // implicit ACS rules are:
            // 'AGO right: A rights overlay G rights'
            // 'right type rule: root, A in AGO granting/denying rights are not changed by other rights'     
      if (l_acLinks[i].getLinkTypeValue(AccessControlLink.OWNER))
          l_l_rights = l_acLinks[i].getFilteredRights(immutableRange, true, true);
      else if (l_acLinks[i].getLinkTypeValue(AccessControlLink.CONTAIN)) // withoutGroupInAGOrights false
          l_l_rights = l_acLinks[i].getFilteredRights(immutableRange, false, true);
      // withoutOtherInAGOrights false
      else l_l_rights = l_acLinks[i].getFilteredRights(immutableRange, false, false);
      
      if (l_l_rights[0].isEmpty() && l_l_rights[1].isEmpty() && l_l_rights[3].isEmpty()) {
            l_rights_.addAll( l_l_rights[2] );

                      // for testing and tracking
                      //System.out.println("dLink on "+i+" has the acLink: "+l_acLinks[i]+
                        //      ", with l_l_rights[2] only.");
                      //if (l_acLinks[i].getRights(AccessControlLink.ACL) != null &&
                        //      ((StringRight[])l_acLinks[i].getRights(AccessControlLink.ACL)).length != 0)
                          //System.out.println(" the acLink has ACL first right: "+
                            //      l_acLinks[i].getRights(AccessControlLink.ACL)[0]);

      }
      else {
          // process the filtered positive immutable rights first
            if (!l_l_rights[0].isEmpty() && !l_l_rights[1].isEmpty())
                l_rights_.addAll( baseUtil_.subtractNegativeNotRemovingInOneACS(
                    l_l_rights[0], l_l_rights[1], acs, forNode) ); // acs may be null
            else if (!l_l_rights[0].isEmpty() ) 
                l_rights_.addAll( l_l_rights[0] );
            
          // add all the other filtered positive rights to the immutable rights
            if (!l_l_rights[2].isEmpty() && !l_l_rights[3].isEmpty())
                l_rights_.addAll( baseUtil_.subtractNegativeNotRemovingInOneACS(
                    l_l_rights[2], l_l_rights[3], acs, forNode) ); // acs may be null
            else if (!l_l_rights[2].isEmpty() ) 
                l_rights_.addAll( l_l_rights[2] );
      } // else      

                      // for testing and tracking
                      //System.out.println("dLink on "+i+" has the acLink: "+l_acLinks[i]+
                        //      ", with the final rights:");
                      //for (Iterator ite = l_rights.iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}
      
       // isTrueForOneStrengthenLinkType() false, so controls the executing rights
      // if the ending node is not a VirtualFolder
      // (where the rights have been copied on the next AccessControlLink)
      if (!forVirtualFolder &&
              !baseUtil_.withExecuteRight(l_rights_.toArray(new StringRight[0]), false)) {

                      // for testing and tracking
                      //System.out.println("\n FactoryUtilities.isActivePath(), "+
                        //"dLink is not active on acLink: "+l_acLinks[i]+", with the 4 initial rights:");
                      //System.out.println("-- l_l_rights[0]:");
                      //for (Iterator ite = l_l_rights[0].iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}
                      //System.out.println("-- l_l_rights[1]:");
                      //for (Iterator ite = l_l_rights[1].iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}
                      //System.out.println("-- l_l_rights[2]:");
                      //for (Iterator ite = l_l_rights[2].iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}          
                      //System.out.println("-- l_l_rights[3]:");
                      //for (Iterator ite = l_l_rights[3].iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}                      
                      
        notActive = true;
        break;
      } // if
      l_rights_.clear();
    } // for i
    
    l_acLinks = null;
    l_l_rights = null;
    acs = null;
    
    if (notActive) {
        return false;
    }
    else return true;

  } // isActivePath()

  /**
  * Construct a JTextPane with the tree of all DisplayableLinks and AccessControlLinks,
  * to be displayed as a comment for the GraphicView associated
  * to _m_l_DisplayableLinks. NO USE.
  * The 'Serif' font is used with the size 12, and 14 for the title.
  * Do not define the size of the pane.
  * @param _gview is the graphic view associated to the view
  * @param _m_l_DisplayableLinks immutable Map of DisplayableLinks lists.
  * @param _view associated to _m_l_DisplayableLinks
  * @return JTextPane as a comment for the GraphicView associated
  *    to _m_l_DisplayableLinks.
  **/
  public static final JTextPane getWhyText(GraphicView _gview,
    Map _m_l_DisplayableLinks, ViewInBase _view) {
    // variables
    JTextPane             result = new JTextPane();
      // concatenated name with an EP name and a Resource name from _view
    Name                  pairKey;
    Name                  acsName1;
    Name                  acsName2;
    DisplayableLinkImpl   dLink;
    AccessControlLinkImpl acLink;
    int                   dLinkCount = 1;
    int                   acLinkCount = 1;
    BaseObject            obj = null;
    AccessControlLinkImpl[] l_acLinks = null;

    // initialize the text font
    result.setFont(new Font("Serif", Font.PLAIN, 12));
    SimpleAttributeSet blueBold = new SimpleAttributeSet();
    StyleConstants.setBold(blueBold, true);
    StyleConstants.setForeground(blueBold, Color.blue);
    SimpleAttributeSet bold = new SimpleAttributeSet();
    StyleConstants.setBold(bold, true);
    SimpleAttributeSet centric = new SimpleAttributeSet();
    StyleConstants.setAlignment(centric, StyleConstants.ALIGN_CENTER);
    StyleConstants.setFontSize(centric, 14);
    StyleConstants.setBold(centric, true);
    SimpleAttributeSet italic = new SimpleAttributeSet();
    StyleConstants.setItalic(italic, true); ;

    appendText(result, lineSepar__, null);
    appendText(result, lineSepar__, null);
    
    if (_m_l_DisplayableLinks.isEmpty()) {
        
        appendText(result, "In this view, there is no authorized access.", bold);


        if (!_gview.getRightsRelatedBaseObjects().isEmpty()) {
          // add the list of all rights-related nodes implied in the current search
          appendText(result, "          -----  -----  -----", null);
          //appendText(result, lineSepar__, null);
          appendText(result, lineSepar__, null);
          appendText(result, "  Access Road has searched all the access paths"+
          " through the applicable nodes, and it has found no path.", null);
          appendText(result, lineSepar__, null);
          appendText(result, "    All the processed nodes outside the view are herein: ", null);
          appendText(result, lineSepar__, null);

          for (Iterator ite = _gview.getRightsRelatedBaseObjects().iterator(); ite.hasNext(); ) {
            obj = (BaseObject)ite.next();
            // AclEntry are listened for changes but not displayed there
            if (obj instanceof AclEntry) continue;
            appendText(result, "        * "+obj.getName(), null);
            appendText(result, lineSepar__, null);
          } // for ite
        } // if        
        
        blueBold = null;
        bold = null;
        centric = null;
        obj = null;
 
                // for testing and tracking
                //try {        
                  //System.out.println("in FactoryUtilities.setWhyText(),"+
                    //"the links are empty - result: "+result.getDocument().getText(0, 20));
                //} catch (BadLocationException b) {
                    //b.printStackTrace();
                //}
        
        return result;
    }

        // for testing and tracking
        //System.out.println("in FactoryUtilities.setWhyText(),"+
        //  "_m_l_DisplayableLinks: "+_m_l_DisplayableLinks);

    for (Iterator ita = _view.getEorM_BaseObjects().keySet().iterator() ;
      ita.hasNext() ; ) {
        acsName1 = (Name) ita.next();

        for (Iterator ito = _view.getEorM_BaseObjects().keySet().iterator() ;
          ito.hasNext() ; ) {
            acsName2 = (Name) ito.next();
            if (acsName1.equals(acsName2)) continue;
            pairKey = ((Name)acsName1.clone()).extended(acsName2);

            // tests each possible pairKey, and writes it in result
            if (_m_l_DisplayableLinks.containsKey(pairKey)) {
              appendText(result, "PATH(S)", bold);
              appendText(result, " FROM "+acsName1.context().toString(), null);
              appendText(result, acsName1.lastComponent(), blueBold);
              appendText(result, " TO "+acsName2.context().toString(), null);
              appendText(result, acsName2.lastComponent().toString()+"."+
                lineSepar__, blueBold);
              dLinkCount = 1;
              for (Iterator iti = ((List)_m_l_DisplayableLinks.get(pairKey)).
                iterator() ; iti.hasNext() ; ) {
                  dLink = (DisplayableLinkImpl) iti.next();
                  appendText(result, "-- PATH number "+dLinkCount+":"+
                    lineSepar__, bold);
                  dLinkCount++;
                  acLinkCount = 1;
                  l_acLinks = dLink.getEorAllACLinks();

                  for (int i = 0 ; i < l_acLinks.length ; i++ ) {
                    acLink = l_acLinks[i];
                    appendText(result, "            Link number "+acLinkCount+":"
                      +lineSepar__, null); // 12 " "
                    acLinkCount++;
                    appendText(result, "                  "+acLink.toSeeWhyText()+
                      lineSepar__, null); // 18 " "
                    if (i != l_acLinks.length - 1) {
                      appendText(result, "            Intermediate node:"+lineSepar__, null); // 12 " "
                      appendText(result, "                  "+dLink.
                        getEorConnectedNodes( acLink )[1].getName().context(), null); // 18 " "
                      appendText(result, dLink.
                        getEorConnectedNodes( acLink )[1].getName().lastComponent()
                        +lineSepar__, bold);
                    } // if
                  } // for i

              } // iti

              appendText(result, lineSepar__, null);
            } // else if (!_m_l_DisplayableLinks.containsKey(pairKey))
        } // ito
    } // ita

    if (!_gview.getRightsRelatedBaseObjects().isEmpty()) {
      // add the list of all rights-related nodes implied in the current search
      appendText(result, "          -----  -----  -----", null);
      //appendText(result, lineSepar__, null);
      appendText(result, lineSepar__, null);
      appendText(result, "  Access Road has found all the access paths."+
              " Some outside-the-view processed nodes are not listed above, "+
              "because the search of path has get no result through them.", null);
      appendText(result, lineSepar__, null);
      appendText(result, "    These nodes are listed herein: ", null);
      appendText(result, lineSepar__, null);

      for (Iterator ite = _gview.getRightsRelatedBaseObjects().iterator(); ite.hasNext(); ) {
        obj = (BaseObject)ite.next();
        // AclEntry are listened for changes but not displayed there
        if (obj instanceof AclEntry) continue;
        appendText(result, "        * "+obj.getName(), null);
        appendText(result, lineSepar__, null);
      } // for ite
    } // if
    
    pairKey = null;
    acsName1 = null;
    acsName2 = null;
    dLink = null;
    acLink = null;
    obj = null;
    l_acLinks = null;
    return result;
  } // setWhyText()
  
  /**
  * Construct a JTextPane with the tree of all DisplayableLinks and AccessControlLinks,
  * to be displayed as a text comment for a GraphicView.
  * Display first the paths RIGHTS USER/ACCESS TARGET, then the paths RIGHTS USER/GROUPMEMBER,
  * and finally the paths GROUPIDMEMBER/GROUPID.
  * Does not display the DisplayableLinks which contain a sequence of 3 nodes
  * which are there 5 times in the previous DisplayableLinks having the same ends,
  * and add a note to say what nodes sequence has been filtered. 
  * Note that this processing has no influence on the paths search in Access Road,
  * since it is done only for the displaying of the result.
  * The 'Serif' font is used with the size 12, and 14 for the title.
  * Does not define the size of the pane.
  * Called by RightsFactory_Facade.getEPRWhyText().
  * Calls linkUtil_.removeTooNumerous3NodesSequence().
  * @param _gview is the graphic view associated to the view
  * @param _m_l_DisplayableLinks immutable Map of DisplayableLinks lists.
  * @param _view associated to _m_l_DisplayableLinks
  * @return JTextPane as a comment for the GraphicView associated
  * to _m_l_DisplayableLinks.
  **/
  public final JTextPane getEPRWhyText(GraphicEPRView _gview,
    Map _m_l_DisplayableLinks, EPRViewInBase _view) {
    // variables
    JTextPane             result = new JTextPane();
      // concatenated name with an EP name and a Resource name from _view 
    Name                  pairKey;
    Name                  epName;
    Name                  targName;
    DisplayableLinkImpl   dLink;
    AccessControlLinkImpl acLink;
    int                   dLinkCount = 1;
    int                   acLinkCount = 1;
    Set                   l_viewNoActorEP = new HashSet(INITIAL_CAPACITY);
    Set                   l_viewTargets;
    Set                   l_filtered3nodesSequences = new HashSet(INITIAL_CAPACITY);
    Set                   l_filtered;
    List                  l_dLinksForPair;
    BaseObject            obj = null;
    BaseObject[]          l_nodes = null;
    AccessControlLinkImpl[] l_acLinks = null;
    
    // initializes l_viewNoActorEP
    l_viewNoActorEP.addAll(_view.getEorM_Sources().keySet());
    l_viewTargets = new HashSet(_view.getEorM_Targets().keySet());
    // removeAll() changes l_viewTargets
    l_viewNoActorEP.removeAll(l_viewTargets);
    l_viewTargets = null;
    
    // initializes the text font
    result.setFont(new Font("Serif", Font.PLAIN, 12));
    SimpleAttributeSet blueBold = new SimpleAttributeSet();
    StyleConstants.setBold(blueBold, true);
    StyleConstants.setForeground(blueBold, Color.blue);
    SimpleAttributeSet bold = new SimpleAttributeSet();
    StyleConstants.setBold(bold, true);
    SimpleAttributeSet centric = new SimpleAttributeSet();
    StyleConstants.setAlignment(centric, StyleConstants.ALIGN_CENTER);
    StyleConstants.setFontSize(centric, 14);
    StyleConstants.setBold(centric, true);
    SimpleAttributeSet italic = new SimpleAttributeSet();
    StyleConstants.setItalic(italic, true); ;
    
    if (_m_l_DisplayableLinks.isEmpty()) {
        
        appendText(result, "In this view, there is no authorized access.", bold);

        // new in v065
        if (_gview.getRightsRelatedBaseObjects() != null &&
                !_gview.getRightsRelatedBaseObjects().isEmpty()) {
          // adds the list of all rights-related nodes implied in the current search
          appendText(result, "          -----  -----  -----", null);
          //appendText(result, lineSepar__, null);
          appendText(result, lineSepar__, null);
          appendText(result, "  Access Road has searched all the access paths"+
          " through the applicable nodes, and it has found no path.", italic);
          appendText(result, lineSepar__, null);
          appendText(result, "    All the  processed nodes outside the view, as intermediate node"+
              " or not, are listed hereinafter with the relevant privileges, if any: ", italic);
          appendText(result, lineSepar__, null);

          for (Iterator ite = _gview.getRightsRelatedBaseObjects().iterator(); ite.hasNext(); ) {
            obj = (BaseObject)ite.next();
            // AclEntry are listened for changings but not displayed there
            if (obj instanceof AclEntry) continue;
            appendText(result, "        * "+obj.getName(), null);
            appendText(result, lineSepar__, null);
          } // for ite
        } // if        
        
        
        pairKey = null;
        dLink = null;
        acLink = null;

        blueBold = null;
        bold = null;
        centric = null;

        l_viewNoActorEP = null;
        obj = null;
 
                // for testing and tracking
                //try {        
                  //System.out.println("in FactoryUtilities.setWhyText(),"+
                    //"the links are empty - result: "+result.getDocument().getText(0, 20));
                //} catch (BadLocationException b) {
                    //b.printStackTrace();
                //}
        
        return result;
    }

        // for testing and tracking
        //System.out.println("in FactoryUtilities.setWhyText(),"+
        //  "_m_l_DisplayableLinks: "+_m_l_DisplayableLinks);

    for (Iterator ita = _view.getEorM_Sources().keySet().iterator() ;
      ita.hasNext() ; ) {
        epName = (Name) ita.next();

        // from an EP to a Resource
        for (Iterator ito = _view.getEorM_Targets().keySet().iterator() ;
          ito.hasNext() ; ) {
            targName = (Name) ito.next();
            if (targName.equals(epName)) continue;
            pairKey = ((Name)epName.clone()).extended(targName);
              
                // for testing and tracking
                //System.out.println("in FactoryUtilities.setWhyText(), "+
                  //      "check of a Resource: "+pairKey);
            
            // tests each possible pairKey, and writes it in result
            if (_m_l_DisplayableLinks.containsKey(pairKey)) {
              appendText(result, "FROM A RIGHT USER TO AN ACCESS TARGET"+lineSepar__, bold);
              appendText(result, " FROM "+epName.context().toString(), null);
              appendText(result, epName.lastComponent()+lineSepar__, blueBold);
              appendText(result, " TO "+targName.context().toString(), null);
              appendText(result, targName.lastComponent().toString()+"."+
                lineSepar__, blueBold);
              dLinkCount = 1;
              l_dLinksForPair = (List)_m_l_DisplayableLinks.get(pairKey);
              if (l_dLinksForPair == null) continue;

              if (l_dLinksForPair.size() > 6) {// update l_dLinksForPair
                  l_filtered = linkUtil_.
                          removeTooNumerous3NodesSequence(l_dLinksForPair);
                  if (l_filtered != null)
                      l_filtered3nodesSequences.addAll( l_filtered );
              } // if
              
              for (Iterator iti = l_dLinksForPair.iterator() ; iti.hasNext() ; ) {
                  dLink = (DisplayableLinkImpl) iti.next();
                  if (dLink == null) continue;

                  //appendText(result, lineSepar__, null);
                  appendText(result, "-- PATH NUMBER "+dLinkCount+":"+
                    lineSepar__, bold);
                  dLinkCount++;
                  acLinkCount = 1;
                  l_acLinks = dLink.getEorAllACLinks();

                  for (int i = 0 ; i < l_acLinks.length ; i++ ) {
                    acLink = l_acLinks[i];

                    appendText(result, "            Link number "+acLinkCount+":"
                      +lineSepar__, null); // 12 " "
                    acLinkCount++;
                    appendText(result, "                  "+acLink.toSeeWhyText()+
                      lineSepar__, null); // 18 " "
                    
                    if (i != l_acLinks.length - 1) {
                      appendText(result, "            Intermediate node:"+lineSepar__, null); // 12 " "
                      appendText(result, "                  "+dLink.
                        getEorConnectedNodes( acLink )[1].getName().context(), null); // 18 " "
                      appendText(result, dLink.
                        getEorConnectedNodes( acLink )[1].getName().lastComponent()
                        +lineSepar__, bold);
                    } // if
                  } // for i

              } // iti

              appendText(result, lineSepar__, null);
            } // if (_m_l_DisplayableLinks.containsKey(pairKey))
        } // ito
              
                // for testing and tracking
                //System.out.println("in FactoryUtilities.setWhyText(), "+
                  //      "end of checking for Resources for: "+epName);  

        // from an Actor to a GroupIDMember
        if (!_view.getEorM_Targets().keySet().contains(epName)) continue;
        
        for (Iterator ito = l_viewNoActorEP.iterator() ;
          ito.hasNext() ; ) {
            targName = (Name) ito.next();
            if (targName.equals(epName)) continue;
            pairKey = ((Name)epName.clone()).extended(targName);
              
                // for testing and tracking
                //System.out.println("in FactoryUtilities.setWhyText(), "+
                  //      "check of a Resource: "+pairKey);
            
            // tests each possible pairKey, and writes it in result
            if (_m_l_DisplayableLinks.containsKey(pairKey)) {
              appendText(result, "FROM AN ACTOR TO A GROUP MEMBER"+lineSepar__, bold);
              appendText(result, " FROM "+epName.context().toString(), null);
              appendText(result, epName.lastComponent()+lineSepar__, blueBold);
              appendText(result, " TO "+targName.context().toString(), null);
              appendText(result, targName.lastComponent().toString()+"."+
                lineSepar__, blueBold);
              dLinkCount = 1;
              l_dLinksForPair = (List)_m_l_DisplayableLinks.get(pairKey);
              if (l_dLinksForPair == null) continue;
              
              if (l_dLinksForPair.size() > 6) {// update l_dLinksForPair
                  l_filtered = linkUtil_.
                          removeTooNumerous3NodesSequence(l_dLinksForPair);
                  if (l_filtered != null)
                      l_filtered3nodesSequences.addAll( l_filtered );
              } // if
              
              for (Iterator iti = l_dLinksForPair.iterator() ; iti.hasNext() ; ) {
                  dLink = (DisplayableLinkImpl) iti.next();
                  if (dLink == null) continue;

                  //appendText(result, lineSepar__, null);
                  appendText(result, "-- PATH NUMBER "+dLinkCount+":"+
                    lineSepar__, bold);
                  dLinkCount++;
                  acLinkCount = 1;
                  l_acLinks = dLink.getEorAllACLinks();

                  for (int i = 0 ; i < l_acLinks.length ; i++ ) {
                    acLink = l_acLinks[i];
                    
                    appendText(result, "            Link number "+acLinkCount+":"
                      +lineSepar__, null); // 12 " "
                    acLinkCount++;
                    appendText(result, "                  "+acLink.toSeeWhyText()+
                      lineSepar__, null); // 18 " "
                    
                    if (i != l_acLinks.length - 1) {
                      appendText(result, "            Intermediate node:"+lineSepar__, null); // 12 " "
                      appendText(result, "                  "+dLink.
                        getEorConnectedNodes( acLink )[1].getName().context(), null); // 18 " "
                      appendText(result, dLink.
                        getEorConnectedNodes( acLink )[1].getName().lastComponent()
                        +lineSepar__, bold);
                    } // if
                  } // for i

              } // iti

              appendText(result, lineSepar__, null);
            } // if (_m_l_DisplayableLinks.containsKey(pairKey))
        } // ito
        
                // for testing and tracking
                //System.out.println("in FactoryUtilities.setWhyText(), "+
                  //      "end of checking for GroupIDMembers for: "+epName);    
    } // ita

    // adds the links GroupIDMember/Group
    Set viewNoActorEP_clone =  new HashSet(l_viewNoActorEP);
    for (Iterator ita = l_viewNoActorEP.iterator() ; ita.hasNext() ; ) {
        epName = (Name) ita.next();

        for (Iterator ito = viewNoActorEP_clone.iterator() ; ito.hasNext() ; ) {
            targName = (Name) ito.next();
            if (targName.equals(epName)) continue;
            pairKey = ((Name)epName.clone()).extended(targName);

            // test each possible pairKey, and write it in result if it is not an actor
            if (!_m_l_DisplayableLinks.containsKey(pairKey)) {
                /*
              appendText(result, "There is no path", bold);
              appendText(result, " from "+epName.context().toString(), null);
              appendText(result, epName.lastComponent(), blueBold);
              appendText(result, " to "+resName.context().toString(), null);
              appendText(result, resName.lastComponent().toString()+"."+
                lineSepar__+lineSepar__, blueBold);
                 * */
            } // if (!_m_l_DisplayableLinks.containsKey(pairKey))
            // if Actor, done in the previous loop
            else if (!(_view.getEorM_Sources().get(targName) instanceof Actor)) {
              appendText(result, "FROM A GROUP MEMBER TO A GROUP"+lineSepar__, bold);
              appendText(result, " FROM "+epName.context().toString(), null);
              appendText(result, epName.lastComponent()+lineSepar__, blueBold);
              appendText(result, " TO "+targName.context().toString(), null);
              appendText(result, targName.lastComponent().toString()+"."+
                lineSepar__, blueBold);
              dLinkCount = 1;
              if (_m_l_DisplayableLinks.get(pairKey) == null) continue;

              for (Iterator iti = ((List)_m_l_DisplayableLinks.get(pairKey)).
                iterator() ; iti.hasNext() ; ) {
                  dLink = (DisplayableLinkImpl) iti.next();
                  if (dLink == null) continue;

                  appendText(result, "-- PATH NUMBER "+dLinkCount+":"+
                    lineSepar__, bold);
                  dLinkCount++;
                  acLinkCount = 1;
                  l_acLinks = dLink.getEorAllACLinks();

                  for (int i = 0 ; i < l_acLinks.length ; i++ ) {
                    acLink = l_acLinks[i];
                    
                    appendText(result, "            Link number "+acLinkCount+":"
                      +lineSepar__, null); // 12 " "
                    acLinkCount++;
                    appendText(result, "                  "+acLink.toSeeWhyText()+
                      lineSepar__, null); // 18 " "
                    
                    if (i != l_acLinks.length - 1) {
                      appendText(result, "            Intermediate node:"+lineSepar__, null); // 12 " "
                      appendText(result, "                  "+dLink.
                        getEorConnectedNodes( acLink )[1].getName().context(), null); // 18 " "
                      appendText(result, dLink.
                        getEorConnectedNodes( acLink )[1].getName().lastComponent()
                        +lineSepar__, bold);
                    } // if
                  } // for i

              } // iti

              appendText(result, lineSepar__, null);
            } // else if (!_m_l_DisplayableLinks.containsKey(pairKey))
        } // ito
    } // ita

    // adds the links VirtualFolder/ACSObject
    for (Iterator ita = _view.getEorM_NoSourceTargets().values().iterator() ; ita.hasNext() ; ) {
        obj = (ACSObject) ita.next();
        if (!(obj instanceof VirtualFolder)) continue;
        epName = (Name)((VirtualFolder)obj).getName();

        for (Iterator ito = ((VirtualFolder)obj).getEorM_BaseObjects().keySet().iterator() ;
          ito.hasNext() ; ) {
            targName = (Name) ito.next();
            if (!_view.getEorM_Targets().keySet().contains(targName)) continue;
            pairKey = ((Name)((Name)epName).clone()).extended(targName);

            // test each possible pairKey, and write it in result
            if (!_m_l_DisplayableLinks.containsKey(pairKey)) {
                /*
              appendText(result, "There is no path", bold);
              appendText(result, " from "+epName.context().toString(), null);
              appendText(result, epName.lastComponent(), blueBold);
              appendText(result, " to "+resName.context().toString(), null);
              appendText(result, resName.lastComponent().toString()+"."+
                lineSepar__+lineSepar__, blueBold);
                 * */
            } // if (!_m_l_DisplayableLinks.containsKey(pairKey))
            // if 
            else {
              appendText(result, "FROM A VIRTUAL FOLDER TO A VIRTUAL MEMBER"+lineSepar__, bold);
              appendText(result, " FROM "+epName.context().toString(), null);
              appendText(result, epName.lastComponent()+lineSepar__, blueBold);
              appendText(result, " TO "+targName.context().toString(), null);
              appendText(result, targName.lastComponent().toString()+"."+
                lineSepar__, blueBold);
              dLinkCount = 1;
              if (_m_l_DisplayableLinks.get(pairKey) == null) continue;

              for (Iterator iti = ((List)_m_l_DisplayableLinks.get(pairKey)).
                iterator() ; iti.hasNext() ; ) {
                  dLink = (DisplayableLinkImpl) iti.next();
                  if (dLink == null) continue;

                  appendText(result, "-- PATH NUMBER "+dLinkCount+":"+
                    lineSepar__, bold);
                  dLinkCount++;
                  acLinkCount = 1;
                  l_acLinks = dLink.getEorAllACLinks();

                  for (int i = 0 ; i < l_acLinks.length ; i++ ) {
                    acLink = l_acLinks[i];

                    appendText(result, "            Link number "+acLinkCount+":"
                      +lineSepar__, null); // 12 " "
                    acLinkCount++;
                    appendText(result, "                  "+acLink.toSeeWhyText()+
                      lineSepar__, null); // 18 " "
                  } // for i
              } // for iti

              appendText(result, lineSepar__, null);
            } // else if (!_m_l_DisplayableLinks.containsKey(pairKey))
        } // ito
    } // ita
    

    if (!_gview.getRightsRelatedBaseObjects().isEmpty()) {
      // adds the list of all rights-related nodes implied in the current search
      appendText(result, "                            ---------------------", null);
      //appendText(result, lineSepar__, null);
      appendText(result, lineSepar__, null);
      //if (viewContainsExecuteAGOotherRight)
        //  appendText(result, "  Access Road has found all the rights, and almost all paths."+
          //    " For paths, and to limit the processing load,"+
            //  " the sole limitation is for the intermediate Ubuntu executables"+
            //  " which run under current account/group, and are executable from the 'other' AGO rights."+
            //  " When these intermediate executables are involved, only the first paths are found, if any."+
            //  " Some nodes outside the view have been processed, but"+
            //  " the search has get no access path through them, and"+
            //  " this is why they are not intermediate nodes.", italic);
      //else
          appendText(result, "  Access Road has found all the rights and all the paths."+
              " Some nodes outside the view have been processed, but"+
              " the search has get no access path through them, and"+
              " this is why they are not intermediate nodes.", italic);
      appendText(result, lineSepar__, null);
      appendText(result, "    All the  processed nodes outside the view, as intermediate node"+
              " or not, are listed hereinafter with the relevant privileges, if any: ", italic);
      appendText(result, lineSepar__, null);

      for (Iterator ite = _gview.getRightsRelatedBaseObjects().iterator(); ite.hasNext(); ) {
        obj = (BaseObject)ite.next();
        // AclEntry and privileges are listened for changings but not displayed there
        if (obj instanceof AclEntry) continue;
        appendText(result, "        * "+obj.getName(), null);
        appendText(result, lineSepar__, null);
      } // for ite
      
      if (!l_filtered3nodesSequences.isEmpty()) {
          appendText(result, lineSepar__, null);
          appendText(result, "    For some pairs RIGHT USER/ACCESS TARGET, too many paths contain"+
                  " the same 3-nodes sequence, and this gives an uselessly large list."+
                  " For these pairs, only 5 of the relevant paths are displaying above."+
                  " This filtering is done AFTER the detection of all the access paths. "+
                  "The 3-nodes sequences which have been filtered in such a way are given hereinafter: ", italic);
          appendText(result, lineSepar__, null);

          for (Iterator ite = l_filtered3nodesSequences.iterator(); ite.hasNext(); ) {
            l_nodes = (BaseObject[])ite.next();
            for (int i = 0; i < l_nodes.length; i++) {
              if (i == 0) appendText(result, "        * "+l_nodes[i]+";", null);
              else appendText(result, " "+l_nodes[i]+";", null);
            } // for i
            appendText(result, lineSepar__, null);
          } // for ite 
       } // if
      
    } // if
    
    // put the comments
    appendText(result, lineSepar__, null);
    appendText(result, lineSepar__, null);
    appendText(result, "Note: User comments may be added and copied (Ctrl+C) like in a text editor. ",
            null); // bold
    appendText(result, "Access Road generates this text at each relevant property change."+
            " The user comments are lost at every text generation and every view saving. ",
            null); // bold

    pairKey = null;
    epName = null;
    targName = null;
    dLink = null;
    acLink = null;

    blueBold = null;
    bold = null;
    centric = null;
    
    l_viewNoActorEP = null;
    obj = null;
    l_filtered3nodesSequences = null;
    l_filtered = null;
    l_nodes = null;
    l_acLinks = null;
        
    return result;
  } // getEPRWhyText()
  
  /**
  * Get a JTextPane with the tree of all DisplayableLinks and AccessControlLinks,
  * to be displayed as a text comment for a GraphicNoThanView.
  * The 'Serif' font is used with the size 12, and 14 for the title.
  * Do not define the size of the pane.
  * Called by RightsFactory_Facade.getNoThanWhyText().
  * @param _gview is the graphic view associated to the view
  * @param _m_l_DisplayableLinks immutable Map of DisplayableLinks lists.
  * @param _view associated to _m_l_DisplayableLinks
  * @return JTextPane as a comment for the GraphicView associated
  * to _m_l_DisplayableLinks.
  **/
  public final JTextPane getNoThanWhyText(GraphicNoThanView _gview,
    Map _m_l_DisplayableLinks, NoThanViewInBase _view) {
    // variables
    JTextPane             result = new JTextPane();
      // concatenated name with an EP name and a Resource name from _view 
    DisplayableLinkImpl   dLink;
    AccessControlLinkImpl acLink;
    int                   acLinkCount = 1;
    List                  l_dLinksForPair;
    BaseObject            obj = null;
    Map.Entry             entry = null;
    AccessControlLinkImpl[] l_acLinks = null;
    
    // initialize the text font
    result.setFont(new Font("Serif", Font.PLAIN, 12));
    SimpleAttributeSet blueBold = new SimpleAttributeSet();
    StyleConstants.setBold(blueBold, true);
    StyleConstants.setForeground(blueBold, Color.blue);
    SimpleAttributeSet bold = new SimpleAttributeSet();
    StyleConstants.setBold(bold, true);
    SimpleAttributeSet centric = new SimpleAttributeSet();
    StyleConstants.setAlignment(centric, StyleConstants.ALIGN_CENTER);
    StyleConstants.setFontSize(centric, 14);
    StyleConstants.setBold(centric, true);
    SimpleAttributeSet italic = new SimpleAttributeSet();
    StyleConstants.setItalic(italic, true); ;
    
    if (_m_l_DisplayableLinks.isEmpty()) {
        
        appendText(result, "In this view, all the non-null right criteria are fulfilled.", bold);

        if (_gview.getRightsRelatedBaseObjects() != null &&
                !_gview.getRightsRelatedBaseObjects().isEmpty()) {
          // add the list of all rights-related nodes implied in the current search
          appendText(result, lineSepar__, null);
          appendText(result, lineSepar__, null);
          appendText(result, "          -----  -----  -----", null);
          appendText(result, lineSepar__, null);
          appendText(result, lineSepar__, null);
          appendText(result, "  Access Road has searched all the access paths"+
          " through the applicable nodes.", italic);
          appendText(result, lineSepar__, null);

          appendText(result, "    All the nodes of the checking perimeter are listed hereafter: ",
                  italic);
          appendText(result, lineSepar__, null);
          for (Iterator ite = _view.getCheckingPerimeter().iterator(); ite.hasNext(); ) {
            obj = (BaseObject)ite.next();
            appendText(result, "        * "+obj.getName(), null);
            appendText(result, lineSepar__, null);
          } // for ite
          
          appendText(result, "    All the intermediate nodes outside the basic right users, "+
                  "and the relevant Privileges, are listed hereafter: ", italic);
          appendText(result, lineSepar__, null);

          for (Iterator ite = _gview.getRightsRelatedBaseObjects().iterator(); ite.hasNext(); ) {
            obj = (BaseObject)ite.next();
            // AclEntry are listened for changes but not displayed there
            if (obj instanceof AclEntry) continue;
            appendText(result, "        * "+obj.getName(), null);
            appendText(result, lineSepar__, null);
          } // for ite
        } // if        
        
        dLink = null;
        acLink = null;

        blueBold = null;
        bold = null;
        centric = null;

        obj = null;
        entry = null;
 
                // for testing and tracking
                //try {        
                  //System.out.println("in FactoryUtilities.getNoThanWhyText(),"+
                    //"the links are empty - result: "+result.getDocument().getText(0, 20));
                //} catch (BadLocationException b) {
                    //b.printStackTrace();
                //}
        
        return result;
    }

                // for testing and tracking
                //System.out.println("in FactoryUtilities.getNoThanWhyText(),"+
                  //"_view.getEorFirstAccessTarget().getName().noOfComponents(): "+
                  //_view.getEorFirstAccessTarget().getName().noOfComponents());

    for (Iterator ita = _m_l_DisplayableLinks.entrySet().iterator() ; ita.hasNext() ; ) {
      entry = (Map.Entry) ita.next();

                // for testing and tracking
                //System.out.println("in FactoryUtilities.getNoThanWhyText(), entry key:"+
                  //      entry.getKey());

      l_dLinksForPair = (List)entry.getValue();
      appendText(result, " NO COMPLIANCE FROM THE RIGHT USER "+((Name)entry.getKey()).
              context(((Name)entry.getKey()).noOfComponents() -
                _view.getEorFirstAccessTarget().getName().noOfComponents() - 1), null);

      appendText(result, lineSepar__, null);
      
      for (Iterator iti = l_dLinksForPair.iterator() ; iti.hasNext() ; ) {
          dLink = (DisplayableLinkImpl) iti.next();
          acLinkCount = 1;
          l_acLinks = dLink.getEorAllACLinks();

          for (int i = 0 ; i < l_acLinks.length ; i++ ) {
            acLink = l_acLinks[i];
            appendText(result, "            Link number "+acLinkCount+":"
              +lineSepar__, null); // 12 " "
            acLinkCount++;
            appendText(result, "                  "+acLink.toSeeWhyText()+
              lineSepar__, null); // 18 " "
            if (i != l_acLinks.length - 1) {
              appendText(result, "            Intermediate node:"+lineSepar__, null); // 12 " "
              appendText(result, "                  "+dLink.
                getEorConnectedNodes( acLink )[1].getName().context(), null); // 18 " "
              appendText(result, dLink.
                getEorConnectedNodes( acLink )[1].getName().lastComponent()
                +lineSepar__, bold);
            } // if
          } // for i
      } // iti

      appendText(result, lineSepar__, null);

        // for testing and tracking
        //System.out.println("in FactoryUtilities.getNoThanWhyText(), "+
          //      "end of checking for Resources for: "+epName);   
    } // ita

    if (_gview.getRightsRelatedBaseObjects() != null &&
            !_gview.getRightsRelatedBaseObjects().isEmpty()) {
      // add the list of all rights-related nodes implied in the current search
      appendText(result, "                            ---------------------", null);
      //appendText(result, lineSepar__, null);
      appendText(result, lineSepar__, null);
      appendText(result, "  For each criterion, Access Road has ended the search "+
              "after having found "+
              "one first access path without compliance.", italic);
      appendText(result, lineSepar__, null);
      
      appendText(result, "    All the nodes of the checking perimeter are listed hereinafter: ", italic);
      appendText(result, lineSepar__, null);
      for (Iterator ite = _view.getCheckingPerimeter().iterator(); ite.hasNext(); ) {
        obj = (BaseObject)ite.next();
        appendText(result, "        * "+obj.getName(), null);
        appendText(result, lineSepar__, null);
      } // for ite
      
      appendText(result, "    All the  processed nodes outside the basic right users, "+
              "as intermediate nodes, and the relevant Privileges"+
              " are listed hereinafter: ", italic);
      appendText(result, lineSepar__, null);
      for (Iterator ite = _gview.getRightsRelatedBaseObjects().iterator(); ite.hasNext(); ) {
        obj = (BaseObject)ite.next();
        // AclEntry are listened for changes but not displayed there
        if (obj instanceof AclEntry) continue;
        appendText(result, "        * "+obj.getName(), null);
        appendText(result, lineSepar__, null);
      } // for ite
      
    } // if
    
    // put the comments
    appendText(result, lineSepar__, null);
    appendText(result, lineSepar__, null);
    appendText(result, "Note: User comments may be added and copied (Ctrl+C) like in a text editor. ",
            centric);
    appendText(result, "Access Road generates this text at each relevant property change."+
            " The user comments are lost at every text generation and every view saving. ",
            centric);

    dLink = null;
    acLink = null;

    blueBold = null;
    bold = null;
    centric = null;
    
    obj = null;
    entry = null;
    l_acLinks = null;
        
    return result;
  } // getNoThanWhyText()
  
  /**
   * Adds a text at the end of a JTextPane.
   * Called in the generic access paths search, then may be called
   * from a view worker thread.
   * @param _textPane to update
   * @param _text to add
   * @param _attributes associated to _text
   */
  public static synchronized final void appendText(JTextPane _textPane, String _text,
    SimpleAttributeSet _attributes) {
      Document doc = _textPane.getDocument();
      try {
        doc.insertString(doc.getLength(), _text, _attributes);
      } catch (BadLocationException b) {
        throw new InternalError("In FactoryUtilities.appendText, exception: "
          +b.toString());
      }
      doc = null;
  } // appendInText

  /**
  * Called by RightsFactory_Facade.getEffectiveRightsForPair().
  * @param _l_rights to analyze
  * @return the first accessing granting or denying right, or null if 
  *           all the rights are null or empty.
  * */
  public final StringRight withSomeAccess(Set<StringRight> _l_rights) {
      if (_l_rights == null) {

                      //for testing and tracking
                      //System.out.println("in FactoryUtilities.withSomeAccess(), "+
                        //"the argument is null. ");  
                      
          return null;
      }
      StringRight result = null;

                      //for testing and tracking
                      //System.out.println("in FactoryUtilities.withSomeAccess(), "+
                        //"rights: "); 
      
      for (Iterator ite = _l_rights.iterator(); ite.hasNext(); ) {
          result = (StringRight)ite.next();

                      //for testing and tracking
                      //System.out.println("----------"+result); 
          
          if (result != null && !result.isEmpty()) {
            break;
          }
      } // for ite
      
      return result;
  } // withSomeAccess()
  
  /**
  * Gets the standardized acronym of a right, if it exists for its metaright.
  * There are acronyms for the metarights
  * '|unknown_rights|', '|modify_from_parent_acs|', '|modify_owner|',
  * '|modify_security_descriptor|', '|is_superadmin|', '|read_from_parent_acs|',
  * '|get_target_rights|', '|unverified_rights|', '|transfer_all_rights_to_third_party|',
  * '|transfer_limited_rights_to_third_party|', '|transfer_all_rights|' and
  * '|transfer_limited_rights|'.
  * For these rights, the acronym is displayed in the GraphicView. The
  * 'SeeWhy' text uses the effective right which may be different if it is an ACS right.
  * @param _st ACS right or metaright
  * @return the acronym. May be null.
  **/
  public final String getRightAcronym(StringRight _st) {

      return (String)M_RIGHTS_ACRONYMS.get(_st.getEorMetaRight().getRight() );

  } // getRightAcronym()

   
  /**
  * Wrapper and caller of gBase.UtilityImpl.subtractNegativeNotRemovingInOneACS().
  * Called by RightsFactory_Facade.getEffectiveRightsForPair() and
  * getFirstShapeEffectiveRights(), 
  * DisplayablelinkUtilities.getDLinksFromNegativePaths(),
  * AclRightsFactoryImpl.getL_aclEntryRights().
  * @param _l_upd_rights to analyze. Not null. Updated by the method.
  * @return an all-positive or all-negative rights set.
  *     May be empty but not null.
  **/
  public final Set<StringRight> substractNegativeFromPositiveRights(
          Set<StringRight> _l_upd_rights) {
    // variables
    Set<StringRight>            l_result = null;
    Set<StringRight>            l_posRights  = null;
    Set<StringRight>            l_negRights = new HashSet(INITIAL_CAPACITY);    
    StringRight                 right = null;

                        // for testing and tracking    
                        System.out.println("");
                        System.out.println(
                                "FactoryUtilities.substractNegativeFromPositiveRights() start.");

    for (Iterator ito = _l_upd_rights.iterator(); ito.hasNext(); ) {
        right = (StringRight)ito.next();
        if (!right.isGrantingRight()) {
            l_negRights.add(right);
            ito.remove();
        } // if
    } // for ito
        
                      // for testing and tracking
                      System.out.println("-- all the positive rights before subtract(): ");
                      for (Iterator ite = _l_upd_rights.iterator(); ite.hasNext(); ) {
                          System.out.println("----- "+ite.next());
                      }
                      System.out.println("-- all the negative rights before subtract(): ");
                      for (Iterator ite = l_negRights.iterator(); ite.hasNext(); ) {
                          System.out.println("----- "+ite.next());
                      }

    // if no positive right, the negative rights are returned
    if (_l_upd_rights.isEmpty() && !l_negRights.isEmpty()) {
        right = null;
        return l_negRights;
    } // if    
    
    // no acs, so generic rights may be returned
    // no node, since the inheritance rights on node are not relevant
    l_posRights = baseUtil_.subtractNegativeNotRemovingInOneACS(
            _l_upd_rights, l_negRights, null, false);

    // if l_posRights is empty, the negative rights are stronger
    if (l_posRights.isEmpty() && !l_negRights.isEmpty()) {
        l_result = l_negRights;
    } // if
    else
        l_result = l_posRights;

                      // for testing and tracking
                      System.out.println("----- resulting rights: ");
                      for (Iterator ite = l_result.iterator(); ite.hasNext(); ) {
                          System.out.println("----- "+ite.next());
                      }
   
    l_posRights  = null;
    l_negRights = null; 
    right = null;
    
    return l_result;
  } // substractNegativeFromPositiveRights()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public final void finalizeForProcess() {
    interpreter_ = null;
    linkUtil_ = null;
    baseUtil_ = null;
    l_rights_.clear();
    l_rights_ = null;
  } // finalizeForProcess()

} // FactoryUtilities
