//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Collection;
import java.util.Collections;
import java.util.ArrayList;
import java.util.Arrays;

import ARoad0.CNot.*;

import ARoad0.gBase.PrivilegeAbst;
import ARoad0.gBase.UtilityImpl;

import ARoad0.gWorkInterface.*;
import ARoad0.ErrW.ProcessError;
import ARoad0.gBaseInterface.*;

/**
 * This class provides detection and filtering operations about links and privileges,
 * for a given ACS.
 * It operates on DisplayableLinkImpls, on the inherited rights and 
 * the typed privileges. An important service is the adding of a DisplayableLinkImpl
 * to the view map for these objects, and its filtering by
 * the AcsAddon subclasses (see the method addDisplayableLink).
 * Another important service is the application of priority rules to filter a link
 * after a comparison with the other links. This service may be defined for
 * the AGO, ACL or Privilege rights. An another service is the handling of the
 * inherited rights of a leaf through a tree. There is also the analysis of the
 * Account/Groups context of a node in a given DisplayableLinkImpl.
 * No direct call to the ACS methods. <P>
 * Each node in a view and each intermediate node has one LinkRights instance
 * for each view it belongs to. The implementation classes of this interface
 * have properties that depend only on the class of the node ACS,
 * then 2 nodes of the same view should use the same LinkRights instance,
 * even if it is not the actual behavior in all cases. But there is an exception
 * for the methods that detects the Account and AcsAddon priorities over all the links.
 * They use RightsMediatorImpl.getAllLinkRights() which returns
 * only one LinkRights instance per implied gWork package, and not one instance 
 * per view node. <P>
 * Each instance of LinkRights is called by only one thread.
 * A node may be in several DisplayableLinks of the same view,
 * but it has only one LinkRights instance for the view.
 * The association between a node and its LinkRights
 * instance is managed by the RightsMediatorImpl of the view. <P>
 * An AcsAddon may overread this class to change the implemented rules,
 * for instance it may be necessary to change the 'ThroughNodesTree' methods
 * for the rights inheritance. <P>
 * During a search, a DisplayableLinkImpl may have to be removed after its insertion
 * in the final result. The reason is the filtering of access paths to apply some priority rules
 * among the AGO, PRI, ACL or BDG rights. Some of these priority rules may be
 * from an AcsAddon, through the call to a method detectAddonPriorityInXXX().
 * These methods have the property to detect a prioritary link
 * which enforces the removing of one or several longuer links in the final result.
 * The inverse is not true, since a long link cannot produce, by its existence,
 * the removing of a shorter link. To overcome this limit in an AcsAddon, 
 * it is possible to use the 3 methods NodeRights.detectAddon2Nodes,
 * detectAddon3Nodes and detectAddonFinalAllNodes.
 * These methods detect some long links, and may remove
 * other links if there is a priority rule to fulfill with. However,
 * this way breaks the semantic of these NodeRights methods. The work is longer
 * and it would produce a code more difficult to maintain. <P>
 * Note : since the AGO rights are only inner rights of an ACS, and since all the nodes
 * of an ACS have LinkRight instances that have the same behavior, it is possible
 * for the AGO rights to use for a node, the LinkRight instance
 * of another node from the same ACS. This is also true for VirtualFolder
 * memberships, but not for the Bridge and Privilege rights. <P>
 * The calls to the ACS are for the methods getManageGroups(), getManageGroupTrees(),
 * getManageLPRIRightsInheritance(), getManageACLRightsInheritance(),
 * getManageAGORightsInheritance(), getManageSecGroupsForActor().
 * This class is used by the gWork and Gui2 classes.
 * @see ARoad0.CNot.AccessControlLinkImpl
 * @version 0.7.0
 */
public class LinkRightsImpl implements LinkRights {

  // table of the acronyms for the standardized rights
  private static Map  M_RIGHTS_ACRONYMS = new HashMap(INITIAL_CAPACITY);

  // --------------------- static initialization -----------------------------
  static {

    M_RIGHTS_ACRONYMS.put("|unknown_rights|", "|unknown_rights|");
    M_RIGHTS_ACRONYMS.put("|modify_from_parent_acs|", "|modify_from_parent_acs|");
    M_RIGHTS_ACRONYMS.put("|modify_owner|", "|modify_owner|");
    M_RIGHTS_ACRONYMS.put("|modify_security_descriptor|", "|modify_security_descriptor|");
    M_RIGHTS_ACRONYMS.put("|is_superadmin|", "|is_superadmin|");

  } // static

  protected AlgorithmInterpreter      interpreter_ = null;
  protected UtilityImpl               utility_ = null;
  protected DisplayableLinkUtilities  linkUtil_ = null;

  /**
  * Only one constructor, called by RightsMediatorImpl after the creation of _interpreter.
  * @param _interpreter algorithm interpreter of this view
  */
  public LinkRightsImpl(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      utility_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
  }

  /**
  * @param _obj Object to compare
  * @return false only if _obj is null or has not the same class, that is, _obj is
  * an instance of LinkRights from another (generic or AcsAddon) gWork package 
  */
  public boolean equals(Object _obj) {
      if (this == _obj)
          return true;
      if (_obj == null)
          return false;
      if (_obj.getClass() != getClass())
          return false;
      else
          return true;
  } // equals()

  /**
   * @return 0 as hashcode
   */
  public int hashCode() {
    //int result = super.hashCode();

    return 0;
  } // hashcode()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
      interpreter_ = null;
      utility_ = null;
      linkUtil_ = null;
  } // finalizeForProcess()

  /**
   * @return the String form of the class, not of the instance
   */
  public String toString() {
      return "LINK_RIGHTS of " + getClass().toString();
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop, and by detectOneHiddenNodeCompoundRights().
   * Overridden by the subclasses.
   * @return true if the detectAddonPriorityInAllLinksAGORights method provides
   * some work in this class.
   */
  public boolean getDetectAddonPriorityInAllLinksAGORights() {
      return false;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop, and by detectOneHiddenNodeCompoundRights().
   * Overridden by the subclasses.
   * @return true if the detectAddonPriorityInAllLinksNonAGORights method provides
   * some work in this class.
   */
  public boolean getDetectAddonPriorityInAllLinksNonAGORights() {
      return false;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the central loops, 
   * and by ThreeNodeRightsFactoryImpl.endsPathsFromGroupIDMemberrWithAclPrivilegeMemberOwnContain().
   * Overridden by the subclasses.
   * @return true if the detectAddonPriorityInLastLinksAGORights method provides
   * some work in this class.
   */
  public boolean getDetectAddonPriorityInLastLinksAGORights() {
      return false;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loops. Overridden by the subclasses.
   * @return true if the detectAddonPriorityInLastLinksNonAGORights method provides
   * some work in this class.
   */
  public boolean getDetectAddonPriorityInLastLinksNonAGORights() {
      return false;
  }

  /**
  * This important method adds a DisplayableLinkImpl in the map, if it does not exist,
  * or adds the rights of each AccessControlLink in the link, to the matched link in the map.
  * This is the main method to add a DisplayableLinkImpl to the map of a view,
  * and CompoundRightsFactoryImpl.detectAcsLinks() is the only method to not use it.
  * If the argument is not a weak link, it is put as the first link in the list
  * if the current first link is weak, since only a first weak link is ridden by GraphicEPRview.
  * The merging of rights is done by DisplayableLinkImpl.mergeRightsWithoutFiltering().<P>
  * Caution: the rights and the comments are merging if there is in the map a
  * DisplayableLinkImpl with the same nodes, and the result is then this previous link.
  * But there is no merging of the _dLink properties since the merging algorithm
  * depends on the semantic of the property. <P>
  * This method may be subclassed by the AcsAddon subclasses for filtering the adding
  * of a DisplayableLinkImpl following the previous DisplayableLinkImpls in the map.
  * Called by AgoRightsFactoryImpl, AclRightsFactoryImpl,
  * PrivilegeRightsFactoryImpl and CompoundRightsFactoryImpl.
  * Calls DisplayableLinkImpl.mergeRightsWithoutFiltering() and
  * mergeCommentsWithoutFiltering().
  * @see ARoad0.gWork.DisplayableLinkUtilities#getDisplayableLink(ARoad0.gBaseInterface.BaseObject, ARoad0.gBaseInterface.BaseObject, ARoad0.gBaseInterface.BaseObject[], java.util.Map) getDisplayableLink()
  * @param _dLink is the new link including some rights and to add to _upd_m_l_DisplayableLinks
  * @param _upd_m_l_DisplayableLinks Map of DisplayableLinks ArrayLists (one per pair) to update.
  * @return DisplayableLinkImpl in _upd_m_l_DisplayableLinks,
  * which may be _dLink, or the similar link in _upd_m_l_DisplayableLinks,
  * after the rights merging but without merging of the properties. Null if the adding
  * is rejected.
  **/
  public DisplayableLinkImpl addDisplayableLink(
     DisplayableLinkImpl _dLink, Map _upd_m_l_DisplayableLinks) {
        // variables
        DisplayableLinkImpl     result = null;
        List                    l_link;
        Name                    pairKey = ((Name)_dLink.getEorFirstEnd().getName()).extended(
                                            _dLink.getEorSecondEnd().getName());

        result = linkUtil_.getDisplayableLink(
          _dLink.getEorFirstEnd(),
          _dLink.getEorSecondEnd(),
          _dLink.getEorNodesPath(),
          _upd_m_l_DisplayableLinks);


                                  //for testing and tracking
                                  //System.out.println("LinkRightsImpl."+
                                    //"addDisplayableLink(), the initial result is "+result);
                                  //System.out.println("--- _dLink path is:");
                                  //if (_dLink.getEorNodesPath() != null) {
                                    //for (int i = 0; i < _dLink.getEorNodesPath().length ; i++) {
                                      //System.out.println("------- for "+i+": "+((BaseObject)_dLink.getEorNodesPath()[i]).getName());
                                    //}
                                  //} // if

        if (result == null) {
          // updates _upd_m_l_DisplayableLinks
          if (!_upd_m_l_DisplayableLinks.containsKey(pairKey) ) {
            l_link = new ArrayList(INITIAL_CAPACITY/3);
            l_link.add(_dLink);
            _upd_m_l_DisplayableLinks.put(pairKey, l_link);
          }
          else {
            // add _dLink in the current list
            l_link = (List)_upd_m_l_DisplayableLinks.get(pairKey);
            if (!_dLink.isWeak() && ((DisplayableLinkImpl)l_link.get(0)).isWeak())
                l_link.add(0, _dLink);
            else
                l_link.add(_dLink);
          } // else
          result = _dLink;
        }
        else { // add the _dLink rights to result
          
          // main job on all rights and comments, on the activated types only,
            // and do not merge the properties
          result.mergeRightsWithoutFiltering(_dLink);
          // all comments are added, even if the relevant type is not activated
          result.mergeCommentsWithoutFiltering(_dLink);

          l_link = (List)_upd_m_l_DisplayableLinks.get(pairKey);
          if (!_dLink.isWeak() && ((DisplayableLinkImpl)l_link.get(0)).isWeak()) {
              l_link.remove(result);
              l_link.add(0, result);
          } // if (_dLink.isWeak())
        } // else

        l_link = null;
        pairKey = null;

                                  //for testing and tracking
                                  //System.out.println("LinkRightsImpl."+
                                    //"addDisplayableLink(), the final result is "+result);
        return result;
  } // addDisplayableLink()

  /**
  * Applies the rule 'AGO right: A rights overlay G rights' to a DisplayableLink
  * that ends with the sequence source/groupIDMember/resource, through the analysis
  * of a list of DisplayableLinkImpls. This is the LinkRightsImpl instance of the last resource. <P>
  * If the last sequence is source/userID/resource in _dLink, this method removes
  * any DisplayableLink in the list containing a sequence source/groupID/resource
  * with the same source and resource, and the method returns false to say
  * there is no priority that forbids the use of the argument _dLink
  * as a valid link. Otherwise, with this last sequence, the method returns true. <P>
  * If the last sequence is source/groupID/resource in _dLink, and if there is
  * a DisplayableLink in the list containing a sequence source/userID/resource
  * with the same source and resource, then the method returns true to say
  * there is a priority that forbids the use of the argument _dLink
  * as a valid link. Otherwise, with this last sequence, the method returns false. <P>
  * The OWNER path may be without workable rights, since the account remains there
  * the single applicable relation.
  * Called by AgoRightsFactoryImpl.detectHiddenChainedGroupsRights().
  * @param _dLink with a Resource as as node of this instance and as second end, and
  * a groupIDMember just before the second end. Returns true
  * if null, empty or with less than 3 nodes. Its AccessControlLinks are not analyzed.
  * @param _upd_l_dLinks to analyze and update, all with the same second end than _dLink.
  * Returns false if it is null or empty. It may be updated only if _dLink contains
  * an UserID as node before the last node. The last AccessControlLink of each
  * DisplayableLinkImpl is analyzed for the types OWNER and CONTAIN. May contain
  * the argument _dLink that is then ignored. May be null.
  * @return true if _dLink is not a valid link
  * @exception InternalError with a list having links without the same _dLink ends, or
   * _dLink does not have a groupIDMember as node just before the last one
  **/
  public boolean detectAccountPriorityInLastLinkAGORights(
    DisplayableLinkImpl _dLink, Collection<DisplayableLinkImpl> _upd_l_dLinks) {

    if (_dLink == null || _dLink.isEmpty() || _dLink.noOfNodes() < 3)
        return true;

    if (!(_dLink.getEorNodeJustBeforeTheSecondEnd() instanceof ImmutableGroupIDMember))
        throw new InternalError("RightsFactoryImpl.detectPriorityInLastLinkAGORights() where"+
                " the argument _dLink does not have a groupIDMember as node just before the last one.");

    if (_upd_l_dLinks == null || _upd_l_dLinks.isEmpty())
        return false;

    ImmutableName           firstEnd = ((ACSObject)_dLink.getEorFirstEnd()).getName();
    ImmutableName           secondEnd =  (Name)((ACSObject)_dLink.getEorSecondEnd()).getName();
    ImmutableGroupIDMember  member = (ImmutableGroupIDMember)_dLink.getEorNodeJustBeforeTheSecondEnd();
    ACSObject               source = (ACSObject)_dLink.getEorAllNodes()[_dLink.noOfNodes() - 3];
    DisplayableLinkImpl     dLink = null;
    boolean                 result = false;
    boolean                 error = false;

    //if (!source.getAcsName().equals(member.getAcsName())) {
     //   firstEnd = null;
       // secondEnd = null;
        //source = null;
        //member = null;
        //return false;
    //} // if
    //if (!source.getAcsName().equals(((ACSObject)_dLink.getEorSecondEnd()).getAcsName())) {
      //  firstEnd = null;
        //secondEnd = null;
        //source = null;
        //member = null;
        //return false;
    //} // if

    // ACS rule: 'AGO right: A rights overlay G rights'
    for (Iterator itg = _upd_l_dLinks.iterator() ; itg.hasNext() ; ) {
          dLink = (DisplayableLinkImpl)itg.next();

          if (!((ACSObject)dLink.getEorFirstEnd()).getName().equals(firstEnd)) {
              error = true;
              break;
          }
          if (!((ACSObject)dLink.getEorSecondEnd()).getName().equals(secondEnd)) {
              error = true;
              break;
          }

          if (dLink.noOfNodes() < 3 || dLink.noOfNodes() < _dLink.noOfNodes() - 2 ||
                  source != dLink.getEorAllNodes()[_dLink.noOfNodes() - 3])
              continue;

          if ((member instanceof ImmutableGroupID) &&
                  dLink.getEorLastLink().getLinkTypeValue(
                    AccessControlLink.OWNER)) {
              // dLink will not be processed since there is an user right.
              // works even if dLink has no owner rights.
              result = true;
              break;
          } // if
          else if ((member instanceof ImmutableUserID) &&
                  dLink.getEorLastLink().getLinkTypeValue(
                    AccessControlLink.CONTAIN)) {
              // even if the owner rights are null,
              // all the dLink with group rights are deleted.
              itg.remove();
          } // if
    } // for itg

    if (error) {
        dLink = null;
        firstEnd = null;
        secondEnd = null;
        throw new InternalError("LinkRightsImpl.detectAccountPriorityInLastLinkAGORights() with"+
                " a list having links without the _dLink 2 ends.");
    } // if

    dLink = null;
    firstEnd = null;
    secondEnd = null;
    member = null;
    source = null;
    return result;
  } // detectAccountPriorityInLastLinkAGORights()

  /**
  * This heavy method is overridden by the AcsAddon, to remove any DisplayableLink
  * which has an AcsAddon object, and contains an AGO right
  * while there is a superior right in another DisplayableLink, following
  * the applicable ACS rules of the AcsAddon ACS.
  * Caution: a smaller DisplayableLink may be in a list after a longuer DisplayableLink.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the final
  * search if getDetectAddonPriorityInAllLinksAGORights() returns true.
  * Overridden method.
  * @param _l_dLinks DisplayableLinkImpls to analyze, all with the same two ends or not,
  * where this LinkRights is for one of the node in at least one DisplayableLinkImpl.
  * Not null.
  * @return the updated argument
  **/
  public Collection detectAddonPriorityInAllLinksAGORights(
          Collection<DisplayableLinkImpl> _l_dLinks) {

                      // for testing and tracking
                      //System.out.println("LinkRightsImpl.detectPriorityInAllLinksAGORights() dLinks at the end:");
                      //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}

      return _l_dLinks;
  } // detectAddonPriorityInAllLinksAGORights()

  /**
  * This heavy method is overridden by the AcsAddon, to remove any DisplayableLink
  * which has an AcsAddon object and contains an ACL right or a Privilege right,
  * while there is a superior right in another DisplayableLink, following
  * the applicable ACS rules of the AcsAddon ACS.
  * Caution: a smaller DisplayableLink may be in a list after a longuer DisplayableLink.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the final
  * search if getDetectAddonPriorityInAllLinksNonAGORights() returns true.
  * Overridden method.
  * @param _l_dLinks DisplayableLinkImpls to analyze, all with the same two ends or not,
  * where this LinkRights is for one of the node in at least one DisplayableLinkImpl.
  * Not null
  * @return the updated argument
  **/
  public Collection detectAddonPriorityInAllLinksNonAGORights(
          Collection<DisplayableLinkImpl> _l_dLinks) {

                      // for testing and tracking
                      //System.out.println("LinkRightsImpl.detectAddonPriorityInAllLinksNonAGORights() dLinks at the end:");
                      //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}

      return _l_dLinks;
  } // detectAddonPriorityInAllLinksNonAGORights()

  /**
  * Overridden by the AcsAddon, to applies any priority rules to a DisplayableLink,
  * about an ACL right or a Privilege right in the last AccessControlLink,
  * through the analysis of a list of DisplayableLinkImpls. The Privilege rights
  * on a VirtualFolder is an exception, since there is no call to this method for them
  * in the xxxRightsFactoryImpl generic methods. But the near method
  * detectAddonPriorityInAllLinksNonAGORights() may be applied to the Privilege
  * rights on a VirtualFolder.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the central
  * loop if getDetectAddonPriorityInLastLinksNonAGORights() returns true,
  * for the central iteration and the final search,
  * and for a collection of DisplayableLinkImpls having from 2 to N+2 nodes.
  * For instance, this method may be overridden to detect the actors connected to the analyzed
  * groupidmember through an executing acl, when these actors do not transfer the AG context.
  * @param _dLink where the second end is the node of this instance.
  * Returns true if null, empty or with less than 3 nodes.
  * @param _upd_l_dLinks to analyze and update, all with the same second end than _dLink.
  * Returns false if it is null or empty. It may be updated. May contain
  * the argument _dLink that is then ignored. May be null.
  * @return true if _dLink is not valid
  **/
  public boolean detectAddonPriorityInLastLinkNonAGORights(
    DisplayableLinkImpl _dLink, Collection<DisplayableLinkImpl> _upd_l_dLinks) {

      return false;
  }

  /**
  * Overridden by the AcsAddon, to apply the ACS priority rules to a DisplayableLink
  * that ends with an access target, for a given Addon ACS that owns this target,
  * through the analysis of a list of DisplayableLinkImpls.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the central
  * loop if getDetectAddonPriorityInLastLinksAGORights() returns true,
  * and for the central iteration and the final search,
  * for a collection of DisplayableLinkImpls having from 2 to N+2 nodes.
  * Overridden method.
  * @param _dLink where the second end is the node of this instance.
  * @param _upd_l_dLinks to analyze and update, all with the same two ends of _dLink.
  * Returns false if it is null or empty. It may be updated. May contain
  * the argument _dLink that is then ignored. May be null.
  * @return true if _dLink is not valid
  * @exception InternalError with some ACS-specific errors
  **/
  public boolean detectAddonPriorityInLastLinkAGORights(
    DisplayableLinkImpl _dLink, Collection<DisplayableLinkImpl> _upd_l_dLinks) {

    return false;
  } // detectAddonPriorityInLastLinkAGORights()

  /**
  * This heavy method removes any DisplayableLink which contains an Account/Group Group
  * relation while there is an Account in another DisplayableLink.
  * Applies the rule 'AGO right: A rights overlay G rights' to a DisplayableLink
  * that ends with the sequence source/groupIDMember/resource.
  * The 'owner' path may be without workable rights. The DisplayableLinkImpl analysis
  * is based on the identity of the nodes in 2 links ('==' operator).
  * The 2 types of DisplayableLinkImpl to remove are listed hereafter: <P>
  * - in dLink to remove, there is a node triplet (ep)-RUN_UNDER or IS_MEMBER-
  * (groupid)-CONTAIN-(resource), and in some dLink2, there is a node triplet
  * (same ep)-RUN_UNDER-(userid)-OWNER-(same resource); <P>
  * - in dLink to remove, there is a node triplet (userid)-IS_MEMBER-(groupid)-CONTAIN-(resource),
  * and in some dLink2, there is a node couple (same userid)-OWNER-(same resource).<P>
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the final loop.
  * Calls removeSomeGroupRightsInAllLinksAGORights().
  * @param _l_dLinks DisplayableLinkImpls to analyze, all items with the same two ends,
  * where this LinkRights is for one of the node in at least one DisplayableLinkImpl.
  * Not null.
  * @return the updated argument.
  **/
  public Collection detectAccountPriorityInAllLinksAGORights(
          Collection<DisplayableLinkImpl> _l_dLinks) {

                      // for testing and tracking
                      //System.out.println("LinkRightsImpl.detectAccountPriorityInAllLinksAGORights() dLinks in the argument:");
                      //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                        // System.out.println("----- "+ite.next());
                      //}

      // variables
    List                    l_dLinks2 = new ArrayList(_l_dLinks);
    Map                     m_ownerLinkTypePositions = new HashMap(INITIAL_CAPACITY);
    Map                     m_containLinkTypePositions = new HashMap(INITIAL_CAPACITY);
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink2 = null;
    BaseObject[]            l_obj = null;
    BaseObject[]            l_obj2 = null;
    boolean                 toRemove = false;
    List                    l_owner = null;
    List                    l_contain = null;
    int                     ownIndex = -1;
    int                     conIndex = -1;

    // initialize
    for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
        dLink = (DisplayableLinkImpl)ite.next();
        m_ownerLinkTypePositions.put(dLink.hashCode(),
                dLink.getOwnerLinkTypePositions());
        m_containLinkTypePositions.put(dLink.hashCode(),
                dLink.getContainLinkTypePositions());
    } // for ite

                  // for testing and tracking
                  //System.out.println("LinkRightsImpl.detectAccountPriorityInAllLinksAGORights() m_ownerLinkTypePositions:");
                  //for (Iterator ite = m_ownerLinkTypePositions.entrySet().iterator(); ite.hasNext(); ) {
                    //  System.out.println("----- "+ite.next());
                  //}

                  // for testing and tracking
                  //System.out.println("LinkRightsImpl.detectAccountPriorityInAllLinksAGORights() m_containLinkTypePositions:");
                  //for (Iterator ite = m_containLinkTypePositions.entrySet().iterator(); ite.hasNext(); ) {
                    //  System.out.println("----- "+ite.next());
                  //}

    for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
        dLink = (DisplayableLinkImpl)ite.next();
        toRemove = false;

        // processing even if one dLink node does not apply the AGO priorities

        l_contain = (List)m_containLinkTypePositions.get( dLink.hashCode() );
        if (l_contain == null) continue;

        for (Iterator ita = l_dLinks2.iterator(); ita.hasNext(); ) {
             dLink2 = (DisplayableLinkImpl)ita.next();
             if (dLink2 == dLink) continue;
             l_owner = (List)m_ownerLinkTypePositions.get( dLink2.hashCode() );

             // to remove dLink, detect an owner relation in dLink2.
             // rule 'AGO right: A rights overlay G rights'
             if (l_owner != null) {
                 l_obj = dLink.getEorAllNodes();
                 l_obj2 = dLink2.getEorAllNodes();

                                            // for testing and tracking
                                      //System.out.println("LinkRightsImpl.detectAccountPriorityInAllLinksAGORights() "+
                                        //      "at the start of the ita loop on l_dLinks2, l_obj2 is not null: ");
                                      //for (int k = 0; k < l_obj2.length; k++) {
                                        //  System.out.println("-------- "+l_obj2[k]);
                                      //}

                 for (Iterator ito = l_owner.iterator(); ito.hasNext(); ) {
                     ownIndex = ((Integer)ito.next()).intValue();

                     if (l_contain != null) {
                         // analyze the first contain relation.
                         // the criteria is comparing 3-nodes dLinks. The cases are:
                         // in dLink, triplet (ep)-RUN_UNDER or IS_MEMBER-
                         //                 (groupid)-CONTAIN-(resource)
                         // in dLink2, triplet (ep)-RUN_UNDER-(userid)-OWNER-(resource)
                         for (Iterator itc = l_contain.iterator(); itc.hasNext(); ) {
                             conIndex = ((Integer)itc.next()).intValue();
                             // The value at the index i of l_contain = dLink.getContainLinkTypePositions()
                             // is connected to the nodes getEorAllNodes()[i] and getEorAllNodes()[i+1].
                             if (conIndex > 0 && ownIndex > 0 &&
                                     l_obj[conIndex+1] instanceof ImmutableResource &&
                                     l_obj[conIndex - 1] == l_obj2[ownIndex - 1] &&
                                     l_obj[conIndex+1] == l_obj2[ownIndex+1]) {
                                 // dLink2 has a owner relation on the same nodes pair
                                 // that dLink has for a contain link for the same ep and resource:
                                 // remove dLink
                                 toRemove = true;

                                      // for testing and tracking
                                      //System.out.println("LinkRightsImpl.detectAccountPriorityInAllLinksAGORights() "+
                                        //"a contain link at "+conIndex+" implies the removing of the dLink "+dLink+
                                        //" since there is a similar owner link at "+ownIndex+" in "+dLink2);

                             }
                             if (toRemove) break;
                         } // for itc

                         // analyze the second contain relation.
                         // the criteria is comparing a 3-nodes dLink and a 2-nodes dLink. The cases are:
                         // in dLink, triplet (userid)-IS_MEMBER-
                         //                 (groupid)-CONTAIN-(resource)
                         // in dLink2, couple (userid)-OWNER-(resource)
                         for (Iterator itc = l_contain.iterator(); itc.hasNext(); ) {
                             conIndex = ((Integer)itc.next()).intValue();
                             // The value at the index i of l_contain = dLink.getContainLinkTypePositions()
                             // is connected to the nodes getEorAllNodes()[i] and getEorAllNodes()[i+1].
                             if (!toRemove && conIndex > 0 &&
                                     l_obj[conIndex+1] instanceof ImmutableResource &&
                                     l_obj[conIndex - 1] == l_obj2[ownIndex] &&
                                     l_obj[conIndex+1] == l_obj2[ownIndex+1]) {
                                 // dLink2 has a owner relation on the same nodes pair
                                 // that dLink has for a contain link for the same resource:
                                 // remove dLink
                                 toRemove = true;

                                      // for testing and tracking
                                      //System.out.println("LinkRightsImpl.detectAccountPriorityInAllLinksAGORights() "+
                                        //"a contain link at "+conIndex+" implies the removing of the dLink "+dLink+
                                        //" since there is a more direct owner link at "+ownIndex+" in "+dLink2);

                             }
                             if (toRemove) break;
                         } // for itc
                     } // if ((l_contain != null
                     if (toRemove) break;

                 } // for ito
             } // if (l_owner2

             l_obj = null;
             l_obj2 = null;
             if (toRemove) break;
        } // for ita

        if (toRemove) {

            ite.remove();
            // l_dLinks2 is also simplified to speed up the processing
            l_dLinks2.remove(dLink);

        } // if
    } // for ite

    l_dLinks2 = null;
    dLink = null;
    dLink2 = null;
    l_obj = null;
    l_obj2 = null;
    l_contain = null;
    m_ownerLinkTypePositions = null;
    m_containLinkTypePositions = null;

                      // for testing and tracking
                      //System.out.println("LinkRightsImpl.detectAccountPriorityInAllLinksAGORights() dLinks at the end:");
                      //for (Iterator ite = _l_dLinks.iterator(); ite.hasNext(); ) {
                        //  System.out.println("----- "+ite.next());
                      //}

    return _l_dLinks;
  } // detectAccountPriorityInAllLinksAGORights()

  /**
   * This important method updates the Account/Groups context of a DisplayableLinkImpl,
   * either from the AG context of the last node having one, or from one of
   * the following nodes in the same ACS, or, if there is no current AG context,
   * through all the connected nodes from the same ACS.
   * As generic rules, the internal AclEntries propagate the AG context so the local
   * AG context of the ACL target is not applied, while the Privileges does not propagate
   * the generic AG context. For instance, this method allows to set the rules for a 3-nodes path
   * 'ep/acl/actor started from xid/alias/virtual folder'.
   * If the intermediate actor runs under the account U and the group G, the generic rule
   * is that the rights of 'ep' on the virtual folder members depend on the U and G rights,
   * and does not depend on the 'ep' AG context.
   * The AG context is a set of GroupIDMembers that may contain, first,
   * an optional couple (UserID, GroupID), and, second, a list of optional GroupIDs. <P>
   * This method has to be called before the use of the AG context of the second end.
   * This means that it the DisplayableLinkImpl has to be updated each time the paths search
   * finds an actor running under its own AG context, or if the context is no more propagated.
   * It is recommanded because the AG context may be needed for analyzing the next nodes
   * of the path. This calling rule is also true for the AcsAddon path search methods.
   * If it is done, any AcsAddon paths search method may read the current AG context
   * of a DisplayableLinkImpl passed in as argument. If it is not done in an AcsAddon,
   * the AcsAddons of the next nodes should call the generic method
   * DisplayableLinkUtilities.getSecondEndAGcontext() to set the AG context properly. <P>
   * If the DisplayableLinkImpl contains only nodes from one ACS, there is one
   * LinkRights instance to use. One call to this method provides a complete result.
   * Otherwise, if the nodes are from different ACSs, the caller of this method
   * has to define a sequence of calls to this method on different LinkRights instances.
   * The aim is to process each node in the right instance and the proper order, and
   * this is done by DisplayableLinkUtilities.getSecondEndAGcontext().
   * In the returned DisplayablelinkImpl, the AG context current position is
   * on the last node to have been analyzed. This position is ridden
   * by the caller to define the next LinkRights instance on which this method
   * has to be called, for the next node(s) to analyze. When this current position is
   * the second end of the DisplayableLinkImpl, the work is completed. <P>
   * Called by ThreeNodesRightsFactoryImpl.selectOneHiddenNodePathsForVirtualFolder(),
   * addPathsFromActorStartsUnderGroupIDMember(), and by
   * DisplayableLinkUtilities.getSecondEndAGcontext().
   * Calls setNextNodeAGrunningContext() that produces the effective work. <P>
   * Overridden method in the AcsAddons. Caution: the overridding in a subclass
   * should work only on the nodes of its AcsAddon.
   * The overridding method may simply delete the AG context of the last analyzed node
   * from a generic ACS or from another AcsAddon.
   * @see ARoad0.gWork.DisplayableLinkUtilities#getSecondEndAGcontext(ARoad0.CNot.DisplayableLinkImpl)
   * @see #setNextNodeAGrunningContext(ARoad0.CNot.DisplayableLinkImpl, int) 
   * @param _upd_dLink the node after the node having the AG context, if any, is the node
   * of this instance. Never null. Its properties should be updated by this method,
   * and they may be null after the updating. The AG context position is set to 
   * the last node having been analyzed.
   * @return set of one UserID and some GroupIDs, as the AG context of the second end
   * of the argument. May be null but never empty.
   * @exception ProcessError if a node is not an ACSObject, if an AccessControLink
   * is applied to a node that has not the right class.
  * */
  public Set<ImmutableGroupIDMember> updateAGrunningContext(
          DisplayableLinkImpl _upd_dLink) throws ProcessError {

                            // for testing and tracking 
                          //if (((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("Amounts view"))
                              //System.out.println("LinkRightsImpl.updateAGrunningContext() starts for "+
                                //  _upd_dLink);

    if (_upd_dLink == null)
        throw new ProcessError("in LinkRightsImpl.updateAGrunningContext(), "+
                        "null DisplayableLinkImpl. ");

    Map             m_properties = _upd_dLink.getProperties();
    boolean         error = false;
    int             currentAGcontextPosition = 0;
    BaseObject[]    l_nodes = _upd_dLink.getEorAllNodes();
    HashSet<ImmutableGroupIDMember> l_result = null;

    if (m_properties != null &&
            m_properties.get(DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) != null ) {

        l_result = (HashSet) m_properties.get(DisplayableLinkImpl.AG_CONTEXT_SET);

        currentAGcontextPosition = ((Integer)m_properties.get(
                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).intValue();
    } // if

                        // for testing and tracking 
                      //if ( ((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("to_ViFi") )
                          //System.out.println("LinkRightsImpl.updateAGrunningContext()"+
                            //  " starts at "+currentAGcontextPosition);

    if (l_nodes.length > currentAGcontextPosition + 1 &&
            (!(l_nodes[currentAGcontextPosition] instanceof ACSObject) ||
            !(l_nodes[currentAGcontextPosition + 1] instanceof ACSObject))) {
        throw new ProcessError("in LinkRightsImpl.updateAGrunningContext()"+
                        ", a node is not an ACSObject. ");
    } // if
    if (l_nodes.length > currentAGcontextPosition + 1 &&
            (!((ACSObject)l_nodes[currentAGcontextPosition]).getAcsName().equals(
            ((ACSObject)l_nodes[currentAGcontextPosition + 1]).getAcsName()))) {

                                // for testing and tracking 
                                //System.out.println(
                                  //      "LinkRightsImpl.updateAGrunningContext() between 2 ACS, "+
                                  //"l_nodes[currentAGcontextPosition]: "+
                                  //l_nodes[currentAGcontextPosition]);

        // for instance, 'bridge' does not transfer the AG context in this generic class
        if (!isTransferingAGcontext(
                _upd_dLink.getEorConnectedLinks(l_nodes[currentAGcontextPosition + 1])[0],
                ((ACSObject)l_nodes[currentAGcontextPosition]).getEorACS(),
                ((ACSObject)l_nodes[currentAGcontextPosition + 1]).getEorACS())) {

            // clear the AG context for the node in the second ACS,
            // and returns null to open a new AG context analysis
            _upd_dLink.addProperty(
                        DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION,
                        new Integer(currentAGcontextPosition + 1));
            _upd_dLink.removeProperty(DisplayableLinkImpl.AG_CONTEXT_SET);
            l_result = null;
        } // if (isTransferingAGcontext()
        else {
            // increment the AG context location on the node in the second ACS,
            // and returns the current AG context
            _upd_dLink.addProperty(
                        DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION,
                        new Integer(currentAGcontextPosition + 1));
        } // else

        l_nodes = null;
        m_properties = null;

                            // for testing and tracking 
                            //System.out.println("end of LinkRightsImpl.updateAGrunningContext()"+
                              //      " for "+_upd_dLink);
                            //if (l_result != null) {
                              //  System.out.println("-- with result:");
                                //for (Object obj : l_result) {
                                  //  System.out.println("------ "+obj);
                                //}
                            //}
                            //else {
                              //  System.out.println("-- with null result.");
                            //}
        
        return l_result;
    } // if

    // from currentAGcontextPosition, analyze all the nodes with the same ACS
    // then return the result of the updated DisplayableLinkImpl
    for (int i = currentAGcontextPosition; i < l_nodes.length - 1; i++) {
        if (!(l_nodes[i] instanceof ACSObject) ||
                !(l_nodes[i + 1] instanceof ACSObject)) {
            error = true;
            break;
        } // if
        if (!((ACSObject)l_nodes[i]).getAcsName().equals(
                ((ACSObject)l_nodes[i + 1]).getAcsName())) {

            // if i != 0, already done in setNextNodeAGrunningContext()
            if (i == 0)
                _upd_dLink.addProperty(
                    DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION,
                    new Integer(i));

            break;

        } // if

                            // for testing and tracking
                            //System.out.println("LinkRightsImpl.updateAGrunningContext()"+
                              //      " calls setNextNodeAGrunningContext() for _nodeIndex = "+
                                //    i+1);

        l_result = setNextNodeAGrunningContext(_upd_dLink, i + 1);

                            // for testing and tracking 
                            //System.out.println("LinkRightsImpl.updateAGrunningContext()"+
                              //      " after setNextNodeAGrunningContext() for _nodeIndex = "+
                                //    i+1);
                            //if (l_result != null) {
                              //  System.out.println("-- with result:");
                                //for (Object obj : l_result) {
                                  //  System.out.println("------ "+obj);
                                //}
                            //}
                            //else {
                              //  System.out.println("-- with null result.");
                            //}

    } // for i
    
    m_properties = null;
    l_nodes = null;

    if (error) {
        throw new ProcessError("in LinkRightsImpl.updateAGrunningContext()"+
                        ", a node is not an ACSObject. ");
    }

                            // for testing and tracking 
                            //System.out.println("end of LinkRightsImpl.updateAGrunningContext()"+
                              //      " for "+_upd_dLink);
                            //if (l_result != null) {
                              //  System.out.println("-- with result:");
                                //for (Object obj : l_result) {
                                  //  System.out.println("------ "+obj);
                                //}
                            //}
                            //else {
                              //  System.out.println("-- with null result.");
                            //}

    return l_result;
  } // updateAGrunningContext()

  /**
   * This method increments the Account/Groups context position, updates the AG context set
   * in the DisplayableLinkImpl and returns this context.
   * For instance, this method allows to implement that for a 3-nodes path
   * 'ep/privilege/actor started from xid/alias/virtual folder',
   * if the intermediate actor runs under the account U and the group G, the generic rule
   * is that the rights of 'ep' on the virtual folder members depend on the U and G rights,
   * and does not depend on the 'ep' AG context.
   * The AG context is a set of GroupIDMembers that may contain, first,
   * an optional couple (UserID, GroupID), and, second, a list of optional GroupIDs.
   * The AG context is returned by this method. It is also stored in the argument. <P>
   * The couple of (UserID, GroupID) contains the 2 GroupIDMembers under which
   * the second end is running, if it is an actor, or which is its AG context as
   * resource or virtual folder.
   * The UserID is given by the last UserID node in the DisplayableLinkImpl
   * that is the source for an AccessControlLink with
   * the types OWNER and through a chain of links that transfers the AG context.
   * The GroupID is given by the last GroupID node in the DisplayableLinkImpl
   * that is the source for an AccessControlLink with the types CONTAIN,
   * and through a chain of links that transfers the AG context. <P>
   * The list of GroupIDs depends on the first node of the DisplayableLinkImpl,
   * if the chain of links transfers the AG context.
   * If the first end is a GroupIDMember, the list contains the groups to which
   * this GroupIDMember is member, if any. 
   * If the first end is an Actor, the list contains the secondary groups of
   * this Actor, if any. <P>
   * If the chain of links does not transfer the first end AG context, the list of
   * GroupIDMembers is from the last intermediate Actor starting without transfered
   * AG context. With the exemple of the 3-nodes path having the form
   * 'ep/privilege/actor started from xid/alias/virtual folder',
   * if the intermediate actor runs under the account U and the group G, 
   * the virtual folder in the path is on the mode VirtualFolderImpl.AG_CONTEXT_DEPENDENT,
   * and the path has to be extended by two paths 'run under/U' and 'run under/G'.
   * Otherwise, if the actor does not run under U, no running group is considered,
   * the virtual folder has the mode PASS_ON, and the 'ep' rights 
   * on the virtual folder members are defined by the other properties of the path
   * 'ep/privilege/actor started from xid/alias/virtual folder/virtual member/resource'. <P>
   * The right proxy property of the actor is never considered. It is ridden
   * in other core algorithms to allow an actor to be an intermediate node,
   * whatever the AG context. <P>
   * Called by updateAGrunningContext(). The caller must have checked
   * the following requisites: <P>
   * - the current node with the AG context is an ACSObject, like the next one, <P>
   * - these two nodes have the same ACS, <P>
   * - the current node with the AG context has the index just before _nodeIndex. <P>
   * The AccessControlLink between theses two implied nodes may allow or not to transfer
   * the AG context, and this is given by isTransferingAGcontext().
   * The returned DisplayableLinkImpl may have a null property for the key
   * DisplayableLinkImpl.AG_CONTEXT_SET, but never for AG_CONTEXT_NODE_POSITION. <P>
   * Overridden method in the AcsAddons, for instance to provide AG context transferings
   * through the privileges. Another case is to reset the AG context of a no-simple
   * DisplayableLinkImpl when an AccessControlLink transfers a null AG context between
   * two nodes.
   * The overriding method may simply delete the AG context of the last analyzed node
   * from a generic ACS or from another AcsAddon.
   * Caution: the overriding in an AcsAddon should work only on nodes from its AcsAddon.
   * Called by updateAGrunningContext().
   * @see CompoundRightsFactoryImpl#selectDirectHiddenReferenceForAlias(ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, ARoad0.gBaseInterface.ACSObject)
   * @see #isTransferingAGcontext(ARoad0.CNot.AccessControlLinkImpl, ARoad0.gBaseInterface.ImmutableACS, ARoad0.gBaseInterface.ImmutableACS)
   * @param _upd_dLink its second end is the node of this instance. Never null.
   * Its properties should be updated by this method, and they may be null after it. 
   * The AG context position is set to the node at _nodeIndex.
   * @param _nodeIndex index of the node to analyze; is the node of this instance.
   * Always greater than 0.
   * @return a clone of the updated AG context, that is a set of one applicable UserID and
   * some GroupIDs, like those under which the last actor is running in its ACS.
   * May be null but never empty.
   * @exception ProcessError if the position of the node to analyze is lesser than 1,
   * and varied cases
   */
  protected HashSet<ImmutableGroupIDMember> setNextNodeAGrunningContext(
          DisplayableLinkImpl _upd_dLink, int _nodeIndex) throws ProcessError {

    if (_upd_dLink == null)
        throw new ProcessError("in LinkRightsImpl.setNextNodeAGrunningContext()"+
                        ", null DisplayableLinkImpl. ");
    if (_nodeIndex < 1)
        throw new ProcessError("in LinkRightsImpl.setNextNodeAGrunningContext()"+
                        ", the position of the node to analyze is lesser than 1.");

                                        // for testing and tracking
                                        //if (((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("Amounts view"))
                                            //System.out.println("LinkRightsImpl.setNextNodeAGrunningContext() starts"+
                                              //  " with _nodeIndex: "+_nodeIndex+" for "+_upd_dLink);

    HashSet<ImmutableGroupIDMember> l_result = null;
    GroupIDMember               member = null;
    Actor                       actor = null;
    ImmutableGroupID[]          l_Grps = null;

    if (_nodeIndex == 1) {
        // if the AG context is transfered, the first end sets the context
        if (isTransferingAGcontext(_upd_dLink.getEorFirstLink(), null, null)) {

            if (_upd_dLink.getEorFirstEnd() instanceof GroupIDMember) {
                member = (GroupIDMember)_upd_dLink.getEorFirstEnd();
                
                                        // for testing and tracking
                                        //System.out.println("LinkRightsImpl.setNextNodeAGrunningContext()"+
                                          //      "adds in the AG context, the GroupIDMember "+member);


                l_result = new HashSet(10);

                l_result.add(member);
                l_Grps = member.getEorL_GroupID();
                if (l_Grps.length != 0)
                    l_result.addAll(Arrays.asList(l_Grps));

            } // if GroupIDMember
            else if (_upd_dLink.getEorFirstEnd() instanceof Actor) {
                actor = (Actor)_upd_dLink.getEorFirstEnd();

                l_result = new HashSet(10);

                if (actor.getEorCurrentUserID() != null)
                    l_result.add(actor.getEorCurrentUserID());
                if (actor.getEorCurrentGroupID() != null)
                    l_result.add(actor.getEorCurrentGroupID());

                l_Grps = actor.getEorL_SecondaryGroup();
                if (l_Grps.length != 0)
                    l_result.addAll(Arrays.asList(l_Grps));

            } // else if Actor
            // else if VirtualFolder: no action

        } // if (isTransferingAGcontext(_upd_dLink.getEorFirstLink()))
        else {
            // if the AG context is not transfered, the second end sets the context
            if (_upd_dLink.getEorAllNodes()[1] instanceof GroupIDMember) {
                member = (GroupIDMember)_upd_dLink.getEorAllNodes()[1];

                l_result = new HashSet(10);

                l_result.add(member);
                l_Grps = member.getEorL_GroupID();
                if (l_Grps.length != 0)
                    l_result.addAll(Arrays.asList(l_Grps));
            }
            else if (_upd_dLink.getEorAllNodes()[1] instanceof Actor) {
                actor = (Actor)_upd_dLink.getEorAllNodes()[1];

                l_result = new HashSet(10);

                if (actor.getEorCurrentUserID() != null)
                    l_result.add(actor.getEorCurrentUserID());
                if (actor.getEorCurrentGroupID() != null)
                    l_result.add(actor.getEorCurrentGroupID());

                l_Grps = actor.getEorL_SecondaryGroup();
                if (l_Grps.length != 0)
                    l_result.addAll(Arrays.asList(l_Grps));

            } // else if (!(_dLink.getEorSecondEnd() instanceof Actor))
            // else if VirtualFolder: no action

        } // else if (isTransferingAGcontext(_upd_dLink.getEorFirstLink()))

        if (l_result != null && !l_result.isEmpty()) {
            _upd_dLink.addProperty(
                DisplayableLinkImpl.AG_CONTEXT_SET,
                l_result.clone());
        }
        else {
            // no use of the DisplayableLinkImpl.AG_CONTEXT_SET key
            l_result = null;
        } // else

        _upd_dLink.addProperty(
                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION,
                new Integer(_nodeIndex));

                        // for testing and tracking
                      //if ( ((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("jerry") ) {
                          //System.out.println("LinkRightsImpl.setNextNodeAGrunningContext() end "+
                            //  "- for _nodeIndex == 1, l_result contains:");
                          //if (l_result != null && !l_result.isEmpty()) {
                            //  for (Object ob : l_result) {
                              //      System.out.println("--- "+((ACSObject)ob).getName());
                              //} // for ob
                          //}
                          //else {
                            //  System.out.println("--- null or empty");
                          //}
                      //} // if

        member = null;
        actor = null;
        l_Grps = null;
        return l_result;

    } // if (_nodeIndex == 1)

    // ---------- _upd_dLink after the first link (_nodeIndex > 1) --------------------

    Integer         index = null;
    Map             m_properties = _upd_dLink.getProperties();
    int             currentAGcontextIndex = 0;
    ACSObject       object = (ACSObject)_upd_dLink.getEorAllNodes()[_nodeIndex];

    if (m_properties == null)
        throw new ProcessError("in LinkRightsImpl.setNextNodeAGrunningContext()"+
                        ", the properties are null for _nodeIndex = "+_nodeIndex);

    index = (Integer)_upd_dLink.getProperties().get(
                                DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION);
    currentAGcontextIndex = index.intValue();

    if (_nodeIndex != currentAGcontextIndex + 1)
        throw new ProcessError("in LinkRightsImpl.setNextNodeAGrunningContext()"+
                        ", the current AG context position is not equal to 1 + "+
                        "the position of the node to analyze");

    // no change if the AG context is transfered
    if (isTransferingAGcontext(_upd_dLink.getEorAllACLinks()[_nodeIndex - 1], null, null)) {
                
                                // for testing and tracking
                                //if (((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("Ammounts view"))
                                  //  System.out.println("LinkRightsImpl.setNextNodeAGrunningContext() finds "+
                                    //  "no change for the set for the acLink "+
                                      //_upd_dLink.getEorAllACLinks()[_nodeIndex - 1]+
                                      //" with the AG context "+
                                      //m_properties.get(DisplayableLinkImpl.AG_CONTEXT_SET));

        // no change for the AG context, and the node position is incremented in all cases.
        // caution: may be null, and then it is not reset like when !isTransferingAGcontext().
        // to get another behavior, this method has to be overridden.
        l_result = (HashSet) m_properties.get(DisplayableLinkImpl.AG_CONTEXT_SET);

    } // if (isTransferingAGcontext(_upd_dLink.getEorAllACLinks()[_nodeIndex - 1]))
    else {
        // if the AG context is not transfered, 'object' sets the AG context

                                // for testing and tracking
                                //if (((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("Ammounts view"))
                                  //  System.out.println("LinkRightsImpl.setNextNodeAGrunningContext() works "+
                                    //  "on "+object.getName()+" at _nodeIndex "+_nodeIndex+
                                      //" because there is no transfering");

        if (object instanceof GroupIDMember) {
            member = (GroupIDMember)object;

            if (m_properties.get(DisplayableLinkImpl.AG_CONTEXT_SET) == null)
                l_result = new HashSet(10);
            else {
                // avoid to create a new map
                l_result = (HashSet) m_properties.get(DisplayableLinkImpl.AG_CONTEXT_SET);
                l_result.clear();
            }

            l_result.add(member);
            l_Grps = member.getEorL_GroupID();
            if (l_Grps.length != 0)
                l_result.addAll(Arrays.asList(l_Grps));
        }
        else if (object instanceof Actor) {
            actor = (Actor)object;

            if (m_properties.get(DisplayableLinkImpl.AG_CONTEXT_SET) == null)
                l_result = new HashSet(10);
            else {
                // avoid to create a new map
                l_result = (HashSet) m_properties.get(DisplayableLinkImpl.AG_CONTEXT_SET);
                l_result.clear();
            }

            if (actor.getEorCurrentUserID() != null)
                l_result.add(actor.getEorCurrentUserID());
            if (actor.getEorCurrentGroupID() != null)
                l_result.add(actor.getEorCurrentGroupID());

            l_Grps = actor.getEorL_SecondaryGroup();
            if (l_Grps.length != 0)
                l_result.addAll(Arrays.asList(l_Grps));

        } // else if (!(_dLink.getEorSecondEnd() instanceof Actor))
        // else if VirtualFolder: no action

        if (l_result != null) { // may be empty
            _upd_dLink.addProperty(
                DisplayableLinkImpl.AG_CONTEXT_SET,
                l_result.clone());
        }
        else {
            _upd_dLink.removeProperty(DisplayableLinkImpl.AG_CONTEXT_SET);
        }

    } // else if (isTransferingAGcontext(_upd_dLink.getEorAllACLinks()[_nodeIndex - 1]))

    _upd_dLink.addProperty(
            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION,
            new Integer(_nodeIndex));


                                    // for testing and tracking
                                  //if ( ((ACSObject)_upd_dLink.getEorSecondEnd()).getNickName().equals("Amounts view") ) {
                                      //System.out.println("end of LinkRightsImpl.setNextNodeAGrunningContext() "+
                                        //  "_nodeIndex != 1, with the AG context node position "+
                                          //_upd_dLink.getProperties().get(
                                            //DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)+
                                          //"- l_result contains:");
                                      //if (l_result != null && !l_result.isEmpty()) {
                                        //  for (Object ob : l_result) {
                                          //      System.out.println("--- "+((ACSObject)ob).getName());
                                          //} // for ob
                                      //}
                                      //else if (l_result != null) {
                                        //  System.out.println("--- empty");
                                      //}
                                      //else if (l_result != null) {
                                        //  System.out.println("--- null");
                                      //}
                                  //} // if

    member = null;
    actor = null;
    l_Grps = null;
    m_properties = null;
    object = null;
    
    return l_result;
  } // setNextNodeAGrunningContext()

  /**
   * Called by setNextNodeAGrunningContext() and updateAGrunningContext().
   * Calls AccessControlLinkImpl.isTransferingAGcontextWithoutACL().
   * In addition to this method, the relation ACL tranfers the AG context
   * for the internal ACLs and the external ACLs.
   * Overridden method in the AcsAddons. Caution: the overriding in an AcsAddon
   * should work only on nodes from its AcsAddon.
   * @see ARoad0.CNot.AccessControlLinkImpl#isTransferingAGcontextWithoutACL()
   * @param _acLink never null
   * @param _firstACS ACS of the _acLink first node. May be null. No use in this version.
   * @param _secondACS ACS of the _acLink second node. May be null. No use in this version.
   * @return true if the Account/Groups context of the starting node is transfered
   * to the last node of the AccessControlLinkImpl
   */
  public boolean isTransferingAGcontext(AccessControlLinkImpl _acLink,
          ImmutableACS _firstACS, ImmutableACS _secondACS) {
      return _acLink.isTransferingAGcontextWithoutACL() ||
            _acLink.getLinkTypeValue(AccessControlLink.ACL);
  }

   /**
    * NO USE - Returns true if a PrivilegeSource uses a given PrivilegeForType.
   * The word '&lt;immutable&gt;' is never taken in account in the type comparison.
   * Caution: works for TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges, but
   * not for TYPED_FOR_SEED privileges.
   * The privileges may be applicable to external objects for
   * the TYPED_FOR_SEED PrivilegeForTypes. The other typed privileges
   * are limited to the ACS objects of their ACS.e.
   * Calls UtilityImpl.getObjectTypeWithoutImmutable().
   * @param _priv to test
   * @param _ps to test
   * @return true if _ps is a source for _priv
   */
  public boolean isSourceForPrivilegeForType(PrivilegeForType _priv,
          ImmutablePrivilegeSource _ps) {
      boolean               result = false;
      String                privType = null;
      String                srType = null;

      if (!_priv.getMainType().equals(PrivilegeAbst.TYPED_FOR_SEED) &&
              _priv.getEorACS() != _ps.getEorACS())
          return false;

      privType = utility_.getObjectTypeWithoutImmutable(_priv.getSourceType());

      if (!(_priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)))
          srType = utility_.getObjectTypeWithoutImmutable(_ps.getType());
      else
          srType = _ps.getName().toString();

      if (srType == null || privType == null)
          result = false;
      else if (srType.equals(privType))
          result = true;

      privType = null;
      srType = null;
      return result;
  } // isSourceForPrivilegeForType()

   /**
   * Returns true if a PrivilegeTarget uses a given PrivilegeForType.
   * The word '&lt;immutable&gt;' is never taken in account in the type comparison.
   * Caution: works for TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges, but
   * not for TYPED_FOR_SEED privileges.
   * The privileges may be applicable to external objects for
   * the TYPED_FOR_SEED PrivilegeForTypes. The other typed privileges
   * are limited to the ACS objects of their ACS.
   * Called by PrivilegeRightsFactoryImpl.getL_typedPrivilege().
   * Calls UtilityImpl.getObjectTypeWithoutImmutable().
   * @param _priv to test. Not null.
   * @param _pt to test. Not null.
   * @return true if _ps is a target for _priv
   */
  public boolean isTargetForPrivilegeForType(
          PrivilegeForType _priv, ImmutablePrivilegeTarget _pt) {
      boolean               result = false;
      String                privType = null;
      String                srType = null;

      if (!_priv.getMainType().equals(PrivilegeAbst.TYPED_FOR_SEED) &&
              _priv.getEorACS() != _pt.getEorACS())
          return false;

      privType = utility_.getObjectTypeWithoutImmutable(_priv.getTargetType());

      if (!(_priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)))
          srType = utility_.getObjectTypeWithoutImmutable(_pt.getType());
      else
          srType = _pt.getName().toString();

      if (srType == null || privType == null)
          result = false;
      else if (srType.equals(privType))
          result = true;

      privType = null;
      srType = null;
      return result;
  } // isTargetForPrivilegeForType()

   /**
   * Gets all the PrivilegeSources using a given PrivilegeForType.
   * This method may produce a heavy processing load if there are many EligibleParties in the ACS.
   * The key word 'immutable' is never taken in account.
   * Caution: works for TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges, but
   * not for the TYPED_FOR_SEED privileges.
   * The arguments may be from different ACS, and the result may then be true.
   * Called by ViewInBase_Facade.populateSketchView().
   * Calls getObjectTypeWithoutImmutable(), ACS.getEorM_EligibleParties().
   * @param _priv to analyze. Returns an empty list if null, empty, from another ACS.
   * @param _acs the privilege acs
   * @return List of PrivilegeSources. May be empty, but never null.
   */
  public List<PrivilegeSource> getEorL_SourcesForPrivilegeForType(
          PrivilegeForType _priv, ImmutableACS _acs) {

                                        // for testing and tracking
                                        //System.out.println(
                                          //      "LinkRightsImpl.getEorL_SourcesForPrivilegeForType()"+
                                            //    " starting for PRI "+_priv.getNickName());

      if (_priv == null || _priv.isEmpty() || _priv.getEorACS() != _acs)
          return Collections.emptyList();
      
      Map                   m_ep = null; 
      // copy to avoid collisions with any eligible party updating in _acs
      // quite fast, but no synchro on _acs.getEorM_EligibleParties()
      m_ep = new HashMap(_acs.getEorM_EligibleParties());

      if (m_ep == null || m_ep.isEmpty())
          return Collections.emptyList();

      List<PrivilegeSource> result = new ArrayList(INITIAL_CAPACITY*3);
      EligibleParty         ep = null;
      String                privType = _priv.getSourceType();
      String                epType = null;

      privType = utility_.getObjectTypeWithoutImmutable(privType);

                                        // for testing and tracking
                                        //System.out.println(
                                          //      "LinkRightsImpl.getEorL_SourcesForPrivilegeForType()"+
                                            //    " PRI type: "+privType);

      for (Object obj : m_ep.values()) {
          ep = (EligibleParty)obj;
          if ( ep.getEorACS() != _acs ) continue;

          if (!(_priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)))
            epType = utility_.getObjectTypeWithoutImmutable(ep.getType());
          else
            epType = ep.getName().toString();

          if (epType.equals(privType)) {

                                        // for testing and tracking
                                        //System.out.println(
                                          //      "LinkRightsImpl.getEorL_SourcesForPrivilegeForType()"+
                                            //    " selected source: "+ep.getNickName());

              result.add(ep);
          } // if
      } // for obj

      ep = null;
      privType = null;
      epType = null;
      return result;
  } // getEorL_SourcesForPrivilegeForType()

   /**
   * Gets all the PrivilegeTargets using a given PrivilegeForType.
   * This method may produce a heavy processing load if there are many resources in the ACS.
   * The key word 'immutable' is never taken in account.
   * Caution: works for TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges, but
   * not for the TYPED_FOR_SEED privileges.
   * The arguments may be from different ACS, and the result may then be true.
   * Called by ViewInBase_Facade.populateSketchView() and
   * PrivilegeRightsFactoryImpl.getL_TargetsOfSourcePrivilege().
   * Calls getObjectTypeWithoutImmutable(), ACS.getEorL_VirtualFolders(),
   * ACS.getEorM_Resources().
   * @param _priv to analyze. Returns an empty list if null, empty, from another ACS.
   * @param _acs the privilege acs
   * @return List of PrivilegeTargets (Resources or VirtualFolders).
   * May be empty, but never null.
   */
  public List<PrivilegeTarget> getEorL_TargetsForPrivilegeForType(
          PrivilegeForType _priv, ImmutableACS _acs) {

      if (_priv == null || _priv.isEmpty() || _priv.getEorACS() != _acs)
          return Collections.emptyList();

      // copy to avoid collisions with any resource or virtual folder updating in _acs
      Map                   m_res = null;
      VirtualFolder[]       l_vf = null;

      if (_acs.getManageResources()) {
            // quite fast, but no synchro on _acs.getEorM_Resources()
            m_res = new HashMap(_acs.getEorM_Resources());
      }
      if (_acs.getManageVirtualFolders()) {
            l_vf = _acs.getEorL_VirtualFolders();
      }

      if ((m_res == null || m_res.isEmpty()) && l_vf == null)
          return Collections.emptyList();

      List<PrivilegeTarget> result = new ArrayList(INITIAL_CAPACITY*3);
      Resource              res = null;
      VirtualFolder         vf = null;
      String                privType = _priv.getTargetType();
      String                type = null;

      privType = utility_.getObjectTypeWithoutImmutable(privType);

      if (m_res != null) {
          for (Object obj : m_res.values()) {
              res = (Resource)obj;
              if ( res.getEorACS() != _acs ) continue;

              if (!(_priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)))
                type = utility_.getObjectTypeWithoutImmutable(res.getType());
              else
                type = res.getName().toString();

              if (type.equals(privType)) result.add(res);
          } // for obj
      } // if (m_res != null)

      if (l_vf != null) {
          for (Object obj : l_vf) {
              vf = (VirtualFolder)obj;
              if ( vf.getEorACS() != _acs ) continue;

              if (!(_priv.getMainType().equals(PrivilegeAbst.TYPED_CLASSED_SPECIFIC)))
                type = utility_.getObjectTypeWithoutImmutable(vf.getType());
              else
                type = vf.getName().toString();

              if (type.equals(privType)) result.add(vf);
          } // for obj
      } // if (l_VirtualFolders_ != null)

                                        // for testing and tracking
                                        //if (result != null)
                                          //  System.out.println("LinkRightsImpl.getEorL_TargetsForPrivilegeForType()"+
                                            //    " list result: "+result.size());
                                        //else
                                          //  System.out.println("LinkRightsImpl.getEorL_TargetsForPrivilegeForType()"+
                                            //                                " with null list result: ");

      res = null;
      vf = null;
      privType = null;
      type = null;
      m_res = null;
      l_vf = null;
      return result;
  } // getEorL_TargetsForPrivilegeForType()

} // LinkRightsImpl

