//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;

import ARoad0.gWorkInterface.*;

import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.CNot.AccessControlLink;
import ARoad0.CNot.AccessControlLinkImpl;

import ARoad0.gBaseInterface.*;
import ARoad0.gBase.UtilityImpl;
import ARoad0.gBase.StringRightImpl;
import ARoad0.gBase.ACSImpl;

/**
 * This class is responsible for getting the rights for a node in a view, and since
 * each implied ACS may be generic or from an AcsAddon, the main job
 * of this class is to give the correct method to use for each ACS and each search.
 * The NodeRights effective class is a NodeRightsImpl or one of its subclasses that
 * is set in an AcsAddon. Thus each node offers to the paths search
 * a set of methods from this NodeRightsImpl and, as an option,
 * another set from the AcsAddon that are called from some NodeRightsImpl methods
 * which are overridden. <P>
 * Overriding this class is the easiest way to change, in an ACS addon, the selection
 * for a node of its direct or inherited ACL or PRI rights, or its AG inherited rights.
 * For changing the selection of the AG direct rights or another relation, it is
 * necessary to overread all the methods detectAddon and selectAddon.
 * Each node in a view and each intermediate node has one NodeRightsImpl instance
 * for each view it belongs to. This class has properties that depend only on
 * the class of the node ACS, then 2 nodes of the same view should use the same NodeRights
 * instance, even if it is not the actual behavior in all cases. Each instance
 * of NodeRightsImpl is called by only one thread. A node may be in several DisplayableLinks
 * of the same view, but it has only one NodeRightsImpl instance for the view.
 * The association between a node and its NodeRights instance is managed
 * by the RightsMediatorImpl of the view. <P>
 * The path search algorithms are divided in 4 sets: <P>
 * - the core algorithms are the larger part of the code, and they are in most
 * of the gWork classes; they work for all the ACS, generic one or from an AcsAddon;
 * they consider the node general interface (EligibleParty or Resource, for instance),
 * but they are independant to the effective properties of the node, <P>
 * - the NodeRightsImpl generic methods: they are called by the core methods
 * to process each node, and they use the generic properties of a node and its environment
 * (like the generic properties of a Resource); they work for all the ACSs,
 * they are always called by the core algorythms,
 * but an AcsAddon may choose to overread them in some rare cases, <P>
 * - the NodeRightsImpl AcsAddon methods: the core methods test through
 * the getXXXAddonYYY methods, if a given AcsAddon method (detectAddon... or selectAddon...)
 * is defined for the node, and if it true, the relevant method is called during
 * the paths search by the core algorithms;
 * these AcsAddon methods are empty in the NodeRightsImpl class
 * (all the getXXXAddonYYY methods return false),
 * and they may be overridden in the subclasses to write a code that calls some
 * specialized AcsAddon methods, <P>
 * - the specialized methods of an AcsAddon are in a NodeRightsImpl subclass,
 * and they are called by an overridden NodeRightsImpl AcsAddon method;
 * there is no name pattern for these methods;
 * they use the generic properties of a node and its environment, but
 * they also handle the original properties of the AcsAddon node.
 * Since an AcsAddon may be based on another AcsAddon, a NodeRightsImpl
 * subclass may have another NodeRightsImpl subclass has super class. <P>
 * This structure is an important part of the AcsAddon pattern. It gives to the AcsAddon
 * designer the full control on the true search algorithms to use in its own ACSs, but
 * it also provides the powerful support of the generic algorithms and all the other
 * AcsAddons, to limit the code he has to add in its AcsAddon. <P>
 * A node is here the end of a view access path, or it may be a potential or
 * effective intermediate node in an access path. Such an intermediate node
 * is an ImmutableSource like Actor, GroupIDMember or VirtualFolder.
 * An simple ACSObject can be an intermediate node in a generic Alias relation,
 * but not in an AcsAddon relation. <P>
 * Generally speaking, the rights depend on the node and its one or two previous
 * nodes in the access path. These rights are based mainly on the classes and properties
 * of these 2 or 3 implied nodes, and naturally also on their ACS properties.
 * The resulting rights depends usually on the NodeRightsImpl algorithms only,
 * but in some rare cases, they may also depend on the single NodeRightsImpl property,
 * defined at the NodeRightsImpl creation, that is simply the node itself. <P>
 * Note: the AG rights are always internal ones, for a given ACS. Since all the nodes
 * of an ACS have NodeRight instances that have the same behavior, it is possible
 * for the AG rights, and only for them, to use for a node, the NodeRight instance
 * of another node of the same ACS. This is also true for VirtualFolder
 * memberships, but not for the Bridge, ACL and Privilege rights.
 * <P>
 * Upper-level methods begin by get...(), with...() and detect...(),
 * middle-level methods begin by select...().
 * @version 0.7.0
 */
public class NodeRightsImpl implements NodeRights {

  /**
   * instance of LinkRightsImpl or one of its subclass
   */
  protected LinkRights                link_ = null;
  protected AlgorithmInterpreter      interpreter_ = null;
  protected DisplayableLinkUtilities  linkUtil_ = null;
  protected UtilityImpl               utility_ = null;
  protected PrivilegeRightsFactoryImpl priFact_ = null;
  protected AclRightsFactoryImpl       aclFact_ = null;

  /**
  * Only one constructor, called by RightsMediatorImpl after the creation of
  * the DisplayableLinkUtilities instance of _interpreter.
  * @param _interpreter algorithm interpreter of this view
  * @param _link to use for processing this node
  */
  public NodeRightsImpl(AlgorithmInterpreter _interpreter, LinkRights _link) {
      interpreter_ = _interpreter;
      link_ = _link;
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
      utility_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
      aclFact_ = (AclRightsFactoryImpl)interpreter_.getCoreAlgorithm("Acl");
      priFact_ = (PrivilegeRightsFactoryImpl)interpreter_.getCoreAlgorithm("Pri");
  }

  /**
   * Called by RightsFactory_Facade. Overridden by the subclasses.
   * @return true if the detectAddon2Nodes method provides some work in this class.
   */
  public boolean getDetectAddon2Nodes() {
      return false;
  }
  
  /**
   * Called by RightsFactory_Facade. Overridden by the subclasses.
   * @return true if the detectAddonFinal2Nodes method provides some work in this class.
   */
  public boolean getDetectAddonFinal2Nodes() {
      return false;
  }
  
   /**
   * Called by RightsFactory_Facade. Overridden by the subclasses.
   * @return true if the detectAddon3Nodes method provides some work in this class.
   */
  public boolean getDetectAddon3Nodes() {
      return false;
  }
  
   /**
   * Called by RightsFactory_Facade. Overridden by the subclasses.
   * @return true if the detectAddonFinal3Nodes method provides some work in this class.
   */
  public boolean getDetectAddonFinal3Nodes() {
      return false;
  }

   /**
   * Called by RightsFactory_Facade. Overridden by the subclasses.
   * @return true if the detectAddonFinalAllNodes method provides some work
   * in this class.
   */
  public boolean getDetectAddonFinalAllNodes() {
      return false;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by ThreeNodesRightsFactoryImpl.
   * Overridden by the subclasses.
   * @return true if the selectAddonLastViewNodeAfterActor method provides some work
   * in this class.
   */
  public boolean getSelectAddonLastViewNodeAfterActor() {
      return false;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by
   * ThreeNodesRightsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilege().
   * Overridden by the subclasses.
   * @return true if the selectAddonLastViewNodeAfterGroupIDMember method provides some work 
   * in this class.
   */
  public boolean getSelectAddonLastViewNodeAfterGroupIDMember() {
      return false;
  }

  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by ThreeNodesRightsFactoryImpl.
   * Overridden by the subclasses.
   * @return true if the selectAddonLastViewNodeAfterVirtualFolder method provides some work
   * in this class.
   */
  public boolean getSelectAddonLastViewNodeAfterVirtualFolder() {
      return false;
  }
  
  /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes, and by ThreeNodesRightsFactoryImpl.
   * Overridden by the subclasses.
   * @return true if the selectAddonNewHiddenNodeForActor method provides some work
   *  in this class.
   */
  public boolean getSelectAddonNewHiddenNodeForActor() {
      return false;
  }
  
   /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes, and by
   * ThreeNodesRightsFactoryImpl.addPathsFromGroupIDMemberAcsAddonRelationActor()
   * and addPathsFromGroupIDMemberAcsAddonRelationNoActor().
   * Overridden by the subclasses.
   * @return true if the selectAddonNewHiddenNodeForGroupIDMember method provides some work
   *  in this class.
   */
  public boolean getSelectAddonNewHiddenNodeForGroupIDMember() {
      return false;
  }
  
   /**
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * central loop on the intermediate nodes, and by ThreeNodesRightsFactoryImpl.
   * Overridden by the subclasses.
   * @return true if the selectAddonNewHiddenNodeForVirtualFolder method provides
   * some work in this class.
   */
  public boolean getSelectAddonNewHiddenNodeForVirtualFolder() {
      return false;
  }

  /**
  * For an AcsAddon, filters for a given node the incompatible rights, if any,
  * in the 2-nodes DisplayableLinks, and adds the AcsAddon specific rights, if any.
  * The specific rights are searching for all sources in the view.
  * Called by RightsFactory_Facade before the detection of the 3-nodes links.
  * Overridden by the subclasses.
  * For a sketch view, this method may return new DisplayableLinks from _center to _node,
  * if any, but also, new DisplayableLinks from the view sources to _center
  * when _node is _center.
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @see #detectAddonFinal2Nodes
  * @param _viewInBase EPRViewInBase
  * @param _node node of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  *   with 'other' links associated to every pair (EP,Resource).
  * */
  public Map detectAddon2Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks, BaseObject _center) {

      return _m_l_DisplayableLinks;
  } // detectAcsAddon2Nodes()

  /**
  * For an AcsAddon, defines the specific final direct rights, if any, for a given node
  * in the 2-nodes DisplayableLinks.
  * This method handles the work to do at the end of the searches on direct rights,
  * like for instance the AGO 'Other' rights on a Linux Ubuntu.
  * The specific rights are searching for all sources in the view.
  * For a sketch view, this method may return new DisplayableLinks from _center to _node,
  * if any, but also, new DisplayableLinks from the view sources to _center
  * when _node is _center.
  * Called by RightsFactory_Facade. Overridden by the subclasses.
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @see #detectAddon2Nodes
  * @param _viewInBase EPRViewInBase
  * @param _node node  of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center is the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks,
  *   with 'other' links associated to every pair (EP,Resource).
  * */
  public Map detectAddonFinal2Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks, BaseObject _center) {

      return _m_l_DisplayableLinks;
  } // detectAcsAddonFinal2Nodes()

  /**
  * For an AcsAddon, filters for a given node the incompatible rights, if any,
  * in the 3-nodes DisplayableLinks, and adds the AcsAddon specific rights, if any..
  * The specific rights are searching for any source in the view.
  * Called by RightsFactory_Facade. Overridden by the subclasses.
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @see #detectAddonFinal3Nodes
  * @param _viewInBase EPRViewInBase
  * @param _node node  of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @return Map of DisplayableLinks
  * @exception InterruptedException
   */
  public Map detectAddon3Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks)  throws InterruptedException {

      return _m_l_DisplayableLinks;
  } // detectAddon3Nodes()

  /**
  * For an AcsAddon, filters the incompatible rights, if any, for a node,
  * and adds the AcsAddon specific rights, if any.
  * This method handles the work to do at the end of the searches on 3-nodes links.
  * Called by RightsFactory_Facade. Overridden by the subclasses.
  * Synchronized to _viewInBase by RightsFactory_Facade.
  * @see #detectAddon3Nodes
  * @param _viewInBase EPRViewInBase
  * @param _node node  of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @return Map of DisplayableLinks
  * @exception InterruptedException
   */
  public Map detectAddonFinal3Nodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node,
        Map _m_l_DisplayableLinks)  throws InterruptedException {
      
      return _m_l_DisplayableLinks;
  } // detectAddonFinal3Nodes()

  /**
  * For an AcsAddon, defines the specific final direct rights, if any, for a given node
  * in the DisplayableLinks having more than 3 nodes.
  * This method handles the work to do at the end of the searches on all the links. <P>
  * Note: to speed up this method, the argument _m_effectiveRights is set from
  * the initial argument _m_l_DisplayableLinks before the first call to this method,
  * and for a given EPR view. _m_effectiveRights is then immutable while
  * _m_l_DisplayableLinks may be updated by this method. This is not the case
  * for the NoThan views.  For the EPR views, this behavior implies this method
  * provides independant changes for each _node for which it is called,
  * so that the initial _m_effectiveRights remains usable.
  * Called by RightsFactory_Facade.
  * @param _viewInBase EPRViewInBase
  * @param _node node of this instance
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @param _m_effectiveRights non-null Map where the keys are the _m_l_DisplayableLinks keys
  * where rights are activated, and the value is a set of StringRights. May be an empty set.
  * @return Map of DisplayableLinks
  * @exception InterruptedException
   */
  public Map detectAddonFinalAllNodes(
        EPRViewInBase _viewInBase, ImmutableTarget _node, Map _m_l_DisplayableLinks,
        Map<ImmutableName, Set<StringRight>> _m_effectiveRights)  throws InterruptedException {
      
      return _m_l_DisplayableLinks;
  } // detectAddonFinalAllNodes()

  /**
  * Gets all the rights of the relevant AclEntries managed by the Resource
  * for a given EligibleParty, without the inherited AclEntries.
  * These AclEntries may be managed by different ACSs, and
  * for each ACS, the negative rights are subtracted from the positive ones.
  * The negative rights from an ACS are then the only resulting rights if
  * the subtraction returns an empty list.
  * With ACLs from several ACS, in the resulting list, the result from each ACS
  * is added without filtering. Then one ACS may provide a negative right
  * while another ACS provides a positive right.
  * Note: this method is sensible since it is indirectly used for both the one-link method,
  * the two-link method and a multiple-links method.
  * Calls BaseUtilityImpl.subtractNegativeNotRemovingInOneACS().
  * Called by AclRightsFactoryImpl.addAclInAccessControlLink().
  * @param _res of this instance, with some AclEntries to filter
  * @param _ep may be associated to some _res AclEntries
  * @param _onlyNonConditionalACLs true to read only the non-conditional AclEntries
  * @return list of positive or negative rights from different ACSs, or null.
  * */
  public List<StringRight> detectL_aclEntryRights(ImmutableResource _res,
    ImmutableEligibleParty _ep, boolean _onlyNonConditionalACLs) {
    
      AclEntry[]                l_resAcl = _res.getL_aclEntry();

      if (l_resAcl.length == 0) return null;

      List<AclEntry>            l_acl = null;
      List                      l_ACS = null;

      // get the _ep AclEntries and ACS
      for (int i = 0; i < l_resAcl.length; i++) {
        if (l_resAcl[i] == null || l_resAcl[i].getEorEP() != _ep )
            continue;
        if (_onlyNonConditionalACLs && l_resAcl[i].isConditionalAclEntry())
            continue;

        if (l_ACS == null)
              l_ACS = new ArrayList(INITIAL_CAPACITY / 5);
        if (!l_ACS.contains(l_resAcl[i].getEorACS())) {
          l_ACS.add(l_resAcl[i].getEorACS());
        } // if

        if (l_acl == null)
          l_acl = new ArrayList(INITIAL_CAPACITY / 5);
        l_acl.add(l_resAcl[i]);

      } // for i

      if (l_acl == null) {
          l_resAcl = null;
          return null;
      } // if

      Set<StringRight>          l_temp = new HashSet(INITIAL_CAPACITY / 5);
      Set<StringRight>          l_tempNeg = new HashSet(INITIAL_CAPACITY / 5);
       List<StringRight>        l_result = null;
      StringRight[]             l_right = null;
      ACS                       acs = null;
      AclEntry                  acl = null;
      boolean                   forNode = false;
      if (_res instanceof Node) forNode = true;

      for (Iterator ita = l_ACS.iterator(); ita.hasNext(); ) {
          acs = (ACS)ita.next();
          // fill l_temp and l_tempNeg with acs ACLs
          for (Iterator ite = l_acl.iterator(); ite.hasNext(); ) {
              acl = (AclEntry)ite.next();
              if (acl.getEorACS() != acs) continue;
              l_right = acl.getL_Rights();
              for (int i = 0; i < l_right.length; i++) {
                  if ( !l_right[i].isGrantingRight() ) 
                      l_tempNeg.add(l_right[i]);
                  else l_temp.add(l_right[i]);                 
              } // for i
          } // for ite on l_acl

          if (l_result == null)
              l_result = new ArrayList(INITIAL_CAPACITY / 5);

          if (l_tempNeg.isEmpty()) {
              l_result.addAll(l_temp);
          } // if
          
          else {
              l_temp = utility_.subtractNegativeNotRemovingInOneACS(
                        l_temp, l_tempNeg, acs, forNode);
              // for each acs, subtract negative rights and do not keep them in l_result
              // if they are weaker
              if (!l_temp.isEmpty()) l_result.addAll(l_temp);
              else l_result.addAll(l_tempNeg);
          } // else
          l_temp.clear();
          l_tempNeg.clear();
      } // for ita on l_acs
      
      l_acl = null;
      l_ACS = null;
      l_temp = null;
      l_tempNeg = null;
      l_resAcl = null;
      acl = null;
      acs = null;
      l_right = null;

      if (l_result == null || l_result.isEmpty()) return null;
      else return l_result;
  } // detectL_aclEntryRights()

  /**
  * Gets all the rights of the relevant direct linked privileges managed
  * by the target for a given source. Inherited privileges are not covered.
  * These privileges are managed by different ACSs, and
  * for each ACS, the negative rights are subtracted from the positive ones.
  * The negative rights from an ACS are then the only resulting rights if
  * the subtraction returns an empty list.
  * With ACLs from several ACS, in the resulting list, the result from each ACS
  * is added without filtering. Then one ACS may provides a negative right
  * while another ACS provides a positive right.
  * Note: this method is sensible since it is indirectly used for both the one-link method,
  * the two-link method and a multiple-links method.
  * Calls PrivilegeRightsFactoryImpl.getL_linkedPrivilege() and
  * BaseUtilityImpl.subtractNegativeNotRemovingInOneACS().
  * Called by PrivilegeRightsFactoryImpl.addPrivilegeInAccessControlLink().
  * @param _targ privilege target of this instance
  * @param _sour privilege source
  * @return list of positive or negative rights from different ACSs, or null.
  * */
  public List<StringRight> detectL_linkedPrivilegeRights(ImmutablePrivilegeTarget _targ,
          ImmutablePrivilegeSource _sour) {

      List<PrivilegeForLinks>   l_pri = priFact_.getL_linkedPrivilege(_sour, _targ);

      if (l_pri == null) return null;

      List                      l_ACS = null;

      // get the ACS
      for (PrivilegeForLinks pri : l_pri) {

        if (l_ACS == null)
              l_ACS = new ArrayList(INITIAL_CAPACITY / 5);
        if (!l_ACS.contains(pri.getEorACS())) {
          l_ACS.add(pri.getEorACS());
        } // if

      } // for pri

      Set<StringRight>          l_temp = new HashSet(INITIAL_CAPACITY / 5);
      Set<StringRight>          l_tempNeg = new HashSet(INITIAL_CAPACITY / 5);
       List<StringRight>        l_result = null;
      StringRight[]             l_right = null;
      ACS                       acs = null;
      boolean                   forNode = false;
      if (_targ instanceof Node) forNode = true;

      for (Iterator ita = l_ACS.iterator(); ita.hasNext(); ) {
          acs = (ACS)ita.next();
          // fill l_temp and l_tempNeg with acs privileges
          for (PrivilegeForLinks pri : l_pri) {
              if (pri.getEorACS() != acs) continue;
              l_right = pri.getL_Rights();
              for (int i = 0; i < l_right.length; i++) {
                  if ( !l_right[i].isGrantingRight() )
                      l_tempNeg.add(l_right[i]);
                  else l_temp.add(l_right[i]);
              } // for i
          } // for ite on l_acl

          if (l_result == null)
              l_result = new ArrayList(INITIAL_CAPACITY / 5);

          if (l_tempNeg.isEmpty()) {
              l_result.addAll(l_temp);
          } // if

          else {
              l_temp = utility_.subtractNegativeNotRemovingInOneACS(
                        l_temp, l_tempNeg, acs, forNode);
              // for each acs, subtract negative rights and do not keep them in l_result
              // if they are weaker
              if (!l_temp.isEmpty()) l_result.addAll(l_temp);
              else l_result.addAll(l_tempNeg);
          } // else
          l_temp.clear();
          l_tempNeg.clear();
      } // for ita on l_acs

      // TO DO ------------------------
      // priority among ACS
      // rule 'sens metarule: parent other-ACS denyings and access-deficiencies delete this embedded-ACS grantings' is implicitly true
      // rule 'sens metarule: no-parent other-ACS denyings and access-deficiencies do not delete this ACS grantings' is implicitly true
      //

      l_pri = null;
      l_ACS = null;
      l_temp = null;
      l_tempNeg = null;
      acs = null;
      l_right = null;

      if (l_result == null || l_result.isEmpty()) return null;
      else return l_result;
  } // detectL_linkedPrivilegeRights()

  /**
  * Returns true if the Account/Group inherited rights in the resources tree allows
  * to access to the resouce.
  * If the leaf is a resource, it is true if the parent UserID/GroupID are
  * the leaf UserID/GroupID, and if the parent rights allow the access to children.
  * If the AG rights inheritance is applicable to the ImmutableSource ACS,
  * this method applies the following generic rules: <P>
  * - 'AGO right inheritance 1: if no access from the  inherited rights,
  * no AGO access to the child if and only if the account or the group is not null' <P>
  * - 'AGO right: A rights overlay G rights' <P>
  * Called by withAccessThroughNodesTreeFromEP(). <P>
  * Caution: To follow the AcsAddon pattern, withAccessThroughNodesTreeFromEP()
  * should be the single caller in the generic gWork package.
  * This method is not private only to let the subclasses overread it,
  * but it is necessary there to call it only from the
  * same method, to not jump withAccessThroughNodesTreeFromEP(). <P>
  * Calls Resource.getAGOAccessFromTheParentAccount() and getAGOAccessFromTheParentGroup().
  * Overridden method by the AcsAddons, like in the AcsAddon Ubuntu for instance.
  * @see ARoad0.gBase.BaseUtilityImpl#withAcrossToDirectoryAGRights(ARoad0.gBaseInterface.StringRight[])
  * @param _res Resource of this instance; may be accessed or not
  *   through its parent tree
  * @param _acc the _res userID or not. May be null. If null and if
  * all the _res parents have no UserID, returns true.
  * @param _grp the _res groupID or not. May be null. If null and if
  * all the _res parents have no GroupID, returns true.
  * @return true if the resource parent tree allows to go through down to the leaf.
  * */
  protected boolean withAgoAccessThroughNodesTree(
    ImmutableResource _res, ImmutableUserID _acc, ImmutableGroupID _grp) {

      boolean result = false;
      ImmutableLeaf res = _res;

      while (res.getEorParent() != null) {
          if (res instanceof Resource &&
                  _acc == ((Directory)res.getEorParent()).getEorUserID()) {

                  // ((Resource)res).getAGOAccessFromTheParentAccount() true
                  // when _acc null and ((Directory)res.getEorParent()).getEorUserID() null
                if (((Resource)res).getAGOAccessFromTheParentAccount()) result = true;
                else result = false;

          } else if (res instanceof Resource &&
                  _grp == ((Directory)res.getEorParent()).getEorGroupID()) {

                if (((Resource)res).getAGOAccessFromTheParentGroup()) result = true;
                else result = false;
          //} else if (res instanceof VirtualFolder) {
            //    result = true;
          } else {
                result = false;

                       //for testing and tracking
                      //System.out.println("in NodeRightsImpl.withAccessThroughNodesTree(), "+
                        //" for 'other', result: "+result+" and res.getAGOAccessFromTheParentOther(): "+
                        //res.getAGOAccessFromTheParentOther());
          } // else

          if (!result) break;
          else res = (ImmutableLeaf)res.getEorParent();
      } // while

                       //for testing and tracking
                      //if (_acc != null && _grp != null)
                          //System.out.println("in NodeRightsImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _acc "+_acc.getName()+" _grp "+_grp.getName());
                      //else if (_acc != null)
                          //System.out.println("in NodeRightsImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _acc "+_acc.getName()+" _grp null");
                      //else if (_grp != null)
                          //System.out.println("in NodeRightsImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _acc null and _grp "+_grp.getName());
                      //else
                          //System.out.println("in NodeRightsImpl.withAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _acc null and _grp null");

      return result;
  } // withAgoAccessThroughNodesTree()

  /**
  * Returns true if the Privilege inherited rights in the resources or 
  * the virtual folders tree allows to access to the PrivilegeTarget
    for the EligibleParties _ep or _grp.
  * Returns true if the resource has no parent.
  * Called by withAccessThroughNodesTreeFromEP(). <P>
  * Caution: To follow the AcsAddon pattern, withAccessThroughNodesTreeFromEP()
  * should be the single caller, from the generic gWork package or not.
  * This method is not private only to let the subclasses overread it,
  * but it is necessary there to call it through the same way, that is
  * to never jump the method withAccessThroughNodesTreeFromEP() which
  * chooses the good arguments for calling this method. <P>
  * Calls ImmutablePrivilegeTarget.getM_TargetInheritedLinkedPRIRights().
  * Overridden method by the AcsAddons.
  * @param _res Resource or VirtualFolder of this instance; may be accessed or not
  *   through its parent tree
  * @param _ep the _res userID or not, or an Actor. May be null.
  * @param _grp the _res groupID or not. May be null.
  * @return true if the parent tree allows to go through down to the leaf.
  * */
  protected boolean withPriAccessThroughNodesTree(
    ImmutablePrivilegeTarget _res, ImmutableEligibleParty _ep, ImmutableGroupID _grp) {

      if (_res.getEorParent() == null) {
          return true;
      }

      boolean               result = false;
      Map<String, StringRight[]> m_rights = _res.getM_TargetInheritedLinkedPRIRights();
      ImmutableEligibleParty member = null;
      String                 name2 = null;
      ImmutableGroupID[]     l_grps = null;

      if (_ep != null && _ep instanceof ImmutableActor) {
          member = (ImmutableActor)_ep;
            // the direct parent has the string '1', and its direct parent
            // has the string '2' as parent order in the last name component, and so on.
            // initialize l_matchingKeys with the names containing the member name
            // in the order of m_rights.keySet() iterator
          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name

          member = null;
          name2 = null;
          return result;
      } // if

      if (_ep != null && _ep instanceof ImmutableUserID) {
          member = (ImmutableUserID)_ep;

          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name
      } // if

      if (!result && _ep != null && _ep instanceof ImmutableUserID &&
              ((ImmutableUserID)_ep).getEorPrimGroup() != null) {
          member = ((ImmutableUserID)_ep).getEorPrimGroup();

          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name
      } // if

      if ( _grp != null) { // _ep is null so l_result is null
          member = _grp;

          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name
      } // if

      if (!result) {
          if (_ep != null && _ep instanceof ImmutableUserID &&
                  _res.getEorACS().getManageGroups()) {

                    l_grps = ((ImmutableUserID)_ep).getEorL_GroupID();
                    for (int i = 0; i < l_grps.length; i++) {

                          for (String name : m_rights.keySet()) {
                                name2 = name.substring(0, name.lastIndexOf(":: "));

                                if (name.equals( ((Name)member.getName()).extended(
                                        // add the parent order
                                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                                    result = true;
                                    break;
                                }
                          } // for name

                          if (!result)
                            break;

                    } // for i

          } else if (_grp != null && _ep instanceof ImmutableGroupID &&
                  _res.getEorACS().getManageGroupTrees()) {

                    l_grps = _grp.getEorL_GroupID();
                    for (int i = 0; i < l_grps.length; i++) {

                          for (String name : m_rights.keySet()) {
                                name2 = name.substring(0, name.lastIndexOf(":: "));

                                if (name.equals( ((Name)member.getName()).extended(
                                        // add the parent order
                                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                                    result = true;
                                    break;
                                }
                          } // for name

                          if (!result)
                            break;

                    } // for i

          } // else
      } // if (!result )

                       //for testing and tracking
                      //if (_ep != null && _grp != null)
                          //System.out.println("in NodeRightsImpl.withPriAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _ep "+_ep.getName()+" _grp "+_grp.getName());
                      //else if (_ep != null)
                          //System.out.println("in NodeRightsImpl.withPriAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _ep "+_ep.getName()+" _grp null");
                      //else if (_grp != null)
                          //System.out.println("in NodeRightsImpl.withPriAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _ep null and _grp "+_grp.getName());
                      //else
                          //System.out.println("in NodeRightsImpl.withPriAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _ep null and _grp null");

      m_rights = null;
      l_grps = null;
      member = null;
      return result;
  } // withPriAccessThroughNodesTree()

  /**
  * Returns true if the AclEntry inherited rights in the resources tree allows
  * to access to the resource for the EligibleParties _ep or _grp.
  * Returns true if the resource has no parent.
  * Called by withAccessThroughNodesTreeFromEP(). <P>
  * Caution: to follow the AcsAddon pattern, withAccessThroughNodesTreeFromEP()
  * should be the single caller of this method, from the generic gWork package or not.
  * This method is not private to let the subclasses overread it,
  * but it is necessary there to call it through the same way, that is
  * to never jump the method withAccessThroughNodesTreeFromEP() which
  * chooses the good arguments for calling this method. <P>
  * Calls ImmutableResource.getM_TargetInheritedACLRights().
  * Overridden method by the AcsAddons.
  * @param _res Resource of this instance; may be accessed or not
  *   through its parent tree
  * @param _ep the _res userID or not, or an Actor. May be null.
  * @param _grp the _res groupID or not. May be null.
  * @return true if the resource parent tree allows to go through down to the leaf.
  * */
  protected boolean withAclAccessThroughNodesTree(
    ImmutableResource _res, ImmutableEligibleParty _ep, ImmutableGroupID _grp) {

      if (_res.getEorParent() == null) {
          return true;
      }

      boolean               result = false;
      Map<String, StringRight[]> m_rights = ((ImmutableResource)_res).
                                                    getM_TargetInheritedACLRights();
      ImmutableEligibleParty member = null;
      String                 name2 = null;
      ImmutableGroupID[]     l_grps = null;

      if (_ep != null && _ep instanceof ImmutableActor) {
          member = (ImmutableActor)_ep;
            // the direct parent has the string '1', and its direct parent
            // has the string '2' as parent order in the last name component, and so on.
            // initialize l_matchingKeys with the names containing the member name
            // in the order of m_rights.keySet() iterator
          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name

          member = null;
          name2 = null;
          return result;
      } // if

      if (_ep != null && _ep instanceof ImmutableUserID) {
          member = (ImmutableUserID)_ep;

          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name
      } // if

      if (!result && _ep != null && _ep instanceof ImmutableUserID &&
              ((ImmutableUserID)_ep).getEorPrimGroup() != null) {
          member = ((ImmutableUserID)_ep).getEorPrimGroup();

          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name
      } // if

      if ( _grp != null) { // _ep is null so l_result is null
          member = _grp;

          for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)member.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                    result = true;
                    break;
                }
          } // for name
      } // if

      if (!result) {
          if (_ep != null && _ep instanceof ImmutableUserID &&
                  _res.getEorACS().getManageGroups()) {

                    l_grps = ((ImmutableUserID)_ep).getEorL_GroupID();
                    for (int i = 0; i < l_grps.length; i++) {

                          for (String name : m_rights.keySet()) {
                                name2 = name.substring(0, name.lastIndexOf(":: "));

                                if (name.equals( ((Name)member.getName()).extended(
                                        // add the parent order
                                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                                    result = true;
                                    break;
                                }
                          } // for name

                          if (!result)
                            break;

                    } // for i

          } else if (_grp != null && _ep instanceof ImmutableGroupID &&
                  _res.getEorACS().getManageGroupTrees()) {

                    l_grps = _grp.getEorL_GroupID();
                    for (int i = 0; i < l_grps.length; i++) {

                          for (String name : m_rights.keySet()) {
                                name2 = name.substring(0, name.lastIndexOf(":: "));

                                if (name.equals( ((Name)member.getName()).extended(
                                        // add the parent order
                                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() ) &&
                                        ((StringRight[])m_rights.get(name)).length != 0 ) {
                                    result = true;
                                    break;
                                }
                          } // for name

                          if (!result)
                            break;

                    } // for i

          } // else
      } // if (!result )


                       //for testing and tracking
                      //if (_ep != null && _grp != null)
                          //System.out.println("in NodeRightsImpl.withAclAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _ep "+_ep.getName()+" _grp "+_grp.getName());
                      //else if (_ep != null)
                          //System.out.println("in NodeRightsImpl.withAclAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //" _ep "+_ep.getName()+" _grp null");
                      //else if (_grp != null)
                          //System.out.println("in NodeRightsImpl.withAclAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _ep null and _grp "+_grp.getName());
                      //else
                          //System.out.println("in NodeRightsImpl.withAclAccessThroughNodesTree(), "+
                            //"result: "+result+" for _res "+_res.getName()+
                            //", _ep null and _grp null");

      m_rights = null;
      l_grps = null;
      member = null;
      name2 = null;
      return result;
  } // withAclAccessThroughNodesTree()

  /**
  * This method is designed to be fast, and to returns true if the resource tree
  * or the virtual folder tree allows the source to access through 
  * the inherited AGO, ACL or PRI rights to the leaf.
  * All the ACSs apply the metarule 'right inheritance in the targets tree: no more
  * than one of the three ACL/AGO/LPRI rights inheritance may be activated'. <P>
  * This method implements the following generic rules for one of the three inherited rights: <P>
  * - 'ACL/AGO/LPRI right inheritance 1: child/node inherits rights from 
  * the direct and indirect parents', <P>
  * - 'ACL/LPRI right inheritance 1: if activated and no access from the inherited rights,
  * no ACL/LPRI access to the child', <P>
  * - 'ACL/AGO/LPRI right inheritance 2: at each access of an account, the first
  * eligible party having inherited rights set them, and the search order is,
  * first the account, second its primary group, and third the account groups', <P>
  * - 'ACL/AGO/LPRI right inheritance 2: at each access of a group, the first
  * eligible party having inherited rights set them, and the search order is
  * first the group, second the group groups', <P>
  * - 'ACL/LPRI right inheritance 2: at each access of an actor, the first
  * eligible party having inherited rights set them, and the search order is,
  * first the actor, second the account of its AG context,
  * third the group of its AG context, fourth the secondary groups of the actor', <P>
  * - 'AGO right inheritance 1: if activated and no access from the  inherited rights,
  * no AGO access to the child if and only if the account or the group is not null', <P>
  * - 'AGO right inheritance 2: at each access of an actor, the first
  * eligible party having inherited rights set them, and the search order is,
  * first the account of its AG context, second the group of its AG context,
  * third the secondary groups of the actor', <P>
  * For the account-to-group primary group relation, these rules imply that the primary group
  * of the account argument delivers its inherited rights, before any group, even
  * the group argument.
  * For an actor, these rules imply that access may be authorize through the actor current userID,
  * even if the AG context of the actor has not to be considered in the access path.
  * In this latter case, the effective rights of the actor are null, but that is the
  * responsability of the consumer of this method. <P>
  * For the group-to-group is_member relation, this method does not apply the rule
  * 'EP: group of groups on N levels, with search of the group inherited rights
  * only for the first is_member level' <P>
  * An AcsAddon may overread this method to change these rules, for example to simplify
  * the method, or to change one of the three called methods
  * withAgo/Pri/AclAccessThroughNodesTree(). It is then recommanded to analyze the need
  * to change the method getL_accessRightsThroughNodesTree(). Any code editing should
  * have a similar changing in the AcsAddon ACS rules. <P>
  * Called by AgoRightsFactoryImpl.detectHiddenChainedGroupsRights(),
  * detectOneHiddenNodeWithCommonAGORights(), detectAGRights(), detectOwnerContainRights(),
  * and selectForGroupIDMemberItsHiddenActorsWithNextLinks(),
  * AclRightsFactoryImpl.addAclInAccessControlLink(),
  * ThreeNodesRightsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeMemberOwnContain(),
  * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
  * Calls the three methods withAgo/Pri/AclAccessThroughNodesTree().
  * @see #getL_accessRightsThroughNodesTree(ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableEligibleParty, ARoad0.gBaseInterface.ImmutableGroupID) 
  * @param _ep Actor, UserID or GroupID to test. If null, returns true.
  * If its ACS does not manage any rights inheritance, returns true.
  * @param _res Resource or VirtualFolder of this instance,
  * which may be accessed or not through its parent tree. Never null.
  * @param _ep_2 the group member which may be accessed by _ep. May be null. If
  * _ep is not an Actor, this argument is not used. Otherwise, in the generic gWork algorithms,
  * only AgoRightsFactoryImpl.detectOneHiddenNodeWithCommonAGORights() uses this argument.
  * Then, for an Actor, this argument replaces the current UserID or the current GroupID
  * of the Actor.
  * @return true if the resource parent tree allows _ep to go through down to the resource,
  * if _ep is null, or if its ACS does not manage any rights inheritance.
  * */
  public boolean withAccessThroughNodesTreeFromEP(
     ImmutableSource _ep, ImmutableLeaf _res, ImmutableGroupIDMember _ep_2) {

    boolean             result = true;
    ImmutableGroupID[]  l_grps = null;

    // process first the LPRI inherited rights
    if (_res.getEorACS().getManageLPRIRightsInheritance() &&
            _res instanceof ImmutablePrivilegeTarget) {

                           // for testing and tracking
                          //System.out.println("in NodeRightsImpl.withAccessThroughNodesTreeFromEP(), "+
                            //"for LPRI rights, and from "+_ep+" to "+_res);

                if (((ImmutablePrivilegeTarget)_res).
                        getEorImmutableL_TargetInheritedLinkedPrivileges() == null)
                    result = false;
                else {
                    if (_ep instanceof Actor) {
                        result = withPriAccessThroughNodesTree((ImmutableResource)_res,
                                (ImmutableEligibleParty)_ep, null);

                        if (!result && _ep_2 == null) {
                           result = withPriAccessThroughNodesTree((ImmutableResource)_res,
                                    ((Actor)_ep).getEorCurrentUserID(), ((Actor)_ep).getEorCurrentGroupID());

                        } else if (!result && _ep_2 instanceof ImmutableUserID) {
                           result = withPriAccessThroughNodesTree((ImmutableResource)_res,
                                    (ImmutableUserID)_ep_2, ((Actor)_ep).getEorCurrentGroupID());

                        } else if (!result && _ep_2 instanceof ImmutableGroupID) {
                           result = withPriAccessThroughNodesTree((ImmutableResource)_res,
                                    ((Actor)_ep).getEorCurrentUserID(), (ImmutableGroupID)_ep_2);

                        } // else if
                    } // if (_ep instanceof Actor)

                    // the method getL_accessRightsThroughNodesTree() is an image of this algorithm
                    else if (_ep instanceof ImmutableGroupIDMember) {
                        if (_ep instanceof ImmutableUserID) {
                            result = withPriAccessThroughNodesTree((ImmutableResource)_res,
                                        (ImmutableUserID)_ep, ((ImmutableUserID)_ep).getEorPrimGroup());
                        } else if (_ep instanceof ImmutableGroupID) {
                            result = withPriAccessThroughNodesTree((ImmutableResource)_res,
                                        null, (ImmutableGroupID)_ep);
                        } // else
                    } // else if (_ep instanceof ImmutableGroupIDMember)

                    l_grps = null;
                } // else

                       // for testing and tracking
                      //System.out.println("in NodeRightsImpl.withAccessThroughNodesTreeFromEP(), "+
                        //"for LPRI rights, and from "+_ep+" to "+_res+" - result: "+result);

                return result;
    } // if (_res.getEorACS().getManageLPRIRightsInheritance())

    // process the ACL inherited rights
    else if (_res.getEorACS().getManageACLRightsInheritance() &&
            _res instanceof ImmutableResource) {

                                       // for testing and tracking
                                      //System.out.println("in NodeRightsImpl."+
                                        //        "withAccessThroughNodesTreeFromEP(), "+
                                          //      "for ACL rights, and from "+_ep+" to "+_res);

                      // the default value
                if (((ImmutableResource)_res).getEorImmutableL_TargetInheritedAclEntries() == null)
                    result = false;
                else {
                    if (_ep instanceof Actor) {
                        result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                                (ImmutableEligibleParty)_ep, null);
                        
                        if (!result && _ep_2 == null) {
                           result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                                    ((Actor)_ep).getEorCurrentUserID(), ((Actor)_ep).getEorCurrentGroupID());

                        } else if (!result && _ep_2 instanceof ImmutableUserID) {
                           result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                                    (ImmutableUserID)_ep_2, ((Actor)_ep).getEorCurrentGroupID());

                        } else if (!result && _ep_2 instanceof ImmutableGroupID) {
                           result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                                    ((Actor)_ep).getEorCurrentUserID(), (ImmutableGroupID)_ep_2);

                        } // else if
                    } // if (_ep instanceof Actor)

                    // the method getL_accessRightsThroughNodesTree() is an image of this algorithm
                    else if (_ep instanceof ImmutableGroupIDMember) {
                        if (_ep instanceof ImmutableUserID) {
                            result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                                        (ImmutableUserID)_ep, ((ImmutableUserID)_ep).getEorPrimGroup());
                        } else if (_ep instanceof ImmutableGroupID) {
                            result = withAclAccessThroughNodesTree((ImmutableResource)_res,
                                        null, (ImmutableGroupID)_ep);
                        } // else
                    } // else if (_ep instanceof ImmutableGroupIDMember)

                    l_grps = null;
                } // else

                                       // for testing and tracking
                                      //System.out.println("in NodeRightsImpl.withAccessThroughNodesTreeFromEP(), "+
                                        //"for the inherited ACL rights, and from "+_ep+" to "+_res+" - result: "+
                                        //result);

                return result;
    } // if (_res.getEorACS().getManageACLRightsInheritance())

    // end of processing for the LPRI and ACL inherited rights
    if (_ep == null || 
            !_ep.getEorACS().getManageAGORightsInheritance() ||
            !(_ep instanceof ImmutableEligibleParty) ||
            !(_res instanceof ImmutableResource))
        return result;

                       // for testing and tracking
                      //System.out.println("in NodeRightsImpl.withAccessThroughNodesTreeFromEP(), "+
                        //"for AGO rights, and from "+_ep+" to "+_res);


    if (_ep instanceof Actor) {

                       // for testing and tracking
                      //System.out.println("in NodekRightsImpl.withAccessThroughNodesTreeFromEP(), "+
                        //"for AGO rights, _ep is an Actor ");

        if (_ep_2 == null) {
           result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                    ((Actor)_ep).getEorCurrentUserID(), ((Actor)_ep).getEorCurrentGroupID());

        } else if (_ep_2 instanceof ImmutableUserID) {
           result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                    (ImmutableUserID)_ep_2, ((Actor)_ep).getEorCurrentGroupID());

        } else if (_ep_2 instanceof ImmutableGroupID) {
           result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                    ((Actor)_ep).getEorCurrentUserID(), (ImmutableGroupID)_ep_2);

        } // else if

        //if (!result && _ep.getEorACS().getManageSecGroupsForActor()) {
          //  l_grps = ((Actor)_ep).getEorL_SecondaryGroup();
            //for (int i = 0; i < l_grps.length; i++) {
              //  result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                //        null, l_grps[i]);
                //if (result) break;
            //} // for i
        //} // if (!result)
    } // if (_ep instanceof Actor)

    // the method getL_accessRightsThroughNodesTree() is an image of this algorithm
    else if (_ep instanceof ImmutableGroupIDMember) {

                       // for testing and tracking
                      //System.out.println("in NodeRightsImpl.withAccessThroughNodesTreeFromEP(), "+
                        //"for AGO rights, _ep is an GroupIDMember ");

        if (_ep instanceof ImmutableUserID) {
            result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                        (ImmutableUserID)_ep, ((ImmutableUserID)_ep).getEorPrimGroup());

            if (!result && _ep.getEorACS().getManageGroups()) {
                l_grps = ((ImmutableGroupIDMember)_ep).getEorL_GroupID();
                for (int i = 0; i < l_grps.length; i++) {
                   result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                            null, l_grps[i]);
                    if (result) break;
                } // for i
            } // if (!result)
        } else if (_ep instanceof ImmutableGroupID) {
            result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                        null, (ImmutableGroupID)_ep);

            if (!result && _ep.getEorACS().getManageGroupTrees()) {
                l_grps = ((ImmutableGroupIDMember)_ep).getEorL_GroupID();
                for (int i = 0; i < l_grps.length; i++) {
                    result = withAgoAccessThroughNodesTree((ImmutableResource)_res,
                            null, l_grps[i]);
                    if (result) break;
                } // for i
            } // if (!result)
        } // else
    } // else if (_ep instanceof ImmutableGroupIDMember)

                       // for testing and tracking
                      //System.out.println("in NodeRightsImpl.withAccessThroughNodesTreeFromEP(), "+
                        //"for AGO rights, and from "+_ep+" to "+_res+" - result: "+result);
    
    l_grps = null;
    return result;
  } // withAccessThroughNodesTreeFromEP()

  /**
  * Returns in an AccessControlLink the inherited rights given by the directory tree
  * or virtual folder tree, as AG, ACL or LPRI rights. This method has to be called on
  * the NodeRightsImpl instance of the argument _res. Comments allow to explain the result.
  * The result uses one of the type AccessControlLink.ACL, LPRI or OWNER.
  * If there is at least a conditional ACL, the result should contain in the comment
  * 'conditional ACL' for the relevant type, while it is not done in the present method.
  * Furthermore, LinkRightsMySQLImpl offers an exemple of overridden method
  * where the conditional group names are added in the comment,
  * like in AclRightsFactoryImpl.addAclInAccessControlLink() for the direct ACLs.
  * This important method has to be called only on the LinkRightsImpl instance of the argument _res.
  * Among the 3 types of inherited rights, an ACS can manage only one type.
  * There is never bottom-up search in the tree of the applicable rights, to provide
  * a fast method. The default algorithm in this method for the ACL or LPRI inherited rights
  * are generally overridden in an AcsAddon gWork package.
  * In the generic gBase classes or in an AcsAddon gBase package, a node has to update
  * the inherited rights of its children, following its own algorithm.
  * For the ACL or LPRI inherited rights, this is always done in an AcsAddon gBase package.
  * The AGO rights inheritance requires to see if each parent node allows or not to go accross it.
  * In this case, to avoid a bottom-up search in this method, each Directory updates
  * the methods getAGOAccessFromTheXXXAccount() of its direct children
  * when its proper AGO rights change. <P>
  * For the ACL and LPRI inherited rights, in the default algorithm of this method,
  * the rights of the leaf are those of the -first- EligibleParty that is linked
  * to the argument _ep or to _grp.
  * In the map of the inherited rights of _res, this method searches
  * among the map keys to find a linked eligible party. It applies a bottom-up
  * test among the parents tree, where the key is the name of an eligible party
  * extended by the order of the parent from _res. For example, if both
  * the _res direct parent and its proper parent deliver two inherited rights sets
  * for _ep to _res in the map, there are two keys '_ep primary group name + 1'
  * and '_ep primary group name + 2' in the inherited rights map of _res,
  * and the returned rights will be the value for the key '_ep primary group name + 1',
  * from the _res direct parent. This is due to the fact that the method
  * ImmutableResource.getM_TargetInheritedACLRights() returns a map where the keySet()
  * method has an iterator in the natural order of the ImmutableNames it uses.
  * This implies that the root of the tree should not give inherited rights
  * to _ep or _grp, while the leaf provides non-null inherited rights through
  * an intermediate parent.
  * In all cases, the method mergeInheritedAclPriRightsThroughNodesTree() is
  * responsible to set the effective ACL or LPRI rights, from the inheritance or not,
  * only if the array of inherited rights is not null (but it may be empty). <P>
  * For the AGO inherited rights, the account or group inherited rights of the resource
  * are returned simply if the tested eligible party is equal to the resource account or group. <P>
  * In the default algorithm of this method, the order of the linked eligible parties
  * to be tested in the inherited rights map is identical for the three types
  * of inherited rights, and this EP search order is: <P>
  * 1/ the account _ep first - and the method ends immediately if _ep is able
  * to deliver some inherited rights to _res <P>
  * 2/ if the previous result is null, the _ep primary group is tested, if any, <P>
  * 3/ otherwise, the group _grp passed in to the method is tested, <P>
  * 4/ if the result is still null, the _ep groups, <P>
  * 5/ at the end, all the _grp groups at the first level are tested. <P>
  * The EP search order does not depend on the distance from _res to the parent
  * which delivers the inherited rights. This is why a parent at three levels above _res
  * will be able to deliver inherited rights to _ep, while the _res direct parent
  * delivers rights to _grp that wont be returned. The EP search order
  * algorithm may be overridden by an AcsAddon. This generic algorithm is trully used
  * only in the rare cases where the _res ACS structure includes all the relevant properties. <P>
  * This method is strongly coupled to withAccessThroughNodesTreeFromEP(), and
  * the rules to follow are explained in the documentation of this method.
  * For the AGO right inheritance, it is recommanded to use
  * withAccessThroughNodesTreeFromEP() before, to check in if an access is allowed.
  * An AcsAddon may overread this method, and then, it is necessary to analyze
  * the need to change also the methods withAcl/Ago/PriAccessThroughNodesTree() and
  * withAccessThroughNodesTreeFromEP(). For example, the AcsAddon Ubuntu Overrides it
  * to add the AG 'Other' rights. No comment is added by this method, but this may be
  * false in an AcsAddon. <P>
  * Called by AgoRightsFactoryImpl.detectAGRights(),
  * addOwnerContainRightsToLastLinkInDisplayableLink(), detectOwnerContainRights(), and
  * selectForGroupIDMemberItsHiddenActorsWithNextLinks(),
  * PrivilegeRightsFactoryImpl.addPrivilegeInAccessControlLink() and
  * AclRightsFactoryImpl.addAclInAccessControlLink().
  * @see #withAccessThroughNodesTreeFromEP(ARoad0.gBaseInterface.ImmutableSource, ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.gBaseInterface.ImmutableGroupIDMember) 
  * @see #getMergedInheritedAclPriRightsAndComments(ARoad0.gBaseInterface.ImmutableSource, ARoad0.gBaseInterface.ImmutableLeaf, ARoad0.CNot.AccessControlLinkImpl, ARoad0.gBaseInterface.StringRight[])
  * @param _res Resource or VirtualFolder of this instance
  * may be accessed or not through its parent tree. May be null.
  * @param _ep the _res userID or not, or an Actor. May be null. If not null and is
  * an account, the account primary group may deliver the inherited rights if
  * the account does not provide it.
  * @param _grp the _res groupID or not. May be null.
  * @return the link with the rights given by the parent and some comments,
  * for one of the type AccessControlLink.ACL, LPRI, or GLOBAL for AG rights.
  * It is not possible to activate more than one of the three types.
  * May be null or without rights. Null if _acc and _grp are both  not null, or if _res,
  * its ACs or its parent is null. Without rights if _res has no parent.
  * */
  public AccessControlLinkImpl getL_accessRightsThroughNodesTree(
          ImmutableLeaf _res, ImmutableEligibleParty _ep, ImmutableGroupID _grp) {

                                   //for testing and tracking
                                  //if (_res != null && _ep != null)
                                    //  System.out.println(
                                    //"start of NodeRightsImpl.getL_accessRightsThroughNodesTree(), "+
                                    //" for _res "+_res.getName()+
                                    //" and for _acc "+_ep.getName());

      if (_res == null || _res.getEorACS() == null) return null;
      if (_res.getEorParent() == null) return new AccessControlLinkImpl();
      if (_ep == null && _grp == null) return null;
      if (_ep != null && _res.getEorACS() != _ep.getEorACS())
          return null;
      if (_grp != null && _res.getEorACS() != _grp.getEorACS())
          return null;

      //if (_ep != null && _grp != null)
        //  throw new InternalError("NodeRightsImpl.getL_accessRightsThroughNodesTree() with "+
          //        "two non-null account and group as arguments.");
      if (_ep != null && !(_ep instanceof ImmutableUserID) &&
              _res.getEorACS().getManageAGORightsInheritance() )
          throw new InternalError("NodeRightsImpl.getL_accessRightsThroughNodesTree() with "+
                  "there are AGO inherited rights with _ep which is not an userID: "+
                  _ep.getName());

      StringRight[]             l_result = null;
      AccessControlLinkImpl     result = null;
      StringRight[]             l_interm = null;
      ImmutableGroupID[]        l_grps = null;
      ImmutableUserID           account = null;
      ImmutableGroupID          grp = null;
      String                    name2 = null;
      Map<String, StringRight[]> m_rights = null;

      // process the AGO rights inheritance first
      if (_res instanceof Resource && _res.getEorACS().getManageAGORightsInheritance() ) {
          if (_ep != null &&
                  _ep == ((Directory)_res.getEorParent()).getEorUserID()) {

                        l_result = ((Resource)_res).getL_inheritedOwnerRights();

          } else if (_ep != null &&
                  ((ImmutableUserID)_ep).getEorPrimGroup() ==
                    ((Directory)_res.getEorParent()).getEorGroupID()) {

                        l_result = ((Resource)_res).getL_inheritedGroupRights();

          } else if (_ep == null &&
                  _grp == ((Directory)_res.getEorParent()).getEorGroupID()) {

                        l_result = ((Resource)_res).getL_inheritedGroupRights();

          } // else if
          

          if ((l_result == null || l_result.length == 0)) {
              if (_ep != null && _res.getEorACS().getManageGroups()) {

                        l_grps = ((ImmutableUserID)_ep).getEorL_GroupID();
                        for (int i = 0; i < l_grps.length; i++) {
                            if (l_grps[i] == ((Directory)_res.getEorParent()).getEorGroupID()) {
                                l_result = ((Resource)_res).getL_inheritedGroupRights();
                                break;
                            } /// if
                        } // for i

              } else if (_grp != null && _res.getEorACS().getManageGroupTrees()) {

                        l_grps = _grp.getEorL_GroupID();
                        for (int i = 0; i < l_grps.length; i++) {
                            if (l_grps[i] == ((Directory)_res.getEorParent()).getEorGroupID()) {
                                l_result = ((Resource)_res).getL_inheritedGroupRights();
                                break;
                            } /// if
                        } // for i

              } // else
          } // if (l_result == null)

          l_grps = null;

          if (l_result == null || l_result.length == 0)
              return null;
          else {
              result = new AccessControlLinkImpl();
              result.setRights(AccessControlLink.GLOBAL, l_result);
              return result;
          } // else

      } // if _res.getEorACS().getManageAGORightsInheritance()
      
      else if (_res instanceof ImmutablePrivilegeTarget &&
              _res.getEorACS().getManageLPRIRightsInheritance()) {

          m_rights = ((ImmutablePrivilegeTarget)_res).
                        getM_TargetInheritedLinkedPRIRights();
      }
      else if (_res instanceof ImmutableResource &&
              _res.getEorACS().getManageACLRightsInheritance()) {

          m_rights = ((ImmutableResource)_res).
                        getM_TargetInheritedACLRights();
      } // else

      if (m_rights.isEmpty()) {

                       // for testing and tracking
                      //if (_res != null && _ep != null)
                        //  System.out.println("NodeRightsImpl.getL_accessRightsThroughNodesTree(), "+
                          //  "empty m_rights for _res: "+_res.getName()+" and _ep "+_ep.getName());
                      //else if (_res != null && _grp != null)
                        //  System.out.println("NodeRightsImpl.getL_accessRightsThroughNodesTree(), "+
                          //  "empty m_rights for _res: "+_res.getName()+" and _grp "+_grp.getName());
                      //else if (_ep != null && _grp != null)
                        //  System.out.println("NodeRightsImpl.getL_accessRightsThroughNodesTree(), "+
                         //   "empty m_rights for _ep: "+_ep.getName()+" and _grp "+_grp.getName());

          m_rights = null;
          return null;
      } // if

      // process the ACL or the LPRI rights inheritance
      if (_ep != null && _ep instanceof ImmutableUserID) {
        account = (ImmutableUserID)_ep;
        // the direct parent has the string '1', and its direct parent
        // has the string '2' as parent order in the last name component, and so on.
        // initialize l_matchingKeys with the names containing the member name
        // in the order of m_rights.keySet() iterator
        for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)account.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() )) {
                l_interm = (StringRight[])m_rights.get(name);
                if (l_interm.length != 0) {
                    l_result = l_interm;

                                //for testing and tracking
                              //System.out.println("NodeRightsImpl.getL_accessRightsThroughNodesTree(),"+
                                //      "for the account: "+shortName);
                              //if (l_result != null) {
                                //    for (int i = 0; i < l_result.length; i++) {
                                  //      System.out.println("---"+l_result[i]);
                                    //}
                              //} // if
                              //else
                                //    System.out.println("--- null result.");

                    break;
                } // if (l_interm.length != 0) {
            } // if (shortName
        } // for name
      } // if
      else if (_ep != null && _ep instanceof ImmutableActor) {
        for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)_ep.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() )) {
                l_interm = (StringRight[])m_rights.get(name);
                if (l_interm.length != 0) {
                    l_result = l_interm;
                    break;
                } // if (l_interm.length != 0) {
            } // if (shortName
        } // for name
          
        if (l_result == null) {
            account = ((ImmutableActor)_ep).getEorCurrentUserID();

            if (account != null) {
                for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)account.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() )) {
                        l_interm = (StringRight[])m_rights.get(name);
                        if (l_interm.length != 0) {
                            l_result = l_interm;
                            break;
                        } // if (l_interm.length != 0) {
                    } // if (shortName
                } // for name
            } // if (member
        } // if (l_result == null)
      } // if _ep instanceof ImmutableActor)

      if (l_result == null && account != null && account.getEorPrimGroup() != null) {
        grp = account.getEorPrimGroup();

        for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)account.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() )) {
                l_interm = (StringRight[])m_rights.get(name);
                if (l_interm.length != 0) {
                    l_result = l_interm;
                    break;
                } // if (l_interm.length != 0) {
            } // if (shortName
        } // for name
      } // if

      if (_grp != null) { // _ep is null so l_result is null
        grp = _grp;

        for (String name : m_rights.keySet()) {
                name2 = name.substring(0, name.lastIndexOf(":: "));

                if (name.equals( ((Name)grp.getName()).extended(
                        // add the parent order
                        name2.substring(name2.lastIndexOf(":: " + 3))).toString() )) {
                l_interm = (StringRight[])m_rights.get(name);
                if (l_interm.length != 0) {
                    l_result = l_interm;
                    break;
                } // if (l_interm.length != 0) {
            } // if (shortName
        } // for name
      } // if

                       //for testing and tracking
                      //if (_res != null && _ep != null)
                        //  System.out.println("NodeRightsImpl.getL_accessRightsThroughNodesTree(), "+
                          //  "result for _res: "+_res.getName()+" and _ep "+_ep.getName());
                      //else if (_res != null && _grp != null)
                        //  System.out.println("NodeRightsImpl.getL_accessRightsThroughNodesTree(), "+
                          //  "result for _res: "+_res.getName()+" and _grp "+_grp.getName());
                      //else if (_ep != null && _grp != null)
                        //  System.out.println("NodeRightsImpl.getL_accessRightsThroughNodesTree(), "+
                          //  "result for _ep: "+_ep.getName()+" and _grp "+_grp.getName());
                      //if (l_result != null) {
                        //    for (int i = 0; i < l_result.length; i++) {
                          //      System.out.println("---"+l_result[i]);
                            //}
                      //} // if
                      //else
                        //  System.out.println("---- null result.");

      l_grps = null;
      m_rights = null;
      account = null;
      grp = null;
      name2 = null;
      l_interm = null;

      if (l_result == null || l_result.length == 0)
          return null;
      else {
          result = new AccessControlLinkImpl();
          if (_res.getEorACS().getManageLPRIRightsInheritance())
              result.setRights(AccessControlLink.LPRI, l_result);
          else
              result.setRights(AccessControlLink.ACL, l_result);

          l_result = null;
          return result;
      } // else
  } // getL_accessRightsThroughNodesTree()

  /**
  * Gets the effective ACL or LPRI rights of the source on the target when
  * there are non-null inherited rights through the nodes tree, and adds
  * some specific comments for the ACL or PRI rights from an AcsAddon,
  * even if there is no inherited rights.
  * The argument _l_transientInheritedRights contains the inherited rights found
  * by getL_accessRightsThroughNodesTree(). In this generic NodeRightsImpl class,
  * these inherited rights are the first found ones in a bottom-up search into the targets tree.
  * In this default method an AcsAddon may overread, the inherited rights are simply 
  * added to the direct rights already set in the argument _upd_aclink, with a
  * special comment for the 'See why' text. This method also adds 'conditional ACL'
  * in the ACL comment, with the condition group names, for each enabled conditional
  * non-inherited AclEntry from the source to the leaf, if the leaf is a Resource. <P>
  * An AcsAddon may always return null even if there are non-null
  * direct rights or non-null inherited rights.
  * Called by PrivilegeRightsFactoryImpl.addPrivilegeInAccessControlLink() and
  * AclRightsFactoryImpl.addAclInAccessControlLink().
  * @param _so source for which the rights on _targ has to be set
  * @param _targ Resource or VirtualFolder of this instance. May be null.
  * @param _upd_acLink contains the direct ACL and/or LPRI rights, if any. Never null.
  * May be empty.
  * @param _l_transientInheritedRights has to be from getL_accessRightsThroughNodesTree().
  * May be null or empty.
  * @return _upd_acLink with the effective ACL or LPRI rights of the source on the target,
  * or null if the inherited rights nullify the result.
  * */
  public AccessControlLinkImpl getMergedInheritedAclPriRightsAndComments(
          ImmutableSource _so, ImmutableLeaf _targ,
          AccessControlLinkImpl _upd_acLink, StringRight[] _l_transientInheritedRights) {

        StringRight     right = null;
        char            open = '[';
        char            close = ']';
        char            space = ' ';

        if (_l_transientInheritedRights != null) {
            if (_targ.getEorACS().getManageACLRightsInheritance()) {
              for (int i = 0; i < _l_transientInheritedRights.length; i++) {
                right = _l_transientInheritedRights[i];
                if (right == null || right.equals(StringRightImpl.EMPTY_INSTANCE)) continue;

                if (right.isGrantingRight()) {
                  _upd_acLink.addRight(AccessControlLink.ACL, right );
                  _upd_acLink.addComment(AccessControlLink.ACL,
                                            AccessControlLink.C_INHERITED_ACL);
                  _upd_acLink.addComment(AccessControlLink.ACL,
                          right.getRight().replace(open, space).replace(close, space) );
                } else {
                  _upd_acLink.addRight(AccessControlLink.NEG_ACL, right);
                  _upd_acLink.addComment(AccessControlLink.NEG_ACL,
                                            AccessControlLink.C_INHERITED_ACL);
                  _upd_acLink.addComment(AccessControlLink.NEG_ACL,
                          right.getRight().replace(open, space).replace(close, space) );
                } // else
              } // for i
            }
            else if (_targ.getEorACS().getManageLPRIRightsInheritance()) {
              for (int i = 0; i < _l_transientInheritedRights.length; i++) {
                right = _l_transientInheritedRights[i];
                if (right == null || right.equals(StringRightImpl.EMPTY_INSTANCE)) continue;

                if (right.isGrantingRight()) {
                  _upd_acLink.addRight(AccessControlLink.LPRI, right );
                  _upd_acLink.addComment(AccessControlLink.LPRI,
                                            AccessControlLink.C_INHERITED_PRI);
                  _upd_acLink.addComment(AccessControlLink.LPRI,
                          right.getRight().replace(open, space).replace(close, space) );
                } else {
                  _upd_acLink.addRight(AccessControlLink.NEG_LPRI, right);
                  _upd_acLink.addComment(AccessControlLink.NEG_LPRI,
                                            AccessControlLink.C_INHERITED_PRI);
                  _upd_acLink.addComment(AccessControlLink.NEG_LPRI,
                          right.getRight().replace(open, space).replace(close, space) );
                } // else
              } // for i
            } // else
        } // if (_l_transientInheritedRights != null)

        right = null;

        return _upd_acLink;
  } // getMergedInheritedAclPriRightsThroughNodesTree()

  /**
   * Detects the view target that is linked to an actor, to extend
   * the current DisplayableLink ended by this actor.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by
   * ThreeNodesRighsFactoryImpl.endsPathsFromActorWithAclBridgePrivilegeAlias() and
   * selectOneHiddenNodePathsForVirtualFolder().
   * Overridden by the subclasses.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _target view node, not node of this instance. Never null.
   * @param _dLink with an actor as second end and node of this instance. Never null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @return new DisplayableLinkImpl that is an extension of _dlink
   * with _target as the second node, or null. Never empty.
  * */
  public DisplayableLinkImpl selectAddonLastViewNodeAfterActor(
          EPRViewInBase _viewInBase, ImmutableTarget _target,
          DisplayableLinkImpl _dLink, Map _m_l_DisplayableLinks) {
      return null;
  } // selectAddonLastViewNodeAfterActor()

  /**
   * Detects the view target that is linked to a groupIDMember, to extend
   * the current DisplayableLink ended by this groupIDMember.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by
   * ThreeNodesRighsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeAlias().
   * Overridden by the subclasses.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _target view node, not node of this instance. Never null.
   * @param _dLink with a groupIDMember as second end and node of this instance. Never null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @return new DisplayableLinkImpl that is an extension of _dlink
   * with _target as the second node, or null. Never empty.
  * */
  public DisplayableLinkImpl selectAddonLastViewNodeAfterGroupIDMember(
          EPRViewInBase _viewInBase, ImmutableTarget _target,
          DisplayableLinkImpl _dLink, Map _m_l_DisplayableLinks) {
      return null;
  } // selectAddonLastViewNodeAfterGroupIDMember()

  /**
   * Detects the view target that is linked to a VirtualFolder, to extend
   * the current DisplayableLink ended by this VirtualFolder.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * final loop on the view nodes, and by
   * ThreeNodesRighsFactoryImpl.endsPathsFromVirtualFolderWithMemberAlias()
   * Overridden by the subclasses.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _target view node, not node of this instance. Never null.
   * @param _dLink with a VirtualFolder as second end and node of this instance. Never null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @return new DisplayableLinkImpl that is an extension of _dlink
   * with _target as the second node, or null. Never empty.
  * */
  public DisplayableLinkImpl selectAddonLastViewNodeAfterVirtualFolder(
          EPRViewInBase _viewInBase, ImmutableResource _target,
          DisplayableLinkImpl _dLink, Map _m_l_DisplayableLinks) {

      return null;
  } // selectAddonLastViewNodeAfterVirtualFolder()

  /**
   * Detects the hidden nodes that are linked to an Actor as access source,
   * to extend the current DisplayableLinks ended by this actor.
   * Returns the no-proxy actors of the _node ACS in _upd_l_NoProxyOrNoExecuteNodes.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * starting and central loops.
   * Two lists are returned to separate the executing right-proxy nodes.
   * The last argument may be updated to provide the second returned value,
   * and it contains only the nodes on which the property changes have to be listened.
   * Overridden and called by the subclasses.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _node node of this instance. Never null.
   * @param _l_dLinks list of links with _node as second end. May be null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @param _upd_l_NoProxyOrNoExecuteNodes updated by the adding of the hidden nodes
   * without executing right or which are not right-proxy nodes, if any.
   * This list is only extended if necessary, as a complement of the returned value of the method.
   * This argument is usually empty at the call of this method, but this is not mandatory.
   * Never null.
   * @return new DisplayableLinkImpls that are extensions or updatings 
   * of some DisplayableLinkImpls in _l_dLinks with an Actor, GroupIDMember
   * or VirtualFolder as the second end, or null. If _l_dLinks is null,
   * returns only simple links. Always null.
  * */
  public List<DisplayableLinkImpl> selectAddonNewHiddenNodeForActor(
          EPRViewInBase _viewInBase, ImmutableActor _node,
          Set<DisplayableLinkImpl> _l_dLinks, Map _m_l_DisplayableLinks,
          List<ACSObject> _upd_l_NoProxyOrNoExecuteNodes) {

      _upd_l_NoProxyOrNoExecuteNodes.addAll(
              ((ACSImpl)_node.getEorACS()).getEorL_NoRightProxyActors());
      return null;
  } // selectAddonNewHiddenNodeForActor()

  /**
   * Detects the hidden nodes that are linked to a GroupIDMember as access source,
   * to extend the current DisplayableLinks ended by this GroupIDMember.
   * Returns the no-proxy actors of the _node ACS in _upd_l_NoProxyOrNoExecuteNodes.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * starting and central loops.
   * Two lists are returned to separate the executing right-proxy nodes.
   * The last argument may be updated to provide the second returned value,
   * and it contains only the nodes on which the property changes have to be listened.
   * Overridden and called by the subclasses.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _node node of this instance. Never null.
   * @param _l_dLinks list of links with _node as second end. May be null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @param _upd_l_NoProxyOrNoExecuteNodes updated by the adding of the hidden nodes
   * without executing right or which are not right-proxy nodes, if any.
   * This list is only extended if necessary, as a complement of the returned value of the method.
   * This argument is usually empty at the call of this method, but this is not mandatory.
   * Never null.
   * @return new DisplayableLinkImpls that are extensions or updatings 
   * of some DisplayableLinkImpls in _l_dLinks with an Actor, GroupIDMember
   * or VirtualFolder as the second end, or null. If _l_dLinks is null,
   * returns only simple links. Always null.
  * */
  public List<DisplayableLinkImpl> selectAddonNewHiddenNodeForGroupIDMember(
          EPRViewInBase _viewInBase, ImmutableGroupIDMember _node,
          Set<DisplayableLinkImpl> _l_dLinks, Map _m_l_DisplayableLinks,
          List<ACSObject> _upd_l_NoProxyOrNoExecuteNodes) {

      _upd_l_NoProxyOrNoExecuteNodes.addAll(
              ((ACSImpl)_node.getEorACS()).getEorL_NoRightProxyActors());
      return null;

  } // selectAddonNewHiddenNodeForGroupIDMember()

  /**
   * Detects the hidden nodes that are linked to a VirtualFolder as access source,
   * to extend the current DisplayableLinks ended by this VirtualFolder.
   * Returns the no-proxy actors of the _node ACS in _upd_l_NoProxyOrNoExecuteNodes.
   * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights() in the
   * starting and central loops.
   * Two lists are returned to separate the executing right-proxy nodes.
   * The last argument may be updated to provide the second returned value,
   * and it contains only the nodes on which the property changes have to be listened.
   * Overridden and called by the subclasses.
   * @param _viewInBase EPRViewInBase. Never null.
   * @param _node node  of this instance. May be a Resource VirtualFolder. Never null.
   * @param _l_dLinks list of links with _node as second end. May be null.
   * @param _m_l_DisplayableLinks immutabke Map of DisplayableLinks lists (one per pair)
   * to never update in this method.
   * @param _upd_l_NoProxyOrNoExecuteNodes updated by the adding of the hidden nodes
   * without executing right or which are not right-proxy nodes, if any.
   * This list is only extended if necessary, as a complement of the returned value of the method.
   * This argument is usually empty at the call of this method, but this is not mandatory.
   * Never null.
   * @return new DisplayableLinkImpls that are extensions or updatings 
   * of some DisplayableLinkImpls in _l_dLinks with an Actor as the second end,
   * or null. If _l_dLinks is null, returns only simple links.
   * Always null.
  * */
  public List<DisplayableLinkImpl> selectAddonNewHiddenNodeForVirtualFolder(
          EPRViewInBase _viewInBase, ImmutableVirtualFolder _node,
          Set<DisplayableLinkImpl> _l_dLinks, Map _m_l_DisplayableLinks,
          List<ACSObject> _upd_l_NoProxyOrNoExecuteNodes) {

      _upd_l_NoProxyOrNoExecuteNodes.addAll(
              ((ACSImpl)_node.getEorACS()).getEorL_NoRightProxyActors());
      return null;
  } // selectAddonNewHiddenNodeForVirtualFolder()

  /**
  * Detects the main GroupID and, recursivelly, the tree of GroupIDs
  * in the _ep ACS, that contains _ep indirectly, and are not in _viewInBase.
  * Detects the MEMBER and GLOBAL relationships without any setting of the link types
  * nor the rights. Puts a comment only for AccessControlLink.CONTAIN, through the
  * adding of '<WEAK>'.
  * The search of indirect relations is provided through up to 40 iterations.
  * Called by detectHiddenChainedGroupsRights().
  * @param _viewInBase EPRViewInBase
  * @param _ep node of this instance; is in _viewInBase
  * @exception InternalError if the number of iterations is up to 40.
  * @return Set of DisplayableLinkImpl for which the first end is _ep,
  *   the second end is not in _viewInBase, and for which the second end and
  *   the intermediate nodes are the actor UserID or GroupIDs that contain _ep.
  *   The rights are not set. May be null but not empty.
  **/
  public Set selectHiddenButNoDirectGroupLinks(
    EPRViewInBase _viewInBase, ImmutableEligibleParty _ep) {
    // variables
    Set                   l_result = null;
    Set                   l_hiddenOwners = new HashSet(INITIAL_CAPACITY / 3);
    Set                   l_initialGroups = null;

    if (_ep instanceof ImmutableGroupIDMember)
      l_initialGroups = selectHiddenDirectGroupIDForGroupIDMember(
        _viewInBase, (ImmutableGroupIDMember)_ep);
    else if (_ep instanceof ImmutableActor && ((ImmutableActor)_ep).getIsRightProxy() )
      l_initialGroups = selectHiddenDirectOwnerContainGlobalForActorAsEP(
        _viewInBase, (ImmutableActor)_ep);

    // it is empty when it is an extended view (i.e. all direct owners are in the view)
    if (l_initialGroups == null || l_initialGroups.isEmpty() ||
            !_ep.getEorACS().getManageGroupTrees())
        return l_result;

    Set                   l_groups;
    Set                   l_nextGroupsToTest = (Set) ((HashSet)l_initialGroups).clone();
    Set                   l_testedGroups = (Set) ((HashSet)l_initialGroups).clone();
    BaseObject[]          l_currentIntermediateNodes;
    Set                   l_l_currentIntermediateNodes;
    Map                   m_l_InternalDisplayableLinks = new HashMap(INITIAL_CAPACITY);
    AgoRightsFactoryImpl  ago = (AgoRightsFactoryImpl)interpreter_.getCoreAlgorithm("Ago");

    ImmutableGroupID[]    l_groups_2;
    DisplayableLinkImpl   dLink;
    ImmutableGroupIDMember      groupmember;
    int iterationsCount = 0;


    // Loop to build up all the possible paths. Stops when there is no new group to test.
    // Each iteration tries to add one new GroupID (or UserID, at the first iteration)
    // to the current paths, through GroupIDMember.getEorL_GroupID().
    // The current paths are put in an internal m_l_InternalDisplayableLinks
    // that has the _m_l_DisplayableLinks structure.
    while (!l_nextGroupsToTest.isEmpty() && iterationsCount < 40) {

      // updates l_nextGroupsToTest and l_groups
      iterationsCount++;

                        // for testing and tracking
                        //System.out.println("in NodeRightsImpl."+
                        //  "selectHiddenButNoDirectGroupLinks(), THE CURRENT STATE IS:");
                        //System.out.println("---_ep: "+_ep.getName());
                        //System.out.println("--- iterationsCount: "+iterationsCount);
                        //System.out.println("------- l_nextGroupsToTest: ");
                        //for (Iterator itetest = l_nextGroupsToTest.iterator(); itetest.hasNext();) {
                        //  System.out.println("--------------- : "+itetest.next().toString());
                        //}

      l_groups = (Set) ((HashSet)l_nextGroupsToTest).clone();
      l_nextGroupsToTest.clear();

      for (Iterator ito = l_groups.iterator(); ito.hasNext();) {

        groupmember = (ImmutableGroupIDMember) ito.next();
        l_testedGroups.add(groupmember);
        l_groups_2 = groupmember.getEorL_GroupID();


                        // for testing and tracking
                        //System.out.println("in NodeRightsImpl."+
                        //  "selectHiddenButNoDirectGroupLinks(), "+
                        //  "before the loop on l_groups_2:");
                        //System.out.println("------- current iteration on group: "+group.getName());
                        //System.out.println("------- group.getEorL_GroupID(): ");
                        //for (int i = 0; i < l_groups_2.length; i++) {
                        //  if (l_groups_2[i] != null) System.out.println("--------------- : "+l_groups_2[i].toString());
                        //}
                        //System.out.println("------- l_testedGroups: ");
                        //for (Iterator itetest = l_testedGroups.iterator(); itetest.hasNext();) {
                        //  System.out.println("--------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("------- l_hiddenOwners: ");
                        //for (Iterator itetest = l_hiddenOwners.iterator(); itetest.hasNext();) {
                        //  System.out.println("--------------- : "+itetest.next().toString());
                        //}
                        //System.out.println("------- l_result: ");
                        //for (Iterator itetest = l_result.iterator(); itetest.hasNext();) {
                        //  System.out.println("--------------- : "+itetest.next().toString());
                        //}

        // gets all the previous intermediate nodes lists
        // that ends by 'group', if they exist.
        // There, m_l_InternalDisplayableLinks has be updated with result
        // in a previous iteration
        l_l_currentIntermediateNodes = linkUtil_.
          getIntermediateNodesListsForPair(
            ((Name)_ep.getName()).extended(groupmember.getName()),
            m_l_InternalDisplayableLinks);

        // it may be empty, so creates it
        if (l_l_currentIntermediateNodes.isEmpty() ) {
          l_currentIntermediateNodes = new BaseObject[] {groupmember};
          l_l_currentIntermediateNodes.add(l_currentIntermediateNodes);
        }

        for (int j = 0; j < l_groups_2.length; j++) {
          if (l_groups_2[j] != null &&
            !_viewInBase.containsSource(l_groups_2[j]) &&
            !l_initialGroups.contains(l_groups_2[j])
            ) {

                            // for testing and tracking
                            //System.out.println("in NodeRightsImpl."+
                            //  "selectHiddenButNoDirectGroupLinks(), start of "+
                            //  "the loop for l_groups_2[j] "+
                            //  l_groups_2[j].getName().lastComponent());

              for (Iterator iti = l_l_currentIntermediateNodes.iterator(); iti.hasNext();) {

                  l_currentIntermediateNodes = (BaseObject[]) iti.next();
                  /*
                  // build up the current path and get the right DisplayableLink list
                  l_currentPath = new ImmutableEligibleParty[
                    l_currentIntermediateNodes.length+1];
                  l_currentPath[0] = _ep;
                  //l_currentPath[l_currentPath.length - 1] = l_groups_2[j];
                  //index = 1;
                  for(int k = 0; k < l_currentIntermediateNodes.length; k++ ) {
                    l_currentPath[k+1] = (ImmutableEligibleParty) l_currentIntermediateNodes[k];
                    //index++;
                  } // for itu
                  */
                            // for testing and tracking
                            //System.out.println("in NodeRightsImpl."+
                            //  "selectHiddenButNoDirectGroupLinks(), the current path "+
                            //  "which is build up is:");
                            //for (int k = 0; k < l_currentIntermediateNodes.length; k++) {
                            //  System.out.println("--------------- : "+l_currentIntermediateNodes[k]);
                            //}

                  // updates the internal state variables
                  l_hiddenOwners.add(l_groups_2[j]);
                  l_nextGroupsToTest.add(l_groups_2[j]);

                  // creates and initialize dLink without rights nor rights type
                  dLink = new DisplayableLinkImpl(_ep, l_groups_2[j],
                     l_currentIntermediateNodes );

                  if (_ep instanceof GroupID)
                      // useless if _ep is not the effective first end, or if
                      // dLink.containsStrengthenLinkTypes() true,
                      // but this cannot be known by this method
                      dLink.getEorFirstLink().addComment(AccessControlLink.CONTAIN,
                              AccessControlLink.C_WEAK_LINK);

                  // set the link types and rights
                  dLink = ago.addOwnerContainRightsToLastLinkInDisplayableLink(
                    dLink);

                            // for testing and tracking
                            //System.out.println("in NodeRightsImpl."+
                            //  "selectHiddenButNoDirectGroupLinks(), "+
                            //  " add new link: "+dLink);
                            //System.out.println("in NodeRightsImpl."+
                            //  "selectHiddenButNoDirectGroupLinks(), m_l_InternalDisplayableLinks:");
                            //for (Iterator ith = m_l_InternalDisplayableLinks.keySet().iterator() ; ith.hasNext() ; ) {
                            //   pairKey = (ImmutableName) ith.next();
                            //   System.out.println("--- for key: "+pairKey);
                            //   for (Iterator itg = ((List)m_l_InternalDisplayableLinks.get(pairKey)).iterator() ; itg.hasNext() ; ) {
                            //    System.out.println("------------"+
                            //      ((DisplayableLinkImpl)itg.next()).toString() );
                            //  } // for itg
                            //} // for ith

                  if (dLink != null) {
                      if (l_result == null)
                          l_result = new HashSet(INITIAL_CAPACITY / 3);

                      // sets result and m_l_InternalDisplayableLinks
                      l_result.add(dLink);
                  } // if
            } // for (Iterator iti

          } // if (l_groups_2[j] != null &&
        } // for j
      } // for ito

    } // while

    if (iterationsCount == 40)
        throw new InternalError("in NodeRightsImpl."+
                          "selectHiddenButNoDirectGroupLinks(), "+
                          "the number of iterations is superior to 40.");

    //ep = null;
    groupmember = null;
    l_groups = null;
    l_groups_2 = null;
    dLink = null;
    l_hiddenOwners = null;
    l_nextGroupsToTest = null;
    l_testedGroups = null;
    l_initialGroups = null;
    l_currentIntermediateNodes = null;
    l_l_currentIntermediateNodes = null;
    m_l_InternalDisplayableLinks = null;
    ago = null;

    return l_result;
  } // selectHiddenButNoDirectGroupLinks()

  /**
  * Detects the current UserID (including 'root'), the current GroupID and
  * the secondary groups under which an actor runs, if they are not in the view.
  * The secondary groups are added only if there is a current UserID or a
  * current GroupID. This is independent to the AGO rights management by the ACS.
  * This method is called on the NodeRights of the actor, while it should be
  * the NodeRights of the result to be comply with the AcsAddon framework.
  * Nonetheless, it works because the NodeRights of the actor is identical
  * to the NodeRights of the result. <P>
  * Note: this method is sensible, since it is called for 3-nodes paths and for
  * the longer paths.
  * Returns null if there is no current UserID, and returns only the relevant GroupIDs
  * if the current UserID is in the view.
  * No control to check up if the actor is a right proxy.
  * Called by selectHiddenDirectOwnerContainLinksForActor(),
  * CompoundRightsFactoryImpl.detectOneHiddenNodeCompoundRights() and
  * ThreeNodesRightsFactoryImpl.detectHiddenCompoundEpRights().
  * @param _viewInBase EPRViewInBase
  * @param _act node of this instance; is in _viewInBase
  * @return the UserID and the GroupIDs under which _res runs and which are not
  * in the view. May be null but not empty.
  **/
  public Set<ImmutableGroupIDMember> selectHiddenDirectOwnerContainGlobalForActorAsEP(
    EPRViewInBase _viewInBase, ImmutableActor _act) {

    // if the current account is null, this actor has the context of its caller
    if (_act.getEorCurrentUserID() == null) return null;

    // variables
    Set<ImmutableGroupIDMember> l_result = null;
    ImmutableGroupID[]          l_groups = _act.getEorL_SecondaryGroup();

    if (!_viewInBase.containsSource(_act.getEorCurrentUserID())) {

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                    //"selectHiddenDirectOwnerContainGlobalForActorAsEP(), "+
                    //"add of its current userID for the actor "+_act.getName());

        l_result = new HashSet(INITIAL_CAPACITY);
        l_result.add(_act.getEorCurrentUserID());
    }
    else {

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                    //"selectHiddenDirectOwnerContainGlobalForActorAsEP(), "+
                    //"its non-null current userID is NOT added for the actor "+_act.getName());
    }

    if (_act.getEorCurrentGroupID() != null &&
         !_viewInBase.containsSource(_act.getEorCurrentGroupID())) {

        if (l_result == null)
            l_result = new HashSet(INITIAL_CAPACITY);
        l_result.add(_act.getEorCurrentGroupID());
    }

    if (l_groups.length != 0 && l_result != null) {
        if (l_result == null)
            l_result = new HashSet(INITIAL_CAPACITY);

        for (int i = 0; i < l_groups.length; i++) {
            if (!_viewInBase.containsSource(l_groups[i]))
                l_result.add(l_groups[i]);
        }
    } // if

    l_groups = null;

    if (l_result != null && l_result.isEmpty()) {
        l_result = null;
        return null;
    }
    else
        return l_result;
  } // selectHiddenDirectOwnerContainGlobalForActorAsEP()

  /**
  * Detects the current UserID (including 'root'), the current GroupID and
  * the secondary groups under which an actor runs, if they are in the view.
  * Returns null if there is no current UserID. No control to check up if the actor
  * is a right proxy.
  * Called by detectAGRights() and 
  * ThreeNodesRightsFactoryImpl.selectOneHiddenNodePathsForVirtualFolder(),
  * selectOneHiddenNodePathsForFirstLinkPrivileges(), addPathsFromActorBridgeActor(),
  * addPathsFromActorAclActor(), addPathsFromActorAcsAddonRelationActor(),
  * addPathsFromGroupIDMemberAclActor(),
  * addPathsFromGroupIDMemberAcsAddonRelationActor().
  * @param _viewInBase EPRViewInBase
  * @param _act node of this instance; is in _viewInBase
  * @return the UserID and the GroupID under which _res runs and which are
  * in the view. May be null but not empty.
  **/
  public Set<ImmutableGroupIDMember> selectDirectOwnerContainGlobalForActorAsEP(
    EPRViewInBase _viewInBase, ImmutableActor _act) {

    // if the current account is null, this actor has the context of its caller
    if (_act.getEorCurrentUserID() == null) return null;

    // variables
    Set<ImmutableGroupIDMember>                 l_result = null;
    ImmutableGroupID[]  l_groups = _act.getEorL_SecondaryGroup();

    if (_viewInBase.containsSource(_act.getEorCurrentUserID())) {
        l_result = new HashSet(INITIAL_CAPACITY);
        l_result.add(_act.getEorCurrentUserID());
    }

    if (_act.getEorCurrentGroupID() != null &&
         _viewInBase.containsSource(_act.getEorCurrentGroupID())) {

        if (l_result == null)
            l_result = new HashSet(INITIAL_CAPACITY);
        l_result.add(_act.getEorCurrentGroupID());
    }

    if (l_groups.length != 0) {
        if (l_result == null)
            l_result = new HashSet(INITIAL_CAPACITY);

        for (int i = 0; i < l_groups.length; i++) {
            if (_viewInBase.containsSource(l_groups[i]))
                l_result.add(l_groups[i]);
        }
    } // if

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectDirectOwnerContainGlobalForActorAsEP(), ");

    l_groups = null;

    if (l_result != null && l_result.isEmpty()) {
        l_result = null;
        return null;
    }
    else
        return l_result;
  } // selectDirectOwnerContainGlobalForActorAsEP()


  /**
  * Detects the UserID and all the GroupIDs in the ACS, that owns or contains
  * an actor directly as a Resource, and are not in _viewInBase.
  * The link type is given in that order:
  * 'owner' includes, in that order:
  * - the userID that owns the actor,
  * 'contain' includes, in that order:
  * - the primary groupID that contains the actor.
  * Then, all the tree of the 'contains' relations are applicable,
  * plus the 'own' relations.
  * Do not process the secondary groups where _act is, since it is
  * an EligibleParty property of Actor.
  * Called by detectHiddenChainedGroupsRights() and
  * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
  * Calls selectHiddenDirectOwnerContainForTarget().
  * @param _viewInBase EPRViewInBase
  * @param _act node of this instance; is in _viewInBase
  * @return Set of UserID and GroupIDs that owns/contain _res and which are not
  * in the view. May be null but not empty.
  **/
  public Set selectHiddenDirectOwnerContainForActorAsResource(
    EPRViewInBase _viewInBase, ImmutableActor _act) {

    return selectHiddenDirectOwnerContainForTarget(_viewInBase, _act);
  } // selectHiddenDirectOwnerContainForActorAsResource()

  /**
  * Detects all the GroupIDs in the ACS, for which
  * a GroupIDMember (even for 'root') is a member, and are not in _viewInBase.
  * Called by selectHiddenButNoDirectGroupLinks() and
  * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
  * @param _viewInBase EPRViewInBase
  * @param _memb node of this instance; is in _viewInBase
  * @return Set of GroupIDs for which _memb is a member and which are not
  * in the view. May be null but not empty.
  **/
  public Set selectHiddenDirectGroupIDForGroupIDMember(
    EPRViewInBase _viewInBase, ImmutableGroupIDMember _memb) {
      // variables
      Set                   l_result = null;
      ImmutableGroupID[]    l_groups = _memb.getEorL_GroupID(); // never null

      for (int i = 0; i < l_groups.length; i++) {

        if (l_groups[i] != null && !_viewInBase.containsSource(l_groups[i])) {
            if (l_result == null)
                l_result = new HashSet(INITIAL_CAPACITY / 3);
            l_result.add(l_groups[i]);

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectHiddenDirectGroupIDForGroupIDMember(), member is "+
                  //  _memb);
                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectHiddenDirectGroupIDForGroupIDMember(), l_groups[i] is "+
                  //  l_groups[i]);
                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectHiddenDirectGroupIDForGroupIDMember(), "+
                  //  _memb.getName().lastComponent()+" is a GroupIDMember owned by: " +
                  //  l_groups[i].getName().lastComponent());

        } // if
      } // for i

      l_groups = null;

      return l_result;
  } // selectHiddenDirectGroupIDForGroupIDMember()


  /**
  * Detects the UserID and the GroupID in the _res ACS,
  * that owns or contains _res directly, and are not in _viewInBase.
  * The link type is given in that order: <P>
  * - the userID that owns the resource, <P>
  * - the groupID that contains the resource. <P>
  * As selectHiddenDirectOwnerContainForActorAsResource(), but without
  * search of the actor secondary groups.
  * Called by detectHiddenChainedGroupsRights()
  * and selectHiddenDirectOwnerContainForActor().
  * @param _viewInBase EPRViewInBase
  * @param _res node of this instance; is in _viewInBase
  * @return Set of UserID and GroupIDs that owns/contain _res and are not
  * in the view. May be null but not empty.
  **/
  public Set selectHiddenDirectOwnerContainForTarget(
    EPRViewInBase _viewInBase, ImmutableResource _res) {
    // variables
    Set l_result = null;

    if (_res.getEorUserID() != null
      && !_viewInBase.containsSource(_res.getEorUserID()) ) {
        if (l_result == null)
            l_result = new HashSet(2);
        l_result.add(_res.getEorUserID() );

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectHiddenDirectOwnerContainForTarget(), "+
                  //  _res.getName().lastComponent()+" is a Resource owned by: " +
                  //  _res.getEorUserID().getName().lastComponent());

    } // if (_res.getEorUserID() != null

    if (_res.getEorGroupID() != null
      && !_viewInBase.containsSource(_res.getEorGroupID()) ) {
        if (l_result == null)
            l_result = new HashSet(2);
        l_result.add(_res.getEorGroupID() );

                  // for testing and tracking
                  //System.out.println("in NodeRightsImpl."+
                  //  "selectHiddenDirectOwnerContainForTarget(), "+
                  //  _res.getName().lastComponent()+" is a Resource contained in: " +
                  //  _res.getEorGroupID().getName().lastComponent());

    } // if (_res.getEorGroupID() != null

    return l_result;
  } // selectHiddenDirectOwnerContainForTarget()

  /**
  * Detects all the actors that are owned or contained by _ep,
  * that are not in _viewInBase and executable from _ep,
  * and that are access sources for another AG context.
  * This means that the actor is in one of the following cases: <P>
  * - at least it has one external ACL as access source, <P>
  * - or at least it has one Bridge as access source, <P>
  * - or at least it has one Privilege as access source, <P>
  * - or at least if is an Alias. <P>
  * No adding of the actor whether _ep is a secondary group for the actor.
  * The ACL, Bridge or Privilege rights of the actor as source are not checked,
  * then they may be without effective rights.
  * The AGO priorities of the ACS are not checked.
  * An AcsAddon may overread this method, for instance to process the inherited AclEntries
  * or Privileges.
  * Note: the actors running under xid, and those that are not proxy, are not selected.
  * Note: if there is an ACL with the same ACS for the source and the target,
  * the AG context is not changed, but the ACL target may be an actor which is
  * a source for ACL, Bridge or Privileges, and then the path has to be explored.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),
  * detectOneHiddenNodeCompoundRights() and by
  * ThreeNodesRightsFactoryImpl.addPathsFromGroupIDMemberOwnContainActor().
  * @param _viewInBase the view to analyze
  * @param _ep node of this instance; may be in _viewInBase or not
  * @return Set of simple DisplayableLinkImpl for which the first end is _ep,
  *   the second end is an actor that is owned or contained by _ep.
  *   The owner/contain rights are set. May be null, but not empty.
  * */
  public Set<DisplayableLinkImpl> selectForGroupIDMemberItsHiddenActorsWithNextLinks(
    EPRViewInBase _viewInBase, ImmutableGroupIDMember _ep) {

    AccessControlLinkImpl acLink;
    Resource[]            l_res = null;
    List                  l_aclRes = null;
    Set<DisplayableLinkImpl>                   l_result = null;
    DisplayableLinkImpl   dLink = null;
    List<Privilege>       l_priv = null;
    StringRight[]         l_rights = null;
    AccessControlLinkImpl inhLinkRights = null;
    boolean               isExternalACLsource = false;
    AclEntry[]            l_acl = null;

    if (_ep instanceof ImmutableUserID)
        l_res = ((ImmutableUserID)_ep).getEorL_UserIDOwnResources();
    else if (_ep instanceof ImmutableGroupID)
        l_res = ((ImmutableGroupID)_ep).getEorL_GroupIDOwnResources() ;

    for (int i = 0; i < l_res.length ; i++) {

        if (!(l_res[i] instanceof ImmutableActor) ||
          !((ImmutableActor)l_res[i]).getIsRightProxy() ||
          _viewInBase.containsSource((ImmutableActor)l_res[i]) ) continue;


        // selects only executable actors from ep
        if (_ep instanceof UserID &&
                !utility_.withExecuteRight(l_res[i].getL_ownerRights(), false))
                        continue;
        if (_ep instanceof GroupID &&
                !utility_.withExecuteRight(l_res[i].getL_groupRights(), false))
                        continue;

        // works because _ep and res[i] have the same ACS
        if (!withAccessThroughNodesTreeFromEP(_ep, l_res[i], null))
                        continue;

        l_aclRes = ((ImmutableActor)l_res[i]).getEorL_AclConnectedResources();
        l_priv = ((ImmutableActor)l_res[i]).getEorL_SourcePrivileges();

        // test if l_res[i] is an external ACL source
        for (Object reso : l_aclRes) {

            l_acl = ((Resource)reso).getL_aclEntry();
            
            for (int j = 0; j < l_acl.length; j++) {
                if (l_acl[j].getEorEP() != l_res[i])
                    continue;

                if (!l_acl[j].getAcsName().equals(l_acl[j].getNameOfEpACS()) ||
                        !l_acl[j].getAcsName().equals(l_acl[j].getNameOfResACS())) {
                    isExternalACLsource = true;
                    break;
                } // if
            } // for i

            if (isExternalACLsource)
                break;
        } // for (reso

        if (isExternalACLsource || // l_res[i] is an external-ACL source
                ((ImmutableActor)l_res[i]).getIsACSBridgeSource() || // l_res[i] is Bridge source
                !l_priv.isEmpty() ||  // l_res[i] is Privilege source
                l_res[i].getEorReferenceAsInternalHardAlias() != null // as alias reference
                ) { 

                acLink = new AccessControlLinkImpl();

                // for _ep as an userID
                if (_ep instanceof ImmutableUserID ) {

                          acLink.setRights(AccessControlLink.OWNER,
                                  l_res[i].getL_ownerRights());

                            // for testing and tracking
                            //System.out.println("in AclRightsFactoryImpl."+
                              //"selectForGroupIDMemberItsHiddenActorsWithAclLinks(), add the link:");
                            //System.out.println("-----------"+dLink);

                // for _ep as GroupID
                } else if (_ep instanceof ImmutableGroupID ) {

                          acLink.setRights(AccessControlLink.CONTAIN,
                                  l_res[i].getL_groupRights());

                          acLink.addComment(AccessControlLink.CONTAIN,
                                  AccessControlLink.C_WEAK_LINK);

                            // for testing and tracking
                            //System.out.println("in AclRightsFactoryImpl."+
                              //"selectForGroupIDMemberItsHiddenActorsWithAclLinks(), add the link:");
                            //System.out.println("-----------"+dLink);
                } // else

                // does not call the l_res[i] LinkRight, and
                // works because _ep and res[i] have the same ACS
                l_rights = null;
                if (_ep instanceof ImmutableUserID)
                  inhLinkRights = getL_accessRightsThroughNodesTree(
                            l_res[i], (ImmutableUserID)_ep, null);
                else if (_ep instanceof ImmutableGroupID)
                  inhLinkRights = getL_accessRightsThroughNodesTree(
                            l_res[i], null, (ImmutableGroupID)_ep);
                l_rights = inhLinkRights.getRights(AccessControlLink.C_INHERITED_AG);

                for (int j = 0; j < l_rights.length; j++) {
                    if (_ep instanceof UserID) {
                        acLink.addRight(AccessControlLink.OWNER, l_rights[j]);
                        acLink.addComment(
                            AccessControlLink.OWNER,
                            inhLinkRights.getLinkTypeComments(AccessControlLink.C_INHERITED_AG));
                    }
                    else {
                        acLink.addRight(AccessControlLink.CONTAIN, l_rights[j]);
                        acLink.addComment(
                            AccessControlLink.CONTAIN,
                            inhLinkRights.getLinkTypeComments(AccessControlLink.C_INHERITED_AG));
                    } // else
                } // for j

                dLink = new DisplayableLinkImpl(_ep, l_res[i], acLink);
                if (l_result == null)
                    l_result = new HashSet(INITIAL_CAPACITY / 3);
                l_result.add(dLink);
        } // if ((l_aclRes != null
    } // for i

    acLink = null;
    l_res = null;
    l_aclRes = null;
    dLink = null;
    l_priv = null;
    l_rights = null;
    inhLinkRights = null;

    return l_result;
  } // selectForGroupIDMemberItsHiddenActorsWithNextLinks()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
      interpreter_ = null;
      link_ = null;
      linkUtil_ = null;
      aclFact_ = null;
      priFact_ = null;
  } // finalizeForProcess()

  /**
   * @return a descriptor
   */
  public String toString() {
      return "NODERIGHTSIMPL of the view "+interpreter_.getSource().getName();
  }

} // NodeRightsImpl

