//Title:        Access Road
//Version:      0.7.0
//Copyright:    2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import ARoad0.gWorkInterface.*;
import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;

import ARoad0.CNot.*;
import ARoad0.Gui2.GraphicView;

import ARoad0.gBaseInterface.*;

import ARoad0.gBase.StringRightImpl;
import ARoad0.gBase.UtilityImpl;

/**
 * This class is responsible for all the core algorithms about Privilege rights analysis.
 * These rights are direct or inherited access rights for a PrivilegeSource (the access user)
 * about a PrivilegeTarget (the access target). They are typed or linked privileges.
 * The inherited rights are only linked privileges.
 * Caution: this class works for the TYPED and the TYPED_CLASSED privileges, but
 * not for the TYPED_FOR_SEED and the TYPED_CLASSED_SPECIFIC privileges.
 * This class is stateless, and RightsMediatorImpl at its creation
 * setups an instance for its ViewInBase.
 * There is an indirect call to the method ACS.getEorL_TargetsForPrivilegeForType().
 * Upper-level methods begin by detect...(),
 * middle-level methods begin by select...(),
 * lower-level methods begin by add...().
 * @see ARoad0.gBase.PrivilegeAbst
 * @see ARoad0.gBaseInterface.PrivilegeForType
 * @see ARoad0.gBaseInterface.PrivilegeForLinks
 * @version 0.7.0
 */
public class PrivilegeRightsFactoryImpl implements CoreAlgorithm {
  private static final int          INITIAL_CAPACITY = 15; // default value for map: 101

  private AlgorithmInterpreter      interpreter_ = null;
  private UtilityImpl               baseUtil_ = null;
  private DisplayableLinkUtilities  linkUtil_ = null;
  private FactoryUtilities          factoryUtil_ = null;
  private Set<PrivilegeTarget>      l_targets_ = new HashSet(INITIAL_CAPACITY);

  /**
  * only one protected constructor
  * @param _interpreter algorithm interpreter
  */
  protected PrivilegeRightsFactoryImpl(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      baseUtil_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
      factoryUtil_ = (FactoryUtilities)interpreter_.getCoreAlgorithm("FactoryUtil");
  }

  /**
  * Defines the effective applicable rights of a PrivilegeSource for a PrivilegeTarget,
  * if there are direct typed or linked privileges between them.
  * Registers all the privileges in the view RightsRelatedBaseObjects set.
  * For a sketch view, only privileges from/to _center are processed.
  * Called by RightsFactory_Facade.detectEPRRights(), detectNoThanRights()
  * and detectSketchRights().
  * Caution: called by detectNoThanRights(), _viewInBase is not the view of _gview,
  * since this method uses transient EPRViewInBases with for each of them,
  * a pair of nodes from the initial NoThanViewInBase.
  * Synchronized to _viewInBase from the GraphicView call.
  * @param _gview is the graphic view to update. No use if _center is not null,
  * and not null otherwise.
  * @param _viewInBase EPRViewInBase to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view, and to update.
  * @param _center the central object of a sketch view. Null if it is not a sketch view.
  * @return Map of DisplayableLinks lists,
  *   with Privilege links associated to every pair (PrivilegeSource, PrivilegeTarget).
  * */
  protected final Map detectPrivilegeRights(GraphicView _gview,
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks, BaseObject _center) {
      // variables
      ImmutableSource         so;
      PrivilegeSource         ps;
      PrivilegeTarget         pt;
      ImmutableTarget         ob;
      AccessControlLinkImpl   acLink;
      HashSet                 l_ObjectToRegister = null;
      LinkRights              linkRights = null;

      if (_center == null)
          l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();

      for (Iterator ite = _viewInBase.getEorM_Sources().values().iterator() ; ite.hasNext() ; ) {
        so = (ImmutableSource) ite.next();
        if (!(so instanceof PrivilegeSource)) continue;
        ps = (PrivilegeSource) so;

        for (Iterator iti = _viewInBase.getEorM_Targets().values().iterator() ;
          iti.hasNext() ; ) {
            ob = (ImmutableTarget) iti.next();
            if (ob == ps || !(ob instanceof PrivilegeTarget)) continue;
            pt = (PrivilegeTarget) ob;

            // for a sketch view
            if (_center != null && _center != ps && _center != pt) continue;

            if (pt == ps) continue;

                                // for testing and tracking
                            //System.out.println("PrivilegeRightsFactoryImpl.detectPrivilegeRights(), "+
                                    //"for ep "+ps.getNickName()+" and res "+pt.getNickName()+
                                    //" call to addPrivilegeInAccessControlLink() ");

            acLink = addPrivilegeInAccessControlLink(pt, ps, null);

            if (acLink != null) {
                linkRights = interpreter_.getLinkRightsInstances().get(
                                pt.getName());
                linkRights.addDisplayableLink(
                        new DisplayableLinkImpl(ps, pt, acLink),
                        _m_l_DisplayableLinks);


                                // for testing and tracking
                            //System.out.println("PrivilegeRightsFactoryImpl.detectPrivilegeRights(), "+
                                    //"for ep "+ps.getNickName()+" and res "+pt.getNickName()+
                                    //" add the DLink: "+new DisplayableLinkImpl(ps, pt, acLink));
            } // if
            
        } // for (Iterator iti

        // to follow the source privilege changes in their sources and targets,
        // even if the privilege is not connected to any target in the view
        if (l_ObjectToRegister != null) {
            for (Object obj : ps.getEorL_SourcePrivilegeForLinks()) {
                l_ObjectToRegister.add(obj);
            } // for obj
            for (Object obj : ps.getEorL_SourcePrivilegeForTypes()) {
                l_ObjectToRegister.add(obj);
            } // for obj
        } // if (l_ObjectToRegister
      }  // for (Iterator ite

      // updates _gview to register the changes in the privileges
      if (l_ObjectToRegister != null)
          _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);

      so = null;
      pt = null;
      ps = null;
      ob = null;
      acLink = null;
      linkRights = null;

                                // for testing and tracking
                            //System.out.println("end of PrivilegeRightsFactoryImpl.detectPrivilegeRights() ");

      return _m_l_DisplayableLinks;
  } // detectPrivilegeRights()

  /**
  * Selects the non-inherited linked privileges which connect the target and the source.
  * Called by NodeRightsImpl.detectL_linkedPrivilegeRights().
  * @param _so with some privileges to filter
  * @param _res to analyze
  * @return the applicable linked privileges, or null. Never empty.
  * */
  public final List<PrivilegeForLinks> getL_linkedPrivilege(
          ImmutablePrivilegeSource _so, ImmutablePrivilegeTarget _res) {

      List<PrivilegeForLinks> l_result = _so.getEorL_SourcePrivilegeForLinks();
      PrivilegeForLinks       priv = null;
      
      if (l_result.isEmpty()) {
          
                            //for testing and tracking
                            //System.out.println(
                                    //"PrivilegeRightsFactoryImpl.getL_linkedPrivilege()"+
                                    //"from "+_so.getNickName()+" to "+_res.getNickName()+
                                    //" has found NO RIGHT.");

          return null;
      } // if

                            //for testing and tracking
                            //System.out.println(
                                    //"PrivilegeRightsFactoryImpl.getL_linkedPrivilege()"+
                                    //"from "+_so.getNickName()+" to "+_res.getNickName()+
                                    //", _so.getEorL_SourcePrivilegeForLinks() returns:");
                            //for (Iterator ite = l_result.listIterator(); ite.hasNext(); ) {
                              //priv = (PrivilegeForLinks)ite.next();
                              //System.out.println("------------- "+priv);
                            //} // for

      for (Iterator ite = l_result.listIterator(); ite.hasNext(); ) {
          priv = (PrivilegeForLinks)ite.next();
          if (!priv.getEorLinkedTargets().contains(_res) ) {
                  ite.remove();
          } // if
      } // for

      priv = null;

                            //for testing and tracking
                            //System.out.println(
                                    //"PrivilegeRightsFactoryImpl.getL_linkedPrivilege()"+
                                    //"from "+_so.getNickName()+" to "+_res.getNickName()+
                                    //", the result is:");
                            //for (Iterator ite = l_result.listIterator(); ite.hasNext(); ) {
                              //priv = (PrivilegeForLinks)ite.next();
                              //System.out.println("------------- "+priv);
                            //} // for


      if (l_result.isEmpty()) {
          l_result = null;
          return null;
      }
      else
          return l_result;
  } // getL_linkedPrivilege()

  /**
  * Gets the typed privileges which connect the target and the source.
  * Caution: works for the TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges,
  * but not for the TYPED_FOR_SEED privileges.
  * Called by addPrivilegeInAccessControlLink().
  * Calls ImmutablePrivilegesource.getEorL_SourcePrivilegeForTypes() and
  * LinkRights.isTargetForPrivilegeForType().
  * @param _so with some privileges to filter
  * @param _pt to analyze
  * @return the applicable typed privileges, or null. Never empty.
  * */
  public final List<PrivilegeForType> getL_typedPrivilege(
          ImmutablePrivilegeSource _so, ImmutablePrivilegeTarget _pt) {

      List<PrivilegeForType> l_result = _so.getEorL_SourcePrivilegeForTypes();
      PrivilegeForType       priv = null;
      LinkRights             linkRights = null;
      
      if (l_result.isEmpty()) return null;

      linkRights = interpreter_.getLinkRightsInstances().get(
                _pt.getName());
      if (linkRights == null) {
            interpreter_.addInstancesForNode(_pt);
            linkRights = interpreter_.getLinkRightsInstances().get(
                _pt.getName());
      } // if

                            // for testing and tracking
                            //System.out.println("PrivilegeRightsFactoryImpl."+
                              //"getL_typedPrivilege(), TPRI list from "+
                              //"_so.getEorL_SourcePrivilegeForTypes(): ");

      for (Iterator ite = l_result.listIterator(); ite.hasNext(); ) {
          priv = (PrivilegeForType)ite.next();

          if (!linkRights.isTargetForPrivilegeForType(priv, _pt) ) {

                            // for testing and tracking
                            //System.out.println("--- removed TPRI "+priv);

                  ite.remove();
          } // if
          else {

                            // for testing and tracking
                            //System.out.println("--- returned TPRI "+priv);
          }
      } // for

      priv = null;

      if (l_result.isEmpty()) {
          l_result = null;
          return null;
      }
      else
          return l_result;
  } // getL_typedPrivilege()

  /**
  * Detects all the Actors and the VirtualFolders that are not in _viewInBase, and
  * are directly accessible from _ep through one privilege. The rights are set.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(),
  * with true for _withExecute, and ThreeNodesRightsFactoryImpl.
  * Note: this method does not take account of the inherited linked privileges in an AcsAddon,
  * and this may be done in NodeRights.selectAddonNewHiddenNodeForGroupIDMember()
  * and selectAddonNewHiddenNodeForActor().
  * Caution: use the internal variable l_targets_.
  * Calls addPrivilegeRightsInSimpleDisplayableLink() and getL_TargetsOfSourcePrivilege().
  * @param _gview is the graphic view to update
  * @param _viewInBase view on which _gview is based
  * @param _ep is in _viewInBase or not
  * @param _withExecute true for detecting only if there is an executing
  *   right in the AccessControlLink. No use if the target is a virtual folder.
  * @return Set of simple DisplayableLinkImpls for which the first end is _ep,
  *   the second end is an Actor or a VirtualFolder not in the view.
  *   May be null, but not empty.
  * */
  public Set<DisplayableLinkImpl> selectDirectHiddenPrivilegeLinksForEP(
        GraphicView _gview,
        EPRViewInBase _viewInBase, ImmutableEligibleParty _ep,
        boolean _withExecute) {

                            // for testing and tracking
                            //System.out.println("in PrivilegeRightsFactoryImpl."+
                              //"selectDirectHiddenPrivilegeLinksForEP(), "+
                              //"start for the EP");
                              //System.out.println("-----------"+_ep.getName());

    if (!(_ep instanceof ImmutablePrivilegeSource))
        return null;

    ImmutablePrivilegeTarget    pt;
    Set<DisplayableLinkImpl>    l_result = null;
    DisplayableLinkImpl         dLink = null;
    // this method returns l_targets_
    Set<PrivilegeTarget>        l_targets = getL_TargetsOfSourcePrivilege(_ep);
    HashSet                     l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();
    NodeRights                  nodeRights = null;

                            // for testing and tracking
                            //System.out.println("in PrivilegeRightsFactoryImpl."+
                              //"selectDirectHiddenPrivilegeLinksForEP(), l_ObjectToRegister:");
                            //System.out.println("-----------"+l_ObjectToRegister);
                            //System.out.println(", and _gview:");
                            //System.out.println("-----------"+_gview.getName());

    // to follow the source privilege changes in their sources and targets,
    // even if the privilege is not connected to any target in the view
    for (Object obj : _ep.getEorL_SourcePrivilegeForLinks()) {

                            // for testing and tracking
                            //System.out.println("in PrivilegeRightsFactoryImpl."+
                              //"selectDirectHiddenPrivilegeLinksForEP(), "+
                              //"_ep.getEorL_SourcePrivilegeForLinks() item:");
                              //System.out.println("-----------"+obj);

        l_ObjectToRegister.add(obj);
    } // for obj
    for (Object obj : _ep.getEorL_SourcePrivilegeForTypes()) {
        l_ObjectToRegister.add(obj);
    } // for obj

    // updates _gview to register the changes in the privileges
    _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);

    if (l_targets == null) {
        return l_result;
    } // if

    for (Iterator iti = l_targets.iterator(); iti.hasNext() ; ) {
            pt = (ImmutablePrivilegeTarget) iti.next();

                            // for testing and tracking
                            //System.out.println("in PrivilegeRightsFactoryImpl."+
                              //"selectDirectHiddenPrivilegeLinksForEP(), analyzes the target:");
                            //System.out.println("-----------"+pt.getName());


            if (!(pt instanceof ImmutableSource) ||
                  (!(pt instanceof ImmutableActor) && !(pt instanceof ImmutableVirtualFolder)) ||
                  _viewInBase.containsSource((ImmutableSource)pt)) continue;

            nodeRights = interpreter_.getNodeRightsInstances().get(pt.getName());
            if (nodeRights == null) {
                interpreter_.addInstancesForNode(pt);
            } // if

            dLink = addPrivilegeRightsInSimpleDisplayableLink(_ep, pt, _withExecute);
            if (dLink != null) {
              if (l_result == null)
                  l_result = new HashSet(INITIAL_CAPACITY / 3);
              l_result.add(dLink);

                            // for testing and tracking
                            //System.out.println("in PrivilegeRightsFactoryImpl."+
                              //"selectDirectHiddenPrivilegeLinksForEP(), add the simple link:");
                            //System.out.println("-----------"+dLink);
            } // if
    } // for iti

    pt = null;
    dLink = null;
    nodeRights = null;
    return l_result;
  } // selectDirectHiddenPrivilegeLinksForEP()

  /**
  * Selects all the direct targets for a PrivilegeSource, through a typed or
  * a linked privilege.
  * Caution: works for TYPED, TYPED_CLASSED, TYPED_CLASSED_SPECIFIC privileges, but
  * not for the TYPED_FOR_SEED privileges.
  * Caution: clear and may return the internal variable l_targets_.
  * Called by selectDirectHiddenPrivilegeLinksForEP().
  * Calls ImmutablePrivilegeSource.getEorL_SourcePrivilegeForTypes() and
  * getEorL_TargetPrivilegeForTypes(), PrivilegeForLinks.getEorLinkedTargets() and
  * LinkRights.getEorL_TargetsForPrivilegeForType() with, as argument,
  * the ACS of the typed privilege, not the _ps ACS.
  * @param _ps to analyze
  * @return all the direct targets of _ps, or null if there is no target
  * */
  public final Set<PrivilegeTarget> getL_TargetsOfSourcePrivilege(
          ImmutablePrivilegeSource _ps) {

      l_targets_.clear();
      LinkRights             linkRights = null;

      linkRights = interpreter_.getLinkRightsInstances().get(
                _ps.getName());
      if (linkRights == null) {
            interpreter_.addInstancesForNode(_ps);
            linkRights = interpreter_.getLinkRightsInstances().get(
                _ps.getName());
      } // if

      for (Object lpri : _ps.getEorL_SourcePrivilegeForLinks()) {
          l_targets_.addAll(((PrivilegeForLinks)lpri).getEorLinkedTargets());
      } // for
      for (Object tpri : _ps.getEorL_SourcePrivilegeForTypes()) {
        l_targets_.addAll(linkRights.getEorL_TargetsForPrivilegeForType(
                                    (PrivilegeForType)tpri, ((PrivilegeForType)tpri).getEorACS()));
      } // for

      if (!l_targets_.isEmpty()) {
          return l_targets_;
      }
      else {
          return null;
      }
  } // getL_TargetsOfSourcePrivilege()

  //
  // --------- sensible methods ----------------------------------------
  //

  /**
  * Creates the simple DisplayableLinkImpl associated to a view object pair,
  * and updates the created AccessControlLink with typed or linked Privilege rights.
  * The main function is to handle the need to have executing rights
  * when the last argument requires it, and this is necessary for the intermediate
  * targets in the DisplayableLinks.
  * Each right is provided by the ACS if the privilege is external,
  * and it may have several ACS which are implied in one DisplayableLinkImpl.
  * So, puts in comments the privilege ACS name for each right.
  * Note : this method is sensible since it is used for both
  * the 3-nodes method and some more-than-3-nodes methods.
  * Called by selectDirectHiddenPrivilegeLinksForEP().
  * Calls addPrivilegeInAccessControlLink().
  * Note: when the argument _withExecute is true and activated,
  * the method is much more complex.
  * AccessControlLinkImpl.getRedondantRights() is called to have all the rights
  * from all the implied ACSs, with one positive or negative right per ACS, and
  * BaseUtilityImpl.withExecuteRight() gets the effective global rights to evaluate
  * if there is an executing right. This means that there is no priorities
  * among the implied ACS.
  * @param _ps is a right user
  * @param _pt is a right target. Has to be known in the AlgorithmInterpreter.
  * @param _withExecute true for returning null if there is no effective executing
  *   right in the link. No use when _pt is a virtual folder.
  * @return a new DisplayableLinkImpl. Null if the arguments are the same,
  * or if there is no privilege to add.
  * */
  public final DisplayableLinkImpl addPrivilegeRightsInSimpleDisplayableLink(
     ImmutablePrivilegeSource _ps, ImmutablePrivilegeTarget _pt, boolean _withExecute) {

        if (_ps == _pt) return null;

        // variables
        DisplayableLinkImpl     result = null;
        AccessControlLinkImpl   acLink = null;

        // updates GLOBAL if there are rights from several ACS
        acLink = addPrivilegeInAccessControlLink(_pt, _ps, acLink);

        if (acLink == null) {
          return null;
        } // if

        if (_withExecute && !(_pt instanceof ImmutableVirtualFolder)) {

            // withExecuteRight is called only if there are not rights from several ACS in acLink
            if (acLink.getLinkTypeComments(AccessControlLink.GLOBAL).
                    indexOf(" ACLs from different ACS: ") == -1) {
                        // apply the _withExecute condition for one ACS
                        if (!baseUtil_.withExecuteRight(
                            (StringRight[])acLink.getRedondantRights().toArray(new StringRight[0]),
                            _pt instanceof Node) ) {

                                        // for testing and tracking
                                        //System.out.println("in PrivilegeRightsFactoryImpl."+
                                        //"addPrivilegeRightsInSimpleDisplayableLink(), ");
                                        //System.out.println("-- the executing condition is not fulfilled,"+
                                          //"with ep: "+_ps.getName()+", res: "+_pt.getName());
                                        //if (acLink.getRedondantRights().isEmpty())
                                          //  System.out.println("-- for no rights");
                                        //else {
                                          //  System.out.println("-- for the rights:");
                                            //for (Object ob : acLink.getRedondantRights()) {
                                              //  System.out.println("-----"+ob);
                                            //}
                                        //} // else


                                result = null;
                        } // if (!BaseUtilityImpl
                        else {
                                result = new DisplayableLinkImpl(_ps, _pt, acLink);
                        } // else if (!BaseUtilityImpl
            } // if (acLink.getLinkTypeComments(
            else {
                // there are rights from several ACS in acLink.
                result = new DisplayableLinkImpl(_ps, _pt, acLink);
            } // else if (acLink.getLinkTypeComments(
        } // if (_withExecute
        else
            result = new DisplayableLinkImpl(_ps, _pt, acLink);

        acLink = null;
        return result;
  } // addPrivilegeRightsInSimpleDisplayableLink()

  /**
  * Adds the privilege rights to an AccessControlLinkImpl. This will allow
  * AccessControlLinkImpl.getRight() to filter them, if there are some negative rights
  * from some ACSs. Reads the positive and negative inherited rights
  * if the target ACS manages the LPRI rights inheritance. In this case,
  * returns the direct rights if there is no inherited rights through the method
  * LinkRights.getL_accessRightsThroughNodesTree(), while
  * the non-null inherited rights may delete all rights or be addded to the direct ones
  * by the method NodeRights.mergeInheritedAclPriRightsAndComments(). <P>
  * If the LPRI inheritance is managed and if NodeRights.withAccessThroughNodesTreeFromEP()
  * returns true, the result is null if the inherited rights array is null.
  * On the other hand, the inherited rights array may be empty, and then the result contains
  * simply the direct rights. When direct rights and inherited rights are not empty,
  * they are merged by NodeRights.mergeInheritedAclPriRightsAndComments().
  * This method allows also to add specific AcsAddon comments for the direct privileges. <P>
  * Puts in comments the rights, and puts also in the GLOBAL comment, the Privilege names
  * and all  the Privilege ACS names if they are more than one. <P>
  * Note: this method is sensible since it is used for both the 2-nodes methods,
  * the 3-nodes methods and a multiple-links method.
  * Called by detectPrivilegeRights(), addPrivilegeRightsInSimpleDisplayableLink(),
  * and CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
  * Calls getL_privilegeRights() and getL_linkedPrivilege(),
  * and the _pt NodeRights.detectL_linkedPrivilegeRights(), mergeInheritedAclPriRightsAndComments(),
  * getL_accessRightsThroughNodesTree() and withAccessThroughNodesTreeFromEP().
  * @param _pt to analyze. Has to be known in the AlgorithmInterpreter.
  * @param _ps to analyze
  * @param _upd_acLink is the AccessControlLink to update. May be null.
  * @return null if no right has been added, otherwise returns the updated _upd_acLink,
  *  or a new AccessControlLinkImpl if _upd_acLink is null.
  **/
  public final AccessControlLinkImpl addPrivilegeInAccessControlLink(
          ImmutablePrivilegeTarget _pt, ImmutablePrivilegeSource _ps,
          AccessControlLinkImpl _upd_acLink) {

    StringRight[]     l_inheritedRights = null;
    AccessControlLinkImpl inhLinkRights = null;
    NodeRights        nodeRights = interpreter_.getNodeRightsInstances().get(_pt.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(_pt);
        nodeRights = interpreter_.getNodeRightsInstances().get(_pt.getName());
    } // if

    if (_pt.getEorACS().getManageLPRIRightsInheritance()) {
      nodeRights = interpreter_.getNodeRightsInstances().get(_pt.getName());

      // this test is redondant, but it avoids an unuseful call
      if (_pt.getEorACS() == _ps.getEorACS() &&
              nodeRights.withAccessThroughNodesTreeFromEP(_ps, _pt, null)) {
          if (!(_ps instanceof ImmutableGroupID))
              inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(
                                _pt, (ImmutableEligibleParty)_ps, null);
          else
              inhLinkRights = nodeRights.getL_accessRightsThroughNodesTree(
                                _pt, null, (ImmutableGroupID)_ps);
          if (inhLinkRights != null)
                  l_inheritedRights = inhLinkRights.getRights(AccessControlLink.C_INHERITED_PRI);
      } // if

      if (_pt.getEorACS() == _ps.getEorACS() &&
              (l_inheritedRights == null || l_inheritedRights.length == 0)) {
          l_inheritedRights = null;
          inhLinkRights = null;
          nodeRights = null;
          return null;
      } // if
    } // if (_pt.getEorACS().getManageLPRIRightsInheritance())

    // l_priv is used only for the comments
    List<PrivilegeForLinks>  l_priv = getL_linkedPrivilege(_ps, _pt);
    List<PrivilegeForType>   l_priv2 = getL_typedPrivilege(_ps, _pt);

    // both positive and negative rights only if the linked privileges are from different ACS
    List<StringRight>        l_lpriRights = nodeRights.detectL_linkedPrivilegeRights(_pt, _ps);

    if (l_lpriRights == null && l_priv2 == null) {
      if (!_pt.getEorACS().getManageLPRIRightsInheritance()) {
          l_inheritedRights = null;
          inhLinkRights = null;
          nodeRights = null;
          l_priv = null;
          l_priv2 = null;
          return null;
      }
      else if (l_inheritedRights == null || l_inheritedRights.length == 0) {
          l_inheritedRights = null;
          inhLinkRights = null;
          nodeRights = null;
          l_priv = null;
          l_priv2 = null;
          return null;
      }
    } // if (l_lpriRights == null && l_priv2 == null)

    // variables
    StringRight[]         l_rights = null;
    StringRight           right = null;
    AccessControlLinkImpl result = null;
    boolean               severalACS = false;
    char                  open = '[';
    char                  close = ']';
    char                  space = ' ';
    Object                obj = null;
    StringBuilder         comment = null;
    Privilege             priv = null;
    Privilege             priv2 = null;

    if (l_lpriRights != null) {
      for (StringRight lright : l_lpriRights) {

            if (lright == null || lright.equals(StringRightImpl.EMPTY_INSTANCE))
                continue;
            
            if (result == null) {
                if (_upd_acLink != null) result = _upd_acLink;
                else result = new AccessControlLinkImpl();
            } // if

            if (lright.isGrantingRight()) {
              result.addRight(AccessControlLink.LPRI, lright );
              result.addComment(AccessControlLink.LPRI,
                      lright.getRight().replace(open, space).replace(close, space) );
            } else {
              result.addRight(AccessControlLink.NEG_LPRI, lright);
              result.addComment(AccessControlLink.NEG_LPRI,
                      lright.getRight().replace(open, space).replace(close, space) );
            } // else

      } // for lright
    } // if (l_lpriRights != null)

    if (l_priv2 != null) {
      for (Iterator ite = l_priv2.iterator(); ite.hasNext(); ) {
            priv = (Privilege)ite.next();

                            //for testing and tracking
                            //System.out.println(
                                    //"PrivilegeRightsFactoryImpl.addPrivilegeInAccessControlLink()"+
                                    //"from "+_ps.getNickName()+" to "+_pt.getNickName()+
                                    //" has found "+priv);

            if (result == null) {
                if (_upd_acLink != null) result = _upd_acLink;
                else result = new AccessControlLinkImpl();
            } // if

            l_rights = priv.getL_Rights();

            for (int i = 0; i < l_rights.length; i++) {
                right = l_rights[i];
                if (right == null || right.equals(StringRightImpl.EMPTY_INSTANCE)) continue;

                if (right.isGrantingRight()) {
                  result.addRight(AccessControlLink.TPRI, right );
                  // put in comments the right
                  result.addComment(AccessControlLink.TPRI,
                          right.getRight().replace(open, space).replace(close, space) );
                } else {
                  result.addRight(AccessControlLink.NEG_TPRI, right);
                  result.addComment(AccessControlLink.NEG_TPRI,
                          right.getRight().replace(open, space).replace(close, space) );
                } // else
            } // for i

      } // for ite
    } // if (l_priv2 != null)
    
    if (l_lpriRights == null && l_priv2 == null) {
        // there are inherited rights
        if (_upd_acLink != null) result = _upd_acLink;
        else result = new AccessControlLinkImpl();
    } // else

    // merge the inherited rights and the direct rights.
    // add some proper comments and some ones from NodeRights.getL_accessRightsThroughNodesTree()
    result = nodeRights.getMergedInheritedAclPriRightsAndComments(
          _ps, _pt, result, l_inheritedRights);
    if (l_inheritedRights != null && result != null) {
          result.mergeCommentsWithoutFiltering(inhLinkRights);
    } // if (l_inheritedRights != null)

    // 'result' is null if the inherited rights merging deletes all rights.
    if (result != null) {

        if (l_priv2 != null) {
            // puts in comments the direct typed privileges names.
            // The typed privileges are always from the source and the target ACS.
            result.setLinkTypeToTrue(AccessControlLink.GLOBAL);
            comment = new StringBuilder("PRI: ");

            for (Iterator ite = l_priv2.iterator(); ite.hasNext(); ) {
                priv = (Privilege)ite.next();
                if (comment.indexOf(priv.getNickName()) == -1)
                        comment.append(priv.getNickName());

            } // for ite

            result.addComment(AccessControlLink.GLOBAL, comment.toString());

        } // if (l_priv2 != null)

        if (l_priv != null) {
            // puts in comments the direct linked privileges ACS names,
            // if there is more than one ACS, and puts also the privilege names.
            result.setLinkTypeToTrue(AccessControlLink.GLOBAL);
            if (comment == null)
                comment = new StringBuilder("PRI: ");

            for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                priv = (Privilege)ite.next();
                if (comment.indexOf(priv.getNickName()) == -1)
                        comment.append(priv.getNickName());

            } // for ite

            result.addComment(AccessControlLink.GLOBAL, comment.toString());

            for (Iterator ite = l_priv.iterator(); ite.hasNext(); ) {
                obj = ite.next();
                if (priv == null) { // first iteration
                    priv = (Privilege)obj;
                    continue;
                }
                else {
                    if (priv2 != null) priv = priv2; // third iteration and over
                    priv2 = (Privilege)obj;
                }

                // detects the first case where there are 2 ACS
                if (priv2.getEorACS() != priv.getEorACS()) {
                    severalACS = true;
                    comment = new StringBuilder(" - PRIs from different ACS: ");
                    if (comment.indexOf(priv.getEorACS().getNickName()) == -1) {
                        comment.append(priv.getEorACS().getNickName());
                    } // if (comment
                } // if

                if (severalACS && comment.indexOf(priv2.getEorACS().getNickName()) == -1) {
                    comment.append(priv2.getEorACS().getNickName());
                    if (ite.hasNext()) comment.append(", ");
                } // if (severalACS
            } // for i

            if (severalACS)
                result.addComment(AccessControlLink.GLOBAL, comment.toString());

        } // if (l_priv != null)
    
    } // if (result != null)

    l_priv = null;
    l_priv2 = null;
    l_lpriRights = null;
    obj = null;
    priv = null;
    priv2 = null;
    comment = null;
    l_rights = null;
    right = null;
    l_inheritedRights = null;
    nodeRights = null;

                        //for testing and tracking
                        //System.out.println(
                                //"PrivilegeRightsFactoryImpl.addPrivilegeInAccessControlLink()"+
                                //"from "+_ps.getNickName()+" to "+_pt.getNickName()+
                                //", result: "+result);

    return result;
  } // addPrivilegeInAccessControlLink()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
    interpreter_ = null;
    linkUtil_ = null;
    baseUtil_ = null;
    factoryUtil_ = null;
    l_targets_.clear();
    l_targets_ = null;
  } // finalizeForProcess()

} // PrivilegeRightsFactoryImpl

