//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import javax.swing.JTextPane;

import ARoad0.CNot.*;
import ARoad0.ErrB.*;
import ARoad0.ErrW.*;
import ARoad0.Gui2.GraphicView;
import ARoad0.Gui2.GraphicEPRView;
import ARoad0.Gui2.GraphicNoThanView;

import ARoad0.gBaseInterface.*;
import ARoad0.gWorkInterface.*;

import ARoad0.gBase.UtilityImpl;
import ARoad0.gBase.EPRViewInBaseImpl;
import ARoad0.gBase.NameImpl;
import ARoad0.gBase.StringRightImpl;

/**
 * This class is the "facade" (GoF design pattern) for all the core algorithms
 * about the sequence of method calls in the access paths search.
 * This work may be summarized as building up new CNot.DisplayableLinkImpls from the old ones,
 * as the result of the analysis of rights and structures. The main parameter
 * to handle is the map of all the current DisplayableLinkImpls, usually named
 * _m_l_DisplayableLinks. At each method call, this map is filled with new DisplayableLinkImpls,
 * and at the end, Gui2.GraphicView reads the map to display the view diagram and
 * the view 'See why' text.<P>
 * This class is stateless. For each ViewInBase, RightsMediatorImpl at its creation
 * setups an instance of this class for its ViewInBase, to be processed in a worker thread.
 * The architecture view and the sketch view are the two exceptions.
 * They are not managed by RightsMediatorImpl. They use specific static methods
 * and they are processed in the event dispatch thread or the main thread.
 * No direct call to the ACS methods.
 * Called by Gui1, Gui2 and gDMak objects. For the EPR and the NoThan ViewInBases,
 * only Gui2.GraphicEPRView and Gui2.GraphiNoThanView are the consumers.
 * @version 0.7.0
 */
public class RightsFactory_Facade implements CoreAlgorithm {
  private static final int          INITIAL_CAPACITY = 15; // default value for map: 101

  // for the getEffectiveRightsForPair() method, for the EPR ViewInBases
  private Set<StringRight>          l_result_ = new HashSet(INITIAL_CAPACITY);
  private Set<StringRight>          l_noActorResult_ = new HashSet(INITIAL_CAPACITY); 
  private Set<String>               l_lastLinkStringResult_ = new HashSet(INITIAL_CAPACITY);

  // for the EPR and NoThan ViewInBases
  private AlgorithmInterpreter      interpreter_ = null;
  private FactoryUtilities          factoryUtil_ = null;
  private DisplayableLinkUtilities  linkUtil_ = null;
  private UtilityImpl               utility_ = null;

  /**
  * only one constructor
  */
  RightsFactory_Facade(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      factoryUtil_ = (FactoryUtilities)interpreter_.getCoreAlgorithm("FactoryUtil");
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
      utility_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
  }

  /**
  * Detects the access paths between BaseObjects in an EPRViewInBase.
  * Returns the DisplayableLinkImpls for every pair of BaseObjects in a view.
  * Whatever the boolean arguments, calls CompoundRightsFactoryImpl.detectBridges()
  * and ThreeNodesRightsFactoryImpl.detectOneHiddenNodeCompoundRights().
  * The calling of other methods in the gWork classes depends on the boolean arguments.
  * The number of called methods may be up to 18.
  * Calls Thread.interrupted() to test the working thread and stops it
  * if necessary. Called by GraphicEPRView.initialize(), that synchronizes on the EPRViewInBase,
  * and called by GraphicEPRView.resetRights(). Called by detectNoThanRights().
  * @param _gview graphic view associated to the view
  * @param _viewInBase the view to analyze
  * @param _detectACL true for detecting the pure AclEntry access paths
  * @param _detectPRI true for detecting the pure Privilege access paths
  * @param _detectAGO true for detecting the AGO access paths
  * @param _detectLargeLinks true for detecting the access paths having
  *     more than 3 nodes
  * @exception ProcessError if there is a loop in the rights search
  * @return Map of DisplayableLinkImpls, structured as m_l_DisplayableLinks_,
  * with one or several access control links associated to each pair (Source, Target).
  * May be empty.
  * */
  public final Map detectEPRRights(GraphicEPRView _gview,
    EPRViewInBase _viewInBase,
    boolean _detectACL, boolean _detectAGO, boolean _detectPRI,
    boolean _detectLargeLinks) throws ProcessError, InterruptedException {

                    // for testing and tracking
                    //System.out.println("call to RightsFactory_Facade.detectEPRRights().");

      // variables
      Map                   result = new HashMap(INITIAL_CAPACITY);
      ImmutableTarget       targ = null;
      NodeRights            nodeRights = null;
      AgoRightsFactoryImpl  ago = (AgoRightsFactoryImpl)interpreter_.getCoreAlgorithm("Ago");
      CompoundRightsFactoryImpl compound = (CompoundRightsFactoryImpl)interpreter_.
                                                getCoreAlgorithm("Compound");
      boolean               detectAcsAddonLinks = false;
      boolean               detectAddon2Nodes = false;
      boolean               detectAddon3Nodes = false;
      boolean               detectAddonFinal2Nodes = false;
      boolean               detectAddonFinal3Nodes = false;
      boolean               detectAddonFinalAllNodes = false;
      // for testing
      ImmutableName key = null;
      DisplayableLinkImpl dLink = null;

      // intialize detectAcsAddonLinks and the other booleans
      // for all the nodes of the view
      for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
          targ = (ImmutableTarget) ita.next();

          // detect the AcsAddon
          if (targ.getEorACS().getClass().toString().indexOf("ARoad0.AcsAddon.") != -1
                  //targ.getEorACS().getIsFromAcsAddon()
                  ) {
              if (!detectAcsAddonLinks)
                      detectAcsAddonLinks = true;
          } // if
          else {
              continue;
          } // else

          nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());

          if (!detectAddon2Nodes && nodeRights.getDetectAddon2Nodes())
              detectAddon2Nodes = true;
          if (!detectAddon3Nodes && nodeRights.getDetectAddon3Nodes())
              detectAddon3Nodes = true;
          if (!detectAddonFinal2Nodes && nodeRights.getDetectAddonFinal2Nodes())
              detectAddonFinal2Nodes = true;
          if (!detectAddonFinal3Nodes && nodeRights.getDetectAddonFinal3Nodes())
              detectAddonFinal3Nodes = true;
          if (!detectAddonFinalAllNodes && nodeRights.getDetectAddonFinalAllNodes())
              detectAddonFinalAllNodes = true;
      } // for ita

      // ---------------- 2-nodes links analyzing ---------------------

      result = compound.detectBridges(_viewInBase, result, null);

      if (_detectACL) {

                    // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls after bridges and before detectACLRights() are:");
                    //ImmutableName key = null;
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+ 
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                    //  "call to detectACLRights()");

        result = ((AclRightsFactoryImpl)interpreter_.getCoreAlgorithm("Acl"))
                .detectACLRights(_viewInBase, result, null);
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
                    // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls after detectACLRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                         //   System.out.println("------------"+
                           //   ((DisplayableLinkImpl)ita.next()).toString() );
                       //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                    //  "call to detectHiddenACLRights()");

      } // if (_detectACL)

                    // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls before detectRootRights() are:");
                    //ImmutableName key;
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //ImmutableName key;            
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls after detectRootRights():");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
        
                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectRights(), "+
                      //"call to detectAGRights()");

      if (_detectAGO) {
          // throw InterruptedException
        result = ago.detectAGRights(_viewInBase, result);
        
                    // for testing and tracking
                    //ImmutableName key;
                    //System.out.println(" ");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls after detectAGRights() and before "+
                      //"detectOneHiddenNodeWithCommonAGORights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"call to detectHiddenDirectOwnerRights()");

                    // for testing and tracking
                    //ImmutableName key;
                    //System.out.println(" ");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls after detectOneHiddenNodeWithCommonAGORights() \n "+
                      //"and before detectHiddenChainedGroupsRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("\n In RightsFactory_Facade.detectEPRRights(), "+
                      //"call to detectHiddenChainedGroupsRights()");

                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls after detectHiddenChainedGroupsRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

        if (detectAcsAddonLinks && detectAddon2Nodes) {
            for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
              targ = (ImmutableTarget) ita.next();
              nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());
              if (nodeRights.getDetectAddon2Nodes())
                  result = nodeRights.detectAddon2Nodes(_viewInBase, targ, result, null);
            } // for ita
        }
        
        if (Thread.interrupted()) {
            throw new InterruptedException();
        }
        

                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls after detectOtherRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"call to detectRootRights()");

        // throws InterruptedException
        result = ago.detectRootRights(_viewInBase, result, null);


                    // for testing and tracking
                    //ImmutableName key = null;
                    //System.out.println(" ");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"the DisplayableLinkImpls before detectOneHiddenNodeCompoundRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
        
      } // if (_detectAGO) {

      // throws InterruptedException
      if (_detectPRI)
          result = ((PrivilegeRightsFactoryImpl)interpreter_.getCoreAlgorithm("Pri")).
              detectPrivilegeRights(_gview, _viewInBase, result, null);

      
                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"after the call to detectPrivilegeRights()");
                    //ImmutableName key;
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

        // in all cases, detect the virtual member and the alias direct access paths
      result = compound.detectVirtualMembers(_viewInBase, result, null);
      result = compound.detectReferencesForAliases(_viewInBase, result, null);
      
                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"after the call to detectVirtualMembers() and detectReferencesForAlias()");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      if (detectAcsAddonLinks && detectAddonFinal2Nodes) {
          for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
              targ = (ImmutableTarget) ita.next();

              nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());
              if (nodeRights.getDetectAddonFinal2Nodes())
                  result = nodeRights.detectAddonFinal2Nodes(_viewInBase, targ, result, null);
          } // for ita
      }

      // ---------------- 3-nodes links analyzing ---------------------
      
                    // for testing and tracking
                    //ImmutableName key;
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"before the 3-nodes links analyzing");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    
      if (_detectAGO) {
            // not called in detectSketchRights - throw InterruptedException
          result = ago.detectOneHiddenNodeWithCommonAGORights(_gview, _viewInBase, result);
      }

        // for Ubuntu, detect the 3 nodes links with the capacity groups
      if (detectAcsAddonLinks && detectAddon3Nodes) {
          for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
              targ = (ImmutableTarget) ita.next();

              nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());
              if (nodeRights.getDetectAddon3Nodes())
                  result = nodeRights.detectAddon3Nodes(_viewInBase, targ, result);
          } // for ita
      }
      
                    // for testing and tracking
                    //ImmutableName key;
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                      //"before ThreeNodesRightsFactoryImpl.detectOneHiddenNodeCompoundRights()");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

        // detect all the 3-nodes access paths and applies priority rules on all links
        // throws InterruptedException
      result = ((ThreeNodesRightsFactoryImpl)interpreter_.getCoreAlgorithm("3nodes")).
              detectOneHiddenNodeCompoundRights(_gview, _viewInBase, result);

        // 
      if (detectAcsAddonLinks && detectAddonFinal3Nodes) {
          for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
              targ = (ImmutableTarget) ita.next();

              nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());
              if (nodeRights.getDetectAddonFinal3Nodes())
                  result = nodeRights.detectAddonFinal3Nodes(_viewInBase, targ, result);
          } // for ita
      } // if (detectAcsAddonLinks)

                            // for testing and tracking
                            //System.out.println("");
                            //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                              //"call to detectHiddenCompoundEpRights()");
                            //ImmutableName key;
                            //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                              //"the DisplayableLinkImpls after detectOneHiddenNodeCompoundRights()"+
                              //" and nodeRights.detectAddonFinal3Nodes() are:");
                            //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                              // key = (ImmutableName) ite.next();
                               //System.out.println("--- for key: "+key);
                               //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                                //System.out.println("------------"+
                                  //((DisplayableLinkImpl)ita.next()).toString() );
                              //} // for ita
                            //} // for ite


      // ---------------- more-than-3-nodes links analyzing ---------------------
      
      if (_detectLargeLinks) {
        if (_detectAGO) {
            // not called in detectSketchRights - throws an InterruptedException
            result = ago.detectHiddenChainedGroupsRights(_gview, _viewInBase, result);
        }

                            // for testing and tracking
                            //System.out.println("");
                            //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                              //"before the call to detectHiddenCompoundEpRights()");
                            //ImmutableName key;
                            //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                              //"the DisplayableLinkImpls after detectHiddenChainedGroupsRights() are:");
                            //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                              // key = (ImmutableName) ite.next();
                               //System.out.println("--- for key: "+key);
                               //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                                //System.out.println("------------"+
                                  //((DisplayableLinkImpl)ita.next()).toString() );
                              //} // for ita
                            //} // for ite

          // the most complex method of this program - throws InterruptedException
        result = compound.detectHiddenCompoundEpRights(_gview, _viewInBase, result);

                            // for testing and tracking
                            //System.out.println("");
                            //System.out.println("RightsFactory_Facade.detectEPRRights(), "+
                              //"after the call to detectHiddenCompoundEpRights()");
                            //ImmutableName key;
                            //System.out.println("In RightsFactory_Facade.detectEPRRights(), "+
                              //"the DisplayableLinkImpls before addFromNegativeAclRightsInOneACS() are:");
                            //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                              // key = (ImmutableName) ite.next();
                               //System.out.println("--- for key: "+key);
                               //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                                //System.out.println("------------"+
                                  //((DisplayableLinkImpl)ita.next()).toString() );
                              //} // for ita
                            //} // for ite

          // detection of the non-simple negative paths in the same ACS
          // and adding of a simple path of the resulting rights.
          // throws InterruptedException
        if (_detectACL || _detectPRI)
            result = compound.addPathFromNegativeRightsInOneACS(_gview, _viewInBase, result);
        
      } // if (_detectLargeLinks)

      result = compound.detectLastSimpleSoftRunUnderDisplayableLinks(_gview, _viewInBase, result);
      
      if (detectAcsAddonLinks && detectAddonFinalAllNodes) {
          // caution: m_effectiveRights remains immutable through the calls
          // to NodeRights.detectAddonFinalAllNodes()
          Map<ImmutableName, Set<StringRight>>
                  m_effectiveRights = getFirstShapeEffectiveRights(result);

          for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
              targ = (ImmutableTarget) ita.next();

              nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());
              if (nodeRights.getDetectAddonFinalAllNodes())
                  result = nodeRights.detectAddonFinalAllNodes(
                          _viewInBase, targ, result, m_effectiveRights);
          } // for ita
      } // if

                            // for testing and tracking
                            //ImmutableName key;
                            //DisplayableLinkImpl dLink;
                            //int count = 0;
                            //System.out.println(" ");
                            //System.out.println("RightsFactory_Facade.detectEPRRights(), "+
                              //"the DisplayableLinkImpls in the final result are:");
                            //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                              // key = (ImmutableName) ite.next();
                               //count = 0;
                               //System.out.println("--- for key: "+key);
                               //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                                 //  dLink = (DisplayableLinkImpl)ita.next();
                                   //System.out.println("--------- at "+count+": "+
                                     //   dLink.toString() );
                                   //System.out.println("-------------------- last link:"+
                                     //   dLink.getEorLastLink().toString() );
                                   //count++;

                                   //if (!ita.hasNext()) {
                                     //  System.out.println("-------- properties of the last link:");
                                       //for (Map.Entry<String, Object> ent : dLink.getProperties().entrySet()) {
                                         //  System.out.println(
                                           //        "------------ key: "+ent.getKey()+
                                             //      "; value: "+ent.getValue());
                                       //} // for ent
                                   //} // if (!ita
                               //} // for ita
                            //} // for ite
                            //System.out.println(" ");

    targ = null;
    nodeRights = null;
    ago = null;
    compound = null;
    return result;
  } // detectEPRRights()

  /**
  * Defines the BRIDGE and ACL relations between all the ACS which are in the view,
  * that are Bridge links between actors, or AclEntries between EligibleParties
  * and resources. Calls CompoundRightsFactoryImpl.detectAcsLinks().
  * Called by GraphicView.initialize(), that synchronizes on _viewInBase,
  * and resetRights(), for the architecture view only.
  * @param _gview is the graphic view associated to the view
  * @param _viewInBase is the view to analyze
  * @exception ProcessError if 
  * @return Map of DisplayableLinkImpls, structured as m_l_DisplayableLinks_,
  *    with one or several access control links associated to each node pair.
  *    May be empty.
  * */
  public static synchronized final Map detectRights(GraphicView _gview, ViewInBase _viewInBase)
          throws ProcessError, InterruptedException {
    // variables
    Map                   result = new HashMap(INITIAL_CAPACITY);

    result = CompoundRightsFactoryImpl.detectAcsLinks(_gview, _viewInBase, result);
    if (Thread.interrupted()) {
        throw new InterruptedException();
    }

                    // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectRights(), "+
                      //"the DisplayableLinkImpls in the final result are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite
                    //System.out.println(" ");

    return result;
  } // detectRights()  

  /**
  * Detects access-control links between BaseObjects in _viewModel,
  * for a sketch view on a central object rights.
  * Detects displayable links for every pair (BaseObject, _center).
  * Does not detect running under currentUID/GID for an Actor.
  * Does not throw any InterruptedException since this method is not called
  * in a worker thread.
  * Uses ViewInBase_Facade.SKETCH_MEDIATOR.
  * Called by ViewInBase_Facade.populateSketchView() in the main thread, for the
  * sketch view only.
  * @param _viewInBase is the view to analyze
  * @param _center is the central object of a sketch view
  * @param _detectAcl true for detecting aclEntry links.
  * @param _detectUserGroupLinks true for detecting
  * userID and groupID ACS-specific access-control links.
  * @exception ProcessError if there is a loop in the rights search
  * @return Map of DisplayableLinkImpls, structured as m_l_DisplayableLinks_,
  * with one or several AccessControlLinks associated to each pair (BaseObject, _center).
  * May be empty.
  * */
  public static synchronized final Map detectSketchRights(EPRViewInBase _viewInBase,
    BaseObject _center, boolean _detectAcl, boolean _detectUserGroupLinks) 
    throws ProcessError, InterruptedException {
      // variables
      Map result = new HashMap(INITIAL_CAPACITY);
      ImmutableTarget   targ = null;
      NodeRights        nodeRights = null;
      AgoRightsFactoryImpl ago = ((AgoRightsFactoryImpl)ViewInBase_Facade.SKETCH_MEDIATOR.
                                    getCoreAlgorithm("Ago"));
      CompoundRightsFactoryImpl  compound = (CompoundRightsFactoryImpl)
                                    ViewInBase_Facade.SKETCH_MEDIATOR.
                                    getCoreAlgorithm("Compound");

      result = compound.detectBridges(_viewInBase, result, _center);
        
      if (_detectAcl) {

                    // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"the DisplayableLinkImpls before detectACLRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"call to detectACLRights()");

        result = ((AclRightsFactoryImpl)ViewInBase_Facade.SKETCH_MEDIATOR.
                getCoreAlgorithm("Acl")).
                detectACLRights(_viewInBase, result, _center);

      } // if (_detectAcl)

                    // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"the DisplayableLinkImpls before detectRootRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      if (_detectUserGroupLinks) {

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"call to detectRootRights()");

          // _center is never 'root' in its ACS
        result = ago.detectRootRights(_viewInBase, result, _center);

                     // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"the DisplayableLinkImpls before detectUserRights() are:");
                    //ImmutableName key;
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                         //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"call to detectUserRights()");

        result = ago.detectUserRights(_viewInBase, result, _center);

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"call to detectGroupRights()");

        result = ago.detectGroupRights(_viewInBase, result, _center);

                    // for testing and tracking
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                      //"the DisplayableLinkImpls after detectGroupRights() are:");
                    //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                        //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

                    // for testing and tracking
                    //System.out.println("");
                    //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                    //  "call to detectHiddenDirectOwnerRights()");

      } // if (_detectUserGroupLinks)

      for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
          targ = (ImmutableTarget) ita.next();
          nodeRights = ViewInBase_Facade.SKETCH_MEDIATOR.getNodeRightsInstances().get(
                  targ.getName());
          if (nodeRights.getDetectAddon2Nodes())
              result = nodeRights.detectAddon2Nodes(_viewInBase, targ, result, _center);
      } // for ita

      // in all cases, detect the direct privilege paths
      result = ((PrivilegeRightsFactoryImpl)ViewInBase_Facade.SKETCH_MEDIATOR.
                getCoreAlgorithm("Pri")).
                    detectPrivilegeRights(null, _viewInBase, result, _center);

      // in all cases, detect the virtual members and the hard alias direct access paths
      result = compound.detectVirtualMembers(_viewInBase, result, _center);
      result = compound.detectHardReferencesForAliases(_viewInBase, result, _center);
      result = compound.detectHardAliasesForReferences(_viewInBase, result, _center);

      for (Iterator ita = _viewInBase.getEorM_Targets().values().iterator() ; ita.hasNext() ; ) {
          targ = (ImmutableTarget) ita.next();

          nodeRights = ViewInBase_Facade.SKETCH_MEDIATOR.getNodeRightsInstances().get(
                  targ.getName());
          if (nodeRights.getDetectAddonFinal2Nodes())
              result = nodeRights.detectAddonFinal2Nodes(_viewInBase, targ, result, _center);
      } // for ita

                            // for testing and tracking
                            //ImmutableName key;
                            //System.out.println("In RightsFactory_Facade.detectSketchRights(), "+
                              //"for "+_center.getNickName()+
                              //", the DisplayableLinkImpls in the final result are:");
                            //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                              // key = (ImmutableName) ite.next();
                               //System.out.println("--- for key: "+key);
                               //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                                //System.out.println("------------"+
                                  //((DisplayableLinkImpl)ita.next()).toString() );
                              //} // for ita
                            //} // for ite
                            //System.out.println(" ");

      ago = null;
      compound = null;
      return result;
  } // detectSketchRights()    

  /**
  * Detects if the No-More-Than and No-Less-Than rights criteria are fulfilled or not.
  * If true, or if the view target is null, or if the two criteria are null,
  * return an empty map. Else, it returns simple DisplayableLinkImpls where
  * each one which demonstrates the break for one criterion.
  * These links includes for the type GLOBAL in their AccesControlLink,
  * the comment AccessControlLink.NO_MORE_COMPLIANCE or AccessControlLink.NO_LESS_COMPLIANCE.<P>
  * At most, there are two DisplayableLinks in the returned map,
  * where the second end is the view target, and the first end is one of view EPs,
  * or one of the members, or one of the running-under actors.<P>
  * This method calls all the methods used by detectEPRRights() for the paths search,
  * excepted for the method CompoundRightsFactoryImpl.detectLastSimpleSoftRunUnderDisplayableLinks().
  * For each elementary pair (EP of the checking perimeter, first target),
  * it searches the rights in creating a dedicated EPRViewInBase with a constant name.
  * If the resulting rights do not comply with the two criterion during the iteration
  * on all the tested EP, the general search is ended. The transient EPRViewInBases
  * are usable because the RightsMediator of the NoThanViewInBase manages
  * all the BaseObjects of the view checking perimeter. They works also because,
  * when some methods find access paths through hidden intermediate nodes which
  * are in the NoThanViewInBase, this simply accelerates the final result and
  * does not change it. <P>
  * This method is called from several worker threads.
  * Called by GraphicNoThanView.initialize(), that synchronizes on _viewInBase,
  * and called by resetRights(), but the result is indeed processed
  * by GraphicNoThanView.resetCriteria().
  * Calls NoThanViewInBase.getCheckingPerimeter(), getMoreCriterion(),
  * getLessCriterion() and all the methods called by detectEPRrights().
  * @param _gview the graphic view associated to the view
  * @param _viewInBase view to analyze
  * @exception ProcessError if there is a loop in the rights search
  * @return Map of DisplayableLinkImpls, structured as m_l_DisplayableLinks_ with lists as values,
  *     with one or several access control links associated to each pair (EP, Resource).
  *     Empty.
  * */
  public final Map detectNoThanRights(GraphicNoThanView _gview,
          NoThanViewInBase _viewInBase) throws ProcessError, InterruptedException {
        
      // variables
      Map                   result = new HashMap(INITIAL_CAPACITY);
      Map                   intermResult = new HashMap(INITIAL_CAPACITY);
      Set<ImmutableSource>  l_testedEPs = _viewInBase.getCheckingPerimeter();
      EPRViewInBaseImpl     view = null;
      Map                   m_baseObj = new HashMap(10);
      ImmutableTarget       target = _viewInBase.getEorFirstAccessTarget();
      Name                  constantName = new NameImpl(
                                "<transient EPRViewInBase for the NoThanViewInBase paths search>");
      StringRight           more = _viewInBase.getMoreCriterion();
      StringRight           less = _viewInBase.getLessCriterion();
      StringRight           unmatch = null;
      boolean               haveMoreResult = false;
      boolean               haveLessResult = false;
      AccessControlLinkImpl acLink = null;
      NodeRights            nodeRights = null;
      AgoRightsFactoryImpl  ago = (AgoRightsFactoryImpl)interpreter_.getCoreAlgorithm("Ago");
      CompoundRightsFactoryImpl  compound = (CompoundRightsFactoryImpl)interpreter_.
                                    getCoreAlgorithm("Compound");
      LinkRights            linkRights = null;
      ImmutableName         key = null;

                        // for testing and tracking
                        //System.out.println("");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //"call for the view "+_viewInBase.getName());


      if (target == null || (more == null && less == null)) {
          intermResult = null;
          l_testedEPs = null;
          m_baseObj = null;
          more = null;
          less = null;
          ago = null;
          compound = null;

          return result;
      }
      else {
          nodeRights = interpreter_.getNodeRightsInstances().get(target.getName());
          linkRights = interpreter_.getLinkRightsInstances().get(target.getName());
      }

      // for each elementary couple (tested EP, target), search of the rights
      // in creating a dedicated EPR view with a constant name,
      // and a dedicated DisplayableLinks map.
      // stop the search at the first match for each of the two criteria.
      for (ImmutableSource so : l_testedEPs) {
          // one view for each couple (ep, _viewInBase target)
          m_baseObj.clear();
          intermResult.clear();
          m_baseObj.put(so.getName(), so);
          m_baseObj.put(target.getName(), target);
          
          try {
              view = new EPRViewInBaseImpl(constantName, m_baseObj, false);
          } catch (CreateError b) {
              throw new InternalError("RightsFactory_Facade.detectNoThanRights(): "+b.toString());
          }

          // ---------------- 2-nodes links analyzing ---------------------

          intermResult = compound.detectBridges(view, intermResult, null);
          
                        // for testing and tracking
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //" the DisplayableLinkImpls before detectACLRights() are:");
                        //ImmutableName key = null;
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite

                        // for testing and tracking
                        //System.out.println("");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                        //  "call to detectACLRights()");

          intermResult = ((AclRightsFactoryImpl)interpreter_.getCoreAlgorithm("Acl")).
                            detectACLRights(view, intermResult, null);

          if (Thread.interrupted()) {
            throw new InterruptedException();
          }

                // no implicit ACS rules - throws InterruptedException
          intermResult = ago.detectAGRights(view, intermResult);

                        // for testing and tracking
                        //ImmutableName key;            
                        //System.out.println(" ");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //"the DisplayableLinkImpls after detectGroupRights() and before "+
                          //"detectOneHiddenNodeWithCommonAGORights() are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                           //key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite

                            // for testing and tracking
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //" the DisplayableLinkImpls before NodeRights.detectAddon2Nodes() are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                             //   System.out.println("------------"+
                               //   ((DisplayableLinkImpl)ita.next()).toString() );
                           //} // for ita
                        //} // for ite

                        // for testing and tracking
                        //System.out.println("");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                        //  "call to detectHiddenACLRights()");


          if (nodeRights.getDetectAddon2Nodes()) {

                            // for testing and tracking
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights() "+
                          //      "for "+so.getName()+", call to detectAddon2Nodes() with the nodeRights "+
                           //     nodeRights.getClass());

              intermResult = nodeRights.detectAddon2Nodes(view, target, intermResult, null);
          }
            
                        // for testing and tracking
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //" the DisplayableLinkImpls before detectAGRights() are:");
                        //ImmutableName key;
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite


                         // for testing and tracking
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                                //"for "+so.getName()+
                          //"the DisplayableLinkImpls before detectUserRights() are:");
                        //ImmutableName key;
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                          // System.out.println("--- for key: "+key);
                          // for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite

                        // for testing and tracking
                        //System.out.println("");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //"call to detectRootRights()");
             // throws InterruptedException
          intermResult = ago.detectRootRights(view, intermResult, null);

          // throws InterruptedException
          intermResult = ((PrivilegeRightsFactoryImpl)interpreter_.getCoreAlgorithm("Pri")).
                            detectPrivilegeRights(_gview, view, intermResult, null);

          intermResult = compound.
                  detectVirtualMembers(view, intermResult, null);
          intermResult = compound.
                  detectReferencesForAliases(view, intermResult, null);

                        // for testing and tracking
                        //ImmutableName key;
                        //System.out.println(" ");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //"the DisplayableLinkImpls before detectAddonFinal2Nodes() are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite

          if (nodeRights.getDetectAddonFinal2Nodes())
                  intermResult = nodeRights.detectAddonFinal2Nodes(view, target, intermResult, null);


            // ---------------- 3-nodes links analyzing ---------------------
            
                        // for testing and tracking
                        //ImmutableName key;
                        //System.out.println(" ");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //"the DisplayableLinkImpls before detectOneHiddenNodeWithCommonAGORights() are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite

                        // for testing and tracking
                        //System.out.println("");
                        //System.out.println("\n In RightsFactory_Facade.detectNoThanRights(), "+
                          //"call to detectHiddenChainedGroupsRights()");


          if (Thread.interrupted()) {
              throw new InterruptedException();
          }

            // not called in detectSketchRights - throws InterruptedException
          intermResult = ago.
              detectOneHiddenNodeWithCommonAGORights(_gview, view, intermResult);

                        // for testing and tracking
                        //ImmutableName key = null;
                        //System.out.println(" ");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //"the DisplayableLinkImpls before nodeRights.detectAddon3Nodes() are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite

            // for Ubuntu, detect the 3 nodes links
          if (nodeRights.getDetectAddon3Nodes())
                  intermResult = nodeRights.detectAddon3Nodes(view, target, intermResult);

                        // for testing and tracking
                        //ImmutableName key;
                        //System.out.println(" ");
                        //System.out.println("RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+", the DisplayableLinkImpls before "+
                          //"ThreeNodesRightsFactoryImpl.detectOneHiddenNodeCompoundRights() are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite
                        //System.out.println(" ");

                // detect 3-nodes access paths - throws InterruptedException
          intermResult = ((ThreeNodesRightsFactoryImpl)interpreter_.getCoreAlgorithm("3nodes")).
                      detectOneHiddenNodeCompoundRights(_gview, view, intermResult);

            // detect the 3 nodes links with the capacity groups
          if (nodeRights.getDetectAddonFinal3Nodes())
                  intermResult = nodeRights.detectAddonFinal3Nodes(view, target, intermResult);


            // ---------------- more-than-3-nodes links analyzing ---------------------

            // not called in detectSketchRights - throws InterruptedException
          intermResult = ago.detectHiddenChainedGroupsRights(_gview, view, intermResult);

            /*
            if (more != null && !haveMoreResult && !fulfilNoThanRightCriteria(intermResult, more, true)) {
              acLink = new AccessControlLinkImpl();
              acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
              acLink.setLinkTypeComments(
                       AccessControlLink.GLOBAL, AccessControlLink.NO_MORE_COMPLIANCE);
              linkUtil_.addDisplayableLink(
                      new DisplayableLinkImpl(ep, target, acLink ), result);
              haveMoreResult = true;
            } // if
            if (less != null && !haveLessResult && !fulfilNoThanRightCriteria(intermResult, less, false)) {
              acLink = new AccessControlLinkImpl();
              acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
              acLink.setLinkTypeComments(
                       AccessControlLink.GLOBAL, AccessControlLink.NO_LESS_COMPLIANCE);
              linkUtil_.addDisplayableLink(
                      new DisplayableLinkImpl(ep, target, acLink ), result);
              haveLessResult = true;
            } // if
               * */

                        // for testing and tracking
                        //System.out.println("");
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //"call to detectHiddenCompoundEpRights()");
                        //ImmutableName key;
                        //System.out.println("In RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //"the DisplayableLinkImpls before detectHiddenCompoundEpRights() are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                           //key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite

              // the most complex method of this program - throws InterruptedException
          intermResult = compound.detectHiddenCompoundEpRights(_gview, view, intermResult);

              // detection of the non-simple negative paths in the same ACS
              // and adding of a simple path of the resulting rights
          //  throws InterruptedException
          intermResult = compound.addPathFromNegativeRightsInOneACS(_gview, view, intermResult);

            // detect the more-than-3 nodes links
          if (nodeRights.getDetectAddonFinalAllNodes())
                  intermResult = nodeRights.detectAddonFinalAllNodes(
                          view, target, intermResult, getFirstShapeEffectiveRights(intermResult));

            // remove all the links from the target
          for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
               key = (ImmutableName) ite.next();
               if (key.subname(0, target.getName().noOfComponents()).equals(target.getName()))
                   ite.remove();
          } // for ite

                        // for testing and tracking;
                        //System.out.println(" ");
                        //System.out.println("RightsFactory_Facade.detectNoThanRights(), "+
                          //      "for "+so.getName()+
                          //"the DisplayableLinkImpls in the final result are:");
                        //for (Iterator ite = intermResult.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)intermResult.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite
                        //System.out.println(" ");
 
            if (more != null && !haveMoreResult) {
              unmatch = fulfilNoThanRightCriteria(intermResult, more, true);

              if (unmatch != null) {
                  acLink = new AccessControlLinkImpl();
                  acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                  acLink.setLinkTypeComments(
                           AccessControlLink.GLOBAL, AccessControlLink.NO_MORE_COMPLIANCE);
                  acLink.addComment(AccessControlLink.GLOBAL, "and it is due to the right '"+
                          unmatch+"'");
                  linkRights.addDisplayableLink(
                          new DisplayableLinkImpl(so, target, acLink ), result);
                  haveMoreResult = true;
              } // if (unmatch
            } // if (more

            if (less != null && !haveLessResult) {
              unmatch = fulfilNoThanRightCriteria(intermResult, less, false);

              if (unmatch != null) {
                  acLink = new AccessControlLinkImpl();
                  acLink.setLinkTypeToTrue(AccessControlLink.GLOBAL);
                  acLink.setLinkTypeComments(
                           AccessControlLink.GLOBAL, AccessControlLink.NO_LESS_COMPLIANCE);
                  acLink.addComment(AccessControlLink.GLOBAL, "and one of the right is '"+
                          unmatch+"'");
                  linkRights.addDisplayableLink(
                          new DisplayableLinkImpl(so, target, acLink ), result);
                  haveLessResult = true;
              } // if (unmatch
            } // if (less
            
            view.finalizeForCopy();
            
            if ((less == null && haveMoreResult) || (more == null && haveLessResult) ||
                    (haveMoreResult && haveLessResult)) break;

            if (Thread.interrupted()) {
                throw new InterruptedException();
            }
      } // ito on so from l_testedEPs

                        // for testing and tracking
                        //ImmutableName key;
                        //System.out.println(" ");
                        //System.out.println("RightsFactory_Facade.detectNoThanRights(), "+
                                //"for "+so.getName()+
                          //"the DisplayableLinkImpls in the final result are:");
                        //for (Iterator ite = result.keySet().iterator() ; ite.hasNext() ; ) {
                           //key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)result.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite
                        //System.out.println(" ");
      
      l_testedEPs = null;
      unmatch = null;
      view = null;
      m_baseObj = null;
      target = null;
      constantName = null;
      intermResult = null;
      more = null;
      less = null;
      acLink = null;
      nodeRights = null;
      ago = null;
      compound = null;
      linkRights = null;
      key = null;
      
      return result;
  } // detectNoThanRights()
  
  /**
  * This important method selects the current effective rights from the DisplayableLinkImpls
  * associated to each BaseObject pair, for an EPR ViewInBase.
  * Puts in each last AccessControlLink the rights synthesis for the 'See why' text.
  * The pair may be: <P>
  * - (EligibleParty, Basic or VirtualFolder), <P>
  * - (GroupIDMember, GroupIDMember), <P>
  * - (Actor, GroupIDMember), for actors with currentUID/GID, <P>
  * - (VirtualFolder, Basic), <P>
  *   in an EPRViewInBase, and : <P>
  * - by the way of a simple 'add' operation, plus some indicators as 'acl',
  *   for positive rights. <P>
  * - by the way of 'remove' operations when a negative acl forbids a right,
  *   or when 'no rights' forbids all rights. <P>
  * The negative-right types are displayed only if the link is simple.
  * These negative rights are then subtracted from the positive rights in the other links. <P>
  * If a DisplayableLinkImpl contain an Actor, adds '<via>' in the result to separate
  * the rights through an actor from the other rights, but this is true only
  * if there is no negative rights in one DisplayableLinkImpl.
  * The SPREAD comments in the last AccessControlLinkImpl of each DisplayableLinkImpl
  * are reserved to the list of all the effective rights, for the SeeWhy text.
  * Called by GraphicEPRView.buildUpArrowsAndRights().
  * The DisplayableLinks must have been set in _m_l_DisplayableLinks.
  * Calls FactoryUtilities.withSomeAccess(), isActivePath(), getRightAcronym(),
  * DisplayableLinkUtilities.getEffectiveRightsAndCommentOnRights(),
  * UtilityImpl.selectHeaders().
  * @param _so BaseObject of the view
  * @param _tg BaseObject of the view
  * @param _m_l_DisplayableLinks immutable Map of DisplayableLinks lists.
  *             May update the last AccessControlLink of each DisplayableLink.
  * @return String rights in DisplayableLinks having _ep and _res as ends,
  *             and having non-null rights. May be empty or null. May contain '<via>'
  *             to indicate the given rights through actors acting as proxies in the paths.
  * @see ARoad0.CNot.AccessControlLink
  * @see ARoad0.gBase.UtilityImpl#subtractNegativeNotRemovingInOneACS(java.util.Set, java.util.Set, ARoad0.gBaseInterface.ImmutableACS, boolean) 
  * @see ARoad0.gWork.DisplayableLinkUtilities#getEffectiveRightsAndCommentOnRights(ARoad0.CNot.DisplayableLinkImpl, boolean)
  **/
   public final String getEffectiveRightsForPair(
    BaseObject _so, BaseObject _tg, Map _m_l_DisplayableLinks) {
       
      Name pairKey = ((Name)_so.getName()).extended(_tg.getName());
      if (!_m_l_DisplayableLinks.containsKey(pairKey) ) return null;
      
       // variables
      StringBuilder             result = new StringBuilder();
      DisplayableLinkImpl       dLink;
      List                      l_dLinks;
      Set<StringRight>[]        l_rights_comment = null;
      StringRight               right = null;
      String                    sright;
      StringBuilder             comment = new StringBuilder();
      boolean                   simpleNegRightOnOneLink = false;
      boolean                   simpleNegRightForPair = false;
      boolean                   viaForLastLinkStringResult = false;

      l_dLinks = (List)_m_l_DisplayableLinks.get(pairKey);//).clone();

                                          // for testing and tracking
                                          //if (_tg.getNickName().equals("localhost"))
                                            //  System.out.println("RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                              //  "start of the method for the pair: "+pairKey);

      // initialize 'simpleNegRightForPair'
      for (Iterator ite = l_dLinks.iterator() ; ite.hasNext() ; ) {
            dLink = (DisplayableLinkImpl)ite.next();      
              
            simpleNegRightOnOneLink = false;
            if (dLink.isSimple() &&
                    dLink.getEorLastLink().isTrueForOneNegativeLinkType()) {
                            simpleNegRightForPair = true;
                            
                                          // for testing and tracking
                                          //System.out.println("-------- "+
                                            //      "simpleNegRightForPair 'true' for dLink "+dLink);
                            break;
            } 
      } // for ite


      // searching for all DisplayableLinks associated to _ep and _res,
      // and put their last rights in l_result_ or l_noActorResult_
      for (Iterator ite = l_dLinks.iterator() ; ite.hasNext() ; ) {
            dLink = (DisplayableLinkImpl)ite.next();      
              
            simpleNegRightOnOneLink = false;
            if (dLink.isSimple() &&
                    dLink.getEorLastLink().isTrueForOneNegativeLinkType()) {
                            simpleNegRightOnOneLink = true;

                                          // for testing and tracking
                                          //System.out.println("-------- "+
                                            //      "simpleNegRightForPair 'true' for dLink "+dLink);
            } // if
            else {

                                          // for testing and tracking
                                          //System.out.println("-------- dLink.isSimple(): "+dLink.isSimple()+
                                            //      " and dLink.getEorLastLink().isTrueForOneNegativeLinkType(): "+
                                              //    dLink.getEorLastLink().isTrueForOneNegativeLinkType());
            } // else
            
            // a negative type is displayed if it is simple
            if (!factoryUtil_.isActivePath(dLink) && !simpleNegRightOnOneLink ) {

                              //System.out.println("RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                //"this is not an active path nor a simple negative acl path: "+dLink);

                 continue;
            }

            // l_rights_comment[0] contains StringRights for the returned String of this method, and
            // l_rights_comment[1] contains StringRights for the 'See why' text
            l_rights_comment = linkUtil_.getEffectiveRightsAndCommentOnRights(dLink, true);
            
            if (factoryUtil_.withSomeAccess(l_rights_comment[0]) == null &&
                    !dLink.getEorLastLink().isTrueForOnePseudoStructuralLinkType() ) {

                              // for testing and tracking
                              //System.out.println("RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                //"this is not an access path nor a negative acl- l_rights_comment[0]: "+
                                //dLink);
                continue;
            }

                              // for testing and tracking
                              //System.out.println("RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                //"l_forCommentRights: ");
                              //for (Iterator ita = l_rights_comment[1].iterator(); ita.hasNext(); ) {
                                  //System.out.println("------------ "+ita.next());
                              //} // for ite

            // ------------ secondary task for this method ---------------------------------
            // for the 'See why' text, from l_rights_comment[1],
            // put in the last AccessControlLink the rights synthesis.
             // do not put the nick name
            // AccessControlLinkImpl adds 'All rights' to the SPREAD comment
            if (l_rights_comment[1] != null) {
                for (Iterator ita = l_rights_comment[1].iterator(); ita.hasNext(); ) {
                    comment.append(((StringRight)ita.next()).getRight());
                    if (ita.hasNext()) comment.append("/");
                } // for ita
            } // if
            dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.SPREAD);
            if (l_rights_comment[1] != null && !l_rights_comment[1].isEmpty()) 
                dLink.getEorLastLink().setLinkTypeComments(
                    AccessControlLink.SPREAD, comment.toString());  
            //else ' (no rights)' is added by AccessControlLinkImpl.

            // ------------ end of the secondary task for this method ----------------------
            
            // adds the effective rights of the last link in one of the two sets,
            // but if 'simpleNegRightForPair' is true, the rights subtraction has to be done
            // on l_noActorResult_ with all the applicable rights
            if (dLink.getEorLastActorInPath() == null || simpleNegRightForPair) {
                l_noActorResult_.addAll(l_rights_comment[0]);
            }
            else { 
                // dLink has actor in the path: '<via>' has to be put in the result
                // note: adding the actor name should be too long and complex
                l_result_.addAll(l_rights_comment[0]);
            } // else

            if (//dLink.isSimple() && 
                    l_rights_comment[0].isEmpty() &&
                    !dLink.containsStrengthenLinkTypes() && // other processing otherwise
                    dLink.getEorLastLink().isTrueForOnePseudoStructuralLinkType()) {
                l_lastLinkStringResult_.addAll(dLink.getEorLastLink().
                        getL_ActivatedPseudoStructuralTypes());

                if (dLink.getEorLastActorInPath() != null)
                    viaForLastLinkStringResult = true;
            } // if
            
            // reinitialization of comment
            comment.delete(0, comment.length());
            
      } // for ite
 
                                      // for testing and tracking
                                      //System.out.println("in RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                        //"l_noActorResult_ size = "+l_noActorResult_.size());

      // filtering of negative rights and set the header rights
      try {
          if (l_noActorResult_.size() > 1) {
            // substract the negative rights.
            // no acs, so generic rights may be returned
            // no node, since the inheritance rights on node are not relevant
            // this method always updates its argument
            if (simpleNegRightForPair)
                l_noActorResult_ = utility_.substractNegativeFromPositiveRights(l_noActorResult_);

            if (l_noActorResult_.size() > 1) {

                                      // for testing and tracking
                                      //System.out.println("in RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                        //"call to selectHeaders() on l_noActorResult_");

                utility_.selectHeaders( (Set<StringRight>) l_noActorResult_);
            } // if
          } // if (l_noActorResult_.size()
          
          if (!l_result_.isEmpty() && l_result_.size() > 1)
            utility_.selectHeaders( (Set<StringRight>) l_result_);
      } catch (BaseError b) {
          throw new InternalError("RightsFactoryImpl.getEffectiveRightsForPair() "+b);
      }

      
                      // for testing and tracking
                      //System.out.println("in RightsFactory_Facade.getEffectiveRightsForPair(), "+
                        //"l_noActorResult_ after filtering: ");
                      //for (Iterator ita = l_noActorResult_.iterator(); ita.hasNext(); ) {
                        //  System.out.println("------------ "+ita.next());
                      //} // for ite
                      //System.out.println("in RightsFactory_Facade.getEffectiveRightsForPair(), "+
                        //"l_result_ after filtering: ");
                      //for (Iterator ita = l_result_.iterator(); ita.hasNext(); ) {
                        //  System.out.println("------------ "+ita.next());
                      //} // for ite
                      //System.out.println("in RightsFactory_Facade.getEffectiveRightsForPair(), "+
                        //"l_lastLinkStringResult_ after filtering: ");
                      //for (Iterator ita = l_lastLinkStringResult_.iterator(); ita.hasNext(); ) {
                        //  System.out.println("------------ "+ita.next());
                      //} // for ite


      if (!l_noActorResult_.isEmpty()) {
          for (Iterator ita = l_noActorResult_.iterator() ; ita.hasNext() ; ) {
            right = (StringRight)ita.next();

            // display first the direct rights
            sright = factoryUtil_.getRightAcronym(right);
            if (sright != null) result.append( sright );
            else result.append( right.getNickName() );
            result.append(" ");

          } // for (Iterator ita
      } // if (!l_noActorResult_.isEmpty())

        // dLink has actor in the path: '<via>' has to be put in the result
      if (!l_result_.isEmpty() && !simpleNegRightForPair) {
          result.append("<via> ");
          for (Iterator ita = l_result_.iterator() ; ita.hasNext() ; ) {
            right = (StringRight)ita.next();
            sright = factoryUtil_.getRightAcronym(right);
            if (sright != null) result.append( sright );
            else result.append( right.getNickName() );
            result.append(" ");
          } // for (Iterator ita
      } // if (!l_result_.isEmpty())

      if (!l_lastLinkStringResult_.isEmpty() && !simpleNegRightForPair) {
          for (Iterator ita = l_lastLinkStringResult_.iterator() ; ita.hasNext() ; ) {
            sright = (String)ita.next();

            if (viaForLastLinkStringResult && result.indexOf("<via>") == -1)
                result.append("<via> ");

                                  // for testing and tracking
                                  //System.out.println(
                                    //        "RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                      //      "adds '<via>' for l_lastLinkStringResult_");

            // display the strengthen rights for dLinks
            // without containsStrengthenLinkTypes() true
            // (for which they are blue lines in the graphic view)
            result.append( sright );
            result.append(" ");

          } // for (Iterator ita
      } // if (!l_lastLinkResult_.isEmpty())

                                  // for testing and tracking
                                  //if (_tg.getNickName().equals("localhost"))
                                    //      System.out.println(
                                      //      "RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                        //    "result: "+result);
                      
      // clear for the next call to this method
      l_noActorResult_.clear(); 
      l_result_.clear();
      l_lastLinkStringResult_.clear();

      comment = null;
      l_dLinks = null;
      dLink = null;
      pairKey = null;
      sright = null;
      right = null;
      l_rights_comment = null;
      
      if (result.length() != 0) return result.toString(); 
      else return null;
  } // getEffectiveRightsForPair()

  /**
  * Constructs a JTextPane with all the DisplayableLinks and AccessControlLinks,
  * to be displayed as a comment for the GraphicView associated
  * to _m_l_DisplayableLinks. NO USE.
  * @param _gview is the graphic view associated to the view
  * @param _m_l_DisplayableLinks is an immutable Map of DisplayableLinks lists.
  * @param _view associated to _m_l_DisplayableLinks
  * @return JTextPane with text, as a comment for the GraphicView
  * associated to _m_l_DisplayableLinks.
  **/
  public static synchronized final JTextPane getWhyText(GraphicView _gview,
    Map _m_l_DisplayableLinks, ViewInBase _view) {

    return FactoryUtilities.getWhyText(_gview, _m_l_DisplayableLinks, _view);
  } // getWhyText()
   
  /**
  * Constructs a JTextPane with all the DisplayableLinks and AccessControlLinks,
  * to be displayed as a comment for the GraphicView associated
  * to _m_l_DisplayableLinks.
  * Called by GraphicView.setWhyText() and GraphicEPRView.setWhyText().
  * Calls directly FactoryUtilities.getEPRWhyText().
  * @param _gview is the graphic view associated to the view
  * @param _m_l_DisplayableLinks is an immutable Map of DisplayableLinks lists.
  * @param _view associated to _m_l_DisplayableLinks
  * @return JTextPane with text, as a comment
  *   for the GraphicView associated to _m_l_DisplayableLinks.
  **/
  public final JTextPane getEPRWhyText(GraphicEPRView _gview,
    Map _m_l_DisplayableLinks, EPRViewInBase _view) {

    return factoryUtil_.getEPRWhyText(_gview, _m_l_DisplayableLinks, _view);
  } // getEPRWhyText()
   
  /**
  * Constructs a JTextPane with all the DisplayableLinks and AccessControlLinks,
  * to be displayed as a comment for the GraphicNoThanView associated
  * to _m_l_DisplayableLinks.
  * Called by GraphicView.getNoThanWhyText().
  * Calls directly FactoryUtilities.getNoThanWhyText().
  * @param _gview is the graphic view associated to the view
  * @param _m_l_DisplayableLinks is an immutable Map of DisplayableLinks lists.
  * @param _view associated to _m_l_DisplayableLinks
  * @return JTextPane with text, as a comment for the GraphicView associated
  * to _m_l_DisplayableLinks.
  **/
  public final JTextPane getNoThanWhyText(GraphicNoThanView _gview,
    Map _m_l_DisplayableLinks, NoThanViewInBase _view) {

    return factoryUtil_.getNoThanWhyText(_gview, _m_l_DisplayableLinks, _view);
  } // getNoThanWhyText()
  
  /**
  * Sets the current effective rights from the DisplayableLinks associated
  * to each BaseObject pair in a view, after a parsing of all the direct and indirect
  * paths from the first end to the second end of the pair. <P>
  * This method is a simplified form of getEffectiveRightsForPair() with a different return.
  * The negative types in the last access control link are processed.
  * Each key into the argument _m_l_DisplayableLinks has to produce a key in the result,
  * and an empty set for a key means it is a structural link. <P>.
  * Called by detectEPRRights(), detectNoThanrights(), and by
  * CompoundRightsFactoryImpl.addPathFromNegativeRightsInOneACS(),
  * by fulfilNoThanRightCriteria() for the NoThan ViewInBases.
  * Calls FactoryUtilities.isActivePath(), UtilityImpl.selectHeaders() and
  * DisplayableLinkUtilities.getEffectiveRightsAndComment().
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  *   associated to the view. No update.
  * @return map where the keys are the _m_l_DisplayableLinks keys where rights are activated,
  *     and the value is a set of StringRights (may be an empty set, never null).
  **/
  public final Map<ImmutableName, Set<StringRight>> getFirstShapeEffectiveRights(
        Map _m_l_DisplayableLinks) {

       // variables
      Map<ImmutableName, Set<StringRight>>   m_result = new HashMap(INITIAL_CAPACITY);
      Map.Entry                 entry = null;
      DisplayableLinkImpl       dLink;
      List                      l_dLinks;
      Set<StringRight>          l_result = null;
      boolean                   simpleNegRightOnOneLink = false;
      boolean                   simpleNegRightForPair = false;


                      // for testing and tracking
                      //System.out.println("RightsFactory_Facade.getFirstShapeEffectiveRights() start. ");

      try {
          for (Iterator ito = _m_l_DisplayableLinks.entrySet().iterator() ; ito.hasNext() ; ) {

              entry = (Map.Entry)ito.next();

              l_dLinks = (List)entry.getValue();
              l_result = null;
              simpleNegRightForPair = false;

              // searching for on all DisplayableLinks, and put their last rights in l_result
              for (Iterator ite = l_dLinks.iterator() ; ite.hasNext() ; ) {
                    dLink = (DisplayableLinkImpl)ite.next();

                    simpleNegRightOnOneLink = false;
                    if (dLink.isSimple() &&
                            dLink.getEorLastLink().isTrueForOneNegativeLinkType()) {
                                    simpleNegRightOnOneLink = true;
                                    simpleNegRightForPair = true;
                    }

                    // NEG_ACL, NEG_LPRI and NEG_TPRI are the only negative types
                    // for which dLink is displayed,
                    // even if it is not simple (not as getEffectiveRightsForPair())
                    if (!factoryUtil_.isActivePath(dLink) && !simpleNegRightOnOneLink ) {

                                      // for testing and tracking
                                      //System.out.println("RightsFactory_Facade.getFirstShapeEffectiveRights(), "+
                                        //"this is not an active path nor a negative acl path: "+dLink);
                        continue;
                    }

                    if (l_result == null)
                        l_result = linkUtil_.
                                getEffectiveRightsAndCommentOnRights(dLink, false)[0];
                    else
                        l_result.addAll(linkUtil_.
                                getEffectiveRightsAndCommentOnRights(dLink, false)[0]);
              } // for ite

              if (l_result == null) continue;

                                          // for testing and tracking
                                          //System.out.println("RightsFactory_Facade.getFirstShapeEffectiveRights(), ");
                                          //System.out.println("simpleNegRightForPair: "+simpleNegRightForPair+
                                            //      " and l_result before substractNegativeFromPositiveRights() for "+
                                              //    entry.getKey());
                                          //for (Iterator ita = l_result.iterator(); ita.hasNext(); ) {
                                            //  System.out.println("--- "+ita.next());
                                          //} // for ita

              // substract negative rights.
                // this method always updates its argument
              if (l_result.size() > 1 && simpleNegRightForPair) {
                  l_result = utility_.substractNegativeFromPositiveRights(l_result);
              }

              if (l_result.size() > 1) {

                                          // for testing and tracking
                                          //System.out.println("in RightsFactory_Facade.getEffectiveRightsForPair(), "+
                                            //"call to selectHeaders() on l_noActorResult_");

                    utility_.selectHeaders(l_result);
              } // if

                                          // for testing and tracking
                                          //System.out.println("RightsFactory_Facade.getFirstShapeEffectiveRights(), ");
                                          //System.out.println("--- adds for the key "+
                                            //      entry.getKey());
                                          //System.out.println("--- l_result: ");
                                          //for (Iterator ita = l_result.iterator(); ita.hasNext(); ) {
                                            //  System.out.println("------ "+ita.next());
                                          //} // for ita

              m_result.put((ImmutableName)((Name)entry.getKey()).clone(), l_result);
          } // for ito
      }
      catch (BaseError b) {
        throw new InternalError("from RightsFactory_Facade.getFirstShapeEffectiveRights(): "+b);
      }

                      // for testing and tracking
                      //System.out.println("RightsFactory_Facade.getFirstShapeEffectiveRights(), "+
                        //"result: ");
                      //for (Map.Entry<ImmutableName, Set<StringRight>> ent : m_result.entrySet()) {
                        //  System.out.println("-- key: "+ent.getKey());
                          //for (StringRight rig : ent.getValue()) {
                            //  System.out.println("---- right: "+rig);
                          //} // for rig
                      //} // for name
                      
      l_dLinks = null;
      dLink = null;
      l_result = null;
      entry = null;
      return m_result;
  } // getFirstShapeEffectiveRights()

  /**
  * Tests the rights in the last AccessControlLink of each DisplayableLink, to
  * say if the criterion right is fulfilled or not. Uses StringRight.compareTo()
  * between the rights, through their metarights if at least one right is a metaright.
  * If _moreCriterion is false and there is an empty _m_l_DisplayableLinks, the returned right
  * is StringRightImpl.EMPTY_INSTANCE. Using StringRight.compareTo(), the
  * right internal key is the first compared property. Thus, this method
  * may detect the criterion right is not fulfilled while the effective right has no
  * direct path to the criterion right, in the network of rights. This is
  * a defensive behavior, which may be considered too restrictive in some cases.
  * On the other hand, the Access Road user gets the effective right
  * in the 'See why' text. He may so decide the gravity to get no fulfillment.
  * Called by detectNoThanRights().
  * @param _m_l_DisplayableLinks is an immutable Map of DisplayableLinks lists
  *     for one couple of BaseObjects in a view
  * @param _right is the more or less criterion
  * @param _moreCriterion true for checking a more criterion, and false for a less criterion
  * @return null if the criterion is fulfilled, or a non-null right if,
  * when _moreCriterion is true, _right is strictly lesser than
  * the returned value, that is one of the effective global rights for the couple of BaseObjects,
  * and also return a non-null right if, when _moreCriterion is false,
  * _right is strictly greater than the returned value, that is a right for the couple
  * of BaseObjects.
  **/
  public final StringRight fulfilNoThanRightCriteria(
          Map _m_l_DisplayableLinks, StringRight _right, boolean _moreCriterion) {
      
     StringRight        result = null;
     StringRight        right = null;
     Set<StringRight>   l_trueRights = null;
     
     if (_moreCriterion) {
         if (_m_l_DisplayableLinks.isEmpty()) {
             return result;
         }
                      // for testing and tracking
                      //System.out.println("RightsFactory_Facade.fulfilNoThanRightCriteria(), "+
                        //"_moreCriteria true for the criterion "+_right);
                      //System.out.println("--- and the metaright: "+_right.getEorMetaRight());

         l_trueRights = getFirstShapeEffectiveRights(_m_l_DisplayableLinks).
                                        values().iterator().next();
         for (Iterator ite = l_trueRights.iterator(); ite.hasNext(); ) {
             right = (StringRight)ite.next();

                      // for testing and tracking
                      //System.out.println("true right.getEorMetaRight(): "+right.getEorMetaRight());
                      
             
             if (right.getEorMetaRight().compareTo(_right.getEorMetaRight()) > 0) {
                 result = right;
                 break;
             }
         } //for ite

                      // for testing and tracking
                      //System.out.println("RightsFactory_Facade.fulfilNoThanRightCriteria(), "+
                        //"result: "+result);
                      
     } // if (_moreCriteria
     else {


                        // for testing and tracking
                        //System.out.println("In RightsFactory_Facade.fulfilNoThanRightCriteria(), "+
                          //" for the less criterion, the DisplayableLinkImpls are:");
                        //ImmutableName key = null;
                        //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                          // key = (ImmutableName) ite.next();
                           //System.out.println("--- for key: "+key);
                           //for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                            //System.out.println("------------"+
                              //((DisplayableLinkImpl)ita.next()).toString() );
                          //} // for ita
                        //} // for ite


         if (_m_l_DisplayableLinks.isEmpty()) {
             return StringRightImpl.EMPTY_INSTANCE;
         }
         l_trueRights = getFirstShapeEffectiveRights(_m_l_DisplayableLinks).
                                        values().iterator().next();
         for (Iterator ite = l_trueRights.iterator(); ite.hasNext(); ) {
             right = (StringRight)ite.next();
             result = right;

                          // for testing and tracking
                          //System.out.println("RightsFactory_Facade.fulfilNoThanRightCriteria(): ");
                          //System.out.println("--- for the less criteria,"+_right+
                            //      " and the current right "+right+", the comparison gives :"+
                              //    right.getEorMetaRight().compareTo(_right.getEorMetaRight()));


             if (right.getEorMetaRight().compareTo(_right.getEorMetaRight()) >= 0) {
                 result = null;

                          // for testing and tracking
                          //System.out.println("RightsFactory_Facade.fulfilNoThanRightCriteria(), "+
                            //      "for the less criteria,"+_right+
                              //    " result is "+result+", due to "+right);
                 
                 break;
             }
         } //for ite         
     } // else
     
                    //if (!result) {
                          // for testing and tracking
                      //    System.out.println("RightsFactory_Facade.fulfilNoThanRightCriteria(), "+
                        //    "true rights with a result "+result+" for the criteria "+_right);
                        // for (Iterator ite = l_trueRights.iterator(); ite.hasNext(); ) {
                          //   right = (StringRight)ite.next();
                          //   System.out.println("-------- "+right);
                        // } //for ite  
                    //} // if
     
     l_trueRights = null;
     right = null;
     return result;
  } // fulfilNoThanRightCriteria()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
      l_result_.clear();
      l_noActorResult_.clear(); 
      l_lastLinkStringResult_.clear();
      l_result_ = null;
      l_noActorResult_ = null; 
      l_lastLinkStringResult_ = null;
      
      // for the EPR and NoThan ViewInBases
      interpreter_ = null;
      factoryUtil_ = null;
      linkUtil_ = null;
      utility_ = null;
  } // finalizeForProcess()
  
} // RightsFactory_Facade
