//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007-2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.HashSet;
import java.util.Set;
import java.util.HashMap;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;

import ARoad0.ErrB.BaseError;
import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.Gui2.TreeManager;

import ARoad0.gBaseInterface.*;
import ARoad0.gWorkInterface.*;

import ARoad0.gBase.UtilityImpl;
import ARoad0.gBase.ACSImpl;
import ARoad0.gBase.AclEntryImpl;
import ARoad0.gBase.GroupIDImpl;

/**
 * This class is a change event adaptor for all the BaseObjects associated to a view,
 * and it is responsible to manage the paths search algorithms for each view.
 * It takes account of the AcsAddons. <P>
 * This is a component of the event management pattern
 * for the views. This class informs the view GraphicViewBaselistenerImpl
 * when the rights may have changed. It is independant to the graphic view,
 * which may be deleted and created several times by GraphicViewBaselistenerImpl.
 * This responsability implies, for each view-related object and through
 * the method setRightsNetwork(), to follow: <P>
 *  - every ACS bridge change, <P>
 *  - every bound property change, including the AclEntry rights, <P>
 *  - every deleted BaseObject which is in an AccessControlLink, <P>
 *  - every new ACS userID with 'root' rights, <P>
 *  - every new ACS actor belonging to its userID, its groupID, <P>
 *    or to a groupID which contains its userID, its groupID, <P>
 *  - every new ACS userID belonging to its groupID, <P>
 *    or to a groupID which contains its userID, its groupID, <P>
 *  - every new ACS actor executed through 'other' rights with a current account
 *    or group, <P>
 *  - every PrivilegeForLinks change in the sources and the targets, <P>
 *  - every specific property change from an AcsAddon object, <P>
 * but to not follow a view-related object when it is a new element of the view,
 * since it is handled directly by the BaseListener.
 * The AclEntries and external linked Privileges are switched when their ACS is saved
 * (see gBase.BaseManagerImpl). To avoid useless right analysis, ActionSave and
 * ActionSaveAll unregister this class before any ACS saving, to not process
 * the property changes from the AclEntries and Privileges.
 * The Gui2.GraphicViewBaseListenerImpl instance that works with
 * the RightsMediatorImpl instance must register itself
 * as a PropertyChangeListener of this instance.
 * It must have a ViewInBase as main event source.
 * The interface methods ACSObjectAddon.getAddonInnerPropertiesToListen() and
 * getAddonParentPropertiesToListen() are used to handle the AcsAddon
 * property change events. The aim is to update any view
 * which contains a node from such an AcsAddon. These methods are ridden by
 * ARoad0.UtilityImpl that is invoked by RightsMediatorImpl. <P> <P>
 * The second responsibility of this class is the selection of algorithms.
 * Following the design pattern Interpreter in the GoF collection,
 * this class manages the path search algorithms of EPRViewInBase, NoThanViewInBase
 * and SketchViewInBase.
 * Each view uses dedicated instances of 8 generic gWork classes and 1 generic gBase
 * class that implement the core algorithms.
 * Each view uses, for each view node and each intermediate node in an access path,
 * one instance of NodeRights and one instance of LinkRights.
 * All these objects are created and managed by this RightsMediatorImpl class.
 * For an EPRViewInBase and a NoThanViewInBase, the instances for the old view nodes
 * are never removed, including the NoThanViewInBase first target,
 * nor the instances for the old intermediate nodes.
 * This is to be ready to the changes in the access paths. <P>
 * An AcsAddon may include some subclasses of NodeRightsImpl and LinkRightsImpl,
 * and they are selected by RightsMediatorImpl when a node from this AcsAddon is implied.
 * The detection of the AcsAddon packages follows the name pattern of these packages.
 * If there is no NodeRights or LinkRights implementation in the AcsAddon gWork
 * package, the generic class is selected. <P>
 * All the algorithm class instances in the gWork and AcsAddon.gWork packages
 * that are necessary in the access paths search are stored in 3 dedicated maps
 * in RightsMediatorImpl. They are called by RightsFactory_Facade - which calls the core
 * algorithm classes - and by the core algorithm classes themself. <P>
 * The instances of this class are created by Gui2.ViewModel and
 * ViewInBase_Facade.createSketchView(). Gui2.GraphicViewBaseListenerImpl
 * is the single listener of an instance.
 * @version 0.7.0
 */
public class RightsMediatorImpl implements RightsMediator {
  private static final int     PRIME = 1000003;
  private static final int     INITIAL_CAPACITY = 15;// default value for map: 101


                    private     int count_ = 0;

  /**
   * The BaseListener main event source which defines the rights network
   * is a ViewInBase
   */
  protected ViewInBase          view_;

  /**
   * Set of hidden BaseObjects which are implied in the rights analysis for the view
   */
  protected Set                 l_RightsNetwork_ = null;

  /**
   * list of PropertyChangeListeners.
   */
  private final Set             l_RightsChangeSupport_ = new HashSet(INITIAL_CAPACITY);
  
  /**
   * last property change event received by propertyChange()
   */
  private PropertyChangeEvent   lastEvent_ = null;

  /**
   * Map of the one instance for each of the 5 classes gWork.xxxRightsFactoryImpl,
   * the 2 gWork.yyyUtilities and gWork.UtilityImpl.
   * This provides to the view a dedicated set of instances for the core algorithms
   * to use in the worker thread of the view.
   * The keys are in L_CORE_ALGORITHMS_KEYS.
   */
  private Map                   m_CoreAlgorithms_ = new HashMap(INITIAL_CAPACITY);

  /**
   * Map of the instances of the NodeRights or its subclasses for the view nodes
   * and the analyzed or intermediate nodes.
   * This provides to each node a dedicated instance. The keys are the node names.
   */
  private Map<ImmutableName, NodeRights>  m_NodeRights_ = new HashMap(INITIAL_CAPACITY*10); // capacity: 150

  /**
   * Map of the instances of the LinkRights or its subclasses for the view nodes
   * and the analyzed or intermediate nodes.
   * This provides to each node a dedicated instance. The keys are the node names.
   */
  private Map<ImmutableName, LinkRights> m_LinkRights_ = new HashMap(INITIAL_CAPACITY*10); // capacity: 150

  //-----------------------------------------------------------------------

  /**
   * The view GraphicViewBaseListener must register itself
   * as a PropertyChangeListener of this RightsNetworkObserver instance.
   * It must have a ViewInBase as main event source.
   * All the CoreAlgorithm instances from the gWork and AcsAddon.gWork packages
   * that are necessary in the access path search are created and stored
   * in the maps m_CoreAlgorithms_, m_NodeRights_ and m_LinkRights_.
   * The NodeRightsImpl instance and the LinkRights instance of each node
   * in the view are created.
   * Called by Gui2.ViewModel and ViewInBase_Facade.createSketchView().
   * @see ARoad0.Gui2.ViewModel#createViewRightsMediator(ARoad0.gBaseInterface.ViewInBase)
   * @param _view is the source for which the rights network is defined
   */
  public RightsMediatorImpl(ViewInBase _view) {

      if (_view == null) {
         throw new NullPointerException("null view from the "+
           "RightsMediatorImpl constructor");
      }

      view_ = _view;

      try {
        // initialize in the right order, since a constructor may use
          // the previously-constructed instances
        m_CoreAlgorithms_.put("BaseUtil", new UtilityImpl());
            // DisplayableLinkUtilities calls back getCoreAlgorithm("BaseUtil")
        m_CoreAlgorithms_.put("LinkUtil", new DisplayableLinkUtilities(this));
            // FactoryUtilities calls back getCoreAlgorithm()
                // for "BaseUtil" and "LinkUtil"
        m_CoreAlgorithms_.put("FactoryUtil", new FactoryUtilities(this));
            // RightsFactory_Facade calls back getCoreAlgorithm()
                // for "BaseUtil", "LinkUtil" and "FactoryUtil"
        m_CoreAlgorithms_.put("Facade", new RightsFactory_Facade(this));
        
        m_CoreAlgorithms_.put("Acl", new AclRightsFactoryImpl(this));
        m_CoreAlgorithms_.put("Ago", new AgoRightsFactoryImpl(this));
        m_CoreAlgorithms_.put("Pri", new PrivilegeRightsFactoryImpl(this));
        
        m_CoreAlgorithms_.put("Compound", new CompoundRightsFactoryImpl(this));
        m_CoreAlgorithms_.put("3nodes", new ThreeNodesRightsFactoryImpl(this));
     

        updateNonCoreAlgorithmsForViewNodes(false);

              // for testing purpose
              //System.out.println("in RightsMediatorImpl constructor,");

      } catch (Exception l) {
        //l.printStackTrace();
      }
  } // constructor

  // ----------------------- PUBLIC METHODS ---------------------------------

  /**
   * Gets the ViewInBase.
   * @return the view of this mediator
   */
  public BaseObject getSource() {
    return view_;
  }

  /**
   * Gets the instance to use for this view for a given core algorithm,
   * as instance of one of the 9 classes RightsFactory_Facade, AclRightsFactoryImpl,
   * AgoRightsFactoryImpl, CompoundRightsFactoryImpl, DisplayableLinkUtilities,
   * PrivilegeRightsFactoryImpl, FactoryUtilities, ThreeNodesRightsFactoryImpl
   * and gBase.UtilityImpl.
   * @param _key belongs to AlgorithmInterpreter.L_CORE_ALGORITHMS_KEYS
   * @return the core algorithm for the view, or null
   */
  public CoreAlgorithm getCoreAlgorithm(String _key) {
    if (_key == null || !L_CORE_ALGORITHMS_KEYS.contains(_key))
        return null;

    return (CoreAlgorithm)m_CoreAlgorithms_.get(_key);
  } // getCoreAlgorithm()

  /**
   * Gets the map of the NodeRights instances to use for each view node, analyzed or
   * intermediate node. This includes the first target of a NoThanViewInBase.
   * The term of 'analyzed' is here applicable to a node
   * that is analyzed by the gWork classes, even if it is not finally detected as
   * an intermediate node to put in a stable DiaplayableLinkImpl.
   * @return unmodifiable map of NodeRights instances, where the node names
   * are the keys
   */
  public Map<ImmutableName, NodeRights> getNodeRightsInstances() {
    return Collections.unmodifiableMap(m_NodeRights_);
  } // getNodeRightsInstances()

  /**
   * Gets the map of the LinkRights instances to use for each view node, analyzed
   * or intermediate node. This includes the first target of a NoThanViewInBase.
   * The term of 'analyzed' is here applicable to a node
   * that is analyzed by the gWork classes, even if it is not finally detected as
   * an intermediate node to put in a stable DiaplayableLinkImpl.
   * @return unmodifiable map of LinkRights instances, where the node names
   * are the keys
   */
  public Map<ImmutableName, LinkRights> getLinkRightsInstances() {
    return Collections.unmodifiableMap(m_LinkRights_);
  } // getLinkRightsInstances()

  /**
   * When the view nodes have changed, this method updates the NodeRightsImpl
   * and the LinkRights instances of the view nodes.
   * For a NoThanViewInBase, it manages all the BaseObjects into the view checking perimeter
   * and the view first target. <P>
   * On the other hand, the analyzed nodes and the intermediate nodes of the view
   * are not processed there.
   * The term of 'analyzed' is here applicable to a node that is processed
   * by the gWork classes in the paths search, even if it is not finally
   * detected as an intermediate node to put in a stable DiaplayableLinkImpl.
   * For an EPRViewInBase and a NoThanViewInBase, the old view nodes are never removed,
   * including the NoThanViewInBase first target, nor the old intermediate nodes.
   * This is to be ready to the possible future changes in the access paths. <P>
   * Called by the constructor, Gui2.ViewModel.changeNonCoreAlgorithms() and
   * ViewInBase_Facade.populateSketchView().
   * When the parameter is false, the previous view nodes and the intermediate nodes
   * keep their NodeRights and LinkRights instances to speed up the processing.
   * @param _clearIntermediateNodes true only for the sketch view, to finalize
   * the NodeRights and LinkRights instances of all the nodes that are
   * not in the view, including all the previous intermediate nodes
   */
  public synchronized void updateNonCoreAlgorithmsForViewNodes(boolean _clearIntermediateNodes) {
    if (_clearIntermediateNodes && !(view_ instanceof SketchViewInBase))
        throw new InternalError("RightsMediatorImpl.updateNonCoreAlgorithmsForViewNodes() "+
                "clears the intermediate nodes only for the sketch views.");
      
    Set l_oldNodeNames = new HashSet(m_NodeRights_.keySet());

                            // for testing and tracking
                          //System.out.println("in RightsMediatorImpl.updateNonCoreAlgorithmsForViewNodes(),"
                            //+" the target of the view "+view_.getName()+" are :");
                          //ImmutableTarget targ = null;
                          //for (Iterator ite = ((EPRViewInBase)view_).getEorM_Targets().values().iterator() ; ite.hasNext() ; ) {
                            //targ = (ImmutableTarget) ite.next();
                            //System.out.println("---"+targ.getName());
                          //}

    // for NoThanViewInBase
    if (view_ instanceof NoThanViewInBase) {
        if (((NoThanViewInBase)view_).getEorFirstAccessTarget() != null)
            addInstancesForNode(((NoThanViewInBase)view_).getEorFirstAccessTarget());

                          // for testing and tracking
                          //System.out.println("and its BaseObjects in the checking perimeter are: ");

        for (Object obj : ((NoThanViewInBase)view_).getCheckingPerimeter()) {

                        //System.out.println("--- "+((BaseObject)obj).getName());

            if (!l_oldNodeNames.contains( ((BaseObject)obj).getName() )) {
                addInstancesForNode((BaseObject)obj);
            }
            else {
                        // for testing and tracking
                        //System.out.println("------ "+
                          //      " NO call to addInstancesForNode() for "+
                            //    ((BaseObject)obj).getName());
            } // else
        } // for obj
    } // if (view_ instanceof NoThanViewInBase)
    // for EPRViewInBase
    else {
                          // for testing and tracking
                          //System.out.println("and its BaseObjects are: ");

        for (Object obj : view_.getEorM_BaseObjects().entrySet()) {
            Map.Entry entry = (Map.Entry)obj;

                        //System.out.println("--- "+((BaseObject)entry.getValue()).getName());

            if (!l_oldNodeNames.contains((ImmutableName)entry.getKey())) {
                addInstancesForNode((BaseObject)entry.getValue());
            }
            else {
                        // for testing and tracking
                        //System.out.println("------ "+
                          //      " NO call to addInstancesForNode() for "+
                            //    ((BaseObject)entry.getValue()).getName());
            } // else
        } // for obj
    } // else if (view_ instanceof NoThanViewInBase)

    if (_clearIntermediateNodes) {
        for (Object obj : l_oldNodeNames) {
            if (!(view_ instanceof NoThanViewInBase)) {
                if (!view_.getEorM_BaseObjects().keySet().contains(obj)) {
                    removeInstancesForNode((ImmutableName)obj);
                } // if
            }
            else {
                if (!view_.getEorM_BaseObjects().keySet().contains(obj) &&
                        obj != ((NoThanViewInBase)view_).getEorFirstAccessTarget()) {
                    removeInstancesForNode((ImmutableName)obj);
                } // if
            } // else
        } // for obj
    } // if


          // for testing purpose
          //System.out.println("end of RightsMediatorImpl.updateNonCoreAlgorithmsForViewNodes()");

    l_oldNodeNames = null;
  } // updateNonCoreAlgorithmsForViewNodes()

  /**
   * Creates a new NodeRights and a new LinkRights instances
   * for a view node, an analyzed or an intermediate node.
   * The term of 'analyzed' is here applicable to a node
   * that is processed by the gWork classes, even if it is not finally detected as
   * an intermediate node to put in a stable DiaplayableLinkImpl.<P>
   * The detection of the AcsAddon packages follows the name pattern of these packages,
   * and does not use the method ACS.getM_AcsVocabulary() to get the AcsAddon name.
   * If there is no NodeRights or LinkRights implementation in the AcsAddon gWork
   * package, the generic class is selected.
   * Updates the maps m_LinkRights_ and m_NodeRights_. <P>
   * Called by updateNonCoreAlgorithmsForViewNodes() and by several methods of
   * ThreeNodesRightsFactoryImpl and CompoundRightsFactoryImpl, each time
   * the instances do not exist for a new processed object.
   * The Gui2.GraphicViewBaseListenerImpl method propertyChange() calls
   * directly this method when it receives a 'FirstAccessTarget' event
   * from a NoThanViewInBase.
   * @param _node node of the view (including the first target of a NoThanViewInBase)
   * or intermediate node. Never null. It may be already known.
   * @return a NodeRights instance
   */
  public NodeRights addInstancesForNode(BaseObject _node) {

      if (m_NodeRights_.get(_node.getName()) != null) {

                    // for testing and tracking
                    //System.out.println("RighsMediatorImpl.addInstancesForNode()"+
                      //      " start for a known node: "+_node.getName());

          return m_NodeRights_.get(_node.getName());
      }

      LinkRights     link = null;
      NodeRights     result = null;
      String         className = null;
      String         packageName = null;
      String         addonName = null;
      String         name = null;
      int            index = -1;

      if (!(_node instanceof ACSObject)) {
          link = new LinkRightsImpl(this);
          result = new NodeRightsImpl(this, link);
      }
      else if (((ACSObject)_node).getEorACS().getClass() == ACSImpl.class) {
          link = new LinkRightsImpl(this);
          result = new NodeRightsImpl(this, link);
      }
      else if (((ACSObject)_node).getEorACS().getClass().
              toString().indexOf("ARoad0.AcsAddon.") != -1) {

          // to remove 'class ' at the start
          className = ((ACSObject)_node).getEorACS().getClass().toString().substring(6);
          index = className.indexOf(".gBase.");
          // the 2 classes may be in the gWork package of the AcsAddon
          name = className.substring(0, index);
          packageName =  name + ".gWork";
          addonName = name.substring(
                  name.substring(0, name.lastIndexOf(".")).length() + 1);
          try {
              link = (LinkRightsImpl) Class.forName(packageName+".LinkRights"+addonName+"Impl").
                      getConstructor(new Class[] {AlgorithmInterpreter.class}).
                      newInstance(new Object[] {this});
          }
          catch (Exception e) {

                    // for testing and tracking
                    System.out.println("RighsMediatorImpl.addInstancesForNode()"+
                            "for the node: "+_node.getName()+
                            ", exception from the LinkRights AcsAddon constructor: "+
                            e.toString());

              link = new LinkRightsImpl(this);
          }

          try {
              result = (NodeRightsImpl) Class.forName(packageName+".NodeRights"+addonName+"Impl").
                      getConstructor(new Class[] {AlgorithmInterpreter.class, LinkRightsImpl.class}).
                      newInstance(new Object[] {this, link});
          }
          catch (Exception e) {
              result = new NodeRightsImpl(this, link);
          }
      } // else
      else
          throw new InternalError("RighsMediatorImpl.addInstancesForNode() "+
                      " with error for "+_node.getName()+", with the ACS "+
                      ((ACSObject)_node).getEorACS().getClass().toString());

      m_LinkRights_.put(
                    _node.getName(),
                    link);
      m_NodeRights_.put(
                    _node.getName(),
                    result);

                            // for testing and tracking
                            //System.out.println("end of RighsMediatorImpl.addInstancesForNode()"+
                              //      "for the node: "+_node.getName()+
                                //    " with the NodeRights class: "+
                                  //  result.getClass()+" and the LinkRights class: "+
                                    //link.getClass());

      link = null;
      className = null;
      packageName = null;
      addonName = null;
      name = null;
      return result;
  } // addInstancesForNode()

  /**
   * Removes the NodeRights and the LinkRights instances
   * for a view node, an analyzed or intermediate node.
   * The term of 'analyzed' is here applicable to a node
   * that is analyzed by the gWork classes, even if it is not finally detected as
   * an intermediate node to put in a stable DiaplayableLinkImpl. <P>
   * Updates the maps m_LinkRights_ and m_NodeRights_.
   * Finalizes the NodeRightsImpl instance.
   * Called by updateNonCoreAlgorithmsForViewNodes() only for the intermediate nodes
   * of a sketch view.
   * @param _name name of the view node, the analyzed or intermediate node
   */
  public void removeInstancesForNode(ImmutableName _name) {

                    // for testing and tracking
                    //System.out.println("Call to RighsMediatorImpl."+
                      //      "removeInstancesForNode() for "+_name);

      ((NodeRights)m_NodeRights_.get(_name)).finalizeForProcess();
      m_LinkRights_.remove(_name);
      m_NodeRights_.remove( _name);
  } // removeInstancesForNode()

  /**
  * Gets the collection of all the LinkRights associated to the nodes in the DisplayableLinkImpls
  * of a collection. When several nodes are from the same ACS package
  * (the generic gBase or an AcsAddon package), this method returns only one LinkRights
  * instance. Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
  * @param _l_dLinks DisplayableLinkImpls to analyze, all with the same two ends or not.
  * Not null.
  * @return the LinkRightsImpl instances. Never null.
  **/
  public final Collection<LinkRights> getAllLinkRights(
          Collection<DisplayableLinkImpl> _l_dLinks) {

      Set<LinkRights>   l_result = new HashSet();
      BaseObject[]      l_nodes = null;
      LinkRights        link = null;

      for (Object obj : _l_dLinks) {
          l_nodes = ((DisplayableLinkImpl)obj).getEorAllNodes();
          for (int i = 0; i < l_nodes.length; i++ ) {

              link = m_LinkRights_.get(l_nodes[i].getName());

              if (link == null)
                  throw new InternalError("RightsMediatorImpl.getAllLinkRights() with"+
                          " null linkRights for the node "+l_nodes[i].getName()+
                          " in the dLink "+obj);

              l_result.add(link);

          } // for i
      } // for obj
      
                            // for testing purpose
                          //System.out.println("RightsMediatorImpl.getAllLinkRights(), "+
                            //      "for the view: "+view_.getName()+", the result is:");

                          //for (Object obj : l_result) {
                            //System.out.println("---------- "+obj.toString());
                          //} // for ite

      l_nodes = null;
      return l_result;
  } // getAllLinkRights()

  /**
   * Registers a listener for the PropertyChange event
   * The class will fire a PropertyChange value whenever the rights
   * have to be updated.
   * Used by ActionSave.save().
   * @param _l  An object to be invoked when a PropertyChange
   *		event is fired.
   */
  public synchronized void addPropertyChangeListener(
    PropertyChangeListener _l) {

    synchronized (l_RightsChangeSupport_) {
      l_RightsChangeSupport_.add(_l);}
  }

  /**
   * Removes a listener for the PropertyChange event
   * Called by ActionSave.save().
   * @param _l the PropertyChange listener to be removed.
   */
  public synchronized void removePropertyChangeListener(
    PropertyChangeListener _l) {

    synchronized (l_RightsChangeSupport_) {
      l_RightsChangeSupport_.remove(_l);}
  }

  /**
   * When the view nodes network has changed, registers the instance,
   * as a RightsNetworkObserver to listen the bound property changes in
   * the network of ACSObjects and properties which are implied in the view.
   * The view already ensures the following of the property changes in its view elements. <P>
   * This method registers for the listening of the intermediate nodes for each access path,
   * and also for the analyzed, non-intermediate nodes which should become
   * intermediates in the future, after a property change. <P>
   * This listening includes the AclEntries and the PrivilegeForLinks,
   * respectively for their rights/conditional groups and their sources/targets.
   * It includes the listening of the specific events from any AcsAddon ACS,
   * or any of its objects, for which the methods
   * BaseUtilityImpl.add/removeAddonInner/ParentPropertyChangeListener() are called.
   * This implies to follow, for each view-related object: <P>
   *  - every ACS bridge change, <P>
   *  - every bound property change, including the AclEntry rights, <P>
   *  - every deleted BaseObject which is in an AccessControlLink, <P>
   *  - every new ACS userID with 'root' rights, <P>
   *  - every new ACS actor belonging to its userID, its groupID, <P>
   *    or to a groupID which contains its userID, its groupID, <P>
   *  - every new ACS userID belonging to its groupID, <P>
   *    or to a groupID which contains its userID, its groupID, <P>
   *  - every new ACS actor executed through 'other' rights with a current account
   *    or group, <P>
   *  - every PrivilegeForLinks change in the sources and the targets, <P>
   *  - every specific property change from an AcsAddon object. <P>
   * The previous listened objects that are not implied in the view
   * are unregistered. This is the most complex method of the class, and it has to
   * follow the evolutions of the properties in the generic gBase ACSObject classes.
   * Calls AclEntryImpl.getIntermediateGroupsToConditionalGroups().
   * Called by Gui2.ViewModel.
   * @param _l_baseObjects the BaseObjects which are intermediate nodes in
   *          the detected access paths of the view, or an intermediate-node AclEntry
   */
  public synchronized void setRightsNetwork(Set _l_baseObjects) {
    Object      obj = null;
    Set         l_old = null;
    Resource    res = null;
    AclEntry[]  l_acl = null;
    EligibleParty   ep = null;
    Resource[]  l_res = null;
    VirtualFolder vf = null;
    ImmutableGroupID[] l_grps = null;
          
                // for testing purpose
              //if (l_RightsNetwork_ != null) {
                  //System.out.println("RightsMediatorImpl.setRightsNetwork(), "+
                          //"for the view: "+view_.getName()+", the old BaseObjects are:");

                  //for (Iterator ite = l_RightsNetwork_.iterator(); ite.hasNext(); ) {
                    //obj = ite.next();
                    //System.out.println("---------- "+obj.toString());
                  //} // for ite
              //} // if
              //else
                  //System.out.println("RightsMediatorImpl.setRightsNetwork(), "+
                          //"for the view: "+view_.getName()+", the old BaseObjects is null.");
          
                // for testing purpose
              //if (_l_baseObjects != null) {
                 //System.out.println("RightsMediatorImpl.setRightsNetwork(), "+
                          //"for the view: "+view_.getName()+", the new BaseObjects are:");

                  //for (Iterator ite = _l_baseObjects.iterator(); ite.hasNext(); ) {
                    //obj = ite.next();
                    //System.out.println("---------- "+obj.toString());
                  //} // for ite
              //} // if
              //else
                  //System.out.println("RightsMediatorImpl.setRightsNetwork(), "+
                          //"for the view: "+view_.getName()+", the new BaseObjects is null.");
    
        if (l_RightsNetwork_ != null) {

            for (Iterator ite = l_RightsNetwork_.iterator(); ite.hasNext(); ) {
                obj = ite.next();
                if (_l_baseObjects != null && !(obj instanceof ImmutableName) &&
                        _l_baseObjects.contains(obj))
                    continue;
                else if (_l_baseObjects != null && (obj instanceof ImmutableName) ) {
                    try {
                        if (_l_baseObjects.contains( ACS_Facade.getAclEntry((ImmutableName)obj) ))
                            continue;
                    } catch (BaseError b) {
                        // the acl has been deleted
                        continue;
                    }
                } // else if

                // for the 'Architecture' view only
                if (obj instanceof ACS) {
                  ((ACS)obj).removePropertyChangeListener("BridgeSources", this);
                  ((ACS)obj).removePropertyChangeListener("ExternalOriented", this);
                  continue;
                } // if

                if (obj instanceof BoundBean )
                    ((BoundBean)obj).removePropertyChangeListener(this);

                if (obj instanceof ACSObject) {

                            // for testing and tracking
                          //System.out.println("in RightsMediatorImpl.setRightsNetwork(), "+
                            //      "the view: "+view_.getName()+" contains the object "+
                              //    obj.toString());
                                    //+" with the ACS:"+((ACSObject)obj).getAcsName() );

                      ACS acs = ((ACS)ACS_Facade.getOpenACS().get( ((ACSObject)obj).getAcsName() ));
                      if (acs != null) {
                          if (acs.isPropertyChangeListener("OneEligibleParty", this))
                              acs.removePropertyChangeListener("OneEligibleParty", this);
                          if (acs.isPropertyChangeListener("BridgeSources", this))
                              acs.removePropertyChangeListener("BridgeSources", this);
                          if (acs.isPropertyChangeListener("OneActorFromOtherInCurrentContext", this))
                              acs.removePropertyChangeListener("OneActorFromOtherInCurrentContext", this);
                      } // if

                      acs = null;

                } // if ACSObject

                // to not follow the update of rights in the ACL
                if (obj instanceof AclEntryImpl) {
                    ((AclEntry)obj).removePropertyChangeListener("AclRights", this);

                      // For a conditional right, the ACL is deleted if a condition group is deleted.
                      // For a conditional right, registers the intermediate groups
                      l_grps = GroupIDImpl.
                              getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)obj);
                      if (l_grps != null) {
                          ((AclEntry)obj).removePropertyChangeListener("ConditionSourceGroups", this);

                          for (int i = 0; i < l_grps.length; i++) {
                            l_grps[i].removePropertyChangeListener("AssociatedGroups", this);
                          } // for i
                      } // if (l_grps
                } // for

                if (obj instanceof EligibleParty) {
                    ep = (EligibleParty) obj;
                    ep.removePropertyChangeListener("AclConnectedResources", this);
                    for (Iterator ita = ep.getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                      l_acl = ((Resource)ita.next()).getL_aclEntry();
                      if (l_acl == null || l_acl.length == 0) continue;

                      for (int i = 0; i < l_acl.length; i++) {
                        if (l_acl[i] != null && l_acl[i].getEorEP() == ep ) {
                              l_acl[i].removePropertyChangeListener("AclRights", this);

                              // for a conditional ACL, registers the intermediate groups
                              l_grps = GroupIDImpl.
                                      getL_IntermediateGroupsOfAclConditionalGroups(
                                            (AclEntryImpl)l_acl[i]);
                              if (l_grps != null) {
                                  l_acl[i].removePropertyChangeListener("ConditionSourceGroups", this);

                                  for (int j = 0; j < l_grps.length; j++) {
                                    l_grps[j].removePropertyChangeListener("AssociatedGroups", this);
                                  } // for i
                              } // if (l_grps
                        } // if
                      } // for
                    } // for ita
                } // if EligibleParty

                if (obj instanceof GroupIDMember) {
                    ep = (EligibleParty) obj;
                    ep.removePropertyChangeListener("AssociatedGroups", this);
                } // if GroupIDMember
                if (obj instanceof GroupID) {
                    ep = (EligibleParty) obj;
                    ep.removePropertyChangeListener("l_Members", this);
                    ep.removePropertyChangeListener("l_GroupIDCurrentActors", this);
                    ep.removePropertyChangeListener("l_GroupIDOwnResources", this);

                    l_res = ((GroupID)ep).getEorL_GroupIDOwnResources();
                    for (int i = 0; i < l_res.length; i++) {
                        l_res[i].removePropertyChangeListener("GroupRights", this);
                    } // for i
                } // if GroupID

                if (obj instanceof UserID) {
                    ep = (EligibleParty) obj;
                    ep.removePropertyChangeListener("l_UserIDCurrentActors", this);
                    ep.removePropertyChangeListener("l_UserIDOwnResources", this);

                    l_res = ((UserID)ep).getEorL_UserIDOwnResources();
                    for (int i = 0; i < l_res.length; i++) {
                        l_res[i].removePropertyChangeListener("OwnerRights", this);
                    } // for i
                } // if UserID

                if (obj instanceof Resource) {
                      res = (Resource)obj;
                      // to follow the ACLs
                      res.removePropertyChangeListener("AclRights", this);
                      if (res.getL_aclEntry() != null) {
                          l_acl = res.getL_aclEntry();
                          for (int i = 0; i < l_acl.length; i++) {
                              l_acl[i].removePropertyChangeListener("AclRights", this);

                              // for a conditional ACL, registers the intermediate groups
                              l_grps = GroupIDImpl.
                                      getL_IntermediateGroupsOfAclConditionalGroups(
                                            (AclEntryImpl)l_acl[i]);
                              if (l_grps != null) {
                                  l_acl[i].removePropertyChangeListener("ConditionSourceGroups", this);

                                  for (int j = 0; j < l_grps.length; j++) {
                                    l_grps[j].removePropertyChangeListener("AssociatedGroups", this);
                                  } // for i
                              } // if (l_grps
                          } // for i
                      } // if
                      // to follow the access accross the nodes tree of each resource
                      while (res.getEorParent() != null) {
                          res.removePropertyChangeListener("FromTheParentAccount", this);
                          res.removePropertyChangeListener("FromTheParentGroup", this);
                          res.removePropertyChangeListener("InheritedOwnerRights", this);
                          res.removePropertyChangeListener("InheritedGroupRights", this);
                          res.removePropertyChangeListener("InheritedTargetPrivileges", this);
                          res.removePropertyChangeListener("InheritedTargetAclEntries", this);

                          res = (Resource)res.getEorParent();
                      } // while
                } // if

                if (obj instanceof ImmutableActor) {
                  res = (Resource)obj;
                  res.removePropertyChangeListener("OwnerRights", this);
                  res.removePropertyChangeListener("GroupRights", this);
                  res.removePropertyChangeListener("RootRights", this);
                  res.removePropertyChangeListener("CurrentUserID", this);
                  res.removePropertyChangeListener("CurrentGroupID", this);
                  res.removePropertyChangeListener("SecondaryGroups", this);
                  res.removePropertyChangeListener("IsRightProxy", this);
                  res.removePropertyChangeListener("ACSBridgeTarget", this);
                  res.removePropertyChangeListener("ActorAlias", this);
                } // if

                if (obj instanceof VirtualFolder) {
                  vf = (VirtualFolder)obj;
                  vf.removePropertyChangeListener("OneBaseObjectInFolder", this);
                  vf.removePropertyChangeListener("OneTargetTypedPrivilege", this);
                  vf.removePropertyChangeListener("OneTargetLinkedPrivilege", this);
                  vf.removePropertyChangeListener("Children", this);
                  vf.removePropertyChangeListener("InheritedTargetPrivileges", this);

                  // to follow the access accross the nodes tree of each virtual folder
                  while (vf.getEorParent() != null) {
                      vf.removePropertyChangeListener("InheritedTargetPrivileges", this);

                      vf = (VirtualFolder)vf.getEorParent();
                  } // while
                } // if VirtualFolder

                // to follow the targets of the linked and typed privileges
                if (obj instanceof PrivilegeSource) {
                  ((PrivilegeSource)obj).removePropertyChangeListener("OneSourceLinkedPrivilege", this);
                  ((PrivilegeSource)obj).removePropertyChangeListener("OneSourceTypedPrivilege", this);

                  for (Object ob : ((PrivilegeSource)obj).getEorL_SourcePrivilegeForLinks()) {
                    ((PrivilegeForLinks)ob).removePropertyChangeListener("OneLinkedPrivilegeTarget", this);
                  } // for ob
                  for (Object ob : ((PrivilegeSource)obj).getEorL_SourcePrivilegeForTypes()) {
                    ((PrivilegeForType)ob).addPropertyChangeListener("OneTypedPrivilegeTarget", this);
                  } // for ob
                } // if (obj instanceof PrivilegeSource)


                  // Gets the specific events to unlisten from an ACS Addon ACSObject,
                  // because these events should change the rights of an ACSObject in the ACS Addon
                if (obj instanceof ACSObject &&
                          ((ACSObject)obj).getEorACS() instanceof AcsAddon) {
                      if (obj instanceof ACSObjectAddon)
                          ((UtilityImpl)m_CoreAlgorithms_.get("BaseUtil")).
                                  removeAddonInnerPropertyChangeListener(
                                  this, (ACSObjectAddon)obj);

                      if (obj instanceof Resource || obj instanceof VirtualFolder ) {
                          ((UtilityImpl)m_CoreAlgorithms_.get("BaseUtil")).
                                  removeAddonParentPropertyChangeListener(
                                  this, (ACSObject)obj);
                      } // if (obj instanceof Resource || obj instanceof VirtualFolder )
                } // if ((ACSObject)obj).getEorACS() instanceof AcsAddon
            } // ite

        } // if (l_RightsNetwork_ != null)

        l_old = l_RightsNetwork_;

        if (_l_baseObjects == null) {
          if (l_RightsNetwork_ != null) l_RightsNetwork_.clear();
          l_RightsNetwork_ = null;
          l_old = null;
          return;
        }

        if (l_RightsNetwork_ != null)
            l_RightsNetwork_.clear();
        else
            l_RightsNetwork_ = new HashSet(INITIAL_CAPACITY*10);

        for (Iterator ite = _l_baseObjects.iterator(); ite.hasNext(); ) {
            obj = ite.next();
            // no registering of the AclEntry instance to handle properly their switch
            // in an ACS saving
            if (obj instanceof AclEntry) l_RightsNetwork_.add(((AclEntry)obj).getName());
            else l_RightsNetwork_.add(obj);
        } // for i

        for (Iterator ite = _l_baseObjects.iterator(); ite.hasNext(); ) {
            obj = ite.next();
            if (l_old != null && !(obj instanceof AclEntry) && l_old.contains(obj))
                continue;
            else if (l_old != null && (obj instanceof AclEntry) &&
                    l_old.contains(  ((AclEntry)obj).getName() )) continue;

            // for the 'Architecture' view only
            if (obj instanceof ACS) {
              ((ACS)obj).addPropertyChangeListener("BridgeSources", this);
              ((ACS)obj).addPropertyChangeListener("ExternalOriented", this);

                    // for testing and tracking
                  //System.out.println("in RightsMediatorImpl.setRightsNetwork(), "+
                    //      "for the view: "+view_.getName()+", register as a change listener for the ACS:"+
                    //      ((ACS)obj).getName());

              //((ACS)obj).addPropertyChangeListener("BridgeTargets", this);
              continue;
            } // if

            if (obj instanceof BaseObject && obj instanceof BoundBean) {
              if (!(obj instanceof ACSObject))
                  ((BoundBean)obj).addPropertyChangeListener(this);
            } else if (!(obj instanceof BaseObject) && !(obj instanceof BoundBean)) {

                                    // for testing and tracking
                                //System.out.println("RightsMediatorImpl.setRightsNetwork(), "+
                                  //      "a new BaseObject is not a BaseObject or a BoundBean: "+obj+"; class: "+obj.getClass());

                throw new InternalError("in RightsMediatorImpl"+
                    ".setRightsNetwork(), a new Object is not a BaseObject"+
                    " nor a BoundBean: "+obj);
            } // else if

            if (obj instanceof ACSObject) {
              ACS acs = ((ACS)ACS_Facade.getOpenACS().get( ((ACSObject)obj).getAcsName() ));

                                    // for testing and tracking
                                //System.out.println("RightsMediatorImpl.setRightsNetwork(), "+
                                  //"for the view: "+view_.getName()+", register as a change listener for the ACS: "+
                                  //acs);
                                //System.out.println("RightsMediatorImpl.setRightsNetwork(), "+
                                  //"for the ACSObject: "+obj);


              if (!acs.isPropertyChangeListener("OneEligibleParty", this))
                  acs.addPropertyChangeListener("OneEligibleParty", this);
              if (!acs.isPropertyChangeListener("BridgeSources", this))
                  acs.addPropertyChangeListener("BridgeSources", this);
              if (!acs.isPropertyChangeListener("OneActorFromOtherInCurrentContext", this))
                  acs.addPropertyChangeListener("OneActorFromOtherInCurrentContext", this);

              // to follow the deletion of an AclEntry, and the update of rights in the ACL
              if (obj instanceof AclEntryImpl) {
                ((AclEntry)obj).getEorResource().addPropertyChangeListener("AclRights", this);
                ((AclEntry)obj).addPropertyChangeListener("AclRights", this);
                
                  // For a conditional right, the ACL is deleted if a condition group is deleted.
                  // For a conditional ACL, registers the intermediate groups
                  l_grps = GroupIDImpl.
                          getL_IntermediateGroupsOfAclConditionalGroups((AclEntryImpl)obj);
                  if (l_grps != null) {
                      ((AclEntry)obj).addPropertyChangeListener("ConditionSourceGroups", this);

                      for (int i = 0; i < l_grps.length; i++) {
                        l_grps[i].addPropertyChangeListener("AssociatedGroups", this);
                      } // for i
                  } // if (l_grps
              } // for

              if (obj instanceof EligibleParty) {
                    ep = (EligibleParty) obj;
                    ep.addPropertyChangeListener("AclConnectedResources", this);
                    for (Iterator ita = ep.getEorL_AclConnectedResources().iterator(); ita.hasNext(); ) {
                      l_acl = ((Resource)ita.next()).getL_aclEntry();
                      if (l_acl == null || l_acl.length == 0) continue;

                      for (int i = 0; i < l_acl.length; i++) {
                        if (l_acl[i] != null && l_acl[i].getEorEP() == ep ) {
                              l_acl[i].addPropertyChangeListener("AclRights", this);

                              // for a conditional ACL, registers the intermediate groups
                              l_grps = GroupIDImpl.
                                      getL_IntermediateGroupsOfAclConditionalGroups(
                                            (AclEntryImpl)l_acl[i]);
                              if (l_grps != null) {
                                  l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                                  for (int j = 0; j < l_grps.length; j++) {
                                    l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                                  } // for i
                              } // if (l_grps
                        } // if
                      } // for
                    } // for ita
              } // if EligibleParty l_Members l_GroupIDCurrentActors

              if (obj instanceof GroupIDMember) {
                    ep = (EligibleParty) obj;
                    ep.addPropertyChangeListener("AssociatedGroups", this);
              } // if GroupIDMember
              if (obj instanceof GroupID) {
                    ep = (EligibleParty) obj;
                    ep.addPropertyChangeListener("l_Members", this);
                    ep.addPropertyChangeListener("l_GroupIDCurrentActors", this);
                    ep.addPropertyChangeListener("l_GroupIDOwnResources", this);

                    l_res = ((GroupID)ep).getEorL_GroupIDOwnResources();
                    for (int i = 0; i < l_res.length; i++) {
                        l_res[i].addPropertyChangeListener("GroupRights", this);
                    } // for i
              } // if GroupID
              if (obj instanceof UserID) {
                    ep = (EligibleParty) obj;
                    ep.addPropertyChangeListener("l_UserIDCurrentActors", this);
                    ep.addPropertyChangeListener("l_UserIDOwnResources", this);

                    l_res = ((UserID)ep).getEorL_UserIDOwnResources();
                    for (int i = 0; i < l_res.length; i++) {
                        l_res[i].addPropertyChangeListener("OwnerRights", this);
                    } // for i
              } // if GroupIDMember

              if (obj instanceof Resource) {
                  res = (Resource)obj;
                  // to follow the ACLs
                  res.addPropertyChangeListener("AclRights", this);
                  if (res.getL_aclEntry() != null) {
                      l_acl = res.getL_aclEntry();
                      for (int i = 0; i < l_acl.length; i++) {
                          l_acl[i].addPropertyChangeListener("AclRights", this);

                          // for a conditional ACL, registers the intermediate groups
                          l_grps = GroupIDImpl.
                                  getL_IntermediateGroupsOfAclConditionalGroups(
                                        (AclEntryImpl)l_acl[i]);
                          if (l_grps != null) {
                              l_acl[i].addPropertyChangeListener("ConditionSourceGroups", this);

                              for (int j = 0; j < l_grps.length; j++) {
                                l_grps[j].addPropertyChangeListener("AssociatedGroups", this);
                              } // for i
                          } // if (l_grps

                      } // for i
                  } // if
                  // to follow the access accross the nodes tree of each resource
                  while (res.getEorParent() != null) {
                      res.addPropertyChangeListener("FromTheParentAccount", this);
                      res.addPropertyChangeListener("FromTheParentGroup", this);
                      res.addPropertyChangeListener("InheritedOwnerRights", this);
                      res.addPropertyChangeListener("InheritedGroupRights", this);
                      res.addPropertyChangeListener("InheritedTargetPrivileges", this);
                      res.addPropertyChangeListener("InheritedTargetAclEntries", this);

                      res = (Resource)res.getEorParent();
                  } // while
              } // if

              // if obj is an actor, it may be implied,
              // for instance because it may became a bridge source
              if (obj instanceof ImmutableActor) {
                  res = (Resource)obj;
                  res.addPropertyChangeListener("OwnerRights", this);
                  res.addPropertyChangeListener("GroupRights", this);
                  res.addPropertyChangeListener("RootRights", this);
                  res.addPropertyChangeListener("CurrentUserID", this);
                  res.addPropertyChangeListener("CurrentGroupID", this);
                  res.addPropertyChangeListener("SecondaryGroups", this);
                  res.addPropertyChangeListener("IsRightProxy", this);
                  res.addPropertyChangeListener("ACSBridgeTarget", this);
                  res.addPropertyChangeListener("ActorAlias", this);
              } // if

              if (obj instanceof VirtualFolder) {
                  vf = (VirtualFolder)obj;
                  vf.addPropertyChangeListener("OneBaseObjectInFolder", this);
                  vf.addPropertyChangeListener("OneTargetTypedPrivilege", this);
                  vf.addPropertyChangeListener("OneTargetLinkedPrivilege", this);
                  vf.addPropertyChangeListener("Children", this);

                  // to follow the access accross the nodes tree of each virtual folder
                  while (vf.getEorParent() != null) {
                      vf.addPropertyChangeListener("InheritedTargetPrivileges", this);

                      vf = (VirtualFolder)vf.getEorParent();
                  } // while
              } // if VirtualFolder

              // to follow the targets of the linked and typed privileges
              if (obj instanceof PrivilegeSource) {
                ((PrivilegeSource)obj).addPropertyChangeListener("OneSourceLinkedPrivilege", this);
                ((PrivilegeSource)obj).addPropertyChangeListener("OneSourceTypedPrivilege", this);

                for (Object ob : ((PrivilegeSource)obj).getEorL_SourcePrivilegeForLinks()) {
                  ((PrivilegeForLinks)ob).addPropertyChangeListener("OneLinkedPrivilegeTarget", this);
                } // for ob
                for (Object ob : ((PrivilegeSource)obj).getEorL_SourcePrivilegeForTypes()) {
                  ((PrivilegeForType)ob).addPropertyChangeListener("OneTypedPrivilegeTarget", this);
                } // for ob
              } // if (obj instanceof PrivilegeSource)

              // Gets the specific events to listen from an ACS Addon ACSObject,
              // because these events should change the rights of an ACSObject in the ACS Addon
              if (obj instanceof ACSObject &&
                      ((ACSObject)obj).getEorACS() instanceof AcsAddon) {
                  if (obj instanceof ACSObjectAddon)
                      ((UtilityImpl)m_CoreAlgorithms_.get("BaseUtil")).
                              addAddonInnerPropertyChangeListener(
                              this, (ACSObjectAddon)obj);

                  if (obj instanceof Resource || obj instanceof VirtualFolder ) {
                      ((UtilityImpl)m_CoreAlgorithms_.get("BaseUtil")).
                              addAddonParentPropertyChangeListener(
                              this, (ACSObject)obj);
                  } // if (obj instanceof Resource || obj instanceof VirtualFolder )
              } // if ((ACSObject)obj).getEorACS() instanceof AcsAddon

                    // for testing and tracking
                  //System.out.println("in RightsMediatorImpl.setRightsNetwork(), "+
                    //      "for the view: "+view_.getName()+", register as a change listener for the ACS:"+
                    //      acs.getName()+" and all its EP for the 'order' property.");

              acs = null;
              //ep = null;
            } // if ACSObject

        } // for ite

    obj = null;
    l_old = null;
    res = null;
    l_acl = null;
    ep = null;
    l_res = null;
    vf = null;
    l_grps = null;
  } // setRightsNetwork()

  /**
   * Manages the events fired by the base objects of the rights network, to
   * update the rights while there is no updating of the view base objects.
   * Most of the events are simply fired with a null property to the listeners,
   * for updating the graphic view.
   * Filters the events about the new UserID having the order 0 (as super-administrator),
   * from the ACS and the UserIDs. Does not fire the 'ImpliedOpenViews' events.
   * Does not fire a 'InheritedXXXRights' event after a 'FromTheParentXXX' event
   * from the same source. Does not fire a 'InheritedTargetPrivileges' or
   * a 'InheritedTargetAclEntries' event if the last event is from the parent. <P>
   * To avoid the processing of events from a closing ACS during the closing of
   * its child ACS, this method uses TreeManager.haveAtLeastOneTreeBaseListener()
   * to see if ActionClose.closeInGUI() has already closed the parent ACS tree
   * in the explorer.
   * @param _evt is the event from the gBase objects
   */
  public void propertyChange(PropertyChangeEvent _evt) {


                            // for testing and tracking
                        //System.out.println("RightsMediatorImpl.propertyChange()"+
                          //      " for the view "+view_.getName()+
                            //    " for the base property: "+_evt.getPropertyName()+
                              //  " from the source: "+_evt.getSource()+
                                //" and the count "+count_+
                                  //    "for the thread "+Thread.currentThread().getName()+
                                    //  " with the priority: "+Thread.currentThread().getPriority());
                        //count_++;
     
    if (view_ == null) {
        
                            // for testing and tracking
                        //System.out.println("RightsMediatorImpl.propertyChange()"+
                          //      " with a finalized instance for the event "+_evt.getPropertyName()+
                            //    " and from the source "+_evt.getSource());
        
        return;
    }
    if (_evt.getSource() instanceof ACS &&
            !TreeManager.haveAtLeastOneTreeBaseListener(((BaseObject)_evt.getSource()).getName())) {

                            // for testing and tracking
                        //System.out.println("RightsMediatorImpl.propertyChange()"+//view_.getName()+
                          //      " for the view "+view_+
                            //    " for the base property: "+_evt.getPropertyName()+
                              //  ", no action since no ACS. ");


        // ActionClose.closeInGUI() has closed the ACS source
        return;
    }
      
    if (lastEvent_ != null && lastEvent_.getSource() == _evt.getSource() ) {
        if (lastEvent_.getPropertyName().equals("FromTheParentAccount") &&
                _evt.getPropertyName().equals("InheritedOwnerRights")) return; 
        else if (lastEvent_.getPropertyName().equals("FromTheParentGroup") &&
                _evt.getPropertyName().equals("InheritedGroupRights")) return; 
    } // if
    
    if (lastEvent_ != null && _evt.getSource() instanceof PrivilegeTarget &&
            lastEvent_.getSource() == ((PrivilegeTarget)_evt.getSource()).getEorParent() ) {
        if (lastEvent_.getPropertyName().equals("InheritedTargetPrivileges") &&
                _evt.getPropertyName().equals("InheritedTargetPrivileges")) return;
    } // if
    if (lastEvent_ != null && _evt.getSource() instanceof Resource &&
            lastEvent_.getSource() == ((Resource)_evt.getSource()).getEorParent() ) {
        if (lastEvent_.getPropertyName().equals("InheritedTargetAclEntries") &&
                _evt.getPropertyName().equals("InheritedTargetAclEntries")) return;
    } // if
    
    lastEvent_ = _evt;
      
      // does not observe some changes in the ACS, the ACS objects
    if (_evt.getPropertyName().equals("ImpliedOpenViews") ||    // from CFolderMemberImpl in Basic
            _evt.getPropertyName().equals("OneBaseObject") ||   // from ACS
            _evt.getPropertyName().equals("OneResource") ||     // from ACS
            _evt.getPropertyName().equals("OneTypedPrivilege") ||   // from ACS
            _evt.getPropertyName().equals("OneLinkedPrivilege") ||  // from ACS
            _evt.getPropertyName().equals("OneResourceVirtualFolder") ||    // from ACS
            _evt.getPropertyName().equals("OneEPVirtualFolder")             // from ACS
            ) {

                            // for testing and tracking
                        //System.out.println("RightsMediatorImpl.propertyChange()"+//view_.getName()+
                          //      " for the view "+view_+
                            //    " for the base property: "+_evt.getPropertyName()+
                              //  ", no action for this event. ");
        return;
    } // if

      // fires most of the events
    else if (!_evt.getPropertyName().equals("OneEligibleParty") &&
      !_evt.getPropertyName().equals("Order") && !(_evt.getSource() instanceof ACS))
        firePropertyChange();
    
    // forgets all the ACS events, excepted for the UserID order at 0
    else if (_evt.getPropertyName().equals("OneEligibleParty") ) { 
      // detect the new instances of UserID with 0 as order value
      if (!(_evt.getSource() instanceof ACS) || _evt.getNewValue() == null) return;
      if (!(_evt.getNewValue() instanceof UserID)) return;
      if ( ((UserID)_evt.getNewValue()).getOrder() != 0) return;
      else firePropertyChange();
    }
    
    // filters the UserID 'Order' event
    else {
      // detects on an UserID the order changes to 0
      if (!(_evt.getSource() instanceof UserID) || _evt.getNewValue() == null) return;
      if ( ((Integer)_evt.getNewValue()).intValue() != 0) return;
      else firePropertyChange();
    } // else
        
  } // propertyChange()

  /**
   * Finalizes the instance.
   * Called by Gui2.ViewModel.removeViewRightsMediator().
   */
  public void finalizeForProcess() {

                // for testing and tracking
                //System.out.println("call to RightsMediatorImpl.finalizeForProcess() for the view"+
                  //      view_.getName());
                //BaseError p = new BaseError();
                //p.printStackTrace();

      // clear l_RightsNetwork_
    setRightsNetwork(null);
    view_ = null;
    l_RightsChangeSupport_.clear();

    ((RightsFactory_Facade)m_CoreAlgorithms_.get("Facade")).finalizeForProcess();
    ((AclRightsFactoryImpl)m_CoreAlgorithms_.get("Acl")).finalizeForProcess();
    ((AgoRightsFactoryImpl)m_CoreAlgorithms_.get("Ago")).finalizeForProcess();
    ((PrivilegeRightsFactoryImpl)m_CoreAlgorithms_.get("Pri")).finalizeForProcess();
    ((CompoundRightsFactoryImpl)m_CoreAlgorithms_.get("Compound")).finalizeForProcess();
    ((ThreeNodesRightsFactoryImpl)m_CoreAlgorithms_.get("3nodes")).finalizeForProcess();
    ((DisplayableLinkUtilities)m_CoreAlgorithms_.get("LinkUtil")).finalizeForProcess();
    ((FactoryUtilities)m_CoreAlgorithms_.get("FactoryUtil")).finalizeForProcess();
    ((UtilityImpl)m_CoreAlgorithms_.get("BaseUtil")).finalizeForProcess();

    m_CoreAlgorithms_.clear();

    for (Object obj : m_NodeRights_.entrySet()) {
        Map.Entry entry = (Map.Entry)obj;
        ((NodeRights)entry.getValue()).finalizeForProcess();
    }

    for (Object obj : m_LinkRights_.entrySet()) {
        Map.Entry entry = (Map.Entry)obj;
        ((LinkRights)entry.getValue()).finalizeForProcess();
    }

    m_NodeRights_.clear();
    m_LinkRights_.clear();
  } // finalizeForProcess()


 /**
 * HashCode of the instance. Use the view hashCode.
 * @return int
 */
  /*public int hashCode(){
        int result = 0;

        if (view_ != null) result = PRIME*result + view_.hashCode();

        return result;
  } // hashCode()
  */
   
  /**
   * no equals: each instance is equals only to itself.
   * @param _obj: object to compare
   * @return true if the two objects have the same views
   */
   /*public synchronized boolean equals(Object _obj) {
    if (this == _obj)
        return true;
    if (_obj == null)
        return false;
    if (_obj.getClass() != getClass())
        return false;

    RightsMediatorImpl obj = (RightsMediatorImpl) _obj;

    if (view_ == null || obj.getSource() == null) {
        if (view_!= null || obj.getSource() != null)
          return false;  // if
    } // if
    else {
      if (!view_.equals( obj.getSource() )) return false; }

    return true;
  } //equals()
  */
  
  /**
   * Get the values of the view.
   * @return view.toString() with the name of the class
   */
  public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("  RIGHTS_MEDIATORIMPL.TOSTRING():");

        sb.append(" view_ = ");
        sb.append(view_);

        return " [" + sb + "] ";
  } // toString()


  // -------------------- PRIVATE METHODS -----------------------------------

  /**
  * Fires an event without name to every registered listener, in any order, with
  * null old and null new properties.
  * The GraphicViewBaseListenerImpl instance of this ViewInBase is the single listener.
  * Called only by propertyChange().
  */
  private void firePropertyChange() {

      if (l_RightsChangeSupport_.isEmpty()) return;
      
      PropertyChangeEvent evt = new PropertyChangeEvent(this,
                                          null, null, null);

      PropertyChangeListener listener;
      
      //if (_propertyName.equals("Rights or No-Rights"))
      Set toFire = (Set) ((HashSet)l_RightsChangeSupport_).clone();

      for (Iterator ite = toFire.iterator() ; ite.hasNext() ; ) {
        listener = (PropertyChangeListener)ite.next();
        if (listener != null) listener.propertyChange(evt);

                    // for testing and tracking
                    //System.out.println("in RightsMediatorImpl, call to "+listener+
                      //".propertyChange()");

      } // for
  } // firePropertyChange()


} // RightsMediatorImpl
