//Title:        Access Road
//Version:      0.7.2
//Copyright:    2011-2012 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Iterator;
import java.util.Arrays;
import java.util.ArrayList;

import ARoad0.Gui2.GraphicView;
import ARoad0.ErrW.ProcessError;
import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.CNot.AccessControlLink;
import ARoad0.CNot.AccessControlLinkImpl;

import ARoad0.gBaseInterface.*;
import ARoad0.gWorkInterface.*;

import ARoad0.gBase.BaseManagerImpl;

/**
 * This class is responsible for providing the core algorithms to detect for a view
 * all the 3-nodes access paths, combining the searches among all types of rights and structures.
 * This class returns the 3-nodes access paths, that is the effective applicable paths
 * from an EligibleParty or a VirtualFolder to a Resource, GroupIDMember or VirtualFolder,
 * when there is exactly one intermediate node.
 * All the AG, ACL, BDG and PRI rights, and all the pseudo-structural links are possible.
 * The AcsAddons are handled to detect they proper rights and structures. <P>
 * The careful study of this class is a good start to understand the search
 * of access paths in Access Road. <P>
 * This class is stateless, having as variables only the instances of the core algorithm
 * classes of the ViewInBase. Each RightsMediatorImpl instance at its creation
 * setups an instance of this class for its ViewInBase.
 * There is no direct call to the ACS methods from the worker thread using this class. <P>
 * The calling sequence of the methods is presented hereinafter: <P>
 * - detectOneHiddenNodeCompoundRights() is the single entry point of the class
 * to analyze a given view; it is called by RightsFactory_Facade, <P>
 * - for each access source of the view, this entry method calls the 4 methods
 * selectOneHiddenNodePathsForXXXX() for respectively the sources which are
 * virtual folders, the privilege links in the first link, the sources which
 * are Actors, or those which are GroupIDMembers, <P>
 * - the 2 first selectOneHiddenNodePathsForXXXX() methods apply proper
 * algorithms for detecting the paths; they calls some of the 4 endsPathFromXXX() methods. <P>
 * - the 2 last selectOneHiddenNodePathsForXXXX() methods call 6 dedicated
 * addPathsFromXXX() methods, and they calls some of the 4 endsPathFromXXX() methods.<P>
 * This structure of methods facilitates the maintenance, even if there is redondancy
 * in the processing of the data. This choice is the opposite of that done for
 * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(), where the search performance
 * is better. Comparing the two approaches for a given issue is a good mean
 * to verify the code quality. Some differences are: <P>
 * - the methods NodeRights.selectAddonNewHiddenNodeForXXX() are called first in
 * CompoundRightsFactoryImpl.detectHiddenCompoundEpRights(), before the other
 * required methods to search new links on the current node. As it is for updating
 * the current access paths in the central search loop, this is not needed in
 * the ThreeNodesRightsFactoryImpl class. <P>
 * - the method DisplayableLinkUtilities.getContainVirtualFolderWithoutMember()
 * implements the rule 'virtual folder: an indirect access to a Resource that is not
  * a virtual member, is always through a virtual member', and it is called in the
 * central and final search loops of CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
 * As it is for analyzing paths of more than 3 nodes, this is not needed in
 * the ThreeNodesRightsFactoryImpl class. <P>
 * - the third final loop in CompoundRightsFactoryImpl.detectHiddenCompoundEpRights()
 * is about the filtering of all access paths to apply some priority rules among
 * the AGO, PRI, ACL, BDG rights. As it covers also the paths resulting from the
 * work of this ThreeNodesRightsFactoryImpl class, this is not needed there.
 * @see CompoundRightsFactoryImpl#detectHiddenCompoundEpRights(ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, java.util.Map) 
 * @version 0.7.2
 */
public final class ThreeNodesRightsFactoryImpl implements CoreAlgorithm {
  private AlgorithmInterpreter      interpreter_ = null;
  //private UtilityImpl               baseUtil_ = null;
  private DisplayableLinkUtilities  linkUtil_ = null;
  private CompoundRightsFactoryImpl compound_ = null;
  private AclRightsFactoryImpl      acl_ = null;
  private PrivilegeRightsFactoryImpl pri_ = null;
  private AgoRightsFactoryImpl      ago_ = null;
  private Set<PrivilegeTarget>      l_targets_ = new HashSet(INITIAL_CAPACITY);

  /**
  * only one protected constructor
  * @param _interpreter algorithm interpreter
  */
  protected ThreeNodesRightsFactoryImpl(AlgorithmInterpreter _interpreter) {
      interpreter_ = _interpreter;
      //baseUtil_ = (UtilityImpl)interpreter_.getCoreAlgorithm("BaseUtil");
      linkUtil_ = (DisplayableLinkUtilities)interpreter_.getCoreAlgorithm("LinkUtil");
      compound_ = (CompoundRightsFactoryImpl)interpreter_.getCoreAlgorithm("Compound");
      acl_ = (AclRightsFactoryImpl)interpreter_.getCoreAlgorithm("Acl");
      pri_ = (PrivilegeRightsFactoryImpl)interpreter_.getCoreAlgorithm("Pri");
      ago_ = (AgoRightsFactoryImpl)interpreter_.getCoreAlgorithm("Ago");
  }

  /**
  * This is the entry method of this class, since it defines all the access paths
  * containing exactly 3 nodes. They are the effective applicable paths in a view
  * from an EligibleParty or a VirtualFolder to a Resource, GroupIDMember or VirtualFolder,
  * when there is exactly one intermediate node.
  * All the AGO, ACL, BDG and PRI rights, and all the structural or pseudo-structural
  * links are analyzed. <P>
  * The ACS Addons, if any, are handled to detect their proper rights and structures.
  * There are 55 basic formats of such 3-nodes access paths. They are derived
  * from the ACSObjects in the generic gBase package, and the set of properties
  * they handle. Including the AcsAddon relations, about 100 3-nodes path formats
  * are possible. <P>
  * These AcsAddon relations are detected in the NodeRights and LinkRights hook methods.
  * To see all the AcsAddon relations, let's go to
  * the 4 selectOneHiddenNodePathsForXXX() methods in this current class.
  * The 55 basic formats, without AcsAddon relations, are listed hereinafter:<P>
  * - ep virtual folder/virtual member/actor started from xid/run under/
  *     current account or group or secondary group, <P>
  * - ep virtual folder/virtual member/actor/pri/resource or virtual folder, <P>
  * - ep virtual folder/virtual member/actor/acl/resource, <P>
  * - ep/pri/actor/acl/resource, <P>
  * - ep/pri/actor started from xid/run under/account or group or secondary group, <P>
  * - ep/pri/actor/bridge/actor, <P>
  * - ep/pri/virtual folder/virtual member/resource. <P>
  * - actor/bridge/actor/acl/resource, <P>
  * - actor/bridge/actor/pri/resource or virtual folder, <P>
  * - actor/bridge/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/bridge/actor/bridge/actor, <P>
  * - actor/acl/actor/acl/resource, <P>
  * - actor/acl/actor/pri/resource or virtual folder, <P>
  * - actor/acl/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/acl/actor/bridge/actor, <P>
  * - actor started from xid/run under/account or group or secondary group/acl/resource, <P>
  * - actor started from xid/run under/account or group or secondary group/pri/
  *         resource or virtual folder, <P>
  * - actor started from xid/run under/account or group or secondary group/member/group, <P>
  * - actor started from xid/run under/account or group or secondary group/
  *         own or contain/resource, <P>
  * - actor started from xid/alias/actor or groupidmember or virtual folder/run under/account or group or secondary group. <P>
  * - account or group/own or contain/actor/acl/resource, <P>
  * - account or group/own or contain/actor/bridge/actor, <P>
  * - account or group/own or contain/actor/pri/resource or virtual folder, <P>
  * - groupidmember/member/group/acl/resource, <P>
  * - groupidmember/member/group/pri/resource or virtual folder, <P>
  * - groupidmember/member/group/member/group, <P>
  * - groupidmember/member/group/contain/resource, <P>
  * - groupidmember/acl/actor/acl/resource, <P>
  * - groupidmember/acl/actor/bridge/actor, <P>
  * - groupidmember/acl/actor/pri/resource or virtual folder, <P>
  * - groupidmember/acl/actor started from xid/
  *         run under/current account or group or secondary group, <P>
  * - virtual folder/alias/virtual folder/virtual member/resource, <P>
  * - virtual folder/alias/virtual folder/alias/resource, <P>
  * - virtual folder/alias/actor started from xid/run under/
  *      current account or group or secondary group, <P>
  * - virtual folder/alias/ep/pri/resource or virtual folder, <P>
  * - virtual folder/alias/ep/acl/resource, <P>
  * - virtual folder/alias/actor/bridge/actor, <P>
  * - virtual folder/alias/ep/alias/reference, <P>
  * - virtual folder/virtual member/resource/alias/reference. <P>
  * - ep/pri/resource/alias/reference, <P>
  * - ep/pri/virtual folder/alias/reference, <P>
  * - actor/alias/actor started from xid/run under/
  *      current account or group or secondary group,<P>
  * - actor/alias/ep/pri/resource or virtual folder,<P>
  * - actor/alias/ep/acl/resource, <P>
  * - actor/alias/reference/alias/reference, <P>
  * - actor/bridge/actor/alias/reference, <P>
  * - actor/acl/resource/alias/reference, <P>
  * - actor started from xid/run under/account or group or secondary group/alias/reference, <P>
  * - groupidmember/alias/actor started from xid/run under/
  *      current account or group or secondary group,<P>
  * - groupidmember/alias/ep/pri/resource or virtual folder,<P>
  * - groupidmember/alias/ep/acl/resource, <P>
  * - groupidmember/alias/reference/alias/reference, <P>
  * - account or group/own or contain/actor/alias/reference, <P>
  * - groupidmember/member/group/alias/reference, <P>
  * - groupidmember/acl/actor/alias/reference, <P>
  * <P>
  * Called by RightsFactory_Facade.detectEPRRights() and detectNoThanRights().
  * Caution: called by detectNoThanRights(), _viewInBase is not the view of _gview,
  * since this method uses one transient EPRViewInBase for each
  * pair of nodes from the initial NoThanViewInBase. <P>
  * Calls selectOneHiddenNodePathsForVirtualFolder(),
  * selectOneHiddenNodePathsForFirstLinkPrivileges(),
  * selectOneHiddenNodePathsForActor(), selectOneHiddenNodePathsForGroupIDMember(),
  * and GraphicView.setRightsRelatedBaseObjects().
  * @see #selectOneHiddenNodePathsForVirtualFolder(ARoad0.gBaseInterface.VirtualFolder, ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, java.util.Map, java.util.HashSet)
  * @see #selectOneHiddenNodePathsForFirstLinkPrivileges(ARoad0.gBaseInterface.EligibleParty, ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, java.util.Map, java.util.HashSet)
  * @see #selectOneHiddenNodePathsForActor(ARoad0.gBaseInterface.Actor, ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, java.util.Map, java.util.HashSet)
  * @see #selectOneHiddenNodePathsForGroupIDMember(ARoad0.gBaseInterface.GroupIDMember, ARoad0.Gui2.GraphicView, ARoad0.gBaseInterface.EPRViewInBase, java.util.Map, java.util.HashSet) 
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
  * */
  public final Map detectOneHiddenNodeCompoundRights(GraphicView _gview,
    EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks) throws InterruptedException {

                                  // for testing and tracking
                                  //System.out.println("Call to ThreeNodesRightsFactoryImpl."+
                                    //"detectOneHiddenNodeCompoundRights() ");

                    // for testing and tracking
                    //System.out.println("In ThreeNodesRightsFactoryImpl."+
                      //"detectOneHiddenNodeCompoundRights(), "+
                      //"the DisplayableLinkImpls at the start are:");
                    //ImmutableName key;
                    //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                      // key = (ImmutableName) ite.next();
                      // System.out.println("--- for key: "+key);
                      // for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                      //  System.out.println("------------"+
                        //  ((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      // variables
      EligibleParty           ep;
      ImmutableSource         so;

      HashSet                 l_ObjectToRegister = _gview.getRightsRelatedBaseObjects();

      // iterate on the view EligibleParties and VirtualFolders
      for (Iterator ite = _viewInBase.getEorM_Sources().values().iterator() ; ite.hasNext() ; ) {

        so = (ImmutableSource) ite.next();
        if (so == null) continue;

        // ITERATE ITE NUMBER 0 on VirtualFolder as first end,
        // for searching
        // ep virtual folder/virtual member/actor started from xid/run under/
            // current account or group or secondary group
        // ep virtual folder/virtual member/actor/pri/resource or virtual folder
        // ep virtual folder/virtual member/actor/acl/resource
        // ep virtual folder/any AcsAddon relation/actor started from xid/run under/
            // current account or group or secondary group
        // ep virtual folder/any AcsAddon relation/actor/pri/resource or virtual folder
        // ep virtual folder/any AcsAddon relation/actor/acl/resource
        // ep virtual folder/virtual member/actor/any AcsAddon relation/target
        // ep virtual folder/any AcsAddon relation/actor/any AcsAddon relation/target
        if (so instanceof VirtualFolder) {
            // Resources VirtualFolders are also processed for Alias relations

            selectOneHiddenNodePathsForVirtualFolder((VirtualFolder)so, _gview,
                    _viewInBase, _m_l_DisplayableLinks, l_ObjectToRegister);

            // end of iteration ite for VirtualFolders
            continue;
        } // end of ITERATE ITE NUMBER 0 on VirtualFolder as first end

        // so is an EligibleParty
        ep = (EligibleParty)so;

        // ITERATE ITE NUMBER 1 on Sources as first end with Privileges,
        // for searching
        // ep/pri/actor/acl/resource
        // ep/pri/actor/any AcsAddon relation/target
        // ep/pri/actor started from xid/run under/account or group or secondary group
        // ep/pri/actor/bridge/actor
        // ep/pri/virtual folder/any AcsAddon relation/resource
        // ep/pri/virtual folder/virtual member/resource
        selectOneHiddenNodePathsForFirstLinkPrivileges(ep, _gview,
                    _viewInBase, _m_l_DisplayableLinks, l_ObjectToRegister);

        // ITERATE ITE NUMBER 2 on Actors as first end, without Privileges in the first link
        if (ep instanceof Actor) {
            selectOneHiddenNodePathsForActor((Actor)ep, _gview,
                    _viewInBase, _m_l_DisplayableLinks, l_ObjectToRegister);
        } // if

        // ITERATE ITE NUMBER 3 on GroupIDMembers as first end, without Privileges in the first link
        if (ep instanceof GroupIDMember) {
            selectOneHiddenNodePathsForGroupIDMember((GroupIDMember)ep, _gview,
                    _viewInBase, _m_l_DisplayableLinks, l_ObjectToRegister);
        } // if
      } // for (Iterator ite

      _gview.setRightsRelatedBaseObjects(l_ObjectToRegister);

                    // for testing and tracking
                    //System.out.println("In ThreeNodesRightsFactoryImpl."+
                      //"detectOneHiddenNodeCompoundRights(), "+
                      //"the DisplayableLinkImpls at the end are:");
                    //ImmutableName key;
                    //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                       //key = (ImmutableName) ite.next();
                       //System.out.println("--- for key: "+key);
                       //for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                          //System.out.println("------------"+
                          //((DisplayableLinkImpl)ita.next()).toString() );
                      //} // for ita
                    //} // for ite

      ep = null;
      so = null;
      l_ObjectToRegister = null;

      return _m_l_DisplayableLinks;
  } // detectOneHiddenNodeCompoundRights()

  /**
  * Defines all the 8+11 3-nodes access paths when the first end is a VirtualFolder.
  * The return paths are listed hereinafter: <P>
  * - ep virtual folder/virtual member/actor started from xid/run under/
  *      current account or group or secondary group,<P>
  * - ep virtual folder/virtual member/actor/pri/resource or virtual folder,<P>
  * - ep virtual folder/virtual member/actor/acl/resource,<P>
  * - ep virtual folder/virtual member/actor/alias/reference,<P>
  * - ep virtual folder/any AcsAddon relation/actor started from xid/run under/
  *      current account or group or secondary group,<P>
  * - ep virtual folder/any AcsAddon relation/actor/pri/resource or virtual folder,<P>
  * - ep virtual folder/any AcsAddon relation/actor/acl/resource.<P>
  * - ep virtual folder/any AcsAddon relation/actor/alias/reference.<P>
  * - ep virtual folder/virtual member/actor/any AcsAddon relation/target, <P>
  * - ep virtual folder/any AcsAddon relation/actor/any AcsAddon relation/target. <P>
  * - virtual folder/alias/virtual folder/virtual member/resource, <P>
  * - virtual folder/alias/virtual folder/alias/reference, <P>
  * - virtual folder/alias/virtual folder/any AcsAddon relation/target <P>
  * - virtual folder/alias/actor started from xid/run under/
  *      current account or group or secondary group, <P>
  * - virtual folder/alias/ep/pri/resource or virtual folder, <P>
  * - virtual folder/alias/ep/acl/resource, <P>
  * - virtual folder/alias/ep/any AcsAddon relation/target, <P>
  * - virtual folder/alias/actor/bridge/actor, <P>
  * - virtual folder/alias/ep/alias/reference, <P>
  * Called by detectOneHiddenNodeCompoundRights().
  * @param _vf the virtual folder as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update.
  * @param _l_ObjectToRegister has to be update for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map selectOneHiddenNodePathsForVirtualFolder(VirtualFolder _vf,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

      // variables
      GroupIDMember           gm = null;
      ImmutableTarget         targ;
      Actor                   actor;
      DisplayableLinkImpl     dLink = null;
      DisplayableLinkImpl     dLink_2 = null;
      Set<DisplayableLinkImpl> l_dLink = null;
      PrivilegeForType        tPriv;
      PrivilegeForLinks       lPriv;
      List<Privilege>         l_priv = null;
      List<DisplayableLinkImpl> l_temporaryLinks = null;
      Set<DisplayableLinkImpl> l_temporaryLinks2 = null;
      List<ACSObject>          l_temporaryLinks3 = new ArrayList(10);
      Set<ImmutableGroupIDMember> l_members = null;
      NodeRights              nodeRights = interpreter_.getNodeRightsInstances().get(_vf.getName());
      NodeRights              nodeRights2 = null;
      LinkRights              linkRights = null;
      AccessControlLinkImpl   acLink = null;

      // node.getSelectAddonNewHiddenNodeForActor() has to be called first.
        // detects the VirtualFolders which have complementary links
        // from their AcsAddon ACS
    if (nodeRights.getSelectAddonNewHiddenNodeForVirtualFolder()) {

            l_temporaryLinks = nodeRights.selectAddonNewHiddenNodeForVirtualFolder(
                        _viewInBase, _vf, null, _m_l_DisplayableLinks, l_temporaryLinks3);
            _l_ObjectToRegister.addAll(l_temporaryLinks3);

            if (l_temporaryLinks != null) {
                for (Iterator itu = l_temporaryLinks.iterator();
                        itu.hasNext(); ) {

                      dLink = (DisplayableLinkImpl)itu.next();
                      _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                      if (!(dLink.getEorSecondEnd() instanceof Actor))
                          itu.remove();
                } // for itu
            } // if (l_temporaryLinks4 != null)
    } // if node.getSelectAddonNewHiddenNodeForVirtualFolder()

    l_dLink = compound_.
            selectDirectHiddenMembersForVirtualFolder(_gview, _viewInBase, _vf);

    if (l_temporaryLinks != null && !l_temporaryLinks.isEmpty()) {
        if (l_dLink != null)
            l_dLink.addAll(l_temporaryLinks);
        else
            l_dLink = new HashSet(l_temporaryLinks);
    } // if

      // virtual members may be actors, since it is an EP_FOLDER VirtualFolder
    if (l_dLink != null) {
        for (Object ob : l_dLink) {
            dLink = (DisplayableLinkImpl)ob;

            // searching for
            // virtual folder/virtual member/resource/alias/reference
            if (((ImmutableBasic)dLink.getEorSecondEnd()).getEorACS().getManageInternalHardAlias() ||
                    _viewInBase.getConnectedToSeveralACS()) {

                l_temporaryLinks2 = compound_.selectDirectReferencesInViewForAlias(
                        _viewInBase, (ACSObject)dLink.getEorSecondEnd());
                if (l_temporaryLinks2 != null) {
                    for (Object obj : l_temporaryLinks) {
                        dLink_2 = linkUtil_.mergeLinks(dLink, (DisplayableLinkImpl)obj);

                        linkRights = interpreter_.getLinkRightsInstances().get(
                                ((DisplayableLinkImpl)obj).getEorSecondEnd().getName());
                        linkRights.addDisplayableLink(
                                dLink_2, _m_l_DisplayableLinks);
                    } // for obj
                } // if (l_temporaryLinks2 != null)
            } // if (getManageInternalHardAlias()

            if (!(dLink.getEorSecondEnd() instanceof EligibleParty))
                continue;
            actor = (Actor)dLink.getEorSecondEnd();

            if (!actor.getIsRightProxy()) continue;

            _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

            linkRights = interpreter_.getLinkRightsInstances().get(
                    dLink.getEorSecondEnd().getName());
            if (linkRights == null) {
                interpreter_.addInstancesForNode(dLink.getEorSecondEnd());
                linkRights = interpreter_.getLinkRightsInstances().get(
                    dLink.getEorSecondEnd().getName());
            } // if

                // l_members may be not null only for 'ep virtual folder/any AcsAddon relation'
                // or if linkRights.updateAGrunningContext() overrides the LinkRightsImpl method.
            try {
                l_members = linkRights.updateAGrunningContext(dLink);
            }
            catch (ProcessError p) {
                throw new InternalError("ThreeNodesRightsFactoryImpl."+
                        "selectOneHiddenNodePathsForVirtualFolder(): "+p.toString());
            }
            if (l_members == null) {
                // searching for
                // ep virtual folder/virtual member/actor started from xid/run under/
                    // current account or group or secondary group
                // ep virtual folder/any AcsAddon relation/actor started from xid/run under/
                    // current account or group or secondary group
                endsPathsFromActorWithStartingUnderXID(dLink, _viewInBase, _m_l_DisplayableLinks);
            } // if (l_members == null)

            // ------------- for ITERATE NUMBER 0 on VirtualFolder as first end

            // searching for
            // ep virtual folder/virtual member/actor/pri/resource or virtual folder - step 1
            // ep virtual folder/any AcsAddon relation/actor/pri/resource or virtual folder - step 1
            l_priv = actor.getEorL_SourcePrivilegeForTypes();
            if (!l_priv.isEmpty()) {

                            // for testing and tracking
                            //System.out.println("In ThreeNodesRightsFactoryImpl."+
                              //"detectOneHiddenNodeCompoundRights(), "+
                              //"getEorL_SourcePrivilegeForTypes() is not empty for "+
                              //actor.getNickName()+" through the searching for "+ep.getNickName());

                for (Object obje : l_priv) {
                    tPriv = (PrivilegeForType)obje;
                    for (Object object : _viewInBase.getEorM_Targets().values()) {
                        if (!(object instanceof ImmutablePrivilegeTarget))
                            continue;

                        linkRights = interpreter_.getLinkRightsInstances().get(
                                ((ImmutablePrivilegeTarget)object).getName());
                        if (linkRights == null) {
                            interpreter_.addInstancesForNode(((ImmutablePrivilegeTarget)object));
                            linkRights = interpreter_.getLinkRightsInstances().get(
                                ((ImmutablePrivilegeTarget)object).getName());
                        } // if

                        if (!linkRights.isTargetForPrivilegeForType(
                                tPriv, (ImmutablePrivilegeTarget)object)) {

                                // for testing and tracking
                                //System.out.println("With the actor "+actor.getNickName()+
                                  //", this typed privilege does not match: "+tPriv.getNickName());

                            continue;
                        } // if

                        dLink_2 = DisplayableLinkImpl.valueOf(
                                dLink, (BaseObject)object, new AccessControlLinkImpl());
                        dLink_2 = linkRights.addDisplayableLink(
                                dLink_2, _m_l_DisplayableLinks);
                        if (dLink_2 == null) continue;

                        if (tPriv.getPositiveRight()) {
                            dLink_2.getEorLastLink().addRightsArray(
                                    AccessControlLink.TPRI, tPriv.getL_Rights());
                        }
                        else {
                            dLink_2.getEorLastLink().addRightsArray(
                                    AccessControlLink.NEG_TPRI, tPriv.getL_Rights());
                        } // else

                    } // for object
                } // for obje
            } // if PrivilegeForTypes for actor

            // ep virtual folder/virtual member/actor/pri/resource or virtual folder - step 2
            // ep virtual folder/any AcsAddon relation/actor/pri/resource or virtual folder - step 2
            l_priv = actor.getEorL_SourcePrivilegeForLinks();
            if (!l_priv.isEmpty()) {
                for (Object obje : l_priv) {
                    lPriv = (PrivilegeForLinks)obje;
                    for (Object object : _viewInBase.getEorM_Targets().values()) {
                        if (!(object instanceof ImmutablePrivilegeTarget))
                            continue;
                        if (!lPriv.getEorLinkedTargets().contains(
                                (ImmutablePrivilegeTarget)object)) {

                                // for testing and tracking
                                //System.out.println("With the actor "+actor.getNickName()+
                                  //", this linked privilege does not match: "+lPriv.getNickName());

                            continue;
                        } // if

                        dLink_2 = DisplayableLinkImpl.valueOf(
                                dLink, (BaseObject)object, new AccessControlLinkImpl());
                        linkRights = interpreter_.getLinkRightsInstances().get(
                                ((BaseObject)object).getName());
                        dLink_2 = linkRights.addDisplayableLink(
                                dLink_2, _m_l_DisplayableLinks);
                        if (dLink_2 == null) continue;

                        if (lPriv.getPositiveRight()) {
                            dLink_2.getEorLastLink().addRightsArray(
                                    AccessControlLink.LPRI, lPriv.getL_Rights());
                        }
                        else {
                            dLink_2.getEorLastLink().addRightsArray(
                                    AccessControlLink.NEG_LPRI, lPriv.getL_Rights());
                        } // else

                    } // for object
                } // for obje
            } // if PrivilegeForLinks for actor

            // ------------- for ITERATE NUMBER 0 on VirtualFolder as first end

            for (Iterator iti = _viewInBase.getEorM_Targets().values().iterator() ;
              iti.hasNext() ; ) {
                targ = (ImmutableTarget) iti.next();
                if (targ == _vf) continue;

                nodeRights2 = interpreter_.getNodeRightsInstances().get(targ.getName());

                // searching for
                // ep virtual folder/virtual member/actor/any AcsAddon relation/target
                // ep virtual folder/any AcsAddon relation/actor/any AcsAddon relation/target
                if (nodeRights2.getSelectAddonLastViewNodeAfterActor()) {
                    dLink_2 = nodeRights2.selectAddonLastViewNodeAfterActor(
                        _viewInBase, targ, dLink, _m_l_DisplayableLinks);

                    if (dLink_2 != null) {
                        linkRights = interpreter_.getLinkRightsInstances().get(
                                targ.getName());
                        linkRights.addDisplayableLink(
                                    dLink_2, _m_l_DisplayableLinks);
                    } // if (dLink_2 != null
                } // if

                if (!(targ instanceof ImmutableResource))
                    continue;

                // searching for
                // ep virtual folder/virtual member/actor/acl/resource
                // ep virtual folder/any AcsAddon relation/actor/acl/resource
                acLink = acl_.addAclInAccessControlLink((ImmutableResource)targ, actor, null);

                if (acLink != null) {
                    dLink_2 = DisplayableLinkImpl.valueOf(dLink, targ, acLink);
                    linkRights = interpreter_.getLinkRightsInstances().get(
                                targ.getName());
                    linkRights.addDisplayableLink(
                                dLink_2,
                                _m_l_DisplayableLinks);
                } // if (acLink != null

                // searching for
                // virtual folder/any AcsAddon relation/actor/alias/reference
                if (actor.getEorACS().getManageInternalHardAlias() ||
                        _viewInBase.getConnectedToSeveralACS()) {

                    l_temporaryLinks2 = compound_.selectDirectReferencesInViewForAlias(
                            _viewInBase, actor);
                    if (l_temporaryLinks2 != null) {
                        for (Object obj : l_temporaryLinks2) {
                            dLink_2 = linkUtil_.mergeLinks(dLink, (DisplayableLinkImpl)obj);

                            linkRights = interpreter_.getLinkRightsInstances().get(
                                ((DisplayableLinkImpl)obj).getEorSecondEnd().getName());
                            linkRights.addDisplayableLink(
                                    dLink_2, _m_l_DisplayableLinks);
                        } // for obj
                    } // if (l_temporaryLinks2 != null
                } // if (actor2.getManageInternalHardAlias()
            } // for iti

        } // for ob on l_dLink
    } // if (l_dLink != null)

    l_dLink = compound_.selectDirectHiddenReferenceForAlias(_gview, _viewInBase, _vf);
    if (l_dLink != null) {
        for (Object link : l_dLink) {
            dLink = (DisplayableLinkImpl)link;

            _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

            // searching for
            // - virtual folder/alias/ep/pri/resource or virtual folder,
            // - virtual folder/alias/ep/acl/resource,
            // - virtual folder/alias/ep/any AcsAddon relation/target,
            // - virtual folder/alias/actor/bridge/actor,
            // - virtual folder/alias/ep/alias/reference
            if (dLink.getEorSecondEnd() instanceof Actor)
                endsPathsFromActorWithAclBridgePrivilegeAlias(dLink, _viewInBase, _m_l_DisplayableLinks);
            else if (dLink.getEorSecondEnd() instanceof GroupIDMember)
                endsPathsFromGroupIDMemberWithAclPrivilegeAlias(dLink, _viewInBase, _m_l_DisplayableLinks);

            // searching for
            // - virtual folder/alias/actor started from xid/run under/
            //      current account or group or secondary group
            if (dLink.getEorSecondEnd() instanceof Actor) {
                endsPathsFromActorWithStartingUnderXID(dLink, _viewInBase, _m_l_DisplayableLinks);
            }

            // searching for
            // - virtual folder/alias/virtual folder/virtual member/resource,
            // - virtual folder/alias/virtual folder/alias/resource,
            // - virtual folder/alias/virtual folder/any AcsAddon relation/target
            if (dLink.getEorSecondEnd() instanceof VirtualFolder) {
                endsPathsFromVirtualFolderWithMemberAlias(dLink, _viewInBase, _m_l_DisplayableLinks);
            }
        } // for link
    } // if (l_dLink != null)

    targ = null;
    gm = null;
    actor = null;
    dLink = null;
    dLink_2 = null;
    tPriv = null;
    lPriv = null;
    l_priv = null;
    l_dLink = null;
    l_members = null;
    nodeRights = null;
    nodeRights2 = null;
    linkRights = null;
    acLink = null;
    l_temporaryLinks = null;
    l_temporaryLinks2 = null;
    l_temporaryLinks3.clear();
    l_temporaryLinks3 = null;

    if (Thread.interrupted()) {
        throw new InterruptedException();
    }

    return _m_l_DisplayableLinks;
  } // selectOneHiddenNodePathsForVirtualFolder()

  /**
  * Defines all the 6+2 3-nodes access paths when the first end is an EligibleParty
  * having Privileges.
  * The return paths are listed hereinafter: <P>
  * - ep/pri/actor/acl/resource, <P>
  * - ep/pri/actor/any AcsAddon relation/target, <P>
  * - ep/pri/actor started from xid/run under/account or group or secondary group, <P>
  * - ep/pri/actor/bridge/actor, <P>
  * - ep/pri/virtual folder/any AcsAddon relation/resource, <P>
  * - ep/pri/virtual folder/virtual member/resource. <P>
  * - ep/pri/resource/alias/reference, <P>
  * - ep/pri/virtual folder/alias/reference. <P>
  * Called by detectOneHiddenNodeCompoundRights(). <P>
  * Caution: when called indirectly by RightsFactory_Facade.detectNoThanRights(),
  * _viewInBase is not the view of _gview,
  * since this method uses transient EPRViewInBases with for each of them,
  * a pair of nodes from the initial NoThanViewInBase. <P>
  * @param _ep the EligibleParty as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map selectOneHiddenNodePathsForFirstLinkPrivileges(EligibleParty _ep,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

        // variables
        GroupIDMember           gm = null;
        ImmutableResource       res;
        ImmutableTarget         targ;
        DisplayableLinkImpl     dLink = null;
        DisplayableLinkImpl     dLink_2 = null;
        Set<DisplayableLinkImpl> l_dLink = null;
        Set<DisplayableLinkImpl> l_temporaryLinks2 = null;
        AccessControlLinkImpl   acLink = null;
        BaseObject              bo;
        Set<ImmutableGroupIDMember>      l_directEP = null;
        String                  rightType = null;
        NodeRights              nodeRights = null;
        LinkRights              linkRights = null;

        l_dLink = pri_.
              selectDirectHiddenPrivilegeLinksForEP(_gview, _viewInBase, _ep, true);
        if (l_dLink != null) {

            for (Object obje : linkUtil_.
                    getAllEndsInDisplayableLinks(l_dLink, false)) {
                bo = (BaseObject)obje;

                _l_ObjectToRegister.add(bo);
                
                for (Iterator ito = linkUtil_.
                  selectDisplayableLinksForSecondEnd(bo, l_dLink).iterator(); ito.hasNext(); ) {

                    dLink = (DisplayableLinkImpl) ito.next();
                        
                                        // for testing and tracking
                                        //System.out.println("In ThreeNodesRightsFactoryImpl."+
                                          //"selectOneHiddenNodePathsForFirstLinkPrivileges(), "+
                                          //"start the analysis of dLink "+dLink);


                    nodeRights = interpreter_.getNodeRightsInstances().get(bo.getName());
                    if (nodeRights == null) {
                        interpreter_.addInstancesForNode(bo);
                        nodeRights = interpreter_.getNodeRightsInstances().get(bo.getName());
                    } // if

                    if (bo instanceof ImmutableActor) {
                        // searching for
                        // ep/pri/actor/acl/resource
                        for (Iterator itu = _viewInBase.getEorM_Targets().values().
                                iterator(); itu.hasNext(); ) {

                            targ = (ImmutableTarget)itu.next();
                            if (targ == bo || targ == _ep) continue;
                            if (!(targ instanceof ImmutableResource))
                                continue;
                            res = (ImmutableResource)targ;

                            acLink = acl_.addAclInAccessControlLink(
                                        res, (ImmutableActor)bo, null);
                            if (acLink == null) continue;

                            dLink_2 = DisplayableLinkImpl.valueOf(dLink, res, acLink);

                            linkRights = interpreter_.getLinkRightsInstances().get(res.getName());
                            linkRights.addDisplayableLink(
                                            dLink_2, _m_l_DisplayableLinks);
                        } // for itu

                        // searching for
                        // ep/pri/actor/any AcsAddon relation/target
                        if (nodeRights.getSelectAddonLastViewNodeAfterActor()) {

                                for (Iterator itu = _viewInBase.getEorM_Targets().values().
                                        iterator(); itu.hasNext(); ) {
                                    targ = (ImmutableTarget)itu.next();
                                    if (targ == bo || targ == _ep) continue;

                                    dLink_2 = nodeRights.selectAddonLastViewNodeAfterActor(
                                        _viewInBase, targ, dLink, _m_l_DisplayableLinks);

                                    if (dLink_2 != null) {
                                        linkRights = interpreter_.getLinkRightsInstances().get(
                                                targ.getName());
                                        linkRights.addDisplayableLink(
                                                    dLink_2, _m_l_DisplayableLinks);
                                    } // if (dLink_2 != null
                                } // for itu

                        } // if node.getSelectAddonNewHiddenNodeForGroupIDMember()

                        // searching for
                        // ep/pri/actor started from xid/run under/
                                // account or group or secondary group
                        l_directEP = nodeRights.
                                selectDirectOwnerContainGlobalForActorAsEP(
                                        _viewInBase, (ImmutableActor)bo);
                        if (l_directEP != null) {
                            for (Object obj : l_directEP) {
                                gm = (GroupIDMember)obj;

                                rightType = null;
                                // set the XID comment if the first end is an actor and the next node,
                                // an actor which runs under its current user/group or secondary group
                                rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                                if (rightType != null && gm instanceof GroupID)
                                  dLink.getEorLastLink().addComment(
                                          rightType, AccessControlLink.C_SET_GROUP);
                                else if (rightType != null && gm instanceof UserID)
                                  dLink.getEorLastLink().addComment(
                                          rightType, AccessControlLink.C_SET_USER);

                                dLink_2 = DisplayableLinkImpl.valueOf(
                                        dLink, gm, new AccessControlLinkImpl());

                                dLink_2.getEorLastLink().setLinkTypeToTrue(
                                        AccessControlLink.RUN_UNDER);
                                // set the GLOBAL type if the first end is an actor, and the next node
                                // is a secpndary group
                                if (gm instanceof GroupID &&
                                  ((ImmutableActor)bo).isMemberToSecondaryGroup((GroupID)gm) ) {
                                    dLink_2.getEorLastLink().setLinkTypeToTrue(
                                            AccessControlLink.GLOBAL);
                                    dLink_2.getEorLastLink().addComment(
                                              AccessControlLink.GLOBAL,
                                              AccessControlLink.C_SECONDARY_GROUP);
                                } // if

                                linkRights = interpreter_.getLinkRightsInstances().get(
                                        gm.getName());
                                linkRights.addDisplayableLink(
                                        dLink_2, _m_l_DisplayableLinks);
                            } // for obje
                        } // if l_directEP != null

                        // searching for
                        // ep/pri/actor/bridge/actor
                        if (((ImmutableActor)bo).getIsACSBridgeSource() &&
                             _viewInBase.containsSource(((ImmutableActor)bo).
                                getEorACSBridgeTarget())) {

                                res = ((ImmutableActor)bo).getEorACSBridgeTarget();

                                dLink_2 = DisplayableLinkImpl.valueOf(
                                        dLink, res, new AccessControlLinkImpl());
                                dLink_2.getEorLastLink().addRight(
                                    AccessControlLink.BRIDGE,
                                    BaseManagerImpl.getInstance().getMetaRight("execute") );

                                linkRights = interpreter_.getLinkRightsInstances().get(
                                        res.getName());
                                linkRights.addDisplayableLink(
                                    dLink_2, _m_l_DisplayableLinks);
                        } // if (((ImmutableActor)bo).getIsACSBridgeSource()
                    } // if (bo instanceof ImmutableActor)

                    else if (bo instanceof ImmutableVirtualFolder) {

                        // searching for
                        // ep/pri/virtual folder/virtual member/resource
                        // ep/pri/virtual folder/alias/reference
                        // ep/pri/virtual folder/any AcsAddon relation/resource
                        endsPathsFromVirtualFolderWithMemberAlias(dLink,
                            _viewInBase, _m_l_DisplayableLinks);

                    } // if (bo instanceof ImmutableVirtualFolder)

                    if (bo instanceof ACSObject) {
                        // searching for
                        // ep/pri/resource/alias/reference
                        // ep/pri/virtual folder/alias/reference
                        if (((ACSObject)bo).getEorACS().getManageInternalHardAlias() ||
                                _viewInBase.getConnectedToSeveralACS()) {

                            l_temporaryLinks2 = compound_.selectDirectReferencesInViewForAlias(
                                    _viewInBase, (ACSObject)bo);
                            if (l_temporaryLinks2 != null) {
                                for (Object obj : l_temporaryLinks2) {
                                    dLink_2 = linkUtil_.mergeLinks(dLink, (DisplayableLinkImpl)obj);

                                    linkRights = interpreter_.getLinkRightsInstances().get(
                                            ((DisplayableLinkImpl)obj).getEorSecondEnd().getName());
                                    linkRights.addDisplayableLink(
                                            dLink_2, _m_l_DisplayableLinks);
                                } // for obj
                            } // if (l_temporaryLinks2
                        } // if (bo.getManageInternalHardAlias()
                    } // if (bo instanceof ACSObject)

                } // for Iteratot ito
            } // for (Object obje : l_dLink) {
        } // if (l_dLink != null)
        else {

                                        // for testing and tracking
                                        //System.out.println("In ThreeNodesRightsFactoryImpl."+
                                          //"selectOneHiddenNodePathsForFirstLinkPrivileges(), "+
                                          //"no dLink for "+_ep.getName());
        } // else

                                        // for testing and tracking
                                        //System.out.println("In ThreeNodesRightsFactoryImpl."+
                                          //"selectOneHiddenNodePathsForFirstLinkPrivileges(), "+
                                          //"the DisplayableLinkImpls at the end are:");
                                        //ImmutableName key;
                                        //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                                           //key = (ImmutableName) ite.next();
                                           //System.out.println("--- for key: "+key);
                                           //for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ;
                                             //   ita.hasNext() ; ) {
                                               //     System.out.println("------------"+
                                                 //     ((DisplayableLinkImpl)ita.next()).toString() );
                                           //} // for ita
                                        //} // for ite


        res = null;
        targ = null;
        gm = null;
        dLink = null;
        dLink_2 = null;
        acLink = null;
        bo = null;
        l_dLink = null;
        l_directEP = null;
        rightType = null;
        nodeRights = null;
        l_temporaryLinks2 = null;
        linkRights = null;

        if (Thread.interrupted()) {
            throw new InterruptedException();
        }

        return _m_l_DisplayableLinks;
    } // selectOneHiddenNodePathsForFirstLinkPrivileges()
  
  /**
  * Defines all the 29+14 3-nodes access paths when the first end is an Actor, without
  * privileges in the first link. The return paths are listed hereinafter: <P>
  * - actor/bridge/actor/acl/resource, <P>
  * - actor/bridge/actor/pri/resource or virtual folder, <P>
  * - actor/bridge/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/bridge/actor/bridge/actor, <P>
  * - actor/bridge/actor/alias/reference, <P>
  * - actor/bridge/actor/any AcsAddon relation/target, <P>
  * - actor/acl/actor/acl/resource, <P>
  * - actor/acl/actor/pri/resource or virtual folder, <P>
  * - actor/acl/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/acl/actor/bridge/actor, <P>
  * - actor/acl/actor/alias/reference, <P>
  * - actor/acl/actor/any AcsAddon relation/target, <P>
  * - actor started from xid/run under/account or group or secondary group/acl/resource, <P>
  * - actor started from xid/run under/account or group or secondary group/pri/
  *         resource or virtual folder, <P>
  * - actor started from xid/run under/account or group or secondary group/member/group, <P>
  * - actor started from xid/run under/account or group or secondary group/
  *         own or contain/resource, <P>
  * - actor started from xid/run under/alias/reference, <P>
  * - actor started from xid/run under/account or group or secondary group/
  *         any AcsAddon relation/target, <P>
  * - actor/any AcsAddon relation/actor/acl/resource, <P>
  * - actor/any AcsAddon relation/actor/pri/resource or virtual folder, <P>
  * - actor/any AcsAddon relation/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/any AcsAddon relation/actor/bridge/actor, <P>
  * - actor/any AcsAddon relation/actor/alias/reference, <P>
  * - actor/any AcsAddon relation/actor/any AcsAddon relation/target, <P>
  * - actor/any AcsAddon relation/account or group/acl/resource, <P>
  * - actor/any AcsAddon relation/account or group/pri/
  *         resource or virtual folder, <P>
  * - actor/any AcsAddon relation/account or group/member/group, <P>
  * - actor/any AcsAddon relation/account or group or secondary group/
  *         own or contain/resource, <P>
  * - actor/any AcsAddon relation/account or group/alias/reference, <P>
  * - actor/any AcsAddon relation/account or group/any AcsAddon relation/target, <P>
  * - actor/any AcsAddon relation/virtual folder/virtual member/resource, <P>
  * - actor/any AcsAddon relation/virtual folder/any AcsAddon relation/resource, <P>
  * - actor/alias/ep/pri/resource or virtual folder, <P>
  * - actor/alias/ep/acl/resource, <P>
  * - actor/alias/ep/run_under/groupidmember, <P>
  * - actorr/alias/ep/any AcsAddon relation/target, <P>
  * - actor/alias/actor/bridge/actor, <P>
  * - actor/alias/ep/alias/reference, <P>
  * - actor/alias/actor started from xid/run under/
  *      current account or group or secondary group, <P>
  * - actor/alias/virtual folder/virtual member/resource,
  * - actor/alias/virtual folder/alias/resource, <P>
  * - actor/alias/virtual folder/any AcsAddon relation/target, <P>
  * - actor/alias/virtual folder/run_under/groupidmember. <P>
  * Called by detectOneHiddenNodeCompoundRights().
  * Calls the 6 methods addPathsFromActorXXXX().
  * @param _act the Actor as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister is _gview.getRightsRelatedBaseObjects() to update
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map selectOneHiddenNodePathsForActor(Actor _act,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    addPathsFromActorBridgeActor(
            _act, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromActorAclActor(
            _act, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromActorStartsUnderGroupIDMember(
            _act, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromActorAcsAddonRelationActor(
            _act, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromActorAcsAddonRelationNoActor(
            _act, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromActorAliasReference(_act, _gview, _viewInBase, _m_l_DisplayableLinks,
            _l_ObjectToRegister);

    if (Thread.interrupted()) {
        throw new InterruptedException();
    }

    return _m_l_DisplayableLinks;
  } // selectOneHiddenNodePathsForActor()
  
  /**
  * Defines all the 26+14 3-nodes access paths when the first end is a GroupIDMember,
  * without privileges in the first link. The return paths are listed hereinafter: <P>
  * - account or group/own or contain/actor/acl/resource, <P>
  * - account or group/own or contain/actor/bridge/actor, <P>
  * - account or group/own or contain/actor/pri/resource or virtual folder, <P>
  * - account or group/own or contain/actor/alias/reference, <P>
  * - account or group/own or contain/actor/any AcsAddon relation/target, <P>
  * - groupidmember/member/group/acl/resource, <P>
  * - groupidmember/member/group/pri/resource or virtual folder, <P>
  * - groupidmember/member/group/member/group, <P>
  * - groupidmember/member/group/contain/resource, <P>
  * - groupidmember/member/group/alias/reference, <P>
  * - groupidmember/member/group/any AcsAddon relation/target, <P>
  * - groupidmember/acl/actor/acl/resource, <P>
  * - groupidmember/acl/actor/bridge/actor, <P>
  * - groupidmember/acl/actor/pri/resource or virtual folder,v
  * - groupidmember/acl/actor started from xid/
  *         run under/current account or group or secondary group, <P>
  * - groupidmember/acl/actor/alias/reference, <P>
  * - groupidmember/acl/actor/any AcsAddon relation/target, <P>
  * - groupidmember/any AcsAddon relation/actor/acl/resource, <P>
  * - groupidmember/any AcsAddon relation/actor/bridge/actor, <P>
  * - groupidmember/any AcsAddon relation/actor/pri/resource or virtual folder, <P>
  * - groupidmember/any AcsAddon relation/actor/alias/reference, <P>
  * - groupidmember/any AcsAddon relation/actor started from xid/
  *         run under/current account or group or secondary group, <P>
  * - groupidmember/any AcsAddon relation/actor/any AcsAddon relation/target, <P>
  * - groupidmember/any AcsAddon relation/account or group/own or contain/resource, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/member/group, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/acl/resource, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/pri/resource or virtual folder, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/alias/reference, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/any AcsAddon relation/target, <P>
  * - groupidmember/any AcsAddon relation/virtual folder/has member/resource, <P>
  * - groupidmember/any AcsAddon relation/virtual folder/any AcsAddon relation/resource, <P>
  * - groupidmember/alias/ep/pri/resource or virtual folder, <P>
  * - groupidmember/alias/ep/acl/resource, <P>
  * - groupidmember/alias/ep/any AcsAddon relation/target, <P>
  * - groupidmember/alias/actor/bridge/actor, <P>
  * - groupidmember/alias/ep/alias/reference, <P>
  * - groupidmember/alias/actor started from xid/run under/
  *      current account or group or secondary group, <P>
  * - groupidmember/alias/virtual folder/virtual member/resource,
  * - groupidmember/alias/virtual folder/alias/resource, <P>
  * - groupidmember/alias/virtual folder/any AcsAddon relation/target. <P>
  * Called by detectOneHiddenNodeCompoundRights().
  * Calls the 6 methods addPathsFromGroupIDMemberXXXX().
  * @param _gm the GroupIDMember as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map selectOneHiddenNodePathsForGroupIDMember(GroupIDMember _gm,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    addPathsFromGroupIDMemberOwnContainActor(
            _gm, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromGroupIDMemberIsMemberGroup(
            _gm, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromGroupIDMemberAclActor(
            _gm, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromGroupIDMemberAcsAddonRelationActor(
            _gm, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromGroupIDMemberAcsAddonRelationNoActor(
            _gm, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);

    addPathsFromGroupIDMemberAliasReference(
            _gm, _gview, _viewInBase, _m_l_DisplayableLinks, _l_ObjectToRegister);
    
    if (Thread.interrupted()) {
        throw new InterruptedException();
    }

    return _m_l_DisplayableLinks;
  } // selectOneHiddenNodePathsForGroupIDMember()

  /**
  * Defines all the 3-nodes access paths when the first end is an Actor with
  * a bridge in the first link. They are listed hereinafter: <P>
  * - actor/bridge/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/bridge/actor/bridge/actor, <P>
  * - actor/bridge/actor/acl/resource, <P>
  * - actor/bridge/actor/pri/resource or virtual folder, <P>
  * - actor/bridge/actor/alias/reference, <P>
  * - actor/bridge/actor/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForActor().
  * Calls endsPathsFromActorWithAclBridgePrivilegeAlias().
  * @param _act the Actor as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromActorBridgeActor(Actor _act,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    // variables
    GroupIDMember           gm = null;
    Set<ImmutableGroupIDMember>  l_directEP = null;
    ImmutableActor          actor2;
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink_2 = null;
    String                  rightType = null;
    NodeRights              nodeRights = null;
    LinkRights              linkRights = null;
        
    if (!_act.getIsACSBridgeSource() ||
                _viewInBase.containsSource(_act.getEorACSBridgeTarget()))
                        return _m_l_DisplayableLinks;

    actor2 = _act.getEorACSBridgeTarget();

    // register it even if not applicable yet
    _l_ObjectToRegister.add(actor2);

    dLink = new DisplayableLinkImpl(_act, actor2, new AccessControlLinkImpl());
    dLink.getEorFirstLink().addRight(
        AccessControlLink.BRIDGE,
        BaseManagerImpl.getInstance().getMetaRight("execute") );

    nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(actor2);
        nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
    } // if

    // searching for
    // actor/bridge/actor started from xid/run under/
            // account or group or secondary group
    l_directEP = nodeRights.
            selectDirectOwnerContainGlobalForActorAsEP(_viewInBase, actor2);
    if (l_directEP != null) {
        for (Object obj : l_directEP) {
            gm = (GroupIDMember)obj;

            rightType = null;
            // set the XID comment if the first end is an actor and the next node,
            // an actor which runs under its current user/group or secondary group
            rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

            if (rightType != null && gm instanceof GroupID)
              dLink.getEorLastLink().addComment(
                      rightType, AccessControlLink.C_SET_GROUP);
            else if (rightType != null && gm instanceof UserID)
              dLink.getEorLastLink().addComment(
                      rightType, AccessControlLink.C_SET_USER);

            dLink_2 = DisplayableLinkImpl.valueOf(
                    dLink, gm, new AccessControlLinkImpl());

            dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
            // set the GLOBAL type if the first end is an actor, and the next node
            // is a secondary group
            if (gm instanceof GroupID &&
              actor2.isMemberToSecondaryGroup((GroupID)gm) ) {
                dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                dLink_2.getEorLastLink().addComment(
                          AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
            } // if

            linkRights = interpreter_.getLinkRightsInstances().get(
                              gm.getName());
            linkRights.addDisplayableLink(
                    dLink_2, _m_l_DisplayableLinks);
        } // for obj
    } // if l_directEP != null

    endsPathsFromActorWithAclBridgePrivilegeAlias(
                    dLink,_viewInBase, _m_l_DisplayableLinks);

    gm = null;
    l_directEP = null;
    actor2 = null;
    dLink = null;
    dLink_2 = null;
    rightType = null;
    nodeRights = null;
    linkRights = null;
    
    return _m_l_DisplayableLinks;
  } // addPathsFromActorBridgeActor()
  
  /**
  * Defines all the 3-nodes access paths when the first end is an Actor with
  * an AclEntry in the first link. They are listed hereinafter: <P>
  * - actor/acl/actor/acl/resource, <P>
  * - actor/acl/actor/pri/resource or virtual folder, <P>
  * - actor/acl/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/acl/actor/bridge/actor, <P>
  * - actor/acl/actor/alias/reference, <P>
  * - actor/acl/actor/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForActor().
  * Calls endsPathsFromActorWithAclBridgePrivilegeAlias().
  * @param _act the Actor as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromActorAclActor(Actor _act,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    // variables
    ImmutableResource       res = null;
    GroupIDMember           gm = null;
    Set<ImmutableGroupIDMember> l_directEP = null;
    ImmutableActor          actor2;
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink_2 = null;
    String                  rightType = null;
    NodeRights              nodeRights = null;
    LinkRights              linkRights = null;
    
    if (_act.getEorL_AclConnectedResources() == null)
        return _m_l_DisplayableLinks;
    
    for (Object obje : _act.getEorL_AclConnectedResources()) {
        res = (ImmutableResource)obje;
        if (!(res instanceof ImmutableActor) || _viewInBase.containsTarget(res))
            continue;
        actor2 = (ImmutableActor)res;

        nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
        if (nodeRights == null) {
            interpreter_.addInstancesForNode(actor2);
            nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
        } // if

        dLink = acl_.addAclBridgeRightsInSimpleDisplayableLink(_act, actor2, true);

        if (dLink != null) {

            _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

            // searching for
            // actor/acl/actor started from xid/run under/
                    // account or group or secondary group
            l_directEP = nodeRights.
                    selectDirectOwnerContainGlobalForActorAsEP(
                            _viewInBase, actor2);
            if (l_directEP != null) {
                for (Object obj : l_directEP) {
                    gm = (GroupIDMember)obj;

                    rightType = null;
                    // set the XID comment if the first end is an actor and the next node,
                    // an actor which runs under its current user/group or secondary group
                    rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                    if (rightType != null && gm instanceof GroupID)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_GROUP);
                    else if (rightType != null && gm instanceof UserID)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_USER);

                    dLink_2 = DisplayableLinkImpl.valueOf(
                            dLink, gm, new AccessControlLinkImpl());

                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                    // set the GLOBAL type if the first end is an actor, and the next node
                    // is a secondary group
                    if (gm instanceof GroupID &&
                      actor2.isMemberToSecondaryGroup((GroupID)gm) ) {
                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink_2.getEorLastLink().addComment(
                                  AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                    } // if

                    linkRights = interpreter_.getLinkRightsInstances().get(
                              gm.getName());
                    linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
                } // for obj
            } // if l_directEP != null

            endsPathsFromActorWithAclBridgePrivilegeAlias(
                    dLink,_viewInBase, _m_l_DisplayableLinks);

        } // if (dLink != null
        
    } // for obje on actor2

    gm = null;
    l_directEP = null;
    actor2 = null;
    dLink = null;
    dLink_2 = null;
    rightType = null;
    nodeRights = null;
    linkRights = null;
    
    return _m_l_DisplayableLinks;
  } // addPathsFromActorAclActor()

  /**
  * Defines all the 3-nodes access paths when the first end is an Actor which
  * starts under a GroupIDMember. They are listed hereinafter: <P>
  * - actor started from xid/run under/account or group or secondary group/acl/resource, <P>
  * - actor started from xid/run under/account or group or secondary group/pri/
  *         resource or virtual folder, <P>
  * - actor started from xid/run under/account or group or secondary group/
  *         own or contain/resource, <P>
  * - actor started from xid/run under/account or group or secondary group/
  *         any AcsAddon relation/target, <P>
  * - actor started from xid/run under/account or group or secondary group/member/group, <P>
  * - actor started from xid/run under/account or group or secondary group/alias/reference, <P>
  * - actor started from xid/alias/actor or groupidmember or virtual folder/run under/account or group or secondary group. <P>
  * Called by selectOneHiddenNodePathsForActor().
  * Calls endsPathsFromGroupIDMemberWithAclPrivilegeAlias().
  * @param _act the Actor as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromActorStartsUnderGroupIDMember(Actor _act,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

                                        // for testing and tracking
                            //System.out.println("ThreeNodesRightsFactoryImpl."+
                              //"addPathsFromActorStartsUnderGroupIDMember(), start for the actor"+
                              //_act.getName());

    // variables
    GroupIDMember           gm = null;
    Set<ImmutableGroupIDMember> l_hiddenDirectEP = null;
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink_2 = null;
    NodeRights              nodeRights = null;
    LinkRights              linkRights = null;
    Set<DisplayableLinkImpl>    l_dLink = null;
    Set<ImmutableGroupIDMember> l_members = null;

    nodeRights = interpreter_.getNodeRightsInstances().get(_act.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(_act);
        nodeRights = interpreter_.getNodeRightsInstances().get(_act.getName());
    } // if

    // works because NodeRights of _act is identical to the NodeRights of gm
    l_hiddenDirectEP = nodeRights.selectHiddenDirectOwnerContainGlobalForActorAsEP(
                                _viewInBase, _act);

    if (l_hiddenDirectEP != null) {
        try {
            for (Object obj : l_hiddenDirectEP) {
                gm = (GroupIDMember)obj;

                dLink = new DisplayableLinkImpl(_act, gm, new AccessControlLinkImpl());

                _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                // set the GLOBAL type if the first end is an actor, and the next node
                // is a secondary group
                if (gm instanceof GroupID &&
                  _act.isMemberToSecondaryGroup((GroupID)gm) ) {
                    dLink.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                    dLink.getEorLastLink().addComment(
                              AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                } // if

                                            // for testing and tracking
                                //System.out.println("ThreeNodesRightsFactoryImpl."+
                                  //"addPathsFromActorStartsUnderGroupIDMember(), call to"+
                                  //"endsPathsFromGroupIDMemberWithAclPrivilegeAlias() "+
                                  //"for the link "+dLink);

                linkRights = interpreter_.getLinkRightsInstances().get(
                        gm.getName());
                if (linkRights == null) {
                    interpreter_.addInstancesForNode(gm);
                    linkRights = interpreter_.getLinkRightsInstances().get(
                        gm.getName());
                } // if

                linkRights.updateAGrunningContext(dLink);

                endsPathsFromGroupIDMemberWithAclPrivilegeAlias(dLink,
                        _viewInBase, _m_l_DisplayableLinks);

            } // for obj

        } catch (ProcessError p) {
            throw new InternalError("ThreeNodesRightsFactoryImpl."+
                    "addPathsFromActorStartsUnderGroupIDMember(): "+p.toString());
        }
    } // if l_hiddenDirectEP != null

    // actor started from xid/alias/actor or groupidmember or virtual folder/run under/groupidmember
    l_dLink = compound_.selectDirectHiddenReferenceForAlias(_gview, _viewInBase, _act);
    if (l_dLink != null) {
        for (Object link : l_dLink) {
            dLink = (DisplayableLinkImpl)link;

            l_members = linkUtil_.getSecondEndAGcontext(dLink);
            if (l_members == null || dLink.getProperties() == null ||
                        dLink.getProperties().get(
                            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) == null)
                continue;

            for (Object memb : l_members) {
                if (!_viewInBase.containsBaseObject((GroupIDMember)memb)) continue;

                dLink_2 = DisplayableLinkImpl.valueOf(dLink, (GroupIDMember)memb,
                        new AccessControlLinkImpl());

                if (memb instanceof GroupID)
                  dLink_2.getEorConnectedLinks(dLink.getEorSecondEnd())[0].
                          addComment(AccessControlLink.IS_ALIAS, AccessControlLink.C_SET_GROUP);
                else if (memb instanceof UserID)
                  dLink_2.getEorConnectedLinks(dLink.getEorSecondEnd())[0].
                          addComment(AccessControlLink.IS_ALIAS, AccessControlLink.C_SET_USER);

                dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);

                // set the GLOBAL type if the first end is an actor, and the next node
                // is a secondary group
                if (memb instanceof GroupID && dLink.getEorSecondEnd() instanceof Actor &&
                  ((Actor)dLink.getEorSecondEnd()).isMemberToSecondaryGroup((GroupID)memb) ) {
                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                    dLink_2.getEorLastLink().addComment(
                              AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                } // if

                                  // for testing and tracking
                                  //System.out.println(
                                    //"in ThreeNodesRightsFactoryImpl.addPathsFromActorStartsUnderGroupIDMember()"+
                                    //" add actor started from xid/alias/actor or groupidmember or virtual folder/run under/groupidmember: "+
                                    //dLink_2);

                linkRights = interpreter_.getLinkRightsInstances().get(
                    ((GroupIDMember)memb).getName());
                linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
            } // for memb

        } // for link
    } // if (l_dLink != null)

    gm = null;
    l_hiddenDirectEP = null;
    dLink = null;
    nodeRights = null;
    dLink_2 = null;
    linkRights = null;
    l_members = null;

    return _m_l_DisplayableLinks;
  } // addPathsFromActorStartsUnderGroupIDMember()
  
  /**
  * Defines all the 3-nodes access paths when the first end is an Actor with
  * an AcsAddon relation in the first link and an Actor as central node.
  * They are listed hereinafter: <P>
  * - actor/any AcsAddon relation/actor/acl/resource, <P>
  * - actor/any AcsAddon relation/actor/pri/resource or virtual folder, <P>
  * - actor/any AcsAddon relation/actor started from xid/run under/
  *         current account or group or secondary group, <P>
  * - actor/any AcsAddon relation/actor/bridge/actor, <P>
  * - actor/any AcsAddon relation/actor/alias/reference, <P>
  * - actor/any AcsAddon relation/actor/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForActor().
  * Calls endsPathsFromActorWithAclBridgePrivilegeAlias().
  * @param _act the Actor as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromActorAcsAddonRelationActor(Actor _act,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    List<DisplayableLinkImpl> l_dLink = null;
    List<ACSObject>          l_temporaryLinks3 = new ArrayList(10);
    GroupIDMember           gm = null;
    Set<ImmutableGroupIDMember> l_directEP = null;
    ImmutableActor          actor2;
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink_2 = null;
    String                  rightType = null;
    NodeRights              nodeRights = null;
    NodeRights              nodeRights2 = null;
    LinkRights              linkRights = null;

    nodeRights = interpreter_.getNodeRightsInstances().get(_act.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(_act);
        nodeRights = interpreter_.getNodeRightsInstances().get(_act.getName());
    } // if

    if (nodeRights.getSelectAddonNewHiddenNodeForActor()) {

        l_dLink = nodeRights.selectAddonNewHiddenNodeForActor(
                _viewInBase, _act, null, _m_l_DisplayableLinks, l_temporaryLinks3);
        _l_ObjectToRegister.addAll(l_temporaryLinks3);

        if (l_dLink != null) {
            for (Object obje : l_dLink) {
                dLink = (DisplayableLinkImpl)obje;
                if (!(dLink.getEorSecondEnd() instanceof ImmutableActor))
                    continue;
                actor2 = (ImmutableActor)dLink.getEorSecondEnd();

                _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                nodeRights2 = interpreter_.getNodeRightsInstances().get(actor2.getName());
                if (nodeRights2 == null) {
                    interpreter_.addInstancesForNode(actor2);
                    nodeRights2 = interpreter_.getNodeRightsInstances().get(actor2.getName());
                } // if

                // searching for
                // actor/any AcsAddon relation/actor started from xid/run under/
                        // account or group or secondary group
                l_directEP = nodeRights2.
                        selectDirectOwnerContainGlobalForActorAsEP(
                                _viewInBase, actor2);
                if (l_directEP != null) {
                    for (Object obj : l_directEP) {
                        gm = (GroupIDMember)obj;

                        rightType = null;
                        // set the XID comment if the first end is an actor and the next node,
                        // an actor which runs under its current user/group or secondary group
                        rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                        if (rightType != null && gm instanceof GroupID)
                          dLink.getEorLastLink().addComment(
                                  rightType, AccessControlLink.C_SET_GROUP);
                        else if (rightType != null && gm instanceof UserID)
                          dLink.getEorLastLink().addComment(
                                  rightType, AccessControlLink.C_SET_USER);

                        dLink_2 = DisplayableLinkImpl.valueOf(
                                dLink, gm, new AccessControlLinkImpl());

                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                        // set the GLOBAL type if the first end is an actor, and the next node
                        // is a secondary group
                        if (gm instanceof GroupID &&
                          actor2.isMemberToSecondaryGroup((GroupID)gm) ) {
                            dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                            dLink_2.getEorLastLink().addComment(
                                      AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                        } // if

                        linkRights = interpreter_.getLinkRightsInstances().get(
                              gm.getName());
                        linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
                    } // for obj
                } // if l_directEP != null

                endsPathsFromActorWithAclBridgePrivilegeAlias(
                        dLink, _viewInBase, _m_l_DisplayableLinks);

            } // for obje
        } // if (l_dLink != null

    } // if node.getSelectAddonNewHiddenNodeForActor()

    l_dLink = null;
    gm = null;
    l_directEP = null;
    actor2 = null;
    dLink = null;
    dLink_2 = null;
    rightType = null;
    nodeRights = null;
    nodeRights2 = null;
    linkRights = null;
    l_temporaryLinks3.clear();
    l_temporaryLinks3 = null;

    return _m_l_DisplayableLinks;
  } // addPathsFromActorAcsAddonRelationActor()

  /**
  * Defines all the 3-nodes access paths when the first end is an Actor with
  * an AcsAddon relation in the first link and no Actor as central node.
  * They are listed hereinafter: <P>
  * - actor/any AcsAddon relation/account or group/acl/resource, <P>
  * - actor/any AcsAddon relation/account or group/pri/
  *         resource or virtual folder, <P>
  * - actor/any AcsAddon relation/account or group/member/group, <P>
  * - actor/any AcsAddon relation/account or group or secondary group/
  *         own or contain/resource, <P>
  * - actor/any AcsAddon relation/account or group/any AcsAddon relation/target, <P>
  * - actor/any AcsAddon relation/virtual folder/virtual member/resource, <P>
  * - actor/any AcsAddon relation/virtual folder/alias/reference, <P>
  * - actor/any AcsAddon relation/virtual folder/any AcsAddon relation/resource. <P>
  * Called by selectOneHiddenNodePathsForActor().
  * Calls endsPathsFromGroupIDMemberWithAclPrivilegeAlias() and
  * endsPathsFromVirtualFolderWithMemberAlias().
  * @param _act the Actor as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromActorAcsAddonRelationNoActor(Actor _act,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    List<DisplayableLinkImpl> l_dLink = null;
    List<ACSObject>           l_temporaryLinks3 = new ArrayList(10);
    DisplayableLinkImpl       dLink = null;
    NodeRights                nodeRights = null;

    nodeRights = interpreter_.getNodeRightsInstances().get(_act.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(_act);
        nodeRights = interpreter_.getNodeRightsInstances().get(_act.getName());
    } // if

    if (nodeRights.getSelectAddonNewHiddenNodeForActor()) {

        l_dLink = nodeRights.selectAddonNewHiddenNodeForActor(
                _viewInBase, _act, null, _m_l_DisplayableLinks, l_temporaryLinks3);
        _l_ObjectToRegister.addAll(l_temporaryLinks3);

        if (l_dLink != null) {
            for (Object obje : l_dLink) {
                dLink = (DisplayableLinkImpl)obje;
                if (dLink.getEorSecondEnd() instanceof Actor)
                    continue;

                _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember) {
                    
                    endsPathsFromGroupIDMemberWithAclPrivilegeAlias(dLink,
                            _viewInBase, _m_l_DisplayableLinks);

                } // if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember)
                else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder) {

                    endsPathsFromVirtualFolderWithMemberAlias(dLink, _viewInBase,
                            _m_l_DisplayableLinks);

                } // if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder)

            } // for obje
        } // if (l_dLink != null

    } // if node.getSelectAddonNewHiddenNodeForActor()

    l_dLink = null;
    dLink = null;
    nodeRights = null;
    l_temporaryLinks3.clear();
    l_temporaryLinks3 = null;

    return _m_l_DisplayableLinks;
  } // addPathsFromActorAcsAddonRelationNoActor()

  /**
  * Defines all the 3-nodes access paths when the first end is an Actor with
  * an alias in the first link. They are listed hereinafter: <P>
  * - actor/alias/ep/pri/resource or virtual folder, <P>
  * - actor/alias/ep/acl/resource, <P>
  * - actorr/alias/ep/any AcsAddon relation/target, <P>
  * - actor/alias/actor/bridge/actor, <P>
  * - actor/alias/ep/alias/reference, <P>
  * - actor/alias/ep/run_under/groupidmember, <P>
  * - actor/alias/actor started from xid/run under/
  *      current account or group or secondary group, <P>
  * - actor/alias/virtual folder/virtual member/resource,
  * - actor/alias/virtual folder/alias/resource, <P>
  * - actor/alias/virtual folder/any AcsAddon relation/target. <P>
  * - actor/alias/virtual folder/run_under/groupidmember, <P>
  * Called by selectOneHiddenNodePathsForActor().
  * Calls endsPathsFromActorWithAclBridgePrivilegeAlias(),
  * endsPathsFromGroupIDMemberWithAclPrivilegeAlias(),
  * endsPathsFromVirtualFolderWithMemberAlias().
  * @param _act the Actor as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister is _gview.getRightsRelatedBaseObjects() to update
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromActorAliasReference(Actor _act,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    // variables
    DisplayableLinkImpl         dLink = null;
    Set<DisplayableLinkImpl>    l_dLink = null;
    Set<ImmutableGroupIDMember> l_members = null;
    LinkRights                  linkRights = interpreter_.getLinkRightsInstances().get(_act.getName());

    l_dLink = compound_.selectDirectHiddenReferenceForAlias(_gview, _viewInBase, _act);
    if (l_dLink != null) {
        for (Object link : l_dLink) {
            dLink = (DisplayableLinkImpl)link;

            _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

            // searching for
            // - actor/alias/ep/pri/resource or virtual folder,
            // - actor/alias/ep/acl/resource,
            // - actorr/alias/ep/any AcsAddon relation/target,
            // - actor/alias/actor/bridge/actor,
            // - actor/alias/ep/alias/reference
            if (dLink.getEorSecondEnd() instanceof Actor)
                endsPathsFromActorWithAclBridgePrivilegeAlias(
                        dLink, _viewInBase, _m_l_DisplayableLinks);
            else if (dLink.getEorSecondEnd() instanceof GroupIDMember)
                endsPathsFromGroupIDMemberWithAclPrivilegeAlias(
                        dLink, _viewInBase, _m_l_DisplayableLinks);

            // searching for
            // - actor/alias/actor started from xid/run under/
            //      current account or group or secondary group
            if (dLink.getEorSecondEnd() instanceof Actor) {

                try {
                    l_members = linkRights.updateAGrunningContext(dLink);
                }
                catch (ProcessError p) {
                    throw new InternalError("ThreeNodesRightsFactoryImpl."+
                            "addPathsFromActorAliasReference(): "+p.toString());
                }
                if (l_members == null) {
                    endsPathsFromActorWithStartingUnderXID(
                            dLink, _viewInBase, _m_l_DisplayableLinks);
                }
            } // if (dLink.getEorSecondEnd() instanceof Actor)

            // searching for
            // - actor/alias/virtual folder/virtual member/resource,
            // - actor/alias/virtual folder/alias/resource,
            // - actor/alias/virtual folder/any AcsAddon relation/target
            if (dLink.getEorSecondEnd() instanceof VirtualFolder) {
                endsPathsFromVirtualFolderWithMemberAlias(
                        dLink, _viewInBase, _m_l_DisplayableLinks);
            }
        } // for link
    } // if (l_dLink != null)

    dLink = null;
    l_dLink = null;
    l_members = null;
    linkRights = null;
    
    return _m_l_DisplayableLinks;
  } // addPathsFromActorAliasReference()

  /**
  * Defines all the 3-nodes access paths when the first end is a GroupIDMember with
  * an own/contain relation in the first link. They are listed hereinafter: <P>
  * - account or group/own or contain/actor/acl/resource, <P>
  * - account or group/own or contain/actor/bridge/actor, <P>
  * - account or group/own or contain/actor/pri/resource or virtual folder, <P>
  * - account or group/own or contain/actor/alias/reference, <P>
  * - account or group/own or contain/actor/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForGroupIDMember().
  * Calls endsPathsFromActorWithAclBridgePrivilegeAlias().
  * @param _gm the GroupIDMember as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromGroupIDMemberOwnContainActor(GroupIDMember _gm,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    // variables
    Set<DisplayableLinkImpl> l_dLink = null;
    ImmutableActor          actor2;
    DisplayableLinkImpl     dLink = null;
    NodeRights              nodeRights = null;
    NodeRights              nodeRights2 = null;

    nodeRights = interpreter_.getNodeRightsInstances().get(_gm.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(_gm);
        nodeRights = interpreter_.getNodeRightsInstances().get(_gm.getName());
    } // if
    
    l_dLink = nodeRights.selectForGroupIDMemberItsHiddenActorsWithNextLinks(_viewInBase, _gm);

    if (l_dLink != null) {
        for (Object obje : l_dLink) {
            dLink = (DisplayableLinkImpl)obje;
            actor2 = (ImmutableActor)dLink.getEorSecondEnd();

            _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

            nodeRights2 = interpreter_.getNodeRightsInstances().get(actor2.getName());
            if (nodeRights2 == null) {
                interpreter_.addInstancesForNode(actor2);
                nodeRights2 = interpreter_.getNodeRightsInstances().get(actor2.getName());
            } // if

            /*
            // no searching for
            // account or group/own or contain/actor started from xid/
                //      run under/current account or group or secondary group
             */

            endsPathsFromActorWithAclBridgePrivilegeAlias(
                    dLink, _viewInBase, _m_l_DisplayableLinks);

        } // for obje
    } // if (l_dLink != null

    l_dLink = null;
    actor2 = null;
    dLink = null;
    nodeRights = null;
    nodeRights2 = null;
    
    return _m_l_DisplayableLinks;
  } // addPathsFromGroupIDMemberOwnContainActor()
  
  /**
  * Defines all the 3-nodes access paths when the first end is a GroupIDMember with
  * an IS_MEMBER relation in the first link. They are listed hereinafter: <P>
  * - groupidmember/member/group/acl/resource, <P>
  * - groupidmember/member/group/pri/resource or virtual folder, <P>
  * - groupidmember/member/group/member/group, <P>
  * - groupidmember/member/group/contain/resource, <P>
  * - groupidmember/member/group/alias/reference, <P>
  * - groupidmember/member/group/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForGroupIDMember().
  * Calls endsPathsFromGroupIDMemberWithAclPrivilegeAlias().
  * @param _gm the GroupIDMember as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromGroupIDMemberIsMemberGroup(GroupIDMember _gm,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    // variables
    ImmutableGroupID[]      l_groups = _gm.getEorL_GroupID();
    DisplayableLinkImpl     dLink = null;

                    // for testing and tracking
                    //System.out.println("start of ThreeNodesRightsFactoryImpl."+
                      //"addPathsFromGroupIDMemberIsMemberGroup() for the member "+
                      //_gm.getName());
    
    if (l_groups.length == 0)
        return _m_l_DisplayableLinks;
    
    for (int i = 0; i < l_groups.length; i++) {
        if (_viewInBase.containsSource(l_groups[i]))
            continue;

                    // for testing and tracking
                    //System.out.println("ThreeNodesRightsFactoryImpl."+
                      //"addPathsFromGroupIDMemberIsMemberGroup() on the _gm group :"+
                      //l_groups[i]);
        
        dLink = new DisplayableLinkImpl(_gm, l_groups[i], new AccessControlLinkImpl());
        dLink.getEorFirstLink().setLinkTypeToTrue(AccessControlLink.IS_MEMBER);

                    // for testing and tracking
                    //System.out.println("in ThreeNodesRightsFactoryImpl."+
                      //"addPathsFromGroupIDMemberIsMemberGroup()"+
                      //", call to endsPathsFromGroupIDMemberWithAclPrivilegeAlias(): "+
                        //"for dLink: "+dLink);

        _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

        endsPathsFromGroupIDMemberWithAclPrivilegeAlias(dLink,
          _viewInBase, _m_l_DisplayableLinks);

    } // for i on l_groups

    dLink = null;
    l_groups = null;
    
    return _m_l_DisplayableLinks;
  } // addPathsFromGroupIDMemberIsMemberGroup()

  /**
  * Defines all the 3-nodes access paths when the first end is a GroupIDMember with
  * an ACL in the first link. They are listed hereinafter: <P>
  * - groupidmember/acl/actor started from xid/run under/current account or group or secondary group, <P>
  * - groupidmember/acl/actor/acl/resource, <P>
  * - groupidmember/acl/actor/bridge/actor, <P>
  * - groupidmember/acl/actor/pri/resource or virtual folder, <P>
  * - groupidmember/acl/actor/alias/reference, <P>
  * - groupidmember/acl/actor/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForGroupIDMember().
  * @param _gm the GroupIDMember as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromGroupIDMemberAclActor(GroupIDMember _gm,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    // variables
    ImmutableResource       res = null;
    GroupIDMember           gm = null;
    Set<ImmutableGroupIDMember> l_directEP = null;
    ImmutableActor          actor2;
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink_2 = null;
    String                  rightType = null;
    NodeRights              nodeRights = null;
    LinkRights              linkRights = null;

    if (_gm.getEorL_AclConnectedResources() == null)
        return _m_l_DisplayableLinks;

    for (Object obje : _gm.getEorL_AclConnectedResources()) {
        res = (ImmutableResource)obje;
        if (!(res instanceof ImmutableActor) || _viewInBase.containsTarget(res))
            continue;
        actor2 = (ImmutableActor)res;

        nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
        if (nodeRights == null) {
            interpreter_.addInstancesForNode(actor2);
            nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
        } // if
        
        dLink = acl_.addAclBridgeRightsInSimpleDisplayableLink( _gm, actor2, true);

        if (dLink != null) {

            _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

            // searching for
            // groupidmember/acl/actor started from xid/run under/
                    // account or group or secondary group
            l_directEP = nodeRights.
                    selectDirectOwnerContainGlobalForActorAsEP(_viewInBase, actor2);
            if (l_directEP != null) {
                for (Object obj : l_directEP) {
                    gm = (GroupIDMember)obj;

                    // continue if the acl transfers the groupidmember context to the actor
                    // and if it is superior to the 'gm' context
                    if (actor2.getAcsName().equals(((ACSObject)dLink.getEorFirstEnd()).getAcsName()) ) {
                        if (gm.getClass() == dLink.getEorFirstEnd().getClass())
                            continue;
                        // the groupidmember, if it is an userID, defines the groupID of actor2
                        else if (gm instanceof GroupID && dLink.getEorFirstEnd() instanceof UserID)
                            continue;
                    } // if (actor2.getAcsName()

                    rightType = null;
                    // set the XID comment if the first end is an actor and the next node,
                    // an actor which runs under its current user/group or secondary group
                    rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                    if (rightType != null && gm instanceof GroupID)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_GROUP);
                    else if (rightType != null && gm instanceof UserID)
                      dLink.getEorLastLink().addComment(
                              rightType, AccessControlLink.C_SET_USER);

                    dLink_2 = DisplayableLinkImpl.valueOf(
                            dLink, gm, new AccessControlLinkImpl());

                    dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                    // set the GLOBAL type if the first end is an actor, and the next node
                    // is a secondary group
                    if (gm instanceof GroupID &&
                      actor2.isMemberToSecondaryGroup((GroupID)gm) ) {
                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                        dLink_2.getEorLastLink().addComment(
                                  AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                    } // if

                    linkRights = interpreter_.getLinkRightsInstances().get(
                              gm.getName());
                    linkRights.addDisplayableLink(
                            dLink_2, _m_l_DisplayableLinks);
                } // for obj
            } // if l_directEP != null

            endsPathsFromActorWithAclBridgePrivilegeAlias(
                    dLink,_viewInBase, _m_l_DisplayableLinks);

        } // if (dLink != null)
    } // if l_hiddenDirectEP != null

    res = null;
    actor2 = null;
    l_directEP = null;
    gm = null;
    dLink = null;
    dLink_2 = null;
    rightType = null;
    nodeRights = null;
    linkRights = null;

    return _m_l_DisplayableLinks;
  } // addPathsFromGroupIDMemberAclActor()
  
  /**
  * Defines all the 3-nodes access paths when the first end is a GroupIDMember with
  * an AcsAddon relation in the first link and an Actor as central node.
  * They are listed hereinafter: <P>
  * - groupidmember/any AcsAddon relation/actor/acl/resource, <P>
  * - groupidmember/any AcsAddon relation/actor/bridge/actor, <P>
  * - groupidmember/any AcsAddon relation/actor/pri/resource or virtual folder, <P>
  * - groupidmember/any AcsAddon relation/actor/
  *         actor started from xid/run under/current account or group or secondary group, <P>
  * - groupidmember/any AcsAddon relation/actor/alias/reference. <P>
  * - groupidmember/any AcsAddon relation/actor/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForGroupIDMember().
  * Calls endsPathsFromActorWithAclBridgePrivilegeAlias().
  * @param _gm the GroupIDMember as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromGroupIDMemberAcsAddonRelationActor(GroupIDMember _gm,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    List<DisplayableLinkImpl> l_dLink = null;
    List<ACSObject>           l_temporaryLinks3 = new ArrayList(10);
    GroupIDMember           gm = null;
    Set<ImmutableGroupIDMember> l_directEP = null;
    ImmutableActor          actor2;
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink_2 = null;
    String                  rightType = null;
    NodeRights              nodeRights = null;
    NodeRights              nodeRights2 = null;
    LinkRights              linkRights = null;

    nodeRights = interpreter_.getNodeRightsInstances().get(_gm.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(_gm);
        nodeRights = interpreter_.getNodeRightsInstances().get(_gm.getName());
    } // if

    if (nodeRights.getSelectAddonNewHiddenNodeForGroupIDMember()) {

        l_dLink = nodeRights.selectAddonNewHiddenNodeForGroupIDMember(
                _viewInBase, _gm, null, _m_l_DisplayableLinks, l_temporaryLinks3);
        _l_ObjectToRegister.addAll(l_temporaryLinks3);

        if (l_dLink != null) {
            for (Object obje : l_dLink) {
                dLink = (DisplayableLinkImpl)obje;
                if (!(dLink.getEorSecondEnd() instanceof ImmutableActor))
                    continue;
                actor2 = (ImmutableActor)dLink.getEorSecondEnd();

                _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                nodeRights2 = interpreter_.getNodeRightsInstances().get(actor2.getName());
                if (nodeRights2 == null) {
                    interpreter_.addInstancesForNode(actor2);
                    nodeRights2 = interpreter_.getNodeRightsInstances().get(actor2.getName());
                } // if

                // searching for
                // groupidmember/any AcsAddon relation/actor started from xid/run under/
                        // account or group or secondary group
                l_directEP = nodeRights2.
                        selectDirectOwnerContainGlobalForActorAsEP(_viewInBase, actor2);
                if (l_directEP != null) {
                    for (Object obj : l_directEP) {
                        gm = (GroupIDMember)obj;

                        if (gm == dLink.getEorFirstEnd())
                            continue;

                        rightType = null;
                        // set the XID comment if the first end is an actor and the next node,
                        // an actor which runs under its current user/group or secondary group
                        rightType = dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

                        if (rightType != null && gm instanceof GroupID)
                          dLink.getEorLastLink().addComment(
                                  rightType, AccessControlLink.C_SET_GROUP);
                        else if (rightType != null && gm instanceof UserID)
                          dLink.getEorLastLink().addComment(
                                  rightType, AccessControlLink.C_SET_USER);

                        dLink_2 = DisplayableLinkImpl.valueOf(
                                dLink, gm, new AccessControlLinkImpl());

                        dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
                        // set the GLOBAL type if the first end is an actor, and the next node
                        // is a secondary group
                        if (gm instanceof GroupID &&
                          actor2.isMemberToSecondaryGroup((GroupID)gm) ) {
                            dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                            dLink_2.getEorLastLink().addComment(
                                      AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
                        } // if

                        linkRights = interpreter_.getLinkRightsInstances().get(
                              gm.getName());
                        linkRights.addDisplayableLink(
                                dLink_2, _m_l_DisplayableLinks);
                    } // for obj
                } // if l_directEP != null

                endsPathsFromActorWithAclBridgePrivilegeAlias(
                        dLink, _viewInBase, _m_l_DisplayableLinks);

            } // for obje
        } // if (l_dLink != null

    } // if node.getSelectAddonNewHiddenNodeForGroupIDMember()

    l_dLink = null;
    gm = null;
    l_directEP = null;
    actor2 = null;
    dLink = null;
    dLink_2 = null;
    rightType = null;
    nodeRights = null;
    nodeRights2 = null;
    linkRights = null;
    l_temporaryLinks3.clear();
    l_temporaryLinks3 = null;

    return _m_l_DisplayableLinks;
  } // addPathsFromGroupIDMemberAcsAddonRelationActor()

  /**
  * Defines all the 3-nodes access paths when the first end is a GroupIDMember with
  * an AcsAddon relation in the first link and no Actor as central node.
  * They are listed hereinafter: <P>
  * - groupidmember/any AcsAddon relation/account or group/own or contain/resource, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/member/group, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/acl/resource, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/pri/resource or virtual folder, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/alias/rereference, <P>
  * - groupidmember/any AcsAddon relation/groupidmember/any AcsAddon relation/target, <P>
  * - groupidmember/any AcsAddon relation/virtual folder/has member/resource, <P>
  * - groupidmember/any AcsAddon relation/virtual folder/alias/reference, <P>
  * - groupidmember/any AcsAddon relation/virtual folder/any AcsAddon relation/resource. <P>
  * Called by selectOneHiddenNodePathsForGroupIDMember().
  * @param _gm the GroupIDMember as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromGroupIDMemberAcsAddonRelationNoActor(GroupIDMember _gm,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    List<DisplayableLinkImpl> l_dLink = null;
    List<ACSObject>           l_temporaryLinks3 = new ArrayList(10);
    DisplayableLinkImpl       dLink = null;
    NodeRights                nodeRights = null;

    nodeRights = interpreter_.getNodeRightsInstances().get(_gm.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(_gm);
        nodeRights = interpreter_.getNodeRightsInstances().get(_gm.getName());
    } // if

    if (nodeRights.getSelectAddonNewHiddenNodeForGroupIDMember()) {

        l_dLink = nodeRights.selectAddonNewHiddenNodeForGroupIDMember(
                _viewInBase, _gm, null, _m_l_DisplayableLinks, l_temporaryLinks3);
        _l_ObjectToRegister.addAll(l_temporaryLinks3);

        if (l_dLink != null) {
            for (Object obje : l_dLink) {
                dLink = (DisplayableLinkImpl)obje;
                if (dLink.getEorSecondEnd() instanceof Actor)
                    continue;

                _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

                if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember) {

                    endsPathsFromGroupIDMemberWithAclPrivilegeAlias(dLink,
                            _viewInBase, _m_l_DisplayableLinks);

                } // if (dLink.getEorSecondEnd() instanceof ImmutableGroupIDMember)
                else if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder) {
                    
                    endsPathsFromVirtualFolderWithMemberAlias(dLink,
                            _viewInBase, _m_l_DisplayableLinks);

                } // if (dLink.getEorSecondEnd() instanceof ImmutableVirtualFolder)

            } // for obje
        } // if (l_dLink != null

    } // if node.getSelectAddonNewHiddenNodeForActor()

    l_dLink = null;
    dLink = null;
    nodeRights = null;
    l_temporaryLinks3.clear();
    l_temporaryLinks3 = null;

    return _m_l_DisplayableLinks;
  } // addPathsFromGroupIDMemberAcsAddonRelationNoActor()

  /**
  * Defines all the 3-nodes access paths when the first end is a GroupIDMember with
  * an alias in the first link. They are listed hereinafter: <P>
  * - groupidmember/alias/ep/pri/resource or virtual folder, <P>
  * - groupidmember/alias/ep/acl/resource, <P>
  * - groupidmember/alias/ep/any AcsAddon relation/target, <P>
  * - groupidmember/alias/actor/bridge/actor, <P>
  * - groupidmember/alias/ep/alias/reference, <P>
  * - groupidmember/alias/actor started from xid/run under/
  *      current account or group or secondary group, <P>
  * - groupidmember/alias/virtual folder/virtual member/resource,
  * - groupidmember/alias/virtual folder/alias/resource, <P>
  * - groupidmember/alias/virtual folder/any AcsAddon relation/target. <P>
  * Called by selectOneHiddenNodePathsForGroupIDMember().
  * Calls endsPathsFromActorWithAclBridgePrivilegeAlias(), endsPathsFromGroupIDMemberWithAclPrivilegeAlias(),
  * endsPathsFromVirtualFolderWithMemberAlias().
  * @param _gm the GroupIDMember as first end of the access paths to find
  * @param _gview is the graphic view to update
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @param _l_ObjectToRegister has to be updated for _gview.getRightsRelatedBaseObjects()
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map addPathsFromGroupIDMemberAliasReference(GroupIDMember _gm,
          GraphicView _gview, EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks,
          HashSet _l_ObjectToRegister) throws InterruptedException {

    // variables
    DisplayableLinkImpl         dLink = null;
    Set<DisplayableLinkImpl>    l_dLink = null;
    Set<ImmutableGroupIDMember> l_members = null;
    LinkRights                  linkRights = interpreter_.getLinkRightsInstances().get(_gm.getName());

    l_dLink = compound_.selectDirectHiddenReferenceForAlias(_gview, _viewInBase, _gm);
    if (l_dLink != null) {
        for (Object link : l_dLink) {
            dLink = (DisplayableLinkImpl)link;

            _l_ObjectToRegister.add( dLink.getEorSecondEnd() );

            // searching for
            // - groupidmember/alias/ep/pri/resource or virtual folder,
            // - groupidmember/alias/ep/acl/resource,
            // - groupidmember/alias/ep/any AcsAddon relation/target,
            // - groupidmember/alias/actor/bridge/actor,
            // - groupidmember/alias/ep/alias/reference
            if (dLink.getEorSecondEnd() instanceof Actor)
                endsPathsFromActorWithAclBridgePrivilegeAlias(dLink, _viewInBase, _m_l_DisplayableLinks);
            else if (dLink.getEorSecondEnd() instanceof GroupIDMember)
                endsPathsFromGroupIDMemberWithAclPrivilegeAlias(dLink, _viewInBase, _m_l_DisplayableLinks);

            // searching for
            // - groupidmember/alias/actor started from xid/run under/
            //      current account or group or secondary group
            if (dLink.getEorSecondEnd() instanceof Actor) {

                try {
                    l_members = linkRights.updateAGrunningContext(dLink);
                }
                catch (ProcessError p) {
                    throw new InternalError("ThreeNodesRightsFactoryImpl."+
                            "addPathsFromGroupIDMemberAliasReference(): "+p.toString());
                }
                if (l_members == null) {
                    endsPathsFromActorWithStartingUnderXID(dLink, _viewInBase, _m_l_DisplayableLinks);
                }
            } // if (dLink.getEorSecondEnd() instanceof Actor)

            // searching for
            // - groupidmember/alias/virtual folder/virtual member/resource,
            // - groupidmember/alias/virtual folder/alias/resource,
            // - groupidmember/alias/virtual folder/any AcsAddon relation/target
            if (dLink.getEorSecondEnd() instanceof VirtualFolder) {
                endsPathsFromVirtualFolderWithMemberAlias(dLink, _viewInBase, _m_l_DisplayableLinks);
            }
        } // for link
    } // if (l_dLink != null)

    dLink = null;
    l_dLink = null;
    l_members = null;
    linkRights = null;
    
    return _m_l_DisplayableLinks;
  } // addPathsFromGroupIDMemberAliasReference()

  /**
  * Ends some 3-nodes access paths from a simple DisplayableLinkImpl where
  * the second end is a VirtualFolder.
  * The return paths are listed hereinafter: <P>
  * - XX/YY/virtual folder/virtual member/resource <P>
  * - XX/YY/virtual folder/alias/reference <P>
  * - XX/YY/virtual folder/any AcsAddon relation/resource <P>
  * Called by the method selectOneHiddenNodePathsForFirstLinkPrivileges(),
  * selectOneHiddenNodePathsForVirtualFolder(),
  * addPathsFromActorAcsAddonRelationNoActor().
  * @param _dLink simple link with a VirtualFolder as second end
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map endsPathsFromVirtualFolderWithMemberAlias(DisplayableLinkImpl _dLink,
      EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks) throws InterruptedException {

    // variables
    VirtualFolder           vf = (VirtualFolder) _dLink.getEorSecondEnd();
    ImmutableResource       res;
    ImmutableTarget         targ;
    DisplayableLinkImpl     dLink = null;
    DisplayableLinkImpl     dLink_2 = null;
    NodeRights              nodeRights = null;
    Set<DisplayableLinkImpl> l_temporaryLinks2 = null;
    LinkRights              linkRights = null;
    //Set<ImmutableGroupIDMember> l_members = null;
    //int                     nextIndex = -1;

    nodeRights = interpreter_.getNodeRightsInstances().get(vf.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(vf);
        nodeRights = interpreter_.getNodeRightsInstances().get(vf.getName());
    } // if

    // searching for
    // XX/YY/virtual folder/virtual member/resource
    for (Object obj : vf.getEorM_BaseObjects().values() ) {
        res = (ImmutableResource)obj;
        if (!_viewInBase.containsTarget(res)) continue;
        if (!res.getEorACS().getManageVirtualFolderRightsPropagationToMembers())
            continue;

                                // for testing and tracking
                                //System.out.println("In ThreeNodesRightsFactoryImpl."+
                                  //"endsPathsFromVirtualFolderWithMemberAlias(), "+
                                  //"detection of XX/YY/virtual folder/virtual member/resource"+
                                  //" for the vf "+vf.getName()+
                                  //" and the resource "+res.getName());


        dLink = DisplayableLinkImpl.valueOf(
                _dLink, res, new AccessControlLinkImpl());

        linkRights = interpreter_.getLinkRightsInstances().get(
                              res.getName());
        dLink_2 = linkRights.addDisplayableLink(
                        dLink, _m_l_DisplayableLinks);

        if (dLink_2 == null) continue;

        // copy of the previous link rights in the virtual_member link
        dLink_2.getEorLastLink().setRights(
            AccessControlLink.HAS_VIRTUAL_MEMBER,
            (StringRight[]) _dLink.getEorFirstLink().getRedondantRights().
                toArray(new StringRight[0]));
    } // for obj

    // searching for
    // XX/YY/virtual folder/alias/reference
    if (vf.getEorACS().getManageInternalHardAlias() ||
            _viewInBase.getConnectedToSeveralACS()) {

        l_temporaryLinks2 = compound_.selectDirectReferencesInViewForAlias(_viewInBase, vf);
        if (l_temporaryLinks2 != null) {
            for (Object ob : l_temporaryLinks2) {
                dLink = linkUtil_.mergeLinks(_dLink, (DisplayableLinkImpl)ob);

                linkRights = interpreter_.getLinkRightsInstances().get(
                              ((DisplayableLinkImpl)ob).getEorSecondEnd().getName());
                linkRights.addDisplayableLink(
                        dLink, _m_l_DisplayableLinks);
            } // for ob
        } // if (l_temporaryLinks2 != null)
    } // if (vf.getManageInternalHardAlias()

    // searching for
    // XX/YY/virtual folder/any AcsAddon relation/resource
    if (nodeRights.getSelectAddonLastViewNodeAfterVirtualFolder()) {

            for (Iterator itu = _viewInBase.getEorM_Targets().values().
                    iterator(); itu.hasNext(); ) {
                targ = (ImmutableTarget)itu.next();
                if (targ == vf || targ == _dLink.getEorFirstEnd()) continue;
                if (!(targ instanceof ImmutableResource))
                    continue;

                dLink = nodeRights.selectAddonLastViewNodeAfterVirtualFolder(
                    _viewInBase, (ImmutableResource)targ, _dLink, _m_l_DisplayableLinks);

                if (dLink != null) {
                    linkRights = interpreter_.getLinkRightsInstances().get(
                              targ.getName());
                    linkRights.addDisplayableLink(
                                dLink, _m_l_DisplayableLinks);
                } // if (dLink != null
            } // for itu

    } // if node.getSelectAddonNewHiddenNodeForGroupIDMember()

                    // for testing and tracking
                    //System.out.println("In ThreeNodesRightsFactoryImpl."+
                      //"endsPathsFromVirtualFolderWithMemberAlias(), "+
                      //"start the detection of XX/YY/virtual folder/virtual member/resource"+
                      //" for the vf "+bo.getName());

    vf = null;
    res = null;
    targ = null;
    dLink = null;
    dLink_2 = null;
    nodeRights = null;
    l_temporaryLinks2 = null;
    linkRights = null;
    //l_members = null;

    return _m_l_DisplayableLinks;
  } // endsPathsFromVirtualFolderWithMemberAlias()

  /**
  * Ends some 3-nodes access paths from a simple DisplayableLinkImpl where
  * the second end is an Actor. The return paths are listed hereinafter: <P>
  * - XX/YY/actor/bridge/actor, <P>
  * - XX/YY/actor/acl/resource, <P>
  * - XX/YY/actor/pri/resource or virtual folder, <P>
  * - XX/YY/actor/alias/reference, <P>
  * - XX/YY/actor/any AcsAddon relation/target. <P>
  * Called by the methods addPathsFromActorXXX() and addPathsFromGroupIDMemberXXX().
  * @param _dLink simple link with an Actor as second end
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map endsPathsFromActorWithAclBridgePrivilegeAlias(DisplayableLinkImpl _dLink,
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks) throws InterruptedException {

    // variables
    GroupIDMember           gm = null;
    ImmutableTarget         targ;
    ImmutableActor          actor2 = (ImmutableActor)_dLink.getEorSecondEnd();
    DisplayableLinkImpl     dLink_2 = null;
    AccessControlLinkImpl   acLink = null;
    NodeRights              nodeRights = null;
    Set<DisplayableLinkImpl>    l_temporaryLinks2 = null;
    LinkRights              linkRights = null;
    //Set<ImmutableGroupIDMember> l_members = null;
    //int                     nextIndex = -1;

    nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
    if (nodeRights == null) {
        interpreter_.addInstancesForNode(actor2);
        nodeRights = interpreter_.getNodeRightsInstances().get(actor2.getName());
    } // if

    // adding the AG context get through an alias relation
    //if (_dLink.getEorLastLink().getLinkTypeValue(AccessControlLink.IS_ALIAS)) {

      //  l_members = linkUtil_.getSecondEndAGcontext(_dLink);
        //if (l_members != null && _dLink.getProperties() != null &&
          //          _dLink.getProperties().get(
            //            DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION) != null) {

            //nextIndex = ((Integer)_dLink.getProperties().get(
              //          DisplayableLinkImpl.AG_CONTEXT_NODE_POSITION)).
                //        intValue();
            // nextIndex has to be the index of the node just before the second end
            //if (nextIndex == _dLink.noOfNodes() - 2) {
              //  for (Object memb : l_members) {
                //    if (_dLink.getEorFirstEnd() == memb ||
                  //              _viewInBase.containsBaseObject((BaseObject)memb)) continue;

                    //if (memb instanceof GroupID) {
                        // noop if already exists
                      //  _dLink.getEorLastLink().addComment(
                        //        AccessControlLink.IS_ALIAS,
                          //      AccessControlLink.C_SET_GROUP + ": "+
                            //        ((GroupIDMember)memb).getNickName());
                    //}
                    //else if (memb instanceof UserID) {
                        // noop if already exists
                      //  _dLink.getEorLastLink().addComment(
                        //        AccessControlLink.IS_ALIAS,
                          //      AccessControlLink.C_SET_USER + ": "+
                            //        ((GroupIDMember)memb).getNickName());
                    //} // else
                //} // for
            //} // if (nextIndex == _dLink.noOfNodes() - 2)
        //} // if (l_members != null
    //} // if (_dLink.getEorLastLink().getLinkTypeValue(AccessControlLink.IS_ALIAS))

    // searching for
    // XX/YY/actor/bridge/actor
    if (actor2.getIsACSBridgeSource() &&
            _viewInBase.containsSource(actor2.getEorACSBridgeTarget())) {

        dLink_2 = DisplayableLinkImpl.valueOf(
                _dLink, actor2.getEorACSBridgeTarget(), new AccessControlLinkImpl());
        dLink_2.getEorLastLink().addRight(
            AccessControlLink.BRIDGE,
            BaseManagerImpl.getInstance().getMetaRight("execute") );

        linkRights = interpreter_.getLinkRightsInstances().get(
                              actor2.getEorACSBridgeTarget().getName());
        linkRights.addDisplayableLink(
                    dLink_2, _m_l_DisplayableLinks);
    } // if (actor2.getIsACSBridgeSource()

    // searching for
    // XX/YY/actor/alias/reference
    if (actor2.getEorACS().getManageInternalHardAlias() ||
            _viewInBase.getConnectedToSeveralACS()) {

        l_temporaryLinks2 = compound_.selectDirectReferencesInViewForAlias(_viewInBase, actor2);
        if (l_temporaryLinks2 != null) {
            for (Object ob : l_temporaryLinks2) {
                dLink_2 = linkUtil_.mergeLinks(_dLink, (DisplayableLinkImpl)ob);

                linkRights = interpreter_.getLinkRightsInstances().get(
                              ((DisplayableLinkImpl)ob).getEorSecondEnd().getName());
                linkRights.addDisplayableLink(
                        dLink_2, _m_l_DisplayableLinks);
            } // for ob
        } // if (l_temporaryLinks2 != null)
    } // if (actor2.getManageInternalHardAlias()

    for (Iterator iti = _viewInBase.getEorM_Targets().values().iterator(); iti.hasNext() ; ) {
        targ = (ImmutableTarget) iti.next();
        if (targ == _dLink.getEorFirstEnd()) continue;

                    // for testing and tracking
                    //System.out.println("in ThreeNodesRightsFactoryImpl."+
                      //"addOneHiddenNodePathsForActor()"+
                      //", loop on the target: "+targ.getName());

        // searching for
        // XX/YY/actor/acl/resource
        if (targ instanceof ImmutableResource) {
            acLink = acl_.addAclInAccessControlLink((ImmutableResource)targ, actor2, null);

            if (acLink != null) {
                dLink_2 = DisplayableLinkImpl.valueOf(_dLink, targ, acLink);

                linkRights = interpreter_.getLinkRightsInstances().get(
                              targ.getName());
                linkRights.addDisplayableLink(
                            dLink_2,
                            _m_l_DisplayableLinks);
            } // if (acLink != null
        } // if (targ instanceof Resource)

        // searching for
        // XX/YY/actor/pri/resource or virtual folder
        if (targ instanceof ImmutablePrivilegeTarget) {
            acLink = pri_.addPrivilegeInAccessControlLink(
                        (ImmutablePrivilegeTarget)targ, actor2, null);

            if (acLink != null) {
                dLink_2 = DisplayableLinkImpl.valueOf(_dLink, targ, acLink);
                linkRights = interpreter_.getLinkRightsInstances().get(
                              targ.getName());
                linkRights.addDisplayableLink(
                            dLink_2,
                            _m_l_DisplayableLinks);
            } // if (acLink != null)
        } // if (targ instanceof ImmutablePrivilegeTarget)

        // searching for
        // XX/YY/actor/any AcsAddon relation/target
        if (nodeRights.getSelectAddonLastViewNodeAfterActor()) {

            dLink_2 = nodeRights.selectAddonLastViewNodeAfterActor(
                _viewInBase, targ, _dLink, _m_l_DisplayableLinks);

            if (dLink_2 != null) {
                linkRights = interpreter_.getLinkRightsInstances().get(
                              targ.getName());
                linkRights.addDisplayableLink(
                            dLink_2, _m_l_DisplayableLinks);
            } // if (dLink_2 != null

        } // if node.getSelectAddonNewHiddenNodeForActor()

    } // for iti on Targets

    gm = null;
    targ = null;
    actor2 = null;
    dLink_2 = null;
    acLink = null;
    nodeRights = null;
    l_temporaryLinks2 = null;
    linkRights = null;
    //l_members = null;
    
    return _m_l_DisplayableLinks;
  } // endsPathsFromActorWithAclBridgePrivilegeAlias()

  /**
  * Ends some 3-nodes access paths from a simple DisplayableLinkImpl where
  * the second end is an Actor without AG context from the first end.
  * The return paths are listed hereinafter: <P>
  * - XX/YY/actor started from xid/run under/
  * current account or group or secondary group, <P>
  * Called by the methods addPathsFromActorXXX() and addPathsFromGroupIDMemberXXX().
  * @param _dLink simple link with an Actor as second end
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map endsPathsFromActorWithStartingUnderXID(DisplayableLinkImpl _dLink,
      EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks) throws InterruptedException {
      // variables
    Actor                   actor = (Actor)_dLink.getEorSecondEnd();
    GroupIDMember           gm = null;
    Set<ImmutableGroupIDMember>  l_directEP = null;
    DisplayableLinkImpl     dLink_2 = null;
    String                  rightType = null;
    NodeRights              nodeRights2 = null;
    LinkRights              linkRights = null;
    //Set<ImmutableGroupIDMember> l_members = null;
    //int                     nextIndex = -1;

    nodeRights2 = interpreter_.getNodeRightsInstances().get(actor.getName());
    if (nodeRights2 == null) {
        interpreter_.addInstancesForNode(actor);
        nodeRights2 = interpreter_.getNodeRightsInstances().get(actor.getName());
    } // if

    l_directEP = nodeRights2.
      selectDirectOwnerContainGlobalForActorAsEP(_viewInBase, actor);
    if (l_directEP != null) {
        for (Object obje : l_directEP) {
            gm = (GroupIDMember)obje;

            //priority = false;
            //if (l_members != null) {
              //  for (Object obj : l_members) {
                //    if (gm.getClass() == obj.getClass()) {
                  //      priority = true;
                    //    break;
                    //}
                //} // for obj
            //} // if (l_members != null
            //if (priority) continue;

            rightType = null;
            // set the XID comment if the first end is an actor and the next node,
            // an actor which runs under its current user/group or secondary group
            rightType = _dLink.getEorLastLink().getL_AllActivatedTypes().get(0);

            if (rightType != null && gm instanceof GroupID)
              _dLink.getEorLastLink().addComment(
                      rightType, AccessControlLink.C_SET_GROUP);
            else if (rightType != null && gm instanceof UserID)
              _dLink.getEorLastLink().addComment(
                      rightType, AccessControlLink.C_SET_USER);

            dLink_2 = DisplayableLinkImpl.valueOf(
                    _dLink, gm, new AccessControlLinkImpl());

            dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.RUN_UNDER);
            // set the GLOBAL type if the first end is an actor, and the next node
            // is a secpndary group
            if (gm instanceof GroupID &&
              actor.isMemberToSecondaryGroup((GroupID)gm) ) {
                dLink_2.getEorLastLink().setLinkTypeToTrue(AccessControlLink.GLOBAL);
                dLink_2.getEorLastLink().addComment(
                          AccessControlLink.GLOBAL, AccessControlLink.C_SECONDARY_GROUP);
            } // if

            linkRights = interpreter_.getLinkRightsInstances().get(
                              gm.getName());
            linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
        } // for obje
    } // if l_directEP != null

    actor = null;
    gm = null;
    l_directEP = null;
    dLink_2 = null;
    rightType = null;
    nodeRights2 = null;
    linkRights = null;
    //l_members = null;

    return _m_l_DisplayableLinks;
  } // endsPathsFromActorWithStartingUnderXID

  /**
  * Ends some 3-nodes access paths from a simple DisplayableLinkImpl where
  * the second end is a GroupIDMember. The return paths are listed hereinafter: <P>
  * - XX/YY/groupidmember/acl/resource, <P>
  * - XX/YY/groupidmember/pri/resource or virtual folder, <P>
  * - XX/YY/groupidmember/member/group, <P>
  * - XX/YY/account or group/own or contain/resource, <P>
  * - XX/YY/groupidmember/alias/reference, <P>
  * - XX/YY/groupidmember/any AcsAddon relation/target. <P>
  * Called by the methods addPathsFromActorXXX() and addPathsFromGroupIDMemberXXX().
  * Calls NodeRights.withAccessThroughNodesTreeFromEP(),
  * detectAccountPriorityInLastLinkAGORights(),
  * NodeRights.selectAddonLastViewNodeAfterGroupIDMember() and
  * AgoRightsFactoryImpl.addOwnerContainRightsToDisplayableLink().
  * @param _dLink simple link with a GroupIDMember as second end
  * @param _viewInBase EPRViewInBase to analyze; not always the view of _gview
  * @param _m_l_DisplayableLinks Map of DisplayableLinks lists (one per pair)
  * associated to the view, and to update
  * @exception InterruptedException
  * @return updated _m_l_DisplayableLinks
   */
  public final Map endsPathsFromGroupIDMemberWithAclPrivilegeAlias(
          DisplayableLinkImpl _dLink,
          EPRViewInBase _viewInBase, Map _m_l_DisplayableLinks) throws InterruptedException {

    // variables
    GroupIDMember           gm = (GroupIDMember)_dLink.getEorSecondEnd();

                          // for testing and tracking
                          //System.out.println("Call to ThreeNodesRightsFactoryImpl."+
                            //"endsPathsFromGroupIDMemberWithAclPrivilegeAlias() "+
                            //"for the first end "+_dLink.getEorFirstEnd()+
                            //"and the second end GroupIDMember "+gm.getName());

    ImmutableTarget         targ;
    GroupID                 gp = null;
    ImmutableSource         so = null;
    DisplayableLinkImpl     dLink_2 = null;
    AccessControlLinkImpl   acLink = null;
    Name                    pairKey = null;
    char                    open = '[';
    char                    close = ']';
    char                    space = ' ';
    boolean                 priority = false;
    LinkRights              linkRights = null;
    NodeRights              nodeRights = null;
    NodeRights              nodeRights2 = null;
    Set<DisplayableLinkImpl> l_temporaryLinks2 = null;
    //Set<ImmutableGroupIDMember> l_members = null;
    //int                     nextIndex = -1;

                          // for testing and tracking
                          //System.out.println("In ThreeNodesRightsFactoryImpl.endsPathsFromGroupIDMemberWithAclPrivilegeAlias() "+
                            //"'nodeRights' has the class "+nodeRights.getClass());
                          //System.out.println(
                            //      "nodeRights.getSelectAddonLastViewNodeAfterGroupIDMember(): "+
                            //nodeRights.getSelectAddonLastViewNodeAfterGroupIDMember());

    nodeRights2 = interpreter_.getNodeRightsInstances().get(gm.getName());
    if (nodeRights2 == null) {
        interpreter_.addInstancesForNode(gm);
        nodeRights2 = interpreter_.getNodeRightsInstances().get(gm.getName());
    } // if
    
    for (Iterator iti = _viewInBase.getEorM_Targets().values().iterator(); iti.hasNext() ; ) {
        targ = (ImmutableTarget) iti.next();

                                  // for testing and tracking
                                  //System.out.println("In ThreeNodesRightsFactoryImpl."+
                                    //    "endsPathsFromGroupIDMemberWithAclPrivilegeAlias(), "+
                                      //  "iteration on the target "+targ.getName());

        if (targ == _dLink.getEorFirstEnd()) continue;

        linkRights = interpreter_.getLinkRightsInstances().get(targ.getName());
        nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());
        //if (linkRights == null) {
          //  interpreter_.addInstancesForNode(targ);
            //linkRights = interpreter_.getLinkRightsInstances().get(targ.getName());
            //nodeRights = interpreter_.getNodeRightsInstances().get(targ.getName());
        //} // if

        if (targ instanceof ImmutableResource) {

                                          // for testing and tracking
                                          //System.out.println("In ThreeNodesRightsFactoryImpl."+
                                            //      "endsPathsFromGroupIDMemberWithAclPrivilegeAlias()"+
                                            //"for the target: "+targ.getName());

            // searching for
            // XX/YY/groupidmember/acl/resource
            acLink = acl_.addAclInAccessControlLink((ImmutableResource)targ, gm, null);

            if (acLink != null) {
                dLink_2 = DisplayableLinkImpl.valueOf(_dLink, targ, acLink);

                                          // for testing and tracking
                                          //System.out.println("In ThreeNodesRightsFactoryImpl."+
                                            //"endsPathsFromGroupIDMemberWithAclPrivilegeAlias(), "+
                                            //"for the target, new link: "+dLink_2);


                linkRights.addDisplayableLink(
                            dLink_2,
                            _m_l_DisplayableLinks);

                            // for testing and tracking
                            //ImmutableName key;
                            //DisplayableLinkImpl dLink;
                            //System.out.println(" ");
                            //System.out.println("ThreeNodesRightsFactoryImpl."+
                              //      "endsPathsFromGroupIDMemberWithAclPrivilegeAlias(), "+
                              //"the DisplayableLinkImpls after XX/YY/groupidmember/acl/resource are:");
                            //for (Iterator ite = _m_l_DisplayableLinks.keySet().iterator() ; ite.hasNext() ; ) {
                              // key = (ImmutableName) ite.next();
                               //System.out.println("--- for key: "+key);
                               //for (Iterator ita = ((List)_m_l_DisplayableLinks.get(key)).iterator() ; ita.hasNext() ; ) {
                                 //  dLink = (DisplayableLinkImpl)ita.next();
                                   //System.out.println("------------"+
                                     //   dLink.toString() );
                                   //System.out.println("-------------------- last link:"+
                                     //   dLink.getEorLastLink().toString() );
                               //} // for ita
                            //} // for ite


            } // if (acLink != null

            // searching for
            // XX/YY/account/own/resource
            if (((ImmutableResource)targ).getEorUserID() == gm ) {

                if (nodeRights.withAccessThroughNodesTreeFromEP(
                                (ImmutableUserID)gm, (ImmutableResource)targ, null)) {

                    dLink_2 = DisplayableLinkImpl.valueOf(
                                _dLink, targ, new AccessControlLinkImpl());

                    dLink_2 = ago_.
                            addOwnerContainRightsToDisplayableLink(dLink_2);

                    if (dLink_2 != null)
                        linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
                } // if (utility_.withAccessThroughNodesTree(
            } // (((ImmutableResource)targ).getEorUserID() != null

            // searching for
            // XX/YY/group/contain/resource
            if (((ImmutableResource)targ).getEorGroupID() == gm ) {

                if (nodeRights.withAccessThroughNodesTreeFromEP(
                                (ImmutableGroupID)gm, (ImmutableResource)targ, null)) {

                    dLink_2 = DisplayableLinkImpl.valueOf(
                                _dLink, targ, new AccessControlLinkImpl());

                    priority = false;
                    pairKey = ((Name)((BaseObject)_dLink.getEorFirstEnd()).
                          getName()).extended(targ.getName());
                    if (_m_l_DisplayableLinks.get(pairKey) != null) {
                        // ACS rule: 'AGO right: A rights overlay G rights'
                        // No new dlink_2 if the userID rights are already applicable
                        priority = linkRights.detectAccountPriorityInLastLinkAGORights(
                                                dLink_2,
                                                (List)_m_l_DisplayableLinks.get(pairKey));
                        if (!priority &&
                                linkRights.getDetectAddonPriorityInLastLinksAGORights()) {
                            priority = linkRights.detectAddonPriorityInLastLinkAGORights(
                                                    dLink_2,
                                                    (List)_m_l_DisplayableLinks.get(pairKey));
                        } // if (!priority
                    } // if (_m_l_DisplayableLinks.get(pairKey)

                    if (!priority) {
                        dLink_2 = ago_.
                                addOwnerContainRightsToDisplayableLink(dLink_2);

                        if (dLink_2 != null)
                            linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
                    } // if (!priority
                } // if (utility_.withAccessThroughNodesTree(

            } // (((ImmutableResource)targ).getEorGroupID() != null

        } // if (targ instanceof ImmutableResource)

        // searching for
        // XX/YY/account or group/pri/resource or virtual folder
        if (targ instanceof ImmutablePrivilegeTarget) {
            acLink = pri_.addPrivilegeInAccessControlLink(
                        (ImmutablePrivilegeTarget)targ, gm, null);

            if (acLink != null) {
                dLink_2 = DisplayableLinkImpl.valueOf(_dLink, targ, acLink);
                linkRights.addDisplayableLink(
                            dLink_2,
                            _m_l_DisplayableLinks);
            } // if (acLink != null)
        } // if (targ instanceof ImmutablePrivilegeTarget)

        // searching for
        // XX/YY/account or group/any AcsAddon relation/target
        if (nodeRights2.getSelectAddonLastViewNodeAfterGroupIDMember()) {

            dLink_2 = nodeRights2.selectAddonLastViewNodeAfterGroupIDMember(
                                _viewInBase, targ, _dLink, _m_l_DisplayableLinks);

            if (dLink_2 != null) {

                          // for testing and tracking
                          //System.out.println("In ThreeNodesRightsFactoryImpl."+
                            //"endsPathsFromGroupIDMemberWithAclPrivilegeAlias(): ");
                          //System.out.println("--- nodeRights2.getSelectAddonLastViewNodeAfterGroupIDMember() is true, "+
                            //"'nodeRights2' has the class "+nodeRights2.getClass());
                          //System.out.println("--- and it founds the dLink: "+dLink_2);

                linkRights.addDisplayableLink(
                            dLink_2, _m_l_DisplayableLinks);
            } // if (dLink_2 != null

        } // if node.getSelectAddonLastViewNodeAfterGroupIDMember()

    } // for iti on Targets

    // second iteration ITI on sources
    for (Iterator iti = _viewInBase.getEorM_Sources().values().iterator() ;
      iti.hasNext() ; ) {
        so = (ImmutableSource) iti.next();
        if (so == null || !(so instanceof GroupID)) continue;
        gp = (GroupID)so;;

        // searching for
        // XX/YY/account or group/member/group
        if (Arrays.asList( gp.getEorL_Members()).contains(gm)) {
            dLink_2 = DisplayableLinkImpl.valueOf(
                        _dLink,
                        gp,
                        new AccessControlLinkImpl());

            // detects the IS_MEMBER and INDIRECT_MEMBER relations
            dLink_2 = ago_.
              addOwnerContainRightsToDisplayableLink(dLink_2);

            linkRights = interpreter_.getLinkRightsInstances().get(
                              gp.getName());
            linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
        } // if (Arrays.asList(

    } // for iti on Sources

    // searching for
    // XX/YY/groupidmember/alias/reference
    if (gm.getEorACS().getManageInternalHardAlias() ||
            _viewInBase.getConnectedToSeveralACS()) {

        l_temporaryLinks2 = compound_.selectDirectReferencesInViewForAlias(_viewInBase, gm);
        if (l_temporaryLinks2 != null) {
            for (Object ob : l_temporaryLinks2) {
                dLink_2 = linkUtil_.mergeLinks(_dLink, (DisplayableLinkImpl)ob);

                linkRights = interpreter_.getLinkRightsInstances().get(
                              ((DisplayableLinkImpl)ob).getEorSecondEnd().getName());
                linkRights.addDisplayableLink(dLink_2, _m_l_DisplayableLinks);
            } // for ob
        } // if (l_temporaryLinks2 != null)
    } // if (gm.getManageInternalHardAlias()

    gm = null;
    targ = null;
    gp = null;
    so = null;
    dLink_2 = null;
    acLink = null;
    pairKey = null;
    linkRights = null;
    nodeRights = null;
    nodeRights2 = null;
    l_temporaryLinks2 = null;
    //l_members = null;
    
    return _m_l_DisplayableLinks;
  } // endsPathsFromGroupIDMemberWithAclPrivilegeAlias()

  /**
   * Finalizes the instance.
   * Called by RightsMediatorImpl.finalizeForProcess().
   */
  public void finalizeForProcess() {
    interpreter_ = null;
    linkUtil_ = null;
    //baseUtil_ = null;
    compound_ = null;
    acl_ = null;
    pri_ = null;
    ago_ = null;
    l_targets_.clear();
    l_targets_ = null;
  } // finalizeForProcess()

} // ThreeNodesRightsFactoryImpl

