//Title:        Access Road
//Version:      0.7.0
//Copyright:    2002, 2004, 2007, 2008, 2009, 2010 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWork;

import ARoad0.gBaseInterface.*;
import ARoad0.gWorkInterface.*;

import ARoad0.Pattern.Singleton;
import ARoad0.ErrB.*;
import ARoad0.ErrW.*;
import ARoad0.gBase.*;
import ARoad0.gDMak.ActionOpenArchitecture;

import java.util.Map;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.util.Collection;

/**
 * This class is the "facade" (GoF pattern) for all the operations
 * about views that are called by Gui1, Gui2 and gDMak objects.
 */
public class ViewInBase_Facade implements Singleton {

  public static final String            NOTHAN_VIEW_INDICATOR = "<NMNL>";
  public static final String            VIEW_FRAME_INDICATOR = " in the set ";
    
  // the only instance of the class, following singleton pattern :
  private static ViewInBase_Facade      instance__ = new ViewInBase_Facade();
  private static final int              INITIAL_CAPACITY = 15; // default value for map: 101

  /**
   * The RightsMediator for the sketch view. Setup by ActionOpenSketcher.openSketcher()
   * and used by RightsFactory_Facade.detectSketchRights() that is a static method.
   * This mediator is finalized by ViewModel.removeViewRightsMediator() that
   * is called by an InternalFrame.openSketcher() inner class on the closing action.
   */
  public static RightsMediatorImpl      SKETCH_MEDIATOR = null;
  /**
   * A static list reserved to populateSketchView().
   */
  public static List<BaseObject>        L_SK_RES = null;
  /**
   * A static list reserved to populateSketchView().
   */
  public static List<ImmutableSource>   L_SK_EP = null;

  static {
      L_SK_RES = new ArrayList(INITIAL_CAPACITY);
      L_SK_EP = new ArrayList(INITIAL_CAPACITY);
  }

  /**
  * only one private constructor
  */
  private ViewInBase_Facade() {
  }

  /**
  * Only method to get an instance.
  */
  public static synchronized ViewInBase_Facade getInstance(){
    return instance__;
  } // getInstance()


  /**
  * @param _name of a view
  * @return boolean true if _name is known as a view name. The view may be
  * opened or closed.
  */
  public static boolean knowsView(ImmutableName _name){
    return BaseManagerImpl.getInstance().knowsView(_name);
  }

  /**
   * getElementsOfOneView(
          ImmutableName _name, boolean _acsNames) throws BaseError
   *
   *
  * @param _name ImmutableName of a view
  * @return boolean true if _name is known as a view name and the view is open.
  */
  public static boolean isOpenView(ImmutableName _name){
    return BaseManagerImpl.getInstance().isOpenView(_name);
  }

  /**
  * Get the elements of a closed or opened ViewInBase.
  * Fire no event. A closed view is not opened by this method.
  * @param _name of an opened or closed ViewInBase.
  * @param _acsNames false to get the view elements names, true to get the names of
  *     the ACS which own the view elements
  * @exception BaseError if the view is not known, 
  *   the open view has not the same name,
  *   or ClassNotFoundException, IOException.
  * @return the element names or the ACS names; if for ACS, it may contains a null item.
  */
  public static Set<ImmutableName> getElementsOfOneView(
          ImmutableName _name, boolean _acsNames) throws BaseError {
    return BaseManagerImpl.getInstance().getElementsOfOneView(_name, _acsNames);
  }

  /**
   * The 'Architecture' view is not returned in the list, to forbid its
   * direct opening by the user.
  * @return unmodifiable List of the closed view names. May be empty but not null.
  */
  public static List getClosedViews() {
    ArrayList list = new ArrayList(BaseManagerImpl.getInstance().getL_ViewNames() );
    // remove the open view names in list
    for(Iterator ite =
      BaseManagerImpl.getInstance().getEorM_OpenViews().keySet().iterator() ;
      ite.hasNext() ; ) {
          list.remove(list.indexOf(ite.next() ) );
    }
    list.remove(ActionOpenArchitecture.ARCHI_VIEW_NAME);
    if (list.isEmpty()) {
        list = null;
        return Collections.EMPTY_LIST;
    }
    else return Collections.unmodifiableList(list);
  }

  /**
   * Check if the component name before the last component equals '<NMNL>'.
   * @param _name of a ViewInBase
   * @return true if it is NoThan view name, even if the view is unknown
   */
  public static boolean isNoThanViewName(ImmutableName _name) {
      if (_name == null || _name.isEmpty() || _name.noOfComponents() < 2) return false;
      else return _name.component(_name.noOfComponents() - 2).equals(NOTHAN_VIEW_INDICATOR);
  } // isNoThanViewName()
  
  /**
   * Do not return the Architecture view and the Sketch view.
   * Call isNoThanViewName() to identity the full views.
  * @return copy of the map of the non-null open Full views. Never null.
  */
  public static Map getOpenFullViews() {
      Map map = new HashMap(BaseManagerImpl.getInstance().getEorM_OpenViews());
      map.remove(ActionOpenArchitecture.ARCHI_VIEW_NAME);
      Map.Entry entry = null;
      for (Iterator ite = map.entrySet().iterator(); ite.hasNext(); ) {
          entry = (Map.Entry)ite.next();
          if (isNoThanViewName((ImmutableName)entry.getKey())) ite.remove(); 
      } // for ite
      entry = null;
      return map;
  }

  /**
   * Do not return the Architecture view and the Sketch view.
   * Call isNoThanViewName() to identity the NoThan views.
   * @return Map of the non-null open NoThan views. Never null.
  */
  public static Map getOpenNoThanViews() {
      Map map = new HashMap(BaseManagerImpl.getInstance().getEorM_OpenViews());
      map.remove(ActionOpenArchitecture.ARCHI_VIEW_NAME);
      Map.Entry entry = null;
      for (Iterator ite = map.entrySet().iterator(); ite.hasNext(); ) {
          entry = (Map.Entry)ite.next();
          if (!isNoThanViewName((ImmutableName)entry.getKey())) ite.remove(); 
      } // for ite
      entry = null;
      return map;
  }

  /**
   * @return unmodifiable Map of the Full and NoThan open views, but
   * without the Architecture view and the Sketch view.
  */
  public static Map getAllStandardOpenViews() {
      Map map = new HashMap(BaseManagerImpl.getInstance().getEorM_OpenViews());
      map.remove(ActionOpenArchitecture.ARCHI_VIEW_NAME);
      return map;
  }
  
  /**
  * @return the 'Architecture' view if open, or null.
  */
  public static ViewInBase getArchitectureView() {
    return (ViewInBase) BaseManagerImpl.getInstance().getEorM_OpenViews().get(
            ActionOpenArchitecture.ARCHI_VIEW_NAME);
  }
  
  /**
   * No use.
  *  Saves all the registered views, even if they are empty.
  *  All open objects (views), but not the closed ones,
  *  are saved in some files (one per open object) in the default directory.
  *  The sketch view is never registered, then not processed there.
  * @exception BaseError if an object is not open, objectName.lastComponent() is null,
  *   FileNotFoundException or IOException.
  */
  public static void saveAllViews() throws BaseError {
    BaseManagerImpl.getInstance().saveAllViews();
  }

  /**
  * Saves the registered view on a file.
  * The sketch view is never registered, then not processed there.
  * @param _name ImmutableName of the view to save
  */
  public static void saveView(ImmutableName _name) {
    // get the view
    ViewInBase view = (ViewInBase) BaseManagerImpl.getInstance().getEorM_OpenViews().get(_name);
    try {
      BaseManagerImpl.getInstance().saveOpenView(view);
    } catch (BaseError b) {
    System.out.println("in gWork.ViewInBase_Facade.saveView(),"+
      " there is a problem: " + b.toString());
    }

    return;
  }

  /**
   * Creates a new classical view for BaseObjects, then save it.
   * Called by ActionOpenArchitecture.openArchitecture().
   * @param _name for the view to create.
   * No test of null nor empty.
   * @param _m_BaseObjects is the list of the BaseObjects in the view. May be null.
   * @exception CreateError or BaseError if _name is already known in BaseManagerImpl,
   * or if _l_BaseObject contains a null, no BaseObject, or empty-name object.
   * @return the new ViewInBase
   */
  public static synchronized ViewInBase createView(ImmutableName _name,
      Map _m_BaseObjects) throws CreateError, BaseError {
        // variables
        ViewInBase view;

        // throw BaseError
        view = (ViewInBase) new ViewInBaseImpl(_name, _m_BaseObjects, true);
        
        // save the view with the initial objects
        BaseManagerImpl.getInstance().saveOpenView(view);

        return view;
  } // createView()

  /**
   * Gets the BaseObjects view; open the registered view if it is closed;
   * creates and saves it if necessary. If the view exists, it may be a subclass of
   * ViewInBase, but the created view is strictly a ViewInBaseImpl instance.
  *  The sketch view is never registered, then not processed there.
   * @param _name ImmutableName for the view to create or open.
   * No test of null nor empty.
   * @param _l_BaseObjects is the list of the BaseObjects in the view.
   *    May be null.
   * @exception DialogError to display to the user.
   * @return ViewInBase
   */
  public static synchronized ViewInBase openOrCreateView(ImmutableName _name,
      List _l_BaseObjects) throws DialogError {
        // variables
        ViewInBase view;
        Map m_BaseObjects = null;
        BaseObject obj;

        try {
          // create the view
          if (!BaseManagerImpl.getInstance().knowsView(_name) ) {
            if (_l_BaseObjects != null) {
                // create m_BaseObjects
                m_BaseObjects = new TreeMap();
                for (Iterator ite = _l_BaseObjects.iterator() ; ite.hasNext() ; ) {
                  obj = (BaseObject)ite.next();
                  m_BaseObjects.put(obj.getName(), obj);
                } // for
            } // if
           
            view = createView(_name, m_BaseObjects);

          } // if (!BaseManagerImpl...knowsView(_name)
          else {
          // open the view or get it
            if (BaseManagerImpl.getInstance().isOpenView(_name) ) {
              view = (ViewInBase) BaseManagerImpl.getInstance().
                  getEorM_OpenViews().get(_name);
              // throw DialogError
            } else view = (ViewInBase) BaseManagerImpl.getInstance().openView(_name);
          } // else
        } catch (BaseError b) {
            System.out.println("in ViewInBase_Facade.openOrCreateView(): "+b);
            b.printStackTrace();
            throw new DialogError(b.toString());
        } catch (CreateError b) {
            System.out.println("in ViewInBase_Facade.openOrCreateView(): "+b);
            b.printStackTrace();
            throw new DialogError(b.toString());
        }

        view = null;
        m_BaseObjects = null;
        obj = null;
        return view;
  } // openOrCreateViewInBase()
  
  /**
   * Creates a new view for EligibleParties and Resources, then registers and saves it.
   * @param _name for the view to create.
   * No test of null nor empty.
   * @param _l_BaseObject is the list of the BaseObjects in the view. May be null.
   * @exception BaseError or CreateError if _name is already known in BaseManagerImpl,
   * or if _l_BaseObject contains a null, no BaseObject, or empty-name object,
   * or if it contains an object that is not an eligible party nor a resource.
   * @return the new EPRViewInBase
   */
  public static synchronized EPRViewInBase createEPRView(ImmutableName _name,
      Set _l_BaseObject) throws CreateError, BaseError {
        // variables
        Map m_BaseObjects = null;
        BaseObject obj;
        EPRViewInBase view;

        if (_l_BaseObject != null) {
            m_BaseObjects = new TreeMap();
            // creates m_BaseObjects
            for (Iterator ite = _l_BaseObject.iterator() ; ite.hasNext() ; ) {
              obj = (BaseObject)ite.next();
              m_BaseObjects.put(obj.getName(), obj);
            } // for           
        } // if

        // throw BaseError
        view = (EPRViewInBase) new EPRViewInBaseImpl(
          _name, m_BaseObjects, true);
        // saves the view with the initial objects
        BaseManagerImpl.getInstance().saveOpenView(view);

        m_BaseObjects = null;
        obj = null;
        return view;
  } // createEPRView()
  
  /**
   * COMMENTS TO DO
   * Creates a new view for EligibleParties and Resources, then registers and saves it.
   * @param _name for the view to create.
   * No test of null nor empty.
   * @param _l_BaseObject is the list of the BaseObjects in the view. May be null.
   * @exception BaseError or CreateError if _name is already known in BaseManagerImpl,
   * or if _l_BaseObject contains a null, no BaseObject, or empty-name object,
   * or if it contains an object that is not an eligible party nor a resource.
   * @return the new EPRViewInBase
   */
  public static synchronized NoThanViewInBase createNoThanView(
          ImmutableName _name, Set _l_BaseObject) throws BaseError, CreateError {
        // variables
        Map m_BaseObjects = null;
        BaseObject obj;
        NoThanViewInBase view;

        if (_l_BaseObject != null) {
            m_BaseObjects = new TreeMap();
            // create m_BaseObjects
            for (Iterator ite = _l_BaseObject.iterator() ; ite.hasNext() ; ) {
              obj = (BaseObject)ite.next();
              m_BaseObjects.put(obj.getName(), obj);
            } // for           
        } // if

        // throw BaseError
        view = new NoThanViewInBaseImpl(_name, m_BaseObjects, true);
        // save the view with the initial objects
        BaseManagerImpl.getInstance().saveOpenView(view);

        m_BaseObjects = null;
        obj = null;
        return view;
  } // createNoThanView()
  
  /**
   * Gets the open EP-to-Resource view; opens the registered view if it is closed;
   * creates it, if it is necessary.
   * _EPList and _ResourceList may contain redondant elements.
   *  The sketch view is never registered, then not processed there.
   * @param _name ImmutableName for the view to create or open.
   * No test of null nor empty.
   * @param _EPList List of the eligible parties in the view.
   * may be null.
   * @param _ResourceList List of the resources in the view.
   * may be null.
   * @exception DialogError to display to the user.
   * @return EPRViewInBase
   */
  public static synchronized EPRViewInBase openOrCreateEPRView(ImmutableName _name,
      List _EPList, List _ResourceList) throws DialogError {
        // variables
        EPRViewInBase view;

            // for testing and tracking
            //System.out.println("ViewInBase_Facade.openOrCreateEPRView() for "+_name);
        
        try {
          // create the view
          if (!BaseManagerImpl.getInstance().knowsView(_name) ) {
            // create l_BaseObjects without duplicate elements
            Set l_BaseObjects = new HashSet(25);
            l_BaseObjects.addAll(_EPList);
            l_BaseObjects.addAll(_ResourceList);
            if (l_BaseObjects.isEmpty()) l_BaseObjects = null;
            
            view = createEPRView(_name, l_BaseObjects);
            
            l_BaseObjects = null;
          } // if (!BaseManagerImpl...knowsView(_name)
          else {
          // open the view or get it
            if (BaseManagerImpl.getInstance().isOpenView(_name) ) {
              view = (EPRViewInBase) BaseManagerImpl.getInstance().
                  getEorM_OpenViews().get(_name);
              // throw DialogError
            } else view = (EPRViewInBase) BaseManagerImpl.getInstance().openView(_name);
          } // else
        } catch (BaseError b) {
            b.printStackTrace();
            System.out.println("in ViewInBase_Facade.openOrCreateEPRView(): "+b);
            throw new DialogError(b.toString());
        } catch (CreateError b) {
            b.printStackTrace();
            System.out.println("in ViewInBase_Facade.openOrCreateEPRView(): "+b);
            throw new DialogError(b.toString());
        }
        
        return view;
  } // openOrCreateEPRViewInBase()

  /**
   * Gets the NoThan view; opens the registered view if it is closed;
   * creates it, if it is necessary.
   * The sketch view is never registered, then not processed there.
   * @param _name ImmutableName for the view to create or open.
   * No test of null nor empty.
   * @param _EPList List of the eligible parties in the view.
   * may be null.
   * @exception DialogError to display to the user.
   * @return the view
   */
  public static synchronized NoThanViewInBase openOrCreateNoThanView(
          ImmutableName _name, List _EPList) throws DialogError {
        // variables
        NoThanViewInBase view;

                    // for testing and tracking
                    //System.out.println("ViewInBase_Facade.openOrCreateNoThanView() for "+_name);
        
        try {
          // create the view
          if (!BaseManagerImpl.getInstance().knowsView(_name) ) {
            // create l_BaseObjects without duplicate elements
            Set l_BaseObjects = new HashSet(25);
            l_BaseObjects.addAll(_EPList);
            if (l_BaseObjects.isEmpty()) l_BaseObjects = null;
            
            view = createNoThanView(_name, l_BaseObjects);
            
            l_BaseObjects = null;
          } // if (!BaseManagerImpl...knowsView(_name)
          else {
          // open the view or get it
            if (BaseManagerImpl.getInstance().isOpenView(_name) ) {
              view = (NoThanViewInBase) BaseManagerImpl.getInstance().
                  getEorM_OpenViews().get(_name);
              // throw DialogError
            } else view = (NoThanViewInBase) BaseManagerImpl.getInstance().openView(_name);
          } // else
        } catch (BaseError b) {
            System.out.println("in ViewInBase_Facade.openOrCreateNoThanView(): "+b);

            b.printStackTrace();
            throw new DialogError(b.toString());
        } catch (CreateError b) {
            System.out.println("in ViewInBase_Facade.openOrCreateNoThanView(): "+b);
            b.printStackTrace();
            throw new DialogError(b.toString());
        }
        
        return view;
  } // openOrCreateNoThanViewInBase()
  
  /**
   * Creates a new sketch view without registering nor saving it in the base.
   * The sketch view does not listen the view elements or the base changes.
   * @param _name for the view to create. No test of null nor empty.
   * @exception BaseError or CreateError if _l_BaseObject contains a null,
   * no BaseObject, or empty-name object,
   * or if it contains an object that is not an eligible party nor a resource.
   * @return the new SketchViewInBase
   */
  public static synchronized SketchViewInBase createSketchView(ImmutableName _name)
      throws BaseError, CreateError {

        // throws BaseError
        SketchViewInBase sketch = new SketchViewInBaseImpl(_name, null);
        return sketch;
  } // createSketchView()
  
  /**
   * Detects the objects to put in a sketch view, around the center,
   * and generates the DisplayableLinks which are applicable.
   * The sketch view does not listen the view elements or the base changes.
   * Called by ActionOpenSketcher.openSelection() for every change to do in the sketcher,
   * so it is designed to be fast. Works in the main thread.
   * Calls RightsFactory_Facade.detectSketchRights().
   * @param _view is the empty sketch view to populate. Never null.
   * @param _center is the center of the sketch view. Never null.
   * @param _type is the _center type, which is "actor", "resource", "ec_resource"
   *    "groupid", "userid", "virtualfolder"
   * @return the map of the DisplayableLinkImpls for this view after the adding of
   *    the relevant elements in _view. Null if the view size is 22 or more.
   */
  public static synchronized Map populateSketchView(SketchViewInBase _view,
      BaseObject _center, String _type) {

    Map                         result = null;
    ACS                         acs;
    Actor                       act;
    ImmutableGroupID[]          l_groups;
    ImmutableGroupIDMember[]    l_groupmembers;
    Collection                  l_virtualmembers;
    Resource[]                  l_res;
    String                      type = _type;
    LinkRights                  linkRights = null;

    L_SK_RES.clear();
    L_SK_EP.clear();

    try {
        L_SK_RES.add(_center);
        _view.setEorL_BaseObjectsForSketcher(L_SK_RES);

        // for the typed privileges of _center
        SKETCH_MEDIATOR.updateNonCoreAlgorithmsForViewNodes(true);
        linkRights = (LinkRights)SKETCH_MEDIATOR.
                                        getLinkRightsInstances().get(_center.getName());

    } catch (BaseError b) {
          throw new InternalError("In ViewInBase_Facade.populateSketchView(): "+
            b.toString());
    }

    // TreeManager does not identify actors under the resources tree in the explorer
    if (_center instanceof Actor) type = "actor";

    if (type.equals("actor") ) {
        // BRIDGE processing
        // Actors as Bridge targets for _center
        if (((Actor)_center).getEorACSBridgeTarget() != null) 
            L_SK_EP.add(
                    ((Actor)_center).getEorACSBridgeTarget());
        // Actors as Bridge-access sources for _center 
        if ( ((Actor)_center).getEorACS().getEorL_OwnActorBridgeTargets().contains(_center)) {
            for (Iterator ite = BaseManagerImpl.getInstance().
                    getEorM_OpenACS().values().iterator(); ite.hasNext(); ) {

                        acs = (ACS) ite.next();
                        if (acs == ((Actor)_center).getEorACS()) continue;
                        if (!acs.getContainsBridgeSource()) continue;
                        for (Iterator ita = acs.getEorL_OwnActorBridgeSources().iterator(); ita.hasNext(); ) {
                            act = (Actor) ita.next();
                            if (act.getEorACSBridgeTarget() == _center)
                                L_SK_EP.add(act);
                        } // for ita

            } // for ite
        } // if
        // UG processing
        // UserIDs as U-access sources for _center
        if ( ((Resource)_center).getEorUserID() != null ) 
            L_SK_EP.add(((Resource)_center).getEorUserID() );
        // GroupIDs as G-access sources for _center
        if ( ((Resource)_center).getEorGroupID() != null )
            L_SK_EP.add(((Resource)_center).getEorGroupID() );

        // current UserIDs as U-access targets for _center
        if ( ((Actor)_center).getEorCurrentUserID() != null ) 
            L_SK_EP.add(((Actor)_center).getEorCurrentUserID() );
        // GroupIDs as G-access targets for _center
        if ( ((Actor)_center).getEorCurrentGroupID() != null )
            L_SK_EP.add(((Actor)_center).getEorCurrentGroupID() );

        // SECONDARY GROUPS processing
        // Groups as secondary groups access targets for _center
        if (((Actor)_center).getEorL_SecondaryGroup().length != 0) {
            l_groups = ((Actor)_center).getEorL_SecondaryGroup();
            for (int i = 0; i < l_groups.length; i++) {
                L_SK_EP.add(l_groups[i]);
            } // for i
        } // if
        // ACL processing
        // Resources as ACL-access targets for _center


                    // for testing and tracking
                //System.out.println("In ViewInBase_Facade.populateSketchView(), for "+_center.getName()+
                  //      " the ACL-connected resources are:");
                //for (Iterator ita = ((ImmutableEligibleParty) _center).
                  //      getEorL_AclConnectedResources().iterator() ; ita.hasNext() ; ) {
                  //System.out.println( ita.next().toString() );
                //}             

        for (Iterator ite = ((ImmutableEligibleParty) _center).
            getEorL_AclConnectedResources().iterator(); ite.hasNext(); ) {
                L_SK_RES.add( (Resource) ite.next());
        } // for ite            
        // EPs as ACL-access sources for _center
        for (Object obj : ((ImmutableResource) _center).getEorL_EPthroughAclEntry()) {
            if (L_SK_EP.contains((EligibleParty)obj) ) continue;
            L_SK_EP.add( (EligibleParty) obj);
        } // for obj
    } // if "actor"
    
    // ACL processing
    else if (type.equals("resource") || type.equals("ec_resource")  ) {
        // EPs as ACL-access sources for _center
        for (Object obj : ((ImmutableResource) _center).getEorL_EPthroughAclEntry()) {
            if (L_SK_EP.contains((EligibleParty)obj) ) continue;
            L_SK_EP.add( (EligibleParty) obj);
        } // for obj
    } // if
    else if (type.equals("groupid") || type.equals("userid")  ) {
        // Resources as ACL-access targets for _center
        for (Iterator ite = ((ImmutableEligibleParty) _center).
            getEorL_AclConnectedResources().iterator(); ite.hasNext(); ) {
                L_SK_RES.add( (Resource) ite.next());
        } // for ite
        // CONTAIN processing
        if (((ImmutableGroupIDMember)_center).getEorL_GroupID().length != 0) {
            l_groups = ((ImmutableGroupIDMember)_center).getEorL_GroupID();
            for (int i = 0; i < l_groups.length; i++) {
                L_SK_EP.add( (EligibleParty) l_groups[i]);
            } // for i                
        }                
    } // if


    // UG processing
    if (type.equals("resource") || type.equals("ec_resource")) {
        // UserIDs as U-access sources and targets for _center
        L_SK_EP.add((EligibleParty) ((Resource)_center).getEorUserID() );
        // GroupIDs as G-access sources and targets for _center
        L_SK_EP.add((EligibleParty) ((Resource)_center).getEorGroupID() );
    } // if        
    else if ( type.equals("userid")  ) {
        // Primary group as access target for _center
        L_SK_EP.add( ((UserID)_center).getEorPrimGroup() );

        // Own resources of the userID are not processed, to keep a short view size.             
        //            Resources as U-access targets for _center
        //count = 0;
        //l_res = ((UserID)_center).getEorL_UserIDOwnResources();
        //for (int i = 0; i < l_res.length; i++) {
          //  _view.addEorResource( (Resource) l_res[i]);
          //  count++;
          //  if(count > 20) break;
        //} // for 

        // size test
        //if (_view.getEorM_BaseObjects().size() > 21) {
          //  acs = null;
          //  act = null;
          //  l_groups = null;
          //  l_groupmembers = null;
          //  l_res = null;        
          //  return result;
        //} // if 

        // Actors as access sources for _center
        l_res = ((UserID)_center).getEorL_UserIDCurrentActors();
        for (int i = 0; i < l_res.length; i++) {
            L_SK_EP.add( (EligibleParty) l_res[i]);
        } // for 

    } // if        
    else if ( type.equals("groupid")  ) {
        // groupIDmember as access target for _center
        l_groupmembers = ((GroupID)_center).getEorL_Members();
        for (int i = 0; i < l_groupmembers.length; i++) {
            L_SK_EP.add((EligibleParty) l_groupmembers[i]);
        } // for

        // Own resources of the groupID are not processed, to keep a short view size.            
        //        Resources as G-access targets for _center
        //l_res = ((GroupID)_center).getEorL_GroupIDOwnResources();
        //count = 0;
        //for (int i = 0; i < l_res.length; i++) {
          //  _view.addEorResource( (Resource) l_res[i]);
          //  count++;
          //  if(count > 20) break;                
        //} // for

        // size test before the last identical test at the end of the method
        if (_view.getEorM_BaseObjects().size() > 21) {
            acs = null;
            act = null;
            l_groups = null;
            l_groupmembers = null;
            l_res = null;        
            return result;
        } // if 

        // Actors as access sources for _center
        l_res = ((GroupID)_center).getEorL_GroupIDCurrentActors();
        for (int i = 0; i < l_res.length; i++) {
            L_SK_EP.add( (EligibleParty) l_res[i]);
        } // for              
        // Actors as access sources for _center as secondary group
        l_res = ((GroupID)_center).getEorL_GroupIDCurrentActorsAsSecondaryGroup();
        for (int i = 0; i < l_res.length; i++) {
            L_SK_EP.add( (EligibleParty) l_res[i]);
        } // for  

    } // if 
    //else if ( type.equals("metaright") || type.equals("acsright") ) {
        // use _view as a ViewInBase, not an EPRViewInBase!
      //  StringRight[] l_rights = ((StringRight)_center).getUpperRights();
        //for (int i = 0; i < l_rights.length; i++) {
          //  _view.addEorBaseObject( l_rights[i]); // --------- NO ---------------------------
        //} // for
        //l_rights = ((StringRight)_center).getLowerRights();
        //for (int i = 0; i < l_rights.length; i++) {
          //  _view.addEorBaseObject( l_rights[i]);
        //} // for
    //} // else

    // Privilege processing
    if (type.equals("actor")) {
        // PrivilegeSources of _center
        for (Object ob : ((Actor)_center).getEorL_TargetPrivilegeForTypes()) {

              L_SK_EP.addAll(
                        linkRights.getEorL_SourcesForPrivilegeForType(
                            (PrivilegeForType)ob,
                            ((PrivilegeForType)ob).getEorACS()) );
        } // for ob
        for (Object ob : ((Actor)_center).getEorL_TargetPrivilegeForLinks()) {
             L_SK_EP.addAll( ((PrivilegeForLinks)ob).getEorLinkedSources() );
        } // for ob

        // PrivilegeTargets of _center
        for (Object ob : ((EligibleParty)_center).getEorL_SourcePrivilegeForTypes()) {
             L_SK_RES.addAll(
                        linkRights.getEorL_TargetsForPrivilegeForType(
                            (PrivilegeForType)ob,
                            ((PrivilegeForType)ob).getEorACS()) );
        } // for ob
        for (Object ob : ((EligibleParty)_center).getEorL_SourcePrivilegeForLinks()) {
            L_SK_RES.addAll( ((PrivilegeForLinks)ob).getEorLinkedTargets() );
        } // for ob 
    }
    else if (type.equals("resource") || type.equals("ec_resource") ) {
        // PrivilegeSources of _center
        for (Object ob : ((Resource)_center).getEorL_TargetPrivilegeForTypes()) {
                L_SK_EP.addAll(
                        linkRights.getEorL_SourcesForPrivilegeForType(
                            (PrivilegeForType)ob,
                            ((PrivilegeForType)ob).getEorACS()) );
        } // for ob
        for (Object ob : ((Resource)_center).getEorL_TargetPrivilegeForLinks()) {
            L_SK_EP.addAll( ((PrivilegeForLinks)ob).getEorLinkedSources() );
        } // for ob
    } // if
    else if (type.equals("groupid") || type.equals("userid")  ) {

                        // for testing and tracking
                    //System.out.println("In ViewInBase_Facade.populateSketchView(), for "+
                      //      _center.getName()+" searching for the TPRI as source:");

        // PrivilegeTargets of _center
        for (Object ob : ((EligibleParty)_center).getEorL_SourcePrivilegeForTypes()) {

                        // for testing and tracking
                    //System.out.println("----- TPRI: "+((PrivilegeForType)ob).getNickName()+
                      //      " for linkRights "+linkRights);


                L_SK_RES.addAll(
                        linkRights.getEorL_TargetsForPrivilegeForType(
                            (PrivilegeForType)ob,
                            ((PrivilegeForType)ob).getEorACS()) );
        } // for ob
        for (Object ob : ((EligibleParty)_center).getEorL_SourcePrivilegeForLinks()) {
            L_SK_RES.addAll( ((PrivilegeForLinks)ob).getEorLinkedTargets() );
        } // for ob
    } // if

    // VirtualFolder processing and Hard Alias processing for all
    if ( type.equals("virtualfolder")  ) {
        // virtual members of _center
        l_virtualmembers = ((VirtualFolder)_center).getEorM_BaseObjects().values();
        for (Object ob : l_virtualmembers) {
            if (ob instanceof EligibleParty)
                L_SK_EP.add((EligibleParty) ob);
            else
                L_SK_RES.add((Resource) ob);
        } // for
        // PrivilegeSources of _center
        for (Object ob : ((VirtualFolder)_center).getEorL_TargetPrivilegeForTypes()) {
                L_SK_EP.addAll(
                        linkRights.getEorL_SourcesForPrivilegeForType(
                            (PrivilegeForType)ob,
                            ((PrivilegeForType)ob).getEorACS()) );
        } // for ob
        for (Object ob : ((VirtualFolder)_center).getEorL_TargetPrivilegeForLinks()) {
            L_SK_EP.addAll( ((PrivilegeForLinks)ob).getEorLinkedSources() );
        } // for ob

        // hard Alias processing
        if (((VirtualFolder)_center).getEorReferenceAsInternalHardAlias() != null)
            L_SK_RES.add((ImmutableSource)
                    ((Actor)_center).getEorReferenceAsInternalHardAlias());
        if (((VirtualFolder)_center).getEorAliasAsInternalHardReference() != null)
            L_SK_EP.add((ImmutableSource)
                    ((VirtualFolder)_center).getEorAliasAsInternalHardReference());
    } // if ( type.equals("virtualfolder")  ) {
    else if (_center instanceof ImmutableBasic) {
        // populate with the virtual folders
        for (Object obj : ((ImmutableBasic)_center).getVirtualFolders()) {
            L_SK_EP.add((ImmutableSource) obj);
        }

        // hard Alias processing
        if (_center instanceof ImmutableSource) {
            if (((ImmutableSource)_center).getEorReferenceAsInternalHardAlias() != null)
                L_SK_RES.add((ImmutableTarget)
                        ((ImmutableSource)_center).getEorReferenceAsInternalHardAlias());
            if (((ImmutableSource)_center).getEorAliasAsInternalHardReference() != null)
                L_SK_EP.add((ImmutableSource)
                        ((ImmutableSource)_center).getEorAliasAsInternalHardReference());
        }
        else if (_center instanceof ImmutableTarget) {
            if (((ImmutableTarget)_center).getEorReferenceAsInternalHardAlias() != null)
                L_SK_RES.add((ImmutableTarget)
                        ((ImmutableTarget)_center).getEorReferenceAsInternalHardAlias());
            if (((ImmutableTarget)_center).getEorAliasAsInternalHardReference() != null)
                L_SK_EP.add((ImmutableSource)
                        ((ImmutableTarget)_center).getEorAliasAsInternalHardReference());
        }
    } // else

    try { 
        L_SK_RES.addAll(L_SK_EP);

        _view.setEorL_BaseObjectsForSketcher(L_SK_RES);

    } catch (BaseError b) {
          throw new InternalError("In ViewInBase_Facade.populateSketchView(): "+
            b.toString());
    } 
    
    if (_view.getEorM_BaseObjects().size() > 21 ) {
        acs = null;
        act = null;
        l_groups = null;
        l_groupmembers = null;
        l_res = null;   
        linkRights = null;
        return result;
    } // if    

                        // for testing and tracking
                    //System.out.println("In ViewInBase_Facade.populateSketchView(), for "+
                      //      _center.getName()+" the objects view are:");
                    //for (Iterator ite = _view.getEorM_BaseObjects().keySet().iterator() ;
                      //     ite.hasNext() ; ) {
                      //System.out.println("-----"+ ite.next().toString() );
                    //}
                    
    try {
        SKETCH_MEDIATOR.updateNonCoreAlgorithmsForViewNodes(true);
        result = RightsFactory_Facade.detectSketchRights(_view, _center, true, true);
    } catch (ProcessError p) {
        throw new InternalError("in ViewInBase_Facade.populateSketchView(): "+p);
    } catch (InterruptedException i) {
        throw new InternalError("in ViewInBase_Facade.populateSketchView(): "+i);
    }

    acs = null;
    act = null;
    l_groups = null;
    l_groupmembers = null;
    l_virtualmembers = null;
    l_res = null;                      
    linkRights = null;
    return result;
  } // populateSketchView()
  
  /**
   * Closes any open view and save it.
   * @param _view ViewInBase to close
   */
  public static synchronized void saveAndCloseOpenView(ViewInBase _view) {
    try {
      BaseManagerImpl.getInstance().saveAndCloseOpenView(_view);
    } catch (BaseError b) {
        throw new InternalError(b.toString());
    }
  } // saveAndCloseOpenView()


  /**
   * Closes any open view without save it. This method is not used for
   * the automatic closings, for example while an ACS is closed and if it contains
   * some nodes of the view.
   * @param _view ViewInBase to close
   * @exception BaseError if _view is not open, ViewName.lastComponent() is null,
   *  or FileNotFoundException, IOException.
   */
  public static synchronized void closeOpenView(ViewInBase _view)
    throws BaseError {
      BaseManagerImpl.getInstance().closeOpenView(_view);
  } // closeOpenView()


  /**
   * Removes a view, and before, close and save it if it is open.
   * @param _name ViewInBase to remove from the base
   * @exception BaseError if the view is unknown, ViewName.lastComponent() is null,
   *  or FileNotFoundException, IOException.
   */
  public static synchronized void removeView(ImmutableName _name)
    throws BaseError {
      BaseManagerImpl.getInstance().removeView(_name);
  }  // removeView(ImmutableName _name)


} // ViewInBase_Facade