//Title:        Access Road
//Version:      0.7.0
//Copyright:    2011 EURL ACCBEE
//Author:       Patrick Thazard
//Description:  a simulator, designer and controller for IT system access controls

/* This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package ARoad0.gWorkInterface;

import java.util.Map;
import java.util.List;
import java.util.Arrays;
import java.util.Collection;

import ARoad0.CNot.DisplayableLinkImpl;
import ARoad0.gBaseInterface.ImmutableName;
import ARoad0.gBaseInterface.BaseObject;

/**
 * This interface is an implementation of the design pattern Interpreter in the GoF collection,
 * to store the paths search algorithms of an EPR view, a NoThan view or the sketch view.
 * For the core algorithms, each view uses a dedicated instance of the 8 gWork classes
 * and 1 gBase class.
 * For the NodeRight and LinkRight classes, each view includes one instance
 * for each node of the view.
 * Each AcsAddon may provide a subclass of NodeRightsImpl or LinkRightsImpl.
 * They have to be used when a node from this AcsAddon is in the view.
 * All the instances in the gBase, gWork and AcsAddon.gWork packages that are necessary
 * in the access paths search are get through this interface.
 * @see ARoad0.gWork.RightsMediatorImpl
 */
public interface AlgorithmInterpreter {

  /**
   *
   * @return the view for which the algorithms are used
   */
  public BaseObject getSource();
  
  /**
   * The keys of the 9 core algorithm classes in gWork, to use in getCoreAlgorithm().
   */
  public final List L_CORE_ALGORITHMS_KEYS = Arrays.asList(new String[] {
      "Facade", "Acl", "Ago", "Pri", "Compound", "3nodes", "LinkUtil", "FactoryUtil",
      "BaseUtil"});

  /**
   * Gets the instance to use for this view for a given core algorithm,
   * as instance of one of the 9 classes RightsFactory_Facade, AclRightsFactoryImpl,
   * AgoRightsFactoryImpl, CompoundRightsFactoryImpl, DisplayableLinkUtilities,
   * PrivilegeRightsFactoryImpl, FactoryUtilities, ThreeNodesRightsFactoryImpl
   * and gBase.UtilityImpl.
   * Theses algorithms are also used for the sketch view.
   * @param _key belongs to L_CORE_ALGORITHMS_KEYS
   * @return the core algorithm for the view, or null
   */
  public CoreAlgorithm getCoreAlgorithm(String _key);

  /**
   * Gets the map of the NodeRights instances to use for each view node or
   * intermediate node. The term of 'intermediate' is here applicable for a node
   * that is analyzed, even if it is not detected as a node to put in an effective
   * DiaplayableLinkImpl of the view.
   * @return unmodifiable map of NodeRights instances, where the node names
   * are the keys
   */
  public Map<ImmutableName, NodeRights> getNodeRightsInstances();

  /**
   * Gets the map of the LinkRights instances to use for each view node or
   * intermediate node. The term of 'intermediate' is here applicable for a node
   * that is analyzed, even if it is not detected as a node to put in an effective
   * DiaplayableLinkImpl of the view.
   * @return unmodifiable map of LinkRights instances, where the node names
   * are the keys
   */
  public Map<ImmutableName, LinkRights> getLinkRightsInstances();
  
  /**
   * When the view nodes have changed, this method updates the NodeRights
   * and the LinkRights instances of each node.
   * When the parameter is false, the previous intermediate nodes keep their
   * NodeRights and LinkRights instances to speed up the processing.
   * @param _clearIntermediateNodes true only for the sketch view, to finalize
   * the NodeRights and linkRights instances of all the nodes that are
   * not in the view, including all the previous intermediate nodes
   */
  public void updateNonCoreAlgorithmsForViewNodes(boolean _clearIntermediateNodes);

  /**
   * Creates a new NodeRights and a new LinkRights instances
   * for a view node or an intermediate node. Updates the maps m_LinkRights_
   * and m_NodeRights_.
   * @param _node node of the view or intermediate node
   * @return a NodeRights instance
   */
  public NodeRights addInstancesForNode(BaseObject _node);

  /**
   * Removes the NodeRights and the LinkRights instances
   * for a view node or an intermediate node. Updates the maps m_LinkRights_
   * and m_NodeRights_. Finalizes the NodeRights instance.
   * @param _name name of the view node or the intermediate node
   */
  public void removeInstancesForNode(ImmutableName _name);

  /**
  * Gets all the LinkRights associated to the nodes in the DisplayableLinkImpls
  * of a collection.
  * Called by CompoundRightsFactoryImpl.detectHiddenCompoundEpRights().
  * @param _l_dLinks DisplayableLinkImpls to analyze, all with the same two ends or not.
  * Not null.
  * @return the LinkRights instances. Never null.
  **/
  public Collection<LinkRights> getAllLinkRights(
          Collection<DisplayableLinkImpl> _l_dLinks);

} // AlgorithmInterpreter
